/*
 * Convert from:
 *     mplus_f12r.bdf, mplus_f12r-jisx0201.bdf, mplus_j12r.bdf
 *     (converter found in https://github.com/uobikiemukot/yaft/tree/master/tools)
 * -------------------------------------------------------------------------------
 * Original License:
 *
 * M+ BITMAP FONTS            Copyright 2002-2005  COZ <coz@users.sourceforge.jp>
 *
 * LICENSE
 *
 * These fonts are free softwares.
 * Unlimited permission is granted to use, copy, and distribute it, with
 * or without modification, either commercially and noncommercially.
 * THESE FONTS ARE PROVIDED "AS IS" WITHOUT WARRANTY.
 * -------------------------------------------------------------------------------
 *
 * Copyright (C) 2014 haru <uobikiemukot at gmail dot com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
struct glyph_t {
    uint32_t code;
    uint8_t width;
    uint16_t bitmap[16];
};

enum {
    CELL_WIDTH = 8,
    CELL_HEIGHT = 16
};

static const struct glyph_t glyphs[] = {
    {32, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {33, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0, 0x0}},
    {34, 1, {0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {35, 1, {0x0, 0x0, 0x0, 0x0, 0x12, 0x12, 0x12, 0x7E, 0x24, 0x24, 0x7E, 0x48, 0x48, 0x48, 0x0, 0x0}},
    {36, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x3E, 0x49, 0x48, 0x38, 0xE, 0x9, 0x49, 0x3E, 0x8, 0x0, 0x0}},
    {37, 1, {0x0, 0x0, 0x0, 0x0, 0x31, 0x4A, 0x4A, 0x34, 0x8, 0x8, 0x16, 0x29, 0x29, 0x46, 0x0, 0x0}},
    {38, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x39, 0x45, 0x42, 0x46, 0x39, 0x0, 0x0}},
    {39, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x8, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x8, 0x8, 0x4, 0x0}},
    {41, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x10, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x10, 0x10, 0x20, 0x0}},
    {42, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x49, 0x2A, 0x1C, 0x2A, 0x49, 0x8, 0x0, 0x0, 0x0}},
    {43, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x7F, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0}},
    {44, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10}},
    {45, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {46, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0}},
    {47, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {48, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {49, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x18, 0x28, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {50, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0xC, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {51, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x1C, 0x2, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {52, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x14, 0x24, 0x44, 0x44, 0x7E, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {53, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x2, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {54, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {55, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x4, 0x4, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {56, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {57, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x3E, 0x2, 0x2, 0x2, 0x4, 0x38, 0x0, 0x0}},
    {58, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {59, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10, 0x0}},
    {60, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {61, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {62, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {63, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x4, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0, 0x0}},
    {64, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x4A, 0x56, 0x52, 0x52, 0x52, 0x4E, 0x20, 0x1E, 0x0, 0x0}},
    {65, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {66, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {67, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {68, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {69, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {70, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {71, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {72, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {73, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {74, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {75, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {76, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {77, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {78, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {79, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {80, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {81, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x66, 0x3C, 0x3, 0x0}},
    {82, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {83, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {84, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {85, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {86, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {87, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x5A, 0x66, 0x66, 0x42, 0x42, 0x0, 0x0}},
    {88, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {89, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {90, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {91, 1, {0x0, 0x0, 0x0, 0xE, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x0}},
    {92, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x2, 0x2, 0x0, 0x0}},
    {93, 1, {0x0, 0x0, 0x0, 0x70, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x70, 0x0}},
    {94, 1, {0x0, 0x0, 0x18, 0x24, 0x42, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {95, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x0}},
    {96, 1, {0x0, 0x20, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {97, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {98, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x0}},
    {99, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {100, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {101, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {102, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {103, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {104, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {105, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {106, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x48, 0x30}},
    {107, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {108, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {109, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x76, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {110, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {111, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {112, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {113, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2}},
    {114, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {115, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {116, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {117, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {118, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x24, 0x24, 0x24, 0x18, 0x18, 0x0, 0x0}},
    {119, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {120, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {121, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {122, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {123, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x8, 0x8, 0x10, 0x10, 0x8, 0x8, 0x10, 0x10, 0xC, 0x0}},
    {124, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {125, 1, {0x0, 0x0, 0x0, 0x30, 0x8, 0x8, 0x10, 0x10, 0x8, 0x8, 0x10, 0x10, 0x8, 0x8, 0x30, 0x0}},
    {126, 1, {0x0, 0x0, 0x0, 0x31, 0x49, 0x46, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {160, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {161, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {162, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x49, 0x48, 0x48, 0x49, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {163, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x3E, 0x61, 0x0, 0x0}},
    {164, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x3C, 0x24, 0x24, 0x3C, 0x24, 0x42, 0x0, 0x0, 0x0}},
    {165, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x22, 0x14, 0x8, 0x7F, 0x8, 0x7F, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {166, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {167, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x42, 0x42, 0x3C, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {168, 1, {0x24, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {169, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x99, 0xA5, 0xA1, 0xA1, 0xA5, 0x99, 0x42, 0x3C, 0x0, 0x0}},
    {170, 1, {0x0, 0x0, 0x1C, 0x2, 0x1E, 0x22, 0x1E, 0x0, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {171, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x12, 0x24, 0x24, 0x48, 0x24, 0x24, 0x12, 0x12, 0x0, 0x0}},
    {172, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {174, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0xB9, 0xA5, 0xA5, 0xB9, 0xA9, 0xA5, 0x42, 0x3C, 0x0, 0x0}},
    {175, 1, {0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {176, 1, {0x10, 0x28, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {177, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x7F, 0x8, 0x8, 0x8, 0x0, 0x7F, 0x0, 0x0, 0x0}},
    {178, 1, {0x0, 0x0, 0x1C, 0x22, 0x2, 0x1C, 0x20, 0x20, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {179, 1, {0x0, 0x0, 0x1C, 0x22, 0x2, 0x1C, 0x2, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {180, 1, {0x0, 0x4, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {181, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x66, 0x59, 0x40, 0x80}},
    {182, 1, {0x0, 0x0, 0x0, 0x3E, 0x7A, 0x7A, 0x7A, 0x7A, 0x3A, 0xA, 0xA, 0xA, 0xA, 0xA, 0xE, 0x0}},
    {183, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {184, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x30}},
    {185, 1, {0x0, 0x0, 0x8, 0x18, 0x28, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {186, 1, {0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {187, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x24, 0x24, 0x12, 0x24, 0x24, 0x48, 0x48, 0x0, 0x0}},
    {188, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x12, 0x16, 0x2A, 0x4E, 0x42, 0x0, 0x0}},
    {189, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x14, 0x1A, 0x22, 0x44, 0x4E, 0x0, 0x0}},
    {190, 1, {0x0, 0x0, 0x0, 0x0, 0x62, 0x12, 0x24, 0x18, 0x68, 0x12, 0x16, 0x2A, 0x4E, 0x42, 0x0, 0x0}},
    {191, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x8, 0x8, 0x30, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {192, 1, {0x30, 0xC, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {193, 1, {0xC, 0x30, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {194, 1, {0x18, 0x24, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {195, 1, {0x32, 0x4C, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {196, 1, {0x24, 0x24, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {197, 1, {0x18, 0x24, 0x18, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {198, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x28, 0x48, 0x48, 0x7F, 0x48, 0x48, 0x48, 0x48, 0x4F, 0x0, 0x0}},
    {199, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x8, 0x30}},
    {200, 1, {0x30, 0xC, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {201, 1, {0xC, 0x30, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {202, 1, {0x18, 0x24, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {203, 1, {0x24, 0x24, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {204, 1, {0x18, 0x6, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {205, 1, {0xC, 0x30, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {206, 1, {0x18, 0x24, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {207, 1, {0x24, 0x24, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {208, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0xF2, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {209, 1, {0x32, 0x4C, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {210, 1, {0x30, 0xC, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {211, 1, {0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {212, 1, {0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {213, 1, {0x32, 0x4C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {214, 1, {0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {215, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x18, 0x24, 0x42, 0x0, 0x0, 0x0, 0x0}},
    {216, 1, {0x0, 0x0, 0x0, 0x2, 0x3A, 0x44, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x22, 0x5C, 0x40, 0x0}},
    {217, 1, {0x30, 0xC, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {218, 1, {0xC, 0x30, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {219, 1, {0x18, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {220, 1, {0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {221, 1, {0xC, 0x30, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {222, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x78, 0x44, 0x42, 0x42, 0x44, 0x78, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {223, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x48, 0x58, 0x44, 0x42, 0x42, 0x52, 0x4C, 0x0, 0x0}},
    {224, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {225, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {226, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {227, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {228, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {229, 1, {0x0, 0x18, 0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {230, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x49, 0x9, 0x3F, 0x48, 0x48, 0x49, 0x3E, 0x0, 0x0}},
    {231, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x8, 0x30}},
    {232, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {233, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {234, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {235, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {236, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {237, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {238, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {239, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {240, 1, {0x0, 0x0, 0x32, 0xC, 0x14, 0x22, 0x2, 0x3E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {241, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {242, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {243, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {244, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {245, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {246, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {247, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x18, 0x0, 0x0, 0x0}},
    {248, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x3C, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x3C, 0x40, 0x0}},
    {249, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {250, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {251, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {252, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {253, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {254, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x22, 0x24, 0x28, 0x30, 0x20, 0x20, 0x20}},
    {255, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {256, 1, {0x0, 0x3C, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {257, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {258, 1, {0x42, 0x42, 0x3C, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {259, 1, {0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {260, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x4, 0x3}},
    {261, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x4, 0x3}},
    {262, 1, {0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {263, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {264, 1, {0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {265, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {266, 1, {0x10, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {267, 1, {0x0, 0x0, 0x10, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {268, 1, {0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {269, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {270, 1, {0x48, 0x30, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {271, 1, {0x24, 0x18, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {272, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0xF2, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {273, 1, {0x0, 0x0, 0x2, 0xF, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {274, 1, {0x0, 0x3C, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {275, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {276, 1, {0x42, 0x42, 0x3C, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {277, 1, {0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {278, 1, {0x10, 0x10, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {279, 1, {0x0, 0x0, 0x10, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {280, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x8, 0x6}},
    {281, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x10, 0xC}},
    {282, 1, {0x24, 0x18, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {283, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {284, 1, {0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {285, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {286, 1, {0x42, 0x42, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {287, 1, {0x0, 0x42, 0x42, 0x3C, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {288, 1, {0x10, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {289, 1, {0x0, 0x0, 0x10, 0x10, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {290, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x8, 0x30}},
    {291, 1, {0x0, 0x0, 0xC, 0x10, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {292, 1, {0x18, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {293, 1, {0x30, 0x48, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {294, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0xFF, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {295, 1, {0x0, 0x0, 0x40, 0xF0, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0, 0x0}},
    {296, 1, {0x32, 0x4C, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {297, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {298, 1, {0x0, 0x3C, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {299, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {300, 1, {0x42, 0x42, 0x3C, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {301, 1, {0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {302, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x6}},
    {303, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x6}},
    {304, 1, {0x8, 0x8, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {305, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {306, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {307, 1, {0x0, 0x0, 0x22, 0x22, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x1A, 0x2, 0x22, 0x1C}},
    {308, 1, {0xC, 0x12, 0x0, 0x0, 0x1F, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {309, 1, {0x0, 0x0, 0xC, 0x12, 0x0, 0x0, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x48, 0x30}},
    {310, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x20, 0xC0}},
    {311, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x20, 0xC0}},
    {312, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x70, 0x70, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {313, 1, {0x18, 0x60, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {314, 1, {0xC, 0x30, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {315, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x8, 0x30}},
    {316, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x30}},
    {317, 1, {0x24, 0x18, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {318, 1, {0x24, 0x18, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {319, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x44, 0x44, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {320, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x10, 0x10, 0x10, 0x14, 0x14, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0}},
    {321, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x48, 0x50, 0x60, 0xC0, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {322, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0xA, 0xC, 0x18, 0x28, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {323, 1, {0xC, 0x30, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {324, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {325, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x20, 0xC0}},
    {326, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0xC0}},
    {327, 1, {0x24, 0x18, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {328, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {329, 1, {0x0, 0x60, 0x20, 0x20, 0x40, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {330, 1, {0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x4C, 0x0, 0x0}},
    {331, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0xC}},
    {332, 1, {0x0, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {333, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {334, 1, {0x42, 0x42, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {335, 1, {0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {336, 1, {0x33, 0x44, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {337, 1, {0x0, 0x0, 0x33, 0x44, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {338, 1, {0x0, 0x0, 0x0, 0x0, 0x37, 0x48, 0x48, 0x48, 0x4E, 0x48, 0x48, 0x48, 0x48, 0x37, 0x0, 0x0}},
    {339, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x49, 0x49, 0x4F, 0x48, 0x48, 0x49, 0x36, 0x0, 0x0}},
    {340, 1, {0xC, 0x30, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {341, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {342, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x20, 0xC0}},
    {343, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x20, 0xC0}},
    {344, 1, {0x24, 0x18, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {345, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {346, 1, {0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {347, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {348, 1, {0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {349, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {350, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x8, 0x30}},
    {351, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x8, 0x30}},
    {352, 1, {0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {353, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {354, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x30}},
    {355, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x8, 0x30}},
    {356, 1, {0x24, 0x18, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {357, 1, {0x24, 0x18, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {358, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0xA, 0xC, 0x18, 0x28, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {359, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x14, 0x18, 0x30, 0x50, 0x10, 0xC, 0x0, 0x0}},
    {360, 1, {0x32, 0x4C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {361, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {362, 1, {0x0, 0x3C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {363, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {364, 1, {0x42, 0x42, 0x3C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {365, 1, {0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {366, 1, {0x18, 0x24, 0x18, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {367, 1, {0x0, 0x0, 0x18, 0x24, 0x18, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {368, 1, {0x33, 0x44, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {369, 1, {0x0, 0x0, 0x33, 0x44, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {370, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {371, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x4, 0x3}},
    {372, 1, {0x18, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x5A, 0x66, 0x66, 0x42, 0x42, 0x0, 0x0}},
    {373, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {374, 1, {0x18, 0x24, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {375, 1, {0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {376, 1, {0x24, 0x24, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {377, 1, {0xC, 0x30, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {378, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {379, 1, {0x10, 0x10, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {380, 1, {0x0, 0x0, 0x10, 0x10, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {381, 1, {0x24, 0x18, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {382, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {383, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {384, 1, {0x0, 0x0, 0x0, 0x40, 0xF0, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x0}},
    {385, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0xA2, 0xA2, 0x22, 0x3C, 0x22, 0x22, 0x22, 0x22, 0x3C, 0x0, 0x0}},
    {386, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {387, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x0}},
    {388, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x20, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x22, 0x22, 0x3C, 0x0, 0x0}},
    {389, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x20, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x3C, 0x0, 0x0}},
    {390, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x2, 0x2, 0x2, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {391, 1, {0x0, 0x0, 0x3, 0x4, 0x3C, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {392, 1, {0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x3C, 0x44, 0x40, 0x40, 0x40, 0x40, 0x44, 0x38, 0x0, 0x0}},
    {393, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0xF2, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {394, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0xA4, 0xA2, 0x22, 0x22, 0x22, 0x22, 0x22, 0x24, 0x38, 0x0, 0x0}},
    {395, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x3E, 0x42, 0x42, 0x42, 0x42, 0x3E, 0x0, 0x0}},
    {396, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {397, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x4, 0x4, 0x38}},
    {398, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x3E, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {399, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {400, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x38, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {401, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x10, 0x10, 0x10, 0x1E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x60, 0x0}},
    {402, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x60, 0x0}},
    {403, 1, {0x0, 0x0, 0x0, 0x3, 0x3C, 0x44, 0x44, 0x40, 0x40, 0x5C, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {404, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {405, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x72, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x44, 0x0, 0x0}},
    {406, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {407, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {408, 1, {0x0, 0x0, 0x0, 0x0, 0x46, 0x4A, 0x50, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {409, 1, {0x0, 0x0, 0x0, 0x30, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {410, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {411, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x28, 0x10, 0x30, 0x48, 0x18, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {412, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x0, 0x0}},
    {413, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x32, 0x32, 0x2A, 0x2A, 0x26, 0x26, 0x22, 0x22, 0xC0, 0x0}},
    {414, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {415, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {416, 1, {0x0, 0x0, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {417, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {418, 1, {0x0, 0x0, 0x0, 0x0, 0x36, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x32, 0x2, 0x2}},
    {419, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x32, 0x2, 0x2}},
    {420, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x51, 0x51, 0x11, 0x1E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {421, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x5C, 0x64, 0x44, 0x44, 0x44, 0x44, 0x64, 0x58, 0x40, 0x40}},
    {422, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x20, 0x3C, 0x22, 0x22, 0x3C, 0x28, 0x24, 0x22, 0x21, 0x0, 0x0}},
    {423, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0xC, 0x30, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {424, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0xC, 0x30, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {425, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x20, 0x10, 0x8, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {426, 1, {0x0, 0x0, 0x0, 0x20, 0x50, 0x50, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0}},
    {427, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x4, 0x18}},
    {428, 1, {0x0, 0x0, 0x0, 0x0, 0x3F, 0x48, 0x48, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {429, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {430, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x6, 0x0}},
    {431, 1, {0x0, 0x0, 0x1, 0x1, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {432, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {433, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {434, 1, {0x0, 0x0, 0x0, 0x0, 0x4C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x48, 0x30, 0x0, 0x0}},
    {435, 1, {0x0, 0x0, 0x0, 0x0, 0x46, 0x45, 0x44, 0x28, 0x28, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {436, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x44, 0x44, 0x44, 0x44, 0x44, 0x24, 0x1C, 0x4, 0x4, 0x38}},
    {437, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x3C, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {438, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x4, 0x8, 0x3C, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {439, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x2, 0x46, 0x3C, 0x0, 0x0}},
    {440, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x10, 0x8, 0x38, 0x40, 0x40, 0x40, 0x62, 0x3C, 0x0, 0x0}},
    {441, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x20, 0x10, 0x8, 0x38, 0x40, 0x40, 0x40, 0x42, 0x3C}},
    {442, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x4, 0x8, 0x1C, 0x2, 0x2, 0x3C, 0x40, 0x42, 0x3C}},
    {443, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x4, 0x8, 0x7E, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {444, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x20, 0x20, 0x3C, 0x2, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {445, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x20, 0x3C, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {446, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x18, 0x4, 0x4, 0x44, 0x38, 0x0, 0x0}},
    {447, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x40}},
    {448, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {449, 1, {0x0, 0x0, 0x0, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x0, 0x0}},
    {450, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {451, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0, 0x0}},
    {452, 1, {0x5, 0x2, 0x0, 0x0, 0x67, 0x51, 0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x54, 0x67, 0x0, 0x0}},
    {453, 1, {0x0, 0x0, 0x5, 0x2, 0x60, 0x50, 0x57, 0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x67, 0x0, 0x0}},
    {454, 1, {0x0, 0x0, 0x5, 0x12, 0x10, 0x10, 0x37, 0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x37, 0x0, 0x0}},
    {455, 1, {0x0, 0x0, 0x0, 0x0, 0x47, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x49, 0x49, 0x76, 0x0, 0x0}},
    {456, 1, {0x0, 0x0, 0x0, 0x1, 0x41, 0x40, 0x43, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x79, 0x2, 0xC}},
    {457, 1, {0x0, 0x0, 0x0, 0x1, 0x61, 0x20, 0x23, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0xF9, 0x2, 0xC}},
    {458, 1, {0x0, 0x0, 0x0, 0x0, 0x4B, 0x49, 0x49, 0x69, 0x69, 0x59, 0x59, 0x4D, 0x4D, 0x4A, 0x0, 0x0}},
    {459, 1, {0x0, 0x0, 0x0, 0x1, 0x49, 0x48, 0x6B, 0x69, 0x69, 0x59, 0x59, 0x59, 0x49, 0x49, 0x12, 0xC}},
    {460, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x0, 0x53, 0x69, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x12, 0xC}},
    {461, 1, {0x24, 0x18, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {462, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {463, 1, {0x24, 0x18, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {464, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {465, 1, {0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {466, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {467, 1, {0x24, 0x18, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {468, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {469, 1, {0x3C, 0x0, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {470, 1, {0x0, 0x3C, 0x0, 0x24, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {471, 1, {0xC, 0x30, 0x0, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {472, 1, {0xC, 0x30, 0x0, 0x24, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {473, 1, {0x24, 0x18, 0x0, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {474, 1, {0x24, 0x18, 0x0, 0x24, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {475, 1, {0x30, 0xC, 0x0, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {476, 1, {0x30, 0xC, 0x0, 0x24, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {477, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {478, 1, {0x3C, 0x0, 0x24, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {479, 1, {0x0, 0x3C, 0x0, 0x24, 0x24, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {480, 1, {0x3C, 0x0, 0x18, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {481, 1, {0x0, 0x3C, 0x0, 0x10, 0x10, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {482, 1, {0x0, 0x3C, 0x0, 0x0, 0x1F, 0x28, 0x48, 0x48, 0x7F, 0x48, 0x48, 0x48, 0x48, 0x4F, 0x0, 0x0}},
    {483, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x3E, 0x49, 0x9, 0x3F, 0x48, 0x48, 0x49, 0x3E, 0x0, 0x0}},
    {484, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x4F, 0x42, 0x3E, 0x0, 0x0}},
    {485, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x4F, 0x42, 0x3C}},
    {486, 1, {0x24, 0x18, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {487, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {488, 1, {0x24, 0x18, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {489, 1, {0x48, 0x30, 0x0, 0x0, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {490, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {491, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {492, 1, {0x0, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {493, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {494, 1, {0x24, 0x18, 0x0, 0x0, 0x7E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x2, 0x46, 0x3C, 0x0, 0x0}},
    {495, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x3E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x2, 0x42, 0x3C}},
    {496, 1, {0x0, 0x0, 0x12, 0xC, 0x0, 0x0, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x48, 0x30}},
    {497, 1, {0x0, 0x0, 0x0, 0x0, 0x67, 0x51, 0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x54, 0x67, 0x0, 0x0}},
    {498, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x50, 0x57, 0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x67, 0x0, 0x0}},
    {499, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x37, 0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x37, 0x0, 0x0}},
    {500, 1, {0xC, 0x30, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {501, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {502, 1, {0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x48, 0x4A, 0x7A, 0x4A, 0x4A, 0x4A, 0x4A, 0x44, 0x0, 0x0}},
    {503, 1, {0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x40, 0x40, 0x0}},
    {504, 1, {0x0, 0x30, 0xC, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {505, 1, {0x0, 0x0, 0x0, 0x30, 0xC, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {506, 1, {0xC, 0x30, 0x18, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {507, 1, {0xC, 0x30, 0x0, 0x18, 0x24, 0x18, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {508, 1, {0x6, 0x18, 0x0, 0x0, 0x1F, 0x28, 0x48, 0x48, 0x7F, 0x48, 0x48, 0x48, 0x48, 0x4F, 0x0, 0x0}},
    {509, 1, {0x0, 0x0, 0x6, 0x18, 0x0, 0x0, 0x3E, 0x49, 0x9, 0x3F, 0x48, 0x48, 0x49, 0x3E, 0x0, 0x0}},
    {510, 1, {0xC, 0x30, 0x0, 0x2, 0x3A, 0x44, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x22, 0x5C, 0x40, 0x0}},
    {511, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x2, 0x3C, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x3C, 0x40, 0x0}},
    {512, 1, {0xCC, 0x22, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {513, 1, {0x0, 0x0, 0xCC, 0x22, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {514, 1, {0x3C, 0x42, 0x42, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {515, 1, {0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {516, 1, {0xCC, 0x22, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {517, 1, {0x0, 0x0, 0xCC, 0x22, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {518, 1, {0x3C, 0x42, 0x42, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {519, 1, {0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {520, 1, {0xCC, 0x22, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {521, 1, {0x0, 0x0, 0xCC, 0x22, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {522, 1, {0x3C, 0x42, 0x42, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {523, 1, {0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {524, 1, {0xCC, 0x22, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {525, 1, {0x0, 0x0, 0xCC, 0x22, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {526, 1, {0x3C, 0x42, 0x42, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {527, 1, {0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {528, 1, {0xCC, 0x22, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {529, 1, {0x0, 0x0, 0xCC, 0x22, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {530, 1, {0x3C, 0x42, 0x42, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {531, 1, {0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {532, 1, {0xCC, 0x22, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {533, 1, {0x0, 0x0, 0xCC, 0x22, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {534, 1, {0x3C, 0x42, 0x42, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {535, 1, {0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {536, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x8, 0x10}},
    {537, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x8, 0x10}},
    {538, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2, 0x4}},
    {539, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x20, 0x40}},
    {540, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0xC, 0x34, 0x2, 0x2, 0x2, 0x3C, 0x0, 0x0}},
    {541, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0xC, 0x32, 0x2, 0xC, 0x30, 0x0, 0x0}},
    {542, 1, {0x0, 0x24, 0x18, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {543, 1, {0x0, 0x24, 0x18, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {544, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {546, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x24, 0x42, 0x42, 0x24, 0x3C, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {547, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x22, 0x24, 0x18, 0x24, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {548, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x2, 0x4}},
    {549, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x2, 0x4}},
    {550, 1, {0x0, 0x0, 0x8, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {551, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {552, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x8, 0x30}},
    {553, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x8, 0x30}},
    {554, 1, {0x3C, 0x0, 0x24, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {555, 1, {0x0, 0x0, 0x3C, 0x0, 0x24, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {556, 1, {0x3C, 0x0, 0x32, 0x4C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {557, 1, {0x0, 0x3C, 0x0, 0x32, 0x4C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {558, 1, {0x0, 0x0, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {559, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {560, 1, {0x3C, 0x0, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {561, 1, {0x0, 0x0, 0x3C, 0x0, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {562, 1, {0x0, 0x0, 0x3E, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {563, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {592, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x7C, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {593, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x46, 0x4A, 0x32, 0x0, 0x0}},
    {594, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x62, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x0}},
    {595, 1, {0x0, 0x0, 0x0, 0x38, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x0}},
    {596, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {597, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x4C, 0x52, 0x3C, 0x40, 0x0}},
    {598, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x34, 0x4C, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x3, 0x0}},
    {599, 1, {0x0, 0x0, 0x0, 0x3, 0x4, 0x4, 0x34, 0x4C, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {600, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {601, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {602, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x39, 0x46, 0x4, 0x4, 0x7C, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {603, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {604, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3C, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {605, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x39, 0x46, 0x4, 0x38, 0x4, 0x4, 0x44, 0x38, 0x0, 0x0}},
    {606, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x5C, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {607, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x4, 0x4, 0x4, 0x1F, 0x4, 0x4, 0x4, 0x48, 0x30}},
    {608, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x34, 0x4C, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x44, 0x38}},
    {609, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x42, 0x3C}},
    {610, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x3E, 0x0, 0x0}},
    {611, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x18}},
    {612, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x18, 0x0, 0x0}},
    {613, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2, 0x2}},
    {614, 1, {0x0, 0x0, 0x0, 0x38, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {615, 1, {0x0, 0x0, 0x0, 0x38, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x4C, 0x0, 0x0}},
    {616, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x18, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {617, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {618, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {619, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {620, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x38, 0x48, 0x3E, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {621, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x6, 0x0}},
    {622, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x10, 0x1F, 0x11, 0x12, 0x14, 0x16, 0x11, 0x11, 0x7D, 0x9, 0x6}},
    {623, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x0, 0x0}},
    {624, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x1, 0x1}},
    {625, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x76, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x41, 0x6, 0x0}},
    {626, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x16, 0x19, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x60, 0x0}},
    {627, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x58, 0x64, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x3, 0x0}},
    {628, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x62, 0x52, 0x4A, 0x46, 0x42, 0x42, 0x0, 0x0}},
    {629, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {630, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x37, 0x48, 0x48, 0x4E, 0x48, 0x48, 0x48, 0x37, 0x0, 0x0}},
    {631, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {632, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8}},
    {633, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {634, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {635, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x4C, 0x34, 0x3, 0x0}},
    {636, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {637, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0}},
    {638, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {639, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {640, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x7C, 0x44, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {641, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x44, 0x7C, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {642, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x7C, 0x40, 0x30}},
    {643, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x60, 0x0}},
    {644, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x60, 0x0}},
    {645, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0}},
    {646, 1, {0x0, 0x0, 0x0, 0x6, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x38, 0x4E, 0x30, 0x0}},
    {647, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {648, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xE, 0x0}},
    {649, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x7F, 0x22, 0x22, 0x26, 0x1A, 0x0, 0x0}},
    {650, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {651, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x42, 0x42, 0x42, 0x42, 0x44, 0x48, 0x30, 0x0, 0x0}},
    {652, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x24, 0x24, 0x24, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {653, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x41, 0x0, 0x0}},
    {654, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x40, 0x40, 0x58, 0x64, 0x42, 0x42, 0x42, 0x42, 0x42}},
    {655, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {656, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x7C, 0x4, 0x3}},
    {657, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x26, 0x49, 0x7E, 0x10, 0x0}},
    {658, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x2, 0x42, 0x3C}},
    {659, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x32, 0x4A, 0x3C}},
    {660, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x4, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {661, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x20, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {662, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x4, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {663, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C}},
    {664, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x5A, 0x5A, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {665, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {666, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x3A, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {667, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3B, 0x44, 0x40, 0x40, 0x4C, 0x44, 0x44, 0x3C, 0x0, 0x0}},
    {668, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {669, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3E, 0x48, 0x30}},
    {670, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x22, 0x12, 0xA, 0x6, 0xA, 0x12, 0x22, 0x2, 0x2}},
    {671, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {672, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x34, 0x4C, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x4}},
    {673, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x4, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {674, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x20, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {675, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x3F, 0x51, 0x52, 0x52, 0x54, 0x54, 0x58, 0x3F, 0x0, 0x0}},
    {676, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x3F, 0x51, 0x52, 0x54, 0x56, 0x51, 0x51, 0x31, 0x9, 0x6}},
    {677, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x3F, 0x51, 0x52, 0x54, 0x58, 0x5A, 0x55, 0x3E, 0x4, 0x0}},
    {678, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x76, 0x28, 0x28, 0x24, 0x22, 0x21, 0x21, 0x1E, 0x0, 0x0}},
    {679, 1, {0x0, 0x0, 0x0, 0x3, 0x24, 0x24, 0x7C, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x1C, 0x4, 0x18}},
    {680, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x76, 0x29, 0x28, 0x28, 0x28, 0x2A, 0x2D, 0x1E, 0x4, 0x0}},
    {681, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x40, 0x6E, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x2, 0xC}},
    {682, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x27, 0x28, 0x28, 0x26, 0x21, 0x21, 0x3E, 0x0, 0x0}},
    {683, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x44, 0x48, 0x50, 0x60, 0x7E, 0x0, 0x0}},
    {684, 1, {0x0, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x36, 0x0, 0x41, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {685, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {688, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {689, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x20, 0x2C, 0x32, 0x22, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {690, 1, {0x0, 0x0, 0x0, 0x4, 0x0, 0x4, 0x4, 0x4, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {691, 1, {0x0, 0x0, 0x0, 0x0, 0x2C, 0x32, 0x20, 0x20, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {692, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4C, 0x34, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {693, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4C, 0x34, 0x3, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {694, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x48, 0x78, 0x44, 0x44, 0x78, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {695, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x54, 0x54, 0x28, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {696, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x3C, 0x4, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {697, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {698, 1, {0x0, 0x0, 0x0, 0x0, 0x12, 0x24, 0x48, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {699, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x10, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {700, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {701, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x10, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {702, 1, {0x0, 0x0, 0x0, 0x30, 0x8, 0x4, 0x4, 0x8, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {703, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x20, 0x20, 0x10, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {704, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x1C, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {705, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x38, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {706, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {707, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x8, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {708, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x41, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {709, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x22, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {710, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {711, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {712, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {713, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {714, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {715, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {716, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {717, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0}},
    {718, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0xC, 0x0}},
    {719, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x0}},
    {720, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x10, 0x0, 0x0, 0x10, 0x38, 0x0, 0x0, 0x0}},
    {721, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {722, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x8, 0x4, 0x4, 0x8, 0x30, 0x0, 0x0, 0x0, 0x0}},
    {723, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x10, 0x20, 0x20, 0x10, 0xC, 0x0, 0x0, 0x0, 0x0}},
    {724, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0, 0x0, 0x0}},
    {725, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {726, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {727, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {728, 1, {0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {729, 1, {0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {730, 1, {0x0, 0x0, 0x18, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {731, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x6}},
    {732, 1, {0x0, 0x0, 0x0, 0x31, 0x49, 0x46, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {733, 1, {0x0, 0x0, 0x33, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {734, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x60, 0xA2, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {735, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x18, 0x24, 0x42, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {736, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x28, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {737, 1, {0x0, 0x0, 0x30, 0x10, 0x10, 0x10, 0x10, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {738, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x40, 0x38, 0x4, 0x78, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {739, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {740, 1, {0x0, 0x0, 0x3C, 0x40, 0x40, 0x38, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {741, 1, {0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {742, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {743, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {744, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x4, 0x4, 0x0, 0x0}},
    {745, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {746, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x0, 0x0}},
    {747, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x78, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {748, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x0, 0x0}},
    {749, 1, {0x0, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {750, 1, {0x0, 0x0, 0x66, 0x22, 0x22, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {884, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x18, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {885, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x18, 0x10, 0x20}},
    {890, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x18}},
    {894, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10}},
    {900, 1, {0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {901, 1, {0x10, 0x10, 0x10, 0x44, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {902, 1, {0x40, 0x40, 0x40, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {903, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {904, 1, {0x40, 0x40, 0x40, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {905, 1, {0x40, 0x40, 0x40, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {906, 1, {0x40, 0x40, 0x40, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {908, 1, {0x40, 0x40, 0x40, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {910, 1, {0x40, 0x40, 0x40, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {911, 1, {0x40, 0x40, 0x40, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0}},
    {912, 1, {0x10, 0x10, 0x10, 0x44, 0x44, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {913, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {914, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {915, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {916, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {917, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {918, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {919, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {920, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {921, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {922, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {923, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x41, 0x0, 0x0}},
    {924, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {925, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {926, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0}},
    {927, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {928, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {929, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {931, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x20, 0x10, 0x8, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {932, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {933, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {934, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x3E, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x7F, 0x0, 0x0}},
    {935, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {936, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {937, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {938, 1, {0x24, 0x24, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {939, 1, {0x24, 0x24, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {940, 1, {0x0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {941, 1, {0x0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x3C, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0}},
    {942, 1, {0x0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {943, 1, {0x0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {944, 1, {0x10, 0x10, 0x10, 0x44, 0x44, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {945, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {946, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x7C, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {947, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x31, 0x49, 0xA, 0x4, 0x8, 0x8, 0x10, 0x10, 0x0, 0x0}},
    {948, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x20, 0x20, 0x18, 0x24, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {949, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {950, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x1C, 0x8, 0x10, 0x20, 0x20, 0x20, 0x20, 0x1C, 0x2, 0x1C}},
    {951, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {952, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x42, 0x42, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {953, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {954, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x24, 0x28, 0x30, 0x30, 0x28, 0x24, 0x22, 0x0, 0x0}},
    {955, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x10, 0x10, 0x8, 0x18, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {956, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x42, 0x40, 0x40}},
    {957, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x44, 0x44, 0x48, 0x50, 0x60, 0x0, 0x0}},
    {958, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x3C, 0x40, 0x40, 0x3C, 0x40, 0x40, 0x3C, 0x2, 0x2, 0x3C}},
    {959, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {960, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {961, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {962, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x20, 0x40, 0x40, 0x40, 0x20, 0x1C, 0x2, 0x2, 0x1C}},
    {963, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x48, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {964, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {965, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {966, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x26, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8}},
    {967, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42}},
    {968, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8}},
    {969, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {970, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {971, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {972, 1, {0x0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {973, 1, {0x0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {974, 1, {0x0, 0x8, 0x8, 0x8, 0x0, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {976, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x42, 0x44, 0x78, 0x44, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {977, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x12, 0xF, 0x22, 0x62, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {978, 1, {0x0, 0x0, 0x0, 0x0, 0x36, 0x49, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {979, 1, {0x0, 0x40, 0x40, 0x0, 0x36, 0x49, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {980, 1, {0x0, 0x36, 0x36, 0x0, 0x36, 0x49, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {981, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8}},
    {982, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {983, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x28, 0x38, 0x28, 0x28, 0x25, 0x43, 0x2, 0xC}},
    {984, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x18, 0x3C}},
    {985, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x18, 0x18}},
    {986, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x40, 0x40, 0x40, 0x30, 0xC, 0x2, 0x2, 0x3C, 0x0, 0x0}},
    {987, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x20, 0x1C, 0x2, 0x1C, 0x0, 0x0}},
    {988, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x40, 0x40, 0x78, 0x48, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {989, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x22, 0x20, 0x20, 0x38, 0x20, 0x20, 0x20, 0x20}},
    {990, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x20, 0x20, 0x20, 0x20, 0x3E, 0x4, 0x4, 0x8, 0x8, 0x0, 0x0}},
    {991, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x7C, 0x8, 0x10, 0x0, 0x0, 0x0}},
    {992, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x6, 0xA, 0x12, 0x26, 0x4A, 0x12, 0x22, 0x4C, 0x0, 0x0}},
    {993, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x10, 0x8, 0x18, 0x2C, 0x14, 0x4, 0x4, 0x4}},
    {994, 1, {0x0, 0x0, 0x0, 0x0, 0xDB, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x1, 0x3E, 0x40, 0x0}},
    {995, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x1, 0x3E, 0x40, 0x0}},
    {996, 1, {0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {997, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1A, 0x22, 0x22, 0x22, 0x1E, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {998, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x22, 0x2, 0x62, 0x1C, 0x0, 0x0}},
    {999, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x3C, 0x52, 0x52, 0x22, 0x2, 0x2, 0x32, 0x4C, 0x0, 0x0}},
    {1000, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0xC, 0x30, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1001, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x24, 0x18, 0x20, 0x40, 0x3E, 0x0, 0x0}},
    {1002, 1, {0x0, 0x0, 0x0, 0x0, 0xC3, 0xA5, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {1003, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x5A, 0x18, 0x18, 0x18, 0x24, 0x42, 0x7E, 0x0, 0x0}},
    {1004, 1, {0x0, 0x0, 0x0, 0x2, 0x3C, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1005, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x3C, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1006, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x8, 0x3E, 0x49, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {1007, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x8, 0x9, 0x3E, 0x48, 0x8, 0x8, 0x8, 0x18, 0x0, 0x0}},
    {1008, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x62, 0x12, 0x14, 0x18, 0x10, 0x30, 0x48, 0x46, 0x0, 0x0}},
    {1009, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x3C, 0x2, 0x0}},
    {1010, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x3C, 0x2, 0x0}},
    {1011, 1, {0x0, 0x0, 0x2, 0x2, 0x0, 0x0, 0xE, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x22, 0x1C, 0x0}},
    {1012, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {1013, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x40, 0x78, 0x40, 0x40, 0x20, 0x1C, 0x0, 0x0}},
    {1014, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0x8, 0x4, 0x3C, 0x4, 0x4, 0x8, 0x70, 0x0, 0x0}},
    {1024, 1, {0x0, 0x30, 0xC, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {1025, 1, {0x24, 0x24, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {1026, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0x11, 0x16, 0x0, 0x0}},
    {1027, 1, {0xC, 0x30, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1028, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x22, 0x1C, 0x0, 0x0}},
    {1029, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1030, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {1031, 1, {0x24, 0x24, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {1032, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {1033, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x48, 0x48, 0x48, 0x4E, 0x49, 0x49, 0x49, 0x49, 0x8E, 0x0, 0x0}},
    {1034, 1, {0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x48, 0x48, 0x7E, 0x49, 0x49, 0x49, 0x49, 0x4E, 0x0, 0x0}},
    {1035, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0x11, 0x11, 0x0, 0x0}},
    {1036, 1, {0xC, 0x30, 0x0, 0x0, 0x40, 0x46, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {1037, 1, {0x0, 0x30, 0xC, 0x0, 0x42, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x42, 0x0, 0x0}},
    {1038, 1, {0x42, 0x42, 0x3C, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x10, 0x30, 0x0, 0x0}},
    {1039, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x7F, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1040, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1041, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {1042, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {1043, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1044, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x12, 0x12, 0x12, 0x22, 0x22, 0x22, 0x42, 0x42, 0xFF, 0x81, 0x81}},
    {1045, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {1046, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x2A, 0x1C, 0x1C, 0x2A, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {1047, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x3C, 0x4, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1048, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x42, 0x0, 0x0}},
    {1049, 1, {0x0, 0x24, 0x18, 0x0, 0x42, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x42, 0x0, 0x0}},
    {1050, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x46, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {1051, 1, {0x0, 0x0, 0x0, 0x0, 0x1E, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x22, 0x22, 0x42, 0x0, 0x0}},
    {1052, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1053, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1054, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1055, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1056, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1057, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1058, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1059, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x10, 0x30, 0x0, 0x0}},
    {1060, 1, {0x0, 0x0, 0x0, 0x8, 0x3E, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1061, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {1062, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7F, 0x1, 0x1}},
    {1063, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {1064, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {1065, 1, {0x0, 0x0, 0x0, 0x0, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0xFF, 0x1, 0x1}},
    {1066, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x10, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0x11, 0x1E, 0x0, 0x0}},
    {1067, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x72, 0x4A, 0x4A, 0x4A, 0x4A, 0x72, 0x0, 0x0}},
    {1068, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {1069, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x2, 0x2, 0x3E, 0x2, 0x2, 0x2, 0x44, 0x38, 0x0, 0x0}},
    {1070, 1, {0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x52, 0x72, 0x52, 0x52, 0x52, 0x52, 0x4C, 0x0, 0x0}},
    {1071, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x42, 0x42, 0x42, 0x3E, 0x12, 0x22, 0x22, 0x42, 0x42, 0x0, 0x0}},
    {1072, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1073, 1, {0x0, 0x0, 0x2, 0x1C, 0x20, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1074, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {1075, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1076, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x12, 0x22, 0x22, 0x42, 0x42, 0x42, 0x7F, 0x41, 0x0}},
    {1077, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {1078, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x1C, 0x1C, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {1079, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3C, 0x4, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1080, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x0, 0x0}},
    {1081, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x0, 0x0}},
    {1082, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {1083, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x12, 0x12, 0x12, 0x12, 0x22, 0x22, 0x42, 0x0, 0x0}},
    {1084, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1085, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1086, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1087, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1088, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40}},
    {1089, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {1090, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1091, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x10, 0x10, 0x20, 0x60}},
    {1092, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8, 0x0}},
    {1093, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {1094, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7F, 0x1, 0x1}},
    {1095, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {1096, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {1097, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0xFF, 0x1, 0x1}},
    {1098, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0x1E, 0x0, 0x0}},
    {1099, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x72, 0x4A, 0x4A, 0x4A, 0x72, 0x0, 0x0}},
    {1100, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {1101, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x2, 0x3E, 0x2, 0x2, 0x44, 0x38, 0x0, 0x0}},
    {1102, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x72, 0x52, 0x52, 0x52, 0x4C, 0x0, 0x0}},
    {1103, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x42, 0x42, 0x42, 0x3E, 0x12, 0x22, 0x42, 0x0, 0x0}},
    {1104, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {1105, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {1106, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x78, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x22, 0x22, 0xC, 0x0}},
    {1107, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1108, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x40, 0x7C, 0x40, 0x40, 0x22, 0x1C, 0x0, 0x0}},
    {1109, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1110, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {1111, 1, {0x0, 0x0, 0x0, 0x24, 0x24, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {1112, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x48, 0x30}},
    {1113, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x48, 0x48, 0x4E, 0x49, 0x49, 0x49, 0x8E, 0x0, 0x0}},
    {1114, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x48, 0x7E, 0x49, 0x49, 0x49, 0x4E, 0x0, 0x0}},
    {1115, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x78, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {1116, 1, {0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x46, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {1117, 1, {0x0, 0x0, 0x30, 0xC, 0x0, 0x0, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x0, 0x0}},
    {1118, 1, {0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x10, 0x10, 0x20, 0x60}},
    {1119, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3E, 0x8, 0x8, 0x8}},
    {1120, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x5A, 0x24, 0x0, 0x0}},
    {1121, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {1122, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x7C, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0x11, 0x1E, 0x0, 0x0}},
    {1123, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x78, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x22, 0x3C, 0x0, 0x0}},
    {1124, 1, {0x0, 0x0, 0x0, 0x0, 0x46, 0x49, 0x50, 0x50, 0x7E, 0x50, 0x50, 0x50, 0x49, 0x46, 0x0, 0x0}},
    {1125, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x49, 0x50, 0x7E, 0x50, 0x50, 0x49, 0x46, 0x0, 0x0}},
    {1126, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x36, 0x2A, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {1127, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x22, 0x36, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {1128, 1, {0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x54, 0x54, 0x52, 0x76, 0x6A, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {1129, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x54, 0x54, 0x52, 0x76, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {1130, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x22, 0x22, 0x14, 0x8, 0x1C, 0x2A, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {1131, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x22, 0x14, 0x8, 0x3E, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {1132, 1, {0x0, 0x0, 0x0, 0x0, 0x5F, 0x51, 0x51, 0x4A, 0x74, 0x4E, 0x55, 0x55, 0x55, 0x55, 0x0, 0x0}},
    {1133, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5F, 0x51, 0x4A, 0x74, 0x4E, 0x55, 0x55, 0x55, 0x0, 0x0}},
    {1134, 1, {0x0, 0x0, 0x24, 0x18, 0x3C, 0x42, 0x2, 0x2, 0x3C, 0x4, 0x2, 0x2, 0x2, 0x3C, 0x40, 0x40}},
    {1135, 1, {0x0, 0x0, 0x0, 0x24, 0x18, 0x0, 0x3C, 0x42, 0x2, 0x3C, 0x4, 0x2, 0x2, 0x3C, 0x40, 0x40}},
    {1136, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1137, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x8}},
    {1138, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1139, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1140, 1, {0x0, 0x0, 0x0, 0x0, 0x43, 0x44, 0x44, 0x44, 0x28, 0x28, 0x28, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {1141, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x48, 0x48, 0x28, 0x28, 0x28, 0x10, 0x10, 0x0, 0x0}},
    {1142, 1, {0x66, 0x11, 0x0, 0x0, 0x43, 0x44, 0x44, 0x44, 0x28, 0x28, 0x28, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {1143, 1, {0x0, 0x0, 0x66, 0x11, 0x0, 0x0, 0x46, 0x48, 0x48, 0x28, 0x28, 0x28, 0x10, 0x10, 0x0, 0x0}},
    {1144, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x50, 0x59, 0x59, 0x59, 0x55, 0x56, 0x54, 0x54, 0x24, 0x8, 0x8}},
    {1145, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x29, 0x59, 0x59, 0x55, 0x56, 0x54, 0x54, 0x24, 0x8, 0x8}},
    {1146, 1, {0x0, 0x0, 0x0, 0x8, 0x3E, 0x49, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x49, 0x3E, 0x8, 0x0}},
    {1147, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x3E, 0x49, 0x41, 0x41, 0x41, 0x41, 0x49, 0x3E, 0x8, 0x0}},
    {1148, 1, {0x3C, 0x43, 0x18, 0x4, 0x8, 0x0, 0x36, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {1149, 1, {0x0, 0x0, 0x0, 0x3C, 0x43, 0x18, 0x4, 0x8, 0x0, 0x36, 0x41, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {1150, 1, {0x0, 0x7F, 0x8, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x5A, 0x24, 0x0, 0x0}},
    {1151, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x5A, 0x24, 0x0, 0x0}},
    {1152, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x3C, 0x4, 0x4, 0x4}},
    {1153, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x3C, 0x4, 0x4, 0x4}},
    {1154, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0xC, 0x44, 0x2A, 0x11, 0x18, 0x24, 0x0, 0x0, 0x0}},
    {1162, 1, {0x0, 0x24, 0x18, 0x0, 0x42, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x43, 0x2, 0x4}},
    {1163, 1, {0x0, 0x0, 0x24, 0x18, 0x0, 0x0, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x63, 0x2, 0x4}},
    {1164, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x70, 0x20, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x26, 0x3C, 0x0, 0x0}},
    {1165, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x70, 0x20, 0x3C, 0x22, 0x22, 0x3C, 0x0, 0x0}},
    {1166, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x4A, 0x44, 0x7A, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1167, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x4A, 0x64, 0x5A, 0x40, 0x40}},
    {1168, 1, {0x0, 0x0, 0x2, 0x2, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1169, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1170, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {1171, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1F, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {1172, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x42, 0xC, 0x0}},
    {1173, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0xC, 0x0}},
    {1174, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x2A, 0x1C, 0x1C, 0x2A, 0x2A, 0x49, 0x49, 0x1, 0x1}},
    {1175, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x1C, 0x1C, 0x2A, 0x49, 0x49, 0x1, 0x1}},
    {1176, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x3C, 0x4, 0x2, 0x2, 0x42, 0x3C, 0x8, 0x30}},
    {1177, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3C, 0x4, 0x2, 0x42, 0x3C, 0x8, 0x30}},
    {1178, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x46, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x2, 0x2}},
    {1179, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x2, 0x2}},
    {1180, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x43, 0x54, 0x54, 0x78, 0x54, 0x54, 0x54, 0x42, 0x41, 0x0, 0x0}},
    {1181, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x43, 0x54, 0x54, 0x78, 0x54, 0x54, 0x42, 0x41, 0x0, 0x0}},
    {1182, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x73, 0x24, 0x28, 0x30, 0x30, 0x28, 0x24, 0x22, 0x21, 0x0, 0x0}},
    {1183, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x23, 0x74, 0x28, 0x30, 0x28, 0x24, 0x22, 0x21, 0x0, 0x0}},
    {1184, 1, {0x0, 0x0, 0x0, 0x0, 0xE0, 0x23, 0x24, 0x28, 0x30, 0x30, 0x28, 0x24, 0x22, 0x21, 0x0, 0x0}},
    {1185, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xE3, 0x24, 0x28, 0x30, 0x28, 0x24, 0x22, 0x21, 0x0, 0x0}},
    {1186, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x43, 0x1, 0x1}},
    {1187, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x43, 0x1, 0x1}},
    {1188, 1, {0x0, 0x0, 0x0, 0x0, 0x47, 0x44, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {1189, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x47, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {1190, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x48, 0x48, 0x48, 0x48, 0x4E, 0x49, 0x49, 0x49, 0x49, 0x1, 0x6}},
    {1191, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x48, 0x48, 0x48, 0x4E, 0x49, 0x49, 0x49, 0x1, 0x6}},
    {1192, 1, {0x0, 0x0, 0x0, 0x20, 0x4C, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x2C, 0x18, 0x6, 0x0}},
    {1193, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x4C, 0x52, 0x52, 0x52, 0x52, 0x2C, 0x18, 0x6, 0x0}},
    {1194, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {1195, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x10, 0xC}},
    {1196, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xC, 0x4, 0x4}},
    {1197, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xC, 0x4, 0x4}},
    {1198, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1199, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {1200, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {1201, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x14, 0x8, 0x8, 0x3E, 0x8, 0x0, 0x0}},
    {1202, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x43, 0x1, 0x1}},
    {1203, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x43, 0x1, 0x1}},
    {1204, 1, {0x0, 0x0, 0x0, 0x0, 0xFA, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3F, 0x1, 0x1}},
    {1205, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFA, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3F, 0x1, 0x1}},
    {1206, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2, 0x3, 0x1, 0x1}},
    {1207, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x3, 0x1, 0x1}},
    {1208, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x52, 0x56, 0x3A, 0x12, 0x12, 0x2, 0x0, 0x0}},
    {1209, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x52, 0x56, 0x3A, 0x12, 0x2, 0x0, 0x0}},
    {1210, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1211, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1212, 1, {0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x52, 0x3E, 0x10, 0x10, 0x12, 0x12, 0xC, 0x0, 0x0}},
    {1213, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x3E, 0x10, 0x10, 0x12, 0xC, 0x0, 0x0}},
    {1214, 1, {0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x52, 0x3E, 0x10, 0x10, 0x12, 0x12, 0xC, 0x8, 0x8}},
    {1215, 1, {0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x52, 0x3E, 0x10, 0x10, 0x12, 0x12, 0xC, 0x8, 0x8}},
    {1216, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {1217, 1, {0x41, 0x41, 0x3E, 0x0, 0x49, 0x49, 0x2A, 0x2A, 0x1C, 0x1C, 0x2A, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {1218, 1, {0x0, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x1C, 0x1C, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {1219, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x46, 0x48, 0x50, 0x60, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x2, 0xC}},
    {1220, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x48, 0x50, 0x60, 0x7C, 0x42, 0x42, 0x42, 0x2, 0xC}},
    {1221, 1, {0x0, 0x0, 0x0, 0x0, 0x1E, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x22, 0x22, 0x43, 0x2, 0x4}},
    {1222, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x12, 0x12, 0x12, 0x12, 0x22, 0x22, 0x43, 0x2, 0x4}},
    {1223, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0xC}},
    {1224, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x2, 0xC}},
    {1225, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x43, 0x2, 0x4}},
    {1226, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x43, 0x2, 0x4}},
    {1227, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2, 0x6, 0x4, 0x4}},
    {1228, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x6, 0x4, 0x4}},
    {1229, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x43, 0x2, 0x4}},
    {1230, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x43, 0x2, 0x4}},
    {1232, 1, {0x42, 0x42, 0x3C, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1233, 1, {0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1234, 1, {0x24, 0x24, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1235, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1236, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x28, 0x48, 0x48, 0x7F, 0x48, 0x48, 0x48, 0x48, 0x4F, 0x0, 0x0}},
    {1237, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x49, 0x9, 0x3F, 0x48, 0x48, 0x49, 0x3E, 0x0, 0x0}},
    {1238, 1, {0x42, 0x42, 0x3C, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {1239, 1, {0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {1240, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {1241, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1242, 1, {0x24, 0x24, 0x0, 0x0, 0x18, 0x24, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {1243, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x7E, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1244, 1, {0x24, 0x24, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x2A, 0x1C, 0x1C, 0x2A, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {1245, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x1C, 0x1C, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {1246, 1, {0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x3C, 0x4, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1247, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3C, 0x4, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1248, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x2, 0x46, 0x3C, 0x0, 0x0}},
    {1249, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x4, 0x8, 0x10, 0x1C, 0x2, 0x2, 0x2, 0x42, 0x3C}},
    {1250, 1, {0x0, 0x3C, 0x0, 0x0, 0x42, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x42, 0x0, 0x0}},
    {1251, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x0, 0x0}},
    {1252, 1, {0x24, 0x24, 0x0, 0x0, 0x42, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x42, 0x0, 0x0}},
    {1253, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x46, 0x46, 0x4A, 0x4A, 0x52, 0x52, 0x62, 0x62, 0x0, 0x0}},
    {1254, 1, {0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1255, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1256, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1257, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1258, 1, {0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1259, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1260, 1, {0x0, 0x0, 0x24, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x3E, 0x2, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1261, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x0, 0x3C, 0x42, 0x2, 0x1E, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1262, 1, {0x0, 0x3E, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x10, 0x30, 0x0, 0x0}},
    {1263, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x10, 0x10, 0x20, 0x60}},
    {1264, 1, {0x24, 0x24, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x10, 0x30, 0x0, 0x0}},
    {1265, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x10, 0x10, 0x20, 0x60}},
    {1266, 1, {0x33, 0x44, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x10, 0x30, 0x0, 0x0}},
    {1267, 1, {0x0, 0x0, 0x33, 0x44, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x10, 0x10, 0x20, 0x60}},
    {1268, 1, {0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {1269, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x2, 0x0, 0x0}},
    {1270, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x70, 0x30, 0x10}},
    {1271, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x70, 0x30, 0x10}},
    {1272, 1, {0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x72, 0x4A, 0x4A, 0x4A, 0x4A, 0x72, 0x0, 0x0}},
    {1273, 1, {0x0, 0x0, 0x24, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x72, 0x4A, 0x4A, 0x4A, 0x72, 0x0, 0x0}},
    {1280, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1281, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1282, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x45, 0x45, 0x45, 0x4D, 0x36, 0x0, 0x0}},
    {1283, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x34, 0x4C, 0x44, 0x44, 0x45, 0x4D, 0x36, 0x0, 0x0}},
    {1284, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4, 0x18, 0x5, 0x5, 0x5, 0x5, 0x2, 0x0, 0x0}},
    {1285, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4, 0x18, 0x5, 0x5, 0x2, 0x0, 0x0}},
    {1286, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4, 0x18, 0x4, 0x4, 0x4, 0x4, 0x3, 0x1, 0x1}},
    {1287, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4, 0x18, 0x4, 0x4, 0x3, 0x1, 0x1}},
    {1288, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x24, 0x24, 0x24, 0x24, 0x25, 0x25, 0x45, 0x45, 0x82, 0x0, 0x0}},
    {1289, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x24, 0x24, 0x24, 0x24, 0x45, 0x45, 0x82, 0x0, 0x0}},
    {1290, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x7C, 0x45, 0x45, 0x45, 0x45, 0x46, 0x0, 0x0}},
    {1291, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x7C, 0x44, 0x45, 0x45, 0x46, 0x0, 0x0}},
    {1292, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1293, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1294, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x9, 0x9, 0x9, 0x9, 0xE, 0x0, 0x0}},
    {1295, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x9, 0x9, 0x9, 0xE, 0x0, 0x0}},
    {1329, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x47, 0x45, 0x38, 0x0, 0x0}},
    {1330, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x7C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1331, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x44, 0x3E, 0x5, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {1332, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4, 0x4, 0x6, 0x5, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {1333, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x7C, 0x42, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1334, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x4, 0x8, 0x10, 0x7E, 0x2, 0x0}},
    {1335, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x2, 0x0}},
    {1336, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x2, 0x0}},
    {1337, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4F, 0x54, 0x54, 0x48, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1338, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x3F, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {1339, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1340, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x2, 0x0}},
    {1341, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x79, 0x49, 0x49, 0x49, 0x49, 0x49, 0x46, 0x0, 0x0}},
    {1342, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {1343, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3E, 0x2, 0x2, 0x0, 0x0}},
    {1344, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x10, 0x8, 0x10, 0x20, 0x40, 0x60, 0x18, 0x6, 0x0, 0x0}},
    {1345, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x24, 0x58, 0x48, 0x36, 0x0, 0x0}},
    {1346, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7, 0x1, 0x0}},
    {1347, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x64, 0x18, 0x24, 0x24, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {1348, 1, {0x0, 0x0, 0x0, 0x0, 0x47, 0x45, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {1349, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x2, 0x42, 0x7C, 0x42, 0x2, 0x2, 0x22, 0x1C, 0x0, 0x0}},
    {1350, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x11, 0x11, 0xE, 0x0}},
    {1351, 1, {0x0, 0x0, 0x0, 0x40, 0x3C, 0x20, 0x40, 0x40, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1352, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1353, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x4, 0x8, 0x70, 0x18, 0x6, 0x0}},
    {1354, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x49, 0x49, 0x49, 0x49, 0x9, 0x9, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {1355, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x12, 0xA, 0xC, 0xC, 0x8, 0x10, 0x7E, 0x2, 0x0}},
    {1356, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x47, 0x45, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {1357, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1358, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x44, 0x44, 0x44, 0x44, 0x44, 0x3C, 0x4, 0x7, 0x1, 0x0}},
    {1359, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1360, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1361, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x3C, 0x42, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {1362, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x58, 0x64, 0x44, 0x44, 0x43, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {1363, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x38, 0x54, 0x54, 0x54, 0x54, 0x54, 0x54, 0x38, 0x10, 0x0, 0x0}},
    {1364, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x12, 0x1C, 0x50, 0x7C, 0x14, 0x10, 0x10, 0x0, 0x0}},
    {1365, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1366, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x48, 0x48, 0x38, 0xE, 0x9, 0x9, 0x9, 0x49, 0x3E, 0x0, 0x0}},
    {1369, 1, {0x10, 0x20, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1370, 1, {0x20, 0x10, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1371, 1, {0x8, 0x18, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1372, 1, {0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1373, 1, {0x20, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1374, 1, {0x38, 0x44, 0x54, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1375, 1, {0x30, 0x50, 0x40, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1377, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x0, 0x0}},
    {1378, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x42, 0x42, 0x40, 0x7E, 0x40, 0x40, 0x40}},
    {1379, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x44, 0x44, 0x3F, 0x4, 0x4, 0x4}},
    {1380, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x58, 0x64, 0x44, 0x44, 0x44, 0x44, 0x44, 0x47, 0x4, 0x4}},
    {1381, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x7E, 0x40, 0x42, 0x42, 0x22, 0x1C, 0x0, 0x0}},
    {1382, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x4C, 0x44, 0x44, 0x44, 0x44, 0x3C, 0x4, 0x4, 0x7}},
    {1383, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x38, 0x4, 0x0}},
    {1384, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x40, 0x7E}},
    {1385, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x58, 0x64, 0x44, 0x4F, 0x54, 0x54, 0x48, 0x40, 0x40, 0x40}},
    {1386, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x3F, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {1387, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x40, 0x40}},
    {1388, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C}},
    {1389, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x79, 0x49, 0x49, 0x49, 0x49, 0x46, 0x40, 0x40, 0x40}},
    {1390, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x8, 0x4, 0x3F, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {1391, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x1E, 0x2, 0x2}},
    {1392, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1393, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x8, 0x3E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3E, 0x0, 0x0}},
    {1394, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x58, 0x64, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4, 0x7}},
    {1395, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x11, 0x10, 0x7F, 0x11, 0x11, 0x11, 0x11, 0x13, 0xD, 0x0, 0x0}},
    {1396, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0x4, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {1397, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x48, 0x30}},
    {1398, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1399, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x3F, 0x0, 0x0}},
    {1400, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {1401, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x10, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x3F, 0x0, 0x0}},
    {1402, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x37, 0x1, 0x1}},
    {1403, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x24, 0x14, 0x18, 0x18, 0x20, 0x20, 0x3F, 0x0, 0x0}},
    {1404, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x58, 0x64, 0x44, 0x44, 0x44, 0x44, 0x44, 0x47, 0x0, 0x0}},
    {1405, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {1406, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x7}},
    {1407, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4A, 0x4D, 0x49, 0x49, 0x49, 0x49, 0x59, 0x29, 0x0, 0x0}},
    {1408, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x40, 0x40}},
    {1409, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x42, 0x3C}},
    {1410, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x0, 0x0}},
    {1411, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4A, 0x4D, 0x49, 0x49, 0x49, 0x49, 0x59, 0x29, 0x8, 0x8}},
    {1412, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x12, 0x1C, 0x10, 0x3C, 0x10, 0x10, 0x18}},
    {1413, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {1414, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x48, 0x48, 0x38, 0xE, 0x9, 0x9, 0x9, 0x49, 0x3E, 0x8, 0x8}},
    {1415, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x3B, 0x0, 0x0}},
    {1417, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {1418, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x62, 0x3C, 0x0, 0x0, 0x0}},
    {1425, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x1C, 0x22, 0x0}},
    {1426, 1, {0x0, 0x0, 0x18, 0x0, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1427, 1, {0xC, 0x30, 0xC, 0x30, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1428, 1, {0x0, 0x18, 0x0, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1429, 1, {0x0, 0x2C, 0x20, 0x2C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1430, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0xC, 0x0}},
    {1431, 1, {0x8, 0x1C, 0x3E, 0x1C, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1432, 1, {0x0, 0x24, 0x52, 0x4A, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1433, 1, {0x0, 0x0, 0xC0, 0x20, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1434, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x2, 0x0}},
    {1435, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x68, 0x68, 0x8, 0x70}},
    {1436, 1, {0x0, 0xC, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1437, 1, {0x0, 0x0, 0x3, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1438, 1, {0x6, 0x8, 0x13, 0x14, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1439, 1, {0x0, 0x42, 0xA5, 0x42, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1440, 1, {0x0, 0x2, 0x5, 0x2, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1441, 1, {0x0, 0x0, 0x0, 0x50, 0x60, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1443, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x38, 0x0}},
    {1444, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x8, 0x0}},
    {1445, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x30, 0x0}},
    {1446, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x64, 0x8, 0x30}},
    {1447, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x4, 0x18}},
    {1448, 1, {0x0, 0x0, 0x30, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1449, 1, {0x0, 0x20, 0x50, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1450, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x1C, 0x8, 0x8}},
    {1451, 1, {0x0, 0x10, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1452, 1, {0x0, 0x4, 0x4, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1453, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x3, 0x0}},
    {1454, 1, {0x0, 0x48, 0xA8, 0xA8, 0x90, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1455, 1, {0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1456, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x0, 0x18, 0x0}},
    {1457, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2A, 0x0, 0x12, 0x0}},
    {1458, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x0, 0x2, 0x0}},
    {1459, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x10, 0x12, 0x0}},
    {1460, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0}},
    {1461, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x66, 0x0, 0x0}},
    {1462, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x0, 0x18}},
    {1463, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0}},
    {1464, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8}},
    {1465, 1, {0x0, 0x0, 0x30, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1467, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x8, 0x2, 0x0}},
    {1468, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1469, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x0}},
    {1470, 1, {0x0, 0x0, 0x3E, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1471, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1472, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1473, 1, {0x0, 0x0, 0x6, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1474, 1, {0x0, 0x0, 0x60, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1475, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x8, 0x0, 0x0, 0x8, 0x1C, 0x8, 0x0, 0x0}},
    {1476, 1, {0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1488, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x22, 0x12, 0x2C, 0x48, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {1489, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7F, 0x0, 0x0}},
    {1490, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x8, 0x8, 0x8, 0x4, 0xC, 0x14, 0x22, 0x42, 0x0, 0x0}},
    {1491, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {1492, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {1493, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1494, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0xC, 0xA, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {1495, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {1496, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x4A, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {1497, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1498, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2}},
    {1499, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x0}},
    {1500, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x3E, 0x2, 0x2, 0x2, 0x2, 0x4, 0x4, 0x8, 0x30, 0x0, 0x0}},
    {1501, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xFE, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3E, 0x0, 0x0}},
    {1502, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4E, 0x31, 0x21, 0x41, 0x41, 0x41, 0x41, 0x41, 0x47, 0x0, 0x0}},
    {1503, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {1504, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x0, 0x0}},
    {1505, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {1506, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0xFE, 0x0, 0x0}},
    {1507, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x72, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2}},
    {1508, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x72, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {1509, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x24, 0x28, 0x30, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20}},
    {1510, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x18, 0x8, 0x4, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {1511, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x22, 0x22, 0x24, 0x24, 0x28, 0x20, 0x20, 0x20, 0x20, 0x20}},
    {1512, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {1513, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {1514, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x62, 0x0, 0x0}},
    {1520, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x77, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x0, 0x0}},
    {1521, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x77, 0x11, 0x11, 0x11, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {1522, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x77, 0x11, 0x11, 0x11, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1523, 1, {0x0, 0x0, 0x2, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1524, 1, {0x0, 0x0, 0x12, 0x24, 0x48, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1548, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1563, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x18, 0x0, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {1567, 1, {0x0, 0x0, 0x0, 0x3E, 0x41, 0x40, 0x20, 0x10, 0x8, 0x8, 0x0, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {1569, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x3C, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {1570, 1, {0x0, 0x1, 0x3E, 0x40, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1571, 1, {0x18, 0x20, 0x18, 0x20, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1572, 1, {0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1573, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0xC, 0x10, 0xC, 0x10}},
    {1574, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x20, 0x0, 0x46, 0x48, 0x48, 0x46, 0x42, 0x3C, 0x0}},
    {1575, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1576, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x8, 0x0, 0x0}},
    {1577, 1, {0x0, 0x0, 0x0, 0x24, 0x0, 0x0, 0x18, 0x24, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1578, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1579, 1, {0x0, 0x0, 0x10, 0x0, 0x28, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1580, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {1581, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {1582, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {1583, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1584, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1585, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1586, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1587, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x48, 0x48, 0x30, 0x0}},
    {1588, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x48, 0x48, 0x30, 0x0}},
    {1589, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {1590, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {1591, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1592, 1, {0x0, 0x0, 0x0, 0x20, 0x24, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1593, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x20, 0x1C, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {1594, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x18, 0x24, 0x20, 0x1C, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {1600, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1601, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1602, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x0, 0x0, 0x1C, 0x12, 0xA, 0x26, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1603, 1, {0x0, 0x0, 0x0, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1604, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x22, 0x42, 0x42, 0x44, 0x3C, 0x0, 0x0}},
    {1605, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x6, 0x1E, 0x20, 0x20, 0x20, 0x20, 0x20}},
    {1606, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1608, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1609, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {1610, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {1611, 1, {0x0, 0xC, 0x30, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1612, 1, {0x0, 0x6, 0x6, 0x32, 0x14, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1613, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x30, 0xC, 0x30, 0x0, 0x0}},
    {1614, 1, {0x0, 0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1615, 1, {0x0, 0x0, 0x10, 0x28, 0x28, 0x1C, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1616, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30}},
    {1617, 1, {0x0, 0x0, 0x0, 0x4, 0x14, 0x58, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1618, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1619, 1, {0x0, 0x0, 0x1, 0x3E, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1620, 1, {0x18, 0x20, 0x18, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1621, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x20}},
    {1632, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x38, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1633, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {1634, 1, {0x0, 0x0, 0x0, 0x22, 0x24, 0x18, 0x10, 0x8, 0x8, 0x4, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {1635, 1, {0x0, 0x0, 0x0, 0x49, 0x52, 0x3C, 0x20, 0x10, 0x10, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {1636, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x18, 0x20, 0x40, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0}},
    {1637, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {1638, 1, {0x0, 0x0, 0x0, 0x40, 0x3C, 0x4, 0x4, 0x2, 0x2, 0x2, 0x1, 0x1, 0x0, 0x0, 0x0, 0x0}},
    {1639, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {1640, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0, 0x0, 0x0}},
    {1641, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x44, 0x4C, 0x34, 0x2, 0x2, 0x1, 0x1, 0x0, 0x0, 0x0, 0x0}},
    {1642, 1, {0x0, 0x0, 0x4, 0x4, 0x48, 0x48, 0x10, 0x10, 0x24, 0x24, 0x40, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {1643, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x38, 0x0, 0x0, 0x0}},
    {1644, 1, {0x0, 0x0, 0x0, 0x18, 0x18, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1645, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x7F, 0x1C, 0x36, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1646, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1648, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1649, 1, {0x4, 0xA, 0x1E, 0x20, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1650, 1, {0x6, 0x8, 0x26, 0x58, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1651, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x6, 0x8, 0x26, 0x58}},
    {1652, 1, {0x18, 0x20, 0x18, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1653, 1, {0x6, 0x8, 0x6, 0x8, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1654, 1, {0x0, 0x0, 0x3, 0x4, 0x3, 0x4, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1655, 1, {0x0, 0x0, 0x33, 0x34, 0x13, 0x24, 0x40, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1656, 1, {0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x3, 0x4, 0x0, 0x46, 0x48, 0x48, 0x46, 0x42, 0x3C, 0x0}},
    {1657, 1, {0x0, 0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1658, 1, {0x0, 0x0, 0x10, 0x0, 0x10, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1659, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {1660, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x8, 0x14, 0x8, 0x0, 0x0}},
    {1661, 1, {0x0, 0x0, 0x28, 0x0, 0x10, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1662, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x14, 0x0, 0x8, 0x0}},
    {1663, 1, {0x0, 0x0, 0x28, 0x0, 0x28, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1664, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x14, 0x0, 0x14, 0x0}},
    {1665, 1, {0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x20, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {1666, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x10, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {1667, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x54, 0x40, 0x22, 0x1C}},
    {1668, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x48, 0x40, 0x48, 0x22, 0x1C}},
    {1669, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {1670, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x40, 0x54, 0x40, 0x48, 0x22, 0x1C}},
    {1671, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x40, 0x54, 0x40, 0x54, 0x22, 0x1C}},
    {1672, 1, {0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1673, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x8, 0x14, 0x8, 0x0, 0x0}},
    {1674, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {1675, 1, {0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {1676, 1, {0x0, 0x0, 0x0, 0x28, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1677, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {1678, 1, {0x0, 0x0, 0x10, 0x0, 0x28, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1679, 1, {0x0, 0x0, 0x28, 0x0, 0x10, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1680, 1, {0x0, 0x0, 0x28, 0x0, 0x28, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1681, 1, {0x0, 0x0, 0x8, 0x8, 0xE, 0xE, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1682, 1, {0x0, 0x0, 0x11, 0xA, 0x4, 0x4, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1683, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x4C, 0x3A, 0x4}},
    {1684, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x32, 0x0}},
    {1685, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0x4, 0x8, 0x48, 0x35, 0x2}},
    {1686, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x12, 0x4, 0x48, 0x32, 0x0}},
    {1687, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1688, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x14, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1689, 1, {0x0, 0x0, 0x0, 0x12, 0x0, 0x12, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1690, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x48, 0x4A, 0x30, 0x0, 0x0}},
    {1691, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x50, 0x55, 0x20, 0x2, 0x0}},
    {1692, 1, {0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x50, 0x55, 0x20, 0x2, 0x0}},
    {1693, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x50, 0x55, 0x20, 0x0, 0x0}},
    {1694, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {1695, 1, {0x0, 0x0, 0x0, 0x24, 0x20, 0x2A, 0x20, 0x2C, 0x32, 0x22, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1696, 1, {0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x18, 0x24, 0x20, 0x1C, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {1697, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1698, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {1699, 1, {0x0, 0x0, 0x4, 0x0, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {1700, 1, {0x0, 0x8, 0x0, 0x12, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1701, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x24, 0x0, 0x10, 0x0}},
    {1702, 1, {0x0, 0xA, 0x0, 0xA, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1703, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x1C, 0x12, 0xA, 0x26, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1704, 1, {0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x1C, 0x12, 0xA, 0x26, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1705, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1707, 1, {0x0, 0x0, 0x0, 0x4, 0xC, 0x1A, 0x24, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1708, 1, {0x0, 0x8, 0x0, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1709, 1, {0x10, 0x0, 0x28, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1710, 1, {0x0, 0x0, 0x0, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3C, 0x0, 0x14, 0x0, 0x8, 0x0}},
    {1711, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1712, 1, {0x0, 0x8, 0x10, 0x24, 0x4C, 0x1A, 0x24, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1713, 1, {0x0, 0x50, 0x6, 0x18, 0x62, 0xC, 0x30, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1714, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {1715, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x10, 0x0, 0x10, 0x0}},
    {1716, 1, {0x40, 0x0, 0xA6, 0x18, 0x62, 0xC, 0x30, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1717, 1, {0x11, 0xA, 0x4, 0x4, 0x0, 0x2, 0x2, 0x2, 0x2, 0x22, 0x42, 0x42, 0x44, 0x3C, 0x0, 0x0}},
    {1718, 1, {0x0, 0x2, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x22, 0x42, 0x42, 0x44, 0x3C, 0x0, 0x0}},
    {1719, 1, {0x4, 0x0, 0x9, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x22, 0x42, 0x42, 0x44, 0x3C, 0x0, 0x0}},
    {1720, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x22, 0x42, 0x3C, 0x0, 0x14, 0x0, 0x8}},
    {1721, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x8}},
    {1722, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1723, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1724, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x3C, 0xA, 0x4}},
    {1725, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {1726, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2C, 0x2A, 0x1A, 0x24, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {1727, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x7E, 0x20, 0x40, 0x54, 0x40, 0x48, 0x22, 0x1C}},
    {1728, 1, {0x0, 0x18, 0x20, 0x18, 0x20, 0x0, 0x10, 0x18, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1729, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2E, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {1730, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x20, 0x0, 0x10, 0x2E, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {1731, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0x0, 0x10, 0x2E, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {1732, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0xE, 0x22, 0x52, 0x24, 0x78, 0x0}},
    {1733, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x1C, 0x48, 0x30, 0x0}},
    {1734, 1, {0x0, 0x0, 0x22, 0x14, 0x8, 0x8, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1735, 1, {0x0, 0x10, 0x28, 0x28, 0x1C, 0x10, 0x20, 0x4C, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1736, 1, {0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1737, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1738, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1739, 1, {0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1740, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {1741, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2A, 0x68, 0xA4, 0x22, 0x22, 0x1C, 0x0}},
    {1742, 1, {0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x10, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {1743, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {1744, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x52, 0x48, 0x44, 0x38, 0x0, 0x10, 0x0, 0x10}},
    {1745, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x52, 0x48, 0x44, 0x38, 0x0, 0x28, 0x0, 0x10}},
    {1746, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x10, 0x20, 0x40, 0x7F, 0x0}},
    {1747, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x40, 0x8, 0x14, 0x10, 0x20, 0x40, 0x7F, 0x0}},
    {1748, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1749, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x18, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1755, 1, {0x0, 0x0, 0x0, 0xA, 0x0, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1765, 1, {0x0, 0x6, 0x6, 0x4, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1766, 1, {0x0, 0x10, 0x20, 0x40, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1776, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x38, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1777, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {1778, 1, {0x0, 0x0, 0x0, 0x22, 0x24, 0x18, 0x10, 0x8, 0x8, 0x4, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {1779, 1, {0x0, 0x0, 0x0, 0x49, 0x52, 0x3C, 0x20, 0x10, 0x10, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {1780, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x60, 0x72, 0x2C, 0x20, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {1781, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x49, 0x55, 0x63, 0x0, 0x0, 0x0}},
    {1782, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x20, 0x20, 0x18, 0x6, 0x18, 0x60, 0x0, 0x0, 0x0, 0x0}},
    {1783, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {1784, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0, 0x0, 0x0}},
    {1785, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x44, 0x4C, 0x34, 0x2, 0x2, 0x1, 0x1, 0x0, 0x0, 0x0, 0x0}},
    {1786, 1, {0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x48, 0x4A, 0x30, 0x0, 0x0}},
    {1787, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x48, 0x4A, 0x30, 0x0, 0x0}},
    {1788, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x18, 0x24, 0x20, 0x1C, 0x20, 0x48, 0x40, 0x42, 0x3C}},
    {1789, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x3C, 0x10, 0x20, 0x0, 0x14, 0x14, 0x14, 0x14}},
    {1790, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x6, 0x3E, 0x40, 0x4A, 0x4A, 0x4A, 0x4A}},
    {1920, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {1921, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x1C, 0x20, 0x40, 0x0, 0x0}},
    {1922, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2A, 0x34, 0x40, 0x80, 0x0, 0x0}},
    {1923, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0xC, 0xC, 0x12, 0x20, 0x40, 0x0, 0x0}},
    {1924, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x54, 0x54, 0x28, 0x0, 0x0}},
    {1925, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x24, 0x18, 0x10, 0x20, 0x0, 0x0}},
    {1926, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x28, 0x30, 0x20, 0x0, 0x0}},
    {1927, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x14, 0x24, 0x0, 0x0}},
    {1928, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {1929, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x4, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {1930, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x2, 0x4, 0x2, 0x4, 0x8, 0x10, 0x0, 0x0}},
    {1931, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x4, 0x18, 0x20, 0x40, 0x0}},
    {1932, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x2C, 0x40, 0x20, 0x40, 0x0, 0x0}},
    {1933, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x12, 0x3C, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {1934, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x8, 0x30, 0x0, 0x0}},
    {1935, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x28, 0x40, 0x40, 0x38, 0x0, 0x0}},
    {1937, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0xC, 0x8, 0x10, 0x1E, 0x0, 0x0}},
    {1938, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0xC, 0xC, 0x12, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {1939, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x2A, 0x2A, 0x14, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {1940, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x16, 0x1A, 0x12, 0x20, 0x40, 0x0, 0x0}},
    {1941, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x2, 0x4, 0x2, 0x4, 0x8, 0x12, 0x0, 0x0}},
    {1942, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x8, 0x8, 0x70, 0x40, 0x40, 0x38, 0x0, 0x0}},
    {1943, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x2, 0x4, 0xC, 0x1A, 0x2A, 0x44, 0x20, 0xC0, 0x0, 0x0}},
    {1944, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0xA, 0x0, 0xC, 0x1A, 0x2A, 0x44, 0x20, 0xC0, 0x0, 0x0}},
    {1945, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x20, 0x48, 0x0, 0x0}},
    {1946, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x10, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {1947, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x2, 0x4, 0x4, 0x18, 0x20, 0x40, 0x0}},
    {1948, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x12, 0xC, 0xC, 0x12, 0x20, 0x40, 0x0, 0x0}},
    {1952, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x2C, 0x40, 0x24, 0x40, 0x0, 0x0}},
    {1953, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0xC, 0x12, 0x2C, 0x40, 0x20, 0x40, 0x0, 0x0}},
    {1954, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x14, 0x24, 0x0, 0x10}},
    {1955, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x4, 0xC, 0x14, 0x24, 0x0, 0x0}},
    {1956, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x4, 0x8, 0x10, 0x8, 0x30, 0x0, 0x0}},
    {1957, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x8, 0x14, 0x14, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {1958, 1, {0x0, 0x10, 0x20, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1959, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1960, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x40}},
    {1961, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x24, 0x48}},
    {1962, 1, {0x0, 0x60, 0x10, 0x20, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1963, 1, {0x0, 0x66, 0x11, 0x22, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1964, 1, {0x0, 0x10, 0x20, 0x40, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1965, 1, {0x22, 0x44, 0x88, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1966, 1, {0x42, 0x24, 0x18, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1967, 1, {0x0, 0x24, 0x40, 0x48, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1968, 1, {0x0, 0x30, 0x48, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {1969, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x28, 0x10, 0x20, 0x40, 0x40, 0x38, 0x0, 0x0}},
    {3585, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {3586, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x64, 0x64, 0x24, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {3587, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x52, 0x72, 0x12, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {3588, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x5A, 0x7A, 0x62, 0x42, 0x42, 0x0, 0x0}},
    {3589, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x4A, 0x42, 0x5A, 0x7A, 0x62, 0x42, 0x42, 0x0, 0x0}},
    {3590, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x52, 0x72, 0x12, 0x22, 0x22, 0x22, 0x72, 0x6C, 0x0, 0x0}},
    {3591, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0xC, 0x4, 0x4, 0x44, 0x24, 0x14, 0xC, 0x0, 0x0}},
    {3592, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x1A, 0x1A, 0xA, 0xA, 0xC, 0x0, 0x0}},
    {3593, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x4, 0x64, 0x64, 0x24, 0x2E, 0x36, 0x0, 0x0}},
    {3594, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x32, 0x34, 0x12, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {3595, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x52, 0x74, 0x12, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {3596, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x39, 0x45, 0x25, 0x45, 0x45, 0x45, 0x6F, 0x6D, 0x0, 0x0}},
    {3597, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x2A, 0x4A, 0x4A, 0x4A, 0x6E, 0x62, 0xE, 0x0}},
    {3598, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x66, 0x7A, 0x16, 0x0}},
    {3599, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x6A, 0x6E, 0x1A, 0x0}},
    {3600, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x20, 0x1C, 0x2, 0x1A, 0x1A, 0x6, 0x2, 0x2A, 0x16}},
    {3601, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x7A, 0x12, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {3602, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x29, 0x55, 0x45, 0x75, 0x75, 0x55, 0x6F, 0x4D, 0x0, 0x0}},
    {3603, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x2A, 0x4A, 0x4A, 0x4A, 0x6F, 0x6B, 0x0, 0x0}},
    {3604, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x5A, 0x5A, 0x4A, 0x72, 0x42, 0x0, 0x0}},
    {3605, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x4A, 0x42, 0x5A, 0x5A, 0x4A, 0x72, 0x42, 0x0, 0x0}},
    {3606, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x32, 0x32, 0x0, 0x0}},
    {3607, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x64, 0x6A, 0x32, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {3608, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x40, 0x78, 0x24, 0x24, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {3609, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x64, 0x64, 0x24, 0x44, 0x44, 0x44, 0x4E, 0x36, 0x0, 0x0}},
    {3610, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x32, 0x12, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {3611, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x32, 0x32, 0x12, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {3612, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x69, 0x69, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {3613, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x69, 0x69, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {3614, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x69, 0x69, 0x29, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {3615, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x69, 0x69, 0x29, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {3616, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x62, 0x62, 0x0, 0x0}},
    {3617, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x62, 0x62, 0x22, 0x22, 0x22, 0x22, 0x72, 0x6C, 0x0, 0x0}},
    {3618, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x62, 0x62, 0x42, 0x32, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {3619, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x40, 0x78, 0x4, 0x4, 0x4, 0xC, 0xC, 0x0, 0x0}},
    {3620, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x32, 0x32, 0x2, 0x0}},
    {3621, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x1A, 0x26, 0x22, 0x32, 0x32, 0x0, 0x0}},
    {3622, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x62, 0x22, 0x22, 0x22, 0x62, 0x62, 0x2, 0x0}},
    {3623, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x4, 0x4, 0x4, 0x4, 0xC, 0xC, 0x0, 0x0}},
    {3624, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x3C, 0x42, 0x5A, 0x7A, 0x62, 0x42, 0x42, 0x0, 0x0}},
    {3625, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x62, 0x62, 0x22, 0x2A, 0x2F, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {3626, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x3C, 0x42, 0x2, 0x1A, 0x26, 0x22, 0x32, 0x32, 0x0, 0x0}},
    {3627, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x66, 0x24, 0x2A, 0x32, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {3628, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x6A, 0x69, 0x29, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {3629, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x62, 0x62, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {3630, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x3C, 0x7A, 0x2, 0x62, 0x62, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {3631, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x74, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {3632, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x38, 0x0, 0x34, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {3634, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {3635, 1, {0x0, 0x0, 0x0, 0x60, 0x60, 0x0, 0x1C, 0x22, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {3647, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x7C, 0x2A, 0x2A, 0x2A, 0x3C, 0x2A, 0x2A, 0x2A, 0x7C, 0x8, 0x0}},
    {3648, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xC, 0xC, 0x0, 0x0}},
    {3649, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x1B, 0x1B, 0x0, 0x0}},
    {3650, 1, {0x0, 0x0, 0x0, 0x3E, 0x41, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x6, 0x6, 0x0, 0x0}},
    {3651, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x54, 0x34, 0x4, 0x4, 0x4, 0x4, 0x4, 0x6, 0x6, 0x0, 0x0}},
    {3652, 1, {0x0, 0x0, 0x0, 0x6C, 0x14, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x6, 0x6, 0x0, 0x0}},
    {3653, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4}},
    {3654, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x72, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x4, 0x8}},
    {3663, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x5A, 0x5A, 0x42, 0x3C, 0x0, 0x0}},
    {3664, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {3665, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x5A, 0x3A, 0x2, 0x3C, 0x0, 0x0}},
    {3666, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x4A, 0x56, 0x52, 0x5A, 0x5A, 0x42, 0x3C, 0x0, 0x0}},
    {3667, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x4A, 0x4A, 0x4A, 0x42, 0x72, 0x72, 0x0, 0x0}},
    {3668, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x3C, 0x40, 0x40, 0x4C, 0x4C, 0x48, 0x3E, 0x0, 0x0}},
    {3669, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1A, 0x16, 0x3C, 0x40, 0x40, 0x4C, 0x4C, 0x48, 0x3E, 0x0, 0x0}},
    {3670, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x2C, 0x12, 0x2, 0x2, 0x32, 0x32, 0x3C, 0x0, 0x0}},
    {3671, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2A, 0x5A, 0x4A, 0x4A, 0x4A, 0x6A, 0x6C, 0x0, 0x0}},
    {3672, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x3C, 0x40, 0x40, 0x46, 0x56, 0x52, 0x6E, 0x0, 0x0}},
    {3673, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x1A, 0x34, 0x50, 0x48, 0x48, 0x64, 0x64, 0x0, 0x0}},
    {3674, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x36, 0x3A, 0x2, 0x2, 0x2, 0x6, 0x4, 0x0, 0x0}},
    {3675, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x50, 0x58, 0x57, 0x54, 0x50, 0x60, 0x60, 0x0, 0x0}},
    {3713, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x62, 0x0, 0x0, 0x0}},
    {3714, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x4, 0x18, 0x20, 0x4C, 0x52, 0x22, 0x0, 0x0, 0x0}},
    {3716, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x32, 0x42, 0x42, 0x42, 0x32, 0x0, 0x0, 0x0}},
    {3719, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x2C, 0x14, 0x4, 0x4, 0x4, 0x4, 0x64, 0x44, 0x44, 0x38}},
    {3720, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x1A, 0x26, 0x2, 0x2, 0x2, 0x0, 0x0, 0x0}},
    {3722, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x64, 0x8, 0x10, 0x20, 0x4C, 0x72, 0x2, 0x2, 0x2}},
    {3725, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x42, 0x42, 0x32, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3732, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x52, 0x22, 0x0, 0x0, 0x0}},
    {3733, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x0, 0x0, 0x0}},
    {3734, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x32, 0x2, 0x2, 0x2}},
    {3735, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8C, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x22, 0x0, 0x0, 0x0}},
    {3737, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x52, 0x22, 0x22, 0x22, 0x22, 0x22, 0x5C, 0x0, 0x0, 0x0}},
    {3738, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x52, 0x22, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3739, 1, {0x0, 0x2, 0x2, 0x2, 0x2, 0x22, 0x52, 0x22, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3740, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x66, 0x86, 0x82, 0x82, 0x82, 0x92, 0x92, 0x6C, 0x0, 0x0, 0x0}},
    {3741, 1, {0x0, 0x2, 0x2, 0x2, 0x2, 0x62, 0x82, 0x82, 0x82, 0x82, 0x92, 0x92, 0x6C, 0x0, 0x0, 0x0}},
    {3742, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0xA2, 0x42, 0x82, 0x82, 0x92, 0x92, 0x6C, 0x0, 0x0, 0x0}},
    {3743, 1, {0x0, 0x2, 0x2, 0x2, 0x2, 0x42, 0xA2, 0x42, 0x82, 0x82, 0x92, 0x92, 0x6C, 0x0, 0x0, 0x0}},
    {3745, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x5C, 0x0, 0x0, 0x0}},
    {3746, 1, {0x0, 0x2, 0x2, 0x2, 0x2, 0x32, 0x42, 0x42, 0x22, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3747, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x2, 0x62, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3749, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x2, 0x32, 0x4A, 0x24, 0x0, 0x0, 0x0}},
    {3751, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3754, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x44, 0x46, 0x7A, 0x2, 0x32, 0x4A, 0x44, 0x0, 0x0, 0x0}},
    {3755, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x96, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x31, 0x0, 0x0, 0x0}},
    {3757, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x46, 0x3A, 0x2, 0x44, 0x38, 0x0, 0x0, 0x0}},
    {3758, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x40, 0x3C, 0x2, 0x62, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3759, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x26, 0x1A, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20}},
    {3760, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x38, 0x0, 0x34, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {3761, 1, {0x61, 0x62, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3762, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x14, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0, 0x0}},
    {3763, 1, {0x0, 0x0, 0xC0, 0xC0, 0x0, 0xC, 0x12, 0xA, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0, 0x0}},
    {3764, 1, {0x7C, 0x44, 0x28, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3765, 1, {0x7C, 0x44, 0x2E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3766, 1, {0x7C, 0x54, 0x28, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3767, 1, {0x7C, 0x54, 0x2E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3768, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8}},
    {3769, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1A, 0xC}},
    {3771, 1, {0x3E, 0x61, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3772, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C}},
    {3773, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x24, 0x18, 0x4, 0x4, 0x4, 0x4, 0x64, 0x44, 0x44, 0x38}},
    {3776, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xC, 0xC, 0x0, 0x0, 0x0}},
    {3777, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x36, 0x36, 0x0, 0x0, 0x0}},
    {3778, 1, {0x0, 0x30, 0x4C, 0x22, 0x18, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x6, 0x6, 0x0, 0x0, 0x0}},
    {3779, 1, {0x0, 0x38, 0x44, 0x54, 0x34, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x6, 0x6, 0x0, 0x0, 0x0}},
    {3780, 1, {0x0, 0x80, 0x4C, 0x34, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x6, 0x6, 0x0, 0x0, 0x0}},
    {3782, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x52, 0x52, 0x22, 0x2, 0x2, 0x2, 0x2, 0x2, 0x4, 0x8}},
    {3784, 1, {0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3785, 1, {0x32, 0x14, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3786, 1, {0x28, 0x55, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3787, 1, {0x8, 0x1C, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3788, 1, {0x2, 0x1C, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3789, 1, {0x0, 0x0, 0x8, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {3792, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {3793, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x5A, 0x3A, 0x2, 0x3C, 0x0, 0x0, 0x0}},
    {3794, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x40, 0x5C, 0x5C, 0x44, 0x38, 0x0, 0x0, 0x0}},
    {3795, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0xA, 0x44, 0x22, 0x22, 0x22, 0x22, 0x22, 0x5C, 0x0, 0x0, 0x0}},
    {3796, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x3C, 0x40, 0x40, 0x4C, 0x4C, 0x48, 0x3E, 0x0, 0x0, 0x0}},
    {3797, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x1A, 0x3C, 0x40, 0x40, 0x4C, 0x4C, 0x48, 0x3E, 0x0, 0x0, 0x0}},
    {3798, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0xA, 0x24, 0x42, 0x42, 0x42, 0x52, 0x52, 0x2C, 0x0, 0x0, 0x0}},
    {3799, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x22, 0x22, 0x22, 0x22, 0x22, 0x62, 0x2, 0x2, 0x2}},
    {3800, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2C, 0x52, 0x52, 0x42, 0x42, 0x26, 0xA, 0x14, 0x2C, 0x30}},
    {3801, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x24, 0x0, 0x0, 0x0}},
    {3804, 1, {0x0, 0x0, 0x0, 0x0, 0xB9, 0xD5, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x26, 0x0, 0x0, 0x0}},
    {3805, 1, {0x0, 0x0, 0x0, 0x0, 0x89, 0xD5, 0x55, 0x55, 0x55, 0x55, 0x55, 0x5D, 0x2E, 0x0, 0x0, 0x0}},
    {3852, 1, {0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4108, 1, {0x38, 0x44, 0x80, 0x80, 0x80, 0x46, 0x3A, 0x2, 0x2, 0x32, 0x42, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {4128, 1, {0x38, 0x44, 0x80, 0x80, 0x80, 0x66, 0x1A, 0x42, 0x7C, 0x40, 0x42, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {4162, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x11, 0x11, 0xE, 0x0, 0x0, 0x0, 0x0}},
    {4163, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x11, 0x1, 0x22, 0x1C, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0, 0x0}},
    {4170, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {4171, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {4172, 1, {0x18, 0x20, 0x18, 0x38, 0x4, 0x4, 0x38, 0x42, 0x7E, 0x2, 0x42, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {4180, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xE, 0x11, 0x2E, 0x20, 0x20, 0x11, 0xE, 0x0, 0x0, 0x0, 0x0}},
    {4256, 1, {0x0, 0x0, 0x0, 0x80, 0xFE, 0x22, 0x40, 0x80, 0x80, 0x80, 0x80, 0x82, 0x46, 0x3A, 0x0, 0x0}},
    {4257, 1, {0x0, 0x0, 0x0, 0x0, 0x68, 0x88, 0x88, 0x88, 0x98, 0x68, 0x8, 0x8, 0x8, 0xE, 0x2, 0x0}},
    {4258, 1, {0x0, 0x0, 0x0, 0x0, 0xF8, 0x88, 0x88, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x2, 0x0}},
    {4259, 1, {0x0, 0x0, 0x0, 0x0, 0x80, 0xFE, 0x12, 0x10, 0x38, 0x44, 0x82, 0x82, 0x44, 0x38, 0x0, 0x0}},
    {4260, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x44, 0x44, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {4261, 1, {0x0, 0x0, 0x0, 0x0, 0xF8, 0x88, 0x88, 0x8, 0xE, 0xA, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {4262, 1, {0x0, 0x0, 0x0, 0x0, 0x80, 0xE0, 0x20, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x22, 0x22, 0x1C, 0x0}},
    {4263, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x50, 0x90, 0x90, 0x9E, 0x92, 0x90, 0x90, 0x50, 0x30, 0x0, 0x0}},
    {4264, 1, {0x0, 0x0, 0x0, 0x40, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {4265, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x58, 0x64, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x4, 0x0, 0x0}},
    {4266, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xE0, 0xA0, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x22, 0x22, 0x1C, 0x0}},
    {4267, 1, {0x0, 0x0, 0x0, 0x80, 0xF8, 0x8, 0x8, 0x8, 0x7E, 0x8A, 0x8A, 0x88, 0x88, 0x70, 0x0, 0x0}},
    {4268, 1, {0x0, 0x0, 0x0, 0x4, 0x7C, 0x40, 0x40, 0x40, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {4269, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x8A, 0x8A, 0x74, 0x0, 0x0}},
    {4270, 1, {0x0, 0x0, 0x0, 0x80, 0xE2, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {4271, 1, {0x0, 0x0, 0x0, 0x40, 0x7A, 0x12, 0x22, 0x22, 0x26, 0x1A, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {4272, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x14, 0x1A, 0x12, 0x92, 0x90, 0x90, 0x90, 0x60, 0x0, 0x0}},
    {4273, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x48, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {4274, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x2, 0x0}},
    {4275, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x8C, 0x8A, 0x72, 0x0, 0x0}},
    {4276, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x92, 0x92, 0x92, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {4277, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x7C, 0x14, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {4278, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {4279, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x4, 0x0, 0x0}},
    {4280, 1, {0x0, 0x0, 0x0, 0x0, 0x34, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {4281, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {4282, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x22, 0x1C, 0x4, 0x8}},
    {4283, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x4A, 0x4A, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {4284, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x78, 0x44, 0x48, 0x40, 0x40}},
    {4285, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x20, 0x10, 0x8, 0x4, 0x42, 0x42, 0x3C}},
    {4286, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x44, 0x7C, 0x40, 0x40, 0x40, 0x7C, 0x4, 0x0}},
    {4287, 1, {0x0, 0x0, 0x0, 0x80, 0xFE, 0x22, 0x22, 0x14, 0x14, 0x8, 0x14, 0x14, 0x22, 0x22, 0x40, 0x80}},
    {4288, 1, {0x0, 0x0, 0x0, 0x0, 0xF0, 0x90, 0x90, 0x10, 0x10, 0x10, 0x12, 0x12, 0x12, 0xC, 0x0, 0x0}},
    {4289, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x78, 0x44, 0x46, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {4290, 1, {0x0, 0x0, 0x0, 0x40, 0x78, 0x44, 0x2, 0x2, 0x2, 0x2, 0x2, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {4291, 1, {0x0, 0x0, 0x0, 0x0, 0x34, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {4292, 1, {0x0, 0x0, 0x0, 0x0, 0x82, 0x82, 0x82, 0x82, 0x44, 0x28, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {4293, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x12, 0x12, 0x1C, 0x10, 0x7C, 0x94, 0x96, 0x90, 0x60, 0x0, 0x0}},
    {4304, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0xC, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {4305, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x30, 0x10, 0x10, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4306, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x48, 0x8, 0x8, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x3C}},
    {4307, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2C, 0x52, 0x52, 0x52, 0x52, 0x4C, 0x20, 0x1C, 0x2}},
    {4308, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x42, 0x42, 0x3C}},
    {4309, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x1C, 0x2, 0x2, 0x42, 0x42, 0x3C}},
    {4310, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x50, 0x50, 0x54, 0x38, 0x1C, 0x12, 0x12, 0x12, 0xC, 0x0, 0x0}},
    {4311, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x32, 0x0, 0x0}},
    {4312, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x0, 0x0}},
    {4313, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x2, 0x2, 0x2, 0x1C, 0x2, 0x2, 0x42, 0x42, 0x3C}},
    {4314, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2A, 0x55, 0x55, 0x55, 0x41, 0x42, 0x40, 0x3E, 0x1}},
    {4315, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x2, 0x2, 0x3E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4316, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x3C, 0x40, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4317, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x54, 0x54, 0x54, 0x54, 0x44, 0x24, 0x0, 0x0}},
    {4318, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x18, 0x4, 0x2, 0x2, 0x1C, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4319, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0xA, 0x52, 0x52, 0x22, 0x2, 0x22, 0x22, 0x1C}},
    {4320, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x38, 0x40, 0x68, 0x54, 0x54, 0x54, 0x54, 0x44, 0x24, 0x0, 0x0}},
    {4321, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x48, 0x50, 0x48, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {4322, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x3C, 0x66, 0x66, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x3C}},
    {4323, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x12, 0x12, 0x2, 0x2, 0x2, 0x22, 0x22, 0x1C}},
    {4324, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x34, 0x4A, 0x4A, 0x4A, 0x4A, 0x34, 0x2, 0x2, 0x1C}},
    {4325, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x4, 0x4, 0x4, 0x1C, 0x24, 0x4, 0x4, 0x4, 0x44, 0x38}},
    {4326, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x54, 0x54, 0x54, 0x44, 0x48, 0x40, 0x38, 0x4}},
    {4327, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x42, 0x42, 0x3C}},
    {4328, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x54, 0x54, 0x4, 0x4, 0x1C, 0x24, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {4329, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x78, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {4330, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x42, 0x42, 0x42, 0x5C, 0x42, 0x42, 0x42, 0x42, 0x3C}},
    {4331, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x2, 0x3E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4332, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x54, 0x54, 0x40, 0x40, 0x7C, 0x44, 0x44, 0x28, 0x28, 0x10, 0x10}},
    {4333, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x18, 0x28, 0x8, 0x8, 0x19, 0x26, 0x1D, 0x4, 0x4, 0x44, 0x38}},
    {4334, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x48, 0x50, 0x78, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {4335, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x20, 0x40}},
    {4336, 1, {0x0, 0x0, 0x10, 0xC, 0x2, 0x12, 0xC, 0x2, 0x12, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4337, 1, {0x0, 0x0, 0x50, 0x4C, 0x42, 0x52, 0x4C, 0x42, 0x52, 0x4C, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {4338, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x7E, 0x0, 0x0}},
    {4339, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x2, 0x4, 0x8, 0x1C, 0x2, 0x2, 0x42, 0x42, 0x3C}},
    {4340, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0xC, 0x2, 0x2, 0x2, 0x1C, 0x2, 0x2, 0x42, 0x42, 0x3C}},
    {4341, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x3E, 0x4, 0x3E, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {4342, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x10, 0x10, 0x10, 0x38, 0x54, 0x54, 0x38, 0x10, 0x10, 0x10, 0x8}},
    {4343, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x4, 0x8, 0x10, 0x20, 0x42, 0x42, 0x3C}},
    {4344, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x42, 0x42, 0x42, 0x64, 0x58, 0x42, 0x42, 0x3C}},
    {4347, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x30, 0x6, 0x6, 0x30, 0x30, 0x0, 0x0, 0x0, 0x0}},
    {4352, 2, {0x0, 0x0, 0x7F80, 0x80, 0x80, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x1800, 0x6000, 0x0, 0x0, 0x0}},
    {4353, 2, {0x0, 0x0, 0x0, 0x7B80, 0x880, 0x880, 0x880, 0x880, 0x1100, 0x1100, 0x2200, 0x4400, 0x800, 0x0, 0x0, 0x0}},
    {4354, 2, {0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x3F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4355, 2, {0x0, 0x3F80, 0x2000, 0x2000, 0x2000, 0x3F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4356, 2, {0x0, 0x3B80, 0x2200, 0x2200, 0x2200, 0x3B80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4357, 2, {0x0, 0x0, 0x3F80, 0x80, 0x3F80, 0x2000, 0x3F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4358, 2, {0x0, 0x3F80, 0x2080, 0x2080, 0x2080, 0x3F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4359, 2, {0x0, 0x2080, 0x2080, 0x3F80, 0x2080, 0x3F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4360, 2, {0x0, 0x2A80, 0x2A80, 0x3B80, 0x2A80, 0x3B80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4361, 2, {0x0, 0x400, 0x400, 0x400, 0xA00, 0x1100, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4362, 2, {0x0, 0x1100, 0x1100, 0x1100, 0x2A80, 0x4440, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4363, 2, {0x0, 0xE00, 0x1100, 0x1100, 0x1100, 0xE00, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4364, 2, {0x0, 0x3F80, 0x100, 0x200, 0xD00, 0x3080, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4365, 2, {0x0, 0x0, 0x3BC0, 0x880, 0x1500, 0x2280, 0x440, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4366, 2, {0xE00, 0x0, 0x3F80, 0x200, 0xD00, 0x3080, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4367, 2, {0x0, 0x3F80, 0x80, 0x3F80, 0x80, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4368, 2, {0x0, 0x3F80, 0x2000, 0x3F80, 0x2000, 0x3F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4369, 2, {0x0, 0x7F80, 0x1200, 0x1200, 0x1200, 0x7F80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4370, 2, {0x0, 0xE00, 0x0, 0x3F80, 0x0, 0x1F00, 0x2080, 0x2080, 0x2080, 0x1F00, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4371, 2, {0x0, 0x21FC, 0x2004, 0x2004, 0x2004, 0x3F84, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4372, 2, {0x0, 0x2040, 0x2040, 0x2040, 0x2040, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4373, 2, {0x0, 0x207C, 0x2040, 0x2040, 0x2040, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4374, 2, {0x0, 0x2044, 0x2044, 0x207C, 0x2044, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4375, 2, {0x0, 0x3E7C, 0x2004, 0x2004, 0x2004, 0x3E04, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4376, 2, {0x0, 0x0, 0x3E40, 0x240, 0x3E40, 0x2040, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4377, 2, {0x0, 0x0, 0x3E7C, 0x204, 0x3E7C, 0x2040, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4378, 2, {0x0, 0x0, 0x3E30, 0x2FC, 0x3E30, 0x2048, 0x3E30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4379, 2, {0x0, 0x0, 0x1FF8, 0x8, 0x1FF8, 0x1000, 0x1FF8, 0x3C0, 0x420, 0x420, 0x3C0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4380, 2, {0x0, 0x3E44, 0x2244, 0x227C, 0x2244, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4381, 2, {0x0, 0x1FFC, 0x1004, 0x1004, 0x1004, 0x1FFC, 0x0, 0x0, 0x0, 0x3E0, 0x410, 0x410, 0x410, 0x3E0, 0x0, 0x0}},
    {4382, 2, {0x0, 0x227C, 0x2204, 0x3E04, 0x2204, 0x3E04, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4383, 2, {0x0, 0x2240, 0x2240, 0x3E40, 0x2240, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4384, 2, {0x0, 0x227C, 0x2240, 0x3E40, 0x2240, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4385, 2, {0x0, 0x2210, 0x2210, 0x3E10, 0x2228, 0x3E44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4386, 2, {0x0, 0x489E, 0x4882, 0x7882, 0x4942, 0x7A22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4387, 2, {0x0, 0x488E, 0x4888, 0x7888, 0x4948, 0x7A2E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4388, 2, {0x0, 0x4924, 0x4924, 0x793C, 0x4AA4, 0x7C7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4389, 2, {0x0, 0x4888, 0x4888, 0x7888, 0x4954, 0x7A22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4390, 2, {0x0, 0x493E, 0x4904, 0x7908, 0x4A94, 0x7C62, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4391, 2, {0x0, 0x227C, 0x2208, 0x3E10, 0x2228, 0x3E44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4392, 2, {0x38, 0x2200, 0x227C, 0x3E10, 0x2228, 0x3E44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4393, 2, {0x0, 0x227C, 0x2240, 0x3E7C, 0x2240, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4394, 2, {0x0, 0x227C, 0x2228, 0x3E28, 0x2228, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4395, 2, {0x0, 0x808, 0xFF8, 0x808, 0xFF8, 0x1C0, 0x220, 0x1C0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4396, 2, {0x0, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x3C0, 0x420, 0x3C0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4397, 2, {0x0, 0x47C, 0x404, 0x404, 0xA04, 0x1104, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4398, 2, {0x0, 0x440, 0x440, 0x440, 0xA40, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4399, 2, {0x0, 0x47C, 0x440, 0x440, 0xA40, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4400, 2, {0x0, 0x47C, 0x404, 0x47C, 0xA40, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4401, 2, {0x0, 0x47C, 0x444, 0x444, 0xA44, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4402, 2, {0x0, 0x444, 0x444, 0x47C, 0xA44, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4403, 2, {0x0, 0x112E, 0x1122, 0x11E2, 0x2922, 0x45E2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4404, 2, {0x0, 0x888, 0x888, 0x888, 0x1554, 0x2222, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4405, 2, {0x0, 0x438, 0x444, 0x444, 0xA44, 0x1138, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4406, 2, {0x0, 0x47C, 0x408, 0x410, 0xA28, 0x1144, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4407, 2, {0x38, 0x400, 0x47C, 0x410, 0xA28, 0x1144, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4408, 2, {0x0, 0x47C, 0x404, 0x47C, 0xA04, 0x1104, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4409, 2, {0x0, 0x47C, 0x440, 0x47C, 0xA40, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4410, 2, {0x0, 0x47C, 0x428, 0x428, 0xA28, 0x117C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4411, 2, {0x0, 0x430, 0x4FC, 0x430, 0xA48, 0x1130, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4412, 2, {0x0, 0x40, 0x40, 0x180, 0x680, 0x1840, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4413, 2, {0x0, 0x120, 0x120, 0x2A0, 0x450, 0x1988, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4414, 2, {0x0, 0x80, 0x80, 0x100, 0x680, 0x1840, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4415, 2, {0x0, 0x480, 0x480, 0x540, 0xA30, 0x118C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4416, 2, {0x0, 0x200, 0x500, 0x880, 0x1040, 0x3FE0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4417, 2, {0x0, 0xE7C, 0x1104, 0x1104, 0x1104, 0xE04, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4418, 2, {0x0, 0xE3C, 0x1120, 0x1120, 0x1120, 0xE3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4419, 2, {0x0, 0xE3E, 0x1122, 0x1122, 0x1122, 0xE3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4420, 2, {0x0, 0xE24, 0x1124, 0x113C, 0x1124, 0xE3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4421, 2, {0x0, 0xE08, 0x1108, 0x1108, 0x1114, 0xE22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4422, 2, {0x0, 0x3820, 0x4450, 0x4488, 0x4504, 0x39FC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4423, 2, {0x0, 0xE38, 0x1144, 0x1144, 0x1144, 0xE38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4424, 2, {0x0, 0xE7C, 0x1108, 0x1110, 0x1128, 0xE44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4425, 2, {0x38, 0xE00, 0x117C, 0x1110, 0x1128, 0xE44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4426, 2, {0x0, 0xE3C, 0x1120, 0x113C, 0x1120, 0xE3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4427, 2, {0x0, 0xE7C, 0x1128, 0x1128, 0x1128, 0xE7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4428, 2, {0x0, 0xE00, 0x1100, 0x1100, 0x1100, 0xE00, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4429, 2, {0x0, 0x3E38, 0x444, 0x844, 0x1444, 0x2238, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4430, 2, {0x0, 0x3FE0, 0xC0, 0x380, 0xC40, 0x3020, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4431, 2, {0x0, 0x0, 0x1FF0, 0x120, 0x2A0, 0xC50, 0x3188, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4432, 2, {0x0, 0x1FF0, 0x60, 0x1C0, 0x620, 0x1810, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4433, 2, {0x0, 0x0, 0x3FE0, 0x840, 0x14A0, 0x2310, 0x18C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4434, 2, {0x1C00, 0x7C, 0x3E04, 0x87C, 0x1404, 0x2204, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4435, 2, {0x1800, 0x30, 0x3EFC, 0x830, 0x1448, 0x2230, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4436, 2, {0x780, 0x0, 0x1FC0, 0x180, 0x640, 0x3820, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4437, 2, {0xE00, 0x0, 0x3F00, 0x200, 0xD80, 0x3070, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4438, 2, {0x0, 0x3E44, 0x1444, 0x147C, 0x1444, 0x3E7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4439, 2, {0x0, 0x1FF8, 0x420, 0x1FF8, 0x3C0, 0x420, 0x3C0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4440, 2, {0x0, 0x1830, 0x7EFC, 0x1C70, 0x2288, 0x2288, 0x1C70, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4441, 2, {0x1FF0, 0x100, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {4447, 2, {0x5555, 0x8000, 0x1, 0xA238, 0x2241, 0xBE40, 0x2241, 0xA238, 0x1, 0x83E0, 0x201, 0x8380, 0x201, 0x8200, 0x1, 0xAAAA}},
    {4931, 1, {0x0, 0x0, 0x7E, 0xC3, 0xC3, 0xFF, 0xC3, 0xC3, 0x7F, 0x3, 0x3, 0x3, 0x3, 0x3, 0x0, 0x0}},
    {4961, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x3, 0x0, 0x0, 0x3, 0x3, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5024, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x24, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x24, 0x78, 0x0, 0x0}},
    {5025, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x44, 0x78, 0x50, 0x48, 0x44, 0x44, 0x43, 0x0, 0x0}},
    {5026, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x49, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5027, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x8, 0x14, 0x12, 0x21, 0x51, 0x51, 0x51, 0x55, 0x56, 0x0, 0x0}},
    {5028, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x48, 0x4B, 0x49, 0x4E, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5029, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {5030, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x20, 0x7F, 0x4, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5031, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x2, 0x72, 0x52, 0x52, 0x52, 0x62, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5032, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x20, 0x22, 0x2A, 0x3E, 0x20, 0x20, 0x20, 0x20, 0x70, 0x0, 0x0}},
    {5033, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C, 0x0, 0x0}},
    {5034, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {5035, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5036, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {5037, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x50, 0x7E, 0x52, 0x14, 0x14, 0x18, 0x0, 0x0}},
    {5038, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x22, 0x2, 0x72, 0x22, 0x3C, 0x20, 0x20, 0x20, 0x20, 0x0, 0x0}},
    {5039, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x11, 0x11, 0x1F, 0x11, 0x11, 0x55, 0x77, 0x0, 0x0}},
    {5040, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x12, 0x1E, 0x12, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0}},
    {5041, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5042, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x24, 0x24, 0x18, 0x5B, 0x59, 0x5A, 0x24, 0x0, 0x0}},
    {5043, 1, {0x0, 0x0, 0x0, 0x0, 0x52, 0x52, 0x52, 0x4A, 0x5A, 0x5A, 0x66, 0x66, 0x42, 0x42, 0x0, 0x0}},
    {5044, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0xA, 0x8, 0x38, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5045, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x52, 0x4C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5046, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x48, 0x4C, 0x4A, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {5047, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {5048, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x64, 0x62, 0x2, 0x42, 0x72, 0x5E, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {5049, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0xA, 0x15, 0x20, 0x7C, 0x54, 0x14, 0x14, 0x14, 0x1C, 0x0, 0x0}},
    {5050, 1, {0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x4A, 0x4A, 0x4E, 0x4E, 0x4A, 0x4A, 0x4A, 0x32, 0x0, 0x0}},
    {5051, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {5052, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x10, 0x38, 0x44, 0x2, 0x2, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5053, 1, {0x0, 0x0, 0x0, 0x0, 0x47, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x72, 0x3C, 0x0, 0x0}},
    {5054, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5055, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x78, 0x20, 0x20, 0x20, 0x20, 0x20, 0x27, 0x22, 0x1E, 0x0, 0x0}},
    {5056, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {5057, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x48, 0x48, 0x44, 0x44, 0x44, 0x42, 0x42, 0x47, 0x0, 0x0}},
    {5058, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {5059, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x42, 0x7E, 0x0, 0x0}},
    {5060, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x72, 0x2, 0x72, 0x22, 0x1E, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {5061, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x48, 0x48, 0x4A, 0x4F, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5062, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x49, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x49, 0x7F, 0x0, 0x0}},
    {5063, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x2, 0x42, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x34, 0x0, 0x0}},
    {5064, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x5A, 0x4A, 0x16, 0x28, 0x48, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5065, 1, {0x0, 0x0, 0x0, 0x0, 0x67, 0x25, 0x24, 0x24, 0x28, 0x28, 0x28, 0x28, 0x28, 0x10, 0x0, 0x0}},
    {5066, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x41, 0x49, 0x55, 0x51, 0x51, 0x51, 0x2E, 0x0, 0x0}},
    {5067, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x3C, 0x40, 0x40, 0x46, 0x42, 0x3C, 0x0, 0x0}},
    {5068, 1, {0x0, 0x0, 0x0, 0x0, 0x77, 0x22, 0x22, 0x22, 0x22, 0x3E, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {5069, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0x9, 0x9, 0x9, 0x9, 0x1, 0x1D, 0x69, 0x59, 0x66, 0x0, 0x0}},
    {5070, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x14, 0x24, 0x44, 0x44, 0x7E, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5071, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x20, 0x20, 0x20, 0x20, 0x3C, 0x22, 0x22, 0x22, 0x7C, 0x0, 0x0}},
    {5072, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x8, 0x8, 0x8, 0xB, 0x5D, 0x68, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {5073, 1, {0x0, 0x0, 0x0, 0x0, 0x72, 0x4D, 0x71, 0x27, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5074, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x44, 0x78, 0x50, 0x48, 0x44, 0x44, 0x4E, 0x0, 0x0}},
    {5075, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {5076, 1, {0x0, 0x0, 0x0, 0x0, 0x55, 0x55, 0x55, 0x49, 0x59, 0x59, 0x65, 0x65, 0x43, 0x41, 0x0, 0x0}},
    {5077, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x4F, 0x40, 0x20, 0x1F, 0x4, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5078, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x13, 0x11, 0xE, 0x0, 0x0}},
    {5079, 1, {0x0, 0x0, 0x0, 0x0, 0x7, 0x2, 0x2, 0x2, 0x2, 0x6, 0xA, 0x12, 0x22, 0x47, 0x0, 0x0}},
    {5080, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x42, 0x2, 0x2, 0x2, 0x6, 0xA, 0x12, 0x22, 0x47, 0x0, 0x0}},
    {5081, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {5082, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5083, 1, {0x0, 0x0, 0x0, 0x0, 0x36, 0x49, 0x62, 0x50, 0x48, 0x48, 0x48, 0x48, 0x28, 0x10, 0x0, 0x0}},
    {5084, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x18, 0x24, 0x22, 0x55, 0x55, 0x55, 0x55, 0x22, 0x0, 0x0}},
    {5085, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {5086, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {5087, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5088, 1, {0x0, 0x0, 0x0, 0x0, 0x7, 0x22, 0x52, 0x7F, 0x12, 0x12, 0x12, 0x12, 0x12, 0xC, 0x0, 0x0}},
    {5089, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x5A, 0x59, 0x5B, 0x18, 0x24, 0x24, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {5090, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5091, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x47, 0x40, 0x40, 0x46, 0x46, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5092, 1, {0x0, 0x0, 0x0, 0x0, 0x7B, 0x51, 0x12, 0x12, 0x14, 0x14, 0x14, 0x18, 0x18, 0x10, 0x0, 0x0}},
    {5093, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x43, 0x4D, 0x74, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {5094, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {5095, 1, {0x0, 0x0, 0x0, 0x0, 0x7, 0x2, 0x2, 0x2, 0x1E, 0x22, 0x42, 0x42, 0x22, 0x1C, 0x0, 0x0}},
    {5096, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x4A, 0x45, 0x40, 0x40, 0x45, 0x4A, 0x48, 0x30, 0x0, 0x0}},
    {5097, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x4C, 0x4C, 0x40, 0x40, 0x4E, 0x44, 0x44, 0x3B, 0x0, 0x0}},
    {5098, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x55, 0x55, 0x53, 0x51, 0x31, 0x11, 0x59, 0x55, 0x62, 0x0, 0x0}},
    {5099, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x72, 0x52, 0x5E, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5100, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x55, 0x51, 0x51, 0x61, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {5101, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x12, 0x12, 0xA, 0x6, 0x2, 0x62, 0x64, 0x38, 0x0, 0x0}},
    {5102, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x46, 0x46, 0x40, 0x40, 0x7C, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {5103, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0x9, 0x1, 0x1, 0x1, 0x5D, 0x49, 0x49, 0x2A, 0x14, 0x0, 0x0}},
    {5104, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x3C, 0x22, 0x22, 0x62, 0x32, 0x2C, 0x0, 0x0}},
    {5105, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x8, 0x1C, 0x22, 0x22, 0x42, 0x42, 0x4A, 0x52, 0x4C, 0x0, 0x0}},
    {5106, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {5107, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x4A, 0x45, 0x40, 0x40, 0x5C, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5108, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {5121, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5122, 1, {0x8, 0x14, 0x8, 0x0, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5123, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5124, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5125, 1, {0x0, 0x0, 0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5126, 1, {0x10, 0x10, 0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5127, 1, {0x24, 0x24, 0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5128, 1, {0x0, 0x0, 0x0, 0x40, 0x60, 0x50, 0x48, 0x54, 0x52, 0x54, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5129, 1, {0x0, 0x0, 0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x52, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5130, 1, {0x0, 0x0, 0x0, 0x2, 0x6, 0xA, 0x12, 0x22, 0x42, 0x22, 0x12, 0xA, 0x6, 0x2, 0x0, 0x0}},
    {5131, 1, {0x10, 0x10, 0x0, 0x2, 0x6, 0xA, 0x12, 0x22, 0x42, 0x22, 0x12, 0xA, 0x6, 0x2, 0x0, 0x0}},
    {5132, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x22, 0x22, 0x22, 0x14, 0x54, 0x54, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5133, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x22, 0x22, 0x22, 0x14, 0x15, 0x15, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5134, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x54, 0x54, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5135, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x15, 0x15, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5136, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x54, 0x54, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5137, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x15, 0x15, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5138, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x18, 0x14, 0x52, 0x51, 0x12, 0x14, 0x18, 0x10, 0x0, 0x0, 0x0}},
    {5139, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x60, 0x50, 0x49, 0x45, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0, 0x0}},
    {5140, 1, {0x0, 0x8, 0x8, 0x0, 0x10, 0x18, 0x14, 0x52, 0x51, 0x12, 0x14, 0x18, 0x10, 0x0, 0x0, 0x0}},
    {5141, 1, {0x0, 0x20, 0x20, 0x0, 0x40, 0x60, 0x50, 0x49, 0x45, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0, 0x0}},
    {5142, 1, {0x20, 0x20, 0x0, 0x40, 0x62, 0x52, 0x48, 0x44, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5143, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x3, 0x5, 0x49, 0x51, 0x9, 0x5, 0x3, 0x1, 0x0, 0x0, 0x0}},
    {5144, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x14, 0x25, 0x45, 0x24, 0x14, 0xC, 0x4, 0x0, 0x0, 0x0}},
    {5145, 1, {0x0, 0x2, 0x2, 0x0, 0x1, 0x3, 0x5, 0x49, 0x51, 0x9, 0x5, 0x3, 0x1, 0x0, 0x0, 0x0}},
    {5146, 1, {0x0, 0x8, 0x8, 0x0, 0x4, 0xC, 0x14, 0x25, 0x45, 0x24, 0x14, 0xC, 0x4, 0x0, 0x0, 0x0}},
    {5147, 1, {0x4, 0x4, 0x0, 0x2, 0x46, 0x4A, 0x12, 0x22, 0x42, 0x22, 0x12, 0xA, 0x6, 0x2, 0x0, 0x0}},
    {5148, 1, {0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5149, 1, {0x0, 0x0, 0x4, 0xA, 0x24, 0x70, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5150, 1, {0x0, 0x0, 0x0, 0x10, 0x28, 0x44, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5151, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5152, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5153, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5154, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5155, 1, {0x0, 0x0, 0x0, 0x30, 0x8, 0x4, 0x4, 0x8, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5156, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5157, 1, {0x0, 0x0, 0x0, 0x9, 0x12, 0x24, 0x48, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5158, 1, {0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5159, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x38, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5160, 1, {0x0, 0x0, 0x0, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5161, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5162, 1, {0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5163, 1, {0x4, 0x2, 0x1, 0x7F, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5164, 1, {0x0, 0x4, 0x2, 0x9, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {5165, 1, {0x0, 0x8, 0x4, 0x42, 0x60, 0x50, 0x48, 0x44, 0x42, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0}},
    {5166, 1, {0x8, 0x4, 0x2, 0x0, 0x6, 0xA, 0x12, 0x22, 0x42, 0x22, 0x12, 0xA, 0x6, 0x2, 0x0, 0x0}},
    {5167, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5168, 1, {0x8, 0x14, 0x8, 0x0, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5169, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5170, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5171, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5172, 1, {0x10, 0x10, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5173, 1, {0x24, 0x24, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5174, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x24, 0x22, 0x24, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5175, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x22, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5176, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5177, 1, {0x10, 0x10, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5178, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x54, 0x54, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5179, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x15, 0x15, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5180, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x54, 0x54, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5181, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x15, 0x15, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5182, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x54, 0x54, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5183, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x15, 0x15, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5184, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x8, 0x4, 0x42, 0x41, 0x2, 0x4, 0x8, 0x10, 0x0, 0x0, 0x0}},
    {5185, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x9, 0x5, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0, 0x0}},
    {5186, 1, {0x0, 0x8, 0x8, 0x0, 0x10, 0x8, 0x4, 0x42, 0x41, 0x2, 0x4, 0x8, 0x10, 0x0, 0x0, 0x0}},
    {5187, 1, {0x0, 0x20, 0x20, 0x0, 0x40, 0x20, 0x10, 0x9, 0x5, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0, 0x0}},
    {5188, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x2, 0x4, 0x48, 0x50, 0x8, 0x4, 0x2, 0x1, 0x0, 0x0, 0x0}},
    {5189, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x21, 0x41, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0}},
    {5190, 1, {0x0, 0x2, 0x2, 0x0, 0x1, 0x2, 0x4, 0x48, 0x50, 0x8, 0x4, 0x2, 0x1, 0x0, 0x0, 0x0}},
    {5191, 1, {0x0, 0x8, 0x8, 0x0, 0x4, 0x8, 0x10, 0x21, 0x41, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0}},
    {5192, 1, {0x24, 0x24, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5193, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5194, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5195, 1, {0x0, 0x0, 0x0, 0x40, 0x50, 0x68, 0x48, 0x48, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5196, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {5197, 1, {0x8, 0x14, 0x8, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x0, 0x0}},
    {5198, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x0, 0x0}},
    {5199, 1, {0x8, 0x8, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x0, 0x0}},
    {5200, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0x2, 0x1, 0x1, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5201, 1, {0x8, 0x8, 0x0, 0x70, 0xC, 0x2, 0x2, 0x1, 0x1, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5202, 1, {0x12, 0x12, 0x0, 0x70, 0xC, 0x2, 0x2, 0x1, 0x1, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5203, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0x2, 0x21, 0x21, 0x21, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5204, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0x2, 0x1, 0x21, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5205, 1, {0x0, 0x0, 0x0, 0x7, 0x18, 0x20, 0x20, 0x40, 0x40, 0x40, 0x20, 0x20, 0x18, 0x7, 0x0, 0x0}},
    {5206, 1, {0x8, 0x8, 0x0, 0x7, 0x18, 0x20, 0x20, 0x40, 0x40, 0x40, 0x20, 0x20, 0x18, 0x7, 0x0, 0x0}},
    {5207, 1, {0x0, 0x0, 0x0, 0x11, 0x11, 0x11, 0x11, 0x11, 0x51, 0x51, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5208, 1, {0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x44, 0x45, 0x45, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5209, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x11, 0x51, 0x51, 0x11, 0x11, 0x11, 0x11, 0x11, 0x0, 0x0}},
    {5210, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x45, 0x45, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {5211, 1, {0x4, 0x4, 0x0, 0xE, 0x11, 0x11, 0x11, 0x51, 0x51, 0x11, 0x11, 0x11, 0x11, 0x11, 0x0, 0x0}},
    {5212, 1, {0x10, 0x10, 0x0, 0x38, 0x44, 0x44, 0x44, 0x45, 0x45, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {5213, 1, {0x0, 0x0, 0x0, 0x1C, 0x2, 0x2, 0x1, 0x41, 0x41, 0x1, 0x1, 0x2, 0x2, 0x1C, 0x0, 0x0}},
    {5214, 1, {0x0, 0x0, 0x0, 0x70, 0x8, 0x8, 0x4, 0x5, 0x5, 0x4, 0x4, 0x8, 0x8, 0x70, 0x0, 0x0}},
    {5215, 1, {0x8, 0x8, 0x0, 0x1C, 0x2, 0x2, 0x1, 0x41, 0x41, 0x1, 0x1, 0x2, 0x2, 0x1C, 0x0, 0x0}},
    {5216, 1, {0x20, 0x20, 0x0, 0x70, 0x8, 0x8, 0x4, 0x5, 0x5, 0x4, 0x4, 0x8, 0x8, 0x70, 0x0, 0x0}},
    {5217, 1, {0x0, 0x0, 0x0, 0x7, 0x8, 0x8, 0x10, 0x50, 0x50, 0x10, 0x10, 0x8, 0x8, 0x7, 0x0, 0x0}},
    {5218, 1, {0x0, 0x0, 0x0, 0x1C, 0x20, 0x20, 0x40, 0x41, 0x41, 0x40, 0x40, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {5219, 1, {0x2, 0x2, 0x0, 0x7, 0x8, 0x8, 0x10, 0x50, 0x50, 0x10, 0x10, 0x8, 0x8, 0x7, 0x0, 0x0}},
    {5220, 1, {0x8, 0x8, 0x0, 0x1C, 0x20, 0x20, 0x40, 0x41, 0x41, 0x40, 0x40, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {5221, 1, {0x0, 0x0, 0x0, 0x7, 0x8, 0x8, 0x50, 0x50, 0x10, 0x50, 0x50, 0x8, 0x8, 0x7, 0x0, 0x0}},
    {5222, 1, {0x0, 0x0, 0x0, 0x1C, 0x20, 0x40, 0x40, 0x40, 0x20, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5223, 1, {0x0, 0x0, 0x0, 0x45, 0x45, 0x45, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5224, 1, {0x0, 0x0, 0x0, 0x39, 0x45, 0x45, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {5225, 1, {0x0, 0x1, 0x1, 0x71, 0xC, 0x2, 0x2, 0x1, 0x1, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5226, 1, {0x0, 0x1, 0x1, 0x1D, 0x20, 0x20, 0x40, 0x40, 0x40, 0x40, 0x40, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {5227, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x3C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5228, 1, {0x10, 0x28, 0x10, 0x38, 0x44, 0x44, 0x44, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5229, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5230, 1, {0x10, 0x10, 0x0, 0x38, 0x44, 0x44, 0x44, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5231, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5232, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5233, 1, {0xA, 0xA, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5234, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x78, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5235, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x78, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5236, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x51, 0x4F, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5237, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x45, 0x3D, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5238, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x51, 0x5E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5239, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x45, 0x79, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5240, 1, {0x4, 0x4, 0x0, 0xE, 0x11, 0x11, 0x51, 0x5E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5241, 1, {0x10, 0x10, 0x0, 0x38, 0x44, 0x44, 0x45, 0x79, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5242, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x41, 0x4F, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5243, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x5, 0x3D, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5244, 1, {0x1, 0x1, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x41, 0x4F, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5245, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x5, 0x3D, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5246, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x5E, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5247, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x79, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5248, 1, {0x10, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x5E, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5249, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x79, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5250, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x10, 0x5E, 0x51, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5251, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x38, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5252, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x1C, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5253, 1, {0x0, 0x0, 0x0, 0x39, 0x45, 0x45, 0x44, 0x3C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5254, 1, {0x0, 0x0, 0x0, 0x39, 0x45, 0x45, 0x44, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5255, 1, {0x0, 0x0, 0x0, 0x5, 0x5, 0x5, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5256, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x40, 0x40, 0x40, 0x78, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5257, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5258, 1, {0x10, 0x28, 0x10, 0x38, 0x44, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5259, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5260, 1, {0x10, 0x10, 0x0, 0x38, 0x44, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5261, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5262, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5263, 1, {0xA, 0xA, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5264, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5265, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5266, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x51, 0x41, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5267, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x45, 0x5, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5268, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x51, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5269, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x45, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5270, 1, {0x4, 0x4, 0x0, 0xE, 0x11, 0x11, 0x51, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5271, 1, {0x10, 0x10, 0x0, 0x38, 0x44, 0x44, 0x45, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5272, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x41, 0x41, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5273, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x5, 0x5, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5274, 1, {0x1, 0x1, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x41, 0x41, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5275, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x5, 0x5, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5276, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x50, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5277, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x41, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5278, 1, {0x10, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x50, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5279, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x41, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5280, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x10, 0x50, 0x51, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5281, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5282, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5283, 1, {0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5284, 1, {0x10, 0x28, 0x10, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5285, 1, {0x0, 0x0, 0x0, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5286, 1, {0x10, 0x10, 0x0, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5287, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {5288, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {5289, 1, {0xA, 0xA, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {5290, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x0, 0x0}},
    {5291, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x0, 0x0}},
    {5292, 1, {0x0, 0x0, 0x0, 0x1F, 0x1, 0x1, 0x41, 0x41, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5293, 1, {0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x5, 0x5, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5294, 1, {0x0, 0x0, 0x0, 0x1F, 0x10, 0x10, 0x50, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5295, 1, {0x0, 0x0, 0x0, 0x7C, 0x40, 0x40, 0x41, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5296, 1, {0x4, 0x4, 0x0, 0x1F, 0x10, 0x10, 0x50, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5297, 1, {0x10, 0x10, 0x0, 0x7C, 0x40, 0x40, 0x41, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5298, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x41, 0x41, 0x1, 0x1, 0x1, 0x1F, 0x0, 0x0}},
    {5299, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x5, 0x5, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {5300, 1, {0x1, 0x1, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x41, 0x41, 0x1, 0x1, 0x1, 0x1F, 0x0, 0x0}},
    {5301, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x5, 0x5, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {5302, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x50, 0x10, 0x10, 0x10, 0x1F, 0x0, 0x0}},
    {5303, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x41, 0x40, 0x40, 0x40, 0x7C, 0x0, 0x0}},
    {5304, 1, {0x10, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x50, 0x10, 0x10, 0x10, 0x1F, 0x0, 0x0}},
    {5305, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x41, 0x40, 0x40, 0x40, 0x7C, 0x0, 0x0}},
    {5306, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x10, 0x50, 0x50, 0x10, 0x10, 0x1F, 0x0, 0x0}},
    {5307, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5308, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0x20, 0x20, 0x10, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5309, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5310, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x8, 0x10, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5311, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x42, 0x7E, 0x0, 0x0}},
    {5312, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x9, 0x9, 0x9, 0x9, 0x6, 0x0, 0x0}},
    {5313, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x8, 0x0, 0x3F, 0x48, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5314, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x48, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5315, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x3F, 0x48, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5316, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x9, 0x9, 0x9, 0x9, 0x7E, 0x0, 0x0}},
    {5317, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x0, 0x6, 0x9, 0x9, 0x9, 0x9, 0x7E, 0x0, 0x0}},
    {5318, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x9, 0x9, 0x0, 0x6, 0x9, 0x9, 0x9, 0x9, 0x7E, 0x0, 0x0}},
    {5319, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x48, 0x48, 0x48, 0x3F, 0x0, 0x0}},
    {5320, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x0, 0x30, 0x48, 0x48, 0x48, 0x48, 0x3F, 0x0, 0x0}},
    {5321, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5E, 0x45, 0x5, 0x5, 0x5, 0x2, 0x0, 0x0}},
    {5322, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x79, 0x15, 0x14, 0x14, 0x14, 0x8, 0x0, 0x0}},
    {5323, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x54, 0x14, 0x14, 0x14, 0xF, 0x0, 0x0}},
    {5324, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x31, 0x49, 0x48, 0x48, 0x48, 0x3E, 0x0, 0x0}},
    {5325, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x48, 0x54, 0x14, 0x14, 0x14, 0xF, 0x0, 0x0}},
    {5326, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x0, 0x31, 0x49, 0x48, 0x48, 0x48, 0x3E, 0x0, 0x0}},
    {5327, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x0, 0x30, 0x48, 0x48, 0x48, 0x48, 0x3F, 0x0, 0x0}},
    {5328, 1, {0x0, 0x0, 0x0, 0x20, 0x50, 0x50, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5329, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5330, 1, {0x0, 0x0, 0x0, 0x2, 0x5, 0x5, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5331, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x1, 0x1, 0x1, 0x1, 0x6, 0x0, 0x0}},
    {5332, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x8, 0x0, 0x3F, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5333, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5334, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x3F, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5335, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x7E, 0x0, 0x0}},
    {5336, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x7E, 0x0, 0x0}},
    {5337, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x9, 0x9, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x7E, 0x0, 0x0}},
    {5338, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x40, 0x40, 0x40, 0x3F, 0x0, 0x0}},
    {5339, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x0, 0x30, 0x40, 0x40, 0x40, 0x40, 0x3F, 0x0, 0x0}},
    {5340, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5E, 0x41, 0x1, 0x1, 0x1, 0x2, 0x0, 0x0}},
    {5341, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x79, 0x5, 0x4, 0x4, 0x4, 0x8, 0x0, 0x0}},
    {5342, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4F, 0x50, 0x10, 0x10, 0x10, 0x8, 0x0, 0x0}},
    {5343, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3D, 0x41, 0x40, 0x40, 0x40, 0x20, 0x0, 0x0}},
    {5344, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0x4F, 0x50, 0x10, 0x10, 0x10, 0x8, 0x0, 0x0}},
    {5345, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x0, 0x3D, 0x41, 0x40, 0x40, 0x40, 0x20, 0x0, 0x0}},
    {5346, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x41, 0x1, 0x1, 0x1, 0x3E, 0x0, 0x0}},
    {5347, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x9, 0x5, 0x4, 0x4, 0x4, 0x78, 0x0, 0x0}},
    {5348, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0x42, 0x41, 0x1, 0x1, 0x1, 0x3E, 0x0, 0x0}},
    {5349, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x9, 0x5, 0x4, 0x4, 0x4, 0x78, 0x0, 0x0}},
    {5350, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x50, 0x10, 0x10, 0x10, 0xF, 0x0, 0x0}},
    {5351, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x21, 0x41, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0}},
    {5352, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0x48, 0x50, 0x10, 0x10, 0x10, 0xF, 0x0, 0x0}},
    {5353, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x0, 0x21, 0x41, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0}},
    {5354, 1, {0x0, 0x0, 0x0, 0x20, 0x40, 0x40, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5355, 1, {0x0, 0x0, 0x0, 0xC, 0x30, 0xC, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5356, 1, {0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x30, 0xC, 0x2, 0xC, 0x30, 0x60, 0x18, 0x6, 0x0, 0x0}},
    {5357, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x20, 0x1C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5358, 1, {0x4, 0xA, 0x4, 0x0, 0x4, 0x4, 0x4, 0x8, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5359, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x8, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5360, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x8, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5361, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x1C, 0x20, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5362, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x1C, 0x20, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5363, 1, {0xA, 0xA, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x1C, 0x20, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5364, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x70, 0x8, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5365, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x70, 0x8, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5366, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x48, 0x7, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5367, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x41, 0x21, 0x1C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5368, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x41, 0x42, 0x1C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5369, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x5, 0x9, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5370, 1, {0x1, 0x1, 0x0, 0x1, 0x1, 0x1, 0x41, 0x42, 0x1C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5371, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x5, 0x9, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5372, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x41, 0x41, 0x7, 0x8, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5373, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x5, 0x5, 0x1C, 0x20, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5374, 1, {0x1, 0x1, 0x0, 0x1, 0x1, 0x1, 0x41, 0x41, 0x7, 0x8, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5375, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x5, 0x5, 0x1C, 0x20, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5376, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x1C, 0x2, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5377, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x41, 0x41, 0x70, 0x8, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5378, 1, {0x10, 0x10, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x1C, 0x2, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5379, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x41, 0x41, 0x70, 0x8, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5380, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x1C, 0x42, 0x41, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5381, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x30, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5382, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x2, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5383, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x18, 0x20, 0x20, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5384, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5385, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x30, 0x8, 0x18, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5386, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x18, 0x20, 0x30, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5387, 1, {0x20, 0x20, 0x0, 0x20, 0x20, 0x20, 0x30, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {5392, 1, {0x0, 0x0, 0x0, 0x30, 0x48, 0x48, 0x48, 0x8, 0x8, 0x8, 0x9, 0x9, 0x9, 0x6, 0x0, 0x0}},
    {5393, 1, {0x0, 0x0, 0x0, 0x6, 0x9, 0x9, 0x9, 0x8, 0x8, 0x8, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5394, 1, {0x8, 0x8, 0x0, 0x6, 0x9, 0x9, 0x9, 0x8, 0x8, 0x8, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5395, 1, {0x0, 0x0, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x3E, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5396, 1, {0x8, 0x8, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x3E, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5397, 1, {0x0, 0x0, 0x0, 0x30, 0x40, 0x40, 0x40, 0x40, 0x3E, 0x1, 0x1, 0x1, 0x1, 0x6, 0x0, 0x0}},
    {5398, 1, {0x8, 0x8, 0x0, 0x30, 0x40, 0x40, 0x40, 0x40, 0x3E, 0x1, 0x1, 0x1, 0x1, 0x6, 0x0, 0x0}},
    {5399, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x14, 0x44, 0x44, 0x4, 0x5, 0x5, 0x5, 0x2, 0x0, 0x0}},
    {5400, 1, {0x0, 0x0, 0x0, 0x20, 0x50, 0x50, 0x50, 0x11, 0x11, 0x10, 0x14, 0x14, 0x14, 0x8, 0x0, 0x0}},
    {5401, 1, {0x0, 0x0, 0x0, 0x2, 0x5, 0x5, 0x5, 0x44, 0x44, 0x4, 0x14, 0x14, 0x14, 0x8, 0x0, 0x0}},
    {5402, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x14, 0x11, 0x11, 0x10, 0x50, 0x50, 0x50, 0x20, 0x0, 0x0}},
    {5403, 1, {0x0, 0x8, 0x8, 0x2, 0x5, 0x5, 0x5, 0x44, 0x44, 0x4, 0x14, 0x14, 0x14, 0x8, 0x0, 0x0}},
    {5404, 1, {0x0, 0x20, 0x20, 0x8, 0x14, 0x14, 0x14, 0x11, 0x11, 0x10, 0x50, 0x50, 0x50, 0x20, 0x0, 0x0}},
    {5405, 1, {0x0, 0x0, 0x0, 0x2, 0x1, 0x1, 0x1, 0x1, 0x4E, 0x50, 0x10, 0x10, 0x10, 0x8, 0x0, 0x0}},
    {5406, 1, {0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0x4, 0x4, 0x39, 0x41, 0x40, 0x40, 0x40, 0x20, 0x0, 0x0}},
    {5407, 1, {0x4, 0x4, 0x0, 0x2, 0x1, 0x1, 0x1, 0x1, 0x4E, 0x50, 0x10, 0x10, 0x10, 0x8, 0x0, 0x0}},
    {5408, 1, {0x20, 0x20, 0x0, 0x8, 0x4, 0x4, 0x4, 0x4, 0x39, 0x41, 0x40, 0x40, 0x40, 0x20, 0x0, 0x0}},
    {5409, 1, {0x0, 0x0, 0x0, 0x8, 0x10, 0x10, 0x10, 0x10, 0x4E, 0x41, 0x1, 0x1, 0x1, 0x2, 0x0, 0x0}},
    {5410, 1, {0x0, 0x0, 0x0, 0x20, 0x40, 0x40, 0x40, 0x40, 0x39, 0x5, 0x4, 0x4, 0x4, 0x8, 0x0, 0x0}},
    {5411, 1, {0x2, 0x2, 0x0, 0x8, 0x10, 0x10, 0x10, 0x10, 0x4E, 0x41, 0x1, 0x1, 0x1, 0x2, 0x0, 0x0}},
    {5412, 1, {0x10, 0x10, 0x0, 0x20, 0x40, 0x40, 0x40, 0x40, 0x39, 0x5, 0x4, 0x4, 0x4, 0x8, 0x0, 0x0}},
    {5413, 1, {0x0, 0x0, 0x0, 0x20, 0x40, 0x40, 0x3E, 0x1, 0x1, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5414, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {5415, 1, {0x20, 0x50, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5416, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5417, 1, {0x20, 0x20, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5418, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5419, 1, {0x2, 0x2, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5420, 1, {0x12, 0x12, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5421, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5422, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5423, 1, {0x0, 0x0, 0x0, 0x1, 0x2, 0x4, 0x8, 0x50, 0x5F, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5424, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x41, 0x7D, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5425, 1, {0x0, 0x0, 0x0, 0x10, 0x8, 0x4, 0x2, 0x41, 0x5F, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5426, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x5, 0x7D, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5427, 1, {0x10, 0x10, 0x0, 0x10, 0x8, 0x4, 0x2, 0x41, 0x5F, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {5428, 1, {0x40, 0x40, 0x0, 0x40, 0x20, 0x10, 0x8, 0x5, 0x7D, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5429, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x5F, 0x50, 0x8, 0x4, 0x2, 0x1, 0x0, 0x0}},
    {5430, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7D, 0x41, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0}},
    {5431, 1, {0x1, 0x1, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x5F, 0x50, 0x8, 0x4, 0x2, 0x1, 0x0, 0x0}},
    {5432, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7D, 0x41, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0}},
    {5433, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x5F, 0x41, 0x2, 0x4, 0x8, 0x10, 0x0, 0x0}},
    {5434, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7D, 0x5, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5435, 1, {0x10, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x5F, 0x41, 0x2, 0x4, 0x8, 0x10, 0x0, 0x0}},
    {5436, 1, {0x40, 0x40, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7D, 0x5, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5437, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x50, 0x1F, 0x41, 0x42, 0x4, 0x8, 0x10, 0x0, 0x0}},
    {5438, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x3C, 0x8, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5439, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x28, 0x54, 0x54, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5440, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5441, 1, {0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5442, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x47, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5443, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x71, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5444, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x11, 0x71, 0x0, 0x0}},
    {5445, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x10, 0x0, 0x38, 0x44, 0x44, 0x44, 0x47, 0x0, 0x0}},
    {5446, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x47, 0x0, 0x0}},
    {5447, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x0, 0x38, 0x44, 0x44, 0x44, 0x47, 0x0, 0x0}},
    {5448, 1, {0x0, 0x0, 0x0, 0x70, 0x8, 0x4, 0x4, 0x8, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5449, 1, {0x20, 0x20, 0x0, 0x70, 0x8, 0x4, 0x4, 0x8, 0x70, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5450, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x70, 0x8, 0x4, 0x4, 0x8, 0x70, 0x0, 0x0}},
    {5451, 1, {0x0, 0x0, 0x0, 0x1C, 0x20, 0x40, 0x40, 0x20, 0x1C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5452, 1, {0x8, 0x8, 0x0, 0x1C, 0x20, 0x40, 0x40, 0x20, 0x1C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5453, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x1C, 0x20, 0x40, 0x40, 0x20, 0x1C, 0x0, 0x0}},
    {5454, 1, {0x2, 0x2, 0x0, 0x7, 0x8, 0x10, 0x10, 0x48, 0x47, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5455, 1, {0x8, 0x8, 0x0, 0x1C, 0x20, 0x40, 0x40, 0x21, 0x1D, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5456, 1, {0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5457, 1, {0x0, 0x0, 0x0, 0x30, 0xC, 0x30, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5458, 1, {0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0xC, 0x30, 0x40, 0x30, 0xC, 0x6, 0x18, 0x60, 0x0, 0x0}},
    {5459, 1, {0x0, 0x0, 0x0, 0x31, 0x49, 0x32, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5460, 1, {0x8, 0x14, 0x8, 0x0, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x32, 0x49, 0x31, 0x0, 0x0}},
    {5461, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x32, 0x49, 0x31, 0x0, 0x0}},
    {5462, 1, {0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x32, 0x49, 0x31, 0x0, 0x0}},
    {5463, 1, {0x0, 0x0, 0x0, 0x20, 0x50, 0x28, 0x4, 0x2, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {5464, 1, {0x8, 0x8, 0x0, 0x20, 0x50, 0x28, 0x4, 0x2, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {5465, 1, {0x0, 0x0, 0x0, 0x2, 0x5, 0xA, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5466, 1, {0x8, 0x8, 0x0, 0x2, 0x5, 0xA, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5467, 1, {0x4, 0x4, 0x0, 0x0, 0x2, 0x5, 0xA, 0x90, 0xA0, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0, 0x0}},
    {5468, 1, {0x8, 0x8, 0x0, 0x0, 0x4, 0xA, 0x14, 0x21, 0x41, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0}},
    {5469, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0x28, 0x40, 0x20, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5470, 1, {0x0, 0x0, 0x0, 0x31, 0x49, 0x71, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {5471, 1, {0x0, 0x0, 0x0, 0x46, 0x49, 0x47, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {5472, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x71, 0x49, 0x31, 0x0, 0x0}},
    {5473, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x47, 0x49, 0x46, 0x0, 0x0}},
    {5474, 1, {0x8, 0x8, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x71, 0x49, 0x31, 0x0, 0x0}},
    {5475, 1, {0x8, 0x8, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x47, 0x49, 0x46, 0x0, 0x0}},
    {5476, 1, {0x0, 0x0, 0x0, 0x30, 0x4C, 0x32, 0x2, 0x1, 0x1, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5477, 1, {0x8, 0x8, 0x0, 0x30, 0x4C, 0x32, 0x2, 0x1, 0x1, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5478, 1, {0x0, 0x0, 0x0, 0x6, 0x19, 0x26, 0x20, 0x40, 0x40, 0x40, 0x20, 0x20, 0x18, 0x7, 0x0, 0x0}},
    {5479, 1, {0x8, 0x8, 0x0, 0x6, 0x19, 0x26, 0x20, 0x40, 0x40, 0x40, 0x20, 0x20, 0x18, 0x7, 0x0, 0x0}},
    {5480, 1, {0x2, 0x2, 0x0, 0x6, 0x9, 0xE, 0x10, 0x50, 0x50, 0x10, 0x10, 0x8, 0x8, 0x7, 0x0, 0x0}},
    {5481, 1, {0x8, 0x8, 0x0, 0x18, 0x24, 0x38, 0x40, 0x41, 0x41, 0x40, 0x40, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {5482, 1, {0x0, 0x0, 0x0, 0x18, 0x34, 0x48, 0x40, 0x40, 0x20, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5483, 1, {0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {5484, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {5485, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0x2, 0x1, 0x79, 0x1, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5486, 1, {0x0, 0x0, 0x0, 0x7, 0x18, 0x20, 0x20, 0x40, 0x4F, 0x40, 0x20, 0x20, 0x18, 0x7, 0x0, 0x0}},
    {5488, 1, {0x0, 0x0, 0x0, 0x55, 0x55, 0x55, 0x55, 0x55, 0x49, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {5489, 1, {0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x49, 0x55, 0x55, 0x55, 0x55, 0x55, 0x0, 0x0}},
    {5490, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0x72, 0x9, 0x9, 0x9, 0x72, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5491, 1, {0x0, 0x0, 0x0, 0x7, 0x18, 0x20, 0x27, 0x48, 0x48, 0x48, 0x27, 0x20, 0x18, 0x7, 0x0, 0x0}},
    {5492, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x3C, 0x4, 0x4, 0x4, 0x5, 0x6, 0x4, 0x0, 0x0}},
    {5493, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x11, 0x1E, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x0, 0x0}},
    {5494, 1, {0x4, 0x4, 0x0, 0xE, 0x11, 0x11, 0x11, 0x1E, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x0, 0x0}},
    {5495, 1, {0x0, 0x0, 0x0, 0x4, 0x6, 0x5, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5496, 1, {0x4, 0x4, 0x0, 0x4, 0x6, 0x5, 0x4, 0x4, 0x4, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5497, 1, {0x0, 0x0, 0x0, 0x10, 0x30, 0x50, 0x10, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5498, 1, {0x10, 0x10, 0x0, 0x10, 0x30, 0x50, 0x10, 0x10, 0x10, 0x1E, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5499, 1, {0x0, 0x0, 0x0, 0x20, 0x60, 0x20, 0x38, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5500, 1, {0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {5501, 1, {0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5509, 1, {0x0, 0x0, 0x0, 0x34, 0x44, 0x36, 0x15, 0x12, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5510, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x3C, 0x14, 0x14, 0x24, 0x24, 0x44, 0x44, 0x0, 0x0}},
    {5511, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x78, 0x50, 0x50, 0x48, 0x48, 0x44, 0x44, 0x0, 0x0}},
    {5512, 1, {0x0, 0x0, 0x0, 0x44, 0x44, 0x24, 0x24, 0x14, 0x14, 0x3C, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5513, 1, {0x0, 0x0, 0x0, 0x44, 0x44, 0x48, 0x48, 0x50, 0x50, 0x78, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5514, 1, {0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x4, 0x4, 0x4, 0x4, 0x5, 0x6, 0x4, 0x0, 0x0}},
    {5515, 1, {0x0, 0x0, 0x0, 0xE, 0x11, 0x11, 0x11, 0x10, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x0, 0x0}},
    {5516, 1, {0x0, 0x0, 0x0, 0x4, 0x6, 0x5, 0x4, 0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {5517, 1, {0x0, 0x0, 0x0, 0x10, 0x30, 0x50, 0x10, 0x10, 0x10, 0x10, 0x11, 0x11, 0x11, 0xE, 0x0, 0x0}},
    {5525, 1, {0x0, 0x0, 0x30, 0x48, 0x48, 0x3C, 0x5, 0x5, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5527, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x5A, 0x4A, 0x32, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {5528, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x5A, 0x52, 0x4C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5529, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x32, 0x4A, 0x5A, 0x42, 0x3C, 0x0, 0x0}},
    {5530, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x4C, 0x52, 0x5A, 0x42, 0x3C, 0x0, 0x0}},
    {5531, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x7E, 0x9, 0x9, 0x9, 0x9, 0x6, 0x0, 0x0}},
    {5532, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x0, 0x3F, 0x48, 0x48, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {5533, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x6, 0x9, 0x9, 0x9, 0x9, 0x7E, 0x0, 0x0}},
    {5534, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x0, 0x30, 0x48, 0x48, 0x48, 0x48, 0x3F, 0x0, 0x0}},
    {5535, 1, {0x0, 0x0, 0x0, 0x3F, 0x0, 0x20, 0x50, 0x50, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5536, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x3F, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5537, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x14, 0x2, 0x3F, 0x40, 0x40, 0x40, 0x40, 0x30, 0x0, 0x0}},
    {5538, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x7E, 0x20, 0x10}},
    {5539, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x0, 0x6, 0x1, 0x1, 0x1, 0x1, 0x7E, 0x20, 0x10}},
    {5540, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x40, 0x40, 0x40, 0x3F, 0x2, 0x4}},
    {5541, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x0, 0x30, 0x40, 0x40, 0x40, 0x40, 0x3F, 0x2, 0x4}},
    {5542, 1, {0x0, 0x0, 0x0, 0x20, 0x40, 0x40, 0x3F, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5543, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x7C, 0x4, 0x4, 0x1F, 0x4, 0x4, 0x0, 0x0}},
    {5544, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1F, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {5545, 1, {0x10, 0x10, 0x0, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1F, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {5546, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x1F, 0x4, 0x4, 0x7C, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5547, 1, {0x4, 0x4, 0x0, 0x4, 0x4, 0x1F, 0x4, 0x4, 0x7C, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5548, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x1F, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {5549, 1, {0x10, 0x10, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x1F, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {5550, 1, {0x0, 0x0, 0x0, 0x20, 0x70, 0x20, 0x3C, 0x8, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5551, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x41, 0x41, 0x62, 0x5C, 0x0, 0x0}},
    {5552, 1, {0x0, 0x0, 0x0, 0x5C, 0x7E, 0x7F, 0x7F, 0x7E, 0x5C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5553, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1D, 0x3F, 0x7F, 0x7F, 0x3F, 0x1D, 0x0, 0x0}},
    {5554, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x5C, 0x7E, 0x7F, 0x7F, 0x7E, 0x5C, 0x0, 0x0}},
    {5555, 1, {0x0, 0x0, 0x0, 0x1D, 0x3F, 0x7F, 0x7F, 0x3F, 0x1D, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5556, 1, {0x0, 0x0, 0x0, 0x7C, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5557, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x4, 0x4, 0x4, 0x7C, 0x0, 0x0}},
    {5558, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x7C, 0x0, 0x0}},
    {5559, 1, {0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {5560, 1, {0x0, 0x0, 0x0, 0x72, 0x52, 0x52, 0x4C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5561, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x32, 0x4A, 0x4A, 0x4E, 0x0, 0x0}},
    {5562, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x4C, 0x52, 0x52, 0x72, 0x0, 0x0}},
    {5563, 1, {0x0, 0x0, 0x0, 0x4E, 0x4A, 0x4A, 0x32, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {5564, 1, {0x0, 0x0, 0x0, 0x41, 0x62, 0x54, 0x48, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5565, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x9, 0x15, 0x23, 0x41, 0x0, 0x0}},
    {5566, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x48, 0x54, 0x62, 0x41, 0x0, 0x0}},
    {5567, 1, {0x0, 0x0, 0x0, 0x41, 0x23, 0x15, 0x9, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5568, 1, {0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x39, 0x45, 0x45, 0x45, 0x3, 0x1, 0x0, 0x0}},
    {5569, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x4E, 0x51, 0x51, 0x51, 0x60, 0x40, 0x0, 0x0}},
    {5570, 1, {0x0, 0x0, 0x0, 0x3, 0x45, 0x45, 0x45, 0x39, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5571, 1, {0x0, 0x0, 0x0, 0x60, 0x51, 0x51, 0x51, 0x4E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {5572, 1, {0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x3E, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5573, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x3E, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {5574, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x18, 0x14, 0x12, 0x14, 0x18, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5575, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x18, 0x54, 0x52, 0x54, 0x18, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5576, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x18, 0x14, 0x52, 0x14, 0x18, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5577, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x18, 0x28, 0x48, 0x28, 0x18, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5578, 1, {0x0, 0x0, 0x0, 0x41, 0x7F, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5579, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x7F, 0x41, 0x0, 0x0}},
    {5580, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x30, 0x28, 0x24, 0x22, 0x24, 0x28, 0x30, 0x20, 0x40, 0x0, 0x0}},
    {5581, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x30, 0x28, 0x34, 0x32, 0x34, 0x28, 0x30, 0x20, 0x40, 0x0, 0x0}},
    {5582, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x30, 0x28, 0x24, 0x32, 0x24, 0x28, 0x30, 0x20, 0x40, 0x0, 0x0}},
    {5583, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0xC, 0x14, 0x24, 0x44, 0x24, 0x14, 0xC, 0x4, 0x2, 0x0, 0x0}},
    {5584, 1, {0x0, 0x0, 0x0, 0x49, 0x49, 0x2A, 0x2A, 0x2A, 0x1C, 0x1C, 0x1C, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5585, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x1C, 0x1C, 0x1C, 0x2A, 0x2A, 0x2A, 0x49, 0x49, 0x0, 0x0}},
    {5586, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x7E, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5587, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x28, 0x24, 0x7E, 0x24, 0x28, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5588, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x34, 0x7E, 0x34, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5589, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x7E, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {5590, 1, {0x0, 0x0, 0x0, 0x49, 0x49, 0x3E, 0x2A, 0x2A, 0x1C, 0x1C, 0x1C, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {5591, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x1C, 0x1C, 0x1C, 0x2A, 0x2A, 0x3E, 0x49, 0x49, 0x0, 0x0}},
    {5592, 1, {0x0, 0x0, 0x0, 0x40, 0x30, 0x18, 0x14, 0x12, 0x7F, 0x12, 0x14, 0x18, 0x30, 0x40, 0x0, 0x0}},
    {5593, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x58, 0x54, 0xFE, 0x54, 0x58, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5594, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x18, 0x74, 0xFE, 0x74, 0x18, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {5595, 1, {0x0, 0x0, 0x0, 0x1, 0x6, 0xC, 0x14, 0x24, 0x7F, 0x24, 0x14, 0xC, 0x6, 0x1, 0x0, 0x0}},
    {5596, 1, {0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {5597, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {5598, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5599, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x52, 0x52, 0x52, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5600, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5601, 1, {0x0, 0x0, 0x0, 0x1E, 0x22, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x1E, 0x0, 0x0}},
    {5602, 1, {0x0, 0x0, 0x0, 0x42, 0x66, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {5603, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x66, 0x42, 0x0, 0x0}},
    {5604, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x22, 0x22, 0x12, 0x22, 0x22, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5605, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x22, 0x2A, 0x1A, 0x2A, 0x22, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5606, 1, {0x0, 0x0, 0x0, 0xF0, 0x8C, 0x82, 0x42, 0x5A, 0x3A, 0x42, 0x42, 0x82, 0x8C, 0xF0, 0x0, 0x0}},
    {5607, 1, {0x0, 0x0, 0x0, 0x1E, 0x22, 0x42, 0x44, 0x44, 0x48, 0x44, 0x44, 0x42, 0x22, 0x1E, 0x0, 0x0}},
    {5608, 1, {0x0, 0x0, 0x0, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {5609, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x0, 0x0}},
    {5610, 1, {0x0, 0x0, 0x0, 0xF8, 0x24, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x24, 0xF8, 0x0, 0x0}},
    {5611, 1, {0x0, 0x0, 0x0, 0xF8, 0x24, 0x22, 0x22, 0x2A, 0x2A, 0x2A, 0x22, 0x22, 0x24, 0xF8, 0x0, 0x0}},
    {5612, 1, {0x0, 0x0, 0x0, 0xF0, 0x4C, 0x42, 0x42, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x4C, 0xF0, 0x0, 0x0}},
    {5613, 1, {0x0, 0x0, 0x0, 0x3E, 0x48, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x48, 0x3E, 0x0, 0x0}},
    {5614, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5615, 1, {0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x2A, 0x14, 0x0, 0x0}},
    {5616, 1, {0x0, 0x0, 0x0, 0x14, 0x2A, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {5617, 1, {0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x2, 0x2, 0x7C, 0x2, 0x2, 0x2, 0x4, 0x78, 0x0, 0x0}},
    {5618, 1, {0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x2, 0x12, 0x7C, 0x12, 0x2, 0x2, 0x4, 0x78, 0x0, 0x0}},
    {5619, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0x2, 0x32, 0x7C, 0x32, 0x2, 0x2, 0xC, 0x70, 0x0, 0x0}},
    {5620, 1, {0x0, 0x0, 0x0, 0x1E, 0x20, 0x40, 0x40, 0x40, 0x3E, 0x40, 0x40, 0x40, 0x20, 0x1E, 0x0, 0x0}},
    {5621, 1, {0x0, 0x0, 0x0, 0x7F, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x2A, 0x14, 0x0, 0x0}},
    {5622, 1, {0x0, 0x0, 0x0, 0x14, 0x2A, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {5623, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5624, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x52, 0x7C, 0x52, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5625, 1, {0x0, 0x0, 0x0, 0xF0, 0x8C, 0x82, 0x82, 0xB2, 0xFC, 0xB2, 0x82, 0x82, 0x8C, 0xF0, 0x0, 0x0}},
    {5626, 1, {0x0, 0x0, 0x0, 0x1E, 0x22, 0x42, 0x42, 0x42, 0x3E, 0x42, 0x42, 0x42, 0x22, 0x1E, 0x0, 0x0}},
    {5627, 1, {0x0, 0x0, 0x0, 0x49, 0x6B, 0x5D, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x2A, 0x14, 0x0, 0x0}},
    {5628, 1, {0x0, 0x0, 0x0, 0x14, 0x2A, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x5D, 0x6B, 0x49, 0x0, 0x0}},
    {5629, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x22, 0x22, 0x7E, 0x22, 0x22, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5630, 1, {0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x22, 0x2A, 0x7E, 0x2A, 0x22, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {5631, 1, {0x0, 0x0, 0x0, 0xF0, 0x8C, 0x82, 0x42, 0x5A, 0xFE, 0x5A, 0x42, 0x82, 0x8C, 0xF0, 0x0, 0x0}},
    {5632, 1, {0x0, 0x0, 0x0, 0x1E, 0x22, 0x42, 0x44, 0x44, 0x7E, 0x44, 0x44, 0x42, 0x22, 0x1E, 0x0, 0x0}},
    {5633, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5634, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {5635, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {5636, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x2, 0x2, 0x2, 0xC, 0x0, 0x0}},
    {5637, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x2, 0x22, 0x22, 0xC, 0x0, 0x0}},
    {5638, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x2, 0x62, 0x62, 0xC, 0x0, 0x0}},
    {5639, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x80, 0x80, 0x80, 0x7E, 0x0, 0x0}},
    {5640, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x49, 0x49, 0x49, 0x2A, 0x14, 0x0, 0x0}},
    {5641, 1, {0x0, 0x0, 0x0, 0x14, 0x2A, 0x49, 0x49, 0x49, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0, 0x0}},
    {5642, 1, {0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x2, 0x2, 0x1C, 0x2, 0x2, 0x2, 0x4, 0x18, 0x0, 0x0}},
    {5643, 1, {0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x2, 0x12, 0x1C, 0x12, 0x2, 0x2, 0x4, 0x18, 0x0, 0x0}},
    {5644, 1, {0x0, 0x0, 0x0, 0xF0, 0xC, 0x2, 0x2, 0x32, 0x7C, 0x32, 0x2, 0x2, 0xC, 0x30, 0x0, 0x0}},
    {5645, 1, {0x0, 0x0, 0x0, 0x18, 0x20, 0x40, 0x40, 0x40, 0x38, 0x40, 0x40, 0x40, 0x20, 0x1E, 0x0, 0x0}},
    {5646, 1, {0x0, 0x0, 0x0, 0x4C, 0x52, 0x52, 0x52, 0x4A, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {5647, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x52, 0x4A, 0x4A, 0x4A, 0x32, 0x0, 0x0}},
    {5648, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x2, 0x32, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {5649, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x12, 0x32, 0x52, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {5650, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x4, 0x2, 0x1A, 0x3E, 0x5A, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {5651, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x42, 0x42, 0x4C, 0x40, 0x40, 0x20, 0x1E, 0x0, 0x0}},
    {5652, 1, {0x0, 0x0, 0x0, 0x0, 0xFE, 0x4, 0x8, 0x10, 0x20, 0x40, 0x98, 0xA4, 0x84, 0x78, 0x0, 0x0}},
    {5653, 1, {0x0, 0x0, 0x0, 0x0, 0xFE, 0x40, 0x20, 0x10, 0x8, 0x4, 0x32, 0x4A, 0x42, 0x3C, 0x0, 0x0}},
    {5654, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x4A, 0x32, 0x4, 0x8, 0x10, 0x20, 0x40, 0xFE, 0x0, 0x0, 0x0}},
    {5668, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x50, 0x68, 0x44, 0x4, 0x4, 0x8, 0x70, 0x0, 0x0}},
    {5669, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x50, 0x68, 0x44, 0x14, 0x14, 0x8, 0x70, 0x0, 0x0}},
    {5670, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x50, 0x68, 0x44, 0x34, 0x34, 0x8, 0x70, 0x0, 0x0}},
    {5671, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x14, 0x2C, 0x44, 0x40, 0x40, 0x20, 0x1C, 0x0, 0x0}},
    {5674, 1, {0x0, 0x0, 0x0, 0x0, 0xF8, 0x20, 0x20, 0x28, 0x34, 0x22, 0x2, 0x2, 0x4, 0x38, 0x0, 0x0}},
    {5675, 1, {0x0, 0x0, 0x0, 0x0, 0xF8, 0x20, 0x20, 0x28, 0x34, 0x22, 0xA, 0xA, 0x4, 0x38, 0x0, 0x0}},
    {5676, 1, {0x0, 0x0, 0x0, 0x0, 0xF8, 0x20, 0x20, 0x28, 0x34, 0x22, 0x1A, 0x1A, 0x4, 0x38, 0x0, 0x0}},
    {5677, 1, {0x0, 0x0, 0x0, 0x0, 0x1F, 0x4, 0x4, 0x14, 0x2C, 0x44, 0x40, 0x40, 0x20, 0x1C, 0x0, 0x0}},
    {5678, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC6, 0x44, 0x82, 0x82, 0x82, 0x44, 0x38, 0x0, 0x0}},
    {5679, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x82, 0x82, 0x82, 0x44, 0xC6, 0x0, 0x0}},
    {5680, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xB8, 0xC4, 0x2, 0x2, 0x2, 0xC4, 0xB8, 0x0, 0x0}},
    {5681, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xB8, 0xC4, 0x12, 0x12, 0x12, 0xC4, 0xB8, 0x0, 0x0}},
    {5682, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xB8, 0xC4, 0x1A, 0x3E, 0x1A, 0xC4, 0xB8, 0x0, 0x0}},
    {5683, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x46, 0x80, 0x80, 0x80, 0x46, 0x3A, 0x0, 0x0}},
    {5686, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0x40, 0x5C, 0x62, 0x1, 0x1, 0x1, 0x62, 0x5C, 0x0, 0x0}},
    {5687, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0x40, 0x5C, 0x62, 0x9, 0x9, 0x9, 0x62, 0x5C, 0x0, 0x0}},
    {5688, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0x40, 0x5C, 0x62, 0xD, 0x1F, 0xD, 0x62, 0x5C, 0x0, 0x0}},
    {5689, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0x2, 0x3A, 0x46, 0x80, 0x80, 0x80, 0x46, 0x3A, 0x0, 0x0}},
    {5696, 1, {0x0, 0x0, 0x0, 0x82, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x28, 0x10, 0x0, 0x0}},
    {5697, 1, {0x0, 0x0, 0x0, 0x10, 0x28, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x82, 0x0, 0x0}},
    {5702, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xFE, 0x4, 0x8, 0x10, 0x20, 0x40, 0xFE, 0x0, 0x0, 0x0, 0x0}},
    {5703, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0xFE, 0x14, 0x18, 0x10, 0x30, 0x50, 0xFE, 0x10, 0x10, 0x0, 0x0}},
    {5704, 1, {0x0, 0x0, 0x0, 0xBA, 0x54, 0x54, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x28, 0x10, 0x0, 0x0}},
    {5705, 1, {0x0, 0x0, 0x0, 0x10, 0x28, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x54, 0x54, 0xBA, 0x0, 0x0}},
    {5722, 1, {0x0, 0x10, 0x10, 0x38, 0x54, 0x50, 0x38, 0x14, 0x54, 0x38, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {5741, 1, {0x0, 0x0, 0x0, 0x62, 0x22, 0x14, 0x14, 0x8, 0x8, 0x8, 0x14, 0x14, 0x22, 0x23, 0x0, 0x0}},
    {5742, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x10, 0x28, 0x0}},
    {5792, 1, {0x0, 0x92, 0xA4, 0xC8, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5793, 1, {0x0, 0xAA, 0x94, 0xA8, 0xD0, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5794, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x0}},
    {5795, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0x82, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x0}},
    {5796, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0x84, 0xB4, 0xB4, 0x84, 0x84, 0x84, 0x84, 0x84, 0x0}},
    {5797, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0xC2, 0xA2, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x0}},
    {5798, 1, {0x0, 0x80, 0x80, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x88, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x0}},
    {5799, 1, {0x0, 0x80, 0x80, 0x80, 0xC0, 0xA0, 0x90, 0xA8, 0xA8, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x0}},
    {5800, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0xC8, 0xA0, 0x90, 0x88, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5801, 1, {0x0, 0x80, 0xC2, 0xA4, 0x98, 0xC2, 0xA4, 0x98, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5802, 1, {0x0, 0x80, 0xC2, 0xA4, 0x98, 0xC0, 0xA0, 0x90, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5803, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0xC8, 0xA0, 0x90, 0x88, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5804, 1, {0x0, 0x10, 0x10, 0x90, 0x50, 0x30, 0x90, 0x58, 0x34, 0x12, 0x18, 0x14, 0x12, 0x10, 0x10, 0x0}},
    {5805, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x18, 0x14, 0x12, 0x18, 0x14, 0x12, 0x10, 0x10, 0x10, 0x0}},
    {5806, 1, {0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x18, 0x28, 0x48, 0x18, 0x28, 0x48, 0x8, 0x8, 0x8, 0x8}},
    {5807, 1, {0x0, 0x10, 0x10, 0x12, 0x14, 0x18, 0x12, 0x34, 0x58, 0x90, 0x30, 0x50, 0x90, 0x10, 0x10, 0x0}},
    {5808, 1, {0x0, 0x10, 0x10, 0x12, 0x14, 0x18, 0x10, 0x30, 0x50, 0x90, 0x30, 0x50, 0x90, 0x10, 0x10, 0x0}},
    {5809, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0x88, 0x90, 0xA0, 0xA0, 0x90, 0x88, 0x84, 0x82, 0x0}},
    {5810, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {5811, 1, {0x0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0x82, 0x0}},
    {5812, 1, {0x0, 0x82, 0x84, 0x88, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5813, 1, {0x0, 0xB2, 0xB4, 0x88, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5814, 1, {0x0, 0x42, 0x44, 0x48, 0x50, 0xE0, 0xE0, 0xE0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0}},
    {5815, 1, {0x0, 0x82, 0x82, 0x44, 0x44, 0x28, 0x28, 0x10, 0x10, 0x28, 0x28, 0x44, 0x44, 0x82, 0x82, 0x0}},
    {5816, 1, {0x0, 0x82, 0x82, 0x44, 0x44, 0x28, 0x28, 0x54, 0x54, 0x28, 0x28, 0x44, 0x44, 0x82, 0x82, 0x0}},
    {5817, 1, {0x0, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {5818, 1, {0x0, 0x82, 0x82, 0x82, 0x82, 0x82, 0xE2, 0x92, 0x8E, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x0}},
    {5819, 1, {0x0, 0x82, 0x82, 0x82, 0x82, 0xE2, 0x92, 0x8E, 0xE2, 0x92, 0x8E, 0x82, 0x82, 0x82, 0x82, 0x0}},
    {5820, 1, {0x0, 0x10, 0x10, 0x10, 0x92, 0x54, 0x38, 0x10, 0x38, 0x54, 0x92, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5821, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x38, 0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5822, 1, {0x0, 0x10, 0x10, 0x10, 0x90, 0x50, 0x30, 0x10, 0x18, 0x14, 0x12, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5823, 1, {0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x30, 0x28, 0x24, 0x22, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0}},
    {5824, 1, {0x0, 0x10, 0x10, 0x10, 0x90, 0x50, 0x38, 0x38, 0x38, 0x14, 0x12, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5825, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5826, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x38, 0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5827, 1, {0x0, 0x0, 0x0, 0x10, 0x20, 0x40, 0x88, 0x44, 0x22, 0x4, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {5828, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x38, 0x54, 0x92, 0x92, 0x54, 0x38, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5829, 1, {0x0, 0x10, 0x10, 0x10, 0x12, 0x14, 0x18, 0x10, 0x30, 0x50, 0x90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5830, 1, {0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x18, 0x28, 0x48, 0x88, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {5831, 1, {0x0, 0x10, 0x18, 0x14, 0x12, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x50, 0x30, 0x10, 0x0}},
    {5832, 1, {0x0, 0x82, 0xC4, 0xA8, 0x90, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x90, 0xA8, 0xC4, 0x82, 0x0}},
    {5833, 1, {0x0, 0x92, 0x92, 0x54, 0x54, 0x38, 0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5834, 1, {0x0, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x8, 0x0, 0x0}},
    {5835, 1, {0x0, 0x40, 0x40, 0x40, 0x40, 0x44, 0x4C, 0x54, 0x64, 0x44, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0}},
    {5836, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5837, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x38, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5838, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x92, 0x54, 0x38, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5839, 1, {0x0, 0x10, 0x38, 0x54, 0x92, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5840, 1, {0x0, 0x8, 0x18, 0x28, 0x48, 0x88, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {5841, 1, {0x0, 0x8, 0x18, 0x28, 0x48, 0x88, 0x8, 0x1C, 0x1C, 0x1C, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {5842, 1, {0x0, 0x60, 0x50, 0x48, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x48, 0x50, 0x60, 0x0}},
    {5843, 1, {0x0, 0x40, 0x40, 0x44, 0x48, 0x50, 0x64, 0x48, 0x50, 0x60, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0}},
    {5844, 1, {0x0, 0xC0, 0xA0, 0x90, 0xA8, 0x90, 0xA0, 0xC0, 0xC0, 0xA0, 0x90, 0xA8, 0x90, 0xA0, 0xC0, 0x0}},
    {5845, 1, {0x0, 0x82, 0x84, 0x88, 0x90, 0xA0, 0xC0, 0x80, 0x80, 0xC0, 0xA0, 0x90, 0x88, 0x84, 0x82, 0x0}},
    {5846, 1, {0x0, 0x82, 0xC6, 0xC6, 0xAA, 0xAA, 0x92, 0x92, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x0}},
    {5847, 1, {0x0, 0x82, 0xC6, 0xAA, 0x92, 0xAA, 0xC6, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x0}},
    {5848, 1, {0x0, 0x92, 0x92, 0x54, 0x54, 0x38, 0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5849, 1, {0x0, 0x38, 0x38, 0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5850, 1, {0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0}},
    {5851, 1, {0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x42, 0xE0, 0xE0, 0xE0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0}},
    {5852, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x44, 0x82, 0x82, 0x44, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {5853, 1, {0x0, 0x82, 0x44, 0x28, 0x10, 0x28, 0x44, 0x82, 0x82, 0x44, 0x28, 0x10, 0x28, 0x44, 0x82, 0x0}},
    {5854, 1, {0x0, 0x82, 0xC6, 0xC6, 0xAA, 0xAA, 0x92, 0x92, 0x92, 0x92, 0xAA, 0xAA, 0xC6, 0xC6, 0x82, 0x0}},
    {5855, 1, {0x0, 0x0, 0x10, 0x28, 0x44, 0x82, 0x82, 0x44, 0x28, 0x10, 0x28, 0x44, 0x82, 0x0, 0x0, 0x0}},
    {5857, 1, {0x0, 0x10, 0x10, 0x10, 0x92, 0x54, 0x38, 0x10, 0x38, 0x54, 0x92, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5859, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x38, 0x54, 0x54, 0x92, 0x92, 0x0}},
    {5860, 1, {0x0, 0x92, 0x92, 0x54, 0x54, 0x38, 0x38, 0x54, 0x54, 0x38, 0x38, 0x54, 0x54, 0x92, 0x92, 0x0}},
    {5861, 1, {0x0, 0x82, 0xC6, 0xAA, 0x92, 0xAA, 0xC6, 0x82, 0x82, 0xC6, 0xAA, 0x92, 0xAA, 0xC6, 0x82, 0x0}},
    {5862, 1, {0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x38, 0x54, 0x54, 0x92, 0x92, 0x0}},
    {5863, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5864, 1, {0x0, 0x10, 0x38, 0x54, 0x92, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x92, 0x54, 0x38, 0x10, 0x0}},
    {5865, 1, {0x0, 0x4, 0xC, 0x14, 0x24, 0x44, 0x24, 0x14, 0xC, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0}},
    {5866, 1, {0x0, 0x40, 0x40, 0xE0, 0x40, 0x44, 0x4C, 0x54, 0x64, 0x44, 0x4, 0xE, 0x4, 0x4, 0x4, 0x0}},
    {5867, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x38, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {5868, 1, {0x0, 0x0, 0x70, 0x70, 0x70, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0x70, 0x70, 0x0, 0x0, 0x0}},
    {5869, 1, {0x0, 0x0, 0x0, 0x38, 0x38, 0x10, 0xD6, 0xFE, 0xD6, 0x10, 0x38, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {5870, 1, {0x0, 0x10, 0x18, 0x14, 0x12, 0x10, 0x10, 0x30, 0x50, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {5871, 1, {0x0, 0x92, 0x54, 0x54, 0x38, 0x38, 0x10, 0x10, 0x10, 0x10, 0x38, 0x38, 0x54, 0x54, 0x92, 0x0}},
    {5872, 1, {0x0, 0x10, 0x10, 0x10, 0x38, 0x54, 0x92, 0x92, 0x92, 0x92, 0x92, 0x54, 0x38, 0x10, 0x10, 0x0}},
    {5986, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x0}},
    {6075, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x1}},
    {6081, 1, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x80, 0x80, 0x80, 0xC0, 0xA0, 0xA0, 0xC0, 0x0, 0x0, 0x0}},
    {6082, 1, {0x18, 0x88, 0x70, 0x0, 0xC0, 0xA0, 0x80, 0x80, 0x80, 0xC0, 0xA0, 0xA0, 0xC0, 0x0, 0x0, 0x0}},
    {6083, 1, {0x98, 0x64, 0x44, 0x18, 0xC0, 0xA0, 0x80, 0x80, 0x80, 0xC0, 0xA0, 0xA0, 0xC0, 0x0, 0x0, 0x0}},
    {6091, 1, {0x0, 0x1, 0x1, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {6102, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x30, 0x0, 0x78, 0x0, 0x30, 0x48, 0x30, 0x0, 0x0, 0x0}},
    {6108, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x20, 0x1C, 0x2, 0x12, 0x24, 0x18, 0x0, 0x0, 0x0}},
    {7680, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x44, 0x82, 0x82, 0xD6, 0x82, 0x82, 0x82, 0x92, 0x28, 0x10}},
    {7681, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x46, 0x3A, 0x8, 0x14, 0x8}},
    {7682, 1, {0x0, 0x10, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x0}},
    {7683, 1, {0x0, 0x0, 0x0, 0x48, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x0}},
    {7684, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x10}},
    {7685, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x8}},
    {7686, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x42, 0x42, 0x42, 0x42, 0x7C, 0x0, 0x3C}},
    {7687, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x0, 0x3C}},
    {7688, 1, {0x8, 0x10, 0x20, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x8, 0x30}},
    {7689, 1, {0x0, 0x0, 0x8, 0x10, 0x20, 0x0, 0x3C, 0x42, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3C, 0x8, 0x30}},
    {7690, 1, {0x0, 0x10, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {7691, 1, {0x0, 0x0, 0x0, 0x12, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7692, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x10}},
    {7693, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x10}},
    {7694, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x0, 0x3C}},
    {7695, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x3C}},
    {7696, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x10, 0x20}},
    {7697, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x8, 0x10}},
    {7698, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x78, 0x30, 0x48}},
    {7699, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x18, 0x24}},
    {7700, 1, {0x10, 0x8, 0x3C, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7701, 1, {0x20, 0x10, 0x8, 0x0, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7702, 1, {0x8, 0x10, 0x3C, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7703, 1, {0x8, 0x10, 0x20, 0x0, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7704, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x18, 0x24}},
    {7705, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x18, 0x24}},
    {7706, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x32, 0x4C}},
    {7707, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x32, 0x4C}},
    {7708, 1, {0x0, 0x22, 0x1C, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x4, 0x18}},
    {7709, 1, {0x0, 0x0, 0x22, 0x1C, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x8, 0x30}},
    {7710, 1, {0x0, 0x10, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {7711, 1, {0x8, 0x0, 0x0, 0xC, 0x10, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {7712, 1, {0x0, 0x0, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x40, 0x4E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7713, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x2, 0x3A, 0x44, 0x44, 0x44, 0x38, 0x20, 0x3C, 0x42, 0x42, 0x3C}},
    {7714, 1, {0x0, 0x0, 0x10, 0x0, 0x44, 0x44, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7715, 1, {0x0, 0x0, 0x0, 0x48, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7716, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x44, 0x0, 0x10}},
    {7717, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x8}},
    {7718, 1, {0x0, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7719, 1, {0x24, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7720, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x62, 0x62, 0x10, 0x60}},
    {7721, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x62, 0x10, 0x60}},
    {7722, 1, {0x0, 0x0, 0x0, 0x0, 0x82, 0x82, 0x82, 0x82, 0xFE, 0x82, 0x82, 0x82, 0x82, 0x82, 0x28, 0x38}},
    {7723, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x4E, 0x51, 0x61, 0x41, 0x41, 0x41, 0x41, 0x41, 0x14, 0x1C}},
    {7724, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x32, 0x4C}},
    {7725, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x32, 0x4C}},
    {7726, 1, {0x4, 0x8, 0x2A, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7727, 1, {0x0, 0x2, 0x4, 0x8, 0x22, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7728, 1, {0x0, 0x4, 0x8, 0x10, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {7729, 1, {0x0, 0x4, 0x8, 0x10, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {7730, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x10}},
    {7731, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x10}},
    {7732, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x3C}},
    {7733, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x0, 0x3C}},
    {7734, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x10}},
    {7735, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x8}},
    {7736, 1, {0x0, 0x0, 0x3C, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x10}},
    {7737, 1, {0x0, 0x0, 0x3C, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x8}},
    {7738, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x3C}},
    {7739, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x1C}},
    {7740, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x18, 0x24}},
    {7741, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x18, 0x24}},
    {7742, 1, {0x0, 0x4, 0x8, 0x10, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7743, 1, {0x0, 0x0, 0x4, 0x8, 0x10, 0x0, 0x76, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {7744, 1, {0x0, 0x0, 0x10, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7745, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x76, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x0, 0x0}},
    {7746, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x66, 0x66, 0x5A, 0x5A, 0x42, 0x42, 0x42, 0x42, 0x0, 0x8}},
    {7747, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x76, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x0, 0x8}},
    {7748, 1, {0x0, 0x0, 0x8, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x0}},
    {7749, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7750, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x8}},
    {7751, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x8}},
    {7752, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x0, 0x3C}},
    {7753, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x3C}},
    {7754, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x62, 0x62, 0x52, 0x52, 0x4A, 0x4A, 0x46, 0x46, 0x42, 0x18, 0x24}},
    {7755, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x18, 0x24}},
    {7756, 1, {0x4, 0x8, 0x32, 0x4C, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7757, 1, {0x2, 0x4, 0x8, 0x32, 0x4C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7758, 1, {0x24, 0x0, 0x32, 0x4C, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7759, 1, {0x0, 0x24, 0x0, 0x32, 0x4C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7760, 1, {0x10, 0x8, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7761, 1, {0x20, 0x10, 0x8, 0x0, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7762, 1, {0x8, 0x10, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7763, 1, {0x8, 0x10, 0x20, 0x0, 0x3C, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7764, 1, {0x0, 0x4, 0x8, 0x10, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {7765, 1, {0x0, 0x0, 0x4, 0x8, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {7766, 1, {0x0, 0x10, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {7767, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {7768, 1, {0x0, 0x10, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {7769, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {7770, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x8}},
    {7771, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x8, 0x0}},
    {7772, 1, {0x0, 0x0, 0x3C, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x8}},
    {7773, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x8, 0x0}},
    {7774, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x48, 0x44, 0x44, 0x42, 0x42, 0x0, 0x3C}},
    {7775, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x3C}},
    {7776, 1, {0x0, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7777, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {7778, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7779, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x8}},
    {7780, 1, {0x24, 0x8, 0x10, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7781, 1, {0x0, 0x20, 0x8, 0x10, 0x20, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {7782, 1, {0x8, 0x22, 0x14, 0x8, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7783, 1, {0x0, 0x8, 0x0, 0x22, 0x14, 0x8, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {7784, 1, {0x0, 0x8, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7785, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x3C, 0x42, 0x40, 0x30, 0xC, 0x2, 0x42, 0x3C, 0x0, 0x8}},
    {7786, 1, {0x0, 0x8, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {7787, 1, {0x0, 0x10, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7788, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8}},
    {7789, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x10}},
    {7790, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x3E}},
    {7791, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x3C}},
    {7792, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8, 0x14, 0x22}},
    {7793, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x12, 0xC, 0x14, 0x22}},
    {7794, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x24}},
    {7795, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x28}},
    {7796, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x32, 0x4C}},
    {7797, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x32, 0x4C}},
    {7798, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x14, 0x22}},
    {7799, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x14, 0x22}},
    {7800, 1, {0x4, 0x8, 0x32, 0x4C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7801, 1, {0x2, 0x4, 0x8, 0x32, 0x4C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7802, 1, {0x24, 0x0, 0x3C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7803, 1, {0x0, 0x0, 0x24, 0x0, 0x3C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7804, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {7805, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x24, 0x24, 0x24, 0x18, 0x18, 0x0, 0x0}},
    {7806, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x8}},
    {7807, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x28, 0x28, 0x28, 0x10, 0x10, 0x0, 0x10}},
    {7808, 1, {0x0, 0x20, 0x10, 0x8, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x5A, 0x66, 0x66, 0x42, 0x42, 0x0, 0x0}},
    {7809, 1, {0x0, 0x0, 0x10, 0x8, 0x4, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {7810, 1, {0x0, 0x4, 0x8, 0x10, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x5A, 0x66, 0x66, 0x42, 0x42, 0x0, 0x0}},
    {7811, 1, {0x0, 0x0, 0x4, 0x8, 0x10, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {7812, 1, {0x0, 0x24, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x5A, 0x66, 0x66, 0x42, 0x42, 0x0, 0x0}},
    {7813, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {7814, 1, {0x0, 0x10, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x54, 0x54, 0x6C, 0x6C, 0x44, 0x44, 0x0, 0x0}},
    {7815, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {7816, 1, {0x0, 0x0, 0x0, 0x0, 0x44, 0x44, 0x44, 0x44, 0x54, 0x54, 0x6C, 0x6C, 0x44, 0x44, 0x0, 0x10}},
    {7817, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x8}},
    {7818, 1, {0x0, 0x10, 0x0, 0x0, 0x44, 0x44, 0x28, 0x28, 0x10, 0x10, 0x28, 0x28, 0x44, 0x44, 0x0, 0x0}},
    {7819, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x44, 0x44, 0x28, 0x10, 0x10, 0x28, 0x44, 0x44, 0x0, 0x0}},
    {7820, 1, {0x0, 0x24, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {7821, 1, {0x0, 0x0, 0x0, 0x24, 0x0, 0x0, 0x42, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {7822, 1, {0x0, 0x8, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {7823, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {7824, 1, {0x8, 0x14, 0x22, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7825, 1, {0x0, 0x8, 0x14, 0x22, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x0}},
    {7826, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x8}},
    {7827, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x8}},
    {7828, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x40, 0x7E, 0x0, 0x3C}},
    {7829, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x7E, 0x0, 0x3C}},
    {7830, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x3C}},
    {7831, 1, {0x0, 0x0, 0x44, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7832, 1, {0x0, 0x8, 0x14, 0x8, 0x0, 0x0, 0x41, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {7833, 1, {0x0, 0x8, 0x14, 0x8, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {7834, 1, {0x0, 0x2, 0x1, 0x1, 0x2, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7835, 1, {0x0, 0x10, 0x0, 0xC, 0x10, 0x10, 0x10, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {7840, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x10}},
    {7841, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x10}},
    {7842, 1, {0x40, 0x20, 0x20, 0x40, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7843, 1, {0x0, 0x8, 0x4, 0x4, 0x8, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7844, 1, {0x2, 0x14, 0x28, 0x44, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7845, 1, {0x0, 0x2, 0x14, 0x28, 0x44, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7846, 1, {0x40, 0x28, 0x14, 0x22, 0x8, 0x14, 0x14, 0x22, 0x22, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {7847, 1, {0x0, 0x40, 0x28, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7848, 1, {0x4, 0x2, 0x14, 0x28, 0x54, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7849, 1, {0x2, 0x1, 0xA, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7850, 1, {0x19, 0x26, 0x8, 0x14, 0x22, 0x8, 0x14, 0x14, 0x22, 0x22, 0x3E, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {7851, 1, {0x19, 0x26, 0x8, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7852, 1, {0x0, 0x8, 0x14, 0x22, 0x8, 0x14, 0x14, 0x22, 0x22, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x0, 0x8}},
    {7853, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x8}},
    {7854, 1, {0x8, 0x10, 0x44, 0x38, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7855, 1, {0x4, 0x8, 0x10, 0x44, 0x38, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7856, 1, {0x20, 0x10, 0x44, 0x38, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7857, 1, {0x20, 0x10, 0x8, 0x22, 0x1C, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7858, 1, {0x10, 0x8, 0xB2, 0x7C, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7859, 1, {0x8, 0x8, 0x10, 0x44, 0x38, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7860, 1, {0x32, 0x4C, 0x82, 0x7C, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x0}},
    {7861, 1, {0x32, 0x4C, 0x0, 0x42, 0x3C, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7862, 1, {0x0, 0x0, 0x44, 0x38, 0x10, 0x28, 0x28, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x0, 0x10}},
    {7863, 1, {0x0, 0x0, 0x0, 0x24, 0x18, 0x0, 0x3C, 0x42, 0x2, 0x3E, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x10}},
    {7864, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x10}},
    {7865, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x10}},
    {7866, 1, {0x10, 0x8, 0x8, 0x10, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7867, 1, {0x10, 0x8, 0x8, 0x10, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7868, 1, {0x0, 0x32, 0x4C, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7869, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7870, 1, {0x2, 0x14, 0x28, 0x44, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7871, 1, {0x0, 0x2, 0x14, 0x28, 0x44, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7872, 1, {0x40, 0x28, 0x14, 0x22, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7873, 1, {0x0, 0x40, 0x28, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7874, 1, {0x4, 0x12, 0x2A, 0x44, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7875, 1, {0x4, 0x2, 0x12, 0x2C, 0x44, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7876, 1, {0x32, 0x4C, 0x10, 0x28, 0x44, 0x7E, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7877, 1, {0x32, 0x4C, 0x10, 0x28, 0x44, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7878, 1, {0x0, 0x8, 0x14, 0x22, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x8}},
    {7879, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x42, 0x7E, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x8}},
    {7880, 1, {0x8, 0x4, 0x4, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7881, 1, {0x0, 0x10, 0x8, 0x8, 0x10, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7882, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x8}},
    {7883, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x8}},
    {7884, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7885, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7886, 1, {0x10, 0x8, 0x8, 0x10, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7887, 1, {0x0, 0x10, 0x8, 0x8, 0x10, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7888, 1, {0x2, 0x14, 0x28, 0x44, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7889, 1, {0x0, 0x2, 0x14, 0x28, 0x44, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7890, 1, {0x40, 0x28, 0x14, 0x22, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7891, 1, {0x0, 0x40, 0x28, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7892, 1, {0x4, 0x12, 0x2A, 0x44, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7893, 1, {0x4, 0x2, 0x12, 0x2C, 0x44, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7894, 1, {0x32, 0x4C, 0x10, 0x28, 0x44, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7895, 1, {0x32, 0x4C, 0x10, 0x28, 0x44, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7896, 1, {0x0, 0x10, 0x28, 0x44, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7897, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7898, 1, {0x8, 0x10, 0x22, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7899, 1, {0x0, 0x0, 0x8, 0x10, 0x22, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7900, 1, {0x20, 0x10, 0xA, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7901, 1, {0x0, 0x0, 0x20, 0x10, 0xA, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7902, 1, {0x10, 0x8, 0xA, 0x11, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7903, 1, {0x0, 0x10, 0x8, 0x8, 0x12, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7904, 1, {0x0, 0x32, 0x4C, 0x2, 0x39, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7905, 1, {0x0, 0x32, 0x4C, 0x0, 0x2, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7906, 1, {0x0, 0x0, 0x2, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x10}},
    {7907, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x1, 0x39, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x10}},
    {7908, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x8}},
    {7909, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x8}},
    {7910, 1, {0x10, 0x8, 0x8, 0x10, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {7911, 1, {0x0, 0x10, 0x8, 0x8, 0x10, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x0}},
    {7912, 1, {0x4, 0x8, 0x12, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7913, 1, {0x0, 0x0, 0x4, 0x8, 0x12, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {7914, 1, {0x40, 0x20, 0x12, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7915, 1, {0x0, 0x0, 0x40, 0x20, 0x12, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {7916, 1, {0x10, 0x8, 0xA, 0x11, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7917, 1, {0x0, 0x10, 0x8, 0x8, 0x12, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {7918, 1, {0x32, 0x4C, 0x2, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {7919, 1, {0x0, 0x32, 0x4C, 0x0, 0x2, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x0}},
    {7920, 1, {0x0, 0x0, 0x2, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x0, 0x10}},
    {7921, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x1, 0x45, 0x46, 0x44, 0x44, 0x44, 0x44, 0x4C, 0x34, 0x0, 0x10}},
    {7922, 1, {0x20, 0x10, 0x8, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {7923, 1, {0x0, 0x20, 0x10, 0x8, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {7924, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8}},
    {7925, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x3C, 0x8}},
    {7926, 1, {0x8, 0x4, 0x4, 0x8, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {7927, 1, {0x0, 0x10, 0x8, 0x8, 0x10, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {7928, 1, {0x0, 0x32, 0x4C, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {7929, 1, {0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1A, 0x2, 0x2, 0x3C}},
    {7936, 1, {0x0, 0x0, 0x30, 0x10, 0x20, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7937, 1, {0x0, 0x0, 0x30, 0x20, 0x10, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7938, 1, {0x0, 0x0, 0x68, 0x28, 0x44, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7939, 1, {0x0, 0x0, 0x68, 0x48, 0x24, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7940, 1, {0x0, 0x0, 0x64, 0x28, 0x48, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7941, 1, {0x0, 0x0, 0x64, 0x48, 0x28, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7942, 1, {0x64, 0x98, 0x0, 0x30, 0x10, 0x20, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7943, 1, {0x64, 0x98, 0x0, 0x30, 0x20, 0x10, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {7944, 1, {0xC0, 0x40, 0x80, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7945, 1, {0xC0, 0x80, 0x40, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7946, 1, {0xD0, 0x50, 0x88, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7947, 1, {0xD0, 0x90, 0x48, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7948, 1, {0xC8, 0x50, 0x90, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7949, 1, {0xC8, 0x90, 0x50, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7950, 1, {0x64, 0x98, 0x60, 0x20, 0x40, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7951, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7952, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7953, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7954, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7955, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7956, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7957, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {7960, 1, {0xC0, 0x40, 0x80, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7961, 1, {0xC0, 0x80, 0x40, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7962, 1, {0xD0, 0x50, 0x88, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7963, 1, {0xD0, 0x90, 0x48, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7964, 1, {0xC8, 0x50, 0x90, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7965, 1, {0xC8, 0x90, 0x50, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {7968, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7969, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7970, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7971, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7972, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7973, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7974, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7975, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {7976, 1, {0xC0, 0x40, 0x80, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7977, 1, {0xC0, 0x80, 0x40, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7978, 1, {0xD0, 0x50, 0x88, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7979, 1, {0xD0, 0x90, 0x48, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7980, 1, {0xC8, 0x50, 0x90, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7981, 1, {0xC8, 0x90, 0x50, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7982, 1, {0x64, 0x98, 0x60, 0x20, 0x40, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7983, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {7984, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7985, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7986, 1, {0x0, 0x0, 0x68, 0x28, 0x44, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7987, 1, {0x0, 0x0, 0x68, 0x48, 0x24, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7988, 1, {0x0, 0x0, 0x64, 0x28, 0x48, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7989, 1, {0x0, 0x0, 0x64, 0x48, 0x28, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7990, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7991, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {7992, 1, {0xC0, 0x40, 0x80, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7993, 1, {0xC0, 0x80, 0x60, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7994, 1, {0xD0, 0x50, 0x88, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7995, 1, {0xD0, 0x90, 0x48, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7996, 1, {0xC8, 0x50, 0x90, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7997, 1, {0xC8, 0x90, 0x50, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7998, 1, {0x64, 0x98, 0x60, 0x20, 0x40, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {7999, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8000, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8001, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8002, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8003, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8004, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8005, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8008, 1, {0xC0, 0x40, 0x80, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8009, 1, {0xC0, 0x80, 0x40, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8010, 1, {0xD0, 0x50, 0x88, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8011, 1, {0xD0, 0x90, 0x48, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8012, 1, {0xC8, 0x50, 0x90, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8013, 1, {0xC8, 0x90, 0x50, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8016, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8017, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8018, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8019, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8020, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8021, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8022, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8023, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8025, 1, {0xC0, 0x80, 0x40, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8027, 1, {0xD0, 0x90, 0x48, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8029, 1, {0xC8, 0x90, 0x50, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8031, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8032, 1, {0x0, 0x0, 0xC, 0x4, 0x8, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8033, 1, {0x0, 0x0, 0xC, 0x8, 0x4, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8034, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8035, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8036, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8037, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8038, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x0, 0x22, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8039, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x0, 0x22, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8040, 1, {0xC0, 0x40, 0x80, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8041, 1, {0xC0, 0x80, 0x40, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8042, 1, {0xD0, 0x50, 0x88, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8043, 1, {0xD0, 0x90, 0x48, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8044, 1, {0xC8, 0x50, 0x90, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8045, 1, {0xC8, 0x90, 0x50, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8046, 1, {0x32, 0x4C, 0x18, 0x8, 0x10, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8047, 1, {0x32, 0x4C, 0x18, 0x10, 0x8, 0x1C, 0x23, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8048, 1, {0x0, 0x0, 0x20, 0x20, 0x10, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {8049, 1, {0x0, 0x0, 0x10, 0x20, 0x20, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {8050, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {8051, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x0, 0x3C, 0x42, 0x40, 0x3C, 0x40, 0x40, 0x42, 0x3C, 0x0, 0x0}},
    {8052, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {8053, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {8054, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8055, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8056, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8057, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8058, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8059, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8060, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8061, 1, {0x0, 0x0, 0x4, 0x8, 0x8, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8064, 1, {0x0, 0x0, 0x30, 0x10, 0x20, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8065, 1, {0x0, 0x0, 0x30, 0x20, 0x10, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8066, 1, {0x0, 0x0, 0x68, 0x28, 0x44, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8067, 1, {0x0, 0x0, 0x68, 0x48, 0x24, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8068, 1, {0x0, 0x0, 0x64, 0x28, 0x48, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8069, 1, {0x0, 0x0, 0x64, 0x48, 0x28, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8070, 1, {0x32, 0x4C, 0x0, 0x30, 0x10, 0x20, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8071, 1, {0x32, 0x4C, 0x0, 0x30, 0x20, 0x10, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8072, 1, {0xC0, 0x40, 0x80, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8073, 1, {0xC0, 0x80, 0x40, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8074, 1, {0xD0, 0x50, 0x88, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8075, 1, {0xD0, 0x90, 0x48, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8076, 1, {0xC8, 0x50, 0x90, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8077, 1, {0xC8, 0x90, 0x50, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8078, 1, {0x64, 0x98, 0x60, 0x20, 0x40, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8079, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8080, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8081, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8082, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8083, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8084, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8085, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8086, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8087, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8088, 1, {0xC0, 0x40, 0x80, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8089, 1, {0xC0, 0x80, 0x40, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8090, 1, {0xD0, 0x50, 0x88, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8091, 1, {0xD0, 0x90, 0x48, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8092, 1, {0xC8, 0x50, 0x90, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8093, 1, {0xC8, 0x90, 0x50, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8094, 1, {0x64, 0x98, 0x60, 0x20, 0x40, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8095, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x20, 0x18}},
    {8096, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8097, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8098, 1, {0x0, 0x0, 0x34, 0x14, 0x22, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8099, 1, {0x0, 0x0, 0x34, 0x24, 0x12, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8100, 1, {0x0, 0x0, 0x32, 0x14, 0x24, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8101, 1, {0x0, 0x0, 0x32, 0x24, 0x14, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8102, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x0, 0x22, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8103, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x0, 0x22, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8104, 1, {0xC0, 0x40, 0x80, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8105, 1, {0xC0, 0x80, 0x40, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8106, 1, {0xD0, 0x50, 0x88, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8107, 1, {0xD0, 0x90, 0x48, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8108, 1, {0xC8, 0x50, 0x90, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8109, 1, {0xC8, 0x90, 0x50, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8110, 1, {0x64, 0x98, 0x60, 0x20, 0x40, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8111, 1, {0x64, 0x98, 0x60, 0x40, 0x20, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8112, 1, {0x0, 0x0, 0x44, 0x44, 0x38, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {8113, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {8114, 1, {0x0, 0x0, 0x20, 0x20, 0x10, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8115, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8116, 1, {0x0, 0x0, 0x10, 0x20, 0x20, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8118, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {8119, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x20, 0x30}},
    {8120, 1, {0x42, 0x42, 0x3C, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8121, 1, {0x0, 0x3C, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8122, 1, {0x80, 0x80, 0x40, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8123, 1, {0x40, 0x80, 0x80, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8124, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8125, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8126, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0xC}},
    {8127, 1, {0x18, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8128, 1, {0x32, 0x4C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8129, 1, {0x32, 0x4C, 0x0, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8130, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8131, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8132, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8134, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x2, 0x2}},
    {8135, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x5C, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x22, 0x32}},
    {8136, 1, {0x80, 0x80, 0x40, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {8137, 1, {0x40, 0x80, 0x80, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {8138, 1, {0x80, 0x80, 0x40, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8139, 1, {0x40, 0x80, 0x80, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8140, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x10, 0xC}},
    {8141, 1, {0x68, 0x28, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8142, 1, {0x64, 0x28, 0x48, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8143, 1, {0x32, 0x4C, 0x0, 0x18, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8144, 1, {0x0, 0x0, 0x44, 0x44, 0x38, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8145, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8146, 1, {0x10, 0x10, 0x8, 0x0, 0x66, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8147, 1, {0x8, 0x10, 0x10, 0x0, 0x66, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8150, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8151, 1, {0x0, 0x32, 0x4C, 0x0, 0x66, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xC, 0x0, 0x0}},
    {8152, 1, {0x22, 0x22, 0x1C, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8153, 1, {0x0, 0x3E, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8154, 1, {0x80, 0x80, 0x40, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8155, 1, {0x40, 0x80, 0x80, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8157, 1, {0x68, 0x48, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8158, 1, {0x64, 0x48, 0x28, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8159, 1, {0x32, 0x4C, 0x0, 0x18, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8160, 1, {0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8161, 1, {0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8162, 1, {0x10, 0x10, 0x8, 0x0, 0x66, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8163, 1, {0x8, 0x10, 0x10, 0x0, 0x66, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8164, 1, {0x0, 0x0, 0x18, 0x8, 0x10, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {8165, 1, {0x0, 0x0, 0x18, 0x10, 0x8, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {8166, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8167, 1, {0x0, 0x32, 0x4C, 0x0, 0x66, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8168, 1, {0x22, 0x22, 0x1C, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8169, 1, {0x0, 0x0, 0x3E, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8170, 1, {0x80, 0x80, 0x40, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8171, 1, {0x40, 0x80, 0x80, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8172, 1, {0xC0, 0x80, 0x40, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {8173, 1, {0x10, 0x10, 0x8, 0x0, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8174, 1, {0x8, 0x10, 0x10, 0x0, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8175, 1, {0x10, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8178, 1, {0x0, 0x0, 0x10, 0x10, 0x8, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8179, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8180, 1, {0x0, 0x0, 0x4, 0x8, 0x8, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8182, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8183, 1, {0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x22, 0x41, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x8, 0xC}},
    {8184, 1, {0x80, 0x80, 0x40, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8185, 1, {0x40, 0x80, 0x80, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8186, 1, {0x80, 0x80, 0x40, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8187, 1, {0x40, 0x80, 0x80, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x0, 0x0}},
    {8188, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x14, 0x77, 0x8, 0x6}},
    {8189, 1, {0x8, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8190, 1, {0x18, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8192, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8193, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8194, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8195, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8196, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8197, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8198, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8199, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8200, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8201, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8202, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8208, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8210, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8211, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8212, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8213, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8214, 1, {0x0, 0x0, 0x0, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x0, 0x0}},
    {8215, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0xFF}},
    {8216, 1, {0x0, 0x0, 0x8, 0x10, 0x10, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8217, 1, {0x0, 0x0, 0x18, 0x8, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8218, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10}},
    {8219, 1, {0x0, 0x0, 0x18, 0x10, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8220, 1, {0x0, 0x0, 0x22, 0x44, 0x44, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8221, 1, {0x0, 0x0, 0x66, 0x22, 0x22, 0x44, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8222, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x22, 0x22, 0x44}},
    {8223, 1, {0x0, 0x0, 0x66, 0x44, 0x44, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8224, 1, {0x0, 0x0, 0x8, 0x1C, 0x8, 0x2A, 0x7F, 0x2A, 0x8, 0x1C, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8225, 1, {0x0, 0x8, 0x1C, 0x8, 0x2A, 0x7F, 0x2A, 0x8, 0x8, 0x2A, 0x7F, 0x2A, 0x8, 0x1C, 0x8, 0x0}},
    {8226, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x7C, 0x7C, 0x7C, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {8227, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x30, 0x38, 0x3C, 0x38, 0x30, 0x20, 0x0, 0x0, 0x0}},
    {8228, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0}},
    {8229, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {8230, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x0, 0x0}},
    {8231, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8240, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0xA4, 0xA4, 0x48, 0x10, 0x10, 0x2A, 0x55, 0x55, 0x8A, 0x0, 0x0}},
    {8241, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0xA0, 0xA0, 0x4C, 0x30, 0xC0, 0x2A, 0x55, 0x55, 0x2A, 0x0, 0x0}},
    {8242, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x18, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8243, 1, {0x0, 0x0, 0x0, 0x0, 0x33, 0x66, 0x44, 0x88, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8244, 1, {0x0, 0x0, 0x0, 0x0, 0x3F, 0x7E, 0x54, 0xA8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8245, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x18, 0x8, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8246, 1, {0x0, 0x0, 0x0, 0x0, 0xCC, 0x66, 0x22, 0x11, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8247, 1, {0x0, 0x0, 0x0, 0x0, 0xFC, 0x7E, 0x2A, 0x15, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8248, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x44}},
    {8249, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x4, 0x8, 0x8, 0x4, 0x4, 0x2, 0x2, 0x0, 0x0}},
    {8250, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x20, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {8251, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x22, 0x14, 0x49, 0x14, 0x22, 0x49, 0x0, 0x0, 0x0, 0x0}},
    {8252, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x0, 0x24, 0x24, 0x0, 0x0}},
    {8253, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x4A, 0x4A, 0xA, 0xC, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0, 0x0}},
    {8254, 1, {0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8255, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x3C}},
    {8256, 1, {0x3C, 0x42, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8257, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x28, 0x44}},
    {8258, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x8, 0x0, 0x22, 0x77, 0x22, 0x0, 0x0, 0x0}},
    {8259, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8260, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {8261, 1, {0x0, 0x0, 0x0, 0xE, 0x8, 0x8, 0x8, 0x8, 0xE, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x0}},
    {8262, 1, {0x0, 0x0, 0x0, 0x70, 0x10, 0x10, 0x10, 0x10, 0x70, 0x10, 0x10, 0x10, 0x10, 0x10, 0x70, 0x0}},
    {8263, 1, {0x0, 0x0, 0x0, 0xEE, 0x11, 0x11, 0x22, 0xCC, 0x88, 0x88, 0x88, 0x0, 0x0, 0x88, 0x0, 0x0}},
    {8264, 1, {0x0, 0x0, 0x0, 0x72, 0xA, 0xA, 0x12, 0x62, 0x42, 0x42, 0x42, 0x0, 0x0, 0x42, 0x0, 0x0}},
    {8265, 1, {0x0, 0x0, 0x0, 0x5C, 0x42, 0x42, 0x44, 0x58, 0x50, 0x50, 0x50, 0x0, 0x0, 0x50, 0x0, 0x0}},
    {8266, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {8267, 1, {0x0, 0x0, 0x0, 0x0, 0xFC, 0x5E, 0x5E, 0x5E, 0x5C, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x0}},
    {8268, 1, {0x0, 0x0, 0x0, 0x0, 0x1E, 0x3A, 0x3A, 0x3A, 0x1E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8269, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x5C, 0x5C, 0x5C, 0x78, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8270, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x49, 0x2A, 0x1C, 0x2A, 0x49, 0x8, 0x0, 0x0}},
    {8271, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x18, 0x10, 0x10, 0x8, 0x0}},
    {8272, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x0, 0x0, 0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0}},
    {8273, 1, {0x0, 0x8, 0x49, 0x2A, 0x1C, 0x2A, 0x49, 0x8, 0x0, 0x8, 0x49, 0x2A, 0x1C, 0x2A, 0x49, 0x8}},
    {8274, 1, {0x0, 0x0, 0x0, 0x0, 0x62, 0x62, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x46, 0x46, 0x0, 0x0}},
    {8287, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8304, 1, {0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8305, 1, {0x0, 0x0, 0x8, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8308, 1, {0x0, 0x0, 0x4, 0xC, 0x14, 0x24, 0x3E, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8309, 1, {0x0, 0x0, 0x3E, 0x20, 0x20, 0x3C, 0x2, 0x2, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8310, 1, {0x0, 0x0, 0x1C, 0x22, 0x20, 0x3C, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8311, 1, {0x0, 0x0, 0x3E, 0x2, 0x2, 0x4, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8312, 1, {0x0, 0x0, 0x1C, 0x22, 0x22, 0x1C, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8313, 1, {0x0, 0x0, 0x1C, 0x22, 0x22, 0x1E, 0x2, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8314, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8315, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8316, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8317, 1, {0x0, 0x0, 0x4, 0x8, 0x8, 0x8, 0x8, 0x8, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8318, 1, {0x0, 0x0, 0x10, 0x8, 0x8, 0x8, 0x8, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8319, 1, {0x0, 0x0, 0x0, 0x0, 0x2C, 0x32, 0x22, 0x22, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8320, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x22, 0x22, 0x1C}},
    {8321, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x18, 0x28, 0x8, 0x8, 0x8, 0x8}},
    {8322, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x2, 0x1C, 0x20, 0x20, 0x3E}},
    {8323, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x2, 0x1C, 0x2, 0x22, 0x1C}},
    {8324, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0xC, 0x14, 0x24, 0x3E, 0x4, 0x4}},
    {8325, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x20, 0x20, 0x3C, 0x2, 0x2, 0x3C}},
    {8326, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x20, 0x3C, 0x22, 0x22, 0x1C}},
    {8327, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x2, 0x2, 0x4, 0x8, 0x8, 0x8}},
    {8328, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x1C, 0x22, 0x22, 0x1C}},
    {8329, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x1E, 0x2, 0x22, 0x1C}},
    {8330, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x0}},
    {8331, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x0, 0x0}},
    {8332, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x3E, 0x0, 0x0}},
    {8333, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x8, 0x8, 0x8, 0x8, 0x8, 0x4}},
    {8334, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x8, 0x8, 0x8, 0x8, 0x8, 0x10}},
    {8352, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x40, 0x40, 0x5E, 0x50, 0x50, 0x3C, 0x10, 0x10, 0x1E, 0x0, 0x0}},
    {8353, 1, {0x0, 0x0, 0x0, 0x0, 0xA, 0x3E, 0x4A, 0x54, 0x54, 0x68, 0x68, 0x72, 0x5C, 0x10, 0x0, 0x0}},
    {8354, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x42, 0x40, 0x56, 0x58, 0x50, 0x50, 0x32, 0x1C, 0x0, 0x0}},
    {8355, 1, {0x0, 0x0, 0x0, 0x0, 0x3F, 0x20, 0x20, 0x20, 0x3E, 0x20, 0x20, 0xFC, 0x20, 0x20, 0x0, 0x0}},
    {8356, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x10, 0x10, 0x7E, 0x10, 0x7E, 0x10, 0x10, 0x3E, 0x61, 0x0, 0x0}},
    {8357, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x76, 0x4D, 0x4D, 0x49, 0x49, 0x59, 0x69, 0x69, 0x0, 0x0}},
    {8358, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x32, 0x7F, 0x2A, 0x7F, 0x26, 0x26, 0x22, 0x22, 0x0, 0x0}},
    {8359, 1, {0x0, 0x0, 0x0, 0x0, 0x68, 0x58, 0x5F, 0x5C, 0x6C, 0x4A, 0x49, 0x49, 0x49, 0x4E, 0x0, 0x0}},
    {8360, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x48, 0x4B, 0x4C, 0x74, 0x54, 0x52, 0x49, 0x49, 0x4E, 0x0, 0x0}},
    {8361, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x7F, 0x2A, 0x7F, 0x36, 0x36, 0x22, 0x22, 0x0, 0x0}},
    {8362, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7A, 0x4A, 0x5A, 0x5A, 0x5A, 0x5A, 0x52, 0x5E, 0x0, 0x0}},
    {8363, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x3E, 0x2, 0x3A, 0x46, 0x42, 0x42, 0x42, 0x46, 0x3A, 0x0, 0x3C}},
    {8364, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x20, 0x7C, 0x20, 0x7C, 0x20, 0x20, 0x12, 0xC, 0x0, 0x0}},
    {8365, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x44, 0x48, 0x50, 0x60, 0xFE, 0x50, 0x48, 0x44, 0x42, 0x0, 0x0}},
    {8366, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0xE, 0x38, 0xE, 0x38, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8367, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x48, 0x48, 0x48, 0x48, 0x4E, 0x4D, 0x4D, 0x4E, 0x74, 0x4, 0x4}},
    {8368, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x14, 0xC, 0x2A, 0x32, 0x31, 0x21, 0x2, 0x4}},
    {8369, 1, {0x0, 0x0, 0x7C, 0x42, 0xFF, 0x42, 0xFF, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {8448, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x51, 0x52, 0x34, 0x8, 0x13, 0x24, 0x44, 0x3, 0x0, 0x0}},
    {8449, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x51, 0x52, 0x34, 0xB, 0x14, 0x22, 0x41, 0x6, 0x0, 0x0}},
    {8450, 1, {0x0, 0x0, 0x0, 0x0, 0x1E, 0x31, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x31, 0x1E, 0x0, 0x0}},
    {8451, 1, {0x0, 0x0, 0x0, 0x60, 0x90, 0x90, 0x60, 0xC, 0x12, 0x10, 0x10, 0x10, 0x12, 0xC, 0x0, 0x0}},
    {8452, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x3C, 0x50, 0x50, 0x50, 0x50, 0x3C, 0x10, 0x1E, 0x0, 0x0}},
    {8453, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x41, 0x42, 0x34, 0x8, 0x16, 0x29, 0x49, 0x6, 0x0, 0x0}},
    {8454, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x41, 0x42, 0x34, 0x8, 0x10, 0x29, 0x49, 0x6, 0x0, 0x0}},
    {8455, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x42, 0x40, 0x40, 0x38, 0x40, 0x40, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8456, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x2, 0x12, 0x1E, 0x12, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {8457, 1, {0x0, 0x0, 0x0, 0x60, 0x90, 0x90, 0x60, 0x0, 0x1E, 0x10, 0x10, 0x1C, 0x10, 0x10, 0x0, 0x0}},
    {8458, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3A, 0x46, 0x42, 0x42, 0x3E, 0x2, 0x3F, 0x42, 0x3C, 0x0}},
    {8459, 1, {0x0, 0x0, 0x0, 0x0, 0x12, 0x75, 0x96, 0x14, 0x1C, 0x34, 0x54, 0x94, 0x94, 0x63, 0x0, 0x0}},
    {8460, 1, {0x0, 0x0, 0x0, 0x0, 0x77, 0xB7, 0xB6, 0xB6, 0x36, 0x3E, 0x36, 0x36, 0x76, 0x67, 0x0, 0x0}},
    {8461, 1, {0x0, 0x0, 0x0, 0x0, 0x71, 0x51, 0x51, 0x51, 0x5F, 0x51, 0x51, 0x51, 0x51, 0x71, 0x0, 0x0}},
    {8462, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x58, 0x64, 0x44, 0x44, 0x44, 0x44, 0x46, 0x0, 0x0}},
    {8463, 1, {0x0, 0x0, 0x0, 0x8, 0x50, 0x60, 0x40, 0xD8, 0x64, 0x44, 0x44, 0x44, 0x44, 0x46, 0x0, 0x0}},
    {8464, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x24, 0x4, 0x4, 0x6, 0xC, 0x14, 0x24, 0x28, 0x30, 0x0}},
    {8465, 1, {0x0, 0x0, 0x0, 0x0, 0x3F, 0x7E, 0x44, 0x24, 0x6, 0x6, 0x6, 0x6, 0x76, 0x5C, 0x0, 0x0}},
    {8466, 1, {0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x3C, 0x10, 0x10, 0x10, 0x10, 0x70, 0x99, 0x66, 0x0, 0x0}},
    {8467, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x24, 0x28, 0x30, 0x60, 0x20, 0x20, 0x22, 0x1C, 0x0, 0x0}},
    {8468, 1, {0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0xFE, 0x48, 0x4A, 0x4D, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {8469, 1, {0x0, 0x0, 0x0, 0x0, 0x71, 0x51, 0x59, 0x59, 0x55, 0x55, 0x53, 0x53, 0x51, 0x71, 0x0, 0x0}},
    {8470, 1, {0x0, 0x0, 0x0, 0x0, 0x92, 0x95, 0x95, 0xD2, 0xD0, 0xB7, 0xB0, 0x90, 0x90, 0x90, 0x0, 0x0}},
    {8471, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0xB9, 0xA5, 0xA5, 0xB9, 0xA1, 0xA1, 0x42, 0x3C, 0x0, 0x0}},
    {8472, 1, {0x0, 0x0, 0x0, 0x0, 0x4E, 0x51, 0x51, 0x21, 0x2E, 0x50, 0x50, 0x50, 0x50, 0x60, 0x0, 0x0}},
    {8473, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x52, 0x52, 0x52, 0x52, 0x5C, 0x50, 0x50, 0x50, 0x70, 0x0, 0x0}},
    {8474, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x32, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x36, 0x1E, 0x1, 0x0}},
    {8475, 1, {0x0, 0x0, 0x0, 0x40, 0x2E, 0x11, 0x31, 0x51, 0x51, 0x5E, 0x54, 0x52, 0x52, 0x21, 0x0, 0x0}},
    {8476, 1, {0x0, 0x0, 0x0, 0x0, 0xEE, 0xB3, 0x33, 0x33, 0x3E, 0x3C, 0x36, 0x36, 0x76, 0xE3, 0x0, 0x0}},
    {8477, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x51, 0x51, 0x51, 0x51, 0x5E, 0x58, 0x54, 0x52, 0x71, 0x0, 0x0}},
    {8478, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x50, 0x4A, 0x44, 0x4A, 0x50, 0x0, 0x0}},
    {8479, 1, {0x0, 0x0, 0x38, 0x8, 0x7C, 0x4A, 0x4A, 0x4A, 0x7C, 0x68, 0x58, 0x48, 0x4C, 0x4A, 0x8, 0x0}},
    {8480, 1, {0x0, 0x0, 0x0, 0x0, 0x71, 0x9B, 0x55, 0x31, 0xD1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8481, 1, {0x0, 0x0, 0x0, 0x0, 0xFE, 0x52, 0x5A, 0x52, 0x5F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8482, 1, {0x0, 0x0, 0x0, 0x0, 0xF1, 0x5B, 0x55, 0x51, 0x51, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8483, 1, {0x0, 0x0, 0x30, 0xC, 0x85, 0x89, 0x4A, 0x52, 0x54, 0x24, 0x28, 0x78, 0x50, 0x80, 0x0, 0x0}},
    {8484, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0xA, 0xA, 0x12, 0x14, 0x24, 0x28, 0x48, 0x50, 0x7E, 0x0, 0x0}},
    {8485, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x8, 0x10, 0x3E, 0x4, 0x8, 0x18, 0x4, 0x2, 0x42, 0x3C, 0x0}},
    {8486, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x22, 0x14, 0x55, 0x77, 0x0, 0x0}},
    {8487, 1, {0x0, 0x0, 0x0, 0x0, 0x77, 0x55, 0x14, 0x22, 0x41, 0x41, 0x41, 0x41, 0x41, 0x3E, 0x0, 0x0}},
    {8488, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x7E, 0x6, 0x3C, 0x26, 0x6, 0x6, 0x66, 0x3C, 0x0, 0x0, 0x0}},
    {8489, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0x8, 0x8, 0x8, 0x8, 0xE, 0x8, 0x0, 0x0}},
    {8490, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x48, 0x50, 0x60, 0x60, 0x50, 0x48, 0x44, 0x0, 0x0}},
    {8491, 1, {0x18, 0x24, 0x18, 0x0, 0x18, 0x24, 0x24, 0x42, 0x42, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8492, 1, {0x0, 0x0, 0x0, 0x0, 0x2E, 0x51, 0x51, 0x51, 0x5E, 0x51, 0x51, 0x51, 0x51, 0x2E, 0x0, 0x0}},
    {8493, 1, {0x0, 0x0, 0x0, 0x6, 0x1E, 0x3C, 0x6C, 0x66, 0x66, 0x6C, 0x60, 0x60, 0x66, 0x3C, 0x18, 0x0}},
    {8494, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x66, 0x66, 0x7E, 0x60, 0x60, 0x66, 0x3C, 0x0, 0x0}},
    {8495, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x48, 0x70, 0x40, 0x44, 0x38, 0x0, 0x0}},
    {8496, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x12, 0x20, 0x20, 0x1C, 0x20, 0x40, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {8497, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x2, 0x4, 0x4, 0x8, 0x3E, 0x10, 0x20, 0x20, 0x0, 0x0}},
    {8498, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x3E, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {8499, 1, {0x0, 0x0, 0x0, 0x0, 0x63, 0x36, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x6A, 0xAA, 0x49, 0x0, 0x0}},
    {8500, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x53, 0x4E, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0}},
    {8501, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x46, 0x22, 0x14, 0x28, 0x64, 0x72, 0x62, 0x0, 0x0}},
    {8502, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x3E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x3F, 0x7E, 0x0, 0x0}},
    {8503, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x3C, 0x1E, 0x2, 0x2, 0x2, 0xE, 0x32, 0x32, 0x32, 0x0, 0x0}},
    {8504, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x3E, 0x1F, 0x2, 0x2, 0x2, 0x2, 0x3, 0x2, 0x0, 0x0}},
    {8505, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x38, 0x18, 0x18, 0x18, 0x18, 0x7E, 0x0, 0x0}},
    {8510, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x55, 0x53, 0x51, 0x50, 0x50, 0x50, 0x50, 0x50, 0x70, 0x0, 0x0}},
    {8513, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x72, 0x2, 0x2, 0x2, 0x42, 0x3C, 0x0, 0x0}},
    {8514, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {8515, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {8516, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x28, 0x44, 0x82, 0x82, 0x82, 0x82, 0x0, 0x0}},
    {8519, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x52, 0x9E, 0xA0, 0xA2, 0x7C, 0x0, 0x0}},
    {8520, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0x3E, 0x12, 0x14, 0x24, 0x28, 0x7C, 0x0, 0x0}},
    {8521, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x0, 0x3E, 0x12, 0x14, 0x14, 0x24, 0x28, 0x28, 0xF0}},
    {8522, 1, {0x0, 0x0, 0x10, 0x10, 0x7C, 0x52, 0x52, 0x52, 0x52, 0x5C, 0x10, 0x10, 0x10, 0x1E, 0x0, 0x0}},
    {8523, 1, {0x0, 0x0, 0x0, 0x0, 0x4E, 0x31, 0x21, 0x51, 0x4E, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {8531, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x1C, 0x12, 0x24, 0x42, 0x4C, 0x0, 0x0}},
    {8532, 1, {0x0, 0x0, 0x0, 0x0, 0x62, 0x12, 0x24, 0x48, 0x78, 0x1C, 0x12, 0x24, 0x42, 0x4C, 0x0, 0x0}},
    {8533, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x1E, 0x18, 0x2C, 0x42, 0x4C, 0x0, 0x0}},
    {8534, 1, {0x0, 0x0, 0x0, 0x0, 0x62, 0x12, 0x24, 0x48, 0x78, 0x1E, 0x18, 0x2C, 0x42, 0x4C, 0x0, 0x0}},
    {8535, 1, {0x0, 0x0, 0x0, 0x0, 0x62, 0x12, 0x24, 0x18, 0x68, 0x1E, 0x18, 0x2C, 0x42, 0x4C, 0x0, 0x0}},
    {8536, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0xF4, 0x28, 0x28, 0x1E, 0x18, 0x2C, 0x42, 0x4C, 0x0, 0x0}},
    {8537, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x16, 0x18, 0x2E, 0x4A, 0x4E, 0x0, 0x0}},
    {8538, 1, {0x0, 0x0, 0x0, 0x0, 0x72, 0x42, 0x64, 0x18, 0x68, 0x16, 0x18, 0x2E, 0x4A, 0x4E, 0x0, 0x0}},
    {8539, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x14, 0x1A, 0x24, 0x4A, 0x4E, 0x0, 0x0}},
    {8540, 1, {0x0, 0x0, 0x0, 0x0, 0x62, 0x12, 0x24, 0x18, 0x68, 0x14, 0x1A, 0x24, 0x4A, 0x4E, 0x0, 0x0}},
    {8541, 1, {0x0, 0x0, 0x0, 0x0, 0x72, 0x42, 0x64, 0x18, 0x68, 0x14, 0x1A, 0x24, 0x4A, 0x4E, 0x0, 0x0}},
    {8542, 1, {0x0, 0x0, 0x0, 0x0, 0x72, 0x12, 0x24, 0x48, 0x48, 0x14, 0x1A, 0x24, 0x4A, 0x4E, 0x0, 0x0}},
    {8543, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x62, 0x24, 0x28, 0x28, 0x10, 0x10, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {8544, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {8545, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x7E, 0x0, 0x0}},
    {8546, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8547, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x29, 0x29, 0x29, 0x29, 0x29, 0x26, 0x26, 0x26, 0x7F, 0x0, 0x0}},
    {8548, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8549, 1, {0x0, 0x0, 0x0, 0x0, 0xFF, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x32, 0x32, 0x32, 0x7F, 0x0, 0x0}},
    {8550, 1, {0x0, 0x0, 0x0, 0x0, 0xFF, 0x55, 0x55, 0x55, 0x55, 0x25, 0x25, 0x25, 0x25, 0x7F, 0x0, 0x0}},
    {8551, 1, {0x0, 0x0, 0x0, 0x0, 0xFF, 0xB5, 0xB5, 0xB5, 0xB5, 0x55, 0x55, 0x55, 0x55, 0xFF, 0x0, 0x0}},
    {8552, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x2A, 0x2A, 0x2A, 0x24, 0x24, 0x2A, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8553, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x22, 0x22, 0x22, 0x1C, 0x1C, 0x22, 0x22, 0x22, 0x7F, 0x0, 0x0}},
    {8554, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x2A, 0x2A, 0x2A, 0x12, 0x12, 0x2A, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8555, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x55, 0x55, 0x55, 0x25, 0x25, 0x55, 0x55, 0x55, 0x7F, 0x0, 0x0}},
    {8556, 1, {0x0, 0x0, 0x0, 0x0, 0x70, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x22, 0x7E, 0x0, 0x0}},
    {8557, 1, {0x0, 0x0, 0x0, 0x0, 0x1A, 0x26, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x22, 0x1C, 0x0, 0x0}},
    {8558, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x24, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x24, 0x78, 0x0, 0x0}},
    {8559, 1, {0x0, 0x0, 0x0, 0x0, 0x63, 0x22, 0x36, 0x36, 0x36, 0x2A, 0x2A, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8560, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x8, 0x18, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {8561, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x0, 0x24, 0x6C, 0x24, 0x24, 0x24, 0x24, 0x7E, 0x0, 0x0}},
    {8562, 1, {0x0, 0x0, 0x0, 0x0, 0x2A, 0x2A, 0x0, 0x2A, 0x7E, 0x2A, 0x2A, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8563, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x0, 0x3B, 0x69, 0x29, 0x29, 0x26, 0x26, 0x7F, 0x0, 0x0}},
    {8564, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x77, 0x22, 0x22, 0x22, 0x14, 0x8, 0x3E, 0x0, 0x0}},
    {8565, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x0, 0x5A, 0x4E, 0x4A, 0x4A, 0x32, 0x32, 0x7F, 0x0, 0x0}},
    {8566, 1, {0x0, 0x0, 0x0, 0x0, 0x5, 0x5, 0x0, 0x55, 0x5F, 0x55, 0x55, 0x25, 0x25, 0x7F, 0x0, 0x0}},
    {8567, 1, {0x0, 0x0, 0x0, 0x0, 0x15, 0x15, 0x0, 0xB5, 0xBF, 0xB5, 0x55, 0x55, 0x55, 0xFF, 0x0, 0x0}},
    {8568, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x0, 0x2A, 0x6A, 0x24, 0x24, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8569, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x77, 0x22, 0x14, 0x8, 0x14, 0x22, 0x77, 0x0, 0x0}},
    {8570, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x0, 0x2A, 0x2E, 0x12, 0x12, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8571, 1, {0x0, 0x0, 0x0, 0x0, 0x5, 0x5, 0x0, 0x55, 0x5F, 0x25, 0x25, 0x55, 0x55, 0x7F, 0x0, 0x0}},
    {8572, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {8573, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1A, 0x26, 0x40, 0x40, 0x40, 0x22, 0x1C, 0x0, 0x0}},
    {8574, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x6, 0x2, 0x1A, 0x26, 0x42, 0x42, 0x42, 0x26, 0x1B, 0x0, 0x0}},
    {8575, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x74, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x7F, 0x0, 0x0}},
    {8576, 1, {0x0, 0x0, 0x0, 0x0, 0x2C, 0x5A, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x4A, 0x5A, 0x2C, 0x0, 0x0}},
    {8577, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x24, 0x32, 0x2A, 0x2A, 0x2A, 0x2A, 0x32, 0x24, 0x78, 0x0, 0x0}},
    {8579, 1, {0x0, 0x0, 0x0, 0x0, 0x58, 0x64, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x44, 0x38, 0x0, 0x0}},
    {8592, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x7F, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8593, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {8594, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0xFE, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8595, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {8596, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x42, 0xFF, 0x42, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8597, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {8598, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x60, 0x50, 0x48, 0x4, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8599, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x6, 0xA, 0x12, 0x20, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8600, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x12, 0xA, 0x6, 0x1E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8601, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x48, 0x50, 0x60, 0x78, 0x0, 0x0, 0x0, 0x0}},
    {8602, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x14, 0x24, 0x7F, 0x24, 0x14, 0x4, 0x4, 0x0, 0x0}},
    {8603, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x28, 0x24, 0xFE, 0x24, 0x28, 0x20, 0x20, 0x0, 0x0}},
    {8604, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x26, 0x79, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8605, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x64, 0x9E, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8606, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x48, 0xFF, 0x48, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8607, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {8608, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x12, 0xFF, 0x12, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8609, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x2A, 0x1C, 0x8, 0x0}},
    {8610, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x11, 0x22, 0x7C, 0x22, 0x11, 0x0, 0x0, 0x0, 0x0}},
    {8611, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x88, 0x44, 0x3E, 0x44, 0x88, 0x0, 0x0, 0x0, 0x0}},
    {8612, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x22, 0x7E, 0x22, 0x12, 0x0, 0x0, 0x0, 0x0}},
    {8613, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x0, 0x0}},
    {8614, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x44, 0x7E, 0x44, 0x48, 0x0, 0x0, 0x0, 0x0}},
    {8615, 1, {0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {8616, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x3E, 0x0, 0x0}},
    {8617, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x11, 0x21, 0x7E, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8618, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x44, 0x42, 0x3F, 0x2, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {8619, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x15, 0x25, 0x7E, 0x24, 0x14, 0x0, 0x0, 0x0, 0x0}},
    {8620, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x54, 0x52, 0x3F, 0x12, 0x14, 0x0, 0x0, 0x0, 0x0}},
    {8621, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x5A, 0xE7, 0x42, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8622, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x2C, 0x4A, 0xFF, 0x4A, 0x2C, 0x8, 0x8, 0x0, 0x0}},
    {8623, 1, {0x0, 0x0, 0x40, 0x40, 0x40, 0x48, 0x58, 0x68, 0x48, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {8624, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x78, 0x24, 0x14, 0x4, 0x4, 0x4, 0x4}},
    {8625, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x1E, 0x24, 0x28, 0x20, 0x20, 0x20, 0x20}},
    {8626, 1, {0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x14, 0x24, 0x78, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8627, 1, {0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x28, 0x24, 0x1E, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8628, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x4, 0x15, 0xE, 0x4, 0x0, 0x0}},
    {8629, 1, {0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x14, 0x24, 0x7C, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8630, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x21, 0xA9, 0x71, 0x20, 0x0, 0x0, 0x0}},
    {8631, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x84, 0x95, 0x8E, 0x4, 0x0, 0x0, 0x0}},
    {8632, 1, {0x0, 0x0, 0x0, 0x7E, 0x0, 0x78, 0x60, 0x50, 0x48, 0x4, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8633, 1, {0x0, 0x0, 0x0, 0x50, 0x60, 0x7E, 0x60, 0x50, 0x0, 0xA, 0x6, 0x7E, 0x6, 0xA, 0x0, 0x0}},
    {8634, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2E, 0x4C, 0x4A, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {8635, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x74, 0x32, 0x52, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {8636, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8637, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8638, 1, {0x0, 0x0, 0x8, 0xC, 0xA, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {8639, 1, {0x0, 0x0, 0x8, 0x18, 0x28, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {8640, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8641, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFE, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8642, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xA, 0xC, 0x8, 0x0, 0x0}},
    {8643, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x28, 0x18, 0x8, 0x0, 0x0}},
    {8644, 1, {0x0, 0x0, 0x0, 0x8, 0x4, 0x7E, 0x4, 0x8, 0x0, 0x10, 0x20, 0x7E, 0x20, 0x10, 0x0, 0x0}},
    {8645, 1, {0x0, 0x0, 0x24, 0x74, 0xAC, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x35, 0x2E, 0x24, 0x0, 0x0}},
    {8646, 1, {0x0, 0x0, 0x0, 0x10, 0x20, 0x7E, 0x20, 0x10, 0x0, 0x8, 0x4, 0x7E, 0x4, 0x8, 0x0, 0x0}},
    {8647, 1, {0x0, 0x0, 0x0, 0x10, 0x20, 0x7E, 0x20, 0x10, 0x0, 0x10, 0x20, 0x7E, 0x20, 0x10, 0x0, 0x0}},
    {8648, 1, {0x0, 0x0, 0x24, 0x7E, 0xA5, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {8649, 1, {0x0, 0x0, 0x0, 0x8, 0x4, 0x7E, 0x4, 0x8, 0x0, 0x8, 0x4, 0x7E, 0x4, 0x8, 0x0, 0x0}},
    {8650, 1, {0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0xA5, 0x7E, 0x24, 0x0, 0x0}},
    {8651, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x7E, 0x0, 0x7E, 0x4, 0x8, 0x0, 0x0, 0x0}},
    {8652, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x7E, 0x0, 0x7E, 0x20, 0x10, 0x0, 0x0, 0x0}},
    {8653, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x28, 0x7E, 0x88, 0x7E, 0x28, 0x8, 0x8, 0x0, 0x0}},
    {8654, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x2C, 0x7E, 0x89, 0x7E, 0x2C, 0x8, 0x8, 0x0, 0x0}},
    {8655, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x14, 0x7E, 0x11, 0x7E, 0x14, 0x10, 0x10, 0x0, 0x0}},
    {8656, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x3E, 0x40, 0x3E, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8657, 1, {0x0, 0x0, 0x8, 0x14, 0x36, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x0, 0x0}},
    {8658, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x7C, 0x2, 0x7C, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8659, 1, {0x0, 0x0, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x36, 0x14, 0x8, 0x0, 0x0}},
    {8660, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x7E, 0x81, 0x7E, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8661, 1, {0x0, 0x0, 0x8, 0x14, 0x36, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x36, 0x14, 0x8, 0x0, 0x0}},
    {8662, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x78, 0x50, 0x68, 0x54, 0xA, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8663, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0xA, 0x16, 0x2A, 0x50, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8664, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0x54, 0x68, 0x50, 0x78, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8665, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x50, 0x2A, 0x16, 0xA, 0x1E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8666, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x3E, 0x40, 0xFE, 0x40, 0x3E, 0x10, 0x0, 0x0, 0x0}},
    {8667, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x7C, 0x2, 0x7F, 0x2, 0x7C, 0x8, 0x0, 0x0, 0x0}},
    {8668, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x42, 0xF5, 0x48, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {8669, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x42, 0xAF, 0x12, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {8670, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {8671, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {8672, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x77, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {8673, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8}},
    {8674, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0xEF, 0x2, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {8675, 1, {0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {8676, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x60, 0x7F, 0x60, 0x50, 0x0, 0x0, 0x0, 0x0}},
    {8677, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xA, 0x6, 0xFE, 0x6, 0xA, 0x0, 0x0, 0x0, 0x0}},
    {8678, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x30, 0x5F, 0x81, 0x5F, 0x30, 0x10, 0x0, 0x0, 0x0}},
    {8679, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x77, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x1C, 0x0, 0x0}},
    {8680, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0xC, 0xFA, 0x81, 0xFA, 0xC, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8681, 1, {0x0, 0x0, 0x1C, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x77, 0x22, 0x14, 0x8, 0x0, 0x0}},
    {8682, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x77, 0x14, 0x14, 0x14, 0x1C, 0x0, 0x1C, 0x14, 0x1C, 0x0, 0x0}},
    {8683, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x63, 0x22, 0x22, 0x22, 0x22, 0x22, 0x63, 0x41, 0x7F, 0x0, 0x0}},
    {8684, 1, {0x0, 0x0, 0x8, 0x14, 0x3E, 0x63, 0x22, 0x22, 0x22, 0x22, 0x22, 0x63, 0x41, 0x7F, 0x0, 0x0}},
    {8685, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x6B, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x6B, 0x49, 0x7F, 0x0, 0x0}},
    {8686, 1, {0x0, 0x8, 0x14, 0x2A, 0x77, 0x22, 0x63, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3E, 0x0, 0x0}},
    {8687, 1, {0x0, 0x8, 0x14, 0x2A, 0x77, 0x22, 0x63, 0x22, 0x22, 0x22, 0x22, 0x63, 0x41, 0x7F, 0x0, 0x0}},
    {8688, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x6C, 0x7A, 0x41, 0x7A, 0x6C, 0x8, 0x0, 0x0, 0x0}},
    {8689, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x40, 0x5C, 0x58, 0x54, 0x42, 0x41, 0x40, 0x40, 0x0, 0x0}},
    {8690, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x41, 0x21, 0x15, 0xD, 0x1D, 0x1, 0x7F, 0x0, 0x0}},
    {8691, 1, {0x0, 0x0, 0x8, 0x14, 0x36, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x36, 0x14, 0x8, 0x0, 0x0}},
    {8693, 1, {0x0, 0x0, 0x24, 0x2E, 0x35, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0xAC, 0x74, 0x24, 0x0, 0x0}},
    {8694, 1, {0x0, 0x4, 0x2, 0xFF, 0x2, 0x4, 0x4, 0x2, 0xFF, 0x2, 0x4, 0x4, 0x2, 0xFF, 0x2, 0x4}},
    {8695, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x44, 0xFF, 0x44, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8696, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x22, 0xFF, 0x22, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8701, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x60, 0xBF, 0x60, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {8702, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x6, 0xFD, 0x6, 0x4, 0x0, 0x0, 0x0, 0x0}},
    {8704, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x3E, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {8705, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x20, 0x20, 0x20, 0x20, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {8706, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x2, 0x2, 0x1E, 0x22, 0x42, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {8707, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {8708, 1, {0x0, 0x0, 0x10, 0x10, 0x7E, 0x12, 0x12, 0x12, 0x7E, 0x12, 0x12, 0x12, 0x12, 0x7E, 0x10, 0x0}},
    {8709, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1D, 0x22, 0x45, 0x49, 0x51, 0x22, 0x5C, 0x0, 0x0}},
    {8710, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x7F, 0x0, 0x0}},
    {8711, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {8712, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x7E, 0x40, 0x40, 0x3E, 0x0, 0x0}},
    {8713, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x48, 0x48, 0x7E, 0x48, 0x48, 0x3E, 0x8, 0x0}},
    {8714, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x7E, 0x40, 0x3E, 0x0, 0x0}},
    {8715, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x7E, 0x2, 0x2, 0x7C, 0x0, 0x0}},
    {8716, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x12, 0x12, 0x7E, 0x12, 0x12, 0x7C, 0x10, 0x0}},
    {8717, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x7E, 0x2, 0x7C, 0x0, 0x0}},
    {8718, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x0, 0x0}},
    {8719, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {8720, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x7F, 0x0, 0x0}},
    {8721, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x20, 0x10, 0x8, 0x8, 0x10, 0x20, 0x42, 0x7E, 0x0, 0x0}},
    {8722, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8723, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x0, 0x8, 0x8, 0x8, 0x7F, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0}},
    {8724, 1, {0x0, 0x0, 0x8, 0x1C, 0x8, 0x0, 0x8, 0x8, 0x8, 0x7F, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0}},
    {8725, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x40, 0x0, 0x0}},
    {8726, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x2, 0x0, 0x0}},
    {8727, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x2A, 0x1C, 0x8, 0x1C, 0x2A, 0x8, 0x0, 0x0, 0x0}},
    {8728, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {8729, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x3C, 0x3C, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {8730, 1, {0x0, 0x0, 0x7, 0x4, 0x4, 0x4, 0x4, 0x4, 0x74, 0x14, 0x14, 0xC, 0xC, 0x4, 0x4, 0x0}},
    {8731, 1, {0x0, 0x0, 0x67, 0x14, 0x64, 0x14, 0x64, 0x4, 0x74, 0x14, 0x14, 0xC, 0xC, 0x4, 0x4, 0x0}},
    {8732, 1, {0x0, 0x0, 0x57, 0x54, 0x74, 0x14, 0x14, 0x4, 0x74, 0x14, 0x14, 0xC, 0xC, 0x4, 0x4, 0x0}},
    {8733, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x48, 0x48, 0x36, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8734, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x49, 0x49, 0x36, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8735, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {8736, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x7E, 0x0, 0x0}},
    {8737, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x14, 0x8, 0x14, 0x24, 0x7E, 0x0, 0x0}},
    {8738, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xA, 0xC, 0x14, 0x64, 0x14, 0xC, 0xA, 0x0, 0x0}},
    {8739, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x0, 0x0}},
    {8740, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x1A, 0x1C, 0x18, 0x18, 0x38, 0x58, 0x18, 0x18, 0x0, 0x0}},
    {8741, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {8742, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x26, 0x24, 0x2C, 0x34, 0x24, 0x64, 0x24, 0x24, 0x0, 0x0}},
    {8743, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x41, 0x41, 0x0, 0x0}},
    {8744, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {8745, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8746, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8747, 1, {0x0, 0x6, 0xA, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x28, 0x30, 0x0}},
    {8748, 1, {0x0, 0xA, 0x15, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x54, 0x68, 0x0}},
    {8749, 1, {0x0, 0x15, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0xAA, 0xD4, 0x0}},
    {8750, 1, {0x0, 0x6, 0xA, 0x8, 0x8, 0x8, 0x1C, 0x2A, 0x2A, 0x1C, 0x8, 0x8, 0x8, 0x28, 0x30, 0x0}},
    {8751, 1, {0x0, 0xA, 0x15, 0x14, 0x14, 0x14, 0x3E, 0x55, 0x55, 0x3E, 0x14, 0x14, 0x14, 0x54, 0x68, 0x0}},
    {8752, 1, {0x0, 0x15, 0x2A, 0x2A, 0x2A, 0x7E, 0xAB, 0xAB, 0xAB, 0x7E, 0x2A, 0x2A, 0x2A, 0xAA, 0xD4, 0x0}},
    {8753, 1, {0x0, 0x6, 0xA, 0x8, 0x8, 0x8, 0x1D, 0x2B, 0x2F, 0x8, 0x8, 0x8, 0x8, 0x28, 0x30, 0x0}},
    {8754, 1, {0x0, 0x6, 0xA, 0x8, 0x8, 0x1D, 0x2B, 0x2F, 0x2A, 0x1C, 0x8, 0x8, 0x8, 0x28, 0x30, 0x0}},
    {8755, 1, {0x0, 0x6, 0xA, 0x8, 0x8, 0x8, 0x5C, 0x6A, 0x7A, 0x2A, 0x1C, 0x8, 0x8, 0x28, 0x30, 0x0}},
    {8756, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x66, 0x66, 0x0, 0x0, 0x0}},
    {8757, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x66, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {8758, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {8759, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x66, 0x0, 0x0, 0x0, 0x66, 0x66, 0x0, 0x0, 0x0}},
    {8760, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8761, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x6, 0x0, 0x78, 0x0, 0x6, 0x6, 0x0, 0x0, 0x0}},
    {8762, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x66, 0x66, 0x0, 0x7E, 0x0, 0x66, 0x66, 0x0, 0x0, 0x0}},
    {8763, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x32, 0x4C, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {8764, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8765, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x32, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8766, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x49, 0x26, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8767, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x48, 0x49, 0x9, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8768, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x8, 0x8, 0x8, 0x10, 0x10, 0x10, 0x8, 0x0, 0x0, 0x0}},
    {8769, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8770, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x0}},
    {8771, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8772, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x7E, 0x8, 0x8, 0x0, 0x0}},
    {8773, 1, {0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8774, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x8, 0x7E, 0x8, 0x8, 0x7E, 0x8, 0x8, 0x0, 0x0}},
    {8775, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x3A, 0x4C, 0x8, 0x7E, 0x8, 0x8, 0x7E, 0x8, 0x8, 0x0, 0x0}},
    {8776, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x0, 0x0}},
    {8777, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3A, 0x4C, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x0, 0x0}},
    {8778, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x32, 0x4C, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8779, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x32, 0x4C, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x0}},
    {8780, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x49, 0x6, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8781, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x3C, 0x0, 0x0, 0x3C, 0x42, 0x0, 0x0, 0x0, 0x0}},
    {8782, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x66, 0x0, 0x0, 0x66, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {8783, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x66, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8784, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8785, 1, {0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {8786, 1, {0x0, 0x0, 0x0, 0x60, 0x60, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x6, 0x6, 0x0, 0x0, 0x0}},
    {8787, 1, {0x0, 0x0, 0x0, 0x6, 0x6, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x60, 0x60, 0x0, 0x0, 0x0}},
    {8788, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5E, 0x0, 0x0, 0x5E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8789, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7A, 0x0, 0x0, 0x7A, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8790, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x24, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8791, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8792, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8793, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8794, 1, {0x0, 0x0, 0x0, 0x42, 0x24, 0x18, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8795, 1, {0x0, 0x0, 0x8, 0x3E, 0x1C, 0x14, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8796, 1, {0x0, 0x0, 0x8, 0x14, 0x22, 0x3E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8797, 1, {0x0, 0x0, 0x12, 0x14, 0x76, 0x54, 0x74, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0}},
    {8798, 1, {0x0, 0x0, 0x0, 0x68, 0x54, 0x54, 0x54, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8799, 1, {0x0, 0x30, 0x48, 0x8, 0x10, 0x10, 0x0, 0x10, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0}},
    {8800, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x7E, 0x8, 0x10, 0x7E, 0x20, 0x40, 0x0, 0x0}},
    {8801, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8802, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x7E, 0x8, 0x7E, 0x10, 0x7E, 0x20, 0x40, 0x0, 0x0}},
    {8803, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8804, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x7E, 0x0, 0x0, 0x0}},
    {8805, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x7E, 0x0, 0x0, 0x0}},
    {8806, 1, {0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0}},
    {8807, 1, {0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0}},
    {8808, 1, {0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0x16, 0x10, 0x7E, 0x10, 0x10, 0x7E, 0x10, 0x10, 0x0}},
    {8809, 1, {0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0x18, 0x68, 0x8, 0x7E, 0x8, 0x8, 0x7E, 0x8, 0x8, 0x0}},
    {8810, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x9, 0x12, 0x24, 0x48, 0x90, 0x48, 0x24, 0x12, 0x9, 0x0, 0x0}},
    {8811, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x90, 0x48, 0x24, 0x12, 0x9, 0x12, 0x24, 0x48, 0x90, 0x0, 0x0}},
    {8812, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x18, 0x18, 0x24, 0x24, 0x24, 0x24, 0x18, 0x18, 0x24, 0x0, 0x0}},
    {8813, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4A, 0x3C, 0x8, 0x8, 0x3C, 0x4A, 0x8, 0x8, 0x0, 0x0}},
    {8814, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0xA, 0xC, 0x8, 0x18, 0x28, 0x18, 0x8, 0xC, 0xA, 0x8, 0x0}},
    {8815, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x50, 0x30, 0x10, 0x18, 0x14, 0x18, 0x10, 0x30, 0x50, 0x10, 0x0}},
    {8816, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0xE, 0x18, 0x68, 0x18, 0xE, 0x8, 0x7E, 0x8, 0x8, 0x0}},
    {8817, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x70, 0x18, 0x16, 0x18, 0x70, 0x10, 0x7E, 0x10, 0x10, 0x0}},
    {8818, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x32, 0x4C, 0x0, 0x0}},
    {8819, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x32, 0x4C, 0x0, 0x0}},
    {8820, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0xE, 0x18, 0x68, 0x18, 0xE, 0x8, 0x3A, 0x4C, 0x8, 0x0}},
    {8821, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x70, 0x18, 0x16, 0x18, 0x70, 0x10, 0x32, 0x5C, 0x10, 0x0}},
    {8822, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x0}},
    {8823, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x0}},
    {8824, 1, {0x0, 0x0, 0x8, 0x8, 0xE, 0x18, 0x68, 0x18, 0xE, 0x68, 0x18, 0xE, 0x18, 0x68, 0x8, 0x0}},
    {8825, 1, {0x0, 0x0, 0x8, 0x8, 0x68, 0x18, 0xE, 0x18, 0x68, 0xE, 0x18, 0x68, 0x18, 0xE, 0x8, 0x0}},
    {8826, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x18, 0x60, 0x18, 0x4, 0x2, 0x2, 0x0, 0x0, 0x0}},
    {8827, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x18, 0x6, 0x18, 0x20, 0x40, 0x40, 0x0, 0x0, 0x0}},
    {8828, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x18, 0x60, 0x18, 0x4, 0x2, 0x0, 0x7E, 0x0, 0x0}},
    {8829, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x18, 0x6, 0x18, 0x20, 0x40, 0x0, 0x7E, 0x0, 0x0}},
    {8830, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x18, 0x60, 0x18, 0x4, 0x2, 0x32, 0x4C, 0x0, 0x0}},
    {8831, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x18, 0x6, 0x18, 0x20, 0x40, 0x32, 0x4C, 0x0, 0x0}},
    {8832, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x12, 0x14, 0x18, 0x70, 0x18, 0x14, 0x12, 0x2, 0x0, 0x0, 0x0}},
    {8833, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x48, 0x28, 0x18, 0xE, 0x18, 0x28, 0x48, 0x40, 0x0, 0x0, 0x0}},
    {8834, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0, 0x0, 0x0}},
    {8835, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x0, 0x0, 0x0}},
    {8836, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x48, 0x48, 0x48, 0x48, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {8837, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x12, 0x12, 0x12, 0x12, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {8838, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x7E, 0x0, 0x0}},
    {8839, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x7E, 0x0, 0x0}},
    {8840, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x48, 0x48, 0x48, 0x48, 0x3E, 0x8, 0x7E, 0x8, 0x0}},
    {8841, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x12, 0x12, 0x12, 0x12, 0x7C, 0x10, 0x7E, 0x10, 0x0}},
    {8842, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x40, 0x3E, 0x8, 0x7E, 0x8, 0x0}},
    {8843, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x2, 0x7C, 0x10, 0x7E, 0x10, 0x0}},
    {8844, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x4A, 0x52, 0x7E, 0x52, 0x4A, 0x42, 0x3C, 0x0, 0x0}},
    {8845, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x5A, 0x5A, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8846, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x52, 0x52, 0x7A, 0x52, 0x52, 0x42, 0x3C, 0x0, 0x0}},
    {8847, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8848, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8849, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x7E, 0x0, 0x0}},
    {8850, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x7E, 0x0, 0x0}},
    {8851, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0, 0x0, 0x0}},
    {8852, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8853, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x2A, 0x49, 0x7F, 0x49, 0x2A, 0x1C, 0x0, 0x0, 0x0}},
    {8854, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x7F, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8855, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x55, 0x49, 0x55, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8856, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x45, 0x49, 0x51, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8857, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x49, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8858, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x49, 0x55, 0x49, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8859, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x49, 0x5D, 0x49, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8860, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x5D, 0x41, 0x5D, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8861, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x5D, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {8862, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x49, 0x49, 0x7F, 0x49, 0x49, 0x7F, 0x0, 0x0, 0x0}},
    {8863, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x7F, 0x41, 0x41, 0x7F, 0x0, 0x0, 0x0}},
    {8864, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x63, 0x55, 0x49, 0x55, 0x63, 0x7F, 0x0, 0x0, 0x0}},
    {8865, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x49, 0x41, 0x41, 0x7F, 0x0, 0x0, 0x0}},
    {8866, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x7E, 0x40, 0x40, 0x40, 0x0, 0x0, 0x0}},
    {8867, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x7E, 0x2, 0x2, 0x2, 0x0, 0x0, 0x0}},
    {8868, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0}},
    {8869, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x7F, 0x0, 0x0, 0x0}},
    {8870, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x3C, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0, 0x0}},
    {8871, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x3C, 0x20, 0x3C, 0x20, 0x20, 0x20, 0x20, 0x0, 0x0}},
    {8872, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x40, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {8873, 1, {0x0, 0x0, 0x0, 0x50, 0x50, 0x50, 0x50, 0x50, 0x5E, 0x50, 0x50, 0x50, 0x50, 0x50, 0x0, 0x0}},
    {8874, 1, {0x0, 0x0, 0x0, 0x54, 0x54, 0x54, 0x54, 0x54, 0x57, 0x54, 0x54, 0x54, 0x54, 0x54, 0x0, 0x0}},
    {8875, 1, {0x0, 0x0, 0x0, 0x50, 0x50, 0x50, 0x50, 0x5E, 0x50, 0x5E, 0x50, 0x50, 0x50, 0x50, 0x0, 0x0}},
    {8876, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x48, 0x48, 0x48, 0x7E, 0x48, 0x48, 0x48, 0x40, 0x40, 0x0, 0x0}},
    {8877, 1, {0x0, 0x0, 0x0, 0x40, 0x48, 0x48, 0x48, 0x7E, 0x48, 0x7E, 0x48, 0x48, 0x48, 0x40, 0x0, 0x0}},
    {8878, 1, {0x0, 0x0, 0x0, 0x50, 0x50, 0x54, 0x54, 0x54, 0x5E, 0x54, 0x54, 0x54, 0x50, 0x50, 0x0, 0x0}},
    {8879, 1, {0x0, 0x0, 0x0, 0x50, 0x54, 0x54, 0x54, 0x5E, 0x54, 0x5E, 0x54, 0x54, 0x54, 0x50, 0x0, 0x0}},
    {8880, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x2, 0x4, 0x18, 0x60, 0x18, 0x4, 0x2, 0xC, 0x0, 0x0, 0x0}},
    {8881, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x20, 0x18, 0x6, 0x18, 0x20, 0x40, 0x30, 0x0, 0x0, 0x0}},
    {8882, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1A, 0x62, 0x1A, 0x6, 0x0, 0x0, 0x0, 0x0}},
    {8883, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x46, 0x58, 0x60, 0x0, 0x0, 0x0, 0x0}},
    {8884, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1A, 0x62, 0x1A, 0x6, 0x0, 0x7E, 0x0, 0x0}},
    {8885, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x46, 0x58, 0x60, 0x0, 0x7E, 0x0, 0x0}},
    {8886, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x5F, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8887, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x7D, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8888, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x7D, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8889, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x0, 0x63, 0x0, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8890, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8891, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x14, 0x14, 0x8, 0x0, 0x3E, 0x0, 0x0}},
    {8892, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x8, 0x14, 0x14, 0x22, 0x22, 0x0, 0x0, 0x0, 0x0}},
    {8893, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x22, 0x22, 0x14, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8894, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x78, 0x48, 0x48, 0x7E, 0x0, 0x0}},
    {8895, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x6, 0xA, 0x12, 0x22, 0x7E, 0x0, 0x0}},
    {8896, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x22, 0x41, 0x41, 0x41, 0x0, 0x0}},
    {8897, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x41, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x0, 0x0}},
    {8898, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x0, 0x0}},
    {8899, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {8900, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8901, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8902, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x3E, 0x1C, 0x14, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8903, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x22, 0x14, 0x3E, 0x14, 0x22, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8904, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x36, 0x2A, 0x36, 0x22, 0x0, 0x0, 0x0}},
    {8905, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x34, 0x28, 0x34, 0x22, 0x0, 0x0, 0x0}},
    {8906, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x16, 0xA, 0x16, 0x22, 0x0, 0x0, 0x0}},
    {8907, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x10, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0}},
    {8908, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0}},
    {8909, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x32, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {8910, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x0, 0x0, 0x0}},
    {8911, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x14, 0x22, 0x22, 0x41, 0x0, 0x0, 0x0}},
    {8912, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x5E, 0x50, 0x50, 0x5E, 0x40, 0x3E, 0x0, 0x0}},
    {8913, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x7A, 0xA, 0xA, 0x7A, 0x2, 0x7C, 0x0, 0x0}},
    {8914, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x5D, 0x55, 0x55, 0x55, 0x0, 0x0, 0x0}},
    {8915, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x55, 0x55, 0x55, 0x5D, 0x41, 0x3E, 0x0, 0x0, 0x0}},
    {8916, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x1C, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x0, 0x0}},
    {8917, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x24, 0x7E, 0x24, 0x24, 0x7E, 0x24, 0x24, 0x0, 0x0}},
    {8918, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x22, 0x10, 0x8, 0x4, 0x2, 0x0, 0x0}},
    {8919, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x10, 0x8, 0x44, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {8920, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x15, 0x2A, 0x54, 0x2A, 0x15, 0x0, 0x0, 0x0, 0x0}},
    {8921, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x2A, 0x15, 0x2A, 0x54, 0x0, 0x0, 0x0, 0x0}},
    {8922, 1, {0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x7E, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x0}},
    {8923, 1, {0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x7E, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x0}},
    {8924, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x0, 0x0, 0x0, 0x0}},
    {8925, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x0, 0x0, 0x0, 0x0}},
    {8926, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x18, 0x60, 0x18, 0x4, 0x2, 0x2, 0x0, 0x0}},
    {8927, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x20, 0x18, 0x6, 0x18, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {8928, 1, {0x0, 0x0, 0x0, 0x8, 0xA, 0xA, 0xC, 0x18, 0x68, 0x18, 0xC, 0xA, 0xA, 0x7E, 0x8, 0x0}},
    {8929, 1, {0x0, 0x0, 0x0, 0x8, 0x48, 0x48, 0x28, 0x18, 0xE, 0x18, 0x28, 0x48, 0x48, 0x7E, 0x8, 0x0}},
    {8930, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x7E, 0x48, 0x48, 0x48, 0x48, 0x7E, 0x8, 0x7E, 0x8, 0x0}},
    {8931, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x7E, 0xA, 0xA, 0xA, 0xA, 0x7E, 0x8, 0x7E, 0x8, 0x0}},
    {8932, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x8, 0x7E, 0x8, 0x0}},
    {8933, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x8, 0x7E, 0x8, 0x0}},
    {8934, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0x6, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x0}},
    {8935, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x18, 0x6, 0x18, 0x60, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x0}},
    {8936, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x18, 0x60, 0x18, 0x4, 0xA, 0x3A, 0x4C, 0x8, 0x8, 0x0}},
    {8937, 1, {0x0, 0x0, 0x40, 0x40, 0x20, 0x18, 0x6, 0x18, 0x20, 0x40, 0x48, 0x3A, 0x4C, 0x8, 0x8, 0x0}},
    {8938, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0xE, 0x1A, 0x6A, 0x1A, 0xE, 0x8, 0x8, 0x0, 0x0}},
    {8939, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x68, 0x58, 0x4E, 0x58, 0x68, 0x8, 0x8, 0x0, 0x0}},
    {8940, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0xE, 0x1A, 0x6A, 0x1A, 0xE, 0x8, 0x7E, 0x8, 0x0}},
    {8941, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x68, 0x58, 0x4E, 0x58, 0x68, 0x8, 0x7E, 0x8, 0x0}},
    {8942, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0}},
    {8943, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xDB, 0xDB, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8944, 1, {0x0, 0x0, 0x0, 0x0, 0x3, 0x3, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0xC0, 0xC0, 0x0, 0x0}},
    {8945, 1, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xC0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x3, 0x3, 0x0, 0x0}},
    {8948, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x42, 0x7E, 0x42, 0x40, 0x3E, 0x0, 0x0, 0x0, 0x0}},
    {8951, 1, {0x0, 0x0, 0x0, 0x7E, 0x0, 0x3E, 0x40, 0x40, 0x7E, 0x40, 0x40, 0x3E, 0x0, 0x0, 0x0, 0x0}},
    {8952, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x7E, 0x40, 0x40, 0x3E, 0x0, 0x7E}},
    {8956, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x42, 0x7E, 0x42, 0x2, 0x7C, 0x0, 0x0, 0x0, 0x0}},
    {8958, 1, {0x0, 0x0, 0x0, 0x7E, 0x0, 0x7C, 0x2, 0x2, 0x7E, 0x2, 0x2, 0x7C, 0x0, 0x0, 0x0, 0x0}},
    {8960, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1D, 0x22, 0x45, 0x49, 0x51, 0x22, 0x5C, 0x0, 0x0, 0x0}},
    {8961, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x90, 0x58, 0x34, 0x12, 0x0, 0x0, 0x0, 0x0}},
    {8962, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {8963, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8964, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x14, 0x8, 0x0, 0x0}},
    {8965, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8966, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x3E, 0x0, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8967, 1, {0x0, 0x0, 0x0, 0x10, 0x8, 0x4, 0x8, 0x10, 0x8, 0x4, 0x8, 0x10, 0x8, 0x4, 0x0, 0x0}},
    {8968, 1, {0x0, 0x0, 0x1E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {8969, 1, {0x0, 0x0, 0x78, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8970, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1E, 0x0, 0x0}},
    {8971, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x78, 0x0, 0x0}},
    {8972, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8973, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {8974, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x7, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8975, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x70, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8976, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {8977, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x3E, 0x22, 0x22, 0x22, 0x3E, 0x41, 0x0, 0x0, 0x0}},
    {8978, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x0, 0x0, 0x0, 0x0}},
    {8979, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x7F, 0x0, 0x0, 0x0, 0x0}},
    {8980, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x22, 0x14, 0x8, 0x0, 0x0}},
    {8981, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x51, 0x22, 0x5C, 0x0, 0x0}},
    {8982, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x1C, 0x2A, 0x7F, 0x2A, 0x1C, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {8983, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0xFF, 0x24, 0x24, 0xFF, 0x24, 0x24, 0x0, 0x0, 0x0, 0x0}},
    {8984, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x55, 0x3E, 0x14, 0x3E, 0x55, 0x22, 0x0, 0x0, 0x0, 0x0}},
    {8985, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8986, 1, {0x0, 0x28, 0x28, 0x28, 0x28, 0x38, 0x54, 0x92, 0x9B, 0x82, 0x44, 0x38, 0x28, 0x28, 0x28, 0x0}},
    {8987, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x24, 0x24, 0x18, 0x18, 0x24, 0x24, 0x3C, 0x7E, 0x0, 0x0}},
    {8988, 1, {0x0, 0x78, 0x40, 0x40, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8989, 1, {0x0, 0x1E, 0x2, 0x2, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8990, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x78, 0x0}},
    {8991, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x1E, 0x0}},
    {8992, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0xA, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {8993, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x28, 0x30, 0x0, 0x0, 0x0, 0x0}},
    {8994, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8995, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8996, 1, {0x0, 0x0, 0x0, 0x0, 0x77, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8997, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x67, 0x10, 0x8, 0x7, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8998, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0xAA, 0x91, 0xAA, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {8999, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x55, 0x49, 0x55, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9000, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x7E, 0x7E, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9003, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x55, 0x89, 0x55, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9004, 1, {0x0, 0x0, 0x0, 0x8, 0x1C, 0x26, 0x63, 0x61, 0x61, 0x63, 0x26, 0x1C, 0x8, 0x0, 0x0, 0x0}},
    {9005, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x18, 0x25, 0x66, 0x9C, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {9006, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x2F, 0x38, 0x60, 0x60, 0x0, 0x0}},
    {9007, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x0, 0x7E, 0x0, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {9008, 1, {0x0, 0x0, 0x0, 0x0, 0x9, 0x1B, 0x12, 0x12, 0x24, 0x24, 0x24, 0x48, 0x48, 0x78, 0x0, 0x0}},
    {9009, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x70, 0xAA, 0xBF, 0xAA, 0x70, 0x20, 0x20, 0x0, 0x0}},
    {9010, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x46, 0x7F, 0x46, 0x58, 0x60, 0x0, 0x0, 0x0}},
    {9011, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x46, 0x7F, 0x0, 0x0, 0x0}},
    {9012, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0, 0x0}},
    {9013, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x24, 0x18, 0x0, 0x0, 0x0}},
    {9014, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x7F, 0x0, 0x0}},
    {9015, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3E, 0x0, 0x0}},
    {9016, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x41, 0x7F, 0x41, 0x41, 0x7F, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9017, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x49, 0x49, 0x41, 0x7F, 0x41, 0x49, 0x49, 0x41, 0x7F, 0x0, 0x0}},
    {9018, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x49, 0x55, 0x63, 0x41, 0x63, 0x55, 0x49, 0x41, 0x7F, 0x0, 0x0}},
    {9019, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x41, 0x49, 0x55, 0x55, 0x49, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9020, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x5D, 0x63, 0x41, 0x41, 0x63, 0x5D, 0x41, 0x7F, 0x0, 0x0}},
    {9021, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x1C, 0x2A, 0x49, 0x49, 0x49, 0x2A, 0x1C, 0x8, 0x8, 0x0, 0x0}},
    {9022, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x49, 0x55, 0x49, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9023, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x4, 0x8, 0x8, 0xFF, 0x10, 0x20, 0x20, 0x40, 0x0, 0x0}},
    {9024, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x20, 0x10, 0x10, 0xFF, 0x8, 0x4, 0x4, 0x2, 0x0, 0x0}},
    {9025, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x43, 0x45, 0x45, 0x49, 0x51, 0x51, 0x61, 0x41, 0x7F, 0x0, 0x0}},
    {9026, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x61, 0x51, 0x51, 0x49, 0x45, 0x45, 0x43, 0x41, 0x7F, 0x0, 0x0}},
    {9027, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x47, 0x59, 0x61, 0x59, 0x47, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9028, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x71, 0x4D, 0x43, 0x4D, 0x71, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9029, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x28, 0x7F, 0x28, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9030, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0xA, 0x7F, 0xA, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9031, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x49, 0x51, 0x7F, 0x51, 0x49, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9032, 1, {0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x49, 0x45, 0x7F, 0x45, 0x49, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9033, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C, 0x22, 0x51, 0x49, 0x45, 0x22, 0x1D, 0x0, 0x0}},
    {9034, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x7F, 0x0, 0x7F, 0x0, 0x0}},
    {9035, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x1C, 0x1C, 0x2A, 0x2A, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {9036, 1, {0x0, 0x0, 0x7F, 0x41, 0x41, 0x41, 0x41, 0x63, 0x63, 0x55, 0x55, 0x49, 0x41, 0x7F, 0x0, 0x0}},
    {9037, 1, {0x0, 0x0, 0x7F, 0x41, 0x41, 0x49, 0x49, 0x55, 0x55, 0x63, 0x63, 0x7F, 0x41, 0x7F, 0x0, 0x0}},
    {9038, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x1C, 0x2A, 0x2A, 0x1C, 0x8, 0x8, 0x7F, 0x0, 0x0}},
    {9039, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9040, 1, {0x0, 0x0, 0x7F, 0x49, 0x5D, 0x6B, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {9041, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9042, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x2A, 0x2A, 0x1C, 0x1C, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9043, 1, {0x0, 0x0, 0x7F, 0x41, 0x49, 0x55, 0x55, 0x63, 0x63, 0x41, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9044, 1, {0x0, 0x0, 0x7F, 0x41, 0x41, 0x7F, 0x63, 0x63, 0x55, 0x55, 0x49, 0x49, 0x41, 0x7F, 0x0, 0x0}},
    {9045, 1, {0x0, 0x0, 0x0, 0x0, 0x7F, 0x8, 0x8, 0x1C, 0x2A, 0x2A, 0x1C, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9046, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x7F, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {9047, 1, {0x0, 0x0, 0x7F, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x6B, 0x5D, 0x49, 0x7F, 0x0, 0x0}},
    {9048, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x0, 0x0}},
    {9049, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x3E, 0x0, 0x3E, 0x0, 0x0}},
    {9050, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x41, 0x22, 0x14, 0x8, 0x0, 0x3E, 0x0, 0x0}},
    {9051, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x3E, 0x0, 0x0}},
    {9052, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x7F, 0x0, 0x0}},
    {9053, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x49, 0x55, 0x49, 0x41, 0x41, 0x41, 0x41, 0x41, 0x0, 0x0}},
    {9054, 1, {0x0, 0x0, 0x7F, 0x41, 0x49, 0x49, 0x49, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9055, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x49, 0x7F, 0x5D, 0x55, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9056, 1, {0x0, 0x0, 0x7F, 0x41, 0x41, 0x49, 0x49, 0x41, 0x41, 0x49, 0x49, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9057, 1, {0x0, 0x24, 0x24, 0x0, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9058, 1, {0x0, 0x24, 0x24, 0x0, 0x0, 0x3E, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {9059, 1, {0x0, 0x24, 0x24, 0x0, 0x0, 0x8, 0x3E, 0x1C, 0x14, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9060, 1, {0x0, 0x24, 0x24, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9061, 1, {0x0, 0x24, 0x24, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0}},
    {9062, 1, {0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {9063, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x1F, 0x28, 0x48, 0x48, 0x48, 0x28, 0x1F, 0x8, 0x8, 0x0, 0x0}},
    {9064, 1, {0x0, 0x24, 0x24, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9065, 1, {0x0, 0x24, 0x24, 0x0, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0}},
    {9066, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9067, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x22, 0x22, 0x14, 0x14, 0x39, 0x4E, 0x0, 0x0, 0x0, 0x0}},
    {9068, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x33, 0x6E, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {9069, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3A, 0x4C, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9070, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x18, 0x8, 0x8, 0x10, 0x0, 0x3E, 0x0, 0x0}},
    {9071, 1, {0x0, 0x0, 0x7F, 0x41, 0x43, 0x45, 0x7F, 0x49, 0x7F, 0x51, 0x61, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9072, 1, {0x0, 0x0, 0x7F, 0x41, 0x49, 0x55, 0x55, 0x45, 0x49, 0x49, 0x41, 0x49, 0x41, 0x7F, 0x0, 0x0}},
    {9073, 1, {0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x22, 0x22, 0x14, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {9074, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4C, 0x0, 0x8, 0x14, 0x14, 0x22, 0x22, 0x0, 0x0, 0x0}},
    {9075, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x6, 0x0, 0x0}},
    {9076, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5C, 0x40, 0x40}},
    {9077, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x22, 0x49, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x0}},
    {9078, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x7E, 0x0, 0x0}},
    {9079, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x7C, 0x40, 0x40, 0x3E, 0x0, 0x7E, 0x0, 0x0}},
    {9080, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x6, 0x0, 0x1E, 0x0, 0x0}},
    {9081, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x41, 0x49, 0x49, 0x49, 0x49, 0x36, 0x0, 0x7F, 0x0, 0x0}},
    {9082, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x32, 0x4A, 0x44, 0x44, 0x44, 0x44, 0x4A, 0x32, 0x0, 0x0}},
    {9087, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1C, 0x3E, 0x1C, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9088, 1, {0x0, 0x0, 0x0, 0x38, 0x4, 0x3C, 0x44, 0x44, 0x3A, 0x0, 0x82, 0x44, 0x28, 0x10, 0x0, 0x0}},
    {9109, 1, {0x0, 0x0, 0x7F, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9110, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x78, 0x70, 0x70, 0x60, 0x60, 0x40, 0x0}},
    {9115, 1, {0x0, 0xE, 0x18, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30}},
    {9116, 1, {0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30}},
    {9117, 1, {0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x18, 0xE}},
    {9118, 1, {0x0, 0x70, 0x18, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC}},
    {9119, 1, {0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC}},
    {9120, 1, {0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0x18, 0x70}},
    {9121, 1, {0x0, 0x3E, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30}},
    {9122, 1, {0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30}},
    {9123, 1, {0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3E}},
    {9124, 1, {0x0, 0x7C, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC}},
    {9125, 1, {0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC}},
    {9126, 1, {0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0x7C}},
    {9127, 1, {0x0, 0x6, 0xC, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9128, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x10, 0x20, 0x40, 0x20, 0x10, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9129, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xC, 0x6}},
    {9130, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9131, 1, {0x0, 0x60, 0x30, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9132, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x8, 0x4, 0x2, 0x4, 0x8, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9133, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x30, 0x60}},
    {9134, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9135, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9136, 1, {0x0, 0x6, 0xC, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x30, 0x60}},
    {9137, 1, {0x0, 0x60, 0x30, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xC, 0x6}},
    {9143, 1, {0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x1, 0x2, 0x2, 0x62, 0x24, 0x14, 0x14, 0x8, 0x0, 0x0}},
    {9144, 1, {0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80}},
    {9145, 1, {0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1}},
    {9146, 1, {0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9147, 1, {0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9148, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0}},
    {9149, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF}},
    {9150, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0, 0x0}},
    {9151, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x3E, 0x0, 0x0}},
    {9163, 1, {0x0, 0x3F, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x0}},
    {9164, 1, {0x0, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x3F, 0x0}},
    {9250, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x18, 0x10, 0x30, 0x5C, 0x12, 0x12, 0x12, 0x12, 0x1C, 0x0, 0x0}},
    {9251, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x7E, 0x0, 0x0}},
    {9252, 1, {0x0, 0x0, 0x0, 0x0, 0x48, 0x68, 0x58, 0x48, 0x0, 0x10, 0x10, 0x10, 0x1E, 0x0, 0x0, 0x0}},
    {9253, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0xA, 0x15, 0x2A, 0x54, 0x28, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {9254, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x42, 0x40, 0x20, 0x10, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0, 0x0}},
    {9280, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0xA, 0xA, 0xA, 0x8, 0x8, 0x28, 0x28, 0x28, 0x38, 0x0, 0x0}},
    {9281, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {9282, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x22, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9283, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {9284, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x2A, 0x2A, 0x2A, 0x8, 0x8, 0x2A, 0x2A, 0x2A, 0x3E, 0x0, 0x0}},
    {9285, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x36, 0x2A, 0x36, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9286, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xE, 0x6E, 0x6E, 0x60, 0x60, 0x6E, 0x6E, 0xE, 0x0, 0x0, 0x0}},
    {9287, 1, {0x0, 0x0, 0x0, 0x0, 0x3, 0x3, 0x3, 0x1B, 0x18, 0x18, 0xD8, 0xC0, 0xC0, 0xC0, 0x0, 0x0}},
    {9288, 1, {0x0, 0x0, 0x0, 0x0, 0x7, 0x57, 0x57, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x0, 0x0}},
    {9289, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6D, 0x6D, 0x6D, 0x6D, 0x6D, 0x6D, 0x6D, 0x6D, 0x6D, 0x0, 0x0}},
    {9290, 1, {0x0, 0x0, 0x0, 0x0, 0xA0, 0xA0, 0x50, 0x28, 0x28, 0x14, 0x14, 0xA, 0x5, 0x5, 0x0, 0x0}},
    {9472, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9473, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9474, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9475, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9476, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xDB, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9477, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xDB, 0xDB, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9478, 1, {0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8}},
    {9479, 1, {0x18, 0x18, 0x18, 0x18, 0x0, 0x0, 0x18, 0x18, 0x18, 0x18, 0x0, 0x0, 0x18, 0x18, 0x18, 0x18}},
    {9480, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xAA, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9481, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xAA, 0xAA, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9482, 1, {0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x0}},
    {9483, 1, {0x18, 0x18, 0x18, 0x0, 0x18, 0x18, 0x18, 0x0, 0x18, 0x18, 0x18, 0x0, 0x18, 0x18, 0x18, 0x0}},
    {9484, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9485, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9486, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9487, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1F, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9488, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9489, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9490, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9491, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9492, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9493, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9494, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9495, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1F, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9496, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9497, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9498, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9499, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF8, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9500, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9501, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9502, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9503, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9504, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9505, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1F, 0x1F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9506, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1F, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9507, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1F, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9508, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9509, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9510, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9511, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9512, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9513, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF8, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9514, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9515, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF8, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9516, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9517, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9518, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9519, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9520, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9521, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9522, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9523, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9524, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9525, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9526, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9527, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9528, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9529, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9530, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9531, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9532, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9533, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9534, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9535, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9536, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9537, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9538, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9539, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9540, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0x1F, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9541, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9542, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9543, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9544, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0xFF, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9545, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0xF8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9546, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0x1F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9547, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xFF, 0xFF, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9548, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xEE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9549, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xEE, 0xEE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9550, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9551, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x0, 0x0, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x0, 0x0}},
    {9552, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9553, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9554, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0x8, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9555, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1F, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9556, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1F, 0x10, 0x17, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9557, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0x8, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9558, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9559, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x4, 0xF4, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9560, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF, 0x8, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9561, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9562, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x17, 0x10, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9563, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0x8, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9564, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9565, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xF4, 0x4, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9566, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF, 0x8, 0xF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9567, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x17, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9568, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x17, 0x10, 0x17, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9569, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xF8, 0x8, 0xF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9570, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xF4, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9571, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xF4, 0x4, 0xF4, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9572, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9573, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9574, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0xF7, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9575, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9576, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9577, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xF7, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9578, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFF, 0x8, 0xFF, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9579, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xFF, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9580, 1, {0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0xF7, 0x0, 0xF7, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14}},
    {9581, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9582, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xE0, 0x10, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9583, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x10, 0xE0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9584, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x4, 0x3, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9585, 1, {0x1, 0x1, 0x2, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x80, 0x80}},
    {9586, 1, {0x80, 0x80, 0x40, 0x40, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x2, 0x2, 0x1, 0x1}},
    {9587, 1, {0x81, 0x81, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x81, 0x81}},
    {9588, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9589, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9590, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9591, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9592, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9593, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9594, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9595, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9596, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9597, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}},
    {9598, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9599, 1, {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {9600, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9601, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF}},
    {9602, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9603, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9604, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9605, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9606, 1, {0x0, 0x0, 0x0, 0x0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9607, 1, {0x0, 0x0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9608, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9609, 1, {0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE}},
    {9610, 1, {0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC}},
    {9611, 1, {0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8}},
    {9612, 1, {0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0}},
    {9613, 1, {0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0, 0xE0}},
    {9614, 1, {0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0}},
    {9615, 1, {0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80}},
    {9616, 1, {0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF}},
    {9617, 1, {0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22}},
    {9618, 1, {0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55}},
    {9619, 1, {0xEE, 0xBB, 0xEE, 0xBB, 0xEE, 0xBB, 0xEE, 0xBB, 0xEE, 0xBB, 0xEE, 0xBB, 0xEE, 0xBB, 0xEE, 0xBB}},
    {9620, 1, {0xFF, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9621, 1, {0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1}},
    {9622, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0}},
    {9623, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF}},
    {9624, 1, {0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9625, 1, {0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9626, 1, {0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF}},
    {9627, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0}},
    {9628, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF}},
    {9629, 1, {0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9630, 1, {0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0}},
    {9631, 1, {0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9632, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x0, 0x0, 0x0}},
    {9633, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0, 0x0}},
    {9634, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x41, 0x41, 0x41, 0x41, 0x3E, 0x0, 0x0, 0x0}},
    {9635, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x5D, 0x5D, 0x5D, 0x41, 0x7F, 0x0, 0x0, 0x0}},
    {9636, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x7F, 0x41, 0x7F, 0x41, 0x7F, 0x0, 0x0, 0x0}},
    {9637, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x55, 0x55, 0x55, 0x55, 0x55, 0x7F, 0x0, 0x0, 0x0}},
    {9638, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x55, 0x7F, 0x55, 0x7F, 0x55, 0x7F, 0x0, 0x0, 0x0}},
    {9639, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x53, 0x49, 0x65, 0x53, 0x49, 0x7F, 0x0, 0x0, 0x0}},
    {9640, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x65, 0x49, 0x53, 0x65, 0x49, 0x7F, 0x0, 0x0, 0x0}},
    {9641, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x55, 0x6B, 0x55, 0x6B, 0x55, 0x7F, 0x0, 0x0, 0x0}},
    {9642, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x3C, 0x3C, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9643, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9644, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x7F, 0x7F, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9645, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9646, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x3C, 0x3C, 0x3C, 0x3C, 0x3C, 0x3C, 0x0, 0x0, 0x0, 0x0}},
    {9647, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x24, 0x24, 0x24, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0}},
    {9648, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x7E, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9649, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x42, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9650, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x3C, 0x3C, 0x7E, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9651, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9652, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x3C, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9653, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9654, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x78, 0x7E, 0x7E, 0x78, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9655, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x46, 0x46, 0x58, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9656, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x3C, 0x3C, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9657, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x2C, 0x2C, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9658, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x78, 0x7E, 0x78, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9659, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x46, 0x58, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9660, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x7E, 0x3C, 0x3C, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9661, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x24, 0x24, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9662, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x3C, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9663, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9664, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1E, 0x7E, 0x7E, 0x1E, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9665, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1A, 0x62, 0x62, 0x1A, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9666, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x3C, 0x3C, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9667, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x34, 0x34, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9668, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1E, 0x7E, 0x1E, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9669, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x1A, 0x62, 0x1A, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9670, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x3E, 0x7F, 0x3E, 0x1C, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {9671, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x41, 0x22, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {9672, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x2A, 0x5D, 0x2A, 0x14, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {9673, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x59, 0x5D, 0x4D, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9674, 1, {0x0, 0x0, 0x0, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x0, 0x0, 0x0}},
    {9675, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9676, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x41, 0x0, 0x41, 0x0, 0x14, 0x0, 0x0, 0x0, 0x0}},
    {9677, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x36, 0x55, 0x55, 0x55, 0x36, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9678, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x49, 0x55, 0x49, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9679, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x3E, 0x7F, 0x7F, 0x7F, 0x3E, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9680, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x32, 0x71, 0x71, 0x71, 0x32, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9681, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x26, 0x47, 0x47, 0x47, 0x26, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9682, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x7F, 0x3E, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9683, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x3E, 0x7F, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9684, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x2E, 0x4F, 0x4F, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9685, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x26, 0x47, 0x47, 0x7F, 0x3E, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9686, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x38, 0x78, 0x78, 0x78, 0x38, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {9687, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0xE, 0xF, 0xF, 0xF, 0xE, 0xC, 0x0, 0x0, 0x0, 0x0}},
    {9688, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xE7, 0xC3, 0xC3, 0xE7, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9689, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xE7, 0xDB, 0xBD, 0xBD, 0xDB, 0xE7, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9690, 1, {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xE7, 0xDB, 0xBD, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9691, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xBD, 0xDB, 0xE7, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},
    {9692, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x20, 0x40, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9693, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x2, 0x1, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9694, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x2, 0xC, 0x0, 0x0, 0x0, 0x0}},
    {9695, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {9696, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x41, 0x41, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9697, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x41, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {9698, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x6, 0xE, 0x1E, 0x3E, 0x7E, 0x0}},
    {9699, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x60, 0x70, 0x78, 0x7C, 0x7E, 0x0}},
    {9700, 1, {0x0, 0x0, 0x7E, 0x7C, 0x78, 0x70, 0x60, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9701, 1, {0x0, 0x0, 0x7E, 0x3E, 0x1E, 0xE, 0x6, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9702, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {9703, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x72, 0x72, 0x72, 0x72, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9704, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x4E, 0x4E, 0x4E, 0x4E, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9705, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x7A, 0x72, 0x62, 0x42, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9706, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x46, 0x4E, 0x5E, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9707, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9708, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x22, 0x2A, 0x41, 0x7F, 0x0, 0x0, 0x0, 0x0}},
    {9709, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x34, 0x34, 0x72, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {9710, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x2C, 0x2C, 0x4E, 0x4E, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {9711, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x81, 0x81, 0x81, 0x81, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0}},
    {9712, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x52, 0x72, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {9713, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x72, 0x52, 0x7E, 0x0, 0x0}},
    {9714, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x4E, 0x4A, 0x7E, 0x0, 0x0}},
    {9715, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x4A, 0x4E, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {9716, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x52, 0x91, 0xF1, 0x81, 0x81, 0x42, 0x3C, 0x0, 0x0}},
    {9717, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x81, 0x81, 0xF1, 0x91, 0x52, 0x3C, 0x0, 0x0}},
    {9718, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x81, 0x81, 0x8F, 0x89, 0x4A, 0x3C, 0x0, 0x0}},
    {9719, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x4A, 0x89, 0x8F, 0x81, 0x81, 0x42, 0x3C, 0x0, 0x0}},
    {9720, 1, {0x0, 0x0, 0x7E, 0x44, 0x48, 0x50, 0x60, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9721, 1, {0x0, 0x0, 0x7E, 0x22, 0x12, 0xA, 0x6, 0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9722, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x60, 0x50, 0x48, 0x44, 0x7E, 0x0}},
    {9723, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0}},
    {9724, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x0}},
    {9725, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x24, 0x24, 0x3C, 0x0, 0x0}},
    {9726, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x3C, 0x3C, 0x3C, 0x0, 0x0}},
    {9727, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x6, 0xA, 0x12, 0x22, 0x7E, 0x0}},
    {9728, 1, {0x0, 0x0, 0x8, 0x49, 0x22, 0x0, 0x1C, 0x3E, 0x3E, 0x1C, 0x0, 0x22, 0x49, 0x8, 0x0, 0x0}},
    {9729, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x76, 0xFF, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9730, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x3E, 0x7F, 0x8, 0x8, 0x8, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {9732, 1, {0x0, 0x0, 0x0, 0x22, 0x22, 0x24, 0x24, 0x28, 0x29, 0x2, 0x30, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {9733, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x7F, 0x1C, 0x36, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9734, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x77, 0x22, 0x2A, 0x36, 0x22, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {9735, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x10, 0xA, 0x6, 0xE, 0x0, 0x0}},
    {9736, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x44, 0x48, 0x50, 0x4A, 0x46, 0x4E, 0x0, 0x0}},
    {9737, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x81, 0x99, 0x99, 0x81, 0x42, 0x3C, 0x0, 0x0}},
    {9738, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x66, 0xA5, 0x42, 0x0, 0x0, 0x0, 0x0}},
    {9739, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0xA5, 0x66, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {9740, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x8, 0x30, 0x48, 0x48, 0x30, 0x0, 0x0, 0x0, 0x0}},
    {9741, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x9, 0x9, 0x6, 0x8, 0x30, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {9742, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x63, 0x0, 0x2A, 0x63, 0x7F, 0x0, 0x0, 0x0}},
    {9743, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x41, 0x5D, 0x3E, 0x49, 0x41, 0x7F, 0x0, 0x0, 0x0}},
    {9744, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x41, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9745, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x7F, 0x45, 0x45, 0x65, 0x59, 0x41, 0x7F, 0x0, 0x0}},
    {9746, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x41, 0x55, 0x49, 0x55, 0x41, 0x7F, 0x0, 0x0}},
    {9747, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x0, 0x0}},
    {9754, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x1F, 0x1F, 0xE, 0x0, 0x0, 0x0, 0x0}},
    {9755, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x7C, 0x7C, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {9756, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x11, 0x19, 0xE, 0x0, 0x0, 0x0, 0x0}},
    {9757, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x1C, 0x34, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {9758, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F, 0x44, 0x4C, 0x38, 0x0, 0x0, 0x0, 0x0}},
    {9759, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x34, 0x1C, 0x4, 0x4, 0x0, 0x0}},
    {9760, 1, {0x0, 0x0, 0xE7, 0xA5, 0x7E, 0x81, 0xA5, 0xA5, 0x81, 0x5A, 0x5A, 0x42, 0xBD, 0xE7, 0x0, 0x0}},
    {9761, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x1C, 0x0, 0x0}},
    {9765, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x14, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9766, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x8, 0x3E, 0x8, 0x18, 0xC, 0x8, 0x8, 0x0, 0x0}},
    {9767, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x12, 0x12, 0x1C, 0x10, 0x54, 0x38, 0x10, 0x38, 0x54, 0x0, 0x0}},
    {9768, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9769, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x8, 0x49, 0x7F, 0x49, 0x8, 0x1C, 0x0, 0x0}},
    {9770, 1, {0x0, 0x0, 0x0, 0x0, 0x1E, 0x21, 0x40, 0xC2, 0xC7, 0xC2, 0x40, 0x21, 0x1E, 0x0, 0x0, 0x0}},
    {9773, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x2, 0x39, 0x79, 0xD, 0x3E, 0x63, 0xC1, 0x0, 0x0}},
    {9774, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x49, 0x49, 0x49, 0x5D, 0x6B, 0x49, 0x3E, 0x0, 0x0}},
    {9784, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x3E, 0x3E, 0x77, 0x3E, 0x3E, 0x49, 0x0, 0x0}},
    {9785, 1, {0x0, 0x0, 0x0, 0x3C, 0x42, 0x81, 0xA5, 0x81, 0x99, 0xA5, 0x81, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {9786, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x42, 0x81, 0xA5, 0x81, 0xA5, 0x99, 0x42, 0x3C, 0x0, 0x0, 0x0}},
    {9787, 1, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x7E, 0xFF, 0xDB, 0xFF, 0xDB, 0xE7, 0x7E, 0x3C, 0x0, 0x0, 0x0}},
    {9788, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x2A, 0x1C, 0x77, 0x1C, 0x2A, 0x49, 0x0, 0x0, 0x0, 0x0}},
    {9789, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0xC, 0x6, 0x5, 0x5, 0x5, 0x5, 0x6, 0xC, 0x38, 0x0, 0x0}},
    {9790, 1, {0x0, 0x0, 0x0, 0x0, 0xE, 0x18, 0x30, 0x50, 0x50, 0x50, 0x50, 0x30, 0x18, 0xE, 0x0, 0x0}},
    {9791, 1, {0x0, 0x0, 0x0, 0x22, 0x22, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x8, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {9792, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x8, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {9793, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x0}},
    {9794, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7, 0x3, 0x5, 0x38, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0, 0x0}},
    {9795, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x3A, 0x46, 0x6, 0xA, 0x7E, 0x2, 0x2, 0x0, 0x0}},
    {9796, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x70, 0x20, 0x2C, 0x32, 0x22, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {9797, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6B, 0x2A, 0x3E, 0x2A, 0x6B, 0x8, 0x1C, 0x14, 0x1C, 0x0, 0x0}},
    {9798, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x59, 0xCB, 0x49, 0x49, 0x49, 0x3E, 0x8, 0x3E, 0x8, 0x8, 0x0}},
    {9799, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x42, 0x42, 0x42, 0x7C, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {9800, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x55, 0x14, 0x14, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {9801, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x41, 0x22, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x0, 0x0}},
    {9802, 1, {0x0, 0x0, 0x0, 0x0, 0x41, 0x3E, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x3E, 0x41, 0x0, 0x0}},
    {9803, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x49, 0x48, 0x30, 0x6, 0x9, 0x49, 0x3E, 0x0, 0x0}},
    {9804, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x12, 0x74, 0x94, 0x94, 0x65, 0x2, 0x0, 0x0}},
    {9805, 1, {0x0, 0x0, 0x0, 0x0, 0x54, 0x7C, 0x55, 0x57, 0x55, 0x55, 0x55, 0x55, 0x6, 0x7C, 0x0, 0x0}},
    {9806, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x22, 0x22, 0x77, 0x0, 0x7F, 0x0, 0x0, 0x0}},
    {9807, 1, {0x0, 0x0, 0x0, 0x0, 0x54, 0x7C, 0x54, 0x54, 0x54, 0x54, 0x54, 0x54, 0x55, 0x3, 0x0, 0x0}},
    {9808, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0x3, 0x45, 0x29, 0x10, 0x28, 0x44, 0x0, 0x0}},
    {9809, 1, {0x0, 0x0, 0x0, 0x0, 0x31, 0x4E, 0x84, 0x8, 0x10, 0x3C, 0x62, 0xA2, 0xA2, 0x1C, 0x0, 0x0}},
    {9810, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2A, 0x54, 0x0, 0x0, 0x2A, 0x54, 0x0, 0x0, 0x0, 0x0}},
    {9811, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x24, 0x7E, 0x24, 0x24, 0x42, 0x42, 0x0, 0x0}},
    {9812, 1, {0x0, 0x8, 0x1C, 0x49, 0x5D, 0x63, 0x41, 0x22, 0x22, 0x22, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9813, 1, {0x0, 0x0, 0x0, 0x8, 0x1C, 0x63, 0x41, 0x22, 0x22, 0x22, 0x41, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9814, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x55, 0x7F, 0x41, 0x41, 0x22, 0x22, 0x41, 0x41, 0x7F, 0x0, 0x0}},
    {9815, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x22, 0x22, 0x22, 0x14, 0x77, 0x41, 0x7F, 0x0, 0x0}},
    {9816, 1, {0x0, 0x0, 0x0, 0x2, 0x1E, 0x22, 0x42, 0x42, 0x3A, 0x12, 0x22, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {9817, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x24, 0x24, 0x42, 0x7E, 0x0, 0x0}},
    {9818, 1, {0x0, 0x8, 0x1C, 0x49, 0x5D, 0x7F, 0x7F, 0x3E, 0x3E, 0x3E, 0x7F, 0x7F, 0x7F, 0x7F, 0x0, 0x0}},
    {9819, 1, {0x0, 0x0, 0x0, 0x8, 0x1C, 0x7F, 0x77, 0x3E, 0x3E, 0x3E, 0x7F, 0x7F, 0x7F, 0x7F, 0x0, 0x0}},
    {9820, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x55, 0x7F, 0x7F, 0x3E, 0x3E, 0x7F, 0x7F, 0x7F, 0x0, 0x0}},
    {9821, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x36, 0x22, 0x36, 0x3E, 0x1C, 0x7F, 0x7F, 0x7F, 0x0, 0x0}},
    {9822, 1, {0x0, 0x0, 0x0, 0x2, 0x1E, 0x36, 0x7E, 0x7E, 0x3E, 0x1E, 0x3E, 0x7E, 0x7E, 0x7E, 0x0, 0x0}},
    {9823, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x3C, 0x3C, 0x18, 0x3C, 0x3C, 0x7E, 0x7E, 0x0, 0x0}},
    {9824, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x1C, 0x3E, 0x7F, 0x7F, 0x7F, 0x3E, 0x8, 0x1C, 0x0, 0x0}},
    {9825, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x49, 0x41, 0x41, 0x41, 0x22, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {9826, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x14, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0, 0x0}},
    {9827, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x1C, 0x1C, 0x8, 0x7F, 0x7F, 0x6B, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {9828, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x14, 0x22, 0x41, 0x41, 0x41, 0x3E, 0x8, 0x1C, 0x0, 0x0}},
    {9829, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x36, 0x7F, 0x7F, 0x7F, 0x7F, 0x3E, 0x1C, 0x8, 0x8, 0x0, 0x0}},
    {9830, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x1C, 0x1C, 0x3E, 0x3E, 0x1C, 0x1C, 0x8, 0x8, 0x0, 0x0}},
    {9831, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x14, 0x1C, 0x8, 0x77, 0x5D, 0x6B, 0x8, 0x8, 0x1C, 0x0, 0x0}},
    {9832, 1, {0x0, 0x0, 0x0, 0x8, 0x51, 0x92, 0x92, 0x92, 0x49, 0x49, 0x49, 0x52, 0x81, 0x7E, 0x0, 0x0}},
    {9833, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x38, 0x78, 0x70, 0x0, 0x0}},
    {9834, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0xC, 0xA, 0xA, 0x8, 0x8, 0x8, 0x38, 0x78, 0x70, 0x0, 0x0}},
    {9835, 1, {0x0, 0x0, 0x0, 0x1C, 0x17, 0x11, 0x11, 0x11, 0x11, 0x71, 0xF1, 0xE7, 0xF, 0xE, 0x0, 0x0}},
    {9836, 1, {0x0, 0x0, 0x0, 0x1C, 0x17, 0x11, 0x1D, 0x17, 0x11, 0x71, 0xF1, 0xE7, 0xF, 0xE, 0x0, 0x0}},
    {9837, 1, {0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x5C, 0x66, 0x46, 0x44, 0x48, 0x50, 0x60, 0x0, 0x0}},
    {9838, 1, {0x0, 0x0, 0x0, 0x40, 0x40, 0x4E, 0x7E, 0x72, 0x42, 0x4E, 0x7E, 0x72, 0x2, 0x2, 0x0, 0x0}},
    {9839, 1, {0x0, 0x0, 0x4, 0x24, 0x26, 0x3E, 0x7C, 0x64, 0x24, 0x26, 0x3E, 0x7C, 0x64, 0x24, 0x20, 0x0}},
    {9840, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x8, 0x8, 0x49, 0x3E, 0x49, 0x8, 0x8, 0x8, 0x14, 0x0, 0x0}},
    {9841, 1, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x14, 0x8, 0x6B, 0x5D, 0x6B, 0x8, 0x8, 0x14, 0x8, 0x0, 0x0}},
    {10088, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0xC, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xC, 0x4, 0x2, 0x0}},
    {10089, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x18, 0xC, 0xC, 0xC, 0xC, 0xC, 0xC, 0x18, 0x10, 0x20, 0x0}},
    {10090, 1, {0x0, 0x0, 0x0, 0x6, 0xC, 0x1C, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x1C, 0xC, 0x6, 0x0}},
    {10091, 1, {0x0, 0x0, 0x0, 0x30, 0x18, 0x1C, 0xE, 0xE, 0xE, 0xE, 0xE, 0xE, 0x1C, 0x18, 0x30, 0x0}},
    {10092, 1, {0x0, 0x0, 0x0, 0x6, 0xC, 0xC, 0x18, 0x18, 0x30, 0x30, 0x18, 0x18, 0xC, 0xC, 0x6, 0x0}},
    {10093, 1, {0x0, 0x0, 0x0, 0x30, 0x18, 0x18, 0xC, 0xC, 0x6, 0x6, 0xC, 0xC, 0x18, 0x18, 0x30, 0x0}},
    {10094, 1, {0x0, 0x0, 0x0, 0x7, 0xE, 0xE, 0x1C, 0x1C, 0x38, 0x38, 0x1C, 0x1C, 0xE, 0xE, 0x7, 0x0}},
    {10095, 1, {0x0, 0x0, 0x0, 0x70, 0x38, 0x38, 0x1C, 0x1C, 0xE, 0xE, 0x1C, 0x1C, 0x38, 0x38, 0x70, 0x0}},
    {10096, 1, {0x0, 0x0, 0x0, 0xF, 0x1E, 0x1E, 0x3C, 0x3C, 0x78, 0x78, 0x3C, 0x3C, 0x1E, 0x1E, 0xF, 0x0}},
    {10097, 1, {0x0, 0x0, 0x0, 0xF0, 0x78, 0x78, 0x3C, 0x3C, 0x1E, 0x1E, 0x3C, 0x3C, 0x78, 0x78, 0xF0, 0x0}},
    {10098, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x8, 0x4, 0x0}},
    {10099, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x10, 0x20, 0x0}},
    {10100, 1, {0x0, 0x0, 0x0, 0x1E, 0x38, 0x30, 0x30, 0x18, 0x18, 0x70, 0x18, 0x18, 0x30, 0x38, 0x1E, 0x0}},
    {10101, 1, {0x0, 0x0, 0x0, 0x78, 0x1C, 0xC, 0xC, 0x18, 0x18, 0xE, 0x18, 0x18, 0xC, 0x1C, 0x78, 0x0}},
    {10193, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x28, 0x28, 0x44, 0x54, 0x82, 0x82, 0x0, 0x0}},
    {10195, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x12, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {10196, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x48, 0x40, 0x40, 0x0, 0x0}},
    {10208, 1, {0x0, 0x0, 0x0, 0x10, 0x28, 0x28, 0x44, 0x44, 0xFE, 0x44, 0x44, 0x28, 0x28, 0x10, 0x0, 0x0}},
    {10214, 1, {0x0, 0x3E, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x3E, 0x0, 0x0}},
    {10215, 1, {0x0, 0x7C, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x7C, 0x0, 0x0}},
    {10216, 1, {0x0, 0x0, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x0, 0x0}},
    {10217, 1, {0x0, 0x0, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x0, 0x0}},
    {10218, 1, {0x0, 0x0, 0xA, 0x14, 0x14, 0x28, 0x28, 0x50, 0x50, 0x28, 0x28, 0x14, 0x14, 0xA, 0x0, 0x0}},
    {10219, 1, {0x0, 0x0, 0x50, 0x28, 0x28, 0x14, 0x14, 0xA, 0xA, 0x14, 0x14, 0x28, 0x28, 0x50, 0x0, 0x0}},
    {10240, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10241, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10242, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10243, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10244, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10245, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10246, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10247, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10248, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10249, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10250, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10251, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10252, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10253, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10254, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10255, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10256, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10257, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10258, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10259, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10260, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10261, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10262, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10263, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10264, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10265, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10266, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10267, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10268, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10269, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10270, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10271, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10272, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10273, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10274, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10275, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10276, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10277, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10278, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10279, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10280, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10281, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10282, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10283, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10284, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10285, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10286, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10287, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10288, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10289, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10290, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10291, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10292, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10293, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10294, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10295, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10296, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10297, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10298, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10299, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10300, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10301, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10302, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10303, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x0}},
    {10304, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10305, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10306, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10307, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10308, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10309, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10310, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10311, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10312, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10313, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10314, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10315, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10316, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10317, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10318, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10319, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10320, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10321, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10322, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10323, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10324, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10325, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10326, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10327, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10328, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10329, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10330, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10331, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10332, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10333, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10334, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10335, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10336, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10337, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10338, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10339, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10340, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10341, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10342, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10343, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10344, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10345, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10346, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10347, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10348, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10349, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10350, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10351, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10352, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10353, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10354, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10355, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10356, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10357, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10358, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10359, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10360, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10361, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10362, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10363, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10364, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10365, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10366, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10367, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x0}},
    {10368, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10369, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10370, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10371, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10372, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10373, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10374, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10375, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10376, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10377, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10378, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10379, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10380, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10381, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10382, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10383, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10384, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10385, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10386, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10387, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10388, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10389, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10390, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10391, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10392, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10393, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10394, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10395, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10396, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10397, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10398, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10399, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10400, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10401, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10402, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10403, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10404, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10405, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10406, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10407, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10408, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10409, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10410, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10411, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10412, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10413, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10414, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10415, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10416, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10417, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10418, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10419, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10420, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10421, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10422, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10423, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10424, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10425, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10426, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10427, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10428, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10429, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10430, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10431, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x0}},
    {10432, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10433, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10434, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10435, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10436, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10437, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10438, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10439, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10440, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10441, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10442, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10443, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10444, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10445, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10446, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10447, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10448, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10449, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10450, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10451, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10452, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10453, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10454, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10455, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10456, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10457, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10458, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10459, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10460, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10461, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10462, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10463, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10464, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10465, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10466, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10467, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10468, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10469, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10470, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10471, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10472, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10473, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10474, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10475, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10476, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10477, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10478, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10479, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10480, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10481, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10482, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10483, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10484, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10485, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10486, 1, {0x0, 0x0, 0x0, 0x22, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10487, 1, {0x0, 0x0, 0x0, 0x32, 0x30, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10488, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10489, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10490, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10491, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10492, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10493, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10494, 1, {0x0, 0x0, 0x0, 0x26, 0x6, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10495, 1, {0x0, 0x0, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x36, 0x36, 0x0, 0x0}},
    {10504, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0}},
    {10505, 1, {0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {10514, 1, {0x0, 0x0, 0x7F, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {10515, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x7F, 0x0, 0x0}},
    {10552, 1, {0x0, 0x0, 0x20, 0x10, 0x8, 0x8, 0x4, 0x4, 0x4, 0x4, 0x28, 0x28, 0x30, 0x3C, 0x0, 0x0}},
    {10553, 1, {0x0, 0x0, 0x4, 0x8, 0x10, 0x10, 0x20, 0x20, 0x20, 0x20, 0x14, 0x14, 0xC, 0x3C, 0x0, 0x0}},
    {10569, 1, {0x0, 0x0, 0x10, 0x38, 0x54, 0x38, 0x54, 0x10, 0x10, 0x10, 0x38, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {10572, 1, {0x0, 0x0, 0x10, 0x18, 0x14, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x0, 0x0}},
    {10573, 1, {0x0, 0x0, 0x10, 0x30, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x14, 0x18, 0x10, 0x0, 0x0}},
    {10575, 1, {0x0, 0x0, 0x20, 0x30, 0x28, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x28, 0x30, 0x20, 0x0, 0x0}},
    {10577, 1, {0x0, 0x0, 0x8, 0x18, 0x28, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x28, 0x18, 0x8, 0x0, 0x0}},
    {10580, 1, {0x0, 0x0, 0x3E, 0x8, 0xC, 0xA, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {10581, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xA, 0xC, 0x8, 0x3E, 0x0, 0x0}},
    {10584, 1, {0x0, 0x0, 0x3E, 0x8, 0x18, 0x28, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {10585, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x28, 0x18, 0x8, 0x3E, 0x0, 0x0}},
    {10588, 1, {0x0, 0x0, 0x10, 0x18, 0x14, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0}},
    {10589, 1, {0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x14, 0x18, 0x10, 0x0, 0x0}},
    {10592, 1, {0x0, 0x0, 0x10, 0x30, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0}},
    {10593, 1, {0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x0, 0x0}},
    {10620, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x40, 0x40, 0x20, 0x1F, 0x20, 0x40, 0x40, 0x20, 0x0, 0x0}},
    {10621, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x1, 0x1, 0x2, 0x7C, 0x2, 0x1, 0x1, 0x2, 0x0, 0x0}},
    {10624, 1, {0x0, 0x0, 0x0, 0x0, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x0, 0x0}},
    {10626, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {10627, 1, {0x0, 0x0, 0x1E, 0x24, 0x24, 0x14, 0x14, 0x24, 0x44, 0x24, 0x14, 0x14, 0x24, 0x24, 0x1E, 0x0}},
    {10628, 1, {0x0, 0x0, 0x78, 0x24, 0x24, 0x28, 0x28, 0x24, 0x22, 0x24, 0x28, 0x28, 0x24, 0x24, 0x78, 0x0}},
    {10629, 1, {0x0, 0x0, 0x0, 0xC, 0x18, 0x18, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x18, 0x18, 0xC, 0x0}},
    {10630, 1, {0x0, 0x0, 0x0, 0x30, 0x18, 0x18, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x18, 0x18, 0x30, 0x0}},
    {10631, 1, {0x0, 0x0, 0x0, 0xC, 0x14, 0x14, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x14, 0x14, 0xC, 0x0}},
    {10632, 1, {0x0, 0x0, 0x0, 0x30, 0x28, 0x28, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x28, 0x28, 0x30, 0x0}},
    {10633, 1, {0x0, 0x0, 0x0, 0x4, 0xC, 0xC, 0x14, 0x14, 0x24, 0x24, 0x14, 0x14, 0xC, 0xC, 0x4, 0x0}},
    {10634, 1, {0x0, 0x0, 0x0, 0x20, 0x30, 0x30, 0x28, 0x28, 0x24, 0x24, 0x28, 0x28, 0x30, 0x30, 0x20, 0x0}},
    {10635, 1, {0x0, 0x0, 0x0, 0xE, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x0, 0xE}},
    {10636, 1, {0x0, 0x0, 0x0, 0x70, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x70, 0x0, 0x70}},
    {10637, 1, {0x0, 0x0, 0x0, 0x1E, 0x18, 0x14, 0x12, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1E, 0x0}},
    {10638, 1, {0x0, 0x0, 0x0, 0x78, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x48, 0x28, 0x18, 0x78, 0x0}},
    {10639, 1, {0x0, 0x0, 0x0, 0x1E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x12, 0x14, 0x18, 0x1E, 0x0}},
    {10640, 1, {0x0, 0x0, 0x0, 0x78, 0x18, 0x28, 0x48, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x78, 0x0}},
    {10641, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x8, 0x10, 0x10, 0x26, 0x26, 0x10, 0x10, 0x8, 0x8, 0x4, 0x0}},
    {10642, 1, {0x0, 0x0, 0x0, 0x20, 0x10, 0x10, 0x8, 0x8, 0x64, 0x64, 0x8, 0x8, 0x10, 0x10, 0x20, 0x0}},
    {10647, 1, {0x0, 0x0, 0x0, 0xC, 0x18, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x18, 0xC, 0x0}},
    {10648, 1, {0x0, 0x0, 0x0, 0x30, 0x18, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x18, 0x30, 0x0}},
    {10649, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x0, 0x1, 0x0, 0x1, 0x0, 0x1, 0x0, 0x0}},
    {10662, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x20, 0x1E, 0x0, 0x0}},
    {10663, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E, 0x20, 0x40, 0x0, 0x0}},
    {10705, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x82, 0xC6, 0xEA, 0xF2, 0xEA, 0xC6, 0x82, 0x0, 0x0}},
    {10706, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x82, 0xC6, 0xAE, 0x9E, 0xAE, 0xC6, 0x82, 0x0, 0x0}},
    {10707, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x82, 0xC6, 0xEE, 0xFE, 0xEE, 0xC6, 0x82, 0x0, 0x0}},
    {10708, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x82, 0xC4, 0xE8, 0xF0, 0xE8, 0xC4, 0x82, 0x0, 0x0}},
    {10709, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x82, 0x46, 0x2E, 0x1E, 0x2E, 0x46, 0x82, 0x0, 0x0}},
    {10710, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFE, 0x44, 0x28, 0x10, 0x28, 0x44, 0xFE, 0x0, 0x0}},
    {10711, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFE, 0x7C, 0x38, 0x10, 0x38, 0x7C, 0xFE, 0x0, 0x0}},
    {10712, 1, {0x0, 0x0, 0x20, 0x10, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x20, 0x0}},
    {10713, 1, {0x0, 0x0, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x20, 0x10, 0x8, 0x0}},
    {10731, 1, {0x0, 0x0, 0x10, 0x38, 0x38, 0x7C, 0x7C, 0xFE, 0x7C, 0x7C, 0x38, 0x38, 0x10, 0x0, 0x0, 0x0}},
    {10741, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x0, 0x0}},
    {10742, 1, {0x0, 0x0, 0x78, 0x0, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {10743, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x20, 0x20, 0x7C, 0x10, 0x8, 0x8, 0x4, 0x4, 0x0, 0x0}},
    {10744, 1, {0x1, 0x1, 0x2, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x0, 0x0}},
    {10745, 1, {0x80, 0x80, 0x40, 0x40, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x4, 0x4, 0x2, 0x2, 0x0, 0x0}},
    {10746, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x28, 0xFE, 0x28, 0x28, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {10747, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x54, 0xFE, 0x54, 0x54, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {10748, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x18, 0x60, 0x80, 0x60, 0x18, 0x4, 0x2, 0x2, 0x0, 0x0}},
    {10749, 1, {0x0, 0x0, 0x0, 0x80, 0x80, 0x40, 0x30, 0xC, 0x2, 0xC, 0x30, 0x40, 0x80, 0x80, 0x0, 0x0}},
    {10782, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0xE, 0x32, 0x42, 0x32, 0xE, 0x2, 0x0, 0x0, 0x0}},
    {10783, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x18, 0x24, 0x24, 0x1C, 0x4, 0x18}},
    {10785, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x30, 0x28, 0x24, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0, 0x0}},
    {10786, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x10, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0, 0x0}},
    {10787, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x44, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0, 0x0}},
    {10788, 1, {0x0, 0x0, 0x0, 0x0, 0x64, 0x98, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0, 0x0}},
    {10789, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x10, 0x0}},
    {10790, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x32, 0x4C}},
    {10792, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x38, 0x7C, 0x10, 0x0}},
    {10793, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x38, 0x8, 0x30, 0x0, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {10794, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x0, 0x10, 0x0, 0x0, 0x0}},
    {10795, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x7C, 0x0, 0x4, 0x0, 0x0, 0x0}},
    {10796, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x7C, 0x0, 0x40, 0x0, 0x0, 0x0}},
    {10799, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x14, 0x8, 0x14, 0x22, 0x0, 0x0, 0x0}},
    {10800, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x0, 0x0, 0x0}},
    {10801, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x0, 0x7C, 0x0}},
    {10802, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x7C, 0x0, 0x0}},
    {10812, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {10813, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {10814, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x14, 0x8, 0x0, 0x8, 0x14, 0x14, 0xC, 0x4, 0x18}},
    {10822, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x18, 0x0, 0x18, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {10823, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x24, 0x0, 0x24, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {10824, 1, {0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x18, 0x0, 0x3C, 0x0, 0x18, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {10825, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x24, 0x0, 0x3C, 0x0, 0x24, 0x24, 0x24, 0x18, 0x0, 0x0}},
    {10826, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x92, 0x92, 0x92, 0x6C, 0x0, 0x0}},
    {10827, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x92, 0x92, 0x92, 0x0, 0x0}},
    {10854, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7F, 0x0, 0x7F, 0x0, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {10855, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x7F, 0x0, 0x7F, 0x0, 0x7F, 0x0, 0x0, 0x0, 0x0}},
    {10858, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x64, 0x98, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {10859, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x64, 0x98, 0x0, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {10860, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x64, 0x98, 0x0, 0xFC, 0x0, 0x64, 0x98, 0x0, 0x0}},
    {10861, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x64, 0x98, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x0, 0x0}},
    {10863, 1, {0x0, 0x0, 0x0, 0x30, 0x48, 0x84, 0x0, 0x64, 0x98, 0x0, 0x64, 0x98, 0x0, 0x0, 0x0, 0x0}},
    {10864, 1, {0x0, 0x0, 0x0, 0x0, 0x64, 0x98, 0x0, 0x64, 0x98, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x0, 0x0}},
    {10865, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {10866, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x0}},
    {10867, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x98, 0x64, 0x0, 0x0, 0x0}},
    {10871, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x48, 0x0}},
    {10891, 1, {0x0, 0x18, 0x60, 0x80, 0x60, 0x18, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x60, 0x18, 0x4, 0x18, 0x60}},
    {10892, 1, {0x0, 0x60, 0x18, 0x4, 0x18, 0x60, 0x0, 0xFC, 0x0, 0xFC, 0x0, 0x18, 0x60, 0x80, 0x60, 0x18}},
    {10895, 1, {0x0, 0x18, 0x60, 0x80, 0x60, 0x18, 0x0, 0x64, 0x98, 0x0, 0x60, 0x18, 0x4, 0x18, 0x60, 0x0}},
    {10896, 1, {0x0, 0x60, 0x18, 0x4, 0x18, 0x60, 0x0, 0x98, 0x64, 0x0, 0x18, 0x60, 0x80, 0x60, 0x18, 0x0}},
    {10897, 1, {0x0, 0xC, 0x30, 0x40, 0x30, 0xC, 0x60, 0x18, 0x4, 0x18, 0x60, 0x0, 0x7C, 0x0, 0x7C, 0x0}},
    {10898, 1, {0x0, 0x60, 0x18, 0x4, 0x18, 0x60, 0xC, 0x30, 0x40, 0x30, 0xC, 0x0, 0x7C, 0x0, 0x7C, 0x0}},
    {10899, 1, {0x0, 0xC, 0x30, 0x40, 0x30, 0xC, 0x40, 0x30, 0xC, 0x60, 0x18, 0x4, 0x18, 0x62, 0xC, 0x30}},
    {10900, 1, {0x0, 0x30, 0xC, 0x2, 0xC, 0x30, 0x2, 0xC, 0x30, 0x6, 0x18, 0x20, 0x18, 0x46, 0x30, 0xC}},
    {10943, 1, {0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x8, 0x8, 0x3E, 0x8, 0x8, 0x0, 0x0}},
    {10944, 1, {0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x0, 0x0}},
    {10945, 1, {0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x22, 0x14, 0x8, 0x14, 0x22, 0x0, 0x0}},
    {10946, 1, {0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x44, 0x28, 0x10, 0x28, 0x44, 0x0, 0x0}},
    {10947, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x7E, 0x0, 0x0}},
    {10948, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x30, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x7E, 0x0, 0x0}},
    {10949, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x0}},
    {10950, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x7E, 0x0, 0x7E, 0x0, 0x0}},
    {10953, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x32, 0x4C, 0x32, 0x4C, 0x0}},
    {10954, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x32, 0x4C, 0x32, 0x4C, 0x0}},
    {10955, 1, {0x0, 0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x4, 0x7E, 0x8, 0x7E, 0x10, 0x0}},
    {10956, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x4, 0x7E, 0x8, 0x7E, 0x10, 0x0}},
    {10963, 1, {0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x0}},
    {10964, 1, {0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0}},
    {10965, 1, {0x0, 0x0, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x3E, 0x40, 0x40, 0x40, 0x3E, 0x0, 0x0}},
    {10966, 1, {0x0, 0x0, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x7C, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x0}},
    {10974, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {10990, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x18, 0x14, 0x10, 0x10, 0x0, 0x0}},
    {10991, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x44, 0x44, 0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {10992, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {10993, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {10994, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x28, 0x28, 0x28, 0x28, 0xFE, 0x28, 0x28, 0x28, 0x28, 0x0, 0x0}},
    {10998, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x10, 0x0, 0x0, 0x10, 0x0, 0x0}},
    {11006, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x28, 0x28, 0x28, 0x28, 0x38, 0x0, 0x0}},
    {11007, 1, {0x0, 0x0, 0x0, 0x0, 0x38, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x38, 0x0, 0x0}},
    {11904, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x180, 0x60, 0x10, 0x1C00, 0x300, 0xC0, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {11905, 2, {0x0, 0x7C, 0x1F80, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {11906, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {11907, 2, {0x0, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1010, 0x1010, 0x1010, 0xFF0}},
    {11908, 2, {0x0, 0x0, 0xFFE0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x10, 0x12, 0xA, 0x6, 0x2}},
    {11909, 2, {0x40, 0x40, 0x80, 0x100, 0x100, 0x300, 0x500, 0x900, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {11910, 2, {0x0, 0xFF0, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x1010, 0x1010, 0x1010, 0x2050, 0x4020}},
    {11911, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x2008, 0x200A, 0x200A, 0x4006, 0x8002}},
    {11912, 2, {0x400, 0x400, 0x7F0, 0x820, 0x840, 0x1000, 0x2000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11913, 2, {0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4, 0x4, 0x14, 0x8}},
    {11914, 2, {0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x7F0, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400}},
    {11915, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x2010, 0x2010, 0x2010, 0x2050, 0x2020, 0x2004, 0x2004, 0x1FFC}},
    {11916, 2, {0x100, 0x1110, 0x920, 0x100, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11917, 2, {0x208, 0x2108, 0x1090, 0x820, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11918, 2, {0x0, 0x0, 0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x440, 0x440, 0x840, 0x840, 0x1042, 0x1042, 0x2042, 0x403E, 0x8000}},
    {11919, 2, {0x220, 0x210, 0x408, 0x804, 0x1002, 0x6422, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x822, 0x822, 0x1022, 0x601E}},
    {11920, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x240, 0x240, 0x240, 0x440, 0x440, 0x840, 0x842, 0x1042, 0x2042, 0xC03E}},
    {11921, 2, {0x220, 0x210, 0x408, 0x804, 0x1002, 0x67E2, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x822, 0x822, 0x1022, 0x601E}},
    {11922, 2, {0x0, 0x0, 0x3FF0, 0x2010, 0x2010, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x2000, 0x2000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {11923, 2, {0x0, 0x100, 0x100, 0x210, 0x410, 0x820, 0x1040, 0x1F80, 0x100, 0x200, 0x420, 0x810, 0x1008, 0x3FFC, 0x4, 0x4}},
    {11924, 2, {0x0, 0x400, 0x400, 0x400, 0x7F0, 0x420, 0x820, 0x820, 0x820, 0x840, 0x1FC0, 0x40, 0x40, 0x40, 0x7FFC, 0x0}},
    {11925, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x8, 0x3FF8, 0x8, 0x0, 0x0, 0x0}},
    {11926, 2, {0x800, 0x800, 0x800, 0x800, 0xC00, 0x4A00, 0x4A00, 0x8800, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800}},
    {11927, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x200, 0x220, 0x290, 0x1248, 0x1224, 0x2214, 0x2210, 0x4200, 0xA00, 0x400}},
    {11928, 2, {0x100, 0x100, 0x100, 0x1FC0, 0x100, 0x100, 0x100, 0x1C0, 0x700, 0x1900, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {11929, 2, {0x800, 0x800, 0x800, 0xFFC, 0x1810, 0x2410, 0x4420, 0x8220, 0x140, 0x80, 0x180, 0x240, 0x420, 0x810, 0x1008, 0xE006}},
    {11931, 2, {0x3FF0, 0x200, 0x200, 0x4200, 0x4200, 0x7FFC, 0x480, 0x480, 0x480, 0x880, 0x880, 0x1084, 0x2084, 0x407C, 0x8000, 0x0}},
    {11932, 2, {0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11933, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1050, 0x1020}},
    {11934, 2, {0x80, 0x80, 0xFE, 0x880, 0x880, 0xFF8, 0x1008, 0x1410, 0x2210, 0x4120, 0x140, 0x80, 0x100, 0x200, 0xC00, 0x7000}},
    {11935, 2, {0x0, 0x1FF0, 0x1010, 0x1210, 0x1110, 0x1110, 0xFFFE, 0x2010, 0x2210, 0x2110, 0x2110, 0x3FFC, 0x10, 0x10, 0xA0, 0x40}},
    {11936, 2, {0x0, 0x10, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2000, 0x2088, 0x3FFC, 0x2080, 0x2040, 0x2020, 0x2010, 0x380E, 0x3004}},
    {11937, 2, {0x2000, 0x1000, 0x1000, 0x8000, 0x4000, 0x4800, 0x800, 0x1000, 0x1000, 0xE000, 0x2000, 0x2000, 0x2000, 0x2000, 0x0, 0x0}},
    {11938, 2, {0x100, 0x100, 0x100, 0x2108, 0x1110, 0x920, 0x100, 0x140, 0x520, 0x910, 0x1108, 0xE104, 0x4104, 0x100, 0x500, 0x200}},
    {11939, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2488, 0x2244, 0x4244, 0x8004, 0x0, 0x0}},
    {11940, 2, {0x10, 0x7C, 0x780, 0xF808, 0x2108, 0x1890, 0x890, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11941, 2, {0x7C, 0x1F90, 0x1110, 0x1110, 0x2108, 0x2104, 0x4002, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11942, 2, {0x200, 0x200, 0x200, 0x2200, 0x1200, 0xA00, 0x200, 0x200, 0x600, 0xA00, 0x1200, 0x6200, 0x2200, 0x200, 0x200, 0x200}},
    {11943, 2, {0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11944, 2, {0x4400, 0x2800, 0x1000, 0x2800, 0x4800, 0x8800, 0x800, 0x1800, 0x2800, 0x4800, 0x8800, 0x800, 0x800, 0x5000, 0x2000, 0x0}},
    {11945, 2, {0xFE00, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0xFE00, 0x1000, 0x1000, 0x1000, 0x1000, 0x1600, 0x3800, 0xC000, 0x0, 0x0}},
    {11946, 2, {0x0, 0x0, 0xFC00, 0x800, 0x1000, 0x1000, 0x1000, 0x5C00, 0x5000, 0x5000, 0x5000, 0x5000, 0x5C00, 0xF000, 0x4000, 0x0}},
    {11947, 2, {0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11948, 2, {0x0, 0x7E00, 0x0, 0x0, 0xFE00, 0x1000, 0x1000, 0x5400, 0x5200, 0x5200, 0x9000, 0x1000, 0x1000, 0x1000, 0x5000, 0x2000}},
    {11949, 2, {0x2000, 0x1000, 0x1000, 0xF800, 0x800, 0x1000, 0x1000, 0x3800, 0x5400, 0x9400, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {11950, 2, {0x0, 0x1040, 0x1040, 0x1040, 0x3EFE, 0x2890, 0x4910, 0x8A10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11951, 2, {0x800, 0x800, 0x1000, 0x1200, 0x2200, 0x7C00, 0x800, 0x1000, 0x2400, 0x7E00, 0x200, 0x0, 0x5200, 0x4900, 0x8900, 0x0}},
    {11952, 2, {0x1000, 0x1000, 0x2000, 0x2400, 0x4400, 0xF800, 0x1000, 0x2000, 0x4000, 0xFC00, 0x4000, 0x0, 0x1C00, 0xE000, 0x4000, 0x0}},
    {11953, 2, {0x3FF8, 0x2008, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2008, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11954, 2, {0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11955, 2, {0x0, 0x3FFC, 0x2004, 0x4008, 0x440, 0x820, 0x1010, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11956, 2, {0x0, 0x7FFE, 0x4444, 0x8448, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x840, 0x842, 0x1042, 0x1042, 0x2042, 0x403E}},
    {11957, 2, {0x0, 0x0, 0x0, 0x3FFC, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2004, 0x0, 0x0}},
    {11958, 2, {0x1010, 0x820, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x200, 0xFFFE, 0x400, 0x800, 0x1000, 0xE000, 0x0, 0x0, 0x0, 0x0}},
    {11959, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11960, 2, {0x440, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x0, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {11961, 2, {0x200, 0x208, 0x3FC8, 0x210, 0x220, 0xFFFE, 0x80, 0x100, 0x200, 0xC00, 0x3000, 0xC000, 0x0, 0x0, 0x0, 0x0}},
    {11962, 2, {0x100, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x108, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {11963, 2, {0x100, 0x100, 0x1FF8, 0x108, 0xFFFE, 0x108, 0x1FF8, 0x100, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11964, 2, {0x7F8, 0x408, 0x508, 0x488, 0x448, 0x468, 0x408, 0x418, 0x428, 0x448, 0x588, 0x888, 0x808, 0x1008, 0x2028, 0x4010}},
    {11965, 2, {0x200, 0xC00, 0x707C, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x7CFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x7CFC, 0x4004}},
    {11966, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11967, 2, {0x820, 0x820, 0xFEFE, 0x820, 0x820, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11968, 2, {0x0, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0xF83E, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x0}},
    {11969, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2000, 0x23E0, 0x2220, 0x2220, 0x4424, 0x4824, 0x901C}},
    {11970, 2, {0x2000, 0x1000, 0x0, 0xFC00, 0x800, 0x1000, 0x1000, 0x3400, 0x5800, 0x9400, 0x1400, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {11971, 2, {0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11972, 2, {0x0, 0x0, 0xFFFE, 0x480, 0x480, 0x480, 0x7FFC, 0x4484, 0x4484, 0x4484, 0x4484, 0x4894, 0x48F4, 0x5004, 0x4004, 0x7FFC}},
    {11973, 2, {0x1FF0, 0x1010, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1290, 0x1290, 0x480, 0x480, 0x882, 0x3082, 0xC07E, 0x0}},
    {11974, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x4108, 0x4128, 0x8010}},
    {11975, 2, {0x400, 0x400, 0x7F0, 0x820, 0x1040, 0x3FF8, 0x5108, 0x1188, 0x1248, 0x1528, 0x1988, 0x1248, 0x1428, 0x1808, 0x1028, 0x1010}},
    {11976, 2, {0x2000, 0x1000, 0x1000, 0x0, 0x0, 0xF000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1400, 0x1800, 0x1000, 0x0, 0x0}},
    {11977, 2, {0x1FF0, 0x1010, 0x1010, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0x0}},
    {11978, 2, {0x0, 0x7F00, 0x4100, 0x4100, 0x4100, 0x7F00, 0x4900, 0x800, 0x800, 0x4F00, 0x4800, 0x4800, 0x4800, 0x4F00, 0xF000, 0x4000}},
    {11979, 2, {0x200, 0x200, 0x200, 0x7FFC, 0x400, 0x900, 0x1100, 0x2100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {11980, 2, {0x2000, 0x1000, 0x1000, 0x0, 0x0, 0xF000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2800, 0x47FE, 0x0, 0x0}},
    {11981, 2, {0x2000, 0x1000, 0x800, 0x4000, 0x2000, 0x1000, 0x0, 0xF000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2800, 0x4400, 0x83FE}},
    {11982, 2, {0x0, 0x2000, 0x1000, 0x0, 0x0, 0xF800, 0x800, 0x1000, 0x2000, 0x7800, 0x800, 0x800, 0x1000, 0x1000, 0x2800, 0xC7FE}},
    {11983, 2, {0x0, 0x3C, 0x24, 0x28, 0x28, 0x30, 0x28, 0x24, 0x24, 0x24, 0x34, 0x28, 0x20, 0x20, 0x20, 0x20}},
    {11984, 2, {0x1000, 0x1000, 0x1E00, 0x2000, 0x2000, 0x7C00, 0x9000, 0x1000, 0xFE00, 0x1000, 0x1000, 0x1200, 0x1400, 0x1800, 0x1000, 0x0}},
    {11985, 2, {0xFF8, 0x800, 0xFF0, 0x800, 0xFF0, 0x800, 0xFFFE, 0x1200, 0x1110, 0x10A0, 0x1040, 0x1220, 0x1418, 0x1806, 0x1000, 0x0}},
    {11986, 2, {0xFF8, 0x800, 0x800, 0xFF0, 0x800, 0x800, 0xFF0, 0x800, 0x800, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008, 0x0}},
    {11987, 2, {0x800, 0x810, 0x820, 0x840, 0x880, 0x900, 0x800, 0xFFFE, 0xA00, 0x900, 0x880, 0x840, 0x920, 0xA18, 0xC06, 0x800}},
    {11988, 2, {0x2000, 0x13FC, 0x1004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {11989, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1000, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1000, 0x1FF8, 0x1008, 0x1008, 0x1FF8}},
    {11990, 2, {0xF800, 0x8800, 0x9000, 0x9000, 0xA000, 0x9000, 0x9000, 0x8800, 0x8800, 0x8800, 0x8800, 0xF000, 0x8000, 0x8000, 0x8000, 0x8000}},
    {11991, 2, {0x7FFC, 0x100, 0xFFFE, 0x9112, 0x8D62, 0x100, 0x1D70, 0x610C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {11992, 2, {0x200, 0x200, 0x7FF8, 0x200, 0x3FF0, 0x200, 0xFFFC, 0x0, 0x1FE0, 0x1020, 0x1FE0, 0x1020, 0x1FE0, 0x1020, 0x1020, 0x10C0}},
    {11993, 2, {0x100, 0x100, 0x104, 0xFFFE, 0x100, 0x110, 0x3FF8, 0x100, 0x104, 0x7FFE, 0x104, 0x104, 0x104, 0x128, 0x110, 0x100}},
    {11994, 2, {0x7FFC, 0x200, 0x400, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1290, 0x240, 0x420, 0x1810, 0x6008, 0x0}},
    {11995, 2, {0x3FF0, 0x2010, 0x2010, 0x2850, 0x2450, 0x2290, 0x2290, 0x2110, 0x2110, 0x2290, 0x2292, 0x244A, 0x484A, 0x4006, 0x8002, 0x0}},
    {11996, 2, {0xFFC0, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x20, 0x20, 0x12, 0xA, 0x6, 0x2, 0x0}},
    {11997, 2, {0x100, 0x280, 0x440, 0xA20, 0x3118, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1008, 0x1190, 0x1060, 0x1210, 0x1408, 0x1804}},
    {11998, 2, {0x800, 0x800, 0x1400, 0x1200, 0x2900, 0x7E00, 0xA200, 0x3E00, 0x2200, 0x3E00, 0x2000, 0x3E00, 0x2000, 0x2000, 0x3E00, 0x2000}},
    {11999, 2, {0x800, 0x800, 0x1400, 0x1200, 0x2900, 0x7E00, 0xA200, 0x3E00, 0x2200, 0x3E00, 0x2000, 0x2800, 0x2400, 0x2A00, 0x3200, 0x2000}},
    {12000, 2, {0x2000, 0x2000, 0x2000, 0x3E00, 0x4400, 0x4800, 0x8000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1200, 0x1400, 0x1800, 0x1000, 0x0}},
    {12001, 2, {0x888, 0x888, 0x1110, 0x2220, 0x1110, 0x888, 0xFFFE, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {12002, 2, {0x7FE0, 0x20, 0x20, 0x1020, 0x1020, 0x1020, 0x1FFC, 0x4, 0x4, 0x4, 0xFFE4, 0x4, 0x4, 0x28, 0x10, 0x0}},
    {12003, 2, {0x1FF0, 0x1010, 0x1F90, 0x1090, 0x1090, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {12004, 2, {0x200, 0x400, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2208, 0x2208, 0x3FF8, 0x510, 0x920, 0x948, 0x117C, 0x2102, 0x4102, 0x80FE}},
    {12005, 2, {0x400, 0x400, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x0}},
    {12006, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1210, 0x1110, 0x1150, 0x1020, 0x1000, 0x1FFC, 0x4, 0x4, 0x7FE4, 0x4, 0x28, 0x10}},
    {12007, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2828, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2828, 0x3FF8, 0x2008}},
    {12008, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE00E}},
    {12009, 2, {0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x440, 0x820, 0x1010}},
    {12010, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x210A, 0x102, 0x102, 0xFE}},
    {12011, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xF01E, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x2010, 0x4010}},
    {12012, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC826, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020}},
    {12013, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0, 0x2928, 0x2548, 0x2388, 0x3FF8, 0x2388, 0x2548, 0x2928, 0x2108, 0x3FF8}},
    {12014, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0, 0x2108, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x8}},
    {12015, 2, {0x200, 0x100, 0x7FFC, 0x1010, 0x820, 0xFFFE, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1112, 0x102, 0xFE}},
    {12016, 2, {0x420, 0x410, 0x410, 0x400, 0xFFFE, 0x480, 0x488, 0x488, 0x490, 0x8A0, 0x8C0, 0x1082, 0x1182, 0x2282, 0x447E, 0x8000}},
    {12017, 2, {0x400, 0x3FF8, 0x2008, 0x3EF8, 0x280, 0x7A80, 0xAFC, 0xFE94, 0xAD4, 0x7AA4, 0x2A4, 0x7AD4, 0xA94, 0xFEFC, 0xA82, 0x79FE}},
    {12018, 2, {0x400, 0xFC0, 0x1080, 0x3FF0, 0xD110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x102, 0xFE}},
    {12019, 2, {0x400, 0x400, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x102, 0x102, 0xFE}},
    {12032, 2, {0x0, 0x0, 0x0, 0x0, 0x4, 0x7FFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12033, 2, {0x100, 0x180, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12034, 2, {0x0, 0x0, 0x400, 0x200, 0x100, 0x80, 0x40, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12035, 2, {0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x400, 0x400, 0x800, 0x1000, 0x2000, 0x0}},
    {12036, 2, {0x40, 0x3FE0, 0x40, 0x80, 0x100, 0x200, 0x400, 0x800, 0x1000, 0x2004, 0x2004, 0x2006, 0x1FFC, 0x0, 0x0, 0x0}},
    {12037, 2, {0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x3C0, 0x40, 0x0, 0x0}},
    {12038, 2, {0x10, 0x1FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x7FFE, 0x0, 0x0, 0x0, 0x0}},
    {12039, 2, {0x0, 0x0, 0x100, 0x80, 0x84, 0x7FFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12040, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x200E, 0x4004, 0x0}},
    {12041, 2, {0x840, 0xC60, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x1040, 0x1044, 0x2044, 0x4046, 0x803C, 0x0}},
    {12042, 2, {0x1E00, 0x100, 0x100, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x200C, 0x4006, 0x8002, 0x0, 0x0}},
    {12043, 2, {0xC0, 0x40, 0x40, 0x640, 0x440, 0x440, 0x420, 0x820, 0x820, 0x810, 0x1010, 0x1008, 0x2004, 0x4006, 0x8002, 0x0}},
    {12044, 2, {0x2004, 0x3FFE, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2024, 0x201C, 0x2008, 0x0}},
    {12045, 2, {0x0, 0x0, 0x0, 0x2002, 0x3FFE, 0x2004, 0x4008, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12046, 2, {0x400, 0x300, 0x100, 0x0, 0x40, 0x80, 0x80, 0x100, 0x100, 0x600, 0x200, 0x200, 0x200, 0x200, 0x0, 0x0}},
    {12047, 2, {0x20, 0xFF0, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x822, 0x1022, 0x2022, 0x401E, 0x0}},
    {12048, 2, {0x2004, 0x3006, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x3FFC, 0x2004, 0x0, 0x0}},
    {12049, 2, {0x4, 0x3FFE, 0x104, 0x104, 0x104, 0x104, 0x104, 0x104, 0x204, 0x204, 0x404, 0x404, 0x844, 0x1038, 0x6010, 0x0}},
    {12050, 2, {0x100, 0x100, 0x100, 0x108, 0x3FFC, 0x108, 0x108, 0x108, 0x108, 0x208, 0x208, 0x408, 0x888, 0x1070, 0x6020, 0x0}},
    {12051, 2, {0x1000, 0x1000, 0x1004, 0x1FFE, 0x1004, 0x2004, 0x4004, 0x4, 0x4, 0x4, 0x4, 0x4, 0x44, 0x38, 0x10, 0x0}},
    {12052, 2, {0x800, 0xC00, 0x800, 0x810, 0x820, 0x840, 0x980, 0xA00, 0xC00, 0x3800, 0xC800, 0x804, 0x804, 0x806, 0x7FC, 0x0}},
    {12053, 2, {0x2008, 0x3FFC, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2008, 0x3FFC, 0x2000, 0x0, 0x0}},
    {12054, 2, {0x7FFC, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2008, 0x3FFC, 0x0, 0x0, 0x0}},
    {12055, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12056, 2, {0x200, 0x200, 0x200, 0x200, 0x200, 0x380, 0x240, 0x230, 0x218, 0x208, 0x200, 0x200, 0x200, 0x200, 0x200, 0x0}},
    {12057, 2, {0x420, 0x7F0, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x4E0, 0x440, 0x400, 0x400, 0x400, 0x400, 0x0}},
    {12058, 2, {0x2004, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000, 0x0}},
    {12059, 2, {0x180, 0x100, 0x100, 0x200, 0x200, 0x400, 0x400, 0x820, 0x810, 0x1008, 0x11FC, 0xFE04, 0x4004, 0x0, 0x0, 0x0}},
    {12060, 2, {0x10, 0x3FF8, 0x810, 0x810, 0x420, 0x420, 0x240, 0x240, 0x180, 0x180, 0x240, 0x420, 0x181E, 0x6004, 0x0, 0x0}},
    {12061, 2, {0x1008, 0x1FFC, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x0, 0x0, 0x0}},
    {12062, 2, {0x2004, 0x3FFE, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x3FFC, 0x2004, 0x0, 0x0}},
    {12063, 2, {0x100, 0x180, 0x100, 0x100, 0x100, 0x108, 0x3FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x104, 0x7FFE, 0x0}},
    {12064, 2, {0x100, 0x180, 0x100, 0x100, 0x100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x108, 0x3FFC, 0x0}},
    {12065, 2, {0x400, 0x410, 0x7F8, 0xC10, 0x1220, 0x2140, 0x180, 0xE70, 0x700E, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12066, 2, {0x200, 0x200, 0x1210, 0xBF8, 0x410, 0x410, 0xA20, 0xA20, 0x1140, 0x2140, 0x4080, 0x140, 0x620, 0x1818, 0xE006, 0x0}},
    {12067, 2, {0x200, 0x200, 0x208, 0x3FC, 0x408, 0x408, 0xC10, 0x1210, 0x2120, 0x4120, 0x40, 0x80, 0x100, 0x600, 0x3800, 0x0}},
    {12068, 2, {0x100, 0x180, 0x100, 0x100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x200E, 0x4004, 0x0}},
    {12069, 2, {0x200, 0x300, 0x200, 0x200, 0x204, 0x7FFE, 0x420, 0x420, 0x820, 0x820, 0x1040, 0xE80, 0x180, 0x670, 0x780E, 0x0}},
    {12070, 2, {0x8, 0x3FFC, 0x10, 0x20, 0xC0, 0x80, 0x84, 0x7FFE, 0x80, 0x80, 0x80, 0x80, 0x480, 0x380, 0x100, 0x0}},
    {12071, 2, {0x0, 0x100, 0x80, 0x2084, 0x3FFE, 0x2004, 0x4008, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12072, 2, {0x20, 0x20, 0x20, 0x24, 0x7FFE, 0x20, 0x20, 0x820, 0x420, 0x420, 0x420, 0x20, 0x120, 0xE0, 0x40, 0x0}},
    {12073, 2, {0x80, 0x80, 0x80, 0x80, 0x80, 0xC90, 0x888, 0x1084, 0x1086, 0x2082, 0x4082, 0x80, 0x480, 0x380, 0x100, 0x0}},
    {12074, 2, {0x200, 0x200, 0x200, 0x208, 0xFFFC, 0x200, 0x200, 0x280, 0x280, 0x480, 0x484, 0x884, 0x1084, 0x2086, 0xC07C, 0x0}},
    {12075, 2, {0x1008, 0x1FFC, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000, 0x0}},
    {12076, 2, {0x100, 0x100, 0x100, 0x2104, 0x2104, 0x2104, 0x2104, 0x2104, 0x3FFC, 0x2104, 0x100, 0x100, 0x200, 0xC00, 0x7000, 0x0}},
    {12077, 2, {0x100, 0x100, 0x100, 0x100, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x7FFC, 0x4004, 0x0}},
    {12078, 2, {0x444, 0x444, 0x888, 0x888, 0x1110, 0x1110, 0x2220, 0x4440, 0x2220, 0x1110, 0x888, 0x888, 0x444, 0x444, 0x444, 0x0}},
    {12079, 2, {0x8, 0x3FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x104, 0x7FFE, 0x0, 0x0, 0x0}},
    {12080, 2, {0x10, 0x7FF8, 0x10, 0x10, 0x10, 0x10, 0x3FF0, 0x2010, 0x2000, 0x2000, 0x2000, 0x2004, 0x2004, 0x2006, 0x1FFC, 0x0}},
    {12081, 2, {0x80, 0x80, 0x80, 0x1084, 0x1FFE, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x10BC, 0x1088, 0x80, 0x80, 0x0}},
    {12082, 2, {0x8, 0x3FFC, 0x100, 0x100, 0x100, 0x100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12083, 2, {0x300, 0x200, 0x400, 0x430, 0x820, 0x1040, 0x7F80, 0x2100, 0x200, 0x420, 0x810, 0x13E8, 0x7C0C, 0x2004, 0x0, 0x0}},
    {12084, 2, {0x100, 0x80, 0x84, 0x1FFE, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000, 0x0}},
    {12085, 2, {0x800, 0x7C00, 0x800, 0x800, 0x1000, 0x2400, 0x3E00, 0x400, 0x8400, 0x4400, 0x2800, 0x1000, 0x2800, 0xC406, 0x83FC, 0x0}},
    {12086, 2, {0x410, 0x410, 0x410, 0x410, 0x410, 0x410, 0x7FFE, 0x410, 0x410, 0x410, 0x410, 0x810, 0x810, 0x1010, 0x2010, 0x0}},
    {12087, 2, {0x120, 0x118, 0x108, 0x100, 0x11E, 0x1E0, 0x7F00, 0x3080, 0x80, 0x80, 0x40, 0x42, 0x22, 0x12, 0xE, 0x0}},
    {12088, 2, {0x10, 0x3FF8, 0x10, 0x10, 0x10, 0x1FF0, 0x1000, 0x1008, 0x3FFC, 0x8, 0x8, 0x8, 0x108, 0xF0, 0x20, 0x0}},
    {12089, 2, {0x8, 0x3FFC, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3FF8, 0x8, 0x0}},
    {12090, 2, {0xC0, 0x80, 0x100, 0x200, 0x400, 0xC0, 0x80, 0x100, 0x200, 0x460, 0x40, 0x80, 0x100, 0x200, 0x400, 0x0}},
    {12091, 2, {0xC0, 0x80, 0x100, 0x200, 0x4C0, 0x880, 0x100, 0x300, 0x500, 0x900, 0x1100, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12092, 2, {0x100, 0x80, 0xC0, 0x40, 0x400, 0x400, 0x2408, 0x2404, 0x2406, 0x2402, 0x4400, 0x8410, 0x410, 0x418, 0x3F0, 0x0}},
    {12093, 2, {0x200, 0x240, 0x220, 0x200, 0x21C, 0x2E0, 0x7F00, 0x218, 0x210, 0x220, 0x140, 0x182, 0x244, 0xC24, 0x701E, 0x0}},
    {12094, 2, {0x10, 0x1038, 0x1FC0, 0x1008, 0x1FFC, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000, 0x0}},
    {12095, 2, {0x10, 0x78, 0x7F80, 0x100, 0x110, 0x3FF8, 0x100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x900, 0x700, 0x200, 0x0}},
    {12096, 2, {0x80, 0x80, 0x84, 0x7FFE, 0x80, 0x90, 0x1FF8, 0x410, 0x410, 0x220, 0x140, 0x80, 0x140, 0x630, 0x380E, 0x0}},
    {12097, 2, {0x100, 0x100, 0x104, 0x1FE, 0x100, 0x110, 0x3FF8, 0x810, 0x420, 0x420, 0x240, 0x180, 0x180, 0xE70, 0x700E, 0x0}},
    {12098, 2, {0x100, 0x80, 0x84, 0x7FFE, 0x820, 0x820, 0x820, 0x440, 0x440, 0x240, 0x280, 0x180, 0x660, 0x1818, 0x6006, 0x0}},
    {12099, 2, {0x20, 0x820, 0x420, 0x220, 0x220, 0x820, 0x420, 0x220, 0x224, 0x3E, 0x3E0, 0x7C20, 0x2020, 0x20, 0x20, 0x0}},
    {12100, 2, {0x20, 0x1070, 0x1F80, 0x1000, 0x1000, 0x1004, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040, 0x2040, 0x2040, 0x4040, 0x8040, 0x0}},
    {12101, 2, {0x100, 0x80, 0x84, 0x7FFE, 0x200, 0x200, 0x208, 0x3FC, 0x208, 0x208, 0x208, 0x408, 0x888, 0x1070, 0x6020, 0x0}},
    {12102, 2, {0x8, 0x1FFC, 0x100, 0x100, 0x100, 0x104, 0x7FFE, 0x240, 0x240, 0x240, 0x240, 0x442, 0x842, 0x1042, 0x603E, 0x0}},
    {12103, 2, {0x1010, 0x1FF8, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0}},
    {12104, 2, {0x2008, 0x3FFC, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x0, 0x0, 0x0}},
    {12105, 2, {0x810, 0xFF8, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0x1090, 0x2070, 0x4020, 0x0}},
    {12106, 2, {0x80, 0x80, 0x80, 0x84, 0x7FFE, 0x80, 0x280, 0x2A0, 0x4A0, 0x490, 0x890, 0x888, 0x1086, 0x6084, 0x80, 0x0}},
    {12107, 2, {0x800, 0x800, 0x804, 0xFFE, 0x1008, 0x2110, 0x4100, 0x100, 0x100, 0x100, 0x180, 0x240, 0x420, 0x1818, 0x6006, 0x0}},
    {12108, 2, {0x80, 0x80, 0x80, 0x80, 0x80, 0x888, 0x8FC, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x884, 0x7FFE, 0x0}},
    {12109, 2, {0x4, 0x7FFE, 0x200, 0x200, 0x210, 0x3F8, 0x410, 0xC10, 0x1220, 0x2120, 0x4140, 0x80, 0x100, 0x600, 0x7800, 0x0}},
    {12110, 2, {0x420, 0x7F0, 0x420, 0x420, 0x420, 0x820, 0x101C, 0x2000, 0x1FF8, 0x810, 0x420, 0x240, 0x180, 0xE70, 0x700C, 0x0}},
    {12111, 2, {0x8, 0x1FFC, 0x1088, 0x1088, 0x1088, 0x1088, 0x7FFE, 0x1108, 0x1108, 0x2108, 0x2208, 0x7FFE, 0x410, 0x9F0, 0x3020, 0x0}},
    {12112, 2, {0x40, 0x1040, 0x1040, 0x1040, 0x104C, 0x1248, 0x1F50, 0x1060, 0x1040, 0x1040, 0x1040, 0x1044, 0x1344, 0x3C46, 0x103C, 0x0}},
    {12113, 2, {0x10, 0x38, 0x3FC0, 0x100, 0x100, 0x11C, 0x3FE0, 0x100, 0x10E, 0x1F0, 0x7F00, 0x102, 0x102, 0x102, 0xFE, 0x0}},
    {12114, 2, {0x10, 0x1038, 0x1FC0, 0x1080, 0x1080, 0x1088, 0x1FFC, 0x1080, 0x1080, 0x1080, 0x1080, 0x1242, 0x1422, 0x7812, 0x200E, 0x0}},
    {12115, 2, {0x800, 0x808, 0xFFC, 0x1000, 0x2FF0, 0x4000, 0xBFF0, 0x10, 0x10, 0x10, 0x10, 0x12, 0xA, 0x6, 0x2, 0x0}},
    {12116, 2, {0x100, 0x100, 0x100, 0x118, 0x110, 0x7D20, 0x540, 0x580, 0x940, 0x920, 0x1110, 0x210C, 0x4906, 0x8702, 0x200, 0x0}},
    {12117, 2, {0x80, 0x80, 0x80, 0x80, 0x88C, 0x888, 0x890, 0x10A0, 0x20C0, 0x80, 0x180, 0x240, 0x420, 0x1818, 0x6006, 0x0}},
    {12118, 2, {0x10, 0x1038, 0x1FE0, 0x1120, 0x1120, 0x1120, 0x1120, 0x1120, 0x1110, 0x1110, 0x1110, 0x2108, 0x2108, 0x4106, 0x8102, 0x0}},
    {12119, 2, {0x440, 0x420, 0x810, 0x100C, 0x2004, 0x4020, 0x420, 0x440, 0x240, 0x280, 0x180, 0x180, 0x240, 0xC30, 0x300E, 0x0}},
    {12120, 2, {0x18, 0x1810, 0x620, 0x1C0, 0x230, 0xC0C, 0x3024, 0x420, 0x440, 0x240, 0x280, 0x180, 0x2C0, 0xC30, 0x300E, 0x0}},
    {12121, 2, {0x10, 0x410, 0x410, 0x410, 0x410, 0x7F0, 0x10, 0x10, 0x3FF0, 0x410, 0x410, 0x410, 0x410, 0x810, 0x3010, 0x0}},
    {12122, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x1044, 0x1FFE, 0x1000, 0x1020, 0x1FF0, 0x1020, 0x1020, 0x1020, 0x1020, 0x2020, 0x4020, 0x0}},
    {12123, 2, {0x4, 0x7FFE, 0x40, 0x840, 0x840, 0x1044, 0x3FFE, 0x140, 0x240, 0x440, 0x840, 0x1040, 0x2240, 0x41C0, 0x80, 0x0}},
    {12124, 2, {0x80, 0x80, 0x1080, 0x1088, 0x1FFC, 0x1080, 0x2080, 0x4080, 0x84, 0x7FFE, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {12125, 2, {0x100, 0x1A0, 0x110, 0x110, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x200E, 0x4004, 0x0}},
    {12126, 2, {0x100, 0x80, 0x84, 0x7FFE, 0x100, 0x100, 0x230, 0x5E0, 0x1E40, 0x880, 0x100, 0x210, 0x5F8, 0x1E0C, 0x804, 0x0}},
    {12127, 2, {0x8, 0x3FFC, 0x100, 0x100, 0x100, 0x110, 0x1FF8, 0x100, 0x120, 0x110, 0x110, 0x100, 0x104, 0x7FFE, 0x0, 0x0}},
    {12128, 2, {0x8, 0x103C, 0x1FE0, 0x1120, 0x1120, 0x1120, 0x1120, 0x1120, 0x1110, 0x1110, 0x1150, 0x1148, 0x2148, 0x43A6, 0x8122, 0x0}},
    {12129, 2, {0x4, 0x7FFE, 0x400, 0x400, 0x420, 0x7F0, 0x820, 0xA20, 0x920, 0x920, 0x1020, 0x1022, 0x1322, 0x7C22, 0x201E, 0x0}},
    {12130, 2, {0x810, 0x810, 0x814, 0x7FFE, 0x810, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x0}},
    {12131, 2, {0x80, 0x1080, 0x1080, 0x1088, 0x1FFC, 0x1080, 0x2080, 0x4080, 0x90, 0x1FF8, 0x80, 0x80, 0x80, 0x84, 0x7FFE, 0x0}},
    {12132, 2, {0x1004, 0x1FFE, 0x1084, 0x1084, 0x1084, 0x1FFC, 0x1084, 0x1084, 0x1084, 0x1FFC, 0x1084, 0x2084, 0x20A4, 0x409C, 0x8088, 0x0}},
    {12133, 2, {0x2004, 0x3FFE, 0x2084, 0x2084, 0x2084, 0x2084, 0x3FFC, 0x2084, 0x2084, 0x2084, 0x2084, 0x2084, 0x3FFC, 0x2004, 0x0, 0x0}},
    {12134, 2, {0x4, 0x7FFE, 0x84, 0x88, 0x80, 0x1080, 0x1088, 0x10FC, 0x1080, 0x1080, 0x1880, 0x1480, 0x2280, 0xC180, 0x80FE, 0x0}},
    {12135, 2, {0x80, 0x844, 0xFFE, 0x4800, 0x2800, 0x2800, 0x800, 0x1800, 0x6800, 0x4800, 0x800, 0x1000, 0x1000, 0x2000, 0x4000, 0x0}},
    {12136, 2, {0x8, 0x7F10, 0x2A2, 0x2244, 0x1428, 0x810, 0x1008, 0x2004, 0x4002, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12137, 2, {0x100, 0x100, 0x1208, 0x1FFC, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x0}},
    {12138, 2, {0x80, 0x80, 0x1084, 0x1FFE, 0x1084, 0x1088, 0x1080, 0x1FF8, 0x1210, 0x1210, 0x1120, 0x20C0, 0x20C0, 0x4330, 0x9C0E, 0x0}},
    {12139, 2, {0x1008, 0x1FFC, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x7FFE, 0x0, 0x0, 0x0}},
    {12140, 2, {0x1010, 0x1FF8, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0}},
    {12141, 2, {0x10, 0x1FF8, 0x20, 0x340, 0x80, 0x84, 0x7FFE, 0x188, 0x290, 0x480, 0x880, 0x1080, 0x2480, 0x4380, 0x100, 0x0}},
    {12142, 2, {0x800, 0x800, 0x808, 0xFFC, 0x1080, 0x2080, 0x80, 0x84, 0x7FFE, 0x80, 0x140, 0x220, 0x410, 0x180E, 0x6004, 0x0}},
    {12143, 2, {0x4, 0x7FFE, 0x100, 0x100, 0x200, 0x408, 0xFFC, 0x1408, 0x2408, 0x4408, 0x408, 0x408, 0x7F8, 0x408, 0x0, 0x0}},
    {12144, 2, {0x10, 0x1FF8, 0x0, 0x0, 0x0, 0x4, 0x7FFE, 0x80, 0xCA0, 0x890, 0x1088, 0x2084, 0x4484, 0x380, 0x100, 0x0}},
    {12145, 2, {0x100, 0x100, 0x2104, 0x3FFE, 0x2104, 0x2204, 0x2244, 0x2224, 0x24F4, 0x3F14, 0x2814, 0x2004, 0x2004, 0x203C, 0x2008, 0x0}},
    {12146, 2, {0x8, 0x3C, 0x1FC0, 0x80, 0x80, 0x84, 0x7FFE, 0x80, 0x1C0, 0x2A0, 0x490, 0x888, 0x1086, 0x6082, 0x80, 0x0}},
    {12147, 2, {0x100, 0x80, 0x2082, 0x3FFE, 0x2004, 0x4008, 0x440, 0x440, 0x440, 0x420, 0x820, 0x810, 0x1008, 0x200C, 0x4006, 0x0}},
    {12148, 2, {0x200, 0x100, 0x80, 0x88, 0x3FFC, 0x0, 0x810, 0x810, 0x420, 0x420, 0x240, 0x240, 0x84, 0x7FFE, 0x0, 0x0}},
    {12149, 2, {0x1000, 0x1060, 0x1040, 0x1244, 0x1F7E, 0x2890, 0x4910, 0x8A10, 0x810, 0x810, 0x810, 0x810, 0x890, 0x870, 0x820, 0x0}},
    {12150, 2, {0x80, 0x108C, 0x888, 0xC90, 0x4A0, 0xC4, 0x7FFE, 0x280, 0x2C0, 0x4A0, 0x890, 0x1088, 0x2086, 0x4082, 0x80, 0x0}},
    {12151, 2, {0x180, 0x100, 0x230, 0x420, 0x1FC0, 0x880, 0x120, 0x210, 0x3FFC, 0x1084, 0x80, 0xC90, 0x888, 0x1086, 0x2082, 0x0}},
    {12152, 2, {0x800, 0x800, 0x808, 0xFFC, 0x1080, 0x2080, 0x84, 0x7FFE, 0x80, 0x1084, 0x1084, 0x1084, 0x1084, 0x1FFC, 0x1004, 0x0}},
    {12153, 2, {0x2004, 0x3FFE, 0x2004, 0x2114, 0x2114, 0x2A94, 0x2464, 0x2224, 0x2524, 0x2954, 0x3094, 0x2104, 0x2024, 0x201C, 0x2008, 0x0}},
    {12154, 2, {0x830, 0x420, 0x244, 0x7FFE, 0x100, 0x100, 0x110, 0x1FF8, 0x100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12155, 2, {0x204, 0x7FFE, 0x204, 0x2244, 0x1224, 0x1224, 0x204, 0x60C, 0x1A34, 0x62C4, 0x204, 0x204, 0x1224, 0xE1C, 0x408, 0x0}},
    {12156, 2, {0x100, 0x100, 0x10C, 0x3FE8, 0x110, 0x124, 0x7FFE, 0x80, 0x100, 0x630, 0xCC0, 0x1704, 0x6404, 0x406, 0x3FC, 0x0}},
    {12157, 2, {0x4, 0x7FFE, 0x80, 0x80, 0x2104, 0x3FFE, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x225C, 0x2008, 0x0}},
    {12158, 2, {0x100, 0x104, 0xFFFE, 0x100, 0x110, 0x3FF8, 0x100, 0x104, 0xFFFE, 0x540, 0x520, 0x910, 0x110C, 0x2106, 0xC102, 0x0}},
    {12159, 2, {0x4, 0x7FFE, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0x820, 0xFE0, 0x824, 0x83E, 0xFE0, 0x7820, 0x20, 0x20, 0x0}},
    {12160, 2, {0x100, 0x108, 0x1FFC, 0x108, 0x7FFE, 0x108, 0x1FF8, 0x108, 0x100, 0x3FF8, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x0}},
    {12161, 2, {0x100, 0x100, 0x2104, 0x3FFE, 0x2104, 0x2184, 0x2244, 0x2434, 0x2914, 0x2104, 0x2184, 0x2244, 0x2424, 0x281C, 0x2008, 0x0}},
    {12162, 2, {0x2008, 0x3FFC, 0x2100, 0x2100, 0x2108, 0x3FFC, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2108, 0x2100, 0x2104, 0x3FFE, 0x2000, 0x0}},
    {12163, 2, {0x180, 0x100, 0x208, 0x1FFC, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x0}},
    {12164, 2, {0x4, 0x7FFE, 0x100, 0x200, 0x420, 0x810, 0x7FF8, 0x2104, 0x100, 0x108, 0x3FFC, 0x100, 0x100, 0x104, 0x7FFE, 0x0}},
    {12165, 2, {0x200, 0x2704, 0x387E, 0x2004, 0x2004, 0x2004, 0x2004, 0x3E7C, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x3FFC, 0x2004, 0x0}},
    {12166, 2, {0x10, 0x78, 0x3F80, 0x80, 0x84, 0x7FFE, 0x80, 0x80, 0x888, 0xFFC, 0x808, 0x808, 0x808, 0xFF8, 0x808, 0x0}},
    {12167, 2, {0x1010, 0x1010, 0x1010, 0x1214, 0x1F7E, 0x1210, 0x2250, 0x5250, 0xA90, 0xAFE, 0x410, 0x810, 0x1010, 0x2010, 0x4010, 0x0}},
    {12168, 2, {0x100, 0x100, 0xA10, 0xFF8, 0x910, 0x890, 0x890, 0x814, 0x7FFE, 0x910, 0x890, 0x890, 0x810, 0x10F0, 0x2020, 0x0}},
    {12169, 2, {0x1010, 0x1FF8, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1200, 0x110C, 0x1130, 0x10C0, 0x1660, 0x7818, 0x2006, 0x0}},
    {12170, 2, {0x400, 0x420, 0x7F0, 0x840, 0x1088, 0x3FFC, 0x5108, 0x1108, 0x1108, 0x1FF8, 0x1008, 0x1002, 0x1002, 0x1002, 0xFFE, 0x0}},
    {12171, 2, {0x810, 0x810, 0x810, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x3E7C, 0x2A54, 0x810, 0x1010, 0x2010, 0x4010, 0x0}},
    {12172, 2, {0x100, 0x1FC, 0x100, 0x100, 0x3FFE, 0x2002, 0x2104, 0x2110, 0x2138, 0x21C0, 0x2F00, 0x2102, 0x2102, 0x4102, 0x80FE, 0x0}},
    {12173, 2, {0x80, 0x80, 0x80, 0x1088, 0x1FFC, 0x1088, 0x1088, 0x1088, 0x1FF8, 0x1088, 0x80, 0x88, 0x84, 0x7FFE, 0x3002, 0x0}},
    {12174, 2, {0xC0, 0x80, 0x100, 0x1208, 0x1FFC, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x7FFE, 0x0, 0x0}},
    {12175, 2, {0x1808, 0x11FC, 0x2000, 0x4000, 0xC04, 0xBFE, 0x1010, 0x3010, 0x5010, 0x9010, 0x1010, 0x1010, 0x1090, 0x1070, 0x1020, 0x0}},
    {12176, 2, {0x100, 0x80, 0x84, 0x7FFE, 0x180, 0x28C, 0x488, 0xC50, 0x1460, 0x2440, 0x4420, 0x490, 0x508, 0xE06, 0x402, 0x0}},
    {12177, 2, {0x4, 0x7FFE, 0x220, 0x220, 0x2224, 0x3FFE, 0x2224, 0x2224, 0x2224, 0x2224, 0x23E4, 0x2224, 0x2004, 0x2004, 0x2004, 0x0}},
    {12178, 2, {0x20, 0x1FF0, 0x1020, 0x1020, 0x1FE0, 0x1020, 0x1020, 0x1FE0, 0x1020, 0x1020, 0x1FE0, 0x14A4, 0x884, 0x3086, 0xC07C, 0x0}},
    {12179, 2, {0x420, 0x7F0, 0x420, 0x844, 0x1FFE, 0x2884, 0x4884, 0xFFC, 0x884, 0x884, 0xFFC, 0x1084, 0x1084, 0x209C, 0x4088, 0x0}},
    {12180, 2, {0x200, 0x104, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x0, 0x1010, 0x1FF8, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0}},
    {12181, 2, {0x440, 0x420, 0x810, 0x100C, 0x2184, 0x240, 0x420, 0x1818, 0x6006, 0xFF0, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x0}},
    {12182, 2, {0x8, 0x3FFC, 0x0, 0x810, 0xFF8, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x460, 0x440, 0x240, 0x284, 0x7FFE, 0x0}},
    {12183, 2, {0x4, 0x7FFE, 0x100, 0x200, 0x40C, 0x1A08, 0x6310, 0x4A0, 0x19C0, 0x62A0, 0x490, 0x188E, 0x6084, 0x480, 0x300, 0x0}},
    {12184, 2, {0x60, 0xC0, 0x188, 0x65C, 0x1A20, 0x1C0, 0x320, 0xC60, 0x1090, 0x330, 0x1C50, 0x90, 0x310, 0x1D20, 0xC0, 0x0}},
    {12185, 2, {0x808, 0xFFC, 0x808, 0x808, 0xFF8, 0x808, 0x808, 0xFF8, 0x808, 0x808, 0xFF8, 0x808, 0x360, 0xC18, 0x3006, 0x0}},
    {12186, 2, {0x80, 0x80, 0x88, 0x1FFC, 0x80, 0x84, 0x7FFE, 0x240, 0x1248, 0x1244, 0x2244, 0x2442, 0x4442, 0x9C0, 0x3080, 0x0}},
    {12187, 2, {0x100, 0x100, 0x110, 0x1FF8, 0x100, 0x104, 0x7FFE, 0x100, 0x900, 0x908, 0x9FC, 0x900, 0x1500, 0x2300, 0x40FE, 0x0}},
    {12188, 2, {0x810, 0xFF8, 0x810, 0x810, 0x810, 0xFF0, 0x890, 0x80, 0x888, 0x8FC, 0x880, 0x1480, 0x1280, 0x2186, 0x40FC, 0x0}},
    {12189, 2, {0x80, 0x910, 0xFF8, 0x810, 0xFF0, 0x810, 0x816, 0xFF4, 0x818, 0x7FF0, 0x50, 0x190, 0x610, 0x18F0, 0x6020, 0x0}},
    {12190, 2, {0x100, 0x104, 0x7FFE, 0x100, 0x1108, 0x1FFC, 0x1108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x104, 0x7FFE, 0x100, 0x100, 0x0}},
    {12191, 2, {0x100, 0x88, 0x3FFC, 0x820, 0x420, 0x240, 0x244, 0x7FFE, 0x100, 0x108, 0x3FFC, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12192, 2, {0x1004, 0x1FFE, 0x1000, 0x1010, 0x17F8, 0x1000, 0x1004, 0x1FFE, 0x1240, 0x1248, 0x1250, 0x2220, 0x2290, 0x4F08, 0x8406, 0x0}},
    {12193, 2, {0x60, 0x180, 0x1E38, 0xC0, 0x1F18, 0x60, 0x1F80, 0x80, 0x1080, 0x1088, 0x18FC, 0x1480, 0x2280, 0x4180, 0x807E, 0x0}},
    {12194, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x2004, 0x2004, 0x2006, 0x1FFC, 0x0}},
    {12195, 2, {0x4, 0x7FFE, 0x240, 0x1244, 0x1FFE, 0x1244, 0x1244, 0x1244, 0x143C, 0x1804, 0x1004, 0x1FFC, 0x1004, 0x1FFC, 0x1004, 0x0}},
    {12196, 2, {0x30, 0xF8, 0x3F00, 0x110, 0x1110, 0x920, 0x944, 0x7FFE, 0x380, 0x540, 0x920, 0x1110, 0x210E, 0x4104, 0x100, 0x0}},
    {12197, 2, {0x1008, 0x1FFC, 0x1108, 0x1108, 0x1FF8, 0x1108, 0x1108, 0x1FF8, 0x1108, 0x100, 0x108, 0x3FFC, 0x100, 0x104, 0x7FFE, 0x0}},
    {12198, 2, {0x80, 0x180, 0x240, 0x420, 0x810, 0x1008, 0x6FF6, 0x100, 0x108, 0x3FFC, 0x100, 0x1118, 0x910, 0x924, 0x7FFE, 0x0}},
    {12199, 2, {0x810, 0xFF8, 0x800, 0xFF0, 0x800, 0xFF0, 0x800, 0x804, 0x7FFE, 0x888, 0x850, 0x820, 0x990, 0xE0E, 0x3804, 0x0}},
    {12200, 2, {0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4024, 0x401C, 0x4008, 0x0}},
    {12201, 2, {0x100, 0x1210, 0x1FF8, 0x1010, 0x1FF0, 0x1000, 0x1FF8, 0x1008, 0x1FF8, 0x1088, 0x84, 0x7FFE, 0x80, 0x80, 0x80, 0x0}},
    {12202, 2, {0x100, 0x110, 0x3FF8, 0x110, 0x7FFE, 0x110, 0x3FF0, 0x110, 0x3188, 0x990, 0x760, 0x920, 0x7118, 0x2706, 0x200, 0x0}},
    {12203, 2, {0x480, 0x440, 0x444, 0xFFE, 0x840, 0x1840, 0x2848, 0x4FFC, 0x840, 0x848, 0xFFC, 0x840, 0x844, 0xFFE, 0x800, 0x0}},
    {12204, 2, {0x4, 0x7FFE, 0x100, 0x2104, 0x3FFE, 0x2104, 0x2944, 0x2524, 0x2524, 0x2104, 0x2944, 0x2524, 0x2524, 0x211C, 0x2108, 0x0}},
    {12205, 2, {0x104, 0xFFFE, 0x100, 0x3FF8, 0x104, 0xFFFE, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x2078, 0x2010, 0x0}},
    {12206, 2, {0x240, 0x240, 0x244, 0x7E7E, 0x240, 0x240, 0x248, 0x3E7C, 0x240, 0x240, 0x244, 0x7E7E, 0x240, 0x240, 0x240, 0x0}},
    {12207, 2, {0x4, 0x7FFE, 0x80, 0x2104, 0x3FFE, 0x2244, 0x2244, 0x23C4, 0x2244, 0x2244, 0x23C4, 0x2244, 0x2244, 0x3FFC, 0x2004, 0x0}},
    {12208, 2, {0x820, 0x824, 0xFFFE, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x0}},
    {12209, 2, {0x200, 0x220, 0x1FF0, 0x424, 0x7FFE, 0x0, 0xFF0, 0x810, 0xFF0, 0x80, 0x3FFC, 0x1080, 0x3FFC, 0x80, 0x80, 0x0}},
    {12210, 2, {0x240, 0x240, 0x240, 0x244, 0x7E7E, 0x240, 0x240, 0x3E7C, 0x240, 0x244, 0x7E7E, 0x240, 0x240, 0x244, 0x7FFE, 0x0}},
    {12211, 2, {0x300, 0x108, 0x7FFC, 0x1020, 0x820, 0x440, 0xFFFE, 0x0, 0x1010, 0x1FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x0}},
    {12212, 2, {0x4, 0x7FFE, 0x80, 0x1108, 0x1FFC, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x360, 0xC18, 0x3006, 0x0}},
    {12213, 2, {0x2008, 0x3FFC, 0x2048, 0x21E8, 0x2F08, 0x2108, 0x2FE8, 0x2928, 0x2928, 0x2FE8, 0x2928, 0x214A, 0x21EA, 0x4F26, 0x8402, 0x0}},
    {12214, 2, {0x3FE0, 0x24, 0x28, 0xD30, 0x392C, 0x922, 0x912, 0x7FEE, 0x924, 0x928, 0x930, 0x112C, 0x1112, 0x210A, 0x4106, 0x0}},
    {12215, 2, {0x80, 0x180, 0x240, 0x520, 0x1898, 0x6FF6, 0x810, 0xFF0, 0x810, 0xFF0, 0x808, 0x990, 0xA60, 0x3C18, 0x1006, 0x0}},
    {12216, 2, {0x830, 0x420, 0x7FFE, 0x80, 0x1110, 0x1FF8, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0}},
    {12217, 2, {0x20, 0x70, 0x3F80, 0x104, 0x7FFE, 0x540, 0x920, 0x3118, 0xC816, 0xFF8, 0x810, 0xFF0, 0x810, 0xFF0, 0x810, 0x0}},
    {12218, 2, {0x1008, 0x1FFC, 0x1100, 0x1110, 0x1FF8, 0x1100, 0x1FF8, 0x1100, 0x1104, 0x1FFE, 0x24, 0x914, 0x2494, 0x2484, 0x4038, 0x0}},
    {12219, 2, {0x810, 0xFF8, 0x810, 0x9F0, 0x2912, 0x3FFE, 0x2004, 0x4FF4, 0x810, 0xFF0, 0x810, 0xFF0, 0x810, 0x870, 0x820, 0x0}},
    {12220, 2, {0x100, 0x84, 0x7FFE, 0x0, 0xFF0, 0x810, 0xFF0, 0x2004, 0x3FFE, 0x2004, 0x27E4, 0x2424, 0x27E4, 0x243C, 0x2008, 0x0}},
    {12221, 2, {0x2200, 0x3F0C, 0x2008, 0x2010, 0x3F26, 0x2044, 0x3F88, 0x2010, 0x7F26, 0x1044, 0x1008, 0x2210, 0xFF20, 0x4140, 0x180, 0x0}},
    {12222, 2, {0x8002, 0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8002, 0x8002, 0x8002, 0x8002, 0x8002, 0x8002, 0x8012, 0x800E, 0x8004, 0x0}},
    {12223, 2, {0x210, 0x2924, 0x2444, 0x32A4, 0x2994, 0x2244, 0x24A4, 0x3FFC, 0x2004, 0x1030, 0x10C0, 0x1700, 0x1804, 0x1004, 0xFFC, 0x0}},
    {12224, 2, {0x4, 0x7FFE, 0x0, 0xFF0, 0x810, 0xFF0, 0x0, 0x3FFC, 0x2004, 0x2444, 0x2284, 0x2FF4, 0x2104, 0x211C, 0x2108, 0x0}},
    {12225, 2, {0x100, 0x1208, 0x1FFC, 0x1108, 0x1108, 0x1FF8, 0x1108, 0x1108, 0x1FF8, 0x1088, 0x190, 0x2A8, 0x4FE, 0x1882, 0x607E, 0x0}},
    {12226, 2, {0x420, 0x7F0, 0x420, 0x840, 0x1FF8, 0x2888, 0x4888, 0xFF8, 0x888, 0x888, 0xFF8, 0x808, 0x2444, 0x2222, 0x4222, 0x0}},
    {12227, 2, {0x100, 0xA10, 0xFF8, 0x810, 0xFF0, 0x810, 0xFF0, 0x808, 0xFFC, 0x800, 0xFFC, 0x2924, 0x2494, 0x4494, 0x18, 0x0}},
    {12228, 2, {0x100, 0x104, 0x1FE, 0x100, 0x3FFC, 0x2214, 0x2924, 0x2444, 0x32A4, 0x2994, 0x2244, 0x2524, 0x2894, 0x3FFC, 0x2004, 0x0}},
    {12229, 2, {0x100, 0x2084, 0x3FFE, 0x2220, 0x2224, 0x3FFE, 0x2224, 0x3FFC, 0x2804, 0x2820, 0x2FA4, 0x2838, 0x29A2, 0x5E22, 0x883E, 0x0}},
    {12230, 2, {0x80, 0x84, 0x7FFE, 0x888, 0x888, 0x1494, 0x23E2, 0x490, 0x1908, 0x63F6, 0x420, 0xB40, 0x30C0, 0x120, 0x1E10, 0x0}},
    {12231, 2, {0x100, 0x4084, 0x7FFE, 0x4410, 0x4410, 0x7F7E, 0x4410, 0x5554, 0x5554, 0x5554, 0x5554, 0x5554, 0x6494, 0xA496, 0x8410, 0x0}},
    {12232, 2, {0x424, 0x7FFE, 0x420, 0x7E0, 0x4, 0x7FFE, 0x100, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x1FF8, 0x1008, 0xC30, 0x700E, 0x0}},
    {12233, 2, {0x70, 0x1F80, 0x108, 0x7FFE, 0x100, 0x7E0, 0x1918, 0x6286, 0xD60, 0x711E, 0xD64, 0x380, 0x560, 0x1918, 0x6306, 0x0}},
    {12234, 2, {0x1008, 0x1FFC, 0x1108, 0x1928, 0x1548, 0x1FF8, 0x1108, 0x100, 0x3FF8, 0x100, 0x100, 0x7FFE, 0x2444, 0x2222, 0x4222, 0x0}},
    {12235, 2, {0x1248, 0x1248, 0xA50, 0xA50, 0x7FFE, 0x1088, 0x2084, 0x5FFE, 0x1084, 0x1184, 0x12C4, 0x14A4, 0x1894, 0x109C, 0x1088, 0x0}},
    {12236, 2, {0x3FFC, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x240, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x242, 0x242, 0x1FE, 0x0}},
    {12237, 2, {0x7E0, 0x420, 0x27E4, 0x2424, 0x27E4, 0x2424, 0x27E4, 0x2424, 0x3E7C, 0x240, 0x7E7E, 0x1248, 0x1248, 0x2248, 0x4248, 0x0}},
    {12238, 2, {0x820, 0x820, 0xFF24, 0x8FE, 0x7E20, 0x24, 0x7FFE, 0x4244, 0x4244, 0x7E48, 0x4228, 0x2410, 0x1E30, 0x704E, 0x2184, 0x0}},
    {12239, 2, {0x608, 0x38FC, 0x2008, 0x3EF8, 0x2008, 0x3FF8, 0x2008, 0x940, 0x2528, 0x2108, 0x2948, 0x252A, 0x210A, 0x2946, 0x7382, 0x0}},
    {12240, 2, {0x100, 0xFF0, 0x810, 0xFF0, 0x810, 0xFF0, 0x810, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x7FFE, 0x420, 0x420, 0x3820, 0x0}},
    {12241, 2, {0x200, 0x104, 0x7FFE, 0x248, 0x3EB0, 0x1528, 0x1528, 0x2534, 0x5922, 0x810, 0xFF0, 0x810, 0xFF0, 0x1010, 0x6010, 0x0}},
    {12242, 2, {0x80, 0x890, 0x8F8, 0x880, 0xFFFE, 0x440, 0x4444, 0x4AA4, 0x5114, 0x7FFC, 0x4444, 0x4AA4, 0x5114, 0x7FFC, 0x4004, 0x0}},
    {12243, 2, {0x1040, 0x844, 0x7F7E, 0x2240, 0x147C, 0x7F04, 0x7C, 0x3E40, 0x227C, 0x3E40, 0x227C, 0x3E40, 0x227E, 0x2E42, 0x243E, 0x0}},
    {12244, 2, {0x840, 0xFE0, 0x840, 0x1080, 0x7FFC, 0x1144, 0x7F7C, 0x544, 0x7F6C, 0x554, 0x7D6C, 0x544, 0x7F7E, 0x502, 0x7CFC, 0x0}},
    {12245, 2, {0x180, 0x660, 0x1818, 0x63C6, 0x0, 0x7BDE, 0x4A52, 0x7BDE, 0x0, 0x3FFC, 0x2244, 0x3FFC, 0x2244, 0x225C, 0x2248, 0x0}},
    {12272, 2, {0xEDB6, 0x0, 0x8102, 0x8102, 0x0, 0x8102, 0x8102, 0x0, 0x8102, 0x8102, 0x0, 0x8102, 0x8102, 0x100, 0x8002, 0xB6DA}},
    {12273, 2, {0xEDB6, 0x0, 0x8002, 0x8002, 0x8002, 0x0, 0x8002, 0x8002, 0x36D8, 0x8002, 0x8002, 0x0, 0x8002, 0x8002, 0x0, 0xEDB6}},
    {12274, 2, {0xEDB6, 0x0, 0x8442, 0x8442, 0x8442, 0x0, 0x8442, 0x8442, 0x0, 0x8442, 0x8442, 0x0, 0x8442, 0x8442, 0x0, 0xEDB6}},
    {12275, 2, {0xEDB6, 0x0, 0x8002, 0x8002, 0x8002, 0x36D8, 0x8002, 0x8002, 0x0, 0x8002, 0xB6DA, 0x0, 0x8002, 0x8002, 0x0, 0xEDB6}},
    {12276, 2, {0xAAAA, 0x0, 0x8002, 0x1550, 0x8002, 0x1010, 0x8002, 0x1010, 0x8002, 0x1010, 0x9012, 0x0, 0x9552, 0x8002, 0x0, 0xAAAA}},
    {12277, 2, {0xDB6B, 0x0, 0x8001, 0x9B69, 0x1008, 0x8001, 0x9009, 0x1008, 0x8001, 0x9009, 0x1008, 0x8001, 0x9009, 0x1008, 0x8001, 0xDB6D}},
    {12278, 2, {0xDBB6, 0x0, 0x8822, 0x8822, 0x0, 0x8822, 0x8822, 0x0, 0x8822, 0x8002, 0x8822, 0xD60, 0x8002, 0x8002, 0x0, 0xDBB6}},
    {12279, 2, {0xDB6D, 0x1, 0x8000, 0x8001, 0xB6D, 0x8800, 0x8001, 0x801, 0x8800, 0x8001, 0x801, 0x8B6C, 0x8001, 0x1, 0x8000, 0xB6DB}},
    {12280, 2, {0xDB6D, 0x1, 0x8000, 0x8001, 0xB6D, 0x8800, 0x8001, 0x801, 0x8800, 0x8001, 0x801, 0x8800, 0x8001, 0x801, 0x8800, 0xB6DB}},
    {12281, 2, {0xB6DB, 0x8000, 0x1, 0x8001, 0x8B6C, 0x801, 0x8001, 0x8800, 0x801, 0x8001, 0x8800, 0xB6D, 0x8001, 0x8000, 0x1, 0xDB6D}},
    {12282, 2, {0xDB6D, 0x1, 0x9000, 0x9001, 0x1, 0x9000, 0x9001, 0x1, 0x9000, 0x9001, 0x1, 0x9000, 0x9B6D, 0x1, 0x8000, 0xB6DB}},
    {12283, 2, {0x0, 0xDB60, 0x8020, 0x0, 0x9B6C, 0x9024, 0x0, 0x9024, 0x9024, 0x0, 0x9024, 0xDB64, 0x0, 0x1004, 0x1B6C, 0x0}},
    {12288, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12289, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1000, 0xC00, 0x600, 0x200, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12290, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC00, 0x1200, 0x1200, 0xC00, 0x0, 0x0, 0x0, 0x0}},
    {12291, 2, {0x0, 0x0, 0x0, 0x240, 0x240, 0x240, 0x480, 0x480, 0x900, 0x900, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12292, 2, {0x0, 0x600, 0x1820, 0x2030, 0x4128, 0x41A8, 0x8144, 0x8D04, 0x9304, 0x4908, 0x4408, 0x2410, 0x1860, 0x180, 0x0, 0x0}},
    {12293, 2, {0x1000, 0x1000, 0x1000, 0x3FC0, 0x2040, 0x4080, 0x8080, 0x900, 0x600, 0x200, 0x100, 0x100, 0x0, 0x0, 0x0, 0x0}},
    {12294, 2, {0x0, 0x20, 0x10, 0x20, 0x3840, 0x4480, 0x4300, 0x8300, 0x8500, 0x8900, 0x9000, 0xA000, 0xC000, 0x0, 0x0, 0x0}},
    {12295, 2, {0x0, 0x780, 0x1860, 0x2010, 0x4008, 0x4008, 0x8004, 0x8004, 0x8004, 0x8004, 0x4008, 0x4008, 0x2010, 0x1860, 0x780, 0x0}},
    {12296, 2, {0x8, 0x8, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x40, 0x20, 0x20, 0x10, 0x10, 0x8, 0x8, 0x0}},
    {12297, 2, {0x2000, 0x2000, 0x1000, 0x1000, 0x800, 0x800, 0x400, 0x400, 0x400, 0x800, 0x800, 0x1000, 0x1000, 0x2000, 0x2000, 0x0}},
    {12298, 2, {0x12, 0x12, 0x24, 0x24, 0x48, 0x48, 0x90, 0x90, 0x48, 0x48, 0x24, 0x24, 0x12, 0x12, 0x0, 0x0}},
    {12299, 2, {0x4800, 0x4800, 0x2400, 0x2400, 0x1200, 0x1200, 0x900, 0x900, 0x1200, 0x1200, 0x2400, 0x2400, 0x4800, 0x4800, 0x0, 0x0}},
    {12300, 2, {0x7C, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12301, 2, {0x0, 0x0, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x3E00, 0x0, 0x0, 0x0}},
    {12302, 2, {0x7E, 0x42, 0x5E, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x70, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12303, 2, {0x0, 0x0, 0x700, 0x500, 0x500, 0x500, 0x500, 0x500, 0x500, 0x500, 0x3D00, 0x2100, 0x3F00, 0x0, 0x0, 0x0}},
    {12304, 2, {0x7C, 0x78, 0x70, 0x70, 0x70, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x70, 0x70, 0x78, 0x7C, 0x0}},
    {12305, 2, {0x1F00, 0xF00, 0x700, 0x700, 0x300, 0x300, 0x300, 0x300, 0x300, 0x300, 0x300, 0x700, 0x700, 0xF00, 0x1F00, 0x0}},
    {12306, 2, {0x7FFE, 0x0, 0x0, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0, 0x0, 0x0}},
    {12307, 2, {0x3FFE, 0x3FFE, 0x3FFE, 0x3FFE, 0x3FFE, 0x0, 0x0, 0x0, 0x3FFE, 0x3FFE, 0x3FFE, 0x3FFE, 0x3FFE, 0x0, 0x0, 0x0}},
    {12308, 2, {0x8, 0x10, 0x20, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x20, 0x10, 0x8}},
    {12309, 2, {0x1000, 0x800, 0x400, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x400, 0x800, 0x1000}},
    {12310, 2, {0x7F0, 0x420, 0x440, 0x480, 0x480, 0x500, 0x500, 0x500, 0x500, 0x500, 0x500, 0x480, 0x480, 0x440, 0x420, 0x7F0}},
    {12311, 2, {0x1FC0, 0x840, 0x440, 0x240, 0x240, 0x140, 0x140, 0x140, 0x140, 0x140, 0x140, 0x240, 0x240, 0x440, 0x840, 0x1FC0}},
    {12312, 2, {0x1000, 0x6000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0x6000, 0x1000}},
    {12313, 2, {0x8000, 0x6000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x6000, 0x8000}},
    {12314, 2, {0xF000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xA000, 0xF000}},
    {12315, 2, {0xF000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0x5000, 0xF000}},
    {12316, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0x8304, 0x7C00, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12317, 2, {0x48, 0x24, 0x12, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12318, 2, {0x1200, 0x2400, 0x4800, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12319, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2400, 0x4800, 0xD800, 0x9000, 0x0}},
    {12320, 2, {0x0, 0x3FFC, 0x3FFC, 0x0, 0x3FFC, 0x3FFC, 0x5596, 0xD9B9, 0xA5A5, 0xA5A5, 0x599A, 0x4182, 0x2424, 0x13C8, 0x810, 0x7E0}},
    {12321, 2, {0xC000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000}},
    {12322, 2, {0x60, 0x40, 0xC40, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x40, 0x40, 0x40, 0x0, 0x0}},
    {12323, 2, {0x4, 0x1004, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1004, 0x1004, 0x4, 0x0, 0x0}},
    {12324, 2, {0x10, 0x10, 0x1020, 0x820, 0x440, 0x240, 0x180, 0xC0, 0x120, 0x210, 0x408, 0x808, 0x3000, 0x0, 0x0, 0x0}},
    {12325, 2, {0x0, 0x420, 0x430, 0x820, 0x1CC0, 0x300, 0x480, 0x840, 0x820, 0x1020, 0x1020, 0x1020, 0x1040, 0x840, 0x780, 0x0}},
    {12326, 2, {0x0, 0x100, 0x80, 0x84, 0x7FFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12327, 2, {0x0, 0x400, 0x400, 0x400, 0x400, 0xFFF0, 0x0, 0x0, 0x0, 0xFFF0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12328, 2, {0x100, 0x100, 0x100, 0x7FFE, 0x0, 0x0, 0x0, 0x3FFC, 0x0, 0x0, 0x0, 0x0, 0x7FFE, 0x0, 0x0, 0x0}},
    {12329, 2, {0x800, 0x800, 0x800, 0xFFC, 0x1010, 0x2810, 0x420, 0x220, 0x140, 0x80, 0xC0, 0x120, 0x210, 0xC10, 0x3000, 0x0}},
    {12330, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6000, 0x9000, 0x9000, 0x6000}},
    {12331, 2, {0x0, 0x6000, 0x9000, 0x9000, 0x6000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12332, 2, {0x0, 0x6, 0x9, 0x9, 0x6, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12333, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x9, 0x9, 0x6}},
    {12334, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC000, 0xC000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12335, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC000, 0xC000, 0x0, 0xC000, 0xC000, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12336, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7070, 0x8888, 0x707, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12337, 2, {0x800, 0xC00, 0x800, 0x1000, 0x1000, 0x2000, 0x4000, 0x8000, 0x8000, 0x4000, 0x2000, 0x1000, 0x800, 0x800, 0x400, 0x400}},
    {12338, 2, {0x800, 0xC00, 0x880, 0x1040, 0x1240, 0x2100, 0x4100, 0x8000, 0x8000, 0x4000, 0x2000, 0x1000, 0x800, 0x800, 0x400, 0x400}},
    {12339, 2, {0x400, 0x600, 0x400, 0x800, 0x800, 0x800, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000, 0x8000}},
    {12340, 2, {0x400, 0x600, 0x440, 0x820, 0x880, 0x840, 0x1040, 0x1000, 0x1000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000, 0x8000}},
    {12341, 2, {0x0, 0x8000, 0x4000, 0x2000, 0x2000, 0x1000, 0x1000, 0x800, 0x800, 0x400, 0x400, 0x200, 0x200, 0x100, 0x100, 0x100}},
    {12342, 2, {0x7C0, 0x1830, 0x2008, 0x5FF4, 0x5FF4, 0x8002, 0x9FF2, 0x9FF2, 0x8102, 0x8102, 0x8102, 0x4104, 0x4104, 0x2108, 0x1830, 0x7C0}},
    {12343, 2, {0x8282, 0x8282, 0x8282, 0x4444, 0x4444, 0x2828, 0x2828, 0x1010, 0x1010, 0x2828, 0x2828, 0x4444, 0x4444, 0x8282, 0x8282, 0x8282}},
    {12344, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {12345, 2, {0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820}},
    {12346, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0x1088, 0x1088, 0xFFFE, 0x1088, 0x1088, 0x1088, 0x1088, 0x1088, 0x2088, 0x2088, 0x4088, 0x8008}},
    {12347, 2, {0x0, 0x0, 0x200, 0x100, 0x80, 0x40, 0x80, 0x100, 0x80, 0x40, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12348, 2, {0x0, 0x0, 0x0, 0x7FF0, 0x4030, 0x4050, 0x4090, 0x4110, 0x4210, 0x4410, 0x4810, 0x5010, 0x6010, 0x7FF0, 0x0, 0x0}},
    {12349, 2, {0x0, 0x0, 0x1000, 0x2860, 0x4490, 0x310, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x8, 0x8, 0x4, 0x0}},
    {12350, 2, {0xDB6D, 0x8001, 0xE00, 0x9189, 0x8071, 0x0, 0x8011, 0x8021, 0x1FF8, 0x8081, 0x8101, 0x1FF8, 0x8401, 0x8800, 0x1, 0xDB6D}},
    {12351, 1, {0x0, 0xFE, 0x82, 0xC6, 0xC6, 0xAA, 0xAA, 0x92, 0xAA, 0xAA, 0xC6, 0xC6, 0x82, 0xFE, 0x0, 0x0}},
    {12353, 2, {0x0, 0x0, 0x0, 0x0, 0x400, 0x200, 0x1380, 0xE00, 0x480, 0x7E0, 0xD10, 0x1610, 0x2610, 0x2A20, 0x10C0, 0x0}},
    {12354, 2, {0x0, 0x400, 0x200, 0x2C0, 0x1F00, 0x480, 0x480, 0x7E0, 0xD10, 0x1508, 0x2208, 0x2608, 0x2A10, 0x1060, 0x0, 0x0}},
    {12355, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1040, 0x1020, 0x1010, 0x1030, 0xA00, 0xC00, 0x400, 0x0, 0x0}},
    {12356, 2, {0x0, 0x0, 0x0, 0x0, 0x4000, 0x2060, 0x2010, 0x2008, 0x2208, 0x2418, 0x1408, 0x1800, 0x800, 0x0, 0x0, 0x0}},
    {12357, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x180, 0x0, 0x1380, 0xC40, 0x40, 0x40, 0x40, 0x80, 0x300, 0x0}},
    {12358, 2, {0x0, 0x200, 0x1C0, 0x0, 0x0, 0x3C0, 0x1C20, 0x20, 0x20, 0x20, 0x20, 0x40, 0x40, 0x80, 0x100, 0x0}},
    {12359, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x180, 0x0, 0x1C0, 0xE80, 0x100, 0x200, 0x700, 0x900, 0x10F0, 0x0}},
    {12360, 2, {0x0, 0x400, 0x380, 0x0, 0x0, 0x11C0, 0xE80, 0x100, 0x200, 0x400, 0xE00, 0x1100, 0x2100, 0x20F8, 0x0, 0x0}},
    {12361, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1040, 0x830, 0xC20, 0x3800, 0xBC0, 0xC20, 0x1820, 0x2820, 0x18C0, 0x800, 0x0}},
    {12362, 2, {0x0, 0x0, 0x800, 0x430, 0x508, 0x60C, 0x1C00, 0x460, 0x590, 0x608, 0xC08, 0x3488, 0xC70, 0x400, 0x0, 0x0}},
    {12363, 2, {0x0, 0x0, 0x800, 0x400, 0x460, 0x4710, 0x3C88, 0x898, 0x880, 0x880, 0x1080, 0x1100, 0x2500, 0x2200, 0x0, 0x0}},
    {12364, 2, {0x0, 0x8, 0x824, 0x410, 0x440, 0x4720, 0x3C90, 0x888, 0x898, 0x880, 0x1080, 0x1100, 0x2500, 0x2200, 0x0, 0x0}},
    {12365, 2, {0x0, 0x200, 0x100, 0x160, 0xF80, 0x80, 0xF0, 0x1F40, 0x20, 0x20, 0x7F0, 0x810, 0x800, 0x600, 0x1C0, 0x0}},
    {12366, 2, {0x0, 0x408, 0x224, 0x2D0, 0x1F00, 0x100, 0x1E0, 0x3E80, 0x40, 0x40, 0xFE0, 0x1020, 0x1000, 0xC00, 0x380, 0x0}},
    {12367, 2, {0x0, 0x80, 0x40, 0x40, 0x80, 0x100, 0x600, 0x800, 0x800, 0x400, 0x200, 0x100, 0x80, 0x40, 0x40, 0x0}},
    {12368, 2, {0x0, 0x80, 0x40, 0x40, 0x88, 0x124, 0x610, 0x800, 0x800, 0x400, 0x200, 0x100, 0x80, 0x40, 0x40, 0x0}},
    {12369, 2, {0x0, 0x80, 0x2040, 0x1020, 0x1020, 0x103C, 0x23E0, 0x2020, 0x2020, 0x2020, 0x2020, 0x1820, 0x1040, 0x1040, 0x80, 0x0}},
    {12370, 2, {0x0, 0x108, 0x40A4, 0x2050, 0x2040, 0x2078, 0x47C0, 0x4040, 0x4040, 0x4040, 0x4040, 0x3040, 0x2080, 0x2080, 0x100, 0x0}},
    {12371, 2, {0x0, 0x0, 0x800, 0x7E0, 0x80, 0x100, 0x0, 0x0, 0x0, 0x0, 0x2000, 0x2000, 0x1010, 0xFE0, 0x0, 0x0}},
    {12372, 2, {0x0, 0x8, 0x804, 0x7F0, 0x88, 0x100, 0x0, 0x0, 0x0, 0x0, 0x2000, 0x2000, 0x1010, 0xFE0, 0x0, 0x0}},
    {12373, 2, {0x0, 0x400, 0x200, 0x130, 0x21C0, 0x1E80, 0x40, 0x40, 0x7E0, 0x820, 0x1000, 0x1000, 0xC00, 0x3C0, 0x0, 0x0}},
    {12374, 2, {0x0, 0x408, 0x204, 0x130, 0x21C8, 0x1E80, 0x40, 0x40, 0x7E0, 0x820, 0x1000, 0x1000, 0xC00, 0x3C0, 0x0, 0x0}},
    {12375, 2, {0x0, 0x0, 0x1000, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800, 0x808, 0x810, 0x460, 0x380, 0x0, 0x0}},
    {12376, 2, {0x0, 0x0, 0x1020, 0x890, 0x840, 0x800, 0x800, 0x800, 0x800, 0x800, 0x808, 0x810, 0x460, 0x380, 0x0, 0x0}},
    {12377, 2, {0x0, 0x100, 0x80, 0x40F8, 0x3F80, 0x80, 0x380, 0x4C0, 0x440, 0x3C0, 0x40, 0x80, 0x80, 0x100, 0x200, 0x0}},
    {12378, 2, {0x0, 0x128, 0x94, 0x40F8, 0x3F80, 0x80, 0x380, 0x4C0, 0x440, 0x3C0, 0x40, 0x80, 0x80, 0x100, 0x200, 0x0}},
    {12379, 2, {0x0, 0x0, 0x80, 0x1040, 0x840, 0x878, 0x4BC0, 0x3C40, 0x840, 0x840, 0x8C0, 0x840, 0x400, 0x3E0, 0x0, 0x0}},
    {12380, 2, {0x0, 0x8, 0x84, 0x1050, 0x848, 0x878, 0x4BC0, 0x3C40, 0x840, 0x840, 0x8C0, 0x840, 0x400, 0x3E0, 0x0, 0x0}},
    {12381, 2, {0x0, 0x0, 0x9C0, 0x640, 0x80, 0x100, 0x238, 0x7C0, 0x3880, 0x100, 0x200, 0x200, 0x200, 0x100, 0xE0, 0x0}},
    {12382, 2, {0x0, 0x0, 0x9C8, 0x664, 0x90, 0x100, 0x238, 0x7C0, 0x3880, 0x100, 0x200, 0x200, 0x200, 0x100, 0xE0, 0x0}},
    {12383, 2, {0x0, 0x0, 0x800, 0x480, 0x700, 0x3C00, 0x8F0, 0x818, 0x820, 0x1000, 0x1000, 0x1100, 0x2100, 0x20F8, 0x0, 0x0}},
    {12384, 2, {0x0, 0x8, 0x824, 0x490, 0x700, 0x3C00, 0x8F0, 0x818, 0x820, 0x1000, 0x1000, 0x1100, 0x2100, 0x20F8, 0x0, 0x0}},
    {12385, 2, {0x0, 0x800, 0x400, 0x4C0, 0x4700, 0x3C00, 0x800, 0x8C0, 0xB20, 0x1C10, 0x1010, 0x10, 0x20, 0xC0, 0x700, 0x0}},
    {12386, 2, {0x0, 0x800, 0x408, 0x4C4, 0x4710, 0x3C08, 0x800, 0x8C0, 0xB20, 0x1C10, 0x1010, 0x10, 0x20, 0xC0, 0x700, 0x0}},
    {12387, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C0, 0x620, 0x3810, 0x10, 0x10, 0x20, 0xC0, 0x700, 0x0}},
    {12388, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x4C10, 0x3008, 0x8, 0x8, 0x8, 0x10, 0x60, 0x380, 0x0, 0x0, 0x0}},
    {12389, 2, {0x0, 0x8, 0x24, 0x10, 0x3E0, 0x4C10, 0x3008, 0x8, 0x8, 0x8, 0x10, 0x60, 0x380, 0x0, 0x0, 0x0}},
    {12390, 2, {0x0, 0x0, 0x38, 0x43C0, 0x3C80, 0x100, 0x100, 0x200, 0x200, 0x200, 0x200, 0x200, 0x100, 0xE0, 0x0, 0x0}},
    {12391, 2, {0x0, 0x0, 0x38, 0x43C0, 0x3C88, 0x124, 0x110, 0x200, 0x200, 0x200, 0x200, 0x200, 0x100, 0xE0, 0x0, 0x0}},
    {12392, 2, {0x0, 0x0, 0x400, 0x200, 0x200, 0x200, 0x270, 0x380, 0x600, 0x800, 0x1000, 0x1000, 0x800, 0x7F0, 0x0, 0x0}},
    {12393, 2, {0x0, 0x0, 0x408, 0x224, 0x210, 0x200, 0x270, 0x380, 0x600, 0x800, 0x1000, 0x1000, 0x800, 0x7F0, 0x0, 0x0}},
    {12394, 2, {0x0, 0x400, 0x200, 0x280, 0x3F00, 0x470, 0x818, 0x820, 0x1040, 0x1040, 0x2040, 0x3C0, 0x460, 0x450, 0x380, 0x0}},
    {12395, 2, {0x0, 0x0, 0x2000, 0x1000, 0x10F0, 0x1318, 0x2020, 0x2000, 0x2000, 0x2000, 0x2200, 0x2A00, 0x11F8, 0x1000, 0x1000, 0x0}},
    {12396, 2, {0x0, 0x0, 0x100, 0x80, 0x80, 0x13E0, 0x1510, 0x1908, 0x1208, 0x2A08, 0x4C08, 0x4C68, 0x5490, 0x2068, 0x0, 0x0}},
    {12397, 2, {0x0, 0x1000, 0x800, 0x800, 0x800, 0x4CE0, 0x3F10, 0xC10, 0x810, 0x1810, 0x1810, 0x28F0, 0x5918, 0x8E0, 0x0, 0x0}},
    {12398, 2, {0x0, 0x0, 0x0, 0x3C0, 0xD20, 0x1110, 0x1108, 0x2208, 0x2208, 0x2408, 0x2410, 0x2810, 0x1060, 0x180, 0x0, 0x0}},
    {12399, 2, {0x0, 0x40, 0x2020, 0x1020, 0x1438, 0x13E0, 0x2020, 0x2020, 0x2020, 0x2020, 0x29E0, 0x1230, 0x1228, 0x11C0, 0x0, 0x0}},
    {12400, 2, {0x0, 0x88, 0x4044, 0x2050, 0x2878, 0x27C0, 0x4040, 0x4040, 0x4040, 0x4040, 0x53C0, 0x2460, 0x2450, 0x2380, 0x0, 0x0}},
    {12401, 2, {0x0, 0x98, 0x4064, 0x2064, 0x2878, 0x27C0, 0x4040, 0x4040, 0x4040, 0x4040, 0x53C0, 0x2460, 0x2450, 0x2380, 0x0, 0x0}},
    {12402, 2, {0x0, 0x0, 0x40, 0x440, 0x3C20, 0x820, 0x830, 0x1028, 0x1020, 0x1020, 0x1040, 0x1040, 0x880, 0x700, 0x0, 0x0}},
    {12403, 2, {0x0, 0x8, 0x44, 0x450, 0x3C28, 0x820, 0x830, 0x1028, 0x1020, 0x1020, 0x1040, 0x1040, 0x880, 0x700, 0x0, 0x0}},
    {12404, 2, {0x0, 0x30, 0x48, 0x448, 0x3C30, 0x820, 0x830, 0x1028, 0x1020, 0x1020, 0x1040, 0x1040, 0x880, 0x700, 0x0, 0x0}},
    {12405, 2, {0x0, 0x0, 0x400, 0x380, 0x80, 0x100, 0x200, 0x200, 0x100, 0x2B0, 0x4C88, 0x389C, 0x2480, 0x300, 0x0, 0x0}},
    {12406, 2, {0x0, 0x0, 0x400, 0x388, 0xA4, 0x110, 0x200, 0x200, 0x100, 0x2B0, 0x4C88, 0x389C, 0x2480, 0x300, 0x0, 0x0}},
    {12407, 2, {0x0, 0x0, 0x400, 0x398, 0xA4, 0x124, 0x218, 0x200, 0x100, 0x2B0, 0x4C88, 0x389C, 0x2480, 0x300, 0x0, 0x0}},
    {12408, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x600, 0x900, 0x1080, 0x2040, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12409, 2, {0x0, 0x0, 0x0, 0x10, 0x48, 0x620, 0x900, 0x1080, 0x2040, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12410, 2, {0x0, 0x0, 0x30, 0x48, 0x48, 0x630, 0x900, 0x1080, 0x2040, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12411, 2, {0x0, 0x0, 0x2230, 0x11C0, 0x1040, 0x1038, 0x23E0, 0x2020, 0x2020, 0x2020, 0x29E0, 0x1230, 0x1228, 0x11C0, 0x0, 0x0}},
    {12412, 2, {0x0, 0x0, 0x4468, 0x2384, 0x2090, 0x2078, 0x47C0, 0x4040, 0x4040, 0x4040, 0x53C0, 0x2460, 0x2450, 0x2380, 0x0, 0x0}},
    {12413, 2, {0x0, 0x18, 0x4464, 0x23A4, 0x2098, 0x2070, 0x47C0, 0x4040, 0x4040, 0x4040, 0x53C0, 0x2460, 0x2450, 0x2380, 0x0, 0x0}},
    {12414, 2, {0x0, 0x100, 0x80, 0x10F0, 0xF80, 0x80, 0x10E0, 0xF80, 0x80, 0x80, 0xF80, 0x10E0, 0x1090, 0xF00, 0x0, 0x0}},
    {12415, 2, {0x0, 0x0, 0x880, 0x780, 0x100, 0x100, 0x220, 0x210, 0xFF0, 0x3418, 0x4424, 0x4820, 0x3040, 0x80, 0x0, 0x0}},
    {12416, 2, {0x0, 0x800, 0x400, 0x2700, 0x1C30, 0x408, 0x1C0C, 0x2400, 0x2400, 0x2C20, 0x1810, 0x810, 0x810, 0x7E0, 0x0, 0x0}},
    {12417, 2, {0x0, 0x0, 0x100, 0x80, 0x1080, 0x13C0, 0x1CA0, 0x1110, 0x2908, 0x2A08, 0x4608, 0x4408, 0x4A10, 0x3060, 0x0, 0x0}},
    {12418, 2, {0x0, 0x200, 0x100, 0x1100, 0xE00, 0x380, 0x1200, 0x2420, 0x1C20, 0x710, 0x410, 0x410, 0x410, 0x220, 0x1C0, 0x0}},
    {12419, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x880, 0x9E0, 0xE10, 0x3C10, 0x460, 0x200, 0x200, 0x100, 0x100, 0x0}},
    {12420, 2, {0x0, 0x100, 0x80, 0x1C0, 0x1800, 0x10F0, 0xB08, 0x4C08, 0x3408, 0x270, 0x200, 0x100, 0x100, 0x80, 0x80, 0x0}},
    {12421, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x21E0, 0x2710, 0x2910, 0x3110, 0x2320, 0x21C0, 0x100, 0x200, 0x0}},
    {12422, 2, {0x0, 0x0, 0x300, 0x4080, 0x21E0, 0x2290, 0x2488, 0x2888, 0x3088, 0x3488, 0x2290, 0x21E0, 0x80, 0x100, 0x200, 0x0}},
    {12423, 2, {0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x120, 0x1C0, 0x100, 0x100, 0x100, 0xF00, 0x11C0, 0x1120, 0xE00, 0x0}},
    {12424, 2, {0x0, 0x200, 0x100, 0x100, 0x130, 0x1C0, 0x100, 0x100, 0x100, 0x100, 0x1F80, 0x2160, 0x2110, 0x1E00, 0x0, 0x0}},
    {12425, 2, {0x0, 0x200, 0x100, 0x1C0, 0x600, 0x800, 0x800, 0x800, 0x9E0, 0xE10, 0x810, 0x810, 0x20, 0xC0, 0x300, 0x0}},
    {12426, 2, {0x0, 0x1000, 0x980, 0xA40, 0xC20, 0xC20, 0x820, 0x820, 0x820, 0x820, 0x40, 0x40, 0x80, 0x100, 0x200, 0x0}},
    {12427, 2, {0x0, 0x0, 0x8C0, 0x740, 0x80, 0x100, 0x200, 0x7C0, 0xC20, 0x1010, 0x2010, 0x390, 0x460, 0x440, 0x380, 0x0}},
    {12428, 2, {0x0, 0x0, 0x1000, 0x800, 0x9C0, 0xA20, 0x1C20, 0x6C20, 0x820, 0x1820, 0x1820, 0x2820, 0x4824, 0x1818, 0x800, 0x0}},
    {12429, 2, {0x0, 0x0, 0x1180, 0xE80, 0x100, 0x200, 0x400, 0xFC0, 0x820, 0x1010, 0x2010, 0x10, 0x20, 0xC0, 0x700, 0x0}},
    {12430, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x800, 0x400, 0x400, 0x6E0, 0x1D10, 0x610, 0xC10, 0xC10, 0x1420, 0x440, 0x0}},
    {12431, 2, {0x0, 0x1000, 0x800, 0x800, 0x800, 0x4CE0, 0x3F10, 0xC08, 0x808, 0x1808, 0x1808, 0x2810, 0x4820, 0x18C0, 0x800, 0x0}},
    {12432, 2, {0x0, 0x0, 0x980, 0x680, 0x100, 0x100, 0x3E0, 0xE10, 0x1208, 0x2408, 0x2408, 0x28C8, 0x1930, 0x10E0, 0x0, 0x0}},
    {12433, 2, {0x0, 0x0, 0x980, 0x700, 0x200, 0x7C0, 0x820, 0x1020, 0x620, 0x9C0, 0x700, 0x800, 0x1CE0, 0x2310, 0x4038, 0x0}},
    {12434, 2, {0x0, 0x400, 0x200, 0x12C0, 0xF00, 0x400, 0x420, 0xF38, 0x18C0, 0x2180, 0x280, 0x480, 0x800, 0x800, 0x7E0, 0x0}},
    {12435, 2, {0x0, 0x200, 0x100, 0x100, 0x200, 0x200, 0x400, 0x400, 0xF00, 0x880, 0x1080, 0x1088, 0x2090, 0x2060, 0x0, 0x0}},
    {12436, 2, {0x0, 0x200, 0x1C8, 0x4, 0x10, 0x3C8, 0x1C20, 0x20, 0x20, 0x20, 0x20, 0x40, 0x40, 0x80, 0x100, 0x0}},
    {12437, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x400, 0x460, 0x3F10, 0x488, 0x898, 0x1080, 0x1100, 0x2500, 0x2200, 0x0, 0x0}},
    {12438, 2, {0x0, 0x0, 0x0, 0x80, 0x40, 0x820, 0x820, 0x838, 0x11E0, 0x1020, 0x1020, 0x1420, 0x1440, 0x840, 0x80, 0x0}},
    {12443, 2, {0x0, 0x1000, 0x4800, 0x2000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12444, 2, {0x0, 0x3000, 0x4800, 0x4800, 0x3000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12445, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC00, 0x300, 0x80, 0xC0, 0x100, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12446, 2, {0x0, 0x0, 0x0, 0x20, 0x90, 0xC40, 0x300, 0x80, 0xC0, 0x100, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12447, 2, {0x0, 0x0, 0x600, 0x180, 0x800, 0x800, 0x800, 0x800, 0xFC0, 0x1020, 0x10, 0x10, 0x20, 0xC0, 0x300, 0x0}},
    {12448, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E00, 0x0, 0x0, 0x0, 0x7E00, 0x0, 0x0, 0x0, 0x0}},
    {12449, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x21E0, 0x1E20, 0x240, 0x180, 0x100, 0x100, 0x200, 0x400, 0x0}},
    {12450, 2, {0x0, 0x0, 0x0, 0x21F0, 0x1E10, 0x20, 0x240, 0x180, 0x100, 0x100, 0x100, 0x200, 0x200, 0x400, 0x800, 0x0}},
    {12451, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x40, 0x80, 0x380, 0x1C80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {12452, 2, {0x0, 0x0, 0x20, 0x20, 0x40, 0x80, 0x300, 0xD00, 0x3080, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {12453, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x1E0, 0x1E20, 0x820, 0x840, 0x40, 0x80, 0x100, 0x200, 0x0}},
    {12454, 2, {0x0, 0x0, 0x200, 0x100, 0x100, 0x1F0, 0x3E10, 0x1010, 0x1020, 0x820, 0x840, 0x40, 0x80, 0x100, 0x200, 0x0}},
    {12455, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xE0, 0xF00, 0x100, 0x100, 0x100, 0x1E0, 0x1E10, 0x0, 0x0}},
    {12456, 2, {0x0, 0x0, 0x0, 0x0, 0x70, 0x1F80, 0x200, 0x100, 0x100, 0x100, 0x100, 0x3F0, 0x3C08, 0x0, 0x0, 0x0}},
    {12457, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x80, 0x10E0, 0xF80, 0x180, 0x280, 0x480, 0x880, 0x1180, 0x80, 0x0}},
    {12458, 2, {0x0, 0x0, 0x80, 0x40, 0x40, 0x2078, 0x1FC0, 0xC0, 0x140, 0x240, 0x440, 0x840, 0x1040, 0xC0, 0x40, 0x0}},
    {12459, 2, {0x0, 0x0, 0x200, 0x100, 0x100, 0x21F0, 0x1F10, 0x210, 0x210, 0x420, 0x420, 0x820, 0x1140, 0x20C0, 0x80, 0x0}},
    {12460, 2, {0x0, 0x8, 0x224, 0x114, 0x110, 0x21F0, 0x1F10, 0x210, 0x210, 0x420, 0x420, 0x820, 0x1140, 0x20C0, 0x80, 0x0}},
    {12461, 2, {0x0, 0x0, 0x400, 0x200, 0x270, 0x380, 0x1D00, 0x100, 0x178, 0x380, 0x3C80, 0x80, 0x40, 0x40, 0x40, 0x0}},
    {12462, 2, {0x0, 0x8, 0x424, 0x214, 0x270, 0x380, 0x1D00, 0x100, 0x178, 0x380, 0x3C80, 0x80, 0x40, 0x40, 0x40, 0x0}},
    {12463, 2, {0x0, 0x0, 0x400, 0x270, 0x390, 0x220, 0x420, 0x440, 0x840, 0x1080, 0x80, 0x100, 0x200, 0x400, 0x800, 0x0}},
    {12464, 2, {0x0, 0x8, 0x824, 0x4D4, 0x750, 0x440, 0x840, 0x880, 0x1080, 0x2100, 0x100, 0x200, 0x400, 0x800, 0x1000, 0x0}},
    {12465, 2, {0x0, 0x0, 0x800, 0x400, 0x400, 0x438, 0xFC0, 0x900, 0x1080, 0x2080, 0x80, 0x100, 0x100, 0x200, 0x400, 0x0}},
    {12466, 2, {0x0, 0x8, 0x824, 0x414, 0x410, 0x438, 0xFC0, 0x900, 0x1080, 0x2080, 0x80, 0x100, 0x100, 0x200, 0x400, 0x0}},
    {12467, 2, {0x0, 0x0, 0x0, 0x0, 0x21F0, 0x1E10, 0x10, 0x10, 0x10, 0x20, 0x20, 0x3E0, 0x1C20, 0x0, 0x0, 0x0}},
    {12468, 2, {0x0, 0x8, 0x24, 0x14, 0x43D0, 0x3C40, 0x40, 0x40, 0x40, 0x80, 0x80, 0x780, 0x3880, 0x0, 0x0, 0x0}},
    {12469, 2, {0x0, 0x0, 0x80, 0x840, 0x440, 0x440, 0x24F8, 0x1F40, 0x440, 0x440, 0x440, 0x80, 0x80, 0x100, 0x200, 0x0}},
    {12470, 2, {0x0, 0x8, 0xA4, 0x854, 0x450, 0x440, 0x24F8, 0x1F40, 0x440, 0x440, 0x440, 0x80, 0x80, 0x100, 0x200, 0x0}},
    {12471, 2, {0x0, 0x0, 0x0, 0xC00, 0x200, 0x0, 0x1800, 0x408, 0x10, 0x20, 0x40, 0x80, 0x1300, 0xC00, 0x0, 0x0}},
    {12472, 2, {0x0, 0x0, 0x8, 0xC24, 0x214, 0x10, 0x1800, 0x408, 0x10, 0x20, 0x40, 0x80, 0x1300, 0xC00, 0x0, 0x0}},
    {12473, 2, {0x0, 0x0, 0x0, 0x10E0, 0xF20, 0x40, 0x40, 0x80, 0x80, 0x140, 0x220, 0xC10, 0x3010, 0x0, 0x0, 0x0}},
    {12474, 2, {0x0, 0x8, 0x24, 0x21D4, 0x1E50, 0x80, 0x80, 0x100, 0x100, 0x280, 0x440, 0x1820, 0x6020, 0x0, 0x0, 0x0}},
    {12475, 2, {0x0, 0x0, 0x800, 0x400, 0x400, 0x478, 0x588, 0x2610, 0x1C20, 0x440, 0x400, 0x400, 0x200, 0x1F0, 0x0, 0x0}},
    {12476, 2, {0x0, 0x8, 0x824, 0x414, 0x410, 0x478, 0x588, 0x2610, 0x1C20, 0x440, 0x400, 0x400, 0x200, 0x1F8, 0x0, 0x0}},
    {12477, 2, {0x0, 0x0, 0x20, 0x1010, 0x810, 0x410, 0x420, 0x420, 0x40, 0x40, 0x80, 0x80, 0x100, 0x200, 0x400, 0x0}},
    {12478, 2, {0x0, 0x8, 0xA4, 0x2054, 0x1050, 0x840, 0x840, 0x840, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x800, 0x0}},
    {12479, 2, {0x0, 0x0, 0x200, 0x100, 0x1F0, 0x210, 0x220, 0x520, 0x8C0, 0x1040, 0x80, 0x100, 0x200, 0xC00, 0x3000, 0x0}},
    {12480, 2, {0x0, 0x8, 0x424, 0x214, 0x3F0, 0x420, 0x420, 0xA40, 0x1140, 0x2080, 0x100, 0x200, 0x400, 0x1800, 0x6000, 0x0}},
    {12481, 2, {0x0, 0x0, 0x40, 0x180, 0x1E00, 0x100, 0xF8, 0x2380, 0x1C80, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x0}},
    {12482, 2, {0x0, 0x8, 0x64, 0x194, 0x1E10, 0x100, 0xF8, 0x2380, 0x1C80, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x0}},
    {12483, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x1220, 0xA20, 0x840, 0x40, 0x80, 0x100, 0x200, 0x0}},
    {12484, 2, {0x0, 0x0, 0x0, 0x410, 0x2208, 0x1108, 0x910, 0x810, 0x20, 0x20, 0x40, 0x80, 0x100, 0x200, 0x0, 0x0}},
    {12485, 2, {0x0, 0x8, 0x24, 0x1094, 0x850, 0x4440, 0x2440, 0x1040, 0x1080, 0x80, 0x100, 0x200, 0x400, 0x800, 0x0, 0x0}},
    {12486, 2, {0x0, 0x0, 0x0, 0x11E0, 0xE00, 0x0, 0x78, 0x2380, 0x1C80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x800, 0x0}},
    {12487, 2, {0x0, 0x8, 0x24, 0x11D4, 0xE10, 0x0, 0x20F0, 0x1F00, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x800, 0x0}},
    {12488, 2, {0x0, 0x0, 0x800, 0x400, 0x400, 0x400, 0x400, 0x700, 0x4C0, 0x420, 0x400, 0x400, 0x400, 0x400, 0x400, 0x0}},
    {12489, 2, {0x0, 0x20, 0x1090, 0x850, 0x840, 0x800, 0x800, 0xE00, 0x980, 0x840, 0x800, 0x800, 0x800, 0x800, 0x800, 0x0}},
    {12490, 2, {0x0, 0x0, 0x100, 0x80, 0x80, 0x20F8, 0x1F80, 0x80, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x800, 0x0}},
    {12491, 2, {0x0, 0x0, 0x0, 0x0, 0xE0, 0x1F00, 0x0, 0x0, 0x0, 0x0, 0x1F0, 0x3E08, 0x0, 0x0, 0x0, 0x0}},
    {12492, 2, {0x0, 0x0, 0x0, 0x10F0, 0xF10, 0x20, 0x20, 0xE40, 0x1C0, 0xC0, 0x120, 0x210, 0xC10, 0x3000, 0x0, 0x0}},
    {12493, 2, {0x0, 0x0, 0x200, 0x100, 0x1E0, 0x1E20, 0x40, 0x80, 0x100, 0x340, 0xD30, 0x3108, 0x100, 0x100, 0x100, 0x0}},
    {12494, 2, {0x0, 0x0, 0x40, 0x20, 0x20, 0x20, 0x40, 0x40, 0x80, 0x80, 0x100, 0x200, 0x400, 0x800, 0x1000, 0x0}},
    {12495, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x880, 0x440, 0x420, 0x810, 0x810, 0x1008, 0x2008, 0x4008, 0x0, 0x0, 0x0}},
    {12496, 2, {0x0, 0x0, 0x8, 0x24, 0x14, 0x890, 0x440, 0x420, 0x810, 0x810, 0x1008, 0x2008, 0x4008, 0x0, 0x0, 0x0}},
    {12497, 2, {0x0, 0x18, 0x24, 0x24, 0x18, 0x880, 0x440, 0x420, 0x810, 0x810, 0x1008, 0x2008, 0x4008, 0x0, 0x0, 0x0}},
    {12498, 2, {0x0, 0x0, 0x0, 0x1000, 0x800, 0x800, 0x8E0, 0xF00, 0x800, 0x800, 0x800, 0x800, 0x400, 0x3F0, 0x0, 0x0}},
    {12499, 2, {0x0, 0x0, 0x10, 0x1048, 0x828, 0x820, 0x800, 0x8E0, 0xF00, 0x800, 0x800, 0x800, 0x400, 0x3F0, 0x0, 0x0}},
    {12500, 2, {0x0, 0x0, 0x30, 0x1048, 0x848, 0x830, 0x800, 0x8E0, 0xF00, 0x800, 0x800, 0x800, 0x400, 0x3F0, 0x0, 0x0}},
    {12501, 2, {0x0, 0x0, 0x0, 0x10F0, 0xF10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x80, 0x100, 0x200, 0x400, 0x0, 0x0}},
    {12502, 2, {0x0, 0x8, 0x24, 0x23D4, 0x1C50, 0x40, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x800, 0x1000, 0x0, 0x0}},
    {12503, 2, {0x0, 0x18, 0x24, 0x23E4, 0x1C58, 0x40, 0x80, 0x80, 0x100, 0x100, 0x200, 0x400, 0x800, 0x1000, 0x0, 0x0}},
    {12504, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x500, 0x4880, 0x3040, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12505, 2, {0x0, 0x0, 0x20, 0x90, 0x50, 0x240, 0x500, 0x4880, 0x3040, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12506, 2, {0x0, 0x0, 0x60, 0x90, 0x90, 0x260, 0x500, 0x4880, 0x3040, 0x30, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12507, 2, {0x0, 0x0, 0x200, 0x100, 0x100, 0x21F0, 0x1F00, 0x100, 0x140, 0x920, 0x910, 0x1110, 0x2100, 0x300, 0x100, 0x0}},
    {12508, 2, {0x0, 0x8, 0x224, 0x114, 0x110, 0x21F0, 0x1F00, 0x100, 0x140, 0x920, 0x910, 0x1110, 0x2100, 0x300, 0x100, 0x0}},
    {12509, 2, {0x0, 0x18, 0x224, 0x124, 0x118, 0x21F0, 0x1F00, 0x100, 0x140, 0x920, 0x910, 0x1110, 0x2100, 0x300, 0x100, 0x0}},
    {12510, 2, {0x0, 0x0, 0x0, 0x0, 0x20F0, 0x1F08, 0x10, 0x20, 0x40, 0xC80, 0x300, 0x80, 0x40, 0x0, 0x0, 0x0}},
    {12511, 2, {0x0, 0x0, 0x400, 0x300, 0xC0, 0x20, 0x800, 0x600, 0x180, 0x40, 0x1000, 0xC00, 0x300, 0xC0, 0x20, 0x0}},
    {12512, 2, {0x0, 0x0, 0x200, 0x100, 0x100, 0x100, 0x200, 0x200, 0x240, 0x420, 0x430, 0x29C8, 0x1E08, 0x0, 0x0, 0x0}},
    {12513, 2, {0x0, 0x0, 0x40, 0x20, 0x20, 0x20, 0x620, 0x1C0, 0x40, 0xA0, 0x90, 0x100, 0x200, 0xC00, 0x3000, 0x0}},
    {12514, 2, {0x0, 0x0, 0x0, 0x10E0, 0xF00, 0x200, 0x200, 0x23F8, 0x1E00, 0x200, 0x200, 0x200, 0x1F0, 0x0, 0x0, 0x0}},
    {12515, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x400, 0x270, 0x390, 0x1E20, 0x140, 0x100, 0x80, 0x80, 0x80, 0x0}},
    {12516, 2, {0x0, 0x0, 0x800, 0x400, 0x478, 0x388, 0x2610, 0x1A20, 0x140, 0x100, 0x100, 0x80, 0x80, 0x80, 0x0, 0x0}},
    {12517, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C0, 0xE40, 0x40, 0x80, 0x1E0, 0x1E10, 0x0, 0x0, 0x0}},
    {12518, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x11C0, 0xE40, 0x40, 0x80, 0x80, 0x1F0, 0x3E08, 0x0, 0x0, 0x0, 0x0}},
    {12519, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1E0, 0xE20, 0x20, 0x1E0, 0xE40, 0x40, 0x1C0, 0xE20, 0x0, 0x0}},
    {12520, 2, {0x0, 0x0, 0x0, 0x1F0, 0x1E10, 0x10, 0x10, 0x1E0, 0xE20, 0x20, 0x20, 0x20, 0x1E0, 0xE10, 0x0, 0x0}},
    {12521, 2, {0x0, 0x0, 0x0, 0x11C0, 0xE00, 0x70, 0x2390, 0x1C10, 0x10, 0x20, 0x20, 0x40, 0x80, 0x300, 0xC00, 0x0}},
    {12522, 2, {0x0, 0x0, 0x40, 0x820, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x40, 0x40, 0x80, 0x100, 0x200, 0x0}},
    {12523, 2, {0x0, 0x0, 0x0, 0x100, 0x880, 0x480, 0x480, 0x488, 0x488, 0x490, 0x890, 0x8A0, 0x10C0, 0x2080, 0x0, 0x0}},
    {12524, 2, {0x0, 0x0, 0x1000, 0x800, 0x800, 0x800, 0x800, 0x800, 0x810, 0x820, 0x840, 0x880, 0xB00, 0xC00, 0x0, 0x0}},
    {12525, 2, {0x0, 0x0, 0x0, 0x0, 0x21F0, 0x1E10, 0x1010, 0x1010, 0x820, 0x820, 0x9E0, 0xE10, 0x800, 0x0, 0x0, 0x0}},
    {12526, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10F0, 0xF10, 0x820, 0x420, 0x440, 0x40, 0x80, 0x100, 0x0}},
    {12527, 2, {0x0, 0x0, 0x0, 0x21F0, 0x1E10, 0x1010, 0x810, 0x820, 0x820, 0x40, 0x40, 0x80, 0x100, 0x200, 0x400, 0x0}},
    {12528, 2, {0x0, 0x0, 0x100, 0x80, 0x80, 0x1F0, 0x1E80, 0x880, 0x480, 0x5F0, 0x3E88, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {12529, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x1E20, 0x240, 0x180, 0x100, 0x100, 0x1F0, 0x3E08, 0x0, 0x0, 0x0, 0x0}},
    {12530, 2, {0x0, 0x0, 0x0, 0x1F0, 0x1E10, 0x10, 0x20, 0x1E0, 0xE20, 0x40, 0x40, 0x80, 0x100, 0x600, 0x1800, 0x0}},
    {12531, 2, {0x0, 0x0, 0x0, 0x0, 0x1800, 0x408, 0x208, 0x10, 0x10, 0x20, 0x40, 0x80, 0x1300, 0xC00, 0x0, 0x0}},
    {12532, 2, {0x0, 0x8, 0x424, 0x214, 0x210, 0x43E0, 0x3C20, 0x2020, 0x1040, 0x1040, 0x80, 0x80, 0x100, 0x200, 0x400, 0x0}},
    {12533, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x11E0, 0xE20, 0x220, 0x220, 0x420, 0x8C0, 0x1040, 0x0}},
    {12534, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x800, 0x400, 0x460, 0x780, 0x880, 0x1080, 0x100, 0x200, 0x400, 0x0}},
    {12535, 2, {0x4, 0x12, 0x8, 0x21F0, 0x1E10, 0x1010, 0x810, 0x820, 0x820, 0x40, 0x40, 0x80, 0x100, 0x200, 0x400, 0x0}},
    {12536, 2, {0x0, 0x8, 0x104, 0x90, 0x88, 0x1F0, 0x1E80, 0x880, 0x480, 0x5F0, 0x3E88, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {12537, 2, {0x0, 0x8, 0x4, 0x10, 0x1E8, 0x1E20, 0x240, 0x180, 0x100, 0x100, 0x1F0, 0x3E08, 0x0, 0x0, 0x0, 0x0}},
    {12538, 2, {0x4, 0x2, 0x8, 0x1F4, 0x1E10, 0x10, 0x20, 0x1E0, 0xE20, 0x40, 0x40, 0x80, 0x100, 0x600, 0x1800, 0x0}},
    {12539, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x300, 0x780, 0x780, 0x300, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12540, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40F8, 0x3F00, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12541, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC00, 0x300, 0x80, 0x40, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12542, 2, {0x0, 0x0, 0x0, 0x20, 0x90, 0xC40, 0x300, 0x80, 0x40, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12543, 2, {0x0, 0x0, 0x21E0, 0x1E10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12549, 2, {0x400, 0x400, 0x800, 0x11F0, 0x3E10, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x280, 0x100, 0x0, 0x0, 0x0}},
    {12550, 2, {0x100, 0x100, 0x200, 0x7F0, 0x10, 0x10, 0x220, 0x140, 0x80, 0x160, 0x620, 0x3800, 0x0, 0x0, 0x0, 0x0}},
    {12551, 2, {0x18, 0x3FEC, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x1008, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {12552, 2, {0x0, 0x30, 0x1FF8, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0xFFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12553, 2, {0x200, 0x200, 0x400, 0x800, 0x1FF8, 0x2108, 0x108, 0x208, 0x208, 0x408, 0x810, 0x1010, 0x120, 0xC0, 0x0, 0x0}},
    {12554, 2, {0xC0, 0xC0, 0x80, 0x80, 0x43FE, 0x3D00, 0x100, 0x100, 0x220, 0x210, 0x5F0, 0x608, 0x8, 0x0, 0x0, 0x0}},
    {12555, 2, {0x1E0, 0xE20, 0x40, 0x40, 0xF8, 0x18, 0x10, 0x10, 0x20, 0x20, 0x240, 0x1C0, 0x100, 0x0, 0x0, 0x0}},
    {12556, 2, {0x80, 0x80, 0x480, 0x480, 0x8F8, 0xF88, 0x88, 0x108, 0x108, 0x210, 0x210, 0x410, 0x8A0, 0x1040, 0x0, 0x0}},
    {12557, 2, {0x210, 0x31C, 0x218, 0x210, 0x410, 0xC20, 0x840, 0x1080, 0x1840, 0x820, 0x410, 0x210, 0x308, 0x8, 0x0, 0x0}},
    {12558, 2, {0x8, 0x3FFE, 0x300, 0x200, 0x7F8, 0x418, 0x10, 0x10, 0x30, 0x20, 0x160, 0x1C0, 0x80, 0x0, 0x0, 0x0}},
    {12559, 2, {0x1E, 0x7E0, 0x400, 0x400, 0x400, 0x400, 0x800, 0x800, 0x800, 0x1000, 0x1000, 0x2000, 0x0, 0x0, 0x0, 0x0}},
    {12560, 2, {0x40, 0x20, 0x20, 0x20, 0x620, 0x420, 0x420, 0x460, 0x5A0, 0xE20, 0x820, 0x20, 0x20, 0x20, 0x20, 0x0}},
    {12561, 2, {0x10, 0x18, 0x38, 0x60, 0xC0, 0x300, 0xC00, 0x600, 0x180, 0x40, 0x30, 0x18, 0xC, 0x0, 0x0, 0x0}},
    {12562, 2, {0x3FFE, 0x1C80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0, 0x0, 0x0, 0x0}},
    {12563, 2, {0x200, 0x100, 0x100, 0x1110, 0x1110, 0x1110, 0x11F0, 0x1F10, 0x100, 0x100, 0x7FFC, 0x2000, 0x0, 0x0, 0x0, 0x0}},
    {12564, 2, {0x40, 0x30, 0x60, 0x80, 0x118, 0x630, 0x60, 0xC0, 0x340, 0x440, 0x1840, 0x40, 0x40, 0x40, 0x40, 0x0}},
    {12565, 2, {0x1C, 0xFEE, 0xC, 0xC, 0x478, 0x780, 0x400, 0x400, 0x400, 0x400, 0x800, 0x1000, 0x0, 0x0, 0x0, 0x0}},
    {12566, 2, {0xFFC0, 0x8040, 0x8040, 0x8040, 0x8040, 0xA040, 0x9840, 0x8640, 0x8140, 0x8040, 0x8040, 0x8040, 0x8040, 0xFFC0, 0x8040, 0x8040}},
    {12567, 2, {0xFC, 0x1F04, 0x104, 0x104, 0x108, 0x108, 0x138, 0x110, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0, 0x0, 0x0}},
    {12568, 2, {0x180, 0x80, 0x80, 0x100, 0x3FFE, 0x100, 0x100, 0x200, 0x3F0, 0x230, 0x20, 0x40, 0xC0, 0x80, 0x0, 0x0}},
    {12569, 2, {0x100, 0x180, 0x180, 0x100, 0x300, 0x200, 0x400, 0x820, 0x1810, 0x1038, 0x3FCC, 0x2004, 0x0, 0x0, 0x0, 0x0}},
    {12570, 2, {0x2030, 0x1020, 0xC40, 0x480, 0x300, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x0, 0x0, 0x0, 0x0}},
    {12571, 2, {0x0, 0x3FFE, 0x40, 0x40, 0x40, 0xFC0, 0x1800, 0x1000, 0x1000, 0xC18, 0x7E0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12572, 2, {0x100, 0x80, 0x80, 0x80, 0x3FC, 0x3C80, 0x80, 0x780, 0x800, 0x1000, 0x1000, 0x1000, 0x810, 0x7E0, 0x0, 0x0}},
    {12573, 2, {0x0, 0xC30, 0x420, 0x420, 0x4FE, 0xF20, 0x7440, 0x440, 0x440, 0x400, 0x400, 0x3F0, 0x0, 0x0, 0x0, 0x0}},
    {12574, 2, {0x1FFC, 0x880, 0x880, 0x1990, 0x1FF8, 0x118, 0x218, 0x210, 0x410, 0x820, 0x1020, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {12575, 2, {0x0, 0x180, 0x680, 0x3880, 0x80, 0x80, 0x40, 0x40, 0x20, 0x10, 0x1C, 0xC, 0x0, 0x0, 0x0, 0x0}},
    {12576, 2, {0x180, 0x180, 0x110, 0x218, 0x218, 0x430, 0xFE0, 0x38C0, 0x180, 0x220, 0x430, 0x1FC8, 0x1008, 0x0, 0x0, 0x0}},
    {12577, 2, {0x38, 0xFD8, 0x10, 0x420, 0x260, 0x1C0, 0xC0, 0x130, 0x218, 0x40C, 0x1804, 0x2000, 0x0, 0x0, 0x0, 0x0}},
    {12578, 2, {0xFF0, 0x10, 0x420, 0x820, 0x840, 0x1FFC, 0x300C, 0x8, 0x10, 0x30, 0x20, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {12579, 2, {0xC00, 0x800, 0x800, 0x1800, 0x1000, 0x11FC, 0x3E0C, 0x8, 0x8, 0x10, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {12580, 2, {0x80, 0x80, 0x80, 0x100, 0x1F8, 0x1F80, 0x280, 0x280, 0x680, 0x480, 0x880, 0x1080, 0x207E, 0x0, 0x0, 0x0}},
    {12581, 2, {0x100, 0xC0, 0xC0, 0x180, 0x100, 0x300, 0x200, 0x400, 0xC00, 0x800, 0x17FC, 0x7800, 0x0, 0x0, 0x0, 0x0}},
    {12582, 2, {0x40, 0x660, 0x440, 0x440, 0x440, 0x440, 0x440, 0x840, 0x840, 0x840, 0x1040, 0x203E, 0x0, 0x0, 0x0, 0x0}},
    {12583, 2, {0x100, 0x180, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {12584, 2, {0x18, 0x18, 0x410, 0x220, 0x160, 0xC0, 0xC0, 0x120, 0x230, 0x418, 0x80C, 0x100C, 0x0, 0x0, 0x0, 0x0}},
    {12585, 2, {0x1010, 0x1818, 0x808, 0x808, 0x808, 0x808, 0x808, 0x808, 0x1008, 0x1008, 0x1FF8, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {12586, 2, {0x0, 0xFFFC, 0x400, 0x400, 0x400, 0x400, 0x7E0, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2140, 0x4080, 0x0}},
    {12587, 2, {0x0, 0xFFF8, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x1080, 0x1080, 0x1080, 0x2080, 0x2080, 0x4080, 0x8078, 0x0}},
    {12588, 2, {0x100, 0x80, 0x0, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {12593, 2, {0x0, 0x0, 0x0, 0x2008, 0x1FFC, 0x808, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0}},
    {12594, 2, {0x0, 0x0, 0x0, 0x4108, 0x3FFC, 0x1108, 0x108, 0x108, 0x108, 0x108, 0x108, 0x108, 0x0, 0x0, 0x0, 0x0}},
    {12595, 2, {0x0, 0x0, 0x0, 0x8260, 0x7F10, 0x2210, 0x210, 0x230, 0x228, 0x248, 0x244, 0x282, 0x301, 0x0, 0x0, 0x0}},
    {12596, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3000, 0x800, 0x800, 0x800, 0x800, 0x808, 0x7FC, 0x0, 0x0, 0x0, 0x0}},
    {12597, 2, {0x0, 0x0, 0x0, 0x0, 0x204, 0xC1FE, 0x20A0, 0x2020, 0x2050, 0x2388, 0x1D04, 0x202, 0x0, 0x0, 0x0, 0x0}},
    {12598, 2, {0x0, 0x0, 0x0, 0x40, 0x38, 0xC402, 0x23FF, 0x2100, 0x2078, 0x2084, 0x2384, 0x1C84, 0x78, 0x0, 0x0, 0x0}},
    {12599, 2, {0x0, 0x0, 0x0, 0x0, 0x7010, 0x1FF8, 0x1000, 0x1000, 0x1000, 0x1000, 0x1008, 0xFFC, 0x0, 0x0, 0x0, 0x0}},
    {12600, 2, {0x0, 0x0, 0x0, 0x0, 0x6288, 0x3FFC, 0x2080, 0x2080, 0x2080, 0x2080, 0x2384, 0x1C7E, 0x0, 0x0, 0x0, 0x0}},
    {12601, 2, {0x0, 0x0, 0x0, 0x2010, 0x1FF8, 0x810, 0x10, 0x3010, 0x1FF8, 0x1000, 0x1000, 0x1010, 0xFF8, 0x0, 0x0, 0x0}},
    {12602, 2, {0x0, 0x0, 0x0, 0x8204, 0x7FFE, 0x2284, 0x204, 0xFF04, 0x4004, 0x4004, 0x4304, 0x3C04, 0x0, 0x0, 0x0, 0x0}},
    {12603, 2, {0x0, 0x0, 0x0, 0x4144, 0x3FFE, 0x1144, 0x144, 0x7FC4, 0x2044, 0x2044, 0x23C4, 0x1C7E, 0x40, 0x0, 0x0, 0x0}},
    {12604, 2, {0x0, 0x0, 0x0, 0x4208, 0x3F04, 0x284, 0x284, 0x7EFC, 0x2084, 0x2084, 0x2384, 0x1CFC, 0x0, 0x0, 0x0, 0x0}},
    {12605, 2, {0x0, 0x0, 0x0, 0x8460, 0x7E10, 0x2410, 0x410, 0xFE30, 0x4028, 0x4048, 0x4684, 0x3902, 0x0, 0x0, 0x0, 0x0}},
    {12606, 2, {0x0, 0x0, 0x0, 0x42C4, 0x3F7E, 0x240, 0x244, 0x7F7E, 0x2040, 0x2040, 0x2344, 0x1C3E, 0x0, 0x0, 0x0, 0x0}},
    {12607, 2, {0x0, 0x0, 0x0, 0x8508, 0x7EFC, 0x400, 0x400, 0xFE84, 0x4048, 0x4048, 0x464A, 0x3BFF, 0x0, 0x0, 0x0, 0x0}},
    {12608, 2, {0x0, 0x0, 0x0, 0x8440, 0x7E38, 0x502, 0x4FF, 0xFE00, 0x4038, 0x4044, 0x4644, 0x3838, 0x0, 0x0, 0x0, 0x0}},
    {12609, 2, {0x0, 0x0, 0x0, 0x3810, 0xFF8, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0xFF8, 0x800, 0x0, 0x0, 0x0}},
    {12610, 2, {0x0, 0x0, 0x20, 0x10, 0x1010, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x0, 0x0, 0x0}},
    {12611, 2, {0x0, 0x0, 0x8, 0x284, 0x2144, 0x1144, 0x1144, 0x1F7C, 0x1144, 0x1144, 0x1144, 0x1F7C, 0x1144, 0x0, 0x0, 0x0}},
    {12612, 2, {0x0, 0x0, 0x0, 0x460, 0x4210, 0x2210, 0x2210, 0x3E10, 0x2228, 0x2228, 0x2244, 0x3E44, 0x2282, 0x0, 0x0, 0x0}},
    {12613, 2, {0x0, 0x0, 0x200, 0x180, 0x80, 0x80, 0x80, 0x140, 0x140, 0x220, 0x410, 0x808, 0x3004, 0x0, 0x0, 0x0}},
    {12614, 2, {0x0, 0x0, 0x0, 0x840, 0x420, 0x420, 0x420, 0xA20, 0xA50, 0x1148, 0x2084, 0xC302, 0x0, 0x0, 0x0, 0x0}},
    {12615, 2, {0x0, 0x0, 0x0, 0x200, 0x100, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0, 0x0}},
    {12616, 2, {0x0, 0x0, 0x0, 0x2010, 0x1FF8, 0x900, 0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x6008, 0x0, 0x0, 0x0}},
    {12617, 2, {0x0, 0x0, 0x0, 0x0, 0x8104, 0x7FFE, 0x2420, 0x420, 0xA50, 0x950, 0x1088, 0x2104, 0xC602, 0x0, 0x0, 0x0}},
    {12618, 2, {0x0, 0x0, 0x400, 0x3E0, 0x0, 0x2008, 0x1FFC, 0x880, 0x140, 0x220, 0x410, 0x808, 0x3004, 0x0, 0x0, 0x0}},
    {12619, 2, {0x0, 0x0, 0x0, 0x1008, 0xFFC, 0x408, 0x8, 0xF8, 0x3F08, 0x1008, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0}},
    {12620, 2, {0x0, 0x0, 0x0, 0x3820, 0xFF0, 0x800, 0x820, 0xFF0, 0x800, 0x800, 0x810, 0x7F8, 0x0, 0x0, 0x0, 0x0}},
    {12621, 2, {0x0, 0x0, 0x0, 0x2010, 0x1FF8, 0x800, 0x0, 0x810, 0x420, 0x220, 0x4248, 0x3FFC, 0x1000, 0x0, 0x0, 0x0}},
    {12622, 2, {0x0, 0x0, 0x0, 0x400, 0x3C0, 0x4008, 0x3FFC, 0x1000, 0x3C0, 0x420, 0x810, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {12623, 2, {0x40, 0x30, 0x10, 0x10, 0x10, 0x12, 0x1F, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12624, 2, {0x20, 0x118, 0xC8, 0x48, 0x48, 0x48, 0x78, 0x48, 0x48, 0x48, 0x48, 0x48, 0x48, 0x48, 0x8, 0x0}},
    {12625, 2, {0x40, 0x30, 0x10, 0x10, 0x12, 0x1F, 0x10, 0x12, 0x1F, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12626, 2, {0x20, 0x118, 0xC8, 0x48, 0x48, 0x78, 0x48, 0x48, 0x78, 0x48, 0x48, 0x48, 0x48, 0x48, 0x8, 0x0}},
    {12627, 2, {0x40, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x3F0, 0x110, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12628, 2, {0x10, 0x8C, 0x64, 0x24, 0x24, 0x24, 0x3E4, 0x124, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x4, 0x0}},
    {12629, 2, {0x40, 0x30, 0x10, 0x10, 0x10, 0x3F0, 0x110, 0x10, 0x3F0, 0x110, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12630, 2, {0x10, 0x8C, 0x64, 0x24, 0x24, 0x3E4, 0x124, 0x24, 0x3E4, 0x124, 0x24, 0x24, 0x24, 0x24, 0x4, 0x0}},
    {12631, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x80, 0x80, 0x80, 0x4082, 0x3FFF, 0x1000, 0x0, 0x0, 0x0}},
    {12632, 2, {0x40, 0x30, 0x10, 0x10, 0x10, 0x10, 0x810, 0x412, 0x41F, 0x5F0, 0x7E10, 0x2010, 0x10, 0x10, 0x10, 0x0}},
    {12633, 2, {0x10, 0x8C, 0x64, 0x24, 0x24, 0x824, 0x424, 0x43C, 0x424, 0x7E4, 0x7C24, 0x2024, 0x24, 0x24, 0x4, 0x0}},
    {12634, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x408, 0x208, 0x208, 0x208, 0x3F8, 0x7E08, 0x2008, 0x8, 0x8, 0x8, 0x0}},
    {12635, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x420, 0x220, 0x220, 0x4222, 0x3FFF, 0x1000, 0x0, 0x0, 0x0}},
    {12636, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4002, 0x3FFF, 0x1080, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {12637, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x8, 0x1E8, 0x7E08, 0x2808, 0x9F8, 0x888, 0x808, 0x808, 0x808, 0x8, 0x0}},
    {12638, 2, {0x20, 0x98, 0x48, 0x48, 0x48, 0x48, 0x48, 0x3C8, 0x7C48, 0x2848, 0xBC8, 0x848, 0x848, 0x848, 0x8, 0x0}},
    {12639, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1E8, 0x7E08, 0x2408, 0x408, 0x408, 0x408, 0x408, 0x8, 0x0}},
    {12640, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4002, 0x3FFF, 0x1220, 0x220, 0x220, 0x220, 0x220, 0x220, 0x220, 0x0}},
    {12641, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4002, 0x3FFF, 0x1000, 0x0, 0x0, 0x0, 0x0}},
    {12642, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1F8, 0x7E08, 0x2008, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {12643, 2, {0x40, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12644, 2, {0xAAAA, 0x1, 0x8000, 0x1, 0x8000, 0x1, 0x913C, 0x1121, 0x9F38, 0x1121, 0x9120, 0x1, 0x8000, 0x1, 0x8000, 0x5555}},
    {12645, 2, {0x0, 0x0, 0x0, 0xC180, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x2344, 0x1C3E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12646, 2, {0x0, 0x0, 0x0, 0xC0C4, 0x207E, 0x2040, 0x2040, 0x2040, 0x2040, 0x2342, 0x1C3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12647, 2, {0x0, 0x0, 0x0, 0xC020, 0x2010, 0x2010, 0x2010, 0x2010, 0x2028, 0x2348, 0x1C84, 0x102, 0x0, 0x0, 0x0, 0x0}},
    {12648, 2, {0x0, 0x0, 0x0, 0xC010, 0x2010, 0x2028, 0x2028, 0x2044, 0x2044, 0x2382, 0x1CFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12649, 2, {0x0, 0x0, 0x0, 0x8850, 0x7FE8, 0x848, 0x848, 0xFC54, 0x4054, 0x4062, 0x3C41, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12650, 2, {0x0, 0x0, 0x0, 0x42C8, 0x3F7C, 0x240, 0x240, 0x7F40, 0x2040, 0x2040, 0x2344, 0x1C3E, 0x0, 0x0, 0x0, 0x0}},
    {12651, 2, {0x0, 0x0, 0x0, 0x8890, 0x7C48, 0xA48, 0xA48, 0xFFD4, 0x4254, 0x4E62, 0x33C1, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12652, 2, {0x0, 0x0, 0x8420, 0x7E20, 0x420, 0x450, 0xFC50, 0x4088, 0x4088, 0x4704, 0x3BFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12653, 2, {0x0, 0x0, 0x0, 0x8400, 0x7E00, 0x504, 0x4FE, 0xFE40, 0x4038, 0x4044, 0x4644, 0x3838, 0x0, 0x0, 0x0, 0x0}},
    {12654, 2, {0x0, 0x0, 0x0, 0x8, 0x6284, 0x3F44, 0x2244, 0x227C, 0x2244, 0x2244, 0x2244, 0x3F7C, 0x2044, 0x0, 0x0, 0x0}},
    {12655, 2, {0x0, 0x0, 0x0, 0x6220, 0x3F10, 0x2210, 0x2210, 0x2210, 0x2228, 0x2228, 0x3E44, 0x2282, 0x0, 0x0, 0x0, 0x0}},
    {12656, 2, {0x0, 0x0, 0x0, 0x6210, 0x3F10, 0x2228, 0x2228, 0x2244, 0x2244, 0x2244, 0x2282, 0x3FFF, 0x2000, 0x0, 0x0, 0x0}},
    {12657, 2, {0x0, 0x3008, 0x1FFC, 0x1008, 0x1008, 0x1008, 0x1FFC, 0x1100, 0x80, 0x7E0, 0x810, 0x1008, 0x1008, 0x810, 0x7E0, 0x0}},
    {12658, 2, {0x0, 0x0, 0x0, 0x400, 0x282, 0x827F, 0x4222, 0x7E02, 0x4202, 0x4202, 0x4202, 0x7E02, 0x4202, 0x0, 0x0, 0x0}},
    {12659, 2, {0x0, 0x0, 0x400, 0x2C8, 0x827C, 0x4240, 0x4240, 0x7E40, 0x4240, 0x4240, 0x4244, 0x7E3E, 0x4200, 0x0, 0x0, 0x0}},
    {12660, 2, {0x0, 0x0, 0x0, 0x1142, 0x88BF, 0x4882, 0x4882, 0x7982, 0x4942, 0x4A42, 0x7A22, 0x4C12, 0x0, 0x0, 0x0, 0x0}},
    {12661, 2, {0x0, 0x0, 0x0, 0x1132, 0x889F, 0x4890, 0x4890, 0x7990, 0x4950, 0x4A50, 0x7A52, 0x4C2F, 0x0, 0x0, 0x0, 0x0}},
    {12662, 2, {0x0, 0x0, 0x0, 0x800, 0x8604, 0x45FE, 0x44A0, 0x7C20, 0x4420, 0x4450, 0x7C88, 0x4706, 0x0, 0x0, 0x0, 0x0}},
    {12663, 2, {0x0, 0x0, 0x400, 0x2C4, 0x827E, 0x4240, 0x4244, 0x7E7E, 0x4240, 0x4240, 0x4244, 0x7E3E, 0x4200, 0x0, 0x0, 0x0}},
    {12664, 2, {0x10, 0x2008, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1108, 0x80, 0x7E0, 0x810, 0x1008, 0x1008, 0x810, 0x7E0, 0x0}},
    {12665, 2, {0x10, 0x2488, 0x1248, 0x1E78, 0x1248, 0x1248, 0x1E78, 0x1248, 0x80, 0x7E0, 0x810, 0x1008, 0x1008, 0x810, 0x7E0, 0x0}},
    {12666, 2, {0x0, 0x0, 0x0, 0x904, 0x4FE, 0x444, 0x404, 0xA04, 0xA04, 0x1104, 0x2104, 0xC084, 0x0, 0x0, 0x0, 0x0}},
    {12667, 2, {0x0, 0x0, 0x0, 0x800, 0x580, 0x440, 0x640, 0xA40, 0xA40, 0x1140, 0x2144, 0x40BE, 0x0, 0x0, 0x0, 0x0}},
    {12668, 2, {0x0, 0x0, 0x0, 0x10C4, 0x87E, 0x840, 0x840, 0x1440, 0x1440, 0x2240, 0x4144, 0x80BE, 0x0, 0x0, 0x0, 0x0}},
    {12669, 2, {0x0, 0x0, 0x4, 0x1002, 0x882, 0x842, 0x87E, 0x1442, 0x1442, 0x2242, 0x217E, 0x40C2, 0x0, 0x0, 0x0, 0x0}},
    {12670, 2, {0x0, 0x0, 0x0, 0x1202, 0x9FF, 0x890, 0x810, 0x1428, 0x1428, 0x2244, 0x4142, 0x8181, 0x0, 0x0, 0x0, 0x0}},
    {12671, 2, {0x0, 0x0, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x820, 0x1010, 0x3FF8, 0x0, 0x0, 0x0, 0x0}},
    {12672, 2, {0x0, 0x0, 0x1818, 0x808, 0x3C3C, 0x4242, 0x4242, 0x4242, 0x4242, 0x4242, 0x4242, 0x3C3C, 0x0, 0x0, 0x0, 0x0}},
    {12673, 2, {0x0, 0x300, 0x100, 0xFE0, 0x1010, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x1010, 0xFE0, 0x0, 0x0, 0x0}},
    {12674, 2, {0x0, 0x3000, 0x1204, 0x79FE, 0x84A0, 0x8420, 0x8420, 0x8450, 0x8450, 0x8488, 0x8504, 0x7A02, 0x0, 0x0, 0x0, 0x0}},
    {12675, 2, {0x0, 0x3000, 0x1020, 0x7820, 0x8450, 0x8450, 0x8488, 0x8488, 0x8488, 0x8504, 0x8504, 0x7BFE, 0x0, 0x0, 0x0, 0x0}},
    {12676, 2, {0x0, 0x2010, 0x1FF8, 0x0, 0x420, 0x4248, 0x3FFC, 0x1100, 0x80, 0x7E0, 0x810, 0x1008, 0x1008, 0x810, 0x7E0, 0x0}},
    {12677, 2, {0x0, 0x0, 0x2040, 0x1C38, 0x8102, 0x7FFF, 0x2000, 0x1C38, 0x2244, 0x2244, 0x2244, 0x1C38, 0x0, 0x0, 0x0, 0x0}},
    {12678, 2, {0x0, 0x0, 0x4008, 0x3FFC, 0x1200, 0x100, 0x7C0, 0x820, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x0, 0x0}},
    {12679, 2, {0x40, 0x30, 0x10, 0x10, 0x10, 0x912, 0x91F, 0x910, 0x9F2, 0x7E1F, 0x2010, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {12680, 2, {0x20, 0x118, 0xC8, 0x48, 0x48, 0x478, 0x248, 0x2248, 0x1278, 0x13C8, 0x7C48, 0x2048, 0x48, 0x48, 0x8, 0x0}},
    {12681, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x208, 0x1108, 0x908, 0x908, 0xBE8, 0x7C08, 0x2008, 0x8, 0x8, 0x8, 0x0}},
    {12682, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x8, 0x1E8, 0x7F08, 0x2978, 0x908, 0x978, 0x908, 0x908, 0x908, 0x8, 0x0}},
    {12683, 2, {0x20, 0x118, 0xC8, 0x48, 0x48, 0x48, 0x348, 0xFC48, 0x55C8, 0x1448, 0x15C8, 0x1448, 0x1448, 0x48, 0x8, 0x0}},
    {12684, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x1E8, 0x7F08, 0x2908, 0x908, 0x908, 0x908, 0x908, 0x8, 0x0}},
    {12685, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x380, 0x100, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12686, 2, {0x20, 0x18, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x208, 0x708, 0x208, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {12688, 2, {0x0, 0x0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12689, 2, {0x0, 0x0, 0x200, 0x200, 0x200, 0x220, 0x240, 0x280, 0x300, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12690, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3FC0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12691, 2, {0x0, 0x0, 0x0, 0x1F80, 0x0, 0x0, 0x0, 0x3FC0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12692, 2, {0x0, 0x0, 0x1FE0, 0x0, 0xFC0, 0x0, 0x0, 0x3FF0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12693, 2, {0x0, 0x0, 0x3FE0, 0x2520, 0x2520, 0x28E0, 0x2020, 0x3FE0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12694, 2, {0x0, 0x0, 0x400, 0x400, 0x780, 0x400, 0x400, 0x400, 0x3FC0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12695, 2, {0x0, 0x400, 0x400, 0x3FC0, 0x2440, 0x2440, 0x3FC0, 0x400, 0x400, 0x400, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12696, 2, {0x0, 0x0, 0x3FC0, 0x400, 0x600, 0x500, 0x400, 0x400, 0x400, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12697, 2, {0x0, 0x0, 0x3FC0, 0x2240, 0x3FC0, 0x2240, 0x3FC0, 0x2240, 0x200, 0x200, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12698, 2, {0x0, 0x0, 0x3F80, 0x300, 0x400, 0x800, 0x1040, 0x2040, 0x1FC0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12699, 2, {0x0, 0x0, 0x3FE0, 0x200, 0x3FE0, 0x2220, 0x2320, 0x24A0, 0x2860, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12700, 2, {0x0, 0x0, 0x3FE0, 0x200, 0x200, 0x200, 0x200, 0x200, 0xA00, 0x400, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12701, 2, {0x0, 0x0, 0x1FC0, 0x200, 0x200, 0x3FE0, 0x200, 0x500, 0x8E0, 0x1040, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12702, 2, {0x0, 0x840, 0x950, 0x3DF0, 0xB50, 0x950, 0x928, 0xD08, 0x30F8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12703, 2, {0x0, 0x400, 0x400, 0x400, 0x400, 0xA00, 0x900, 0x1080, 0x2040, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12704, 2, {0x0, 0x1000, 0x1000, 0x2000, 0x7FF0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1A0, 0x260, 0x240, 0x180}},
    {12705, 2, {0x0, 0x0, 0x3FF0, 0x410, 0x410, 0x410, 0x410, 0x410, 0x420, 0x520, 0x4C0, 0x400, 0x3400, 0x4C00, 0x4C00, 0x3400}},
    {12706, 2, {0x0, 0x40, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x10C0, 0x1340, 0x1C40, 0x40, 0x340, 0x4C0, 0x4C0, 0x340}},
    {12707, 2, {0x0, 0x0, 0x880, 0x880, 0x1100, 0x1100, 0x2200, 0x2200, 0x4400, 0x2200, 0x2200, 0x1130, 0x1148, 0x8C8, 0x8B0, 0x0}},
    {12708, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0xFFF8, 0x1008, 0x1010, 0x1010, 0x1000, 0x1000, 0x1000, 0x1000, 0xFF0, 0x0, 0x0}},
    {12709, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0xFFF8, 0x1008, 0x1010, 0x1010, 0x1000, 0x1030, 0x1048, 0x1048, 0xFF0, 0x10, 0x0}},
    {12710, 2, {0x0, 0x0, 0xFFF8, 0x200, 0x200, 0x200, 0x1E00, 0x2000, 0x4000, 0x4000, 0x2020, 0x1040, 0x880, 0x500, 0x200, 0x0}},
    {12711, 2, {0x0, 0x0, 0xFFF8, 0x200, 0x200, 0x200, 0x1E00, 0x2000, 0x4060, 0x4090, 0x4090, 0x4050, 0x2060, 0x1F90, 0x0, 0x0}},
    {12712, 2, {0x0, 0x0, 0x2020, 0x1040, 0x1040, 0x880, 0x880, 0x500, 0x500, 0x3F8, 0x500, 0x880, 0x3060, 0xC018, 0x0, 0x0}},
    {12713, 2, {0x0, 0x2010, 0x1020, 0x840, 0x480, 0x100, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x1A00, 0x2600, 0x2600, 0x1A00}},
    {12714, 2, {0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0xFFF8, 0x10, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {12715, 2, {0x0, 0x0, 0x2020, 0x1040, 0x1040, 0x880, 0x880, 0x500, 0x500, 0x200, 0x518, 0x8A4, 0x3064, 0xC018, 0x8, 0x0}},
    {12716, 2, {0x0, 0x0, 0x0, 0xFFF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x10, 0x0, 0x0}},
    {12717, 2, {0x0, 0x0, 0xFFF8, 0x1240, 0x1240, 0x1240, 0x1240, 0x1240, 0x1240, 0x1240, 0x1240, 0x1240, 0x223C, 0xC200, 0x0, 0x0}},
    {12718, 2, {0x0, 0x0, 0xFFF8, 0x1200, 0x1200, 0x2200, 0x3FF0, 0x210, 0x210, 0x210, 0x410, 0x6410, 0x9820, 0x9020, 0x6840, 0x0}},
    {12719, 2, {0x0, 0x400, 0x400, 0x840, 0x840, 0x1080, 0x1080, 0x3F00, 0x200, 0x460, 0x890, 0x1090, 0x3FE0, 0x20, 0x10, 0x0}},
    {12720, 2, {0x0, 0x3060, 0x880, 0x500, 0xFFF8, 0x8208, 0x8208, 0x8208, 0x8208, 0x8208, 0x8208, 0x8208, 0x8208, 0x8208, 0x8228, 0x10}},
    {12721, 2, {0x0, 0xFFF8, 0x200, 0xFFF8, 0x8208, 0x8E08, 0x9008, 0xA008, 0xA008, 0xA028, 0x9048, 0x8F88, 0x8008, 0x8028, 0x10, 0x0}},
    {12722, 2, {0x0, 0x0, 0xFFF8, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0xFFF8, 0x0}},
    {12723, 2, {0x0, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x1A00, 0x2600, 0x2600, 0x1A00, 0x0}},
    {12724, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1000, 0x2000, 0x7E00, 0x200, 0x200, 0x200, 0x1400, 0x800, 0x0, 0x0}},
    {12725, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1000, 0x2000, 0x7E00, 0x1200, 0x2200, 0x4200, 0x9400, 0x800, 0x0, 0x0}},
    {12726, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFC00, 0x2000, 0x4000, 0xF800, 0x800, 0x800, 0x2800, 0x1000, 0x0, 0x0}},
    {12727, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E00, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x8000, 0x0, 0x0}},
    {12784, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4000, 0x2E00, 0x3200, 0x2400, 0x4400, 0x800, 0x1000, 0x2000, 0x4000, 0x0}},
    {12785, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6000, 0x1000, 0x0, 0x6100, 0x1200, 0x400, 0x4800, 0x3000, 0x0}},
    {12786, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8E00, 0x7200, 0x400, 0x400, 0x800, 0x1400, 0x2200, 0xC200, 0x0}},
    {12787, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x3000, 0x2800, 0x2400, 0x2000, 0x2000, 0x2000, 0x0}},
    {12788, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8E00, 0x7200, 0x400, 0x3400, 0x800, 0x1400, 0x2200, 0xC200, 0x0}},
    {12789, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x480, 0x240, 0x220, 0x420, 0x810, 0x1010, 0x0, 0x0, 0x0}},
    {12790, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4000, 0x4000, 0x4000, 0x4E00, 0x7000, 0x4000, 0x4000, 0x4000, 0x3E00, 0x0}},
    {12791, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8E00, 0x7200, 0x400, 0x400, 0x800, 0x1000, 0x2000, 0xC000, 0x0}},
    {12792, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x500, 0x8C0, 0x30, 0x0, 0x0, 0x0}},
    {12793, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2000, 0x1000, 0x1E00, 0xF000, 0x1000, 0x5400, 0x9200, 0x1000, 0x5000, 0x3000, 0x0}},
    {12794, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x100, 0x100, 0x100, 0x200, 0x220, 0x14D0, 0xF08, 0x0}},
    {12795, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8C00, 0x7000, 0x0, 0x8E00, 0x7200, 0x200, 0x400, 0x800, 0x7000, 0x0}},
    {12796, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8800, 0x4400, 0x4400, 0x4400, 0x400, 0x400, 0x800, 0x3000, 0x0}},
    {12797, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x240, 0x120, 0x120, 0x120, 0x120, 0x124, 0x224, 0x428, 0x830, 0x0}},
    {12798, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8000, 0x4000, 0x4000, 0x4000, 0x4400, 0x4800, 0x5000, 0x6000, 0x0}},
    {12799, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF0, 0x1F10, 0x1010, 0x1010, 0x1020, 0x1020, 0x11F0, 0x1E00, 0x0, 0x0}},
    {12800, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4FF2, 0x4012, 0x4012, 0x4012, 0x4012, 0x4012, 0x4012, 0x4012, 0x4002, 0x2004, 0x1818, 0x0}},
    {12801, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4402, 0x4402, 0x4402, 0x4402, 0x4402, 0x4402, 0x47FA, 0x4002, 0x4002, 0x2004, 0x1818, 0x0}},
    {12802, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4002, 0x4FF2, 0x4802, 0x4802, 0x4802, 0x4802, 0x4802, 0x4FF2, 0x4002, 0x2004, 0x1818, 0x0}},
    {12803, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4FF2, 0x4012, 0x4012, 0x4FF2, 0x4802, 0x4802, 0x4802, 0x4FFA, 0x4002, 0x2004, 0x1818, 0x0}},
    {12804, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4002, 0x4FF2, 0x4812, 0x4812, 0x4812, 0x4812, 0x4FF2, 0x4002, 0x4002, 0x2004, 0x1818, 0x0}},
    {12805, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4812, 0x4812, 0x4812, 0x4FF2, 0x4812, 0x4812, 0x4812, 0x4FF2, 0x4002, 0x2004, 0x1818, 0x0}},
    {12806, 2, {0x0, 0x1818, 0x2004, 0x4082, 0x4082, 0x4082, 0x4142, 0x4142, 0x4222, 0x4412, 0x480A, 0x4002, 0x4002, 0x2004, 0x1818, 0x0}},
    {12807, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x43C2, 0x4422, 0x4812, 0x4812, 0x4812, 0x4812, 0x4422, 0x43C2, 0x4002, 0x2004, 0x1818, 0x0}},
    {12808, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x5FF2, 0x4102, 0x4102, 0x4282, 0x4282, 0x4442, 0x4822, 0x5012, 0x4002, 0x2004, 0x1818, 0x0}},
    {12809, 2, {0x0, 0x1818, 0x2004, 0x43E2, 0x4002, 0x4002, 0x4FFA, 0x4082, 0x4142, 0x4222, 0x4412, 0x480A, 0x4002, 0x2004, 0x1818, 0x0}},
    {12810, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4FF2, 0x4012, 0x4012, 0x5FF2, 0x4012, 0x4012, 0x4012, 0x4012, 0x4012, 0x2004, 0x1818, 0x0}},
    {12811, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4FF2, 0x4802, 0x4802, 0x4FF2, 0x4802, 0x4802, 0x4802, 0x4FF2, 0x4002, 0x2004, 0x1818, 0x0}},
    {12812, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x4002, 0x4FF2, 0x4002, 0x4422, 0x4422, 0x4242, 0x4242, 0x5FFA, 0x4002, 0x2004, 0x1818, 0x0}},
    {12813, 2, {0x0, 0x1818, 0x2004, 0x4002, 0x43C2, 0x4002, 0x5FFA, 0x4002, 0x43C2, 0x4422, 0x4422, 0x4422, 0x43C2, 0x2004, 0x1818, 0x0}},
    {12814, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FA2, 0x40A2, 0x413E, 0x4222, 0x4422, 0x4822, 0x5022, 0x6022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12815, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5022, 0x5022, 0x5022, 0x503E, 0x5022, 0x5FE2, 0x4022, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12816, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FA2, 0x5022, 0x5022, 0x503E, 0x5022, 0x5FA2, 0x4022, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12817, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FA2, 0x40A2, 0x40A2, 0x5FBE, 0x5022, 0x5022, 0x5FA2, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12818, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x4022, 0x5F22, 0x5122, 0x513E, 0x5122, 0x5F22, 0x4022, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12819, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5122, 0x5122, 0x5F22, 0x513E, 0x5122, 0x5F22, 0x4022, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12820, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x4422, 0x4422, 0x4A22, 0x4A3E, 0x5122, 0x60A2, 0x4022, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12821, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x4E22, 0x5122, 0x5122, 0x513E, 0x5122, 0x5122, 0x4E22, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12822, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FE2, 0x4222, 0x4222, 0x453E, 0x4522, 0x48A2, 0x5062, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12823, 2, {0x0, 0x1818, 0x2024, 0x4722, 0x4022, 0x5FE2, 0x4222, 0x423E, 0x4522, 0x48A2, 0x5062, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12824, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FA2, 0x40A2, 0x7F22, 0x423E, 0x4422, 0x4822, 0x5022, 0x6022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12825, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FA2, 0x5022, 0x5FA2, 0x503E, 0x5022, 0x5FA2, 0x4022, 0x6022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12826, 2, {0x0, 0x1818, 0x2024, 0x4022, 0x5FA2, 0x4022, 0x50A2, 0x493E, 0x4922, 0x4BE2, 0x7C22, 0x4022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12827, 2, {0x0, 0x1818, 0x2024, 0x4F22, 0x4022, 0x7FE2, 0x4022, 0x4F3E, 0x50A2, 0x50A2, 0x4F22, 0x6022, 0x4022, 0x2024, 0x1818, 0x0}},
    {12828, 2, {0x0, 0x3018, 0x4004, 0x9FF2, 0x8102, 0x8282, 0x8C62, 0xB01A, 0x8002, 0xBFFA, 0x8102, 0x8102, 0x8102, 0x4104, 0x3118, 0x0}},
    {12832, 2, {0x0, 0x2008, 0x4004, 0x8002, 0x8002, 0x8002, 0x8012, 0xBFFA, 0x8002, 0x8002, 0x8002, 0x8002, 0x8002, 0x4004, 0x2008, 0x0}},
    {12833, 2, {0x0, 0x2008, 0x4004, 0x8022, 0x9FF2, 0x8002, 0x8002, 0x8002, 0x8002, 0x8012, 0xBFFA, 0x8002, 0x8002, 0x4004, 0x2008, 0x0}},
    {12834, 2, {0x0, 0x2008, 0x4024, 0x9FF2, 0x8002, 0x8002, 0x8042, 0x8FE2, 0x8002, 0x8002, 0x8012, 0xBFFA, 0x8002, 0x4004, 0x2008, 0x0}},
    {12835, 2, {0x0, 0x2008, 0x4004, 0x8002, 0xBFFA, 0xA44A, 0xA44A, 0xA44A, 0xA44A, 0xA44A, 0xBFFA, 0x8002, 0x8002, 0x4004, 0x2008, 0x0}},
    {12836, 2, {0x0, 0x2008, 0x4024, 0x9FF2, 0x8202, 0x8202, 0x8222, 0x9FF2, 0x8422, 0x8422, 0x8422, 0x8422, 0xBFFA, 0x4004, 0x2008, 0x0}},
    {12837, 2, {0x0, 0x2008, 0x4004, 0x8202, 0x8102, 0x8012, 0xBFFA, 0x8002, 0x8002, 0x8442, 0x8422, 0x8812, 0x9012, 0x4004, 0x2008, 0x0}},
    {12838, 2, {0x0, 0x2008, 0x4204, 0x8202, 0x8202, 0x8212, 0x827A, 0x8382, 0xBE02, 0x8202, 0x8202, 0x8222, 0x8222, 0x41E4, 0x2008, 0x0}},
    {12839, 2, {0x0, 0x2008, 0x4084, 0x8482, 0x8482, 0x8482, 0x8482, 0x8482, 0x8842, 0x8822, 0x901A, 0xA012, 0x8002, 0x4004, 0x2008, 0x0}},
    {12840, 2, {0x0, 0x2008, 0x4204, 0x8202, 0x8242, 0x9FE2, 0x8242, 0x8242, 0x8242, 0x8442, 0x844A, 0x884A, 0x903A, 0x4004, 0x2008, 0x0}},
    {12841, 2, {0x0, 0x2008, 0x4104, 0x8102, 0x8102, 0x8112, 0xBFFA, 0x8102, 0x8102, 0x8102, 0x8102, 0x8102, 0x8102, 0x4004, 0x2008, 0x0}},
    {12842, 2, {0x0, 0x3018, 0x47F4, 0x8412, 0x8412, 0x87F2, 0x8412, 0x8412, 0x87F2, 0x8412, 0x8412, 0x8852, 0x9022, 0x4004, 0x3018, 0x0}},
    {12843, 2, {0x0, 0x3018, 0x4104, 0x8102, 0x8912, 0x8922, 0x8942, 0x9102, 0x8282, 0x8282, 0x8442, 0x8822, 0xB01A, 0x4004, 0x3018, 0x0}},
    {12844, 2, {0x0, 0x3018, 0x4104, 0x8102, 0x8112, 0xBDA2, 0x8542, 0x8542, 0x8922, 0x8922, 0x9112, 0xA50A, 0x8202, 0x4004, 0x3018, 0x0}},
    {12845, 2, {0x0, 0x3018, 0x4104, 0x8102, 0x8102, 0xBFFA, 0x8102, 0x8382, 0x8542, 0x8922, 0xB11A, 0x8112, 0x8102, 0x4004, 0x3018, 0x0}},
    {12846, 2, {0x0, 0x3018, 0x4104, 0x8282, 0x8442, 0x8822, 0xB7DA, 0x8102, 0x9FF2, 0x8102, 0x8922, 0x8542, 0xBFFA, 0x4004, 0x3018, 0x0}},
    {12847, 2, {0x0, 0x3018, 0x4104, 0x8102, 0x8102, 0x8102, 0x9FF2, 0x8102, 0x8102, 0x8102, 0x8102, 0x8102, 0xBFFA, 0x4004, 0x3018, 0x0}},
    {12848, 2, {0x0, 0x3018, 0x4004, 0x8FE2, 0x8822, 0x8822, 0x8822, 0x8FE2, 0x8822, 0x8822, 0x8822, 0x8FE2, 0x8822, 0x4004, 0x3018, 0x0}},
    {12849, 2, {0x0, 0x3018, 0x4844, 0x8942, 0xBDF2, 0x8A42, 0x9C42, 0x9BFA, 0xA842, 0xA8E2, 0x8952, 0x8A4A, 0x8842, 0x4004, 0x3018, 0x0}},
    {12850, 2, {0x0, 0x3018, 0x4204, 0xBFFA, 0x8402, 0x87E2, 0x8C22, 0x97E2, 0xA422, 0x87E2, 0x8422, 0x8422, 0x8462, 0x4004, 0x3018, 0x0}},
    {12851, 2, {0x0, 0x3018, 0x5044, 0x8842, 0xBE42, 0x8442, 0x89F2, 0x8C42, 0x9A42, 0xA842, 0x8842, 0x8842, 0x8BFA, 0x4004, 0x3018, 0x0}},
    {12852, 2, {0x0, 0x3018, 0x4404, 0x87F2, 0x8422, 0x8A42, 0x9182, 0x8102, 0x87F2, 0xBC12, 0x8412, 0x8412, 0x87F2, 0x4004, 0x3018, 0x0}},
    {12853, 2, {0x0, 0x3018, 0x4844, 0xA9F2, 0xBC42, 0xA842, 0xABFA, 0x8C22, 0x9BFA, 0xA922, 0x88A2, 0x8822, 0x88E2, 0x4004, 0x3018, 0x0}},
    {12854, 2, {0x0, 0x3018, 0x4014, 0x9E12, 0x93FA, 0x9E12, 0x9232, 0x9E52, 0x9292, 0x9F12, 0x8012, 0x9252, 0xA122, 0x4004, 0x3018, 0x0}},
    {12855, 2, {0x0, 0x3018, 0x51F4, 0x8912, 0xBF12, 0x8312, 0x85F2, 0x8CA2, 0x9AA2, 0xAAA2, 0x88AA, 0x892A, 0x8A1A, 0x4004, 0x3018, 0x0}},
    {12856, 2, {0x0, 0x3018, 0x5214, 0x8912, 0x8922, 0x9FFA, 0x910A, 0xA102, 0x9FF2, 0x8112, 0x8212, 0x8412, 0x9862, 0x4004, 0x3018, 0x0}},
    {12857, 2, {0x0, 0x3018, 0x44A4, 0x8492, 0x8882, 0x88FA, 0x9B82, 0xA882, 0x8882, 0x8842, 0x884A, 0x882A, 0x881A, 0x4004, 0x3018, 0x0}},
    {12858, 2, {0x0, 0x3018, 0x4034, 0xBBC2, 0xA84A, 0xAA4A, 0xA952, 0xAFFA, 0xB842, 0xA842, 0x8042, 0x8042, 0x81C2, 0x4004, 0x3018, 0x0}},
    {12859, 2, {0x0, 0x3018, 0x4914, 0x84A2, 0x9FFA, 0x900A, 0x87E2, 0x8042, 0x8082, 0x9FFA, 0x8082, 0x8082, 0x8382, 0x4004, 0x3018, 0x0}},
    {12860, 2, {0x0, 0x3018, 0x5F24, 0x9422, 0x9F3A, 0x9142, 0x9F52, 0x940A, 0x9F82, 0x8002, 0x9FF2, 0x9292, 0xBFFA, 0x4004, 0x3018, 0x0}},
    {12861, 2, {0x0, 0x3018, 0x4104, 0x8182, 0x8242, 0x8422, 0x889A, 0xB082, 0x84E2, 0x8482, 0x8482, 0x8482, 0x9FFA, 0x4004, 0x3018, 0x0}},
    {12862, 2, {0x0, 0x3018, 0x5104, 0x8BFA, 0x844A, 0x9FE2, 0x8C3A, 0x87E2, 0x8422, 0x87E2, 0x8422, 0x87E2, 0x9C3A, 0x4004, 0x3018, 0x0}},
    {12863, 2, {0x0, 0x3018, 0x4884, 0x89F2, 0x8892, 0xBE92, 0x8922, 0x8922, 0x8FFA, 0x8AAA, 0x8AAA, 0x8CCA, 0x8912, 0x4004, 0x3018, 0x0}},
    {12864, 2, {0x0, 0x3018, 0x48FC, 0x8F92, 0x92A2, 0xAA42, 0x9422, 0x8BD2, 0x900A, 0xAFF2, 0x84A2, 0x8A92, 0xB10A, 0x4004, 0x3018, 0x0}},
    {12865, 2, {0x0, 0x3018, 0x4444, 0x8442, 0x8842, 0x8FFA, 0x9842, 0xA8E2, 0x88E2, 0x8952, 0x8A4A, 0x8C42, 0x8842, 0x4004, 0x3018, 0x0}},
    {12866, 2, {0x0, 0x3018, 0x4204, 0x8402, 0x8FE2, 0x8822, 0x8822, 0x8FE2, 0x8822, 0x8FE2, 0x8822, 0x8822, 0x8FE2, 0x4004, 0x3018, 0x0}},
    {12867, 2, {0x0, 0x3018, 0x5FF4, 0x8202, 0x8442, 0x8822, 0x9FF2, 0x8102, 0x8102, 0x8FE2, 0x8102, 0x8102, 0xBFF2, 0x4004, 0x3018, 0x0}},
    {12881, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C22, 0xA262, 0x8222, 0x8222, 0x9C22, 0xA022, 0xA022, 0xBEFA, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12882, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x820A, 0x820A, 0x9C72, 0xA082, 0xA082, 0xBEFA, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12883, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x820A, 0x8232, 0x9C0A, 0xA00A, 0xA08A, 0xBE72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12884, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C92, 0xA292, 0x8292, 0x82FA, 0x9C12, 0xA012, 0xA012, 0xBE12, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12885, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9CFA, 0xA282, 0x8282, 0x8272, 0x9C0A, 0xA00A, 0xA08A, 0xBE72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12886, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x8282, 0x82F2, 0x9C8A, 0xA08A, 0xA08A, 0xBE72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12887, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9CFA, 0xA28A, 0x820A, 0x8212, 0x9C12, 0xA022, 0xA022, 0xBE22, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12888, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x828A, 0x8272, 0x9C8A, 0xA08A, 0xA08A, 0xBE72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12889, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x828A, 0x828A, 0x9C7A, 0xA00A, 0xA08A, 0xBE72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12890, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x828A, 0x8C8A, 0x828A, 0x828A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12891, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C22, 0xA262, 0x8222, 0x8C22, 0x8222, 0x8222, 0xA222, 0x9CFA, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12892, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x820A, 0x8C0A, 0x8272, 0x8282, 0xA282, 0x9CFA, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12893, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x820A, 0x8C32, 0x820A, 0x820A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12894, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C92, 0xA292, 0x8292, 0x8CFA, 0x8212, 0x8212, 0xA212, 0x9C12, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12895, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9CFA, 0xA282, 0x8282, 0x8C72, 0x820A, 0x820A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12896, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4002, 0x5FF2, 0x4012, 0x4012, 0x4012, 0x4012, 0x4012, 0x4012, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12897, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4002, 0x4802, 0x4802, 0x4802, 0x4802, 0x4802, 0x4802, 0x4FF2, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12898, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4FF2, 0x4802, 0x4802, 0x4802, 0x4802, 0x4802, 0x4FF2, 0x4002, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12899, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4FF2, 0x4012, 0x4012, 0x4FF2, 0x4802, 0x4802, 0x4802, 0x4FFA, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12900, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4FF2, 0x4812, 0x4812, 0x4812, 0x4812, 0x4812, 0x4FF2, 0x4002, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12901, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4812, 0x4812, 0x4812, 0x4FF2, 0x4812, 0x4812, 0x4812, 0x4FF2, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12902, 2, {0x0, 0xFF0, 0x1008, 0x2084, 0x4082, 0x4082, 0x4142, 0x4142, 0x4222, 0x4222, 0x4412, 0x480A, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12903, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x43C2, 0x4422, 0x4812, 0x4812, 0x4812, 0x4812, 0x4422, 0x43C2, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12904, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4002, 0x4FFA, 0x4082, 0x4082, 0x4142, 0x4222, 0x4412, 0x480A, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12905, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x43E2, 0x4002, 0x4FFA, 0x4082, 0x4142, 0x4222, 0x4412, 0x480A, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12906, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4FF2, 0x4012, 0x4012, 0x5FF2, 0x4012, 0x4012, 0x4012, 0x4012, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12907, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4FE2, 0x4802, 0x4802, 0x4FE2, 0x4802, 0x4802, 0x4802, 0x4FF2, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12908, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x4002, 0x4FF2, 0x4002, 0x4422, 0x4222, 0x4242, 0x4242, 0x5FFA, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12909, 2, {0x0, 0xFF0, 0x1008, 0x2004, 0x47E2, 0x4002, 0x5FFA, 0x4002, 0x43C2, 0x4422, 0x4422, 0x43C2, 0x2004, 0x1008, 0xFF0, 0x0}},
    {12910, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FA2, 0x40A2, 0x413E, 0x4122, 0x4222, 0x4422, 0x4822, 0x5022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12911, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5022, 0x5022, 0x5022, 0x503E, 0x5022, 0x5FE2, 0x4022, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12912, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FA2, 0x5022, 0x5022, 0x503E, 0x5022, 0x5FA2, 0x4022, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12913, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FA2, 0x40A2, 0x40A2, 0x5FBE, 0x5022, 0x5022, 0x5FA2, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12914, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x4022, 0x5F22, 0x5122, 0x513E, 0x5122, 0x5122, 0x5F22, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12915, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5122, 0x5122, 0x5F22, 0x513E, 0x5122, 0x5122, 0x5F22, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12916, 2, {0x0, 0xFF0, 0x1028, 0x2224, 0x4222, 0x4422, 0x4622, 0x493E, 0x50A2, 0x6022, 0x4022, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12917, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x4E22, 0x5122, 0x5122, 0x513E, 0x5122, 0x5122, 0x4E22, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12918, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FE2, 0x4222, 0x4222, 0x453E, 0x48A2, 0x5062, 0x6022, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12919, 2, {0x0, 0xFF0, 0x1028, 0x27A4, 0x4022, 0x5FE2, 0x4222, 0x453E, 0x48A2, 0x5062, 0x6022, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12920, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FA2, 0x40A2, 0x4122, 0x5E3E, 0x4422, 0x4822, 0x5022, 0x6022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12921, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FA2, 0x5022, 0x5FA2, 0x503E, 0x5022, 0x5FA2, 0x4022, 0x6022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12922, 2, {0x0, 0xFF0, 0x1028, 0x2024, 0x5FA2, 0x4022, 0x50A2, 0x493E, 0x4922, 0x4BE2, 0x7C22, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12923, 2, {0x0, 0xFF0, 0x1028, 0x2F24, 0x4022, 0x7FE2, 0x4022, 0x4F3E, 0x50A2, 0x50A2, 0x4F22, 0x4022, 0x2024, 0x1028, 0xFF0, 0x0}},
    {12927, 2, {0x0, 0x1FF0, 0x2008, 0x5EF4, 0x8C42, 0x8C80, 0xED0E, 0xE02, 0xD02, 0x8C82, 0x8C42, 0x5EF4, 0x2008, 0x1FF0, 0x0, 0x0}},
    {12928, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x8002, 0x8002, 0x8012, 0xBFFA, 0x8002, 0x8002, 0x8002, 0x8002, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12929, 2, {0x0, 0x1FF0, 0x2008, 0x4024, 0x9FF2, 0x8002, 0x8002, 0x8002, 0x8002, 0x8012, 0xBFFA, 0x8002, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12930, 2, {0x0, 0x1FF0, 0x2008, 0x4024, 0x9FF2, 0x8002, 0x8042, 0x8FE2, 0x8002, 0x8012, 0xBFFA, 0x8002, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12931, 2, {0x0, 0x1FF0, 0x2008, 0x5FF4, 0x9292, 0x9292, 0x9292, 0x9292, 0x94F2, 0x9812, 0x9012, 0x9FF2, 0x5014, 0x2008, 0x1FF0, 0x0}},
    {12932, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9FF2, 0x8102, 0x8102, 0x8FE2, 0x8222, 0x8222, 0x8222, 0x8222, 0x7FFC, 0x2008, 0x1FF0, 0x0}},
    {12933, 2, {0x0, 0x1FF0, 0x2208, 0x4104, 0x8002, 0xBFFA, 0x8002, 0x8442, 0x8422, 0x8812, 0x900A, 0xA00A, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12934, 2, {0x0, 0xFF0, 0x1008, 0x2204, 0x4202, 0x4232, 0x43C2, 0x5E02, 0x4202, 0x4202, 0x4212, 0x4212, 0x21F4, 0x1008, 0xFF0, 0x0}},
    {12935, 2, {0x0, 0x1FF0, 0x2008, 0x4084, 0x8482, 0x8482, 0x8442, 0x8442, 0x8822, 0x8822, 0x9012, 0xA00A, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12936, 2, {0x0, 0x1FF0, 0x2208, 0x4204, 0x8202, 0x9FC2, 0x8242, 0x8242, 0x8242, 0x8442, 0x844A, 0x884A, 0x503C, 0x2008, 0x1FF0, 0x0}},
    {12937, 2, {0x0, 0xFE0, 0x1010, 0x2108, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x2008, 0x1010, 0xFE0, 0x0}},
    {12938, 2, {0x0, 0x1FF0, 0x2008, 0x47F4, 0x8412, 0x8412, 0x87F2, 0x8412, 0x8412, 0x87F2, 0x8412, 0x8412, 0x4854, 0x3028, 0x1FF0, 0x0}},
    {12939, 2, {0x0, 0x1FF0, 0x2008, 0x4104, 0x8912, 0x8922, 0x8942, 0x9102, 0x8282, 0x8282, 0x8442, 0x8822, 0x701C, 0x2008, 0x1FF0, 0x0}},
    {12940, 2, {0x0, 0x1FF0, 0x2108, 0x4104, 0x8112, 0xBDA2, 0x8542, 0x8542, 0x8922, 0x8922, 0x9112, 0xA50A, 0x4204, 0x2008, 0x1FF0, 0x0}},
    {12941, 2, {0x0, 0x1FF0, 0x2108, 0x4104, 0x8102, 0xBFFA, 0x8102, 0x8382, 0x8542, 0x8922, 0xB11A, 0x8112, 0x4104, 0x2008, 0x1FF0, 0x0}},
    {12942, 2, {0x0, 0x1FF0, 0x2108, 0x4284, 0x8442, 0x8822, 0xB7DA, 0x8102, 0x9FF2, 0x8102, 0x8922, 0x8542, 0x7FFC, 0x2008, 0x1FF0, 0x0}},
    {12943, 2, {0x0, 0x1FF0, 0x2108, 0x4104, 0x8102, 0x8102, 0x9FF2, 0x8102, 0x8102, 0x8102, 0x8102, 0x8102, 0x7FFC, 0x2008, 0x1FF0, 0x0}},
    {12944, 2, {0x0, 0x7F0, 0x808, 0x17F4, 0x2412, 0x2412, 0x2412, 0x27F2, 0x2412, 0x2412, 0x2412, 0x27F2, 0x1414, 0x808, 0x7F0, 0x0}},
    {12945, 2, {0x0, 0x1FF0, 0x2848, 0x4944, 0xBDF2, 0x8A42, 0x9C42, 0x9BFA, 0xA842, 0xA8E2, 0x8952, 0x8A4A, 0x4844, 0x2008, 0x1FF0, 0x0}},
    {12946, 2, {0x0, 0x1FF0, 0x2208, 0x7FFC, 0x8402, 0x87E2, 0x8C22, 0x97E2, 0xA422, 0x87E2, 0x8422, 0x8422, 0x4464, 0x2008, 0x1FF0, 0x0}},
    {12947, 2, {0x0, 0x1FF0, 0x3048, 0x4844, 0xBE42, 0x8442, 0x89F2, 0x8C42, 0x9A42, 0xA842, 0x8842, 0x8842, 0x4BFC, 0x2008, 0x1FF0, 0x0}},
    {12948, 2, {0x0, 0x1FF0, 0x2408, 0x47F4, 0x8422, 0x8A42, 0x9182, 0x8102, 0x87F2, 0xBC12, 0x8412, 0x8412, 0x47F4, 0x2008, 0x1FF0, 0x0}},
    {12949, 2, {0x0, 0x1FF0, 0x2848, 0x69F4, 0xBC42, 0xA842, 0xABFA, 0x8C22, 0x9BFA, 0xA922, 0x88A2, 0x8822, 0x48E4, 0x2008, 0x1FF0, 0x0}},
    {12950, 2, {0x0, 0x1FF0, 0x2018, 0x5E14, 0x93FA, 0x9E12, 0x9232, 0x9E52, 0x9292, 0x9F12, 0x8012, 0x9252, 0x6124, 0x2008, 0x1FF0, 0x0}},
    {12951, 2, {0x0, 0x1FF0, 0x2008, 0x51F4, 0x8912, 0xBF12, 0x8312, 0x85F2, 0x8CA2, 0x9AA2, 0xAAAA, 0x892A, 0x4A1C, 0x2008, 0x1FF0, 0x0}},
    {12952, 2, {0x0, 0x1FF0, 0x3218, 0x4914, 0x8922, 0x9FFA, 0x910A, 0xA102, 0x9FF2, 0x8112, 0x8212, 0x8412, 0x5864, 0x2008, 0x1FF0, 0x0}},
    {12953, 2, {0x0, 0x1FF0, 0x2488, 0x7854, 0x8892, 0xBE92, 0x88A2, 0x8DA2, 0x9AAA, 0xAACA, 0x8892, 0x8992, 0x4E74, 0x2008, 0x1FF0, 0x0}},
    {12954, 2, {0x0, 0xFF0, 0x1008, 0x3FFC, 0x4912, 0x4FF2, 0x4912, 0x4FF2, 0x4102, 0x5FFA, 0x410A, 0x420A, 0x244C, 0x1838, 0xFF0, 0x0}},
    {12955, 2, {0x0, 0x1FF0, 0x2208, 0x4204, 0x8202, 0xBFFA, 0x8442, 0x8442, 0x8842, 0x8E82, 0x8182, 0x8262, 0x5C14, 0x2008, 0x1FF0, 0x0}},
    {12956, 2, {0x0, 0x1FF0, 0x3048, 0x4BFC, 0x88A2, 0x83FA, 0xBA4A, 0x8BFA, 0x8AAA, 0x8AEA, 0x8A1A, 0x9402, 0x63FC, 0x2008, 0x1FF0, 0x0}},
    {12957, 2, {0x0, 0x1FF0, 0x2BF8, 0x4914, 0x89F2, 0x8912, 0x9BFA, 0xAC8A, 0x8B52, 0x8DEA, 0x8922, 0x88C2, 0x4F3C, 0x2008, 0x1FF0, 0x0}},
    {12958, 2, {0x0, 0xFF0, 0x1208, 0x3CFC, 0x508A, 0x508A, 0x5E8A, 0x508A, 0x508A, 0x528A, 0x5CB2, 0x5082, 0x2084, 0x1008, 0xFF0, 0x0}},
    {12959, 2, {0x0, 0x1FF0, 0x3088, 0x4844, 0xA3FA, 0x9442, 0x8442, 0x8842, 0x89F2, 0xB042, 0x9042, 0x9042, 0x53FC, 0x2008, 0x1FF0, 0x0}},
    {12960, 2, {0x0, 0x1FF8, 0x2004, 0x41FA, 0xBE41, 0x89F9, 0x8909, 0x89F9, 0x8909, 0x8FF9, 0xB109, 0x81F9, 0x4092, 0x230C, 0x1FF8, 0x0}},
    {12961, 2, {0x0, 0x1FF0, 0x2448, 0x4444, 0x8842, 0x8FFA, 0x9842, 0xA8E2, 0x88E2, 0x8952, 0x8A4A, 0x8C42, 0x4844, 0x2008, 0x1FF0, 0x0}},
    {12962, 2, {0x0, 0x1FF0, 0x2008, 0x5FFC, 0x900A, 0xA412, 0x87E2, 0x8402, 0x87F2, 0x8012, 0x9FD2, 0x8012, 0x40A4, 0x2048, 0x1FF0, 0x0}},
    {12963, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x8002, 0x9FF2, 0x8102, 0x8102, 0x89E2, 0x8902, 0x8902, 0x9FF2, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12964, 2, {0x0, 0x1FF0, 0x2108, 0x4104, 0x8102, 0x8102, 0x81F2, 0x8102, 0x8102, 0x8102, 0x8102, 0x8102, 0x7FFC, 0x2008, 0x1FF0, 0x0}},
    {12965, 2, {0x0, 0x7F0, 0x888, 0x1084, 0x2FFA, 0x288A, 0x288A, 0x288A, 0x2FFA, 0x288A, 0x2082, 0x2082, 0x1084, 0x808, 0x7F0, 0x0}},
    {12966, 2, {0x0, 0x1FF0, 0x2008, 0x7FFC, 0x8102, 0x8102, 0x8182, 0x8142, 0x8122, 0x8122, 0x8102, 0x8102, 0x4104, 0x2008, 0x1FF0, 0x0}},
    {12967, 2, {0x0, 0x1FF0, 0x2208, 0x4204, 0xBFFA, 0x8202, 0x8402, 0x87F2, 0x8882, 0x8882, 0x9082, 0xA082, 0x4FFC, 0x2008, 0x1FF0, 0x0}},
    {12968, 2, {0x0, 0x1FF0, 0x2108, 0x4104, 0xBFFA, 0x8202, 0x8402, 0x87E2, 0x8C22, 0x9422, 0xA422, 0x87E2, 0x4424, 0x2008, 0x1FF0, 0x0}},
    {12969, 2, {0x0, 0xFF0, 0x1008, 0x3FFC, 0x5202, 0x53F2, 0x5482, 0x5FFA, 0x5082, 0x5142, 0x5222, 0x5412, 0x3FFC, 0x1008, 0xFF0, 0x0}},
    {12970, 2, {0x0, 0x1FF0, 0x2008, 0x4084, 0x9FFA, 0x900A, 0xA7E2, 0x8002, 0x9FFA, 0x8082, 0x88A2, 0x8892, 0x5294, 0x2108, 0x1FF0, 0x0}},
    {12971, 2, {0x0, 0xFF0, 0x1918, 0x24A4, 0x5FFA, 0x500A, 0x47E2, 0x4042, 0x4082, 0x5FFA, 0x4082, 0x4082, 0x2384, 0x1008, 0xFF0, 0x0}},
    {12972, 2, {0x0, 0x1FF0, 0x3E48, 0x6844, 0xBE7A, 0xA282, 0xBEA2, 0xA812, 0xBF02, 0x8002, 0x9FF2, 0x9292, 0x7FFC, 0x2008, 0x1FF0, 0x0}},
    {12973, 2, {0x0, 0x1FF0, 0x2108, 0x4184, 0x8242, 0x8422, 0x889A, 0xB082, 0x84E2, 0x8482, 0x8482, 0x8482, 0x5FFC, 0x2008, 0x1FF0, 0x0}},
    {12974, 2, {0x0, 0xFF0, 0x1108, 0x2BFC, 0x444A, 0x5FE2, 0x4C3A, 0x47E2, 0x4422, 0x47E2, 0x4422, 0x47E2, 0x3C3C, 0x1008, 0xFF0, 0x0}},
    {12975, 2, {0x0, 0xFF0, 0x1888, 0x29F4, 0x4892, 0x7E92, 0x4922, 0x4922, 0x4FFA, 0x4AAA, 0x4AAA, 0x4CCA, 0x2914, 0x1008, 0xFF0, 0x0}},
    {12976, 2, {0x0, 0x1FF0, 0x2108, 0x7FFC, 0x8482, 0x84F2, 0x8992, 0x9A52, 0xAD22, 0x88A2, 0x8842, 0x88A2, 0x4B1C, 0x2008, 0x1FF0, 0x0}},
    {12977, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x8282, 0x8CF2, 0x828A, 0x828A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12978, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9CFA, 0xA28A, 0x820A, 0x8C12, 0x8212, 0x8222, 0xA222, 0x9C22, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12979, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x828A, 0x8C72, 0x828A, 0x828A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12980, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0x9C72, 0xA28A, 0x828A, 0x8C8A, 0x827A, 0x820A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12981, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA472, 0xA48A, 0xA48A, 0xBE8A, 0x848A, 0x848A, 0x848A, 0x8472, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12982, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA422, 0xA462, 0xA422, 0xBE22, 0x8422, 0x8422, 0x8422, 0x84FA, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12983, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA472, 0xA48A, 0xA40A, 0xBE0A, 0x8472, 0x8482, 0x8482, 0x84FA, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12984, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA472, 0xA48A, 0xA40A, 0xBE32, 0x840A, 0x840A, 0x848A, 0x8472, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12985, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA492, 0xA492, 0xA492, 0xBEFA, 0x8412, 0x8412, 0x8412, 0x8412, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12986, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA4FA, 0xA482, 0xA482, 0xBE72, 0x840A, 0x840A, 0x848A, 0x8472, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12987, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA472, 0xA48A, 0xA482, 0xBEF2, 0x848A, 0x848A, 0x848A, 0x8472, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12988, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA4FA, 0xA48A, 0xA40A, 0xBE12, 0x8412, 0x8422, 0x8422, 0x8422, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12989, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA472, 0xA48A, 0xA48A, 0xBE72, 0x848A, 0x848A, 0x848A, 0x8472, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12990, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xA472, 0xA48A, 0xA48A, 0xBE8A, 0x847A, 0x840A, 0x848A, 0x8472, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12991, 2, {0x0, 0x1FF0, 0x2008, 0x4004, 0xBE72, 0xA08A, 0xA08A, 0x9C8A, 0x828A, 0x828A, 0xA28A, 0x9C72, 0x4004, 0x2008, 0x1FF0, 0x0}},
    {12992, 2, {0x0, 0x107C, 0x3044, 0x1044, 0x1044, 0x107C, 0x1044, 0x1044, 0x1044, 0x107C, 0x1044, 0x1044, 0x1044, 0x1044, 0x7C94, 0x88}},
    {12993, 2, {0x0, 0x7C, 0x1844, 0x2444, 0x2444, 0x47C, 0x444, 0x844, 0x844, 0x87C, 0x1044, 0x1044, 0x1044, 0x2044, 0x2094, 0x3C88}},
    {12994, 2, {0x0, 0x7C, 0x1844, 0x2444, 0x2444, 0x47C, 0x444, 0x444, 0x1844, 0x47C, 0x444, 0x444, 0x444, 0x2444, 0x2494, 0x1888}},
    {12995, 2, {0x0, 0x7C, 0x2444, 0x2444, 0x2444, 0x247C, 0x2444, 0x2444, 0x3C44, 0x47C, 0x444, 0x444, 0x444, 0x444, 0x494, 0x488}},
    {12996, 2, {0x0, 0x7C, 0x3C44, 0x2044, 0x2044, 0x207C, 0x2044, 0x2044, 0x3844, 0x47C, 0x444, 0x444, 0x444, 0x2444, 0x2494, 0x1888}},
    {12997, 2, {0x0, 0x7C, 0x1844, 0x2444, 0x2444, 0x207C, 0x2044, 0x2044, 0x3844, 0x247C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2494, 0x1888}},
    {12998, 2, {0x0, 0x7C, 0x3C44, 0x444, 0x444, 0x47C, 0x844, 0x844, 0x844, 0x107C, 0x1044, 0x1044, 0x1044, 0x1044, 0x1094, 0x1088}},
    {12999, 2, {0x0, 0x7C, 0x1844, 0x2444, 0x2444, 0x247C, 0x2444, 0x2444, 0x1844, 0x247C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2494, 0x1888}},
    {13000, 2, {0x0, 0x7C, 0x1844, 0x2444, 0x2444, 0x247C, 0x2444, 0x2444, 0x1844, 0x47C, 0x444, 0x444, 0x444, 0x2444, 0x2494, 0x1888}},
    {13001, 2, {0x0, 0x3E, 0x2322, 0x64A2, 0x24A2, 0x24BE, 0x24A2, 0x24A2, 0x24A2, 0x24BE, 0x24A2, 0x24A2, 0x24A2, 0x24A2, 0x24CA, 0x7344}},
    {13002, 2, {0x0, 0x3E, 0x2222, 0x6622, 0x2222, 0x223E, 0x2222, 0x2222, 0x2222, 0x223E, 0x2222, 0x2222, 0x2222, 0x2222, 0x224A, 0x7744}},
    {13003, 2, {0x0, 0x3E, 0x2322, 0x64A2, 0x24A2, 0x20BE, 0x20A2, 0x2122, 0x2122, 0x213E, 0x2222, 0x2222, 0x2222, 0x2422, 0x244A, 0x77C4}},
    {13008, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4024, 0x4144, 0x4184, 0x4104, 0x2208, 0x2408, 0x1010, 0xC60, 0x380, 0x0}},
    {13009, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2048, 0x2088, 0x4104, 0x4304, 0x4504, 0x4904, 0x2108, 0x2108, 0x1010, 0xC60, 0x380, 0x0}},
    {13010, 2, {0x0, 0x380, 0xC60, 0x1110, 0x2108, 0x2FE8, 0x4824, 0x4824, 0x4024, 0x4044, 0x2048, 0x2088, 0x1110, 0xC60, 0x380, 0x0}},
    {13011, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2008, 0x47C4, 0x4104, 0x4104, 0x4104, 0x2FE8, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13012, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2088, 0x2FE8, 0x4184, 0x4284, 0x4484, 0x4884, 0x2288, 0x2188, 0x1010, 0xC60, 0x380, 0x0}},
    {13013, 2, {0x0, 0x380, 0xC60, 0x1110, 0x2108, 0x2FE8, 0x4124, 0x4224, 0x4444, 0x4844, 0x2288, 0x2108, 0x1010, 0xC60, 0x380, 0x0}},
    {13014, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2208, 0x23C8, 0x4F04, 0x4104, 0x41F4, 0x4FC4, 0x2048, 0x2048, 0x1010, 0xC60, 0x380, 0x0}},
    {13015, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2208, 0x23E8, 0x4224, 0x4424, 0x4844, 0x4044, 0x2088, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13016, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2408, 0x27E8, 0x4484, 0x4884, 0x5084, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13017, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4024, 0x4024, 0x4044, 0x4044, 0x2FE8, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13018, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2248, 0x2248, 0x4FF4, 0x4244, 0x4244, 0x4244, 0x2088, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13019, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2608, 0x2108, 0x4C04, 0x4224, 0x4044, 0x4084, 0x2908, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13020, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FC8, 0x4044, 0x4044, 0x4084, 0x4184, 0x2248, 0x2448, 0x1010, 0xC60, 0x380, 0x0}},
    {13021, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2208, 0x2FE8, 0x4224, 0x4244, 0x4284, 0x4204, 0x2208, 0x21E8, 0x1010, 0xC60, 0x380, 0x0}},
    {13022, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2828, 0x4424, 0x4424, 0x4044, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13023, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2208, 0x23E8, 0x4224, 0x4624, 0x4544, 0x48C4, 0x20A8, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13024, 2, {0x0, 0x380, 0xC60, 0x1010, 0x20E8, 0x2F88, 0x40F4, 0x5F84, 0x4084, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13025, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2208, 0x2928, 0x4524, 0x4424, 0x4044, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13026, 2, {0x0, 0x380, 0xC60, 0x1010, 0x27C8, 0x2008, 0x4FE4, 0x4084, 0x4084, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13027, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2208, 0x2208, 0x4204, 0x4304, 0x4284, 0x4244, 0x2208, 0x2208, 0x1010, 0xC60, 0x380, 0x0}},
    {13028, 2, {0x0, 0x380, 0xC60, 0x1110, 0x2088, 0x2FE8, 0x4084, 0x4084, 0x4084, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13029, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x21E8, 0x4E04, 0x4004, 0x40F4, 0x4F04, 0x2008, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13030, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FC8, 0x4044, 0x4644, 0x4184, 0x4184, 0x2248, 0x2448, 0x1010, 0xC60, 0x380, 0x0}},
    {13031, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FC8, 0x4044, 0x4084, 0x4104, 0x4344, 0x2528, 0x2108, 0x1110, 0xC60, 0x380, 0x0}},
    {13032, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2028, 0x4024, 0x4024, 0x4044, 0x4084, 0x2108, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13033, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2288, 0x4244, 0x4424, 0x5824, 0x4004, 0x2008, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13034, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2808, 0x4984, 0x4E04, 0x4804, 0x4804, 0x2808, 0x27C8, 0x1010, 0xC60, 0x380, 0x0}},
    {13035, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4024, 0x4024, 0x4024, 0x4044, 0x2048, 0x2088, 0x1110, 0xC60, 0x380, 0x0}},
    {13036, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2208, 0x4504, 0x4884, 0x4044, 0x4024, 0x2008, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13037, 2, {0x0, 0x380, 0xC60, 0x1110, 0x2108, 0x2FE8, 0x4104, 0x4544, 0x4924, 0x4124, 0x2508, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13038, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4024, 0x4344, 0x4084, 0x4044, 0x2048, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13039, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2308, 0x20C8, 0x4604, 0x4184, 0x4004, 0x4604, 0x21C8, 0x2018, 0x1010, 0xC60, 0x380, 0x0}},
    {13040, 2, {0x0, 0x380, 0xC60, 0x1110, 0x2108, 0x2108, 0x4204, 0x4204, 0x4444, 0x4424, 0x27E8, 0x2828, 0x1010, 0xC60, 0x380, 0x0}},
    {13041, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2028, 0x4024, 0x4224, 0x4144, 0x40C4, 0x20A8, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13042, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2FC8, 0x2208, 0x43C4, 0x5E04, 0x4204, 0x4204, 0x2208, 0x21E8, 0x1010, 0xC60, 0x380, 0x0}},
    {13043, 2, {0x0, 0x380, 0xC60, 0x1210, 0x2208, 0x2FE8, 0x4224, 0x4144, 0x4104, 0x4104, 0x2088, 0x2088, 0x1090, 0xC60, 0x380, 0x0}},
    {13044, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2008, 0x47E4, 0x4024, 0x4044, 0x4084, 0x2FE8, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13045, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x27E8, 0x4024, 0x47E4, 0x4024, 0x4024, 0x27E8, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13046, 2, {0x0, 0x380, 0xC60, 0x1010, 0x27C8, 0x2008, 0x4FE4, 0x4024, 0x4024, 0x4044, 0x2048, 0x2088, 0x1310, 0xC60, 0x380, 0x0}},
    {13047, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2248, 0x2248, 0x4244, 0x4244, 0x4244, 0x4044, 0x2088, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13048, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2108, 0x2508, 0x4504, 0x4514, 0x4524, 0x4544, 0x2588, 0x2908, 0x1010, 0xC60, 0x380, 0x0}},
    {13049, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2408, 0x2408, 0x4404, 0x4424, 0x4444, 0x4484, 0x2508, 0x2608, 0x1010, 0xC60, 0x380, 0x0}},
    {13050, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x28E8, 0x4F24, 0x4824, 0x4844, 0x49E4, 0x2E08, 0x2808, 0x1010, 0xC60, 0x380, 0x0}},
    {13051, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4824, 0x4824, 0x4024, 0x4044, 0x2048, 0x2088, 0x1110, 0xC60, 0x380, 0x0}},
    {13052, 2, {0x0, 0x380, 0xC60, 0x1090, 0x2088, 0x2FE8, 0x4484, 0x4484, 0x5FE4, 0x4084, 0x2088, 0x2088, 0x1090, 0xC60, 0x380, 0x0}},
    {13053, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4024, 0x4344, 0x4084, 0x4044, 0x3FF8, 0x2008, 0x1010, 0xC60, 0x380, 0x0}},
    {13054, 2, {0x0, 0x380, 0xC60, 0x1010, 0x2008, 0x2FE8, 0x4024, 0x47E4, 0x4024, 0x4044, 0x2088, 0x2308, 0x1010, 0xC60, 0x380, 0x0}},
    {13056, 2, {0x7C04, 0x40A, 0x154A, 0x1924, 0x1120, 0x1220, 0x2000, 0x0, 0x40, 0x40, 0x40, 0x7C60, 0x50, 0x40, 0x40, 0x0}},
    {13057, 2, {0x3E00, 0x210, 0xA50, 0xC50, 0x854, 0x854, 0x1058, 0x0, 0x0, 0x3E00, 0x27C, 0x204, 0x218, 0x410, 0x1820, 0x0}},
    {13058, 2, {0x3E00, 0x240, 0xA20, 0xC04, 0x818, 0x860, 0x1000, 0x0, 0x13E, 0x2282, 0x528A, 0x90C, 0x408, 0x8, 0x10, 0x0}},
    {13059, 2, {0x3E00, 0x200, 0xA00, 0xC7C, 0x800, 0x800, 0x1000, 0x0, 0x0, 0x800, 0x2800, 0x2800, 0x2A00, 0x2A00, 0x2C00, 0x0}},
    {13060, 2, {0x200, 0x438, 0x800, 0x1800, 0x2800, 0x87C, 0x800, 0x0, 0x102, 0x81E9, 0x4124, 0xA20, 0x3020, 0xC040, 0x180, 0x0}},
    {13061, 2, {0x200, 0x440, 0x820, 0x1804, 0x2818, 0x860, 0x800, 0x0, 0x400, 0x3800, 0x800, 0x3E00, 0x800, 0x800, 0x1000, 0x0}},
    {13062, 2, {0x800, 0x3E00, 0x2208, 0x227C, 0x218, 0x428, 0x848, 0x0, 0x0, 0x2000, 0x1000, 0x200, 0xC00, 0x3000, 0x0, 0x0}},
    {13063, 2, {0x0, 0x7CF8, 0x1008, 0x1010, 0x1020, 0x1050, 0x7C88, 0x0, 0x4022, 0x7829, 0x4824, 0x8BB0, 0x828, 0x1020, 0x6020, 0x0}},
    {13064, 2, {0x0, 0x3E00, 0x800, 0x87C, 0x800, 0x800, 0x3E00, 0x0, 0x1000, 0x3E00, 0x1200, 0x127C, 0x1200, 0x1200, 0x2400, 0x0}},
    {13065, 2, {0x400, 0x3E40, 0x420, 0xC04, 0x1418, 0x2460, 0xC00, 0x0, 0x0, 0x3E00, 0x200, 0x400, 0x800, 0x1400, 0x2200, 0x0}},
    {13066, 2, {0x400, 0x3E00, 0x400, 0xC7C, 0x1400, 0x2400, 0xC00, 0x0, 0x800, 0x800, 0x1000, 0x2000, 0x2200, 0x3E00, 0x200, 0x0}},
    {13067, 2, {0x1004, 0x3E08, 0x1210, 0x1230, 0x1250, 0x1210, 0x2410, 0x0, 0x2400, 0x2400, 0x2400, 0x2400, 0x400, 0x800, 0x3000, 0x0}},
    {13068, 2, {0x1038, 0x3E00, 0x127C, 0x1204, 0x1204, 0x1208, 0x2430, 0x0, 0x20, 0x20, 0x2A20, 0x2A30, 0x228, 0x420, 0x1820, 0x0}},
    {13069, 2, {0x1000, 0x3E7C, 0x1244, 0x1244, 0x1244, 0x127C, 0x2400, 0x0, 0x2400, 0x2400, 0x2400, 0x247C, 0x400, 0x800, 0x3000, 0x0}},
    {13070, 2, {0x4100, 0xFCBE, 0x4A22, 0x4822, 0x4822, 0x483E, 0x9000, 0x0, 0x0, 0x4000, 0x2000, 0x400, 0x1800, 0x6000, 0x0, 0x0}},
    {13071, 2, {0x4080, 0xFA50, 0x4908, 0x4802, 0x480C, 0x4830, 0x9000, 0x0, 0x0, 0x7C00, 0x400, 0x400, 0x2800, 0x1000, 0x800, 0x0}},
    {13072, 2, {0x2040, 0x7D20, 0x1080, 0x7C00, 0x1000, 0x800, 0x800, 0x0, 0x204, 0x7D2, 0x248, 0x240, 0x240, 0x240, 0x480, 0x0}},
    {13073, 2, {0x4080, 0xFA5C, 0x2100, 0xF800, 0x2000, 0x103E, 0x1000, 0x0, 0x0, 0x0, 0x0, 0x7C00, 0x0, 0x0, 0x0, 0x0}},
    {13074, 2, {0x1000, 0x3E00, 0x800, 0x3E38, 0x808, 0x47C, 0x400, 0x0, 0x2400, 0x2400, 0x2400, 0x247C, 0x400, 0x800, 0x3000, 0x0}},
    {13075, 2, {0x4080, 0xFA44, 0x2114, 0xF814, 0x2015, 0x1015, 0x1016, 0x0, 0x4080, 0x7A40, 0x4900, 0xA83E, 0x1800, 0x1000, 0x6000, 0x0}},
    {13076, 2, {0x1000, 0x3E00, 0x800, 0x3E00, 0x800, 0x400, 0x400, 0x0, 0x0, 0x7C, 0x44, 0x44, 0x44, 0x7C, 0x0, 0x0}},
    {13077, 2, {0x2000, 0x7CF8, 0x1088, 0x7C88, 0x1088, 0x8F8, 0x800, 0x0, 0x4132, 0x7482, 0x5274, 0x9014, 0x1015, 0x2027, 0x4041, 0x0}},
    {13078, 2, {0x4002, 0xFBC2, 0x2242, 0xFA54, 0x2248, 0x13D4, 0x1020, 0x0, 0x200, 0x208, 0x228, 0xFB28, 0x2AA, 0x22A, 0x22C, 0x0}},
    {13079, 2, {0x1000, 0x3E3E, 0x822, 0x3E22, 0x822, 0x43E, 0x400, 0x0, 0x8, 0xF808, 0x8AA8, 0x8AAC, 0x82A, 0x1048, 0x6188, 0x0}},
    {13080, 2, {0x408C, 0x7A40, 0x491E, 0x8802, 0x802, 0x1004, 0x6008, 0x0, 0x1000, 0x1000, 0x2000, 0x4000, 0x4400, 0x7C00, 0x400, 0x0}},
    {13081, 2, {0x408C, 0x7A40, 0x491E, 0x8802, 0x802, 0x1004, 0x6008, 0x0, 0x2200, 0x2210, 0x4208, 0x8302, 0x8A8C, 0xFA30, 0xA00, 0x0}},
    {13082, 2, {0x2000, 0x3C20, 0x24A0, 0x44A0, 0x4A8, 0x8A8, 0x30B0, 0x0, 0x4110, 0xFCAF, 0x4A49, 0x50C9, 0x4149, 0x4049, 0x384F, 0x0}},
    {13083, 2, {0x1000, 0x1E7C, 0x1244, 0x2244, 0x244, 0x47C, 0x1800, 0x0, 0x10, 0x7C, 0x8, 0x3E10, 0x38, 0x54, 0x10, 0x0}},
    {13084, 2, {0x1000, 0x1E00, 0x2400, 0x47C, 0x400, 0x400, 0x800, 0x0, 0x0, 0x3E00, 0x200, 0x400, 0x800, 0x1400, 0x2200, 0x0}},
    {13085, 2, {0x0, 0x3E10, 0x250, 0x250, 0x254, 0x254, 0x3E58, 0x0, 0x800, 0x800, 0x3E00, 0x800, 0x800, 0x1000, 0x2000, 0x0}},
    {13086, 2, {0x0, 0x3E00, 0x200, 0x27C, 0x200, 0x200, 0x3E00, 0x0, 0x840, 0x3EA0, 0x8A0, 0x2A40, 0x2A00, 0x800, 0x1800, 0x0}},
    {13087, 2, {0x1404, 0x3E08, 0x1410, 0x1430, 0x450, 0x810, 0x1010, 0x0, 0x1000, 0x1E10, 0x1250, 0x2250, 0x254, 0x454, 0x1858, 0x0}},
    {13088, 2, {0x1400, 0x3E40, 0x1420, 0x1404, 0x418, 0x860, 0x1000, 0x0, 0x1008, 0xE008, 0x2010, 0xFBA0, 0x2022, 0x203E, 0x4002, 0x0}},
    {13089, 2, {0x2090, 0x1490, 0x4490, 0x2490, 0x810, 0x7020, 0xC0, 0x0, 0x102, 0x81E9, 0x4124, 0xA20, 0x3020, 0xC040, 0x180, 0x0}},
    {13090, 2, {0x1000, 0x3E40, 0x1220, 0x1404, 0x1018, 0x1060, 0xE00, 0x0, 0x400, 0x3800, 0x800, 0x3E00, 0x800, 0x800, 0x1000, 0x0}},
    {13091, 2, {0x1000, 0x3E40, 0x1220, 0x1404, 0x1018, 0x1060, 0xE00, 0x0, 0x1000, 0x1000, 0x1000, 0x1800, 0x1400, 0x1000, 0x1000, 0x0}},
    {13092, 2, {0x4080, 0x7A40, 0x4900, 0xA83E, 0x1800, 0x1000, 0x6000, 0x0, 0x0, 0x7C00, 0x400, 0x800, 0x1000, 0x2800, 0x4400, 0x0}},
    {13093, 2, {0x1C20, 0x90, 0x3E40, 0x800, 0x800, 0x800, 0x1000, 0x0, 0x20, 0x14, 0x44, 0x24, 0x8, 0x70, 0x0, 0x0}},
    {13094, 2, {0x2100, 0x2480, 0x2200, 0x3000, 0x2800, 0x2000, 0x2000, 0x0, 0x0, 0x10, 0x50, 0x50, 0x54, 0x54, 0x58, 0x0}},
    {13095, 2, {0x1000, 0x1000, 0x1000, 0x1800, 0x1400, 0x1000, 0x1000, 0x0, 0x0, 0x40, 0x20, 0x4, 0x18, 0x60, 0x0, 0x0}},
    {13096, 2, {0x800, 0x800, 0x3E00, 0x800, 0x800, 0x1000, 0x2000, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x10, 0x20, 0x0}},
    {13097, 2, {0x400, 0x400, 0x454, 0x454, 0x404, 0x808, 0x1030, 0x0, 0x800, 0x800, 0x800, 0xC00, 0xA00, 0x800, 0x800, 0x0}},
    {13098, 2, {0x4, 0x8, 0x1410, 0x1230, 0x1250, 0x2210, 0x10, 0x0, 0x0, 0x2A00, 0x2A00, 0x200, 0x200, 0x400, 0x1800, 0x0}},
    {13099, 2, {0x100, 0x280, 0x5280, 0x493E, 0x4800, 0x8800, 0x0, 0x0, 0x4008, 0xFA08, 0x4908, 0x502C, 0x40CA, 0x4308, 0x3808, 0x0}},
    {13100, 2, {0x100, 0x280, 0x5280, 0x493E, 0x4800, 0x8800, 0x0, 0x0, 0x0, 0x5400, 0x5400, 0x400, 0x400, 0x800, 0x3000, 0x0}},
    {13101, 2, {0x100, 0x480, 0x5200, 0x483E, 0x4800, 0x8800, 0x0, 0x0, 0x0, 0x4020, 0x40A0, 0x44A0, 0x48A8, 0x50A8, 0x60B0, 0x0}},
    {13102, 2, {0x813E, 0x9A82, 0xE28A, 0x810C, 0x8008, 0x8008, 0x7810, 0x0, 0x200, 0xFA08, 0xA28, 0x1328, 0x22AA, 0x522A, 0x8A2C, 0x0}},
    {13103, 2, {0x8110, 0x9A9E, 0xE292, 0x8122, 0x8002, 0x8004, 0x7818, 0x0, 0x0, 0x1000, 0x5000, 0x5000, 0x5400, 0x5400, 0x5800, 0x0}},
    {13104, 2, {0x2040, 0x26A0, 0x38A0, 0x2040, 0x2000, 0x2000, 0x1E00, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x7C, 0x0}},
    {13105, 2, {0x4040, 0x4D20, 0x7080, 0x4000, 0x4000, 0x4000, 0x3C00, 0x0, 0x0, 0x10, 0x50, 0x50, 0x54, 0x54, 0x58, 0x0}},
    {13106, 2, {0x0, 0x7C00, 0x4F8, 0x408, 0x430, 0x820, 0x3040, 0x0, 0xE022, 0x29, 0xEAA4, 0x2AB0, 0x20A8, 0x4120, 0x8620, 0x0}},
    {13107, 2, {0x0, 0x3E00, 0x208, 0x210, 0x230, 0x450, 0x1810, 0x0, 0x20, 0x20, 0x20, 0x3E30, 0x28, 0x20, 0x20, 0x0}},
    {13108, 2, {0x100, 0xF480, 0x122A, 0x102A, 0x1002, 0x2004, 0xC018, 0x0, 0x4000, 0x2804, 0x8814, 0x4FD4, 0x1115, 0xE115, 0x7D6, 0x0}},
    {13109, 2, {0x38, 0x3E00, 0x27C, 0x204, 0x204, 0x408, 0x1830, 0x0, 0x0, 0x2000, 0x1000, 0x200, 0xC00, 0x3000, 0x0, 0x0}},
    {13110, 2, {0x40, 0x2078, 0x5048, 0x888, 0x408, 0x10, 0x60, 0x0, 0x4000, 0x7804, 0x4814, 0xABD4, 0x1815, 0x1015, 0x6016, 0x0}},
    {13111, 2, {0x100, 0x2280, 0x5280, 0x900, 0x400, 0x0, 0x0, 0x0, 0x0, 0x24, 0x14, 0x14, 0x4, 0x8, 0x30, 0x0}},
    {13112, 2, {0x100, 0x229C, 0x5280, 0x900, 0x400, 0x3E, 0x0, 0x0, 0x2000, 0x2600, 0x3800, 0x2000, 0x2000, 0x2000, 0x1E00, 0x0}},
    {13113, 2, {0x0, 0x1010, 0x2850, 0x450, 0x254, 0x54, 0x58, 0x0, 0x0, 0x2A00, 0x2A00, 0x200, 0x200, 0x400, 0x1800, 0x0}},
    {13114, 2, {0x100, 0x22A0, 0x5290, 0x902, 0x40C, 0x30, 0x0, 0x0, 0x0, 0x3E00, 0x200, 0x400, 0x800, 0x1400, 0x2200, 0x0}},
    {13115, 2, {0x100, 0x2280, 0x5280, 0x93E, 0x400, 0x0, 0x0, 0x0, 0x2040, 0x1520, 0x4480, 0x2400, 0x800, 0x7000, 0x0, 0x0}},
    {13116, 2, {0x100, 0x2480, 0x5200, 0x83E, 0x400, 0x0, 0x0, 0x0, 0x1000, 0x1E00, 0x1200, 0x2A00, 0x600, 0x400, 0x1800, 0x0}},
    {13117, 2, {0x1082, 0x7D44, 0x1148, 0x5498, 0x5428, 0x1008, 0x3008, 0x0, 0x20, 0x2020, 0x1020, 0x230, 0xC28, 0x3020, 0x20, 0x0}},
    {13118, 2, {0x1040, 0x7D24, 0x1094, 0x5414, 0x5415, 0x1015, 0x3016, 0x0, 0x1000, 0x1000, 0x1000, 0x1800, 0x1400, 0x1000, 0x1000, 0x0}},
    {13119, 2, {0x800, 0x3E00, 0x800, 0x2A00, 0x2A00, 0x800, 0x1800, 0x0, 0x0, 0x40, 0x20, 0x4, 0x18, 0x60, 0x0, 0x0}},
    {13120, 2, {0x1080, 0x7D50, 0x1148, 0x5482, 0x540C, 0x1030, 0x3000, 0x0, 0x1080, 0x1240, 0x1100, 0x1800, 0x1400, 0x1000, 0x1000, 0x0}},
    {13121, 2, {0x800, 0x3E00, 0x800, 0x2A7C, 0x2A00, 0x800, 0x1800, 0x0, 0x0, 0x800, 0x2800, 0x2800, 0x2A00, 0x2A00, 0x2C00, 0x0}},
    {13122, 2, {0x800, 0x3E00, 0x800, 0x2A7C, 0x2A00, 0x800, 0x1800, 0x0, 0x0, 0x2000, 0x1000, 0x200, 0xC00, 0x3000, 0x0, 0x0}},
    {13123, 2, {0x4, 0x3E08, 0x210, 0x230, 0x1450, 0x810, 0x410, 0x0, 0x1000, 0x1E7C, 0x1244, 0x244, 0x244, 0x47C, 0x1800, 0x0}},
    {13124, 2, {0x4, 0x3E08, 0x210, 0x230, 0x1450, 0x810, 0x410, 0x0, 0x0, 0x800, 0x2800, 0x2800, 0x2A00, 0x2A00, 0x2C00, 0x0}},
    {13125, 2, {0x0, 0x3E00, 0x254, 0x254, 0x1404, 0x808, 0x430, 0x0, 0x0, 0x0, 0x1400, 0x1200, 0x1200, 0x2200, 0x0, 0x0}},
    {13126, 2, {0x0, 0x3E10, 0x250, 0x250, 0x1454, 0x854, 0x458, 0x0, 0x1000, 0x1E00, 0x1200, 0x2200, 0x200, 0x400, 0x1800, 0x0}},
    {13127, 2, {0x0, 0x7C80, 0x440, 0x408, 0x2830, 0x10C0, 0x800, 0x0, 0x0, 0x4020, 0x2B90, 0x8882, 0x4B8C, 0x10B0, 0xE380, 0x0}},
    {13128, 2, {0x1820, 0x43C, 0x3024, 0x844, 0x4, 0x3008, 0xC30, 0x0, 0x0, 0x3E40, 0x2220, 0x2204, 0x2218, 0x3E60, 0x0, 0x0}},
    {13129, 2, {0x1800, 0x400, 0x3000, 0x800, 0x0, 0x3000, 0xC00, 0x0, 0x48, 0x48, 0x48, 0x48, 0x8, 0x10, 0x60, 0x0}},
    {13130, 2, {0x1848, 0x448, 0x3048, 0x848, 0x8, 0x3010, 0xC60, 0x0, 0x100, 0x484, 0x5214, 0x4814, 0x49D5, 0x8815, 0x16, 0x0}},
    {13131, 2, {0x400, 0x400, 0x400, 0x2800, 0x1000, 0x2800, 0x4000, 0x0, 0x204, 0x7D2, 0x248, 0x240, 0x240, 0x240, 0x480, 0x0}},
    {13132, 2, {0x902, 0xBE9, 0x924, 0x5120, 0x2120, 0x5120, 0x8240, 0x0, 0x2000, 0x2080, 0x2040, 0x3008, 0x2830, 0x20C0, 0x2000, 0x0}},
    {13133, 2, {0x200, 0x200, 0x200, 0x147C, 0x800, 0x1400, 0x2000, 0x0, 0x1000, 0x1010, 0x1050, 0x1850, 0x1454, 0x1054, 0x1058, 0x0}},
    {13134, 2, {0x1000, 0x1000, 0x3E00, 0x127C, 0x1400, 0x1000, 0x1000, 0x0, 0x2100, 0x2480, 0x2200, 0x3000, 0x2800, 0x2000, 0x2000, 0x0}},
    {13135, 2, {0x1000, 0x1000, 0x3E00, 0x127C, 0x1400, 0x1000, 0x1000, 0x0, 0x0, 0x800, 0x2800, 0x2800, 0x2A00, 0x2A00, 0x2C00, 0x0}},
    {13136, 2, {0x7C, 0x1C04, 0x414, 0x418, 0x410, 0x3E10, 0x20, 0x0, 0x0, 0x2000, 0x1000, 0x200, 0xC00, 0x3000, 0x0, 0x0}},
    {13137, 2, {0x2400, 0x2400, 0x2454, 0x2454, 0x404, 0x808, 0x3030, 0x0, 0x1000, 0x1010, 0x1050, 0x1850, 0x1454, 0x1054, 0x1058, 0x0}},
    {13138, 2, {0x2400, 0x2400, 0x2400, 0x2400, 0x400, 0x800, 0x3000, 0x0, 0x38, 0x0, 0x7C, 0x4, 0x4, 0x8, 0x30, 0x0}},
    {13139, 2, {0x1102, 0x5135, 0x51C5, 0x5102, 0x5500, 0x5500, 0x58F0, 0x0, 0x0, 0x0, 0x0, 0x7C00, 0x0, 0x0, 0x0, 0x0}},
    {13140, 2, {0x0, 0x800, 0x2800, 0x287C, 0x2A00, 0x2A00, 0x2C00, 0x0, 0x80, 0xFA44, 0x914, 0x814, 0x815, 0x1015, 0x6016, 0x0}},
    {13141, 2, {0x0, 0x2000, 0x2000, 0x2200, 0x2400, 0x2800, 0x3000, 0x0, 0x10, 0x10, 0x20, 0x40, 0x44, 0x7C, 0x4, 0x0}},
    {13142, 2, {0x0, 0x4080, 0x4040, 0x4408, 0x4830, 0x50C0, 0x6000, 0x0, 0x8820, 0x8E94, 0x9442, 0xC400, 0xA401, 0x8406, 0x8818, 0x0}},
    {13143, 2, {0x0, 0x3E00, 0x2254, 0x2254, 0x204, 0x408, 0x1830, 0x0, 0x1000, 0x1000, 0x1000, 0x1800, 0x1400, 0x1000, 0x1000, 0x0}},
    {13144, 2, {0x10, 0x1810, 0x2410, 0x241E, 0x2410, 0x2410, 0x247C, 0x2444, 0x2444, 0x2444, 0x247C, 0x2444, 0x2400, 0x2454, 0x18AA, 0xAA}},
    {13145, 2, {0x10, 0x1010, 0x3010, 0x101E, 0x1010, 0x1010, 0x107C, 0x1044, 0x1044, 0x1044, 0x107C, 0x1044, 0x1000, 0x1054, 0x7CAA, 0xAA}},
    {13146, 2, {0x10, 0x1810, 0x2410, 0x241E, 0x410, 0x410, 0x87C, 0x844, 0x844, 0x1044, 0x107C, 0x1044, 0x2000, 0x2054, 0x3CAA, 0xAA}},
    {13147, 2, {0x10, 0x1810, 0x2410, 0x241E, 0x410, 0x410, 0x47C, 0x1844, 0x444, 0x444, 0x47C, 0x444, 0x2400, 0x2454, 0x18AA, 0xAA}},
    {13148, 2, {0x10, 0x2410, 0x2410, 0x241E, 0x2410, 0x2410, 0x247C, 0x3C44, 0x444, 0x444, 0x47C, 0x444, 0x400, 0x454, 0x4AA, 0xAA}},
    {13149, 2, {0x10, 0x3C10, 0x2010, 0x201E, 0x2010, 0x2010, 0x207C, 0x3844, 0x444, 0x444, 0x47C, 0x444, 0x2400, 0x2454, 0x18AA, 0xAA}},
    {13150, 2, {0x10, 0x1810, 0x2410, 0x241E, 0x2010, 0x2010, 0x207C, 0x3844, 0x2444, 0x2444, 0x247C, 0x2444, 0x2400, 0x2454, 0x18AA, 0xAA}},
    {13151, 2, {0x10, 0x3C10, 0x410, 0x41E, 0x410, 0x810, 0x87C, 0x844, 0x1044, 0x1044, 0x107C, 0x1044, 0x1000, 0x1054, 0x10AA, 0xAA}},
    {13152, 2, {0x10, 0x1810, 0x2410, 0x241E, 0x2410, 0x2410, 0x247C, 0x1844, 0x2444, 0x2444, 0x247C, 0x2444, 0x2400, 0x2454, 0x18AA, 0xAA}},
    {13153, 2, {0x10, 0x1810, 0x2410, 0x241E, 0x2410, 0x2410, 0x247C, 0x1844, 0x444, 0x444, 0x47C, 0x444, 0x2400, 0x2454, 0x18AA, 0xAA}},
    {13154, 2, {0x10, 0x2610, 0x6910, 0x291E, 0x2910, 0x2910, 0x297C, 0x2944, 0x2944, 0x2944, 0x297C, 0x2944, 0x2900, 0x2954, 0x76AA, 0xAA}},
    {13155, 2, {0x10, 0x2210, 0x6610, 0x221E, 0x2210, 0x2210, 0x227C, 0x2244, 0x2244, 0x2244, 0x227C, 0x2244, 0x2200, 0x2254, 0x77AA, 0xAA}},
    {13156, 2, {0x10, 0x2610, 0x6910, 0x291E, 0x2110, 0x2110, 0x227C, 0x2244, 0x2244, 0x2444, 0x247C, 0x2444, 0x2800, 0x2854, 0x7FAA, 0xAA}},
    {13157, 2, {0x10, 0x2610, 0x6910, 0x291E, 0x2110, 0x2110, 0x217C, 0x2644, 0x2144, 0x2144, 0x217C, 0x2144, 0x2900, 0x2954, 0x76AA, 0xAA}},
    {13158, 2, {0x10, 0x2910, 0x6910, 0x291E, 0x2910, 0x2910, 0x297C, 0x2F44, 0x2144, 0x2144, 0x217C, 0x2144, 0x2100, 0x2154, 0x71AA, 0xAA}},
    {13159, 2, {0x10, 0x2F10, 0x6810, 0x281E, 0x2810, 0x2810, 0x287C, 0x2E44, 0x2144, 0x2144, 0x217C, 0x2144, 0x2900, 0x2954, 0x76AA, 0xAA}},
    {13160, 2, {0x10, 0x2610, 0x6910, 0x291E, 0x2810, 0x2810, 0x287C, 0x2E44, 0x2944, 0x2944, 0x297C, 0x2944, 0x2900, 0x2954, 0x76AA, 0xAA}},
    {13161, 2, {0x10, 0x2F10, 0x6110, 0x211E, 0x2110, 0x2210, 0x227C, 0x2244, 0x2444, 0x2444, 0x247C, 0x2444, 0x2400, 0x2454, 0x74AA, 0xAA}},
    {13162, 2, {0x10, 0x2610, 0x6910, 0x291E, 0x2910, 0x2910, 0x297C, 0x2644, 0x2944, 0x2944, 0x297C, 0x2944, 0x2900, 0x2954, 0x76AA, 0xAA}},
    {13163, 2, {0x10, 0x2610, 0x6910, 0x291E, 0x2910, 0x2910, 0x297C, 0x2644, 0x2144, 0x2144, 0x217C, 0x2144, 0x2900, 0x2954, 0x76AA, 0xAA}},
    {13164, 2, {0x8, 0x6308, 0x9488, 0x948F, 0x1488, 0x1488, 0x24BE, 0x24A2, 0x24A2, 0x44A2, 0x44BE, 0x44A2, 0x8480, 0x84AA, 0xF355, 0x55}},
    {13165, 2, {0x8, 0x6208, 0x9608, 0x920F, 0x1208, 0x1208, 0x223E, 0x2222, 0x2222, 0x4222, 0x423E, 0x4222, 0x8200, 0x822A, 0xF755, 0x55}},
    {13166, 2, {0x8, 0x6308, 0x9488, 0x948F, 0x1088, 0x1088, 0x213E, 0x2122, 0x2122, 0x4222, 0x423E, 0x4222, 0x8400, 0x842A, 0xF7D5, 0x55}},
    {13167, 2, {0x8, 0x6308, 0x9488, 0x948F, 0x1088, 0x1088, 0x20BE, 0x2322, 0x20A2, 0x40A2, 0x40BE, 0x40A2, 0x8480, 0x84AA, 0xF355, 0x55}},
    {13168, 2, {0x8, 0x6488, 0x9488, 0x948F, 0x1488, 0x1488, 0x24BE, 0x27A2, 0x20A2, 0x40A2, 0x40BE, 0x40A2, 0x8080, 0x80AA, 0xF0D5, 0x55}},
    {13169, 2, {0x0, 0x0, 0x43C0, 0x4220, 0x4220, 0x4220, 0x53CC, 0x6A12, 0x4A02, 0x4A0E, 0x4A12, 0x4A12, 0x4A16, 0x4A0A, 0x0, 0x0}},
    {13170, 2, {0x0, 0x0, 0x200, 0x200, 0x200, 0x200, 0x3A3C, 0x4642, 0x4202, 0x423E, 0x4242, 0x4242, 0x4646, 0x3A3A, 0x0, 0x0}},
    {13171, 2, {0x0, 0x0, 0x842, 0x1442, 0x1442, 0x1442, 0x2242, 0x2242, 0x3E42, 0x2242, 0x4142, 0x4142, 0x4142, 0x413C, 0x0, 0x0}},
    {13172, 2, {0x0, 0x0, 0x4000, 0x4000, 0x4000, 0x4000, 0x7194, 0x4A5A, 0x4852, 0x49D0, 0x4A50, 0x4A50, 0x4AD0, 0x7150, 0x0, 0x0}},
    {13173, 2, {0x0, 0x0, 0x82, 0x82, 0x82, 0x82, 0x7882, 0x8444, 0x8444, 0x8444, 0x8428, 0x8428, 0x8410, 0x7810, 0x0, 0x0}},
    {13174, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C3C, 0x6242, 0x4240, 0x4240, 0x4240, 0x4240, 0x6242, 0x5C3C, 0x4000, 0x4000}},
    {13179, 2, {0x0, 0xA, 0x9, 0xFE09, 0x1008, 0x927F, 0x5248, 0x5448, 0x1078, 0xFE55, 0x1055, 0x1095, 0x10D6, 0x10B5, 0x110A, 0x10}},
    {13180, 2, {0x0, 0x1E30, 0xAEE, 0xEA4A, 0xAAFA, 0xAA4A, 0xEA4A, 0xB6CA, 0xA0EA, 0xAEDA, 0xEB4A, 0xAA4A, 0xE4E, 0xA4A, 0x0, 0x0}},
    {13181, 2, {0x0, 0x0, 0x10FE, 0x1010, 0x1010, 0xFE50, 0x1050, 0x1050, 0x285E, 0x2850, 0x2850, 0x4450, 0x4450, 0x4450, 0x82FE, 0x0}},
    {13182, 2, {0x0, 0x1, 0xE81, 0xEA42, 0xAA02, 0xAA94, 0xEE55, 0xAA17, 0xAA21, 0xAA20, 0xEE6F, 0xAA29, 0xA29, 0x122F, 0x2629, 0x0}},
    {13183, 2, {0x2A05, 0x7F84, 0x2A3F, 0x7702, 0xA23A, 0x2712, 0x2AB9, 0x0, 0x1044, 0x28F4, 0x441E, 0xBA24, 0xD4, 0xFE44, 0x2444, 0x7A5F}},
    {13184, 2, {0x0, 0x0, 0x10, 0x28, 0x28, 0x28, 0x5C44, 0x6244, 0x427C, 0x4244, 0x4282, 0x6282, 0x5C82, 0x4082, 0x4000, 0x4000}},
    {13185, 2, {0x0, 0x0, 0x10, 0x28, 0x28, 0x28, 0x5844, 0x6444, 0x447C, 0x4444, 0x4482, 0x4482, 0x4482, 0x4482, 0x0, 0x0}},
    {13186, 2, {0x0, 0x0, 0x10, 0x28, 0x28, 0x28, 0x1144, 0x1144, 0x117C, 0x1244, 0x2282, 0x2282, 0x2282, 0x3D82, 0x4000, 0x4000}},
    {13187, 2, {0x0, 0x0, 0x10, 0x28, 0x28, 0x28, 0x6844, 0x5444, 0x547C, 0x5444, 0x5482, 0x5482, 0x5482, 0x5482, 0x0, 0x0}},
    {13188, 2, {0x0, 0x0, 0x8010, 0x8028, 0x8028, 0x8028, 0x8444, 0x8844, 0x907C, 0xA044, 0xD082, 0x8882, 0x8482, 0x8282, 0x0, 0x0}},
    {13189, 2, {0x0, 0x0, 0x407C, 0x4042, 0x4042, 0x4042, 0x4242, 0x447C, 0x4842, 0x5042, 0x6842, 0x4442, 0x4242, 0x417C, 0x0, 0x0}},
    {13190, 2, {0x0, 0x0, 0x417C, 0x4142, 0x6342, 0x6342, 0x5542, 0x557C, 0x4942, 0x4942, 0x4142, 0x4142, 0x4142, 0x417C, 0x0, 0x0}},
    {13191, 2, {0x0, 0x0, 0x3C7C, 0x4242, 0x4242, 0x4042, 0x4042, 0x4E7C, 0x4242, 0x4242, 0x4242, 0x4242, 0x4642, 0x3A7C, 0x0, 0x0}},
    {13192, 2, {0x0, 0x0, 0xC, 0x4, 0x4, 0x4, 0x71E4, 0x8A14, 0x8014, 0x81F4, 0x8214, 0x8214, 0x8A34, 0x71DE, 0x0, 0x0}},
    {13193, 2, {0x0, 0x0, 0x800C, 0x8004, 0x8004, 0x8004, 0x9364, 0x9494, 0xA414, 0xC474, 0xA494, 0x9494, 0x8CB4, 0x8F5E, 0x0, 0x0}},
    {13194, 2, {0x0, 0x0, 0xFE, 0x80, 0x80, 0x80, 0x5C80, 0x62F8, 0x4280, 0x4280, 0x4280, 0x6280, 0x5C80, 0x4080, 0x4000, 0x4000}},
    {13195, 2, {0x0, 0x0, 0xFE, 0x80, 0x80, 0x80, 0x58F8, 0x6480, 0x4480, 0x4480, 0x4480, 0x4480, 0x4480, 0x4480, 0x0, 0x0}},
    {13196, 2, {0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x227C, 0x2240, 0x2240, 0x2240, 0x4440, 0x4440, 0x4440, 0x7B40, 0x8000, 0x8000}},
    {13197, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x1134, 0x1148, 0x1148, 0x1148, 0x2230, 0x2220, 0x2238, 0x3DC4, 0x4044, 0x4038}},
    {13198, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x6868, 0x5490, 0x5490, 0x5490, 0x5460, 0x5440, 0x5470, 0x5488, 0x88, 0x70}},
    {13199, 2, {0x0, 0x0, 0x4000, 0x4000, 0x4000, 0x4004, 0x4234, 0x4448, 0x4848, 0x5048, 0x6830, 0x4420, 0x4238, 0x4144, 0x44, 0x38}},
    {13200, 2, {0x0, 0x4100, 0x4100, 0x4100, 0x4100, 0x4100, 0x413E, 0x7F02, 0x4104, 0x4108, 0x4110, 0x4120, 0x4140, 0x417E, 0x0, 0x0}},
    {13201, 2, {0x0, 0x8440, 0x8440, 0x8440, 0x8440, 0x8440, 0x945E, 0x97C2, 0xA444, 0xC448, 0xA448, 0x9450, 0x8C60, 0x8C7E, 0x0, 0x0}},
    {13202, 2, {0x0, 0x8A40, 0x8A40, 0xDA40, 0xDA40, 0xAA40, 0xAA5E, 0x8BC2, 0x8A44, 0x8A44, 0x8A48, 0x8A48, 0x8A50, 0x8A5E, 0x0, 0x0}},
    {13203, 2, {0x0, 0x0, 0x7240, 0x8A40, 0x8A40, 0x8240, 0x825E, 0x83C2, 0x9E44, 0x8A44, 0x8A48, 0x8A48, 0x8A50, 0x7A5E, 0x0, 0x0}},
    {13204, 2, {0x0, 0x0, 0xFC80, 0x2480, 0x2480, 0x2480, 0x24BE, 0x2782, 0x2484, 0x2488, 0x2488, 0x2490, 0x24A0, 0x24BE, 0x0, 0x0}},
    {13205, 2, {0x0, 0x0, 0x8, 0x14, 0x14, 0x24, 0x1128, 0x1148, 0x1148, 0x2250, 0x2252, 0x2262, 0x2264, 0x5DD8, 0x4000, 0x4000}},
    {13206, 2, {0x0, 0x0, 0x8, 0x14, 0x14, 0x24, 0x7628, 0x4948, 0x4948, 0x4950, 0x4952, 0x4962, 0x4964, 0x48D8, 0x0, 0x0}},
    {13207, 2, {0x0, 0x0, 0x208, 0x214, 0x214, 0x224, 0x3A28, 0x4648, 0x4248, 0x4250, 0x4252, 0x4262, 0x4664, 0x3BD8, 0x0, 0x0}},
    {13208, 2, {0x0, 0x0, 0x4008, 0x4014, 0x4014, 0x4024, 0x4228, 0x4448, 0x4848, 0x5050, 0x6852, 0x4462, 0x4264, 0x41D8, 0x0, 0x0}},
    {13209, 2, {0x0, 0x0, 0x1C00, 0x2200, 0x2000, 0x2000, 0x78C8, 0x2134, 0x2124, 0x2124, 0x2124, 0x2124, 0x2124, 0x2124, 0x0, 0x0}},
    {13210, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x59D8, 0x6524, 0x4524, 0x4524, 0x4524, 0x4524, 0x4524, 0x4524, 0x0, 0x0}},
    {13211, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22A4, 0x22DA, 0x2292, 0x4292, 0x4492, 0x4492, 0x4492, 0x7B92, 0x8000, 0x8000}},
    {13212, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3468, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x2A54, 0x0, 0x0}},
    {13213, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C68, 0x2254, 0x4054, 0x4054, 0x4054, 0x4054, 0x2254, 0x1C54, 0x0, 0x0}},
    {13214, 2, {0x0, 0x0, 0x4000, 0x4000, 0x4000, 0x4000, 0x4268, 0x4454, 0x4854, 0x5054, 0x6854, 0x4454, 0x4254, 0x4154, 0x0, 0x0}},
    {13215, 2, {0x1E, 0x2, 0x1E, 0x10, 0x1E, 0x0, 0x68D0, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x0, 0x0}},
    {13216, 2, {0x1E, 0x2, 0x1E, 0x10, 0x1E, 0x0, 0x38D0, 0x44A8, 0x80A8, 0x80A8, 0x80A8, 0x80A8, 0x44A8, 0x38A8, 0x0, 0x0}},
    {13217, 2, {0x3C, 0x4, 0x3C, 0x20, 0x3C, 0x0, 0x2CC0, 0x3320, 0x2220, 0x2220, 0x2220, 0x2220, 0x2220, 0x2220, 0x0, 0x0}},
    {13218, 2, {0x1E, 0x2, 0x801E, 0x8010, 0x801E, 0x8000, 0x84D0, 0x88A8, 0x90A8, 0xA0A8, 0xD0A8, 0x88A8, 0x84A8, 0x82A8, 0x0, 0x0}},
    {13219, 2, {0x1E, 0x2, 0xE, 0x2, 0x1E, 0x0, 0x68D0, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x0, 0x0}},
    {13220, 2, {0x1E, 0x2, 0xE, 0x2, 0x1E, 0x0, 0x38D0, 0x44A8, 0x80A8, 0x80A8, 0x80A8, 0x80A8, 0x44A8, 0x38A8, 0x0, 0x0}},
    {13221, 2, {0x3C, 0x4, 0x1C, 0x4, 0x3C, 0x0, 0x2CC0, 0x3320, 0x2220, 0x2220, 0x2220, 0x2220, 0x2220, 0x2220, 0x0, 0x0}},
    {13222, 2, {0x1E, 0x2, 0x800E, 0x8002, 0x801E, 0x8000, 0x84D0, 0x88A8, 0x90A8, 0xA0A8, 0xD0A8, 0x88A8, 0x84A8, 0x82A8, 0x0, 0x0}},
    {13223, 2, {0x2, 0xD004, 0xA808, 0xA810, 0xA820, 0xA840, 0xA880, 0xA978, 0xAA84, 0x280, 0x478, 0x804, 0x1084, 0x2078, 0x4000, 0x8000}},
    {13224, 2, {0x10, 0xD02E, 0xA822, 0xA84E, 0xA888, 0xA88E, 0xA900, 0xAA78, 0xAC84, 0x480, 0x878, 0x1004, 0x1084, 0x2078, 0x4000, 0x8000}},
    {13225, 2, {0x0, 0x0, 0x7C00, 0x4200, 0x4200, 0x4200, 0x42F0, 0x7D08, 0x4008, 0x40F8, 0x4108, 0x4108, 0x4118, 0x40E8, 0x0, 0x0}},
    {13226, 2, {0x0, 0x0, 0x81C0, 0x8120, 0x8120, 0x8120, 0x853C, 0x89C2, 0x9102, 0xA11E, 0xD122, 0x8922, 0x8526, 0x831A, 0x0, 0x0}},
    {13227, 2, {0x0, 0x0, 0x89C0, 0x8920, 0xD920, 0xD920, 0xA93C, 0xA9C2, 0x8902, 0x891E, 0x8922, 0x8922, 0x8926, 0x891A, 0x0, 0x0}},
    {13228, 2, {0x0, 0x0, 0x71C0, 0x8920, 0x8920, 0x8120, 0x813C, 0x81C2, 0x9D02, 0x891E, 0x8922, 0x8922, 0x8926, 0x791A, 0x0, 0x0}},
    {13229, 2, {0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x5B1A, 0x64A6, 0x40A2, 0x47A2, 0x48A2, 0x48A2, 0x49A6, 0x46DA, 0x0, 0x0}},
    {13230, 2, {0x82, 0x84, 0x84, 0xB188, 0xCA88, 0x9A90, 0xAA90, 0xAAA0, 0xAAAC, 0x9DD2, 0x90, 0x10C, 0x102, 0x212, 0x40C, 0x800}},
    {13231, 2, {0x90, 0x9E, 0x92, 0xB1AE, 0xCAA8, 0x9AAE, 0xAAC0, 0xAAC0, 0xAA98, 0x9DA4, 0xA0, 0x118, 0x104, 0x224, 0x218, 0x400}},
    {13232, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5C78, 0x6284, 0x4280, 0x4270, 0x4208, 0x4204, 0x6284, 0x5C78, 0x4000, 0x4000}},
    {13233, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5878, 0x6484, 0x4480, 0x4470, 0x4408, 0x4404, 0x4484, 0x4478, 0x0, 0x0}},
    {13234, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x113C, 0x1142, 0x1140, 0x1138, 0x2204, 0x2202, 0x2242, 0x3DBC, 0x4000, 0x4000}},
    {13235, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6878, 0x5484, 0x5480, 0x5470, 0x5408, 0x5404, 0x5484, 0x5478, 0x0, 0x0}},
    {13236, 2, {0x0, 0x0, 0x104, 0x104, 0x104, 0x104, 0xB904, 0xC488, 0x8488, 0x8488, 0x8450, 0xC450, 0xB820, 0x8020, 0x8000, 0x8000}},
    {13237, 2, {0x0, 0x0, 0x104, 0x104, 0x104, 0x104, 0x5888, 0x6488, 0x4488, 0x4450, 0x4450, 0x4420, 0x4420, 0x4420, 0x0, 0x0}},
    {13238, 2, {0x0, 0x0, 0x82, 0x82, 0x82, 0x82, 0x2282, 0x2244, 0x2244, 0x2244, 0x4428, 0x4428, 0x4410, 0x7B10, 0x8000, 0x8000}},
    {13239, 2, {0x0, 0x0, 0x82, 0x82, 0x82, 0x82, 0x3482, 0x2A44, 0x2A44, 0x2A44, 0x2A28, 0x2A28, 0x2A10, 0x2A10, 0x0, 0x0}},
    {13240, 2, {0x0, 0x0, 0x8104, 0x8104, 0x8104, 0x8104, 0x8504, 0x8888, 0x9088, 0xA088, 0xD050, 0x8850, 0x8420, 0x8220, 0x0, 0x0}},
    {13241, 2, {0x0, 0x0, 0x8282, 0x8282, 0xC682, 0xC682, 0xAA82, 0xAA44, 0x9244, 0x9244, 0x8228, 0x8228, 0x8210, 0x8210, 0x0, 0x0}},
    {13242, 2, {0x0, 0x0, 0x222, 0x222, 0x222, 0x122, 0x5914, 0x6514, 0x4494, 0x4494, 0x44B4, 0x64C8, 0x5848, 0x4048, 0x4000, 0x4000}},
    {13243, 2, {0x0, 0x0, 0x222, 0x222, 0x222, 0x122, 0x5914, 0x6514, 0x4494, 0x4494, 0x44B4, 0x44C8, 0x4448, 0x4448, 0x0, 0x0}},
    {13244, 2, {0x0, 0x0, 0x222, 0x222, 0x222, 0x122, 0x2514, 0x2514, 0x2494, 0x2494, 0x48B4, 0x48C8, 0x4848, 0x7648, 0x8000, 0x8000}},
    {13245, 2, {0x0, 0x0, 0x222, 0x222, 0x222, 0x122, 0x6914, 0x5514, 0x5494, 0x5494, 0x54B4, 0x54C8, 0x5448, 0x5448, 0x0, 0x0}},
    {13246, 2, {0x0, 0x0, 0x8222, 0x8222, 0x8222, 0x8122, 0x8514, 0x8914, 0x9094, 0xA094, 0xD0B4, 0x88C8, 0x8448, 0x8348, 0x0, 0x0}},
    {13247, 2, {0x0, 0x0, 0x8322, 0x8322, 0xC722, 0xC722, 0xAA94, 0xAA94, 0x9294, 0x9294, 0x82B4, 0x8248, 0x8248, 0x8248, 0x0, 0x0}},
    {13248, 2, {0x0, 0x0, 0x4070, 0x4088, 0x4104, 0x4104, 0x4904, 0x4904, 0x5104, 0x6104, 0x5088, 0x4888, 0x4652, 0x47DE, 0x0, 0x0}},
    {13249, 2, {0x0, 0x0, 0x8870, 0x8888, 0xD904, 0xD904, 0xA904, 0xA904, 0x8904, 0x8904, 0x8888, 0x8888, 0x8A52, 0x8BDE, 0x0, 0x0}},
    {13250, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30D0, 0x48A8, 0x8A8, 0x38A8, 0x48A8, 0x48A8, 0x58A8, 0x2AAA, 0x0, 0x0}},
    {13251, 2, {0x0, 0x0, 0x3C00, 0x2200, 0x2200, 0x2200, 0x2234, 0x3C4C, 0x2244, 0x2144, 0x2144, 0x214C, 0x2234, 0x3C04, 0x4, 0x4}},
    {13252, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C38, 0x2244, 0x4080, 0x4080, 0x4080, 0x4080, 0x2244, 0x1C38, 0x0, 0x0}},
    {13253, 2, {0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C74, 0x428C, 0x4084, 0x4084, 0x4084, 0x4084, 0x428C, 0x3C74, 0x0, 0x0}},
    {13254, 2, {0x10, 0x7020, 0x8840, 0x8080, 0x8100, 0x8A00, 0x7601, 0xA4D, 0x1252, 0x2292, 0x4312, 0x828C, 0x248, 0x24E, 0x11, 0xE}},
    {13255, 2, {0x0, 0x0, 0x3C00, 0x4200, 0x8100, 0x8000, 0x8070, 0x8088, 0x8088, 0x8088, 0x8088, 0x8188, 0x4288, 0x3C72, 0x0, 0x0}},
    {13256, 2, {0x0, 0x0, 0x4F8, 0x484, 0x484, 0x484, 0x7484, 0x8CF8, 0x8484, 0x8482, 0x8482, 0x8482, 0x8C84, 0x74F8, 0x0, 0x0}},
    {13257, 2, {0x0, 0x0, 0x1E00, 0x2100, 0x4100, 0x4000, 0x4042, 0x4042, 0x4022, 0x43A4, 0x4114, 0x4118, 0x2308, 0x1D10, 0x10, 0x60}},
    {13258, 2, {0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2C78, 0x3284, 0x2204, 0x227C, 0x2284, 0x2284, 0x228C, 0x2274, 0x0, 0x0}},
    {13259, 2, {0x0, 0x0, 0x427C, 0x4242, 0x4242, 0x4242, 0x4242, 0x7E7C, 0x4240, 0x4240, 0x4240, 0x4240, 0x4240, 0x4240, 0x0, 0x0}},
    {13260, 2, {0x0, 0x0, 0x0, 0x1000, 0x1000, 0x0, 0x30B8, 0x10C4, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x7C84, 0x0, 0x0}},
    {13261, 2, {0x0, 0x0, 0x8282, 0x8484, 0x8888, 0x9090, 0xA0A0, 0xC0C0, 0xC0C0, 0xA0A0, 0x9090, 0x8888, 0x8484, 0x8282, 0x0, 0x0}},
    {13262, 2, {0x0, 0x0, 0x4504, 0x4504, 0x458C, 0x498C, 0x5154, 0x6154, 0x5124, 0x4924, 0x4504, 0x4504, 0x4504, 0x4504, 0x0, 0x0}},
    {13263, 2, {0x0, 0x0, 0x4000, 0x4020, 0x4020, 0x4020, 0x4278, 0x4420, 0x4820, 0x5020, 0x6820, 0x4424, 0x4224, 0x4118, 0x0, 0x0}},
    {13264, 2, {0x0, 0x0, 0x6000, 0x2000, 0x2000, 0x2000, 0x2000, 0x21D8, 0x2124, 0x2124, 0x2124, 0x2124, 0x2124, 0xF924, 0x0, 0x0}},
    {13265, 2, {0x0, 0x0, 0x6000, 0x2000, 0x2000, 0x2000, 0x2178, 0x2184, 0x2104, 0x2104, 0x2104, 0x2104, 0xF904, 0x0, 0x0, 0x0}},
    {13266, 2, {0x0, 0x0, 0x6000, 0x2000, 0x2000, 0x2002, 0x271A, 0x28A4, 0x28A4, 0x28A4, 0x2898, 0x2890, 0x289C, 0xF722, 0x22, 0x1C}},
    {13267, 2, {0x0, 0x6000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2204, 0x2108, 0x2090, 0x2060, 0x2060, 0x2090, 0x2108, 0xFA04, 0x0, 0x0}},
    {13268, 2, {0x0, 0x0, 0x80, 0x80, 0x80, 0x80, 0x34B8, 0x2AC4, 0x2A84, 0x2A84, 0x2A84, 0x2A84, 0x2AC4, 0x2AB8, 0x0, 0x0}},
    {13269, 2, {0x0, 0x0, 0x18, 0x88, 0x88, 0x8, 0x6988, 0x5488, 0x5488, 0x5488, 0x5488, 0x5488, 0x5488, 0x57FE, 0x0, 0x0}},
    {13270, 2, {0x0, 0x0, 0xC, 0x4, 0x4, 0x4, 0xD1C4, 0xAA24, 0xAA24, 0xAA24, 0xAA24, 0xAA24, 0xAA24, 0xA9DF, 0x0, 0x0}},
    {13271, 2, {0x0, 0x0, 0x7C42, 0x4242, 0x4242, 0x4242, 0x4242, 0x7C7E, 0x4042, 0x4042, 0x4042, 0x4042, 0x4042, 0x4042, 0x0, 0x0}},
    {13272, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x58D0, 0x64A8, 0x44A8, 0x44A8, 0x44A8, 0x78A8, 0x40A8, 0x44AA, 0x4000, 0x4000}},
    {13273, 2, {0x0, 0x0, 0xE722, 0x94A2, 0x94B6, 0x94B6, 0x94AA, 0xE72A, 0x8422, 0x8422, 0x8422, 0x8422, 0x8422, 0x8422, 0x0, 0x0}},
    {13274, 2, {0x0, 0x0, 0x7C7C, 0x4242, 0x4242, 0x4242, 0x4242, 0x7C7C, 0x4048, 0x4048, 0x4044, 0x4044, 0x4042, 0x4042, 0x0, 0x0}},
    {13275, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C2C, 0x4232, 0x4020, 0x3820, 0x420, 0x220, 0x4220, 0x3C20, 0x0, 0x0}},
    {13276, 2, {0x0, 0x0, 0x1E00, 0x2100, 0x4000, 0x4000, 0x3082, 0xC82, 0x344, 0xC4, 0xA8, 0x40A8, 0x2110, 0x1E10, 0x0, 0x0}},
    {13277, 2, {0x0, 0x0, 0x88A0, 0x88A0, 0x88A0, 0x88A0, 0x44BC, 0x4522, 0x4D22, 0x4D22, 0x5522, 0x2222, 0x2222, 0x223C, 0x0, 0x0}},
    {13280, 2, {0x0, 0x107C, 0x3044, 0x1044, 0x1044, 0x1044, 0x1044, 0x107C, 0x1044, 0x1044, 0x1044, 0x1044, 0x1044, 0x107C, 0x7C44, 0x44}},
    {13281, 2, {0x0, 0x187C, 0x2444, 0x2444, 0x444, 0x444, 0x844, 0x87C, 0x844, 0x1044, 0x1044, 0x1044, 0x2044, 0x207C, 0x3C44, 0x44}},
    {13282, 2, {0x0, 0x187C, 0x2444, 0x2444, 0x444, 0x444, 0x444, 0x187C, 0x444, 0x444, 0x444, 0x444, 0x2444, 0x247C, 0x1844, 0x44}},
    {13283, 2, {0x0, 0x247C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x3C7C, 0x444, 0x444, 0x444, 0x444, 0x444, 0x47C, 0x444, 0x44}},
    {13284, 2, {0x0, 0x3C7C, 0x2044, 0x2044, 0x2044, 0x2044, 0x2044, 0x387C, 0x444, 0x444, 0x444, 0x444, 0x2444, 0x247C, 0x1844, 0x44}},
    {13285, 2, {0x0, 0x187C, 0x2444, 0x2444, 0x2044, 0x2044, 0x2044, 0x387C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x247C, 0x1844, 0x44}},
    {13286, 2, {0x0, 0x3C7C, 0x444, 0x444, 0x444, 0x844, 0x844, 0x87C, 0x1044, 0x1044, 0x1044, 0x1044, 0x1044, 0x107C, 0x1044, 0x44}},
    {13287, 2, {0x0, 0x187C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x187C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x247C, 0x1844, 0x44}},
    {13288, 2, {0x0, 0x187C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x187C, 0x444, 0x444, 0x444, 0x444, 0x2444, 0x247C, 0x1844, 0x44}},
    {13289, 2, {0x0, 0x267C, 0x6944, 0x2944, 0x2944, 0x2944, 0x2944, 0x297C, 0x2944, 0x2944, 0x2944, 0x2944, 0x2944, 0x297C, 0x7644, 0x44}},
    {13290, 2, {0x0, 0x227C, 0x6644, 0x2244, 0x2244, 0x2244, 0x2244, 0x227C, 0x2244, 0x2244, 0x2244, 0x2244, 0x2244, 0x227C, 0x7744, 0x44}},
    {13291, 2, {0x0, 0x267C, 0x6944, 0x2944, 0x2144, 0x2144, 0x2244, 0x227C, 0x2244, 0x2444, 0x2444, 0x2444, 0x2844, 0x287C, 0x7F44, 0x44}},
    {13292, 2, {0x0, 0x267C, 0x6944, 0x2944, 0x2144, 0x2144, 0x2144, 0x267C, 0x2144, 0x2144, 0x2144, 0x2144, 0x2944, 0x297C, 0x7644, 0x44}},
    {13293, 2, {0x0, 0x297C, 0x6944, 0x2944, 0x2944, 0x2944, 0x2944, 0x2F7C, 0x2144, 0x2144, 0x2144, 0x2144, 0x2144, 0x217C, 0x7144, 0x44}},
    {13294, 2, {0x0, 0x2F7C, 0x6844, 0x2844, 0x2844, 0x2844, 0x2844, 0x2E7C, 0x2144, 0x2144, 0x2144, 0x2144, 0x2944, 0x297C, 0x7644, 0x44}},
    {13295, 2, {0x0, 0x267C, 0x6944, 0x2944, 0x2844, 0x2844, 0x2844, 0x2E7C, 0x2944, 0x2944, 0x2944, 0x2944, 0x2944, 0x297C, 0x7644, 0x44}},
    {13296, 2, {0x0, 0x2F7C, 0x6144, 0x2144, 0x2144, 0x2244, 0x2244, 0x227C, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x247C, 0x7444, 0x44}},
    {13297, 2, {0x0, 0x267C, 0x6944, 0x2944, 0x2944, 0x2944, 0x2944, 0x267C, 0x2944, 0x2944, 0x2944, 0x2944, 0x2944, 0x297C, 0x7644, 0x44}},
    {13298, 2, {0x0, 0x267C, 0x6944, 0x2944, 0x2944, 0x2944, 0x2944, 0x267C, 0x2144, 0x2144, 0x2144, 0x2144, 0x2944, 0x297C, 0x7644, 0x44}},
    {13299, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x14A4, 0x14A4, 0x24A4, 0x24BC, 0x24A4, 0x44A4, 0x44A4, 0x44A4, 0x84A4, 0x84BC, 0xF324, 0x24}},
    {13300, 2, {0x0, 0x623C, 0x9624, 0x9224, 0x1224, 0x1224, 0x2224, 0x223C, 0x2224, 0x4224, 0x4224, 0x4224, 0x8224, 0x823C, 0xF724, 0x24}},
    {13301, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x10A4, 0x10A4, 0x2124, 0x213C, 0x2124, 0x4224, 0x4224, 0x4224, 0x8424, 0x843C, 0xF7A4, 0x24}},
    {13302, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x10A4, 0x10A4, 0x20A4, 0x233C, 0x20A4, 0x40A4, 0x40A4, 0x40A4, 0x84A4, 0x84BC, 0xF324, 0x24}},
    {13303, 2, {0x0, 0x64BC, 0x94A4, 0x94A4, 0x14A4, 0x14A4, 0x24A4, 0x27BC, 0x20A4, 0x40A4, 0x40A4, 0x40A4, 0x80A4, 0x80BC, 0xF0A4, 0x24}},
    {13304, 2, {0x0, 0x67BC, 0x9424, 0x9424, 0x1424, 0x1424, 0x2424, 0x273C, 0x20A4, 0x40A4, 0x40A4, 0x40A4, 0x84A4, 0x84BC, 0xF324, 0x24}},
    {13305, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x1424, 0x1424, 0x2424, 0x273C, 0x24A4, 0x44A4, 0x44A4, 0x44A4, 0x84A4, 0x84BC, 0xF324, 0x24}},
    {13306, 2, {0x0, 0x67BC, 0x90A4, 0x90A4, 0x10A4, 0x1124, 0x2124, 0x213C, 0x2224, 0x4224, 0x4224, 0x4224, 0x8224, 0x823C, 0xF224, 0x24}},
    {13307, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x14A4, 0x14A4, 0x24A4, 0x233C, 0x24A4, 0x44A4, 0x44A4, 0x44A4, 0x84A4, 0x84BC, 0xF324, 0x24}},
    {13308, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x14A4, 0x14A4, 0x24A4, 0x233C, 0x20A4, 0x40A4, 0x40A4, 0x40A4, 0x84A4, 0x84BC, 0xF324, 0x24}},
    {13309, 2, {0x0, 0x633C, 0x94A4, 0x94A4, 0x14A4, 0x14A4, 0x14A4, 0x64BC, 0x14A4, 0x14A4, 0x14A4, 0x14A4, 0x94A4, 0x94BC, 0x6324, 0x24}},
    {13310, 2, {0x0, 0x623C, 0x9624, 0x9224, 0x1224, 0x1224, 0x1224, 0x623C, 0x1224, 0x1224, 0x1224, 0x1224, 0x9224, 0x923C, 0x6724, 0x24}},
    {13312, 2, {0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {13313, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2288, 0x2248, 0x2428, 0x2828, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13314, 2, {0x100, 0x100, 0x1F8, 0x3F00, 0x104, 0x104, 0xFC, 0x1000, 0x1020, 0x1020, 0x163C, 0xF8E0, 0x1022, 0x1122, 0x161E, 0x1800}},
    {13315, 2, {0x100, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0x4004, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {13316, 2, {0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x2100, 0x2100, 0x4100, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {13317, 2, {0x0, 0x10, 0x10, 0x2020, 0x1020, 0x840, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0x4000, 0x0}},
    {13318, 2, {0x10, 0xF8, 0x1F00, 0x1000, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x1000, 0x1FFC, 0x1004, 0x2004, 0x2004, 0x4028, 0x8010}},
    {13319, 2, {0x200, 0x200, 0x3FE0, 0x420, 0x420, 0x822, 0x3022, 0xC01E, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13320, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x3FF8, 0x10, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC, 0x0}},
    {13321, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13322, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x0, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13323, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13324, 2, {0x800, 0x800, 0x1FFC, 0x2000, 0x4100, 0x8100, 0x1170, 0x1390, 0x1D10, 0xF110, 0x1150, 0x1122, 0x1102, 0x1002, 0xFFE, 0x0}},
    {13325, 2, {0xFC, 0x7F00, 0x208, 0x2108, 0x1110, 0x1020, 0x0, 0x3FF8, 0x10, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC, 0x0}},
    {13326, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE00E, 0x0, 0x3FF8, 0x10, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13327, 2, {0x0, 0x3FFC, 0x100, 0x100, 0x1FF0, 0x210, 0x410, 0x7FFE, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13328, 2, {0x1000, 0x1000, 0xFEFC, 0x2284, 0x4488, 0x2850, 0x1020, 0x2858, 0xC586, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13329, 2, {0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE18E, 0xC40, 0x300, 0x80, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13330, 2, {0x0, 0x7FF8, 0x408, 0x850, 0x3020, 0xDFF8, 0x1008, 0x1008, 0x1FF8, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13331, 2, {0x0, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x50, 0x20, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13332, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13333, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13334, 2, {0x820, 0x820, 0x820, 0xFFA0, 0x820, 0x820, 0x7F20, 0x20, 0x20, 0x7F20, 0x4120, 0x4122, 0x4122, 0x7F22, 0x411E, 0x0}},
    {13335, 2, {0x208, 0x3FD0, 0x220, 0xFFFE, 0x200, 0xC70, 0x3F84, 0xC804, 0x7FC, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13336, 2, {0x100, 0x1110, 0x920, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13337, 2, {0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13338, 2, {0x7FFC, 0x200, 0x3FF0, 0x410, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13339, 2, {0x840, 0x840, 0x17FC, 0x3080, 0x5100, 0x93F8, 0x1508, 0x1108, 0x11F8, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13340, 2, {0x1040, 0x1440, 0x1240, 0x1040, 0xFEF8, 0x1048, 0x1048, 0x9248, 0x5448, 0x1048, 0x3848, 0x548A, 0x928A, 0x110A, 0x5206, 0x2400}},
    {13341, 2, {0x1100, 0x1100, 0xF9F8, 0x1208, 0x19E8, 0x3128, 0xD1E8, 0x1008, 0x5050, 0x2020, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13342, 2, {0x820, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1010, 0x2FE8, 0x4044, 0x8382, 0xC08, 0x1008, 0xFF8}},
    {13343, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {13344, 2, {0x7A0, 0xF820, 0x48A0, 0x2520, 0x7F20, 0x220, 0x7A2, 0xFC22, 0x4422, 0xC1E, 0x0, 0x3FF0, 0x1C0, 0x1E04, 0x2004, 0x1FFC}},
    {13345, 2, {0x2000, 0x27FC, 0x2040, 0x23F8, 0xFAA8, 0x2AA8, 0x2AA8, 0x2A18, 0x2840, 0x2FFC, 0x28A0, 0x2912, 0x2A0A, 0x4802, 0x47FE, 0x8000}},
    {13346, 2, {0x203C, 0x17C0, 0x8244, 0x4928, 0x13F8, 0xE010, 0x27FE, 0x2020, 0x20A0, 0x40, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {13347, 2, {0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2080, 0x27F0, 0x2490, 0x27F0, 0x2490, 0x3FFC, 0x2080, 0x2FF8, 0x4060, 0x4380, 0x8404, 0x3FC}},
    {13348, 2, {0x2040, 0x23F8, 0x2040, 0x21F0, 0xF910, 0x29F0, 0x2910, 0x29F0, 0x2910, 0x29F0, 0x2910, 0x2BFA, 0x28A2, 0x4912, 0x47FE, 0x8000}},
    {13349, 2, {0x1008, 0x29C8, 0x455E, 0xB94A, 0x11CA, 0x7D4A, 0x11CA, 0x5552, 0x3D6A, 0xE384, 0x0, 0x3FF0, 0x1C0, 0x1E04, 0x2004, 0x1FFC}},
    {13350, 2, {0x3BB8, 0x2208, 0x3BB8, 0x2488, 0xFFFE, 0x1050, 0x2788, 0xC106, 0x1FF0, 0x100, 0x300, 0x3FF0, 0x1C0, 0x1E04, 0x2004, 0x1FFC}},
    {13351, 2, {0x80, 0x1080, 0x880, 0x480, 0x480, 0x80, 0x7FFC, 0x184, 0x288, 0x490, 0x880, 0x3080, 0xC080, 0x80, 0x280, 0x100}},
    {13352, 2, {0x0, 0x7CFC, 0x404, 0x2828, 0x1010, 0x808, 0xFEFE, 0x1212, 0x1414, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x5050, 0x2020}},
    {13353, 2, {0x2424, 0x2424, 0x2424, 0x2424, 0x7E7E, 0x2424, 0x2424, 0x2424, 0xFEFE, 0x2424, 0x2424, 0x2424, 0x2424, 0x4444, 0x4444, 0x8484}},
    {13354, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x288, 0x440, 0x820, 0x3018, 0xC006}},
    {13355, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x0, 0x820, 0x440, 0x2288, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x2008, 0x3FF8, 0x8}},
    {13356, 2, {0x200, 0x100, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008, 0x0, 0x1110, 0x1110, 0x1110, 0x2110, 0x2112, 0x4112, 0x800E}},
    {13357, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13358, 2, {0x200, 0x100, 0xFFFE, 0x1010, 0x2448, 0x4444, 0x1FF0, 0x440, 0x1FF0, 0x440, 0x7FFC, 0x490, 0xC60, 0x3518, 0xC606, 0x400}},
    {13359, 2, {0x100, 0x7FFC, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x1FE0, 0x100, 0x7FFC, 0x920, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE0}},
    {13360, 2, {0x800, 0x9FC, 0x904, 0x1104, 0x1104, 0x3104, 0x3104, 0x5104, 0x9104, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1000}},
    {13361, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1050, 0x2188, 0xC606, 0x3820, 0xC0, 0x700, 0x3818, 0x60, 0x380, 0x7C00, 0x0}},
    {13362, 2, {0x840, 0x840, 0x840, 0x1040, 0x1040, 0x37FE, 0x3040, 0x5040, 0x9040, 0x10A0, 0x10A0, 0x1090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {13363, 2, {0x800, 0x800, 0xBFC, 0x1090, 0x1090, 0x3090, 0x3090, 0x5090, 0x9090, 0x1090, 0x1090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {13364, 2, {0x880, 0x880, 0x880, 0x10FC, 0x1104, 0x3104, 0x3284, 0x5448, 0x9028, 0x1010, 0x1010, 0x1020, 0x1040, 0x1080, 0x1100, 0x1600}},
    {13365, 2, {0x800, 0xBFC, 0x840, 0x1040, 0x1040, 0x3040, 0x3FFE, 0x5040, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080}},
    {13366, 2, {0x800, 0x800, 0xBFC, 0x1204, 0x1204, 0x3204, 0x3204, 0x53FC, 0x9204, 0x1200, 0x1200, 0x1202, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {13367, 2, {0x800, 0xFF0, 0x810, 0x1010, 0x1210, 0x3210, 0x3210, 0x53FC, 0x9004, 0x1004, 0x1004, 0x17F4, 0x1004, 0x1004, 0x1028, 0x1010}},
    {13368, 2, {0x880, 0x880, 0x880, 0x10FC, 0x1104, 0x3108, 0x3240, 0x5440, 0x9040, 0x10A0, 0x10A0, 0x1090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {13369, 2, {0x900, 0x900, 0x9FC, 0x1200, 0x1400, 0x31F8, 0x3000, 0x5000, 0x93F8, 0x1008, 0x1008, 0x1008, 0x100A, 0x100A, 0x1006, 0x1002}},
    {13370, 2, {0x1120, 0x1120, 0x1120, 0x2220, 0x2220, 0x6620, 0x6620, 0xAA20, 0x2220, 0x2250, 0x2250, 0x2250, 0x2248, 0x2288, 0x2284, 0x2302}},
    {13371, 2, {0x820, 0x820, 0x820, 0x1020, 0x13FE, 0x3222, 0x3222, 0x5222, 0x9252, 0x124A, 0x128A, 0x1302, 0x1202, 0x1202, 0x120A, 0x1204}},
    {13372, 2, {0x808, 0x868, 0xB88, 0x1088, 0x1088, 0x3088, 0x3088, 0x53FE, 0x9088, 0x1088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {13373, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1328, 0x32A8, 0x32A8, 0x5248, 0x9248, 0x12A8, 0x12A8, 0x132A, 0x140A, 0x140A, 0x1806, 0x1002}},
    {13374, 2, {0x800, 0x800, 0xBF8, 0x1000, 0x1000, 0x3000, 0x37FC, 0x5120, 0x9120, 0x1120, 0x1120, 0x1120, 0x1222, 0x1222, 0x141E, 0x1800}},
    {13375, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x33F8, 0x3040, 0x5040, 0x9040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080}},
    {13376, 2, {0x840, 0x840, 0x840, 0x1040, 0x17FE, 0x3040, 0x3040, 0x50A0, 0x90A0, 0x10A0, 0x1110, 0x1110, 0x1288, 0x1248, 0x1444, 0x1802}},
    {13377, 2, {0x840, 0x840, 0xA48, 0x1248, 0x1248, 0x3248, 0x33F8, 0x5048, 0x9040, 0x1040, 0x1444, 0x1444, 0x1444, 0x1444, 0x17FC, 0x1004}},
    {13378, 2, {0x1000, 0x17FC, 0x1040, 0x2040, 0x2040, 0x67FC, 0x6444, 0xA444, 0x24A4, 0x2494, 0x2514, 0x2604, 0x2404, 0x2404, 0x2414, 0x2408}},
    {13379, 2, {0x1220, 0x1220, 0x1220, 0x2420, 0x257C, 0x6924, 0x6E24, 0xA224, 0x2424, 0x2424, 0x2924, 0x2F24, 0x2144, 0x2044, 0x2094, 0x2108}},
    {13380, 2, {0x800, 0x9F8, 0x908, 0x1148, 0x1128, 0x3128, 0x37FE, 0x5208, 0x9248, 0x1228, 0x1228, 0x13FE, 0x1008, 0x1008, 0x1050, 0x1020}},
    {13381, 2, {0x900, 0x900, 0x900, 0x11FE, 0x1240, 0x3240, 0x3440, 0x5040, 0x907C, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {13382, 2, {0x800, 0xBFC, 0xA04, 0x1204, 0x1204, 0x3204, 0x33FC, 0x5090, 0x9090, 0x1090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E, 0x1800}},
    {13383, 2, {0x880, 0x880, 0x8F8, 0x1108, 0x1210, 0x3420, 0x33FC, 0x5004, 0x9004, 0x1004, 0x11FC, 0x1004, 0x1004, 0x1004, 0x13FC, 0x1004}},
    {13384, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1080, 0x3120, 0x3120, 0x5220, 0x93FC, 0x1020, 0x1128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {13385, 2, {0x1000, 0x17F8, 0x1408, 0x2408, 0x27F8, 0x6488, 0x6488, 0xA4E8, 0x2528, 0x2528, 0x26A8, 0x244A, 0x244A, 0x288A, 0x2906, 0x3202}},
    {13386, 2, {0x840, 0x820, 0x820, 0x17FE, 0x1090, 0x3090, 0x3090, 0x5294, 0x9292, 0x1292, 0x1492, 0x1090, 0x1110, 0x1110, 0x1250, 0x1420}},
    {13387, 2, {0x10A0, 0x10A0, 0x10A0, 0x20A0, 0x27FC, 0x64A4, 0x64A4, 0xA4A4, 0x24A4, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x27FC, 0x2404}},
    {13388, 2, {0x800, 0xBFE, 0xA00, 0x1200, 0x13FC, 0x3220, 0x3220, 0x52F8, 0x9220, 0x1220, 0x1220, 0x13FC, 0x1200, 0x1200, 0x13FE, 0x1000}},
    {13389, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x2040, 0x63F8, 0x6040, 0xA040, 0x27FE, 0x20E0, 0x2150, 0x2150, 0x2248, 0x2444, 0x2842, 0x2040}},
    {13390, 2, {0x1000, 0x1248, 0x1248, 0x2490, 0x2490, 0x6248, 0x6248, 0xA000, 0x2040, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2004}},
    {13391, 2, {0x840, 0x820, 0xBFE, 0x1202, 0x1414, 0x3010, 0x3010, 0x53FE, 0x9010, 0x1110, 0x1090, 0x1090, 0x1010, 0x1010, 0x1050, 0x1020}},
    {13392, 2, {0x840, 0x840, 0x840, 0x107C, 0x1040, 0x3040, 0x37FE, 0x5000, 0x9040, 0x1040, 0x1250, 0x1248, 0x1444, 0x1844, 0x1140, 0x1080}},
    {13393, 2, {0x840, 0x840, 0x880, 0x13FC, 0x1204, 0x3204, 0x3204, 0x53FC, 0x9204, 0x1204, 0x13FC, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {13394, 2, {0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x100, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x280, 0x440, 0x1830, 0xE00E}},
    {13395, 2, {0x820, 0x840, 0x888, 0x1104, 0x13FE, 0x3042, 0x3040, 0x57FE, 0x9080, 0x10FC, 0x1144, 0x1144, 0x1228, 0x1210, 0x1468, 0x1186}},
    {13396, 2, {0x800, 0xBFE, 0x820, 0x1040, 0x11FC, 0x3104, 0x3104, 0x5124, 0x9124, 0x1124, 0x1124, 0x1144, 0x1050, 0x1088, 0x1104, 0x1204}},
    {13397, 2, {0x820, 0x820, 0xA22, 0x1222, 0x1222, 0x33FE, 0x3040, 0x5040, 0x90FC, 0x1104, 0x1284, 0x1448, 0x1050, 0x1020, 0x10C0, 0x1700}},
    {13398, 2, {0x1110, 0x1110, 0x1110, 0x2210, 0x22FE, 0x6692, 0x6692, 0xAA92, 0x2292, 0x22FE, 0x2292, 0x2210, 0x2210, 0x2210, 0x2210, 0x2210}},
    {13399, 2, {0x880, 0x900, 0xBF8, 0x1208, 0x13F8, 0x3208, 0x33FA, 0x520A, 0x920C, 0x17F8, 0x1018, 0x1028, 0x1048, 0x1188, 0x1E28, 0x1010}},
    {13400, 2, {0x1014, 0x1012, 0x1010, 0x2FFE, 0x2010, 0x6290, 0x6292, 0xA292, 0x2FD2, 0x2294, 0x2294, 0x2288, 0x228A, 0x249A, 0x2426, 0x2842}},
    {13401, 2, {0x1000, 0x1FFE, 0x1480, 0x2490, 0x2490, 0x6790, 0x6490, 0xA490, 0x2490, 0x2790, 0x2490, 0x2492, 0x24D2, 0x2F92, 0x208E, 0x2080}},
    {13402, 2, {0x1000, 0x17DE, 0x1252, 0x2252, 0x2254, 0x67D4, 0x6258, 0xA254, 0x2252, 0x2FD2, 0x2252, 0x225A, 0x2254, 0x2450, 0x2550, 0x2890}},
    {13403, 2, {0x840, 0x840, 0xFFE, 0x1040, 0x1040, 0x37FC, 0x3444, 0x5444, 0x97FC, 0x1040, 0x10E0, 0x1150, 0x1248, 0x1C46, 0x1040, 0x1040}},
    {13404, 2, {0x1208, 0x121C, 0x1260, 0x2240, 0x2FC0, 0x6240, 0x627E, 0xA2C8, 0x2348, 0x2E48, 0x2248, 0x2248, 0x2248, 0x2248, 0x2A88, 0x2508}},
    {13405, 2, {0x1000, 0x17FC, 0x1040, 0x2040, 0x23F8, 0x6040, 0x6040, 0xA7FC, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2110, 0x2210, 0x2210, 0x2410}},
    {13406, 2, {0x1040, 0x1040, 0x1244, 0x2244, 0x24A8, 0x6110, 0x6208, 0xAC06, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2110, 0x2210, 0x2210, 0x2410}},
    {13407, 2, {0x808, 0x81C, 0x9E0, 0x1100, 0x1100, 0x31FE, 0x3110, 0x5110, 0x9110, 0x1110, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204, 0x1402}},
    {13408, 2, {0x880, 0xAFC, 0xAA4, 0x12A4, 0x12A8, 0x3290, 0x32A8, 0x50C6, 0x9020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {13409, 2, {0x1000, 0x17BC, 0x14A4, 0x24A4, 0x24A4, 0x67A4, 0x64A4, 0xA4A4, 0x24A4, 0x27A4, 0x2434, 0x2528, 0x24A0, 0x2560, 0x2620, 0x2020}},
    {13410, 2, {0x900, 0x900, 0x91E, 0x1392, 0x1112, 0x3112, 0x3112, 0x57D2, 0x9112, 0x1112, 0x121A, 0x1294, 0x17D0, 0x1250, 0x1010, 0x1010}},
    {13411, 2, {0x910, 0x910, 0x910, 0x1110, 0x17BC, 0x3110, 0x3110, 0x5338, 0x93B8, 0x1554, 0x1554, 0x1992, 0x1110, 0x1110, 0x1110, 0x1110}},
    {13412, 2, {0x1208, 0x1208, 0x1FFE, 0x2208, 0x2208, 0x6000, 0x67FC, 0xA444, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2404}},
    {13413, 2, {0x800, 0xBF8, 0xA08, 0x13F8, 0x1208, 0x33F8, 0x3100, 0x5100, 0x93FC, 0x1494, 0x1894, 0x1124, 0x1244, 0x1484, 0x1128, 0x1210}},
    {13414, 2, {0x840, 0x820, 0xBFC, 0x1204, 0x1204, 0x33FC, 0x3200, 0x5228, 0x9224, 0x13FE, 0x1220, 0x1250, 0x1250, 0x1488, 0x1504, 0x1A02}},
    {13415, 2, {0x1038, 0x13C0, 0x1040, 0x27FC, 0x20E0, 0x6150, 0x6248, 0xAC46, 0x23F0, 0x2020, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2140, 0x2080}},
    {13416, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3100, 0x3100, 0x53F8, 0x9440, 0x1040, 0x1FFE, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1C06}},
    {13417, 2, {0x800, 0xBFE, 0xA52, 0x1252, 0x1252, 0x33FE, 0x3020, 0x5040, 0x90FC, 0x1104, 0x1288, 0x1050, 0x1020, 0x1040, 0x1180, 0x1600}},
    {13418, 2, {0x840, 0x820, 0xFFE, 0x1108, 0x1090, 0x3060, 0x3198, 0x5606, 0x9108, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x1108, 0x1208, 0x1408}},
    {13419, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1288, 0x3446, 0x33F0, 0x5010, 0x90A0, 0x1040, 0x1148, 0x1504, 0x1514, 0x1514, 0x18F0, 0x1000}},
    {13420, 2, {0x1210, 0x1210, 0x1210, 0x2210, 0x2F7E, 0x6510, 0x6510, 0xA510, 0x257C, 0x2944, 0x2544, 0x2244, 0x2244, 0x2544, 0x257C, 0x2844}},
    {13421, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x27FC, 0x6444, 0x6444, 0xA7FC, 0x2000, 0x2000, 0x2FFE, 0x2208, 0x2208, 0x2208, 0x2408, 0x2808}},
    {13422, 2, {0x800, 0xBFE, 0xA22, 0x1020, 0x13FE, 0x3020, 0x31FC, 0x5124, 0x91FC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {13423, 2, {0x800, 0xFFE, 0x840, 0x1080, 0x13FC, 0x3204, 0x3204, 0x53FC, 0x9204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1000, 0x1108, 0x1204}},
    {13424, 2, {0x900, 0x9F8, 0xA08, 0x13F0, 0x1010, 0x37FE, 0x3080, 0x5144, 0x9668, 0x10B0, 0x1128, 0x1668, 0x10A4, 0x1122, 0x16A0, 0x1040}},
    {13425, 2, {0x800, 0xBF8, 0x890, 0x1060, 0x17FC, 0x30A4, 0x3128, 0x5220, 0x9460, 0x1040, 0x17FC, 0x10E0, 0x1150, 0x1248, 0x1C46, 0x1040}},
    {13426, 2, {0x1040, 0x1040, 0x1FFE, 0x2040, 0x2040, 0x67FC, 0x6514, 0xA4A4, 0x25F4, 0x2444, 0x2444, 0x25F4, 0x2444, 0x2444, 0x2454, 0x2408}},
    {13427, 2, {0x900, 0x900, 0xBFE, 0x1442, 0x1252, 0x33F2, 0x3082, 0x5114, 0x9608, 0x13FE, 0x1442, 0x1A52, 0x13F2, 0x1082, 0x1114, 0x1608}},
    {13428, 2, {0xBF8, 0x808, 0x9F8, 0x1008, 0x13F8, 0x3000, 0x37FC, 0x5444, 0x93F8, 0x1248, 0x1258, 0x1040, 0x13F8, 0x1110, 0x10E0, 0x171E}},
    {13429, 2, {0x820, 0x810, 0x9FE, 0x1100, 0x1120, 0x3520, 0x333C, 0x5150, 0x9190, 0x1310, 0x15FE, 0x1110, 0x1128, 0x1228, 0x1244, 0x1482}},
    {13430, 2, {0x900, 0x900, 0x9FC, 0x1200, 0x15F8, 0x3000, 0x33F8, 0x5088, 0x92A8, 0x11C8, 0x17F8, 0x108A, 0x11CA, 0x16AA, 0x1086, 0x1082}},
    {13431, 2, {0x1110, 0x1212, 0x1454, 0x27D8, 0x2010, 0x6012, 0x67D2, 0xA44E, 0x2440, 0x27D2, 0x2454, 0x2458, 0x27D0, 0x2452, 0x2552, 0x248E}},
    {13432, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x13F8, 0x3208, 0x33F8, 0x5000, 0x9FFE, 0x1008, 0x13E8, 0x1228, 0x13E8, 0x1008, 0x1028, 0x1010}},
    {13433, 2, {0x100, 0x280, 0x440, 0xFE0, 0x3118, 0xDFF6, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFE0, 0x820, 0xFE0, 0x820, 0xFE0}},
    {13434, 2, {0x8A0, 0x890, 0x9FE, 0x1320, 0x15FC, 0x3120, 0x31FC, 0x5120, 0x91FE, 0x1100, 0x13FC, 0x1088, 0x109E, 0x1102, 0x120A, 0x1404}},
    {13435, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2894, 0x6050, 0x6524, 0xA54A, 0x28FA, 0x2300, 0x2040, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2004}},
    {13436, 2, {0x1010, 0x1010, 0x103C, 0x2FA4, 0x2AC4, 0x6AA8, 0x6A90, 0xAFA8, 0x2AC6, 0x2A80, 0x2ABC, 0x2AA4, 0x2FA4, 0x28A4, 0x203C, 0x2024}},
    {13437, 2, {0x1040, 0x1444, 0x1248, 0x2FFE, 0x2802, 0x63F8, 0x6208, 0xA208, 0x23F8, 0x2040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x2000}},
    {13438, 2, {0x840, 0x8FC, 0x908, 0x1650, 0x1120, 0x30C0, 0x3320, 0x5FFE, 0x9000, 0x13FC, 0x1000, 0x13FC, 0x1000, 0x13FC, 0x1204, 0x13FC}},
    {13439, 2, {0x1248, 0x1490, 0x1248, 0x2000, 0x23F8, 0x6248, 0x63F8, 0xA248, 0x23F8, 0x2040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {13440, 2, {0x1080, 0x1040, 0x17FC, 0x2110, 0x20A0, 0x67FC, 0x6444, 0xA5F4, 0x2444, 0x25F4, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x2408}},
    {13441, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x25F4, 0x6444, 0x67FC, 0xA514, 0x24A4, 0x25F4, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2404}},
    {13442, 2, {0x908, 0x908, 0xFFE, 0x1108, 0x1044, 0x33F4, 0x3048, 0x5050, 0x97FE, 0x1080, 0x11FC, 0x1304, 0x15FC, 0x1904, 0x11FC, 0x1104}},
    {13443, 2, {0x814, 0xBFE, 0x810, 0x13D0, 0x1010, 0x33D0, 0x3010, 0x53D0, 0x9250, 0x13D0, 0x1250, 0x13C8, 0x124A, 0x13CA, 0x1186, 0x1242}},
    {13444, 2, {0x910, 0x914, 0x9D2, 0x1250, 0x127E, 0x3550, 0x3090, 0x52A8, 0x9128, 0x1144, 0x1244, 0x1482, 0x1008, 0x12A4, 0x1252, 0x1452}},
    {13445, 2, {0x820, 0xBFE, 0x840, 0x10F8, 0x1348, 0x3030, 0x33D0, 0x503C, 0x91C4, 0x1048, 0x1030, 0x13C4, 0x10A8, 0x1190, 0x16A8, 0x10C6}},
    {13446, 2, {0x840, 0x8A0, 0x918, 0x16E6, 0x1000, 0x33F8, 0x3208, 0x53F8, 0x9000, 0x17BC, 0x1084, 0x14A4, 0x1294, 0x14A4, 0x1294, 0x1108}},
    {13447, 2, {0x1100, 0x1100, 0x13FC, 0x26A8, 0x2AA8, 0x62A8, 0x67FC, 0xA2A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000, 0x22A8, 0x2254, 0x2454, 0x2000}},
    {13448, 2, {0x1008, 0x1788, 0x1088, 0x2110, 0x2FDE, 0x6494, 0x64A4, 0xA794, 0x2494, 0x2794, 0x2494, 0x24C8, 0x2788, 0x2C94, 0x20A4, 0x20C2}},
    {13449, 2, {0x840, 0xA48, 0x950, 0x17FC, 0x1404, 0x31F0, 0x3110, 0x5110, 0x91F0, 0x1040, 0x17FC, 0x10E0, 0x1150, 0x1248, 0x1446, 0x1040}},
    {13450, 2, {0x1000, 0x17BC, 0x1294, 0x26B4, 0x2AD4, 0x64A4, 0x65AC, 0xA840, 0x2108, 0x2108, 0x25A8, 0x252C, 0x2528, 0x25A8, 0x2E7E, 0x2000}},
    {13451, 2, {0x1480, 0x1482, 0x149C, 0x2FD0, 0x2490, 0x6790, 0x649E, 0xA794, 0x2494, 0x2494, 0x2FD4, 0x2014, 0x2514, 0x28A4, 0x3024, 0x2044}},
    {13452, 2, {0x1110, 0x1554, 0x15B8, 0x2910, 0x22A8, 0x6444, 0x6000, 0xA7FC, 0x2404, 0x29E8, 0x2120, 0x2120, 0x2120, 0x2222, 0x2422, 0x281E}},
    {13453, 2, {0x880, 0x910, 0xBF8, 0x1210, 0x14A4, 0x37BC, 0x3080, 0x5110, 0x93E0, 0x1040, 0x1088, 0x17FC, 0x1044, 0x1248, 0x1544, 0x1080}},
    {13454, 2, {0x800, 0xBF8, 0x880, 0x17FC, 0x1110, 0x3208, 0x3CA6, 0x50A0, 0x97BC, 0x10A0, 0x13B8, 0x10A0, 0x17BC, 0x10A0, 0x10A0, 0x10A0}},
    {13455, 2, {0x840, 0x820, 0xFFE, 0x1492, 0x13FC, 0x3090, 0x33FC, 0x5090, 0x97FE, 0x1108, 0x1204, 0x1462, 0x1010, 0x1180, 0x1060, 0x1010}},
    {13456, 2, {0x910, 0x910, 0x9DE, 0x12A8, 0x1444, 0x3100, 0x31FC, 0x5204, 0x95F4, 0x1114, 0x11F4, 0x1114, 0x11F4, 0x1114, 0x1028, 0x1010}},
    {13457, 2, {0x1000, 0x17FE, 0x1428, 0x2624, 0x257E, 0x6448, 0x64C8, 0xA47E, 0x2648, 0x257E, 0x2448, 0x2448, 0x257E, 0x2640, 0x27FE, 0x2000}},
    {13458, 2, {0x10A0, 0x14A4, 0x12A8, 0x20A0, 0x27FC, 0x6110, 0x60A0, 0xA7FC, 0x2040, 0x23F8, 0x2040, 0x27FC, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {13459, 2, {0x1010, 0x1410, 0x127C, 0x2010, 0x20FE, 0x6044, 0x6628, 0xA2FE, 0x2210, 0x227C, 0x2210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x2800}},
    {13460, 2, {0x1000, 0x17FC, 0x14A4, 0x24A4, 0x27FC, 0x6200, 0x63FE, 0xA442, 0x2BFA, 0x224A, 0x23FA, 0x2052, 0x207A, 0x278A, 0x2214, 0x2008}},
    {13461, 2, {0x1040, 0x1020, 0x17FE, 0x2028, 0x2224, 0x627E, 0x64C8, 0xA748, 0x217E, 0x2248, 0x24C8, 0x2F7E, 0x2148, 0x2248, 0x247E, 0x2840}},
    {13462, 2, {0x1110, 0x1110, 0x17FC, 0x2110, 0x23F8, 0x6248, 0x63F8, 0xA248, 0x23F8, 0x2040, 0x27FC, 0x2444, 0x2454, 0x25F4, 0x2414, 0x240C}},
    {13463, 2, {0x1110, 0x17FC, 0x1110, 0x21F0, 0x2110, 0x61F0, 0x6110, 0xA7FE, 0x2402, 0x2BFC, 0x2110, 0x23F8, 0x2040, 0x23F8, 0x2040, 0x27FC}},
    {13464, 2, {0x1000, 0x17BC, 0x1514, 0x2794, 0x2488, 0x6788, 0x6514, 0xA7A2, 0x2080, 0x2040, 0x27FC, 0x2000, 0x2208, 0x2110, 0x2FFE, 0x2000}},
    {13465, 2, {0x1108, 0x1110, 0x123E, 0x22A2, 0x24BE, 0x6F22, 0x613E, 0xA208, 0x24BE, 0x2FAA, 0x20AA, 0x202A, 0x2AAA, 0x2AAE, 0x2808, 0x2008}},
    {13466, 2, {0x103E, 0x17C0, 0x1244, 0x2128, 0x23F8, 0x6040, 0x67FC, 0xA000, 0x23F8, 0x2008, 0x23F8, 0x2008, 0x23F8, 0x2544, 0x252A, 0x28FA}},
    {13467, 2, {0x1000, 0x17BC, 0x1084, 0x24A4, 0x2294, 0x64A4, 0x6120, 0xA210, 0x27FC, 0x2A20, 0x23FC, 0x2220, 0x23FC, 0x2220, 0x23FE, 0x2200}},
    {13468, 2, {0x940, 0xA4C, 0xA64, 0x1244, 0x134C, 0x3264, 0x32A4, 0x53AC, 0x92A4, 0x12A4, 0x17FE, 0x1000, 0x1110, 0x1108, 0x1204, 0x1404}},
    {13469, 2, {0x908, 0xFFE, 0x908, 0x1000, 0x13FC, 0x3294, 0x3294, 0x53FC, 0x9028, 0x13FE, 0x1220, 0x1324, 0x12A8, 0x1292, 0x142A, 0x18C6}},
    {13470, 2, {0x820, 0xA22, 0xBFE, 0x1090, 0x11FE, 0x3310, 0x35FE, 0x5110, 0x91FE, 0x1110, 0x11FE, 0x1100, 0x13DE, 0x1252, 0x1272, 0x1206}},
    {13471, 2, {0x1040, 0x17FE, 0x1000, 0x23FC, 0x2240, 0x63F8, 0x6240, 0xA3F8, 0x2240, 0x23FC, 0x2004, 0x2554, 0x20A8, 0x2310, 0x2D48, 0x2186}},
    {13472, 2, {0x1100, 0x11BC, 0x1524, 0x27A8, 0x2490, 0x6AA8, 0x6146, 0xA2A0, 0x27BC, 0x20A0, 0x23B8, 0x20A0, 0x27BC, 0x20A0, 0x2FFE, 0x2000}},
    {13473, 2, {0x1182, 0x1E3C, 0x1088, 0x2A90, 0x2522, 0x643C, 0x6788, 0xAA12, 0x223E, 0x2F88, 0x2208, 0x2AAC, 0x2AAA, 0x2BCA, 0x2EA8, 0x2090}},
    {13474, 2, {0x1040, 0x10A0, 0x1110, 0x2208, 0x2DF6, 0x6000, 0x6EEE, 0xAAAA, 0x2EEE, 0x2000, 0x2FFE, 0x2922, 0x2FFE, 0x2922, 0x292A, 0x2804}},
    {13475, 2, {0x1040, 0x1020, 0x17FE, 0x24A8, 0x24FE, 0x6550, 0x677C, 0xA550, 0x257C, 0x2550, 0x257E, 0x2540, 0x28A4, 0x2A8A, 0x328A, 0x2478}},
    {13476, 2, {0xBFC, 0x908, 0x9F8, 0x1108, 0x11F8, 0x310E, 0x37F8, 0x5008, 0x97FE, 0x1294, 0x139C, 0x1294, 0x139C, 0x12D6, 0x17BC, 0x1084}},
    {13477, 2, {0x1150, 0x13F8, 0x1554, 0x27FC, 0x2554, 0x67FC, 0x6554, 0xA7FC, 0x2000, 0x27FC, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x2FFE}},
    {13478, 2, {0x17FC, 0x1444, 0x17FC, 0x2444, 0x27FC, 0x6000, 0x6FBE, 0xAAAA, 0x2FBE, 0x2AAA, 0x2FBE, 0x2040, 0x2040, 0x27FC, 0x2040, 0x2FFE}},
    {13479, 2, {0x1418, 0x1214, 0x1F90, 0x253E, 0x2AA8, 0x6DE8, 0x68BE, 0xAFA8, 0x2228, 0x2FBE, 0x2AA8, 0x2AA8, 0x2BA8, 0x28BE, 0x2AA0, 0x2920}},
    {13480, 2, {0x1040, 0x15F6, 0x1912, 0x29B2, 0x2D56, 0x69B2, 0x6912, 0xADF6, 0x2080, 0x2FFE, 0x2208, 0x25F4, 0x2912, 0x21F0, 0x2104, 0x20FC}},
    {13481, 2, {0x11F0, 0x1210, 0x13E0, 0x2020, 0x27FC, 0x6188, 0x6650, 0xA1A8, 0x2664, 0x21A0, 0x2248, 0x2FBE, 0x2AAA, 0x2FBE, 0x228A, 0x2FBE}},
    {13482, 2, {0x380, 0xC60, 0x3EF8, 0xE8A6, 0x3EF8, 0x2288, 0x3EF8, 0x28A0, 0x3EF8, 0x1000, 0x11FC, 0x5CA0, 0x5154, 0x50B8, 0x5D54, 0xE0B2}},
    {13483, 2, {0x0, 0xFF8, 0x800, 0x800, 0xFF0, 0x800, 0x800, 0xFFFE, 0x440, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {13484, 2, {0x0, 0x7DFC, 0x20, 0x20, 0x20, 0xFCF8, 0x2820, 0x2820, 0x2820, 0x2820, 0x29FC, 0x2800, 0x2800, 0x4802, 0x47FE, 0x8000}},
    {13485, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x2244, 0x2244, 0x3E7C, 0x1428, 0x1428, 0x1428, 0x1428, 0x144A, 0x254A, 0x268A, 0x4486, 0x8100}},
    {13486, 2, {0x0, 0x1FE, 0x7820, 0x40, 0xA2, 0x334, 0xFC58, 0x2894, 0x2B34, 0x2852, 0x2892, 0x2B50, 0x2822, 0x4802, 0x47FE, 0x8000}},
    {13487, 2, {0x1088, 0x13FE, 0x9088, 0x5400, 0x59FC, 0x10A8, 0xFCA8, 0x2BFE, 0x28A8, 0x28A8, 0x29FC, 0x2820, 0x2BFE, 0x4C20, 0x4820, 0x8020}},
    {13488, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x0, 0x1FF0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x7FFC, 0x0}},
    {13489, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3218, 0xC206, 0x7F0, 0x810, 0x1410, 0x2220, 0x140, 0x80, 0x300, 0x1C00, 0xE000}},
    {13490, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {13491, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x4104, 0x5144, 0x4924, 0x4924, 0x5554, 0x638C, 0x4104, 0x4104, 0x4104, 0x4104, 0x4114, 0x4108}},
    {13492, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3118, 0xD116, 0x1110, 0x1110, 0x1FF0, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {13493, 2, {0x400, 0x800, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x0, 0x0, 0x440, 0x820, 0x1010, 0x2008, 0x4004}},
    {13494, 2, {0x440, 0x440, 0x820, 0x1010, 0x2108, 0xC206, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {13495, 2, {0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1290, 0x1450, 0x1010, 0x1010, 0x7FFC, 0x0, 0x440, 0x820, 0x1010, 0x2008}},
    {13496, 2, {0x820, 0x420, 0x440, 0x7FFC, 0x200, 0x600, 0x1908, 0x6290, 0x4A0, 0x18C0, 0x61A0, 0x298, 0xC86, 0x3080, 0xC280, 0x100}},
    {13497, 2, {0x1010, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0xFEFE, 0x2828, 0x4444, 0x8282}},
    {13498, 2, {0x0, 0x7FFC, 0x4004, 0x4824, 0x4444, 0x5FF4, 0x4004, 0x4884, 0x4884, 0x4944, 0x4A24, 0x4804, 0x4FF4, 0x4004, 0x4014, 0x4008}},
    {13499, 2, {0x7FFC, 0x4004, 0x5FF4, 0x4004, 0x5FF4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {13500, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0xFE0, 0x100, 0x7FFC, 0x4104, 0x5144, 0x4924, 0x5554, 0x638C, 0x4104, 0x4114, 0x4108}},
    {13501, 2, {0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0x8A0, 0x848, 0x7F8}},
    {13502, 2, {0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x2040, 0x1040, 0xFEFE, 0x2120, 0x3CAC, 0x24F4, 0x27A4, 0x44AC, 0x54A2, 0x887E}},
    {13503, 2, {0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x800, 0x7F28, 0x1424, 0x497E, 0x55C8, 0x7F7E, 0x848, 0x7F7E, 0x5548, 0x5D7E, 0x4340}},
    {13504, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x200, 0x200, 0x200, 0x200, 0x280, 0x240, 0x220, 0x210, 0x200, 0x200, 0x200, 0x200}},
    {13505, 2, {0x0, 0x0, 0x7FFE, 0x4002, 0x8444, 0x820, 0x1010, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {13506, 2, {0x0, 0x7FFE, 0x4002, 0x8024, 0x3E20, 0x1FE, 0x44, 0x7E44, 0x14C8, 0x1428, 0x1410, 0x146A, 0x2586, 0x2402, 0x43FE, 0x8000}},
    {13507, 2, {0x0, 0x7FFE, 0x4002, 0x9004, 0x800, 0x7F1C, 0xE0, 0x3E20, 0x20, 0x3E3E, 0x1E0, 0x3E22, 0x2222, 0x2222, 0x3E1E, 0x2200}},
    {13508, 2, {0x1010, 0x2210, 0x7F10, 0x4210, 0x94FE, 0xF792, 0x892, 0x3610, 0xC928, 0x30A8, 0xC428, 0x1928, 0x6248, 0xC4A, 0x308A, 0xC106}},
    {13509, 2, {0x0, 0x4000, 0x23FE, 0x2020, 0x20, 0x820, 0x820, 0x1020, 0x1020, 0xE020, 0x2020, 0x2020, 0x2020, 0x2020, 0x20A0, 0x40}},
    {13510, 2, {0x40, 0x4040, 0x20A0, 0x20A0, 0x110, 0xA08, 0xC06, 0x1110, 0x1120, 0xE140, 0x2180, 0x2100, 0x2104, 0x2104, 0x20FC, 0x0}},
    {13511, 2, {0x10, 0x4078, 0x27C0, 0x2040, 0x40, 0x840, 0x840, 0x17FE, 0x1040, 0xE0A0, 0x20A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {13512, 2, {0x0, 0x4000, 0x27FE, 0x2040, 0x40, 0x880, 0x880, 0x11F8, 0x1308, 0xE508, 0x2908, 0x2108, 0x2108, 0x2108, 0x21F8, 0x108}},
    {13513, 2, {0x10, 0x4078, 0x23C0, 0x2040, 0x40, 0x7FE, 0x1040, 0x1040, 0x2040, 0xE3F8, 0x2208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {13514, 2, {0x0, 0x47FC, 0x2404, 0x2404, 0x5F4, 0x404, 0x1404, 0x15F4, 0x2514, 0xE514, 0x2514, 0x25F4, 0x2404, 0x2404, 0x2414, 0x408}},
    {13515, 2, {0x110, 0x4110, 0x2110, 0x2110, 0x7FC, 0x110, 0x1110, 0x1110, 0x2110, 0xEFFE, 0x2000, 0x2110, 0x2108, 0x2208, 0x2404, 0x804}},
    {13516, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x208, 0x406, 0x13F8, 0x1040, 0x2040, 0xE040, 0x23F8, 0x2040, 0x2040, 0x2040, 0x27FE, 0x0}},
    {13517, 2, {0x80, 0x4040, 0x2040, 0x27FE, 0x80, 0x110, 0x1208, 0x17FC, 0x2124, 0xE120, 0x2120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {13518, 2, {0x0, 0x43F8, 0x2008, 0x21F8, 0x8, 0xBF8, 0x800, 0x17FC, 0x1404, 0xE3F0, 0x2110, 0x2110, 0x20A0, 0x2040, 0x21B0, 0x60E}},
    {13519, 2, {0x0, 0x47FC, 0x2404, 0x24A4, 0x514, 0x60C, 0x1404, 0x15F4, 0x2514, 0xE514, 0x2514, 0x25F4, 0x2514, 0x2404, 0x2414, 0x408}},
    {13520, 2, {0x40, 0x4040, 0x23F8, 0x2040, 0x40, 0x7FC, 0x1110, 0x1208, 0x2444, 0xE040, 0x23F8, 0x2040, 0x2040, 0x2040, 0x27FC, 0x0}},
    {13521, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x40, 0xFFE, 0x1110, 0x10A0, 0x27FC, 0xE040, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x40}},
    {13522, 2, {0x2040, 0x1040, 0x478, 0xBC4, 0x7044, 0x103C, 0x1100, 0x1280, 0xC60, 0x3118, 0xC106, 0x990, 0x1148, 0x2148, 0x500, 0x200}},
    {13523, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x40, 0x7FC, 0x1514, 0x14A4, 0x25F4, 0xE444, 0x2444, 0x25F4, 0x2444, 0x2444, 0x2454, 0x408}},
    {13524, 2, {0x108, 0x4208, 0x2788, 0x2488, 0x7BE, 0x488, 0x1788, 0x14C8, 0x24A8, 0xEFA8, 0x2188, 0x2288, 0x2488, 0x2888, 0x22A8, 0x110}},
    {13525, 2, {0x14, 0x4012, 0x2010, 0x27FE, 0x410, 0x410, 0x17F2, 0x1492, 0x2492, 0xE5B4, 0x26D4, 0x2488, 0x254A, 0x293A, 0x2A26, 0x1042}},
    {13526, 2, {0x0, 0x47FC, 0x2444, 0x27FC, 0x444, 0x7FC, 0x1248, 0x1248, 0x2FFE, 0xE248, 0x2248, 0x27FC, 0x2040, 0x2FFE, 0x2040, 0x40}},
    {13527, 2, {0x100, 0x47BC, 0x2114, 0x2394, 0x114, 0x7A4, 0x114C, 0x1080, 0x2110, 0xE3E0, 0x2048, 0x27FC, 0x2044, 0x2248, 0x2544, 0x80}},
    {13528, 2, {0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2108, 0x23C8, 0x2448, 0x2A88, 0x2108, 0x2608, 0x3BE8, 0x210A, 0x2FEA, 0x414A, 0x4126, 0x9FF2}},
    {13529, 2, {0x200, 0x100, 0x100, 0x1FF0, 0x4044, 0x4844, 0x4484, 0x4284, 0x4104, 0x4284, 0x4444, 0x4824, 0x5024, 0x4004, 0x7FFC, 0x4}},
    {13530, 2, {0x4, 0x4, 0x7F04, 0x824, 0x824, 0x824, 0x824, 0x824, 0x824, 0x824, 0x824, 0xF24, 0xF004, 0x4004, 0x14, 0x8}},
    {13531, 2, {0x0, 0x1FC, 0x44, 0x7E44, 0x844, 0x844, 0x844, 0x844, 0x844, 0x844, 0xE84, 0xF084, 0x4104, 0x104, 0x228, 0x410}},
    {13532, 2, {0x800, 0x800, 0x10FC, 0x1024, 0x2224, 0x4224, 0xFC24, 0x424, 0x824, 0x1024, 0x2424, 0x4224, 0xFE44, 0x4244, 0x94, 0x108}},
    {13533, 2, {0x2204, 0x2204, 0x2204, 0x2224, 0x7F24, 0x2224, 0x2224, 0x2224, 0xFFA4, 0x2224, 0x2224, 0x2224, 0x2204, 0x4204, 0x4214, 0x8208}},
    {13534, 2, {0x800, 0x800, 0x8FC, 0x7F24, 0x824, 0x824, 0x824, 0x7F24, 0x824, 0x824, 0x824, 0xFFA4, 0x844, 0x844, 0x894, 0x908}},
    {13535, 2, {0x404, 0x404, 0x404, 0x3FA4, 0x24A4, 0x2524, 0x2424, 0x3FA4, 0x20A4, 0x2924, 0x2524, 0x2224, 0x2504, 0x4884, 0x9094, 0x8}},
    {13536, 2, {0x804, 0x804, 0x804, 0x824, 0xFA4, 0x824, 0x824, 0x824, 0x7F24, 0x4124, 0x4124, 0x4124, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {13537, 2, {0x2004, 0x2004, 0x3F04, 0x4824, 0x8824, 0x824, 0xFF24, 0x824, 0x824, 0x4924, 0x4924, 0x4904, 0x4F04, 0x7904, 0x114, 0x8}},
    {13538, 2, {0x1004, 0x1004, 0x1F04, 0x2124, 0x6124, 0x9224, 0xC24, 0x1224, 0x2124, 0xC0A4, 0x3F24, 0x2124, 0x2104, 0x2104, 0x3F14, 0x2108}},
    {13539, 2, {0x804, 0x804, 0x1404, 0x2224, 0x4124, 0xBEA4, 0x24, 0x24, 0x3E24, 0x2224, 0x2224, 0x2224, 0x2204, 0x3E04, 0x2214, 0x8}},
    {13540, 2, {0x804, 0x804, 0x804, 0xFFA4, 0x824, 0x824, 0x7F24, 0x24, 0x24, 0x7F24, 0x4124, 0x4124, 0x4104, 0x7F04, 0x4114, 0x8}},
    {13541, 2, {0x804, 0x4904, 0x2904, 0x2924, 0x2A24, 0x824, 0xFF24, 0x124, 0x124, 0x124, 0x7F24, 0x124, 0x104, 0x104, 0xFF14, 0x108}},
    {13542, 2, {0x4, 0xFFC4, 0x804, 0x824, 0x1024, 0x7F24, 0x4124, 0x4124, 0x4124, 0x7F24, 0x4124, 0x4124, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {13543, 2, {0x804, 0x804, 0x1404, 0x2224, 0x5124, 0x88A4, 0x7E24, 0x224, 0x424, 0x824, 0x7F24, 0x4124, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {13544, 2, {0x804, 0x804, 0x804, 0xFFA4, 0x824, 0x2A24, 0x2A24, 0x2A24, 0x5D24, 0x88A4, 0x824, 0x1424, 0x1204, 0x2104, 0x4114, 0x8008}},
    {13545, 2, {0x1004, 0x1004, 0x1F04, 0x2124, 0x4224, 0xBFA4, 0x24A4, 0x24A4, 0x3FA4, 0x24A4, 0x24A4, 0x3FA4, 0x2484, 0x4484, 0x4294, 0x8108}},
    {13546, 2, {0x1000, 0x800, 0x7EFC, 0x4224, 0x4224, 0x7E24, 0x4224, 0x4224, 0x7E24, 0x4024, 0x4824, 0x4424, 0x4A44, 0x5244, 0x6094, 0x4108}},
    {13547, 2, {0x804, 0x804, 0xFFC4, 0x1224, 0x2124, 0x40A4, 0xBF24, 0x1224, 0x1224, 0x1224, 0x7FA4, 0x1224, 0x1204, 0x2204, 0x2214, 0x4208}},
    {13548, 2, {0x804, 0x4904, 0x2A04, 0x7F24, 0x824, 0x1024, 0xFFA4, 0x2224, 0x4124, 0x80A4, 0x7F24, 0x1124, 0x1104, 0x2104, 0x4A14, 0x8408}},
    {13549, 2, {0x4, 0x7F04, 0x4104, 0x7F24, 0x4124, 0x7F24, 0x2024, 0x7FA4, 0x88A4, 0x48A4, 0x54A4, 0x40A4, 0x7E84, 0x84, 0x514, 0x208}},
    {13550, 2, {0x4, 0x3F84, 0x2484, 0x24A4, 0x2EA4, 0x24A4, 0x3FA4, 0x20A4, 0x2EA4, 0x2AA4, 0x2AA4, 0x2EA4, 0x2084, 0x4084, 0x4294, 0x8108}},
    {13551, 2, {0x2484, 0x2484, 0x4904, 0x9224, 0x4924, 0x24A4, 0x24A4, 0x24, 0x7FA4, 0x48A4, 0x48A4, 0x7FA4, 0x4884, 0x4884, 0x7F94, 0x4088}},
    {13552, 2, {0x4, 0xFF84, 0x804, 0x824, 0x3E24, 0xAAA4, 0xAAA4, 0xBEA4, 0xAAA4, 0xAAA4, 0xAAA4, 0xBEA4, 0x8084, 0x8F84, 0xF094, 0x8}},
    {13553, 2, {0x804, 0x1404, 0x2204, 0x4124, 0xBEA4, 0x24, 0x78A4, 0x4AA4, 0x4AA4, 0x7AA4, 0x4AA4, 0x4AA4, 0x7A84, 0x4884, 0x4A94, 0x5908}},
    {13554, 2, {0x802, 0x402, 0x3FC2, 0x2042, 0x2052, 0x3FD2, 0x2012, 0x2012, 0x3FD2, 0x3552, 0x3552, 0x5FD2, 0x5542, 0x5542, 0x944A, 0x10C4}},
    {13555, 2, {0x8, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x48, 0xFF48, 0x848, 0x848, 0x2F08, 0x2828, 0x2810, 0x5800, 0x4FFE, 0x8000}},
    {13556, 2, {0x4, 0xFF84, 0x804, 0x1024, 0x7F24, 0x5524, 0x5524, 0x5524, 0x5524, 0x4B24, 0x824, 0xFFA4, 0x1404, 0x2204, 0x4114, 0x8088}},
    {13557, 2, {0x4, 0xF784, 0x9484, 0x94A4, 0xF7A4, 0x24, 0x7F24, 0x24, 0xFFA4, 0x2024, 0x4024, 0x7F24, 0x104, 0x104, 0xA14, 0x408}},
    {13558, 2, {0x2004, 0x2F04, 0xF504, 0x2524, 0x7524, 0x2524, 0xF524, 0x2924, 0x2324, 0x824, 0x824, 0xFFA4, 0x1404, 0x2204, 0x4114, 0x8088}},
    {13559, 2, {0x804, 0x1004, 0x7F04, 0x4124, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x7F24, 0x824, 0xFFA4, 0x1C24, 0x2A04, 0x4904, 0x8814, 0x808}},
    {13560, 2, {0x4, 0x7704, 0x5504, 0x5524, 0x7724, 0x1424, 0x1424, 0x7724, 0x4124, 0x4124, 0x7724, 0x1424, 0x1404, 0x1F04, 0xF014, 0x4008}},
    {13561, 2, {0x4, 0x7FC4, 0x4404, 0x54A4, 0x6524, 0x4424, 0x4A24, 0x5124, 0x60A4, 0x4424, 0x54A4, 0x6524, 0x4404, 0x4A04, 0x9114, 0x2088}},
    {13562, 2, {0x804, 0xF84, 0x804, 0x7FA4, 0x48A4, 0x4F24, 0x78A4, 0x47A4, 0x4824, 0x4424, 0x7FA4, 0x4124, 0x5A04, 0x4404, 0x8A14, 0x3108}},
    {13563, 2, {0x4, 0xFF84, 0x9484, 0x94A4, 0xFFA4, 0x24, 0x24, 0xFFA4, 0x824, 0x824, 0x4F24, 0x4824, 0x4804, 0x4F84, 0xF814, 0x8}},
    {13564, 2, {0x804, 0x804, 0xFF84, 0x1C24, 0x2B24, 0xC8A4, 0x1424, 0x2224, 0x4924, 0x88A4, 0x4924, 0x2A24, 0x4904, 0x8884, 0x2814, 0x1008}},
    {13565, 2, {0x804, 0x4904, 0x4904, 0x7F24, 0x1024, 0x824, 0xFFA4, 0x80A4, 0x3F24, 0x24, 0xFFA4, 0x824, 0x2A04, 0x4904, 0xA894, 0x1008}},
    {13566, 2, {0x2, 0x7FC2, 0x4042, 0x7FC2, 0x4212, 0x4A92, 0x4712, 0x4A92, 0x5052, 0x4A12, 0x4FD2, 0x5212, 0x4202, 0xBFE2, 0x820A, 0x204}},
    {13567, 2, {0x704, 0x7804, 0x804, 0xFFA4, 0x2A24, 0x4924, 0x94A4, 0x2224, 0x4924, 0x88A4, 0x4924, 0x2A24, 0x4904, 0x8884, 0x2814, 0x1008}},
    {13568, 2, {0x4, 0x7F04, 0x4104, 0x7F24, 0x4124, 0x7F24, 0x824, 0xFFA4, 0x24, 0x7F24, 0x4124, 0x7F24, 0x804, 0x4904, 0x8894, 0x1808}},
    {13569, 2, {0x504, 0x7204, 0x1284, 0x5124, 0x2124, 0x5EA4, 0x80A4, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x2124, 0x1204, 0x784, 0x7814, 0x2008}},
    {13570, 2, {0x404, 0x2484, 0x1504, 0x7FD4, 0x1514, 0x2494, 0x4454, 0x2094, 0x2094, 0x3BF4, 0x4894, 0xAA84, 0x13E4, 0x2084, 0x4094, 0x8088}},
    {13571, 2, {0x1080, 0x10F8, 0x2510, 0x7BFC, 0x1124, 0x2524, 0x7DFC, 0x102, 0x5502, 0x94FE, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {13572, 2, {0x2, 0x7FC2, 0xA02, 0x7FC2, 0x4A52, 0x4A52, 0x7FD2, 0x412, 0x2492, 0x1512, 0xFFD2, 0xE12, 0x1502, 0x2482, 0x444A, 0x404}},
    {13573, 2, {0x402, 0x3F82, 0x482, 0x7FC2, 0x492, 0x3F92, 0x412, 0x5552, 0x75D2, 0x4452, 0x7FD2, 0x4452, 0x75C2, 0x5542, 0x554A, 0x8444}},
    {13574, 2, {0x4, 0xF784, 0x2104, 0xA524, 0xF7A4, 0x6324, 0xB5A4, 0x2124, 0x7FA4, 0x40A4, 0x40A4, 0x7FA4, 0x4084, 0x4084, 0x7F94, 0x4088}},
    {13575, 2, {0x904, 0xF644, 0x1284, 0xA124, 0x40A4, 0xF764, 0x1524, 0x1524, 0x79A4, 0x4024, 0x8F24, 0xF124, 0x1504, 0x1204, 0x5514, 0x2908}},
    {13576, 2, {0x2204, 0x2204, 0xFF84, 0x2224, 0x1424, 0x1424, 0xF7A4, 0x1424, 0x1424, 0x7724, 0x1424, 0x1424, 0xF784, 0x1404, 0x1414, 0x1408}},
    {13577, 2, {0x2, 0x7FC2, 0x4042, 0x7FCA, 0x400A, 0x5FCA, 0x420A, 0x5FCA, 0x524A, 0x5FCA, 0x420A, 0x7FEA, 0x62A2, 0x6FA2, 0xA0AA, 0x2064}},
    {13578, 2, {0x402, 0xFFE2, 0x2, 0x7FC2, 0x4052, 0x5F52, 0x5152, 0x7FD2, 0x12, 0x3F92, 0x2092, 0x3F92, 0x2082, 0x3F82, 0xA, 0xFFE4}},
    {13579, 2, {0x4, 0x7A04, 0x5204, 0x7BA4, 0x4A24, 0x7B24, 0x54A4, 0x78A4, 0x24, 0x7F24, 0x5524, 0x5524, 0x5504, 0x5784, 0xFC14, 0x8}},
    {13580, 2, {0xFC04, 0x4B84, 0x7884, 0x4AA4, 0x7924, 0x4D24, 0xFAA4, 0xCA4, 0x7A4, 0xF824, 0x28A4, 0x4D24, 0x1A04, 0x2904, 0xC894, 0x808}},
    {13581, 2, {0x4204, 0x7BC4, 0xA504, 0x10A4, 0x7FA4, 0x40A4, 0x7FA4, 0x40A4, 0x7FA4, 0x40A4, 0x7FA4, 0x2124, 0xFFC4, 0x2104, 0x4114, 0x8108}},
    {13582, 2, {0x5210, 0x7EFE, 0x2844, 0x7E28, 0xA2FE, 0x3E10, 0x2010, 0x3E7C, 0x2210, 0x3E10, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {13583, 2, {0x2, 0x7FC2, 0x1002, 0x3F8A, 0x508A, 0x108A, 0x1F8A, 0xA, 0xFBEA, 0x208A, 0x79EA, 0x492A, 0xCB22, 0x4922, 0x79EA, 0x4924}},
    {13584, 2, {0x2102, 0x2102, 0x3DE2, 0x5282, 0x8852, 0x7FD2, 0xA12, 0x7BD2, 0xA12, 0x3B92, 0xA12, 0x7BD2, 0xA02, 0xFC2, 0xF80A, 0x4004}},
    {13585, 2, {0x820, 0xFFFE, 0x2820, 0x7DFC, 0x4488, 0x7C50, 0x41FE, 0x7C20, 0x45FC, 0x7C20, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {13586, 2, {0x4442, 0x7FC2, 0x1202, 0x3FC2, 0x6212, 0xBFD2, 0x2212, 0x3FD2, 0x2212, 0x3FD2, 0x2012, 0x7FD2, 0x5142, 0x6EC2, 0x4A4A, 0x4EC4}},
    {13587, 2, {0x420, 0xE20, 0x7820, 0x820, 0x8FC, 0x824, 0x824, 0xFF24, 0x824, 0x824, 0x824, 0x844, 0x844, 0x884, 0x928, 0xA10}},
    {13588, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x2244, 0x227C, 0x4A44, 0x8400, 0x1FF0, 0x2008, 0x2008, 0x2008, 0x1FF0, 0x0}},
    {13589, 2, {0x2020, 0x2020, 0x3F20, 0x4020, 0x80FC, 0x7E24, 0x24, 0x24, 0x7C24, 0x424, 0x524, 0x524, 0x344, 0x144, 0xA8, 0x110}},
    {13590, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x2244, 0x227C, 0x4A44, 0x8400, 0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x10}},
    {13591, 2, {0x1210, 0x1210, 0x1210, 0x7F90, 0x12FE, 0x1292, 0x7F92, 0x5212, 0x5212, 0x7FD2, 0x1252, 0x1252, 0x2352, 0x22A2, 0x422A, 0x8244}},
    {13592, 2, {0x1020, 0x1020, 0x9220, 0x9220, 0x92FC, 0x9224, 0xFE24, 0x1024, 0x1024, 0x9224, 0x9224, 0x9244, 0x9E44, 0xF284, 0x128, 0x210}},
    {13593, 2, {0x4220, 0x2420, 0x20, 0x7E20, 0x24FC, 0x2424, 0x2424, 0x2424, 0xFF24, 0x2424, 0x2424, 0x2424, 0x2444, 0x4444, 0x44A8, 0x8510}},
    {13594, 2, {0x420, 0xE20, 0x3820, 0x820, 0x8FC, 0xFF24, 0x824, 0x824, 0x3E24, 0x2224, 0x2224, 0x2224, 0x2244, 0x3E44, 0x22A8, 0x110}},
    {13595, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {13596, 2, {0x220, 0xF20, 0xF020, 0x120, 0x91FC, 0x4A24, 0x24, 0xFE24, 0x424, 0x824, 0xF24, 0xF844, 0x4844, 0x884, 0x2928, 0x1210}},
    {13597, 2, {0x10, 0xFF90, 0x810, 0x810, 0xFF90, 0x88FE, 0x8892, 0xCC92, 0xAA92, 0xAA92, 0xDD92, 0x9992, 0x8892, 0x88A2, 0x8AAA, 0x8144}},
    {13598, 2, {0x220, 0xF20, 0x7820, 0x820, 0xFF7C, 0x824, 0x2A24, 0x2A24, 0xEBA4, 0x2A24, 0x2AA4, 0x6B24, 0xAA44, 0x844, 0x894, 0x908}},
    {13599, 2, {0x820, 0x2820, 0x3EF8, 0x4828, 0x8A8, 0x3E48, 0x84A, 0xEAA, 0x7926, 0x202, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {13600, 2, {0x20, 0x7E20, 0x4220, 0x7E20, 0x42FC, 0x7E24, 0x2024, 0x7F24, 0x8924, 0x4924, 0x5524, 0x4124, 0x7D44, 0x144, 0xAA8, 0x510}},
    {13601, 2, {0x820, 0x1020, 0x7F20, 0x5520, 0x49FC, 0x5524, 0x4124, 0x7F24, 0x24, 0x4424, 0x7724, 0x4424, 0x5544, 0x6644, 0x44A8, 0x110}},
    {13602, 2, {0x1080, 0x10F8, 0x2510, 0x7BFC, 0x1124, 0x2524, 0x7DFC, 0x102, 0x5502, 0x94FE, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {13603, 2, {0x8, 0x3F88, 0x2488, 0x3F88, 0x24BE, 0x3F8A, 0xA, 0xFBEA, 0xAAAA, 0xAAAA, 0xFBEA, 0xAAAA, 0xAAB2, 0xFBF2, 0x2A, 0x44}},
    {13604, 2, {0x1020, 0x820, 0x7F20, 0x4120, 0x86FC, 0x7824, 0x4924, 0x2A24, 0xFF24, 0x2A24, 0x4124, 0xBE24, 0x2A44, 0x3E44, 0x2AA8, 0x3F10}},
    {13605, 2, {0x10, 0xFF10, 0xA510, 0xA510, 0xFF7E, 0x4412, 0x9492, 0xF712, 0x492, 0xF392, 0x9012, 0xF492, 0x9712, 0xF422, 0x94AA, 0xB3C4}},
    {13606, 2, {0x2210, 0xFF90, 0x810, 0x7F10, 0x107E, 0xFF92, 0x2212, 0x4912, 0xBE92, 0x2212, 0x3E12, 0x2212, 0x3E92, 0x2522, 0x2A2A, 0x3144}},
    {13607, 2, {0x810, 0xF90, 0x810, 0x7F90, 0x48FE, 0x7E12, 0x4892, 0x5F92, 0x5512, 0x5F12, 0x5512, 0x5F12, 0x4012, 0xBFA2, 0x2AAA, 0x7FC4}},
    {13608, 2, {0x1000, 0x1000, 0x3FFC, 0x4524, 0xBDC4, 0x514, 0x1CF4, 0x6404, 0x1FC4, 0x1044, 0x1FC4, 0x1044, 0x1FC4, 0x1044, 0x1154, 0x1088}},
    {13609, 2, {0x1000, 0x1000, 0x3FFC, 0x2204, 0x5FC4, 0x8204, 0x3FE4, 0x404, 0x884, 0x1FC4, 0x4, 0x1FC4, 0x1544, 0x1544, 0x3FE8, 0x10}},
    {13610, 2, {0x1000, 0x1000, 0x3FFC, 0x2204, 0x4204, 0x83F4, 0x204, 0x3FE4, 0x2AA4, 0x2524, 0x3264, 0x2524, 0x2AA4, 0x3FE4, 0x2028, 0x10}},
    {13611, 2, {0x400, 0x400, 0x400, 0x400, 0xFFFE, 0x480, 0x480, 0x484, 0x488, 0x890, 0x8E0, 0x1082, 0x1082, 0x2082, 0x407E, 0x8000}},
    {13612, 2, {0x2030, 0x21C0, 0x3E08, 0x2008, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {13613, 2, {0x0, 0x43F8, 0x4208, 0x4208, 0x43F8, 0x7A08, 0x4208, 0x43F8, 0x4000, 0x47FE, 0x4040, 0x4A40, 0x5278, 0x6240, 0x4540, 0x8FE}},
    {13614, 2, {0x5110, 0x5112, 0xFA54, 0x53D8, 0x7010, 0x2012, 0xFBD2, 0xAA4E, 0xFA40, 0x23D2, 0xFA54, 0x2258, 0xFBD0, 0x2252, 0x5252, 0x8ACE}},
    {13615, 2, {0x0, 0x7FFC, 0x4000, 0x47F0, 0x4410, 0x4410, 0x480C, 0x5000, 0x47F0, 0x4410, 0x4410, 0x4410, 0x47F0, 0x4000, 0x7FFE, 0x0}},
    {13616, 2, {0x0, 0x7FFC, 0x4220, 0x4220, 0x43A0, 0x44B0, 0x44A8, 0x44A4, 0x4AA4, 0x5120, 0x4120, 0x4220, 0x4420, 0x4820, 0x7FFE, 0x0}},
    {13617, 2, {0x0, 0x7FFC, 0x4000, 0x4280, 0x4CB8, 0x4888, 0x4EB8, 0x4888, 0x4FF8, 0x4140, 0x4220, 0x4410, 0x5808, 0x4000, 0x7FFE, 0x0}},
    {13618, 2, {0x0, 0x7FFC, 0x4220, 0x4520, 0x4AA0, 0x5F78, 0x6928, 0x4F28, 0x4928, 0x4F28, 0x482A, 0x4A2A, 0x4D46, 0x4880, 0x7FFE, 0x0}},
    {13619, 2, {0x0, 0x7FFC, 0x4200, 0x4538, 0x4AA8, 0x5F68, 0x6928, 0x4F28, 0x4928, 0x4F2A, 0x482A, 0x4A4A, 0x4D46, 0x4880, 0x7FFE, 0x0}},
    {13620, 2, {0x0, 0x7FFC, 0x4000, 0x4FF8, 0x4888, 0x4FF8, 0x4888, 0x4FF8, 0x4220, 0x4FF8, 0x4220, 0x5FFC, 0x4410, 0x4808, 0x7FFE, 0x0}},
    {13621, 2, {0x7FFE, 0x4000, 0x5E78, 0x5248, 0x5E78, 0x5144, 0x4F3C, 0x4220, 0x4FF8, 0x4220, 0x4220, 0x5FFC, 0x4220, 0x4410, 0x4808, 0x7FFE}},
    {13622, 2, {0x7FFE, 0x4420, 0x7FBC, 0x4A64, 0x7F98, 0x4066, 0x5F7C, 0x5110, 0x5FFE, 0x5144, 0x5F7C, 0x4444, 0x7FFC, 0x4428, 0x4444, 0x7FFE}},
    {13623, 2, {0x0, 0x7FFC, 0x4000, 0x4FF8, 0x4080, 0x4080, 0x4FF8, 0x4888, 0x4948, 0x4A28, 0x4C28, 0x4808, 0x4828, 0x4810, 0x7FFE, 0x0}},
    {13624, 2, {0x7FFE, 0x4000, 0x403E, 0x4FC0, 0x487C, 0x4840, 0x4BFE, 0x4A42, 0x4A78, 0x4AC4, 0x4A3C, 0x5200, 0x54F0, 0x6494, 0x490C, 0x7FFE}},
    {13625, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x11FC, 0x1044, 0xFE44, 0x1044, 0x1044, 0x1084, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {13626, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x1100, 0x1120, 0x1020, 0xFFFE, 0x1020, 0x1020, 0x1022, 0x1002, 0x1002, 0xFFE, 0x0}},
    {13627, 2, {0x0, 0x7FFC, 0x100, 0x300, 0x560, 0x1918, 0x6104, 0x100, 0x0, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {13628, 2, {0x10A0, 0x1090, 0xFDFE, 0x1320, 0x1DFC, 0x3120, 0xD1FC, 0x1120, 0x51FE, 0x2100, 0x0, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {13629, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13630, 2, {0x0, 0x3FE0, 0x2020, 0x2020, 0x2020, 0x2020, 0x2020, 0x2020, 0x2140, 0x2080, 0x2000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {13631, 2, {0x0, 0xFFFE, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1070, 0x1190, 0x1610, 0x1810, 0x10, 0x10, 0x10}},
    {13632, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1052, 0x1022, 0x1002, 0xFFE, 0x0}},
    {13633, 2, {0x100, 0x3200, 0xC7C, 0x1244, 0x6944, 0x844, 0xFFC4, 0x1044, 0x2444, 0x2444, 0x4844, 0x8A54, 0x1148, 0x3F40, 0x1140, 0x40}},
    {13634, 2, {0x0, 0x3FFE, 0x2080, 0x2080, 0x2140, 0x2220, 0x2410, 0x2908, 0x3086, 0x2080, 0x2FF0, 0x2010, 0x2020, 0x4020, 0x4040, 0x8080}},
    {13635, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x27FC, 0x2110, 0x2110, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2110, 0x2210, 0x4210, 0x4410, 0x8810}},
    {13636, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x2FFC, 0x2080, 0x2100, 0x2210, 0x2420, 0x2FC0, 0x2080, 0x2100, 0x2210, 0x4408, 0x4FFC, 0x8404}},
    {13637, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2040, 0x4040, 0x4040, 0x8040}},
    {13638, 2, {0x0, 0x3FFE, 0x2000, 0x2040, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x27FC, 0x2404, 0x2404, 0x2404, 0x4404, 0x47FC, 0x8404}},
    {13639, 2, {0x0, 0x3FFE, 0x2000, 0x2080, 0x2040, 0x2040, 0x2FFC, 0x2000, 0x2408, 0x2208, 0x2210, 0x2110, 0x2120, 0x4000, 0x5FFE, 0x8000}},
    {13640, 2, {0x0, 0x3FFE, 0x2080, 0x2040, 0x2000, 0x3FFE, 0x2080, 0x2100, 0x2284, 0x2688, 0x2A50, 0x3220, 0x2210, 0x4288, 0x4306, 0x8200}},
    {13641, 2, {0x0, 0x3FFE, 0x2080, 0x2080, 0x2140, 0x2220, 0x2410, 0x2BE8, 0x3006, 0x2000, 0x27F0, 0x2410, 0x2410, 0x4410, 0x47F0, 0x8410}},
    {13642, 2, {0x0, 0x3FFE, 0x2000, 0x2048, 0x2044, 0x2FFE, 0x2040, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x4444, 0x4454, 0x8408}},
    {13643, 2, {0x0, 0x3FFE, 0x2080, 0x2140, 0x2220, 0x2410, 0x2808, 0x37F6, 0x2080, 0x2080, 0x2FF8, 0x2080, 0x2490, 0x42A0, 0x5FFE, 0x8000}},
    {13644, 2, {0x0, 0x3FFE, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2000, 0x2FFC, 0x2010, 0x3FFE, 0x2410, 0x4210, 0x4250, 0x8020}},
    {13645, 2, {0x0, 0x3FFE, 0x2280, 0x2240, 0x2240, 0x27FC, 0x2440, 0x2C40, 0x37F8, 0x2440, 0x2440, 0x27F8, 0x2440, 0x4440, 0x47FC, 0x8400}},
    {13646, 2, {0x0, 0x3FFE, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x3FFE, 0x2200, 0x2408, 0x2FFC, 0x2000, 0x2FFC, 0x2924, 0x4924, 0x7FFE, 0x8000}},
    {13647, 2, {0x0, 0x3FFE, 0x2000, 0x3F88, 0x2048, 0x2F28, 0x2908, 0x2948, 0x2928, 0x2F0E, 0x2078, 0x2908, 0x2608, 0x4388, 0x5C08, 0x8808}},
    {13648, 2, {0x0, 0x3FFE, 0x2100, 0x2638, 0x2408, 0x2738, 0x2408, 0x27F8, 0x2200, 0x27FC, 0x2804, 0x2554, 0x2554, 0x4804, 0x4028, 0x8010}},
    {13649, 2, {0x0, 0x3FFE, 0x2204, 0x2204, 0x3FC4, 0x2204, 0x2FBE, 0x2004, 0x2FA4, 0x2894, 0x2F94, 0x2884, 0x2504, 0x4384, 0x5C14, 0x8808}},
    {13650, 2, {0x3FFE, 0x2408, 0x2210, 0x3FFE, 0x2080, 0x2FFC, 0x2080, 0x3FFE, 0x20A0, 0x2F28, 0x2224, 0x3FFE, 0x4228, 0x5F12, 0x822A, 0x6C6}},
    {13651, 2, {0x0, 0x3FFE, 0x2000, 0x27FC, 0x2554, 0x24E4, 0x27FC, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x24A4, 0x2852, 0x47FC, 0x4040, 0x8FFE}},
    {13652, 2, {0x3FFE, 0x2010, 0x3F10, 0x2AFE, 0x2E10, 0x2AFE, 0x2EAA, 0x2AAA, 0x3EFE, 0x2200, 0x3EFE, 0x2A20, 0x5E14, 0x4AA2, 0xBF28, 0x218}},
    {13653, 2, {0x100, 0x100, 0x200, 0x420, 0x810, 0x1008, 0x3FFC, 0x1004, 0x100, 0x100, 0x200, 0x420, 0x810, 0x1008, 0x3FFC, 0x1004}},
    {13654, 2, {0x400, 0x820, 0x1010, 0x3FF8, 0x810, 0x1020, 0x2448, 0x7EFC, 0x204, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {13655, 2, {0x4100, 0x2210, 0xFF90, 0x810, 0x7F20, 0x820, 0xFFA0, 0x1040, 0x1048, 0x3E84, 0x42FC, 0x8404, 0x800, 0x1400, 0x23FE, 0xC000}},
    {13656, 2, {0x400, 0x820, 0x1010, 0x3FF8, 0x810, 0x1020, 0x2448, 0x7EFC, 0x100, 0x6C0, 0x1830, 0xE10E, 0x920, 0x1110, 0x2508, 0x200}},
    {13657, 2, {0x1080, 0x1080, 0x20F0, 0x4510, 0xFE20, 0x29FC, 0x4524, 0xA324, 0x3D24, 0x45FC, 0x4450, 0xA858, 0x1094, 0x2892, 0x4512, 0x820E}},
    {13658, 2, {0x0, 0x3FF0, 0x1010, 0x1210, 0x1120, 0x920, 0x820, 0x4440, 0x2440, 0x2280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {13659, 2, {0x0, 0x3FF0, 0x20, 0x1E40, 0x180, 0x660, 0x1818, 0xE004, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC006}},
    {13660, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC006}},
    {13661, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {13662, 2, {0x0, 0x3F00, 0x217C, 0x2124, 0x3F24, 0x2424, 0x2424, 0x2428, 0x3F28, 0x3528, 0x3510, 0x5510, 0x5528, 0x9728, 0x444, 0x482}},
    {13663, 2, {0x800, 0x800, 0xF00, 0x8FC, 0x844, 0x7F44, 0x5544, 0x5544, 0x5528, 0x7F28, 0x5510, 0x5510, 0x5528, 0x5128, 0x4544, 0x4282}},
    {13664, 2, {0x4140, 0x2120, 0x3FE, 0xA20, 0x17FC, 0xEA20, 0x23FC, 0x2220, 0x23FE, 0x200, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {13665, 2, {0x800, 0xF00, 0x87C, 0xFFA4, 0x80A4, 0x7F24, 0x1424, 0x2228, 0x4928, 0x1628, 0x2110, 0xFE10, 0x2228, 0x2228, 0x3E44, 0x2282}},
    {13666, 2, {0x1000, 0x1E00, 0x107C, 0xFF44, 0x8144, 0x7E44, 0x44, 0x7E28, 0x4228, 0x7E28, 0x4210, 0x7E10, 0x4228, 0x7E28, 0x2444, 0x4282}},
    {13667, 2, {0x40, 0x840, 0x820, 0x1020, 0x1010, 0x2008, 0x4004, 0x9FF2, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0}},
    {13668, 2, {0x100, 0x100, 0x7900, 0x4900, 0x4BF0, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x7910, 0x4A12, 0x212, 0x412, 0x80E, 0x1000}},
    {13669, 2, {0x40, 0x40, 0x7C40, 0x4440, 0x4440, 0x4440, 0x4440, 0x4440, 0x44A0, 0x44A0, 0x7CA0, 0x4510, 0x110, 0x208, 0x404, 0x802}},
    {13670, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {13671, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4900, 0x4900, 0x7900, 0x4900, 0x200, 0x200, 0x400, 0x800}},
    {13672, 2, {0x0, 0x1F0, 0x7910, 0x4910, 0x4910, 0x4910, 0x4990, 0x4950, 0x4950, 0x4910, 0x7910, 0x4912, 0x212, 0x212, 0x40E, 0x800}},
    {13673, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x20, 0x20, 0x20, 0x20}},
    {13674, 2, {0x40, 0x40, 0x7840, 0x4BFC, 0x4A44, 0x4C48, 0x4840, 0x4840, 0x48A0, 0x48A0, 0x78A0, 0x4920, 0x122, 0x222, 0x41E, 0x800}},
    {13675, 2, {0x80, 0x40, 0x7840, 0x4FFE, 0x4900, 0x4900, 0x4900, 0x49F8, 0x4908, 0x4908, 0x7908, 0x4908, 0x208, 0x208, 0x450, 0x820}},
    {13676, 2, {0x20, 0x120, 0x7920, 0x4910, 0x4910, 0x4A48, 0x4A48, 0x4C44, 0x4882, 0x4880, 0x7910, 0x4908, 0x208, 0x7FC, 0x204, 0x0}},
    {13677, 2, {0x10, 0x78, 0x7BC0, 0x4840, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x4840, 0x48A0, 0x78A0, 0x4890, 0x110, 0x108, 0x204, 0x402}},
    {13678, 2, {0x20, 0x110, 0x7908, 0x4A04, 0x4A04, 0x4C10, 0x4910, 0x4910, 0x48A0, 0x48A0, 0x7840, 0x48A0, 0x110, 0x208, 0x404, 0x802}},
    {13679, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x4284, 0x4444, 0x4824, 0x57D4, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4014, 0x4008}},
    {13680, 2, {0x8, 0x3C, 0x79E0, 0x4820, 0x4820, 0x483C, 0x49E0, 0x4820, 0x4820, 0x483E, 0x7BE0, 0x4820, 0x22, 0x22, 0x22, 0x1E}},
    {13681, 2, {0x90, 0x88, 0x7888, 0x4880, 0x4FFE, 0x48A0, 0x48A0, 0x48A0, 0x48A0, 0x48A0, 0x7920, 0x4922, 0x122, 0x222, 0x21E, 0x400}},
    {13682, 2, {0x40, 0x40, 0xF040, 0x9040, 0x97FE, 0x9040, 0x90E0, 0x90E0, 0x9150, 0x9150, 0xF248, 0x9444, 0x842, 0x40, 0x40, 0x40}},
    {13683, 2, {0x0, 0x8, 0x7908, 0x4888, 0x4A52, 0x4A52, 0x4A22, 0x4A22, 0x4A52, 0x4A92, 0x7B0A, 0x4A0A, 0x202, 0x3FE, 0x2, 0x0}},
    {13684, 2, {0x0, 0x0, 0x7BFE, 0x4A02, 0x4C04, 0x4800, 0x49F0, 0x4910, 0x4910, 0x4910, 0x7910, 0x4912, 0x212, 0x212, 0x40E, 0x800}},
    {13685, 2, {0x0, 0x3FE, 0x7820, 0x4820, 0x4820, 0x4820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0x20, 0x7FE, 0x0}},
    {13686, 2, {0x0, 0x3FC, 0x7840, 0x4840, 0x4840, 0x4840, 0x4BF8, 0x4888, 0x4888, 0x4888, 0x7888, 0x4908, 0x108, 0x108, 0x7FE, 0x0}},
    {13687, 2, {0x20, 0x20, 0x7840, 0x49FC, 0x4904, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x7904, 0x4904, 0x104, 0x1FC, 0x104}},
    {13688, 2, {0x80, 0x40, 0x7840, 0x4800, 0x4FFC, 0x4800, 0x4808, 0x4A08, 0x4A08, 0x4910, 0x7910, 0x4910, 0x120, 0x20, 0xFFE, 0x0}},
    {13689, 2, {0x40, 0x48, 0x7844, 0x4840, 0x4BFE, 0x4880, 0x4880, 0x48FC, 0x4944, 0x4944, 0x7948, 0x4A50, 0x220, 0x450, 0x888, 0x306}},
    {13690, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7CF8, 0x0, 0x0, 0xFFFE, 0x800, 0x1000, 0x1FF0, 0x10, 0x10, 0xA0, 0x40}},
    {13691, 2, {0x200, 0x100, 0x0, 0x7FFC, 0x100, 0x300, 0xD60, 0x3118, 0xC104, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {13692, 2, {0x100, 0x100, 0x79FE, 0x4A02, 0x4A02, 0x4C8A, 0x4852, 0x4A22, 0x4A52, 0x4A8A, 0x7A02, 0x4BFE, 0x2, 0x2, 0x14, 0x8}},
    {13693, 2, {0x8, 0x88, 0x7848, 0x4A28, 0x4A28, 0x4A08, 0x4A08, 0x4A08, 0x4A08, 0x4A08, 0x4A10, 0x7A98, 0x4B24, 0x224, 0x42, 0x82}},
    {13694, 2, {0x100, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x0, 0x40, 0x3E44, 0x2248, 0x2270, 0x2240, 0x2242, 0x3E42, 0x223E}},
    {13695, 2, {0x70, 0x1F80, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC006, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {13696, 2, {0x40, 0x40, 0x7BF8, 0x4848, 0x4FFE, 0x4848, 0x4BF8, 0x4840, 0x4840, 0x4BF8, 0x7840, 0x4840, 0x7FC, 0x40, 0x40, 0x40}},
    {13697, 2, {0x8, 0x190, 0x7860, 0x4898, 0x4B04, 0x4840, 0x4840, 0x4FFE, 0x4880, 0x4890, 0x7910, 0x4920, 0x248, 0x484, 0x9FC, 0x84}},
    {13698, 2, {0x40, 0x40, 0x7BF8, 0x4A48, 0x4A48, 0x4BF8, 0x4A48, 0x4A48, 0x4BF8, 0x4840, 0x7848, 0x4830, 0x64, 0x194, 0x60C, 0x4}},
    {13699, 2, {0x8, 0x3C, 0x7BC0, 0x4A00, 0x4A00, 0x4BFE, 0x4A00, 0x4A00, 0x4A00, 0x4AFC, 0x7A84, 0x4A84, 0x484, 0x484, 0x8FC, 0x84}},
    {13700, 2, {0x104, 0x124, 0x7924, 0x4924, 0x4924, 0x4924, 0x4BB4, 0x4B6C, 0x4D24, 0x4924, 0x7924, 0x4924, 0x124, 0x224, 0x204, 0x404}},
    {13701, 2, {0x28, 0x24, 0x7824, 0x4820, 0x4BFE, 0x4A20, 0x4A24, 0x4A24, 0x4A24, 0x4BA8, 0x7A28, 0x4A10, 0x212, 0x42A, 0x446, 0x882}},
    {13702, 2, {0x50, 0x50, 0x7850, 0x4850, 0x4BFE, 0x4A52, 0x4A52, 0x4A52, 0x4A52, 0x4BFE, 0x7A52, 0x4A52, 0x252, 0x252, 0x3FE, 0x202}},
    {13703, 2, {0x0, 0x3FE, 0x7820, 0x4820, 0x4840, 0x4BFE, 0x4A52, 0x4A52, 0x4A52, 0x4A52, 0x7A52, 0x4A52, 0x252, 0x242, 0x20A, 0x204}},
    {13704, 2, {0x208, 0x3FD0, 0x220, 0xFFFE, 0x200, 0xC70, 0x3F84, 0xC804, 0x7FC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {13705, 2, {0x20, 0x20, 0x7850, 0x4888, 0x4924, 0x4A22, 0x4C20, 0x4920, 0x493C, 0x4920, 0x7920, 0x4920, 0x120, 0x7FE, 0x0, 0x0}},
    {13706, 2, {0x0, 0x3FC, 0x7804, 0x4804, 0x49FC, 0x4804, 0x4804, 0x4BFC, 0x4808, 0x4808, 0x4FFE, 0x7908, 0x4888, 0x88, 0x28, 0x10}},
    {13707, 2, {0x100, 0x280, 0x440, 0x1830, 0xE44E, 0x440, 0x440, 0x440, 0x840, 0x1040, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {13708, 2, {0x880, 0x880, 0x13F0, 0x3090, 0x5090, 0x9092, 0x1112, 0x120E, 0x1400, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13709, 2, {0x40, 0x440, 0x240, 0x1240, 0x840, 0x9FC, 0x7E40, 0x40, 0x40, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13710, 2, {0x80, 0x7C88, 0x44B0, 0x44C4, 0x4484, 0x7C7C, 0x100, 0x280, 0x440, 0x1830, 0xE44E, 0x440, 0x440, 0x440, 0x840, 0x1040}},
    {13711, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4820, 0x4820, 0x79FC, 0x4820, 0x20, 0x20, 0x3FE, 0x0}},
    {13712, 2, {0x7CF8, 0x4488, 0x4488, 0x7CF8, 0x0, 0x3FF0, 0x210, 0x210, 0xFFFE, 0x210, 0x410, 0x3FF0, 0x810, 0x1000, 0x2000, 0xC000}},
    {13713, 2, {0x14, 0x12, 0x7810, 0x4FFE, 0x4810, 0x4A90, 0x4A92, 0x4A92, 0x4FD2, 0x4A94, 0x7A94, 0x4A88, 0x28A, 0x49A, 0x426, 0x842}},
    {13714, 2, {0x108, 0x88, 0x7890, 0x4BFC, 0x4824, 0x4824, 0x4BFC, 0x4A20, 0x4A20, 0x4BFE, 0x7862, 0x48A2, 0x12A, 0x224, 0x420, 0x20}},
    {13715, 2, {0x100, 0x1F8, 0x7B08, 0x4C90, 0x4860, 0x4998, 0x4E46, 0x4840, 0x4BF8, 0x4840, 0x7BF8, 0x4840, 0x7FC, 0x40, 0x40, 0x40}},
    {13716, 2, {0x0, 0x1FC, 0x7820, 0x4820, 0x4820, 0x4BFE, 0x4850, 0x4888, 0x4904, 0x4AFA, 0x7C88, 0x4888, 0x88, 0x88, 0xF8, 0x88}},
    {13717, 2, {0x40, 0x20, 0x79FC, 0x4800, 0x4888, 0x4850, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x79FC, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {13718, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13719, 2, {0x40, 0x248, 0x7A48, 0x4A48, 0x4BF8, 0x4840, 0x48A0, 0x4910, 0x4A88, 0x4C46, 0x7840, 0x4BF0, 0x10, 0x20, 0x20, 0x40}},
    {13720, 2, {0x0, 0x3FC, 0x7A00, 0x4A00, 0x4AF8, 0x4A00, 0x4A00, 0x4BFC, 0x4AA0, 0x4AA4, 0x7AA8, 0x4A90, 0x290, 0x488, 0x4A4, 0x8C2}},
    {13721, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x227C, 0x4A44, 0x8400, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13722, 2, {0x810, 0x810, 0x17FE, 0x3010, 0x5210, 0x9110, 0x1110, 0x1050, 0x1020, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13723, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13724, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13725, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x920, 0x3118, 0xC106, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13726, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x4904, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x4A22, 0x7A52, 0x4A8A, 0x30A, 0x202, 0x20A, 0x204}},
    {13727, 2, {0x1C, 0x3E0, 0xF084, 0x9244, 0x9148, 0x9110, 0x97FE, 0x9402, 0x9000, 0x93F8, 0xF108, 0x9110, 0xA0, 0x40, 0x1B0, 0x60E}},
    {13728, 2, {0x20, 0x20, 0x7BFE, 0x4850, 0x4888, 0x4904, 0x4AFA, 0x4800, 0x4800, 0x4BFE, 0x7820, 0x4924, 0x122, 0x222, 0xA0, 0x40}},
    {13729, 2, {0x80, 0x40, 0xF7FE, 0x9120, 0x9120, 0x923C, 0x9244, 0x9664, 0x9294, 0x9348, 0xF248, 0x9230, 0x220, 0x250, 0x288, 0x306}},
    {13730, 2, {0x80, 0x40, 0xEFFE, 0xA800, 0xA890, 0xA890, 0xAA90, 0xAA92, 0xAADC, 0xAA90, 0xEA90, 0xAA92, 0x1292, 0x12D2, 0x270E, 0x0}},
    {13731, 2, {0x40, 0x20, 0x79FC, 0x4800, 0x4908, 0x4890, 0x4800, 0x4BFE, 0x4800, 0x4800, 0x79FC, 0x4904, 0x104, 0x104, 0x1FC, 0x104}},
    {13732, 2, {0x0, 0x1FC, 0x7808, 0x4810, 0x4A22, 0x4AAA, 0x4A72, 0x4A22, 0x4A72, 0x4AAA, 0x7B26, 0x4AA2, 0x242, 0x202, 0x3FE, 0x2}},
    {13733, 2, {0x0, 0x7FC, 0xF404, 0x9434, 0x95C4, 0x9444, 0x9444, 0x97FC, 0x9444, 0x94E4, 0xF554, 0x964C, 0x444, 0x444, 0x7FC, 0x404}},
    {13734, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9248, 0x93F8, 0x9248, 0x93F8, 0x9040, 0xF0E0, 0x9150, 0x248, 0xC46, 0x40, 0x40}},
    {13735, 2, {0x1C, 0x3E0, 0x7A20, 0x4BFE, 0x4A20, 0x4A92, 0x4B0A, 0x4A06, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104, 0x104, 0x1FC, 0x104}},
    {13736, 2, {0x0, 0x1F8, 0x7808, 0x4808, 0x49F8, 0x4808, 0x4808, 0x4BFE, 0x4820, 0x4A22, 0x7974, 0x48A8, 0x124, 0x222, 0xA0, 0x40}},
    {13737, 2, {0x0, 0xFC0, 0xF4BC, 0x94A4, 0x94A4, 0x97A4, 0x94A4, 0x94A8, 0x97A8, 0x94A8, 0xF490, 0x95D0, 0xEA8, 0xA8, 0xC4, 0x82}},
    {13738, 2, {0x14, 0x12, 0x7810, 0x4FFE, 0x4810, 0x4810, 0x4BD2, 0x4A52, 0x4A52, 0x4A54, 0x7BD4, 0x4808, 0xEA, 0x71A, 0x226, 0x42}},
    {13739, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x4BFE, 0x4888, 0x4944, 0x4A42, 0x48F8, 0x7988, 0x4A50, 0x20, 0x50, 0x188, 0x606}},
    {13740, 2, {0x80, 0x31C, 0x7A04, 0x4A04, 0x4B9C, 0x4A04, 0x4A04, 0x4BFC, 0x4800, 0x4800, 0x7BF8, 0x4908, 0x90, 0x60, 0x198, 0x606}},
    {13741, 2, {0x0, 0x3FC, 0xF040, 0x9040, 0x97FE, 0x90A0, 0x9110, 0x9208, 0x9446, 0x9040, 0xF048, 0x9264, 0x252, 0x452, 0x140, 0x80}},
    {13742, 2, {0x40, 0x40, 0xF0A0, 0x9110, 0x9208, 0x9DF6, 0x9000, 0x97FC, 0x94A4, 0x94A4, 0xF7FC, 0x94A4, 0x4A4, 0x4A4, 0x414, 0x408}},
    {13743, 2, {0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2028, 0x2010, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13744, 2, {0x80, 0x7C88, 0x44B0, 0x44C4, 0x4484, 0x7C7C, 0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2028, 0x2010}},
    {13745, 2, {0x100, 0xFFFE, 0x880, 0x10F8, 0x3148, 0xD328, 0x1490, 0x1060, 0x1198, 0x1606, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13746, 2, {0x1000, 0x1000, 0xFE7C, 0x2244, 0x4444, 0x3444, 0x844, 0x147C, 0x2244, 0xC000, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13747, 2, {0x1080, 0x880, 0x47FC, 0x2110, 0xA10, 0x11A0, 0x7060, 0x1090, 0x1308, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13748, 2, {0x110, 0x110, 0x7BFE, 0x4910, 0x4880, 0x4880, 0x49FC, 0x4954, 0x4A54, 0x4C94, 0x78A4, 0x4924, 0x244, 0x84, 0x128, 0x210}},
    {13749, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x4908, 0x49F8, 0x4908, 0x49F8, 0x4908, 0x7908, 0x4FFE, 0x0, 0x110, 0x208, 0x404}},
    {13750, 2, {0x0, 0x7FC, 0xF0A0, 0x90A0, 0x97FC, 0x94A4, 0x94A4, 0x94A4, 0x97FC, 0x9040, 0xF040, 0x97FC, 0x40, 0x40, 0xFFE, 0x0}},
    {13751, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4800, 0x4BFE, 0x7820, 0x4920, 0x13C, 0x120, 0x2A0, 0x47E}},
    {13752, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x4A78, 0x4BC0, 0x4A44, 0x4A3C, 0x4A00, 0x7AFC, 0x4A80, 0x3FE, 0x480, 0x480, 0x8FC}},
    {13753, 2, {0x10, 0x110, 0x7908, 0x4A04, 0x4DFA, 0x4888, 0x4908, 0x4A28, 0x4C10, 0x4BF8, 0x7AA8, 0x4AA8, 0x2A8, 0x2A8, 0xFFE, 0x0}},
    {13754, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9080, 0x97FC, 0x9110, 0x9208, 0x95F6, 0xF110, 0x9110, 0x1F0, 0x110, 0x110, 0x1F0}},
    {13755, 2, {0x0, 0x3F8, 0xF090, 0x9060, 0x97FC, 0x90A4, 0x9128, 0x9220, 0x9460, 0x9040, 0xF7FC, 0x90E0, 0x150, 0x248, 0xC46, 0x40}},
    {13756, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4888, 0x48F8, 0x4888, 0x4888, 0x48F8, 0x4820, 0x7BFE, 0x4870, 0xA8, 0x124, 0x222, 0x20}},
    {13757, 2, {0x0, 0x3FE, 0x7820, 0x4840, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x0, 0x88, 0x104}},
    {13758, 2, {0x0, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x1010, 0x820, 0xFFFE, 0x100, 0x2108, 0x2108, 0x3FF8, 0x408, 0x800, 0x3000, 0xC000}},
    {13759, 2, {0x0, 0xFDE, 0xF492, 0x9492, 0x9494, 0x9794, 0x9498, 0x9494, 0x9792, 0x9492, 0xF492, 0x95DA, 0xE94, 0x90, 0x90, 0x90}},
    {13760, 2, {0x40, 0x40, 0xE148, 0xA148, 0xA94A, 0xAA52, 0xAAD6, 0xAB5A, 0xA94A, 0xAA52, 0xED6A, 0xAF7A, 0x94A, 0x842, 0xFFE, 0x2}},
    {13761, 2, {0x0, 0x3DE, 0x7A52, 0x4A52, 0x4BDE, 0x4800, 0x49FC, 0x4800, 0x4BFE, 0x4880, 0x7900, 0x49FC, 0x4, 0x4, 0x28, 0x10}},
    {13762, 2, {0x20, 0x20, 0x7924, 0x4922, 0x4A2A, 0x4830, 0x48C0, 0x4F00, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x104, 0x1FC, 0x104}},
    {13763, 2, {0x0, 0x3FC, 0x7A04, 0x4BFC, 0x4A20, 0x4BFE, 0x4A10, 0x4A8A, 0x4B06, 0x4800, 0x7BFC, 0x4A04, 0x3FC, 0x204, 0x3FC, 0x204}},
    {13764, 2, {0x0, 0x3FC, 0x7A94, 0x4A94, 0x4A94, 0x4BFC, 0x4880, 0x4840, 0x4FFE, 0x4880, 0x78F8, 0x4888, 0x108, 0x108, 0x228, 0x410}},
    {13765, 2, {0x100, 0x13C, 0x7924, 0x4FE4, 0x4924, 0x493C, 0x4924, 0x4BA4, 0x4AA4, 0x4ABC, 0x7AA4, 0x4AA4, 0x3A4, 0x44, 0x54, 0x88}},
    {13766, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48F8, 0x4820, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x7820, 0x4BFE, 0x20, 0x20, 0x20, 0x20}},
    {13767, 2, {0x0, 0x3DE, 0x7A42, 0x4A42, 0x4A42, 0x4BDE, 0x4A00, 0x4A3E, 0x4BD2, 0x4A12, 0x7A14, 0x4BD4, 0x208, 0x214, 0x224, 0x242}},
    {13768, 2, {0x0, 0x3FC, 0x7888, 0x4850, 0x4BFE, 0x4A22, 0x4BFE, 0x4A22, 0x4BFE, 0x4A22, 0x7840, 0x4BFE, 0x42, 0x82, 0x114, 0x208}},
    {13769, 2, {0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x400, 0xFE0, 0x1440, 0x380, 0x1C70, 0xEFEE, 0x820, 0xFE0, 0x820}},
    {13770, 2, {0x0, 0x3C78, 0x2448, 0x2448, 0x2448, 0x2448, 0x3C78, 0x0, 0x0, 0x7EFC, 0x4284, 0x4284, 0x4284, 0x4284, 0x7EFC, 0x4284}},
    {13771, 2, {0x80, 0x80, 0x79FC, 0x4A04, 0x4C08, 0x4BFE, 0x4A00, 0x4A80, 0x4AFC, 0x4B20, 0x7A20, 0x4BFE, 0x420, 0x450, 0x888, 0x1306}},
    {13772, 2, {0x40, 0x20, 0x79FE, 0x4902, 0x4A04, 0x49FC, 0x4840, 0x4888, 0x49FC, 0x4824, 0x7820, 0x49FC, 0x20, 0x20, 0x3FE, 0x0}},
    {13773, 2, {0x38, 0x3C0, 0xF040, 0x9040, 0x97FC, 0x9150, 0x9248, 0x9446, 0x93F8, 0x9208, 0xF208, 0x93F8, 0x208, 0x208, 0x3F8, 0x208}},
    {13774, 2, {0x200, 0x200, 0xF7BC, 0x92A4, 0x92A4, 0x94A4, 0x95BC, 0x9840, 0x9040, 0x97FC, 0xF0E0, 0x9150, 0x248, 0x444, 0x842, 0x40}},
    {13775, 2, {0x0, 0x3FE, 0x7810, 0x4820, 0x4868, 0x48A4, 0x4B22, 0x4820, 0x4800, 0x4BFE, 0x7820, 0x4920, 0x13C, 0x120, 0x120, 0x7FE}},
    {13776, 2, {0x40, 0x40, 0x7BFC, 0x4840, 0x48A0, 0x4910, 0x4A88, 0x4C46, 0x4800, 0x4BF8, 0x7AA8, 0x4AA8, 0x2A8, 0x2A8, 0xFFE, 0x0}},
    {13777, 2, {0x100, 0x3BC, 0xEE10, 0xA210, 0xA23C, 0xAF90, 0xA210, 0xA23C, 0xA210, 0xAF90, 0xE27E, 0xA210, 0x210, 0x210, 0x410, 0x810}},
    {13778, 2, {0x80, 0x40, 0xF7FE, 0x9000, 0x93F8, 0x9208, 0x9208, 0x9208, 0x93F8, 0x90A4, 0xF128, 0x9310, 0x510, 0x948, 0x186, 0x100}},
    {13779, 2, {0x10, 0x420, 0xF2F8, 0x9288, 0x9088, 0x90F8, 0x9E80, 0x9280, 0x92FC, 0x9284, 0xF284, 0x92FC, 0x284, 0x500, 0x8FE, 0x0}},
    {13780, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x4A78, 0x4BC4, 0x4A3C, 0x4A40, 0x4A20, 0x7BFE, 0x4A88, 0x250, 0x420, 0x458, 0x986}},
    {13781, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x49FC, 0x4888, 0x4BFE, 0x4820, 0x49FC, 0x4924, 0x79FC, 0x4924, 0x3FE, 0x104, 0x114, 0x108}},
    {13782, 2, {0x3C, 0x3C0, 0x7844, 0x4A24, 0x4928, 0x4900, 0x4840, 0x4B9C, 0x4A04, 0x4A04, 0x7B9C, 0x4A04, 0x204, 0x204, 0x3FC, 0x204}},
    {13783, 2, {0x40, 0x7C80, 0x45FC, 0x4524, 0x4524, 0x7DFC, 0x2124, 0x2144, 0x7DFC, 0x2490, 0x2510, 0x27FE, 0x4410, 0x4410, 0x9410, 0x810}},
    {13784, 2, {0x28, 0x24, 0x7BFE, 0x4820, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x7808, 0x4BFE, 0x88, 0x48, 0x48, 0x18}},
    {13785, 2, {0x80, 0x80, 0x79FC, 0x4A44, 0x4954, 0x49F4, 0x4884, 0x4928, 0x4A90, 0x49FC, 0x7A44, 0x4D54, 0x1F4, 0x84, 0x128, 0x210}},
    {13786, 2, {0x0, 0x7FC, 0xF0A0, 0x90A0, 0x97FC, 0x94A4, 0x94A4, 0x97FC, 0x9040, 0x9040, 0xF7FC, 0x90E0, 0x150, 0x248, 0xC46, 0x40}},
    {13787, 2, {0x210, 0x210, 0xF3DE, 0x9528, 0x9884, 0x9038, 0x93C0, 0x9040, 0x9040, 0x97FC, 0xF040, 0x90A0, 0xA0, 0x110, 0x208, 0xC06}},
    {13788, 2, {0x80, 0x40, 0xF7FC, 0x9080, 0x9108, 0x93F0, 0x9060, 0x9184, 0x97FE, 0x9002, 0xF7FC, 0x9444, 0x7FC, 0x444, 0x7FC, 0x404}},
    {13789, 2, {0x104, 0x84, 0x7888, 0x4800, 0x4BFE, 0x4820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x7BFE, 0x4800, 0x2A4, 0x252, 0x452, 0x0}},
    {13790, 2, {0x8, 0x8, 0x7788, 0x527E, 0x524A, 0x544A, 0x5748, 0x5D7C, 0x5554, 0x5554, 0x7554, 0x5754, 0x548, 0x88, 0x114, 0x22}},
    {13791, 2, {0x2080, 0x1040, 0x87FC, 0x4040, 0x1040, 0x23F8, 0xE040, 0x2040, 0x27FC, 0x2000, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {13792, 2, {0x840, 0x840, 0x7E84, 0x13FE, 0x2202, 0x14FC, 0x884, 0x1484, 0x62FC, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {13793, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x27F8, 0x6110, 0xA13C, 0x2104, 0x2214, 0x2408, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {13794, 2, {0x1C8, 0x708, 0x7108, 0x5FE8, 0x511E, 0x57CA, 0x554A, 0x57CA, 0x554A, 0x57CA, 0x710A, 0x57CA, 0x112, 0x1D2, 0xE2A, 0x444}},
    {13795, 2, {0x2A8, 0x2A8, 0x77FC, 0x52A8, 0x52AA, 0x54E6, 0x5800, 0x57FC, 0x5444, 0x5040, 0x73F8, 0x5248, 0x248, 0x248, 0x258, 0x40}},
    {13796, 2, {0x28, 0x24, 0x3FFE, 0x2220, 0x2324, 0x2224, 0x2FA8, 0x2218, 0x2A92, 0x522A, 0x4646, 0x8082, 0xFF8, 0x808, 0xFF8, 0x808}},
    {13797, 2, {0x200, 0x11E, 0x77D2, 0x5012, 0x57D4, 0x5454, 0x57D8, 0x5014, 0x57D2, 0x5092, 0x7112, 0x51DA, 0x714, 0x110, 0x510, 0x210}},
    {13798, 2, {0x10, 0x1F20, 0xEACE, 0xAAAA, 0xAEAA, 0xAAAA, 0xAAAA, 0xAEAA, 0xAAAA, 0xAAAA, 0xEBEA, 0xAEAC, 0x1A28, 0x248, 0x248, 0x288}},
    {13799, 2, {0x40, 0x20, 0x7BFE, 0x4A8A, 0x4904, 0x4BFE, 0x4840, 0x4888, 0x49FC, 0x4824, 0x7820, 0x49FC, 0x20, 0x20, 0x3FE, 0x0}},
    {13800, 2, {0x0, 0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13801, 2, {0x0, 0x3DE, 0x7842, 0x4A52, 0x494A, 0x4A52, 0x4842, 0x4820, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104, 0x104, 0x1FC, 0x104}},
    {13802, 2, {0x90, 0x92, 0x7AD4, 0x4A9A, 0x4AD2, 0x4F4E, 0x4888, 0x49F0, 0x4820, 0x4844, 0x7BFE, 0x4822, 0x128, 0x224, 0x4A2, 0x40}},
    {13803, 2, {0x100, 0x110, 0x7BDC, 0x4A54, 0x4D54, 0x4A88, 0x4908, 0x4AF4, 0x4C02, 0x4800, 0x7BFC, 0x4840, 0x150, 0x248, 0x544, 0x80}},
    {13804, 2, {0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x200, 0xFE0, 0x34C0, 0x700, 0x3880, 0x3F8, 0x1D30, 0x1C0, 0x3E00}},
    {13805, 2, {0x1C, 0x1E0, 0x7820, 0x4924, 0x48A8, 0x4BFE, 0x4870, 0x48A8, 0x4924, 0x4A22, 0x7840, 0x4824, 0x2A2, 0x28A, 0x488, 0x78}},
    {13806, 2, {0x2020, 0x2010, 0x3DFE, 0x4510, 0x457C, 0xA914, 0x1914, 0x117E, 0x2114, 0x7D14, 0xA57C, 0x2510, 0x2528, 0x3E28, 0x244, 0x482}},
    {13807, 2, {0x108, 0x1FD0, 0x120, 0x7FFC, 0x200, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {13808, 2, {0x124, 0x124, 0x7224, 0x5424, 0x5954, 0x514A, 0x5292, 0x5610, 0x5A10, 0x5250, 0x725C, 0x5250, 0x250, 0x2B0, 0x29E, 0x300}},
    {13809, 2, {0x140, 0x120, 0x7BFE, 0x4E20, 0x4BFC, 0x4A20, 0x4BFC, 0x4A20, 0x4BFE, 0x4A40, 0x7840, 0x4FFE, 0x150, 0x248, 0xC46, 0x40}},
    {13810, 2, {0x40, 0x3F8, 0x7848, 0x4FFE, 0x4848, 0x4BF8, 0x4840, 0x4FFE, 0x4800, 0x4BF8, 0x7A48, 0x4BF8, 0x248, 0x3F8, 0x0, 0x7FE}},
    {13811, 2, {0x210, 0x210, 0x7BDE, 0x4A28, 0x4D44, 0x48A0, 0x4910, 0x4A08, 0x4DF6, 0x4800, 0x7800, 0x4BF8, 0x208, 0x208, 0x3F8, 0x208}},
    {13812, 2, {0x79E, 0x492, 0xF79E, 0x9492, 0x979E, 0x9402, 0x94F2, 0x9492, 0x94F2, 0x9492, 0xF4F2, 0x9492, 0x492, 0x532, 0x402, 0x406}},
    {13813, 2, {0x210, 0x290, 0xEA90, 0xAB3E, 0xB222, 0xA544, 0xA890, 0xB210, 0xA290, 0xAB10, 0xEA28, 0xB228, 0x528, 0x8C4, 0x1044, 0x82}},
    {13814, 2, {0x24, 0x3A8, 0x7892, 0x4D14, 0x4A08, 0x4C04, 0x4BBA, 0x48A8, 0x48C6, 0x4B80, 0x7A3C, 0x4B84, 0xA8, 0x90, 0x2A8, 0x144}},
    {13815, 2, {0x20, 0x3FE, 0x7A52, 0x49FC, 0x4850, 0x49FC, 0x4850, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x4924, 0x124, 0x58, 0x84, 0x302}},
    {13816, 2, {0x280, 0x48E, 0xE8E0, 0xB140, 0xA040, 0xA24E, 0xA5F4, 0xAC44, 0xB444, 0xA554, 0xE554, 0xA574, 0x594, 0x404, 0x414, 0x408}},
    {13817, 2, {0x140, 0x120, 0x73FE, 0x5620, 0x5BFC, 0x5220, 0x53FC, 0x5220, 0x5220, 0x53FE, 0x7240, 0x5024, 0x522, 0x50A, 0x8F8, 0x0}},
    {13818, 2, {0x0, 0x3F8, 0x7080, 0x57FC, 0x5110, 0x5208, 0x5CA6, 0x50A0, 0x57BC, 0x50A0, 0x73B8, 0x50A0, 0x7BC, 0xA0, 0xA0, 0xA0}},
    {13819, 2, {0x0, 0x4F8, 0x7288, 0x52E8, 0x50A8, 0x50A8, 0x5DFC, 0x5504, 0x5574, 0x5554, 0x7574, 0x5504, 0x514, 0x508, 0xA00, 0x11FE}},
    {13820, 2, {0xA0, 0x4A4, 0xF2A8, 0x90A0, 0x97FC, 0x9110, 0x90A0, 0x97FC, 0x9040, 0x93F8, 0xF040, 0x97FC, 0x150, 0x248, 0xC46, 0x40}},
    {13821, 2, {0x1100, 0x2BDE, 0x4A52, 0x73D4, 0x2A52, 0x4BD2, 0x7A1A, 0x1294, 0x2350, 0xC240, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13822, 2, {0x510, 0x518, 0xEFA8, 0xA528, 0xA76A, 0xA2AA, 0xAFAA, 0xAAAC, 0xAAAC, 0xAFA8, 0xE238, 0xAFAA, 0x22A, 0x22A, 0x226, 0x220}},
    {13823, 2, {0x0, 0x79E, 0x7492, 0x579E, 0x5492, 0x579E, 0x5402, 0x55FA, 0x5412, 0x55D2, 0x7552, 0x55D2, 0x412, 0x452, 0x42A, 0x404}},
    {13824, 2, {0x80, 0x338, 0x7A08, 0x4BB8, 0x4A08, 0x4A08, 0x4BF8, 0x4840, 0x4A48, 0x4950, 0x4FFE, 0x78E0, 0x4950, 0x248, 0xC46, 0x40}},
    {13825, 2, {0x0, 0x7FC, 0xF4A4, 0x97FC, 0x9040, 0x93F8, 0x9040, 0x97FC, 0x9110, 0x90A0, 0xF3F8, 0x9040, 0xFFE, 0x40, 0x40, 0x40}},
    {13826, 2, {0x108, 0x90, 0x7BFE, 0x4908, 0x4908, 0x4A52, 0x4B9C, 0x4908, 0x4A52, 0x4BDE, 0x7840, 0x48A4, 0x2AA, 0x28A, 0x478, 0x0}},
    {13827, 2, {0x88, 0x3FE, 0x7888, 0x4820, 0x4BFE, 0x4840, 0x48F8, 0x4988, 0x4A88, 0x48F8, 0x7840, 0x4824, 0x2A2, 0x28A, 0x488, 0x78}},
    {13828, 2, {0x200, 0x2FE, 0xE2AA, 0xA2AA, 0xAFAA, 0xA2FE, 0xA620, 0xA710, 0xAAFE, 0xAA20, 0xF220, 0xA23C, 0x224, 0x244, 0x254, 0x288}},
    {13829, 2, {0x240, 0x240, 0xE4AE, 0xA890, 0xB300, 0xA2E0, 0xA45E, 0xAC44, 0xB5E4, 0xA444, 0xE554, 0xA4E4, 0x444, 0x464, 0x584, 0x40C}},
    {13830, 2, {0x40, 0xFFE, 0xE040, 0xA7FC, 0xA000, 0xA3F8, 0xA208, 0xAFFE, 0xA802, 0xA7FC, 0xE110, 0xA3F8, 0x40, 0x7FC, 0x40, 0xFFE}},
    {13831, 2, {0x110, 0x554, 0xF5B8, 0x9910, 0x92A8, 0x9444, 0x9000, 0x97FC, 0x9444, 0x9040, 0xF7FC, 0x90E0, 0x150, 0x248, 0xC46, 0x40}},
    {13832, 2, {0x400, 0x57C, 0xF604, 0x94A8, 0x9390, 0x907C, 0x9414, 0x9794, 0x9A50, 0x9250, 0x9FDC, 0xF250, 0x9550, 0x4B0, 0x89E, 0x100}},
    {13833, 2, {0x210, 0x210, 0xF3DE, 0x9528, 0x9084, 0x9000, 0x97BC, 0x94A4, 0x97A4, 0x94A4, 0xF7A4, 0x9434, 0x528, 0x6A0, 0x420, 0x20}},
    {13834, 2, {0x7FC, 0x40, 0xEFFE, 0xA842, 0xA358, 0xA040, 0xA358, 0xA080, 0xA040, 0xAFFE, 0xE000, 0xA7FC, 0x0, 0x7FC, 0x404, 0x7FC}},
    {13835, 2, {0x7DE, 0x512, 0x77D2, 0x5454, 0x57C8, 0x5514, 0x57E2, 0x5000, 0x53FC, 0x5204, 0x73FC, 0x5204, 0x3FC, 0x204, 0x3FC, 0x108}},
    {13836, 2, {0x202, 0x3C2, 0xE202, 0xAFEA, 0xAA2A, 0xAB8A, 0xAE4A, 0xA9CA, 0xA80A, 0xAFEA, 0xEA0A, 0xAD4A, 0xA82, 0xDC2, 0x12AA, 0x2584}},
    {13837, 2, {0x108, 0x14C, 0xE52A, 0xA548, 0xA88E, 0xA138, 0xA288, 0xA7EA, 0xBC8A, 0xA7EC, 0xE48C, 0xA7E8, 0x48A, 0x7FA, 0x426, 0x442}},
    {13838, 2, {0x400, 0x2FE, 0xE202, 0xA802, 0xAB92, 0xA892, 0xAFDE, 0xAAAA, 0xABCA, 0xAAAA, 0xEB92, 0xAAD2, 0xFAA, 0x8C6, 0x882, 0x806}},
    {13839, 2, {0x2110, 0x17FE, 0x1080, 0x3F8, 0xF208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE, 0x90, 0x3EE4, 0x2284, 0x3E7C}},
    {13840, 2, {0x10, 0x7C8, 0x753E, 0x57C0, 0x5514, 0x57D2, 0x5522, 0x5500, 0x57D4, 0x5054, 0x7554, 0x5548, 0x548, 0x454, 0x54, 0x1A2}},
    {13841, 2, {0x410, 0x410, 0x1F7C, 0xE414, 0xA414, 0xAEFE, 0xA014, 0xAE14, 0xA07C, 0xBF10, 0xE454, 0x438, 0x1654, 0x2512, 0x1450, 0x820}},
    {13842, 2, {0xC38, 0x71E0, 0x1020, 0x13FE, 0xFD24, 0x31FC, 0x3924, 0x55FC, 0x5020, 0x93FE, 0x1000, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {13843, 2, {0x7BC, 0x4A4, 0xF7BC, 0x94A4, 0x97BC, 0x9444, 0x95F4, 0x9444, 0x95F4, 0x9554, 0xF5F4, 0x9554, 0x5F4, 0x4E4, 0x554, 0x44C}},
    {13844, 2, {0x40, 0x7FE, 0x7492, 0x53FC, 0x5090, 0x53FC, 0x5090, 0x57FE, 0x5108, 0x53FC, 0x750A, 0x51F8, 0x108, 0x1F8, 0x90, 0x108}},
    {13845, 2, {0x90, 0x7FE, 0x7890, 0x4BE8, 0x4A88, 0x4BCE, 0x4A50, 0x4BD4, 0x4AA2, 0x4BE2, 0x7800, 0x4BFC, 0x294, 0x294, 0x7FE, 0x0}},
    {13846, 2, {0x200, 0x11E, 0xEFC4, 0xA448, 0xA29E, 0xAFD2, 0xA912, 0xAA1E, 0xAC92, 0xA91E, 0xEA12, 0xAC52, 0x89E, 0x900, 0x120C, 0x412}},
    {13847, 2, {0x7BC, 0x4A4, 0xF7BC, 0x94A4, 0x97BC, 0x94A4, 0x97BC, 0x9120, 0x93FE, 0x9220, 0xF7FC, 0x9A20, 0x3FC, 0x220, 0x3FE, 0x200}},
    {13848, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13849, 2, {0x7BC, 0x4A4, 0x77BC, 0x5000, 0x57FE, 0x5400, 0x5590, 0x549E, 0x57D4, 0x5564, 0x75D4, 0x5554, 0x5D4, 0x568, 0xBC8, 0x1054}},
    {13850, 2, {0xFBE, 0x8A2, 0xEFBE, 0xA8A2, 0xAFBE, 0xA802, 0xAB92, 0xA892, 0xAFDE, 0xAAAA, 0xEBCA, 0xAAAA, 0xB92, 0xAD2, 0xFAA, 0x8C4}},
    {13851, 2, {0x40, 0xA0, 0x7318, 0x5DF6, 0x5000, 0x53F8, 0x5208, 0x53F8, 0x5110, 0x57DC, 0x7284, 0x4FFC, 0x450, 0x7DC, 0x452, 0x4CE}},
    {13852, 2, {0x5118, 0x4B94, 0x7C3E, 0xD3A8, 0x7C7E, 0x53A8, 0x7C3E, 0x53A8, 0x7EBE, 0x43A0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {13853, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4FC4, 0x4044, 0x4844, 0x4484, 0x4284, 0x4104, 0x4284, 0x4444, 0x4844, 0x4004, 0x7FFC, 0x4004}},
    {13854, 2, {0x0, 0x7FFC, 0x4204, 0x4204, 0x4204, 0x5FE4, 0x4224, 0x4224, 0x4224, 0x4424, 0x4424, 0x4824, 0x4944, 0x5084, 0x7FFC, 0x4004}},
    {13855, 2, {0x0, 0x7FFC, 0x4004, 0x4444, 0x4444, 0x4444, 0x5FF4, 0x4444, 0x4444, 0x4444, 0x4844, 0x4844, 0x5044, 0x4004, 0x7FFC, 0x4004}},
    {13856, 2, {0x0, 0x7FFC, 0x4004, 0x4204, 0x4144, 0x4144, 0x4484, 0x54A4, 0x5514, 0x6654, 0x4444, 0x4BC4, 0x5004, 0x4004, 0x7FFC, 0x4004}},
    {13857, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x5FF4, 0x5114, 0x5114, 0x5FF4, 0x5114, 0x5114, 0x5FF4, 0x5114, 0x4104, 0x4104, 0x7FFC, 0x4004}},
    {13858, 2, {0x0, 0x7FFC, 0x4104, 0x4924, 0x4524, 0x4544, 0x5FF4, 0x4284, 0x4284, 0x4284, 0x4494, 0x4894, 0x5074, 0x4004, 0x7FFC, 0x4004}},
    {13859, 2, {0x7FFC, 0x4004, 0x47C4, 0x4444, 0x47C4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x48A4, 0x4844, 0x7FFC, 0x4004}},
    {13860, 2, {0x0, 0x7FFC, 0x4144, 0x6554, 0x5554, 0x4774, 0x6414, 0x57F4, 0x4414, 0x5774, 0x6554, 0x4554, 0x4954, 0x5014, 0x7FFC, 0x4004}},
    {13861, 2, {0xFFFE, 0x8002, 0x861A, 0xF8E2, 0xAA42, 0x9492, 0x90E2, 0xBE4A, 0xC8FA, 0xBE22, 0x88AA, 0xAB26, 0xBE62, 0x8002, 0xFFFE, 0x8002}},
    {13862, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x11FC, 0xFE44, 0x1044, 0x1044, 0x1044, 0x1084, 0x1684, 0x1884, 0xE104, 0x4104, 0x228, 0x410}},
    {13863, 2, {0x1000, 0x11F8, 0x1008, 0x1050, 0x1020, 0xFC10, 0x13FE, 0x1022, 0x1024, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {13864, 2, {0x2020, 0x2020, 0x2020, 0x2020, 0x23FE, 0xFA22, 0x2222, 0x2222, 0x2252, 0x224A, 0x228A, 0x3B02, 0xE202, 0x4202, 0x20A, 0x204}},
    {13865, 2, {0x1010, 0x1110, 0x1110, 0x1112, 0x1112, 0xFD14, 0x11D8, 0x1110, 0x1110, 0x1110, 0x1110, 0x1D12, 0xE152, 0x4192, 0x10E, 0x0}},
    {13866, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x1020, 0xFC3C, 0x11E0, 0x1020, 0x1020, 0x103E, 0x13E0, 0x1C20, 0xE022, 0x4022, 0x22, 0x1E}},
    {13867, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0x11FC, 0xFC88, 0x1088, 0x1088, 0x13FE, 0x1088, 0x1088, 0x1C88, 0xE088, 0x4108, 0x108, 0x208}},
    {13868, 2, {0x2100, 0x2100, 0x2100, 0x21FC, 0xFA04, 0x2484, 0x2084, 0x2104, 0x2144, 0x2224, 0x27F4, 0x3A14, 0xE004, 0x4004, 0x28, 0x10}},
    {13869, 2, {0x2080, 0x2080, 0x213E, 0x2102, 0x2242, 0xFA42, 0x27C2, 0x2082, 0x2082, 0x2102, 0x2142, 0x3A22, 0xE7E2, 0x4222, 0xA, 0x4}},
    {13870, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1124, 0x1120, 0x1120, 0x1110, 0x1D10, 0xE208, 0x4208, 0x404, 0x802}},
    {13871, 2, {0x1080, 0x1080, 0x1100, 0x11FE, 0x1200, 0xFC40, 0x1040, 0x13FC, 0x1044, 0x1044, 0x1084, 0x1C84, 0xE104, 0x4104, 0x228, 0x410}},
    {13872, 2, {0x1010, 0x1110, 0x1090, 0x1090, 0x1010, 0xFD10, 0x1090, 0x1090, 0x1010, 0x101E, 0x13F0, 0x1C10, 0xE010, 0x4010, 0x10, 0x10}},
    {13873, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1124, 0x1124, 0x13FE, 0x1D04, 0xE104, 0x4104, 0x114, 0x108}},
    {13874, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0x23FC, 0xFA40, 0x2240, 0x2240, 0x23F8, 0x2248, 0x2248, 0x3A48, 0xE488, 0x448A, 0x90A, 0x206}},
    {13875, 2, {0x440, 0x444, 0x7C48, 0x470, 0x440, 0xC42, 0x3442, 0xC43E, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {13876, 2, {0x100, 0x110, 0x7910, 0x4910, 0x4910, 0x4928, 0x4924, 0x7944, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {13877, 2, {0x1040, 0x1040, 0x1040, 0x17FE, 0x1080, 0xFCA0, 0x1120, 0x11FC, 0x1324, 0x1524, 0x1124, 0x1D24, 0xE134, 0x4128, 0x20, 0x20}},
    {13878, 2, {0x810, 0x810, 0x1EFC, 0x2210, 0x5290, 0xC90, 0x8FE, 0x1010, 0x2010, 0x4110, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13879, 2, {0x1110, 0x1110, 0x13FC, 0x1110, 0x1110, 0xFC00, 0x1208, 0x1208, 0x1110, 0x1110, 0x10A0, 0x1C40, 0xE0A0, 0x4110, 0x208, 0xC06}},
    {13880, 2, {0x0, 0xFFFE, 0x1040, 0x1044, 0x3E48, 0x4270, 0xA440, 0x1842, 0x1042, 0x213E, 0xC100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13881, 2, {0x880, 0x888, 0x2890, 0x2EA0, 0x28C0, 0x2884, 0x2E84, 0xF07C, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {13882, 2, {0x2120, 0x2128, 0x2124, 0x2224, 0xFA20, 0x263E, 0x2BE0, 0x2224, 0x2224, 0x2228, 0x2228, 0x3A10, 0xE212, 0x422A, 0x246, 0x282}},
    {13883, 2, {0x2000, 0x23FE, 0x2202, 0x2222, 0x2222, 0xFA22, 0x23FE, 0x2222, 0x2222, 0x2252, 0x224A, 0x3A8A, 0xE302, 0x4202, 0x3FE, 0x202}},
    {13884, 2, {0x1040, 0x1044, 0x11F4, 0x1048, 0x1050, 0xFDFE, 0x1040, 0x1080, 0x11FE, 0x1240, 0x1480, 0x1CFC, 0xE004, 0x4004, 0x28, 0x10}},
    {13885, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x40, 0x844, 0x844, 0x7F44, 0x828, 0x82A, 0xF12, 0xF02A, 0x4046, 0x82}},
    {13886, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0x2404, 0xF800, 0x21FC, 0x2020, 0x2020, 0x2020, 0x23FE, 0x3820, 0xE020, 0x4020, 0xA0, 0x40}},
    {13887, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1020, 0x1020, 0x11FC, 0x1C20, 0xE020, 0x4020, 0x3FE, 0x0}},
    {13888, 2, {0x2080, 0x2040, 0x27FC, 0x2404, 0x2110, 0xFA08, 0x2484, 0x2080, 0x23F0, 0x2090, 0x2090, 0x3910, 0xE112, 0x4212, 0x40E, 0x800}},
    {13889, 2, {0x1080, 0x1080, 0x10FE, 0x1100, 0x12FC, 0xFC84, 0x10A4, 0x1094, 0x13FE, 0x1084, 0x1124, 0x1D14, 0xE1FE, 0x4004, 0x28, 0x10}},
    {13890, 2, {0x1104, 0x1084, 0x1088, 0x1010, 0x11FC, 0xFD04, 0x1104, 0x1104, 0x11FC, 0x1050, 0x1050, 0x1C90, 0xE092, 0x4112, 0x20E, 0x400}},
    {13891, 2, {0x1010, 0x1038, 0x11C0, 0x1100, 0x1100, 0xFDFC, 0x1110, 0x1110, 0x1110, 0x1110, 0x13FE, 0x1C00, 0xE090, 0x4108, 0x204, 0x402}},
    {13892, 2, {0x1110, 0x1112, 0x1114, 0x11D8, 0x1110, 0xFD12, 0x1152, 0x118E, 0x1120, 0x1020, 0x11FC, 0x1C20, 0xE020, 0x4020, 0x3FE, 0x0}},
    {13893, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0x2208, 0xFBF8, 0x2040, 0x2040, 0x27FC, 0x20E0, 0x2150, 0x3A48, 0xE444, 0x4842, 0x40, 0x40}},
    {13894, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0x2444, 0xF840, 0x23FE, 0x2040, 0x2090, 0x2090, 0x2110, 0x3920, 0xE224, 0x4242, 0x4FE, 0x842}},
    {13895, 2, {0x1000, 0x13FE, 0x1040, 0x1080, 0x1144, 0xFE24, 0x1068, 0x12B0, 0x1130, 0x12A8, 0x1068, 0x1CA4, 0xE122, 0x4220, 0xA0, 0x40}},
    {13896, 2, {0x2040, 0x2040, 0x207C, 0x2040, 0x23FC, 0xFA44, 0x2270, 0x23C0, 0x2244, 0x223C, 0x2200, 0x3AF0, 0xE490, 0x4492, 0x912, 0x120E}},
    {13897, 2, {0x1050, 0x1050, 0x1050, 0x11FC, 0x1154, 0xFD54, 0x1154, 0x11FC, 0x1154, 0x1154, 0x1D54, 0xE3FE, 0x4000, 0x88, 0x104, 0x202}},
    {13898, 2, {0x2000, 0x21FC, 0x2100, 0x21F8, 0x2100, 0xF9F8, 0x2100, 0x2FFE, 0x2280, 0x2248, 0x2250, 0x3A20, 0xE210, 0x4288, 0x306, 0x200}},
    {13899, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x1088, 0xFCF8, 0x1088, 0x1088, 0x10F8, 0x1088, 0x1C88, 0xE3FE, 0x4000, 0x88, 0x104, 0x202}},
    {13900, 2, {0x1020, 0x1020, 0x11FC, 0x1024, 0x13FE, 0xFC24, 0x11FC, 0x1020, 0x1124, 0x1124, 0x1174, 0x1DAC, 0xE124, 0x4124, 0x224, 0x424}},
    {13901, 2, {0x2000, 0x27BC, 0x2084, 0x2294, 0x2108, 0xFA94, 0x24A4, 0x2840, 0x2000, 0x27BC, 0x20A4, 0x3AA4, 0xE128, 0x4290, 0x4A8, 0x846}},
    {13902, 2, {0x2014, 0x2012, 0x27FE, 0x2410, 0xF410, 0x27F0, 0x2492, 0x2492, 0x27D2, 0x2554, 0x3554, 0xE488, 0x454A, 0xA1A, 0x826, 0x1042}},
    {13903, 2, {0x1100, 0x1100, 0x11FE, 0x1200, 0x15FC, 0xFD04, 0x11FC, 0x1104, 0x11FC, 0x1080, 0x10FC, 0x1D08, 0xE290, 0x4060, 0x198, 0x606}},
    {13904, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0xFC00, 0x13FE, 0x1020, 0x1040, 0x11FC, 0x1154, 0x1D54, 0xE154, 0x4154, 0x154, 0x10C}},
    {13905, 2, {0xF0, 0x3F00, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x4124, 0x41F4, 0x5E14, 0x4008, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {13906, 2, {0xF8, 0x7E08, 0x4A50, 0x4A20, 0x7EFC, 0x4A24, 0x4A28, 0x7E20, 0xA0, 0x140, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13907, 2, {0x100, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x820, 0x3EF8, 0x820, 0xE20, 0x78FC, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13908, 2, {0x1040, 0x1040, 0x11FC, 0x1084, 0x13FE, 0xFC00, 0x11FC, 0x1104, 0x11FC, 0x1020, 0x13FE, 0x1C20, 0xE220, 0x43FE, 0x20, 0x20}},
    {13909, 2, {0x2000, 0x23FE, 0x2222, 0x23FE, 0x2222, 0xFBFE, 0x2000, 0x21FC, 0x2104, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x4104, 0x114, 0x108}},
    {13910, 2, {0x1004, 0x101E, 0x13E0, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1222, 0x1E2A, 0xE2FA, 0x4202, 0x20A, 0x204}},
    {13911, 2, {0x2000, 0x23F8, 0x2248, 0x2248, 0x23F8, 0xFA48, 0x2248, 0x23F8, 0x2000, 0x27FE, 0x2240, 0x3A44, 0xE228, 0x4290, 0x308, 0x206}},
    {13912, 2, {0x2000, 0x23FE, 0x2050, 0x2050, 0x23FE, 0xFA52, 0x2252, 0x23FE, 0x2000, 0x2040, 0x27FE, 0x3888, 0xE190, 0x4060, 0xD8, 0x304}},
    {13913, 2, {0x2120, 0x1110, 0x97FE, 0x4128, 0x4934, 0x928, 0x1132, 0xE124, 0x223A, 0x2262, 0x259E, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {13914, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0x11FC, 0xFC20, 0x13FE, 0x1000, 0x10A0, 0x10FC, 0x1120, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {13915, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x1C08, 0xE3FE, 0x4088, 0x48, 0x48, 0x18}},
    {13916, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FEC, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {13917, 2, {0x2040, 0x21FC, 0x2044, 0x2094, 0x2108, 0xFBDE, 0x214A, 0x214A, 0x2294, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x41FC, 0x104, 0x10C}},
    {13918, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1104, 0x11FC, 0x1020, 0x13FE, 0x1C70, 0xE0A8, 0x4124, 0x222, 0x20}},
    {13919, 2, {0x2100, 0x209E, 0x23D2, 0x2252, 0x2252, 0xFBDE, 0x2252, 0x2252, 0x23D2, 0x221E, 0x2292, 0x3A52, 0xE2D2, 0x4352, 0x22A, 0x44}},
    {13920, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13921, 2, {0x2124, 0x2124, 0x2224, 0x2424, 0x2954, 0xF94A, 0x2292, 0x2610, 0x2A10, 0x2250, 0x225C, 0x3A50, 0xE250, 0x42B0, 0x29E, 0x300}},
    {13922, 2, {0x2088, 0x2088, 0x23FE, 0x2088, 0x2088, 0xF8F8, 0x2020, 0x23FE, 0x2222, 0x2332, 0x22AA, 0x3B76, 0xE222, 0x4222, 0x22A, 0x224}},
    {13923, 2, {0x2000, 0x2F3E, 0x2120, 0x213C, 0xF120, 0x2F3C, 0x2820, 0x28FE, 0x2850, 0x2F52, 0x3154, 0xE148, 0x4148, 0x144, 0xA62, 0x440}},
    {13924, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0x2242, 0xFA78, 0x23C4, 0x223C, 0x2208, 0x22F0, 0x2324, 0x3AA8, 0xE5FE, 0x4420, 0x8A0, 0x40}},
    {13925, 2, {0x2080, 0x2040, 0x27FC, 0x2110, 0xFA48, 0x2484, 0x2120, 0x23F0, 0x2010, 0x2088, 0x2150, 0x3B20, 0xE510, 0x4948, 0x186, 0x100}},
    {13926, 2, {0x808, 0x1448, 0x2228, 0x5C88, 0x8848, 0x7F0E, 0x8F8, 0x2A08, 0x4908, 0x1808, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13927, 2, {0x2000, 0x23FE, 0x2050, 0x23DE, 0x2252, 0xFA52, 0x23DE, 0x2050, 0x23FE, 0x2222, 0x2222, 0x3BFE, 0xE222, 0x4222, 0x3FE, 0x202}},
    {13928, 2, {0x2248, 0x2248, 0x2554, 0x275C, 0x2248, 0xFD54, 0x275C, 0x2248, 0x2FFE, 0x2220, 0x2228, 0x3A28, 0xE514, 0x4494, 0x42C, 0x844}},
    {13929, 2, {0x2040, 0x2040, 0x27FC, 0x20A0, 0x2514, 0xFA08, 0x27FC, 0x2A0A, 0x23F8, 0x2208, 0x23F8, 0x3840, 0xE248, 0x4444, 0x944, 0x80}},
    {13930, 2, {0x2020, 0x23FE, 0x2020, 0x21FC, 0x2000, 0xFBFE, 0x2202, 0x21FC, 0x2000, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x4088, 0x3FE, 0x0}},
    {13931, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0x11FC, 0xFD20, 0x11FC, 0x1120, 0x11FE, 0x1100, 0x1C20, 0xE3FE, 0x4070, 0xA8, 0x326, 0x20}},
    {13932, 2, {0x814, 0x812, 0x7F10, 0x2AFE, 0x2A10, 0x5D10, 0x1C28, 0x2A28, 0x4944, 0x8882, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {13933, 2, {0x2000, 0x2FFE, 0x2800, 0x2A28, 0xF948, 0x2BEE, 0x2892, 0x2884, 0x2AA0, 0x2AA8, 0x3BE8, 0xE888, 0x4894, 0x914, 0x924, 0x1242}},
    {13934, 2, {0x2210, 0x2210, 0x23DE, 0x2228, 0x2544, 0xF8A0, 0x2110, 0x2208, 0x25F6, 0x2000, 0x2000, 0x3BF8, 0xE208, 0x4208, 0x3F8, 0x208}},
    {13935, 2, {0x820, 0x3E20, 0x8F8, 0x7F28, 0x2228, 0x496A, 0x3E2A, 0x856, 0x7E82, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x820, 0x1010, 0x3FF8}},
    {13936, 2, {0x1000, 0x1EFC, 0x2224, 0x7F44, 0xA99C, 0x3F50, 0x297C, 0x3F90, 0x29FE, 0x2910, 0x4510, 0x8210, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {13937, 2, {0x20, 0x7C3C, 0x4420, 0x7DFE, 0x122, 0xFD38, 0x41E4, 0x7D1C, 0x570, 0x2A54, 0x148C, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {13938, 2, {0x2040, 0x2020, 0x27FE, 0x2028, 0x2224, 0xF27E, 0x24C8, 0x2748, 0x217E, 0x2248, 0x24C8, 0x377E, 0xE148, 0x4248, 0x47E, 0x840}},
    {13939, 2, {0x2014, 0x2012, 0x27FE, 0x2410, 0xF5D0, 0x2414, 0x25D4, 0x2558, 0x25CA, 0x2416, 0x3442, 0xE824, 0x42A2, 0x28A, 0x478, 0x0}},
    {13940, 2, {0x2100, 0x21F8, 0x2208, 0x27FE, 0x2288, 0xFB24, 0x23FE, 0x2200, 0x22FC, 0x2200, 0x22FC, 0x3A00, 0xE2FC, 0x4484, 0x4FC, 0x884}},
    {13941, 2, {0x1020, 0x13FE, 0x1000, 0x10F8, 0x1088, 0xFCF8, 0x1000, 0x13FE, 0x1202, 0x12FA, 0x128A, 0x1EFA, 0xE226, 0x41F8, 0x20, 0x3FE}},
    {13942, 2, {0x100, 0xFFFE, 0x1010, 0x1FF0, 0x0, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {13943, 2, {0x2000, 0x23FE, 0x2202, 0x23FE, 0x2202, 0xFBFE, 0x2108, 0x2252, 0x239C, 0x2108, 0x2252, 0x3BDE, 0xE000, 0x42A4, 0x252, 0x452}},
    {13944, 2, {0x2128, 0x21AA, 0x216C, 0x2228, 0xFAFE, 0x2644, 0x2628, 0x2AFE, 0x2210, 0x227C, 0x2210, 0x3AFE, 0xE210, 0x4228, 0x244, 0x282}},
    {13945, 2, {0x2088, 0x2088, 0x23FE, 0x2088, 0x23FE, 0xFA52, 0x23FE, 0x2000, 0x23FE, 0x2242, 0x20F8, 0x3908, 0xE290, 0x4060, 0xC0, 0x700}},
    {13946, 2, {0x27FE, 0x2420, 0x25FC, 0x2524, 0x25FC, 0xFC20, 0x27FE, 0x2504, 0x25FC, 0x2504, 0x25FC, 0x3D04, 0xE5FC, 0x4488, 0x504, 0x7FE}},
    {13947, 2, {0x2020, 0x27FE, 0x2400, 0x25FC, 0x2524, 0xF5FC, 0x2524, 0x25FC, 0x2420, 0x25FC, 0x2420, 0x37FE, 0xE524, 0x4AFA, 0x820, 0x13FE}},
    {13948, 2, {0x23F8, 0x2248, 0x2248, 0x23F8, 0x2248, 0xFA48, 0x23F8, 0x2000, 0x2FBE, 0x2AAA, 0x2AAA, 0x3FBE, 0xEAAA, 0x4AAA, 0xFBE, 0x8A2}},
    {13949, 2, {0x2020, 0x2020, 0x27FE, 0x2020, 0x23AE, 0xF924, 0x23AE, 0x22AA, 0x23AE, 0x22AA, 0x23AE, 0x3850, 0xE050, 0x4088, 0x104, 0x602}},
    {13950, 2, {0x22A0, 0x244C, 0x24A4, 0x260C, 0xF4A4, 0x264C, 0x24A4, 0x2FFE, 0x2802, 0x23F0, 0x3020, 0xE040, 0x47FC, 0x40, 0x140, 0x80}},
    {13951, 2, {0x2110, 0x27FC, 0x2040, 0x23F8, 0x2040, 0xF7FC, 0x2000, 0x23D4, 0x2112, 0x27FE, 0x2350, 0x3534, 0xE78C, 0x42CA, 0x456, 0x8C2}},
    {13952, 2, {0x2108, 0x21EC, 0x210A, 0x27E8, 0xF528, 0x25DE, 0x2728, 0x24EA, 0x240A, 0x27EA, 0x340C, 0xE7EC, 0x4A2A, 0xBEA, 0x1156, 0x7E2}},
    {13953, 2, {0x1010, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x4444, 0xEEEE, 0x4444, 0x6464, 0xDEDE, 0x0, 0xFEFE, 0x2828, 0x2A28, 0x4C4A, 0x8886}},
    {13954, 2, {0x2040, 0x27FC, 0x2248, 0x23F8, 0xF040, 0x2FFE, 0x2AAA, 0x23B8, 0x2110, 0x27FC, 0x3110, 0xEFFE, 0x4128, 0x310, 0xD48, 0x186}},
    {13955, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8444, 0x2388, 0x2448, 0x2008, 0x3FF8, 0x820, 0x440, 0x7FFC}},
    {13956, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x5112, 0x9FF4, 0x1110, 0x1FF0, 0x0, 0x7FFC, 0x104, 0x11F0, 0x2900, 0x47FE}},
    {13957, 2, {0x200, 0x7E0, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x100, 0x1100, 0x1FF8, 0x2100, 0x5FF0, 0x100, 0xFFFE, 0x0}},
    {13958, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0xFF0, 0x1420, 0x2240, 0x4180, 0xE60, 0x701C}},
    {13959, 2, {0x440, 0x2288, 0x2108, 0x2288, 0x2448, 0x3FF8, 0x1210, 0x2208, 0x47E4, 0x820, 0x1440, 0x2280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {13960, 2, {0x1010, 0x1010, 0x1010, 0x1E10, 0x1210, 0x2210, 0x22FE, 0x5210, 0x8A10, 0x410, 0x410, 0x810, 0x810, 0x1010, 0x2010, 0x4010}},
    {13961, 2, {0x0, 0x3FF0, 0x810, 0x820, 0x83C, 0x1204, 0x17E4, 0x3854, 0x4588, 0x680, 0x39F8, 0x610, 0x1920, 0xC0, 0x300, 0x3C00}},
    {13962, 2, {0x2020, 0x2010, 0x3C10, 0x45FE, 0x4820, 0xA824, 0x1044, 0x28F8, 0x4E12, 0x1222, 0x2444, 0x5588, 0x810, 0x1028, 0x2044, 0xC182}},
    {13963, 2, {0x2000, 0x21FC, 0x3D24, 0x4524, 0x4974, 0xA924, 0x11FC, 0x2904, 0x4F74, 0x1354, 0x2554, 0x5574, 0x904, 0x1204, 0x2214, 0xC408}},
    {13964, 2, {0x2000, 0x21FC, 0x3D24, 0x4524, 0x49FC, 0xA924, 0x1124, 0x29FC, 0x4E20, 0x13FE, 0x2470, 0x54A8, 0x924, 0x1222, 0x2020, 0xC020}},
    {13965, 2, {0x2020, 0x21FC, 0x3D24, 0x45FC, 0x4820, 0xABFE, 0x1000, 0x29FC, 0x4F04, 0x13FC, 0x2504, 0x55FC, 0x904, 0x11FC, 0x2088, 0xC104}},
    {13966, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x0, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x8}},
    {13967, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x820, 0x820, 0x7FFC, 0x820, 0x820, 0x1020, 0x1020, 0x2020}},
    {13968, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x2C0, 0x420, 0x1810, 0x6008, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {13969, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {13970, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x2120, 0x1110, 0x1110, 0x2928, 0x4544, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {13971, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3110, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100, 0x100}},
    {13972, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC826, 0x440, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {13973, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3458, 0xC446, 0x7FF8, 0x448, 0x3FF8, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2048, 0x4040}},
    {13974, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {13975, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {13976, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x7DFC, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x1050, 0x1050, 0x2888, 0x2488, 0x4504, 0x8202}},
    {13977, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x227C, 0x4A44, 0x8500, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {13978, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3218, 0xC7E6, 0x820, 0x3440, 0x280, 0x300, 0x1FF0, 0xE810, 0x810, 0xFF0, 0x810}},
    {13979, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x38D8, 0xCF06, 0x810, 0x7F0, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {13980, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3098, 0xC0A6, 0x3FF8, 0x80, 0x3E80, 0x2248, 0x3E50, 0x24, 0xE54, 0x718C, 0x604}},
    {13981, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC806, 0x840, 0x13F8, 0x3040, 0x57FC, 0x9040, 0x13F8, 0x1040, 0x17FC, 0x1000}},
    {13982, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC826, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2A68, 0x48A4, 0x8922, 0x820, 0x820}},
    {13983, 2, {0x800, 0x1FC0, 0x2080, 0x7FF0, 0xA010, 0x3FF0, 0x2088, 0x3CF4, 0x2084, 0x3D7C, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {13984, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3458, 0xC286, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x2488, 0x2244, 0x4244}},
    {13985, 2, {0x1010, 0x1010, 0xFEFE, 0x1010, 0x2828, 0x4444, 0xFEFE, 0x404, 0x7474, 0x5454, 0x5454, 0x7474, 0x5454, 0x404, 0x1414, 0x808}},
    {13986, 2, {0x1080, 0x1080, 0x1080, 0x2080, 0x2FFE, 0x6108, 0x6108, 0xA208, 0x2210, 0x2610, 0x21A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C04}},
    {13987, 2, {0x200, 0x200, 0xFFFE, 0x840, 0x1C80, 0x380, 0x470, 0x1808, 0x6200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {13988, 2, {0x1048, 0x1044, 0x1044, 0x1040, 0xFE4E, 0x23F0, 0x2240, 0x2240, 0x4240, 0x2440, 0x1420, 0x822, 0x1412, 0x220A, 0x4206, 0x8002}},
    {13989, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFC20, 0x2420, 0x2420, 0x27FE, 0x2420, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {13990, 2, {0x1000, 0x11FE, 0x1048, 0x1048, 0xFC48, 0x2448, 0x2448, 0x2448, 0x2448, 0x4848, 0x2848, 0x1048, 0x2888, 0x4488, 0x8508, 0x208}},
    {13991, 2, {0x1020, 0x1010, 0x1010, 0x10FE, 0xFC80, 0x2480, 0x2480, 0x2480, 0x2480, 0x4880, 0x2880, 0x1080, 0x2880, 0x4500, 0x8500, 0x200}},
    {13992, 2, {0x1000, 0x13F0, 0x1090, 0x1090, 0xFC90, 0x2490, 0x2490, 0x27F0, 0x2490, 0x4890, 0x2890, 0x1092, 0x288A, 0x448A, 0x8486, 0x82}},
    {13993, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0xFC84, 0x25FE, 0x2484, 0x2484, 0x2484, 0x49FE, 0x2884, 0x1084, 0x2884, 0x4484, 0x8494, 0x88}},
    {13994, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x243C, 0x25E0, 0x2420, 0x2420, 0x483E, 0x2BE0, 0x1020, 0x2822, 0x4422, 0x8422, 0x1E}},
    {13995, 2, {0x2000, 0x23FC, 0x2084, 0x2088, 0xF888, 0x4890, 0x489C, 0x4884, 0x8944, 0x4944, 0x3128, 0x1128, 0x2A10, 0x4A28, 0x8444, 0x182}},
    {13996, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0xFD04, 0x2604, 0x2484, 0x2444, 0x2444, 0x4814, 0x2824, 0x1044, 0x2984, 0x4404, 0x8428, 0x10}},
    {13997, 2, {0x1040, 0x1050, 0x1048, 0x1048, 0xFC40, 0x25FE, 0x2450, 0x2450, 0x2450, 0x4850, 0x2890, 0x1090, 0x2892, 0x4512, 0x850E, 0x200}},
    {13998, 2, {0x1000, 0x1000, 0x11FE, 0x1102, 0xFD02, 0x2400, 0x2478, 0x2448, 0x2448, 0x4848, 0x2848, 0x104A, 0x284A, 0x448A, 0x8486, 0x100}},
    {13999, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0xFB28, 0x4AA8, 0x4AA8, 0x4A48, 0x8A48, 0x4AA8, 0x32A8, 0x132A, 0x2C0A, 0x440A, 0x8806, 0x1002}},
    {14000, 2, {0x1000, 0x11FE, 0x1010, 0x1010, 0xFC20, 0x2420, 0x2468, 0x24A4, 0x2522, 0x4A22, 0x2820, 0x1020, 0x2820, 0x4400, 0x83FE, 0x0}},
    {14001, 2, {0x1008, 0x101C, 0x10E0, 0x1080, 0xFC80, 0x2480, 0x2480, 0x24FE, 0x2488, 0x4888, 0x2888, 0x1088, 0x2888, 0x4488, 0x85FE, 0x0}},
    {14002, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFC3E, 0x2420, 0x2420, 0x2420, 0x25FC, 0x4904, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14003, 2, {0x2100, 0x2100, 0x2100, 0x211C, 0xFBD4, 0x4954, 0x4954, 0x4954, 0x8954, 0x4954, 0x3154, 0x1154, 0x2954, 0x4A5C, 0x8354, 0x480}},
    {14004, 2, {0x2040, 0x2040, 0x2040, 0x23FE, 0xF880, 0x48A0, 0x4920, 0x49FC, 0x8B24, 0x4D24, 0x3124, 0x1124, 0x2934, 0x4928, 0x8020, 0x20}},
    {14005, 2, {0x1040, 0x1040, 0x107C, 0x1084, 0xFD88, 0x2650, 0x2420, 0x2450, 0x2488, 0x4B06, 0x2860, 0x1010, 0x2808, 0x44C0, 0x8420, 0x10}},
    {14006, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x2508, 0x25F8, 0x2500, 0x2500, 0x49FC, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14007, 2, {0x2020, 0x2020, 0x2050, 0x2050, 0xF888, 0x4904, 0x4A22, 0x4820, 0x48A8, 0x90A4, 0x5124, 0x2122, 0x3222, 0x4820, 0x48A0, 0x8040}},
    {14008, 2, {0x1000, 0x11FC, 0x1004, 0x1004, 0xFDF4, 0x2404, 0x2404, 0x25F4, 0x2514, 0x4914, 0x29F4, 0x1114, 0x2804, 0x4404, 0x8428, 0x10}},
    {14009, 2, {0x2000, 0x2080, 0x231E, 0x2252, 0xFA52, 0x4A52, 0x4A52, 0x4A52, 0x4A52, 0x92D2, 0x535A, 0x2254, 0x3090, 0x4890, 0x4910, 0x8210}},
    {14010, 2, {0x1000, 0x11FC, 0x1044, 0x1048, 0xFC8E, 0x2482, 0x257A, 0x2614, 0x2420, 0x4820, 0x2BFE, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {14011, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x400, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x3008}},
    {14012, 2, {0x1000, 0x1000, 0x11FC, 0x1124, 0xFD24, 0x2524, 0x2524, 0x2524, 0x25FC, 0x4924, 0x2924, 0x1124, 0x2924, 0x4524, 0x85FC, 0x104}},
    {14013, 2, {0x0, 0x7DFE, 0x1040, 0x1080, 0x11FC, 0x1004, 0x1E04, 0xE004, 0x4228, 0x210, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14014, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x2504, 0x25FC, 0x2450, 0x2450, 0x4850, 0x2850, 0x1092, 0x2892, 0x4512, 0x820E, 0x400}},
    {14015, 2, {0x1040, 0x1040, 0x10FC, 0x1084, 0xFD04, 0x26F4, 0x2494, 0x2494, 0x24F4, 0x4884, 0x2894, 0x1088, 0x2882, 0x4482, 0x847E, 0x0}},
    {14016, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFE10, 0x2420, 0x25FC, 0x2404, 0x2404, 0x4804, 0x29FC, 0x1004, 0x2804, 0x4404, 0x85FC, 0x4}},
    {14017, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFD24, 0x24A4, 0x24A8, 0x2420, 0x27FE, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x8420, 0x20}},
    {14018, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x257C, 0x2554, 0x2554, 0x2554, 0x4954, 0x297C, 0x1140, 0x2940, 0x4642, 0x8242, 0x43E}},
    {14019, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x824, 0x7F44, 0x854, 0xA88, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14020, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFC40, 0x2440, 0x24A0, 0x24A2, 0x25A4, 0x4A98, 0x2C90, 0x1088, 0x2888, 0x44A4, 0x84C2, 0x80}},
    {14021, 2, {0x1000, 0x13FE, 0x1088, 0x1088, 0xFCF8, 0x2488, 0x2488, 0x24F8, 0x2488, 0x4888, 0x288E, 0x13F8, 0x2808, 0x4408, 0x8408, 0x8}},
    {14022, 2, {0x1000, 0x11FC, 0x1044, 0x1048, 0xFC5E, 0x2482, 0x248A, 0x2524, 0x2420, 0x4BFE, 0x2820, 0x1070, 0x28A8, 0x4524, 0x8222, 0x20}},
    {14023, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFD00, 0x2510, 0x2510, 0x2510, 0x257C, 0x4910, 0x2910, 0x1110, 0x2910, 0x4610, 0x82FE, 0x400}},
    {14024, 2, {0x1040, 0x1044, 0x11F4, 0x1048, 0xFC50, 0x25FE, 0x2440, 0x2480, 0x25FE, 0x4A40, 0x2C80, 0x10FC, 0x2804, 0x4404, 0x8428, 0x10}},
    {14025, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFDFC, 0x2554, 0x2554, 0x2554, 0x2554, 0x4954, 0x298C, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14026, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFC88, 0x2488, 0x2526, 0x2620, 0x25FE, 0x4820, 0x2870, 0x10A8, 0x2924, 0x4622, 0x8420, 0x20}},
    {14027, 2, {0x2040, 0x2040, 0x20A0, 0x20A0, 0xF910, 0x4A48, 0x4C46, 0x4840, 0x4944, 0x9152, 0x524A, 0x244A, 0x3048, 0x4840, 0x4940, 0x8080}},
    {14028, 2, {0x2040, 0x2040, 0x2090, 0x2108, 0xFBFC, 0x4824, 0x4920, 0x4920, 0x49FC, 0x9220, 0x5020, 0x27FE, 0x3020, 0x4820, 0x4820, 0x8020}},
    {14029, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x2420, 0x2524, 0x2524, 0x2524, 0x4AAA, 0x2820, 0x1050, 0x2850, 0x4488, 0x8104, 0x202}},
    {14030, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0xFC84, 0x2484, 0x24FC, 0x2400, 0x2400, 0x49FE, 0x2902, 0x1102, 0x2902, 0x4502, 0x85FE, 0x102}},
    {14031, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0xFDFE, 0x2420, 0x2520, 0x24B2, 0x24B4, 0x4868, 0x28A8, 0x1124, 0x2A22, 0x4420, 0x84A0, 0x40}},
    {14032, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x2524, 0x2524, 0x25FC, 0x2524, 0x4924, 0x29FC, 0x1120, 0x28A0, 0x4440, 0x80B0, 0x30E}},
    {14033, 2, {0x800, 0xF7C, 0x2844, 0x3F44, 0x6128, 0x9210, 0xC68, 0x3186, 0xC000, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14034, 2, {0x1000, 0x11FE, 0x1000, 0x1000, 0xFCFC, 0x2484, 0x2484, 0x2484, 0x24FC, 0x4800, 0x2884, 0x1044, 0x2848, 0x4400, 0x85FE, 0x0}},
    {14035, 2, {0x2184, 0x2068, 0x2030, 0x20C8, 0xFB04, 0x4840, 0x4BFE, 0x48A0, 0x4920, 0x93FC, 0x5524, 0x2124, 0x3134, 0x4928, 0x4820, 0x8020}},
    {14036, 2, {0x1040, 0x1040, 0x10FC, 0x1088, 0xFD50, 0x2420, 0x24D8, 0x2726, 0x24F8, 0x4820, 0x28F8, 0x1020, 0x2BFE, 0x4420, 0x8020, 0x20}},
    {14037, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFCF8, 0x2400, 0x25FC, 0x2524, 0x2524, 0x4924, 0x29FC, 0x1100, 0x2900, 0x4502, 0x8502, 0xFE}},
    {14038, 2, {0x2040, 0x2040, 0x2088, 0x2104, 0xFBFE, 0x4802, 0x4888, 0x4944, 0x8A42, 0x48F8, 0x3188, 0x1250, 0x2820, 0x4850, 0x8188, 0x606}},
    {14039, 2, {0x1020, 0x1020, 0x10A8, 0x10A8, 0xFCA8, 0x2574, 0x2622, 0x2420, 0x2420, 0x49FC, 0x2820, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {14040, 2, {0x2020, 0x2020, 0x23FE, 0x2020, 0xF820, 0x4BFE, 0x4A02, 0x4C04, 0x49F8, 0x9010, 0x5020, 0x23FE, 0x3020, 0x4820, 0x48A0, 0x8040}},
    {14041, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFC88, 0x2450, 0x27FE, 0x2420, 0x2420, 0x4820, 0x29FC, 0x1020, 0x2820, 0x4420, 0x8420, 0x20}},
    {14042, 2, {0x1000, 0x11F8, 0x1008, 0x10D0, 0xFC20, 0x25FC, 0x2524, 0x2524, 0x25FC, 0x4924, 0x2924, 0x11FC, 0x2924, 0x4524, 0x8524, 0x10C}},
    {14043, 2, {0x1040, 0x1080, 0x11F8, 0x1108, 0xFDF8, 0x2508, 0x25FA, 0x250A, 0x250C, 0x4BF8, 0x2818, 0x1028, 0x2848, 0x4588, 0x8628, 0x10}},
    {14044, 2, {0x1040, 0x1040, 0x24FE, 0x6488, 0xA548, 0x2450, 0x2420, 0x2450, 0x228C, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14045, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x2504, 0x2504, 0x25FC, 0x2504, 0x4904, 0x29FC, 0x1000, 0x2888, 0x4484, 0x8102, 0x202}},
    {14046, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x2420, 0x2420, 0x27FE, 0x2488, 0x4888, 0x2BFE, 0x1088, 0x2888, 0x4508, 0x8108, 0x208}},
    {14047, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFE10, 0x25FC, 0x2524, 0x2524, 0x2524, 0x4924, 0x2BFE, 0x1050, 0x2850, 0x4488, 0x8104, 0x602}},
    {14048, 2, {0x2100, 0x20BE, 0x2282, 0x2202, 0xFA02, 0x4AFA, 0x4A8A, 0x4A8A, 0x4AFA, 0x928A, 0x528A, 0x22FA, 0x3202, 0x4A02, 0x4A0A, 0x8204}},
    {14049, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x25F8, 0x2400, 0x2400, 0x27FE, 0x4890, 0x2890, 0x1090, 0x2912, 0x4512, 0x820E, 0x400}},
    {14050, 2, {0x201C, 0x21E0, 0x2020, 0x2020, 0xFBFE, 0x48A8, 0x4924, 0x4A22, 0x89F8, 0x4888, 0x3090, 0x10BE, 0x2902, 0x4902, 0x8214, 0x408}},
    {14051, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFCF8, 0x2400, 0x25FC, 0x2504, 0x2524, 0x4924, 0x2924, 0x1124, 0x2850, 0x4448, 0x8084, 0x304}},
    {14052, 2, {0x2000, 0x23FE, 0x2000, 0x2000, 0xFBDE, 0x4A52, 0x4A52, 0x4A52, 0x4B5A, 0x92D6, 0x5252, 0x2252, 0x3252, 0x4A52, 0x4A52, 0x82D6}},
    {14053, 2, {0x1020, 0x1022, 0x13B4, 0x10A8, 0xFCA8, 0x2524, 0x26A2, 0x2440, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14054, 2, {0x2110, 0x2110, 0x2110, 0x2110, 0xF7BC, 0x5110, 0x5110, 0x5338, 0x53B8, 0x9554, 0x5554, 0x2992, 0x2110, 0x5110, 0x5110, 0x8110}},
    {14055, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x2400, 0x2400, 0x27FE, 0x2440, 0x4880, 0x28FC, 0x1184, 0x2A84, 0x4484, 0x84FC, 0x84}},
    {14056, 2, {0x2108, 0x2088, 0x2090, 0x23FC, 0xF840, 0x4840, 0x4BF8, 0x4840, 0x8840, 0x4FFE, 0x30A0, 0x10A0, 0x2920, 0x4922, 0x8222, 0x41E}},
    {14057, 2, {0x1040, 0x1020, 0x13FE, 0x1080, 0xFD04, 0x27FE, 0x2402, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x11FC, 0x2904, 0x4504, 0x8514, 0x108}},
    {14058, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x2524, 0x2622, 0x25FC, 0x2524, 0x49FC, 0x2924, 0x11FC, 0x2822, 0x4422, 0x841E, 0x0}},
    {14059, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0xFCFC, 0x2484, 0x2484, 0x24FC, 0x2440, 0x48FE, 0x292A, 0x124A, 0x2892, 0x4522, 0x844A, 0x84}},
    {14060, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xF820, 0x4BFE, 0x4888, 0x4904, 0x4A22, 0x9020, 0x51FC, 0x2020, 0x3020, 0x4820, 0x4BFE, 0x8000}},
    {14061, 2, {0x1088, 0x1088, 0x1088, 0x13FE, 0xFC88, 0x2488, 0x27FE, 0x2400, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14062, 2, {0x1040, 0x1020, 0x13FE, 0x1222, 0xFCA4, 0x2528, 0x2450, 0x2488, 0x2504, 0x48F8, 0x2888, 0x1088, 0x2850, 0x4420, 0x80D8, 0x306}},
    {14063, 2, {0x2210, 0x2210, 0x2390, 0x24BC, 0xF914, 0x57D4, 0x5554, 0x5554, 0x5554, 0x97D4, 0x52A4, 0x22AC, 0x22C2, 0x5482, 0x547E, 0x8800}},
    {14064, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD10, 0x25FE, 0x2510, 0x254A, 0x2586, 0x4800, 0x29FC, 0x1104, 0x29FC, 0x4504, 0x85FC, 0x104}},
    {14065, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x2524, 0x2524, 0x25FC, 0x2400, 0x4BFE, 0x2920, 0x1122, 0x2914, 0x4548, 0x8584, 0x102}},
    {14066, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0xFDFC, 0x2504, 0x2504, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x1104, 0x29FC, 0x4400, 0x8488, 0x104}},
    {14067, 2, {0x100, 0x7FFC, 0x820, 0x440, 0x7FFC, 0x4104, 0x1FF0, 0x1110, 0x1150, 0x1120, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {14068, 2, {0x2092, 0x2124, 0x2248, 0x2124, 0xF892, 0x4840, 0x4880, 0x4BFE, 0x4A02, 0x928A, 0x5252, 0x2222, 0x3252, 0x4A8A, 0x4BFE, 0x8202}},
    {14069, 2, {0x101E, 0x13E0, 0x1044, 0x1124, 0xFC88, 0x2420, 0x25FC, 0x2524, 0x2524, 0x49FC, 0x2924, 0x1124, 0x2BFE, 0x4504, 0x8514, 0x108}},
    {14070, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x2480, 0x2480, 0x27FE, 0x2490, 0x4910, 0x2952, 0x1254, 0x2AA8, 0x4428, 0x8844, 0x82}},
    {14071, 2, {0xC48, 0x7044, 0x13FE, 0xFC50, 0x3050, 0x5892, 0x9492, 0x110E, 0x1200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14072, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x25FC, 0x2400, 0x2400, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2800, 0x4488, 0x8450, 0x3FE}},
    {14073, 2, {0x2100, 0x21FE, 0x2110, 0x2210, 0xFAFE, 0x4A92, 0x4E92, 0x4AFE, 0x4A92, 0x9292, 0x52FE, 0x2290, 0x3250, 0x4A20, 0x4A58, 0x8286}},
    {14074, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFD04, 0x2488, 0x27FE, 0x2400, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14075, 2, {0x2020, 0x2020, 0x23FE, 0x2050, 0xF888, 0x4B26, 0x49FC, 0x4820, 0x4820, 0x93FE, 0x5000, 0x2020, 0x31FC, 0x4820, 0x4820, 0x83FE}},
    {14076, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x27FE, 0x2420, 0x24A0, 0x252C, 0x4924, 0x2924, 0x11AC, 0x2924, 0x4524, 0x85FC, 0x104}},
    {14077, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xF820, 0x4924, 0x48A8, 0x4820, 0x4BFE, 0x9070, 0x50A8, 0x20A8, 0x3124, 0x4A22, 0x4820, 0x8020}},
    {14078, 2, {0x2008, 0x200C, 0x200A, 0x23FE, 0xFA08, 0x4A08, 0x4AE8, 0x4A0A, 0x4A0A, 0x92EC, 0x52AC, 0x22A8, 0x32EA, 0x4A1A, 0x4A26, 0x8442}},
    {14079, 2, {0x20A8, 0x20A8, 0x23FE, 0x20A8, 0xF8A8, 0x4800, 0x4BFE, 0x4A22, 0x8820, 0x49FC, 0x3124, 0x1124, 0x2934, 0x4928, 0x8020, 0x20}},
    {14080, 2, {0x2100, 0x2100, 0x213C, 0x2124, 0xFBA4, 0x4924, 0x493C, 0x4BA4, 0x4B64, 0x9524, 0x593C, 0x2124, 0x3124, 0x4924, 0x493C, 0x8124}},
    {14081, 2, {0x1020, 0x1020, 0x13FE, 0x1070, 0xFCA8, 0x2524, 0x2622, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x11FC, 0x2904, 0x4400, 0x83FE, 0x0}},
    {14082, 2, {0x808, 0x7E08, 0x8FE, 0xFE08, 0x848, 0x7E28, 0x808, 0xF28, 0xF010, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14083, 2, {0x2090, 0x2088, 0x20FE, 0x2100, 0xF900, 0x4B3C, 0x4D00, 0x4900, 0x493C, 0x9100, 0x5100, 0x213C, 0x3124, 0x4924, 0x493C, 0x8124}},
    {14084, 2, {0x2000, 0x27FC, 0x2404, 0x2434, 0xF5C4, 0x5444, 0x55F4, 0x5554, 0x5554, 0x95F4, 0x5444, 0x2454, 0x25F4, 0x5494, 0x5806, 0x9002}},
    {14085, 2, {0x1040, 0x1020, 0x13FE, 0x1040, 0xFC88, 0x25F0, 0x2420, 0x2444, 0x25FE, 0x4802, 0x29FC, 0x1124, 0x29FC, 0x4524, 0x85FC, 0x104}},
    {14086, 2, {0x2010, 0x2092, 0x2452, 0x2254, 0xFA10, 0x48FE, 0x4810, 0x4E38, 0x8A54, 0x4A92, 0x3310, 0x1210, 0x2A10, 0x4500, 0x88FE, 0x0}},
    {14087, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xFBFE, 0x4820, 0x49FC, 0x4820, 0x8BFE, 0x4800, 0x3040, 0x1024, 0x2AA2, 0x4A8A, 0x8488, 0x78}},
    {14088, 2, {0x20, 0x7E7C, 0x14C4, 0x838, 0xFEC6, 0x2A20, 0x48FC, 0xA844, 0x128C, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14089, 2, {0x2124, 0x2248, 0x2248, 0x2124, 0xFBF8, 0x4A08, 0x4BF8, 0x4800, 0x4BF8, 0x9248, 0x5248, 0x23F8, 0x3202, 0x4A02, 0x49FE, 0x8000}},
    {14090, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x4A48, 0x4A48, 0x4AFC, 0x8A48, 0x4A48, 0x33FE, 0x12A4, 0x2AA8, 0x4490, 0x84C8, 0x886}},
    {14091, 2, {0x2104, 0x2084, 0x2088, 0x2000, 0xFBFE, 0x4800, 0x4888, 0x4904, 0x8A02, 0x49FC, 0x3154, 0x1154, 0x2954, 0x4954, 0x87FE, 0x0}},
    {14092, 2, {0x1000, 0x1078, 0xFE48, 0x1048, 0x7C48, 0x86, 0xFE00, 0x82FC, 0x7C44, 0x1044, 0xFC28, 0x2428, 0x6810, 0x3028, 0x4844, 0x8582}},
    {14093, 2, {0x1020, 0x1010, 0x11FE, 0x1110, 0xFD7C, 0x2514, 0x25FE, 0x2514, 0x257C, 0x4910, 0x297C, 0x1144, 0x2944, 0x4644, 0x827C, 0x444}},
    {14094, 2, {0x101E, 0x13E0, 0x1122, 0x1094, 0xFC40, 0x2488, 0x25F0, 0x2420, 0x24C4, 0x49FE, 0x2822, 0x1020, 0x2BFE, 0x4450, 0x8488, 0x306}},
    {14095, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0xFCFC, 0x2400, 0x25FE, 0x2502, 0x25FE, 0x4902, 0x29FE, 0x1102, 0x29FE, 0x4448, 0x8484, 0x102}},
    {14096, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x2400, 0x27FE, 0x2400, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2888, 0x4450, 0x83FE, 0x0}},
    {14097, 2, {0x4048, 0x2FFE, 0x40, 0x87FC, 0x5444, 0x17FC, 0xE444, 0x27FC, 0x2444, 0x2000, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {14098, 2, {0x1040, 0x119C, 0x1104, 0x1104, 0xFDDC, 0x2504, 0x2504, 0x25FC, 0x2488, 0x4888, 0x2BFE, 0x1088, 0x2888, 0x4488, 0x8508, 0x208}},
    {14099, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC20, 0x2450, 0x2488, 0x2504, 0x26FA, 0x4800, 0x2800, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {14100, 2, {0x1080, 0x1040, 0x13FC, 0x1108, 0xFC90, 0x27FE, 0x2400, 0x25F8, 0x2508, 0x4908, 0x29F8, 0x1090, 0x2890, 0x4512, 0x8212, 0x40E}},
    {14101, 2, {0x2010, 0x2010, 0x22FE, 0x2110, 0xF97C, 0x4854, 0x487C, 0x4B54, 0x897C, 0x4910, 0x31FE, 0x1110, 0x2910, 0x4A90, 0x847E, 0x0}},
    {14102, 2, {0x2090, 0x2090, 0x27FE, 0x2090, 0xFBFC, 0x4A94, 0x4BFC, 0x4A94, 0x4BFC, 0x9000, 0x51F8, 0x2108, 0x31F8, 0x4908, 0x49F8, 0x8108}},
    {14103, 2, {0x2040, 0x2088, 0x21FC, 0x2108, 0xFA52, 0x4BFE, 0x4850, 0x4988, 0x4E26, 0x90C0, 0x5310, 0x2064, 0x3388, 0x4830, 0x48C0, 0x8700}},
    {14104, 2, {0x1020, 0x103C, 0x1020, 0x11FE, 0xFD22, 0x2538, 0x25E4, 0x251C, 0x2500, 0x497C, 0x2944, 0x117C, 0x2944, 0x457C, 0x8544, 0x2FE}},
    {14105, 2, {0x2040, 0x2020, 0x23FE, 0x2250, 0xFA50, 0x4BFE, 0x4A52, 0x4A52, 0x8BFE, 0x4A00, 0x3292, 0x12D4, 0x2A98, 0x4492, 0x84D2, 0x88E}},
    {14106, 2, {0x2040, 0x2020, 0x27FE, 0x2402, 0xF900, 0x49FE, 0x4A10, 0x4E20, 0x4A7C, 0x9244, 0x5244, 0x227C, 0x3244, 0x4A44, 0x4A7C, 0x8244}},
    {14107, 2, {0x2108, 0x2108, 0x2108, 0x27C8, 0xF11E, 0x57D2, 0x5564, 0x57C0, 0x5548, 0x97C8, 0x5108, 0x27C8, 0x2114, 0x5114, 0x5124, 0x8142}},
    {14108, 2, {0x2108, 0x2108, 0x2108, 0x27D0, 0xF11E, 0x5794, 0x5124, 0x5FD4, 0x5214, 0x93D4, 0x5254, 0x2248, 0x2248, 0x54D4, 0x5424, 0x8842}},
    {14109, 2, {0x2020, 0x2050, 0x2088, 0x2124, 0xFAFA, 0x4810, 0x4820, 0x4BFE, 0x4850, 0x91FC, 0x5154, 0x218C, 0x3104, 0x49FC, 0x4904, 0x81FC}},
    {14110, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0xA88, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {14111, 2, {0x2008, 0x2388, 0x2088, 0x2110, 0xF7DE, 0x5294, 0x52A4, 0x5394, 0x5294, 0x9394, 0x5294, 0x22C8, 0x2388, 0x5694, 0x50A4, 0x80C2}},
    {14112, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0xFC88, 0x25FE, 0x2710, 0x2510, 0x25FE, 0x4910, 0x2910, 0x11FE, 0x2910, 0x4510, 0x85FE, 0x100}},
    {14113, 2, {0x2124, 0x2124, 0x2224, 0x2424, 0xF954, 0x494A, 0x4A92, 0x4E10, 0x4A10, 0x9250, 0x525C, 0x2250, 0x3250, 0x4AB0, 0x4A9E, 0x8300}},
    {14114, 2, {0x1020, 0x11FC, 0x1124, 0x13FE, 0xFD24, 0x25FC, 0x2420, 0x25FC, 0x2524, 0x49FC, 0x2840, 0x13FE, 0x2888, 0x45D0, 0x8470, 0x38C}},
    {14115, 2, {0x2110, 0x2114, 0x21D2, 0x2250, 0xFA7E, 0x4D50, 0x4890, 0x4AA8, 0x4928, 0x9144, 0x5244, 0x2482, 0x3008, 0x4AA4, 0x4A52, 0x8452}},
    {14116, 2, {0x2000, 0x23FE, 0x2050, 0x23FE, 0xFA52, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x91FC, 0x5104, 0x21FC, 0x3020, 0x4BFE, 0x4820, 0x8020}},
    {14117, 2, {0x2000, 0x23FE, 0x2252, 0x2252, 0xFBFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x9104, 0x51FC, 0x2104, 0x31FC, 0x4888, 0x4904, 0x8202}},
    {14118, 2, {0x2000, 0x23FC, 0x2004, 0x23FC, 0xF804, 0x4BFC, 0x4800, 0x4BBC, 0x4924, 0x97BC, 0x5008, 0x23FE, 0x3108, 0x4888, 0x48A8, 0x8010}},
    {14119, 2, {0x2000, 0x27FE, 0x2400, 0x2628, 0xF548, 0x57EE, 0x5492, 0x5484, 0x56A0, 0x96A8, 0x57E8, 0x2488, 0x2494, 0x5514, 0x5924, 0x9242}},
    {14120, 2, {0x2000, 0x23FE, 0x2202, 0x23FE, 0xFA10, 0x4A54, 0x4A38, 0x4A54, 0x8A82, 0x4A50, 0x327C, 0x1290, 0x2A10, 0x45FE, 0x8410, 0x810}},
    {14121, 2, {0x2040, 0x2080, 0x23FE, 0x2222, 0xFAAA, 0x4A22, 0x4BFE, 0x4A72, 0x4AAA, 0x9202, 0x5020, 0x23FE, 0x3020, 0x4850, 0x4888, 0x8306}},
    {14122, 2, {0x804, 0x7F78, 0x2240, 0xFF40, 0x87E, 0x7F48, 0x2A48, 0x4948, 0xA888, 0x1400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {14123, 2, {0x2288, 0x22A8, 0x26A8, 0x2BF0, 0xF01E, 0x5224, 0x55D4, 0x5C14, 0x5414, 0x95D4, 0x5554, 0x2548, 0x2568, 0x5554, 0x5614, 0x8422}},
    {14124, 2, {0x2100, 0x21F8, 0x2208, 0x27FE, 0xFA88, 0x4B24, 0x4BFE, 0x4A00, 0x8AFC, 0x4A00, 0x32FC, 0x1200, 0x2AFC, 0x4484, 0x84FC, 0x884}},
    {14125, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFE02, 0x24F8, 0x2488, 0x2488, 0x24F8, 0x4800, 0x29FC, 0x1124, 0x29FC, 0x4524, 0x85FC, 0x104}},
    {14126, 2, {0x2008, 0x23E8, 0x2288, 0x23EE, 0xFA28, 0x4BF4, 0x4A82, 0x4BE2, 0x8800, 0x49FC, 0x3154, 0x1154, 0x2954, 0x4954, 0x87FE, 0x0}},
    {14127, 2, {0x2000, 0x27DE, 0x2492, 0x2492, 0xF5D2, 0x555E, 0x5550, 0x55D0, 0x5492, 0x9492, 0x57CE, 0x2000, 0x2000, 0x5524, 0x5492, 0x8892}},
    {14128, 2, {0x2124, 0x2174, 0x2258, 0x22DA, 0xFBFE, 0x4954, 0x4ADA, 0x4BFE, 0x88A2, 0x4820, 0x33FE, 0x1070, 0x28A8, 0x4924, 0x8222, 0x20}},
    {14129, 2, {0x2042, 0x239C, 0x2210, 0x23DE, 0xFA94, 0x4A94, 0x4C20, 0x4BFC, 0x4A04, 0x93FC, 0x5204, 0x23FC, 0x3204, 0x4BFC, 0x4908, 0x8204}},
    {14130, 2, {0x2040, 0x2FFE, 0x2400, 0x27FC, 0xF000, 0x57FC, 0x5404, 0x57FC, 0x5080, 0x9C9C, 0x57D4, 0x3D54, 0x355C, 0x5C96, 0x5556, 0xAE22}},
    {14131, 2, {0x2080, 0x2040, 0x27FC, 0x2000, 0xFBF8, 0x4AA8, 0x4BF8, 0x4840, 0x4A44, 0x9118, 0x5244, 0x20A8, 0x3190, 0x4A88, 0x4CA6, 0x80C0}},
    {14132, 2, {0x2108, 0x27FE, 0x2108, 0x23FC, 0xFA94, 0x4BFC, 0x4800, 0x4FFE, 0x4C02, 0x91F8, 0x5108, 0x21F8, 0x3108, 0x49F8, 0x4908, 0x81F8}},
    {14133, 2, {0x2040, 0x27FE, 0x2000, 0x23FC, 0xFA40, 0x4BF8, 0x4A40, 0x4BF8, 0x4A40, 0x93FC, 0x5004, 0x2554, 0x30A8, 0x4B10, 0x4D48, 0x8186}},
    {14134, 2, {0x21F0, 0x2210, 0x27FC, 0x2204, 0xFBFC, 0x4A24, 0x4BB8, 0x4A22, 0x49FE, 0x9210, 0x57FC, 0x2244, 0x33FC, 0x48D0, 0x494A, 0x863E}},
    {14135, 2, {0xF7FC, 0x1248, 0x17FC, 0xFC46, 0x86EC, 0xF554, 0x16EC, 0x1554, 0x56EC, 0x2444, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {14136, 2, {0x820, 0xFFFE, 0x2820, 0x7DFC, 0x4488, 0x7C50, 0x41FE, 0x7C20, 0x45FC, 0x7C20, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {14137, 2, {0x23DE, 0x2252, 0x23DE, 0x2252, 0xFBDE, 0x4A52, 0x4BDE, 0x4890, 0x49FE, 0x9110, 0x53FE, 0x2510, 0x31FE, 0x4910, 0x49FE, 0x8100}},
    {14138, 2, {0x2528, 0x27BE, 0x2948, 0x27BE, 0xFB18, 0x4DAA, 0x4946, 0x4BFC, 0x4A04, 0x93FC, 0x5204, 0x23FC, 0x3204, 0x4BFC, 0x4908, 0x8204}},
    {14139, 2, {0x2444, 0x24E4, 0x2A0A, 0x2EEE, 0xF404, 0x5AEA, 0x5E0E, 0x50E0, 0x5AAA, 0x9AEA, 0x5040, 0x2024, 0x2522, 0x550A, 0x5908, 0x80F8}},
    {14140, 2, {0x21FC, 0x2124, 0x21FC, 0x2124, 0xFBFE, 0x4AAA, 0x4BFE, 0x4AAA, 0x4BFE, 0x9000, 0x53FE, 0x228A, 0x30F8, 0x4888, 0x4888, 0x83FE}},
    {14141, 2, {0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x0, 0x1FE0, 0x40, 0x80, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {14142, 2, {0x20, 0xFC20, 0x450, 0x850, 0x1088, 0x1104, 0x1602, 0x1888, 0x3088, 0xD088, 0x1088, 0x1088, 0x1088, 0x1108, 0x5108, 0x2208}},
    {14143, 2, {0x0, 0xFDF8, 0x408, 0x850, 0x1020, 0x1010, 0x15FE, 0x1822, 0x3024, 0xD020, 0x1020, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {14144, 2, {0x0, 0x3E7C, 0x2204, 0x2A08, 0x2608, 0x2210, 0x2210, 0xFF7E, 0x4210, 0x5210, 0x4A10, 0x4210, 0x7F90, 0x210, 0x1450, 0x820}},
    {14145, 2, {0x1000, 0x1FF8, 0x2008, 0x5FC8, 0x9048, 0x1FC8, 0x1048, 0x1FE8, 0x10, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {14146, 2, {0x104, 0xFD0E, 0x530, 0x920, 0x17E0, 0x1120, 0x153E, 0x1964, 0x31A4, 0xD724, 0x1124, 0x1124, 0x1124, 0x1124, 0x5544, 0x2284}},
    {14147, 2, {0x80, 0xFC80, 0x4F8, 0x908, 0x1210, 0x11FC, 0x1524, 0x1924, 0x3124, 0xD1FC, 0x1050, 0x1050, 0x1090, 0x1092, 0x5112, 0x220E}},
    {14148, 2, {0x2020, 0x1020, 0xFE20, 0x1FC, 0x7C20, 0x44A8, 0x7CA8, 0xA8, 0x7CA8, 0x4A8, 0x8F8, 0x1E22, 0xF022, 0x1022, 0x501E, 0x2000}},
    {14149, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0x7C44, 0x828, 0x1E28, 0xF010, 0x1028, 0x5044, 0x2182}},
    {14150, 2, {0x20, 0xFDFE, 0x420, 0x9FC, 0x1000, 0x11FC, 0x1504, 0x19FC, 0x3088, 0xD3FE, 0x1000, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {14151, 2, {0x2020, 0x1010, 0xFEFE, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x0, 0x7C7C, 0x808, 0x1010, 0x1EFE, 0xF010, 0x1010, 0x5050, 0x2020}},
    {14152, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x0, 0x7C7C, 0x808, 0xFEFE, 0x1010, 0x3030}},
    {14153, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0}},
    {14154, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x3FF0, 0x1010, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {14155, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FE0, 0x40, 0x80, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {14156, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4202, 0x8204, 0x7E0, 0x420, 0x840, 0x1040, 0x80, 0x180, 0x240, 0x420, 0x1810, 0x600C}},
    {14157, 2, {0x200, 0x100, 0x7FFE, 0x5002, 0x9004, 0x1FF8, 0x1000, 0x2000, 0x3FF8, 0x8, 0x8, 0xFFC8, 0x8, 0x8, 0x50, 0x20}},
    {14158, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x100, 0x100, 0x7FF8, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x4800, 0x87FE, 0x0}},
    {14159, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x280, 0x440, 0x820, 0x3458, 0xC446, 0x440, 0x440, 0x840, 0x840, 0x1040, 0x2040}},
    {14160, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4082, 0x8484, 0x440, 0x840, 0x1020, 0x2210, 0x4208, 0x8406, 0x840, 0x1020, 0x3FF0, 0x1010}},
    {14161, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x100, 0x7FFC, 0x400, 0x400, 0x7F0, 0x410, 0x810, 0x810, 0x1010, 0x20A0, 0x4040}},
    {14162, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {14163, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {14164, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x100, 0x7FFC, 0x820, 0x1010, 0x2448, 0x4444, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {14165, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8104, 0x900, 0x2828, 0x2824, 0x47E4, 0x0, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {14166, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x87E4, 0x1840, 0x6480, 0x320, 0xC40, 0x71F8, 0x608, 0x1A10, 0x120, 0xC0, 0x700, 0x7800}},
    {14167, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8104, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1104, 0x1088, 0x1050, 0x1220, 0x1418, 0x1806}},
    {14168, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8104, 0x3FF8, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {14169, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {14170, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x801C, 0x7BE0, 0x840, 0x1040, 0x2278, 0x7A40, 0xA40, 0x2BFC, 0x1000, 0x2800, 0x47FE, 0x8000}},
    {14171, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8804, 0xFFC, 0x1040, 0x3040, 0x53F8, 0x9248, 0x1248, 0x12A8, 0x1318, 0x1208, 0x13F8, 0x1208}},
    {14172, 2, {0x200, 0x100, 0x7FFE, 0x4442, 0x8444, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {14173, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x820, 0x820, 0x7EFC, 0x820, 0x1C30, 0x1A70, 0x2A68, 0x48A4, 0x8922, 0x820, 0x820}},
    {14174, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x7FFC, 0x440, 0x440, 0x3C78, 0x2008, 0x2008, 0x3C78, 0x440, 0x440, 0xFFFE, 0x0}},
    {14175, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xBFFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {14176, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1210, 0x100, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {14177, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x7F00, 0x22FC, 0x3E44, 0x2244, 0x3E44, 0x2228, 0x2228, 0x2F10, 0xF228, 0x4244, 0x282}},
    {14178, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x80F4, 0x3F00, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x4124, 0x41F4, 0x5E14, 0x4008}},
    {14179, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8804, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x200, 0xFFFE, 0x440, 0x1830, 0xE00E}},
    {14180, 2, {0x100, 0x7FFE, 0x4822, 0xBFFC, 0x820, 0x3FF8, 0x820, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x2008, 0x2018}},
    {14181, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8424, 0x2420, 0x2524, 0x24A8, 0x3C20, 0x7FE, 0x470, 0xFCA8, 0x2524, 0x2622, 0x2420, 0x4420}},
    {14182, 2, {0x100, 0x7FFE, 0x4902, 0x9FFC, 0x3100, 0x5FF0, 0x9100, 0x1FF0, 0x1100, 0x1FF8, 0x1000, 0x3FE0, 0x840, 0x8F8, 0x1008, 0x6030}},
    {14183, 2, {0x100, 0x7FFE, 0x4402, 0xBFFC, 0x820, 0x7C0, 0x7838, 0x1110, 0x1110, 0x2928, 0x4544, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {14184, 2, {0x200, 0x100, 0x7FFE, 0x4102, 0x9FF4, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {14185, 2, {0x100, 0x7FFE, 0x4002, 0x8104, 0x3FF8, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x24C4, 0x22A8, 0x4498, 0x5886, 0x8180}},
    {14186, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9044, 0x1040, 0x7C40, 0x11F8, 0xFE48, 0x4448, 0x28C8, 0x7C48, 0x10AA, 0xFEAA, 0x1106, 0x1202}},
    {14187, 2, {0x100, 0x7FFE, 0x4002, 0x8104, 0x79F0, 0x4A90, 0x5060, 0x6198, 0x5606, 0x49F0, 0x4A40, 0x6BF8, 0x5440, 0x43F8, 0x4040, 0x47FC}},
    {14188, 2, {0x2020, 0x1020, 0xFE20, 0x8220, 0x11FC, 0x7C20, 0x1020, 0x7C20, 0x11FE, 0xFE20, 0x1050, 0x7C50, 0x4488, 0x4488, 0x7D04, 0x4602}},
    {14189, 2, {0x100, 0x7FFE, 0x4002, 0x9004, 0x13F8, 0x1208, 0x5BF8, 0x5608, 0x53F8, 0x9140, 0x13F8, 0x1440, 0x11F0, 0x1040, 0x17FC, 0x1000}},
    {14190, 2, {0x100, 0x7FFE, 0x4002, 0x8804, 0x2A00, 0x3E7C, 0x4944, 0x7F44, 0x44, 0x3E28, 0x28, 0x7F10, 0x810, 0x2A28, 0x4944, 0x1882}},
    {14191, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8074, 0x1F80, 0x200, 0xFFFE, 0x820, 0x37D8, 0xC446, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27D8}},
    {14192, 2, {0x100, 0x7FFE, 0x4202, 0x9FF4, 0x1490, 0x1250, 0x1130, 0x1FF0, 0x100, 0xFFFE, 0x9212, 0x2FE8, 0x4404, 0x7E0, 0x820, 0x3060}},
    {14193, 2, {0x100, 0x7FFE, 0x408A, 0x8BFC, 0x4888, 0x49FC, 0x4954, 0x79FC, 0x800, 0xBFE, 0xFA42, 0x4878, 0x4888, 0x4950, 0x8860, 0x980}},
    {14194, 2, {0x100, 0x7FFE, 0x4112, 0x97FC, 0x5110, 0x53F8, 0x52A8, 0x77FC, 0x1414, 0x11F0, 0xF010, 0x57FC, 0x5444, 0x53F8, 0x9248, 0x1258}},
    {14195, 2, {0x200, 0x200, 0x200, 0x200, 0xFFFC, 0x200, 0x200, 0x1290, 0x1248, 0x1224, 0x2224, 0x2212, 0x4212, 0x8200, 0xA00, 0x400}},
    {14196, 2, {0x8, 0x7C08, 0x8, 0x1FE, 0x8, 0xFE08, 0x2488, 0x2448, 0x2448, 0x2408, 0x2408, 0x2428, 0x2412, 0x4402, 0x43FE, 0x8000}},
    {14197, 2, {0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x7FFC, 0x20, 0x20, 0xFFFE, 0x820, 0x420, 0x420, 0xA0, 0x40}},
    {14198, 2, {0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x440, 0x442, 0x842, 0x303E, 0xC000, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {14199, 2, {0x3FF0, 0x10, 0x1FF0, 0x10, 0x3FF0, 0x40, 0x7DF0, 0x1150, 0x1C94, 0xE14C, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {14200, 2, {0x100, 0x1110, 0x2108, 0x4504, 0x800, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x440, 0x842, 0x1042, 0x603E}},
    {14201, 2, {0x100, 0x1110, 0x2168, 0x4384, 0x1C00, 0xE800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2108, 0x2108, 0x3FF8, 0x492, 0x188A, 0xE07E}},
    {14202, 2, {0x2210, 0x1410, 0xFF90, 0x1410, 0x1414, 0x7F52, 0x1552, 0xFF92, 0x1510, 0x7F14, 0x1404, 0x3608, 0x5510, 0x94A0, 0x1440, 0x1580}},
    {14203, 2, {0x2210, 0x2210, 0xFF90, 0x2210, 0x3E54, 0x852, 0x7F52, 0x4990, 0x7F10, 0x814, 0x7F04, 0x808, 0x7F08, 0x810, 0xF20, 0xF0C0}},
    {14204, 2, {0x2000, 0x21FC, 0x2020, 0x2020, 0xFC20, 0x2820, 0x2BFE, 0x2820, 0x2820, 0x2820, 0x2820, 0x28A0, 0x2842, 0x4802, 0x47FE, 0x8000}},
    {14205, 2, {0x2020, 0x2020, 0x2020, 0x2024, 0xFDA8, 0x28B0, 0x28B0, 0x2928, 0x2924, 0x2A24, 0x28A0, 0x2840, 0x2802, 0x4802, 0x47FE, 0x8000}},
    {14206, 2, {0x2040, 0x2040, 0x2040, 0x23FC, 0xFC80, 0x2880, 0x28F8, 0x2920, 0x2920, 0x2A20, 0x2820, 0x29FC, 0x2802, 0x4802, 0x47FE, 0x8000}},
    {14207, 2, {0x2020, 0x2020, 0x21FC, 0x2124, 0xFD28, 0x2920, 0x29F8, 0x2908, 0x2950, 0x2920, 0x2A50, 0x2A88, 0x2D0A, 0x4802, 0x47FE, 0x8000}},
    {14208, 2, {0x2040, 0x2040, 0x2040, 0x23FE, 0xFC80, 0x2890, 0x2912, 0x2954, 0x2A90, 0x2C28, 0x2824, 0x2844, 0x2982, 0x4802, 0x47FE, 0x8000}},
    {14209, 2, {0x2000, 0x21F8, 0x2108, 0x21F8, 0xFD08, 0x29F8, 0x2800, 0x2918, 0x29E0, 0x2904, 0x2904, 0x28FC, 0x2802, 0x4802, 0x47FE, 0x8000}},
    {14210, 2, {0x201C, 0x21E0, 0x2044, 0x2124, 0xFC88, 0x2820, 0x2820, 0x2BFE, 0x2848, 0x28C8, 0x2830, 0x2868, 0x2986, 0x4802, 0x47FE, 0x8000}},
    {14211, 2, {0x2840, 0x2448, 0x4244, 0x9244, 0x1040, 0x29FE, 0x4450, 0x8250, 0x7C50, 0x4450, 0x4450, 0x4452, 0x4452, 0x7C92, 0x448E, 0x100}},
    {14212, 2, {0x2040, 0x2040, 0x23FC, 0x20A0, 0xFD10, 0x2A48, 0x2DF4, 0x2950, 0x29F0, 0x2950, 0x29F4, 0x2844, 0x283E, 0x4802, 0x47FE, 0x8000}},
    {14213, 2, {0x2000, 0x21FC, 0x2124, 0x21FC, 0xFD24, 0x29FC, 0x2820, 0x29FE, 0x2870, 0x28A8, 0x2924, 0x2A20, 0x2822, 0x4802, 0x47FE, 0x8000}},
    {14214, 2, {0x2020, 0x2020, 0x21FC, 0x2050, 0xFC88, 0x2BFE, 0x2808, 0x29E8, 0x2928, 0x2928, 0x29E8, 0x2928, 0x2808, 0x481A, 0x4802, 0x87FE}},
    {14215, 2, {0x2000, 0x21FC, 0x2124, 0x21FC, 0xFD24, 0x29FC, 0x2800, 0x2BFE, 0x2922, 0x2914, 0x2948, 0x2984, 0x2902, 0x4802, 0x47FE, 0x8000}},
    {14216, 2, {0x2040, 0x2020, 0x21FC, 0x2104, 0xFD04, 0x29FC, 0x2900, 0x29FC, 0x2B54, 0x2B54, 0x2DFC, 0x2954, 0x2954, 0x490E, 0x4802, 0x87FE}},
    {14217, 2, {0x2088, 0x2050, 0x23FE, 0x2020, 0xFDFC, 0x2840, 0x2BFE, 0x2880, 0x29FC, 0x2A20, 0x2C20, 0x2BFE, 0x2802, 0x4802, 0x47FE, 0x8000}},
    {14218, 2, {0x1040, 0x5440, 0x5440, 0x7C40, 0x92FC, 0x9250, 0xFE50, 0x50, 0x7C50, 0x50, 0xFE50, 0x1052, 0x5492, 0x9292, 0x510E, 0x2200}},
    {14219, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x2FF8, 0x2408, 0x2410, 0x2220, 0x2140, 0x2080, 0x4140, 0x4630, 0x980E}},
    {14220, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x27F8, 0x2010, 0x2020, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x4040, 0x4140, 0x8080}},
    {14221, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2100, 0x2100, 0x2FE0, 0x2120, 0x2520, 0x2220, 0x2320, 0x24A2, 0x4422, 0x481E, 0x9000}},
    {14222, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2FF8, 0x2888, 0x2888, 0x2888, 0x2FF8, 0x2800, 0x2802, 0x4802, 0x4802, 0x87FE}},
    {14223, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2078, 0x2780, 0x2080, 0x2040, 0x2FF8, 0x2010, 0x2060, 0x4180, 0x4600, 0x99FE}},
    {14224, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2080, 0x2040, 0x2FFC, 0x2000, 0x2408, 0x2210, 0x2120, 0x4000, 0x5FFE, 0x8000}},
    {14225, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2888, 0x2490, 0x22A0, 0x2080, 0x22A0, 0x2492, 0x288A, 0x4082, 0x407E, 0x8000}},
    {14226, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2100, 0x2200, 0x2FF8, 0x2808, 0x2808, 0x2FF8, 0x2808, 0x2808, 0x4FF8, 0x4808, 0x8808, 0xFF8}},
    {14227, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2860, 0x2F88, 0x2808, 0x27F8, 0x2000, 0x2FF8, 0x2808, 0x4FF8, 0x4808, 0x8FF8, 0x808}},
    {14228, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2100, 0x23F0, 0x2C20, 0x2240, 0x2180, 0x2240, 0x2CF8, 0x2108, 0x4690, 0x4060, 0x8180, 0xE00}},
    {14229, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2220, 0x2410, 0x2808, 0x37F6, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x4410, 0x4410, 0x8450, 0x420}},
    {14230, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x3FFC, 0x2080, 0x2FF8, 0x2888, 0x28A8, 0x29D0, 0x22A0, 0x4490, 0x588C, 0x8080}},
    {14231, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2090, 0x2088, 0x2FFC, 0x2080, 0x2888, 0x25D0, 0x22A0, 0x2490, 0x488C, 0x4280, 0x8100}},
    {14232, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2FFC, 0x2100, 0x2308, 0x2488, 0x2D50, 0x2260, 0x25E0, 0x2950, 0x4250, 0x4448, 0x9944, 0x80}},
    {14233, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x20F0, 0x3F00, 0x21E0, 0x2F00, 0x21F4, 0x3F04, 0x20FC, 0x2088, 0x4ED0, 0x42A0, 0x8498, 0x1986}},
    {14234, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2038, 0x27C0, 0x2040, 0x2FFE, 0x2040, 0x2140, 0x265C, 0x2444, 0x475C, 0x4444, 0x87FC, 0x404}},
    {14235, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2288, 0x2490, 0x28E4, 0x3284, 0x247C, 0x2C00, 0x34FC, 0x2484, 0x44FC, 0x4484, 0x84FC, 0x484}},
    {14236, 2, {0x3FFC, 0x2004, 0x3FFC, 0x220C, 0x2470, 0x2810, 0x30FE, 0x2238, 0x2454, 0x2C92, 0x3420, 0x25FE, 0x4444, 0x44C8, 0x8438, 0x5C6}},
    {14237, 2, {0x24, 0x3FFE, 0x2220, 0x23A4, 0x2224, 0x3FA8, 0x2B10, 0x529A, 0x4626, 0x8042, 0x1FF8, 0x1008, 0x1FF8, 0x1000, 0x2000, 0x4000}},
    {14238, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2448, 0x2848, 0x3054, 0x22A4, 0x2500, 0x2C10, 0x3490, 0x249C, 0x4490, 0x4550, 0x863E, 0x400}},
    {14239, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2220, 0x2FF8, 0x2140, 0x2FF8, 0x2A28, 0x2DD8, 0x2808, 0x2FF8, 0x2080, 0x5FFC, 0x4140, 0x8630, 0x180C}},
    {14240, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x2FF8, 0x2140, 0x2A28, 0x27F0, 0x2C18, 0x37F4, 0x2410, 0x27F0, 0x4080, 0x4888, 0x9284, 0x100}},
    {14241, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2738, 0x2210, 0x2F7C, 0x2000, 0x2738, 0x2210, 0x23BC, 0x2EC8, 0x4130, 0x4310, 0x8D48, 0x186}},
    {14242, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {14243, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x540, 0x820, 0x3018, 0xDFF6, 0x210, 0x210, 0x210, 0x410, 0x8A0, 0x1040}},
    {14244, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x0}},
    {14245, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x600, 0x100, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {14246, 2, {0x1000, 0x10F0, 0x1090, 0x1090, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5C92, 0x6492, 0x112, 0x10E, 0x200}},
    {14247, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x3FF0, 0x410, 0x420, 0x87C, 0x804, 0x1004, 0x2004, 0x4028, 0x8010}},
    {14248, 2, {0x1000, 0x11FC, 0x1004, 0x1008, 0x5410, 0x5420, 0x5420, 0x57FE, 0x5420, 0x5420, 0x5420, 0x5C20, 0x6420, 0x20, 0xA0, 0x40}},
    {14249, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0xFE0, 0x820, 0x820, 0xA20, 0x920, 0x922, 0x1022, 0x1022, 0x201E, 0x4000}},
    {14250, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x1FF0, 0x10, 0x10, 0xFFFE, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14251, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x100, 0x7FF8, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x4800, 0x87FE}},
    {14252, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {14253, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14254, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x800, 0x800, 0x1FF8, 0x1008, 0x2010, 0x4020, 0x40, 0xC80, 0x300, 0x80, 0x40}},
    {14255, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x3FF0, 0x2000, 0x2004, 0x2004, 0x1FFC, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14256, 2, {0x1080, 0x1080, 0x1080, 0x11FC, 0x5520, 0x5620, 0x5420, 0x5420, 0x57FE, 0x5420, 0x5420, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {14257, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {14258, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x55FC, 0x5524, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5524, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {14259, 2, {0x1008, 0x1088, 0x1048, 0x1048, 0x5408, 0x5488, 0x5448, 0x5448, 0x5408, 0x540E, 0x55F8, 0x5C08, 0x6408, 0x8, 0x8, 0x8}},
    {14260, 2, {0x1004, 0x101E, 0x11F0, 0x1110, 0x5510, 0x5510, 0x5510, 0x55FE, 0x5510, 0x5510, 0x5510, 0x5D08, 0x650A, 0x14A, 0x186, 0x102}},
    {14261, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x7FFC, 0x820, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {14262, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x5488, 0x5488, 0x5488, 0x5488, 0x5488, 0x5450, 0x5450, 0x5C20, 0x6450, 0x88, 0x104, 0x602}},
    {14263, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {14264, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x100, 0x300, 0x560, 0x910, 0x3108, 0xC104, 0x100, 0x100}},
    {14265, 2, {0x1000, 0x1000, 0x13FE, 0x1008, 0x5408, 0x55E8, 0x5528, 0x5528, 0x5528, 0x5528, 0x55E8, 0x5D28, 0x6408, 0x8, 0x28, 0x10}},
    {14266, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x800, 0x800, 0x1FF0, 0x2100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {14267, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x1000, 0x1FF0, 0x10, 0x10, 0xA0, 0x40}},
    {14268, 2, {0x0, 0x3FF0, 0x20, 0xC0, 0x7D04, 0x588, 0x950, 0x1120, 0x2118, 0xC506, 0x200, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14269, 2, {0x2028, 0x2024, 0x2024, 0x2020, 0xABFE, 0xAA20, 0xAA24, 0xAA24, 0xAA24, 0xAA28, 0xAA28, 0xBA90, 0xCB12, 0x22A, 0x46, 0x82}},
    {14270, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x5504, 0x55FC, 0x5500, 0x5540, 0x5544, 0x5548, 0x5570, 0x5D40, 0x6542, 0x242, 0x23E, 0x400}},
    {14271, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x1010, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {14272, 2, {0x70, 0x1F80, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC006, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14273, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0x557C, 0x5510, 0x5510, 0x5510, 0x5510, 0x55FE, 0x5510, 0x5D10, 0x6510, 0x210, 0x210, 0x410}},
    {14274, 2, {0x1000, 0x10FC, 0x1084, 0x10A4, 0x5494, 0x5484, 0x5484, 0x57FE, 0x5504, 0x5544, 0x5524, 0x5D04, 0x65FE, 0x4, 0x28, 0x10}},
    {14275, 2, {0x1000, 0x11FC, 0x1004, 0x1004, 0x55F4, 0x5404, 0x5404, 0x55F4, 0x5514, 0x5514, 0x55F4, 0x5D14, 0x6404, 0x4, 0x28, 0x10}},
    {14276, 2, {0x1084, 0x1044, 0x1048, 0x1000, 0x55FE, 0x5420, 0x5420, 0x54FC, 0x5420, 0x5420, 0x55FE, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {14277, 2, {0x1000, 0x1088, 0x1050, 0x1124, 0x5554, 0x558C, 0x5504, 0x55FC, 0x5400, 0x5490, 0x5490, 0x5C90, 0x6510, 0x112, 0x212, 0x40E}},
    {14278, 2, {0x2090, 0x2094, 0x2114, 0x2118, 0xAB10, 0xAD32, 0xA952, 0xA90E, 0xA920, 0xA820, 0xABFE, 0xB820, 0xC820, 0x20, 0x20, 0x20}},
    {14279, 2, {0x1020, 0x1020, 0x1124, 0x1124, 0x5524, 0x5524, 0x56AA, 0x5472, 0x5420, 0x5420, 0x55FC, 0x5C20, 0x6420, 0x20, 0x3FE, 0x0}},
    {14280, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0x55FE, 0x5420, 0x5520, 0x54B2, 0x54B4, 0x5468, 0x54A8, 0x5D24, 0x6222, 0x20, 0xA0, 0x40}},
    {14281, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xAC24, 0xA920, 0xA9FC, 0xA920, 0xAA20, 0xA820, 0xABFE, 0xB820, 0xC820, 0x20, 0x20, 0x20}},
    {14282, 2, {0x1008, 0x103C, 0x11C0, 0x1004, 0x5544, 0x54A8, 0x5400, 0x55F8, 0x5410, 0x5420, 0x57FE, 0x5C20, 0x6420, 0x20, 0xA0, 0x40}},
    {14283, 2, {0x1104, 0x1084, 0x1088, 0x1010, 0x55FC, 0x5504, 0x5504, 0x5504, 0x55FC, 0x5450, 0x5450, 0x5C90, 0x6492, 0x112, 0x20E, 0x400}},
    {14284, 2, {0x1090, 0x1088, 0x1080, 0x13FE, 0x54A0, 0x54A8, 0x54B0, 0x54A4, 0x54A8, 0x54B0, 0x5524, 0x5D2A, 0x6532, 0x222, 0x2DE, 0x400}},
    {14285, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0x5504, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5520, 0x5522, 0x5D14, 0x6508, 0x144, 0x182, 0x100}},
    {14286, 2, {0x2008, 0x203C, 0x23C0, 0x2044, 0xAA24, 0xA928, 0xA900, 0xA840, 0xABFE, 0xA888, 0xA908, 0xBB90, 0xC860, 0x50, 0x188, 0x604}},
    {14287, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0x5544, 0x5622, 0x54F8, 0x5408, 0x5410, 0x5420, 0x55FC, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {14288, 2, {0x1048, 0x1048, 0x11FE, 0x1048, 0x5448, 0x5420, 0x5410, 0x55FE, 0x5480, 0x5480, 0x5480, 0x5C80, 0x6480, 0x80, 0xFC, 0x0}},
    {14289, 2, {0x2020, 0x2020, 0x23FE, 0x2020, 0xA820, 0xABFE, 0xAA02, 0xAC04, 0xA9F8, 0xA810, 0xA820, 0xBBFE, 0xC820, 0x20, 0xA0, 0x40}},
    {14290, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x408, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {14291, 2, {0x2184, 0x2068, 0x2030, 0x20C8, 0xAB04, 0xA840, 0xABFE, 0xA8A0, 0xA920, 0xABFC, 0xAD24, 0xB924, 0xC934, 0x128, 0x20, 0x20}},
    {14292, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0xC60, 0x3118, 0xCFE6, 0x40, 0x80, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {14293, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x20, 0x7920, 0x48A0, 0x4A20, 0x4920, 0x483C, 0x4BE0, 0x7820, 0x4820, 0x20, 0x20}},
    {14294, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {14295, 2, {0x100, 0x100, 0x4104, 0x5144, 0x5554, 0x638C, 0x4104, 0x7DF4, 0x5554, 0x5554, 0x5554, 0x6594, 0x4D34, 0x4104, 0x7FFC, 0x4}},
    {14296, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x5574, 0x5524, 0x55FC, 0x5504, 0x5574, 0x5554, 0x5554, 0x5D74, 0x6504, 0x204, 0x214, 0x408}},
    {14297, 2, {0x1088, 0x1088, 0x1088, 0x13FE, 0x5488, 0x5488, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x5504, 0x5DFC, 0x6504, 0x104, 0x1FC, 0x104}},
    {14298, 2, {0x1020, 0x1020, 0xFDFC, 0x1020, 0x3020, 0x39F8, 0x5488, 0x9050, 0x1020, 0x1050, 0x118C, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14299, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1088, 0x1088, 0xFBFE, 0x1088, 0x1088, 0x1CF8, 0xF088, 0x1088, 0x1088, 0x50F8, 0x2088}},
    {14300, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x40, 0x7C40, 0x4FC, 0x484, 0x7D28, 0x4020, 0x40A8, 0x7CA4, 0x4A4, 0x524, 0x28A0, 0x1040}},
    {14301, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0x5508, 0x5490, 0x5400, 0x57FE, 0x5400, 0x5400, 0x55FC, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {14302, 2, {0x1048, 0x1044, 0x105E, 0x11E0, 0x5428, 0x5412, 0x546A, 0x5596, 0x5448, 0x545E, 0x55E0, 0x5C24, 0x6428, 0x12, 0x6A, 0x186}},
    {14303, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x820, 0x7FFC, 0x820, 0xFFFE, 0x1010, 0x2FC8, 0x4844, 0x8FC2, 0x810, 0x810, 0x7F0}},
    {14304, 2, {0x2000, 0x23FE, 0x2202, 0x228A, 0xAA52, 0xABFE, 0xAA22, 0xAA22, 0xAAAA, 0xAAAA, 0xAAAA, 0xBAFA, 0xCA02, 0x202, 0x20A, 0x204}},
    {14305, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0x8A0, 0x848, 0x7F8}},
    {14306, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x3FF8, 0x480, 0x840, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x50, 0x20}},
    {14307, 2, {0x2104, 0x2114, 0x2114, 0x2114, 0xABD2, 0xA922, 0xA92A, 0xAB48, 0xAB88, 0xAD50, 0xAD10, 0xB914, 0xC922, 0x17E, 0x122, 0x100}},
    {14308, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x1110, 0x9A0, 0x540, 0x1930, 0xE50E, 0x200}},
    {14309, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x400, 0x820, 0x1FF0, 0x210, 0xFFFE, 0x8A0, 0x3318, 0xCC46, 0x380, 0x1C30, 0x1C0, 0x3E00}},
    {14310, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x7FFC, 0x4104, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {14311, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x3EF8, 0x2288, 0x3EF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x800, 0xFF0, 0x10, 0xA0, 0x40}},
    {14312, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x55FC, 0x5400, 0x57FE, 0x5420, 0x5440, 0x55FC, 0x5554, 0x5D54, 0x6554, 0x154, 0x154, 0x10C}},
    {14313, 2, {0x20, 0x7F20, 0x413E, 0x7F44, 0x48A4, 0x7F24, 0x4828, 0x4510, 0x5328, 0x6146, 0x0, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14314, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x5400, 0x57FE, 0x5520, 0x5D22, 0x6514, 0x148, 0x184, 0x102}},
    {14315, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0x55FC, 0x5504, 0x55FC, 0x5420, 0x5432, 0x57B4, 0x54A8, 0x5D28, 0x6524, 0x224, 0x4A2, 0x40}},
    {14316, 2, {0x100, 0x4104, 0x7FFC, 0x500, 0x3978, 0x2108, 0x3D78, 0x2108, 0x3FF8, 0x100, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {14317, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x5510, 0x55FE, 0x5510, 0x554A, 0x5586, 0x5400, 0x55FC, 0x5D04, 0x65FC, 0x104, 0x1FC, 0x104}},
    {14318, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x5488, 0x5504, 0x5622, 0x5428, 0x5424, 0x57FE, 0x5420, 0x5C50, 0x6450, 0x88, 0x104, 0x202}},
    {14319, 2, {0x100, 0x4104, 0x7FFC, 0x200, 0x100, 0x7FFE, 0x4802, 0x9FF4, 0x2820, 0x7C0, 0x1830, 0xE00E, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {14320, 2, {0x2000, 0x23FE, 0x2202, 0x2000, 0xA9FC, 0xA904, 0xA9FC, 0xA904, 0xA9FC, 0xA840, 0xA820, 0xBBFE, 0xC800, 0x88, 0x104, 0x202}},
    {14321, 2, {0x2104, 0x2084, 0x2088, 0x2000, 0xABFE, 0xA820, 0xA820, 0xA9FC, 0xA820, 0xA820, 0xABFE, 0xB800, 0xCAA4, 0x252, 0x452, 0x0}},
    {14322, 2, {0x1000, 0x11FE, 0x1110, 0x1120, 0x557C, 0x5544, 0x557C, 0x5544, 0x557C, 0x5510, 0x5510, 0x5D54, 0x6552, 0x292, 0x250, 0x420}},
    {14323, 2, {0x2000, 0x23FE, 0x2050, 0x2050, 0xABFE, 0xAA52, 0xAA52, 0xABFE, 0xA820, 0xA820, 0xABFE, 0xB870, 0xC8A8, 0x124, 0x622, 0x20}},
    {14324, 2, {0x2020, 0x2040, 0x23FC, 0x2224, 0xAA24, 0xABFC, 0xAA24, 0xAA44, 0xABFC, 0xA840, 0xA8A8, 0xB8B4, 0xC93C, 0x122, 0x222, 0x41E}},
    {14325, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {14326, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x3FFC, 0x2080, 0x2100, 0x27F0, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x4080, 0x4490, 0x8A88, 0x1104}},
    {14327, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x5420, 0x5450, 0x5488, 0x5504, 0x56FA, 0x5400, 0x5400, 0x5DFC, 0x6504, 0x104, 0x1FC, 0x104}},
    {14328, 2, {0x2020, 0x2040, 0x21FC, 0x2104, 0xA9FC, 0xA904, 0xA9FC, 0xA840, 0xABFE, 0xA888, 0xA924, 0xBA22, 0xCDFC, 0x20, 0x20, 0x20}},
    {14329, 2, {0x2010, 0x3F20, 0x2ACE, 0x2AAA, 0xAEAA, 0xAAAA, 0xAAAA, 0xAEAA, 0xAAAA, 0xAAAA, 0xABEA, 0xBEAC, 0xDA28, 0x248, 0x248, 0x288}},
    {14330, 2, {0x100, 0x4104, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0xFFFE, 0x2108, 0x3FF8, 0x1110, 0x1FF0, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {14331, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {14332, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x820, 0x820, 0x7F3E, 0x844, 0x3EA4, 0x824, 0xFF28, 0x1028, 0x1E10, 0x2228, 0x4A44, 0x8482}},
    {14333, 2, {0x1000, 0x13FE, 0x1050, 0x11FC, 0x5554, 0x5554, 0x55FC, 0x5400, 0x55FC, 0x5400, 0x57FE, 0x5C20, 0x64A8, 0x124, 0x2A2, 0x40}},
    {14334, 2, {0x1010, 0x11FE, 0x1110, 0x117C, 0x5514, 0x55FE, 0x5514, 0x557C, 0x5510, 0x557C, 0x5554, 0x5D7C, 0x6554, 0x27C, 0x254, 0x44C}},
    {14335, 2, {0x2088, 0x2088, 0x23FE, 0x2088, 0xA824, 0xABFE, 0xA850, 0xA888, 0xA904, 0xAA8A, 0xA888, 0xBBFE, 0xC888, 0x88, 0x108, 0x208}},
    {14336, 2, {0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4924, 0x849C, 0x100, 0x2108, 0x2108, 0x3FF8}},
    {14337, 2, {0x2040, 0x2088, 0x21FC, 0x2108, 0xAA52, 0xABFE, 0xA850, 0xA988, 0xAE26, 0xA8C0, 0xAB10, 0xB864, 0xCB88, 0x30, 0xC0, 0x700}},
    {14338, 2, {0x2108, 0x2108, 0x2108, 0x27D0, 0xA91E, 0xAF94, 0xA924, 0xAFD4, 0xAA14, 0xABD4, 0xAA54, 0xBA48, 0xCA48, 0x4D4, 0x424, 0x842}},
    {14339, 2, {0x2040, 0x2020, 0x23FE, 0x2088, 0xA850, 0xABFE, 0xAA22, 0xAAFA, 0xAA22, 0xAAFA, 0xAA8A, 0xBA8A, 0xCAFA, 0x202, 0x20A, 0x204}},
    {14340, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {14341, 2, {0x100, 0x4104, 0x7FFC, 0x200, 0x13FC, 0x2400, 0x43F8, 0x8A08, 0x13F8, 0x3208, 0x53F8, 0x9100, 0x13F8, 0x1510, 0x10E0, 0x171C}},
    {14342, 2, {0x2000, 0x23DE, 0x2252, 0x2252, 0xABDE, 0xA800, 0xA9FC, 0xA924, 0xA9FC, 0xA924, 0xA9FC, 0xB820, 0xCBFE, 0x20, 0x20, 0x20}},
    {14343, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x2050, 0x1048, 0xFE48, 0x1FE, 0x7C50, 0x4450, 0x7C50, 0x1050, 0x5450, 0x5292, 0x9292, 0x310E}},
    {14344, 2, {0x2000, 0x23DE, 0x2252, 0x23DE, 0xAA52, 0xABDE, 0xAA02, 0xAAFA, 0xAA22, 0xAA72, 0xAA22, 0xBA22, 0xCAFA, 0x202, 0x20A, 0x204}},
    {14345, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0x55FE, 0x5400, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x55FC, 0x5C20, 0x65FC, 0x20, 0x3FE, 0x0}},
    {14346, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0xAA42, 0xAA78, 0xABC4, 0xAA3C, 0xAA00, 0xAA50, 0xAA54, 0xBB54, 0xC4D8, 0x450, 0x9FE, 0x0}},
    {14347, 2, {0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x7C7C, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1000, 0x1FF8, 0x8, 0x70}},
    {14348, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x20, 0x2220, 0xFFBE, 0x2242, 0x3E94, 0x2210, 0x3E10, 0x2210, 0xFFA8, 0x2428, 0x4244, 0x8282}},
    {14349, 2, {0x100, 0x4104, 0x7FFC, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {14350, 2, {0x20A0, 0x2090, 0x21FE, 0x2320, 0xA9FC, 0xA920, 0xA9FC, 0xA920, 0xA9FE, 0xA900, 0xA820, 0xBBFE, 0xC870, 0xA8, 0x326, 0x20}},
    {14351, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0x55FC, 0x54A8, 0x54A8, 0x57FE, 0x54A8, 0x54A8, 0x55FC, 0x5C20, 0x63FE, 0x20, 0x20, 0x20}},
    {14352, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x70, 0x1F80, 0x200, 0xFFFE, 0x820, 0x37D8, 0xC446, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27D8}},
    {14353, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2448, 0x3FF8, 0x280, 0x7EFC, 0x280, 0x3EF8, 0x280, 0x7EFC, 0x280, 0x280}},
    {14354, 2, {0x2002, 0x2FE2, 0x2282, 0x2FEA, 0xAAAA, 0xAAAA, 0xAFEA, 0xA80A, 0xABCA, 0xA80A, 0xAFEA, 0xB90A, 0xC542, 0x522, 0x92A, 0x304}},
    {14355, 2, {0x2088, 0x23FE, 0x2088, 0x23FE, 0xAA02, 0xA9FC, 0xA800, 0xABFE, 0xA840, 0xA8A2, 0xAB54, 0xB8B8, 0xCB54, 0x92, 0x350, 0x20}},
    {14356, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x40, 0x3E20, 0x22FC, 0x2288, 0x3E50, 0x21FE, 0x2020, 0x3E20, 0x52FC, 0x5220, 0x9E20, 0x1220}},
    {14357, 2, {0x100, 0x4104, 0x7FFC, 0x40, 0x7880, 0x4BF8, 0x5248, 0x63F8, 0x5248, 0x4BF8, 0x4848, 0x68B4, 0x50BC, 0x4122, 0x4222, 0x441E}},
    {14358, 2, {0x4104, 0x7FFC, 0x820, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x240, 0x3C50, 0x848, 0xFFFE, 0x850, 0x7E24, 0x854, 0x198C}},
    {14359, 2, {0x2040, 0x2080, 0x23FE, 0x2222, 0xAAAA, 0xAA22, 0xABFE, 0xAA72, 0xAAAA, 0xAA02, 0xA820, 0xBBFE, 0xC850, 0x88, 0x104, 0x602}},
    {14360, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0xA00, 0x33B8, 0x2208, 0x3A38, 0x2388, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {14361, 2, {0x2020, 0x23FE, 0x2000, 0x21FC, 0xA904, 0xA9FC, 0xA800, 0xABFE, 0xAA02, 0xA9FC, 0xA840, 0xBBA4, 0xC8D8, 0x334, 0xD2, 0x330}},
    {14362, 2, {0x2144, 0x2144, 0x2554, 0x2364, 0xA94E, 0xABE4, 0xA804, 0xAA2C, 0xA944, 0xABE4, 0xA884, 0xBBE4, 0xC884, 0xE4, 0x714, 0x8}},
    {14363, 2, {0x1088, 0x13FE, 0x10A8, 0x1090, 0x55FE, 0x5720, 0x55FC, 0x5520, 0x55FC, 0x5520, 0x55FE, 0x5D00, 0x63FC, 0x88, 0x70, 0x38E}},
    {14364, 2, {0x2200, 0x22BE, 0x2302, 0x2254, 0xA9C8, 0xA83E, 0xAA0A, 0xABCA, 0xAD28, 0xA928, 0xAFEE, 0xB928, 0xCAA8, 0x258, 0x44E, 0x80}},
    {14365, 2, {0x1088, 0x11DC, 0x1088, 0x13DE, 0x5488, 0x5554, 0x5622, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5DFC, 0x6504, 0x1FC, 0x88, 0x104}},
    {14366, 2, {0x2220, 0x4AA0, 0x8ABE, 0x1FC4, 0x20A4, 0x6F24, 0xA028, 0x2FA8, 0x2210, 0x2728, 0x2244, 0x2F82, 0x100, 0x2108, 0x2108, 0x3FF8}},
    {14367, 2, {0x1050, 0x1190, 0x109E, 0x13EA, 0x548A, 0x55D2, 0x568A, 0x54A4, 0x5450, 0x5488, 0x5726, 0x5CA8, 0x6470, 0xA8, 0x124, 0x60}},
    {14368, 2, {0x2040, 0x207C, 0x2040, 0x23FC, 0xAA44, 0xABF0, 0xAA44, 0xAAFC, 0xAAA8, 0xAAF8, 0xAAA8, 0xBAF8, 0xCA00, 0x5FC, 0x554, 0xBFE}},
    {14369, 2, {0x2020, 0x23FE, 0x2000, 0x23FC, 0xAA20, 0xABF8, 0xAA20, 0xABF8, 0xAA20, 0xABFE, 0xA802, 0xBAAA, 0xC854, 0x188, 0x6A4, 0xC2}},
    {14370, 2, {0x2108, 0x3FF8, 0x200, 0x1FF0, 0x1110, 0x1FF0, 0x1210, 0x1FF0, 0x4A8, 0x8BA, 0x3082, 0xC27E, 0x100, 0x7FFC, 0xC48, 0x763E}},
    {14371, 2, {0x2000, 0x23FE, 0x2222, 0x22CC, 0xAA44, 0xABEE, 0xAA44, 0xAAEE, 0xAB54, 0xAA44, 0xAA10, 0xBA90, 0xCA9E, 0x490, 0x490, 0xBFE}},
    {14372, 2, {0x2020, 0x23FE, 0x2000, 0x21DC, 0xA954, 0xA9DC, 0xA888, 0xABFE, 0xA888, 0xABFE, 0xA888, 0xBBFE, 0xC894, 0x188, 0x2A4, 0xC2}},
    {14373, 2, {0x27FC, 0x2444, 0x27FC, 0x2444, 0xAFFC, 0xA800, 0xAFBE, 0xAAAA, 0xAFBE, 0xAAAA, 0xAFBE, 0xB840, 0xC840, 0x7FC, 0x40, 0xFFE}},
    {14374, 2, {0x20A8, 0x21FC, 0x22AA, 0x23FE, 0xAAAA, 0xABFE, 0xAAAA, 0xABFE, 0xA800, 0xABFE, 0xA800, 0xB9FC, 0xC904, 0x1FC, 0x88, 0x3FE}},
    {14375, 2, {0x2020, 0x23FE, 0x2248, 0x23FE, 0xAA48, 0xAAEC, 0xAB5A, 0xAA48, 0xAA50, 0xABDE, 0xAA50, 0xBBDC, 0xCA50, 0x5DE, 0x450, 0x850}},
    {14376, 2, {0x2154, 0x2124, 0x23DE, 0x2154, 0xABAE, 0xAD54, 0xAFFE, 0xAC92, 0xAFFE, 0xAA44, 0xABDE, 0xBA44, 0xCBD4, 0x20C, 0x284, 0x34C}},
    {14377, 2, {0x80, 0x80, 0x2140, 0x2220, 0x2410, 0x2000, 0x3FFC, 0x0, 0x920, 0x920, 0x920, 0x1120, 0x1122, 0x2122, 0x411E, 0x8000}},
    {14378, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x1F8, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {14379, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4400, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0x600C}},
    {14380, 2, {0x0, 0x7CF8, 0x1088, 0x1088, 0x1088, 0x1CF8, 0xE000, 0x4100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {14381, 2, {0x0, 0x7C7C, 0x1010, 0x1010, 0x1010, 0x1010, 0xFEFE, 0x0, 0x0, 0x7C7C, 0x1010, 0x1010, 0x1010, 0x1E10, 0xF0FE, 0x4000}},
    {14382, 2, {0x2200, 0x1400, 0x7F7C, 0x1410, 0x5510, 0x3610, 0xFF90, 0x10, 0x3E10, 0x2210, 0x2210, 0x3E10, 0x2210, 0x2210, 0x3EFE, 0x2200}},
    {14383, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x0}},
    {14384, 2, {0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x227C, 0x4A44, 0x8400, 0x1FF0, 0x10, 0x10, 0x1FF0, 0x1000, 0x1004, 0x1004, 0xFFC}},
    {14385, 2, {0x3FF0, 0x10, 0x3FF0, 0x2004, 0x1FFC, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {14386, 2, {0x840, 0x840, 0x840, 0x7F42, 0x4944, 0x4948, 0x4950, 0x4960, 0x4940, 0x4940, 0x4940, 0x4B42, 0x842, 0x842, 0x83E, 0x800}},
    {14387, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {14388, 2, {0x1000, 0x1000, 0x11FC, 0x7C24, 0x5424, 0x54A4, 0x54A4, 0x54A4, 0x5524, 0x5424, 0x5424, 0x5C44, 0x1044, 0x1084, 0x1114, 0x1208}},
    {14389, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {14390, 2, {0x1020, 0x1020, 0x1040, 0x7CF8, 0x5488, 0x5488, 0x5488, 0x5488, 0x5488, 0x5488, 0x5488, 0x5C8A, 0x108A, 0x110A, 0x1106, 0x1200}},
    {14391, 2, {0x1000, 0x13FC, 0x1084, 0x7C88, 0x5488, 0x5490, 0x549C, 0x5484, 0x5544, 0x5544, 0x5528, 0x5D28, 0x1210, 0x1228, 0x1444, 0x1182}},
    {14392, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x55FC, 0x5420, 0x5420, 0x5420, 0x57FE, 0x5420, 0x5450, 0x5C50, 0x1088, 0x1088, 0x1104, 0x1202}},
    {14393, 2, {0x1020, 0x1020, 0x1050, 0x7C50, 0x5488, 0x5504, 0x5602, 0x5488, 0x5488, 0x5488, 0x5488, 0x5C88, 0x1088, 0x1108, 0x1108, 0x1208}},
    {14394, 2, {0x1020, 0x1020, 0x1020, 0x7CA8, 0x54A4, 0x54A2, 0x5522, 0x5520, 0x5624, 0x5424, 0x5428, 0x5C08, 0x1010, 0x1020, 0x10C0, 0x1300}},
    {14395, 2, {0x200, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x100, 0x280, 0xC80, 0x1FF8, 0x2888, 0x4888, 0x88A8, 0x890, 0x80, 0x80}},
    {14396, 2, {0x1004, 0x100E, 0x10F0, 0x7C80, 0x5480, 0x5480, 0x54FE, 0x5488, 0x5488, 0x5488, 0x5488, 0x5C88, 0x1088, 0x1108, 0x1108, 0x1208}},
    {14397, 2, {0x1040, 0x1040, 0x107C, 0x7C84, 0x5588, 0x5650, 0x5420, 0x5450, 0x5488, 0x5706, 0x5460, 0x5C10, 0x1008, 0x10C0, 0x1020, 0x1010}},
    {14398, 2, {0x1000, 0x1EF8, 0x2288, 0x62A8, 0x9492, 0x882, 0x307E, 0xC100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {14399, 2, {0x880, 0x888, 0x2890, 0x2EA0, 0x28C0, 0x2884, 0x2E84, 0xF17C, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {14400, 2, {0x100, 0x2108, 0x1108, 0x910, 0x920, 0x100, 0x7FFC, 0x4104, 0x4104, 0x4384, 0x4544, 0x4924, 0x4104, 0x4104, 0x4114, 0x4008}},
    {14401, 2, {0x1020, 0x1020, 0x13FE, 0x7C50, 0x5488, 0x5504, 0x56FA, 0x5400, 0x57FE, 0x5440, 0x5480, 0x5DFC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {14402, 2, {0x4, 0x7FA4, 0x1024, 0x3F24, 0x5124, 0xA04, 0xC04, 0x3114, 0xC108, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {14403, 2, {0x1010, 0x1110, 0x1110, 0x7D28, 0x5544, 0x5500, 0x55FE, 0x5400, 0x54A8, 0x54A8, 0x54A8, 0x5CA8, 0x10A8, 0x10AA, 0x112A, 0x1206}},
    {14404, 2, {0x1040, 0x1040, 0x1080, 0x7CFE, 0x5502, 0x5602, 0x55F2, 0x5512, 0x5512, 0x55F2, 0x5512, 0x5D12, 0x11F2, 0x1002, 0x1014, 0x1008}},
    {14405, 2, {0x1020, 0x1020, 0x107C, 0x7C84, 0x5548, 0x5430, 0x5420, 0x5448, 0x5590, 0x543E, 0x5442, 0x5DA4, 0x1018, 0x1010, 0x1060, 0x1180}},
    {14406, 2, {0x1080, 0x1040, 0x13FC, 0x7D00, 0x5500, 0x5500, 0x55F8, 0x5400, 0x5550, 0x5550, 0x5550, 0x5D50, 0x1250, 0x1252, 0x1452, 0x180E}},
    {14407, 2, {0x1000, 0x13FE, 0x1120, 0x7D28, 0x5528, 0x55E8, 0x5528, 0x5528, 0x5528, 0x55E8, 0x5528, 0x5D2A, 0x113A, 0x13EA, 0x1026, 0x1020}},
    {14408, 2, {0x1080, 0x1080, 0x10F8, 0x7D08, 0x5610, 0x55FC, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5450, 0x5C50, 0x1090, 0x1092, 0x1112, 0x120E}},
    {14409, 2, {0x2010, 0x2388, 0x2088, 0xF8FE, 0xA890, 0xAB90, 0xAA24, 0xAA24, 0xAA38, 0xAB88, 0xA890, 0xB890, 0x20A4, 0x20BE, 0x2282, 0x2100}},
    {14410, 2, {0x1040, 0x1040, 0x107C, 0x7C84, 0x5508, 0x5600, 0x5420, 0x55CE, 0x5502, 0x5502, 0x55CE, 0x5D02, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {14411, 2, {0x1020, 0x1020, 0x13FE, 0x7C50, 0x5488, 0x5524, 0x5622, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x5DFC, 0x1022, 0x1022, 0x101E, 0x1000}},
    {14412, 2, {0x1000, 0x11FE, 0x1102, 0x7D7A, 0x5502, 0x557A, 0x5400, 0x54FC, 0x5484, 0x54FC, 0x5484, 0x5CFC, 0x1084, 0x1084, 0x1094, 0x1088}},
    {14413, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0xFFFE, 0x820, 0x3118, 0xDFF6, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {14414, 2, {0x1040, 0x1048, 0x1084, 0x7DFE, 0x5420, 0x57FE, 0x5488, 0x5524, 0x5642, 0x5588, 0x5410, 0x5C62, 0x1184, 0x1018, 0x1060, 0x1380}},
    {14415, 2, {0x2040, 0x20A0, 0x2110, 0xFA08, 0xADF6, 0xA800, 0xABC4, 0xAA54, 0xAA54, 0xABD4, 0xAA54, 0xBA54, 0x23D4, 0x2244, 0x2254, 0x22C8}},
    {14416, 2, {0x2108, 0x2088, 0x2090, 0xFBFE, 0xA800, 0xABC4, 0xAA54, 0xAA54, 0xABD4, 0xAA54, 0xAA54, 0xBBD4, 0x2254, 0x2244, 0x2254, 0x22C8}},
    {14417, 2, {0x610, 0x3810, 0x852, 0x7E54, 0x890, 0x1C28, 0x2A28, 0x4844, 0x882, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100}},
    {14418, 2, {0x1008, 0x101C, 0x11F0, 0x7D10, 0x5510, 0x55FE, 0x5510, 0x557C, 0x5544, 0x557C, 0x5544, 0x5D7C, 0x1144, 0x1244, 0x127C, 0x1444}},
    {14419, 2, {0x2000, 0x23FE, 0x2222, 0xF820, 0xABFE, 0xA820, 0xA9FC, 0xA924, 0xA9FC, 0xA924, 0xA9FC, 0xB820, 0x23FE, 0x2020, 0x2020, 0x2020}},
    {14420, 2, {0x40, 0x7F40, 0x127E, 0xC88, 0xFF48, 0x1948, 0x2A50, 0x4820, 0xA850, 0x118C, 0x100, 0x3FF8, 0x2108, 0x2128, 0x2110, 0x100}},
    {14421, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x54A8, 0x5420, 0x55FC, 0x5524, 0x5524, 0x5524, 0x57FE, 0x5C20, 0x1050, 0x1088, 0x1104, 0x1202}},
    {14422, 2, {0x1008, 0x103C, 0x11E0, 0x7C20, 0x57FE, 0x5420, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x55FC, 0x5C20, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {14423, 2, {0x2004, 0x201E, 0x23F0, 0xFA1E, 0xAA10, 0xAAFE, 0xAA92, 0xAA98, 0xAAF2, 0xAA8E, 0xAA80, 0xBAB8, 0x22A8, 0x24AA, 0x254A, 0x2A86}},
    {14424, 2, {0x1088, 0x1050, 0x1000, 0x7DFE, 0x5450, 0x55FC, 0x5454, 0x57FE, 0x5454, 0x55FC, 0x5450, 0x5CD8, 0x1154, 0x1252, 0x1050, 0x1050}},
    {14425, 2, {0x1020, 0x1040, 0x11FC, 0x7D04, 0x5554, 0x5524, 0x5554, 0x5504, 0x55FC, 0x5400, 0x5512, 0x5DD4, 0x1118, 0x1152, 0x1192, 0x110E}},
    {14426, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x55FC, 0x5488, 0x57FE, 0x5420, 0x55FC, 0x5524, 0x55FC, 0x5D24, 0x13FE, 0x1104, 0x1114, 0x1108}},
    {14427, 2, {0x1088, 0x1088, 0x13FE, 0x7CA8, 0x5410, 0x57FE, 0x5480, 0x5480, 0x54FC, 0x5400, 0x54A8, 0x5CA8, 0x10A8, 0x10AA, 0x112A, 0x1206}},
    {14428, 2, {0x2100, 0x2110, 0x23DC, 0xFA54, 0xAD54, 0xAA88, 0xA908, 0xAAF4, 0xAC02, 0xA800, 0xABFC, 0xB840, 0x2150, 0x2248, 0x2544, 0x2080}},
    {14429, 2, {0x2020, 0x247C, 0x2284, 0xF948, 0xA830, 0xA8CE, 0xAE10, 0xAA7C, 0xAA10, 0xAA7C, 0xAA10, 0xBAFE, 0x2210, 0x2210, 0x25FE, 0x2800}},
    {14430, 2, {0x2020, 0x21FC, 0x2124, 0xFBFE, 0xA924, 0xA9FC, 0xA820, 0xA9FC, 0xA924, 0xA9FC, 0xA840, 0xBBFE, 0x2088, 0x21D0, 0x2070, 0x238C}},
    {14431, 2, {0x2090, 0x2090, 0x27FE, 0xF890, 0xABFC, 0xAA94, 0xABFC, 0xAA94, 0xABFC, 0xA800, 0xA9F8, 0xB908, 0x21F8, 0x2108, 0x21F8, 0x2108}},
    {14432, 2, {0x2020, 0x27A4, 0x20A8, 0xFA92, 0xA914, 0xAA08, 0xADF4, 0xA802, 0xABF8, 0xAA08, 0xAA08, 0xBBF8, 0x2208, 0x2110, 0x2FFE, 0x2000}},
    {14433, 2, {0x2000, 0x23FE, 0x2200, 0xFA04, 0xABF4, 0xAA04, 0xAAEE, 0xAAA4, 0xAAA4, 0xAAEC, 0xAA04, 0xBAA4, 0x2244, 0x2274, 0x2594, 0x2008}},
    {14434, 2, {0x21FC, 0x2104, 0x21FC, 0xF904, 0xA9FC, 0xA800, 0xABFE, 0xAA52, 0xABFE, 0xA800, 0xABFE, 0xB880, 0x20FC, 0x2084, 0x2114, 0x2208}},
    {14435, 2, {0x1020, 0x820, 0x7F3E, 0x20, 0x223E, 0x1402, 0xFFBE, 0x88A0, 0x93E, 0x7F20, 0x493E, 0x4920, 0x4D3E, 0x4A20, 0x822, 0x81E}},
    {14436, 2, {0x2050, 0x2252, 0x2154, 0xF850, 0xABFE, 0xA888, 0xA850, 0xABFE, 0xA820, 0xA9FC, 0xA820, 0xBBFE, 0x20A8, 0x2124, 0x2622, 0x2020}},
    {14437, 2, {0x203E, 0x27C0, 0x2244, 0xF928, 0xABF8, 0xA840, 0xAFFC, 0xA800, 0xABF8, 0xA808, 0xABF8, 0xB808, 0x23F8, 0x2544, 0x252A, 0x28FA}},
    {14438, 2, {0x1020, 0x1010, 0x11FE, 0x7D02, 0x54FC, 0x54A8, 0x5530, 0x54FC, 0x5584, 0x54FC, 0x5484, 0x5CFC, 0x1084, 0x10FC, 0x1048, 0x1084}},
    {14439, 2, {0x2040, 0x2020, 0x23FE, 0xFA8A, 0xA904, 0xA812, 0xABD4, 0xA848, 0xAA8A, 0xA974, 0xAA02, 0xB9FC, 0x2104, 0x21FC, 0x2088, 0x23FE}},
    {14440, 2, {0x2148, 0x214C, 0x22AA, 0xF808, 0xAFFE, 0xA948, 0xAB68, 0xA94A, 0xAB6A, 0xA94C, 0xAB6C, 0xB948, 0x216A, 0x239A, 0x2026, 0x2042}},
    {14441, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x3FF8, 0x2108, 0x2128, 0x2110, 0x100}},
    {14442, 2, {0x203E, 0x27C4, 0x2128, 0xFBFE, 0xAA02, 0xABFE, 0xAAA0, 0xAAEE, 0xAAAA, 0xAAEE, 0xABA8, 0xBA46, 0x22F8, 0x2548, 0x2430, 0x29CE}},
    {14443, 2, {0x1000, 0x11FC, 0x2104, 0x21FC, 0x4504, 0x45FC, 0xF880, 0x9FE, 0x1222, 0x1122, 0x2952, 0x4502, 0xFDFA, 0x4402, 0x14, 0x8}},
    {14444, 2, {0x1110, 0x2120, 0x4544, 0x7978, 0x1110, 0x2524, 0x7D7C, 0x524, 0x110, 0xFFFE, 0x2080, 0x3E88, 0x2252, 0x2222, 0x4A5A, 0x8586}},
    {14445, 2, {0x4410, 0x4924, 0x5E78, 0x4410, 0x4A28, 0x5F7C, 0x4000, 0x7FFE, 0x4410, 0x4924, 0x5E78, 0x4410, 0x4A28, 0x5F7C, 0x4000, 0x7FFE}},
    {14446, 2, {0x4920, 0x4920, 0xAABE, 0xEBA0, 0x4940, 0xAABC, 0xEB80, 0x2900, 0xFFB8, 0x2408, 0x2508, 0x2508, 0x228A, 0x528A, 0x4586, 0x8882}},
    {14447, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2010, 0x20F8, 0x2780, 0x2080, 0x2080, 0x20FE, 0x3F80, 0x2080, 0x4082, 0x4082, 0x8082, 0x7E}},
    {14448, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x27F8, 0x2080, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x2080, 0x4080, 0x4080, 0x8280, 0x100}},
    {14449, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2100, 0x2100, 0x23F0, 0x2210, 0x2420, 0x2820, 0x2040, 0x20C0, 0x4120, 0x4210, 0x8C08, 0x3006}},
    {14450, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2090, 0x2088, 0x20FE, 0x2F80, 0x2080, 0x2088, 0x2090, 0x2060, 0x2042, 0x41A2, 0x4E1A, 0x8006}},
    {14451, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x203C, 0x27C0, 0x2440, 0x2440, 0x2440, 0x27FE, 0x2420, 0x2420, 0x2412, 0x450A, 0x4606, 0x8402}},
    {14452, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2040, 0x2040, 0x2FFC, 0x2000, 0x2408, 0x2208, 0x2210, 0x2110, 0x4120, 0x4000, 0x9FFE, 0x0}},
    {14453, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2140, 0x2220, 0x2490, 0x2848, 0x3046, 0x27F0, 0x2010, 0x2020, 0x4340, 0x4080, 0x8040, 0x20}},
    {14454, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x27F8, 0x2408, 0x2408, 0x2408, 0x2408, 0x27F8, 0x2408, 0x4120, 0x4210, 0x8408, 0x804}},
    {14455, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2FFC, 0x21C0, 0x22A0, 0x22A0, 0x2490, 0x2888, 0x57F4, 0x4080, 0x8080, 0x80}},
    {14456, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2200, 0x23FE, 0x2500, 0x2900, 0x31F8, 0x2100, 0x2100, 0x21FC, 0x4100, 0x4100, 0x8100, 0x100}},
    {14457, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x4208, 0x4208, 0x8FFE, 0x0}},
    {14458, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x4000, 0x4000, 0x9FFE, 0x0}},
    {14459, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2180, 0x263C, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x25A4, 0x26B4, 0x2528, 0x4120, 0x4220, 0x8420}},
    {14460, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x2084, 0x2048, 0x2030, 0x41D4, 0x5E0C, 0x8004}},
    {14461, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2040, 0x27F8, 0x2048, 0x3FFE, 0x2048, 0x27F8, 0x2040, 0x27FC, 0x4040, 0x5FFE, 0x8040, 0x40}},
    {14462, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2484, 0x2488, 0x2450, 0x2420, 0x4518, 0x4606, 0x8400}},
    {14463, 2, {0x100, 0x80, 0x3FFE, 0x2208, 0x2108, 0x2110, 0x2FFE, 0x2040, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2084, 0x4100, 0x4600, 0x9800}},
    {14464, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2040, 0x2FFE, 0x2040, 0x27FC, 0x2444, 0x2454, 0x24E8, 0x2150, 0x4248, 0x4C46, 0x8040, 0x40}},
    {14465, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x21F0, 0x2210, 0x2D20, 0x20D0, 0x2320, 0x2C7C, 0x2184, 0x2688, 0x4050, 0x4060, 0x8180, 0xE00}},
    {14466, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2080, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x2400, 0x2400, 0x27FC, 0x4404, 0x4404, 0x87FC, 0x404}},
    {14467, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2080, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2482, 0x2444, 0x2428, 0x4510, 0x4608, 0x8406}},
    {14468, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2FFC, 0x2000, 0x23F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x2000, 0x4208, 0x4110, 0x9FFE}},
    {14469, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27FC, 0x2040, 0x23F8, 0x2040, 0x2FFE, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x4210, 0x4210, 0x8410}},
    {14470, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2040, 0x27FC, 0x4040, 0x4040, 0x8FFE}},
    {14471, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2080, 0x2FFC, 0x2000, 0x27F8, 0x2000, 0x27F8, 0x2000, 0x27F8, 0x2408, 0x4408, 0x47F8, 0x8408}},
    {14472, 2, {0x100, 0x80, 0x3FFE, 0x2280, 0x2240, 0x27FE, 0x2440, 0x2C40, 0x37FC, 0x2440, 0x2440, 0x27FC, 0x4440, 0x4440, 0x87FE, 0x400}},
    {14473, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x218C, 0x2E70, 0x2210, 0x2210, 0x3F90, 0x261E, 0x2770, 0x2A90, 0x2A12, 0x5212, 0x4212, 0x820E}},
    {14474, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2140, 0x2220, 0x2410, 0x3FFE, 0x2010, 0x27D0, 0x2450, 0x47D0, 0x4010, 0x8050, 0x20}},
    {14475, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2278, 0x2488, 0x2550, 0x2C24, 0x3448, 0x259E, 0x2422, 0x2464, 0x2598, 0x4410, 0x4460, 0x8580}},
    {14476, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2EFC, 0x2A08, 0x2AE8, 0x2CA8, 0x2AA8, 0x2AA8, 0x2AA8, 0x2AE8, 0x4EA8, 0x4808, 0x8828, 0x810}},
    {14477, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2110, 0x2FFE, 0x2110, 0x2040, 0x23F8, 0x2248, 0x2248, 0x2FFE, 0x20A0, 0x4110, 0x4608, 0x9806}},
    {14478, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2038, 0x27C0, 0x2040, 0x2FFE, 0x2040, 0x2140, 0x265C, 0x2444, 0x275C, 0x4444, 0x47FC, 0x8404}},
    {14479, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x20A0, 0x2318, 0x2DF6, 0x2000, 0x2784, 0x24A4, 0x27A4, 0x24A4, 0x27A4, 0x4484, 0x4494, 0x8588}},
    {14480, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x23F0, 0x2210, 0x23D0, 0x2250, 0x2250, 0x2FFC, 0x2804, 0x29E4, 0x2924, 0x49E4, 0x4814, 0x8808}},
    {14481, 2, {0x80, 0x3FFE, 0x2280, 0x2240, 0x27FC, 0x2440, 0x2FF8, 0x3440, 0x27F8, 0x2440, 0x27FC, 0x2480, 0x4080, 0x5FFE, 0x8080, 0x80}},
    {14482, 2, {0x80, 0x3FFE, 0x2040, 0x207E, 0x2040, 0x2FFE, 0x2842, 0x2BF8, 0x2842, 0x29FE, 0x2904, 0x29FC, 0x4904, 0x51FC, 0x9104, 0x27FE}},
    {14483, 2, {0x80, 0x3FFE, 0x2040, 0x27FC, 0x2110, 0x2FFE, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x4040, 0x4FFE, 0x8040, 0x40}},
    {14484, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2448, 0x2848, 0x3054, 0x22A4, 0x2500, 0x2C10, 0x3490, 0x249C, 0x4490, 0x4550, 0x863E, 0x400}},
    {14485, 2, {0x100, 0x80, 0x3FFE, 0x2228, 0x2224, 0x257E, 0x2948, 0x2E48, 0x22FE, 0x2548, 0x2948, 0x2F7E, 0x4248, 0x4448, 0x887E, 0x1040}},
    {14486, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2078, 0x2F80, 0x2490, 0x22A0, 0x3FFE, 0x2490, 0x2808, 0x37F6, 0x4490, 0x47F0, 0x8490, 0x7F0}},
    {14487, 2, {0x80, 0x3FFE, 0x2038, 0x27C0, 0x2080, 0x2FFC, 0x2110, 0x27F8, 0x3916, 0x21F0, 0x2000, 0x2FFE, 0x4802, 0x49F2, 0x8912, 0x9F6}},
    {14488, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2120, 0x2FFC, 0x2120, 0x2FF8, 0x2128, 0x3FFE, 0x2128, 0x2FF8, 0x2000, 0x4A48, 0x4924, 0x9124}},
    {14489, 2, {0x80, 0x3FFE, 0x2510, 0x2510, 0x3FD0, 0x253C, 0x2714, 0x2214, 0x2F94, 0x2A94, 0x2F94, 0x2214, 0x4FA4, 0x4224, 0x8354, 0x1C88}},
    {14490, 2, {0x80, 0x3FFE, 0x2080, 0x20F8, 0x2080, 0x2FFC, 0x2884, 0x2BF0, 0x2888, 0x2BF8, 0x2A48, 0x2BF8, 0x4A48, 0x57FC, 0x9084, 0x230C}},
    {14491, 2, {0x80, 0x3FFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2DF6, 0x2000, 0x27BC, 0x24A4, 0x24A4, 0x27BC, 0x4110, 0x42A8, 0x8444, 0x882}},
    {14492, 2, {0x80, 0x3FFE, 0x2208, 0x3FFE, 0x2208, 0x23F8, 0x2000, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x4040, 0x4524, 0x850A, 0x8FA}},
    {14493, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2706, 0x2938, 0x3208, 0x2F88, 0x2ABE, 0x2F88, 0x2A9C, 0x2FAA, 0x4008, 0x5548, 0x9548, 0x2008}},
    {14494, 2, {0x80, 0x7FFE, 0x4000, 0x7E7C, 0x5224, 0x4A14, 0x5224, 0x4FF8, 0x4888, 0x4FF8, 0x4888, 0x4FF8, 0x4220, 0x5FFC, 0x4410, 0x8808}},
    {14495, 2, {0x0, 0x20, 0xF820, 0x820, 0x1020, 0x113C, 0x2120, 0x7920, 0x920, 0x920, 0x4920, 0x31FC, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {14496, 2, {0x0, 0x3FC, 0xFA04, 0xA04, 0x12F4, 0x1294, 0x2294, 0x7A94, 0xAF4, 0xA94, 0x4A04, 0x3214, 0x1208, 0x2C00, 0x43FE, 0x8000}},
    {14497, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8418, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {14498, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x41A8, 0x5D28, 0x5510, 0x5D28, 0x4246, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {14499, 2, {0x2420, 0xFF20, 0x247E, 0x7EC4, 0x8228, 0x7A10, 0x4A28, 0x7AC6, 0x400, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {14500, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x1040, 0x1040, 0x1F40, 0x1140, 0x1120, 0x2122, 0x2112, 0x410A, 0x4A06, 0x8402}},
    {14501, 2, {0x20, 0x7F28, 0x4124, 0x4124, 0x5D20, 0x413E, 0x41E0, 0x5D20, 0x5520, 0x5520, 0x5520, 0x5D24, 0x4114, 0x4114, 0x450C, 0x4204}},
    {14502, 2, {0x20, 0xFF28, 0x224, 0x7A24, 0x4A20, 0x7A2E, 0xF0, 0xFF20, 0x220, 0x7A20, 0x4A20, 0x4A10, 0x7A12, 0x20A, 0xA06, 0x402}},
    {14503, 2, {0x0, 0x7FF0, 0x10, 0x10, 0x3FF0, 0x2000, 0x2000, 0x3FF8, 0x8, 0x8, 0xFFE8, 0x8, 0x8, 0x8, 0x50, 0x20}},
    {14504, 2, {0x0, 0x7DFC, 0x420, 0x420, 0x420, 0x7C20, 0x4020, 0x43FE, 0x4020, 0x7C20, 0x420, 0x420, 0x420, 0x420, 0x2820, 0x1020}},
    {14505, 2, {0x40, 0x7C40, 0x480, 0x4FC, 0x504, 0x7E04, 0x4004, 0x4104, 0x4084, 0x7C44, 0x444, 0x404, 0x404, 0x404, 0x2828, 0x1010}},
    {14506, 2, {0x0, 0xF9FC, 0x800, 0x800, 0x800, 0x7BFE, 0x4080, 0x4080, 0x4100, 0x79FC, 0x804, 0x804, 0x804, 0x804, 0x5028, 0x2010}},
    {14507, 2, {0x20, 0xF820, 0x820, 0x820, 0x924, 0x7924, 0x4124, 0x4124, 0x4124, 0x7924, 0x924, 0x924, 0x924, 0x9FC, 0x5004, 0x2000}},
    {14508, 2, {0x40, 0xF840, 0x840, 0x840, 0xBFE, 0x7840, 0x4080, 0x4090, 0x4090, 0x7920, 0x920, 0xA48, 0xA44, 0xC84, 0x51FE, 0x2082}},
    {14509, 2, {0x20, 0x7C20, 0x420, 0x43E, 0x420, 0x7C20, 0x4020, 0x41FC, 0x4084, 0x7C88, 0x448, 0x450, 0x420, 0x450, 0x2988, 0x1606}},
    {14510, 2, {0x80, 0xF880, 0x8FE, 0x900, 0xA20, 0x7920, 0x412C, 0x4174, 0x43A4, 0x7924, 0x934, 0x928, 0x922, 0x902, 0x50FE, 0x2000}},
    {14511, 2, {0x0, 0xFBFC, 0xA04, 0xA04, 0xA04, 0x7BFC, 0x4220, 0x4220, 0x43FE, 0x7A20, 0xA20, 0xA10, 0xA12, 0xA8A, 0x5306, 0x2202}},
    {14512, 2, {0x10, 0xF810, 0x810, 0x9FE, 0x912, 0x7914, 0x4110, 0x41FC, 0x4144, 0x7944, 0x928, 0x928, 0x910, 0xA28, 0x5244, 0x2482}},
    {14513, 2, {0x20, 0xF840, 0x888, 0x904, 0xBFE, 0x7802, 0x4020, 0x4020, 0x4128, 0x7924, 0xA24, 0xA22, 0xC22, 0x820, 0x50A0, 0x2040}},
    {14514, 2, {0x0, 0x7CF8, 0x408, 0x7CF8, 0x4080, 0x7CF8, 0x408, 0x408, 0x2850, 0x1020, 0x0, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {14515, 2, {0x80, 0x7C80, 0x4FE, 0x500, 0x600, 0x7C1C, 0x41E0, 0x4120, 0x4122, 0x7D24, 0x528, 0x510, 0x510, 0x548, 0x2984, 0x1102}},
    {14516, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x2000, 0x2000, 0x3FFC, 0x4, 0x3FE4, 0x2AA4, 0x24A4, 0x2B24, 0x24A4, 0x3FE8, 0x10}},
    {14517, 2, {0x20, 0xF820, 0x850, 0x888, 0x904, 0x7A02, 0x41FC, 0x4000, 0x4000, 0x79FC, 0x904, 0x904, 0x904, 0x904, 0x51FC, 0x2104}},
    {14518, 2, {0x0, 0xFBFE, 0x820, 0x820, 0x840, 0x79FC, 0x4104, 0x4104, 0x4104, 0x79FC, 0x904, 0x904, 0x904, 0x904, 0x51FC, 0x2104}},
    {14519, 2, {0x10, 0x7C92, 0x492, 0x492, 0x4FE, 0x7C20, 0x4020, 0x407E, 0x4082, 0x7D42, 0x424, 0x428, 0x410, 0x420, 0x2840, 0x1180}},
    {14520, 2, {0x20, 0xF720, 0x1120, 0x113E, 0x1144, 0xF744, 0x8444, 0x84A4, 0x8428, 0xF728, 0x1110, 0x1110, 0x1128, 0x1148, 0xAA84, 0x4502}},
    {14521, 2, {0x40, 0x7C20, 0x5FC, 0x400, 0x488, 0x7C50, 0x43FE, 0x4020, 0x4020, 0x7C20, 0x5FC, 0x420, 0x420, 0x420, 0x2820, 0x1020}},
    {14522, 2, {0x40, 0xF820, 0x9FC, 0x800, 0x888, 0x7850, 0x43FE, 0x4040, 0x4040, 0x7BFE, 0x888, 0x988, 0x850, 0x830, 0x50C8, 0x2304}},
    {14523, 2, {0x1C, 0xF9E0, 0x820, 0x820, 0xBFE, 0x78A8, 0x4124, 0x4242, 0x4040, 0x7BFE, 0x888, 0x908, 0x8D0, 0x830, 0x5048, 0x2184}},
    {14524, 2, {0x20, 0x7D24, 0x4A4, 0x4A8, 0x420, 0x7DFC, 0x4124, 0x4124, 0x4174, 0x7DAC, 0x524, 0x524, 0x524, 0x504, 0x2914, 0x1108}},
    {14525, 2, {0x0, 0xF7EE, 0x1242, 0x1242, 0x1242, 0xF3CE, 0x8248, 0x8248, 0x83C8, 0xF24E, 0x1242, 0x12E2, 0x1742, 0x1042, 0xA054, 0x4048}},
    {14526, 2, {0x0, 0xFBFE, 0x850, 0x850, 0x9FC, 0x7954, 0x4154, 0x4154, 0x41FC, 0x7820, 0x820, 0x9FC, 0x820, 0x820, 0x53FE, 0x2000}},
    {14527, 2, {0x80, 0xF8BC, 0x884, 0x908, 0x9FE, 0x7B20, 0x4520, 0x417C, 0x4190, 0x7910, 0x9FE, 0x910, 0x928, 0x928, 0x5144, 0x2182}},
    {14528, 2, {0x20, 0xF820, 0xBFE, 0x820, 0x9FC, 0x7904, 0x41FC, 0x4104, 0x41FC, 0x7904, 0x9FC, 0x904, 0xBFE, 0x888, 0x5104, 0x2202}},
    {14529, 2, {0x88, 0xF888, 0xBFE, 0x888, 0x888, 0x7BFE, 0x4080, 0x41FC, 0x4324, 0x7DFC, 0x924, 0x9FC, 0x924, 0x924, 0x5124, 0x210C}},
    {14530, 2, {0x0, 0xFBFE, 0x800, 0x9FC, 0x904, 0x7904, 0x41FC, 0x4000, 0x43FE, 0x7A8A, 0xA52, 0xBFE, 0xA22, 0xA22, 0x522A, 0x2204}},
    {14531, 2, {0x200, 0xF11E, 0x17C2, 0x1202, 0x1482, 0xF7DE, 0x8050, 0x8790, 0x8490, 0xF79E, 0x1482, 0x1782, 0x1482, 0x1482, 0xA494, 0x4588}},
    {14532, 2, {0x200, 0xF97E, 0x840, 0xC7C, 0xA44, 0x78FC, 0x4140, 0x467E, 0x4220, 0x7820, 0xBFE, 0x870, 0x8A8, 0x924, 0x5222, 0x2020}},
    {14533, 2, {0x0, 0xFBDE, 0x888, 0xAA8, 0xBFE, 0x7988, 0x42DA, 0x44A6, 0x41F8, 0x7908, 0x908, 0x9F8, 0x908, 0x908, 0x51F8, 0x2108}},
    {14534, 2, {0x7FC, 0xF248, 0x17FC, 0x1C46, 0x16EC, 0xF554, 0x86EC, 0x8554, 0x86EC, 0xF454, 0x1408, 0x17FC, 0x1040, 0x13F8, 0xA040, 0x4FFE}},
    {14535, 2, {0x800, 0xFE0, 0x1020, 0x1FC0, 0x40, 0x7FFC, 0x900, 0x3200, 0xCC80, 0x3080, 0xCFF8, 0x888, 0x888, 0x8A8, 0x890, 0x80}},
    {14536, 2, {0x2020, 0x3C3C, 0x4444, 0x7878, 0x808, 0xFEFE, 0x5050, 0xA8A8, 0x4848, 0xBEBE, 0x2A2A, 0x2A2A, 0x2A2A, 0x2E2E, 0x808, 0x808}},
    {14537, 2, {0x4, 0x4, 0x7F08, 0x810, 0x820, 0x844, 0x804, 0x808, 0x810, 0x822, 0x842, 0xF04, 0xF008, 0x4010, 0x60, 0x180}},
    {14538, 2, {0x400, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE10E, 0x600, 0x3840, 0x188, 0x610, 0x3820, 0xC0, 0x700, 0x7800}},
    {14539, 2, {0x1100, 0x1104, 0x1104, 0x1108, 0x7FD0, 0x1122, 0x1902, 0x1504, 0x1108, 0xFFD0, 0x1122, 0x1102, 0x1104, 0x2108, 0x2110, 0x4160}},
    {14540, 2, {0x800, 0x804, 0x804, 0x808, 0xF90, 0x822, 0x802, 0x804, 0x7F08, 0x4110, 0x4122, 0x4102, 0x4104, 0x4108, 0x7F10, 0x4160}},
    {14541, 2, {0x800, 0x804, 0x1004, 0x2208, 0x4110, 0xFFA2, 0x82, 0x4, 0x7F08, 0x4110, 0x4122, 0x4102, 0x4104, 0x4108, 0x7F10, 0x4160}},
    {14542, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x1150, 0x2188, 0x4644, 0x1990, 0x620, 0x18C0, 0x700, 0x3800}},
    {14543, 2, {0x800, 0x804, 0xFF84, 0x1408, 0x2210, 0x4122, 0xBE82, 0x2204, 0x3E08, 0x2210, 0x3E22, 0x802, 0xFF84, 0x808, 0x810, 0x860}},
    {14544, 2, {0x800, 0x1004, 0x7F04, 0x4108, 0x7F10, 0x4122, 0x7F02, 0x804, 0x808, 0xE890, 0x2D22, 0x2A02, 0x4904, 0x8888, 0x2810, 0x1060}},
    {14545, 2, {0x202, 0x3FC2, 0x2204, 0x2F88, 0x2290, 0x3FC2, 0x2282, 0x2F84, 0x2208, 0x2F90, 0x2A82, 0x2F82, 0x4A84, 0x4F88, 0x8A90, 0x9A0}},
    {14546, 2, {0x4100, 0x2204, 0x4, 0xFF88, 0x8890, 0xAAA2, 0x9C82, 0x8884, 0xFF88, 0x10, 0x7F22, 0x4102, 0x7F04, 0x4108, 0x7F10, 0x4160}},
    {14547, 2, {0x802, 0xDC2, 0x2844, 0x3D48, 0x4490, 0xA942, 0x3242, 0xC404, 0x2488, 0x1510, 0x7FC2, 0xE02, 0x1504, 0x2488, 0xC450, 0x420}},
    {14548, 2, {0x800, 0x800, 0x13FE, 0x2020, 0x4820, 0x820, 0x1020, 0x3020, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {14549, 2, {0x840, 0x840, 0x1040, 0x2040, 0x4840, 0xFFE, 0x1040, 0x3040, 0x5040, 0x90A0, 0x10A0, 0x1090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {14550, 2, {0x840, 0x840, 0x1040, 0x2040, 0x4BFE, 0x840, 0x1040, 0x30A0, 0x50A0, 0x90A0, 0x1110, 0x1110, 0x1288, 0x1248, 0x1444, 0x1802}},
    {14551, 2, {0x840, 0x820, 0x1020, 0x21FC, 0x4904, 0x904, 0x1104, 0x31FC, 0x5104, 0x9100, 0x1100, 0x1100, 0x1100, 0x1200, 0x1200, 0x1400}},
    {14552, 2, {0x880, 0x880, 0x1100, 0x21FC, 0x4A04, 0xC04, 0x11E4, 0x3124, 0x5124, 0x9124, 0x11E4, 0x1124, 0x1004, 0x1004, 0x1028, 0x1010}},
    {14553, 2, {0x1040, 0x1040, 0x2040, 0x4040, 0x97FC, 0x1444, 0x2444, 0x6444, 0xA444, 0x27FC, 0x2444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2404}},
    {14554, 2, {0x800, 0xBFC, 0x1204, 0x2204, 0x4AF4, 0xA04, 0x1204, 0x32F4, 0x5294, 0x9294, 0x1294, 0x12F4, 0x1204, 0x1204, 0x1214, 0x1208}},
    {14555, 2, {0x840, 0x840, 0x10A0, 0x2110, 0x4A08, 0xC06, 0x13F8, 0x3000, 0x5000, 0x93F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {14556, 2, {0x1002, 0x17E2, 0x2242, 0x424A, 0x924A, 0x124A, 0x224A, 0x67EA, 0xA24A, 0x224A, 0x224A, 0x224A, 0x2242, 0x2442, 0x244A, 0x2844}},
    {14557, 2, {0x828, 0x824, 0x1024, 0x2020, 0x4BFE, 0x820, 0x1124, 0x3124, 0x5124, 0x93A8, 0x1128, 0x1110, 0x1212, 0x122A, 0x1446, 0x1082}},
    {14558, 2, {0x890, 0x890, 0x1108, 0x2148, 0x4A44, 0xC92, 0x1108, 0x33FC, 0x5094, 0x9090, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {14559, 2, {0x820, 0x820, 0x1020, 0x23FE, 0x4820, 0x820, 0x11FC, 0x3000, 0x5000, 0x91FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {14560, 2, {0x1040, 0x1040, 0x207C, 0x4084, 0x9588, 0x1650, 0x2420, 0x6450, 0xA488, 0x2706, 0x2460, 0x2410, 0x2408, 0x20C0, 0x2020, 0x2010}},
    {14561, 2, {0x1110, 0x1110, 0x2110, 0x4210, 0x92FE, 0x1692, 0x2692, 0x6A92, 0xA292, 0x22FE, 0x2292, 0x2210, 0x2210, 0x2210, 0x2210, 0x2210}},
    {14562, 2, {0x908, 0x888, 0x1090, 0x23FC, 0x4824, 0x824, 0x13FC, 0x3220, 0x5220, 0x93FE, 0x1062, 0x10A2, 0x112A, 0x1224, 0x1420, 0x1020}},
    {14563, 2, {0x1040, 0x1040, 0x2040, 0x47FE, 0x9040, 0x1450, 0x2248, 0x6248, 0xA554, 0x28A2, 0x20A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x2802}},
    {14564, 2, {0x850, 0x848, 0x105E, 0x23E4, 0x4828, 0x832, 0x10CA, 0x3706, 0x5050, 0x9048, 0x105E, 0x13E4, 0x1028, 0x1032, 0x10CA, 0x1706}},
    {14565, 2, {0x1110, 0x1110, 0x2110, 0x427C, 0x9210, 0x1610, 0x2AFE, 0x6208, 0xA208, 0x22FE, 0x2208, 0x2248, 0x2228, 0x2208, 0x2228, 0x2210}},
    {14566, 2, {0x838, 0xBC0, 0x1040, 0x2040, 0x4FFC, 0x950, 0x1248, 0x3486, 0x5080, 0x97FC, 0x1110, 0x1210, 0x11A0, 0x1060, 0x1090, 0x1308}},
    {14567, 2, {0x1100, 0x1208, 0x2448, 0x4788, 0x9108, 0x1248, 0x27C8, 0x6008, 0xA7C8, 0x2448, 0x27C8, 0x2448, 0x27CA, 0x244A, 0x244A, 0x24C6}},
    {14568, 2, {0x1210, 0x1210, 0x2210, 0x4210, 0x9F7E, 0x1510, 0x2510, 0x6510, 0xA57C, 0x2944, 0x2544, 0x2244, 0x2244, 0x2544, 0x257C, 0x2844}},
    {14569, 2, {0x1110, 0x1110, 0x2110, 0x4110, 0x97BC, 0x1110, 0x2110, 0x6338, 0xA3B8, 0x2554, 0x2554, 0x2992, 0x2110, 0x2110, 0x2110, 0x2110}},
    {14570, 2, {0x83C, 0xFC0, 0x1244, 0x2128, 0x4800, 0xBFC, 0x1080, 0x3080, 0x57FE, 0x9100, 0x11F8, 0x1288, 0x1250, 0x1420, 0x18D8, 0x1306}},
    {14571, 2, {0x808, 0x83C, 0x11E0, 0x2020, 0x4BFE, 0x820, 0x11FC, 0x3124, 0x51FC, 0x9124, 0x11FC, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {14572, 2, {0x800, 0x9F8, 0x1108, 0x2108, 0x49F8, 0x800, 0x17FE, 0x3108, 0x51F8, 0x9108, 0x11F8, 0x1108, 0x113E, 0x17C8, 0x1008, 0x1008}},
    {14573, 2, {0x800, 0x888, 0x1252, 0x2222, 0x4A52, 0xA8A, 0x1202, 0x33FE, 0x5088, 0x9144, 0x127A, 0x1088, 0x1150, 0x1020, 0x10D8, 0x1706}},
    {14574, 2, {0x820, 0x820, 0x13FE, 0x2020, 0x4820, 0xBFE, 0x128A, 0x3252, 0x52FA, 0x9222, 0x1222, 0x12FA, 0x1222, 0x1222, 0x122A, 0x1204}},
    {14575, 2, {0x800, 0xBFE, 0x1202, 0x2202, 0x4BFE, 0xA10, 0x1292, 0x3254, 0x52FE, 0x9282, 0x12FE, 0x1282, 0x12FE, 0x1482, 0x148A, 0x1884}},
    {14576, 2, {0x81C, 0x9E0, 0x1020, 0x2124, 0x48A8, 0xBFE, 0x1070, 0x30A8, 0x5124, 0x9222, 0x1040, 0x1024, 0x12A2, 0x128A, 0x1488, 0x1078}},
    {14577, 2, {0x820, 0xBFE, 0x1020, 0x21FC, 0x4820, 0xBFE, 0x1000, 0x31FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {14578, 2, {0x1210, 0x1210, 0x2A90, 0x4A90, 0x9FBE, 0x1024, 0x2054, 0x6F94, 0xA014, 0x2014, 0x2F94, 0x2888, 0x2888, 0x2F94, 0x2894, 0x2022}},
    {14579, 2, {0x1100, 0x1100, 0x23FC, 0x46A8, 0x9AA8, 0x12A8, 0x27FC, 0x62A8, 0xA2A8, 0x22A8, 0x2FFE, 0x2000, 0x22A8, 0x2254, 0x2454, 0x2000}},
    {14580, 2, {0x1110, 0x1110, 0x27FC, 0x4110, 0x9110, 0x1FFE, 0x2040, 0x67FC, 0xA444, 0x27FC, 0x2444, 0x27FC, 0x2000, 0x2110, 0x2208, 0x2404}},
    {14581, 2, {0x1010, 0x1410, 0x227C, 0x4010, 0x90FE, 0x1044, 0x2628, 0x62FE, 0xA210, 0x227C, 0x2210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x2800}},
    {14582, 2, {0x1040, 0x1FFE, 0x2000, 0x47FC, 0x9404, 0x15F4, 0x2514, 0x67FC, 0xA000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2000, 0x2FFE}},
    {14583, 2, {0x1080, 0x1040, 0x27FC, 0x4444, 0x92A8, 0x1294, 0x2474, 0x6000, 0xA3F8, 0x22A8, 0x2FFE, 0x2000, 0x27FC, 0x2040, 0x2140, 0x2080}},
    {14584, 2, {0x1508, 0x1548, 0x2DA8, 0x453E, 0x9FA8, 0x1248, 0x2448, 0x6ABE, 0xA308, 0x2588, 0x2988, 0x237E, 0x2540, 0x2920, 0x251E, 0x2200}},
    {14585, 2, {0x1040, 0x1020, 0x27FE, 0x44A8, 0x94FE, 0x1550, 0x277C, 0x6550, 0xA57C, 0x2550, 0x257E, 0x2540, 0x28A4, 0x2A8A, 0x328A, 0x2478}},
    {14586, 2, {0x400, 0x400, 0x400, 0x400, 0x2510, 0x2488, 0x2444, 0x4444, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400}},
    {14587, 2, {0x1010, 0x820, 0x440, 0x280, 0x100, 0x280, 0xC60, 0x3018, 0xC006, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14588, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0x1844, 0x5544, 0x5144, 0x5144, 0x9144, 0x1244, 0x1044, 0x1084, 0x1084, 0x1104, 0x1228, 0x1410}},
    {14589, 2, {0x200, 0x200, 0xFFFE, 0x420, 0x820, 0x1C40, 0x380, 0xC70, 0x7008, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14590, 2, {0x1000, 0x13FC, 0x1104, 0x1144, 0x1924, 0x54A8, 0x5088, 0x5088, 0x9050, 0x1050, 0x1020, 0x1020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {14591, 2, {0x1040, 0x1040, 0x1080, 0x18FC, 0x5504, 0x5204, 0x5004, 0x9104, 0x1084, 0x1044, 0x1044, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {14592, 2, {0x300, 0x80, 0x2140, 0x2220, 0x2410, 0x2808, 0x3006, 0x2000, 0x3FFC, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14593, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14594, 2, {0x0, 0x3FE0, 0x820, 0x840, 0xC78, 0xA08, 0x1210, 0x1120, 0x20C0, 0x4330, 0x8C0E, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14595, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1880, 0x5480, 0x5080, 0x50FC, 0x9084, 0x1084, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {14596, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1820, 0x5420, 0x51FC, 0x5124, 0x9124, 0x1124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020, 0x1020}},
    {14597, 2, {0x7FF8, 0x200, 0x4200, 0x4200, 0x7FFC, 0x480, 0x880, 0x1084, 0x2084, 0xC07C, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14598, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1900, 0x55FC, 0x5144, 0x5144, 0x9144, 0x1128, 0x1128, 0x1110, 0x1110, 0x1228, 0x1244, 0x1482}},
    {14599, 2, {0x1010, 0x1078, 0x13C0, 0x1040, 0x1840, 0x5440, 0x5040, 0x57FE, 0x9040, 0x10A0, 0x10A0, 0x1090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {14600, 2, {0x1000, 0x1040, 0x1020, 0x1010, 0x1890, 0x5480, 0x5080, 0x5284, 0x9282, 0x1282, 0x1282, 0x1488, 0x1088, 0x1088, 0x1078, 0x1000}},
    {14601, 2, {0x1000, 0x13FC, 0x1010, 0x1910, 0x5510, 0x5110, 0x5210, 0x93FE, 0x1030, 0x1050, 0x1090, 0x1110, 0x1210, 0x1410, 0x1050, 0x1020}},
    {14602, 2, {0x1004, 0x1308, 0x10D0, 0x1020, 0x1850, 0x5488, 0x5104, 0x5000, 0x9004, 0x1104, 0x1088, 0x1050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {14603, 2, {0x1010, 0x1090, 0x1090, 0x1888, 0x5508, 0x5104, 0x5204, 0x95FA, 0x1088, 0x1088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {14604, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0x1BFE, 0x5488, 0x5088, 0x5088, 0x9088, 0x10F8, 0x1088, 0x1088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {14605, 2, {0x0, 0x7DFE, 0x1040, 0x1080, 0x11FC, 0x1004, 0x1E04, 0xE004, 0x4028, 0x110, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14606, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x2244, 0x227C, 0x4A44, 0x8400, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14607, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x1B10, 0x54A0, 0x5040, 0x50A0, 0x9318, 0x1C06, 0x10C0, 0x1020, 0x1010, 0x1180, 0x1060, 0x1010}},
    {14608, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14609, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x1BFE, 0x5440, 0x5040, 0x5080, 0x90FC, 0x1184, 0x1284, 0x1084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {14610, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1850, 0x5450, 0x5088, 0x5124, 0x9222, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {14611, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x1BFC, 0x5440, 0x50E0, 0x50E0, 0x9150, 0x1150, 0x1248, 0x15F4, 0x1842, 0x1040, 0x1040, 0x1040}},
    {14612, 2, {0x1088, 0x1088, 0x1088, 0x1908, 0x557E, 0x5308, 0x5508, 0x9148, 0x1128, 0x1128, 0x1108, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {14613, 2, {0x1020, 0x1020, 0x1124, 0x1124, 0x1924, 0x5524, 0x51FC, 0x5020, 0x9020, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {14614, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1A02, 0x5404, 0x5000, 0x5000, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {14615, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1210, 0x2208, 0xC226, 0x1290, 0x1248, 0x2248, 0xA00, 0x400}},
    {14616, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x1A10, 0x5420, 0x53FC, 0x5004, 0x9004, 0x1004, 0x11FC, 0x1004, 0x1004, 0x1004, 0x13FC, 0x1004}},
    {14617, 2, {0x3FF8, 0x2008, 0x2008, 0x3FC8, 0x2048, 0x2048, 0x3FC8, 0x2008, 0x2008, 0x3FF8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14618, 2, {0x1080, 0x1080, 0x113C, 0x1A00, 0x5480, 0x5080, 0x517E, 0x9308, 0x1508, 0x1108, 0x1108, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {14619, 2, {0x1080, 0x1088, 0x109C, 0x1970, 0x5510, 0x5310, 0x5510, 0x9110, 0x11FE, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x117C, 0x1100}},
    {14620, 2, {0x2028, 0x2024, 0x2024, 0x2020, 0x33FE, 0xAA20, 0xA224, 0xA224, 0xA324, 0x22A8, 0x2228, 0x2210, 0x2212, 0x242A, 0x2446, 0x2882}},
    {14621, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1840, 0x5488, 0x5104, 0x53FE, 0x9092, 0x1090, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {14622, 2, {0x1040, 0x1020, 0x13FE, 0x1A02, 0x5400, 0x5038, 0x53C0, 0x9040, 0x1040, 0x107C, 0x17C0, 0x1040, 0x1042, 0x1042, 0x103E, 0x1000}},
    {14623, 2, {0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14624, 2, {0x7F84, 0x824, 0x1024, 0x3F24, 0x5124, 0xA24, 0x404, 0x804, 0x3014, 0xC008, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14625, 2, {0x2004, 0x2004, 0x2FC4, 0x2214, 0x3214, 0xABD4, 0xA254, 0xA454, 0xA654, 0x2554, 0x2894, 0x2094, 0x2104, 0x2204, 0x2414, 0x2808}},
    {14626, 2, {0x1040, 0x1040, 0x17FE, 0x1080, 0x1900, 0x55FC, 0x5304, 0x5504, 0x99FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {14627, 2, {0x1010, 0x820, 0x7FFC, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4120, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14628, 2, {0x1020, 0x1020, 0x13FE, 0x1222, 0x1A22, 0x57FE, 0x5222, 0x5222, 0x93FE, 0x1020, 0x1120, 0x10A0, 0x1040, 0x10A0, 0x1118, 0x1606}},
    {14629, 2, {0x1040, 0x1020, 0x1020, 0x17FE, 0x1840, 0x5484, 0x5108, 0x53F0, 0x9022, 0x1044, 0x1188, 0x1610, 0x1020, 0x1050, 0x1188, 0x1604}},
    {14630, 2, {0x2040, 0x2040, 0x2FFE, 0x3040, 0xA840, 0xA7FC, 0xA444, 0xA444, 0x27FC, 0x2040, 0x2240, 0x2140, 0x2080, 0x2140, 0x2230, 0x2C0E}},
    {14631, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1900, 0x55FE, 0x5100, 0x5100, 0x917C, 0x1144, 0x1144, 0x1144, 0x1144, 0x127C, 0x1244, 0x1400}},
    {14632, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0x1BFE, 0x5488, 0x5088, 0x5088, 0x9088, 0x17FE, 0x1000, 0x1088, 0x1084, 0x1104, 0x1202, 0x1402}},
    {14633, 2, {0x800, 0xFF0, 0x1020, 0x2C40, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14634, 2, {0x0, 0xFFFE, 0x1000, 0x10F8, 0x3E88, 0x4288, 0xA4A8, 0x1894, 0x1084, 0x207C, 0xC100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14635, 2, {0x400, 0x400, 0x7FFC, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0x850, 0xA20, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14636, 2, {0x1020, 0x1020, 0x11FC, 0x1820, 0x5420, 0x5020, 0x53FE, 0x9000, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {14637, 2, {0x1010, 0x1078, 0x13C0, 0x1040, 0x1840, 0x57FE, 0x50A0, 0x5110, 0x9208, 0x1516, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {14638, 2, {0x2200, 0x1400, 0x19FC, 0x2420, 0x4C20, 0x14F8, 0x2420, 0x4420, 0x5FC, 0x2800, 0x1100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14639, 2, {0x1000, 0x13FE, 0x1202, 0x1252, 0x1A8A, 0x5706, 0x5202, 0x52FA, 0x928A, 0x128A, 0x128A, 0x12FA, 0x128A, 0x1202, 0x120A, 0x1204}},
    {14640, 2, {0x900, 0x900, 0x11FE, 0x1280, 0x3480, 0x50FC, 0x9080, 0x1080, 0x10FC, 0x1080, 0x1080, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14641, 2, {0x1000, 0x13FC, 0x1000, 0x1000, 0x19F8, 0x5508, 0x5108, 0x5108, 0x91F8, 0x1000, 0x1108, 0x1088, 0x1090, 0x1000, 0x17FE, 0x1000}},
    {14642, 2, {0x100, 0x100, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x888, 0x28A4, 0x2824, 0x47E0}},
    {14643, 2, {0x1000, 0x13FE, 0x1010, 0x1820, 0x5468, 0x50A4, 0x5322, 0x9020, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {14644, 2, {0x2000, 0x2FFE, 0x2480, 0x2490, 0x3490, 0xAF90, 0xA490, 0xA490, 0xA490, 0x2790, 0x2490, 0x2492, 0x24D2, 0x2F92, 0x208E, 0x2080}},
    {14645, 2, {0x40, 0x7E40, 0x47E, 0x888, 0xE88, 0x1948, 0xE850, 0x820, 0x850, 0x2888, 0x1306, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14646, 2, {0x1090, 0x1088, 0x1080, 0x13FE, 0x18A0, 0x54A8, 0x50B0, 0x50A4, 0x90A8, 0x10B0, 0x1124, 0x112A, 0x1132, 0x1222, 0x12DE, 0x1400}},
    {14647, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0x1A88, 0x5446, 0x51F0, 0x5010, 0x9020, 0x1040, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {14648, 2, {0x1040, 0x1040, 0x13FC, 0x1840, 0x5440, 0x5040, 0x57FE, 0x9090, 0x1294, 0x1292, 0x1492, 0x1912, 0x1110, 0x1210, 0x1450, 0x1820}},
    {14649, 2, {0x1050, 0x1048, 0x1048, 0x1840, 0x57FE, 0x5040, 0x5244, 0x9164, 0x1168, 0x10D0, 0x1150, 0x1248, 0x1444, 0x1042, 0x1140, 0x1080}},
    {14650, 2, {0x2080, 0x2040, 0x2FFE, 0x2200, 0x3200, 0xAA00, 0xA3F8, 0xA000, 0xA248, 0x2248, 0x2248, 0x2248, 0x2248, 0x244A, 0x244A, 0x2806}},
    {14651, 2, {0x600, 0x387C, 0x2244, 0x3264, 0x2A54, 0x2244, 0x2A54, 0x3248, 0x440, 0x1840, 0x6100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14652, 2, {0x1110, 0x1110, 0x1110, 0x1390, 0x193C, 0x5514, 0x5114, 0x57D4, 0x9114, 0x1114, 0x1214, 0x1294, 0x17D4, 0x1224, 0x1054, 0x1088}},
    {14653, 2, {0x1020, 0x1020, 0x13FE, 0x1820, 0x55FC, 0x5040, 0x53FE, 0x9048, 0x1088, 0x10FE, 0x1108, 0x1148, 0x1228, 0x1408, 0x1028, 0x1010}},
    {14654, 2, {0x1000, 0x13FE, 0x1202, 0x1252, 0x1888, 0x5504, 0x5020, 0x5020, 0x93FE, 0x1020, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020, 0x1020}},
    {14655, 2, {0x1040, 0x1080, 0x17FE, 0x1110, 0x1A48, 0x5446, 0x53F8, 0x5248, 0x9248, 0x13F8, 0x1248, 0x1248, 0x13F8, 0x1042, 0x1042, 0x103E}},
    {14656, 2, {0x1040, 0x1040, 0x13FC, 0x1040, 0x1840, 0x55F8, 0x5108, 0x51F8, 0x9108, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {14657, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1290, 0x1290, 0x5290, 0x5FF0, 0x4000, 0x7FFC, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14658, 2, {0x1000, 0x13F8, 0x1208, 0x13F8, 0x1A08, 0x57F8, 0x5000, 0x53FC, 0x9010, 0x17FE, 0x1010, 0x1210, 0x1110, 0x1010, 0x1050, 0x1020}},
    {14659, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0x1BDE, 0x5652, 0x5252, 0x53DE, 0x9202, 0x1202, 0x1202, 0x1202, 0x1202, 0x1202, 0x120A, 0x1204}},
    {14660, 2, {0x1020, 0x1020, 0x11FC, 0x1820, 0x5420, 0x53FE, 0x5088, 0x9144, 0x1242, 0x10F8, 0x1188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {14661, 2, {0x101C, 0x13E0, 0x1084, 0x1A44, 0x5548, 0x5110, 0x57FE, 0x9402, 0x1000, 0x13F8, 0x1108, 0x1110, 0x10A0, 0x1040, 0x11B0, 0x160E}},
    {14662, 2, {0x1040, 0x1040, 0x13F8, 0x1848, 0x5448, 0x57FE, 0x5048, 0x9048, 0x13F8, 0x1444, 0x12E8, 0x1150, 0x1248, 0x1446, 0x1140, 0x1080}},
    {14663, 2, {0x1050, 0x1048, 0x105E, 0x13E4, 0x1828, 0x5432, 0x50CA, 0x5706, 0x9050, 0x1048, 0x105E, 0x13E4, 0x1028, 0x1032, 0x10CA, 0x1706}},
    {14664, 2, {0x1020, 0x1020, 0xFE3E, 0x9220, 0x9220, 0x92FC, 0x9A84, 0x9484, 0x10FC, 0x1100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {14665, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0x1908, 0x5490, 0x5000, 0x53FE, 0x9000, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {14666, 2, {0x2000, 0x27BC, 0x24A4, 0x24A4, 0x34A4, 0xAFBC, 0xA4A4, 0xA4A4, 0xA4A4, 0x27BC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x2AA4, 0x314C}},
    {14667, 2, {0x840, 0x820, 0x17FE, 0x1040, 0x30A4, 0x51A8, 0x9290, 0x1488, 0x10C4, 0x1082, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14668, 2, {0x1100, 0x1100, 0x11FC, 0x1A04, 0x5484, 0x5494, 0x52A4, 0x9084, 0x17FC, 0x1084, 0x11C4, 0x12A4, 0x1494, 0x1084, 0x10A8, 0x1010}},
    {14669, 2, {0x1108, 0x1108, 0x17FE, 0x1108, 0x1908, 0x55F8, 0x5108, 0x5108, 0x91F8, 0x1108, 0x1108, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {14670, 2, {0xC80, 0x7080, 0x11FC, 0xFD24, 0x3224, 0x3844, 0x5444, 0x5084, 0x9104, 0x1228, 0x1110, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14671, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0x1A94, 0x5694, 0x5294, 0x53FC, 0x9294, 0x1294, 0x1294, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {14672, 2, {0x100, 0x280, 0xC60, 0x3018, 0xC7C6, 0x0, 0x3EF8, 0x2288, 0x2288, 0x3EA8, 0x90, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14673, 2, {0x0, 0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFE, 0x4102, 0x8884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14674, 2, {0x1008, 0x103C, 0x13C0, 0x1044, 0x1A24, 0x5528, 0x5100, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x10A8, 0x1124, 0x1222, 0x1420, 0x1020}},
    {14675, 2, {0x1040, 0x1040, 0x13FC, 0x18A0, 0x5510, 0x5208, 0x57FE, 0x9008, 0x13C8, 0x1248, 0x1248, 0x1248, 0x13C8, 0x1008, 0x1028, 0x1010}},
    {14676, 2, {0x1040, 0x1020, 0x13FE, 0x1880, 0x5504, 0x53FE, 0x5002, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {14677, 2, {0x100, 0x910, 0x920, 0x12C0, 0x430, 0x1908, 0x6914, 0x910, 0x12A0, 0x440, 0x1830, 0x600C, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14678, 2, {0x90, 0x7890, 0x4908, 0x4908, 0x6A44, 0x5C52, 0x4888, 0xFDF8, 0x4808, 0x4840, 0x4824, 0x4AA2, 0x4A8A, 0x4C88, 0x8878, 0x1800}},
    {14679, 2, {0x11FC, 0x1104, 0x1104, 0x11FC, 0x1904, 0x5504, 0x51FC, 0x5040, 0x9020, 0x13FE, 0x1088, 0x1088, 0x1050, 0x1020, 0x10D8, 0x1706}},
    {14680, 2, {0x2200, 0x2200, 0x227E, 0x2712, 0x3212, 0xAA52, 0xA252, 0xAFD2, 0xA252, 0x2292, 0x2412, 0x2512, 0x2FA2, 0x24A2, 0x204A, 0x2084}},
    {14681, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0x1924, 0x5524, 0x51FC, 0x5020, 0x9040, 0x1020, 0x10A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {14682, 2, {0x2040, 0x2140, 0x265C, 0x3444, 0xAC44, 0xA75C, 0xA444, 0xA444, 0x27FC, 0x2444, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x2404, 0x2802}},
    {14683, 2, {0x1000, 0x13FC, 0x1008, 0x1810, 0x5420, 0x53AE, 0x52A2, 0x92AA, 0x12A4, 0x12A4, 0x13AA, 0x1032, 0x10A0, 0x1040, 0x17FE, 0x1000}},
    {14684, 2, {0x1000, 0x13FE, 0x1222, 0x1BFE, 0x5622, 0x53FE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {14685, 2, {0x2000, 0x277E, 0x2124, 0x3124, 0xA93C, 0xA724, 0xA424, 0xA43C, 0x2424, 0x2724, 0x2126, 0x217C, 0x2104, 0x2104, 0x2A04, 0x2404}},
    {14686, 2, {0x1104, 0x1088, 0x1000, 0x13FE, 0x1840, 0x54C2, 0x5122, 0x5254, 0x9098, 0x1138, 0x1254, 0x1094, 0x1112, 0x1610, 0x1050, 0x1020}},
    {14687, 2, {0x1100, 0x11F8, 0x1208, 0x13F0, 0x1810, 0x57FE, 0x5080, 0x5144, 0x9668, 0x10B0, 0x1128, 0x1668, 0x10A4, 0x1122, 0x16A0, 0x1040}},
    {14688, 2, {0x1000, 0x13FE, 0x1022, 0x1920, 0x553C, 0x5120, 0x52FE, 0x9400, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {14689, 2, {0x1000, 0x13FC, 0x1204, 0x13FC, 0x1A00, 0x5698, 0x52E0, 0x5284, 0x927C, 0x1220, 0x13FE, 0x1270, 0x14A8, 0x1524, 0x1A22, 0x1020}},
    {14690, 2, {0x1108, 0x1088, 0x1090, 0x17FE, 0x1890, 0x5490, 0x53FC, 0x5294, 0x9294, 0x130C, 0x1204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {14691, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x920, 0x37D8, 0xC106, 0x1FF0, 0x100, 0xFE0, 0x100, 0x3FF8, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14692, 2, {0x20, 0x7E20, 0x1450, 0x888, 0x7F44, 0x2A22, 0x49F8, 0x8808, 0x2810, 0x1020, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14693, 2, {0x11FC, 0x1124, 0x1124, 0x11FC, 0x1924, 0x5524, 0x51FC, 0x5020, 0x9020, 0x13FE, 0x1222, 0x122A, 0x12FA, 0x120A, 0x1202, 0x1206}},
    {14694, 2, {0x3FFC, 0x2080, 0x2FF8, 0x2080, 0x2490, 0x2490, 0x2AA8, 0x3144, 0x2220, 0x2410, 0x3FFE, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14695, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x19FC, 0x5504, 0x5104, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1000, 0x1088, 0x1104}},
    {14696, 2, {0x1020, 0x10F8, 0xFC20, 0x2420, 0x25FC, 0x4800, 0x2820, 0x10F8, 0x2820, 0x4420, 0x81FC, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14697, 2, {0x1040, 0x1044, 0x13F4, 0x1848, 0x5450, 0x57FE, 0x5040, 0x9080, 0x11F8, 0x1308, 0x1508, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {14698, 2, {0x1020, 0x1124, 0x10A8, 0x1020, 0x1BFE, 0x54A8, 0x5124, 0x5202, 0x9040, 0x17FE, 0x1088, 0x1108, 0x1190, 0x1060, 0x1198, 0x1604}},
    {14699, 2, {0x1040, 0x1020, 0x17FE, 0x1000, 0x19F8, 0x5508, 0x51F8, 0x5000, 0x93FC, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {14700, 2, {0x2080, 0x2040, 0x27FC, 0x3208, 0xA910, 0xA7FE, 0xA402, 0xA884, 0x2040, 0x27FC, 0x2100, 0x21F8, 0x2108, 0x2208, 0x2228, 0x2410}},
    {14701, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x19FC, 0x5440, 0x53FE, 0x5088, 0x9124, 0x1222, 0x1124, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {14702, 2, {0x1080, 0x1080, 0x11FC, 0x1244, 0x1954, 0x55F4, 0x5084, 0x5128, 0x9290, 0x11FC, 0x1244, 0x1554, 0x11F4, 0x1084, 0x1128, 0x1210}},
    {14703, 2, {0x7C, 0x7F80, 0x1108, 0x890, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x8, 0x1FF8, 0x8, 0x3FF8, 0x208, 0x5124, 0x9FE4}},
    {14704, 2, {0x10A0, 0x132E, 0x1222, 0x1222, 0x1BAE, 0x5622, 0x5222, 0x53FE, 0x9020, 0x13FE, 0x1104, 0x1088, 0x1050, 0x1020, 0x10D8, 0x1306}},
    {14705, 2, {0x1000, 0x13FE, 0x1228, 0x1228, 0x1AEE, 0x5628, 0x5228, 0x52EE, 0x9228, 0x1228, 0x12EE, 0x1228, 0x1228, 0x1228, 0x13FE, 0x1000}},
    {14706, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x1010, 0x2008, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14707, 2, {0x2000, 0x27FE, 0x2420, 0x2440, 0x35FC, 0xAD04, 0xA5FC, 0xA504, 0xA5FC, 0x2524, 0x2420, 0x24A8, 0x2924, 0x2A22, 0x30A0, 0x2040}},
    {14708, 2, {0x2004, 0x203E, 0x27E0, 0x343E, 0xAC20, 0xA5FE, 0xA522, 0xA538, 0x25E2, 0x251E, 0x2500, 0x2978, 0x2A48, 0x324A, 0x248A, 0x2906}},
    {14709, 2, {0x2000, 0x27FE, 0x2402, 0x25FA, 0x3442, 0xACC2, 0xA52A, 0xA46A, 0xA4B2, 0x252A, 0x246A, 0x24AA, 0x2522, 0x2442, 0x27FE, 0x2402}},
    {14710, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1210, 0x2108, 0xC906, 0x2828, 0x2824, 0x47E4, 0x0}},
    {14711, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2FFE, 0x2248, 0x2248, 0x2554, 0x28A2, 0x2318, 0x2C06, 0x2080, 0x4A44, 0x4A4A, 0x920A, 0x1F8}},
    {14712, 2, {0x0, 0xFDFC, 0x8524, 0x8524, 0xFDFC, 0x9124, 0x9124, 0x91FC, 0xFD04, 0x9000, 0x9020, 0x9014, 0x9552, 0xB542, 0xCE48, 0x8438}},
    {14713, 2, {0x1010, 0x5450, 0x3888, 0x1104, 0xFEFA, 0x1048, 0x3848, 0x5488, 0x9128, 0x1210, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14714, 2, {0x1020, 0x1020, 0x11FC, 0x1820, 0x5420, 0x53FE, 0x5080, 0x9104, 0x13FE, 0x1002, 0x13FC, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {14715, 2, {0x28, 0x24, 0x3FFE, 0x20A0, 0x2F24, 0x2224, 0x2FA8, 0x2218, 0x2B12, 0x52AA, 0x4A46, 0x8482, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14716, 2, {0x2000, 0x27FC, 0x2040, 0x3278, 0xAA40, 0xAFFE, 0xA100, 0xA1FC, 0x2200, 0x23FC, 0x2004, 0x2554, 0x2554, 0x2804, 0x2028, 0x2010}},
    {14717, 2, {0x1020, 0x13FE, 0x1020, 0x19FC, 0x5420, 0x53FE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {14718, 2, {0x1000, 0x13FE, 0x1248, 0x1248, 0x1BFE, 0x5648, 0x5220, 0x53FE, 0x9240, 0x12FC, 0x1344, 0x1244, 0x127C, 0x1200, 0x13FE, 0x1000}},
    {14719, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x8100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {14720, 2, {0x2020, 0x247C, 0x2284, 0x3148, 0xA830, 0xA0CE, 0xA610, 0xA27C, 0x2210, 0x227C, 0x2210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x2800}},
    {14721, 2, {0x40, 0x27FC, 0x1040, 0x13F8, 0x248, 0x73F8, 0x1248, 0x13F8, 0x1040, 0x17FC, 0x2840, 0x47FE, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14722, 2, {0x1110, 0x920, 0x3FFC, 0x2004, 0x4FE8, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x2118, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {14723, 2, {0x1040, 0x1020, 0x13FC, 0x1800, 0x5508, 0x5090, 0x53FE, 0x9220, 0x12A0, 0x12FC, 0x1320, 0x1220, 0x12FC, 0x1420, 0x1420, 0x19FE}},
    {14724, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2220, 0x2F7C, 0x2220, 0x2770, 0x2AA8, 0x3224, 0x2080, 0x2048, 0x4A44, 0x4A14, 0x91F0, 0x0}},
    {14725, 2, {0x22A8, 0x22A8, 0x27FC, 0x32A8, 0xAAAA, 0xA4E6, 0xA800, 0xA7FC, 0x2444, 0x2040, 0x23F8, 0x2248, 0x2248, 0x2248, 0x2258, 0x2040}},
    {14726, 2, {0x1040, 0x107C, 0x1040, 0x1BFE, 0x5642, 0x5278, 0x53C4, 0x923C, 0x1208, 0x12F0, 0x1324, 0x12A8, 0x15FE, 0x1420, 0x18A0, 0x1040}},
    {14727, 2, {0x1040, 0x1020, 0x13FE, 0x1250, 0x1A50, 0x57FE, 0x5252, 0x5252, 0x93FE, 0x1200, 0x1292, 0x12D4, 0x1298, 0x1492, 0x14D2, 0x188E}},
    {14728, 2, {0x2208, 0x2110, 0x2000, 0x27FC, 0x3040, 0xABF8, 0xA040, 0xA7FC, 0xA000, 0x23F8, 0x22A8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000}},
    {14729, 2, {0x1020, 0x13FE, 0x1020, 0x19FC, 0x5400, 0x53FE, 0x5202, 0x91FC, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1088, 0x13FE, 0x1000}},
    {14730, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0x19FC, 0x54A8, 0x50A8, 0x53FE, 0x90A8, 0x10A8, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {14731, 2, {0x2040, 0x2040, 0x27FC, 0x3110, 0xAA48, 0xA446, 0xA3F8, 0xA050, 0x27FC, 0x2080, 0x23F8, 0x2D08, 0x21F8, 0x2108, 0x21F8, 0x2108}},
    {14732, 2, {0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE, 0xC20, 0x703C, 0x1DE2, 0xF022, 0x1C1E, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14733, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0x1BDE, 0x5488, 0x5050, 0x53FE, 0x9020, 0x1124, 0x1124, 0x11FC, 0x1024, 0x1040, 0x1080, 0x1100}},
    {14734, 2, {0x1020, 0x11FC, 0x1024, 0x13FE, 0x1824, 0x55FC, 0x5020, 0x53FE, 0x9000, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1000, 0x13FE}},
    {14735, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0x3420, 0xACA4, 0xA39C, 0xA000, 0xA110, 0x27FC, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2208, 0x2404}},
    {14736, 2, {0x1014, 0x17FE, 0x1010, 0x1BD0, 0x5410, 0x53D0, 0x5010, 0x93D0, 0x1250, 0x13D0, 0x1250, 0x13C8, 0x124A, 0x13CA, 0x1186, 0x1242}},
    {14737, 2, {0x2008, 0x2788, 0x2088, 0x3110, 0xAFDE, 0xA494, 0xA4A4, 0xA794, 0x2494, 0x2794, 0x2494, 0x24C8, 0x2788, 0x2C94, 0x20A4, 0x20C2}},
    {14738, 2, {0x1000, 0x13FE, 0x1252, 0x1252, 0x1BFE, 0x5420, 0x53FE, 0x5000, 0x91FC, 0x1000, 0x11FC, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {14739, 2, {0x1110, 0x1114, 0x11D2, 0x1250, 0x1A7E, 0x5550, 0x5090, 0x52A8, 0x9128, 0x1144, 0x1244, 0x1482, 0x1008, 0x12A4, 0x1252, 0x1452}},
    {14740, 2, {0x3E10, 0x2214, 0x3E12, 0x2210, 0x3EFE, 0x10, 0x3E28, 0x2228, 0x3E28, 0x2244, 0x3E44, 0x2282, 0x2700, 0x4884, 0x4812, 0x87F2}},
    {14741, 2, {0x2020, 0x2720, 0x253C, 0x2544, 0x35A8, 0xAE10, 0xA528, 0xA546, 0xA538, 0x2510, 0x2550, 0x267E, 0x2490, 0x247C, 0x2410, 0x24FE}},
    {14742, 2, {0x2000, 0x279E, 0x2492, 0x279E, 0x3492, 0xAF9E, 0xA442, 0xA422, 0xA40A, 0x2546, 0x2542, 0x2552, 0x2632, 0x2402, 0x240A, 0x2404}},
    {14743, 2, {0x2110, 0x2110, 0x27BC, 0x3110, 0xABB8, 0xA554, 0xA912, 0xA000, 0x23F8, 0x2000, 0x27FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {14744, 2, {0x1020, 0x7E20, 0x423E, 0x7E44, 0x42A4, 0x7E24, 0x1028, 0xFF28, 0x2010, 0x3E28, 0x4244, 0x8682, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14745, 2, {0x11FC, 0x1124, 0x11FC, 0x1924, 0x55FC, 0x5020, 0x53FE, 0x922A, 0x12FA, 0x120A, 0x1246, 0x1020, 0x10A4, 0x128A, 0x128A, 0x1478}},
    {14746, 2, {0x4010, 0x2008, 0x2108, 0xF8BE, 0x80, 0x5014, 0x2208, 0xFA3E, 0x2208, 0x2A48, 0x7A3C, 0x2AA8, 0x2AA8, 0x5288, 0x4188, 0x8008}},
    {14747, 2, {0xA00, 0x33B8, 0x2208, 0x3BB8, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x820, 0x1210, 0x2108, 0xC906, 0x2828, 0x2824, 0x47E4, 0x0}},
    {14748, 2, {0x2108, 0x27FE, 0x2148, 0x2220, 0x37FE, 0xAA20, 0xA3FC, 0xA220, 0xA3FC, 0x2220, 0x23FE, 0x2200, 0x27F8, 0x2110, 0x20E0, 0x2F1E}},
    {14749, 2, {0x3C20, 0xD040, 0x4DFC, 0x3104, 0xC9FC, 0x1504, 0x25FC, 0xCC50, 0x1450, 0x2492, 0xD50E, 0xA00, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14750, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0xFFFE, 0x2, 0x3FF0, 0x100, 0xFFFE, 0x10, 0x7DFC, 0x4490, 0x7D30, 0x4894, 0x87F2}},
    {14751, 2, {0x31FC, 0xCB24, 0x2D24, 0x31FC, 0xC924, 0x1524, 0x25FC, 0xCC20, 0x15FC, 0x2420, 0xD420, 0xBFE, 0x100, 0x4884, 0x4812, 0x87F2}},
    {14752, 2, {0x2284, 0x2284, 0x2AA4, 0x26C4, 0x329E, 0xAFE4, 0xA004, 0xA454, 0xA28C, 0x2FE4, 0x2104, 0x27C4, 0x2104, 0x21C4, 0x2E14, 0x2008}},
    {14753, 2, {0x2444, 0x24E4, 0x28A8, 0x3AAA, 0xAEEE, 0xA4A4, 0xAAAA, 0xAEEE, 0x2242, 0x2040, 0x2FFE, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {14754, 2, {0x1108, 0x1108, 0x11EE, 0x1294, 0x1842, 0x5400, 0x53DE, 0x5252, 0x93D2, 0x1252, 0x13D2, 0x121A, 0x1294, 0x1350, 0x1210, 0x1010}},
    {14755, 2, {0x2100, 0x479C, 0x8280, 0x17C0, 0x203E, 0x6788, 0xA488, 0x2788, 0x2108, 0x27C8, 0x2908, 0x2FE8, 0x2110, 0x4884, 0x4812, 0x87F2}},
    {14756, 2, {0x1040, 0xFE40, 0x107C, 0x7C84, 0x28, 0xFE20, 0x8220, 0x7C50, 0x88, 0x7D26, 0x4410, 0x7C54, 0x4542, 0x2F4A, 0xF238, 0x4000}},
    {14757, 2, {0x2040, 0x2FFE, 0x2842, 0x27FC, 0x3040, 0xABF8, 0xA040, 0xA7FC, 0xA040, 0x27FC, 0x24A4, 0x27FC, 0x2040, 0x2524, 0x252A, 0x28FA}},
    {14758, 2, {0x2000, 0x279E, 0x2492, 0x379E, 0xAC92, 0xA79E, 0xA482, 0xA532, 0x2512, 0x25B2, 0x2512, 0x25F2, 0x24A2, 0x252A, 0x261A, 0x2404}},
    {14759, 2, {0x2108, 0x2528, 0x27BE, 0x2948, 0x37BE, 0xAB18, 0xA5AA, 0xA946, 0xA3F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {14760, 2, {0x27BC, 0x24A4, 0x27BC, 0x24A4, 0x37BC, 0xAC44, 0xA5F4, 0xA444, 0xA5F4, 0x2554, 0x25F4, 0x2554, 0x25F4, 0x24E4, 0x2554, 0x244C}},
    {14761, 2, {0x201C, 0x2EE0, 0x2A54, 0x2A28, 0x3A7C, 0xAC10, 0xAAFE, 0xAA00, 0xAAFC, 0x2A04, 0x2A7C, 0x2C04, 0x28FC, 0x2814, 0x294A, 0x2A38}},
    {14762, 2, {0x2248, 0x22E8, 0x2308, 0x36EA, 0xAA1C, 0xA2E8, 0xA208, 0xA2E8, 0x25B4, 0x28E2, 0x2000, 0x23F8, 0x2110, 0x20E0, 0x2318, 0x2C06}},
    {14763, 2, {0x2528, 0x27BE, 0x2948, 0x37BE, 0xAB18, 0xA5AA, 0xA946, 0xA3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {14764, 2, {0x2000, 0x27FC, 0x24A4, 0x24A4, 0x37FC, 0xAA10, 0xA2A8, 0xA4BE, 0xAF68, 0x223C, 0x24A8, 0x2FBC, 0x2028, 0x2AA8, 0x2ABE, 0x2020}},
    {14765, 2, {0x27FC, 0x2040, 0x2FFE, 0x2842, 0x3358, 0xA840, 0xA358, 0xA000, 0xAEEE, 0x2AAA, 0x2EEE, 0x2000, 0x2FFE, 0x2248, 0x2554, 0x2FFE}},
    {14766, 2, {0x240, 0x220, 0x3F8, 0x7E00, 0x200, 0x3F8, 0x3E00, 0x200, 0x2FC, 0x7F10, 0x120, 0x140, 0x184, 0xE44, 0x7034, 0xC}},
    {14767, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x40, 0x840, 0x844, 0x1044, 0x2244, 0x7F28, 0x128, 0x12, 0xF32, 0xF04A, 0x4086, 0x302}},
    {14768, 2, {0x2250, 0x2248, 0x2248, 0x5540, 0x88C0, 0x40, 0xFFFE, 0x40, 0x40, 0x24, 0x28, 0x12, 0x32, 0xCA, 0x306, 0x2}},
    {14769, 2, {0x40, 0x48, 0x7F44, 0x44, 0x40, 0xFFFE, 0x40, 0x844, 0x844, 0x7F44, 0x828, 0x82A, 0xF12, 0xF02A, 0x4046, 0x82}},
    {14770, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x1040, 0x1044, 0x1F44, 0x1144, 0x1128, 0x112A, 0x2112, 0x252A, 0x4246, 0x8082}},
    {14771, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x40, 0x844, 0x844, 0x7F44, 0x828, 0x82A, 0xF12, 0xF02A, 0x4046, 0x82}},
    {14772, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x2040, 0x3E40, 0x4244, 0xB444, 0xC44, 0x3228, 0xC128, 0x7E12, 0x4232, 0x424A, 0x7E86, 0x4302}},
    {14773, 2, {0x828, 0x4824, 0x4824, 0x7F20, 0x482E, 0x88F0, 0x820, 0xFFA4, 0x1824, 0x1C28, 0x2A28, 0x2910, 0x4912, 0x882A, 0x846, 0x882}},
    {14774, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x40, 0x7F40, 0x1044, 0x2244, 0x7F44, 0x928, 0x3E28, 0x812, 0xF32, 0xF04A, 0x86, 0x302}},
    {14775, 2, {0x28, 0xFC24, 0x424, 0x6820, 0x102E, 0xFEF0, 0x9220, 0x9224, 0xFE24, 0x9228, 0x9228, 0xFE10, 0x9212, 0x922A, 0x9246, 0x8682}},
    {14776, 2, {0x828, 0x824, 0xFA4, 0x820, 0x7F2E, 0x41F0, 0x7F20, 0x4124, 0x7F24, 0x4928, 0x828, 0xFF90, 0x812, 0x82A, 0x846, 0x882}},
    {14777, 2, {0x20, 0x7C28, 0x4424, 0x7C24, 0x4420, 0x7C3E, 0x1E0, 0xFE24, 0x4024, 0x7E28, 0xAA28, 0x2A10, 0x4A32, 0x924A, 0x2A86, 0x4502}},
    {14778, 2, {0x28, 0x7FA4, 0x4024, 0x5F20, 0x402E, 0x7FF0, 0x5520, 0x5224, 0x5924, 0x50A8, 0x4228, 0x7F90, 0x5212, 0x4A2A, 0x8246, 0x682}},
    {14779, 2, {0x28, 0xEF24, 0x2124, 0xA920, 0x652E, 0xA9F0, 0x2120, 0x1024, 0x7E24, 0x4228, 0x4228, 0x7E10, 0x4212, 0x422A, 0x7E46, 0x4282}},
    {14780, 2, {0x828, 0x824, 0x7F24, 0x1020, 0x3E2E, 0x1070, 0xFFA0, 0x2224, 0x4124, 0x90A8, 0x2628, 0x2210, 0x3612, 0x222A, 0x3E46, 0x2282}},
    {14781, 2, {0x2014, 0x2012, 0xFC10, 0x23FE, 0x4010, 0x7BD0, 0x4A50, 0xCBD2, 0x7812, 0x4BD2, 0x4814, 0x7AAC, 0x4D4A, 0x4AAA, 0x4816, 0x5822}},
    {14782, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2480, 0x24F0, 0x2790, 0x3C90, 0x2492, 0x44B2, 0x4402, 0x83FE, 0x0}},
    {14783, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2080, 0x2FF8, 0x21C0, 0x22A0, 0x22A0, 0x4490, 0x4888, 0x9084, 0x80}},
    {14784, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2FF0, 0x2210, 0x2220, 0x2338, 0x2288, 0x4450, 0x4420, 0x88D8, 0x1306}},
    {14785, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x2080, 0x3FFE, 0x2100, 0x4210, 0x4408, 0x8FFC, 0x404}},
    {14786, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x2FFC, 0x2884, 0x2884, 0x2FFC, 0x4884, 0x4884, 0x8FFC, 0x804}},
    {14787, 2, {0x1004, 0x1004, 0x1084, 0x1084, 0xFE84, 0x1084, 0x1084, 0x1284, 0x1C8C, 0x30B4, 0xD0C4, 0x1084, 0x1004, 0x1004, 0x5004, 0x2004}},
    {14788, 2, {0x1000, 0x1000, 0x1104, 0x1104, 0xFD04, 0x1104, 0x1104, 0x1504, 0x1904, 0x3104, 0xD104, 0x1104, 0x1104, 0x11FC, 0x5004, 0x2000}},
    {14789, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0xFE44, 0x1044, 0x1044, 0x1244, 0x1C44, 0x3044, 0xD044, 0x1084, 0x1084, 0x1104, 0x5228, 0x2410}},
    {14790, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x1124, 0x1524, 0x1924, 0x3124, 0xD124, 0x1134, 0x1128, 0x1020, 0x5020, 0x2020}},
    {14791, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x1104, 0x1504, 0x1974, 0x3104, 0xD104, 0x1104, 0x1104, 0x1104, 0x5114, 0x2108}},
    {14792, 2, {0x1000, 0x11FC, 0x1004, 0x1004, 0xFC04, 0x10FC, 0x1080, 0x1480, 0x1900, 0x31FC, 0xD004, 0x1004, 0x1004, 0x1004, 0x5028, 0x2010}},
    {14793, 2, {0x1100, 0x1100, 0x11FC, 0x1200, 0xFC00, 0x11F8, 0x1000, 0x1400, 0x1BF8, 0x3008, 0xD008, 0x1008, 0x100A, 0x100A, 0x5006, 0x2002}},
    {14794, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x11FC, 0x1524, 0x1924, 0x3124, 0xD124, 0x1134, 0x1128, 0x1020, 0x5020, 0x2020}},
    {14795, 2, {0x1040, 0x1048, 0x1044, 0x1044, 0xFC40, 0x13FE, 0x1040, 0x1440, 0x1840, 0x30A0, 0xD0A0, 0x1090, 0x1110, 0x1108, 0x5204, 0x2402}},
    {14796, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x103C, 0x11E0, 0x1420, 0x1820, 0x303E, 0xD3E0, 0x1020, 0x1022, 0x1022, 0x5022, 0x201E}},
    {14797, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC80, 0x1080, 0x1080, 0x14FC, 0x1884, 0x3084, 0xD084, 0x1084, 0x1104, 0x1104, 0x5228, 0x2410}},
    {14798, 2, {0x1000, 0x11FE, 0x1008, 0x1088, 0xFC88, 0x1088, 0x1108, 0x15FE, 0x1818, 0x3028, 0xD048, 0x1088, 0x1108, 0x1208, 0x5028, 0x2010}},
    {14799, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFA94, 0x2254, 0x2224, 0x2A24, 0x3254, 0xE254, 0x2294, 0x2304, 0x2204, 0x2204, 0xA214, 0x4208}},
    {14800, 2, {0x2000, 0x2000, 0x27BC, 0x20A4, 0xF8A4, 0x24A4, 0x22A4, 0x2AA8, 0x3128, 0xE128, 0x2290, 0x2290, 0x24A8, 0x2828, 0xA044, 0x4082}},
    {14801, 2, {0x1020, 0x1120, 0x10A0, 0x10A0, 0xFC20, 0x13FE, 0x1022, 0x1422, 0x1842, 0x3052, 0xD04A, 0x108A, 0x1082, 0x1102, 0x5214, 0x2408}},
    {14802, 2, {0x2040, 0x2020, 0x2020, 0x23FE, 0xFA02, 0x2404, 0x2000, 0x2890, 0x3090, 0xE090, 0x2088, 0x2108, 0x2108, 0x2104, 0xA204, 0x4402}},
    {14803, 2, {0x1008, 0x103C, 0x13D0, 0x1290, 0xFE90, 0x1290, 0x1290, 0x1690, 0x1A90, 0x3288, 0xD288, 0x12C8, 0x12A4, 0x14D4, 0x5492, 0x2800}},
    {14804, 2, {0x2028, 0x2024, 0x2024, 0x2020, 0xFBFE, 0x2220, 0x2224, 0x2A24, 0x3224, 0xE228, 0x2228, 0x2210, 0x2212, 0x242A, 0xA446, 0x4882}},
    {14805, 2, {0x2000, 0x2080, 0x231E, 0x2252, 0xFA52, 0x2252, 0x2252, 0x2A52, 0x3252, 0xE2D2, 0x235A, 0x2254, 0x2090, 0x2090, 0xA110, 0x4210}},
    {14806, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x1100, 0x1500, 0x197C, 0x3108, 0xD110, 0x1120, 0x1142, 0x1282, 0x527E, 0x2400}},
    {14807, 2, {0x1090, 0x1090, 0x1090, 0x1290, 0xFE92, 0x12D4, 0x1298, 0x1690, 0x1A90, 0x3290, 0xD290, 0x1292, 0x12D2, 0x1712, 0x520E, 0x2000}},
    {14808, 2, {0x880, 0x888, 0x2E90, 0x28E4, 0x2884, 0x2E7C, 0xF000, 0xE0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14809, 2, {0x1000, 0x1040, 0x1020, 0x1028, 0xFC08, 0x1088, 0x1490, 0x1A94, 0x32A2, 0xD2A2, 0x14C2, 0x1088, 0x1188, 0x1288, 0x5478, 0x2000}},
    {14810, 2, {0x1000, 0x1000, 0x13FC, 0x1100, 0xFD00, 0x11F0, 0x1110, 0x1510, 0x1990, 0x3250, 0xD250, 0x1210, 0x1212, 0x1292, 0x5312, 0x220E}},
    {14811, 2, {0x440, 0x2440, 0x247C, 0x2480, 0x2490, 0x2508, 0x408, 0xE0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14812, 2, {0x1000, 0x13FE, 0x1210, 0x1210, 0xFE10, 0x12FE, 0x1292, 0x1692, 0x1A92, 0x3292, 0xD29A, 0x1294, 0x1210, 0x1210, 0x53FE, 0x2000}},
    {14813, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x2244, 0x2A7C, 0x4400, 0x80E0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14814, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x11FE, 0x1040, 0x1440, 0x187C, 0x30A4, 0xD0A8, 0x1128, 0x1110, 0x1228, 0x5444, 0x2082}},
    {14815, 2, {0x2200, 0x217E, 0x2140, 0x2440, 0xFA40, 0x2240, 0x2040, 0x2940, 0x3140, 0xE240, 0x2640, 0x2240, 0x2240, 0x2240, 0xA27E, 0x4000}},
    {14816, 2, {0x1020, 0x1040, 0x1088, 0x1104, 0xFBFE, 0x1002, 0x1420, 0x1820, 0x3128, 0xD124, 0x1224, 0x1222, 0x1422, 0x1020, 0x50A0, 0x2040}},
    {14817, 2, {0x1020, 0x1010, 0x1010, 0x13FE, 0xFC20, 0x1042, 0x1084, 0x15F8, 0x1810, 0x3022, 0xD0C4, 0x1308, 0x1010, 0x1028, 0x50C4, 0x2302}},
    {14818, 2, {0x2000, 0x27FC, 0x2404, 0x2444, 0xFC44, 0x2444, 0x27FC, 0x2C44, 0x3444, 0xE4A4, 0x2494, 0x2514, 0x2604, 0x2404, 0xA7FC, 0x4404}},
    {14819, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x1020, 0x13FE, 0x1870, 0x30A8, 0xD0A8, 0x1124, 0x1124, 0x1222, 0x5020, 0x2020}},
    {14820, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC40, 0x1088, 0x1504, 0x1BFE, 0x3092, 0xD090, 0x1090, 0x1090, 0x1112, 0x1112, 0x520E, 0x2400}},
    {14821, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x1020, 0x13FE, 0x1890, 0x3090, 0xD090, 0x1090, 0x1112, 0x1112, 0x5212, 0x240E}},
    {14822, 2, {0x2080, 0x2080, 0x2100, 0x21FC, 0xFA04, 0x2404, 0x23E4, 0x2A24, 0x3224, 0xE3E4, 0x2224, 0x2224, 0x23E4, 0x2004, 0xA028, 0x4010}},
    {14823, 2, {0x1040, 0x1020, 0x13FC, 0x1040, 0xFC90, 0x1108, 0x13FC, 0x1004, 0x1950, 0x3150, 0xD150, 0x1150, 0x1250, 0x1252, 0x5452, 0x280E}},
    {14824, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x11FE, 0x1100, 0x1500, 0x197C, 0x3144, 0xD144, 0x1144, 0x1144, 0x127C, 0x5244, 0x2400}},
    {14825, 2, {0x2000, 0x27FC, 0x2404, 0x2404, 0xF5F4, 0x2444, 0x2444, 0x25F4, 0x3554, 0xE554, 0x2554, 0x2554, 0x2574, 0x2446, 0xA446, 0x4842}},
    {14826, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0xFA08, 0x11FE, 0x1500, 0x197C, 0x3144, 0xD144, 0x1154, 0x1148, 0x1142, 0x1242, 0x523E, 0x2400}},
    {14827, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFC40, 0x11FC, 0x1154, 0x1554, 0x1954, 0x3154, 0xD154, 0x1154, 0x1154, 0x1144, 0x5114, 0x2108}},
    {14828, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1D52, 0xF20E, 0x4400, 0xE0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14829, 2, {0x7CFC, 0x1088, 0x1090, 0x1088, 0x1E84, 0xF094, 0x4088, 0xE0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14830, 2, {0x1000, 0x1040, 0x119C, 0x1104, 0xFD04, 0x1104, 0x1104, 0x15DC, 0x1904, 0x3104, 0xD104, 0x1104, 0x1104, 0x11FC, 0x5104, 0x2000}},
    {14831, 2, {0x1040, 0x1040, 0x1044, 0x11F4, 0xFC48, 0x1050, 0x13FE, 0x1040, 0x1880, 0x3184, 0xD298, 0x14E0, 0x1082, 0x1082, 0x507E, 0x2000}},
    {14832, 2, {0x1008, 0x103C, 0x13C0, 0x1200, 0xFE20, 0x1220, 0x1220, 0x13FE, 0x1820, 0x3020, 0xD128, 0x1124, 0x1222, 0x1422, 0x50A0, 0x2040}},
    {14833, 2, {0x808, 0x81C, 0x1470, 0x2210, 0x4110, 0xBEFC, 0x10, 0x10, 0x3E10, 0x22FE, 0x2210, 0x2210, 0x2210, 0x3E10, 0x2250, 0x20}},
    {14834, 2, {0x2080, 0x2040, 0x27FC, 0x2404, 0xF848, 0x2040, 0x2040, 0x2BFC, 0x30E0, 0xE150, 0x2150, 0x2248, 0x2444, 0x2842, 0xA040, 0x4040}},
    {14835, 2, {0x1088, 0xFD04, 0x12FA, 0x1C48, 0xF088, 0x1128, 0x3210, 0xE0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14836, 2, {0x1010, 0x10D8, 0x1394, 0x1094, 0xFC90, 0x13FE, 0x1090, 0x1494, 0x1894, 0x30D8, 0xD398, 0x1090, 0x10AA, 0x10CA, 0x5286, 0x2102}},
    {14837, 2, {0x1000, 0x13FE, 0x1010, 0x1020, 0xFC68, 0x10A4, 0x1322, 0x1020, 0x1800, 0x31FC, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {14838, 2, {0x1040, 0x1080, 0x11F8, 0x1108, 0xFDF8, 0x1108, 0x11FA, 0x150A, 0x190C, 0x33F8, 0xD018, 0x1028, 0x1048, 0x1188, 0x5628, 0x2010}},
    {14839, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x1050, 0x1488, 0x1904, 0x32FA, 0xD488, 0x1088, 0x1088, 0x1088, 0x50F8, 0x2088}},
    {14840, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDF8, 0x1108, 0x1108, 0x15F8, 0x1800, 0x33FC, 0xD008, 0x1030, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {14841, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1000, 0x15FE, 0x1808, 0x33FE, 0xD008, 0x1108, 0x1088, 0x1008, 0x5028, 0x2010}},
    {14842, 2, {0x1020, 0x1022, 0x13B4, 0x10A8, 0xFCA8, 0x1124, 0x12A2, 0x1040, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {14843, 2, {0x1000, 0x13FE, 0x1040, 0x1080, 0xFD44, 0x1224, 0x1068, 0x12B0, 0x1930, 0x32A8, 0xD068, 0x10A4, 0x1122, 0x1220, 0x50A0, 0x2040}},
    {14844, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x1020, 0x1420, 0x1BFE, 0x3040, 0xD0A4, 0x11A8, 0x1290, 0x1488, 0x50C6, 0x2080}},
    {14845, 2, {0x1000, 0x11FC, 0x1104, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1524, 0x1974, 0x3154, 0xD154, 0x1174, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {14846, 2, {0x1080, 0x1080, 0x10FC, 0x1154, 0xFA54, 0x10A4, 0x1524, 0x1A44, 0x3094, 0xD108, 0x1040, 0x10A4, 0x12AA, 0x128A, 0x5478, 0x2000}},
    {14847, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFD54, 0x1222, 0x1442, 0x1020, 0x1820, 0x3120, 0xD13C, 0x1120, 0x12A0, 0x1260, 0x543E, 0x2800}},
    {14848, 2, {0x103C, 0x11E0, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x1424, 0x1BFE, 0x3024, 0xD1FC, 0x1070, 0x10A8, 0x1124, 0x1622, 0x5020, 0x2020}},
    {14849, 2, {0x1040, 0x1040, 0x13FC, 0x1040, 0xFC40, 0x11F8, 0x1108, 0x15F8, 0x1908, 0x31F8, 0xD108, 0x11F8, 0x1108, 0x1108, 0x57FE, 0x2000}},
    {14850, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x11FC, 0x1524, 0x19FC, 0x3020, 0xD070, 0x10A8, 0x1124, 0x1622, 0x5020, 0x2020}},
    {14851, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x11FC, 0x1400, 0x19FC, 0x3008, 0xD010, 0x13FE, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {14852, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0xFDF4, 0x2444, 0x2444, 0x2FFC, 0x3404, 0xE5F4, 0x2514, 0x2514, 0x25F4, 0x2404, 0xA414, 0x4808}},
    {14853, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCA8, 0x1020, 0x11FC, 0x1524, 0x1924, 0x3154, 0xD14C, 0x1184, 0x1104, 0x1104, 0x5114, 0x2108}},
    {14854, 2, {0x2110, 0x2110, 0x2110, 0x2110, 0xF7BC, 0x2110, 0x2110, 0x2338, 0x33B8, 0xE554, 0x2554, 0x2992, 0x2110, 0x2110, 0xA110, 0x4110}},
    {14855, 2, {0x1020, 0x49FC, 0x2124, 0x9F8, 0x7148, 0x1230, 0x144C, 0xE0, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {14856, 2, {0x1040, 0x1020, 0x17FE, 0x1108, 0xF890, 0x1060, 0x1198, 0x1E06, 0x3108, 0xD1F8, 0x1108, 0x11F8, 0x1108, 0x1108, 0x5208, 0x2408}},
    {14857, 2, {0x1000, 0x13FC, 0x1204, 0x13FC, 0xFE20, 0x13FE, 0x1210, 0x128A, 0x1B06, 0x3000, 0xD3FC, 0x1204, 0x13FC, 0x1204, 0x53FC, 0x2204}},
    {14858, 2, {0x1040, 0x1040, 0x13FE, 0x1080, 0xFDFC, 0x1220, 0x15FE, 0x1000, 0x19FC, 0x3104, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x5114, 0x2108}},
    {14859, 2, {0x1040, 0x1044, 0x13F4, 0x1048, 0xF850, 0x17FE, 0x1040, 0x1880, 0x31F8, 0xD308, 0x1508, 0x11F8, 0x1108, 0x1108, 0x51F8, 0x2108}},
    {14860, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x48, 0x7F86, 0x4100, 0x82FC, 0x3C44, 0x844, 0x7E28, 0x828, 0x7F10, 0x828, 0x2844, 0x1182}},
    {14861, 2, {0x808, 0x3EFE, 0x808, 0x7F48, 0x828, 0x3E08, 0x828, 0x7F10, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {14862, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0xFDFC, 0x1154, 0x1154, 0x1554, 0x1954, 0x312C, 0xD020, 0x13FE, 0x1050, 0x1088, 0x5104, 0x2202}},
    {14863, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1080, 0x1480, 0x1BFE, 0x3090, 0xD110, 0x1152, 0x1254, 0x12A8, 0x1428, 0x5844, 0x2082}},
    {14864, 2, {0x2100, 0x217C, 0x2144, 0x2244, 0xF244, 0x267C, 0x2A10, 0x2210, 0x32FE, 0xE238, 0x2254, 0x2254, 0x2294, 0x2312, 0xA210, 0x4210}},
    {14865, 2, {0x1000, 0x1088, 0x1252, 0x1222, 0xFE52, 0x128A, 0x1202, 0x17FE, 0x1888, 0x3144, 0xD27A, 0x1088, 0x1150, 0x1020, 0x50D8, 0x2706}},
    {14866, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x1326, 0x11FC, 0x1420, 0x1820, 0x33FE, 0xD000, 0x1020, 0x11FC, 0x1020, 0x5020, 0x23FE}},
    {14867, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC88, 0x1104, 0x1222, 0x1120, 0x19FC, 0x3220, 0xD020, 0x17FE, 0x1020, 0x1020, 0x5020, 0x2020}},
    {14868, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0xF410, 0x2410, 0x25D0, 0x2412, 0x3412, 0xE5D4, 0x2554, 0x2548, 0x25DA, 0x242A, 0xA846, 0x5082}},
    {14869, 2, {0x1000, 0x13F8, 0x1108, 0x1110, 0xF91C, 0x12E4, 0x12A4, 0x1A54, 0x34A8, 0xD100, 0x13FC, 0x1294, 0x1294, 0x1294, 0x57FE, 0x2000}},
    {14870, 2, {0x2008, 0x2008, 0x27C8, 0x2210, 0xFA1E, 0x24A4, 0x27D4, 0x2914, 0x3114, 0xE7D4, 0x2114, 0x2108, 0x21C8, 0x2E14, 0xA414, 0x4022}},
    {14871, 2, {0x1040, 0x1040, 0x107E, 0x1040, 0xFBFC, 0x1004, 0x15FC, 0x1804, 0x33FC, 0xD040, 0x1240, 0x127C, 0x1240, 0x1540, 0x54FE, 0x2800}},
    {14872, 2, {0x1020, 0x1020, 0x10A4, 0x10A2, 0xFD2A, 0x1010, 0x1460, 0x1980, 0x36FC, 0xD084, 0x10FC, 0x1084, 0x10FC, 0x1084, 0x50FC, 0x2084}},
    {14873, 2, {0x101E, 0x13E0, 0x1122, 0x1094, 0xFC40, 0x1088, 0x11F0, 0x1420, 0x18C4, 0x31FE, 0xD022, 0x1020, 0x13FE, 0x1050, 0x5088, 0x2306}},
    {14874, 2, {0x1090, 0x1090, 0x13FE, 0x1090, 0xFD00, 0x11FC, 0x1204, 0x11E4, 0x1924, 0x31E4, 0xD124, 0x11E4, 0x1124, 0x1004, 0x5028, 0x2010}},
    {14875, 2, {0x1000, 0x13DE, 0x1252, 0x13DE, 0xFA52, 0x13DE, 0x1202, 0x1A22, 0x3222, 0xD222, 0x1252, 0x124A, 0x128A, 0x1202, 0x520A, 0x2204}},
    {14876, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFA42, 0x1278, 0x13C4, 0x1A3C, 0x3240, 0xD220, 0x13FE, 0x1288, 0x1250, 0x1420, 0x5458, 0x2986}},
    {14877, 2, {0x1000, 0x13FE, 0x1202, 0x1202, 0xFBFE, 0x1210, 0x1292, 0x1A54, 0x32FE, 0xD282, 0x12FE, 0x1282, 0x12FE, 0x1482, 0x548A, 0x2884}},
    {14878, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x13FE, 0x1440, 0x1888, 0x31F0, 0xD024, 0x13FE, 0x1022, 0x1124, 0x52A2, 0x2040}},
    {14879, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1000, 0x13FE, 0x1400, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1088, 0x1050, 0x53FE, 0x2000}},
    {14880, 2, {0x1000, 0x13FE, 0x1202, 0x1000, 0xFDFC, 0x1104, 0x11FC, 0x1504, 0x19FC, 0x3040, 0xD020, 0x13FE, 0x1000, 0x1088, 0x5104, 0x2202}},
    {14881, 2, {0x2000, 0x27FE, 0x2402, 0x25FA, 0xFC42, 0x24C2, 0x252A, 0x2C6A, 0x34B2, 0xE52A, 0x246A, 0x24AA, 0x2522, 0x2442, 0xA7FE, 0x4402}},
    {14882, 2, {0x2110, 0x2212, 0x2454, 0x27D8, 0xF810, 0x2012, 0x27D2, 0x2C4E, 0x3440, 0xE7D2, 0x2454, 0x2458, 0x27D0, 0x2452, 0xA552, 0x448E}},
    {14883, 2, {0x2120, 0x2110, 0x2208, 0x2486, 0xF910, 0x23F8, 0x2008, 0x2800, 0x37BC, 0xE084, 0x24A4, 0x2294, 0x24A4, 0x2084, 0xA294, 0x4108}},
    {14884, 2, {0x2000, 0x27FE, 0x2450, 0x2488, 0xFDFE, 0x2690, 0x24FC, 0x2C90, 0x34FC, 0xE490, 0x2490, 0x24FE, 0x2480, 0x2400, 0xA7FE, 0x4000}},
    {14885, 2, {0x2008, 0x2108, 0x2108, 0x2208, 0xF2BE, 0x24AA, 0x2F2A, 0x212A, 0x322A, 0xE4AA, 0x2FB6, 0x20A2, 0x2022, 0x2AA2, 0xAAAA, 0x4824}},
    {14886, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x11FC, 0x1520, 0x19FE, 0x3100, 0xD3F8, 0x1108, 0x1090, 0x1060, 0x5198, 0x2606}},
    {14887, 2, {0x1040, 0x1048, 0x1084, 0x11FE, 0xFC20, 0x13FE, 0x1088, 0x1504, 0x1BFE, 0x3524, 0xD124, 0x11FC, 0x1124, 0x1124, 0x51FC, 0x2104}},
    {14888, 2, {0x1040, 0x13BE, 0x1212, 0x1292, 0xFE52, 0x12AA, 0x1324, 0x1040, 0x1BFE, 0x3222, 0xD222, 0x13FE, 0x1222, 0x1222, 0x53FE, 0x2202}},
    {14889, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFAEC, 0x1224, 0x1244, 0x1A44, 0x32EE, 0xD224, 0x12A4, 0x12A4, 0x124E, 0x1460, 0x549E, 0x2900}},
    {14890, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1020, 0x1524, 0x18A8, 0x3020, 0xD3FE, 0x1090, 0x1090, 0x1112, 0x5212, 0x240E}},
    {14891, 2, {0x2090, 0x2290, 0x229C, 0x22A0, 0xFAD0, 0x2288, 0x2288, 0x2880, 0x3000, 0xE3F8, 0x22A8, 0x22A8, 0x22A8, 0x22A8, 0xAFFE, 0x4000}},
    {14892, 2, {0x2040, 0x2040, 0x23F8, 0x2040, 0xF7FE, 0x2000, 0x23F8, 0x2208, 0x3208, 0xE3F8, 0x20A4, 0x2128, 0x2310, 0x2D48, 0xA186, 0x4100}},
    {14893, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0xFBFC, 0x1200, 0x12F8, 0x1A88, 0x32F8, 0xD200, 0x13FC, 0x1324, 0x15FC, 0x1502, 0x5902, 0x20FE}},
    {14894, 2, {0x2020, 0x2240, 0x24EC, 0x24A4, 0xF4A4, 0x24E4, 0x24A4, 0x24A4, 0x36E4, 0xE40C, 0x2110, 0x2110, 0x2110, 0x2112, 0xA212, 0x440E}},
    {14895, 2, {0x2200, 0x211E, 0x27D2, 0x2012, 0xF7D4, 0x2454, 0x27D8, 0x2014, 0x37D2, 0xE092, 0x2112, 0x21DA, 0x2714, 0x2110, 0xA510, 0x4210}},
    {14896, 2, {0x1040, 0x1020, 0x13FE, 0x1088, 0xFD24, 0x1242, 0x1090, 0x15F8, 0x1808, 0x3044, 0xD0A8, 0x1190, 0x1288, 0x14A4, 0x50C2, 0x2080}},
    {14897, 2, {0x2040, 0x20FC, 0x2108, 0x2650, 0xF920, 0x20C0, 0x2320, 0x2FFE, 0x3000, 0xE3FC, 0x2000, 0x23FC, 0x2000, 0x23FC, 0xA204, 0x43FC}},
    {14898, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF840, 0x2020, 0x27FC, 0x2C04, 0x3444, 0xE040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0xAC46, 0x4040}},
    {14899, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC20, 0x11FC, 0x1050, 0x1488, 0x1BFE, 0x3008, 0xD1E8, 0x1128, 0x11E8, 0x1008, 0x5028, 0x2010}},
    {14900, 2, {0x2110, 0x2110, 0x2110, 0x2FDE, 0xF120, 0x27C0, 0x245C, 0x27C4, 0x3448, 0xE7C8, 0x2110, 0x2FD0, 0x2122, 0x2122, 0xA11E, 0x4100}},
    {14901, 2, {0x1108, 0x1090, 0x17FE, 0x1000, 0xFBC4, 0x1254, 0x13D4, 0x1A54, 0x33D4, 0xD244, 0x12CC, 0x1000, 0x13FC, 0x1084, 0x5114, 0x2608}},
    {14902, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x1500, 0x19FE, 0x3100, 0xD1FE, 0x1002, 0x12AA, 0x12AA, 0x5202, 0x200C}},
    {14903, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x1020, 0x11FC, 0x1524, 0x19FC, 0x3020, 0xD3FE, 0x1020, 0x11FC, 0x1020, 0x53FE, 0x2000}},
    {14904, 2, {0x1040, 0x1020, 0x13FE, 0x1242, 0xFC28, 0x1294, 0x12AA, 0x14CA, 0x18F8, 0x3300, 0xD020, 0x1124, 0x1124, 0x1124, 0x51FC, 0x2004}},
    {14905, 2, {0x1088, 0x13DE, 0x1088, 0x11DC, 0xFC88, 0x13DE, 0x1088, 0x1400, 0x1BFC, 0x3004, 0xD004, 0x11FC, 0x1004, 0x1004, 0x53FC, 0x2004}},
    {14906, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x10F8, 0x1420, 0x1BFE, 0x3222, 0xD332, 0x12AA, 0x1376, 0x1222, 0x1222, 0x522A, 0x2224}},
    {14907, 2, {0x804, 0x7F78, 0x840, 0x3E40, 0x2A7E, 0x3E48, 0x2A48, 0xFF48, 0x8A8, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {14908, 2, {0x40, 0x7CF8, 0x5548, 0x5430, 0x7DCE, 0x54F8, 0x5488, 0x7CF8, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {14909, 2, {0x2004, 0x27C4, 0x2004, 0x2394, 0xFA94, 0x2394, 0x2014, 0x2FD4, 0x3554, 0xE554, 0x27D4, 0x2554, 0x2544, 0x27C4, 0xA454, 0x4008}},
    {14910, 2, {0x2208, 0x2110, 0x27FC, 0x2040, 0xFBF8, 0x2040, 0x27FC, 0x2840, 0x3020, 0xE1C0, 0x2044, 0x2768, 0x2150, 0x2248, 0xAD46, 0x4080}},
    {14911, 2, {0x2040, 0x207C, 0x2040, 0x27FE, 0xF442, 0x2478, 0x25C2, 0x243E, 0x3500, 0xE538, 0x25A8, 0x26AA, 0x2AA6, 0x2940, 0xB2FE, 0x4400}},
    {14912, 2, {0x2040, 0x207C, 0x2040, 0x27FE, 0xF442, 0x2478, 0x25C2, 0x24BE, 0x3480, 0xE5FC, 0x2554, 0x26D4, 0x24B4, 0x292E, 0xAA24, 0x504C}},
    {14913, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFCA0, 0x1090, 0x11FE, 0x1520, 0x1B20, 0x31FC, 0xD120, 0x11FC, 0x1120, 0x1120, 0x51FE, 0x2100}},
    {14914, 2, {0x2204, 0x2204, 0x2784, 0xF494, 0x290C, 0x27C4, 0x2564, 0x3554, 0xE7C4, 0x2546, 0x257C, 0x27C4, 0x2544, 0x2544, 0xA444, 0x48C4}},
    {14915, 2, {0x2040, 0x2020, 0x27FE, 0x2492, 0xFBFC, 0x2090, 0x23FC, 0x2890, 0x37FE, 0xE108, 0x2204, 0x2462, 0x2010, 0x2180, 0xA060, 0x4010}},
    {14916, 2, {0x1020, 0x1124, 0x1124, 0x11FC, 0xFC00, 0x11FC, 0x1124, 0x15FC, 0x1924, 0x31FC, 0xD040, 0x1024, 0x12A2, 0x128A, 0x5478, 0x2000}},
    {14917, 2, {0x1000, 0x13FE, 0x1050, 0x13DE, 0xFE52, 0x1252, 0x13DE, 0x1050, 0x1BFE, 0x3222, 0xD222, 0x13FE, 0x1222, 0x1222, 0x53FE, 0x2202}},
    {14918, 2, {0x2214, 0x2112, 0x2FD2, 0x2010, 0xF07E, 0x2790, 0x2490, 0x2490, 0x3790, 0xE128, 0x25A8, 0x2568, 0x2928, 0x214A, 0xA54A, 0x4286}},
    {14919, 2, {0x1020, 0x11FC, 0x1024, 0x13FE, 0xFC24, 0x11FC, 0x1420, 0x1BFE, 0x3000, 0xD1FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x5000, 0x23FE}},
    {14920, 2, {0x1040, 0x1020, 0x13FE, 0x128A, 0xFD04, 0x11FC, 0x1444, 0x1BFE, 0x3044, 0xD1FC, 0x1080, 0x11FC, 0x1284, 0x1484, 0x50FC, 0x2084}},
    {14921, 2, {0x2040, 0x20A0, 0x2118, 0x26E6, 0xF800, 0x23F8, 0x2208, 0x2BF8, 0x3000, 0xE7BC, 0x2084, 0x24A4, 0x2294, 0x24A4, 0xA294, 0x4108}},
    {14922, 2, {0x101C, 0x13E0, 0x1044, 0x1124, 0xFC88, 0x11FC, 0x1154, 0x1554, 0x19FC, 0x3008, 0xD7FE, 0x1108, 0x1088, 0x1088, 0x5028, 0x2010}},
    {14923, 2, {0x2040, 0x23F8, 0x2048, 0x27FE, 0xF048, 0x23F8, 0x2040, 0x2554, 0x375C, 0xE444, 0x27FC, 0x2444, 0x275C, 0x2554, 0xA554, 0x4844}},
    {14924, 2, {0x1090, 0x17FC, 0x1094, 0x13FC, 0xFA90, 0x13FE, 0x1112, 0x1BFA, 0x350E, 0xD1F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x5090, 0x2108}},
    {14925, 2, {0x1210, 0x1210, 0x13DE, 0x1528, 0xFCA4, 0x1020, 0x13FE, 0x1020, 0x1BFE, 0x3222, 0xD22A, 0x1274, 0x10A8, 0x1124, 0x5622, 0x2020}},
    {14926, 2, {0x2008, 0x2788, 0x248C, 0x278A, 0xF488, 0x27BE, 0x2008, 0x2788, 0x3488, 0xE788, 0x2494, 0x2794, 0x2494, 0x2494, 0xA494, 0x45A2}},
    {14927, 2, {0x1000, 0x11FC, 0x1124, 0x11AC, 0xFD74, 0x1124, 0x11FC, 0x1420, 0x19FC, 0x3020, 0xD3FE, 0x1000, 0x12A4, 0x1252, 0x5452, 0x2000}},
    {14928, 2, {0x2210, 0x2210, 0x23DE, 0x2528, 0xF8C4, 0x2040, 0x23FC, 0x2840, 0x37FC, 0xE010, 0x2FFE, 0x2010, 0x2210, 0x2110, 0xA050, 0x4020}},
    {14929, 2, {0x2000, 0x23F8, 0x2080, 0x27FC, 0xF110, 0x2208, 0x2CA6, 0x20A0, 0x37BC, 0xE0A0, 0x23B8, 0x20A0, 0x27BC, 0x20A0, 0xA0A0, 0x40A0}},
    {14930, 2, {0x2110, 0x2110, 0x27BC, 0x2110, 0xFBB8, 0x2554, 0x2912, 0x2800, 0x33F8, 0xE000, 0x27FC, 0x2040, 0x2248, 0x2444, 0xA942, 0x4080}},
    {14931, 2, {0x2C78, 0x4448, 0x6C48, 0x4486, 0x7D78, 0x48, 0x7C30, 0x1048, 0xFCA4, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {14932, 2, {0x4000, 0x4FDC, 0x4854, 0x4854, 0xEFD4, 0x4A94, 0x4AA6, 0x4FC0, 0x6A9C, 0xCA94, 0x4FD4, 0x4814, 0x5288, 0x5248, 0x6454, 0xC822}},
    {14933, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0xFA22, 0x11AC, 0x1420, 0x19AC, 0x3050, 0xD0C8, 0x1326, 0x10F8, 0x1008, 0x10D0, 0x5020, 0x2010}},
    {14934, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0xFD24, 0x11FC, 0x1524, 0x1BFE, 0x3000, 0xD1FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x5000, 0x23FE}},
    {14935, 2, {0x2444, 0x27FC, 0x2120, 0x23FE, 0xF620, 0x2BFC, 0x2220, 0x23FC, 0x3220, 0xE3FE, 0x2200, 0x27FC, 0x2108, 0x211E, 0xA202, 0x4C1C}},
    {14936, 2, {0x1104, 0x1088, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x13FE, 0x1454, 0x1992, 0x3090, 0xD3FE, 0x1090, 0x10D4, 0x138A, 0x5096, 0x21A2}},
    {14937, 2, {0x2080, 0x2040, 0x27FE, 0x2512, 0xFBFC, 0x2110, 0x27FC, 0x2910, 0x37FE, 0xE110, 0x2248, 0x25F4, 0x2842, 0x2040, 0xA7FC, 0x4000}},
    {14938, 2, {0x2110, 0x27FC, 0x2110, 0x27FC, 0xF404, 0x23F8, 0x2000, 0x27FC, 0x3080, 0xE144, 0x26A8, 0x2170, 0x26A8, 0x2126, 0xA6A0, 0x4040}},
    {14939, 2, {0x27FC, 0x2444, 0x27FC, 0xF444, 0x25F4, 0x2554, 0x25F4, 0x3554, 0xE5F4, 0x244C, 0x27FC, 0x2414, 0x27FC, 0x2514, 0xA4B4, 0x47FC}},
    {14940, 2, {0x2010, 0x27D0, 0x2510, 0x27DE, 0xF450, 0x27E8, 0x2504, 0x27C4, 0x3000, 0xE3F8, 0x22A8, 0x22A8, 0x22A8, 0x22A8, 0xAFFE, 0x4000}},
    {14941, 2, {0x2040, 0x27FE, 0x2000, 0x21F8, 0xF908, 0x21F8, 0x2000, 0x2FFE, 0x3402, 0xE3F8, 0x2180, 0x2644, 0x21B8, 0x2668, 0xA1A6, 0x4E60}},
    {14942, 2, {0x2110, 0x2554, 0x25B8, 0x2910, 0xF2A8, 0x2444, 0x2000, 0x27FC, 0x3444, 0xE040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0xAC46, 0x4040}},
    {14943, 2, {0x1040, 0x1020, 0x13FE, 0x1252, 0xFDFC, 0x1050, 0x11FC, 0x1450, 0x1BFE, 0x3088, 0xD124, 0x12FA, 0x1020, 0x1124, 0x52A2, 0x2040}},
    {14944, 2, {0x2080, 0x2040, 0x27FC, 0x24A0, 0xF7FC, 0x24A4, 0x27FC, 0x2400, 0x3524, 0xE5A8, 0x2530, 0x25A4, 0x251C, 0x2800, 0xAAA4, 0x5452}},
    {14945, 2, {0x2000, 0x27FC, 0x2100, 0xF3F8, 0x2508, 0x2108, 0x21F8, 0x3000, 0xEFBE, 0x2208, 0x279E, 0x2492, 0x2CB2, 0x2492, 0xA79E, 0x4492}},
    {14946, 2, {0x2108, 0x27FE, 0x2108, 0x2000, 0xFBFC, 0x2294, 0x2294, 0x2BFC, 0x3028, 0xE3FE, 0x2220, 0x2324, 0x22A8, 0x2292, 0xA42A, 0x48C6}},
    {14947, 2, {0x27BC, 0x24A4, 0x2294, 0xF4A4, 0x2FFE, 0x2842, 0x27FC, 0x3040, 0xE3F8, 0x2248, 0x23F8, 0x2248, 0x23F8, 0x2040, 0xA7FC, 0x4040}},
    {14948, 2, {0x2148, 0x27C8, 0x2108, 0x27D0, 0xF55E, 0x27E4, 0x2554, 0x27D4, 0x3554, 0xE114, 0x27D4, 0x2208, 0x23C8, 0x2254, 0xA554, 0x48A2}},
    {14949, 2, {0x2114, 0x2112, 0x27D2, 0x2110, 0xF110, 0x2FFE, 0x2290, 0x26D2, 0x3292, 0xE6D2, 0x2294, 0x26D4, 0x228A, 0x22CA, 0xAF16, 0x4022}},
    {14950, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x11FC, 0x1520, 0x19FE, 0x3100, 0xD3FE, 0x1252, 0x128A, 0x1376, 0x5252, 0x2276}},
    {14951, 2, {0x11FC, 0x1104, 0x11FC, 0x1104, 0xFDFC, 0x1088, 0x11FC, 0x1488, 0x1BFE, 0x3088, 0xD124, 0x12AA, 0x1070, 0x10A8, 0x5124, 0x2060}},
    {14952, 2, {0x1088, 0x13DE, 0x1088, 0x11DC, 0xFC88, 0x13DE, 0x1088, 0x17FC, 0x1804, 0x31FC, 0xD004, 0x13FC, 0x1040, 0x12A4, 0x528A, 0x247A}},
    {14953, 2, {0x2400, 0x249E, 0x2504, 0x2608, 0xF45E, 0x2452, 0x23D2, 0x211E, 0x3112, 0xED5E, 0x2552, 0x2592, 0x295E, 0x2920, 0xB52C, 0x4212}},
    {14954, 2, {0x2100, 0x213E, 0x2208, 0x2290, 0xF4BE, 0x2F22, 0x2122, 0x223E, 0x34A2, 0xEFBE, 0x20A2, 0x2022, 0x2ABE, 0x2A80, 0xA814, 0x4022}},
    {14955, 2, {0x2042, 0x239C, 0x2210, 0x23DE, 0xFA94, 0x2294, 0x2420, 0x2BFC, 0x3204, 0xE3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0xA108, 0x4204}},
    {14956, 2, {0x2108, 0x21CC, 0x210A, 0x27E8, 0xF52E, 0x25D8, 0x2748, 0x24CA, 0x340A, 0xE54C, 0x276C, 0x2548, 0x254A, 0x297A, 0xA7A6, 0x4042}},
    {14957, 2, {0x22A0, 0x244C, 0x24A4, 0x260C, 0xF4A4, 0x264C, 0x24A4, 0x2FFE, 0x3802, 0xE3F0, 0x2020, 0x2040, 0x27FC, 0x2040, 0xA140, 0x4080}},
    {14958, 2, {0x2140, 0x226C, 0x2244, 0x236C, 0xFAA4, 0x23AC, 0x22A4, 0x2FFE, 0x3108, 0xE2F4, 0x2442, 0x23F8, 0x2040, 0x27FC, 0xA040, 0x40C0}},
    {14959, 2, {0x2210, 0x23D0, 0x241E, 0x2FE4, 0xF554, 0x2FF4, 0x2548, 0x27D4, 0x31B2, 0xE220, 0x27C0, 0x2088, 0x27FC, 0x2044, 0xA350, 0x4488}},
    {14960, 2, {0x2110, 0x27FE, 0x2110, 0x23F8, 0xFA08, 0x23F8, 0x2000, 0x2FBC, 0x34A4, 0xE7BC, 0x2040, 0x2FFE, 0x2150, 0x2248, 0xAC46, 0x4040}},
    {14961, 2, {0x101E, 0x13E0, 0x1122, 0x1094, 0xFBFE, 0x1252, 0x17FE, 0x1804, 0x33C4, 0xD27E, 0x13C4, 0x1254, 0x13CC, 0x1284, 0x52D4, 0x2348}},
    {14962, 2, {0x2110, 0x2FFE, 0x2110, 0x27BC, 0xFCA4, 0x27BC, 0x2140, 0x2920, 0x33FE, 0xE220, 0x27FC, 0x2A20, 0x23FC, 0x2220, 0xA3FE, 0x4200}},
    {14963, 2, {0x2108, 0x2294, 0x23DE, 0x26B4, 0xFBDE, 0x2294, 0x23DE, 0x2A94, 0x33DE, 0xE210, 0x27FC, 0x2108, 0x2090, 0x2060, 0xA198, 0x4606}},
    {14964, 2, {0x27BC, 0x24A4, 0x27BC, 0x24A4, 0xF7BC, 0x24A4, 0x27BC, 0x2120, 0x33FE, 0xE220, 0x27FC, 0x2A20, 0x23FC, 0x2220, 0xA3FE, 0x4200}},
    {14965, 2, {0x2108, 0x21EC, 0x210A, 0x2FEA, 0xF928, 0x2BC8, 0x293E, 0x2FE8, 0x3948, 0xEBE8, 0x2A28, 0x2B68, 0x2AA8, 0x2BF4, 0xAAB4, 0x52A2}},
    {14966, 2, {0x2210, 0x23DE, 0x2528, 0x2084, 0xF7FE, 0x2248, 0x27FC, 0x2C46, 0x36EC, 0xE554, 0x26EC, 0x2554, 0x26EC, 0x2444, 0xA454, 0x4408}},
    {14967, 2, {0x2040, 0x27FE, 0x2492, 0x23FC, 0xF890, 0x23FC, 0x2090, 0x2FFE, 0x3108, 0xE3FC, 0x2522, 0x21F8, 0x2120, 0x21FE, 0xA002, 0x4556}},
    {14968, 2, {0x11FC, 0x1124, 0x11FC, 0x1124, 0xFBFE, 0x12AA, 0x17FE, 0x1AAA, 0x33FE, 0xD020, 0x13FE, 0x1242, 0x10F8, 0x1150, 0x5020, 0x23FE}},
    {14969, 2, {0x11FC, 0x1124, 0x11FC, 0x1124, 0xFBFE, 0x12AA, 0x17FE, 0x1AAA, 0x33FE, 0xD000, 0x13FE, 0x128A, 0x10F8, 0x1088, 0x5088, 0x23FE}},
    {14970, 2, {0x10, 0x410, 0x4410, 0x44FE, 0x4410, 0x4410, 0x44FC, 0x7744, 0x4444, 0x4444, 0x4428, 0x4428, 0x5510, 0x6628, 0x4444, 0x182}},
    {14971, 2, {0x2010, 0x2010, 0x7E10, 0x42FE, 0x8410, 0x7F10, 0x40FC, 0x5E44, 0x5244, 0x5244, 0x5A28, 0x5428, 0x5010, 0x9428, 0x9844, 0x1182}},
    {14972, 2, {0x1010, 0x1010, 0x3E10, 0x42FE, 0xA410, 0x1810, 0x10FC, 0x2444, 0xC844, 0x1F44, 0x2128, 0xD228, 0xC10, 0x828, 0x3044, 0xC182}},
    {14973, 2, {0x1010, 0x1010, 0xFE10, 0x10FE, 0x1010, 0xFE10, 0x92FC, 0x9244, 0xFE44, 0x1044, 0x3828, 0x5428, 0x9210, 0x1028, 0x1044, 0x1182}},
    {14974, 2, {0x210, 0xF10, 0x7810, 0x8FE, 0xFF10, 0x2A10, 0x2AFC, 0xFF44, 0x2A44, 0x2A44, 0xFF28, 0x828, 0x810, 0xE28, 0x7044, 0x2182}},
    {14975, 2, {0x1020, 0x1020, 0x2020, 0x3F3E, 0x4044, 0x8044, 0x3E44, 0x2A4, 0x428, 0x828, 0x1010, 0x2010, 0x2128, 0x2128, 0x1F44, 0x82}},
    {14976, 2, {0x40, 0x7C40, 0x1040, 0x1080, 0x10FE, 0x1108, 0x1288, 0xFC88, 0x1088, 0x1050, 0x1050, 0x1020, 0x1050, 0x1088, 0x5104, 0x2202}},
    {14977, 2, {0x810, 0x810, 0x1410, 0x141E, 0x2210, 0x5210, 0x88FC, 0x844, 0x44, 0x7E44, 0x228, 0x428, 0x410, 0x828, 0x844, 0x82}},
    {14978, 2, {0x2010, 0x2010, 0x7F10, 0x801E, 0x10, 0x7E10, 0xFC, 0x44, 0xFC44, 0x444, 0x428, 0x528, 0x510, 0x328, 0x144, 0x82}},
    {14979, 2, {0x10, 0x10, 0xFF90, 0x21E, 0x210, 0x7A10, 0x4AFC, 0x4A44, 0x4A44, 0x4A44, 0x7A28, 0x4A28, 0x210, 0x228, 0xA44, 0x482}},
    {14980, 2, {0x2820, 0x2820, 0x2820, 0xFE3E, 0x2A44, 0x2A44, 0xFE44, 0xA8A4, 0xA828, 0xFF28, 0x2910, 0x2910, 0x2D28, 0x4A48, 0x4884, 0x8902}},
    {14981, 2, {0x7FFC, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2118, 0x100, 0x1FC, 0x100, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {14982, 2, {0x210, 0x710, 0x7810, 0x481E, 0x4810, 0x4810, 0x7EFC, 0x4844, 0x4844, 0x4844, 0x4828, 0x4428, 0x4510, 0x6328, 0x4944, 0x482}},
    {14983, 2, {0x820, 0x2820, 0x2820, 0x3E3E, 0x4844, 0x844, 0x844, 0x7FA4, 0x1428, 0x1428, 0x1410, 0x1410, 0x2528, 0x2648, 0x4484, 0x8102}},
    {14984, 2, {0x810, 0x810, 0x81E, 0x7F10, 0x810, 0x8FC, 0xFF44, 0x44, 0x844, 0x828, 0x7F28, 0x810, 0x810, 0xF28, 0xF044, 0x4082}},
    {14985, 2, {0x820, 0x820, 0x1420, 0x223E, 0x4144, 0xBEC4, 0x44, 0xA4, 0x3E28, 0x2228, 0x2210, 0x2210, 0x2228, 0x3E48, 0x2284, 0x102}},
    {14986, 2, {0x10, 0x7E10, 0x1210, 0x121E, 0xFF10, 0x1210, 0x12FC, 0x7E44, 0x2044, 0x2044, 0x7E28, 0x6228, 0xA210, 0x2228, 0x3E44, 0x2282}},
    {14987, 2, {0x10, 0x7F10, 0x4110, 0x7F1E, 0x4110, 0x7F10, 0xFC, 0x44, 0x7F44, 0x844, 0x828, 0xFFA8, 0x810, 0x828, 0x844, 0x882}},
    {14988, 2, {0x10, 0xFC10, 0x410, 0x681E, 0x1010, 0xFE10, 0x92FC, 0x9244, 0xFE44, 0x9244, 0x9228, 0xFE28, 0x9210, 0x9228, 0x9244, 0x8682}},
    {14989, 2, {0x820, 0x820, 0x7F20, 0x83E, 0x844, 0x7F44, 0x4144, 0x82A4, 0x3C28, 0x428, 0x810, 0xF10, 0x7828, 0x828, 0x2844, 0x1082}},
    {14990, 2, {0x810, 0x810, 0x810, 0xFF9E, 0x810, 0x2A10, 0x2AFC, 0x2A44, 0x5D44, 0x88C4, 0x828, 0x1428, 0x1210, 0x2128, 0x4144, 0x8082}},
    {14991, 2, {0x1010, 0x2010, 0x7F10, 0x491E, 0x4910, 0x7F10, 0x49FC, 0x5144, 0x7F44, 0x2444, 0x4428, 0xFFA8, 0x410, 0x428, 0x444, 0x482}},
    {14992, 2, {0x810, 0x410, 0x3F10, 0x211E, 0x2110, 0x3F10, 0x24FC, 0x2644, 0x2544, 0x3F44, 0x2428, 0x2428, 0x4A10, 0x4928, 0x9144, 0x2082}},
    {14993, 2, {0x810, 0x810, 0xFF90, 0x141E, 0x2210, 0x4910, 0x88FC, 0x7F44, 0x4944, 0x7F44, 0x4928, 0x7F28, 0x810, 0xA28, 0xC44, 0x882}},
    {14994, 2, {0x1010, 0x6710, 0x4110, 0x411E, 0x7710, 0x4110, 0x41FC, 0x7F44, 0x1444, 0x1444, 0x1428, 0x1428, 0x2510, 0x2628, 0x4444, 0x8082}},
    {14995, 2, {0x210, 0xF10, 0x7810, 0x81E, 0xFF90, 0x810, 0x2AFC, 0x2A44, 0xEBC4, 0x2A44, 0x2AA8, 0x6B28, 0xAA10, 0x828, 0x844, 0x882}},
    {14996, 2, {0x2210, 0x2210, 0x2210, 0x221E, 0xF790, 0x2210, 0x22FC, 0x6644, 0x7744, 0xAAC4, 0xAA28, 0x3228, 0x2210, 0x2228, 0x2244, 0x2282}},
    {14997, 2, {0x810, 0x810, 0xF9E, 0x810, 0x7F10, 0x41FC, 0x7F44, 0x4144, 0x7F44, 0x4928, 0x828, 0xFF90, 0x810, 0x828, 0x844, 0x882}},
    {14998, 2, {0x10, 0x7E10, 0x210, 0x3E1E, 0x210, 0xFF90, 0x8FC, 0x944, 0x4944, 0x2A44, 0x1C28, 0x2A28, 0xC910, 0x928, 0x2844, 0x1082}},
    {14999, 2, {0x1010, 0x810, 0x7F1E, 0x10, 0x2210, 0x14FC, 0xFF44, 0x44, 0x44, 0x3E28, 0x2228, 0x2210, 0x2210, 0x3E28, 0x2244, 0x82}},
    {15000, 2, {0x820, 0x820, 0x1420, 0x223E, 0x4144, 0xBEC4, 0x844, 0x8A4, 0x7F28, 0x4928, 0x4910, 0x7F10, 0x828, 0x1448, 0x2284, 0x4102}},
    {15001, 2, {0x20, 0x7F20, 0x4920, 0x493E, 0x7F44, 0x4944, 0x4944, 0x7FA4, 0x828, 0xFF28, 0x1C10, 0x2A10, 0x4928, 0x8848, 0x884, 0x902}},
    {15002, 2, {0x2420, 0x2420, 0x7E20, 0x243E, 0x2444, 0xFF44, 0x44, 0x7EA4, 0x4228, 0x4228, 0x7E10, 0x4210, 0x4228, 0x7E48, 0x4284, 0x102}},
    {15003, 2, {0x2210, 0x2210, 0xFF10, 0x221E, 0x3E10, 0x2210, 0x3EFC, 0x2244, 0x2244, 0xFF44, 0x4028, 0x5428, 0x6210, 0x4028, 0x7E44, 0x82}},
    {15004, 2, {0x1010, 0x9210, 0x9210, 0x921E, 0xFE10, 0x10, 0xFEFC, 0x1044, 0x2044, 0xFE44, 0xAA28, 0xAA28, 0xAA10, 0xAA28, 0xAA44, 0x8682}},
    {15005, 2, {0x810, 0x810, 0xFF90, 0x81E, 0x7F10, 0x4910, 0x6BFC, 0x5D44, 0x4944, 0x7F44, 0x828, 0x1C28, 0x2A10, 0x4928, 0x8844, 0x882}},
    {15006, 2, {0x10, 0xFF10, 0x8110, 0xBD1E, 0x8110, 0xBD10, 0x81FC, 0x44, 0x7E44, 0x4244, 0x7E28, 0x4228, 0x7E10, 0x4228, 0x7E44, 0x4282}},
    {15007, 2, {0x1020, 0x820, 0xFF20, 0x3E, 0x3E44, 0x2244, 0x3E44, 0xA4, 0x7F28, 0x4128, 0x4110, 0x7F10, 0x4128, 0x4148, 0x7F84, 0x4102}},
    {15008, 2, {0x4410, 0x2810, 0x10, 0xFF1E, 0x2810, 0xFE10, 0x2AFC, 0xFF44, 0x2A44, 0xFE44, 0x2828, 0x6C28, 0xAA10, 0x2928, 0x2844, 0x2882}},
    {15009, 2, {0x1010, 0x810, 0x7F10, 0x411E, 0x810, 0x3E10, 0x8FC, 0x3E44, 0x844, 0x7F44, 0x828, 0x3E28, 0x2210, 0x2228, 0x3E44, 0x2282}},
    {15010, 2, {0x2810, 0x4B10, 0x4910, 0x491E, 0x6B10, 0x4910, 0x49FC, 0x7F44, 0x844, 0x7F44, 0x128, 0x3228, 0xC10, 0x1228, 0x6144, 0x82}},
    {15011, 2, {0x1020, 0x820, 0x7F20, 0x3E, 0x3E44, 0x2244, 0x3E44, 0xA4, 0x7F28, 0x4128, 0x5D10, 0x5510, 0x5D28, 0x4128, 0x4544, 0x4282}},
    {15012, 2, {0x20, 0x7F20, 0x4920, 0x7F3E, 0x4944, 0x7F44, 0x2A44, 0x2AA4, 0xFFA8, 0x2A28, 0x2A10, 0x7F10, 0x828, 0xFFA8, 0x844, 0x882}},
    {15013, 2, {0x1010, 0x1E10, 0x1010, 0x7F1E, 0x5110, 0x5C10, 0x72FC, 0x4E44, 0x4044, 0x5F44, 0x5128, 0x5F28, 0x5110, 0x5F28, 0x5144, 0xBF82}},
    {15014, 2, {0x1020, 0x820, 0xFF20, 0x813E, 0x44, 0x7E44, 0x844, 0x7EA4, 0x4A28, 0x7E28, 0x4A10, 0x7E10, 0x28, 0x2448, 0x4284, 0x8102}},
    {15015, 2, {0x810, 0x1410, 0x2210, 0x5D1E, 0x8090, 0x3E10, 0x22FC, 0x3E44, 0x44, 0x7744, 0x1128, 0x5528, 0x3310, 0x5528, 0x1144, 0x3382}},
    {15016, 2, {0x2210, 0x1410, 0xFF9E, 0x810, 0x7F10, 0x8FC, 0xFFC4, 0x4944, 0x2A44, 0xFFA8, 0x28, 0x7F10, 0x4110, 0x4128, 0x7F44, 0x4182}},
    {15017, 2, {0x408, 0x3F88, 0x488, 0x7FEE, 0x488, 0x3F88, 0x43C, 0x5554, 0x75D4, 0x4454, 0x7FD4, 0x4448, 0x75C8, 0x5554, 0x5554, 0x8462}},
    {15018, 2, {0x10, 0x7F10, 0x1010, 0xFF9E, 0x2210, 0x4110, 0x94FC, 0x1444, 0xF7C4, 0x1444, 0x7728, 0x1428, 0xF790, 0x1428, 0x1444, 0x1482}},
    {15019, 2, {0x1010, 0x1F10, 0x1010, 0xFF9E, 0x8090, 0x7F10, 0x2AFC, 0x5544, 0xA244, 0x7F44, 0xA2A8, 0x3E28, 0x2210, 0x3E28, 0x2244, 0x3E82}},
    {15020, 2, {0x10, 0xEE10, 0x2210, 0xAA1E, 0x6610, 0xAA10, 0xFC, 0x2444, 0x7F44, 0xC844, 0x7E28, 0x4828, 0x7E10, 0x4828, 0x7F44, 0x4082}},
    {15021, 2, {0x1010, 0x1F10, 0x101E, 0xFF10, 0x9110, 0xFCFC, 0x9144, 0xBF44, 0xAA44, 0xBE28, 0xAA28, 0xBE10, 0x8010, 0x7F28, 0x5544, 0xFF82}},
    {15022, 2, {0x8, 0x7FC8, 0x408, 0xFFEE, 0x8428, 0x3588, 0x41C, 0x3594, 0x14, 0x14, 0xEEF4, 0xAAA8, 0xAAA8, 0xAAB4, 0xEEF4, 0xAAA2}},
    {15023, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE10E, 0x100, 0x1FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {15024, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE10E, 0x1110, 0x920, 0x540, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {15025, 2, {0x80, 0x2080, 0x10F8, 0x108, 0xFE10, 0x9FC, 0x924, 0x4924, 0x2924, 0x1124, 0x13FE, 0x2850, 0x2850, 0x4888, 0x8104, 0x202}},
    {15026, 2, {0x40, 0x440, 0x240, 0x1240, 0x840, 0x840, 0x1FC, 0x7E40, 0x40, 0x40, 0x40, 0x7FF8, 0x8, 0x8, 0x8, 0x8}},
    {15027, 2, {0x40, 0x440, 0x240, 0x1240, 0x840, 0x9FC, 0x7E40, 0x40, 0x40, 0x0, 0x1FF0, 0x2008, 0x2008, 0x2008, 0x1FF0, 0x0}},
    {15028, 2, {0x10, 0x7C10, 0x90, 0x50, 0x10, 0xFE90, 0x2850, 0x2810, 0x281E, 0x29F0, 0x2810, 0x2A10, 0x4C10, 0x4810, 0x8010, 0x10}},
    {15029, 2, {0x1010, 0x1010, 0x9490, 0x5850, 0x1010, 0x7C90, 0x1050, 0x1010, 0xFE1E, 0x11F0, 0x1010, 0x1010, 0x2010, 0x2010, 0x4010, 0x8010}},
    {15030, 2, {0x10, 0xFE10, 0x1090, 0x1050, 0x2010, 0x7C90, 0x4450, 0x4410, 0x441E, 0x7DF0, 0x4410, 0x4410, 0x4410, 0x7C10, 0x4410, 0x10}},
    {15031, 2, {0x10, 0xFE10, 0x90, 0x50, 0x7C10, 0x4490, 0x4450, 0x4410, 0x7C1E, 0x1F0, 0x4410, 0x2410, 0x2810, 0xE10, 0xF010, 0x4010}},
    {15032, 2, {0x2208, 0x2248, 0xFFA8, 0x2228, 0x2208, 0x3E48, 0x2228, 0x2228, 0x3E0E, 0x2278, 0x2208, 0xFF08, 0x8, 0x1408, 0x2208, 0x4108}},
    {15033, 2, {0x808, 0x7F48, 0x4928, 0xFFA8, 0x4908, 0x7F48, 0x828, 0x7F28, 0x490E, 0x7F78, 0x1008, 0xFF08, 0x2108, 0x7208, 0xC08, 0xF308}},
    {15034, 2, {0x808, 0x848, 0xFFA8, 0x828, 0xEB88, 0xAAC8, 0xEBA8, 0xAAA8, 0xEB8E, 0xAAF8, 0xEB88, 0x1408, 0x1208, 0x2208, 0x4108, 0x8108}},
    {15035, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x440, 0x1240, 0x87E, 0xFFC0, 0x40, 0x40}},
    {15036, 2, {0x0, 0x4, 0xFF78, 0x440, 0x440, 0x7440, 0x547E, 0x5448, 0x5448, 0x5448, 0x7448, 0x5448, 0x448, 0x488, 0x1488, 0x908}},
    {15037, 2, {0x0, 0xFFFE, 0x0, 0xC, 0x7CF0, 0x4480, 0x4480, 0x4480, 0x7CFE, 0x4488, 0x4488, 0x4488, 0x7C88, 0x4508, 0x108, 0x208}},
    {15038, 2, {0x1000, 0x1004, 0x1E78, 0x2240, 0x6240, 0x9440, 0x87E, 0x1448, 0x2248, 0x4148, 0xBE48, 0x2248, 0x2248, 0x2288, 0x3E88, 0x2308}},
    {15039, 2, {0x800, 0x4904, 0x4978, 0x4940, 0x7F40, 0x1040, 0x207E, 0x4848, 0x8848, 0x4948, 0x4948, 0x7F48, 0x848, 0x1088, 0x2088, 0x4108}},
    {15040, 2, {0x0, 0x7C04, 0x4478, 0x7C40, 0x4440, 0x7C40, 0x7E, 0xFE48, 0x8248, 0x8248, 0xFE48, 0x8248, 0x8248, 0xFE88, 0x8288, 0x108}},
    {15041, 2, {0x1100, 0x9102, 0xAABC, 0xBBA0, 0x9120, 0xAAA0, 0xBBBE, 0x8024, 0xFFA4, 0x9224, 0x9224, 0xA1A4, 0x8024, 0xFFC4, 0x44, 0x84}},
    {15042, 2, {0x1000, 0x804, 0x7F78, 0x4040, 0x5240, 0x5240, 0x7F7E, 0x5248, 0x5248, 0x5E48, 0x4048, 0x5548, 0x5548, 0x5588, 0x8088, 0x108}},
    {15043, 2, {0x2020, 0x1020, 0x1020, 0xFF20, 0x2020, 0x2020, 0x3E20, 0x2220, 0x2250, 0x2250, 0x2250, 0x2250, 0x4288, 0x4A88, 0x8504, 0x202}},
    {15044, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x3FF8, 0x4808, 0x8408, 0x7FE8, 0x1008, 0x1F88, 0x1088, 0x2088, 0x2088, 0x42A8, 0x8110}},
    {15045, 2, {0x2020, 0x1020, 0x1050, 0x88, 0xFD04, 0x2202, 0x20F8, 0x3C00, 0x2400, 0x27FE, 0x2488, 0x2488, 0x2488, 0x4488, 0x5508, 0x8A08}},
    {15046, 2, {0x2020, 0x1020, 0x50, 0xFE50, 0x2088, 0x2104, 0x3E22, 0x2420, 0x24A8, 0x24A4, 0x2524, 0x2522, 0x4622, 0x5420, 0x88A0, 0x40}},
    {15047, 2, {0x1010, 0x810, 0xFF28, 0x2044, 0x3E82, 0x2210, 0x2208, 0x4220, 0x4A10, 0x8408, 0x0, 0x3FF8, 0x8, 0x8, 0x8, 0x8}},
    {15048, 2, {0x1010, 0x810, 0xFF28, 0x2044, 0x3E82, 0x2210, 0x2208, 0x4220, 0x4A10, 0x8408, 0x1FF0, 0x2008, 0x2008, 0x2008, 0x1FF0, 0x0}},
    {15049, 2, {0x2040, 0x1040, 0x10FC, 0x104, 0xFE08, 0x21FE, 0x2100, 0x3D7C, 0x2544, 0x2544, 0x2554, 0x2548, 0x2542, 0x4542, 0x553E, 0x8A00}},
    {15050, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFC00, 0x21FE, 0x2008, 0x3C08, 0x25E8, 0x2528, 0x2528, 0x2528, 0x25E8, 0x4408, 0x5428, 0x8810}},
    {15051, 2, {0x2080, 0x1080, 0x10FE, 0x120, 0xFE20, 0x21FC, 0x2124, 0x3D24, 0x25FC, 0x2524, 0x2524, 0x27FE, 0x2504, 0x4504, 0x5514, 0x8908}},
    {15052, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE00, 0x2000, 0x21FE, 0x3C10, 0x2410, 0x2490, 0x249C, 0x2490, 0x2490, 0x4490, 0x55FE, 0x8800}},
    {15053, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFC90, 0x2090, 0x215E, 0x3D64, 0x2754, 0x2554, 0x2554, 0x2548, 0x2548, 0x4554, 0x5514, 0x8922}},
    {15054, 2, {0x2100, 0x1100, 0x13FE, 0xFC00, 0x23FE, 0x2222, 0x39FC, 0x2820, 0x29FC, 0x2924, 0x29FC, 0x2924, 0x29FC, 0x4820, 0x5BFE, 0x8020}},
    {15055, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFDFC, 0x2050, 0x21FC, 0x3D54, 0x2554, 0x25FC, 0x2440, 0x27FE, 0x2488, 0x4590, 0x5470, 0x8B8C}},
    {15056, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x2038, 0x23C0, 0x3C04, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {15057, 2, {0x40, 0x40, 0x7C40, 0x4440, 0x45FC, 0x4444, 0x4444, 0x7C44, 0x4444, 0x4484, 0x4484, 0x4484, 0x7D04, 0x4504, 0x228, 0x410}},
    {15058, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {15059, 2, {0x40, 0x40, 0x7C80, 0x44FE, 0x4500, 0x4600, 0x44FC, 0x7C08, 0x4410, 0x4420, 0x4440, 0x4480, 0x7D02, 0x4502, 0xFE, 0x0}},
    {15060, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x200, 0x3F8, 0x200, 0x200, 0x200, 0xFFFE, 0x0}},
    {15061, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x840, 0x1042, 0x2042, 0xC03E}},
    {15062, 2, {0x200, 0x200, 0x3F8, 0x200, 0x200, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {15063, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x1FE0, 0x40, 0x180, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {15064, 2, {0x0, 0x7FFC, 0x100, 0x300, 0x570, 0x190C, 0xE102, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {15065, 2, {0x20, 0x20, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x4420, 0x7C20, 0x47FE, 0x4420, 0x4450, 0x4450, 0x7C88, 0x4488, 0x104, 0x202}},
    {15066, 2, {0x800, 0x1FF8, 0x2248, 0x4448, 0x888, 0x1108, 0x2228, 0x410, 0x1FF0, 0x1010, 0x1010, 0x1FD0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {15067, 2, {0x0, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x4504, 0x45FC, 0x7C50, 0x4450, 0x4450, 0x4450, 0x4492, 0x7C92, 0x4512, 0x20E, 0x400}},
    {15068, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x4908, 0x4908, 0x79F8, 0x4908, 0x4908, 0x4908, 0x4908, 0x79F8, 0x4800, 0x0, 0x7FE, 0x0}},
    {15069, 2, {0x8, 0x3C, 0x7BE0, 0x4A20, 0x4A20, 0x4A20, 0x4A20, 0x7BFE, 0x4A20, 0x4A10, 0x4A10, 0x4A12, 0x7A0A, 0x4A8A, 0x326, 0x212}},
    {15070, 2, {0x20, 0x124, 0x7924, 0x4924, 0x49FC, 0x4820, 0x4820, 0x7BFC, 0x4904, 0x4888, 0x4888, 0x4850, 0x7820, 0x4850, 0x188, 0x606}},
    {15071, 2, {0x0, 0xF8, 0x7C88, 0x4488, 0x4488, 0x4488, 0x4506, 0x7E00, 0x4400, 0x44FC, 0x4484, 0x4484, 0x7C84, 0x4484, 0xFC, 0x84}},
    {15072, 2, {0x20, 0x20, 0x7D24, 0x44A4, 0x44A8, 0x4420, 0x45FC, 0x7C20, 0x4420, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {15073, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE30E, 0xC0, 0xE00, 0x180, 0x40}},
    {15074, 2, {0x20, 0x20, 0x7820, 0x49FC, 0x4820, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x4840, 0x4840, 0x4888, 0x7904, 0x4BFE, 0x102, 0x0}},
    {15075, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FFA, 0x102, 0x102, 0xFE}},
    {15076, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x80, 0x1F00, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {15077, 2, {0x40, 0x40, 0x7CFC, 0x4504, 0x4688, 0x4450, 0x4420, 0x7C40, 0x4480, 0x45FC, 0x4684, 0x4484, 0x7C84, 0x4484, 0xFC, 0x84}},
    {15078, 2, {0x840, 0x840, 0x10FC, 0x1088, 0x3550, 0x5420, 0x94D8, 0x1706, 0x14F8, 0x1488, 0x1488, 0x14F8, 0x1488, 0x1088, 0x10F8, 0x1088}},
    {15079, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1110, 0x920, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {15080, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x4C44, 0x7840, 0x4BFE, 0x4888, 0x4888, 0x4908, 0x78D0, 0x4820, 0x50, 0x88, 0x304}},
    {15081, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x820, 0xFFFE, 0x820, 0x820, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {15082, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x3018, 0xD3F6, 0x1210, 0x1210, 0x13F0, 0x1210, 0x1210, 0x13F0, 0x1210}},
    {15083, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2448, 0x1448, 0x1450, 0x440, 0xFFFE, 0x0}},
    {15084, 2, {0x80, 0x80, 0x7900, 0x49FC, 0x4A04, 0x4C04, 0x7BE4, 0x4A24, 0x4A24, 0x4BE4, 0x4A24, 0x7A24, 0x4BE4, 0x4, 0x28, 0x10}},
    {15085, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x40, 0x7FFC, 0x840, 0x440, 0x140, 0x80}},
    {15086, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {15087, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x1842, 0xE03E}},
    {15088, 2, {0x80, 0x40, 0x7BF8, 0x4A08, 0x4A08, 0x4BF8, 0x4A08, 0x7A08, 0x4BF8, 0x4A40, 0x4A44, 0x4A28, 0x7A10, 0x4A88, 0x306, 0x200}},
    {15089, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x440, 0x2448, 0x4444, 0x8842, 0x1140, 0x2080}},
    {15090, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x4820, 0x4BFE, 0x7A02, 0x4C04, 0x49F8, 0x4810, 0x4820, 0x7BFE, 0x4820, 0x20, 0xA0, 0x40}},
    {15091, 2, {0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x2530, 0x260E, 0x2000, 0x2FF8, 0x2808, 0x4FF8, 0x4808, 0x8FF8, 0x808}},
    {15092, 2, {0x40, 0x44, 0x7BF4, 0x4848, 0x4848, 0x4FFE, 0x7820, 0x4840, 0x4BF8, 0x4910, 0x4A20, 0x7DFE, 0x4820, 0x20, 0xA0, 0x40}},
    {15093, 2, {0x90, 0x90, 0x7890, 0x4F9E, 0x4890, 0x4890, 0x4890, 0x7B9C, 0x4890, 0x4890, 0x4890, 0x4F9E, 0x7890, 0x4890, 0x90, 0x90}},
    {15094, 2, {0x0, 0x1FC, 0x7804, 0x48FC, 0x4804, 0x49FC, 0x7800, 0x4BFE, 0x4A22, 0x4820, 0x49FC, 0x7924, 0x4924, 0x134, 0x128, 0x20}},
    {15095, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1010, 0x2008, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {15096, 2, {0x40, 0x40, 0xF7FE, 0x9040, 0x93F8, 0x9048, 0x97FE, 0xF048, 0x93F8, 0x9040, 0x9240, 0x927C, 0xF240, 0x9540, 0x4FE, 0x800}},
    {15097, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x618, 0x38E0, 0x2080, 0x3E80, 0x22FC, 0x3E90, 0x2090, 0x2090, 0x4090, 0x8110}},
    {15098, 2, {0x820, 0x820, 0x820, 0x1450, 0x2288, 0x4924, 0x820, 0x1450, 0x2288, 0x4106, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {15099, 2, {0x40, 0x40, 0xF0A0, 0x9110, 0x9208, 0x9DF6, 0x9000, 0xF7FC, 0x94A4, 0x94A4, 0x97FC, 0x94A4, 0xF4A4, 0x94A4, 0x414, 0x408}},
    {15100, 2, {0x82, 0x82, 0x7A82, 0x4BEA, 0x4C8A, 0x488A, 0x4FFA, 0x788A, 0x4BEA, 0x4AAA, 0x4AAA, 0x4AAA, 0x7AA2, 0x4AE2, 0x8A, 0x84}},
    {15101, 2, {0x0, 0x1F8, 0x7808, 0x4808, 0x49F8, 0x4808, 0x4808, 0x7BFE, 0x4820, 0x4A22, 0x4974, 0x48A8, 0x7924, 0x4A22, 0xA0, 0x40}},
    {15102, 2, {0x20, 0x124, 0x78A4, 0x48A8, 0x4820, 0x4BFE, 0x4A02, 0x7A02, 0x4AFA, 0x4A8A, 0x4A8A, 0x4A8A, 0x7AFA, 0x4A02, 0x20A, 0x204}},
    {15103, 2, {0xA0, 0x90, 0x7880, 0x49FE, 0x4910, 0x4B10, 0x4DFC, 0x7910, 0x4910, 0x49FC, 0x4910, 0x4910, 0x7910, 0x49FE, 0x100, 0x100}},
    {15104, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4800, 0x49FC, 0x4808, 0x4810, 0x7BFE, 0x4820, 0x20, 0xA0, 0x40}},
    {15105, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2040, 0x10A0, 0x8110, 0x42E8, 0x1406, 0x23F8, 0xE208, 0x2208, 0x23F8, 0x2208}},
    {15106, 2, {0x1FC, 0x124, 0x7924, 0x49FC, 0x4924, 0x4924, 0x79FC, 0x4820, 0x4820, 0x4BFE, 0x4A22, 0x7A2A, 0x4AFA, 0x20A, 0x202, 0x206}},
    {15107, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x1000, 0x2008, 0x7FFC, 0x4, 0x3FF8, 0x2008, 0x3FF8, 0x1108, 0x2290, 0xC60, 0xF01E}},
    {15108, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FFC, 0x2080, 0x2FF8, 0x2080, 0x3FFC, 0x2220, 0x4A28, 0x5424, 0xA4A2, 0x840}},
    {15109, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0xFFFE, 0x800, 0xFF8, 0x1008, 0x2028, 0xC010}},
    {15110, 2, {0x0, 0x3FC, 0x7A04, 0x4BFC, 0x4A20, 0x4BFE, 0x4A10, 0x7A8A, 0x4B06, 0x4800, 0x4BFC, 0x4A04, 0x7BFC, 0x4A04, 0x3FC, 0x204}},
    {15111, 2, {0x80, 0x80, 0x78F8, 0x4908, 0x4A10, 0x4DFC, 0x7924, 0x4924, 0x4924, 0x4924, 0x4BFE, 0x7850, 0x4850, 0x88, 0x104, 0x602}},
    {15112, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x4908, 0x7908, 0x49F8, 0x4800, 0x4BFC, 0x4A94, 0x7A94, 0x4A94, 0x294, 0x7FE, 0x0}},
    {15113, 2, {0x0, 0x3FE, 0x7820, 0x4840, 0x49FC, 0x4954, 0x7954, 0x4954, 0x4954, 0x492C, 0x4820, 0x7BFE, 0x4850, 0x88, 0x104, 0x202}},
    {15114, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0xFC, 0x7F08, 0x1110, 0x3FF8, 0x200, 0xFFFE, 0x800, 0x17F0, 0x2220, 0xC1C0, 0xE3C}},
    {15115, 2, {0x80, 0xBC, 0x7884, 0x4908, 0x49FE, 0x4B20, 0x4D20, 0x797C, 0x4990, 0x4910, 0x49FE, 0x4910, 0x7928, 0x4928, 0x144, 0x182}},
    {15116, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x920, 0x1110, 0x2308}},
    {15117, 2, {0x28, 0x24, 0x7BFE, 0x4820, 0x49FC, 0x4924, 0x79FC, 0x4924, 0x49FC, 0x4924, 0x4808, 0x7BFE, 0x4888, 0x48, 0x48, 0x18}},
    {15118, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1248, 0x3CF0, 0x820, 0x1450, 0x3EF8, 0x0, 0x2488, 0x2244, 0x4244}},
    {15119, 2, {0x0, 0x3FE, 0x7800, 0x49FC, 0x4904, 0x4904, 0x79FC, 0x4800, 0x4BFE, 0x4A8A, 0x4A52, 0x7BFE, 0x4A22, 0x222, 0x22A, 0x204}},
    {15120, 2, {0x0, 0x3FC, 0x7890, 0x4A94, 0x4998, 0x4890, 0x4FFE, 0x7800, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x1F8, 0x108}},
    {15121, 2, {0x200, 0x11E, 0xF7D2, 0x9012, 0x97D4, 0x9454, 0x97D8, 0xF014, 0x97D2, 0x9092, 0x9112, 0x91DA, 0xF714, 0x9110, 0x510, 0x210}},
    {15122, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4824, 0x4BFE, 0x7850, 0x4888, 0x4904, 0x4A8A, 0x4888, 0x7BFE, 0x4888, 0x88, 0x108, 0x208}},
    {15123, 2, {0x0, 0x3FE, 0x7850, 0x49FC, 0x4954, 0x4954, 0x49FC, 0x7800, 0x49FC, 0x4800, 0x4BFE, 0x4820, 0x78A8, 0x4924, 0x2A2, 0x40}},
    {15124, 2, {0x0, 0x7BC, 0xF084, 0x94A4, 0x9294, 0x94A4, 0x9050, 0xF188, 0x9626, 0x90C0, 0x9310, 0x9064, 0xF388, 0x9030, 0xC0, 0x700}},
    {15125, 2, {0x222, 0x112, 0x7914, 0x483E, 0x4F88, 0x4888, 0x4908, 0x793E, 0x4B88, 0x4D48, 0x4908, 0x493E, 0x7908, 0x4908, 0x108, 0x108}},
    {15126, 2, {0x210, 0x110, 0xF79E, 0x94A4, 0x94D4, 0x9794, 0x9408, 0xF414, 0x9462, 0x9800, 0x93F8, 0x9208, 0xF208, 0x9208, 0x3F8, 0x208}},
    {15127, 2, {0x110, 0x114, 0x79D2, 0x4A50, 0x4A7E, 0x4D50, 0x4890, 0x7AA8, 0x4928, 0x4944, 0x4A44, 0x4C82, 0x7808, 0x4AA4, 0x252, 0x452}},
    {15128, 2, {0x20, 0x1FC, 0x7824, 0x4BFE, 0x4824, 0x49FC, 0x4820, 0x7AAA, 0x4BAE, 0x4A22, 0x4BFE, 0x4A22, 0x7BAE, 0x4AAA, 0x2AA, 0x422}},
    {15129, 2, {0xE, 0x3F0, 0x7844, 0x4A24, 0x4908, 0x4BF8, 0x7A08, 0x4BFC, 0x4A04, 0x4BFE, 0x4A02, 0x7AAA, 0x4AAA, 0x502, 0x414, 0x808}},
    {15130, 2, {0x208, 0x108, 0xF7C8, 0x9210, 0x949E, 0x9FD4, 0x9064, 0xF794, 0x9494, 0x9794, 0x9494, 0x9788, 0xF488, 0x9494, 0x4A4, 0x5C2}},
    {15131, 2, {0x40, 0xA0, 0xF118, 0x96E6, 0x9000, 0x93F8, 0x9208, 0xF3F8, 0x9000, 0x97BC, 0x9084, 0x94A4, 0xF294, 0x94A4, 0x294, 0x108}},
    {15132, 2, {0x0, 0xFEFE, 0x2020, 0x4444, 0xFEFE, 0x1212, 0x7C7C, 0x1010, 0x1EFE, 0xE000, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {15133, 2, {0x104, 0x88, 0x7800, 0x4BFE, 0x4A22, 0x4AAA, 0x4A72, 0x7A22, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x104}},
    {15134, 2, {0x20, 0x222, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4880, 0x49FE, 0x4A22, 0x7D52, 0x4902, 0x1FA, 0x6}},
    {15135, 2, {0x4044, 0x4024, 0x4028, 0x7EFE, 0xAA92, 0x2AD6, 0x2ABA, 0x4A92, 0x4AFE, 0x9200, 0x127C, 0x2244, 0x227C, 0x4244, 0x947C, 0x844}},
    {15136, 2, {0x210, 0x210, 0xF3D0, 0x945E, 0x98A4, 0x93D4, 0x9054, 0xF054, 0x93C8, 0x9048, 0x9054, 0x97E2, 0xF000, 0x92A4, 0x252, 0x452}},
    {15137, 2, {0x0, 0x1FC, 0x7820, 0x4BFE, 0x4A22, 0x49AC, 0x7820, 0x49AC, 0x4850, 0x48C8, 0x4B26, 0x78F8, 0x4808, 0xD0, 0x20, 0x10}},
    {15138, 2, {0x104, 0x88, 0x7BFE, 0x4820, 0x49FC, 0x4820, 0x7BFE, 0x4854, 0x4992, 0x4890, 0x4BFE, 0x7890, 0x48D4, 0x38A, 0x96, 0x1A2}},
    {15139, 2, {0x284, 0x284, 0xFAA4, 0x96C4, 0x929E, 0x97C4, 0x9004, 0xF454, 0x928C, 0x97C4, 0x9104, 0x97C4, 0xF104, 0x91C4, 0xE14, 0x8}},
    {15140, 2, {0x0, 0x3FE, 0x7A02, 0x4BFE, 0x4A02, 0x4BFE, 0x7908, 0x4A52, 0x4B9C, 0x4908, 0x4A52, 0x7BDE, 0x4800, 0x2A4, 0x252, 0x452}},
    {15141, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x1110, 0x1FF0, 0x4104, 0x7FFC, 0x200, 0xFFFE, 0x920, 0x3558, 0xDFF6, 0x540, 0x920}},
    {15142, 2, {0x88, 0x3FE, 0x78A8, 0x4890, 0x49FE, 0x4B20, 0x79FC, 0x4920, 0x49FC, 0x4920, 0x49FE, 0x7900, 0x4BFC, 0x88, 0x70, 0x38E}},
    {15143, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0xA20, 0x3458, 0xCF86, 0x120, 0x1FF0, 0x950, 0x1320}},
    {15144, 2, {0x110, 0x110, 0xF110, 0x97BC, 0x9110, 0x9110, 0x9FFE, 0xF2A8, 0x92A8, 0x96EC, 0x9AAA, 0x92AA, 0xF2A8, 0x95A8, 0x448, 0x898}},
    {15145, 2, {0x88, 0x3DE, 0x7888, 0x49DC, 0x4888, 0x4BDE, 0x4888, 0x7BFC, 0x4804, 0x49FC, 0x4804, 0x4BFC, 0x7840, 0x4AA4, 0x28A, 0x47A}},
    {15146, 2, {0xFE0, 0x820, 0xFE0, 0x820, 0xFE0, 0x7C7C, 0x4444, 0x7C7C, 0x4544, 0xFFFE, 0x8802, 0xFF0, 0x1010, 0x1FE0, 0x20, 0xFFFE}},
    {15147, 2, {0x440, 0xFFFE, 0x440, 0x17D2, 0xF01C, 0x17D0, 0x3452, 0xD7CE, 0x0, 0x4444, 0x8222, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {15148, 2, {0x7BC, 0x4A4, 0xF7BC, 0x94A4, 0x97BC, 0x94A4, 0x97BC, 0xF120, 0x93FE, 0x9220, 0x97FC, 0x9A20, 0xF3FC, 0x9220, 0x3FE, 0x200}},
    {15149, 2, {0x1E, 0x3E0, 0x7922, 0x4894, 0x4BFE, 0x4A52, 0x7BFE, 0x4804, 0x4BC4, 0x4A7E, 0x4BC4, 0x7A54, 0x4BCC, 0x284, 0x2D4, 0x348}},
    {15150, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x4450, 0xFE48, 0x44FE, 0x7D90, 0x10FC, 0xFE90, 0x92FC, 0xFE90, 0x1090, 0xFEFE, 0x2880}},
    {15151, 2, {0x80, 0x3FFE, 0x800, 0xEFF8, 0xA000, 0xAFF8, 0xA808, 0xEFF8, 0xA220, 0xBD5C, 0xB7F4, 0xBC94, 0xF5DC, 0x1C96, 0x17F6, 0x2CA2}},
    {15152, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {15153, 2, {0x2850, 0x3E7C, 0x4890, 0xFEFE, 0x1428, 0x1528, 0x264A, 0x4486, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {15154, 2, {0x88, 0x7C50, 0x1000, 0x11FC, 0x7C50, 0x5488, 0x5504, 0x7DFC, 0x5554, 0x5554, 0x7D54, 0x93FE, 0x6000, 0x2000, 0x5FFE, 0x8000}},
    {15155, 2, {0x80, 0x3C80, 0x2480, 0x25FC, 0x2520, 0x3E20, 0x2420, 0x2420, 0x27FE, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {15156, 2, {0x90, 0x7890, 0x4890, 0x4890, 0x4BFC, 0x7890, 0x4890, 0x4890, 0x4890, 0x7BFE, 0x4800, 0x4890, 0x4888, 0x4908, 0x4904, 0x9A04}},
    {15157, 2, {0x40, 0x3C20, 0x2420, 0x25FE, 0x2400, 0x3C88, 0x2504, 0x2602, 0x2488, 0x3C88, 0x2450, 0x2450, 0x2420, 0x4450, 0x5488, 0x8B06}},
    {15158, 2, {0x20, 0x7920, 0x4920, 0x49FC, 0x4920, 0x7A20, 0x4820, 0x4BFE, 0x4800, 0x7800, 0x49FC, 0x4904, 0x4904, 0x4904, 0x49FC, 0x9904}},
    {15159, 2, {0x0, 0x7BFE, 0x4A22, 0x4A22, 0x4A22, 0x7BFE, 0x4A22, 0x4A62, 0x4A72, 0x7AAA, 0x4B22, 0x4A22, 0x4A22, 0x4A02, 0x4BFE, 0x9A02}},
    {15160, 2, {0x88, 0x3C88, 0x25FE, 0x2488, 0x2488, 0x3C50, 0x2450, 0x2494, 0x2594, 0x3E98, 0x2490, 0x24B2, 0x24D2, 0x4492, 0x548E, 0x8880}},
    {15161, 2, {0x80, 0x3C80, 0x24F8, 0x2508, 0x2610, 0x3DFC, 0x2424, 0x2424, 0x25FE, 0x3C24, 0x2424, 0x25FC, 0x2424, 0x4420, 0x54A0, 0x8840}},
    {15162, 2, {0x20, 0x7924, 0x48A8, 0x49FC, 0x4840, 0x7BFE, 0x4908, 0x4A24, 0x4C22, 0x79FC, 0x4924, 0x4924, 0x4934, 0x4928, 0x4820, 0x9820}},
    {15163, 2, {0x88, 0x7888, 0x4BFE, 0x48A8, 0x4810, 0x7BFE, 0x4880, 0x4880, 0x48FC, 0x7800, 0x48A8, 0x48A8, 0x48A8, 0x48AA, 0x492A, 0x9A06}},
    {15164, 2, {0x1040, 0x1040, 0x7EFE, 0x1080, 0x21FC, 0x3E84, 0x24FC, 0x6484, 0xBCFC, 0x2440, 0x247C, 0x3CC4, 0x2528, 0x2410, 0x2468, 0x2D86}},
    {15165, 2, {0x0, 0x3CF8, 0x2488, 0x24F8, 0x2400, 0x3DDC, 0x2554, 0x25DC, 0x2400, 0x3DFE, 0x2480, 0x24FC, 0x2404, 0x4404, 0x5428, 0x8810}},
    {15166, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1070}},
    {15167, 2, {0x108, 0x7888, 0x4BC8, 0x4810, 0x4BDE, 0x7A64, 0x4BD4, 0x4814, 0x4BD4, 0x7854, 0x4894, 0x48E8, 0x4B88, 0x4894, 0x4A94, 0x9922}},
    {15168, 2, {0x110, 0x7910, 0x49DE, 0x4AA8, 0x4C44, 0x7800, 0x4BFC, 0x4820, 0x4928, 0x7928, 0x4928, 0x4AB4, 0x4C62, 0x4820, 0x4BFE, 0x9800}},
    {15169, 2, {0x1000, 0x1000, 0x13F8, 0x1088, 0xFC90, 0x1090, 0x30A0, 0x38BC, 0x5484, 0x5484, 0x9104, 0x1104, 0x1104, 0x1204, 0x1228, 0x1410}},
    {15170, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFC20, 0x10A8, 0x30A4, 0x38A4, 0x5522, 0x5122, 0x9222, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {15171, 2, {0x1020, 0x1020, 0x1040, 0x1040, 0xFC84, 0x1084, 0x3108, 0x3BF8, 0x5510, 0x5410, 0x9020, 0x1040, 0x1088, 0x1104, 0x13FE, 0x1102}},
    {15172, 2, {0x1000, 0x13F0, 0x1090, 0x1090, 0xFC90, 0x1090, 0x3090, 0x3BF0, 0x5490, 0x5490, 0x9090, 0x1092, 0x108A, 0x108A, 0x1086, 0x1082}},
    {15173, 2, {0x1000, 0x1000, 0x13FC, 0x1000, 0xFC00, 0x1000, 0x3000, 0x39F8, 0x5400, 0x5400, 0x9000, 0x1000, 0x1000, 0x17FE, 0x1000, 0x1000}},
    {15174, 2, {0x0, 0x3FE0, 0x420, 0x47C, 0x804, 0x3014, 0xC108, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {15175, 2, {0x1040, 0x1040, 0x1080, 0x1088, 0xFD04, 0x13FE, 0x3002, 0x3890, 0x5490, 0x5490, 0x9090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {15176, 2, {0x1040, 0x1040, 0x1040, 0x13F8, 0xFC48, 0x1048, 0x3048, 0x3848, 0x57FE, 0x5440, 0x90A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1406}},
    {15177, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {15178, 2, {0x1010, 0x1110, 0x1110, 0x1110, 0xFD10, 0x11FE, 0x3100, 0x3900, 0x5500, 0x55F8, 0x9108, 0x1108, 0x1108, 0x1208, 0x1208, 0x1408}},
    {15179, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3020, 0x39FC, 0x5420, 0x5420, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {15180, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0xFDFC, 0x1120, 0x3220, 0x3820, 0x5420, 0x53FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {15181, 2, {0x4080, 0x2080, 0x89C, 0x17E0, 0xE080, 0x2084, 0x2084, 0x207C, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {15182, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFE94, 0x1254, 0x3224, 0x3A24, 0x5654, 0x5254, 0x9294, 0x1304, 0x1204, 0x1204, 0x1214, 0x1208}},
    {15183, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0xFC20, 0x1020, 0x30FC, 0x3820, 0x5420, 0x55FE, 0x9022, 0x1022, 0x102A, 0x1024, 0x1020, 0x1020}},
    {15184, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC106, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {15185, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x3020, 0x3BFE, 0x5470, 0x54A8, 0x90A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {15186, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x1108, 0x31F8, 0x3900, 0x5500, 0x55FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15187, 2, {0x1020, 0x1020, 0x1040, 0x1088, 0xFD04, 0x13FE, 0x3002, 0x3888, 0x5488, 0x53FE, 0x9088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {15188, 2, {0x1008, 0x103C, 0x13E0, 0x1020, 0xFD24, 0x10A4, 0x30A8, 0x3820, 0x57FE, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {15189, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x3020, 0x3BFE, 0x5420, 0x5440, 0x9040, 0x1088, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {15190, 2, {0x10A0, 0x10A8, 0x10A4, 0x1124, 0xFD20, 0x133E, 0x3BE0, 0x3520, 0x5120, 0x5120, 0x9120, 0x1120, 0x1114, 0x1114, 0x110C, 0x1104}},
    {15191, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {15192, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x1202, 0x31FC, 0x3800, 0x5400, 0x55FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15193, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x33FE, 0x3808, 0x5408, 0x55FE, 0x9008, 0x1088, 0x1048, 0x1008, 0x1028, 0x1010}},
    {15194, 2, {0x1080, 0x1080, 0x10FE, 0x1100, 0xFE00, 0x101C, 0x31E0, 0x3920, 0x5522, 0x5524, 0x9128, 0x1110, 0x1110, 0x1148, 0x1184, 0x1102}},
    {15195, 2, {0x1004, 0x101E, 0x11E0, 0x1100, 0xFD06, 0x1178, 0x3150, 0x3950, 0x5552, 0x5554, 0x9148, 0x1148, 0x1144, 0x1254, 0x1262, 0x1440}},
    {15196, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0xFBFE, 0x1020, 0x3124, 0x3924, 0x5524, 0x53A8, 0x9128, 0x1110, 0x1212, 0x122A, 0x1446, 0x1082}},
    {15197, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x200, 0x200, 0x1FE0, 0x420, 0x420, 0x822, 0x1022, 0x201E, 0x4000}},
    {15198, 2, {0x1020, 0x1020, 0x1124, 0x1122, 0xFA2A, 0x1010, 0x3060, 0x3B80, 0x5440, 0x5040, 0x93FC, 0x1044, 0x1084, 0x1084, 0x1114, 0x1208}},
    {15199, 2, {0x440, 0x440, 0x3FF8, 0x440, 0x440, 0x440, 0xFFFE, 0x820, 0x3118, 0xC106, 0x1FF0, 0x300, 0x5C0, 0x1930, 0x6108, 0x100}},
    {15200, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x3020, 0x3BFE, 0x5490, 0x5490, 0x9090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {15201, 2, {0x1020, 0x1020, 0x1040, 0x11FC, 0xFD04, 0x1104, 0x3194, 0x3954, 0x5524, 0x5524, 0x9154, 0x1194, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15202, 2, {0x2004, 0x27C4, 0x2284, 0x2294, 0xFA94, 0x2294, 0x2294, 0x77D4, 0x6A94, 0xA294, 0xA294, 0x2294, 0x2284, 0x2484, 0x2494, 0x2888}},
    {15203, 2, {0x1100, 0x1080, 0x10BC, 0x1204, 0xFE04, 0x12F4, 0x3294, 0x3A94, 0x5694, 0x5294, 0x92F4, 0x1294, 0x1204, 0x1204, 0x1214, 0x1208}},
    {15204, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFD44, 0x1124, 0x3124, 0x390C, 0x5500, 0x55FE, 0x9002, 0x1002, 0x13FA, 0x1002, 0x1014, 0x1008}},
    {15205, 2, {0x1040, 0x1040, 0x10F8, 0x1088, 0xFD10, 0x13FC, 0x3104, 0x3924, 0x5524, 0x5124, 0x9124, 0x1124, 0x1050, 0x1088, 0x1104, 0x1202}},
    {15206, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC00, 0x1038, 0x33C0, 0x3840, 0x5440, 0x507C, 0x93C0, 0x1040, 0x1042, 0x1042, 0x103E, 0x1000}},
    {15207, 2, {0x200, 0x400, 0x1FF0, 0x1110, 0x1090, 0xFFFE, 0x1010, 0x1210, 0x2150, 0x4020, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15208, 2, {0x2000, 0x27DE, 0x2092, 0x2492, 0xFC94, 0x2494, 0x27D8, 0x7094, 0x6992, 0xA192, 0xA292, 0x229A, 0x2494, 0x2890, 0x2290, 0x2110}},
    {15209, 2, {0x1004, 0x101E, 0x11E0, 0x1022, 0xFD12, 0x1094, 0x3080, 0x3808, 0x55FE, 0x5408, 0x9088, 0x1048, 0x1048, 0x1008, 0x1028, 0x1010}},
    {15210, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x3124, 0x3924, 0x55FC, 0x5524, 0x9124, 0x11FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {15211, 2, {0x1020, 0x1020, 0x10A8, 0x10A8, 0xFCA8, 0x1174, 0x3222, 0x3820, 0x5420, 0x55FC, 0x9020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {15212, 2, {0x1000, 0x13FE, 0x1040, 0x1080, 0xFD44, 0x1224, 0x3068, 0x38B0, 0x5530, 0x5228, 0x9068, 0x10A4, 0x1122, 0x1220, 0x10A0, 0x1040}},
    {15213, 2, {0x2002, 0x23E2, 0x2222, 0x222A, 0xFA2A, 0x23EA, 0x210A, 0x710A, 0x6BEA, 0xA12A, 0xA12A, 0x212A, 0x2222, 0x2222, 0x24AA, 0x2844}},
    {15214, 2, {0x2124, 0x2124, 0x2248, 0x2248, 0xF924, 0x2124, 0x2000, 0x77FE, 0x6880, 0xA1FC, 0xA104, 0x2288, 0x2450, 0x2020, 0x20C0, 0x2700}},
    {15215, 2, {0x1000, 0x13FE, 0x1120, 0x1128, 0xFD28, 0x11E8, 0x3128, 0x3928, 0x5528, 0x55E8, 0x9128, 0x112A, 0x113A, 0x13EA, 0x1026, 0x1020}},
    {15216, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF17C, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2388, 0x2558, 0x920, 0x1110, 0x610C}},
    {15217, 2, {0x1000, 0x11FE, 0x1100, 0x117C, 0xFD54, 0x1154, 0x317C, 0x3954, 0x5554, 0x557C, 0x9110, 0x1110, 0x1110, 0x1100, 0x11FE, 0x1000}},
    {15218, 2, {0x1000, 0x1050, 0x1048, 0x1084, 0xFD24, 0x1020, 0x3050, 0x3888, 0x5506, 0x52F8, 0x9088, 0x1088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {15219, 2, {0x1020, 0x1022, 0x11FA, 0x1024, 0xFC24, 0x13FE, 0x3010, 0x3820, 0x55FC, 0x5488, 0x9110, 0x12FE, 0x1410, 0x1010, 0x1050, 0x1020}},
    {15220, 2, {0x1080, 0x12FC, 0x12A4, 0x12A4, 0xFEA8, 0x1290, 0x32A8, 0x38C6, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {15221, 2, {0x1100, 0x1100, 0x11FC, 0x1204, 0xFA04, 0x15F4, 0x3154, 0x3954, 0x55F4, 0x5154, 0x9154, 0x11F4, 0x1114, 0x1004, 0x1028, 0x1010}},
    {15222, 2, {0x1104, 0x1104, 0x17FE, 0x1104, 0xFD44, 0x1020, 0x37FE, 0x3880, 0x5480, 0x54FC, 0x9084, 0x1084, 0x1104, 0x1104, 0x1214, 0x1408}},
    {15223, 2, {0x11FC, 0x1124, 0x1124, 0x11FC, 0xFD24, 0x1124, 0x31FC, 0x3840, 0x5440, 0x53FE, 0x9042, 0x1082, 0x1082, 0x1102, 0x1214, 0x1408}},
    {15224, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFA10, 0x11FC, 0x3124, 0x3924, 0x5524, 0x51FC, 0x9050, 0x1058, 0x1094, 0x1092, 0x1112, 0x120E}},
    {15225, 2, {0x1010, 0x1388, 0x1088, 0x10FE, 0xFC90, 0x1390, 0x3224, 0x3A24, 0x5638, 0x5388, 0x9090, 0x1090, 0x10A4, 0x10BE, 0x1282, 0x1100}},
    {15226, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x1124, 0x3222, 0x39FC, 0x5524, 0x55FC, 0x9124, 0x11FC, 0x1022, 0x1022, 0x101E, 0x1000}},
    {15227, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0xFC20, 0x13FE, 0x3202, 0x3A02, 0x56FA, 0x528A, 0x928A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {15228, 2, {0x1020, 0x1022, 0x13B4, 0x10A8, 0xFCA8, 0x1124, 0x32A2, 0x3840, 0x55FC, 0x5504, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15229, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0xFE1C, 0x12F0, 0x3290, 0x3A90, 0x56FE, 0x5290, 0x9290, 0x1290, 0x12AA, 0x12CA, 0x14A6, 0x1812}},
    {15230, 2, {0x1000, 0x13FE, 0x1202, 0x1202, 0xFBFE, 0x1210, 0x3292, 0x3A92, 0x5692, 0x52FE, 0x9210, 0x1292, 0x1292, 0x1492, 0x14FE, 0x1802}},
    {15231, 2, {0x2000, 0x23FC, 0x2204, 0x23FC, 0xFA04, 0x23FC, 0x2000, 0x7180, 0x6E3E, 0xA422, 0xA422, 0x2422, 0x25A2, 0x262A, 0x2424, 0x2020}},
    {15232, 2, {0x1000, 0x11FC, 0x1008, 0x1010, 0xFA22, 0x12AA, 0x3272, 0x3A22, 0x5672, 0x52AA, 0x9326, 0x12A2, 0x1242, 0x1202, 0x13FE, 0x1002}},
    {15233, 2, {0x1184, 0x1068, 0x1030, 0x10C8, 0xFB24, 0x1020, 0x33FE, 0x3840, 0x54FC, 0x5184, 0x92FC, 0x1084, 0x10FC, 0x1084, 0x1094, 0x1088}},
    {15234, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x3118, 0xC106, 0x1FF0, 0x540, 0x920, 0x1110}},
    {15235, 2, {0x101C, 0x11E0, 0x1020, 0x1020, 0xFBFE, 0x10A8, 0x3124, 0x3A42, 0x5440, 0x53FE, 0x9088, 0x1108, 0x10D0, 0x1030, 0x1048, 0x1184}},
    {15236, 2, {0x1040, 0x1020, 0x13FE, 0x1080, 0xFC80, 0x11FE, 0x3142, 0x3A22, 0x55FA, 0x5082, 0x90F2, 0x1092, 0x1112, 0x1152, 0x122A, 0x1004}},
    {15237, 2, {0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC906, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {15238, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x128A, 0x3050, 0x39FC, 0x5420, 0x54F8, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {15239, 2, {0x1000, 0x13FC, 0x1040, 0x1040, 0xFBFE, 0x10A0, 0x3110, 0x3A08, 0x5446, 0x5040, 0x9048, 0x1264, 0x1252, 0x1452, 0x1140, 0x1080}},
    {15240, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x1104, 0x32FA, 0x3800, 0x5400, 0x53FE, 0x9020, 0x1124, 0x1122, 0x1222, 0x10A0, 0x1040}},
    {15241, 2, {0x2110, 0x2110, 0x2110, 0x227C, 0xFA10, 0x2692, 0x3654, 0x6A10, 0x62FE, 0xA210, 0xA228, 0x2228, 0x2228, 0x2244, 0x2244, 0x2282}},
    {15242, 2, {0x2008, 0x27C8, 0x2048, 0x2288, 0xF93E, 0x2088, 0x27D8, 0x715C, 0x692C, 0xA12A, 0xA148, 0x2108, 0x2108, 0x2108, 0x2508, 0x2208}},
    {15243, 2, {0x2100, 0x211E, 0x27F2, 0x2112, 0xFA14, 0x23D4, 0x2258, 0x7654, 0x6BD2, 0xA252, 0xA252, 0x23DA, 0x2254, 0x2250, 0x2250, 0x22D0}},
    {15244, 2, {0x1000, 0x17FE, 0x1040, 0x1080, 0xFBFC, 0x1294, 0x3294, 0x3AF4, 0x5694, 0x5294, 0x92F4, 0x1294, 0x1294, 0x1294, 0x13FC, 0x1204}},
    {15245, 2, {0x820, 0xFFFE, 0x0, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2618, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15246, 2, {0x1000, 0x13FE, 0x1010, 0x1020, 0xFC68, 0x10A4, 0x3322, 0x3820, 0x5400, 0x51FC, 0x9154, 0x1154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {15247, 2, {0x2040, 0x2040, 0x27FC, 0x20E0, 0xF950, 0x2248, 0x2C46, 0x73F8, 0x6800, 0xA000, 0xA7FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {15248, 2, {0x1020, 0x1020, 0x10A4, 0x10A2, 0xFD2A, 0x1010, 0x3860, 0x3580, 0x56FC, 0x5084, 0x90FC, 0x1084, 0x10FC, 0x1084, 0x10FC, 0x1084}},
    {15249, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x13FE, 0x3020, 0x38A0, 0x552C, 0x5524, 0x9124, 0x11AC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {15250, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFDFC, 0x1154, 0x3154, 0x3954, 0x55FC, 0x5420, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {15251, 2, {0x1100, 0x117E, 0x1110, 0x1110, 0xFD90, 0x115E, 0x3B22, 0x352A, 0x5126, 0x5142, 0x9152, 0x110C, 0x1104, 0x1104, 0x117E, 0x1100}},
    {15252, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFBFE, 0x1020, 0x31FC, 0x3924, 0x55FC, 0x5124, 0x91FC, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {15253, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0xFDFC, 0x1154, 0x3154, 0x3954, 0x5554, 0x552C, 0x9020, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {15254, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x1020, 0x31FC, 0x3924, 0x5524, 0x55FC, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {15255, 2, {0x20, 0x7E20, 0x1450, 0x888, 0x7F44, 0x2A22, 0x49F8, 0x8808, 0x2810, 0x1020, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15256, 2, {0x2008, 0x2788, 0x2088, 0x2290, 0xF91E, 0x2114, 0x27E4, 0x7154, 0x6B14, 0xA314, 0xA514, 0x2508, 0x2908, 0x2114, 0x2524, 0x2242}},
    {15257, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFBDE, 0x1000, 0x31FC, 0x3800, 0x57FE, 0x5080, 0x9100, 0x11FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {15258, 2, {0x100, 0x1F8, 0x100, 0x3FF8, 0x2108, 0x22C8, 0x2D28, 0x2288, 0x2448, 0x3FF8, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15259, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5400, 0x53FE, 0x9020, 0x1120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {15260, 2, {0x1000, 0x13FE, 0x1200, 0x1224, 0xFAA4, 0x12A4, 0x32E4, 0x3A3E, 0x57E4, 0x52A4, 0x92A4, 0x12A4, 0x132E, 0x1200, 0x13FE, 0x1000}},
    {15261, 2, {0x2000, 0x229E, 0x2252, 0x2252, 0xF552, 0x2112, 0x2292, 0x7252, 0x6C32, 0xA3D2, 0xA25A, 0x2254, 0x2250, 0x2250, 0x23D0, 0x2250}},
    {15262, 2, {0x1020, 0x1120, 0x10BE, 0x1042, 0xFC84, 0x1310, 0x3110, 0x3928, 0x5544, 0x5482, 0x90FC, 0x1084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {15263, 2, {0x2008, 0x2008, 0x27E8, 0x2108, 0xF93E, 0x2208, 0x23D8, 0x765C, 0x6A6C, 0xA26A, 0xA248, 0x2248, 0x23C8, 0x2248, 0x2008, 0x2008}},
    {15264, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC00, 0x13FE, 0x3A02, 0x3424, 0x5020, 0x53FE, 0x9070, 0x10A8, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {15265, 2, {0x800, 0x8FC, 0x2A80, 0x2CF8, 0x4888, 0x8F8, 0x1480, 0x22FC, 0x4000, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {15266, 2, {0x1080, 0x10BC, 0x1084, 0x1108, 0xFDFE, 0x1320, 0x3520, 0x397C, 0x5590, 0x5110, 0x91FE, 0x1110, 0x1128, 0x1128, 0x1144, 0x1182}},
    {15267, 2, {0x0, 0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {15268, 2, {0x200, 0x100, 0x7FFC, 0x4924, 0x1290, 0x2448, 0x1830, 0x6FEC, 0x820, 0x820, 0xFE0, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {15269, 2, {0x1040, 0x1048, 0x1084, 0x11FE, 0xFC20, 0x13FE, 0x3888, 0x3504, 0x53FE, 0x5524, 0x9124, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {15270, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFBFE, 0x1242, 0x348C, 0x39F0, 0x5420, 0x5044, 0x93FE, 0x1022, 0x1128, 0x1224, 0x14A4, 0x1040}},
    {15271, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x3100, 0x39FE, 0x5500, 0x55FE, 0x9002, 0x12AA, 0x12AA, 0x1402, 0x1014, 0x1008}},
    {15272, 2, {0x1000, 0x11FC, 0x1124, 0x11FC, 0xFD24, 0x11FC, 0x3888, 0x3544, 0x5242, 0x50F8, 0x9188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {15273, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5440, 0x9020, 0x10A4, 0x128A, 0x128A, 0x1478, 0x1000}},
    {15274, 2, {0x1000, 0x13B8, 0x1128, 0x1128, 0xFD3A, 0x11AA, 0x3726, 0x3A40, 0x5420, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {15275, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC24, 0x11FC, 0x3820, 0x35FC, 0x5020, 0x53FE, 0x9020, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15276, 2, {0x1090, 0x1088, 0x1104, 0x1242, 0xFC88, 0x11FC, 0x3804, 0x3400, 0x53DE, 0x5042, 0x9252, 0x114A, 0x1252, 0x1042, 0x114A, 0x1084}},
    {15277, 2, {0x1000, 0x13FC, 0x1008, 0x10D0, 0xFC20, 0x13FE, 0x328A, 0x3A52, 0x56FA, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {15278, 2, {0x1080, 0x13DE, 0x108A, 0x11CA, 0xFC8A, 0x13D2, 0x30A6, 0x3808, 0x54F0, 0x5420, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x10A0, 0x1040}},
    {15279, 2, {0x2000, 0x27FE, 0x2402, 0x25FA, 0xFC42, 0x24C2, 0x252A, 0x746A, 0x6CB2, 0xA52A, 0xA46A, 0x24AA, 0x2522, 0x2442, 0x27FE, 0x2402}},
    {15280, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFD54, 0x1124, 0x3154, 0x3904, 0x55FC, 0x5400, 0x9112, 0x11D4, 0x1118, 0x1152, 0x1192, 0x110E}},
    {15281, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC20, 0x10A0, 0x38FC, 0x3520, 0x5020, 0x53FE, 0x9050, 0x1050, 0x1090, 0x1092, 0x1112, 0x120E}},
    {15282, 2, {0x1080, 0x1080, 0x11FC, 0x1244, 0xFD54, 0x11F4, 0x3884, 0x3528, 0x5290, 0x51FC, 0x9244, 0x1554, 0x11F4, 0x1084, 0x1128, 0x1210}},
    {15283, 2, {0x2044, 0x2724, 0x2528, 0x2500, 0xFD7C, 0x2710, 0x2510, 0x7510, 0x6DFE, 0xA710, 0xA510, 0x2528, 0x2528, 0x2544, 0x2B44, 0x2082}},
    {15284, 2, {0x10A0, 0x112C, 0x1124, 0x1124, 0xFDAC, 0x1124, 0x3124, 0x39FC, 0x5420, 0x55FC, 0x9088, 0x1050, 0x1020, 0x1050, 0x1088, 0x1306}},
    {15285, 2, {0x2220, 0x2120, 0x213E, 0x27A0, 0xFA40, 0x223C, 0x23A4, 0x72B4, 0x6AAC, 0xA2A4, 0xA2FE, 0x22A4, 0x24A4, 0x24A4, 0x29A4, 0x304C}},
    {15286, 2, {0x2440, 0x225E, 0x2292, 0x2012, 0xF7D2, 0x211E, 0x2552, 0x7552, 0x6D52, 0xA7DE, 0xA152, 0x2112, 0x2212, 0x2222, 0x242A, 0x2844}},
    {15287, 2, {0x1050, 0x1048, 0x13FC, 0x10A0, 0xFD22, 0x121E, 0x3110, 0x39E0, 0x5504, 0x54FC, 0x9000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC}},
    {15288, 2, {0x1044, 0x1024, 0x1228, 0x1100, 0xFD7C, 0x1010, 0x3010, 0x3B7E, 0x5510, 0x5110, 0x9128, 0x1124, 0x1144, 0x1280, 0x147E, 0x1000}},
    {15289, 2, {0x2008, 0x2008, 0x27C8, 0x2210, 0xFA1E, 0x24A4, 0x27D4, 0x7114, 0x6914, 0xA7D4, 0xA114, 0x2108, 0x21C8, 0x2E14, 0x2414, 0x2022}},
    {15290, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x37D8, 0xC106, 0x0, 0x7BBC, 0x4AA4, 0x4AA4, 0x4AA4, 0x4AA4, 0x7BBC, 0x4AA4}},
    {15291, 2, {0x1000, 0x11FC, 0x10A4, 0x1088, 0xFC50, 0x1020, 0x30D8, 0x3B26, 0x5420, 0x51FC, 0x9124, 0x1124, 0x11FC, 0x1022, 0x13FE, 0x1002}},
    {15292, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0xFA04, 0x13FC, 0x3200, 0x3BDC, 0x5644, 0x5354, 0x92CC, 0x1244, 0x12CC, 0x1554, 0x1444, 0x18CC}},
    {15293, 2, {0x2100, 0x223C, 0x27A4, 0x24A4, 0xF6A4, 0x25A4, 0x24C2, 0x7F80, 0x6CBC, 0xA4A4, 0xA6A4, 0x25A4, 0x2494, 0x2488, 0x2494, 0x29A2}},
    {15294, 2, {0x1000, 0x7E7C, 0x4244, 0x7E7C, 0x4244, 0x7E7C, 0x4844, 0x4444, 0x5A94, 0x6108, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15295, 2, {0x1000, 0x11FC, 0x1124, 0x11FC, 0xFD24, 0x11FC, 0x30A8, 0x38A8, 0x57FE, 0x54A8, 0x90A8, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {15296, 2, {0x2210, 0x2110, 0x2110, 0x27BE, 0xFA40, 0x2200, 0x23BE, 0x728A, 0x6A88, 0xA2A8, 0xA2AE, 0x22A8, 0x24A8, 0x25A8, 0x285E, 0x3080}},
    {15297, 2, {0x2020, 0x2710, 0x257C, 0x2500, 0xFD44, 0x2628, 0x2500, 0x75FE, 0x6D00, 0xA500, 0xA57C, 0x2644, 0x2444, 0x2444, 0x247C, 0x2444}},
    {15298, 2, {0x22A8, 0x22A8, 0x27FC, 0x22A8, 0xFAAA, 0x24E6, 0x2800, 0x77FC, 0x6C44, 0xA040, 0xA3F8, 0x2248, 0x2248, 0x2248, 0x2258, 0x2040}},
    {15299, 2, {0x1020, 0x1020, 0x13FE, 0x1070, 0xFCAC, 0x1322, 0x3850, 0x3488, 0x5124, 0x5222, 0x9124, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {15300, 2, {0x820, 0x7EFC, 0x870, 0x1CA8, 0x2B26, 0xC820, 0x87C, 0x3F80, 0x2000, 0x3FFC, 0x2000, 0x2FF8, 0x2888, 0x4FF8, 0x4802, 0x87FE}},
    {15301, 2, {0x2040, 0x2020, 0x27FE, 0x2420, 0xFC40, 0x25FC, 0x2524, 0x75FC, 0x6D44, 0xA5FC, 0xA450, 0x2490, 0x2BFE, 0x2810, 0x3010, 0x2010}},
    {15302, 2, {0x1040, 0x1020, 0x13FC, 0x1000, 0xFD08, 0x1090, 0x33FE, 0x3A20, 0x56A0, 0x52FC, 0x9320, 0x1220, 0x12FC, 0x1420, 0x1420, 0x19FE}},
    {15303, 2, {0x1050, 0x1050, 0x13DE, 0x1050, 0xFDDC, 0x1050, 0x33DE, 0x3850, 0x5450, 0x5420, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {15304, 2, {0x2010, 0x2010, 0x24FE, 0x2210, 0xFA10, 0x20FE, 0x2092, 0x7692, 0x6AFE, 0xA210, 0xA238, 0x2254, 0x2292, 0x2210, 0x2510, 0x28FE}},
    {15305, 2, {0x1020, 0x103E, 0x1020, 0x11FE, 0xFD20, 0x113C, 0x31E2, 0x391E, 0x5504, 0x5578, 0x9192, 0x1154, 0x12FE, 0x1210, 0x1450, 0x1020}},
    {15306, 2, {0x2000, 0x27FC, 0x2040, 0x2278, 0xF240, 0x2FFE, 0x2100, 0x71FC, 0x6A00, 0xA3FC, 0xA004, 0x2554, 0x2554, 0x2804, 0x2028, 0x2010}},
    {15307, 2, {0x4000, 0x4CFE, 0x7010, 0x4220, 0x42FC, 0x3E84, 0x1084, 0x10FC, 0xFE84, 0x30FC, 0x3884, 0x5484, 0x52FC, 0x9048, 0x1084, 0x1102}},
    {15308, 2, {0x2000, 0x227E, 0x2110, 0x2130, 0xF84A, 0x209A, 0x262C, 0x724C, 0x6A9A, 0xA22A, 0xA24A, 0x2288, 0x2228, 0x2510, 0x28FE, 0x2000}},
    {15309, 2, {0x2090, 0x2290, 0x229E, 0x22A2, 0xFAD4, 0x23A8, 0x2094, 0x70A4, 0x6FFE, 0xA284, 0xA2A4, 0x2294, 0x2294, 0x2284, 0x2494, 0x2888}},
    {15310, 2, {0x2208, 0x2108, 0x2108, 0x27C8, 0xF83E, 0x2788, 0x2018, 0x779C, 0x682C, 0xA7AA, 0xA4C8, 0x2488, 0x2488, 0x2788, 0x2488, 0x2008}},
    {15311, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x8084, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15312, 2, {0x40, 0x7BFC, 0x4880, 0x49F8, 0x5220, 0x4DFC, 0x4888, 0x68F8, 0x5088, 0x40F8, 0x4088, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {15313, 2, {0x2000, 0x2F3E, 0x2120, 0x213C, 0xF120, 0x2F3C, 0x2820, 0x78FE, 0x6850, 0xAF52, 0xA154, 0x2148, 0x2148, 0x2144, 0x2A62, 0x2440}},
    {15314, 2, {0x200E, 0x27F0, 0x2044, 0x2224, 0xF908, 0x23FC, 0x2204, 0x73FC, 0x6A04, 0xA3FC, 0xA204, 0x23FC, 0x2090, 0x2110, 0x2212, 0x2C0E}},
    {15315, 2, {0x2040, 0x20A0, 0x2118, 0x26E6, 0xF800, 0x23F8, 0x2208, 0x73F8, 0x6800, 0xA7BC, 0xA084, 0x24A4, 0x2294, 0x24A4, 0x2294, 0x2108}},
    {15316, 2, {0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE, 0xC20, 0x703C, 0x1DE2, 0xF022, 0x1C1E, 0x100, 0xFFFE, 0x920, 0x711C}},
    {15317, 2, {0x2240, 0x2242, 0x225C, 0x27F0, 0xFA50, 0x23D0, 0x225E, 0x73D4, 0x6A54, 0xA254, 0xA7F4, 0x2014, 0x2294, 0x2464, 0x2824, 0x2044}},
    {15318, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1154, 0x3154, 0x39FC, 0x5400, 0x53FE, 0x9040, 0x1024, 0x12A2, 0x128A, 0x1488, 0x1078}},
    {15319, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0xFCA4, 0x27BC, 0x2444, 0x7444, 0x6DF4, 0xA444, 0xA4E4, 0x2554, 0x264C, 0x2444, 0x2454, 0x2408}},
    {15320, 2, {0x2108, 0x2108, 0x27C8, 0x2108, 0xF91E, 0x2392, 0x2024, 0x7380, 0x6808, 0xA7C8, 0xA108, 0x2108, 0x2594, 0x2954, 0x2524, 0x2242}},
    {15321, 2, {0x2208, 0x2108, 0x27C8, 0x2210, 0xF49E, 0x2FD4, 0x2064, 0x7794, 0x6C94, 0xA794, 0xA494, 0x2788, 0x2488, 0x2494, 0x24A4, 0x25C2}},
    {15322, 2, {0x1210, 0x1210, 0x13DE, 0x1228, 0xFD44, 0x10A0, 0x3110, 0x3A08, 0x55F6, 0x5000, 0x9000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {15323, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1000, 0x33FE, 0x3A52, 0x5450, 0x51DC, 0x9104, 0x11DC, 0x1050, 0x1050, 0x13FE, 0x1000}},
    {15324, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC20, 0x13FE, 0x3088, 0x3888, 0x5554, 0x5222, 0x9020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {15325, 2, {0x2010, 0x2790, 0x24BC, 0x24A4, 0xF4C4, 0x27A8, 0x2110, 0x7128, 0x6946, 0xA5C0, 0xA53C, 0x2524, 0x2524, 0x25A4, 0x2E3C, 0x2024}},
    {15326, 2, {0x2080, 0x209E, 0x2112, 0x2152, 0xFA5E, 0x2780, 0x209E, 0x7112, 0x6A52, 0xA7DE, 0xA052, 0x201E, 0x2552, 0x2552, 0x2412, 0x2016}},
    {15327, 2, {0x1010, 0x1010, 0xFEFE, 0x3838, 0x5454, 0x9282, 0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x1110, 0x9A0, 0x540, 0x1930, 0xE30E}},
    {15328, 2, {0x1020, 0x13FE, 0x1048, 0x10F0, 0xFC20, 0x13FE, 0x3242, 0x3890, 0x55F8, 0x5028, 0x9120, 0x11FC, 0x1220, 0x17FE, 0x1020, 0x1020}},
    {15329, 2, {0x1040, 0x1040, 0xFEFC, 0x3884, 0x5504, 0x92F4, 0x2894, 0x4494, 0x92F4, 0x1084, 0x5494, 0x3888, 0x5482, 0x9282, 0x507E, 0x2000}},
    {15330, 2, {0x1000, 0x13DE, 0x1252, 0x13DE, 0xFE10, 0x1252, 0x31CE, 0x3800, 0x5488, 0x53FE, 0x9088, 0x1088, 0x13FE, 0x1088, 0x1104, 0x1202}},
    {15331, 2, {0x2108, 0x2FFE, 0x2108, 0x2090, 0xF108, 0x23FC, 0x2C8A, 0x7128, 0x6A50, 0xA040, 0xA7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {15332, 2, {0x1010, 0x1010, 0x1010, 0xFEFE, 0x1010, 0x5454, 0x5454, 0x5454, 0xBABA, 0x1010, 0x3838, 0x5454, 0x9292, 0x1010, 0x1010, 0x1010}},
    {15333, 2, {0x1010, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x5454, 0x7C7C, 0x5454, 0x7C7C, 0x1010, 0x3838, 0x5454, 0x9292, 0x1010, 0x1010, 0x1010}},
    {15334, 2, {0x2110, 0x27FC, 0x2110, 0x21F0, 0xF910, 0x21F0, 0x2110, 0x77FC, 0x6910, 0xA248, 0xA446, 0x23F8, 0x20E0, 0x2158, 0x2644, 0x2040}},
    {15335, 2, {0x2100, 0x213C, 0x2114, 0x27D4, 0xF914, 0x2124, 0x2FCC, 0x7100, 0x693C, 0xA524, 0xA5E4, 0x253C, 0x2724, 0x2500, 0x24FE, 0x2800}},
    {15336, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFBFE, 0x1252, 0x3252, 0x3BFE, 0x5524, 0x50A8, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {15337, 2, {0x10, 0x7E10, 0x22FE, 0x1C10, 0x227C, 0x7714, 0x557C, 0x2250, 0x557E, 0x812, 0x7F12, 0x1C16, 0x2A28, 0x4928, 0x8844, 0x882}},
    {15338, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0xFA22, 0x11AC, 0x3020, 0x39AC, 0x5450, 0x50C8, 0x9326, 0x10F8, 0x1008, 0x10D0, 0x1020, 0x1010}},
    {15339, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFA42, 0x1278, 0x33C4, 0x3A3C, 0x5600, 0x53FE, 0x9240, 0x13A4, 0x1258, 0x15B4, 0x1452, 0x19B0}},
    {15340, 2, {0x1010, 0xFEFE, 0x5454, 0x9292, 0x40, 0x7BF8, 0x848, 0x17FE, 0x2048, 0x7BF8, 0x840, 0xBF8, 0x5040, 0x27FC, 0x5840, 0x87FE}},
    {15341, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFA42, 0x13F8, 0x3244, 0x3AFC, 0x56A8, 0x52F8, 0x92A8, 0x12F8, 0x1220, 0x15FC, 0x1444, 0x198C}},
    {15342, 2, {0x1000, 0x11FC, 0x1154, 0x1154, 0xFDFC, 0x1080, 0x31FE, 0x3A42, 0x55F2, 0x5152, 0x9152, 0x11F2, 0x104A, 0x13FA, 0x100A, 0x1004}},
    {15343, 2, {0x1080, 0x11FC, 0x1288, 0x1070, 0xFBAE, 0x10F8, 0x3120, 0x39FC, 0x5420, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {15344, 2, {0x11FC, 0x1154, 0x1154, 0x11FC, 0xFC20, 0x13FE, 0x3020, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x17FE}},
    {15345, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x9002, 0x1FF8, 0x2100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x920, 0x711C}},
    {15346, 2, {0x2110, 0x2110, 0x27BC, 0x2110, 0xFBB8, 0x2554, 0x2912, 0x7000, 0x6BF8, 0xA000, 0xA7FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {15347, 2, {0x2288, 0x2288, 0x27E8, 0xFA90, 0x241E, 0x27E4, 0x7454, 0x6854, 0xAF54, 0xA554, 0x2554, 0x2748, 0x2548, 0x2054, 0x2294, 0x2122}},
    {15348, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC20, 0x13FE, 0x3202, 0x3888, 0x5524, 0x5450, 0x9088, 0x11FC, 0x128A, 0x1088, 0x10F8, 0x1088}},
    {15349, 2, {0x2110, 0x27FE, 0x2110, 0x23F8, 0xF840, 0x27FC, 0x2040, 0x73F8, 0x6A48, 0xA3F8, 0xA248, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x2FFE}},
    {15350, 2, {0x1020, 0x1088, 0x1252, 0x1326, 0xFE52, 0x128A, 0x3222, 0x3BFE, 0x5488, 0x5144, 0x927A, 0x1088, 0x1150, 0x1020, 0x10D8, 0x1706}},
    {15351, 2, {0x2128, 0x21AA, 0x216C, 0x2228, 0xFAFE, 0x2644, 0x3628, 0x6AFE, 0x6210, 0xA27C, 0xA210, 0x22FE, 0x2210, 0x2228, 0x2244, 0x2282}},
    {15352, 2, {0x2040, 0x23F8, 0x2048, 0x27FE, 0xF848, 0x23F8, 0x2040, 0x77FE, 0x6800, 0xA2A4, 0xA452, 0x23F8, 0x22A8, 0x22A8, 0x2FFE, 0x2000}},
    {15353, 2, {0x2004, 0x2FE4, 0x2284, 0x2FE4, 0xFABE, 0x2AA4, 0x2FE4, 0x7004, 0x6FD4, 0xA00C, 0xAFEC, 0x2104, 0x2544, 0x2524, 0x2924, 0x230C}},
    {15354, 2, {0x40, 0x7E40, 0x487E, 0x7EA0, 0x4310, 0x7EFE, 0x48AA, 0x48AA, 0x7EFE, 0x0, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15355, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x8212, 0xD20, 0x32C0, 0xDA0, 0x3290, 0xD88, 0x3100, 0xFFFE, 0x920, 0x711C}},
    {15356, 2, {0x1088, 0x13FE, 0x1088, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3840, 0x5488, 0x55FC, 0x9000, 0x11FC, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {15357, 2, {0x1020, 0x17FE, 0x1402, 0x13FC, 0xFE24, 0x13FC, 0x3200, 0x3AFC, 0x5684, 0x52FC, 0x9284, 0x12FC, 0x1284, 0x14FC, 0x1448, 0x1884}},
    {15358, 2, {0x2028, 0x24FE, 0x2228, 0x22FE, 0xF8AA, 0x20FE, 0x2EAA, 0x72FE, 0x6A00, 0xA27C, 0xA244, 0x227C, 0x2244, 0x227C, 0x2500, 0x28FE}},
    {15359, 2, {0x1124, 0x1248, 0x1124, 0x1000, 0xFBFC, 0x1294, 0x3264, 0x3A94, 0x57FC, 0x5248, 0x9368, 0x1248, 0x136A, 0x124A, 0x1246, 0x1362}},
    {15360, 2, {0x1050, 0x1190, 0x109E, 0x13EA, 0xFC8A, 0x11D2, 0x328A, 0x38A4, 0x5450, 0x5088, 0x9326, 0x10A8, 0x1070, 0x10A8, 0x1124, 0x1060}},
    {15361, 2, {0x2000, 0x27FC, 0x2100, 0xFBF8, 0x2508, 0x2108, 0x71F8, 0x6800, 0xAFBE, 0xA208, 0x279E, 0x2492, 0x2CB2, 0x2492, 0x279E, 0x2492}},
    {15362, 2, {0x2020, 0x23FE, 0x2202, 0x241C, 0xFBE0, 0x2124, 0x20A8, 0x73FE, 0x6924, 0xA202, 0xA5FC, 0x2124, 0x21FC, 0x2124, 0x21FC, 0x2104}},
    {15363, 2, {0x2206, 0x2438, 0x2F88, 0x2892, 0xFFA4, 0x28BC, 0x2F88, 0x7214, 0x6FBE, 0xAA8A, 0xAA88, 0x2AAC, 0x2AAA, 0x2BCA, 0x2228, 0x2210}},
    {15364, 2, {0x2014, 0x2792, 0x249E, 0x24F0, 0xF794, 0x248A, 0x2496, 0x77E2, 0x6C94, 0xA492, 0xA79E, 0x2070, 0x2514, 0x248A, 0x2896, 0x2062}},
    {15365, 2, {0x2020, 0x2010, 0x24FE, 0x2244, 0xFA28, 0x20FE, 0x2092, 0x76FE, 0x6A92, 0xA2BA, 0xA2AA, 0x22BA, 0x2286, 0x2500, 0x28FE, 0x2000}},
    {15366, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x8282, 0x3EF8, 0x280, 0x7EFC, 0x280, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {15367, 2, {0x2114, 0x2112, 0x27D2, 0x2110, 0xF110, 0x2FFE, 0x2290, 0x76D2, 0x6A92, 0xA6D2, 0xA294, 0x26D4, 0x228A, 0x22CA, 0x2F16, 0x2022}},
    {15368, 2, {0x1010, 0xFEFE, 0x5454, 0x9292, 0x40, 0x27FC, 0x1040, 0x13F8, 0x248, 0x73F8, 0x1248, 0x13F8, 0x1040, 0x17FC, 0x2840, 0x47FE}},
    {15369, 2, {0x2010, 0x2010, 0x2F7E, 0x2410, 0xF43C, 0x2910, 0x2F7E, 0x7224, 0x6A4A, 0xAF30, 0xA210, 0x227E, 0x2338, 0x2E54, 0x2492, 0x2010}},
    {15370, 2, {0x2108, 0x2108, 0x27AA, 0x212C, 0xF108, 0x2FD4, 0x2012, 0x74A2, 0x6948, 0xA108, 0xA7AA, 0x212C, 0x2108, 0x2194, 0x2E24, 0x2442}},
    {15371, 2, {0x2100, 0x213E, 0x2508, 0x25D0, 0xF53E, 0x2522, 0x2FE2, 0x703E, 0x6922, 0xA57E, 0xA562, 0x25A2, 0x28BE, 0x2100, 0x2214, 0x2C22}},
    {15372, 2, {0x11FC, 0x1020, 0x13FE, 0x1222, 0xFDAC, 0x1020, 0x31AC, 0x3890, 0x55FE, 0x5320, 0x95FC, 0x1120, 0x11FC, 0x1120, 0x11FE, 0x1100}},
    {15373, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {15374, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0xFDFE, 0x1310, 0x35FE, 0x3910, 0x55FE, 0x5510, 0x91FE, 0x1100, 0x13DE, 0x1252, 0x1272, 0x1206}},
    {15375, 2, {0x1108, 0x11EE, 0x1294, 0x1020, 0xFBFE, 0x1000, 0x30F8, 0x3888, 0x54F8, 0x5000, 0x93FE, 0x1202, 0x12FA, 0x128A, 0x12FA, 0x1206}},
    {15376, 2, {0x2008, 0x273E, 0x2590, 0x255E, 0xFD64, 0x265E, 0x2500, 0x75DE, 0x6D52, 0xA55E, 0xA552, 0x265E, 0x2452, 0x2456, 0x24A0, 0x251E}},
    {15377, 2, {0x2020, 0x2294, 0x228A, 0x247A, 0xF800, 0x228A, 0x3144, 0x6A28, 0x6AAA, 0xA186, 0xA040, 0x2FFE, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {15378, 2, {0x22A0, 0x244C, 0x24A4, 0x260C, 0xFCA4, 0x264C, 0x34A4, 0x6FFE, 0x6802, 0xA3F0, 0xA020, 0x2040, 0x27FC, 0x2040, 0x2140, 0x2080}},
    {15379, 2, {0x27D0, 0x2510, 0x27DE, 0x2450, 0xF7E8, 0x2504, 0x27C4, 0x73F8, 0x6A08, 0xA3F8, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x2208}},
    {15380, 2, {0x2248, 0x22E8, 0x2308, 0x26EA, 0xFA1C, 0x22E8, 0x3208, 0x6AE8, 0x65B4, 0xA8E2, 0xA000, 0x23F8, 0x2110, 0x20E0, 0x2318, 0x2C06}},
    {15381, 2, {0x2110, 0x27FC, 0x2040, 0x23F8, 0xF840, 0x27FC, 0x2000, 0x73D4, 0x6912, 0xA7FE, 0xA350, 0x2534, 0x238C, 0x208A, 0x2296, 0x2102}},
    {15382, 2, {0x13D0, 0x1290, 0x13DE, 0x1268, 0xFBC4, 0x1284, 0x33F8, 0x3908, 0x55F8, 0x5108, 0x91F8, 0x1108, 0x11F8, 0x1090, 0x1112, 0x160E}},
    {15383, 2, {0x2014, 0x27FE, 0x2490, 0x24D4, 0xFC94, 0x27FA, 0x35AA, 0x6A96, 0x61A2, 0xA3F8, 0xA208, 0x23F8, 0x2040, 0x227C, 0x2540, 0x28FE}},
    {15384, 2, {0x2080, 0x20F8, 0x2108, 0x21F0, 0xF810, 0x27FE, 0x2554, 0x73BC, 0x6FCA, 0xA13E, 0xA388, 0x256A, 0x2088, 0x27FE, 0x2088, 0x2108}},
    {15385, 2, {0x2518, 0x2514, 0x2F90, 0x253E, 0xF728, 0x2268, 0x2FBE, 0x7AA8, 0x6FA8, 0xA23E, 0xAFA8, 0x2228, 0x3FE8, 0x253E, 0x28A0, 0x3060}},
    {15386, 2, {0x220C, 0x210A, 0x27C8, 0x229E, 0xF554, 0x26F4, 0x245E, 0x77D4, 0x6914, 0xA7DE, 0xA554, 0x2554, 0x25D4, 0x245E, 0x2550, 0x2490}},
    {15387, 2, {0x2110, 0x2FFE, 0x2554, 0x24E4, 0xF8A8, 0x2AAA, 0x2EEE, 0x74A4, 0x6AAA, 0xAEEE, 0xA242, 0x2FFE, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {15388, 2, {0x2200, 0x221E, 0x2FC4, 0x2208, 0xF79E, 0x2492, 0x2792, 0x749E, 0x6F92, 0xA49E, 0xA792, 0x2492, 0x2FDE, 0x2500, 0x248C, 0x2892}},
    {15389, 2, {0x40, 0x3E40, 0x2240, 0x227C, 0x2284, 0x2288, 0x2320, 0x3E20, 0x2020, 0x2050, 0x2050, 0x2250, 0x2C88, 0x3088, 0x2104, 0x202}},
    {15390, 2, {0x820, 0x820, 0x820, 0x87E, 0x4942, 0x4984, 0x4910, 0x4910, 0x4910, 0x4910, 0x4928, 0x4F28, 0x7948, 0x44, 0x84, 0x102}},
    {15391, 2, {0x2020, 0x2020, 0x7F20, 0x807E, 0x42, 0x7E84, 0x10, 0x10, 0xFC10, 0x410, 0x428, 0x528, 0x528, 0x344, 0x144, 0x82}},
    {15392, 2, {0x1020, 0x820, 0x820, 0xFF7E, 0x42, 0x84, 0x3C10, 0x2410, 0x2410, 0x2410, 0x2428, 0x2528, 0x4648, 0x4444, 0x8084, 0x102}},
    {15393, 2, {0x820, 0x820, 0x1420, 0x127E, 0x2142, 0x4084, 0x9210, 0x1210, 0x1210, 0x1210, 0x1228, 0x1228, 0x2228, 0x2244, 0x4244, 0x8282}},
    {15394, 2, {0x20, 0x7E20, 0x2420, 0x247E, 0x2442, 0x2484, 0x2410, 0xFF10, 0x2410, 0x2410, 0x2428, 0x2428, 0x2448, 0x2444, 0x4484, 0x8502}},
    {15395, 2, {0x1420, 0x1420, 0x1420, 0x547E, 0x5542, 0x5E84, 0x5410, 0x5410, 0x5410, 0x5410, 0x5428, 0x5428, 0x5D48, 0x7644, 0xC484, 0x102}},
    {15396, 2, {0x20, 0x20, 0xFFA0, 0x27E, 0x242, 0x7A84, 0x4A10, 0x4A10, 0x4A10, 0x4A10, 0x7A28, 0x4A28, 0x248, 0x244, 0xA84, 0x502}},
    {15397, 2, {0x1220, 0x5220, 0x5220, 0x527E, 0x5242, 0xFF84, 0x5210, 0x5210, 0x5210, 0x5E10, 0x4028, 0x4028, 0x4048, 0x7FC4, 0x84, 0x102}},
    {15398, 2, {0x820, 0x820, 0x820, 0x7F7E, 0x842, 0x884, 0x810, 0xFF90, 0x1010, 0x1010, 0x2428, 0x4228, 0xFF48, 0x4144, 0x84, 0x102}},
    {15399, 2, {0x1040, 0x1040, 0x2040, 0x287C, 0x4484, 0xFE88, 0x4320, 0x20, 0x7C20, 0x4450, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {15400, 2, {0x20, 0x20, 0x7E20, 0x427E, 0x4242, 0x4284, 0x4210, 0x7E10, 0x4210, 0x10, 0x2428, 0x2228, 0x4248, 0x4244, 0x8084, 0x102}},
    {15401, 2, {0x820, 0x820, 0x820, 0x7F7E, 0x4942, 0x4984, 0x4910, 0x4910, 0x7F10, 0x4810, 0x828, 0xA28, 0xF48, 0xF144, 0x4084, 0x102}},
    {15402, 2, {0x820, 0x820, 0x820, 0x7F7E, 0x842, 0x884, 0xFF10, 0x10, 0x810, 0x810, 0x7F28, 0x828, 0x848, 0xF44, 0xF084, 0x4102}},
    {15403, 2, {0x820, 0x2820, 0x2820, 0x3E7E, 0x4842, 0x884, 0x810, 0x7F10, 0x1410, 0x1410, 0x1428, 0x1428, 0x2548, 0x2644, 0x4484, 0x8102}},
    {15404, 2, {0x2020, 0x2020, 0x2020, 0x7F7E, 0x4142, 0x8184, 0x7910, 0x4910, 0x4910, 0x7910, 0x4928, 0x4928, 0x7928, 0x144, 0xA44, 0x482}},
    {15405, 2, {0x1040, 0x1040, 0xFE40, 0x287C, 0x4484, 0x8288, 0x7D20, 0x20, 0xFE20, 0x2050, 0x4050, 0x7C50, 0x488, 0x488, 0x2904, 0x1202}},
    {15406, 2, {0x20, 0x7F20, 0x4020, 0x407E, 0x5E42, 0x4084, 0x4010, 0x7F10, 0x5010, 0x5110, 0x5A28, 0x5428, 0x5228, 0x9144, 0x9844, 0x1082}},
    {15407, 2, {0x20, 0xFF20, 0x20, 0x7E, 0x7E42, 0x4284, 0x4210, 0x4210, 0x7E10, 0x10, 0x4228, 0x2228, 0x2428, 0xF44, 0xF044, 0x4082}},
    {15408, 2, {0x820, 0x820, 0x7F20, 0x87E, 0x842, 0x2A84, 0x2A10, 0x2A10, 0x5D10, 0x8890, 0x828, 0x1428, 0x1428, 0x2244, 0x4244, 0x8082}},
    {15409, 2, {0x720, 0x7820, 0x820, 0x87E, 0x7F42, 0x884, 0x7F10, 0x4910, 0x4910, 0x7F10, 0x4928, 0x4928, 0x7F28, 0x844, 0x844, 0x882}},
    {15410, 2, {0x420, 0x520, 0x4A0, 0xFFBE, 0x422, 0x444, 0x7550, 0x5590, 0x5510, 0x7310, 0x290, 0x32A8, 0xC5A8, 0x8C4, 0x1044, 0x82}},
    {15411, 2, {0x40, 0xFE40, 0x2840, 0x287C, 0x2884, 0xEE88, 0x8320, 0x8220, 0x8220, 0xEE50, 0x2850, 0x2850, 0x2E88, 0xF088, 0x4104, 0x202}},
    {15412, 2, {0x1020, 0x820, 0x7F20, 0x7E, 0x2242, 0x1484, 0xFF10, 0x10, 0x10, 0x3E10, 0x2228, 0x2228, 0x2248, 0x3E44, 0x2284, 0x102}},
    {15413, 2, {0x1020, 0x820, 0x20, 0x7F7E, 0x42, 0x2284, 0x2210, 0x5510, 0x8890, 0x810, 0xFF28, 0x828, 0x848, 0x844, 0x884, 0x902}},
    {15414, 2, {0x2220, 0x2220, 0x3A20, 0x4B3E, 0x4AA2, 0xAA44, 0x1250, 0x2890, 0x4790, 0x8010, 0x3F10, 0x2128, 0x2128, 0x2144, 0x3F44, 0x2182}},
    {15415, 2, {0x20, 0x20, 0xFFA0, 0x227E, 0x2242, 0x3A84, 0x4A10, 0x4B90, 0x4A10, 0xAA10, 0x1228, 0x1228, 0x22A8, 0x2344, 0x4244, 0x8082}},
    {15416, 2, {0x820, 0x820, 0x1420, 0x227E, 0x4142, 0xBE84, 0x810, 0x810, 0x7F10, 0x4910, 0x4928, 0x7F28, 0x848, 0x1444, 0x2284, 0x4102}},
    {15417, 2, {0x110, 0x190, 0x150, 0x7F9E, 0x4112, 0x4122, 0x5D08, 0x4148, 0x4148, 0x5D48, 0x5588, 0x5594, 0x5D54, 0x4154, 0x42D4, 0x8462}},
    {15418, 2, {0x20, 0x7720, 0x5120, 0x513E, 0x5122, 0x7744, 0x4050, 0x4090, 0x7710, 0x4110, 0x4910, 0x7528, 0x4228, 0x4244, 0x4544, 0x4982}},
    {15419, 2, {0x820, 0x820, 0xFF20, 0x83E, 0x822, 0x7E44, 0x50, 0x7E90, 0x4210, 0x4210, 0x7E10, 0x4228, 0x2428, 0xF44, 0xF044, 0x4082}},
    {15420, 2, {0x20, 0xFF20, 0x8820, 0x887E, 0xFE42, 0x8884, 0xAA10, 0xAA10, 0xDD10, 0x8810, 0x9428, 0x9228, 0xA128, 0xC044, 0xFF44, 0x82}},
    {15421, 2, {0x20, 0xFFA0, 0x8020, 0xBF7E, 0xA142, 0xBF84, 0xA110, 0xBF10, 0x8810, 0xFF90, 0x9228, 0xB228, 0x8C28, 0x9244, 0xA144, 0xFF82}},
    {15422, 2, {0x820, 0x1020, 0x7E20, 0x427E, 0x7E42, 0x4284, 0x7E10, 0x1010, 0x810, 0xFF10, 0x2028, 0x3E28, 0x2248, 0x4244, 0x4A84, 0x8502}},
    {15423, 2, {0xE40, 0xF040, 0x2240, 0x927C, 0x4484, 0x2088, 0x4520, 0xF820, 0x1020, 0x2450, 0xFE50, 0x1050, 0xFE88, 0x2888, 0x4504, 0x8202}},
    {15424, 2, {0x2020, 0x2020, 0x7F20, 0x917E, 0x5542, 0x7D84, 0x2110, 0x4A10, 0xA410, 0x7F10, 0x9128, 0x5528, 0x7D28, 0x2144, 0x4A44, 0x8482}},
    {15425, 2, {0x820, 0x2A20, 0x2A20, 0x3E7E, 0x4942, 0x4984, 0x7F10, 0x10, 0x3E10, 0x10, 0x7F28, 0x828, 0x2A48, 0x4944, 0x2884, 0x1102}},
    {15426, 2, {0x410, 0x210, 0x3FD0, 0x221E, 0x2F92, 0x22A4, 0x3FC8, 0x2288, 0x2F88, 0x2208, 0x3248, 0x2A94, 0x2714, 0x4A94, 0x5254, 0x8622}},
    {15427, 2, {0x820, 0x1420, 0x2220, 0x497E, 0xBEC2, 0x484, 0x810, 0xFF90, 0x1410, 0x7F10, 0x5528, 0x6328, 0x4148, 0x7F44, 0x4184, 0x7F02}},
    {15428, 2, {0x20, 0x7F20, 0x4920, 0x6B7E, 0x5D42, 0x4984, 0x7F10, 0x810, 0x7F10, 0x810, 0xF28, 0xF028, 0x128, 0x5544, 0x5444, 0x4082}},
    {15429, 2, {0x820, 0xFFA0, 0x820, 0x7F7E, 0x42, 0xFF84, 0x8090, 0x7F10, 0x10, 0x7F10, 0x4128, 0x7F28, 0x4128, 0x2244, 0xFC4, 0xF082}},
    {15430, 2, {0x820, 0xF20, 0x820, 0x7F7E, 0x4942, 0x4C84, 0x7910, 0x4710, 0x4010, 0x5F10, 0x4028, 0x5F28, 0x5128, 0x5F44, 0x8A44, 0x3F82}},
    {15431, 2, {0x820, 0x7F20, 0x2A20, 0x2A3E, 0x5D22, 0x844, 0xFF50, 0x90, 0x7F10, 0x4110, 0x5D10, 0x5528, 0x5D28, 0x4144, 0x7F44, 0x4182}},
    {15432, 2, {0x2220, 0x2220, 0xF7A0, 0x227E, 0x7742, 0xAA84, 0x2250, 0x10, 0x7F10, 0x10, 0xFFA8, 0x828, 0x4928, 0x88C4, 0x2844, 0x1082}},
    {15433, 2, {0x1420, 0x5520, 0x3620, 0x147E, 0xFF42, 0x2284, 0x1410, 0x7F10, 0x810, 0x7F10, 0x828, 0xFF28, 0x1C28, 0x2A44, 0x4944, 0x882}},
    {15434, 2, {0xFF20, 0x1020, 0x7E20, 0x427E, 0x7E42, 0x4284, 0xFF10, 0x9110, 0x4A10, 0xA510, 0x1C28, 0x2028, 0x7E28, 0xA244, 0x1C44, 0xE282}},
    {15435, 2, {0x3F90, 0x410, 0x7FD0, 0x445E, 0xB592, 0x424, 0x3588, 0x1208, 0x3FC8, 0x6408, 0xBF88, 0x2414, 0x3F94, 0x2414, 0x3FD4, 0x2022}},
    {15436, 2, {0x7F88, 0x2108, 0x3F08, 0x2108, 0x3F1E, 0x21D2, 0xFF24, 0x100, 0xFFC8, 0x5288, 0x7388, 0x5288, 0x7394, 0x5AD4, 0xF7A4, 0x10C2}},
    {15437, 2, {0x4848, 0x4448, 0x4048, 0x5E4E, 0xA0AA, 0xAEB2, 0xC0C4, 0x4E44, 0x4044, 0x8E84, 0xAAA4, 0xEAEA, 0xA0A, 0xAEAA, 0xAAAA, 0x10}},
    {15438, 2, {0x7F20, 0x5520, 0x7F20, 0x2A7E, 0x4942, 0x2A84, 0x6110, 0x3E10, 0x4410, 0xFF10, 0x4928, 0x7F28, 0x4928, 0x7F44, 0x44, 0xAA82}},
    {15439, 2, {0x0, 0x1FF0, 0x0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {15440, 2, {0x0, 0x10, 0x3F08, 0x8, 0x0, 0xFFFE, 0x0, 0x400, 0x400, 0x2700, 0x2400, 0x2400, 0x2400, 0x2780, 0xF800, 0x4000}},
    {15441, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x200, 0x400, 0xFF0, 0x1010, 0x6820, 0x440, 0x280, 0x300, 0x1C00, 0xE000}},
    {15442, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x10, 0x50, 0x20}},
    {15443, 2, {0x820, 0x820, 0x850, 0x4850, 0x4E88, 0x4924, 0x4A12, 0x4810, 0x49FC, 0x4804, 0x4808, 0x4888, 0x4E50, 0xF020, 0x10, 0x10}},
    {15444, 2, {0x440, 0x444, 0x1748, 0x1470, 0x1442, 0x1742, 0x783E, 0x200, 0x7F0, 0x810, 0x3420, 0x240, 0x80, 0x300, 0xC00, 0x7000}},
    {15445, 2, {0x1000, 0x11FC, 0x1004, 0x10FC, 0x5004, 0x5DFC, 0x5000, 0x53FE, 0x5222, 0x5020, 0x51FC, 0x5D24, 0x7124, 0xC134, 0x128, 0x20}},
    {15446, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x5088, 0x5C88, 0x5154, 0x5222, 0x5000, 0x5020, 0x53FE, 0x5C20, 0x7020, 0xC020, 0x20, 0x20}},
    {15447, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x51FC, 0x5D04, 0x51FC, 0x5104, 0x51FC, 0x5028, 0x5024, 0x5DFE, 0x7050, 0xC088, 0x104, 0x202}},
    {15448, 2, {0x3FFE, 0x2914, 0x2F12, 0x2910, 0x2F7E, 0x2010, 0x2F28, 0x2928, 0x2F28, 0x2944, 0x2F44, 0x2982, 0x4B40, 0x447C, 0x8440, 0x1FFE}},
    {15449, 2, {0x0, 0x3FE, 0xFC40, 0x2040, 0x2080, 0x3C80, 0x4500, 0x45FC, 0x6404, 0x9404, 0x804, 0x804, 0x1004, 0x2004, 0x4028, 0x8010}},
    {15450, 2, {0x0, 0xFC, 0xFC84, 0x2084, 0x2084, 0x3CFC, 0x4484, 0x4484, 0x6484, 0x94FC, 0x884, 0x884, 0x1104, 0x2104, 0x4214, 0x8408}},
    {15451, 2, {0x0, 0x0, 0xFDFE, 0x2020, 0x2020, 0x3C7C, 0x4444, 0x4484, 0x6484, 0x9548, 0xA28, 0x810, 0x1010, 0x2020, 0x4040, 0x8080}},
    {15452, 2, {0x0, 0x20, 0xFC20, 0x2020, 0x2020, 0x3C20, 0x4520, 0x453C, 0x6520, 0x9520, 0x920, 0x920, 0x12A0, 0x2260, 0x443E, 0x8800}},
    {15453, 2, {0x10, 0x10, 0xFF90, 0x2210, 0x227E, 0x3A12, 0x4A12, 0x4B92, 0x4A12, 0xAA12, 0x1212, 0x1212, 0x2292, 0x2322, 0x422A, 0x8044}},
    {15454, 2, {0x80, 0x40, 0xFC40, 0x2000, 0x23FC, 0x3C00, 0x4408, 0x4508, 0x6508, 0x9490, 0x890, 0x890, 0x10A0, 0x2020, 0x47FE, 0x8000}},
    {15455, 2, {0x10, 0x10, 0xFC10, 0x21FE, 0x2112, 0x3D14, 0x4510, 0x45FC, 0x6544, 0x9544, 0x928, 0x928, 0x1110, 0x2228, 0x4244, 0x8482}},
    {15456, 2, {0x20, 0x20, 0xFC20, 0x2020, 0x23FE, 0x3C20, 0x4420, 0x4420, 0x65FC, 0x9504, 0x904, 0x904, 0x1104, 0x2104, 0x41FC, 0x8104}},
    {15457, 2, {0x20, 0x120, 0xFD20, 0x21FC, 0x2120, 0x3E20, 0x4420, 0x47FE, 0x6490, 0x9490, 0x890, 0x890, 0x1112, 0x2112, 0x4212, 0x840E}},
    {15458, 2, {0x40, 0x20, 0xFDFC, 0x2104, 0x2104, 0x3DFC, 0x4504, 0x4504, 0x65FC, 0x9520, 0x922, 0x914, 0x1108, 0x2144, 0x4182, 0x8100}},
    {15459, 2, {0x4, 0x1E, 0xFDE0, 0x2022, 0x2112, 0x3C94, 0x4480, 0x4420, 0x65FE, 0x9444, 0x884, 0x9C8, 0x1030, 0x2028, 0x40C4, 0x8302}},
    {15460, 2, {0x80, 0x86, 0xF898, 0x2090, 0x23D0, 0x3890, 0x489E, 0x4994, 0x69D4, 0x9AB4, 0xA94, 0x1494, 0x1094, 0x2094, 0x4094, 0x80A4}},
    {15461, 2, {0x20, 0x20, 0xF9FC, 0x2020, 0x2020, 0x3BFE, 0x4888, 0x4944, 0x6A42, 0x98F8, 0x988, 0x1250, 0x1020, 0x2050, 0x4188, 0x8606}},
    {15462, 2, {0x20, 0x20, 0xFDFC, 0x2050, 0x2088, 0x3D04, 0x47FE, 0x4408, 0x65E8, 0x9528, 0x928, 0x9E8, 0x1128, 0x2008, 0x4028, 0x8010}},
    {15463, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2504, 0x3900, 0x49DE, 0x4A52, 0x6A52, 0x9B52, 0xC9A, 0x1094, 0x1110, 0x2112, 0x4212, 0x840E}},
    {15464, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2404, 0x3800, 0x4BFE, 0x4820, 0x6820, 0x9920, 0x93C, 0x1120, 0x12A0, 0x2260, 0x443E, 0x8800}},
    {15465, 2, {0x20, 0x20, 0xF9FC, 0x2020, 0x2020, 0x3BFE, 0x4802, 0x4894, 0x6850, 0x9910, 0x890, 0x13FE, 0x1028, 0x2044, 0x4082, 0x8302}},
    {15466, 2, {0x0, 0x1FC, 0xFD04, 0x21FC, 0x2110, 0x3DFE, 0x4510, 0x454A, 0x6586, 0x9400, 0x9FC, 0x904, 0x11FC, 0x2104, 0x41FC, 0x8104}},
    {15467, 2, {0x20, 0xFC20, 0x23FE, 0x2020, 0x21FC, 0x3D24, 0x45AC, 0x4574, 0x6524, 0x95FC, 0x820, 0x870, 0x10A8, 0x2124, 0x4222, 0x8020}},
    {15468, 2, {0x0, 0x3F8, 0xFA48, 0x2248, 0x23F8, 0x3A48, 0x4A48, 0x4BF8, 0x6800, 0x9FFE, 0xA40, 0x1244, 0x1228, 0x2290, 0x4308, 0x8206}},
    {15469, 2, {0x80, 0x338, 0xFA28, 0x2228, 0x23A8, 0x3A46, 0x4A00, 0x4A7C, 0x6BA4, 0x9A24, 0xA28, 0x13A8, 0x1610, 0x2228, 0x4244, 0x8282}},
    {15470, 2, {0x20, 0x20, 0xFDFC, 0x2124, 0x21FC, 0x3C20, 0x47FE, 0x4400, 0x65FC, 0x9504, 0x924, 0x924, 0x1124, 0x2050, 0x4088, 0x8304}},
    {15471, 2, {0x20, 0x124, 0xFD24, 0x2124, 0x21FC, 0x3C00, 0x47FE, 0x4400, 0x65FC, 0x9504, 0x904, 0x9FC, 0x1088, 0x2050, 0x43FE, 0x8000}},
    {15472, 2, {0x40, 0xFC20, 0x23FE, 0x2242, 0x2020, 0x3DFC, 0x4488, 0x4450, 0x67FE, 0x9420, 0x820, 0x9FC, 0x1020, 0x2020, 0x4020, 0x8020}},
    {15473, 2, {0x20, 0x40, 0xFBFC, 0x2224, 0x2224, 0x3BFC, 0x4A24, 0x4A44, 0x6BFC, 0x9840, 0x8A8, 0x10B4, 0x113C, 0x2122, 0x4222, 0x841E}},
    {15474, 2, {0x208, 0x108, 0xF910, 0x2000, 0x27FC, 0x3800, 0x4910, 0x4A08, 0x6C04, 0x9BF8, 0xAA8, 0x12A8, 0x12A8, 0x22A8, 0x4FFE, 0x8000}},
    {15475, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1010, 0x2FE8, 0xC206, 0xFE0, 0x440, 0x3F80}},
    {15476, 2, {0x20, 0x3FE, 0xF820, 0x21FC, 0x2020, 0x3BFE, 0x4800, 0x49FC, 0x6904, 0x99FC, 0x904, 0x11FC, 0x1104, 0x21FC, 0x4088, 0x8104}},
    {15477, 2, {0x20, 0x1FC, 0xFC20, 0x2088, 0x23FE, 0x3C88, 0x45FC, 0x4504, 0x65FC, 0x9504, 0x9FC, 0x904, 0x11FC, 0x2088, 0x4104, 0x8202}},
    {15478, 2, {0x82, 0x82, 0xFBF4, 0x2088, 0x2080, 0x3BE2, 0x4802, 0x4BE4, 0x6A28, 0x9A20, 0xBE0, 0x1222, 0x1142, 0x2074, 0x4788, 0x8210}},
    {15479, 2, {0xA0, 0x90, 0xFDFE, 0x2320, 0x21FC, 0x3D20, 0x45FC, 0x4520, 0x65FE, 0x9500, 0x820, 0xBFE, 0x1070, 0x20A8, 0x4326, 0x8020}},
    {15480, 2, {0x10, 0x3C8, 0xFA7E, 0x2240, 0x2262, 0x3BD4, 0x4A00, 0x4A3E, 0x6BC8, 0x9A48, 0xE7E, 0x1248, 0x1248, 0x23C8, 0x4248, 0x8008}},
    {15481, 2, {0x0, 0x3FE, 0xFA22, 0x22CC, 0x2244, 0x3BEE, 0x4A44, 0x4AEE, 0x6B54, 0x9A44, 0xA10, 0x1290, 0x129E, 0x2490, 0x4490, 0x8BFE}},
    {15482, 2, {0x40, 0x7C, 0xF840, 0x23FC, 0x2244, 0x3BF0, 0x4A44, 0x4AFC, 0x6AA8, 0x9AF8, 0xAA8, 0x12F8, 0x1200, 0x25FC, 0x4554, 0x8BFE}},
    {15483, 2, {0x40, 0x1FFE, 0xE400, 0x47FC, 0x4000, 0x77FC, 0x5404, 0x97FC, 0x5110, 0x2EAE, 0x2BFA, 0x2E4A, 0x4AEE, 0x4E4A, 0x8BFA, 0x1650}},
    {15484, 2, {0x0, 0x3E78, 0x2248, 0x2248, 0x2248, 0x2286, 0x2200, 0x3EFC, 0x2044, 0x2044, 0x2028, 0x2228, 0x2C10, 0x3028, 0x2044, 0x182}},
    {15485, 2, {0x800, 0x878, 0x848, 0x7F48, 0x4948, 0x886, 0x800, 0x14FC, 0x1444, 0x1444, 0x1428, 0x2528, 0x2610, 0x4428, 0x8044, 0x182}},
    {15486, 2, {0x1000, 0x878, 0x848, 0xFF48, 0x848, 0x1086, 0x2200, 0x7CFC, 0x944, 0x1244, 0x2428, 0xC828, 0x1410, 0x2228, 0xC044, 0x182}},
    {15487, 2, {0x1000, 0x1078, 0xFE48, 0x1048, 0x1048, 0x7C86, 0x0, 0xFC, 0xFE44, 0x8244, 0x8228, 0xBA28, 0x8210, 0x8228, 0x8A44, 0x8582}},
    {15488, 2, {0x0, 0x7F78, 0x4048, 0x4048, 0x5E48, 0x4086, 0x4000, 0x7EFC, 0x5044, 0x5144, 0x5A28, 0x5428, 0x5210, 0x5128, 0x9844, 0x1182}},
    {15489, 2, {0x2000, 0x1078, 0xFE48, 0x8248, 0x2848, 0x4486, 0x8200, 0xFC, 0x7C44, 0x1044, 0x1028, 0x1028, 0x1E10, 0xF028, 0x4044, 0x182}},
    {15490, 2, {0x1000, 0x9278, 0x5248, 0x5448, 0x1048, 0xFE86, 0x8200, 0x82FC, 0xBA44, 0xAA44, 0xAA28, 0xAA28, 0xBA10, 0x8228, 0x8A44, 0x8582}},
    {15491, 2, {0x1000, 0x1078, 0x1E48, 0x1048, 0x7E48, 0x4286, 0x7E00, 0x42FC, 0x7E44, 0x44, 0x2028, 0x3E28, 0x2010, 0x2228, 0x2C44, 0x3182}},
    {15492, 2, {0x1000, 0x1078, 0xFE48, 0x1048, 0x7C48, 0x86, 0xFE00, 0x82FC, 0x44, 0x7C44, 0x1028, 0x7C28, 0x1010, 0x1E28, 0xF044, 0x4182}},
    {15493, 2, {0x2000, 0x1078, 0xFE48, 0x8248, 0x1048, 0x7C86, 0x1000, 0x7CFC, 0x1044, 0xFE44, 0x1028, 0x7C28, 0x4410, 0x4428, 0x7C44, 0x4582}},
    {15494, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4148, 0x8284, 0xC60, 0x3118, 0xCFE6, 0x40, 0x680, 0x180}},
    {15495, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4848, 0x9FF4, 0x2820, 0x7C0, 0x1830, 0xE18E, 0xC40, 0x300}},
    {15496, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x8086, 0x7FF8, 0x808, 0x1030, 0x3FF8, 0xD008, 0x1FF8}},
    {15497, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x2044, 0xCE44, 0xAA28, 0xEE28, 0xAA10, 0xEE28, 0x2844, 0x4982}},
    {15498, 2, {0x1000, 0x1F38, 0x1028, 0xFFA8, 0x80A8, 0x7F46, 0x2A00, 0x557C, 0xA224, 0x7F24, 0xA2A8, 0x3E28, 0x2210, 0x3E28, 0x2244, 0x3E82}},
    {15499, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2008, 0x2008, 0x3FF8, 0x0, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {15500, 2, {0x800, 0x1DF8, 0x7048, 0x1048, 0x1050, 0x1C5C, 0x7084, 0x1084, 0x1084, 0x1D04, 0xF128, 0x1212, 0x1002, 0x1002, 0xFFE, 0x0}},
    {15501, 2, {0x1010, 0x1038, 0x11C0, 0xFE40, 0x1078, 0x11C0, 0xFC40, 0x447C, 0x47C0, 0x2844, 0x2844, 0x1044, 0x183C, 0x2400, 0x43FE, 0x8000}},
    {15502, 2, {0x408, 0xE1C, 0x70E0, 0x1020, 0x1020, 0x1C3C, 0x70E0, 0x1020, 0x1020, 0x1E3C, 0xF0E0, 0x1022, 0x1422, 0x1822, 0x101E, 0x0}},
    {15503, 2, {0x408, 0x243C, 0x25E0, 0x2420, 0x2420, 0x3E3C, 0x21E0, 0x2020, 0x3C20, 0x243E, 0x27E0, 0x2420, 0x2422, 0x4422, 0x4422, 0x841E}},
    {15504, 2, {0x820, 0x1C20, 0x7050, 0x1050, 0x1088, 0x1D44, 0x7222, 0x10F8, 0x1008, 0x1C10, 0xF010, 0x1022, 0x1022, 0x1002, 0xFFE, 0x0}},
    {15505, 2, {0x820, 0x1C20, 0x7020, 0x1024, 0x11A8, 0x1CB0, 0x70B0, 0x1128, 0x1128, 0x1D24, 0xF224, 0x10A0, 0x1040, 0x1002, 0xFFE, 0x0}},
    {15506, 2, {0x2008, 0x201C, 0x40E0, 0x7E20, 0x8220, 0x23C, 0x7AE0, 0x4A20, 0x4A20, 0x4A3C, 0x7AE0, 0x4A22, 0x222, 0x222, 0x141E, 0x800}},
    {15507, 2, {0x828, 0x1C24, 0x7024, 0x11FE, 0x1120, 0x1D28, 0x71A8, 0x1150, 0x1114, 0x1D34, 0xF14C, 0x1104, 0x1200, 0x1002, 0xFFE, 0x0}},
    {15508, 2, {0x1004, 0x80E, 0x7F70, 0x10, 0x2210, 0x141C, 0xFF70, 0x810, 0x810, 0x81E, 0x7F70, 0x810, 0x812, 0x812, 0x812, 0x80E}},
    {15509, 2, {0x408, 0x1E1C, 0xE0E0, 0x2220, 0x9220, 0x543C, 0x40E0, 0x420, 0xFF20, 0x43C, 0x44E0, 0x2422, 0x2422, 0x422, 0x141E, 0x800}},
    {15510, 2, {0x804, 0x490E, 0x2970, 0x2A10, 0x810, 0x7F1C, 0x4170, 0x4110, 0x7F10, 0x411E, 0x4170, 0x7F10, 0x4112, 0x4112, 0x4512, 0x420E}},
    {15511, 2, {0x408, 0x1E1C, 0xE0E0, 0x220, 0x9220, 0x4C3C, 0xE0, 0xFE20, 0x420, 0x83C, 0xEE0, 0xF822, 0x4822, 0x822, 0x281E, 0x1000}},
    {15512, 2, {0x8, 0x7E1C, 0x42E0, 0x4220, 0x7E20, 0x423C, 0x42E0, 0x7E20, 0x4220, 0x423C, 0x7EE0, 0x22, 0x2422, 0x2222, 0x411E, 0x8100}},
    {15513, 2, {0x840, 0x1C20, 0x71FE, 0x1040, 0x1088, 0x1DFC, 0x7004, 0x10A8, 0x10A8, 0x1CA8, 0xF12A, 0x1226, 0x1000, 0x1002, 0xFFE, 0x0}},
    {15514, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x92A, 0x7230, 0x10C0, 0x1700, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {15515, 2, {0x2000, 0x3E7C, 0x4844, 0x844, 0xFF44, 0x1444, 0x227C, 0x4000, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {15516, 2, {0x1008, 0x521C, 0x54E0, 0x9020, 0x2820, 0x443C, 0x82E0, 0x1020, 0x1020, 0x523C, 0x54E0, 0x9022, 0x2822, 0x2422, 0x421E, 0x8000}},
    {15517, 2, {0x850, 0x1C48, 0x70FE, 0x1190, 0x1290, 0x1CFE, 0x7090, 0x1090, 0x10FE, 0x1C90, 0xF090, 0x10FE, 0x1080, 0x1002, 0xFFE, 0x0}},
    {15518, 2, {0x1004, 0x80E, 0x70, 0x7F10, 0x10, 0x221C, 0x2270, 0x5510, 0x8890, 0x81E, 0xFF70, 0x810, 0x812, 0x812, 0x812, 0x80E}},
    {15519, 2, {0x2204, 0x220E, 0x7F70, 0x2210, 0x2A10, 0x81C, 0x3E70, 0x2A10, 0x2A10, 0x2A1E, 0xFF70, 0x810, 0x1412, 0x1212, 0x2212, 0x400E}},
    {15520, 2, {0x820, 0x820, 0x7E7E, 0x8C4, 0xFF28, 0x1C10, 0x2A28, 0x48C6, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {15521, 2, {0x804, 0x80E, 0x7F70, 0x1C10, 0x2A10, 0x491C, 0x70, 0x7E10, 0x10, 0xFF1E, 0x870, 0x4A10, 0x4912, 0x8912, 0x2812, 0x100E}},
    {15522, 2, {0x1004, 0x80E, 0x7E70, 0x4210, 0x4210, 0x7E1C, 0x4270, 0x4010, 0x7F10, 0x551E, 0x5570, 0x7F10, 0xD512, 0x5512, 0x5512, 0x430E}},
    {15523, 2, {0x5008, 0x961C, 0x92E0, 0x9220, 0xD620, 0x923C, 0x92E0, 0xFE20, 0x1020, 0xFE3C, 0x2E0, 0x6422, 0x1822, 0x2422, 0xC21E, 0x0}},
    {15524, 2, {0x1004, 0x80E, 0x7F70, 0x10, 0x3E10, 0x221C, 0xFF70, 0x2210, 0x3E10, 0x189E, 0x3570, 0x5210, 0x9112, 0x1492, 0x1812, 0x100E}},
    {15525, 2, {0x1004, 0x80E, 0x7F70, 0x4810, 0x5E10, 0x4A1C, 0x7F70, 0x4A10, 0x5E10, 0x481E, 0x5E70, 0x5210, 0x5212, 0x5212, 0x5E12, 0x920E}},
    {15526, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1106, 0x7C18, 0x44E0, 0x7C20, 0x4438, 0x7CE0, 0x1020, 0xFE3C, 0x11E0, 0x1022, 0x1022, 0x101E}},
    {15527, 2, {0x1000, 0x3BF0, 0xE150, 0x2110, 0x20A0, 0x3840, 0xE1B0, 0x264E, 0x21F0, 0x3950, 0xE1F0, 0x2048, 0x23FA, 0x200A, 0x2002, 0x1FFE}},
    {15528, 2, {0x2408, 0x241C, 0xFEE0, 0x2420, 0x20, 0xFE3C, 0x24E0, 0x3C20, 0x2420, 0x3C3C, 0x24E0, 0x2E22, 0xF422, 0x422, 0x41E, 0x400}},
    {15529, 2, {0x1000, 0x3BF8, 0xE208, 0x23F8, 0x2208, 0x3BF8, 0xE000, 0x27BC, 0x2084, 0x3CA4, 0xE294, 0x24A4, 0x2296, 0x210A, 0x2002, 0x1FFE}},
    {15530, 2, {0xA04, 0x740E, 0x1570, 0x5210, 0x2210, 0x211C, 0x5D70, 0x8010, 0x7E10, 0x421E, 0x4270, 0x7E10, 0x4212, 0x2412, 0xF12, 0xF00E}},
    {15531, 2, {0x2808, 0xAA1C, 0x6CE0, 0x2820, 0xFE20, 0x443C, 0x28E0, 0xFE20, 0x1020, 0x7C3C, 0x10E0, 0xFE22, 0x1022, 0x2822, 0x441E, 0x8200}},
    {15532, 2, {0x2808, 0x241C, 0x7EE0, 0xC820, 0x4820, 0x7E3C, 0x48E0, 0x4820, 0x7E20, 0x483C, 0x48E0, 0x7E22, 0x4022, 0xAA22, 0xAA1E, 0x0}},
    {15533, 2, {0x804, 0x290E, 0x2970, 0x4A10, 0x1410, 0x221C, 0x4170, 0x10, 0x2210, 0x2A9E, 0xB770, 0x6210, 0x2212, 0x5212, 0x4512, 0x888E}},
    {15534, 2, {0x1210, 0x3A10, 0xE3DE, 0x2528, 0x2084, 0x3840, 0xE0A0, 0x2110, 0x22E8, 0x3C04, 0xE1F0, 0x2110, 0x2112, 0x21F2, 0x2002, 0x1FFE}},
    {15535, 2, {0x4, 0x7A0E, 0x5270, 0x7B90, 0x4A10, 0x7B1C, 0x54F0, 0x7890, 0x10, 0x7F1E, 0x5570, 0x5510, 0x5512, 0x5792, 0xFC12, 0xE}},
    {15536, 2, {0x804, 0xFF0E, 0x9170, 0x4A10, 0xA510, 0x1C1C, 0x70, 0x7F10, 0x5510, 0xFF9E, 0x70, 0x7F10, 0x812, 0x812, 0x2812, 0x100E}},
    {15537, 2, {0x1124, 0x3A48, 0xE124, 0x23FC, 0x2294, 0x3A64, 0xE294, 0x23FC, 0x2248, 0x3B68, 0xE248, 0x236A, 0x2246, 0x2362, 0x2002, 0x1FFE}},
    {15538, 2, {0x7F04, 0x490E, 0x7F70, 0x4910, 0xFF90, 0xAA9C, 0xFFF0, 0xAA90, 0xFF90, 0x1E, 0xFFF0, 0xA290, 0x3E12, 0x2212, 0x2792, 0xF80E}},
    {15539, 2, {0x820, 0x1C20, 0xF120, 0x9120, 0x91FC, 0x9120, 0xFE20, 0x9020, 0x93FE, 0x9020, 0x9050, 0x8A50, 0xA688, 0xD288, 0x904, 0x202}},
    {15540, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x410, 0x7FD0, 0x4450, 0x4450, 0x7FD2, 0x444A, 0x40A, 0x406, 0x402}},
    {15541, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x2490, 0x1510, 0x7FD0, 0x4050, 0x7FD0, 0x4052, 0x7FCA, 0x404A, 0x4146, 0x4082}},
    {15542, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x7F0, 0x410, 0x7FD0, 0x5550, 0x4A52, 0x64CA, 0x4A4A, 0x5546, 0x7FC2}},
    {15543, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x2A90, 0x3B90, 0x10, 0x3F90, 0x2490, 0x3F92, 0x248A, 0xFFEA, 0x406, 0x402}},
    {15544, 2, {0x0, 0x2000, 0x13FC, 0x1008, 0x8010, 0x4020, 0x4840, 0x880, 0x1080, 0x1100, 0xE100, 0x2202, 0x2202, 0x2202, 0x21FE, 0x0}},
    {15545, 2, {0x0, 0x21F0, 0x1110, 0x1110, 0x8110, 0x4110, 0x4910, 0x910, 0x1110, 0x1110, 0xE210, 0x2212, 0x2212, 0x2412, 0x240E, 0x800}},
    {15546, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8040, 0x404E, 0x4870, 0xFC0, 0x1040, 0x1040, 0xE040, 0x2042, 0x2042, 0x2042, 0x203E, 0x0}},
    {15547, 2, {0x840, 0x840, 0x840, 0x1044, 0x1768, 0x3170, 0x3160, 0x5150, 0x9150, 0x1248, 0x1248, 0x1444, 0x1842, 0x1040, 0x1140, 0x1080}},
    {15548, 2, {0x0, 0x2008, 0x1208, 0x1208, 0x8208, 0x4110, 0x4910, 0x910, 0x10A0, 0x10A0, 0xE040, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {15549, 2, {0x0, 0x23FC, 0x1084, 0x1084, 0x8084, 0x4284, 0x4A84, 0xA84, 0x1484, 0x1084, 0xE084, 0x2084, 0x2104, 0x2104, 0x2228, 0x410}},
    {15550, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x100, 0x3D84, 0x588, 0x950, 0x920, 0x1110, 0x2108, 0xC106, 0x500, 0x200}},
    {15551, 2, {0x40, 0x2020, 0x1020, 0x13FE, 0x8200, 0x4200, 0x4A00, 0xA00, 0x1200, 0x1200, 0xE200, 0x2200, 0x2200, 0x2400, 0x2400, 0x800}},
    {15552, 2, {0x90, 0x2088, 0x1088, 0x1080, 0x80BE, 0x47C0, 0x4080, 0x1044, 0x1044, 0x2048, 0xE030, 0x2022, 0x2052, 0x218A, 0x2606, 0x2}},
    {15553, 2, {0x0, 0x23FE, 0x1200, 0x1220, 0x8220, 0x4220, 0x4A20, 0xA20, 0x1220, 0x1250, 0xE250, 0x2250, 0x2488, 0x2488, 0x2904, 0x1202}},
    {15554, 2, {0x20, 0x2120, 0x1120, 0x1110, 0x8110, 0x4248, 0x4248, 0x1444, 0x1882, 0x2080, 0xE110, 0x2108, 0x2208, 0x27FC, 0x2204, 0x0}},
    {15555, 2, {0x80, 0x2080, 0x13F0, 0x1090, 0x8090, 0x4112, 0x4912, 0xA4E, 0x1440, 0x1040, 0xEFFE, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {15556, 2, {0x80, 0x2080, 0x1080, 0x10FC, 0x8104, 0x4108, 0x4240, 0x1440, 0x1040, 0x20A0, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {15557, 2, {0x0, 0x2000, 0x17FE, 0x1020, 0x8020, 0x4040, 0x4840, 0x8D0, 0x1148, 0x1244, 0xE442, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {15558, 2, {0x10, 0x2110, 0x1090, 0x1090, 0x8010, 0x4210, 0x4910, 0x910, 0x101E, 0x17F0, 0xE010, 0x2010, 0x2010, 0x2010, 0x2010, 0x10}},
    {15559, 2, {0x20, 0x2110, 0x1108, 0x1204, 0x8204, 0x4410, 0x4110, 0x1110, 0x10A0, 0x20A0, 0xE040, 0x20A0, 0x2110, 0x2208, 0x2404, 0x802}},
    {15560, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FC, 0x4040, 0x4950, 0x950, 0x1150, 0x1150, 0xE152, 0x2252, 0x224E, 0x2440, 0x2040, 0x40}},
    {15561, 2, {0x0, 0x21FC, 0x1104, 0x1104, 0x8104, 0x41FC, 0x4904, 0x904, 0x1104, 0x11FC, 0xE104, 0x2104, 0x2204, 0x2204, 0x2414, 0x808}},
    {15562, 2, {0x80, 0x2080, 0x1100, 0x11FE, 0x8210, 0x4210, 0x4510, 0x1110, 0x1110, 0x20A0, 0xE0A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0x406}},
    {15563, 2, {0x8, 0x203C, 0x13C0, 0x1200, 0x8200, 0x4200, 0x4A00, 0xBFC, 0x1220, 0x1220, 0xE220, 0x2220, 0x2220, 0x2220, 0x2FFE, 0x0}},
    {15564, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x8444, 0x47FC, 0x4444, 0x1444, 0x1444, 0x27FC, 0xE444, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {15565, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x8080, 0x40A0, 0x4920, 0x9FC, 0x1324, 0x1524, 0xE124, 0x2124, 0x2134, 0x2128, 0x2020, 0x20}},
    {15566, 2, {0x40, 0x2040, 0x1080, 0x1110, 0x8208, 0x47FC, 0x4004, 0x1110, 0x1110, 0x27FE, 0xE110, 0x2110, 0x2210, 0x2210, 0x2410, 0x810}},
    {15567, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8444, 0x4444, 0x4444, 0x1444, 0x17FC, 0x2040, 0xE240, 0x2140, 0x2080, 0x2140, 0x2230, 0xC0E}},
    {15568, 2, {0x110, 0x2110, 0x1110, 0x1110, 0x87D0, 0x4110, 0x4310, 0x1390, 0x1550, 0x2510, 0xE910, 0x2112, 0x2112, 0x2112, 0x210E, 0x100}},
    {15569, 2, {0x0, 0x2000, 0x13F8, 0x12A8, 0x82A8, 0x42A8, 0x4AA8, 0xAA8, 0x12A8, 0x12A8, 0xE2A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000, 0x0}},
    {15570, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x80A0, 0x4110, 0x4208, 0x1426, 0x1020, 0x2040, 0xE080, 0x2108, 0x2204, 0x27FE, 0x2202, 0x0}},
    {15571, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FE, 0x4080, 0x4880, 0x900, 0x13FC, 0x1504, 0xE904, 0x2104, 0x2104, 0x2104, 0x21FC, 0x104}},
    {15572, 2, {0x4, 0x2004, 0x1784, 0x1084, 0x80BE, 0x4484, 0x4284, 0x12A4, 0x1114, 0x2114, 0xE284, 0x2284, 0x2484, 0x2804, 0x2014, 0x8}},
    {15573, 2, {0x0, 0x2108, 0x1088, 0x1090, 0x8000, 0x43FE, 0x4800, 0x800, 0x1000, 0x11FC, 0xE000, 0x2000, 0x2000, 0x2000, 0x23FE, 0x0}},
    {15574, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x8080, 0x4080, 0x4940, 0x944, 0x1348, 0x1530, 0xE920, 0x2110, 0x2108, 0x2144, 0x2182, 0x100}},
    {15575, 2, {0x0, 0x23F8, 0x1108, 0x1150, 0x80A0, 0x4040, 0x41B0, 0x160E, 0x1040, 0x2040, 0xE3F8, 0x2040, 0x2040, 0x2040, 0x27FC, 0x0}},
    {15576, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x8080, 0x4110, 0x4208, 0x17FC, 0x1124, 0x2120, 0xE120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {15577, 2, {0x40, 0x2080, 0x1108, 0x1204, 0x87FE, 0x4002, 0x4800, 0xBFC, 0x1204, 0x13FC, 0xE204, 0x23FC, 0x2204, 0x2204, 0x2214, 0x208}},
    {15578, 2, {0x28, 0x2024, 0x1024, 0x1020, 0x83FE, 0x4220, 0x4A24, 0xA24, 0x1324, 0x12A8, 0xE228, 0x2210, 0x2212, 0x242A, 0x2446, 0x882}},
    {15579, 2, {0x40, 0x2140, 0x165C, 0x1444, 0x8444, 0x475C, 0x4444, 0x1444, 0x17FC, 0x2444, 0xE0A0, 0x20A0, 0x2110, 0x2208, 0x2404, 0x802}},
    {15580, 2, {0x90, 0x2090, 0x1090, 0x1110, 0x81FE, 0x4310, 0x4D10, 0x938, 0x1138, 0x1154, 0xE154, 0x2192, 0x2110, 0x2110, 0x2110, 0x110}},
    {15581, 2, {0x100, 0x2108, 0x111C, 0x1270, 0x8210, 0x4610, 0x4A10, 0x1210, 0x12FE, 0x2210, 0xE210, 0x2210, 0x2210, 0x2210, 0x227C, 0x200}},
    {15582, 2, {0x110, 0x2110, 0x1110, 0x1210, 0x82FE, 0x4692, 0x5692, 0x1A92, 0x1292, 0x22FE, 0xE292, 0x2210, 0x2210, 0x2210, 0x2210, 0x210}},
    {15583, 2, {0x440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x820, 0x1110, 0x2928, 0xC546, 0x380, 0x540, 0x920, 0x3110, 0x500, 0x200}},
    {15584, 2, {0x10, 0x2410, 0x1210, 0x1210, 0x80FE, 0x4010, 0x4610, 0x1228, 0x1224, 0x2244, 0xE242, 0x2282, 0x2200, 0x2500, 0x28FE, 0x0}},
    {15585, 2, {0x408, 0x4208, 0x2208, 0x2008, 0xFE, 0x8E08, 0x4248, 0x5228, 0x1228, 0x2208, 0xE208, 0x2228, 0x2210, 0x2500, 0x28FE, 0x0}},
    {15586, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x8040, 0x47FE, 0x48A0, 0x910, 0x1208, 0x1516, 0xE110, 0x2110, 0x2110, 0x2210, 0x2210, 0x410}},
    {15587, 2, {0x40, 0x2040, 0x1044, 0x13F4, 0x8048, 0x4050, 0x47FE, 0x1040, 0x1080, 0x2184, 0xE298, 0x24E0, 0x2882, 0x2082, 0x207E, 0x0}},
    {15588, 2, {0x110, 0x4910, 0x2510, 0x2510, 0x110, 0x8FFE, 0x4110, 0x5110, 0x1598, 0x2554, 0xE954, 0x2210, 0x2210, 0x2410, 0x2810, 0x1010}},
    {15589, 2, {0x0, 0x27FC, 0x1040, 0x1040, 0x83F8, 0x4040, 0x4040, 0x17FC, 0x1110, 0x2110, 0xEFFE, 0x2110, 0x2110, 0x2210, 0x2210, 0x410}},
    {15590, 2, {0x14, 0x4012, 0x2010, 0x2FFE, 0x10, 0x8290, 0x4292, 0x5292, 0x1FD2, 0x2294, 0xE294, 0x2288, 0x228A, 0x249A, 0x2426, 0x842}},
    {15591, 2, {0x0, 0x2FFE, 0x1480, 0x1490, 0x8490, 0x4790, 0x4490, 0x1490, 0x1490, 0x2790, 0xE490, 0x2492, 0x24D2, 0x2F92, 0x208E, 0x80}},
    {15592, 2, {0x124, 0x2124, 0x1248, 0x1248, 0x8124, 0x4124, 0x4000, 0x17FE, 0x1080, 0x21FC, 0xE104, 0x2288, 0x2450, 0x2020, 0x20C0, 0x700}},
    {15593, 2, {0x0, 0x407C, 0x2044, 0x2E44, 0x464, 0x8454, 0x4454, 0x5444, 0x15FE, 0x2444, 0xE744, 0x3C44, 0x2044, 0x2084, 0x2094, 0x108}},
    {15594, 2, {0x0, 0x27FC, 0x1020, 0x1040, 0x80D0, 0x4148, 0x5244, 0x1444, 0x1040, 0x2000, 0xE3F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {15595, 2, {0x100, 0x108, 0x7D10, 0x5A0, 0x960, 0x3118, 0xC506, 0x200, 0xC00, 0x3078, 0x2008, 0x3C78, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {15596, 2, {0x100, 0x2100, 0x11FE, 0x1200, 0x8400, 0x41FE, 0x4822, 0x824, 0x1120, 0x113C, 0xE120, 0x2120, 0x2120, 0x22A0, 0x247E, 0x800}},
    {15597, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x43F8, 0x4040, 0x1040, 0x17FC, 0x20E0, 0xE150, 0x2248, 0x2444, 0x2842, 0x2040, 0x40}},
    {15598, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4200, 0x4A20, 0xA22, 0x13B2, 0x12B4, 0xE2A8, 0x22A8, 0x2524, 0x2524, 0x2AA2, 0x40}},
    {15599, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x8110, 0x40A0, 0x4FFE, 0x1040, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {15600, 2, {0x40, 0x2020, 0x13FC, 0x1040, 0x8090, 0x4108, 0x4BFC, 0x804, 0x1110, 0x1110, 0xE7FE, 0x2110, 0x2110, 0x2110, 0x2210, 0x410}},
    {15601, 2, {0x0, 0x2780, 0x14BC, 0x1524, 0x8524, 0x4624, 0x4524, 0x14BC, 0x14A4, 0x24A4, 0xE6A4, 0x2524, 0x243C, 0x2424, 0x2400, 0x400}},
    {15602, 2, {0x20, 0x4820, 0x2420, 0x2020, 0x820, 0x85FE, 0x4420, 0x5020, 0x1220, 0x2450, 0xEC50, 0x2450, 0x2488, 0x24C8, 0x2524, 0x202}},
    {15603, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8040, 0x4040, 0x4BF8, 0xA08, 0x1208, 0x13F8, 0xE0A0, 0x20A0, 0x2122, 0x2122, 0x221E, 0x400}},
    {15604, 2, {0x40, 0x2028, 0x1088, 0x1290, 0x82A4, 0x42C2, 0x448A, 0x1188, 0x1678, 0x2000, 0xE040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {15605, 2, {0x38, 0x23C0, 0x1040, 0x17FC, 0x80E0, 0x4150, 0x5248, 0x1C46, 0x13F0, 0x2020, 0xE040, 0x27FC, 0x2040, 0x2040, 0x2140, 0x80}},
    {15606, 2, {0xE8, 0x4F08, 0x2228, 0x2928, 0x548, 0x8008, 0x47C8, 0x5088, 0x1108, 0x2108, 0xE1E8, 0x2F08, 0x210A, 0x210A, 0x250A, 0x206}},
    {15607, 2, {0x100, 0x2100, 0x11FC, 0x12A4, 0x84A4, 0x4124, 0x4A44, 0x894, 0x1108, 0x1000, 0xE3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x204}},
    {15608, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8108, 0x4000, 0x4890, 0x892, 0x1114, 0x1318, 0xE510, 0x2132, 0x2152, 0x2112, 0x210E, 0x100}},
    {15609, 2, {0x200, 0x4200, 0x22FE, 0x2210, 0xF90, 0x8210, 0x4610, 0x577C, 0x1690, 0x2A90, 0xEA10, 0x3210, 0x2210, 0x22FE, 0x2200, 0x200}},
    {15610, 2, {0x240, 0x4240, 0x227E, 0x2280, 0xF00, 0x8278, 0x4208, 0x5308, 0x1210, 0x26FE, 0xEA10, 0x2210, 0x2210, 0x2210, 0x2A50, 0x420}},
    {15611, 2, {0x110, 0x2110, 0x1110, 0x17FC, 0x8110, 0x4110, 0x4FFE, 0x1000, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {15612, 2, {0x50, 0x48, 0xFFFE, 0x40, 0x3E40, 0x2224, 0x3E28, 0x12, 0xE2A, 0x70C6, 0x30A, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {15613, 2, {0x20, 0x2720, 0x1120, 0x113E, 0x8140, 0x4740, 0x4490, 0x1410, 0x1410, 0x2754, 0xE152, 0x2152, 0x2192, 0x2110, 0x2550, 0x220}},
    {15614, 2, {0x208, 0x2108, 0x1110, 0x17FC, 0x8040, 0x4040, 0x43F8, 0x1040, 0x1080, 0x2FFE, 0xE120, 0x2120, 0x2220, 0x2222, 0x2422, 0x81E}},
    {15615, 2, {0x40, 0x2020, 0x13FC, 0x1040, 0x8090, 0x4108, 0x4BFC, 0x844, 0x1040, 0x17FC, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {15616, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8000, 0x403C, 0x4BC0, 0x820, 0x1010, 0x13FC, 0xE008, 0x2010, 0x2060, 0x2180, 0x2240, 0x43E}},
    {15617, 2, {0x0, 0x23F8, 0x1040, 0x1040, 0x8FFE, 0x4110, 0x4208, 0x1516, 0x13F8, 0x2110, 0xE110, 0x27FC, 0x2110, 0x2110, 0x2210, 0x410}},
    {15618, 2, {0x20, 0x4020, 0x2020, 0x2EF8, 0x4A8, 0x84A8, 0x44F8, 0x54A8, 0x14A8, 0x24A8, 0xE5FE, 0x2688, 0x3888, 0x2088, 0x20A8, 0x90}},
    {15619, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8202, 0x43FE, 0x4220, 0x1220, 0x13B2, 0x22B4, 0xE2A8, 0x24A8, 0x2524, 0x2924, 0x32A2, 0x40}},
    {15620, 2, {0x0, 0x27FE, 0x1402, 0x1492, 0x848A, 0x450A, 0x4402, 0x1402, 0x14F2, 0x2492, 0xE492, 0x2492, 0x2492, 0x2492, 0x27FE, 0x402}},
    {15621, 2, {0x100, 0x108, 0x7D10, 0x5A0, 0x960, 0x3118, 0xC506, 0x200, 0x3FF8, 0x2448, 0x2828, 0x37D8, 0x2448, 0x27C8, 0x2008, 0x3FF8}},
    {15622, 2, {0x0, 0x23F8, 0x1008, 0x11F8, 0x8008, 0x43F8, 0x4000, 0x17FE, 0x1442, 0x2844, 0xE3F8, 0x2248, 0x2248, 0x2248, 0x2258, 0x40}},
    {15623, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xD6EA, 0x52AC, 0x52A8, 0x5AA8, 0x56EC, 0x52AA, 0x52AA, 0x94C8, 0x1088, 0x1088, 0x52A8, 0x2110}},
    {15624, 2, {0x10, 0x2038, 0x13C0, 0x1200, 0x83FC, 0x4220, 0x4220, 0x1220, 0x1FFE, 0x2000, 0xE040, 0x2444, 0x2444, 0x2444, 0x27FC, 0x4}},
    {15625, 2, {0x40, 0x2090, 0x1108, 0x13FC, 0x8108, 0x4210, 0x44A4, 0x17BE, 0x1082, 0x2000, 0xE3FC, 0x2000, 0x2000, 0x2000, 0x27FE, 0x0}},
    {15626, 2, {0x444, 0x2244, 0x1284, 0x1014, 0x87D4, 0x4114, 0x4554, 0x1554, 0x1554, 0x27D4, 0xE154, 0x2114, 0x2204, 0x2204, 0x2414, 0x808}},
    {15627, 2, {0x40, 0x2040, 0x13F8, 0x1048, 0x87FE, 0x4048, 0x4BF8, 0x840, 0x1248, 0x1248, 0xE2E8, 0x2358, 0x2248, 0x2248, 0x2448, 0x848}},
    {15628, 2, {0x210, 0x4210, 0x2210, 0x2210, 0xF7E, 0x8510, 0x4510, 0x5510, 0x157C, 0x2944, 0xE544, 0x2244, 0x2244, 0x2544, 0x257C, 0x844}},
    {15629, 2, {0x3F0, 0x4010, 0x2020, 0x2044, 0xE58, 0x83F0, 0x4250, 0x5248, 0x15E8, 0x2448, 0xE444, 0x2BF4, 0x3042, 0x2040, 0x2140, 0x80}},
    {15630, 2, {0x40, 0x2040, 0x17FC, 0x10A0, 0x8110, 0x4208, 0x45F6, 0x1000, 0x1000, 0x27FC, 0xE040, 0x2248, 0x2244, 0x2444, 0x2140, 0x80}},
    {15631, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8110, 0x4208, 0x4404, 0x13F8, 0x1040, 0x2040, 0xE3F8, 0x2040, 0x2040, 0x2040, 0x27FC, 0x0}},
    {15632, 2, {0x0, 0x40FE, 0x2F10, 0x2910, 0x950, 0x895E, 0x4F52, 0x4952, 0x1952, 0x2F7E, 0xE902, 0x2902, 0x2902, 0x2F02, 0x2914, 0x8}},
    {15633, 2, {0x40, 0x2020, 0x17FE, 0x1400, 0x8504, 0x4488, 0x47FE, 0x1420, 0x1524, 0x2524, 0xE524, 0x25FC, 0x2844, 0x2880, 0x3100, 0x200}},
    {15634, 2, {0x80, 0x2040, 0x17FE, 0x1090, 0x8298, 0x4494, 0x4114, 0x1230, 0x1400, 0x2110, 0xE110, 0x2FFE, 0x2110, 0x2110, 0x2210, 0x410}},
    {15635, 2, {0x0, 0x23FC, 0x1244, 0x1244, 0x83FC, 0x4244, 0x4A44, 0xBFC, 0x1000, 0x1040, 0xE024, 0x2522, 0x250A, 0x2908, 0x20F8, 0x0}},
    {15636, 2, {0x80, 0x2080, 0x11F8, 0x1208, 0x8410, 0x43FC, 0x4804, 0x9FC, 0x1004, 0x13FC, 0xE004, 0x2040, 0x2124, 0x252A, 0x250A, 0x8F8}},
    {15637, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8000, 0x4038, 0x4BC0, 0x840, 0x1040, 0x17FC, 0xE0E0, 0x2150, 0x2248, 0x2444, 0x2842, 0x40}},
    {15638, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x81F0, 0x4040, 0x43F8, 0x1248, 0x1248, 0x23F8, 0xE040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x40}},
    {15639, 2, {0x0, 0x23FE, 0x1050, 0x1050, 0x83FE, 0x4252, 0x4A52, 0xBFE, 0x1000, 0x1040, 0xE7FE, 0x2088, 0x2190, 0x2060, 0x20D8, 0x304}},
    {15640, 2, {0x0, 0x27FE, 0x1402, 0x15FA, 0x8402, 0x45FA, 0x4402, 0x1000, 0x13FC, 0x2204, 0xE3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x204}},
    {15641, 2, {0x8, 0x203C, 0x17C0, 0x1040, 0x8040, 0x4FFE, 0x4040, 0x1140, 0x165C, 0x2444, 0xE444, 0x275C, 0x2444, 0x2444, 0x27FC, 0x404}},
    {15642, 2, {0x208, 0x2110, 0x1000, 0x17FE, 0x8080, 0x4184, 0x4244, 0x14A8, 0x1130, 0x2270, 0xE4A8, 0x2128, 0x2224, 0x2C22, 0x20A0, 0x40}},
    {15643, 2, {0x8, 0x2008, 0x17C8, 0x1210, 0x821E, 0x44A4, 0x47D4, 0x1114, 0x1114, 0x27D4, 0xE114, 0x2108, 0x21C8, 0x2E14, 0x2414, 0x22}},
    {15644, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8404, 0x47FC, 0x4400, 0x1400, 0x17FC, 0x26A4, 0xEAA4, 0x2BFC, 0x2AA4, 0x2AA4, 0x32A4, 0x20C}},
    {15645, 2, {0x604, 0x7844, 0x844, 0xFF44, 0x1C44, 0x2A44, 0xC914, 0xA88, 0xC60, 0x3118, 0xC926, 0x540, 0x920, 0x1110, 0x2508, 0x200}},
    {15646, 2, {0x8, 0x203C, 0x13C0, 0x1044, 0x8224, 0x4128, 0x4900, 0xA7C, 0x1244, 0x1244, 0xE27C, 0x2244, 0x2244, 0x2244, 0x227C, 0x244}},
    {15647, 2, {0x0, 0x23FE, 0x1200, 0x12FC, 0x8284, 0x42FC, 0x4A84, 0xAFC, 0x1200, 0x12FC, 0xE208, 0x2210, 0x23FE, 0x2410, 0x2450, 0x820}},
    {15648, 2, {0x0, 0x8FF8, 0x4010, 0x20, 0x8040, 0x577A, 0x554A, 0x156A, 0x3552, 0x5552, 0xD76A, 0x504A, 0x5142, 0x5082, 0x5FFE, 0x2}},
    {15649, 2, {0x38, 0x23C0, 0x1040, 0x1040, 0x87FC, 0x4150, 0x5248, 0x1446, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {15650, 2, {0x90, 0x9090, 0x4890, 0x190, 0x917C, 0x4954, 0x4954, 0x354, 0x2554, 0x497C, 0xD954, 0x4910, 0x4910, 0x4910, 0x4910, 0x110}},
    {15651, 2, {0x20, 0x9020, 0x4BFE, 0x20, 0x91FC, 0x4824, 0x49FC, 0x120, 0x25FE, 0x4822, 0xD82A, 0x4854, 0x4850, 0x4888, 0x4904, 0x202}},
    {15652, 2, {0x210, 0x4208, 0x2200, 0x227E, 0xF90, 0x8210, 0x4210, 0x5710, 0x16FC, 0x2A10, 0xEA10, 0x2210, 0x2210, 0x22FE, 0x2200, 0x200}},
    {15653, 2, {0x410, 0x8410, 0x4454, 0x438, 0x9510, 0x557C, 0x5544, 0x1544, 0x357C, 0x5544, 0xD544, 0x577C, 0x5944, 0x4044, 0x4054, 0x48}},
    {15654, 2, {0x40, 0x2020, 0x17FE, 0x1402, 0x8000, 0x41F8, 0x4908, 0x908, 0x11F8, 0x1000, 0xE3FC, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {15655, 2, {0x40, 0x2080, 0x13FC, 0x1204, 0x83FC, 0x4204, 0x43FC, 0x1204, 0x13FC, 0x2000, 0xE040, 0x2024, 0x2522, 0x250A, 0x28F8, 0x0}},
    {15656, 2, {0x44, 0x2724, 0x1528, 0x1500, 0x857C, 0x4710, 0x4510, 0x1510, 0x15FE, 0x2710, 0xE510, 0x2528, 0x2528, 0x2544, 0x2B44, 0x82}},
    {15657, 2, {0x124, 0x2248, 0x1248, 0x1124, 0x83F8, 0x4208, 0x4BF8, 0x800, 0x13F8, 0x1248, 0xE248, 0x23F8, 0x2202, 0x2202, 0x21FE, 0x0}},
    {15658, 2, {0x40, 0x2080, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x4BF8, 0xA08, 0x13F8, 0x1040, 0xE7FE, 0x20E0, 0x2150, 0x2248, 0x2446, 0x40}},
    {15659, 2, {0x4000, 0x2FDE, 0x2252, 0x8252, 0x5FF4, 0x4254, 0x258, 0x2FD4, 0x2412, 0x47D2, 0xCC52, 0x4C5A, 0x5454, 0x67D0, 0x4450, 0x10}},
    {15660, 2, {0x108, 0x2208, 0x1788, 0x1488, 0x87BE, 0x4488, 0x4788, 0x14C8, 0x14A8, 0x2FA8, 0xE188, 0x2288, 0x2488, 0x2888, 0x22A8, 0x110}},
    {15661, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x80A0, 0x4110, 0x4A08, 0x800, 0x13F8, 0x1208, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x208}},
    {15662, 2, {0x0, 0x27FC, 0x1404, 0x1404, 0x87FC, 0x4420, 0x4524, 0x14A8, 0x15FC, 0x2504, 0xE5FC, 0x2504, 0x25FC, 0x2904, 0x2914, 0x1108}},
    {15663, 2, {0x20, 0x27A4, 0x14A8, 0x14B0, 0x84A2, 0x47A2, 0x449E, 0x1480, 0x14BE, 0x27A2, 0xE4A2, 0x24BE, 0x24A2, 0x28A2, 0x2ABE, 0x1122}},
    {15664, 2, {0x110, 0x2212, 0x1454, 0x17D8, 0x8010, 0x4012, 0x47D2, 0x144E, 0x1440, 0x27D2, 0xE454, 0x2458, 0x27D0, 0x2452, 0x2552, 0x48E}},
    {15665, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x80A0, 0x4110, 0x4248, 0x1040, 0x17FC, 0x2040, 0xE444, 0x22A8, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {15666, 2, {0x4, 0x203E, 0x17E0, 0x143E, 0x8420, 0x45FE, 0x4522, 0x1538, 0x15E2, 0x251E, 0xE500, 0x2978, 0x2A48, 0x324A, 0x248A, 0x906}},
    {15667, 2, {0x448, 0x4444, 0x2440, 0x24FE, 0x450, 0x9F54, 0x4458, 0x5452, 0x1454, 0x2458, 0xE452, 0x2754, 0x3C9A, 0x2092, 0x212E, 0x240}},
    {15668, 2, {0x20, 0x24A4, 0x14A8, 0x1520, 0x8450, 0x4488, 0x4504, 0x17FE, 0x1420, 0x24A4, 0xE528, 0x2450, 0x2488, 0x2504, 0x27FE, 0x0}},
    {15669, 2, {0x40, 0x2028, 0x1090, 0x12A4, 0x82C2, 0x448A, 0x4188, 0x1678, 0x1000, 0x23FC, 0xE294, 0x2294, 0x2294, 0x2294, 0x2FFE, 0x0}},
    {15670, 2, {0x80, 0x2080, 0x17FE, 0x1522, 0x8110, 0x43FC, 0x5620, 0x1A20, 0x13FC, 0x2220, 0xE220, 0x23FC, 0x2220, 0x2220, 0x23FE, 0x200}},
    {15671, 2, {0x10, 0x4F90, 0x2510, 0x2510, 0x77C, 0x8554, 0x4754, 0x5510, 0x1510, 0x25A8, 0xE728, 0x2D28, 0x214A, 0x214A, 0x2186, 0x100}},
    {15672, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x84A4, 0x47BC, 0x4404, 0x1444, 0x1444, 0x2444, 0xE4A4, 0x2494, 0x2514, 0x2404, 0x2414, 0x408}},
    {15673, 2, {0x10, 0x2092, 0x1452, 0x1254, 0x8210, 0x40FE, 0x4010, 0x1638, 0x1254, 0x2292, 0xE310, 0x2210, 0x2210, 0x2500, 0x28FE, 0x0}},
    {15674, 2, {0x20, 0x8820, 0x4420, 0x5FC, 0x9020, 0x48A8, 0x48A8, 0x4A8, 0x2574, 0x4A22, 0xD850, 0x4850, 0x4888, 0x4888, 0x4904, 0x202}},
    {15675, 2, {0x4, 0x4FC4, 0x2004, 0x2004, 0x7BE, 0x8484, 0x4484, 0x54A4, 0x1794, 0x2014, 0xE884, 0x2484, 0x2504, 0x21C4, 0x2E14, 0x8}},
    {15676, 2, {0x40, 0x2020, 0x17FE, 0x1402, 0x8100, 0x41FE, 0x4210, 0x1620, 0x1A7C, 0x2244, 0xE244, 0x227C, 0x2244, 0x2244, 0x227C, 0x244}},
    {15677, 2, {0x8, 0x403C, 0x27C0, 0x2040, 0x248, 0x8150, 0x4FFE, 0x5150, 0x1248, 0x2C46, 0xE080, 0x2044, 0x2A42, 0x2A12, 0x31F0, 0x0}},
    {15678, 2, {0x108, 0x4108, 0x2FC8, 0x2110, 0x11E, 0x8FD4, 0x4864, 0x4014, 0x1794, 0x2094, 0xE114, 0x21C8, 0x2F08, 0x2114, 0x2524, 0x242}},
    {15679, 2, {0x0, 0x407E, 0x2F90, 0x2220, 0x27C, 0x8244, 0x4244, 0x527C, 0x1244, 0x227C, 0xE244, 0x2244, 0x227C, 0x2A00, 0x2428, 0x44}},
    {15680, 2, {0x420, 0x4220, 0x223E, 0x2F40, 0x4A0, 0x8420, 0x473C, 0x5550, 0x1510, 0x2510, 0xE5FE, 0x2510, 0x2528, 0x2928, 0x2B44, 0x1082}},
    {15681, 2, {0x108, 0x2108, 0x17FE, 0x1148, 0x8020, 0x47FE, 0x4880, 0x908, 0x13FC, 0x1004, 0xE150, 0x2150, 0x2150, 0x2252, 0x2252, 0x44E}},
    {15682, 2, {0x40, 0x2020, 0x13FE, 0x1200, 0x8248, 0x4248, 0x4BFE, 0xA48, 0x1248, 0x1248, 0xE278, 0x2200, 0x2554, 0x252A, 0x2A2A, 0x1000}},
    {15683, 2, {0x410, 0x8408, 0x4408, 0x1F7E, 0x9542, 0x5542, 0x5520, 0x1524, 0x3F28, 0x4430, 0xC420, 0x4522, 0x47A2, 0x5CA2, 0x481E, 0x0}},
    {15684, 2, {0x14, 0x2012, 0x1012, 0x17FE, 0x8010, 0x43D0, 0x4A50, 0xBD4, 0x1014, 0x17D4, 0xE114, 0x27F8, 0x210A, 0x228A, 0x2456, 0x822}},
    {15685, 2, {0x40, 0x20A0, 0x1110, 0x1248, 0x8DF6, 0x4020, 0x4840, 0xBF8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {15686, 2, {0x8, 0x4028, 0x2F28, 0x293E, 0x928, 0x8948, 0x4908, 0x4F7E, 0x1900, 0x2900, 0xE93E, 0x2922, 0x2F22, 0x2922, 0x203E, 0x22}},
    {15687, 2, {0x1A4, 0x4E24, 0x2224, 0x2224, 0x27E, 0x8FA4, 0x4224, 0x5224, 0x1224, 0x2FBC, 0xE8A4, 0x28A4, 0x28A4, 0x2FA4, 0x28BC, 0x24}},
    {15688, 2, {0x88, 0x9088, 0x4888, 0x88, 0x93DE, 0x4948, 0x4948, 0x148, 0x255E, 0x4A52, 0xD952, 0x4892, 0x4892, 0x495E, 0x4952, 0x200}},
    {15689, 2, {0x88, 0x9088, 0x4888, 0x88, 0x93DE, 0x4888, 0x4888, 0x19C, 0x25DC, 0x4AAA, 0xDAAA, 0x4CC8, 0x4888, 0x4888, 0x4888, 0x88}},
    {15690, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x5000, 0x1FFE, 0x1480, 0x27BC, 0xE494, 0x2794, 0x24D4, 0x2F88, 0x2094, 0xA2}},
    {15691, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x87FC, 0x4000, 0x4BF8, 0xA08, 0x13F8, 0x1208, 0xE3F8, 0x2208, 0x23F8, 0x2110, 0x2208, 0x404}},
    {15692, 2, {0x408, 0x841C, 0x49F0, 0x1110, 0xA510, 0x45FE, 0x4910, 0x197C, 0x2944, 0x497C, 0xC944, 0x497C, 0x4944, 0x4A44, 0x4A7C, 0xC44}},
    {15693, 2, {0x0, 0x279E, 0x1492, 0x179E, 0x8492, 0x479E, 0x4442, 0x1422, 0x140A, 0x2546, 0xE542, 0x2552, 0x2632, 0x2402, 0x240A, 0x404}},
    {15694, 2, {0x79E, 0x2492, 0x179E, 0x1492, 0x879E, 0x4402, 0x44F2, 0x1492, 0x14F2, 0x2492, 0xE4F2, 0x2492, 0x2492, 0x2532, 0x2402, 0x406}},
    {15695, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8040, 0x47FC, 0x4110, 0x1110, 0x12A8, 0x2444, 0xE040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x40}},
    {15696, 2, {0x4000, 0x2FFE, 0x2800, 0x8A28, 0x4948, 0x4BEE, 0x892, 0x2884, 0x2AA0, 0x4AA8, 0xCBE8, 0x4888, 0x4894, 0x4914, 0x4924, 0x1242}},
    {15697, 2, {0x200, 0x4200, 0x27BC, 0x22A4, 0x4A4, 0x85BC, 0x4800, 0x53F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {15698, 2, {0x90, 0x27FC, 0x1094, 0x13FC, 0x8290, 0x43FE, 0x4912, 0xBFA, 0x150E, 0x11F8, 0xE108, 0x21F8, 0x2108, 0x21F8, 0x2090, 0x108}},
    {15699, 2, {0x80, 0x4040, 0x2FFE, 0x2802, 0x3F8, 0x8208, 0x43F8, 0x5208, 0x13F8, 0x2000, 0xEFFC, 0x2240, 0x227C, 0x2540, 0x24FE, 0x800}},
    {15700, 2, {0x40, 0x2040, 0x17FC, 0x1110, 0x8248, 0x4446, 0x43F8, 0x1050, 0x17FC, 0x2080, 0xE3F8, 0x2D08, 0x21F8, 0x2108, 0x21F8, 0x108}},
    {15701, 2, {0x80, 0x2110, 0x13F8, 0x1210, 0x84A4, 0x47BC, 0x4040, 0x10A0, 0x1358, 0x2C46, 0xE048, 0x2264, 0x2252, 0x2452, 0x2140, 0x80}},
    {15702, 2, {0x100, 0x2100, 0x17DE, 0x1252, 0x8692, 0x4112, 0x429E, 0x14C0, 0x1110, 0x23E0, 0xE048, 0x27FC, 0x2254, 0x2448, 0x2944, 0x80}},
    {15703, 2, {0x43FC, 0x2240, 0xBF8, 0xA40, 0x13F8, 0x2240, 0xE3FC, 0x2004, 0x2554, 0x2008, 0x108, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {15704, 2, {0x810, 0xA14, 0x6AD4, 0x2C58, 0x2A54, 0x4A94, 0xA950, 0x1020, 0x810, 0xA14, 0x6AD4, 0x2C58, 0x2A54, 0x4A94, 0xA952, 0x1020}},
    {15705, 2, {0x40, 0x27FC, 0x1040, 0x13F8, 0x8000, 0x43F8, 0x4208, 0x13F8, 0x1110, 0x2FFE, 0xE000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {15706, 2, {0x110, 0x2208, 0x1414, 0x11A0, 0x8060, 0x4198, 0x4644, 0x10A0, 0x1318, 0x2DF6, 0xE040, 0x27FC, 0x2248, 0x2150, 0x2FFE, 0x0}},
    {15707, 2, {0x110, 0x20A0, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FC, 0x1248, 0x1150, 0x2FFE, 0xE000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {15708, 2, {0x10, 0x2410, 0x12FE, 0x1210, 0x8010, 0x48FE, 0x4482, 0x14AA, 0x1192, 0x22BA, 0xEE92, 0x22FE, 0x2292, 0x2292, 0x228A, 0x84}},
    {15709, 2, {0x80, 0x8040, 0x5FFE, 0x148, 0x8230, 0x4690, 0x5B08, 0x206, 0x27FC, 0x4524, 0xC614, 0x45EC, 0x4524, 0x45E4, 0x4414, 0x408}},
    {15710, 2, {0xC8, 0x4708, 0x212A, 0x212C, 0xFC8, 0x8108, 0x4388, 0x5554, 0x1914, 0x2122, 0xE040, 0x2024, 0x2522, 0x250A, 0x2908, 0xF8}},
    {15711, 2, {0x288, 0x42A8, 0x26A8, 0x2BF0, 0x1E, 0x8224, 0x45D4, 0x5C14, 0x1414, 0x25D4, 0xE554, 0x2548, 0x2568, 0x2554, 0x2614, 0x422}},
    {15712, 2, {0x40, 0x47FE, 0x2402, 0x2914, 0x208, 0x87FC, 0x4404, 0x57FC, 0x1420, 0x2524, 0xE524, 0x25FC, 0x2420, 0x2924, 0x2924, 0x11FC}},
    {15713, 2, {0x3F8, 0x2040, 0x17FC, 0x1444, 0x8358, 0x4040, 0x4358, 0x1200, 0x13F8, 0x2408, 0xEBC8, 0x2248, 0x23E8, 0x2212, 0x2202, 0x1FE}},
    {15714, 2, {0x0, 0x47FC, 0x2040, 0x2FFE, 0x842, 0x8358, 0x4040, 0x5358, 0x1000, 0x27FC, 0xE444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x7FC}},
    {15715, 2, {0x10, 0x9F10, 0x5110, 0x1F3E, 0x9122, 0x5F44, 0x4810, 0x1F90, 0x2490, 0x5490, 0xDAA8, 0x50A8, 0x5EA8, 0x40C4, 0x4544, 0x282}},
    {15716, 2, {0x0, 0x27BC, 0x1084, 0x14A4, 0x8294, 0x44A4, 0x4040, 0x17FC, 0x1040, 0x2248, 0xE248, 0x2554, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {15717, 2, {0x80, 0x2040, 0x17FC, 0x1484, 0x8050, 0x4528, 0x5554, 0x1994, 0x16F0, 0x2000, 0xE3F8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x0}},
    {15718, 2, {0x20, 0x21FE, 0x1440, 0x12FC, 0x8110, 0x42FE, 0x4000, 0x16FC, 0x1284, 0x22FC, 0xE284, 0x22FC, 0x2284, 0x228C, 0x2500, 0x8FE}},
    {15719, 2, {0x110, 0x27FE, 0x1110, 0x13F8, 0x8208, 0x43F8, 0x4A08, 0xBF8, 0x1100, 0x13FC, 0xE444, 0x2AA4, 0x2204, 0x23F4, 0x2014, 0x8}},
    {15720, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x108, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {15721, 2, {0x108, 0x27FE, 0x1108, 0x1000, 0x8488, 0x4FFE, 0x4488, 0x14F8, 0x1400, 0x27FC, 0xE040, 0x2FFE, 0x2150, 0x2248, 0x2C46, 0x40}},
    {15722, 2, {0x208, 0x2110, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FC, 0x1000, 0x1440, 0x227C, 0xE084, 0x2228, 0x2C20, 0x2450, 0x2488, 0x106}},
    {15723, 2, {0x1F8, 0x2010, 0x17FE, 0x10A0, 0x8040, 0x47BC, 0x4084, 0x11CE, 0x1738, 0x2108, 0xE318, 0x27FC, 0x2404, 0x27FC, 0x2404, 0x7FC}},
    {15724, 2, {0x0, 0x9DF8, 0x5488, 0x1490, 0x14FC, 0x1D54, 0x9524, 0x5554, 0x1688, 0x1C00, 0x15FC, 0x5554, 0x9554, 0x2554, 0x27FE, 0x4C00}},
    {15725, 2, {0x40, 0x2044, 0x13F8, 0x1050, 0x87FE, 0x4040, 0x51FC, 0x1704, 0x11FC, 0x2104, 0xE1FC, 0x2000, 0x2524, 0x2492, 0x2892, 0x0}},
    {15726, 2, {0x40, 0x27FC, 0x10A0, 0x1318, 0x8D16, 0x43F8, 0x4110, 0x11F0, 0x1040, 0x23F8, 0xE248, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x40}},
    {15727, 2, {0x4000, 0x2FF8, 0x2808, 0x8BE8, 0x4908, 0x4BE8, 0xA28, 0x2BE8, 0x2A28, 0x4BE8, 0xCA08, 0x4BFA, 0x4A0A, 0x53FA, 0x5556, 0x2032}},
    {15728, 2, {0x140, 0x224C, 0x1264, 0x1244, 0x834C, 0x4264, 0x4AA4, 0xBAC, 0x12A4, 0x12A4, 0xE7FE, 0x2000, 0x2110, 0x2108, 0x2204, 0x404}},
    {15729, 2, {0x4, 0x4FE4, 0x2284, 0x2FE4, 0xABE, 0x8AA4, 0x4FE4, 0x5004, 0x17D4, 0x200C, 0xEFEC, 0x2104, 0x2544, 0x2524, 0x2924, 0x30C}},
    {15730, 2, {0x100, 0x4200, 0x27FC, 0x2AA8, 0x2A8, 0x8FFE, 0x42A8, 0x52A8, 0x1FFE, 0x2208, 0xE3BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0xC08}},
    {15731, 2, {0x40, 0x4FFE, 0x2802, 0x27BC, 0x4A4, 0x8294, 0x44E4, 0x50A0, 0x1318, 0x2C46, 0xE090, 0x2320, 0x20C8, 0x2310, 0x2060, 0x380}},
    {15732, 2, {0xC, 0x800A, 0x57FE, 0x1408, 0x95E8, 0x5D48, 0x454A, 0x5EA, 0x3D2A, 0x552C, 0xD5EC, 0x554C, 0x654A, 0x49EA, 0x4816, 0x1022}},
    {15733, 2, {0x7E0, 0x225C, 0x13C4, 0x1254, 0x83C8, 0x4268, 0x47D4, 0x1064, 0x103C, 0x27C0, 0xE144, 0x2268, 0x20D0, 0x2148, 0x2646, 0x40}},
    {15734, 2, {0x110, 0x4114, 0x27D2, 0x2110, 0xFFE, 0x8290, 0x4450, 0x5FF4, 0x1494, 0x27F4, 0xE494, 0x27E8, 0x248A, 0x27FA, 0x2426, 0x42}},
    {15735, 2, {0x2010, 0x1108, 0x1088, 0x7CBE, 0x200, 0x2914, 0x1108, 0xFC3E, 0x1088, 0x1088, 0x7D1C, 0x1308, 0x1108, 0x2108, 0x2108, 0x4008}},
    {15736, 2, {0x40, 0x20A0, 0x1110, 0x1208, 0x85F4, 0x4802, 0x43B8, 0x12A8, 0x13B8, 0x2000, 0xE7FC, 0x24A4, 0x27FC, 0x24A4, 0x24A4, 0x40C}},
    {15737, 2, {0x428, 0x8428, 0x44FE, 0x428, 0x9500, 0x557C, 0x5544, 0x157C, 0x3544, 0x557C, 0xD510, 0x577E, 0x5910, 0x4028, 0x4044, 0x82}},
    {15738, 2, {0x210, 0x2210, 0x17DE, 0x1528, 0x88C4, 0x4080, 0x4BFC, 0xA24, 0x13FC, 0x1244, 0xE3FC, 0x2120, 0x2220, 0x27FE, 0x2020, 0x20}},
    {15739, 2, {0x4050, 0x2E96, 0x2A92, 0x8A92, 0x4AD6, 0x4E92, 0xA92, 0x2AFE, 0x2A10, 0x4EFC, 0xCA44, 0x4A44, 0x4A28, 0x5210, 0x5228, 0x26C6}},
    {15740, 2, {0x0, 0x27FC, 0x1404, 0x1080, 0x8338, 0x4208, 0x43B8, 0x1208, 0x13F8, 0x2200, 0xE7FC, 0x2804, 0x2554, 0x2554, 0x2828, 0x10}},
    {15741, 2, {0x3F8, 0x2248, 0x1248, 0x13F8, 0x8248, 0x4248, 0x43F8, 0x1000, 0x1FBE, 0x2AAA, 0xEAAA, 0x2FBE, 0x2AAA, 0x2AAA, 0x2FBE, 0x8A2}},
    {15742, 2, {0x200, 0x823C, 0x5FA4, 0x224, 0x8FBC, 0x4224, 0x5FE4, 0x3C, 0x2FA4, 0x48A4, 0xCFBC, 0x4890, 0x4F98, 0x48A8, 0x48AA, 0x9C6}},
    {15743, 2, {0x80, 0x2100, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x43F8, 0x1208, 0x1410, 0x2FBE, 0xE8A2, 0x28A2, 0x2FBE, 0x28A2, 0x28A2, 0xFBE}},
    {15744, 2, {0x0, 0x4EEE, 0x2A22, 0x2AAA, 0xA66, 0x8EAA, 0x4A10, 0x4A28, 0x1A44, 0x2F92, 0xEA20, 0x2AC8, 0x2A32, 0x2AC4, 0x2A18, 0x16E0}},
    {15745, 2, {0x390, 0x9C10, 0x411E, 0x1510, 0x8A7E, 0x4852, 0x425C, 0x270, 0x3F52, 0x424E, 0xD240, 0x4A5C, 0x4254, 0x4A54, 0x4496, 0x122}},
    {15746, 2, {0x88, 0x9088, 0x49FC, 0x88, 0x9088, 0x4BFE, 0x4820, 0x1FC, 0x2524, 0x49FC, 0xD924, 0x49FC, 0x4800, 0x4888, 0x4904, 0x202}},
    {15747, 2, {0x80, 0x9080, 0x49FE, 0x354, 0x9154, 0x4954, 0x4BFE, 0x154, 0x2554, 0x4954, 0xDBFE, 0x4800, 0x4954, 0x492A, 0x4A2A, 0x0}},
    {15748, 2, {0x7BC, 0x24A4, 0x17BC, 0x14A4, 0x87BC, 0x4484, 0x44F4, 0x1514, 0x1624, 0x25B4, 0xE514, 0x25B4, 0x2514, 0x25F4, 0x2414, 0x408}},
    {15749, 2, {0x42A0, 0x244C, 0x24A4, 0x860C, 0x44A4, 0x464C, 0x4A4, 0x2FFE, 0x2802, 0x43F0, 0xC020, 0x4040, 0x47FC, 0x4040, 0x4140, 0x80}},
    {15750, 2, {0x410, 0x4210, 0x227A, 0x2F12, 0x14, 0x8F7E, 0x4008, 0x4F10, 0x103C, 0x2F64, 0xE9A4, 0x293C, 0x2924, 0x2F24, 0x293C, 0x24}},
    {15751, 2, {0x4228, 0x2228, 0x227C, 0x8228, 0x4F28, 0x42FE, 0x610, 0x277C, 0x2AD4, 0x4A7C, 0xD254, 0x427C, 0x4200, 0x4228, 0x4244, 0x282}},
    {15752, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4FFE, 0x4842, 0x1040, 0x23F8, 0xE248, 0x2248, 0x23F8, 0x2044, 0x27FC, 0x4}},
    {15753, 2, {0x308, 0xBC10, 0x44BE, 0x1122, 0x8432, 0x492A, 0x5E22, 0x426, 0x2920, 0x5F3E, 0xC402, 0x5F02, 0x447A, 0x4A02, 0x510A, 0x2084}},
    {15754, 2, {0x0, 0x4FBE, 0x2A08, 0x2A10, 0xA3E, 0x8FA2, 0x48A2, 0x48BE, 0x18A2, 0x2FBE, 0xEA22, 0x2A22, 0x2A3E, 0x2A00, 0x2F94, 0x22}},
    {15755, 2, {0x508, 0x5548, 0x2DA8, 0x253E, 0xFA8, 0x8248, 0x4448, 0x4ABE, 0x1308, 0x2588, 0xE988, 0x237E, 0x2540, 0x2920, 0x251E, 0x200}},
    {15756, 2, {0x44, 0x9F44, 0x4AAA, 0xAEE, 0x8E44, 0x4AAA, 0x4AEE, 0xE00, 0x2A28, 0x4AAA, 0xCBAA, 0x4EEE, 0x5A28, 0x4228, 0x4248, 0x288}},
    {15757, 2, {0x200, 0x4222, 0x2792, 0x2494, 0x93E, 0x87C8, 0x4548, 0x555E, 0x17C8, 0x2548, 0xE57E, 0x27C8, 0x2008, 0x2AA8, 0x2AA8, 0x8}},
    {15758, 2, {0x1FC, 0x4954, 0x25FC, 0x2400, 0x1FE, 0x8000, 0x5CFC, 0x4484, 0x14FC, 0x2452, 0xE4CC, 0x2548, 0x2464, 0x2B00, 0x30FE, 0x0}},
    {15759, 2, {0x1C, 0x4EE0, 0x2A54, 0x2A28, 0xA7C, 0x8C10, 0x4AFE, 0x4A00, 0x1AFC, 0x2A04, 0xEA7C, 0x2C04, 0x28FC, 0x2814, 0x294A, 0xA38}},
    {15760, 2, {0x40, 0x2020, 0x17FE, 0x14A8, 0x84FE, 0x4550, 0x477C, 0x1550, 0x157C, 0x2550, 0xE57E, 0x2540, 0x28A4, 0x2A8A, 0x328A, 0x478}},
    {15761, 2, {0x210, 0x8210, 0x4510, 0x89E, 0x9052, 0x6FA2, 0x4008, 0x8, 0x3DC8, 0x5548, 0xD548, 0x5DD4, 0x4894, 0x4894, 0x5554, 0x2222}},
    {15762, 2, {0x1F3E, 0x9122, 0x5F3E, 0x1122, 0x9F3E, 0x5002, 0x57FA, 0x1482, 0x37F2, 0x5482, 0xD7F2, 0x5482, 0x57FA, 0x500A, 0x5AAA, 0x1014}},
    {15763, 2, {0x210, 0x9FFE, 0x4214, 0xA, 0x7FE, 0x1408, 0x95E8, 0x5D48, 0x5EA, 0x3D2A, 0x15EC, 0x554C, 0xA54A, 0x9EA, 0x816, 0x1022}},
    {15764, 2, {0x0, 0x43F8, 0x22A8, 0x22A8, 0x3F8, 0x8000, 0x4FBE, 0x4AAA, 0x1AAA, 0x2FBE, 0xE040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {15765, 2, {0x210, 0x43DE, 0x2528, 0x2884, 0x7BC, 0x84A4, 0x47BC, 0x54A4, 0x17BC, 0x2404, 0xE5F4, 0x2514, 0x25F4, 0x2514, 0x25F4, 0x40C}},
    {15766, 2, {0x880, 0x89F8, 0x4A88, 0x1070, 0x138E, 0x3400, 0xB5FC, 0x55AC, 0x1574, 0x15FC, 0x1420, 0x55FC, 0x9420, 0x13FE, 0x1154, 0x122A}},
    {15767, 2, {0x2028, 0x7F78, 0xAA14, 0x7F7C, 0x2A54, 0x3FB2, 0x600, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x184, 0x3D48, 0x930, 0x730E}},
    {15768, 2, {0x210, 0x43DE, 0x2528, 0x2884, 0x7FC, 0x80A0, 0x47FC, 0x54A4, 0x17FC, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x2040, 0x2FFE, 0x40}},
    {15769, 2, {0x284, 0xAAA4, 0x56D4, 0x29E, 0xA7D4, 0x5124, 0x5224, 0x55E, 0x2984, 0x52C4, 0xF4C4, 0x51BE, 0x52A0, 0x5490, 0x528E, 0x100}},
    {15770, 2, {0x7FE, 0x2090, 0x17FE, 0x1492, 0x87FE, 0x4280, 0x47DE, 0x1292, 0x1392, 0x211E, 0xE7D2, 0x2552, 0x27DE, 0x2112, 0x2FD2, 0x126}},
    {15771, 2, {0x1080, 0x8880, 0x49FE, 0x3D00, 0x12A0, 0x112C, 0x9D34, 0x55AC, 0x1534, 0x1554, 0x15DC, 0x5554, 0xA7FE, 0x2488, 0x4D04, 0x8202}},
    {15772, 2, {0x2, 0xBEFA, 0x628A, 0x3EFA, 0x2082, 0x31C6, 0xAAAA, 0x6082, 0x3FFE, 0x2AAA, 0x2AAA, 0x7FFE, 0xAAAA, 0x2AAA, 0x32CA, 0x4104}},
    {15773, 2, {0x40, 0x9020, 0x4BFE, 0x2A4, 0x92BE, 0x4BE8, 0x4ABE, 0x2A8, 0x26BE, 0x4AA8, 0xDABE, 0x4AA0, 0x4A14, 0x4B52, 0x4D4A, 0xA38}},
    {15774, 2, {0x3F8, 0x2208, 0x13F8, 0x1208, 0x83F8, 0x4040, 0x4FFE, 0x14A4, 0x17BC, 0x2110, 0xE7FC, 0x2110, 0x2FFE, 0x2124, 0x2318, 0xD86}},
    {15775, 2, {0x47FC, 0x2444, 0x27FC, 0x8444, 0x4FFE, 0x4AAA, 0xFBE, 0x2AAA, 0x2FBE, 0x4220, 0xC7C0, 0x4084, 0x4FFE, 0x4042, 0x4548, 0x884}},
    {15776, 2, {0x0, 0x9EFE, 0x54AA, 0x1EAA, 0x94FE, 0x5E10, 0x547C, 0x1410, 0x3FFE, 0x4128, 0xD57C, 0x5510, 0x55FE, 0x5110, 0x4110, 0x610}},
    {15777, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2928, 0x2928, 0x2948, 0x3108, 0x2288, 0x2248, 0x244A, 0x242A, 0x482A, 0x5006, 0x8002}},
    {15778, 2, {0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC206, 0x200, 0x3FE0, 0x220, 0x420, 0x420, 0x822, 0x3022, 0xC01E}},
    {15779, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x0, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15780, 2, {0x1080, 0x1080, 0x1080, 0x10FC, 0x5504, 0x5904, 0x5284, 0x9448, 0x1028, 0x1010, 0x1010, 0x2820, 0x2440, 0x4480, 0x4100, 0x8600}},
    {15781, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x5420, 0x5820, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x2820, 0x2420, 0x4420, 0x4020, 0x8020}},
    {15782, 2, {0x1000, 0x11F8, 0x1008, 0x1050, 0x5420, 0x5810, 0x53FE, 0x9022, 0x1024, 0x1020, 0x1020, 0x2820, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {15783, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0x5522, 0x5922, 0x5122, 0x9122, 0x1152, 0x114A, 0x118A, 0x2902, 0x2502, 0x4502, 0x410A, 0x8104}},
    {15784, 2, {0x0, 0x7FFC, 0x80, 0x100, 0x300, 0x560, 0x918, 0x3104, 0xC102, 0x100, 0x100, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {15785, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0x5522, 0x5A24, 0x5020, 0x9020, 0x1050, 0x1050, 0x1050, 0x2850, 0x2490, 0x4492, 0x4112, 0x820E}},
    {15786, 2, {0x0, 0x7FFC, 0x100, 0x300, 0x560, 0x1918, 0x6104, 0x0, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15787, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x2104, 0x12C8, 0xC30, 0xF00C, 0x102, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15788, 2, {0x1020, 0x1020, 0x1020, 0x15FE, 0x5820, 0x5020, 0x5020, 0x91FC, 0x1070, 0x10A8, 0x10A8, 0x2924, 0x2524, 0x4222, 0x4020, 0x8020}},
    {15789, 2, {0x1010, 0x1090, 0x1090, 0x1110, 0x557E, 0x5A52, 0x5392, 0x9092, 0x1112, 0x1112, 0x1252, 0x2BD2, 0x2462, 0x4422, 0x404A, 0x8084}},
    {15790, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2008, 0x2018, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {15791, 2, {0x1000, 0x1080, 0x131E, 0x1252, 0x5652, 0x5A52, 0x5252, 0x9252, 0x1252, 0x12D2, 0x135A, 0x2A54, 0x2490, 0x4490, 0x4110, 0x8210}},
    {15792, 2, {0x1088, 0x1088, 0x1088, 0x1488, 0x5BFE, 0x5088, 0x5088, 0x9088, 0x1088, 0x10F8, 0x1088, 0x2888, 0x2488, 0x4488, 0x40F8, 0x8088}},
    {15793, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x55FE, 0x5820, 0x5070, 0x9070, 0x10A8, 0x10A8, 0x1124, 0x2AFA, 0x2420, 0x4420, 0x4020, 0x8020}},
    {15794, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x5524, 0x59FC, 0x5124, 0x9124, 0x1124, 0x11FC, 0x1124, 0x2924, 0x2524, 0x4524, 0x4104, 0x820C}},
    {15795, 2, {0x100, 0x3FF0, 0x110, 0xFFFE, 0x110, 0x3FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x100, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {15796, 2, {0x1020, 0x1020, 0x107C, 0x1084, 0x5548, 0x5830, 0x5020, 0x9048, 0x1190, 0x103E, 0x1042, 0x29A4, 0x2418, 0x4410, 0x4060, 0x8180}},
    {15797, 2, {0x810, 0x878, 0x13C0, 0x3040, 0x5040, 0x9FFE, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {15798, 2, {0x1020, 0x1020, 0x1020, 0x17FE, 0x5A42, 0x5444, 0x50A0, 0x90A2, 0x11A4, 0x1298, 0x1490, 0x2888, 0x2484, 0x44A2, 0x40C0, 0x8080}},
    {15799, 2, {0x1080, 0x1080, 0x11FC, 0x1504, 0x5A04, 0x51E4, 0x5124, 0x9124, 0x11E4, 0x1124, 0x1124, 0x29E4, 0x2504, 0x4404, 0x4028, 0x8010}},
    {15800, 2, {0x1040, 0x1040, 0x11FC, 0x1444, 0x5884, 0x5084, 0x5128, 0x9210, 0x1108, 0x13DE, 0x114A, 0x294A, 0x254A, 0x4252, 0x435A, 0x84A4}},
    {15801, 2, {0x1000, 0x11FC, 0x1020, 0x1420, 0x5BFE, 0x5000, 0x5020, 0x9022, 0x13B2, 0x10B4, 0x10A8, 0x2928, 0x2524, 0x4224, 0x44A2, 0x8040}},
    {15802, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x5420, 0x5BFE, 0x5050, 0x9088, 0x1104, 0x12FA, 0x1488, 0x2888, 0x2488, 0x4488, 0x40F8, 0x8088}},
    {15803, 2, {0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE10E, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100, 0x4888, 0x4444, 0x8444}},
    {15804, 2, {0x1020, 0x1040, 0x1088, 0x1504, 0x5BFE, 0x5082, 0x5080, 0x91FC, 0x1220, 0x1020, 0x13FE, 0x2820, 0x2450, 0x4488, 0x4104, 0x8602}},
    {15805, 2, {0x1000, 0x11FC, 0x1104, 0x1504, 0x5904, 0x51FC, 0x5020, 0x9020, 0x13FE, 0x1222, 0x1252, 0x2A8A, 0x270A, 0x4202, 0x420A, 0x8204}},
    {15806, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {15807, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x5420, 0x5BFE, 0x5202, 0x9404, 0x11F8, 0x1010, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {15808, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x4104, 0x7FF0, 0x410, 0x410, 0x410, 0xFFD0, 0x412, 0x40A, 0x406, 0x402}},
    {15809, 2, {0x1000, 0x11F8, 0x1008, 0x10D0, 0x5420, 0x59FC, 0x5124, 0x9124, 0x11FC, 0x1124, 0x1124, 0x29FC, 0x2524, 0x4524, 0x4124, 0x810C}},
    {15810, 2, {0x0, 0x7E7C, 0x4804, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC4, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15811, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xEFEE, 0x800, 0xFE0, 0x800, 0xFE0, 0x800, 0xFFFE, 0x1110, 0x14E0, 0x181C}},
    {15812, 2, {0x1000, 0x11FC, 0x1124, 0x1524, 0x59FC, 0x5124, 0x5124, 0x91FC, 0x1020, 0x13FE, 0x1070, 0x28A8, 0x2524, 0x4222, 0x4020, 0x8020}},
    {15813, 2, {0x7F00, 0x227C, 0x3E04, 0x2228, 0x3E10, 0x23A8, 0xFE44, 0x200, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15814, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x55FC, 0x5900, 0x51FC, 0x9104, 0x11FC, 0x1020, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x4020, 0x8020}},
    {15815, 2, {0x2040, 0x2040, 0x3C40, 0x45FE, 0xA840, 0x1090, 0x2090, 0xC892, 0x1152, 0x3D54, 0x4690, 0x8428, 0x2828, 0x1044, 0x2084, 0xC102}},
    {15816, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0x5488, 0x5924, 0x5222, 0x91FC, 0x1124, 0x11FC, 0x1124, 0x29FC, 0x2422, 0x4422, 0x401E, 0x8000}},
    {15817, 2, {0x4, 0x3F84, 0x2084, 0x2084, 0x3FBE, 0x2004, 0x2F84, 0x2024, 0x3FD4, 0x2214, 0x2A44, 0x2A84, 0x5204, 0x4504, 0x8894, 0x1048}},
    {15818, 2, {0x820, 0x820, 0x820, 0x7EFC, 0x820, 0x1C30, 0x1A70, 0x2A68, 0x48A4, 0x8922, 0x820, 0x820, 0x0, 0x4888, 0x4444, 0x8444}},
    {15819, 2, {0x1020, 0x10A2, 0x10A2, 0x1124, 0x5450, 0x5888, 0x5304, 0x9022, 0x1020, 0x10A4, 0x10A4, 0x2928, 0x2450, 0x4488, 0x4104, 0x8202}},
    {15820, 2, {0x1000, 0x11FC, 0x1004, 0x14FC, 0x5804, 0x51FC, 0x5000, 0x93FE, 0x1222, 0x1020, 0x11FC, 0x2924, 0x2524, 0x4534, 0x4128, 0x8020}},
    {15821, 2, {0x1020, 0x1020, 0x1050, 0x1488, 0x5904, 0x52FA, 0x5000, 0x91FC, 0x1154, 0x1154, 0x11FC, 0x2954, 0x2554, 0x4554, 0x4104, 0x810C}},
    {15822, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x5504, 0x59FC, 0x5080, 0x91FE, 0x1222, 0x1122, 0x1152, 0x2902, 0x25FA, 0x4402, 0x4014, 0x8008}},
    {15823, 2, {0x640, 0x387C, 0x890, 0x7E10, 0x8FE, 0x1C10, 0x2A10, 0x4810, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15824, 2, {0x1040, 0x119C, 0x1104, 0x1104, 0x55DC, 0x5904, 0x5104, 0x91FC, 0x1020, 0x1020, 0x11FC, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {15825, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xE00E, 0x3FF8, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {15826, 2, {0x11FC, 0x1124, 0x1124, 0x15FC, 0x5924, 0x5124, 0x51FC, 0x9020, 0x1020, 0x13FE, 0x1222, 0x2A2A, 0x26FA, 0x420A, 0x4202, 0x8206}},
    {15827, 2, {0x1080, 0x1080, 0x10FC, 0x1554, 0x5AD4, 0x50B4, 0x512C, 0x9244, 0x1094, 0x1108, 0x1040, 0x28A4, 0x26AA, 0x428A, 0x4478, 0x8000}},
    {15828, 2, {0x1008, 0x103C, 0x13C0, 0x1444, 0x5A24, 0x5128, 0x5100, 0x927C, 0x1244, 0x1244, 0x127C, 0x2A44, 0x2644, 0x4244, 0x427C, 0x8244}},
    {15829, 2, {0x2210, 0x1210, 0x1410, 0xFF92, 0x1452, 0x1454, 0x7F58, 0x5590, 0x5510, 0x6328, 0x4128, 0x7F28, 0x4148, 0x4144, 0x7F84, 0x4102}},
    {15830, 2, {0x0, 0x7DFC, 0x4444, 0x4444, 0x7C44, 0x4444, 0x4494, 0x7C88, 0x1100, 0x14FC, 0x5884, 0x5084, 0x9084, 0x2884, 0x44FC, 0x8084}},
    {15831, 2, {0x0, 0x2FBC, 0x28A4, 0x2FA4, 0x283C, 0x2FA2, 0x28A2, 0x4F9E, 0x8100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15832, 2, {0x1040, 0x1088, 0x1104, 0x13FE, 0x5402, 0x59FC, 0x5104, 0x9104, 0x11FC, 0x1020, 0x13FE, 0x2870, 0x24A8, 0x4524, 0x4222, 0x8020}},
    {15833, 2, {0x1108, 0x1088, 0x1090, 0x17FE, 0x5800, 0x53C4, 0x5254, 0x9254, 0x13D4, 0x1254, 0x1254, 0x2BD4, 0x2654, 0x4244, 0x4254, 0x82C8}},
    {15834, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x55FC, 0x5904, 0x51FC, 0x9000, 0x13FE, 0x1202, 0x11FC, 0x2820, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {15835, 2, {0x800, 0xBF8, 0x1208, 0x3208, 0x53F8, 0x9040, 0x17FE, 0x10E0, 0x1150, 0x1248, 0x1C46, 0x1040, 0x0, 0x4888, 0x4444, 0x8444}},
    {15836, 2, {0x2080, 0x2040, 0x27FE, 0x2090, 0xAA98, 0xB494, 0xA114, 0xA230, 0x2440, 0x2040, 0x27FE, 0x2040, 0x50A0, 0x4910, 0x4208, 0x8C06}},
    {15837, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x5488, 0x5904, 0x5222, 0x9028, 0x1024, 0x13FE, 0x1020, 0x2850, 0x2450, 0x4488, 0x4104, 0x8202}},
    {15838, 2, {0x2000, 0x23FE, 0x2200, 0x22FC, 0xAA84, 0xB2FC, 0xA284, 0xA2FC, 0x2200, 0x22FC, 0x2208, 0x2210, 0x53FE, 0x4C10, 0x4450, 0x8820}},
    {15839, 2, {0x1008, 0x1210, 0x117C, 0x1544, 0x5844, 0x507C, 0x5740, 0x9140, 0x117E, 0x1142, 0x1142, 0x297E, 0x2542, 0x4280, 0x447E, 0x8000}},
    {15840, 2, {0x110, 0x920, 0x12E0, 0xC18, 0x7104, 0x910, 0x12A0, 0xC60, 0x701C, 0x1010, 0x3EFC, 0x4210, 0x1490, 0x8FE, 0x3010, 0xC010}},
    {15841, 2, {0x0, 0x7EFE, 0x4280, 0x7EFC, 0x4284, 0x7EFC, 0x4280, 0x7EFE, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {15842, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x800, 0x37F0, 0xC080, 0x1FFC, 0x100, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {15843, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0x55FC, 0x5820, 0x53FE, 0x9000, 0x10A0, 0x10FC, 0x1120, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {15844, 2, {0x1000, 0x1078, 0xFE48, 0x1048, 0x7C48, 0x86, 0xFE00, 0x82FC, 0x7C44, 0x1044, 0x1228, 0x5428, 0xA810, 0x2428, 0x4444, 0x8182}},
    {15845, 2, {0x1000, 0x15F8, 0x5410, 0x5820, 0x9024, 0x29A8, 0x44B0, 0x90A8, 0x1528, 0x5924, 0x5222, 0x90A0, 0x2840, 0x4400, 0x81FC, 0x0}},
    {15846, 2, {0x20, 0xFF20, 0x24A4, 0x3CA4, 0x24A8, 0x3D20, 0x2420, 0x2450, 0x2F50, 0xF488, 0x504, 0x602, 0x400, 0x4888, 0x4444, 0x8444}},
    {15847, 2, {0x1000, 0x11FE, 0x1110, 0x1520, 0x597C, 0x5144, 0x517C, 0x9144, 0x117C, 0x1110, 0x1110, 0x2954, 0x2552, 0x4292, 0x4250, 0x8420}},
    {15848, 2, {0x1020, 0x1222, 0x13FE, 0x1480, 0x59F8, 0x5288, 0x5070, 0x938E, 0x1020, 0x11FC, 0x1020, 0x29FC, 0x2420, 0x43FE, 0x4020, 0x8020}},
    {15849, 2, {0x0, 0x27DC, 0x2504, 0x2784, 0x249C, 0x2790, 0x2512, 0x47CE, 0x0, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {15850, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0x55FC, 0x5920, 0x51FC, 0x9120, 0x11FE, 0x1100, 0x13FC, 0x2888, 0x249E, 0x4502, 0x420A, 0x8404}},
    {15851, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x8084, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {15852, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x5400, 0x59FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1020, 0x2BFE, 0x2450, 0x4488, 0x4104, 0x8202}},
    {15853, 2, {0x2100, 0x13F0, 0x1510, 0xE0, 0x775C, 0x11F0, 0x1040, 0x13FC, 0x2840, 0x47FE, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {15854, 2, {0x1090, 0x1090, 0x17FE, 0x1090, 0x57FC, 0x5A94, 0x53FC, 0x9294, 0x13FC, 0x1000, 0x11F8, 0x2908, 0x25F8, 0x4508, 0x41F8, 0x8108}},
    {15855, 2, {0x1020, 0x1020, 0x55FC, 0x5840, 0x90F8, 0x1040, 0x2BFE, 0x4488, 0x9124, 0x1222, 0x7CF8, 0x1020, 0x11FC, 0x1E20, 0xF020, 0x4020}},
    {15856, 2, {0x1040, 0x1020, 0x13FE, 0x1400, 0x5954, 0x5124, 0x5154, 0x91FC, 0x1020, 0x13FE, 0x1242, 0x2A92, 0x26FA, 0x420A, 0x4202, 0x8206}},
    {15857, 2, {0x1010, 0x1010, 0x2210, 0x7F7C, 0x110, 0x3E10, 0x2210, 0x3EFE, 0x2210, 0x3E20, 0x2244, 0x2AFE, 0x2442, 0x0, 0x2488, 0x4244}},
    {15858, 2, {0x21C8, 0x2708, 0x2108, 0x2FE8, 0xA91E, 0xB7CA, 0xA54A, 0xA7CA, 0x254A, 0x27CA, 0x210A, 0x27CA, 0x5112, 0x49D2, 0x4E2A, 0x8444}},
    {15859, 2, {0x3FFE, 0x2248, 0x24FE, 0x2D90, 0x36FC, 0x2490, 0x24FC, 0x2490, 0x24FE, 0x2480, 0x2040, 0x2248, 0x4450, 0x40A0, 0x8318, 0xC06}},
    {15860, 2, {0x3FFE, 0x2104, 0x2E38, 0x2210, 0x2FBC, 0x2718, 0x2AB4, 0x3252, 0x2000, 0x2080, 0x2488, 0x2890, 0x2140, 0x4220, 0x4C18, 0xB006}},
    {15861, 2, {0x1000, 0x11FC, 0x1124, 0x15AC, 0x5974, 0x5124, 0x51FC, 0x9020, 0x11FC, 0x1020, 0x13FE, 0x2800, 0x26A4, 0x4252, 0x4452, 0x8000}},
    {15862, 2, {0x1110, 0x17FC, 0x2110, 0x2FFE, 0x6100, 0xA3F8, 0x2648, 0x2BF8, 0x2248, 0x23F8, 0x2248, 0x2258, 0x0, 0x4888, 0x4444, 0x8444}},
    {15863, 2, {0x1000, 0x13DE, 0x1252, 0x17DE, 0x5A10, 0x5252, 0x51CE, 0x9000, 0x1088, 0x13FE, 0x1088, 0x2888, 0x27FE, 0x4488, 0x4104, 0x8202}},
    {15864, 2, {0x1000, 0x11FC, 0x1104, 0x15FC, 0x5904, 0x51FC, 0x5000, 0x91FC, 0x10A8, 0x13FE, 0x10A8, 0x29FC, 0x2420, 0x43FE, 0x4020, 0x8020}},
    {15865, 2, {0x1088, 0x13FE, 0x1088, 0x1020, 0x543E, 0x5820, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {15866, 2, {0x2020, 0x3E20, 0x443E, 0xA844, 0x4684, 0x4228, 0x7610, 0x4228, 0x7E44, 0x102, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {15867, 2, {0x1080, 0x1080, 0x11FE, 0x1354, 0x5554, 0x5954, 0x53FE, 0x9154, 0x1154, 0x1154, 0x17FE, 0x2800, 0x2554, 0x452A, 0x422A, 0x8000}},
    {15868, 2, {0x440, 0xFFFE, 0x440, 0x440, 0x17D0, 0x1012, 0xF7DC, 0x1450, 0x3452, 0xD7D2, 0x100E, 0x100, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {15869, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0x55FC, 0x5820, 0x53FE, 0x9124, 0x10A8, 0x13FE, 0x1000, 0x29FC, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {15870, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0xAA42, 0xB278, 0xA3C4, 0xA23C, 0x2200, 0x23FE, 0x2240, 0x23A4, 0x5258, 0x4DB4, 0x4452, 0x89B0}},
    {15871, 2, {0x1020, 0x1020, 0x1050, 0x1488, 0x5904, 0x52FA, 0x5000, 0x9000, 0x11DC, 0x1154, 0x1154, 0x29DC, 0x2488, 0x4488, 0x4154, 0x8222}},
    {15872, 2, {0x1020, 0x13FE, 0x1000, 0x14F8, 0x5888, 0x50F8, 0x5000, 0x93FE, 0x1202, 0x12FA, 0x128A, 0x2AFA, 0x2626, 0x41F8, 0x4020, 0x83FE}},
    {15873, 2, {0x1050, 0x1252, 0x1154, 0x1450, 0x5BFE, 0x5088, 0x5050, 0x93FE, 0x1020, 0x11FC, 0x1020, 0x2BFE, 0x24A8, 0x4124, 0x4622, 0x8020}},
    {15874, 2, {0x110, 0x27FC, 0x1080, 0x348, 0x70B0, 0x1368, 0x10A4, 0x1320, 0x1040, 0x2FFE, 0x4100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {15875, 2, {0x10, 0x7F10, 0x4910, 0x6B1E, 0x5D10, 0x4910, 0x7F7C, 0x844, 0x7F44, 0x844, 0xF7C, 0xF044, 0x0, 0x4888, 0x4444, 0x8444}},
    {15876, 2, {0x2040, 0x2020, 0x23FE, 0x2088, 0xA852, 0xB7AC, 0xA2AA, 0xA2A8, 0x25AC, 0x2000, 0x2104, 0x21FC, 0x5104, 0x49FC, 0x4104, 0x8204}},
    {15877, 2, {0x2210, 0x2210, 0x23DE, 0x2528, 0xA884, 0xB000, 0xA7BC, 0xA4A4, 0x27A4, 0x24A4, 0x27A4, 0x2434, 0x5528, 0x4EA0, 0x4420, 0x8020}},
    {15878, 2, {0x1120, 0x11FC, 0x1220, 0x17FE, 0x5800, 0x51FC, 0x5104, 0x91FC, 0x1050, 0x13DE, 0x1050, 0x29DC, 0x2450, 0x43DE, 0x4050, 0x8050}},
    {15879, 2, {0x1088, 0x11DC, 0x1088, 0x13DE, 0x5488, 0x5954, 0x5222, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x29FC, 0x2504, 0x45FC, 0x4088, 0x8104}},
    {15880, 2, {0x810, 0x2448, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x2448, 0x3F7E, 0x2040, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {15881, 2, {0x2108, 0x27C8, 0xAAAA, 0xB12C, 0xAFE8, 0x2108, 0x57D4, 0x4912, 0x8922, 0x0, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {15882, 2, {0x2108, 0x2108, 0x27CE, 0x2112, 0xA924, 0xB7DE, 0xA552, 0xA55E, 0x27D2, 0x211E, 0x2392, 0x255E, 0x5940, 0x4914, 0x4112, 0x8122}},
    {15883, 2, {0x203C, 0x21E0, 0x3D24, 0x44A8, 0xA9FE, 0x50A8, 0x2124, 0x4202, 0x91FC, 0x5524, 0x5524, 0x59FC, 0x9124, 0x2924, 0x25FC, 0x4504}},
    {15884, 2, {0x11FC, 0x1020, 0x13FE, 0x1222, 0x55AC, 0x5820, 0x51AC, 0x9090, 0x11FE, 0x1320, 0x15FC, 0x2920, 0x25FC, 0x4520, 0x41FE, 0x8100}},
    {15885, 2, {0x2108, 0x21EC, 0x210A, 0x27E8, 0xAD28, 0xB5DE, 0xA728, 0xA4EA, 0x240A, 0x27EA, 0x240C, 0x27EC, 0x5A2A, 0x4BEA, 0x5156, 0x87E2}},
    {15886, 2, {0x13FC, 0x1108, 0x11F8, 0x1508, 0x59F8, 0x510E, 0x57F8, 0x9008, 0x17FE, 0x1294, 0x139C, 0x2A94, 0x279C, 0x42D6, 0x47BC, 0x8084}},
    {15887, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3532, 0x2F3C, 0x2120, 0x2738, 0x2120, 0x2F3C, 0x2488, 0x4950, 0x4630, 0xB80E}},
    {15888, 2, {0x2850, 0xFE48, 0x28FE, 0x3990, 0x1290, 0x7CFE, 0x5490, 0x7C90, 0x10FE, 0x7C90, 0x1090, 0xFEFE, 0x2880, 0x4400, 0x2488, 0x4244}},
    {15889, 2, {0x2FEC, 0x4824, 0x6BAC, 0x4824, 0x6BAC, 0x4AA4, 0xFFFE, 0x8822, 0x7EFC, 0x2C68, 0x4AA4, 0x100, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {15890, 2, {0xFC, 0x7F00, 0x208, 0x2108, 0x1110, 0x1020, 0xF8, 0x3F00, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x3FF8, 0x0}},
    {15891, 2, {0xFC, 0x7F00, 0x208, 0x2108, 0x1110, 0x1020, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {15892, 2, {0xF8, 0x7F00, 0x2110, 0x1110, 0x1220, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0xFF0, 0x810}},
    {15893, 2, {0x228, 0x77E, 0x3848, 0x2AFE, 0x2A48, 0x2A7E, 0x2A48, 0x2A7E, 0x2A40, 0x297E, 0x2922, 0x289C, 0x2862, 0x4830, 0x480E, 0x8800}},
    {15894, 2, {0x0, 0x28F8, 0x2488, 0x4288, 0x4088, 0x84F8, 0x488, 0x4888, 0x2888, 0x10F8, 0x1888, 0x2488, 0x2488, 0x4088, 0x83FE, 0x0}},
    {15895, 2, {0x840, 0x1020, 0x2850, 0x4488, 0x300, 0x1CE0, 0xE01E, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2028, 0x2010}},
    {15896, 2, {0x840, 0x1020, 0x2850, 0x4488, 0x300, 0x1CE0, 0xE01E, 0x1FF0, 0x1290, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x500, 0x200}},
    {15897, 2, {0x840, 0x1020, 0x2850, 0x4488, 0x300, 0x1CE0, 0xE21E, 0x1FD0, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0}},
    {15898, 2, {0x8, 0x8488, 0x4450, 0x2820, 0x1050, 0x2888, 0x4500, 0x8408, 0x88, 0x488, 0x4450, 0x2850, 0x1020, 0x2850, 0x4488, 0x8506}},
    {15899, 2, {0x400, 0x25FC, 0x2444, 0x2444, 0x2444, 0x3C44, 0x494, 0x488, 0xFD00, 0x24FC, 0x2484, 0x2484, 0x2484, 0x4484, 0x44FC, 0x8484}},
    {15900, 2, {0x420, 0x2410, 0x2410, 0x25FE, 0x2502, 0x3E54, 0x488, 0x504, 0xFC00, 0x24FC, 0x2420, 0x2420, 0x2420, 0x4420, 0x45FE, 0x8400}},
    {15901, 2, {0x440, 0x2440, 0x2440, 0x247C, 0x2444, 0x3E88, 0x20A0, 0x2120, 0x3C20, 0x2450, 0x2450, 0x2448, 0x2488, 0x4484, 0x4504, 0x8602}},
    {15902, 2, {0x410, 0x2490, 0x2490, 0x2490, 0x2490, 0x3EFE, 0x2080, 0x2080, 0x3C80, 0x24F8, 0x2488, 0x2488, 0x2488, 0x4488, 0x4488, 0x8508}},
    {15903, 2, {0x480, 0x2484, 0x2498, 0x24E0, 0x2482, 0x3E82, 0x207E, 0x2000, 0x3CFC, 0x2484, 0x2484, 0x24FC, 0x2484, 0x4484, 0x44FC, 0x8484}},
    {15904, 2, {0x890, 0x4890, 0x4890, 0x4A92, 0x4994, 0x7C98, 0x4090, 0x4198, 0x7A94, 0x4C92, 0x4890, 0x4890, 0x4912, 0x4912, 0x4A12, 0x8C0E}},
    {15905, 2, {0x820, 0x4920, 0x4920, 0x49FC, 0x4920, 0x7E20, 0x4020, 0x43FE, 0x7870, 0x48A8, 0x48A8, 0x4924, 0x4924, 0x4A22, 0x4820, 0x8820}},
    {15906, 2, {0x844, 0x4844, 0x49FE, 0x4844, 0x4844, 0x7DFE, 0x4040, 0x40FE, 0x7992, 0x4AFE, 0x4892, 0x48FE, 0x4892, 0x4892, 0x4892, 0x8886}},
    {15907, 2, {0x850, 0x4A52, 0x4954, 0x4850, 0x4BFE, 0x7C88, 0x4050, 0x43FE, 0x7820, 0x49FC, 0x4820, 0x4BFE, 0x48A8, 0x4924, 0x4E22, 0x8820}},
    {15908, 2, {0x4090, 0x2288, 0x2288, 0xFABE, 0x280, 0x53D4, 0x2208, 0xFA3E, 0x2388, 0x2288, 0x729C, 0x2288, 0x2288, 0x4488, 0x4488, 0x8888}},
    {15909, 2, {0x9F0, 0x4A10, 0x4FFC, 0x4A04, 0x4BFC, 0x7E24, 0x43B8, 0x4222, 0x79FE, 0x4A10, 0x4FFC, 0x4A44, 0x4BFC, 0x48D0, 0x494A, 0x8E3E}},
    {15910, 2, {0x0, 0x3FF8, 0x80, 0x1080, 0x1080, 0x2080, 0x3FFC, 0x80, 0x1080, 0x880, 0x680, 0x180, 0xF0, 0x8E, 0x280, 0x100}},
    {15911, 2, {0x0, 0x7EFC, 0x484, 0x2484, 0x24FC, 0x2484, 0x4484, 0x7EFC, 0xC92, 0x1492, 0x1494, 0x2488, 0x4488, 0x84A4, 0x14C2, 0x880}},
    {15912, 2, {0x804, 0x804, 0x4884, 0x4884, 0x7E84, 0x4884, 0x8884, 0x884, 0xE8C, 0xF894, 0x48E4, 0x884, 0x804, 0x804, 0x804, 0x804}},
    {15913, 2, {0x800, 0x8FC, 0x4820, 0x4820, 0x7E20, 0x4820, 0x8820, 0x9FE, 0xE20, 0xF820, 0x4820, 0x820, 0x820, 0x820, 0x820, 0x820}},
    {15914, 2, {0x1104, 0x1124, 0x5124, 0x5124, 0x7D24, 0x5124, 0x9124, 0x1124, 0x1D24, 0xF124, 0x5124, 0x1124, 0x1124, 0x1224, 0x1204, 0x1404}},
    {15915, 2, {0x804, 0x80E, 0x48F0, 0x4880, 0x7E80, 0x4880, 0x88FE, 0x888, 0xE88, 0xF888, 0x4888, 0x888, 0x888, 0x908, 0x908, 0xA08}},
    {15916, 2, {0x1020, 0x1020, 0x5020, 0x53FE, 0x7C20, 0x5020, 0x91FC, 0x1124, 0x1D24, 0xF124, 0x5124, 0x1134, 0x1128, 0x1020, 0x1020, 0x1020}},
    {15917, 2, {0x1000, 0x1000, 0x51FC, 0x5124, 0x7D24, 0x5124, 0x9124, 0x1124, 0x1DFC, 0xF100, 0x5100, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {15918, 2, {0x1010, 0x1090, 0x5090, 0x5088, 0x7D08, 0x5104, 0x9204, 0x15FA, 0x1C88, 0xF088, 0x5088, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {15919, 2, {0x808, 0x888, 0x4848, 0x4848, 0x7E08, 0x4888, 0x8848, 0x848, 0xE08, 0xF80E, 0x49F8, 0x808, 0x808, 0x808, 0x808, 0x808}},
    {15920, 2, {0x1020, 0x1010, 0x5010, 0x51FE, 0x7D02, 0x5204, 0x9080, 0x1088, 0x1C90, 0xF0A0, 0x50C0, 0x1082, 0x1082, 0x1082, 0x107E, 0x1000}},
    {15921, 2, {0x1080, 0x1080, 0x50FE, 0x5100, 0x7A20, 0x5120, 0x912C, 0x1174, 0x1BA4, 0x3124, 0xD134, 0x1128, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {15922, 2, {0x840, 0x840, 0x4840, 0x487E, 0x7EA0, 0x48A0, 0x8920, 0x83C, 0xE20, 0xF820, 0x4820, 0x83E, 0x820, 0x820, 0x820, 0x820}},
    {15923, 2, {0x1020, 0x1020, 0x5050, 0x5050, 0x7C88, 0x5124, 0x9212, 0x1010, 0x1DFC, 0xF004, 0x5008, 0x1088, 0x1050, 0x1020, 0x1010, 0x1010}},
    {15924, 2, {0x0, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15925, 2, {0x1020, 0x1020, 0x5020, 0x53FE, 0x7C20, 0x5020, 0x91FC, 0x1000, 0x1C00, 0xF1FC, 0x5104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15926, 2, {0x1000, 0x11FC, 0x5104, 0x5124, 0x7D24, 0x5124, 0x91FC, 0x1124, 0x1D24, 0xF154, 0x514C, 0x118C, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {15927, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x900, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15928, 2, {0x1008, 0x101C, 0x51E0, 0x5100, 0x7D00, 0x51FE, 0x9100, 0x1100, 0x1D7C, 0xF144, 0x5144, 0x1144, 0x1144, 0x127C, 0x1244, 0x1400}},
    {15929, 2, {0x1004, 0x101E, 0x51E0, 0x5022, 0x7D12, 0x5094, 0x9080, 0x1008, 0x1DFE, 0xF008, 0x5088, 0x1048, 0x1048, 0x1008, 0x1028, 0x1010}},
    {15930, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x92A, 0x7230, 0x10C0, 0x1700, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15931, 2, {0x1020, 0x1040, 0x5088, 0x5104, 0x7DFE, 0x5002, 0x9080, 0x10FC, 0x1D20, 0xF020, 0x53FE, 0x1020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {15932, 2, {0x1040, 0x1040, 0x50FC, 0x5088, 0x7D50, 0x5020, 0x90D8, 0x1326, 0x1CF8, 0xF020, 0x50F8, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {15933, 2, {0x800, 0x8FC, 0x4884, 0x4884, 0x7EFC, 0x4884, 0x8884, 0x8FC, 0xE84, 0xF884, 0x48FC, 0x800, 0x848, 0x844, 0x882, 0x902}},
    {15934, 2, {0x1000, 0x13FC, 0x5044, 0x5144, 0x7944, 0x5284, 0x9094, 0x1108, 0x1A40, 0x3020, 0xD0A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {15935, 2, {0x1020, 0x1020, 0x53FE, 0x5020, 0x7DFC, 0x5000, 0x93FE, 0x1202, 0x1C04, 0xF0F0, 0x5090, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E}},
    {15936, 2, {0x1020, 0x1020, 0x53FE, 0x5020, 0x7820, 0x53FE, 0x9202, 0x1404, 0x19F8, 0x3010, 0xD020, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {15937, 2, {0x1040, 0x1020, 0x53FE, 0x5000, 0x7C00, 0x51FC, 0x9104, 0x1104, 0x1DFC, 0xF020, 0x5128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {15938, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15939, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x540, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15940, 2, {0x1000, 0x11FC, 0x5124, 0x51FC, 0x7D24, 0x51FC, 0x9088, 0x1144, 0x1E42, 0xF0F8, 0x5188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {15941, 2, {0x1080, 0x1080, 0x50FC, 0x5154, 0x7AD4, 0x50B4, 0x912C, 0x1244, 0x1894, 0x3108, 0xD040, 0x10A4, 0x12AA, 0x128A, 0x1478, 0x1000}},
    {15942, 2, {0x1000, 0x13FE, 0x5222, 0x5020, 0x7DFE, 0x5020, 0x91FC, 0x1124, 0x1DFC, 0xF124, 0x51FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {15943, 2, {0x1000, 0x11FC, 0x5104, 0x51FC, 0x7D04, 0x51FC, 0x9080, 0x11FE, 0x1E22, 0xF122, 0x5152, 0x1102, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {15944, 2, {0x820, 0x820, 0x7E7E, 0x8C4, 0xFF28, 0x1C10, 0x2A28, 0x48C6, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15945, 2, {0x1000, 0x11FE, 0x5110, 0x5120, 0x7D7C, 0x5144, 0x917C, 0x1144, 0x1D7C, 0xF110, 0x5110, 0x1154, 0x1152, 0x1292, 0x1250, 0x1420}},
    {15946, 2, {0x1040, 0x1040, 0x53FE, 0x52A2, 0x7C90, 0x51FE, 0x9120, 0x1320, 0x1DFC, 0xF120, 0x5120, 0x11FC, 0x1120, 0x1120, 0x11FE, 0x1100}},
    {15947, 2, {0x1120, 0x113C, 0x5144, 0x52A8, 0x7A10, 0x56A8, 0x9AC6, 0x1280, 0x1A88, 0x32B0, 0xD284, 0x1288, 0x1232, 0x1204, 0x1218, 0x1260}},
    {15948, 2, {0x1040, 0x1020, 0x53FC, 0x5000, 0x7D08, 0x5090, 0x93FE, 0x1220, 0x1EA0, 0xF2FC, 0x5320, 0x1220, 0x12FC, 0x1420, 0x1420, 0x19FE}},
    {15949, 2, {0x1040, 0x1020, 0x53FE, 0x5088, 0x7C50, 0x53FE, 0x9222, 0x12FA, 0x1E22, 0xF2FA, 0x528A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {15950, 2, {0x1000, 0x11FC, 0x5124, 0x51FC, 0x7D24, 0x51FC, 0x9040, 0x1088, 0x1DF0, 0xF020, 0x5044, 0x13FE, 0x1022, 0x1124, 0x1222, 0x1060}},
    {15951, 2, {0x1088, 0x1088, 0x53FE, 0x5088, 0x7CF8, 0x5020, 0x91FC, 0x1124, 0x1DFC, 0xF020, 0x53FE, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {15952, 2, {0x1110, 0x2110, 0xCAA8, 0x1444, 0x3240, 0xD27C, 0x1540, 0x14FE, 0x900, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {15953, 2, {0x1014, 0x17FE, 0x5010, 0x53D0, 0x7810, 0x53D0, 0x9010, 0x13D0, 0x1A50, 0x33D0, 0xD250, 0x13C8, 0x124A, 0x13CA, 0x1186, 0x1242}},
    {15954, 2, {0x1050, 0x1252, 0x5154, 0x5050, 0x7BFE, 0x5088, 0x9050, 0x13FE, 0x1820, 0x31FC, 0xD020, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1602}},
    {15955, 2, {0x1020, 0x1020, 0x51FC, 0x5020, 0x7BFE, 0x5108, 0x939C, 0x1108, 0x1988, 0x363E, 0xD000, 0x13FE, 0x1090, 0x1090, 0x1112, 0x120E}},
    {15956, 2, {0x1000, 0x13FE, 0x5000, 0x51FC, 0x7D24, 0x51FC, 0x9124, 0x13FE, 0x1C00, 0xF1FC, 0x5124, 0x11FC, 0x1124, 0x11FC, 0x1000, 0x13FE}},
    {15957, 2, {0x7DFE, 0x20, 0xFC, 0xFE84, 0x28FC, 0x2884, 0x28FC, 0x4A84, 0x4CFC, 0x8848, 0x1184, 0x1FF8, 0x2100, 0xFFFE, 0x100, 0x100}},
    {15958, 2, {0x1020, 0x1124, 0x50A8, 0x51FC, 0x7840, 0x53FE, 0x9088, 0x1104, 0x1AFA, 0x3040, 0xD1A4, 0x1058, 0x11B4, 0x1052, 0x1390, 0x1030}},
    {15959, 2, {0x1020, 0x13FE, 0x5020, 0x51FC, 0x7C00, 0x53FE, 0x9002, 0x11FC, 0x1C20, 0xF3FE, 0x5000, 0x13FE, 0x1004, 0x13BE, 0x12A4, 0x138C}},
    {15960, 2, {0x1000, 0x11FC, 0x5020, 0x53FE, 0x7E22, 0x51AC, 0x9020, 0x11AC, 0x1C00, 0xF3FE, 0x5020, 0x11FC, 0x1154, 0x1154, 0x1154, 0x110C}},
    {15961, 2, {0x1088, 0x12AA, 0x52DC, 0x5488, 0x7954, 0x5222, 0x93FE, 0x1202, 0x1DFC, 0x3104, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {15962, 2, {0x1088, 0x12AA, 0x52DC, 0x5488, 0x7954, 0x5222, 0x93FE, 0x1202, 0x18F8, 0x3088, 0xD0F8, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {15963, 2, {0x2000, 0x27FE, 0xA040, 0xA2EE, 0xFAAA, 0xABEA, 0xAAAE, 0x2AE8, 0x3BAA, 0xECE6, 0x2288, 0x25F4, 0x2290, 0x2060, 0x2198, 0x2606}},
    {15964, 2, {0x820, 0xA28, 0x924, 0x924, 0x820, 0x7EFE, 0x820, 0x820, 0x850, 0x1450, 0x1250, 0x1288, 0x2088, 0x2104, 0x4204, 0x8402}},
    {15965, 2, {0x4, 0x441E, 0x29F0, 0x1110, 0x2910, 0x4910, 0x8910, 0x9FE, 0x1910, 0x2910, 0x4910, 0x8908, 0x90A, 0x94A, 0x5186, 0x2102}},
    {15966, 2, {0x8, 0x441C, 0x29E0, 0x1100, 0x2900, 0x4900, 0x89FE, 0x910, 0x1910, 0x2910, 0x4910, 0x8910, 0x910, 0xA10, 0x5210, 0x2410}},
    {15967, 2, {0x40, 0x4440, 0x2840, 0x13F8, 0x2848, 0x4848, 0x8848, 0x848, 0x1FFE, 0x2840, 0x48A0, 0x88A0, 0x910, 0x910, 0x5208, 0x2406}},
    {15968, 2, {0x20, 0x4420, 0x2820, 0x13FE, 0x2820, 0x4924, 0x8924, 0x924, 0x1924, 0x29FC, 0x4824, 0x8820, 0x822, 0x822, 0x501E, 0x2000}},
    {15969, 2, {0x40, 0x4420, 0x2820, 0x13FE, 0x2840, 0x4840, 0x8888, 0x908, 0x1BF0, 0x2820, 0x4840, 0x8888, 0x904, 0xBFC, 0x5104, 0x2000}},
    {15970, 2, {0x200, 0x8A00, 0x5200, 0x221E, 0x57D2, 0x9252, 0x1252, 0x1252, 0x3252, 0x5252, 0x9252, 0x1252, 0x1452, 0x145E, 0xA952, 0x5080}},
    {15971, 2, {0x80, 0x8880, 0x50F8, 0x2108, 0x5310, 0x94A0, 0x1040, 0x10A0, 0x3318, 0x5C06, 0x90C0, 0x1020, 0x1010, 0x1180, 0xA060, 0x4010}},
    {15972, 2, {0x20, 0x4420, 0x2820, 0x11FC, 0x2820, 0x4820, 0x8820, 0xBFE, 0x1820, 0x2840, 0x4840, 0x8888, 0x904, 0xBFE, 0x5102, 0x2000}},
    {15973, 2, {0x40, 0x4420, 0x2800, 0x13FE, 0x2820, 0x4820, 0x8820, 0x820, 0x19FC, 0x2820, 0x4820, 0x8820, 0x820, 0x820, 0x53FE, 0x2000}},
    {15974, 2, {0x0, 0x8BFC, 0x5084, 0x2084, 0x5084, 0x9104, 0x1114, 0x1208, 0x3400, 0x51FC, 0x9104, 0x1104, 0x1104, 0x1104, 0xA1FC, 0x4104}},
    {15975, 2, {0x20, 0x4420, 0x2820, 0x11FC, 0x2924, 0x4924, 0x8924, 0x924, 0x1924, 0x2BFE, 0x4820, 0x8850, 0x850, 0x888, 0x5104, 0x2202}},
    {15976, 2, {0x20, 0x4420, 0x2820, 0x1020, 0x29FC, 0x4924, 0x8924, 0x924, 0x1924, 0x29FC, 0x4924, 0x8924, 0x924, 0x924, 0x51FC, 0x2104}},
    {15977, 2, {0x0, 0x45FC, 0x2884, 0x1088, 0x2850, 0x4820, 0x8850, 0x888, 0x1B26, 0x2820, 0x49FC, 0x8820, 0x820, 0x820, 0x53FE, 0x2000}},
    {15978, 2, {0x0, 0x97BC, 0x54A4, 0x24A4, 0x54A4, 0x94A4, 0x14A4, 0x1FFE, 0x34A4, 0x54A4, 0x94A4, 0x14A4, 0x14A4, 0x15A4, 0xA854, 0x5088}},
    {15979, 2, {0x20, 0x4420, 0x2BFE, 0x1020, 0x29FC, 0x4824, 0x89FC, 0x920, 0x19FE, 0x2822, 0x482A, 0x8854, 0x850, 0x888, 0x5104, 0x2202}},
    {15980, 2, {0x40, 0x8840, 0x5FFE, 0x2040, 0x5040, 0x97FC, 0x1444, 0x1444, 0x37FC, 0x5040, 0x9240, 0x1140, 0x1080, 0x1140, 0xA230, 0x4C0E}},
    {15981, 2, {0x20, 0x4420, 0x29FC, 0x1124, 0x2924, 0x49FC, 0x8924, 0x924, 0x19FC, 0x2820, 0x4824, 0x8818, 0x832, 0x8CA, 0x5306, 0x2002}},
    {15982, 2, {0x0, 0x45FE, 0x2820, 0x1020, 0x2840, 0x49FC, 0x8904, 0x904, 0x1904, 0x29FC, 0x4904, 0x8904, 0x904, 0x904, 0x51FC, 0x2104}},
    {15983, 2, {0x90, 0x4490, 0x2890, 0x1110, 0x29FE, 0x4B10, 0x8D10, 0x938, 0x1938, 0x2954, 0x4954, 0x8992, 0x910, 0x910, 0x5110, 0x2110}},
    {15984, 2, {0x40, 0x4444, 0x2A44, 0x1148, 0x2950, 0x4840, 0x8FFE, 0x890, 0x1890, 0x2890, 0x4890, 0x8892, 0x912, 0x912, 0x520E, 0x2400}},
    {15985, 2, {0x0, 0x45FC, 0x2904, 0x1104, 0x2904, 0x49FC, 0x8820, 0x820, 0x1920, 0x293C, 0x4920, 0x8920, 0xAA0, 0xA60, 0x543E, 0x2800}},
    {15986, 2, {0x20, 0x4422, 0x29FA, 0x1024, 0x2824, 0x4BFE, 0x8810, 0x820, 0x19FC, 0x2888, 0x4910, 0x8AFE, 0xC10, 0x810, 0x5050, 0x2020}},
    {15987, 2, {0x0, 0x47FE, 0x2840, 0x1040, 0x29FC, 0x4884, 0x8884, 0xBFE, 0x1800, 0x2800, 0x49FC, 0x8904, 0x904, 0x904, 0x51FC, 0x2104}},
    {15988, 2, {0x0, 0x45FE, 0x2820, 0x1020, 0x29FC, 0x4924, 0x8924, 0x9FC, 0x1924, 0x2924, 0x49FC, 0x8920, 0x8A0, 0x840, 0x50B0, 0x230E}},
    {15989, 2, {0x0, 0x45FE, 0x2800, 0x1092, 0x2924, 0x4A48, 0x8924, 0x892, 0x1800, 0x29FE, 0x4820, 0x8820, 0x820, 0x820, 0x53FE, 0x2000}},
    {15990, 2, {0x0, 0x8806, 0x53B8, 0x2088, 0x5088, 0x9108, 0x113E, 0x1388, 0x3088, 0x5288, 0x9288, 0x113E, 0x1100, 0x1280, 0xA47E, 0x4800}},
    {15991, 2, {0x4210, 0x2414, 0x1812, 0x2412, 0x5210, 0x10FE, 0xFF10, 0x2010, 0x2810, 0x7E28, 0xAA28, 0x2A28, 0x2A44, 0x2E44, 0x884, 0x902}},
    {15992, 2, {0x20, 0x4440, 0x29FC, 0x1104, 0x2904, 0x49FC, 0x8904, 0x904, 0x19FC, 0x2850, 0x4850, 0x8850, 0x892, 0x892, 0x510E, 0x2200}},
    {15993, 2, {0x40, 0x8820, 0x53FE, 0x2220, 0x52FC, 0x9224, 0x1224, 0x13FE, 0x3224, 0x5224, 0x92FC, 0x1220, 0x1250, 0x1450, 0xA488, 0x4906}},
    {15994, 2, {0x20, 0x4422, 0x2BB4, 0x10A8, 0x28A8, 0x4924, 0x8AA2, 0x840, 0x19FC, 0x2904, 0x4904, 0x89FC, 0x904, 0x904, 0x51FC, 0x2104}},
    {15995, 2, {0x1C, 0x45E0, 0x2820, 0x1020, 0x2BFE, 0x48A8, 0x8924, 0xA42, 0x1840, 0x2BFE, 0x4888, 0x8908, 0x8D0, 0x830, 0x5048, 0x2184}},
    {15996, 2, {0x80, 0x4480, 0x28FE, 0x1102, 0x2A42, 0x4A4A, 0x8952, 0x842, 0x1BFE, 0x2842, 0x48E2, 0x8952, 0xA4A, 0x842, 0x5054, 0x2008}},
    {15997, 2, {0x50, 0x4448, 0x285E, 0x13E4, 0x2828, 0x4832, 0x88CA, 0xF06, 0x1850, 0x2848, 0x485E, 0x8BE4, 0x828, 0x832, 0x50CA, 0x2706}},
    {15998, 2, {0x110, 0x8910, 0x53D8, 0x2254, 0x5452, 0x9A90, 0x1110, 0x12FE, 0x3C00, 0x5000, 0x93F8, 0x1208, 0x1208, 0x1208, 0xA3F8, 0x4208}},
    {15999, 2, {0x20, 0x4420, 0x283E, 0x1020, 0x29FC, 0x4904, 0x89FC, 0x904, 0x19FC, 0x2924, 0x4820, 0x8BFE, 0x820, 0x820, 0x5020, 0x2020}},
    {16000, 2, {0x80, 0x4480, 0x28FC, 0x1154, 0x2A54, 0x48A4, 0x8924, 0xA44, 0x1894, 0x2908, 0x4840, 0x88A4, 0xAAA, 0xA8A, 0x5478, 0x2000}},
    {16001, 2, {0x92, 0x4524, 0x2A48, 0x1124, 0x2892, 0x4840, 0x8880, 0xBFE, 0x1A02, 0x2A8A, 0x4A52, 0x8A22, 0xA52, 0xA8A, 0x53FE, 0x2202}},
    {16002, 2, {0x14, 0x8812, 0x5010, 0x27FE, 0x5410, 0x9410, 0x15D0, 0x1412, 0x3412, 0x55D4, 0x9554, 0x1548, 0x15DA, 0x142A, 0xA846, 0x5082}},
    {16003, 2, {0x88, 0x4488, 0x2BFE, 0x1088, 0x2900, 0x49FC, 0x8904, 0xA04, 0x19F4, 0x2914, 0x4914, 0x89F4, 0x914, 0x804, 0x5028, 0x2010}},
    {16004, 2, {0x40, 0x44A0, 0x2910, 0x1208, 0x2DF6, 0x4800, 0x8BC4, 0xA54, 0x1A54, 0x2BD4, 0x4A54, 0x8A54, 0xBD4, 0xA44, 0x5254, 0x22C8}},
    {16005, 2, {0x80, 0x8880, 0x51FC, 0x2204, 0x5408, 0x9BFE, 0x1200, 0x1280, 0x32FC, 0x5320, 0x9220, 0x13FE, 0x1420, 0x1450, 0xA888, 0x5306}},
    {16006, 2, {0x210, 0x4510, 0x287E, 0x1310, 0x2928, 0x49A8, 0x8944, 0x800, 0x19FC, 0x2904, 0x4924, 0x8924, 0x924, 0x850, 0x5088, 0x2304}},
    {16007, 2, {0x10, 0x9014, 0x5FD2, 0x2912, 0x6910, 0xA97E, 0x2B90, 0x2A90, 0x6A90, 0xAB98, 0x2928, 0x2928, 0x2924, 0x2FC4, 0xA044, 0x4082}},
    {16008, 2, {0x20, 0x4524, 0x2924, 0x11FC, 0x2820, 0x4BFE, 0x8800, 0x820, 0x19FC, 0x2924, 0x4924, 0x89FC, 0x820, 0x824, 0x53FE, 0x2102}},
    {16009, 2, {0x1000, 0x1078, 0xFE48, 0x1048, 0x7C48, 0x86, 0xFE00, 0x82FC, 0x7C44, 0x44, 0x1428, 0xFE28, 0x1010, 0x2828, 0x4444, 0x8182}},
    {16010, 2, {0x104, 0x4488, 0x2800, 0x13FE, 0x2820, 0x4820, 0x89FC, 0x820, 0x1820, 0x2BFE, 0x4840, 0x8824, 0xAA2, 0xA8A, 0x5488, 0x2078}},
    {16011, 2, {0x120, 0x8910, 0x5208, 0x2486, 0x5910, 0x93F8, 0x1008, 0x1000, 0x37BC, 0x5084, 0x94A4, 0x1294, 0x14A4, 0x1084, 0xA294, 0x4108}},
    {16012, 2, {0x110, 0x88A0, 0x5000, 0x27FE, 0x50A0, 0x93F8, 0x10A8, 0x17FE, 0x30A8, 0x53F8, 0x90A0, 0x11B0, 0x12A8, 0x14A6, 0xA0A0, 0x40A0}},
    {16013, 2, {0x40, 0x8820, 0x53FE, 0x2202, 0x5414, 0x91E0, 0x1100, 0x1100, 0x31FC, 0x5110, 0x9110, 0x17FE, 0x1000, 0x1090, 0xA108, 0x4204}},
    {16014, 2, {0x20, 0x97FE, 0x5420, 0x25FC, 0x5424, 0x97FE, 0x1424, 0x15FC, 0x3420, 0x55FC, 0x9524, 0x15FC, 0x1524, 0x19FC, 0xA924, 0x512C}},
    {16015, 2, {0x20, 0x45FC, 0x2924, 0x13FE, 0x2924, 0x49FC, 0x8820, 0x9FC, 0x1924, 0x29FC, 0x4840, 0x8BFE, 0x888, 0x9D0, 0x5070, 0x238C}},
    {16016, 2, {0x124, 0x4648, 0x2924, 0x1000, 0x29FC, 0x4924, 0x89FC, 0x924, 0x19FC, 0x2820, 0x4BFE, 0x8870, 0x8A8, 0x924, 0x5622, 0x2020}},
    {16017, 2, {0x40, 0x4488, 0x29FC, 0x1108, 0x2A52, 0x4BFE, 0x8850, 0x988, 0x1E26, 0x28C0, 0x4B10, 0x8864, 0xB88, 0x830, 0x50C0, 0x2700}},
    {16018, 2, {0x0, 0x97BC, 0x5084, 0x24A4, 0x5294, 0x94A4, 0x1050, 0x1188, 0x3626, 0x50C0, 0x9310, 0x1064, 0x1388, 0x1030, 0xA0C0, 0x4700}},
    {16019, 2, {0x20, 0x47FE, 0x2820, 0x11FC, 0x2820, 0x4BFE, 0x8800, 0x9FC, 0x1904, 0x29FC, 0x4904, 0x89FC, 0x904, 0x9FC, 0x5088, 0x2104}},
    {16020, 2, {0xE, 0x8BF0, 0x5044, 0x2224, 0x5108, 0x93F8, 0x1208, 0x13FC, 0x3204, 0x53FE, 0x9202, 0x12AA, 0x12AA, 0x1502, 0xA414, 0x4808}},
    {16021, 2, {0x78, 0x97C0, 0x5248, 0x2150, 0x57FC, 0x9150, 0x1248, 0x1406, 0x33F8, 0x5248, 0x9248, 0x13F8, 0x1248, 0x1248, 0xA3F8, 0x4208}},
    {16022, 2, {0x8, 0x9788, 0x5088, 0x2110, 0x6FDE, 0xA494, 0x24A4, 0x2794, 0x6494, 0xA794, 0x2494, 0x24C8, 0x2788, 0x2C94, 0xA0A4, 0x40C2}},
    {16023, 2, {0x0, 0x47DE, 0x2A52, 0x1252, 0x2BDE, 0x4800, 0x89FC, 0x924, 0x19FC, 0x2924, 0x49FC, 0x8820, 0xBFE, 0x820, 0x5020, 0x2020}},
    {16024, 2, {0x90, 0x4488, 0x29FE, 0x1110, 0x2B10, 0x4DFE, 0x8910, 0x910, 0x19FE, 0x2910, 0x4910, 0x89FE, 0x900, 0xAA4, 0x5252, 0x2452}},
    {16025, 2, {0x80, 0x8840, 0x57FE, 0x2402, 0x51F8, 0x9148, 0x17FE, 0x1228, 0x33FC, 0x5010, 0x93F8, 0x1248, 0x1248, 0x12A8, 0xA110, 0x4608}},
    {16026, 2, {0x10, 0x9410, 0x527C, 0x2010, 0x50FE, 0x9044, 0x1628, 0x12FE, 0x3210, 0x527C, 0x9210, 0x12FE, 0x1210, 0x1210, 0xA5FE, 0x4800}},
    {16027, 2, {0x3FC, 0x8A40, 0x53F8, 0x2240, 0x53F8, 0x9240, 0x13FC, 0x1004, 0x3554, 0x5008, 0x9110, 0x1FFE, 0x1110, 0x1110, 0xA210, 0x4410}},
    {16028, 2, {0x90, 0x8BFC, 0x5294, 0x23FC, 0x5294, 0x93FC, 0x1000, 0x13FC, 0x3200, 0x52F8, 0x9200, 0x13FE, 0x1520, 0x1514, 0xA548, 0x4986}},
    {16029, 2, {0x10, 0x97D0, 0x5510, 0x27DE, 0x5450, 0x97E8, 0x1504, 0x17C4, 0x3000, 0x53F8, 0x92A8, 0x12A8, 0x12A8, 0x12A8, 0xAFFE, 0x4000}},
    {16030, 2, {0x140, 0x8A4C, 0x5264, 0x2244, 0x534C, 0x9264, 0x12A4, 0x13AC, 0x32A4, 0x52A4, 0x97FE, 0x1000, 0x1110, 0x1108, 0xA204, 0x4404}},
    {16031, 2, {0x0, 0x97BC, 0x5084, 0x24A4, 0x5294, 0x94A4, 0x1120, 0x1210, 0x37FC, 0x5A20, 0x93FC, 0x1220, 0x13FC, 0x1220, 0xA3FE, 0x4200}},
    {16032, 2, {0x80, 0x9040, 0x5FFE, 0x2912, 0x6208, 0xA584, 0x2E0E, 0x2A74, 0x6A54, 0xAB54, 0x2AD4, 0x2A54, 0x2A52, 0x2D54, 0xB75C, 0x4084}},
    {16033, 2, {0x0, 0x97FE, 0x5444, 0x2598, 0x6488, 0xA7DE, 0x2488, 0x25DC, 0x66AA, 0xA488, 0x2420, 0x2520, 0x253C, 0x2920, 0xA920, 0x57FE}},
    {16034, 2, {0x7FC, 0x9040, 0x57FE, 0x2442, 0x5B5C, 0x9040, 0x175C, 0x1120, 0x33FC, 0x5620, 0x9BFC, 0x1220, 0x13FC, 0x1220, 0xA3FE, 0x4200}},
    {16035, 2, {0x108, 0x91EC, 0x510A, 0x27E8, 0x6528, 0xA5DE, 0x2728, 0x24EA, 0x640A, 0xA7EA, 0x240C, 0x27EC, 0x2A2A, 0x2BEA, 0xB156, 0x47E2}},
    {16036, 2, {0x148, 0x894C, 0x52AA, 0x2008, 0x57FE, 0x9148, 0x1368, 0x114A, 0x336A, 0x514C, 0x936C, 0x1148, 0x116A, 0x139A, 0xA026, 0x4042}},
    {16037, 2, {0x1F0, 0x8A10, 0x57FC, 0x2204, 0x53FC, 0x9224, 0x13B8, 0x1222, 0x31FE, 0x5210, 0x97FC, 0x1244, 0x13FC, 0x10D0, 0xA14A, 0x463E}},
    {16038, 2, {0x44, 0x9F44, 0x5AAA, 0x2AEE, 0x6E44, 0xAAAA, 0x2AEE, 0x2E00, 0x6A28, 0xAAAA, 0x2BAA, 0x2EEE, 0x3A28, 0x2228, 0xA248, 0x4288}},
    {16039, 2, {0x7BC, 0x94A4, 0x57BC, 0x2000, 0x57FC, 0x9080, 0x13F8, 0x1208, 0x33F8, 0x5208, 0x93F8, 0x1110, 0x1208, 0x17BC, 0xA4A4, 0x47BC}},
    {16040, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x1020, 0x1040, 0x7C40, 0x1040, 0x1080, 0x1088, 0x1084, 0x1D04, 0xE3FE, 0x4102, 0x2, 0x0}},
    {16041, 2, {0x4, 0x4, 0xFE84, 0x1084, 0x1084, 0x1084, 0x7C84, 0x1084, 0x108C, 0x1094, 0x10E4, 0x1E84, 0xF004, 0x4004, 0x4, 0x4}},
    {16042, 2, {0x40, 0x40, 0xFE40, 0x1040, 0x1040, 0x1050, 0x7C48, 0x1044, 0x1044, 0x1040, 0x1040, 0x1E40, 0xF040, 0x4040, 0x40, 0x40}},
    {16043, 2, {0x8, 0x8, 0xFE08, 0x1088, 0x1088, 0x1088, 0x7C88, 0x1088, 0x1088, 0x1088, 0x1088, 0x1E08, 0xF008, 0x4008, 0x28, 0x10}},
    {16044, 2, {0x0, 0x1F0, 0xFD10, 0x1110, 0x1110, 0x1110, 0x7D90, 0x1150, 0x1150, 0x1110, 0x1110, 0x1D12, 0xE212, 0x4212, 0x40E, 0x800}},
    {16045, 2, {0x0, 0x1FC, 0xFC04, 0x1008, 0x1010, 0x1020, 0x7C20, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {16046, 2, {0x0, 0x1FC, 0xFC00, 0x1000, 0x1000, 0x13FE, 0x1040, 0x7C40, 0x1080, 0x10FC, 0x1004, 0x1004, 0x1C04, 0xE004, 0x4028, 0x10}},
    {16047, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {16048, 2, {0x0, 0x3FE, 0xFC20, 0x1020, 0x1020, 0x11FC, 0x7D24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D34, 0xE128, 0x4020, 0x20, 0x20}},
    {16049, 2, {0x1000, 0x17F8, 0x2110, 0x613C, 0xA104, 0x2204, 0x2414, 0x2808, 0x2000, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16050, 2, {0x0, 0x1F8, 0xFC48, 0x1048, 0x1048, 0x1048, 0x7C48, 0x11F8, 0x1088, 0x1088, 0x1088, 0x1C88, 0xE088, 0x4088, 0x3FE, 0x0}},
    {16051, 2, {0x20, 0x20, 0xFE20, 0x103E, 0x1020, 0x1020, 0x7C20, 0x11FC, 0x1084, 0x1088, 0x1048, 0x1E50, 0xF020, 0x4050, 0x188, 0x606}},
    {16052, 2, {0x20, 0x28, 0xFE24, 0x1024, 0x1020, 0x11FE, 0x7C20, 0x1020, 0x1020, 0x1050, 0x1050, 0x1E50, 0xF088, 0x4088, 0x104, 0x202}},
    {16053, 2, {0x80, 0x80, 0xFC80, 0x10FC, 0x1104, 0x1108, 0x1240, 0x7C40, 0x1040, 0x10A0, 0x10A0, 0x1110, 0x1D10, 0xE208, 0x4404, 0x802}},
    {16054, 2, {0x8, 0x88, 0xFE48, 0x1048, 0x1008, 0x1088, 0x7C48, 0x1048, 0x1008, 0x100E, 0x11F8, 0x1E08, 0xF008, 0x4008, 0x8, 0x8}},
    {16055, 2, {0x20, 0x28, 0xFC24, 0x1020, 0x11FC, 0x1020, 0x7C70, 0x1070, 0x10A8, 0x10A8, 0x1124, 0x1D24, 0xE222, 0x4020, 0x20, 0x20}},
    {16056, 2, {0x800, 0x9F0, 0x1110, 0x3190, 0x5150, 0x9152, 0x1112, 0x120E, 0x400, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16057, 2, {0x20, 0x20, 0xFC40, 0x1088, 0x1104, 0x13FE, 0x7C02, 0x1088, 0x1088, 0x13FE, 0x1088, 0x1C88, 0xE108, 0x4108, 0x208, 0x408}},
    {16058, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1104, 0x11FC, 0x1104, 0x7D04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {16059, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x1020, 0x7DFC, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D34, 0xE128, 0x4020, 0x20, 0x20}},
    {16060, 2, {0x0, 0x3FE, 0xFC22, 0x1024, 0x1020, 0x1020, 0x7D20, 0x113C, 0x1120, 0x1120, 0x1120, 0x1D20, 0xE2A0, 0x4260, 0x43E, 0x800}},
    {16061, 2, {0x0, 0x1FE, 0xFC10, 0x1010, 0x1020, 0x1020, 0x7C68, 0x10A4, 0x1122, 0x1222, 0x1020, 0x1C20, 0xE020, 0x4000, 0x3FE, 0x0}},
    {16062, 2, {0x0, 0x3FC, 0xFA04, 0x2204, 0x2204, 0x22F4, 0xFA94, 0x2294, 0x2294, 0x2294, 0x22F4, 0x3A94, 0xE204, 0x4204, 0x214, 0x208}},
    {16063, 2, {0x0, 0x1FE, 0xFD10, 0x1110, 0x1110, 0x117C, 0x1144, 0x7D44, 0x1144, 0x1144, 0x117C, 0x1110, 0x1D10, 0xE110, 0x41FE, 0x0}},
    {16064, 2, {0x20, 0x20, 0xFBFE, 0x2222, 0x2222, 0x23FE, 0x2222, 0xFA22, 0x23FE, 0x2020, 0x2120, 0x38A0, 0xE040, 0x40A0, 0x118, 0x606}},
    {16065, 2, {0x0, 0x0, 0xFDFC, 0x1104, 0x1104, 0x1174, 0x7D54, 0x1154, 0x1154, 0x1154, 0x1174, 0x1D04, 0xE104, 0x41FC, 0x104, 0x0}},
    {16066, 2, {0x104, 0x84, 0xFC88, 0x1000, 0x13FE, 0x1088, 0x7C88, 0x1088, 0x1088, 0x17FE, 0x1088, 0x1C88, 0xE108, 0x4108, 0x208, 0x408}},
    {16067, 2, {0x440, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16068, 2, {0x8, 0x3C, 0xFDC0, 0x1000, 0x1000, 0x13FE, 0x7C20, 0x1020, 0x1120, 0x113C, 0x1120, 0x1D20, 0xE2A0, 0x4260, 0x43E, 0x800}},
    {16069, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x1202, 0x10F8, 0x7C00, 0x1000, 0x13FE, 0x1020, 0x1040, 0x1C88, 0xE104, 0x43FE, 0x102}},
    {16070, 2, {0x40, 0x80, 0xF908, 0x2204, 0x27FE, 0x2002, 0x2000, 0xFBFC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x3A04, 0xE204, 0x214, 0x208}},
    {16071, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x1202, 0x7DFC, 0x1020, 0x1020, 0x1020, 0x11FC, 0x1C20, 0xE020, 0x4020, 0x3FE, 0x0}},
    {16072, 2, {0x8, 0x1C, 0xFDE0, 0x1100, 0x1100, 0x11FE, 0x7D00, 0x1100, 0x117C, 0x1144, 0x1144, 0x1D44, 0xE144, 0x427C, 0x244, 0x400}},
    {16073, 2, {0x20, 0x20, 0xF920, 0x213C, 0x2120, 0x2120, 0x27FE, 0xF800, 0x2020, 0x2124, 0x2124, 0x3A28, 0xE410, 0x4020, 0xC0, 0x700}},
    {16074, 2, {0x80, 0x80, 0xFCF8, 0x1108, 0x1210, 0x11FC, 0x1124, 0x7D24, 0x1124, 0x11FC, 0x1050, 0x1050, 0x1C90, 0xE092, 0x4112, 0x20E}},
    {16075, 2, {0x20, 0x20, 0xFDFE, 0x1020, 0x1020, 0x11FC, 0x7D24, 0x1124, 0x11FC, 0x1020, 0x1070, 0x1CA8, 0xE124, 0x4222, 0x20, 0x20}},
    {16076, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x12FA, 0x7C20, 0x1020, 0x13FE, 0x1020, 0x1128, 0x1D24, 0xE222, 0x4422, 0xA0, 0x40}},
    {16077, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x1000, 0x7C00, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1C20, 0xE050, 0x4088, 0x104, 0x202}},
    {16078, 2, {0x100, 0x100, 0x7FFC, 0x100, 0xFD7C, 0x1110, 0x1110, 0x7D7C, 0x1110, 0x1D10, 0xE2FE, 0x4280, 0x440, 0x820, 0x3018, 0xC006}},
    {16079, 2, {0x0, 0x1FC, 0xFC04, 0x1004, 0x10FC, 0x1004, 0x7C04, 0x11FC, 0x1020, 0x10A4, 0x10A4, 0x1D28, 0xE050, 0x4088, 0x104, 0x202}},
    {16080, 2, {0x40, 0x40, 0xFC88, 0x1104, 0x13FE, 0x1002, 0x1088, 0x7D44, 0x1242, 0x10F8, 0x1188, 0x1250, 0x1C20, 0xE050, 0x4188, 0x606}},
    {16081, 2, {0x1C, 0xE0, 0xFC20, 0x11FE, 0x1070, 0x10A8, 0x7D24, 0x1222, 0x10F8, 0x1010, 0x1020, 0x1DFE, 0xE020, 0x4020, 0xA0, 0x40}},
    {16082, 2, {0x0, 0x3FE, 0xFA02, 0x221A, 0x22E2, 0x2222, 0x2222, 0xFBFE, 0x2222, 0x2272, 0x22AA, 0x3B26, 0xE222, 0x4222, 0x3FE, 0x202}},
    {16083, 2, {0x0, 0x7E0, 0xFA5E, 0x2252, 0x2252, 0x23D2, 0x2252, 0xFA54, 0x23D4, 0x2254, 0x2248, 0x3AE8, 0xE754, 0x4054, 0x62, 0x40}},
    {16084, 2, {0x8, 0x3C, 0xFDE0, 0x1020, 0x13FE, 0x10A8, 0x10A8, 0x7CA8, 0x13FE, 0x10A8, 0x10A8, 0x13FE, 0x1C20, 0xE020, 0x41FC, 0x0}},
    {16085, 2, {0x0, 0x3FE, 0xFA02, 0x2202, 0x23FE, 0x2210, 0x2292, 0xFA92, 0x2292, 0x22FE, 0x2210, 0x3A92, 0xE292, 0x4492, 0x4FE, 0x802}},
    {16086, 2, {0x40, 0x40, 0xFBF8, 0x2048, 0x2048, 0x27FE, 0x2048, 0xF848, 0x23F8, 0x2444, 0x22E8, 0x3950, 0xE248, 0x4446, 0x140, 0x80}},
    {16087, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16088, 2, {0x40, 0x20, 0xFBFE, 0x2222, 0x2020, 0x21FC, 0x2124, 0xF924, 0x21FC, 0x2124, 0x2124, 0x39FC, 0xE124, 0x4020, 0x20, 0x20}},
    {16089, 2, {0x40, 0x20, 0xFBFE, 0x2080, 0x2104, 0x23FE, 0xF802, 0x21FC, 0x2104, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x4104, 0x114, 0x108}},
    {16090, 2, {0x0, 0x3DE, 0xFA52, 0x2252, 0x2252, 0x23DE, 0x2252, 0xFA52, 0x2252, 0x23DE, 0x2252, 0x3A52, 0xE252, 0x4252, 0x552, 0x8A6}},
    {16091, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x11F8, 0x1108, 0x7D08, 0x11F8, 0x1080, 0x11FC, 0x1254, 0x1C94, 0xE124, 0x4244, 0x94, 0x108}},
    {16092, 2, {0x88, 0x88, 0xF888, 0x2088, 0x2154, 0x2222, 0x2442, 0xF820, 0x2020, 0x2120, 0x213C, 0x3920, 0xE2A0, 0x4260, 0x43E, 0x800}},
    {16093, 2, {0x82, 0x82, 0xFBE2, 0x2082, 0x208A, 0x23EA, 0x22AA, 0xFAAA, 0x23EA, 0x208A, 0x21CA, 0x3AAA, 0xE4A2, 0x4082, 0x8A, 0x84}},
    {16094, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2202, 0x23FE, 0x2200, 0xFA00, 0x23FE, 0x2352, 0x2552, 0x3DFE, 0xE552, 0x4552, 0x952, 0x106}},
    {16095, 2, {0x40, 0x40, 0xFBFE, 0x2080, 0x21FC, 0x2220, 0x25FE, 0xF800, 0x21FC, 0x2104, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x114, 0x108}},
    {16096, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2088, 0x2104, 0xFA22, 0x2028, 0x2024, 0x23FE, 0x2020, 0x3850, 0xE050, 0x4088, 0x104, 0x202}},
    {16097, 2, {0x48, 0x148, 0xFD48, 0x13FE, 0x1148, 0x1148, 0x7D78, 0x1100, 0x11FE, 0x1020, 0x13FE, 0x1C70, 0xE0A8, 0x4124, 0x622, 0x20}},
    {16098, 2, {0x210, 0x110, 0xF910, 0x27BE, 0x2228, 0x2248, 0x23AA, 0xFAAE, 0x22BA, 0x22EA, 0x22AE, 0x3AA8, 0xE4AA, 0x44A2, 0x9A2, 0x101E}},
    {16099, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20F8, 0x2088, 0x20F8, 0xF888, 0x2088, 0x23FE, 0x2100, 0x3948, 0xE184, 0x4100, 0x1FE, 0x0}},
    {16100, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x2100, 0x21FC, 0x2104, 0xFA04, 0x21F4, 0x2114, 0x2114, 0x39F4, 0xE114, 0x4004, 0x28, 0x10}},
    {16101, 2, {0x88, 0x48, 0xFC50, 0x13FE, 0x1050, 0x1050, 0x7DFC, 0x1154, 0x1154, 0x118C, 0x1104, 0x1DFC, 0xE104, 0x4104, 0x1FC, 0x104}},
    {16102, 2, {0x1FC, 0x124, 0xFD24, 0x11FC, 0x1124, 0x1124, 0x7DFC, 0x1020, 0x1020, 0x13FE, 0x1222, 0x1E2A, 0xE2FA, 0x420A, 0x202, 0x206}},
    {16103, 2, {0x104, 0x104, 0xF9DE, 0x2244, 0x2554, 0x209E, 0x2104, 0xFA24, 0x2424, 0x23FE, 0x2070, 0x38A8, 0xE124, 0x4622, 0x20, 0x20}},
    {16104, 2, {0x0, 0x7EFC, 0x4844, 0x7E54, 0x4328, 0x7E90, 0x4828, 0x7EC6, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16105, 2, {0x110, 0xA0, 0xF800, 0x27FE, 0x20A0, 0x23F8, 0xF8A8, 0x27FE, 0x20A8, 0x23F8, 0x20A0, 0x39B0, 0xE2A8, 0x4CA6, 0xA0, 0xA0}},
    {16106, 2, {0xA0, 0x90, 0xFDFE, 0x1320, 0x15FC, 0x1120, 0x7DFC, 0x1120, 0x11FE, 0x1100, 0x13FC, 0x1C88, 0xE09E, 0x4102, 0x20A, 0x404}},
    {16107, 2, {0x0, 0x1FC, 0xFD24, 0x11FC, 0x1124, 0x11FC, 0x7CA8, 0x10A8, 0x13FE, 0x10A8, 0x10A8, 0x1DFC, 0xE020, 0x43FE, 0x20, 0x20}},
    {16108, 2, {0x1FC, 0x20, 0xFBFE, 0x2222, 0x21AC, 0x2020, 0xF9AC, 0x2000, 0x21FC, 0x2000, 0x23FE, 0x3880, 0xE0FC, 0x4004, 0x28, 0x10}},
    {16109, 2, {0x40, 0x20, 0xFBFE, 0x2020, 0x2048, 0x22F2, 0xF924, 0x2252, 0x20F8, 0x2008, 0x2020, 0x3BFE, 0xE020, 0x4020, 0x20, 0x20}},
    {16110, 2, {0x100, 0x110, 0xFBDC, 0x2254, 0x2554, 0x2288, 0xF908, 0x22F4, 0x2402, 0x2000, 0x23FC, 0x3840, 0xE150, 0x4248, 0x544, 0x80}},
    {16111, 2, {0x40, 0x7C, 0xF840, 0x23FE, 0x2242, 0x2278, 0xFBC4, 0x223C, 0x2208, 0x22F0, 0x2324, 0x3AA8, 0xE5FE, 0x4420, 0x8A0, 0x40}},
    {16112, 2, {0x88, 0x3DE, 0xFC88, 0x11DC, 0x1088, 0x13DE, 0x7C88, 0x1000, 0x13FC, 0x1004, 0x1004, 0x1DFC, 0xE004, 0x4004, 0x3FC, 0x4}},
    {16113, 2, {0x20, 0x47C, 0xFA84, 0x2148, 0x2030, 0x20CE, 0x2610, 0xFA7C, 0x2210, 0x227C, 0x2210, 0x3AFE, 0xE210, 0x4210, 0x5FE, 0x800}},
    {16114, 2, {0x20, 0x1FC, 0xFD24, 0x13FE, 0x1124, 0x11FC, 0x7C20, 0x11FC, 0x1124, 0x11FC, 0x1040, 0x1DFE, 0xE088, 0x41D0, 0x70, 0x38C}},
    {16115, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x2042, 0x20E2, 0xFB82, 0x208A, 0x208A, 0x27EA, 0x218A, 0x39CA, 0xE2AA, 0x4482, 0x8A, 0x84}},
    {16116, 2, {0x0, 0x1F8, 0xFD08, 0x11F8, 0x1108, 0x11F8, 0x7C00, 0x13FC, 0x1294, 0x13FC, 0x1000, 0x1DF8, 0xE090, 0x4060, 0x198, 0x606}},
    {16117, 2, {0x3FE, 0x202, 0xFBFE, 0x2200, 0x22FC, 0x2208, 0x23FE, 0xFA10, 0x2230, 0x23CE, 0x2242, 0x3A84, 0xE4EE, 0x4784, 0x884, 0x18C}},
    {16118, 2, {0x108, 0x1EE, 0xFA94, 0x2042, 0x2020, 0x21FC, 0x2024, 0xFBFE, 0x2024, 0x21FC, 0x2020, 0x39FE, 0xE020, 0x43FE, 0x20, 0x20}},
    {16119, 2, {0x208, 0x208, 0xEFBE, 0x4208, 0x4208, 0x4FBE, 0x4AAA, 0xEAAA, 0x4AAA, 0x4BAE, 0x4208, 0x671C, 0xCAAA, 0x1248, 0x208, 0x208}},
    {16120, 2, {0x0, 0x7BC, 0xF908, 0x2528, 0x27BC, 0x2318, 0x25AA, 0xF946, 0x2000, 0x23F8, 0x2208, 0x3A08, 0xE3F8, 0x4208, 0x208, 0x3F8}},
    {16121, 2, {0xFDFC, 0x1040, 0x21A2, 0x7C54, 0xA5B8, 0x2454, 0x3D92, 0x2460, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16122, 2, {0x3FFE, 0x2104, 0x2E38, 0x2210, 0x2FBC, 0x2718, 0x2AB4, 0x3252, 0x2000, 0x2FFC, 0x2080, 0x27F8, 0x20A0, 0x4090, 0x5FFE, 0x8000}},
    {16123, 2, {0x108, 0x88, 0xFBC8, 0x2010, 0x23DE, 0x2264, 0x23D4, 0xF814, 0x23D4, 0x2054, 0x2094, 0x38E8, 0xE388, 0x4094, 0x294, 0x122}},
    {16124, 2, {0x0, 0x3FE, 0xFC50, 0x13FE, 0x1252, 0x13FE, 0x7C00, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE020, 0x43FE, 0x20, 0x20}},
    {16125, 2, {0xA0, 0x90, 0xFDFE, 0x1320, 0x11FC, 0x1120, 0x7DFC, 0x1120, 0x11FE, 0x1100, 0x13DE, 0x1E52, 0xE252, 0x4272, 0x202, 0x206}},
    {16126, 2, {0x100, 0xFFFE, 0x1048, 0x24FE, 0x7990, 0x10FC, 0x2290, 0x7CFC, 0x890, 0x30FE, 0xC080, 0x7FFC, 0x100, 0x3FF8, 0x120, 0xFFFE}},
    {16127, 2, {0x0, 0x3F8, 0xFAA8, 0x22A8, 0x23F8, 0x2100, 0xFBFC, 0x2484, 0x23E4, 0x22A4, 0x22A4, 0x3BE4, 0xE094, 0x47F4, 0x14, 0x8}},
    {16128, 2, {0x10, 0x410, 0xFA7C, 0x2010, 0x20FE, 0x2044, 0x2628, 0xFAFE, 0x2210, 0x227C, 0x2210, 0x3AFE, 0xE210, 0x4210, 0x5FE, 0x800}},
    {16129, 2, {0x104, 0x88, 0xFBFE, 0x2020, 0x21FC, 0x2020, 0xFBFE, 0x2054, 0x2192, 0x2090, 0x23FE, 0x3890, 0xE0D4, 0x438A, 0x96, 0x1A2}},
    {16130, 2, {0xA00, 0x33B8, 0x2208, 0x3BB8, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x8002, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {16131, 2, {0x124, 0x248, 0xF924, 0x2000, 0x23FC, 0x2294, 0x2264, 0xFA94, 0x23FC, 0x2248, 0x2368, 0x3A48, 0xE36A, 0x424A, 0x246, 0x362}},
    {16132, 2, {0x124, 0x124, 0xFAAA, 0x23AE, 0x2124, 0x22AA, 0xFBAE, 0x2024, 0x27FE, 0x2010, 0x27D4, 0x3D54, 0xE7CA, 0x454A, 0x7D6, 0x22}},
    {16133, 2, {0x20, 0x10, 0xFBFE, 0x2244, 0x22FE, 0x2244, 0x23FE, 0xFA10, 0x22FE, 0x2292, 0x22FE, 0x3A92, 0xE2FE, 0x4200, 0x244, 0x482}},
    {16134, 2, {0x88, 0x2AA, 0xFADC, 0x2488, 0x2154, 0x2222, 0x27FE, 0xFC02, 0x21F8, 0x2040, 0x2040, 0x39F8, 0xE050, 0x4048, 0x3FE, 0x0}},
    {16135, 2, {0x20, 0x222, 0xFBFE, 0x2090, 0x21FE, 0x2310, 0x25FE, 0xF910, 0x21FE, 0x2110, 0x21FE, 0x2100, 0x3BDE, 0xE252, 0x272, 0x206}},
    {16136, 2, {0x0, 0x7FC, 0xF4A4, 0x24A4, 0x27FC, 0x2210, 0x22A8, 0xF4BE, 0x2F68, 0x223C, 0x24A8, 0x2FBC, 0x3028, 0xCAA8, 0xABE, 0x20}},
    {16137, 2, {0x210, 0x710, 0x3810, 0x2A7C, 0x2A54, 0x2A54, 0x2A10, 0x2A28, 0x2A28, 0x2948, 0x294A, 0x2886, 0x2840, 0x4A30, 0x4D0E, 0x8900}},
    {16138, 2, {0x1008, 0x101C, 0x10E8, 0xFEA8, 0x10A8, 0x54A8, 0x54A8, 0x54A8, 0x54A8, 0x7CA4, 0x10A4, 0x10A4, 0x14A2, 0x1928, 0x1134, 0x224}},
    {16139, 2, {0x1008, 0x101C, 0x10E8, 0x10A8, 0xFEA8, 0x10A8, 0x10A8, 0x10A8, 0x7CA8, 0x44A4, 0x44A4, 0x44A4, 0x44A2, 0x7D28, 0x4534, 0x224}},
    {16140, 2, {0x808, 0x1C1C, 0x70E8, 0x50A8, 0x50A8, 0x50A8, 0x58A8, 0x54A8, 0x54A8, 0x54A4, 0x50A4, 0x50A4, 0x58A2, 0x5528, 0x9D34, 0x624}},
    {16141, 2, {0x1008, 0x101C, 0x10E8, 0xFEA8, 0x10A8, 0x7CA8, 0x10A8, 0xFEA8, 0x30A8, 0x38A4, 0x54A4, 0x54A4, 0x90A2, 0x1128, 0x1134, 0x1224}},
    {16142, 2, {0x220, 0x710, 0x3810, 0x2AFE, 0x2A00, 0x2A44, 0x2A82, 0x2A28, 0x2A10, 0x2928, 0x2944, 0x2884, 0x2840, 0x4A30, 0x4D0E, 0x8900}},
    {16143, 2, {0x408, 0x1E1C, 0xE0E8, 0x22A8, 0x92A8, 0x54A8, 0x48A8, 0x10A8, 0xFCA8, 0x24A4, 0x44A4, 0x28A4, 0x10A2, 0x2928, 0x4534, 0x8224}},
    {16144, 2, {0x804, 0x40E, 0x7F74, 0x4154, 0x4154, 0x7F54, 0x4054, 0x7F54, 0x5554, 0x5552, 0x7F52, 0x5552, 0x5550, 0x5594, 0x819A, 0x312}},
    {16145, 2, {0x1008, 0x101C, 0xFEE8, 0x10A8, 0xFEA8, 0x92A8, 0xD6A8, 0xBAA8, 0x92A8, 0xFEA4, 0x10A4, 0x38A4, 0x54A2, 0x9328, 0x1134, 0x1224}},
    {16146, 2, {0x27C, 0x744, 0x387C, 0x2A44, 0x2A7C, 0x2A00, 0x2AFE, 0x2A40, 0x2A7E, 0x29AA, 0x2952, 0x28AA, 0x2844, 0x4A30, 0x4D0E, 0x8900}},
    {16147, 2, {0x888, 0x1C50, 0x7000, 0x53FE, 0x5050, 0x51FC, 0x5854, 0x57FE, 0x5454, 0x55FC, 0x5050, 0x50D8, 0x5954, 0x5652, 0x9C50, 0x450}},
    {16148, 2, {0x8, 0x1C, 0x7CE8, 0x44A8, 0x7CA8, 0x44A8, 0x7CA8, 0xA8, 0x7CA8, 0x54A4, 0x54A4, 0x54A4, 0x56A2, 0xF928, 0x134, 0x224}},
    {16149, 2, {0x1008, 0x921C, 0x54E8, 0xFEA8, 0x82A8, 0x7CA8, 0x44A8, 0x7CA8, 0xA8, 0x7CA4, 0x54A4, 0x7CA4, 0x54A2, 0x5528, 0x7D34, 0x4624}},
    {16150, 2, {0x11F0, 0x3A10, 0xE3E0, 0xA020, 0xA7FC, 0xA188, 0xB650, 0xA9A8, 0xAE64, 0xA9A0, 0xA248, 0xAFBE, 0xBAAA, 0xAFBE, 0x3A8A, 0xFBE}},
    {16151, 2, {0x0, 0x1FC, 0xFE80, 0x1080, 0x1080, 0x10F8, 0x1088, 0x1088, 0x1148, 0x1128, 0x1128, 0x1108, 0x1108, 0x114A, 0x518A, 0x2106}},
    {16152, 2, {0x1000, 0x11FC, 0x1080, 0x1080, 0x5480, 0x54F8, 0x5488, 0x5488, 0x5548, 0x5528, 0x5528, 0x5D08, 0x6508, 0x14A, 0x18A, 0x106}},
    {16153, 2, {0x1000, 0x10FC, 0x1040, 0xFE40, 0x9240, 0x9278, 0x9248, 0x9248, 0x92A8, 0x9298, 0x9A88, 0x9488, 0x10AA, 0x10CA, 0x108A, 0x1006}},
    {16154, 2, {0x1000, 0x8FC, 0x840, 0xFF40, 0x40, 0x78, 0x3C48, 0x2448, 0x24A8, 0x2498, 0x2488, 0x2588, 0x46AA, 0x44CA, 0x808A, 0x6}},
    {16155, 2, {0x0, 0x7EFC, 0x2440, 0x2440, 0x2440, 0x2478, 0x2448, 0xFF48, 0x24A8, 0x2498, 0x2488, 0x2488, 0x24AA, 0x24CA, 0x448A, 0x8406}},
    {16156, 2, {0x100, 0x1FF0, 0x1110, 0x1110, 0xFFFE, 0x280, 0x1C70, 0xE00E, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16157, 2, {0x1000, 0x1EF8, 0x2288, 0x62A8, 0x9492, 0x882, 0x307E, 0xC000, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16158, 2, {0x0, 0x7EFC, 0x840, 0x840, 0x4A40, 0x2A78, 0x2C48, 0x848, 0xFFA8, 0x898, 0x888, 0x888, 0x8AA, 0x8CA, 0x88A, 0x806}},
    {16159, 2, {0x1000, 0x10FC, 0x2040, 0x7E40, 0x4240, 0x4278, 0x4248, 0x4248, 0x7EA8, 0x4298, 0x4288, 0x4288, 0x42AA, 0x7ECA, 0x428A, 0x6}},
    {16160, 2, {0x1000, 0x8FC, 0x840, 0x7F40, 0x4140, 0x8278, 0x2048, 0x2048, 0x3EA8, 0x2098, 0x2088, 0x2088, 0x26AA, 0x38CA, 0x208A, 0x6}},
    {16161, 2, {0x800, 0x28FC, 0x2840, 0x3E40, 0x4840, 0x8878, 0x848, 0x7F48, 0x18A8, 0x1C98, 0x2A88, 0x2A88, 0x48AA, 0x88CA, 0x88A, 0x806}},
    {16162, 2, {0x0, 0xFEFC, 0x9040, 0x9040, 0x9040, 0xBC78, 0xA448, 0xA448, 0xA4A8, 0xA498, 0xBC88, 0x9088, 0x90AA, 0x90CA, 0xFE8A, 0x6}},
    {16163, 2, {0x0, 0xFEFC, 0x1040, 0x1040, 0x2040, 0x7C78, 0x4448, 0x4448, 0x44A8, 0x7C98, 0x4488, 0x4488, 0x44AA, 0x7CCA, 0x448A, 0x6}},
    {16164, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16165, 2, {0x1000, 0x12FC, 0x7A40, 0x1440, 0x1440, 0xFF78, 0x1048, 0x2048, 0x7EA8, 0x9098, 0x2088, 0x3C88, 0x4AA, 0x4CA, 0x148A, 0x806}},
    {16166, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4400, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16167, 2, {0x0, 0xFCFC, 0x440, 0x6840, 0x1040, 0xFE78, 0x9248, 0x9248, 0xFEA8, 0x9298, 0x9288, 0xFE88, 0x92AA, 0x92CA, 0x928A, 0x8606}},
    {16168, 2, {0x800, 0x8FC, 0x1440, 0x2240, 0x5140, 0x8878, 0x7E48, 0x248, 0x4A8, 0x898, 0x7E88, 0x4288, 0x42AA, 0x7ECA, 0x428A, 0x6}},
    {16169, 2, {0x400, 0x57C, 0x4A0, 0x7F20, 0x4420, 0x4438, 0x7428, 0x5528, 0x5558, 0x5648, 0x5648, 0x5448, 0x655A, 0x4D6A, 0x934A, 0x106}},
    {16170, 2, {0x800, 0x8FC, 0x7F40, 0x840, 0x840, 0x2A78, 0x2A48, 0x2A48, 0x5DA8, 0x8898, 0x888, 0x1488, 0x14AA, 0x22CA, 0x428A, 0x8006}},
    {16171, 2, {0x0, 0x7EFC, 0x4A40, 0x4A40, 0x7E40, 0x4A78, 0x4A48, 0x7E48, 0x8A8, 0xFF98, 0x1888, 0x2C88, 0x2AAA, 0x4ACA, 0x888A, 0x806}},
    {16172, 2, {0x1000, 0x10FC, 0x7C40, 0x1040, 0x1040, 0xFE78, 0x4448, 0x2848, 0x7CA8, 0x1098, 0x1088, 0xFE88, 0x10AA, 0x10CA, 0x108A, 0x1006}},
    {16173, 2, {0x1020, 0x103E, 0x7C20, 0x11FC, 0x1104, 0x1D04, 0xE1FC, 0x104, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16174, 2, {0x1000, 0x10FC, 0xFE40, 0x1040, 0x7C40, 0x1478, 0xFE48, 0x1448, 0x7CA8, 0x1098, 0xFC88, 0x2488, 0x44AA, 0x28CA, 0x308A, 0xC806}},
    {16175, 2, {0x800, 0x87C, 0xFFA0, 0x820, 0x820, 0x7F38, 0x4928, 0x7F28, 0x4958, 0x7F48, 0x848, 0x1C48, 0x2A5A, 0x496A, 0x884A, 0x806}},
    {16176, 2, {0x800, 0x107C, 0x7F20, 0x4920, 0x4920, 0x7F38, 0x4928, 0x4928, 0x7F58, 0x1448, 0x2448, 0xFF48, 0x45A, 0x46A, 0x44A, 0x406}},
    {16177, 2, {0x444, 0x7C58, 0x462, 0x1C42, 0xE43E, 0x0, 0x7FFC, 0x4824, 0x1010, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16178, 2, {0x0, 0xFEFC, 0x1040, 0x2040, 0xFE40, 0xAA78, 0xAA48, 0xAA48, 0xA2A8, 0x9698, 0x1088, 0xFE88, 0x10AA, 0x28CA, 0x448A, 0x8206}},
    {16179, 2, {0x1000, 0x10FC, 0x5240, 0x5140, 0x9540, 0x878, 0x3048, 0xC048, 0x7EA8, 0x4298, 0x7E88, 0x4288, 0x7EAA, 0x42CA, 0x7E8A, 0x4206}},
    {16180, 2, {0x0, 0x7CFC, 0x5440, 0x5440, 0x7C40, 0x5478, 0x5448, 0x7C48, 0x10A8, 0xFE98, 0x9288, 0x9688, 0xBEAA, 0x82CA, 0x8A8A, 0x8406}},
    {16181, 2, {0x0, 0x3EFC, 0x2220, 0x3E20, 0x2238, 0x3E28, 0x48, 0xFF68, 0x858, 0x84A, 0x2F4A, 0x286A, 0x2846, 0x5800, 0x4FFE, 0x8000}},
    {16182, 2, {0x1000, 0x10FC, 0x2840, 0x4440, 0x8240, 0x7C78, 0x48, 0xE248, 0xAAA8, 0xAA98, 0xEA88, 0xAA88, 0xAAAA, 0xE2CA, 0xAA8A, 0xA406}},
    {16183, 2, {0x1000, 0x92FC, 0x9240, 0x9240, 0xFE40, 0x78, 0xFE48, 0x1048, 0x20A8, 0xFE98, 0xAA88, 0xAA88, 0xAAAA, 0xAACA, 0xAA8A, 0x8606}},
    {16184, 2, {0x1000, 0x87C, 0xFFA0, 0x80A0, 0x1420, 0x2238, 0x4928, 0x1428, 0x2258, 0x4148, 0xBEC8, 0x2248, 0x225A, 0x226A, 0x3E4A, 0x2206}},
    {16185, 2, {0x4400, 0x44FC, 0xFF40, 0x5440, 0x840, 0xFF78, 0x4048, 0x4048, 0x7EA8, 0x98, 0x5488, 0x5488, 0x55AA, 0x56CA, 0x948A, 0x6}},
    {16186, 2, {0x1000, 0x8FC, 0x7F40, 0x4840, 0x5E40, 0x4A78, 0x7F48, 0x4A48, 0x5EA8, 0x4898, 0x5E88, 0x5288, 0x52AA, 0x52CA, 0x5E8A, 0x9206}},
    {16187, 2, {0x1110, 0x2110, 0xCAA8, 0x1444, 0x3240, 0xD27C, 0x1540, 0x14FE, 0x800, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16188, 2, {0x0, 0xFEFC, 0x2840, 0xFE40, 0xAA40, 0xAA78, 0xFE48, 0x48, 0x7CA8, 0x98, 0xFE88, 0x1088, 0x54AA, 0x92CA, 0x508A, 0x2006}},
    {16189, 2, {0x1000, 0x10FC, 0xFE40, 0x1040, 0xBA40, 0x5478, 0xBA48, 0x1048, 0xBAA8, 0x5498, 0xBA88, 0x1088, 0x28AA, 0x24CA, 0x428A, 0x8006}},
    {16190, 2, {0x800, 0x47C, 0x7FA0, 0x5220, 0x5220, 0x7FB8, 0x52A8, 0x52A8, 0x7FD8, 0x5248, 0x5248, 0x5BC8, 0x525A, 0x52EA, 0x9B4A, 0x1206}},
    {16191, 2, {0x1000, 0x87C, 0x7F20, 0x2220, 0xFF20, 0x38, 0x7F28, 0x4928, 0x7F58, 0x4948, 0x7F48, 0x848, 0x7F5A, 0x86A, 0xF4A, 0xF006}},
    {16192, 2, {0x2200, 0x147C, 0xFF20, 0x1420, 0x7F20, 0x5538, 0x6328, 0x5D28, 0x4158, 0x7F48, 0x848, 0xFF48, 0x145A, 0x226A, 0x414A, 0x8006}},
    {16193, 2, {0x3FFE, 0x2104, 0x2E38, 0x2210, 0x2FBC, 0x2718, 0x2AB4, 0x3252, 0x2000, 0x3FFE, 0x2400, 0x27F0, 0x2490, 0x4852, 0x4B12, 0x8C0E}},
    {16194, 2, {0x800, 0x147C, 0x2220, 0x5D20, 0x80A0, 0x3E38, 0x2228, 0x3E28, 0x58, 0x7748, 0x1148, 0x5548, 0x335A, 0x556A, 0x114A, 0x3306}},
    {16195, 2, {0xE00, 0xF0FC, 0x9240, 0x5440, 0xFE40, 0x3878, 0x5448, 0x8248, 0x7CA8, 0x5498, 0x5488, 0x7C88, 0x54AA, 0x54CA, 0x7C8A, 0x4406}},
    {16196, 2, {0x240C, 0xFF70, 0x2440, 0x3C40, 0x247E, 0x3C48, 0x2448, 0xFF48, 0x2448, 0x4288, 0x0, 0x7FFC, 0x800, 0xFE0, 0x1124, 0x1C9C}},
    {16197, 2, {0xA00, 0x74FC, 0x1540, 0x5240, 0x2240, 0x2178, 0x5D48, 0x8048, 0x7EA8, 0x4298, 0x4288, 0x7E88, 0x42AA, 0x24CA, 0xF8A, 0xF006}},
    {16198, 2, {0x1000, 0x92FC, 0x5440, 0xFE40, 0x8240, 0x7C78, 0x4448, 0x7C48, 0xA8, 0x7C98, 0x5488, 0x7C88, 0x54AA, 0x54CA, 0x7C8A, 0x4406}},
    {16199, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0xBFF4, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16200, 2, {0x48F8, 0x7050, 0x4420, 0x3DFC, 0x2024, 0x7CB8, 0x90A0, 0x7D7E, 0x2A00, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16201, 2, {0x0, 0x7E40, 0x487E, 0x7E90, 0x4308, 0x7EFE, 0x48AA, 0x7EFE, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {16202, 2, {0x4900, 0x5D7C, 0x6120, 0xDDA0, 0x4320, 0x5D38, 0x4128, 0x5D28, 0xB6D8, 0x9C48, 0x48, 0x7F48, 0x25A, 0x346A, 0x1C4A, 0x6306}},
    {16203, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x0, 0x7FFC, 0x800, 0xFE0, 0x1124, 0x1C9C}},
    {16204, 2, {0x7F00, 0x87C, 0xFFA0, 0x88A0, 0x6B20, 0x1038, 0x6F28, 0x4528, 0x5558, 0x6B48, 0x1048, 0x7F48, 0x495A, 0x7F6A, 0x494A, 0x7F06}},
    {16205, 2, {0x2200, 0x22FC, 0x2240, 0x2240, 0xFF40, 0x2278, 0x2248, 0x2248, 0x3EA8, 0x2298, 0x2288, 0x2288, 0x22AA, 0x3ECA, 0x228A, 0x6}},
    {16206, 2, {0x2210, 0x2210, 0xFF10, 0x2210, 0x3EFE, 0x2292, 0x3E92, 0x2210, 0x2228, 0xFF28, 0x4028, 0x5428, 0x6248, 0x404A, 0x7E8A, 0x106}},
    {16207, 2, {0x2040, 0x10A0, 0x8318, 0x4DF6, 0x1000, 0xE3F8, 0x2208, 0x23F8, 0x0, 0x1010, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16208, 2, {0x4888, 0x4850, 0x4800, 0x49FE, 0xFC50, 0x49FC, 0x4854, 0x4BFE, 0x4854, 0x79FC, 0x4850, 0x48D8, 0x4954, 0x7A52, 0x4850, 0x50}},
    {16209, 2, {0x4BFE, 0x4850, 0x4BFE, 0x4A52, 0xFE52, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20}},
    {16210, 2, {0x10, 0x7C10, 0x5490, 0x5490, 0x7CFC, 0x5490, 0x5510, 0x7C10, 0x1010, 0xFEFC, 0x1210, 0x1210, 0x2210, 0x2210, 0x4AFE, 0x8400}},
    {16211, 2, {0x100, 0x1100, 0x1FF8, 0x2100, 0x5FF0, 0x100, 0xFFFE, 0x810, 0x4850, 0x7E7C, 0x4890, 0x8810, 0x7E7C, 0x810, 0xF10, 0xF0FE}},
    {16212, 2, {0x2810, 0x2810, 0xAA50, 0x6C7C, 0x2850, 0xFE90, 0x1010, 0x307C, 0x4A10, 0x9C10, 0x2C10, 0xDAFE, 0x2900, 0xC8C0, 0x283E, 0x1000}},
    {16213, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FF8, 0x8, 0x1008, 0x1FFE, 0x2, 0x2, 0x14, 0x8}},
    {16214, 2, {0x40, 0x40, 0x7C40, 0x5440, 0x55FC, 0x5444, 0x5444, 0x7C44, 0x5444, 0x5484, 0x5484, 0x5484, 0x7D04, 0x4504, 0x228, 0x410}},
    {16215, 2, {0x20, 0x20, 0x7C20, 0x5420, 0x5420, 0x5420, 0x5420, 0x7C20, 0x5450, 0x5450, 0x5450, 0x5488, 0x7C88, 0x4504, 0x204, 0x402}},
    {16216, 2, {0x40, 0x20, 0x7C00, 0x55FC, 0x5400, 0x5400, 0x54F0, 0x7C90, 0x5490, 0x5490, 0x5490, 0x5492, 0x7C92, 0x4512, 0x10E, 0x200}},
    {16217, 2, {0x40, 0x40, 0x7C80, 0x5488, 0x5504, 0x57FE, 0x5402, 0x7C90, 0x5490, 0x5490, 0x5490, 0x5490, 0x7D12, 0x4512, 0x20E, 0x400}},
    {16218, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1100, 0x3FF8, 0x4100, 0x100, 0xFFFE, 0x100, 0x100}},
    {16219, 2, {0x80, 0x80, 0x7D00, 0x55FC, 0x5604, 0x5404, 0x55E4, 0x7D24, 0x5524, 0x5524, 0x55E4, 0x5524, 0x7C04, 0x4404, 0x28, 0x10}},
    {16220, 2, {0x1000, 0x1EF8, 0x2288, 0x62A8, 0x9492, 0x882, 0x307E, 0xC000, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {16221, 2, {0x7F04, 0x844, 0x1E44, 0x2244, 0x5444, 0x844, 0x300C, 0xC000, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {16222, 2, {0x0, 0x3EF8, 0x2288, 0x2288, 0x2288, 0x3EF8, 0x0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {16223, 2, {0x4, 0x1E, 0x7DE0, 0x5422, 0x5512, 0x5494, 0x5480, 0x7C08, 0x55FE, 0x5408, 0x5488, 0x5448, 0x7C48, 0x4408, 0x28, 0x10}},
    {16224, 2, {0x20, 0x20, 0x7C20, 0x57FE, 0x5420, 0x5420, 0x5524, 0x7D24, 0x5524, 0x56AA, 0x5420, 0x5450, 0x7C50, 0x4488, 0x104, 0x202}},
    {16225, 2, {0x40, 0x80, 0xF9FC, 0xA924, 0xA924, 0xA9FC, 0xF924, 0xA944, 0xA9FC, 0xA890, 0xA910, 0xFBFE, 0x8810, 0x10, 0x10, 0x10}},
    {16226, 2, {0x40, 0x80, 0xFBFC, 0xA910, 0xAA48, 0xAC46, 0xFBF8, 0xAA48, 0xAA48, 0xABF8, 0xAA48, 0xFA48, 0x8BF8, 0x42, 0x42, 0x3E}},
    {16227, 2, {0x14, 0x12, 0xF810, 0xAFFE, 0xA810, 0xA810, 0xFBD2, 0xAA52, 0xAA52, 0xAA54, 0xABD4, 0xF808, 0x88EA, 0x71A, 0x226, 0x42}},
    {16228, 2, {0x80, 0x80, 0x7CFE, 0x5502, 0x5622, 0x54AA, 0x5472, 0x7C22, 0x55FE, 0x5422, 0x5472, 0x54AA, 0x7D22, 0x4422, 0xA, 0x4}},
    {16229, 2, {0x0, 0xFC, 0x7C48, 0x5430, 0x55FE, 0x5452, 0x5494, 0x7D10, 0x5630, 0x5420, 0x55FE, 0x5470, 0x7CA8, 0x4524, 0x222, 0x20}},
    {16230, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {16231, 2, {0x84, 0x44, 0xF848, 0xABFE, 0xA884, 0xA884, 0xF908, 0xA94A, 0xAA52, 0xAB9C, 0xA884, 0xF908, 0x8908, 0x252, 0x3DE, 0x42}},
    {16232, 2, {0x108, 0x88, 0xF890, 0xABFC, 0xA840, 0xA9F8, 0xF840, 0xABFE, 0xA880, 0xA900, 0xA9FC, 0xFA20, 0x8C20, 0x820, 0x3FE, 0x0}},
    {16233, 2, {0x40, 0x88, 0xF9FC, 0xA908, 0xAA52, 0xABFE, 0xF850, 0xA988, 0xAE26, 0xA8C0, 0xAB10, 0xF864, 0x8B88, 0x30, 0xC0, 0x700}},
    {16234, 2, {0x104, 0x88, 0xF800, 0xABFE, 0xAA22, 0xAAAA, 0xFA72, 0xAA22, 0xABFE, 0xA800, 0xA9FC, 0xF904, 0x89FC, 0x104, 0x1FC, 0x104}},
    {16235, 2, {0x100, 0xFFFE, 0x1048, 0x24FE, 0x7990, 0x10FC, 0x2290, 0x7CFC, 0x890, 0x30FE, 0xC080, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {16236, 2, {0x0, 0xFEFE, 0x0, 0x7C7C, 0x4444, 0x4444, 0x7C7C, 0x0, 0xFEFE, 0x9292, 0x9292, 0xFEFE, 0x9292, 0x9292, 0xFEFE, 0x8282}},
    {16237, 2, {0x0, 0x1FC, 0xF820, 0xABFE, 0xAA22, 0xA9AC, 0xF820, 0xA9AC, 0xA800, 0xABFE, 0xA820, 0xF9FC, 0x8954, 0x154, 0x154, 0x10C}},
    {16238, 2, {0xFE0, 0x920, 0xFE0, 0x920, 0xFE0, 0x7C7C, 0x5454, 0x7C7C, 0x5454, 0x7D7C, 0x1100, 0x1FF8, 0x2100, 0x5FF0, 0x100, 0xFFFE}},
    {16239, 2, {0x222, 0x3FE, 0xF890, 0xA9FE, 0xAB10, 0xADFE, 0xF910, 0xA9FE, 0xA910, 0xA9FE, 0xA900, 0xFBFE, 0x8A8A, 0x376, 0x252, 0x276}},
    {16240, 2, {0x0, 0x777C, 0x5514, 0x5514, 0x5510, 0x7710, 0x5550, 0x555C, 0x5550, 0x7750, 0x5550, 0x5550, 0x5550, 0xB550, 0x89BE, 0x1300}},
    {16241, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9008, 0x5208, 0x5208, 0x1208, 0x3208, 0x5208, 0x9238, 0x13C8, 0x2208, 0x2008, 0x4008, 0x8008}},
    {16242, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FE, 0x5080, 0x1100, 0x3200, 0x53F8, 0x9008, 0x1008, 0x2008, 0x2008, 0x4050, 0x8020}},
    {16243, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5040, 0x1040, 0x3040, 0x5FFE, 0x9040, 0x1040, 0x2040, 0x2040, 0x4140, 0x8080}},
    {16244, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9010, 0x5010, 0x57FE, 0x1010, 0x3010, 0x5210, 0x9120, 0x10A0, 0x2040, 0x20A0, 0x4318, 0x8C06}},
    {16245, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x5040, 0x1040, 0x37FC, 0x5040, 0x9040, 0x1040, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {16246, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53F8, 0x5008, 0x1008, 0x3008, 0x53F8, 0x9200, 0x1202, 0x2202, 0x2202, 0x41FE, 0x8000}},
    {16247, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x5044, 0x5768, 0x1170, 0x3160, 0x5150, 0x9250, 0x1248, 0x2444, 0x2842, 0x4140, 0x8080}},
    {16248, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x53FC, 0x1244, 0x3448, 0x50A0, 0x90A0, 0x1120, 0x2120, 0x2222, 0x4422, 0x881E}},
    {16249, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x901C, 0x53E0, 0x5200, 0x1200, 0x33F8, 0x5288, 0x9288, 0x1250, 0x2420, 0x2450, 0x4888, 0x9306}},
    {16250, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97F8, 0x5108, 0x5110, 0x1120, 0x313C, 0x5204, 0x9288, 0x1250, 0x2420, 0x2450, 0x4888, 0x9306}},
    {16251, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x57FC, 0x5040, 0x1040, 0x37F8, 0x5208, 0x9110, 0x10A0, 0x2040, 0x20A0, 0x4318, 0x8C06}},
    {16252, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9004, 0x57E4, 0x5024, 0x1024, 0x33E4, 0x5204, 0x9404, 0x17E4, 0x2024, 0x2024, 0x4144, 0x8084}},
    {16253, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x5040, 0x57FE, 0x1040, 0x3040, 0x53F8, 0x9208, 0x1208, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {16254, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5248, 0x5248, 0x1248, 0x33F8, 0x5048, 0x9040, 0x1444, 0x2444, 0x2444, 0x47FC, 0x8004}},
    {16255, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x901C, 0x53E0, 0x5290, 0x1290, 0x3290, 0x5290, 0x9290, 0x1288, 0x24C8, 0x24A4, 0x48D4, 0x9092}},
    {16256, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0xA080, 0x6FFC, 0x2080, 0x2080, 0x6080, 0xBFFE, 0x2100, 0x2210, 0x4408, 0x4FFC, 0x8404}},
    {16257, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9050, 0x5048, 0x5040, 0x17FC, 0x30E0, 0x5160, 0x9150, 0x1248, 0x2444, 0x2842, 0x4040, 0x8040}},
    {16258, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9080, 0x5110, 0x5208, 0x17FC, 0x3204, 0x5000, 0x93F8, 0x1208, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {16259, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x57FC, 0x5404, 0x1808, 0x3200, 0x5210, 0x9260, 0x1380, 0x2200, 0x2204, 0x4204, 0x81FC}},
    {16260, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x57FC, 0x1444, 0x3444, 0x5444, 0x97FC, 0x1444, 0x2444, 0x2444, 0x47FC, 0x8404}},
    {16261, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x905C, 0x53E0, 0x5240, 0x1240, 0x33FC, 0x5044, 0x90C4, 0x1154, 0x2248, 0x2C40, 0x4040, 0x8040}},
    {16262, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x53F8, 0x5248, 0x1248, 0x33F8, 0x5248, 0x9248, 0x1FFE, 0x2208, 0x2208, 0x4228, 0x8210}},
    {16263, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0xA7F8, 0x6408, 0x2408, 0x27F8, 0x6408, 0xA408, 0x27F8, 0x2408, 0x4408, 0x47F8, 0x8408}},
    {16264, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9108, 0x5108, 0x5108, 0x12FE, 0x3208, 0x5648, 0x9A28, 0x1228, 0x2208, 0x2208, 0x4228, 0x8210}},
    {16265, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x50A0, 0x5110, 0x1248, 0x3426, 0x5020, 0x93F8, 0x1008, 0x2110, 0x20A0, 0x4040, 0x8020}},
    {16266, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2040, 0xAFFE, 0x6040, 0x27FC, 0x2444, 0x6454, 0xA4E8, 0x2150, 0x2248, 0x4C46, 0x4040, 0x8040}},
    {16267, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x9110, 0x5210, 0x52FE, 0x1610, 0x3A38, 0x5238, 0x9254, 0x1254, 0x2292, 0x2210, 0x4210, 0x8210}},
    {16268, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5240, 0x53FC, 0x1440, 0x3040, 0x5FFE, 0x9120, 0x1120, 0x2120, 0x2222, 0x4422, 0x981E}},
    {16269, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x5FFE, 0x5040, 0x1080, 0x37FC, 0x54A4, 0x94A4, 0x14A4, 0x24A4, 0x24A4, 0x4494, 0x8408}},
    {16270, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57BC, 0x54A4, 0x14A4, 0x34A4, 0x54A4, 0x94E4, 0x1404, 0x2404, 0x2404, 0x47FC, 0x8404}},
    {16271, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x57FC, 0x5040, 0x1040, 0x3FFE, 0x5000, 0x9040, 0x17FC, 0x2040, 0x2040, 0x5FFE, 0x8000}},
    {16272, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9004, 0x5FC4, 0x5214, 0x1214, 0x37D4, 0x5454, 0x9A54, 0x1194, 0x2084, 0x2104, 0x4214, 0x8C08}},
    {16273, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x90A0, 0x5110, 0x5208, 0x1DF6, 0x3000, 0x5000, 0x93F8, 0x1208, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {16274, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x9100, 0x511E, 0x5FD2, 0x1252, 0x3252, 0x5452, 0x9492, 0x1292, 0x2112, 0x229E, 0x4452, 0x8840}},
    {16275, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F0, 0x5210, 0x53F0, 0x1000, 0x37F8, 0x5408, 0x97F8, 0x1408, 0x27F8, 0x2408, 0x4428, 0x8410}},
    {16276, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53F8, 0x5208, 0x1208, 0x33F8, 0x5000, 0x97FC, 0x1404, 0x2404, 0x2404, 0x47FC, 0x8404}},
    {16277, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x57FC, 0x5208, 0x1110, 0x3FFE, 0x5040, 0x9040, 0x17FC, 0x2040, 0x2040, 0x4040, 0x8040}},
    {16278, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5444, 0x17FC, 0x3444, 0x57FC, 0x9040, 0x17FC, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {16279, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x9080, 0x57F8, 0x5408, 0x17F8, 0x3408, 0x57F8, 0x9482, 0x1444, 0x2428, 0x2510, 0x4608, 0x8406}},
    {16280, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5040, 0x53F8, 0x1248, 0x33F8, 0x5248, 0x93F8, 0x1240, 0x2140, 0x2080, 0x4360, 0x8C1E}},
    {16281, 2, {0x80, 0x40, 0x1FFE, 0x1010, 0x9720, 0x50C0, 0x5130, 0x1688, 0x3FFE, 0x5100, 0x9240, 0x17F8, 0x2A48, 0x2268, 0x4250, 0x8040}},
    {16282, 2, {0x80, 0x40, 0x1FFE, 0x1120, 0x9124, 0x55E8, 0x5530, 0x1522, 0x35A2, 0x5E1E, 0x9000, 0x13F8, 0x2000, 0x2000, 0x4FFE, 0x8000}},
    {16283, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5008, 0x51F8, 0x1008, 0x33F8, 0x5000, 0x97FC, 0x1404, 0x23F8, 0x2110, 0x40E0, 0x871C}},
    {16284, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x57FC, 0x1040, 0x3248, 0x5248, 0x9554, 0x1842, 0x20A0, 0x2110, 0x4608, 0x9806}},
    {16285, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x3FFC, 0xA080, 0x6FF8, 0x2888, 0x2FF8, 0x6888, 0xAFF8, 0x2080, 0x3FFE, 0x4080, 0x4080, 0x8080}},
    {16286, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FFC, 0xA120, 0x6FFC, 0x2924, 0x2924, 0x6A1C, 0xAC04, 0x2FFC, 0x2804, 0x4804, 0x4FFC, 0x8804}},
    {16287, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x527C, 0x5240, 0x1240, 0x3FFE, 0x5040, 0x9144, 0x1248, 0x2410, 0x2060, 0x4380, 0x9C00}},
    {16288, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9880, 0x557C, 0x5210, 0x1D10, 0x3110, 0x517C, 0x9310, 0x1D10, 0x2110, 0x2110, 0x4A7E, 0x8400}},
    {16289, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x911E, 0x57D2, 0x5112, 0x1112, 0x3FF2, 0x5112, 0x9292, 0x145A, 0x2FD4, 0x2450, 0x4010, 0x8010}},
    {16290, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x5FFE, 0x5040, 0x1248, 0x3248, 0x5554, 0x98E2, 0x1150, 0x2248, 0x2C46, 0x4040, 0x8040}},
    {16291, 2, {0x80, 0x3FFE, 0x2080, 0x2140, 0xA220, 0x6410, 0x2808, 0x37F6, 0x6080, 0xA080, 0x2FF8, 0x2080, 0x2490, 0x42A0, 0x5FFE, 0x8000}},
    {16292, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x57FC, 0x5404, 0x1110, 0x3208, 0x5404, 0x93F8, 0x1040, 0x2040, 0x2040, 0x47FC, 0x8000}},
    {16293, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9200, 0x53BC, 0x5214, 0x1214, 0x3FD4, 0x5014, 0x9208, 0x1688, 0x2A48, 0x2214, 0x4A14, 0x8422}},
    {16294, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x57FC, 0x5208, 0x1110, 0x3FFE, 0x5000, 0x93F8, 0x1208, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {16295, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5248, 0x53F8, 0x1248, 0x33F8, 0x5040, 0x97FC, 0x10E0, 0x2150, 0x2248, 0x4C46, 0x8040}},
    {16296, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0xA140, 0x6220, 0x2410, 0x3FFE, 0x6010, 0xA7D0, 0x2450, 0x27D0, 0x4010, 0x4050, 0x8020}},
    {16297, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9038, 0x53C0, 0x5040, 0x1FFE, 0x3150, 0x5154, 0x9758, 0x1154, 0x2354, 0x254C, 0x4040, 0x8040}},
    {16298, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9248, 0x5454, 0x5862, 0x1380, 0x3FF8, 0x5208, 0x93F8, 0x1208, 0x23F8, 0x2208, 0x43F8, 0x8208}},
    {16299, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x93FC, 0x5204, 0x5204, 0x13FC, 0x3200, 0x53FC, 0x9354, 0x1554, 0x25FC, 0x2954, 0x4144, 0x810C}},
    {16300, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9424, 0x57A8, 0x5432, 0x1522, 0x361E, 0x5100, 0x93FC, 0x1204, 0x23FC, 0x2204, 0x43FC, 0x8204}},
    {16301, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2310, 0xAE10, 0x6254, 0x2254, 0x3F58, 0x6690, 0xA610, 0x2B28, 0x2A28, 0x5244, 0x4244, 0x8282}},
    {16302, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9080, 0x53F8, 0x5208, 0x13F8, 0x3208, 0x53F8, 0x9000, 0x17FC, 0x2040, 0x23FC, 0x4040, 0x8FFE}},
    {16303, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5248, 0x53F8, 0x1248, 0x33F8, 0x5000, 0x97FE, 0x1244, 0x2228, 0x2290, 0x4308, 0x8206}},
    {16304, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x223C, 0xA224, 0x6FA4, 0x223C, 0x2224, 0x6FA4, 0xA8BC, 0x28A4, 0x2FA4, 0x48C4, 0x4054, 0x8088}},
    {16305, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2F50, 0xA124, 0x6A18, 0x2408, 0x3BF6, 0x6120, 0xA120, 0x2FFC, 0x2120, 0x4224, 0x4424, 0x881C}},
    {16306, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F90, 0xA21E, 0x6FA4, 0x2224, 0x3FD4, 0x6414, 0xA794, 0x2488, 0x2488, 0x4894, 0x4AA4, 0x9142}},
    {16307, 2, {0x100, 0x80, 0x3FFE, 0x2880, 0x24FC, 0xA124, 0x6420, 0x3850, 0x298C, 0x6100, 0xBFFE, 0x2210, 0x2720, 0x40C0, 0x4330, 0x8C08}},
    {16308, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x57FC, 0x5150, 0x1248, 0x3446, 0x53F8, 0x9208, 0x13F8, 0x2208, 0x23F8, 0x4208, 0x8FFE}},
    {16309, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9108, 0x5FC8, 0x5008, 0x1788, 0x303E, 0x5788, 0x9008, 0x1788, 0x2488, 0x2488, 0x4788, 0x8488}},
    {16310, 2, {0x80, 0x3FFE, 0x2000, 0x23F8, 0xA208, 0x63C8, 0x2248, 0x2FFE, 0x6802, 0xB3FC, 0x2208, 0x23F8, 0x2208, 0x43F8, 0x4208, 0x8218}},
    {16311, 2, {0x80, 0x3FFE, 0x2000, 0x203E, 0xAFC0, 0x687C, 0x2840, 0x2BFE, 0x6A42, 0xAA78, 0x2AC4, 0x2A3C, 0x3200, 0x54F0, 0x6494, 0x890C}},
    {16312, 2, {0x80, 0x40, 0x1FFE, 0x1020, 0x97FE, 0x5000, 0x51F8, 0x1108, 0x31F8, 0x5000, 0x97FE, 0x1402, 0x25FA, 0x250A, 0x45FA, 0x8406}},
    {16313, 2, {0x100, 0x80, 0x3FFE, 0x2120, 0x2128, 0xA5B0, 0x6524, 0x2524, 0x259C, 0x6E40, 0xA040, 0x27FC, 0x20E0, 0x4150, 0x4248, 0x8C46}},
    {16314, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27FC, 0xA440, 0x67F8, 0x2440, 0x27F8, 0x6440, 0xA7FC, 0x2004, 0x2AA4, 0x4AA4, 0x5014, 0x8008}},
    {16315, 2, {0x80, 0x40, 0x1FFE, 0x1010, 0x9728, 0x5544, 0x5582, 0x167C, 0x3510, 0x5510, 0x95FE, 0x1510, 0x2754, 0x2492, 0x4552, 0x8420}},
    {16316, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x29F8, 0xA508, 0x65F8, 0x2108, 0x3DF8, 0x6524, 0xA518, 0x2548, 0x2584, 0x4504, 0x4A00, 0x91FE}},
    {16317, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2880, 0xA5F8, 0x6108, 0x2108, 0x2DF8, 0x6500, 0xA5FC, 0x2504, 0x2504, 0x45FC, 0x4A00, 0x91FE}},
    {16318, 2, {0x80, 0x40, 0x1FFE, 0x1208, 0x9110, 0x5FFE, 0x5120, 0x17FC, 0x3124, 0x5FFE, 0x9124, 0x17FC, 0x2330, 0x2528, 0x5926, 0x8120}},
    {16319, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2100, 0xAFBE, 0x68A2, 0x2FA4, 0x28A8, 0x6FA4, 0xAA22, 0x2922, 0x2AAA, 0x4CA4, 0x4820, 0x8020}},
    {16320, 2, {0x100, 0x80, 0x3FFE, 0x2020, 0x2E2E, 0xA424, 0x64A4, 0x24A4, 0x24A4, 0x6EAE, 0xA424, 0x2424, 0x2444, 0x4644, 0x588E, 0x8100}},
    {16321, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x93F8, 0x5208, 0x53F8, 0x1208, 0x33F8, 0x5080, 0x97FC, 0x1210, 0x2448, 0x29F6, 0x4040, 0x8040}},
    {16322, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x9550, 0x5550, 0x57DE, 0x1024, 0x37D4, 0x5114, 0x9114, 0x17D4, 0x2108, 0x2108, 0x4FD4, 0x8022}},
    {16323, 2, {0x80, 0x40, 0x1FFE, 0x12A8, 0x92A8, 0x57FC, 0x52AA, 0x14E6, 0x3800, 0x57FC, 0x9444, 0x13F8, 0x2248, 0x2248, 0x4258, 0x8040}},
    {16324, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FDC, 0xAA14, 0x6BD4, 0x2D16, 0x2920, 0x6FDC, 0xA914, 0x2A94, 0x2C54, 0x4808, 0x4FD4, 0x8022}},
    {16325, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2410, 0xA89C, 0x7090, 0x2290, 0x25FE, 0x6C10, 0xB490, 0x249C, 0x2490, 0x4550, 0x463E, 0x8400}},
    {16326, 2, {0x80, 0x1FFE, 0x1000, 0x1038, 0x93C0, 0x5040, 0x17FC, 0x3150, 0x5248, 0x9486, 0x1040, 0x17FC, 0x20A2, 0x2314, 0x4D48, 0x8186}},
    {16327, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97BC, 0x5084, 0x54A4, 0x1294, 0x34A4, 0x5100, 0x93FC, 0x1204, 0x23FC, 0x2204, 0x43FC, 0x8204}},
    {16328, 2, {0x100, 0x80, 0x3FFE, 0x2228, 0x2224, 0xA57E, 0x6948, 0x2E48, 0x22FE, 0x6548, 0xA948, 0x2F7E, 0x2248, 0x4448, 0x487E, 0x9040}},
    {16329, 2, {0x80, 0x3FFE, 0x2040, 0x27FC, 0xA444, 0x67FC, 0x2040, 0x3FFE, 0x6404, 0xA7FC, 0x2404, 0x27FC, 0x2404, 0x47FC, 0x4208, 0x8404}},
    {16330, 2, {0x80, 0x3FFE, 0x2000, 0x2F78, 0xA210, 0x6A50, 0x2F7C, 0x2630, 0x6B54, 0xB28C, 0x27F8, 0x2408, 0x27F8, 0x4408, 0x47F8, 0x8408}},
    {16331, 2, {0x80, 0x1FFE, 0x1000, 0x13F8, 0x9208, 0x53F8, 0x1000, 0x37BC, 0x54A4, 0x97BC, 0x1040, 0x1FFE, 0x2150, 0x2248, 0x4C46, 0x8040}},
    {16332, 2, {0x80, 0x3FFE, 0x2080, 0x2140, 0xA220, 0x6410, 0x3BEC, 0x2000, 0x6F78, 0xA948, 0x2948, 0x2F78, 0x2220, 0x4550, 0x4888, 0x9104}},
    {16333, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x287C, 0xAF14, 0x7124, 0x2FCC, 0x2AA8, 0x6FBE, 0xAAC8, 0x2F88, 0x2AFE, 0x4A88, 0x5088, 0xA188}},
    {16334, 2, {0x80, 0x1FFE, 0x1040, 0x17FC, 0x9150, 0x5FFE, 0x1110, 0x37FC, 0x5404, 0x97FC, 0x1404, 0x17FC, 0x2404, 0x27FC, 0x4110, 0x8208}},
    {16335, 2, {0x80, 0x3FFE, 0x2000, 0x27F8, 0xA088, 0x6FFE, 0x2108, 0x27F8, 0x6310, 0xA4A0, 0x2FFC, 0x3040, 0x23F8, 0x4040, 0x4FFE, 0x8040}},
    {16336, 2, {0x80, 0x3FFE, 0x2210, 0x223E, 0xAF44, 0x62A8, 0x2210, 0x3FA8, 0x625E, 0xA222, 0x2A54, 0x2B08, 0x2A10, 0x4A60, 0x5600, 0xA3FE}},
    {16337, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0xA492, 0x628A, 0x2492, 0x2220, 0x67FE, 0xAC20, 0x37FE, 0x2420, 0x27FE, 0x4420, 0x47FE, 0x8400}},
    {16338, 2, {0x80, 0x3FFE, 0x2040, 0x2FFC, 0xA040, 0x67FC, 0x2000, 0x3FFE, 0x6002, 0xA7F8, 0x2040, 0x3FFE, 0x2008, 0x4F7E, 0x4948, 0x8F18}},
    {16339, 2, {0x80, 0x3FFE, 0x2040, 0x27FC, 0xA248, 0x6248, 0x2554, 0x28A2, 0x6318, 0xADF6, 0x2110, 0x21F0, 0x2240, 0x427C, 0x4540, 0x88FE}},
    {16340, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27FC, 0xA444, 0x67FC, 0x2444, 0x27FC, 0x6000, 0xAFBE, 0x2AAA, 0x2FBE, 0x4AAA, 0x4FBE, 0x88A2}},
    {16341, 2, {0x80, 0x3FFE, 0x2000, 0x2F3C, 0xA924, 0x6F3C, 0x2924, 0x2F3C, 0x69E4, 0xAA24, 0x2C44, 0x2B64, 0x2A24, 0x4B64, 0x4A24, 0x8BEC}},
    {16342, 2, {0x80, 0x3FFE, 0x2040, 0x207C, 0xA040, 0x67FE, 0x2442, 0x25F8, 0x6444, 0xA4FC, 0x24A8, 0x24F8, 0x24A8, 0x49FC, 0x4954, 0x93FE}},
    {16343, 2, {0x80, 0x3FFE, 0x2000, 0x2300, 0xBCFE, 0x6410, 0x3F7C, 0x2E44, 0x757C, 0xA444, 0x2E7C, 0x2A44, 0x2A7C, 0x5328, 0x5244, 0xA082}},
    {16344, 2, {0x80, 0x3FFE, 0x2208, 0x2AAA, 0xA248, 0x6514, 0x28A2, 0x2FFE, 0x6802, 0xA3F8, 0x2208, 0x23F8, 0x2000, 0x47FC, 0x4404, 0x87FC}},
    {16345, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x23F8, 0xA2A8, 0x63F8, 0x2000, 0x2FBE, 0x6AAA, 0xAFBE, 0x2040, 0x2FFE, 0x40A0, 0x4318, 0x8C06}},
    {16346, 2, {0x80, 0x3FFE, 0x2000, 0x2FFE, 0xA912, 0x6FFE, 0x2228, 0x2444, 0x69FE, 0xAF48, 0x227E, 0x2548, 0x2F7E, 0x4048, 0x557E, 0x9540}},
    {16347, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0xA514, 0x671C, 0x2554, 0x2FFE, 0x6920, 0xAFFC, 0x2924, 0x2FFC, 0x2A24, 0x4BB8, 0x5222, 0xA39E}},
    {16348, 2, {0x80, 0x3FFE, 0x2100, 0x27F8, 0xA548, 0x64A8, 0x27F8, 0x2210, 0x67F8, 0xBD2E, 0x27F8, 0x2000, 0x2FFC, 0x4924, 0x4FFC, 0x8924}},
    {16349, 2, {0x400, 0x800, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x200, 0x400, 0x820, 0x1010, 0x2008, 0x7FFC, 0x2004}},
    {16350, 2, {0x1008, 0x103C, 0x21E0, 0x7C20, 0x4420, 0x443C, 0x45E0, 0x4420, 0x7C20, 0x443E, 0x47E0, 0x4420, 0x4422, 0x7C22, 0x4422, 0x1E}},
    {16351, 2, {0x1020, 0x1020, 0x2040, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {16352, 2, {0x1040, 0x1044, 0x2244, 0x7948, 0x4950, 0x4840, 0x4FFE, 0x4890, 0x7890, 0x4890, 0x4890, 0x4892, 0x4912, 0x7912, 0x4A0E, 0x400}},
    {16353, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x2448, 0x1450, 0xC60, 0x1450, 0x6448, 0x842, 0x1042, 0x603E}},
    {16354, 2, {0x0, 0x7CFC, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7D14, 0x208, 0x800, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {16355, 2, {0x1000, 0x13F8, 0x2208, 0x7BF8, 0x4A08, 0x4BF8, 0x4900, 0x4BFC, 0x7C44, 0x4A44, 0x4AA4, 0x4A04, 0x4BF4, 0x7804, 0x4828, 0x10}},
    {16356, 2, {0x1040, 0x1040, 0x23FC, 0x7840, 0x49F8, 0x4880, 0x4BFC, 0x4910, 0x7A08, 0x4DF6, 0x4910, 0x4910, 0x49F0, 0x7910, 0x4910, 0x1F0}},
    {16357, 2, {0x1040, 0x1040, 0x23FE, 0x7AA2, 0x4890, 0x49FE, 0x4910, 0x4B10, 0x7DFE, 0x4910, 0x4910, 0x49FE, 0x4910, 0x7910, 0x49FE, 0x100}},
    {16358, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x8486, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16359, 2, {0x1020, 0x13FE, 0x2020, 0x79FC, 0x4800, 0x4BFE, 0x4802, 0x49FC, 0x7820, 0x4BFE, 0x4800, 0x4BFE, 0x4804, 0x7BBE, 0x4AA4, 0x38C}},
    {16360, 2, {0x1000, 0x13FE, 0x2222, 0x7ACC, 0x4A44, 0x4BEE, 0x4A44, 0x4AEE, 0x7B54, 0x4A44, 0x4A10, 0x4A10, 0x4A5E, 0x7A50, 0x4A50, 0x5FE}},
    {16361, 2, {0x1124, 0x10A8, 0x23FE, 0x7A02, 0x48F8, 0x4888, 0x4BFE, 0x4AAA, 0x7A72, 0x4BFE, 0x4820, 0x4BFE, 0x4820, 0x7FFE, 0x4AA4, 0x452}},
    {16362, 2, {0x820, 0x820, 0x820, 0x7F78, 0x4928, 0x4A28, 0x4828, 0x7EA8, 0x4268, 0x4228, 0x5258, 0x4C48, 0x444A, 0x4A8A, 0x9286, 0x2102}},
    {16363, 2, {0x2080, 0x2084, 0x3E98, 0x20E4, 0x2684, 0x387C, 0x80, 0x1FFC, 0x1084, 0x1088, 0x17F0, 0x1210, 0x2120, 0x20C0, 0x4330, 0x1C0E}},
    {16364, 2, {0x800, 0x800, 0x87C, 0x7F54, 0x4954, 0x4A54, 0x4854, 0x7E54, 0x427C, 0x4240, 0x5240, 0x4C40, 0x4442, 0x4A42, 0x923E, 0x2000}},
    {16365, 2, {0xC10, 0x7010, 0x4010, 0x40FE, 0x4092, 0x7E94, 0x4890, 0x48FC, 0x58A4, 0x48A4, 0x4CA8, 0x4AA8, 0x4890, 0x4928, 0x8944, 0x282}},
    {16366, 2, {0x1010, 0x1010, 0x1010, 0x7CFE, 0x5492, 0x5494, 0x5490, 0x54FC, 0x54A4, 0xFEA4, 0x10A8, 0x10A8, 0x2890, 0x2528, 0x4544, 0x8282}},
    {16367, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FFC, 0x2104, 0x2FE8, 0x2420, 0x4240, 0x4180, 0x9E7C}},
    {16368, 2, {0x2010, 0x1010, 0x10, 0xFEFE, 0x92, 0x2894, 0x4490, 0x82FC, 0x4A4, 0x44A4, 0x28A8, 0x10A8, 0x2890, 0x4528, 0x8544, 0x282}},
    {16369, 2, {0x100C, 0x11F0, 0xFD00, 0x11FE, 0x1D10, 0xF210, 0x1210, 0x3410, 0x100, 0x3FFC, 0x2104, 0x2FE8, 0x2420, 0x4240, 0x4180, 0x9E7C}},
    {16370, 2, {0x1010, 0x1010, 0xFE10, 0x10FE, 0x2892, 0x4494, 0xFE90, 0x4FC, 0x74A4, 0x54A4, 0x54A8, 0x74A8, 0x5490, 0x528, 0x1544, 0xA82}},
    {16371, 2, {0x82A, 0x82A, 0x854, 0x7F54, 0x492A, 0x4A2A, 0x482A, 0x7E00, 0x42FE, 0x4292, 0x5292, 0x4CFE, 0x4492, 0x4A92, 0x92FE, 0x2082}},
    {16372, 2, {0x1010, 0x1010, 0xFF10, 0x24FE, 0x4292, 0x8194, 0x7E90, 0x24FC, 0x3CA4, 0x24A4, 0x3CA8, 0x24A8, 0x2E90, 0xF528, 0x544, 0x682}},
    {16373, 2, {0x800, 0x8FE, 0x810, 0x7F20, 0x49FE, 0x4AAA, 0x48AA, 0x7EAA, 0x42A2, 0x4296, 0x5210, 0x4CFE, 0x4410, 0x4A28, 0x9244, 0x2082}},
    {16374, 2, {0x2010, 0x1010, 0xFE10, 0xFE, 0x4492, 0x2894, 0xFE90, 0xA2FC, 0x10A4, 0xFEA4, 0x20A8, 0x3CA8, 0x2490, 0x4528, 0x5544, 0x8A82}},
    {16375, 2, {0x4410, 0x2410, 0x2810, 0xFEFE, 0x1092, 0x7C94, 0x1090, 0xFEFC, 0x20A4, 0x3EA4, 0x48A8, 0x48A8, 0x8E90, 0x7928, 0x2144, 0x282}},
    {16376, 2, {0x87C, 0x844, 0x87C, 0x7F44, 0x497C, 0x4A00, 0x48FE, 0x7EAA, 0x42FE, 0x4200, 0x52FC, 0x4C44, 0x4428, 0x4A10, 0x9268, 0x2186}},
    {16377, 2, {0x1010, 0x1210, 0x117C, 0x7C10, 0x54FE, 0x5044, 0x5328, 0x7DFE, 0x4510, 0x657C, 0x5910, 0x49FE, 0x5510, 0x9510, 0x22FE, 0x400}},
    {16378, 2, {0x7F08, 0x4108, 0x7F08, 0x417E, 0x7F4A, 0x224A, 0x7F48, 0x227C, 0xFFD4, 0x2254, 0x4954, 0xAAD4, 0x1C48, 0x2A88, 0x4914, 0x1822}},
    {16379, 2, {0x0, 0x1FF0, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16380, 2, {0x0, 0x3FF8, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16381, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC406, 0x800, 0xFE0, 0x20, 0x140, 0x80, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16382, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF0, 0x100, 0x100, 0x500, 0x200, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16383, 2, {0x0, 0x7FFC, 0x104, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16384, 2, {0x1000, 0x9F8, 0x4108, 0x2508, 0x928, 0x1110, 0x7102, 0x1102, 0x10FE, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16385, 2, {0x200, 0xFFFE, 0x400, 0xFF0, 0x1810, 0x2FF0, 0xC810, 0xFF0, 0x810, 0x830, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16386, 2, {0x100, 0x7FFC, 0x4004, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16387, 2, {0x0, 0x7FFC, 0x200, 0xD08, 0x7190, 0x2A0, 0xCC0, 0x71A0, 0x698, 0x1A86, 0xE100, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16388, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8924, 0x2848, 0x4994, 0xE10, 0x77F0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16389, 2, {0x2020, 0x1020, 0xFC20, 0x9FC, 0x1020, 0x3820, 0x5420, 0x9020, 0x13FE, 0x1000, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16390, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7FFC, 0x0, 0x1108, 0x26D0, 0x1830, 0xE00E, 0x3FF8, 0x2448, 0xFFFE}},
    {16391, 2, {0x1080, 0x1080, 0x11FC, 0xFE20, 0x1020, 0x11FC, 0x7C20, 0x4524, 0x4524, 0x7DFC, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16392, 2, {0x7DFC, 0x4420, 0x7D20, 0x4520, 0x7DFE, 0x5050, 0x4890, 0x5512, 0x620E, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16393, 2, {0x70, 0x1F80, 0x200, 0xFFFE, 0x820, 0x37D8, 0xC446, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27D8, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {16394, 2, {0x27BC, 0x14A4, 0x9294, 0x44A4, 0x4040, 0x11B0, 0x2E4E, 0xE390, 0x2060, 0x2388, 0x2070, 0x380, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {16395, 2, {0x2100, 0x21FE, 0x2220, 0x243E, 0xF820, 0x23FE, 0x22AA, 0x2252, 0x2326, 0x3A52, 0xE2AA, 0x43FE, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {16396, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7D7C, 0x1110, 0x7D7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x2448, 0xFFFE}},
    {16397, 2, {0x1040, 0xFEF8, 0x2948, 0xFE30, 0x1CE, 0x7CF8, 0x4420, 0x7DFC, 0x4488, 0x7CF8, 0x1088, 0xFEF8, 0x1050, 0x3FF8, 0x2448, 0xFFFE}},
    {16398, 2, {0x0, 0x3FE, 0x7C40, 0x4440, 0x4480, 0x7C80, 0x4500, 0x45FC, 0x7C04, 0x4404, 0x4404, 0x4404, 0x7C04, 0x4404, 0x28, 0x10}},
    {16399, 2, {0x1000, 0x1000, 0x1FFC, 0x2004, 0x2004, 0x5FE4, 0x9024, 0x1FE4, 0x1024, 0x1FE4, 0x1024, 0x1FE4, 0x1024, 0x4, 0x28, 0x10}},
    {16400, 2, {0x8, 0x7E08, 0x4210, 0x4220, 0x4244, 0x7E04, 0x4208, 0x4210, 0x4222, 0x7E42, 0x4204, 0x4204, 0x4208, 0x7E10, 0x4220, 0xC0}},
    {16401, 2, {0x0, 0x1F8, 0x7C88, 0x44C8, 0x44A8, 0x7CA8, 0x4488, 0x4450, 0x7C50, 0x4450, 0x4420, 0x4420, 0x7C50, 0x4488, 0x104, 0x202}},
    {16402, 2, {0x8, 0x3C, 0x7DE0, 0x4420, 0x4420, 0x7C20, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {16403, 2, {0x40, 0x40, 0x7C40, 0x4440, 0x45F8, 0x7C48, 0x4448, 0x4448, 0x7D48, 0x44C8, 0x4448, 0x44A8, 0x7CAA, 0x450A, 0x206, 0x402}},
    {16404, 2, {0x0, 0x0, 0x7DFC, 0x4444, 0x4444, 0x7D44, 0x4544, 0x4544, 0x7D44, 0x4644, 0x4444, 0x4484, 0x7C84, 0x4504, 0x228, 0x410}},
    {16405, 2, {0x4, 0x3E4, 0x7824, 0x4824, 0x4824, 0x79E4, 0x4904, 0x4904, 0x7A04, 0x4BE4, 0x4824, 0x4824, 0x7824, 0x4824, 0x144, 0x84}},
    {16406, 2, {0x0, 0x0, 0x7DFC, 0x4420, 0x4420, 0x7C20, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4450, 0x4450, 0x7C88, 0x4488, 0x104, 0x202}},
    {16407, 2, {0x40, 0x40, 0x7840, 0x4BF8, 0x4848, 0x7848, 0x4848, 0x4848, 0x7BFE, 0x4840, 0x48A0, 0x48A0, 0x7910, 0x4910, 0x208, 0x406}},
    {16408, 2, {0x0, 0x3FC, 0x7890, 0x4890, 0x4890, 0x7890, 0x4890, 0x4FFE, 0x7890, 0x4890, 0x4890, 0x4890, 0x7910, 0x4910, 0x210, 0x410}},
    {16409, 2, {0x0, 0x100, 0x797E, 0x4912, 0x4912, 0x7912, 0x49D2, 0x4F12, 0x7912, 0x4912, 0x4952, 0x4992, 0x7922, 0x4822, 0x4A, 0x84}},
    {16410, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xC00, 0x70FC, 0x4084, 0x4084, 0x4C84, 0x7094, 0x4088, 0x80}},
    {16411, 2, {0x100, 0x100, 0x7900, 0x49FC, 0x4AA4, 0x7CA4, 0x48A4, 0x4924, 0x7924, 0x4A44, 0x4C44, 0x4884, 0x7904, 0x4A04, 0x428, 0x10}},
    {16412, 2, {0x800, 0x800, 0x1FF8, 0x2248, 0x4448, 0x888, 0x1128, 0x2210, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16413, 2, {0x10, 0x110, 0x7910, 0x4912, 0x4912, 0x7914, 0x49D8, 0x4910, 0x7910, 0x4910, 0x4910, 0x4912, 0x7952, 0x4992, 0x10E, 0x0}},
    {16414, 2, {0x10, 0x110, 0x7C90, 0x4490, 0x4410, 0x7D10, 0x4490, 0x4490, 0x7C10, 0x441E, 0x47F0, 0x4410, 0x7C10, 0x4410, 0x10, 0x10}},
    {16415, 2, {0x90, 0x90, 0x7890, 0x4BFC, 0x4894, 0x7894, 0x4BFC, 0x4A90, 0x7A90, 0x4BFE, 0x4892, 0x4892, 0x791A, 0x4914, 0x210, 0x410}},
    {16416, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x0}},
    {16417, 2, {0x20, 0x20, 0x7C20, 0x4420, 0x443E, 0x7C20, 0x4420, 0x4420, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x7D04, 0x4504, 0x1FC, 0x104}},
    {16418, 2, {0x40, 0x40, 0x7880, 0x48FC, 0x4920, 0x7A20, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x4850, 0x4850, 0x7888, 0x4888, 0x104, 0x202}},
    {16419, 2, {0x0, 0x40, 0x7820, 0x4828, 0x4808, 0x7888, 0x4890, 0x4A94, 0x7AA2, 0x4AA2, 0x4CC2, 0x4888, 0x7988, 0x4A88, 0x478, 0x0}},
    {16420, 2, {0x1020, 0x1020, 0x3F30, 0x4128, 0xA224, 0x1422, 0x820, 0x3020, 0xDFF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16421, 2, {0x8, 0x1C, 0x7DE0, 0x4500, 0x4520, 0x7D20, 0x4520, 0x45FE, 0x7C20, 0x4420, 0x44A8, 0x44A4, 0x7D22, 0x4622, 0xA0, 0x40}},
    {16422, 2, {0x20, 0x20, 0x7820, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x4820, 0x79FC, 0x4904, 0x4904, 0x4904, 0x7904, 0x4904, 0x1FC, 0x104}},
    {16423, 2, {0x110, 0x110, 0x7910, 0x4910, 0x4BFC, 0x7910, 0x4910, 0x4910, 0x7910, 0x4FFE, 0x4800, 0x4910, 0x7908, 0x4A08, 0x404, 0x804}},
    {16424, 2, {0x8, 0x3C, 0x79E0, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x4820, 0x7820, 0x49FC, 0x4904, 0x4904, 0x7904, 0x4904, 0x1FC, 0x104}},
    {16425, 2, {0x80, 0x80, 0x78F8, 0x4908, 0x4B10, 0x7CA0, 0x4840, 0x48A0, 0x7918, 0x4A06, 0x4DF8, 0x4908, 0x7908, 0x4908, 0x1F8, 0x108}},
    {16426, 2, {0x80, 0x80, 0x793C, 0x4A00, 0x4C80, 0x7880, 0x497E, 0x4B08, 0x7D08, 0x4908, 0x4908, 0x4908, 0x7908, 0x4908, 0x128, 0x110}},
    {16427, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A08, 0x7C06, 0x4BF8, 0x4800, 0x7800, 0x4BF8, 0x4A08, 0x4A08, 0x7A08, 0x4A08, 0x3F8, 0x208}},
    {16428, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A08, 0x7C06, 0x4BF8, 0x4840, 0x7840, 0x4840, 0x4BF8, 0x4840, 0x7840, 0x4840, 0x7FE, 0x0}},
    {16429, 2, {0x20, 0x10, 0x7C10, 0x45FE, 0x4420, 0x7C24, 0x4444, 0x44F8, 0x7C12, 0x4422, 0x4444, 0x4588, 0x7C10, 0x4428, 0x44, 0x182}},
    {16430, 2, {0x80, 0x40, 0x7BFC, 0x4900, 0x4900, 0x7900, 0x49F8, 0x4800, 0x7950, 0x4950, 0x4950, 0x4950, 0x7A50, 0x4A52, 0x452, 0x80E}},
    {16431, 2, {0x48, 0x44, 0x7BFE, 0x4840, 0x4840, 0x7BFC, 0x4A44, 0x4A44, 0x7BFC, 0x4A44, 0x4A44, 0x4BFC, 0x7A44, 0x4A44, 0x254, 0x208}},
    {16432, 2, {0x88, 0x7884, 0x4902, 0x4A22, 0x4820, 0x7850, 0x4888, 0x4904, 0x7A02, 0x49FC, 0x4904, 0x4904, 0x7904, 0x4904, 0x1FC, 0x104}},
    {16433, 2, {0x80, 0x80, 0x79F8, 0x4910, 0x4AA0, 0x7840, 0x49B0, 0x4E4E, 0x79F0, 0x4840, 0x49F0, 0x4840, 0x7FFC, 0x4840, 0x40, 0x40}},
    {16434, 2, {0x100, 0x100, 0x79FC, 0x4A00, 0x4DF8, 0x7908, 0x4948, 0x4928, 0x7BFE, 0x4908, 0x4A48, 0x4A28, 0x7BFC, 0x4808, 0x50, 0x20}},
    {16435, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x4820, 0x79FC, 0x4924, 0x4924, 0x79FC, 0x4820, 0x4870, 0x48A8, 0x7924, 0x4A22, 0x20, 0x20}},
    {16436, 2, {0x0, 0x3FC, 0x7800, 0x4924, 0x4A48, 0x7C90, 0x4A48, 0x4924, 0x7800, 0x4BFC, 0x4840, 0x4840, 0x7840, 0x4840, 0x7FE, 0x0}},
    {16437, 2, {0x20, 0x40, 0x7888, 0x4904, 0x4BFE, 0x7882, 0x4880, 0x49FC, 0x7A20, 0x4820, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x602}},
    {16438, 2, {0x80, 0x40, 0x7BF8, 0x4A08, 0x4A08, 0x7BF8, 0x4A08, 0x4A08, 0x7BF8, 0x4A40, 0x4A44, 0x4A28, 0x7A10, 0x4A88, 0x306, 0x200}},
    {16439, 2, {0x88, 0x88, 0x7888, 0x49C8, 0x489E, 0x788A, 0x488A, 0x4BEA, 0x788A, 0x488A, 0x490A, 0x494A, 0x7BEA, 0x4912, 0x2A, 0x44}},
    {16440, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16441, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x7820, 0x4924, 0x4924, 0x7924, 0x4AAA, 0x4820, 0x4850, 0x7850, 0x4888, 0x104, 0x202}},
    {16442, 2, {0x1080, 0x20FC, 0x6910, 0xAA90, 0x2860, 0x2890, 0x230C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16443, 2, {0x20, 0x20, 0x79FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4800, 0x7BFE, 0x4880, 0x4900, 0x49FC, 0x7804, 0x4804, 0x28, 0x10}},
    {16444, 2, {0x0, 0x3FC, 0x7A00, 0x4A00, 0x4AF8, 0x7A00, 0x4A00, 0x4BFC, 0x7AA0, 0x4AA4, 0x4AA8, 0x4A90, 0x7A90, 0x4C88, 0x4A4, 0x8C2}},
    {16445, 2, {0x4, 0xE, 0x7BB8, 0x4888, 0x4888, 0x7928, 0x492E, 0x4BA8, 0x78A8, 0x4AA8, 0x4AA8, 0x493E, 0x7900, 0x4A80, 0x47E, 0x800}},
    {16446, 2, {0x0, 0xFEFC, 0x1044, 0x2028, 0x7E10, 0xA228, 0x3EC6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16447, 2, {0x104, 0x10E, 0xF130, 0x9120, 0x97E0, 0xF120, 0x913E, 0x9164, 0xF1A4, 0x9724, 0x9124, 0x9124, 0xF124, 0x9124, 0x544, 0x284}},
    {16448, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208}},
    {16449, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x4800, 0x79FC, 0x4904, 0x4904, 0x7904, 0x49FC, 0x4820, 0x48A8, 0x7924, 0x4A22, 0xA0, 0x40}},
    {16450, 2, {0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16451, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x4820, 0x79FC, 0x4820, 0x4820, 0x7BFE, 0x4840, 0x48A4, 0x49A8, 0x7A90, 0x4C88, 0xC6, 0x80}},
    {16452, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x7BFE, 0x4888, 0x4850, 0x79FC, 0x4820, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20, 0x20}},
    {16453, 2, {0x0, 0x1FC, 0x7808, 0x4810, 0x4820, 0x7BFE, 0x4820, 0x48A0, 0x7840, 0x49FC, 0x4954, 0x4954, 0x7954, 0x4954, 0x7FE, 0x0}},
    {16454, 2, {0x40, 0x80, 0x7BFC, 0x4910, 0x4A48, 0x7C46, 0x4BF8, 0x4A48, 0x7A48, 0x4BF8, 0x4A48, 0x4A48, 0x7BF8, 0x4842, 0x42, 0x3E}},
    {16455, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0xA96C, 0x4994, 0xE10, 0x77F0, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16456, 2, {0x1040, 0x840, 0x3E7E, 0x2288, 0x2350, 0x3E20, 0x2058, 0x4186, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16457, 2, {0x1048, 0x844, 0x3E7E, 0x23C0, 0x2224, 0x3E28, 0x2212, 0x206A, 0x4186, 0x9FF2, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16458, 2, {0x0, 0x3F8, 0x7948, 0x4910, 0x48A0, 0x7840, 0x49B0, 0x4E0E, 0x79F8, 0x4908, 0x49F8, 0x4908, 0x79F8, 0x4908, 0x1F8, 0x108}},
    {16459, 2, {0x48, 0x148, 0x7948, 0x4BFE, 0x4948, 0x7948, 0x4978, 0x4900, 0x79FE, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x4924, 0x622, 0x20}},
    {16460, 2, {0x0, 0x3FE, 0x7A22, 0x4BFE, 0x4A22, 0x7BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4904, 0x114, 0x108}},
    {16461, 2, {0x14, 0x12, 0xF010, 0x97FE, 0x9410, 0xF410, 0x95D0, 0x9412, 0xF412, 0x95D4, 0x9554, 0x9548, 0xF5DA, 0x942A, 0x846, 0x1082}},
    {16462, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4800, 0x7BFE, 0x4A02, 0x49FC, 0x4820, 0x7820, 0x4820, 0xA0, 0x40}},
    {16463, 2, {0x0, 0x3FE, 0x7850, 0x4850, 0x4BFE, 0x7A52, 0x4A52, 0x4BFE, 0x7800, 0x4840, 0x4FFE, 0x4888, 0x7990, 0x4860, 0xD8, 0x304}},
    {16464, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48A8, 0x7820, 0x49FC, 0x4924, 0x7924, 0x4924, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x202}},
    {16465, 2, {0x0, 0x1F8, 0x7908, 0x49F8, 0x4908, 0x79F8, 0x4800, 0x4BFE, 0x7900, 0x49FC, 0x4A54, 0x4C94, 0x7924, 0x4A44, 0xA8, 0x110}},
    {16466, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x7800, 0x4FFE, 0x4908, 0x79F8, 0x4908, 0x49F8, 0x4908, 0x793E, 0x4FC8, 0x8, 0x8}},
    {16467, 2, {0x0, 0x88, 0x7A52, 0x4A22, 0x4A52, 0x7A8A, 0x4A02, 0x4BFE, 0x7888, 0x4944, 0x4A7A, 0x4888, 0x7950, 0x4820, 0xD8, 0x706}},
    {16468, 2, {0x1E, 0x3E0, 0x7844, 0x4924, 0x4888, 0x79FC, 0x4840, 0x4840, 0x7BFE, 0x4880, 0x48FC, 0x4944, 0x7928, 0x4A10, 0x68, 0x186}},
    {16469, 2, {0x0, 0x3FC, 0x7A04, 0x4BFC, 0x4A20, 0x7BFE, 0x4A10, 0x4A8A, 0x7B06, 0x4800, 0x4BFC, 0x4A04, 0x7BFC, 0x4A04, 0x3FC, 0x204}},
    {16470, 2, {0x20, 0x124, 0x78A8, 0x4820, 0x4BFE, 0x78A8, 0x4924, 0x4A02, 0x7840, 0x4BFE, 0x4888, 0x4908, 0x7990, 0x4860, 0x198, 0x604}},
    {16471, 2, {0x20, 0x124, 0x7924, 0x4924, 0x49FC, 0x7800, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x7888, 0x4850, 0x3FE, 0x0}},
    {16472, 2, {0x8, 0x3C, 0x7BC0, 0x4804, 0x4A44, 0x7928, 0x49FC, 0x4A20, 0x7820, 0x4BFE, 0x4820, 0x4924, 0x7924, 0x4924, 0x1FC, 0x4}},
    {16473, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x4BFE, 0x7A02, 0x4840, 0x4BFE, 0x4888, 0x7990, 0x4860, 0x98, 0x304}},
    {16474, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x7800, 0x4BFC, 0x4A04, 0x7BFC, 0x4A04, 0x4BFC, 0x4A04, 0x7BFC, 0x4890, 0x108, 0x204}},
    {16475, 2, {0x20, 0x40, 0x7BFC, 0x4A24, 0x4A24, 0x7BFC, 0x4A24, 0x4A44, 0x7BFC, 0x4840, 0x48A8, 0x48B4, 0x793C, 0x4922, 0x222, 0x41E}},
    {16476, 2, {0x0, 0x1FC, 0x7D04, 0x45FC, 0x4504, 0x7DFC, 0x4420, 0x4524, 0x7CA8, 0x4420, 0x47FE, 0x4490, 0x7C90, 0x4512, 0x212, 0x40E}},
    {16477, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16478, 2, {0x820, 0x1450, 0x2288, 0x820, 0x1450, 0x2288, 0xFFFE, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0xFF0, 0x810}},
    {16479, 2, {0x108, 0x88, 0x7890, 0x4BFC, 0x4840, 0x79F8, 0x4840, 0x4BFE, 0x7880, 0x4900, 0x49FC, 0x4A20, 0x7C20, 0x4820, 0x3FE, 0x0}},
    {16480, 2, {0x110, 0xA0, 0x7800, 0x4FFE, 0x48A0, 0x7BF8, 0x48A8, 0x4FFE, 0x78A8, 0x4BF8, 0x48A0, 0x49B0, 0x7AA8, 0x4CA6, 0xA0, 0xA0}},
    {16481, 2, {0x0, 0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x7BDE, 0x4A02, 0x4A22, 0x7A22, 0x4A22, 0x4A52, 0x4A4A, 0x7A8A, 0x4A02, 0x20A, 0x204}},
    {16482, 2, {0x210, 0x110, 0xF110, 0x97BE, 0x9240, 0xF200, 0x93BE, 0x928A, 0xF288, 0x92A8, 0x92AE, 0x92A8, 0xF4A8, 0x95A8, 0x85E, 0x1080}},
    {16483, 2, {0x20, 0x7BFE, 0x4820, 0x49FC, 0x4924, 0x79FC, 0x4924, 0x49FC, 0x7822, 0x4BFE, 0x4808, 0x4BFE, 0x7908, 0x4888, 0x28, 0x10}},
    {16484, 2, {0x40, 0x20, 0x7BFE, 0x4888, 0x4850, 0x7BFE, 0x4A22, 0x4AFA, 0x7A22, 0x4AFA, 0x4A8A, 0x4A8A, 0x7AFA, 0x4A02, 0x20A, 0x204}},
    {16485, 2, {0x0, 0x3FE, 0x7A48, 0x4A48, 0x4BFE, 0x7A48, 0x4A20, 0x4BFE, 0x7A40, 0x4AFC, 0x4B44, 0x4A44, 0x7A7C, 0x4A00, 0x3FE, 0x0}},
    {16486, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x7A78, 0x4BC4, 0x4A3C, 0x7A00, 0x4AFC, 0x4A84, 0x4AFC, 0x7A84, 0x4CFC, 0x484, 0x9FE}},
    {16487, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48F8, 0x7820, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x49FC, 0x4820, 0x7BFE, 0x4850, 0x88, 0x306}},
    {16488, 2, {0x200, 0x11E, 0xF7D2, 0x9012, 0x97D4, 0xF454, 0x97D8, 0x9014, 0xF7D2, 0x9092, 0x9112, 0x91DA, 0xF714, 0x9110, 0x510, 0x210}},
    {16489, 2, {0x40, 0x40, 0x78F8, 0x4908, 0x4A10, 0x79FC, 0x4924, 0x4924, 0x79FC, 0x4924, 0x4924, 0x49FC, 0x7800, 0x4954, 0x12A, 0x22A}},
    {16490, 2, {0x200, 0x202, 0xE23C, 0xAFA0, 0xA220, 0xEFA0, 0xAABE, 0xAFA4, 0xEAA4, 0xAFA4, 0xA224, 0xAFA4, 0xE224, 0xA244, 0x244, 0x284}},
    {16491, 2, {0x0, 0x1FC, 0x7924, 0x49AC, 0x4974, 0x7924, 0x49FC, 0x4820, 0x79FC, 0x4820, 0x4BFE, 0x4800, 0x7AA4, 0x4A52, 0x452, 0x0}},
    {16492, 2, {0x104, 0x88, 0x7800, 0x4BFE, 0x4A22, 0x7AAA, 0x4A72, 0x4A22, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x104}},
    {16493, 2, {0x28, 0x24, 0x7BFE, 0x4850, 0x4888, 0x7904, 0x4A02, 0x4998, 0x7954, 0x4910, 0x4BBE, 0x4910, 0x7AA8, 0x4AA8, 0x444, 0x882}},
    {16494, 2, {0x0, 0x7BC, 0xF108, 0x9528, 0x97BC, 0xF318, 0x95AA, 0x9946, 0xF000, 0x93F8, 0x9208, 0x9208, 0xF3F8, 0x9208, 0x208, 0x3F8}},
    {16495, 2, {0x40, 0xA0, 0xF118, 0x96E6, 0x9000, 0xF3F8, 0x9208, 0x93F8, 0xF000, 0x97BC, 0x9084, 0x94A4, 0xF294, 0x94A4, 0x294, 0x108}},
    {16496, 2, {0x0, 0xFE, 0xF010, 0x9720, 0x927C, 0xF244, 0x9244, 0x927C, 0xF244, 0x927C, 0x93C4, 0x9E44, 0xF07C, 0x9028, 0x44, 0x82}},
    {16497, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4BFE, 0x7908, 0x4B9C, 0x4908, 0x7988, 0x4E3E, 0x4800, 0x4BFE, 0x7890, 0x4890, 0x112, 0x20E}},
    {16498, 2, {0x0, 0x3FE, 0x7A52, 0x4A52, 0x4BFE, 0x7800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104, 0x202}},
    {16499, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4824, 0x7BFE, 0x4850, 0x4888, 0x7904, 0x4A8A, 0x4888, 0x4BFE, 0x7888, 0x4888, 0x108, 0x208}},
    {16500, 2, {0x40, 0xFFE, 0xF000, 0x97FC, 0x9404, 0xF5F4, 0x9514, 0x97FC, 0xF000, 0x93F8, 0x9208, 0x93F8, 0xF208, 0x93F8, 0x0, 0xFFE}},
    {16501, 2, {0x0, 0x1FC, 0x7954, 0x4954, 0x49FC, 0x7800, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x4862, 0x7894, 0x4988, 0x2A4, 0xC2}},
    {16502, 2, {0x108, 0x208, 0xF7C8, 0x9450, 0x97DE, 0xF464, 0x97D4, 0x9214, 0xF114, 0x97D4, 0x9214, 0x93C8, 0xF248, 0x9454, 0x554, 0x8A2}},
    {16503, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16504, 2, {0x90, 0x3FC, 0xF294, 0x93FC, 0x9294, 0xF3FC, 0x9000, 0x93FC, 0xF200, 0x92F8, 0x9200, 0x93FE, 0xF520, 0x9514, 0x548, 0x986}},
    {16505, 2, {0x10, 0x3C8, 0x7A7E, 0x4A40, 0x4A62, 0x7BD4, 0x4A00, 0x4A3E, 0x7BC8, 0x4A48, 0x4E7E, 0x4A48, 0x7A48, 0x4BC8, 0x248, 0x8}},
    {16506, 2, {0x0, 0x1FC, 0x7954, 0x49FC, 0x4820, 0x79FC, 0x4820, 0x4BFE, 0x7888, 0x4850, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {16507, 2, {0x444, 0x4E4, 0xE8A8, 0xAAAA, 0xAEEE, 0xE4A4, 0xAAAA, 0xAEEE, 0xE242, 0xA040, 0xAFFE, 0xA0E0, 0xE150, 0xA248, 0xC46, 0x40}},
    {16508, 2, {0x1FC, 0x7924, 0x49AC, 0x4924, 0x49FC, 0x7820, 0x49FC, 0x4820, 0x7BFE, 0x4954, 0x4A2A, 0x4820, 0x79FC, 0x4820, 0x3FE, 0x0}},
    {16509, 2, {0x124, 0x248, 0x7924, 0x4800, 0x4BFC, 0x7A94, 0x4A64, 0x4A94, 0x7BFC, 0x4A48, 0x4B68, 0x4A48, 0x7B6A, 0x4A4A, 0x246, 0x362}},
    {16510, 2, {0x108, 0x7FE, 0x7908, 0x4800, 0x4BFC, 0x7A94, 0x4A94, 0x4BFC, 0x7828, 0x4BFE, 0x4A20, 0x4B24, 0x7AA8, 0x4A92, 0x42A, 0x8C6}},
    {16511, 2, {0x7F14, 0x6B12, 0x5D10, 0x7F7E, 0x810, 0x7F10, 0x828, 0xFFA8, 0x5544, 0x9FF2, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16512, 2, {0x110, 0x7FC, 0xF040, 0x93F8, 0x9040, 0xF7FC, 0x9000, 0x93D4, 0xF112, 0x97FE, 0x9350, 0x9534, 0xF38C, 0x908A, 0x296, 0x102}},
    {16513, 2, {0x1F0, 0x210, 0xF7FC, 0x9204, 0x93FC, 0xF224, 0x93B8, 0x9222, 0xF1FE, 0x9210, 0x97FC, 0x9244, 0xF3FC, 0x90D0, 0x14A, 0x63E}},
    {16514, 2, {0x7BC, 0x4A4, 0xF7BC, 0x94A4, 0x97BC, 0xF4A4, 0x97BC, 0x9120, 0xF3FE, 0x9220, 0x97FC, 0x9A20, 0xF3FC, 0x9220, 0x3FE, 0x200}},
    {16515, 2, {0x3C, 0x7C0, 0xF244, 0x9128, 0x97FC, 0xF4A4, 0x97FC, 0x9008, 0xF788, 0x94FE, 0x9788, 0x94A8, 0xF798, 0x9508, 0x5A8, 0x690}},
    {16516, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x78A0, 0x49FE, 0x4B20, 0x7DFC, 0x4920, 0x49FE, 0x4800, 0x79FC, 0x4888, 0x70, 0x78E}},
    {16517, 2, {0x444, 0x4E4, 0xEA0A, 0xAEEE, 0xA404, 0xEAEA, 0xAE0E, 0xA0E0, 0xEAAA, 0xAAEA, 0xA040, 0xA7FC, 0xE444, 0xA7FC, 0x44, 0xFFE}},
    {16518, 2, {0x40, 0x7C40, 0x440, 0x287C, 0x1084, 0x884, 0xFF04, 0x1A44, 0x2824, 0x2824, 0x4804, 0x4804, 0x8804, 0x804, 0x2828, 0x1010}},
    {16519, 2, {0x0, 0x7CFC, 0x424, 0x2824, 0x1024, 0x824, 0xFE24, 0x1AFC, 0x2844, 0x2844, 0x4844, 0x4844, 0x8844, 0x844, 0x29FE, 0x1000}},
    {16520, 2, {0x10, 0x7C10, 0x410, 0x28FE, 0x1092, 0x892, 0xFE92, 0x1A92, 0x28FE, 0x2892, 0x4810, 0x4814, 0x8812, 0x9FE, 0x2882, 0x1000}},
    {16521, 2, {0x80, 0xF8BC, 0x884, 0x5108, 0x21FE, 0x1320, 0xFD20, 0x357C, 0x3190, 0x5110, 0x51FE, 0x5110, 0x9128, 0x1128, 0x5144, 0x2182}},
    {16522, 2, {0x80, 0xF8F8, 0x908, 0x53FE, 0x2512, 0x1122, 0xFDFE, 0x3440, 0x30A2, 0x5354, 0x5098, 0x5334, 0x9054, 0x1092, 0x5350, 0x2020}},
    {16523, 2, {0x1000, 0x10FC, 0xFE48, 0x1030, 0x7DFE, 0x4452, 0x7C94, 0x1110, 0x1230, 0x7C20, 0x11FE, 0x5070, 0x7CA8, 0x1124, 0x1222, 0x1020}},
    {16524, 2, {0x20, 0xF9FC, 0x888, 0x5050, 0x23FE, 0x1000, 0xFDFC, 0x3524, 0x31FC, 0x5124, 0x51FC, 0x5020, 0x91FC, 0x1020, 0x53FE, 0x2000}},
    {16525, 2, {0x1FC, 0xF904, 0x9FC, 0x5104, 0x21FC, 0x1088, 0xFDFC, 0x3488, 0x33FE, 0x5088, 0x5124, 0x52AA, 0x9070, 0x10A8, 0x5124, 0x2060}},
    {16526, 2, {0x528, 0xFFBE, 0x948, 0x57BE, 0x2318, 0x15AA, 0xFD46, 0x37FC, 0x3204, 0x53FC, 0x5204, 0x53FC, 0x9204, 0x13FC, 0x5108, 0x2204}},
    {16527, 2, {0x2000, 0x21FC, 0x3C44, 0x5044, 0x9044, 0x1044, 0x1094, 0xFE88, 0x1100, 0x10FC, 0x1084, 0x2884, 0x2484, 0x4484, 0x40FC, 0x8084}},
    {16528, 2, {0x2020, 0x2020, 0x2124, 0x7D24, 0x5124, 0x9124, 0x11FC, 0xFC20, 0x1020, 0x1124, 0x1124, 0x2924, 0x2524, 0x4524, 0x41FC, 0x8004}},
    {16529, 2, {0x2048, 0x2048, 0x2048, 0x7D48, 0x514A, 0x916C, 0x1148, 0xFD48, 0x1148, 0x1148, 0x1148, 0x294A, 0x256A, 0x478A, 0x4106, 0x8000}},
    {16530, 2, {0x2020, 0x2020, 0x3C20, 0x53FE, 0x9020, 0x1020, 0x11FC, 0xFE00, 0x1000, 0x11FC, 0x1104, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {16531, 2, {0x2000, 0x21FC, 0x2104, 0x7D04, 0x51FC, 0x9104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x2850, 0x2490, 0x4492, 0x4112, 0x820E}},
    {16532, 2, {0x2020, 0x2020, 0x21FC, 0x7C20, 0x5020, 0x93FE, 0x1088, 0xFC50, 0x11FC, 0x1020, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x4020, 0x8020}},
    {16533, 2, {0x2080, 0x20FC, 0x2104, 0x7DF8, 0x5008, 0x93FE, 0x1040, 0xFCA2, 0x1334, 0x1058, 0x1094, 0x2B34, 0x2452, 0x4490, 0x4350, 0x8020}},
    {16534, 2, {0x0, 0xFFFE, 0x200, 0x210, 0x408, 0x404, 0x800, 0xFF8, 0x1808, 0x2808, 0x4808, 0x8808, 0x808, 0x808, 0xFF8, 0x808}},
    {16535, 2, {0x40, 0x40, 0xFE40, 0x1044, 0x1044, 0x2048, 0x3E50, 0x6260, 0x6240, 0xA2C0, 0x2340, 0x2244, 0x3E44, 0x2244, 0x3C, 0x0}},
    {16536, 2, {0x0, 0x3FC, 0xFD04, 0x1144, 0x1124, 0x20A8, 0x3C88, 0x6488, 0x6450, 0xA450, 0x2420, 0x2420, 0x3C50, 0x2488, 0x2104, 0x602}},
    {16537, 2, {0x0, 0x7C, 0xFE44, 0x1048, 0x1048, 0x2050, 0x3E48, 0x6248, 0x6244, 0xA244, 0x2244, 0x2268, 0x3E50, 0x2240, 0x40, 0x40}},
    {16538, 2, {0x8, 0x88, 0xFC50, 0x1020, 0x1050, 0x2088, 0x3D00, 0x6408, 0x6488, 0xA488, 0x2450, 0x2450, 0x3C20, 0x2450, 0x2088, 0x306}},
    {16539, 2, {0x0, 0x1F8, 0xFC08, 0x1050, 0x1020, 0x2010, 0x3DFE, 0x6422, 0x6424, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x20A0, 0x40}},
    {16540, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x2020, 0x3C20, 0x65FC, 0x6420, 0xA420, 0x2420, 0x27FE, 0x3C20, 0x2420, 0x2020, 0x20}},
    {16541, 2, {0x50, 0x48, 0xFC48, 0x1040, 0x105E, 0x21E0, 0x3C40, 0x6444, 0x6444, 0xA448, 0x2430, 0x2422, 0x3C52, 0x248A, 0x2306, 0x2}},
    {16542, 2, {0x0, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {16543, 2, {0x800, 0x800, 0x7F7C, 0x944, 0x1144, 0x1144, 0x257C, 0x4200, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {16544, 2, {0x40, 0x40, 0xFC80, 0x10FC, 0x1120, 0x2220, 0x3C20, 0x6420, 0x67FE, 0xA420, 0x2450, 0x2450, 0x3C88, 0x2488, 0x2104, 0x202}},
    {16545, 2, {0x8, 0x1C, 0xFCE0, 0x1080, 0x1080, 0x2080, 0x3C80, 0x64FE, 0x6488, 0xA488, 0x2488, 0x2488, 0x3C88, 0x2488, 0x21FE, 0x0}},
    {16546, 2, {0x40, 0x40, 0xFC7C, 0x1084, 0x1188, 0x2250, 0x3C20, 0x6450, 0x6488, 0xA706, 0x2460, 0x2410, 0x3C08, 0x24C0, 0x2020, 0x10}},
    {16547, 2, {0x90, 0x90, 0xF890, 0x2290, 0x2292, 0x42D4, 0x7A98, 0xCA90, 0x4A90, 0x4A90, 0x4A90, 0x4A92, 0x7AD2, 0x4712, 0x20E, 0x0}},
    {16548, 2, {0x88, 0x88, 0xFC88, 0x1108, 0x117E, 0x2308, 0x3D08, 0x6548, 0x6528, 0xA528, 0x2508, 0x2508, 0x3D08, 0x2508, 0x2128, 0x110}},
    {16549, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1104, 0x21FC, 0x3D20, 0x6520, 0x65FE, 0xA520, 0x2520, 0x2510, 0x3D12, 0x254A, 0x2186, 0x102}},
    {16550, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x2212, 0x3C20, 0x6440, 0x6588, 0xA410, 0x2420, 0x2444, 0x3D88, 0x2410, 0x2060, 0x380}},
    {16551, 2, {0x80, 0x80, 0xFC80, 0x11FE, 0x1102, 0x2204, 0x3C20, 0x6420, 0x6528, 0xA524, 0x2624, 0x2622, 0x3C22, 0x2420, 0x20A0, 0x40}},
    {16552, 2, {0x8, 0x1C, 0xFDE0, 0x1100, 0x1100, 0x2100, 0x3DFE, 0x6510, 0x6510, 0xA530, 0x2518, 0x2514, 0x3D12, 0x2210, 0x210, 0x410}},
    {16553, 2, {0x0, 0x1FC, 0xFD04, 0x1124, 0x1124, 0x2124, 0x3DFC, 0x6524, 0x6524, 0xA554, 0x254C, 0x258C, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {16554, 2, {0x90, 0x90, 0xFC90, 0x1292, 0x1194, 0x2098, 0x3C90, 0x6598, 0x6694, 0xA492, 0x2490, 0x2490, 0x3D12, 0x2512, 0x2212, 0x40E}},
    {16555, 2, {0x80, 0x80, 0xF8FC, 0x2108, 0x2290, 0x4060, 0x7998, 0x4E26, 0xC820, 0x49FC, 0x4820, 0x4A20, 0x7BFE, 0x4820, 0x20, 0x20}},
    {16556, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4400, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {16557, 2, {0x40, 0x20, 0xFC20, 0x13FE, 0x1000, 0x2088, 0x3D04, 0x6602, 0x6488, 0xA488, 0x2450, 0x2450, 0x3C20, 0x2450, 0x2088, 0x306}},
    {16558, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x800, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {16559, 2, {0x20, 0x10, 0xFC10, 0x11FE, 0x1100, 0x2110, 0x3D10, 0x6510, 0x657C, 0xA510, 0x2510, 0x2510, 0x3D10, 0x2210, 0x2FE, 0x400}},
    {16560, 2, {0x42, 0xE2, 0xFB82, 0x208A, 0x208A, 0x408A, 0x7BEA, 0x488A, 0xC98A, 0x49CA, 0x4AAA, 0x4A82, 0x7C82, 0x4882, 0x8A, 0x84}},
    {16561, 2, {0x104, 0x84, 0xFC88, 0x1010, 0x11FC, 0x2104, 0x3D04, 0x6504, 0x65FC, 0xA450, 0x2450, 0x2490, 0x3C92, 0x2512, 0x220E, 0x400}},
    {16562, 2, {0x820, 0x7E20, 0x8FC, 0xFF24, 0x1024, 0x2444, 0x7E54, 0x2288, 0x100, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {16563, 2, {0x20, 0x20, 0xFCA8, 0x10A8, 0x10A8, 0x2174, 0x3E22, 0x6420, 0x6420, 0xA5FC, 0x2420, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {16564, 2, {0x40, 0x20, 0xFDFE, 0x1000, 0x1000, 0x20FC, 0x3C00, 0x6400, 0x64FC, 0xA400, 0x2400, 0x24FC, 0x3C84, 0x2484, 0x20FC, 0x84}},
    {16565, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x13FE, 0x2000, 0x3C00, 0x65FC, 0x6504, 0xA524, 0x2524, 0x2524, 0x3D54, 0x2488, 0x2104, 0x602}},
    {16566, 2, {0x88, 0x88, 0xF888, 0x21C8, 0x209E, 0x408A, 0x788A, 0x4BEA, 0xC88A, 0x488A, 0x490A, 0x494A, 0x7BEA, 0x4912, 0x2A, 0x44}},
    {16567, 2, {0x1C, 0x1E0, 0xFC20, 0x1020, 0x13FE, 0x2020, 0x3CA8, 0x64AA, 0x67AC, 0xA4A8, 0x24AA, 0x25AA, 0x3EA6, 0x2420, 0x2020, 0x20}},
    {16568, 2, {0x14, 0x12, 0xF810, 0x27FE, 0x2010, 0x4010, 0x7BD2, 0xCA52, 0x4A52, 0x4A54, 0x4BD4, 0x4808, 0x78EA, 0x471A, 0x226, 0x42}},
    {16569, 2, {0x20, 0x40, 0xFCFE, 0x1124, 0x1298, 0x2050, 0x3C60, 0x6780, 0x6420, 0xA424, 0x24A4, 0x24A8, 0x3D30, 0x2450, 0x2088, 0x306}},
    {16570, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x11FC, 0x2124, 0x3D24, 0x65FC, 0x6420, 0xA7FE, 0x2470, 0x24A8, 0x3D24, 0x2222, 0x20, 0x20}},
    {16571, 2, {0x0, 0xFC, 0xFC80, 0x10F8, 0x1080, 0x20F8, 0x3C80, 0x67FE, 0x6540, 0xA524, 0x2528, 0x2510, 0x3D08, 0x2544, 0x2182, 0x100}},
    {16572, 2, {0x20, 0x20, 0xF850, 0x2088, 0x2144, 0x4222, 0x79F8, 0x4808, 0xC850, 0x4820, 0x48A4, 0x4A82, 0x7A8A, 0x4A8A, 0x478, 0x0}},
    {16573, 2, {0x20, 0x20, 0xFC3E, 0x1020, 0x11FC, 0x2104, 0x3DFC, 0x6504, 0x65FC, 0xA524, 0x2420, 0x27FE, 0x3C20, 0x2420, 0x2020, 0x20}},
    {16574, 2, {0x20, 0x20, 0xF820, 0x23FE, 0x2020, 0x4124, 0x7924, 0x4924, 0xCAAA, 0x4870, 0x48A8, 0x48A8, 0x7924, 0x4A22, 0x420, 0x20}},
    {16575, 2, {0x20, 0x22, 0xFBB4, 0x20A8, 0x20A8, 0x4124, 0x7AA2, 0x4840, 0xC9FC, 0x4904, 0x4904, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {16576, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x4024, 0x7BFE, 0x4824, 0xC9FC, 0x4840, 0x4BFE, 0x4884, 0x79C8, 0x4830, 0xCC, 0x302}},
    {16577, 2, {0x0, 0x1FE, 0xFC48, 0x1048, 0x1048, 0x21CE, 0x3D02, 0x6502, 0x6502, 0xA5CE, 0x2448, 0x2448, 0x3C48, 0x2448, 0x21FE, 0x0}},
    {16578, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x2104, 0x3D04, 0x65FC, 0x6400, 0xA512, 0x25D4, 0x2518, 0x3D10, 0x2552, 0x2192, 0x10E}},
    {16579, 2, {0x0, 0x3FE, 0xFA02, 0x228A, 0x2252, 0x43FE, 0x7A42, 0x4A22, 0xCBFE, 0x4A82, 0x4A82, 0x4A82, 0x7AFA, 0x4A02, 0x20A, 0x204}},
    {16580, 2, {0x40, 0x20, 0xFBFE, 0x2000, 0x2000, 0x41FC, 0x7904, 0x4904, 0xC904, 0x49FC, 0x4820, 0x48A8, 0x7924, 0x4A22, 0xA0, 0x40}},
    {16581, 2, {0x104, 0x88, 0xFC50, 0x11FC, 0x1124, 0x2124, 0x3DFC, 0x6524, 0x6524, 0xA5FC, 0x2420, 0x2420, 0x3DFE, 0x2420, 0x2020, 0x20}},
    {16582, 2, {0x84, 0x84, 0xFDFE, 0x1084, 0x1084, 0x20FC, 0x3C84, 0x6484, 0x64FC, 0xA484, 0x2484, 0x25FE, 0x3C00, 0x2448, 0x2084, 0x102}},
    {16583, 2, {0x20, 0x40, 0xFCFC, 0x1084, 0x10FC, 0x2084, 0x3CFC, 0x6420, 0x6420, 0xA5FE, 0x2522, 0x2522, 0x3D2A, 0x2524, 0x2020, 0x20}},
    {16584, 2, {0x110, 0x112, 0xFDD4, 0x1118, 0x1152, 0x2192, 0x3D2E, 0x6440, 0x65FC, 0xA504, 0x2504, 0x25FC, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {16585, 2, {0x0, 0x1FC, 0xFD04, 0x11FC, 0x1110, 0x21FE, 0x3D10, 0x654A, 0x6586, 0xA400, 0x25FC, 0x2504, 0x3DFC, 0x2504, 0x21FC, 0x104}},
    {16586, 2, {0x90, 0x90, 0xFBFC, 0x2090, 0x2000, 0x41F8, 0x7908, 0x4908, 0xC9F8, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x7FE, 0x0}},
    {16587, 2, {0x40, 0xA0, 0xF910, 0x2208, 0x25F6, 0x4000, 0x7BC4, 0x4A54, 0xCA54, 0x4BD4, 0x4A54, 0x4A54, 0x7BD4, 0x4A44, 0x254, 0x2C8}},
    {16588, 2, {0x0, 0x3FE, 0xFC50, 0x1050, 0x11FC, 0x2154, 0x3D54, 0x6554, 0x65FC, 0xA420, 0x2420, 0x25FC, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {16589, 2, {0x104, 0x88, 0xF800, 0x23FE, 0x2040, 0x40C2, 0x7922, 0xCA54, 0x4898, 0x4938, 0x4A54, 0x4894, 0x7912, 0x4610, 0x50, 0x20}},
    {16590, 2, {0x0, 0x40, 0xFBFE, 0x2088, 0x2104, 0x4642, 0x78FC, 0x4B08, 0xC8B0, 0x48D0, 0x4F3E, 0x48C2, 0x7B24, 0x4818, 0x60, 0x780}},
    {16591, 2, {0x0, 0x1FC, 0xFD04, 0x11FC, 0x1104, 0x21FC, 0x3C20, 0x6520, 0x65FE, 0xA620, 0x2420, 0x25FC, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {16592, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2088, 0x4104, 0x7A22, 0x4828, 0xC824, 0x4BFE, 0x4820, 0x4850, 0x7850, 0x4888, 0x104, 0x202}},
    {16593, 2, {0x100, 0x23C, 0xF7A4, 0x24A4, 0x26A4, 0x45A4, 0x74C2, 0xDF80, 0x54BC, 0x54A4, 0x56A4, 0x55A4, 0x7494, 0x5488, 0x494, 0x9A2}},
    {16594, 2, {0x0, 0x3FE, 0xF800, 0x21FC, 0x2104, 0x4104, 0x79FC, 0x4800, 0xCBFE, 0x4A8A, 0x4A52, 0x4BFE, 0x7A22, 0x4A22, 0x22A, 0x204}},
    {16595, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x21FC, 0x4088, 0x7BFE, 0x4820, 0xC9FC, 0x4924, 0x49FC, 0x4924, 0x7BFE, 0x4904, 0x114, 0x108}},
    {16596, 2, {0x20, 0x10, 0xFDFE, 0x1102, 0x1000, 0x20FC, 0x3C84, 0x6484, 0x64FC, 0xA400, 0x25FE, 0x2502, 0x3D02, 0x2502, 0x21FE, 0x102}},
    {16597, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {16598, 2, {0x40, 0x80, 0xF9FC, 0x2104, 0x2104, 0x41FC, 0x7900, 0xC9FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x7AAA, 0x4402, 0x14, 0x8}},
    {16599, 2, {0x2, 0x3E2, 0xF822, 0x21EA, 0x202A, 0x47FA, 0x788A, 0x4AAA, 0xC9CA, 0x488A, 0x49CA, 0x4AA2, 0x7C92, 0x4882, 0x28A, 0x104}},
    {16600, 2, {0x0, 0x3DC, 0xFA94, 0x23D4, 0x2254, 0x43C8, 0x7A88, 0x4A94, 0xCBE2, 0x4800, 0x4820, 0x49FC, 0x7820, 0x4820, 0x3FE, 0x0}},
    {16601, 2, {0x40, 0x20, 0xFBFE, 0x2250, 0x2250, 0x43FE, 0x7A52, 0xCA52, 0x4BFE, 0x4A00, 0x4A92, 0x4AD4, 0x7A98, 0x4492, 0x4D2, 0x88E}},
    {16602, 2, {0x420, 0x220, 0xF23E, 0x2F40, 0x24A0, 0x4420, 0x773C, 0xD550, 0x5510, 0x5510, 0x55FE, 0x5510, 0x7528, 0x5928, 0xB44, 0x1082}},
    {16603, 2, {0x10, 0x10, 0xFAFE, 0x2110, 0x217C, 0x4054, 0x787C, 0x4B54, 0xC97C, 0x4910, 0x49FE, 0x4910, 0x7910, 0x4A90, 0x47E, 0x0}},
    {16604, 2, {0x0, 0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {16605, 2, {0x208, 0x248, 0xF748, 0x2290, 0x229E, 0x4FE4, 0x7114, 0xD214, 0x5FD4, 0x5494, 0x5914, 0x51C8, 0x7F08, 0x5114, 0x514, 0x222}},
    {16606, 2, {0x840, 0x7F40, 0x22F8, 0x3E48, 0x48, 0x3EC8, 0x448, 0x7EAA, 0x8AA, 0x2906, 0x1202, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {16607, 2, {0x288, 0x288, 0xF7C8, 0x2290, 0x229E, 0x4FD4, 0x7024, 0xD7D4, 0x5454, 0x57D4, 0x5454, 0x57C8, 0x7448, 0x5454, 0x464, 0x4C2}},
    {16608, 2, {0x210, 0x110, 0xF110, 0x27BE, 0x2240, 0x4200, 0x73BE, 0xD28A, 0x5288, 0x52A8, 0x52AE, 0x52A8, 0x74A8, 0x55A8, 0x85E, 0x1080}},
    {16609, 2, {0x0, 0x3DE, 0xF888, 0x22A8, 0x23FE, 0x4188, 0x7ADA, 0x4CA6, 0xC9F8, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x1F8, 0x108}},
    {16610, 2, {0x210, 0x210, 0xF510, 0x24BE, 0x2822, 0x4744, 0x7210, 0xD210, 0x5F10, 0x5210, 0x5AA8, 0x5728, 0x7228, 0x5344, 0xC44, 0x82}},
    {16611, 2, {0xE, 0x3F0, 0xF844, 0x2224, 0x2108, 0x43F8, 0x7A08, 0xCBFC, 0x4A04, 0x4BFE, 0x4A02, 0x4AAA, 0x7AAA, 0x4502, 0x414, 0x808}},
    {16612, 2, {0x40, 0x3F8, 0xF048, 0x27FE, 0x2048, 0x43F8, 0x7040, 0xD554, 0x575C, 0x5444, 0x57FC, 0x5444, 0x775C, 0x5554, 0x554, 0x844}},
    {16613, 2, {0x20, 0x1FC, 0xFC88, 0x1050, 0x13FE, 0x2000, 0x3DFC, 0x6524, 0x65FC, 0xA524, 0x25FC, 0x2420, 0x3DFC, 0x2420, 0x23FE, 0x0}},
    {16614, 2, {0x840, 0x7F40, 0x227E, 0x3E88, 0x148, 0x3E48, 0x450, 0x7E20, 0x850, 0x288C, 0x1000, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {16615, 2, {0x80, 0x79F8, 0x4A88, 0x4870, 0x538E, 0x4920, 0x49FC, 0x6A20, 0x50F8, 0x4020, 0xFFFE, 0x400, 0xFF8, 0x3808, 0xCFF8, 0x808}},
    {16616, 2, {0x0, 0x3BE, 0xF8A2, 0x20A2, 0x20BE, 0x4388, 0x7A08, 0xCA3E, 0x4A2A, 0x4BAA, 0x48AA, 0x48BE, 0x7888, 0x408A, 0x57E, 0x202}},
    {16617, 2, {0x50, 0x1FC, 0xFD54, 0x11FC, 0x1154, 0x21FC, 0x3C00, 0x65FE, 0x6500, 0xA57C, 0x2500, 0x25FE, 0x3D50, 0x2554, 0x2148, 0x266}},
    {16618, 2, {0x20, 0x3FE, 0xF800, 0x23FC, 0x2204, 0x42F4, 0x7A94, 0x4BFC, 0xC800, 0x49F8, 0x4908, 0x49F8, 0x7908, 0x49F8, 0x0, 0x3FE}},
    {16619, 2, {0x100, 0x1F8, 0xFA08, 0x27FE, 0x2288, 0x4324, 0x7BFE, 0xCA00, 0x4AFC, 0x4A00, 0x4AFC, 0x4A00, 0x7AFC, 0x4484, 0x4FC, 0x884}},
    {16620, 2, {0x40, 0x248, 0xFA48, 0x23F8, 0x2080, 0x43F8, 0x7A48, 0x4BF8, 0xCA48, 0x4BF8, 0x4848, 0x48B4, 0x78BC, 0x4922, 0x222, 0x41E}},
    {16621, 2, {0x14, 0x12, 0xF7FE, 0x2410, 0x25D0, 0x4414, 0x75D4, 0xD558, 0x55CA, 0x5416, 0x5442, 0x5824, 0x72A2, 0x528A, 0x478, 0x0}},
    {16622, 2, {0x10, 0x410, 0xF27C, 0x2010, 0x20FE, 0x4044, 0x7628, 0xD2FE, 0x5210, 0x527C, 0x5210, 0x52FE, 0x7210, 0x5210, 0x5FE, 0x800}},
    {16623, 2, {0x20, 0x40, 0xF8F8, 0x228A, 0x218C, 0x40F8, 0x798C, 0x4A8A, 0xC8F8, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x4924, 0x622, 0x20}},
    {16624, 2, {0x80, 0x40, 0xF7FE, 0x2402, 0x2A24, 0x43BC, 0x74A4, 0xDAA8, 0x5510, 0x52E8, 0x5404, 0x5BFA, 0x7040, 0x5248, 0x444, 0xC0}},
    {16625, 2, {0x114, 0x112, 0xF7D2, 0x2110, 0x2110, 0x4FFE, 0x7290, 0xD6D2, 0x5292, 0x56D2, 0x5294, 0x56D4, 0x728A, 0x52CA, 0xF16, 0x22}},
    {16626, 2, {0x100, 0x3DC, 0xFA54, 0x2366, 0x2240, 0x47DC, 0x7A54, 0x4B48, 0xCA54, 0x4CA2, 0x4800, 0x4BFC, 0x7A94, 0x4A94, 0x7FE, 0x0}},
    {16627, 2, {0x124, 0x248, 0xF924, 0x2000, 0x23FC, 0x4294, 0x7A64, 0x4A94, 0xCBFC, 0x4A48, 0x4B68, 0x4A48, 0x7B6A, 0x4A4A, 0x246, 0x362}},
    {16628, 2, {0x410, 0x210, 0xF07A, 0x2712, 0x2014, 0x477E, 0x7008, 0xD710, 0x503C, 0x5764, 0x55A4, 0x553C, 0x7524, 0x5724, 0x53C, 0x24}},
    {16629, 2, {0x40, 0x7FE, 0xF000, 0x23FC, 0x2240, 0x43F8, 0x7240, 0xD3F8, 0x5240, 0x53FC, 0x5004, 0x5554, 0x70A8, 0x5310, 0xD48, 0x186}},
    {16630, 2, {0x40, 0x20, 0xFDFE, 0x1000, 0x11FC, 0x2154, 0x3DFC, 0x6420, 0x6522, 0xA48C, 0x2522, 0x2454, 0x3CC8, 0x2544, 0x2252, 0x60}},
    {16631, 2, {0x40, 0xFFE, 0xF400, 0x27FC, 0x2000, 0x47FC, 0x7404, 0xD7FC, 0x5080, 0x5C9C, 0x57D4, 0x5D54, 0x755C, 0x5C96, 0x1556, 0x2E22}},
    {16632, 2, {0x148, 0x14C, 0xFAAA, 0x2008, 0x27FE, 0x4148, 0x7B68, 0x494A, 0xCB6A, 0x494C, 0x4B6C, 0x4948, 0x796A, 0x4B9A, 0x26, 0x42}},
    {16633, 2, {0x3DE, 0x252, 0xFBDE, 0x2252, 0x23DE, 0x4222, 0x7AFA, 0x4A22, 0xCAFA, 0x4AAA, 0x4AFA, 0x4AAA, 0x7AFA, 0x4A72, 0x2AA, 0x226}},
    {16634, 2, {0x20, 0x7FE, 0xF488, 0x27DE, 0x2488, 0x45DC, 0x76AA, 0xD488, 0x5450, 0x57DE, 0x5450, 0x55DC, 0x7450, 0x5BDE, 0x850, 0x1050}},
    {16635, 2, {0x3FE, 0x20, 0xF7FE, 0x2422, 0x21AC, 0x4280, 0x77DE, 0xD292, 0x5392, 0x511E, 0x57D2, 0x5552, 0x77DE, 0x5112, 0xFD2, 0x126}},
    {16636, 2, {0x1020, 0x820, 0x820, 0x7E20, 0x420, 0x828, 0x824, 0x1C22, 0x2A22, 0x4A20, 0x8820, 0x820, 0x820, 0x820, 0x820, 0x820}},
    {16637, 2, {0x2020, 0x1020, 0x1020, 0xFBFE, 0x820, 0x1020, 0x1020, 0x39FC, 0x5484, 0x9488, 0x1048, 0x1050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {16638, 2, {0x2010, 0x1110, 0x1110, 0xF912, 0x912, 0x1114, 0x11D8, 0x3910, 0x5510, 0x9510, 0x1110, 0x1112, 0x1152, 0x1192, 0x110E, 0x1000}},
    {16639, 2, {0x2020, 0x1020, 0x1020, 0xF820, 0x9FC, 0x1020, 0x1020, 0x3820, 0x55FE, 0x9420, 0x1050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {16640, 2, {0x2090, 0x1088, 0x1088, 0xF880, 0x8BC, 0x13C0, 0x1080, 0x3888, 0x5488, 0x9490, 0x1060, 0x1044, 0x10A4, 0x1114, 0x120C, 0x1004}},
    {16641, 2, {0x2048, 0x1148, 0x1148, 0xF948, 0x948, 0x13FE, 0x1148, 0x3948, 0x5548, 0x9548, 0x1178, 0x1100, 0x1100, 0x1100, 0x11FE, 0x1000}},
    {16642, 2, {0x2020, 0x1020, 0x1020, 0xF820, 0x9FC, 0x1124, 0x1124, 0x3924, 0x5524, 0x95FC, 0x1124, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {16643, 2, {0x2020, 0x1020, 0x1020, 0xF9FC, 0x924, 0x1124, 0x1924, 0x3524, 0x5124, 0x93FE, 0x1020, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {16644, 2, {0x2000, 0x11FC, 0x1104, 0xF924, 0x924, 0x1124, 0x11FC, 0x3924, 0x5524, 0x9554, 0x114C, 0x118C, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16645, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC006, 0x3FF8, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {16646, 2, {0x2008, 0x103C, 0x11E0, 0xF820, 0x820, 0x13FE, 0x1020, 0x3820, 0x5420, 0x95FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16647, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x904, 0x11FC, 0x1000, 0x3800, 0x57FE, 0x9420, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {16648, 2, {0x2000, 0x13FC, 0x1000, 0xF800, 0x9F8, 0x1108, 0x1108, 0x3908, 0x55F8, 0x9400, 0x1108, 0x1088, 0x1090, 0x1000, 0x17FE, 0x1000}},
    {16649, 2, {0x2010, 0x10D8, 0x1394, 0xF894, 0x890, 0x13FE, 0x1090, 0x3894, 0x5494, 0x90D8, 0x1398, 0x1090, 0x10AA, 0x10CA, 0x1286, 0x1102}},
    {16650, 2, {0x2020, 0x1020, 0x1020, 0xFBFE, 0x820, 0x1020, 0x1020, 0x39FC, 0x5440, 0x9420, 0x10A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {16651, 2, {0x2020, 0x1020, 0x13FE, 0xF850, 0x888, 0x1124, 0x1222, 0x39FC, 0x5524, 0x95FC, 0x1124, 0x11FC, 0x1022, 0x1022, 0x101E, 0x1000}},
    {16652, 2, {0x4000, 0x23DE, 0x2042, 0xF94A, 0x1084, 0x114A, 0x2252, 0x3420, 0x6800, 0xABDE, 0x2052, 0x2152, 0x2094, 0x2148, 0x2254, 0x2422}},
    {16653, 2, {0x2088, 0x1088, 0x1088, 0xFBFE, 0x888, 0x1088, 0x17FE, 0x3800, 0x55FC, 0x9504, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16654, 2, {0x2020, 0x1020, 0x11FC, 0xF850, 0x888, 0x1104, 0x13FE, 0x3808, 0x55E8, 0x9528, 0x1128, 0x11E8, 0x1128, 0x1008, 0x1028, 0x1010}},
    {16655, 2, {0x2088, 0x1088, 0x13FE, 0xF888, 0x800, 0x101C, 0x19E0, 0x3420, 0x5020, 0x93FE, 0x1020, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {16656, 2, {0x1000, 0x5400, 0x54FC, 0x7C84, 0x9284, 0x9284, 0xFE84, 0xFC, 0x7C80, 0x80, 0xFE80, 0x1082, 0x5482, 0x9282, 0x507E, 0x2000}},
    {16657, 2, {0x201C, 0x13E0, 0x1220, 0xFBFE, 0xA20, 0x1292, 0x130A, 0x3A06, 0x55FC, 0x9504, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16658, 2, {0x2020, 0x1020, 0x1050, 0xF888, 0x944, 0x1222, 0x11F8, 0x3808, 0x5450, 0x9420, 0x10A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {16659, 2, {0x2020, 0x1040, 0x11FC, 0xF904, 0x9FC, 0x1104, 0x11FC, 0x3800, 0x57FE, 0x9420, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {16660, 2, {0x2004, 0x101E, 0x13E0, 0xF820, 0x9FC, 0x1124, 0x1924, 0x35FC, 0x5020, 0x93FE, 0x1222, 0x122A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {16661, 2, {0x2020, 0x1022, 0x13B4, 0xF8A8, 0x8A8, 0x1124, 0x12A2, 0x3840, 0x55FC, 0x9524, 0x1124, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {16662, 2, {0x4040, 0x20A0, 0x2110, 0xFA08, 0x15F6, 0x1000, 0x23C4, 0x3254, 0x6A54, 0xABD4, 0x2254, 0x2254, 0x23D4, 0x2244, 0x2254, 0x22C8}},
    {16663, 2, {0x4050, 0x2252, 0x2252, 0xFA52, 0x13DE, 0x1202, 0x2202, 0x33FE, 0x6A02, 0xAA02, 0x23DE, 0x2252, 0x2252, 0x2252, 0x2452, 0x2802}},
    {16664, 2, {0x2020, 0x1010, 0x11FE, 0xF884, 0x848, 0x11FE, 0x1102, 0x3A24, 0x5410, 0x95FE, 0x1040, 0x107C, 0x1044, 0x1084, 0x1094, 0x1108}},
    {16665, 2, {0x2000, 0x13FE, 0x1202, 0xF800, 0x9FC, 0x1104, 0x11FC, 0x3904, 0x55FC, 0x9440, 0x1020, 0x13FE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {16666, 2, {0x4090, 0x2090, 0x27FE, 0xF890, 0x13FC, 0x1294, 0x23FC, 0x3294, 0x6BFC, 0xA800, 0x21F8, 0x2108, 0x21F8, 0x2108, 0x21F8, 0x2108}},
    {16667, 2, {0x2020, 0x11FC, 0x1124, 0xFBFE, 0x924, 0x11FC, 0x1020, 0x39FC, 0x5524, 0x95FC, 0x1040, 0x13FE, 0x1088, 0x11D0, 0x1070, 0x138C}},
    {16668, 2, {0x2040, 0x1020, 0x13FE, 0xF800, 0x954, 0x1124, 0x1954, 0x35FC, 0x5020, 0x93FE, 0x1242, 0x1292, 0x12FA, 0x120A, 0x1202, 0x1206}},
    {16669, 2, {0x2020, 0x1020, 0x13FE, 0xF820, 0x9FC, 0x1040, 0x1BFE, 0x3488, 0x5104, 0x9242, 0x159C, 0x1104, 0x11DC, 0x1104, 0x11FC, 0x1104}},
    {16670, 2, {0x4100, 0x2110, 0x23DC, 0xFA54, 0x1554, 0x1288, 0x2108, 0x32F4, 0x6C02, 0xA800, 0x23FC, 0x2040, 0x2150, 0x2248, 0x2544, 0x2080}},
    {16671, 2, {0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE, 0xC20, 0x703C, 0x1DE2, 0xF022, 0x1C1E, 0x0, 0xFFFE, 0x1110, 0x2308}},
    {16672, 2, {0x2020, 0x13FE, 0x1000, 0xFBFC, 0xA04, 0x12F4, 0x1294, 0x3BFC, 0x5400, 0x95F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x1000, 0x13FE}},
    {16673, 2, {0x2080, 0x10F8, 0x1108, 0xFBFE, 0x944, 0x1192, 0x11FE, 0x3900, 0x557C, 0x9500, 0x117C, 0x1100, 0x117C, 0x1244, 0x127C, 0x1444}},
    {16674, 2, {0x4080, 0x2040, 0x27FC, 0xF910, 0x10A4, 0x1F58, 0x2554, 0x3552, 0x6B58, 0xA800, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2408}},
    {16675, 2, {0x2020, 0x13FE, 0x1020, 0xF9FC, 0x800, 0x13FE, 0x1252, 0x3BFE, 0x5504, 0x95FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {16676, 2, {0x4108, 0x2108, 0x27CE, 0xF912, 0x1124, 0x17DE, 0x2552, 0x355E, 0x6FD2, 0xA91E, 0x2392, 0x255E, 0x2940, 0x2114, 0x2112, 0x2122}},
    {16677, 2, {0x47FC, 0x2040, 0x2FFE, 0xF842, 0x1358, 0x1040, 0x2358, 0x3000, 0x6EEE, 0xAAAA, 0x2EEE, 0x2000, 0x2FFE, 0x2248, 0x2554, 0x2FFE}},
    {16678, 2, {0x400, 0xEFC, 0x7804, 0x808, 0x810, 0xFF20, 0x820, 0x1820, 0x1C20, 0x2A20, 0x2A20, 0x4820, 0x8820, 0x820, 0x8A0, 0x840}},
    {16679, 2, {0x400, 0xE00, 0x79FC, 0x844, 0x848, 0xFF48, 0x850, 0x185E, 0x1C42, 0x2A42, 0x2A82, 0x4882, 0x8882, 0x902, 0x914, 0xA08}},
    {16680, 2, {0x800, 0x1DFC, 0xF020, 0x1020, 0x1020, 0xFC20, 0x1020, 0x33FE, 0x3820, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {16681, 2, {0x420, 0xE28, 0x7824, 0x824, 0x820, 0xFF2E, 0x8F0, 0x1820, 0x1C20, 0x2A20, 0x2A20, 0x4810, 0x8812, 0x80A, 0x806, 0x802}},
    {16682, 2, {0x840, 0x1C40, 0xF080, 0x10FC, 0x1104, 0xFE04, 0x1004, 0x3104, 0x3884, 0x5444, 0x5444, 0x9004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {16683, 2, {0x400, 0xE00, 0x78FC, 0x804, 0x804, 0xFF04, 0x804, 0x18FC, 0x1C84, 0x2A80, 0x2A80, 0x4880, 0x8882, 0x882, 0x87E, 0x800}},
    {16684, 2, {0x820, 0x1C20, 0xF020, 0x1120, 0x112C, 0xFD34, 0x1164, 0x33A4, 0x3924, 0x5534, 0x5528, 0x9122, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {16685, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x1020, 0xFC20, 0x1020, 0x33FE, 0x3820, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {16686, 2, {0x820, 0x1C20, 0xF020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x3820, 0x35FE, 0x5420, 0x5050, 0x9050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {16687, 2, {0x400, 0xEFC, 0x7848, 0x848, 0x848, 0xFF48, 0x848, 0x19FE, 0x1C48, 0x2A48, 0x2A48, 0x4848, 0x8848, 0x848, 0x888, 0x908}},
    {16688, 2, {0x800, 0x1DFE, 0xF008, 0x1088, 0x1088, 0xFC88, 0x1108, 0x31FE, 0x3818, 0x5428, 0x5448, 0x9088, 0x1108, 0x1208, 0x1028, 0x1010}},
    {16689, 2, {0x840, 0x1C20, 0xF020, 0x13FE, 0x1080, 0xFC80, 0x1080, 0x30FC, 0x3884, 0x5484, 0x5484, 0x9084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {16690, 2, {0x820, 0x1C20, 0xF020, 0x11FE, 0x1122, 0xFD22, 0x1122, 0x3122, 0x3952, 0x554A, 0x558A, 0x9102, 0x1102, 0x1102, 0x110A, 0x1104}},
    {16691, 2, {0x820, 0x1C20, 0xF020, 0x1120, 0x1120, 0xFD20, 0x113C, 0x3120, 0x3920, 0x5520, 0x5520, 0x9120, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {16692, 2, {0x800, 0x1CFC, 0xF084, 0x1084, 0x1084, 0xFCFC, 0x1084, 0x3084, 0x3884, 0x54FC, 0x5484, 0x9084, 0x1104, 0x1104, 0x1214, 0x1408}},
    {16693, 2, {0x70, 0x1F80, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC006, 0x100, 0x1FF0, 0x1100, 0x1100, 0xFFFE, 0x100, 0x100}},
    {16694, 2, {0x890, 0x1C90, 0xF090, 0x13FC, 0x1094, 0xFC94, 0x13FC, 0x3A90, 0x3690, 0x53FE, 0x5092, 0x9092, 0x111A, 0x1114, 0x1210, 0x1410}},
    {16695, 2, {0x800, 0x1C00, 0xF3FE, 0x1020, 0x1020, 0xFC40, 0x1040, 0x38FC, 0x3584, 0x5284, 0x5484, 0x9084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {16696, 2, {0x820, 0x1C20, 0xF040, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x3104, 0x39FC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16697, 2, {0xF8, 0x3F00, 0x100, 0x7FFC, 0x920, 0x1110, 0x2288, 0xCC64, 0x3018, 0xC1E6, 0x1F00, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {16698, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x11FC, 0xFC24, 0x11FC, 0x3120, 0x39FE, 0x5422, 0x542A, 0x9054, 0x1050, 0x1088, 0x1104, 0x1202}},
    {16699, 2, {0x890, 0x1C90, 0xF090, 0x1292, 0x1194, 0xFC98, 0x1090, 0x3198, 0x3A94, 0x5492, 0x5490, 0x9090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {16700, 2, {0x800, 0x1DFE, 0xF020, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x3124, 0x3954, 0x554C, 0x558C, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16701, 2, {0x800, 0x1DFC, 0xF050, 0x1050, 0x1050, 0xFDFC, 0x1154, 0x3154, 0x3954, 0x5554, 0x555C, 0x9184, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16702, 2, {0x828, 0x1C24, 0xF024, 0x1020, 0x13FE, 0xFC20, 0x1124, 0x3124, 0x3924, 0x57A8, 0x5528, 0x9110, 0x1212, 0x122A, 0x1446, 0x1082}},
    {16703, 2, {0x820, 0x1C20, 0xF1FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5420, 0x5424, 0x9018, 0x1032, 0x10CA, 0x1306, 0x1002}},
    {16704, 2, {0x802, 0x1C02, 0xF3E2, 0x110A, 0x110A, 0xFDEA, 0x112A, 0x3A2A, 0x372A, 0x52AA, 0x544A, 0x904A, 0x1082, 0x1102, 0x120A, 0x1404}},
    {16705, 2, {0x100, 0x7FFC, 0x4004, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38, 0x1E0, 0x1F00, 0x100, 0x7FFC, 0x540, 0x1930, 0x610C}},
    {16706, 2, {0x880, 0x1C80, 0xF0F8, 0x1108, 0x1310, 0xFCA0, 0x1040, 0x38A0, 0x3518, 0x5206, 0x55F8, 0x9108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {16707, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4400, 0xE0, 0x1F00, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {16708, 2, {0x840, 0x1C40, 0xF1FC, 0x1044, 0x1084, 0xFC84, 0x1128, 0x3210, 0x3908, 0x57DE, 0x554A, 0x914A, 0x114A, 0x1252, 0x135A, 0x14A4}},
    {16709, 2, {0x820, 0x1C20, 0xF1FC, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x3000, 0x3820, 0x5420, 0x55FC, 0x9020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {16710, 2, {0x810, 0x1C10, 0xF210, 0x113E, 0x1122, 0xFC44, 0x1010, 0x3890, 0x3490, 0x5310, 0x5128, 0x9128, 0x1128, 0x1144, 0x1044, 0x1082}},
    {16711, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0xE0, 0x1F00, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {16712, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x10D0, 0x2F08, 0xC106, 0x1FF0, 0x540, 0x920, 0x1110, 0x6108, 0x100}},
    {16713, 2, {0x800, 0x1DFE, 0xF048, 0x1048, 0x1048, 0xFC48, 0x1248, 0x314A, 0x394A, 0x554C, 0x5448, 0x9048, 0x1048, 0x1048, 0x13FE, 0x1000}},
    {16714, 2, {0x840, 0x1C20, 0xF3FE, 0x1200, 0x1200, 0xFE0C, 0x12F0, 0x3220, 0x3A20, 0x563E, 0x53E0, 0x9220, 0x1422, 0x1422, 0x181E, 0x1000}},
    {16715, 2, {0x880, 0x1C80, 0xF0F8, 0x1108, 0x1210, 0xFDFC, 0x1124, 0x3124, 0x3924, 0x55FC, 0x5450, 0x9050, 0x1090, 0x1092, 0x1112, 0x120E}},
    {16716, 2, {0x800, 0x1CFC, 0xF084, 0x1084, 0x10FC, 0xFC00, 0x11FE, 0x3102, 0x3902, 0x55FE, 0x5502, 0x9102, 0x11FE, 0x1102, 0x110A, 0x1104}},
    {16717, 2, {0x800, 0x1C06, 0xF3B8, 0x1088, 0x1088, 0xFD08, 0x113E, 0x3B88, 0x3488, 0x5288, 0x5288, 0x913E, 0x1100, 0x1280, 0x147E, 0x1800}},
    {16718, 2, {0x81C, 0x1DE0, 0xF020, 0x1020, 0x13FE, 0xFCA8, 0x1124, 0x3222, 0x39F8, 0x5488, 0x5490, 0x90BE, 0x1102, 0x1102, 0x1214, 0x1408}},
    {16719, 2, {0x800, 0x1DFC, 0xF104, 0x1104, 0x11FC, 0xFD00, 0x1118, 0x31E0, 0x3938, 0x55E0, 0x553C, 0x91E0, 0x1122, 0x1222, 0x121E, 0x1400}},
    {16720, 2, {0x800, 0x1DF8, 0xF108, 0x1108, 0x11F8, 0xFD08, 0x1108, 0x31F8, 0x3908, 0x5508, 0x55F8, 0x9090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {16721, 2, {0x808, 0x1C3C, 0xF3C0, 0x1044, 0x1224, 0xFD28, 0x1100, 0x3040, 0x3BFE, 0x5488, 0x5108, 0x9390, 0x1060, 0x1050, 0x1188, 0x1604}},
    {16722, 2, {0x890, 0x1C90, 0xF3FE, 0x1090, 0x1090, 0xFC40, 0x1020, 0x33FE, 0x3900, 0x5500, 0x5500, 0x9100, 0x1100, 0x1100, 0x11F8, 0x1000}},
    {16723, 2, {0x1200, 0x3A00, 0xE23C, 0x27D4, 0x2454, 0xF854, 0x2754, 0x3554, 0x6D54, 0x6554, 0xA748, 0x2548, 0x2048, 0x2054, 0x2294, 0x2122}},
    {16724, 2, {0x892, 0x1C92, 0xF124, 0x1248, 0x1124, 0xFC92, 0x1092, 0x3000, 0x39FE, 0x5522, 0x5522, 0x91FE, 0x1122, 0x1122, 0x11FE, 0x1102}},
    {16725, 2, {0x800, 0x1DFE, 0xF102, 0x1102, 0x11FE, 0xFD10, 0x1110, 0x31FE, 0x3910, 0x5510, 0x557E, 0x9142, 0x1142, 0x1242, 0x127E, 0x1442}},
    {16726, 2, {0x820, 0x1C20, 0xF3FE, 0x1050, 0x1088, 0xFD24, 0x1222, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x91FC, 0x1022, 0x1022, 0x101E, 0x1000}},
    {16727, 2, {0x81C, 0x1DE0, 0xF020, 0x1020, 0x13FE, 0xFCA8, 0x1124, 0x3A42, 0x3440, 0x53FE, 0x5088, 0x9108, 0x10D0, 0x1030, 0x1048, 0x1184}},
    {16728, 2, {0x820, 0x1C20, 0xF020, 0x13FE, 0x1020, 0xFD24, 0x1124, 0x3124, 0x3AAA, 0x5470, 0x54A8, 0x90A8, 0x1124, 0x1222, 0x1420, 0x1020}},
    {16729, 2, {0x800, 0x1DFC, 0xF104, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x31FC, 0x3800, 0x5512, 0x55D4, 0x9118, 0x1110, 0x1152, 0x1192, 0x110E}},
    {16730, 2, {0x820, 0x1D24, 0xF0A8, 0x1020, 0x11FC, 0xFC40, 0x13FE, 0x3888, 0x3504, 0x52FA, 0x5488, 0x9088, 0x10A8, 0x1092, 0x1082, 0x107E}},
    {16731, 2, {0x8FC, 0x1C84, 0xF084, 0x10FC, 0x1084, 0xFC84, 0x10FC, 0x3000, 0x39FE, 0x5502, 0x5502, 0x91FE, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {16732, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x13FE, 0xFCA8, 0x10A8, 0x30A8, 0x3BFE, 0x54A8, 0x50A8, 0x93FE, 0x1020, 0x1020, 0x11FC, 0x1000}},
    {16733, 2, {0x840, 0x1C20, 0xF020, 0x13FE, 0x1202, 0xFC94, 0x1108, 0x3204, 0x3800, 0x55FC, 0x5420, 0x9020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {16734, 2, {0x800, 0x1DF8, 0xF108, 0x11F8, 0x1108, 0xFDF8, 0x1000, 0x31FC, 0x3810, 0x57FE, 0x5410, 0x9110, 0x1090, 0x1010, 0x1050, 0x1020}},
    {16735, 2, {0x40, 0x1848, 0x7084, 0x11FE, 0x1020, 0xFBFE, 0x1088, 0x3124, 0x3A42, 0x5588, 0x5010, 0x9062, 0x1184, 0x1018, 0x1060, 0x1380}},
    {16736, 2, {0x800, 0x1DFC, 0xF104, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x39FC, 0x3400, 0x53FE, 0x5020, 0x9120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {16737, 2, {0x800, 0x1DFE, 0xF020, 0x1040, 0x11FC, 0xFD04, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x9104, 0x11FC, 0x1000, 0x1088, 0x1104}},
    {16738, 2, {0x820, 0x1C20, 0xF3FE, 0x1222, 0x13FE, 0xFE22, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {16739, 2, {0x820, 0x1C40, 0xF1FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3000, 0x3BFE, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {16740, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x30A0, 0x392C, 0x5524, 0x5524, 0x91AC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {16741, 2, {0x800, 0x1DFC, 0xF104, 0x11FC, 0x1104, 0xFDFC, 0x1080, 0x31FE, 0x3A22, 0x5522, 0x5552, 0x9102, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {16742, 2, {0x888, 0x1C88, 0xF3FE, 0x1088, 0x1088, 0xFC00, 0x11FC, 0x3124, 0x3924, 0x5524, 0x55FC, 0x9124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {16743, 2, {0x840, 0x1C20, 0xF3FC, 0x1000, 0x1108, 0xFC90, 0x13FE, 0x3000, 0x39F8, 0x5508, 0x5508, 0x91F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {16744, 2, {0x81C, 0x1DE0, 0xF020, 0x1020, 0x13FE, 0xFCA8, 0x1124, 0x3222, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {16745, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x11FC, 0xFC88, 0x1050, 0x3020, 0x3858, 0x5586, 0x5400, 0x91FE, 0x1102, 0x11FE, 0x1084, 0x1102}},
    {16746, 2, {0x820, 0x1C20, 0xF1FC, 0x1124, 0x11FC, 0xFC20, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x5524, 0x9124, 0x1124, 0x1050, 0x1088, 0x1304}},
    {16747, 2, {0x840, 0x1C20, 0xF3FE, 0x1000, 0x1000, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5420, 0x90A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {16748, 2, {0x904, 0x1C84, 0xF088, 0x1000, 0x13FE, 0xFC00, 0x1088, 0x3104, 0x3A02, 0x55FC, 0x5554, 0x9154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {16749, 2, {0x820, 0x1C10, 0xF1FE, 0x1084, 0x1048, 0xFDFE, 0x1102, 0x3224, 0x3810, 0x55FE, 0x5440, 0x907C, 0x1044, 0x1084, 0x1094, 0x1108}},
    {16750, 2, {0x820, 0x1C20, 0xF050, 0x1048, 0x10A4, 0xFDFE, 0x1284, 0x30FC, 0x3884, 0x54FC, 0x5480, 0x90FC, 0x1144, 0x1144, 0x127C, 0x1044}},
    {16751, 2, {0x820, 0x1C10, 0xF1FE, 0x1110, 0x117C, 0xFD14, 0x11FE, 0x3114, 0x397C, 0x5510, 0x557C, 0x9144, 0x1144, 0x1244, 0x127C, 0x1444}},
    {16752, 2, {0x800, 0x1DFC, 0xF124, 0x1124, 0x11FC, 0xFD24, 0x1154, 0x318C, 0x3904, 0x55FC, 0x5440, 0x9020, 0x10A4, 0x128A, 0x128A, 0x1478}},
    {16753, 2, {0x820, 0x1D24, 0xF124, 0x1124, 0x11FC, 0xFC00, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1088, 0x1050, 0x13FE, 0x1000}},
    {16754, 2, {0x840, 0x1C44, 0xF1F8, 0x1050, 0x13FE, 0xFC40, 0x10F8, 0x3182, 0x3A7E, 0x5400, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {16755, 2, {0x880, 0x1C80, 0xF1FC, 0x1244, 0x1154, 0xFDF4, 0x1084, 0x3128, 0x3A90, 0x55FC, 0x5244, 0x9554, 0x11F4, 0x1084, 0x1128, 0x1210}},
    {16756, 2, {0x20, 0x1820, 0x73FE, 0x1020, 0x13FE, 0xFA42, 0x148C, 0x31F0, 0x3820, 0x5444, 0x53FE, 0x9022, 0x1128, 0x1224, 0x14A4, 0x1040}},
    {16757, 2, {0x904, 0x1C84, 0xF088, 0x1000, 0x13FE, 0xFC20, 0x1020, 0x39FC, 0x3420, 0x5420, 0x53FE, 0x9000, 0x12A4, 0x1252, 0x1452, 0x1000}},
    {16758, 2, {0x800, 0x1DFE, 0xF100, 0x117E, 0x1100, 0xFDFE, 0x1154, 0x3148, 0x3964, 0x5542, 0x5504, 0x91FE, 0x1144, 0x1224, 0x1204, 0x140C}},
    {16759, 2, {0x888, 0x1C88, 0xF3FE, 0x1088, 0x10A8, 0xFC50, 0x1088, 0x3174, 0x3A22, 0x5420, 0x55FC, 0x9020, 0x10A8, 0x1124, 0x12A2, 0x1040}},
    {16760, 2, {0x888, 0x1C88, 0xF3FE, 0x1088, 0x1040, 0xFC20, 0x11FC, 0x3888, 0x3450, 0x53FE, 0x5020, 0x9020, 0x11FC, 0x1020, 0x1020, 0x1020}},
    {16761, 2, {0x820, 0x1DFC, 0xF124, 0x13FE, 0x1124, 0xFDFC, 0x1020, 0x31FC, 0x3924, 0x55FC, 0x5040, 0x93FE, 0x1088, 0x11D0, 0x1070, 0x138C}},
    {16762, 2, {0x800, 0x1DFC, 0xF050, 0x11FC, 0x1154, 0xFD54, 0x11FC, 0x3000, 0x39FC, 0x5400, 0x57FE, 0x9020, 0x10A8, 0x1124, 0x12A2, 0x1040}},
    {16763, 2, {0x840, 0x1C20, 0xF3FE, 0x1000, 0x1154, 0xFD24, 0x1154, 0x39FC, 0x3420, 0x53FE, 0x5242, 0x9292, 0x12FA, 0x120A, 0x1202, 0x1206}},
    {16764, 2, {0x800, 0x1DF8, 0xF108, 0x11F8, 0x1108, 0xFDF8, 0x1000, 0x3BFC, 0x3694, 0x53FC, 0x5000, 0x91F8, 0x1090, 0x1060, 0x1198, 0x1606}},
    {16765, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x81E4, 0x1F00, 0x100, 0x7FFC, 0x540, 0x1930, 0x610C}},
    {16766, 2, {0x0, 0x1BDE, 0x7088, 0x12A8, 0x13FE, 0xF988, 0x12DA, 0x34A6, 0x39F8, 0x5508, 0x5108, 0x91F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {16767, 2, {0x888, 0x1FFE, 0xF088, 0x1000, 0x11FC, 0xFCA8, 0x10A8, 0x33FE, 0x38A8, 0x54A8, 0x55FC, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {16768, 2, {0x800, 0x1DFC, 0xF124, 0x11AC, 0x1174, 0xFD24, 0x11FC, 0x3820, 0x35FC, 0x5020, 0x53FE, 0x9000, 0x12A4, 0x1252, 0x1452, 0x1000}},
    {16769, 2, {0x800, 0x1DFC, 0xF154, 0x11FC, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3888, 0x5450, 0x55FC, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {16770, 2, {0x1020, 0x3BFE, 0xE288, 0x2250, 0x23FE, 0xFA50, 0x22FC, 0x3254, 0x6BFE, 0x6254, 0xA2FC, 0x2250, 0x22D8, 0x2554, 0x2652, 0x2850}},
    {16771, 2, {0x2110, 0x17FC, 0x40, 0x73F8, 0x1208, 0x13F8, 0x1208, 0x2FFE, 0x4000, 0x1E0, 0x1F00, 0x100, 0x7FFC, 0x540, 0x1930, 0x610C}},
    {16772, 2, {0x810, 0x1DFE, 0xF000, 0x11FE, 0x1102, 0xFD7A, 0x114A, 0x39FE, 0x3400, 0x54FC, 0x5084, 0x90FC, 0x1084, 0x10FC, 0x1000, 0x11FE}},
    {16773, 2, {0xA20, 0x1D3E, 0xF042, 0x1094, 0x1310, 0xFD28, 0x1046, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {16774, 2, {0x8F8, 0x1C88, 0xF088, 0x10F8, 0x1000, 0xFDDC, 0x1154, 0x3154, 0x39DC, 0x5420, 0x57FE, 0x9070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {16775, 2, {0x1000, 0xE7FC, 0x2484, 0x25E4, 0x2524, 0xFDE4, 0x2524, 0x35E4, 0x6D04, 0x65F4, 0xA504, 0x25F4, 0x2414, 0x2556, 0x2556, 0x2832}},
    {16776, 2, {0x1000, 0xE4FE, 0x2282, 0x22FE, 0x2080, 0xF8AA, 0x2E9C, 0x32AA, 0x6A80, 0x62A8, 0xA2BE, 0x22C8, 0x22BE, 0x2288, 0x2508, 0x28FE}},
    {16777, 2, {0x1000, 0x3BFC, 0xE294, 0x2294, 0x23FC, 0xF910, 0x2252, 0x33DC, 0x6812, 0x63CE, 0xA240, 0x23D2, 0x225C, 0x23D0, 0x2252, 0x22CE}},
    {16778, 2, {0x94, 0x1B98, 0x7092, 0x118E, 0x1280, 0xF9FC, 0x1124, 0x31FC, 0x3924, 0x55FC, 0x5088, 0x91FC, 0x1088, 0x13FE, 0x1088, 0x1104}},
    {16779, 2, {0x1080, 0x39F8, 0xE208, 0x27FE, 0x2252, 0xFBDE, 0x2050, 0x335E, 0x6952, 0x67DE, 0xA152, 0x235E, 0x2150, 0x27D2, 0x2152, 0x233E}},
    {16780, 2, {0x394, 0x1910, 0x77BE, 0x1118, 0x13AA, 0xFD46, 0x1020, 0x33FE, 0x3800, 0x57FC, 0x5000, 0x93FC, 0x1000, 0x13FC, 0x1204, 0x13FC}},
    {16781, 2, {0x1210, 0x3910, 0xE7DE, 0x2010, 0x245E, 0xFA82, 0x27DE, 0x3010, 0x6FDE, 0x6450, 0xA7DE, 0x2450, 0x27DE, 0x2450, 0x2452, 0x24CE}},
    {16782, 2, {0x1148, 0x394C, 0xE2AA, 0x2008, 0x27FE, 0xF948, 0x2368, 0x314A, 0x6B6A, 0x614C, 0xA36C, 0x2148, 0x216A, 0x239A, 0x2026, 0x2042}},
    {16783, 2, {0x924, 0x1CA8, 0xF3FE, 0x10A8, 0x1326, 0xFDFC, 0x1124, 0x31FC, 0x3924, 0x55FC, 0x5488, 0x93FE, 0x1088, 0x17FE, 0x1104, 0x1202}},
    {16784, 2, {0x100, 0x1FF0, 0x200, 0x7FFC, 0x820, 0x37D8, 0xC106, 0x3FF8, 0x1110, 0x7D7C, 0x2020, 0xFEFE, 0x5454, 0xBABA, 0x1010, 0x5454}},
    {16785, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2008, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {16786, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1210, 0x2208, 0x1FE0, 0x220, 0xA20, 0x420, 0x620, 0x922, 0x1022, 0x201E, 0x4000}},
    {16787, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1110, 0x2208, 0x440, 0x820, 0x1FF0, 0x450, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {16788, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x2108, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100}},
    {16789, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x100, 0x3FF0, 0x110, 0x110, 0x110, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {16790, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2208, 0x200, 0x7FFC, 0x400, 0x480, 0x880, 0x1100, 0x2220, 0x4410, 0x8FF8, 0x408}},
    {16791, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x0, 0x7C88, 0x4488, 0x4488, 0x4488, 0x4498, 0x7CE8, 0x4488, 0x8, 0x8}},
    {16792, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x100, 0x100}},
    {16793, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {16794, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2018}},
    {16795, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x18, 0x79E0, 0x4840, 0x4840, 0x487E, 0x4BC0, 0x4840, 0x7844, 0x4844, 0x3C}},
    {16796, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9414, 0x400, 0x7FFC, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0x850, 0x820}},
    {16797, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2108, 0x200, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {16798, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x10F0, 0x1F00, 0x1004, 0x1004, 0xFFC}},
    {16799, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x2288, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {16800, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x3FFC, 0x2100, 0x2100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x0}},
    {16801, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9214, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1108, 0x1090, 0x1260, 0x1418, 0x1806}},
    {16802, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x1040, 0xFDF8, 0x1048, 0x1448, 0x1848, 0x33FE, 0xD040, 0x10A0, 0x1110, 0x5208, 0x2406}},
    {16803, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2908, 0x2890, 0x4860, 0x4A18, 0x8C06}},
    {16804, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1012, 0x1FF4, 0x1018, 0x7FF0, 0x190, 0xE10, 0xF050, 0x20}},
    {16805, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x7FFC, 0x200, 0xD08, 0x7190, 0x2A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {16806, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x910, 0x920, 0x12C0, 0x430, 0x1908, 0x6914, 0x910, 0x12A0, 0x440, 0x1830, 0x600C}},
    {16807, 2, {0x100, 0x7FFE, 0x4822, 0x9074, 0x1F80, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x400, 0x7FFC, 0x820, 0x1C40, 0x380, 0x7C78}},
    {16808, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0x7FFC, 0x440, 0x1830, 0xE00E}},
    {16809, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x440, 0x1830, 0xE00E}},
    {16810, 2, {0x1040, 0x1020, 0x95FC, 0x5504, 0x5850, 0x1088, 0xFD04, 0x2800, 0x28F8, 0x2820, 0x2820, 0x2820, 0x29FC, 0x4800, 0x4802, 0x87FE}},
    {16811, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x1FF8, 0x2A48, 0x4448, 0xA88, 0x1108, 0x22A8, 0x410, 0x100, 0x4884, 0x4812, 0x87F2}},
    {16812, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0xFE0, 0x820, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {16813, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x7F7C, 0x810, 0x7F10, 0x4910, 0x7F10, 0x49FE, 0x7F10, 0x810, 0xFF90, 0x810, 0x810}},
    {16814, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x23F8, 0x1008, 0x11F8, 0x8008, 0x43F8, 0x5000, 0x17FC, 0x2404, 0xE3F8, 0x2110, 0x20E0, 0x271C}},
    {16815, 2, {0x100, 0x7FFE, 0x4822, 0xBEFC, 0x1248, 0xC30, 0x32C8, 0x0, 0x3EF8, 0x1248, 0xC30, 0x32C8, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {16816, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x3FF8, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2F78, 0x2948, 0x2F78, 0x2000, 0x3FFC}},
    {16817, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2040, 0x1248, 0x90B0, 0x4108, 0x4604, 0x1040, 0x2244, 0xE448, 0x20A0, 0x2318, 0x2C06}},
    {16818, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x3FF8, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x24C4, 0x22A8, 0x4498, 0x5886, 0x8180}},
    {16819, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x23FC, 0x1080, 0x1144, 0x668, 0x70B0, 0x1128, 0x1668, 0x10A4, 0x1124, 0x16A0, 0x2840, 0x47FE}},
    {16820, 2, {0x100, 0x7FFE, 0x4822, 0x9FD4, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {16821, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x820, 0x7FFC, 0x5114, 0x4924, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16822, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x2488, 0x2244, 0x4244}},
    {16823, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x1FF0, 0x220, 0x7FFC, 0x488, 0x7A80, 0x100, 0x3FF8, 0x2828, 0x37D8, 0x2448, 0x27C8, 0x2018}},
    {16824, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9094, 0x7EA0, 0x444, 0x2828, 0x17D0, 0x2008, 0xDFF6, 0x1010, 0x1FF0, 0x820, 0x440, 0x7FFC}},
    {16825, 2, {0x100, 0x7FFE, 0x4922, 0x9FF4, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {16826, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0xF8, 0x3F00, 0x1110, 0x920, 0xFFFE, 0x920, 0x3018, 0xDFF6, 0x1110, 0x1FF0, 0x1110, 0x1FF0}},
    {16827, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x7C20, 0x420, 0xFF3E, 0x2244, 0x3EA4, 0x2228, 0x3E28, 0x2390, 0xFE28, 0x4244, 0x282}},
    {16828, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x21F8, 0x1108, 0x11E8, 0x128, 0x73FC, 0x1204, 0x12F4, 0x1294, 0x12F4, 0x1214, 0x2A08, 0x47FE}},
    {16829, 2, {0x100, 0x7FFE, 0x4822, 0x91FC, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x2FF8, 0x2300, 0x2C90, 0x2360, 0x5CD8, 0x4346, 0x9CC0}},
    {16830, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {16831, 2, {0x100, 0x7FFE, 0x4822, 0x9054, 0x9FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x904, 0x9FC, 0xF820, 0x4BFE, 0x48A8, 0x4924, 0x8A22}},
    {16832, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x2200, 0xFF7E, 0x2224, 0x3E3C, 0x824, 0x7F24, 0x493C, 0x7F24, 0x826, 0xFF7C, 0x804, 0x804}},
    {16833, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x2044, 0x3E28, 0x44FE, 0xFF10, 0x4910, 0x7F7C, 0x4910, 0x7F10, 0xFE, 0x5510, 0x4A90, 0x8A90}},
    {16834, 2, {0x200, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x820, 0x440, 0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {16835, 2, {0x1000, 0x800, 0x8FC, 0x7E84, 0x84, 0x484, 0x4484, 0x44FC, 0x2484, 0x2880, 0x2880, 0xE82, 0xF082, 0x4082, 0x7E, 0x0}},
    {16836, 2, {0x1040, 0x840, 0x840, 0x7EFE, 0x80, 0x500, 0x447C, 0x4408, 0x2410, 0x2820, 0x2840, 0xE80, 0xF102, 0x4102, 0xFE, 0x0}},
    {16837, 2, {0x1050, 0x848, 0x848, 0x7E40, 0x5E, 0x5E0, 0x4440, 0x4444, 0x2444, 0x2848, 0x2830, 0xE22, 0xF052, 0x408A, 0x306, 0x2}},
    {16838, 2, {0x1008, 0x888, 0x848, 0x7E48, 0x8, 0x488, 0x4448, 0x4448, 0x2408, 0x280E, 0x29F8, 0xE08, 0xF008, 0x4008, 0x8, 0x8}},
    {16839, 2, {0x2000, 0x11FC, 0x1104, 0xFD04, 0x104, 0x9FC, 0x8920, 0x8920, 0x49FE, 0x4920, 0x5120, 0x5110, 0x1D12, 0xE14A, 0x4186, 0x102}},
    {16840, 2, {0x2000, 0x11FC, 0x1084, 0xFC88, 0x50, 0x820, 0x88D8, 0x8B06, 0x4820, 0x4820, 0x51FC, 0x5020, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {16841, 2, {0x2000, 0x1000, 0x13FE, 0xFC20, 0x20, 0x840, 0x8840, 0x88FC, 0x4984, 0x4A84, 0x5484, 0x5084, 0x1C84, 0xE084, 0x40FC, 0x84}},
    {16842, 2, {0x2020, 0x1010, 0x1010, 0xFDFE, 0x100, 0x910, 0x8910, 0x8910, 0x4910, 0x4920, 0x5128, 0x5124, 0x1E44, 0xE2FE, 0x4442, 0x800}},
    {16843, 2, {0x1020, 0x820, 0x87C, 0x7E84, 0x148, 0x430, 0x4428, 0x4448, 0x259E, 0x2822, 0x2842, 0xEA4, 0xF018, 0x4010, 0x60, 0x180}},
    {16844, 2, {0x2020, 0x1124, 0x10A4, 0xFCA8, 0x20, 0x9FC, 0x8904, 0x8904, 0x49FC, 0x4904, 0x5104, 0x51FC, 0x1D04, 0xE104, 0x4114, 0x108}},
    {16845, 2, {0x2000, 0x11FC, 0x1104, 0xFD04, 0x104, 0x9FC, 0x8820, 0x8820, 0x4920, 0x493C, 0x5120, 0x5120, 0x1EA0, 0xE260, 0x443E, 0x800}},
    {16846, 2, {0x2048, 0x1048, 0x1048, 0xFDFE, 0x48, 0x848, 0x89FE, 0x8800, 0x48FC, 0x4884, 0x5084, 0x50FC, 0x1C84, 0xE084, 0x40FC, 0x84}},
    {16847, 2, {0x2040, 0x1020, 0x13FE, 0xFC00, 0x1FC, 0x904, 0x89FC, 0x8800, 0x49FC, 0x4808, 0x5010, 0x53FE, 0x1C20, 0xE020, 0x40A0, 0x40}},
    {16848, 2, {0x2020, 0x1020, 0x11FC, 0xFC24, 0x24, 0xBFE, 0x8824, 0x8824, 0x49FC, 0x4A22, 0x5174, 0x50A8, 0x1D24, 0xE222, 0x40A0, 0x40}},
    {16849, 2, {0x2040, 0x1080, 0x11FC, 0xFD24, 0x124, 0x9FC, 0x8924, 0x8944, 0x49FC, 0x4890, 0x5110, 0x53FE, 0x1C10, 0xE010, 0x4010, 0x10}},
    {16850, 2, {0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {16851, 2, {0xBFE, 0x1040, 0x61FC, 0x904, 0x11FC, 0x6104, 0x5FC, 0x904, 0x31FC, 0xC088, 0x304, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE}},
    {16852, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x200, 0x711C, 0x57D4, 0x7014, 0x545C, 0x7296, 0x51D6, 0xB622}},
    {16853, 2, {0x0, 0x21FC, 0x1020, 0x13FE, 0xFA22, 0x1AC, 0x820, 0x89AC, 0x8800, 0x4BFE, 0x5020, 0x51FC, 0x1D54, 0xE154, 0x4154, 0x10C}},
    {16854, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x800, 0x1FF8, 0x2108, 0x4108, 0x208, 0x208, 0x408, 0x1808, 0x6050, 0x20}},
    {16855, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x440, 0x820, 0x1010, 0x2108, 0xC206, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {16856, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x7FFC, 0x400, 0xFF0, 0x810, 0x820, 0x1020, 0x1FC0, 0x40, 0xFFFE, 0x0}},
    {16857, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7F08, 0x108, 0x108, 0x3F08, 0x2008, 0x4008, 0x7F08, 0x108, 0x108, 0xA08, 0x408}},
    {16858, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x840, 0x840, 0x1040, 0x3050, 0x5048, 0x9044, 0x1044, 0x1040, 0x1040, 0x1040}},
    {16859, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {16860, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8808, 0x800, 0xFFC, 0x1004, 0x2108, 0x4100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {16861, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8040, 0x420, 0x400, 0x4FE, 0xFF00, 0x210, 0x160, 0x184, 0xE44, 0x7034, 0xC}},
    {16862, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1010, 0x1010, 0xFFFE, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {16863, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9020, 0x1020, 0x3E20, 0x22FC, 0x4224, 0xA424, 0x1444, 0x844, 0x1084, 0x2114, 0x4208}},
    {16864, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8200, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {16865, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {16866, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8100, 0x100, 0xFFFE, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {16867, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x280, 0x440, 0x920, 0x3118, 0xC106, 0x100, 0x920, 0x1110, 0x2108, 0x500, 0x200}},
    {16868, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x4104, 0x4284, 0x4444, 0x4824, 0x4004, 0x4014, 0x4008}},
    {16869, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {16870, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8100, 0x1FF0, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {16871, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x2008}},
    {16872, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE10E, 0x1FF0, 0x100, 0x1100, 0x1FF8, 0x100, 0x100}},
    {16873, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0xB0, 0x1C2, 0xE32, 0x700E}},
    {16874, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FFC, 0x400, 0x820, 0x1FF0, 0x110, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {16875, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FFC, 0x2080, 0x2080, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x2080, 0x2080, 0x3FFE, 0x0}},
    {16876, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1100, 0x1100, 0x3FF8, 0x4100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {16877, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x200, 0x208, 0x3FD0, 0x220, 0xFFFE, 0x100, 0x610, 0x8E0, 0x3F00, 0xC808, 0x808, 0x7F8}},
    {16878, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x810, 0x878, 0x13C0, 0x3040, 0x5040, 0x9FFE, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000}},
    {16879, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x103E, 0xFFD0, 0x10, 0x10}},
    {16880, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3118, 0xC506, 0x200, 0x3FF8, 0x0}},
    {16881, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x2208, 0x23C8, 0x2448, 0x2A88, 0x2108, 0x2288, 0x2408, 0x3FF8, 0x2008}},
    {16882, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x3FF8, 0x100, 0x1110, 0x920, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {16883, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4548, 0x8220, 0x3F8, 0x7E00, 0x3F0, 0x3E00, 0x3FC, 0xFE20, 0x140, 0x84, 0x764, 0x781C}},
    {16884, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0xFFFE, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2848, 0x3038, 0x2008, 0x3FF8, 0x2008}},
    {16885, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x200C, 0x11F0, 0x1100, 0x8100, 0x49FE, 0x4910, 0x1110, 0xE110, 0x2210, 0x2210, 0x2410}},
    {16886, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x1FF0, 0x2020, 0x5FF8, 0x1108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x2108, 0x2128, 0x4010}},
    {16887, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3F04, 0x2104, 0x2124, 0x3F24, 0x824, 0x824, 0x7FA4, 0x8A4, 0x10A4, 0x1084, 0x2294, 0x4108}},
    {16888, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0}},
    {16889, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x408, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {16890, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {16891, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2008, 0x3FF8, 0x2000, 0x2070, 0x2F80, 0x20F0, 0x2F80, 0x20F8, 0x5F82, 0x4082, 0x807E}},
    {16892, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {16893, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9508, 0x100C, 0x11F0, 0xFD00, 0x1100, 0x11FE, 0x1D10, 0xF110, 0x1110, 0x1210, 0x5210, 0x2410}},
    {16894, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {16895, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2388, 0x560, 0x1918, 0xE106, 0x100}},
    {16896, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3F84, 0x2A84, 0x2A94, 0x2A94, 0xFFD4, 0x2A94, 0x2A94, 0x2A94, 0x2A84, 0x2094, 0x2188}},
    {16897, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x2488, 0x2488, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {16898, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1010, 0x820, 0x7FFC, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020}},
    {16899, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x23FC, 0x1084, 0x1084, 0x8084, 0x4914, 0x4A08, 0x15FC, 0xE104, 0x2104, 0x2104, 0x21FC, 0x2104}},
    {16900, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9FE0, 0x4044, 0x5194, 0x4924, 0x43C4, 0x4544, 0x4924, 0x5514, 0x4204, 0x7FFC, 0x4}},
    {16901, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8528, 0x90, 0x7FFC, 0x80, 0x3E88, 0x2288, 0x2250, 0x3E50, 0x24, 0x754, 0x788C, 0x2104}},
    {16902, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x20A0, 0x2090, 0x2FFC, 0x2120, 0x4210, 0x4408, 0x9806}},
    {16903, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x1FC, 0x100, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {16904, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3EFC, 0x22A4, 0x22A4, 0x3EA4, 0x22FC, 0x2280, 0x3E80, 0x2282, 0x2282, 0x4A7E, 0x8400}},
    {16905, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x3FF8, 0x820, 0xFFFE, 0x400, 0x7FFC, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {16906, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1010, 0x2208, 0x47E4, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {16907, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x104, 0x7D98, 0x960, 0x3118, 0xC506, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {16908, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x9F8, 0x908, 0x4108, 0x21F8, 0x2908, 0x908, 0x11F8, 0x7108, 0x1108, 0x17FE, 0x1000}},
    {16909, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0xFFFE, 0x1000, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16910, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1220, 0x7FFC, 0x1220, 0x13E0, 0x1000, 0x1FF8, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {16911, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {16912, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9548, 0x1040, 0x13FE, 0x5880, 0x5490, 0x5110, 0x9152, 0x1254, 0x14A8, 0x1028, 0x1044, 0x1082}},
    {16913, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1F90, 0x1090, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {16914, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2448, 0x2288, 0x2FE8, 0x2108, 0x3FF8, 0x2108, 0x2128, 0x2010}},
    {16915, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2104, 0x5284, 0x1454, 0x1FE8}},
    {16916, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7CF8, 0x4408, 0x4408, 0x7CF8, 0x4000, 0x41F8, 0x7C88, 0x4050, 0x4020, 0x7C50, 0x4088, 0x4306}},
    {16917, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1100, 0x11F8, 0x2900, 0x47FE, 0x8000}},
    {16918, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xC10, 0x7090, 0x1050, 0x1010, 0xFC90, 0x3050, 0x381E, 0x55F0, 0x5010, 0x9010, 0x1010, 0x1010}},
    {16919, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x910, 0xA08, 0x17FC, 0x3110, 0x5208, 0x9484, 0x11F8, 0x1308, 0x1490, 0x1060, 0x1198, 0x1606}},
    {16920, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x440, 0xFFFE, 0x1020, 0x1020, 0x2448, 0x7CF8, 0x810, 0x1020, 0x2448, 0x7EFC, 0x204}},
    {16921, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x1008, 0x2408, 0x79FE, 0x1008, 0x2488, 0x7E48, 0x48, 0x5408, 0x4A28, 0x8A10}},
    {16922, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x1010, 0x3FF8, 0x8, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {16923, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0xF7C, 0x844, 0x844, 0x7F44, 0x5528, 0x5528, 0x7F10, 0x5510, 0x5528, 0x5544, 0x4382}},
    {16924, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xBF8, 0x808, 0x11F8, 0x3008, 0x53F8, 0x9000, 0x17FC, 0x1404, 0x13F8, 0x1110, 0x10E0, 0x171C}},
    {16925, 2, {0x2040, 0x3E7E, 0x4890, 0xBFFC, 0x104, 0x11F0, 0x1100, 0x2FFE, 0x4000, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {16926, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8}},
    {16927, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8808, 0x3F04, 0x2124, 0x3F24, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x1204, 0x2114, 0x4088}},
    {16928, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xFC, 0x7F00, 0x2208, 0x1110, 0x3FF8, 0x200, 0xFFFE, 0x400, 0xFF0, 0x1220, 0x21C0, 0xCE3C}},
    {16929, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x1FF8, 0x2A48, 0x4448, 0xA88, 0x1108, 0x22A8, 0x410, 0x100, 0x4884, 0x4812, 0x87F2}},
    {16930, 2, {0x2040, 0x3E7E, 0x4890, 0x9000, 0x2008, 0x7FFC, 0x4, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {16931, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x20, 0x7F20, 0x43E, 0x444, 0x27A4, 0x2428, 0x2428, 0x2410, 0x2F28, 0xF044, 0x4082}},
    {16932, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1088, 0x1088, 0xFC88, 0x13FE, 0x3088, 0x3888, 0x54F8, 0x5488, 0x9088, 0x1088, 0x10F8, 0x1088}},
    {16933, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1010, 0x1F10, 0x2210, 0x7FBE, 0xA492, 0x3F92, 0x2492, 0x3F92, 0x2492, 0x24A2, 0x42AA, 0x8144}},
    {16934, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x1010, 0x2208, 0x47E4, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {16935, 2, {0x2040, 0x3E7E, 0x4890, 0x80F8, 0x7F00, 0x2210, 0x1120, 0x1FE0, 0x1020, 0x1FF0, 0x1010, 0x1FFC, 0x2004, 0x2AA4, 0x4AA4, 0x800C}},
    {16936, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7DF8, 0x4508, 0x49F8, 0x5108, 0x49F8, 0x4524, 0x4528, 0x5510, 0x4948, 0x4186, 0x4100}},
    {16937, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x804, 0x1404, 0x2224, 0x5D24, 0x8024, 0x1124, 0x924, 0x4A24, 0x27A4, 0xF804, 0x4014, 0x8}},
    {16938, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4824, 0x4444, 0x5FF4, 0x4104, 0x410C}},
    {16939, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1020, 0x1020, 0x25FC, 0x7D24, 0x924, 0x1154, 0x254C, 0x7D84, 0x104, 0x5504, 0x5514, 0x8108}},
    {16940, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x4044, 0x27FE, 0x2040, 0x87FC, 0x4444, 0x57FC, 0x1444, 0xE7FC, 0x2444, 0x2444, 0x2454, 0x408}},
    {16941, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x400, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x290, 0x4A8, 0x8FA, 0x308A, 0xC07E}},
    {16942, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x1080, 0x10FC, 0xFD04, 0x2F4, 0x4894, 0x4894, 0x48F4, 0x1094, 0x1E04, 0xF028, 0x4010}},
    {16943, 2, {0x2080, 0x3EFE, 0x4910, 0xBFF8, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x1010}},
    {16944, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1020, 0x1124, 0x10A8, 0xFC20, 0x11FC, 0x1104, 0x1DFC, 0xF104, 0x51FC, 0x1104, 0x5114, 0x2108}},
    {16945, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0xFC, 0x7F00, 0x2208, 0x1110, 0xC20, 0x3078, 0x2008, 0x3C78, 0x2008, 0x3FF8, 0x2008}},
    {16946, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1078, 0x3E48, 0x2248, 0x2A86, 0x2300, 0xFEFC, 0x2244, 0x2A44, 0x2228, 0x4210, 0x4A28, 0x84C6}},
    {16947, 2, {0x2040, 0x3F7E, 0x4890, 0x9FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x0, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x6AD4, 0x408}},
    {16948, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1020, 0x840, 0x7FFC, 0x480, 0x3FF0, 0x490, 0xFFFE, 0x490, 0x3FF0, 0xCA0, 0x3498, 0xC486}},
    {16949, 2, {0x2040, 0x3F7E, 0x4890, 0x8548, 0x820, 0x3218, 0xC446, 0xFE0, 0x20, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x2A54, 0x4488}},
    {16950, 2, {0x1040, 0x1040, 0x3F7E, 0x4A90, 0x8508, 0x7FFE, 0x4002, 0x8924, 0x1290, 0x2448, 0x820, 0x3FF8, 0xC826, 0x820, 0xFE0, 0x820}},
    {16951, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0x1080, 0x10F8, 0xFD08, 0x1290, 0x3060, 0x3998, 0x5606, 0x51F8, 0x9108, 0x1108, 0x11F8, 0x1108}},
    {16952, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1004, 0x1F24, 0x2214, 0x7F84, 0xA4A4, 0x3F94, 0x2486, 0x3FBC, 0x2484, 0x2484, 0x4284, 0x8104}},
    {16953, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFE, 0x4802, 0x8BF4, 0x1080, 0x33F8, 0x5208, 0x9208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {16954, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x10FC, 0x2488, 0x7D50, 0x820, 0x10D8, 0x2706, 0x7C60, 0x10, 0x54C0, 0x5430, 0x8008}},
    {16955, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x28, 0x7824, 0x4BFE, 0x4820, 0x79FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x4924, 0x990C}},
    {16956, 2, {0x2040, 0x3F7E, 0x5090, 0x8A08, 0x100, 0x7FFE, 0x4202, 0x8924, 0x2848, 0x4994, 0xE10, 0x77F0, 0x100, 0x2108, 0x2108, 0x3FF8}},
    {16957, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x7FFC, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE}},
    {16958, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1020, 0x1124, 0xFCA8, 0x1020, 0x31FC, 0x3904, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {16959, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x4004, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {16960, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x3FFC, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x23E0, 0x2000, 0x4A48, 0x4924, 0x9124}},
    {16961, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x10, 0xFF90, 0x2228, 0x3E44, 0x2292, 0x3E08, 0x227C, 0x2384, 0xFE08, 0x230, 0x208, 0x204}},
    {16962, 2, {0x2040, 0x3E7E, 0x4890, 0x9FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x400, 0x820, 0x1FC0, 0x310, 0x3FF8, 0x1120, 0x2510, 0x4208}},
    {16963, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x448, 0x7DFE, 0x2488, 0x2448, 0x2448, 0x4408, 0x8418}},
    {16964, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7840, 0x4BF8, 0x5040, 0x67FC, 0x5110, 0x4A88, 0x4CF4, 0x6910, 0x52A0, 0x4040, 0x41B0, 0x4E0E}},
    {16965, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9548, 0x1040, 0x7C40, 0x11F8, 0xFE48, 0x4448, 0x28C8, 0x7C48, 0x10AA, 0xFEAA, 0x1106, 0x1202}},
    {16966, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4924, 0x8494, 0x8}},
    {16967, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0x1104, 0x7FC4, 0x1124, 0x1F24, 0x1124, 0x1F24, 0x1124, 0x7FE4, 0x2924, 0x3084, 0x2014, 0x3FC8}},
    {16968, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2420, 0x7E20, 0x243E, 0xFF44, 0xA4, 0x7E24, 0x4228, 0x7E28, 0x4210, 0x7E28, 0x4244, 0x4682}},
    {16969, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2000, 0x3E1C, 0x44E8, 0x7EA8, 0xAAA8, 0x3EA8, 0x2AA8, 0x3EA8, 0x2AA4, 0x2B54, 0x2B74, 0x4612}},
    {16970, 2, {0x2080, 0x3EFE, 0x5510, 0x8820, 0x1FF0, 0x210, 0xFFFE, 0x920, 0x3FF8, 0xD116, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x100}},
    {16971, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x11FC, 0x1124, 0xFDFC, 0x2500, 0x257C, 0x4944, 0x297C, 0x1144, 0x297C, 0x4644, 0x827C, 0x444}},
    {16972, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7890, 0x4BFC, 0x4880, 0x5144, 0x5668, 0x48B0, 0x4928, 0x4E68, 0x68A4, 0x5122, 0x46A0, 0x4040}},
    {16973, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4084, 0x4A24, 0x4A14, 0x5254, 0x41C4, 0x4014, 0x4008}},
    {16974, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1110, 0x920, 0x3FFC, 0x2004, 0x4FE8, 0x820, 0xFE0, 0x100, 0x7FFC, 0x540, 0x1930, 0x610C}},
    {16975, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x1EFC, 0x2224, 0x7F44, 0xA99C, 0x3F50, 0x297C, 0x3F90, 0x29FE, 0x2910, 0x4510, 0x8210}},
    {16976, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1020, 0xFE20, 0x1020, 0x7DFC, 0x5420, 0x7C20, 0x54F8, 0x7C88, 0x1088, 0xFE88, 0x10F8, 0x1088}},
    {16977, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x3FFC, 0x4AA0, 0xAA0, 0x7FFC, 0xAA0, 0xAA0, 0x7FFC, 0x0, 0x2488, 0x2244, 0x4244}},
    {16978, 2, {0x2080, 0x3EFE, 0x4910, 0x80F8, 0x3F00, 0x1110, 0x920, 0xFFFE, 0x920, 0x3018, 0xDFF6, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1010}},
    {16979, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x11FC, 0x1124, 0xFDFC, 0x3124, 0x39FC, 0x5420, 0x53FE, 0x90A8, 0x1124, 0x1222, 0x1020}},
    {16980, 2, {0x2040, 0x3F7E, 0x5090, 0x8808, 0x90, 0x88, 0xFBFE, 0x2048, 0x3C32, 0x45CE, 0x4890, 0xA888, 0x13FE, 0x1048, 0x2032, 0xC1CE}},
    {16981, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x440, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {16982, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x810, 0x7F10, 0x828, 0x7F44, 0x4992, 0x7F08, 0x497C, 0x7F04, 0x808, 0xFFB0, 0x808, 0x804}},
    {16983, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x11F8, 0x1108, 0xFDF8, 0x1000, 0x13FC, 0x1108, 0x1DF8, 0xF108, 0x11F8, 0x110E, 0x57F8, 0x2008}},
    {16984, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x1020, 0x13FE, 0xFA02, 0x14FC, 0x3000, 0x39FE, 0x5420, 0x50A8, 0x9124, 0x12A2, 0x1040}},
    {16985, 2, {0x2040, 0x3E7E, 0x4890, 0xBFF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1010, 0x3FF8}},
    {16986, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFE, 0x4002, 0xBFF4, 0x800, 0xFE0, 0x1124, 0x16A4, 0x181C}},
    {16987, 2, {0x2040, 0x3E7E, 0x4890, 0x8400, 0x3FF8, 0x2928, 0x2548, 0x2FE8, 0x2548, 0x2828, 0x100, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006}},
    {16988, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x20A0, 0x2090, 0x7DFE, 0x9120, 0x1320, 0x11FC, 0xFD20, 0x11FC, 0x2920, 0x2520, 0x41FE, 0x8100}},
    {16989, 2, {0x2040, 0x3F7E, 0x4890, 0x9518, 0x820, 0x3FF8, 0x100, 0x7FFC, 0x240, 0x3C50, 0x848, 0xFFFE, 0x850, 0x7E24, 0x854, 0x198C}},
    {16990, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x28, 0x3FFE, 0x2020, 0x2FA4, 0x2028, 0x2F90, 0x48AA, 0x4FC6, 0x8282, 0x900, 0x2828, 0x47E4}},
    {16991, 2, {0x2040, 0x3E7E, 0x4890, 0x8120, 0x11FC, 0x1320, 0xFDF8, 0x1120, 0x15F8, 0x1920, 0x31F8, 0xD100, 0x13F8, 0x1110, 0x50E0, 0x271E}},
    {16992, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2220, 0x4AA0, 0x8ABE, 0x1FC4, 0x20A4, 0x6F24, 0xA028, 0x2F28, 0x2910, 0x29A8, 0x2944, 0x3082}},
    {16993, 2, {0x2080, 0x3EFE, 0x4910, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE}},
    {16994, 2, {0x2040, 0x3E7E, 0x4890, 0x8100, 0x7FFC, 0x1110, 0x2928, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8}},
    {16995, 2, {0x2040, 0x3E7E, 0x4890, 0x8200, 0x1CF0, 0x1010, 0x1EF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2004, 0x4924, 0x2494, 0x2494, 0x4008}},
    {16996, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x4920, 0x2A20, 0xFF3E, 0x2A44, 0x49A4, 0x1024, 0xFE28, 0x2228, 0x6410, 0x1828, 0x2444, 0xC282}},
    {16997, 2, {0x2040, 0x3F7E, 0x4890, 0x8548, 0x11FE, 0x1122, 0xFD7A, 0x118A, 0x1552, 0x1922, 0x3152, 0xD1FE, 0x1040, 0x12A4, 0x528A, 0x247A}},
    {16998, 2, {0x2040, 0x3E7E, 0x4890, 0x8FF8, 0x4140, 0x2FF8, 0x2948, 0xFF8, 0xE100, 0x2FFC, 0x2410, 0x2BE8, 0x3224, 0x23E0, 0x5208, 0x8FFE}},
    {16999, 2, {0x2040, 0x3E7E, 0x4890, 0xF784, 0x4284, 0x62A4, 0x54A4, 0x69A4, 0x1424, 0x2224, 0xDD24, 0x824, 0x7F24, 0x4A04, 0x2F94, 0xF008}},
    {17000, 2, {0x2040, 0x3E7E, 0x4890, 0x9000, 0x878, 0x7F48, 0x2248, 0x1486, 0x7F00, 0x10FC, 0x6944, 0x1644, 0x6D28, 0x1490, 0x6428, 0x18C6}},
    {17001, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x13FE, 0x1200, 0xFE7C, 0x1244, 0x327C, 0x3A00, 0x56EE, 0x52AA, 0x92EE, 0x1200, 0x13FE}},
    {17002, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFEFE, 0x9292, 0xFEFE, 0x9292, 0xFEFE, 0x8282}},
    {17003, 2, {0x2040, 0x3E7E, 0x4890, 0x8640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4908, 0x2C0, 0x1D30, 0xE92E, 0x540, 0x920, 0x1310}},
    {17004, 2, {0x2080, 0x3EFE, 0x4910, 0x8080, 0x3FFE, 0x2020, 0x2F28, 0x3110, 0x2A0A, 0x277C, 0x294A, 0x3784, 0x2878, 0x4F48, 0x4130, 0x874C}},
    {17005, 2, {0x2080, 0x3EFE, 0x4910, 0x9084, 0x2108, 0x1084, 0x3FF8, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x2948, 0x2528, 0x294A, 0x2526, 0x3182}},
    {17006, 2, {0x2040, 0x3E7E, 0x4890, 0xFC7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4484, 0x4FE4, 0x5904, 0x4FC4, 0x4904, 0x4FC4, 0x4904, 0x4FEC}},
    {17007, 2, {0x2040, 0x3E7E, 0x4890, 0xBF7C, 0x2444, 0x3F28, 0x2110, 0x3F28, 0x2446, 0x3FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {17008, 2, {0x2040, 0x3E7E, 0x4890, 0xBE20, 0x2210, 0x3AFE, 0x2A00, 0x7F28, 0x4144, 0x3E82, 0x2228, 0x3E28, 0x2210, 0x3E28, 0x2244, 0x2682}},
    {17009, 2, {0x2040, 0x3F7E, 0x4890, 0xA508, 0x1000, 0x3FE, 0xFA00, 0xA7C, 0x1244, 0x2A7C, 0x7200, 0xAAEE, 0x2AAA, 0x22EE, 0x2200, 0x23FE}},
    {17010, 2, {0x2040, 0x3E7E, 0x4890, 0x8000, 0xF7FC, 0x1248, 0x17FC, 0xFC46, 0x86EC, 0x8554, 0xF6EC, 0x1554, 0x16EC, 0x1444, 0x5454, 0x2408}},
    {17011, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2044, 0x3E28, 0x44FE, 0xFF10, 0x4910, 0x7F7C, 0x4910, 0x7F10, 0xFE, 0x5510, 0x4A90, 0x8A90}},
    {17012, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x0, 0x3EF8, 0x2288, 0x3EF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x490, 0x1C60, 0x671C}},
    {17013, 2, {0x2040, 0x3F7E, 0x4890, 0x9FF8, 0x1490, 0x1250, 0x1FF0, 0x200, 0xFFFE, 0xAA0, 0x3458, 0xCFE6, 0x1550, 0x7C0, 0x4A0, 0x670}},
    {17014, 2, {0x2040, 0x3E7E, 0x4890, 0xA448, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x2448, 0x3F7E, 0x2040, 0x1FF0, 0x820, 0x7C0, 0xF83E}},
    {17015, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1E78, 0x1248, 0x3FFC, 0x2020, 0x2E20, 0x223C, 0x3FC8, 0x2928, 0x2F28, 0x4990, 0x5F28, 0x8144}},
    {17016, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8808, 0x4110, 0x2788, 0xF83E, 0x5394, 0x2008, 0xFBBE, 0x2008, 0x739C, 0x2288, 0x4388, 0x8288}},
    {17017, 2, {0x2040, 0x3E7E, 0x4890, 0xBFF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x3BB8, 0x2AA8, 0x3BB8, 0x0, 0x7FFC, 0x1110, 0x2928, 0xFFFE}},
    {17018, 2, {0x800, 0x800, 0x4AFC, 0x2A20, 0x2C20, 0x820, 0xFE20, 0x1820, 0x1C20, 0x2A20, 0x2A20, 0x4820, 0x8820, 0x820, 0x9FE, 0x800}},
    {17019, 2, {0x1004, 0x101E, 0x95F0, 0x5510, 0x5910, 0x1110, 0xFD10, 0x31FE, 0x3910, 0x5510, 0x5510, 0x9108, 0x110A, 0x114A, 0x1186, 0x1102}},
    {17020, 2, {0x804, 0x80E, 0x4AF0, 0x2A80, 0x2C80, 0x880, 0xFEFE, 0x1888, 0x1C88, 0x2A88, 0x2A88, 0x4888, 0x8888, 0x908, 0x908, 0xA08}},
    {17021, 2, {0x1020, 0x1020, 0x9524, 0x54A4, 0x58A8, 0x1020, 0xFDFC, 0x3020, 0x3820, 0x5420, 0x57FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {17022, 2, {0x440, 0x444, 0x7C78, 0x440, 0x442, 0x1C42, 0xE43E, 0x0, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {17023, 2, {0x1008, 0x103C, 0x95E0, 0x5420, 0x5924, 0x10A4, 0xFCA8, 0x3020, 0x39FE, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {17024, 2, {0x1020, 0x1020, 0x9420, 0x5420, 0x5BFE, 0x1020, 0xFC20, 0x3020, 0x39FC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {17025, 2, {0x1020, 0x1120, 0x9520, 0x55FC, 0x5920, 0x1220, 0xFC20, 0x33FE, 0x3890, 0x5490, 0x5490, 0x9090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {17026, 2, {0x1020, 0x1020, 0x97FE, 0x5420, 0x5820, 0x11FC, 0xFD24, 0x3124, 0x3924, 0x552C, 0x5470, 0x90A8, 0x1124, 0x1622, 0x1020, 0x1020}},
    {17027, 2, {0x2000, 0x27FC, 0x2404, 0xAC04, 0x75F4, 0x2444, 0xFC44, 0x25F4, 0x7554, 0x6D54, 0xAD54, 0x2554, 0x2574, 0x2446, 0x2446, 0x2842}},
    {17028, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4400, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {17029, 2, {0x1040, 0x1040, 0x94FC, 0x5504, 0x5A88, 0x1050, 0xFC20, 0x3040, 0x3880, 0x55FC, 0x5684, 0x9084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {17030, 2, {0x400, 0x3878, 0x2008, 0x2008, 0x3C78, 0x2008, 0x2008, 0x3FF8, 0x0, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {17031, 2, {0x1028, 0x1024, 0x95FE, 0x5420, 0x5820, 0x11FC, 0xFD24, 0x3124, 0x39FC, 0x5524, 0x5524, 0x91FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {17032, 2, {0x1080, 0x1080, 0x94FE, 0x5500, 0x5AFC, 0x1084, 0xFCA4, 0x3094, 0x3BFE, 0x5484, 0x5524, 0x9114, 0x11FE, 0x1004, 0x1028, 0x1010}},
    {17033, 2, {0x1040, 0x1020, 0x95FC, 0x5504, 0x5A28, 0x1020, 0xFC20, 0x31FC, 0x3870, 0x5468, 0x54A8, 0x90A4, 0x1124, 0x1222, 0x1020, 0x1020}},
    {17034, 2, {0x1000, 0x11FC, 0x9504, 0x5504, 0x59FC, 0x1100, 0xFD1C, 0x3170, 0x391C, 0x5570, 0x551E, 0x91F0, 0x1110, 0x1112, 0x1112, 0x120E}},
    {17035, 2, {0x1048, 0x1148, 0x9548, 0x5548, 0x5948, 0x11C8, 0xFC7E, 0x3048, 0x3BC8, 0x5548, 0x5548, 0x9148, 0x1148, 0x1148, 0x125C, 0x1440}},
    {17036, 2, {0x1020, 0x1010, 0x9410, 0x55FE, 0x5900, 0x1148, 0xFD4A, 0x314A, 0x396C, 0x5548, 0x5548, 0x914A, 0x114A, 0x126A, 0x1246, 0x1400}},
    {17037, 2, {0xC80, 0x70FC, 0x1124, 0xFD24, 0x1224, 0x3844, 0x5484, 0x9128, 0x1010, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {17038, 2, {0x1020, 0x1124, 0x94A8, 0x5420, 0x59FC, 0x1040, 0xFDFE, 0x3088, 0x3904, 0x56FA, 0x5488, 0x9088, 0x10A8, 0x1092, 0x1082, 0x107E}},
    {17039, 2, {0x1020, 0x10A4, 0x94A4, 0x5528, 0x5850, 0x1088, 0xFD04, 0x3024, 0x3820, 0x54A8, 0x54A8, 0x9130, 0x1050, 0x1088, 0x1104, 0x1202}},
    {17040, 2, {0x1000, 0x11FC, 0x9504, 0x5504, 0x59FC, 0x1104, 0xFD04, 0x31FC, 0x3800, 0x5512, 0x55D4, 0x9118, 0x1110, 0x1152, 0x1192, 0x110E}},
    {17041, 2, {0x1020, 0x1124, 0x94A4, 0x54A8, 0x5820, 0x11FC, 0xFD04, 0x3104, 0x3974, 0x5554, 0x5554, 0x9154, 0x1174, 0x1104, 0x1114, 0x1108}},
    {17042, 2, {0x10A0, 0x1090, 0x9480, 0x55FE, 0x5910, 0x1310, 0xFDFC, 0x3110, 0x3910, 0x55FC, 0x5510, 0x9110, 0x1110, 0x11FE, 0x1100, 0x1100}},
    {17043, 2, {0x1000, 0x11FC, 0x9504, 0x5504, 0x59FC, 0x1104, 0xFD04, 0x31FC, 0x3800, 0x57FE, 0x5420, 0x9120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {17044, 2, {0x2088, 0x2088, 0x23FE, 0xA888, 0x70A8, 0x2020, 0xF9FC, 0x2124, 0x7124, 0x6924, 0xABFE, 0x2020, 0x2050, 0x2088, 0x2104, 0x2202}},
    {17045, 2, {0x2010, 0x2010, 0x207C, 0xAB14, 0x71FE, 0x2114, 0xFA7C, 0x2210, 0x777C, 0x6910, 0xA9FE, 0x2510, 0x2210, 0x2300, 0x24FE, 0x2800}},
    {17046, 2, {0x2020, 0x2020, 0x23FE, 0xA820, 0x7020, 0x23FE, 0xFA8A, 0x2252, 0x72FA, 0x6A22, 0xAA22, 0x22FA, 0x2222, 0x2222, 0x222A, 0x2204}},
    {17047, 2, {0x1020, 0x1040, 0x95FC, 0x5504, 0x59FC, 0x1104, 0xFDFC, 0x3000, 0x3BFE, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {17048, 2, {0x1020, 0x1020, 0x95FE, 0x5522, 0x59FE, 0x1122, 0xFDFE, 0x3000, 0x38FC, 0x5484, 0x54FC, 0x9084, 0x10FC, 0x1084, 0x1094, 0x1088}},
    {17049, 2, {0x0, 0xFDFE, 0x2048, 0x4048, 0x7DFE, 0xC448, 0x4488, 0x7D08, 0x8, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {17050, 2, {0x2010, 0x2220, 0x217C, 0xA944, 0x7044, 0x207C, 0xFB40, 0x2140, 0x717C, 0x6944, 0xA944, 0x217C, 0x2144, 0x2280, 0x247E, 0x2000}},
    {17051, 2, {0x2000, 0x245C, 0x2294, 0xA914, 0x7294, 0x2454, 0xF926, 0x2100, 0x77DC, 0x6914, 0xAD94, 0x2554, 0x2948, 0x2108, 0x2514, 0x2222}},
    {17052, 2, {0x1020, 0x13FE, 0x9420, 0x55FC, 0x5924, 0x11FC, 0xFD24, 0x31FC, 0x3822, 0x57FE, 0x5408, 0x93FE, 0x1108, 0x1088, 0x1028, 0x1010}},
    {17053, 2, {0x101C, 0x11E0, 0x9420, 0x5524, 0x58A8, 0x13FE, 0xFC70, 0x30A8, 0x3924, 0x5622, 0x5440, 0x9024, 0x12A2, 0x128A, 0x1488, 0x1078}},
    {17054, 2, {0x2040, 0x2020, 0x23FE, 0xA888, 0x7050, 0x23FE, 0xFA22, 0x22FA, 0x7222, 0x6AFA, 0xAA8A, 0x228A, 0x22FA, 0x2202, 0x220A, 0x2204}},
    {17055, 2, {0x2088, 0x2088, 0x23FE, 0xA888, 0x7088, 0x20F8, 0xF820, 0x23FE, 0x7222, 0x6B32, 0xAAAA, 0x2376, 0x2222, 0x2222, 0x222A, 0x2224}},
    {17056, 2, {0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {17057, 2, {0x1000, 0x11F8, 0x9508, 0x55F8, 0x5908, 0x11F8, 0xFC00, 0x33FC, 0x3A94, 0x57FC, 0x5400, 0x91F8, 0x1090, 0x1060, 0x1198, 0x1606}},
    {17058, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8418, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {17059, 2, {0x1088, 0x1088, 0x95FC, 0x5488, 0x5888, 0x13FE, 0xFC20, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x91FC, 0x1000, 0x1088, 0x1104, 0x1202}},
    {17060, 2, {0x1000, 0x13FE, 0x9450, 0x57FE, 0x5A52, 0x13FE, 0xFC00, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {17061, 2, {0x2020, 0x21FC, 0x2024, 0xABFE, 0x7024, 0x21FC, 0xF820, 0x22AA, 0x73AE, 0x6A22, 0xABFE, 0x2222, 0x23AE, 0x22AA, 0x22AA, 0x2422}},
    {17062, 2, {0x1020, 0x13FE, 0x9420, 0x55FC, 0x5800, 0x13FE, 0xFE02, 0x31FC, 0x3800, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x1088, 0x13FE, 0x1000}},
    {17063, 2, {0x2090, 0x27FC, 0x2094, 0xABFC, 0x7290, 0x23FE, 0xF912, 0x23FA, 0x750E, 0x69F8, 0xA908, 0x21F8, 0x2108, 0x21F8, 0x2090, 0x2108}},
    {17064, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0x755C, 0x5394, 0x77D4, 0x5114, 0x7396, 0x5556, 0xB122}},
    {17065, 2, {0x103C, 0x13E0, 0x9524, 0x54A8, 0x5BFE, 0x10A8, 0xFD24, 0x3202, 0x39FC, 0x5524, 0x5524, 0x91FC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {17066, 2, {0x1088, 0x1088, 0x97FE, 0x5488, 0x59FC, 0x1124, 0xFDFC, 0x3124, 0x39FC, 0x5420, 0x57FE, 0x9222, 0x122A, 0x12FA, 0x120A, 0x1206}},
    {17067, 2, {0x2000, 0x23FC, 0x2294, 0xAA94, 0x73FC, 0x2000, 0xF890, 0x279E, 0x7090, 0x6890, 0xAB9C, 0x2090, 0x2090, 0x279E, 0x2090, 0x2090}},
    {17068, 2, {0xDFC, 0x7124, 0x11FC, 0x1124, 0xFDFC, 0x3020, 0x3BFE, 0x54A8, 0x5124, 0x8222, 0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {17069, 2, {0x1020, 0x13FE, 0x9420, 0x55FC, 0x5800, 0x13FE, 0xFC02, 0x31FC, 0x3820, 0x57FE, 0x5400, 0x93FE, 0x1004, 0x13BE, 0x12A4, 0x138C}},
    {17070, 2, {0x2000, 0x23DE, 0x2042, 0xAA52, 0x714A, 0x2252, 0xF890, 0x2108, 0x73FE, 0x6D10, 0xA9FE, 0x2110, 0x21FE, 0x2110, 0x21FE, 0x2100}},
    {17071, 2, {0x1020, 0x1010, 0x95FE, 0x5528, 0x597C, 0x1128, 0xFDFE, 0x3110, 0x397C, 0x5554, 0x557C, 0x9154, 0x117C, 0x1200, 0x1244, 0x1482}},
    {17072, 2, {0x2210, 0x2110, 0x203A, 0xAB92, 0x7014, 0x23BE, 0xF808, 0x2390, 0x703C, 0x6BE4, 0xAAA4, 0x22BC, 0x22A4, 0x23A4, 0x22BC, 0x2024}},
    {17073, 2, {0x2148, 0x214C, 0x22AA, 0xA808, 0x77FE, 0x2148, 0xFB68, 0x214A, 0x736A, 0x694C, 0xAB6C, 0x2148, 0x216A, 0x239A, 0x2026, 0x2042}},
    {17074, 2, {0x10, 0xFE08, 0x28FE, 0xFE00, 0xAA44, 0xAA28, 0xFEFE, 0x1088, 0x54A8, 0x38BE, 0xFEC8, 0x1088, 0x38BE, 0x5508, 0x9108, 0x127E}},
    {17075, 2, {0x2020, 0x23FE, 0x2248, 0xABFE, 0x7248, 0x22EC, 0xFB5A, 0x2248, 0x7250, 0x6BDE, 0xAA50, 0x23DC, 0x2250, 0x25DE, 0x2450, 0x2850}},
    {17076, 2, {0x1088, 0x1088, 0x94EE, 0x5554, 0x5A22, 0x1050, 0xFC88, 0x3174, 0x3A02, 0x55DC, 0x5554, 0x91DC, 0x1088, 0x1088, 0x1154, 0x1222}},
    {17077, 2, {0x1040, 0x1040, 0x2040, 0x2440, 0x45F8, 0xF848, 0x1048, 0x2048, 0x4848, 0xFC48, 0x448, 0x4A, 0x548A, 0x548A, 0x8106, 0x200}},
    {17078, 2, {0x1020, 0x1020, 0x2020, 0x2520, 0x452C, 0xF934, 0x1164, 0x23A4, 0x4924, 0xFD34, 0x528, 0x122, 0x5522, 0x5502, 0x80FE, 0x0}},
    {17079, 2, {0x808, 0x808, 0x1008, 0x1208, 0x22FE, 0x7C08, 0x818, 0x1018, 0x2428, 0x7E28, 0x248, 0x88, 0x5208, 0x4908, 0x8928, 0x10}},
    {17080, 2, {0x1000, 0x11FC, 0x2000, 0x2400, 0x4400, 0xFBFE, 0x1040, 0x2040, 0x4880, 0xFCFC, 0x404, 0x4, 0x5404, 0x5404, 0x8028, 0x10}},
    {17081, 2, {0x1008, 0x103C, 0x21E0, 0x2420, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4820, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x8020, 0x20}},
    {17082, 2, {0x1000, 0x11FE, 0x2040, 0x2440, 0x4440, 0xF87C, 0x1084, 0x2084, 0x4884, 0xFD04, 0x5FC, 0x8, 0x5408, 0x5408, 0x81FE, 0x0}},
    {17083, 2, {0x1040, 0x1040, 0x2040, 0x247E, 0x4482, 0xF884, 0x1120, 0x2220, 0x4820, 0xFC50, 0x450, 0x48, 0x5488, 0x5484, 0x8102, 0x200}},
    {17084, 2, {0x1010, 0x1010, 0x2010, 0x2490, 0x4490, 0xF890, 0x109E, 0x2090, 0x4890, 0xFC90, 0x490, 0x90, 0x5490, 0x5490, 0x83FE, 0x0}},
    {17085, 2, {0x1020, 0x1020, 0x2020, 0x25FC, 0x4424, 0xF824, 0x1024, 0x2024, 0x4BFE, 0xFC20, 0x450, 0x50, 0x5488, 0x5488, 0x8104, 0x202}},
    {17086, 2, {0x1020, 0x1020, 0x2020, 0x2420, 0x45FE, 0xF820, 0x1070, 0x2070, 0x48A8, 0xFCA8, 0x524, 0x124, 0x5622, 0x5420, 0x8020, 0x20}},
    {17087, 2, {0x1020, 0x1020, 0x2020, 0x2420, 0x45FC, 0xF820, 0x1020, 0x2020, 0x4BFE, 0xFC20, 0x450, 0x50, 0x5488, 0x5488, 0x8104, 0x202}},
    {17088, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x2110, 0x23E0, 0x2040, 0x2088, 0x27FC, 0x2044, 0x4250, 0x4448, 0x8944, 0x80}},
    {17089, 2, {0x1040, 0x1020, 0x2000, 0x25FC, 0x4400, 0xF800, 0x10F0, 0x2090, 0x4890, 0xFC90, 0x490, 0x92, 0x5492, 0x5512, 0x810E, 0x200}},
    {17090, 2, {0x804, 0x844, 0x1028, 0x1210, 0x2228, 0x7C44, 0x880, 0x1004, 0x2444, 0x7E44, 0x228, 0x28, 0x5210, 0x4928, 0x8944, 0x82}},
    {17091, 2, {0x1008, 0x103C, 0x21E0, 0x2420, 0x4420, 0xF83C, 0x11E0, 0x2020, 0x4820, 0xFC3E, 0x7E0, 0x20, 0x5422, 0x5422, 0x8022, 0x1E}},
    {17092, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x4594, 0xF954, 0x1154, 0x2124, 0x4924, 0xFD54, 0x554, 0x194, 0x5504, 0x5504, 0x8114, 0x108}},
    {17093, 2, {0x1010, 0x1010, 0x2090, 0x2490, 0x44FE, 0xF890, 0x1110, 0x2010, 0x4810, 0xFDFE, 0x410, 0x10, 0x5410, 0x5410, 0x8010, 0x10}},
    {17094, 2, {0x1040, 0x1020, 0x2020, 0x25FC, 0x4504, 0xF904, 0x1104, 0x21FC, 0x4904, 0xFD00, 0x500, 0x100, 0x5500, 0x5600, 0x8200, 0x400}},
    {17095, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x4504, 0xF904, 0x11FC, 0x2120, 0x4920, 0xFD20, 0x520, 0x110, 0x5510, 0x5508, 0x8204, 0x402}},
    {17096, 2, {0x800, 0x7F7C, 0x1144, 0x3228, 0xC10, 0x1228, 0x6346, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {17097, 2, {0x1040, 0x1020, 0x2020, 0x2BFE, 0x4A02, 0xF404, 0x1000, 0x2090, 0x4890, 0xFC90, 0x488, 0x108, 0x5508, 0x5504, 0x8204, 0x402}},
    {17098, 2, {0x1000, 0x1000, 0x21FE, 0x2410, 0x4410, 0xF810, 0x1090, 0x2090, 0x489E, 0xFC90, 0x490, 0x90, 0x5490, 0x5490, 0x83FE, 0x0}},
    {17099, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x4504, 0xF9FC, 0x1120, 0x2120, 0x49FE, 0xFD20, 0x520, 0x110, 0x5512, 0x554A, 0x8186, 0x102}},
    {17100, 2, {0x1000, 0x11FE, 0x2100, 0x2500, 0x4500, 0xF9FC, 0x1104, 0x2104, 0x4904, 0xFD04, 0x5FC, 0x100, 0x5500, 0x5500, 0x81FE, 0x0}},
    {17101, 2, {0x1000, 0x1000, 0x21FE, 0x2408, 0x4408, 0xF9E8, 0x1128, 0x2128, 0x4928, 0xFD28, 0x5E8, 0x128, 0x5408, 0x5408, 0x8028, 0x10}},
    {17102, 2, {0x1000, 0x1000, 0x21FC, 0x2504, 0x4504, 0xF904, 0x11FC, 0x2104, 0x4904, 0xFD04, 0x5FC, 0x104, 0x5400, 0x5400, 0x83FE, 0x0}},
    {17103, 2, {0x1080, 0x1080, 0x2080, 0x24FE, 0x4540, 0xF940, 0x1240, 0x207C, 0x4840, 0xFC40, 0x440, 0x7C, 0x5440, 0x5440, 0x8040, 0x40}},
    {17104, 2, {0x1028, 0x1024, 0x2024, 0x2820, 0x4BFE, 0xF220, 0x1224, 0x2224, 0x4A24, 0xFA28, 0xA28, 0x290, 0xAB12, 0xAA2A, 0x8046, 0x82}},
    {17105, 2, {0x1000, 0x11FE, 0x2020, 0x2420, 0x4420, 0xF9FE, 0x1122, 0x2122, 0x4952, 0xFD4A, 0x58A, 0x102, 0x5502, 0x5502, 0x810A, 0x104}},
    {17106, 2, {0x1000, 0x11F8, 0x2008, 0x2450, 0x4420, 0xF810, 0x13FE, 0x2062, 0x48A4, 0xFCA0, 0x520, 0x220, 0x5420, 0x5420, 0x80A0, 0x40}},
    {17107, 2, {0x1080, 0x1080, 0x20F8, 0x2508, 0x4610, 0xFC20, 0x11FC, 0x2004, 0x4804, 0xFC04, 0x5FC, 0x4, 0x5404, 0x5404, 0x81FC, 0x4}},
    {17108, 2, {0x1000, 0x11FE, 0x2010, 0x2410, 0x4420, 0xF820, 0x1068, 0x20A4, 0x4922, 0xFE22, 0x420, 0x20, 0x5420, 0x5400, 0x83FE, 0x0}},
    {17109, 2, {0x818, 0x13E0, 0x3040, 0x57FC, 0x9040, 0x13F8, 0x200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {17110, 2, {0x1020, 0x1020, 0x20FC, 0x2424, 0x45FE, 0xF824, 0x10FC, 0x2020, 0x4820, 0xFCFC, 0x420, 0x20, 0x55FE, 0x5420, 0x8020, 0x20}},
    {17111, 2, {0x1000, 0x11FE, 0x2120, 0x2520, 0x4520, 0xF9FC, 0x1104, 0x2104, 0x4904, 0xFD04, 0x5FC, 0x120, 0x5520, 0x5520, 0x81FE, 0x0}},
    {17112, 2, {0x1020, 0x1020, 0x23FE, 0x2820, 0x4820, 0xF1FC, 0x1020, 0x2020, 0x4BFE, 0xF870, 0x8A8, 0xA8, 0xA924, 0xAA22, 0x8420, 0x20}},
    {17113, 2, {0x1000, 0x13FE, 0x2108, 0x2508, 0x4508, 0xF9F8, 0x1108, 0x2108, 0x49F8, 0xFD08, 0x508, 0x11E, 0x57E8, 0x5408, 0x8008, 0x8}},
    {17114, 2, {0x1000, 0x13DE, 0x2042, 0x2842, 0x4A52, 0xF14A, 0x114A, 0x2042, 0x48C6, 0xF94A, 0xA52, 0x42, 0xA842, 0xA842, 0x814A, 0x84}},
    {17115, 2, {0x1020, 0x1020, 0x2124, 0x24A4, 0x44A8, 0xF820, 0x1020, 0x23FE, 0x4870, 0xFCA8, 0x4A8, 0x124, 0x5524, 0x5622, 0x8020, 0x20}},
    {17116, 2, {0x1000, 0x8FC, 0x3E44, 0x2228, 0x2210, 0x3E28, 0x2246, 0x4420, 0x9FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {17117, 2, {0x1040, 0x1024, 0x2224, 0x2508, 0x4508, 0xF810, 0x1040, 0x2040, 0x4BFE, 0xFC88, 0x508, 0x190, 0x5460, 0x5450, 0x8188, 0x604}},
    {17118, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x4504, 0xF9AC, 0x1154, 0x2154, 0x4954, 0xFD54, 0x5AC, 0x104, 0x5504, 0x5504, 0x8114, 0x108}},
    {17119, 2, {0x1000, 0x11F8, 0x2008, 0x2408, 0x45F8, 0xF900, 0x1104, 0x2104, 0x48FC, 0xFC40, 0x420, 0xA4, 0xAA8A, 0xAA8A, 0x8478, 0x0}},
    {17120, 2, {0x1028, 0x1024, 0x21FE, 0x2420, 0x4420, 0xF9FC, 0x1124, 0x2124, 0x49FC, 0xFD24, 0x524, 0x1FC, 0x5524, 0x5524, 0x8124, 0x10C}},
    {17121, 2, {0x1020, 0x1020, 0x2050, 0x2888, 0x4904, 0xF2FA, 0x1020, 0x2020, 0x4BFE, 0xF820, 0x928, 0x124, 0xAA22, 0xAC22, 0x80A0, 0x40}},
    {17122, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1648, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17123, 2, {0x2010, 0x2010, 0x3F24, 0x4044, 0xBEF8, 0x2210, 0x3220, 0x2A44, 0xFF7E, 0x2212, 0x5210, 0x4A54, 0x7F52, 0x292, 0x1450, 0x820}},
    {17124, 2, {0x840, 0x840, 0x7F40, 0x87C, 0x3E84, 0x884, 0xFF04, 0x1044, 0x2224, 0x7C24, 0x904, 0xFF84, 0x884, 0x4904, 0xA8A8, 0x1010}},
    {17125, 2, {0x1000, 0x11FC, 0x2124, 0x2524, 0x45FC, 0xF924, 0x1124, 0x21FC, 0x4820, 0xFC20, 0x5FC, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {17126, 2, {0x1080, 0x1080, 0x20FE, 0x2500, 0x46FC, 0xF884, 0x10A4, 0x2094, 0x49FE, 0xFC84, 0x524, 0x114, 0x55FE, 0x5404, 0x8028, 0x10}},
    {17127, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x45FC, 0xF910, 0x1110, 0x21FE, 0x4910, 0xFD10, 0x57C, 0x144, 0x5544, 0x5544, 0x817C, 0x244}},
    {17128, 2, {0x1040, 0x1020, 0x21FC, 0x2400, 0x4508, 0xF890, 0x1000, 0x23FE, 0x4800, 0xFC00, 0x5FC, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {17129, 2, {0x1000, 0x1040, 0x239C, 0x2A04, 0x4A04, 0xF39C, 0x1204, 0x2204, 0x4BFC, 0xFC90, 0x490, 0x90, 0x5490, 0x5512, 0x8212, 0x40E}},
    {17130, 2, {0x2000, 0x277E, 0x4504, 0x4504, 0x9574, 0xE554, 0x2654, 0x4554, 0x9554, 0xF554, 0x574, 0x1554, 0xAE04, 0xA404, 0x8414, 0x408}},
    {17131, 2, {0x1080, 0x1080, 0x20F8, 0x2508, 0x4610, 0xF9FC, 0x1024, 0x2024, 0x49FE, 0xFC24, 0x424, 0x1FC, 0x5424, 0x5420, 0x80A0, 0x40}},
    {17132, 2, {0x1000, 0x13FC, 0x2040, 0x2840, 0x4BFE, 0xF0A0, 0x1110, 0x2208, 0x4C46, 0xF840, 0x848, 0x264, 0xAA52, 0xAC52, 0x8140, 0x80}},
    {17133, 2, {0x1040, 0x1020, 0x23FE, 0x2880, 0x4904, 0xF3FE, 0x1002, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x5504, 0x8114, 0x108}},
    {17134, 2, {0x1020, 0x1020, 0x2050, 0x2488, 0x4504, 0xFA02, 0x10F8, 0x2020, 0x4820, 0xFDFC, 0x420, 0x124, 0x54A4, 0x54A8, 0x83FE, 0x0}},
    {17135, 2, {0x1048, 0x844, 0x3E7E, 0x23C0, 0x2224, 0x3E28, 0x2212, 0x206A, 0x4446, 0x8882, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17136, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x1110, 0x2248, 0xCF86, 0x220, 0x1FF0, 0x948, 0x1320}},
    {17137, 2, {0x1020, 0x1020, 0x2020, 0x2BFE, 0x4820, 0xF124, 0x1124, 0x2124, 0x4AAA, 0xF870, 0x8A8, 0xA8, 0xA924, 0xAA22, 0x8420, 0x20}},
    {17138, 2, {0x1000, 0x11F0, 0x2110, 0x2910, 0x49F0, 0xF040, 0x13F8, 0x2248, 0x4A48, 0xFBF8, 0xA48, 0x248, 0xABFA, 0xA842, 0x8042, 0x3E}},
    {17139, 2, {0x1050, 0x1052, 0x23DC, 0x2450, 0x44D2, 0xFB4E, 0x1000, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x5504, 0x8114, 0x108}},
    {17140, 2, {0x1000, 0x10FC, 0x2048, 0x2430, 0x45FE, 0xF852, 0x1094, 0x2110, 0x4A30, 0xFC20, 0x5FE, 0x70, 0x54A8, 0x5524, 0x8222, 0x20}},
    {17141, 2, {0x1000, 0x11FC, 0x2104, 0x25FC, 0x4504, 0xF9FC, 0x1080, 0x21FE, 0x4A22, 0xFD22, 0x552, 0x102, 0x55FA, 0x5402, 0x8014, 0x8}},
    {17142, 2, {0x1000, 0x13FE, 0x2020, 0x2440, 0x45FC, 0xF904, 0x1104, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x104, 0x55FC, 0x5400, 0x8088, 0x104}},
    {17143, 2, {0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x98C6, 0x440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17144, 2, {0x1008, 0x101C, 0x21F0, 0x2510, 0x4510, 0xF9FE, 0x1110, 0x217C, 0x4944, 0xFD7C, 0x544, 0x17C, 0xA944, 0xAA44, 0x827C, 0x444}},
    {17145, 2, {0x1000, 0x13FE, 0x2000, 0x29FC, 0x4904, 0xF104, 0x11FC, 0x2000, 0x4BFE, 0xFA22, 0xA22, 0x3FE, 0xAA22, 0xAA22, 0x83FE, 0x202}},
    {17146, 2, {0x1088, 0x11C8, 0x2708, 0x2908, 0x492A, 0xF12A, 0x17AC, 0x2148, 0x4B08, 0xFB88, 0xD54, 0x514, 0xA914, 0xA924, 0x8124, 0x142}},
    {17147, 2, {0x1020, 0x1020, 0x23FE, 0x2820, 0x4820, 0xF3FE, 0x128A, 0x2252, 0x4AFA, 0xFA22, 0xA22, 0x2FA, 0xAA22, 0xAA22, 0x822A, 0x204}},
    {17148, 2, {0x1000, 0x11F8, 0x2088, 0x2490, 0x44FC, 0xF954, 0x1124, 0x2154, 0x4A88, 0xFC00, 0x5FC, 0x154, 0xA954, 0xA954, 0x83FE, 0x0}},
    {17149, 2, {0x2104, 0x2104, 0x4104, 0x47C4, 0x913E, 0xE104, 0x27C4, 0x4024, 0x9114, 0xF114, 0x7C4, 0x1104, 0xA904, 0xA1C4, 0x8E14, 0x408}},
    {17150, 2, {0x1000, 0x1040, 0x23FE, 0x2888, 0x4904, 0xF642, 0x10FC, 0x2308, 0x48B0, 0xF8D0, 0xF3E, 0xC2, 0xAB24, 0xA818, 0x8060, 0x780}},
    {17151, 2, {0x1000, 0x11FC, 0x2124, 0x2924, 0x49FC, 0xF124, 0x1124, 0x21FC, 0x4800, 0xFBFE, 0x920, 0x122, 0xA914, 0xA948, 0x8184, 0x102}},
    {17152, 2, {0x10C4, 0x1704, 0x2124, 0x2914, 0x4914, 0xF784, 0x1124, 0x2114, 0x4B94, 0xFB46, 0xD3C, 0x504, 0xA904, 0xA904, 0x8104, 0x104}},
    {17153, 2, {0x1000, 0x13FE, 0x2050, 0x2850, 0x4BFE, 0xF252, 0x1252, 0x23FE, 0x4800, 0xFC40, 0x7FE, 0x88, 0x5590, 0x5460, 0x80D8, 0x304}},
    {17154, 2, {0x1000, 0x13FC, 0x2204, 0x2204, 0x4BFC, 0xFA00, 0x13FC, 0x2240, 0x4A88, 0xFBFC, 0xA24, 0x1220, 0xADFC, 0xA420, 0x8820, 0x13FE}},
    {17155, 2, {0x1082, 0x1082, 0x22A2, 0x29CA, 0x488A, 0xF3EA, 0x122A, 0x222A, 0x4BEA, 0xFA2A, 0xA2A, 0x3EA, 0xAA22, 0xAA22, 0x82AA, 0x244}},
    {17156, 2, {0x11FC, 0x1020, 0x21FC, 0x2524, 0x4554, 0xF904, 0x112C, 0x2020, 0x483C, 0xFC20, 0x5F8, 0x88, 0x5450, 0x5420, 0x80D8, 0x306}},
    {17157, 2, {0x1020, 0x1010, 0x21FE, 0x2510, 0x457C, 0xF914, 0x11FE, 0x2114, 0x497C, 0xFD10, 0x57C, 0x144, 0xA944, 0xAA44, 0x827C, 0x444}},
    {17158, 2, {0x1020, 0x1040, 0x23FC, 0x2A24, 0x4A24, 0xF3FC, 0x1224, 0x2244, 0x4BFC, 0xF840, 0x8A8, 0xB4, 0xA93C, 0xA922, 0x8222, 0x41E}},
    {17159, 2, {0x1020, 0x1020, 0x23FE, 0x2820, 0x4BFE, 0xF242, 0x148C, 0x21F0, 0x4820, 0xF844, 0xBFE, 0x22, 0xA928, 0xAA24, 0x84A4, 0x40}},
    {17160, 2, {0x1000, 0x11FC, 0x2104, 0x29FC, 0x4904, 0xF1FC, 0x1000, 0x23DE, 0x4842, 0xFA52, 0x94A, 0xC6, 0xAB5A, 0xA842, 0x814A, 0x84}},
    {17161, 2, {0x1000, 0x13FE, 0x2252, 0x2A52, 0x4BFE, 0xF000, 0x1000, 0x23FE, 0x4820, 0xF820, 0x93C, 0x120, 0xA920, 0xA920, 0x87FE, 0x0}},
    {17162, 2, {0x1008, 0x103C, 0x23C0, 0x2804, 0x4A44, 0xF128, 0x11FC, 0x2220, 0x4820, 0xFBFE, 0x820, 0x124, 0xA924, 0xA924, 0x81FC, 0x4}},
    {17163, 2, {0x1088, 0x1088, 0x23FE, 0x2888, 0x4820, 0xF050, 0x1088, 0x2104, 0x4AFA, 0xFC00, 0x400, 0x1FC, 0x5504, 0x5504, 0x81FC, 0x104}},
    {17164, 2, {0x1000, 0x13DE, 0x2042, 0x2A52, 0x494A, 0xF252, 0x1042, 0x2020, 0x49FC, 0xFD04, 0x504, 0x1FC, 0x5504, 0x5504, 0x81FC, 0x104}},
    {17165, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x44F8, 0xF820, 0x11FC, 0x2124, 0x49FC, 0xFC20, 0x5FC, 0x20, 0x55FC, 0x5420, 0x83FE, 0x0}},
    {17166, 2, {0x40, 0x7CF8, 0x5548, 0x5430, 0x7DCE, 0x54F8, 0x5488, 0x7CF8, 0x440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17167, 2, {0x1040, 0x1020, 0x23FE, 0x2A42, 0x4828, 0xF294, 0x12AA, 0x24CA, 0x48F8, 0xFB00, 0x820, 0x124, 0xA924, 0xA924, 0x81FC, 0x4}},
    {17168, 2, {0x100E, 0x13F0, 0x2044, 0x2A24, 0x4908, 0xF3FC, 0x1204, 0x23FC, 0x4A04, 0xFBFC, 0xA04, 0x3FC, 0xA890, 0xA910, 0x8212, 0xC0E}},
    {17169, 2, {0x1000, 0x13DC, 0x2294, 0x2BD4, 0x4A54, 0xF3C8, 0x1288, 0x2294, 0x4BE2, 0xF800, 0x820, 0x1FC, 0xA820, 0xA820, 0x83FE, 0x0}},
    {17170, 2, {0x1040, 0x1020, 0x23FE, 0x2250, 0x4A50, 0xFBFE, 0x1252, 0x2252, 0x4BFE, 0xFA00, 0xA92, 0x12D4, 0xAA98, 0xA492, 0x84D2, 0x88E}},
    {17171, 2, {0x2078, 0x3F48, 0x404C, 0xBE80, 0x2A78, 0xFF28, 0x4A10, 0x7F28, 0x444, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17172, 2, {0x1000, 0x13FE, 0x2200, 0x2A7C, 0x4A44, 0xF244, 0x127C, 0x2200, 0x4AEE, 0xFAAA, 0xAAA, 0x2AA, 0xAAEE, 0xAA00, 0x83FE, 0x0}},
    {17173, 2, {0x2040, 0x2020, 0x47FE, 0x4490, 0x9490, 0xE490, 0x27FC, 0x4490, 0x94D8, 0xF5B8, 0x5B4, 0x16D4, 0xA892, 0xA890, 0x9090, 0x90}},
    {17174, 2, {0x10A0, 0x1090, 0x21FE, 0x2B20, 0x49FC, 0xF120, 0x11FC, 0x2120, 0x49FE, 0xFD00, 0x420, 0x3FE, 0x5470, 0x54A8, 0x8326, 0x20}},
    {17175, 2, {0x2100, 0x2100, 0x43FC, 0x46A8, 0x9AA8, 0xE2A8, 0x27FC, 0x42A8, 0x92A8, 0xF2A8, 0xFFE, 0x1000, 0xAAA8, 0xA254, 0x8454, 0x0}},
    {17176, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4546, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17177, 2, {0x1088, 0x1088, 0x21FC, 0x2488, 0x4488, 0xFBFE, 0x1020, 0x21FC, 0x4924, 0xFDFC, 0x524, 0x1FC, 0x5400, 0x5488, 0x8104, 0x202}},
    {17178, 2, {0x1000, 0x13FE, 0x2050, 0x2850, 0x4BFE, 0xF252, 0x1252, 0x23FE, 0x4924, 0xF8A8, 0xBFE, 0x70, 0xA8A8, 0xA924, 0x8622, 0x20}},
    {17179, 2, {0x1000, 0x1E0C, 0x22F0, 0x5420, 0x2844, 0x10F8, 0x2810, 0xFF24, 0xFE, 0x7E12, 0x50, 0x7E54, 0x92, 0x7F12, 0x4250, 0x7E20}},
    {17180, 2, {0x1050, 0x1252, 0x2154, 0x2850, 0x4BFE, 0xF088, 0x1050, 0x23FE, 0x4820, 0xF9FC, 0x820, 0x3FE, 0xA8A8, 0xA924, 0x8622, 0x20}},
    {17181, 2, {0x2110, 0x2110, 0x47BC, 0x4110, 0x93B8, 0xE554, 0x2912, 0x4000, 0x93F8, 0xF000, 0x7FC, 0x1040, 0xAA48, 0xA444, 0x8942, 0x80}},
    {17182, 2, {0x1020, 0x1020, 0x2050, 0x2488, 0x4504, 0xFAFA, 0x1000, 0x2000, 0x49DC, 0xFD54, 0x554, 0x1DC, 0x5488, 0x5488, 0x8154, 0x222}},
    {17183, 2, {0x1010, 0x13C8, 0x227E, 0x2A40, 0x4A62, 0xF3D4, 0x1200, 0x223E, 0x4BC8, 0xFA48, 0xE7E, 0x248, 0xAA48, 0xABC8, 0x8248, 0x8}},
    {17184, 2, {0x290, 0x7FD0, 0x421E, 0x5EA2, 0x4288, 0x5D08, 0x5548, 0x5ED4, 0x8462, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {17185, 2, {0x1090, 0x13FC, 0x2294, 0x2BFC, 0x4A94, 0xF3FC, 0x1000, 0x23FC, 0x4800, 0xFBFC, 0xA04, 0x3FC, 0xA908, 0xA890, 0x87FE, 0x0}},
    {17186, 2, {0x1000, 0x11FC, 0x2020, 0x2BFE, 0x4A22, 0xF1AC, 0x1020, 0x21AC, 0x4850, 0xF8C8, 0xB26, 0xF8, 0xA808, 0xA8D0, 0x8020, 0x10}},
    {17187, 2, {0x1108, 0x1108, 0x21EE, 0x2A94, 0x4842, 0xF010, 0x11FE, 0x2102, 0x48F8, 0xFC88, 0x4F8, 0x80, 0x54FC, 0x5484, 0x80FC, 0x84}},
    {17188, 2, {0x2000, 0x27FC, 0x4248, 0x4444, 0x9FFE, 0xE444, 0x26EC, 0x4554, 0x96EC, 0xF444, 0x6EC, 0x1554, 0xAEEC, 0xA444, 0x8454, 0x408}},
    {17189, 2, {0x103E, 0x17C0, 0x2244, 0x2128, 0x4BF8, 0xF840, 0x17FC, 0x2000, 0x4BF8, 0xF808, 0xBF8, 0x1008, 0xABF8, 0xA544, 0x852A, 0x8FA}},
    {17190, 2, {0x1000, 0x13DE, 0x2042, 0x2A52, 0x494A, 0xF252, 0x1090, 0x2108, 0x4BFE, 0xFD10, 0x9FE, 0x110, 0xA9FE, 0xA910, 0x81FE, 0x100}},
    {17191, 2, {0x1020, 0x13FE, 0x2020, 0x29FC, 0x4800, 0xF3FE, 0x1002, 0x21FC, 0x4820, 0xFBFE, 0x800, 0x3FE, 0xA804, 0xABBE, 0x82A4, 0x38C}},
    {17192, 2, {0x1088, 0x1088, 0x23FE, 0x2888, 0x4910, 0xF1DE, 0x1252, 0x2554, 0x4A88, 0xF974, 0xA02, 0x1FC, 0xA820, 0xA924, 0x8222, 0x60}},
    {17193, 2, {0x1108, 0x17FE, 0x2108, 0x2800, 0x4BFC, 0xF294, 0x1294, 0x23FC, 0x4828, 0xFBFE, 0xA20, 0x324, 0xAAA8, 0xAA92, 0x842A, 0x8C6}},
    {17194, 2, {0x13DE, 0x1252, 0x23DE, 0x2A52, 0x4BDE, 0xF242, 0x127A, 0x228A, 0x4B12, 0xFADA, 0xA8A, 0x2DA, 0xAA8A, 0xAAFA, 0x820A, 0x204}},
    {17195, 2, {0x13D0, 0x1290, 0x23DE, 0x2A68, 0x4BC4, 0xF284, 0x13F8, 0x2108, 0x49F8, 0xFD08, 0x5F8, 0x108, 0x55F8, 0x5490, 0x8112, 0x60E}},
    {17196, 2, {0x1110, 0x1090, 0x23DE, 0x2810, 0x4A5E, 0xF182, 0x13DE, 0x2010, 0x4BDE, 0xFA50, 0xBDE, 0x250, 0xABDE, 0xAA50, 0x8252, 0x2CE}},
    {17197, 2, {0x1088, 0x1088, 0x23FE, 0x28A8, 0x4890, 0xF1FE, 0x1120, 0x23FC, 0x4D20, 0xF9FC, 0x920, 0x1FE, 0xA900, 0xAAA4, 0x8252, 0x452}},
    {17198, 2, {0x11FC, 0x1020, 0x23FE, 0x2A22, 0x49AC, 0xF020, 0x11AC, 0x2000, 0x489E, 0xFBF2, 0x89E, 0x1D2, 0xAABE, 0xAC92, 0x809E, 0x92}},
    {17199, 2, {0x1088, 0x13FE, 0x2088, 0x2BDE, 0x4A52, 0xF3DE, 0x10A0, 0x2090, 0x49FE, 0xF920, 0xBFC, 0x520, 0xA9FC, 0xA920, 0x81FE, 0x100}},
    {17200, 2, {0x13FC, 0x1108, 0x21F8, 0x2908, 0x49F8, 0xF10E, 0x17F8, 0x2008, 0x4FFE, 0xFA94, 0xB9C, 0x294, 0xAB9C, 0xAAD6, 0x87BC, 0x84}},
    {17201, 2, {0x4080, 0x7FFC, 0x4800, 0x8FF8, 0xA000, 0xEFF8, 0x4808, 0x4FF8, 0xA100, 0xFC9C, 0x15D4, 0x5C14, 0x555C, 0x5C96, 0x95D6, 0x2C22}},
    {17202, 2, {0x1088, 0x13FE, 0x2088, 0x29F0, 0x4820, 0xF3FE, 0x1020, 0x21FC, 0x49AC, 0xFD74, 0x5FC, 0x20, 0x55FC, 0x5420, 0x83FE, 0x154}},
    {17203, 2, {0x27FC, 0x2248, 0x47FC, 0x4C46, 0x96EC, 0xE554, 0x26EC, 0x4554, 0x96EC, 0xF454, 0x408, 0x17FC, 0xA840, 0xA3F8, 0x8050, 0x7FC}},
    {17204, 2, {0x4080, 0x7FFC, 0x4800, 0x8FF8, 0xA000, 0xEFF8, 0x4808, 0x4FF8, 0xA220, 0xFD5C, 0x17F4, 0x5C94, 0x55DC, 0x5C96, 0x97F6, 0x2CA2}},
    {17205, 2, {0x13FE, 0x1202, 0x23FE, 0x2A92, 0x4A54, 0xF292, 0x12FE, 0x22AA, 0x4AFE, 0xFB22, 0xAFA, 0x2AA, 0xAAFA, 0xAA22, 0x85FA, 0x4}},
    {17206, 2, {0x1010, 0x1010, 0x2010, 0x2410, 0x45FE, 0xF810, 0x1030, 0x2030, 0x4050, 0xFC50, 0x4090, 0x110, 0x1C10, 0xE010, 0x4050, 0x20}},
    {17207, 2, {0x1020, 0x1020, 0x2020, 0x2420, 0x45FC, 0xF924, 0x1124, 0x2124, 0x4124, 0xFDFC, 0x4124, 0x124, 0x1D24, 0xE124, 0x41FC, 0x104}},
    {17208, 2, {0x1020, 0x1020, 0x2020, 0x25FC, 0x4420, 0xF840, 0x13FE, 0x2040, 0x4080, 0xFDFC, 0x4004, 0x88, 0x1C50, 0xE020, 0x4010, 0x10}},
    {17209, 2, {0x1000, 0x13FC, 0x2204, 0x2204, 0x4A04, 0xFAF4, 0x1294, 0x2294, 0x4294, 0xFA94, 0x42F4, 0x294, 0x1A04, 0xE204, 0x4214, 0x208}},
    {17210, 2, {0x1000, 0x13FE, 0x2088, 0x2488, 0x44F8, 0xF888, 0x1088, 0x20F8, 0x4088, 0xFC88, 0x408E, 0x3F8, 0x1C08, 0xE008, 0x4008, 0x8}},
    {17211, 2, {0x1000, 0x13DE, 0x2042, 0x2042, 0x4A52, 0xF94A, 0x114A, 0x2042, 0x40C6, 0xF94A, 0x4252, 0x42, 0x1842, 0xE042, 0x414A, 0x84}},
    {17212, 2, {0x1104, 0x1084, 0x2088, 0x2410, 0x45FC, 0xF904, 0x1104, 0x2104, 0x41FC, 0xFC50, 0x4050, 0x90, 0x1C92, 0xE112, 0x420E, 0x400}},
    {17213, 2, {0x1008, 0x103C, 0x23C0, 0x2044, 0x4A24, 0xF928, 0x1100, 0x2020, 0x43FE, 0xF870, 0x40A8, 0xA8, 0x1924, 0xE222, 0x4420, 0x20}},
    {17214, 2, {0x1020, 0x1020, 0x23FE, 0x2020, 0x4820, 0xFBFE, 0x128A, 0x2252, 0x42FA, 0xFA22, 0x4222, 0x2FA, 0x1A22, 0xE222, 0x422A, 0x204}},
    {17215, 2, {0x1000, 0x13FE, 0x2000, 0x21FC, 0x4904, 0xF904, 0x11FC, 0x2000, 0x43FE, 0xFA22, 0x4222, 0x3FE, 0x1A22, 0xE222, 0x43FE, 0x202}},
    {17216, 2, {0x1090, 0x1290, 0x229C, 0x22A0, 0x4AD0, 0xFA88, 0x1288, 0x2080, 0x4000, 0xFBF8, 0x42A8, 0x2A8, 0x1AA8, 0xE2A8, 0x4FFE, 0x0}},
    {17217, 2, {0x1088, 0x1448, 0x2250, 0x20FC, 0x4820, 0xF850, 0x1694, 0x2238, 0x4250, 0xFA98, 0x4234, 0x254, 0x1A90, 0xE220, 0x45FE, 0x800}},
    {17218, 2, {0x2000, 0x20FC, 0x3C20, 0x5020, 0x9020, 0x1020, 0xFE20, 0x11FE, 0x1020, 0x5420, 0x5420, 0x5420, 0x5C20, 0x6420, 0x4A0, 0x40}},
    {17219, 2, {0x8, 0x1FC, 0x7E00, 0x2208, 0x1108, 0x1110, 0x800, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {17220, 2, {0x2020, 0x2020, 0x3C20, 0x5020, 0x903E, 0x1020, 0xFE20, 0x1020, 0x11FC, 0x5504, 0x5504, 0x5504, 0x5D04, 0x6504, 0x5FC, 0x104}},
    {17221, 2, {0x2020, 0x2020, 0x3C50, 0x5050, 0x9088, 0x1124, 0xFE12, 0x1010, 0x11FC, 0x5404, 0x5408, 0x5488, 0x5C50, 0x6420, 0x410, 0x10}},
    {17222, 2, {0x2040, 0x2020, 0x3C20, 0x51FE, 0x9102, 0x1204, 0xFE00, 0x1000, 0x11FE, 0x5420, 0x5420, 0x5420, 0x5C20, 0x6420, 0x4A0, 0x40}},
    {17223, 2, {0x2004, 0x201E, 0x3CE0, 0x5000, 0x9010, 0x1008, 0xFEFE, 0x1002, 0x1004, 0x5408, 0x5410, 0x5420, 0x5C40, 0x64A0, 0x51E, 0x0}},
    {17224, 2, {0x2000, 0x21FC, 0x3C20, 0x5020, 0x9124, 0x10A4, 0xFEA8, 0x1020, 0x13FE, 0x5420, 0x5420, 0x5420, 0x5C20, 0x6420, 0x420, 0x20}},
    {17225, 2, {0x2000, 0x2000, 0x3CFC, 0x5084, 0x9084, 0x1084, 0xFE84, 0x1084, 0x10FC, 0x5484, 0x5400, 0x5448, 0x5C44, 0x6484, 0x482, 0x102}},
    {17226, 2, {0x2040, 0x2020, 0x3C20, 0x51FE, 0x9000, 0x1088, 0xFF04, 0x1202, 0x1088, 0x5488, 0x5450, 0x5450, 0x5C20, 0x6450, 0x488, 0x306}},
    {17227, 2, {0x2008, 0x203C, 0x3DE0, 0x5020, 0x91FE, 0x10A8, 0xFEA8, 0x10A8, 0x11FE, 0x54A8, 0x54A8, 0x55FE, 0x5C20, 0x6420, 0x5FC, 0x0}},
    {17228, 2, {0x2040, 0x2020, 0x3CFC, 0x5000, 0x9088, 0x1048, 0xFE50, 0x11FE, 0x1000, 0x5400, 0x54FC, 0x5484, 0x5C84, 0x6484, 0x4FC, 0x84}},
    {17229, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x2044, 0x3C44, 0x5028, 0xFE28, 0x1010, 0x5428, 0x7C44, 0x182}},
    {17230, 2, {0x4088, 0x43FE, 0x7888, 0xA020, 0x23FE, 0x2250, 0xFBFE, 0x2252, 0x23FE, 0xAA80, 0xAAFE, 0xAA80, 0xBAFE, 0xCC02, 0xAAA, 0x4}},
    {17231, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2028, 0x2010}},
    {17232, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {17233, 2, {0x0, 0x3FF8, 0x2228, 0x2AA8, 0x2448, 0x2AA8, 0x3128, 0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {17234, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {17235, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x3FFC, 0x40, 0x1040, 0x2040, 0x3FFE, 0x140, 0x640, 0x1840, 0x6140, 0x80}},
    {17236, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x200, 0x200, 0x7FFC, 0x400, 0x480, 0x880, 0x1100, 0x2220, 0x4410, 0x8FF8, 0x408}},
    {17237, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0xF0, 0x3F00, 0x2100, 0x2100, 0x3FFC, 0x2080, 0x2040, 0x2A24, 0x3114, 0x210C}},
    {17238, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1090, 0x1F90, 0x1050, 0x1024, 0x1004, 0xFFC}},
    {17239, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x200, 0x100, 0x7FFC, 0x400, 0x820, 0x1FC0, 0x80, 0x320, 0xC10, 0x3FF8, 0x8}},
    {17240, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x1110, 0x920, 0x540, 0x7FFC, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {17241, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x1000, 0x3FFC, 0x4000, 0x9FF0, 0x1210, 0x1110, 0xFFFE, 0x2210, 0x2110, 0x3FFC, 0x10, 0x60}},
    {17242, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1108, 0x1090, 0x1260, 0x1418, 0x1806}},
    {17243, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x2FE8, 0x2108, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {17244, 2, {0x7FFC, 0x4444, 0x7FFC, 0x900, 0x880, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000}},
    {17245, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0x104, 0x7D98, 0x960, 0x3118, 0xC506, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {17246, 2, {0x7FFC, 0x4004, 0x4AA4, 0x4444, 0x4AA4, 0x50, 0x48, 0xFFFE, 0x40, 0x3E40, 0x2224, 0x3E28, 0x12, 0xE2A, 0x70C6, 0x302}},
    {17247, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x1010, 0x2208, 0x47E4, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {17248, 2, {0x7FFC, 0x4444, 0x7FFC, 0x100, 0x80, 0xFFFE, 0x440, 0x1390, 0x1450, 0x1FF0, 0x100, 0x7FFC, 0x4444, 0x4FE4, 0x4424, 0x400C}},
    {17249, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2848, 0x2F70, 0x4844, 0x4B44, 0x8C3C}},
    {17250, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x1000, 0x3FFC, 0x4AA0, 0xAA0, 0x7FFC, 0xAA0, 0xAA0, 0x7FFC, 0x0, 0x2488, 0x2244, 0x4244}},
    {17251, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {17252, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x100, 0xFFFE, 0x288, 0x7D70, 0x2548, 0x4D66, 0x9010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x2010}},
    {17253, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x7FFE, 0x4000, 0x5F3C, 0x4410, 0x7F7E, 0x4E34, 0x5552, 0x6090, 0x44F8, 0x4480, 0xBFFE}},
    {17254, 2, {0x7FFC, 0x4444, 0x7FFC, 0x1000, 0x828, 0x7F24, 0x147E, 0x4948, 0x55C8, 0x7F7E, 0x848, 0x7F7E, 0x5148, 0x5548, 0x5D7E, 0x4340}},
    {17255, 2, {0x1088, 0x1048, 0x5050, 0x5000, 0x7DFC, 0x5020, 0x9020, 0x10FC, 0x1C20, 0xF020, 0x51FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {17256, 2, {0x4420, 0x2420, 0x2820, 0xFDFE, 0x1020, 0x1020, 0x7DFC, 0x1124, 0x1124, 0xFD24, 0x1124, 0x1134, 0x2128, 0x2020, 0x4020, 0x8020}},
    {17257, 2, {0x4440, 0x2440, 0x2840, 0xFE80, 0x10FE, 0x1108, 0x7E88, 0x1088, 0x1088, 0xFE50, 0x1050, 0x1020, 0x2050, 0x2088, 0x4104, 0x8202}},
    {17258, 2, {0x4420, 0x2420, 0x2820, 0xFDFE, 0x1020, 0x1020, 0x7C20, 0x11FC, 0x1070, 0xFCA8, 0x10A8, 0x1124, 0x2124, 0x2222, 0x4020, 0x8020}},
    {17259, 2, {0x4420, 0x2410, 0x2810, 0xFDFE, 0x1102, 0x1204, 0x7C80, 0x1088, 0x1090, 0xFCA0, 0x10C0, 0x1082, 0x2082, 0x2082, 0x407E, 0x8000}},
    {17260, 2, {0x4400, 0x25FC, 0x2820, 0xFC20, 0x1124, 0x10A4, 0x7CA8, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x1020, 0x2020, 0x2020, 0x4020, 0x8020}},
    {17261, 2, {0x4400, 0x24FC, 0x2884, 0xFEA4, 0x1094, 0x1084, 0x7C84, 0x13FE, 0x1104, 0xFF44, 0x1124, 0x1104, 0x21FE, 0x2004, 0x4028, 0x8010}},
    {17262, 2, {0x4450, 0x2450, 0x2850, 0xFD52, 0x10D4, 0x1058, 0x7C50, 0x1058, 0x10D4, 0xFD52, 0x1050, 0x1050, 0x2092, 0x2092, 0x4112, 0x820E}},
    {17263, 2, {0x4440, 0x2440, 0x28FC, 0xFD04, 0x1208, 0x11FE, 0x7D00, 0x117C, 0x1144, 0xFD44, 0x1154, 0x1148, 0x2142, 0x2242, 0x423E, 0x8400}},
    {17264, 2, {0x4400, 0x25FC, 0x2904, 0xFD24, 0x1124, 0x1124, 0x7DFC, 0x1124, 0x1124, 0xFD54, 0x114C, 0x118C, 0x2104, 0x2104, 0x41FC, 0x8104}},
    {17265, 2, {0x4420, 0x2420, 0x2850, 0xFC88, 0x1104, 0x12FA, 0x7C20, 0x1020, 0x13FE, 0xFC20, 0x1128, 0x1124, 0x2222, 0x2422, 0x40A0, 0x8040}},
    {17266, 2, {0x8800, 0x4840, 0x539C, 0xFA04, 0x2204, 0x239C, 0xFA04, 0x2204, 0x23FC, 0xF890, 0x2090, 0x2090, 0x2090, 0x4112, 0x4212, 0x840E}},
    {17267, 2, {0x8800, 0x4BDE, 0x5042, 0xF94A, 0x2084, 0x214A, 0xFA52, 0x2420, 0x2000, 0xFBDE, 0x2052, 0x2152, 0x2094, 0x4148, 0x4254, 0x8422}},
    {17268, 2, {0x881C, 0x49E0, 0x5020, 0xF820, 0x23FE, 0x20A8, 0xF924, 0x2242, 0x2040, 0xFBFE, 0x2088, 0x2108, 0x20D0, 0x4030, 0x4048, 0x8184}},
    {17269, 2, {0x4440, 0x2440, 0x28F8, 0xFD08, 0x1010, 0x11FC, 0x7C24, 0x1024, 0x13FE, 0xFC24, 0x1024, 0x11FC, 0x2024, 0x2020, 0x40A0, 0x8040}},
    {17270, 2, {0x4420, 0x2420, 0x2BFE, 0xFC20, 0x11FC, 0x1124, 0x7DFC, 0x1124, 0x11FC, 0xFC20, 0x1070, 0x10A8, 0x2124, 0x2622, 0x4020, 0x8020}},
    {17271, 2, {0x4440, 0x2440, 0x29FC, 0xFC84, 0x13FE, 0x1000, 0x7DFC, 0x1104, 0x11FC, 0xFC20, 0x13FE, 0x1020, 0x2220, 0x23FE, 0x4020, 0x8020}},
    {17272, 2, {0x4428, 0x2424, 0x2BFE, 0xFC20, 0x11FC, 0x1124, 0x7DFC, 0x1124, 0x11FC, 0xFD24, 0x1008, 0x13FE, 0x2088, 0x2048, 0x4048, 0x8018}},
    {17273, 2, {0x4420, 0x2440, 0x29FC, 0xFD04, 0x11FC, 0x1104, 0x7DFC, 0x1104, 0x11FC, 0xFC28, 0x1024, 0x13FE, 0x2050, 0x2088, 0x4104, 0x8602}},
    {17274, 2, {0x8820, 0x4820, 0x51FC, 0xF924, 0x21FC, 0x2020, 0xFBFE, 0x2222, 0x23FE, 0xF820, 0x2040, 0x2024, 0x22A2, 0x428A, 0x4478, 0x8000}},
    {17275, 2, {0x8800, 0x4BDE, 0x5252, 0xFBDE, 0x2210, 0x2252, 0xF9CE, 0x2000, 0x2088, 0xFBFE, 0x2088, 0x2088, 0x23FE, 0x4088, 0x4104, 0x8202}},
    {17276, 2, {0x8800, 0x4BDE, 0x5088, 0xFAA8, 0x23FE, 0x2188, 0xFADA, 0x24A6, 0x21F8, 0xF908, 0x2108, 0x21F8, 0x2108, 0x4108, 0x41F8, 0x8108}},
    {17277, 2, {0x8800, 0x4BFE, 0x5222, 0xFACC, 0x2244, 0x23EE, 0xFA44, 0x22EE, 0x2354, 0xFA44, 0x2210, 0x2290, 0x229E, 0x4490, 0x4490, 0x8BFE}},
    {17278, 2, {0x0, 0x7EEE, 0x2422, 0x2422, 0x24AA, 0x2466, 0x2422, 0xFF22, 0x2426, 0x246A, 0x24B2, 0x2422, 0x2422, 0x2422, 0x44AA, 0x8444}},
    {17279, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x204, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {17280, 2, {0x14, 0xEE12, 0x2212, 0x2210, 0xAAFE, 0x6690, 0x2292, 0x2292, 0x2692, 0x6A94, 0xB294, 0x22B4, 0x22CA, 0x228A, 0xAA16, 0x4422}},
    {17281, 2, {0x1000, 0x11DC, 0x1044, 0x1044, 0xFF54, 0x10CC, 0x1044, 0x1044, 0x7C4C, 0x44D4, 0x4564, 0x4444, 0x4444, 0x7C44, 0x4554, 0x88}},
    {17282, 2, {0x40, 0xEE40, 0x22FC, 0x2284, 0xAB04, 0x66F4, 0x2294, 0x2294, 0x26F4, 0x6A84, 0xB294, 0x2288, 0x2282, 0x2282, 0xAA7E, 0x4400}},
    {17283, 2, {0x1000, 0x11DC, 0x1044, 0x7C44, 0x5554, 0x54CC, 0x7C44, 0x5444, 0x544C, 0x54D4, 0xFF64, 0x4444, 0x4444, 0x4444, 0x5554, 0x4888}},
    {17284, 2, {0x0, 0xFDDC, 0x2444, 0x2444, 0x2554, 0x24CC, 0x5444, 0x4844, 0x804C, 0x7CD4, 0x4564, 0x4444, 0x4444, 0x4444, 0x7D54, 0x4488}},
    {17285, 2, {0x1000, 0x11DC, 0x2044, 0x7C44, 0x4554, 0x44CC, 0x4444, 0x4444, 0x7C4C, 0x44D4, 0x4564, 0x4444, 0x4444, 0x7C44, 0x4554, 0x88}},
    {17286, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x304, 0x280, 0xC60, 0x3218, 0xC106, 0x1FE0, 0x20, 0x640, 0x180, 0x40}},
    {17287, 2, {0x1000, 0x21DC, 0x7C44, 0x4444, 0x6554, 0x54CC, 0x5444, 0xFC44, 0x444C, 0x64D4, 0x5564, 0x5444, 0x4444, 0x4444, 0x5554, 0x8888}},
    {17288, 2, {0x20, 0xEE20, 0x22FC, 0x22A4, 0xAAA4, 0x66FC, 0x22A4, 0x22A4, 0x26FC, 0x6A20, 0xB224, 0x2218, 0x2212, 0x226A, 0xAB86, 0x4402}},
    {17289, 2, {0xC, 0xEE0A, 0x2208, 0x22FE, 0xAA88, 0x6688, 0x2288, 0x228A, 0x26EA, 0x6A8A, 0xB28C, 0x228C, 0x228A, 0x228A, 0xAA96, 0x4522}},
    {17290, 2, {0x1000, 0x10EE, 0x1E22, 0x2222, 0x62AA, 0x9466, 0x822, 0x1422, 0x2226, 0x416A, 0xBEB2, 0x2222, 0x2222, 0x2222, 0x3EAA, 0x2244}},
    {17291, 2, {0x800, 0x28EE, 0x2822, 0x3E22, 0x48AA, 0x866, 0xFF22, 0x22, 0x26, 0x3E6A, 0x22B2, 0x2222, 0x2222, 0x3E22, 0x22AA, 0x44}},
    {17292, 2, {0x0, 0xEE7C, 0x2244, 0x2244, 0xAA44, 0x6644, 0x227C, 0x2210, 0x2610, 0x6A50, 0xB25E, 0x2250, 0x2250, 0x2270, 0xAA9E, 0x4500}},
    {17293, 2, {0x28, 0xEC24, 0x25FE, 0x2420, 0xB420, 0x6DFC, 0x2524, 0x2524, 0x25FC, 0x6D24, 0xB524, 0x25FC, 0x2524, 0x2524, 0xB524, 0x490C}},
    {17294, 2, {0x0, 0xEDFE, 0x24A0, 0x24A8, 0xB4A8, 0x6CE8, 0x24A8, 0x24A8, 0x24A8, 0x6CE8, 0xB4A8, 0x24AA, 0x24BA, 0x25EA, 0xB426, 0x4820}},
    {17295, 2, {0x800, 0x8EE, 0x1422, 0x2222, 0x51AA, 0x8866, 0x7E22, 0x222, 0x426, 0x86A, 0x7EB2, 0x4222, 0x4222, 0x7E22, 0x42AA, 0x44}},
    {17296, 2, {0x1000, 0x11DC, 0xFE44, 0x1044, 0x7D54, 0x44CC, 0x7C44, 0x4444, 0x7C4C, 0x54D4, 0x1164, 0xFE44, 0x1044, 0x1044, 0x1154, 0x1088}},
    {17297, 2, {0x0, 0xEEFC, 0x2204, 0x2204, 0xAA7C, 0x6604, 0x2204, 0x22FE, 0x2610, 0x6A92, 0xB254, 0x2238, 0x2254, 0x2292, 0xAA50, 0x4420}},
    {17298, 2, {0x48, 0xEE44, 0x225E, 0x23E0, 0xAA28, 0x6612, 0x226A, 0x2396, 0x2648, 0x6A5E, 0xB3E0, 0x2224, 0x2228, 0x2212, 0xAA6A, 0x4586}},
    {17299, 2, {0x10, 0xEE10, 0x22DA, 0x2254, 0xAA54, 0x6692, 0x2350, 0x2220, 0x26FE, 0x6A82, 0xB282, 0x22FE, 0x2282, 0x2282, 0xAAFE, 0x4482}},
    {17300, 2, {0x28, 0xEC24, 0x27FE, 0x2420, 0xB5FC, 0x6D24, 0x25FC, 0x2524, 0x25FC, 0x6D24, 0xB408, 0x27FE, 0x2488, 0x2448, 0xB448, 0x4818}},
    {17301, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0x3EF8, 0x208, 0x1248, 0xA28, 0x1248, 0x2288, 0xA28, 0x410}},
    {17302, 2, {0x4400, 0x24EE, 0x2822, 0xFE22, 0x92AA, 0xD666, 0xBA22, 0x9222, 0xFE26, 0x6A, 0x7CB2, 0x4422, 0x7C22, 0x4422, 0x7CAA, 0x4444}},
    {17303, 2, {0x600, 0x78EE, 0x1022, 0xFFA2, 0x22AA, 0x5D66, 0x94A2, 0x1C22, 0x26, 0x7F6A, 0x41B2, 0x5D22, 0x5522, 0x5D22, 0x41AA, 0x4344}},
    {17304, 2, {0x800, 0x7F7E, 0x912, 0xFF92, 0x95A, 0x7F36, 0x812, 0xAA92, 0xEB96, 0x88BA, 0xFFD2, 0x8892, 0xEB92, 0xAA92, 0xAA92, 0x8B6}},
    {17305, 2, {0x2000, 0x11DC, 0xFE44, 0x8244, 0x7D54, 0x54CC, 0x9844, 0x7C44, 0xC44C, 0x7CD4, 0x4564, 0x7C44, 0x4444, 0x7C44, 0x2954, 0x4488}},
    {17306, 2, {0x88, 0xEDDC, 0x2488, 0x25DC, 0xB488, 0x6DDC, 0x2488, 0x25FC, 0x2404, 0x6CFC, 0xB404, 0x25FC, 0x2420, 0x2554, 0xB54A, 0x4A3A}},
    {17307, 2, {0x200, 0x208, 0x3FD0, 0x220, 0xFFFE, 0x200, 0xC70, 0x3F84, 0xC804, 0x7FC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {17308, 2, {0x1010, 0x1210, 0x7E12, 0x147E, 0x1414, 0xFE14, 0x8FE, 0x1008, 0x2030, 0x7E22, 0xA064, 0x22B8, 0x2D22, 0x3022, 0x201E, 0x0}},
    {17309, 2, {0x220, 0x3FC0, 0x280, 0xFFFE, 0x220, 0xFC0, 0x3408, 0xC3F8, 0x0, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x6AD4, 0x408}},
    {17310, 2, {0x100, 0x108, 0x1FD0, 0x120, 0x7FFC, 0x200, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x0, 0x3FF8, 0x8, 0x8, 0x8}},
    {17311, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFC40, 0x25FC, 0x2554, 0x2554, 0x2554, 0x4954, 0x2954, 0x1154, 0x2954, 0x4544, 0x8114, 0x108}},
    {17312, 2, {0x2000, 0x23FE, 0x2020, 0x2040, 0xFDFC, 0x2554, 0x2554, 0xA554, 0x6554, 0x2554, 0x3544, 0x2D16, 0x250A, 0x4402, 0x43FE, 0x8000}},
    {17313, 2, {0x0, 0xFFFE, 0x200, 0x400, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {17314, 2, {0x1040, 0x1040, 0x1080, 0xFCFE, 0x1100, 0x7A00, 0x10FC, 0xFC08, 0x1010, 0x3820, 0x3440, 0x5480, 0x9102, 0x1102, 0x10FE, 0x1000}},
    {17315, 2, {0x1008, 0x1048, 0x1128, 0xFD28, 0x1128, 0x7908, 0x1108, 0xFD08, 0x1108, 0x3908, 0x3548, 0x5594, 0x9114, 0x1022, 0x1042, 0x1082}},
    {17316, 2, {0x1000, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x7900, 0x1100, 0xFDFE, 0x1102, 0x397A, 0x354A, 0x554A, 0x927A, 0x1202, 0x1414, 0x1008}},
    {17317, 2, {0x1000, 0x1050, 0x1048, 0xFC84, 0x1124, 0x7820, 0x1050, 0xFC88, 0x1106, 0x3AF8, 0x3488, 0x5488, 0x9088, 0x1088, 0x10F8, 0x1088}},
    {17318, 2, {0x1020, 0x10A4, 0x10A4, 0xFD28, 0x1050, 0x7888, 0x1104, 0xFC24, 0x1020, 0x38A8, 0x34A8, 0x5530, 0x9050, 0x1088, 0x1104, 0x1202}},
    {17319, 2, {0x1040, 0x1020, 0x10FC, 0xFC00, 0x1088, 0x7848, 0x1050, 0xFDFE, 0x1000, 0x3800, 0x34FC, 0x5484, 0x9084, 0x1084, 0x10FC, 0x1084}},
    {17320, 2, {0x1020, 0x1020, 0x13FE, 0xFC50, 0x1088, 0x7924, 0x1222, 0xFDFC, 0x1124, 0x39FC, 0x3524, 0x55FC, 0x9022, 0x1022, 0x101E, 0x1000}},
    {17321, 2, {0x1092, 0x1092, 0x1124, 0xFE48, 0x1124, 0x7892, 0x1092, 0xFC00, 0x11FE, 0x3922, 0x3522, 0x55FE, 0x9122, 0x1122, 0x11FE, 0x1102}},
    {17322, 2, {0x1000, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x79FC, 0x1000, 0xFDFE, 0x1008, 0x3BFE, 0x3408, 0x5508, 0x9088, 0x1008, 0x1028, 0x1010}},
    {17323, 2, {0x1050, 0x1124, 0x1154, 0xFD04, 0x11FC, 0x7850, 0x1088, 0xFD44, 0x1078, 0x3888, 0x3488, 0x5550, 0x9020, 0x1050, 0x1088, 0x1106}},
    {17324, 2, {0x1000, 0x13FE, 0x1252, 0xFE52, 0x13FE, 0x7820, 0x1020, 0xFDFC, 0x1020, 0x3820, 0x37FE, 0x5440, 0x9088, 0x1104, 0x13FE, 0x1102}},
    {17325, 2, {0x1010, 0x1050, 0x1250, 0xFD7C, 0x1090, 0x7810, 0x10FE, 0xFF00, 0x117C, 0x3944, 0x3544, 0x517C, 0x5144, 0x9100, 0x12FE, 0x1400}},
    {17326, 2, {0x1040, 0x1020, 0x11FC, 0xFC88, 0x1050, 0x79FC, 0x1104, 0xFD24, 0x11FC, 0x3924, 0x3574, 0x5554, 0x9154, 0x1174, 0x1104, 0x110C}},
    {17327, 2, {0x1088, 0x1088, 0x13FE, 0xFC88, 0x10F8, 0x7820, 0x11FC, 0xFD24, 0x11FC, 0x3820, 0x35FC, 0x5420, 0x93FE, 0x1050, 0x1088, 0x1306}},
    {17328, 2, {0x1142, 0x115C, 0x13F0, 0xFD50, 0x1150, 0x795E, 0x13F4, 0xFC14, 0x11D4, 0x3954, 0x3554, 0x55D4, 0x9154, 0x1154, 0x11E4, 0x1004}},
    {17329, 2, {0x2000, 0x23FC, 0x2294, 0xFA94, 0x23FC, 0x7110, 0x2252, 0xFBDC, 0x2012, 0x73CE, 0x6A40, 0xA3D2, 0x225C, 0x23D0, 0x2252, 0x22CE}},
    {17330, 2, {0x0, 0xFF00, 0x24FC, 0x2404, 0x3C08, 0x2410, 0x2420, 0x3C20, 0x2440, 0x2440, 0x2E80, 0xF482, 0x4482, 0x482, 0x47E, 0x400}},
    {17331, 2, {0x0, 0xFF7C, 0x2444, 0x2444, 0x3C44, 0x247C, 0x2444, 0x3C44, 0x2444, 0x247C, 0x2E44, 0xF444, 0x4444, 0x484, 0x494, 0x508}},
    {17332, 2, {0x48, 0xFF48, 0x2448, 0x2448, 0x3CFC, 0x2448, 0x2448, 0x3C48, 0x25FE, 0x2448, 0x2E48, 0xF448, 0x4448, 0x488, 0x488, 0x508}},
    {17333, 2, {0x40, 0xFF20, 0x2428, 0x2408, 0x3C08, 0x2448, 0x2454, 0x3D52, 0x2552, 0x2560, 0x2E60, 0xF444, 0x44C4, 0x544, 0x63C, 0x400}},
    {17334, 2, {0x20, 0xFC20, 0x4820, 0x49FC, 0x7924, 0x4924, 0x4924, 0x79FC, 0x4924, 0x4924, 0x4D24, 0x79FC, 0xC924, 0x820, 0x820, 0x820}},
    {17335, 2, {0x20, 0xFD20, 0x4920, 0x49FC, 0x7920, 0x4A20, 0x4820, 0x7BFE, 0x4870, 0x48A8, 0x4CA8, 0x7924, 0xC924, 0xA22, 0x820, 0x820}},
    {17336, 2, {0x0, 0xFDFE, 0x4820, 0x4820, 0x78FC, 0x4844, 0x4844, 0x79FE, 0x4800, 0x4800, 0x4CFC, 0x7884, 0xC884, 0x884, 0x8FC, 0x884}},
    {17337, 2, {0xF8, 0x3F00, 0x1110, 0x920, 0xFFFE, 0x920, 0x3118, 0xC106, 0x3FF8, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20}},
    {17338, 2, {0x82, 0xFC82, 0x4A82, 0x4BEA, 0x7C8A, 0x488A, 0x4FFA, 0x788A, 0x4BEA, 0x4AAA, 0x4EAA, 0x7AAA, 0xCAA2, 0xAE2, 0x88A, 0x884}},
    {17339, 2, {0x0, 0xFDFC, 0x4924, 0x4924, 0x7974, 0x4924, 0x49FC, 0x7904, 0x4974, 0x4954, 0x4D54, 0x7974, 0xC904, 0xA04, 0xA14, 0xC08}},
    {17340, 2, {0x0, 0xFDF8, 0x4808, 0x4808, 0x79F8, 0x4808, 0x4808, 0x7BFE, 0x4820, 0x4A22, 0x4D74, 0x78A8, 0xC924, 0xA22, 0x8A0, 0x840}},
    {17341, 2, {0x1C, 0xFDE0, 0x4920, 0x4920, 0x79FE, 0x4910, 0x494A, 0x7986, 0x4802, 0x49FC, 0x4D04, 0x7904, 0xC9FC, 0x904, 0x904, 0x9FC}},
    {17342, 2, {0x20, 0xFC20, 0x4850, 0x4888, 0x7904, 0x4AFA, 0x4800, 0x79FC, 0x4954, 0x4954, 0x4DFC, 0x7954, 0xC954, 0x954, 0x904, 0x90C}},
    {17343, 2, {0x8, 0xFCC8, 0x4B08, 0x4908, 0x792A, 0x492A, 0x4BAC, 0x7948, 0x4908, 0x4B88, 0x4D54, 0x7914, 0xC914, 0x924, 0x924, 0x942}},
    {17344, 2, {0x82, 0xFC82, 0x4BE2, 0x4882, 0x788A, 0x4BEA, 0x4AAA, 0x7AAA, 0x4BEA, 0x488A, 0x4DCA, 0x7AAA, 0xCCA2, 0x882, 0x88A, 0x884}},
    {17345, 2, {0x40, 0xFC20, 0x4BFE, 0x4A42, 0x7820, 0x49FC, 0x4888, 0x7850, 0x4BFE, 0x4820, 0x4C20, 0x79FC, 0xC820, 0x820, 0x820, 0x820}},
    {17346, 2, {0xA0, 0xFD2C, 0x4924, 0x4924, 0x79AC, 0x4924, 0x4924, 0x79FC, 0x4820, 0x49FC, 0x4C88, 0x7850, 0xC820, 0x850, 0x888, 0xB06}},
    {17347, 2, {0x0, 0xFDFE, 0x4902, 0x4800, 0x78FC, 0x4884, 0x48FC, 0x7884, 0x48FC, 0x4820, 0x4C10, 0x79FE, 0xC800, 0x848, 0x884, 0x902}},
    {17348, 2, {0x40, 0xFC20, 0x4BFE, 0x4A8A, 0x7904, 0x4BFE, 0x4840, 0x7888, 0x49FC, 0x4824, 0x4C20, 0x79FC, 0xC820, 0x820, 0xBFE, 0x800}},
    {17349, 2, {0x80, 0xFC88, 0x49EE, 0x492A, 0x7AAA, 0x4944, 0x4884, 0x797A, 0x4A00, 0x4800, 0x4DFE, 0x7820, 0xC8A8, 0x924, 0xAA2, 0x840}},
    {17350, 2, {0x50, 0xFC50, 0x4BFE, 0x4850, 0x7BFE, 0x4A52, 0x4BFE, 0x7A52, 0x4BFE, 0x4800, 0x4DFC, 0x7904, 0xC9FC, 0x904, 0x9FC, 0x904}},
    {17351, 2, {0x0, 0xFDFC, 0x4850, 0x49FC, 0x7954, 0x4954, 0x49FC, 0x7800, 0x49FC, 0x4800, 0x4FFE, 0x7820, 0xC8A8, 0x924, 0xAA2, 0x840}},
    {17352, 2, {0x108, 0xF908, 0x5254, 0x5254, 0x74BC, 0x5708, 0x5110, 0x7224, 0x54BE, 0x578A, 0x5808, 0x70AC, 0xD56A, 0x154A, 0x1428, 0x1010}},
    {17353, 2, {0x0, 0xFDFC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4888, 0x7954, 0x49DC, 0x4888, 0x4D54, 0x79DC, 0xC800, 0x9A8, 0x954, 0xA54}},
    {17354, 2, {0x110, 0xFC90, 0x4BDE, 0x4810, 0x7A5E, 0x4982, 0x4BDE, 0x7810, 0x4BDE, 0x4A50, 0x4FDE, 0x7A50, 0xCBDE, 0xA50, 0xA52, 0xACE}},
    {17355, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x540, 0x920, 0x3118, 0xC106, 0x2578, 0x1348, 0xA178, 0xA544, 0x1D3C, 0x100}},
    {17356, 2, {0x840, 0x820, 0x1010, 0x2008, 0x4006, 0x9FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {17357, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {17358, 2, {0x0, 0x3C00, 0x24F0, 0x2490, 0x2490, 0x3C90, 0x2490, 0x24D0, 0x24B0, 0x3C90, 0x2490, 0x2492, 0x2492, 0x4512, 0x550E, 0x8A00}},
    {17359, 2, {0x0, 0x3DFC, 0x2420, 0x2420, 0x2420, 0x3C20, 0x27FE, 0x2420, 0x2420, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x54A0, 0x8840}},
    {17360, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2424, 0x3C24, 0x2424, 0x2424, 0x27FE, 0x3C20, 0x2450, 0x2450, 0x2488, 0x4488, 0x5504, 0x8A02}},
    {17361, 2, {0x200, 0x400, 0xFF8, 0x1210, 0x6920, 0x4C0, 0x300, 0xC00, 0x7100, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2288, 0x2448, 0x2018}},
    {17362, 2, {0x8, 0x3C88, 0x2488, 0x2488, 0x2488, 0x3CFE, 0x2480, 0x2480, 0x2480, 0x3CFC, 0x2484, 0x2484, 0x2484, 0x4504, 0x5504, 0x8A04}},
    {17363, 2, {0x0, 0x3CFC, 0x2400, 0x2400, 0x3400, 0x2DFE, 0x2450, 0x2450, 0x2C50, 0x3450, 0x2450, 0x2452, 0x2452, 0x4492, 0x548E, 0x8900}},
    {17364, 2, {0x0, 0x3DF8, 0x2448, 0x2448, 0x2448, 0x3C48, 0x2448, 0x25F8, 0x2488, 0x3C88, 0x2488, 0x2488, 0x2488, 0x4488, 0x57FE, 0x8800}},
    {17365, 2, {0x8, 0x3C3C, 0x25E0, 0x2420, 0x2420, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x55FC, 0x8800}},
    {17366, 2, {0x4, 0x3DE4, 0x2424, 0x2424, 0x2424, 0x3DE4, 0x2504, 0x2504, 0x2504, 0x3DE4, 0x2424, 0x2424, 0x2424, 0x4424, 0x5544, 0x8884}},
    {17367, 2, {0x80, 0x3C80, 0x24FC, 0x2500, 0x2600, 0x3CF8, 0x2400, 0x2400, 0x25F8, 0x3C08, 0x2408, 0x2408, 0x240A, 0x440A, 0x5406, 0x8802}},
    {17368, 2, {0x0, 0x3DFE, 0x2528, 0x2528, 0x2528, 0x3D28, 0x2528, 0x2528, 0x252A, 0x3D2A, 0x254A, 0x2546, 0x2580, 0x4500, 0x55FE, 0x8800}},
    {17369, 2, {0x20, 0x3C20, 0x2420, 0x25FE, 0x2522, 0x3E24, 0x2420, 0x2420, 0x2450, 0x3C50, 0x2450, 0x2450, 0x2490, 0x4492, 0x5512, 0x8A0E}},
    {17370, 2, {0x20, 0x3C20, 0x2420, 0x24A8, 0x24A4, 0x3CA2, 0x2522, 0x2520, 0x2624, 0x3C24, 0x2428, 0x2408, 0x2410, 0x4420, 0x54C0, 0x8B00}},
    {17371, 2, {0x40, 0x3C40, 0x2480, 0x24FC, 0x2504, 0x3E04, 0x2484, 0x2444, 0x2444, 0x3C14, 0x2424, 0x2444, 0x2584, 0x4404, 0x5428, 0x8810}},
    {17372, 2, {0x0, 0x7BFC, 0x4884, 0x4888, 0x4888, 0x7890, 0x489C, 0x4884, 0x4944, 0x7944, 0x4928, 0x4928, 0x4A10, 0x4A28, 0x4C44, 0x9982}},
    {17373, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2420, 0x3C40, 0x27FE, 0x2440, 0x2480, 0x3DFC, 0x2404, 0x2488, 0x2450, 0x4420, 0x5410, 0x8810}},
    {17374, 2, {0x20, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x3C20, 0x2420, 0x25FC, 0x2470, 0x3CA8, 0x24A8, 0x2524, 0x2524, 0x4622, 0x5420, 0x8820}},
    {17375, 2, {0x40, 0x3C20, 0x2428, 0x2408, 0x2408, 0x3C48, 0x2454, 0x2552, 0x2552, 0x3D60, 0x2660, 0x2444, 0x24C4, 0x4544, 0x563C, 0x8800}},
    {17376, 2, {0x40, 0x7820, 0x4820, 0x4800, 0x4BFC, 0x7800, 0x4808, 0x4908, 0x4908, 0x7890, 0x4890, 0x4890, 0x48A0, 0x4820, 0x4BFE, 0x9800}},
    {17377, 2, {0x0, 0x3CFC, 0x2400, 0x2400, 0x2400, 0x3DFE, 0x2420, 0x2420, 0x24A8, 0x3CA4, 0x2524, 0x2522, 0x2622, 0x4420, 0x54A0, 0x8840}},
    {17378, 2, {0x10, 0x3C10, 0x2410, 0x24FE, 0x2492, 0x3C94, 0x2490, 0x24FC, 0x24A4, 0x3CA4, 0x24A8, 0x24A8, 0x2490, 0x4528, 0x5544, 0x8A82}},
    {17379, 2, {0x0, 0x3CF8, 0x2488, 0x2488, 0x2488, 0x3CF8, 0x2488, 0x2488, 0x2488, 0x3CF8, 0x2488, 0x2488, 0x2488, 0x4488, 0x57FE, 0x8800}},
    {17380, 2, {0x0, 0x3DFC, 0x2404, 0x2404, 0x25F4, 0x3C04, 0x2404, 0x25F4, 0x2514, 0x3D14, 0x25F4, 0x2514, 0x2404, 0x4404, 0x5428, 0x8810}},
    {17381, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2524, 0x3D24, 0x25FC, 0x2524, 0x2524, 0x3D24, 0x27FE, 0x2504, 0x2504, 0x4504, 0x5514, 0x8908}},
    {17382, 2, {0x8, 0x3C3C, 0x25E0, 0x2420, 0x2420, 0x3FFE, 0x2420, 0x2420, 0x2420, 0x3DFC, 0x2504, 0x2504, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {17383, 2, {0x20, 0x3C20, 0x247C, 0x2484, 0x2548, 0x3C30, 0x2420, 0x2448, 0x2590, 0x3C3E, 0x2442, 0x25A4, 0x2418, 0x4410, 0x5460, 0x8980}},
    {17384, 2, {0x40, 0x3C80, 0x25F8, 0x2508, 0x2508, 0x3D08, 0x25F8, 0x2500, 0x2500, 0x3D00, 0x25FC, 0x2504, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {17385, 2, {0x20, 0x3C20, 0x2450, 0x2450, 0x2488, 0x3D04, 0x26FA, 0x2400, 0x2400, 0x3CF8, 0x2488, 0x2488, 0x2488, 0x4488, 0x54F8, 0x8888}},
    {17386, 2, {0x0, 0x3DFE, 0x2488, 0x2488, 0x2488, 0x3CF8, 0x2488, 0x2488, 0x24F8, 0x3C88, 0x2488, 0x249E, 0x27E8, 0x4408, 0x5408, 0x8808}},
    {17387, 2, {0x90, 0x7890, 0x4890, 0x4910, 0x49FE, 0x7B10, 0x4D10, 0x4938, 0x4938, 0x7954, 0x4954, 0x4992, 0x4910, 0x4910, 0x4910, 0x9910}},
    {17388, 2, {0x20, 0x3C20, 0x2448, 0x2484, 0x25FE, 0x3C12, 0x2490, 0x2490, 0x24FE, 0x3D10, 0x2410, 0x25FE, 0x2410, 0x4410, 0x5410, 0x8810}},
    {17389, 2, {0x20, 0x7920, 0x4920, 0x49FC, 0x4920, 0x7A20, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x48A8, 0x4924, 0x4924, 0x4A22, 0x4820, 0x9820}},
    {17390, 2, {0x40, 0x3C40, 0x25FC, 0x2444, 0x2444, 0x3C84, 0x2494, 0x2508, 0x2400, 0x3D02, 0x2484, 0x2448, 0x2400, 0x4448, 0x5484, 0x8902}},
    {17391, 2, {0x20, 0x7820, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x49FC, 0x4840, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x4A8A, 0x4C78, 0x9800}},
    {17392, 2, {0x0, 0x7BFC, 0x4844, 0x4944, 0x4944, 0x7A84, 0x4894, 0x4908, 0x4A40, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x4A8A, 0x4C78, 0x9800}},
    {17393, 2, {0x0, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x7900, 0x4900, 0x49FE, 0x4902, 0x797A, 0x494A, 0x494A, 0x4A7A, 0x4A02, 0x4C14, 0x9808}},
    {17394, 2, {0x108, 0x3C88, 0x2490, 0x25FC, 0x2424, 0x3C24, 0x25FC, 0x2520, 0x2520, 0x3DFE, 0x2462, 0x24A2, 0x252A, 0x4624, 0x5420, 0x8820}},
    {17395, 2, {0x104, 0x790E, 0x4930, 0x4920, 0x4FE0, 0x7920, 0x493E, 0x4964, 0x49A4, 0x7F24, 0x4924, 0x4924, 0x4924, 0x4924, 0x4944, 0x9B84}},
    {17396, 2, {0x20, 0x3D24, 0x24A4, 0x24A8, 0x2420, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2504, 0x4504, 0x5514, 0x8908}},
    {17397, 2, {0x90, 0x7888, 0x4880, 0x4BFE, 0x48A0, 0x78A8, 0x48B0, 0x48A4, 0x48A8, 0x78B0, 0x4924, 0x492A, 0x4932, 0x4A22, 0x4ADE, 0x9C00}},
    {17398, 2, {0x110, 0x3D12, 0x2514, 0x25D8, 0x2510, 0x3D12, 0x2552, 0x258E, 0x2520, 0x3C20, 0x25FC, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {17399, 2, {0x0, 0x3DFC, 0x2504, 0x25FC, 0x2504, 0x3DFC, 0x2400, 0x2400, 0x25FC, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x4420, 0x5420, 0x8820}},
    {17400, 2, {0x0, 0x7BFC, 0x4840, 0x4840, 0x49F8, 0x7888, 0x4888, 0x4BFE, 0x4800, 0x7800, 0x49F8, 0x4908, 0x4908, 0x4908, 0x49F8, 0x9908}},
    {17401, 2, {0x0, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2504, 0x3D04, 0x25FC, 0x2450, 0x2490, 0x4492, 0x5512, 0x8A0E}},
    {17402, 2, {0x40, 0x3C40, 0x24FC, 0x2488, 0x2550, 0x3C20, 0x24D8, 0x2726, 0x24F8, 0x3C20, 0x24F8, 0x2420, 0x27FE, 0x4420, 0x5420, 0x8820}},
    {17403, 2, {0x1020, 0x1020, 0x2020, 0x4420, 0xFEFC, 0x220, 0x7C20, 0x4420, 0x45FE, 0x7C20, 0x4420, 0x7C48, 0x4484, 0x45FE, 0x5482, 0x4800}},
    {17404, 2, {0x48, 0x3C44, 0x245E, 0x25E0, 0x2428, 0x3C12, 0x246A, 0x2596, 0x2448, 0x3C5E, 0x25E0, 0x2424, 0x2428, 0x4412, 0x546A, 0x8986}},
    {17405, 2, {0x40, 0x3C20, 0x25FC, 0x2400, 0x2508, 0x3C90, 0x2400, 0x27FE, 0x2400, 0x3C00, 0x25FC, 0x2504, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {17406, 2, {0x20, 0x3C20, 0x25FE, 0x2420, 0x24FC, 0x3C40, 0x25FE, 0x2488, 0x2524, 0x3E22, 0x24F8, 0x2420, 0x25FE, 0x4420, 0x5420, 0x8820}},
    {17407, 2, {0x1040, 0x840, 0x3E7E, 0x2288, 0x2350, 0x3E20, 0x2058, 0x4186, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {17408, 2, {0x20, 0x3C20, 0x25FC, 0x2450, 0x2488, 0x3D04, 0x27FE, 0x2408, 0x25E8, 0x3D28, 0x2528, 0x25E8, 0x2528, 0x4408, 0x5428, 0x8810}},
    {17409, 2, {0x0, 0x7BFE, 0x4840, 0x4880, 0x4944, 0x7A24, 0x4868, 0x4AB0, 0x4930, 0x7AA8, 0x4868, 0x48A4, 0x4922, 0x4A20, 0x48A0, 0x9840}},
    {17410, 2, {0x0, 0x79F8, 0x4808, 0x4808, 0x49F8, 0x7808, 0x4808, 0x4BFE, 0x4820, 0x7A22, 0x4974, 0x48A8, 0x4924, 0x4A22, 0x48A0, 0x9840}},
    {17411, 2, {0x0, 0x7BFE, 0x4A02, 0x4A1A, 0x4AE2, 0x7A22, 0x4A22, 0x4BFE, 0x4A22, 0x7A72, 0x4AAA, 0x4B26, 0x4A22, 0x4A22, 0x4BFE, 0x9A02}},
    {17412, 2, {0x40, 0x3C40, 0x247C, 0x2484, 0x2508, 0x3E00, 0x2420, 0x25CE, 0x2502, 0x3D02, 0x25CE, 0x2502, 0x2502, 0x4502, 0x55FE, 0x8902}},
    {17413, 2, {0x0, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x3D00, 0x2540, 0x257C, 0x2590, 0x3D10, 0x25FE, 0x2510, 0x2528, 0x4528, 0x5544, 0x8A82}},
    {17414, 2, {0x8, 0x783C, 0x4BC0, 0x4844, 0x4A24, 0x7928, 0x4900, 0x4820, 0x4BFE, 0x7870, 0x48A8, 0x48A8, 0x4924, 0x4A22, 0x4C20, 0x9820}},
    {17415, 2, {0x80, 0x3C80, 0x24FC, 0x2554, 0x2654, 0x3C94, 0x2524, 0x2454, 0x2488, 0x3C00, 0x25FC, 0x2504, 0x25FC, 0x4504, 0x55FC, 0x8904}},
    {17416, 2, {0x40, 0x7840, 0x4BFC, 0x4840, 0x4840, 0x79F8, 0x4908, 0x49F8, 0x4908, 0x79F8, 0x4908, 0x49F8, 0x4908, 0x4908, 0x4FFE, 0x9800}},
    {17417, 2, {0x92, 0x3C92, 0x2524, 0x2648, 0x2524, 0x3C92, 0x2492, 0x2400, 0x25FE, 0x3D22, 0x2522, 0x25FE, 0x2522, 0x4522, 0x55FE, 0x8902}},
    {17418, 2, {0x0, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2400, 0x3D12, 0x25D4, 0x2518, 0x2510, 0x4552, 0x5592, 0x890E}},
    {17419, 2, {0x90, 0x7890, 0x4908, 0x4948, 0x4A44, 0x7C92, 0x4908, 0x4BFC, 0x4844, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x4A8A, 0x4C78, 0x9800}},
    {17420, 2, {0x440, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0xFFFE, 0x1010, 0x2008, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {17421, 2, {0x20, 0x3C20, 0x27FE, 0x2420, 0x2420, 0x3DFC, 0x2400, 0x2400, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2400, 0x4488, 0x5450, 0x8BFE}},
    {17422, 2, {0x0, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x7904, 0x4904, 0x49FC, 0x4800, 0x7BFE, 0x4820, 0x4920, 0x493C, 0x4920, 0x4AA0, 0x9C7E}},
    {17423, 2, {0x20, 0x3C20, 0x27FE, 0x2420, 0x25FC, 0x3C40, 0x27FE, 0x2488, 0x2504, 0x3EFA, 0x2488, 0x2488, 0x24F8, 0x4488, 0x5488, 0x88F8}},
    {17424, 2, {0x88, 0x71C8, 0x5708, 0x5108, 0x512A, 0x712A, 0x57AC, 0x5148, 0x5308, 0x7388, 0x5554, 0x5514, 0x5914, 0x5124, 0x5124, 0xB142}},
    {17425, 2, {0x48, 0x3D48, 0x2548, 0x27FE, 0x2548, 0x3D48, 0x2578, 0x2500, 0x25FE, 0x3C20, 0x27FE, 0x2470, 0x24A8, 0x4524, 0x5622, 0x8820}},
    {17426, 2, {0x0, 0x3C40, 0x25FE, 0x2488, 0x2504, 0x3E42, 0x24F8, 0x2708, 0x2490, 0x3C70, 0x27BE, 0x2442, 0x25A4, 0x4418, 0x5460, 0x8B80}},
    {17427, 2, {0x0, 0x3DFE, 0x2410, 0x2420, 0x24FC, 0x3C84, 0x2484, 0x24FC, 0x2484, 0x3CFC, 0x2484, 0x2484, 0x24FC, 0x4448, 0x5484, 0x8902}},
    {17428, 2, {0x20, 0x3C10, 0x25FE, 0x2502, 0x2502, 0x3DFE, 0x2500, 0x2500, 0x25FE, 0x3DAA, 0x26AA, 0x24FE, 0x24AA, 0x44AA, 0x54AA, 0x8886}},
    {17429, 2, {0x0, 0x3CF8, 0x2488, 0x2488, 0x24F8, 0x3C00, 0x25FE, 0x2488, 0x24F8, 0x3C88, 0x24F8, 0x2488, 0x249E, 0x47E8, 0x5408, 0x8808}},
    {17430, 2, {0x0, 0x7BFC, 0x4A04, 0x4A04, 0x4BFC, 0x7A20, 0x4B24, 0x4AA8, 0x4A20, 0x7BFE, 0x4A70, 0x4AA8, 0x4AA8, 0x4D24, 0x4A22, 0x9820}},
    {17431, 2, {0x20, 0x3C20, 0x24FA, 0x2424, 0x2428, 0x3DFE, 0x2420, 0x2440, 0x24FC, 0x3D44, 0x2644, 0x247C, 0x2444, 0x4444, 0x547C, 0x8844}},
    {17432, 2, {0x1E, 0x7BE0, 0x4844, 0x4924, 0x4888, 0x79FC, 0x4840, 0x4840, 0x4BFE, 0x7880, 0x48FC, 0x4944, 0x4928, 0x4A10, 0x4868, 0x9986}},
    {17433, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x48F8, 0x7820, 0x49FC, 0x4924, 0x4924, 0x79FC, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x9820}},
    {17434, 2, {0x0, 0x7BDE, 0x4A52, 0x4A52, 0x4A52, 0x7BD2, 0x4A52, 0x4A52, 0x4A52, 0x7BD2, 0x4A1A, 0x4A94, 0x4A50, 0x4AB0, 0x4B10, 0x9810}},
    {17435, 2, {0x20, 0x3C20, 0x25FE, 0x2450, 0x2488, 0x3D04, 0x27FE, 0x2488, 0x24F8, 0x3C88, 0x24F8, 0x2488, 0x249E, 0x47E8, 0x5408, 0x8808}},
    {17436, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x4BFE, 0x4888, 0x4904, 0x9A02}},
    {17437, 2, {0x20, 0x3D24, 0x24A8, 0x2420, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2504, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x4400, 0x5488, 0x8904}},
    {17438, 2, {0x0, 0x7BDE, 0x4842, 0x4842, 0x4BDE, 0x7A10, 0x4A10, 0x4BDE, 0x4842, 0x7A52, 0x494A, 0x494A, 0x4A52, 0x4842, 0x4A94, 0x9908}},
    {17439, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x49FC, 0x7888, 0x4BFE, 0x4820, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x4BFE, 0x4904, 0x4914, 0x9908}},
    {17440, 2, {0x88, 0x3C88, 0x27FE, 0x24A8, 0x2410, 0x3FFE, 0x2480, 0x2480, 0x24FC, 0x3C00, 0x24A8, 0x24A8, 0x24A8, 0x44AA, 0x552A, 0x8A06}},
    {17441, 2, {0x200, 0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x100, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2288, 0x2448, 0x2018}},
    {17442, 2, {0x80, 0x7880, 0x49FC, 0x4A44, 0x4954, 0x79F4, 0x4884, 0x4928, 0x4A90, 0x79FC, 0x4A44, 0x4D54, 0x49F4, 0x4884, 0x4928, 0x9A10}},
    {17443, 2, {0x0, 0x3CF8, 0x2488, 0x2488, 0x24F8, 0x3C00, 0x25FC, 0x2504, 0x25FC, 0x3D04, 0x25FC, 0x2504, 0x25FC, 0x4450, 0x5488, 0x8904}},
    {17444, 2, {0x84, 0x3C44, 0x2448, 0x25FE, 0x2420, 0x3CFC, 0x2420, 0x25FE, 0x2440, 0x3C80, 0x24FE, 0x2510, 0x2610, 0x4410, 0x55FE, 0x8800}},
    {17445, 2, {0x90, 0x7888, 0x4904, 0x4A42, 0x4C88, 0x79FC, 0x4804, 0x4800, 0x4BDE, 0x7842, 0x4A52, 0x494A, 0x4A52, 0x4842, 0x494A, 0x9884}},
    {17446, 2, {0x20, 0x3C20, 0x24FC, 0x2420, 0x2420, 0x3DFE, 0x2440, 0x2484, 0x25FE, 0x3C02, 0x25FC, 0x2554, 0x2554, 0x4554, 0x57FE, 0x8800}},
    {17447, 2, {0x40, 0x7820, 0x4BFC, 0x4800, 0x49F8, 0x7908, 0x49F8, 0x4800, 0x4BFC, 0x7A04, 0x4AF4, 0x4A94, 0x4AF4, 0x4A04, 0x4A14, 0x9A08}},
    {17448, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0x7C44, 0x4428, 0x7C28, 0x4410, 0x7C28, 0x4444, 0x4D82}},
    {17449, 2, {0x20, 0x3D24, 0x2524, 0x2524, 0x25FC, 0x3C00, 0x27FE, 0x2400, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2488, 0x4450, 0x57FE, 0x8800}},
    {17450, 2, {0xA0, 0x7890, 0x49FE, 0x4B20, 0x4DFC, 0x7920, 0x49FC, 0x4920, 0x49FE, 0x7900, 0x4BFC, 0x4888, 0x489E, 0x4902, 0x4A0A, 0x9C04}},
    {17451, 2, {0x124, 0x7924, 0x4A24, 0x4C24, 0x4954, 0x794A, 0x4A92, 0x4E10, 0x4A10, 0x7A50, 0x4A5C, 0x4A50, 0x4A50, 0x4AB0, 0x4A9E, 0x9B00}},
    {17452, 2, {0x90, 0x7890, 0x4FFE, 0x4890, 0x4BFC, 0x7A94, 0x4BFC, 0x4A94, 0x4BFC, 0x7800, 0x49F8, 0x4908, 0x49F8, 0x4908, 0x49F8, 0x9908}},
    {17453, 2, {0x2A8, 0x72A8, 0x57FC, 0x52A8, 0x52AA, 0x74E6, 0x5800, 0x57FC, 0x5444, 0x7040, 0x53F8, 0x5248, 0x5248, 0x5248, 0x5258, 0xB040}},
    {17454, 2, {0x40, 0x7820, 0x4BFC, 0x4800, 0x4908, 0x7890, 0x4BFE, 0x4A20, 0x4AA0, 0x7AFC, 0x4B20, 0x4A20, 0x4AFC, 0x4A20, 0x4C20, 0x99FE}},
    {17455, 2, {0x0, 0x3DFC, 0x2524, 0x25FC, 0x2524, 0x3DFC, 0x2440, 0x2488, 0x25F0, 0x3C20, 0x2444, 0x27FE, 0x2422, 0x4524, 0x5622, 0x8860}},
    {17456, 2, {0x120, 0x793C, 0x4944, 0x4AA8, 0x4A10, 0x7EA8, 0x4AC6, 0x4ABC, 0x4AA4, 0x7ABC, 0x4AA4, 0x4ABC, 0x4A24, 0x4A24, 0x4A24, 0x9A2C}},
    {17457, 2, {0x40, 0x7820, 0x4BFE, 0x4888, 0x4850, 0x7BFE, 0x4A22, 0x4AFA, 0x4A22, 0x7AFA, 0x4A8A, 0x4A8A, 0x4AFA, 0x4A02, 0x4A0A, 0x9A04}},
    {17458, 2, {0x0, 0x7BDE, 0x4842, 0x4A52, 0x494A, 0x7A52, 0x4842, 0x4820, 0x49FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x9904}},
    {17459, 2, {0x40, 0x3C40, 0x24F8, 0x2508, 0x2610, 0x3DFC, 0x2524, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x25FC, 0x2400, 0x4554, 0x552A, 0x8A2A}},
    {17460, 2, {0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {17461, 2, {0x88, 0x3C88, 0x25FC, 0x2488, 0x2488, 0x3FFE, 0x2420, 0x25FC, 0x2524, 0x3DFC, 0x2524, 0x25FC, 0x2400, 0x4488, 0x5504, 0x8A02}},
    {17462, 2, {0x0, 0x7BDE, 0x4888, 0x4AA8, 0x4BFE, 0x7988, 0x4ADA, 0x4CA6, 0x49F8, 0x7908, 0x4908, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x9908}},
    {17463, 2, {0x0, 0x7BDE, 0x4A52, 0x4A52, 0x4BDE, 0x7800, 0x49FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x9820}},
    {17464, 2, {0x20, 0x3DFC, 0x2424, 0x27FE, 0x2424, 0x3DFC, 0x2420, 0x27FE, 0x2400, 0x3DFC, 0x2524, 0x25FC, 0x2524, 0x45FC, 0x5400, 0x8BFE}},
    {17465, 2, {0x20, 0x79FC, 0x4824, 0x4BFE, 0x4824, 0x79FC, 0x4820, 0x4AAA, 0x4BAE, 0x7A22, 0x4BFE, 0x4A22, 0x4BAE, 0x4AAA, 0x4AAA, 0x9C22}},
    {17466, 2, {0x0, 0x7BFE, 0x4850, 0x4BFE, 0x4A52, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4820, 0x4BFE, 0x4820, 0x9820}},
    {17467, 2, {0x20, 0x7BFE, 0x4820, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x49FC, 0x4888, 0x7850, 0x4BFE, 0x4820, 0x49FC, 0x4820, 0x4820, 0x9820}},
    {17468, 2, {0x80, 0x7BDE, 0x488A, 0x49CA, 0x488A, 0x7BD2, 0x48A6, 0x4840, 0x4888, 0x79F0, 0x4824, 0x4BFE, 0x4822, 0x4924, 0x4AA2, 0x9840}},
    {17469, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x4888, 0x78F8, 0x4820, 0x4BFE, 0x4A22, 0x7B32, 0x4AAA, 0x4B76, 0x4A22, 0x4A22, 0x4A2A, 0x9A24}},
    {17470, 2, {0x0, 0x3DFC, 0x2554, 0x25FC, 0x2420, 0x3DFC, 0x2420, 0x27FE, 0x2488, 0x3C50, 0x25FC, 0x2420, 0x27FE, 0x4420, 0x5420, 0x8820}},
    {17471, 2, {0x40, 0x7880, 0x4BFE, 0x4A22, 0x4AAA, 0x7A22, 0x4BFE, 0x4A72, 0x4AAA, 0x7A02, 0x4820, 0x4BFE, 0x4820, 0x4850, 0x4888, 0x9B06}},
    {17472, 2, {0x10, 0x7BC8, 0x4A7E, 0x4A40, 0x4A62, 0x7BD4, 0x4A00, 0x4A3E, 0x4BC8, 0x7A48, 0x4E7E, 0x4A48, 0x4A48, 0x4BC8, 0x4A48, 0x9808}},
    {17473, 2, {0x3AB8, 0x2108, 0x3AB8, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {17474, 2, {0x140, 0x7A4C, 0x4A64, 0x4A44, 0x4B4C, 0x7A64, 0x4AA4, 0x4BAC, 0x4AA4, 0x7AA4, 0x4FFE, 0x4800, 0x4910, 0x4908, 0x4A04, 0x9C04}},
    {17475, 2, {0x88, 0x3DFE, 0x2488, 0x25FE, 0x3502, 0x2CFC, 0x2400, 0x25FE, 0x2C20, 0x3450, 0x25AA, 0x245C, 0x25AA, 0x4448, 0x55A8, 0x8810}},
    {17476, 2, {0x40, 0x3CF8, 0x2488, 0x24F8, 0x2488, 0x3CF8, 0x2400, 0x25FC, 0x2524, 0x3DFC, 0x2524, 0x25FC, 0x2400, 0x47FE, 0x5488, 0x8908}},
    {17477, 2, {0x88, 0x7BFE, 0x4888, 0x49FC, 0x4954, 0x7954, 0x49FC, 0x4800, 0x4BFE, 0x7A42, 0x48F8, 0x4888, 0x4948, 0x4830, 0x4860, 0x9B80}},
    {17478, 2, {0x108, 0x7890, 0x4BFC, 0x4840, 0x49F8, 0x7880, 0x4BFC, 0x4948, 0x49FC, 0x7B0A, 0x4DF8, 0x4908, 0x49FA, 0x4934, 0x498C, 0x9902}},
    {17479, 2, {0x42, 0x7B9C, 0x4A10, 0x4BDE, 0x4A94, 0x7A94, 0x4C20, 0x4BFC, 0x4A04, 0x7BFC, 0x4A04, 0x4BFC, 0x4A04, 0x4BFC, 0x4908, 0x9A04}},
    {17480, 2, {0x124, 0x7974, 0x4A58, 0x4ADA, 0x4BFE, 0x7954, 0x4ADA, 0x4BFE, 0x48A2, 0x7820, 0x4BFE, 0x4870, 0x48A8, 0x4924, 0x4A22, 0x9820}},
    {17481, 2, {0xE040, 0xBFFE, 0xA400, 0xA7FC, 0xA000, 0xE7FC, 0xA404, 0xA7FC, 0xA040, 0xEE4E, 0xABEA, 0xAEAA, 0xAAAE, 0xAE4A, 0x2AAA, 0x7710}},
    {17482, 2, {0x10, 0x787C, 0x4A54, 0x497C, 0x4910, 0x78FE, 0x4844, 0x4B7C, 0x4944, 0x797C, 0x4944, 0x497C, 0x4928, 0x4944, 0x4A80, 0x9C7E}},
    {17483, 2, {0x20, 0x7BFE, 0x4800, 0x49DC, 0x4954, 0x79DC, 0x4888, 0x4BFE, 0x4888, 0x7BFE, 0x4888, 0x4BFE, 0x4894, 0x4988, 0x4AA4, 0x98C2}},
    {17484, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A22, 0x4AFA, 0x4A22, 0x4AFA, 0x7AAA, 0x4AFA, 0x4AAA, 0x4AFA, 0x4A72, 0x4AAA, 0x9A26}},
    {17485, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A52, 0x4BDE, 0x498C, 0x4A52, 0x7840, 0x4FFE, 0x4888, 0x4990, 0x4860, 0x48D8, 0x9B04}},
    {17486, 2, {0x148, 0x794C, 0x4AAA, 0x4808, 0x4FFE, 0x7948, 0x4B68, 0x494A, 0x4B6A, 0x794C, 0x4B6C, 0x4948, 0x496A, 0x4B9A, 0x4826, 0x9842}},
    {17487, 2, {0x88, 0x7BFE, 0x4888, 0x4BDE, 0x4A52, 0x7BDE, 0x48A0, 0x4890, 0x49FE, 0x7920, 0x4BFC, 0x4D20, 0x49FC, 0x4920, 0x49FE, 0x9900}},
    {17488, 2, {0x40, 0x7E40, 0x4840, 0x48FE, 0x4880, 0x7E80, 0x4310, 0x4210, 0x4258, 0x7E54, 0x4894, 0x4892, 0x4912, 0x4810, 0x7E50, 0x20}},
    {17489, 2, {0x28, 0x7EAA, 0x486C, 0x4828, 0x48FE, 0x7E44, 0x4228, 0x42FE, 0x4210, 0x7E7C, 0x4810, 0x48FE, 0x4810, 0x4828, 0x7E44, 0x82}},
    {17490, 2, {0x0, 0xFE00, 0x11FC, 0x2024, 0x2824, 0x44A4, 0xFEA4, 0x12A4, 0x1124, 0x1024, 0x7C24, 0x1044, 0x1044, 0x1E84, 0xF114, 0x4208}},
    {17491, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0x7FFE, 0x4002, 0xBFFC, 0x820, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC}},
    {17492, 2, {0x20, 0x4020, 0x9820, 0x8BFE, 0x8820, 0x8820, 0x89FC, 0xD924, 0x8924, 0x8924, 0x8924, 0x8934, 0xF928, 0x8820, 0x20, 0x20}},
    {17493, 2, {0x600, 0x38F8, 0x2008, 0x3EF8, 0x2008, 0x3FF8, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x210A, 0x410A, 0x4106, 0x8102}},
    {17494, 2, {0x20, 0x4020, 0x9BFE, 0x8820, 0x89FC, 0x8824, 0x8BFE, 0xD824, 0x89FC, 0x8820, 0x8920, 0x893E, 0xF920, 0x8AA0, 0x27E, 0x400}},
    {17495, 2, {0x0, 0x41F8, 0x9908, 0x89F8, 0x8908, 0x89F8, 0x8800, 0xDBFE, 0x8900, 0x89FC, 0x8A54, 0x8C94, 0xF924, 0x8A44, 0xA8, 0x110}},
    {17496, 2, {0x108, 0x4088, 0x9890, 0x8BFC, 0x8840, 0x89F8, 0x8840, 0xDBFE, 0x8880, 0x8900, 0x89FC, 0x8A20, 0xFC20, 0x8820, 0x3FE, 0x0}},
    {17497, 2, {0x800, 0x1CFC, 0x7084, 0x1084, 0x1084, 0xFDFE, 0x1084, 0x1084, 0x7C84, 0x45FE, 0x4484, 0x4484, 0x4484, 0x7C84, 0x4494, 0x88}},
    {17498, 2, {0x800, 0x1CFC, 0x7084, 0x1084, 0x1084, 0xFCFC, 0x1084, 0x1084, 0x7C84, 0x44FC, 0x4484, 0x4484, 0x4484, 0x7C84, 0x4514, 0x208}},
    {17499, 2, {0x804, 0x1C1E, 0x71F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x11FE, 0x7D10, 0x4510, 0x4510, 0x4508, 0x450A, 0x7D4A, 0x45A6, 0x112}},
    {17500, 2, {0x848, 0x1D48, 0x7148, 0x13FE, 0x1148, 0xFD48, 0x1178, 0x1100, 0x7DFE, 0x4420, 0x47FE, 0x4470, 0x44A8, 0x7D24, 0x4622, 0x20}},
    {17501, 2, {0x0, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x0, 0xFFFE, 0x9022, 0x1020, 0x3DF8, 0x4420, 0xA520, 0x1920, 0x11FC, 0x2020, 0x4020}},
    {17502, 2, {0x3FFC, 0x2490, 0x2960, 0x2610, 0x3888, 0x2290, 0x2560, 0x2210, 0x2C08, 0x3FFC, 0x1010, 0x3EFC, 0x4210, 0x1490, 0x18FE, 0xE010}},
    {17503, 2, {0xF10, 0xF020, 0x117C, 0x8944, 0x427C, 0x44, 0xFF7C, 0x8100, 0x42FE, 0x7710, 0x9210, 0x5A7C, 0x2F10, 0x2210, 0x42FE, 0x8200}},
    {17504, 2, {0x800, 0x11FC, 0x3E04, 0x2208, 0x3210, 0x2A20, 0x2A20, 0xFE20, 0x2220, 0x3220, 0x2A20, 0x2A20, 0x2220, 0x4220, 0x4AA0, 0x8440}},
    {17505, 2, {0x1000, 0x21F8, 0x7C88, 0x44C8, 0x64A8, 0x54A8, 0x5488, 0xFC50, 0x4450, 0x6450, 0x5420, 0x5420, 0x4450, 0x4488, 0x5504, 0x8A02}},
    {17506, 2, {0x1000, 0x2000, 0x7DFC, 0x4490, 0x6490, 0x5490, 0x5490, 0xFC90, 0x4490, 0x6490, 0x5490, 0x5492, 0x4492, 0x4492, 0x550E, 0x8A00}},
    {17507, 2, {0x804, 0x1004, 0x3F08, 0x2110, 0x2920, 0x2544, 0x2504, 0xFF08, 0x2110, 0x2922, 0x2542, 0x2504, 0x2108, 0x4110, 0x4520, 0x82C0}},
    {17508, 2, {0x820, 0x1020, 0x3E50, 0x2250, 0x3288, 0x2B44, 0x2A22, 0xFE20, 0x2200, 0x33FC, 0x2A04, 0x2A08, 0x2208, 0x4210, 0x4A10, 0x8420}},
    {17509, 2, {0x800, 0x10FC, 0x3E44, 0x2248, 0x3248, 0x2A50, 0x2A5C, 0xFE44, 0x2244, 0x3244, 0x2A68, 0x2AA8, 0x2290, 0x4298, 0x4B24, 0x8442}},
    {17510, 2, {0x840, 0x1040, 0x3E80, 0x22FC, 0x3304, 0x2A04, 0x2AF4, 0xFE94, 0x2294, 0x3294, 0x2AF4, 0x2A94, 0x2204, 0x4204, 0x4A28, 0x8410}},
    {17511, 2, {0x1088, 0x2088, 0x7888, 0x4908, 0x697E, 0x5B08, 0x4D08, 0xF948, 0x4928, 0x4928, 0x6908, 0x5908, 0x4908, 0x4908, 0x4928, 0x9910}},
    {17512, 2, {0x1080, 0x2080, 0x7CFE, 0x4500, 0x6620, 0x5520, 0x552C, 0xFD74, 0x47A4, 0x6524, 0x5534, 0x5528, 0x4522, 0x4502, 0x54FE, 0x8800}},
    {17513, 2, {0x810, 0x1010, 0x3E10, 0x2210, 0x32FE, 0x2A10, 0x2A10, 0xFE10, 0x227C, 0x3244, 0x2A44, 0x2A44, 0x2244, 0x4244, 0x4A7C, 0x8444}},
    {17514, 2, {0x820, 0x1020, 0x3E50, 0x2250, 0x3288, 0x2B04, 0x2AFA, 0xFE00, 0x2200, 0x32F8, 0x2A88, 0x2A88, 0x2288, 0x4288, 0x4AF8, 0x8488}},
    {17515, 2, {0x884, 0x1048, 0x3E00, 0x22FC, 0x3248, 0x2A48, 0x2A48, 0xFE48, 0x23FE, 0x3248, 0x2A48, 0x2A48, 0x2248, 0x4288, 0x4A88, 0x8508}},
    {17516, 2, {0x850, 0x1050, 0x3E50, 0x2352, 0x32D4, 0x2A58, 0x2A50, 0xFE58, 0x22D4, 0x3352, 0x2A50, 0x2A50, 0x2292, 0x4292, 0x4B12, 0x860E}},
    {17517, 2, {0x1000, 0x2100, 0x7CBE, 0x4488, 0x6408, 0x5608, 0x5508, 0xFD48, 0x4448, 0x6488, 0x5788, 0x5488, 0x4488, 0x44FE, 0x5480, 0x8800}},
    {17518, 2, {0x1000, 0x20F8, 0x7C88, 0x4488, 0x6488, 0x5488, 0x5526, 0xFE20, 0x45FE, 0x6420, 0x5470, 0x54A8, 0x4524, 0x4622, 0x5420, 0x8820}},
    {17519, 2, {0x1108, 0x2088, 0x7C90, 0x45FC, 0x6424, 0x5424, 0x55FC, 0xFD20, 0x4520, 0x65FE, 0x5462, 0x54A2, 0x452A, 0x4624, 0x5420, 0x8820}},
    {17520, 2, {0x810, 0x1010, 0x3E90, 0x229E, 0x3290, 0x2A90, 0x2BFE, 0xFE00, 0x2210, 0x3292, 0x2A92, 0x2B14, 0x2204, 0x4208, 0x4A30, 0x85C0}},
    {17521, 2, {0x1040, 0x2020, 0x7BFE, 0x4A02, 0x6D04, 0x5900, 0x49DE, 0xFA52, 0x4A52, 0x4B52, 0x6C9A, 0x5894, 0x4910, 0x4912, 0x4A12, 0x9C0E}},
    {17522, 2, {0x820, 0x1020, 0x3E3E, 0x2220, 0x32FC, 0x2A84, 0x2AFC, 0xFE84, 0x22FC, 0x32A4, 0x2A20, 0x2BFE, 0x2220, 0x4220, 0x4A20, 0x8420}},
    {17523, 2, {0x1020, 0x2020, 0x7C50, 0x4488, 0x6504, 0x56FA, 0x5400, 0xFDFC, 0x4554, 0x6554, 0x55FC, 0x5554, 0x4554, 0x4554, 0x5504, 0x890C}},
    {17524, 2, {0x844, 0x1044, 0x3EFE, 0x2244, 0x3244, 0x2A7C, 0x2A44, 0xFE44, 0x227C, 0x3244, 0x2A44, 0x2AFE, 0x2200, 0x4228, 0x4A44, 0x8482}},
    {17525, 2, {0x1000, 0x21FE, 0x7D02, 0x457A, 0x6502, 0x557A, 0x5400, 0xFCFC, 0x4484, 0x64FC, 0x5484, 0x54FC, 0x4484, 0x4484, 0x5494, 0x8888}},
    {17526, 2, {0x810, 0x1010, 0x3EFE, 0x2210, 0x327C, 0x2A10, 0x2AFE, 0xFE00, 0x227C, 0x3244, 0x2A7C, 0x2A44, 0x227C, 0x4244, 0x4A54, 0x8448}},
    {17527, 2, {0x1020, 0x2010, 0x7DFE, 0x4502, 0x6502, 0x55FE, 0x5500, 0xFD7E, 0x4542, 0x657E, 0x5542, 0x557E, 0x4542, 0x4542, 0x554A, 0x8A44}},
    {17528, 2, {0x820, 0x1010, 0x3EFE, 0x2282, 0x3200, 0x2A7C, 0x2A00, 0xFE00, 0x22FE, 0x3210, 0x2A10, 0x2A54, 0x2252, 0x4292, 0x4A50, 0x8420}},
    {17529, 2, {0x1040, 0x2020, 0x7BFE, 0x4A22, 0x68A4, 0x5928, 0x4850, 0xF888, 0x4904, 0x48F8, 0x6888, 0x5888, 0x4850, 0x4820, 0x48D8, 0x9B06}},
    {17530, 2, {0x1000, 0x23FC, 0x7A04, 0x4A34, 0x6BC4, 0x5A44, 0x4BF4, 0xFB54, 0x4B54, 0x4BF4, 0x6A44, 0x5A54, 0x4BF4, 0x4A94, 0x4C06, 0x9802}},
    {17531, 2, {0x1108, 0x2088, 0x7888, 0x4BDE, 0x6910, 0x5920, 0x49DE, 0xF942, 0x4944, 0x4944, 0x695E, 0x5944, 0x4A44, 0x4A44, 0x4CD4, 0x9808}},
    {17532, 2, {0x1040, 0x2040, 0x7CFE, 0x4522, 0x64AA, 0x54FA, 0x5442, 0xFC94, 0x4548, 0x64FE, 0x5522, 0x56AA, 0x44FA, 0x4442, 0x5494, 0x8908}},
    {17533, 2, {0x1000, 0x21FC, 0x7904, 0x49FC, 0x6904, 0x59FC, 0x4800, 0xFBDE, 0x4842, 0x4A52, 0x694A, 0x58C6, 0x4B5A, 0x4842, 0x494A, 0x9884}},
    {17534, 2, {0x10A0, 0x2090, 0x79FE, 0x4B20, 0x69FC, 0x5920, 0x49FC, 0xF920, 0x49FE, 0x4900, 0x6BF8, 0x5908, 0x4890, 0x4860, 0x4998, 0x9E06}},
    {17535, 2, {0x1020, 0x2010, 0x7BFE, 0x4A02, 0x6880, 0x58FE, 0x4910, 0xFB20, 0x4D7C, 0x4944, 0x6944, 0x597C, 0x4944, 0x4944, 0x497C, 0x9944}},
    {17536, 2, {0x1110, 0x2110, 0x79DE, 0x4AA8, 0x6C44, 0x5888, 0x4888, 0xF908, 0x4B7E, 0x4D08, 0x6948, 0x5928, 0x4908, 0x4908, 0x4928, 0x9910}},
    {17537, 2, {0x1000, 0x23DE, 0x7842, 0x4A52, 0x694A, 0x5A52, 0x4842, 0xF820, 0x49FC, 0x4904, 0x6904, 0x59FC, 0x4904, 0x4904, 0x49FC, 0x9904}},
    {17538, 2, {0x1090, 0x2290, 0x7A9E, 0x4AA2, 0x6AD4, 0x5BA8, 0x4894, 0xF8A4, 0x4FFE, 0x4A84, 0x6AA4, 0x5A94, 0x4A94, 0x4A84, 0x4C94, 0x9888}},
    {17539, 2, {0x1010, 0x2050, 0x7A50, 0x497C, 0x6890, 0x5810, 0x48FE, 0xFB00, 0x497C, 0x4944, 0x6944, 0x597C, 0x4944, 0x4900, 0x4AFE, 0x9C00}},
    {17540, 2, {0x1000, 0x2000, 0x7BFE, 0x4A02, 0x69FC, 0x5800, 0x48F8, 0xF888, 0x48F8, 0x4800, 0x69FC, 0x5924, 0x49FC, 0x4924, 0x49FC, 0x9904}},
    {17541, 2, {0x1020, 0x2124, 0x7CA8, 0x45FC, 0x6440, 0x57FE, 0x5488, 0xFD04, 0x47FE, 0x6504, 0x55FC, 0x5504, 0x45FC, 0x4504, 0x55FC, 0x8904}},
    {17542, 2, {0x1108, 0x2088, 0x7BE8, 0x4910, 0x6A9E, 0x5FD4, 0x4864, 0xFBD4, 0x4A54, 0x4BD4, 0x6A54, 0x5BC8, 0x4A48, 0x4A54, 0x4A64, 0x9AC2}},
    {17543, 2, {0x1040, 0x2020, 0x7BFE, 0x4A02, 0x69FC, 0x5800, 0x48F8, 0xF888, 0x48F8, 0x4800, 0x69FC, 0x5924, 0x49FC, 0x4924, 0x49FC, 0x9904}},
    {17544, 2, {0x1020, 0x21FC, 0x7C20, 0x4488, 0x67FE, 0x5488, 0x55FC, 0xFD04, 0x45FC, 0x6504, 0x55FC, 0x5504, 0x45FC, 0x4488, 0x5504, 0x8A02}},
    {17545, 2, {0x1080, 0x2100, 0x7BFE, 0x4D54, 0x6954, 0x5BFE, 0x4954, 0xF954, 0x4BFE, 0x4904, 0x69DE, 0x5A44, 0x4954, 0x489E, 0x4904, 0x9A04}},
    {17546, 2, {0x440, 0x2440, 0x22FC, 0x2288, 0x4310, 0x40FC, 0xBEA4, 0x10A4, 0x20A4, 0x3EFC, 0x280, 0x280, 0x282, 0x282, 0x147E, 0x800}},
    {17547, 2, {0x1040, 0x1040, 0x10FC, 0x7C88, 0x5510, 0x54FC, 0x54A4, 0x54A4, 0x54A4, 0xFEFC, 0x1080, 0x1080, 0x2882, 0x2482, 0x447E, 0x8000}},
    {17548, 2, {0x20, 0xFF20, 0x817C, 0x44, 0x7E88, 0x427C, 0x7E54, 0x4254, 0x7E54, 0x107C, 0x840, 0xFF40, 0x42, 0x2442, 0x423E, 0x8100}},
    {17549, 2, {0x2040, 0x1040, 0xFEFC, 0x88, 0x4510, 0x28FC, 0xFEA4, 0xA2A4, 0x10A4, 0xFEFC, 0x2080, 0x3C80, 0x2482, 0x4482, 0x547E, 0x8800}},
    {17550, 2, {0x2420, 0x2420, 0xFF7C, 0x2444, 0x1288, 0xFE7C, 0x2454, 0x4254, 0x8154, 0x247C, 0x2440, 0xFF40, 0x2442, 0x2442, 0x443E, 0x8400}},
    {17551, 2, {0x4440, 0x2440, 0x28FC, 0xFE88, 0x9310, 0xD6FC, 0xBAA4, 0x92A4, 0xFEA4, 0xFC, 0x7C80, 0x4480, 0x7C82, 0x4482, 0x7C7E, 0x4400}},
    {17552, 2, {0x2220, 0xFFA0, 0x227C, 0x7F44, 0x5588, 0x7F7C, 0x54, 0xFFD4, 0x80D4, 0x3E7C, 0x2240, 0x3E40, 0x2242, 0x3E42, 0x223E, 0x3E00}},
    {17553, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x810, 0x820, 0x0, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {17554, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x8, 0x48, 0x188, 0x608, 0x1808, 0x6008, 0x8, 0x50, 0x20}},
    {17555, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0xFFFE, 0x400, 0x400, 0x800, 0x1FF0, 0x10, 0x10, 0x10, 0xA0, 0x40}},
    {17556, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x40, 0x440, 0x440, 0x420, 0x820, 0x820, 0x1010, 0x1010, 0x2008, 0x4004, 0x8002}},
    {17557, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x100, 0x100, 0x1110, 0x1108, 0x2104, 0x2104, 0x4102, 0x8102, 0x500, 0x200}},
    {17558, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x7FF0, 0x410, 0x410, 0x410, 0xFFD0, 0x412, 0x40A, 0x406, 0x402, 0x400}},
    {17559, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x800, 0x1FFC, 0x2000, 0x4000, 0x9FE0, 0x20, 0xC0, 0x300, 0xC04, 0x1004, 0xFFC}},
    {17560, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x400, 0x400, 0xFF0, 0x1010, 0x2820, 0x4440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {17561, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x100, 0x100, 0x100, 0x1F8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {17562, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x100, 0x1110, 0x1108, 0x2104, 0x2114, 0x4120, 0x40, 0x180, 0xE00, 0xF000}},
    {17563, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x10FC, 0x2204, 0x4204, 0xFC04, 0x804, 0x1004, 0x2204, 0x4104, 0xFF04, 0x4128, 0x10}},
    {17564, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {17565, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x8, 0x1810, 0x660, 0x180, 0x660, 0x1818, 0x6C24, 0x240, 0x180, 0x660, 0x1818, 0xE006}},
    {17566, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FFC, 0x2104, 0x4108, 0x280, 0x280, 0x480, 0x882, 0x1082, 0x207E, 0x4000}},
    {17567, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1210, 0x1110, 0x1010, 0xFFFE, 0x1010, 0x2010, 0x2010, 0x4050, 0x8020}},
    {17568, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {17569, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7F08, 0x108, 0x108, 0x3F08, 0x2008, 0x4008, 0x7F08, 0x108, 0x108, 0xA08, 0x408}},
    {17570, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0xC00, 0x70FC, 0x4084, 0x4084, 0x4084, 0x4C84, 0x7094, 0x4088, 0x80, 0x80}},
    {17571, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1010, 0x7FFC, 0x1010, 0x1010, 0x7FFC, 0x1010, 0x1010, 0x1050, 0x1020}},
    {17572, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {17573, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {17574, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1F0, 0x3E00, 0x200, 0x100, 0x7FF8, 0x60, 0x180, 0xE00, 0x3000, 0x4800, 0x87FE, 0x0}},
    {17575, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {17576, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x800, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {17577, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0xF0, 0x1F00, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100, 0x100}},
    {17578, 2, {0x820, 0xFFFE, 0x820, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {17579, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x850, 0x848, 0x1040, 0x307C, 0x57C0, 0x9040, 0x1020, 0x1020, 0x1014, 0x100C, 0x1004}},
    {17580, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x100, 0x1110, 0x2108, 0x4104, 0x8502, 0x200}},
    {17581, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x0, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {17582, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8004, 0xFE0, 0x820, 0xA20, 0x920, 0x922, 0x1022, 0x201E, 0x4000}},
    {17583, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4202, 0x8204, 0x1FE0, 0x220, 0xC20, 0x420, 0xA20, 0x1122, 0x2022, 0x401E}},
    {17584, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FFE, 0x2000, 0x2FFC, 0x2080, 0x2080, 0x27F8, 0x2080, 0x2080, 0x2FFC, 0x2000, 0x3FFE}},
    {17585, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x3FF0, 0x20, 0xC0, 0x7D04, 0x588, 0x950, 0x1120, 0x2118, 0xC506, 0x200, 0xFFFE, 0x0}},
    {17586, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1008, 0x101C, 0xFEE0, 0x1220, 0x2220, 0x2220, 0x643E, 0x15E0, 0x822, 0x1422, 0x2222, 0xC01E}},
    {17587, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0xFE0, 0x820, 0x1020, 0x611C, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {17588, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x400, 0x7FFC, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0x850, 0x820}},
    {17589, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x11F8, 0xFC08, 0x2410, 0x2420, 0x25FE, 0x4420, 0x2420, 0x1820, 0x1820, 0x24A0, 0xC240}},
    {17590, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0xB0, 0x1C2, 0xE32, 0x700E}},
    {17591, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xF0, 0x1F00, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {17592, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x200, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {17593, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {17594, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x400, 0x820, 0x1FC0, 0x80, 0x320, 0xC10, 0x3FF8, 0x108, 0x1110, 0x2108, 0x4504, 0x200}},
    {17595, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x87C, 0x804, 0x4904, 0x4904, 0x497C, 0x4940, 0x4940, 0x4F42, 0x7942, 0x3E, 0x0}},
    {17596, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2008}},
    {17597, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2008, 0x2AA8, 0x2AA8, 0x2448, 0x2448, 0x2AA8, 0x2AA8, 0x2008, 0x2028, 0x2010}},
    {17598, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0xFFFE, 0x200, 0x500, 0x904, 0x1888, 0x2850, 0xC820, 0xA18, 0xC06, 0x800}},
    {17599, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {17600, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {17601, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x20, 0x7FFC, 0x820, 0x420, 0x4A0, 0x40}},
    {17602, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {17603, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {17604, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4400, 0x280C, 0x11F0, 0x2900, 0xC900, 0x19FE, 0x2910, 0x4910, 0x8910, 0xA10, 0x5210, 0x2410}},
    {17605, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x200C, 0x11F0, 0x1100, 0x8100, 0x49FE, 0x4910, 0x1110, 0xE110, 0x2210, 0x2210, 0x2410}},
    {17606, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x100C, 0x11F0, 0xFD00, 0x1100, 0x11FE, 0x1D10, 0xF110, 0x1110, 0x1210, 0x5210, 0x2410}},
    {17607, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x440, 0x2448, 0x4444, 0x8842, 0x840, 0x1140, 0x2080}},
    {17608, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x10F8, 0x2088, 0x4108, 0x8A06, 0x1000, 0x31F8, 0x5088, 0x9050, 0x1020, 0x10D8, 0x1706}},
    {17609, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7E7C, 0x2444, 0x2448, 0x4450, 0x7F48, 0x1444, 0x2444, 0x4454, 0x8448, 0x1440, 0x840}},
    {17610, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x200, 0x200, 0x3FF0, 0x410, 0x810, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0}},
    {17611, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2008, 0x103C, 0x13E0, 0x8220, 0x4A20, 0x4BFE, 0x1220, 0xE210, 0x2212, 0x228A, 0x2306, 0x2202}},
    {17612, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0}},
    {17613, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {17614, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1080, 0x1080, 0x14FC, 0x5504, 0x5A04, 0x5084, 0x9044, 0x1044, 0x2804, 0x2404, 0x4428, 0x8010}},
    {17615, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x300, 0xD60, 0x3118, 0xC104, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {17616, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x440, 0x820, 0x3118, 0xC106, 0x3FF8, 0x100, 0x1110, 0x1114, 0x1FF4, 0x104, 0xFC}},
    {17617, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2000, 0x17FC, 0x9110, 0x4110, 0x4110, 0x17FE, 0x2110, 0xE110, 0x2210, 0x2210, 0x2410, 0x810}},
    {17618, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1FF8, 0x2008, 0x5FC8, 0x9248, 0x1248, 0x1FC8, 0x1248, 0x1248, 0x1FC8, 0x28, 0x10}},
    {17619, 2, {0x820, 0xFFFE, 0xA20, 0x200, 0x7FFC, 0x200, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x1FFC, 0x2020, 0x4420, 0x8220, 0x2A0, 0x40}},
    {17620, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x108, 0x1FD0, 0x120, 0xFFFE, 0x100, 0x1FF0, 0x820, 0x30C0, 0xCFFC, 0x80, 0x280, 0x100}},
    {17621, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23F0, 0x1000, 0x1000, 0x7FC, 0x7120, 0x1120, 0x1224, 0x1224, 0x141C, 0x2800, 0x47FE, 0x0}},
    {17622, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1210, 0x2208, 0x1FF0, 0x410, 0x410, 0x810, 0x10A0, 0x2040}},
    {17623, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x820, 0x1110, 0x2208, 0xC446, 0x820, 0x1FF0, 0x110, 0x884, 0x4892, 0x4812, 0x87F0}},
    {17624, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x840, 0x1F60, 0x2250, 0x5448, 0x840, 0x37FE, 0xC000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {17625, 2, {0x820, 0xFFFE, 0x820, 0x70, 0x1F80, 0x100, 0xFFFE, 0x100, 0x920, 0x7924, 0x928, 0x932, 0x1922, 0xE91E, 0x100, 0x100}},
    {17626, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x21F8, 0x4108, 0x8908, 0x11F8, 0x3108, 0x5108, 0x91F8, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {17627, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2080, 0x2490, 0x2490, 0x27F0, 0x2080, 0x4888, 0x4888, 0x8FF8}},
    {17628, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2008, 0x103C, 0x13E0, 0x8220, 0x4A20, 0x4BFE, 0x1220, 0xE210, 0x2212, 0x228A, 0x2326, 0x2212}},
    {17629, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x1FE0, 0x40, 0x180, 0x7FFC, 0x100, 0x500, 0x200, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {17630, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x20A0, 0x2090, 0x2FFC, 0x2120, 0x4210, 0x4408, 0x9806}},
    {17631, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x27C8, 0x2108, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x47C8, 0x4028, 0x8010}},
    {17632, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x104, 0x7D98, 0x960, 0x3118, 0xC506, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {17633, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2820, 0x2000, 0x3E7C, 0x4844, 0x8844, 0x844, 0xFF44, 0x844, 0x1444, 0x227C, 0x4144, 0x8000}},
    {17634, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {17635, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x4104, 0x5144, 0x4924, 0x5554, 0x6594, 0x4104, 0x4114, 0x4108}},
    {17636, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x1FF8, 0x2248, 0x4448, 0x888, 0x1108, 0x2228, 0x410, 0x100, 0x4884, 0x4812, 0x87F2}},
    {17637, 2, {0x600, 0x100, 0x6C0, 0x1830, 0xE44E, 0x3FF8, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100}},
    {17638, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {17639, 2, {0x440, 0xFFFE, 0x440, 0x0, 0x3FF0, 0x20, 0xC0, 0x793C, 0x4904, 0x4968, 0x4910, 0x7928, 0x144, 0x500, 0x200, 0xFFFE}},
    {17640, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0x8804, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {17641, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x11FE, 0xFC20, 0x1020, 0x31FC, 0x3884, 0x5488, 0x5050, 0x9020, 0x10D8, 0x1306}},
    {17642, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FF8, 0x2248, 0x4488, 0x908, 0x1228, 0x2410}},
    {17643, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x3FF8, 0x440, 0x820, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1F90, 0x10, 0x50, 0x20}},
    {17644, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x1FC, 0x100, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {17645, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x104, 0x3204, 0xC24, 0x3224, 0xC924, 0x824, 0xFFA4, 0x824, 0x2A24, 0x4904, 0xA894, 0x1008}},
    {17646, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3E40, 0x240, 0x147E, 0x888, 0xFF48, 0xA48, 0x1850, 0x2850, 0x4820, 0x850, 0x2888, 0x1104}},
    {17647, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1210, 0xFFFE, 0x2090, 0x3FFC, 0x10, 0x60}},
    {17648, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {17649, 2, {0x820, 0xFFFE, 0x820, 0x440, 0x274C, 0x2470, 0x2442, 0x2F42, 0xF03E, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {17650, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x908, 0x890, 0x1000, 0x13FC, 0x3204, 0x5204, 0x93FC, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {17651, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x9F8, 0x1108, 0x21F8, 0x4508, 0x9F8, 0x1924, 0x2928, 0x4910, 0x948, 0x986, 0x900}},
    {17652, 2, {0x820, 0xFFFE, 0x820, 0x0, 0xFFFE, 0x200, 0x3FF8, 0x2488, 0x24A8, 0x2110, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {17653, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7EFC, 0x4284, 0x4284, 0x4284, 0x7EFC, 0x4284}},
    {17654, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x808, 0x808, 0xFF88, 0x848, 0x7F48, 0x4948, 0x7F48, 0x1C48, 0x2A48, 0x4908, 0x8828, 0x810}},
    {17655, 2, {0x820, 0x820, 0xFFFE, 0x8A0, 0x2100, 0x123C, 0x9480, 0x4100, 0x437E, 0x1508, 0x2108, 0xE108, 0x2108, 0x2108, 0x2128, 0x2110}},
    {17656, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x7EF8, 0x1220, 0x2420, 0x1820, 0x65FC, 0x820, 0x7E20, 0x1220, 0x2420, 0x1820, 0x6420}},
    {17657, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0xE60, 0x3118, 0xDFE6, 0x1020, 0x1FE0, 0x1020, 0x1FE0, 0x1008, 0x1190, 0x1460, 0x1810}},
    {17658, 2, {0x820, 0xFFFE, 0x820, 0x1010, 0x820, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {17659, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0xBE24, 0x20, 0x1FC, 0x7E48, 0x14C8, 0x1430, 0x146A, 0x2586, 0x2402, 0x43FE}},
    {17660, 2, {0x820, 0xFFFE, 0x820, 0x1088, 0x1084, 0x13FE, 0xFCA4, 0x24A8, 0x24B4, 0x24A8, 0x48B2, 0x28A4, 0x10AA, 0x2932, 0x4522, 0x82DE}},
    {17661, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2000, 0x10FC, 0xFE04, 0x4, 0x7C04, 0xFC, 0x7C80, 0x80, 0x7C80, 0x4482, 0x7C82, 0x447E}},
    {17662, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1248, 0x4444, 0x2014, 0x860, 0x7380, 0x2000, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {17663, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FE0, 0x440, 0x280, 0x7FFC, 0x4444, 0x4284, 0x5FF4, 0x4104, 0x7FFC, 0x4104, 0x4104, 0x7FFC}},
    {17664, 2, {0x440, 0xFFFE, 0x440, 0x1038, 0x13C0, 0x2040, 0x27FC, 0x6150, 0xA248, 0x2446, 0x2080, 0x2FFE, 0x2110, 0x2320, 0x20E0, 0x2718}},
    {17665, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {17666, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4202, 0x8104, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x100, 0x3FF8, 0x100, 0x100, 0x100}},
    {17667, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2000, 0x11F8, 0xFD08, 0x908, 0x11F8, 0x3908, 0x5508, 0x95F8, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {17668, 2, {0x820, 0xFFFE, 0x820, 0xFC, 0x7F00, 0x2208, 0x1110, 0x2020, 0x3FF8, 0x4100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x3FF8}},
    {17669, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x20, 0x1F20, 0x1120, 0x1F20, 0x0, 0xFFFC, 0x20, 0x1F20, 0x1120, 0x1F20, 0x60}},
    {17670, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4400, 0x29FC, 0x1124, 0x29FC, 0x4924, 0x99FC, 0x2820, 0x49FC, 0x8820, 0x820, 0x53FE, 0x2000}},
    {17671, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE}},
    {17672, 2, {0x440, 0xFFFE, 0x440, 0x3FF8, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {17673, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xC1C, 0x71E0, 0x1020, 0x1020, 0xFDFE, 0x3070, 0x38A8, 0x54A8, 0x5124, 0x9222, 0x1020, 0x1020}},
    {17674, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x108, 0xFC90, 0x11FC, 0x1090, 0x1090, 0x7C90, 0x13FE, 0x1090, 0x1C90, 0xF110, 0x4110, 0x210}},
    {17675, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x2420, 0x7CA8, 0x8A4, 0x1122, 0x2522, 0x7E28, 0x10, 0x5420, 0x54C0, 0x8300}},
    {17676, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x820, 0x3118, 0xCFE6, 0x100, 0x100}},
    {17677, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF8, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x4, 0x2924, 0x4494, 0x8}},
    {17678, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x17FC, 0x1040, 0x3F8, 0x7248, 0x1248, 0x13F8, 0x1150, 0x1248, 0x1444, 0x2840, 0x47FE}},
    {17679, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x2240, 0x13F8, 0x1440, 0x40, 0x77FC, 0x1000, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE, 0x0}},
    {17680, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2120, 0x1110, 0x17FC, 0x8090, 0x4064, 0x539C, 0x1120, 0x2110, 0xE7FC, 0x2090, 0x2064, 0x239C}},
    {17681, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x808, 0x1448, 0x2228, 0x5D08, 0x88C8, 0x828, 0x7F0E, 0x8F8, 0x2A08, 0x4908, 0xA888, 0x1008}},
    {17682, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x80, 0x7840, 0x4BFC, 0x5108, 0x6090, 0x57FE, 0x4800, 0x49F8, 0x6908, 0x5108, 0x41F8, 0x4108}},
    {17683, 2, {0x820, 0xFFFE, 0x820, 0x2000, 0x1040, 0xFE40, 0xFE, 0x7C90, 0x110, 0x7C10, 0x1FE, 0x7C10, 0x4410, 0x4410, 0x7C10, 0x4410}},
    {17684, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {17685, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x28, 0x7824, 0x4BFE, 0x4820, 0x79FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x4924, 0x990C}},
    {17686, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7840, 0x4BF8, 0x5040, 0x67FC, 0x5110, 0x4A88, 0x4CF4, 0x6910, 0x52A0, 0x4040, 0x41B0, 0x4E0E}},
    {17687, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2088, 0x1050, 0xFBFE, 0x820, 0x1020, 0x39FC, 0x5420, 0x9420, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {17688, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2000, 0x20F8, 0x3C88, 0x5088, 0x10F8, 0xFC88, 0x1088, 0x54F8, 0x5488, 0x5C88, 0x75FE, 0x400}},
    {17689, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1080, 0x2880, 0x44FC, 0xB904, 0x1204, 0x7C84, 0x1044, 0x9444, 0x5004, 0x1E04, 0xE028, 0x4010}},
    {17690, 2, {0x820, 0xFFFE, 0x820, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x2630, 0x240E, 0x2FF8, 0x4808, 0x4FF8, 0x8808, 0x818}},
    {17691, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23FC, 0x2040, 0x21F8, 0x2108, 0x3DF8, 0x2108, 0x21F8, 0x2508, 0x29F8, 0x3000, 0x2090, 0x108}},
    {17692, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0xFC20, 0x49FC, 0x7924, 0x4924, 0x79FC, 0x4924, 0x4D24, 0xFBFE, 0x904, 0x914, 0x908}},
    {17693, 2, {0x440, 0xFFFE, 0x440, 0x844, 0x2F78, 0x2842, 0x2F3E, 0xF400, 0x820, 0x1FC0, 0x310, 0xC08, 0x3FFC, 0x104, 0x1510, 0x2208}},
    {17694, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x200, 0x4444, 0x2F88, 0x1110, 0x2248, 0x4FE4, 0x20, 0x100, 0xFFFE, 0x100, 0x100}},
    {17695, 2, {0x440, 0xFFFE, 0x440, 0x20, 0xC40, 0x70FC, 0x1308, 0x1090, 0xFC60, 0x3090, 0x3B3E, 0x5442, 0x55A4, 0x9018, 0x1060, 0x1380}},
    {17696, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1124, 0xFCA8, 0x1020, 0x31FC, 0x3904, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {17697, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x1FE0, 0x2020, 0x7FF8, 0xA108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x2488, 0x2244, 0x4244}},
    {17698, 2, {0x820, 0xFFFE, 0x820, 0x1450, 0x2288, 0x820, 0x1450, 0x2288, 0x1FF0, 0x0, 0xFFFE, 0x1000, 0x1FF8, 0x8, 0x50, 0x20}},
    {17699, 2, {0x820, 0xFFFE, 0x820, 0x304, 0x3C24, 0x424, 0x7FA4, 0x1624, 0x2504, 0x440C, 0x1100, 0x1FF8, 0x2100, 0xFFFE, 0x100, 0x100}},
    {17700, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x13FE, 0x2020, 0x7C40, 0x45FC, 0x4504, 0x4504, 0x7DFC, 0x4504, 0x4504, 0x7DFC, 0x4504}},
    {17701, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF0, 0x110, 0xFFFC, 0x110, 0x3FF0, 0x920, 0x2548, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x4108}},
    {17702, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1100, 0x11FC, 0x1200, 0xFDF8, 0x3148, 0x3928, 0x57FE, 0x5148, 0x9228, 0x13FC, 0x1050, 0x1020}},
    {17703, 2, {0x820, 0xFFFE, 0x820, 0x304, 0x3C24, 0x424, 0x7FA4, 0x1624, 0x2504, 0x440C, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {17704, 2, {0x820, 0xFFFE, 0x820, 0x1050, 0x1048, 0x24FE, 0x4590, 0x7A90, 0x10FC, 0x2490, 0x4490, 0x78FC, 0x890, 0x1090, 0x20FE, 0xC080}},
    {17705, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x820, 0xFFBE, 0x844, 0x7FA4, 0x4924, 0x7F28, 0x1C28, 0x2A10, 0x4928, 0x8844, 0x882}},
    {17706, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x100, 0x2928, 0x1110, 0x2928, 0x100, 0x2928, 0x1110, 0x2AA8, 0x440, 0x1830, 0xE00E}},
    {17707, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x7C20, 0x7FE, 0x1820, 0x51FC, 0x5D24, 0x512C, 0x5070, 0x5CA8, 0x7124, 0xC222, 0x20}},
    {17708, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {17709, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x4104, 0x1490, 0x1428, 0x23E8, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x2118}},
    {17710, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x220C, 0x2270, 0xFF40, 0x2240, 0x3E40, 0x227E, 0x3E48, 0x2248, 0xFF48, 0x2248, 0x4148, 0x8088}},
    {17711, 2, {0x820, 0xFFFE, 0x820, 0x10, 0x7F10, 0x1410, 0x14FE, 0x7F10, 0x5510, 0x557C, 0x5744, 0x6144, 0x7F44, 0x4144, 0x7F7C, 0x4144}},
    {17712, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x4004, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {17713, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x11F8, 0x1108, 0xFDF8, 0x1000, 0x13FC, 0x1108, 0x1DF8, 0xF108, 0x11F8, 0x110E, 0x57F8, 0x2008}},
    {17714, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x90, 0x7EA0, 0x444, 0x2828, 0x17D0, 0x2008, 0xDFF6, 0x1010, 0x1FF0, 0x820, 0x440, 0x7FFC}},
    {17715, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1028, 0x1E24, 0x32FE, 0x4A20, 0xA450, 0x1850, 0x1088, 0x2106, 0xC000, 0x2488, 0x2244, 0x4244}},
    {17716, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x2A00, 0x3E7C, 0x4944, 0x7F44, 0x44, 0x3E28, 0x28, 0x7F10, 0x810, 0x2A28, 0x4944, 0x1882}},
    {17717, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4104, 0x5FF4, 0x4384, 0x4544, 0x4924, 0x5124, 0x410C}},
    {17718, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x15FC, 0x1800, 0x37FE, 0xD020, 0x113C, 0x1120, 0x52A0, 0x247E}},
    {17719, 2, {0x820, 0xFFFE, 0x820, 0x0, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x100, 0x6C0, 0x1A30, 0xE10E, 0x1FE0, 0x40, 0x80}},
    {17720, 2, {0x440, 0xFFFE, 0x440, 0x120, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006, 0x1028, 0x1424, 0xFEFE, 0x1020, 0x2850, 0x4488, 0x8106}},
    {17721, 2, {0x820, 0xFFFE, 0x820, 0x108, 0x7890, 0x4BFC, 0x4880, 0x5144, 0x5668, 0x48B0, 0x4928, 0x4E68, 0x68A4, 0x5122, 0x46A0, 0x4040}},
    {17722, 2, {0x440, 0xFFFE, 0x440, 0x7BFC, 0x4880, 0x49F8, 0x5220, 0x55FC, 0x4800, 0x49F8, 0x4908, 0x69F8, 0x5108, 0x41F8, 0x4108, 0x4118}},
    {17723, 2, {0x820, 0xFFFE, 0x820, 0x1040, 0xFE40, 0x80, 0x7CFE, 0x4508, 0x7E88, 0x88, 0x7C50, 0x850, 0xFE20, 0x1050, 0x5088, 0x2306}},
    {17724, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7EFC, 0x1224, 0x52A4, 0x9324, 0x264C, 0x4090, 0x850, 0x2E5C, 0x2850, 0x2E50, 0xF1FE, 0x0}},
    {17725, 2, {0x820, 0xFFFE, 0x820, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x5100, 0x13FC, 0x2444, 0xEAA4, 0x2204, 0x23F4, 0x2014, 0x2008}},
    {17726, 2, {0x820, 0xFFFE, 0x820, 0x23F8, 0x2208, 0x23F8, 0xFA08, 0x23F8, 0x2900, 0x33FC, 0x6444, 0xAAA4, 0x2204, 0x23F4, 0xA014, 0x4008}},
    {17727, 2, {0x820, 0xFFFE, 0x820, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8, 0x20, 0xFFFE, 0x820, 0x460}},
    {17728, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x7F20, 0x83E, 0x3E42, 0x94, 0x3E10, 0x10, 0x7F10, 0x828, 0x2A28, 0x4944, 0x1882}},
    {17729, 2, {0x820, 0xFFFE, 0x820, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x540, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {17730, 2, {0x820, 0xFFFE, 0x820, 0x80, 0x2110, 0x1208, 0x17FC, 0x4, 0x73F8, 0x1208, 0x13F8, 0x10C4, 0x11A8, 0x1290, 0x2C88, 0x47FE}},
    {17731, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x78A0, 0x4910, 0x4BF8, 0x5506, 0x51F0, 0x4900, 0x49F0, 0x4900, 0x6FFC, 0x5100, 0x4210, 0x43F8}},
    {17732, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3518, 0xC206, 0x1FF0, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {17733, 2, {0x820, 0xFFFE, 0x820, 0x80, 0x112C, 0x1124, 0xFDAC, 0x2524, 0x25FC, 0x4820, 0x29F8, 0x1088, 0x2850, 0x4420, 0x80D8, 0x306}},
    {17734, 2, {0x440, 0xFFFE, 0x440, 0x20A0, 0x2110, 0x2208, 0xF5F6, 0x2000, 0x6784, 0x74A4, 0xAFA4, 0xA4A4, 0x27A4, 0x2484, 0x2494, 0x2588}},
    {17735, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2080, 0x2100, 0x2FF8, 0x2888, 0x2FF8, 0x2908, 0x2FF8, 0x4254, 0x445C, 0x9842, 0x603E}},
    {17736, 2, {0x820, 0xFFFE, 0x820, 0x48, 0x948, 0x4BFE, 0x4948, 0x7978, 0x900, 0x9FE, 0xF820, 0x4BFE, 0x4870, 0x48A8, 0x8B26, 0x820}},
    {17737, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x440, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {17738, 2, {0x820, 0xFFFE, 0x920, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE}},
    {17739, 2, {0x820, 0xFFFE, 0x820, 0x1040, 0x1020, 0x13FE, 0x1220, 0x7EFC, 0x1224, 0x13FE, 0x1224, 0x1EFC, 0xE220, 0x4FC, 0x484, 0x8FC}},
    {17740, 2, {0x820, 0xFFFE, 0x820, 0x0, 0xCA0, 0x7090, 0x11FE, 0x1120, 0xFB20, 0x31FC, 0x3920, 0x55FC, 0x5120, 0x9120, 0x11FE, 0x1100}},
    {17741, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0xFFFE, 0x488, 0x1A70, 0xEC18, 0x806, 0x3FF8, 0x2448, 0x2FE8, 0x3458, 0x27C8, 0x2028, 0x2010}},
    {17742, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1FC, 0x7C00, 0x44F8, 0x4488, 0x7CF8, 0x4400, 0x45FC, 0x7D24, 0x45FC, 0x4524, 0x7DFC, 0x4504}},
    {17743, 2, {0x440, 0xFFFE, 0x440, 0x1010, 0x820, 0x3FF8, 0x100, 0x7FFC, 0x240, 0x3C50, 0x848, 0xFFFE, 0x850, 0x7E24, 0x854, 0x198C}},
    {17744, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x21FC, 0x1000, 0xFCF8, 0x888, 0x10F8, 0x3800, 0x55FC, 0x9524, 0x11FC, 0x1124, 0x11FC, 0x1104}},
    {17745, 2, {0x440, 0xFFFE, 0x440, 0x1C, 0x11E0, 0x1044, 0x2524, 0x7C88, 0x820, 0x1020, 0x25FC, 0x7C48, 0xC8, 0x5430, 0x5468, 0x8184}},
    {17746, 2, {0x820, 0xFFFE, 0x920, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0xFFFE, 0x1000, 0x7EF8, 0x2288, 0x46F8}},
    {17747, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x5082, 0x9EFC, 0x2288, 0x5450, 0x2820, 0x17D8, 0x2006, 0xDFF0, 0x100, 0x1110, 0x2308}},
    {17748, 2, {0x820, 0xFFFE, 0x820, 0x2100, 0x411C, 0x8280, 0x1440, 0x2BBE, 0x7108, 0xA7C8, 0x2108, 0x2948, 0x2508, 0x21C8, 0x2E28, 0x2410}},
    {17749, 2, {0x820, 0xFFFE, 0x820, 0x11F8, 0x908, 0x9F8, 0x4108, 0x23FC, 0x2A94, 0xBFC, 0x1000, 0x73F8, 0x1090, 0x1060, 0x1198, 0x1606}},
    {17750, 2, {0x820, 0xFFFE, 0x820, 0x400C, 0x2F70, 0x2110, 0x224, 0x278, 0xE312, 0x2E7E, 0x2210, 0x2254, 0x2A92, 0x2430, 0x5000, 0x8FFE}},
    {17751, 2, {0x820, 0xFFFE, 0xA20, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x820, 0x1020}},
    {17752, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x13FC, 0x2840, 0x45F8, 0x9308, 0x9F8, 0xFD08, 0x5F8, 0x908, 0x51F8, 0x2000, 0x1090, 0x108}},
    {17753, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x7F28, 0x824, 0xFFFE, 0x1220, 0x3FA4, 0x6424, 0xBFA8, 0x2418, 0x3F92, 0x242A, 0x3FC6, 0x2082}},
    {17754, 2, {0x820, 0xFFFE, 0x820, 0x3FFC, 0x2080, 0x27F0, 0x2490, 0x27F0, 0x2080, 0x3FFC, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x2220, 0x3FFC}},
    {17755, 2, {0x440, 0xFFFE, 0x440, 0x2020, 0x1020, 0xFDFC, 0x20, 0x78A8, 0xA8, 0x78A8, 0x174, 0x7A22, 0x4850, 0x4888, 0x7904, 0x4A02}},
    {17756, 2, {0x820, 0xFFFE, 0x820, 0x7CF8, 0x2448, 0x1428, 0x2448, 0x100, 0x7FFC, 0x820, 0x1450, 0x2288, 0x100, 0xFFFE, 0x100, 0x100}},
    {17757, 2, {0x820, 0xFFFE, 0x820, 0x1E, 0xDE0, 0x7044, 0x1124, 0x10A8, 0xFDFC, 0x3124, 0x39FC, 0x5524, 0x53FE, 0x9104, 0x1114, 0x1108}},
    {17758, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x47C4, 0x4444, 0x47C4, 0x4444, 0x47D4, 0x44A4, 0x4544, 0x462C}},
    {17759, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x2FEC, 0x4824, 0x4BA4, 0x682C, 0x4BA4, 0x6AAC, 0x4BA4, 0x4824, 0xFFFE, 0x1010, 0x2008, 0x4004}},
    {17760, 2, {0x820, 0xFFFE, 0x820, 0x80, 0x3FFE, 0x2020, 0x2F28, 0x3110, 0x2A0A, 0x277C, 0x294A, 0x3784, 0x2878, 0x4F48, 0x4130, 0x874C}},
    {17761, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x47C4, 0x4444, 0x47C4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {17762, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x88, 0x7C50, 0x13FE, 0x1020, 0x21FC, 0x3C40, 0x67FE, 0xA480, 0x257C, 0x2610, 0x3C10, 0x25FE}},
    {17763, 2, {0x820, 0xFFFE, 0x820, 0x23F8, 0x1208, 0x3F8, 0xFA08, 0xBF8, 0x1100, 0x2BFC, 0x7444, 0xAAA4, 0x2A04, 0x23F4, 0x2014, 0x2008}},
    {17764, 2, {0x820, 0xFFFE, 0x820, 0x7700, 0x557C, 0x7744, 0x48, 0x7F50, 0x4948, 0x7F44, 0x4944, 0x7F54, 0x848, 0xFFC0, 0x840, 0x840}},
    {17765, 2, {0x440, 0xFFFE, 0x440, 0x1110, 0x2548, 0x4924, 0xBFFA, 0x2108, 0x3558, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3558, 0x2118}},
    {17766, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x1000, 0x1FF8, 0x820, 0xFE0, 0x7938, 0x4FA8, 0x7AA8, 0x4AB8, 0x792A, 0x4AAA, 0x9C46}},
    {17767, 2, {0x440, 0xFFFE, 0x440, 0x3E20, 0x2210, 0x3AFE, 0x2A00, 0x7F28, 0x4144, 0x3E82, 0x2228, 0x3E28, 0x2210, 0x3E28, 0x2244, 0x2682}},
    {17768, 2, {0x440, 0xFFFE, 0x440, 0x124, 0x3CA8, 0x25FC, 0x2440, 0x3DFE, 0x2488, 0x2544, 0x3E8A, 0x24F0, 0x2424, 0x45FC, 0x54A8, 0x8964}},
    {17769, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x1020, 0x13FE, 0x7850, 0x128A, 0x31FC, 0x3A8A, 0x54F8, 0x5088, 0x90F8, 0x1124, 0x1222, 0x1060}},
    {17770, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x23F8, 0x12A8, 0x93F8, 0x4040, 0x43F8, 0x1040, 0x27FC, 0xE110, 0x23F8, 0x2040, 0x27FC, 0x2040}},
    {17771, 2, {0x820, 0xFFFE, 0x8A0, 0x111E, 0x13D2, 0x5A5E, 0x53D0, 0x5A52, 0xE3CE, 0x1208, 0x27E4, 0x4822, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {17772, 2, {0x820, 0xFFFE, 0x820, 0x3FE, 0xF800, 0x9FC, 0x924, 0xF9FC, 0x8124, 0x87FE, 0xF800, 0x9FC, 0x924, 0x9FC, 0x5124, 0x27FE}},
    {17773, 2, {0x440, 0xFFFE, 0x440, 0x1010, 0x1054, 0x7C58, 0x1090, 0xFE28, 0x44, 0x5492, 0x9254, 0x7C58, 0x1090, 0x1C28, 0xE044, 0x4182}},
    {17774, 2, {0x440, 0xFFFE, 0x440, 0x1100, 0x11FC, 0x7E44, 0x1154, 0xFEE4, 0x4444, 0x29F4, 0x7C44, 0x10E4, 0xFF54, 0x1044, 0x1054, 0x1008}},
    {17775, 2, {0x820, 0xFFFE, 0x820, 0x0, 0xFE1C, 0x28E8, 0xFEA8, 0xAAA8, 0xFEA8, 0xA8, 0x7CA8, 0xA4, 0xFEA4, 0x554A, 0x9278, 0x3008}},
    {17776, 2, {0x440, 0xFFFE, 0x440, 0x101C, 0x13E0, 0x1524, 0x54A8, 0x5BFE, 0x5124, 0x9202, 0x11FC, 0x1124, 0x29FC, 0x2524, 0x45FC, 0x8104}},
    {17777, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7DFE, 0x20, 0xFC, 0x7E84, 0x42FC, 0x4284, 0x7EFC, 0x4284, 0x24FC, 0xF00, 0xF048, 0x4084}},
    {17778, 2, {0x820, 0xFFFE, 0x8A0, 0x3FFE, 0x2040, 0x2FFE, 0x2404, 0x25F4, 0x2514, 0x27FC, 0x2000, 0x27FC, 0x2000, 0x4FFE, 0x4248, 0x84C4}},
    {17779, 2, {0x820, 0xFFFE, 0x820, 0x0, 0xF7FC, 0x1248, 0x17FC, 0xFC46, 0x86EC, 0x8554, 0xF6EC, 0x1554, 0x16EC, 0x1444, 0x5454, 0x2408}},
    {17780, 2, {0x440, 0xFFFE, 0x440, 0xA0, 0xC90, 0x71FE, 0x1120, 0x13FC, 0xFD20, 0x31FC, 0x3920, 0x55FE, 0x5000, 0x92A4, 0x1252, 0x1452}},
    {17781, 2, {0x440, 0xFFFE, 0x450, 0x1020, 0x10FC, 0xFE84, 0x10FC, 0x1084, 0x7CFC, 0x80, 0xFE, 0x7C80, 0x44FE, 0x4402, 0x7D52, 0x4406}},
    {17782, 2, {0x440, 0xFFFE, 0x440, 0x2090, 0x10FE, 0xFD90, 0x2FC, 0x7890, 0xFC, 0x7890, 0xFC, 0x7880, 0x49FC, 0x4888, 0x7870, 0x4B8E}},
    {17783, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3492, 0x2120, 0x27C0, 0x2110, 0x4FF8, 0x44A4, 0x8990}},
    {17784, 2, {0x440, 0xFFFE, 0x440, 0x7F10, 0x4910, 0x6BFE, 0x5D10, 0x7F10, 0x87C, 0x7F00, 0x800, 0xF7C, 0xF044, 0x44, 0x557C, 0x5544}},
    {17785, 2, {0x440, 0xFFFE, 0x440, 0x2010, 0x1048, 0xFE40, 0x44FE, 0x4590, 0xAA90, 0x10FC, 0xFE90, 0x10FC, 0x5490, 0x5290, 0x92FE, 0x3080}},
    {17786, 2, {0x820, 0xFFFE, 0x820, 0x2448, 0x1450, 0x7FFC, 0x820, 0x440, 0x7FFC, 0x100, 0xFFFE, 0x2200, 0x3E7C, 0x23A4, 0x7E18, 0x266}},
    {17787, 2, {0x440, 0xFFFE, 0x440, 0x3F7E, 0x1224, 0x1B34, 0x12A4, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {17788, 2, {0x820, 0xFFFE, 0x820, 0x4450, 0xFE48, 0x44FE, 0x7D90, 0x1290, 0xFEFE, 0x9290, 0xFE90, 0x10FE, 0xFE90, 0x2890, 0x44FE, 0x8280}},
    {17789, 2, {0x820, 0xFFFE, 0x820, 0x1FC, 0x7D54, 0x45FC, 0x4480, 0x7DFC, 0x1244, 0x11F4, 0x5D54, 0x51F4, 0x5044, 0x5DF4, 0xE014, 0x8}},
    {17790, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x1E78, 0x1248, 0x3FFC, 0x2020, 0x2E20, 0x223C, 0x3FC8, 0x2928, 0x2F28, 0x4990, 0x5F28, 0x8144}},
    {17791, 2, {0x440, 0xFFFE, 0x450, 0x4420, 0xFEFC, 0x4484, 0x7CFC, 0x1084, 0xFEFC, 0x9280, 0xFEFE, 0x1080, 0xFEFE, 0x2802, 0x4552, 0x8206}},
    {17792, 2, {0x820, 0xFFFE, 0x820, 0x3C78, 0x2448, 0xFFFE, 0x200, 0x1FF0, 0x1490, 0x1250, 0x1FF0, 0x820, 0x1010, 0x7C7C, 0x4444, 0x7C7C}},
    {17793, 2, {0x2810, 0xFEFE, 0x2844, 0x28, 0x7CFE, 0x5400, 0x7C7C, 0x5444, 0x7C7C, 0x1044, 0xFE7C, 0x9210, 0x920A, 0xBAAA, 0x8AA4, 0x851C}},
    {17794, 2, {0x820, 0xFFFE, 0x2928, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0xF0, 0x3F00, 0x100, 0xFFFE, 0x100, 0x300}},
    {17795, 2, {0x820, 0xFFFE, 0xA20, 0x7FFC, 0x800, 0x1FF0, 0x6810, 0xFF0, 0x4444, 0xFEFE, 0x5454, 0xFEFE, 0x2020, 0x7C7C, 0xC4C4, 0x7D7C}},
    {17796, 2, {0x820, 0xFFFE, 0x820, 0x4450, 0xFE48, 0x44FE, 0x7D90, 0x10FC, 0xFE90, 0x92FC, 0xFE90, 0x1090, 0xFEFE, 0x2880, 0x4448, 0x8224}},
    {17797, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x3BB8, 0x2AA8, 0x3BB8, 0x0, 0x7FFC, 0x1110, 0x2928, 0xFFFE}},
    {17798, 2, {0x440, 0xFFFE, 0x440, 0x20, 0xFBFE, 0x2154, 0xF9DC, 0xA888, 0xA9FC, 0xD888, 0x89FC, 0x8888, 0xFBFE, 0x898A, 0xFEA4, 0x88C2}},
    {17799, 2, {0x820, 0xFFFE, 0x2A28, 0xF3BC, 0x2508, 0x739C, 0xAD6A, 0x27C8, 0x0, 0xFFFE, 0x950A, 0x6B70, 0x5504, 0x7F38, 0x2482, 0x3FBC}},
    {17800, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x2418, 0x24E0, 0x2704, 0x2404, 0x4404, 0x43FC, 0x8000}},
    {17801, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x23F8, 0x2040, 0x2040, 0x2FFE, 0x4040, 0x4140, 0x8080}},
    {17802, 2, {0x20, 0x43E, 0x420, 0x85FE, 0x4522, 0x2938, 0x29E0, 0x1122, 0x111E, 0x2900, 0x2978, 0x4548, 0x4548, 0x824A, 0x24A, 0x486}},
    {17803, 2, {0x410, 0x790, 0x410, 0x3F90, 0x2490, 0x2710, 0x3C10, 0x2490, 0x2390, 0x2028, 0x2F28, 0x2928, 0x2964, 0x49A4, 0x9124, 0x2042}},
    {17804, 2, {0x404, 0x784, 0x404, 0x3FD4, 0x2454, 0x2714, 0x3C54, 0x23D4, 0x2014, 0x2F14, 0x2914, 0x2904, 0x4944, 0x4984, 0x9114, 0x2008}},
    {17805, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2100, 0x21FC, 0x2210, 0x2510, 0x28A0, 0x4040, 0x41B0, 0x860C}},
    {17806, 2, {0x40, 0x7840, 0x487C, 0x5040, 0x53FC, 0x6244, 0x5270, 0x4BC0, 0x4A44, 0x4A3C, 0x6A00, 0x52F0, 0x4490, 0x4492, 0x4912, 0x520E}},
    {17807, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x22FC, 0x2200, 0x27F8, 0x2D48, 0x3348, 0x22E8, 0x449E, 0x4908, 0x8218}},
    {17808, 2, {0x410, 0xE1E, 0x3010, 0x20FE, 0x2092, 0x209C, 0x3EF0, 0x2492, 0x248E, 0x2480, 0x24B8, 0x24A8, 0x24A8, 0x452A, 0x452A, 0x8246}},
    {17809, 2, {0x800, 0xF00, 0x87C, 0x7F44, 0x4944, 0x4E44, 0x787C, 0x4944, 0x4744, 0x4044, 0x4E7C, 0x4A44, 0x4A02, 0x5202, 0x91FE, 0x2000}},
    {17810, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x2FF8, 0x2000, 0x27F0, 0x2410, 0x47F0, 0x4220, 0x9FFC}},
    {17811, 2, {0x1010, 0x101E, 0x9210, 0x92FE, 0x9292, 0x929C, 0xFEF0, 0x1092, 0x108E, 0x9280, 0x92B8, 0x92A8, 0x9EA8, 0xF32A, 0x12A, 0x246}},
    {17812, 2, {0x1020, 0x103E, 0x1020, 0x7DFE, 0x1122, 0x1138, 0x11E0, 0xFD22, 0x111E, 0x1100, 0x2178, 0x4948, 0xFD48, 0x464A, 0x24A, 0x486}},
    {17813, 2, {0x800, 0xF00, 0x878, 0x7F48, 0x4948, 0x4E78, 0x7848, 0x4948, 0x4778, 0x4048, 0x4E48, 0x4AFC, 0x4A02, 0x5202, 0x91FE, 0x2000}},
    {17814, 2, {0x800, 0xF00, 0x87C, 0x7F54, 0x4954, 0x4E7C, 0x7854, 0x4954, 0x477C, 0x4010, 0x4E10, 0x4A10, 0x4A12, 0x5202, 0x91FE, 0x2000}},
    {17815, 2, {0x1FC, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2028, 0x2FFE, 0x2820, 0x2BA4, 0x2828, 0x2BAC, 0x2A94, 0x4BAC, 0x9044}},
    {17816, 2, {0x1040, 0x1E40, 0x1078, 0x7E88, 0x5310, 0x5CFC, 0x70A4, 0x52A4, 0x4EFC, 0x4020, 0x5C50, 0x5458, 0x5454, 0x5692, 0x9492, 0x210E}},
    {17817, 2, {0x3F90, 0x2418, 0x3F14, 0x2112, 0x3F10, 0x247C, 0x3F80, 0x2080, 0x2FFC, 0x2884, 0x28F0, 0x2F84, 0x287C, 0x49E0, 0x4924, 0x921C}},
    {17818, 2, {0x1020, 0x1E40, 0x10FC, 0x7E84, 0x5284, 0x5CFC, 0x7080, 0x52FE, 0x4E80, 0x40FE, 0x5C02, 0x54AA, 0x54AA, 0x5602, 0x9414, 0x2008}},
    {17819, 2, {0x810, 0xF10, 0x87C, 0x7F10, 0x49FE, 0x4EA2, 0x7844, 0x4978, 0x4714, 0x40FE, 0x4E10, 0x4A54, 0x4AB2, 0x5202, 0x91FE, 0x2000}},
    {17820, 2, {0x1E3C, 0x1020, 0x7EFE, 0x52A2, 0x5CB8, 0x72E4, 0x4E9C, 0x4080, 0x5CB8, 0x952A, 0x2346, 0x5FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {17821, 2, {0x8, 0x7C, 0x780, 0x7900, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2100, 0x110, 0x1F8, 0x7F04, 0x2004}},
    {17822, 2, {0x808, 0x808, 0x808, 0x7F08, 0x49FE, 0x4908, 0x4908, 0x4908, 0x7F48, 0x4828, 0x828, 0xA08, 0xF08, 0xF108, 0x4028, 0x10}},
    {17823, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1FF8, 0x0, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {17824, 2, {0x1000, 0x11F0, 0x1110, 0x7D10, 0x5510, 0x5510, 0x5590, 0x5550, 0x7D50, 0x5110, 0x1110, 0x1512, 0x1D12, 0xE612, 0x420E, 0x400}},
    {17825, 2, {0x1000, 0x11F8, 0x1048, 0x7C48, 0x5448, 0x5448, 0x5448, 0x55F8, 0x7C88, 0x5088, 0x1088, 0x1488, 0x1C88, 0xE488, 0x43FE, 0x0}},
    {17826, 2, {0x1020, 0x1020, 0x1020, 0x7CA8, 0x54A4, 0x54A2, 0x5522, 0x5520, 0x7E24, 0x5024, 0x1028, 0x1408, 0x1E10, 0xE220, 0x40C0, 0x300}},
    {17827, 2, {0x1008, 0x101C, 0x11F0, 0x7D50, 0x5550, 0x5550, 0x5550, 0x5550, 0x7D50, 0x5148, 0x1148, 0x1548, 0x1D44, 0xE644, 0x4242, 0x400}},
    {17828, 2, {0x1000, 0x10F8, 0x3E88, 0x4288, 0xA4A8, 0x1894, 0x1084, 0x207C, 0xC100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {17829, 2, {0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x100, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {17830, 2, {0x1020, 0x1020, 0x1124, 0x7D24, 0x5524, 0x5524, 0x55FC, 0x5420, 0x7C20, 0x5124, 0x1124, 0x1524, 0x1D24, 0xE524, 0x41FC, 0x4}},
    {17831, 2, {0x1000, 0x1000, 0x10FC, 0x7C84, 0x5484, 0x5484, 0x54FC, 0x5484, 0x7C84, 0x5084, 0x10FC, 0x1484, 0x1E00, 0xE200, 0x41FE, 0x0}},
    {17832, 2, {0x1000, 0x1000, 0x11FE, 0x7C20, 0x5420, 0x5440, 0x5440, 0x54FC, 0x7C84, 0x5184, 0x1284, 0x1484, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {17833, 2, {0x1040, 0x1020, 0x1028, 0x7C08, 0x5408, 0x5448, 0x5454, 0x5552, 0x7D52, 0x5160, 0x1260, 0x1444, 0x1CC4, 0xE544, 0x423C, 0x0}},
    {17834, 2, {0x880, 0x888, 0x2E90, 0x28A0, 0x28C0, 0x2884, 0x2E84, 0xF07C, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {17835, 2, {0x1010, 0x1050, 0x1050, 0x7C88, 0x5488, 0x5504, 0x56FA, 0x5420, 0x7C20, 0x5020, 0x11FC, 0x1420, 0x1E20, 0xE220, 0x4020, 0x20}},
    {17836, 2, {0x1000, 0x11FC, 0x1124, 0x7D24, 0x5524, 0x55FC, 0x5524, 0x5524, 0x7D24, 0x51FC, 0x1124, 0x1420, 0x1C20, 0xE420, 0x4020, 0x20}},
    {17837, 2, {0x1110, 0x910, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x3018, 0xC106, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {17838, 2, {0x1020, 0x1020, 0x1040, 0x7DFC, 0x5504, 0x5504, 0x5574, 0x5554, 0x7D54, 0x5154, 0x1154, 0x1574, 0x1D04, 0xE504, 0x4114, 0x108}},
    {17839, 2, {0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17840, 2, {0x1004, 0x101E, 0x11E0, 0x7D00, 0x5506, 0x5578, 0x5550, 0x5550, 0x7D52, 0x1154, 0x1548, 0x1D48, 0xE544, 0x4254, 0x262, 0x440}},
    {17841, 2, {0x1000, 0x11FE, 0x1100, 0x7D00, 0x55FE, 0x5510, 0x5510, 0x557C, 0x7D10, 0x5110, 0x1110, 0x15FE, 0x1D00, 0xE500, 0x41FE, 0x0}},
    {17842, 2, {0x1080, 0x1080, 0x11FC, 0x7D04, 0x5604, 0x55E4, 0x5524, 0x5524, 0x7DE4, 0x5124, 0x1124, 0x15E4, 0x1D04, 0xE404, 0x4028, 0x10}},
    {17843, 2, {0x1020, 0x1010, 0x11FE, 0x7D02, 0x5400, 0x541C, 0x54E0, 0x5420, 0x7C20, 0x503E, 0x11E0, 0x1420, 0x1E22, 0xE222, 0x401E, 0x0}},
    {17844, 2, {0x1050, 0x1050, 0x1050, 0x7D52, 0x54D4, 0x5458, 0x5450, 0x5458, 0x7CD4, 0x5152, 0x1050, 0x1450, 0x1E92, 0xE292, 0x4112, 0x20E}},
    {17845, 2, {0x1020, 0x1020, 0x1020, 0x7DFC, 0x5524, 0x5524, 0x5524, 0x5524, 0x7DFC, 0x5124, 0x1020, 0x1428, 0x1C24, 0xE5FE, 0x4082, 0x0}},
    {17846, 2, {0x1000, 0x13FE, 0x1040, 0x7C80, 0x5544, 0x5624, 0x5468, 0x54B0, 0x7D30, 0x5228, 0x1068, 0x14A4, 0x1D22, 0xE620, 0x40A0, 0x40}},
    {17847, 2, {0x2184, 0x2068, 0x2030, 0xF8C8, 0xAB04, 0xA840, 0xABFE, 0xA8A0, 0xF920, 0xA3FC, 0x2524, 0x2924, 0x3934, 0xE928, 0x4020, 0x20}},
    {17848, 2, {0x240, 0x3C50, 0x848, 0xFFFE, 0x848, 0xE50, 0x7824, 0x854, 0x198C, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17849, 2, {0x2088, 0x2088, 0x2088, 0xFB8E, 0xA888, 0xA888, 0xA800, 0xA9FC, 0xF820, 0xA020, 0x23FE, 0x2820, 0x3820, 0xE820, 0x4020, 0x20}},
    {17850, 2, {0x1080, 0x1080, 0x11F8, 0x2108, 0x2A90, 0x6C60, 0xA998, 0x2E46, 0x2840, 0x2BF8, 0x2A48, 0x2A48, 0x2BF8, 0x2044, 0x27FC, 0x2004}},
    {17851, 2, {0x1020, 0x1010, 0x11FE, 0x7C20, 0x5448, 0x5484, 0x55FE, 0x5402, 0x7CA8, 0x50A8, 0x10A8, 0x14A8, 0x1EA8, 0xE2AA, 0x412A, 0x206}},
    {17852, 2, {0x2100, 0x2100, 0x211E, 0xFB92, 0xA912, 0xA912, 0xA912, 0xAFD2, 0xF912, 0xA112, 0x221A, 0x2A94, 0x3FD0, 0xEA50, 0x4010, 0x10}},
    {17853, 2, {0x308, 0x3C08, 0x448, 0x7F48, 0xC48, 0x1648, 0x2508, 0x4428, 0x510, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {17854, 2, {0x1020, 0x1010, 0x11FE, 0x7D02, 0x5604, 0x54F8, 0x5400, 0x5400, 0x7DFE, 0x5050, 0x1050, 0x1450, 0x1E92, 0xE292, 0x410E, 0x200}},
    {17855, 2, {0xC80, 0x70FC, 0x1124, 0xFD24, 0x1224, 0x3844, 0x5484, 0x9128, 0x1010, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17856, 2, {0x2020, 0x2020, 0x21FC, 0xF820, 0xA820, 0xABFE, 0xA888, 0xA944, 0xFA42, 0xA0F8, 0x2188, 0x2A50, 0x3820, 0xE850, 0x4188, 0x606}},
    {17857, 2, {0x1020, 0x1020, 0x11FC, 0x7C50, 0x5488, 0x5504, 0x57FE, 0x5408, 0x7DE8, 0x5128, 0x1128, 0x15E8, 0x1D28, 0xE408, 0x4028, 0x10}},
    {17858, 2, {0x1020, 0x1020, 0x103C, 0x7C20, 0x55FC, 0x5524, 0x5530, 0x55E0, 0x7D24, 0x511C, 0x1100, 0x1570, 0x1D50, 0xE552, 0x4292, 0x50E}},
    {17859, 2, {0x1048, 0x1044, 0x105E, 0x7DE0, 0x5428, 0x5412, 0x546A, 0x5596, 0x7C48, 0x505E, 0x11E0, 0x1424, 0x1E28, 0xE212, 0x406A, 0x186}},
    {17860, 2, {0x1050, 0x1088, 0x1154, 0x7C20, 0x54D8, 0x5706, 0x5408, 0x54F0, 0x7C80, 0x5080, 0x10FC, 0x1490, 0x1C90, 0xE510, 0x4210, 0x10}},
    {17861, 2, {0x1000, 0x10FC, 0x1080, 0x7CF8, 0x5480, 0x54F8, 0x5480, 0x57FE, 0x7D40, 0x5124, 0x1128, 0x1510, 0x1D08, 0xE544, 0x4182, 0x100}},
    {17862, 2, {0x1020, 0x1010, 0x11FE, 0x7D02, 0x5604, 0x54F8, 0x5488, 0x5488, 0x7CF8, 0x5080, 0x1080, 0x14FC, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {17863, 2, {0x1080, 0x1080, 0x10FE, 0x7D02, 0x5622, 0x54AA, 0x5472, 0x5422, 0x7DFE, 0x5022, 0x1072, 0x14AA, 0x1D22, 0xE422, 0x400A, 0x4}},
    {17864, 2, {0x1040, 0x1020, 0x13FE, 0x7C80, 0x5480, 0x5480, 0x54FC, 0x5400, 0x7CFC, 0x5084, 0x10FC, 0x1484, 0x1EFC, 0xE284, 0x40FC, 0x84}},
    {17865, 2, {0x10F8, 0x1088, 0x1088, 0x7CF8, 0x5488, 0x5488, 0x54F8, 0x5400, 0x7DFC, 0x5104, 0x1104, 0x15FC, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {17866, 2, {0x1020, 0x10A2, 0x10A2, 0x7D24, 0x5450, 0x5488, 0x5704, 0x5422, 0x7C20, 0x50A4, 0x10A4, 0x1528, 0x1C50, 0xE488, 0x4104, 0x202}},
    {17867, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x5488, 0x54F8, 0x5488, 0x5488, 0x7CF8, 0x5020, 0x13FE, 0x1470, 0x1CA8, 0xE524, 0x4222, 0x20}},
    {17868, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5420, 0x5520, 0x7DFE, 0x5220, 0x1020, 0x15FC, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {17869, 2, {0xFE0, 0x1020, 0x1FC0, 0x40, 0xFFFE, 0xC10, 0x7320, 0xDC0, 0x72B0, 0xC8E, 0xF300, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {17870, 2, {0x1000, 0x13FE, 0x1050, 0x7C50, 0x55FC, 0x5554, 0x5554, 0x5554, 0x7DFC, 0x5020, 0x1020, 0x15FC, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {17871, 2, {0x1040, 0x1020, 0x11FE, 0x7D22, 0x54A4, 0x5528, 0x5450, 0x5488, 0x7D04, 0x50F8, 0x1088, 0x1488, 0x1E50, 0xE220, 0x40D8, 0x306}},
    {17872, 2, {0x1040, 0xFEFE, 0x2120, 0x3CAC, 0x24F4, 0x27A4, 0x44AC, 0x54A2, 0x887E, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17873, 2, {0x1000, 0x10FC, 0x1084, 0x7CFC, 0x5484, 0x54FC, 0x5484, 0x5420, 0x7C10, 0x51FE, 0x1000, 0x1484, 0x1E48, 0xE200, 0x41FE, 0x0}},
    {17874, 2, {0x2108, 0x2090, 0x2000, 0xFBFC, 0xA890, 0xA890, 0xA892, 0xAC92, 0xFA94, 0xA298, 0x2090, 0x2890, 0x3890, 0xE890, 0x47FE, 0x0}},
    {17875, 2, {0x2080, 0x2080, 0x20FC, 0xF954, 0xAAD4, 0xA8B4, 0xA92C, 0xAA44, 0xF894, 0xA108, 0x2040, 0x28A4, 0x3AAA, 0xEA8A, 0x4478, 0x0}},
    {17876, 2, {0x1080, 0x10BC, 0x1084, 0x7D08, 0x55FE, 0x5720, 0x5520, 0x557C, 0x7D90, 0x5110, 0x11FE, 0x1510, 0x1D28, 0xE528, 0x4144, 0x182}},
    {17877, 2, {0x100, 0x7FFC, 0x5004, 0x3EF8, 0x4288, 0xA2AA, 0x1492, 0x187E, 0xE100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {17878, 2, {0x20A8, 0x20A8, 0x23FE, 0xF8A8, 0xA8A8, 0xA800, 0xABFE, 0xAA22, 0xF820, 0xA1FC, 0x2124, 0x2924, 0x3934, 0xE928, 0x4020, 0x20}},
    {17879, 2, {0x2210, 0x2210, 0x23DE, 0xFD28, 0xA884, 0xA800, 0xABFC, 0xA890, 0xF890, 0xA090, 0x27FE, 0x2890, 0x3890, 0xE910, 0x4210, 0x410}},
    {17880, 2, {0x1020, 0x1020, 0x10FC, 0x7C20, 0x5420, 0x55FE, 0x5440, 0x5484, 0x7DFE, 0x5002, 0x11FC, 0x1554, 0x1D54, 0xE554, 0x43FE, 0x0}},
    {17881, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1110, 0x2FE8, 0xC926, 0xFE0, 0x110, 0x3FF8, 0x1008}},
    {17882, 2, {0x1028, 0x1024, 0x13FE, 0x7C20, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x7DFC, 0x5124, 0x1008, 0x17FE, 0x1C88, 0xE448, 0x4048, 0x18}},
    {17883, 2, {0x2120, 0x213C, 0x2144, 0xFAA8, 0xAA10, 0xAEA8, 0xAAC6, 0xAABC, 0xFAA4, 0xA2BC, 0x22A4, 0x2ABC, 0x3A24, 0xEA24, 0x4224, 0x22C}},
    {17884, 2, {0x2040, 0x23BE, 0x2212, 0xFA92, 0xAA52, 0xAAAA, 0xAB24, 0xA840, 0xFBFE, 0xA222, 0x2222, 0x2BFE, 0x3A22, 0xEA22, 0x43FE, 0x202}},
    {17885, 2, {0x2020, 0x1020, 0xFDFC, 0x924, 0x11FC, 0x3924, 0x55FC, 0x9020, 0x1020, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17886, 2, {0x0, 0x3FF8, 0x100, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006, 0x1020, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {17887, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17888, 2, {0x2040, 0x1040, 0xFEFE, 0x2100, 0x3DFE, 0x2422, 0x24A0, 0x44BC, 0x5560, 0x8A3E, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {17889, 2, {0x2000, 0x23FE, 0x2020, 0xF93C, 0xA920, 0xAFFE, 0xA880, 0xA8FE, 0xF900, 0xA1FE, 0x2002, 0x2AAA, 0x3AAA, 0xEC02, 0x4014, 0x8}},
    {17890, 2, {0x1010, 0x1050, 0x1250, 0x7D7C, 0x5490, 0x5410, 0x54FE, 0x5700, 0x7D7C, 0x5144, 0x1144, 0x157C, 0x1D44, 0xE500, 0x42FE, 0x400}},
    {17891, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xFE0, 0x820, 0xFE0, 0x0, 0x1020, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {17892, 2, {0x1010, 0x11FE, 0x1110, 0x7D7C, 0x5514, 0x55FE, 0x5514, 0x557C, 0x7D10, 0x117C, 0x1554, 0x1D7C, 0xE554, 0x427C, 0x254, 0x44C}},
    {17893, 2, {0x2124, 0x2124, 0x2224, 0xFC24, 0xA954, 0xA94A, 0xAA92, 0xAE10, 0xFA10, 0xA250, 0x225C, 0x2A50, 0x3A50, 0xEAB0, 0x429E, 0x300}},
    {17894, 2, {0x2020, 0x247C, 0x2284, 0xF948, 0xA830, 0xA8CE, 0xAE10, 0xAA7C, 0xFA10, 0xA27C, 0x2210, 0x2AFE, 0x3A10, 0xEA10, 0x45FE, 0x800}},
    {17895, 2, {0x1020, 0x1010, 0x11FE, 0x7D10, 0x557C, 0x5514, 0x55FE, 0x5514, 0x7D7C, 0x1110, 0x1592, 0x1D54, 0xE538, 0x4254, 0x292, 0x430}},
    {17896, 2, {0x1040, 0x1040, 0x10F8, 0x7D08, 0x5610, 0x55FC, 0x5524, 0x5524, 0x7DFC, 0x5124, 0x1124, 0x15FC, 0x1C00, 0xE554, 0x412A, 0x22A}},
    {17897, 2, {0x200C, 0x200A, 0x2008, 0xFBFE, 0xAA08, 0xAA48, 0xAA6A, 0xAA4A, 0xFBFA, 0xA24C, 0x224C, 0x2AEC, 0x3B5A, 0xEC4A, 0x44D6, 0x822}},
    {17898, 2, {0x80, 0x3FFE, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x23E0, 0x2000, 0x2A48, 0x3124, 0x2080, 0x2FF8, 0x4888, 0x4FF8, 0x8084, 0x1FFE}},
    {17899, 2, {0x1020, 0x1010, 0x11FE, 0x7D48, 0x5548, 0x5548, 0x55FE, 0x5548, 0x7D48, 0x51EC, 0x115A, 0x1568, 0x1D48, 0xE548, 0x4248, 0x48}},
    {17900, 2, {0x2100, 0x13F0, 0x1510, 0xE0, 0x775C, 0x11F0, 0x1040, 0x13FC, 0x2840, 0x47FE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {17901, 2, {0x2020, 0x2020, 0x21FC, 0xF924, 0xA9FC, 0xA820, 0xABFE, 0xAA22, 0xFBFE, 0xA020, 0x2040, 0x2824, 0x3AA2, 0xEA8A, 0x4478, 0x0}},
    {17902, 2, {0x2020, 0x2020, 0x23FE, 0xF820, 0xAAAA, 0xA924, 0xAAAA, 0xA820, 0xFAAA, 0xA124, 0x22AA, 0x2850, 0x3850, 0xE888, 0x4104, 0x602}},
    {17903, 2, {0x2028, 0x2224, 0x217E, 0xF9C8, 0xA948, 0xA87E, 0xAB48, 0xA948, 0xF97E, 0xA148, 0x2148, 0x297E, 0x3940, 0xEA80, 0x447E, 0x0}},
    {17904, 2, {0x2000, 0x21FC, 0x2124, 0xFBFE, 0xA924, 0xA9FC, 0xA800, 0xA9FC, 0xF904, 0xA1FC, 0x2104, 0x29FC, 0x3904, 0xE9FC, 0x4088, 0x104}},
    {17905, 2, {0x2050, 0x2252, 0x2154, 0xF850, 0xABFE, 0xA888, 0xA850, 0xABFE, 0xF820, 0xA1FC, 0x2020, 0x2BFE, 0x3850, 0xE888, 0x4104, 0x602}},
    {17906, 2, {0x2020, 0x2124, 0x20A8, 0xFBFE, 0xA8A8, 0xA924, 0xAA22, 0xA904, 0xF904, 0xA1DE, 0x2244, 0x2D54, 0x389E, 0xE904, 0x4204, 0x404}},
    {17907, 2, {0x2010, 0x23D2, 0x2054, 0xF948, 0xA88A, 0xA904, 0xAAFA, 0xAC00, 0xF9FC, 0xA104, 0x2104, 0x29FC, 0x3904, 0xE888, 0x47FE, 0x0}},
    {17908, 2, {0x2110, 0x2110, 0x21DE, 0xFAA8, 0xAC44, 0xA806, 0xABB8, 0xA888, 0xF908, 0xA3BE, 0x2088, 0x2A88, 0x393E, 0xEAC0, 0x443E, 0x0}},
    {17909, 2, {0x1000, 0x13FE, 0x1000, 0x7DFC, 0x5524, 0x55FC, 0x5524, 0x57FE, 0x7C00, 0x51FC, 0x1124, 0x15FC, 0x1D24, 0xE5FC, 0x4000, 0x3FE}},
    {17910, 2, {0x1088, 0x13FE, 0x1088, 0x7DFC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x7C80, 0x51FE, 0x1222, 0x1552, 0x1D02, 0xE5FA, 0x400A, 0x4}},
    {17911, 2, {0x1040, 0x1020, 0x11FC, 0x7C88, 0x5450, 0x57FE, 0x5400, 0x55FC, 0x7D04, 0x51FC, 0x1104, 0x15FC, 0x1C20, 0xE694, 0x428A, 0x47A}},
    {17912, 2, {0x100, 0xFFFE, 0x1048, 0x24FE, 0x7990, 0x10FC, 0x2290, 0x7CFC, 0x890, 0x30FE, 0xC180, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {17913, 2, {0x2108, 0x2090, 0x23FE, 0xF908, 0xA908, 0xAA52, 0xAB9C, 0xA908, 0xFA52, 0xA3DE, 0x2040, 0x28A4, 0x3AAA, 0xEA8A, 0x4478, 0x0}},
    {17914, 2, {0x18, 0x7BE0, 0x1040, 0x2278, 0x7A40, 0xA40, 0x2BFC, 0x1000, 0x2FFE, 0x4000, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17915, 2, {0x2108, 0x2108, 0x21EE, 0xFA94, 0xA842, 0xA800, 0xABDE, 0xAA52, 0xFBD2, 0xA252, 0x23D2, 0x2A1A, 0x3A94, 0xEB50, 0x4210, 0x10}},
    {17916, 2, {0x201C, 0x21E0, 0x2020, 0xFBFE, 0xA820, 0xA9FC, 0xA9AC, 0xA974, 0xF9FC, 0xA020, 0x21FC, 0x2820, 0x3BFE, 0xE800, 0x4154, 0x22A}},
    {17917, 2, {0x10A0, 0x2090, 0x45FE, 0x7920, 0x13FC, 0x2520, 0x7DFC, 0x120, 0x55FE, 0x8100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {17918, 2, {0x2000, 0x23FE, 0x2202, 0xFBFE, 0xAA02, 0xABFE, 0xA908, 0xAA52, 0xFB9C, 0xA108, 0x2252, 0x2BDE, 0x3800, 0xEAA4, 0x4252, 0x452}},
    {17919, 2, {0x2040, 0x2020, 0x23FE, 0xFA22, 0xA954, 0xA94A, 0xAA3A, 0xA800, 0xF9FC, 0xA154, 0x23FE, 0x2800, 0x39FC, 0xE820, 0x40A0, 0x40}},
    {17920, 2, {0x800, 0x3FF0, 0x2010, 0x3FF0, 0x2000, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x1120, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17921, 2, {0x828, 0x7F24, 0x820, 0xFFFE, 0x1220, 0x3FA4, 0x6424, 0xBF28, 0x2412, 0x3FAA, 0x146, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {17922, 2, {0x2088, 0x208C, 0x23EA, 0xF888, 0xA888, 0xABFE, 0xA948, 0xAB6A, 0xF94A, 0xA36A, 0x214C, 0x2B6C, 0x394A, 0xE96A, 0x4796, 0x22}},
    {17923, 2, {0x21FC, 0x2124, 0x21AC, 0xF924, 0xA9FC, 0xA820, 0xA9FC, 0xA820, 0xFBFE, 0xA154, 0x222A, 0x2820, 0x39FC, 0xE820, 0x43FE, 0x0}},
    {17924, 2, {0x108, 0x1FD0, 0x120, 0x7FFC, 0x200, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17925, 2, {0x2044, 0x4258, 0xFF62, 0x142, 0x7E3E, 0x4200, 0x7E4C, 0x4272, 0x7E42, 0x423E, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17926, 2, {0x1020, 0x13FE, 0x1222, 0x7DFC, 0x5420, 0x54F8, 0x5420, 0x55FC, 0x7C20, 0x11FC, 0x1554, 0x1DFC, 0xE440, 0x42A4, 0x28A, 0x47A}},
    {17927, 2, {0x100, 0x7FFC, 0x820, 0x7C0, 0x1930, 0xEFEE, 0x920, 0xFE0, 0x110, 0x3FF8, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17928, 2, {0x2FE8, 0x2948, 0x2BE8, 0xF95E, 0xA80A, 0xABEA, 0xAAAA, 0xABEA, 0xFAAA, 0xABEA, 0x288A, 0x2BEA, 0x3AAA, 0xEBEA, 0x4A2A, 0x1276}},
    {17929, 2, {0x3FF8, 0x100, 0x7FFC, 0x440, 0x1930, 0xEFEE, 0x920, 0xFE0, 0x110, 0x3FF8, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17930, 2, {0x2210, 0x23DE, 0x2528, 0xF884, 0xABFC, 0xAA94, 0xAA94, 0xABFC, 0xF828, 0xA3FE, 0x2220, 0x2B24, 0x3AA8, 0xEA92, 0x442A, 0x8C6}},
    {17931, 2, {0x2148, 0x214C, 0x22AA, 0xF808, 0xABFE, 0xA948, 0xAB68, 0xA94A, 0xFB6A, 0xA14A, 0x236C, 0x294C, 0x396A, 0xEB8A, 0x4016, 0x22}},
    {17932, 2, {0x3FFC, 0x2420, 0x2FF8, 0x2520, 0x3FFC, 0x2200, 0x27F0, 0x2A10, 0x33F0, 0x3FFC, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17933, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x7FFC, 0x4124, 0x5FEC, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {17934, 2, {0x2444, 0x28E8, 0x2E0E, 0xFCE4, 0xAA0A, 0xAEEE, 0xA8A0, 0xAAEA, 0xF800, 0xA3F8, 0x2008, 0x2BF8, 0x3A00, 0xEBFC, 0x4004, 0x38}},
    {17935, 2, {0x800, 0x83E, 0x1022, 0x7F24, 0x5524, 0x5528, 0x5524, 0x5524, 0x5522, 0x5522, 0x5522, 0x5534, 0x57A8, 0xFC20, 0x20, 0x20}},
    {17936, 2, {0x800, 0x800, 0x1078, 0x7F28, 0x5528, 0x5528, 0x556C, 0x556A, 0x55AA, 0x552A, 0x5528, 0x5528, 0x57C8, 0xFC48, 0x98, 0x100}},
    {17937, 2, {0x830, 0x808, 0x1000, 0x7F70, 0x5510, 0x551A, 0x55DA, 0x555C, 0x5558, 0x5558, 0x5554, 0x5554, 0x5794, 0xFC12, 0x50, 0x20}},
    {17938, 2, {0x810, 0x810, 0x1020, 0x7F44, 0x55FE, 0x5528, 0x5544, 0x55A2, 0x553C, 0x5544, 0x5544, 0x55A8, 0x5790, 0xFC28, 0x44, 0x82}},
    {17939, 2, {0x810, 0x810, 0x101E, 0x7F22, 0x5524, 0x5540, 0x5510, 0x5566, 0x5542, 0x5542, 0x5566, 0x5542, 0x57C2, 0xFC42, 0x7E, 0x42}},
    {17940, 2, {0x820, 0x810, 0x107E, 0x7F42, 0x5510, 0x551C, 0x5524, 0x5554, 0x5508, 0x5536, 0x55C0, 0x553C, 0x57A4, 0xFC24, 0x3C, 0x24}},
    {17941, 2, {0x1200, 0x1100, 0x211C, 0x4FC0, 0x8000, 0x1000, 0x23BE, 0x6288, 0xA288, 0x2288, 0x2288, 0x22A8, 0x24C8, 0x2488, 0x2828, 0x2010}},
    {17942, 2, {0x1000, 0x1200, 0x221C, 0x43C0, 0x8500, 0x1100, 0x213E, 0x6FC8, 0xA108, 0x2108, 0x2548, 0x2548, 0x25C8, 0x2648, 0x2028, 0x2010}},
    {17943, 2, {0x1100, 0x1540, 0x254E, 0x47C0, 0x8000, 0x17C0, 0x221E, 0x6444, 0xA784, 0x2104, 0x2244, 0x2FE4, 0x2104, 0x2544, 0x2934, 0x2308}},
    {17944, 2, {0x2100, 0x2FE0, 0x410E, 0x47C0, 0x8100, 0x1FE0, 0x221E, 0x6444, 0xA784, 0x2104, 0x2244, 0x2FE4, 0x2104, 0x2544, 0x2934, 0x2308}},
    {17945, 2, {0x1100, 0x17C0, 0x224E, 0x4FE0, 0x8000, 0x17C0, 0x245E, 0x67C4, 0xA104, 0x23C4, 0x2504, 0x27E4, 0x2104, 0x27C4, 0x2554, 0x2548}},
    {17946, 2, {0x200, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {17947, 2, {0x2000, 0x11F0, 0x110, 0xFD10, 0x910, 0x1110, 0x1110, 0x3510, 0x5910, 0x9510, 0x1510, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {17948, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x9FC, 0x1124, 0x1124, 0x3524, 0x5924, 0x9524, 0x1524, 0x1134, 0x1128, 0x1020, 0x1020, 0x1020}},
    {17949, 2, {0x2050, 0x1048, 0x48, 0xFC40, 0x87C, 0x13C0, 0x1040, 0x3440, 0x5840, 0x9420, 0x1420, 0x1020, 0x1014, 0x1014, 0x100C, 0x1004}},
    {17950, 2, {0x2000, 0x13F8, 0x8, 0xFC08, 0x908, 0x1108, 0x1108, 0x35FE, 0x5802, 0x9402, 0x1402, 0x13FA, 0x1002, 0x1002, 0x1014, 0x1008}},
    {17951, 2, {0x2000, 0x11FC, 0x24, 0xFC24, 0x824, 0x11FC, 0x1120, 0x3520, 0x5920, 0x95FE, 0x1422, 0x1022, 0x102A, 0x1024, 0x1020, 0x1020}},
    {17952, 2, {0x2040, 0x1040, 0x40, 0xFC40, 0x9FE, 0x1040, 0x1040, 0x3484, 0x58A4, 0x94A4, 0x1528, 0x1128, 0x1210, 0x1428, 0x1044, 0x1182}},
    {17953, 2, {0x2080, 0x2088, 0x20B0, 0x3EC0, 0x2084, 0x2684, 0x387C, 0x2200, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {17954, 2, {0x2040, 0x1020, 0x20, 0xF9FE, 0x902, 0x1204, 0x1000, 0x3400, 0x59FE, 0x9420, 0x1420, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {17955, 2, {0x2090, 0x1090, 0x90, 0xF290, 0x1292, 0x22D4, 0x2298, 0x6A90, 0xB290, 0x2A90, 0x2290, 0x2292, 0x23D2, 0x2E12, 0x240E, 0x2000}},
    {17956, 2, {0x2040, 0x1050, 0x48, 0xF840, 0xBFC, 0x1040, 0x14E0, 0x38E0, 0x5550, 0x9150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040, 0x1040}},
    {17957, 2, {0x2000, 0x11FC, 0x124, 0xFD24, 0x924, 0x11FC, 0x1124, 0x3524, 0x5924, 0x95FC, 0x1524, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {17958, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x904, 0x11FC, 0x1100, 0x3540, 0x5944, 0x9548, 0x1570, 0x1140, 0x1142, 0x1242, 0x123E, 0x1400}},
    {17959, 2, {0x2020, 0x1020, 0x20, 0xFDFC, 0x924, 0x1124, 0x1124, 0x3524, 0x5924, 0x97FE, 0x1420, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {17960, 2, {0x2040, 0x1020, 0x20, 0xFBFE, 0x800, 0x1088, 0x1504, 0x3A02, 0x5488, 0x9088, 0x1050, 0x1050, 0x1020, 0x1050, 0x1088, 0x1306}},
    {17961, 2, {0x2080, 0x1080, 0x1FC, 0xFD04, 0xA04, 0x11E4, 0x1124, 0x3524, 0x59E4, 0x9524, 0x1524, 0x11E4, 0x1104, 0x1004, 0x1028, 0x1010}},
    {17962, 2, {0x2040, 0x1020, 0x20, 0xFBFE, 0x840, 0x1088, 0x1504, 0x3BFE, 0x5492, 0x9090, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {17963, 2, {0x800, 0x800, 0x7F7C, 0x1144, 0x2144, 0x1A44, 0xC7C, 0x3200, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {17964, 2, {0x2028, 0x1024, 0x24, 0xF820, 0xBFE, 0x1020, 0x1124, 0x3524, 0x5924, 0x97A8, 0x1528, 0x1110, 0x1212, 0x122A, 0x1446, 0x1082}},
    {17965, 2, {0x2000, 0x11FE, 0x20, 0xFC20, 0x848, 0x1084, 0x11FE, 0x3422, 0x5820, 0x9420, 0x15FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {17966, 2, {0x100, 0x100, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0x100, 0xFFFE, 0x1200, 0x1110, 0x10A0, 0x1040, 0x1430, 0x180E, 0x1000}},
    {17967, 2, {0x2020, 0x1124, 0xA4, 0xFCA8, 0x820, 0x11FC, 0x1104, 0x3504, 0x59FC, 0x9504, 0x1504, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {17968, 2, {0x2004, 0x100E, 0x3B8, 0xF088, 0x1088, 0x2128, 0x212E, 0x6BA8, 0xB0A8, 0x2AA8, 0x22A8, 0x213E, 0x2100, 0x2280, 0x247E, 0x2800}},
    {17969, 2, {0x100, 0xFFFE, 0x820, 0x1110, 0x2288, 0xC60, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {17970, 2, {0x2000, 0x11FC, 0x4, 0xF8FC, 0x804, 0x11FC, 0x1000, 0x33FE, 0x5602, 0x99F8, 0x1488, 0x1088, 0x1050, 0x1020, 0x10D8, 0x1306}},
    {17971, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x904, 0x1202, 0x10F8, 0x3420, 0x5820, 0x95FC, 0x1420, 0x1124, 0x10A4, 0x10A8, 0x13FE, 0x1000}},
    {17972, 2, {0x2208, 0x1228, 0x228, 0xF228, 0x1FA4, 0x2244, 0x2254, 0x6E92, 0xB710, 0x2AA0, 0x2A20, 0x3228, 0x2244, 0x22FC, 0x2244, 0x2200}},
    {17973, 2, {0x2000, 0x11FC, 0x4, 0xF804, 0x9FC, 0x1004, 0x1404, 0x3BFE, 0x5420, 0x9222, 0x1174, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {17974, 2, {0x2000, 0x11FC, 0x8, 0xF010, 0x1222, 0x22AA, 0x2272, 0x6A22, 0xB272, 0x2AAA, 0x2326, 0x22A2, 0x2242, 0x2202, 0x23FE, 0x2002}},
    {17975, 2, {0x2020, 0x1124, 0xA4, 0xFCA8, 0x820, 0x11FC, 0x1124, 0x3524, 0x5974, 0x95AC, 0x1524, 0x1124, 0x1124, 0x1104, 0x1114, 0x1108}},
    {17976, 2, {0x2080, 0x1040, 0x7FE, 0xF120, 0x1120, 0x223C, 0x2244, 0x6E64, 0xB294, 0x2B48, 0x2248, 0x2230, 0x2220, 0x2250, 0x2288, 0x2306}},
    {17977, 2, {0x2040, 0x1020, 0x3FE, 0xF800, 0x888, 0x1088, 0x1554, 0x3A22, 0x5400, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {17978, 2, {0x2040, 0x1020, 0x3FE, 0xF202, 0x1404, 0x2000, 0x23FE, 0x6820, 0xB020, 0x2920, 0x213C, 0x2120, 0x22A0, 0x2260, 0x243E, 0x2800}},
    {17979, 2, {0x2040, 0x1020, 0x3FE, 0xF880, 0x904, 0x13FE, 0x1002, 0x35FC, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {17980, 2, {0x2040, 0x1020, 0x3FE, 0xF202, 0x1504, 0x2100, 0x21DE, 0x6A52, 0xB252, 0x2B52, 0x249A, 0x2094, 0x2110, 0x2112, 0x2212, 0x240E}},
    {17981, 2, {0x2000, 0x1040, 0x39C, 0xFA04, 0xA04, 0x139C, 0x1204, 0x3604, 0x5BFC, 0x9490, 0x1490, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {17982, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0xC04, 0x11F8, 0x1108, 0x3508, 0x59F8, 0x9500, 0x1500, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {17983, 2, {0x2000, 0x13FE, 0x202, 0xF202, 0x13FE, 0x2210, 0x2292, 0x6A92, 0xB292, 0x2AFE, 0x2210, 0x2292, 0x2292, 0x2492, 0x24FE, 0x2802}},
    {17984, 2, {0x2000, 0x13DE, 0x252, 0xF252, 0x1252, 0x23DE, 0x2252, 0x6A52, 0xB252, 0x2BDE, 0x2252, 0x2252, 0x2252, 0x2252, 0x2552, 0x28A6}},
    {17985, 2, {0x2048, 0x1044, 0x5E, 0xFDE0, 0x828, 0x1012, 0x106A, 0x3596, 0x5848, 0x945E, 0x15E0, 0x1024, 0x1028, 0x1012, 0x106A, 0x1186}},
    {17986, 2, {0x2090, 0x1090, 0x108, 0xF148, 0x1244, 0x2492, 0x2108, 0x6BFC, 0xB044, 0x2820, 0x20A4, 0x2282, 0x228A, 0x228A, 0x2478, 0x2000}},
    {17987, 2, {0x2040, 0x1040, 0x3FE, 0xF880, 0x9FC, 0x1220, 0x15FE, 0x3000, 0x55FC, 0x9904, 0x15FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {17988, 2, {0x2008, 0x103C, 0x1E0, 0xF820, 0x820, 0x13FE, 0x1020, 0x34A0, 0x592C, 0x9524, 0x1524, 0x11AC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {17989, 2, {0x2000, 0x13FE, 0x50, 0xF850, 0xBFE, 0x1252, 0x1652, 0x3BFE, 0x5400, 0x9040, 0x17FE, 0x1088, 0x1190, 0x1060, 0x10D8, 0x1304}},
    {17990, 2, {0x2010, 0x13D4, 0x58, 0xFA52, 0x98C, 0x1088, 0x1504, 0x3AFA, 0x5420, 0x9020, 0x11FE, 0x1020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {17991, 2, {0x2000, 0x13FE, 0x20, 0xFC40, 0x9FC, 0x1154, 0x1154, 0x3554, 0x5954, 0x952C, 0x1420, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {17992, 2, {0x2100, 0x1178, 0x108, 0xF208, 0x12FE, 0x2640, 0x2A40, 0x6A7C, 0xB290, 0x2A10, 0x22FE, 0x2210, 0x2228, 0x2228, 0x2244, 0x2282}},
    {17993, 2, {0x2008, 0x101C, 0x1F0, 0xFD10, 0x910, 0x11FE, 0x1110, 0x357C, 0x5944, 0x957C, 0x1544, 0x117C, 0x1144, 0x1244, 0x127C, 0x1444}},
    {17994, 2, {0x20A8, 0x10A8, 0x3FE, 0xF8A8, 0x8A8, 0x1000, 0x13FE, 0x3622, 0x5820, 0x95FC, 0x1524, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {17995, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0xC04, 0x11FC, 0x1000, 0x35FC, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x1000, 0x13FE, 0x1000}},
    {17996, 2, {0x2020, 0x1020, 0x1FC, 0xF924, 0x9FC, 0x1020, 0x13FE, 0x3400, 0x59FC, 0x9504, 0x1524, 0x1124, 0x1124, 0x1050, 0x1088, 0x1304}},
    {17997, 2, {0x2000, 0x13FA, 0xC, 0xF84A, 0xB88, 0x112A, 0x1126, 0x3522, 0x5BF8, 0x952A, 0x152C, 0x112A, 0x1128, 0x122A, 0x1226, 0x1422}},
    {17998, 2, {0x201E, 0x13E0, 0x122, 0xFC94, 0x840, 0x1088, 0x11F0, 0x3420, 0x58C4, 0x95FE, 0x1422, 0x1020, 0x13FE, 0x1050, 0x1088, 0x1306}},
    {17999, 2, {0x2028, 0x1024, 0x3FE, 0xF820, 0x9FC, 0x1124, 0x11FC, 0x3524, 0x59FC, 0x9524, 0x1408, 0x13FE, 0x1088, 0x1048, 0x1048, 0x1018}},
    {18000, 2, {0x2000, 0x13FE, 0x0, 0xF9FC, 0x904, 0x1104, 0x15FC, 0x3800, 0x57FE, 0x928A, 0x1252, 0x13FE, 0x1222, 0x1222, 0x122A, 0x1204}},
    {18001, 2, {0x2080, 0x1040, 0x7FE, 0xF000, 0x13F8, 0x2208, 0x2FFE, 0x6A08, 0xB3F8, 0x28A4, 0x2128, 0x2310, 0x2510, 0x2948, 0x2186, 0x2100}},
    {18002, 2, {0x2040, 0x1020, 0x3FE, 0xF840, 0x888, 0x11F0, 0x1020, 0x3444, 0x59FE, 0x9402, 0x15FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1104}},
    {18003, 2, {0x2000, 0x13FE, 0x252, 0xFA52, 0xBFE, 0x1020, 0x1420, 0x39FC, 0x5420, 0x9020, 0x17FE, 0x1040, 0x1088, 0x1104, 0x13FE, 0x1102}},
    {18004, 2, {0x2000, 0x13FE, 0x200, 0xFA7C, 0xA44, 0x1244, 0x167C, 0x3A00, 0x56EE, 0x92AA, 0x12AA, 0x12AA, 0x12EE, 0x1200, 0x13FE, 0x1000}},
    {18005, 2, {0x2124, 0x1124, 0x224, 0xF4A4, 0x1154, 0x214A, 0x2392, 0x6D08, 0xB108, 0x2928, 0x212E, 0x2128, 0x2128, 0x2158, 0x214E, 0x2180}},
    {18006, 2, {0x2020, 0x1222, 0x222, 0xF3FE, 0x1000, 0x23DE, 0x2252, 0x6A52, 0xB3DE, 0x2A52, 0x2252, 0x23DE, 0x2252, 0x2252, 0x255A, 0x28A4}},
    {18007, 2, {0x2040, 0x1020, 0x3FE, 0xF888, 0x850, 0x13FE, 0x1622, 0x3AFA, 0x5622, 0x92FA, 0x128A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {18008, 2, {0x2014, 0x1012, 0x10, 0xF7FE, 0x1410, 0x2490, 0x24D2, 0x6C92, 0xB7F2, 0x2C94, 0x2494, 0x25C8, 0x2AAA, 0x289A, 0x31A6, 0x2042}},
    {18009, 2, {0x2000, 0x13FE, 0x202, 0xF3FE, 0x1210, 0x2254, 0x2238, 0x6A54, 0xB282, 0x2A50, 0x227C, 0x2290, 0x2210, 0x25FE, 0x2410, 0x2810}},
    {18010, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C, 0x200, 0xFFFE, 0x680, 0x1C48, 0xE530, 0x60E}},
    {18011, 2, {0x2040, 0x1020, 0x3FE, 0xF888, 0x924, 0x1242, 0x1090, 0x35F8, 0x5808, 0x9444, 0x14A8, 0x1190, 0x1288, 0x14A4, 0x10C2, 0x1080}},
    {18012, 2, {0x2020, 0x147C, 0x284, 0xF148, 0x1030, 0x20CE, 0x2610, 0x6A7C, 0xB210, 0x2A7C, 0x2210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x2800}},
    {18013, 2, {0x100, 0xFFFE, 0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0xB82, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {18014, 2, {0x2020, 0x17A4, 0xA8, 0xF292, 0x1114, 0x2208, 0x25F4, 0x6802, 0xB3F8, 0x2A08, 0x2208, 0x23F8, 0x2208, 0x2110, 0x2FFE, 0x2000}},
    {18015, 2, {0x2000, 0x17FC, 0x444, 0xF5F4, 0x1494, 0x27FC, 0x2404, 0x6DF4, 0xB514, 0x2DF4, 0x2444, 0x24F4, 0x2544, 0x25FC, 0x2444, 0x27FC}},
    {18016, 2, {0x2000, 0x17FE, 0x400, 0xF628, 0x1548, 0x27EE, 0x2492, 0x6C84, 0xB6A0, 0x2EA8, 0x27E8, 0x2488, 0x2494, 0x2514, 0x2924, 0x3242}},
    {18017, 2, {0x2020, 0x11FC, 0x124, 0xF9FC, 0x820, 0x13FE, 0x1000, 0x35FC, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {18018, 2, {0x2104, 0x1088, 0x0, 0xFBFE, 0xA22, 0x12AA, 0x1272, 0x3222, 0x57FE, 0x9800, 0x15FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {18019, 2, {0x2000, 0x11FC, 0x40, 0xFBFE, 0x888, 0x1104, 0x1652, 0x3050, 0x57DE, 0x9850, 0x15DC, 0x1050, 0x13DE, 0x1050, 0x1050, 0x1050}},
    {18020, 2, {0x2000, 0x127C, 0x144, 0xF174, 0x1054, 0x2054, 0x26FE, 0x6A82, 0xB2BA, 0x2AAA, 0x22BA, 0x2282, 0x228A, 0x2284, 0x2500, 0x28FE}},
    {18021, 2, {0x2000, 0x11FC, 0x20, 0xFBFE, 0xA22, 0x11AC, 0x1420, 0x39AC, 0x5450, 0x90C8, 0x1326, 0x10F8, 0x1008, 0x10D0, 0x1020, 0x1010}},
    {18022, 2, {0x2088, 0x13FE, 0x88, 0xF9FC, 0x954, 0x1154, 0x15FC, 0x3800, 0x57FE, 0x9242, 0x10F8, 0x1088, 0x1148, 0x1030, 0x1060, 0x1380}},
    {18023, 2, {0x201C, 0x11E0, 0x20, 0xFBFE, 0x820, 0x11FC, 0x11AC, 0x3574, 0x59FC, 0x9420, 0x15FC, 0x1020, 0x13FE, 0x1000, 0x1154, 0x122A}},
    {18024, 2, {0x21FC, 0x1124, 0x1FC, 0xF924, 0x9FC, 0x1020, 0x13FE, 0x34A8, 0x5924, 0x9642, 0x1420, 0x13FE, 0x1060, 0x1194, 0x1688, 0x10C6}},
    {18025, 2, {0x2088, 0x13FE, 0x88, 0xFBFE, 0xA02, 0x11FC, 0x1400, 0x3BFE, 0x5440, 0x90A2, 0x1354, 0x10B8, 0x1354, 0x1092, 0x1350, 0x1020}},
    {18026, 2, {0x2288, 0x2108, 0xFABE, 0x2008, 0x729C, 0xA92A, 0x2288, 0xFFFE, 0xA20, 0x3118, 0xCFE6, 0x288, 0x450, 0x1C20, 0x6518, 0x606}},
    {18027, 2, {0x2094, 0x1398, 0x92, 0xFD8E, 0xA80, 0x11FC, 0x1124, 0x35FC, 0x5924, 0x95FC, 0x1488, 0x11FC, 0x1088, 0x13FE, 0x1088, 0x1104}},
    {18028, 2, {0x23DE, 0x1252, 0x3DE, 0xFA52, 0xBDE, 0x1252, 0x17DE, 0x398C, 0x5652, 0x9040, 0x17FE, 0x1088, 0x1190, 0x1060, 0x10D8, 0x1304}},
    {18029, 2, {0x2040, 0x17FE, 0x492, 0xF3FC, 0x1090, 0x23FC, 0x2090, 0x6FFE, 0xB108, 0x2BFC, 0x250A, 0x21F8, 0x2240, 0x227C, 0x2540, 0x28FE}},
    {18030, 2, {0x2088, 0x13FE, 0x88, 0xFBDE, 0xA52, 0x13DE, 0x14A0, 0x3890, 0x55FE, 0x9120, 0x13FC, 0x1520, 0x11FC, 0x1120, 0x11FE, 0x1100}},
    {18031, 2, {0x2014, 0x17FE, 0x490, 0xF4D4, 0x1494, 0x27FA, 0x25AA, 0x6A96, 0xB1A2, 0x2BF8, 0x2208, 0x23F8, 0x2040, 0x227C, 0x2540, 0x28FE}},
    {18032, 2, {0x220C, 0x110A, 0x7C8, 0xF29E, 0x1554, 0x26F4, 0x245E, 0x6FD4, 0xB114, 0x2FDE, 0x2554, 0x2554, 0x25D4, 0x245E, 0x2550, 0x2490}},
    {18033, 2, {0x23FE, 0x1202, 0x3FE, 0xFA92, 0xA54, 0x1292, 0x16FE, 0x3AAA, 0x56FE, 0x9322, 0x12FA, 0x12AA, 0x12FA, 0x1222, 0x15FA, 0x1004}},
    {18034, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {18035, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {18036, 2, {0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xC826, 0xFE0, 0x808, 0x808, 0x7F8}},
    {18037, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {18038, 2, {0x10, 0xFE10, 0x2810, 0x29FE, 0x2922, 0xFE20, 0xAA50, 0xAA50, 0xAAD2, 0xAACC, 0xAF48, 0xC244, 0x8244, 0x8252, 0xFE60, 0x8240}},
    {18039, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {18040, 2, {0x8, 0x7E08, 0x4208, 0x42FE, 0x7E08, 0x4208, 0x4288, 0x7E48, 0x4248, 0x4208, 0x7E08, 0x2428, 0x2412, 0x2402, 0x43FE, 0x8000}},
    {18041, 2, {0x0, 0x7EFC, 0x2484, 0x2484, 0x24FC, 0x2484, 0x2484, 0xFFFC, 0x2484, 0x2484, 0x24FC, 0x2450, 0x2450, 0x2492, 0x4492, 0x850E}},
    {18042, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x55FC, 0x5904, 0x5104, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x2850, 0x2450, 0x4492, 0x4112, 0x820E}},
    {18043, 2, {0x800, 0x87C, 0xB44, 0x7C44, 0x487C, 0x4844, 0x4844, 0x7F7C, 0x944, 0x1944, 0x297C, 0x4928, 0x8B28, 0x84A, 0x84A, 0x886}},
    {18044, 2, {0x0, 0x7EFC, 0x1284, 0x1284, 0x22FC, 0x2A84, 0x4484, 0x80FC, 0x3E84, 0x2284, 0x22FC, 0x2250, 0x2250, 0x3E92, 0x2292, 0x10E}},
    {18045, 2, {0x100, 0x100, 0xFFFE, 0x280, 0x2448, 0x1830, 0x3FF8, 0xD016, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {18046, 2, {0x2000, 0x10FC, 0x1084, 0xFE84, 0x82FC, 0x484, 0x4084, 0x40FC, 0x7C84, 0x4084, 0x40FC, 0x4050, 0x4C50, 0x7092, 0x4092, 0x10E}},
    {18047, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x10FC, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x3884, 0x34FC, 0x5250, 0x9050, 0x1092, 0x1092, 0x110E}},
    {18048, 2, {0x800, 0x107C, 0x3E44, 0x2244, 0x327C, 0x2A44, 0x2A44, 0xFE7C, 0x2244, 0x3244, 0x2A7C, 0x2A28, 0x2228, 0x424A, 0x4A4A, 0x8486}},
    {18049, 2, {0x800, 0x87C, 0x1044, 0x7F44, 0x417C, 0x4144, 0x5D44, 0x557C, 0x5544, 0x5544, 0x557C, 0x5D28, 0x4128, 0x414A, 0x454A, 0x4286}},
    {18050, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x10FC, 0x1084, 0xFE84, 0x28FC, 0x2884, 0x6C84, 0xAAFC, 0x2850, 0x2850, 0x4892, 0x4892, 0x990E}},
    {18051, 2, {0x800, 0x87C, 0xF44, 0x844, 0x87C, 0x7E44, 0x4244, 0x427C, 0x7A44, 0x4A44, 0x4A7C, 0x7A28, 0x4228, 0x424A, 0x7E4A, 0x4286}},
    {18052, 2, {0x0, 0x7E7C, 0x244, 0x3E44, 0x27C, 0xFFC4, 0x844, 0x97C, 0x4944, 0x2A44, 0x1C7C, 0x2A28, 0xC928, 0x94A, 0x284A, 0x1086}},
    {18053, 2, {0x800, 0x87C, 0x844, 0xFF44, 0x87C, 0x2A44, 0x2A44, 0x2A7C, 0x5D44, 0x8844, 0x1C7C, 0x2A28, 0x4928, 0x884A, 0x84A, 0x886}},
    {18054, 2, {0x2200, 0x227C, 0x7F44, 0x2244, 0x2A7C, 0x844, 0x3E44, 0x2A7C, 0x2A44, 0x2A44, 0xFF7C, 0x828, 0x1428, 0x124A, 0x224A, 0x4086}},
    {18055, 2, {0x1000, 0x107C, 0x5244, 0x5144, 0x957C, 0x844, 0x3044, 0xC07C, 0x7E44, 0x4244, 0x7E7C, 0x4228, 0x7E28, 0x424A, 0x7E4A, 0x4286}},
    {18056, 2, {0x0, 0xFF7C, 0x1444, 0x1444, 0x7F7C, 0x5544, 0x5544, 0x7F7C, 0x844, 0x844, 0x7F7C, 0x828, 0x828, 0xF4A, 0xF04A, 0x4086}},
    {18057, 2, {0x2000, 0x2F7C, 0xF544, 0x2544, 0x757C, 0x2544, 0xF544, 0x297C, 0x2344, 0x844, 0x87C, 0xFFA8, 0x1428, 0x224A, 0x414A, 0x8086}},
    {18058, 2, {0x2200, 0x147C, 0x44, 0x7F44, 0x7C, 0x1444, 0x2244, 0x417C, 0x44, 0x7F44, 0x557C, 0x5528, 0x5528, 0x57CA, 0xFC4A, 0x86}},
    {18059, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x44, 0x7F44, 0x417C, 0x7F44, 0x4144, 0x7F7C, 0x4128, 0x7F28, 0x144A, 0x224A, 0x4286}},
    {18060, 2, {0x800, 0x107C, 0x7F44, 0x4144, 0x7F7C, 0x4144, 0x7F44, 0x107C, 0xFFC4, 0x2244, 0x497C, 0x88A8, 0x7F68, 0x84A, 0x84A, 0x886}},
    {18061, 2, {0x800, 0xFF7C, 0x844, 0x7E44, 0x87C, 0xFF44, 0x44, 0x7E7C, 0x4244, 0x7E44, 0x427C, 0x7E28, 0x4228, 0x7E4A, 0x244A, 0x4286}},
    {18062, 2, {0x800, 0x87C, 0x7F44, 0x1044, 0x3E7C, 0x1044, 0xFFC4, 0x227C, 0x4144, 0x90C4, 0x267C, 0x2228, 0x3628, 0x224A, 0x3E4A, 0x2286}},
    {18063, 2, {0x800, 0x4A7C, 0x2C44, 0x844, 0xFF7C, 0x2A44, 0x4944, 0x887C, 0x4244, 0x7744, 0x927C, 0x5A28, 0x2F28, 0x224A, 0x424A, 0x8286}},
    {18064, 2, {0x4000, 0x777C, 0xA544, 0x2544, 0xFD7C, 0x2744, 0x5544, 0x807C, 0x7E44, 0x4244, 0x427C, 0x7E28, 0x4228, 0x424A, 0x7E4A, 0x4286}},
    {18065, 2, {0x0, 0xFEFC, 0xAA84, 0xAA84, 0xFEFC, 0x84, 0x7C84, 0x44FC, 0x7C84, 0x4484, 0x7CFC, 0x4450, 0x7C50, 0x92, 0x2892, 0x450E}},
    {18066, 2, {0x1000, 0x87C, 0x7F44, 0x2244, 0xFF7C, 0x44, 0x7F44, 0x497C, 0x7F44, 0x4944, 0x7F7C, 0x828, 0x7F28, 0x84A, 0xF4A, 0xF086}},
    {18067, 2, {0x0, 0xFEFC, 0x2884, 0xFE84, 0xAAFC, 0xFE84, 0x84, 0x7CFC, 0x4484, 0x7C84, 0x44FC, 0x7C50, 0x1050, 0xFE92, 0x1092, 0x110E}},
    {18068, 2, {0x2000, 0x10FC, 0xFE84, 0x8284, 0x7CFC, 0x5484, 0x9884, 0x7CFC, 0xC484, 0x7C84, 0x44FC, 0x7C50, 0x4450, 0x7C92, 0x2892, 0x450E}},
    {18069, 2, {0x7700, 0x7C, 0x7744, 0x5544, 0x777C, 0x5544, 0x844, 0x7F7C, 0x5444, 0x7F44, 0x557C, 0x7F28, 0x5528, 0x5E4A, 0x554A, 0x9B86}},
    {18070, 2, {0x0, 0xFEFC, 0x1084, 0xFE84, 0x92FC, 0x5484, 0x84, 0xEEFC, 0xAA84, 0xEE84, 0xFC, 0xFE50, 0x5450, 0xBA92, 0x1092, 0xFF0E}},
    {18071, 2, {0x2008, 0x2048, 0x7848, 0x4848, 0x9044, 0x7C84, 0x54A4, 0x5522, 0x7C20, 0x5440, 0x5440, 0x7C48, 0x5484, 0x55FE, 0x4482, 0x8C00}},
    {18072, 2, {0x1200, 0x1380, 0x249C, 0x4900, 0x87C0, 0x1540, 0x257E, 0x67C8, 0xA548, 0x2548, 0x27C8, 0x2548, 0x2548, 0x2548, 0x2468, 0x28D0}},
    {18073, 2, {0x1000, 0x10FE, 0x3C00, 0x2400, 0x487C, 0xBE44, 0x2A44, 0x2A7C, 0x3E44, 0x2A44, 0x2A7C, 0x3E44, 0x2A00, 0x4A00, 0x42FE, 0x8600}},
    {18074, 2, {0x1010, 0x1050, 0x3C50, 0x247C, 0x4890, 0xBE10, 0x2A10, 0x2AFE, 0x3E28, 0x2A28, 0x2A28, 0x3E28, 0x2A4A, 0x4A4A, 0x4286, 0x8700}},
    {18075, 2, {0x1010, 0x1050, 0x3C50, 0x247C, 0x4850, 0xBE90, 0x2A10, 0x2AFE, 0x3E00, 0x2A00, 0x2A7C, 0x3E44, 0x2A44, 0x4A44, 0x427C, 0x8644}},
    {18076, 2, {0x1020, 0x1040, 0x3CFE, 0x2492, 0x4892, 0xBEFE, 0x2A92, 0x2AA2, 0x3EFE, 0x2A48, 0x2A88, 0x3FFE, 0x2A08, 0x4A08, 0x4208, 0x8608}},
    {18077, 2, {0x2040, 0x2020, 0x79FE, 0x4800, 0x9088, 0x7C88, 0x5554, 0x5622, 0x7C00, 0x5420, 0x55FE, 0x7C20, 0x5420, 0x5420, 0x4420, 0x8C20}},
    {18078, 2, {0x2020, 0x2020, 0x7820, 0x4BFE, 0x9020, 0x7D24, 0x5524, 0x5524, 0x7EAA, 0x5470, 0x54A8, 0x7CA8, 0x5524, 0x5622, 0x4420, 0x8C20}},
    {18079, 2, {0x1000, 0x10FC, 0x3C84, 0x24FC, 0x4884, 0xBEFC, 0x2A00, 0x2AFE, 0x3E08, 0x2BFE, 0x2A08, 0x3E88, 0x2A48, 0x4A08, 0x4228, 0x8610}},
    {18080, 2, {0x2000, 0x21FC, 0x7904, 0x4904, 0x91FC, 0x7D04, 0x5504, 0x55FC, 0x7C00, 0x5512, 0x55D4, 0x7D18, 0x5510, 0x5552, 0x4592, 0x8D0E}},
    {18081, 2, {0x2000, 0x21FC, 0x7924, 0x4924, 0x91FC, 0x7D24, 0x5524, 0x55FC, 0x7C20, 0x5410, 0x5454, 0x7D42, 0x5542, 0x554A, 0x4638, 0x8C00}},
    {18082, 2, {0x2048, 0x2148, 0x7948, 0x4BFE, 0x9148, 0x7D48, 0x5578, 0x5500, 0x7DFE, 0x5420, 0x57FE, 0x7C70, 0x54A8, 0x5524, 0x4622, 0x8C20}},
    {18083, 2, {0x1000, 0x10FC, 0x3C84, 0x2484, 0x48FC, 0xBE84, 0x2A84, 0x2AFC, 0x3E00, 0x2BFE, 0x2A20, 0x3EA0, 0x2ABC, 0x4AA0, 0x4360, 0x863E}},
    {18084, 2, {0x1000, 0x107C, 0x3C54, 0x2454, 0x487C, 0xBE54, 0x2A54, 0x2A7C, 0x3E10, 0x2AFE, 0x2A92, 0x3E96, 0x2ABE, 0x4A82, 0x428A, 0x8684}},
    {18085, 2, {0x1000, 0x10EE, 0x3C22, 0x2422, 0x48EE, 0xBE88, 0x2A88, 0x2AEE, 0x3E22, 0x2AAA, 0x2A66, 0x3E66, 0x2AAA, 0x4A22, 0x42AA, 0x8644}},
    {18086, 2, {0x2008, 0x203C, 0x79D0, 0x491C, 0x9110, 0x7D7C, 0x5554, 0x5558, 0x7D70, 0x5554, 0x554C, 0x7D40, 0x5678, 0x54A8, 0x442A, 0x8C46}},
    {18087, 2, {0x2000, 0x21DC, 0x7844, 0x4954, 0x90CC, 0x7D54, 0x5428, 0x5444, 0x7D92, 0x5420, 0x54C8, 0x7C32, 0x54C4, 0x5418, 0x4460, 0x8D80}},
    {18088, 2, {0x2012, 0x21D4, 0x7848, 0x4A8A, 0x9104, 0x7E02, 0x55DC, 0x5454, 0x7C62, 0x55C0, 0x551E, 0x7DC2, 0x5454, 0x5448, 0x4554, 0x8CA2}},
    {18089, 2, {0x2018, 0x21E0, 0x7840, 0x4BFE, 0x9088, 0x7D74, 0x5652, 0x5470, 0x7C00, 0x55FC, 0x5504, 0x7D74, 0x5554, 0x5574, 0x4504, 0x8D0C}},
    {18090, 2, {0x2000, 0x21FC, 0x7954, 0x4954, 0x91FC, 0x7C00, 0x57FE, 0x5400, 0x7DFC, 0x5504, 0x55FC, 0x7C62, 0x5494, 0x5588, 0x46A4, 0x8CC2}},
    {18091, 2, {0x1020, 0x7E20, 0x4A7E, 0x7EC4, 0x1028, 0xFF10, 0x2028, 0x3EC6, 0x5200, 0x1FE0, 0x2040, 0x5FF0, 0x9110, 0x1FF0, 0x1110, 0x2130}},
    {18092, 2, {0x2020, 0x2124, 0x78A8, 0x4BFE, 0x90A8, 0x7D24, 0x5622, 0x5508, 0x7DC8, 0x565E, 0x5548, 0x7CA8, 0x54BE, 0x5508, 0x4508, 0x8E08}},
    {18093, 2, {0x2088, 0x23FE, 0x7888, 0x4800, 0x91DC, 0x7D54, 0x55DC, 0x5488, 0x7DFE, 0x5690, 0x54FC, 0x7C90, 0x54FC, 0x5490, 0x44FE, 0x8C80}},
    {18094, 2, {0x2000, 0x1000, 0x11FC, 0xFE44, 0x48, 0x7C48, 0x50, 0x7C5E, 0x42, 0x7C42, 0x4482, 0x4482, 0x4482, 0x7D02, 0x4514, 0x208}},
    {18095, 2, {0x1004, 0x804, 0x804, 0xFFA4, 0x24, 0x7F24, 0x24, 0x7F24, 0x24, 0x7F24, 0x4124, 0x4124, 0x4124, 0x7F04, 0x4114, 0x8}},
    {18096, 2, {0x2080, 0x1080, 0x1080, 0xFE84, 0x88, 0x7C90, 0xA0, 0x7CC0, 0x80, 0x7C80, 0x4480, 0x4484, 0x4484, 0x7C84, 0x447C, 0x0}},
    {18097, 2, {0x2020, 0x1020, 0x1020, 0xFE20, 0x20, 0x7CA8, 0xA4, 0x7CA4, 0x122, 0x7D22, 0x4622, 0x4420, 0x4420, 0x7C20, 0x44A0, 0x40}},
    {18098, 2, {0x2008, 0x1008, 0x1010, 0xFE20, 0x40, 0x7C88, 0x8, 0x7C10, 0x20, 0x7C44, 0x4484, 0x4408, 0x4410, 0x7C20, 0x44C0, 0x300}},
    {18099, 2, {0x2020, 0x1020, 0x20, 0xFDFE, 0x20, 0x7820, 0x20, 0x79FC, 0x84, 0x7888, 0x4848, 0x4850, 0x4820, 0x7850, 0x4988, 0x606}},
    {18100, 2, {0x2000, 0x10FC, 0x1084, 0xFE84, 0x84, 0x7CFC, 0x84, 0x7C84, 0x84, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7C84, 0x4514, 0x208}},
    {18101, 2, {0x2008, 0x1088, 0x1048, 0xFE48, 0x8, 0x7C88, 0x48, 0x7C48, 0x8, 0x7C0E, 0x45F8, 0x4408, 0x4408, 0x7C08, 0x4408, 0x8}},
    {18102, 2, {0x2000, 0x1000, 0x11FC, 0xFE20, 0x20, 0x7C20, 0x20, 0x7DFE, 0x20, 0x7C20, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {18103, 2, {0x2090, 0x1090, 0x108, 0xFD08, 0x204, 0x7C02, 0x3FC, 0x7880, 0x100, 0x79F8, 0x4808, 0x4808, 0x4808, 0x7808, 0x4850, 0x20}},
    {18104, 2, {0x2020, 0x1020, 0x1050, 0xFE50, 0x88, 0x7D04, 0x202, 0x7C88, 0x88, 0x7C88, 0x4488, 0x4488, 0x4488, 0x7D08, 0x4508, 0x208}},
    {18105, 2, {0x2010, 0x1110, 0x110, 0xFD12, 0x112, 0x7914, 0x1D8, 0x7910, 0x110, 0x7910, 0x4910, 0x4912, 0x4952, 0x7992, 0x490E, 0x0}},
    {18106, 2, {0x2040, 0x1040, 0x1040, 0xFE80, 0xFE, 0x7D08, 0x288, 0x7C88, 0x88, 0x7C50, 0x4450, 0x4420, 0x4450, 0x7C88, 0x4504, 0x202}},
    {18107, 2, {0x200, 0x7F0, 0x1820, 0x6240, 0x980, 0x600, 0x1900, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {18108, 2, {0x2000, 0x11F8, 0x1048, 0xFE48, 0x48, 0x7C48, 0x48, 0x7DF8, 0x88, 0x7C88, 0x4488, 0x4488, 0x4488, 0x7C88, 0x47FE, 0x0}},
    {18109, 2, {0x2008, 0x103C, 0x11E0, 0xFE20, 0x20, 0x7C3C, 0x1E0, 0x7C20, 0x20, 0x7C3E, 0x47E0, 0x4420, 0x4422, 0x7C22, 0x4422, 0x1E}},
    {18110, 2, {0x2008, 0x103C, 0x11E0, 0xFE20, 0x20, 0x7C20, 0x20, 0x7DFE, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x45FC, 0x0}},
    {18111, 2, {0x2040, 0x1040, 0x1040, 0xFE7C, 0x84, 0x7C88, 0x120, 0x7C20, 0x20, 0x7C50, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {18112, 2, {0x2008, 0x101C, 0x1E0, 0xFD00, 0x100, 0x79FC, 0x144, 0x7944, 0x144, 0x7928, 0x4928, 0x4910, 0x4910, 0x7A28, 0x4A44, 0x482}},
    {18113, 2, {0x2000, 0x10FC, 0x1084, 0xFE84, 0x84, 0x7DFE, 0x84, 0x7C84, 0x84, 0x7DFE, 0x4484, 0x4484, 0x4484, 0x7C84, 0x4494, 0x88}},
    {18114, 2, {0x2008, 0x1068, 0x388, 0xFC88, 0x88, 0x7888, 0x88, 0x7BFE, 0x88, 0x7888, 0x4888, 0x4888, 0x4908, 0x7908, 0x4A08, 0x408}},
    {18115, 2, {0x2000, 0x1000, 0x11FC, 0xFC00, 0x0, 0x7C00, 0x3FE, 0x7C90, 0x90, 0x7C90, 0x4490, 0x4490, 0x4492, 0x7D12, 0x4512, 0x20E}},
    {18116, 2, {0x4200, 0x2200, 0x23C, 0xFBA4, 0x2A4, 0x72A4, 0x4A4, 0x74A4, 0xAB4, 0x7128, 0x5120, 0x5122, 0x5222, 0x7222, 0x541E, 0x800}},
    {18117, 2, {0x2000, 0x11FC, 0x1124, 0xFD24, 0x124, 0x7DFC, 0x124, 0x7D24, 0x124, 0x7DFC, 0x4524, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {18118, 2, {0x2020, 0x1020, 0x1020, 0xFE20, 0x1FC, 0x7D24, 0x124, 0x7D24, 0x124, 0x7DFC, 0x4524, 0x4524, 0x4524, 0x7D24, 0x45FC, 0x104}},
    {18119, 2, {0x2000, 0x10F8, 0x1088, 0xFE88, 0x88, 0x7C88, 0x106, 0x7E00, 0x0, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x104}},
    {18120, 2, {0x2040, 0x1040, 0x1080, 0xFCFC, 0x120, 0x7E20, 0x20, 0x7C20, 0x1FE, 0x7C20, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {18121, 2, {0x2000, 0x11FC, 0x1104, 0xFF04, 0x104, 0x7DFC, 0x120, 0x7D20, 0x1FE, 0x7D20, 0x4520, 0x4510, 0x4512, 0x7D4A, 0x4586, 0x102}},
    {18122, 2, {0x2000, 0x1000, 0x10FC, 0xFE84, 0x84, 0x7C84, 0x84, 0x7C84, 0xFC, 0x7C84, 0x4400, 0x4448, 0x4444, 0x7C84, 0x4482, 0x102}},
    {18123, 2, {0x2028, 0x1024, 0x24, 0xFC20, 0x1FE, 0x7920, 0x124, 0x7924, 0x124, 0x7928, 0x4928, 0x4918, 0x4912, 0x7A1A, 0x4A26, 0x442}},
    {18124, 2, {0x2040, 0x1040, 0x10FC, 0xFE84, 0x104, 0x7EF4, 0x94, 0x7C94, 0xF4, 0x7C84, 0x4494, 0x4488, 0x4482, 0x7C82, 0x447E, 0x0}},
    {18125, 2, {0x2090, 0x1090, 0x90, 0xFBFC, 0x94, 0x7894, 0x3FC, 0x7A90, 0x290, 0x7BFE, 0x4892, 0x4892, 0x491A, 0x7914, 0x4A10, 0x410}},
    {18126, 2, {0x2040, 0x1020, 0x20, 0xFBFE, 0x202, 0x7C04, 0x0, 0x7890, 0x90, 0x7890, 0x4888, 0x4908, 0x4908, 0x7904, 0x4A04, 0x402}},
    {18127, 2, {0x2000, 0x10FC, 0x1084, 0xFE84, 0x84, 0x7CFC, 0x80, 0x7CA0, 0xA2, 0x7CA4, 0x44B8, 0x44A0, 0x44A2, 0x7D22, 0x451E, 0x200}},
    {18128, 2, {0x0, 0x3FF8, 0x8, 0x7FC8, 0x8, 0x3F88, 0x2088, 0x3FA8, 0x110, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {18129, 2, {0x2000, 0x1040, 0x20, 0xFC28, 0x8, 0x7888, 0x90, 0x7A94, 0x2A2, 0x7AA2, 0x4CC2, 0x4888, 0x4988, 0x7A88, 0x4478, 0x0}},
    {18130, 2, {0x0, 0x7DFE, 0x1080, 0x10FC, 0x1E04, 0xF028, 0x110, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {18131, 2, {0x480, 0x2480, 0x24FC, 0x24A0, 0x2510, 0x400, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {18132, 2, {0x2000, 0x13FE, 0x1020, 0xFE20, 0x40, 0x7DFC, 0x154, 0x7D54, 0x154, 0x7D54, 0x4554, 0x4554, 0x4554, 0x7D44, 0x4514, 0x108}},
    {18133, 2, {0x2020, 0x1020, 0x3FE, 0xFE22, 0x222, 0x7BFE, 0x222, 0x7A22, 0x3FE, 0x7820, 0x4920, 0x48A0, 0x4840, 0x78A0, 0x4918, 0x606}},
    {18134, 2, {0x2020, 0x1020, 0x1FC, 0xFD24, 0x124, 0x79FC, 0x124, 0x7924, 0x1FC, 0x7820, 0x4824, 0x4818, 0x4832, 0x78CA, 0x4B06, 0x2}},
    {18135, 2, {0x40, 0x7E20, 0x49FE, 0x4800, 0x4800, 0x7EFC, 0x4200, 0x4200, 0x42FC, 0x7E00, 0x4800, 0x48FC, 0x4884, 0x4884, 0x7EFC, 0x84}},
    {18136, 2, {0x2080, 0x1088, 0x9C, 0xFD70, 0x110, 0x7B10, 0x510, 0x7910, 0x1FE, 0x7910, 0x4910, 0x4910, 0x4910, 0x7910, 0x497C, 0x100}},
    {18137, 2, {0x2090, 0x1090, 0x90, 0xFD10, 0x1FE, 0x7B10, 0x510, 0x7938, 0x138, 0x7954, 0x4954, 0x4992, 0x4910, 0x7910, 0x4910, 0x110}},
    {18138, 2, {0x800, 0x7EF8, 0x848, 0x3E48, 0x888, 0x7EA8, 0x910, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {18139, 2, {0x2000, 0x1000, 0x11FC, 0xFF04, 0x104, 0x7D74, 0x154, 0x7D54, 0x154, 0x7D54, 0x4574, 0x4504, 0x4504, 0x7DFC, 0x4504, 0x0}},
    {18140, 2, {0x2020, 0x1020, 0x40, 0xFDFC, 0x104, 0x7904, 0x194, 0x7954, 0x124, 0x7924, 0x4954, 0x4994, 0x4904, 0x7904, 0x49FC, 0x104}},
    {18141, 2, {0x2000, 0x11FC, 0x24, 0xFC24, 0x1FC, 0x7844, 0x44, 0x7BFE, 0x0, 0x7820, 0x4820, 0x4924, 0x4924, 0x7924, 0x49FC, 0x4}},
    {18142, 2, {0x2000, 0x11FC, 0x1104, 0xFDFC, 0x104, 0x7DFC, 0x0, 0x7C00, 0x1FC, 0x7C20, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x4420, 0x20}},
    {18143, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x20, 0x7820, 0x124, 0x7924, 0x124, 0x7AAA, 0x4820, 0x4850, 0x4850, 0x7888, 0x4904, 0x202}},
    {18144, 2, {0x2000, 0x11FE, 0x1000, 0xFE00, 0xFC, 0x7C84, 0x84, 0x7C84, 0xFC, 0x7C00, 0x4484, 0x4444, 0x4448, 0x7C00, 0x45FE, 0x0}},
    {18145, 2, {0x201C, 0x13E0, 0x220, 0xFE20, 0x3FE, 0x7A20, 0x290, 0x7B0A, 0x206, 0x79FA, 0x4908, 0x4908, 0x4908, 0x7908, 0x49F8, 0x108}},
    {18146, 2, {0x201C, 0x11E0, 0x20, 0xFC20, 0x3FE, 0x78A8, 0x124, 0x7A22, 0x0, 0x78F8, 0x4888, 0x4888, 0x4888, 0x788A, 0x490A, 0x206}},
    {18147, 2, {0x2020, 0x1020, 0x1FC, 0xFD24, 0x1FC, 0x7924, 0x1FC, 0x7800, 0x3FE, 0x7880, 0x4900, 0x49FC, 0x4804, 0x7804, 0x4828, 0x10}},
    {18148, 2, {0x2000, 0x10FC, 0x1084, 0xFE84, 0x84, 0x7C84, 0xFC, 0x7C10, 0x10, 0x7C90, 0x449E, 0x4490, 0x4490, 0x7CD0, 0x453E, 0x200}},
    {18149, 2, {0x2184, 0x1068, 0x30, 0xFCC8, 0x304, 0x7840, 0x3FE, 0x78A0, 0x120, 0x7BFC, 0x4D24, 0x4924, 0x4934, 0x7928, 0x4820, 0x20}},
    {18150, 2, {0x2000, 0x1050, 0x1048, 0xFE84, 0x124, 0x7C20, 0x50, 0x7C88, 0x106, 0x7EF8, 0x4488, 0x4488, 0x4488, 0x7C88, 0x44F8, 0x88}},
    {18151, 2, {0x2010, 0x1110, 0x1110, 0xFD92, 0x154, 0x7D10, 0x1FE, 0x7D10, 0x138, 0x7D54, 0x4592, 0x4510, 0x4510, 0x7D10, 0x45FE, 0x0}},
    {18152, 2, {0x2040, 0x1020, 0x11FC, 0xFE00, 0x88, 0x7C50, 0x3FE, 0x7C20, 0x20, 0x7C20, 0x45FC, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {18153, 2, {0x2000, 0x11FE, 0x1048, 0xFE48, 0x48, 0x7DCE, 0x102, 0x7D02, 0x102, 0x7DCE, 0x4448, 0x4448, 0x4448, 0x7C48, 0x45FE, 0x0}},
    {18154, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0x88, 0x7904, 0x202, 0x79FC, 0x4, 0x79FC, 0x4900, 0x49FC, 0x4804, 0x7804, 0x4828, 0x10}},
    {18155, 2, {0x2000, 0x10FC, 0x80, 0xFCF8, 0x80, 0x78F8, 0x80, 0x7BFE, 0x140, 0x7924, 0x4928, 0x4910, 0x4908, 0x7944, 0x4982, 0x100}},
    {18156, 2, {0x2080, 0x1080, 0x1FC, 0xFD04, 0x284, 0x78F4, 0x144, 0x7844, 0x3FC, 0x7844, 0x4954, 0x4954, 0x49F4, 0x7804, 0x4828, 0x10}},
    {18157, 2, {0x2020, 0x1020, 0x1FC, 0xFC20, 0x20, 0x7BFE, 0x88, 0x7850, 0x1FC, 0x7820, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20}},
    {18158, 2, {0x2088, 0x1088, 0x10E8, 0xFD2C, 0x12A, 0x7EA8, 0x48, 0x7CA0, 0x11E, 0x7E00, 0x44FC, 0x4484, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {18159, 2, {0x2000, 0x11FE, 0x102, 0xFD02, 0x1FE, 0x7910, 0x110, 0x79FE, 0x110, 0x7910, 0x497E, 0x4942, 0x4942, 0x7A42, 0x4A7E, 0x442}},
    {18160, 2, {0x2000, 0x11FC, 0x1104, 0xFF04, 0x1FC, 0x7D04, 0x104, 0x7DFC, 0x0, 0x7D12, 0x45D4, 0x4518, 0x4510, 0x7D52, 0x4592, 0x10E}},
    {18161, 2, {0x2020, 0x1020, 0x1FC, 0xFD24, 0x124, 0x7924, 0x1FC, 0x7820, 0x40, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x7A8A, 0x4478, 0x0}},
    {18162, 2, {0x2020, 0x1040, 0x1FC, 0xFD04, 0x1FC, 0x7904, 0x1FC, 0x7820, 0x20, 0x7BFE, 0x4A22, 0x4A22, 0x4A2A, 0x7A24, 0x4820, 0x20}},
    {18163, 2, {0x2020, 0x1020, 0x3FE, 0xFC50, 0x88, 0x7924, 0x222, 0x79FC, 0x124, 0x79FC, 0x4924, 0x49FC, 0x4822, 0x7822, 0x481E, 0x0}},
    {18164, 2, {0x2020, 0x1020, 0x1FC, 0xFC50, 0x88, 0x7904, 0x3FE, 0x7808, 0x1E8, 0x7928, 0x4928, 0x49E8, 0x4928, 0x7808, 0x4828, 0x10}},
    {18165, 2, {0x201C, 0x13E0, 0x44, 0xF924, 0xA8, 0x7890, 0x3FE, 0x7A02, 0x0, 0x79FC, 0x4884, 0x4888, 0x4850, 0x7820, 0x48D8, 0x306}},
    {18166, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7820, 0x3FE, 0x7820, 0x0, 0x7820, 0x4BFE, 0x4870, 0x48A8, 0x7924, 0x4A22, 0x20}},
    {18167, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0x504, 0x7900, 0x1DE, 0x7A52, 0x252, 0x7B52, 0x4C9A, 0x4894, 0x4910, 0x7912, 0x4A12, 0x40E}},
    {18168, 2, {0x2020, 0x1010, 0x1FE, 0xFD00, 0x110, 0x7910, 0x11E, 0x7910, 0x110, 0x797E, 0x4942, 0x4942, 0x4942, 0x7A42, 0x4A7E, 0x442}},
    {18169, 2, {0x2080, 0x10FC, 0x104, 0xFDF8, 0x8, 0x7BFE, 0x40, 0x78A2, 0x334, 0x7858, 0x4894, 0x4B34, 0x4852, 0x7890, 0x4B50, 0x20}},
    {18170, 2, {0x21FC, 0x1124, 0x124, 0xFDFC, 0x124, 0x7924, 0x1FC, 0x7850, 0x88, 0x7904, 0x4A8A, 0x4888, 0x4888, 0x7888, 0x4908, 0x208}},
    {18171, 2, {0x2008, 0x101C, 0x1F0, 0xFD10, 0x110, 0x79FE, 0x110, 0x797C, 0x144, 0x797C, 0x4944, 0x497C, 0x4944, 0x7A44, 0x427C, 0x444}},
    {18172, 2, {0x2040, 0x119C, 0x104, 0xFD04, 0x1DC, 0x7904, 0x104, 0x79FC, 0x0, 0x7800, 0x49FC, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {18173, 2, {0x2008, 0x103C, 0x1E0, 0xFC20, 0x20, 0x7BFE, 0x20, 0x78A0, 0x12C, 0x7924, 0x4924, 0x49AC, 0x4924, 0x7924, 0x49FC, 0x104}},
    {18174, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7820, 0x3FE, 0x7840, 0x88, 0x79F0, 0x4824, 0x4BFE, 0x4822, 0x7924, 0x4AA2, 0x40}},
    {18175, 2, {0x2000, 0x13FE, 0x0, 0xFDFC, 0x104, 0x7904, 0x1FC, 0x7800, 0x3FE, 0x7A8A, 0x4A52, 0x4BFE, 0x4A22, 0x7A22, 0x4A2A, 0x204}},
    {18176, 2, {0x4000, 0x23DE, 0x14A, 0xF94A, 0x35A, 0x756A, 0x14A, 0x72D6, 0x20, 0x7020, 0x5120, 0x513C, 0x5120, 0x7120, 0x57FE, 0x0}},
    {18177, 2, {0x2020, 0x11FC, 0x220, 0xFBFE, 0x200, 0x7A20, 0x3FC, 0x7A20, 0x3FE, 0x7A00, 0x4BE0, 0x4820, 0x4BFE, 0x7850, 0x4888, 0x306}},
    {18178, 2, {0x2020, 0x1040, 0x1FC, 0xFD04, 0x1FC, 0x7904, 0x1FC, 0x7840, 0x3FE, 0x7888, 0x4924, 0x4A22, 0x4DFC, 0x7820, 0x4820, 0x20}},
    {18179, 2, {0x2050, 0x1048, 0xFEFE, 0x90, 0x7DFE, 0x90, 0x7CFE, 0x90, 0x7C90, 0x44FE, 0x7C80, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0}},
    {18180, 2, {0x2000, 0x11FE, 0x102, 0xFDFE, 0x110, 0x7954, 0x138, 0x7954, 0x182, 0x7950, 0x497C, 0x4990, 0x4910, 0x7AFE, 0x4A10, 0x410}},
    {18181, 2, {0x2020, 0x1222, 0x3FE, 0xFC90, 0x88, 0x79FE, 0x310, 0x7D10, 0x1FE, 0x7910, 0x4910, 0x49FE, 0x4910, 0x7910, 0x49FE, 0x100}},
    {18182, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3492, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x2FF8, 0x2000, 0x4FF8, 0x4808, 0x8FF8}},
    {18183, 2, {0x2040, 0x1020, 0x3FE, 0xFC88, 0x124, 0x7A42, 0x90, 0x79F8, 0x8, 0x7844, 0x48A8, 0x4990, 0x4A88, 0x7CA4, 0x48C2, 0x80}},
    {18184, 2, {0x2124, 0x1248, 0x124, 0xFC00, 0x1FC, 0x7924, 0x1FC, 0x7924, 0x1FC, 0x7820, 0x4BFE, 0x4870, 0x48A8, 0x7924, 0x4622, 0x20}},
    {18185, 2, {0x2020, 0x1020, 0x3FE, 0xFC70, 0xAC, 0x7B22, 0x50, 0x7888, 0x124, 0x7A22, 0x4924, 0x48A8, 0x4924, 0x7A22, 0x48A0, 0x40}},
    {18186, 2, {0x2090, 0x1090, 0x7FE, 0xF890, 0x3FC, 0x7A94, 0x3FC, 0x7A94, 0x3FC, 0x7800, 0x49F8, 0x4908, 0x49F8, 0x7908, 0x49F8, 0x108}},
    {18187, 2, {0x2020, 0x11FC, 0x124, 0xFDFC, 0x20, 0x7BFE, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {18188, 2, {0x1204, 0x1104, 0x2008, 0x2B8A, 0x4812, 0xF3BC, 0x1004, 0x2388, 0x4812, 0xFBBE, 0xA82, 0x280, 0xAAAA, 0xABAA, 0x82A2, 0x0}},
    {18189, 2, {0x2020, 0x1020, 0x1FC, 0xFC20, 0x154, 0x7888, 0x104, 0x7AFA, 0x88, 0x78F8, 0x4888, 0x48F8, 0x4820, 0x78A8, 0x4924, 0x60}},
    {18190, 2, {0x2088, 0x12AA, 0x2DC, 0xFC88, 0x154, 0x7A22, 0x0, 0x7BFE, 0x242, 0x7C44, 0x4BFC, 0x4844, 0x4844, 0x7884, 0x4914, 0x208}},
    {18191, 2, {0x4020, 0x2720, 0x5FE, 0xFD40, 0x57C, 0x7690, 0x57E, 0x7500, 0x57C, 0x7544, 0x557C, 0x5644, 0x547C, 0x7444, 0x5444, 0x44C}},
    {18192, 2, {0x40, 0x7BFC, 0x4880, 0x51F8, 0x4A20, 0x4DFC, 0x6888, 0x52F8, 0x4188, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {18193, 2, {0x2000, 0x13FE, 0x50, 0xFDDC, 0x104, 0x7904, 0x1DC, 0x7850, 0x50, 0x7BFE, 0x4840, 0x4820, 0x48A4, 0x7A8A, 0x4A8A, 0x478}},
    {18194, 2, {0x2040, 0x1080, 0x3FE, 0xFE22, 0x2AA, 0x7A22, 0x3FE, 0x7A72, 0x2AA, 0x7A02, 0x4820, 0x4BFE, 0x4850, 0x7888, 0x4904, 0x602}},
    {18195, 2, {0x2088, 0x13FE, 0x88, 0xFC00, 0x148, 0x7BFE, 0x148, 0x7978, 0x100, 0x79FC, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x4B26, 0x20}},
    {18196, 2, {0x2020, 0x10FE, 0x240, 0xFD7C, 0x90, 0x797E, 0x0, 0x7B7C, 0x144, 0x797C, 0x4944, 0x497C, 0x4944, 0x794C, 0x4A80, 0x47E}},
    {18197, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0x1FC, 0x7924, 0x1FC, 0x7924, 0x1FC, 0x7820, 0x4BFE, 0x4A22, 0x4A2A, 0x7AFA, 0x4A0A, 0x206}},
    {18198, 2, {0x2000, 0x13FE, 0x50, 0xFBFE, 0x252, 0x7BFE, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {18199, 2, {0x4014, 0x2012, 0x7FE, 0xF410, 0x5D0, 0x7414, 0x5D4, 0x7558, 0x5CA, 0x7416, 0x5442, 0x5824, 0x52A2, 0x728A, 0x5478, 0x0}},
    {18200, 2, {0x4288, 0x2288, 0x7E8, 0xFA90, 0x41E, 0x77E4, 0x454, 0x7854, 0x754, 0x7554, 0x5554, 0x5748, 0x5548, 0x7054, 0x5294, 0x122}},
    {18201, 2, {0x2000, 0x13FE, 0x202, 0xFBFE, 0x202, 0x7BFE, 0x108, 0x7A52, 0x39C, 0x7908, 0x4A52, 0x4BDE, 0x4800, 0x7AA4, 0x4A52, 0x452}},
    {18202, 2, {0x4400, 0x22FE, 0x2AA, 0xF8AA, 0xFE, 0x7010, 0xE54, 0x7292, 0x210, 0x7254, 0x5292, 0x5210, 0x5210, 0x7500, 0x58FE, 0x0}},
    {18203, 2, {0x2080, 0x11F8, 0x210, 0xFDFC, 0x124, 0x79FC, 0x92, 0x7B0E, 0x80, 0x79F8, 0x4A10, 0x49FC, 0x4924, 0x79FC, 0x4892, 0x30E}},
    {18204, 2, {0x2040, 0x107C, 0x40, 0xFBFE, 0x202, 0x79FC, 0xA8, 0x7954, 0x288, 0x79FC, 0x4A8A, 0x48F8, 0x4888, 0x78F8, 0x4888, 0xF8}},
    {18205, 2, {0x2044, 0x1284, 0x2EA, 0xFE4E, 0x2A4, 0x7AEE, 0x222, 0x7BFE, 0x244, 0x7A84, 0x4AEA, 0x4A4E, 0x4AA4, 0x7AEE, 0x4A22, 0x3FE}},
    {18206, 2, {0x4040, 0x2020, 0x3FE, 0xF888, 0x52, 0x77AC, 0x2AA, 0x72A8, 0x5AC, 0x7000, 0x5104, 0x51FC, 0x5104, 0x71FC, 0x5104, 0x204}},
    {18207, 2, {0x2088, 0x108C, 0x3EA, 0xF888, 0x88, 0x7BFE, 0x148, 0x7B6A, 0x14A, 0x7B6A, 0x494C, 0x4B6C, 0x494A, 0x796A, 0x4B96, 0x22}},
    {18208, 2, {0x2042, 0x139C, 0x210, 0xFBDE, 0x294, 0x7A94, 0x420, 0x7BFC, 0x204, 0x7BFC, 0x4A04, 0x4BFC, 0x4A04, 0x7BFC, 0x4908, 0x204}},
    {18209, 2, {0x43FC, 0x2040, 0x1F8, 0xF908, 0x1F8, 0x7108, 0x7FE, 0x7442, 0x2A4, 0x728A, 0x547A, 0x5080, 0x51F8, 0x7288, 0x5070, 0x38E}},
    {18210, 2, {0x2020, 0x13FE, 0x222, 0xFDFC, 0x20, 0x78F8, 0x20, 0x79FC, 0x20, 0x79FC, 0x4954, 0x49FC, 0x4840, 0x7AA4, 0x4A8A, 0x47A}},
    {18211, 2, {0x8, 0x201C, 0x11E0, 0x1100, 0x100, 0x100, 0xF1FE, 0x1110, 0x1110, 0x1110, 0x1110, 0x1510, 0x1910, 0x1210, 0x210, 0x410}},
    {18212, 2, {0x110, 0x4110, 0x2110, 0x2110, 0x7FC, 0x110, 0xE110, 0x2110, 0x2110, 0x2FFE, 0x2000, 0x2910, 0x3108, 0x2208, 0x404, 0x804}},
    {18213, 2, {0x0, 0x407C, 0x2F44, 0x2144, 0x27C, 0x244, 0xE444, 0x2744, 0x217C, 0x2500, 0x2500, 0x2AFE, 0x3200, 0x2500, 0x8FE, 0x0}},
    {18214, 2, {0x110, 0x2110, 0x17FE, 0x1110, 0x200, 0x3FC, 0xF404, 0x13E4, 0x1224, 0x13E4, 0x1224, 0x17E4, 0x1A24, 0x1004, 0x28, 0x10}},
    {18215, 2, {0x0, 0x23DE, 0x114A, 0x114A, 0x35A, 0x56A, 0xF14A, 0x12D6, 0x1020, 0x1020, 0x1120, 0x153C, 0x1920, 0x1120, 0x7FE, 0x0}},
    {18216, 2, {0x20, 0x2020, 0x11FC, 0x1124, 0x1FC, 0x20, 0xF3FE, 0x1222, 0x13FE, 0x1020, 0x1040, 0x1424, 0x1AA2, 0x128A, 0x478, 0x0}},
    {18217, 2, {0x110, 0x4110, 0x27FC, 0x2110, 0x1F0, 0x208, 0xE2EA, 0x26AC, 0x22A8, 0x22AA, 0x26EA, 0x2A06, 0x3000, 0x22A4, 0x252, 0x452}},
    {18218, 2, {0x40, 0x2440, 0x2240, 0x4A40, 0x89F8, 0x1448, 0x2248, 0x4148, 0xBC48, 0x2448, 0x2448, 0x244A, 0x248A, 0x3C8A, 0x2506, 0x200}},
    {18219, 2, {0x0, 0x2400, 0x22F8, 0x4A20, 0x8820, 0x1420, 0x2220, 0x4120, 0xBC20, 0x2420, 0x2420, 0x2420, 0x2420, 0x3DFC, 0x2400, 0x0}},
    {18220, 2, {0x20, 0x2420, 0x2220, 0x4A20, 0x89FC, 0x1524, 0x2324, 0x4124, 0xBD24, 0x25FC, 0x2524, 0x2524, 0x2524, 0x3D24, 0x25FC, 0x104}},
    {18221, 2, {0x100, 0x1F8, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x820, 0x1110, 0x2288, 0xC60, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {18222, 2, {0x20, 0x2410, 0x22FE, 0x4A82, 0x8914, 0x1450, 0x227E, 0x4150, 0xBC90, 0x2410, 0x25FE, 0x2410, 0x2410, 0x3C10, 0x2410, 0x10}},
    {18223, 2, {0x40, 0x2440, 0x227E, 0x4A82, 0x8922, 0x14AA, 0x2272, 0x4122, 0xBCFE, 0x2422, 0x2472, 0x24AA, 0x2522, 0x3C22, 0x240A, 0x4}},
    {18224, 2, {0x20, 0x4840, 0x45FC, 0x9504, 0x11FC, 0x2904, 0x45FC, 0x8040, 0x7BFE, 0x4888, 0x4924, 0x4A22, 0x4DFC, 0x7820, 0x4820, 0x20}},
    {18225, 2, {0x0, 0x49F8, 0x4508, 0x95F8, 0x1108, 0x29F8, 0x4400, 0x83FC, 0x7A94, 0x4BFC, 0x4800, 0x49F8, 0x4890, 0x7860, 0x4998, 0x606}},
    {18226, 2, {0x124, 0x4A48, 0x4524, 0x9400, 0x13FC, 0x2A94, 0x4664, 0x8294, 0x7BFC, 0x4A48, 0x4B68, 0x4A48, 0x4B6A, 0x7A4A, 0x4A46, 0x362}},
    {18227, 2, {0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {18228, 2, {0x20, 0xFE20, 0x20, 0x3E, 0x7C20, 0x4420, 0x4420, 0x45FC, 0x7C84, 0x88, 0x4448, 0x2850, 0x2E20, 0xF050, 0x4188, 0x606}},
    {18229, 2, {0x10, 0xFE10, 0x10, 0x1FE, 0x7D12, 0x4514, 0x4510, 0x45FC, 0x7D44, 0x144, 0x4528, 0x2928, 0x2F10, 0xF228, 0x4244, 0x482}},
    {18230, 2, {0x20, 0xFE20, 0x3C, 0x44, 0x7CA8, 0x4410, 0x4428, 0x44D6, 0x7C10, 0xFC, 0x4410, 0x2890, 0x2EFE, 0xF010, 0x4010, 0x10}},
    {18231, 2, {0x0, 0xFEFC, 0x4, 0x7C, 0x7C04, 0x44FC, 0x4400, 0x45FE, 0x7D02, 0xFC, 0x4444, 0x2844, 0x2E28, 0xF010, 0x4068, 0x186}},
    {18232, 2, {0x0, 0xFEFE, 0x82, 0x82, 0x7CFE, 0x4480, 0x448C, 0x44B0, 0x7C9C, 0xB0, 0x449E, 0x28F0, 0x2E92, 0xF112, 0x410E, 0x200}},
    {18233, 2, {0x20, 0xFE20, 0x1FE, 0x20, 0x7C20, 0x45FC, 0x4524, 0x4524, 0x7DFC, 0x20, 0x4470, 0x28A8, 0x2F24, 0xF222, 0x4020, 0x20}},
    {18234, 2, {0x48, 0xFE48, 0x48, 0x1FE, 0x7C48, 0x4448, 0x45FE, 0x4400, 0x7CFC, 0x84, 0x4484, 0x28FC, 0x2E84, 0xF084, 0x40FC, 0x84}},
    {18235, 2, {0x0, 0xFEFC, 0x84, 0x84, 0x7CFC, 0x4484, 0x4484, 0x44FC, 0x7C00, 0x1FE, 0x4420, 0x28A0, 0x2EBC, 0xF0A0, 0x4160, 0x23E}},
    {18236, 2, {0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x98C6, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7FFC}},
    {18237, 2, {0x20, 0xFE50, 0x88, 0x104, 0x7EFA, 0x4400, 0x45E2, 0x452A, 0x7D2A, 0x1EA, 0x452A, 0x292A, 0x2FEA, 0xF122, 0x412A, 0x164}},
    {18238, 2, {0x20, 0xFE10, 0x1FE, 0x0, 0x7C84, 0x4448, 0x45FE, 0x4400, 0x7CFC, 0x84, 0x4484, 0x28FC, 0x2E84, 0xF084, 0x40FC, 0x84}},
    {18239, 2, {0x0, 0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7FFC}},
    {18240, 2, {0x20, 0xFEDE, 0x8A, 0xCA, 0x7CAA, 0x44D2, 0x44A6, 0x4400, 0x7CFE, 0x92, 0x4492, 0x28FE, 0x2E92, 0xF092, 0x40FE, 0x82}},
    {18241, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE}},
    {18242, 2, {0x7E78, 0x4848, 0x7E8E, 0x4200, 0x42F8, 0x7E48, 0x4830, 0x7ECE, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7FFC}},
    {18243, 2, {0x0, 0xFEFE, 0x82, 0xFE, 0x7C82, 0x44FE, 0x4444, 0x44AA, 0x7CEE, 0x44, 0x44AA, 0x28EE, 0x2E00, 0xF0D4, 0x40AA, 0x12A}},
    {18244, 2, {0x108, 0xFA94, 0x3DE, 0x6B4, 0x7BDE, 0x5294, 0x53DE, 0x5294, 0x73DE, 0x210, 0x97FC, 0x5108, 0x6090, 0x3860, 0xC198, 0x606}},
    {18245, 2, {0x20, 0xFE20, 0x820, 0x1020, 0x3020, 0x4A20, 0x9DFC, 0x2820, 0x4C20, 0x9A20, 0x2A20, 0x4820, 0x8820, 0x820, 0x2BFE, 0x1000}},
    {18246, 2, {0x800, 0x1FF8, 0x2248, 0x4488, 0x950, 0x1220, 0x7FFC, 0x200, 0xD10, 0x72A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {18247, 2, {0x0, 0x25FC, 0x2220, 0x4240, 0x80A4, 0x2538, 0x2450, 0x2498, 0x2538, 0x2454, 0x2494, 0x2550, 0x2422, 0x4402, 0x43FE, 0x8000}},
    {18248, 2, {0x1C, 0xFDE0, 0x1100, 0x2100, 0x61FE, 0x9500, 0x3900, 0x517C, 0x9944, 0x3544, 0x5554, 0x9148, 0x1142, 0x1242, 0x523E, 0x2400}},
    {18249, 2, {0x440, 0x2448, 0x1450, 0xFFFE, 0x8002, 0x0, 0x7FFC, 0x200, 0xD10, 0x72A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {18250, 2, {0x20, 0xFC20, 0x13FE, 0x2020, 0x61FC, 0x9424, 0x3BFE, 0x5024, 0x99FC, 0x3420, 0x5520, 0x913E, 0x1120, 0x12A0, 0x527E, 0x2400}},
    {18251, 2, {0x40, 0xFC20, 0x13FE, 0x2202, 0x6000, 0x95FC, 0x3800, 0x5000, 0x9BFE, 0x3420, 0x5528, 0x9124, 0x1222, 0x1422, 0x50A0, 0x2040}},
    {18252, 2, {0x0, 0xFBDE, 0x2042, 0x414A, 0xA084, 0x154A, 0x3A52, 0x5420, 0x9800, 0x37DE, 0x5052, 0x9152, 0x1094, 0x1148, 0x5254, 0x2422}},
    {18253, 2, {0x0, 0xFBFE, 0x2222, 0x4020, 0xA1FE, 0x1420, 0x39FC, 0x5124, 0x99FC, 0x3524, 0x51FC, 0x9020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {18254, 2, {0x20, 0xFD24, 0x1124, 0x2124, 0x61FC, 0x9400, 0x3BFE, 0x5020, 0x9840, 0x35FC, 0x5554, 0x9154, 0x1154, 0x1154, 0x5154, 0x210C}},
    {18255, 2, {0x20, 0xFDFC, 0x1124, 0x23FE, 0x6124, 0x95FC, 0x3820, 0x51FC, 0x9924, 0x35FC, 0x5440, 0x93FE, 0x1088, 0x11D0, 0x5070, 0x238C}},
    {18256, 2, {0x10, 0xFB90, 0x22FE, 0x42A0, 0xA2BE, 0x1748, 0x3ABE, 0x5280, 0x9ABE, 0x36A2, 0x52BE, 0x9322, 0x123E, 0x1222, 0x5222, 0x2226}},
    {18257, 2, {0x20, 0xFDFC, 0x1088, 0x2050, 0x63FE, 0x9400, 0x39FC, 0x5124, 0x99FC, 0x3524, 0x55FC, 0x9020, 0x11FC, 0x1020, 0x53FE, 0x2000}},
    {18258, 2, {0x3F0, 0xFD2E, 0x11E2, 0x212A, 0x61E4, 0x9534, 0x3BEA, 0x5032, 0x981E, 0x35E0, 0x54A2, 0x9134, 0x1068, 0x10A4, 0x5322, 0x2020}},
    {18259, 2, {0x124, 0xFA48, 0x2124, 0x4000, 0xA3FC, 0x1694, 0x3A64, 0x5294, 0x9BFC, 0x3648, 0x5368, 0x9248, 0x136A, 0x124A, 0x5246, 0x2362}},
    {18260, 2, {0x88, 0xFBFE, 0x2088, 0x43DE, 0xA252, 0x17DE, 0x38A0, 0x5090, 0x99FE, 0x3520, 0x53FC, 0x9520, 0x11FC, 0x1120, 0x51FE, 0x2100}},
    {18261, 2, {0x3FC, 0xF908, 0x11F8, 0x2108, 0x65F8, 0x990E, 0x37F8, 0x5008, 0x97FE, 0x3A94, 0x579C, 0x9294, 0x139C, 0x12D6, 0x57BC, 0x2084}},
    {18262, 2, {0x1008, 0x201C, 0xD5F0, 0x950, 0x5150, 0x2150, 0xD150, 0x1950, 0x2950, 0xC948, 0x1948, 0x2948, 0xC944, 0xA44, 0x5242, 0x2400}},
    {18263, 2, {0xC20, 0x3020, 0xC220, 0x1420, 0x4BFE, 0x3020, 0xC870, 0x1470, 0x24A8, 0xCCA8, 0x1524, 0x2622, 0xC420, 0x420, 0x2820, 0x1020}},
    {18264, 2, {0xC00, 0x30F8, 0xC288, 0x1488, 0x4888, 0x3106, 0xCA00, 0x15FC, 0x2484, 0xCC84, 0x1448, 0x2450, 0xC420, 0x450, 0x2888, 0x1306}},
    {18265, 2, {0xC00, 0x31FE, 0xC300, 0x1504, 0x4944, 0x3128, 0xC928, 0x1510, 0x2510, 0xCD28, 0x1528, 0x2544, 0xC584, 0x500, 0x29FE, 0x1000}},
    {18266, 2, {0x1000, 0x21FC, 0xD504, 0x904, 0x5104, 0x21FC, 0xD100, 0x1940, 0x2944, 0xC948, 0x1970, 0x2940, 0xC942, 0xA42, 0x523E, 0x2400}},
    {18267, 2, {0x1010, 0x2010, 0xD410, 0x9FE, 0x5112, 0x2114, 0xD110, 0x19FC, 0x2944, 0xC944, 0x1928, 0x2928, 0xC910, 0xA28, 0x5244, 0x2482}},
    {18268, 2, {0x1040, 0x2040, 0xD488, 0x904, 0x53FE, 0x2002, 0xD088, 0x1944, 0x2A42, 0xC8F8, 0x1988, 0x2A50, 0xC820, 0x850, 0x5188, 0x2606}},
    {18269, 2, {0xC20, 0x3020, 0xC3FC, 0x1450, 0x4888, 0x3104, 0xCBFE, 0x1408, 0x25E8, 0xCD28, 0x1528, 0x25E8, 0xC528, 0x408, 0x2828, 0x1010}},
    {18270, 2, {0x1020, 0x2020, 0xD43E, 0x820, 0x51FE, 0x2122, 0xD138, 0x19E0, 0x2922, 0xC91E, 0x1900, 0x2978, 0xCA48, 0xA4A, 0x548A, 0x2906}},
    {18271, 2, {0x1080, 0x209E, 0xD7EA, 0x88A, 0x51CA, 0x208A, 0xD3EA, 0x1892, 0x28A6, 0xC820, 0x1BFE, 0x2820, 0xC850, 0x888, 0x5104, 0x2602}},
    {18272, 2, {0x1000, 0x21FE, 0xD510, 0x920, 0x517C, 0x2144, 0xD17C, 0x1944, 0x297C, 0xC910, 0x1910, 0x2954, 0xC952, 0xA92, 0x5250, 0x2420}},
    {18273, 2, {0x1088, 0x2088, 0xD7FE, 0x888, 0x5088, 0x20F8, 0xD020, 0x1BFE, 0x2A22, 0xCB32, 0x1AAA, 0x2B76, 0xCA22, 0xA22, 0x522A, 0x2224}},
    {18274, 2, {0x1000, 0x21F8, 0xD508, 0x9F8, 0x5108, 0x21F8, 0xD000, 0x1BFC, 0x2A94, 0xCBFC, 0x1800, 0x29F8, 0xC890, 0x860, 0x5198, 0x2606}},
    {18275, 2, {0x1200, 0x217E, 0xD440, 0xC7C, 0x5244, 0x20FC, 0xD140, 0x1E7E, 0x2A20, 0xC820, 0x1BFE, 0x2870, 0xC8A8, 0x924, 0x5222, 0x2020}},
    {18276, 2, {0xC20, 0x3020, 0xC3FC, 0x1420, 0x4954, 0x3088, 0xC904, 0x16FA, 0x2488, 0xCCF8, 0x1488, 0x24F8, 0xC420, 0x4A8, 0x2924, 0x1060}},
    {18277, 2, {0x1040, 0x21FC, 0xD504, 0x9FC, 0x5104, 0x21FC, 0xD104, 0x19FC, 0x2820, 0xC924, 0x18A8, 0x2924, 0xC820, 0xBFE, 0x5020, 0x2020}},
    {18278, 2, {0x1040, 0x2080, 0xD7FE, 0xA52, 0x5252, 0x27FE, 0xD000, 0x181E, 0x2BE0, 0xC8A2, 0x1922, 0x2A74, 0xC8A8, 0x924, 0x5622, 0x2020}},
    {18279, 2, {0x40, 0x7C40, 0x4480, 0x44FC, 0x7D04, 0x4604, 0x4484, 0x7C44, 0x4444, 0x4414, 0x7C24, 0x44, 0x2984, 0x2404, 0x4428, 0x8010}},
    {18280, 2, {0x400, 0x25FC, 0x2488, 0x2450, 0x2420, 0x2450, 0x8C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18281, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x7D04, 0x4612, 0x4420, 0x7C40, 0x4588, 0x4410, 0x7C20, 0x44, 0x2988, 0x2410, 0x4460, 0x8380}},
    {18282, 2, {0x0, 0x7C00, 0x45FE, 0x4412, 0x7C14, 0x4410, 0x4490, 0x7C90, 0x449E, 0x4490, 0x7C90, 0x90, 0x28D0, 0x2530, 0x451E, 0x8200}},
    {18283, 2, {0x80, 0x7C80, 0x4480, 0x44FE, 0x7D40, 0x4540, 0x4640, 0x7C7C, 0x4440, 0x4440, 0x7C40, 0x7C, 0x2840, 0x2440, 0x4440, 0x8040}},
    {18284, 2, {0x40, 0x7C20, 0x4400, 0x45FC, 0x7C20, 0x4420, 0x4420, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x20, 0x2820, 0x2420, 0x47FE, 0x8000}},
    {18285, 2, {0x80, 0x7C80, 0x4500, 0x45FC, 0x7E04, 0x4404, 0x45E4, 0x7D24, 0x4524, 0x4524, 0x7DE4, 0x124, 0x2804, 0x2404, 0x4428, 0x8010}},
    {18286, 2, {0x20, 0x7C10, 0x4410, 0x45FE, 0x7C20, 0x4420, 0x4444, 0x7C84, 0x45F8, 0x4410, 0x7C20, 0x44, 0x2882, 0x25FE, 0x4482, 0x8000}},
    {18287, 2, {0x80, 0x7C80, 0x44FE, 0x4500, 0x7E20, 0x4520, 0x452C, 0x7D74, 0x47A4, 0x4524, 0x7D34, 0x128, 0x2922, 0x2502, 0x44FE, 0x8000}},
    {18288, 2, {0x20, 0x7C20, 0x4420, 0x45FC, 0x7C20, 0x4420, 0x47FE, 0x7C08, 0x4408, 0x45FE, 0x7C08, 0x88, 0x2848, 0x2408, 0x4428, 0x8010}},
    {18289, 2, {0x200, 0x1FF0, 0x810, 0x1030, 0x7EFC, 0x2244, 0x468C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18290, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4490, 0x4490, 0x7C88, 0x4484, 0x4532, 0x7D08, 0x200, 0x2860, 0x2418, 0x4404, 0x8000}},
    {18291, 2, {0x800, 0xF7C, 0x2844, 0x7F28, 0xA110, 0x1628, 0x3846, 0xDFF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18292, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFE, 0x5012, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18293, 2, {0x28, 0x7C24, 0x45FE, 0x4420, 0x7C20, 0x45FC, 0x4524, 0x7D24, 0x45FC, 0x4524, 0x7D24, 0x1FC, 0x2924, 0x2524, 0x4524, 0x810C}},
    {18294, 2, {0x40, 0x7C20, 0x47FE, 0x4400, 0x7C00, 0x45FC, 0x4504, 0x7D04, 0x45FC, 0x4420, 0x7D28, 0x124, 0x2A22, 0x2422, 0x44A0, 0x8040}},
    {18295, 2, {0x2000, 0x3E7C, 0x4844, 0x844, 0xFF44, 0x247C, 0x4200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18296, 2, {0x88, 0x7C88, 0x47FE, 0x4488, 0x7CA8, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x45FC, 0x7C84, 0x88, 0x2850, 0x2420, 0x44D8, 0x8306}},
    {18297, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x7D04, 0x4900, 0x49DE, 0x7A52, 0x4A52, 0x4B52, 0x7C9A, 0x94, 0x5110, 0x4912, 0x8A12, 0x40E}},
    {18298, 2, {0x1020, 0x810, 0x7EFC, 0x448, 0x2448, 0x1830, 0x64CC, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18299, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x7DFC, 0x4510, 0x4510, 0x7DFE, 0x4510, 0x4510, 0x7D7C, 0x144, 0x2944, 0x2544, 0x457C, 0x8244}},
    {18300, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x7DFC, 0x4420, 0x47FE, 0x7C00, 0x45FC, 0x4504, 0x7DFC, 0x104, 0x29FC, 0x2504, 0x4514, 0x8108}},
    {18301, 2, {0x0, 0x79FE, 0x4910, 0x4910, 0x797C, 0x4910, 0x4910, 0x79FE, 0x4900, 0x4910, 0x7910, 0x17C, 0x5110, 0x4A10, 0x8AFE, 0x400}},
    {18302, 2, {0x3F50, 0x48, 0xFFFE, 0x840, 0x2F20, 0x2812, 0x2F0A, 0xF006, 0x3FFA, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820}},
    {18303, 2, {0x500, 0x3978, 0x2108, 0x3D78, 0x2208, 0x3FF8, 0x820, 0x3FF8, 0xD016, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18304, 2, {0x80, 0x7CBC, 0x4484, 0x4508, 0x7DFE, 0x4720, 0x4520, 0x7D7C, 0x4590, 0x4510, 0x7DFE, 0x110, 0x2928, 0x2528, 0x4544, 0x8182}},
    {18305, 2, {0x0, 0x7BFE, 0x4A00, 0x4AFC, 0x7A84, 0x4AFC, 0x4A84, 0x7AFC, 0x4A20, 0x4BFE, 0x7A48, 0x2C8, 0x5230, 0x4A48, 0x8A84, 0x3FE}},
    {18306, 2, {0x0, 0x7DFC, 0x4420, 0x4440, 0x7DFC, 0x4554, 0x4554, 0x7D54, 0x4544, 0x452C, 0x7C20, 0x3FE, 0x2850, 0x2488, 0x4504, 0x8202}},
    {18307, 2, {0x8, 0x7008, 0x57C8, 0x5210, 0x721E, 0x54A4, 0x57D4, 0x7114, 0x5114, 0x57D4, 0x7114, 0x108, 0x51C8, 0x4E14, 0x8414, 0x22}},
    {18308, 2, {0x2040, 0xFEFE, 0x2120, 0x3CAC, 0x24F4, 0x47AC, 0x54A2, 0x887E, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18309, 2, {0x0, 0x79FE, 0x4900, 0x497E, 0x7900, 0x49FE, 0x4954, 0x7948, 0x4964, 0x4942, 0x7900, 0x17E, 0x5142, 0x4A42, 0x8A7E, 0x442}},
    {18310, 2, {0x20, 0x7D24, 0x44A8, 0x4420, 0x7DFC, 0x4504, 0x4504, 0x7DFC, 0x4504, 0x45FC, 0x7D04, 0x104, 0x29FC, 0x2400, 0x4488, 0x8104}},
    {18311, 2, {0x820, 0x3E20, 0x8F8, 0x7F28, 0x2A68, 0x5D2A, 0x856, 0x3E82, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {18312, 2, {0x1E, 0x7BE0, 0x4944, 0x48A8, 0x79F8, 0x4908, 0x4908, 0x79FC, 0x4904, 0x4904, 0x79FE, 0x102, 0x52AA, 0x4AAA, 0x8D0A, 0x4}},
    {18313, 2, {0x20, 0x7D24, 0x44A8, 0x45FC, 0x7C40, 0x47FE, 0x4488, 0x7DFC, 0x468A, 0x44F8, 0x7C88, 0xF8, 0x2888, 0x24F8, 0x4450, 0x8088}},
    {18314, 2, {0x7C, 0x7840, 0x4BFE, 0x4A42, 0x7A78, 0x4BC4, 0x4A3C, 0x7AA8, 0x4B54, 0x4AF8, 0x7B8C, 0x2FA, 0x5288, 0x4AF8, 0x8A88, 0x4F8}},
    {18315, 2, {0x200, 0x713C, 0x57A4, 0x5024, 0x74BC, 0x5324, 0x57A4, 0x713C, 0x5124, 0x57A4, 0x713C, 0x590, 0x5558, 0x4968, 0x812A, 0x346}},
    {18316, 2, {0x8, 0x7C3C, 0x45E0, 0x5420, 0x5420, 0x57FE, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5504, 0x1104, 0x2904, 0x2504, 0x45FC, 0x8104}},
    {18317, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x55FC, 0x5420, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x1104, 0x29FC, 0x2504, 0x4114, 0x8108}},
    {18318, 2, {0xF8, 0x7C88, 0x4488, 0x54F8, 0x5488, 0x5488, 0x54F8, 0x5400, 0x55FC, 0x5504, 0x5504, 0x11FC, 0x2904, 0x2504, 0x45FC, 0x8104}},
    {18319, 2, {0x0, 0x7DFC, 0x4524, 0x4524, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x5420, 0x5410, 0x5454, 0x1142, 0x2942, 0x254A, 0x4238, 0x8000}},
    {18320, 2, {0x20, 0x7C20, 0x47FE, 0x5420, 0x55FC, 0x5440, 0x57FE, 0x5488, 0x5504, 0x56FA, 0x5488, 0x1088, 0x28F8, 0x2488, 0x4488, 0x80F8}},
    {18321, 2, {0x800, 0x800, 0x87C, 0x7E10, 0x810, 0x810, 0xFF10, 0x1410, 0x1410, 0x5610, 0x5510, 0x9510, 0x2410, 0x2410, 0x54FE, 0x8800}},
    {18322, 2, {0x800, 0x8FE, 0x882, 0x7E82, 0x8BA, 0x882, 0xFF82, 0x14BA, 0x14AA, 0x56AA, 0x55AA, 0x94BA, 0x2482, 0x2482, 0x548A, 0x8884}},
    {18323, 2, {0x800, 0x8FE, 0x800, 0x7E52, 0x8A4, 0x928, 0xFF94, 0x144A, 0x1400, 0x56FE, 0x5510, 0x9510, 0x2410, 0x2410, 0x55FE, 0x8800}},
    {18324, 2, {0x800, 0x8FE, 0x882, 0x7E82, 0x8FE, 0x880, 0xFF8C, 0x14B0, 0x149C, 0x56B0, 0x559E, 0x94F0, 0x2492, 0x2512, 0x550E, 0x8A00}},
    {18325, 2, {0x800, 0x87C, 0x810, 0x7EFE, 0x892, 0x854, 0xFF10, 0x1454, 0x1400, 0x56FE, 0x5510, 0x94FE, 0x24AA, 0x24AA, 0x54AA, 0x8886}},
    {18326, 2, {0x850, 0x848, 0x848, 0x7E40, 0x85C, 0x9E0, 0xFE40, 0x840, 0x2820, 0x2824, 0x2E14, 0x280C, 0x2804, 0x5800, 0x4FFE, 0x8000}},
    {18327, 2, {0x808, 0x808, 0x808, 0x7E08, 0x8FE, 0x818, 0xFE18, 0x828, 0x2828, 0x2848, 0x2E48, 0x2888, 0x2818, 0x5800, 0x4FFE, 0x8000}},
    {18328, 2, {0x840, 0x840, 0x87C, 0x7E80, 0x900, 0x800, 0xFEF8, 0x810, 0x2820, 0x2840, 0x2E84, 0x2884, 0x287C, 0x5800, 0x4FFE, 0x8000}},
    {18329, 2, {0x1000, 0x13FE, 0x1020, 0x7C20, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x1124, 0x5124, 0x5D24, 0x5134, 0x5128, 0x7020, 0x5020, 0x8FFE}},
    {18330, 2, {0x810, 0x810, 0x8FE, 0x7E10, 0x810, 0x8FC, 0xFF04, 0x804, 0x2868, 0x2810, 0x2E28, 0x2844, 0x2882, 0x5800, 0x4FFE, 0x8000}},
    {18331, 2, {0x800, 0x9FE, 0x820, 0x7EA0, 0x8BC, 0x8A4, 0xFEA4, 0x8FC, 0x2804, 0x2804, 0x2E04, 0x2828, 0x2810, 0x5800, 0x4FFE, 0x8000}},
    {18332, 2, {0x800, 0x800, 0x9FE, 0x7E10, 0x820, 0x820, 0xFE68, 0x864, 0x28A2, 0x2922, 0x2E20, 0x2820, 0x2820, 0x5800, 0x4FFE, 0x8000}},
    {18333, 2, {0x800, 0x800, 0x8FC, 0x7E20, 0x820, 0x820, 0xFE20, 0x8FC, 0x2828, 0x2824, 0x2E24, 0x2820, 0x29FE, 0x5800, 0x4FFE, 0x8000}},
    {18334, 2, {0x1050, 0x1050, 0x1050, 0x7DFC, 0x1054, 0x1054, 0xFDFC, 0x1150, 0x1150, 0x51FE, 0x5C52, 0x5052, 0x505A, 0x7094, 0x5110, 0x8FFE}},
    {18335, 2, {0x820, 0x820, 0x840, 0x7EFC, 0x884, 0x884, 0xFE84, 0x8FC, 0x2884, 0x2884, 0x2E84, 0x28FC, 0x2884, 0x5800, 0x4FFE, 0x8000}},
    {18336, 2, {0x800, 0x9FE, 0x820, 0x7E20, 0x840, 0x87C, 0xFEC4, 0x944, 0x2844, 0x2844, 0x2E44, 0x287C, 0x2844, 0x5800, 0x4FFE, 0x8000}},
    {18337, 2, {0x800, 0x800, 0x8FC, 0x7E84, 0x884, 0x8FC, 0xFE84, 0x884, 0x28FC, 0x2884, 0x2E00, 0x29FE, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {18338, 2, {0x840, 0x840, 0x840, 0x7EFE, 0x8A0, 0x920, 0xFE3C, 0x820, 0x2820, 0x283E, 0x2E20, 0x2820, 0x2820, 0x5800, 0x4FFE, 0x8000}},
    {18339, 2, {0x808, 0x81C, 0x8E0, 0x7E80, 0x880, 0x8FE, 0xFE90, 0x890, 0x28B0, 0x2898, 0x2E94, 0x2910, 0x2910, 0x5A00, 0x4FFE, 0x8000}},
    {18340, 2, {0x1080, 0x1080, 0x1100, 0x7DFC, 0x1204, 0x1004, 0xFDE4, 0x1124, 0x1124, 0x5124, 0x5DE4, 0x5124, 0x5004, 0x7028, 0x5010, 0x8FFE}},
    {18341, 2, {0x810, 0x810, 0x9FE, 0x7E20, 0x840, 0x8FC, 0xFF44, 0x87C, 0x2844, 0x287C, 0x2E44, 0x2854, 0x2848, 0x5800, 0x4FFE, 0x8000}},
    {18342, 2, {0x820, 0x820, 0x83C, 0x7E44, 0x8A8, 0x810, 0xFF28, 0x844, 0x2882, 0x287C, 0x2E44, 0x2844, 0x287C, 0x5800, 0x4FFE, 0x8000}},
    {18343, 2, {0x808, 0x81C, 0x8E0, 0x7E80, 0x8FE, 0x880, 0xFE80, 0x8BC, 0x28A4, 0x28A4, 0x2EA4, 0x293C, 0x2924, 0x5A00, 0x4FFE, 0x8000}},
    {18344, 2, {0x820, 0x820, 0x8FC, 0x7E20, 0x820, 0x9FE, 0xFE00, 0x820, 0x28FC, 0x2820, 0x2E20, 0x2820, 0x29FE, 0x5800, 0x4FFE, 0x8000}},
    {18345, 2, {0x810, 0x810, 0x828, 0x7E44, 0x882, 0x87C, 0xFF00, 0x800, 0x287C, 0x2844, 0x2E44, 0x287C, 0x2844, 0x5800, 0x4FFE, 0x8000}},
    {18346, 2, {0x800, 0x8FE, 0x880, 0x7E84, 0x8B8, 0x8A0, 0xFFBE, 0x8A4, 0x28A4, 0x28A4, 0x2EC4, 0x2880, 0x28FE, 0x5800, 0x4FFE, 0x8000}},
    {18347, 2, {0x1000, 0x11FC, 0x1044, 0x7C44, 0x13FE, 0x1044, 0xFC44, 0x11FC, 0x1080, 0x50FC, 0x5D84, 0x5284, 0x50FC, 0x7084, 0x5000, 0x8FFE}},
    {18348, 2, {0x80C, 0x870, 0x810, 0x7EFE, 0x838, 0x854, 0xFE92, 0x800, 0x287C, 0x2824, 0x2E2E, 0x2842, 0x288A, 0x5904, 0x4FFE, 0x8000}},
    {18349, 2, {0x1020, 0x1048, 0x1084, 0x7DFE, 0x1002, 0x1048, 0xFC84, 0x1142, 0x107C, 0x5084, 0x5D68, 0x5010, 0x5068, 0x7184, 0x5000, 0x8FFE}},
    {18350, 2, {0x810, 0x810, 0x828, 0x7E44, 0x8BA, 0x810, 0xFEFE, 0x810, 0x2858, 0x2894, 0x2F12, 0x2850, 0x2820, 0x5800, 0x4FFE, 0x8000}},
    {18351, 2, {0x80E, 0x8F0, 0x822, 0x7E92, 0x854, 0x800, 0xFF7C, 0x808, 0x2810, 0x28FE, 0x2E10, 0x2810, 0x2850, 0x5820, 0x4FFE, 0x8000}},
    {18352, 2, {0x1020, 0x1020, 0x11FC, 0x7C20, 0x1020, 0x13FE, 0xFC50, 0x1050, 0x1154, 0x5152, 0x5E52, 0x5090, 0x5090, 0x7130, 0x5000, 0x8FFE}},
    {18353, 2, {0x828, 0x844, 0x892, 0x7E10, 0x828, 0x844, 0xFE82, 0x97C, 0x2844, 0x2844, 0x2E44, 0x287C, 0x2844, 0x5800, 0x4FFE, 0x8000}},
    {18354, 2, {0x1010, 0x10D8, 0x1394, 0x7C94, 0x1090, 0x13FE, 0xFC94, 0x1094, 0x10D8, 0x5392, 0x5CAA, 0x50CA, 0x5286, 0x7102, 0x5000, 0x8FFE}},
    {18355, 2, {0x820, 0x810, 0x8FC, 0x7E00, 0x884, 0x848, 0xFEFE, 0x800, 0x28FC, 0x2884, 0x2E84, 0x28FC, 0x2884, 0x5800, 0x4FFE, 0x8000}},
    {18356, 2, {0x11FC, 0x1124, 0x1124, 0x7D74, 0x1124, 0x11FC, 0xFD04, 0x1174, 0x1154, 0x5154, 0x5D74, 0x5104, 0x5204, 0x7214, 0x5408, 0x8FFE}},
    {18357, 2, {0x1000, 0x11EE, 0x1022, 0x7CAA, 0x1044, 0x10AA, 0xFD32, 0x1200, 0x11EE, 0x5022, 0x5CAA, 0x5044, 0x50AA, 0x7112, 0x5220, 0x8FFE}},
    {18358, 2, {0x11F8, 0x1108, 0x1108, 0x7DF8, 0x1108, 0x1108, 0xFDF8, 0x1080, 0x11FC, 0x5254, 0x5C94, 0x5124, 0x5244, 0x7094, 0x5108, 0x8FFE}},
    {18359, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x1100, 0x1110, 0xFD54, 0x1154, 0x117C, 0x5110, 0x5D54, 0x5154, 0x5254, 0x747C, 0x5000, 0x8FFE}},
    {18360, 2, {0x1000, 0x10F8, 0x1000, 0x7C00, 0x11FC, 0x1020, 0xFD24, 0x1222, 0x1060, 0x5000, 0x5DFC, 0x5104, 0x5104, 0x71FC, 0x5000, 0x8FFE}},
    {18361, 2, {0x820, 0x840, 0x8FC, 0x7E84, 0x8FC, 0x884, 0xFEFC, 0x800, 0x29FE, 0x2820, 0x2EFC, 0x2820, 0x29FE, 0x5800, 0x4FFE, 0x8000}},
    {18362, 2, {0x81E, 0x9F0, 0x910, 0x7F10, 0x9FE, 0x910, 0xFF7E, 0x942, 0x297E, 0x2942, 0x2F7E, 0x2942, 0x297E, 0x5A00, 0x4FFE, 0x8000}},
    {18363, 2, {0x11F8, 0x1050, 0x1020, 0x7DFC, 0x1124, 0x11FC, 0xFD24, 0x11FC, 0x1124, 0x5040, 0x5DFC, 0x5044, 0x5084, 0x7128, 0x5210, 0x8FFE}},
    {18364, 2, {0x820, 0x83C, 0x844, 0x7E78, 0x808, 0x8FE, 0xFF20, 0x852, 0x28AC, 0x2858, 0x2EAC, 0x284A, 0x28AA, 0x5810, 0x4FFE, 0x8000}},
    {18365, 2, {0x810, 0x810, 0x87C, 0x7E10, 0x8FE, 0x8A2, 0xFE44, 0x878, 0x2814, 0x28FE, 0x2E10, 0x2854, 0x28B2, 0x5800, 0x4FFE, 0x8000}},
    {18366, 2, {0x101E, 0x13E0, 0x123C, 0x7A20, 0x12FE, 0x12A2, 0xFAB8, 0x12E4, 0x129C, 0x5280, 0x5AB8, 0x5328, 0x554A, 0x7286, 0x5000, 0x8FFE}},
    {18367, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x920, 0x17D0, 0x2108, 0xDFF6, 0x100, 0x11F8, 0x2900, 0x47FE}},
    {18368, 2, {0x844, 0x828, 0x8FE, 0x7E10, 0x87C, 0x810, 0xFEFE, 0x820, 0x2820, 0x287E, 0x2E88, 0x2908, 0x287E, 0x5800, 0x4FFE, 0x8000}},
    {18369, 2, {0x1124, 0x1248, 0x1124, 0x7C00, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x11FC, 0x5020, 0x5DFE, 0x5070, 0x50A8, 0x7124, 0x5220, 0x8FFE}},
    {18370, 2, {0x10FC, 0x1084, 0x10FC, 0x7C84, 0x10FC, 0x1000, 0xFDFE, 0x114A, 0x11FE, 0x5000, 0x5CFC, 0x5048, 0x5030, 0x70CC, 0x5302, 0x8FFE}},
    {18371, 2, {0x1040, 0x1088, 0x11FC, 0x7D08, 0x1252, 0x13FE, 0xFC50, 0x118C, 0x1622, 0x51C8, 0x5C32, 0x51CC, 0x5030, 0x71C0, 0x5000, 0x8FFE}},
    {18372, 2, {0x1020, 0x13FE, 0x1020, 0x7DFC, 0x1020, 0x13FE, 0xFC00, 0x11FC, 0x1104, 0x51FC, 0x5D04, 0x51FC, 0x5104, 0x71FC, 0x5088, 0x8FFE}},
    {18373, 2, {0x804, 0x7F78, 0x840, 0x3E40, 0x2A7E, 0x3E48, 0x2A48, 0xFF48, 0x988, 0x3FF0, 0x100, 0xFFFC, 0x1100, 0x11F8, 0x2900, 0x47FE}},
    {18374, 2, {0x11FC, 0x1124, 0x11FC, 0x7D24, 0x11FC, 0x10A8, 0xFCA8, 0x13FE, 0x10A8, 0x50A8, 0x5DFC, 0x5020, 0x53FE, 0x7020, 0x5020, 0x8FFE}},
    {18375, 2, {0x894, 0x918, 0x9DE, 0x7E94, 0x95A, 0x9DE, 0xFE94, 0x9FE, 0x2890, 0x28D4, 0x2EAA, 0x2916, 0x2A22, 0x5800, 0x4FFE, 0x8000}},
    {18376, 2, {0x1094, 0x11D2, 0x1090, 0x7DFE, 0x1010, 0x11D4, 0xFD54, 0x11D4, 0x1014, 0x53E8, 0x5C8A, 0x51CA, 0x5096, 0x73E2, 0x5000, 0x8FFE}},
    {18377, 2, {0x11FC, 0x1154, 0x1154, 0x7DFC, 0x1080, 0x11FC, 0xFE44, 0x11F4, 0x1154, 0x5154, 0x5DF4, 0x5044, 0x51F4, 0x7014, 0x5008, 0x8FFE}},
    {18378, 2, {0x1020, 0x103E, 0x1020, 0x7DFE, 0x1122, 0x11F8, 0xFD22, 0x11FE, 0x1120, 0x5152, 0x5DAC, 0x525A, 0x52AA, 0x7448, 0x5090, 0x8FFE}},
    {18379, 2, {0x1040, 0x107C, 0x1084, 0x7D08, 0x13FE, 0x1144, 0xFD92, 0x117C, 0x1100, 0x517C, 0x5D00, 0x517C, 0x5244, 0x727C, 0x5400, 0x8FFE}},
    {18380, 2, {0x1048, 0x104C, 0x11FA, 0x7C48, 0x13FE, 0x10A8, 0xFDFA, 0x134A, 0x15EC, 0x514C, 0x5DEA, 0x514A, 0x51F6, 0x7122, 0x5000, 0x8FFE}},
    {18381, 2, {0x1040, 0x11FC, 0x1124, 0x7D94, 0x114C, 0x1124, 0xFDFC, 0x1020, 0x13FE, 0x528A, 0x5D24, 0x53FE, 0x5080, 0x70FC, 0x5104, 0x8FFE}},
    {18382, 2, {0x1088, 0x11DC, 0x1088, 0x7DDE, 0x1088, 0x1154, 0xFE22, 0x11FC, 0x1104, 0x51FC, 0x5D04, 0x51FC, 0x5104, 0x71FC, 0x5088, 0x8FFE}},
    {18383, 2, {0x1124, 0x1174, 0x1258, 0x7BFE, 0x1154, 0x12DA, 0xFBFE, 0x1082, 0x1020, 0x53FE, 0x5870, 0x50A8, 0x5124, 0x7222, 0x5020, 0x8FFE}},
    {18384, 2, {0x13FE, 0x1200, 0x1222, 0x7ECC, 0x1244, 0x13EE, 0xFE44, 0x12EE, 0x1354, 0x5210, 0x5E90, 0x529E, 0x5490, 0x73FE, 0x5000, 0x8FFE}},
    {18385, 2, {0x1020, 0x1050, 0x1088, 0x7D74, 0x1202, 0x11FC, 0xFD54, 0x11FC, 0x1000, 0x51FC, 0x5D54, 0x51FC, 0x5154, 0x710C, 0x5000, 0x8FFE}},
    {18386, 2, {0x1088, 0x13FE, 0x1088, 0x7BDE, 0x1252, 0x13DE, 0xF8A0, 0x11FE, 0x1120, 0x53FC, 0x5D20, 0x51FC, 0x5120, 0x71FE, 0x5100, 0x8FFE}},
    {18387, 2, {0x0, 0x7C00, 0x45FE, 0x4420, 0x4420, 0x7C20, 0x1020, 0x1020, 0x1020, 0x5C20, 0x5020, 0x5020, 0x5020, 0x5C20, 0xE0A0, 0x40}},
    {18388, 2, {0x40, 0x7E40, 0x4240, 0x4240, 0x4240, 0x7E60, 0x850, 0x848, 0x4E44, 0x4844, 0x4840, 0x4840, 0x4E40, 0xF040, 0x40, 0x40}},
    {18389, 2, {0x0, 0x7DF8, 0x4488, 0x44C8, 0x44A8, 0x7CA8, 0x1088, 0x1050, 0x5C50, 0x5050, 0x5020, 0x5020, 0x5C50, 0xE088, 0x104, 0x202}},
    {18390, 2, {0x20, 0x7C20, 0x4420, 0x4420, 0x4524, 0x7D24, 0x1124, 0x1124, 0x5D24, 0x5124, 0x5124, 0x5124, 0x5D24, 0xE1FC, 0x4, 0x0}},
    {18391, 2, {0x4, 0x7C1E, 0x45F0, 0x4510, 0x4510, 0x7D10, 0x1110, 0x11FE, 0x1110, 0x5D10, 0x5110, 0x5108, 0x510A, 0x5D4A, 0xE186, 0x102}},
    {18392, 2, {0x80, 0x7C40, 0x4440, 0x47FC, 0x4400, 0x7C00, 0x11F0, 0x1110, 0x5D10, 0x5110, 0x5110, 0x5112, 0x5D12, 0xE212, 0x20E, 0x400}},
    {18393, 2, {0x0, 0x7900, 0x497E, 0x4912, 0x4912, 0x7912, 0x11D2, 0x1712, 0x5112, 0x5D12, 0x5152, 0x5192, 0x5922, 0xE022, 0x4A, 0x84}},
    {18394, 2, {0x0, 0x7CFC, 0x4400, 0x4400, 0x4400, 0x7DFE, 0x1048, 0x1048, 0x5C48, 0x5048, 0x5048, 0x5048, 0x5C88, 0xE088, 0x108, 0x208}},
    {18395, 2, {0x20, 0x7C20, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x11FC, 0x1124, 0x5D24, 0x5124, 0x5124, 0x5134, 0x5D28, 0xE020, 0x20, 0x20}},
    {18396, 2, {0x20, 0x7C20, 0x4420, 0x45FE, 0x4522, 0x7D22, 0x1122, 0x1122, 0x1152, 0x5D4A, 0x518A, 0x5102, 0x5102, 0x5D02, 0xE10A, 0x104}},
    {18397, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7D06, 0x1200, 0x11FC, 0x1084, 0x5C84, 0x5048, 0x5050, 0x5020, 0x5C50, 0xE088, 0x306}},
    {18398, 2, {0x20, 0x7C20, 0x4420, 0x44A8, 0x44A4, 0x7CA2, 0x1122, 0x1120, 0x1224, 0x5C24, 0x5028, 0x5008, 0x5010, 0x5C20, 0xE0C0, 0x300}},
    {18399, 2, {0x200, 0x77C, 0x3844, 0x2044, 0x2044, 0x207C, 0x3F90, 0x2410, 0x2450, 0x2C5C, 0x2450, 0x2650, 0x2550, 0x44B0, 0x449E, 0x8500}},
    {18400, 2, {0x28, 0x7C24, 0x4424, 0x4420, 0x45FE, 0x7D20, 0x1124, 0x1124, 0x1124, 0x5D28, 0x5128, 0x5168, 0x5194, 0x5D14, 0xE02C, 0x44}},
    {18401, 2, {0x4, 0x7C1E, 0x45F0, 0x4510, 0x4510, 0x7D10, 0x1110, 0x11FE, 0x1110, 0x5D10, 0x5110, 0x5108, 0x510A, 0x5D4A, 0xE1A6, 0x112}},
    {18402, 2, {0x80, 0x7C80, 0x4480, 0x44FE, 0x4502, 0x7D04, 0x1220, 0x1020, 0x5CA8, 0x50A4, 0x5124, 0x5122, 0x5E22, 0xE020, 0xA0, 0x40}},
    {18403, 2, {0x20, 0x7C28, 0x4424, 0x4420, 0x45FC, 0x7C20, 0x1070, 0x1070, 0x5CA8, 0x50A8, 0x5124, 0x5124, 0x5E22, 0xE020, 0x20, 0x20}},
    {18404, 2, {0x40, 0x7C20, 0x4428, 0x4408, 0x4408, 0x7C48, 0x1054, 0x1152, 0x5D52, 0x5160, 0x5260, 0x5044, 0x5CC4, 0xE144, 0x23C, 0x0}},
    {18405, 2, {0x0, 0x7DF8, 0x4408, 0x4450, 0x4420, 0x7C10, 0x13FE, 0x1062, 0x5CA4, 0x50A0, 0x5120, 0x5220, 0x5C20, 0xE020, 0xA0, 0x40}},
    {18406, 2, {0x8, 0x7C3C, 0x45E0, 0x4420, 0x4420, 0x7DFE, 0x1040, 0x1040, 0x5C7C, 0x50A4, 0x50A8, 0x5128, 0x5D10, 0xE228, 0x444, 0x82}},
    {18407, 2, {0x0, 0x7C00, 0x45FC, 0x4524, 0x4524, 0x7D24, 0x1124, 0x1124, 0x11FC, 0x5D24, 0x5124, 0x5124, 0x5124, 0x5D24, 0xE1FC, 0x104}},
    {18408, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x7DFC, 0x1120, 0x1120, 0x11FE, 0x5D20, 0x5120, 0x5110, 0x5112, 0x5D4A, 0xE186, 0x102}},
    {18409, 2, {0x20, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x7C20, 0x1020, 0x13FE, 0x5C20, 0x5040, 0x5040, 0x5088, 0x5D04, 0xE3FE, 0x102, 0x0}},
    {18410, 2, {0x4, 0x7C1E, 0x45E0, 0x4400, 0x4420, 0x7C10, 0x11FE, 0x1004, 0x1008, 0x5C10, 0x5020, 0x5040, 0x5080, 0x5D40, 0xE23E, 0x0}},
    {18411, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x100, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {18412, 2, {0x8, 0x7C1C, 0x45E0, 0x4500, 0x4500, 0x7D00, 0x1100, 0x11FE, 0x5D10, 0x5110, 0x5110, 0x5110, 0x5D10, 0xE110, 0x7FE, 0x0}},
    {18413, 2, {0x80, 0x7C80, 0x4480, 0x44FE, 0x4540, 0x7D40, 0x1240, 0x107C, 0x5C40, 0x5040, 0x5040, 0x507C, 0x5C40, 0xE040, 0x40, 0x40}},
    {18414, 2, {0x90, 0x7C94, 0x4492, 0x4512, 0x4510, 0x7F7E, 0x1510, 0x1110, 0x1110, 0x5D28, 0x5128, 0x5128, 0x5128, 0x5D44, 0xE144, 0x182}},
    {18415, 2, {0x8, 0x7C3C, 0x45E0, 0x4420, 0x4420, 0x7DFE, 0x1020, 0x1020, 0x5C20, 0x51FC, 0x5104, 0x5104, 0x5D04, 0xE104, 0x1FC, 0x104}},
    {18416, 2, {0x40, 0x7C40, 0x449C, 0x4500, 0x4640, 0x7C40, 0x10BE, 0x1188, 0x1288, 0x5C88, 0x5088, 0x5088, 0x5088, 0x5C88, 0xE0A8, 0x90}},
    {18417, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x4420, 0x7DFC, 0x1124, 0x1124, 0x5D24, 0x512C, 0x5070, 0x50A8, 0x5D24, 0xE622, 0x20, 0x20}},
    {18418, 2, {0x40, 0x7C20, 0x4420, 0x47FE, 0x4440, 0x7C88, 0x1104, 0x13FE, 0x1092, 0x5C90, 0x5090, 0x5090, 0x5112, 0x5D12, 0xE20E, 0x400}},
    {18419, 2, {0x0, 0x3EEE, 0x2222, 0x2222, 0x22AA, 0x3E66, 0x822, 0x866, 0x28AA, 0x2E22, 0x2822, 0x28AA, 0x2844, 0x5800, 0x4FFE, 0x8000}},
    {18420, 2, {0x0, 0x7DFE, 0x4500, 0x4500, 0x457C, 0x7D00, 0x1100, 0x11FE, 0x5D50, 0x5152, 0x5154, 0x5148, 0x5D48, 0xE244, 0x252, 0x460}},
    {18421, 2, {0x28, 0x7C24, 0x4424, 0x4420, 0x45FE, 0x7C20, 0x1120, 0x10B2, 0x5CB4, 0x5068, 0x50A8, 0x5124, 0x5E22, 0xE020, 0xA0, 0x40}},
    {18422, 2, {0x20, 0x7C20, 0x44A8, 0x44A8, 0x44A8, 0x7D74, 0x1222, 0x1020, 0x1020, 0x5DFC, 0x5020, 0x5020, 0x5020, 0x5C20, 0xE3FE, 0x0}},
    {18423, 2, {0x104, 0x790E, 0x4930, 0x4920, 0x4FE0, 0x7920, 0x113E, 0x1164, 0x1DA4, 0x5724, 0x5124, 0x5124, 0x5124, 0x5D24, 0xE144, 0x384}},
    {18424, 2, {0x20, 0x7C20, 0x447C, 0x44C4, 0x4528, 0x7C90, 0x1060, 0x1180, 0x5C48, 0x5048, 0x51FE, 0x5048, 0x5C48, 0xE088, 0x88, 0x108}},
    {18425, 2, {0x4, 0x7C1E, 0x45E0, 0x4422, 0x4512, 0x7C94, 0x1080, 0x1008, 0x11FE, 0x5C08, 0x5088, 0x5048, 0x5048, 0x5C08, 0xE028, 0x10}},
    {18426, 2, {0x0, 0x7EFC, 0x4284, 0x4284, 0x42FC, 0x7E84, 0x884, 0x8FC, 0x4E84, 0x4884, 0x48FC, 0x4800, 0x4E48, 0xF044, 0x82, 0x102}},
    {18427, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x4504, 0x7E02, 0x10F8, 0x1020, 0x5C20, 0x53FE, 0x5020, 0x5124, 0x5D22, 0xE222, 0xA0, 0x40}},
    {18428, 2, {0x8, 0x7C0C, 0x44EA, 0x440A, 0x4408, 0x7DFE, 0x1048, 0x1048, 0x1148, 0x5D68, 0x5148, 0x5148, 0x514A, 0x5D6A, 0xE386, 0x102}},
    {18429, 2, {0x40, 0x7C20, 0x45FC, 0x4440, 0x4490, 0x7D08, 0x13FC, 0x1004, 0x5D50, 0x5150, 0x5150, 0x5150, 0x5E50, 0xE252, 0x452, 0x80E}},
    {18430, 2, {0x0, 0x7BDE, 0x4842, 0x494A, 0x4884, 0x794A, 0x1252, 0x1420, 0x1C00, 0x53DE, 0x5052, 0x5152, 0x5094, 0x5D48, 0xE254, 0x422}},
    {18431, 2, {0x0, 0x7DF8, 0x4408, 0x4408, 0x45F8, 0x7C08, 0x1008, 0x13FE, 0x5C20, 0x5222, 0x5174, 0x50A8, 0x5D24, 0xE222, 0xA0, 0x40}},
    {18432, 2, {0x20, 0x7D24, 0x44A4, 0x44A8, 0x4420, 0x7DFC, 0x1104, 0x1104, 0x1174, 0x5D54, 0x5154, 0x5154, 0x5174, 0x5D04, 0xE114, 0x108}},
    {18433, 2, {0x40, 0x7890, 0x4908, 0x4BFC, 0x4908, 0x7A10, 0x14A4, 0x17BE, 0x5082, 0x5C00, 0x53FC, 0x5000, 0x5800, 0xE000, 0x7FE, 0x0}},
    {18434, 2, {0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x1100, 0x2900, 0x47FE}},
    {18435, 2, {0x0, 0x7BFE, 0x4820, 0x4820, 0x4BFE, 0x7A22, 0x1222, 0x1332, 0x52AA, 0x5EAA, 0x5376, 0x5266, 0x5A22, 0xE222, 0x22A, 0x204}},
    {18436, 2, {0x50, 0x7C50, 0x4450, 0x45FC, 0x4554, 0x7D54, 0x1154, 0x11FC, 0x5D54, 0x5154, 0x5154, 0x53FE, 0x5C00, 0xE050, 0x88, 0x104}},
    {18437, 2, {0x0, 0x7DFE, 0x4502, 0x451A, 0x4562, 0x7D22, 0x1122, 0x11FA, 0x5D22, 0x5132, 0x516A, 0x51A6, 0x5D22, 0xE122, 0x1FE, 0x102}},
    {18438, 2, {0x0, 0x7CFC, 0x4480, 0x44F8, 0x4480, 0x7CF8, 0x1080, 0x13FE, 0x1140, 0x5D24, 0x5128, 0x5110, 0x5108, 0x5D44, 0xE182, 0x100}},
    {18439, 2, {0x0, 0x7DFE, 0x4502, 0x4502, 0x45FE, 0x7D10, 0x1152, 0x1152, 0x5D52, 0x517E, 0x5110, 0x5152, 0x5D52, 0xE252, 0x27E, 0x402}},
    {18440, 2, {0x20, 0x7820, 0x49FC, 0x4824, 0x4824, 0x7BFE, 0x1024, 0x1024, 0x51FC, 0x5E22, 0x5174, 0x50A8, 0x5924, 0xE222, 0xA0, 0x40}},
    {18441, 2, {0x20, 0x7C10, 0x45FE, 0x4502, 0x4604, 0x7CF8, 0x1088, 0x1088, 0x5CF8, 0x5080, 0x5080, 0x50FC, 0x5C84, 0xE084, 0xFC, 0x84}},
    {18442, 2, {0x90, 0x7C90, 0x4490, 0x479E, 0x4490, 0x7C90, 0x1090, 0x139C, 0x1090, 0x5C90, 0x5090, 0x539E, 0x5090, 0x5C90, 0xE090, 0x90}},
    {18443, 2, {0x40, 0x7C80, 0x45FC, 0x4524, 0x4524, 0x7DFC, 0x1124, 0x1144, 0x5DFC, 0x5090, 0x5110, 0x53FE, 0x5C10, 0xE010, 0x10, 0x10}},
    {18444, 2, {0x20, 0x7C22, 0x47B4, 0x44A8, 0x44A8, 0x7D24, 0x12A2, 0x1040, 0x5DFC, 0x5104, 0x51FC, 0x5104, 0x5DFC, 0xE104, 0x114, 0x108}},
    {18445, 2, {0x0, 0x7BDE, 0x4A42, 0x4A42, 0x4A42, 0x7BDE, 0x1200, 0x123E, 0x53D2, 0x5E12, 0x5214, 0x53D4, 0x5A08, 0xE214, 0x224, 0x242}},
    {18446, 2, {0x0, 0x7DFE, 0x4502, 0x4502, 0x45FE, 0x7D00, 0x11FE, 0x1120, 0x5D44, 0x51FE, 0x5112, 0x5110, 0x5EFE, 0xE210, 0x410, 0x9FE}},
    {18447, 2, {0x10, 0x7BD4, 0x4858, 0x4A52, 0x498C, 0x7888, 0x1104, 0x12FA, 0x1020, 0x5C20, 0x53FE, 0x5020, 0x5050, 0x5C88, 0xE104, 0x202}},
    {18448, 2, {0x0, 0x7DFC, 0x4504, 0x45FC, 0x4504, 0x7DFC, 0x1104, 0x11FC, 0x5C28, 0x5024, 0x53FE, 0x5050, 0x5C50, 0xE088, 0x104, 0x202}},
    {18449, 2, {0x20, 0x7820, 0x4BFE, 0x4850, 0x4888, 0x7B26, 0x11FC, 0x1020, 0x1C20, 0x53FE, 0x5000, 0x5020, 0x51FC, 0x5C20, 0xE020, 0x3FE}},
    {18450, 2, {0x80, 0x7C9E, 0x4492, 0x45D2, 0x4492, 0x7C9E, 0x1092, 0x11D2, 0x5D52, 0x515E, 0x5152, 0x5152, 0x5DD2, 0xE022, 0x2A, 0x44}},
    {18451, 2, {0x88, 0x7C48, 0x4450, 0x47FE, 0x4450, 0x7C50, 0x11FC, 0x1154, 0x5D54, 0x518C, 0x5104, 0x51FC, 0x5D04, 0xE104, 0x1FC, 0x104}},
    {18452, 2, {0x0, 0x7C84, 0x4448, 0x4400, 0x45FE, 0x7C20, 0x1044, 0x11A8, 0x1030, 0x5C58, 0x5198, 0x5034, 0x5052, 0x5D90, 0xE050, 0x20}},
    {18453, 2, {0x90, 0x7892, 0x4AD4, 0x4A98, 0x4A92, 0x7AD2, 0x170E, 0x1020, 0x1C20, 0x53FE, 0x5070, 0x50A8, 0x5124, 0x5E22, 0xE020, 0x20}},
    {18454, 2, {0x20, 0x7D20, 0x44BE, 0x4442, 0x4484, 0x7F10, 0x1110, 0x1128, 0x1144, 0x5C82, 0x50FC, 0x5084, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {18455, 2, {0x20, 0x7C40, 0x45FC, 0x4504, 0x45FC, 0x7D04, 0x11FC, 0x1104, 0x5DFC, 0x5028, 0x5024, 0x53FE, 0x5C50, 0xE088, 0x104, 0x202}},
    {18456, 2, {0x20, 0x7C40, 0x45FC, 0x4504, 0x4554, 0x7D24, 0x1154, 0x1104, 0x11FC, 0x5C00, 0x5112, 0x51D4, 0x5118, 0x5D52, 0xE192, 0x10E}},
    {18457, 2, {0x20, 0x7C10, 0x45FE, 0x4484, 0x4448, 0x7DFE, 0x1102, 0x1224, 0x1010, 0x5DFE, 0x5040, 0x507C, 0x5044, 0x5C84, 0xE094, 0x108}},
    {18458, 2, {0x80, 0xF040, 0x97FC, 0x9404, 0x9110, 0xF208, 0x203C, 0x23D0, 0xA290, 0xBA90, 0xA290, 0xA288, 0xA288, 0xB4A4, 0xC4D4, 0x892}},
    {18459, 2, {0x8, 0x783C, 0x4BC0, 0x4804, 0x4A44, 0x7928, 0x11FC, 0x1220, 0x1C20, 0x53FE, 0x5020, 0x5124, 0x5124, 0x5D24, 0xE1FC, 0x4}},
    {18460, 2, {0x88, 0x7C88, 0x45FE, 0x4488, 0x4400, 0x7DFE, 0x1088, 0x1088, 0x10F8, 0x5C88, 0x50F8, 0x5088, 0x509E, 0x5DE8, 0xE008, 0x8}},
    {18461, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x44F8, 0x7C00, 0x11FC, 0x1104, 0x5DFC, 0x5104, 0x51FC, 0x5104, 0x5DFC, 0xE050, 0x88, 0x104}},
    {18462, 2, {0x14, 0xF012, 0x9010, 0x97FE, 0x9410, 0xF490, 0x24D2, 0x2492, 0xA7F2, 0xBC94, 0xA494, 0xA5C8, 0xBAAA, 0xE89A, 0x11A6, 0x42}},
    {18463, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x882, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {18464, 2, {0x1450, 0x7FFC, 0x2452, 0x47CE, 0x8000, 0x7FFC, 0x4104, 0x3FF8, 0x2108, 0x3FF8, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {18465, 2, {0x20, 0x7C3C, 0x4420, 0x45FE, 0x4522, 0x7D38, 0x11E4, 0x111C, 0x1100, 0x5D7C, 0x5144, 0x517C, 0x5144, 0x5D7C, 0xE144, 0x2FE}},
    {18466, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {18467, 2, {0x0, 0x7BDE, 0x4A52, 0x4BDE, 0x4A10, 0x7A52, 0x11CE, 0x1000, 0x1C88, 0x53FE, 0x5088, 0x5088, 0x53FE, 0x5C88, 0xE104, 0x202}},
    {18468, 2, {0x0, 0x7DDC, 0x4554, 0x4554, 0x45DC, 0x7C00, 0x11FC, 0x1124, 0x11FC, 0x5D24, 0x51FC, 0x5020, 0x53FE, 0x5C20, 0xE020, 0x20}},
    {18469, 2, {0x108, 0xF108, 0x9548, 0x9390, 0x911E, 0xF7D4, 0x2564, 0x2554, 0xA554, 0xBF54, 0xA5D4, 0xA548, 0xA548, 0xB554, 0xC464, 0x4C2}},
    {18470, 2, {0x80, 0x7CEE, 0x454A, 0x444A, 0x45FA, 0x7C4E, 0x10AA, 0x1100, 0x10FC, 0x5C84, 0x5084, 0x50FC, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {18471, 2, {0x0, 0x7BFC, 0x4A04, 0x4A04, 0x4BFC, 0x7A22, 0x12FC, 0x1228, 0x13FE, 0x5E20, 0x52FC, 0x5384, 0x52FC, 0x5C84, 0xE4FC, 0x884}},
    {18472, 2, {0x80, 0x7CF8, 0x4508, 0x47FE, 0x4544, 0x7D92, 0x11FE, 0x1100, 0x5D7C, 0x5100, 0x517C, 0x5100, 0x5D7C, 0xE244, 0x27C, 0x444}},
    {18473, 2, {0x20, 0xF13C, 0x9120, 0x97FE, 0x9010, 0xF014, 0x23FE, 0x2210, 0xA3F0, 0xBA54, 0xA254, 0xA2D4, 0xBB68, 0xE44A, 0x896, 0x322}},
    {18474, 2, {0x44, 0x7A84, 0x4AEA, 0x4A4E, 0x4AA4, 0x7AEE, 0x1222, 0x13FE, 0x1244, 0x5E84, 0x52EA, 0x524E, 0x52A4, 0x5EEE, 0xE222, 0x3FE}},
    {18475, 2, {0x3F0, 0x792E, 0x49E2, 0x492A, 0x49E4, 0x7934, 0x13EA, 0x1032, 0x1C1E, 0x53E0, 0x50A2, 0x5134, 0x5068, 0x5CA4, 0xE322, 0x20}},
    {18476, 2, {0xA0, 0xF74E, 0x915A, 0x952A, 0x922A, 0xF21C, 0x25DA, 0x280A, 0xA1EA, 0xB92A, 0xA1EA, 0xA22C, 0xA148, 0xB068, 0xC388, 0x8}},
    {18477, 2, {0x108, 0xF108, 0x97CE, 0x9112, 0x9124, 0xF7DE, 0x2552, 0x255E, 0xA7D2, 0xB91E, 0xA392, 0xA55E, 0xA940, 0xB114, 0xC112, 0x122}},
    {18478, 2, {0x124, 0x78A8, 0x49FC, 0x4840, 0x4BFE, 0x7908, 0x13FC, 0x1522, 0x51FC, 0x5D20, 0x51FC, 0x5120, 0x59FE, 0xE002, 0x2AA, 0x4}},
    {18479, 2, {0x40, 0xF0A0, 0x9110, 0x9208, 0x9DF6, 0xF000, 0x2EEE, 0x2AAA, 0xAEEE, 0xB800, 0xAFFE, 0xA922, 0xAFFE, 0xB922, 0xC92A, 0x804}},
    {18480, 2, {0x88, 0x7BFE, 0x4888, 0x4BDE, 0x4A52, 0x7BDE, 0x10A0, 0x1090, 0x11FE, 0x5D20, 0x53FC, 0x5520, 0x51FC, 0x5D20, 0xE1FE, 0x100}},
    {18481, 2, {0x3FE, 0x7A02, 0x4BFE, 0x4A92, 0x4A54, 0x7A92, 0x12FE, 0x12AA, 0x52FE, 0x5F22, 0x52FA, 0x52AA, 0x5AFA, 0xE222, 0x5FA, 0x4}},
    {18482, 2, {0x1020, 0x2020, 0x7C50, 0x4450, 0x7C88, 0x4524, 0x7E12, 0x4410, 0x45FC, 0xFC04, 0xC08, 0x1488, 0x2450, 0x4420, 0x9410, 0x810}},
    {18483, 2, {0x1020, 0x2020, 0x7D24, 0x4524, 0x7D24, 0x4524, 0x7DFC, 0x4420, 0x4420, 0xFD24, 0xD24, 0x1524, 0x2524, 0x4524, 0x95FC, 0x804}},
    {18484, 2, {0x1020, 0x2010, 0x7C10, 0x4400, 0x7DFE, 0x4400, 0x7C04, 0x4484, 0x4484, 0xFC48, 0xC48, 0x1448, 0x2450, 0x4410, 0x95FE, 0x800}},
    {18485, 2, {0x1044, 0x2044, 0x7C44, 0x4484, 0x7CBE, 0x4584, 0x7E84, 0x44A4, 0x4494, 0xFC94, 0xC84, 0x1484, 0x2484, 0x4484, 0x9494, 0x888}},
    {18486, 2, {0x1040, 0x2040, 0x7C80, 0x44FC, 0x7D20, 0x4620, 0x7C20, 0x4420, 0x47FE, 0xFC20, 0xC50, 0x1450, 0x2488, 0x4488, 0x9504, 0xA02}},
    {18487, 2, {0x1050, 0x2050, 0x7C50, 0x4552, 0x7CD4, 0x4458, 0x7C50, 0x4458, 0x44D4, 0xFD52, 0xC50, 0x1450, 0x2492, 0x4492, 0x9512, 0xA0E}},
    {18488, 2, {0x1020, 0x2020, 0x7DFC, 0x4450, 0x7C88, 0x4504, 0x7EFA, 0x4400, 0x47FE, 0xFC40, 0xC80, 0x15FC, 0x2404, 0x4404, 0x9428, 0x810}},
    {18489, 2, {0x1020, 0x2010, 0x7C10, 0x45FE, 0x7C20, 0x4424, 0x7C44, 0x44F8, 0x4412, 0xFC22, 0xC44, 0x1588, 0x2410, 0x4428, 0x9444, 0x982}},
    {18490, 2, {0x1080, 0x2040, 0x7C5E, 0x4502, 0x7D02, 0x457A, 0x7D4A, 0x454A, 0x454A, 0xFD4A, 0xD7A, 0x154A, 0x2502, 0x4502, 0x950A, 0x904}},
    {18491, 2, {0x1020, 0x2010, 0x7DFE, 0x4502, 0x7C48, 0x4484, 0x7D02, 0x44FC, 0x4404, 0xFCFC, 0xC80, 0x14FC, 0x2404, 0x4404, 0x9428, 0x810}},
    {18492, 2, {0x1020, 0x2050, 0x7C88, 0x4504, 0x7EFA, 0x4400, 0x7DE2, 0x452A, 0x452A, 0xFDEA, 0xD2A, 0x152A, 0x25EA, 0x4522, 0x952A, 0x964}},
    {18493, 2, {0x1020, 0x2124, 0x7924, 0x4924, 0x79FC, 0x4800, 0x7BFE, 0x4800, 0x49FC, 0xF904, 0x1904, 0x29FC, 0x4888, 0x8850, 0x2BFE, 0x1000}},
    {18494, 2, {0x1000, 0x23DE, 0x7A52, 0x4BDE, 0x7A52, 0x4BDE, 0x7A02, 0x4A22, 0x4A22, 0xFA22, 0x1A52, 0x2A4A, 0x4A8A, 0x8A02, 0x2A0A, 0x1204}},
    {18495, 2, {0x1020, 0x21FC, 0x7D24, 0x45FC, 0x7C20, 0x47FE, 0x7C00, 0x45FC, 0x4504, 0xFDFC, 0xD04, 0x15FC, 0x2504, 0x45FC, 0x9488, 0x904}},
    {18496, 2, {0x1010, 0x21FE, 0x7C00, 0x45FE, 0x7D02, 0x457A, 0x7D4A, 0x45FE, 0x4400, 0xFCFC, 0xC84, 0x14FC, 0x2484, 0x44FC, 0x9400, 0x9FE}},
    {18497, 2, {0x1110, 0x2090, 0x7BDE, 0x4810, 0x7A5E, 0x4982, 0x7BDE, 0x4810, 0x4BDE, 0xFA50, 0x1BDE, 0x2A50, 0x4BDE, 0x8A50, 0x2A52, 0x12CE}},
    {18498, 2, {0x804, 0x804, 0x844, 0xFF44, 0x844, 0x7F44, 0x4944, 0x7F44, 0x4944, 0x7F4C, 0x874, 0xFF44, 0x804, 0x804, 0x804, 0x804}},
    {18499, 2, {0x1040, 0x1040, 0x1040, 0xFE40, 0x11FC, 0x7C44, 0x5444, 0x7C44, 0x5444, 0x7C84, 0x1084, 0xFE84, 0x1104, 0x1104, 0x1228, 0x1410}},
    {18500, 2, {0x1000, 0x10F0, 0x1090, 0xFE90, 0x1090, 0x7C90, 0x5490, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE92, 0x1092, 0x1112, 0x110E, 0x1200}},
    {18501, 2, {0x1082, 0x1092, 0x1092, 0xFE92, 0x1092, 0x7C92, 0x5492, 0x7C92, 0x5492, 0x7C92, 0x1092, 0xFE92, 0x1092, 0x1112, 0x1102, 0x1202}},
    {18502, 2, {0x1010, 0x1090, 0x1090, 0xFE88, 0x1088, 0x7D24, 0x5524, 0x7E22, 0x5420, 0x7C40, 0x1048, 0xFE44, 0x1084, 0x11FE, 0x1082, 0x1000}},
    {18503, 2, {0x1000, 0x1000, 0x11FC, 0xFE00, 0x1000, 0x7C00, 0x57FE, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE90, 0x1112, 0x1112, 0x120E, 0x1400}},
    {18504, 2, {0x1008, 0x1088, 0x1050, 0xFE20, 0x1050, 0x7C88, 0x5500, 0x7C08, 0x5488, 0x7C88, 0x1050, 0xFE50, 0x1020, 0x1050, 0x1088, 0x1306}},
    {18505, 2, {0x1040, 0x1020, 0x1000, 0xFDFC, 0x1000, 0x7C00, 0x54F0, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE92, 0x1092, 0x1112, 0x110E, 0x1200}},
    {18506, 2, {0x1008, 0x101C, 0x11E0, 0xFD00, 0x1100, 0x7DFC, 0x5544, 0x7D44, 0x5544, 0x7D28, 0x1128, 0xFF10, 0x1110, 0x1228, 0x1244, 0x1482}},
    {18507, 2, {0x1020, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x7C20, 0x5420, 0x7DFC, 0x5484, 0x7C88, 0x1048, 0xFE50, 0x1020, 0x1050, 0x1188, 0x1606}},
    {18508, 2, {0x1040, 0x1040, 0x1040, 0xFC40, 0x13FE, 0x7C40, 0x5480, 0x7C90, 0x5490, 0x7D20, 0x1120, 0xFE48, 0x1244, 0x1484, 0x11FE, 0x1082}},
    {18509, 2, {0x1020, 0x1020, 0x1020, 0xFC20, 0x11FC, 0x7C20, 0x5420, 0x7C20, 0x55FE, 0x7C20, 0x1050, 0xFE50, 0x1088, 0x1088, 0x1104, 0x1202}},
    {18510, 2, {0x1020, 0x1010, 0x1010, 0xFEFE, 0x1082, 0x7C82, 0x5482, 0x7CFE, 0x5482, 0x7C80, 0x1080, 0xFE80, 0x1080, 0x1100, 0x1100, 0x1200}},
    {18511, 2, {0x1008, 0x11C8, 0x1048, 0xFE48, 0x1048, 0x7DC8, 0x5508, 0x7D08, 0x5510, 0x7DD0, 0x1054, 0xFE54, 0x1052, 0x105E, 0x1282, 0x1100}},
    {18512, 2, {0x1020, 0x1010, 0x1010, 0xFEFE, 0x1082, 0x7D04, 0x5440, 0x7C48, 0x5450, 0x7C60, 0x1040, 0xFE42, 0x1042, 0x1042, 0x103E, 0x1000}},
    {18513, 2, {0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x7DFC, 0x5520, 0x7D20, 0x55FE, 0x7D20, 0x1120, 0xFD10, 0x1112, 0x114A, 0x1186, 0x1102}},
    {18514, 2, {0x1000, 0x1000, 0x11FC, 0xFD24, 0x1124, 0x7D24, 0x5524, 0x7D24, 0x55FC, 0x7D24, 0x1124, 0xFD24, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {18515, 2, {0x1100, 0x1100, 0x211C, 0x47C0, 0x8100, 0x17C0, 0x257E, 0x67C8, 0xA548, 0x27C8, 0x2108, 0x2FE8, 0x2108, 0x2108, 0x2128, 0x2110}},
    {18516, 2, {0x1020, 0x1020, 0x107C, 0xFC84, 0x1148, 0x7C30, 0x5420, 0x7C48, 0x5590, 0x7C3E, 0x1042, 0xFDA4, 0x1018, 0x1010, 0x1060, 0x1180}},
    {18517, 2, {0x1000, 0x11F8, 0x1010, 0xFC20, 0x1024, 0x7DA8, 0x54B0, 0x7CA8, 0x5528, 0x7D24, 0x1222, 0xFCA0, 0x1040, 0x1000, 0x11FC, 0x1000}},
    {18518, 2, {0x1000, 0x1000, 0x11FC, 0xFE00, 0x1050, 0x7C88, 0x5524, 0x7C20, 0x5420, 0x7DFC, 0x1020, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {18519, 2, {0x7DF0, 0x1110, 0x1190, 0x1D52, 0xF20E, 0x500, 0x1FF0, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {18520, 2, {0x1080, 0x1080, 0x10FE, 0xFD02, 0x1102, 0x7EFA, 0x54AA, 0x7CAA, 0x54FA, 0x7CAA, 0x10AA, 0xFEFA, 0x108A, 0x1002, 0x1014, 0x1008}},
    {18521, 2, {0x1020, 0x1010, 0x10FC, 0xFE84, 0x1084, 0x7CFC, 0x5484, 0x7C84, 0x54FC, 0x7CA2, 0x10A4, 0xFE98, 0x1090, 0x1088, 0x10C4, 0x1082}},
    {18522, 2, {0x1020, 0x1040, 0x11FC, 0xFD04, 0x1104, 0x7DFC, 0x5504, 0x7D04, 0x55FC, 0x7C50, 0x1050, 0xFE50, 0x1092, 0x1092, 0x110E, 0x1200}},
    {18523, 2, {0x1010, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x5454, 0x5454, 0x7C7C, 0x5454, 0x5454, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x1010, 0x1010}},
    {18524, 2, {0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x7DFE, 0x5488, 0x7D04, 0x5622, 0x7C20, 0x11FC, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {18525, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0x2504, 0xF900, 0xA9DE, 0xFA52, 0xAA52, 0xFB52, 0x249A, 0xF894, 0x2110, 0x2112, 0x2212, 0x240E}},
    {18526, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x1110, 0x3FF8, 0xC106, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {18527, 2, {0x1020, 0x1040, 0x10FE, 0xFC92, 0x1092, 0x7CFE, 0x5492, 0x7CA2, 0x54FE, 0x7C48, 0x1088, 0xFDFE, 0x1008, 0x1008, 0x1008, 0x1008}},
    {18528, 2, {0x1020, 0x1020, 0x103E, 0xFE20, 0x1020, 0x7CFC, 0x5484, 0x7CFC, 0x5484, 0x7CFC, 0x1084, 0xFEFC, 0x1000, 0x1048, 0x1084, 0x1102}},
    {18529, 2, {0x1110, 0x1112, 0x11D4, 0xFD18, 0x1152, 0x7D92, 0x552E, 0x7C40, 0x55FC, 0x7D04, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {18530, 2, {0x1040, 0x1020, 0x11FE, 0xFD02, 0x1102, 0x7DFE, 0x5500, 0x7D00, 0x55FE, 0x7DAA, 0x11AA, 0xFEFE, 0x12AA, 0x12AA, 0x14A2, 0x1086}},
    {18531, 2, {0x1000, 0x13FE, 0x1222, 0xFC20, 0x11FE, 0x7C20, 0x55FC, 0x7D24, 0x55FC, 0x7D24, 0x11FC, 0xFC20, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {18532, 2, {0x200, 0x3FD0, 0x260, 0xFFFE, 0x1C10, 0xEFF0, 0x810, 0xFF0, 0x100, 0x7FFC, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100, 0x100}},
    {18533, 2, {0x1040, 0x1020, 0x13FE, 0xFE02, 0x1050, 0x7C88, 0x5524, 0x7C50, 0x5488, 0x7D04, 0x12FA, 0xFC88, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {18534, 2, {0x21F8, 0x2088, 0x2070, 0xF98C, 0x2000, 0xFBDE, 0xAA52, 0xF98C, 0xAA52, 0xF820, 0x23FE, 0xF870, 0x20A8, 0x2124, 0x2622, 0x2020}},
    {18535, 2, {0x1040, 0x1080, 0x11FC, 0xFD04, 0x1104, 0x7DFC, 0x5500, 0x7DFE, 0x5500, 0x7DFE, 0x1002, 0xFEAA, 0x12AA, 0x1402, 0x1014, 0x1008}},
    {18536, 2, {0x820, 0x440, 0x7FFC, 0x100, 0xFFFE, 0x400, 0x1FF0, 0xE100, 0x3FFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {18537, 2, {0x1020, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x7D04, 0x55FC, 0x7D04, 0x55FC, 0x7D04, 0x11FC, 0xFD04, 0x13FE, 0x1088, 0x1104, 0x1202}},
    {18538, 2, {0x1020, 0x1010, 0x11FE, 0xFD02, 0x1102, 0x7DFE, 0x5500, 0x7DEE, 0x5522, 0x7DAA, 0x1166, 0xFD22, 0x1166, 0x12AA, 0x1222, 0x1466}},
    {18539, 2, {0x1020, 0x1220, 0x117C, 0xFCA4, 0x1018, 0x7CE6, 0x5710, 0x7D7E, 0x5510, 0x7D7C, 0x1110, 0xFDFE, 0x1110, 0x1110, 0x12FE, 0x1400}},
    {18540, 2, {0x2000, 0x21F8, 0x2108, 0xF9F8, 0x2108, 0xF9F8, 0xA800, 0xFBFC, 0xAA94, 0xFBFC, 0x2000, 0xF9F8, 0x2090, 0x2060, 0x2198, 0x2606}},
    {18541, 2, {0x1084, 0x1048, 0x11FE, 0xFE10, 0x10FC, 0x7C20, 0x55FE, 0x7C40, 0x547C, 0x7CA4, 0x1124, 0xFEFC, 0x1044, 0x1044, 0x11FE, 0x1000}},
    {18542, 2, {0x1124, 0x1124, 0x1224, 0xFCA4, 0x1154, 0x7D4A, 0x5792, 0x7D08, 0x5508, 0x7D28, 0x112E, 0xFD28, 0x1128, 0x1158, 0x114E, 0x1180}},
    {18543, 2, {0x1020, 0x1040, 0x11FC, 0xFD24, 0x117C, 0x7D8C, 0x5554, 0x7D24, 0x5554, 0x7DFC, 0x1040, 0xFC24, 0x12A2, 0x128A, 0x1488, 0x1078}},
    {18544, 2, {0x878, 0xFF48, 0x886, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x8104, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {18545, 2, {0x2000, 0x23FE, 0x2200, 0xFA7C, 0x2244, 0xFA44, 0xAA7C, 0xFA00, 0xAAEE, 0xFAAA, 0x22AA, 0xFAAA, 0x22EE, 0x2200, 0x23FE, 0x2000}},
    {18546, 2, {0x1000, 0x13DE, 0x1252, 0xFE52, 0x13DE, 0x7C00, 0x55FC, 0x7D24, 0x55FC, 0x7D24, 0x11FC, 0xFC20, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {18547, 2, {0x2140, 0x2144, 0x2158, 0xFBF0, 0x2150, 0xF9D0, 0xA95E, 0xF9D4, 0xA954, 0xF954, 0x23F4, 0xF814, 0x2154, 0x2234, 0x2414, 0x2024}},
    {18548, 2, {0x1020, 0x11FC, 0x1088, 0xFC50, 0x13FE, 0x7C00, 0x55FC, 0x7D24, 0x55FC, 0x7D24, 0x11FC, 0xFC20, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {18549, 2, {0x1044, 0x1224, 0x1128, 0xFC7E, 0x1010, 0x7C28, 0x574A, 0x7D1C, 0x5528, 0x7D4C, 0x111A, 0xFD2A, 0x1148, 0x1110, 0x12FE, 0x1400}},
    {18550, 2, {0x2010, 0x23C8, 0x227E, 0xFA40, 0x2262, 0xFBD4, 0xAA00, 0xFA3E, 0xABC8, 0xFA48, 0x267E, 0xFA48, 0x2248, 0x23C8, 0x2248, 0x2008}},
    {18551, 2, {0x1088, 0x13FE, 0x1088, 0xFC20, 0x11FC, 0x7C20, 0x57FE, 0x7C40, 0x5488, 0x7DFC, 0x1000, 0xFDFC, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {18552, 2, {0x1088, 0x13FE, 0x1088, 0xFDFC, 0x1154, 0x7D54, 0x55FC, 0x7C28, 0x5424, 0x7DFE, 0x1120, 0xFD24, 0x1114, 0x114A, 0x1196, 0x1122}},
    {18553, 2, {0x1020, 0x11FE, 0x1020, 0xFDFE, 0x1102, 0x7EFC, 0x54A4, 0x7CFC, 0x54A4, 0x7CFC, 0x1000, 0xFDFE, 0x10A2, 0x10BC, 0x1160, 0x123E}},
    {18554, 2, {0x2088, 0x23DE, 0x2088, 0xF9DC, 0x2088, 0xFBDE, 0xA888, 0xFBFC, 0xA804, 0xF9FC, 0x2004, 0xFBFC, 0x2040, 0x22A4, 0x228A, 0x247A}},
    {18555, 2, {0x2040, 0x21FC, 0x2124, 0xF994, 0x214C, 0xF924, 0xA9FC, 0xF820, 0xABFE, 0xFA8A, 0x2124, 0xFBFE, 0x2080, 0x20FC, 0x2104, 0x220C}},
    {18556, 2, {0x2000, 0x21FC, 0x2020, 0xFBFE, 0x2222, 0xF9AC, 0xA820, 0xF9AC, 0xA800, 0xFBEE, 0x22AA, 0xFAAA, 0x22AA, 0x22AA, 0x23EE, 0x22AA}},
    {18557, 2, {0x2528, 0x27BE, 0x2948, 0xFFBE, 0x2318, 0xFDAA, 0xA946, 0xFBFC, 0xAA04, 0xFBFC, 0x2204, 0xFBFC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {18558, 2, {0x2108, 0x21EC, 0x210A, 0xFFEA, 0x2928, 0xFBC8, 0xA93E, 0xFFE8, 0xA948, 0xFBE8, 0x2A28, 0xFB68, 0x2AA8, 0x2BF4, 0x2AB4, 0x32A2}},
    {18559, 2, {0x21FC, 0x2020, 0x23FE, 0xFA22, 0x21AC, 0xF820, 0xA9AC, 0xF800, 0xABBE, 0xFAAA, 0x23BE, 0xF800, 0x23FE, 0x2124, 0x22AA, 0x27FE}},
    {18560, 2, {0x1040, 0x1040, 0x1040, 0xFEFE, 0x2080, 0x2900, 0x48F8, 0x7E08, 0x810, 0x820, 0xE40, 0xF880, 0x4902, 0x902, 0x8FE, 0x800}},
    {18561, 2, {0x1000, 0x10FC, 0x1084, 0xFE84, 0x2084, 0x28FC, 0x4884, 0x7E84, 0x884, 0x8FC, 0xE84, 0xF884, 0x4884, 0x884, 0x914, 0xA08}},
    {18562, 2, {0x1040, 0x1020, 0x1020, 0xFE00, 0x21FE, 0x2800, 0x4804, 0x7E84, 0x884, 0x848, 0xE48, 0xF848, 0x4850, 0x810, 0xBFE, 0x800}},
    {18563, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x1008, 0x810, 0x660, 0x180, 0xE70, 0xF00E}},
    {18564, 2, {0x1010, 0x24FE, 0x4244, 0x7E28, 0xFE, 0x7E10, 0x427C, 0x7E10, 0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820}},
    {18565, 2, {0x200, 0x1C78, 0x1008, 0x1E78, 0x1008, 0x1FF8, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2888, 0x4850, 0x4A30, 0x8C0E}},
    {18566, 2, {0x0, 0x1FE, 0x7D00, 0x457C, 0x4500, 0x1FE, 0x3954, 0x2948, 0x2964, 0x2942, 0x2B08, 0x2DFE, 0x4948, 0x4128, 0x8208, 0x418}},
    {18567, 2, {0x2000, 0x11FE, 0x7D00, 0x457C, 0x4500, 0x1FE, 0x3954, 0x2948, 0x2964, 0x2942, 0x2B08, 0x2DFE, 0x4948, 0x4128, 0x8208, 0x418}},
    {18568, 2, {0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2868, 0x4910, 0x4A88, 0x8C64, 0x3FF8, 0xD556, 0x1390, 0x1FF0, 0xA20, 0x920, 0xFE0}},
    {18569, 2, {0x1118, 0x27C8, 0x2448, 0x36D8, 0x2548, 0x26C8, 0x37D8, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x4430, 0x860E}},
    {18570, 2, {0x0, 0x23F8, 0x1010, 0x1020, 0x40, 0x40, 0xF7FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080, 0x2800, 0x47FE, 0x0}},
    {18571, 2, {0x0, 0x23FC, 0x1090, 0x1090, 0x90, 0x90, 0xF090, 0x1090, 0x1090, 0x1110, 0x1110, 0x1210, 0x1410, 0x2800, 0x47FE, 0x0}},
    {18572, 2, {0x40, 0x4040, 0x2040, 0x2FFE, 0x40, 0x40, 0xE7FC, 0x2444, 0x2444, 0x2444, 0x2454, 0x2448, 0x2040, 0x2040, 0x5000, 0x8FFE}},
    {18573, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x100, 0x100, 0xF1F8, 0x1108, 0x1108, 0x1108, 0x1208, 0x1208, 0x1450, 0x2820, 0x47FE, 0x0}},
    {18574, 2, {0x0, 0x23FC, 0x1090, 0x1090, 0x90, 0x90, 0xF7FE, 0x1090, 0x1090, 0x1110, 0x1110, 0x1210, 0x1410, 0x2800, 0x47FE, 0x0}},
    {18575, 2, {0x10, 0x2110, 0x1090, 0x1090, 0x10, 0x210, 0xF110, 0x1110, 0x101E, 0x17F0, 0x1010, 0x1010, 0x1010, 0x2810, 0x47FE, 0x0}},
    {18576, 2, {0x0, 0x21F8, 0x1108, 0x1108, 0x1F8, 0x108, 0xF108, 0x11F8, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000, 0x2800, 0x47FE, 0x0}},
    {18577, 2, {0x8, 0x203C, 0x13E0, 0x1220, 0x220, 0x220, 0xF3FE, 0x1220, 0x1210, 0x1212, 0x128A, 0x1326, 0x1212, 0x2800, 0x47FE, 0x0}},
    {18578, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x110, 0x208, 0xF414, 0x1110, 0x10A0, 0x1040, 0x10A0, 0x1110, 0x1208, 0x2800, 0x47FE, 0x0}},
    {18579, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x404, 0x1FC, 0xF020, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x10A0, 0x2840, 0x47FE, 0x0}},
    {18580, 2, {0x40, 0x20A0, 0x1110, 0x1208, 0xDF6, 0x0, 0xF000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1000, 0x2800, 0x47FE, 0x0}},
    {18581, 2, {0x28, 0x2024, 0x1024, 0x1020, 0x3FE, 0x220, 0xF224, 0x1224, 0x13A8, 0x1228, 0x1212, 0x121A, 0x1426, 0x1442, 0x2880, 0x47FE}},
    {18582, 2, {0x40, 0x2040, 0x13F8, 0x1048, 0x7FE, 0x48, 0xF3F8, 0x1040, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x1040, 0x2840, 0x47FE, 0x0}},
    {18583, 2, {0x120, 0x4120, 0x2120, 0x2FFC, 0x924, 0x924, 0xE924, 0x2FFC, 0x2924, 0x2924, 0x2924, 0x2FFC, 0x2000, 0x5000, 0x8FFE, 0x0}},
    {18584, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x414, 0x10, 0xF3FE, 0x1010, 0x1110, 0x1090, 0x1090, 0x1010, 0x1010, 0x1050, 0x2820, 0x47FE}},
    {18585, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x3FC, 0x204, 0xF204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1000, 0x1108, 0x1204, 0x2800, 0x47FE}},
    {18586, 2, {0x0, 0x4FFE, 0x2040, 0x27FC, 0x444, 0x7FC, 0xE444, 0x27FC, 0x2040, 0x2740, 0x20C0, 0x20B0, 0x2108, 0x2204, 0x5000, 0x8FFE}},
    {18587, 2, {0x0, 0x23FC, 0x1204, 0x1294, 0x30C, 0x204, 0xF2F4, 0x1294, 0x1294, 0x12F4, 0x1204, 0x1214, 0x1208, 0x2800, 0x47FE, 0x0}},
    {18588, 2, {0x3F8, 0x2008, 0x11F8, 0x1008, 0x3F8, 0x0, 0xF7FC, 0x1444, 0x13F8, 0x1248, 0x1248, 0x1268, 0x1250, 0x2840, 0x47FE, 0x0}},
    {18589, 2, {0x0, 0x27FE, 0x1090, 0x1090, 0x39C, 0x204, 0xF204, 0x139C, 0x1090, 0x1090, 0x1090, 0x17FE, 0x1000, 0x2800, 0x47FE, 0x0}},
    {18590, 2, {0x210, 0x4210, 0x2210, 0x2F7C, 0x210, 0x630, 0xE738, 0x2AD4, 0x2A52, 0x3292, 0x2310, 0x2210, 0x2210, 0x5000, 0x8FFE, 0x0}},
    {18591, 2, {0x420, 0x4220, 0x2020, 0x2FBE, 0x448, 0x488, 0xE728, 0x2528, 0x2510, 0x2910, 0x2928, 0x3544, 0x2284, 0x5000, 0x8FFE, 0x0}},
    {18592, 2, {0x40, 0x4244, 0x2244, 0x24A8, 0x110, 0x208, 0xEC06, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2210, 0x2210, 0x5410, 0x8FFE, 0x0}},
    {18593, 2, {0x290, 0x4290, 0x2FFC, 0x2290, 0x2F0, 0x200, 0xE3F8, 0x2040, 0x2040, 0x27FC, 0x2150, 0x2248, 0x2444, 0x5040, 0x8FFE, 0x0}},
    {18594, 2, {0x40, 0x40A0, 0x2110, 0x2288, 0xC46, 0x3F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2200, 0x25FC, 0x2504, 0x29FC, 0x5000, 0x8FFE}},
    {18595, 2, {0x20, 0x20FE, 0x134C, 0x1070, 0x7A0, 0x1FE, 0xF000, 0x11FC, 0x1000, 0x11FC, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x2800, 0x47FE}},
    {18596, 2, {0x390, 0x4E18, 0x2214, 0x2210, 0xF7E, 0x610, 0xE738, 0x2AB4, 0x2A54, 0x3250, 0x2290, 0x2210, 0x2210, 0x5000, 0x8FFE, 0x0}},
    {18597, 2, {0x0, 0x27FC, 0x1040, 0x1278, 0x240, 0xFFE, 0xF000, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE}},
    {18598, 2, {0x80, 0x2040, 0x17FC, 0x1040, 0x90, 0x5E4, 0xF248, 0x14A4, 0x11F0, 0x1010, 0x1040, 0x17FC, 0x1040, 0x1040, 0x2840, 0x47FE}},
    {18599, 2, {0x0, 0x27BC, 0x1084, 0x14A4, 0x294, 0x4A4, 0xF050, 0x1188, 0x1626, 0x10C0, 0x1310, 0x1064, 0x1388, 0x1030, 0x28C0, 0x47FE}},
    {18600, 2, {0x224, 0x4224, 0x2424, 0x2924, 0x154, 0x24A, 0xE692, 0x2A10, 0x2250, 0x225E, 0x2250, 0x2250, 0x22B0, 0x229E, 0x5300, 0x8FFE}},
    {18601, 2, {0x10, 0x5F90, 0x2210, 0x221E, 0xFA4, 0x4C4, 0xE494, 0x3FD4, 0x2008, 0x2F88, 0x2894, 0x28A2, 0x2FC2, 0x5000, 0x8FFE, 0x0}},
    {18602, 2, {0x78, 0x4780, 0x2080, 0x2FFC, 0x120, 0x210, 0xE5E8, 0x2926, 0x27F8, 0x2408, 0x25E8, 0x2528, 0x25E8, 0x2418, 0x5000, 0x8FFE}},
    {18603, 2, {0x720, 0x25FE, 0x1540, 0x157C, 0x690, 0x57E, 0xF500, 0x157C, 0x1544, 0x157C, 0x1644, 0x147C, 0x1444, 0x144C, 0x2800, 0x47FE}},
    {18604, 2, {0x3C, 0x47C0, 0x2044, 0x2224, 0x108, 0x7FE, 0xE402, 0x2A08, 0x2208, 0x23BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x5408, 0x8FFE}},
    {18605, 2, {0x208, 0x4110, 0x27FC, 0x2040, 0x3F8, 0x40, 0xE7FC, 0x2000, 0x2440, 0x227C, 0x2084, 0x2228, 0x2C20, 0x2450, 0x5488, 0x8FFE}},
    {18606, 2, {0x0, 0x47FC, 0x2040, 0x2FFE, 0x842, 0x358, 0xE040, 0x2358, 0x2000, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x5000, 0x8FFE}},
    {18607, 2, {0x244, 0x4448, 0x20B0, 0x230C, 0xC42, 0x244, 0xE4A8, 0x2318, 0x3C06, 0x2408, 0x27BE, 0x2888, 0x3528, 0x223E, 0x5408, 0x8FFE}},
    {18608, 2, {0x7BC, 0x2084, 0x14A4, 0x1294, 0x4A4, 0x120, 0xF210, 0x17FC, 0x1A20, 0x13FC, 0x1220, 0x13FC, 0x1220, 0x13FE, 0x2A00, 0x47FE}},
    {18609, 2, {0x40, 0x27FC, 0x1040, 0x13F8, 0x0, 0x7FC, 0xF4A4, 0x17FC, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2910, 0x47FE}},
    {18610, 2, {0x7BC, 0x24A4, 0x17BC, 0x14A4, 0x7BC, 0x140, 0xF3FC, 0x1240, 0x17FC, 0x1A40, 0x13FE, 0x1290, 0x1060, 0x1398, 0x2800, 0x47FE}},
    {18611, 2, {0x0, 0x3C7C, 0x2444, 0x2448, 0x2448, 0x2450, 0x2448, 0x2448, 0x2444, 0x2444, 0x2444, 0x2568, 0x4650, 0x4440, 0x8040, 0x40}},
    {18612, 2, {0x0, 0x77C, 0x7844, 0x848, 0x848, 0x850, 0x848, 0xFFC8, 0x844, 0x844, 0x1044, 0x1068, 0x2050, 0x2040, 0x4040, 0x8040}},
    {18613, 2, {0x0, 0x7E7C, 0x44, 0x48, 0x48, 0xFF50, 0x1048, 0x1048, 0x2044, 0x2044, 0x4444, 0x4268, 0xFF50, 0x4140, 0x40, 0x40}},
    {18614, 2, {0x800, 0x87C, 0x844, 0x848, 0x7F48, 0x850, 0x1848, 0x1C48, 0x2A44, 0x2944, 0x4844, 0x8868, 0x850, 0x840, 0x840, 0x840}},
    {18615, 2, {0x0, 0x3F3E, 0x2122, 0x2124, 0x2924, 0x2528, 0x2524, 0x2124, 0xFFE2, 0x2122, 0x2122, 0x2134, 0x2128, 0x4120, 0x4520, 0x8220}},
    {18616, 2, {0x0, 0x3F3E, 0x2122, 0x2124, 0x2124, 0x3F28, 0x2124, 0x2124, 0x3F22, 0x2122, 0x2122, 0x2134, 0x27A8, 0xF820, 0x4020, 0x20}},
    {18617, 2, {0x0, 0x7FBE, 0x4022, 0x4024, 0x4024, 0x7F28, 0x4124, 0x4124, 0x4122, 0x4122, 0x7F22, 0x4034, 0x4028, 0x4020, 0x7FA0, 0x20}},
    {18618, 2, {0x800, 0x87C, 0x4944, 0x4948, 0x4948, 0x4950, 0x7F48, 0x848, 0x844, 0x4944, 0x4944, 0x4968, 0x4F50, 0x7940, 0x40, 0x40}},
    {18619, 2, {0x0, 0x7FBE, 0x4822, 0x4824, 0x4824, 0x7F28, 0x4124, 0x4124, 0x4122, 0x4122, 0x7F22, 0x4834, 0x4828, 0x4820, 0x7FA0, 0x20}},
    {18620, 2, {0x1200, 0x123E, 0x1222, 0x1224, 0x7F24, 0x1228, 0x1224, 0x1224, 0x1222, 0xFFA2, 0x22, 0x1234, 0x1128, 0x2120, 0x20A0, 0x40A0}},
    {18621, 2, {0x440, 0x440, 0x7FFC, 0x440, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1004, 0x1004, 0xFFC}},
    {18622, 2, {0x800, 0x287C, 0x2844, 0x3E48, 0x4848, 0x850, 0x848, 0x7F48, 0x1444, 0x1444, 0x1444, 0x1468, 0x2550, 0x2640, 0x4440, 0x8040}},
    {18623, 2, {0x1000, 0x83E, 0x822, 0xFFA4, 0x80A4, 0x1128, 0x1024, 0xFFA4, 0x2222, 0x2222, 0x4222, 0x3434, 0x828, 0x1420, 0x2220, 0xC120}},
    {18624, 2, {0x1000, 0x107C, 0x3F44, 0x2148, 0x4248, 0xBFD0, 0x2048, 0x2F48, 0x2944, 0x2944, 0x2D44, 0x2A68, 0x2850, 0x4A40, 0x4C40, 0x8840}},
    {18625, 2, {0x0, 0xF7BE, 0x10A2, 0x10A4, 0x94A4, 0x52A8, 0x52A4, 0x10A4, 0x31A2, 0x52A2, 0x94A2, 0x10B4, 0x10A8, 0x10A0, 0x52A0, 0x2120}},
    {18626, 2, {0x800, 0x87C, 0x844, 0x7F48, 0x848, 0x3E50, 0x848, 0x7F48, 0x1844, 0x1C44, 0x2A44, 0x2A68, 0x4850, 0x8840, 0x840, 0x840}},
    {18627, 2, {0x2000, 0x207C, 0x3F44, 0x4848, 0x8848, 0x850, 0x7F48, 0x848, 0x844, 0x2E44, 0x2844, 0x2868, 0x2F50, 0xF040, 0x4040, 0x40}},
    {18628, 2, {0x800, 0x87C, 0x1444, 0x2248, 0x4148, 0xBE50, 0x848, 0x848, 0x7F44, 0x844, 0x2A44, 0x2968, 0x4950, 0x8840, 0x2840, 0x1040}},
    {18629, 2, {0x0, 0x3FBE, 0x2022, 0x2024, 0x2F24, 0x2028, 0x2024, 0x3FA4, 0x2822, 0x28A2, 0x2D22, 0x2A34, 0x2928, 0x48A0, 0x4C20, 0x8820}},
    {18630, 2, {0x0, 0x7FBE, 0x22, 0x24A4, 0x4924, 0x9228, 0x4924, 0x24A4, 0x22, 0x7FA2, 0x822, 0x834, 0x828, 0xBA0, 0xFC20, 0x20}},
    {18631, 2, {0x800, 0xA7C, 0x944, 0x848, 0x7F48, 0x850, 0x848, 0x4948, 0x2A44, 0x844, 0x1C44, 0x2A68, 0xC950, 0x840, 0x2840, 0x1040}},
    {18632, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x7CF8, 0x0, 0x7DFC, 0x5524, 0x5524, 0x5524, 0x7DFC, 0x4100, 0x4100, 0x4D02, 0x7102, 0x40FE}},
    {18633, 2, {0x2040, 0x1248, 0x4444, 0x2014, 0x860, 0x7380, 0x2000, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2002, 0x1FFE}},
    {18634, 2, {0x800, 0x493E, 0x2922, 0x2A24, 0x824, 0x7F28, 0x4124, 0x4124, 0x5D22, 0x5522, 0x5522, 0x5534, 0x5D28, 0x4120, 0x4520, 0x4220}},
    {18635, 2, {0x100, 0x7BE, 0x7822, 0x8A4, 0x44A4, 0x2528, 0x2024, 0x424, 0x7FA2, 0xC22, 0x1622, 0x1534, 0x24A8, 0x4420, 0x8420, 0x420}},
    {18636, 2, {0x800, 0x41E, 0x7FD2, 0x4012, 0x4394, 0x5E14, 0x5218, 0x5214, 0x5FD2, 0x5212, 0x5212, 0x521A, 0x5554, 0x5950, 0x94D0, 0x250}},
    {18637, 2, {0x800, 0x83E, 0xFFA2, 0x1424, 0x2224, 0x4928, 0x88A4, 0x7F24, 0x4922, 0x7F22, 0x4922, 0x7F34, 0x828, 0xA20, 0xC20, 0x820}},
    {18638, 2, {0x2480, 0x24BE, 0x4922, 0x9224, 0x4924, 0x24A8, 0x24A4, 0x24, 0x7FA2, 0x48A2, 0x48A2, 0x7FB4, 0x48A8, 0x48A0, 0x7FA0, 0x40A0}},
    {18639, 2, {0x1500, 0x153E, 0x1522, 0x2524, 0x2FA4, 0x6528, 0x6524, 0xA524, 0x2522, 0x2522, 0x3FE2, 0x2034, 0x2528, 0x24A0, 0x28A0, 0x3020}},
    {18640, 2, {0x2200, 0x227C, 0x7F44, 0x2248, 0x2A48, 0x850, 0x3E48, 0x2A48, 0x2A44, 0x2A44, 0xFF44, 0x868, 0x1450, 0x1240, 0x2240, 0x4040}},
    {18641, 2, {0x0, 0x3FBE, 0x2022, 0x2FA4, 0x2AA4, 0x2AA8, 0x2FA4, 0x2AA4, 0x2AA2, 0x2FA2, 0x2222, 0x2FB4, 0x4228, 0x43A0, 0x9C20, 0x820}},
    {18642, 2, {0x800, 0x83E, 0x7F22, 0x4924, 0x7F24, 0x828, 0xFFA4, 0x24, 0x7F22, 0x4122, 0x4922, 0x4934, 0x4928, 0x1420, 0x2220, 0xC120}},
    {18643, 2, {0x0, 0xFFBE, 0x22, 0x7F24, 0x4124, 0x4128, 0x7F24, 0x24, 0xFFA2, 0xA2A2, 0x94A2, 0xFFB4, 0x88A8, 0x88A0, 0x8AA0, 0x8120}},
    {18644, 2, {0x800, 0x497C, 0x2A44, 0x848, 0x7F48, 0x4150, 0x4148, 0x7F48, 0x4144, 0x7F44, 0x4144, 0x4168, 0x7F50, 0x40, 0x2240, 0x4140}},
    {18645, 2, {0x0, 0x7FBE, 0x4822, 0x7F24, 0x4824, 0x7F28, 0x4824, 0x4824, 0x7FA2, 0xA2, 0xAAA2, 0xAAB4, 0xAAA8, 0x80A0, 0x520, 0x220}},
    {18646, 2, {0x0, 0x7E78, 0x4848, 0x7E48, 0x4878, 0x7E00, 0x48FC, 0x48A4, 0x7EA4, 0x2A4, 0xAAFC, 0xAA80, 0xAA80, 0x8282, 0x1482, 0x87E}},
    {18647, 2, {0x800, 0x103E, 0x7F22, 0x4124, 0x7F24, 0x4128, 0x7F24, 0x1024, 0xFFA2, 0x2222, 0x4922, 0x88B4, 0x7F68, 0x820, 0x820, 0x820}},
    {18648, 2, {0x800, 0x493E, 0x2A22, 0xFFA4, 0x80A4, 0x3E28, 0x2224, 0x2224, 0x3E22, 0x822, 0x7F22, 0x834, 0x828, 0xF20, 0xF020, 0x4020}},
    {18649, 2, {0x400, 0x445E, 0x4452, 0x7FD2, 0x14, 0x7BD4, 0x4A58, 0x4A54, 0x7BD2, 0x4A52, 0x4A52, 0x7BDA, 0x4A54, 0x4A50, 0xAB50, 0x1490}},
    {18650, 2, {0x800, 0x7F3E, 0x4922, 0xFFA4, 0x4924, 0x7F28, 0x824, 0x7F24, 0x4922, 0x7F22, 0x1022, 0xFF34, 0x2128, 0x7220, 0xC20, 0xF320}},
    {18651, 2, {0x800, 0x83E, 0xFFA2, 0x1C24, 0x2B24, 0xC8A8, 0x1424, 0x2224, 0x4922, 0x88A2, 0x4922, 0x2A34, 0x4928, 0x88A0, 0x2820, 0x1020}},
    {18652, 2, {0x800, 0xF3E, 0x822, 0x7FA4, 0x48A4, 0x4E28, 0x7924, 0x4724, 0x4022, 0x5F22, 0x5122, 0x5F34, 0x5128, 0x5F20, 0x51A0, 0xBE20}},
    {18653, 2, {0x0, 0x3FDE, 0x2052, 0x2052, 0x3FD4, 0x2234, 0x2FD8, 0x2294, 0x3FF2, 0x2212, 0x2FD2, 0x385A, 0x2FD4, 0x4850, 0x4FD0, 0x8850}},
    {18654, 2, {0x0, 0xF7BE, 0x94A2, 0x94A4, 0xF7A4, 0x2228, 0x1424, 0xFFA4, 0x822, 0x4922, 0x4922, 0x7F34, 0x928, 0x1020, 0x2020, 0x4020}},
    {18655, 2, {0x0, 0xF7BE, 0x2122, 0xA524, 0xF7A4, 0x6328, 0xB5A4, 0x2124, 0x7FA2, 0x40A2, 0x40A2, 0x7FB4, 0x40A8, 0x40A0, 0x7FA0, 0x40A0}},
    {18656, 2, {0x1100, 0x111E, 0x7BD2, 0x1112, 0xFFF4, 0x1114, 0x2A98, 0x4454, 0x3FB2, 0x2092, 0x2092, 0x3F9A, 0x2094, 0x2090, 0x3F90, 0x2090}},
    {18657, 2, {0x4100, 0x223E, 0xFFA2, 0x824, 0x7F24, 0x828, 0xFFA4, 0x1524, 0x64A2, 0x2422, 0xFFA2, 0x2434, 0x3528, 0xE2A0, 0x25A0, 0x68A0}},
    {18658, 2, {0x2280, 0x229E, 0x27D2, 0xFA92, 0x2294, 0x7294, 0x27F8, 0xF814, 0x27D2, 0x7452, 0x6C52, 0xA7DA, 0x2454, 0x2450, 0x27D0, 0x2450}},
    {18659, 2, {0x2480, 0x151E, 0x7FD2, 0x4052, 0x1F14, 0x1114, 0x7FD8, 0x5554, 0x4E52, 0x7FD2, 0x412, 0x7FDA, 0x414, 0xFFF0, 0x2A50, 0x4530}},
    {18660, 2, {0x7BC0, 0x4A5E, 0x7BD2, 0x4A52, 0x7BD4, 0x1414, 0x3FD8, 0x6414, 0xBF92, 0x2412, 0x3FD2, 0x1A, 0x3F94, 0x1090, 0xF10, 0xF090}},
    {18661, 2, {0x20, 0xFFA0, 0x1420, 0x1422, 0x7F22, 0x5524, 0x5528, 0x5530, 0x5720, 0x6120, 0x4120, 0x7F22, 0x4122, 0x4122, 0x7F1E, 0x4100}},
    {18662, 2, {0x20, 0xFFA0, 0x1420, 0x1420, 0x7F7C, 0x5524, 0x5524, 0x5524, 0x5724, 0x6124, 0x4124, 0x7F24, 0x4144, 0x4144, 0x7F94, 0x4108}},
    {18663, 2, {0x20, 0xFFA8, 0x1424, 0x1424, 0x7F20, 0x553E, 0x55E0, 0x5520, 0x5720, 0x6120, 0x4120, 0x7F20, 0x4114, 0x4114, 0x7F0C, 0x4104}},
    {18664, 2, {0x10, 0xFF90, 0x1410, 0x141C, 0x7F70, 0x5510, 0x5510, 0x5510, 0x571E, 0x61F0, 0x4110, 0x7F12, 0x4112, 0x4112, 0x7F0E, 0x4100}},
    {18665, 2, {0x20, 0xFE20, 0x2820, 0x29FC, 0xFE20, 0xAAA8, 0xAAA8, 0xAAA8, 0xAEA8, 0xC2A8, 0x82F8, 0xFE22, 0x8222, 0x8222, 0xFE1E, 0x8200}},
    {18666, 2, {0x10, 0xFE10, 0x2810, 0x28FE, 0xFE10, 0xAA10, 0xAAFE, 0xAA92, 0xAE92, 0xC292, 0x8292, 0xFE9A, 0x8294, 0x8210, 0xFE10, 0x8210}},
    {18667, 2, {0x10, 0xFE10, 0x2810, 0x28FE, 0xFE10, 0xAA10, 0xAAFC, 0xAA44, 0xAE44, 0xC244, 0x8228, 0xFE28, 0x8210, 0x8228, 0xFE44, 0x8282}},
    {18668, 2, {0x48, 0xFE44, 0x2844, 0x2840, 0xFE5E, 0xABE0, 0xAA40, 0xAA44, 0xAE44, 0xC248, 0x8230, 0xFE22, 0x8252, 0x828A, 0xFF06, 0x8202}},
    {18669, 2, {0x10, 0xFE14, 0x2812, 0x2812, 0xFE10, 0xAAFE, 0xAA10, 0xAA10, 0xAE10, 0xC228, 0x8228, 0xFE28, 0x8228, 0x8244, 0xFE44, 0x8282}},
    {18670, 2, {0x28, 0xFE24, 0x2824, 0x2820, 0xFEFE, 0xAA20, 0xAA20, 0xAA3C, 0xAE34, 0xC254, 0x8254, 0xFE54, 0x8288, 0x8288, 0xFF14, 0x8222}},
    {18671, 2, {0x0, 0xFFBC, 0x1424, 0x1424, 0x7F24, 0x553C, 0x5524, 0x5524, 0x5724, 0x613C, 0x4124, 0x7F24, 0x4124, 0x4124, 0x7F7E, 0x4100}},
    {18672, 2, {0x0, 0xFFBE, 0x1420, 0x1420, 0x7F20, 0x553E, 0x5522, 0x5522, 0x5722, 0x6122, 0x413E, 0x7F20, 0x4120, 0x4120, 0x7F3E, 0x4100}},
    {18673, 2, {0x20, 0xFFA0, 0x1420, 0x147C, 0x7F44, 0x5584, 0x5574, 0x5554, 0x5754, 0x6154, 0x4174, 0x7F54, 0x4104, 0x4104, 0x7F28, 0x4110}},
    {18674, 2, {0x20, 0xFE20, 0x2840, 0x2848, 0xFE84, 0xAAFC, 0xAA04, 0xAA48, 0xAE48, 0xC2FE, 0x8248, 0xFE48, 0x8248, 0x8288, 0xFE88, 0x8308}},
    {18675, 2, {0x0, 0xFEFC, 0x2804, 0x2804, 0xFEFC, 0xAA04, 0xAA04, 0xAAF4, 0xAE94, 0xC294, 0x8294, 0xFEF4, 0x8204, 0x8204, 0xFE14, 0x8208}},
    {18676, 2, {0x0, 0xFE44, 0x2828, 0x2892, 0xFEAA, 0xAAC6, 0xAA82, 0xAAFE, 0xAE28, 0xC228, 0x8228, 0xFE28, 0x8248, 0x824A, 0xFE8A, 0x8306}},
    {18677, 2, {0x0, 0xFEFE, 0x2848, 0x2848, 0xFE78, 0xAA48, 0xAA48, 0xAA78, 0xAE48, 0xC248, 0x824E, 0xFEF8, 0x8208, 0x8208, 0xFE08, 0x8208}},
    {18678, 2, {0x8, 0xFE1C, 0x2870, 0x2810, 0xFE10, 0xAAFE, 0xAA10, 0xAA10, 0xAE10, 0xC27C, 0x8244, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {18679, 2, {0x10, 0xFE50, 0x2850, 0x287C, 0xFE90, 0xAA10, 0xAA10, 0xAAFE, 0xAE10, 0xC238, 0x8238, 0xFE54, 0x8254, 0x8292, 0xFE10, 0x8210}},
    {18680, 2, {0x838, 0x13C0, 0x3040, 0x57FC, 0x9040, 0x1040, 0x13F8, 0x1000, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8}},
    {18681, 2, {0x50, 0xFE58, 0x2854, 0x2894, 0xFE90, 0xAA9C, 0xABF0, 0xAA94, 0xAE94, 0xC294, 0x8298, 0xFE98, 0x8294, 0x8294, 0xFEAC, 0x82C4}},
    {18682, 2, {0x0, 0xFE7C, 0x2844, 0x2844, 0xFE7C, 0xAA00, 0xAAFE, 0xAA82, 0xAE82, 0xC2FE, 0x8282, 0xFE82, 0x82FE, 0x8282, 0xFE8A, 0x8284}},
    {18683, 2, {0x10, 0xFE10, 0x2828, 0x2844, 0xFE92, 0xAA08, 0xAA7C, 0xAA04, 0xAE08, 0xC210, 0x827C, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {18684, 2, {0x20, 0xFE10, 0x28FE, 0x2800, 0xFE00, 0xAA7C, 0xAA44, 0xAA44, 0xAE7C, 0xC210, 0x8254, 0xFE52, 0x8292, 0x8210, 0xFE50, 0x8220}},
    {18685, 2, {0x2000, 0x3E7C, 0x4844, 0x844, 0xFF44, 0x1444, 0x227C, 0x4000, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8}},
    {18686, 2, {0x20, 0xFE10, 0x28FE, 0x2882, 0xFF04, 0xAA7C, 0xAA44, 0xAA44, 0xAE7C, 0xC244, 0x8244, 0xFE7C, 0x8244, 0x8244, 0xFFFE, 0x8200}},
    {18687, 2, {0x0, 0xFE7C, 0x2810, 0x2810, 0xFEFE, 0xAA10, 0xAA10, 0xAA50, 0xAE20, 0xC2FE, 0x82AA, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {18688, 2, {0x44, 0xFE44, 0x29FE, 0x2844, 0xFE54, 0xAA10, 0xABFE, 0xAA20, 0xAE40, 0xC27E, 0x82C2, 0xFF42, 0x8242, 0x8242, 0xFE7E, 0x8242}},
    {18689, 2, {0x48, 0xFE48, 0x29FC, 0x2848, 0xFE78, 0xAA48, 0xAA78, 0xAA48, 0xAE48, 0xC3FE, 0x8280, 0xFEA8, 0x82C4, 0x8280, 0xFEFE, 0x8200}},
    {18690, 2, {0x88, 0xFE88, 0x2BFE, 0x2888, 0xFE88, 0xAAF8, 0xAA88, 0xAA88, 0xAEF8, 0xC220, 0x83FE, 0xFE70, 0x82A8, 0x8324, 0xFE22, 0x8220}},
    {18691, 2, {0x40, 0xFE20, 0x29FE, 0x2800, 0xFE84, 0xAA48, 0xABFE, 0xAA00, 0xAEFC, 0xC284, 0x8284, 0xFEFC, 0x8284, 0x8284, 0xFEFC, 0x8284}},
    {18692, 2, {0x0, 0xFEFE, 0x2810, 0x2820, 0xFEFE, 0xAAAA, 0xAAAA, 0xAABA, 0xAEAA, 0xC2AA, 0x82BA, 0xFEAA, 0x82AA, 0x82AA, 0xFEFE, 0x8282}},
    {18693, 2, {0x10, 0xFE10, 0x2828, 0x2844, 0xFE82, 0xAA7C, 0xAA00, 0xAAE2, 0xAEAA, 0xC2AA, 0x82EA, 0xFEAA, 0x82AA, 0x82E2, 0xFEAA, 0x82A4}},
    {18694, 2, {0x10, 0xFED4, 0x2848, 0x294A, 0xFE84, 0xAA84, 0xAB02, 0xAAF8, 0xAE20, 0xC220, 0x83FC, 0xFE20, 0x8250, 0x8248, 0xFE84, 0x8304}},
    {18695, 2, {0x20, 0xFE10, 0x28FE, 0x2882, 0xFE28, 0xAA44, 0xAA92, 0xAA28, 0xAE44, 0xC282, 0x827C, 0xFE04, 0x8228, 0x8210, 0xFE28, 0x82C4}},
    {18696, 2, {0x44, 0xFE82, 0x297C, 0x2820, 0xFE40, 0xAA7C, 0xAA04, 0xAA28, 0xAE10, 0xC2FE, 0x82AA, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {18697, 2, {0x10, 0xFE0C, 0x2824, 0x28AA, 0xFEB0, 0xAB24, 0xAA64, 0xAB9C, 0xAE00, 0xC2FE, 0x82AA, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {18698, 2, {0x44, 0xFE44, 0x28FE, 0x2844, 0xFE00, 0xAAFE, 0xAA44, 0xAA44, 0xAE7C, 0xC244, 0x827C, 0xFE44, 0x824E, 0x82F4, 0xFE04, 0x8204}},
    {18699, 2, {0x4408, 0x240C, 0x280A, 0xFE08, 0x29FE, 0x2808, 0xFE08, 0xAAEA, 0xAAAA, 0xAEAA, 0xC2EC, 0xFE0C, 0x822A, 0x82CA, 0xFE16, 0x8222}},
    {18700, 2, {0x20, 0xFE20, 0x2850, 0x2848, 0xFEA4, 0xABFE, 0xAA84, 0xAAFC, 0xAE84, 0xC2FC, 0x8280, 0xFEFC, 0x8344, 0x8344, 0xFE7C, 0x8244}},
    {18701, 2, {0x88, 0xF888, 0x23FE, 0x2088, 0xF888, 0xA8F8, 0xA820, 0xABFE, 0xDA22, 0x8B32, 0x8AAA, 0xFB76, 0x8A22, 0x8A22, 0xFA2A, 0x8A24}},
    {18702, 2, {0xFC, 0xFE28, 0x2810, 0x29FE, 0xFE52, 0xAA94, 0xAB50, 0xAA20, 0xAEFE, 0xC2AA, 0x82C6, 0xFEBA, 0x82AA, 0x82BA, 0xFE82, 0x8286}},
    {18703, 2, {0x28, 0xFEEE, 0x2828, 0x28EE, 0xFE28, 0xAAEE, 0xAA28, 0xAA06, 0xAEF8, 0xC210, 0x82FE, 0xFE10, 0x82FE, 0x8210, 0xFE50, 0x8220}},
    {18704, 2, {0x0, 0xFBDE, 0x2088, 0x22A8, 0xFBFE, 0xA988, 0xAADA, 0xACA6, 0xD9F8, 0x8908, 0x8908, 0xF9F8, 0x8908, 0x8908, 0xF9F8, 0x8908}},
    {18705, 2, {0x28, 0xFE28, 0x287C, 0x2828, 0xFE28, 0xAAFE, 0xAA10, 0xAA7C, 0xAE54, 0xC27C, 0x8254, 0xFE7C, 0x8200, 0x8228, 0xFE44, 0x8282}},
    {18706, 2, {0x124, 0xF924, 0x22AA, 0x23AE, 0xF924, 0xAAAA, 0xABAE, 0xA924, 0xDBFE, 0x8910, 0x8914, 0xF914, 0x8A8A, 0x8A4A, 0xFA16, 0x8C22}},
    {18707, 2, {0x48, 0xFE48, 0x29FE, 0x2848, 0xFEFE, 0xAA82, 0xAB7C, 0xAA00, 0xAEFE, 0xC220, 0x82D2, 0xFE2C, 0x82D8, 0x822C, 0xFECA, 0x8230}},
    {18708, 2, {0x4204, 0x3FD8, 0x8A90, 0x4F9E, 0x2A94, 0x5FD4, 0x8224, 0xFFFE, 0x480, 0x3FF8, 0x2488, 0x2878, 0x3008, 0x3FF8, 0x2008, 0x3FF8}},
    {18709, 2, {0x124, 0xF974, 0x2258, 0x22DA, 0xFBFE, 0xA954, 0xAADA, 0xABFE, 0xD8A2, 0x8820, 0x8BFE, 0xF870, 0x88A8, 0x8924, 0xFA22, 0x8820}},
    {18710, 2, {0x1FC, 0xF904, 0x21FC, 0x2104, 0xF9FC, 0xA888, 0xA9FC, 0xA888, 0xDBFE, 0x8888, 0x8924, 0xFAAA, 0x8870, 0x88A8, 0xF924, 0x8860}},
    {18711, 2, {0x820, 0x7F7C, 0x14C4, 0x7F38, 0x22D6, 0x3E38, 0x2250, 0x7F7C, 0x810, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x37B8, 0x2008, 0x3FF8}},
    {18712, 2, {0x148, 0xF94C, 0x22AA, 0x2008, 0xFBFE, 0xA948, 0xAB68, 0xA94A, 0xDB6A, 0x894A, 0x8B6C, 0xF94C, 0x896A, 0x8B8A, 0xF816, 0x8822}},
    {18713, 2, {0xF8, 0xF908, 0x21F0, 0x2010, 0xFBFE, 0xA8C4, 0xAB28, 0xA8D4, 0xDB32, 0x88D0, 0x8A28, 0xFFBE, 0x8AAA, 0x8FBE, 0xFA8A, 0x8FBE}},
    {18714, 2, {0x0, 0x3DFC, 0x2524, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x25FC, 0x2420, 0x3C20, 0x25FC, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {18715, 2, {0x1004, 0x1004, 0x2884, 0x2484, 0x4284, 0xBC84, 0x1084, 0x1084, 0xFE8C, 0x1094, 0x94E4, 0x5884, 0x5004, 0x1E04, 0xF004, 0x4004}},
    {18716, 2, {0x1020, 0x1020, 0x2820, 0x243C, 0x41E0, 0xB820, 0x1020, 0x1020, 0xFC3E, 0x13E0, 0x9420, 0x5822, 0x5022, 0x1C22, 0xE01E, 0x4000}},
    {18717, 2, {0x1000, 0x1040, 0x299E, 0x2512, 0x4112, 0xB912, 0x1112, 0x1112, 0xFD12, 0x1112, 0x955A, 0x5994, 0x5110, 0x1C10, 0xE010, 0x4010}},
    {18718, 2, {0x1040, 0x1050, 0x2848, 0x2448, 0x4040, 0xB9FE, 0x1050, 0x1050, 0xFC50, 0x1050, 0x9490, 0x5890, 0x5092, 0x1D12, 0xE10E, 0x4200}},
    {18719, 2, {0x1000, 0x1000, 0x29FE, 0x2502, 0x4302, 0xBC00, 0x1078, 0x1048, 0xFE48, 0x1048, 0x9448, 0x584A, 0x504A, 0x1E8A, 0xF086, 0x4100}},
    {18720, 2, {0x1010, 0x1010, 0x2810, 0x2490, 0x4290, 0xBC90, 0x109E, 0x1090, 0xFE90, 0x1090, 0x9490, 0x5890, 0x5090, 0x1E90, 0xF1FE, 0x4000}},
    {18721, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4124, 0xB924, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x97FE, 0x5904, 0x5104, 0x1D04, 0xE114, 0x4108}},
    {18722, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4104, 0xB9FC, 0x1040, 0x1040, 0xFDFC, 0x1044, 0x9444, 0x5884, 0x5084, 0x1D04, 0xE228, 0x4410}},
    {18723, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x4124, 0xB8A4, 0x10A8, 0x1020, 0xFDFE, 0x1020, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE0A0, 0x4040}},
    {18724, 2, {0x1020, 0x1010, 0x2810, 0x25FE, 0x4220, 0xBC24, 0x1044, 0x10F8, 0xFE12, 0x1022, 0x9444, 0x5988, 0x5010, 0x1E28, 0xF044, 0x4182}},
    {18725, 2, {0x1040, 0x1040, 0x28FC, 0x2504, 0x4208, 0xB9FE, 0x1100, 0x117C, 0xFD44, 0x1144, 0x9554, 0x5948, 0x5142, 0x1E42, 0xE23E, 0x4400}},
    {18726, 2, {0x10A8, 0x10A4, 0x28A4, 0x2520, 0x412E, 0xBBF0, 0x1520, 0x1124, 0xFD24, 0x1124, 0x9528, 0x5910, 0x5132, 0x1D4A, 0xE186, 0x4102}},
    {18727, 2, {0x1000, 0x1000, 0x29FC, 0x2504, 0x4104, 0xB974, 0x1154, 0x1154, 0xFD54, 0x1154, 0x9574, 0x5904, 0x5104, 0x1DFC, 0xE104, 0x4000}},
    {18728, 2, {0x1004, 0x101E, 0x29E0, 0x2500, 0x4106, 0xB978, 0x1150, 0x1150, 0xFD52, 0x1154, 0x9548, 0x5948, 0x5144, 0x1E54, 0xE262, 0x4440}},
    {18729, 2, {0x1020, 0x1010, 0x29FE, 0x2502, 0x4000, 0xB81C, 0x10E0, 0x1020, 0xFC20, 0x103E, 0x95E0, 0x5820, 0x5022, 0x1C22, 0xE01E, 0x4000}},
    {18730, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x4088, 0xB888, 0x1126, 0x1220, 0xFDFE, 0x1020, 0x9470, 0x58A8, 0x5124, 0x1E22, 0xE020, 0x4020}},
    {18731, 2, {0x1020, 0x1124, 0x2924, 0x2524, 0x41FC, 0xB820, 0x1050, 0x1088, 0xFD44, 0x1222, 0x9420, 0x59F8, 0x5008, 0x1C10, 0xE010, 0x4020}},
    {18732, 2, {0x2010, 0x2210, 0x5110, 0x4910, 0x8454, 0x7A52, 0x2252, 0x2090, 0xF810, 0x2114, 0x2E04, 0xB208, 0x6208, 0x3A10, 0xE220, 0xC0}},
    {18733, 2, {0x2184, 0x2068, 0x5030, 0x48C8, 0x8304, 0x7840, 0x23FE, 0x20A0, 0xF920, 0x23FC, 0x2D24, 0xB124, 0x6134, 0x3928, 0xE020, 0x20}},
    {18734, 2, {0x1010, 0x1010, 0x2890, 0x249E, 0x4090, 0xB890, 0x13FE, 0x1000, 0xFC10, 0x1092, 0x9492, 0x5914, 0x5208, 0x1C10, 0xE060, 0x4380}},
    {18735, 2, {0x1002, 0x13E2, 0x2944, 0x2548, 0x4140, 0xB942, 0x1142, 0x13E4, 0xFD48, 0x1140, 0x9542, 0x5942, 0x5144, 0x1E44, 0xE248, 0x4450}},
    {18736, 2, {0x1080, 0x1080, 0x7DFC, 0x1244, 0x1048, 0x1CA0, 0xE110, 0x6C8, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0x540, 0xFFFE, 0x0}},
    {18737, 2, {0x1020, 0x1124, 0x2924, 0x25FC, 0x4000, 0xB80C, 0x11F0, 0x1100, 0xFD00, 0x11FE, 0x9510, 0x5910, 0x5110, 0x1E10, 0xE210, 0x4410}},
    {18738, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x4020, 0xB820, 0x13FE, 0x1048, 0xFC48, 0x114C, 0x954A, 0x5A4A, 0x5088, 0x1C88, 0xE128, 0x4210}},
    {18739, 2, {0x2080, 0x2040, 0x57FE, 0x4920, 0x8120, 0x7A3C, 0x2244, 0x2664, 0xFA94, 0x2348, 0x2A48, 0xB230, 0x6220, 0x3A50, 0xE288, 0x306}},
    {18740, 2, {0x2000, 0x21FC, 0x5008, 0x4810, 0x8222, 0x7AAA, 0x2272, 0x2222, 0xFA72, 0x22AA, 0x2B26, 0xB2A2, 0x6242, 0x3A02, 0xE3FE, 0x2}},
    {18741, 2, {0x1090, 0x1090, 0x2890, 0x279E, 0x4090, 0xB890, 0x1090, 0x139C, 0xFC90, 0x1090, 0x9490, 0x5B9E, 0x5090, 0x1C90, 0xE090, 0x4090}},
    {18742, 2, {0x1020, 0x1020, 0x2BFE, 0x2450, 0x4088, 0xB924, 0x1222, 0x11FC, 0xFD24, 0x11FC, 0x9524, 0x59FC, 0x5022, 0x1C22, 0xE01E, 0x4000}},
    {18743, 2, {0x2000, 0x23FE, 0x5200, 0x4AFC, 0x8284, 0x7AFC, 0x2284, 0x22FC, 0xFA20, 0x23FE, 0x2A48, 0xB2C8, 0x6230, 0x3A48, 0xE284, 0x3FE}},
    {18744, 2, {0x2080, 0x20FC, 0x5104, 0x49F8, 0x8008, 0x7BFE, 0x2040, 0x20A2, 0xFB34, 0x2058, 0x2894, 0xB334, 0x6052, 0x3890, 0xE350, 0x20}},
    {18745, 2, {0x1040, 0x1020, 0x2BFE, 0x2622, 0x40A4, 0xB928, 0x1050, 0x1088, 0xFD04, 0x10F8, 0x9488, 0x5888, 0x5050, 0x1C20, 0xE0D8, 0x4306}},
    {18746, 2, {0x1000, 0x1002, 0x2BDC, 0x2490, 0x4090, 0xB910, 0x11DE, 0x1354, 0xFD54, 0x1154, 0x9554, 0x59D4, 0x5154, 0x1C24, 0xE024, 0x4044}},
    {18747, 2, {0x1040, 0x1040, 0x2BFE, 0x2480, 0x41FC, 0xBA20, 0x15FE, 0x1000, 0xFDFC, 0x1104, 0x95FC, 0x5904, 0x51FC, 0x1D04, 0xE114, 0x4108}},
    {18748, 2, {0x2040, 0x2080, 0x51FC, 0x4904, 0x81FC, 0x7904, 0x21FC, 0x2020, 0xF832, 0x23B4, 0x28A8, 0xB128, 0x6124, 0x3A24, 0xE4A2, 0x40}},
    {18749, 2, {0x1000, 0x11FC, 0x2954, 0x2554, 0x4154, 0xB9FC, 0x1020, 0x10A8, 0xFCA4, 0x1124, 0x9420, 0x58A8, 0x50A4, 0x1D24, 0xE020, 0x4020}},
    {18750, 2, {0x101E, 0x11E0, 0x2822, 0x2512, 0x4094, 0xB880, 0x1020, 0x11CE, 0xFD02, 0x1102, 0x95CE, 0x5902, 0x5102, 0x1D02, 0xE1FE, 0x4102}},
    {18751, 2, {0x1080, 0x13DE, 0x288A, 0x25CA, 0x408A, 0xBBD2, 0x10A6, 0x1008, 0xFCF0, 0x1020, 0x95FC, 0x5820, 0x53FE, 0x1C20, 0xE0A0, 0x4040}},
    {18752, 2, {0x1020, 0x1124, 0x2924, 0x25FC, 0x4020, 0xBBFE, 0x1000, 0x1020, 0xFDFC, 0x1124, 0x9524, 0x59FC, 0x5020, 0x1C24, 0xE3FE, 0x4102}},
    {18753, 2, {0x2000, 0x23FC, 0x5008, 0x48D0, 0x8020, 0x7BFE, 0x228A, 0x2252, 0xFAFA, 0x2222, 0x2A22, 0xB3FE, 0x6222, 0x3A22, 0xE3FE, 0x202}},
    {18754, 2, {0x2000, 0x23FE, 0x5202, 0x4CFC, 0x8000, 0x79FC, 0x2040, 0x20A4, 0xFB38, 0x2058, 0x2894, 0xB334, 0x6052, 0x3890, 0xE350, 0x20}},
    {18755, 2, {0x1040, 0x1040, 0x2BFE, 0x2692, 0x4088, 0xB9FE, 0x1310, 0x1510, 0xFDFE, 0x1110, 0x9510, 0x59FE, 0x5110, 0x1D10, 0xE1FE, 0x4100}},
    {18756, 2, {0x2000, 0x23FE, 0x5200, 0x4A48, 0x8248, 0x7AFC, 0x2248, 0x2248, 0xFBFE, 0x2200, 0x2AFC, 0xB284, 0x62FC, 0x3A84, 0xE2FC, 0x484}},
    {18757, 2, {0x100, 0x7FFC, 0x100, 0x7D7C, 0x1110, 0x7D7C, 0x1110, 0x1CFE, 0xE100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18758, 2, {0x100, 0xFFFE, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7D7C, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {18759, 2, {0x1020, 0x1010, 0x29FE, 0x2502, 0x4102, 0xB9FE, 0x1100, 0x11EE, 0xFD22, 0x11AA, 0x9566, 0x5922, 0x5166, 0x1EAA, 0xE222, 0x4466}},
    {18760, 2, {0x2088, 0x2088, 0x53FE, 0x4888, 0x8024, 0x7BFE, 0x2050, 0x2088, 0xF904, 0x228A, 0x2888, 0xB3FE, 0x6088, 0x3888, 0xE108, 0x208}},
    {18761, 2, {0x1000, 0x13FE, 0x2A02, 0x2440, 0x40F8, 0xB908, 0x1210, 0x11FC, 0xFD24, 0x1124, 0x95FC, 0x5850, 0x5058, 0x1C96, 0xE112, 0x420E}},
    {18762, 2, {0x1020, 0x13FE, 0x2820, 0x25FC, 0x4020, 0xBBFE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1DFC, 0xE088, 0x4104}},
    {18763, 2, {0x1020, 0x1010, 0x29FE, 0x2400, 0x42FC, 0xBC84, 0x10FC, 0x1000, 0xFEFC, 0x1008, 0x9410, 0x5850, 0x5020, 0x1E00, 0xF154, 0x4252}},
    {18764, 2, {0x200, 0x100, 0x7FFC, 0x4004, 0x3C78, 0x2448, 0x3C78, 0x2448, 0x3D78, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18765, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x982, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18766, 2, {0x23FE, 0x2200, 0x52FC, 0x4A00, 0x83FE, 0x7A94, 0x22A8, 0x22C4, 0xFA00, 0x22FC, 0x2A84, 0xB2FC, 0x6284, 0x3CFC, 0xE484, 0x88C}},
    {18767, 2, {0x2280, 0x228E, 0x54EA, 0x494A, 0x804A, 0x724A, 0x25EA, 0x2C4A, 0xF44A, 0x256A, 0xA54A, 0x754E, 0x2568, 0x3788, 0xC408, 0x408}},
    {18768, 2, {0x20, 0x7E7C, 0x14C4, 0x838, 0xFEC6, 0x2A20, 0x48FC, 0xA844, 0x118C, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18769, 2, {0x1020, 0x1020, 0x2BFE, 0x24A8, 0x40A8, 0xB974, 0x1222, 0x1070, 0xFCA8, 0x1146, 0x9478, 0x5888, 0x5150, 0x1C20, 0xE050, 0x4180}},
    {18770, 2, {0x7CF8, 0x488, 0x4F8, 0x7C20, 0x41FC, 0x7D24, 0x5FC, 0x428, 0x2BFC, 0x16C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18771, 2, {0x1088, 0x1088, 0x29FE, 0x2488, 0x40F8, 0xB888, 0x10F8, 0x1088, 0xFDFE, 0x1088, 0x9524, 0x5AFA, 0x5020, 0x1C20, 0xE1FE, 0x4000}},
    {18772, 2, {0x2000, 0x23FE, 0x5050, 0x4850, 0x83FE, 0x7A52, 0x2252, 0x23FE, 0xF924, 0x20A8, 0x2BFE, 0xB070, 0x60A8, 0x3924, 0xE622, 0x20}},
    {18773, 2, {0x2108, 0x2108, 0x5548, 0x4B90, 0x811E, 0x77D4, 0x2564, 0x2554, 0xF554, 0x2754, 0xA5D4, 0x7548, 0x2548, 0x3554, 0xC464, 0x4C2}},
    {18774, 2, {0x1088, 0x1050, 0x2BFE, 0x2450, 0x41FC, 0xB954, 0x118C, 0x1174, 0xFD04, 0x11FC, 0x9420, 0x5BFE, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {18775, 2, {0x2108, 0x2108, 0x57C8, 0x4908, 0x811E, 0x7B92, 0x2024, 0x2380, 0xF808, 0x27C8, 0x2908, 0xB108, 0x6594, 0x3954, 0xE124, 0x342}},
    {18776, 2, {0x1088, 0x1088, 0x2954, 0x2622, 0x4088, 0xB888, 0x1154, 0x1222, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {18777, 2, {0x2044, 0x2724, 0x5528, 0x4D7E, 0x8510, 0x7630, 0x254A, 0x251A, 0xF52C, 0x254C, 0xA51A, 0x762A, 0x2448, 0x3408, 0xC428, 0x410}},
    {18778, 2, {0x3FF8, 0x2008, 0x2EE8, 0x2AA8, 0x2EE8, 0x2AA8, 0x2EE8, 0x2AA8, 0x2FE8, 0x2288, 0x2448, 0x3BB8, 0x210A, 0x4FEA, 0x4546, 0x9FF2}},
    {18779, 2, {0x1080, 0x13DE, 0x288A, 0x25CA, 0x408A, 0xBBD2, 0x10A6, 0x1040, 0xFC88, 0x11F0, 0x9424, 0x5BFE, 0x5022, 0x1D24, 0xE2A2, 0x4040}},
    {18780, 2, {0x2000, 0x23DE, 0x5252, 0x4BDE, 0x8252, 0x7BDE, 0x2222, 0x2222, 0xFAFA, 0x2222, 0x2A72, 0xB2AA, 0x6326, 0x3A22, 0xE22A, 0x204}},
    {18781, 2, {0x2020, 0x23FE, 0x5250, 0x4BFE, 0x8252, 0x7BFE, 0x2294, 0x22D8, 0xFA92, 0x22CE, 0x2A00, 0xB284, 0x6298, 0x3CE2, 0xE482, 0x87E}},
    {18782, 2, {0x2108, 0x2208, 0x27C8, 0x5450, 0x8FDE, 0x7464, 0x27D4, 0x2214, 0xF914, 0x27D4, 0x2A14, 0xB3C8, 0x6248, 0x3C54, 0xE554, 0x8A2}},
    {18783, 2, {0x2200, 0x227C, 0x5244, 0x4A7C, 0x8744, 0x7A7C, 0x2220, 0x237E, 0xFA92, 0x2652, 0x2A6A, 0xB242, 0x627A, 0x3A02, 0xE214, 0x608}},
    {18784, 2, {0x2088, 0x2088, 0x53FE, 0x4888, 0x8014, 0x7812, 0x23FE, 0x2210, 0xFAF4, 0x2214, 0x2AF4, 0xB2A8, 0x62AA, 0x3AFA, 0xE216, 0x422}},
    {18785, 2, {0x2114, 0x2554, 0x5554, 0x4DFE, 0x8554, 0x77D4, 0x245C, 0x2440, 0xF77E, 0x2510, 0xA5FE, 0x7510, 0x2538, 0x3554, 0xC592, 0x910}},
    {18786, 2, {0x810, 0xFF7E, 0x810, 0x7F7C, 0x4104, 0x7F28, 0x2210, 0x1F28, 0xE144, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18787, 2, {0x2C78, 0x4448, 0x6C48, 0x4486, 0x7D78, 0x48, 0x7C30, 0x1048, 0xFD84, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18788, 2, {0x11FC, 0x1020, 0x2BFE, 0x2622, 0x41AC, 0xB820, 0x11AC, 0x1040, 0xFCFC, 0x1104, 0x96F4, 0x5894, 0x50F4, 0x1C8A, 0xE082, 0x407E}},
    {18789, 2, {0x2020, 0x23FE, 0x5088, 0x4850, 0x83FE, 0x7A50, 0x23FC, 0x2254, 0xFBFE, 0x2254, 0x2BFC, 0xB250, 0x62D8, 0x3D54, 0xE652, 0x850}},
    {18790, 2, {0x2020, 0x21FE, 0x5440, 0x4AFC, 0x8110, 0x7AFE, 0x2000, 0x26FC, 0xFA84, 0x22FC, 0x2A84, 0xB2FC, 0x6284, 0x3A8C, 0xE500, 0x8FE}},
    {18791, 2, {0x2040, 0x207C, 0x5040, 0x4BFE, 0x8202, 0x79FC, 0x20A8, 0x2154, 0xFA88, 0x21FC, 0x2A8A, 0xB0F8, 0x6088, 0x38F8, 0xE088, 0xF8}},
    {18792, 2, {0x2000, 0x23DE, 0x5252, 0x4BDE, 0x8252, 0x7BDE, 0x2202, 0x22FA, 0xFA8A, 0x22FA, 0x2A02, 0xB2FA, 0x628A, 0x3AFA, 0xE202, 0x206}},
    {18793, 2, {0x2288, 0x22A8, 0x56A8, 0x4BF0, 0x801E, 0x7224, 0x25D4, 0x2C14, 0xF414, 0x25D4, 0xA554, 0x7548, 0x2568, 0x3554, 0xC614, 0x422}},
    {18794, 2, {0x2000, 0x23FE, 0x5202, 0x4BFE, 0x8202, 0x7BFE, 0x2108, 0x2252, 0xFB9C, 0x2108, 0x2A52, 0xB3DE, 0x6000, 0x3AA4, 0xE252, 0x452}},
    {18795, 2, {0x2110, 0x2114, 0x57D2, 0x4910, 0x8FFE, 0x7290, 0x2450, 0x2FF4, 0xF494, 0x27F4, 0xA494, 0x77E8, 0x248A, 0x37FA, 0xC426, 0x42}},
    {18796, 2, {0x1088, 0x13FE, 0x2888, 0x2424, 0x43FE, 0xB820, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x95FC, 0x5924, 0x5008, 0x1DFE, 0xE088, 0x4058}},
    {18797, 2, {0x1414, 0x1010, 0xFEFE, 0x1010, 0x5454, 0x3838, 0x5454, 0x9292, 0x3130, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18798, 2, {0x2000, 0x23FE, 0x5050, 0x4BFE, 0x8252, 0x7BFE, 0x2000, 0x23DE, 0xFA92, 0x23D2, 0x2A52, 0xB3D2, 0x629A, 0x3A94, 0xE3D0, 0x10}},
    {18799, 2, {0x2000, 0x23FC, 0x5294, 0x4A94, 0x83FC, 0x7910, 0x2252, 0x23DC, 0xF812, 0x23CE, 0x2A40, 0xB3D2, 0x625C, 0x3BD0, 0xE252, 0x2CE}},
    {18800, 2, {0x2040, 0x2020, 0x57FE, 0x4900, 0x8100, 0x79FC, 0x2000, 0x2088, 0xFBFE, 0x22AA, 0x2AAA, 0xB3FE, 0x6088, 0x38AA, 0xE3FE, 0x22}},
    {18801, 2, {0x2020, 0x23FE, 0x5202, 0x4840, 0x819C, 0x7904, 0x21DC, 0x2104, 0xF9FC, 0x2100, 0x2BFE, 0xB402, 0x62AA, 0x3AAA, 0xE414, 0x8}},
    {18802, 2, {0x2008, 0x27E8, 0x5408, 0x4DC8, 0x855E, 0x7552, 0x25E4, 0x2400, 0xF7E8, 0x26A8, 0xA6A8, 0x77E8, 0x2414, 0x37F4, 0xC024, 0x42}},
    {18803, 2, {0x23FC, 0x2040, 0x51F8, 0x4908, 0x81F8, 0x7908, 0x27FE, 0x2442, 0xFAA4, 0x228A, 0x2C7A, 0xB080, 0x61F8, 0x3A88, 0xE070, 0x38E}},
    {18804, 2, {0x2020, 0x2222, 0x53FE, 0x4890, 0x81FE, 0x7B10, 0x25FE, 0x2110, 0xF9FE, 0x2110, 0x29FE, 0xB100, 0x63DE, 0x3A52, 0xE272, 0x206}},
    {18805, 2, {0x2010, 0x2010, 0x577E, 0x4D10, 0x85FE, 0x7544, 0x25EE, 0x2744, 0xF564, 0x25DE, 0xA500, 0x75FE, 0x2728, 0x352A, 0xC04A, 0x86}},
    {18806, 2, {0x2000, 0x23FE, 0x5222, 0x4ACC, 0x8244, 0x7BEE, 0x2244, 0x22EE, 0xFB54, 0x2244, 0x2A10, 0xB210, 0x625E, 0x3A50, 0xE250, 0x5FE}},
    {18807, 2, {0x2442, 0x2772, 0x2AA2, 0x511A, 0x8BEA, 0x722A, 0x23EA, 0x222A, 0xFBEA, 0x222A, 0x2BEA, 0xB14A, 0x67F2, 0x3942, 0xE24A, 0x444}},
    {18808, 2, {0xF7FC, 0x1248, 0x17FC, 0xFC46, 0x86EC, 0xF554, 0x16EC, 0x1554, 0x57EC, 0x26C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {18809, 2, {0x2080, 0x23DE, 0x508A, 0x49D2, 0x80A6, 0x7BD0, 0x209C, 0x2050, 0xFBFE, 0x2050, 0x2AAA, 0xB3FE, 0x6252, 0x3AAA, 0xE3FE, 0x2}},
    {18810, 2, {0x2000, 0x23FC, 0x3810, 0x2110, 0x4110, 0x7910, 0xA210, 0x23FE, 0xF830, 0x2050, 0x2090, 0x2110, 0x2A10, 0x3410, 0x2050, 0x20}},
    {18811, 2, {0x1000, 0x1000, 0x3DFC, 0x2020, 0x4020, 0xBC20, 0x1020, 0x13FE, 0xFC50, 0x1050, 0x1050, 0x1090, 0x1492, 0x1912, 0x120E, 0x400}},
    {18812, 2, {0x1080, 0x1080, 0x3C80, 0x20FC, 0x4154, 0xBE54, 0x1054, 0x1094, 0xFC94, 0x1124, 0x1224, 0x1044, 0x1444, 0x1884, 0x1128, 0x10}},
    {18813, 2, {0x2040, 0x2048, 0x3A44, 0x2240, 0x43FE, 0x7880, 0xA080, 0x20FC, 0xF944, 0x2144, 0x2148, 0x2250, 0x2A20, 0x3450, 0x2888, 0x306}},
    {18814, 2, {0x2000, 0x23FE, 0x3A02, 0x2484, 0x4080, 0x78FC, 0xA080, 0x2100, 0xF9FC, 0x2004, 0x2004, 0x23F4, 0x2804, 0x3004, 0x2028, 0x10}},
    {18815, 2, {0x1008, 0x101C, 0x3DE0, 0x2100, 0x4100, 0xBDFE, 0x1120, 0x1120, 0xFD3C, 0x1124, 0x1124, 0x1124, 0x1524, 0x1A44, 0x1254, 0x488}},
    {18816, 2, {0x2000, 0x27BC, 0x3884, 0x2084, 0x44A4, 0x7A94, 0xA294, 0x2084, 0xF98C, 0x2294, 0x24A4, 0x2084, 0x2884, 0x3084, 0x2294, 0x108}},
    {18817, 2, {0x2100, 0x211E, 0x3912, 0x27D2, 0x4114, 0x7914, 0xA7D8, 0x2114, 0xF912, 0x27D2, 0x2112, 0x211A, 0x2914, 0x3210, 0x2210, 0x410}},
    {18818, 2, {0x2040, 0x2020, 0x3BFC, 0x2204, 0x4204, 0x7BFC, 0xA200, 0x23DC, 0xFA44, 0x2354, 0x22CC, 0x2244, 0x2ACC, 0x3554, 0x2444, 0x8CC}},
    {18819, 2, {0x2088, 0x2088, 0x3BFE, 0x20A8, 0x4022, 0x79FA, 0xA024, 0x2028, 0xFBFE, 0x2040, 0x20FC, 0x2184, 0x2AFC, 0x3484, 0x20FC, 0x84}},
    {18820, 2, {0x1020, 0x13FE, 0x3C20, 0x21FC, 0x4124, 0xBDFC, 0x1124, 0x11FC, 0xFC22, 0x13FE, 0x1050, 0x1094, 0x1588, 0x1AA8, 0x14C4, 0x82}},
    {18821, 2, {0x2088, 0x2050, 0x3BFE, 0x2020, 0x41FC, 0x7820, 0xA3FE, 0x2124, 0xF8A8, 0x23FE, 0x2000, 0x21FC, 0x2904, 0x3104, 0x21FC, 0x104}},
    {18822, 2, {0x23DE, 0x2252, 0x3BDE, 0x2252, 0x43DE, 0x78A0, 0xA1FE, 0x2320, 0xFDFC, 0x2120, 0x21FE, 0x2000, 0x29FC, 0x3088, 0x2070, 0x78E}},
    {18823, 2, {0x10, 0x3E50, 0x2050, 0x2048, 0x3C88, 0x20A4, 0x3D24, 0x2242, 0x2040, 0xFE80, 0x10F8, 0x2008, 0x4408, 0xFE08, 0x250, 0x20}},
    {18824, 2, {0x40, 0x3E40, 0x2040, 0x2040, 0x3DFE, 0x2040, 0x3C40, 0x2040, 0x20FC, 0xFE90, 0x1090, 0x2110, 0x4510, 0xFE10, 0x2FE, 0x0}},
    {18825, 2, {0x0, 0x3EFE, 0x20AA, 0x20AA, 0x3CAA, 0x20AA, 0x3CAA, 0x20AA, 0x20AA, 0xFEAE, 0x10C2, 0x2082, 0x4482, 0xFEFE, 0x282, 0x0}},
    {18826, 2, {0x20, 0x3E20, 0x2020, 0x203C, 0x3C20, 0x2020, 0x3DFE, 0x2000, 0x2020, 0xFE20, 0x10A8, 0x20A4, 0x4522, 0xFE22, 0x2A0, 0x40}},
    {18827, 2, {0x88, 0x3C88, 0x2088, 0x3BD0, 0x209E, 0x3BE4, 0x2094, 0x27D4, 0xF914, 0x21D4, 0x2154, 0x4948, 0xF948, 0x4A54, 0x2D4, 0x422}},
    {18828, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {18829, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4404, 0x47F4, 0x4804, 0x57C4, 0x4184, 0x4204, 0x4414, 0x4414, 0x43F4, 0x400C}},
    {18830, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FC4, 0x4004, 0x5FE4, 0x4504, 0x4504, 0x4504, 0x4924, 0x50E4, 0x4014, 0x4008}},
    {18831, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4284, 0x4464, 0x581C, 0x4444, 0x4444, 0x4444, 0x4844, 0x5054, 0x4008}},
    {18832, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x40E4, 0x5F04, 0x4104, 0x5FE4, 0x4104, 0x7FFC, 0x4104, 0x4514, 0x4208}},
    {18833, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4204, 0x4FC4, 0x4244, 0x4244, 0x5FF4, 0x4204, 0x4504, 0x4484, 0x4854, 0x5008}},
    {18834, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FE4, 0x4104, 0x4204, 0x47C4, 0x4C44, 0x5444, 0x4444, 0x47C4, 0x4454, 0x4008}},
    {18835, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x41F4, 0x4104, 0x4104, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {18836, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5FF4, 0x4004, 0x400C}},
    {18837, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4824, 0x4444, 0x5FF4, 0x4444, 0x4444, 0x7FFC, 0x4444, 0x4844, 0x4844, 0x504C}},
    {18838, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x400C}},
    {18839, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4204, 0x4FE4, 0x4AA4, 0x4AA4, 0x4AA4, 0x4AA4, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {18840, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4914, 0x48A4, 0x4A44, 0x4C24, 0x481C}},
    {18841, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x7FFC, 0x4044, 0x5FF4, 0x4844, 0x4444, 0x4154, 0x4088}},
    {18842, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x40E4, 0x4F04, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {18843, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x47C4, 0x4884, 0x5FE4, 0x4124, 0x5FF4, 0x4124, 0x4FE4, 0x4104, 0x4304, 0x400C}},
    {18844, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4204, 0x4FE4, 0x4424, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FF4, 0x4008}},
    {18845, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4284, 0x4AA4, 0x4A94, 0x5494, 0x4484, 0x4994, 0x4008}},
    {18846, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x49F4, 0x4444, 0x5044, 0x49F4, 0x4444, 0x4844, 0x5844, 0x4BFC, 0x4804, 0x400C}},
    {18847, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4924, 0x4924, 0x5554, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x400C}},
    {18848, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4204, 0x4514, 0x59A4, 0x4AC4, 0x45A4, 0x5A94, 0x4494, 0x5A84, 0x410C}},
    {18849, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x5FF4, 0x5014, 0x47C4, 0x4444, 0x47C4, 0x4404, 0x47E4, 0x4424, 0x47E4, 0x400C}},
    {18850, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x5084, 0x4884, 0x63E4, 0x5084, 0x4884, 0x57F4, 0x7104, 0x5224, 0x57F4, 0x5014, 0x400C}},
    {18851, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7E7C, 0x4104, 0x5FF4, 0x4824, 0x4444, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x400C}},
    {18852, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4EE4, 0x4AA4, 0x4444, 0x4AA4, 0x5104, 0x4EE4, 0x4AA4, 0x4444, 0x4AA4, 0x510C}},
    {18853, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4444, 0x4444, 0x5EF4, 0x4444, 0x4EE4, 0x5554, 0x644C, 0x4444, 0x4454, 0x4008}},
    {18854, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4284, 0x4C64, 0x77DC, 0x4104, 0x5FF4, 0x4924, 0x4544, 0x7FFC, 0x4004, 0x400C}},
    {18855, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4204, 0x4CE4, 0x4824, 0x4EE4, 0x4824, 0x4FE4, 0x4284, 0x4494, 0x5874, 0x400C}},
    {18856, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4544, 0x4924, 0x410C}},
    {18857, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4404, 0x4FC4, 0x5444, 0x4394, 0x5C68}},
    {18858, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4404, 0x4FF4, 0x5914, 0x6A94, 0x4FD4, 0x4054, 0x402C}},
    {18859, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4444, 0x5FF4, 0x4444, 0x4104, 0x4FE4, 0x4924, 0x5FF4, 0x4284, 0x4444, 0x482C}},
    {18860, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4404, 0x5EF4, 0x4454, 0x4E54, 0x4454, 0x5EB4, 0x4504, 0x5FF4, 0x4284, 0x4444, 0x482C}},
    {18861, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x40E4, 0x4F04, 0x4104, 0x5FF4, 0x4544, 0x4924, 0x57D4, 0x4444, 0x47C4, 0x4454, 0x47C8}},
    {18862, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4FE4, 0x4104, 0x4FE4, 0x4204, 0x5FF4, 0x4444, 0x4FE4, 0x545C, 0x67C4, 0x4444, 0x47CC}},
    {18863, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x7F44, 0x4844, 0x527C, 0x5F94, 0x4454, 0x5F54, 0x4424, 0x4754, 0x7C94, 0x510C}},
    {18864, 2, {0x783C, 0x4824, 0x7A3C, 0x4BA4, 0x7CBC, 0x4FC4, 0x5544, 0x47C4, 0x4544, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4114, 0x40F4}},
    {18865, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x41E4, 0x5E24, 0x4544, 0x4F84, 0x4884, 0x4FC4, 0x4844, 0x4FE4, 0x4824, 0x5524, 0x656C}},
    {18866, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x47C4, 0x4844, 0x5FF4, 0x6A24, 0x4C94, 0x4FF4, 0x4804, 0x4BE4, 0x4A24, 0x53EC}},
    {18867, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x5204, 0x5774, 0x6D54, 0x7754, 0x5564, 0x6F54, 0x7C54, 0x5654, 0x5564, 0x66C4, 0x444C}},
    {18868, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x4AA4, 0x4FE4, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4444, 0x4FE4, 0x4104, 0x5FF4, 0x410C}},
    {18869, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4924, 0x5FF4, 0x711C, 0x5BB4, 0x5554, 0x5BB4, 0x5554, 0x5BB4, 0x511C}},
    {18870, 2, {0x2000, 0x13FC, 0x1004, 0x4204, 0x47C4, 0x4844, 0x5084, 0x4FE4, 0x4124, 0x5FF4, 0x4124, 0x4FE4, 0x4104, 0x4504, 0x4214, 0x4008}},
    {18871, 2, {0x2000, 0x13FC, 0x1004, 0x4004, 0x4104, 0x4924, 0x4924, 0x5554, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {18872, 2, {0x2000, 0x17FC, 0x4, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x5FF4, 0x5114, 0x5154, 0x57D4, 0x5054, 0x5034, 0x400C}},
    {18873, 2, {0x20, 0x7C20, 0x4420, 0x4820, 0x4820, 0x5020, 0x4BFE, 0x4820, 0x4420, 0x4420, 0x4420, 0x6820, 0x5020, 0x4020, 0x4020, 0x4020}},
    {18874, 2, {0x0, 0x7C00, 0x45FE, 0x4820, 0x4820, 0x5020, 0x4820, 0x4820, 0x4420, 0x4420, 0x4420, 0x6820, 0x5020, 0x4020, 0x40A0, 0x4040}},
    {18875, 2, {0x0, 0x7DFC, 0x4404, 0x4808, 0x4810, 0x5020, 0x4820, 0x4BFE, 0x4420, 0x4420, 0x4420, 0x6820, 0x5020, 0x4020, 0x40A0, 0x4040}},
    {18876, 2, {0x40, 0x7C40, 0x4440, 0x4BF8, 0x4848, 0x5048, 0x4848, 0x4848, 0x47FE, 0x4440, 0x44A0, 0x68A0, 0x5110, 0x4110, 0x4208, 0x4406}},
    {18877, 2, {0x0, 0x7DF8, 0x4408, 0x4850, 0x4820, 0x5010, 0x4BFE, 0x4822, 0x4424, 0x4420, 0x4420, 0x6820, 0x5020, 0x4020, 0x40A0, 0x4040}},
    {18878, 2, {0x1040, 0x1040, 0x2080, 0x7C88, 0x4504, 0x47FE, 0x7C02, 0x4090, 0x4090, 0x7C90, 0x4490, 0x4490, 0x7D12, 0x4512, 0x20E, 0x400}},
    {18879, 2, {0x20, 0x7C20, 0x4420, 0x4820, 0x49FC, 0x5124, 0x4924, 0x4924, 0x4524, 0x45FC, 0x4524, 0x6820, 0x5020, 0x4020, 0x4020, 0x4020}},
    {18880, 2, {0x40, 0x7C40, 0x4440, 0x4840, 0x4BFE, 0x5040, 0x4880, 0x4890, 0x4490, 0x4520, 0x4520, 0x6A48, 0x5244, 0x4484, 0x41FE, 0x4082}},
    {18881, 2, {0x80, 0x7C80, 0x4500, 0x49FC, 0x4A04, 0x5404, 0x49E4, 0x4924, 0x4524, 0x4524, 0x45E4, 0x6924, 0x5004, 0x4004, 0x4028, 0x4010}},
    {18882, 2, {0x0, 0x7DFC, 0x4444, 0x4844, 0x4844, 0x5044, 0x4894, 0x4888, 0x4500, 0x44FC, 0x4484, 0x6884, 0x5084, 0x4084, 0x40FC, 0x4084}},
    {18883, 2, {0x0, 0x7C00, 0x45FC, 0x4924, 0x4924, 0x5124, 0x4924, 0x4924, 0x45FC, 0x4524, 0x4524, 0x6924, 0x5124, 0x4124, 0x41FC, 0x4104}},
    {18884, 2, {0x80, 0x7C80, 0x44F8, 0x4908, 0x4B10, 0x54A0, 0x4840, 0x48A0, 0x4518, 0x4606, 0x45F8, 0x6908, 0x5108, 0x4108, 0x41F8, 0x4108}},
    {18885, 2, {0x40, 0x7840, 0x4BFC, 0x4840, 0x53F8, 0x5048, 0x63F8, 0x5240, 0x4BFC, 0x4844, 0x4854, 0x68A8, 0x50A0, 0x4110, 0x4208, 0x4406}},
    {18886, 2, {0x90, 0x7C90, 0x4490, 0x4890, 0x4BFC, 0x5090, 0x4890, 0x4890, 0x4490, 0x47FE, 0x4400, 0x6890, 0x5088, 0x4108, 0x4104, 0x4204}},
    {18887, 2, {0x40, 0x7820, 0x4820, 0x4BFE, 0x5040, 0x5040, 0x60A0, 0x50A2, 0x49A4, 0x4A98, 0x4C90, 0x6888, 0x5084, 0x40A2, 0x40C0, 0x4080}},
    {18888, 2, {0x0, 0x7BFE, 0x4850, 0x5050, 0x5050, 0x61FC, 0x5154, 0x4954, 0x4954, 0x4954, 0x695C, 0x5184, 0x4104, 0x4104, 0x41FC, 0x4104}},
    {18889, 2, {0x0, 0x7DFC, 0x4524, 0x4924, 0x49FC, 0x5124, 0x4924, 0x49FC, 0x4420, 0x4420, 0x45FC, 0x6820, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {18890, 2, {0x20, 0x7D20, 0x4520, 0x49FC, 0x4920, 0x5220, 0x4820, 0x4BFE, 0x4400, 0x4400, 0x45FC, 0x6904, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {18891, 2, {0x0, 0x7DFC, 0x4504, 0x4904, 0x49FC, 0x5104, 0x4904, 0x49FC, 0x4504, 0x4504, 0x45FC, 0x6850, 0x5090, 0x4092, 0x4112, 0x420E}},
    {18892, 2, {0x8, 0x783C, 0x4BC0, 0x5044, 0x5224, 0x6128, 0x5100, 0x4840, 0x4BFE, 0x4888, 0x6908, 0x5390, 0x4060, 0x4050, 0x4188, 0x4604}},
    {18893, 2, {0x88, 0x7C84, 0x4502, 0x4A22, 0x4820, 0x5050, 0x4888, 0x4904, 0x4602, 0x45FC, 0x4504, 0x6904, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {18894, 2, {0x0, 0x7CFC, 0x4484, 0x4884, 0x48FC, 0x5000, 0x49FE, 0x4902, 0x4502, 0x45FE, 0x4502, 0x6902, 0x51FE, 0x4102, 0x410A, 0x4104}},
    {18895, 2, {0x40, 0x7C40, 0x44FC, 0x4888, 0x4950, 0x5020, 0x48D8, 0x4B26, 0x44F8, 0x4420, 0x44F8, 0x6820, 0x53FE, 0x4020, 0x4020, 0x4020}},
    {18896, 2, {0x40, 0x7C20, 0x47FE, 0x4800, 0x49FC, 0x5104, 0x49FC, 0x4800, 0x45FC, 0x4408, 0x4410, 0x6BFE, 0x5020, 0x4020, 0x40A0, 0x4040}},
    {18897, 2, {0x40, 0x7820, 0x4BFE, 0x5202, 0x5404, 0x6000, 0x53FE, 0x4820, 0x4820, 0x4920, 0x693C, 0x5120, 0x42A0, 0x4260, 0x443E, 0x4800}},
    {18898, 2, {0x20, 0x7C20, 0x4420, 0x4BFE, 0x4820, 0x5124, 0x4924, 0x4924, 0x46AA, 0x4470, 0x44A8, 0x68A8, 0x5124, 0x4222, 0x4420, 0x4020}},
    {18899, 2, {0x0, 0x7DFC, 0x4524, 0x4924, 0x4974, 0x5124, 0x49FC, 0x4904, 0x4574, 0x4554, 0x4554, 0x6974, 0x5104, 0x4204, 0x4214, 0x4408}},
    {18900, 2, {0x20, 0x7C20, 0x4450, 0x4888, 0x4944, 0x5222, 0x49F8, 0x4808, 0x4450, 0x4420, 0x44A4, 0x6A82, 0x528A, 0x428A, 0x4478, 0x4000}},
    {18901, 2, {0x14, 0x7812, 0x4810, 0x57FE, 0x5010, 0x6010, 0x53D2, 0x4A52, 0x4A52, 0x4A54, 0x6BD4, 0x5008, 0x40EA, 0x471A, 0x4226, 0x4042}},
    {18902, 2, {0x48, 0x7C44, 0x445E, 0x49E0, 0x4828, 0x5012, 0x486A, 0x4996, 0x4448, 0x445E, 0x45E0, 0x6824, 0x5028, 0x4012, 0x406A, 0x4186}},
    {18903, 2, {0x100, 0x7D00, 0x45FE, 0x4A00, 0x4DFC, 0x5104, 0x49FC, 0x4904, 0x45FC, 0x4480, 0x44FC, 0x6908, 0x5290, 0x4060, 0x4198, 0x4606}},
    {18904, 2, {0x100, 0x79F8, 0x4A08, 0x53F0, 0x5010, 0x67FE, 0x5080, 0x4944, 0x4E68, 0x48B0, 0x6928, 0x5668, 0x40A4, 0x4122, 0x46A0, 0x4040}},
    {18905, 2, {0x40, 0x7A48, 0x4A48, 0x5248, 0x53F8, 0x6040, 0x50A0, 0x4910, 0x4A88, 0x4C86, 0x6890, 0x52A8, 0x4294, 0x4494, 0x4280, 0x4100}},
    {18906, 2, {0x20, 0x7C10, 0x45FE, 0x4800, 0x48FC, 0x5084, 0x48FC, 0x4800, 0x45FE, 0x4502, 0x457A, 0x694A, 0x517A, 0x4102, 0x410A, 0x4104}},
    {18907, 2, {0x40, 0x7820, 0x4BFC, 0x5108, 0x5090, 0x63FE, 0x5202, 0x4C44, 0x4820, 0x4BFC, 0x6880, 0x50F8, 0x4088, 0x4108, 0x4128, 0x4210}},
    {18908, 2, {0x20, 0x7C10, 0x45FE, 0x4910, 0x497C, 0x5114, 0x49FE, 0x4914, 0x457C, 0x4510, 0x457C, 0x6944, 0x5144, 0x4244, 0x427C, 0x4444}},
    {18909, 2, {0x20, 0x7D24, 0x4524, 0x49FC, 0x4820, 0x53FE, 0x4800, 0x4820, 0x45FC, 0x4524, 0x4524, 0x69FC, 0x5020, 0x4024, 0x43FE, 0x4102}},
    {18910, 2, {0x0, 0x7DFE, 0x4520, 0x49FC, 0x4920, 0x51FC, 0x4920, 0x4920, 0x45FE, 0x4402, 0x46AA, 0x6AAA, 0x52AA, 0x4202, 0x4014, 0x4008}},
    {18911, 2, {0x1E, 0x7DE0, 0x4422, 0x4912, 0x4894, 0x5080, 0x4820, 0x49CE, 0x4502, 0x4502, 0x45CE, 0x6902, 0x5102, 0x4102, 0x41FE, 0x4102}},
    {18912, 2, {0x20, 0x7BFE, 0x4820, 0x51FC, 0x5124, 0x61FC, 0x5124, 0x49FC, 0x4822, 0x4BFE, 0x6808, 0x53FE, 0x4108, 0x4088, 0x4028, 0x4010}},
    {18913, 2, {0x10, 0x7DFE, 0x4510, 0x497C, 0x4914, 0x51FE, 0x4914, 0x497C, 0x4510, 0x457C, 0x4554, 0x697C, 0x5154, 0x427C, 0x4254, 0x444C}},
    {18914, 2, {0x0, 0x7BFE, 0x4A00, 0x527C, 0x5244, 0x6244, 0x527C, 0x4A00, 0x4AEE, 0x4AAA, 0x6AAA, 0x52AA, 0x42EE, 0x4200, 0x43FE, 0x4000}},
    {18915, 2, {0x0, 0x7DFC, 0x4450, 0x49FC, 0x4954, 0x5154, 0x49FC, 0x4800, 0x45FC, 0x4400, 0x47FE, 0x6820, 0x50A8, 0x4124, 0x42A2, 0x4040}},
    {18916, 2, {0x50, 0x7A52, 0x4954, 0x5050, 0x53FE, 0x6088, 0x5050, 0x4BFE, 0x4820, 0x49FC, 0x6820, 0x53FE, 0x4050, 0x4088, 0x4104, 0x4602}},
    {18917, 2, {0x20, 0x7820, 0x4BFE, 0x5020, 0x51FC, 0x6124, 0x51FC, 0x4924, 0x49FC, 0x4822, 0x6BFE, 0x5042, 0x4024, 0x42A2, 0x428A, 0x4478}},
    {18918, 2, {0xE, 0x7BF0, 0x4844, 0x5224, 0x5108, 0x63F8, 0x5208, 0x4BFC, 0x4A04, 0x4BFE, 0x6A02, 0x52AA, 0x42AA, 0x4502, 0x4414, 0x4808}},
    {18919, 2, {0x104, 0x7888, 0x4BFE, 0x4820, 0x51FC, 0x5020, 0x63FE, 0x5054, 0x4992, 0x4890, 0x4BFE, 0x6890, 0x50D4, 0x438A, 0x4096, 0x41A2}},
    {18920, 2, {0x50, 0x7A52, 0x4954, 0x4850, 0x53FE, 0x5088, 0x6050, 0x53FE, 0x4820, 0x49FC, 0x4820, 0x6BFE, 0x50A8, 0x4124, 0x4622, 0x4020}},
    {18921, 2, {0x8, 0x7B88, 0x4888, 0x5110, 0x57DE, 0x6294, 0x52A4, 0x4B94, 0x4A94, 0x4B94, 0x6A94, 0x52C8, 0x4388, 0x4694, 0x40A4, 0x40C2}},
    {18922, 2, {0x20, 0x7810, 0x4BFE, 0x4A10, 0x52FE, 0x5254, 0x6254, 0x52BA, 0x4A28, 0x4A44, 0x4A82, 0x6A20, 0x52FE, 0x4444, 0x4438, 0x48C6}},
    {18923, 2, {0x20, 0x7BFE, 0x4800, 0x51FC, 0x5104, 0x61FC, 0x5000, 0x4BFE, 0x4A02, 0x49FC, 0x6840, 0x53A4, 0x40D8, 0x4334, 0x40D2, 0x4330}},
    {18924, 2, {0x40, 0x7820, 0x4BFE, 0x5202, 0x51FC, 0x6148, 0x5250, 0x49FC, 0x4B04, 0x49FC, 0x6904, 0x51FC, 0x4104, 0x41FC, 0x4088, 0x4104}},
    {18925, 2, {0x40, 0x7880, 0x4BFE, 0x5222, 0x528A, 0x6252, 0x5326, 0x4A52, 0x4AAA, 0x4BFE, 0x6840, 0x5024, 0x42A2, 0x428A, 0x4488, 0x4078}},
    {18926, 2, {0x20, 0x7BFE, 0x4A22, 0x51FC, 0x5020, 0x60F8, 0x5020, 0x49FC, 0x4820, 0x49FC, 0x6954, 0x51FC, 0x4040, 0x42A4, 0x428A, 0x447A}},
    {18927, 2, {0x1F0, 0x7A10, 0x4FFC, 0x5204, 0x53FC, 0x6224, 0x53B8, 0x4A22, 0x49FE, 0x4A10, 0x6FFC, 0x5244, 0x43FC, 0x40D0, 0x414A, 0x463E}},
    {18928, 2, {0x406, 0xE438, 0xAF08, 0xA910, 0xB224, 0xAFBC, 0xCA88, 0xAA94, 0xAFBE, 0xAA8A, 0xAA88, 0xAFAC, 0xC02A, 0x954A, 0x9548, 0x8018}},
    {18929, 2, {0x2050, 0x2048, 0x2080, 0x20FE, 0xFD90, 0x2690, 0x24FC, 0x2490, 0x2490, 0x24FC, 0x2590, 0x2690, 0x4490, 0x40FE, 0x8080, 0x80}},
    {18930, 2, {0x50, 0x7C48, 0x1080, 0x10FE, 0x1190, 0x1290, 0x10FC, 0xFE90, 0x1090, 0x10FC, 0x1090, 0x1090, 0x1090, 0x10FE, 0x1080, 0x1080}},
    {18931, 2, {0x850, 0x848, 0x880, 0x8FE, 0x4990, 0x4E90, 0x48FC, 0x4890, 0x4890, 0x48FC, 0x4890, 0x4E90, 0x5890, 0xE0FE, 0x80, 0x80}},
    {18932, 2, {0x1050, 0x1048, 0x1080, 0xFEFE, 0x1190, 0x1290, 0xFCFC, 0x490, 0x490, 0x48FC, 0x2890, 0x1090, 0x2890, 0x44FE, 0x8280, 0x80}},
    {18933, 2, {0x1050, 0x1048, 0x1080, 0xFEFE, 0x9390, 0x1290, 0x10FC, 0x2890, 0x2890, 0x28FC, 0x2890, 0x4A90, 0x4C90, 0x88FE, 0x80, 0x80}},
    {18934, 2, {0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1200, 0x100, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {18935, 2, {0x2050, 0x2048, 0x3C80, 0x44FE, 0xA990, 0x1290, 0x28FC, 0x4490, 0x8290, 0x30FC, 0x890, 0x90, 0x6090, 0x10FE, 0x880, 0x80}},
    {18936, 2, {0x1050, 0x1048, 0x1080, 0x10FE, 0xFD90, 0x1290, 0x10FC, 0x1090, 0x7C90, 0x44FC, 0x4490, 0x4490, 0x4490, 0x7CFE, 0x4480, 0x80}},
    {18937, 2, {0x100, 0x80, 0x3FFE, 0x2250, 0x2248, 0x24FE, 0x2490, 0x2D90, 0x36FE, 0x2490, 0x2490, 0x24FE, 0x4490, 0x4490, 0x84FE, 0x480}},
    {18938, 2, {0x828, 0x824, 0x840, 0xFF7E, 0x10C8, 0x1148, 0x207E, 0x3E48, 0x6248, 0xA27E, 0x2248, 0x2248, 0x2248, 0x3E7E, 0x2240, 0x40}},
    {18939, 2, {0x850, 0x848, 0x1480, 0x22FE, 0x4190, 0xBE90, 0xFC, 0x90, 0x3E90, 0x22FC, 0x2290, 0x2290, 0x2290, 0x3EFE, 0x2280, 0x80}},
    {18940, 2, {0x1050, 0x5048, 0x5080, 0x7CFE, 0x5190, 0x9290, 0x10FC, 0xFE90, 0x90, 0xFC, 0x7C90, 0x4490, 0x4490, 0x44FE, 0x7C80, 0x4480}},
    {18941, 2, {0x628, 0x7824, 0x840, 0x4A7E, 0x2AC8, 0x2D48, 0x87E, 0xFF48, 0x1848, 0x1C7E, 0x2A48, 0x2A48, 0x4848, 0x887E, 0x840, 0x840}},
    {18942, 2, {0x1050, 0x1048, 0x2840, 0x44FE, 0x8290, 0x7D90, 0x12FC, 0x1090, 0xFE90, 0x10FC, 0x1090, 0x7C90, 0x4490, 0x44FE, 0x7C80, 0x4480}},
    {18943, 2, {0x2428, 0x2424, 0x7E40, 0x247E, 0x24C8, 0xFF48, 0x7E, 0x7E48, 0x4248, 0x427E, 0x7E48, 0x4248, 0x4248, 0x7E7E, 0x4240, 0x40}},
    {18944, 2, {0x2050, 0x1048, 0x2880, 0x44FE, 0x9190, 0x1290, 0x54FC, 0x3890, 0x1090, 0xFEFC, 0x1090, 0x3890, 0x5490, 0x94FE, 0x1080, 0x1080}},
    {18945, 2, {0x7E20, 0x243E, 0x1848, 0xFFA8, 0x2910, 0x4A28, 0x9846, 0x900, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {18946, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x904, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {18947, 2, {0x28, 0x7E24, 0x4240, 0x7E7E, 0x42C8, 0x7F48, 0x87E, 0xFF48, 0x8148, 0x107E, 0xFF48, 0x2248, 0x6448, 0x187E, 0x2440, 0xC240}},
    {18948, 2, {0x828, 0x1424, 0x2240, 0x497E, 0xBEC8, 0x548, 0x87E, 0xFFC8, 0x1448, 0x7F7E, 0x5548, 0x6348, 0x4148, 0x7F7E, 0x4140, 0x7F40}},
    {18949, 2, {0x1028, 0x1024, 0xFE40, 0x107E, 0xAAC8, 0x4548, 0x827E, 0x7D48, 0x4448, 0x7C7E, 0x4448, 0x7C48, 0x1048, 0x547E, 0x9240, 0x3040}},
    {18950, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0xA00, 0x1FF8, 0x3100, 0x5FF0, 0x9100, 0x1FF0, 0x1100, 0x1FF8}},
    {18951, 2, {0x2850, 0x2448, 0x7EFE, 0xC990, 0x7EFC, 0x4890, 0x7EFC, 0x4890, 0x7EFE, 0x4080, 0x7DFC, 0x488, 0x2850, 0x1020, 0x2858, 0xC586}},
    {18952, 2, {0x2050, 0x1048, 0xFE80, 0x82FE, 0x7D90, 0x5690, 0x98FC, 0x7C90, 0xC490, 0x7CFC, 0x4490, 0x7C90, 0x4490, 0x7CFE, 0x2880, 0x4480}},
    {18953, 2, {0x28, 0x7724, 0x5540, 0x777E, 0x55C8, 0x7748, 0x557E, 0x7748, 0x2248, 0x557E, 0x1048, 0xFF48, 0x2248, 0x647E, 0x1C40, 0xE340}},
    {18954, 2, {0x880, 0x1FFC, 0x3080, 0xDFF8, 0x1080, 0x1FFC, 0x1428, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x1004, 0x1E7C, 0x1244, 0x2244}},
    {18955, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0xF0, 0x3F00, 0x200, 0x3F8, 0x7E02, 0x202, 0x1FE}},
    {18956, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x844, 0x1044, 0x603C}},
    {18957, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x820, 0x3FF8, 0x820, 0x820, 0xFFFE, 0x1020, 0x2020, 0x4020}},
    {18958, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x7F10, 0x110, 0x7F10, 0x4020, 0x7F24, 0x142, 0xAFE, 0x442}},
    {18959, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06}},
    {18960, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x100, 0x6C0, 0x1830, 0xE7CE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {18961, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x100, 0xFFFE, 0x100, 0x7FF8, 0x108, 0x3FF8, 0x2100, 0x3FFC, 0x284, 0x44C, 0x1830}},
    {18962, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x2244, 0x4A94, 0x408}},
    {18963, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {18964, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x1110, 0x920, 0xFFFE, 0x440, 0x842, 0x3042, 0xC03E}},
    {18965, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x11FC, 0x4820, 0x2020, 0xBFE, 0x7020, 0x1020, 0x10A0, 0x40}},
    {18966, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x400, 0x7FFC, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0x850, 0x820}},
    {18967, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0xFC, 0x7F00, 0x1210, 0x920, 0x3FF0, 0x40, 0xFFFE, 0x100, 0x500, 0x200}},
    {18968, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {18969, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x2038, 0x13C0, 0x8040, 0x4840, 0x17FE, 0xE040, 0x2040, 0x23FC}},
    {18970, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x800, 0x1FF8, 0x6488, 0x908, 0x3628, 0x110, 0x4894, 0x87F2, 0x0}},
    {18971, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x100C, 0xFDF0, 0x1100, 0x39FE, 0x5510, 0x9210, 0x1210, 0x1410}},
    {18972, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x4A94, 0x8508}},
    {18973, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {18974, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x1F8, 0x7F00, 0x100, 0x1FF0, 0x1110, 0x7FFC, 0x4124, 0x41F4, 0x5E14, 0x4008}},
    {18975, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2000, 0x1040, 0x83F8, 0x4040, 0x17FC, 0x2040, 0xE3F8, 0x2040, 0x2FFE}},
    {18976, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2104, 0x5284, 0x1454, 0x1FE8}},
    {18977, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x1FE0, 0x40, 0x5194, 0x4924, 0x5514, 0x4204, 0x7FFC, 0x4}},
    {18978, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1240, 0x1420, 0x1FF8}},
    {18979, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x820, 0x7FFC, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100}},
    {18980, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2040, 0x1040, 0x7FC, 0xF444, 0x17FC, 0x1444, 0x17FC, 0x2800, 0x47FE}},
    {18981, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x900, 0x1FF8, 0x3100, 0x5FF0, 0x9100, 0x1FFC, 0x1000, 0x3FF0, 0x820, 0x7C0, 0xF83E}},
    {18982, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x200, 0x3FF8, 0x820, 0x7FFC, 0x4104, 0x3FF8, 0x400, 0x7F0, 0x810, 0x1050, 0x2020}},
    {18983, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2020, 0x1050, 0xFE88, 0x526, 0x4410, 0x28FC, 0x2E08, 0xF070, 0x4008}},
    {18984, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x80, 0x2100, 0x13F8, 0x208, 0xF3F8, 0x1200, 0x13FC, 0x1204, 0x13FC, 0x2800, 0x47FE}},
    {18985, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2020, 0x1020, 0xFDFC, 0x924, 0x11FC, 0x3924, 0x55FC, 0x9020, 0x1020}},
    {18986, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x11FC, 0x1124, 0x11FC, 0xFD24, 0x11FC, 0x1020, 0x1DFC, 0xE020, 0x43FE, 0x0}},
    {18987, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x100, 0x7FFC, 0x1110, 0x2BA8, 0x4564, 0x1918, 0xE3E6, 0xE40, 0x180, 0xE40}},
    {18988, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2000, 0x1110, 0x8110, 0x4FBC, 0x1110, 0xE3B8, 0x2554, 0x2912, 0x2110}},
    {18989, 2, {0x10, 0x7F10, 0x854, 0xFFB8, 0x8890, 0x6B7C, 0x844, 0x6B44, 0x7C, 0x7E44, 0x44, 0xFF7C, 0x1044, 0x2244, 0x7F54, 0x148}},
    {18990, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x9414, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x3FF8, 0x8, 0x1FF8, 0x8, 0x3FF8}},
    {18991, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x20, 0x7BFE, 0x4820, 0x51FC, 0x6124, 0x51FC, 0x4924, 0x49FC, 0x6870, 0x50A8, 0x4124}},
    {18992, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x2110, 0x17FC, 0x9110, 0x41F0, 0x4840, 0xBF8, 0x1248, 0xE3F8, 0x2040, 0x2FFE, 0x2040}},
    {18993, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x1200, 0x27FC, 0x4208, 0x8BF8, 0x1208, 0x33F8, 0x5100, 0x93F8, 0x1510, 0x10E0, 0x171E}},
    {18994, 2, {0x20, 0x7F20, 0x83C, 0xFFC4, 0x8888, 0x6B7E, 0x852, 0x6B52, 0x7E, 0xFF90, 0x82C, 0x7F2A, 0x5528, 0x554A, 0x554A, 0x4386}},
    {18995, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2692, 0x450E}},
    {18996, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x5508, 0x3608, 0x7F7E, 0x1408, 0x7F48, 0x828, 0x3E28, 0x808, 0xF28, 0x7810}},
    {18997, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0xA7FC, 0x10A0, 0x17FC, 0x84A4, 0x47FC, 0x5208, 0x13F8, 0x2208, 0xE3F8, 0x2040, 0x27FC, 0x2040}},
    {18998, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x820, 0x1FF0, 0x10, 0x4FE4, 0x4AA4, 0x4964, 0x7FFC, 0x440, 0xFC7C, 0x2444, 0x4444}},
    {18999, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x1108, 0x7BCE, 0x1108, 0x3B88, 0x557E, 0xBF92, 0x2094, 0x3F88, 0x2094, 0x21A2}},
    {19000, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x18, 0x7BE0, 0x4950, 0x53F8, 0x6040, 0x53F8, 0x4808, 0x49F8, 0x6840, 0x5294, 0x44F4}},
    {19001, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x1080, 0x11FC, 0x2288, 0x2870, 0x6B8E, 0xA9FC, 0x29AC, 0x2974, 0x29FC, 0x2820, 0x2154}},
    {19002, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x9FF4, 0x440, 0xFE0, 0x0, 0x7C7C, 0x1010, 0xFEFE, 0x9292, 0x5454, 0xFEFE, 0x2828, 0x7C7C}},
    {19003, 2, {0x7C7C, 0x1010, 0xFEFE, 0x9292, 0x7C7C, 0x5454, 0x7C7C, 0x0, 0x7C7C, 0x1010, 0xFEFE, 0x9292, 0x7C7C, 0x5454, 0x7C7C, 0x0}},
    {19004, 2, {0x1044, 0x11FE, 0xFE54, 0x1048, 0x7CFE, 0x1190, 0xFEFE, 0x90, 0x7CFE, 0x4490, 0x7CFE, 0x4480, 0x7DFE, 0x4444, 0x5438, 0x49C6}},
    {19005, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440, 0x3FF0, 0x10, 0x10, 0x3FF0, 0x2000, 0x2004, 0x2004, 0x1FFC}},
    {19006, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2120, 0x2120, 0x2F3C, 0x2120, 0x2120, 0x2738, 0x2120, 0x2120, 0x4F3C, 0x4120, 0x8120, 0x120}},
    {19007, 2, {0x0, 0x3FF8, 0x100, 0x200, 0xFFFE, 0x820, 0x3018, 0xC446, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440}},
    {19008, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {19009, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x3FF0, 0x200, 0x7FF8, 0x202, 0xDFE, 0x7040, 0x1C78, 0xF1C4, 0x1C3C}},
    {19010, 2, {0x20, 0xFE20, 0x1020, 0x2050, 0xFE50, 0xAA88, 0xAB44, 0xBA22, 0xAA20, 0xAAF8, 0xBA08, 0xAA08, 0xAA10, 0xFE10, 0x8220, 0x20}},
    {19011, 2, {0x10, 0xFE90, 0x1090, 0x2092, 0xFE92, 0xAA94, 0xAAF4, 0xBA98, 0xAA90, 0xAA90, 0xBA90, 0xAA92, 0xAAB2, 0xFED2, 0x828E, 0x0}},
    {19012, 2, {0x0, 0xFE7C, 0x1044, 0x2044, 0xFE54, 0xAA54, 0xAA54, 0xBA54, 0xAA54, 0xAA54, 0xBA10, 0xAA28, 0xAA28, 0xFE4A, 0x828A, 0x106}},
    {19013, 2, {0x10, 0xFE10, 0x1010, 0x20FE, 0xFE92, 0xAA94, 0xAA90, 0xBAFC, 0xAAA4, 0xAAA4, 0xBAA8, 0xAAA8, 0xAA90, 0xFEA8, 0x8344, 0x282}},
    {19014, 2, {0x40, 0xFE40, 0x1040, 0x207E, 0xFEA0, 0xAAA0, 0xAB20, 0xBA3E, 0xAA20, 0xAA20, 0xBA20, 0xAA3E, 0xAA20, 0xFE20, 0x8220, 0x20}},
    {19015, 2, {0x10, 0xFE10, 0x1010, 0x2010, 0xFE1E, 0xAA10, 0xAA10, 0xBA10, 0xAAFE, 0xAA82, 0xBA82, 0xAA82, 0xAA82, 0xFE82, 0x82FE, 0x82}},
    {19016, 2, {0x40, 0xFE40, 0x107E, 0x2080, 0xFF7C, 0xAA44, 0xAA64, 0xBA54, 0xABFE, 0xAA44, 0xBAA4, 0xAA94, 0xAAFE, 0xFE04, 0x8228, 0x10}},
    {19017, 2, {0x14, 0xFE12, 0x10FE, 0x2010, 0xFE10, 0xAAFE, 0xAA92, 0xBA92, 0xAAFE, 0xAA92, 0xBA92, 0xAAFE, 0xAA92, 0xFE92, 0x8292, 0x86}},
    {19018, 2, {0x20, 0xFE10, 0x10FE, 0x2082, 0xFE40, 0xAA4E, 0xAA6A, 0xBAAA, 0xAAAA, 0xAAAA, 0xBB6C, 0xAA28, 0xAA4A, 0xFE4A, 0x828A, 0x106}},
    {19019, 2, {0x10, 0xFE08, 0x10FE, 0x20A4, 0xFEA4, 0xAAA4, 0xAAFE, 0xBAA4, 0xAAA4, 0xAAF6, 0xBAAC, 0xAAB4, 0xAAA4, 0xFEA4, 0x8324, 0x24}},
    {19020, 2, {0x28, 0xFE24, 0x1040, 0x207E, 0xFEC8, 0xAB48, 0xAA7E, 0xBA48, 0xAA48, 0xAA7E, 0xBA48, 0xAA48, 0xAA7E, 0xFE40, 0x82AA, 0x12A}},
    {19021, 2, {0x10, 0xFE10, 0x10FE, 0x2010, 0xFEAA, 0xAA44, 0xAA82, 0xBB7C, 0xAA44, 0xAA7C, 0xBA44, 0xAA7C, 0xAA10, 0xFE54, 0x8292, 0x30}},
    {19022, 2, {0x10, 0xFE10, 0x1028, 0x2044, 0xFE82, 0xAA7C, 0xAA00, 0xBAEE, 0xAAAA, 0xAAAA, 0xBAEE, 0xAA44, 0xAA44, 0xFEAA, 0x82AA, 0x112}},
    {19023, 2, {0x24, 0xFEFE, 0x1024, 0x2000, 0xFEFE, 0xAAAA, 0xAAAA, 0xBAFE, 0xAA14, 0xAAFE, 0xBA90, 0xAAD2, 0xAAAC, 0xFEAA, 0x8296, 0x122}},
    {19024, 2, {0x2220, 0x2220, 0xFF20, 0x227E, 0x3E40, 0x880, 0x7F3C, 0x4904, 0x4908, 0x7F10, 0x820, 0xFFC0, 0x842, 0x842, 0x83E, 0x800}},
    {19025, 2, {0x2200, 0x227E, 0xFF02, 0x2202, 0x3E02, 0x83E, 0x7F20, 0x4920, 0x4940, 0x7F7E, 0x802, 0xFF82, 0x802, 0x802, 0x814, 0x808}},
    {19026, 2, {0x2200, 0x227E, 0xFF10, 0x2210, 0x3E10, 0x810, 0x7F10, 0x49FE, 0x4910, 0x7F10, 0x810, 0xFF90, 0x810, 0x810, 0x850, 0x820}},
    {19027, 2, {0x2880, 0x2880, 0xFC80, 0x28FE, 0x3902, 0x1242, 0x7C42, 0x5482, 0x54A2, 0x7D12, 0x13FA, 0xFD0A, 0x1002, 0x1002, 0x1014, 0x1008}},
    {19028, 2, {0x2800, 0x28F8, 0xFE88, 0x2888, 0x3888, 0x1106, 0x7E00, 0x55FC, 0x5484, 0x7C84, 0x1048, 0xFE50, 0x1020, 0x1050, 0x1088, 0x1306}},
    {19029, 2, {0x2800, 0x2840, 0xFD9E, 0x2912, 0x3912, 0x1112, 0x7D12, 0x5512, 0x5512, 0x7D12, 0x115A, 0xFD94, 0x1110, 0x1010, 0x1010, 0x1010}},
    {19030, 2, {0x2820, 0x2820, 0xFE20, 0x28A8, 0x38A4, 0x10A2, 0x7D22, 0x5520, 0x5624, 0x7C24, 0x1028, 0xFE08, 0x1010, 0x1020, 0x10C0, 0x1300}},
    {19031, 2, {0x2800, 0x28FC, 0xFE84, 0x2884, 0x3884, 0x10FC, 0x7C84, 0x5484, 0x5484, 0x7CFC, 0x1084, 0xFE84, 0x1104, 0x1104, 0x1214, 0x1408}},
    {19032, 2, {0x2200, 0x2200, 0xFF7E, 0x2208, 0x3E10, 0x810, 0x7F24, 0x4944, 0x4978, 0x7F08, 0x810, 0xFF94, 0x822, 0x87E, 0x822, 0x800}},
    {19033, 2, {0x2820, 0x2810, 0xFE10, 0x29FE, 0x3820, 0x1020, 0x7C44, 0x5484, 0x55F8, 0x7C10, 0x1020, 0xFE44, 0x1082, 0x11FE, 0x1082, 0x1000}},
    {19034, 2, {0x2804, 0x281E, 0xFDF0, 0x2910, 0x3910, 0x1110, 0x7D10, 0x55FE, 0x5510, 0x7D10, 0x1110, 0xFD08, 0x110A, 0x114A, 0x11A6, 0x1112}},
    {19035, 2, {0x2840, 0x2820, 0xFC28, 0x2808, 0x3808, 0x1048, 0x7C54, 0x5552, 0x5552, 0x7D60, 0x1260, 0xFC44, 0x10C4, 0x1144, 0x123C, 0x1000}},
    {19036, 2, {0x2820, 0x2820, 0xFE20, 0x2820, 0x39FC, 0x1124, 0x7D24, 0x5524, 0x5524, 0x7DFC, 0x1124, 0xFF24, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {19037, 2, {0x2808, 0x281C, 0xFDF0, 0x2950, 0x3950, 0x1150, 0x7D50, 0x5550, 0x5550, 0x7D48, 0x1148, 0xFD68, 0x1154, 0x1274, 0x1252, 0x1400}},
    {19038, 2, {0x2820, 0x2820, 0xFE20, 0x2820, 0x383E, 0x1020, 0x7C20, 0x5420, 0x55FC, 0x7D04, 0x1104, 0xFF04, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {19039, 2, {0x2820, 0x2820, 0xFDFE, 0x2820, 0x39FC, 0x1024, 0x7DFC, 0x5520, 0x55FE, 0x7C22, 0x102A, 0xFC54, 0x1050, 0x1088, 0x1104, 0x1202}},
    {19040, 2, {0x5000, 0x5006, 0xFBB8, 0x5088, 0x7088, 0x2108, 0xF93E, 0xAB88, 0xA888, 0xFA88, 0x2288, 0xF93E, 0x2100, 0x2280, 0x247E, 0x2800}},
    {19041, 2, {0x2820, 0x2820, 0xFC20, 0x2BFE, 0x3820, 0x1020, 0x7D24, 0x5524, 0x5524, 0x7EAA, 0x1020, 0xFC50, 0x1050, 0x1088, 0x1104, 0x1202}},
    {19042, 2, {0x5104, 0x510E, 0xF930, 0x5120, 0x77E0, 0x2120, 0xF93E, 0xA964, 0xA9A4, 0xFF24, 0x2124, 0xF924, 0x2124, 0x2124, 0x2544, 0x2284}},
    {19043, 2, {0x2820, 0x2820, 0xFC50, 0x2888, 0x3904, 0x12FA, 0x7C20, 0x5420, 0x57FE, 0x7C20, 0x1128, 0xFD24, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {19044, 2, {0x2800, 0x28FC, 0xFE84, 0x2884, 0x38FC, 0x1084, 0x7C84, 0x54FC, 0x5484, 0x7C84, 0x10FC, 0xFE50, 0x1050, 0x1092, 0x1112, 0x120E}},
    {19045, 2, {0x5004, 0x500E, 0xFBB8, 0x5088, 0x7088, 0x2128, 0xF92E, 0xABA8, 0xA8A8, 0xFAA8, 0x22A8, 0xF93E, 0x2100, 0x2280, 0x247E, 0x2800}},
    {19046, 2, {0x1080, 0x20FC, 0x6910, 0xAA90, 0x2860, 0x2890, 0x230C, 0x440, 0x7FFC, 0x440, 0x7C0, 0x100, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {19047, 2, {0x2840, 0x2820, 0xFDFE, 0x2800, 0x3800, 0x10FC, 0x7C00, 0x5400, 0x54FC, 0x7C00, 0x1000, 0xFCFC, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {19048, 2, {0x2800, 0x28FC, 0xFE80, 0x28F8, 0x3880, 0x10F8, 0x7C80, 0x57FE, 0x5540, 0x7D24, 0x1128, 0xFF10, 0x1108, 0x1144, 0x1182, 0x1100}},
    {19049, 2, {0x5040, 0x5020, 0xFBFE, 0x5202, 0x7504, 0x2100, 0xF9DE, 0xAA52, 0xAA52, 0xFB52, 0x249A, 0xF894, 0x2110, 0x2112, 0x2212, 0x240E}},
    {19050, 2, {0x2820, 0x2810, 0xFDFE, 0x2902, 0x3A04, 0x10F8, 0x7C88, 0x5488, 0x54F8, 0x7C80, 0x1080, 0xFCFC, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {19051, 2, {0x5000, 0x53FE, 0xF820, 0x5020, 0x73FE, 0x2222, 0xFA22, 0xAB32, 0xAAAA, 0xFAAA, 0x2376, 0xFA66, 0x2222, 0x2222, 0x222A, 0x2204}},
    {19052, 2, {0x2820, 0x2820, 0xFDFE, 0x2820, 0x38FC, 0x1040, 0x7DFE, 0x5488, 0x5524, 0x7E22, 0x10F8, 0xFC20, 0x11FE, 0x1020, 0x1020, 0x1020}},
    {19053, 2, {0x2820, 0x2820, 0xFDFC, 0x2850, 0x3888, 0x1104, 0x7FFE, 0x5408, 0x55E8, 0x7D28, 0x1128, 0xFDE8, 0x1128, 0x1008, 0x1028, 0x1010}},
    {19054, 2, {0x2800, 0x29F8, 0xFC08, 0x2808, 0x39F8, 0x1008, 0x7C08, 0x57FE, 0x5420, 0x7E22, 0x1174, 0xFCA8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {19055, 2, {0x5000, 0x53FC, 0xF808, 0x5010, 0x7020, 0x23AE, 0xFAA2, 0xAAAA, 0xAAA4, 0xFAA4, 0x23AA, 0xF832, 0x20A0, 0x2040, 0x27FE, 0x2000}},
    {19056, 2, {0x2800, 0x28F8, 0xFE88, 0x2888, 0x38F8, 0x1000, 0x7DFE, 0x5488, 0x54F8, 0x7C88, 0x10F8, 0xFE88, 0x109E, 0x13E8, 0x1008, 0x1008}},
    {19057, 2, {0x5040, 0x50A0, 0xF910, 0x5208, 0x75F6, 0x2000, 0xFBC4, 0xAA54, 0xAA54, 0xFBD4, 0x2254, 0xFA54, 0x23D4, 0x2244, 0x2254, 0x22C8}},
    {19058, 2, {0x2840, 0x2820, 0xFDFE, 0x2948, 0x3948, 0x11FE, 0x7D48, 0x5578, 0x5500, 0x7DFC, 0x1144, 0xFD48, 0x1128, 0x1210, 0x1228, 0x14C6}},
    {19059, 2, {0x2840, 0x2820, 0xFDFE, 0x2922, 0x38A4, 0x1128, 0x7C50, 0x5488, 0x5504, 0x7CF8, 0x1088, 0xFC88, 0x1050, 0x1020, 0x10D8, 0x1306}},
    {19060, 2, {0x2880, 0x289E, 0xFC92, 0x29D2, 0x3892, 0x109E, 0x7C92, 0x55D2, 0x5552, 0x7D5E, 0x1152, 0xFD52, 0x11D2, 0x1022, 0x102A, 0x1044}},
    {19061, 2, {0x2800, 0x2BFE, 0xFE22, 0x2820, 0x3BFE, 0x1020, 0x7DFC, 0x5524, 0x55FC, 0x7D24, 0x11FC, 0xFC20, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {19062, 2, {0x2820, 0x2924, 0xFD24, 0x29FC, 0x3820, 0x11FE, 0x7C00, 0x5420, 0x55FC, 0x7D24, 0x1124, 0xFDFC, 0x1020, 0x1024, 0x13FE, 0x1102}},
    {19063, 2, {0x2820, 0x2810, 0xFDFE, 0x2884, 0x3848, 0x11FE, 0x7D02, 0x5624, 0x5410, 0x7DFE, 0x1040, 0xFC7C, 0x1044, 0x1084, 0x1094, 0x1108}},
    {19064, 2, {0x2888, 0x2888, 0xFDFE, 0x2888, 0x3800, 0x11FE, 0x7C88, 0x5488, 0x54F8, 0x7C88, 0x10F8, 0xFC88, 0x109E, 0x11E8, 0x1008, 0x1008}},
    {19065, 2, {0x2800, 0x28FE, 0xFE00, 0x287C, 0x3844, 0x107C, 0x7C00, 0x54FE, 0x5482, 0x7CAA, 0x1092, 0xFEFE, 0x1092, 0x1092, 0x1092, 0x1086}},
    {19066, 2, {0x5090, 0x5088, 0xF904, 0x5242, 0x7488, 0x21FC, 0xF804, 0xA800, 0xABDE, 0xF842, 0x2252, 0xF94A, 0x2252, 0x2042, 0x214A, 0x2084}},
    {19067, 2, {0x2800, 0x29FE, 0xFD20, 0x29FC, 0x3920, 0x11FC, 0x7D20, 0x5520, 0x55FE, 0x7C02, 0x12AA, 0xFEAA, 0x12AA, 0x1202, 0x1014, 0x1008}},
    {19068, 2, {0x2820, 0x2A20, 0xFD7C, 0x28A4, 0x3818, 0x10E6, 0x7F10, 0x557E, 0x5510, 0x7D7C, 0x1110, 0xFDFE, 0x1110, 0x1110, 0x12FE, 0x1400}},
    {19069, 2, {0x5000, 0x53FE, 0xFA00, 0x527C, 0x7244, 0x2244, 0xFA7C, 0xAA00, 0xAAEE, 0xFAAA, 0x22AA, 0xFAAA, 0x22EE, 0x2200, 0x23FE, 0x2000}},
    {19070, 2, {0x5040, 0x5020, 0xFBFE, 0x5200, 0x7248, 0x2248, 0xFBFE, 0xAA48, 0xAA48, 0xFA48, 0x2278, 0xFA00, 0x2554, 0x252A, 0x2A2A, 0x3000}},
    {19071, 2, {0x2820, 0x29FC, 0xFC20, 0x2888, 0x3BFE, 0x1088, 0x7DFC, 0x5504, 0x55FC, 0x7D04, 0x11FC, 0xFD04, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {19072, 2, {0x2820, 0x29FE, 0xFF02, 0x28F8, 0x3888, 0x10F8, 0x7C80, 0x54FC, 0x5484, 0x7CFC, 0x1000, 0xFE20, 0x1054, 0x1142, 0x114A, 0x1238}},
    {19073, 2, {0x2850, 0x2952, 0xFCD4, 0x2850, 0x39FE, 0x1088, 0x7C50, 0x55FE, 0x5420, 0x7CFC, 0x1020, 0xFDFE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {19074, 2, {0x5208, 0x5208, 0xFFBE, 0x5208, 0x7208, 0x2FBE, 0xFAAA, 0xAAAA, 0xAAAA, 0xFBAE, 0x2208, 0xFF1C, 0x2AAA, 0x3248, 0x2208, 0x2208}},
    {19075, 2, {0x5108, 0x5088, 0xFBC8, 0x5010, 0x73DE, 0x2264, 0xFBD4, 0xA814, 0xABD4, 0xF854, 0x2094, 0xF8E8, 0x2388, 0x2094, 0x2294, 0x2122}},
    {19076, 2, {0x2888, 0x2888, 0xFDFC, 0x2888, 0x3888, 0x13FE, 0x7C20, 0x55FC, 0x5524, 0x7DFC, 0x1124, 0xFDFC, 0x1000, 0x1088, 0x1104, 0x1202}},
    {19077, 2, {0x2800, 0x29FC, 0xFD54, 0x2954, 0x39FC, 0x1080, 0x7DFE, 0x5642, 0x55F2, 0x7D52, 0x1152, 0xFDF2, 0x104A, 0x13FA, 0x100A, 0x1004}},
    {19078, 2, {0x2850, 0x29FC, 0xFD54, 0x29FC, 0x3954, 0x11FC, 0x7C00, 0x55FC, 0x5400, 0x7DFC, 0x1104, 0xFDFC, 0x1088, 0x1050, 0x13FE, 0x1000}},
    {19079, 2, {0x2888, 0x2BFE, 0xFC88, 0x2824, 0x3BFE, 0x1020, 0x7DFC, 0x5524, 0x55FC, 0x7D24, 0x11FC, 0xFD24, 0x1008, 0x11FE, 0x1088, 0x1058}},
    {19080, 2, {0x2810, 0x287C, 0xFE54, 0x297C, 0x3910, 0x10FE, 0x7C00, 0x577C, 0x5544, 0x7D7C, 0x1140, 0xFD7C, 0x1144, 0x117C, 0x1280, 0x147E}},
    {19081, 2, {0x5124, 0x5248, 0xF924, 0x5000, 0x73FC, 0x2294, 0xFA64, 0xAA94, 0xABFC, 0xFA48, 0x2368, 0xFA48, 0x236A, 0x224A, 0x2246, 0x2362}},
    {19082, 2, {0x5110, 0x5090, 0xFBDE, 0x5010, 0x725E, 0x2182, 0xFBDE, 0xA810, 0xABDE, 0xFA50, 0x23DE, 0xFA50, 0x23DE, 0x2250, 0x2252, 0x22CE}},
    {19083, 2, {0x2810, 0x287C, 0xFE54, 0x297C, 0x3910, 0x10FE, 0x7C44, 0x577C, 0x5544, 0x7D7C, 0x1144, 0xFD7C, 0x1128, 0x1144, 0x1280, 0x147E}},
    {19084, 2, {0x51F0, 0x5210, 0xFBFC, 0x5604, 0x73FC, 0x2224, 0xFBB8, 0xAA22, 0xA9FE, 0xFA10, 0x27FC, 0xFA44, 0x23FC, 0x20D0, 0x214A, 0x263E}},
    {19085, 2, {0x53DE, 0x5252, 0xFBDE, 0x5252, 0x73DE, 0x2222, 0xFAFA, 0xAA22, 0xAAFA, 0xFAAA, 0x22FA, 0xFAAA, 0x22FA, 0x2272, 0x22AA, 0x2226}},
    {19086, 2, {0x5222, 0x53FE, 0xF890, 0x51FE, 0x7310, 0x25FE, 0xF910, 0xA9FE, 0xA910, 0xF9FE, 0x2100, 0xFBFE, 0x228A, 0x2376, 0x2252, 0x2276}},
    {19087, 2, {0x1010, 0x1010, 0x7E10, 0x22FE, 0xFF92, 0x92, 0x7E92, 0x42AA, 0x7EAA, 0x8CA, 0x7E82, 0x882, 0x4882, 0x7E82, 0x88A, 0x884}},
    {19088, 2, {0x1040, 0x1020, 0x7C28, 0x2408, 0xFE08, 0x48, 0x7C54, 0x4552, 0x7D52, 0x1160, 0x7E60, 0x1044, 0x50C4, 0x7D44, 0x123C, 0x1000}},
    {19089, 2, {0x1020, 0x1010, 0x7E10, 0x22FE, 0xFF82, 0x84, 0x7E40, 0x4244, 0x7E48, 0x850, 0x7E60, 0x842, 0x4842, 0x7E42, 0x83E, 0x800}},
    {19090, 2, {0x1020, 0x1010, 0x7E00, 0x22FE, 0xFF10, 0x10, 0x7E10, 0x4210, 0x7E7C, 0x810, 0x7E10, 0x810, 0x4810, 0x7E10, 0x8FE, 0x800}},
    {19091, 2, {0x1020, 0x1020, 0x7E20, 0x2220, 0xFF3C, 0x20, 0x7E20, 0x4220, 0x7EFC, 0x884, 0x7E84, 0x884, 0x4884, 0x7E84, 0x8FC, 0x884}},
    {19092, 2, {0x1028, 0x1024, 0x7DFE, 0x2420, 0xFE20, 0x1FC, 0x7D24, 0x4524, 0x7DFC, 0x1124, 0x7D24, 0x11FC, 0x5124, 0x7D24, 0x1124, 0x110C}},
    {19093, 2, {0x1080, 0x1080, 0x7CFE, 0x2502, 0xFE22, 0xAA, 0x7C72, 0x4422, 0x7DFE, 0x1022, 0x7C72, 0x10AA, 0x5122, 0x7C22, 0x100A, 0x1004}},
    {19094, 2, {0x1040, 0x1040, 0x7E7E, 0x2280, 0xFF7C, 0x44, 0x7E7C, 0x4244, 0x7E7C, 0x820, 0x7E3C, 0x844, 0x48A8, 0x7E10, 0x828, 0x8C6}},
    {19095, 2, {0x1000, 0x11DC, 0x7D44, 0x2544, 0xFF44, 0x1DC, 0x7D00, 0x4500, 0x7DDC, 0x1114, 0x7D14, 0x11D4, 0x5108, 0x7D08, 0x1114, 0x1122}},
    {19096, 2, {0x1000, 0x10FC, 0x7C84, 0x2484, 0xFEFC, 0x84, 0x7C84, 0x44FC, 0x7C00, 0x11FE, 0x7C10, 0x1090, 0x509C, 0x7C90, 0x1150, 0x123E}},
    {19097, 2, {0x1014, 0x1012, 0x7DFE, 0x2410, 0xFEFE, 0x92, 0x7CFE, 0x4492, 0x7CFE, 0x1092, 0x7C04, 0x11FE, 0x5044, 0x7C24, 0x1024, 0x100C}},
    {19098, 2, {0x1000, 0x11FC, 0x7D04, 0x25FC, 0xFF04, 0x1FC, 0x7C00, 0x47DE, 0x7C42, 0x1252, 0x7D4A, 0x10C6, 0x535A, 0x7C42, 0x114A, 0x1084}},
    {19099, 2, {0x101E, 0x10F0, 0x7C92, 0x2454, 0xFEFE, 0x54, 0x7C92, 0x4500, 0x7CFE, 0x1092, 0x7C92, 0x10FE, 0x5092, 0x7C92, 0x10FE, 0x1082}},
    {19100, 2, {0x1040, 0x1078, 0x7C88, 0x25FE, 0xFEA4, 0xD2, 0x7CFE, 0x4480, 0x7CBC, 0x1080, 0x7CBC, 0x1080, 0x50BC, 0x7D24, 0x113C, 0x1224}},
    {19101, 2, {0x1044, 0x11FE, 0x7C54, 0x2448, 0xFEFE, 0x190, 0x7CFE, 0x4490, 0x7CFE, 0x1090, 0x7CFE, 0x1080, 0x51FE, 0x7C44, 0x1038, 0x11C6}},
    {19102, 2, {0x0, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE44E, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0xFFFE}},
    {19103, 2, {0x1000, 0x11F8, 0xFE88, 0x1050, 0x1020, 0x7C50, 0x558C, 0x5450, 0x55DC, 0x5450, 0x55DC, 0x5C50, 0x11DC, 0x1050, 0x13FE, 0x1000}},
    {19104, 2, {0x1640, 0xF87C, 0x9084, 0xFE28, 0x3220, 0x5250, 0x968C, 0x1000, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0xFFFE}},
    {19105, 2, {0x1120, 0x1CBE, 0x7042, 0x5094, 0x7D90, 0x34A8, 0x5C46, 0x9000, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0xFFFE}},
    {19106, 2, {0x1220, 0x1D3E, 0x7042, 0x5094, 0x5310, 0x5128, 0x7D46, 0x1450, 0x35DC, 0x3450, 0x55DC, 0x9850, 0x11DC, 0x1050, 0x13FE, 0x1000}},
    {19107, 2, {0x1220, 0x113E, 0x1042, 0xFC94, 0x1310, 0x1128, 0x7C46, 0x5450, 0x55DC, 0x5450, 0x5DDC, 0x3050, 0x51DC, 0x9050, 0x13FE, 0x1000}},
    {19108, 2, {0x281E, 0x29F0, 0x2892, 0xEE54, 0x29FE, 0x2854, 0xEE92, 0x2900, 0x28FE, 0x2892, 0xEE92, 0x28FE, 0x2892, 0x2E92, 0xF0FE, 0x4082}},
    {19109, 2, {0x1000, 0x1EFC, 0x1044, 0xFF28, 0x8110, 0x7E28, 0xC6, 0x7E28, 0x42EE, 0x7E28, 0x42EE, 0x7E28, 0x42EE, 0x7E28, 0x24FE, 0x4200}},
    {19110, 2, {0x2000, 0x1000, 0x7CF8, 0x20, 0x4420, 0x2820, 0xFE20, 0x20, 0x7C20, 0x4420, 0x4420, 0x7C20, 0x4420, 0x4420, 0x7DFC, 0x4400}},
    {19111, 2, {0x2020, 0x1020, 0x7C20, 0xA0, 0x44AC, 0x28B4, 0xFEE4, 0x1A4, 0x7CA4, 0x44B4, 0x44A8, 0x7CA2, 0x44A2, 0x4482, 0x7C7E, 0x4400}},
    {19112, 2, {0x2040, 0x1040, 0x7C40, 0xFC, 0x4484, 0x2904, 0xFE04, 0x84, 0x7C44, 0x4444, 0x4444, 0x7C04, 0x4404, 0x4404, 0x7C28, 0x4410}},
    {19113, 2, {0x2020, 0x1020, 0x7C20, 0x50, 0x4450, 0x2888, 0xFF44, 0x222, 0x7C20, 0x44F8, 0x4408, 0x7C08, 0x4410, 0x4410, 0x7C20, 0x4420}},
    {19114, 2, {0x1000, 0x7E7C, 0x1244, 0x2244, 0x4A7C, 0x8444, 0x100, 0x3FF8, 0x820, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {19115, 2, {0x2000, 0x10FE, 0x7C00, 0x52, 0x44A4, 0x2928, 0xFE94, 0x4A, 0x7C00, 0x44FE, 0x4410, 0x7C10, 0x4410, 0x4410, 0x7DFE, 0x4400}},
    {19116, 2, {0x2020, 0x1020, 0x7DFE, 0x20, 0x4420, 0x29FE, 0xFF02, 0x204, 0x7CF8, 0x4410, 0x4420, 0x7DFE, 0x4420, 0x4420, 0x7CA0, 0x4440}},
    {19117, 2, {0x2020, 0x1010, 0x7CFE, 0x0, 0x4444, 0x2828, 0xFEFE, 0x0, 0x7C7C, 0x4444, 0x4444, 0x7C7C, 0x4444, 0x4444, 0x7C7C, 0x4444}},
    {19118, 2, {0x2040, 0x1040, 0x7CFE, 0x122, 0x44AA, 0x28FA, 0xFE42, 0x94, 0x7D48, 0x44FE, 0x4522, 0x7EAA, 0x44FA, 0x4442, 0x7C94, 0x4508}},
    {19119, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4102, 0x9FF4, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {19120, 2, {0x2040, 0x1020, 0x79FC, 0x88, 0x4850, 0x33FE, 0xFC00, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x4A94, 0x7A8A, 0x4C7A}},
    {19121, 2, {0x0, 0x1FE, 0xFC20, 0x440, 0x5FC, 0x504, 0x504, 0x7DFC, 0x4104, 0x41FC, 0x4104, 0x4504, 0x59FC, 0x6088, 0x4104, 0x202}},
    {19122, 2, {0x0, 0xFE, 0xFE10, 0x2820, 0x28FC, 0x2884, 0x2884, 0x28FC, 0x2884, 0x28FC, 0x2884, 0x2A84, 0x4CFC, 0x4848, 0x8084, 0x102}},
    {19123, 2, {0x0, 0xFE, 0x7E10, 0x1220, 0x12FC, 0xFF84, 0x1284, 0x12FC, 0x7E84, 0x12FC, 0x1084, 0x1084, 0x20FC, 0x2048, 0x4084, 0x8102}},
    {19124, 2, {0x800, 0x8FE, 0x810, 0x7F20, 0x497C, 0x844, 0x844, 0x147C, 0x1444, 0x147C, 0x1444, 0x2544, 0x267C, 0x4400, 0x8028, 0x44}},
    {19125, 2, {0x0, 0x7EFE, 0x410, 0x2420, 0x24FC, 0x2484, 0x4484, 0x7FFC, 0xC84, 0x14FC, 0x1484, 0x2484, 0x44FC, 0x8448, 0x1484, 0x902}},
    {19126, 2, {0x0, 0x7CFE, 0x4410, 0x4420, 0x44FC, 0x7C84, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x4484, 0x4EFC, 0xF048, 0x84, 0x102}},
    {19127, 2, {0x2000, 0x20FE, 0x4010, 0x7E20, 0x82FC, 0x284, 0x7A84, 0x4AFC, 0x4A84, 0x4AFC, 0x7A84, 0x4A84, 0x2FC, 0x248, 0x1484, 0x902}},
    {19128, 2, {0x0, 0x7EFE, 0x4210, 0x4220, 0x7EFC, 0x4884, 0x4884, 0x48FC, 0x7E84, 0x48FC, 0x4884, 0x4884, 0x4AFC, 0x4A48, 0x6684, 0x4302}},
    {19129, 2, {0x0, 0x7EFE, 0x810, 0x820, 0x10FC, 0x1484, 0x3284, 0x51FC, 0x9084, 0x10FC, 0x1084, 0x84, 0xEFC, 0xF048, 0x4084, 0x102}},
    {19130, 2, {0x1000, 0x10FE, 0x1010, 0x7E20, 0x10FC, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x10FC, 0x2084, 0x2884, 0x44FC, 0xFE48, 0x4284, 0x102}},
    {19131, 2, {0x1000, 0x10FE, 0x2010, 0x4420, 0xFEFC, 0x284, 0x2484, 0x24FC, 0x2484, 0xFEFC, 0x2484, 0x2484, 0x24FC, 0x4448, 0x4484, 0x8502}},
    {19132, 2, {0x1000, 0x10FE, 0x9210, 0x9220, 0x92FC, 0x9284, 0xFE84, 0x10FC, 0x1084, 0x92FC, 0x9284, 0x9284, 0x9EFC, 0xF248, 0x84, 0x102}},
    {19133, 2, {0x0, 0x7CFE, 0x4410, 0x4420, 0x44FC, 0x7C84, 0x84, 0xFEFC, 0x2084, 0x40FC, 0x7C84, 0x484, 0x4FC, 0x448, 0x2884, 0x1102}},
    {19134, 2, {0x10, 0xFF10, 0x828, 0x1028, 0x7E44, 0x4282, 0x4208, 0x7E10, 0x4264, 0x7E08, 0x4212, 0x4264, 0x7E08, 0x2430, 0x42C0, 0x8100}},
    {19135, 2, {0x800, 0x8FE, 0x1010, 0x7F20, 0x417C, 0x4144, 0x6544, 0x557C, 0x4944, 0x497C, 0x5544, 0x6544, 0x417C, 0x7F00, 0x4128, 0x44}},
    {19136, 2, {0x0, 0x7EFE, 0x4210, 0x4220, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x4084, 0x48FC, 0x4484, 0x4A84, 0x52FC, 0x6048, 0x4084, 0x102}},
    {19137, 2, {0x1000, 0x11FE, 0x2020, 0x7C40, 0x45FC, 0x4504, 0x7D04, 0x45FC, 0x4504, 0x7DFC, 0x4504, 0x4504, 0x45FC, 0x7C88, 0x4504, 0x202}},
    {19138, 2, {0x1000, 0x10FE, 0x3E10, 0x4220, 0xA47C, 0x1844, 0x1044, 0x247C, 0xC844, 0x1F7C, 0x2144, 0xD244, 0xC7C, 0x800, 0x3028, 0xC044}},
    {19139, 2, {0x0, 0x7F7E, 0x4010, 0x4020, 0x5E7C, 0x4044, 0x4044, 0x7F7C, 0x5044, 0x517C, 0x5A44, 0x5444, 0x527C, 0x9100, 0x9828, 0x1044}},
    {19140, 2, {0x4400, 0x24FE, 0x2810, 0x20, 0x7CFC, 0x4484, 0x4484, 0x44FC, 0x7C84, 0x28FC, 0x2884, 0x2884, 0x2AFC, 0x4C48, 0x8884, 0x102}},
    {19141, 2, {0x1000, 0x8FE, 0x7F10, 0x4120, 0x497C, 0x844, 0x844, 0x7F7C, 0x844, 0x187C, 0x1C44, 0x2A44, 0x497C, 0x8800, 0x828, 0x844}},
    {19142, 2, {0x400, 0x53E, 0x488, 0x410, 0x7FBE, 0x4422, 0x4422, 0x753E, 0x5522, 0x553E, 0x5522, 0x5222, 0x52BE, 0xA680, 0x8994, 0x10A2}},
    {19143, 2, {0xC00, 0x30FE, 0xC610, 0x1820, 0x637C, 0xC44, 0xF044, 0x7C, 0x7E44, 0x427C, 0xFF44, 0x4244, 0xFF7C, 0x4200, 0x4A28, 0x4444}},
    {19144, 2, {0x800, 0x497E, 0x4910, 0x4920, 0x7F7C, 0x844, 0x1444, 0x227C, 0x5144, 0x88FC, 0x844, 0x7E44, 0x27C, 0x400, 0x428, 0x844}},
    {19145, 2, {0xC00, 0x30FE, 0xC210, 0x1420, 0x48FC, 0x3084, 0xC884, 0x14FC, 0x2484, 0xCCFC, 0x1484, 0x2484, 0xC4FC, 0x448, 0x2884, 0x1102}},
    {19146, 2, {0x0, 0xFEFE, 0x810, 0x1020, 0x347C, 0x5244, 0x9144, 0x107C, 0x44, 0x7E7C, 0x4244, 0x4244, 0x427C, 0x4200, 0x7E28, 0x4244}},
    {19147, 2, {0xC00, 0x70FE, 0x1010, 0xFE20, 0x38FC, 0x5484, 0x9284, 0xFC, 0x1084, 0xFCFC, 0x2484, 0x4484, 0x28FC, 0x1048, 0x2884, 0xC502}},
    {19148, 2, {0x800, 0x10FE, 0x7F10, 0x4920, 0x497C, 0x7F44, 0x4944, 0x497C, 0x7F44, 0x147C, 0x2444, 0xFF44, 0x47C, 0x400, 0x428, 0x444}},
    {19149, 2, {0x1400, 0x14FE, 0x1410, 0xF720, 0x147C, 0x1444, 0x1444, 0x777C, 0x1444, 0x147C, 0x1444, 0xF744, 0x147C, 0x1400, 0x1428, 0x1444}},
    {19150, 2, {0x0, 0xF77E, 0x1110, 0x5520, 0x227C, 0x5544, 0x8944, 0x107C, 0x44, 0xF77C, 0x1144, 0x5544, 0x227C, 0x5500, 0x8928, 0x1044}},
    {19151, 2, {0x2200, 0x227E, 0xFF90, 0x2220, 0x227C, 0x3E44, 0x2244, 0x227C, 0x3E44, 0x227C, 0x2244, 0xFF44, 0x7C, 0x1400, 0x2228, 0x4144}},
    {19152, 2, {0x2200, 0x227E, 0x2210, 0x2220, 0xF7FC, 0x2244, 0x2244, 0x667C, 0x7744, 0xAAFC, 0xAA44, 0x3244, 0x227C, 0x2200, 0x2228, 0x2244}},
    {19153, 2, {0x1000, 0x10FE, 0xFE10, 0x1020, 0x28FC, 0x4484, 0xFE84, 0x4FC, 0x7484, 0x54FC, 0x5484, 0x7484, 0x54FC, 0x448, 0x1484, 0x902}},
    {19154, 2, {0x0, 0x6FE, 0x7810, 0x4820, 0x7F7C, 0x4844, 0x4544, 0x537C, 0x6144, 0x7C, 0x7E44, 0x4244, 0x7E7C, 0x4200, 0x7E28, 0x4244}},
    {19155, 2, {0x2000, 0x10FE, 0x7C10, 0x20, 0x44FC, 0x2484, 0x2884, 0xFEFC, 0x84, 0xFC, 0x7C84, 0x4484, 0x44FC, 0x4448, 0x7C84, 0x4502}},
    {19156, 2, {0x2000, 0x2F7E, 0xF510, 0x2520, 0x757C, 0x2544, 0xF544, 0x297C, 0x2344, 0x87C, 0x844, 0xFFC4, 0x147C, 0x2200, 0x4128, 0x80C4}},
    {19157, 2, {0x2000, 0x10FE, 0xFE10, 0x20, 0x44FC, 0x2884, 0xFE84, 0x82FC, 0x1084, 0x7CFC, 0x5484, 0x5484, 0x54FC, 0x5C48, 0x1084, 0x1102}},
    {19158, 2, {0x2200, 0x22FE, 0xFF10, 0x2220, 0x3E7C, 0x2244, 0x3E44, 0x227C, 0x2244, 0xFF7C, 0x4044, 0x5444, 0x627C, 0x4000, 0x7E28, 0x44}},
    {19159, 2, {0x0, 0x777E, 0x5110, 0x5120, 0x517C, 0x7744, 0x4044, 0x407C, 0x7744, 0x417C, 0x4944, 0x7544, 0x427C, 0x4200, 0x4528, 0x4944}},
    {19160, 2, {0x0, 0x7CFE, 0x4410, 0x7C20, 0x44FC, 0x7C84, 0x2084, 0x7EFC, 0x9284, 0x52FC, 0x6A84, 0x4284, 0x7AFC, 0x248, 0x1484, 0x902}},
    {19161, 2, {0x800, 0x8FE, 0x7F10, 0x820, 0x2A7C, 0x2244, 0xFFC4, 0x227C, 0x2A44, 0x87C, 0x7F44, 0x1C44, 0x2A7C, 0x4900, 0x8928, 0x844}},
    {19162, 2, {0x0, 0x3EFE, 0x2210, 0x2220, 0x3A7C, 0x2A44, 0x2A44, 0x7F7C, 0x4144, 0x5D7C, 0x5544, 0x5544, 0x5D7C, 0x4100, 0x4528, 0x4244}},
    {19163, 2, {0x2000, 0x2F3E, 0x2108, 0x4110, 0x5FBE, 0xC822, 0x4822, 0x4FBE, 0x5222, 0x423E, 0x5FA2, 0x4222, 0x453E, 0x4480, 0x4894, 0x5022}},
    {19164, 2, {0x800, 0x2A3E, 0x2A08, 0xAA90, 0xCCBE, 0xDDA2, 0xFFA2, 0xAABE, 0xAAA2, 0xDDBE, 0xFFA2, 0x99A2, 0x88BE, 0x8F80, 0xF094, 0x22}},
    {19165, 2, {0x600, 0x78FE, 0x810, 0x820, 0xFF7C, 0x1C44, 0x2A44, 0x497C, 0xBE44, 0x227C, 0x2244, 0x3E44, 0x227C, 0x2200, 0x3E28, 0x2244}},
    {19166, 2, {0x4400, 0x24FE, 0x2810, 0x20, 0xFEFC, 0x1084, 0x1084, 0x7CFC, 0x1084, 0x10FC, 0xFE84, 0x84, 0xAAFC, 0xAA48, 0xAA84, 0x102}},
    {19167, 2, {0x0, 0x7CFE, 0x4410, 0x4420, 0x7CFC, 0x84, 0xFE84, 0x82FC, 0xFE84, 0x82FC, 0xFE84, 0x8284, 0xFEFC, 0x2848, 0x4484, 0x8502}},
    {19168, 2, {0x1800, 0x60FE, 0xC10, 0x7020, 0xCFC, 0x7084, 0x84, 0xFEFC, 0x1084, 0x34FC, 0x5284, 0x9084, 0x7CFC, 0x4448, 0x7C84, 0x4502}},
    {19169, 2, {0x2200, 0x147E, 0xFF90, 0x1420, 0x147C, 0x7F44, 0x1544, 0xFFFC, 0x1544, 0x7F7C, 0x1444, 0x3644, 0x557C, 0x9480, 0x1428, 0x1444}},
    {19170, 2, {0x300, 0x7C7E, 0x4710, 0x4420, 0x5F7C, 0x5544, 0x5644, 0x5C7C, 0x5544, 0x537C, 0x5044, 0x5744, 0x557C, 0x5500, 0xA9A8, 0x44}},
    {19171, 2, {0xE00, 0xF0FE, 0x2210, 0x9220, 0x44FC, 0x2084, 0x4484, 0xF8FC, 0x1084, 0x24FC, 0xFE84, 0x1084, 0xFEFC, 0x2848, 0x4484, 0x8302}},
    {19172, 2, {0x0, 0xFF7E, 0x8110, 0x20, 0x7E7C, 0x4244, 0x7E44, 0x427C, 0x7E44, 0x107C, 0x844, 0xFF44, 0x7C, 0x2400, 0x4228, 0x8144}},
    {19173, 2, {0x800, 0x10FE, 0x7F10, 0x4920, 0x497C, 0x7F44, 0x497C, 0x4944, 0x7F7C, 0x1044, 0x1A7C, 0x2CA8, 0x2FC4, 0x4882, 0x4802, 0x87FE}},
    {19174, 2, {0x800, 0x83E, 0x7F08, 0x810, 0x83E, 0xFFA2, 0x2022, 0x413E, 0xFFA2, 0x40BE, 0x22, 0x7F22, 0x553E, 0x5500, 0x5794, 0xFC22}},
    {19175, 2, {0x1000, 0x20FE, 0x7C10, 0x4420, 0x7CFC, 0x4484, 0x7C84, 0x10FC, 0xFE84, 0x28FC, 0x5484, 0x9284, 0x7CFC, 0x1048, 0x1084, 0x1102}},
    {19176, 2, {0x1200, 0x123E, 0x1208, 0x7A10, 0x13FE, 0x7CA2, 0x10A2, 0xFCBE, 0x22A2, 0x3ABE, 0x2922, 0x2922, 0x293E, 0x4A80, 0x5A94, 0x84A2}},
    {19177, 2, {0x1000, 0x223E, 0x7F08, 0x4210, 0x94BE, 0xF7A2, 0x822, 0x363E, 0xC922, 0x30BE, 0xC422, 0x1922, 0x623E, 0xC00, 0x3014, 0xC022}},
    {19178, 2, {0x1000, 0x10FE, 0xFE10, 0x1020, 0xBAFC, 0x5484, 0xBA84, 0x10FC, 0xBA84, 0x54FC, 0xBA84, 0x1084, 0x28FC, 0x2448, 0x4284, 0x8102}},
    {19179, 2, {0x800, 0x7F7E, 0x4910, 0xFFA0, 0x497C, 0x7F44, 0x844, 0x7F7C, 0x4944, 0x7F7C, 0x1044, 0xFF44, 0x217C, 0x7200, 0xC28, 0xF344}},
    {19180, 2, {0x0, 0xF7BE, 0x2108, 0xA510, 0xF7BE, 0x6322, 0xB5A2, 0x213E, 0x7FA2, 0x40BE, 0x40A2, 0x7FA2, 0x40BE, 0x4080, 0x7F94, 0x40A2}},
    {19181, 2, {0x1000, 0x7CFE, 0x5410, 0x7C20, 0x10FC, 0xFE84, 0x84, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7C84, 0x44FC, 0x7C48, 0x2884, 0x4502}},
    {19182, 2, {0x0, 0x7F7E, 0x110, 0x3F20, 0x17C, 0x7F44, 0x44, 0x777C, 0x2544, 0xF77C, 0x244, 0xFFC4, 0x427C, 0x2200, 0x2A28, 0x444}},
    {19183, 2, {0x4100, 0x211E, 0xFE4, 0xF108, 0x429E, 0x4452, 0x7FF2, 0x505E, 0x5752, 0x555E, 0x5552, 0x5752, 0x555E, 0x5040, 0xA14C, 0x92}},
    {19184, 2, {0x20, 0xF924, 0x20A8, 0x43FE, 0xF8A8, 0x8924, 0x8A22, 0xF904, 0x8904, 0xF9DE, 0x8A44, 0x8D54, 0xF89E, 0x104, 0x5204, 0x8C04}},
    {19185, 2, {0x0, 0x7FBE, 0x808, 0x7F90, 0x54BE, 0x54A2, 0x54A2, 0x41BE, 0x22, 0x7FBE, 0x822, 0x7FA2, 0x54BE, 0x5480, 0x5494, 0x41A2}},
    {19186, 2, {0x280, 0x25E, 0x7FC4, 0x4208, 0x5E1E, 0x4292, 0x5E92, 0x571E, 0x5D52, 0x42DE, 0x8852, 0x492, 0x545E, 0x5140, 0x8F0C, 0x12}},
    {19187, 2, {0x0, 0x7F7E, 0x5510, 0x5520, 0x7F7C, 0x2044, 0x7F44, 0x917C, 0x7D44, 0x557C, 0x5544, 0x7D44, 0x117C, 0x7D00, 0x528, 0x244}},
    {19188, 2, {0x2200, 0x223E, 0xF788, 0x2210, 0x773E, 0xAAA2, 0x2262, 0x3E, 0x7F22, 0x3E, 0xFFA2, 0x822, 0x493E, 0x4880, 0xA894, 0x1022}},
    {19189, 2, {0x1000, 0x7F3E, 0x4908, 0x6510, 0x533E, 0x4922, 0x7F22, 0x83E, 0xFFA2, 0xA2BE, 0x4922, 0xFFA2, 0x203E, 0x3F00, 0x4114, 0x8322}},
    {19190, 2, {0x2080, 0x2A9E, 0xF5E4, 0x2A88, 0x71DE, 0x6AB2, 0xA492, 0x2A9E, 0x2092, 0x41E, 0x412, 0x7FD2, 0xA1E, 0x1100, 0x208C, 0x4052}},
    {19191, 2, {0x0, 0xFFBE, 0x808, 0xEB90, 0xAABE, 0xEBA2, 0x822, 0xFFBE, 0x822, 0xEBBE, 0xAAA2, 0xEBA2, 0x83E, 0xF80, 0xF814, 0x4022}},
    {19192, 2, {0x0, 0x7FDE, 0x404A, 0x434A, 0x5C4A, 0x444A, 0x5F4A, 0x554A, 0x5552, 0x5F52, 0x446A, 0x4544, 0x5F40, 0x4922, 0x801E, 0x0}},
    {19193, 2, {0x0, 0x7FC0, 0x407E, 0x4348, 0x5C48, 0x4448, 0x5F48, 0x5548, 0x5548, 0x5F48, 0x4448, 0x457E, 0x5F40, 0x4922, 0x801E, 0x0}},
    {19194, 2, {0x1000, 0x11FC, 0x1104, 0xFD34, 0x11C4, 0x1144, 0x21F4, 0x2954, 0x2954, 0x29F4, 0x5144, 0x5554, 0x5DF4, 0x8516, 0x206, 0x402}},
    {19195, 2, {0x0, 0x7FDE, 0x4052, 0x4352, 0x5C52, 0x445E, 0x5F52, 0x5552, 0x5552, 0x5F52, 0x445E, 0x4540, 0x5F40, 0x4922, 0x801E, 0x0}},
    {19196, 2, {0x10, 0x7F10, 0x4110, 0x4D7C, 0x7114, 0x5114, 0x7D14, 0x55FE, 0x7D10, 0x5128, 0x5524, 0x7D44, 0x4480, 0x4042, 0x803E, 0x0}},
    {19197, 2, {0x0, 0x7FC6, 0x4058, 0x4348, 0x5C4E, 0x4458, 0x5F4E, 0x5578, 0x554A, 0x5F4A, 0x444A, 0x4546, 0x5F40, 0x4922, 0x801E, 0x0}},
    {19198, 2, {0x40, 0x7C20, 0x4428, 0x4C08, 0x7448, 0x5454, 0x5552, 0x7D62, 0x5660, 0x7C44, 0x54C4, 0x553C, 0x5200, 0x7D00, 0x4482, 0x807E}},
    {19199, 2, {0x1E, 0x7FCA, 0x404A, 0x434A, 0x5C52, 0x4466, 0x5F40, 0x555E, 0x5552, 0x5F52, 0x445E, 0x4552, 0x5F40, 0x4922, 0x801E, 0x0}},
    {19200, 2, {0x48, 0x7C48, 0x4488, 0x4CBE, 0x752A, 0x55CA, 0x544A, 0x7C8A, 0x54AA, 0x7DF2, 0x5412, 0x542A, 0x5244, 0x7D00, 0x4482, 0x807E}},
    {19201, 2, {0x10, 0x7F08, 0x417E, 0x4D42, 0x7120, 0x5124, 0x7D38, 0x5520, 0x7D22, 0x5122, 0x551E, 0x7D00, 0x4480, 0x4042, 0x803E, 0x0}},
    {19202, 2, {0x28, 0x7C24, 0x4420, 0x4DFC, 0x7520, 0x5524, 0x5524, 0x7D28, 0x5528, 0x7D10, 0x5594, 0x552C, 0x5244, 0x7D00, 0x4482, 0x807E}},
    {19203, 2, {0x10, 0x7F08, 0x4100, 0x4D7C, 0x7100, 0x5144, 0x7D24, 0x5528, 0x7D28, 0x5110, 0x55FE, 0x7D00, 0x4480, 0x4042, 0x803E, 0x0}},
    {19204, 2, {0x14, 0x7C12, 0x4410, 0x4DFE, 0x7510, 0x5514, 0x5514, 0x7DD4, 0x5508, 0x7D0A, 0x551A, 0x5526, 0x5242, 0x7D00, 0x4482, 0x807E}},
    {19205, 2, {0x40, 0x7C40, 0x45FC, 0x4C84, 0x7484, 0x5528, 0x5610, 0x7D08, 0x57DE, 0x7D4A, 0x554A, 0x5652, 0x525A, 0x7DE4, 0x4482, 0x807E}},
    {19206, 2, {0x44, 0x7D84, 0x4494, 0x4C94, 0x77F4, 0x5494, 0x5594, 0x7DD4, 0x56B4, 0x7C94, 0x5484, 0x5494, 0x5288, 0x7D00, 0x4482, 0x807E}},
    {19207, 2, {0x4400, 0x25FC, 0x2904, 0x134, 0x7DC4, 0x4544, 0x45F4, 0x4554, 0x7D54, 0x29F4, 0x2944, 0x2954, 0x2BF4, 0x4D16, 0x8A06, 0x402}},
    {19208, 2, {0x0, 0x7F44, 0x4128, 0x4D00, 0x717C, 0x5144, 0x7D44, 0x557C, 0x7D28, 0x5128, 0x554A, 0x7D86, 0x4480, 0x4042, 0x803E, 0x0}},
    {19209, 2, {0x20, 0x7C20, 0x45FC, 0x4C20, 0x7420, 0x57FE, 0x5450, 0x7C50, 0x5554, 0x7D52, 0x5652, 0x5490, 0x5290, 0x7D30, 0x4482, 0x807E}},
    {19210, 2, {0x10, 0x7F10, 0x41FE, 0x4D10, 0x7154, 0x5154, 0x7DBA, 0x5510, 0x7D28, 0x5124, 0x5542, 0x7D80, 0x4480, 0x4042, 0x803E, 0x0}},
    {19211, 2, {0x20, 0x7C20, 0x44F8, 0x4C20, 0x75FC, 0x5488, 0x5544, 0x7C78, 0x5488, 0x7D50, 0x5420, 0x5450, 0x5288, 0x7D08, 0x4482, 0x807E}},
    {19212, 2, {0x1000, 0x1DFC, 0x1104, 0x7D34, 0x55C4, 0x5944, 0x71F4, 0x5554, 0x4D54, 0x41F4, 0x5D44, 0x5554, 0x57F4, 0x9516, 0x2206, 0x402}},
    {19213, 2, {0x80, 0x7C80, 0x44FC, 0x4D54, 0x7654, 0x54A4, 0x5554, 0x7C88, 0x5420, 0x7C14, 0x5552, 0x554A, 0x5238, 0x7D00, 0x4482, 0x807E}},
    {19214, 2, {0x14, 0x7C12, 0x47FE, 0x4C10, 0x7410, 0x55D4, 0x5554, 0x7D54, 0x55CC, 0x7C0A, 0x54CA, 0x5716, 0x5222, 0x7D40, 0x4482, 0x807E}},
    {19215, 2, {0x20, 0x7F10, 0x417C, 0x4D44, 0x7128, 0x51FE, 0x7D00, 0x557C, 0x7D44, 0x5144, 0x557C, 0x7D44, 0x4480, 0x4042, 0x803E, 0x0}},
    {19216, 2, {0x18, 0x7CE0, 0x4420, 0x4DFC, 0x7470, 0x54A8, 0x5524, 0x7C40, 0x57FE, 0x7C88, 0x5590, 0x5470, 0x538C, 0x7D00, 0x4482, 0x807E}},
    {19217, 2, {0x7C, 0x7F54, 0x417C, 0x4D54, 0x717C, 0x5100, 0x7D7C, 0x5544, 0x7D7C, 0x5144, 0x557C, 0x7D44, 0x44CC, 0x4042, 0x803E, 0x0}},
    {19218, 2, {0x20, 0x7DFC, 0x4504, 0x4C50, 0x7488, 0x5524, 0x5450, 0x7C88, 0x5506, 0x7CF8, 0x5408, 0x5450, 0x5220, 0x7D50, 0x448A, 0x807E}},
    {19219, 2, {0x20, 0x7F10, 0x41FE, 0x4D44, 0x7128, 0x51FE, 0x7D00, 0x557C, 0x7D44, 0x517C, 0x5544, 0x7D7C, 0x4480, 0x4042, 0x803E, 0x0}},
    {19220, 2, {0x20, 0x7C50, 0x4488, 0x4D04, 0x74F8, 0x5400, 0x55C4, 0x7D54, 0x55D4, 0x7D54, 0x55D4, 0x5544, 0x5354, 0x7D48, 0x4482, 0x807E}},
    {19221, 2, {0x0, 0x7DFC, 0x4504, 0x4D34, 0x75C4, 0x5544, 0x7DF4, 0x5554, 0x5554, 0x7DF4, 0x5544, 0x5554, 0x7DF4, 0x4916, 0x8206, 0x402}},
    {19222, 2, {0x10, 0x7F20, 0x417C, 0x4D44, 0x717C, 0x5144, 0x7D7C, 0x5500, 0x7D7C, 0x5110, 0x557C, 0x7D10, 0x44FC, 0x4042, 0x803E, 0x0}},
    {19223, 2, {0x0, 0x7CF8, 0x4488, 0x7C98, 0x44E8, 0x7CA8, 0xA8, 0xFEF8, 0x40A8, 0x7EF8, 0xAAA8, 0x2AAA, 0x4AAA, 0x92FA, 0x2A86, 0x4502}},
    {19224, 2, {0x20, 0x7DFC, 0x4400, 0x4CF8, 0x7488, 0x54F8, 0x5400, 0x7DFC, 0x5504, 0x7D74, 0x5554, 0x5574, 0x530C, 0x7D00, 0x4482, 0x807E}},
    {19225, 2, {0xE, 0x7F70, 0x4112, 0x4D44, 0x7120, 0x517C, 0x7D90, 0x5510, 0x7DFE, 0x5110, 0x5554, 0x7D7C, 0x4480, 0x4042, 0x803E, 0x0}},
    {19226, 2, {0x0, 0x7DFC, 0x4520, 0x4DF8, 0x7520, 0x55F8, 0x5520, 0x7DFC, 0x5404, 0x7D54, 0x5554, 0x5404, 0x5228, 0x7D10, 0x4482, 0x807E}},
    {19227, 2, {0x1FC, 0x7D24, 0x45FC, 0x4D24, 0x75FC, 0x54A8, 0x54A8, 0x7FFE, 0x54A8, 0x7CA8, 0x55FC, 0x5420, 0x53FE, 0x7D20, 0x44A2, 0x807E}},
    {19228, 2, {0x40, 0x7C20, 0x45FC, 0x4C50, 0x7524, 0x5554, 0x55FC, 0x7C20, 0x55FC, 0x7D44, 0x5554, 0x5574, 0x530C, 0x7D00, 0x4482, 0x807E}},
    {19229, 2, {0x28, 0x7F7C, 0x4128, 0x4DFE, 0x7110, 0x517C, 0x7D54, 0x557C, 0x7D54, 0x517C, 0x5500, 0x7D44, 0x4482, 0x4042, 0x803E, 0x0}},
    {19230, 2, {0x20, 0x7DFE, 0x4420, 0x4CFC, 0x7400, 0x55FE, 0x5402, 0x7CF8, 0x5420, 0x7DFE, 0x5404, 0x55DE, 0x5354, 0x7DCC, 0x4482, 0x807E}},
    {19231, 2, {0x82, 0x7D7A, 0x452A, 0x4DAA, 0x754A, 0x559A, 0x5542, 0x7CA2, 0x55F2, 0x7C4A, 0x55F2, 0x5552, 0x52E6, 0x7DF8, 0x4482, 0x807E}},
    {19232, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x4, 0x7FE8, 0x824, 0x7212, 0x120E, 0xFFD0, 0x122A, 0x221E}},
    {19233, 2, {0x200, 0x3FD0, 0x260, 0xFFFE, 0x1C10, 0xEFF0, 0x810, 0xFF0, 0x4, 0x7FE8, 0x824, 0x7212, 0x120E, 0xFFD0, 0x122A, 0x221E}},
    {19234, 2, {0x800, 0x800, 0x14FC, 0x1224, 0x2924, 0x7E24, 0xA224, 0x3E24, 0x2224, 0x3E24, 0x2024, 0x2844, 0x2444, 0x2A84, 0x3328, 0x2210}},
    {19235, 2, {0x820, 0x820, 0x1420, 0x127E, 0x2940, 0x7E80, 0xA23C, 0x3E04, 0x2208, 0x3E10, 0x2020, 0x2840, 0x2442, 0x2A42, 0x323E, 0x2000}},
    {19236, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2360, 0x2C98, 0x37F6, 0x2410, 0x27F0, 0x2410, 0x27F4, 0x4468, 0x4518, 0x8604}},
    {19237, 2, {0x420, 0x210, 0x13E, 0xFFC0, 0x22, 0x11A, 0x286, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19238, 2, {0x800, 0x878, 0x1448, 0x1248, 0x2948, 0x7E86, 0xA300, 0x3EFC, 0x2244, 0x3E44, 0x2028, 0x2828, 0x2410, 0x2A28, 0x3244, 0x2182}},
    {19239, 2, {0x800, 0x800, 0x14FC, 0x1200, 0x2900, 0x7E00, 0xA3FE, 0x3E50, 0x2250, 0x3E50, 0x2050, 0x2850, 0x2452, 0x2A92, 0x328E, 0x2100}},
    {19240, 2, {0x1080, 0x1080, 0x2880, 0x25FE, 0x5202, 0xFC82, 0x4482, 0x7DE2, 0x44A2, 0x7CA2, 0x412A, 0x512A, 0x4A1A, 0x5402, 0x6414, 0x4008}},
    {19241, 2, {0x0, 0x7FFC, 0x100, 0x360, 0xD18, 0x7104, 0x280, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19242, 2, {0x800, 0x800, 0x14FE, 0x1210, 0x2910, 0x7E20, 0xA220, 0x3E68, 0x2264, 0x3EA2, 0x2122, 0x2820, 0x2420, 0x2A20, 0x3220, 0x2020}},
    {19243, 2, {0x804, 0x81E, 0x14F0, 0x1290, 0x2990, 0x7E90, 0xA290, 0x3EFE, 0x2290, 0x3E90, 0x2090, 0x2888, 0x248A, 0x2AAA, 0x32D6, 0x208A}},
    {19244, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x5324, 0xFD24, 0x4524, 0x7D24, 0x4524, 0x7FFE, 0x4020, 0x5050, 0x4850, 0x5488, 0x6504, 0x4202}},
    {19245, 2, {0x0, 0x7DFC, 0x4444, 0x4444, 0x7C94, 0x108, 0x280, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19246, 2, {0x1008, 0x1048, 0x2928, 0x2528, 0x5328, 0xFD08, 0x4508, 0x7D08, 0x4508, 0x7D08, 0x4148, 0x5194, 0x4914, 0x5422, 0x6442, 0x4082}},
    {19247, 2, {0x810, 0x810, 0x1410, 0x1210, 0x291E, 0x7E10, 0xA210, 0x3E10, 0x22FE, 0x3E82, 0x2082, 0x2882, 0x2482, 0x2A82, 0x32FE, 0x2082}},
    {19248, 2, {0x800, 0x8FE, 0x1422, 0x1222, 0x2922, 0x7E22, 0xA24A, 0x3E44, 0x2280, 0x3E7E, 0x2042, 0x2842, 0x2442, 0x2A42, 0x327E, 0x2042}},
    {19249, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x53FE, 0xFC70, 0x44A8, 0x7CA8, 0x4524, 0x7D24, 0x4222, 0x50F8, 0x4820, 0x5420, 0x6420, 0x4020}},
    {19250, 2, {0x840, 0x840, 0x1480, 0x12FC, 0x2904, 0x7E04, 0xA2F4, 0x3E94, 0x2294, 0x3E94, 0x20F4, 0x2894, 0x2404, 0x2A04, 0x3228, 0x2010}},
    {19251, 2, {0x1020, 0x1020, 0x2924, 0x24A4, 0x52A8, 0xFC20, 0x45FC, 0x7C20, 0x4420, 0x7C20, 0x43FE, 0x5020, 0x4820, 0x5420, 0x6420, 0x4020}},
    {19252, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x5220, 0xFC20, 0x4420, 0x7DFC, 0x4470, 0x7CA8, 0x40A8, 0x5124, 0x4924, 0x5622, 0x6420, 0x4020}},
    {19253, 2, {0x1020, 0x1010, 0x2810, 0x25FE, 0x5220, 0xFC24, 0x4444, 0x7CF8, 0x4412, 0x7C22, 0x4044, 0x5188, 0x4810, 0x5428, 0x6444, 0x4182}},
    {19254, 2, {0x800, 0x8FC, 0x1484, 0x1284, 0x29FC, 0x7E84, 0xA284, 0x3EFC, 0x22A2, 0x3EA4, 0x2098, 0x2890, 0x2488, 0x2AA4, 0x32C2, 0x2080}},
    {19255, 2, {0x820, 0x820, 0x147C, 0x1284, 0x2948, 0x7E30, 0xA220, 0x3E48, 0x2390, 0x3E3E, 0x2042, 0x29A4, 0x2418, 0x2A10, 0x3260, 0x2180}},
    {19256, 2, {0x7F7C, 0x804, 0x1E28, 0x2210, 0x5428, 0x944, 0x3280, 0xCD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19257, 2, {0x1000, 0x11FE, 0x2820, 0x2420, 0x5248, 0xFC84, 0x45FE, 0x7C22, 0x4420, 0x7C20, 0x41FC, 0x5020, 0x4820, 0x5420, 0x67FE, 0x4000}},
    {19258, 2, {0x1104, 0x1084, 0x2888, 0x2400, 0x53FE, 0xFC20, 0x4420, 0x7DFC, 0x4420, 0x7C20, 0x43FE, 0x5020, 0x4820, 0x5420, 0x6420, 0x4020}},
    {19259, 2, {0x1028, 0x1024, 0x283E, 0x25E0, 0x5220, 0xFC3C, 0x45E0, 0x7C20, 0x443E, 0x7DE0, 0x4022, 0x5024, 0x4818, 0x546A, 0x6586, 0x4002}},
    {19260, 2, {0x800, 0x87C, 0x1444, 0x1244, 0x297C, 0x7E00, 0xA2FE, 0x3E82, 0x2282, 0x3EFE, 0x2082, 0x2882, 0x24FE, 0x2A82, 0x328A, 0x2084}},
    {19261, 2, {0x884, 0x844, 0x1448, 0x1200, 0x29FC, 0x7E84, 0xA284, 0x3E84, 0x22FC, 0x3E48, 0x2048, 0x2848, 0x2448, 0x2A8A, 0x328A, 0x2106}},
    {19262, 2, {0x1088, 0x1048, 0x2850, 0x25FC, 0x5224, 0xFC24, 0x45FC, 0x7D20, 0x4520, 0x7DFE, 0x4022, 0x5022, 0x486A, 0x54A4, 0x6520, 0x4020}},
    {19263, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x53FC, 0xFD00, 0x451C, 0x7D70, 0x451C, 0x7D70, 0x411E, 0x51F0, 0x4910, 0x5512, 0x6512, 0x420E}},
    {19264, 2, {0x1000, 0x1078, 0x2A28, 0x2528, 0x5128, 0xF828, 0x4878, 0x7B28, 0x4928, 0x792A, 0x412A, 0x5126, 0x4922, 0x5A80, 0x6C7E, 0x4000}},
    {19265, 2, {0x100C, 0xFEF0, 0x1080, 0x1CFE, 0xF088, 0x1108, 0x3288, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19266, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x52F8, 0xFC00, 0x45FC, 0x7D24, 0x4524, 0x7D24, 0x41FC, 0x5100, 0x4900, 0x5502, 0x6502, 0x40FE}},
    {19267, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x5144, 0xFA22, 0x49F8, 0x7808, 0x4850, 0x7820, 0x40A4, 0x5282, 0x4A8A, 0x5A8A, 0x6C78, 0x4000}},
    {19268, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x53FC, 0xFD24, 0x45FC, 0x7C20, 0x45FC, 0x7C24, 0x43FE, 0x5024, 0x49FC, 0x5420, 0x64A0, 0x4040}},
    {19269, 2, {0x1000, 0x11FC, 0x2904, 0x2524, 0x5324, 0xFDFC, 0x4524, 0x7D24, 0x4574, 0x7D54, 0x4154, 0x5174, 0x4904, 0x5504, 0x65FC, 0x4104}},
    {19270, 2, {0x7F20, 0x4830, 0x7E28, 0x4224, 0x7E20, 0x4920, 0x7E80, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19271, 2, {0x1080, 0x1080, 0x28FC, 0x2504, 0x5244, 0xFC44, 0x47FC, 0x7C44, 0x45F4, 0x7C04, 0x41F4, 0x5114, 0x49F4, 0x5404, 0x6428, 0x4010}},
    {19272, 2, {0x1010, 0x1010, 0x287C, 0x2714, 0x51FE, 0xF914, 0x4A7C, 0x7A10, 0x4F7C, 0x7910, 0x41FE, 0x5510, 0x4A10, 0x5300, 0x64FE, 0x4800}},
    {19273, 2, {0x1020, 0x1020, 0x28FE, 0x2440, 0x52FC, 0xFD10, 0x46FE, 0x7C00, 0x44FC, 0x7C84, 0x40FC, 0x5084, 0x48FC, 0x5484, 0x6494, 0x4088}},
    {19274, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x52A8, 0xFC20, 0x45FC, 0x7D24, 0x4524, 0x7D24, 0x43FE, 0x5020, 0x4850, 0x5488, 0x6504, 0x4202}},
    {19275, 2, {0x1100, 0x117C, 0x2944, 0x2644, 0x5244, 0xFE7C, 0x4A10, 0x7A10, 0x4AFE, 0x7A38, 0x4254, 0x5254, 0x4A94, 0x5B12, 0x6A10, 0x4210}},
    {19276, 2, {0x107C, 0xFF44, 0x107C, 0x7E44, 0x427C, 0x7F44, 0x28C, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19277, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x5220, 0xFCF8, 0x4420, 0x7C20, 0x45FC, 0x7C00, 0x41FC, 0x5154, 0x4954, 0x5554, 0x67FE, 0x4000}},
    {19278, 2, {0x1048, 0x1148, 0x2948, 0x27FE, 0x5148, 0xF948, 0x4978, 0x7900, 0x49FE, 0x7820, 0x43FE, 0x5070, 0x48A8, 0x5924, 0x6E22, 0x4020}},
    {19279, 2, {0x1040, 0x1020, 0x29FE, 0x2502, 0x5102, 0xF9FE, 0x4900, 0x7900, 0x49FE, 0x79AA, 0x41AA, 0x52FE, 0x4AAA, 0x5AAA, 0x6CA2, 0x4086}},
    {19280, 2, {0x1104, 0x1088, 0x2800, 0x27FE, 0x5020, 0xF820, 0x49FC, 0x7820, 0x4820, 0x7BFE, 0x4040, 0x5024, 0x4AA2, 0x5A8A, 0x6C88, 0x4078}},
    {19281, 2, {0x1088, 0x1050, 0x2800, 0x25FE, 0x5250, 0xFDFC, 0x4454, 0x7DFE, 0x4454, 0x7DFC, 0x4050, 0x50D8, 0x4954, 0x5652, 0x6450, 0x4050}},
    {19282, 2, {0x1010, 0x1020, 0x28FC, 0x2484, 0x52FC, 0xFC84, 0x44FC, 0x7C84, 0x44FC, 0x7C20, 0x4010, 0x5054, 0x4942, 0x554A, 0x6638, 0x4000}},
    {19283, 2, {0x1020, 0x1124, 0x2924, 0x2524, 0x53FC, 0xFC00, 0x47FE, 0x7C00, 0x45FC, 0x7D04, 0x4104, 0x51FC, 0x4888, 0x5450, 0x67FE, 0x4000}},
    {19284, 2, {0x1010, 0x1220, 0x297C, 0x2544, 0x5044, 0xF87C, 0x4B40, 0x7940, 0x497C, 0x7944, 0x4144, 0x517C, 0x4944, 0x5A80, 0x6C7E, 0x4000}},
    {19285, 2, {0x100C, 0xFEF0, 0x5480, 0x7CFE, 0x5488, 0xFF08, 0x1288, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {19286, 2, {0x1040, 0x1020, 0x29FE, 0x2500, 0x5148, 0xF948, 0x49FE, 0x7948, 0x4948, 0x7948, 0x4178, 0x5100, 0x4AD4, 0x52AA, 0x652A, 0x4800}},
    {19287, 2, {0x1040, 0x1020, 0x29FC, 0x2488, 0x5050, 0xFBFE, 0x4800, 0x79FC, 0x4904, 0x79FC, 0x4104, 0x51FC, 0x4890, 0x5892, 0x6912, 0x460E}},
    {19288, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x5104, 0xF9FC, 0x4820, 0x7BFE, 0x4800, 0x79FC, 0x4104, 0x51FC, 0x4820, 0x5924, 0x6A22, 0x4060}},
    {19289, 2, {0x1000, 0x13DE, 0x2888, 0x26A8, 0x53FE, 0xF988, 0x4ADA, 0x7CA6, 0x49F8, 0x7908, 0x4108, 0x51F8, 0x4908, 0x5908, 0x69F8, 0x4108}},
    {19290, 2, {0x1020, 0x11FC, 0x2888, 0x2450, 0x53FE, 0xFC00, 0x45FC, 0x7D24, 0x45FC, 0x7D24, 0x41FC, 0x5020, 0x49FC, 0x5420, 0x67FE, 0x4000}},
    {19291, 2, {0x1008, 0x1388, 0x2888, 0x2510, 0x57DE, 0xFA94, 0x4AA4, 0x7B94, 0x4A94, 0x7B94, 0x4294, 0x52C8, 0x4B88, 0x5E94, 0x68A4, 0x40C2}},
    {19292, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x5354, 0xFC88, 0x4504, 0x7EFA, 0x4488, 0x7CF8, 0x4088, 0x50F8, 0x4820, 0x54A8, 0x6524, 0x4060}},
    {19293, 2, {0x1020, 0x1050, 0x2888, 0x2574, 0x5202, 0xFDFC, 0x4524, 0x7DAC, 0x4524, 0x7DFC, 0x4000, 0x50F8, 0x4888, 0x54F8, 0x6488, 0x40F8}},
    {19294, 2, {0x1000, 0x11FC, 0x2954, 0x25FC, 0x5220, 0xFDFC, 0x4420, 0x7FFE, 0x4488, 0x7C50, 0x41FC, 0x5020, 0x4BFE, 0x5420, 0x6420, 0x4020}},
    {19295, 2, {0x1088, 0x13FE, 0x2888, 0x2400, 0x5148, 0xFBFE, 0x4948, 0x7978, 0x4900, 0x79FC, 0x4020, 0x53FE, 0x4870, 0x58A8, 0x6B26, 0x4020}},
    {19296, 2, {0x1020, 0x13FE, 0x2A88, 0x2650, 0x53FE, 0xFA50, 0x4AFC, 0x7A54, 0x4BFE, 0x7A54, 0x42FC, 0x5250, 0x4AD8, 0x5554, 0x6652, 0x4850}},
    {19297, 2, {0x103C, 0x13C0, 0x2944, 0x24A4, 0x51F8, 0xF820, 0x4BFC, 0x7800, 0x49F8, 0x7808, 0x41F8, 0x5008, 0x49F8, 0x5AC4, 0x6AAA, 0x447A}},
    {19298, 2, {0x1040, 0x1020, 0x2BFE, 0x2622, 0x5154, 0xF94A, 0x4A3A, 0x7800, 0x49FC, 0x7954, 0x43FE, 0x5000, 0x49FC, 0x5820, 0x68A0, 0x4040}},
    {19299, 2, {0x1040, 0x1020, 0x2BFE, 0x2488, 0x5052, 0xFFAC, 0x4AAA, 0x7AA8, 0x4DAC, 0x7800, 0x4104, 0x51FC, 0x4904, 0x59FC, 0x6904, 0x4204}},
    {19300, 2, {0x1010, 0x107C, 0x2A54, 0x257C, 0x5110, 0xF8FE, 0x4800, 0x7B7C, 0x4944, 0x797C, 0x4140, 0x517C, 0x4944, 0x597C, 0x6A80, 0x447E}},
    {19301, 2, {0x1108, 0x1090, 0x2BFC, 0x2440, 0x51F8, 0xF880, 0x4BFC, 0x7948, 0x49FC, 0x7B0A, 0x45F8, 0x5108, 0x49FA, 0x5934, 0x698C, 0x4102}},
    {19302, 2, {0x1088, 0x13FE, 0x2888, 0x2412, 0x52FE, 0xF810, 0x4CFE, 0x7A92, 0x4AFE, 0x7892, 0x41FE, 0x5292, 0x4E04, 0x5AFE, 0x6A44, 0x422C}},
    {19303, 2, {0x1020, 0x13FE, 0x2A48, 0x27FE, 0x5248, 0xFAEC, 0x4B5A, 0x7A48, 0x4A20, 0x7A44, 0x4278, 0x5214, 0x4AFE, 0x5A10, 0x6A54, 0x44B2}},
    {19304, 2, {0x1222, 0x13FE, 0x2890, 0x25FE, 0x5310, 0xFDFE, 0x4910, 0x79FE, 0x4910, 0x79FE, 0x4100, 0x53FE, 0x4A8A, 0x5B76, 0x6A52, 0x4276}},
    {19305, 2, {0x1020, 0x13FE, 0x2A48, 0x27FE, 0x5248, 0xFAEC, 0x4B5A, 0x7A48, 0x4A50, 0x7BDE, 0x4250, 0x53DC, 0x4A50, 0x55DE, 0x6450, 0x4850}},
    {19306, 2, {0x2000, 0x21FC, 0x2104, 0x3DFC, 0x4504, 0x4BFE, 0x8000, 0x21FC, 0x2124, 0x21FC, 0x2124, 0x25FC, 0x2820, 0x31FC, 0x2020, 0x3FE}},
    {19307, 2, {0x4084, 0x4448, 0x59FE, 0x6020, 0x4240, 0x42FC, 0x3E84, 0x84, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484}},
    {19308, 2, {0x4092, 0x4524, 0x5924, 0x6092, 0x4200, 0x42FE, 0x3E10, 0x20, 0x7CFC, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7CFC, 0x4484}},
    {19309, 2, {0x4400, 0x28FE, 0xFE10, 0x1020, 0x20FC, 0x7C84, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7C84, 0x44FC, 0x4448, 0x7C84, 0x4502}},
    {19310, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1010, 0x820, 0xFFFE, 0x200, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8}},
    {19311, 2, {0x420, 0x7820, 0x1124, 0xFF24, 0x1124, 0x3924, 0x55FC, 0x9020, 0x7C20, 0x4524, 0x4524, 0x7D24, 0x4524, 0x4524, 0x7DFC, 0x4404}},
    {19312, 2, {0x420, 0x7820, 0x11FC, 0xFE40, 0x10F8, 0x3840, 0x57FE, 0x9088, 0x7D24, 0x4622, 0x44F8, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x4420}},
    {19313, 2, {0x420, 0x7924, 0x10A4, 0xFEA8, 0x1020, 0x39FC, 0x5524, 0x9124, 0x7D74, 0x45AC, 0x4524, 0x7D24, 0x4524, 0x4504, 0x7D14, 0x4508}},
    {19314, 2, {0x420, 0x7820, 0x11FC, 0xFE50, 0x1088, 0x3904, 0x57FE, 0x9008, 0x7DE8, 0x4528, 0x4528, 0x7DE8, 0x4528, 0x4408, 0x7C28, 0x4410}},
    {19315, 2, {0x7C40, 0x11FC, 0xFD24, 0x39FC, 0x5524, 0xFDFC, 0x2468, 0x1BA2, 0x651E, 0x7FF8, 0x920, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE0}},
    {19316, 2, {0x0, 0x1FF8, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x4, 0x2924, 0x4494, 0x108, 0xFFFE, 0x100, 0x100, 0x100}},
    {19317, 2, {0x0, 0x7E78, 0x4848, 0x7E48, 0x4848, 0x7E48, 0x4868, 0x4858, 0x7E58, 0x248, 0xAA48, 0xAA4A, 0xAA4A, 0x828A, 0x1486, 0x900}},
    {19318, 2, {0x0, 0x7EFE, 0x4848, 0x7E48, 0x4848, 0x7E48, 0x4848, 0x4848, 0x7E48, 0x248, 0xAA48, 0xAA48, 0xAA88, 0x8288, 0x1508, 0xA08}},
    {19319, 2, {0x8, 0x7F1C, 0x4870, 0x7F10, 0x4810, 0x7F1C, 0x4870, 0x4810, 0x7F90, 0x9E, 0xAAF0, 0xAA90, 0xAA92, 0x8092, 0x512, 0x20E}},
    {19320, 2, {0x10, 0x7E88, 0x4884, 0x7E84, 0x4902, 0x7E02, 0x4888, 0x4888, 0x7E88, 0x250, 0xAA50, 0xAA20, 0xAA50, 0x8288, 0x1504, 0xA02}},
    {19321, 2, {0x0, 0x7C40, 0x519E, 0x7D12, 0x5112, 0x7D12, 0x5112, 0x5112, 0x7D12, 0x512, 0x555A, 0x5594, 0x5510, 0x8410, 0x1410, 0x810}},
    {19322, 2, {0x20, 0x7E10, 0x4810, 0x7EFE, 0x4800, 0x7E00, 0x4878, 0x4848, 0x7E48, 0x248, 0xAA48, 0xAA4A, 0xAA4A, 0x828A, 0x1486, 0x900}},
    {19323, 2, {0x10, 0x7E50, 0x4850, 0x7E50, 0x4888, 0x7E88, 0x4904, 0x4AFA, 0x7E48, 0x248, 0xAA48, 0xAA48, 0xAA88, 0x8288, 0x1528, 0xA10}},
    {19324, 2, {0x0, 0x7F3C, 0x4800, 0x7F00, 0x4800, 0x7F7E, 0x4824, 0x4824, 0x7FA4, 0xA4, 0xAAA4, 0xAAA4, 0xAAA4, 0x80A4, 0x544, 0x284}},
    {19325, 2, {0x8, 0x7F28, 0x4828, 0x7F28, 0x483E, 0x7F28, 0x4848, 0x4808, 0x7F88, 0xFE, 0xAA88, 0xAA88, 0xAA88, 0x8088, 0x508, 0x208}},
    {19326, 2, {0x20, 0x7E28, 0x4824, 0x7E24, 0x4820, 0x7E20, 0x48FC, 0x4820, 0x7E20, 0x250, 0xAA50, 0xAA50, 0xAA88, 0x8288, 0x1504, 0xA02}},
    {19327, 2, {0x10, 0x7E10, 0x4850, 0x7E50, 0x487C, 0x7E90, 0x4910, 0x4810, 0x7EFE, 0x210, 0xAA28, 0xAA28, 0xAA28, 0x8244, 0x1444, 0x882}},
    {19328, 2, {0x40, 0x7E40, 0x48FC, 0x7E84, 0x4904, 0x7EF4, 0x4894, 0x4894, 0x7EF4, 0x284, 0xAA94, 0xAA88, 0xAA82, 0x8282, 0x147E, 0x800}},
    {19329, 2, {0x20, 0x7E20, 0x4840, 0x7E48, 0x4884, 0x7EFC, 0x4804, 0x4848, 0x7E48, 0x2FE, 0xAA48, 0xAA48, 0xAA48, 0x8288, 0x1488, 0x908}},
    {19330, 2, {0x28, 0x7E24, 0x4824, 0x7E20, 0x48FE, 0x7E20, 0x4820, 0x483C, 0x7E34, 0x254, 0xAA54, 0xAA54, 0xAA88, 0x8288, 0x1514, 0x822}},
    {19331, 2, {0x10, 0x7E10, 0x4810, 0x7E7E, 0x4810, 0x7E10, 0x4810, 0x48FE, 0x7E10, 0x210, 0xAA20, 0xAA28, 0xAA44, 0x82FE, 0x1442, 0x800}},
    {19332, 2, {0x10, 0x7F08, 0x4808, 0x7F7E, 0x4810, 0x7F10, 0x4824, 0x4824, 0x7FB8, 0x88, 0xAA90, 0xAA90, 0xAAA4, 0x80BE, 0x502, 0x200}},
    {19333, 2, {0x14, 0x7E12, 0x4812, 0x7E10, 0x48FE, 0x7E90, 0x4892, 0x4892, 0x7E92, 0x294, 0xAA94, 0xAAB4, 0xAACA, 0x828A, 0x1416, 0x822}},
    {19334, 2, {0x0, 0x7EFE, 0x4808, 0x7E10, 0x4834, 0x7E52, 0x4890, 0x4810, 0x7E00, 0x210, 0xAA10, 0xAAFE, 0xAA10, 0x8210, 0x1410, 0x810}},
    {19335, 2, {0x20, 0x7E20, 0x48FC, 0x7E24, 0x49FE, 0x7E24, 0x48FC, 0x4820, 0x7E20, 0x2FC, 0xAA20, 0xAA20, 0xABFE, 0x8220, 0x1420, 0x820}},
    {19336, 2, {0x10, 0x7F10, 0x481E, 0x7F22, 0x4854, 0x7F08, 0x4810, 0x4864, 0x7F88, 0x9E, 0xAAA2, 0xAAC2, 0xAA94, 0x8088, 0x510, 0x260}},
    {19337, 2, {0x40, 0x7E40, 0x487E, 0x7E80, 0x4900, 0x7E1C, 0x48E0, 0x48A0, 0x7EA2, 0x2A4, 0xAAA8, 0xAA90, 0xAA90, 0x82A8, 0x14C4, 0x882}},
    {19338, 2, {0x20, 0x7C20, 0x51FE, 0x7C20, 0x51FC, 0x7C24, 0x51FC, 0x5120, 0x7DFE, 0x422, 0x542A, 0x5454, 0x5450, 0x8488, 0x1504, 0xA02}},
    {19339, 2, {0x2, 0x7C02, 0x53E2, 0x7C8A, 0x508A, 0x7CEA, 0x50AA, 0x512A, 0x7DAA, 0x56A, 0x564A, 0x544A, 0x5482, 0x8482, 0x150A, 0xA04}},
    {19340, 2, {0x90, 0xF890, 0xA090, 0xF910, 0xA1FE, 0xFB10, 0xA510, 0xA138, 0xF938, 0x954, 0xA954, 0xA992, 0xA910, 0x8910, 0x910, 0x3110}},
    {19341, 2, {0x3FF8, 0x2100, 0x3FF0, 0x2100, 0x3FF0, 0x2100, 0x3FFC, 0x4924, 0x8694, 0x108, 0xFFFE, 0x488, 0x850, 0x3A20, 0xCC18, 0x806}},
    {19342, 2, {0xC2, 0x7C34, 0x5018, 0x7C64, 0x5182, 0x7C20, 0x51FE, 0x5050, 0x7C90, 0x5FE, 0x5692, 0x5492, 0x549A, 0x8494, 0x1410, 0x810}},
    {19343, 2, {0x0, 0x7EFE, 0x4810, 0x7E10, 0x487C, 0x7E24, 0x4824, 0x48FE, 0x7E00, 0x200, 0xAA7C, 0xAA44, 0xAA44, 0x8244, 0x147C, 0x844}},
    {19344, 2, {0x0, 0x7EFE, 0x4882, 0x7E82, 0x48AA, 0x7EC6, 0x4882, 0x48BA, 0x7EAA, 0x2AA, 0xAABA, 0xAAAA, 0xAA82, 0x8282, 0x148A, 0x884}},
    {19345, 2, {0x4, 0x7E1E, 0x48E0, 0x7E12, 0x488A, 0x7E4C, 0x4840, 0x4808, 0x7EFE, 0x208, 0xAA88, 0xAA48, 0xAA28, 0x8208, 0x1428, 0x810}},
    {19346, 2, {0x14, 0x7E12, 0x48FE, 0x7E10, 0x4810, 0x7EFE, 0x4892, 0x4892, 0x7EFE, 0x292, 0xAA92, 0xAAFE, 0xAA92, 0x8292, 0x1492, 0x886}},
    {19347, 2, {0x20, 0x7E20, 0x483E, 0x7E20, 0x48FC, 0x7E84, 0x48FC, 0x4884, 0x7EFC, 0x2A4, 0xAA20, 0xABFE, 0xAA20, 0x8220, 0x1420, 0x820}},
    {19348, 2, {0x4, 0x7E1E, 0x48F0, 0x7E10, 0x48FE, 0x7E54, 0x4854, 0x4854, 0x7EFE, 0x254, 0xAA54, 0xAAFE, 0xAA10, 0x8210, 0x14FE, 0x800}},
    {19349, 2, {0x0, 0xFBDE, 0xA042, 0xF94A, 0xA084, 0xF94A, 0xA252, 0xA420, 0xF800, 0xBDE, 0xA852, 0xA952, 0xA894, 0x8948, 0xA54, 0x3422}},
    {19350, 2, {0x7C, 0x7E44, 0x4844, 0x7E7C, 0x4844, 0x7E44, 0x487C, 0x4800, 0x7EFE, 0x282, 0xAA82, 0xAAFE, 0xAA82, 0x8282, 0x14FE, 0x882}},
    {19351, 2, {0x10, 0x7E92, 0x4892, 0x7E92, 0x48FE, 0x7E00, 0x48FE, 0x4880, 0x7EBE, 0x288, 0xAA88, 0xAAFE, 0xAA88, 0x8308, 0x1508, 0xA08}},
    {19352, 2, {0x0, 0x7E20, 0x48CE, 0x7E82, 0x4882, 0x7EEE, 0x4882, 0x4882, 0x7EFE, 0x228, 0xAA28, 0xAA28, 0xAA48, 0x824A, 0x148A, 0x906}},
    {19353, 2, {0x8, 0x7E0C, 0x480A, 0x7E08, 0x49FE, 0x7E08, 0x4808, 0x48EA, 0x7EAA, 0x2AA, 0xAAEC, 0xAA0C, 0xAA2A, 0x82CA, 0x1416, 0x822}},
    {19354, 2, {0x10, 0x7E10, 0x487E, 0x7E10, 0x4810, 0x7EFE, 0x4824, 0x4842, 0x7E90, 0x23C, 0xAA44, 0xAAA4, 0xAA28, 0x8210, 0x1428, 0x8C6}},
    {19355, 2, {0x14, 0xFB94, 0xA294, 0xFAA4, 0xA2BE, 0xFAA4, 0xA364, 0xA2B4, 0xFAAC, 0xAAC, 0xAAA4, 0xAAA4, 0xAB24, 0x8A24, 0xA24, 0x322C}},
    {19356, 2, {0x14, 0x7E54, 0x4854, 0x7EFE, 0x4854, 0x7E54, 0x485C, 0x4840, 0x7E7E, 0x210, 0xAAFE, 0xAA10, 0xAA38, 0x8254, 0x1492, 0x810}},
    {19357, 2, {0x0, 0x7EFE, 0x4892, 0x7E10, 0x48FE, 0x7E10, 0x487C, 0x4854, 0x7E7C, 0x254, 0xAA7C, 0xAA10, 0xAAFE, 0x8210, 0x1410, 0x810}},
    {19358, 2, {0x20, 0x7E20, 0x49FE, 0x7E20, 0x48FC, 0x7E40, 0x49FE, 0x4888, 0x7F04, 0x2FA, 0xAA88, 0xAA88, 0xAAF8, 0x8288, 0x1488, 0x8F8}},
    {19359, 2, {0x40, 0x7C20, 0x51FE, 0x7D22, 0x50A4, 0x7D28, 0x5050, 0x5088, 0x7D04, 0x4F8, 0x5488, 0x5488, 0x5450, 0x8420, 0x14D8, 0xB06}},
    {19360, 2, {0x0, 0x7EFE, 0x4800, 0x7E7C, 0x4844, 0x7E44, 0x487C, 0x4800, 0x7EFE, 0x292, 0xAA92, 0xAAFE, 0xAA92, 0x8292, 0x14FE, 0x882}},
    {19361, 2, {0x40, 0x7E40, 0x487E, 0x7E80, 0x487C, 0x7E44, 0x487C, 0x4844, 0x7E7C, 0x220, 0xAA3C, 0xAA44, 0xAAA8, 0x8210, 0x1428, 0x8C6}},
    {19362, 2, {0x4, 0x7E1E, 0x48F0, 0x7E10, 0x4810, 0x7EFE, 0x4810, 0x4850, 0x7E96, 0x292, 0xAA92, 0xAAD6, 0xAA92, 0x8292, 0x14FE, 0x882}},
    {19363, 2, {0x0, 0x7EFC, 0x4844, 0x7E38, 0x4844, 0x7EEE, 0x48AA, 0x4844, 0x7EAA, 0x210, 0xAAFE, 0xAA38, 0xAA54, 0x8292, 0x1410, 0x810}},
    {19364, 2, {0x20, 0x7E20, 0x49FE, 0x7F52, 0x4848, 0x7EFE, 0x4890, 0x4990, 0x7EFE, 0x290, 0xAA90, 0xAAFE, 0xAA90, 0x8290, 0x14FE, 0x880}},
    {19365, 2, {0x0, 0x7EFE, 0x4800, 0x7E7C, 0x4844, 0x7E7C, 0x4800, 0x48FE, 0x7E82, 0x2AA, 0xAA92, 0xAAFE, 0xAA92, 0x8292, 0x1492, 0x886}},
    {19366, 2, {0x20, 0x7E10, 0x48FE, 0x7E00, 0x487C, 0x7E44, 0x487C, 0x4800, 0x7EFE, 0x282, 0xAABA, 0xAAAA, 0xAABA, 0x8282, 0x148A, 0x884}},
    {19367, 2, {0x1010, 0x1028, 0x1044, 0xFE82, 0x11FE, 0x7C90, 0x44FC, 0x7C90, 0x44FC, 0x7C90, 0x10FE, 0xFE02, 0x12AA, 0x12AA, 0x1414, 0x1008}},
    {19368, 2, {0x20, 0x7E10, 0x48FE, 0x7EA2, 0x4810, 0x7EFE, 0x4844, 0x4828, 0x7EFE, 0x210, 0xAA10, 0xAA7C, 0xAA10, 0x8210, 0x1410, 0x810}},
    {19369, 2, {0x0, 0x7E7C, 0x4844, 0x7E74, 0x4854, 0x7EFE, 0x4882, 0x487C, 0x7E44, 0x27C, 0xAA44, 0xAA7C, 0xAA44, 0x8244, 0x1454, 0x848}},
    {19370, 2, {0x84, 0x7C84, 0x50DE, 0x7D44, 0x5154, 0x7EDE, 0x5044, 0x5084, 0x7D24, 0x420, 0x55FE, 0x5470, 0x54A8, 0x8524, 0x1622, 0x820}},
    {19371, 2, {0x20, 0xF9FC, 0xA124, 0xFBFE, 0xA124, 0xF9FC, 0xA020, 0xA1FC, 0xF924, 0x9FC, 0xA840, 0xABFE, 0xA888, 0x89D0, 0x870, 0x338C}},
    {19372, 2, {0x48, 0x7E48, 0x49FE, 0x7E48, 0x4800, 0x7EFC, 0x4884, 0x48FC, 0x7E84, 0x2FC, 0xAA20, 0xABFE, 0xAA20, 0x8250, 0x1488, 0x906}},
    {19373, 2, {0xE, 0x7EF0, 0x4822, 0x7E92, 0x4844, 0x7EFE, 0x4882, 0x48FE, 0x7E82, 0x2FE, 0xAA82, 0xAAFE, 0xAA28, 0x8248, 0x148A, 0x906}},
    {19374, 2, {0x20, 0x7C20, 0x53FE, 0x7CA8, 0x50A8, 0x7D74, 0x5222, 0x5070, 0x7CA8, 0x546, 0x5678, 0x5488, 0x5550, 0x8420, 0x1450, 0x980}},
    {19375, 2, {0x88, 0xF888, 0xA088, 0xFBD0, 0xA09E, 0xFBE4, 0xA094, 0xA7D4, 0xF914, 0x9D4, 0xA954, 0xA948, 0xA948, 0x8A54, 0xAD4, 0x3422}},
    {19376, 2, {0x20, 0x7E10, 0x48FE, 0x7E44, 0x4828, 0x7EFE, 0x4882, 0x4892, 0x7EFE, 0x292, 0xAABA, 0xAAAA, 0xAAAA, 0x82BA, 0x1482, 0x886}},
    {19377, 2, {0x0, 0xFA7E, 0xA110, 0xF930, 0xA04A, 0xF89A, 0xA62C, 0xA24C, 0xFA9A, 0xA2A, 0xAA4A, 0xAA88, 0xAA28, 0x8D10, 0x8FE, 0x3000}},
    {19378, 2, {0x28, 0x7E28, 0x487C, 0x7E28, 0x4828, 0x7EFE, 0x4810, 0x487C, 0x7E54, 0x27C, 0xAA54, 0xAA7C, 0xAA00, 0x8228, 0x1444, 0x882}},
    {19379, 2, {0x3C, 0xFBE0, 0xA124, 0xF8A8, 0xA3FE, 0xF8A8, 0xA124, 0xA202, 0xF9FC, 0x924, 0xA924, 0xA9FC, 0xA924, 0x8924, 0x9FC, 0x3104}},
    {19380, 2, {0x10, 0x7ED4, 0x4848, 0x7F4A, 0x4884, 0x7E84, 0x497A, 0x4800, 0x7EFC, 0x284, 0xAAFC, 0xAA00, 0xAA84, 0x8248, 0x15FE, 0x800}},
    {19381, 2, {0x20, 0x7DFC, 0x5088, 0x7C50, 0x53FE, 0x7C00, 0x51FC, 0x5124, 0x7DFC, 0x524, 0x55FC, 0x5420, 0x55FC, 0x8420, 0x17FE, 0x800}},
    {19382, 2, {0x50, 0x7E48, 0x48FE, 0x7F90, 0x48FC, 0x7E90, 0x48FC, 0x4890, 0x7EFE, 0x280, 0xAA10, 0xAAFE, 0xAA38, 0x8254, 0x1492, 0x810}},
    {19383, 2, {0x0, 0x7EFE, 0x48AA, 0x7EAA, 0x48FE, 0x7E40, 0x48FE, 0x4922, 0x7EFA, 0x2AA, 0xAAAA, 0xAAFA, 0xAA22, 0x82FA, 0x140A, 0x804}},
    {19384, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x5102, 0x9FF4, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x492C}},
    {19385, 2, {0x20, 0xF93C, 0xA120, 0xFBFE, 0xA010, 0xF814, 0xA3FE, 0xA210, 0xFBF0, 0xA54, 0xAA54, 0xAAD4, 0xAB68, 0x8C4A, 0x896, 0x3322}},
    {19386, 2, {0x40, 0xF820, 0xA3FE, 0xF888, 0xA052, 0xFFAC, 0xA2AA, 0xA2A8, 0xFDAC, 0x800, 0xA904, 0xA9FC, 0xA904, 0x89FC, 0x904, 0x3204}},
    {19387, 2, {0x20, 0xFBFE, 0xA020, 0xF9FC, 0xA000, 0xFBFE, 0xA002, 0xA1FC, 0xF820, 0xBFE, 0xA800, 0xABFE, 0xA804, 0x8BBE, 0xAA4, 0x338C}},
    {19388, 2, {0x12, 0x7E54, 0x48A8, 0x7E44, 0x4892, 0x7E54, 0x4898, 0x4824, 0x7EC2, 0x284, 0xAAEE, 0xAB24, 0xAAB4, 0x825E, 0x1444, 0x884}},
    {19389, 2, {0x40, 0xF820, 0xA3FE, 0xFA50, 0xA3FE, 0xFA52, 0xA3FE, 0xA200, 0xFA94, 0xAD8, 0xAA90, 0xAAD2, 0xAA8E, 0x8C00, 0x954, 0x322A}},
    {19390, 2, {0x110, 0xF890, 0xA3DE, 0xF810, 0xA25E, 0xF982, 0xA3DE, 0xA010, 0xFBDE, 0xA50, 0xABDE, 0xAA50, 0xABDE, 0x8A50, 0xA52, 0x32CE}},
    {19391, 2, {0x20, 0xFBFE, 0xA24A, 0xF9FE, 0xA048, 0xF9FE, 0xA048, 0xA3FE, 0xF884, 0x9FE, 0xAA84, 0xA8FC, 0xA920, 0x893C, 0xAA0, 0x347E}},
    {19392, 2, {0x3FC, 0xF908, 0xA1F8, 0xF908, 0xA1F8, 0xF90E, 0xA7F8, 0xA008, 0xFFFE, 0xA94, 0xAB9C, 0xAA94, 0xAB9C, 0x8AD6, 0xFBC, 0x3084}},
    {19393, 2, {0x40, 0xFFE, 0xF200, 0xA3F8, 0xF000, 0xA7F8, 0xF408, 0xA7F8, 0xF220, 0x1D5C, 0x57F4, 0xBC94, 0x95DC, 0x1C96, 0x57F6, 0x2CA2}},
    {19394, 2, {0x3FF8, 0x2100, 0x3FF0, 0x2100, 0x3FF8, 0x9248, 0x7EFC, 0x4890, 0x7EFC, 0x4890, 0x7EFE, 0xAAAA, 0x100, 0xFFFE, 0x920, 0x711C}},
    {19395, 2, {0x40, 0xF820, 0x820, 0x4BFE, 0x4A02, 0x4C44, 0x4840, 0x7FFE, 0x488, 0x488, 0x1D08, 0xE4D0, 0x4420, 0x450, 0x2888, 0x1304}},
    {19396, 2, {0x0, 0xF8FC, 0x884, 0x4884, 0x4884, 0x48FC, 0x4820, 0x7C20, 0x5FE, 0x522, 0x1D52, 0xE58A, 0x450A, 0x502, 0x290A, 0x1104}},
    {19397, 2, {0x0, 0xF1FE, 0x1084, 0x50FC, 0x5084, 0x50FC, 0x5086, 0x7BFC, 0x804, 0xBDE, 0x3852, 0xC952, 0x894, 0x948, 0x2A54, 0x1422}},
    {19398, 2, {0x20, 0x7F20, 0x4120, 0x7920, 0x4920, 0xFFA0, 0x80A0, 0x7F20, 0x4120, 0x7F20, 0x4120, 0x7F22, 0x4122, 0x4122, 0x451E, 0x4200}},
    {19399, 2, {0x20, 0x7C20, 0x4420, 0x7420, 0x54FC, 0xFE24, 0x8224, 0x7C24, 0x4424, 0x7C44, 0x4444, 0x7C44, 0x4484, 0x4484, 0x5528, 0x4A10}},
    {19400, 2, {0x0, 0x7CFC, 0x4400, 0x7400, 0x5400, 0xFEFE, 0x8250, 0x7C50, 0x4450, 0x7C50, 0x4450, 0x7C52, 0x4452, 0x4492, 0x548E, 0x4900}},
    {19401, 2, {0x40, 0x7C40, 0x4440, 0x747C, 0x5444, 0xFE88, 0x82A0, 0x7D20, 0x4420, 0x7C50, 0x4450, 0x7C48, 0x4488, 0x4484, 0x5504, 0x4A02}},
    {19402, 2, {0x0, 0x7C00, 0x45FE, 0x7404, 0x5404, 0xFEF4, 0x8294, 0x7C94, 0x4494, 0x7C94, 0x44F4, 0x7C94, 0x4404, 0x4404, 0x5414, 0x4808}},
    {19403, 2, {0x50, 0x7C48, 0x4448, 0x7440, 0x55FE, 0xFE40, 0x8240, 0x7C7C, 0x44A4, 0x7CA4, 0x44A8, 0x7D28, 0x4510, 0x4628, 0x5444, 0x4882}},
    {19404, 2, {0x0, 0x7CFE, 0x4482, 0x7482, 0x54FE, 0xFEA0, 0x82A0, 0x7CA0, 0x44F8, 0x7CA8, 0x44A8, 0x7CA8, 0x4528, 0x452A, 0x564A, 0x4886}},
    {19405, 2, {0x20, 0x7C20, 0x4450, 0x7450, 0x5488, 0xFF24, 0x8212, 0x7C10, 0x45FC, 0x7C04, 0x4408, 0x7C88, 0x4450, 0x4420, 0x5410, 0x4810}},
    {19406, 2, {0x20, 0x7C20, 0x4450, 0x7450, 0x5488, 0xFF04, 0x82FA, 0x7C20, 0x4420, 0x7C20, 0x45FC, 0x7C20, 0x4420, 0x4420, 0x5420, 0x4820}},
    {19407, 2, {0x4, 0x7C1E, 0x44F0, 0x7410, 0x5410, 0xFEFE, 0x8210, 0x7C10, 0x4410, 0x7CFE, 0x4482, 0x7C82, 0x4482, 0x4482, 0x54FE, 0x4882}},
    {19408, 2, {0x10, 0x7C10, 0x4410, 0x74FE, 0x5492, 0xFE92, 0x82AA, 0x7CC6, 0x4492, 0x7C92, 0x44AA, 0x7CC6, 0x4482, 0x4482, 0x548A, 0x4884}},
    {19409, 2, {0x20, 0x7C20, 0x4524, 0x74A4, 0x54A8, 0xFE20, 0x83FE, 0x7C50, 0x4450, 0x7C50, 0x4450, 0x7C52, 0x4452, 0x4492, 0x548E, 0x4900}},
    {19410, 2, {0x20, 0x7C20, 0x444E, 0x7480, 0x5520, 0xFE20, 0x825E, 0x7CC4, 0x4544, 0x7C44, 0x4444, 0x7C44, 0x4444, 0x4444, 0x5454, 0x4848}},
    {19411, 2, {0x20, 0x7C20, 0x44FC, 0x7420, 0x5420, 0xFE20, 0x83FE, 0x7C00, 0x4420, 0x7C20, 0x44FC, 0x7C20, 0x4420, 0x4420, 0x55FE, 0x4800}},
    {19412, 2, {0x20, 0x7C10, 0x45FE, 0x7502, 0x5400, 0xFE1C, 0x82E0, 0x7C20, 0x4420, 0x7C3E, 0x45E0, 0x7C20, 0x4422, 0x4422, 0x541E, 0x4800}},
    {19413, 2, {0x0, 0x7C06, 0x45D8, 0x7448, 0x5448, 0xFE88, 0x829E, 0x7DC8, 0x4448, 0x7D48, 0x4548, 0x7C9E, 0x4480, 0x4540, 0x563E, 0x4C00}},
    {19414, 2, {0x20, 0x7C10, 0x44FC, 0x7484, 0x5484, 0xFEFC, 0x8284, 0x7C84, 0x44FC, 0x7CA2, 0x44A4, 0x7C98, 0x4490, 0x4488, 0x54C4, 0x4882}},
    {19415, 2, {0x110, 0x7D12, 0x4514, 0x75D8, 0x5510, 0xFF12, 0x8352, 0x7D8E, 0x4520, 0x7C20, 0x45FC, 0x7C20, 0x4420, 0x4420, 0x57FE, 0x4800}},
    {19416, 2, {0x20, 0x7C10, 0x45FE, 0x7502, 0x5604, 0xFEF8, 0x8200, 0x7C00, 0x45FE, 0x7C50, 0x4450, 0x7C50, 0x4492, 0x4492, 0x550E, 0x4A00}},
    {19417, 2, {0x14, 0x7C12, 0x44FE, 0x7410, 0x5410, 0xFEFE, 0x8292, 0x7C92, 0x44FE, 0x7C92, 0x4492, 0x7CFE, 0x4492, 0x4492, 0x5492, 0x4886}},
    {19418, 2, {0x20, 0x7C44, 0x44F8, 0x7410, 0x5424, 0xFEFE, 0x8202, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x5494, 0x4888}},
    {19419, 2, {0x40, 0x7C20, 0x45FC, 0x7504, 0x5480, 0xFE9C, 0x82D4, 0x7D54, 0x4554, 0x7D54, 0x46D8, 0x7C50, 0x4494, 0x4494, 0x5514, 0x4A0C}},
    {19420, 2, {0x0, 0x7CFC, 0x4484, 0x7484, 0x54FC, 0xFE84, 0x8284, 0x7CFC, 0x4440, 0x7CFE, 0x452A, 0x7E4A, 0x4492, 0x4522, 0x544A, 0x4884}},
    {19421, 2, {0x20, 0x7C20, 0x44FE, 0x7440, 0x54FC, 0xFF10, 0x82FE, 0x7C00, 0x44FC, 0x7C84, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x5494, 0x4888}},
    {19422, 2, {0x0, 0x7CF8, 0x4488, 0x7488, 0x54E8, 0xFEA8, 0x82A8, 0x7DFC, 0x4504, 0x7D74, 0x4554, 0x7D54, 0x4574, 0x4504, 0x5514, 0x4908}},
    {19423, 2, {0x0, 0x7C84, 0x4448, 0x7400, 0x55FE, 0xFE20, 0x8244, 0x7DA8, 0x4430, 0x7C58, 0x4598, 0x7C34, 0x4452, 0x4590, 0x5450, 0x4820}},
    {19424, 2, {0x0, 0x7CF8, 0x4488, 0x7488, 0x54F8, 0xFE88, 0x8288, 0x7CF8, 0x4400, 0x7DFC, 0x4554, 0x7D54, 0x4554, 0x4554, 0x57FE, 0x4800}},
    {19425, 2, {0x88, 0x7C50, 0x4400, 0x75FE, 0x5450, 0xFFFC, 0x8254, 0x7DFE, 0x4454, 0x7DFC, 0x4450, 0x7CD8, 0x4554, 0x4652, 0x5450, 0x4850}},
    {19426, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x37FA, 0x2208, 0x23C8, 0x2248, 0x2FFE, 0x2A0A, 0x23F8, 0x4208, 0x4228, 0x8210}},
    {19427, 2, {0x20, 0x7DFC, 0x4524, 0x75FC, 0x5420, 0xFFFE, 0x8200, 0x7DFC, 0x4504, 0x7DFC, 0x4504, 0x7DFC, 0x4504, 0x45FC, 0x5488, 0x4904}},
    {19428, 2, {0x10, 0x7C28, 0x4444, 0x74BA, 0x5500, 0xFEFE, 0x8292, 0x7CD6, 0x4492, 0x7CFE, 0x4400, 0x7C7C, 0x4444, 0x447C, 0x5444, 0x487C}},
    {19429, 2, {0x0, 0x7CFE, 0x4482, 0x74FE, 0x5482, 0xFEFE, 0x8244, 0x7CAA, 0x44EE, 0x7C44, 0x44AA, 0x7CEE, 0x4400, 0x44D4, 0x54AA, 0x492A}},
    {19430, 2, {0x24, 0x7CFE, 0x4424, 0x7400, 0x54FE, 0xFEAA, 0x82AA, 0x7CFE, 0x4414, 0x7CFE, 0x4490, 0x7CD2, 0x44AC, 0x44AA, 0x5496, 0x4922}},
    {19431, 2, {0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x17D0, 0x1450, 0x17D0, 0x1010, 0x1030}},
    {19432, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FEC, 0x200, 0xFFFE, 0x1830, 0xE00E}},
    {19433, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FEC, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2002, 0x1FFE}},
    {19434, 2, {0x2000, 0x11FE, 0xFE10, 0x20, 0x7CFC, 0x4484, 0x7C84, 0xFC, 0xFE84, 0x82FC, 0xBA84, 0xAA84, 0xAAFC, 0xBA48, 0x8284, 0x8702}},
    {19435, 2, {0x20, 0x7C10, 0x45FE, 0x7C00, 0x44FC, 0x7C84, 0x10FC, 0xFE00, 0x1FE, 0x7D02, 0x457A, 0x7D4A, 0x117A, 0x5502, 0x930A, 0x3104}},
    {19436, 2, {0x1000, 0xFEEE, 0x44AA, 0x7CAA, 0xEE, 0xFE00, 0xAA7C, 0xBA54, 0x827C, 0xFE54, 0x447C, 0x7C10, 0x10FE, 0xFE10, 0x1010, 0x1010}},
    {19437, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {19438, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x3FE0, 0x420, 0x47C, 0x804, 0x3014, 0xC008}},
    {19439, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x100, 0x1110, 0x2108, 0x4164, 0x380, 0x1C00, 0xE000}},
    {19440, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x280, 0x440, 0x1C70, 0xE44E, 0x440, 0x840, 0x1040}},
    {19441, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x7FFC, 0x100, 0x770, 0x190C, 0xE102, 0x100}},
    {19442, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2002, 0x2002, 0x1FFE}},
    {19443, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x440, 0x820, 0x3218, 0xC446, 0x820, 0x1FF0, 0x810}},
    {19444, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0xFE0, 0x1020, 0x601C, 0x1FE0, 0x820, 0x4C0, 0x780, 0xF87E}},
    {19445, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x100, 0x3FF8, 0x100, 0x820, 0xFFFE, 0x820, 0x1020}},
    {19446, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x0}},
    {19447, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x90, 0xFFFE, 0x1088, 0x7E50, 0x1024, 0x20D4, 0xC70C}},
    {19448, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x4100, 0x21FC, 0xA44, 0x1048, 0xE0A0, 0x2118, 0x2606}},
    {19449, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3FF0, 0x10, 0x1FF0, 0x10, 0x7FFE, 0x4002, 0x9FF4, 0x820, 0x7C0, 0xF83E}},
    {19450, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x18, 0x3FE0, 0x2100, 0x3FFC, 0x2080, 0x2C64, 0x301C, 0x1FF0, 0x1010, 0x1010, 0x1FF0}},
    {19451, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1100, 0x1FF8, 0x2100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {19452, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x1400, 0x820, 0x14A4, 0x6CA8, 0x1550, 0x6488, 0x1906}},
    {19453, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x3FF8, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {19454, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3FF8, 0x2108, 0x27C8, 0x2108, 0x2FE8, 0x2008, 0x27C8, 0x4448, 0x47C8, 0x8018}},
    {19455, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x100, 0x3FF8, 0x820, 0x1450, 0x100, 0x7FFC, 0x100}},
    {19456, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1FF0, 0x100, 0xFFFE, 0x920, 0x711C, 0x200, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {19457, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x1110, 0x9A0, 0x540, 0x1930, 0xE30E}},
    {19458, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x18, 0x3FE0, 0x1210, 0x920, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {19459, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x820, 0x440, 0x7FFC, 0x440, 0x2448, 0x1450, 0xFFFE}},
    {19460, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x4904, 0x4892, 0x87F2}},
    {19461, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820}},
    {19462, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1FF0, 0x220, 0x140, 0x7FFC, 0x1884, 0x6288, 0x100, 0xFFFE, 0x920, 0x711C}},
    {19463, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x200, 0x1FD0, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x810}},
    {19464, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x800, 0x37F0, 0xC080, 0x1FFC}},
    {19465, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1078, 0x3E48, 0x2248, 0x2A86, 0xFF78, 0x2248, 0x2A28, 0x4210, 0x4A28, 0x84C6}},
    {19466, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x110, 0xFFFE, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x20, 0xFFFE, 0x820, 0x460}},
    {19467, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4404, 0x3FF8, 0x820, 0x7C0, 0x7838}},
    {19468, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x1088, 0x2154, 0x4A22, 0x113C, 0x3120, 0x52A0, 0x147E}},
    {19469, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x7FFC, 0x910, 0x3108, 0xC6C4, 0x1930, 0xE92E, 0x5C0, 0x1930, 0x6108, 0x300}},
    {19470, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820}},
    {19471, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x828, 0x7F24, 0x820, 0xFFFE, 0x1220, 0x3FA4, 0x6424, 0xBF28, 0x2412, 0x3FAA, 0x2046}},
    {19472, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {19473, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8514, 0x7FFC, 0x200, 0x3FF8, 0x2448, 0x2458}},
    {19474, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0x7FFC, 0x820, 0x7FFE, 0x4002, 0x9FF4, 0x600, 0x1920, 0x6C0, 0x38A0, 0x318}},
    {19475, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x200, 0x1FF0, 0x1490, 0x1250, 0xFFFE, 0x9212, 0x2FE8, 0x4404, 0x7E0, 0x820, 0x3060}},
    {19476, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x800, 0x3E78, 0x2248, 0x2A86, 0x7E78, 0x2248, 0x2A30, 0x464C, 0x3FF8, 0x2448, 0xFFFE}},
    {19477, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1FC, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x4FFC, 0x4914, 0xBFFE}},
    {19478, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x2850, 0x3E7C, 0x4890, 0xFEFE, 0x1528, 0x264A, 0x5FF6, 0x1490, 0x1250, 0x1FF0, 0x820}},
    {19479, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8402, 0x8782, 0x8882, 0x97C2, 0x8442, 0x8FE2, 0x8922, 0x8FE2, 0x8922, 0x8F12, 0x81F6}},
    {19480, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8002, 0xBEFA, 0x924A, 0x8A2A, 0x934A, 0x8282, 0x8C62, 0xB19A, 0x8602, 0x8072, 0x9F86}},
    {19481, 2, {0x10, 0xFE10, 0x10, 0x7CFE, 0x4410, 0x7C10, 0xFC, 0xFE44, 0x8244, 0xAA44, 0x9228, 0xFE28, 0x9210, 0x9228, 0x9244, 0x8682}},
    {19482, 2, {0x20, 0xFE10, 0x10, 0x7CFE, 0x4444, 0x7C44, 0x44, 0xFE44, 0x8244, 0xAA28, 0x9228, 0xFE10, 0x9210, 0x9228, 0x9244, 0x8682}},
    {19483, 2, {0x0, 0xFE00, 0x1FE, 0x7C40, 0x4440, 0x7C78, 0x48, 0xFE48, 0x8268, 0xAA98, 0x9298, 0xFE88, 0x928A, 0x92AA, 0x92CA, 0x8686}},
    {19484, 2, {0x0, 0xEFEE, 0x2002, 0x27C2, 0x2442, 0xE7CE, 0x8008, 0x8FE8, 0x8828, 0xEAAE, 0x2922, 0x2FE2, 0x2922, 0x2922, 0xA92A, 0x4864}},
    {19485, 2, {0x44, 0xFE24, 0x28, 0x7CFE, 0x4492, 0x7CD6, 0xBA, 0xFE92, 0x82FE, 0xAA00, 0x927C, 0xFE44, 0x927C, 0x9244, 0x927C, 0x8644}},
    {19486, 2, {0x713C, 0x17C4, 0x719C, 0x47F0, 0x725C, 0x17C4, 0x5A54, 0x23C8, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x5FF4, 0x410C}},
    {19487, 2, {0x3FF8, 0x0, 0xFFFE, 0x820, 0x1010, 0x3FF8, 0x208, 0x1FF0, 0x1110, 0x1FF0, 0x1210, 0x1FF0, 0x4A8, 0x8BA, 0x3082, 0xC07E}},
    {19488, 2, {0x810, 0x1010, 0x7F10, 0x497C, 0x4954, 0x7F54, 0x497C, 0x4954, 0x7F54, 0x107C, 0x1A10, 0x2C90, 0x2F90, 0x4810, 0x4802, 0x87FE}},
    {19489, 2, {0x808, 0x1008, 0x7F28, 0x4928, 0x493E, 0x7F48, 0x4908, 0x4908, 0x7F7E, 0x1008, 0x1A08, 0x2C94, 0x2F92, 0x4820, 0x4842, 0x87FE}},
    {19490, 2, {0x1020, 0x2040, 0x7C8E, 0x5500, 0x5420, 0x7C5E, 0x54C4, 0x5544, 0x7C44, 0x1044, 0x2A44, 0x2D54, 0x4F48, 0x4802, 0x87FE, 0x0}},
    {19491, 2, {0x1000, 0x21FC, 0x7D04, 0x5504, 0x55AC, 0x7D54, 0x5554, 0x5554, 0x7D54, 0x11AC, 0x2B04, 0x2D14, 0x4F08, 0x4802, 0x87FE, 0x0}},
    {19492, 2, {0x1020, 0x1040, 0xFDFC, 0x1124, 0x7D24, 0x55FC, 0x7D24, 0x5524, 0x7DFC, 0x1040, 0x3868, 0x54B2, 0x92BE, 0x1120, 0x1122, 0x121E}},
    {19493, 2, {0x50, 0x7FFC, 0x40, 0x3E48, 0x2250, 0x3E20, 0x54, 0x7FFC, 0x1114, 0x1FF0, 0x1210, 0x1FF0, 0x4A8, 0x8BA, 0x3082, 0xC07E}},
    {19494, 2, {0x800, 0x1008, 0x7F10, 0x497E, 0x494A, 0x7F7E, 0x4952, 0x497E, 0x7F28, 0x1048, 0x1A7E, 0x2C88, 0x2F88, 0x4808, 0x4802, 0x87FE}},
    {19495, 2, {0x1020, 0x203C, 0x7C20, 0x54FE, 0x54A2, 0x7CB8, 0x54E4, 0x549C, 0x7C80, 0x10B8, 0x2AA8, 0x2D2A, 0x4F46, 0x4A02, 0x87FE, 0x0}},
    {19496, 2, {0x1000, 0x21FC, 0x7D24, 0x55FC, 0x5500, 0x7D7C, 0x5544, 0x557C, 0x7D44, 0x117C, 0x2A44, 0x2D7C, 0x4F00, 0x4802, 0x87FE, 0x0}},
    {19497, 2, {0x1020, 0x1040, 0x7DFC, 0x1294, 0x1494, 0xFFFC, 0x10A4, 0x20A4, 0x7EFC, 0xA220, 0x2244, 0x3E6A, 0x226E, 0x22A0, 0x3EA2, 0x231E}},
    {19498, 2, {0x1020, 0x9240, 0x54FC, 0xAA94, 0x4494, 0x82FC, 0x7CA4, 0x44A4, 0x44FC, 0x7C20, 0x4444, 0x446A, 0x7C6E, 0x44A0, 0x44A2, 0x4D1E}},
    {19499, 2, {0x1010, 0x207C, 0x7C10, 0x54FE, 0x5444, 0x7CEE, 0x5444, 0x54EE, 0x7C00, 0x10FE, 0x2A28, 0x2D2A, 0x4F46, 0x4882, 0x87FE, 0x0}},
    {19500, 2, {0x10FC, 0x2028, 0x7C10, 0x55FE, 0x5452, 0x7C90, 0x55FE, 0x54AA, 0x7CC6, 0x10BA, 0x2AAA, 0x2DBA, 0x4F84, 0x4802, 0x87FE, 0x0}},
    {19501, 2, {0x20, 0xFE40, 0x10FC, 0xFE94, 0xAA94, 0xAAFC, 0xAAA4, 0x86A4, 0xFC, 0xFE20, 0x1044, 0xFE6A, 0xAA6E, 0xAAA0, 0xAAA2, 0x871E}},
    {19502, 2, {0x2088, 0x43FE, 0xF820, 0xA9FC, 0xA820, 0xFBFE, 0xA854, 0xA992, 0xF890, 0x23FE, 0x2090, 0x54D4, 0x5B8A, 0x5E96, 0x51A2, 0x8FFE}},
    {19503, 2, {0x1100, 0x215C, 0x7D84, 0x5558, 0x54C8, 0x7D3E, 0x55CA, 0x56A8, 0x7DEC, 0x10A8, 0x2B68, 0x2D5E, 0x4F80, 0x4802, 0x87FE, 0x0}},
    {19504, 2, {0x20, 0x7C40, 0x10FC, 0xFE94, 0x9294, 0x54FC, 0x10A4, 0x54A4, 0xFC, 0xFE20, 0x1044, 0xFE6A, 0xAA6E, 0xAAA0, 0xAAA2, 0x871E}},
    {19505, 2, {0x20, 0xFE40, 0x10FC, 0xFE94, 0x9294, 0x54FC, 0xA4, 0xEEA4, 0xAAFC, 0xEE20, 0x44, 0xFE6A, 0x546E, 0xBAA0, 0x10A2, 0xFF1E}},
    {19506, 2, {0x1020, 0x1020, 0x3E20, 0x4220, 0x8420, 0x7F20, 0x4920, 0x4920, 0x7F20, 0x4920, 0x4920, 0x7F22, 0x22, 0x5522, 0x4A9E, 0x8A80}},
    {19507, 2, {0x2000, 0x2000, 0x7CFE, 0x4410, 0x8810, 0x7E10, 0x5210, 0x5210, 0x7E10, 0x5210, 0x5210, 0x7E10, 0x10, 0xAA10, 0xAA50, 0x20}},
    {19508, 2, {0x2040, 0x2040, 0x7880, 0x48FC, 0x9100, 0x7E00, 0x54F8, 0x5408, 0x7C10, 0x5420, 0x5440, 0x7C80, 0x84, 0xAA84, 0xAA7C, 0x0}},
    {19509, 2, {0x2000, 0x21FC, 0x7C04, 0x4408, 0x8810, 0x7E20, 0x5220, 0x53FE, 0x7E20, 0x5220, 0x5220, 0x7E20, 0x20, 0xAA20, 0xAAA0, 0x40}},
    {19510, 2, {0x2020, 0x2010, 0x7C10, 0x4400, 0x89FE, 0x7E40, 0x5240, 0x5240, 0x7E40, 0x5240, 0x5240, 0x7E40, 0x40, 0xAA40, 0xAA7E, 0x0}},
    {19511, 2, {0x2020, 0x2020, 0x7820, 0x49FE, 0x9020, 0x7C20, 0x5420, 0x54FC, 0x7C20, 0x5420, 0x5420, 0x7DFE, 0x20, 0xAA20, 0xAA20, 0x20}},
    {19512, 2, {0x2008, 0x2048, 0x7848, 0x4848, 0x9044, 0x7C84, 0x54A4, 0x5522, 0x7C20, 0x5440, 0x5440, 0x7C48, 0x84, 0xABFE, 0xAA82, 0x0}},
    {19513, 2, {0x2028, 0x2024, 0x7C24, 0x4420, 0x882E, 0x7EF0, 0x5220, 0x5224, 0x7E24, 0x5228, 0x5228, 0x7E10, 0x32, 0xAA4A, 0xAB86, 0x2}},
    {19514, 2, {0x2004, 0x200E, 0x7C70, 0x4440, 0x8840, 0x7E40, 0x527E, 0x5248, 0x7E48, 0x5248, 0x5248, 0x7E48, 0x48, 0xAA88, 0xAA88, 0x108}},
    {19515, 2, {0x1020, 0x1020, 0x3E20, 0x423E, 0x8444, 0x7F44, 0x4944, 0x49A4, 0x7F28, 0x4928, 0x4910, 0x7F10, 0x28, 0x5528, 0x4AC4, 0x8A82}},
    {19516, 2, {0x2020, 0x2020, 0x7820, 0x4850, 0x9050, 0x7C88, 0x5544, 0x5622, 0x7C20, 0x54F8, 0x5408, 0x7C08, 0x10, 0xAA10, 0xAA20, 0x20}},
    {19517, 2, {0x2020, 0x2020, 0x7820, 0x4BFE, 0x9020, 0x7C20, 0x55FC, 0x5524, 0x7D24, 0x5524, 0x5524, 0x7D34, 0x128, 0xAA20, 0xAA20, 0x20}},
    {19518, 2, {0x2000, 0x2000, 0x78FC, 0x48A4, 0x90A4, 0x7CA4, 0x54A4, 0x54A4, 0x7CFC, 0x5480, 0x5480, 0x7C80, 0x82, 0xAA82, 0xAA7E, 0x0}},
    {19519, 2, {0x2040, 0x2040, 0x787E, 0x4880, 0x9120, 0x7CA0, 0x54AC, 0x54B4, 0x7DE4, 0x54A4, 0x54B4, 0x7CA8, 0xA2, 0xAA82, 0xAA7E, 0x0}},
    {19520, 2, {0x2020, 0x2020, 0x7820, 0x48FC, 0x90A4, 0x7CA4, 0x54A4, 0x54A4, 0x7CA4, 0x55FE, 0x5420, 0x7C50, 0x50, 0xAA88, 0xAB04, 0x202}},
    {19521, 2, {0x2020, 0x2020, 0x7820, 0x4BFE, 0x9050, 0x7C50, 0x5488, 0x5524, 0x7E22, 0x5420, 0x55FC, 0x7C20, 0x20, 0xAA20, 0xAA20, 0x20}},
    {19522, 2, {0x2010, 0x2050, 0x7850, 0x4890, 0x90FC, 0x7D54, 0x55D4, 0x5494, 0x7C94, 0x5514, 0x5554, 0x7DD4, 0x54, 0xAA24, 0xAA34, 0x48}},
    {19523, 2, {0x2010, 0x2010, 0x7C50, 0x4450, 0x887C, 0x7E90, 0x5310, 0x5210, 0x7EFE, 0x5210, 0x5228, 0x7E28, 0x28, 0xAA44, 0xAA44, 0x82}},
    {19524, 2, {0x2008, 0x201C, 0x79F0, 0x4950, 0x9150, 0x7D50, 0x5550, 0x5550, 0x7D50, 0x5548, 0x5548, 0x7D68, 0x154, 0xAA74, 0xAA52, 0x400}},
    {19525, 2, {0x2020, 0x2020, 0x7820, 0x4BFE, 0x9020, 0x7C20, 0x5420, 0x55FC, 0x7C70, 0x54A8, 0x54A8, 0x7D24, 0x124, 0xAA22, 0xAA20, 0x20}},
    {19526, 2, {0x2000, 0x21FC, 0x7844, 0x4848, 0x908E, 0x7C82, 0x557A, 0x5614, 0x7C20, 0x5420, 0x57FE, 0x7C20, 0x20, 0xAA20, 0xAAA0, 0x40}},
    {19527, 2, {0x2000, 0x2000, 0x7CFC, 0x4484, 0x8884, 0x7E84, 0x52FC, 0x5284, 0x7E84, 0x5284, 0x52FC, 0x7E84, 0x0, 0xAA00, 0xABFE, 0x0}},
    {19528, 2, {0x2000, 0x20FC, 0x7800, 0x4800, 0x9000, 0x7DFC, 0x5420, 0x5420, 0x7CA8, 0x54A4, 0x5524, 0x7D22, 0x222, 0xAA20, 0xAAA0, 0x40}},
    {19529, 2, {0x2000, 0x20F8, 0x7888, 0x4888, 0x9088, 0x7CF8, 0x5488, 0x5488, 0x7C88, 0x54F8, 0x5488, 0x7C88, 0x88, 0xAA88, 0xABFE, 0x0}},
    {19530, 2, {0x2020, 0x2020, 0x7924, 0x48A4, 0x90A8, 0x7C20, 0x5420, 0x57FE, 0x7C70, 0x54A8, 0x54A8, 0x7D24, 0x124, 0xAA22, 0xAA20, 0x20}},
    {19531, 2, {0x2048, 0x2048, 0x7848, 0x4848, 0x91FE, 0x7C48, 0x5448, 0x5448, 0x7C48, 0x55FE, 0x5400, 0x7C48, 0x44, 0xAA84, 0xAA82, 0x102}},
    {19532, 2, {0x2000, 0x20FE, 0x7C90, 0x4490, 0x8890, 0x7EBC, 0x52A4, 0x52A4, 0x7EA4, 0x52A4, 0x52BC, 0x7E90, 0x90, 0xAA90, 0xAAFE, 0x0}},
    {19533, 2, {0x2000, 0x21FE, 0x7840, 0x4840, 0x9040, 0x7C78, 0x5488, 0x54A8, 0x7C98, 0x5508, 0x5548, 0x7C30, 0x10, 0xAA10, 0xABFE, 0x0}},
    {19534, 2, {0x2000, 0x21FE, 0x7800, 0x4800, 0x90FC, 0x7C84, 0x5484, 0x54FC, 0x7C84, 0x5484, 0x54FC, 0x7C84, 0x0, 0xAA00, 0xABFE, 0x0}},
    {19535, 2, {0x2000, 0x21FE, 0x7800, 0x4800, 0x90FC, 0x7C84, 0x5484, 0x5484, 0x7CFC, 0x5400, 0x5484, 0x7C44, 0x48, 0xAA00, 0xABFE, 0x0}},
    {19536, 2, {0x2008, 0x203C, 0x79C0, 0x4804, 0x9144, 0x7CA8, 0x5400, 0x55F8, 0x7C10, 0x5420, 0x57FE, 0x7C20, 0x20, 0xAA20, 0xAAA0, 0x40}},
    {19537, 2, {0x2080, 0x208C, 0x78B0, 0x48A0, 0x93E0, 0x7CA0, 0x54BE, 0x54E8, 0x7CA8, 0x57A8, 0x54A8, 0x7CA8, 0xA8, 0xAAA8, 0xAAC8, 0x188}},
    {19538, 2, {0x2000, 0x207C, 0x7C44, 0x4444, 0x887C, 0x7E00, 0x52FE, 0x5292, 0x7E92, 0x5292, 0x52FE, 0x7E80, 0x80, 0xAA82, 0xAA82, 0x7E}},
    {19539, 2, {0x2000, 0x2006, 0x79D8, 0x4848, 0x9048, 0x7C88, 0x549E, 0x55C8, 0x7C48, 0x5548, 0x5548, 0x7C9E, 0x80, 0xAB40, 0xAA3E, 0x400}},
    {19540, 2, {0x2088, 0x2088, 0x7888, 0x4910, 0x915E, 0x7D64, 0x5754, 0x5554, 0x7D54, 0x5554, 0x5554, 0x7D48, 0x148, 0xAB54, 0xAB14, 0x122}},
    {19541, 2, {0x2080, 0x2080, 0x78FE, 0x4900, 0x92FC, 0x7C84, 0x54A4, 0x5494, 0x7DFE, 0x5484, 0x5524, 0x7D14, 0x1FE, 0xAA04, 0xAA28, 0x10}},
    {19542, 2, {0x2000, 0x20FE, 0x7C28, 0x4428, 0x8828, 0x7EEE, 0x52AA, 0x52AA, 0x7EAA, 0x52AA, 0x52EE, 0x7E28, 0x28, 0xAA28, 0xAA48, 0x88}},
    {19543, 2, {0x800, 0xF7C, 0x2844, 0x7F28, 0xA110, 0x1628, 0x3846, 0xCFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {19544, 2, {0x304, 0x3C24, 0x424, 0x7FA4, 0x1624, 0x2504, 0x440C, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {19545, 2, {0x2080, 0x2080, 0x789C, 0x48F4, 0x9094, 0x7C94, 0x57F4, 0x5414, 0x7C94, 0x5494, 0x55C8, 0x7EA8, 0x88, 0xAA94, 0xAA94, 0x1A2}},
    {19546, 2, {0x2000, 0x21F8, 0x7808, 0x4808, 0x90F8, 0x7C08, 0x5408, 0x55FC, 0x7C20, 0x5524, 0x54A8, 0x7C70, 0xA8, 0xAB24, 0xAAA2, 0x40}},
    {19547, 2, {0x2014, 0x2012, 0x7810, 0x49FE, 0x9010, 0x7C10, 0x55D2, 0x5552, 0x7D52, 0x5554, 0x55D4, 0x7C08, 0xCA, 0xAB1A, 0xAA26, 0x42}},
    {19548, 2, {0x2048, 0x2048, 0x7848, 0x49FE, 0x9048, 0x7C48, 0x55FE, 0x5400, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {19549, 2, {0x2020, 0x2040, 0x78FE, 0x4892, 0x9092, 0x7CFE, 0x5492, 0x54A2, 0x7CFE, 0x5448, 0x5488, 0x7DFE, 0x8, 0xAA08, 0xAA08, 0x8}},
    {19550, 2, {0x2020, 0x2020, 0x79FE, 0x4850, 0x9088, 0x7D04, 0x56FA, 0x5400, 0x7C00, 0x55FE, 0x5420, 0x7CA4, 0xA2, 0xAB22, 0xAA20, 0x60}},
    {19551, 2, {0x2000, 0x20FE, 0x7C82, 0x4482, 0x88FE, 0x7E88, 0x5288, 0x52FE, 0x7E88, 0x5288, 0x52BE, 0x7EA2, 0xA2, 0xAAA2, 0xAABE, 0x122}},
    {19552, 2, {0x2048, 0x2044, 0x785E, 0x49E0, 0x9028, 0x7C12, 0x546A, 0x5596, 0x7C48, 0x545E, 0x55E0, 0x7C24, 0x28, 0xAA12, 0xAA6A, 0x186}},
    {19553, 2, {0x2080, 0x2080, 0x78FE, 0x4902, 0x9222, 0x7CAA, 0x5472, 0x5422, 0x7DFE, 0x5422, 0x5472, 0x7CAA, 0x122, 0xAA22, 0xAA0A, 0x4}},
    {19554, 2, {0x2080, 0x2080, 0x78F8, 0x4908, 0x9210, 0x7DFC, 0x5424, 0x5424, 0x7DFE, 0x5424, 0x5424, 0x7DFC, 0x24, 0xAA20, 0xAAA0, 0x40}},
    {19555, 2, {0x2040, 0x2020, 0x79FE, 0x4800, 0x9088, 0x7C88, 0x5554, 0x5622, 0x7C00, 0x5420, 0x55FE, 0x7C20, 0x20, 0xAA20, 0xAA20, 0x20}},
    {19556, 2, {0x2040, 0x2040, 0x787C, 0x4884, 0x9108, 0x7E00, 0x5420, 0x54CE, 0x7C82, 0x5482, 0x54EE, 0x7C82, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {19557, 2, {0x2404, 0x3FA4, 0x4424, 0xFFA4, 0x424, 0x3F84, 0x2494, 0x2FE8, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {19558, 2, {0x2050, 0x2048, 0x7880, 0x48FE, 0x9190, 0x7E90, 0x54FC, 0x5490, 0x7C90, 0x54FC, 0x5490, 0x7C90, 0x90, 0xAAFE, 0xAA80, 0x80}},
    {19559, 2, {0x2020, 0x2124, 0x78A8, 0x4820, 0x91FC, 0x7C40, 0x57FE, 0x5488, 0x7D04, 0x56FA, 0x5488, 0x7C88, 0xA8, 0xAA92, 0xAA82, 0x7E}},
    {19560, 2, {0x2082, 0x2082, 0x73E2, 0x908A, 0x208A, 0xFBEA, 0xAAAA, 0xAAAA, 0xFAAA, 0xAAEA, 0xA88A, 0xF9CA, 0x2A2, 0xAC82, 0xA88A, 0x84}},
    {19561, 2, {0x2000, 0x20FE, 0x7C82, 0x44C6, 0x88AA, 0x7EFE, 0x52A2, 0x5292, 0x7EFE, 0x52A2, 0x52A2, 0x7EBA, 0x82, 0xAA82, 0xAA8A, 0x84}},
    {19562, 2, {0x2010, 0x2010, 0x79FE, 0x4810, 0x9010, 0x7CFE, 0x5410, 0x5410, 0x7DFE, 0x5420, 0x5452, 0x7CD4, 0x148, 0xAA44, 0xAA62, 0x40}},
    {19563, 2, {0x404, 0x7FC4, 0x424, 0x3FA4, 0x24A4, 0x3FA4, 0x1504, 0x2494, 0x4FE8, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488}},
    {19564, 2, {0x2000, 0x21FE, 0x7812, 0x4890, 0x909C, 0x7C90, 0x557E, 0x5600, 0x7CFC, 0x5484, 0x54FC, 0x7C84, 0xFC, 0xAA84, 0xAA94, 0x88}},
    {19565, 2, {0x2080, 0x20FE, 0x7890, 0x4890, 0x90D0, 0x7CBC, 0x5594, 0x5694, 0x7CAC, 0x54A4, 0x54B4, 0x7C88, 0x88, 0xAA88, 0xAAFE, 0x80}},
    {19566, 2, {0x2080, 0x209E, 0x73EA, 0x908A, 0x21CA, 0xF88A, 0xABEA, 0xA892, 0xF8A6, 0xA820, 0xABFE, 0xF820, 0x50, 0xA888, 0xA904, 0x602}},
    {19567, 2, {0x7E20, 0x243E, 0x1848, 0xFFA8, 0x2910, 0x4A28, 0x9846, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {19568, 2, {0x2008, 0x203C, 0x79E0, 0x4820, 0x93FE, 0x7C20, 0x55FC, 0x5524, 0x7DFC, 0x5524, 0x55FC, 0x7C20, 0x1FC, 0xAA20, 0xABFE, 0x0}},
    {19569, 2, {0x2020, 0x2010, 0x7CFE, 0x4400, 0x8844, 0x7E28, 0x52FE, 0x5282, 0x7E10, 0x527C, 0x5254, 0x7E54, 0x54, 0xAA5C, 0xAA10, 0x10}},
    {19570, 2, {0x2040, 0x207C, 0x7884, 0x48F8, 0x9008, 0x7DFE, 0x5440, 0x54A4, 0x7D38, 0x5458, 0x5494, 0x7D34, 0x54, 0xAA92, 0xAB50, 0x20}},
    {19571, 2, {0x2000, 0x21FE, 0x7900, 0x497C, 0x9154, 0x7D54, 0x557C, 0x5554, 0x7D54, 0x557C, 0x5510, 0x7D10, 0x17C, 0xAA10, 0xAA10, 0x4FE}},
    {19572, 2, {0x2040, 0x205E, 0x7840, 0x4860, 0x915E, 0x7D52, 0x5552, 0x555E, 0x7C52, 0x5452, 0x5452, 0x7C5E, 0x52, 0xAA40, 0xAA5E, 0x40}},
    {19573, 2, {0x2050, 0x2088, 0x7924, 0x4840, 0x9090, 0x7DF8, 0x5408, 0x5400, 0x7DDC, 0x5444, 0x5554, 0x7CCC, 0x154, 0xAA44, 0xAA54, 0xC8}},
    {19574, 2, {0x2100, 0x209E, 0x79D2, 0x4954, 0x9154, 0x7DD8, 0x5554, 0x5552, 0x7DD2, 0x5512, 0x559A, 0x7D54, 0x1D0, 0xAB50, 0xAA10, 0x10}},
    {19575, 2, {0xF10, 0x810, 0x7F90, 0x48A0, 0x4E3E, 0x7944, 0x4F24, 0x5124, 0x5F24, 0x7528, 0x5F28, 0x5510, 0x5F28, 0x4028, 0xAAC4, 0x82}},
    {19576, 2, {0x20A0, 0x212C, 0x7924, 0x4924, 0x91AC, 0x7D24, 0x5524, 0x55FC, 0x7C20, 0x55FC, 0x5488, 0x7C50, 0x20, 0xAA50, 0xAA88, 0x306}},
    {19577, 2, {0x2044, 0x2024, 0x7C28, 0x44FE, 0x8810, 0x7E7C, 0x5210, 0x52FE, 0x7E20, 0x5220, 0x527C, 0x7E50, 0x90, 0xAB10, 0xAAFE, 0x0}},
    {19578, 2, {0x2020, 0x2124, 0x7924, 0x4924, 0x91FC, 0x7C00, 0x57FE, 0x5400, 0x7DFC, 0x5504, 0x5504, 0x7DFC, 0x88, 0xAA50, 0xABFE, 0x0}},
    {19579, 2, {0x2000, 0x207C, 0x7C44, 0x4474, 0x8854, 0x7EFE, 0x5282, 0x527C, 0x7E44, 0x527C, 0x5244, 0x7E7C, 0x44, 0xAA44, 0xAA54, 0x48}},
    {19580, 2, {0x2000, 0x21FE, 0x7902, 0x4902, 0x91FE, 0x7D24, 0x5524, 0x557E, 0x7D24, 0x5524, 0x55FE, 0x7D52, 0x154, 0xAA48, 0xAA64, 0x442}},
    {19581, 2, {0x2020, 0x2020, 0x7850, 0x4848, 0x90A4, 0x7DFE, 0x5684, 0x54FC, 0x7C84, 0x54FC, 0x5480, 0x7CFC, 0x144, 0xAB44, 0xAA7C, 0x44}},
    {19582, 2, {0x2020, 0x21FC, 0x7924, 0x4BFE, 0x9124, 0x7DFC, 0x5420, 0x55FC, 0x7D24, 0x55FC, 0x5440, 0x7DFE, 0x88, 0xABD0, 0xAA70, 0x38C}},
    {19583, 2, {0x2100, 0x2102, 0x711C, 0x97D0, 0x2110, 0xFFD0, 0xAD5E, 0xAFD4, 0xFD54, 0xAFD4, 0xA914, 0xFFD4, 0x114, 0xA924, 0xA924, 0x144}},
    {19584, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFE6, 0x1042, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {19585, 2, {0x2004, 0x23E4, 0x7224, 0x9224, 0x23EE, 0xFA04, 0xAAE4, 0xAA04, 0xFA0C, 0xABF4, 0xAA44, 0xFA64, 0x554, 0xAD44, 0xAA54, 0xC8}},
    {19586, 2, {0x2210, 0x2110, 0x7110, 0x97BE, 0x2240, 0xFA00, 0xABBE, 0xAA8A, 0xFA88, 0xAAA8, 0xAAAE, 0xFAA8, 0x4A8, 0xADA8, 0xA85E, 0x1080}},
    {19587, 2, {0x2092, 0x2124, 0x7892, 0x4800, 0x90FE, 0x7C92, 0x54FE, 0x5492, 0x7CFE, 0x5410, 0x55FE, 0x7C38, 0x54, 0xAA92, 0xAB10, 0x10}},
    {19588, 2, {0x2208, 0x2208, 0x73C8, 0x9210, 0x27DE, 0xFA54, 0xAB64, 0xAAD4, 0xFFF4, 0xAA54, 0xAD54, 0xFCC8, 0x7C8, 0xA854, 0xAAA4, 0x142}},
    {19589, 2, {0x2100, 0x213C, 0x7124, 0x93A4, 0x213C, 0xF924, 0xA924, 0xABBC, 0xF924, 0xA924, 0xA93C, 0xF928, 0x2A8, 0xAACA, 0xAC4A, 0x886}},
    {19590, 2, {0x2020, 0x2020, 0x7C78, 0x4448, 0x8890, 0x7E7C, 0x5254, 0x5254, 0x7E7C, 0x5254, 0x5254, 0x7E7C, 0x0, 0xAAAA, 0xAAAA, 0x0}},
    {19591, 2, {0x2010, 0x2210, 0x797C, 0x4910, 0x907C, 0x7C54, 0x5754, 0x557C, 0x7D10, 0x5538, 0x5554, 0x7D92, 0x110, 0xAA90, 0xAA7E, 0x0}},
    {19592, 2, {0x2020, 0x2010, 0x79FE, 0x4948, 0x9148, 0x7D48, 0x55FE, 0x5548, 0x7D48, 0x55EC, 0x555A, 0x7D68, 0x148, 0xAB48, 0xAA48, 0x48}},
    {19593, 2, {0x2140, 0x2144, 0x7158, 0x93F0, 0x2150, 0xF9D0, 0xA95E, 0xA9D4, 0xF954, 0xA954, 0xABF4, 0xF814, 0x154, 0xAA34, 0xAC14, 0x24}},
    {19594, 2, {0x2010, 0x2390, 0x72FE, 0x92A0, 0x22BE, 0xFB48, 0xAABE, 0xAA80, 0xFABE, 0xAAA2, 0xAABE, 0xFB22, 0x23E, 0xAA22, 0xAA22, 0x226}},
    {19595, 2, {0x2088, 0x2088, 0x79DC, 0x4888, 0x93DE, 0x7C88, 0x5554, 0x5622, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {19596, 2, {0x2108, 0x2108, 0x77C8, 0x9108, 0x211E, 0xFB92, 0xA824, 0xAB80, 0xF808, 0xAFC8, 0xA908, 0xF908, 0x594, 0xA954, 0xA924, 0x342}},
    {19597, 2, {0x1124, 0x20A8, 0x79FC, 0x4840, 0x6BFE, 0x5908, 0x4A84, 0xFCFA, 0x4910, 0x4BFC, 0x6D24, 0x59FC, 0x4924, 0x49FC, 0x4800, 0x9954}},
    {19598, 2, {0x2008, 0x2388, 0x7088, 0x9110, 0x27DE, 0xFA94, 0xAAA4, 0xAB94, 0xFA94, 0xAB94, 0xAA94, 0xFAC8, 0x388, 0xAE94, 0xA8A4, 0xC2}},
    {19599, 2, {0x2048, 0x216A, 0x78DC, 0x4848, 0x90B4, 0x7D22, 0x55FE, 0x5502, 0x7C20, 0x5420, 0x54FC, 0x7C24, 0x44, 0xAA44, 0xAA94, 0x108}},
    {19600, 2, {0x2020, 0x203E, 0x7820, 0x49FE, 0x9122, 0x7DFC, 0x5522, 0x557E, 0x7D54, 0x557C, 0x5554, 0x7D7C, 0x110, 0xAAFE, 0xAA22, 0x4C6}},
    {19601, 2, {0x2104, 0x2088, 0x7BFE, 0x4820, 0x91FC, 0x7C20, 0x57FE, 0x5454, 0x7D92, 0x5490, 0x57FE, 0x7C90, 0xD4, 0xAB8A, 0xAA96, 0x1A2}},
    {19602, 2, {0x2100, 0x21DE, 0x724A, 0x948A, 0x23EA, 0xFAB6, 0xAAA4, 0xABF4, 0xFABE, 0xAAA4, 0xABE4, 0xFABE, 0x2A4, 0xAAA4, 0xAA24, 0x464}},
    {19603, 2, {0x2020, 0x2020, 0x7850, 0x4888, 0x9104, 0x7EFA, 0x5400, 0x5400, 0x7DDC, 0x5554, 0x5554, 0x7DDC, 0x88, 0xAA88, 0xAB54, 0x222}},
    {19604, 2, {0x2000, 0x23FE, 0x7000, 0x91FC, 0x2124, 0xF9FC, 0xA924, 0xABFE, 0xF800, 0xA9FC, 0xA924, 0xF9FC, 0x124, 0xA9FC, 0xA800, 0x3FE}},
    {19605, 2, {0x2084, 0x2048, 0x79FE, 0x4848, 0x914A, 0x7CCC, 0x5448, 0x55FE, 0x7C00, 0x54FC, 0x5484, 0x7C84, 0xFC, 0xAA84, 0xAA84, 0xFC}},
    {19606, 2, {0x2020, 0x23FE, 0x7020, 0x91FC, 0x2000, 0xFBFE, 0xA802, 0xA9FC, 0xF820, 0xABFE, 0xA800, 0xFBFE, 0x4, 0xABBE, 0xAAA4, 0x38C}},
    {19607, 2, {0x2120, 0x2120, 0x7256, 0x9248, 0x2580, 0xF970, 0xAA2E, 0xAA24, 0xFEF4, 0xAA24, 0xAAAC, 0xFA74, 0x224, 0xAA34, 0xAAC4, 0x20C}},
    {19608, 2, {0x2108, 0x2108, 0x79EE, 0x4A94, 0x9042, 0x7C10, 0x55FE, 0x5502, 0x7CF8, 0x5488, 0x54F8, 0x7C80, 0xFC, 0xAA84, 0xAAFC, 0x84}},
    {19609, 2, {0x2108, 0x2108, 0x71EE, 0x9294, 0x2042, 0xF800, 0xABDE, 0xAA52, 0xFBD2, 0xAA52, 0xABD2, 0xFA1A, 0x294, 0xAB50, 0xAA10, 0x10}},
    {19610, 2, {0x2088, 0x2088, 0x73EE, 0x9092, 0x2084, 0xFBEE, 0xAAAA, 0xAAAE, 0xFBEA, 0xA88E, 0xA9CA, 0xFAAE, 0x4A0, 0xA88C, 0xA88A, 0x92}},
    {19611, 2, {0x2088, 0x23FE, 0x7888, 0x49FC, 0x9154, 0x7DFC, 0x5400, 0x57FE, 0x7E02, 0x54F8, 0x5488, 0x7CF8, 0x88, 0xAAF8, 0xAA88, 0xF8}},
    {19612, 2, {0x3C3C, 0x4848, 0xFCFC, 0x5454, 0x7C7C, 0x5454, 0x7C7C, 0xAAAA, 0x3C3C, 0x4848, 0xFCFC, 0x5454, 0x7C7C, 0x5454, 0x7C7C, 0xAAAA}},
    {19613, 2, {0x2020, 0x2020, 0x7C50, 0x4450, 0x8888, 0x7D04, 0x5602, 0x54F8, 0x7C88, 0x5488, 0x54A8, 0x7C90, 0x82, 0x1C82, 0xE07E, 0x4000}},
    {19614, 2, {0x2040, 0x2020, 0x7C20, 0x4400, 0x89FE, 0x7C00, 0x5404, 0x5484, 0x7C84, 0x5448, 0x5448, 0x7C48, 0x50, 0x1C10, 0xE3FE, 0x4000}},
    {19615, 2, {0x2000, 0x2040, 0x7D9E, 0x4512, 0x8912, 0x7D12, 0x55D2, 0x5512, 0x7D12, 0x5512, 0x555A, 0x7D94, 0x110, 0x1C10, 0xE010, 0x4010}},
    {19616, 2, {0x2020, 0x2020, 0x7DFE, 0x4420, 0x88FC, 0x7C40, 0x55FE, 0x5488, 0x7D04, 0x56FA, 0x5488, 0x7C88, 0xF8, 0x1C88, 0xE088, 0x40F8}},
    {19617, 2, {0x2088, 0x2048, 0x7C50, 0x45FE, 0x8850, 0x7C50, 0x55FC, 0x5554, 0x7D54, 0x558C, 0x5504, 0x7DFC, 0x104, 0x1D04, 0xE1FC, 0x4104}},
    {19618, 2, {0x124, 0x3CA8, 0x25FC, 0x2440, 0x27FE, 0x3D08, 0x2684, 0x24FA, 0x2510, 0x3FFC, 0x2524, 0x25FC, 0x2524, 0x45FC, 0x5400, 0x8BFE}},
    {19619, 2, {0xF10, 0x810, 0x7F90, 0x48A0, 0x4E3E, 0x7944, 0x4F24, 0x5124, 0x5F24, 0x7528, 0x5F28, 0x5510, 0x5F28, 0x4028, 0xBF44, 0x82}},
    {19620, 2, {0x4000, 0x44FE, 0x7A28, 0x8828, 0x14FE, 0xFAAA, 0xAAAA, 0xA8AA, 0xF9AE, 0xAAC2, 0xAE82, 0xFAFE, 0x282, 0x1A82, 0xE2FE, 0x4082}},
    {19621, 2, {0x200, 0x3FE0, 0x422, 0x1822, 0xE21E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {19622, 2, {0x1010, 0x2010, 0x7F10, 0x4110, 0x7F7E, 0x4110, 0x7F10, 0x4010, 0x7F90, 0x4028, 0x7FA8, 0xA8, 0xAAA8, 0xAAC4, 0x8144, 0x682}},
    {19623, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1100, 0x11FE, 0x1100, 0x1DFE, 0xE002, 0x42AA, 0x2AA, 0x202, 0xC}},
    {19624, 2, {0x40, 0x80, 0x7DFC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1100, 0x11FE, 0x1100, 0x11FE, 0x1C02, 0xE2AA, 0x42AA, 0x202, 0xC}},
    {19625, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2200, 0x27F0, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x2400, 0x27FC, 0x2400, 0x47FC, 0x4004, 0x8AA4, 0xAAC}},
    {19626, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1100, 0x11FE, 0x2900, 0x25FE, 0x2402, 0x42AA, 0x42AA, 0x8202, 0xC}},
    {19627, 2, {0x80, 0xFDF8, 0x2908, 0x29F8, 0x2908, 0x29F8, 0x2900, 0x29FC, 0x2900, 0x29FC, 0x2804, 0x2954, 0x2A5C, 0x4802, 0x47FE, 0x8000}},
    {19628, 2, {0x1004, 0x201E, 0x7CF0, 0x4490, 0x7C90, 0x4490, 0x7C90, 0x40FE, 0x7E90, 0x4090, 0x7E90, 0x288, 0xAA8A, 0xAAAA, 0x82C6, 0xC82}},
    {19629, 2, {0x420, 0xE40, 0xF0FC, 0x9084, 0x90FC, 0x9084, 0xFCFC, 0x9080, 0x90FE, 0x9080, 0x90FE, 0x8802, 0x8AAA, 0xC6AA, 0x8302, 0xC}},
    {19630, 2, {0x20, 0x7C40, 0xFC, 0x84, 0xFC, 0xFE84, 0x28FC, 0x2880, 0x28FE, 0x2880, 0x28FE, 0x2A02, 0x4EAA, 0x4AAA, 0x8202, 0xC}},
    {19631, 2, {0x20, 0x7E40, 0x22FC, 0x2484, 0x24FC, 0x2884, 0x2EFC, 0x2280, 0x2AFE, 0x2480, 0x44FE, 0x4A02, 0x92AA, 0x20AA, 0x4102, 0xC}},
    {19632, 2, {0x40, 0x80, 0x79FC, 0x104, 0x1FC, 0x104, 0xFDFC, 0x2100, 0x21FE, 0x2100, 0x41FE, 0x5002, 0x8AAA, 0xFAAA, 0xA02, 0xC}},
    {19633, 2, {0x1020, 0x2010, 0x7F00, 0x41FE, 0x7F20, 0x4120, 0x7F3C, 0x4024, 0x7FA4, 0x4024, 0x7FA4, 0xA4, 0xAAA4, 0xAAA4, 0x8154, 0x688}},
    {19634, 2, {0x20, 0x2840, 0x24FC, 0x2484, 0x42FC, 0x5284, 0x90FC, 0x1080, 0x20FE, 0x2080, 0x48FE, 0x4402, 0xFCAA, 0x44AA, 0x102, 0xC}},
    {19635, 2, {0x1020, 0x840, 0x8FC, 0xFF84, 0xFC, 0x84, 0x3CFC, 0x2480, 0x24FE, 0x2480, 0x24FE, 0x2502, 0x46AA, 0x44AA, 0x8102, 0xC}},
    {19636, 2, {0x1040, 0x1080, 0x11FC, 0x1504, 0x59FC, 0x5104, 0x51FC, 0x9100, 0x11FE, 0x1100, 0x11FE, 0x2802, 0x26AA, 0x42AA, 0x4202, 0x800C}},
    {19637, 2, {0x100, 0x1110, 0x21E8, 0x4F04, 0xF200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {19638, 2, {0x1FF0, 0x0, 0xFFFE, 0x844, 0x323C, 0xDFF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {19639, 2, {0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4924, 0x8494, 0x108, 0xFFFE, 0x920, 0x711C}},
    {19640, 2, {0x1040, 0x1080, 0x11FC, 0x2904, 0x25FC, 0x4304, 0xA9FC, 0x2900, 0x29FE, 0x2900, 0x29FE, 0x2802, 0x4AAA, 0x4AAA, 0x8A02, 0xC}},
    {19641, 2, {0x40, 0x7E80, 0x9FC, 0x904, 0x11FC, 0x1504, 0x33FC, 0x5100, 0x91FE, 0x1100, 0x11FE, 0x2, 0xEAA, 0xF2AA, 0x4202, 0xC}},
    {19642, 2, {0x4840, 0x4880, 0x49FC, 0x4904, 0xFDFC, 0x4904, 0x49FC, 0x4900, 0x79FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x7AAA, 0x4A02, 0xC}},
    {19643, 2, {0x2040, 0x1080, 0x1FC, 0xFD04, 0x21FC, 0x4104, 0x49FC, 0x8900, 0xF1FE, 0x1100, 0x21FE, 0x2002, 0x4AAA, 0xFAAA, 0xA02, 0xC}},
    {19644, 2, {0x1040, 0x1080, 0x51FC, 0x5104, 0x7DFC, 0x5104, 0x91FC, 0x1100, 0x11FE, 0x7D00, 0x11FE, 0x1002, 0x12AA, 0x1EAA, 0xE202, 0x400C}},
    {19645, 2, {0x40, 0x80, 0xF9FC, 0x2104, 0x21FC, 0x4104, 0x79FC, 0x4900, 0xC9FE, 0x4900, 0x49FE, 0x4802, 0x7AAA, 0x4AAA, 0x202, 0xC}},
    {19646, 2, {0x0, 0x7DFE, 0x1080, 0x10FC, 0x1E04, 0xF028, 0x210, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {19647, 2, {0x20, 0x3E40, 0x22FC, 0x2284, 0x22FC, 0x3E84, 0x20FC, 0x2880, 0x28FE, 0x2E80, 0x28FE, 0x2802, 0x4AAA, 0x4CAA, 0x8902, 0xC}},
    {19648, 2, {0x1010, 0x2010, 0x7C50, 0x4450, 0x7C7C, 0x4490, 0x7D10, 0x4010, 0x7EFE, 0x4010, 0x7E28, 0x228, 0xAA28, 0xAA44, 0x8244, 0xC82}},
    {19649, 2, {0x1050, 0x2048, 0x7C48, 0x4440, 0x7DFE, 0x4440, 0x7C40, 0x407C, 0x7EA4, 0x40A4, 0x7EA8, 0x328, 0xAB10, 0xAA28, 0x8244, 0xC82}},
    {19650, 2, {0x20, 0xFC40, 0x24FC, 0x2484, 0x44FC, 0x5484, 0x88FC, 0x80, 0x7CFE, 0x4480, 0x44FE, 0x4402, 0x44AA, 0x7CAA, 0x4502, 0xC}},
    {19651, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C, 0x800, 0x1EF8, 0x2288, 0x5498, 0x882, 0x307E}},
    {19652, 2, {0x1420, 0x1440, 0x14FC, 0x5484, 0x55FC, 0x5E84, 0x54FC, 0x5480, 0x54FE, 0x5480, 0x54FE, 0x5402, 0x5DAA, 0x76AA, 0xC502, 0xC}},
    {19653, 2, {0x1000, 0x2000, 0x7F3E, 0x4122, 0x7F22, 0x4122, 0x7F22, 0x403E, 0x7FA2, 0x4000, 0x7F94, 0x94, 0xAAA2, 0xAAA2, 0x8142, 0x600}},
    {19654, 2, {0x1040, 0x1080, 0x21FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4900, 0x79FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x7AAA, 0x4A02, 0xC}},
    {19655, 2, {0x20, 0x3C40, 0x24FC, 0x3484, 0x2CFC, 0x2484, 0x24FC, 0xFE80, 0x44FE, 0x6480, 0x54FE, 0x4402, 0x7EAA, 0x4AA, 0x2902, 0x100C}},
    {19656, 2, {0x1040, 0x2040, 0x7CFC, 0x4484, 0x7D04, 0x46F4, 0x7C94, 0x4094, 0x7EF4, 0x4084, 0x7E94, 0x288, 0xAA82, 0xAA82, 0x827E, 0xC00}},
    {19657, 2, {0x2040, 0x2080, 0x3DFC, 0x4504, 0xA9FC, 0x1104, 0x29FC, 0x4500, 0x83FE, 0x3100, 0x9FE, 0x2, 0x62AA, 0x12AA, 0xA02, 0xC}},
    {19658, 2, {0x2820, 0x2440, 0x24FC, 0x2084, 0xFEFC, 0x2084, 0x20FC, 0x3E80, 0x42FE, 0x4480, 0x54FE, 0x8802, 0x16AA, 0x22AA, 0x4202, 0xC}},
    {19659, 2, {0x2040, 0x2080, 0x21FC, 0xF904, 0xA9FC, 0xA904, 0xA9FC, 0xA900, 0xF9FE, 0xA100, 0x21FE, 0x2802, 0x3AAA, 0xEAAA, 0x4202, 0xC}},
    {19660, 2, {0x1040, 0x1080, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x79FC, 0x4900, 0x49FE, 0x7900, 0x49FE, 0x7802, 0x4AAA, 0x4AAA, 0x4A02, 0x580C}},
    {19661, 2, {0x2410, 0x2420, 0x247C, 0x2444, 0x7F7C, 0x2444, 0x247C, 0x2440, 0x247E, 0xFF40, 0x7E, 0x2402, 0x24AA, 0x42AA, 0x4302, 0x820C}},
    {19662, 2, {0x40, 0x4080, 0x99FC, 0x8904, 0x89FC, 0x8904, 0x89FC, 0xD900, 0x89FE, 0x8900, 0x89FE, 0x8802, 0xFAAA, 0x8AAA, 0x202, 0xC}},
    {19663, 2, {0x1040, 0x1080, 0x11FC, 0x7D04, 0x11FC, 0x1104, 0xFDFC, 0x100, 0x11FE, 0x1100, 0x7DFE, 0x1002, 0x12AA, 0x1EAA, 0xE202, 0x400C}},
    {19664, 2, {0x4100, 0x21FC, 0xA44, 0x1048, 0xE0A0, 0x2118, 0x2606, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {19665, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x20FC, 0x3E84, 0x22FC, 0x6280, 0xBEFE, 0x2280, 0x22FE, 0x3E02, 0x22AA, 0x22AA, 0x2302, 0x260C}},
    {19666, 2, {0x820, 0x7F28, 0x824, 0xFFFE, 0x1020, 0x7F20, 0x4124, 0x7F24, 0x4124, 0x7FA8, 0x4028, 0x7F90, 0x92, 0xAAAA, 0xAAC6, 0x182}},
    {19667, 2, {0x1020, 0x1240, 0x7EFC, 0x1484, 0x14FC, 0xFF84, 0x8FC, 0x1080, 0x20FE, 0x7E80, 0xA0FE, 0x2202, 0x2CAA, 0x30AA, 0x2102, 0xC}},
    {19668, 2, {0x600, 0x38F8, 0x2008, 0x3CF8, 0x2008, 0x3FF8, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {19669, 2, {0xE20, 0xF040, 0x22FC, 0x9284, 0x54FC, 0x84, 0xFCFC, 0x880, 0x10FE, 0x1680, 0x18FE, 0xF002, 0x52AA, 0x12AA, 0x5202, 0x200C}},
    {19670, 2, {0x1000, 0x21FE, 0x7CA0, 0x44A8, 0x7CA8, 0x44E8, 0x7CA8, 0x40A8, 0x7EA8, 0x40E8, 0x7EA8, 0x2AA, 0xAABA, 0xABEA, 0x8226, 0xC20}},
    {19671, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C, 0x240, 0x3C50, 0x848, 0xFFFE, 0x850, 0x7E24, 0x854, 0x198C}},
    {19672, 2, {0x2010, 0x40D8, 0xF394, 0x9094, 0xF090, 0x93FE, 0xF090, 0x8094, 0xF894, 0x80D8, 0xFB98, 0x890, 0xA8AA, 0xA8CA, 0x8A86, 0x3102}},
    {19673, 2, {0x810, 0x820, 0x7F7C, 0x4944, 0x7F7C, 0x4944, 0x7F7C, 0x40, 0xFF7E, 0x2040, 0x407E, 0x7E02, 0x2AA, 0x2AA, 0x1502, 0x80C}},
    {19674, 2, {0x40, 0x7C80, 0x45FC, 0x7D04, 0x45FC, 0x7D04, 0x1FC, 0x100, 0x7DFE, 0x1100, 0x11FE, 0xFE02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {19675, 2, {0x20, 0x7E40, 0x42FC, 0x4284, 0x7EFC, 0x4284, 0x42FC, 0x7E80, 0x10FE, 0x1080, 0xFEFE, 0x2202, 0x64AA, 0x18AA, 0x2502, 0xC20C}},
    {19676, 2, {0x1020, 0x1040, 0x28FC, 0x4484, 0x82FC, 0x7C84, 0x10FC, 0x1080, 0xFEFE, 0x1080, 0x54FE, 0x5202, 0x92AA, 0x10AA, 0x5102, 0x200C}},
    {19677, 2, {0x2020, 0x1040, 0x7CFC, 0x84, 0x44FC, 0x2884, 0xFC, 0xFE80, 0xFE, 0x80, 0x7CFE, 0x4402, 0x44AA, 0x44AA, 0x7D02, 0x440C}},
    {19678, 2, {0x810, 0x820, 0x7F7C, 0x1044, 0x3E7C, 0x1044, 0xFFFC, 0x2240, 0x497E, 0x88C0, 0x3E7E, 0x802, 0x7FAA, 0x8AA, 0x902, 0x80C}},
    {19679, 2, {0x20, 0xE40, 0xEAFC, 0xAA84, 0xAAFC, 0xAE84, 0xEAFC, 0xAA80, 0xAAFE, 0xAE80, 0xEAFE, 0xAA02, 0xAAA, 0x12AA, 0x1302, 0x260C}},
    {19680, 2, {0x210, 0xF20, 0x787C, 0x844, 0xFF7C, 0x2A44, 0x2A7C, 0xFF40, 0x2A7E, 0x2A40, 0xFF7E, 0x802, 0x8AA, 0xEAA, 0x7102, 0x200C}},
    {19681, 2, {0x2820, 0x2440, 0x40FC, 0x7E84, 0xC8FC, 0x4884, 0x7EFC, 0x4880, 0x48FE, 0x7E80, 0x48FE, 0x4802, 0x48AA, 0x7EAA, 0x4102, 0x400C}},
    {19682, 2, {0x4840, 0x4880, 0x49FC, 0xFD04, 0x49FC, 0x4904, 0x79FC, 0x4900, 0x79FE, 0x4900, 0x49FE, 0xFC02, 0x2AA, 0x52AA, 0x4A02, 0x840C}},
    {19683, 2, {0x850, 0x7F48, 0x840, 0xFFFE, 0x2040, 0x3E50, 0x2222, 0x465A, 0x8286, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19684, 2, {0x1010, 0x9220, 0x527C, 0x5444, 0x107C, 0xFE44, 0x927C, 0x9240, 0xBA7E, 0xD640, 0x927E, 0x9202, 0x92AA, 0x82AA, 0x8B02, 0x840C}},
    {19685, 2, {0x1020, 0x9240, 0x92FC, 0xFE84, 0x10FC, 0x2884, 0x44FC, 0x9280, 0x8FE, 0xFC80, 0x4FE, 0x802, 0x52AA, 0x22AA, 0x1202, 0xC}},
    {19686, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0xFC, 0x7C84, 0xFC, 0x7C80, 0x44FE, 0x7C80, 0x44FE, 0x7C02, 0x2AA, 0x1EAA, 0xE202, 0x400C}},
    {19687, 2, {0x2100, 0x4178, 0xF108, 0x9208, 0xF2FE, 0x9640, 0xFA40, 0x827C, 0xFA90, 0x8210, 0xFAFE, 0xA10, 0xAA28, 0xAA28, 0x8A44, 0x3282}},
    {19688, 2, {0x1020, 0x2040, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x80, 0xFEFE, 0x1080, 0x10FE, 0x7C02, 0x12AA, 0x1EAA, 0xF202, 0x400C}},
    {19689, 2, {0x20, 0xFE40, 0x28FC, 0x2884, 0xFEFC, 0xAA84, 0xAAFC, 0xFE80, 0x10FE, 0xFC80, 0x24FE, 0x4402, 0x2AAA, 0x12AA, 0x2A02, 0xC40C}},
    {19690, 2, {0x1020, 0x9240, 0x92FC, 0x9284, 0xFEFC, 0x84, 0xFEFC, 0x1080, 0x20FE, 0xFE80, 0xAAFE, 0xAA02, 0xAAAA, 0xAAAA, 0xAB02, 0x860C}},
    {19691, 2, {0xA10, 0x7420, 0x157C, 0x5244, 0x227C, 0x2144, 0x417C, 0xBE40, 0x87E, 0x840, 0x7F7E, 0x802, 0x14AA, 0x12AA, 0x2302, 0x400C}},
    {19692, 2, {0x2840, 0x2880, 0xFDFC, 0x2904, 0x39FC, 0x1104, 0x7DFC, 0x5500, 0x55FE, 0x7D00, 0x11FE, 0xFC02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {19693, 2, {0x20, 0xEE40, 0xAAFC, 0xAA84, 0xAAFC, 0xEA84, 0xAAFC, 0xAA80, 0xEAFE, 0x8A80, 0xCAFE, 0xAE02, 0xD8AA, 0x88AA, 0x902, 0x80C}},
    {19694, 2, {0x20, 0x7C10, 0x45E0, 0x7C22, 0x45B4, 0x7CA8, 0x4524, 0x56A2, 0x8A40, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19695, 2, {0x1C20, 0x7040, 0x10FC, 0xFE84, 0x10FC, 0x7C84, 0x54FC, 0x7C80, 0x54FE, 0x7C80, 0x10FE, 0x7C02, 0x10AA, 0x1EAA, 0xF102, 0x400C}},
    {19696, 2, {0x1008, 0x1790, 0x14BC, 0x24A4, 0x24BC, 0x67A4, 0xA23C, 0x2220, 0x2FBE, 0x2720, 0x26BE, 0x2A82, 0x2AAA, 0x322A, 0x2202, 0x220C}},
    {19697, 2, {0x20, 0x7E7C, 0x14C4, 0x838, 0xFEC6, 0x2A20, 0x48FC, 0xA844, 0x128C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19698, 2, {0x11FE, 0x1100, 0xFD7C, 0x1100, 0x1DFE, 0xF154, 0x1248, 0x5464, 0x2242, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19699, 2, {0x1020, 0x5440, 0x54FC, 0x7C84, 0x92FC, 0x9284, 0xFEFC, 0x80, 0x7CFE, 0x80, 0xFEFE, 0x1002, 0x54AA, 0x92AA, 0x5102, 0x200C}},
    {19700, 2, {0x20, 0xFE40, 0xAAFC, 0xAA84, 0xAAFC, 0xFE84, 0x10FC, 0x5480, 0x52FE, 0x9280, 0x10FE, 0x5402, 0x52AA, 0x92AA, 0x1102, 0x100C}},
    {19701, 2, {0x1104, 0x1108, 0x111E, 0x7D12, 0x13DE, 0x5552, 0x555E, 0x5550, 0xBB5E, 0x1150, 0x395E, 0x5542, 0x9252, 0x1252, 0x154A, 0x1884}},
    {19702, 2, {0x100E, 0x20F0, 0x7C22, 0x4492, 0x7C44, 0x4420, 0x7C44, 0x40F8, 0x7E10, 0x4024, 0x7EFE, 0x210, 0xAAFE, 0xAA28, 0x8244, 0xC82}},
    {19703, 2, {0x1110, 0x2110, 0xCAA8, 0x1444, 0x3240, 0xD27C, 0x1540, 0x14FE, 0xA00, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19704, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2100, 0x27F0, 0x2410, 0x27F0, 0x2410, 0x27FC, 0x4004, 0x4AA4, 0x800C}},
    {19705, 2, {0x2040, 0x4020, 0xF3FE, 0x9202, 0xF008, 0x93EE, 0xF008, 0x801C, 0xFBE4, 0x8154, 0xF948, 0x954, 0xA964, 0xA942, 0x893E, 0x3200}},
    {19706, 2, {0x810, 0x1420, 0x227C, 0x4944, 0xBEFC, 0x444, 0x87C, 0xFFC0, 0x147E, 0x7F40, 0x557E, 0x6302, 0x412A, 0x7F2A, 0x4142, 0x7F0C}},
    {19707, 2, {0x804, 0x7F78, 0x840, 0x3E40, 0x2A7E, 0x3E48, 0x2A48, 0xFF48, 0xA88, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19708, 2, {0x1000, 0x20FE, 0x7C80, 0x4480, 0x7CBC, 0x44A4, 0x7CA4, 0x40BC, 0x7E80, 0x40EE, 0x7EAA, 0x2AA, 0xAAEE, 0xAA80, 0x82FE, 0xC00}},
    {19709, 2, {0x1020, 0x9240, 0x92FC, 0xFE84, 0x28FC, 0x2484, 0x7EFC, 0xC880, 0x48FE, 0x7E80, 0x48FE, 0x7E02, 0x48AA, 0x48AA, 0x7F02, 0x400C}},
    {19710, 2, {0xA10, 0x7420, 0x157C, 0x5244, 0x227C, 0x2144, 0x5D7C, 0x8040, 0x7E7E, 0x4240, 0x427E, 0x7E02, 0x42AA, 0x24AA, 0xF02, 0xF00C}},
    {19711, 2, {0x2410, 0x2620, 0x357C, 0x5444, 0x5F7C, 0x7444, 0x947C, 0x6440, 0x2A7E, 0x4940, 0x917E, 0x2, 0xAAAA, 0xAAAA, 0xAB02, 0xC}},
    {19712, 2, {0x1010, 0x820, 0x7F7C, 0x2244, 0xFF7C, 0x44, 0x7F7C, 0x4940, 0x7F7E, 0x4940, 0x7F7E, 0x802, 0x7F2A, 0x82A, 0xF42, 0xF00C}},
    {19713, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0x7CFC, 0x5484, 0x7CFC, 0x5480, 0x7CFE, 0x1080, 0xFEFE, 0x9202, 0x92AA, 0xBAAA, 0x8B02, 0x840C}},
    {19714, 2, {0x2010, 0x3E20, 0x427C, 0xFF44, 0x497C, 0x4944, 0x7F7C, 0x1040, 0x287E, 0xD540, 0x267E, 0xCD02, 0x15AA, 0x24AA, 0xD502, 0x80C}},
    {19715, 2, {0x1020, 0x1040, 0x7CFC, 0x1084, 0xFEFC, 0x4484, 0xEEFC, 0x4480, 0x64FE, 0xDE80, 0xFE, 0xFE02, 0x28AA, 0x2AAA, 0x4D02, 0x880C}},
    {19716, 2, {0xF10, 0xF020, 0x117C, 0x8944, 0x427C, 0x44, 0xFF7C, 0x8140, 0x427E, 0x7740, 0x927E, 0x5A02, 0x2F2A, 0x222A, 0x4242, 0x820C}},
    {19717, 2, {0x810, 0x7F20, 0x87C, 0x2244, 0xFFFC, 0x2244, 0x7F7C, 0x4140, 0x7F7E, 0x4140, 0x7F7E, 0x4102, 0x7F2A, 0x222A, 0x4142, 0x808C}},
    {19718, 2, {0x2820, 0xAA40, 0x6CFC, 0x2884, 0xFEFC, 0x4484, 0x28FC, 0xFE80, 0x10FE, 0x7C80, 0x10FE, 0xFE02, 0x10AA, 0x28AA, 0x4502, 0x820C}},
    {19719, 2, {0x10, 0x7F20, 0x87C, 0xFFC4, 0x88FC, 0x6B44, 0x87C, 0x6B40, 0x147E, 0x2A40, 0x457E, 0xBE02, 0x2AA, 0x4AA, 0x1902, 0x40C}},
    {19720, 2, {0x1020, 0x2040, 0xFEFC, 0x9284, 0xD6FC, 0xBA84, 0xFEFC, 0x9280, 0xBAFE, 0xD680, 0x82FE, 0x1002, 0xFEAA, 0x28AA, 0x4502, 0x820C}},
    {19721, 2, {0x20, 0xFE40, 0xAAFC, 0xAA84, 0xFEFC, 0x84, 0xFEFC, 0x80, 0x7CFE, 0x4480, 0x7CFE, 0x2A02, 0x64AA, 0xA2AA, 0x3102, 0x200C}},
    {19722, 2, {0x4108, 0x2210, 0xFFBC, 0x824, 0x7F3C, 0x824, 0xFFBC, 0x1520, 0x64BE, 0x2420, 0xFFBE, 0x2402, 0x352A, 0xE2AA, 0x2582, 0x688C}},
    {19723, 2, {0x2000, 0x43FC, 0xF294, 0x9294, 0xF3FC, 0x9000, 0xF7FE, 0x8000, 0xFBFC, 0x8204, 0xFBFC, 0x8A2, 0xA914, 0xAB08, 0x8D44, 0x3182}},
    {19724, 2, {0x1048, 0x2048, 0x7DFE, 0x4448, 0x7CFE, 0x4482, 0x7D7C, 0x4000, 0x7EFE, 0x4020, 0x7ED2, 0x22C, 0xAAD8, 0xAA2C, 0x82CA, 0xC30}},
    {19725, 2, {0x10, 0x20, 0x777C, 0x5544, 0x557C, 0x7744, 0x557C, 0x5540, 0x777E, 0x5540, 0x557E, 0x7702, 0x2A, 0x552A, 0x88C2, 0xC}},
    {19726, 2, {0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {19727, 2, {0x2208, 0x2210, 0xFFBC, 0x2224, 0x3E3C, 0x4124, 0x5D3C, 0xD5E0, 0x553E, 0x5520, 0xDD7E, 0x4182, 0x2A, 0x552A, 0x4A82, 0x8A8C}},
    {19728, 2, {0x808, 0xFF90, 0x84BC, 0x7724, 0x2A3C, 0x7724, 0x2ABC, 0xFFA0, 0x3E, 0x7F20, 0x413E, 0x7F02, 0x412A, 0x7F2A, 0x2202, 0x410C}},
    {19729, 2, {0x7F08, 0x4910, 0x7F3C, 0x4924, 0xFFBC, 0xAAA4, 0xFFBC, 0xAAA0, 0xFFBE, 0x20, 0xFFBE, 0xA282, 0x3E2A, 0x222A, 0x2782, 0xF80C}},
    {19730, 2, {0x7F08, 0x810, 0xFFBC, 0x88A4, 0x6B3C, 0x824, 0x6B3C, 0x20, 0xEFBE, 0xAAA0, 0xEFBE, 0x2, 0xFFAA, 0x492A, 0xAA82, 0xFF8C}},
    {19731, 2, {0x10, 0xFFA0, 0x87C, 0x844, 0x864, 0x7F54, 0x4944, 0x494C, 0x4940, 0x497E, 0x4902, 0x4D02, 0x4A7A, 0x802, 0x80A, 0x804}},
    {19732, 2, {0x2020, 0x1040, 0xFC, 0xFE84, 0xA4, 0x2884, 0x4494, 0x8288, 0x480, 0x44FE, 0x2802, 0x1002, 0x29FA, 0x4402, 0x8414, 0x8}},
    {19733, 2, {0x7F04, 0x844, 0x1E44, 0x2244, 0x5444, 0x844, 0x320C, 0xDFE0, 0x1420, 0x1260, 0x1000, 0x1FF8, 0x8, 0x7F88, 0x28, 0x10}},
    {19734, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CA4, 0x1084, 0xFE94, 0x88, 0x7C80, 0x44FE, 0x7C02, 0x4402, 0x7DFA, 0x4402, 0x5414, 0x4808}},
    {19735, 2, {0x20, 0x7E40, 0x42FC, 0x7E84, 0x42A4, 0x7E84, 0x4294, 0x7E88, 0x1480, 0x12FE, 0xFF02, 0x1002, 0x29FA, 0x2402, 0x4214, 0x8008}},
    {19736, 2, {0x310, 0x7C20, 0x477C, 0x4444, 0x5F64, 0x5554, 0x5644, 0x5C4C, 0x5540, 0x537E, 0x5002, 0x5702, 0x557A, 0x5502, 0xA98A, 0x4}},
    {19737, 2, {0x208, 0x7910, 0x493E, 0x4BA2, 0x4832, 0x7AAA, 0x4122, 0x47E6, 0x7920, 0x693E, 0xAB82, 0xA902, 0x297A, 0x3902, 0x290A, 0x104}},
    {19738, 2, {0x1040, 0x1020, 0x1E20, 0x10FC, 0x1000, 0xFE00, 0x9278, 0xC648, 0xAA48, 0xD648, 0xAA48, 0xC64A, 0x924A, 0xFE8A, 0x8286, 0x100}},
    {19739, 2, {0x1010, 0x1092, 0x1E52, 0x1054, 0x1010, 0xFEFE, 0x9282, 0xC682, 0xAAFE, 0xD682, 0xAA82, 0xC6FE, 0x9282, 0xFE82, 0x828A, 0x84}},
    {19740, 2, {0x1010, 0x1020, 0x1EFE, 0x1092, 0x1092, 0xFEFE, 0x9292, 0xC692, 0xAAFE, 0xD620, 0xAA24, 0xC65A, 0x925E, 0xFE90, 0x8292, 0x10E}},
    {19741, 2, {0x1000, 0x107C, 0x1E44, 0x107C, 0x1044, 0xFE7C, 0x9200, 0xC7FE, 0xAAA0, 0xD6EE, 0xAAAA, 0xC6EA, 0x92BA, 0xFFE4, 0x8224, 0x2A}},
    {19742, 2, {0x1000, 0x11DC, 0x1E44, 0x1154, 0x10CC, 0xFF54, 0x9250, 0xC648, 0xAAFE, 0xD790, 0xAAFE, 0xC690, 0x92FE, 0xFE90, 0x82FE, 0x80}},
    {19743, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2000, 0x23F0, 0x2210, 0x4290, 0x4452, 0x980E}},
    {19744, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2038, 0x27C0, 0x2040, 0x5FFE, 0x4110, 0x8E0E}},
    {19745, 2, {0x0, 0x7FFE, 0x240, 0x1248, 0xA50, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2848, 0x2F70, 0x4844, 0x4B44, 0x8C3C}},
    {19746, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2040, 0x2040, 0x2FFE, 0x4150, 0x4248, 0x8C46}},
    {19747, 2, {0x804, 0x7F1E, 0x54F0, 0x5410, 0x7F92, 0x5552, 0x5554, 0x7F10, 0x54FE, 0x5410, 0x5E10, 0x5410, 0x5410, 0x5C10, 0x5650, 0x8020}},
    {19748, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x27DE, 0x2240, 0x23FC, 0x2440, 0x4BF8, 0x4040, 0x9FFE}},
    {19749, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2400, 0x3FBC, 0x24A4, 0x44A4, 0x4ABC, 0x9124}},
    {19750, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2040, 0x3FFE, 0x2000, 0x27F8, 0x2000, 0x47F8, 0x4408, 0x87F8}},
    {19751, 2, {0x81C, 0x7F70, 0x5410, 0x54FE, 0x7F38, 0x5554, 0x5592, 0x7F10, 0x5420, 0x54FE, 0x5E24, 0x5444, 0x5428, 0x5C10, 0x5628, 0x80C4}},
    {19752, 2, {0x800, 0x7F7E, 0x5448, 0x5450, 0x7F5C, 0x5554, 0x5554, 0x7F5C, 0x5454, 0x5454, 0x5E5C, 0x5448, 0x545C, 0x5CAA, 0x568A, 0x8118}},
    {19753, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x3FFC, 0x2140, 0x2FF8, 0x2948, 0x2FF8, 0x2000, 0x5FFC, 0x4490, 0x8988}},
    {19754, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2800, 0x24F8, 0x2408, 0x3E78, 0x2208, 0x24FE, 0x2E54, 0x5538, 0x4454, 0x84B2}},
    {19755, 2, {0x800, 0x7F7C, 0x5410, 0x54FE, 0x7F92, 0x5554, 0x5510, 0x7F54, 0x5428, 0x5454, 0x5E8A, 0x547C, 0x5404, 0x5C08, 0x5630, 0x8008}},
    {19756, 2, {0x848, 0x844, 0xFF44, 0x2A40, 0x2A5E, 0x5DE0, 0x8840, 0x1C40, 0x2A40, 0x5040, 0x1E20, 0x2222, 0x5412, 0x80A, 0x1406, 0x6002}},
    {19757, 2, {0x808, 0x808, 0xFF08, 0x2A08, 0x2AFE, 0x5D08, 0x8818, 0x1C18, 0x2A28, 0x5028, 0x1E48, 0x2288, 0x5408, 0x808, 0x1428, 0x6010}},
    {19758, 2, {0x810, 0x810, 0xFF10, 0x2A10, 0x2A92, 0x5D92, 0x8892, 0x1C92, 0x2A92, 0x5092, 0x1E92, 0x2292, 0x5492, 0x8FE, 0x1402, 0x6000}},
    {19759, 2, {0x840, 0x840, 0xFF40, 0x2A7E, 0x2AAA, 0x5D2A, 0x882A, 0x1C4A, 0x2A4A, 0x5092, 0x1F12, 0x2222, 0x5422, 0x842, 0x1494, 0x6008}},
    {19760, 2, {0x848, 0x844, 0xFF44, 0x2A40, 0x2A5E, 0x5DE0, 0x8840, 0x1C44, 0x2A44, 0x5048, 0x1E30, 0x2222, 0x5452, 0x88A, 0x1506, 0x6002}},
    {19761, 2, {0x820, 0x810, 0xFF10, 0x2AFE, 0x2A82, 0x5D04, 0x8840, 0x1C48, 0x2A50, 0x5060, 0x1E40, 0x2242, 0x5442, 0x842, 0x143E, 0x6000}},
    {19762, 2, {0x810, 0x810, 0xFF10, 0x2AFE, 0x2A10, 0x5D10, 0x8810, 0x1C7C, 0x2A10, 0x5038, 0x1E38, 0x2254, 0x5454, 0x892, 0x1410, 0x6010}},
    {19763, 2, {0x820, 0x810, 0xFF10, 0x2AFE, 0x2A82, 0x5D04, 0x8800, 0x1C28, 0x2A28, 0x5028, 0x1E28, 0x2228, 0x5444, 0x844, 0x1444, 0x6082}},
    {19764, 2, {0x810, 0x810, 0xFF10, 0x2A10, 0x2A1E, 0x5D10, 0x8810, 0x1C10, 0x2AFE, 0x5082, 0x1E82, 0x2282, 0x5482, 0x882, 0x14FE, 0x6082}},
    {19765, 2, {0x1084, 0x1048, 0xFE00, 0x54FC, 0x5448, 0xBA48, 0x1048, 0x3848, 0x55FE, 0xA248, 0x3C48, 0x4448, 0xA848, 0x1088, 0x2888, 0xC108}},
    {19766, 2, {0x1040, 0x1040, 0xFEFC, 0x5488, 0x5550, 0xBA20, 0x10D8, 0x3B26, 0x54F8, 0xA220, 0x3CF8, 0x4420, 0xABFE, 0x1020, 0x2820, 0xC020}},
    {19767, 2, {0x1040, 0x1020, 0xFE20, 0x55FE, 0x5502, 0xBA04, 0x10F8, 0x3800, 0x5400, 0xA3FE, 0x3C50, 0x4450, 0xA890, 0x1092, 0x2912, 0xC20E}},
    {19768, 2, {0x80E, 0x8F0, 0xFF22, 0x2A92, 0x2A54, 0x5D00, 0x887C, 0x1C08, 0x2A10, 0x5010, 0x1EFE, 0x2210, 0x5410, 0x810, 0x1450, 0x6020}},
    {19769, 2, {0x800, 0x87C, 0xFF54, 0x2A54, 0x2A7C, 0x5D54, 0x8854, 0x1C7C, 0x2A10, 0x50FE, 0x1E10, 0x2238, 0x5454, 0x892, 0x1410, 0x6010}},
    {19770, 2, {0x820, 0x810, 0xFF7C, 0x2A00, 0x2A44, 0x5D28, 0x8800, 0x1CFE, 0x2A00, 0x5000, 0x1E7C, 0x2244, 0x5444, 0x844, 0x147C, 0x6044}},
    {19771, 2, {0x640, 0x387C, 0x894, 0x7E24, 0x2CD4, 0x4A08, 0x100, 0x7FFC, 0x1110, 0x2BA8, 0x4564, 0x1918, 0xE3E6, 0xE40, 0x180, 0xE40}},
    {19772, 2, {0x1048, 0x1044, 0xFE5E, 0x55E0, 0x5428, 0xBA12, 0x106A, 0x3996, 0x5448, 0xA25E, 0x3DE0, 0x4424, 0xA828, 0x1012, 0x286A, 0xC186}},
    {19773, 2, {0x1020, 0x1040, 0xFEFE, 0x5492, 0x5492, 0xBAFE, 0x1092, 0x38A2, 0x54FE, 0xA248, 0x3C88, 0x45FE, 0xA808, 0x1008, 0x2808, 0xC008}},
    {19774, 2, {0x1084, 0x1044, 0xFE48, 0x55FE, 0x5420, 0xBAFC, 0x1020, 0x39FE, 0x5440, 0xA280, 0x3CFE, 0x4510, 0xAA10, 0x1410, 0x29FE, 0xC000}},
    {19775, 2, {0x1000, 0x11FE, 0xFF02, 0x5478, 0x5400, 0xBAFC, 0x1040, 0x38A4, 0x5538, 0xA258, 0x3C94, 0x4534, 0xA852, 0x1090, 0x2950, 0xC020}},
    {19776, 2, {0x1010, 0x1092, 0xFE54, 0x5410, 0x54FE, 0xBA82, 0x1082, 0x38FE, 0x5482, 0xA2FE, 0x3C82, 0x4482, 0xA8FE, 0x1000, 0x2844, 0xC082}},
    {19777, 2, {0x1048, 0x1148, 0xFF5E, 0x556A, 0x5554, 0xBBC8, 0x1050, 0x3864, 0x55C4, 0xA37E, 0x3D44, 0x4564, 0xA954, 0x1144, 0x2954, 0xC248}},
    {19778, 2, {0x1020, 0x1010, 0xFEFE, 0x5444, 0x5428, 0xBAFE, 0x1082, 0x3892, 0x54FE, 0xA292, 0x3CBA, 0x44AA, 0xA8AA, 0x10BA, 0x2882, 0xC086}},
    {19779, 2, {0x828, 0x828, 0xFF7C, 0x2A28, 0x2A28, 0x5DFE, 0x8810, 0x1C7C, 0x2A54, 0x507C, 0x1E54, 0x227C, 0x5400, 0x828, 0x1444, 0x6082}},
    {19780, 2, {0x1050, 0x11FC, 0xFF54, 0x55FC, 0x5554, 0xBBFC, 0x1000, 0x39FC, 0x5400, 0xA3FC, 0x3D04, 0x45FC, 0xA888, 0x1050, 0x2BFE, 0xC000}},
    {19781, 2, {0x820, 0x7F20, 0x83E, 0x3E44, 0x8A8, 0x7F10, 0x1028, 0x1E46, 0x2280, 0xFFFE, 0x11C8, 0x2AB4, 0xDD0, 0x764E, 0x180, 0x1E40}},
    {19782, 2, {0x1048, 0x1048, 0xFEFE, 0x5448, 0x54FE, 0xBA82, 0x117C, 0x3800, 0x54FE, 0xA220, 0x3CD2, 0x442C, 0xA8D8, 0x102C, 0x28CA, 0xC030}},
    {19783, 2, {0x810, 0x410, 0x7FD0, 0x4910, 0x493E, 0x4924, 0x7FD4, 0x4914, 0x5914, 0x5D94, 0x6B54, 0x4D08, 0x4908, 0x4914, 0x8914, 0x922}},
    {19784, 2, {0x1110, 0x23E0, 0x4048, 0xBFC, 0x1100, 0x33F0, 0x5510, 0x10E0, 0x1110, 0x1288, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x4D54, 0x9412}},
    {19785, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3492, 0x2140, 0x2630, 0x3BEE, 0x2000, 0x2FA4, 0x28A4, 0x4FA4, 0x48A4, 0x898C}},
    {19786, 2, {0x88, 0x7888, 0x49FC, 0x4888, 0x6888, 0x5BFE, 0x4820, 0xFDFC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x4800, 0x4888, 0x4904, 0x9A02}},
    {19787, 2, {0x2820, 0x2820, 0x7DFE, 0x2840, 0x2840, 0xFEFC, 0x1084, 0x7D84, 0x56FC, 0x5484, 0x7C84, 0x54FC, 0x5484, 0x7C84, 0x2894, 0x4488}},
    {19788, 2, {0x2820, 0x2820, 0x7C20, 0x29FC, 0x2820, 0xFEA8, 0x10A8, 0x7CA8, 0x5574, 0x5622, 0x7C50, 0x5450, 0x5488, 0x7C88, 0x2904, 0x4602}},
    {19789, 2, {0x2820, 0x2810, 0x7DFE, 0x2800, 0x28FC, 0xFE84, 0x10FC, 0x7C00, 0x54FC, 0x5408, 0x7C10, 0x55FE, 0x5410, 0x7C10, 0x2850, 0x4420}},
    {19790, 2, {0x2820, 0x2924, 0x7D24, 0x2924, 0x29FC, 0xFE00, 0x13FE, 0x7C20, 0x5440, 0x55FC, 0x7D54, 0x5554, 0x5554, 0x7D54, 0x2954, 0x450C}},
    {19791, 2, {0x2800, 0x29DC, 0x7C44, 0x2954, 0x28CC, 0xFF54, 0x1028, 0x7C44, 0x5592, 0x5420, 0x7CC8, 0x5432, 0x54C4, 0x7C18, 0x2860, 0x4580}},
    {19792, 2, {0x2800, 0x29DC, 0x7D54, 0x2954, 0x29DC, 0xFE00, 0x11FC, 0x7D24, 0x55FC, 0x5524, 0x7DFC, 0x5420, 0x57FE, 0x7C20, 0x2820, 0x4420}},
    {19793, 2, {0xC00, 0x7000, 0x11FC, 0xFE24, 0x3824, 0x54A4, 0x90A4, 0x28A4, 0x5524, 0x9024, 0x5424, 0x3844, 0x5444, 0x9284, 0x5114, 0x2208}},
    {19794, 2, {0xC00, 0x70FC, 0x1084, 0xFE84, 0x3884, 0x5484, 0x9084, 0x28FC, 0x5484, 0x9084, 0x5484, 0x3884, 0x5484, 0x9284, 0x50FC, 0x2084}},
    {19795, 2, {0xC10, 0x7010, 0x1090, 0xFE90, 0x38FE, 0x5490, 0x9110, 0x2810, 0x5410, 0x91FE, 0x5410, 0x3810, 0x5410, 0x9210, 0x5010, 0x2010}},
    {19796, 2, {0xC20, 0x7020, 0x13FE, 0xFE20, 0x39FC, 0x5524, 0x91FC, 0x2924, 0x55FC, 0x9020, 0x5470, 0x38A8, 0x5524, 0x9622, 0x5020, 0x2020}},
    {19797, 2, {0xC00, 0x71FC, 0x1104, 0xFF04, 0x39FC, 0x5510, 0x9110, 0x29FE, 0x5510, 0x9110, 0x557C, 0x3944, 0x5544, 0x9344, 0x517C, 0x2244}},
    {19798, 2, {0x7E7C, 0x4804, 0x7E28, 0x4210, 0x7E28, 0x48FC, 0x7F00, 0x100, 0x7FFC, 0x910, 0x3288, 0xCC64, 0x3938, 0xC546, 0x920, 0x3310}},
    {19799, 2, {0xC00, 0x71FC, 0x1000, 0xFEF8, 0x3888, 0x5488, 0x90F8, 0x2800, 0x55FC, 0x9124, 0x5524, 0x39FC, 0x5524, 0x9324, 0x51FC, 0x2104}},
    {19800, 2, {0xC00, 0x70FC, 0x1084, 0xFEFC, 0x3884, 0x54FC, 0x9000, 0x29FE, 0x5480, 0x90FE, 0x552A, 0x3A4A, 0x5492, 0x9322, 0x5054, 0x2088}},
    {19801, 2, {0xC00, 0x7020, 0x10FE, 0xFE44, 0x3882, 0x5520, 0x907C, 0x2888, 0x5450, 0x9068, 0x559E, 0x3822, 0x54D4, 0x9208, 0x5030, 0x21C0}},
    {19802, 2, {0xC1E, 0x71E0, 0x1022, 0xFF12, 0x3894, 0x5480, 0x9020, 0x29CE, 0x5502, 0x9102, 0x55CE, 0x3902, 0x5502, 0x9302, 0x51FE, 0x2102}},
    {19803, 2, {0xC00, 0x71EE, 0x114A, 0xFFEA, 0x392A, 0x55E4, 0x9144, 0x294A, 0x55F0, 0x9000, 0x5410, 0x38FE, 0x5410, 0x9210, 0x51FE, 0x2000}},
    {19804, 2, {0xC50, 0x71FC, 0x1154, 0xFFFC, 0x3954, 0x55FC, 0x9000, 0x29FE, 0x5500, 0x917C, 0x5500, 0x39FE, 0x5550, 0x9354, 0x5148, 0x2266}},
    {19805, 2, {0x20, 0x7F20, 0x4920, 0x6B20, 0x5D20, 0x4920, 0x7F20, 0x820, 0x7F20, 0x820, 0xF20, 0xF022, 0x22, 0x5522, 0x4A9E, 0x8000}},
    {19806, 2, {0x4, 0x7F04, 0x4904, 0x6B24, 0x5D24, 0x4924, 0x7F24, 0x824, 0x7F24, 0x824, 0xF24, 0xF024, 0x4, 0x5504, 0x4A94, 0x8008}},
    {19807, 2, {0x0, 0xFE7C, 0x9210, 0xD610, 0xBA10, 0x9210, 0xFE10, 0x10FE, 0xFE10, 0x1010, 0x1E10, 0xE010, 0x210, 0xAA10, 0xA810, 0x8010}},
    {19808, 2, {0x40, 0xFE40, 0x9240, 0xD6FC, 0xBA84, 0x9304, 0xFE04, 0x1084, 0xFE44, 0x1044, 0x1E44, 0xE004, 0x204, 0xAA04, 0xA828, 0x8010}},
    {19809, 2, {0x0, 0x3FF8, 0x100, 0x7FFC, 0x440, 0x1830, 0xE00E, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE, 0x4444}},
    {19810, 2, {0x10, 0xFE10, 0x9210, 0xD67C, 0xBA10, 0x9210, 0xFE10, 0x10FE, 0xFE10, 0x1038, 0x1E38, 0xE054, 0x254, 0xAA92, 0xA810, 0x8010}},
    {19811, 2, {0x0, 0xFE00, 0x92FC, 0xD684, 0xBA84, 0x9284, 0xFEFC, 0x1084, 0xFE84, 0x1084, 0x1EFC, 0xE084, 0x200, 0xAA00, 0xA9FE, 0x8000}},
    {19812, 2, {0x3FF8, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x7FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE, 0x4444}},
    {19813, 2, {0x40, 0xFE40, 0x927C, 0xD684, 0xBB08, 0x92FC, 0xFEA4, 0x10A4, 0xFEA4, 0x10FC, 0x1E80, 0xE080, 0x282, 0xAA82, 0xA87E, 0x8000}},
    {19814, 2, {0x0, 0xFE9C, 0x9248, 0xD608, 0xBA88, 0x9248, 0xFE5E, 0x1008, 0xFE28, 0x1048, 0x1EC8, 0xE048, 0x248, 0xAA48, 0xA848, 0x8018}},
    {19815, 2, {0x820, 0x7F28, 0x824, 0xFFFE, 0x20, 0x7F20, 0x4924, 0x6B24, 0x4924, 0x7F28, 0x828, 0x7F10, 0x812, 0xFFAA, 0x46, 0xAA82}},
    {19816, 2, {0x48, 0xFE44, 0x9240, 0xD6FE, 0xBA50, 0x9254, 0xFE58, 0x1052, 0xFE54, 0x1058, 0x1E52, 0xE054, 0x29A, 0xAA92, 0xA92E, 0x8240}},
    {19817, 2, {0x304, 0x3C24, 0x424, 0x7FA4, 0x1624, 0x2504, 0x440C, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE, 0x4444}},
    {19818, 2, {0x0, 0xFEFC, 0x9284, 0xD684, 0xBAFC, 0x9284, 0xFE84, 0x10FC, 0xFE00, 0x1090, 0x1E92, 0xE0F4, 0x298, 0xAA92, 0xA8D2, 0x808E}},
    {19819, 2, {0x100, 0x7FFC, 0x5004, 0x3EF8, 0x4288, 0xA2AA, 0x1492, 0x187E, 0xE000, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0xFFFE, 0x2448}},
    {19820, 2, {0x10, 0xFE10, 0x92DA, 0xD654, 0xBA54, 0x9292, 0xFF50, 0x1020, 0xFEFE, 0x1082, 0x1E82, 0xE0FE, 0x282, 0xAA82, 0xA8FE, 0x8082}},
    {19821, 2, {0x20, 0xFE20, 0x92FA, 0xD624, 0xBA28, 0x93FE, 0xFE20, 0x1040, 0xFEFC, 0x1144, 0x1E44, 0xE07C, 0x244, 0xAA44, 0xA87C, 0x8044}},
    {19822, 2, {0x0, 0xFE7C, 0x9244, 0xD67C, 0xBA44, 0x927C, 0xFE00, 0x10FE, 0xFE40, 0x107E, 0x1EAA, 0xE12A, 0x24A, 0xAA92, 0xA82A, 0x8044}},
    {19823, 2, {0x8, 0xFE1C, 0x92E0, 0xD620, 0xBBFE, 0x9220, 0xFEFC, 0x10A4, 0xFEFC, 0x10A4, 0x1EFC, 0xE020, 0x2FC, 0xAA20, 0xA9FE, 0x8000}},
    {19824, 2, {0x40, 0xFE7E, 0x9280, 0xD77C, 0xBA44, 0x927C, 0xFE44, 0x107C, 0xFE00, 0x10FE, 0x1E40, 0xE0FE, 0x32A, 0xAA4A, 0xA892, 0x8026}},
    {19825, 2, {0x10, 0xFE10, 0x92FE, 0xD610, 0xBA7C, 0x9200, 0xFE7C, 0x1044, 0xFE7C, 0x1028, 0x1EFE, 0xE000, 0x27C, 0xAA44, 0xA87C, 0x8044}},
    {19826, 2, {0x7C, 0xFE44, 0x9244, 0xD67C, 0xBA00, 0x92EE, 0xFEAA, 0x10AA, 0xFEEE, 0x1010, 0x1EFE, 0xE038, 0x254, 0xAA92, 0xA910, 0x8010}},
    {19827, 2, {0x10, 0xFE28, 0x9244, 0xD6BA, 0xBB00, 0x92FE, 0xFE92, 0x10D6, 0xFE92, 0x10FE, 0x1E00, 0xE07C, 0x244, 0xAA7C, 0xA844, 0x807C}},
    {19828, 2, {0x0, 0xFEFE, 0x92AA, 0xD6AA, 0xBAEE, 0x9228, 0xFEEE, 0x10AA, 0xFEAA, 0x10EE, 0x1EAA, 0xE0AA, 0x2EE, 0xAAA8, 0xA82A, 0x801E}},
    {19829, 2, {0x2040, 0x3F7E, 0x4890, 0x9FF8, 0x1490, 0x1250, 0x1130, 0xFFFE, 0x820, 0x3FF8, 0xD556, 0x1390, 0x1FF0, 0x100, 0x7FFC, 0x2448}},
    {19830, 2, {0x2000, 0x21FC, 0x2154, 0x7D54, 0x45DC, 0x8450, 0x75DC, 0x5554, 0x5554, 0x55DC, 0x7554, 0x5554, 0x5DC, 0x550, 0x2852, 0x103E}},
    {19831, 2, {0x1000, 0x11FC, 0x1154, 0x7D54, 0x11DC, 0x1050, 0xFDDC, 0x154, 0x1154, 0x11DC, 0x7D54, 0x1154, 0x11DC, 0x1D50, 0xE052, 0x403E}},
    {19832, 2, {0x7C10, 0x1052, 0xFE94, 0x3828, 0x5444, 0x9FF2, 0x1290, 0x1EF0, 0x280, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EFA, 0x282, 0x1FE}},
    {19833, 2, {0x201C, 0x7DC8, 0x917E, 0xFDC8, 0x2818, 0x5FF0, 0x9290, 0x1EF0, 0x280, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EFA, 0x282, 0x1FE}},
    {19834, 2, {0x21F8, 0x2108, 0x21F8, 0x250A, 0xF5FA, 0x250A, 0x25FA, 0x2402, 0x279E, 0x2090, 0x3090, 0xEF9E, 0x4492, 0x492, 0x492, 0x892}},
    {19835, 2, {0x1C44, 0x1444, 0x1CEE, 0x5544, 0x5DEE, 0x5544, 0x5DEE, 0x4144, 0x7744, 0x1400, 0x14FE, 0xF702, 0x557E, 0x5502, 0x55FE, 0x9502}},
    {19836, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8FF8, 0x4824, 0x4FE4, 0x4824, 0x7FFC, 0x440, 0xFC7C, 0x2444}},
    {19837, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E28, 0x2418, 0xFEE6, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x400, 0x820, 0x1FF0, 0x810}},
    {19838, 2, {0x810, 0x17FE, 0x3210, 0xD150, 0x1020, 0x810, 0xFF10, 0x8FE, 0x7E10, 0xFC, 0x7E44, 0x4248, 0x7E28, 0x2410, 0x1E28, 0xE0C6}},
    {19839, 2, {0x810, 0xFF10, 0x8FC, 0x7E10, 0xFC, 0x7E44, 0x4228, 0x7E10, 0x2428, 0x1EC6, 0xE100, 0x1FC, 0x100, 0x3FF8, 0x2008, 0x3FF8}},
    {19840, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4204, 0x7E48, 0x2438, 0x1E44, 0xE100, 0x6C0, 0x1830, 0xE7CE, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {19841, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E38, 0x2444, 0x7E82, 0x118, 0x7D60, 0x918, 0x3FF6, 0xD010, 0x1FF0, 0x1010, 0x1FF0}},
    {19842, 2, {0x2020, 0xCE20, 0x8220, 0xEE7C, 0x8244, 0x8284, 0xFE04, 0x84, 0x9244, 0x9224, 0xDA24, 0x9204, 0xDA04, 0x9204, 0x9328, 0xD910}},
    {19843, 2, {0x2010, 0xCE10, 0x8210, 0xEE28, 0x8228, 0x8244, 0xFE44, 0xA2, 0x9210, 0x9210, 0xDA7C, 0x9204, 0xDA08, 0x9208, 0x9310, 0xD910}},
    {19844, 2, {0x2000, 0xCEFC, 0x8210, 0xEE10, 0x8294, 0x8254, 0xFE58, 0x10, 0x93FE, 0x9210, 0xDA10, 0x9210, 0xDA10, 0x9210, 0x9310, 0xD910}},
    {19845, 2, {0x2020, 0xCE20, 0x823C, 0xEE44, 0x82A8, 0x8210, 0xFE28, 0x44, 0x9282, 0x927C, 0xDA44, 0x9244, 0xDA44, 0x9244, 0x937C, 0xD944}},
    {19846, 2, {0x2028, 0xCE24, 0x8240, 0xEE7E, 0x82C8, 0x8348, 0xFE7E, 0x48, 0x9248, 0x927E, 0xDA48, 0x9248, 0xDA48, 0x927E, 0x9340, 0xD940}},
    {19847, 2, {0x2010, 0xCE10, 0x82FE, 0xEE92, 0x82FE, 0x8292, 0xFEFE, 0x0, 0x927C, 0x9244, 0xDA7C, 0x9244, 0xDA7C, 0x9244, 0x9344, 0xD94C}},
    {19848, 2, {0x2014, 0xCE12, 0x83FE, 0xEE10, 0x82FE, 0x8292, 0xFEFE, 0x92, 0x92FE, 0x9292, 0xDA04, 0x93FE, 0xDA44, 0x9224, 0x9324, 0xD90C}},
    {19849, 2, {0x2020, 0xCEDE, 0x828A, 0xEECA, 0x82AA, 0x82D2, 0xFEA6, 0x0, 0x92FE, 0x9292, 0xDA92, 0x92FE, 0xDA92, 0x9292, 0x93FE, 0xD982}},
    {19850, 2, {0x2000, 0x7CFC, 0x4424, 0x7C24, 0x4424, 0x7C24, 0x24, 0xFEFC, 0x9244, 0xFE44, 0x9244, 0xFE44, 0x44, 0xFE44, 0x45FE, 0x8400}},
    {19851, 2, {0x2000, 0x7CFC, 0x4444, 0x7C48, 0x4448, 0x7C50, 0x5C, 0xFE44, 0x9244, 0xFE44, 0x9268, 0xFEA8, 0x90, 0xFE98, 0x4524, 0x8442}},
    {19852, 2, {0x2040, 0x7C40, 0x44FC, 0x7C84, 0x4504, 0x7EF4, 0x94, 0xFE94, 0x92F4, 0xFE84, 0x9294, 0xFE88, 0x82, 0xFE82, 0x447E, 0x8400}},
    {19853, 2, {0x40, 0xFEF8, 0x2088, 0x28F8, 0x4488, 0xFCF8, 0x1400, 0x11FC, 0x1124, 0x7DFC, 0x1124, 0x11FC, 0x1C00, 0xE3FE, 0x4088, 0x108}},
    {19854, 2, {0x1010, 0x3E10, 0x2228, 0x3E28, 0x2244, 0x3E82, 0x7C, 0x7F00, 0x4900, 0x7F7C, 0x4944, 0x7F44, 0x44, 0xFFC4, 0x227C, 0x4244}},
    {19855, 2, {0x4420, 0x287C, 0x44, 0xFC7C, 0x1444, 0x147C, 0xFC00, 0x90FE, 0x9092, 0xFEFE, 0x3292, 0x32FE, 0x5A00, 0x54FE, 0x9044, 0x1084}},
    {19856, 2, {0x2010, 0x7C28, 0x4444, 0x7CBA, 0x4500, 0x7CFE, 0x92, 0xFED6, 0x9292, 0xFEFE, 0x9200, 0xFE7C, 0x44, 0xFE7C, 0x4444, 0x847C}},
    {19857, 2, {0x2020, 0x7CF8, 0x4420, 0x7DFC, 0x4504, 0x7CF8, 0xA8, 0xFEF8, 0x92A8, 0xFEF8, 0x9244, 0xFEFC, 0x20, 0xFEB8, 0x44A0, 0x857C}},
    {19858, 2, {0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2208, 0x3FF8, 0x2448, 0x4388, 0x9C68}},
    {19859, 2, {0x2040, 0x2020, 0x23FE, 0xF888, 0xA852, 0xAFAC, 0xAAAA, 0xAAA8, 0xADAC, 0xA800, 0xA904, 0xB9FC, 0x2104, 0x21FC, 0x2104, 0x2204}},
    {19860, 2, {0x1020, 0x5020, 0x5C20, 0x50A0, 0xFEAC, 0x28B4, 0xAAE4, 0xD7A4, 0x82A4, 0xFEB4, 0xAAA8, 0xAAA2, 0xD6A2, 0x9E82, 0xE27E, 0x200}},
    {19861, 2, {0x800, 0x2800, 0x2F7C, 0x2854, 0xFF54, 0x1454, 0x5554, 0x6B54, 0x417C, 0x7F40, 0x5540, 0x5540, 0x6B42, 0x4F42, 0x713E, 0x100}},
    {19862, 2, {0x810, 0x2810, 0x2F10, 0x2828, 0xFF28, 0x1444, 0x5544, 0x6BA2, 0x4110, 0x7F10, 0x557C, 0x5504, 0x6B08, 0x4F08, 0x7110, 0x110}},
    {19863, 2, {0x800, 0x287E, 0x2F04, 0x2804, 0xFF74, 0x1454, 0x5554, 0x6B54, 0x4154, 0x7F54, 0x5574, 0x5554, 0x6B04, 0x4F04, 0x7114, 0x108}},
    {19864, 2, {0x820, 0x2810, 0x2F10, 0x28FE, 0xFF00, 0x1404, 0x5544, 0x6B44, 0x4124, 0x7F24, 0x5528, 0x5528, 0x6B08, 0x4F10, 0x71FE, 0x100}},
    {19865, 2, {0x800, 0x287E, 0x2F40, 0x2840, 0xFF40, 0x147C, 0x5544, 0x6B44, 0x4144, 0x7F44, 0x557C, 0x5540, 0x6B40, 0x4F40, 0x717E, 0x100}},
    {19866, 2, {0x810, 0x2810, 0x2F10, 0x287C, 0xFF54, 0x1454, 0x5554, 0x6B54, 0x417C, 0x7F54, 0x5510, 0x5514, 0x6B12, 0x4F7E, 0x7122, 0x100}},
    {19867, 2, {0x7F04, 0x844, 0x1E44, 0x2244, 0x5444, 0x844, 0x310C, 0xC9E0, 0x900, 0x7FFC, 0x440, 0x2AA8, 0x3FF8, 0x2448, 0x2AA8, 0x3FF8}},
    {19868, 2, {0x1010, 0x5050, 0x5C50, 0x507C, 0xFE50, 0x2890, 0xAA10, 0xD6FE, 0x8200, 0xFE00, 0xAA7C, 0xAA44, 0xD644, 0x9E44, 0xE27C, 0x244}},
    {19869, 2, {0x1020, 0x5020, 0x5C20, 0x51FC, 0xFE20, 0x28A8, 0xAAA8, 0xD6A8, 0x8374, 0xFE22, 0xAA50, 0xAA50, 0xD688, 0x9E88, 0xE304, 0x202}},
    {19870, 2, {0x824, 0x2824, 0x2F7E, 0x2824, 0xFF24, 0x143C, 0x5524, 0x6B24, 0x413C, 0x7F24, 0x5524, 0x55FE, 0x6B00, 0x4F24, 0x7122, 0x142}},
    {19871, 2, {0x1020, 0x5020, 0x5C3C, 0x5044, 0xFE48, 0x2880, 0xAA20, 0xD6CC, 0x8284, 0xFE84, 0xAACC, 0xAA84, 0xD684, 0x9E84, 0xE2FC, 0x284}},
    {19872, 2, {0x28, 0x3FFE, 0x2020, 0x2FA4, 0x2028, 0x2F90, 0x48AA, 0x4FC6, 0x8082, 0x100, 0x11F8, 0x1100, 0xFFFE, 0x440, 0x2AA8, 0x3FF8}},
    {19873, 2, {0x1002, 0x50E2, 0x5CA2, 0x50AA, 0xFEAA, 0x28EA, 0xAAAA, 0xD6AA, 0x82EA, 0xFEAA, 0xAAAA, 0xAAA2, 0xD6E2, 0x9E02, 0xE24A, 0x2A4}},
    {19874, 2, {0x100C, 0x500A, 0x5C08, 0x50FE, 0xFE88, 0x2888, 0xAAFA, 0xD68A, 0x828A, 0xFEEC, 0xAAAC, 0xAAAC, 0xD6EA, 0x9E8A, 0xE296, 0x322}},
    {19875, 2, {0x1010, 0x5054, 0x5C54, 0x5054, 0xFE7C, 0x2800, 0xAAFE, 0xD600, 0x827C, 0xFE44, 0xAA44, 0xAA7C, 0xD644, 0x9E28, 0xE2FE, 0x200}},
    {19876, 2, {0x1000, 0x507C, 0x5C44, 0x5074, 0xFE54, 0x28FE, 0xAA82, 0xD67C, 0x8244, 0xFE7C, 0xAA44, 0xAA7C, 0xD644, 0x9E44, 0xE254, 0x248}},
    {19877, 2, {0x1020, 0x503C, 0x5C20, 0x50FE, 0xFEA2, 0x28B8, 0xAAE4, 0xD69C, 0x8280, 0xFEBC, 0xAAA4, 0xAABC, 0xD6A4, 0x9EBC, 0xE2A4, 0x37E}},
    {19878, 2, {0x1020, 0x51FC, 0x5C20, 0x50FC, 0xFE20, 0x29FE, 0xAA00, 0xD6FC, 0x8284, 0xFEFC, 0xAA84, 0xAAFC, 0xD684, 0x9EFC, 0xE248, 0x284}},
    {19879, 2, {0x1010, 0x5010, 0x5C7E, 0x5010, 0xFEFE, 0x2844, 0xAAEE, 0xD644, 0x8264, 0xFEDE, 0xAA00, 0xAAFE, 0xD628, 0x9E28, 0xE24A, 0x286}},
    {19880, 2, {0x1010, 0x5010, 0x5C28, 0x5044, 0xFE82, 0x287C, 0xAA00, 0xD6EE, 0x82AA, 0xFEAA, 0xAAEE, 0xAA44, 0xD644, 0x9EAA, 0xE2AA, 0x312}},
    {19881, 2, {0x1020, 0x5010, 0x5DFE, 0x5054, 0xFFA8, 0x28AC, 0xABAA, 0xD6A8, 0x83AC, 0xFE00, 0xAA84, 0xAAFC, 0xD684, 0x9EFC, 0xE284, 0x304}},
    {19882, 2, {0x104C, 0x51FA, 0x5C48, 0x51FE, 0xFE08, 0x2868, 0xAA5A, 0xD6FA, 0x83AA, 0xFEFC, 0xAAAC, 0xAAFC, 0xD6AA, 0x9EFA, 0xE296, 0x222}},
    {19883, 2, {0x13DE, 0x5252, 0x5FDE, 0x5000, 0xFFFE, 0x2A00, 0xAAC8, 0xD64E, 0x83EA, 0xFEB2, 0xAAEA, 0xAAAA, 0xD6EA, 0x9EB4, 0xE3E4, 0x22A}},
    {19884, 2, {0x0, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x1010, 0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E}},
    {19885, 2, {0x1400, 0x12FE, 0x1210, 0x1010, 0xFF90, 0x147C, 0x1410, 0x1510, 0x1510, 0x16FE, 0x2600, 0x2402, 0x2C02, 0x5402, 0x43FE, 0x8000}},
    {19886, 2, {0x440, 0x420, 0xFFFE, 0x910, 0x1160, 0x2184, 0x4F04, 0x80FC, 0x0, 0x1FF0, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0x600C}},
    {19887, 2, {0x2080, 0x27FC, 0x2404, 0x27BC, 0xF0A0, 0x56BE, 0x52A2, 0x57B6, 0x52AA, 0x96AA, 0x50B6, 0x26A2, 0x22BE, 0x57A0, 0x52A2, 0x867E}},
    {19888, 2, {0x1000, 0xFF86, 0x80B8, 0xF788, 0x1408, 0xD7CE, 0x5458, 0xF6C8, 0x5548, 0xD54E, 0x16F8, 0xD448, 0x57CA, 0xF40A, 0x544A, 0xCFC6}},
    {19889, 2, {0x4080, 0x47FC, 0x7404, 0x97BC, 0x50A0, 0x26BE, 0x52A2, 0x8FB6, 0xAAA, 0x66AA, 0x10B6, 0x6A2, 0x82BE, 0x47A0, 0x22A2, 0x67E}},
    {19890, 2, {0x1040, 0x13FE, 0x1202, 0x7FDE, 0x5450, 0x575E, 0x7D52, 0x57DA, 0x5556, 0x5756, 0xFC5A, 0x4752, 0x455E, 0x47D0, 0x5552, 0x4B3E}},
    {19891, 2, {0x810, 0x808, 0x1400, 0x227E, 0x5D00, 0x8080, 0x7F38, 0x5528, 0x7F28, 0x28, 0x7F28, 0x552A, 0x7F2A, 0x554A, 0x5546, 0x4380}},
    {19892, 2, {0x840, 0x2A7C, 0x4894, 0x1410, 0x2128, 0x46C4, 0x1830, 0xEFEE, 0x0, 0x3BB8, 0x2AA8, 0x7FFC, 0x4444, 0x7FFC, 0x4444, 0x444C}},
    {19893, 2, {0x1004, 0x103E, 0x2BE0, 0x463C, 0xBA20, 0x2FC, 0xFEA4, 0xAAB8, 0xFEE4, 0x29C, 0xFE80, 0xAAB8, 0xFEA8, 0xAAAA, 0xAB2A, 0x8646}},
    {19968, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {19969, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {19970, 2, {0x0, 0xFFFE, 0x400, 0x400, 0x800, 0x800, 0x1000, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xA0, 0x40}},
    {19971, 2, {0x200, 0x200, 0x200, 0x200, 0x200, 0x23C, 0x7C0, 0xFA00, 0x200, 0x200, 0x200, 0x204, 0x204, 0x204, 0x1FC, 0x0}},
    {19972, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {19973, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {19974, 2, {0x0, 0xFFFE, 0x200, 0x200, 0x400, 0x400, 0x800, 0x800, 0x1000, 0x2000, 0x4000, 0x8000, 0x0, 0x0, 0x0, 0x0}},
    {19975, 2, {0x0, 0x0, 0xFFFE, 0x400, 0x400, 0x400, 0x7F0, 0x410, 0x410, 0x810, 0x810, 0x1010, 0x1010, 0x2010, 0x40A0, 0x8040}},
    {19976, 2, {0x80, 0x80, 0x80, 0x80, 0x7FFC, 0x80, 0x1080, 0x1080, 0x880, 0x480, 0x280, 0x100, 0x280, 0x440, 0x1830, 0xE00E}},
    {19977, 2, {0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0, 0x0}},
    {19978, 2, {0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x3F8, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0xFFFE, 0x0}},
    {19979, 2, {0x0, 0xFFFE, 0x200, 0x200, 0x200, 0x200, 0x240, 0x220, 0x210, 0x208, 0x208, 0x200, 0x200, 0x200, 0x200, 0x200}},
    {19980, 2, {0x0, 0xFFFE, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020, 0x8020}},
    {19981, 2, {0x0, 0x7FFC, 0x80, 0x80, 0x100, 0x100, 0x340, 0x520, 0x910, 0x1108, 0x2104, 0x4104, 0x8100, 0x100, 0x100, 0x100}},
    {19982, 2, {0x1000, 0x1000, 0x1000, 0x1FFC, 0x1000, 0x2000, 0x2000, 0x3FF8, 0x8, 0x8, 0x8, 0xFFC8, 0x8, 0x8, 0x50, 0x20}},
    {19983, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x2100, 0x21FC, 0x2104, 0x2104, 0x2104, 0x3FFC, 0x4, 0x4, 0x4, 0x4, 0x28, 0x10}},
    {19984, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x2100, 0x21FC, 0x2100, 0x2100, 0x2100, 0x3FFC, 0x4, 0x4, 0x4, 0x4, 0x28, 0x10}},
    {19985, 2, {0x0, 0x3FF0, 0x210, 0x210, 0x210, 0x210, 0x210, 0x3FF0, 0x410, 0x410, 0x410, 0x410, 0x410, 0x410, 0xFFFE, 0x0}},
    {19986, 2, {0x0, 0x3FF8, 0x108, 0x108, 0x908, 0x508, 0x208, 0x308, 0x488, 0x488, 0x808, 0x1050, 0x2020, 0x0, 0xFFFE, 0x0}},
    {19987, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x200, 0x200, 0xFFFE, 0x400, 0x800, 0xFF0, 0x10, 0x20, 0x640, 0x180, 0x40, 0x20}},
    {19988, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {19989, 2, {0x0, 0x7FFC, 0x80, 0x100, 0x100, 0x340, 0x520, 0x910, 0x3108, 0xC104, 0x100, 0x100, 0x100, 0x0, 0xFFFE, 0x0}},
    {19990, 2, {0x220, 0x1220, 0x1220, 0x1220, 0x1220, 0xFFFE, 0x1220, 0x1220, 0x1220, 0x1220, 0x13E0, 0x1000, 0x1000, 0x1000, 0x1FFC, 0x0}},
    {19991, 2, {0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {19992, 2, {0x10, 0x78, 0x1F80, 0x1000, 0x1000, 0x1000, 0x1000, 0x1FFC, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0xFFFE, 0x0}},
    {19993, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x4104, 0x4284, 0x4244, 0x4424, 0x4824, 0x4004, 0x4004, 0x4014, 0x4008}},
    {19994, 2, {0x440, 0x440, 0x440, 0x440, 0x4444, 0x2444, 0x2448, 0x1448, 0x1450, 0x1460, 0x440, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {19995, 2, {0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x1450, 0x1250, 0x2288, 0x2088, 0x4104, 0x8202, 0x0, 0xFFFE, 0x0}},
    {19996, 2, {0x200, 0x200, 0x200, 0x7FFC, 0x400, 0x900, 0x1100, 0x2100, 0x3FF8, 0x100, 0x920, 0x1110, 0x2108, 0x4104, 0x500, 0x200}},
    {19997, 2, {0x810, 0x810, 0x1020, 0x1020, 0x2244, 0x4284, 0x7CF8, 0x408, 0x810, 0x1020, 0x2040, 0x7EFC, 0x0, 0x0, 0xFFFE, 0x0}},
    {19998, 2, {0x0, 0x3FF0, 0x20, 0x40, 0x80, 0x7D04, 0x588, 0x950, 0x920, 0x1110, 0x2108, 0x4106, 0x8500, 0x200, 0xFFFE, 0x0}},
    {19999, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {20000, 2, {0x440, 0x440, 0x440, 0x444, 0x7C48, 0x450, 0x460, 0x440, 0x442, 0x442, 0x1C42, 0xE43E, 0x4400, 0x0, 0xFFFE, 0x0}},
    {20001, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x4924, 0x4924, 0x4924, 0x4924, 0x4FE4, 0x4024, 0x4004, 0x4014, 0x4008}},
    {20002, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {20003, 2, {0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x7C7C, 0x4444, 0x4444, 0x4444, 0x7C7C, 0x444, 0x840, 0x840, 0x1040, 0x2040, 0x4040}},
    {20004, 2, {0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x4AA4, 0x4A94, 0x5114, 0x4204, 0x4004, 0x4014, 0x4008}},
    {20005, 2, {0x0, 0x7FFC, 0x240, 0x1248, 0xA50, 0x240, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {20006, 2, {0x1010, 0x820, 0x440, 0x7FFC, 0x440, 0x440, 0x440, 0x4444, 0x2444, 0x1448, 0x1450, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {20007, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1110, 0x910, 0x920, 0x100, 0xFFFE, 0x1200, 0x1110, 0x10A0, 0x1040, 0x1430, 0x180E, 0x1000}},
    {20008, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {20009, 2, {0x40, 0x40, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x10C0, 0x1340, 0x1C40, 0x1040, 0x40, 0x40, 0x40, 0x40}},
    {20010, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2108, 0xC106, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {20011, 2, {0x0, 0x2010, 0x1020, 0x840, 0x480, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {20012, 2, {0x100, 0x100, 0x100, 0x1100, 0x900, 0x900, 0x100, 0x300, 0x500, 0x900, 0x3100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {20013, 2, {0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100, 0x100}},
    {20014, 2, {0x0, 0x7FF0, 0x410, 0x410, 0x410, 0x7FD0, 0x410, 0x410, 0x410, 0xFFD0, 0x412, 0x40A, 0x40A, 0x406, 0x402, 0x400}},
    {20015, 2, {0x100, 0x10C, 0x170, 0x780, 0x7900, 0x10C, 0x170, 0x780, 0x7900, 0x10C, 0x170, 0x780, 0x7900, 0x100, 0x100, 0x100}},
    {20016, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {20017, 2, {0x440, 0x440, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x4C44, 0x547C, 0x6444, 0x440, 0x840, 0x840, 0x1040, 0x2040, 0xC040}},
    {20018, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x4104, 0x4104, 0x7FFC, 0x4104, 0x100, 0x100, 0x100}},
    {20019, 2, {0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x4844, 0x1040, 0x2040, 0x4040}},
    {20020, 2, {0x880, 0x880, 0x880, 0x49FE, 0x4940, 0x4A20, 0x4C20, 0x4800, 0x49FC, 0x4924, 0x4924, 0x4924, 0x4924, 0x9FC, 0x904, 0x800}},
    {20021, 2, {0x440, 0x2448, 0x1450, 0x440, 0xFFFE, 0x0, 0x820, 0x440, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {20022, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x400, 0x200, 0x100, 0x80, 0x40, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {20023, 2, {0x0, 0x0, 0x0, 0x8, 0x2008, 0x1010, 0x810, 0x420, 0x440, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {20024, 2, {0x200, 0x200, 0x200, 0x200, 0x7FE0, 0x220, 0x1220, 0xA20, 0x420, 0x420, 0xA20, 0x922, 0x1122, 0x2022, 0x401E, 0x8000}},
    {20025, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1210, 0x1110, 0x1110, 0x1010, 0xFFFE, 0x1010, 0x1010, 0x1010, 0x2010, 0x2010, 0x4050, 0x8020}},
    {20026, 2, {0x100, 0x2100, 0x1100, 0x1100, 0x100, 0x7FF8, 0x208, 0x208, 0x288, 0x448, 0x448, 0x808, 0x1008, 0x2008, 0x4050, 0x8020}},
    {20027, 2, {0x200, 0x100, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {20028, 2, {0x820, 0x820, 0x820, 0x820, 0x7FFC, 0x820, 0xA20, 0x920, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {20029, 2, {0x0, 0xFFFE, 0x0, 0x0, 0x3EF8, 0x2288, 0x2288, 0x2288, 0x32C8, 0x2AA8, 0x2AA8, 0x2288, 0x2288, 0x2288, 0x2AA8, 0x2490}},
    {20030, 2, {0x2208, 0x1108, 0x1110, 0x20, 0xFFFE, 0x820, 0x1110, 0x2108, 0xCFE6, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x100}},
    {20031, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x200, 0x200, 0x400, 0x400, 0x800, 0x1000, 0x2000}},
    {20032, 2, {0x0, 0x0, 0x0, 0x800, 0x800, 0x1400, 0x1400, 0x2400, 0x4200, 0x100, 0xC0, 0x30, 0xE, 0x0, 0x0, 0x0}},
    {20033, 2, {0x0, 0x0, 0x3F00, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x80, 0x80, 0x40, 0x20, 0x10, 0x8, 0x6}},
    {20034, 2, {0x0, 0x10, 0x2010, 0x2010, 0x1020, 0x1020, 0x820, 0x840, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {20035, 2, {0x0, 0x3FF0, 0x410, 0x420, 0x420, 0x440, 0x47C, 0x404, 0x804, 0x804, 0x804, 0x1004, 0x1004, 0x2004, 0x4028, 0x8010}},
    {20036, 2, {0x0, 0x10, 0x10, 0x10, 0x10, 0x220, 0x120, 0xC0, 0x40, 0xA0, 0x110, 0x2210, 0x2400, 0x3800, 0x2000, 0x0}},
    {20037, 2, {0x400, 0x400, 0x400, 0x7E0, 0x820, 0x820, 0x1040, 0x2040, 0x4080, 0x80, 0x140, 0x240, 0x420, 0x810, 0x3008, 0xC006}},
    {20038, 2, {0x10, 0x810, 0x810, 0x1020, 0x1020, 0x2040, 0x7FC0, 0x2080, 0x80, 0x140, 0x140, 0x220, 0x420, 0x810, 0x3008, 0xC006}},
    {20039, 2, {0x20, 0x3F0, 0x3E00, 0x200, 0x200, 0x200, 0x200, 0x3FC, 0xFE00, 0x200, 0x200, 0x204, 0x204, 0x204, 0x1FC, 0x0}},
    {20040, 2, {0x100, 0x100, 0x200, 0x200, 0x440, 0x840, 0x1080, 0x2080, 0x4100, 0x200, 0x420, 0x810, 0x1008, 0x3FFC, 0x1004, 0x0}},
    {20041, 2, {0x200, 0x110, 0x2110, 0x2010, 0x1020, 0x1020, 0x820, 0x840, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {20042, 2, {0x0, 0x8, 0x1FC, 0x7E00, 0x8, 0x2008, 0x1010, 0x1010, 0x820, 0x840, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {20043, 2, {0x200, 0x100, 0x100, 0x0, 0x7FF8, 0x10, 0x20, 0x40, 0x80, 0x100, 0x200, 0x400, 0x1800, 0x2400, 0x43FE, 0x0}},
    {20044, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1050, 0x1020, 0x1000, 0x1FFC, 0x4, 0x4, 0x7FE4, 0x4, 0x28, 0x10}},
    {20045, 2, {0x800, 0x800, 0xFFE, 0x1200, 0x1200, 0x2200, 0x43F8, 0x8200, 0x200, 0x200, 0x3FC, 0x200, 0x200, 0x200, 0x200, 0x200}},
    {20046, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x1110, 0x910, 0x920, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {20047, 2, {0x10, 0x1F8, 0x7E00, 0x0, 0x200, 0x100, 0x100, 0x7FF8, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x4800, 0x87FE, 0x0}},
    {20048, 2, {0x20, 0xF0, 0x1F00, 0x1000, 0x1100, 0x2100, 0x2100, 0x3FFC, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x500, 0x200}},
    {20049, 2, {0x8, 0x1FC, 0x7F00, 0x100, 0x904, 0x908, 0x1190, 0x21A0, 0x4540, 0x8540, 0x920, 0x1110, 0x2108, 0x4106, 0x8100, 0x100}},
    {20050, 2, {0x20, 0xF0, 0x1F00, 0x1000, 0x1000, 0x1FF8, 0x1080, 0x1080, 0x1080, 0x1080, 0xFFFE, 0x0, 0x800, 0x1000, 0x2000, 0x4000}},
    {20051, 2, {0x20, 0xF0, 0x1F00, 0x1000, 0x1000, 0x1FF8, 0x1080, 0x1080, 0x1080, 0x1080, 0xFFFE, 0x0, 0x40, 0x20, 0x10, 0x8}},
    {20052, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC826, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020}},
    {20053, 2, {0x8, 0xFC, 0x3F80, 0x2080, 0x2FF8, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x2080, 0x2FF8, 0x2888, 0x4888, 0x48A8, 0x8890, 0x80}},
    {20054, 2, {0x70, 0x1F80, 0x100, 0x100, 0xFFFE, 0x100, 0x920, 0x924, 0x7928, 0x930, 0x922, 0x1922, 0xE91E, 0x100, 0x100, 0x100}},
    {20055, 2, {0xF8, 0x3F00, 0x100, 0x7FFC, 0x920, 0x920, 0xFFFE, 0x920, 0x920, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {20056, 2, {0xF8, 0x3F00, 0x100, 0x7FFC, 0x100, 0x924, 0xF938, 0x922, 0x3922, 0xCB9E, 0x540, 0x920, 0x3118, 0xC106, 0x100, 0x100}},
    {20057, 2, {0x0, 0x7FF0, 0x20, 0x40, 0x80, 0x100, 0x200, 0x400, 0x800, 0x1000, 0x1000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {20058, 2, {0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1004, 0x1004, 0x1004, 0xFFC, 0x0}},
    {20059, 2, {0x0, 0x0, 0xFFFC, 0x4, 0x8, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {20060, 2, {0x0, 0x0, 0x1000, 0x1030, 0x10D0, 0x1710, 0x3810, 0xD010, 0x1010, 0x1050, 0x1020, 0x1002, 0x1002, 0x1002, 0xFFE, 0x0}},
    {20061, 2, {0x400, 0x400, 0x400, 0x400, 0x7FE0, 0x420, 0x420, 0x420, 0x820, 0x820, 0x820, 0x1022, 0x1022, 0x2022, 0x401E, 0x8000}},
    {20062, 2, {0x800, 0x800, 0x1FFC, 0x1000, 0x2000, 0x4000, 0x9FE0, 0x20, 0x40, 0x180, 0x600, 0x804, 0x1004, 0x1004, 0xFFC, 0x0}},
    {20063, 2, {0x100, 0x100, 0x1100, 0x1130, 0x11D0, 0x1710, 0x3910, 0xD110, 0x1110, 0x1150, 0x1120, 0x1102, 0x1002, 0x1002, 0xFFE, 0x0}},
    {20064, 2, {0x0, 0x7FF8, 0x8, 0x8, 0x808, 0x408, 0x208, 0x208, 0x68, 0x188, 0xE08, 0x7008, 0x2008, 0x8, 0x50, 0x20}},
    {20065, 2, {0x200, 0x200, 0x420, 0x820, 0x1040, 0x3F80, 0x108, 0x208, 0x410, 0x1FE0, 0x40, 0x80, 0x100, 0x600, 0x1800, 0xE000}},
    {20066, 2, {0x820, 0x820, 0x820, 0x820, 0x4920, 0x4920, 0x4920, 0x4920, 0x4920, 0x4920, 0x4920, 0x4F22, 0x7922, 0x22, 0x1E, 0x0}},
    {20067, 2, {0x840, 0x840, 0x1040, 0x1040, 0x2240, 0x4240, 0xFC40, 0x440, 0x840, 0x1040, 0x2440, 0x4242, 0xFE42, 0x4242, 0x3E, 0x0}},
    {20068, 2, {0x0, 0xFFFE, 0x200, 0x240, 0x220, 0x210, 0x208, 0x200, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20069, 2, {0x8, 0x1FC, 0x7E00, 0x10, 0x1010, 0x820, 0x440, 0x80, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20070, 2, {0x220, 0x210, 0x208, 0x3FE0, 0x220, 0x220, 0x220, 0x220, 0xFFFC, 0x204, 0x204, 0x204, 0x204, 0x228, 0x210, 0x200}},
    {20071, 2, {0x40, 0x440, 0x240, 0x1240, 0x840, 0x9FC, 0x7E40, 0x40, 0x40, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {20072, 2, {0x820, 0x820, 0x1020, 0x2220, 0x4120, 0xFFA0, 0xA0, 0x20, 0x7F20, 0x4120, 0x4120, 0x4122, 0x4122, 0x4122, 0x7F1E, 0x4100}},
    {20073, 2, {0x820, 0x820, 0x820, 0x820, 0xFA0, 0x820, 0x820, 0x820, 0x7F20, 0x4120, 0x4120, 0x4122, 0x4122, 0x4122, 0x7F1E, 0x4100}},
    {20074, 2, {0x0, 0xFDFC, 0x524, 0x524, 0x924, 0x924, 0x11FC, 0x1124, 0x2124, 0x2124, 0x4124, 0x41FC, 0x4002, 0x4002, 0x3FFE, 0x0}},
    {20075, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x1244, 0x227C, 0x4A44, 0x8400, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20076, 2, {0x0, 0x3FFC, 0x2000, 0x3FF8, 0x2008, 0x3FF8, 0x2000, 0x3FFC, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20077, 2, {0x0, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x0, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20078, 2, {0xC00, 0x70FC, 0x4084, 0x4484, 0x5C84, 0x6494, 0x4488, 0x880, 0x1080, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20079, 2, {0xF8, 0x7F00, 0x1110, 0x920, 0xFFFE, 0x100, 0x100, 0x500, 0x200, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20080, 2, {0x0, 0x7FFC, 0x4, 0x888, 0x480, 0x2480, 0x1080, 0x1080, 0x100, 0xFFFE, 0x100, 0x2C0, 0x420, 0x810, 0x3008, 0xC004}},
    {20081, 2, {0x440, 0xE40, 0x3840, 0x840, 0x840, 0xFF40, 0x840, 0x840, 0x3E40, 0x2240, 0x2240, 0x2242, 0x2242, 0x3E42, 0x223E, 0x0}},
    {20082, 2, {0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {20083, 2, {0x220, 0xF20, 0xF020, 0x120, 0x9120, 0x4A20, 0x20, 0xFE20, 0x420, 0x820, 0xF20, 0xF822, 0x4822, 0x822, 0x281E, 0x1000}},
    {20084, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {20085, 2, {0x1020, 0x820, 0x7F20, 0x20, 0x2220, 0x1420, 0xFFA0, 0x820, 0x820, 0x820, 0x7F20, 0x822, 0x822, 0x822, 0x81E, 0x800}},
    {20086, 2, {0x120, 0x110, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x0, 0x3FF0, 0x1C0, 0x1E04, 0x2004, 0x1FFC}},
    {20087, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x922, 0x1228, 0x7030, 0x10C0, 0x1700, 0x0, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {20088, 2, {0x800, 0x87C, 0x2844, 0x2E64, 0x3A54, 0xEAFE, 0x2A44, 0x2AA4, 0x2A94, 0x2EFE, 0x2804, 0x282A, 0x2012, 0x2002, 0x1FFE, 0x0}},
    {20089, 2, {0x820, 0x820, 0xFFA0, 0x820, 0x7F20, 0x4120, 0x7F20, 0x4120, 0x7F20, 0x4920, 0x820, 0xFFA2, 0x822, 0x822, 0x81E, 0x800}},
    {20090, 2, {0x618, 0x38E0, 0x2080, 0x3EFC, 0x2290, 0x3E90, 0x2090, 0x4110, 0x8000, 0x7FF0, 0x60, 0x180, 0x600, 0x1804, 0x2004, 0x1FFC}},
    {20091, 2, {0x1010, 0x810, 0xFF28, 0x2044, 0x3E82, 0x2210, 0x2208, 0x4220, 0x4A10, 0x8408, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {20092, 2, {0x2080, 0x1040, 0x87FC, 0x4040, 0x1040, 0x23F8, 0xE040, 0x2040, 0x27FC, 0x2000, 0x7FF0, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC}},
    {20093, 2, {0x108, 0x1FD0, 0x120, 0x7FFC, 0x200, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x0, 0x3FF0, 0x1C0, 0x1E04, 0x2004, 0x1FFC}},
    {20094, 2, {0x1040, 0x1040, 0x1080, 0xFEFE, 0x1100, 0x7E00, 0x44FC, 0x7C08, 0x4410, 0x7C20, 0x1040, 0xFE80, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {20095, 2, {0xF20, 0xF020, 0x1120, 0x8920, 0x4220, 0x1020, 0x2220, 0x7C20, 0x820, 0x1120, 0xFFA0, 0x8A2, 0x4A22, 0x8922, 0x281E, 0x1000}},
    {20096, 2, {0x400, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {20097, 2, {0x820, 0x820, 0xFFA0, 0x87E, 0x7F40, 0x4180, 0x7F7C, 0x4104, 0x7F08, 0x10, 0x7F20, 0x840, 0xFFC2, 0x842, 0x83E, 0x800}},
    {20098, 2, {0xF20, 0xF020, 0x4920, 0x2220, 0xFE20, 0x1420, 0x820, 0xFF20, 0x9520, 0xBD20, 0x8320, 0xBD22, 0x9522, 0x8922, 0x951E, 0xA300}},
    {20099, 2, {0x820, 0x4920, 0x2A20, 0xFFA0, 0x2A20, 0x4920, 0x88A0, 0x4120, 0x4120, 0x77A0, 0x9120, 0x5522, 0x27A2, 0x4122, 0x811E, 0x100}},
    {20100, 2, {0x820, 0xFFA0, 0x820, 0x7F20, 0x20, 0xFFA0, 0x80A0, 0x7F20, 0x20, 0x7F20, 0x4120, 0x7F22, 0x4122, 0x2222, 0xF9E, 0xF000}},
    {20101, 2, {0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x20}},
    {20102, 2, {0x0, 0x7FF8, 0x10, 0x20, 0x40, 0x180, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {20103, 2, {0x800, 0x800, 0x800, 0xFFC, 0x1004, 0x1108, 0x2110, 0x4100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {20104, 2, {0x0, 0x1FF0, 0x10, 0x220, 0x140, 0x80, 0x7FFE, 0x82, 0x84, 0x80, 0x80, 0x80, 0x80, 0x80, 0x280, 0x100}},
    {20105, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x108, 0x108, 0x7FFE, 0x108, 0x108, 0x3FF8, 0x108, 0x100, 0x500, 0x200}},
    {20106, 2, {0x100, 0x100, 0x7FFC, 0x1110, 0x920, 0x3FF8, 0x108, 0x108, 0xFFFE, 0x108, 0x108, 0x3FF8, 0x108, 0x100, 0x500, 0x200}},
    {20107, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x500, 0x200}},
    {20108, 2, {0x0, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0, 0x0, 0x0}},
    {20109, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {20110, 2, {0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {20111, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0xFFFE, 0x400, 0x400, 0x800, 0xFF0, 0x10, 0x10, 0x10, 0x10, 0xA0, 0x40}},
    {20112, 2, {0x0, 0x3FF8, 0x200, 0x200, 0x200, 0xFFFE, 0x400, 0x400, 0x800, 0xFF0, 0x10, 0x10, 0x10, 0x10, 0xA0, 0x40}},
    {20113, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x200, 0x400, 0x400, 0x840, 0x1020, 0x2010, 0x7FF8, 0x2008, 0x8}},
    {20114, 2, {0x0, 0x7FFC, 0x400, 0x400, 0x400, 0xFF0, 0x810, 0x810, 0x1010, 0x1020, 0x1FE0, 0x20, 0x40, 0x40, 0xFFFE, 0x0}},
    {20115, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0xFFFE, 0x820, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020, 0x8020}},
    {20116, 2, {0x0, 0x7FFC, 0x200, 0x200, 0x200, 0x200, 0x3FF0, 0x410, 0x410, 0x410, 0x410, 0x810, 0x810, 0x810, 0xFFFE, 0x0}},
    {20117, 2, {0x820, 0x820, 0x820, 0x820, 0x7FFC, 0x820, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {20118, 2, {0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0xFFFE, 0x0, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0xFFFE, 0x0}},
    {20119, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0}},
    {20120, 2, {0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0, 0x0, 0xFFFE, 0x0}},
    {20121, 2, {0x0, 0x7FFC, 0x400, 0x400, 0x400, 0xFF0, 0x810, 0xA10, 0x1110, 0x1020, 0x420, 0x220, 0x40, 0x40, 0xFFFE, 0x0}},
    {20122, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x4444, 0x2444, 0x2448, 0x1448, 0x1450, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {20123, 2, {0x440, 0x440, 0x2444, 0x2748, 0x2470, 0x2442, 0x2442, 0x2F42, 0xF03E, 0x4000, 0x0, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0}},
    {20124, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {20125, 2, {0x400, 0x820, 0x1010, 0x3FF8, 0x810, 0x1020, 0x2448, 0x7EFC, 0x204, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0xFFFE, 0x0}},
    {20126, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x3C78, 0x2008, 0x2008, 0x2008, 0x2008, 0x3C78, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {20127, 2, {0x0, 0x3FF0, 0x20, 0x40, 0x180, 0x793C, 0x4904, 0x4944, 0x4928, 0x4910, 0x7928, 0x144, 0x500, 0x200, 0xFFFE, 0x0}},
    {20128, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {20129, 2, {0x200, 0x100, 0x100, 0x0, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1FFC, 0x0}},
    {20130, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x0, 0x0, 0xFE0, 0x820, 0x820, 0x820, 0x820, 0x822, 0x1022, 0x1022, 0x201E, 0x4000}},
    {20131, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {20132, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x0, 0x1010, 0x1008, 0x2024, 0x4824, 0x440, 0x280, 0x100, 0x280, 0xC40, 0x3030, 0xC00E}},
    {20133, 2, {0x200, 0x100, 0x0, 0x7FFC, 0x200, 0x400, 0x820, 0x1040, 0x3F88, 0x110, 0x620, 0x1840, 0x60A0, 0x310, 0xC08, 0x7004}},
    {20134, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x440, 0x440, 0x2450, 0x2448, 0x2448, 0x4444, 0x4444, 0x8844, 0x840, 0x1040, 0x2140, 0x4080}},
    {20135, 2, {0x200, 0x100, 0x7FFC, 0x0, 0x820, 0x420, 0x440, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {20136, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF0, 0x60, 0x180, 0x100, 0x100, 0x500, 0x200}},
    {20137, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {20138, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4844, 0x940, 0x1080, 0x7FE0, 0x20, 0x20, 0x10, 0x10, 0x8, 0x6}},
    {20139, 2, {0x200, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x1FE0, 0x40, 0x180, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {20140, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {20141, 2, {0x200, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {20142, 2, {0x200, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8FE4, 0x820, 0x820, 0x1022, 0x2022, 0xC01E}},
    {20143, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {20144, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {20145, 2, {0x200, 0x100, 0xFFFE, 0x800, 0x800, 0x13F8, 0x1208, 0x3208, 0x53F8, 0x9208, 0x1208, 0x13F8, 0x1208, 0x1000, 0x17FC, 0x1000}},
    {20146, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x100, 0x920, 0x1110, 0x2108, 0x4504, 0x200}},
    {20147, 2, {0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x80F4, 0x3F00, 0x200, 0x3F8, 0x7E02, 0x202, 0x1FE}},
    {20148, 2, {0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4102, 0x9FF4, 0x100, 0x7FFC, 0x200, 0x3FE0, 0x822, 0x701E}},
    {20149, 2, {0x100, 0xFFFE, 0x0, 0x840, 0x7DF8, 0x848, 0xCC8, 0x784A, 0x8AA, 0x2906, 0x1302, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {20150, 2, {0x100, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE}},
    {20151, 2, {0x100, 0x80, 0x3FFC, 0x410, 0x220, 0x3FFE, 0x2240, 0x2FF8, 0x2248, 0x3FFE, 0x2248, 0x2FF8, 0x2660, 0x4A50, 0x524C, 0x8240}},
    {20152, 2, {0x2088, 0x1048, 0xFC50, 0x1FC, 0x7D24, 0x4524, 0x7DFC, 0x124, 0x7D24, 0x9FC, 0x1020, 0x1C20, 0xF3FE, 0x1020, 0x5020, 0x2020}},
    {20153, 2, {0x100, 0xFFFE, 0x0, 0x2FEC, 0x4824, 0x6BAC, 0x4824, 0x6BAC, 0x4AA4, 0xFFFE, 0x9012, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {20154, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x820, 0x1010, 0x2008, 0x4004, 0x8002}},
    {20155, 2, {0x800, 0x800, 0x800, 0x1000, 0x1000, 0x3000, 0x3000, 0x5000, 0x9000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {20156, 2, {0x100, 0x100, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006, 0x0, 0x0, 0xFFFE, 0x0}},
    {20157, 2, {0x100, 0x100, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006, 0x200, 0x100, 0x80, 0x80, 0x0, 0x0}},
    {20158, 2, {0x80, 0x80, 0x80, 0x2080, 0x2080, 0x2140, 0x2140, 0x2220, 0x2220, 0x2410, 0x2808, 0x3006, 0x2000, 0x2000, 0x3FFE, 0x0}},
    {20159, 2, {0x800, 0x800, 0xBFC, 0x1008, 0x1010, 0x3020, 0x3040, 0x5080, 0x9080, 0x1100, 0x1100, 0x1202, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {20160, 2, {0x820, 0x820, 0x820, 0x1020, 0x1020, 0x3020, 0x37FE, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {20161, 2, {0x800, 0x800, 0x800, 0x13FC, 0x1000, 0x3000, 0x3000, 0x5000, 0x9000, 0x1000, 0x1000, 0x1000, 0x17FE, 0x1000, 0x1000, 0x1000}},
    {20162, 2, {0x840, 0x840, 0x840, 0x1040, 0x13FC, 0x3044, 0x3044, 0x5044, 0x9044, 0x1084, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {20163, 2, {0x800, 0x800, 0xBFE, 0x1020, 0x1020, 0x3020, 0x3020, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {20164, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x2080, 0x2080, 0x2080, 0x2080, 0x2080, 0x2140, 0x2140, 0x2220, 0x4220, 0x4410, 0x8808, 0x1006}},
    {20165, 2, {0x800, 0xBFC, 0x904, 0x1104, 0x1104, 0x3088, 0x3088, 0x5088, 0x9050, 0x1050, 0x1020, 0x1020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {20166, 2, {0x840, 0x840, 0x840, 0x1040, 0x1040, 0x3050, 0x3048, 0x5044, 0x9044, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20167, 2, {0x900, 0x900, 0x900, 0x1100, 0x17F0, 0x3110, 0x3110, 0x5110, 0x9110, 0x1110, 0x1110, 0x1212, 0x1212, 0x1412, 0x180E, 0x1000}},
    {20168, 2, {0x800, 0x820, 0x920, 0x1120, 0x1120, 0x3120, 0x3120, 0x5120, 0x9120, 0x1110, 0x1110, 0x1210, 0x1208, 0x1408, 0x1804, 0x1002}},
    {20169, 2, {0x800, 0x9F0, 0x910, 0x1110, 0x1110, 0x3110, 0x3110, 0x5110, 0x9110, 0x1110, 0x1110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {20170, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1210, 0x2108, 0xC106, 0x0, 0x1FF0, 0x10, 0x20, 0x20, 0x40, 0x80, 0x100}},
    {20171, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC826, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {20172, 2, {0x100, 0x100, 0x280, 0x240, 0x420, 0x810, 0x1108, 0x6104, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {20173, 2, {0x800, 0x800, 0xBF8, 0x1088, 0x1090, 0x3090, 0x30A0, 0x50BC, 0x9084, 0x1084, 0x1104, 0x1104, 0x1104, 0x1204, 0x1228, 0x1410}},
    {20174, 2, {0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x850, 0x1450, 0x1250, 0x1288, 0x2088, 0x2104, 0x4204, 0x8402}},
    {20175, 2, {0x840, 0x840, 0x840, 0x1040, 0x1040, 0x3080, 0x3080, 0x5080, 0x9100, 0x1110, 0x1108, 0x1208, 0x17FC, 0x1204, 0x1004, 0x1000}},
    {20176, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2108, 0xC106, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {20177, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC806, 0x830, 0x8C0, 0xF00, 0x808, 0x808, 0x808, 0x7F8, 0x0}},
    {20178, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC606, 0x180, 0x40, 0x0, 0xC00, 0x300, 0xC0, 0x20, 0x0}},
    {20179, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0xFE0, 0x820, 0x820, 0x8A0, 0x844, 0x804, 0x804, 0x7FC, 0x0}},
    {20180, 2, {0x800, 0xBF8, 0x808, 0x1010, 0x1020, 0x3040, 0x3040, 0x57FE, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080}},
    {20181, 2, {0x840, 0x840, 0x840, 0x1040, 0x1040, 0x3FFE, 0x3040, 0x5040, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000, 0x1000}},
    {20182, 2, {0x840, 0x840, 0xA40, 0x1258, 0x1268, 0x32C8, 0x3348, 0x5E48, 0x9248, 0x1248, 0x1258, 0x1242, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {20183, 2, {0x810, 0x810, 0x810, 0x1010, 0x13FE, 0x3010, 0x3010, 0x5110, 0x9110, 0x1090, 0x1090, 0x1050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {20184, 2, {0x810, 0x810, 0x810, 0x1010, 0x17FE, 0x3010, 0x3010, 0x5010, 0x9110, 0x1090, 0x1090, 0x1010, 0x1010, 0x1010, 0x1050, 0x1020}},
    {20185, 2, {0x1040, 0x1040, 0x1040, 0x2040, 0x2444, 0x6444, 0x6444, 0xA444, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2004, 0x2000}},
    {20186, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2108, 0xC106, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {20187, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x3040, 0x3040, 0x507E, 0x97C0, 0x1040, 0x1040, 0x1042, 0x1042, 0x1042, 0x103E, 0x1000}},
    {20188, 2, {0x800, 0x800, 0x800, 0x13FE, 0x1020, 0x3020, 0x3020, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000, 0x1000}},
    {20189, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006, 0x1FF0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x7FFC, 0x0}},
    {20190, 2, {0x800, 0x800, 0xBFC, 0x1044, 0x1044, 0x3244, 0x3244, 0x5244, 0x9484, 0x1084, 0x1084, 0x1104, 0x1104, 0x1204, 0x1428, 0x1810}},
    {20191, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x3040, 0x3040, 0x57FE, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20192, 2, {0x800, 0xBFC, 0x840, 0x1040, 0x1040, 0x3040, 0x3040, 0x5FFE, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20193, 2, {0x880, 0x880, 0x900, 0x11FE, 0x1200, 0x3400, 0x31F8, 0x5008, 0x9010, 0x1060, 0x1080, 0x1100, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {20194, 2, {0x880, 0x880, 0x900, 0x11FC, 0x1204, 0x3204, 0x3404, 0x5904, 0x9084, 0x1044, 0x1044, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {20195, 2, {0x890, 0x888, 0x888, 0x1080, 0x10BE, 0x37C0, 0x3080, 0x5040, 0x9040, 0x1040, 0x1020, 0x1022, 0x1012, 0x100A, 0x1006, 0x1002}},
    {20196, 2, {0x100, 0x100, 0x280, 0x440, 0xA20, 0x1110, 0x2108, 0xC006, 0x1FF0, 0x10, 0x20, 0x440, 0x280, 0x100, 0x80, 0x80}},
    {20197, 2, {0x10, 0x410, 0x2210, 0x2110, 0x2110, 0x2010, 0x2010, 0x2020, 0x2020, 0x2020, 0x2440, 0x2850, 0x3088, 0x2104, 0x202, 0x402}},
    {20198, 2, {0x820, 0x820, 0x820, 0x1020, 0x1020, 0x3128, 0x3124, 0x5124, 0x9222, 0x1222, 0x1222, 0x1422, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {20199, 2, {0x200, 0x200, 0x200, 0x3F8, 0x200, 0x200, 0x200, 0xFFFE, 0x100, 0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {20200, 2, {0x800, 0x800, 0xBFC, 0x1000, 0x1000, 0x3000, 0x3000, 0x51F8, 0x9000, 0x1000, 0x1000, 0x1000, 0x1000, 0x17FE, 0x1000, 0x1000}},
    {20201, 2, {0x840, 0x840, 0x840, 0x1040, 0x1040, 0x3040, 0x307C, 0x5040, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {20202, 2, {0x880, 0x848, 0xA48, 0x1208, 0x1208, 0x3110, 0x3110, 0x5110, 0x90A0, 0x10A0, 0x1040, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {20203, 2, {0x840, 0x840, 0x880, 0x1080, 0x1100, 0x3110, 0x3210, 0x5420, 0x9020, 0x1040, 0x1080, 0x1108, 0x1204, 0x17FE, 0x1202, 0x1000}},
    {20204, 2, {0x900, 0x880, 0x8BC, 0x1204, 0x1204, 0x3204, 0x3204, 0x5204, 0x9204, 0x1204, 0x1204, 0x1204, 0x1204, 0x1204, 0x1214, 0x1208}},
    {20205, 2, {0x800, 0x800, 0xBF8, 0x1088, 0x1088, 0x3288, 0x3188, 0x5088, 0x90C8, 0x10A8, 0x1098, 0x110E, 0x1108, 0x1208, 0x1450, 0x1820}},
    {20206, 2, {0x808, 0x83C, 0xBC0, 0x1200, 0x1200, 0x3200, 0x33F8, 0x5288, 0x9288, 0x1288, 0x1250, 0x1250, 0x1220, 0x1450, 0x1488, 0x1906}},
    {20207, 2, {0x840, 0x840, 0x840, 0x1148, 0x1144, 0x3242, 0x3242, 0x5448, 0x9048, 0x1048, 0x1010, 0x1010, 0x1020, 0x1040, 0x1180, 0x1600}},
    {20208, 2, {0x800, 0x880, 0xB3C, 0x1224, 0x1224, 0x3224, 0x3224, 0x5224, 0x9224, 0x1224, 0x12B4, 0x1328, 0x1220, 0x1020, 0x1020, 0x1020}},
    {20209, 2, {0x840, 0x840, 0x8A0, 0x10A0, 0x1110, 0x3288, 0x3446, 0x5040, 0x9000, 0x13F8, 0x1008, 0x1010, 0x1010, 0x1020, 0x1020, 0x1040}},
    {20210, 2, {0x840, 0x840, 0x840, 0x1040, 0x17FC, 0x3444, 0x3444, 0x5444, 0x9444, 0x1444, 0x17FC, 0x1444, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20211, 2, {0x820, 0xA20, 0xA20, 0x1222, 0x1222, 0x3224, 0x33A8, 0x5230, 0x9220, 0x1220, 0x1220, 0x1222, 0x12A2, 0x1322, 0x121E, 0x1000}},
    {20212, 2, {0x800, 0x9FC, 0x904, 0x1104, 0x1104, 0x31FC, 0x3104, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x1204, 0x1204, 0x1414, 0x1808}},
    {20213, 2, {0x1100, 0x1100, 0x1100, 0x23F8, 0x2240, 0x6240, 0x6440, 0xA040, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040}},
    {20214, 2, {0x820, 0x820, 0x920, 0x1120, 0x11FC, 0x3220, 0x3220, 0x5420, 0x9020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {20215, 2, {0x840, 0x840, 0x8A0, 0x10A0, 0x1110, 0x3208, 0x3406, 0x5110, 0x9110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {20216, 2, {0x810, 0x878, 0xFC0, 0x1040, 0x1040, 0x3040, 0x3040, 0x57FE, 0x9040, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {20217, 2, {0x840, 0x840, 0x840, 0x17FC, 0x1040, 0x3040, 0x3040, 0x53FC, 0x9040, 0x1040, 0x1040, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20218, 2, {0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {20219, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x3040, 0x3040, 0x5FFE, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000}},
    {20220, 2, {0x800, 0x800, 0xBFE, 0x1020, 0x1020, 0x3020, 0x3020, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {20221, 2, {0x810, 0x890, 0x890, 0x1088, 0x1108, 0x3104, 0x3204, 0x55FA, 0x9088, 0x1088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {20222, 2, {0x800, 0xFFE, 0x880, 0x1080, 0x1080, 0x30F8, 0x3108, 0x5108, 0x9108, 0x1208, 0x13F8, 0x1010, 0x1010, 0x1010, 0x17FE, 0x1000}},
    {20223, 2, {0x880, 0x840, 0x840, 0x17FE, 0x1100, 0x3100, 0x3100, 0x51F8, 0x9108, 0x1108, 0x1108, 0x1108, 0x1208, 0x1208, 0x1450, 0x1820}},
    {20224, 2, {0x820, 0x920, 0x920, 0x1110, 0x1110, 0x3248, 0x3248, 0x5444, 0x9882, 0x1080, 0x1110, 0x1108, 0x1208, 0x17FC, 0x1204, 0x1000}},
    {20225, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {20226, 2, {0x1040, 0x1040, 0x1040, 0x2FFE, 0x2040, 0x6040, 0x67FC, 0xA444, 0x2444, 0x2444, 0x2444, 0x2454, 0x2448, 0x2040, 0x2040, 0x2040}},
    {20227, 2, {0x800, 0xBF8, 0x808, 0x1110, 0x10A0, 0x3040, 0x37FE, 0x5042, 0x9044, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080}},
    {20228, 2, {0x800, 0xBFC, 0x844, 0x1044, 0x1044, 0x33FC, 0x3240, 0x5240, 0x9440, 0x17FE, 0x1042, 0x1042, 0x104A, 0x1044, 0x1040, 0x1040}},
    {20229, 2, {0x840, 0x840, 0x840, 0x17FC, 0x1040, 0x3248, 0x3248, 0x5248, 0x9248, 0x13F8, 0x1048, 0x1040, 0x1042, 0x1042, 0x103E, 0x1000}},
    {20230, 2, {0x900, 0x900, 0x900, 0x11FC, 0x12A4, 0x34A4, 0x30A4, 0x5124, 0x9124, 0x1244, 0x1444, 0x1084, 0x1104, 0x1204, 0x1428, 0x1010}},
    {20231, 2, {0x800, 0x9F0, 0x910, 0x1110, 0x1110, 0x320E, 0x3400, 0x53F8, 0x9108, 0x1110, 0x1090, 0x10A0, 0x1040, 0x10A0, 0x1318, 0x1C06}},
    {20232, 2, {0x800, 0x880, 0x840, 0x1020, 0x1120, 0x3100, 0x3100, 0x5504, 0x9502, 0x1502, 0x1502, 0x1908, 0x1108, 0x1108, 0x10F8, 0x1000}},
    {20233, 2, {0x880, 0x840, 0x840, 0x13FC, 0x1000, 0x3000, 0x31F0, 0x5110, 0x9110, 0x1110, 0x1110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {20234, 2, {0x800, 0xBF8, 0x888, 0x1088, 0x1088, 0x37FE, 0x3088, 0x5088, 0x9088, 0x13F8, 0x1088, 0x1080, 0x1100, 0x1100, 0x1200, 0x1400}},
    {20235, 2, {0x1000, 0x17F8, 0x1108, 0x2110, 0x2110, 0x6120, 0x613C, 0xA104, 0x2104, 0x2288, 0x2288, 0x2250, 0x2420, 0x2450, 0x2888, 0x2306}},
    {20236, 2, {0x800, 0xBFE, 0xA00, 0x1200, 0x12F8, 0x3288, 0x3288, 0x5288, 0x9288, 0x12A8, 0x1290, 0x1282, 0x1482, 0x1482, 0x187E, 0x1000}},
    {20237, 2, {0x800, 0xBFC, 0x840, 0x1040, 0x1040, 0x3040, 0x33F8, 0x5088, 0x9088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {20238, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1040, 0x3040, 0x3040, 0x57F8, 0x9208, 0x1208, 0x1110, 0x10A0, 0x1040, 0x10A0, 0x1118, 0x1606}},
    {20239, 2, {0x840, 0x850, 0x848, 0x1048, 0x1040, 0x37FE, 0x3040, 0x5040, 0x90A0, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {20240, 2, {0x890, 0x888, 0x888, 0x1080, 0x10BE, 0x37C0, 0x3080, 0x5044, 0x9044, 0x1048, 0x1030, 0x1022, 0x1052, 0x118A, 0x1606, 0x1002}},
    {20241, 2, {0x840, 0x840, 0x840, 0x1040, 0x17FE, 0x3040, 0x30E0, 0x50E0, 0x9150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040, 0x1040}},
    {20242, 2, {0x808, 0x81C, 0x9E0, 0x1100, 0x1100, 0x3100, 0x31FE, 0x5110, 0x9110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {20243, 2, {0x800, 0x800, 0xBFE, 0x1020, 0x1020, 0x3040, 0x3040, 0x50D0, 0x9148, 0x1244, 0x1444, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20244, 2, {0x800, 0x800, 0xBFE, 0x1202, 0x1404, 0x3000, 0x31F0, 0x5110, 0x9110, 0x1110, 0x1110, 0x1112, 0x1212, 0x1212, 0x140E, 0x1800}},
    {20245, 2, {0x840, 0x840, 0x840, 0x13FC, 0x1040, 0x3040, 0x3040, 0x57FE, 0x9040, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {20246, 2, {0x880, 0x880, 0x880, 0x1080, 0x17FE, 0x3080, 0x3100, 0x51FC, 0x9144, 0x1244, 0x1248, 0x1450, 0x1420, 0x1850, 0x1088, 0x1306}},
    {20247, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2828, 0xC826, 0x820, 0x820, 0x820, 0x1450, 0x1250, 0x2288, 0x4104, 0x8202}},
    {20248, 2, {0x890, 0x888, 0x888, 0x1080, 0x17FE, 0x30A0, 0x30A0, 0x50A0, 0x90A0, 0x10A0, 0x1120, 0x1122, 0x1122, 0x1222, 0x121E, 0x1400}},
    {20249, 2, {0x840, 0x840, 0x840, 0x1040, 0x1244, 0x3244, 0x3248, 0x5450, 0x9040, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {20250, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x0, 0x7FFC, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {20251, 2, {0x800, 0xBFC, 0xA00, 0x1208, 0x1288, 0x3250, 0x3250, 0x5220, 0x9220, 0x1250, 0x1250, 0x1288, 0x1308, 0x1200, 0x13FE, 0x1000}},
    {20252, 2, {0x880, 0x880, 0xBF0, 0x1090, 0x1090, 0x3112, 0x3112, 0x524E, 0x9440, 0x1040, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20253, 2, {0x800, 0x800, 0xBFC, 0x1000, 0x1000, 0x3000, 0x37FE, 0x5040, 0x9040, 0x1080, 0x1080, 0x1108, 0x1204, 0x17FE, 0x1202, 0x1000}},
    {20254, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x1110, 0x910, 0x920, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {20255, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1040, 0x3040, 0x33FC, 0x5040, 0x9040, 0x17FE, 0x1042, 0x1042, 0x104A, 0x1044, 0x1040, 0x1040}},
    {20256, 2, {0x840, 0x840, 0x840, 0x13F8, 0x1040, 0x3080, 0x37FE, 0x5080, 0x9100, 0x13F8, 0x1008, 0x1110, 0x10A0, 0x1040, 0x1020, 0x1020}},
    {20257, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1080, 0x30A0, 0x3120, 0x53FC, 0x9020, 0x1020, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {20258, 2, {0x800, 0xBFC, 0x810, 0x1110, 0x1110, 0x3110, 0x3210, 0x53FE, 0x9030, 0x1050, 0x1090, 0x1110, 0x1210, 0x1410, 0x1050, 0x1020}},
    {20259, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1248, 0x3248, 0x3248, 0x5248, 0x9248, 0x12A8, 0x12A8, 0x1120, 0x1120, 0x1222, 0x1422, 0x181E}},
    {20260, 2, {0x880, 0x880, 0x900, 0x11FE, 0x1200, 0x3440, 0x3040, 0x53FC, 0x9044, 0x1044, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {20261, 2, {0x900, 0x908, 0x908, 0x1110, 0x1120, 0x3140, 0x3100, 0x57FE, 0x9140, 0x1120, 0x1120, 0x1110, 0x1108, 0x1144, 0x1182, 0x1100}},
    {20262, 2, {0x840, 0x840, 0x8A0, 0x10A0, 0x1110, 0x3208, 0x3406, 0x5110, 0x9120, 0x1140, 0x1180, 0x1100, 0x1104, 0x1104, 0x10FC, 0x1000}},
    {20263, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x3404, 0x3BF2, 0x5210, 0x9210, 0x1210, 0x1250, 0x1220, 0x1204, 0x1204, 0x11FC, 0x1000}},
    {20264, 2, {0x900, 0x900, 0x9FC, 0x1204, 0x1204, 0x3504, 0x3884, 0x5084, 0x9014, 0x1024, 0x1044, 0x1384, 0x1104, 0x1004, 0x1028, 0x1010}},
    {20265, 2, {0x880, 0x840, 0x840, 0x17FC, 0x1110, 0x3110, 0x3110, 0x5110, 0x9110, 0x10A0, 0x10A0, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1C06}},
    {20266, 2, {0x840, 0xA40, 0x940, 0x1140, 0x1040, 0x37FC, 0x3044, 0x5044, 0x9084, 0x10A4, 0x1094, 0x1114, 0x1104, 0x1204, 0x1428, 0x1810}},
    {20267, 2, {0x840, 0x820, 0x820, 0x1000, 0x13FE, 0x3202, 0x3404, 0x5000, 0x9000, 0x1000, 0x1000, 0x1000, 0x1000, 0x13FE, 0x1000, 0x1000}},
    {20268, 2, {0x800, 0x9FC, 0x904, 0x1104, 0x1104, 0x3104, 0x31FC, 0x5124, 0x9120, 0x1120, 0x1110, 0x1110, 0x1208, 0x1208, 0x1404, 0x1802}},
    {20269, 2, {0x880, 0x840, 0x840, 0x17FE, 0x1040, 0x3080, 0x3108, 0x5210, 0x97E0, 0x1040, 0x1080, 0x1108, 0x1204, 0x17FE, 0x1202, 0x1000}},
    {20270, 2, {0x1200, 0x1200, 0x127C, 0x2224, 0x2FA4, 0x64A4, 0x64A4, 0xA4A8, 0x24A8, 0x2928, 0x2510, 0x2210, 0x2528, 0x2948, 0x3084, 0x2102}},
    {20271, 2, {0x840, 0x840, 0x880, 0x13FC, 0x1204, 0x3204, 0x3204, 0x5204, 0x93FC, 0x1204, 0x1204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20272, 2, {0x840, 0x840, 0x840, 0x1040, 0x17FE, 0x3040, 0x3040, 0x5040, 0x93F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20273, 2, {0x840, 0x840, 0x8A0, 0x10A0, 0x1110, 0x3208, 0x3446, 0x5040, 0x9150, 0x1148, 0x1248, 0x1244, 0x1444, 0x1040, 0x1140, 0x1080}},
    {20274, 2, {0x800, 0xBFC, 0xA04, 0x1204, 0x1204, 0x33FC, 0x3200, 0x5240, 0x9244, 0x1248, 0x1270, 0x1240, 0x1442, 0x1442, 0x183E, 0x1000}},
    {20275, 2, {0x848, 0xA48, 0xA48, 0x1248, 0x1248, 0x37FE, 0x3248, 0x5248, 0x9248, 0x1248, 0x1278, 0x1200, 0x1200, 0x1200, 0x13FE, 0x1000}},
    {20276, 2, {0x840, 0xA44, 0x948, 0x1150, 0x1040, 0x37FC, 0x3040, 0x5040, 0x9040, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20277, 2, {0x1000, 0x1000, 0x17FC, 0x24A4, 0x24A4, 0x64A4, 0x64A4, 0xA4A4, 0x24A4, 0x2524, 0x251C, 0x2604, 0x2404, 0x2404, 0x27FC, 0x2404}},
    {20278, 2, {0x840, 0x840, 0x8A0, 0x10A0, 0x1110, 0x3248, 0x3426, 0x5020, 0x93F8, 0x1008, 0x1010, 0x1110, 0x10A0, 0x1040, 0x1020, 0x1020}},
    {20279, 2, {0x1040, 0x1040, 0x1040, 0x2040, 0x27FC, 0x6444, 0x6444, 0xA444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2404}},
    {20280, 2, {0x840, 0x840, 0x840, 0x17FC, 0x1444, 0x3444, 0x3444, 0x57FC, 0x9444, 0x1444, 0x1444, 0x17FC, 0x1444, 0x1040, 0x1040, 0x1040}},
    {20281, 2, {0x800, 0x9F8, 0x908, 0x1108, 0x1108, 0x31F8, 0x3108, 0x5108, 0x9108, 0x11F8, 0x1108, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {20282, 2, {0x800, 0xBFC, 0x804, 0x1004, 0x17F4, 0x3004, 0x3004, 0x53E4, 0x9224, 0x1224, 0x1224, 0x13E4, 0x1004, 0x1004, 0x1014, 0x1008}},
    {20283, 2, {0x800, 0xBF8, 0x840, 0x1040, 0x1248, 0x3148, 0x3150, 0x5040, 0x97FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20284, 2, {0x1008, 0x1108, 0x1088, 0x2448, 0x2448, 0x6408, 0x6408, 0xA408, 0x2410, 0x2410, 0x2410, 0x2528, 0x2624, 0x2444, 0x2082, 0x2102}},
    {20285, 2, {0x1400, 0x1400, 0x1400, 0x243C, 0x2FA4, 0x64A4, 0x64A4, 0xA4A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x28A4, 0x28BC, 0x32A4, 0x2100}},
    {20286, 2, {0x800, 0xBFC, 0x820, 0x1020, 0x1040, 0x3040, 0x30D0, 0x5148, 0x9244, 0x1444, 0x1040, 0x1040, 0x1040, 0x1000, 0x17FE, 0x1000}},
    {20287, 2, {0x800, 0x800, 0xBFC, 0x1204, 0x1204, 0x3204, 0x3204, 0x5204, 0x93FC, 0x1204, 0x1000, 0x1090, 0x1088, 0x1104, 0x1202, 0x1402}},
    {20288, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1208, 0x3208, 0x33F8, 0x5200, 0x9200, 0x13FC, 0x1204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20289, 2, {0x840, 0x840, 0x880, 0x1080, 0x1110, 0x3208, 0x37FC, 0x5004, 0x9000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20290, 2, {0x800, 0x800, 0xBFE, 0x1020, 0x1020, 0x3020, 0x3120, 0x5120, 0x913C, 0x1120, 0x1120, 0x1120, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {20291, 2, {0x1000, 0x1000, 0x17FC, 0x2444, 0x2444, 0x6444, 0x6444, 0xA7FC, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2404, 0x2000}},
    {20292, 2, {0x908, 0x908, 0x908, 0x1108, 0x17FE, 0x3108, 0x3108, 0x5108, 0x9108, 0x11F8, 0x1108, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {20293, 2, {0x840, 0x840, 0x840, 0x13FC, 0x1040, 0x3040, 0x3040, 0x57FE, 0x90E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040}},
    {20294, 2, {0x800, 0x800, 0xBFC, 0x1204, 0x1204, 0x3204, 0x33FC, 0x5204, 0x9204, 0x1204, 0x13FC, 0x1204, 0x1000, 0x1000, 0x17FE, 0x1000}},
    {20295, 2, {0x840, 0x820, 0x820, 0x13FE, 0x1202, 0x3404, 0x3000, 0x5000, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {20296, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1080, 0x30A0, 0x3120, 0x51FC, 0x9324, 0x1524, 0x1124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {20297, 2, {0x840, 0x840, 0x840, 0x1040, 0x13FC, 0x3040, 0x3040, 0x5040, 0x97FE, 0x1040, 0x1080, 0x1080, 0x1110, 0x1208, 0x17FC, 0x1204}},
    {20298, 2, {0x820, 0x820, 0x820, 0x13FE, 0x1222, 0x3224, 0x3220, 0x53FC, 0x9284, 0x1288, 0x1248, 0x1250, 0x1220, 0x1450, 0x1488, 0x1906}},
    {20299, 2, {0x800, 0xBFC, 0x884, 0x1084, 0x1084, 0x3104, 0x3114, 0x5208, 0x9400, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20300, 2, {0x1120, 0x1120, 0x1120, 0x2522, 0x2522, 0x65E4, 0x6528, 0xA530, 0x2520, 0x2520, 0x2520, 0x2522, 0x25E2, 0x3E22, 0x281E, 0x2000}},
    {20301, 2, {0x880, 0x840, 0x840, 0x1000, 0x17FC, 0x3000, 0x3008, 0x5208, 0x9208, 0x1110, 0x1110, 0x1110, 0x1120, 0x1020, 0x1FFE, 0x1000}},
    {20302, 2, {0x808, 0x83C, 0xBE0, 0x1220, 0x1220, 0x3220, 0x3220, 0x53FE, 0x9220, 0x1210, 0x1210, 0x1212, 0x120A, 0x128A, 0x1326, 0x1212}},
    {20303, 2, {0x880, 0x840, 0x800, 0x17FC, 0x1040, 0x3040, 0x3040, 0x5040, 0x97FC, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {20304, 2, {0x880, 0x880, 0x880, 0x1080, 0x17FE, 0x3080, 0x3100, 0x5100, 0x91FC, 0x1220, 0x1220, 0x1420, 0x1420, 0x1820, 0x13FE, 0x1000}},
    {20305, 2, {0x840, 0x840, 0x840, 0x1040, 0x17FE, 0x3080, 0x3080, 0x5100, 0x93FC, 0x1504, 0x1904, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20306, 2, {0x840, 0x840, 0x840, 0x13F8, 0x1248, 0x3248, 0x3248, 0x5248, 0x9248, 0x17FE, 0x1040, 0x10A0, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {20307, 2, {0x840, 0x840, 0x840, 0x1040, 0x17FC, 0x3040, 0x30E0, 0x50E0, 0x9150, 0x1150, 0x1248, 0x15F4, 0x1842, 0x1040, 0x1040, 0x1040}},
    {20308, 2, {0x840, 0x840, 0x840, 0x107E, 0x1040, 0x3040, 0x3040, 0x5040, 0x93FC, 0x1204, 0x1204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20309, 2, {0x800, 0x800, 0x17FE, 0x1008, 0x2008, 0x33C8, 0x5248, 0x9248, 0x1248, 0x1248, 0x13C8, 0x1248, 0x1008, 0x1008, 0x1028, 0x1010}},
    {20310, 2, {0x1000, 0x1080, 0x1040, 0x2028, 0x2128, 0x6108, 0x6110, 0xA514, 0x2522, 0x2522, 0x2542, 0x2988, 0x2108, 0x2308, 0x24F8, 0x2800}},
    {20311, 2, {0x840, 0x820, 0x820, 0x13FE, 0x1202, 0x3404, 0x3100, 0x5108, 0x9110, 0x1120, 0x11C0, 0x1102, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {20312, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x0, 0x7FFC, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {20313, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2FE8, 0xC106, 0x100, 0x3FF8, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {20314, 2, {0x1040, 0x1040, 0x1240, 0x2240, 0x23FC, 0x6240, 0x6440, 0xA040, 0x2FFE, 0x2040, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x2404, 0x2802}},
    {20315, 2, {0x1120, 0x1120, 0x1120, 0x2FFC, 0x2124, 0x6124, 0x67FC, 0xA520, 0x2920, 0x2FFE, 0x2122, 0x2122, 0x222A, 0x2224, 0x2420, 0x2820}},
    {20316, 2, {0x900, 0x900, 0x900, 0x11FE, 0x1280, 0x3280, 0x3480, 0x50F8, 0x9080, 0x1080, 0x1080, 0x10FC, 0x1080, 0x1080, 0x1080, 0x1080}},
    {20317, 2, {0x1100, 0x1100, 0x1200, 0x23FC, 0x2404, 0x6804, 0x63E4, 0xA224, 0x2224, 0x2224, 0x2224, 0x23E4, 0x2224, 0x2004, 0x2028, 0x2010}},
    {20318, 2, {0x1000, 0x13F8, 0x1000, 0x2000, 0x2FFE, 0x6080, 0x6080, 0xAFFE, 0x2108, 0x2210, 0x2610, 0x21A0, 0x2060, 0x2090, 0x2308, 0x2C04}},
    {20319, 2, {0x880, 0x880, 0x8F8, 0x1108, 0x1310, 0x34A0, 0x3040, 0x50A0, 0x9318, 0x1C06, 0x10C0, 0x1020, 0x1010, 0x1180, 0x1060, 0x1010}},
    {20320, 2, {0x880, 0x880, 0x880, 0x11FE, 0x1102, 0x3204, 0x3420, 0x5020, 0x9128, 0x1124, 0x1224, 0x1222, 0x1422, 0x1020, 0x10A0, 0x1040}},
    {20321, 2, {0x1040, 0x1040, 0x10A0, 0x20A0, 0x2110, 0x6208, 0x6444, 0xA842, 0x2040, 0x2444, 0x2444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2004}},
    {20322, 2, {0x800, 0xBFE, 0xA00, 0x1200, 0x1200, 0x33FC, 0x3204, 0x5204, 0x9204, 0x1204, 0x13FC, 0x1200, 0x1200, 0x1200, 0x13FE, 0x1000}},
    {20323, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x2444, 0x67FC, 0x6444, 0xA444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x2444, 0x2844, 0x2854, 0x3008}},
    {20324, 2, {0x800, 0x800, 0xFFC, 0x1100, 0x1100, 0x31F0, 0x3110, 0x5110, 0x9190, 0x1250, 0x1250, 0x1210, 0x1212, 0x1292, 0x1312, 0x120E}},
    {20325, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2FE8, 0xC006, 0x210, 0x1110, 0x910, 0x920, 0x820, 0x40, 0x7FFC, 0x0}},
    {20326, 2, {0x800, 0x800, 0xFFE, 0x1040, 0x1040, 0x3080, 0x3080, 0x51F8, 0x9308, 0x1508, 0x1908, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {20327, 2, {0x840, 0x840, 0x840, 0x107C, 0x1040, 0x3040, 0x3040, 0x57FE, 0x9040, 0x1040, 0x1050, 0x1048, 0x1044, 0x1040, 0x1040, 0x1040}},
    {20328, 2, {0x880, 0x880, 0x9FC, 0x1104, 0x1204, 0x35F4, 0x3114, 0x5114, 0x9114, 0x11F4, 0x1104, 0x1128, 0x1112, 0x1102, 0x10FE, 0x1000}},
    {20329, 2, {0x1000, 0x17FC, 0x1404, 0x2404, 0x25F4, 0x6444, 0x6444, 0xA5F4, 0x2554, 0x2554, 0x2554, 0x2554, 0x2574, 0x2446, 0x2446, 0x2842}},
    {20330, 2, {0x800, 0x800, 0xBFC, 0x1204, 0x1204, 0x32F4, 0x3294, 0x5294, 0x9294, 0x1294, 0x12F4, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1000}},
    {20331, 2, {0x880, 0x880, 0x9F8, 0x1208, 0x1510, 0x30A0, 0x3040, 0x50A0, 0x9318, 0x1C06, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20332, 2, {0x840, 0x840, 0x844, 0x13F4, 0x1048, 0x3050, 0x37FE, 0x5040, 0x9080, 0x1184, 0x1298, 0x14E0, 0x1882, 0x1082, 0x107E, 0x1000}},
    {20333, 2, {0x900, 0x900, 0x9F8, 0x1310, 0x14A0, 0x3040, 0x31B0, 0x564E, 0x9040, 0x13FC, 0x1040, 0x1440, 0x17FE, 0x1040, 0x1040, 0x1040}},
    {20334, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x3406, 0x33F8, 0x5000, 0x9000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20335, 2, {0xA08, 0x908, 0x910, 0x1000, 0x17FC, 0x3040, 0x3040, 0x53F8, 0x9040, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20336, 2, {0x800, 0xBFE, 0x820, 0x1020, 0x1040, 0x31FC, 0x3104, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20337, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0xEFEE, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE}},
    {20338, 2, {0x880, 0x880, 0x8FC, 0x1104, 0x1288, 0x3450, 0x3020, 0x5040, 0x91FC, 0x1304, 0x1D04, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20339, 2, {0x840, 0x840, 0xBFC, 0x1040, 0x1040, 0x3040, 0x3FFE, 0x5000, 0x9040, 0x1040, 0x13FC, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {20340, 2, {0x800, 0xBFE, 0x908, 0x1108, 0x1108, 0x31F8, 0x3108, 0x5108, 0x91F8, 0x1108, 0x1108, 0x111E, 0x17E8, 0x1008, 0x1008, 0x1008}},
    {20341, 2, {0x904, 0x884, 0x888, 0x1000, 0x13FE, 0x3088, 0x3088, 0x5088, 0x9088, 0x17FE, 0x1088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {20342, 2, {0x840, 0x840, 0x840, 0x1FFE, 0x1040, 0x3040, 0x37FC, 0x5000, 0x9000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20343, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3208, 0x3208, 0x53F8, 0x9244, 0x1248, 0x1230, 0x1220, 0x1210, 0x1288, 0x1306, 0x1200}},
    {20344, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x37FE, 0x3040, 0x5040, 0x9040, 0x13F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20345, 2, {0x1100, 0x1100, 0x13F8, 0x2208, 0x2410, 0x6BFE, 0x6200, 0xA2F8, 0x2288, 0x2288, 0x22A8, 0x2290, 0x2282, 0x2482, 0x247E, 0x2800}},
    {20346, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x3406, 0x33F8, 0x5040, 0x9040, 0x1040, 0x13F8, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {20347, 2, {0x890, 0x890, 0x890, 0x1492, 0x1294, 0x3098, 0x3090, 0x5198, 0x9294, 0x1492, 0x1090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {20348, 2, {0x1080, 0x1040, 0x1040, 0x27FC, 0x2000, 0x6110, 0x6208, 0xA404, 0x2110, 0x2110, 0x20A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {20349, 2, {0x1020, 0x1020, 0x1420, 0x227E, 0x2242, 0x6084, 0x6010, 0xA110, 0x2110, 0x2E10, 0x2228, 0x2228, 0x2248, 0x2244, 0x2084, 0x2102}},
    {20350, 2, {0x810, 0x910, 0x908, 0x1208, 0x1404, 0x3BFA, 0x3208, 0x5208, 0x93F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x1228, 0x1210}},
    {20351, 2, {0x1040, 0x1040, 0x1FFE, 0x2040, 0x2040, 0x67FC, 0x6444, 0xA444, 0x27FC, 0x2040, 0x2240, 0x2140, 0x2080, 0x2140, 0x2230, 0x2C0E}},
    {20352, 2, {0x1004, 0x1FC4, 0x1484, 0x2494, 0x2494, 0x6494, 0x6494, 0xAFD4, 0x2494, 0x2494, 0x2494, 0x2494, 0x2484, 0x2884, 0x2894, 0x3088}},
    {20353, 2, {0x1040, 0x1240, 0x1240, 0x23FC, 0x2440, 0x6840, 0x6040, 0xA7FE, 0x2120, 0x2120, 0x2120, 0x2120, 0x2222, 0x2222, 0x241E, 0x2800}},
    {20354, 2, {0x840, 0x820, 0xBFE, 0x1200, 0x1200, 0x320C, 0x32F0, 0x5220, 0x9220, 0x123E, 0x13E0, 0x1220, 0x1422, 0x1422, 0x181E, 0x1000}},
    {20355, 2, {0x1000, 0x17FC, 0x1404, 0x2404, 0x2404, 0x67FC, 0x6000, 0xA248, 0x2248, 0x2248, 0x2248, 0x2248, 0x224A, 0x244A, 0x244A, 0x2806}},
    {20356, 2, {0x1000, 0x17FE, 0x1040, 0x2080, 0x2110, 0x6208, 0x67FC, 0xA044, 0x2040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x2000}},
    {20357, 2, {0x840, 0x820, 0x820, 0x17FE, 0x1040, 0x3084, 0x3108, 0x53F0, 0x9022, 0x1044, 0x1188, 0x1610, 0x1020, 0x1050, 0x1188, 0x1604}},
    {20358, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x1110, 0x1110, 0x2928, 0x4544, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100, 0x100}},
    {20359, 2, {0x1040, 0x1040, 0x17FE, 0x2040, 0x27FC, 0x6044, 0x63FC, 0xA240, 0x2440, 0x27FE, 0x2042, 0x20AA, 0x20A4, 0x2110, 0x2208, 0x2C06}},
    {20360, 2, {0x1080, 0x1080, 0x11F8, 0x2210, 0x2520, 0x60C0, 0x6090, 0xA320, 0x2C7C, 0x2084, 0x2108, 0x2690, 0x2060, 0x2040, 0x2180, 0x2E00}},
    {20361, 2, {0x1040, 0x1040, 0x17FC, 0x20A0, 0x2110, 0x6208, 0x6DF6, 0xA000, 0x27FC, 0x2080, 0x2100, 0x23F8, 0x2008, 0x2008, 0x2050, 0x2020}},
    {20362, 2, {0x840, 0x844, 0xA44, 0x1148, 0x1150, 0x3040, 0x37FE, 0x5090, 0x9090, 0x1090, 0x1090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {20363, 2, {0x1004, 0x1004, 0x17C4, 0x2214, 0x2214, 0x63D4, 0x6254, 0xA454, 0x2654, 0x2554, 0x2894, 0x2094, 0x2104, 0x2204, 0x2414, 0x2808}},
    {20364, 2, {0x100, 0x280, 0x440, 0x920, 0x3098, 0xCFE6, 0x40, 0x80, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x820, 0x1010, 0x3FF8, 0x8}},
    {20365, 2, {0x840, 0x840, 0x840, 0x13FC, 0x1040, 0x3040, 0x37FE, 0x5010, 0x9010, 0x17FE, 0x1010, 0x1210, 0x1110, 0x1110, 0x1050, 0x1020}},
    {20366, 2, {0x840, 0x840, 0xA48, 0x1148, 0x1150, 0x3040, 0x37FE, 0x5040, 0x90E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040}},
    {20367, 2, {0x840, 0xA40, 0xA40, 0x13FC, 0x1440, 0x3040, 0x3040, 0x57FE, 0x90E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040}},
    {20368, 2, {0x840, 0x840, 0x880, 0x1100, 0x13FC, 0x3294, 0x3294, 0x5294, 0x9294, 0x1294, 0x1294, 0x1294, 0x1294, 0x1294, 0x1FFE, 0x1000}},
    {20369, 2, {0x840, 0x840, 0xFFE, 0x1080, 0x1100, 0x31FC, 0x3304, 0x5504, 0x99FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {20370, 2, {0x840, 0x820, 0x820, 0x13FE, 0x1202, 0x3444, 0x3040, 0x53FE, 0x9088, 0x1088, 0x1108, 0x10D0, 0x1020, 0x1050, 0x1088, 0x1304}},
    {20371, 2, {0x840, 0x840, 0xBF8, 0x1048, 0x17FE, 0x3048, 0x33F8, 0x5040, 0x9040, 0x17FC, 0x1040, 0x1040, 0x1FFE, 0x1040, 0x1040, 0x1040}},
    {20372, 2, {0x840, 0x840, 0x890, 0x1108, 0x13FC, 0x3024, 0x3120, 0x5120, 0x91FC, 0x1220, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {20373, 2, {0x1000, 0x17FC, 0x1040, 0x2040, 0x2080, 0x67FC, 0x64A4, 0xA4A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x2484, 0x2414, 0x2408}},
    {20374, 2, {0x100, 0x280, 0x440, 0x820, 0x37D8, 0xC006, 0x0, 0x3FF8, 0x2488, 0x2488, 0x2488, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {20375, 2, {0x1000, 0x17FC, 0x1404, 0x2404, 0x25F4, 0x6404, 0x6404, 0xA5F4, 0x2514, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2404, 0x2414, 0x2408}},
    {20376, 2, {0x840, 0x820, 0x820, 0x13FE, 0x1202, 0x3404, 0x3038, 0x53C0, 0x9040, 0x1040, 0x107E, 0x17C0, 0x1040, 0x1042, 0x1042, 0x103E}},
    {20377, 2, {0x828, 0x824, 0x824, 0x1020, 0x17FE, 0x3020, 0x3020, 0x57E0, 0x9120, 0x1110, 0x1110, 0x1110, 0x11CA, 0x170A, 0x1206, 0x1002}},
    {20378, 2, {0x880, 0x880, 0x900, 0x11FC, 0x1204, 0x3404, 0x33E4, 0x5224, 0x9224, 0x13E4, 0x1224, 0x1224, 0x13E4, 0x1004, 0x1028, 0x1010}},
    {20379, 2, {0x910, 0x910, 0x910, 0x1110, 0x13FC, 0x3110, 0x3110, 0x5110, 0x9110, 0x17FE, 0x1000, 0x1110, 0x1108, 0x1208, 0x1404, 0x1804}},
    {20380, 2, {0x880, 0x900, 0xBF8, 0x1208, 0x1288, 0x3248, 0x3248, 0x5FFE, 0x9208, 0x1288, 0x1248, 0x1248, 0x1208, 0x1208, 0x1428, 0x1810}},
    {20381, 2, {0x880, 0x840, 0x840, 0x17FE, 0x1080, 0x3080, 0x3140, 0x5144, 0x9348, 0x1530, 0x1920, 0x1110, 0x1108, 0x1144, 0x1182, 0x1100}},
    {20382, 2, {0x900, 0x900, 0x900, 0x111E, 0x17D2, 0x3252, 0x3252, 0x5252, 0x9252, 0x1492, 0x1292, 0x1112, 0x1292, 0x125E, 0x1452, 0x1800}},
    {20383, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1080, 0x3080, 0x317C, 0x5108, 0x9310, 0x1510, 0x11FE, 0x1110, 0x1110, 0x1110, 0x1150, 0x1120}},
    {20384, 2, {0x840, 0x840, 0x840, 0x17FC, 0x1040, 0x3248, 0x3150, 0x5040, 0x9FFE, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {20385, 2, {0x1000, 0x17FE, 0x10A0, 0x20A0, 0x20A0, 0x67FC, 0x64A4, 0xA4A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x27FC, 0x2404, 0x2000}},
    {20386, 2, {0x800, 0xFFC, 0x840, 0x1040, 0x13F8, 0x3248, 0x3248, 0x53F8, 0x9248, 0x1248, 0x1FFE, 0x1208, 0x1208, 0x1208, 0x1228, 0x1210}},
    {20387, 2, {0x1000, 0x13F8, 0x1208, 0x2208, 0x2208, 0x6208, 0x63F8, 0xA000, 0x2000, 0x27FC, 0x2404, 0x2404, 0x2404, 0x2404, 0x27FC, 0x2404}},
    {20388, 2, {0x880, 0x888, 0xBE8, 0x1090, 0x10A0, 0x37FE, 0x3080, 0x5100, 0x93FC, 0x1480, 0x1900, 0x11F8, 0x1008, 0x1008, 0x1050, 0x1020}},
    {20389, 2, {0x880, 0x880, 0x8BC, 0x13C0, 0x1050, 0x3024, 0x30D4, 0x530C, 0x9000, 0x17FE, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {20390, 2, {0x840, 0x840, 0x87E, 0x1040, 0x1040, 0x33F8, 0x3208, 0x5248, 0x9248, 0x1248, 0x1248, 0x1248, 0x10A0, 0x1110, 0x1208, 0x1404}},
    {20391, 2, {0x1004, 0x17C4, 0x1444, 0x2454, 0x2554, 0x6554, 0x6554, 0xA554, 0x2554, 0x2554, 0x2554, 0x2104, 0x2284, 0x2244, 0x2414, 0x2808}},
    {20392, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x37FE, 0x30A0, 0x5110, 0x9208, 0x1516, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {20393, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x3406, 0x31F0, 0x5000, 0x9000, 0x17FC, 0x1040, 0x1080, 0x1110, 0x1208, 0x17FC, 0x1204}},
    {20394, 2, {0x880, 0x840, 0xFFE, 0x1208, 0x1110, 0x30A0, 0x3040, 0x51B0, 0x960E, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {20395, 2, {0x880, 0x840, 0x1FFE, 0x1200, 0x1200, 0x3200, 0x33FC, 0x5080, 0x9080, 0x1FFE, 0x1108, 0x1208, 0x1390, 0x1060, 0x1198, 0x1604}},
    {20396, 2, {0x1040, 0x1040, 0x1040, 0x27FC, 0x2484, 0x6888, 0x6140, 0xA144, 0x2348, 0x2530, 0x2920, 0x2110, 0x2108, 0x2144, 0x2182, 0x2100}},
    {20397, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3220, 0x3220, 0x5210, 0x9208, 0x14C6, 0x1420, 0x1800, 0x11C0, 0x1030, 0x1008, 0x1000}},
    {20398, 2, {0x900, 0x900, 0x9FC, 0x1200, 0x15F8, 0x3108, 0x3148, 0x5128, 0x97FE, 0x1108, 0x1248, 0x1228, 0x13FC, 0x1008, 0x1050, 0x1020}},
    {20399, 2, {0x800, 0xBF8, 0x808, 0x1008, 0x17FE, 0x3100, 0x3100, 0x53F8, 0x9440, 0x1040, 0x1FFE, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1C06}},
    {20400, 2, {0x800, 0xBF8, 0x888, 0x1088, 0x17FE, 0x3088, 0x3088, 0x53F8, 0x9100, 0x1100, 0x13F8, 0x1508, 0x1908, 0x1108, 0x11F8, 0x1108}},
    {20401, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1208, 0x33F8, 0x3000, 0x5000, 0x97FC, 0x1040, 0x1040, 0x13F8, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {20402, 2, {0x800, 0xBFC, 0xA00, 0x1200, 0x12F8, 0x3200, 0x3200, 0x53FC, 0x92A0, 0x12A4, 0x12A8, 0x1290, 0x1290, 0x1488, 0x14A4, 0x18C2}},
    {20403, 2, {0x840, 0x840, 0xA48, 0x1248, 0x1248, 0x3554, 0x38E2, 0x5040, 0x9040, 0x17FC, 0x1040, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {20404, 2, {0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x3F04, 0x924, 0x924, 0x3F24, 0x1124, 0x1124, 0x17A4, 0x7804, 0x200C}},
    {20405, 2, {0x800, 0xBF8, 0x808, 0x11F8, 0x1008, 0x33F8, 0x3000, 0x57FC, 0x9404, 0x13F0, 0x1110, 0x1110, 0x10A0, 0x1040, 0x11B0, 0x160E}},
    {20406, 2, {0x800, 0x9F8, 0x908, 0x1108, 0x1108, 0x3108, 0x31F8, 0x5040, 0x9080, 0x13FC, 0x1204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20407, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3200, 0x3200, 0x53FC, 0x9204, 0x12F4, 0x1294, 0x1294, 0x14F4, 0x1404, 0x1828, 0x1010}},
    {20408, 2, {0x800, 0xFFC, 0x800, 0x1000, 0x13F8, 0x3208, 0x3208, 0x5208, 0x9208, 0x13F8, 0x1000, 0x1208, 0x1108, 0x1110, 0x1FFE, 0x1000}},
    {20409, 2, {0x1000, 0x100C, 0x1770, 0x2110, 0x2110, 0x6210, 0x627C, 0xA710, 0x2110, 0x2510, 0x2510, 0x227C, 0x2200, 0x2500, 0x28FE, 0x3000}},
    {20410, 2, {0x840, 0xA48, 0xA48, 0x1248, 0x13F8, 0x3040, 0x30A0, 0x5110, 0x9288, 0x1446, 0x1040, 0x13F0, 0x1010, 0x1020, 0x1020, 0x1040}},
    {20411, 2, {0xA08, 0x908, 0x910, 0x1020, 0x13F8, 0x3208, 0x3208, 0x5208, 0x93F8, 0x10A0, 0x10A0, 0x1120, 0x1122, 0x1222, 0x141E, 0x1800}},
    {20412, 2, {0x820, 0x820, 0xBFE, 0x1020, 0x1020, 0x33FE, 0x3202, 0x5404, 0x91F8, 0x1010, 0x1020, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {20413, 2, {0xBF8, 0xA48, 0xA48, 0x13F8, 0x1248, 0x3248, 0x33F8, 0x5080, 0x9080, 0x17FC, 0x1084, 0x1104, 0x1104, 0x1204, 0x1428, 0x1810}},
    {20414, 2, {0x840, 0x844, 0xBF4, 0x1048, 0x1048, 0x37FE, 0x3020, 0x5040, 0x93F8, 0x1110, 0x1220, 0x15FE, 0x1820, 0x1020, 0x10A0, 0x1040}},
    {20415, 2, {0x1000, 0x17FE, 0x1040, 0x27FC, 0x2444, 0x6444, 0x67FC, 0xA444, 0x2444, 0x27FC, 0x2240, 0x2140, 0x2080, 0x2140, 0x2230, 0x2C0E}},
    {20416, 2, {0x808, 0x83C, 0xBC0, 0x1044, 0x1224, 0x3128, 0x3100, 0x5040, 0x93FE, 0x1088, 0x1108, 0x1390, 0x1060, 0x1050, 0x1188, 0x1604}},
    {20417, 2, {0x800, 0x8F8, 0x888, 0x1088, 0x12F8, 0x3200, 0x3200, 0x53FC, 0x9044, 0x1044, 0x17FE, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {20418, 2, {0x804, 0x81E, 0xBE0, 0x1080, 0x1108, 0x33F0, 0x3020, 0x5048, 0x9184, 0x17FE, 0x1022, 0x1128, 0x1124, 0x1222, 0x14A2, 0x1040}},
    {20419, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1208, 0x3208, 0x33F8, 0x5040, 0x9040, 0x1240, 0x127C, 0x1240, 0x1240, 0x1540, 0x18FE, 0x1000}},
    {20420, 2, {0x1028, 0x11A4, 0x1724, 0x2120, 0x2120, 0x6FFE, 0x6120, 0xA124, 0x21A4, 0x2328, 0x2D28, 0x2110, 0x2112, 0x212A, 0x2546, 0x2282}},
    {20421, 2, {0x850, 0x848, 0x848, 0x1040, 0x17FE, 0x3040, 0x3244, 0x5164, 0x9168, 0x10D0, 0x1150, 0x1248, 0x1444, 0x1042, 0x1140, 0x1080}},
    {20422, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x35F6, 0x3040, 0x5040, 0x97FC, 0x1040, 0x1250, 0x1248, 0x1444, 0x1844, 0x1140, 0x1080}},
    {20423, 2, {0x800, 0xC40, 0xABE, 0x1108, 0x1288, 0x3488, 0x3088, 0x50BE, 0x9188, 0x1288, 0x1488, 0x1088, 0x1088, 0x1088, 0x153E, 0x1200}},
    {20424, 2, {0x840, 0xA40, 0xA40, 0x13FC, 0x1440, 0x3840, 0x3040, 0x57FE, 0x9000, 0x1000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20425, 2, {0x800, 0xFFC, 0x880, 0x1080, 0x13F8, 0x3108, 0x3108, 0x57FE, 0x9000, 0x1000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20426, 2, {0x840, 0x840, 0x888, 0x1104, 0x13FE, 0x3002, 0x3088, 0x5144, 0x9242, 0x10F8, 0x1188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {20427, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3000, 0x37FC, 0x5444, 0x9444, 0x1444, 0x17FC, 0x1400, 0x1402, 0x1402, 0x13FE, 0x1000}},
    {20428, 2, {0x1048, 0x1044, 0x17FE, 0x2040, 0x2040, 0x67FC, 0x6444, 0xA444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x2454, 0x2408}},
    {20429, 2, {0x880, 0x840, 0xBF8, 0x1208, 0x1208, 0x33F8, 0x3208, 0x5208, 0x93F8, 0x1240, 0x1244, 0x1228, 0x1210, 0x1288, 0x1306, 0x1200}},
    {20430, 2, {0x1000, 0x11F8, 0x1108, 0x2908, 0x2508, 0x45F8, 0x9108, 0x1108, 0x1108, 0x11F8, 0x2908, 0x2508, 0x4508, 0x8108, 0x7FE, 0x0}},
    {20431, 2, {0x840, 0xA48, 0x948, 0x1150, 0x1040, 0x33F8, 0x3208, 0x5208, 0x93F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x1228, 0x1210}},
    {20432, 2, {0x1084, 0x11C4, 0x1704, 0x2114, 0x2114, 0x6114, 0x6FD4, 0xA114, 0x2314, 0x2394, 0x2554, 0x2504, 0x2904, 0x2104, 0x2114, 0x2108}},
    {20433, 2, {0x1000, 0x17F8, 0x1010, 0x21A0, 0x2040, 0x67FC, 0x6444, 0xA444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x2454, 0x2408}},
    {20434, 2, {0x840, 0x820, 0xBFE, 0x1202, 0x1404, 0x31F8, 0x3000, 0x5000, 0x93FE, 0x1090, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {20435, 2, {0x800, 0xBFC, 0x800, 0x1124, 0x1248, 0x3490, 0x3248, 0x5124, 0x9000, 0x13FC, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {20436, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3208, 0x3208, 0x53F8, 0x9208, 0x1208, 0x13F8, 0x1120, 0x1120, 0x1222, 0x1422, 0x181E}},
    {20437, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2844, 0x6040, 0x6040, 0xA7FC, 0x20E0, 0x2150, 0x2150, 0x2248, 0x2444, 0x2842, 0x2040, 0x2040}},
    {20438, 2, {0x1000, 0x17FC, 0x1020, 0x2040, 0x20D0, 0x6148, 0x6644, 0xA040, 0x2000, 0x23F8, 0x2208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {20439, 2, {0x910, 0x908, 0xA04, 0x1444, 0x1040, 0x30A0, 0x3110, 0x5208, 0x9C06, 0x13F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20440, 2, {0x808, 0x83C, 0xBC0, 0x1004, 0x1244, 0x3128, 0x3000, 0x53F8, 0x9010, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {20441, 2, {0x1208, 0x1110, 0x10E0, 0x2110, 0x2208, 0x6080, 0x6FFE, 0xA140, 0x2240, 0x27FC, 0x2A44, 0x2244, 0x2254, 0x2248, 0x2040, 0x2040}},
    {20442, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x27FC, 0x6444, 0x6444, 0xA7FC, 0x2040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x2000}},
    {20443, 2, {0x900, 0x900, 0x9F0, 0x1210, 0x1220, 0x37FC, 0x3A44, 0x5244, 0x9244, 0x13FC, 0x12A4, 0x10A0, 0x1120, 0x1222, 0x1422, 0x181E}},
    {20444, 2, {0x1040, 0x1040, 0x17FC, 0x2444, 0x27FC, 0x6444, 0x67FC, 0xA000, 0x3FFE, 0x2100, 0x2200, 0x23FC, 0x2004, 0x2004, 0x2028, 0x2010}},
    {20445, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1208, 0x33F8, 0x3040, 0x5040, 0x97FC, 0x10E0, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040}},
    {20446, 2, {0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2248, 0x2208, 0x2A28, 0x2410}},
    {20447, 2, {0x1040, 0x1080, 0x1110, 0x2208, 0x27FC, 0x6104, 0x6100, 0xA3F8, 0x2440, 0x2040, 0x2FFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {20448, 2, {0x840, 0x840, 0x840, 0x17FE, 0x1040, 0x3248, 0x3248, 0x5248, 0x9554, 0x18A2, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {20449, 2, {0x840, 0x820, 0xBFE, 0x1000, 0x1000, 0x31FC, 0x3000, 0x5000, 0x91FC, 0x1000, 0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20450, 2, {0x880, 0x880, 0x9F8, 0x1310, 0x14A0, 0x3040, 0x31B0, 0x5E4E, 0x9080, 0x1320, 0x1044, 0x1188, 0x1610, 0x1020, 0x10C0, 0x1700}},
    {20451, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3000, 0x3000, 0x57FC, 0x9040, 0x1040, 0x1FFE, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {20452, 2, {0x908, 0x888, 0x890, 0x13FC, 0x1024, 0x3024, 0x33FC, 0x5220, 0x9220, 0x13FE, 0x1062, 0x10A2, 0x112A, 0x1224, 0x1420, 0x1020}},
    {20453, 2, {0x1040, 0x1040, 0x1FFE, 0x2040, 0x27FC, 0x6444, 0x6444, 0xA7FC, 0x2444, 0x2444, 0x27FC, 0x2040, 0x3FFE, 0x2040, 0x2040, 0x2040}},
    {20454, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x23F8, 0x6080, 0x67FE, 0xA090, 0x2110, 0x21FE, 0x2210, 0x2290, 0x2450, 0x2810, 0x2050, 0x2020}},
    {20455, 2, {0x820, 0x820, 0x820, 0x13FE, 0x1020, 0x3020, 0x3020, 0x51FC, 0x9040, 0x1020, 0x10A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {20456, 2, {0x800, 0xBFC, 0x890, 0x1294, 0x1198, 0x3090, 0x33FE, 0x5200, 0x9200, 0x1200, 0x1200, 0x1200, 0x1200, 0x1400, 0x1400, 0x1800}},
    {20457, 2, {0x1000, 0x1FFE, 0x1120, 0x2120, 0x2120, 0x6FFC, 0x6924, 0xA924, 0x2924, 0x2AA4, 0x2A54, 0x2C4C, 0x2884, 0x2804, 0x2814, 0x2808}},
    {20458, 2, {0x1000, 0x1FFE, 0x1000, 0x2000, 0x27BC, 0x64A4, 0x64A4, 0xA4A4, 0x26B4, 0x25AC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x25AC}},
    {20459, 2, {0x840, 0x840, 0x840, 0x17FC, 0x1040, 0x3248, 0x3148, 0x5150, 0x9FFE, 0x10E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040}},
    {20460, 2, {0x1088, 0x11C8, 0x1F08, 0x2108, 0x2108, 0x6FD0, 0x6110, 0xA310, 0x2390, 0x2554, 0x2564, 0x2922, 0x217E, 0x2122, 0x2102, 0x2100}},
    {20461, 2, {0x840, 0x840, 0x8A0, 0x10A0, 0x1110, 0x3208, 0x35F6, 0x5000, 0x9088, 0x1048, 0x1248, 0x1150, 0x1110, 0x1020, 0x17FE, 0x1000}},
    {20462, 2, {0x1080, 0x1080, 0x11F8, 0x2108, 0x2A90, 0x6C60, 0xA998, 0x2E26, 0x28C0, 0x2B10, 0x2820, 0x28C8, 0x2B10, 0x2060, 0x2180, 0x2600}},
    {20463, 2, {0x1040, 0x1020, 0x17FE, 0x2400, 0x2488, 0x6488, 0x6508, 0xA57E, 0x2708, 0x2548, 0x2528, 0x2528, 0x2908, 0x2908, 0x3128, 0x2110}},
    {20464, 2, {0x1000, 0x1180, 0x1700, 0x211E, 0x2112, 0x6FD2, 0x6112, 0xA312, 0x2392, 0x2352, 0x2552, 0x2512, 0x291E, 0x2112, 0x2100, 0x2100}},
    {20465, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3208, 0x33F8, 0x5208, 0x93F8, 0x1208, 0x1208, 0x1FFE, 0x1000, 0x1110, 0x1208, 0x1404}},
    {20466, 2, {0x1210, 0x1110, 0x1110, 0x27D0, 0x203E, 0x6292, 0x6252, 0xA452, 0x2892, 0x2292, 0x2112, 0x2112, 0x22A2, 0x24A2, 0x284A, 0x2084}},
    {20467, 2, {0x890, 0x890, 0x890, 0x179E, 0x1090, 0x3090, 0x3090, 0x539C, 0x9090, 0x1090, 0x1090, 0x179E, 0x1090, 0x1090, 0x1090, 0x1090}},
    {20468, 2, {0x10A0, 0x1090, 0x10BC, 0x27C8, 0x2050, 0x6064, 0x6194, 0xAE0C, 0x20A0, 0x2090, 0x20BC, 0x27C8, 0x2050, 0x2064, 0x2194, 0x2E0C}},
    {20469, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x2040, 0x63F8, 0x6040, 0xA040, 0x2FFE, 0x2088, 0x2150, 0x2320, 0x2510, 0x2948, 0x2186, 0x2100}},
    {20470, 2, {0x900, 0x900, 0x9DC, 0x1114, 0x1114, 0x3114, 0x37D4, 0x5014, 0x9114, 0x1114, 0x1588, 0x1548, 0x1948, 0x1114, 0x1514, 0x1222}},
    {20471, 2, {0x1000, 0x17BE, 0x14AA, 0x24AA, 0x24AA, 0x67AA, 0x64AA, 0xA4AA, 0x24BE, 0x27A0, 0x24A0, 0x24A0, 0x24A2, 0x28A2, 0x2AA2, 0x311E}},
    {20472, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x13F8, 0x3080, 0x37FC, 0x5110, 0x9248, 0x1446, 0x11F0, 0x1040, 0x17FC, 0x1040, 0x1040, 0x1040}},
    {20473, 2, {0x800, 0xFFE, 0x890, 0x1090, 0x1090, 0x339C, 0x3204, 0x5204, 0x9204, 0x139C, 0x1090, 0x1090, 0x1090, 0x1090, 0x17FE, 0x1000}},
    {20474, 2, {0x840, 0x880, 0xBFC, 0x1110, 0x1248, 0x3446, 0x33F8, 0x5248, 0x9248, 0x13F8, 0x1248, 0x1248, 0x13F8, 0x1042, 0x1042, 0x103E}},
    {20475, 2, {0x880, 0x880, 0x9F8, 0x1208, 0x1510, 0x30E0, 0x3318, 0x5C06, 0x93F8, 0x1248, 0x1248, 0x13F8, 0x1248, 0x1248, 0x13F8, 0x1208}},
    {20476, 2, {0x880, 0x840, 0xFFE, 0x1100, 0x1208, 0x37FC, 0x3004, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1208, 0x1228, 0x1210}},
    {20477, 2, {0x1010, 0x1090, 0x1710, 0x243E, 0x2422, 0x6444, 0x6790, 0xA510, 0x2510, 0x2510, 0x2528, 0x2528, 0x2528, 0x2944, 0x2944, 0x3082}},
    {20478, 2, {0x840, 0x880, 0xBFC, 0x1224, 0x1224, 0x33FC, 0x3224, 0x5244, 0x93FC, 0x1090, 0x1110, 0x17FE, 0x1010, 0x1010, 0x1010, 0x1010}},
    {20479, 2, {0x840, 0x840, 0x87C, 0x1040, 0x13FC, 0x3244, 0x3270, 0x53C0, 0x9244, 0x123C, 0x1200, 0x12F0, 0x1490, 0x1492, 0x1912, 0x120E}},
    {20480, 2, {0x800, 0x9FC, 0x900, 0x11F8, 0x1100, 0x31F8, 0x3100, 0x5FFE, 0x9280, 0x1248, 0x1250, 0x1220, 0x1210, 0x1288, 0x1306, 0x1200}},
    {20481, 2, {0x900, 0x900, 0x900, 0x13DE, 0x1292, 0x3492, 0x3092, 0x5092, 0x93F2, 0x1092, 0x1092, 0x1152, 0x115E, 0x1220, 0x1420, 0x1800}},
    {20482, 2, {0x884, 0x884, 0x908, 0x13DE, 0x1108, 0x3108, 0x3108, 0x5108, 0x97BE, 0x1108, 0x1108, 0x1108, 0x1108, 0x1208, 0x1208, 0x1408}},
    {20483, 2, {0x1110, 0x1110, 0x13D8, 0x2254, 0x2452, 0x6A90, 0x6110, 0xA2FE, 0x2C00, 0x2000, 0x23F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {20484, 2, {0x908, 0x890, 0x860, 0x1198, 0x1644, 0x3040, 0x37FE, 0x5080, 0x91FC, 0x1304, 0x15FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {20485, 2, {0x880, 0x840, 0xBFC, 0x1000, 0x1110, 0x3110, 0x32A8, 0x5444, 0x9000, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20486, 2, {0x1000, 0x17FC, 0x1040, 0x2040, 0x27FC, 0x6444, 0x6444, 0xA664, 0x2554, 0x2554, 0x26EC, 0x24CC, 0x2444, 0x2444, 0x2454, 0x2408}},
    {20487, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2A04, 0x6200, 0x63BC, 0xA4A4, 0x24A4, 0x26A4, 0x2934, 0x2128, 0x2222, 0x2222, 0x2422, 0x281E}},
    {20488, 2, {0x840, 0x840, 0x840, 0x17FC, 0x1040, 0x3248, 0x3248, 0x5248, 0x9554, 0x10E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040}},
    {20489, 2, {0x100, 0x280, 0x440, 0xA20, 0x3118, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x2FF8, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {20490, 2, {0x890, 0x890, 0x908, 0x1148, 0x1244, 0x3492, 0x3108, 0x53FC, 0x9044, 0x1020, 0x10A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {20491, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x27FC, 0x6444, 0x6444, 0xA5F4, 0x2514, 0x2514, 0x2514, 0x25F4, 0x2514, 0x2404, 0x27FC, 0x2404}},
    {20492, 2, {0x840, 0x820, 0xBFE, 0x1202, 0x1404, 0x31F8, 0x3108, 0x5108, 0x91F8, 0x1100, 0x1100, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20493, 2, {0x880, 0x840, 0xBFC, 0x1000, 0x1208, 0x3108, 0x3110, 0x57FE, 0x9000, 0x1000, 0x13FC, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20494, 2, {0x890, 0x890, 0x890, 0x13FC, 0x1294, 0x3294, 0x3294, 0x53FC, 0x9294, 0x1294, 0x1294, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {20495, 2, {0x1080, 0x1080, 0x11F8, 0x2108, 0x2A90, 0x6C60, 0xA998, 0x2E46, 0x2850, 0x2848, 0x2BFC, 0x2840, 0x28A0, 0x2110, 0x2208, 0x2404}},
    {20496, 2, {0x840, 0x840, 0x10FC, 0x1088, 0x3550, 0x5420, 0x94D8, 0x1706, 0x1420, 0x14A4, 0x14A4, 0x1528, 0x1450, 0x1048, 0x1084, 0x1302}},
    {20497, 2, {0x1000, 0x17BC, 0x14A4, 0x24A4, 0x27BC, 0x64A4, 0x64A4, 0xA7BC, 0x2404, 0x2404, 0x2404, 0x2404, 0x2404, 0x2404, 0x2414, 0x2408}},
    {20498, 2, {0x1004, 0x1004, 0x1FC4, 0x2214, 0x2214, 0x6494, 0x67D4, 0xA114, 0x2114, 0x27D4, 0x2114, 0x2114, 0x21C4, 0x2E04, 0x2414, 0x2008}},
    {20499, 2, {0x840, 0xA44, 0xA44, 0x1448, 0x10A0, 0x3110, 0x3608, 0x5044, 0x9040, 0x1248, 0x1248, 0x1450, 0x10A0, 0x1110, 0x1208, 0x1C06}},
    {20500, 2, {0x1000, 0x17FC, 0x1404, 0x2404, 0x27FC, 0x6420, 0x6524, 0xA524, 0x2524, 0x25FC, 0x2420, 0x2524, 0x2524, 0x2924, 0x29FC, 0x3004}},
    {20501, 2, {0x810, 0x878, 0xBC0, 0x1040, 0x1040, 0x37FC, 0x3248, 0x5248, 0x9FFE, 0x1248, 0x1248, 0x17FC, 0x1040, 0x1040, 0x17FC, 0x1000}},
    {20502, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x1040, 0x37FE, 0x3110, 0x50A0, 0x97FC, 0x1040, 0x1040, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20503, 2, {0x1000, 0x17BC, 0x14A4, 0x24A4, 0x24A4, 0x67BC, 0x64A4, 0xA4A4, 0x24A4, 0x27BC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x2AA4, 0x314C}},
    {20504, 2, {0x1040, 0x1248, 0x1148, 0x2150, 0x2040, 0x67FC, 0x6404, 0xA404, 0x25F4, 0x2514, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x2408}},
    {20505, 2, {0x1000, 0x13F8, 0x1008, 0x2808, 0x2FFE, 0x6900, 0x6900, 0xABF8, 0x2C40, 0x2840, 0x2FFE, 0x2840, 0x28A0, 0x2110, 0x2208, 0x2C06}},
    {20506, 2, {0x840, 0x840, 0xBFC, 0x10A0, 0x1110, 0x3208, 0x37FE, 0x5008, 0x93C8, 0x1248, 0x1248, 0x1248, 0x13C8, 0x1008, 0x1028, 0x1010}},
    {20507, 2, {0x908, 0x908, 0xFFE, 0x1108, 0x1108, 0x31F8, 0x3108, 0x5108, 0x91F8, 0x1108, 0x1108, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {20508, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x25F4, 0x6444, 0x6444, 0xA7FC, 0x2404, 0x25F4, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x2808}},
    {20509, 2, {0x810, 0x810, 0xFF90, 0x810, 0x7F10, 0x4110, 0x7F10, 0x4110, 0x7F10, 0x4910, 0x828, 0xFFA8, 0x848, 0x844, 0x884, 0x902}},
    {20510, 2, {0x880, 0x840, 0xFFE, 0x1000, 0x1000, 0x33F8, 0x3208, 0x5208, 0x93F8, 0x1040, 0x1250, 0x1248, 0x1444, 0x1844, 0x1140, 0x1080}},
    {20511, 2, {0x910, 0x910, 0x910, 0x17FC, 0x1110, 0x3110, 0x3FFE, 0x5000, 0x93F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20512, 2, {0x940, 0x920, 0x920, 0x13FE, 0x1220, 0x3620, 0x3BFC, 0x5220, 0x9220, 0x13FC, 0x1220, 0x1220, 0x1220, 0x13FE, 0x1200, 0x1200}},
    {20513, 2, {0x9F8, 0x908, 0x908, 0x11F8, 0x1108, 0x3108, 0x31F8, 0x5000, 0x93FC, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20514, 2, {0x1040, 0x1040, 0x17FE, 0x2040, 0x23F8, 0x6048, 0x67FE, 0xA048, 0x23F8, 0x2040, 0x2240, 0x227C, 0x2240, 0x2540, 0x24FE, 0x2800}},
    {20515, 2, {0x1410, 0x1210, 0x1210, 0x2FA0, 0x243E, 0x6444, 0x67A4, 0xA4A4, 0x24A8, 0x24A8, 0x2490, 0x2490, 0x2AA8, 0x2928, 0x3044, 0x2082}},
    {20516, 2, {0x820, 0x820, 0xBFE, 0x1040, 0x11FC, 0x3504, 0x3504, 0x55FC, 0x9504, 0x15FC, 0x1504, 0x1504, 0x15FC, 0x1400, 0x17FE, 0x1000}},
    {20517, 2, {0x840, 0x820, 0x820, 0x13FE, 0x1202, 0x3494, 0x3108, 0x5204, 0x9000, 0x11FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {20518, 2, {0x1040, 0x1248, 0x1150, 0x27FC, 0x2080, 0x6100, 0x6FFE, 0xA210, 0x2408, 0x2BF4, 0x3212, 0x2210, 0x2250, 0x2224, 0x2204, 0x21FC}},
    {20519, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2804, 0x63F0, 0x6000, 0xA000, 0x27FC, 0x2040, 0x2250, 0x2248, 0x2444, 0x2844, 0x2140, 0x2080}},
    {20520, 2, {0x800, 0xBFC, 0xA04, 0x1204, 0x13FC, 0x3220, 0x3220, 0x53FE, 0x9220, 0x1220, 0x12FC, 0x1284, 0x1284, 0x1484, 0x14FC, 0x1884}},
    {20521, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x13F8, 0x3040, 0x37FE, 0x5000, 0x93F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1228, 0x1210}},
    {20522, 2, {0x800, 0x840, 0xB9C, 0x1204, 0x1204, 0x339C, 0x3204, 0x5204, 0x93FC, 0x1090, 0x1090, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {20523, 2, {0x1040, 0x1040, 0x10A0, 0x2110, 0x2208, 0x6DF6, 0x6000, 0xA7FC, 0x24A4, 0x24A4, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x2414, 0x2408}},
    {20524, 2, {0x840, 0x840, 0x87E, 0x1040, 0x13FC, 0x3204, 0x33FC, 0x5204, 0x93FC, 0x1244, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1040}},
    {20525, 2, {0x838, 0xBC0, 0x840, 0x1040, 0x17FC, 0x3150, 0x3248, 0x5486, 0x9080, 0x17FC, 0x1110, 0x1210, 0x11A0, 0x1060, 0x1090, 0x1308}},
    {20526, 2, {0x800, 0xBF8, 0xA48, 0x1248, 0x13F8, 0x3248, 0x3248, 0x53F8, 0x9040, 0x17FC, 0x10E0, 0x1150, 0x1248, 0x1446, 0x1040, 0x1040}},
    {20527, 2, {0x1208, 0x1228, 0x1228, 0x2228, 0x2FA4, 0x6244, 0x6254, 0xA692, 0x2710, 0x2AA0, 0x2A20, 0x3228, 0x2244, 0x22FC, 0x2244, 0x2200}},
    {20528, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x2040, 0x6FFE, 0x6208, 0xA484, 0x2882, 0x21F8, 0x2208, 0x2510, 0x20A0, 0x2040, 0x21B0, 0x2E0E}},
    {20529, 2, {0x1000, 0x17FC, 0x1404, 0x2404, 0x27FC, 0x6404, 0x6404, 0xA7FC, 0x2000, 0x2420, 0x2422, 0x27AC, 0x2430, 0x2422, 0x25A2, 0x261E}},
    {20530, 2, {0x1040, 0x1040, 0x1FFE, 0x2040, 0x2040, 0x67FC, 0x6444, 0xA7FC, 0x2444, 0x27FC, 0x24E4, 0x2150, 0x2248, 0x2444, 0x3842, 0x2040}},
    {20531, 2, {0x1040, 0x1040, 0x1FFE, 0x2040, 0x27FC, 0x6444, 0x67FC, 0xA040, 0x27FC, 0x2044, 0x3FFE, 0x2044, 0x27FC, 0x2040, 0x2140, 0x2080}},
    {20532, 2, {0x840, 0x840, 0xFFE, 0x10A0, 0x1150, 0x3248, 0x35F6, 0x5040, 0x9150, 0x1110, 0x17FE, 0x1110, 0x1110, 0x1110, 0x1210, 0x1410}},
    {20533, 2, {0x810, 0x818, 0xBD4, 0x1014, 0x1010, 0x37FE, 0x3010, 0x5090, 0x9090, 0x12D0, 0x1290, 0x1290, 0x128A, 0x12EA, 0x1706, 0x1202}},
    {20534, 2, {0x800, 0xBFC, 0xA04, 0x1204, 0x13FC, 0x3204, 0x33FC, 0x5204, 0x9204, 0x13FC, 0x1000, 0x17FE, 0x1000, 0x1108, 0x1204, 0x1402}},
    {20535, 2, {0x840, 0x840, 0xFFC, 0x10A0, 0x1110, 0x3208, 0x35F6, 0x5000, 0x9000, 0x17FC, 0x1040, 0x1248, 0x1244, 0x1444, 0x1140, 0x1080}},
    {20536, 2, {0x808, 0x83C, 0xBC0, 0x1044, 0x1224, 0x3128, 0x3100, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x10A8, 0x1124, 0x1222, 0x1420, 0x1020}},
    {20537, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x35F6, 0x3040, 0x5040, 0x93F8, 0x1248, 0x1248, 0x13F8, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {20538, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x13F8, 0x3040, 0x37FE, 0x5000, 0x93F8, 0x1208, 0x1248, 0x1248, 0x1248, 0x10A0, 0x1110, 0x1608}},
    {20539, 2, {0x1000, 0x1FDE, 0x1492, 0x2492, 0x2494, 0x6794, 0x6498, 0xA494, 0x2792, 0x2492, 0x2492, 0x25DA, 0x2E94, 0x2090, 0x2090, 0x2090}},
    {20540, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x1040, 0x33F8, 0x3208, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1208, 0x1FFE, 0x1000}},
    {20541, 2, {0x840, 0x840, 0x8A0, 0x1110, 0x1208, 0x35F6, 0x3040, 0x5040, 0x97FC, 0x1040, 0x1040, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20542, 2, {0x1000, 0x10FE, 0x1410, 0x2420, 0x247C, 0x6444, 0x6754, 0xA454, 0x2454, 0x2454, 0x2554, 0x2654, 0x2420, 0x2028, 0x2044, 0x2082}},
    {20543, 2, {0x880, 0x840, 0xBFC, 0x1000, 0x1108, 0x3090, 0x37FE, 0x5040, 0x9040, 0x17FE, 0x1088, 0x1108, 0x1090, 0x1060, 0x1198, 0x1604}},
    {20544, 2, {0x1110, 0x1110, 0x17FC, 0x2110, 0x2150, 0x6040, 0x63F8, 0xA248, 0x2248, 0x2248, 0x2FFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2406}},
    {20545, 2, {0x103E, 0x17C0, 0x1084, 0x2444, 0x2208, 0x6040, 0x67FC, 0xA444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x3FFE, 0x2404, 0x2414, 0x2408}},
    {20546, 2, {0x908, 0x888, 0x890, 0x17FE, 0x1000, 0x33C4, 0x3254, 0x5254, 0x93D4, 0x1254, 0x1254, 0x13D4, 0x1254, 0x1244, 0x1254, 0x12C8}},
    {20547, 2, {0x800, 0xBFE, 0xA00, 0x12FC, 0x1284, 0x32FC, 0x3284, 0x52FC, 0x9220, 0x13FE, 0x1248, 0x12C8, 0x1230, 0x1248, 0x1284, 0x13FE}},
    {20548, 2, {0x1000, 0x17FC, 0x1040, 0x2080, 0x27FC, 0x64A4, 0x64A4, 0xA4A4, 0x2484, 0x244C, 0x2040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {20549, 2, {0x1008, 0x103C, 0x17C0, 0x2040, 0x2FFE, 0x6040, 0x67FC, 0xA444, 0x27FC, 0x2444, 0x27FC, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x2000}},
    {20550, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x23F8, 0x6080, 0x6FFE, 0xA110, 0x2208, 0x27FC, 0x2A0A, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8}},
    {20551, 2, {0x1000, 0x17BC, 0x1484, 0x2484, 0x2484, 0x67BC, 0x6400, 0xA400, 0x27BC, 0x2424, 0x2424, 0x27A8, 0x2410, 0x2428, 0x2444, 0x2482}},
    {20552, 2, {0x800, 0xBF8, 0xA08, 0x13F8, 0x1208, 0x33F8, 0x3100, 0x53FC, 0x9444, 0x1244, 0x12A4, 0x1204, 0x13F4, 0x1004, 0x1028, 0x1010}},
    {20553, 2, {0x880, 0x880, 0xBF8, 0x1108, 0x17FE, 0x3000, 0x33F8, 0x5208, 0x93F8, 0x1040, 0x17FC, 0x1040, 0x1440, 0x17FE, 0x1040, 0x1040}},
    {20554, 2, {0x804, 0x81E, 0xBE0, 0x1020, 0x11FC, 0x3124, 0x3124, 0x51FC, 0x9020, 0x13FE, 0x1222, 0x122A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {20555, 2, {0x800, 0xBFC, 0xA04, 0x1204, 0x13FC, 0x3200, 0x3284, 0x5248, 0x92FC, 0x1248, 0x1248, 0x13FE, 0x1448, 0x1448, 0x1888, 0x1108}},
    {20556, 2, {0x1108, 0x1108, 0x17FE, 0x2108, 0x2148, 0x6040, 0x67FE, 0xA080, 0x2100, 0x23FC, 0x2504, 0x2904, 0x2104, 0x2104, 0x21FC, 0x2104}},
    {20557, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x13F8, 0x3208, 0x3208, 0x53F8, 0x9000, 0x17FE, 0x1040, 0x1240, 0x1278, 0x1240, 0x1540, 0x18FE}},
    {20558, 2, {0x800, 0xBF8, 0xA48, 0x1248, 0x13F8, 0x3248, 0x3248, 0x53F8, 0x9000, 0x17FE, 0x1240, 0x1244, 0x1228, 0x1290, 0x1308, 0x1206}},
    {20559, 2, {0x1080, 0x1040, 0x17FC, 0x2404, 0x2404, 0x67FC, 0x6400, 0xA400, 0x27FC, 0x26A4, 0x2AA4, 0x2BFC, 0x2AA4, 0x2AA4, 0x32A4, 0x220C}},
    {20560, 2, {0x840, 0x820, 0xBFC, 0x1000, 0x1108, 0x3090, 0x33FE, 0x5210, 0x9220, 0x12C8, 0x1210, 0x1224, 0x12C8, 0x1410, 0x1460, 0x1980}},
    {20561, 2, {0x2000, 0x2FF8, 0x2808, 0x4868, 0x4B88, 0x4888, 0xC888, 0x4BE8, 0x4AA8, 0x4AA8, 0x4BE8, 0x488A, 0x48AA, 0x50EA, 0x5726, 0x6002}},
    {20562, 2, {0x800, 0xBF8, 0xA08, 0x13F8, 0x1208, 0x33F8, 0x3000, 0x5FFE, 0x9200, 0x13FC, 0x1494, 0x1924, 0x1244, 0x1484, 0x1128, 0x1210}},
    {20563, 2, {0x800, 0xBFC, 0xA04, 0x1204, 0x13FC, 0x3200, 0x33FC, 0x5240, 0x9288, 0x13FC, 0x1224, 0x1220, 0x15FC, 0x1420, 0x1820, 0x13FE}},
    {20564, 2, {0x800, 0xBDE, 0xA52, 0x1252, 0x13DE, 0x3000, 0x31FC, 0x5000, 0x97FE, 0x1080, 0x1100, 0x11FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {20565, 2, {0xA20, 0xA24, 0xBA8, 0x1230, 0x12A2, 0x3322, 0x325E, 0x5080, 0x93FC, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20566, 2, {0x840, 0x844, 0xBF4, 0x1048, 0x1050, 0x37FE, 0x3040, 0x5080, 0x91F8, 0x1308, 0x1508, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {20567, 2, {0x820, 0x820, 0x924, 0x1122, 0x122A, 0x3030, 0x30C0, 0x5700, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {20568, 2, {0x800, 0xBF8, 0xA08, 0x1208, 0x1208, 0x3208, 0x33F8, 0x5000, 0x9000, 0x17BC, 0x14A4, 0x14A4, 0x14A4, 0x14A4, 0x17BC, 0x14A4}},
    {20569, 2, {0x880, 0x840, 0xBFC, 0x1000, 0x1108, 0x3090, 0x37FE, 0x5442, 0x9844, 0x13FC, 0x1244, 0x1244, 0x1254, 0x1248, 0x1040, 0x1040}},
    {20570, 2, {0x1210, 0x1210, 0x1210, 0x2FA0, 0x223E, 0x6244, 0x6224, 0xAFA4, 0x28A4, 0x28A8, 0x28A8, 0x2890, 0x2FA8, 0x2828, 0x2044, 0x2082}},
    {20571, 2, {0x1008, 0x103C, 0x17C0, 0x2040, 0x2040, 0x6FFE, 0x6040, 0xA140, 0x265C, 0x2444, 0x2444, 0x275C, 0x2444, 0x2444, 0x27FC, 0x2404}},
    {20572, 2, {0x1080, 0x1040, 0x17FC, 0x2000, 0x23F8, 0x6208, 0x63F8, 0xA000, 0x2FFE, 0x2802, 0x23F8, 0x2040, 0x2040, 0x2040, 0x2140, 0x2080}},
    {20573, 2, {0x890, 0x892, 0xF9C, 0x1090, 0x1192, 0x368E, 0x3000, 0x53FC, 0x9204, 0x13FC, 0x1204, 0x13FC, 0x1204, 0x1204, 0x1214, 0x1208}},
    {20574, 2, {0x890, 0xA90, 0xA90, 0x17FE, 0x1290, 0x3290, 0x32F0, 0x5200, 0x93FC, 0x1040, 0x17FE, 0x10E0, 0x1150, 0x1248, 0x1C46, 0x1040}},
    {20575, 2, {0x840, 0x880, 0xBF8, 0x1208, 0x13F8, 0x3208, 0x33F8, 0x5000, 0x97FC, 0x1040, 0x1040, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x1000}},
    {20576, 2, {0x800, 0xBFE, 0x850, 0x1050, 0x13FE, 0x3252, 0x3252, 0x53FE, 0x9000, 0x1040, 0x17FE, 0x1088, 0x1190, 0x1060, 0x10D8, 0x1304}},
    {20577, 2, {0x908, 0x908, 0xFFE, 0x1108, 0x11F8, 0x3108, 0x31F8, 0x5108, 0x9108, 0x17FE, 0x1250, 0x1288, 0x1304, 0x1200, 0x13FC, 0x1000}},
    {20578, 2, {0x1390, 0x1E10, 0x1210, 0x2212, 0x2252, 0x6FD4, 0x6258, 0xA690, 0x2710, 0x2AA8, 0x2A28, 0x3228, 0x2248, 0x2244, 0x2284, 0x2302}},
    {20579, 2, {0x880, 0x840, 0xFFC, 0x1000, 0x1208, 0x3110, 0x3FFE, 0x5000, 0x93F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {20580, 2, {0x908, 0x888, 0x890, 0x17FE, 0x1090, 0x3090, 0x33FC, 0x5294, 0x9294, 0x130C, 0x1204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {20581, 2, {0x1020, 0x1020, 0x10FC, 0x2E24, 0x23FE, 0x6224, 0x64FC, 0xA420, 0x2EFC, 0x2220, 0x2220, 0x2BFE, 0x2420, 0x2620, 0x29FE, 0x3000}},
    {20582, 2, {0x800, 0xBFE, 0x822, 0x1120, 0x113C, 0x3120, 0x32FE, 0x5400, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {20583, 2, {0x1000, 0x1080, 0x17FC, 0x2110, 0x2208, 0x6C86, 0x61F8, 0xA610, 0x2160, 0x21A0, 0x2E7E, 0x2184, 0x2648, 0x2030, 0x20C0, 0x2F00}},
    {20584, 2, {0x1120, 0x1124, 0x15A8, 0x2530, 0x2524, 0x65A4, 0x6E1C, 0xA040, 0x2040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x2040}},
    {20585, 2, {0x900, 0x900, 0xBF8, 0x1208, 0x1410, 0x3BFC, 0x3204, 0x53FC, 0x9204, 0x13FC, 0x1204, 0x13FC, 0x1000, 0x1108, 0x1204, 0x1402}},
    {20586, 2, {0x1000, 0x17FC, 0x1000, 0x23F8, 0x2208, 0x6208, 0x63F8, 0xA000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2404}},
    {20587, 2, {0x1110, 0x1110, 0x1210, 0x247C, 0x2910, 0x6110, 0x62FE, 0xA608, 0x2A08, 0x22FE, 0x2208, 0x2248, 0x2228, 0x2208, 0x2228, 0x2210}},
    {20588, 2, {0x1100, 0x1100, 0x13FC, 0x2494, 0x2B94, 0x6164, 0x6254, 0xA484, 0x2128, 0x2210, 0x2040, 0x2124, 0x252A, 0x250A, 0x28F8, 0x2000}},
    {20589, 2, {0x800, 0xFFE, 0x840, 0x1080, 0x13FC, 0x3294, 0x3294, 0x52F4, 0x9294, 0x1294, 0x12F4, 0x1294, 0x1294, 0x1294, 0x13FC, 0x1204}},
    {20590, 2, {0x800, 0x9F8, 0x908, 0x1108, 0x11F8, 0x3000, 0x37FE, 0x5108, 0x91F8, 0x1108, 0x11F8, 0x1108, 0x113E, 0x17C8, 0x1008, 0x1008}},
    {20591, 2, {0x880, 0x840, 0xFFE, 0x1000, 0x13F8, 0x3208, 0x3208, 0x5208, 0x93F8, 0x10A4, 0x1128, 0x1310, 0x1510, 0x1948, 0x1186, 0x1100}},
    {20592, 2, {0x900, 0x93C, 0xFD4, 0x1114, 0x1394, 0x3114, 0x37D4, 0x5124, 0x914C, 0x1040, 0x17FE, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1C06}},
    {20593, 2, {0x103C, 0x17E0, 0x1420, 0x2420, 0x27FE, 0x6420, 0x65FC, 0xA504, 0x2504, 0x25FC, 0x2504, 0x25FC, 0x2904, 0x2904, 0x31FC, 0x2104}},
    {20594, 2, {0x800, 0xBFC, 0xA44, 0x1244, 0x13FC, 0x3244, 0x3244, 0x53FC, 0x9000, 0x1040, 0x1024, 0x1522, 0x150A, 0x1908, 0x10F8, 0x1000}},
    {20595, 2, {0x1040, 0x1444, 0x1444, 0x2444, 0x27FC, 0x6000, 0x6FFE, 0xA040, 0x2080, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x240C}},
    {20596, 2, {0x1002, 0x17C2, 0x1442, 0x2452, 0x27D2, 0x6452, 0x6452, 0xA7D2, 0x2452, 0x2452, 0x27D2, 0x2002, 0x2282, 0x2242, 0x244A, 0x2804}},
    {20597, 2, {0x1040, 0x107E, 0x1040, 0x2040, 0x27FC, 0x6404, 0x6404, 0xA7FC, 0x2404, 0x27FC, 0x2404, 0x2404, 0x27FC, 0x2110, 0x2208, 0x2404}},
    {20598, 2, {0x1000, 0x17FC, 0x1444, 0x2444, 0x27FC, 0x6444, 0x6444, 0xA7FC, 0x2040, 0x2FFE, 0x2842, 0x2852, 0x2BF2, 0x2912, 0x280A, 0x2804}},
    {20599, 2, {0x840, 0x8A0, 0x910, 0x1208, 0x15F6, 0x3000, 0x33C4, 0x5254, 0x9254, 0x13D4, 0x1254, 0x1254, 0x13D4, 0x1244, 0x1254, 0x12C8}},
    {20600, 2, {0x9C0, 0x8A0, 0x910, 0x1208, 0x15F6, 0x3000, 0x33CA, 0x524A, 0x9254, 0x13D4, 0x1268, 0x1254, 0x13D4, 0x124A, 0x124A, 0x12C0}},
    {20601, 2, {0x840, 0x840, 0x10FC, 0x1088, 0x3550, 0x5420, 0x94D8, 0x1706, 0x14F8, 0x14A8, 0x14A8, 0x14A8, 0x14F8, 0x10A8, 0x10A8, 0x10F8}},
    {20602, 2, {0x910, 0x910, 0xBD8, 0x1254, 0x1452, 0x3A90, 0x3110, 0x52FE, 0x9C00, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x13F8}},
    {20603, 2, {0x820, 0x924, 0x8A8, 0x1020, 0x13FE, 0x30A8, 0x3124, 0x5202, 0x9040, 0x17FE, 0x1088, 0x1108, 0x1190, 0x1060, 0x1198, 0x1604}},
    {20604, 2, {0x840, 0x840, 0x87E, 0x1040, 0x13FC, 0x3004, 0x31FC, 0x5004, 0x93FC, 0x1040, 0x1240, 0x127C, 0x1240, 0x1540, 0x14FE, 0x1800}},
    {20605, 2, {0x840, 0xA40, 0x940, 0x17F8, 0x1088, 0x3090, 0x30FC, 0x5104, 0x9108, 0x13FE, 0x1202, 0x142A, 0x1AAA, 0x1282, 0x1414, 0x1008}},
    {20606, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x2248, 0x6FFE, 0x6208, 0xA000, 0x27FC, 0x2404, 0x2444, 0x2444, 0x2444, 0x20B0, 0x230C, 0x2C02}},
    {20607, 2, {0x840, 0xA48, 0x950, 0x1040, 0x17FE, 0x3402, 0x3804, 0x51F0, 0x9000, 0x1000, 0x17FC, 0x1080, 0x1110, 0x1208, 0x17FC, 0x1204}},
    {20608, 2, {0x1040, 0x1080, 0x17FC, 0x2444, 0x2444, 0x67FC, 0x6444, 0xA484, 0x27FC, 0x2080, 0x2150, 0x2164, 0x227C, 0x2442, 0x2842, 0x303E}},
    {20609, 2, {0x1140, 0x165C, 0x1444, 0x2444, 0x275C, 0x6444, 0x6444, 0xA7FC, 0x2040, 0x27FC, 0x2208, 0x2110, 0x20A0, 0x2040, 0x21B0, 0x2E0E}},
    {20610, 2, {0x1004, 0x103E, 0x17E0, 0x243E, 0x2420, 0x65FE, 0x6522, 0xA538, 0x25E2, 0x251E, 0x2500, 0x2978, 0x2A48, 0x324A, 0x248A, 0x2906}},
    {20611, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x13F8, 0x3040, 0x37FE, 0x5080, 0x9110, 0x13E0, 0x1048, 0x17FC, 0x1044, 0x1248, 0x1544, 0x1080}},
    {20612, 2, {0x1040, 0x1040, 0x1FFE, 0x2444, 0x22A8, 0x6110, 0x6208, 0xAC06, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208}},
    {20613, 2, {0x850, 0x848, 0xFFE, 0x1040, 0x17FC, 0x3444, 0x37FC, 0x5444, 0x97FC, 0x1444, 0x1008, 0x1FFE, 0x1208, 0x1108, 0x1128, 0x1010}},
    {20614, 2, {0x1000, 0x17FE, 0x1420, 0x2440, 0x25FC, 0x6504, 0x65FC, 0xA504, 0x25FC, 0x2524, 0x2420, 0x24A8, 0x2924, 0x2A22, 0x30A0, 0x2040}},
    {20615, 2, {0x908, 0x908, 0xFFE, 0x1108, 0x1000, 0x37FE, 0x3108, 0x51F8, 0x9108, 0x11F8, 0x1108, 0x113E, 0x17C8, 0x1008, 0x1008, 0x1008}},
    {20616, 2, {0x1000, 0x17FC, 0x10A0, 0x20A0, 0x27FC, 0x64A4, 0x64A4, 0xA7FC, 0x2040, 0x2040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {20617, 2, {0x800, 0xBFE, 0xA00, 0x12FC, 0x1200, 0x33FE, 0x32A8, 0x5290, 0x92C8, 0x1286, 0x1208, 0x13FE, 0x1288, 0x1448, 0x1408, 0x1818}},
    {20618, 2, {0x800, 0x9F8, 0x908, 0x1108, 0x11F8, 0x3000, 0x33FC, 0x5204, 0x93FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1090, 0x1108, 0x1204}},
    {20619, 2, {0x910, 0x910, 0xFFC, 0x1110, 0x13F8, 0x3110, 0x37FC, 0x5040, 0x93F8, 0x1248, 0x13F8, 0x1248, 0x17FE, 0x1208, 0x1228, 0x1210}},
    {20620, 2, {0x800, 0xBFC, 0xA40, 0x13F8, 0x1240, 0x33F8, 0x3240, 0x5240, 0x93FC, 0x1004, 0x1554, 0x1554, 0x1554, 0x1404, 0x1028, 0x1010}},
    {20621, 2, {0x1080, 0x1040, 0x17FC, 0x2208, 0x2110, 0x67FE, 0x6402, 0xA884, 0x2040, 0x27FC, 0x2100, 0x21F8, 0x2108, 0x2208, 0x2228, 0x2410}},
    {20622, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x23F8, 0x6208, 0x63F8, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2FFE, 0x2110, 0x2208, 0x2404}},
    {20623, 2, {0x1040, 0x1020, 0x17FE, 0x2420, 0x25FC, 0x6424, 0x67FE, 0xA424, 0x25FC, 0x2420, 0x25FC, 0x2504, 0x2504, 0x2904, 0x29FC, 0x3104}},
    {20624, 2, {0x1080, 0x1040, 0x17FE, 0x2000, 0x23F8, 0x6208, 0x63F8, 0xA000, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x25F4, 0x2404, 0x2414, 0x2408}},
    {20625, 2, {0x1208, 0x1208, 0x13BE, 0x2488, 0x2AA8, 0x613E, 0x6208, 0xA448, 0x2848, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x2040}},
    {20626, 2, {0x83C, 0xFC0, 0xA44, 0x1128, 0x1080, 0x3110, 0x33E0, 0x5040, 0x9188, 0x13FC, 0x1044, 0x1040, 0x17FC, 0x10A0, 0x1110, 0x160E}},
    {20627, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2402, 0x67FE, 0x6400, 0xA7DE, 0x2442, 0x2652, 0x254A, 0x24C6, 0x2B5A, 0x2842, 0x314A, 0x2084}},
    {20628, 2, {0xA08, 0x910, 0x800, 0x17FE, 0x10A0, 0x37FC, 0x30A4, 0x5FFE, 0x90A4, 0x17FC, 0x10A0, 0x11B0, 0x12A8, 0x1CA6, 0x10A0, 0x10A0}},
    {20629, 2, {0x1080, 0x1080, 0x17FE, 0x2522, 0x2110, 0x63FC, 0x6620, 0xAA20, 0x23FC, 0x2220, 0x2220, 0x23FC, 0x2220, 0x2220, 0x23FE, 0x2200}},
    {20630, 2, {0x840, 0x840, 0x8A0, 0x1190, 0x1248, 0x37FE, 0x3A08, 0x53F8, 0x9208, 0x13F8, 0x1200, 0x13F8, 0x1508, 0x1508, 0x19F8, 0x1108}},
    {20631, 2, {0x880, 0x840, 0xFFC, 0x1080, 0x1108, 0x33F0, 0x3060, 0x5184, 0x97FE, 0x1002, 0x17FC, 0x1444, 0x17FC, 0x1444, 0x17FC, 0x1404}},
    {20632, 2, {0x100, 0x280, 0x440, 0x820, 0x3118, 0xD116, 0x1110, 0x2928, 0x4544, 0x1110, 0x2928, 0x4544, 0xFFFE, 0x100, 0x100, 0x100}},
    {20633, 2, {0x888, 0x888, 0xBFE, 0x1088, 0x1088, 0x37FE, 0x3080, 0x51FC, 0x9324, 0x15FC, 0x1124, 0x11FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {20634, 2, {0x1410, 0x1210, 0x1210, 0x2FA0, 0x203E, 0x6544, 0x64A4, 0xA8A4, 0x3128, 0x2528, 0x2210, 0x2210, 0x2528, 0x2928, 0x3044, 0x2082}},
    {20635, 2, {0x840, 0x820, 0xBFE, 0x1202, 0x1494, 0x3108, 0x3264, 0x5090, 0x9108, 0x1204, 0x15FA, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {20636, 2, {0x1004, 0x103E, 0x17C0, 0x2244, 0x2124, 0x6128, 0x6200, 0xA3FC, 0x2440, 0x2040, 0x2FFE, 0x2040, 0x2444, 0x2444, 0x27FC, 0x2004}},
    {20637, 2, {0x800, 0xBF8, 0xA08, 0x13F8, 0x1208, 0x33F8, 0x3000, 0x5000, 0x97BC, 0x1084, 0x14A4, 0x1294, 0x14A4, 0x1084, 0x1294, 0x1108}},
    {20638, 2, {0x908, 0x888, 0x890, 0x13FC, 0x1040, 0x31F8, 0x3040, 0x53FE, 0x9080, 0x1100, 0x11FC, 0x1220, 0x1420, 0x1820, 0x13FE, 0x1000}},
    {20639, 2, {0x920, 0x910, 0xA08, 0x1486, 0x1910, 0x33F8, 0x3008, 0x5000, 0x97BC, 0x1084, 0x14A4, 0x1294, 0x14A4, 0x1084, 0x1294, 0x1108}},
    {20640, 2, {0x1204, 0x1104, 0x1104, 0x2FC4, 0x203E, 0x6784, 0x6004, 0xA7A4, 0x2014, 0x2794, 0x2484, 0x2484, 0x2484, 0x2784, 0x2494, 0x2008}},
    {20641, 2, {0x1410, 0x1208, 0x1208, 0x2FBE, 0x2000, 0x6082, 0x68A2, 0xA8A2, 0x2492, 0x2512, 0x2514, 0x2114, 0x2204, 0x2188, 0x2E3E, 0x2400}},
    {20642, 2, {0x1080, 0x1040, 0x17FC, 0x2404, 0x2000, 0x63F8, 0x6080, 0xA144, 0x2668, 0x20B0, 0x2128, 0x2668, 0x20A4, 0x2122, 0x26A0, 0x2040}},
    {20643, 2, {0x840, 0x840, 0xFFC, 0x1040, 0x13F8, 0x3080, 0x37FC, 0x5110, 0x9248, 0x1446, 0x1248, 0x1150, 0x1248, 0x1444, 0x1140, 0x1080}},
    {20644, 2, {0x1110, 0x1114, 0x17D2, 0x2110, 0x2FFE, 0x6210, 0x6210, 0xAFD4, 0x2414, 0x2914, 0x2FD4, 0x2108, 0x21CA, 0x2F1A, 0x2526, 0x2142}},
    {20645, 2, {0x840, 0xA48, 0x950, 0x1040, 0x17FE, 0x3402, 0x3804, 0x53F8, 0x9208, 0x1208, 0x13F8, 0x10A0, 0x1120, 0x1124, 0x1224, 0x141C}},
    {20646, 2, {0x800, 0x9F8, 0x908, 0x11E8, 0x1128, 0x37FE, 0x3402, 0x51F8, 0x9108, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x1108, 0x1128, 0x1110}},
    {20647, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2814, 0x63E0, 0x6200, 0xA200, 0x23FC, 0x2210, 0x2210, 0x2FFE, 0x2000, 0x2110, 0x2208, 0x2404}},
    {20648, 2, {0x1020, 0x1020, 0x18FA, 0x2424, 0x2428, 0x61FE, 0x6020, 0xBC40, 0x24FC, 0x2544, 0x2644, 0x247C, 0x2544, 0x2644, 0x247C, 0x2044}},
    {20649, 2, {0x1028, 0x1024, 0x1740, 0x217E, 0x21C8, 0x6948, 0x657E, 0xA548, 0x2248, 0x227E, 0x2548, 0x2548, 0x2948, 0x207E, 0x2040, 0x2040}},
    {20650, 2, {0x1080, 0x1110, 0x13F8, 0x2210, 0x24A4, 0x6FBE, 0x6042, 0xA0A0, 0x2318, 0x2C46, 0x2190, 0x2664, 0x2188, 0x2630, 0x21C0, 0x2E00}},
    {20651, 2, {0x1000, 0x17FC, 0x1444, 0x27FC, 0x2444, 0x67FC, 0x6100, 0xA210, 0x27E0, 0x20C8, 0x2304, 0x2FFE, 0x2042, 0x2248, 0x2544, 0x2882}},
    {20652, 2, {0x820, 0xA22, 0xBFE, 0x1090, 0x1088, 0x31FE, 0x3310, 0x5510, 0x91FE, 0x1110, 0x1110, 0x11FE, 0x1110, 0x1110, 0x11FE, 0x1100}},
    {20653, 2, {0x1020, 0x17FE, 0x1420, 0x25FC, 0x2424, 0x67FE, 0x6424, 0xA5FC, 0x2420, 0x25FC, 0x2524, 0x25FC, 0x2524, 0x29FC, 0x2924, 0x312C}},
    {20654, 2, {0x8A0, 0x8A0, 0xFFE, 0x10A0, 0x17FC, 0x34A4, 0x37FC, 0x54A4, 0x97FC, 0x1000, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208}},
    {20655, 2, {0x840, 0x880, 0xBFC, 0x1244, 0x12F4, 0x3314, 0x32A4, 0x5244, 0x92A4, 0x13FC, 0x1040, 0x1024, 0x1522, 0x150A, 0x1908, 0x10F8}},
    {20656, 2, {0x1040, 0x1444, 0x1444, 0x27FC, 0x2000, 0x67BC, 0x64A4, 0xA4A4, 0x27BC, 0x24A4, 0x24A4, 0x27BC, 0x24A4, 0x24A4, 0x2AB4, 0x3148}},
    {20657, 2, {0x1124, 0x1124, 0x1224, 0x2424, 0x2954, 0x614A, 0x6292, 0xA610, 0x2A10, 0x2250, 0x225C, 0x2250, 0x2250, 0x22B0, 0x229E, 0x2300}},
    {20658, 2, {0x1210, 0x1210, 0x1FD0, 0x2220, 0x2FBE, 0x6244, 0xBFA4, 0x2424, 0x2424, 0x27A8, 0x24A8, 0x2490, 0x24A8, 0x29A8, 0x2844, 0x3082}},
    {20659, 2, {0x1040, 0x1FFE, 0x1040, 0x27FC, 0x2444, 0x67FC, 0x6444, 0xA7FC, 0x2042, 0x2FFE, 0x2010, 0x2FFE, 0x2210, 0x2110, 0x2150, 0x2020}},
    {20660, 2, {0x1000, 0x17FC, 0x1400, 0x24F8, 0x2488, 0x6488, 0x64F8, 0xA400, 0x25DC, 0x2554, 0x2554, 0x2554, 0x25DC, 0x2400, 0x27FE, 0x2000}},
    {20661, 2, {0x820, 0xBFE, 0x820, 0x11FC, 0x1020, 0x33FE, 0x3000, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {20662, 2, {0x1014, 0x1012, 0x1010, 0x27FE, 0x2410, 0x6490, 0x64D2, 0xA492, 0x27F2, 0x2494, 0x2494, 0x25C8, 0x2AAA, 0x289A, 0x31A6, 0x2042}},
    {20663, 2, {0x900, 0x9FC, 0xA00, 0x15F8, 0x1108, 0x31F8, 0x3108, 0x51F8, 0x9000, 0x17FE, 0x1100, 0x13FC, 0x14A4, 0x1124, 0x1254, 0x1088}},
    {20664, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x2554, 0x6248, 0x6554, 0xA040, 0x2554, 0x2248, 0x2554, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {20665, 2, {0x880, 0x840, 0xBF8, 0x1110, 0x10A0, 0x37FE, 0x3000, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x13F8, 0x1120, 0x1122, 0x1222, 0x1C1E}},
    {20666, 2, {0x900, 0x910, 0xBDC, 0x1254, 0x1554, 0x3288, 0x3108, 0x52F4, 0x9402, 0x1000, 0x13FC, 0x1040, 0x1150, 0x1248, 0x1544, 0x1080}},
    {20667, 2, {0x840, 0x880, 0xBFC, 0x1294, 0x1264, 0x3294, 0x33FC, 0x5090, 0x9108, 0x1284, 0x14FA, 0x1108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {20668, 2, {0x1110, 0x1110, 0x17FE, 0x2110, 0x21F0, 0x6040, 0x63F8, 0xA248, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x20A0, 0x2110, 0x260C}},
    {20669, 2, {0x1040, 0x17FC, 0x1110, 0x20A0, 0x2FFE, 0x6000, 0x63F8, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040}},
    {20670, 2, {0x1000, 0x11FE, 0x1010, 0x2420, 0x24FC, 0x6484, 0x6784, 0xA4FC, 0x2484, 0x24FC, 0x2584, 0x2684, 0x24FC, 0x2048, 0x2084, 0x2102}},
    {20671, 2, {0x1000, 0x17FC, 0x1040, 0x2278, 0x2240, 0x6FFE, 0x6100, 0xA1FC, 0x2200, 0x23FC, 0x2004, 0x2554, 0x2554, 0x2804, 0x2028, 0x2010}},
    {20672, 2, {0x12A8, 0x12A8, 0x17FC, 0x22A8, 0x22AA, 0x64E6, 0x6800, 0xA7FC, 0x2444, 0x2040, 0x23F8, 0x2248, 0x2248, 0x2248, 0x2258, 0x2040}},
    {20673, 2, {0x1008, 0x103C, 0x17C0, 0x2040, 0x2248, 0x6150, 0x6FFE, 0xA150, 0x2248, 0x2C46, 0x2080, 0x2044, 0x2A42, 0x2A12, 0x31F0, 0x2000}},
    {20674, 2, {0x1040, 0x13F8, 0x1248, 0x2FFE, 0x2248, 0x63F8, 0x6040, 0xA3F8, 0x2248, 0x23F8, 0x2080, 0x2FFE, 0x2110, 0x20E0, 0x2318, 0x2C04}},
    {20675, 2, {0x1110, 0x1110, 0x1FFE, 0x2110, 0x2400, 0x67FC, 0x6804, 0xB7F4, 0x2494, 0x27F4, 0x2494, 0x27F4, 0x2494, 0x2494, 0x2434, 0x2008}},
    {20676, 2, {0x1000, 0x1FFE, 0x10A0, 0x27FC, 0x24A4, 0x64A4, 0x67FC, 0xA000, 0x27FC, 0x2000, 0x2FFE, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {20677, 2, {0x1110, 0x1110, 0x1FFE, 0x2110, 0x21F0, 0x6040, 0x67FC, 0xA444, 0x27FC, 0x2040, 0x2FFE, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x2000}},
    {20678, 2, {0x1020, 0x1820, 0x15FE, 0x2420, 0x21FC, 0x6124, 0x6DFC, 0xA524, 0x25FC, 0x2420, 0x25FE, 0x2420, 0x2420, 0x2A20, 0x31FE, 0x2000}},
    {20679, 2, {0x800, 0xFBC, 0x884, 0x14A4, 0x1294, 0x34A4, 0x3050, 0x5188, 0x9626, 0x10C0, 0x1310, 0x1064, 0x1388, 0x1030, 0x10C0, 0x1700}},
    {20680, 2, {0x13F8, 0x1208, 0x13F8, 0x2208, 0x23F8, 0x6000, 0x67FC, 0xA4A4, 0x24A4, 0x27FC, 0x2000, 0x23F8, 0x2110, 0x20E0, 0x2318, 0x2C06}},
    {20681, 2, {0x100, 0x280, 0x440, 0x820, 0x37D8, 0xC006, 0x0, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x820, 0x820, 0x1450, 0x2288, 0x4104}},
    {20682, 2, {0x1000, 0x17FC, 0x10A0, 0x27FC, 0x24A4, 0x67FC, 0x6040, 0xAFFE, 0x2110, 0x2208, 0x27F6, 0x2A10, 0x23F0, 0x2204, 0x2204, 0x21FC}},
    {20683, 2, {0x840, 0x8A0, 0x910, 0x1248, 0x1DF6, 0x3020, 0x3040, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1110, 0x1208}},
    {20684, 2, {0x1210, 0x1210, 0x1F90, 0x2220, 0x2FBE, 0x6AC4, 0x6AA4, 0xAFA4, 0x2AA4, 0x2AA8, 0x2FA8, 0x2210, 0x3FA8, 0x2228, 0x2244, 0x2282}},
    {20685, 2, {0x11C8, 0x1708, 0x1108, 0x2FE8, 0x211E, 0x67CA, 0x654A, 0xA7CA, 0x254A, 0x27CA, 0x210A, 0x27CA, 0x2112, 0x21D2, 0x2E2A, 0x2444}},
    {20686, 2, {0x1000, 0x17BC, 0x14A4, 0x27BC, 0x2420, 0x64A4, 0x639C, 0xA000, 0x2110, 0x27FC, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2208, 0x2404}},
    {20687, 2, {0x900, 0x9F0, 0xA10, 0x17FC, 0x1A24, 0x3244, 0x33FC, 0x5080, 0x9144, 0x16A8, 0x1130, 0x1668, 0x10A8, 0x1124, 0x16A2, 0x1040}},
    {20688, 2, {0x888, 0x850, 0xBFE, 0x1020, 0x11FC, 0x3020, 0x33FE, 0x5124, 0x90A8, 0x17FE, 0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {20689, 2, {0x1038, 0x13C0, 0x1080, 0x27FE, 0x2110, 0x6208, 0x65F6, 0xA910, 0x21F0, 0x2000, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x25F4, 0x240C}},
    {20690, 2, {0x1080, 0x1040, 0x17FC, 0x2514, 0x2208, 0x63F8, 0x6088, 0xA7FE, 0x2088, 0x23F8, 0x2100, 0x23F8, 0x2508, 0x2908, 0x21F8, 0x2108}},
    {20691, 2, {0x840, 0xBF8, 0xA48, 0x13F8, 0x1040, 0x37FE, 0x3000, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1110, 0x1208}},
    {20692, 2, {0x908, 0x890, 0xFFE, 0x1090, 0x13FC, 0x3294, 0x331C, 0x5204, 0x93FC, 0x1204, 0x13FC, 0x1008, 0x17FE, 0x1108, 0x10A8, 0x1010}},
    {20693, 2, {0x10A0, 0x14A4, 0x12A8, 0x20A0, 0x27FC, 0x6110, 0x60A0, 0xA7FC, 0x2040, 0x23F8, 0x2040, 0x27FC, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {20694, 2, {0x1040, 0x17FC, 0x1040, 0x23F8, 0x2000, 0x63F8, 0x6208, 0xA3F8, 0x2110, 0x2FFE, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {20695, 2, {0x888, 0xAAA, 0xADC, 0x1488, 0x1154, 0x3222, 0x3000, 0x53FE, 0x9242, 0x1444, 0x13FC, 0x1044, 0x1044, 0x1084, 0x1114, 0x1208}},
    {20696, 2, {0x2208, 0x3248, 0x2A48, 0x4A90, 0x421E, 0x5FE4, 0xD054, 0x5054, 0x5754, 0x5554, 0x5554, 0x5548, 0x5748, 0x5054, 0x5154, 0x50A2}},
    {20697, 2, {0x1110, 0x1110, 0x17FC, 0x2110, 0x2110, 0x6FFE, 0x6040, 0xA7FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2000, 0x2110, 0x2208, 0x2404}},
    {20698, 2, {0x1040, 0x1040, 0x17FC, 0x20A0, 0x2514, 0x6208, 0x67FC, 0xAA0A, 0x23F8, 0x2208, 0x23F8, 0x2040, 0x2248, 0x2444, 0x2944, 0x2080}},
    {20699, 2, {0x1510, 0x1510, 0x1510, 0x2F90, 0x253E, 0x6722, 0x6544, 0xA790, 0x2510, 0x2510, 0x2F90, 0x2028, 0x2528, 0x28A4, 0x3044, 0x2082}},
    {20700, 2, {0x1020, 0x17A4, 0x10A8, 0x2292, 0x2114, 0x6208, 0x65F4, 0xA802, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2110, 0x2FFE, 0x2000}},
    {20701, 2, {0x17FC, 0x1404, 0x17FC, 0x2400, 0x25F8, 0x6410, 0x67FE, 0xA420, 0x2460, 0x279C, 0x2484, 0x2508, 0x29DE, 0x2F08, 0x3108, 0x2318}},
    {20702, 2, {0x1004, 0x103E, 0x17C0, 0x2244, 0x2128, 0x63F8, 0x6208, 0xA3FC, 0x2204, 0x23FE, 0x2402, 0x26AA, 0x2AAA, 0x2A02, 0x3014, 0x2008}},
    {20703, 2, {0x1248, 0x1248, 0x1554, 0x275C, 0x2248, 0x6554, 0x675C, 0xA248, 0x2FFE, 0x2220, 0x2228, 0x2228, 0x2514, 0x2494, 0x242C, 0x2844}},
    {20704, 2, {0x878, 0xFC0, 0xA48, 0x1150, 0x17FC, 0x3150, 0x3248, 0x5406, 0x93F8, 0x1248, 0x1248, 0x13F8, 0x1248, 0x1248, 0x13F8, 0x1208}},
    {20705, 2, {0x1040, 0x1040, 0x1FFE, 0x2040, 0x27FC, 0x6444, 0x67FC, 0xA444, 0x27FC, 0x2044, 0x2FFE, 0x2042, 0x2024, 0x2522, 0x250A, 0x28F8}},
    {20706, 2, {0x103C, 0x17C0, 0x1044, 0x2224, 0x2108, 0x67FE, 0x6402, 0xAA08, 0x2208, 0x23BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x2408, 0x2808}},
    {20707, 2, {0x1110, 0x1110, 0x17BC, 0x2110, 0x2FFE, 0x6110, 0x62A8, 0xA444, 0x23FA, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {20708, 2, {0x800, 0xBDE, 0xA52, 0x1252, 0x13DE, 0x3000, 0x31FC, 0x5124, 0x91FC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {20709, 2, {0x1040, 0x1040, 0x17FC, 0x2040, 0x2FFE, 0x6110, 0x67BC, 0xA110, 0x2190, 0x263E, 0x2000, 0x2FFE, 0x2120, 0x2122, 0x2222, 0x2C1E}},
    {20710, 2, {0x1414, 0x1212, 0x1F92, 0x2010, 0x207E, 0x6F90, 0x6890, 0xA890, 0x2F90, 0x2228, 0x2B28, 0x2AA8, 0x3228, 0x224A, 0x2A4A, 0x2486}},
    {20711, 2, {0x904, 0x888, 0x800, 0x13FE, 0x1222, 0x32AA, 0x3272, 0x5222, 0x93FE, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {20712, 2, {0x1040, 0x17FC, 0x1040, 0x2110, 0x2FFE, 0x6110, 0x63F8, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x2208, 0x2404}},
    {20713, 2, {0x179E, 0x1492, 0x179E, 0x2492, 0x279E, 0x6402, 0x64F2, 0xA492, 0x24F2, 0x2492, 0x24F2, 0x2492, 0x2492, 0x2532, 0x2402, 0x2406}},
    {20714, 2, {0x800, 0xBFC, 0x848, 0x1030, 0x13FE, 0x3052, 0x3094, 0x5350, 0x9020, 0x13FE, 0x1252, 0x128A, 0x1376, 0x1252, 0x1272, 0x1206}},
    {20715, 2, {0x800, 0xFFE, 0x890, 0x1090, 0x139C, 0x3204, 0x3204, 0x539C, 0x9090, 0x1090, 0x17FE, 0x1040, 0x1524, 0x150A, 0x190A, 0x10F8}},
    {20716, 2, {0x1140, 0x1120, 0x13FE, 0x2220, 0x2620, 0x6BFC, 0x6220, 0xA220, 0x23FC, 0x2220, 0x2220, 0x23FE, 0x2200, 0x2524, 0x2492, 0x2892}},
    {20717, 2, {0x1000, 0x17BC, 0x1108, 0x2528, 0x27BC, 0x6318, 0x65AA, 0xA946, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8}},
    {20718, 2, {0x840, 0xBF8, 0x910, 0x10A0, 0x17FC, 0x3000, 0x33F8, 0x5248, 0x93F8, 0x1248, 0x13F8, 0x1040, 0x13F8, 0x1040, 0x17FE, 0x1000}},
    {20719, 2, {0x1040, 0x1248, 0x1150, 0x27FC, 0x2150, 0x6248, 0x6444, 0xA208, 0x2208, 0x23BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x2408, 0x2808}},
    {20720, 2, {0x1010, 0x1010, 0xFEFE, 0x1010, 0xFEFE, 0x9292, 0x9292, 0x3838, 0x5454, 0x9292, 0x1010, 0x1110, 0x280, 0xC60, 0x3018, 0xC006}},
    {20721, 2, {0x840, 0x820, 0xBFE, 0x1202, 0x1202, 0x33FE, 0x3250, 0x5248, 0x92FE, 0x1390, 0x12FE, 0x1290, 0x14FE, 0x1490, 0x18FE, 0x1080}},
    {20722, 2, {0x1000, 0x17FE, 0x10A0, 0x27FC, 0x24A4, 0x64A4, 0x67FC, 0xA208, 0x2208, 0x23BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x2408, 0x2808}},
    {20723, 2, {0x1000, 0x17FC, 0x10A0, 0x20A0, 0x27FC, 0x64A4, 0x64A4, 0xA7FC, 0x2248, 0x2150, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {20724, 2, {0x1000, 0x17BC, 0x14A4, 0x27BC, 0x24A4, 0x67BC, 0x6404, 0xA5F4, 0x2514, 0x25F4, 0x2514, 0x25F4, 0x2514, 0x2404, 0x2414, 0x2408}},
    {20725, 2, {0x1000, 0x1FFE, 0x1000, 0x27FC, 0x2444, 0x67FC, 0x6444, 0xAFFE, 0x2000, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2000, 0x2FFE}},
    {20726, 2, {0x1000, 0x17FC, 0x14A4, 0x24A4, 0x27BC, 0x60A0, 0x67BC, 0xA4A4, 0x24A4, 0x27BC, 0x24A4, 0x24A4, 0x27BC, 0x24A2, 0x20A2, 0x207E}},
    {20727, 2, {0x1108, 0x17FE, 0x1108, 0x2000, 0x2488, 0x6FFE, 0x6488, 0xA4F8, 0x2400, 0x27FC, 0x2040, 0x2FFE, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {20728, 2, {0x1110, 0x1110, 0x17BC, 0x2110, 0x23B8, 0x6554, 0x6912, 0xA000, 0x23F8, 0x2000, 0x27FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {20729, 2, {0x1000, 0x1FFE, 0x1120, 0x2FFC, 0x2924, 0x6FFC, 0x6000, 0xA7F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2210, 0x2408}},
    {20730, 2, {0x13F8, 0x1208, 0x1208, 0x23F8, 0x2000, 0x67BC, 0x64A4, 0xA4A4, 0x27BC, 0x2040, 0x2FFE, 0x2160, 0x2250, 0x2448, 0x3846, 0x2040}},
    {20731, 2, {0x1010, 0x1008, 0x1788, 0x24BE, 0x2480, 0x6494, 0x6788, 0xA47E, 0x2408, 0x2788, 0x26BE, 0x2A88, 0x2A88, 0x2B88, 0x3288, 0x2008}},
    {20732, 2, {0x10A0, 0x10A0, 0x17FC, 0x24A4, 0x27FC, 0x64A4, 0x67FC, 0xA000, 0x27FC, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x20A0, 0x2FFE}},
    {20733, 2, {0x10C8, 0x1708, 0x112A, 0x212C, 0x2FC8, 0x6108, 0x6388, 0xA554, 0x2914, 0x2122, 0x2040, 0x2024, 0x2522, 0x250A, 0x2908, 0x20F8}},
    {20734, 2, {0x83C, 0xBC0, 0x844, 0x1224, 0x1108, 0x37FE, 0x3442, 0x52A4, 0x928A, 0x147A, 0x1080, 0x10F8, 0x1108, 0x1290, 0x1060, 0x179E}},
    {20735, 2, {0x1080, 0x1040, 0x17FE, 0x2512, 0x2BFC, 0x6110, 0x67FC, 0xA110, 0x27FE, 0x2110, 0x2248, 0x25F4, 0x2842, 0x2040, 0x27FC, 0x2000}},
    {20736, 2, {0x1110, 0x17FE, 0x1040, 0x23FC, 0x2040, 0x67FE, 0xA080, 0x2728, 0x2124, 0x2FFE, 0x2120, 0x21A4, 0x2F1A, 0x2112, 0x276E, 0x0}},
    {20737, 2, {0x1140, 0x1120, 0x13FE, 0x2220, 0x27FC, 0x6A20, 0x63FC, 0xA220, 0x23FE, 0x2200, 0x279E, 0x2492, 0x2492, 0x24F2, 0x2402, 0x2406}},
    {20738, 2, {0x890, 0xBFC, 0xA94, 0x13FC, 0x1294, 0x33FC, 0x3000, 0x53FC, 0x9200, 0x12F8, 0x1200, 0x13FE, 0x1520, 0x1514, 0x1548, 0x1986}},
    {20739, 2, {0x1040, 0x1FFE, 0x1000, 0x27FC, 0x2404, 0x65F4, 0x6514, 0xA7FC, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2000, 0x2FFE}},
    {20740, 2, {0x880, 0x840, 0xBFC, 0x1108, 0x1090, 0x37FE, 0x3000, 0x53FC, 0x9204, 0x13FC, 0x1204, 0x13FC, 0x1040, 0x1524, 0x150A, 0x18FA}},
    {20741, 2, {0x1040, 0x1248, 0x1150, 0x27FE, 0x2402, 0x6804, 0x63F8, 0xA208, 0x23F8, 0x2000, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2404}},
    {20742, 2, {0x1288, 0x1288, 0x17E8, 0x2290, 0x241E, 0x67E4, 0x6454, 0xA854, 0x2754, 0x2554, 0x2554, 0x2748, 0x2548, 0x2054, 0x2294, 0x2122}},
    {20743, 2, {0x800, 0xBFC, 0xA94, 0x1294, 0x13FC, 0x3000, 0x37FE, 0x5000, 0x93FC, 0x1204, 0x13FC, 0x10A2, 0x1114, 0x1308, 0x1D44, 0x1182}},
    {20744, 2, {0x1040, 0x10A0, 0x1318, 0x2DF6, 0x2000, 0x67FC, 0x6554, 0xA4E4, 0x27FC, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208}},
    {20745, 2, {0x1040, 0x1040, 0x10A0, 0x2110, 0x2208, 0x65F6, 0x6800, 0xA7BC, 0x24A4, 0x24A4, 0x27BC, 0x2210, 0x2210, 0x2528, 0x2944, 0x3082}},
    {20746, 2, {0x1110, 0x1110, 0x17BC, 0x2110, 0x23B8, 0x6554, 0x6912, 0xA000, 0x27FC, 0x2044, 0x2240, 0x2278, 0x2240, 0x2540, 0x28FE, 0x3000}},
    {20747, 2, {0x1200, 0x13F8, 0x1410, 0x27FE, 0x2D48, 0x7624, 0x67FE, 0xA400, 0x25FC, 0x2400, 0x25FC, 0x2400, 0x25FC, 0x2904, 0x29FC, 0x3104}},
    {20748, 2, {0x1210, 0x1410, 0x1F90, 0x2890, 0x2FBE, 0x68A4, 0x6FD4, 0xA414, 0x2214, 0x2FD4, 0x2414, 0x2788, 0x2488, 0x2894, 0x2A94, 0x3122}},
    {20749, 2, {0x820, 0x888, 0xA52, 0x1326, 0x1252, 0x328A, 0x3222, 0x53FE, 0x9088, 0x1144, 0x127A, 0x1088, 0x1150, 0x1020, 0x10D8, 0x1706}},
    {20750, 2, {0x1114, 0x17D2, 0x1110, 0x2FFE, 0x2110, 0x6FD0, 0x6112, 0xA7D2, 0x2552, 0x27D4, 0x2554, 0x27C8, 0x210A, 0x2FDA, 0x2126, 0x2142}},
    {20751, 2, {0x1080, 0x10DC, 0x1284, 0x23D4, 0x2448, 0x6A94, 0x6324, 0xAC40, 0x2248, 0x2150, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {20752, 2, {0x1080, 0x1040, 0x17FE, 0x2402, 0x2BFC, 0x6148, 0x6250, 0xA7FC, 0x2204, 0x27FC, 0x2A04, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {20753, 2, {0x800, 0xBFE, 0xA02, 0x13FE, 0x1202, 0x33FE, 0x3108, 0x5252, 0x939C, 0x1108, 0x1252, 0x13DE, 0x1000, 0x12A4, 0x1252, 0x1452}},
    {20754, 2, {0x1000, 0x17FC, 0x1040, 0x2FFE, 0x2842, 0x6358, 0x6040, 0xA358, 0x2000, 0x2FFE, 0x2040, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x240C}},
    {20755, 2, {0x1040, 0x1FFE, 0x1040, 0x27FC, 0x2000, 0x63F8, 0x6208, 0xAFFE, 0x2802, 0x27FC, 0x2110, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x2FFE}},
    {20756, 2, {0x1040, 0x17FC, 0x1040, 0x23FC, 0x2000, 0x6FFE, 0x6002, 0xA3F8, 0x2040, 0x27FE, 0x2000, 0x27FE, 0x2008, 0x27FE, 0x24A8, 0x2798}},
    {20757, 2, {0x1080, 0x1040, 0x17FC, 0x2110, 0x20A4, 0x6F58, 0x6554, 0xA552, 0x2B58, 0x2000, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2408}},
    {20758, 2, {0x1010, 0x17D0, 0x1510, 0x27DE, 0x2450, 0x67E8, 0x6504, 0xA7C4, 0x2000, 0x23F8, 0x22A8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000}},
    {20759, 2, {0x1400, 0x157C, 0x1604, 0x24A8, 0x2390, 0x607C, 0x6414, 0xA794, 0x2A50, 0x2250, 0x2FDC, 0x2250, 0x2550, 0x24B0, 0x289E, 0x2100}},
    {20760, 2, {0x840, 0xBF8, 0x848, 0x17FE, 0x1048, 0x33F8, 0x3040, 0x57FE, 0x9000, 0x12A4, 0x1452, 0x13F8, 0x12A8, 0x12A8, 0x1FFE, 0x1000}},
    {20761, 2, {0x1010, 0x14FE, 0x1292, 0x22FE, 0x2010, 0x61FE, 0x6000, 0xAEFE, 0x2282, 0x22FE, 0x2280, 0x22FE, 0x2282, 0x22FE, 0x2500, 0x28FE}},
    {20762, 2, {0x888, 0x888, 0xBFE, 0x1088, 0x13FE, 0x3252, 0x33FE, 0x5000, 0x93FE, 0x1242, 0x10F8, 0x1108, 0x1290, 0x1060, 0x10C0, 0x1700}},
    {20763, 2, {0x1100, 0x1200, 0x17FC, 0x2AA8, 0x22A8, 0x6FFE, 0x62A8, 0xA2A8, 0x2FFE, 0x2208, 0x23BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x2C08}},
    {20764, 2, {0x1080, 0x1040, 0x17FC, 0x2444, 0x22A8, 0x6294, 0x6474, 0xA000, 0x23F8, 0x22A8, 0x2FFE, 0x2000, 0x27FC, 0x2040, 0x2140, 0x2080}},
    {20765, 2, {0x1110, 0x1554, 0x15B8, 0x2910, 0x22A8, 0x6444, 0x6000, 0xA7FC, 0x2444, 0x2040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {20766, 2, {0x1000, 0x17FC, 0x1248, 0x2444, 0x2FFE, 0x6444, 0x66EC, 0xA554, 0x26EC, 0x2444, 0x26EC, 0x2554, 0x26EC, 0x2444, 0x2454, 0x2408}},
    {20767, 2, {0xA48, 0x950, 0xFFE, 0x1402, 0x11F8, 0x3108, 0x33FC, 0x5204, 0x93FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1108, 0x1204, 0x1402}},
    {20768, 2, {0x924, 0xA48, 0x924, 0x1000, 0x13FC, 0x3294, 0x3264, 0x5294, 0x93FC, 0x1248, 0x1368, 0x1248, 0x136A, 0x124A, 0x1246, 0x1362}},
    {20769, 2, {0x13F8, 0x1248, 0x1248, 0x23F8, 0x2248, 0x6248, 0x63F8, 0xA000, 0x2FBE, 0x2AAA, 0x2AAA, 0x2FBE, 0x2AAA, 0x2AAA, 0x2FBE, 0x28A2}},
    {20770, 2, {0x1020, 0x103E, 0x1020, 0x27FE, 0x2422, 0x65F8, 0x6422, 0xA5FE, 0x2524, 0x25FC, 0x2524, 0x25FC, 0x2840, 0x2AA4, 0x328A, 0x247A}},
    {20771, 2, {0x1040, 0x1020, 0x17FE, 0x2488, 0x25FC, 0x6488, 0x67FE, 0xA420, 0x25FC, 0x2524, 0x25FC, 0x2524, 0x29FC, 0x2888, 0x3104, 0x2202}},
    {20772, 2, {0xBF8, 0xA08, 0xBF8, 0x1208, 0x13F8, 0x3110, 0x37FC, 0x5110, 0x9FFE, 0x1110, 0x1248, 0x1554, 0x10E0, 0x1150, 0x1248, 0x10C0}},
    {20773, 2, {0x820, 0xBFE, 0x820, 0x11FC, 0x1000, 0x33FE, 0x3252, 0x53FE, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {20774, 2, {0x1080, 0x1040, 0x17FC, 0x24A0, 0x27FC, 0x64A4, 0x67FC, 0xA400, 0x2524, 0x25A8, 0x2530, 0x25A4, 0x251C, 0x2800, 0x2AA4, 0x3452}},
    {20775, 2, {0x1208, 0x1FBE, 0x1208, 0x2FBE, 0x2208, 0x6514, 0x68A2, 0xA7FC, 0x2404, 0x27FC, 0x2404, 0x27FC, 0x2404, 0x27FC, 0x2110, 0x2208}},
    {20776, 2, {0x842, 0xB9C, 0xA10, 0x13DE, 0x1294, 0x3294, 0x3420, 0x53FC, 0x9204, 0x13FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1108, 0x1204}},
    {20777, 2, {0x1000, 0x1F7C, 0x1944, 0x297C, 0x2F44, 0x6944, 0x697C, 0xAF20, 0x297E, 0x29AA, 0x2F2A, 0x204A, 0x2A92, 0x2922, 0x294A, 0x3104}},
    {20778, 2, {0x17FC, 0x1080, 0x13F8, 0x2208, 0x23F8, 0x6208, 0x6FFE, 0xA882, 0x2548, 0x2514, 0x28F4, 0x2100, 0x23F8, 0x2508, 0x20F0, 0x2F0E}},
    {20779, 2, {0x1040, 0x17FE, 0x1000, 0x21F8, 0x2108, 0x61F8, 0x6000, 0xA7FE, 0x2402, 0x23F8, 0x2180, 0x2644, 0x21B8, 0x2668, 0x21A6, 0x2E60}},
    {20780, 2, {0x1200, 0x123C, 0x1FA4, 0x2224, 0x2FBC, 0x6224, 0x7FE4, 0xA03C, 0x2FA4, 0x28A4, 0x2FBC, 0x2890, 0x2F98, 0x28A8, 0x28AA, 0x29C6}},
    {20781, 2, {0x1400, 0x123C, 0x1FA4, 0x2024, 0x28BC, 0x6524, 0x6FA4, 0xA23C, 0x2224, 0x2FA4, 0x223C, 0x2B10, 0x2A98, 0x32A8, 0x2A2A, 0x2446}},
    {20782, 2, {0x1000, 0x17FE, 0x1444, 0x2598, 0x2488, 0x67DE, 0x6488, 0xA5DC, 0x26AA, 0x2488, 0x2420, 0x2520, 0x253C, 0x2920, 0x2920, 0x37FE}},
    {20783, 2, {0x1248, 0x1150, 0x17FC, 0x2080, 0x2FFE, 0x6208, 0x67FC, 0xAA42, 0x23F8, 0x2240, 0x23F8, 0x2240, 0x23FC, 0x2004, 0x2554, 0x2008}},
    {20784, 2, {0x1110, 0x1FFE, 0x1110, 0x203C, 0x2FC0, 0x6488, 0x6250, 0xA7F8, 0x2408, 0x27FC, 0x2404, 0x27FE, 0x2802, 0x2AAA, 0x32AA, 0x2004}},
    {20785, 2, {0x1420, 0x1220, 0x1FBE, 0x2520, 0x223E, 0x7F82, 0x603E, 0xAFA0, 0x28BE, 0x2FA0, 0x28BE, 0x2FA0, 0x28BE, 0x28A0, 0x28A2, 0x299E}},
    {20786, 2, {0x1410, 0x1210, 0x127A, 0x2F12, 0x2014, 0x6F7E, 0x6008, 0xAF10, 0x203C, 0x2F64, 0x29A4, 0x293C, 0x2924, 0x2F24, 0x293C, 0x2024}},
    {20787, 2, {0x9F0, 0xA10, 0xFFC, 0x1204, 0x13FC, 0x3224, 0x33B8, 0x5222, 0x91FE, 0x1210, 0x17FC, 0x1244, 0x13FC, 0x10D0, 0x114A, 0x163E}},
    {20788, 2, {0x1040, 0x17FC, 0x1000, 0x23B8, 0x22A8, 0x63B8, 0x6110, 0xA7FC, 0x2110, 0x27FC, 0x2110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {20789, 2, {0x1080, 0x11F8, 0x1288, 0x2070, 0x238E, 0x6800, 0x69FC, 0xA924, 0x29AC, 0x2924, 0x29FC, 0x2820, 0x29FC, 0x2020, 0x23FE, 0x2154}},
    {20790, 2, {0xA22, 0xBFE, 0x890, 0x11FE, 0x1310, 0x35FE, 0x3110, 0x51FE, 0x9110, 0x11FE, 0x1100, 0x13FE, 0x128A, 0x1376, 0x1252, 0x1276}},
    {20791, 2, {0x17BE, 0x1000, 0x17BC, 0x24A4, 0x26B4, 0x64A4, 0x6040, 0xA7FE, 0x24A0, 0x27FC, 0x24A4, 0x27FC, 0x2512, 0x29DC, 0x2912, 0x31CE}},
    {20792, 2, {0x1000, 0x17FC, 0x14A4, 0x24A4, 0x27FC, 0x6210, 0x62A8, 0xA4BE, 0x2F68, 0x223C, 0x24A8, 0x2FBC, 0x2028, 0x2AA8, 0x2ABE, 0x2020}},
    {20793, 2, {0x1528, 0x17BE, 0x1948, 0x27BE, 0x2318, 0x65AA, 0x6946, 0xA3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {20794, 2, {0x1518, 0x1514, 0x1F90, 0x253E, 0x2728, 0x6268, 0x6FBE, 0xAAA8, 0x2FA8, 0x223E, 0x2FA8, 0x2228, 0x3FE8, 0x253E, 0x28A0, 0x3060}},
    {20795, 2, {0x1248, 0x1150, 0x17FC, 0x2404, 0x21F0, 0x6110, 0x67FC, 0xA554, 0x24E4, 0x27FC, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x22A4, 0x2452}},
    {20796, 2, {0x17BC, 0x14A4, 0x17BC, 0x2000, 0x27FE, 0x6400, 0x6590, 0xA49E, 0x27D4, 0x2564, 0x25D4, 0x2554, 0x25D4, 0x2568, 0x2BC8, 0x3054}},
    {20797, 2, {0x17FC, 0x1444, 0x17FC, 0x2444, 0x2FFE, 0x6AAA, 0x6FBE, 0xAAAA, 0x2FBE, 0x2220, 0x27C0, 0x2084, 0x2FFE, 0x2042, 0x2548, 0x2884}},
    {20798, 2, {0x1040, 0x17FC, 0x1248, 0x23F8, 0x2040, 0x6FFE, 0x6AAA, 0xA3B8, 0x2110, 0x27FC, 0x2110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {20799, 2, {0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x1042, 0x1042, 0x2042, 0x403E, 0x8000}},
    {20800, 2, {0x0, 0x0, 0xFFFE, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x1044, 0x1044, 0x2044, 0x403C, 0x8000}},
    {20801, 2, {0x100, 0x100, 0x200, 0x440, 0x820, 0x1010, 0x3FF8, 0x448, 0x440, 0x440, 0x440, 0x840, 0x842, 0x1042, 0x2042, 0xC03E}},
    {20802, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x1100, 0x1100, 0x2100, 0x3FFE, 0x280, 0x480, 0x480, 0x884, 0x1084, 0x207C, 0xC000}},
    {20803, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x440, 0x840, 0x842, 0x1042, 0x2042, 0x403E, 0x8000}},
    {20804, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x440, 0x440, 0x440, 0x840, 0x840, 0x1042, 0x2042, 0xC03E}},
    {20805, 2, {0x200, 0x100, 0xFFFE, 0x400, 0x400, 0x820, 0x1010, 0x3FF8, 0x448, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {20806, 2, {0x440, 0x440, 0x4444, 0x2448, 0x1450, 0x1440, 0x460, 0xC50, 0x1448, 0x2444, 0x4440, 0x842, 0x842, 0x1042, 0x203E, 0x4000}},
    {20807, 2, {0x820, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2828, 0x3FF8, 0x440, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {20808, 2, {0x100, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x440, 0x440, 0x440, 0x840, 0x842, 0x1042, 0x203E, 0xC000}},
    {20809, 2, {0x100, 0x2108, 0x1108, 0x910, 0x920, 0x100, 0xFFFE, 0x440, 0x440, 0x440, 0x440, 0x842, 0x842, 0x1042, 0x203E, 0xC000}},
    {20810, 2, {0x440, 0x440, 0x820, 0x1110, 0x2108, 0xC206, 0x440, 0x820, 0x1FF0, 0x450, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {20811, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x440, 0x440, 0x840, 0x1042, 0x2042, 0xC03E}},
    {20812, 2, {0x440, 0x440, 0x820, 0x1010, 0x2008, 0xDFF6, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {20813, 2, {0x800, 0x800, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x2108, 0x3FF8, 0x2288, 0x480, 0x880, 0x1082, 0x2082, 0xC07E}},
    {20814, 2, {0x8, 0x1FC, 0x7F00, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2288, 0x4A0, 0x890, 0x1092, 0x2082, 0xC07E}},
    {20815, 2, {0x10, 0xF8, 0x1F80, 0x1080, 0x10F8, 0x1080, 0x1080, 0x1080, 0x1FFE, 0x1120, 0x1120, 0x1120, 0x2222, 0x2222, 0x441E, 0x9800}},
    {20816, 2, {0x0, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x0, 0x440, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {20817, 2, {0x1010, 0x820, 0x440, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1450, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {20818, 2, {0x0, 0x600, 0x38F8, 0x2008, 0x2008, 0x3CF8, 0x2008, 0x2008, 0x3FF8, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {20819, 2, {0x0, 0x7EFE, 0x810, 0x810, 0x4890, 0x4890, 0x4890, 0x7EFE, 0x810, 0x810, 0x1428, 0x1428, 0x2548, 0x264A, 0x448A, 0x8106}},
    {20820, 2, {0x800, 0x800, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x2108, 0x3FF8, 0x2288, 0x4A0, 0x890, 0x1092, 0x2082, 0xC07E}},
    {20821, 2, {0x0, 0x3C78, 0x2448, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2008, 0x3FF8, 0x2448, 0x440, 0x440, 0x840, 0x1042, 0x2042, 0xC03E}},
    {20822, 2, {0x200, 0x100, 0xFFFE, 0x820, 0x1110, 0x2208, 0x4444, 0x820, 0x1FF0, 0x450, 0x440, 0x840, 0x840, 0x1042, 0x2042, 0xC03E}},
    {20823, 2, {0x200, 0x100, 0xFFFE, 0x820, 0x1010, 0x2008, 0x5FF4, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {20824, 2, {0x20, 0x7C20, 0x48, 0x84, 0x1FE, 0xFE02, 0x2400, 0x24FC, 0x2484, 0x2484, 0x24FC, 0x2484, 0x2402, 0x4402, 0x43FE, 0x8000}},
    {20825, 2, {0x810, 0x810, 0xFF10, 0x810, 0x810, 0x7F7E, 0x4110, 0x4110, 0x7F10, 0x1410, 0x1410, 0x1410, 0x2412, 0x2402, 0x43FE, 0x8000}},
    {20826, 2, {0x100, 0x1110, 0x910, 0x920, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {20827, 2, {0x808, 0x81C, 0xFF70, 0x810, 0x810, 0x7F10, 0x41FE, 0x4110, 0x7F10, 0x1410, 0x1410, 0x1410, 0x2412, 0x2402, 0x43FE, 0x8000}},
    {20828, 2, {0x100, 0x1200, 0x67DC, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4444, 0x57C4, 0x601C, 0x440, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {20829, 2, {0x808, 0x828, 0xFF24, 0x844, 0x842, 0x7F80, 0x417C, 0x4124, 0x7F24, 0x1444, 0x1444, 0x1494, 0x250A, 0x2402, 0x43FE, 0x8000}},
    {20830, 2, {0x838, 0x8E0, 0xFF20, 0x838, 0x8E0, 0x7F20, 0x413C, 0x41E0, 0x7F24, 0x1424, 0x1424, 0x141C, 0x2402, 0x2402, 0x43FE, 0x8000}},
    {20831, 2, {0x810, 0x2850, 0x2850, 0x3E7C, 0x4890, 0x810, 0x810, 0x7EFE, 0x1428, 0x1428, 0x1428, 0x1428, 0x254A, 0x264A, 0x4486, 0x8100}},
    {20832, 2, {0x100, 0x1210, 0x17D2, 0xF454, 0x1458, 0x17D0, 0x1452, 0x3452, 0xD7CE, 0x1000, 0x440, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {20833, 2, {0x800, 0x8FE, 0xFF10, 0x820, 0x87C, 0x7F44, 0x4144, 0x417C, 0x7F44, 0x1444, 0x147C, 0x1444, 0x2402, 0x2402, 0x43FE, 0x8000}},
    {20834, 2, {0x820, 0x820, 0x820, 0x7EFE, 0x820, 0x820, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x1450, 0x1450, 0x2550, 0x2692, 0x4492, 0x810E}},
    {20835, 2, {0x1000, 0x11FE, 0xFD00, 0x117C, 0x1154, 0x7D7C, 0x4554, 0x457C, 0x7D10, 0x297C, 0x2910, 0x2AFE, 0x2C00, 0x4802, 0x47FE, 0x8000}},
    {20836, 2, {0x1010, 0x11FE, 0x9528, 0x557C, 0x5928, 0x11FE, 0xFD10, 0x297C, 0x2954, 0x297C, 0x2954, 0x2A7C, 0x2A28, 0x4C44, 0x4802, 0x87FE}},
    {20837, 2, {0x400, 0x200, 0x100, 0x100, 0x100, 0x280, 0x280, 0x280, 0x440, 0x440, 0x820, 0x820, 0x1010, 0x2010, 0x4008, 0x8006}},
    {20838, 2, {0x400, 0x4200, 0x4100, 0x4100, 0x4080, 0x4080, 0x4140, 0x4140, 0x4220, 0x4220, 0x4410, 0x4808, 0x5004, 0x4000, 0x7FFE, 0x0}},
    {20839, 2, {0x400, 0x200, 0x200, 0x7FFC, 0x4104, 0x4104, 0x4284, 0x4284, 0x4444, 0x4824, 0x5014, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20840, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2FE8, 0xC106, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x100, 0x7FFC, 0x0}},
    {20841, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x5144, 0x4924, 0x4924, 0x5554, 0x6594, 0x4104, 0x4104, 0x4114, 0x4008}},
    {20842, 2, {0x600, 0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x3E24, 0x2224, 0x2248, 0x3E48, 0x2290, 0x3E48, 0x2248, 0x2224, 0x2624}},
    {20843, 2, {0x0, 0x40, 0x440, 0x440, 0x440, 0x440, 0x440, 0x420, 0x820, 0x820, 0x810, 0x1010, 0x1008, 0x2008, 0x2004, 0x4002}},
    {20844, 2, {0x80, 0x480, 0x480, 0x840, 0x840, 0x1020, 0x2010, 0x4208, 0x8206, 0x400, 0x440, 0x820, 0x1020, 0x3FF0, 0x1010, 0x0}},
    {20845, 2, {0x200, 0x100, 0x80, 0x80, 0x0, 0xFFFE, 0x0, 0x0, 0x440, 0x420, 0x810, 0x808, 0x1008, 0x2004, 0x4004, 0x0}},
    {20846, 2, {0x440, 0x440, 0x820, 0x820, 0x1010, 0x2008, 0xDFF6, 0x400, 0x800, 0xFE0, 0x20, 0x20, 0x20, 0x20, 0x140, 0x80}},
    {20847, 2, {0x2010, 0x1010, 0x820, 0x840, 0x0, 0x7FF8, 0x208, 0x208, 0x208, 0x208, 0x228, 0x210, 0x200, 0x200, 0x200, 0x200}},
    {20848, 2, {0x1010, 0x810, 0x420, 0x440, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0}},
    {20849, 2, {0x420, 0x420, 0x420, 0x420, 0x3FFC, 0x420, 0x420, 0x420, 0x420, 0x420, 0x7FFE, 0x0, 0x420, 0x810, 0x1008, 0x2004}},
    {20850, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x440, 0x440, 0x820, 0x820, 0x1010, 0x2008, 0xC006}},
    {20851, 2, {0x1010, 0x810, 0x820, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {20852, 2, {0x408, 0x208, 0x2108, 0x1110, 0x810, 0x820, 0x40, 0xFFFE, 0x0, 0x440, 0x420, 0x810, 0x808, 0x1008, 0x2004, 0x4004}},
    {20853, 2, {0x20, 0xF0, 0x1F00, 0x1000, 0x1000, 0x1FF8, 0x1080, 0x1080, 0x1080, 0x1080, 0xFFFE, 0x0, 0x840, 0x1020, 0x2010, 0x4008}},
    {20854, 2, {0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {20855, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x440, 0x820, 0x1010, 0x2008}},
    {20856, 2, {0x440, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x440, 0x820, 0x1010, 0x2008}},
    {20857, 2, {0x820, 0x420, 0x440, 0xFFFE, 0x1020, 0x1020, 0x2040, 0x2448, 0x4488, 0x78F0, 0x810, 0x1020, 0x1428, 0x2244, 0x7EFC, 0x204}},
    {20858, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {20859, 2, {0x820, 0x440, 0x288, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x440, 0x830, 0x344E, 0xC444, 0x440, 0x440, 0x840, 0x1040}},
    {20860, 2, {0x1020, 0x840, 0x7FFC, 0x480, 0x480, 0x3FF0, 0x490, 0xFFFE, 0x490, 0x3FF0, 0xCC0, 0x14A0, 0x2498, 0xC486, 0x480, 0x480}},
    {20861, 2, {0x820, 0x440, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {20862, 2, {0x820, 0x440, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x820, 0x7FFC, 0x820, 0x820, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {20863, 2, {0x810, 0x420, 0xFFFE, 0x820, 0x3EF8, 0x828, 0x3EAA, 0x866, 0x3892, 0x0, 0x1FF0, 0x0, 0x7FFC, 0x400, 0x810, 0x1FF8}},
    {20864, 2, {0x444, 0x7C78, 0x442, 0x1C3E, 0xE400, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x1010, 0x2008}},
    {20865, 2, {0x4400, 0x25FE, 0x2902, 0x7D02, 0x55FE, 0x5524, 0x7D24, 0x557E, 0x5524, 0x7D24, 0x11FE, 0x1152, 0xFD54, 0x1248, 0x1264, 0x1442}},
    {20866, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20867, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x5FF4, 0x4004, 0x4004, 0x4004, 0x5FF4, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20868, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0xFFFE, 0x2008, 0x2008, 0x2008, 0xFFFE, 0x2008, 0x2008, 0x2008, 0x2008, 0x2028, 0x2010}},
    {20869, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x4104, 0x4104, 0x4284, 0x4244, 0x4424, 0x4814, 0x5014, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20870, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20871, 2, {0x200, 0x200, 0xFFFE, 0x400, 0x400, 0xFF0, 0x810, 0x1810, 0x2810, 0x4810, 0x8810, 0x810, 0x810, 0x810, 0x850, 0x820}},
    {20872, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x4824, 0x4444, 0x4284, 0x4104, 0x4284, 0x4444, 0x4824, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20873, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0xFFFE, 0x2008, 0x2008, 0x2008, 0x2028, 0x2010}},
    {20874, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2408, 0x2028, 0x2010}},
    {20875, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20876, 2, {0x0, 0x1E78, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0xFFFE, 0x1248, 0x1248, 0x1248, 0x1248, 0x1248, 0x1288, 0x26A8, 0x4110}},
    {20877, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0xFFFE, 0x2008, 0x2008, 0x2008, 0x2028, 0x2010}},
    {20878, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1F90, 0x1090, 0x1090, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {20879, 2, {0x0, 0x7FFC, 0x4004, 0x4444, 0x4824, 0x5014, 0x67CC, 0x4444, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4004, 0x4014, 0x4008}},
    {20880, 2, {0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {20881, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x17D0, 0x1010, 0x17D0, 0x1010, 0x1050, 0x1020}},
    {20882, 2, {0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x2008, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {20883, 2, {0x820, 0x820, 0x7FFC, 0x820, 0x3FF8, 0x820, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x2008, 0x2028, 0x2010}},
    {20884, 2, {0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x2FE8, 0x0, 0x7DFC, 0x4420, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x4420, 0xA0, 0x40}},
    {20885, 2, {0x7FFC, 0x4004, 0x5FF4, 0x4004, 0x5FF4, 0x400, 0xFE0, 0x1020, 0x3FF8, 0x5108, 0x1108, 0x1FF8, 0x280, 0x482, 0x1882, 0x607E}},
    {20886, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {20887, 2, {0x0, 0x0, 0x7FFE, 0x4002, 0x8004, 0x0, 0xFC0, 0x840, 0x840, 0x840, 0x840, 0x1042, 0x1042, 0x2042, 0x403E, 0x8000}},
    {20888, 2, {0x100, 0x100, 0x100, 0x3FFC, 0x2104, 0x4108, 0x100, 0x280, 0x280, 0x280, 0x480, 0x480, 0x884, 0x1084, 0x207C, 0x4000}},
    {20889, 2, {0x0, 0x7FFE, 0x4002, 0x9004, 0x1000, 0x1FF8, 0x1000, 0x2000, 0x3FF8, 0x8, 0x8, 0xFFC8, 0x8, 0x8, 0x50, 0x20}},
    {20890, 2, {0x0, 0x0, 0x7FFE, 0x4002, 0x8104, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {20891, 2, {0x0, 0x7FFE, 0x4002, 0x8204, 0x200, 0x3FF8, 0x400, 0x900, 0x1100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {20892, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x4204, 0x8208, 0x500, 0x508, 0x890, 0x18A0, 0x2840, 0x4820, 0x8810, 0xA08, 0xC06, 0x800}},
    {20893, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {20894, 2, {0x0, 0x7FFE, 0x4002, 0x8104, 0x1110, 0x920, 0x540, 0x100, 0x7FFC, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {20895, 2, {0x0, 0x7FFE, 0x4102, 0x8204, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x10F0, 0x1F00, 0x1004, 0x1004, 0xFFC, 0x0}},
    {20896, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x3E08, 0x8, 0xFE, 0x7E08, 0x1448, 0x1428, 0x1408, 0x142A, 0x2412, 0x2402, 0x43FE, 0x8000}},
    {20897, 2, {0x0, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x7FFC, 0x200, 0xD08, 0x7190, 0x2A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {20898, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x7FFC, 0x200, 0x608, 0x1910, 0x6AA0, 0x4C0, 0x1BA0, 0x6298, 0xC86, 0x3080, 0xC280, 0x100}},
    {20899, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x7F00, 0x22FC, 0x2244, 0x3E44, 0x2244, 0x2228, 0x3E28, 0x2210, 0x2790, 0xFA28, 0x4244, 0x282}},
    {20900, 2, {0x0, 0x7FFE, 0x4002, 0x8804, 0xFE0, 0x1020, 0x2040, 0x5FF8, 0x1108, 0x1108, 0x1FF8, 0x280, 0x4A0, 0x892, 0x1082, 0x607E}},
    {20901, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1210, 0x100, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {20902, 2, {0x0, 0x0, 0x7FFE, 0x4002, 0x8024, 0x3E38, 0x20, 0x20, 0x7EF8, 0x1408, 0x1450, 0x1420, 0x1454, 0x2484, 0x2404, 0x43FC}},
    {20903, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x820, 0x820, 0x820, 0x7EFC, 0x820, 0x1C30, 0x1A70, 0x2A68, 0x48A4, 0x8922, 0x820, 0x820}},
    {20904, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {20905, 2, {0x0, 0x7FFE, 0x4002, 0x8204, 0x1CF0, 0x1010, 0x1EF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2004, 0x4924, 0x2494, 0x2494, 0x4008}},
    {20906, 2, {0xFFFE, 0x8442, 0x3FF8, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x920, 0x3FF8, 0xD116, 0x1130, 0x100}},
    {20907, 2, {0x0, 0x4000, 0x2000, 0x2000, 0x0, 0x800, 0x800, 0x1000, 0x1000, 0xE000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x0}},
    {20908, 2, {0x400, 0x400, 0xFF0, 0x1010, 0x2820, 0x4440, 0x380, 0xC60, 0x3018, 0xC006, 0x700, 0xC0, 0x20, 0xE00, 0x180, 0x40}},
    {20909, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x1010, 0x2608, 0xC186, 0x40, 0x0, 0xC00, 0x300, 0xC0, 0x20}},
    {20910, 2, {0x0, 0x4000, 0x27FC, 0x2040, 0x40, 0x840, 0x840, 0x1040, 0x1040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x2000, 0x0}},
    {20911, 2, {0x0, 0x47F0, 0x2010, 0x2010, 0x210, 0xA10, 0xA10, 0x13FC, 0x1004, 0xE004, 0x2004, 0x27F4, 0x2004, 0x2004, 0x2028, 0x10}},
    {20912, 2, {0x40, 0x4040, 0x2040, 0x2044, 0x68, 0x770, 0x1160, 0x1150, 0x2150, 0xE248, 0x2248, 0x2444, 0x2842, 0x2040, 0x2140, 0x80}},
    {20913, 2, {0x0, 0x47FE, 0x2080, 0x2080, 0x80, 0x8F8, 0x908, 0x1108, 0x1108, 0x2208, 0xE3F8, 0x2010, 0x2010, 0x2010, 0x27FE, 0x0}},
    {20914, 2, {0x40, 0x4040, 0x2040, 0x2040, 0x7FC, 0x444, 0x1444, 0x1444, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2040, 0x2040, 0x2040, 0x40}},
    {20915, 2, {0x80, 0x4080, 0x2080, 0x27F8, 0x88, 0x88, 0x1088, 0x1088, 0x2FFE, 0xE080, 0x2140, 0x2140, 0x2220, 0x2410, 0x2808, 0x1006}},
    {20916, 2, {0x0, 0x43FC, 0x2010, 0x2110, 0x110, 0x910, 0xA10, 0x13FE, 0x1030, 0xE050, 0x2090, 0x2110, 0x2210, 0x2410, 0x2050, 0x20}},
    {20917, 2, {0x0, 0x47F8, 0x2408, 0x2408, 0x408, 0x408, 0x17F8, 0x1120, 0x2120, 0xE120, 0x2120, 0x2220, 0x2222, 0x2422, 0x281E, 0x1000}},
    {20918, 2, {0x40, 0x4040, 0x2080, 0x2110, 0x208, 0x7FC, 0x1004, 0x1000, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {20919, 2, {0x40, 0x4040, 0x20A0, 0x20A0, 0x110, 0x248, 0x1426, 0x1020, 0x23F8, 0xE008, 0x2010, 0x2110, 0x20A0, 0x2040, 0x2020, 0x20}},
    {20920, 2, {0x40, 0x4244, 0x2148, 0x2150, 0x40, 0x7FC, 0x1040, 0x1040, 0x2040, 0xEFFE, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {20921, 2, {0x90, 0x4088, 0x2088, 0x2080, 0x7FE, 0x80, 0x1100, 0x11FC, 0x2144, 0xE244, 0x2248, 0x2450, 0x2420, 0x2850, 0x2088, 0x306}},
    {20922, 2, {0x0, 0x43FC, 0x2204, 0x2204, 0x204, 0xBFC, 0xA20, 0x1220, 0x13FE, 0xE220, 0x2220, 0x2210, 0x2212, 0x228A, 0x2306, 0x202}},
    {20923, 2, {0x40, 0x4040, 0x2040, 0x27FE, 0x80, 0x920, 0x920, 0x1220, 0x13FC, 0xE020, 0x2128, 0x2124, 0x2222, 0x2422, 0x20A0, 0x40}},
    {20924, 2, {0x40, 0x4240, 0x2240, 0x23FC, 0x440, 0x40, 0x1040, 0x17FE, 0x2120, 0xE120, 0x2120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {20925, 2, {0x4, 0x4004, 0x2FC4, 0x2224, 0x224, 0x3A4, 0x12A4, 0x14A4, 0x26A4, 0xE5A4, 0x2924, 0x2124, 0x2204, 0x2204, 0x2414, 0x808}},
    {20926, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x208, 0x1406, 0x13F8, 0x1000, 0x2000, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {20927, 2, {0x40, 0x4040, 0x23F8, 0x2048, 0x7FE, 0x48, 0x13F8, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x40}},
    {20928, 2, {0x100, 0x4100, 0x23F0, 0x2410, 0x20, 0x17FC, 0x1044, 0x1044, 0x2FFE, 0x2044, 0xE044, 0x27FC, 0x2044, 0x2040, 0x2140, 0x80}},
    {20929, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x40, 0x7FC, 0x1444, 0x1444, 0x27FC, 0xE040, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {20930, 2, {0x100, 0x4100, 0x21F0, 0x2210, 0x220, 0x7FC, 0x1A44, 0x1244, 0x2244, 0xE3FC, 0x22A4, 0x20A0, 0x2120, 0x2222, 0x2422, 0x81E}},
    {20931, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x208, 0x5F6, 0x1040, 0x1040, 0x27FC, 0xE040, 0x2250, 0x2248, 0x2444, 0x2844, 0x2140, 0x80}},
    {20932, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x3F8, 0x48, 0x17FE, 0x1048, 0x23F8, 0xE080, 0x27FE, 0x2108, 0x2390, 0x2060, 0x2198, 0x604}},
    {20933, 2, {0x0, 0x47FC, 0x2444, 0x2444, 0x7FC, 0x444, 0x1444, 0x15F4, 0x2514, 0xE514, 0x2514, 0x25F4, 0x2514, 0x2404, 0x27FC, 0x404}},
    {20934, 2, {0x140, 0x4120, 0x2120, 0x23FE, 0x220, 0x1620, 0x1BFC, 0x1220, 0x2220, 0x23FC, 0xE220, 0x2220, 0x2220, 0x23FE, 0x2200, 0x200}},
    {20935, 2, {0x208, 0x4228, 0x2228, 0x2228, 0xFA4, 0x244, 0x1254, 0x1692, 0x2710, 0xEAA0, 0x2A20, 0x3228, 0x2244, 0x22FC, 0x2244, 0x200}},
    {20936, 2, {0x3C, 0x47C0, 0x2244, 0x2128, 0x0, 0xBF8, 0x848, 0x1048, 0x17FE, 0xE048, 0x2048, 0x23F8, 0x2048, 0x2040, 0x2140, 0x80}},
    {20937, 2, {0x80, 0x4040, 0x2FFE, 0x2000, 0x0, 0x3F8, 0x1208, 0x1208, 0x23F8, 0xE040, 0x2250, 0x2248, 0x2444, 0x2844, 0x2140, 0x80}},
    {20938, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x3F8, 0x1040, 0x17FE, 0x1000, 0x23F8, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x2228, 0x210}},
    {20939, 2, {0x0, 0x47FC, 0x2444, 0x2444, 0x5F4, 0x444, 0x1444, 0x17FC, 0x2404, 0xE5F4, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x808}},
    {20940, 2, {0x40, 0x4040, 0x23F8, 0x2040, 0x40, 0x7FE, 0x1110, 0x1288, 0x2484, 0xE1F0, 0x2310, 0x24A0, 0x2040, 0x20A0, 0x2310, 0xC0C}},
    {20941, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x3F8, 0xA48, 0xBF8, 0x1248, 0x13F8, 0xE040, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {20942, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x208, 0x406, 0x11F0, 0x1040, 0x2040, 0xE3F8, 0x2040, 0x2248, 0x2148, 0x2150, 0x2FFE, 0x0}},
    {20943, 2, {0x14, 0x4012, 0x2010, 0x27FE, 0x410, 0x410, 0x15D0, 0x1412, 0x2412, 0xE5D4, 0x2554, 0x2548, 0x25DA, 0x242A, 0x2846, 0x1082}},
    {20944, 2, {0x0, 0x47FC, 0x20A0, 0x20A0, 0x7FC, 0x4A4, 0x14A4, 0x14A4, 0x27FC, 0xE040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {20945, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x3F8, 0x80, 0x17FC, 0x1110, 0x2208, 0xE5F6, 0x2040, 0x23F8, 0x2040, 0x20A0, 0x2110, 0x608}},
    {20946, 2, {0x40, 0x4248, 0x2248, 0x2248, 0x3F8, 0x0, 0x17FC, 0x1000, 0x23F8, 0xE208, 0x2208, 0x23F8, 0x2110, 0x20A0, 0x27FE, 0x0}},
    {20947, 2, {0x0, 0x47FC, 0x20A0, 0x20A0, 0x7FC, 0x4A4, 0x14A4, 0x17FC, 0x2040, 0xE040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {20948, 2, {0x40, 0x4040, 0x20A0, 0x2190, 0x248, 0x17FE, 0x1A08, 0x13F8, 0x2208, 0x23F8, 0xE200, 0x23F8, 0x2508, 0x2508, 0x29F8, 0x108}},
    {20949, 2, {0x0, 0x47FC, 0x2404, 0x2000, 0x3F8, 0xA08, 0xBF8, 0x1208, 0x13F8, 0xE080, 0x2040, 0x27FC, 0x2000, 0x2110, 0x2208, 0x404}},
    {20950, 2, {0x4140, 0x2120, 0x3FE, 0xA20, 0x17FC, 0xEA20, 0x23FC, 0x2220, 0x23FE, 0x200, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {20951, 2, {0x40, 0x4444, 0x27FC, 0x2120, 0x110, 0x3FE, 0x1220, 0x1620, 0x2BFC, 0xE220, 0x2220, 0x23FC, 0x2220, 0x2220, 0x23FE, 0x200}},
    {20952, 2, {0x480, 0x4482, 0x249C, 0x2FD0, 0x490, 0x790, 0x149E, 0x1794, 0x2494, 0xE494, 0x2FD4, 0x2014, 0x2514, 0x28A4, 0x3024, 0x44}},
    {20953, 2, {0x0, 0x47FC, 0x24A4, 0x27FC, 0x40, 0x3F8, 0x1040, 0x17FC, 0x2110, 0xE0A0, 0x23F8, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x40}},
    {20954, 2, {0x110, 0x4110, 0x27BC, 0x2110, 0x3B8, 0x554, 0x1912, 0x1000, 0x23F8, 0xE000, 0x27FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x80}},
    {20955, 2, {0x40, 0x4FFE, 0x2000, 0x27FC, 0x404, 0x5F4, 0x1514, 0x15F4, 0x2404, 0xE7FC, 0x2000, 0x23F8, 0x2000, 0x2FFE, 0x2248, 0x4C4}},
    {20956, 2, {0x40, 0x4FFE, 0x2000, 0x27FC, 0x404, 0x5F4, 0x1514, 0x15F4, 0x2404, 0xE7FC, 0x2008, 0x23F0, 0x2040, 0x2FFE, 0x2150, 0xE4E}},
    {20957, 2, {0x400, 0x457C, 0x2604, 0x24A8, 0x390, 0x107C, 0x1414, 0x1794, 0x2A50, 0x2250, 0xEFDC, 0x2250, 0x2550, 0x24B0, 0x289E, 0x100}},
    {20958, 2, {0x800, 0x8BDE, 0x4A92, 0x4A92, 0xBD2, 0x2A5E, 0x2A50, 0x2BD0, 0x4A92, 0xCA92, 0x4BCE, 0x5000, 0x4000, 0x4524, 0x4492, 0x892}},
    {20959, 2, {0x40, 0x4FFE, 0x2040, 0x27FC, 0x0, 0x7FC, 0x14A4, 0x17FC, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {20960, 2, {0x0, 0xFE0, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1022, 0x2022, 0x2022, 0x401E, 0x8000}},
    {20961, 2, {0x0, 0xFE0, 0x820, 0x820, 0x820, 0xA20, 0x920, 0x8A0, 0x8A0, 0x820, 0x1020, 0x1022, 0x2022, 0x2022, 0x401E, 0x8000}},
    {20962, 2, {0x200, 0x200, 0x400, 0x1FE0, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1022, 0x2022, 0x2022, 0x401E, 0x8000}},
    {20963, 2, {0x200, 0x200, 0x200, 0x200, 0x1FE0, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1022, 0x2022, 0x2022, 0x401E, 0x8000}},
    {20964, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2FC8, 0x2048, 0x2848, 0x2488, 0x2288, 0x2108, 0x2288, 0x244A, 0x284A, 0x400A, 0x4006, 0x8002}},
    {20965, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x23E0, 0x2220, 0x2220, 0x2220, 0x2220, 0x2422, 0x4422, 0x4822, 0x901E}},
    {20966, 2, {0x1000, 0x10F0, 0x1090, 0x1E90, 0x1290, 0x2290, 0x2290, 0x5290, 0x9494, 0x1494, 0x894, 0x90C, 0x1400, 0x2300, 0x40FE, 0x8000}},
    {20967, 2, {0x0, 0x3FF8, 0x2008, 0x2108, 0x2108, 0x2108, 0x2FE8, 0x2928, 0x2928, 0x2928, 0x29A8, 0x294A, 0x210A, 0x410A, 0x4106, 0x8002}},
    {20968, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2FC8, 0x2008, 0x2008, 0x3FE8, 0x2208, 0x2488, 0x2848, 0x3FEA, 0x282A, 0x400A, 0x4006, 0x8002}},
    {20969, 2, {0x0, 0x3FF8, 0x2008, 0x2108, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x2388, 0x2348, 0x2528, 0x252A, 0x290A, 0x510A, 0x4106, 0x8002}},
    {20970, 2, {0x0, 0x3FF8, 0x2008, 0x2088, 0x2088, 0x2088, 0x2088, 0x2488, 0x24E8, 0x2488, 0x2488, 0x248A, 0x248A, 0x5FFA, 0x4006, 0x8002}},
    {20971, 2, {0x200, 0x400, 0x3FE0, 0x2020, 0x2420, 0x22A0, 0x2040, 0x3FF8, 0x8, 0x1F88, 0x10A8, 0x1090, 0x1082, 0x2082, 0x407E, 0x8000}},
    {20972, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2FE8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x244A, 0x244A, 0x47CA, 0x4446, 0x8002}},
    {20973, 2, {0x838, 0xBC0, 0x1040, 0x3040, 0x57FE, 0x9040, 0x1040, 0x1040, 0x13F8, 0x0, 0xFE0, 0x820, 0x820, 0x1022, 0x2022, 0xC01E}},
    {20974, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2FE8, 0x2108, 0x2208, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x244A, 0x244A, 0x47CA, 0x4446, 0x8002}},
    {20975, 2, {0x1000, 0x92F0, 0x9290, 0x9290, 0xFE90, 0x90, 0xFE90, 0x290, 0x290, 0x7E90, 0x4090, 0x4092, 0x4692, 0x5912, 0x610E, 0x200}},
    {20976, 2, {0x0, 0x3FF8, 0x2208, 0x27C8, 0x2448, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x2FE8, 0x2108, 0x27CA, 0x210A, 0x4FEA, 0x4006, 0x8002}},
    {20977, 2, {0x800, 0x4900, 0x4978, 0x7F48, 0x48, 0xFF48, 0x48, 0x7F48, 0x4148, 0x4148, 0x7F48, 0x4248, 0x244A, 0xF4A, 0xF08A, 0x4106}},
    {20978, 2, {0x0, 0x3FF8, 0x2448, 0x2288, 0x3FF8, 0x2288, 0x2FE8, 0x22A8, 0x3FF8, 0x22A8, 0x2FE8, 0x228A, 0x26CA, 0x4AAA, 0x5296, 0x8282}},
    {20979, 2, {0x90, 0x7EA0, 0x444, 0x2BA8, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x440, 0x7FFC, 0x0, 0xFE0, 0x820, 0x822, 0x1022, 0x601E}},
    {20980, 2, {0x0, 0x43FC, 0x2240, 0xBF8, 0xA40, 0x13F8, 0x2240, 0xE3FC, 0x2004, 0x2554, 0x2008, 0xFE0, 0x820, 0x822, 0x1022, 0x601E}},
    {20981, 2, {0x0, 0x0, 0x0, 0x0, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x7FFC, 0x4}},
    {20982, 2, {0x0, 0x10, 0x1010, 0x820, 0x4444, 0x4284, 0x4104, 0x4284, 0x4444, 0x4824, 0x5014, 0x6014, 0x4004, 0x4004, 0x7FFC, 0x4}},
    {20983, 2, {0x100, 0x100, 0x100, 0x100, 0x4FE4, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4004, 0x4004, 0x7FFC, 0x4}},
    {20984, 2, {0x0, 0xFE0, 0x820, 0x820, 0x820, 0x820, 0x820, 0x783C, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {20985, 2, {0x0, 0x7C7C, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4004, 0x4004, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {20986, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x108, 0x100, 0x100, 0x4104, 0x4104, 0x4104, 0x4104, 0x7FFC, 0x4}},
    {20987, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {20988, 2, {0x100, 0x100, 0x100, 0x110, 0x5D24, 0x4544, 0x4584, 0x4944, 0x4924, 0x5114, 0x6114, 0x4504, 0x4204, 0x4004, 0x7FFC, 0x4}},
    {20989, 2, {0x0, 0x3FF0, 0x20, 0x40, 0x5194, 0x4924, 0x4544, 0x4104, 0x4544, 0x4924, 0x5114, 0x4504, 0x4204, 0x4004, 0x7FFC, 0x4}},
    {20990, 2, {0x0, 0x3FF0, 0x20, 0x40, 0x4184, 0x5D74, 0x5594, 0x5554, 0x5524, 0x5D54, 0x4194, 0x4504, 0x4204, 0x4004, 0x7FFC, 0x4}},
    {20991, 2, {0x440, 0x2448, 0x1450, 0x440, 0xFFFE, 0x0, 0x440, 0x2288, 0x2FE8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {20992, 2, {0x0, 0x0, 0x3FF8, 0x208, 0x208, 0x208, 0x208, 0x208, 0x208, 0x408, 0x408, 0x808, 0x808, 0x1008, 0x2050, 0x4020}},
    {20993, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x8, 0x48, 0x88, 0x108, 0x208, 0x408, 0x808, 0x1008, 0x6008, 0x8, 0x50, 0x20}},
    {20994, 2, {0x4, 0x4, 0x4, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x4, 0x4, 0x4, 0x14, 0x8}},
    {20995, 2, {0x0, 0x3FF8, 0x208, 0x208, 0x208, 0x2208, 0x2208, 0x2208, 0x4208, 0x208, 0x408, 0x408, 0x808, 0x808, 0x1050, 0x6020}},
    {20996, 2, {0x0, 0x7FF0, 0x410, 0x410, 0x2410, 0x1410, 0xC10, 0x410, 0x610, 0x590, 0x870, 0x81E, 0x1010, 0x1010, 0x20A0, 0xC040}},
    {20997, 2, {0x0, 0x7FE0, 0x420, 0x420, 0x420, 0x420, 0x2428, 0x2424, 0x4422, 0x8422, 0x820, 0x820, 0x1020, 0x1020, 0x2140, 0xC080}},
    {20998, 2, {0x40, 0x440, 0x420, 0x820, 0x1010, 0x2008, 0x4004, 0x9FE2, 0x420, 0x420, 0x420, 0x820, 0x820, 0x1020, 0x2140, 0x4080}},
    {20999, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0x1044, 0x1E44, 0xF044, 0x1044, 0x1044, 0x1244, 0x1444, 0x1884, 0x1084, 0x104, 0x228, 0x410}},
    {21000, 2, {0x4, 0x104, 0x104, 0x2124, 0x1124, 0xA24, 0xA24, 0x424, 0x424, 0xA24, 0xA24, 0x1124, 0x2104, 0x4004, 0x8014, 0x8}},
    {21001, 2, {0x1004, 0x1004, 0x2004, 0x3FA4, 0x4024, 0x8024, 0x3F24, 0x224, 0x424, 0x824, 0x1024, 0x2024, 0x4084, 0x4084, 0x3F94, 0x8}},
    {21002, 2, {0x4, 0x7F04, 0x804, 0x824, 0x824, 0x824, 0x824, 0xFFA4, 0x824, 0x824, 0x824, 0x824, 0x804, 0x804, 0x814, 0x808}},
    {21003, 2, {0x4, 0x704, 0x7804, 0x824, 0x824, 0x824, 0x824, 0xFFA4, 0x824, 0x824, 0x1024, 0x1024, 0x2004, 0x2004, 0x4014, 0x8008}},
    {21004, 2, {0x404, 0x404, 0x404, 0x424, 0xFFA4, 0x424, 0x424, 0x424, 0x4424, 0x2424, 0x2424, 0x424, 0x404, 0x404, 0x1414, 0x808}},
    {21005, 2, {0x400, 0x400, 0xFF0, 0x1010, 0x2020, 0x4040, 0xBFF8, 0x8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x8, 0x3FF8, 0x8}},
    {21006, 2, {0x2004, 0x2004, 0x2004, 0x3FA4, 0x54A4, 0x94A4, 0x14A4, 0x24A4, 0x24A4, 0x48A4, 0x88A4, 0x10A4, 0x2084, 0x4084, 0x8514, 0x208}},
    {21007, 2, {0x2004, 0x2004, 0x3F84, 0x4024, 0x8024, 0x3F24, 0x24, 0x24, 0x7F24, 0x124, 0x124, 0x124, 0x144, 0x144, 0xD4, 0x48}},
    {21008, 2, {0x4, 0x3F04, 0x2104, 0x2124, 0x2924, 0x2524, 0x2524, 0x2124, 0xFFE4, 0x2124, 0x2124, 0x2124, 0x2104, 0x4104, 0x4514, 0x8208}},
    {21009, 2, {0x4, 0x7F84, 0x1204, 0x1224, 0x1224, 0x1224, 0x1224, 0xFFE4, 0x1224, 0x1224, 0x1224, 0x1224, 0x2204, 0x2204, 0x4214, 0x8208}},
    {21010, 2, {0x1004, 0x1404, 0x1204, 0x1024, 0x13A4, 0xFC24, 0x1024, 0x1124, 0x1124, 0x1224, 0xC24, 0x8A4, 0x1484, 0x2284, 0xC194, 0x88}},
    {21011, 2, {0x4, 0x3F04, 0x4, 0x24, 0x24, 0xFFA4, 0x1224, 0x1224, 0x1224, 0x1224, 0x1224, 0x22A4, 0x2304, 0x4204, 0x8014, 0x8}},
    {21012, 2, {0x804, 0x804, 0x804, 0x7F24, 0x924, 0x924, 0x924, 0x924, 0xFFE4, 0x824, 0x824, 0x1424, 0x1204, 0x2104, 0x4114, 0x8008}},
    {21013, 2, {0x0, 0x3FF8, 0x208, 0x208, 0x408, 0x850, 0x3020, 0x0, 0x7EFC, 0x1224, 0x1224, 0x1224, 0x2244, 0x2244, 0x4A94, 0x8508}},
    {21014, 2, {0x4, 0x3F04, 0x2104, 0x2124, 0x2124, 0x3F24, 0x2124, 0x2124, 0x2124, 0x3F24, 0x2124, 0x2124, 0x2104, 0x4104, 0x4514, 0x8208}},
    {21015, 2, {0x4, 0x7F84, 0x804, 0x824, 0x1F24, 0x1124, 0x2124, 0x2124, 0x5224, 0x8A24, 0x424, 0x424, 0x804, 0x1004, 0x2014, 0x4008}},
    {21016, 2, {0x1004, 0x804, 0x804, 0xFFA4, 0x224, 0x4224, 0x2224, 0x1424, 0x1424, 0x824, 0x824, 0x1424, 0x2404, 0x4204, 0x8214, 0x8}},
    {21017, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x1424, 0x1204, 0x2104, 0x4094, 0x8088}},
    {21018, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x6524, 0x5524, 0x5524, 0x4924, 0x4924, 0x5524, 0x5524, 0x6524, 0x4104, 0x4104, 0x4514, 0x4208}},
    {21019, 2, {0x804, 0x804, 0x1404, 0x1224, 0x2124, 0x40A4, 0xBE24, 0x2224, 0x2224, 0x2224, 0x2A24, 0x2424, 0x2084, 0x2084, 0x1F94, 0x8}},
    {21020, 2, {0x1202, 0x1202, 0x1202, 0x7F82, 0x1292, 0x1292, 0x7F92, 0x5212, 0x5212, 0x7FD2, 0x1252, 0x1252, 0x2342, 0x2282, 0x420A, 0x8204}},
    {21021, 2, {0x2000, 0x1000, 0x1FC, 0xFC44, 0x844, 0x1044, 0x1044, 0x3444, 0x5844, 0x9444, 0x1444, 0x1084, 0x1084, 0x1104, 0x1228, 0x1410}},
    {21022, 2, {0x4, 0x3F04, 0x2104, 0x2124, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x2124, 0x2124, 0x2124, 0x2784, 0xF804, 0x4014, 0x8}},
    {21023, 2, {0x4, 0x7F04, 0x1104, 0x1124, 0x2124, 0x2524, 0x4224, 0x8024, 0x3F24, 0x2124, 0x2124, 0x2124, 0x2104, 0x3F04, 0x2114, 0x8}},
    {21024, 2, {0x2, 0x7BC2, 0x4A42, 0x4A4A, 0x4A4A, 0x4A4A, 0x4A4A, 0xFFEA, 0x4A4A, 0x4A4A, 0x4A4A, 0x4A4A, 0x4A42, 0x5A42, 0x854A, 0x884}},
    {21025, 2, {0x2, 0x7F82, 0x4082, 0x4082, 0x4092, 0x7F92, 0x4412, 0x4412, 0x7FD2, 0x4412, 0x4412, 0x4212, 0x4242, 0x5142, 0x60CA, 0x4044}},
    {21026, 2, {0x804, 0x804, 0x1404, 0x1224, 0x2124, 0x48A4, 0x8424, 0x424, 0x7F24, 0x124, 0x2224, 0x1424, 0x804, 0x404, 0x414, 0x8}},
    {21027, 2, {0x804, 0x804, 0x804, 0x1024, 0x1224, 0x2124, 0x7FA4, 0x20A4, 0x24, 0x3F24, 0x2124, 0x2124, 0x2104, 0x2104, 0x3F14, 0x2108}},
    {21028, 2, {0x404, 0x4484, 0x2484, 0x2524, 0x424, 0x7FA4, 0x424, 0x424, 0x424, 0xFFE4, 0x424, 0x824, 0x804, 0x1004, 0x2014, 0x4008}},
    {21029, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x4124, 0x7F24, 0x24, 0x1024, 0x1024, 0x1F24, 0x1124, 0x1124, 0x2104, 0x2104, 0x4A14, 0x8408}},
    {21030, 2, {0x804, 0x804, 0x804, 0x7F24, 0x824, 0x824, 0x824, 0xFFA4, 0x824, 0x1024, 0x1024, 0x2224, 0x4104, 0xFF84, 0x4094, 0x8}},
    {21031, 2, {0x800, 0x800, 0x8FC, 0x7F24, 0x824, 0x824, 0x824, 0xFFA4, 0x1024, 0x1024, 0x2424, 0x4224, 0xFF44, 0x4144, 0x94, 0x108}},
    {21032, 2, {0x1004, 0x1004, 0x3F04, 0x2124, 0x4124, 0xBD24, 0x2524, 0x2524, 0x3D24, 0x2124, 0x2524, 0x2224, 0x2084, 0x2084, 0x1F94, 0x8}},
    {21033, 2, {0x104, 0x784, 0x7C04, 0x424, 0x424, 0x424, 0xFFA4, 0xC24, 0x1624, 0x1524, 0x24A4, 0x2424, 0x4404, 0x8404, 0x414, 0x408}},
    {21034, 2, {0x4, 0x7F04, 0x5504, 0x5524, 0x5524, 0x5524, 0x5524, 0xFFA4, 0x5524, 0x5524, 0x5524, 0x5524, 0x5504, 0x5104, 0x4514, 0x4208}},
    {21035, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x4124, 0x7F24, 0x1024, 0x1024, 0xFF24, 0x1124, 0x1124, 0x1124, 0x2104, 0x2104, 0x4A14, 0x8408}},
    {21036, 2, {0x1404, 0x1204, 0x1004, 0x1724, 0x7824, 0x1024, 0x17A4, 0xF824, 0x1124, 0x1224, 0xC24, 0x8A4, 0x1484, 0x2284, 0xC194, 0x88}},
    {21037, 2, {0x4, 0x7F04, 0x204, 0x424, 0xC24, 0x1224, 0x2124, 0xC0A4, 0x24, 0x7F24, 0x824, 0x824, 0x804, 0xF04, 0xF014, 0x4008}},
    {21038, 2, {0x204, 0xF04, 0x7804, 0x824, 0x824, 0xFFA4, 0x824, 0x824, 0x824, 0x7F24, 0x4124, 0x4124, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {21039, 2, {0x4, 0x7F84, 0x4, 0x24, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x2124, 0x4, 0x784, 0xF814, 0x4008}},
    {21040, 2, {0x4, 0xFF84, 0x804, 0x1024, 0x2224, 0x4124, 0xFFA4, 0x8A4, 0x824, 0x824, 0x7F24, 0x824, 0x804, 0xF84, 0xF814, 0x4008}},
    {21041, 2, {0x2400, 0x2400, 0x25F8, 0x2448, 0x7E48, 0x2448, 0x254C, 0x254A, 0xFE4A, 0x2448, 0x2448, 0x2448, 0x2448, 0x4488, 0x44A8, 0x8510}},
    {21042, 2, {0x804, 0x804, 0x7F04, 0x824, 0x824, 0x824, 0xFFA4, 0x24, 0x824, 0x824, 0x7F24, 0x824, 0x804, 0xF04, 0xF014, 0x4008}},
    {21043, 2, {0x804, 0x804, 0xFF84, 0x1424, 0x2224, 0x4124, 0xBEA4, 0x24, 0xFFA4, 0x1024, 0x2024, 0x7F24, 0x104, 0x104, 0xA14, 0x408}},
    {21044, 2, {0x4, 0x7F04, 0x1104, 0x1224, 0x23A4, 0x20A4, 0x4AA4, 0x8924, 0x824, 0xFFA4, 0x1824, 0x2C24, 0x2A04, 0x4904, 0x8894, 0x808}},
    {21045, 2, {0x4, 0xFF84, 0x2204, 0x2224, 0x3E24, 0x2224, 0x2224, 0x3E24, 0x2224, 0x2224, 0x27A4, 0xFA24, 0x4204, 0x204, 0x214, 0x208}},
    {21046, 2, {0x404, 0x2404, 0x2404, 0x3FA4, 0x4424, 0x424, 0xFFE4, 0x424, 0x424, 0x3FA4, 0x24A4, 0x24A4, 0x2684, 0x2504, 0x414, 0x408}},
    {21047, 2, {0x2, 0x3FC2, 0x2042, 0x2042, 0x3FD2, 0x2212, 0x2212, 0x2212, 0x3FD2, 0x5252, 0x5252, 0x5252, 0x9342, 0x1282, 0x20A, 0x204}},
    {21048, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x820, 0x1140, 0x2080}},
    {21049, 2, {0x104, 0x6204, 0x1404, 0x824, 0x1424, 0x2224, 0x4924, 0x824, 0xFFA4, 0x824, 0x2A24, 0x2924, 0x4884, 0x8884, 0x2814, 0x1008}},
    {21050, 2, {0x804, 0x804, 0xFF84, 0x824, 0x824, 0x7F24, 0x4924, 0x4924, 0x4D24, 0x4A24, 0x1C24, 0x2A24, 0x4904, 0x8884, 0x814, 0x808}},
    {21051, 2, {0x1004, 0x804, 0x4, 0xFFA4, 0x824, 0x1124, 0x2124, 0x7E24, 0x424, 0x924, 0x1224, 0x2424, 0xCC04, 0x1204, 0x2114, 0xC088}},
    {21052, 2, {0x1000, 0x1000, 0x10FC, 0x7E24, 0x1024, 0x10A4, 0x10A4, 0xFEA4, 0x10A4, 0x1124, 0x2024, 0x2824, 0x4444, 0xFE44, 0x4294, 0x108}},
    {21053, 2, {0x804, 0x804, 0x1404, 0x2224, 0x4124, 0xBEA4, 0x24, 0x24, 0xFFA4, 0x1024, 0x1024, 0x2224, 0x4104, 0xFF04, 0x4114, 0x8}},
    {21054, 2, {0x804, 0x804, 0x804, 0xFFA4, 0x824, 0x4924, 0x2924, 0x2A24, 0xFFA4, 0x824, 0x824, 0x1424, 0x1204, 0x2104, 0x4114, 0x8008}},
    {21055, 2, {0x804, 0x4904, 0x4904, 0x4924, 0x7F24, 0x1024, 0x1024, 0x3F24, 0x4124, 0xA124, 0x1224, 0x1424, 0x804, 0x1004, 0x2014, 0xC008}},
    {21056, 2, {0x804, 0x4904, 0x4904, 0x4924, 0x7F24, 0x24, 0x7F24, 0x124, 0x124, 0x3F24, 0x2024, 0x2024, 0x2304, 0x2C04, 0x3014, 0x8}},
    {21057, 2, {0x4, 0x3E04, 0x2204, 0x2224, 0x2224, 0x49A4, 0x8824, 0x824, 0xFFA4, 0x824, 0x1C24, 0x2A24, 0x4904, 0x8884, 0x814, 0x808}},
    {21058, 2, {0x1004, 0x804, 0xFF84, 0x224, 0x6424, 0x1824, 0x2624, 0xC124, 0x22A4, 0x2224, 0x2224, 0x2224, 0x2204, 0x4204, 0x4214, 0x8208}},
    {21059, 2, {0x4204, 0x2204, 0x2404, 0xFF24, 0x924, 0x924, 0xFF24, 0x8824, 0x8824, 0xFFA4, 0x18A4, 0x28A4, 0x4A84, 0x8904, 0x814, 0x808}},
    {21060, 2, {0x4, 0x7F84, 0x4, 0x24A4, 0x4924, 0x9224, 0x4924, 0x24A4, 0x24, 0x7FA4, 0x824, 0x824, 0x804, 0xB84, 0xFC14, 0x8}},
    {21061, 2, {0x4, 0x7F84, 0x4, 0x24, 0x3F24, 0x2124, 0x2124, 0x2124, 0x3F24, 0x24, 0x2124, 0x1124, 0x1204, 0x784, 0x7814, 0x2008}},
    {21062, 2, {0x804, 0x404, 0x3F04, 0x2124, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x2024, 0x2424, 0x2224, 0x2504, 0x2904, 0x3014, 0x2008}},
    {21063, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x24, 0x2204, 0x2104, 0x4094, 0x8088}},
    {21064, 2, {0x4, 0x3F04, 0x2104, 0x2124, 0x3F24, 0x24, 0x7FA4, 0x40A4, 0x40A4, 0x7FA4, 0x40A4, 0x40A4, 0x7F84, 0x4084, 0x4294, 0x4108}},
    {21065, 2, {0x804, 0x804, 0x2A04, 0x2A24, 0x2A24, 0x5D24, 0x88A4, 0x824, 0x824, 0x7F24, 0x824, 0x824, 0x804, 0xF84, 0xF814, 0x4008}},
    {21066, 2, {0x804, 0x4904, 0x2904, 0x2A24, 0x824, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x4104, 0x4104, 0x4514, 0x4208}},
    {21067, 2, {0x808, 0x808, 0xFF48, 0x848, 0x848, 0x7F48, 0x4148, 0x4148, 0x7F48, 0x1448, 0x1408, 0x1428, 0x2412, 0x2402, 0x43FE, 0x8000}},
    {21068, 2, {0x804, 0x804, 0xFF84, 0x824, 0x824, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x824, 0x1C24, 0x2A24, 0x4904, 0x8804, 0x814, 0x808}},
    {21069, 2, {0x1010, 0x810, 0x820, 0xFFFE, 0x0, 0x3E08, 0x2248, 0x2248, 0x3E48, 0x2248, 0x2248, 0x3E48, 0x2208, 0x2208, 0x2A28, 0x2410}},
    {21070, 2, {0x204, 0x6404, 0x1804, 0x1624, 0x2124, 0xCA24, 0x924, 0x824, 0xFFA4, 0x1824, 0x1C24, 0x2A24, 0x2904, 0x4884, 0x8894, 0x808}},
    {21071, 2, {0x4200, 0x2400, 0xFC, 0x7E24, 0x2424, 0x24A4, 0x24A4, 0x24A4, 0xFEA4, 0x2524, 0x2424, 0x2424, 0x2444, 0x4444, 0x4494, 0x8508}},
    {21072, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x4124, 0x7F24, 0x824, 0x824, 0xFFA4, 0x88A4, 0x94A4, 0xA2A4, 0xC284, 0x8084, 0x8294, 0x8108}},
    {21073, 2, {0x804, 0x804, 0x1404, 0x1224, 0x2124, 0x40A4, 0xBE24, 0x24, 0x1124, 0x924, 0x4924, 0x2224, 0x2204, 0x784, 0x7814, 0x2008}},
    {21074, 2, {0x1204, 0x1204, 0x1204, 0x7F24, 0x1224, 0x1224, 0xFFA4, 0x24, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x2104, 0x2104, 0x3F14, 0x2108}},
    {21075, 2, {0xC80, 0x7080, 0x11FC, 0xFD24, 0x3224, 0x3844, 0x5484, 0x9104, 0x1228, 0x1010, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21076, 2, {0x4, 0x3F04, 0x2104, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x1024, 0x3FA4, 0x4AA4, 0x92A4, 0x2484, 0x4884, 0x1294, 0x2108}},
    {21077, 2, {0x1202, 0x1202, 0x1202, 0xF3C2, 0x1212, 0x1212, 0x1212, 0x7392, 0x1212, 0x1212, 0x1212, 0xF3D2, 0x1202, 0x1202, 0x120A, 0x1204}},
    {21078, 2, {0x1004, 0x804, 0x7F84, 0x24, 0x2124, 0x1224, 0xFFE4, 0x24, 0x24, 0x3F24, 0x2124, 0x2124, 0x2104, 0x3F04, 0x2114, 0x8}},
    {21079, 2, {0x2404, 0x2204, 0x2F04, 0xF024, 0x1424, 0x924, 0x3524, 0xCB24, 0x2424, 0x2F24, 0xF024, 0x1224, 0x1404, 0x904, 0x3514, 0xC308}},
    {21080, 2, {0x2204, 0x2204, 0xFF84, 0x2224, 0x2224, 0x3E24, 0x2224, 0x2224, 0x3E24, 0x2224, 0x2224, 0xFFA4, 0x4, 0x2204, 0x4114, 0x8088}},
    {21081, 2, {0x4200, 0x2400, 0x1F8, 0x7E48, 0x2448, 0x2448, 0x254C, 0x254A, 0xFE4A, 0x2448, 0x2448, 0x2448, 0x2448, 0x4488, 0x44A8, 0x8510}},
    {21082, 2, {0x402, 0x402, 0xFFC2, 0x402, 0x3F92, 0x2492, 0x3F92, 0x412, 0x3F92, 0x492, 0xFFD2, 0x492, 0x3F82, 0x402, 0x140A, 0x804}},
    {21083, 2, {0x2, 0x7FC2, 0x4042, 0x5142, 0x4A52, 0x7FD2, 0x4452, 0x4452, 0x5552, 0x5552, 0x5552, 0x5F52, 0x4042, 0x4042, 0x414A, 0x4084}},
    {21084, 2, {0x804, 0x404, 0x7F84, 0x40A4, 0x2024, 0x3BA4, 0x2AA4, 0x2AA4, 0x2AA4, 0x5AA4, 0x4BA4, 0x8A24, 0x1204, 0x1284, 0x2314, 0x4208}},
    {21085, 2, {0x2004, 0x3E04, 0x4204, 0x7C24, 0x424, 0xFFA4, 0x824, 0x924, 0x4924, 0x2A24, 0x1C24, 0x2A24, 0xC904, 0x904, 0x2814, 0x1008}},
    {21086, 2, {0x804, 0x804, 0x7F84, 0x1424, 0x2224, 0x4124, 0xFFE4, 0x124, 0x7924, 0x4924, 0x4924, 0x4924, 0x7904, 0x104, 0x514, 0x208}},
    {21087, 2, {0x4, 0x7B84, 0x884, 0x2AA4, 0x1124, 0x2AA4, 0x44A4, 0x8824, 0x24, 0x7BA4, 0x8A4, 0x2AA4, 0x1104, 0x2A84, 0x4494, 0x8808}},
    {21088, 2, {0x1004, 0x804, 0xFF84, 0x24, 0x24, 0x7F24, 0x4124, 0x4124, 0x4124, 0x7F24, 0x824, 0x2A24, 0x4904, 0x8884, 0x2814, 0x1008}},
    {21089, 2, {0x804, 0x904, 0x4904, 0x4A24, 0x8824, 0x1424, 0x2224, 0x4924, 0x8824, 0x4924, 0x4924, 0x8A24, 0x1404, 0x2204, 0x4114, 0x8088}},
    {21090, 2, {0x4, 0xFF84, 0x804, 0x1024, 0x2824, 0xC4A4, 0xD24, 0x5624, 0x2524, 0xD524, 0xCA4, 0x14A4, 0x2404, 0xC404, 0x1414, 0x808}},
    {21091, 2, {0x804, 0x804, 0x1404, 0x2224, 0x4124, 0xBEA4, 0x824, 0x824, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x804, 0x1404, 0x2214, 0x4108}},
    {21092, 2, {0x1004, 0x804, 0xFF84, 0x224, 0x6424, 0x1824, 0x2624, 0xC124, 0x22A4, 0x3E24, 0x2224, 0x2224, 0x3E04, 0x4204, 0x4214, 0x8208}},
    {21093, 2, {0x4, 0x7E04, 0x204, 0x3E24, 0x224, 0xFFA4, 0x824, 0x924, 0x4924, 0x2A24, 0x1C24, 0x2A24, 0xC904, 0x904, 0x2814, 0x1008}},
    {21094, 2, {0x804, 0x804, 0xFF84, 0x1424, 0x2224, 0x4924, 0x88A4, 0x7F24, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x804, 0xA04, 0xC14, 0x808}},
    {21095, 2, {0x4, 0x3F84, 0x2084, 0x20A4, 0x3FA4, 0x2424, 0x2424, 0x3FA4, 0x2424, 0x2424, 0x3FA4, 0x50A4, 0x5084, 0x9084, 0x1F94, 0x1088}},
    {21096, 2, {0x804, 0x804, 0xFF84, 0x824, 0x7F24, 0x824, 0xFFA4, 0x824, 0xFFA4, 0x1024, 0x2024, 0x7F24, 0xA104, 0x2104, 0x3F14, 0x2108}},
    {21097, 2, {0x702, 0x7802, 0x802, 0xFF82, 0x2A12, 0x2A92, 0xEB12, 0x2A92, 0x6A92, 0xA992, 0x1C12, 0x2A12, 0x4902, 0x8882, 0x80A, 0x804}},
    {21098, 2, {0x820, 0xFFFE, 0x0, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2618, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21099, 2, {0x402, 0x202, 0x3FE2, 0x2482, 0x2492, 0x3FF2, 0x2492, 0x2792, 0x2012, 0x2FD2, 0x2052, 0x2492, 0x2282, 0x4102, 0x468A, 0x9844}},
    {21100, 2, {0x804, 0x4904, 0x4904, 0x4924, 0x7F24, 0x24, 0xFFA4, 0x824, 0x1024, 0x7F24, 0x5524, 0x5524, 0x5504, 0x5504, 0x5514, 0x4308}},
    {21101, 2, {0x4, 0x3F84, 0x2084, 0x20A4, 0x3FA4, 0x2024, 0x3FA4, 0x2424, 0x2924, 0x2FA4, 0x2224, 0x2FA4, 0x2204, 0x4384, 0x5E14, 0x8808}},
    {21102, 2, {0x4, 0x7F04, 0x4104, 0x4124, 0x7924, 0x4924, 0x4924, 0xFFA4, 0x80A4, 0xBEA4, 0xA2A4, 0xA2A4, 0xBE84, 0x8084, 0x8294, 0x8108}},
    {21103, 2, {0x2, 0x7FC2, 0x2, 0x3F82, 0x2092, 0x2092, 0x3F92, 0x12, 0x7FD2, 0x4452, 0x4452, 0x7FD2, 0x4442, 0x4442, 0x7FCA, 0x4044}},
    {21104, 2, {0x382, 0x3C02, 0x402, 0x7FC2, 0x1512, 0x1512, 0xFFD2, 0x1512, 0x1512, 0x7FD2, 0xE12, 0x1512, 0x2482, 0x4442, 0x840A, 0x404}},
    {21105, 2, {0x800, 0x800, 0x147E, 0x2212, 0x4112, 0xBED2, 0x852, 0x852, 0x7F52, 0x4992, 0x4912, 0x7F12, 0x822, 0x1422, 0x224A, 0x4184}},
    {21106, 2, {0x1004, 0x804, 0xFF84, 0x80A4, 0x924, 0x7F24, 0x824, 0x7F24, 0x824, 0xFFA4, 0x824, 0x7F24, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {21107, 2, {0x2204, 0x2204, 0xFF84, 0x2224, 0x824, 0x1424, 0x2224, 0x4124, 0xBEA4, 0x24, 0x24, 0x7F24, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {21108, 2, {0x804, 0x4904, 0x4904, 0x7F24, 0x24, 0xFF24, 0x24, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x4224, 0x2404, 0xF04, 0xF014, 0x4008}},
    {21109, 2, {0x804, 0xC04, 0x1204, 0x2924, 0x44A4, 0xBF24, 0x2124, 0x3F24, 0x2124, 0x3F24, 0x2024, 0x3F24, 0x5104, 0x5104, 0x9F14, 0x1108}},
    {21110, 2, {0x2004, 0x3F04, 0x4104, 0x7E24, 0x224, 0xFFA4, 0x1024, 0x28A4, 0xCD24, 0x1624, 0x2524, 0xCD24, 0x1484, 0x2484, 0xD414, 0x808}},
    {21111, 2, {0x804, 0x404, 0x3F84, 0x1124, 0xA24, 0x3FA4, 0x2224, 0x2A24, 0x2FA4, 0x3224, 0x2224, 0x2FA4, 0x2204, 0x4204, 0x4394, 0x9C08}},
    {21112, 2, {0x804, 0xFF84, 0x804, 0x7F24, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x8A4, 0xFFA4, 0x224, 0xFFA4, 0x4204, 0x2204, 0xA14, 0x408}},
    {21113, 2, {0x4, 0x7704, 0x1104, 0x5524, 0x3324, 0x5524, 0xC24, 0x3324, 0xC4A4, 0x1824, 0x6224, 0xCA4, 0x7104, 0x604, 0x1814, 0xE008}},
    {21114, 2, {0x820, 0x820, 0x7E7E, 0x8C4, 0xFF28, 0x1C10, 0x2A28, 0x48C6, 0x0, 0x3FFE, 0x2000, 0x2FF8, 0x2208, 0x4208, 0x4428, 0x9810}},
    {21115, 2, {0x202, 0x2222, 0x2222, 0x3FEA, 0xA, 0x3DEA, 0x252A, 0x252A, 0x3DEA, 0x252A, 0x252A, 0x3DEA, 0x2522, 0x2522, 0x55AA, 0x8A44}},
    {21116, 2, {0x1004, 0x2204, 0x7F04, 0x4224, 0x94A4, 0xF7A4, 0x824, 0x3624, 0xC924, 0x30A4, 0xC424, 0x1924, 0x6204, 0xC04, 0x3014, 0xC008}},
    {21117, 2, {0x2, 0xFFE2, 0xA02, 0x7FC2, 0x4A52, 0x4A52, 0x7FD2, 0x12, 0x7FD2, 0x12, 0xFFF2, 0x412, 0x2482, 0x4442, 0x942A, 0x804}},
    {21118, 2, {0x2, 0x7FC2, 0x4002, 0x4F82, 0x4892, 0x4892, 0x4F92, 0x4012, 0x5DD2, 0x5552, 0x5552, 0x5552, 0x5DC2, 0x4002, 0x7FEA, 0x4}},
    {21119, 2, {0x4904, 0x9204, 0x4904, 0x24, 0x7F24, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x824, 0xFFA4, 0x1C24, 0x2A04, 0x4904, 0x8814, 0x808}},
    {21120, 2, {0x4, 0x7F04, 0x1204, 0xC24, 0xFFA4, 0x14A4, 0x2524, 0xD424, 0x824, 0xFFA4, 0x94A4, 0xA2A4, 0xDD84, 0x9484, 0x9C94, 0x8188}},
    {21121, 2, {0x1202, 0x1102, 0x3FC2, 0x2202, 0x6212, 0xBFD2, 0x2212, 0x2212, 0x3FD2, 0x2212, 0x2212, 0x3FD2, 0x2002, 0x5482, 0x4A4A, 0x8A44}},
    {21122, 2, {0x2, 0x7FF2, 0x4002, 0x514A, 0x4A4A, 0x5F7A, 0x449A, 0x442A, 0x550A, 0x554A, 0x5F4A, 0x444A, 0x44A2, 0x48A2, 0x492A, 0x9204}},
    {21123, 2, {0x804, 0x7F04, 0x904, 0xFFA4, 0x924, 0x7F24, 0x824, 0xFFA4, 0x24, 0x7F24, 0x4924, 0x7F24, 0x4904, 0x7F04, 0x14, 0xFF88}},
    {21124, 2, {0x4204, 0x4204, 0x7BC4, 0x4524, 0xA8A4, 0x1424, 0x2224, 0x4124, 0xBEA4, 0x24, 0x24, 0x7F24, 0x4104, 0x4104, 0x7F14, 0x4108}},
    {21125, 2, {0x4, 0x7F04, 0x5504, 0x5524, 0x7F24, 0x2024, 0x7FA4, 0x90A4, 0x7CA4, 0x54A4, 0x54A4, 0x7CA4, 0x1284, 0xFE84, 0x294, 0x108}},
    {21126, 2, {0x202, 0x3FF2, 0x2882, 0x250A, 0x3FEA, 0x250A, 0x2FCA, 0x254A, 0x3FEA, 0x254A, 0x2FCA, 0x250A, 0x2D82, 0x5542, 0x652A, 0x8504}},
    {21127, 2, {0x804, 0xF04, 0x804, 0x7FA4, 0x48A4, 0x7E24, 0x4924, 0x4724, 0x4024, 0x7FA4, 0x4824, 0x5524, 0x4A04, 0xB704, 0xA94, 0x3608}},
    {21128, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21129, 2, {0x802, 0x77C2, 0x5242, 0x4A42, 0x5552, 0x6492, 0xA12, 0x1112, 0x3F92, 0xC452, 0x3F12, 0x492, 0x2502, 0x17C2, 0x7C0A, 0x2004}},
    {21130, 2, {0x402, 0xB02, 0x30C2, 0xDF22, 0x12, 0x7FD2, 0x5552, 0x4E52, 0x7FD2, 0x12, 0x3F92, 0x2092, 0x3F82, 0x2082, 0x3F8A, 0x2084}},
    {21131, 2, {0x4, 0x7F04, 0x4104, 0x7F24, 0x24, 0xF7A4, 0x94A4, 0xF7A4, 0x824, 0x824, 0xFFA4, 0x1C24, 0x2A04, 0x4904, 0x8894, 0x808}},
    {21132, 2, {0x202, 0x13C2, 0x1202, 0x7FEA, 0x10A, 0x14A, 0x3FEA, 0x210A, 0x3F0A, 0x254A, 0x254A, 0x2D4A, 0x3682, 0x44A2, 0x896A, 0x3224}},
    {21133, 2, {0x804, 0x804, 0x1404, 0x2224, 0x4124, 0xBEA4, 0x24, 0x24, 0x7724, 0x5524, 0x5524, 0x7724, 0x2204, 0x2204, 0x5514, 0x8888}},
    {21134, 2, {0x800, 0x800, 0x14FC, 0x2224, 0x4124, 0xBE24, 0x24, 0x24, 0x7724, 0x5524, 0x5524, 0x7724, 0x2244, 0x2244, 0x5594, 0x8908}},
    {21135, 2, {0x804, 0x4904, 0x2A04, 0xFFA4, 0x80A4, 0x3E24, 0x2224, 0x2224, 0x3E24, 0x24, 0x7F24, 0x4924, 0x7F04, 0x4904, 0x7F14, 0x4108}},
    {21136, 2, {0x1102, 0x7FC2, 0x1502, 0x1202, 0x3FD2, 0x6412, 0xBF92, 0x2412, 0x3F92, 0x2412, 0x3FD2, 0x2012, 0x7F82, 0x1102, 0xE0A, 0x7184}},
    {21137, 2, {0x802, 0x402, 0x7FC2, 0x1102, 0xA52, 0xF592, 0x5552, 0x5532, 0xB592, 0x12, 0x2092, 0x3F92, 0x2082, 0x3F82, 0x208A, 0x4084}},
    {21138, 2, {0x800, 0x800, 0x147E, 0x2212, 0x4112, 0xBE52, 0x52, 0x52, 0x7752, 0x5592, 0x5512, 0x7712, 0x2222, 0x2222, 0x554A, 0x8884}},
    {21139, 2, {0x802, 0x3F82, 0x2082, 0x3F82, 0x2092, 0x3F92, 0x12, 0x7FD2, 0x4452, 0x7FD2, 0x4452, 0x7FD2, 0x2, 0xFFE2, 0x110A, 0x2104}},
    {21140, 2, {0x800, 0x800, 0x147C, 0x2224, 0x4124, 0xBE24, 0xA4, 0x64, 0x7724, 0x5534, 0x552C, 0x7726, 0x2224, 0x2224, 0x5544, 0x888C}},
    {21141, 2, {0x844, 0x7384, 0x4204, 0x7BE4, 0x52A4, 0x52A4, 0x8424, 0x7FA4, 0x40A4, 0x7FA4, 0x40A4, 0x7FA4, 0x4084, 0x7F84, 0x2114, 0x4088}},
    {21142, 2, {0x3E02, 0x4202, 0xFF82, 0x4082, 0x7F92, 0x4492, 0x7712, 0x4452, 0x3FD2, 0x4212, 0xFF92, 0x4892, 0x7F82, 0x1A02, 0x294A, 0xC7C4}},
    {21143, 2, {0x5282, 0x7BE2, 0x9482, 0x7BEA, 0x318A, 0x5AAA, 0x946A, 0x3FCA, 0x204A, 0x3FCA, 0x204A, 0x3FCA, 0x2042, 0x3FC2, 0x108A, 0x2044}},
    {21144, 2, {0x202, 0x3FF2, 0x2482, 0x3FEA, 0x248A, 0x2ECA, 0x35AA, 0x248A, 0x250A, 0x3DEA, 0x250A, 0x3DCA, 0x2502, 0x5DE2, 0x450A, 0x8504}},
    {21145, 2, {0x1F02, 0x2102, 0x3E02, 0x202, 0x7FD2, 0x1892, 0x6512, 0x1A92, 0x6652, 0x1A12, 0x2492, 0xFBF2, 0xAAA2, 0xFBE2, 0x28AA, 0xFBE4}},
    {21146, 2, {0x7FC2, 0x4042, 0x7FC2, 0x5242, 0x4A92, 0x5252, 0x5FD2, 0x5552, 0x5FD2, 0x6452, 0x5F52, 0x5552, 0x5F42, 0x4442, 0xBF4A, 0x84}},
    {21147, 2, {0x200, 0x200, 0x200, 0x200, 0x7FF8, 0x208, 0x208, 0x208, 0x208, 0x408, 0x408, 0x808, 0x808, 0x1088, 0x2050, 0x4020}},
    {21148, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFF20, 0x1120, 0x1120, 0x1120, 0x1120, 0x1120, 0x2120, 0x2122, 0x2122, 0x4122, 0x4A1E, 0x8400}},
    {21149, 2, {0x40, 0x40, 0x40, 0xFC40, 0x5FC, 0x444, 0x4844, 0x2844, 0x1044, 0x1084, 0x2884, 0x2484, 0x4504, 0x8104, 0x228, 0x410}},
    {21150, 2, {0x200, 0x200, 0x200, 0x200, 0x7FF0, 0x210, 0x210, 0x1218, 0x1214, 0x2412, 0x4412, 0x810, 0x810, 0x1010, 0x20A0, 0x4040}},
    {21151, 2, {0x40, 0x40, 0x40, 0xFE40, 0x11FC, 0x1044, 0x1044, 0x1044, 0x1044, 0x1084, 0x1084, 0x1E84, 0xF104, 0x4104, 0x228, 0x410}},
    {21152, 2, {0x1000, 0x1000, 0x1000, 0x107C, 0xFE44, 0x1244, 0x1244, 0x1244, 0x1244, 0x1244, 0x1244, 0x1244, 0x2244, 0x227C, 0x4A44, 0x8400}},
    {21153, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x24C0, 0x300, 0xCC0, 0x3230, 0xC20E, 0x1FF0, 0x210, 0x410, 0x410, 0x810, 0x10A0, 0x2040}},
    {21154, 2, {0x20, 0x20, 0xFF20, 0x2020, 0x20FC, 0x2024, 0x3E24, 0x2224, 0x2224, 0x2224, 0x2224, 0x2224, 0x4244, 0x4244, 0x94A8, 0x910}},
    {21155, 2, {0x100, 0x1110, 0x1108, 0x2124, 0x4144, 0x180, 0xE00, 0xF100, 0x100, 0x3FF8, 0x208, 0x208, 0x408, 0x808, 0x1050, 0x6020}},
    {21156, 2, {0x220, 0x720, 0x3820, 0x2020, 0x20FC, 0x2024, 0x3F24, 0x2424, 0x2424, 0x2424, 0x2424, 0x2444, 0x2444, 0x4484, 0x4528, 0x8210}},
    {21157, 2, {0x1020, 0x820, 0x820, 0xFF20, 0xFC, 0x24, 0x3C24, 0x2424, 0x2424, 0x2424, 0x2424, 0x2544, 0x4644, 0x4484, 0x8128, 0x210}},
    {21158, 2, {0x200, 0x200, 0x3FF0, 0x410, 0x810, 0x10A0, 0x6040, 0x1020, 0x1020, 0x7EFC, 0x1224, 0x1224, 0x2244, 0x2244, 0x4A94, 0x8508}},
    {21159, 2, {0x420, 0xE20, 0x7820, 0x820, 0x8FC, 0x7E24, 0x824, 0x824, 0x824, 0xFF24, 0x824, 0x844, 0x844, 0x884, 0x2928, 0x1210}},
    {21160, 2, {0x40, 0x40, 0x7C40, 0x40, 0x1FC, 0x44, 0xFE44, 0x2044, 0x2044, 0x2084, 0x4884, 0x4484, 0xFD04, 0x4504, 0x228, 0x410}},
    {21161, 2, {0x20, 0x7C20, 0x4420, 0x4420, 0x44FC, 0x7C24, 0x4424, 0x4424, 0x7C24, 0x4424, 0x4424, 0x4444, 0x4E44, 0xF084, 0x128, 0x210}},
    {21162, 2, {0x1000, 0x1000, 0xFEFC, 0x2284, 0x4488, 0x2850, 0x1020, 0x2858, 0xC586, 0x200, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21163, 2, {0x820, 0x820, 0x820, 0x7E20, 0x8FC, 0x824, 0x824, 0xFF24, 0x1024, 0x1024, 0x2024, 0x4444, 0xFE44, 0x4284, 0x128, 0x210}},
    {21164, 2, {0x2020, 0x2020, 0x2020, 0x7E20, 0x42FC, 0x8224, 0x7A24, 0x4A24, 0x4A24, 0x4A24, 0x7A24, 0x4A44, 0x244, 0x284, 0x1528, 0xA10}},
    {21165, 2, {0x20, 0x7E20, 0x1220, 0x1220, 0x22FC, 0x2A24, 0x4424, 0x8024, 0x3E24, 0x2224, 0x2224, 0x2224, 0x2244, 0x3E44, 0x22A8, 0x110}},
    {21166, 2, {0x820, 0x820, 0x4820, 0x4820, 0x7EFC, 0x4824, 0x8824, 0x824, 0xFF24, 0x824, 0x824, 0x1444, 0x2244, 0x4284, 0x8128, 0x210}},
    {21167, 2, {0x20, 0x20, 0xFF20, 0x1020, 0x10FC, 0x1024, 0x3E24, 0x2224, 0x6224, 0xA224, 0x2224, 0x2244, 0x3E44, 0x2284, 0x128, 0x210}},
    {21168, 2, {0x1020, 0x1020, 0x2020, 0x7E20, 0x42FC, 0x4224, 0x4224, 0x4224, 0x7E24, 0x4224, 0x4224, 0x4224, 0x4244, 0x7E44, 0x42A8, 0x110}},
    {21169, 2, {0x10, 0x7FD0, 0x4010, 0x4010, 0x5F7E, 0x4812, 0x4812, 0x4F12, 0x4912, 0x4912, 0x4912, 0x4912, 0x4912, 0x5122, 0x952A, 0x2244}},
    {21170, 2, {0x20, 0x7E20, 0x420, 0x820, 0x18FC, 0x2424, 0x4224, 0x8124, 0x24, 0x7E24, 0x824, 0x844, 0x844, 0xE84, 0xF128, 0x4210}},
    {21171, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4202, 0x8204, 0x3FF0, 0x210, 0x210, 0x410, 0x410, 0x810, 0x10A0, 0x2040}},
    {21172, 2, {0x2208, 0x1108, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8204, 0x200, 0x7FF8, 0x208, 0x208, 0x408, 0x408, 0x808, 0x1050, 0x6020}},
    {21173, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x820, 0x1210, 0x2208, 0xDFF6, 0x410, 0x410, 0x810, 0x10A0, 0x2040}},
    {21174, 2, {0x420, 0xE20, 0x7020, 0x4020, 0x40FC, 0x7F24, 0x4024, 0x4024, 0x5E24, 0x5224, 0x5224, 0x5224, 0x5244, 0x9E44, 0x92A8, 0x110}},
    {21175, 2, {0x2220, 0x1220, 0x1420, 0x7F20, 0x8FC, 0x824, 0x3E24, 0x824, 0x824, 0x7F24, 0x824, 0x844, 0x1044, 0x1084, 0x2128, 0x4210}},
    {21176, 2, {0x820, 0x820, 0x7F20, 0x820, 0x8FC, 0x824, 0xFFA4, 0x24, 0x824, 0x824, 0x7F24, 0x824, 0x844, 0xF44, 0xF0A8, 0x4110}},
    {21177, 2, {0x1020, 0x820, 0x20, 0xFF20, 0xFC, 0x2424, 0x4224, 0x8124, 0x2424, 0x1424, 0x824, 0x1444, 0x2244, 0x4284, 0x8128, 0x210}},
    {21178, 2, {0x1010, 0x1010, 0x2410, 0x427E, 0xFF12, 0x912, 0x4812, 0x4812, 0x7F12, 0x8822, 0x822, 0xFFA2, 0x842, 0x842, 0x894, 0x908}},
    {21179, 2, {0x10, 0x7F90, 0x4010, 0x407E, 0x7F92, 0x4412, 0x4412, 0x5F12, 0x4412, 0x4422, 0x7FA2, 0x4022, 0x4042, 0x7FC2, 0x94, 0x108}},
    {21180, 2, {0x820, 0x820, 0x820, 0xFF20, 0x8FC, 0x824, 0x7E24, 0x24, 0x24, 0x7E24, 0x4224, 0x4224, 0x4244, 0x7E44, 0x42A8, 0x110}},
    {21181, 2, {0x7F84, 0x824, 0x1024, 0x3F24, 0x5124, 0xA24, 0x404, 0x804, 0x3014, 0xC208, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21182, 2, {0x1020, 0x820, 0x820, 0xFF20, 0x8FC, 0x1024, 0x2224, 0x7C24, 0x924, 0x1224, 0x2424, 0xC824, 0x1444, 0x2244, 0xC0A8, 0x110}},
    {21183, 2, {0x840, 0x840, 0x7DF8, 0x848, 0xC48, 0x19C8, 0x684A, 0x8AA, 0x2886, 0x1302, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21184, 2, {0x820, 0x820, 0x820, 0x7F20, 0x8FC, 0x824, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x1424, 0x1424, 0x2544, 0x2644, 0x44A8, 0x8110}},
    {21185, 2, {0x20, 0x7F20, 0x20, 0x2520, 0x4AFC, 0x9224, 0x4924, 0x24A4, 0x24, 0x7F24, 0x824, 0x824, 0x844, 0xF44, 0xF8A8, 0x110}},
    {21186, 2, {0x820, 0x2820, 0x2820, 0x3E20, 0x48FC, 0x824, 0xFF24, 0x24, 0x24, 0x3E24, 0x2224, 0x2224, 0x2244, 0x3E44, 0x22A8, 0x110}},
    {21187, 2, {0x820, 0x820, 0x7F20, 0x820, 0x8FC, 0x7F24, 0x4124, 0x8224, 0x3C24, 0x424, 0x824, 0xF44, 0x7844, 0x884, 0x2928, 0x1210}},
    {21188, 2, {0x2010, 0x2010, 0x3F90, 0x4010, 0xBF7E, 0x2112, 0x2912, 0x2512, 0xFFD2, 0x2112, 0x4912, 0x4512, 0x7F92, 0x122, 0xA2A, 0x444}},
    {21189, 2, {0x820, 0x820, 0xFFA0, 0x820, 0x8FC, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x824, 0x1C24, 0x2A44, 0x4944, 0x8884, 0x928, 0xA10}},
    {21190, 2, {0x1020, 0x820, 0x7E20, 0x4220, 0x42FC, 0x7E24, 0x4224, 0x4224, 0x7E24, 0x4024, 0x4824, 0x4444, 0x4A44, 0x5284, 0x6128, 0x4210}},
    {21191, 2, {0x3FF0, 0x660, 0x180, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x2208, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21192, 2, {0x10, 0x7E10, 0x210, 0x3410, 0x87E, 0x7F12, 0x4912, 0x4912, 0x7F12, 0x4912, 0x4912, 0x7F12, 0x4912, 0x4922, 0x492A, 0x4344}},
    {21193, 2, {0x2020, 0x2020, 0x3E20, 0x4420, 0x88FC, 0x7F24, 0x4924, 0x4924, 0x4924, 0x7F44, 0x1444, 0x1494, 0x250A, 0x2402, 0x43FE, 0x8000}},
    {21194, 2, {0x820, 0x820, 0xFF20, 0x87C, 0x824, 0x7F24, 0x4124, 0x4124, 0x7F24, 0x1444, 0x1454, 0x1488, 0x2502, 0x2402, 0x43FE, 0x8000}},
    {21195, 2, {0x20, 0x3E20, 0x2220, 0x2220, 0x3EFC, 0x24, 0x7F24, 0x4124, 0x4924, 0x4924, 0x4924, 0x4944, 0x1444, 0x1284, 0x2128, 0xC210}},
    {21196, 2, {0x820, 0x4920, 0x2A20, 0x7F20, 0x8FC, 0x1024, 0xFFA4, 0x2224, 0x4124, 0xBEA4, 0x2224, 0x2A24, 0x2444, 0x2044, 0x2CA8, 0x3110}},
    {21197, 2, {0x1010, 0x810, 0xFF90, 0x10, 0x7E, 0x7F12, 0x4112, 0x4112, 0x4112, 0x7F12, 0x812, 0x2A12, 0x4912, 0x88A2, 0x282A, 0x1044}},
    {21198, 2, {0x820, 0x820, 0x7E20, 0x820, 0x8FC, 0xFF24, 0x2224, 0x4124, 0x88A4, 0x824, 0x7E24, 0x824, 0x844, 0xF44, 0xF8A8, 0x4110}},
    {21199, 2, {0x1020, 0x820, 0x7F20, 0x20, 0x22FC, 0x1424, 0xFFA4, 0x24, 0x24, 0x3E24, 0x2224, 0x2224, 0x2244, 0x3E44, 0x22A8, 0x110}},
    {21200, 2, {0x20, 0x3E20, 0x420, 0x820, 0x7EFC, 0x824, 0x824, 0x2824, 0x1024, 0x7F24, 0x5524, 0x5544, 0x5544, 0x5784, 0xFD28, 0x210}},
    {21201, 2, {0x820, 0x820, 0x820, 0xFF20, 0x8FC, 0x2A24, 0x2A24, 0x2A24, 0x5D24, 0x8824, 0x1C24, 0x2A44, 0x4944, 0x8884, 0x928, 0xA10}},
    {21202, 2, {0x2220, 0x2220, 0xFFA0, 0x2220, 0x3EFC, 0x824, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x824, 0xFFC4, 0x844, 0x884, 0x928, 0xA10}},
    {21203, 2, {0x4410, 0x4490, 0x7510, 0x4610, 0x54FE, 0x6492, 0x4B92, 0x1012, 0x7F12, 0x4112, 0x4112, 0x7F12, 0x4112, 0x4122, 0x7F2A, 0x4144}},
    {21204, 2, {0x20, 0xFFA0, 0x820, 0x1020, 0x7F7C, 0x5524, 0x5524, 0x5D24, 0x5524, 0x5524, 0x5D24, 0x5524, 0x5544, 0x5544, 0x7F94, 0x4108}},
    {21205, 2, {0xF20, 0x7820, 0x820, 0xFFA0, 0x87C, 0x7F24, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x824, 0x7F24, 0x844, 0xFC4, 0xF894, 0x4108}},
    {21206, 2, {0x20, 0xFF20, 0x8120, 0xBD20, 0x81FC, 0xBD24, 0x8124, 0x24, 0x7E24, 0x4224, 0x7E24, 0x4224, 0x7E44, 0x4244, 0x7EA8, 0x4310}},
    {21207, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3E20, 0x2220, 0x3EFC, 0x2224, 0x3E24, 0x2224, 0x2F44, 0xF044, 0x4094, 0x108}},
    {21208, 2, {0x2220, 0x2220, 0xFF20, 0x2220, 0x3EFC, 0x2224, 0x3E24, 0x2224, 0x2224, 0xFF24, 0x4024, 0x5444, 0x6244, 0x4084, 0x7F28, 0x210}},
    {21209, 2, {0x40, 0xFC40, 0x4FC, 0x6888, 0x1150, 0x820, 0xFED8, 0x1306, 0x3440, 0x3040, 0x53FC, 0x5044, 0x9084, 0x1084, 0x5114, 0x2208}},
    {21210, 2, {0x5210, 0x5210, 0xFF90, 0x5210, 0x5E7E, 0x4012, 0x7F12, 0x12, 0x7F12, 0x4112, 0x4912, 0x4912, 0x4912, 0x1422, 0x222A, 0xC144}},
    {21211, 2, {0x20, 0x3E20, 0x2220, 0x2220, 0x3EFC, 0x24, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x7F44, 0x1444, 0x22A8, 0x4110}},
    {21212, 2, {0x2820, 0x4420, 0x9220, 0x2020, 0x48FC, 0xFC24, 0x424, 0x24, 0xEE24, 0x2224, 0xAA24, 0x6624, 0xAA44, 0x2244, 0xAAA8, 0x4510}},
    {21213, 2, {0x20, 0x7924, 0x48A8, 0x4820, 0x49FC, 0x7840, 0x4BFE, 0x4888, 0x4944, 0x7A42, 0x4DF8, 0x4848, 0x4888, 0x4888, 0x4928, 0x9A10}},
    {21214, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4102, 0x8104, 0x3FF8, 0x208, 0x208, 0x408, 0x808, 0x1050, 0x6020}},
    {21215, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1010, 0x2208, 0xDFF6, 0x410, 0x850, 0x3020}},
    {21216, 2, {0x20, 0x7720, 0x1120, 0x5520, 0x337C, 0x5524, 0xC24, 0x3324, 0xC4A4, 0x1824, 0x6224, 0xCA4, 0x7144, 0x644, 0x1894, 0xE108}},
    {21217, 2, {0x20, 0xFFA0, 0x1420, 0x7F20, 0x55FC, 0x5524, 0x7F24, 0x24, 0x7F24, 0x24, 0xFFA4, 0x824, 0x2A44, 0x4944, 0xA8A8, 0x1110}},
    {21218, 2, {0x1040, 0x7C40, 0x11F8, 0xFE48, 0x44C8, 0x9248, 0x7CAA, 0x10AA, 0x1D06, 0xE202, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21219, 2, {0x810, 0x7F10, 0x810, 0x7F10, 0x87E, 0xFF92, 0x12, 0x7F12, 0x4112, 0x7F12, 0x4112, 0x7F12, 0x4112, 0x7F22, 0x222A, 0x4144}},
    {21220, 2, {0x2220, 0x2220, 0xFFA0, 0x2220, 0x3EFC, 0x824, 0x7F24, 0x4924, 0x7F24, 0x824, 0x7F24, 0x844, 0x7F44, 0x884, 0xF28, 0xF210}},
    {21221, 2, {0x7CF8, 0x488, 0x4F8, 0x7C20, 0x41FC, 0x7D24, 0x5FC, 0x428, 0x2BFC, 0x1204, 0x200, 0x7FF8, 0x408, 0x808, 0x1050, 0x6020}},
    {21222, 2, {0x4920, 0x9220, 0x4920, 0x20, 0x7F7C, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x824, 0xFFA4, 0x1C24, 0x2A44, 0x4944, 0x8894, 0x908}},
    {21223, 2, {0x2020, 0x2020, 0x3F20, 0x4820, 0x88FC, 0x7F24, 0x1424, 0x2224, 0x7F24, 0xA424, 0x3F24, 0x2424, 0x3F44, 0x2444, 0x3FA8, 0x2110}},
    {21224, 2, {0x2020, 0x3E20, 0x4220, 0xFF20, 0x49FC, 0x4924, 0x7F24, 0x1024, 0x2824, 0xD524, 0x2624, 0xCD24, 0x1544, 0x2444, 0xD4A8, 0x910}},
    {21225, 2, {0x5220, 0xFFA0, 0x5220, 0x5E20, 0x40FC, 0x7F24, 0x24, 0x7E24, 0x4224, 0x7E24, 0x4224, 0x7E44, 0x4244, 0x7E84, 0x2528, 0x4210}},
    {21226, 2, {0x610, 0x7810, 0x1010, 0xFF90, 0x227E, 0x5D12, 0x9492, 0x1C12, 0x12, 0x7F12, 0x4112, 0x5D12, 0x5512, 0x5D22, 0x412A, 0x4344}},
    {21227, 2, {0xF10, 0xF810, 0x4910, 0x2A7E, 0xFF92, 0x2A12, 0x4912, 0x8092, 0x7F12, 0x4922, 0x4922, 0x7F22, 0x4942, 0x4942, 0x7F94, 0x4108}},
    {21228, 2, {0x810, 0x4910, 0x2A10, 0x7F10, 0x107E, 0xFF92, 0x2212, 0x5112, 0xA292, 0x3C12, 0x912, 0x7F12, 0x812, 0x4922, 0x88AA, 0x1844}},
    {21229, 2, {0x1010, 0x810, 0x7F10, 0x2210, 0xFF7E, 0x12, 0x7F12, 0x4912, 0x7F12, 0x4912, 0x7F12, 0x812, 0x7F12, 0x822, 0xF2A, 0xF044}},
    {21230, 2, {0x810, 0xF10, 0x810, 0x7F90, 0x48FE, 0x7E12, 0x4912, 0x4712, 0x4012, 0x7F92, 0x4812, 0x5512, 0x4A12, 0xB722, 0xAAA, 0x3644}},
    {21231, 2, {0x820, 0xFF20, 0x20, 0xFF20, 0x817C, 0xBD24, 0xA524, 0xFF24, 0x24, 0x7E24, 0x4224, 0x7E24, 0x4244, 0x7E44, 0x94, 0xFF08}},
    {21232, 2, {0x1000, 0x10FE, 0x7E92, 0x1292, 0x22FE, 0x2A92, 0x4492, 0x80FE, 0x4410, 0xFF08, 0x5528, 0x5520, 0x5564, 0x55AA, 0x5928, 0xB318}},
    {21233, 2, {0x2210, 0x2210, 0xFF90, 0x2210, 0x7F7E, 0x4912, 0x7F12, 0x4912, 0x7F12, 0x812, 0xFF92, 0x8892, 0x8A92, 0xBEA2, 0x82AA, 0x81C4}},
    {21234, 2, {0x7F20, 0x820, 0xFF20, 0x8FC, 0x7F24, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x844, 0x7F44, 0x894, 0xFF08, 0x0, 0x2448, 0x4224}},
    {21235, 2, {0x710, 0x7810, 0x810, 0xFF90, 0x87E, 0x7F12, 0x6B12, 0x5D12, 0x7F12, 0x812, 0x7F12, 0x812, 0xFF92, 0x22, 0x552A, 0x8AC4}},
    {21236, 2, {0x810, 0xF90, 0x810, 0x7F90, 0x48FE, 0x7E12, 0x4892, 0x5F92, 0x5512, 0x5F12, 0x5512, 0x5F12, 0x4812, 0x5522, 0xB2AA, 0x5E44}},
    {21237, 2, {0x10, 0x7F90, 0x4A10, 0x7F90, 0x4A7E, 0x5F12, 0x5512, 0x5F12, 0x5512, 0x5F12, 0x4412, 0x7F92, 0x6492, 0xAEA2, 0xA2AA, 0x2144}},
    {21238, 2, {0x2210, 0x4F90, 0x821E, 0x14A4, 0x2FD4, 0x6014, 0xAF94, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x200, 0x7FFC, 0x404, 0x1828, 0x6010}},
    {21239, 2, {0x810, 0xFF90, 0x10, 0x7710, 0x557E, 0x7712, 0x2212, 0x7F12, 0x2212, 0x7F12, 0x2212, 0xFF92, 0x2892, 0x6522, 0xB22A, 0x2144}},
    {21240, 2, {0x2210, 0xFF90, 0x2210, 0xF790, 0x94FE, 0xF792, 0x2812, 0x2412, 0x7F92, 0x4812, 0xFF12, 0x4812, 0x7F12, 0x4822, 0x7FAA, 0x4044}},
    {21241, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4004, 0x8004, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x28, 0x10}},
    {21242, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x8008, 0x408, 0x208, 0x108, 0x108, 0x8, 0x8, 0x8, 0x50, 0x20}},
    {21243, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x8008, 0x1F88, 0x8, 0x8, 0x8, 0x3FC8, 0x8, 0x8, 0x50, 0x20}},
    {21244, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4404, 0x8404, 0x3F84, 0x484, 0x484, 0x884, 0x894, 0x1094, 0x2074, 0x4004, 0x28, 0x10}},
    {21245, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4208, 0x8208, 0x208, 0x508, 0x488, 0x488, 0x848, 0x1048, 0x2008, 0x50, 0x20}},
    {21246, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4208, 0x8208, 0x408, 0x408, 0x888, 0x1048, 0x3FE8, 0x1028, 0x8, 0x50, 0x20}},
    {21247, 2, {0x800, 0x800, 0x800, 0x1FFC, 0x1244, 0x2244, 0x4244, 0x8444, 0x444, 0x884, 0x1084, 0x2104, 0x4204, 0x404, 0x828, 0x1010}},
    {21248, 2, {0x800, 0x800, 0xFFC, 0x1004, 0x2004, 0x4804, 0x8404, 0x204, 0x24, 0xC4, 0x304, 0x1C04, 0x804, 0x4, 0x28, 0x10}},
    {21249, 2, {0x800, 0x800, 0x800, 0x1FF8, 0x1108, 0x2908, 0x4508, 0x8208, 0x308, 0x488, 0x448, 0x838, 0x100E, 0x2008, 0x4050, 0x20}},
    {21250, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4804, 0x8844, 0x884, 0x904, 0xE04, 0x814, 0x814, 0x814, 0x7F4, 0x4, 0x28, 0x10}},
    {21251, 2, {0x800, 0x800, 0xFFC, 0x1004, 0x2104, 0x4104, 0x9104, 0x1284, 0x1244, 0x1424, 0x1824, 0x1004, 0x1FE4, 0x4, 0x28, 0x10}},
    {21252, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4204, 0x8104, 0x4, 0x3FF4, 0x804, 0x804, 0x804, 0x804, 0xFE4, 0x4, 0x28, 0x10}},
    {21253, 2, {0x800, 0x800, 0x1FF0, 0x1010, 0x2010, 0x5F90, 0x9090, 0x1090, 0x1090, 0x1F90, 0x1050, 0x1020, 0x1004, 0x1004, 0xFFC, 0x0}},
    {21254, 2, {0x800, 0x800, 0x800, 0x1FFC, 0x1244, 0x2244, 0x4244, 0x9C44, 0x644, 0x984, 0x10C4, 0x2124, 0x4214, 0x404, 0x828, 0x1010}},
    {21255, 2, {0x1000, 0x1000, 0x3FFC, 0x2204, 0x4204, 0x87E4, 0x824, 0x1024, 0x2444, 0x244, 0x284, 0x104, 0x604, 0x1804, 0x6028, 0x10}},
    {21256, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4088, 0x8888, 0x2528, 0x2228, 0x2528, 0x28A8, 0x2028, 0x3FE8, 0x8, 0x50, 0x20}},
    {21257, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4004, 0xBFE4, 0x204, 0x1244, 0xA44, 0xA84, 0x204, 0x7FF4, 0x204, 0x204, 0x228, 0x210}},
    {21258, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4204, 0x9244, 0xA84, 0x204, 0x7FF4, 0x704, 0xA84, 0x1244, 0x6224, 0x204, 0x228, 0x10}},
    {21259, 2, {0x1000, 0x1000, 0x3FFC, 0x2004, 0x4804, 0x8FC4, 0x1204, 0x204, 0x7FF4, 0x204, 0x1244, 0x1244, 0x1FC4, 0x44, 0x28, 0x10}},
    {21260, 2, {0x1000, 0x1000, 0x3FFC, 0x2204, 0x4504, 0x8884, 0x1044, 0x6FB4, 0x4, 0x1FC4, 0x1044, 0x1044, 0x1FC4, 0x4, 0x28, 0x10}},
    {21261, 2, {0x1000, 0x1000, 0x1FFC, 0x2244, 0x4224, 0xBFF4, 0x204, 0x3FE4, 0x2224, 0x3FE4, 0x2224, 0x3FE4, 0x2224, 0x2224, 0x22B4, 0x2048}},
    {21262, 2, {0x1000, 0x1000, 0x1FFC, 0x2404, 0x7FE4, 0x8884, 0x1244, 0x6234, 0x1FC4, 0x1244, 0x1FC4, 0x1244, 0x1FC4, 0x224, 0x1F4, 0x8}},
    {21263, 2, {0x1040, 0x1040, 0xFEFC, 0x2884, 0x4504, 0x82F4, 0x7C94, 0x94, 0xFEF4, 0x2084, 0x4094, 0x7C88, 0x482, 0x482, 0x287E, 0x1000}},
    {21264, 2, {0x1000, 0x1000, 0x1FFC, 0x2004, 0x7FE4, 0x8004, 0x1FC4, 0x1044, 0x1FC4, 0x4, 0x3FE4, 0x2224, 0x3FE4, 0x2224, 0x3FF4, 0x2028}},
    {21265, 2, {0x1000, 0x1000, 0x3FFC, 0x2804, 0x5EF4, 0x9214, 0x1E14, 0x12F4, 0x1E84, 0x1284, 0x7EF4, 0xA14, 0x1214, 0x2A54, 0x4428, 0x10}},
    {21266, 2, {0x1000, 0x1000, 0x1FFC, 0x2884, 0x4884, 0xBFE4, 0xA84, 0x504, 0x884, 0x1744, 0x6034, 0x1FC4, 0x1044, 0x1044, 0x1FD4, 0x1048}},
    {21267, 2, {0x1000, 0x1000, 0x3FFC, 0x4004, 0xBDE4, 0x2524, 0x2524, 0x3D34, 0x2604, 0x25E4, 0x3C24, 0x20A4, 0x2844, 0x2CA4, 0x3328, 0x2010}},
    {21268, 2, {0x1000, 0x1000, 0x3FFC, 0x2804, 0x5EF4, 0x9294, 0x1E94, 0x12F4, 0x1E04, 0x12F4, 0x7E94, 0xA94, 0x12F4, 0x2A04, 0x4428, 0x10}},
    {21269, 2, {0x1000, 0x1000, 0x1000, 0x1008, 0x1030, 0x10C0, 0x1300, 0x1C00, 0x1000, 0x1000, 0x1000, 0x1004, 0x1004, 0x1004, 0xFFC, 0x0}},
    {21270, 2, {0x880, 0x880, 0x884, 0x1088, 0x1090, 0x30A0, 0x30C0, 0x5080, 0x9180, 0x1280, 0x1480, 0x1082, 0x1082, 0x1082, 0x107E, 0x1000}},
    {21271, 2, {0x440, 0x440, 0x440, 0x444, 0x448, 0x7C50, 0x460, 0x440, 0x440, 0x440, 0x440, 0x442, 0x1C42, 0xE442, 0x443E, 0x400}},
    {21272, 2, {0x4092, 0x4124, 0x4248, 0x4924, 0x4892, 0x5040, 0x6080, 0x43FE, 0x4202, 0x428A, 0x4252, 0x4A22, 0x4A52, 0x4A8A, 0x3BFE, 0x202}},
    {21273, 2, {0x0, 0x3E40, 0x2240, 0x3E44, 0x2248, 0x3E50, 0x60, 0xFF40, 0x844, 0x844, 0x2F44, 0x283C, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {21274, 2, {0x0, 0x7FFC, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x7FFE, 0x0}},
    {21275, 2, {0x0, 0x7FFC, 0x4200, 0x4200, 0x47F0, 0x4410, 0x4820, 0x5020, 0x4040, 0x40A0, 0x4110, 0x4608, 0x5804, 0x4000, 0x7FFE, 0x0}},
    {21276, 2, {0x0, 0x7FFC, 0x4080, 0x4080, 0x4480, 0x44F0, 0x4790, 0x5C90, 0x44D0, 0x44A4, 0x4484, 0x4404, 0x43FC, 0x4000, 0x7FFE, 0x0}},
    {21277, 2, {0x0, 0x7FFC, 0x4080, 0x4080, 0x4080, 0x4FF8, 0x4888, 0x4888, 0x4888, 0x4888, 0x48A8, 0x4890, 0x4080, 0x4080, 0x7FFE, 0x0}},
    {21278, 2, {0x0, 0x7FFC, 0x4000, 0x4000, 0x4FF8, 0x4080, 0x4080, 0x4080, 0x4080, 0x4080, 0x4080, 0x5FFC, 0x4000, 0x4000, 0x7FFE, 0x0}},
    {21279, 2, {0x0, 0x7FFC, 0x4000, 0x4100, 0x4080, 0x5FFC, 0x4000, 0x43E0, 0x4220, 0x4220, 0x4224, 0x4424, 0x481C, 0x5000, 0x7FFE, 0x0}},
    {21280, 2, {0x0, 0x7FFC, 0x4000, 0x4038, 0x47C0, 0x4400, 0x4400, 0x47FC, 0x4420, 0x4420, 0x4420, 0x4820, 0x4820, 0x5020, 0x7FFE, 0x0}},
    {21281, 2, {0x0, 0x7FFC, 0x4000, 0x4000, 0x4FF8, 0x4080, 0x4080, 0x47F0, 0x4080, 0x4080, 0x4080, 0x4FF8, 0x4000, 0x4000, 0x7FFE, 0x0}},
    {21282, 2, {0x0, 0x7FFC, 0x4400, 0x4400, 0x47F8, 0x4928, 0x5128, 0x6248, 0x4448, 0x4888, 0x5108, 0x4250, 0x4420, 0x4000, 0x7FFE, 0x0}},
    {21283, 2, {0x0, 0x7FFC, 0x4000, 0x4000, 0x4FF8, 0x4888, 0x4888, 0x4FF8, 0x4888, 0x4888, 0x4FF8, 0x4888, 0x4080, 0x4080, 0x7FFE, 0x0}},
    {21284, 2, {0x0, 0x7FFC, 0x4000, 0x4000, 0x5FFC, 0x4080, 0x4080, 0x4FF8, 0x4080, 0x40A0, 0x4090, 0x5FFC, 0x4000, 0x4000, 0x7FFE, 0x0}},
    {21285, 2, {0x0, 0x7FFC, 0x4200, 0x4410, 0x4808, 0x5FFC, 0x4004, 0x4410, 0x4410, 0x7FFE, 0x4410, 0x4410, 0x4810, 0x5010, 0x7FFE, 0x0}},
    {21286, 2, {0x0, 0x7FFC, 0x4420, 0x4420, 0x5FA0, 0x4878, 0x5228, 0x5FA8, 0x4228, 0x43A8, 0x5E2A, 0x4A2A, 0x4246, 0x4280, 0x7FFE, 0x0}},
    {21287, 2, {0x0, 0x7FFC, 0x4080, 0x4080, 0x5FFC, 0x4080, 0x4490, 0x4490, 0x4AA8, 0x5144, 0x4120, 0x4220, 0x4410, 0x4810, 0x7FFE, 0x0}},
    {21288, 2, {0x0, 0x7FFC, 0x4220, 0x5220, 0x5220, 0x5E20, 0x43FC, 0x4220, 0x7E20, 0x5220, 0x5220, 0x52F8, 0x6200, 0x4200, 0x7FFE, 0x0}},
    {21289, 2, {0x0, 0x7FFC, 0x4080, 0x4888, 0x4888, 0x4FF8, 0x4080, 0x5FFC, 0x4000, 0x4080, 0x4FF8, 0x4080, 0x5FFC, 0x4000, 0x7FFE, 0x0}},
    {21290, 2, {0x0, 0x7FFC, 0x4240, 0x4240, 0x7E7C, 0x4240, 0x4240, 0x5E78, 0x4240, 0x4240, 0x7E7C, 0x4240, 0x4240, 0x4240, 0x7FFE, 0x0}},
    {21291, 2, {0x0, 0x7FFC, 0x4400, 0x4FFC, 0x5124, 0x6244, 0x44A8, 0x4910, 0x4FF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4000, 0x7FFE, 0x0}},
    {21292, 2, {0x7FFE, 0x4080, 0x4140, 0x4220, 0x4C18, 0x73F6, 0x4000, 0x4FA4, 0x48A4, 0x4FA4, 0x48A4, 0x4FA4, 0x4884, 0x4A94, 0x4908, 0x7FFE}},
    {21293, 2, {0x0, 0x7FFC, 0x4420, 0x5F20, 0x4420, 0x5F78, 0x5528, 0x5F28, 0x5528, 0x5F28, 0x442A, 0x5F2A, 0x4446, 0x4480, 0x7FFE, 0x0}},
    {21294, 2, {0x0, 0x7FFC, 0x4080, 0x4FF8, 0x4888, 0x4FF8, 0x4080, 0x7FFE, 0x4000, 0x4FF8, 0x4888, 0x4968, 0x4210, 0x4C08, 0x7FFE, 0x0}},
    {21295, 2, {0x0, 0x7FFC, 0x40A0, 0x4890, 0x45FE, 0x6120, 0x57FC, 0x4520, 0x49FC, 0x7920, 0x4920, 0x49FE, 0x4900, 0x4000, 0x7FFE, 0x0}},
    {21296, 2, {0x7FFE, 0x4000, 0x5F7C, 0x5144, 0x5F7C, 0x4000, 0x4FF8, 0x4888, 0x4FF8, 0x4888, 0x4FF8, 0x4080, 0x7FFE, 0x4080, 0x4080, 0x7FFE}},
    {21297, 2, {0x7FFE, 0x4080, 0x4FF8, 0x4888, 0x4FF8, 0x4080, 0x7FFE, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4410, 0x4808, 0x7FFE}},
    {21298, 2, {0x0, 0x7FFC, 0x4080, 0x5FFC, 0x4220, 0x4C18, 0x77F6, 0x4410, 0x47F0, 0x4000, 0x4F78, 0x4948, 0x4F78, 0x4000, 0x7FFE, 0x0}},
    {21299, 2, {0x7FFE, 0x4080, 0x4140, 0x4220, 0x4C18, 0x73E6, 0x4000, 0x5F7C, 0x5144, 0x5144, 0x5F7C, 0x4410, 0x4A28, 0x5144, 0x4000, 0x7FFE}},
    {21300, 2, {0x7FFE, 0x4840, 0x4F7C, 0x5490, 0x6FF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4220, 0x5FFC, 0x4420, 0x4820, 0x7FFE}},
    {21301, 2, {0x7FFE, 0x4080, 0x4FF8, 0x4080, 0x4FF8, 0x4948, 0x4FF8, 0x4410, 0x47F0, 0x4410, 0x47F0, 0x4410, 0x47F0, 0x4220, 0x4410, 0x7FFE}},
    {21302, 2, {0x7FFE, 0x4410, 0x5FFC, 0x4490, 0x4FFC, 0x5880, 0x6FF8, 0x4880, 0x4FF8, 0x4880, 0x4FFC, 0x4408, 0x4738, 0x4408, 0x7FFE, 0x0}},
    {21303, 2, {0x7FFE, 0x4948, 0x4F78, 0x4948, 0x4F78, 0x4948, 0x4F78, 0x4480, 0x47F8, 0x4C80, 0x57F0, 0x6480, 0x47FC, 0x4400, 0x7FFE, 0x0}},
    {21304, 2, {0x0, 0x0, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1FFE, 0x0, 0x0}},
    {21305, 2, {0x0, 0x7FFC, 0x4440, 0x4440, 0x4440, 0x4440, 0x4440, 0x4440, 0x4444, 0x4444, 0x4844, 0x483C, 0x5000, 0x4000, 0x7FFE, 0x0}},
    {21306, 2, {0x0, 0x7FFC, 0x4000, 0x4010, 0x4410, 0x4220, 0x4140, 0x4080, 0x4140, 0x4220, 0x4410, 0x4810, 0x4000, 0x4000, 0x7FFE, 0x0}},
    {21307, 2, {0x0, 0x7FFC, 0x4400, 0x4400, 0x4FF8, 0x5080, 0x4080, 0x5FFC, 0x4080, 0x4140, 0x4220, 0x4410, 0x5808, 0x4000, 0x7FFE, 0x0}},
    {21308, 2, {0x0, 0x7FFC, 0x4080, 0x4140, 0x4220, 0x4410, 0x5BEC, 0x4000, 0x47F0, 0x4410, 0x4410, 0x47F0, 0x4410, 0x4000, 0x7FFE, 0x0}},
    {21309, 2, {0x7FFE, 0x4000, 0x4FF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4100, 0x5FFE, 0x4210, 0x4720, 0x40C0, 0x4330, 0x4C08, 0x4000, 0x7FFE}},
    {21310, 2, {0x0, 0x7FFC, 0x4080, 0x4FF8, 0x4808, 0x4FF8, 0x4800, 0x4FFC, 0x5924, 0x6924, 0x4FFC, 0x4924, 0x492C, 0x4000, 0x7FFE, 0x0}},
    {21311, 2, {0x0, 0x7FFC, 0x4000, 0x4410, 0x7FFE, 0x4410, 0x4100, 0x7FFE, 0x4400, 0x4FF8, 0x5408, 0x6408, 0x47F8, 0x4000, 0x7FFE, 0x0}},
    {21312, 2, {0x0, 0x7FFC, 0x4000, 0x4FF0, 0x4810, 0x4810, 0x4FF0, 0x4000, 0x5E78, 0x5248, 0x5248, 0x5E78, 0x5248, 0x4000, 0x7FFE, 0x0}},
    {21313, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21314, 2, {0x0, 0x7FF0, 0x410, 0x410, 0x410, 0x410, 0x410, 0xFFD0, 0x410, 0x410, 0x412, 0x40A, 0x40A, 0x406, 0x402, 0x400}},
    {21315, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21316, 2, {0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0xFFFE, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840}},
    {21317, 2, {0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x2110, 0x2010, 0x4010}},
    {21318, 2, {0x200, 0x200, 0x3FE0, 0x420, 0x420, 0x822, 0x1022, 0x211E, 0x4100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21319, 2, {0x120, 0x7A0, 0x3C20, 0x420, 0x420, 0x420, 0x420, 0xFFFE, 0x420, 0x420, 0x420, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {21320, 2, {0x800, 0x800, 0x800, 0x1FF8, 0x1100, 0x2100, 0x4100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21321, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x920, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {21322, 2, {0x100, 0x2108, 0x1108, 0x910, 0x920, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21323, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x920, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x820, 0x820, 0xFE0, 0x820}},
    {21324, 2, {0x2010, 0x2490, 0x2490, 0x2490, 0x2490, 0x2490, 0xFFFE, 0x2490, 0x2490, 0x2490, 0x2490, 0x2490, 0x2490, 0x2490, 0x2490, 0x2010}},
    {21325, 2, {0x4, 0x3F84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x3FFC, 0x2084, 0x2080, 0x2080, 0x2080, 0x2080, 0x20FE, 0x2000, 0x2000}},
    {21326, 2, {0x880, 0x888, 0x1090, 0x30E0, 0x5180, 0x9684, 0x1084, 0x107C, 0x1100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21327, 2, {0x2080, 0x2080, 0x2080, 0x2080, 0xFBF0, 0x2090, 0x2090, 0x2298, 0x2294, 0x2492, 0x2892, 0x2090, 0x2110, 0x2110, 0x2250, 0x2420}},
    {21328, 2, {0x2000, 0x21FC, 0x2100, 0x2100, 0x2100, 0x2100, 0x2100, 0x3FFC, 0x104, 0x104, 0x104, 0x104, 0x104, 0xFF04, 0x4, 0x0}},
    {21329, 2, {0x200, 0x400, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2208, 0x2208, 0x3FF8, 0x880, 0x1080, 0xFFFE, 0x80, 0x80, 0x80, 0x80}},
    {21330, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x820, 0x820, 0x1450, 0x2288, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {21331, 2, {0x100, 0x1FC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1110, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {21332, 2, {0x2040, 0x2040, 0x23FC, 0x2084, 0xF884, 0x2114, 0x2208, 0x2010, 0x2410, 0x2FBE, 0x2492, 0x2492, 0x2492, 0x28A2, 0x2AAA, 0x3144}},
    {21333, 2, {0x1010, 0x820, 0x440, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {21334, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x404, 0x288, 0x1280, 0x880, 0x900, 0xFFFE, 0x240, 0x420, 0x1810, 0x6008}},
    {21335, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x4824, 0x4444, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4114, 0x4008}},
    {21336, 2, {0x2208, 0x1108, 0x1110, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {21337, 2, {0x2210, 0x2210, 0xFF90, 0x2210, 0x3E10, 0x2210, 0x3E7E, 0x2210, 0x2210, 0xFFD0, 0x4010, 0x5210, 0x6110, 0x4010, 0x7F90, 0x10}},
    {21338, 2, {0x2050, 0x2048, 0x2FFE, 0x2040, 0x27FC, 0xFC44, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x2008, 0x2FFE, 0x2208, 0x2108, 0x2128, 0x2010}},
    {21339, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {21340, 2, {0x200, 0x200, 0x200, 0x200, 0x200, 0x280, 0x240, 0x220, 0x210, 0x210, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200}},
    {21341, 2, {0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0xF87E, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840, 0x840}},
    {21342, 2, {0x200, 0x100, 0x100, 0x0, 0xFFFE, 0x200, 0x200, 0x240, 0x220, 0x210, 0x210, 0x200, 0x200, 0x200, 0x200, 0x200}},
    {21343, 2, {0x40, 0x40, 0x7C40, 0x4440, 0x4440, 0x4460, 0x4450, 0x4448, 0x4444, 0x4444, 0x7C40, 0x4440, 0x40, 0x40, 0x40, 0x40}},
    {21344, 2, {0x100, 0x100, 0x100, 0x1FE, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21345, 2, {0x200, 0x200, 0x3F8, 0x200, 0x200, 0x200, 0xFFFE, 0x200, 0x200, 0x240, 0x220, 0x210, 0x208, 0x200, 0x200, 0x200}},
    {21346, 2, {0x100, 0x100, 0x1FC, 0x100, 0x100, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {21347, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21348, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x2008}},
    {21349, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2928, 0x2528, 0x2248, 0x2548, 0x2888, 0x2148, 0x2228, 0x3FF8, 0x2008}},
    {21350, 2, {0x840, 0x840, 0x840, 0x7F40, 0x840, 0x860, 0xFFD0, 0x848, 0x844, 0x7F44, 0x840, 0x840, 0x840, 0xFC0, 0xF040, 0x40}},
    {21351, 2, {0x20, 0x7FA0, 0x4820, 0x4820, 0x4820, 0x7F30, 0x4128, 0x4124, 0x4122, 0x7F22, 0x4820, 0x4820, 0x4820, 0x7FA0, 0x20, 0x20}},
    {21352, 2, {0x100, 0x100, 0x1FC, 0x100, 0x1FF0, 0x1010, 0x1F90, 0x1090, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {21353, 2, {0x0, 0x3F8, 0x208, 0x208, 0x208, 0x208, 0x208, 0x208, 0x208, 0x208, 0x228, 0x210, 0x200, 0x200, 0x200, 0x200}},
    {21354, 2, {0x0, 0x3F8, 0x208, 0x208, 0x208, 0x288, 0x248, 0x248, 0x208, 0x208, 0x228, 0x210, 0x200, 0x200, 0x200, 0x200}},
    {21355, 2, {0x0, 0x7FF8, 0x208, 0x208, 0x208, 0x208, 0x208, 0x208, 0x228, 0x210, 0x200, 0x200, 0x200, 0x200, 0xFFFE, 0x0}},
    {21356, 2, {0x0, 0x600, 0x78FC, 0x4084, 0x4084, 0x4084, 0x4084, 0x4084, 0x4084, 0x4084, 0x4084, 0x4EA8, 0x7090, 0x80, 0x80, 0x80}},
    {21357, 2, {0x0, 0x0, 0xFEFC, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1E84, 0xF0A8, 0x4090, 0x80, 0x80, 0x80}},
    {21358, 2, {0x7C, 0x3F80, 0x2000, 0x2000, 0x3FFE, 0x2000, 0x2000, 0x27F0, 0x2410, 0x2410, 0x2450, 0x2424, 0x2404, 0x4404, 0x43FC, 0x8000}},
    {21359, 2, {0xC00, 0x70FC, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x4C84, 0x74A8, 0x490, 0x880, 0x880, 0x1080, 0x2080}},
    {21360, 2, {0x0, 0x600, 0x78FC, 0x4084, 0x4084, 0x4084, 0x7E84, 0x4084, 0x4084, 0x4084, 0x4084, 0x4EA8, 0x7090, 0x80, 0x80, 0x80}},
    {21361, 2, {0x400, 0x400, 0xFF0, 0x1010, 0x2020, 0x5FFC, 0x1000, 0x13F0, 0x1210, 0x1210, 0x1250, 0x1220, 0x2204, 0x2204, 0x41FC, 0x8000}},
    {21362, 2, {0x0, 0x0, 0x7F7C, 0x1144, 0x1144, 0x2144, 0x2544, 0x4244, 0x8044, 0x3F44, 0x2144, 0x2154, 0x2148, 0x2140, 0x3F40, 0x2140}},
    {21363, 2, {0x0, 0x7E7C, 0x4244, 0x4244, 0x7E44, 0x4244, 0x4244, 0x7E44, 0x4044, 0x4844, 0x4454, 0x4A48, 0x5240, 0x6040, 0x40, 0x40}},
    {21364, 2, {0x800, 0x800, 0x87C, 0x7F44, 0x844, 0x844, 0x844, 0xFFC4, 0x1044, 0x1044, 0x2444, 0x4254, 0xFF48, 0x4140, 0x40, 0x40}},
    {21365, 2, {0xC00, 0x70FC, 0x4484, 0x4484, 0x64A4, 0x5494, 0x5494, 0x4484, 0x4484, 0x4C84, 0x74A8, 0x490, 0x880, 0x880, 0x1080, 0x2080}},
    {21366, 2, {0x1000, 0x1000, 0x3E7C, 0x4244, 0xA444, 0x1844, 0x1044, 0x2444, 0xC844, 0x1F44, 0x2144, 0xD254, 0xC48, 0x840, 0x3040, 0xC040}},
    {21367, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC826, 0x820, 0x8A8, 0x848, 0x808, 0x7F8}},
    {21368, 2, {0x2000, 0x2000, 0x3F7C, 0x4844, 0x8844, 0x844, 0x7F44, 0x844, 0x844, 0x2E44, 0x2844, 0x2854, 0x2F48, 0xF040, 0x4040, 0x40}},
    {21369, 2, {0x800, 0x800, 0x107C, 0x7F44, 0x5544, 0x5544, 0x5544, 0x5544, 0x5544, 0x5544, 0x5544, 0x5554, 0x57C8, 0xFC40, 0x40, 0x40}},
    {21370, 2, {0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3118, 0xC506, 0x200, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1050, 0x1024, 0x1004, 0xFFC}},
    {21371, 2, {0x1400, 0x1200, 0x217C, 0x4944, 0x844, 0x1444, 0x2244, 0x4144, 0xBE44, 0x2244, 0x2244, 0x2254, 0x2248, 0x3E40, 0x2240, 0x40}},
    {21372, 2, {0x80, 0xFC80, 0x29F8, 0x2A08, 0x2DFE, 0x2900, 0x2978, 0x2948, 0x2968, 0x2950, 0x2944, 0x2A44, 0x2C3C, 0x4802, 0x47FE, 0x8000}},
    {21373, 2, {0x800, 0x1000, 0x7E7C, 0x4244, 0x7E44, 0x4244, 0x7E44, 0x44, 0x4044, 0x4044, 0x7E44, 0x4054, 0x4048, 0x4C40, 0x7040, 0x40}},
    {21374, 2, {0x0, 0x7700, 0x557C, 0x5544, 0x7744, 0x44, 0x7F44, 0x44, 0xFFC4, 0x2044, 0x3F44, 0x154, 0x148, 0x140, 0xA40, 0x440}},
    {21375, 2, {0x800, 0x73DE, 0x4252, 0x4A52, 0x4A52, 0x4BD2, 0x4A52, 0x4A52, 0x4A52, 0x5BD2, 0x6A12, 0x4A9A, 0x1254, 0x12B0, 0x2310, 0x4210}},
    {21376, 2, {0x800, 0x800, 0x7F7C, 0x1C44, 0x2A44, 0x4944, 0x1444, 0x2244, 0x4944, 0x8844, 0x2A44, 0x1C54, 0x2A48, 0xC940, 0x2840, 0x1040}},
    {21377, 2, {0x808, 0x1448, 0x2228, 0x5C88, 0x8848, 0x7F0E, 0x8F8, 0x2A08, 0x4908, 0x1808, 0x7FFC, 0x104, 0x114, 0x108, 0x100, 0x100}},
    {21378, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {21379, 2, {0x400, 0x400, 0xFF0, 0x1010, 0x2020, 0x5FFC, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {21380, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x27F0, 0x2410, 0x2410, 0x2410, 0x2410, 0x2450, 0x2420, 0x2404, 0x2404, 0x4404, 0x43FC, 0x8000}},
    {21381, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x4040, 0x4140, 0x8080}},
    {21382, 2, {0x0, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2084, 0x2084, 0x2104, 0x2104, 0x2204, 0x2204, 0x4404, 0x4828, 0x9010}},
    {21383, 2, {0x0, 0x3FFE, 0x2000, 0x2010, 0x20F8, 0x2780, 0x2080, 0x2080, 0x20FE, 0x3F80, 0x2080, 0x2082, 0x2082, 0x4082, 0x407E, 0x8000}},
    {21384, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x2FFC, 0x2080, 0x2080, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x2080, 0x2080, 0x4080, 0x4080, 0x8080}},
    {21385, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x2FFE, 0x2100, 0x2100, 0x21F8, 0x2108, 0x2108, 0x2108, 0x2208, 0x2208, 0x4408, 0x4850, 0x9020}},
    {21386, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x27FE, 0x2010, 0x2210, 0x2210, 0x2410, 0x27FE, 0x2050, 0x2090, 0x2110, 0x4210, 0x4C50, 0x8020}},
    {21387, 2, {0x0, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x2080, 0x2090, 0x2088, 0x2088, 0x4080, 0x5FFE, 0x8000}},
    {21388, 2, {0x0, 0x3FFE, 0x2000, 0x2080, 0x2090, 0x2088, 0x2080, 0x2FFC, 0x2080, 0x2140, 0x2140, 0x2220, 0x2220, 0x4410, 0x4808, 0x9006}},
    {21389, 2, {0x0, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2FFC, 0x2100, 0x2240, 0x2440, 0x27FC, 0x2040, 0x2040, 0x4FFE, 0x4040, 0x8040, 0x40}},
    {21390, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x203C, 0x27C0, 0x2440, 0x2440, 0x2440, 0x27FE, 0x2420, 0x2420, 0x2412, 0x450A, 0x4646, 0x8422}},
    {21391, 2, {0x0, 0x3FFE, 0x2000, 0x2200, 0x2200, 0x23FE, 0x2480, 0x2480, 0x28FC, 0x3080, 0x2080, 0x20FC, 0x2080, 0x4080, 0x4080, 0x8080}},
    {21392, 2, {0x0, 0x3FFE, 0x2000, 0x2220, 0x2210, 0x2200, 0x2FFC, 0x2240, 0x2240, 0x2248, 0x2450, 0x2460, 0x4842, 0x48C2, 0x933E, 0x2000}},
    {21393, 2, {0x0, 0x3FFE, 0x2000, 0x2020, 0x2FA0, 0x28A2, 0x28A4, 0x28A8, 0x28B0, 0x28A0, 0x28A0, 0x28A2, 0x2FA2, 0x48A2, 0x401E, 0x8000}},
    {21394, 2, {0x0, 0x3FFE, 0x2000, 0x2100, 0x2100, 0x23FC, 0x2240, 0x2440, 0x2040, 0x3FFE, 0x2040, 0x2444, 0x2444, 0x4444, 0x47FC, 0x8004}},
    {21395, 2, {0x0, 0x3FFE, 0x2040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x2000, 0x2040, 0x2040, 0x27FC, 0x2040, 0x4040, 0x4FFE, 0x8000}},
    {21396, 2, {0x0, 0x3FFE, 0x2000, 0x2000, 0x27FC, 0x2080, 0x2110, 0x2208, 0x27FC, 0x2044, 0x2040, 0x27FC, 0x2040, 0x4040, 0x4FFE, 0x8000}},
    {21397, 2, {0x0, 0x3FFE, 0x2000, 0x2F84, 0x28A4, 0x28A4, 0x2AA4, 0x2AA4, 0x2AA4, 0x2AA4, 0x2AA4, 0x2224, 0x2504, 0x4484, 0x4894, 0x9008}},
    {21398, 2, {0x0, 0x3FFE, 0x2410, 0x2408, 0x3FFE, 0x2490, 0x24A0, 0x24C8, 0x2490, 0x24A4, 0x24C8, 0x2492, 0x28E2, 0x4982, 0x567E, 0xA000}},
    {21399, 2, {0x0, 0x3FFE, 0x2080, 0x2040, 0x27FC, 0x2000, 0x2110, 0x20A0, 0x2FFE, 0x2040, 0x2040, 0x27FC, 0x2040, 0x4040, 0x4040, 0x8040}},
    {21400, 2, {0x0, 0x3FFE, 0x2000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2040, 0x2040, 0x27FC, 0x4040, 0x4040, 0x8FFE}},
    {21401, 2, {0x0, 0x3FFE, 0x2040, 0x2040, 0x2FFC, 0x2040, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2040, 0x2FFE, 0x4040, 0x4040, 0x8040}},
    {21402, 2, {0x0, 0x3FFE, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2000, 0x27F8, 0x2010, 0x2020, 0x2FFE, 0x4040, 0x4140, 0x8080}},
    {21403, 2, {0x0, 0x3FFE, 0x2000, 0x200C, 0x2F70, 0x2940, 0x2940, 0x2940, 0x297E, 0x2948, 0x2948, 0x2948, 0x2F48, 0x4088, 0x4088, 0x8108}},
    {21404, 2, {0x0, 0x3FFE, 0x2000, 0x2038, 0x27C0, 0x2040, 0x27FC, 0x2248, 0x2248, 0x2FFE, 0x2248, 0x2248, 0x2FFE, 0x4040, 0x4040, 0x87FC}},
    {21405, 2, {0x0, 0x3FFE, 0x2210, 0x2210, 0x2FFC, 0x2210, 0x2210, 0x3FFE, 0x2000, 0x27F8, 0x2408, 0x2408, 0x47F8, 0x4408, 0x8408, 0x7F8}},
    {21406, 2, {0x0, 0x3FFE, 0x2120, 0x2120, 0x2120, 0x3F3E, 0x2120, 0x2120, 0x2F3C, 0x2120, 0x2120, 0x3F3E, 0x2120, 0x4120, 0x4120, 0x8120}},
    {21407, 2, {0x0, 0x3FFE, 0x2080, 0x2100, 0x27F0, 0x2410, 0x2410, 0x27F0, 0x2410, 0x2410, 0x27F0, 0x2080, 0x2490, 0x4888, 0x5284, 0x8100}},
    {21408, 2, {0x0, 0x3FFE, 0x2000, 0x2F84, 0x2884, 0x28A4, 0x2FA4, 0x28A4, 0x2FA4, 0x28A4, 0x28A4, 0x2FA4, 0x2024, 0x4904, 0x4894, 0x9088}},
    {21409, 2, {0x0, 0x3FFE, 0x2080, 0x2100, 0x27FC, 0x2404, 0x27FC, 0x2404, 0x27FC, 0x2040, 0x2744, 0x2168, 0x2250, 0x4448, 0x5946, 0x8080}},
    {21410, 2, {0x0, 0x3FFE, 0x2200, 0x2200, 0x227C, 0x2F44, 0x2244, 0x267C, 0x2744, 0x2AC4, 0x2A7C, 0x3244, 0x2244, 0x4244, 0x427C, 0x8244}},
    {21411, 2, {0x0, 0x3FFE, 0x20A0, 0x2090, 0x3FFE, 0x2140, 0x2630, 0x380E, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x4FF8, 0x4888, 0x8080, 0x80}},
    {21412, 2, {0x0, 0x3FFE, 0x2000, 0x230C, 0x2E70, 0x2210, 0x2210, 0x3FFE, 0x2210, 0x2738, 0x26B4, 0x2A54, 0x2A52, 0x5290, 0x4210, 0x8210}},
    {21413, 2, {0x0, 0x3FFE, 0x2000, 0x2890, 0x2510, 0x2FBE, 0x2222, 0x2244, 0x2A90, 0x2A90, 0x2F90, 0x2210, 0x4228, 0x4428, 0x8444, 0x882}},
    {21414, 2, {0x3FFE, 0x2000, 0x3FFC, 0x2100, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2200, 0x47F8, 0x4A08, 0x81F0, 0x1E0E}},
    {21415, 2, {0x0, 0x3FFE, 0x2080, 0x2FFC, 0x2080, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x7FFE, 0x4210, 0x8408}},
    {21416, 2, {0x0, 0x3FFE, 0x2008, 0x3F88, 0x2008, 0x2F08, 0x297E, 0x2908, 0x2908, 0x2F48, 0x2028, 0x2908, 0x2608, 0x4388, 0x5C28, 0x8810}},
    {21417, 2, {0x0, 0x3FFE, 0x2000, 0x2F7E, 0x2910, 0x2910, 0x2F50, 0x2950, 0x297E, 0x2F10, 0x2828, 0x2A28, 0x2928, 0x4B4A, 0x4D4A, 0x8886}},
    {21418, 2, {0x0, 0x3FFE, 0x2110, 0x2FFE, 0x2110, 0x21F0, 0x2040, 0x27FC, 0x2444, 0x27FC, 0x2040, 0x2FFE, 0x2040, 0x47FC, 0x4040, 0x8FFE}},
    {21419, 2, {0x0, 0x3FFE, 0x2210, 0x2210, 0x2F90, 0x221E, 0x2FA4, 0x2224, 0x3FD4, 0x2414, 0x2794, 0x2488, 0x2488, 0x4894, 0x4AA4, 0x9142}},
    {21420, 2, {0x3FFE, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2220, 0x27B0, 0x2D28, 0x27FE, 0x3800, 0x47F8, 0x4408, 0x87F8, 0x408}},
    {21421, 2, {0x0, 0x3FFE, 0x2010, 0x2F94, 0x2892, 0x2F90, 0x28FE, 0x2F90, 0x2010, 0x2F90, 0x2890, 0x2F90, 0x28A8, 0x4FA8, 0x48C4, 0x8982}},
    {21422, 2, {0x0, 0x3FFE, 0x2900, 0x290C, 0x3FB0, 0x2920, 0x2920, 0x2F3E, 0x2928, 0x2F28, 0x2928, 0x2928, 0x5FA8, 0x4048, 0x8A88, 0x1108}},
    {21423, 2, {0x0, 0x3FFE, 0x2210, 0x2210, 0x3FFE, 0x2718, 0x2AB4, 0x3252, 0x2210, 0x2210, 0x2080, 0x2444, 0x2452, 0x4412, 0x43F0, 0x8000}},
    {21424, 2, {0x0, 0x7FFE, 0x4208, 0x5248, 0x4A50, 0x4A9E, 0x5FE4, 0x5054, 0x5754, 0x5554, 0x5548, 0x5748, 0x5048, 0x5154, 0x5094, 0x8022}},
    {21425, 2, {0x0, 0x3FFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2DF6, 0x2000, 0x27BC, 0x24A4, 0x24A4, 0x27BC, 0x2110, 0x42A8, 0x4444, 0x8882}},
    {21426, 2, {0x0, 0x3FFE, 0x2110, 0x2FFE, 0x2110, 0x23F8, 0x2248, 0x23F8, 0x2248, 0x23F8, 0x2040, 0x27FC, 0x2454, 0x45F4, 0x4414, 0x840C}},
    {21427, 2, {0x1104, 0x884, 0x888, 0x3FFE, 0x2000, 0x2710, 0x211E, 0x2FD4, 0x24A4, 0x2794, 0x2494, 0x2788, 0x24E8, 0x5F94, 0x40A4, 0x80C2}},
    {21428, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x2000, 0x27FC, 0x2444, 0x27FC, 0x4444, 0x47FC, 0x8040}},
    {21429, 2, {0x3FFE, 0x2080, 0x2FF8, 0x2808, 0x2FF8, 0x4490, 0x4A88, 0x8100, 0x7EFE, 0x4890, 0x5EBC, 0x52A4, 0x5EBC, 0x4890, 0x6AD4, 0x9930}},
    {21430, 2, {0x200, 0x200, 0x200, 0x200, 0x400, 0x400, 0x400, 0x800, 0x840, 0x1020, 0x1010, 0x2010, 0x7FF8, 0x2008, 0x8, 0x0}},
    {21431, 2, {0x200, 0x200, 0x200, 0xFFFE, 0x400, 0x400, 0x440, 0x840, 0x880, 0x1080, 0x1100, 0x2210, 0x4408, 0x8FFC, 0x404, 0x4}},
    {21432, 2, {0x808, 0x808, 0x808, 0x808, 0x808, 0x1010, 0x1010, 0x1010, 0x2020, 0x2424, 0x2424, 0x4242, 0xFEFE, 0x4242, 0x202, 0x0}},
    {21433, 2, {0x200, 0x200, 0x3FE0, 0x220, 0x220, 0x420, 0x422, 0x822, 0x311E, 0xC100, 0x200, 0x420, 0x810, 0x1008, 0x3FFC, 0x1004}},
    {21434, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x920, 0x3118, 0xC206, 0x200, 0x420, 0x810, 0x1008, 0x3FFC, 0x1004, 0x4}},
    {21435, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x200, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8}},
    {21436, 2, {0x200, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1108, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x100, 0x500, 0x200}},
    {21437, 2, {0x200, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x1010, 0x1010, 0x1010, 0x2020, 0x2424, 0x4242, 0xFEFE, 0x4242, 0x0}},
    {21438, 2, {0x0, 0x7FF0, 0x410, 0x410, 0x410, 0x3F90, 0x410, 0x410, 0x7FD0, 0x410, 0x810, 0x1112, 0x208A, 0x7FCA, 0x46, 0x2}},
    {21439, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {21440, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x400, 0x820, 0x1010, 0x3FF8, 0x8}},
    {21441, 2, {0x200, 0x440, 0x820, 0x1FF0, 0x200, 0x7FFC, 0x820, 0x1010, 0x27C8, 0xC006, 0x0, 0xFE0, 0x0, 0x0, 0x3FF8, 0x0}},
    {21442, 2, {0x200, 0x440, 0x820, 0x1FF0, 0x200, 0x7FFC, 0x820, 0x1090, 0x2308, 0xCC46, 0x180, 0x610, 0x1820, 0xC0, 0x700, 0x7800}},
    {21443, 2, {0x400, 0x820, 0x1FF0, 0x1020, 0x2448, 0x7EFC, 0x100, 0x6C0, 0x18B0, 0xE30E, 0xC40, 0x3190, 0x620, 0x38C0, 0x700, 0x7800}},
    {21444, 2, {0x400, 0x820, 0x1FF0, 0x1020, 0x2448, 0x7EFC, 0x100, 0x6C0, 0x1830, 0xE00E, 0xFE0, 0x0, 0x1FF0, 0x0, 0x7FFC, 0x0}},
    {21445, 2, {0x400, 0x820, 0x1FF0, 0x1020, 0x2448, 0x7EFC, 0x100, 0x6C0, 0x1830, 0xE20E, 0x220, 0x1290, 0x1248, 0x2248, 0xA00, 0x400}},
    {21446, 2, {0xE, 0x3F0, 0x7844, 0x224, 0x108, 0x3FE, 0xFE42, 0x1040, 0x13FE, 0x2080, 0x28FC, 0x4544, 0xFD28, 0x4210, 0x468, 0x186}},
    {21447, 2, {0x10, 0x210, 0x797C, 0x14, 0xFE, 0x14, 0xFB7C, 0x2192, 0x2154, 0x4138, 0x5154, 0x4992, 0xF930, 0x4900, 0x2FE, 0x400}},
    {21448, 2, {0x0, 0x3FF0, 0x1010, 0x1010, 0x1020, 0x820, 0x820, 0x440, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {21449, 2, {0x0, 0x3FF0, 0x1010, 0x1210, 0x1120, 0x920, 0x820, 0x440, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {21450, 2, {0x0, 0x3FE0, 0x820, 0x820, 0x840, 0x840, 0xCF8, 0xA08, 0xA08, 0x910, 0x1110, 0x10A0, 0x2040, 0x20A0, 0x4318, 0x8C06}},
    {21451, 2, {0x200, 0x200, 0x200, 0xFFFE, 0x400, 0x400, 0xFF0, 0xA10, 0x1210, 0x1120, 0x2140, 0x4080, 0x8140, 0x620, 0x1818, 0x6006}},
    {21452, 2, {0x0, 0x0, 0xFDFC, 0x484, 0x4484, 0x4484, 0x2888, 0x2888, 0x1050, 0x1050, 0x2820, 0x2820, 0x4450, 0x4488, 0x8104, 0x202}},
    {21453, 2, {0x10, 0xF8, 0x3F00, 0x2000, 0x2000, 0x3FF8, 0x2408, 0x2410, 0x2210, 0x2220, 0x2140, 0x2080, 0x4140, 0x4220, 0x8C18, 0x3006}},
    {21454, 2, {0x400, 0x400, 0x45FC, 0x4484, 0x4484, 0x4484, 0x4488, 0x4488, 0x4C50, 0x5450, 0x6420, 0x4420, 0x450, 0x488, 0x504, 0x602}},
    {21455, 2, {0x200, 0x200, 0x3FF0, 0x210, 0x210, 0x210, 0xFFFE, 0x200, 0x200, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {21456, 2, {0xFC, 0x3F00, 0x100, 0x100, 0x7FFE, 0x200, 0x400, 0x7F8, 0xA08, 0xA08, 0x1110, 0x20A0, 0x4040, 0x80A0, 0x318, 0xC06}},
    {21457, 2, {0x100, 0x1110, 0x1108, 0x2200, 0x3FFC, 0x200, 0x400, 0x7F8, 0xA08, 0x908, 0x1110, 0x10A0, 0x2040, 0x40A0, 0x318, 0x1C06}},
    {21458, 2, {0x0, 0x3FF0, 0x20, 0x1E40, 0x180, 0x660, 0x1818, 0xE004, 0x0, 0x3EF8, 0x288, 0x2288, 0x1450, 0x820, 0x34D8, 0xC306}},
    {21459, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {21460, 2, {0x800, 0x800, 0xEFC, 0x844, 0x844, 0x844, 0xFF44, 0x28, 0x828, 0x828, 0x4A10, 0x4910, 0x8928, 0x848, 0x2884, 0x1102}},
    {21461, 2, {0x0, 0x3EF8, 0x208, 0x2288, 0x1450, 0x820, 0x34D0, 0xC308, 0x0, 0x3EF8, 0x288, 0x2288, 0x1450, 0x820, 0x34D8, 0xC306}},
    {21462, 2, {0x0, 0xFF80, 0x22FC, 0x2244, 0x3E44, 0x2244, 0x2244, 0x3E44, 0x2228, 0x2228, 0x27A8, 0xFA10, 0x4210, 0x228, 0x244, 0x282}},
    {21463, 2, {0x8, 0x1FC, 0x7E10, 0x2210, 0x1120, 0x7FFE, 0x4002, 0x8004, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC006}},
    {21464, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4444, 0x0, 0x3FF0, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {21465, 2, {0x800, 0x800, 0x14FC, 0x2244, 0x4144, 0xBE44, 0x844, 0x828, 0x7F28, 0x828, 0x2A10, 0x2910, 0x4928, 0x8848, 0x2884, 0x1102}},
    {21466, 2, {0x0, 0x3E7C, 0x2204, 0x2204, 0x2204, 0x3E7C, 0x2000, 0x2000, 0x3EFC, 0x2044, 0x2044, 0x3E28, 0x2010, 0x2028, 0x20C4, 0x2302}},
    {21467, 2, {0x1008, 0x101C, 0x95E0, 0x5900, 0x1100, 0x7DFC, 0x1144, 0x1144, 0xFF44, 0x1128, 0x1128, 0x1110, 0x2110, 0x2228, 0x4244, 0x8482}},
    {21468, 2, {0x200, 0x100, 0x7FFE, 0x4112, 0x9124, 0x22C0, 0x430, 0x180C, 0xE002, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {21469, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0x7C94, 0x1088, 0x1080, 0x10FC, 0xFEA4, 0x10A4, 0x10A8, 0x24A8, 0x4290, 0xFEA8, 0x42C4, 0x82}},
    {21470, 2, {0x0, 0x7F00, 0x417C, 0x4124, 0x7F24, 0x4024, 0x5F24, 0x4028, 0x4028, 0x7FA8, 0x4410, 0x5510, 0x94A8, 0xA4A8, 0x1444, 0x882}},
    {21471, 2, {0x100, 0x500, 0x3978, 0x2108, 0x2108, 0x3D78, 0x2108, 0x2108, 0x3FF8, 0x100, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {21472, 2, {0x1FE0, 0x440, 0x380, 0xC40, 0x7EFC, 0x2244, 0x1C38, 0x6244, 0x0, 0xFFFE, 0x9012, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {21473, 2, {0x1000, 0x1F00, 0x107C, 0xFFA4, 0x80A4, 0x7F24, 0x2A24, 0x5528, 0xA228, 0x7F28, 0xA290, 0x3E10, 0x2228, 0x3E28, 0x2244, 0x3E82}},
    {21474, 2, {0x2448, 0x1450, 0xFFFE, 0x820, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x2200, 0x3EFC, 0x2244, 0x3E28, 0x2390, 0xFE28, 0x2C6}},
    {21475, 2, {0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x0}},
    {21476, 2, {0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21477, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x9F88, 0x1088, 0x1088, 0x1088, 0x1088, 0x1F88, 0x1088, 0x8, 0x50, 0x20}},
    {21478, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x200, 0x200, 0x7FF8, 0x208, 0x408, 0x408, 0x808, 0x1008, 0x2050, 0x4020}},
    {21479, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FF8, 0x408, 0x408, 0x808, 0x808, 0x1008, 0x2050, 0xC020}},
    {21480, 2, {0x0, 0x0, 0x7DFC, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x7C44, 0x4484, 0x84, 0x104, 0x228, 0x410}},
    {21481, 2, {0x0, 0xFC, 0x7C84, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x7C94, 0x4488, 0x80, 0x80, 0x80, 0x80}},
    {21482, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x0, 0x820, 0x810, 0x1008, 0x2004, 0x4004}},
    {21483, 2, {0x4, 0x4, 0x7C84, 0x4484, 0x4484, 0x4484, 0x4484, 0x4484, 0x448C, 0x44B4, 0x7CC4, 0x4484, 0x4, 0x4, 0x4, 0x4}},
    {21484, 2, {0x0, 0x7FF8, 0x408, 0x408, 0x808, 0x808, 0x1050, 0x2020, 0x4000, 0x9FF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {21485, 2, {0x0, 0x20, 0x7920, 0x4920, 0x4920, 0x4920, 0x4920, 0x4920, 0x4920, 0x4910, 0x7910, 0x4A10, 0x208, 0x408, 0x804, 0x1002}},
    {21486, 2, {0x0, 0x0, 0x7BFE, 0x4820, 0x4820, 0x4820, 0x4820, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0xA0, 0x40}},
    {21487, 2, {0x0, 0x7FFE, 0x10, 0x10, 0x1F10, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1F10, 0x1110, 0x10, 0x10, 0x50, 0x20}},
    {21488, 2, {0x200, 0x200, 0x400, 0x820, 0x1010, 0x2008, 0x7FFC, 0x2004, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21489, 2, {0x40, 0x40, 0x7C40, 0x4442, 0x4444, 0x4448, 0x4450, 0x4460, 0x4440, 0x44C0, 0x7D40, 0x4442, 0x42, 0x42, 0x3E, 0x0}},
    {21490, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x1100, 0x900, 0x600, 0x580, 0x1870, 0xE00E}},
    {21491, 2, {0x200, 0x200, 0x200, 0xFFFE, 0x400, 0x400, 0x800, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8808, 0x808, 0x808, 0xFF8, 0x808}},
    {21492, 2, {0x200, 0x200, 0x3FE0, 0x220, 0x420, 0x422, 0x822, 0x301E, 0xC000, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21493, 2, {0x0, 0x7FFC, 0x4000, 0x4000, 0x4000, 0x47F0, 0x4410, 0x4410, 0x4410, 0x4410, 0x47F0, 0x4410, 0x4000, 0x4000, 0x7FFE, 0x0}},
    {21494, 2, {0x20, 0x20, 0x7820, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21495, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x800, 0x1000, 0x1FF0, 0x10, 0x10, 0x10, 0xA0, 0x40}},
    {21496, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x1F88, 0x1088, 0x1088, 0x1088, 0x1088, 0x1F88, 0x1088, 0x28, 0x10}},
    {21497, 2, {0x0, 0x7F8, 0x7A08, 0x4A08, 0x4A08, 0x4910, 0x4910, 0x4910, 0x48A0, 0x48A0, 0x7840, 0x4840, 0xA0, 0x110, 0x208, 0xC06}},
    {21498, 2, {0x200, 0x100, 0x7880, 0x4880, 0x4840, 0x4840, 0x4840, 0x48A0, 0x48A0, 0x48A0, 0x7910, 0x4910, 0x208, 0x208, 0x404, 0x802}},
    {21499, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4BFC, 0x4844, 0x4844, 0x4844, 0x4844, 0x4884, 0x7884, 0x4884, 0x104, 0x104, 0x228, 0x410}},
    {21500, 2, {0x0, 0x3FC, 0x7804, 0x4804, 0x4804, 0x4814, 0x4824, 0x4844, 0x4884, 0x4904, 0x4A04, 0x7804, 0x4804, 0x4, 0x28, 0x10}},
    {21501, 2, {0x0, 0x1F0, 0x7910, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x7A10, 0x4A12, 0x212, 0x412, 0x40E, 0x800}},
    {21502, 2, {0x0, 0x3FF8, 0x60, 0x180, 0x100, 0x100, 0x100, 0x500, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21503, 2, {0x0, 0x0, 0x7DFC, 0x4420, 0x4420, 0x4420, 0x4420, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20, 0x3FE, 0x0}},
    {21504, 2, {0x8, 0x3C, 0x7DE0, 0x4420, 0x4420, 0x4420, 0x4420, 0x47FE, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20, 0x20, 0x20}},
    {21505, 2, {0x0, 0x1FC, 0x7820, 0x4820, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0xA0, 0x40}},
    {21506, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21507, 2, {0x80, 0x80, 0x7900, 0x49FE, 0x4A00, 0x4C00, 0x49F8, 0x4808, 0x4810, 0x4860, 0x7880, 0x4900, 0x202, 0x202, 0x1FE, 0x0}},
    {21508, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x24C0, 0x300, 0xCC0, 0x3030, 0xC00E, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21509, 2, {0x0, 0x0, 0x7EFC, 0x4284, 0x4284, 0x4284, 0x4284, 0x4284, 0x4284, 0x4284, 0x4284, 0x4284, 0x7EFC, 0x4284, 0x0, 0x0}},
    {21510, 2, {0x20, 0x20, 0x7840, 0x4840, 0x4884, 0x4884, 0x4908, 0x4BF8, 0x4910, 0x4810, 0x7820, 0x4840, 0x88, 0x104, 0x3FE, 0x102}},
    {21511, 2, {0x0, 0x3F8, 0x7808, 0x4810, 0x4820, 0x4840, 0x4840, 0x4FFE, 0x4840, 0x4840, 0x7840, 0x4840, 0x40, 0x40, 0x140, 0x80}},
    {21512, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21513, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x3FF8, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21514, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {21515, 2, {0x10, 0x10, 0x7810, 0x4810, 0x4BFE, 0x4810, 0x4810, 0x4810, 0x4910, 0x4890, 0x7890, 0x4810, 0x10, 0x10, 0x50, 0x20}},
    {21516, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x2FF4, 0x2004, 0x2004, 0x27E4, 0x2424, 0x2424, 0x2424, 0x27E4, 0x2424, 0x2004, 0x2014, 0x2008}},
    {21517, 2, {0x200, 0x200, 0x7F0, 0x810, 0x1420, 0x6240, 0x280, 0x100, 0x600, 0x1FF8, 0xE808, 0x808, 0x808, 0x808, 0xFF8, 0x808}},
    {21518, 2, {0x10, 0xF8, 0x1F00, 0x1000, 0x1000, 0x1FFE, 0x1000, 0x1000, 0x1000, 0x17F8, 0x1408, 0x2408, 0x2408, 0x4408, 0x87F8, 0x408}},
    {21519, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x1100, 0x900, 0x600, 0x580, 0x1870, 0xE00E}},
    {21520, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4840, 0x4840, 0x4FFC, 0x4840, 0x4840, 0x4840, 0x7840, 0x4840, 0x40, 0x40, 0xFFE, 0x0}},
    {21521, 2, {0x200, 0x400, 0x800, 0x7FFC, 0x4004, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4004, 0x4014, 0x4008}},
    {21522, 2, {0x8, 0x3C, 0x7BC0, 0x4840, 0x4840, 0x4840, 0x4840, 0x487E, 0x4BC0, 0x4840, 0x7840, 0x4842, 0x42, 0x42, 0x3E, 0x0}},
    {21523, 2, {0x0, 0x7FE, 0x7840, 0x4840, 0x4840, 0x4840, 0x4850, 0x4848, 0x4844, 0x4844, 0x7840, 0x4840, 0x40, 0x40, 0x40, 0x40}},
    {21524, 2, {0x20, 0x20, 0x7820, 0x4920, 0x492C, 0x4934, 0x4964, 0x4BA4, 0x4924, 0x4934, 0x7928, 0x4922, 0x122, 0x102, 0xFE, 0x0}},
    {21525, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21526, 2, {0x0, 0x202, 0x7904, 0x4888, 0x4850, 0x4820, 0x4820, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21527, 2, {0x0, 0x3F0, 0x7810, 0x4810, 0x4910, 0x4910, 0x4910, 0x49FC, 0x4804, 0x4804, 0x7804, 0x4BF4, 0x4, 0x4, 0x28, 0x10}},
    {21528, 2, {0x80, 0x80, 0x7880, 0x49FC, 0x4920, 0x4A20, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21529, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4A44, 0x4A44, 0x4A48, 0x4C50, 0x4840, 0x48A0, 0x78A0, 0x4910, 0x110, 0x208, 0x404, 0x802}},
    {21530, 2, {0x0, 0x3F8, 0x7888, 0x4888, 0x4888, 0x4FFE, 0x4888, 0x4888, 0x4888, 0x4BF8, 0x7888, 0x4880, 0x100, 0x100, 0x200, 0x400}},
    {21531, 2, {0x0, 0x3FF8, 0x408, 0x408, 0xFFFE, 0x408, 0x408, 0x3FF8, 0x800, 0x800, 0x1FF8, 0x1808, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {21532, 2, {0x0, 0x3F8, 0x7888, 0x4888, 0x4888, 0x4888, 0x4888, 0x4FF8, 0x4908, 0x4908, 0x7908, 0x4908, 0x108, 0x108, 0x7FE, 0x0}},
    {21533, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC006, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21534, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21535, 2, {0x40, 0x40, 0x78A0, 0x48A0, 0x4910, 0x4A88, 0x4C46, 0x4840, 0x4800, 0x4BF8, 0x7808, 0x4810, 0x10, 0x20, 0x20, 0x40}},
    {21536, 2, {0x40, 0x50, 0x7848, 0x4848, 0x4840, 0x4FFE, 0x4840, 0x4840, 0x48A0, 0x48A0, 0x48A0, 0x7910, 0x4910, 0x208, 0x404, 0x802}},
    {21537, 2, {0x20, 0x220, 0x7A20, 0x4A22, 0x4A22, 0x4A24, 0x4BA8, 0x4A30, 0x4A20, 0x4A20, 0x7A20, 0x4A22, 0x2A2, 0x322, 0x21E, 0x0}},
    {21538, 2, {0x100, 0x80, 0x888, 0x4804, 0x4812, 0x4812, 0x87F0, 0x0, 0x0, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {21539, 2, {0x0, 0x40, 0xF020, 0x9010, 0x9090, 0x9080, 0x9080, 0x9284, 0x9282, 0x9282, 0xF282, 0x9488, 0x88, 0x88, 0x78, 0x0}},
    {21540, 2, {0x40, 0x40, 0x78A0, 0x48A0, 0x4910, 0x4A08, 0x4C06, 0x4910, 0x4910, 0x4910, 0x7910, 0x4910, 0x110, 0x210, 0x210, 0x410}},
    {21541, 2, {0x0, 0x0, 0x7BFE, 0x4810, 0x4810, 0x4820, 0x4820, 0x4868, 0x4864, 0x48A4, 0x7922, 0x4A22, 0x420, 0x20, 0x20, 0x20}},
    {21542, 2, {0x0, 0x7FFC, 0x80, 0x100, 0x300, 0x560, 0x918, 0x3104, 0xC102, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21543, 2, {0x0, 0x3F8, 0x7A48, 0x4A48, 0x4A48, 0x4A48, 0x4A48, 0x4BF8, 0x4A08, 0x4A00, 0x7A00, 0x4A02, 0x202, 0x202, 0x1FE, 0x0}},
    {21544, 2, {0x40, 0x40, 0x7840, 0x4BFC, 0x4840, 0x4A48, 0x4A48, 0x4A48, 0x4A48, 0x4BF8, 0x7848, 0x4840, 0x42, 0x42, 0x3E, 0x0}},
    {21545, 2, {0x10, 0x90, 0x7890, 0x4888, 0x4908, 0x4904, 0x4A04, 0x4DFA, 0x4888, 0x4888, 0x4888, 0x7888, 0x4908, 0x108, 0x228, 0x410}},
    {21546, 2, {0xA0, 0xA0, 0x78A0, 0x4924, 0x4924, 0x4B28, 0x4D30, 0x4920, 0x4960, 0x49A0, 0x7920, 0x4922, 0x122, 0x122, 0x11E, 0x100}},
    {21547, 2, {0x100, 0x100, 0x280, 0x440, 0xA20, 0x3118, 0xC106, 0xFE0, 0x40, 0x80, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21548, 2, {0x8, 0x1C, 0x79E0, 0x4900, 0x4900, 0x4900, 0x49FE, 0x4910, 0x4910, 0x4910, 0x7910, 0x4910, 0x110, 0x210, 0x210, 0x410}},
    {21549, 2, {0x80, 0x40, 0x7840, 0x4FFC, 0x4800, 0x4800, 0x49F0, 0x4910, 0x4910, 0x4910, 0x7910, 0x4912, 0x212, 0x212, 0x40E, 0x800}},
    {21550, 2, {0x80, 0x80, 0x7900, 0x4910, 0x4A08, 0x4FFC, 0x4804, 0x4920, 0x4920, 0x4920, 0x7920, 0x4920, 0x222, 0x222, 0x41E, 0x800}},
    {21551, 2, {0x100, 0x80, 0x1FFC, 0x1004, 0x1004, 0x1004, 0x1FFC, 0x1000, 0x1000, 0x1000, 0x17FC, 0x2404, 0x2404, 0x4404, 0x87FC, 0x404}},
    {21552, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4BFE, 0x4840, 0x4880, 0x4890, 0x4890, 0x4920, 0x7920, 0x4A48, 0x244, 0x484, 0x1FE, 0x82}},
    {21553, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x49FC, 0x4884, 0x4888, 0x7848, 0x4850, 0x20, 0x50, 0x188, 0x606}},
    {21554, 2, {0x4, 0x3E4, 0x7824, 0x4824, 0x4824, 0x49E4, 0x4904, 0x4904, 0x4A04, 0x4BE4, 0x7824, 0x4824, 0x24, 0x24, 0x144, 0x84}},
    {21555, 2, {0x0, 0x7E0, 0x420, 0x1420, 0x17E0, 0x1000, 0x1000, 0x1FF0, 0x110, 0x110, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {21556, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {21557, 2, {0x40, 0x40, 0x7840, 0x4948, 0x4944, 0x4A42, 0x4A42, 0x4C48, 0x4848, 0x4848, 0x7810, 0x4810, 0x20, 0x40, 0x180, 0x600}},
    {21558, 2, {0x40, 0x20, 0x20, 0xFBFE, 0x8A22, 0x8A22, 0x8A32, 0x8A4A, 0x8A4A, 0x8A86, 0xFB06, 0x8A02, 0x202, 0x202, 0x20A, 0x204}},
    {21559, 2, {0x40, 0x40, 0x7840, 0x4BF8, 0x4848, 0x4848, 0x4848, 0x4848, 0x4FFE, 0x4840, 0x78A0, 0x48A0, 0x110, 0x110, 0x208, 0x406}},
    {21560, 2, {0x0, 0x7F8, 0x7908, 0x4910, 0x4910, 0x4920, 0x493C, 0x4904, 0x4904, 0x4A88, 0x7A88, 0x4A50, 0x420, 0x450, 0x888, 0x306}},
    {21561, 2, {0x80, 0x80, 0x7880, 0x48FC, 0x4904, 0x4908, 0x4A40, 0x4C40, 0x4840, 0x48A0, 0x78A0, 0x4910, 0x110, 0x208, 0x404, 0x802}},
    {21562, 2, {0x0, 0x1F0, 0x7910, 0x4910, 0x4910, 0x4A0E, 0x4C00, 0x4BF8, 0x4908, 0x4910, 0x7890, 0x48A0, 0x40, 0xA0, 0x318, 0xC06}},
    {21563, 2, {0x100, 0x100, 0x7900, 0x49FC, 0x4AA4, 0x4CA4, 0x48A4, 0x4924, 0x4924, 0x4A44, 0x7C44, 0x4884, 0x104, 0x204, 0x428, 0x10}},
    {21564, 2, {0x10, 0x7D0, 0x7850, 0x4890, 0x4910, 0x4910, 0x4950, 0x4990, 0x4910, 0x4F10, 0x7910, 0x4912, 0x112, 0x112, 0x50E, 0x200}},
    {21565, 2, {0x20, 0x20, 0x7920, 0x4920, 0x49FC, 0x4920, 0x4A20, 0x4820, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21566, 2, {0x0, 0x7FFC, 0x200, 0x200, 0x3FF0, 0x410, 0x410, 0x810, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21567, 2, {0x100, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21568, 2, {0x0, 0x3FC, 0x7810, 0x4910, 0x4910, 0x4910, 0x4A10, 0x4BFE, 0x4830, 0x4850, 0x7890, 0x4910, 0x210, 0x410, 0x50, 0x20}},
    {21569, 2, {0x40, 0x40, 0x7880, 0x48FC, 0x4904, 0x4A04, 0x4884, 0x4844, 0x4844, 0x4814, 0x4824, 0x7844, 0x4984, 0x4, 0x28, 0x10}},
    {21570, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21571, 2, {0x0, 0x3FE, 0x7A00, 0x4A00, 0x4AF8, 0x4A88, 0x4A88, 0x4A88, 0x4A88, 0x4AA8, 0x7A90, 0x4A82, 0x482, 0x482, 0x87E, 0x1000}},
    {21572, 2, {0x0, 0x3FE0, 0x420, 0x47C, 0x804, 0x1014, 0x6108, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21573, 2, {0x80, 0x40, 0x7840, 0x4FFC, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x48A0, 0x78A0, 0x4840, 0xA0, 0x110, 0x208, 0xC06}},
    {21574, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {21575, 2, {0x100, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21576, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {21577, 2, {0x0, 0x7E0, 0x420, 0x1420, 0x17E0, 0x1000, 0x1000, 0x1FF0, 0x10, 0x10, 0xFFFE, 0x0, 0x440, 0x820, 0x1010, 0x2008}},
    {21578, 2, {0x100, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21579, 2, {0x40, 0x40, 0x7840, 0x4BFC, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x4840, 0x48A0, 0x78A0, 0x4910, 0x110, 0x208, 0x404, 0x802}},
    {21580, 2, {0x10, 0x110, 0x7890, 0x4890, 0x4810, 0x4910, 0x4890, 0x4890, 0x4810, 0x481E, 0x7BF0, 0x4810, 0x10, 0x10, 0x10, 0x10}},
    {21581, 2, {0x0, 0x0, 0x79FC, 0x4800, 0x4800, 0x4800, 0x4BFE, 0x4820, 0x4820, 0x4840, 0x7840, 0x4888, 0x104, 0x3FE, 0x102, 0x0}},
    {21582, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4924, 0x4920, 0x4920, 0x7910, 0x4910, 0x208, 0x208, 0x404, 0x802}},
    {21583, 2, {0x8, 0x68, 0x7B88, 0x4888, 0x4888, 0x4888, 0x4888, 0x4BFE, 0x4888, 0x4888, 0x7888, 0x4888, 0x108, 0x108, 0x208, 0x408}},
    {21584, 2, {0x40, 0x40, 0x7840, 0x4BFC, 0x4A44, 0x4A44, 0x4A44, 0x4A44, 0x4AA4, 0x4A94, 0x7B14, 0x4A04, 0x204, 0x204, 0x214, 0x208}},
    {21585, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC006, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21586, 2, {0x0, 0x3FC, 0x7840, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x4840, 0x48A0, 0x48A0, 0x78A0, 0x4920, 0x122, 0x222, 0x41E, 0x800}},
    {21587, 2, {0x90, 0x90, 0x7BFE, 0x4890, 0x4890, 0x4800, 0x4BF8, 0x4810, 0x4820, 0x4840, 0x7880, 0x4900, 0x202, 0x202, 0x1FE, 0x0}},
    {21588, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4FFE, 0x4840, 0x4840, 0x48A0, 0x48A0, 0x48A0, 0x7910, 0x4910, 0x288, 0x248, 0x444, 0x802}},
    {21589, 2, {0x0, 0x3FC, 0x7A00, 0x4A08, 0x4A88, 0x4A50, 0x4A50, 0x4A20, 0x4A20, 0x4A50, 0x7A50, 0x4A88, 0x308, 0x200, 0x3FE, 0x0}},
    {21590, 2, {0x0, 0x3FE, 0x7A00, 0x4A20, 0x4A20, 0x4A20, 0x4AFC, 0x4A24, 0x4A24, 0x4A24, 0x7A44, 0x4A44, 0x484, 0x484, 0x928, 0x1210}},
    {21591, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x7924, 0x4850, 0x48, 0x84, 0x102, 0x602}},
    {21592, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2108, 0x2108, 0x2108, 0x2108, 0x22C8, 0x430, 0x1808, 0x6004}},
    {21593, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x4104, 0x4284, 0x4444, 0x4824, 0x4004, 0x4014, 0x4008}},
    {21594, 2, {0x40, 0x40, 0x7C40, 0x4480, 0x44FE, 0x4508, 0x4688, 0x4488, 0x4488, 0x4450, 0x7C50, 0x4420, 0x50, 0x88, 0x104, 0x202}},
    {21595, 2, {0x40, 0x40, 0x78A0, 0x48A0, 0x4910, 0x4A08, 0x4C06, 0x49F0, 0x4910, 0x4910, 0x7950, 0x4920, 0x104, 0x104, 0xFC, 0x0}},
    {21596, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x4904, 0x4904, 0x4914, 0x4908, 0x4900, 0x49FE, 0x7802, 0x4802, 0x3FA, 0x2, 0x14, 0x8}},
    {21597, 2, {0x40, 0x20, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x4900, 0x4900, 0x497C, 0x4908, 0x7910, 0x4920, 0x142, 0x282, 0x27E, 0x400}},
    {21598, 2, {0x0, 0x1FC, 0x7804, 0x4804, 0x4BF4, 0x4804, 0x4804, 0x49F4, 0x4914, 0x4914, 0x7914, 0x49F4, 0x4, 0x4, 0x14, 0x8}},
    {21599, 2, {0x80, 0x40, 0x7840, 0x4BFE, 0x4840, 0x4840, 0x4888, 0x4908, 0x4BF0, 0x4820, 0x7840, 0x4888, 0x104, 0x3FE, 0x102, 0x0}},
    {21600, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4FFC, 0x4840, 0x48E0, 0x48E0, 0x4950, 0x4950, 0x4A48, 0x7DF4, 0x4842, 0x40, 0x40, 0x40}},
    {21601, 2, {0x0, 0x3FC, 0x7A04, 0x4A04, 0x4A04, 0x4BFC, 0x4A20, 0x4A20, 0x4BFE, 0x4A20, 0x7A20, 0x4A10, 0x212, 0x28A, 0x306, 0x202}},
    {21602, 2, {0x0, 0x3FC, 0x7A04, 0x4A04, 0x4A04, 0x4BFC, 0x4A00, 0x4A40, 0x4A44, 0x4A48, 0x7A70, 0x4A40, 0x442, 0x442, 0x83E, 0x1000}},
    {21603, 2, {0x0, 0x1F8, 0x7908, 0x4948, 0x4928, 0x4908, 0x4908, 0x4FFE, 0x4A08, 0x4A88, 0x7A48, 0x4A08, 0x3FE, 0x8, 0x50, 0x20}},
    {21604, 2, {0x40, 0x40, 0x78A0, 0x48A0, 0x4910, 0x4A48, 0x4C26, 0x4820, 0x4BF8, 0x4808, 0x7810, 0x4910, 0xA0, 0x40, 0x20, 0x20}},
    {21605, 2, {0x40, 0x40, 0x7840, 0x4BF8, 0x4A48, 0x4A48, 0x4BF8, 0x4A48, 0x4A48, 0x4A48, 0x4FFE, 0x7A08, 0x4A08, 0x208, 0x228, 0x210}},
    {21606, 2, {0x20, 0x120, 0x7120, 0x5220, 0x52FC, 0x54A4, 0x5724, 0x5124, 0x5224, 0x5224, 0x74A4, 0x47A4, 0xC4, 0x44, 0x94, 0x108}},
    {21607, 2, {0x8, 0x3C, 0x7BE0, 0x4A20, 0x4A20, 0x4A20, 0x4A20, 0x4BFE, 0x4A20, 0x4A10, 0x7A10, 0x4A12, 0x20A, 0x28A, 0x326, 0x212}},
    {21608, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x4008, 0x4028, 0x8010}},
    {21609, 2, {0x0, 0x1FC, 0x7800, 0x4800, 0x4800, 0x4BFE, 0x4820, 0x4820, 0x4928, 0x4924, 0x7A24, 0x4A22, 0x422, 0x20, 0xA0, 0x40}},
    {21610, 2, {0x0, 0x3FC, 0x7A04, 0x4A04, 0x4A04, 0x4A04, 0x4BFC, 0x4890, 0x4890, 0x4890, 0x7890, 0x4912, 0x112, 0x212, 0x40E, 0x800}},
    {21611, 2, {0x20, 0x20, 0x7820, 0x4820, 0x483E, 0x4820, 0x4820, 0x4820, 0x49FC, 0x4904, 0x4904, 0x7904, 0x4904, 0x104, 0x1FC, 0x104}},
    {21612, 2, {0x0, 0x0, 0x7BFC, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x4A9C, 0x7B04, 0x4A04, 0x204, 0x3FC, 0x204}},
    {21613, 2, {0x48, 0x248, 0x7A48, 0x4A48, 0x4A48, 0x4FFE, 0x4A48, 0x4A48, 0x4A48, 0x4A48, 0x7A78, 0x4A00, 0x200, 0x200, 0x3FE, 0x0}},
    {21614, 2, {0x0, 0x0, 0x79FC, 0x4904, 0x4904, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4904, 0x7800, 0x4890, 0x88, 0x104, 0x202, 0x402}},
    {21615, 2, {0x0, 0x1FC, 0x7820, 0x4820, 0x4924, 0x48A4, 0x48A8, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21616, 2, {0x440, 0x440, 0x2444, 0x2758, 0x2460, 0x2442, 0x2F42, 0xF03E, 0x4000, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21617, 2, {0x8, 0x3C, 0x7BD0, 0x4A90, 0x4A90, 0x4A90, 0x4A90, 0x4A90, 0x4A90, 0x4A88, 0x7A88, 0x4AC8, 0x2A4, 0x4D4, 0x492, 0x800}},
    {21618, 2, {0x90, 0x90, 0x7890, 0x4A90, 0x4A92, 0x4AD4, 0x4A98, 0x4A90, 0x4A90, 0x4A90, 0x7A90, 0x4A92, 0x3D2, 0xE12, 0x40E, 0x0}},
    {21619, 2, {0x40, 0x40, 0x7840, 0x4BFC, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x48E0, 0x4950, 0x7950, 0x4A48, 0x444, 0x842, 0x40, 0x40}},
    {21620, 2, {0x80, 0x80, 0x7900, 0x49FC, 0x4A04, 0x4C04, 0x49E4, 0x4924, 0x4924, 0x4924, 0x79E4, 0x4924, 0x4, 0x4, 0x28, 0x10}},
    {21621, 2, {0x0, 0x0, 0x7BFE, 0x4808, 0x4808, 0x49E8, 0x4928, 0x4928, 0x4928, 0x4928, 0x79E8, 0x4928, 0x8, 0x8, 0x28, 0x10}},
    {21622, 2, {0x200, 0x200, 0xE27C, 0xA224, 0xAFA4, 0xA4A4, 0xA4A4, 0xA4A8, 0xA4A8, 0xA928, 0xE510, 0xA210, 0x528, 0x948, 0x1084, 0x102}},
    {21623, 2, {0x0, 0x7FC, 0xF444, 0x9444, 0x9444, 0x97FC, 0x9444, 0x9444, 0x9444, 0x97FC, 0xF444, 0x9040, 0x40, 0x40, 0x40, 0x40}},
    {21624, 2, {0x0, 0x3FC, 0x7820, 0x4820, 0x4840, 0x4840, 0x48D0, 0x4948, 0x4A44, 0x4C44, 0x7840, 0x4840, 0x40, 0x0, 0x7FE, 0x0}},
    {21625, 2, {0x40, 0x40, 0x7A40, 0x4A40, 0x4BFC, 0x4A40, 0x4C40, 0x4840, 0x4FFE, 0x4840, 0x78A0, 0x48A0, 0x110, 0x208, 0x404, 0x802}},
    {21626, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x4908, 0x49F8, 0x4800, 0x4BFE, 0x4880, 0x4900, 0x79F8, 0x4808, 0x8, 0x8, 0x50, 0x20}},
    {21627, 2, {0x40, 0x40, 0x7840, 0x4BF8, 0x4A48, 0x4A48, 0x4A48, 0x4BF8, 0x4A48, 0x4A48, 0x7A48, 0x4BF8, 0x248, 0x40, 0x40, 0x40}},
    {21628, 2, {0x8, 0x3C, 0x7BE0, 0x4820, 0x4924, 0x48A4, 0x48A8, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0xA0, 0x40}},
    {21629, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x37D8, 0xC006, 0x0, 0x3EF8, 0x2288, 0x2288, 0x2288, 0x3EA8, 0x2290, 0x80, 0x80}},
    {21630, 2, {0x0, 0x0, 0x79FC, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x4904, 0x79FC, 0x4904, 0x0, 0x0, 0x3FE, 0x0}},
    {21631, 2, {0x20, 0x20, 0x7820, 0x49FC, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x4840, 0x7840, 0x4888, 0x104, 0x3FE, 0x102, 0x0}},
    {21632, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x4908, 0x49F8, 0x4908, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x108, 0x108, 0x7FE, 0x0}},
    {21633, 2, {0x108, 0x108, 0x7908, 0x4908, 0x4FFE, 0x4908, 0x4908, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x108, 0x108, 0x1F8, 0x108}},
    {21634, 2, {0x0, 0x3FE, 0x7A10, 0x4A10, 0x4A10, 0x4AFE, 0x4A92, 0x4A92, 0x4A92, 0x4A92, 0x7A9A, 0x4A94, 0x210, 0x210, 0x3FE, 0x0}},
    {21635, 2, {0x120, 0x120, 0x7920, 0x4AA0, 0x4AAC, 0x4EB4, 0x4AE4, 0x4BA4, 0x4AA4, 0x4AB4, 0x7AA8, 0x4AA2, 0x2A2, 0x282, 0x27E, 0x200}},
    {21636, 2, {0x20, 0x20, 0x7924, 0x4924, 0x4924, 0x4924, 0x49FC, 0x4824, 0x4820, 0x4820, 0x7A22, 0x4A22, 0x222, 0x222, 0x3FE, 0x2}},
    {21637, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21638, 2, {0x80, 0x80, 0x79FC, 0x4904, 0x4A04, 0x4DF4, 0x4914, 0x4914, 0x4914, 0x49F4, 0x7904, 0x4928, 0x112, 0x102, 0xFE, 0x0}},
    {21639, 2, {0x0, 0x40, 0x7820, 0x4828, 0x4808, 0x4888, 0x4890, 0x4A94, 0x4AA2, 0x4AA2, 0x7CC2, 0x4888, 0x188, 0x288, 0x478, 0x0}},
    {21640, 2, {0x90, 0x90, 0x7890, 0x4BFC, 0x4894, 0x4894, 0x4BFC, 0x4A90, 0x4A90, 0x4BFE, 0x7892, 0x4892, 0x11A, 0x114, 0x210, 0x410}},
    {21641, 2, {0x40, 0x40, 0x7840, 0x4BF8, 0x4A48, 0x4A48, 0x4A48, 0x4A48, 0x4A48, 0x4FFE, 0x7840, 0x48A0, 0xA0, 0x110, 0x208, 0x406}},
    {21642, 2, {0x38, 0x3C0, 0x7840, 0x4840, 0x4840, 0x4FFC, 0x4840, 0x48E0, 0x48E0, 0x4950, 0x7950, 0x4A48, 0x444, 0x842, 0x40, 0x40}},
    {21643, 2, {0x100, 0x100, 0x7900, 0x49FE, 0x4A80, 0x4A80, 0x4C80, 0x48F8, 0x4880, 0x4880, 0x7880, 0x48FC, 0x80, 0x80, 0x80, 0x80}},
    {21644, 2, {0x400, 0xE00, 0x7800, 0x87C, 0x844, 0xFF44, 0x844, 0x1844, 0x1C44, 0x2A44, 0x2A44, 0x4844, 0x887C, 0x844, 0x800, 0x800}},
    {21645, 2, {0x20, 0x20, 0x7840, 0x4888, 0x4904, 0x4BFE, 0x4802, 0x4800, 0x49FC, 0x4904, 0x7904, 0x4904, 0x104, 0x104, 0x1FC, 0x104}},
    {21646, 2, {0x1040, 0x1040, 0x1E50, 0x2248, 0x2244, 0x5444, 0x840, 0x1440, 0x23FE, 0xC000, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21647, 2, {0x40, 0x20, 0x7810, 0x49E0, 0x4822, 0x4834, 0x4BB8, 0x48B0, 0x48A8, 0x48A8, 0x7928, 0x4924, 0x224, 0x422, 0xA0, 0x40}},
    {21648, 2, {0x88, 0x88, 0x7888, 0x4908, 0x497E, 0x4B08, 0x4D08, 0x4948, 0x4928, 0x4928, 0x7908, 0x4908, 0x108, 0x108, 0x128, 0x110}},
    {21649, 2, {0x100, 0x100, 0x793E, 0x4908, 0x4FC8, 0x4908, 0x4908, 0x4948, 0x4988, 0x4F08, 0x7908, 0x4908, 0x108, 0x108, 0x528, 0x210}},
    {21650, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7CF8, 0x0, 0x0, 0xFC0, 0x840, 0x840, 0x840, 0x1040, 0x1042, 0x2042, 0xC03E}},
    {21651, 2, {0x0, 0x0, 0x7BFC, 0x4900, 0x4900, 0x49F0, 0x4910, 0x4910, 0x4990, 0x4A50, 0x7A50, 0x4A10, 0x212, 0x292, 0x312, 0x20E}},
    {21652, 2, {0x40, 0x40, 0x7840, 0x487C, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x4840, 0x4840, 0x7850, 0x4848, 0x44, 0x40, 0x40, 0x40}},
    {21653, 2, {0x20, 0x20, 0x7820, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x49FC, 0x4904, 0x7904, 0x4904, 0x104, 0x104, 0x1FC, 0x104}},
    {21654, 2, {0x400, 0x400, 0xF400, 0x943C, 0x9FA4, 0x94A4, 0x94A4, 0x94A4, 0x94A4, 0x94A4, 0xF4A4, 0x94A4, 0x8A4, 0x8BC, 0x12A4, 0x2100}},
    {21655, 2, {0x40, 0x40, 0x7840, 0x4840, 0x4BFE, 0x4840, 0x4880, 0x4880, 0x48FE, 0x4910, 0x7910, 0x4A10, 0x210, 0x410, 0x1FE, 0x0}},
    {21656, 2, {0x40, 0x40, 0x7840, 0x4BFE, 0x4880, 0x48A0, 0x4920, 0x49FC, 0x4B24, 0x4D24, 0x7924, 0x4924, 0x134, 0x128, 0x20, 0x20}},
    {21657, 2, {0x90, 0x88, 0x7888, 0x4880, 0x4BFE, 0x48A0, 0x48A0, 0x48A4, 0x48A4, 0x4928, 0x7928, 0x4932, 0x222, 0x262, 0x49E, 0x800}},
    {21658, 2, {0x80, 0x80, 0x78F8, 0x4908, 0x4B10, 0x4CA0, 0x4840, 0x48A0, 0x4B18, 0x4C06, 0x78C0, 0x4820, 0x10, 0x180, 0x60, 0x10}},
    {21659, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x4C04, 0x4800, 0x4800, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20, 0xA0, 0x40}},
    {21660, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x4C04, 0x4900, 0x4908, 0x4910, 0x4920, 0x79C0, 0x4902, 0x102, 0x102, 0xFE, 0x0}},
    {21661, 2, {0x108, 0x108, 0x7108, 0x5210, 0x5294, 0x54A4, 0x57BC, 0x5108, 0x5108, 0x5210, 0x7420, 0x57BC, 0x0, 0x0, 0xFFE, 0x0}},
    {21662, 2, {0x0, 0x3FC, 0x7890, 0x4890, 0x4890, 0x4890, 0x4890, 0x4FFE, 0x4890, 0x4890, 0x7890, 0x4890, 0x110, 0x110, 0x210, 0x410}},
    {21663, 2, {0x0, 0x3FE, 0x7820, 0x4820, 0x4840, 0x49FC, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x4904, 0x104, 0x1FC, 0x104}},
    {21664, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x103E, 0xFFD0, 0x10, 0x10}},
    {21665, 2, {0x0, 0x3FE, 0x7908, 0x4908, 0x4908, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x4908, 0x7908, 0x491E, 0x7E8, 0x8, 0x8, 0x8}},
    {21666, 2, {0x0, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x400, 0x800, 0xFF0, 0x10, 0x10, 0xA0, 0x40}},
    {21667, 2, {0x40, 0x44, 0x7A44, 0x4948, 0x4950, 0x4840, 0x4FFE, 0x4890, 0x4890, 0x4890, 0x7890, 0x4892, 0x112, 0x112, 0x20E, 0x400}},
    {21668, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x4C04, 0x4838, 0x4BC0, 0x4840, 0x4840, 0x787E, 0x4FC0, 0x40, 0x42, 0x42, 0x3E}},
    {21669, 2, {0x0, 0x3FE, 0x7820, 0x4840, 0x4888, 0x4904, 0x4BFE, 0x4822, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x7FE, 0x0}},
    {21670, 2, {0x40, 0x40, 0xF7FE, 0x9040, 0x97FC, 0x9044, 0x93FC, 0x9240, 0x9440, 0x97FE, 0xF042, 0x90AA, 0xA4, 0x110, 0x208, 0xC06}},
    {21671, 2, {0x4, 0x4, 0xEFC4, 0xA214, 0xA214, 0xA3D4, 0xA254, 0xA454, 0xA654, 0xA554, 0xE894, 0xA094, 0x104, 0x204, 0x414, 0x808}},
    {21672, 2, {0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21673, 2, {0x104, 0x84, 0x7888, 0x4800, 0x4BFE, 0x4820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21674, 2, {0x40, 0x40, 0x7A48, 0x4948, 0x4950, 0x4840, 0x4FFE, 0x4840, 0x48E0, 0x4950, 0x7950, 0x4A48, 0x444, 0x842, 0x40, 0x40}},
    {21675, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x2244, 0x2244, 0x3E7C, 0x2A00, 0x2828, 0x2824, 0x2844, 0x2442, 0x2482, 0x4200, 0x4180, 0x807E}},
    {21676, 2, {0x80, 0x40, 0xF040, 0x97FC, 0x9000, 0x9110, 0x9208, 0x9404, 0x9110, 0x9110, 0xF0A0, 0x9040, 0xA0, 0x110, 0x208, 0xC06}},
    {21677, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x49FC, 0x4800, 0x4800, 0x49FC, 0x7904, 0x4904, 0x104, 0x104, 0x1FC, 0x104}},
    {21678, 2, {0x20, 0x120, 0x7920, 0x49FC, 0x4920, 0x4A20, 0x4820, 0x4BFE, 0x4870, 0x48A8, 0x78A8, 0x4924, 0x124, 0x222, 0x20, 0x20}},
    {21679, 2, {0x80, 0x80, 0x78F8, 0x4908, 0x4B10, 0x4CA0, 0x4840, 0x48A0, 0x4918, 0x4A06, 0x7DF8, 0x4908, 0x108, 0x108, 0x1F8, 0x108}},
    {21680, 2, {0x80, 0x80, 0x79FC, 0x4904, 0x4A04, 0x49E4, 0x4924, 0x4924, 0x49E4, 0x4924, 0x7924, 0x49E4, 0x104, 0x4, 0x28, 0x10}},
    {21681, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x4904, 0x49FC, 0x7904, 0x4904, 0x104, 0x1FC, 0x104}},
    {21682, 2, {0x208, 0x108, 0x7910, 0x4800, 0x4BFC, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x4840, 0x7840, 0x48A0, 0xA0, 0x110, 0x208, 0x406}},
    {21683, 2, {0x40, 0x20, 0xF020, 0x97FE, 0x9040, 0x9084, 0x9108, 0x93F0, 0x9022, 0x9044, 0xF188, 0x9610, 0x20, 0x50, 0x188, 0x604}},
    {21684, 2, {0x40, 0x40, 0x7840, 0x4BFE, 0x4880, 0x4890, 0x4890, 0x4912, 0x4952, 0x4954, 0x7A90, 0x4A28, 0x428, 0x44, 0x84, 0x102}},
    {21685, 2, {0x20, 0x20, 0x7BFE, 0x4850, 0x4888, 0x4904, 0x4AFA, 0x4800, 0x4BFE, 0x4840, 0x7880, 0x49FC, 0x4, 0x4, 0x28, 0x10}},
    {21686, 2, {0x8, 0x3C, 0x79E0, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x49FC, 0x7904, 0x4904, 0x104, 0x104, 0x1FC, 0x104}},
    {21687, 2, {0x90, 0x90, 0x7890, 0x4A92, 0x4994, 0x4898, 0x4890, 0x4998, 0x4A94, 0x4C92, 0x7890, 0x4890, 0x112, 0x112, 0x212, 0x40E}},
    {21688, 2, {0x50, 0x48, 0x40, 0x3FFE, 0x2040, 0x2040, 0x2F44, 0x2044, 0x2044, 0x2F28, 0x2928, 0x2912, 0x2F32, 0x494A, 0x4086, 0x8102}},
    {21689, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x4C44, 0x4840, 0x4BFE, 0x4888, 0x4888, 0x7908, 0x48D0, 0x20, 0x50, 0x88, 0x304}},
    {21690, 2, {0x0, 0x3FE, 0x7800, 0x4800, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x79FC, 0x4904, 0x0, 0x0, 0x3FE, 0x0}},
    {21691, 2, {0x90, 0x90, 0x7890, 0x4910, 0x49FE, 0x4B10, 0x4D10, 0x4938, 0x4938, 0x4954, 0x7954, 0x4992, 0x110, 0x110, 0x110, 0x110}},
    {21692, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1F90, 0x1090, 0x1090, 0x7FFC, 0x4004, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4004, 0x400C}},
    {21693, 2, {0x0, 0x7FC, 0xF404, 0x9444, 0x9444, 0x9444, 0x97FC, 0x9444, 0x9444, 0x94A4, 0xF494, 0x9514, 0x604, 0x404, 0x7FC, 0x404}},
    {21694, 2, {0x40, 0x40, 0x7844, 0x4BF4, 0x4848, 0x4850, 0x4FFE, 0x4840, 0x4880, 0x4984, 0x4A98, 0x7CE0, 0x4882, 0x82, 0x7E, 0x0}},
    {21695, 2, {0x100, 0x17C, 0xF124, 0x9224, 0x9224, 0x96FE, 0x9A24, 0x9224, 0x9224, 0x927C, 0xF224, 0x9220, 0x220, 0x220, 0x240, 0x280}},
    {21696, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x280, 0x444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {21697, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7C7C, 0x4444, 0x4444, 0x4444, 0x4444, 0x7C7C, 0x4444}},
    {21698, 2, {0x0, 0x7FE, 0x7890, 0x4890, 0x4890, 0x4BFC, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x7A9C, 0x4B04, 0x204, 0x204, 0x3FC, 0x204}},
    {21699, 2, {0x0, 0x3FC, 0x7A04, 0x4A04, 0x4AF4, 0x4A04, 0x4A04, 0x4AF4, 0x4A94, 0x4A94, 0x7A94, 0x4AF4, 0x204, 0x204, 0x214, 0x208}},
    {21700, 2, {0x110, 0x110, 0x7910, 0x4910, 0x4BFC, 0x4910, 0x4910, 0x4910, 0x4910, 0x4FFE, 0x7800, 0x4910, 0x108, 0x208, 0x404, 0x804}},
    {21701, 2, {0x100, 0x100, 0x7900, 0x49FE, 0x4A02, 0x4A02, 0x4D12, 0x48A2, 0x4A4A, 0x4AAA, 0x7B1A, 0x4A0A, 0x3FA, 0x2, 0x14, 0x8}},
    {21702, 2, {0x80, 0x80, 0xF1F8, 0x9210, 0x9520, 0x90C0, 0x9090, 0x9320, 0x9C7C, 0x9084, 0xF108, 0x9690, 0x60, 0x40, 0x180, 0xE00}},
    {21703, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x4800, 0x4820, 0x4820, 0x79FC, 0x4820, 0x20, 0x20, 0x3FE, 0x0}},
    {21704, 2, {0x40, 0x40, 0xA0, 0xF8A0, 0x8910, 0x8A08, 0x8C06, 0x89F4, 0x8800, 0x8800, 0x8BF8, 0xFA08, 0x8A08, 0x208, 0x3F8, 0x208}},
    {21705, 2, {0x840, 0x848, 0x844, 0x7F44, 0x840, 0x840, 0xFFFE, 0x40, 0x44, 0x3E44, 0x2228, 0x222A, 0x2212, 0x3E2A, 0x2246, 0x82}},
    {21706, 2, {0x40, 0x40, 0xF7FE, 0x9080, 0x9100, 0x91FC, 0x9304, 0x9504, 0x99FC, 0x9104, 0xF104, 0x91FC, 0x104, 0x104, 0x114, 0x108}},
    {21707, 2, {0x210, 0x250, 0x7250, 0x525C, 0x5F74, 0x52D4, 0x5254, 0x5254, 0x5254, 0x525C, 0x7250, 0x5352, 0xE42, 0x442, 0x3E, 0x0}},
    {21708, 2, {0x8, 0x3C, 0xF7C0, 0x9400, 0x941C, 0x95E0, 0x9520, 0x9522, 0x9524, 0x9528, 0xF510, 0x9510, 0x508, 0x944, 0x982, 0x1100}},
    {21709, 2, {0x40, 0x40, 0x7880, 0x4BFC, 0x4A04, 0x4A04, 0x4AF4, 0x4A94, 0x4A94, 0x4A94, 0x7A94, 0x4AF4, 0x204, 0x204, 0x214, 0x208}},
    {21710, 2, {0x110, 0x110, 0x7BFC, 0x4910, 0x4910, 0x4800, 0x4A08, 0x4A08, 0x4910, 0x4910, 0x78A0, 0x4840, 0xA0, 0x110, 0x208, 0xC06}},
    {21711, 2, {0x0, 0x3F8, 0x7A08, 0x4A08, 0x4BF8, 0x4A08, 0x4A08, 0x4BF8, 0x4A44, 0x4A48, 0x7A30, 0x4A20, 0x210, 0x288, 0x306, 0x200}},
    {21712, 2, {0x0, 0x3FE, 0x7A00, 0x4A00, 0x4BFC, 0x4A20, 0x4A20, 0x4AF8, 0x4A20, 0x4A20, 0x7A20, 0x4BFC, 0x200, 0x200, 0x3FE, 0x0}},
    {21713, 2, {0x0, 0x3FC, 0x7890, 0x4890, 0x4890, 0x4892, 0x4C92, 0x4A94, 0x4A94, 0x4A98, 0x7890, 0x4890, 0x90, 0x90, 0x7FE, 0x0}},
    {21714, 2, {0x10, 0x210, 0x7910, 0x4910, 0x48FE, 0x4810, 0x4B10, 0x4928, 0x4924, 0x4944, 0x7942, 0x4982, 0x100, 0x280, 0x47E, 0x0}},
    {21715, 2, {0x80, 0x80, 0x78BC, 0x4BC0, 0x4850, 0x4824, 0x48D4, 0x4B0C, 0x4800, 0x4BFE, 0x7890, 0x4890, 0x112, 0x112, 0x20E, 0x400}},
    {21716, 2, {0x110, 0x112, 0x7914, 0x49D8, 0x4910, 0x4912, 0x4952, 0x498E, 0x4920, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21717, 2, {0x20, 0x124, 0x7924, 0x4924, 0x49FC, 0x4840, 0x4840, 0x48FC, 0x4904, 0x4A84, 0x7848, 0x4850, 0x20, 0x40, 0x80, 0x300}},
    {21718, 2, {0x80, 0x80, 0x78FE, 0x4910, 0x4A10, 0x4810, 0x48FC, 0x4890, 0x4890, 0x4890, 0x7BFE, 0x4810, 0x10, 0x10, 0x10, 0x10}},
    {21719, 2, {0x90, 0x94, 0x7914, 0x4918, 0x4B10, 0x4D32, 0x4952, 0x490E, 0x4920, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21720, 2, {0x80, 0x80, 0x793C, 0x4A00, 0x4C80, 0x4880, 0x497E, 0x4B08, 0x4D08, 0x4908, 0x7908, 0x4908, 0x108, 0x108, 0x128, 0x110}},
    {21721, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A08, 0x4C06, 0x49F0, 0x4800, 0x4800, 0x4FFC, 0x7840, 0x4880, 0x110, 0x208, 0x7FC, 0x204}},
    {21722, 2, {0x0, 0x1F0, 0x7910, 0x4910, 0x4910, 0x4910, 0x4A4E, 0x4C40, 0x4BFC, 0x4840, 0x78E0, 0x4950, 0x248, 0x446, 0x40, 0x40}},
    {21723, 2, {0x80, 0x7C88, 0x44B0, 0x44C4, 0x4484, 0x7C7C, 0x0, 0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040}},
    {21724, 2, {0x40, 0x20, 0x7BFE, 0x4904, 0x4888, 0x4850, 0x4820, 0x48D8, 0x4B06, 0x4888, 0x7888, 0x4888, 0x88, 0x108, 0x108, 0x208}},
    {21725, 2, {0x40, 0x40, 0xF040, 0x97FC, 0x9484, 0x9888, 0x9140, 0x9144, 0x9348, 0x9530, 0xF920, 0x9110, 0x108, 0x144, 0x182, 0x100}},
    {21726, 2, {0x40, 0x40, 0x7890, 0x4908, 0x4BFC, 0x4824, 0x4920, 0x4920, 0x49FC, 0x4A20, 0x7820, 0x4FFE, 0x20, 0x20, 0x20, 0x20}},
    {21727, 2, {0x220, 0x220, 0xF220, 0x943C, 0x9544, 0x9F44, 0x9284, 0x9224, 0x9414, 0x9F94, 0xF404, 0x9004, 0x184, 0xE04, 0x428, 0x10}},
    {21728, 2, {0x20, 0x120, 0x7920, 0x49FE, 0x4A20, 0x4C20, 0x4820, 0x4BFE, 0x4800, 0x4800, 0x79FC, 0x4904, 0x104, 0x104, 0x1FC, 0x104}},
    {21729, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {21730, 2, {0x0, 0x3FE, 0x7820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x4888, 0x4888, 0x7BFE, 0x4888, 0x88, 0x108, 0x108, 0x208}},
    {21731, 2, {0x0, 0x3FC, 0x7800, 0x4800, 0x49F8, 0x4908, 0x4908, 0x4908, 0x49F8, 0x4800, 0x7908, 0x4888, 0x90, 0x0, 0x7FE, 0x0}},
    {21732, 2, {0x110, 0x108, 0x7900, 0x4FFE, 0x4948, 0x4950, 0x4964, 0x4948, 0x4950, 0x4A64, 0x7A48, 0x4A52, 0x462, 0x442, 0x9BE, 0x1000}},
    {21733, 2, {0x0, 0x7FFC, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x0, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x50, 0x20}},
    {21734, 2, {0x28, 0x1A4, 0xF724, 0x9120, 0x9120, 0x9FFE, 0x9120, 0x9124, 0x91A4, 0x9328, 0xFD28, 0x9110, 0x112, 0x12A, 0x546, 0x282}},
    {21735, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x4848, 0x4848, 0x494C, 0x794A, 0x4A4A, 0x88, 0x88, 0x128, 0x210}},
    {21736, 2, {0x20, 0x124, 0x78A4, 0x48A8, 0x4820, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104, 0x104, 0x114, 0x108}},
    {21737, 2, {0x0, 0x7FC, 0xF444, 0x9444, 0x97FC, 0x9444, 0x9444, 0x97FC, 0x9040, 0x9040, 0xF7FC, 0x9040, 0x40, 0x40, 0xFFE, 0x0}},
    {21738, 2, {0x0, 0xFDE, 0x7252, 0x5252, 0x5254, 0x57D4, 0x5258, 0x5254, 0x5252, 0x5FD2, 0x7252, 0x525A, 0x254, 0x450, 0x550, 0x890}},
    {21739, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x4904, 0x49FC, 0x4820, 0x4820, 0x4920, 0x493C, 0x7920, 0x4920, 0x2A0, 0x260, 0x43E, 0x800}},
    {21740, 2, {0x100, 0x100, 0x79FE, 0x4A04, 0x4A04, 0x4EF4, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x7AF4, 0x4A94, 0x204, 0x204, 0x214, 0x208}},
    {21741, 2, {0x0, 0x3C78, 0x2448, 0x2448, 0x2448, 0x3C78, 0x0, 0x110, 0x108, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {21742, 2, {0x40, 0x44, 0x7BF4, 0x4848, 0x4848, 0x4FFE, 0x4820, 0x4840, 0x4BF8, 0x4910, 0x7A20, 0x4DFE, 0x820, 0x20, 0xA0, 0x40}},
    {21743, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x79FC, 0x4850, 0x90, 0x92, 0x112, 0x20E}},
    {21744, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C24, 0x4920, 0x49FC, 0x4920, 0x4A20, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21745, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x4820, 0x4BFE, 0x4A02, 0x4C04, 0x49F8, 0x4810, 0x7820, 0x4BFE, 0x20, 0x20, 0xA0, 0x40}},
    {21746, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21747, 2, {0x208, 0x21C, 0x7260, 0x5240, 0x5FC0, 0x5240, 0x527E, 0x52C8, 0x5348, 0x5E48, 0x7248, 0x5248, 0x248, 0x248, 0xA88, 0x508}},
    {21748, 2, {0x80, 0x40, 0x7BF8, 0x4A08, 0x4A08, 0x4BF8, 0x4A08, 0x4A08, 0x4BF8, 0x4A40, 0x7A44, 0x4A28, 0x210, 0x288, 0x306, 0x200}},
    {21749, 2, {0x2, 0x3E2, 0x7A22, 0x4A2A, 0x4A2A, 0x4BEA, 0x490A, 0x490A, 0x4BEA, 0x492A, 0x792A, 0x492A, 0x222, 0x222, 0x4AA, 0x844}},
    {21750, 2, {0x88, 0x88, 0x7888, 0x4B8E, 0x4888, 0x4888, 0x4800, 0x49FC, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x20, 0x20}},
    {21751, 2, {0x8, 0x3C, 0x7BC0, 0x4844, 0x4A24, 0x4928, 0x4900, 0x4810, 0x4BFE, 0x4810, 0x7910, 0x4890, 0x90, 0x10, 0x50, 0x20}},
    {21752, 2, {0x8, 0x3C, 0x7BC0, 0x4844, 0x4A24, 0x4928, 0x4900, 0x4840, 0x4BFE, 0x4888, 0x7908, 0x4B90, 0x60, 0x50, 0x188, 0x604}},
    {21753, 2, {0x8, 0x3C, 0x7BC0, 0x4804, 0x4A44, 0x4928, 0x4800, 0x4BF8, 0x4810, 0x4820, 0x4FFE, 0x7820, 0x4820, 0x20, 0xA0, 0x40}},
    {21754, 2, {0x48, 0x44, 0xF7FE, 0x9040, 0x9040, 0x93FC, 0x9244, 0x9244, 0x93FC, 0x9244, 0xF244, 0x93FC, 0x244, 0x244, 0x254, 0x208}},
    {21755, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4800, 0x4800, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x20, 0x20, 0x20, 0x20}},
    {21756, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x4800, 0x4BFC, 0x7808, 0x4830, 0x20, 0x20, 0xA0, 0x40}},
    {21757, 2, {0x0, 0x7FE, 0xF040, 0x97FC, 0x9444, 0x9444, 0x97FC, 0x9444, 0x9444, 0x97FC, 0xF240, 0x9140, 0x80, 0x140, 0x230, 0xC0E}},
    {21758, 2, {0x104, 0x84, 0x7888, 0x4810, 0x49FC, 0x4904, 0x4904, 0x4904, 0x49FC, 0x4850, 0x7850, 0x4890, 0x92, 0x112, 0x20E, 0x400}},
    {21759, 2, {0x800, 0x800, 0x7F7C, 0x1144, 0x1144, 0x2544, 0x427C, 0x0, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x50, 0x20}},
    {21760, 2, {0x1C, 0x1E0, 0x7820, 0x4820, 0x4BFE, 0x48A8, 0x4924, 0x4A22, 0x49F8, 0x4888, 0x7890, 0x48BE, 0x102, 0x102, 0x214, 0x408}},
    {21761, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x4800, 0x49FC, 0x4800, 0x4800, 0x49FC, 0x4800, 0x7800, 0x49FC, 0x104, 0x104, 0x1FC, 0x104}},
    {21762, 2, {0x110, 0x108, 0xF204, 0x9444, 0x9040, 0x90A0, 0x9110, 0x9208, 0x9C06, 0x93F8, 0xF208, 0x9208, 0x208, 0x208, 0x3F8, 0x208}},
    {21763, 2, {0x100, 0x100, 0x7BF8, 0x4A08, 0x4C10, 0x4BFC, 0x4A44, 0x4A44, 0x4BFC, 0x4A44, 0x7A44, 0x4BFC, 0x244, 0x444, 0x454, 0x808}},
    {21764, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x7800, 0x4888, 0x84, 0x102, 0x202}},
    {21765, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A88, 0x4C46, 0x49F0, 0x4810, 0x4820, 0x4840, 0x7BF8, 0x4A08, 0x208, 0x208, 0x3F8, 0x208}},
    {21766, 2, {0x40, 0x40, 0x7888, 0x4904, 0x4BFE, 0x4802, 0x4888, 0x4944, 0x4A42, 0x48F8, 0x7988, 0x4A50, 0x20, 0x50, 0x188, 0x606}},
    {21767, 2, {0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2488, 0x2450, 0x2530, 0x260E, 0x2000, 0x2FF8, 0x4808, 0x4808, 0x8FF8, 0x808}},
    {21768, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x4800, 0x4BFC, 0x4A44, 0x4A44, 0x4A44, 0x7BFC, 0x4A00, 0x202, 0x202, 0x1FE, 0x0}},
    {21769, 2, {0x40, 0x80, 0xF110, 0x9208, 0x97FC, 0x9104, 0x9100, 0x93F8, 0x9440, 0x9040, 0x9FFE, 0xF040, 0x90A0, 0x110, 0x208, 0xC06}},
    {21770, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x4924, 0x4924, 0x4924, 0x4AAA, 0x7820, 0x4850, 0x50, 0x88, 0x104, 0x202}},
    {21771, 2, {0x20, 0x20, 0x7850, 0x4888, 0x4904, 0x4AFA, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x7928, 0x4924, 0x222, 0x422, 0xA0, 0x40}},
    {21772, 2, {0x4, 0xE, 0x7BB8, 0x4888, 0x4888, 0x4928, 0x492E, 0x4BA8, 0x48A8, 0x4AA8, 0x7AA8, 0x493E, 0x100, 0x280, 0x47E, 0x800}},
    {21773, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x49F8, 0x4800, 0x4800, 0x4BFE, 0x4890, 0x7890, 0x4890, 0x112, 0x112, 0x20E, 0x400}},
    {21774, 2, {0x42, 0xE2, 0x7B82, 0x488A, 0x488A, 0x488A, 0x4FEA, 0x488A, 0x498A, 0x49CA, 0x7AAA, 0x4A82, 0x482, 0x82, 0x8A, 0x84}},
    {21775, 2, {0x184, 0x68, 0x7830, 0x48C8, 0x4B04, 0x4840, 0x4BFE, 0x48A0, 0x4920, 0x4BFC, 0x7D24, 0x4924, 0x134, 0x128, 0x20, 0x20}},
    {21776, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2080, 0x2FF8, 0x2808, 0x4808, 0x4808, 0x8FF8, 0x808}},
    {21777, 2, {0x40, 0x40, 0xF248, 0x9248, 0x9248, 0x9554, 0x98E2, 0x9040, 0x9040, 0x97FC, 0xF040, 0x9040, 0x40, 0x40, 0xFFE, 0x0}},
    {21778, 2, {0x0, 0xFFE, 0xF0A0, 0x90A0, 0x97FC, 0x94A4, 0x94A4, 0x94A4, 0x951C, 0x9604, 0xF404, 0x97FC, 0x404, 0x404, 0x7FC, 0x404}},
    {21779, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x7820, 0x4BFE, 0x20, 0x20, 0x20}},
    {21780, 2, {0x0, 0x7FC, 0xF040, 0x9040, 0x93F8, 0x9088, 0x9088, 0x97FE, 0x9000, 0x9000, 0xF3F8, 0x9208, 0x208, 0x208, 0x3F8, 0x208}},
    {21781, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x20, 0x20, 0x20, 0x20}},
    {21782, 2, {0x0, 0x3FC, 0x7890, 0x4890, 0x4890, 0x4BFC, 0x4A94, 0x4A94, 0x4A94, 0x4A94, 0x7BFC, 0x4890, 0x90, 0x90, 0x7FE, 0x0}},
    {21783, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x4920, 0x793C, 0x4920, 0x2A0, 0x260, 0x43E, 0x800}},
    {21784, 2, {0x0, 0xFEFC, 0x1044, 0x2044, 0x3E28, 0x6210, 0xA228, 0x3E44, 0x2282, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21785, 2, {0x10, 0x890, 0x7510, 0x5212, 0x5552, 0x5954, 0x5158, 0x5190, 0x5310, 0x5528, 0x7928, 0x5128, 0x148, 0x144, 0xA84, 0x502}},
    {21786, 2, {0x0, 0x1FC, 0x7804, 0x48FC, 0x4804, 0x49FC, 0x4800, 0x4BFE, 0x4A02, 0x49F8, 0x7888, 0x4888, 0x50, 0x20, 0xD8, 0x306}},
    {21787, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9040, 0x97FE, 0x9080, 0x90F8, 0x9108, 0xF310, 0x94A0, 0x40, 0xA0, 0x318, 0xC06}},
    {21788, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {21789, 2, {0x0, 0x1FC, 0x7820, 0x4820, 0x4BFE, 0x4800, 0x4800, 0x49FC, 0x4904, 0x4924, 0x7924, 0x4924, 0x154, 0x88, 0x104, 0x602}},
    {21790, 2, {0x104, 0x124, 0x7914, 0x4914, 0x4FC4, 0x4924, 0x4914, 0x4954, 0x4986, 0x4F3C, 0x7904, 0x4904, 0x104, 0x104, 0x504, 0x204}},
    {21791, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x400, 0x820, 0x1FF0, 0x10, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {21792, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4800, 0x4BFE, 0x4A02, 0x4C44, 0x4840, 0x4BFC, 0x7844, 0x4884, 0x84, 0x104, 0x228, 0x410}},
    {21793, 2, {0x0, 0x3FE, 0x7850, 0x4850, 0x4850, 0x4BFE, 0x4A52, 0x4A52, 0x4A52, 0x4AAA, 0x7AA6, 0x4B02, 0x202, 0x202, 0x20A, 0x204}},
    {21794, 2, {0x20, 0x124, 0x78A4, 0x48A8, 0x4820, 0x49FC, 0x4904, 0x4924, 0x4924, 0x4924, 0x7924, 0x4924, 0x50, 0x88, 0x104, 0x202}},
    {21795, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4840, 0x7840, 0x487E, 0x7C0, 0x42, 0x42, 0x3E}},
    {21796, 2, {0x100, 0x100, 0xF1F0, 0x9210, 0x9420, 0x9BF8, 0x9248, 0x9248, 0x9248, 0x9248, 0xF7FE, 0x90A0, 0xA0, 0x110, 0x208, 0xC06}},
    {21797, 2, {0x10, 0x410, 0x7A28, 0x4A28, 0x4844, 0x4992, 0x4908, 0x4908, 0x4AFE, 0x4E02, 0x7A04, 0x4A44, 0x228, 0x210, 0x208, 0x8}},
    {21798, 2, {0x10, 0x210, 0x7910, 0x4910, 0x4C54, 0x4A52, 0x4A52, 0x4890, 0x4810, 0x4914, 0x7E04, 0x4A08, 0x208, 0x210, 0x220, 0xC0}},
    {21799, 2, {0x0, 0x7BC, 0x74A4, 0x54A4, 0x54A4, 0x57A4, 0x54A4, 0x54A4, 0x54A4, 0x57A4, 0x7434, 0x5528, 0x4A0, 0x560, 0x620, 0x20}},
    {21800, 2, {0x0, 0x7BC, 0x74A4, 0x5524, 0x5524, 0x563C, 0x5524, 0x54A4, 0x54A4, 0x54BC, 0x76A4, 0x5524, 0x424, 0x424, 0x47E, 0x400}},
    {21801, 2, {0x1C, 0x1E0, 0x7820, 0x4820, 0x4BFE, 0x48A8, 0x4924, 0x4A42, 0x4840, 0x4BFE, 0x7888, 0x4908, 0xD0, 0x30, 0x48, 0x184}},
    {21802, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9080, 0x97FC, 0x9110, 0x9248, 0x9446, 0xF1F0, 0x9040, 0x7FC, 0x40, 0x40, 0x40}},
    {21803, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A08, 0x4C06, 0x49F0, 0x4840, 0x4840, 0x4BF8, 0x7840, 0x4A48, 0x148, 0x150, 0xFFE, 0x0}},
    {21804, 2, {0x40, 0x40, 0x787C, 0x4840, 0x4BFC, 0x4A44, 0x4A70, 0x4BC0, 0x4A44, 0x4A3C, 0x7A00, 0x4AF0, 0x490, 0x492, 0x912, 0x120E}},
    {21805, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4888, 0x48F8, 0x4888, 0x4888, 0x48F8, 0x4888, 0x7888, 0x4BFE, 0x0, 0x88, 0x104, 0x202}},
    {21806, 2, {0x900, 0x880, 0xFFC, 0x1080, 0x3FF8, 0x5080, 0x9FF8, 0x1080, 0x1FFC, 0x0, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {21807, 2, {0xA0, 0x90, 0x7880, 0x49FE, 0x4910, 0x4B10, 0x4DFC, 0x4910, 0x4910, 0x49FC, 0x7910, 0x4910, 0x110, 0x1FE, 0x100, 0x100}},
    {21808, 2, {0x2, 0x7E2, 0x7422, 0x542A, 0x57EA, 0x548A, 0x548A, 0x548A, 0x57EA, 0x56AA, 0x76AA, 0x5AAA, 0xAA2, 0x12E2, 0x8A, 0x84}},
    {21809, 2, {0x1F8, 0x108, 0x7908, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x4800, 0x4BFC, 0x4A04, 0x7A04, 0x4BFC, 0x204, 0x204, 0x3FC, 0x204}},
    {21810, 2, {0x0, 0x40, 0x7B9C, 0x4A04, 0x4A04, 0x4B9C, 0x4A04, 0x4A04, 0x4BFC, 0x4890, 0x7890, 0x4890, 0x90, 0x112, 0x212, 0x40E}},
    {21811, 2, {0x40, 0x20, 0x7BFC, 0x4A04, 0x4A04, 0x4BFC, 0x4A00, 0x4A28, 0x4A24, 0x4BFE, 0x7A20, 0x4A50, 0x250, 0x488, 0x504, 0xA02}},
    {21812, 2, {0x208, 0x108, 0xF110, 0x97FC, 0x9040, 0x9040, 0x93F8, 0x9040, 0x9080, 0x9FFE, 0xF120, 0x9120, 0x220, 0x222, 0x422, 0x81E}},
    {21813, 2, {0x40, 0x80, 0x7BFC, 0x4910, 0x4A48, 0x4C46, 0x4BF8, 0x4A48, 0x4A48, 0x4BF8, 0x4A48, 0x7A48, 0x4BF8, 0x42, 0x42, 0x3E}},
    {21814, 2, {0x88, 0x88, 0x7888, 0x4BFE, 0x4888, 0x4888, 0x4FFE, 0x4800, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104, 0x104, 0x1FC, 0x104}},
    {21815, 2, {0x40, 0x20, 0x7BFE, 0x4880, 0x4904, 0x4BFE, 0x4802, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104, 0x104, 0x114, 0x108}},
    {21816, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A88, 0x4C46, 0x4800, 0x4BF0, 0x4810, 0x4820, 0x7840, 0x4924, 0x522, 0x50A, 0x908, 0xF8}},
    {21817, 2, {0x108, 0x88, 0x7808, 0x4FD4, 0x4A14, 0x4A22, 0x4A00, 0x4BC8, 0x4A44, 0x4A44, 0x7A40, 0x4A50, 0x248, 0x444, 0x544, 0x880}},
    {21818, 2, {0x90, 0x90, 0x7890, 0x4BFC, 0x4A94, 0x4A94, 0x4A94, 0x4BFC, 0x4A94, 0x4A94, 0x4A94, 0x7FFE, 0x4800, 0x90, 0x108, 0x204}},
    {21819, 2, {0x40, 0x40, 0x7840, 0x4FFC, 0x4840, 0x4A48, 0x4A48, 0x4A48, 0x4D54, 0x48E0, 0x7950, 0x4950, 0x248, 0x444, 0x842, 0x40}},
    {21820, 2, {0x80, 0x40, 0x7BFC, 0x4800, 0x4908, 0x4890, 0x4FFE, 0x4840, 0x4840, 0x4FFE, 0x7888, 0x4908, 0x90, 0x60, 0x198, 0x604}},
    {21821, 2, {0x100, 0x106, 0xF138, 0x9120, 0x97A0, 0x9120, 0x913E, 0x9324, 0x93A4, 0x9564, 0xF524, 0x9924, 0x124, 0x124, 0x124, 0x144}},
    {21822, 2, {0x10, 0x78, 0xF3C0, 0x9040, 0x9040, 0x97FC, 0x9248, 0x9248, 0x97FE, 0x9248, 0xF248, 0x97FC, 0x40, 0x40, 0x7FC, 0x0}},
    {21823, 2, {0x100, 0x100, 0xF3FC, 0x9494, 0x9894, 0x9124, 0x9244, 0x9484, 0x9128, 0x9210, 0xF040, 0x9124, 0x52A, 0x50A, 0x8F8, 0x0}},
    {21824, 2, {0x0, 0x3FE, 0x7A10, 0x4A10, 0x4AFE, 0x4A10, 0x4A10, 0x4BFE, 0x4A00, 0x4A10, 0x7A10, 0x4AFE, 0x210, 0x410, 0x5FE, 0x800}},
    {21825, 2, {0x0, 0x7FC, 0xF444, 0x9444, 0x95F4, 0x9444, 0x9444, 0x97FC, 0x9404, 0x95F4, 0xF514, 0x9514, 0x5F4, 0x404, 0x414, 0x808}},
    {21826, 2, {0xE8, 0xF08, 0xE228, 0xA928, 0xA548, 0xA008, 0xA7C8, 0xA088, 0xA108, 0xA108, 0xE1E8, 0xAF08, 0x10A, 0x10A, 0x50A, 0x206}},
    {21827, 2, {0x20, 0x120, 0x793C, 0x4920, 0x4920, 0x4FFE, 0x4800, 0x4BFC, 0x4A04, 0x4BFC, 0x7A04, 0x4BFC, 0x204, 0x204, 0x214, 0x208}},
    {21828, 2, {0x0, 0x3FE, 0x7840, 0x4880, 0x4944, 0x4A24, 0x4868, 0x4AB0, 0x4930, 0x4AA8, 0x7868, 0x48A4, 0x122, 0x220, 0xA0, 0x40}},
    {21829, 2, {0x20, 0x20, 0x783E, 0x4820, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4924, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20, 0x20}},
    {21830, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x3FF8, 0x2008, 0x2448, 0x2828, 0x37D8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {21831, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x3FF8, 0x2108, 0x2FE8, 0x2108, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {21832, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x4BFE, 0x4888, 0x4850, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x20, 0x20, 0x20, 0x20}},
    {21833, 2, {0x110, 0x110, 0x7110, 0x5110, 0x57BC, 0x5110, 0x5110, 0x5338, 0x53B8, 0x5554, 0x7554, 0x5992, 0x110, 0x110, 0x110, 0x110}},
    {21834, 2, {0x0, 0xEFC, 0xEA08, 0xAA08, 0xAAE8, 0xAAA8, 0xACA8, 0xAAA8, 0xAAA8, 0xAAA8, 0xEAE8, 0xAAA8, 0xC08, 0x808, 0x828, 0x810}},
    {21835, 2, {0x8, 0x3C, 0x7BC0, 0x4844, 0x4A24, 0x4928, 0x4900, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x48A8, 0x124, 0x222, 0x420, 0x20}},
    {21836, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x4C94, 0x4908, 0x4A04, 0x4800, 0x49FC, 0x7820, 0x4820, 0x20, 0x20, 0x7FE, 0x0}},
    {21837, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x4800, 0x49FC, 0x4808, 0x7810, 0x4BFE, 0x20, 0x20, 0xA0, 0x40}},
    {21838, 2, {0x2000, 0x21FE, 0x2020, 0x3E20, 0x49FC, 0x8844, 0x844, 0x9FE, 0xFE00, 0x800, 0x8FC, 0x884, 0x884, 0x884, 0x8FC, 0x884}},
    {21839, 2, {0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x4008, 0x4008, 0x4FC8, 0x4848, 0x4848, 0x4FC8, 0x4848, 0x4008, 0x4028, 0x4010}},
    {21840, 2, {0x80, 0x40, 0xF7FC, 0x9000, 0x9110, 0x9110, 0x92A8, 0x9444, 0x9000, 0x9040, 0xF7FE, 0x9040, 0x40, 0x40, 0x40, 0x40}},
    {21841, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4824, 0x4BFE, 0x4824, 0x49FC, 0x4820, 0x7920, 0x493E, 0x120, 0x2A0, 0x27E, 0x400}},
    {21842, 2, {0x0, 0x3FE, 0x7A02, 0x4A02, 0x4BFE, 0x4A10, 0x4A92, 0x4A92, 0x4A92, 0x4AFE, 0x7A10, 0x4A92, 0x292, 0x492, 0x4FE, 0x802}},
    {21843, 2, {0x1040, 0x840, 0x3E7E, 0x2288, 0x2348, 0x3E50, 0x2020, 0x4058, 0x4186, 0x8000, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21844, 2, {0x1048, 0x844, 0x3E7E, 0x23C0, 0x2224, 0x3E28, 0x2212, 0x206A, 0x4186, 0x8002, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21845, 2, {0x100, 0x100, 0x79FC, 0x4A04, 0x4D04, 0x49E4, 0x4A84, 0x4884, 0x4FF4, 0x4884, 0x7AA4, 0x4AA4, 0x3E4, 0x4, 0x28, 0x10}},
    {21846, 2, {0x40, 0x244, 0xF244, 0x9448, 0x90A0, 0x9110, 0x9608, 0x9044, 0x9040, 0x9248, 0xF248, 0x9450, 0xA0, 0x110, 0x208, 0xC06}},
    {21847, 2, {0x80, 0x80, 0x78F8, 0x4908, 0x4A10, 0x4C00, 0x4840, 0x4B9C, 0x4A04, 0x4A04, 0x7B9C, 0x4A04, 0x204, 0x204, 0x3FC, 0x204}},
    {21848, 2, {0x80, 0x40, 0xF7FE, 0x9402, 0x9A04, 0x9200, 0x93BC, 0x94A4, 0x94A4, 0x96A4, 0xF934, 0x9128, 0x222, 0x222, 0x422, 0x81E}},
    {21849, 2, {0x440, 0x440, 0x2444, 0x2758, 0x2460, 0x2442, 0x2442, 0x2F42, 0xF03E, 0x0, 0x3E7C, 0x2244, 0x2244, 0x2244, 0x3E7C, 0x2244}},
    {21850, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x200, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {21851, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4824, 0x4BFE, 0x4824, 0x49FC, 0x4840, 0x7BFE, 0x4884, 0x1C8, 0x30, 0xCC, 0x302}},
    {21852, 2, {0x0, 0x7BC, 0x7084, 0x5294, 0x5108, 0x5294, 0x54A4, 0x5840, 0x5000, 0x57BC, 0x70A4, 0x52A4, 0x128, 0x290, 0x4A8, 0x846}},
    {21853, 2, {0x0, 0x180, 0x7700, 0x511E, 0x5112, 0x5FD2, 0x5112, 0x5312, 0x5392, 0x5352, 0x7552, 0x5512, 0x91E, 0x112, 0x100, 0x100}},
    {21854, 2, {0x0, 0x7FE, 0x7890, 0x4890, 0x4890, 0x4B9C, 0x4A04, 0x4A04, 0x4A04, 0x4B9C, 0x7890, 0x4890, 0x90, 0x90, 0x7FE, 0x0}},
    {21855, 2, {0x820, 0x420, 0x3F20, 0x213E, 0x2144, 0x2144, 0x3F44, 0x20A4, 0x2028, 0x3F28, 0x3110, 0x5110, 0x5128, 0x9F28, 0x1144, 0x82}},
    {21856, 2, {0x60C, 0x3870, 0x2040, 0x2040, 0x3F7E, 0x2448, 0x2448, 0x2448, 0x4488, 0x8100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21857, 2, {0x90, 0x90, 0xF090, 0x979E, 0x9090, 0x9090, 0x9090, 0x939C, 0x9090, 0x9090, 0xF090, 0x979E, 0x90, 0x90, 0x90, 0x90}},
    {21858, 2, {0x0, 0x3FE, 0x7820, 0x4820, 0x4BFE, 0x4A22, 0x4A22, 0x4B32, 0x4AAA, 0x4AAA, 0x7B76, 0x4A66, 0x222, 0x222, 0x22A, 0x204}},
    {21859, 2, {0x400, 0x400, 0x77DE, 0x5912, 0x5112, 0x5112, 0x5FD2, 0x5112, 0x5112, 0x55D2, 0x7512, 0x551A, 0x5D4, 0x1E10, 0x810, 0x10}},
    {21860, 2, {0x40, 0x80, 0x7BFC, 0x4A24, 0x4A24, 0x4BFC, 0x4A24, 0x4A44, 0x4BFC, 0x4890, 0x7910, 0x4FFE, 0x10, 0x10, 0x10, 0x10}},
    {21861, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A08, 0x4DF6, 0x4840, 0x4840, 0x4FFC, 0x4840, 0x7840, 0x4BF8, 0x208, 0x208, 0x3F8, 0x208}},
    {21862, 2, {0x120, 0x110, 0x7110, 0x517E, 0x5780, 0x5104, 0x5144, 0x51C4, 0x5124, 0x5324, 0x7528, 0x5128, 0x108, 0x110, 0x57E, 0x200}},
    {21863, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x7924, 0x4924, 0x124, 0x50, 0x88, 0x304}},
    {21864, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x104, 0x114, 0x108}},
    {21865, 2, {0x110, 0x110, 0xF7D0, 0x9110, 0x9110, 0x9118, 0x9FF4, 0x9012, 0x9112, 0x9110, 0xF7D0, 0x9110, 0x110, 0x1D0, 0xE10, 0x410}},
    {21866, 2, {0x210, 0x210, 0x7220, 0x527C, 0x5FC4, 0x5244, 0x5244, 0x52C4, 0x537C, 0x5E44, 0x7244, 0x5244, 0x244, 0x244, 0xA7C, 0x444}},
    {21867, 2, {0x40, 0x44, 0x7BF4, 0x4848, 0x4850, 0x4FFE, 0x4840, 0x4880, 0x49F8, 0x4B08, 0x7D08, 0x49F8, 0x108, 0x108, 0x1F8, 0x108}},
    {21868, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1110, 0x920, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {21869, 2, {0x210, 0x210, 0x7210, 0x5FBC, 0x5210, 0x5510, 0x557E, 0x5790, 0x5120, 0x513C, 0x71C4, 0x5F08, 0x528, 0x110, 0x108, 0x104}},
    {21870, 2, {0x40, 0x40, 0xF27C, 0x9240, 0x9240, 0x9240, 0x9FFE, 0x9000, 0x9444, 0x9444, 0xF4A4, 0x9494, 0x514, 0x404, 0x7FC, 0x4}},
    {21871, 2, {0x0, 0x3FE, 0x7A02, 0x4A02, 0x4BFE, 0x4A22, 0x4A22, 0x4AFA, 0x4A22, 0x4A32, 0x7A2A, 0x4BFE, 0x202, 0x202, 0x3FE, 0x202}},
    {21872, 2, {0x0, 0x3FE, 0x7A52, 0x4A52, 0x4A52, 0x4BFE, 0x4820, 0x4840, 0x48FC, 0x4904, 0x7A88, 0x4850, 0x20, 0x40, 0x180, 0x600}},
    {21873, 2, {0x20, 0x124, 0x7924, 0x4924, 0x49FC, 0x4800, 0x4800, 0x4BFE, 0x4840, 0x4880, 0x78FC, 0x4984, 0x284, 0x484, 0xFC, 0x84}},
    {21874, 2, {0x88, 0x88, 0x7908, 0x4BDE, 0x4A52, 0x4A62, 0x4A42, 0x4A52, 0x4BCA, 0x4A4A, 0x7A42, 0x4A42, 0x242, 0x3C2, 0x254, 0x8}},
    {21875, 2, {0x40, 0x248, 0xF150, 0x97FC, 0x9080, 0x9100, 0x9FFE, 0x9210, 0x9408, 0x9BF4, 0xF212, 0x9210, 0x250, 0x224, 0x204, 0x1FC}},
    {21876, 2, {0x104, 0x88, 0x7850, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x7820, 0x4820, 0x3FE, 0x20, 0x20, 0x20}},
    {21877, 2, {0x10, 0x410, 0x7210, 0x52FE, 0x5892, 0x5494, 0x5490, 0x52FC, 0x52A4, 0x54A4, 0x7CA8, 0x54A8, 0x490, 0x528, 0x544, 0x282}},
    {21878, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x4800, 0x4BFE, 0x4820, 0x4820, 0x4920, 0x793C, 0x4920, 0x2A0, 0x260, 0x43E, 0x800}},
    {21879, 2, {0x100, 0x9E, 0x7BD2, 0x4A52, 0x4A54, 0x4BD4, 0x4A58, 0x4A54, 0x4BD2, 0x4A12, 0x7A92, 0x4A5A, 0x2D4, 0x350, 0x210, 0x10}},
    {21880, 2, {0x40, 0x40, 0x7BF8, 0x4848, 0x4FFE, 0x4848, 0x4BF8, 0x4840, 0x4A48, 0x4A48, 0x7AE8, 0x4B58, 0x248, 0x248, 0x448, 0x848}},
    {21881, 2, {0x0, 0x3FC, 0x7A04, 0x4A04, 0x4BFC, 0x4A20, 0x4A20, 0x4BFE, 0x4A20, 0x4A20, 0x7AFC, 0x4A84, 0x284, 0x484, 0x4FC, 0x884}},
    {21882, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4800, 0x4BFE, 0x4900, 0x49FE, 0x7A4A, 0x4C92, 0x122, 0x242, 0x94, 0x108}},
    {21883, 2, {0x100, 0x3FF8, 0x820, 0x440, 0x7FFE, 0x4102, 0x9FF4, 0x1110, 0x1150, 0x1120, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21884, 2, {0x40, 0x20, 0x79FC, 0x4800, 0x4888, 0x4850, 0x4BFE, 0x4A22, 0x4C24, 0x49FC, 0x4924, 0x7924, 0x4934, 0x128, 0x20, 0x20}},
    {21885, 2, {0x40, 0xA0, 0x7910, 0x4A08, 0x4DF6, 0x4800, 0x4BF8, 0x4A08, 0x4A08, 0x4BF8, 0x7910, 0x4910, 0x7FC, 0x110, 0x210, 0x410}},
    {21886, 2, {0x88, 0x1C8, 0x7708, 0x5108, 0x512A, 0x512A, 0x57AC, 0x5148, 0x5308, 0x5388, 0x7554, 0x5514, 0x914, 0x124, 0x124, 0x142}},
    {21887, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48F8, 0x4888, 0x48F8, 0x4888, 0x4888, 0x4BFE, 0x7900, 0x4948, 0x184, 0x100, 0x1FE, 0x0}},
    {21888, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4880, 0x48F8, 0x4908, 0x4A90, 0x4860, 0x4998, 0x7E06, 0x49F8, 0x108, 0x108, 0x1F8, 0x108}},
    {21889, 2, {0x1FC, 0x124, 0x7924, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x7A22, 0x4A2A, 0x2FA, 0x20A, 0x202, 0x206}},
    {21890, 2, {0x0, 0x3F8, 0x7A48, 0x4A48, 0x4BF8, 0x4A48, 0x4A48, 0x4BF8, 0x4800, 0x4FFE, 0x7A40, 0x4A44, 0x228, 0x290, 0x308, 0x206}},
    {21891, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x9040, 0x97FC, 0x9514, 0x94A4, 0x95F4, 0x9444, 0xF444, 0x95F4, 0x444, 0x444, 0x454, 0x408}},
    {21892, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x1110, 0x920, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21893, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x4840, 0x4820, 0x4BFE, 0x7800, 0x4904, 0x88, 0x0, 0x7FE, 0x0}},
    {21894, 2, {0x1010, 0x1010, 0x1010, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x0, 0x0, 0x7C7C, 0x4444, 0x4444, 0x4444, 0x4444, 0x7C7C, 0x4444}},
    {21895, 2, {0x104, 0x104, 0x77C4, 0x5104, 0x5114, 0x57D4, 0x5554, 0x5554, 0x57D4, 0x5114, 0x7394, 0x5554, 0x944, 0x104, 0x114, 0x108}},
    {21896, 2, {0x220, 0x224, 0x7BA8, 0x4A30, 0x4AA2, 0x4B22, 0x4A5E, 0x4880, 0x4BFC, 0x4A04, 0x7A04, 0x4BFC, 0x204, 0x204, 0x3FC, 0x204}},
    {21897, 2, {0x100, 0x178, 0x7108, 0x5208, 0x52FE, 0x5640, 0x5A40, 0x527C, 0x5290, 0x5210, 0x72FE, 0x5210, 0x228, 0x228, 0x244, 0x282}},
    {21898, 2, {0x14, 0x12, 0x7010, 0x57FE, 0x5410, 0x5410, 0x55D0, 0x5412, 0x5412, 0x55D4, 0x7554, 0x5548, 0x5DA, 0x42A, 0x846, 0x1082}},
    {21899, 2, {0x48, 0x148, 0x7948, 0x4BFE, 0x4948, 0x4948, 0x4978, 0x4900, 0x49FE, 0x4820, 0x7BFE, 0x4870, 0xA8, 0x124, 0x622, 0x20}},
    {21900, 2, {0x0, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x0, 0x1008, 0x1108, 0x1108, 0x5548, 0x5328, 0x9108, 0x2108, 0x2108, 0x4108, 0x8108}},
    {21901, 2, {0x90, 0x92, 0x7AD4, 0x4A98, 0x4A92, 0x4AD2, 0x4F0E, 0x4820, 0x4820, 0x4BFE, 0x7870, 0x48A8, 0x124, 0x622, 0x20, 0x20}},
    {21902, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x49E4, 0x4924, 0x4924, 0x4BFE, 0x4A02, 0x4AFA, 0x7A8A, 0x4A8A, 0x2FA, 0x202, 0x20A, 0x204}},
    {21903, 2, {0x108, 0x108, 0xF7FE, 0x9108, 0x9148, 0x9040, 0x97FE, 0x9080, 0x9100, 0x93FC, 0xF504, 0x9904, 0x104, 0x104, 0x1FC, 0x104}},
    {21904, 2, {0x100, 0x11E, 0x77F2, 0x5112, 0x5214, 0x53D4, 0x5258, 0x5654, 0x5BD2, 0x5252, 0x7252, 0x53DA, 0x254, 0x250, 0x250, 0x2D0}},
    {21905, 2, {0x80, 0x40, 0x7BFC, 0x4800, 0x4908, 0x4890, 0x4FFE, 0x4800, 0x49F8, 0x4908, 0x7908, 0x49F8, 0x108, 0x108, 0x1F8, 0x108}},
    {21906, 2, {0x88, 0x88, 0x79EC, 0x492A, 0x4A28, 0x4D48, 0x4888, 0x497E, 0x4E00, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104, 0x104, 0x1FC}},
    {21907, 2, {0x0, 0x3FE, 0x7850, 0x4850, 0x4BFE, 0x4A52, 0x4A52, 0x4BFE, 0x4800, 0x4840, 0x4FFE, 0x7888, 0x4990, 0x60, 0xD8, 0x304}},
    {21908, 2, {0x0, 0x3FC, 0x7A04, 0x4A04, 0x4BFC, 0x4A00, 0x4BFC, 0x4A40, 0x4A88, 0x4BFC, 0x7A24, 0x4A20, 0x5FC, 0x420, 0x820, 0x13FE}},
    {21909, 2, {0x0, 0x7FE, 0x7840, 0x4880, 0x4BFC, 0x4A94, 0x4A94, 0x4AF4, 0x4A94, 0x4A94, 0x7AF4, 0x4A94, 0x294, 0x294, 0x3FC, 0x204}},
    {21910, 2, {0x108, 0x108, 0x7108, 0x5108, 0x57BE, 0x5108, 0x5308, 0x5388, 0x557E, 0x5522, 0x7922, 0x5122, 0x122, 0x122, 0x13E, 0x122}},
    {21911, 2, {0x0, 0x3FE, 0x7A22, 0x4820, 0x4BFE, 0x4820, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x3FE, 0x20, 0x20, 0x20}},
    {21912, 2, {0x40, 0x444, 0xF444, 0x9444, 0x97FC, 0x9000, 0x9FFE, 0x9040, 0x9080, 0x97FC, 0xF4A4, 0x94A4, 0x4A4, 0x4A4, 0x4A4, 0x40C}},
    {21913, 2, {0x100, 0x1F8, 0xF208, 0x93F0, 0x9010, 0x97FE, 0x9080, 0x9144, 0x9668, 0x90B0, 0xF128, 0x9668, 0xA4, 0x122, 0x6A0, 0x40}},
    {21914, 2, {0x100, 0x100, 0xF3F0, 0x9410, 0x9820, 0x97FC, 0x9404, 0x94A4, 0x9514, 0x9444, 0xF040, 0x9FFE, 0xA0, 0x110, 0x208, 0xC06}},
    {21915, 2, {0x3C, 0x7C0, 0x7A44, 0x4928, 0x4800, 0x4BFC, 0x4880, 0x4880, 0x4FFE, 0x4900, 0x79F8, 0x4A88, 0x250, 0x420, 0x8D8, 0x306}},
    {21916, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21917, 2, {0x0, 0x3F8, 0x7A08, 0x4BF8, 0x4A08, 0x4BF8, 0x4900, 0x4BFC, 0x4C44, 0x4A44, 0x7AA4, 0x4A04, 0x3F4, 0x4, 0x28, 0x10}},
    {21918, 2, {0x80, 0x100, 0x7BDE, 0x4A52, 0x4A52, 0x4BD2, 0x4A52, 0x4A52, 0x4BD2, 0x4812, 0x7A5A, 0x4B94, 0x210, 0x250, 0x1D0, 0x10}},
    {21919, 2, {0x0, 0x3FE, 0x7A22, 0x4BFE, 0x4A22, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104, 0x104, 0x114, 0x108}},
    {21920, 2, {0x8, 0x3C, 0x79E0, 0x4820, 0x4BFE, 0x4820, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x1FC, 0x20, 0x3FE, 0x0}},
    {21921, 2, {0x40, 0x40, 0x79FC, 0x4884, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x220, 0x3FE, 0x20, 0x20}},
    {21922, 2, {0x4, 0x1E, 0x7BE0, 0x4820, 0x4820, 0x4FFE, 0x4820, 0x48A0, 0x4B2E, 0x4A22, 0x7A22, 0x4BAE, 0x222, 0x222, 0x3FE, 0x202}},
    {21923, 2, {0x100, 0x100, 0x79FC, 0x4A04, 0x4C04, 0x49E4, 0x4924, 0x4924, 0x79E4, 0x4804, 0x28, 0x10, 0x2488, 0x2244, 0x4244, 0x8004}},
    {21924, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4800, 0x4BFE, 0x4820, 0x7820, 0x49FC, 0x20, 0x20, 0x3FE, 0x0}},
    {21925, 2, {0x40, 0x20, 0x7BFE, 0x4A48, 0x4A48, 0x4BFE, 0x4A48, 0x4A78, 0x4A00, 0x4AFC, 0x7A44, 0x4A48, 0x228, 0x410, 0x468, 0x986}},
    {21926, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7D7C, 0x4544, 0x4544, 0x7D7C, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {21927, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x49FC, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104, 0x0, 0x3FE, 0x0}},
    {21928, 2, {0x80, 0x40, 0xF7FE, 0x9000, 0x93F8, 0x9208, 0x93F8, 0x9000, 0x97FE, 0x9402, 0xF9F4, 0x9110, 0x110, 0x212, 0x412, 0x80E}},
    {21929, 2, {0x1C0, 0xA0, 0x7910, 0x4A08, 0x4DF6, 0x4800, 0x4BCA, 0x4A4A, 0x4A54, 0x4BD4, 0x7A68, 0x4A54, 0x3D4, 0x24A, 0x24A, 0x2C0}},
    {21930, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3D78, 0x2548, 0x3D78, 0x100, 0xFFFE, 0x1200, 0x1110, 0x10A0, 0x1040, 0x1430, 0x180E, 0x1000}},
    {21931, 2, {0x80, 0x9E, 0x7BEA, 0x488A, 0x49CA, 0x488A, 0x4BEA, 0x4892, 0x48A6, 0x4820, 0x7BFE, 0x4820, 0x50, 0x88, 0x104, 0x602}},
    {21932, 2, {0xF8, 0x3F00, 0x200, 0xFFFE, 0x440, 0xFE0, 0x3458, 0xC7C6, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2028, 0x2010}},
    {21933, 2, {0x40, 0x20, 0x7BFC, 0x4800, 0x4908, 0x4890, 0x4BFE, 0x4A10, 0x4A20, 0x4AC8, 0x7A10, 0x4A24, 0x2C8, 0x410, 0x460, 0x980}},
    {21934, 2, {0x0, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {21935, 2, {0x20, 0x20, 0x7BFE, 0x4850, 0x48A8, 0x4924, 0x4AFA, 0x4820, 0x48A8, 0x4888, 0x7BFE, 0x4888, 0x88, 0x88, 0x108, 0x208}},
    {21936, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A88, 0x4C46, 0x4BF8, 0x4A08, 0x4BF8, 0x4A08, 0x7BF8, 0x4A44, 0x228, 0x290, 0x308, 0x204}},
    {21937, 2, {0x0, 0x7FE, 0xF400, 0x95FC, 0x9524, 0x9524, 0x95FC, 0x9524, 0x9524, 0x95FC, 0xF420, 0x9420, 0x5FC, 0x820, 0x820, 0x13FE}},
    {21938, 2, {0x110, 0x110, 0x7210, 0x529E, 0x54A2, 0x5F22, 0x5142, 0x5212, 0x548A, 0x5FCA, 0x7442, 0x5002, 0xA82, 0x942, 0x1154, 0x8}},
    {21939, 2, {0x40, 0x40, 0xF7FC, 0x90E0, 0x9150, 0x9248, 0x9446, 0x93F8, 0x9208, 0x93F8, 0xF208, 0x93F8, 0x208, 0x0, 0xFFE, 0x0}},
    {21940, 2, {0x14, 0x12, 0x77FE, 0x5410, 0x5410, 0x57F0, 0x5492, 0x5492, 0x57D2, 0x5554, 0x7554, 0x5488, 0x54A, 0xA1A, 0x826, 0x1042}},
    {21941, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4888, 0x4800, 0x49FC, 0x4924, 0x4924, 0x4924, 0x79FC, 0x4924, 0x124, 0x124, 0x1FC, 0x104}},
    {21942, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {21943, 2, {0x20, 0x1FC, 0x7820, 0x4888, 0x4BFE, 0x4888, 0x4800, 0x49FC, 0x4904, 0x4924, 0x7924, 0x4924, 0x124, 0x50, 0x88, 0x304}},
    {21944, 2, {0x120, 0x110, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {21945, 2, {0x40, 0x40, 0xF7FC, 0x90A0, 0x9110, 0x9248, 0x9DF6, 0x9040, 0x9040, 0x97FC, 0xF000, 0x9040, 0x3F8, 0x40, 0x40, 0xFFE}},
    {21946, 2, {0x106, 0x178, 0x7110, 0x5222, 0x5244, 0x56F8, 0x5A14, 0x5222, 0x52FE, 0x5212, 0x7210, 0x5254, 0x292, 0x312, 0x250, 0x220}},
    {21947, 2, {0x40, 0xA0, 0x7910, 0x4A08, 0x4DF6, 0x4800, 0x4BC4, 0x4A54, 0x4A54, 0x4BD4, 0x7A54, 0x4A54, 0x3D4, 0x244, 0x254, 0x2C8}},
    {21948, 2, {0x80, 0x80, 0x79F8, 0x4A08, 0x4C10, 0x4BFC, 0x4804, 0x49FC, 0x4804, 0x4BFC, 0x7804, 0x4840, 0x124, 0x52A, 0x50A, 0x8F8}},
    {21949, 2, {0x20, 0x124, 0x78A8, 0x4820, 0x4BFE, 0x48A8, 0x4924, 0x4A02, 0x4840, 0x4FFE, 0x7888, 0x4908, 0x190, 0x60, 0x198, 0x604}},
    {21950, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4102, 0x9104, 0x1FF0, 0x2100, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {21951, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7C7C, 0x4444, 0x4444, 0x7C7C, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {21952, 2, {0x1000, 0x1078, 0xFE48, 0x1048, 0x7C48, 0x86, 0xFE00, 0x82FC, 0x44, 0x7C44, 0x28, 0x7C28, 0x4410, 0x4428, 0x7C44, 0x4582}},
    {21953, 2, {0x4, 0x1E, 0x7BF0, 0x4A1E, 0x4A10, 0x4AFE, 0x4A92, 0x4A98, 0x4AF2, 0x4A8E, 0x7A80, 0x4AB8, 0x2A8, 0x4AA, 0x54A, 0xA86}},
    {21954, 2, {0x8, 0x3C, 0x7BC0, 0x4804, 0x4A44, 0x4928, 0x49FC, 0x4A20, 0x4820, 0x4BFE, 0x7820, 0x4924, 0x124, 0x124, 0x1FC, 0x4}},
    {21955, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x4800, 0x4BFE, 0x4A02, 0x7AFA, 0x4A8A, 0x2FA, 0x202, 0x20A, 0x204}},
    {21956, 2, {0x7FC, 0x80, 0xF3F8, 0x9208, 0x93F8, 0x9208, 0x93F8, 0x9208, 0x93F8, 0x9080, 0xF1F8, 0x9310, 0x4A0, 0x40, 0x1B0, 0xE0E}},
    {21957, 2, {0x40, 0x80, 0x7BF8, 0x4A08, 0x4BF8, 0x4A08, 0x4BF8, 0x4A08, 0x4BF8, 0x4850, 0x7848, 0x4FFE, 0xA0, 0x110, 0x208, 0xC06}},
    {21958, 2, {0x40, 0x40, 0xF0A0, 0x9190, 0x9248, 0x97FE, 0x9A08, 0x93F8, 0x9208, 0x93F8, 0xF200, 0x93F8, 0x508, 0x508, 0x9F8, 0x108}},
    {21959, 2, {0x100, 0x100, 0x7FFC, 0x1110, 0x2928, 0x4544, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {21960, 2, {0x124, 0x248, 0x7A48, 0x4924, 0x4BF8, 0x4A08, 0x4BF8, 0x4800, 0x4BF8, 0x4A48, 0x7A48, 0x4BF8, 0x202, 0x202, 0x1FE, 0x0}},
    {21961, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x4840, 0x4888, 0x49F0, 0x7824, 0x4BFE, 0x22, 0x124, 0x2A2, 0x40}},
    {21962, 2, {0x0, 0x1FC, 0x7820, 0x4820, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x88, 0x104, 0x202}},
    {21963, 2, {0x40, 0x1FC, 0x7844, 0x4894, 0x4908, 0x4BDE, 0x494A, 0x494A, 0x4A94, 0x49FC, 0x7904, 0x49FC, 0x104, 0x1FC, 0x104, 0x10C}},
    {21964, 2, {0x208, 0x108, 0xF110, 0x9000, 0x97FC, 0x9000, 0x9110, 0x9208, 0x9404, 0x93F8, 0xF2A8, 0x92A8, 0x2A8, 0x2A8, 0xFFE, 0x0}},
    {21965, 2, {0x440, 0x25E, 0x7292, 0x5012, 0x57D2, 0x511E, 0x5552, 0x5552, 0x5552, 0x57DE, 0x7152, 0x5112, 0x212, 0x222, 0x42A, 0x844}},
    {21966, 2, {0x0, 0x1FE, 0x7920, 0x49FC, 0x4920, 0x49FC, 0x4920, 0x4920, 0x49FE, 0x4802, 0x7AAA, 0x4AAA, 0x2AA, 0x202, 0x14, 0x8}},
    {21967, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48A8, 0x4850, 0x4888, 0x4924, 0x4A22, 0x49FC, 0x7820, 0x48A8, 0x124, 0x222, 0xA0, 0x40}},
    {21968, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C24, 0x49FC, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x7820, 0x49FC, 0x104, 0x104, 0x1FC, 0x104}},
    {21969, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x4BFE, 0x4880, 0x4904, 0x4BFE, 0x4802, 0x7BFC, 0x4A94, 0x294, 0x294, 0x7FE, 0x0}},
    {21970, 2, {0x110, 0x110, 0xF7FC, 0x9110, 0x9040, 0x90A0, 0x9110, 0x9208, 0x95F6, 0x9000, 0xF000, 0x93F8, 0x208, 0x208, 0x3F8, 0x208}},
    {21971, 2, {0x3F0, 0x110, 0xF0E0, 0x9318, 0x9000, 0x97BC, 0x94A4, 0x9318, 0x94A4, 0x9040, 0xF7FC, 0x90E0, 0x150, 0x248, 0xC46, 0x40}},
    {21972, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9208, 0x93F8, 0x9208, 0x93F8, 0x9208, 0xF3F8, 0x9208, 0xFFE, 0x110, 0x208, 0x404}},
    {21973, 2, {0x0, 0x3FE, 0x7A00, 0x4AFC, 0x4A00, 0x4BFE, 0x4AA8, 0x4A90, 0x4AC8, 0x4A86, 0x7A08, 0x4BFE, 0x288, 0x448, 0x408, 0x818}},
    {21974, 2, {0x140, 0x65C, 0xF444, 0x9444, 0x975C, 0x9444, 0x9444, 0x97FC, 0x9040, 0x97FC, 0xF208, 0x9110, 0xA0, 0x40, 0x1B0, 0xE0E}},
    {21975, 2, {0x0, 0x1FC, 0x7904, 0x49E4, 0x4924, 0x4BFE, 0x4A02, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104, 0x104, 0x114, 0x108}},
    {21976, 2, {0x1E, 0x3E0, 0x7922, 0x4894, 0x4840, 0x4888, 0x49F0, 0x4820, 0x48C4, 0x49FE, 0x7822, 0x4820, 0x3FE, 0x50, 0x88, 0x306}},
    {21977, 2, {0x40, 0x20, 0x7BFC, 0x4908, 0x4890, 0x4BFE, 0x4A02, 0x4C44, 0x4820, 0x4BFC, 0x7880, 0x48F8, 0x88, 0x108, 0x128, 0x210}},
    {21978, 2, {0x40, 0x80, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x4900, 0x49FE, 0x4900, 0x49FE, 0x7802, 0x4AAA, 0x2AA, 0x402, 0x14, 0x8}},
    {21979, 2, {0x108, 0x90, 0x7800, 0x4BFE, 0x48A0, 0x4BFC, 0x48A4, 0x4FFE, 0x48A4, 0x4BFC, 0x78A0, 0x49B0, 0x2A8, 0xCA6, 0xA0, 0xA0}},
    {21980, 2, {0x40, 0x44, 0x79F8, 0x4850, 0x4BFE, 0x4840, 0x48F8, 0x4982, 0x4A7E, 0x4C00, 0x79FC, 0x4904, 0x1FC, 0x104, 0x1FC, 0x104}},
    {21981, 2, {0x0, 0xFFE, 0xF000, 0x93F8, 0x9208, 0x9208, 0x93F8, 0x9000, 0x97FC, 0x9514, 0xF4A4, 0x97FC, 0x444, 0x444, 0x454, 0x408}},
    {21982, 2, {0x84, 0x44, 0x7848, 0x4BFE, 0x4884, 0x4884, 0x4908, 0x494A, 0x4A52, 0x4B9C, 0x7884, 0x4908, 0x108, 0x252, 0x3DE, 0x42}},
    {21983, 2, {0x108, 0x88, 0x7890, 0x4BFC, 0x4840, 0x49F8, 0x4840, 0x4BFE, 0x4880, 0x4900, 0x79FC, 0x4A20, 0x420, 0x820, 0x3FE, 0x0}},
    {21984, 2, {0xE40, 0xF040, 0x2278, 0x9288, 0x5550, 0x20, 0x7C50, 0x888, 0x1106, 0x10F8, 0x1E88, 0xF088, 0x5088, 0x1088, 0x50F8, 0x2088}},
    {21985, 2, {0x90, 0x88, 0x7904, 0x4A42, 0x4C88, 0x49FC, 0x4804, 0x4800, 0x4BDE, 0x4842, 0x7A52, 0x494A, 0x252, 0x42, 0x14A, 0x84}},
    {21986, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x4800, 0x4BFC, 0x7A94, 0x4A94, 0x294, 0x294, 0x7FE, 0x0}},
    {21987, 2, {0x0, 0x3E7C, 0x2204, 0x2204, 0x3EF4, 0x4, 0x7F04, 0x5574, 0x5554, 0x5554, 0x7F54, 0x5554, 0x5574, 0x5504, 0x5114, 0x4308}},
    {21988, 2, {0x20, 0x124, 0x7924, 0x49FC, 0x4820, 0x4BFE, 0x4800, 0x4820, 0x49FC, 0x4924, 0x7924, 0x49FC, 0x20, 0x24, 0x3FE, 0x102}},
    {21989, 2, {0x40, 0x80, 0xF3F8, 0x9208, 0x93F8, 0x9208, 0x93F8, 0x9080, 0x97FC, 0x9110, 0xF248, 0x9444, 0xBFA, 0x40, 0x40, 0x40}},
    {21990, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x97FC, 0x9484, 0x9918, 0x93E0, 0x9040, 0x9088, 0xF7FC, 0x9044, 0x250, 0x448, 0x948, 0x80}},
    {21991, 2, {0x1000, 0x7E78, 0x1248, 0x2A48, 0x4578, 0x8280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x2488, 0x2488, 0x3FF8, 0x2488, 0x24A8, 0x2010}},
    {21992, 2, {0x80, 0x880, 0xE4FE, 0xA100, 0xAAFC, 0xA484, 0xA4A4, 0xA094, 0xA3FE, 0xA484, 0xED24, 0xA514, 0x5FE, 0x404, 0x428, 0x10}},
    {21993, 2, {0x20, 0x124, 0x78A8, 0x4820, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x0, 0x88, 0x104}},
    {21994, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9080, 0x97FC, 0x9110, 0x9268, 0x95C6, 0xF040, 0x93F8, 0xE0, 0x158, 0x644, 0x40}},
    {21995, 2, {0x0, 0x3FC, 0x7908, 0x49F8, 0x4908, 0x49F8, 0x490E, 0x4FF8, 0x4808, 0x4FBC, 0x78A4, 0x4AA4, 0x128, 0x290, 0x4A8, 0x846}},
    {21996, 2, {0x110, 0x110, 0xF7FE, 0x9110, 0x9100, 0x92FE, 0x9208, 0x9608, 0x9AE8, 0x92A8, 0xF2A8, 0x92E8, 0x2A8, 0x208, 0x228, 0x210}},
    {21997, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x90, 0x3EE4, 0x2284, 0x3E7C}},
    {21998, 2, {0x0, 0xFE, 0xEE28, 0xAA28, 0xAA28, 0xAAFE, 0xAAAA, 0xAEAA, 0xAAAA, 0xAAAA, 0xEAAE, 0xAAC2, 0xE82, 0xA82, 0xFE, 0x82}},
    {21999, 2, {0x0, 0x7FC, 0xF444, 0x9444, 0x97FC, 0x9444, 0x94A4, 0x9514, 0x9404, 0x97FC, 0xF040, 0x9024, 0x522, 0x50A, 0x908, 0xF8}},
    {22000, 2, {0x200, 0x3FC, 0x7524, 0x5524, 0x5D24, 0x55FC, 0x5524, 0x5524, 0x5574, 0x5554, 0x7554, 0x5554, 0x574, 0x504, 0x5FC, 0x504}},
    {22001, 2, {0x20, 0x50, 0x7888, 0x4974, 0x4A02, 0x48F8, 0x4888, 0x48F8, 0x480C, 0x49F0, 0x7820, 0x49FC, 0x20, 0x3FE, 0x20, 0x60}},
    {22002, 2, {0x90, 0x108, 0xF294, 0x9060, 0x9198, 0x9646, 0x90FC, 0x9308, 0x90B0, 0x90D0, 0xF73E, 0x90C2, 0x324, 0x18, 0x60, 0x780}},
    {22003, 2, {0xE, 0x3F0, 0x7844, 0x4A24, 0x4908, 0x4BFE, 0x4A42, 0x4840, 0x4BFE, 0x4880, 0x78FC, 0x4944, 0x128, 0x210, 0x468, 0x186}},
    {22004, 2, {0x108, 0x90, 0x7BFC, 0x4840, 0x4840, 0x49F8, 0x4840, 0x4840, 0x4BFC, 0x48A0, 0x78A8, 0x48B4, 0x13C, 0x122, 0x222, 0x41E}},
    {22005, 2, {0x1FC, 0x808, 0xE450, 0xA420, 0xA1FC, 0xA124, 0xBD24, 0xA5FC, 0xA524, 0xA5FC, 0xE524, 0xA524, 0x52C, 0xA00, 0x11FE, 0x0}},
    {22006, 2, {0x0, 0x1FC, 0x7924, 0x49FC, 0x4924, 0x49FC, 0x48A8, 0x48A8, 0x4BFE, 0x48A8, 0x78A8, 0x49FC, 0x20, 0x3FE, 0x20, 0x20}},
    {22007, 2, {0x108, 0x108, 0x7108, 0x57D0, 0x511E, 0x5794, 0x5124, 0x5FD4, 0x5214, 0x53D4, 0x7254, 0x5248, 0x248, 0x4D4, 0x424, 0x842}},
    {22008, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x8000, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {22009, 2, {0x20, 0x820, 0x75FE, 0x5420, 0x51FC, 0x5124, 0x5DFC, 0x5524, 0x55FC, 0x5420, 0x75FE, 0x5420, 0x420, 0xA20, 0x11FE, 0x0}},
    {22010, 2, {0x20, 0x222, 0x7BFE, 0x4890, 0x4888, 0x49FE, 0x4B10, 0x4D10, 0x49FE, 0x4910, 0x7910, 0x49FE, 0x110, 0x110, 0x1FE, 0x100}},
    {22011, 2, {0x40, 0x20, 0x7BFE, 0x4A00, 0x4A48, 0x4A48, 0x4BFE, 0x4A48, 0x4A48, 0x4A48, 0x7A78, 0x4A00, 0x554, 0x52A, 0xA2A, 0x1000}},
    {22012, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4BFE, 0x50, 0x88, 0x104, 0x202}},
    {22013, 2, {0x108, 0x108, 0xF7C8, 0x9108, 0x911E, 0x97D2, 0x9564, 0x9540, 0x97C8, 0x9108, 0xF388, 0x9548, 0x954, 0x114, 0x124, 0x142}},
    {22014, 2, {0x420, 0x220, 0x723E, 0x5F40, 0x54A0, 0x5420, 0x573C, 0x5550, 0x5510, 0x5510, 0x75FE, 0x5510, 0x528, 0x928, 0xB44, 0x1082}},
    {22015, 2, {0x40, 0xA0, 0x7910, 0x4A48, 0x4DF6, 0x4820, 0x4840, 0x4BF8, 0x4A08, 0x4BF8, 0x7A08, 0x4BF8, 0x208, 0x3F8, 0x110, 0x208}},
    {22016, 2, {0x80, 0x40, 0xF7FC, 0x9110, 0x90A0, 0x97FC, 0x9444, 0x95F4, 0x9444, 0x95F4, 0xF514, 0x9514, 0x5F4, 0x404, 0x414, 0x408}},
    {22017, 2, {0x14, 0x12, 0x7010, 0x57FE, 0x5410, 0x5490, 0x54D2, 0x5492, 0x57F2, 0x5494, 0x7494, 0x55C8, 0xAAA, 0x89A, 0x11A6, 0x42}},
    {22018, 2, {0x40, 0x785E, 0x4852, 0x4A52, 0x4A52, 0x4A52, 0x7A5E, 0x240, 0x7A5E, 0x4AD2, 0x4B52, 0x4A52, 0x4852, 0x785E, 0x4852, 0x40}},
    {22019, 2, {0x20, 0x7FE, 0x7420, 0x55FC, 0x5424, 0x57FE, 0x5424, 0x55FC, 0x5420, 0x55FC, 0x7524, 0x55FC, 0x524, 0x9FC, 0x924, 0x112C}},
    {22020, 2, {0x100, 0x3F8, 0x7A08, 0x4BF8, 0x4A08, 0x4BF8, 0x4A00, 0x4BFE, 0x4A00, 0x4BFE, 0x7842, 0x4FFA, 0x164, 0x658, 0x1846, 0x40}},
    {22021, 2, {0x0, 0x7BE, 0x7488, 0x5488, 0x57A8, 0x54A8, 0x54A8, 0x57BE, 0x5410, 0x5518, 0x7498, 0x55A8, 0x6AA, 0x44A, 0x46, 0x80}},
    {22022, 2, {0x110, 0x110, 0xF7FE, 0x9110, 0x91F0, 0x9040, 0x93F8, 0x9248, 0x93F8, 0x9040, 0xF7FC, 0x9040, 0xFFE, 0xA0, 0x110, 0x60C}},
    {22023, 2, {0x80, 0x110, 0xF3F8, 0x9210, 0x94A4, 0x9FBE, 0x9042, 0x90A0, 0x9318, 0x9C46, 0xF190, 0x9664, 0x188, 0x630, 0x1C0, 0xE00}},
    {22024, 2, {0xA0, 0xA0, 0x77FE, 0x50A0, 0x57FC, 0x54A4, 0x57FC, 0x54A4, 0x57FC, 0x5000, 0x73F8, 0x5208, 0x3F8, 0x208, 0x3F8, 0x208}},
    {22025, 2, {0x100, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0xFFFE, 0x1000, 0x7F7C, 0x1144, 0x2144, 0x457C, 0x8244}},
    {22026, 2, {0x20, 0x222, 0x7A22, 0x4BFE, 0x4800, 0x4BFE, 0x4A20, 0x4BFC, 0x4A20, 0x4BFE, 0x7A00, 0x4A20, 0x5FC, 0x420, 0xBFE, 0x0}},
    {22027, 2, {0x208, 0x248, 0xE748, 0xA290, 0xA29E, 0xAFE4, 0xA114, 0xA214, 0xAFD4, 0xA494, 0xE914, 0xA1C8, 0xF08, 0x114, 0x514, 0x222}},
    {22028, 2, {0x0, 0xFFE, 0xE0A0, 0xA7FC, 0xA4A4, 0xA4A4, 0xA7FC, 0xA000, 0xA7FC, 0xA000, 0xEFFE, 0xA040, 0x248, 0x444, 0x942, 0x80}},
    {22029, 2, {0x40, 0x3F8, 0xF248, 0x9FFE, 0x9248, 0x93F8, 0x9040, 0x93F8, 0x9248, 0x93F8, 0xF080, 0x9FFE, 0x110, 0xE0, 0x318, 0xC04}},
    {22030, 2, {0x7FC, 0x80, 0x7BF8, 0x4A08, 0x4BF8, 0x4A08, 0x4BF8, 0x4A08, 0x4BF8, 0x4910, 0x797C, 0x4F80, 0x90, 0x62, 0x1B2, 0xE0E}},
    {22031, 2, {0x1000, 0x11DC, 0x1144, 0x1144, 0xFD44, 0x11DC, 0x1100, 0x1100, 0x7DDC, 0x4514, 0x4514, 0x45D4, 0x4508, 0x7D08, 0x4514, 0x122}},
    {22032, 2, {0x0, 0x7BC, 0xF084, 0x94A4, 0x9294, 0x94A4, 0x9050, 0x9188, 0x9626, 0x90C0, 0xF310, 0x9064, 0x388, 0x30, 0xC0, 0x700}},
    {22033, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x4A78, 0x4BC4, 0x4A3C, 0x4A08, 0x4AF0, 0x7B24, 0x4AA8, 0x5FE, 0x420, 0x8A0, 0x40}},
    {22034, 2, {0x88, 0x3DE, 0x7888, 0x49DC, 0x4888, 0x4BDE, 0x4888, 0x4800, 0x4BFC, 0x4804, 0x7804, 0x49FC, 0x4, 0x4, 0x3FC, 0x4}},
    {22035, 2, {0x0, 0xFFE, 0xE82A, 0xA826, 0xAFFE, 0xA822, 0xABAA, 0xAAAA, 0xABAA, 0xA832, 0xE996, 0xAE2E, 0x846, 0x882, 0xFFE, 0x802}},
    {22036, 2, {0x0, 0x3FE, 0x7A00, 0x4A7C, 0x4A44, 0x4A44, 0x4A7C, 0x4A00, 0x4AEE, 0x4AAA, 0x7AAA, 0x4AAA, 0x2EE, 0x200, 0x3FE, 0x0}},
    {22037, 2, {0x0, 0x7FC, 0xF040, 0x9278, 0x9240, 0x9FFE, 0x9100, 0x91FC, 0x9200, 0x93FC, 0xF004, 0x9554, 0x554, 0x804, 0x28, 0x10}},
    {22038, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4820, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104, 0x1FC, 0x88, 0x104}},
    {22039, 2, {0x1110, 0x920, 0x7FFE, 0x5012, 0x9FF4, 0x0, 0x20F0, 0x3F04, 0x2004, 0x1FFC, 0x0, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8}},
    {22040, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x4A78, 0x4BC4, 0x4A3C, 0x4A00, 0x4A50, 0x7A54, 0x4B54, 0x4D8, 0x450, 0x9FE, 0x0}},
    {22041, 2, {0x410, 0x210, 0xE8FE, 0xA492, 0xA2FC, 0xA4A4, 0xACA8, 0xA510, 0xA528, 0xA6C6, 0xE080, 0xAFFE, 0x110, 0x320, 0xE0, 0x718}},
    {22042, 2, {0x200, 0x27C, 0x7444, 0x597C, 0x5144, 0x527C, 0x5600, 0x5AFC, 0x5208, 0x53FE, 0x7208, 0x5288, 0x248, 0x208, 0x228, 0x210}},
    {22043, 2, {0x40, 0x20, 0xF7FE, 0x9490, 0x9490, 0x9490, 0x97FC, 0x9490, 0x94D8, 0x95B8, 0xF5B4, 0x96D4, 0x892, 0x890, 0x1090, 0x90}},
    {22044, 2, {0x40, 0x7FC, 0x7150, 0x5150, 0x52E8, 0x5444, 0x50E0, 0x5358, 0x5C46, 0x5080, 0x71F8, 0x5208, 0x4D0, 0x20, 0xD0, 0x708}},
    {22045, 2, {0x204, 0x204, 0x7784, 0x5494, 0x590C, 0x57C4, 0x5564, 0x5554, 0x57C4, 0x5546, 0x757C, 0x57C4, 0x544, 0x544, 0x444, 0x8C4}},
    {22046, 2, {0x290, 0x290, 0x77D0, 0x5290, 0x53BC, 0x5114, 0x57D4, 0x5554, 0x5554, 0x57D4, 0x7114, 0x57D4, 0x124, 0x124, 0x154, 0x188}},
    {22047, 2, {0x200, 0x23E, 0xEF62, 0xA264, 0xA2A4, 0xAFE8, 0xA124, 0xA224, 0xA7A2, 0xACA2, 0xF4A2, 0xA7B4, 0x4A8, 0x4A0, 0x7A0, 0x4A0}},
    {22048, 2, {0x7FC, 0x80, 0xF3F8, 0x9208, 0x93F8, 0x9208, 0x93F8, 0x9208, 0x97FE, 0x94A2, 0xF89C, 0x97F0, 0x50, 0x24, 0xD4, 0x30C}},
    {22049, 2, {0x20, 0x124, 0x78A8, 0x4BFE, 0x4A02, 0x48F8, 0x4888, 0x4888, 0x48F8, 0x4820, 0x7820, 0x49FC, 0x20, 0x20, 0x3FE, 0x0}},
    {22050, 2, {0x0, 0x7DE, 0x7542, 0x5554, 0x57C8, 0x5544, 0x557E, 0x57CA, 0x510A, 0x5108, 0x77C8, 0x5108, 0x108, 0x1C8, 0xE28, 0x410}},
    {22051, 2, {0x40, 0x444, 0x7444, 0x57FC, 0x5000, 0x57BC, 0x54A4, 0x54A4, 0x57BC, 0x54A4, 0x74A4, 0x57BC, 0x4A4, 0x4A4, 0xAB4, 0x1148}},
    {22052, 2, {0x0, 0xFBE, 0xE8A2, 0xAAAA, 0xAAAA, 0xAAAA, 0xA514, 0xA8A2, 0xA080, 0xAFFE, 0xE110, 0xA210, 0x3A0, 0x60, 0x198, 0xE04}},
    {22053, 2, {0x210, 0x210, 0x7450, 0x595C, 0x5150, 0x5250, 0x56FE, 0x5A00, 0x5210, 0x5250, 0x725C, 0x5250, 0x2B0, 0x290, 0x30E, 0x200}},
    {22054, 2, {0x3FF8, 0x2008, 0x3FF8, 0x1010, 0x2008, 0x5FF4, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38}},
    {22055, 2, {0x80, 0x40, 0xF7FE, 0x9402, 0x9894, 0x9050, 0x9524, 0x954A, 0x98FA, 0x9300, 0xF040, 0x9444, 0x444, 0x444, 0x7FC, 0x4}},
    {22056, 2, {0x40, 0x40, 0x77FC, 0x5044, 0x5FFE, 0x5044, 0x57FC, 0x5040, 0x5554, 0x54E4, 0x7444, 0x57FC, 0x4E4, 0x954, 0xA4C, 0x1044}},
    {22057, 2, {0x88, 0x3FE, 0x7888, 0x4800, 0x49FC, 0x48A8, 0x48A8, 0x4BFE, 0x48A8, 0x48A8, 0x79FC, 0x4820, 0x3FE, 0x20, 0x20, 0x20}},
    {22058, 2, {0x0, 0x3FE, 0x7A52, 0x4A52, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x88, 0x104, 0x202}},
    {22059, 2, {0x110, 0x114, 0x79D2, 0x4A50, 0x4A7E, 0x4D50, 0x4890, 0x4AA8, 0x4928, 0x4944, 0x7A44, 0x4C82, 0x8, 0x2A4, 0x252, 0x452}},
    {22060, 2, {0x0, 0x3F8, 0xF208, 0x93F8, 0x9208, 0x93F8, 0x9000, 0x9FFE, 0x9480, 0x97BC, 0xF494, 0x9794, 0x4D4, 0xF88, 0x94, 0xA2}},
    {22061, 2, {0x104, 0x104, 0x77E8, 0x5110, 0x5100, 0x57C4, 0x5004, 0x57C8, 0x5450, 0x5442, 0x77C2, 0x5444, 0x284, 0xE8, 0xF10, 0x420}},
    {22062, 2, {0x88, 0x2AA, 0x7ADC, 0x4C88, 0x4954, 0x4A22, 0x4800, 0x4BFE, 0x4A42, 0x4C44, 0x7BFC, 0x4844, 0x44, 0x84, 0x114, 0x208}},
    {22063, 2, {0x40, 0x3F8, 0xF048, 0x97FE, 0x9048, 0x93F8, 0x9040, 0x9554, 0x975C, 0x9444, 0xF7FC, 0x9444, 0x75C, 0x554, 0x554, 0x844}},
    {22064, 2, {0x248, 0x248, 0x7554, 0x575C, 0x5248, 0x5554, 0x575C, 0x5248, 0x5FFE, 0x5220, 0x7228, 0x5228, 0x514, 0x494, 0x42C, 0x844}},
    {22065, 2, {0x0, 0x7FC, 0xF404, 0x97FC, 0x9400, 0x95FC, 0x9420, 0x95FC, 0x9524, 0x95FC, 0xF420, 0x97FE, 0x62A, 0x6FA, 0xA0A, 0x1206}},
    {22066, 2, {0x200, 0x23C, 0xFA4, 0xE224, 0xA224, 0xAFBC, 0xA8A4, 0xAFA4, 0xA8A4, 0xAFBC, 0xA224, 0xE224, 0xAFA4, 0x224, 0x244, 0x28C}},
    {22067, 2, {0x20, 0x1FC, 0x7924, 0x49FC, 0x4820, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104, 0x1FC, 0x88, 0x104}},
    {22068, 2, {0x90, 0x294, 0xF2D8, 0x9292, 0x92D2, 0x970E, 0x91F0, 0x9210, 0x97FC, 0x9A44, 0xF3FC, 0x9244, 0x3FC, 0x444, 0x454, 0x808}},
    {22069, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4BFE, 0x4908, 0x4B9C, 0x4908, 0x4988, 0x4E3E, 0x7800, 0x4BFE, 0x90, 0x90, 0x112, 0x20E}},
    {22070, 2, {0x480, 0x482, 0xE49C, 0xAFD0, 0xA490, 0xA790, 0xA49E, 0xA794, 0xA494, 0xA494, 0xEFD4, 0xA014, 0x514, 0x8A4, 0x1024, 0x44}},
    {22071, 2, {0x40, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4820, 0x4924, 0x78A8, 0x4924, 0x20, 0x3FE, 0x20, 0x20}},
    {22072, 2, {0x80, 0x80, 0x79FE, 0x4B54, 0x4D54, 0x4954, 0x4BFE, 0x4954, 0x4954, 0x4954, 0x4FFE, 0x7800, 0x4954, 0x12A, 0x22A, 0x0}},
    {22073, 2, {0x40, 0x40, 0xF7FC, 0x90A0, 0x9514, 0x9208, 0x97FC, 0x9A0A, 0x93F8, 0x9208, 0xF3F8, 0x9040, 0x248, 0x444, 0x944, 0x80}},
    {22074, 2, {0x38, 0x3C0, 0xF080, 0x97FE, 0x9110, 0x9208, 0x95F6, 0x9910, 0x91F0, 0x9000, 0xF7FC, 0x9404, 0x5F4, 0x514, 0x5F4, 0x40C}},
    {22075, 2, {0x40, 0x7FC, 0xF040, 0x93F8, 0x9000, 0x93F8, 0x9208, 0x93F8, 0x9110, 0x9FFE, 0xF000, 0x93F8, 0x208, 0x208, 0x3F8, 0x208}},
    {22076, 2, {0x3E7C, 0x2244, 0x3E7C, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {22077, 2, {0x0, 0x3DE, 0x7A52, 0x4A52, 0x4BDE, 0x4800, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x3FE, 0x20, 0x20, 0x20}},
    {22078, 2, {0x0, 0x3FE, 0x7850, 0x4BFE, 0x4A52, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x20, 0x3FE, 0x20, 0x20}},
    {22079, 2, {0x0, 0x7FC, 0xF444, 0x9554, 0x94E4, 0x9444, 0x97FC, 0x9040, 0x97FC, 0x9040, 0xF040, 0x9FFE, 0x0, 0x524, 0x492, 0x892}},
    {22080, 2, {0x0, 0x3DE, 0x7A52, 0x4BDE, 0x4A10, 0x4A52, 0x49CE, 0x4800, 0x4888, 0x4BFE, 0x7888, 0x4888, 0x7FE, 0x88, 0x104, 0x202}},
    {22081, 2, {0x0, 0x7FE, 0xF090, 0x9090, 0x939C, 0x9204, 0x9204, 0x939C, 0x9090, 0x9090, 0xF7FE, 0x9040, 0x524, 0x50A, 0x90A, 0xF8}},
    {22082, 2, {0x108, 0x90, 0xF7FE, 0x9090, 0x93FC, 0x9294, 0x931C, 0x9204, 0x93FC, 0x9204, 0xF3FC, 0x9008, 0x7FE, 0x108, 0xA8, 0x10}},
    {22083, 2, {0x3C, 0x3E0, 0x7924, 0x48A8, 0x4BFE, 0x48A8, 0x4924, 0x4A02, 0x49FC, 0x4924, 0x7924, 0x49FC, 0x124, 0x124, 0x1FC, 0x104}},
    {22084, 2, {0x80, 0x3DE, 0x788A, 0x49CA, 0x488A, 0x4BD2, 0x48A6, 0x4840, 0x4888, 0x49F0, 0x7824, 0x4BFE, 0x22, 0x124, 0x2A2, 0x40}},
    {22085, 2, {0x4, 0x3E, 0xF7C0, 0x9244, 0x9128, 0x93F8, 0x9208, 0x93FC, 0x9204, 0x93FE, 0xF402, 0x96AA, 0xAAA, 0xA02, 0x1014, 0x8}},
    {22086, 2, {0x0, 0x7BC, 0x7108, 0x5528, 0x57BC, 0x5318, 0x55AA, 0x5946, 0x5000, 0x53F8, 0x7208, 0x5208, 0x3F8, 0x208, 0x208, 0x3F8}},
    {22087, 2, {0x20, 0x1FC, 0x7888, 0x4850, 0x4BFE, 0x4800, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x1FC, 0x20, 0x3FE, 0x0}},
    {22088, 2, {0x414, 0x212, 0xEF92, 0xA010, 0xA07E, 0xAF90, 0xA890, 0xA890, 0xAF90, 0xA228, 0xEB28, 0xAAA8, 0x1228, 0x24A, 0xA4A, 0x486}},
    {22089, 2, {0x8, 0x788, 0x7088, 0x5110, 0x5FDE, 0x5494, 0x54A4, 0x5794, 0x5494, 0x5794, 0x7494, 0x54C8, 0x788, 0xC94, 0xA4, 0xC2}},
    {22090, 2, {0x0, 0x3FC, 0x7848, 0x4830, 0x4BFE, 0x4852, 0x4894, 0x4B50, 0x4820, 0x4BFE, 0x7A52, 0x4A8A, 0x376, 0x252, 0x272, 0x206}},
    {22091, 2, {0x208, 0x108, 0x77C8, 0x5010, 0x579E, 0x5494, 0x57A4, 0x5014, 0x5794, 0x5094, 0x7114, 0x51C8, 0x708, 0x114, 0x524, 0x242}},
    {22092, 2, {0x208, 0x110, 0xF000, 0x97FC, 0x9444, 0x9554, 0x94E4, 0x9444, 0x97FC, 0x9000, 0xF3F8, 0x9208, 0x3F8, 0x208, 0x3F8, 0x208}},
    {22093, 2, {0x140, 0x120, 0xF3FE, 0x9220, 0x9620, 0x9BFC, 0x9220, 0x9220, 0x93FC, 0x9220, 0xF220, 0x93FE, 0x200, 0x524, 0x492, 0x892}},
    {22094, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4800, 0x4BFE, 0x4A02, 0x49FC, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x104, 0x88, 0x3FE, 0x0}},
    {22095, 2, {0x40, 0xA0, 0xF118, 0x96E6, 0x9000, 0x93F8, 0x9208, 0x93F8, 0x9000, 0x97BC, 0xF084, 0x94A4, 0x294, 0x4A4, 0x294, 0x108}},
    {22096, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x0, 0x3FFC, 0x100, 0x100, 0x7FFE, 0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244}},
    {22097, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4820, 0x4924, 0x48A8, 0x7924, 0x4820, 0x3FE, 0x20, 0x20, 0x20}},
    {22098, 2, {0x20, 0x124, 0x78A8, 0x4BFE, 0x48A8, 0x4924, 0x4A22, 0x4904, 0x4904, 0x49DE, 0x7A44, 0x4D54, 0x9E, 0x104, 0x204, 0x404}},
    {22099, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x4A78, 0x4BC4, 0x4A3C, 0x4A00, 0x4A50, 0x7B54, 0x4B54, 0x5DC, 0x450, 0xBFE, 0x0}},
    {22100, 2, {0x20, 0x7A4, 0xF0A8, 0x9292, 0x9114, 0x9208, 0x95F4, 0x9802, 0x93F8, 0x9208, 0xF208, 0x93F8, 0x208, 0x110, 0xFFE, 0x0}},
    {22101, 2, {0xF8, 0x3F20, 0x940, 0x1FE0, 0x1020, 0x1FF0, 0x1010, 0x1FFC, 0x2004, 0x2AA4, 0x400C, 0xBFF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {22102, 2, {0x0, 0xFBE, 0xF208, 0x971C, 0x9208, 0x9FBE, 0x9040, 0x90A0, 0x9318, 0x9C46, 0xF020, 0x93F8, 0x8, 0x10, 0x20, 0x40}},
    {22103, 2, {0xA0, 0x4A4, 0xF2A8, 0x90A0, 0x97FC, 0x9110, 0x90A0, 0x97FC, 0x9040, 0x93F8, 0xF040, 0x97FC, 0xA0, 0x110, 0x208, 0xC06}},
    {22104, 2, {0x0, 0xFFE, 0xE800, 0xAA28, 0xA948, 0xABEE, 0xA892, 0xA884, 0xAAA0, 0xAAA8, 0xEBE8, 0xA888, 0x894, 0x914, 0x924, 0x1242}},
    {22105, 2, {0x40, 0xA0, 0xF110, 0x9248, 0x9DF6, 0x90A0, 0x9248, 0x92A8, 0x93F8, 0x9040, 0xF7FC, 0x9484, 0x524, 0x5F4, 0x414, 0x40C}},
    {22106, 2, {0x0, 0x3FC, 0x7804, 0x4BFC, 0x4804, 0x4BFC, 0x4800, 0x4BBC, 0x4924, 0x4FBC, 0x7808, 0x4BFE, 0x108, 0x88, 0xA8, 0x10}},
    {22107, 2, {0x20, 0x20, 0x793C, 0x4920, 0x4920, 0x4FFE, 0x4800, 0x4820, 0x4AAA, 0x4A22, 0x7BFE, 0x4A72, 0x2AA, 0x222, 0x3FE, 0x2}},
    {22108, 2, {0x100, 0x3F0, 0xF410, 0x9FFC, 0x9444, 0x97FC, 0x9444, 0x97FC, 0x9000, 0x9FFE, 0xF000, 0x93F8, 0x208, 0x3F8, 0x208, 0x3F8}},
    {22109, 2, {0x108, 0x108, 0xE214, 0xA294, 0xA4BC, 0xAF08, 0xA110, 0xA224, 0xA4BE, 0xAF8A, 0xE088, 0xA02C, 0xAAA, 0xACA, 0x828, 0x10}},
    {22110, 2, {0x40, 0x40, 0xF0A0, 0x9110, 0x9208, 0x95F6, 0x9800, 0x97BC, 0x94A4, 0x94A4, 0xF7BC, 0x9210, 0x210, 0x528, 0x944, 0x1082}},
    {22111, 2, {0x100, 0x80, 0x3FFE, 0x2248, 0x24FE, 0x2D90, 0x36FC, 0x2490, 0x24FC, 0x2490, 0x24FE, 0x2480, 0x4FFC, 0x4804, 0x8FFC, 0x804}},
    {22112, 2, {0x10, 0x410, 0x7A7C, 0x4810, 0x48FE, 0x4844, 0x4E28, 0x4AFE, 0x4A10, 0x4A7C, 0x7A10, 0x4AFE, 0x210, 0x210, 0x5FE, 0x800}},
    {22113, 2, {0x100, 0x1F8, 0x7A08, 0x4FFE, 0x4A88, 0x4B24, 0x4BFE, 0x4A00, 0x4AFC, 0x4A00, 0x7AFC, 0x4A00, 0x2FC, 0x484, 0x4FC, 0x884}},
    {22114, 2, {0x80, 0x100, 0xF7FC, 0x9444, 0x9554, 0x9444, 0x97FC, 0x94E4, 0x9554, 0x9404, 0xF040, 0x97FE, 0xA0, 0x110, 0x208, 0xC06}},
    {22115, 2, {0x0, 0x3F8, 0x7AA8, 0x4AA8, 0x4BF8, 0x4900, 0x4BFC, 0x4C84, 0x4BE4, 0x4AA4, 0x7AA4, 0x4BE4, 0x94, 0x7F4, 0x14, 0x8}},
    {22116, 2, {0x110, 0x110, 0xF7BC, 0x9110, 0x93B8, 0x9554, 0x9912, 0x9000, 0x93F8, 0x9000, 0xF7FC, 0x9040, 0x248, 0x444, 0x942, 0x80}},
    {22117, 2, {0x90, 0x3FC, 0x7A94, 0x4BFC, 0x4A94, 0x4BFC, 0x4800, 0x4BFC, 0x4A00, 0x4AF8, 0x7A00, 0x4BFE, 0x520, 0x514, 0x548, 0x986}},
    {22118, 2, {0x20, 0x13C, 0x7920, 0x4FFE, 0x4810, 0x4814, 0x4BFE, 0x4A10, 0x4BF0, 0x4A54, 0x7A54, 0x4AD4, 0x368, 0x44A, 0x896, 0x322}},
    {22119, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x4820, 0x7BFE, 0x4A22, 0x22A, 0x2FA, 0x20A, 0x206}},
    {22120, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x120, 0x110, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006, 0x3E7C, 0x2244, 0x2244, 0x3E7C}},
    {22121, 2, {0x0, 0x7FFC, 0x100, 0x3D78, 0x2548, 0x2548, 0x3D78, 0x100, 0x7FFC, 0x100, 0x3D78, 0x2548, 0x2548, 0x3D78, 0x100, 0xFFFE}},
    {22122, 2, {0x3F8, 0x208, 0xF208, 0x93F8, 0x9000, 0x97BC, 0x94A4, 0x94A4, 0x97BC, 0x9040, 0xFFFE, 0x9160, 0x250, 0x448, 0x1846, 0x40}},
    {22123, 2, {0x80, 0x40, 0x7BFC, 0x4908, 0x4890, 0x4FFE, 0x4800, 0x4BFC, 0x4A04, 0x4BFC, 0x7A04, 0x4BFC, 0x40, 0x524, 0x50A, 0x8FA}},
    {22124, 2, {0x210, 0x210, 0xF3DE, 0x9528, 0x9884, 0x9000, 0x97FC, 0x9040, 0x9248, 0x9248, 0xF248, 0x9554, 0x8E2, 0x40, 0xFFE, 0x0}},
    {22125, 2, {0x108, 0x208, 0x77C8, 0x5450, 0x57DE, 0x5464, 0x57D4, 0x5214, 0x5114, 0x57D4, 0x7214, 0x53C8, 0x248, 0x454, 0x554, 0x8A2}},
    {22126, 2, {0x0, 0x3FC, 0x7A94, 0x4A94, 0x4BFC, 0x4800, 0x4FFE, 0x4800, 0x4BFC, 0x4A04, 0x7BFC, 0x48A2, 0x114, 0x308, 0xD44, 0x182}},
    {22127, 2, {0x3C, 0x3C0, 0x7044, 0x5224, 0x5108, 0x57FE, 0x5442, 0x52A4, 0x528A, 0x547A, 0x7080, 0x50F8, 0x108, 0x290, 0x60, 0x79E}},
    {22128, 2, {0x40, 0x20, 0x77FE, 0x5028, 0x5224, 0x527E, 0x54C8, 0x5748, 0x517E, 0x5248, 0x74C8, 0x577E, 0x148, 0x248, 0x47E, 0x840}},
    {22129, 2, {0x40, 0x7C, 0x7840, 0x4BFE, 0x4A42, 0x4A78, 0x4BC4, 0x4A3C, 0x4A00, 0x4BFE, 0x7A40, 0x4BA4, 0x258, 0x5B4, 0x452, 0x9B0}},
    {22130, 2, {0x40, 0xA0, 0xF318, 0x9DF6, 0x9000, 0x97FC, 0x9554, 0x94E4, 0x97FC, 0x9000, 0xF3F8, 0x9208, 0x3F8, 0x208, 0x3F8, 0x208}},
    {22131, 2, {0x20, 0x3E, 0x7820, 0x4BFE, 0x4A22, 0x4BF8, 0x4A22, 0x4AFE, 0x4A84, 0x4AFC, 0x7A00, 0x4AFC, 0x220, 0x5FE, 0x448, 0x986}},
    {22132, 2, {0x20, 0x1FC, 0x7820, 0x4888, 0x4BFE, 0x4888, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x88, 0x104, 0x202}},
    {22133, 2, {0x2110, 0x17FE, 0x1080, 0x3F8, 0xF208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {22134, 2, {0x88, 0x3FE, 0x7888, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4880, 0x49FE, 0x7A22, 0x4D52, 0x102, 0x1FA, 0xA, 0x4}},
    {22135, 2, {0x210, 0x110, 0x7790, 0x503E, 0x54A2, 0x5344, 0x5F90, 0x5010, 0x5790, 0x5490, 0x77A8, 0x54A8, 0x4A8, 0x7C4, 0x44, 0x82}},
    {22136, 2, {0x200, 0x23E, 0xE208, 0xBFD0, 0xA23E, 0xAAA2, 0xAAA2, 0xAABE, 0xAAA2, 0xAFBE, 0xE222, 0xA222, 0x2BE, 0x300, 0x214, 0x22}},
    {22137, 2, {0x40, 0x248, 0xF150, 0x97FE, 0x9402, 0x9804, 0x93F8, 0x9208, 0x93F8, 0x9000, 0xF7FC, 0x9444, 0x7FC, 0x444, 0x7FC, 0x404}},
    {22138, 2, {0x200, 0x102, 0x77DC, 0x5010, 0x5450, 0x5290, 0x57DE, 0x5114, 0x5114, 0x57D4, 0x7114, 0x5594, 0x554, 0x924, 0x524, 0x244}},
    {22139, 2, {0x80, 0x40, 0xF7FE, 0x9512, 0x9BFC, 0x9110, 0x97FC, 0x9110, 0x97FE, 0x9110, 0xF248, 0x95F4, 0x842, 0x40, 0x7FC, 0x0}},
    {22140, 2, {0x10, 0x8, 0x7788, 0x54BE, 0x5480, 0x5494, 0x5788, 0x547E, 0x5408, 0x5788, 0x76BE, 0x5A88, 0xA88, 0xB88, 0x1288, 0x8}},
    {22141, 2, {0x800, 0xFF00, 0x8FE, 0x7E10, 0x10, 0x7E30, 0x4238, 0x7E54, 0x2492, 0xFF10, 0x10, 0x7E10, 0x4210, 0x4200, 0x7EFE, 0x4200}},
    {22142, 2, {0x40, 0x20, 0x7BFE, 0x4A8A, 0x4944, 0x4A22, 0x49FC, 0x4888, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x1FC, 0x104, 0x1FC, 0x104}},
    {22143, 2, {0x3DE, 0x42, 0x7A52, 0x494A, 0x4A52, 0x4820, 0x4BFE, 0x4888, 0x4888, 0x4954, 0x7A22, 0x4820, 0x3FE, 0x20, 0x20, 0x20}},
    {22144, 2, {0x40, 0x20, 0x7BFE, 0x4A22, 0x4954, 0x494A, 0x4A3A, 0x4800, 0x49FC, 0x4954, 0x7BFE, 0x4800, 0x1FC, 0x20, 0xA0, 0x40}},
    {22145, 2, {0x0, 0x3DE, 0x7842, 0x4A52, 0x494A, 0x4A52, 0x4890, 0x4908, 0x4BFE, 0x4D10, 0x79FE, 0x4910, 0x1FE, 0x110, 0x1FE, 0x100}},
    {22146, 2, {0x10, 0x7D0, 0x7510, 0x57DE, 0x5450, 0x57E8, 0x5504, 0x57C4, 0x5000, 0x53F8, 0x72A8, 0x52A8, 0x2A8, 0x2A8, 0xFFE, 0x0}},
    {22147, 2, {0x400, 0x2FE, 0x7AAA, 0x48AA, 0x48FE, 0x4810, 0x4E54, 0x4A92, 0x4A10, 0x4A54, 0x7A92, 0x4A10, 0x210, 0x500, 0x8FE, 0x0}},
    {22148, 2, {0x88, 0x3FE, 0x78A8, 0x4890, 0x49FE, 0x4B20, 0x49FC, 0x4920, 0x49FC, 0x4920, 0x79FE, 0x4900, 0x3FC, 0x88, 0x70, 0x38E}},
    {22149, 2, {0x0, 0x7FC, 0xE040, 0xAFFE, 0xA842, 0xA358, 0xA040, 0xA358, 0xA000, 0xAFFE, 0xE040, 0xA7FC, 0x4A4, 0x4A4, 0x4A4, 0x40C}},
    {22150, 2, {0x108, 0x7FE, 0xF108, 0x9020, 0x97FE, 0x9000, 0x91F8, 0x9108, 0x91F8, 0x9000, 0xF7FE, 0x9402, 0x5FA, 0x50A, 0x5FA, 0x406}},
    {22151, 2, {0x110, 0x110, 0x7110, 0x57BC, 0x5110, 0x5110, 0x5FFE, 0x52A8, 0x52A8, 0x56EC, 0x7AAA, 0x52AA, 0x2A8, 0x5A8, 0x448, 0x898}},
    {22152, 2, {0x0, 0xFFE, 0xE808, 0xABC8, 0xAA4C, 0xABCA, 0xAA48, 0xABDE, 0xA808, 0xABC8, 0xEA48, 0xABD4, 0xA54, 0xBD4, 0xA54, 0x12E2}},
    {22153, 2, {0x284, 0x284, 0xEAA4, 0xA6C4, 0xA29E, 0xAFE4, 0xA004, 0xA454, 0xA28C, 0xAFE4, 0xE104, 0xA7C4, 0x104, 0x1C4, 0xE14, 0x8}},
    {22154, 2, {0x80, 0x3F8, 0xF208, 0x93F8, 0x9208, 0x93F8, 0x9000, 0x97FC, 0x9444, 0x97FC, 0xF444, 0x97FC, 0x0, 0xFFE, 0x110, 0x210}},
    {22155, 2, {0x40, 0x7FC, 0x7040, 0x53FC, 0x5000, 0x57FE, 0x5002, 0x53F8, 0x5040, 0x57FE, 0x7000, 0x57FE, 0x8, 0x7FE, 0x4A8, 0x798}},
    {22156, 2, {0x80, 0x40, 0x77FC, 0x5110, 0x50A4, 0x5F58, 0x5554, 0x5552, 0x5B58, 0x5000, 0x7208, 0x53F8, 0x208, 0x3F8, 0x208, 0x408}},
    {22157, 2, {0x40, 0x3F8, 0x7048, 0x57FE, 0x5048, 0x53F8, 0x5040, 0x57FE, 0x5000, 0x52A4, 0x7452, 0x53F8, 0x2A8, 0x2A8, 0xFFE, 0x0}},
    {22158, 2, {0x40, 0x7FE, 0xF000, 0x91F8, 0x9108, 0x91F8, 0x9000, 0x97FE, 0x9402, 0x93F8, 0xF180, 0x9644, 0x1B8, 0x668, 0x1A6, 0xE60}},
    {22159, 2, {0x40, 0x7FC, 0xF040, 0x97FE, 0x9402, 0x9BFC, 0x9248, 0x93F8, 0x9248, 0x93F8, 0xF000, 0x97FC, 0x244, 0x278, 0x540, 0x8FE}},
    {22160, 2, {0x124, 0xA8, 0x7BFE, 0x4A8A, 0x48F8, 0x490C, 0x49F0, 0x4902, 0x48FE, 0x4800, 0x79FC, 0x4904, 0x1FC, 0x104, 0x1FC, 0x104}},
    {22161, 2, {0x3C, 0x7C0, 0xF040, 0x9FFE, 0x9040, 0x97FC, 0x9554, 0x94E4, 0x97FC, 0x9040, 0xF7FC, 0x9040, 0xFFE, 0x0, 0x524, 0x892}},
    {22162, 2, {0x40, 0x20, 0x77FE, 0x5488, 0x57DE, 0x5488, 0x55DC, 0x56AA, 0x5488, 0x5440, 0x7488, 0x55F0, 0x420, 0x448, 0x9FC, 0x1084}},
    {22163, 2, {0x80, 0x40, 0x77FE, 0x5402, 0x5A24, 0x53BC, 0x54A4, 0x5AA8, 0x5510, 0x52E8, 0x7404, 0x5BFA, 0x40, 0x248, 0x444, 0xC0}},
    {22164, 2, {0x40, 0x7FC, 0x7040, 0x57FE, 0x5402, 0x5BFC, 0x5248, 0x53F8, 0x5248, 0x53F8, 0x7108, 0x57FC, 0x244, 0x278, 0x540, 0x8FE}},
    {22165, 2, {0x100, 0x3F0, 0x7410, 0x5FFC, 0x5444, 0x57FC, 0x5444, 0x57FC, 0x5524, 0x5892, 0x73F8, 0x5208, 0x3F8, 0x208, 0x3F8, 0x208}},
    {22166, 2, {0x88, 0x3DE, 0x7888, 0x49DC, 0x4888, 0x4BDE, 0x4888, 0x4BFC, 0x4804, 0x49FC, 0x7804, 0x4BFC, 0x40, 0x2A4, 0x28A, 0x47A}},
    {22167, 2, {0x3F8, 0x208, 0xF3F8, 0x9208, 0x93F8, 0x9110, 0x97FC, 0x9110, 0x9FFE, 0x9110, 0xF248, 0x9554, 0xE0, 0x150, 0x248, 0xC0}},
    {22168, 2, {0x7FC, 0x80, 0xE3F8, 0xA208, 0xA3F8, 0xA208, 0xAFFE, 0xA882, 0xA548, 0xA514, 0xE8F4, 0xA100, 0x3F8, 0x508, 0xF0, 0xF0E}},
    {22169, 2, {0x20, 0x120, 0x793C, 0x4920, 0x4FFE, 0x4890, 0x4A94, 0x4B6C, 0x4A04, 0x4BFC, 0x7A94, 0x4A94, 0x36C, 0x204, 0x3FC, 0x4}},
    {22170, 2, {0x0, 0x3E7C, 0x2244, 0x3E7C, 0x0, 0x3FFC, 0x2100, 0x3FF8, 0x2008, 0x3FF8, 0x2100, 0x3FFC, 0x0, 0x3E7C, 0x2244, 0x3E7C}},
    {22171, 2, {0x444, 0x4E4, 0xE8A8, 0xAAAA, 0xAEEE, 0xA4A4, 0xAAAA, 0xAEEE, 0xA242, 0xA040, 0xEFFE, 0xA0E0, 0x150, 0x248, 0xC46, 0x40}},
    {22172, 2, {0x1FC, 0x124, 0x79AC, 0x4924, 0x49FC, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x4954, 0x7A2A, 0x4820, 0x1FC, 0x20, 0x3FE, 0x0}},
    {22173, 2, {0x40, 0x20, 0x77FE, 0x5488, 0x55FC, 0x5488, 0x57FE, 0x5420, 0x55FC, 0x5524, 0x75FC, 0x5524, 0x9FC, 0x888, 0x1104, 0x202}},
    {22174, 2, {0x100, 0xFFFE, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x0, 0x7C7C, 0x4444, 0x7C7C}},
    {22175, 2, {0x50, 0x190, 0x789E, 0x4BEA, 0x488A, 0x49D2, 0x4A8A, 0x48A4, 0x4850, 0x4888, 0x7B26, 0x48A8, 0x70, 0xA8, 0x124, 0x60}},
    {22176, 2, {0x202, 0x5E2, 0x74A2, 0x56AA, 0x552A, 0x566A, 0x510A, 0x528A, 0x544A, 0x5BAA, 0x710A, 0x57CA, 0x122, 0x542, 0x1EA, 0xE04}},
    {22177, 2, {0x508, 0x508, 0xEFBE, 0xA508, 0xA708, 0xA208, 0xAFBE, 0xAA80, 0xAA88, 0xAF88, 0xE23E, 0xAF88, 0x208, 0x208, 0x23E, 0x200}},
    {22178, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x9452, 0x2FE8, 0x4444, 0x1FF0, 0x440, 0x7FFC, 0x508, 0x1890, 0xEA60, 0xC18}},
    {22179, 2, {0x3EF8, 0x2288, 0x3EF8, 0x0, 0xFFFE, 0x100, 0x1FF0, 0x1010, 0x1110, 0x2C0, 0xC30, 0x7008, 0x0, 0x3EF8, 0x2288, 0x3EF8}},
    {22180, 2, {0x20, 0x7FE, 0x7488, 0x57DE, 0x5488, 0x55DC, 0x56AA, 0x548C, 0x55F0, 0x5420, 0x75FC, 0x5420, 0x7FE, 0x820, 0x8A0, 0x1040}},
    {22181, 2, {0x110, 0x110, 0x77FC, 0x5110, 0x51F0, 0x5208, 0x52EA, 0x56AC, 0x52A8, 0x52AA, 0x76EA, 0x5A06, 0x0, 0x2A4, 0x252, 0x452}},
    {22182, 2, {0x0, 0x7FE, 0x7444, 0x5598, 0x5488, 0x57DE, 0x5488, 0x55DC, 0x56AA, 0x5488, 0x7420, 0x5520, 0x53C, 0x920, 0x920, 0x17FE}},
    {22183, 2, {0x40, 0x7C, 0x7840, 0x4BFC, 0x4A44, 0x4BF0, 0x4A44, 0x4AFC, 0x4AA8, 0x4AF8, 0x7AA8, 0x4AF8, 0x200, 0x5FC, 0x554, 0xBFE}},
    {22184, 2, {0x210, 0x110, 0x77DE, 0x5010, 0x545E, 0x5282, 0x57DE, 0x5010, 0x57DE, 0x5450, 0x77DE, 0x5450, 0x7DE, 0x450, 0x452, 0x4CE}},
    {22185, 2, {0x414, 0x412, 0xE9FE, 0xAA10, 0xB2FE, 0xBC92, 0xA4FE, 0xA892, 0xB2FE, 0xBE92, 0xE204, 0x81FE, 0x2A44, 0x2A24, 0x2024, 0xC}},
    {22186, 2, {0x3DE, 0x252, 0x7BDE, 0x4A52, 0x4BDE, 0x4A42, 0x4A7A, 0x4A8A, 0x4B12, 0x4ADA, 0x7A8A, 0x4ADA, 0x28A, 0x2FA, 0x20A, 0x204}},
    {22187, 2, {0x400, 0x23C, 0xEFA4, 0xA024, 0xA8BC, 0xA524, 0xAFA4, 0xA23C, 0xA224, 0xAFA4, 0xE23C, 0xAB10, 0xA98, 0x12A8, 0xA2A, 0x446}},
    {22188, 2, {0x100, 0x13E, 0x7508, 0x55D0, 0x553E, 0x5522, 0x5FE2, 0x503E, 0x5122, 0x557E, 0x7562, 0x55A2, 0x8BE, 0x100, 0x214, 0xC22}},
    {22189, 2, {0x800, 0xFF7E, 0x808, 0x7E08, 0x10, 0x7E34, 0x4252, 0x7E92, 0x2410, 0xFF10, 0x7C, 0x7E44, 0x4244, 0x4244, 0x7E7C, 0x4244}},
    {22190, 2, {0x1100, 0x2BDE, 0x4A52, 0x73D4, 0x2A52, 0x4BD2, 0x7A1A, 0x1294, 0x2350, 0xCA40, 0x3FFC, 0x2004, 0x27E4, 0x2424, 0x27E4, 0x200C}},
    {22191, 2, {0x7FC, 0x40, 0xF7FE, 0x9442, 0x9B5C, 0x9040, 0x975C, 0x9120, 0x93FC, 0x9620, 0xFBFC, 0x9220, 0x3FC, 0x220, 0x3FE, 0x200}},
    {22192, 2, {0x40, 0x20, 0x7BFE, 0x4A48, 0x4BFE, 0x4A48, 0x4AEC, 0x4B5A, 0x4A48, 0x4BFE, 0x7A20, 0x4A7C, 0x2C4, 0x344, 0x47C, 0x844}},
    {22193, 2, {0x108, 0x1EC, 0x710A, 0x57E8, 0x5528, 0x55DE, 0x5728, 0x54EA, 0x540A, 0x57EA, 0x740C, 0x57EC, 0xA2A, 0xBEA, 0x1156, 0x7E2}},
    {22194, 2, {0x2000, 0x11DC, 0xFD54, 0x154, 0x7DDC, 0x4400, 0x7DFC, 0x124, 0x7DFC, 0x924, 0x11FC, 0x1C20, 0xF3FE, 0x1020, 0x5020, 0x2020}},
    {22195, 2, {0x3AB8, 0x2108, 0x3AB8, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0x9104, 0x1FF0, 0x2100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {22196, 2, {0x3E7C, 0x2244, 0x3E7C, 0x0, 0x3FFE, 0x2000, 0x2E20, 0x223C, 0x3FA8, 0x2948, 0x2F28, 0x2928, 0x2F10, 0x49D0, 0x7F28, 0x8144}},
    {22197, 2, {0x1F0, 0x210, 0x77FC, 0x5204, 0x53FC, 0x5224, 0x53B8, 0x5222, 0x51FE, 0x5210, 0x77FC, 0x5244, 0x3FC, 0xD0, 0x14A, 0x63E}},
    {22198, 2, {0x3DE, 0x252, 0x7BDE, 0x4A52, 0x4BDE, 0x4A52, 0x4BDE, 0x498C, 0x4A52, 0x4840, 0x7FFE, 0x4888, 0x190, 0x60, 0xD8, 0x304}},
    {22199, 2, {0x40, 0x7FC, 0xF000, 0x93B8, 0x92A8, 0x93B8, 0x9110, 0x97FC, 0x9110, 0x97FC, 0xF110, 0x9FFE, 0x128, 0x310, 0xD48, 0x186}},
    {22200, 2, {0x8, 0xFE8, 0xE928, 0xAB68, 0xA9AE, 0xA928, 0xAFE8, 0xA108, 0xAFE8, 0xA11E, 0xE112, 0xBFF2, 0x12, 0x1552, 0x155E, 0x2012}},
    {22201, 2, {0x408, 0x40E, 0xE788, 0xAA3E, 0xA22A, 0xA23C, 0xBFAA, 0xA226, 0xA220, 0xAABE, 0xEAA8, 0xABBA, 0xCAC, 0x5E, 0x48, 0x98}},
    {22202, 2, {0x1FC, 0x954, 0x75FC, 0x5488, 0x5050, 0x53FE, 0x5C40, 0x54A2, 0x5534, 0x5458, 0x75B4, 0x5452, 0x592, 0x450, 0xA20, 0x11FE}},
    {22203, 2, {0x0, 0x7BDE, 0x4892, 0x4912, 0x4BD2, 0x4A52, 0x7A5E, 0x3C0, 0x7A5E, 0x4BD2, 0x4A52, 0x4A52, 0x4BD2, 0x781E, 0x4992, 0x240}},
    {22204, 2, {0x3C, 0x7C0, 0x7244, 0x5128, 0x57FC, 0x54A4, 0x57FC, 0x5008, 0x5788, 0x54FE, 0x7788, 0x54A8, 0x798, 0x508, 0x5A8, 0x690}},
    {22205, 2, {0x110, 0x7D0, 0x711E, 0x57A4, 0x5154, 0x5FD4, 0x5208, 0x5394, 0x54A2, 0x5BFC, 0x7204, 0x53FC, 0x204, 0x3FC, 0x108, 0x204}},
    {22206, 2, {0x88, 0x3FE, 0x7888, 0x4BDE, 0x4A52, 0x4BDE, 0x48A0, 0x4890, 0x49FE, 0x4920, 0x7BFC, 0x4D20, 0x1FC, 0x120, 0x1FE, 0x100}},
    {22207, 2, {0x108, 0x7FE, 0x7128, 0x53FE, 0x5220, 0x57FC, 0x5A20, 0x53FC, 0x5220, 0x53FE, 0x7080, 0x531C, 0x204, 0x39C, 0x204, 0x3FC}},
    {22208, 2, {0x208, 0x208, 0xE27E, 0xAF88, 0xA23E, 0xAFAA, 0xAABE, 0xAFAA, 0xAABE, 0xAF8A, 0xE27E, 0xAF84, 0x27E, 0x224, 0x214, 0x20C}},
    {22209, 2, {0x3FC, 0x108, 0x79F8, 0x4908, 0x49F8, 0x490E, 0x4FF8, 0x4808, 0x4FFE, 0x4A94, 0x7B9C, 0x4A94, 0x39C, 0x2D6, 0x7BC, 0x84}},
    {22210, 2, {0x3EF8, 0x2288, 0x3EF8, 0x0, 0xFFFE, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010, 0x3EF8, 0x2288, 0x3EF8}},
    {22211, 2, {0x414, 0x212, 0xEFA0, 0xA53E, 0xA564, 0xAAA4, 0xA03E, 0xA224, 0xAFA4, 0xA23E, 0xEB24, 0xAAA4, 0x12A4, 0x23E, 0xA20, 0x420}},
    {22212, 2, {0x418, 0x214, 0xEF90, 0xA53E, 0xAAA8, 0xADE8, 0xA8BE, 0xAFA8, 0xA228, 0xAFBE, 0xEAA8, 0xAAA8, 0xBA8, 0x8BE, 0xAA0, 0x920}},
    {22213, 2, {0x0, 0xEEFE, 0xAA82, 0xAA82, 0xEEFE, 0xA4, 0xFEA4, 0x92FE, 0xFEA4, 0x92A4, 0xFEFE, 0x10A8, 0xFEAA, 0x1124, 0x1132, 0x1220}},
    {22214, 2, {0x88, 0x3FE, 0x7888, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x4BFE, 0x4A2A, 0x4AFA, 0x7A26, 0x49FC, 0x124, 0x1FC, 0x22, 0x3FE}},
    {22215, 2, {0xFBE, 0x0, 0x77BC, 0x54A4, 0x56B4, 0x54A4, 0x5040, 0x57FE, 0x54A0, 0x57FC, 0x74A4, 0x57FC, 0x512, 0x9DC, 0x912, 0x11CE}},
    {22216, 2, {0x108, 0x7FE, 0x7108, 0x5390, 0x513C, 0x57D4, 0x52B4, 0x555C, 0x5396, 0x5122, 0x73FC, 0x5000, 0x7FE, 0x80, 0x104, 0x3FE}},
    {22217, 2, {0x0, 0x7FC, 0x74A4, 0x54A4, 0x57FC, 0x5210, 0x52A8, 0x54BE, 0x5F68, 0x523C, 0x74A8, 0x5FBC, 0x28, 0xAA8, 0xABE, 0x20}},
    {22218, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x9292, 0x1EF0, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x488, 0xC50, 0x3530, 0xC60E}},
    {22219, 2, {0x528, 0x7BE, 0x7948, 0x57BE, 0x5318, 0x55AA, 0x5946, 0x53FC, 0x5204, 0x53FC, 0x7204, 0x53FC, 0x204, 0x3FC, 0x108, 0x204}},
    {22220, 2, {0x110, 0xFFE, 0xE510, 0xA400, 0xAF06, 0xB138, 0xAF88, 0xAA88, 0xAABE, 0xAF88, 0xEA9C, 0xAF9A, 0x28, 0x1548, 0x1548, 0x8}},
    {22221, 2, {0x1010, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x2828, 0xFEFE, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444}},
    {22222, 2, {0x44, 0xE28, 0xEAFE, 0xAA92, 0xAED6, 0xAABA, 0xAAFE, 0xAE00, 0xAA7C, 0xAA44, 0xEE44, 0xA07C, 0xA44, 0x944, 0x117C, 0x44}},
    {22223, 2, {0x2810, 0x29FE, 0xFE10, 0x28FC, 0x3800, 0x10FC, 0x7C84, 0x54FC, 0x7C48, 0x11FE, 0x7C00, 0x10FC, 0xFE84, 0x2884, 0x44FC, 0x8284}},
    {22224, 2, {0x108, 0x1EC, 0xE10A, 0xAFEA, 0xA928, 0xABC8, 0xA93E, 0xAFE8, 0xA948, 0xABE8, 0xEA28, 0xAB68, 0xAA8, 0xBF4, 0xAB4, 0x12A2}},
    {22225, 2, {0x7FC, 0x404, 0x77FC, 0x5524, 0x54A8, 0x5524, 0x55FC, 0x5554, 0x55FC, 0x5644, 0x75F4, 0x5554, 0x9F4, 0x844, 0x13F4, 0x8}},
    {22226, 2, {0x110, 0x7FE, 0x7110, 0x57BC, 0x54A4, 0x57BC, 0x54A4, 0x57BC, 0x5444, 0x57FC, 0x7554, 0x55F4, 0x444, 0x4E4, 0x554, 0x448}},
    {22227, 2, {0x80, 0x3DE, 0x788A, 0x49D2, 0x48A6, 0x4BD0, 0x489C, 0x4850, 0x4BFE, 0x4850, 0x7AAA, 0x4BFE, 0x252, 0x2AA, 0x3FE, 0x2}},
    {22228, 2, {0x40, 0x7FC, 0x7248, 0x53F8, 0x5040, 0x5FFE, 0x5AAA, 0x53B8, 0x5110, 0x57FC, 0x7110, 0x4FFE, 0x128, 0x310, 0xD48, 0x186}},
    {22229, 2, {0x790, 0x51E, 0x77A8, 0x5484, 0x57BE, 0x552A, 0x57BE, 0x5000, 0x53F8, 0x5208, 0x73F8, 0x5208, 0x3F8, 0xA0, 0x124, 0x61C}},
    {22230, 2, {0x210, 0x7DE, 0xE928, 0xAFFE, 0xA912, 0xAFFE, 0xA228, 0xA444, 0xA9FE, 0xAF48, 0xE27E, 0xA548, 0xF7E, 0x48, 0x157E, 0x1540}},
    {22231, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22232, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4FC4, 0x4844, 0x4844, 0x4844, 0x4FC4, 0x4804, 0x4824, 0x4824, 0x47E4, 0x4004, 0x4014, 0x4008}},
    {22233, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x7F84, 0x4084, 0x4084, 0x4084, 0x4084, 0x7F84, 0x4084, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22234, 2, {0x0, 0x7FFC, 0x4004, 0x4104, 0x4104, 0x4104, 0x4104, 0x4284, 0x4244, 0x4444, 0x4424, 0x4824, 0x5024, 0x4004, 0x7FFC, 0x4004}},
    {22235, 2, {0x0, 0x0, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x4844, 0x483C, 0x5004, 0x6004, 0x4004, 0x7FFC, 0x4004, 0x0}},
    {22236, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4104, 0x4104, 0x4204, 0x4204, 0x4444, 0x4824, 0x5FF4, 0x4814, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22237, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4FE4, 0x4044, 0x4084, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4104, 0x4504, 0x4204, 0x7FFC, 0x4004}},
    {22238, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x0}},
    {22239, 2, {0x200, 0x200, 0x400, 0x7FFC, 0x4004, 0x4824, 0x4444, 0x4284, 0x4104, 0x4284, 0x4444, 0x4824, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22240, 2, {0x0, 0x7FFC, 0x4004, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4284, 0x4244, 0x4424, 0x4814, 0x5014, 0x4004, 0x7FFC, 0x4004}},
    {22241, 2, {0x0, 0x7FFC, 0x4204, 0x4204, 0x4204, 0x7FFC, 0x4444, 0x4444, 0x4884, 0x4684, 0x4104, 0x4284, 0x4444, 0x4824, 0x7FFC, 0x4004}},
    {22242, 2, {0x0, 0x7FFC, 0x4004, 0x4084, 0x4084, 0x5FF4, 0x4084, 0x4184, 0x4284, 0x4484, 0x4884, 0x5084, 0x4284, 0x4104, 0x7FFC, 0x4004}},
    {22243, 2, {0x0, 0x7FFC, 0x4004, 0x4084, 0x4084, 0x5FF4, 0x4084, 0x4884, 0x4484, 0x4484, 0x4084, 0x4084, 0x4284, 0x4104, 0x7FFC, 0x4004}},
    {22244, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4924, 0x4924, 0x4FE4, 0x4114, 0x4114, 0x4114, 0x40F4, 0x4004, 0x7FFC, 0x4004}},
    {22245, 2, {0x0, 0x7FFC, 0x4004, 0x4204, 0x4104, 0x5FF4, 0x4004, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4454, 0x4834, 0x5004, 0x7FFC, 0x4004}},
    {22246, 2, {0x0, 0x7FFC, 0x4004, 0x4104, 0x4104, 0x5D14, 0x45A4, 0x4544, 0x4924, 0x4914, 0x5114, 0x6504, 0x4204, 0x4004, 0x7FFC, 0x4004}},
    {22247, 2, {0x0, 0x7FFC, 0x4004, 0x4444, 0x4424, 0x4814, 0x5014, 0x6004, 0x47C4, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x7FFC, 0x4004}},
    {22248, 2, {0x0, 0x7FFC, 0x4004, 0x5FF4, 0x4084, 0x4104, 0x4104, 0x4384, 0x4544, 0x4924, 0x5114, 0x6114, 0x4104, 0x4104, 0x7FFC, 0x4004}},
    {22249, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4FE4, 0x4004, 0x4004, 0x5FF4, 0x4204, 0x4444, 0x4824, 0x5FF4, 0x4814, 0x4004, 0x7FFC, 0x4004}},
    {22250, 2, {0x400, 0x800, 0x7FFC, 0x4004, 0x4084, 0x4894, 0x4514, 0x4224, 0x4544, 0x4884, 0x5144, 0x4224, 0x4C24, 0x4004, 0x7FFC, 0x4004}},
    {22251, 2, {0x0, 0x7FFC, 0x4404, 0x4404, 0x47F4, 0x4954, 0x4954, 0x5254, 0x4294, 0x4494, 0x4914, 0x42A4, 0x4444, 0x4004, 0x7FFC, 0x4004}},
    {22252, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4444, 0x4444, 0x4444, 0x7FFC, 0x4004}},
    {22253, 2, {0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4004, 0x4004, 0x5FF4, 0x4484, 0x4484, 0x4484, 0x4894, 0x4894, 0x5074, 0x4004, 0x7FFC, 0x4004}},
    {22254, 2, {0x0, 0x7FFC, 0x4504, 0x4504, 0x4924, 0x5944, 0x6984, 0x4904, 0x4F14, 0x4914, 0x4914, 0x48F4, 0x4804, 0x4004, 0x7FFC, 0x4004}},
    {22255, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x5FF4, 0x4104, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22256, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4384, 0x4344, 0x4524, 0x4914, 0x5114, 0x4104, 0x4104, 0x7FFC, 0x4004}},
    {22257, 2, {0x400, 0x800, 0x7FFC, 0x4204, 0x4204, 0x47F4, 0x4814, 0x5224, 0x4144, 0x4084, 0x4144, 0x4624, 0x5804, 0x4004, 0x7FFC, 0x4004}},
    {22258, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x5FF4, 0x4444, 0x4444, 0x4444, 0x7FFC, 0x4444, 0x4444, 0x4844, 0x5044, 0x7FFC, 0x4004}},
    {22259, 2, {0x0, 0x7FFC, 0x4004, 0x4104, 0x4484, 0x42A4, 0x5224, 0x4844, 0x4444, 0x4284, 0x4104, 0x46C4, 0x5834, 0x4004, 0x7FFC, 0x4004}},
    {22260, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4114, 0x4114, 0x4154, 0x4124, 0x4104, 0x7FFC, 0x4004}},
    {22261, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x4284, 0x4444, 0x4824, 0x549C, 0x6504, 0x4604, 0x4424, 0x4424, 0x43E4, 0x4004, 0x7FFC, 0x4004}},
    {22262, 2, {0x0, 0x7FFC, 0x4004, 0x4444, 0x4824, 0x5114, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22263, 2, {0x0, 0x7FFC, 0x4004, 0x40E4, 0x4F04, 0x4104, 0x4104, 0x5FF4, 0x4304, 0x4584, 0x4944, 0x5124, 0x4104, 0x4004, 0x7FFC, 0x4004}},
    {22264, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x5FF4, 0x4104, 0x4104, 0x49E4, 0x4904, 0x4904, 0x4904, 0x5FF4, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22265, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x4284, 0x4444, 0x4A24, 0x511C, 0x6FE4, 0x4024, 0x4044, 0x4284, 0x4104, 0x4084, 0x7FFC, 0x4004}},
    {22266, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4FE4, 0x4824, 0x4824, 0x4824, 0x4FE4, 0x4824, 0x4004, 0x7FFC, 0x4004}},
    {22267, 2, {0x0, 0x7FFC, 0x4004, 0x5FE4, 0x5024, 0x5024, 0x5FE4, 0x5104, 0x5FF4, 0x5104, 0x5094, 0x5454, 0x5834, 0x4004, 0x7FFC, 0x4004}},
    {22268, 2, {0x0, 0x7FFC, 0x4204, 0x4204, 0x4444, 0x4824, 0x5FF4, 0x4014, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4824, 0x4004, 0x7FFC, 0x4004}},
    {22269, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x5FF4, 0x4104, 0x4104, 0x4FE4, 0x4104, 0x4144, 0x4124, 0x5FF4, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {22270, 2, {0x0, 0x7FFC, 0x4204, 0x4204, 0x47E4, 0x4C44, 0x5284, 0x4104, 0x46C4, 0x783C, 0x4304, 0x4084, 0x4604, 0x4184, 0x7FFC, 0x4004}},
    {22271, 2, {0x0, 0x7FFC, 0x4204, 0x4204, 0x5FF4, 0x4404, 0x4FE4, 0x5424, 0x67E4, 0x4424, 0x47E4, 0x4424, 0x44A4, 0x4444, 0x7FFC, 0x4004}},
    {22272, 2, {0x7FFC, 0x4004, 0x4444, 0x4824, 0x5214, 0x4104, 0x5FF4, 0x4404, 0x4404, 0x47E4, 0x4424, 0x4824, 0x48A4, 0x5044, 0x7FFC, 0x4004}},
    {22273, 2, {0x0, 0x7FFC, 0x4104, 0x5FF4, 0x4004, 0x4FE4, 0x4004, 0x4FE4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4004, 0x7FFC, 0x4004}},
    {22274, 2, {0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4204, 0x4514, 0x59A4, 0x42C4, 0x44A4, 0x59A4, 0x4294, 0x4494, 0x5A84, 0x4104, 0x7FFC, 0x4004}},
    {22275, 2, {0x0, 0x7FFC, 0x4144, 0x4124, 0x5FF4, 0x4104, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4964, 0x4004, 0x7FFC, 0x4004}},
    {22276, 2, {0x0, 0x7FFC, 0x4004, 0x5FF4, 0x4204, 0x4FE4, 0x4424, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4004, 0x7FFC, 0x4004}},
    {22277, 2, {0x3FF0, 0x20, 0x640, 0x180, 0x7FFC, 0x4444, 0x4284, 0x5FF4, 0x4104, 0x4104, 0x7FFC, 0x4104, 0x4104, 0x4104, 0x7FFC, 0x4004}},
    {22278, 2, {0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5FF4, 0x5014, 0x5114, 0x5114, 0x5294, 0x4444, 0x4824, 0x7FFC, 0x4004}},
    {22279, 2, {0x0, 0x7FFC, 0x4104, 0x4284, 0x4444, 0x4BA4, 0x5014, 0x4FE4, 0x4AA4, 0x4AA4, 0x4FE4, 0x4AA4, 0x4AA4, 0x4864, 0x7FFC, 0x4004}},
    {22280, 2, {0x0, 0x7FFC, 0x4924, 0x4544, 0x5FF4, 0x4204, 0x7FFC, 0x4824, 0x57D4, 0x644C, 0x44C4, 0x4424, 0x43E4, 0x4004, 0x7FFC, 0x4004}},
    {22281, 2, {0x0, 0x7FFC, 0x4104, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4444, 0x4284, 0x4FE4, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x7FFC, 0x4004}},
    {22282, 2, {0x7FFC, 0x4104, 0x4FE4, 0x4104, 0x4FE4, 0x4104, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4864, 0x7FFC}},
    {22283, 2, {0x0, 0x7FFC, 0x40A4, 0x4094, 0x7FFC, 0x4084, 0x5E94, 0x5294, 0x52A4, 0x5EA4, 0x4054, 0x46B4, 0x5914, 0x4004, 0x7FFC, 0x4004}},
    {22284, 2, {0x0, 0x7FFC, 0x4104, 0x5114, 0x5114, 0x5FF4, 0x4004, 0x7FFC, 0x4104, 0x5FF4, 0x5294, 0x5294, 0x5034, 0x4004, 0x7FFC, 0x4004}},
    {22285, 2, {0x7FFC, 0x4204, 0x4FE4, 0x4424, 0x5FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4104, 0x5FF4, 0x4104, 0x5104, 0x5FFC, 0x4104, 0x7FFC}},
    {22286, 2, {0x7FFC, 0x4404, 0x4824, 0x5FF4, 0x4014, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4444, 0x4824, 0x7FFC, 0x4004}},
    {22287, 2, {0x0, 0x7FFC, 0x4924, 0x4544, 0x5FF4, 0x4204, 0x7FFC, 0x4824, 0x57D4, 0x604C, 0x47C4, 0x4424, 0x43E4, 0x4004, 0x7FFC, 0x4004}},
    {22288, 2, {0x0, 0x7FFC, 0x4004, 0x5FF4, 0x5294, 0x5294, 0x5FF4, 0x4104, 0x7FFC, 0x4404, 0x47F4, 0x4814, 0x5054, 0x6024, 0x7FFC, 0x4004}},
    {22289, 2, {0x7FFC, 0x4124, 0x5FF4, 0x4104, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4044, 0x5FF4, 0x4844, 0x4544, 0x4084, 0x7FFC}},
    {22290, 2, {0x0, 0x7FFC, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4114, 0x46A4, 0x5A44, 0x4224, 0x7FFC, 0x4004}},
    {22291, 2, {0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5FF4, 0x5014, 0x5FF4, 0x5014, 0x5FF4, 0x5014, 0x5FF4, 0x4444, 0x4824, 0x7FFC}},
    {22292, 2, {0x7FFC, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4204, 0x4444, 0x4FE4, 0x4004, 0x4FE4, 0x4AA4, 0x4AA4, 0x5FF4, 0x4004, 0x7FFC, 0x4004}},
    {22293, 2, {0x7FFC, 0x4104, 0x4104, 0x4FE4, 0x4124, 0x5FF4, 0x4124, 0x4FE4, 0x4104, 0x5FF4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x7FFC}},
    {22294, 2, {0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4104, 0x7FFC, 0x4104, 0x5FF4, 0x5014, 0x57D4, 0x5454, 0x5FF4, 0x4004, 0x7FFC, 0x4004}},
    {22295, 2, {0x7FFC, 0x4204, 0x4444, 0x4FE4, 0x4104, 0x7FFC, 0x4204, 0x5FF4, 0x5454, 0x57D4, 0x5454, 0x57D4, 0x5454, 0x5FF4, 0x4004, 0x7FFC}},
    {22296, 2, {0x7FFC, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4114, 0x5FF4, 0x4044, 0x5FF4, 0x4844, 0x44C4, 0x7FFC}},
    {22297, 2, {0x0, 0x7FFC, 0x4444, 0x4824, 0x57D4, 0x4104, 0x4FE4, 0x4AA4, 0x4BA4, 0x4AA4, 0x4BA4, 0x4AA4, 0x4FE4, 0x4004, 0x7FFC, 0x4004}},
    {22298, 2, {0x7FFC, 0x4104, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x5FF4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4444, 0x4824, 0x7FFC}},
    {22299, 2, {0x7FFC, 0x4004, 0x5FF4, 0x5294, 0x5FF4, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4444, 0x4284, 0x4FE4, 0x4104, 0x7FFC, 0x4104, 0x7FFC}},
    {22300, 2, {0x7FFC, 0x4004, 0x5FF4, 0x5294, 0x5FF4, 0x4004, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4514, 0x4CA4, 0x7444, 0x4624, 0x7FFC}},
    {22301, 2, {0x0, 0xFFFE, 0x8002, 0x9112, 0x9392, 0xA82A, 0xBBBA, 0x9012, 0xABAA, 0xB83A, 0x8382, 0xAAAA, 0xABAA, 0x8002, 0xFFFE, 0x8002}},
    {22302, 2, {0xFFFE, 0x9112, 0xA3A2, 0xB83A, 0x9392, 0xA82A, 0xBBBA, 0x8282, 0xABAA, 0x8102, 0xBFFA, 0x8542, 0x8922, 0x9112, 0xFFFE, 0x8002}},
    {22303, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22304, 2, {0x1080, 0x1080, 0x1080, 0x1080, 0x1080, 0xFE80, 0x1080, 0x1080, 0x1080, 0x1080, 0x1680, 0x1884, 0xE084, 0x4084, 0x7C, 0x0}},
    {22305, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x120, 0x110, 0x110, 0x100, 0x100, 0xFFFE, 0x0}},
    {22306, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x1020, 0xFE20, 0x1020, 0x1020, 0x1020, 0x1020, 0x1620, 0x1820, 0xE020, 0x4020, 0xA0, 0x40}},
    {22307, 2, {0x0, 0x3FF0, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE10E, 0x100, 0x1FF0, 0x100, 0x100, 0x100, 0x7FFC, 0x0}},
    {22308, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0xFE60, 0x1050, 0x1048, 0x1044, 0x1044, 0x1640, 0x1840, 0xE040, 0x4040, 0x40, 0x40}},
    {22309, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x440, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {22310, 2, {0x1200, 0x1100, 0x1080, 0x1080, 0x1040, 0xFC40, 0x1040, 0x10A0, 0x10A0, 0x10A0, 0x1110, 0x1D10, 0xE208, 0x4208, 0x404, 0x802}},
    {22311, 2, {0x0, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x2080, 0x2080, 0x2080, 0x2080, 0x4080, 0x5FFE, 0x8000}},
    {22312, 2, {0x200, 0x200, 0x400, 0xFFFE, 0x800, 0x840, 0x1040, 0x3040, 0x57FC, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {22313, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x1020, 0xFC20, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {22314, 2, {0x1040, 0x1040, 0x1080, 0x10FE, 0x1100, 0xFE00, 0x10FC, 0x1008, 0x1010, 0x1020, 0x1040, 0x1C80, 0xE102, 0x4102, 0xFE, 0x0}},
    {22315, 2, {0x1008, 0x103C, 0x13C0, 0x1040, 0x1040, 0xFE40, 0x1040, 0x107E, 0x13C0, 0x1040, 0x1640, 0x1842, 0xE042, 0x4042, 0x3E, 0x0}},
    {22316, 2, {0x2000, 0x21FC, 0x2000, 0x2000, 0x2000, 0xFBFE, 0x2080, 0x2080, 0x2100, 0x21FC, 0x2004, 0x3804, 0xE004, 0x4004, 0x28, 0x10}},
    {22317, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22318, 2, {0x1000, 0x1000, 0x11F8, 0x1008, 0x1008, 0xFC08, 0x1008, 0x11F8, 0x1108, 0x1100, 0x1100, 0x1D00, 0xE102, 0x4102, 0xFE, 0x0}},
    {22319, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1100, 0x1100, 0x1100, 0x1D00, 0xE102, 0x4102, 0xFE, 0x0}},
    {22320, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0x112C, 0xFD34, 0x1164, 0x13A4, 0x1124, 0x1134, 0x1128, 0x1D22, 0xE122, 0x4102, 0xFE, 0x0}},
    {22321, 2, {0x20, 0xC20, 0x7020, 0x1020, 0x1020, 0x11FC, 0x1020, 0xFE20, 0x1020, 0x1020, 0x1020, 0x1020, 0x2020, 0x23FE, 0x4000, 0x8000}},
    {22322, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x1020, 0xFC20, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {22323, 2, {0x1104, 0x1124, 0x1124, 0x1124, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D24, 0xE124, 0x4224, 0x204, 0x404}},
    {22324, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0x1104, 0xFD04, 0x1204, 0x1484, 0x1044, 0x1024, 0x1024, 0x1C04, 0xE004, 0x4004, 0x28, 0x10}},
    {22325, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0xFE40, 0x107C, 0x1040, 0x1040, 0x1040, 0x1640, 0x1840, 0xE040, 0x4040, 0x7FE, 0x0}},
    {22326, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22327, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0x1040, 0xFE40, 0x1050, 0x1048, 0x1044, 0x1044, 0x1640, 0x1840, 0xE040, 0x4040, 0x40, 0x40}},
    {22328, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D24, 0xE124, 0x41FC, 0x4, 0x0}},
    {22329, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1200, 0xFE00, 0x1200, 0x1200, 0x1200, 0x1200, 0x1E00, 0xE200, 0x4200, 0x400, 0x400, 0x800}},
    {22330, 2, {0x2000, 0x23F0, 0x2020, 0x2040, 0x2080, 0xF900, 0x23FC, 0x2124, 0x2124, 0x2124, 0x2224, 0x3A44, 0xE444, 0x4884, 0x128, 0x210}},
    {22331, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1100, 0xFD00, 0x11FE, 0x1110, 0x1110, 0x1110, 0x1110, 0x1D10, 0xE110, 0x4210, 0x210, 0x410}},
    {22332, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22333, 2, {0x1100, 0x1100, 0x1100, 0x11FC, 0x12A4, 0xFCA4, 0x10A4, 0x1124, 0x1124, 0x1244, 0x1C44, 0xE084, 0x4104, 0x204, 0x428, 0x10}},
    {22334, 2, {0x1000, 0x17F8, 0x1108, 0x1110, 0x1110, 0xFD20, 0x113C, 0x1104, 0x1104, 0x1288, 0x1E88, 0xE250, 0x4420, 0x450, 0x888, 0x306}},
    {22335, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0x1088, 0xFD04, 0x1202, 0x1088, 0x1088, 0x1088, 0x1088, 0x1C88, 0xE088, 0x4108, 0x108, 0x208}},
    {22336, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0x1120, 0xFD20, 0x113C, 0x1120, 0x1120, 0x1120, 0x1120, 0x1D20, 0xE120, 0x4120, 0x7FE, 0x0}},
    {22337, 2, {0x1004, 0x101E, 0x11F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x11FE, 0x1110, 0x1110, 0x1110, 0x1D08, 0xE10A, 0x414A, 0x186, 0x102}},
    {22338, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1100, 0xFDFC, 0x1144, 0x1144, 0x1144, 0x1128, 0x1128, 0x1D10, 0xE110, 0x4228, 0x244, 0x482}},
    {22339, 2, {0x1000, 0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x1090, 0x1090, 0x1090, 0x1090, 0x1C90, 0xE112, 0x4112, 0x20E, 0x400}},
    {22340, 2, {0x2000, 0x21F0, 0x2110, 0x2110, 0x2110, 0xFA0E, 0x2400, 0x23F8, 0x2108, 0x2110, 0x2090, 0x38A0, 0xE040, 0x40A0, 0x318, 0xC06}},
    {22341, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0x1088, 0xFD44, 0x1222, 0x1020, 0x1000, 0x11FC, 0x1004, 0x1C08, 0xE008, 0x4010, 0x10, 0x20}},
    {22342, 2, {0x1040, 0x1040, 0x1040, 0x1080, 0x10FE, 0xFD08, 0x1288, 0x1088, 0x1088, 0x1050, 0x1050, 0x1C20, 0xE050, 0x4088, 0x104, 0x202}},
    {22343, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0x1104, 0xFE04, 0x1084, 0x1044, 0x1044, 0x1014, 0x1024, 0x1C44, 0xE184, 0x4004, 0x28, 0x10}},
    {22344, 2, {0x1000, 0x1000, 0x13FC, 0x1204, 0x1408, 0xFC00, 0x11F0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1D12, 0xE212, 0x4212, 0x40E, 0x800}},
    {22345, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1020, 0xFD24, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1024, 0x1C20, 0xE022, 0x4022, 0x1E, 0x0}},
    {22346, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1080, 0xFC80, 0x1080, 0x10FC, 0x1084, 0x1084, 0x1084, 0x1C84, 0xE104, 0x4104, 0x228, 0x410}},
    {22347, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0x1108, 0xFD04, 0x1204, 0x15FA, 0x1088, 0x1088, 0x1088, 0x1C88, 0xE108, 0x4108, 0x228, 0x410}},
    {22348, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6140, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22349, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x1148, 0xFD28, 0x1128, 0x1108, 0x17FE, 0x1108, 0x1108, 0x1D08, 0xE108, 0x4208, 0x228, 0x410}},
    {22350, 2, {0x1080, 0x1080, 0x1080, 0x10FC, 0x1104, 0xFD08, 0x1240, 0x1440, 0x1040, 0x10A0, 0x10A0, 0x1D10, 0xE110, 0x4208, 0x404, 0x802}},
    {22351, 2, {0x1000, 0x1000, 0x13FE, 0x1010, 0x1010, 0xFC20, 0x1020, 0x1068, 0x1064, 0x10A4, 0x1D22, 0xE222, 0x4420, 0x20, 0x20, 0x20}},
    {22352, 2, {0x100, 0x1110, 0x1110, 0x1110, 0x1110, 0x2928, 0x2524, 0x4544, 0x8100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22353, 2, {0x1080, 0x1040, 0x1040, 0x13FC, 0x1000, 0xFC00, 0x11F0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1D12, 0xE112, 0x4212, 0x20E, 0x400}},
    {22354, 2, {0x2080, 0x2088, 0x20B0, 0x3EC0, 0x2080, 0x2084, 0x2684, 0x387C, 0x2100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22355, 2, {0x820, 0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFFFE, 0x1020, 0x2020, 0x4120, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22356, 2, {0x100, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22357, 2, {0x78, 0x1FA0, 0x1120, 0x1120, 0x1120, 0x1110, 0x2110, 0x2108, 0x4004, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22358, 2, {0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x440, 0x442, 0x842, 0x303E, 0xC100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22359, 2, {0x1040, 0x1040, 0x1040, 0x13FC, 0x1044, 0xFC44, 0x1044, 0x1044, 0x13FE, 0x1040, 0x10A0, 0x1CA0, 0xE110, 0x4208, 0x404, 0x802}},
    {22360, 2, {0x1000, 0x17FE, 0x1080, 0x1080, 0x1080, 0xFCF8, 0x1108, 0x1108, 0x1108, 0x1208, 0x13F8, 0x1C10, 0xE010, 0x4010, 0x7FE, 0x0}},
    {22361, 2, {0x0, 0x7FFE, 0x888, 0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x888, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22362, 2, {0x400, 0x25FC, 0x2504, 0x2488, 0x2450, 0x2420, 0x2458, 0x2586, 0x400, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22363, 2, {0x1000, 0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x1020, 0x1020, 0x1040, 0x1040, 0x1C88, 0xE104, 0x43FE, 0x102, 0x0}},
    {22364, 2, {0x2000, 0x23FE, 0x2200, 0x2220, 0x2220, 0xFA20, 0x22FC, 0x2224, 0x2224, 0x2224, 0x2244, 0x3A44, 0xE484, 0x4484, 0x928, 0x1210}},
    {22365, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1C50, 0xE048, 0x4084, 0x102, 0x602}},
    {22366, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x1114, 0x1108, 0x1100, 0x11FE, 0x1002, 0x1C02, 0xE3FA, 0x4002, 0x14, 0x8}},
    {22367, 2, {0x1080, 0x1040, 0x1040, 0x17FC, 0x1110, 0xFD10, 0x1110, 0x1110, 0x1110, 0x10A0, 0x10A0, 0x1C40, 0xE0A0, 0x4110, 0x208, 0xC06}},
    {22368, 2, {0x40, 0x7C40, 0x4440, 0x4840, 0x44A0, 0x44A0, 0x5510, 0x4908, 0x4204, 0x4502, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22369, 2, {0x2020, 0x2020, 0x2020, 0x23FE, 0x2222, 0xFA24, 0x2220, 0x23FC, 0x2284, 0x2288, 0x2248, 0x3A50, 0xE220, 0x4450, 0x488, 0x906}},
    {22370, 2, {0x1020, 0x1020, 0x1124, 0x10A4, 0x10A8, 0xFC20, 0x11FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {22371, 2, {0x100, 0x1110, 0x920, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22372, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x11FC, 0x1124, 0x1124, 0x1124, 0x1DFC, 0xE124, 0x4020, 0x20, 0x20}},
    {22373, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x1108, 0xFDF8, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108, 0x1D08, 0xE108, 0x4108, 0x7FE, 0x0}},
    {22374, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1D04, 0xE000, 0x4000, 0x3FE, 0x0}},
    {22375, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x1020, 0xFC40, 0x1040, 0x10FC, 0x1184, 0x1284, 0x1484, 0x1C84, 0xE084, 0x4084, 0xFC, 0x84}},
    {22376, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x1100, 0x1108, 0x1110, 0x1120, 0x11C0, 0x1D02, 0xE102, 0x4102, 0xFE, 0x0}},
    {22377, 2, {0x1108, 0x1108, 0x1108, 0x1108, 0x17FE, 0xFD08, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108, 0x1D08, 0xE108, 0x4108, 0x1F8, 0x108}},
    {22378, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x1124, 0xFCA4, 0x10A8, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {22379, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x103E, 0xFC20, 0x1020, 0x1020, 0x11FC, 0x1104, 0x1104, 0x1D04, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22380, 2, {0x2008, 0x203C, 0x23D0, 0x2290, 0x2290, 0xFA90, 0x2290, 0x2290, 0x2290, 0x2288, 0x2288, 0x3AC8, 0xE2A4, 0x44D4, 0x492, 0x800}},
    {22381, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0xFDFC, 0x1100, 0x1140, 0x1144, 0x1148, 0x1170, 0x1D40, 0xE142, 0x4242, 0x23E, 0x400}},
    {22382, 2, {0x1020, 0x1020, 0x1040, 0x1040, 0x1088, 0xFD04, 0x13FE, 0x1002, 0x1000, 0x11FC, 0x1104, 0x1D04, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22383, 2, {0x1000, 0x13FC, 0x1020, 0x1020, 0x1040, 0xFE40, 0x10D0, 0x1148, 0x1244, 0x1444, 0x1640, 0x1840, 0xE040, 0x4000, 0x7FE, 0x0}},
    {22384, 2, {0x1000, 0x11FE, 0x1102, 0x1102, 0x1102, 0xFD7A, 0x114A, 0x114A, 0x114A, 0x114A, 0x117A, 0x1D4A, 0xE102, 0x4102, 0x10A, 0x104}},
    {22385, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x13FE, 0x1020, 0x1C50, 0xE050, 0x4088, 0x104, 0x202}},
    {22386, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0x1094, 0xFC94, 0x13FC, 0x1290, 0x1290, 0x13FE, 0x1092, 0x1C92, 0xE11A, 0x4114, 0x210, 0x410}},
    {22387, 2, {0x2020, 0x2120, 0x2120, 0x2220, 0x22FC, 0xF4A4, 0x2724, 0x2124, 0x2224, 0x2224, 0x24A4, 0x37A4, 0xE0C4, 0x4044, 0x94, 0x108}},
    {22388, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x820, 0x1010, 0x2108, 0x4104, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22389, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1100, 0xFD00, 0x1100, 0x11FE, 0x1110, 0x1110, 0x1110, 0x1D10, 0xE110, 0x4110, 0x7FE, 0x0}},
    {22390, 2, {0x2000, 0x21F8, 0x2108, 0x2148, 0x2128, 0xF908, 0x2108, 0x27FE, 0x2208, 0x2288, 0x2248, 0x3A08, 0xE3FE, 0x4008, 0x50, 0x20}},
    {22391, 2, {0x2000, 0x2000, 0x27FE, 0x2008, 0x2008, 0xFBC8, 0x2248, 0x2248, 0x2248, 0x2248, 0x23C8, 0x3A48, 0xE008, 0x4008, 0x28, 0x10}},
    {22392, 2, {0x1080, 0x1080, 0x1100, 0x11FC, 0x1204, 0xFC04, 0x11E4, 0x1124, 0x1124, 0x1124, 0x11E4, 0x1D24, 0xE004, 0x4004, 0x28, 0x10}},
    {22393, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x1000, 0x1090, 0x1090, 0x1090, 0x1088, 0x1D08, 0xE108, 0x4104, 0x204, 0x402}},
    {22394, 2, {0x1050, 0x1048, 0x1048, 0x1040, 0x13FE, 0xFC80, 0x1080, 0x10FC, 0x1144, 0x1144, 0x1128, 0x1D28, 0xE210, 0x4228, 0x444, 0x182}},
    {22395, 2, {0x1008, 0x103C, 0x13E0, 0x1220, 0x1220, 0xFE20, 0x1220, 0x13FE, 0x1220, 0x1210, 0x1210, 0x1E12, 0xE20A, 0x428A, 0x326, 0x212}},
    {22396, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1100, 0xFD00, 0x11FE, 0x1110, 0x1110, 0x1130, 0x1118, 0x1D14, 0xE112, 0x4210, 0x210, 0x410}},
    {22397, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0x1088, 0xFD24, 0x1212, 0x1010, 0x11FC, 0x1004, 0x1008, 0x1C88, 0xE050, 0x4020, 0x10, 0x10}},
    {22398, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x1000, 0x1000, 0x13FE, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {22399, 2, {0x2088, 0x2088, 0x2088, 0x2108, 0x217E, 0xFB08, 0x2508, 0x2148, 0x2128, 0x2128, 0x2108, 0x3908, 0xE108, 0x4108, 0x128, 0x110}},
    {22400, 2, {0x1008, 0x103C, 0x13E0, 0x1020, 0x1124, 0xFCA4, 0x10A8, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {22401, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0x1120, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1020, 0x1050, 0x1C50, 0xE088, 0x4088, 0x104, 0x202}},
    {22402, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x7FFC, 0x920, 0x920, 0xFFFE, 0x920, 0x920, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x0}},
    {22403, 2, {0x1080, 0x1040, 0x1040, 0x1000, 0x13FC, 0xFC00, 0x1008, 0x1108, 0x1108, 0x1090, 0x1090, 0x1C90, 0xE0A0, 0x4020, 0x7FE, 0x0}},
    {22404, 2, {0x440, 0x420, 0xFFFE, 0x910, 0x1120, 0x11C0, 0x2304, 0x4D04, 0x80FC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22405, 2, {0x1090, 0x1088, 0x1088, 0x1080, 0x13FE, 0xFCA0, 0x10A0, 0x10A4, 0x10A4, 0x10A8, 0x1128, 0x1D32, 0xE122, 0x4262, 0x29E, 0x400}},
    {22406, 2, {0x2040, 0x2040, 0x207E, 0x2040, 0x2040, 0xFBFC, 0x2204, 0x2204, 0x2204, 0x23FC, 0x2204, 0x3A00, 0xE200, 0x4400, 0x400, 0x800}},
    {22407, 2, {0x2000, 0x2000, 0x23DE, 0x2252, 0x2252, 0xFA52, 0x2252, 0x2252, 0x2272, 0x2202, 0x2202, 0x3A02, 0xE202, 0x43FE, 0x202, 0x0}},
    {22408, 2, {0x890, 0x888, 0x1080, 0x309E, 0x57E0, 0x9080, 0x1044, 0x1034, 0x110C, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22409, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0x1204, 0xFDF4, 0x1114, 0x1114, 0x1114, 0x11F4, 0x1104, 0x1D28, 0xE112, 0x4102, 0xFE, 0x0}},
    {22410, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0x2204, 0xFBFC, 0x2220, 0x2220, 0x23FE, 0x2220, 0x2220, 0x3A10, 0xE212, 0x428A, 0x306, 0x202}},
    {22411, 2, {0x4, 0x7F84, 0x1224, 0x1224, 0x1224, 0xFFA4, 0x1224, 0x2204, 0x2214, 0x4108, 0x8100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22412, 2, {0x2000, 0x23FE, 0x2202, 0x2202, 0x22FA, 0xFA02, 0x2202, 0x22FA, 0x228A, 0x228A, 0x228A, 0x3AFA, 0xE202, 0x4202, 0x20A, 0x204}},
    {22413, 2, {0x1020, 0x1020, 0x1040, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1D04, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22414, 2, {0x2080, 0x2080, 0x21F8, 0x2208, 0x2510, 0xF8A0, 0x2040, 0x20A0, 0x2318, 0x2C06, 0x23F8, 0x3A08, 0xE208, 0x4208, 0x3F8, 0x208}},
    {22415, 2, {0x1020, 0x1020, 0x11FC, 0x1024, 0x13FE, 0xFC24, 0x11FC, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {22416, 2, {0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22417, 2, {0x2040, 0x2040, 0x20FC, 0x2108, 0x2290, 0xF860, 0x2048, 0x2190, 0x263E, 0x2042, 0x2084, 0x3B48, 0xE030, 0x4020, 0xC0, 0x700}},
    {22418, 2, {0x400, 0x820, 0x1010, 0x3FF8, 0x810, 0x1020, 0x2448, 0x7EFC, 0x204, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22419, 2, {0x2040, 0x2020, 0x2020, 0x27FE, 0x2040, 0xF884, 0x2108, 0x23F0, 0x2022, 0x2044, 0x3988, 0xE610, 0x4020, 0x50, 0x188, 0x604}},
    {22420, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22421, 2, {0x7C, 0x3F80, 0x2000, 0x3FFE, 0x2000, 0x2FF8, 0x2808, 0x2808, 0x2FF8, 0x2080, 0x2080, 0x2FFC, 0x4080, 0x4080, 0xBFFE, 0x0}},
    {22422, 2, {0x1040, 0x1080, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x1100, 0x1100, 0x1100, 0x11FC, 0x1D04, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22423, 2, {0x1090, 0x1090, 0x1090, 0x1292, 0x1194, 0xFC98, 0x1090, 0x1198, 0x1294, 0x1492, 0x1090, 0x1C90, 0xE112, 0x4112, 0x212, 0x40E}},
    {22424, 2, {0x1090, 0x1094, 0x1092, 0x1112, 0x1110, 0xFF7E, 0x1510, 0x1110, 0x1110, 0x1128, 0x1128, 0x1D28, 0xE128, 0x4144, 0x144, 0x182}},
    {22425, 2, {0x1020, 0x1020, 0x1124, 0x10A4, 0x10A8, 0xFC20, 0x13FE, 0x1090, 0x1090, 0x1090, 0x1090, 0x1C92, 0xE112, 0x4112, 0x20E, 0x400}},
    {22426, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x1010, 0x1010, 0x1010, 0x7CFC, 0x1010, 0x1010, 0x1E10, 0xF0FE, 0x4000}},
    {22427, 2, {0x1000, 0x11F0, 0x1110, 0x1110, 0x1110, 0xFD10, 0x124E, 0x1440, 0x13FC, 0x1040, 0x1CE0, 0xE150, 0x4248, 0x446, 0x40, 0x40}},
    {22428, 2, {0x1000, 0x13F8, 0x1088, 0x1090, 0x10BC, 0xFD04, 0x1114, 0x1248, 0x1040, 0x17FE, 0x1040, 0x1CE0, 0xE150, 0x4248, 0x446, 0x40}},
    {22429, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0x1208, 0xFDFE, 0x1100, 0x117C, 0x1144, 0x1144, 0x1154, 0x1D48, 0xE142, 0x4242, 0x23E, 0x400}},
    {22430, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0x2000, 0xF838, 0x23C0, 0x2040, 0x2040, 0x207C, 0x23C0, 0x3840, 0xE042, 0x4042, 0x3E, 0x0}},
    {22431, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0x13FE, 0xFC20, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {22432, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x11F8, 0xFD08, 0x1108, 0x11F8, 0x1144, 0x1148, 0x1130, 0x1D20, 0xE110, 0x4148, 0x186, 0x100}},
    {22433, 2, {0x8A0, 0x890, 0x1080, 0x30FC, 0x5780, 0x9048, 0x1050, 0x1024, 0x1054, 0x118C, 0x1604, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22434, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1100, 0xFDFE, 0x1100, 0x1100, 0x117C, 0x1144, 0x1144, 0x1D44, 0xE144, 0x427C, 0x244, 0x400}},
    {22435, 2, {0x1000, 0x13FE, 0x1000, 0x1000, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1D04, 0xE000, 0x4000, 0x3FE, 0x0}},
    {22436, 2, {0x2000, 0x27FE, 0x2040, 0x2080, 0x2110, 0xFA08, 0x27FC, 0x2044, 0x2040, 0x2040, 0x27FC, 0x3840, 0xE040, 0x4040, 0xFFE, 0x0}},
    {22437, 2, {0x2040, 0x2040, 0x20A0, 0x2110, 0xFA08, 0x2406, 0x23F8, 0x2000, 0x2000, 0x23F8, 0x2208, 0x3A08, 0xE208, 0x4208, 0x3F8, 0x208}},
    {22438, 2, {0x0, 0x3FF0, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x2208, 0x2190, 0x2C60, 0x311C, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22439, 2, {0x2040, 0x2040, 0x2080, 0x23FC, 0x2204, 0xFA04, 0x22F4, 0x2294, 0x2294, 0x2294, 0x2294, 0x3AF4, 0xE204, 0x4204, 0x214, 0x208}},
    {22440, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1414, 0xFC10, 0x1010, 0x13FE, 0x1010, 0x1110, 0x1090, 0x1C90, 0xE010, 0x4010, 0x50, 0x20}},
    {22441, 2, {0x0, 0x7FFC, 0x440, 0x2448, 0x1448, 0x1450, 0x440, 0xFFFE, 0x0, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22442, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0x11FE, 0xFC88, 0x1088, 0x1088, 0x1088, 0x13FE, 0x1088, 0x1C88, 0xE108, 0x4108, 0x208, 0x408}},
    {22443, 2, {0x840, 0x840, 0x7DF8, 0x848, 0xC48, 0x19C8, 0x684A, 0x8AA, 0x2886, 0x1102, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22444, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0x11FE, 0xFC88, 0x1088, 0x1088, 0x1088, 0x13FE, 0x1000, 0x1C88, 0xE084, 0x4104, 0x202, 0x402}},
    {22445, 2, {0x2000, 0x23FC, 0x2090, 0x2090, 0x2090, 0xF892, 0x2492, 0x2294, 0x2294, 0x2298, 0x2090, 0x3890, 0xE090, 0x4090, 0x7FE, 0x0}},
    {22446, 2, {0x2040, 0x2040, 0x27FC, 0x20A0, 0x2110, 0xFA08, 0x2DF6, 0x2000, 0x27FC, 0x2080, 0x2100, 0x3BF8, 0xE008, 0x4008, 0x50, 0x20}},
    {22447, 2, {0x1010, 0x1210, 0x1110, 0x1110, 0x10FE, 0xFC10, 0x1310, 0x1128, 0x1124, 0x1144, 0x1142, 0x1D82, 0xE100, 0x4280, 0x47E, 0x0}},
    {22448, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0x1020, 0xFDFE, 0x1000, 0x1000, 0x11FE, 0x1020, 0x1028, 0x1C24, 0xE022, 0x4020, 0x20, 0x20}},
    {22449, 2, {0x1020, 0x1124, 0x10A4, 0x10A4, 0x10A8, 0xFC20, 0x13FC, 0x1004, 0x1004, 0x1004, 0x11FC, 0x1C04, 0xE004, 0x4004, 0x3FC, 0x4}},
    {22450, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0xFC00, 0x11FC, 0x1004, 0x1004, 0x11FC, 0x1100, 0x1D00, 0xE102, 0x4102, 0xFE, 0x0}},
    {22451, 2, {0x2080, 0x2080, 0x213C, 0x2200, 0x2480, 0xF880, 0x217E, 0x2308, 0x2508, 0x2108, 0x2108, 0x3908, 0xE108, 0x4108, 0x128, 0x110}},
    {22452, 2, {0x2040, 0x2020, 0x2020, 0x27FE, 0x2000, 0xF800, 0x2090, 0x2254, 0x2224, 0x2224, 0x2254, 0x3A94, 0xE204, 0x4204, 0x3FC, 0x4}},
    {22453, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1202, 0xFC44, 0x1040, 0x13FE, 0x1088, 0x1088, 0x1108, 0x1CD0, 0xE020, 0x4050, 0x88, 0x304}},
    {22454, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0x1088, 0xFC50, 0x13FE, 0x1020, 0x1020, 0x1020, 0x11FC, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {22455, 2, {0x1000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0xFE08, 0x1208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1D20, 0xE120, 0x4222, 0x422, 0x181E}},
    {22456, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1404, 0xFDF8, 0x1000, 0x1000, 0x13FE, 0x1090, 0x1090, 0x1C90, 0xE112, 0x4112, 0x20E, 0x400}},
    {22457, 2, {0x2100, 0x211E, 0x2112, 0x2FD2, 0x2114, 0xF914, 0x27D8, 0x2114, 0x2112, 0x2FD2, 0x2112, 0x391A, 0xE214, 0x4210, 0x410, 0x810}},
    {22458, 2, {0x1008, 0x103C, 0x11C0, 0x1004, 0x1144, 0xFCA8, 0x1000, 0x11F8, 0x1010, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {22459, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1C00, 0xE088, 0x4084, 0x102, 0x202}},
    {22460, 2, {0x10F8, 0x2088, 0x4108, 0x8A06, 0x1000, 0x31F8, 0x5088, 0x9050, 0x1020, 0x10D8, 0x1706, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22461, 2, {0x200C, 0x11F0, 0x8100, 0x4900, 0x9FE, 0x1110, 0xE110, 0x2210, 0x2210, 0x2510, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22462, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0xFDFC, 0x1000, 0x1000, 0x11FC, 0x1020, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20, 0x20}},
    {22463, 2, {0x2040, 0x2020, 0x23FE, 0x2200, 0x2200, 0xFAFE, 0x2204, 0x2228, 0x2210, 0x23FE, 0x3A12, 0xE214, 0x4210, 0x410, 0x450, 0x820}},
    {22464, 2, {0xF8, 0x3F00, 0x100, 0xFFFE, 0x100, 0x924, 0xF938, 0x922, 0x3922, 0xC91E, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22465, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0xFC20, 0x1050, 0x1088, 0x1144, 0x1222, 0x1020, 0x1DF8, 0xE008, 0x4010, 0x10, 0x20}},
    {22466, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x1D20, 0xE0A0, 0x4040, 0xB0, 0x30E}},
    {22467, 2, {0x2040, 0x2080, 0x2110, 0x2208, 0x27FC, 0xF904, 0x2100, 0x23F8, 0x2440, 0x2040, 0x27FE, 0x3840, 0xE0A0, 0x4110, 0x208, 0xC06}},
    {22468, 2, {0x2100, 0x21F8, 0x2308, 0x2490, 0x2060, 0xF998, 0x2646, 0x2040, 0x23F8, 0x2040, 0x23F8, 0x3840, 0xE7FC, 0x4040, 0x40, 0x40}},
    {22469, 2, {0x40, 0x7820, 0x4BFE, 0x5080, 0x48F8, 0x4488, 0x5508, 0x4928, 0x4210, 0x4100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22470, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x1210, 0xFDFC, 0x1124, 0x1124, 0x11FC, 0x1124, 0x1124, 0x1DFC, 0xE124, 0x4224, 0x214, 0x408}},
    {22471, 2, {0x1000, 0x11F8, 0x1008, 0x10D0, 0x1020, 0xFDFC, 0x1124, 0x1124, 0x11FC, 0x1124, 0x1124, 0x1DFC, 0xE124, 0x4124, 0x124, 0x10C}},
    {22472, 2, {0x1040, 0x1040, 0x1088, 0x1104, 0x13FE, 0xFC02, 0x1088, 0x1144, 0x1242, 0x10F8, 0x1D88, 0xE250, 0x4020, 0x50, 0x188, 0x606}},
    {22473, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x1124, 0x1124, 0x1124, 0x12AA, 0x1020, 0x1C50, 0xE050, 0x4088, 0x104, 0x202}},
    {22474, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x104, 0x7D88, 0x950, 0x1120, 0x2518, 0xC206, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22475, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x1020, 0x1020, 0x11FC, 0x1C20, 0xE020, 0x4020, 0x3FE, 0x0}},
    {22476, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1120, 0x1122, 0x1D14, 0xE108, 0x4144, 0x182, 0x100}},
    {22477, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0x10FC, 0xFC00, 0x11FE, 0x1102, 0x1102, 0x11FE, 0x1102, 0x1D02, 0xE1FE, 0x4102, 0x10A, 0x104}},
    {22478, 2, {0x2028, 0x2024, 0x2020, 0x27FE, 0x2420, 0xFC20, 0x2424, 0x27A4, 0x24A4, 0x24A8, 0x24A8, 0x3C90, 0xE692, 0x492A, 0x846, 0x1082}},
    {22479, 2, {0x2008, 0x201C, 0x2770, 0x2110, 0x2110, 0xFA50, 0x225C, 0x2750, 0x2150, 0x2550, 0x2550, 0x3A7C, 0xE200, 0x4500, 0x8FE, 0x1000}},
    {22480, 2, {0x1000, 0x11FC, 0x1004, 0x10FC, 0x1004, 0xFDFC, 0x1000, 0x13FE, 0x1202, 0x11F8, 0x1088, 0x1C88, 0xE050, 0x4020, 0xD8, 0x306}},
    {22481, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22482, 2, {0x1008, 0x103C, 0x13C0, 0x1044, 0x1224, 0xFD28, 0x1100, 0x1010, 0x13FE, 0x1010, 0x1110, 0x1C90, 0xE090, 0x4010, 0x50, 0x20}},
    {22483, 2, {0x1008, 0x103C, 0x13C0, 0x1124, 0x1098, 0xFC60, 0x1390, 0x1010, 0x13FE, 0x1010, 0x1110, 0x1C90, 0xE090, 0x4010, 0x50, 0x20}},
    {22484, 2, {0x2048, 0x2044, 0x27FE, 0x2040, 0x2040, 0xFBFC, 0x2244, 0x2244, 0x23FC, 0x2244, 0x2244, 0x3BFC, 0xE244, 0x4244, 0x254, 0x208}},
    {22485, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0xFDFC, 0x1000, 0x1000, 0x13FE, 0x1020, 0x1020, 0x1DFC, 0xE020, 0x4020, 0x3FE, 0x0}},
    {22486, 2, {0x2108, 0x2108, 0x27FE, 0x2108, 0x2108, 0xF800, 0x2090, 0x2092, 0x2114, 0x2318, 0x2510, 0x3932, 0xE152, 0x4112, 0x10E, 0x100}},
    {22487, 2, {0x1010, 0x1010, 0x1090, 0x109E, 0x1090, 0xFC90, 0x13FE, 0x1000, 0x1010, 0x1092, 0x1092, 0x1D14, 0xE208, 0x4010, 0x60, 0x380}},
    {22488, 2, {0x2008, 0x2008, 0x2788, 0x2488, 0x24FE, 0xFC88, 0x2488, 0x27C8, 0x24A8, 0x24A8, 0x2488, 0x3C88, 0xE788, 0x4488, 0x28, 0x10}},
    {22489, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0xF800, 0x2000, 0x27FC, 0x2404, 0x2444, 0x2444, 0x3C44, 0xE4B4, 0x4108, 0x604, 0x1802}},
    {22490, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0x2208, 0xFBF8, 0x2040, 0x2040, 0x27FC, 0x2444, 0x24A4, 0x3D14, 0xE614, 0x4404, 0x414, 0x408}},
    {22491, 2, {0x2000, 0x27FC, 0x2404, 0x24A4, 0x2514, 0xFE0C, 0x2404, 0x25F4, 0x2514, 0x2514, 0x2514, 0x3DF4, 0xE514, 0x4404, 0x414, 0x408}},
    {22492, 2, {0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0x920, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22493, 2, {0x2020, 0x2020, 0x2050, 0x2088, 0x2144, 0xFA22, 0x21F8, 0x2008, 0x2050, 0x2020, 0x20A4, 0x3A82, 0xE28A, 0x428A, 0x478, 0x0}},
    {22494, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1404, 0xFC00, 0x13FE, 0x1020, 0x1020, 0x1120, 0x113C, 0x1D20, 0xE2A0, 0x4260, 0x43E, 0x800}},
    {22495, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0x2010, 0xF810, 0x23D2, 0x2252, 0x2252, 0x2254, 0x23D4, 0x3808, 0xE0EA, 0x471A, 0x226, 0x42}},
    {22496, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x11FC, 0xFD00, 0x11FC, 0x1104, 0x11FC, 0x1020, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20, 0x20}},
    {22497, 2, {0x2000, 0x27FE, 0x2090, 0x2090, 0x2090, 0xFB9C, 0x2204, 0x2204, 0x2204, 0x239C, 0x2090, 0x3890, 0xE090, 0x4090, 0x7FE, 0x0}},
    {22498, 2, {0x2040, 0x2248, 0x2150, 0x2040, 0x23F8, 0xF880, 0x27FC, 0x2110, 0x2208, 0x25F4, 0x2912, 0x3910, 0xE150, 0x4124, 0x104, 0xFC}},
    {22499, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x1088, 0xFC88, 0x1154, 0x1222, 0x1000, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {22500, 2, {0x2040, 0x2080, 0x23FC, 0x2224, 0x2224, 0xFBFC, 0x2224, 0x2244, 0x23FC, 0x2090, 0x3910, 0xE7FE, 0x4010, 0x10, 0x10, 0x10}},
    {22501, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC20, 0x13FE, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1D04, 0xE1FC, 0x4104, 0x114, 0x108}},
    {22502, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0x2504, 0xF900, 0x21DE, 0x2252, 0x2252, 0x2352, 0x249A, 0x3894, 0xE110, 0x4112, 0x212, 0x40E}},
    {22503, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x1D08, 0xE3FE, 0x4000, 0x90, 0x108, 0x204}},
    {22504, 2, {0x2040, 0x2040, 0x20A0, 0x2110, 0x2208, 0xFDF6, 0x2000, 0x27FC, 0x24A4, 0x24A4, 0x27FC, 0x3CA4, 0xE4A4, 0x44A4, 0x414, 0x408}},
    {22505, 2, {0x1080, 0x1080, 0x11F8, 0x1108, 0x1210, 0xFDFC, 0x1024, 0x1024, 0x13FE, 0x1024, 0x1024, 0x1DFC, 0xE024, 0x4020, 0xA0, 0x40}},
    {22506, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1202, 0xFC94, 0x1108, 0x1204, 0x1000, 0x11FC, 0x1020, 0x1C20, 0xE020, 0x4020, 0x7FE, 0x0}},
    {22507, 2, {0x2020, 0x2124, 0x20A4, 0x20A8, 0x2020, 0xFBFE, 0x2202, 0x2202, 0x22FA, 0x228A, 0x228A, 0x3A8A, 0xE2FA, 0x4202, 0x20A, 0x204}},
    {22508, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x23F8, 0xFA48, 0x23F8, 0x2248, 0x23F8, 0x2040, 0x20E0, 0x3950, 0xE248, 0x4C46, 0x40, 0x40}},
    {22509, 2, {0x2040, 0x2040, 0x23F8, 0x2048, 0xF848, 0x27FE, 0x2048, 0x2048, 0x23F8, 0x2444, 0x22E8, 0x3950, 0xE248, 0x4446, 0x140, 0x80}},
    {22510, 2, {0x1020, 0x1122, 0x1122, 0x1224, 0x1050, 0xFC88, 0x1304, 0x1022, 0x1020, 0x1124, 0x1D24, 0xE228, 0x4050, 0x88, 0x104, 0x602}},
    {22511, 2, {0x2040, 0x2080, 0x27FE, 0x2110, 0x2248, 0xFC46, 0x23F8, 0x2248, 0x2248, 0x23F8, 0x2248, 0x3A48, 0xE3F8, 0x4042, 0x42, 0x3E}},
    {22512, 2, {0x1008, 0x103C, 0x13C0, 0x1044, 0x1224, 0xFD28, 0x1100, 0x1020, 0x13FE, 0x1070, 0x10A8, 0x1CA8, 0xE124, 0x4222, 0x420, 0x20}},
    {22513, 2, {0x2100, 0x2100, 0x21DC, 0x2114, 0xF114, 0x2114, 0x27D4, 0x2014, 0x2114, 0x2114, 0x3588, 0xE548, 0x4948, 0x114, 0x514, 0x222}},
    {22514, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC40, 0x13FE, 0x1088, 0x1124, 0x1222, 0x10F8, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {22515, 2, {0x2080, 0x2080, 0x20F8, 0x2108, 0xFA10, 0x2400, 0x2040, 0x239C, 0x2204, 0x2204, 0x239C, 0x3A04, 0xE204, 0x4204, 0x3FC, 0x204}},
    {22516, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x2040, 0xFBF8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x3BF8, 0xE208, 0x4208, 0xFFE, 0x0}},
    {22517, 2, {0x2008, 0x203C, 0x21E0, 0x2020, 0x2020, 0xFBFE, 0x2124, 0x2124, 0x27FE, 0x2124, 0x2124, 0x3BFE, 0xE020, 0x4020, 0x3FE, 0x0}},
    {22518, 2, {0x1040, 0x1040, 0x7C40, 0x1040, 0x11F8, 0xFE48, 0x48, 0x4548, 0x92C8, 0x1048, 0x7C68, 0x105A, 0x108A, 0x1C8A, 0xE106, 0x4202}},
    {22519, 2, {0x1040, 0x1040, 0x7C40, 0x11F8, 0x1048, 0xFE48, 0x4448, 0x2948, 0x7CC8, 0x1048, 0x10A8, 0xFEA8, 0x108A, 0x110A, 0x1106, 0x1202}},
    {22520, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0xFDFC, 0x1080, 0x1080, 0x11FE, 0x124A, 0x144A, 0x1C92, 0xE122, 0x4242, 0x94, 0x108}},
    {22521, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0x1108, 0xFC90, 0x1000, 0x13FE, 0x1000, 0x1000, 0x11FC, 0x1D04, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22522, 2, {0x820, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x820, 0x1110, 0x2108, 0xCFE6, 0x100, 0x100, 0x3FF8}},
    {22523, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x1000, 0x11FC, 0x1008, 0x1C10, 0xE3FE, 0x4020, 0x20, 0xA0, 0x40}},
    {22524, 2, {0x2040, 0x2040, 0x23FC, 0x20A0, 0x2110, 0xFA08, 0x27FE, 0x2008, 0x23C8, 0x2248, 0x2248, 0x3A48, 0xE3C8, 0x4008, 0x28, 0x10}},
    {22525, 2, {0x2000, 0x23F8, 0x2008, 0x21F8, 0x2008, 0xFBF8, 0x2000, 0x27FE, 0x2442, 0x2844, 0x23F8, 0x3A48, 0xE248, 0x4248, 0x258, 0x40}},
    {22526, 2, {0x0, 0x7F7C, 0x2244, 0x3E44, 0x2228, 0x3E10, 0x23A8, 0xFE44, 0x282, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22527, 2, {0x23F8, 0x1208, 0x8208, 0x4BF8, 0xA40, 0x1248, 0xE270, 0x2244, 0x2444, 0x253C, 0x900, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22528, 2, {0x1000, 0x13FE, 0x1202, 0x1202, 0x13FE, 0xFE10, 0x1292, 0x1292, 0x1292, 0x12FE, 0x1210, 0x1E92, 0xE292, 0x4492, 0x4FE, 0x802}},
    {22529, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1070, 0x1CA8, 0xE124, 0x4222, 0x20, 0x20}},
    {22530, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22531, 2, {0x1020, 0x810, 0x7EFE, 0x1020, 0x1E3C, 0x1224, 0x1224, 0x2244, 0x4A94, 0x8508, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22532, 2, {0x2000, 0x2040, 0x239C, 0x2204, 0x2204, 0xFB9C, 0x2204, 0x2204, 0x23FC, 0x2090, 0x2090, 0x3890, 0xE090, 0x4112, 0x212, 0x40E}},
    {22533, 2, {0x7F7C, 0x4844, 0x7F44, 0x4148, 0x4128, 0x7F30, 0x4810, 0x4828, 0x7F44, 0x82, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22534, 2, {0x2140, 0x2120, 0x2120, 0x23FE, 0x2220, 0xFE20, 0x2BFC, 0x2220, 0x2220, 0x23FC, 0x2220, 0x3A20, 0xE220, 0x43FE, 0x200, 0x200}},
    {22535, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0}},
    {22536, 2, {0x2000, 0x23FE, 0x2202, 0x228A, 0x2252, 0xFBFE, 0x2222, 0x2222, 0x22AA, 0x22AA, 0x22AA, 0x3AFA, 0xE202, 0x4202, 0x20A, 0x204}},
    {22537, 2, {0x2040, 0x2020, 0x23FE, 0x2080, 0x2104, 0xFBFE, 0x2002, 0x21FC, 0x2104, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x4104, 0x114, 0x108}},
    {22538, 2, {0x0, 0x7FFC, 0x440, 0x3C78, 0x2008, 0x2008, 0x3C78, 0x440, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22539, 2, {0x2000, 0x23DE, 0x2252, 0x2252, 0x2252, 0xFBDE, 0x2252, 0x2252, 0x2252, 0x23DE, 0x2252, 0x3A52, 0xE252, 0x4252, 0x552, 0x8A6}},
    {22540, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0xFC44, 0x2444, 0x25F4, 0x2514, 0x2514, 0x2514, 0x3DF4, 0xE514, 0x4404, 0x7FC, 0x404}},
    {22541, 2, {0x2100, 0x2100, 0x21F0, 0x2210, 0xFA20, 0x27FC, 0x2A44, 0x2244, 0x2244, 0x23FC, 0x22A4, 0x38B0, 0xE128, 0x422A, 0x422, 0x81E}},
    {22542, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0x2020, 0xFBFE, 0x2088, 0x2144, 0x2242, 0x20F8, 0x2188, 0x3A50, 0xE020, 0x4050, 0x188, 0x606}},
    {22543, 2, {0x820, 0x810, 0x9FE, 0x7E40, 0x187C, 0x2C44, 0x2A84, 0x4894, 0x8908, 0x800, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22544, 2, {0x1000, 0x13FE, 0x1210, 0x1210, 0x12FE, 0xFE10, 0x1210, 0x13FE, 0x1200, 0x1210, 0x1210, 0x1EFE, 0xE210, 0x4410, 0x5FE, 0x800}},
    {22545, 2, {0x100C, 0xFEF0, 0x2080, 0x4880, 0x7EFE, 0x888, 0xE88, 0xF908, 0x4A08, 0x908, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22546, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1000, 0x1112, 0x11D4, 0x1D18, 0xE110, 0x4152, 0x192, 0x10E}},
    {22547, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0xFC00, 0x11FE, 0x1100, 0x117C, 0x1110, 0x1110, 0x1DFE, 0xE110, 0x4210, 0x210, 0x410}},
    {22548, 2, {0x1000, 0x13FE, 0x1202, 0x1252, 0x1088, 0xFD04, 0x1020, 0x1020, 0x13FE, 0x1020, 0x1070, 0x1CA8, 0xE124, 0x4222, 0x20, 0x20}},
    {22549, 2, {0x20, 0x7BFE, 0x4840, 0x48FC, 0x5184, 0x4AFC, 0x4884, 0x68FC, 0x5084, 0x418C, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22550, 2, {0x2124, 0x2248, 0x2490, 0x2248, 0x2124, 0xF880, 0x2100, 0x27FC, 0x2404, 0x2514, 0x24A4, 0x3C44, 0xE4A4, 0x4514, 0x7FC, 0x404}},
    {22551, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1088, 0xFD04, 0x1222, 0x1028, 0x1024, 0x13FE, 0x1020, 0x1C50, 0xE050, 0x4088, 0x104, 0x202}},
    {22552, 2, {0x2040, 0x2248, 0x2150, 0x2040, 0x23F8, 0xF880, 0x27FC, 0x2110, 0x2248, 0x2444, 0x3842, 0xE3F8, 0x4040, 0x40, 0x7FC, 0x0}},
    {22553, 2, {0x2000, 0x27FC, 0x20A0, 0x20A0, 0x27FC, 0xFCA4, 0x24A4, 0x24A4, 0x27FC, 0x2040, 0x3840, 0xE7FC, 0x4040, 0x40, 0xFFE, 0x0}},
    {22554, 2, {0x1000, 0x13FE, 0x1222, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {22555, 2, {0x2000, 0x23FE, 0x2000, 0x21FC, 0x2104, 0xF904, 0x21FC, 0x2000, 0x23FE, 0x2222, 0x2222, 0x3BFE, 0xE222, 0x4222, 0x3FE, 0x202}},
    {22556, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x11FC, 0xFD24, 0x11AC, 0x1174, 0x1124, 0x11FC, 0x1020, 0x1C70, 0xE0A8, 0x4124, 0x222, 0x20}},
    {22557, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x11E4, 0xFD24, 0x1124, 0x13FE, 0x1202, 0x12FA, 0x1E8A, 0xE28A, 0x42FA, 0x202, 0x20A, 0x204}},
    {22558, 2, {0x2090, 0x2290, 0x2290, 0x27FE, 0x2290, 0xFA90, 0x22F0, 0x2200, 0x23FC, 0x2040, 0x27FE, 0x38E0, 0xE150, 0x4248, 0xC46, 0x40}},
    {22559, 2, {0x2080, 0x20FC, 0x2104, 0x21F8, 0x2008, 0xFBFE, 0x2040, 0x20A2, 0x2334, 0x2058, 0x2094, 0x3B34, 0xE052, 0x4090, 0x350, 0x20}},
    {22560, 2, {0x2100, 0x2178, 0x2108, 0x2208, 0xFAFE, 0x2640, 0x2A40, 0x227C, 0x2290, 0x2210, 0x22FE, 0x3A10, 0xE228, 0x4228, 0x244, 0x282}},
    {22561, 2, {0x800, 0xBF8, 0x1208, 0x33F8, 0x5040, 0x97FC, 0x1150, 0x1248, 0x1446, 0x1140, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22562, 2, {0x2100, 0x217C, 0x2144, 0x2244, 0xFA44, 0x267C, 0x2A10, 0x2210, 0x22FE, 0x2238, 0x2254, 0x3A54, 0xE294, 0x4312, 0x210, 0x210}},
    {22563, 2, {0x11FC, 0x1124, 0x1124, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1222, 0x1E2A, 0xE2FA, 0x420A, 0x202, 0x206}},
    {22564, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1000, 0x13FE, 0x1020, 0x1D20, 0xE13C, 0x4120, 0x2A0, 0x47E}},
    {22565, 2, {0x40, 0x7F40, 0x127E, 0xC88, 0xFF48, 0x1948, 0x2A50, 0x4820, 0xA850, 0x118C, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22566, 2, {0x1110, 0x1112, 0x11D4, 0x1118, 0x1152, 0xFD92, 0x112E, 0x1040, 0x11FC, 0x1104, 0x1104, 0x1DFC, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22567, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x11FC, 0xFD54, 0x1154, 0x1154, 0x1154, 0x112C, 0x1C20, 0xE3FE, 0x4050, 0x88, 0x104, 0x202}},
    {22568, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0xFDFC, 0x1080, 0x11FE, 0x1222, 0x1122, 0x1152, 0x1D02, 0xE1FA, 0x4002, 0x14, 0x8}},
    {22569, 2, {0x2100, 0x213C, 0x2100, 0x2100, 0xF9BC, 0x2564, 0x2524, 0x253C, 0x2524, 0x2124, 0x2124, 0x393C, 0xE124, 0x4100, 0x17E, 0x100}},
    {22570, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x10F8, 0xFC88, 0x10F8, 0x1088, 0x1088, 0x13FE, 0x1100, 0x1D48, 0xE184, 0x4100, 0x1FE, 0x0}},
    {22571, 2, {0x2000, 0x2088, 0x2252, 0x2222, 0x2252, 0xFA8A, 0x2202, 0x23FE, 0x2088, 0x2144, 0x227A, 0x3888, 0xE150, 0x4020, 0xD8, 0x706}},
    {22572, 2, {0x2040, 0x20A0, 0x2110, 0x2208, 0x25F6, 0xF800, 0x23C4, 0x2254, 0x2254, 0x23D4, 0x2254, 0x3A54, 0xE3D4, 0x4244, 0x254, 0x2C8}},
    {22573, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1000, 0x13FE, 0x1020, 0x1020, 0x1DFC, 0xE020, 0x4020, 0x3FE, 0x0}},
    {22574, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0x13DE, 0xFC00, 0x11FC, 0x1000, 0x13FE, 0x1080, 0x1100, 0x1DFC, 0xE004, 0x4004, 0x28, 0x10}},
    {22575, 2, {0x100, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x820, 0x3EF8, 0x820, 0xE20, 0x78FC, 0x0, 0xFFFE, 0x440, 0x842, 0x1042, 0x603E}},
    {22576, 2, {0x2000, 0x23FE, 0x2200, 0x22FC, 0x2284, 0xFAFC, 0x2284, 0x22FC, 0x2220, 0x23FE, 0x2248, 0x3AC8, 0xE230, 0x4248, 0x284, 0x3FE}},
    {22577, 2, {0x1000, 0x10FC, 0x7C84, 0x1084, 0x1094, 0xFE88, 0x4480, 0x28FC, 0x7CA4, 0x10A4, 0x10A8, 0xFEA8, 0x1090, 0x10A8, 0x10C4, 0x1082}},
    {22578, 2, {0x7E7C, 0x4244, 0x7E44, 0x4244, 0x7E44, 0x4844, 0x4454, 0x5A48, 0x6240, 0x4040, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22579, 2, {0x2000, 0x23FC, 0x2224, 0x2224, 0x23FC, 0xFA00, 0x22FC, 0x2284, 0x2284, 0x22FC, 0x2284, 0x3AFC, 0xE484, 0x4484, 0x8FC, 0x1084}},
    {22580, 2, {0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0xFBF8, 0x2000, 0x2FFE, 0x2200, 0x23FC, 0x2494, 0x3924, 0xE244, 0x4484, 0x128, 0x210}},
    {22581, 2, {0x2040, 0x2044, 0x23F4, 0x2048, 0xF850, 0x27FE, 0x2040, 0x2080, 0x21F8, 0x2B08, 0x3508, 0x21F8, 0xC108, 0x108, 0x1F8, 0x108}},
    {22582, 2, {0x2040, 0x2040, 0x23FE, 0x2080, 0x21FC, 0xFA20, 0x25FE, 0x2000, 0x21FC, 0x2104, 0x21FC, 0x3904, 0xE1FC, 0x4104, 0x114, 0x108}},
    {22583, 2, {0x2080, 0x2040, 0x27FC, 0x2000, 0x2208, 0xF910, 0x2FFE, 0x2000, 0x23F8, 0x2208, 0x2208, 0x3BF8, 0xE208, 0x4208, 0x3F8, 0x208}},
    {22584, 2, {0x2000, 0x2FF8, 0x2808, 0x2868, 0xFB88, 0x2888, 0x2888, 0x2BE8, 0x2AA8, 0x2AA8, 0x3BE8, 0xE88A, 0x48AA, 0x10EA, 0x1726, 0x2002}},
    {22585, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1C20, 0xE1FC, 0x4020, 0x3FE, 0x0}},
    {22586, 2, {0x23F8, 0x2248, 0x2248, 0x23F8, 0x2248, 0xFA48, 0x23F8, 0x20A0, 0x2110, 0x2208, 0x2516, 0x3910, 0xE110, 0x4110, 0x210, 0x410}},
    {22587, 2, {0x2040, 0x13F8, 0x9048, 0x4FFE, 0x4048, 0x13F8, 0x2040, 0xE3FC, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22588, 2, {0x808, 0x808, 0x7E08, 0x8FE, 0xFE08, 0x848, 0x7E28, 0x808, 0xF28, 0xF010, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22589, 2, {0x1000, 0x13FE, 0x1252, 0x1252, 0x13FE, 0xFC00, 0x1000, 0x13FE, 0x1020, 0x1020, 0x113C, 0x1D20, 0xE120, 0x4120, 0x7FE, 0x0}},
    {22590, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x23F8, 0xF880, 0x27FE, 0x2110, 0x2208, 0x27FC, 0x3A0A, 0xE208, 0x43F8, 0x208, 0x208, 0x3F8}},
    {22591, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0x2410, 0xFC10, 0x25D0, 0x2412, 0x2412, 0x25D4, 0x2554, 0x3D48, 0xE5DA, 0x442A, 0x846, 0x1082}},
    {22592, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0x23FC, 0xFA00, 0x2284, 0x2248, 0x22FC, 0x2248, 0x2248, 0x3BFE, 0xE448, 0x4448, 0x888, 0x1108}},
    {22593, 2, {0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x600C, 0x1830, 0x6C0, 0x1830, 0x610C, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22594, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF110, 0x2FFE, 0x2110, 0x2208, 0x27F4, 0x2A12, 0x3210, 0xE3F0, 0x4204, 0x204, 0x1FC, 0x0}},
    {22595, 2, {0x2108, 0x2108, 0x27FE, 0x2148, 0xF820, 0x27FE, 0x2200, 0x2200, 0x23FC, 0x2000, 0x2248, 0x3A48, 0xE248, 0x444A, 0x44A, 0x846}},
    {22596, 2, {0x2000, 0x23FC, 0x2294, 0x2294, 0x2294, 0xFBFC, 0x2080, 0x2040, 0x27FE, 0x2080, 0x20F8, 0x3888, 0xE108, 0x4108, 0x228, 0x410}},
    {22597, 2, {0x2080, 0x2338, 0x2228, 0x2228, 0x23A8, 0xFA46, 0x2200, 0x227C, 0x23A4, 0x2224, 0x2228, 0x3BA8, 0xE610, 0x4228, 0x244, 0x282}},
    {22598, 2, {0x2080, 0x2040, 0x27FE, 0x2090, 0xFA94, 0x2492, 0x2000, 0x23F8, 0x2008, 0x23F8, 0x2200, 0x3BFC, 0xE004, 0x4004, 0x28, 0x10}},
    {22599, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1404, 0xFDFC, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE104, 0x4000, 0x3FE, 0x0}},
    {22600, 2, {0x0, 0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22601, 2, {0x2040, 0x2248, 0x2150, 0x2248, 0x24A4, 0xF918, 0x2606, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x3BF8, 0xE208, 0x4208, 0x228, 0x210}},
    {22602, 2, {0x2020, 0x2040, 0x23FC, 0x2224, 0x2224, 0xFBFC, 0x2224, 0x2244, 0x23FC, 0x2040, 0x20A8, 0x38B4, 0xE13C, 0x4122, 0x222, 0x41E}},
    {22603, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {22604, 2, {0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0xFBF8, 0x2000, 0x2000, 0x27BC, 0x2084, 0x24A4, 0x3A94, 0xE4A4, 0x4084, 0x294, 0x108}},
    {22605, 2, {0x20, 0x7924, 0x48A8, 0x4820, 0x49FC, 0x7840, 0x4BFE, 0x4888, 0x4924, 0x7A22, 0x4C20, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x9800}},
    {22606, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1050, 0xFC88, 0x1124, 0x1050, 0x1088, 0x1104, 0x12FA, 0x1C88, 0xE088, 0x4088, 0xF8, 0x88}},
    {22607, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0xFC00, 0x13FE, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1DFC, 0xE088, 0x4050, 0x3FE, 0x0}},
    {22608, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC20, 0x13FE, 0x1040, 0x1088, 0x11F0, 0x1C24, 0xE3FE, 0x4022, 0x124, 0x2A2, 0x40}},
    {22609, 2, {0x2200, 0x147C, 0xFF44, 0x87C, 0x4944, 0x497C, 0x7F44, 0x844, 0x1094, 0x2108, 0x4100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22610, 2, {0x2010, 0x2010, 0x2710, 0x257C, 0x2510, 0xF510, 0x25FE, 0x2708, 0x2508, 0x25FE, 0x2508, 0x3548, 0xE728, 0x4008, 0x28, 0x10}},
    {22611, 2, {0x1000, 0x13FE, 0x1202, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x1104, 0x11FC, 0x1040, 0x1C20, 0xE3FE, 0x4000, 0x88, 0x104, 0x202}},
    {22612, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0x2040, 0xF8A0, 0x2110, 0x2208, 0x25F6, 0x2000, 0x2000, 0x3BF8, 0xE208, 0x4208, 0x3F8, 0x208}},
    {22613, 2, {0x2120, 0x2110, 0x2208, 0x2486, 0x2910, 0xFBF8, 0x2008, 0x2000, 0x27BC, 0x2084, 0x24A4, 0x3A94, 0xE4A4, 0x4084, 0x294, 0x108}},
    {22614, 2, {0x2038, 0x23C0, 0x2040, 0x27FC, 0x2150, 0xF954, 0x2758, 0x2154, 0x2354, 0x254C, 0x20E0, 0x3950, 0xE248, 0x4444, 0x842, 0x40}},
    {22615, 2, {0x2040, 0x10A0, 0x9110, 0x4208, 0x4DF6, 0x1040, 0x27FC, 0xE040, 0x2248, 0x2444, 0x2942, 0x2080, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22616, 2, {0x2040, 0x2020, 0x27FE, 0x2420, 0x25FC, 0xFC24, 0x27FE, 0x2424, 0x25FC, 0x2420, 0x25FC, 0x3D04, 0xE504, 0x4904, 0x9FC, 0x1104}},
    {22617, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x1000, 0x13FE, 0x1202, 0x1EFA, 0xE28A, 0x42FA, 0x202, 0x20A, 0x204}},
    {22618, 2, {0x2000, 0x27FE, 0x2402, 0x2804, 0x27FE, 0xF880, 0x2184, 0x2244, 0x26A8, 0x2130, 0x3AF0, 0xE4A8, 0x4128, 0x224, 0xCA2, 0x40}},
    {22619, 2, {0x2000, 0x27FC, 0x20A0, 0x20A0, 0x27FC, 0xFCA4, 0x24A4, 0x27FC, 0x2040, 0x2040, 0x27FC, 0x38E0, 0xE150, 0x4248, 0xC46, 0x40}},
    {22620, 2, {0x2000, 0x27FE, 0x2402, 0x29F4, 0x2000, 0xFBF8, 0x2080, 0x2144, 0x2668, 0x20B0, 0x3928, 0xE668, 0x40A4, 0x122, 0x6A0, 0x40}},
    {22621, 2, {0x2040, 0x2020, 0x23FC, 0x2108, 0x2090, 0xFBFE, 0x2202, 0x2444, 0x2020, 0x23FC, 0x2080, 0x38F8, 0xE088, 0x4108, 0x128, 0x210}},
    {22622, 2, {0x200, 0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1110, 0x2108, 0xDFF6, 0x100, 0x100, 0x7FFC}},
    {22623, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x840, 0x1F48, 0x6250, 0x1462, 0x1842, 0xE03E, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22624, 2, {0x2010, 0x2420, 0x22F8, 0x2288, 0xF888, 0x20F8, 0x2E80, 0x2280, 0x22FC, 0x2284, 0x2284, 0x3AFC, 0xE284, 0x4500, 0x8FE, 0x0}},
    {22625, 2, {0x1118, 0x11E2, 0x1102, 0x10FE, 0x1000, 0xFCFC, 0x1284, 0x12FC, 0x1284, 0x12FC, 0x1284, 0x1EFC, 0xE200, 0x43FE, 0x108, 0x204}},
    {22626, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1100, 0x11FE, 0x1100, 0x11FE, 0x1C02, 0xE2AA, 0x42AA, 0x402, 0x14, 0x8}},
    {22627, 2, {0x2000, 0x13F8, 0x9208, 0x4208, 0x4BF8, 0x800, 0x17FC, 0xE040, 0x23F8, 0x2040, 0x27FC, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22628, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x11F8, 0xFC00, 0x13FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1E04, 0xE3FC, 0x4090, 0x108, 0x204}},
    {22629, 2, {0x2000, 0x23FE, 0x2000, 0x21FC, 0x2104, 0xF904, 0x21FC, 0x2000, 0x23FE, 0x228A, 0x2252, 0x3BFE, 0xE222, 0x4222, 0x22A, 0x204}},
    {22630, 2, {0x7C20, 0x45FE, 0x4820, 0x51FC, 0x4924, 0x45FC, 0x4524, 0x55FC, 0x4820, 0x43FE, 0x4020, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22631, 2, {0x2208, 0x2108, 0x2110, 0x2000, 0x27FC, 0xF800, 0x2110, 0x2208, 0x2404, 0x23F8, 0x22A8, 0x3AA8, 0xE2A8, 0x42A8, 0xFFE, 0x0}},
    {22632, 2, {0x2090, 0x2090, 0x23FC, 0x2090, 0x2090, 0xF890, 0x27FE, 0x2090, 0x2148, 0x2244, 0x244A, 0x3954, 0xE14A, 0x424A, 0x140, 0x80}},
    {22633, 2, {0x2100, 0x2100, 0x21FE, 0x2200, 0x25F8, 0xF908, 0x2108, 0x21F8, 0x2000, 0x23FC, 0x2294, 0x3A94, 0xE294, 0x4294, 0x7FE, 0x0}},
    {22634, 2, {0x101E, 0x11E0, 0x1022, 0x1112, 0x1094, 0xFC80, 0x1020, 0x11CE, 0x1102, 0x1102, 0x11CE, 0x1D02, 0xE102, 0x4102, 0x1FE, 0x102}},
    {22635, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x23F8, 0xFA08, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x3A08, 0xEFFE, 0x4110, 0x208, 0x404}},
    {22636, 2, {0x2000, 0x27FE, 0x2420, 0x2440, 0x25FC, 0xFD04, 0x25FC, 0x2504, 0x25FC, 0x2524, 0x3C20, 0xE4A8, 0x4924, 0xA22, 0x10A0, 0x40}},
    {22637, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x11F8, 0xFD08, 0x1108, 0x11F8, 0x1000, 0x13FC, 0x1294, 0x1E94, 0xE294, 0x4294, 0x7FE, 0x0}},
    {22638, 2, {0x2084, 0x2104, 0x23C4, 0x2244, 0x23DE, 0xFA44, 0x23C4, 0x2264, 0x2254, 0x27D4, 0x20C4, 0x3944, 0xE244, 0x4444, 0x154, 0x88}},
    {22639, 2, {0x2040, 0x23BE, 0x2212, 0x2292, 0x2252, 0xFAAA, 0x2324, 0x2040, 0x23FE, 0x2222, 0x2222, 0x3BFE, 0xE222, 0x4222, 0x3FE, 0x202}},
    {22640, 2, {0x2100, 0x11FC, 0x9200, 0x45F8, 0x4148, 0x1128, 0x23FE, 0xE148, 0x2228, 0x23FC, 0x2050, 0x120, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22641, 2, {0x1000, 0x7E7C, 0x4244, 0x7E7C, 0x4244, 0x7E7C, 0x4844, 0x4444, 0x5A94, 0x6108, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22642, 2, {0x2100, 0x21FC, 0x2200, 0x25F8, 0x2108, 0xF9F8, 0x2108, 0x21F8, 0x2000, 0x27FE, 0x2100, 0x3BFC, 0xE4A4, 0x4124, 0x254, 0x88}},
    {22643, 2, {0x2020, 0x247C, 0x2284, 0x2148, 0x2030, 0xF8CE, 0x2610, 0x227C, 0x2210, 0x227C, 0x2210, 0x3AFE, 0xE210, 0x4210, 0x5FE, 0x800}},
    {22644, 2, {0x2020, 0x2222, 0x2222, 0x23FE, 0x2000, 0xFBDE, 0x2252, 0x2252, 0x23DE, 0x2252, 0x2252, 0x3BDE, 0xE252, 0x4252, 0x55A, 0x8A4}},
    {22645, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2F78, 0x2842, 0x2F3E, 0x2080, 0x2FF8, 0x4080, 0x7FFE, 0x8000}},
    {22646, 2, {0x2040, 0x2020, 0x23FE, 0x2250, 0x2250, 0xFBFE, 0x2252, 0x2252, 0x23FE, 0x2200, 0x2292, 0x3AD4, 0xE298, 0x4492, 0x4D2, 0x88E}},
    {22647, 2, {0x2020, 0x2020, 0x203E, 0x2020, 0x2020, 0xFBFE, 0x2202, 0x2326, 0x228A, 0x2252, 0x2326, 0x3A52, 0xE28A, 0x4326, 0x3FE, 0x202}},
    {22648, 2, {0x2000, 0x23FE, 0x2200, 0x227C, 0x2244, 0xFA44, 0x227C, 0x2200, 0x22EE, 0x22AA, 0x22AA, 0x3AAA, 0xE2EE, 0x4200, 0x3FE, 0x0}},
    {22649, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22650, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0x2FBC, 0x2318, 0x26B4, 0x2A52, 0x2080, 0x2080, 0x2FFC, 0x4080, 0x4080, 0x9FFE, 0x0}},
    {22651, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x1000, 0xFDFC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1C20, 0xE3FE, 0x4050, 0x88, 0x104, 0x202}},
    {22652, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x11FC, 0x1022, 0x13FE, 0x1C08, 0xE3FE, 0x4108, 0x88, 0x28, 0x10}},
    {22653, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x2554, 0xFA48, 0x2554, 0x2040, 0x2554, 0x2248, 0x2554, 0x38A0, 0xE0A0, 0x4110, 0x208, 0xC06}},
    {22654, 2, {0x1040, 0xFE40, 0x40, 0x7DF8, 0x4448, 0x7C48, 0xC8, 0x7C48, 0x8AA, 0xFEAA, 0x1106, 0x5202, 0x2100, 0x3FF8, 0x100, 0xFFFE}},
    {22655, 2, {0x1020, 0x11FC, 0x1124, 0x13FE, 0x1124, 0xFDFC, 0x1020, 0x11FC, 0x1124, 0x11FC, 0x1C40, 0xE3FE, 0x4088, 0x1D0, 0x70, 0x38C}},
    {22656, 2, {0x2000, 0x27FC, 0x2404, 0x27FC, 0x2420, 0xFCA8, 0x2470, 0x24A8, 0x2504, 0x24A0, 0x3CFC, 0xE520, 0x4420, 0xBFE, 0x820, 0x1020}},
    {22657, 2, {0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0xF800, 0x27FC, 0x24A4, 0x24A4, 0x27FC, 0x2000, 0x3BF8, 0xE110, 0x40E0, 0x318, 0xC06}},
    {22658, 2, {0x1000, 0x13FE, 0x1050, 0x11FC, 0x1154, 0xFD54, 0x11FC, 0x1000, 0x11FC, 0x1000, 0x13FE, 0x1C20, 0xE0A8, 0x4124, 0x2A2, 0x40}},
    {22659, 2, {0x2080, 0x2040, 0x23F8, 0x2110, 0xF8A0, 0x27FE, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x3BF8, 0xE120, 0x4122, 0x222, 0xC1E}},
    {22660, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0x2410, 0xFC90, 0x24D2, 0x2492, 0x27F2, 0x2494, 0x3C94, 0xE5C8, 0x4AAA, 0x89A, 0x11A6, 0x42}},
    {22661, 2, {0x7F7C, 0x4904, 0x7F28, 0x4910, 0x7F7E, 0x812, 0x7F14, 0x810, 0xF50, 0xF020, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22662, 2, {0x22A8, 0x22A8, 0x27FC, 0x22A8, 0xFAAA, 0x24E6, 0x2800, 0x27FC, 0x2444, 0x2040, 0x23F8, 0x3A48, 0xE248, 0x4248, 0x258, 0x40}},
    {22663, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0x13FE, 0xFC00, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {22664, 2, {0x2248, 0x2248, 0x2FE8, 0x2248, 0x23DE, 0xF24A, 0x23CA, 0x224A, 0x224A, 0x2FEA, 0x240A, 0x354A, 0xE632, 0x4412, 0x7EA, 0x44}},
    {22665, 2, {0x2020, 0x27FE, 0x2420, 0x25FC, 0xF424, 0x27FE, 0x2424, 0x25FC, 0x2420, 0x25FC, 0x3524, 0xE5FC, 0x4524, 0x9FC, 0x924, 0x112C}},
    {22666, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x882, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22667, 2, {0x2040, 0x2088, 0x21FC, 0x2108, 0x2252, 0xFBFE, 0x2050, 0x2188, 0x2626, 0x20C0, 0x2310, 0x3864, 0xE388, 0x4030, 0xC0, 0x700}},
    {22668, 2, {0x2040, 0x2020, 0x23FE, 0x2200, 0x2248, 0xFA48, 0x23FE, 0x2248, 0x2248, 0x2248, 0x3A78, 0xE200, 0x4554, 0x52A, 0xA2A, 0x1000}},
    {22669, 2, {0x1000, 0x7EFC, 0x4210, 0x7E90, 0x4290, 0x7EFE, 0x10, 0x4028, 0x7E28, 0x4048, 0x4C8A, 0x7106, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22670, 2, {0x2200, 0x211E, 0x27D2, 0x2012, 0x27D4, 0xFC54, 0x27D8, 0x2014, 0x27D2, 0x2092, 0x2112, 0x39DA, 0xE714, 0x4110, 0x510, 0x210}},
    {22671, 2, {0x420, 0x247E, 0x2494, 0x3D48, 0x430, 0x5C8, 0x7C08, 0x25FE, 0x2488, 0x2448, 0x4418, 0x8100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22672, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x10F8, 0xFC20, 0x11FC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1C20, 0xE1FC, 0x4020, 0x3FE, 0x0}},
    {22673, 2, {0x2040, 0x2020, 0x23FE, 0x2088, 0x2050, 0xFBFE, 0x2222, 0x22FA, 0x2222, 0x22FA, 0x228A, 0x3A8A, 0xE2FA, 0x4202, 0x20A, 0x204}},
    {22674, 2, {0x2040, 0x2020, 0x23FE, 0x2088, 0x2050, 0xFBFE, 0x2252, 0x228A, 0x2306, 0x22FA, 0x228A, 0x3A8A, 0xE2FA, 0x4202, 0x20A, 0x204}},
    {22675, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1010, 0x2108, 0xCFE6, 0x100, 0x100, 0x3FF8}},
    {22676, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0x1088, 0xFDFE, 0x1310, 0x1510, 0x11FE, 0x1110, 0x1110, 0x1DFE, 0xE110, 0x4110, 0x1FE, 0x100}},
    {22677, 2, {0x2000, 0x27FC, 0x2040, 0x2278, 0xF240, 0x2FFE, 0x2100, 0x21FC, 0x2200, 0x23FC, 0x3004, 0xE554, 0x4554, 0x804, 0x28, 0x10}},
    {22678, 2, {0x2040, 0x20A0, 0x2110, 0x2208, 0x2DF6, 0xF000, 0x23F8, 0x2208, 0x23F8, 0x2000, 0x27FC, 0x3444, 0xE7FC, 0x4444, 0x7FC, 0x404}},
    {22679, 2, {0x2104, 0x2088, 0x2000, 0x23FE, 0x2222, 0xFBFE, 0x2222, 0x2222, 0x23FE, 0x2000, 0x21FC, 0x3904, 0xE1FC, 0x4104, 0x1FC, 0x104}},
    {22680, 2, {0x2110, 0x2110, 0x2110, 0x2FDE, 0xF120, 0x27C0, 0x245C, 0x27C4, 0x2448, 0x27C8, 0x3110, 0xEFD0, 0x4122, 0x122, 0x11E, 0x100}},
    {22681, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x2248, 0xF950, 0x27FE, 0x2000, 0x23F8, 0x2208, 0x22E8, 0x3AA8, 0xE2E8, 0x4208, 0x3F8, 0x208}},
    {22682, 2, {0x2278, 0x2128, 0x246C, 0x22AA, 0x202A, 0xF948, 0x2298, 0x2040, 0x2040, 0x27FC, 0x20E0, 0x3950, 0xE248, 0x4C46, 0x40, 0x40}},
    {22683, 2, {0x2004, 0x27C4, 0x2444, 0x2444, 0x27DE, 0xFC04, 0x25C4, 0x2404, 0x2414, 0x27EC, 0x3C84, 0xE4C4, 0x4AA4, 0xA84, 0x1494, 0x188}},
    {22684, 2, {0x88, 0x7C50, 0x45FC, 0x4840, 0x50A2, 0x4B54, 0x44B8, 0x6754, 0x5892, 0x4350, 0x4020, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22685, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0x23FE, 0xF908, 0x239C, 0x2108, 0x2188, 0x263E, 0x2000, 0x3BFE, 0xE090, 0x4090, 0x112, 0x20E}},
    {22686, 2, {0x2208, 0x2110, 0x2000, 0x27FC, 0x2444, 0xFD54, 0x24E4, 0x2444, 0x27FC, 0x2000, 0x23F8, 0x3A08, 0xE3F8, 0x4208, 0x3F8, 0x208}},
    {22687, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0x2242, 0xFA78, 0x23C4, 0x223C, 0x2200, 0x2250, 0x2254, 0x3B54, 0xE4D8, 0x4450, 0x9FE, 0x0}},
    {22688, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0x13DE, 0xFC00, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {22689, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0x11FC, 0xFC20, 0x13FE, 0x1124, 0x10A8, 0x13FE, 0x1000, 0x1DFC, 0xE104, 0x4104, 0x1FC, 0x104}},
    {22690, 2, {0x2024, 0x23A8, 0x2092, 0x2514, 0xF208, 0x2C04, 0x27BA, 0x20A8, 0x27C6, 0x2400, 0x37BC, 0xE084, 0x40A8, 0x90, 0x528, 0x244}},
    {22691, 2, {0x20A0, 0x24A4, 0x22A8, 0x20A0, 0x27FC, 0xF910, 0x20A0, 0x27FC, 0x2040, 0x23F8, 0x3840, 0xE7FC, 0x40A0, 0x110, 0x208, 0xC06}},
    {22692, 2, {0x2040, 0x23F8, 0x2248, 0x23F8, 0xF840, 0x27FE, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x3BF8, 0xE208, 0x43F8, 0x110, 0x208}},
    {22693, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0x13FE, 0xFC00, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1C20, 0xE1FC, 0x4020, 0x3FE, 0x0}},
    {22694, 2, {0x103C, 0x13E0, 0x1124, 0x10A8, 0x13FE, 0xFCA8, 0x1124, 0x1202, 0x11FC, 0x1124, 0x1124, 0x1DFC, 0xE124, 0x4124, 0x1FC, 0x104}},
    {22695, 2, {0x2038, 0x23C0, 0x2080, 0x27FE, 0x2110, 0xFA08, 0x25F6, 0x2910, 0x21F0, 0x2000, 0x27FC, 0x3C04, 0xE5F4, 0x4514, 0x5F4, 0x40C}},
    {22696, 2, {0x0, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x2448, 0x4224, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22697, 2, {0x2210, 0x2110, 0x2FD0, 0x2010, 0x27BE, 0xFCA4, 0x27D4, 0x2014, 0x2794, 0x2094, 0x2114, 0x39C8, 0xEF08, 0x4114, 0x514, 0x222}},
    {22698, 2, {0x840, 0x7F40, 0x7E, 0x3E88, 0x2348, 0x3E48, 0x50, 0x3E50, 0x420, 0x7F50, 0x888, 0x2904, 0x1100, 0x3FF8, 0x100, 0xFFFE}},
    {22699, 2, {0x2108, 0x2090, 0x27FE, 0x2090, 0x23FC, 0xFA94, 0x231C, 0x2204, 0x23FC, 0x2204, 0x3BFC, 0xE008, 0x47FE, 0x108, 0xA8, 0x10}},
    {22700, 2, {0x40, 0x3C78, 0x2488, 0x28F0, 0x3010, 0x2BFE, 0x2440, 0x25A4, 0x3458, 0x29B4, 0x2052, 0x21B0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22701, 2, {0x2028, 0x2024, 0x23FE, 0x2220, 0x2224, 0xFBA4, 0x2298, 0x2490, 0x25AA, 0x2846, 0x23FA, 0x3AA8, 0xE2A8, 0x42A8, 0xFFE, 0x0}},
    {22702, 2, {0x40, 0x7BFC, 0x4880, 0x51F8, 0x6220, 0x55FC, 0x4888, 0x48F8, 0x6888, 0x50F8, 0x4088, 0x4100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22703, 2, {0x1040, 0x13FE, 0x1880, 0x55F8, 0x5220, 0x95FC, 0x1088, 0x10F8, 0x1088, 0x10F8, 0x1088, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22704, 2, {0x2000, 0x23FE, 0x2050, 0x23FE, 0x2252, 0xFBFE, 0x2000, 0x21FC, 0x2104, 0x21FC, 0x2104, 0x39FC, 0xE020, 0x43FE, 0x20, 0x20}},
    {22705, 2, {0x2020, 0x27A4, 0x20A8, 0x2292, 0x2114, 0xFA08, 0x25F4, 0x2802, 0x23F8, 0x2208, 0x2208, 0x3BF8, 0xE208, 0x4110, 0xFFE, 0x0}},
    {22706, 2, {0x2100, 0x2100, 0x23FC, 0x26A8, 0x2AA8, 0xF2A8, 0x27FC, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x3000, 0xE2A8, 0x4254, 0x454, 0x0}},
    {22707, 2, {0x1020, 0x11FC, 0x1020, 0x1088, 0x13FE, 0xFC88, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1D04, 0xE1FC, 0x4088, 0x104, 0x202}},
    {22708, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0x2110, 0xF7FE, 0x2040, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x37FC, 0xE000, 0x4110, 0x208, 0x404}},
    {22709, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0x1222, 0xFDAC, 0x1020, 0x11AC, 0x1000, 0x11FC, 0x1C00, 0xE3FE, 0x4040, 0x88, 0x1FC, 0x84}},
    {22710, 2, {0x2010, 0x2410, 0x227C, 0x2010, 0x20FE, 0xF844, 0x2628, 0x22FE, 0x2210, 0x227C, 0x2210, 0x3AFE, 0xE210, 0x4210, 0x5FE, 0x800}},
    {22711, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0x11FC, 0xFCA8, 0x10A8, 0x13FE, 0x10A8, 0x10A8, 0x11FC, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {22712, 2, {0x2108, 0x2108, 0x27FE, 0x2108, 0x2044, 0xFBF4, 0x2048, 0x2050, 0x27FE, 0x2080, 0x21FC, 0x3B04, 0xE5FC, 0x4904, 0x1FC, 0x104}},
    {22713, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0x24A4, 0xFFBC, 0x2404, 0x25F4, 0x2514, 0x25F4, 0x2514, 0x3DF4, 0xE514, 0x4404, 0x414, 0x408}},
    {22714, 2, {0x2080, 0x2100, 0x27FC, 0x2444, 0x2554, 0xFC44, 0x27FC, 0x24E4, 0x2554, 0x2404, 0x3840, 0xE7FE, 0x40A0, 0x110, 0x208, 0xC06}},
    {22715, 2, {0x2040, 0x2040, 0x27FC, 0x2248, 0xFA48, 0x2554, 0x2FFE, 0x2000, 0x23F8, 0x2208, 0x22E8, 0x3AA8, 0xE2E8, 0x4208, 0x3F8, 0x208}},
    {22716, 2, {0x800, 0xFF78, 0x848, 0x7F48, 0x4986, 0x7F00, 0x4978, 0x7F48, 0x828, 0xFF90, 0x4928, 0x7F44, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22717, 2, {0x2210, 0x2410, 0x2F90, 0x2890, 0x2FBE, 0xF8A4, 0x2FD4, 0x2414, 0x2214, 0x2FD4, 0x2414, 0x3788, 0xE488, 0x4894, 0xA94, 0x1122}},
    {22718, 2, {0x600, 0xF8FC, 0x4A84, 0x24FC, 0x3884, 0xCCFC, 0x36A2, 0xCA94, 0x12A8, 0x6AC4, 0x482, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22719, 2, {0x2000, 0x27FC, 0x24A4, 0x27FC, 0x2040, 0xFBF8, 0x2040, 0x27FC, 0x2110, 0x20A0, 0x3BF8, 0xE040, 0x4FFE, 0x40, 0x40, 0x40}},
    {22720, 2, {0x2010, 0x2008, 0x2788, 0x24BE, 0xF480, 0x2494, 0x2788, 0x247E, 0x2408, 0x2788, 0x36BE, 0xEA88, 0x4A88, 0xB88, 0x1288, 0x8}},
    {22721, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3D20, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22722, 2, {0x3F78, 0x2148, 0x3F48, 0x2086, 0x2A78, 0x3F48, 0x2A48, 0x3F28, 0x2010, 0x4A28, 0x9146, 0x2000, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22723, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x13FE, 0x1000, 0x11FC, 0x1124, 0x1DFC, 0xE124, 0x41FC, 0x0, 0x3FE}},
    {22724, 2, {0x0, 0x2788, 0x2088, 0xFABE, 0x2108, 0x77D8, 0x6B5C, 0xA52A, 0x2948, 0x2508, 0x2208, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22725, 2, {0x200, 0x100, 0xFFFE, 0x1048, 0x24FE, 0x7990, 0x10FC, 0x2290, 0x7CFC, 0x890, 0x30FE, 0xC080, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22726, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {22727, 2, {0x2040, 0x2FFE, 0x2000, 0x27FC, 0x2404, 0xFDF4, 0x2514, 0x27FC, 0x2000, 0x23F8, 0x2208, 0x3BF8, 0xE208, 0x43F8, 0x0, 0xFFE}},
    {22728, 2, {0x2040, 0x2FFE, 0x2000, 0x27FC, 0x2404, 0xFDF4, 0x2514, 0x25F4, 0x2404, 0x27FC, 0x2008, 0x3BF0, 0xE040, 0x4FFE, 0x150, 0xE4E}},
    {22729, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0x2242, 0xFA78, 0x23C4, 0x223C, 0x2200, 0x23FE, 0x2240, 0x3BA4, 0xE258, 0x45B4, 0x452, 0x9B0}},
    {22730, 2, {0x2020, 0x2020, 0x27FE, 0x2020, 0x23FE, 0xFA52, 0x2252, 0x23FE, 0x2040, 0x2020, 0x27FE, 0x38A4, 0xE1A8, 0x4290, 0x488, 0xC6}},
    {22731, 2, {0x2040, 0x2248, 0x2150, 0x27FE, 0x2402, 0xF804, 0x23F8, 0x2208, 0x23F8, 0x2000, 0x27FC, 0x3C44, 0xE7FC, 0x4444, 0x7FC, 0x404}},
    {22732, 2, {0x2080, 0x2040, 0x27FE, 0x2110, 0xFA08, 0x2514, 0x23F8, 0x2110, 0x23F8, 0x2110, 0x27FC, 0x3928, 0xE310, 0x4D48, 0x186, 0x100}},
    {22733, 2, {0x420C, 0x2FB0, 0x8220, 0x4FA0, 0xABE, 0x2FA8, 0x4AA8, 0xCFA8, 0x4228, 0x5FA8, 0x4248, 0x288, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22734, 2, {0x203C, 0x27C0, 0x2040, 0x2FFE, 0xF040, 0x27FC, 0x2554, 0x24E4, 0x27FC, 0x2040, 0x37FC, 0xE040, 0x4FFE, 0x0, 0x524, 0x892}},
    {22735, 2, {0x2010, 0x27D0, 0x2510, 0x27DE, 0xFC50, 0x27E8, 0x2504, 0x27C4, 0x2000, 0x23F8, 0x22A8, 0x3AA8, 0xE2A8, 0x42A8, 0xFFE, 0x0}},
    {22736, 2, {0x7FFC, 0x1110, 0x2108, 0x7FFC, 0xB55A, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3458, 0x2018, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22737, 2, {0x800, 0xF00, 0x87C, 0xFFA4, 0x80A4, 0x7F24, 0x4924, 0x94A8, 0x2210, 0x7F28, 0xA244, 0x3E02, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22738, 2, {0x2110, 0x27FC, 0x2110, 0x2040, 0x23F8, 0xF840, 0x27FC, 0x2080, 0x2110, 0x23F8, 0x2000, 0x3BF8, 0xE2A8, 0x42A8, 0xFFE, 0x0}},
    {22739, 2, {0x0, 0x3FFE, 0x288C, 0x2F8A, 0x2888, 0x2FBE, 0x2008, 0x2F88, 0x2888, 0x2F94, 0x2894, 0x29A2, 0x4040, 0x47FC, 0x8040, 0x1FFE}},
    {22740, 2, {0x2040, 0x27FC, 0x2040, 0x23FC, 0x2000, 0xF7FE, 0x2002, 0x23F8, 0x2040, 0x27FE, 0x2000, 0x37FE, 0xE008, 0x47FE, 0x4A8, 0x798}},
    {22741, 2, {0x2040, 0x27FE, 0x2000, 0x21F8, 0xF908, 0x21F8, 0x2000, 0x27FE, 0x2402, 0x23F8, 0x2980, 0x3644, 0xE1B8, 0x4668, 0x1A6, 0xE60}},
    {22742, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0x1222, 0xFDAC, 0x1020, 0x11AC, 0x1000, 0x13FE, 0x1020, 0x1DFC, 0xE154, 0x4154, 0x154, 0x10C}},
    {22743, 2, {0x2040, 0x23F8, 0x2048, 0x27FE, 0x2048, 0xFBF8, 0x2040, 0x27FE, 0x2000, 0x22A4, 0x2452, 0x3BF8, 0xE2A8, 0x42A8, 0xFFE, 0x0}},
    {22744, 2, {0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0x3EF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22745, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0x25FC, 0xFC88, 0x27FE, 0x2420, 0x25FC, 0x2524, 0x3DFC, 0xE524, 0x45FC, 0x400, 0x488, 0x904}},
    {22746, 2, {0x2040, 0x207C, 0x2040, 0x23FC, 0x2244, 0xFBF0, 0x2244, 0x22FC, 0x22A8, 0x22F8, 0x22A8, 0x3AF8, 0xE200, 0x45FC, 0x554, 0xBFE}},
    {22747, 2, {0x27BC, 0x24A4, 0x27BC, 0x24A4, 0x27BC, 0xFC84, 0x24F4, 0x2514, 0x2624, 0x25B4, 0x2514, 0x3DB4, 0xE514, 0x45F4, 0x414, 0x408}},
    {22748, 2, {0x11FC, 0x1104, 0x11FC, 0x1104, 0x13FE, 0xFC20, 0x13FE, 0x1222, 0x11AC, 0x1020, 0x11FC, 0x1C00, 0xE3FE, 0x4040, 0x88, 0x1FC}},
    {22749, 2, {0x2020, 0x29FC, 0x2524, 0x21FC, 0xF820, 0x23FE, 0x2D04, 0x25FC, 0x2504, 0x25FC, 0x2504, 0x3DFC, 0xE488, 0x4504, 0xBFE, 0x1000}},
    {22750, 2, {0x2080, 0x2040, 0x2FFE, 0x2000, 0x27FC, 0xFCA4, 0x27FC, 0x2248, 0x2150, 0x2208, 0x2444, 0x38A8, 0xE110, 0x4308, 0xD46, 0x180}},
    {22751, 2, {0x840, 0x7E7C, 0x2440, 0xFF7C, 0x4, 0x7E7C, 0x4240, 0x7E7C, 0x4240, 0x7E7C, 0x4242, 0x463E, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22752, 2, {0x2210, 0x2110, 0x27DE, 0x2010, 0x245E, 0xFA82, 0x27DE, 0x2010, 0x27DE, 0x2450, 0x27DE, 0x3C50, 0xE7DE, 0x4450, 0x452, 0x4CE}},
    {22753, 2, {0x1000, 0x1F7C, 0x1024, 0xFFA4, 0x80A8, 0x7F10, 0x2A28, 0x5546, 0xA210, 0x7F10, 0xA2BC, 0x3E10, 0x2210, 0x3E10, 0x22FE, 0x3E00}},
    {22754, 2, {0x2000, 0x27FE, 0x2444, 0x2598, 0x2488, 0xF7DE, 0x2488, 0x25DC, 0x26AA, 0x2488, 0x2420, 0x3520, 0xE53C, 0x4920, 0x920, 0x17FE}},
    {22755, 2, {0x4044, 0x5F44, 0x4AAA, 0x4AEE, 0x4E44, 0xEAAA, 0x4AEE, 0x4E00, 0x4A28, 0x4AAA, 0x6BAA, 0x4EEE, 0x9A28, 0x228, 0x248, 0x288}},
    {22756, 2, {0x2040, 0x27FC, 0x2000, 0x23B8, 0xF2A8, 0x23B8, 0x2110, 0x27FC, 0x2110, 0x27FC, 0x3110, 0xEFFE, 0x4128, 0x310, 0xD48, 0x186}},
    {22757, 2, {0x27FE, 0x2400, 0x25FC, 0x25AC, 0x2574, 0xFDFC, 0x2420, 0x25FC, 0x2420, 0x27FE, 0x3C00, 0xE554, 0x4820, 0x9FC, 0x1020, 0x3FE}},
    {22758, 2, {0x2110, 0x2FFE, 0x2110, 0x27BC, 0x24A4, 0xF7BC, 0x2140, 0x2120, 0x23FE, 0x2220, 0x27FC, 0x3A20, 0xE3FC, 0x4220, 0x3FE, 0x200}},
    {22759, 2, {0x27BC, 0x24A4, 0x27BC, 0x2000, 0x27FE, 0xFC00, 0x2590, 0x249E, 0x27D4, 0x2564, 0x25D4, 0x3D54, 0xE5D4, 0x4568, 0xBC8, 0x1054}},
    {22760, 2, {0x0, 0x3EF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x0, 0x3EF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {22761, 2, {0x27FC, 0x2040, 0x2FFE, 0x2842, 0x2358, 0xF500, 0x2FBC, 0x2524, 0x2724, 0x223C, 0x2FA4, 0x3AA4, 0xEFBC, 0x4224, 0x1FA4, 0x24C}},
    {22762, 2, {0x2444, 0x28E8, 0x2E0E, 0x24E4, 0xFA0A, 0x2EEE, 0x20A0, 0x2AEA, 0x2000, 0x23F8, 0x2008, 0x3BF8, 0xE200, 0x43FC, 0x4, 0x38}},
    {22763, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x0}},
    {22764, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x0}},
    {22765, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3FF8, 0x108, 0x108, 0x128, 0x110, 0x100, 0x100, 0x100}},
    {22766, 2, {0x820, 0x820, 0x820, 0x4820, 0x2820, 0x2820, 0xBFE, 0x1820, 0x2820, 0x4820, 0x8820, 0x820, 0x820, 0x820, 0x9FC, 0x800}},
    {22767, 2, {0x420, 0x2420, 0x2420, 0x2420, 0x2420, 0x3C20, 0x7FE, 0x420, 0xFC20, 0x2420, 0x2420, 0x2420, 0x2420, 0x45FC, 0x4400, 0x8400}},
    {22768, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {22769, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8004, 0x1010, 0x10E0, 0x1F08, 0x1008, 0x1008, 0xFF8}},
    {22770, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8444, 0x440, 0x440, 0x440, 0x844, 0x1044, 0x203C}},
    {22771, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8004, 0xFE0, 0x820, 0x820, 0x1022, 0x2022, 0xC01E}},
    {22772, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {22773, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x0, 0x1010, 0x1010, 0xFEFE, 0x1010, 0x1010, 0x1010, 0x7C7C, 0x0}},
    {22774, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8444, 0x440, 0x2448, 0x1448, 0x1450, 0x440, 0x7FFC}},
    {22775, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4442, 0x8444, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x440, 0x440, 0xFFFE}},
    {22776, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0xBFFC, 0x440, 0x2448, 0x1448, 0x1450, 0x440, 0xFFFE, 0x0}},
    {22777, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE}},
    {22778, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4442, 0x8444, 0x3C78, 0x2008, 0x3C78, 0x440, 0x440, 0xFFFE, 0x0}},
    {22779, 2, {0x1000, 0x13FE, 0x1022, 0x1120, 0x113C, 0xFD20, 0x12FE, 0x1400, 0x11FC, 0x1104, 0x11FC, 0x1D04, 0xE1FC, 0x4104, 0x114, 0x108}},
    {22780, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0xBFFC, 0x440, 0x3C78, 0x2008, 0x3C78, 0x440, 0xFFFE, 0x0}},
    {22781, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0xFFFE, 0x2, 0x3FF0, 0x100, 0xFFFE, 0x0, 0x7FFC, 0x10, 0x7DFC, 0x4490, 0x7C30}},
    {22782, 2, {0x800, 0x87E, 0xFF40, 0x840, 0x840, 0x7E7C, 0x44, 0x7E44, 0x4244, 0x4244, 0x7E7C, 0x4240, 0x2440, 0xF40, 0xF07E, 0x4000}},
    {22783, 2, {0x1088, 0x1050, 0x13FE, 0x1050, 0x11FC, 0xFF54, 0x118C, 0x1174, 0x1104, 0x11FC, 0x1C08, 0x73FE, 0x2108, 0x88, 0xA8, 0x10}},
    {22784, 2, {0x100, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x104, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x8, 0x3EFE, 0x2248, 0x3E18}},
    {22785, 2, {0x100, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x2FE8, 0x2108, 0x27C8, 0x2448, 0x3FF8}},
    {22786, 2, {0x400, 0x400, 0x400, 0xFF0, 0x810, 0x1010, 0x3020, 0x4820, 0x8440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22787, 2, {0x0, 0x3FF0, 0x410, 0x420, 0x420, 0x440, 0x47C, 0x404, 0xBE4, 0x824, 0x944, 0x1084, 0x1144, 0x2224, 0x4C28, 0x8010}},
    {22788, 2, {0x1040, 0x1040, 0x1040, 0x1E40, 0x1260, 0x2250, 0x2248, 0x5244, 0x9444, 0x1440, 0x840, 0x840, 0x1440, 0x2300, 0x40FE, 0x8000}},
    {22789, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x6440, 0x380, 0x1C70, 0xE10E, 0x100, 0x1FF0, 0x100, 0x2100, 0x3FF8, 0x100, 0x100, 0x100}},
    {22790, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x6440, 0x380, 0x1C70, 0xE10E, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {22791, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x6440, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {22792, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x6440, 0x380, 0x1C70, 0xE10E, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x1110, 0x6108, 0x100}},
    {22793, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4444, 0x800, 0xFF0, 0x1010, 0x2820, 0x4440, 0x380, 0xC60, 0x3018, 0xC006}},
    {22794, 2, {0x400, 0x400, 0x400, 0x2FF0, 0x2810, 0x1010, 0x3020, 0x4820, 0x8440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22795, 2, {0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x820, 0x1210, 0x2208, 0x47E4, 0x820, 0x1440, 0x2280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {22796, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x1210, 0x2208, 0x47E4, 0x8822, 0x1440, 0x2280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {22797, 2, {0x1000, 0x1000, 0x1FFC, 0x2000, 0x5FF0, 0x9010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFF0, 0x1820, 0x6440, 0x380, 0x1C70, 0xE00E}},
    {22798, 2, {0x100, 0x1110, 0x1110, 0x2928, 0x4544, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x800, 0xFF0, 0x1420, 0x2240, 0x4180, 0xE60, 0x701C}},
    {22799, 2, {0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE00E}},
    {22800, 2, {0x1000, 0x1FE0, 0x2020, 0x7FF8, 0xA308, 0x2488, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FF0, 0x2820, 0x7C0, 0xF83E}},
    {22801, 2, {0x2108, 0x27C8, 0x2008, 0x6B9A, 0xB02C, 0x2388, 0x2008, 0x5394, 0x4A92, 0x8BA2, 0x400, 0xFE0, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {22802, 2, {0x0, 0xFFFE, 0x1100, 0x13DC, 0x5254, 0x5BD4, 0x525C, 0x53D0, 0x5A52, 0xE3CE, 0x2408, 0x4FE4, 0x9442, 0x380, 0x1C70, 0xE00E}},
    {22803, 2, {0x17C0, 0x621C, 0x47C4, 0x4444, 0x77DC, 0x4444, 0x47C4, 0x745C, 0x47C4, 0x0, 0x2408, 0x4FE4, 0x9442, 0x380, 0x1C70, 0xE00E}},
    {22804, 2, {0x440, 0xFFFE, 0x1100, 0x13DC, 0x5254, 0x5BD4, 0x525C, 0x53D0, 0x5A52, 0xE3CE, 0x2408, 0x4FE4, 0x9442, 0x380, 0x1C70, 0xE00E}},
    {22805, 2, {0x200, 0x200, 0x200, 0x200, 0x7F8, 0x408, 0x808, 0x1010, 0x2410, 0x4220, 0x140, 0x80, 0x100, 0x200, 0xC00, 0x7000}},
    {22806, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x3E40, 0x2260, 0x4250, 0x4248, 0xA444, 0x1444, 0x840, 0x840, 0x1040, 0x2040, 0x4040, 0x8040}},
    {22807, 2, {0x1000, 0x1000, 0x10F8, 0x1E88, 0x1288, 0x2288, 0x2288, 0x5288, 0x8AA8, 0x490, 0x480, 0x882, 0x882, 0x1082, 0x207E, 0x4000}},
    {22808, 2, {0x1000, 0x1000, 0x107C, 0x1E44, 0x1244, 0x2244, 0x2244, 0x5244, 0x8A44, 0x444, 0x444, 0x854, 0x848, 0x1040, 0x2040, 0x4040}},
    {22809, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2FE8, 0x2208, 0x2208, 0x27C8, 0x2448, 0x2848, 0x3488, 0x228A, 0x210A, 0x420A, 0x4406, 0x8802}},
    {22810, 2, {0x200, 0x200, 0x7F0, 0x820, 0x3840, 0x480, 0x340, 0xC80, 0x71F8, 0x208, 0xC10, 0x3220, 0x140, 0x180, 0xE00, 0x7000}},
    {22811, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x1FF0, 0x10, 0x10, 0xFFFE, 0x200, 0x7F0, 0x810, 0x3420, 0x240, 0x180, 0xE00, 0x7000}},
    {22812, 2, {0x200, 0x100, 0xFFFE, 0x880, 0x880, 0x10F8, 0x1108, 0x3148, 0x5228, 0x9510, 0x1110, 0x10A0, 0x1040, 0x10A0, 0x1118, 0x1606}},
    {22813, 2, {0x2020, 0x2020, 0x2120, 0x3D20, 0x25FC, 0x4520, 0x4620, 0xA420, 0x1420, 0x9FC, 0x820, 0x1020, 0x1020, 0x2020, 0x43FE, 0x8000}},
    {22814, 2, {0x1020, 0x1020, 0x3F30, 0x4128, 0xA224, 0x1422, 0x820, 0x3020, 0xC020, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {22815, 2, {0x2020, 0x2020, 0x407C, 0x7E44, 0x82A8, 0x210, 0x7A20, 0x4AD0, 0x4A1E, 0x4A22, 0x7A64, 0x4A94, 0x208, 0x210, 0x1420, 0x8C0}},
    {22816, 2, {0x1020, 0x1020, 0x3E40, 0x427E, 0xA482, 0x1902, 0x107A, 0x284A, 0xCF4A, 0x114A, 0x327A, 0x4A4A, 0x402, 0x802, 0x3014, 0xC008}},
    {22817, 2, {0x1010, 0x1010, 0x3E10, 0x42FE, 0xA410, 0x1810, 0x107C, 0x2800, 0xCF00, 0x117C, 0x3244, 0x4A44, 0x444, 0x844, 0x307C, 0xC044}},
    {22818, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFE, 0x4202, 0x87F4, 0x810, 0x3420, 0x240, 0x180, 0xE00, 0x7000}},
    {22819, 2, {0x100, 0xFFFE, 0x1010, 0x3FF8, 0x0, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFE, 0x4402, 0x8FF4, 0x3420, 0x2C0, 0x700, 0x7800}},
    {22820, 2, {0x200, 0xFF0, 0x3460, 0x380, 0x1D00, 0x7FFE, 0x4002, 0xBFFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x1010, 0x2008}},
    {22821, 2, {0x20, 0x7E20, 0x4A7C, 0x4A44, 0x7EA8, 0x4A10, 0x4A20, 0x7ED0, 0x81E, 0xFF22, 0x1864, 0x2C94, 0x2A08, 0x4A10, 0x8820, 0x8C0}},
    {22822, 2, {0x2220, 0x2220, 0xFF7C, 0x2244, 0x3EA8, 0x2210, 0x3E20, 0x22D0, 0x221E, 0xFF22, 0x4064, 0x5494, 0x6208, 0x4010, 0x7E20, 0xC0}},
    {22823, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {22824, 2, {0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x7F80, 0x80, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22825, 2, {0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {22826, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0xA20, 0x1110, 0x2108, 0xC006}},
    {22827, 2, {0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {22828, 2, {0x100, 0x100, 0x100, 0x1FF0, 0x110, 0x110, 0x110, 0x110, 0x110, 0xFFFE, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22829, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {22830, 2, {0x100, 0x100, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22831, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x100, 0x3FF8, 0x208, 0x208, 0x408, 0x808, 0x1050, 0x6020}},
    {22832, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC446, 0x440, 0x440, 0x440, 0x440, 0x840, 0x840, 0x1040, 0x2040}},
    {22833, 2, {0x100, 0x1100, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x600C}},
    {22834, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {22835, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xCFE6, 0x0, 0x0, 0x0, 0x0, 0x7FFC, 0x0}},
    {22836, 2, {0x80, 0x80, 0x880, 0x480, 0x2480, 0x1080, 0x1080, 0x80, 0xFFFE, 0x100, 0x140, 0x220, 0x410, 0x808, 0x3004, 0xC004}},
    {22837, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x500, 0x200}},
    {22838, 2, {0x820, 0x820, 0x820, 0x820, 0x820, 0x7EFE, 0x820, 0x820, 0x850, 0x1450, 0x1250, 0x1288, 0x2088, 0x2104, 0x4204, 0x8402}},
    {22839, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x108, 0x108, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x284, 0x44C, 0x820, 0x3018, 0xC006}},
    {22840, 2, {0x100, 0x100, 0x7FFC, 0x440, 0x820, 0x17D0, 0x2008, 0xC006, 0x3FF8, 0x400, 0x800, 0xFF0, 0x10, 0x10, 0xA0, 0x40}},
    {22841, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x1110, 0x920, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22842, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC046, 0x40, 0xFFFE, 0x40, 0x1040, 0x840, 0x840, 0x140, 0x80}},
    {22843, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {22844, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC926, 0x920, 0x920, 0x920, 0x920, 0x1120, 0x1120, 0x2120, 0x4020}},
    {22845, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x0, 0x7FFC, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {22846, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x1110, 0x1110, 0x1110, 0x2928, 0x4544, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22847, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xDFF6, 0x1110, 0x1110, 0x1110, 0x1FF0, 0x1004, 0x1004, 0xFFC, 0x0}},
    {22848, 2, {0x0, 0x7FFC, 0x100, 0x300, 0x560, 0x1918, 0x6104, 0x0, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22849, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xDFF6, 0x1000, 0x1220, 0x1140, 0x1080, 0x1140, 0x1220, 0x1000, 0x1FF8, 0x0}},
    {22850, 2, {0x800, 0x800, 0xFE0, 0x1020, 0x2040, 0x5FF0, 0x9110, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22851, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC0F6, 0x3F00, 0x2100, 0x2100, 0x3FFC, 0x2080, 0x2040, 0x2A24, 0x3114, 0x210C}},
    {22852, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x920, 0x3118, 0xDFF6, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF4, 0x104, 0x104, 0xFC}},
    {22853, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC606, 0x38F8, 0x2088, 0x2288, 0x2288, 0x2E88, 0x32A8, 0x2490, 0x480, 0x880}},
    {22854, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xDFF6, 0x1000, 0x1000, 0x1FE0, 0x1020, 0x1020, 0x1FE0, 0x1000, 0x1000, 0x1FF8}},
    {22855, 2, {0x100, 0x100, 0x3FF8, 0x280, 0x440, 0x820, 0xFFFE, 0x10, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x10, 0x50, 0x20}},
    {22856, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xCFE6, 0x0, 0x0, 0x7FFC, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {22857, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x440, 0x920, 0x3118, 0xCFE6, 0x100, 0x100, 0x3FF8, 0x100, 0x100}},
    {22858, 2, {0x80, 0xFF8, 0x4080, 0x5FFC, 0x4080, 0x4FF8, 0x4080, 0x5FFC, 0x4000, 0x7F00, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22859, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {22860, 2, {0x100, 0x100, 0x1FC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22861, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x820, 0x1010, 0x2448, 0xC446, 0x440, 0x840, 0x840, 0x1040, 0x2040}},
    {22862, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x920, 0x3118, 0xCFE6, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {22863, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x440, 0x820, 0x37D8, 0xC106, 0x1FF0, 0x100, 0x2C0, 0xC30, 0x3008}},
    {22864, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2288, 0x2448, 0x2828, 0x2108, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {22865, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x824, 0x7F44, 0x854, 0x988, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22866, 2, {0x200, 0x200, 0x7FFC, 0x440, 0xA20, 0x3118, 0xDFF6, 0x400, 0x840, 0x1F80, 0x110, 0x620, 0x3840, 0x1A0, 0xE10, 0x7008}},
    {22867, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3218, 0xC7E6, 0x1840, 0x580, 0x680, 0x39F8, 0x610, 0x1920, 0xC0, 0x300, 0x3C00}},
    {22868, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x920, 0x3118, 0xCFE6, 0x100, 0x920, 0x820, 0x7FFC, 0x820, 0x820, 0x1020, 0x1020, 0x2020}},
    {22869, 2, {0x200, 0x100, 0x7FFC, 0x440, 0x2450, 0x4448, 0x8944, 0x1080, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22870, 2, {0x840, 0x4840, 0x28FC, 0x904, 0x1A88, 0x2850, 0x4820, 0x8840, 0x980, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22871, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0xFF0, 0x3418, 0xC7E6, 0x400, 0x7E0, 0x400, 0xFFFE, 0x400, 0x810, 0x1FF8, 0x8}},
    {22872, 2, {0x2420, 0x2420, 0x2420, 0x3DFC, 0x420, 0xFC20, 0x2420, 0x24F8, 0x4500, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22873, 2, {0x400, 0x820, 0x1FF0, 0x210, 0xFFFE, 0x820, 0x3118, 0xDFF6, 0x1110, 0x1FF0, 0x1110, 0x1FF4, 0x104, 0x104, 0xFC, 0x0}},
    {22874, 2, {0xFC, 0x7F00, 0x2208, 0x1510, 0x840, 0x1F80, 0x320, 0xC10, 0x3FF8, 0x108, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22875, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC006, 0x3CF8, 0x2488, 0x3CF8, 0x2488, 0x24F8, 0x3C88, 0x2488, 0x128, 0x210}},
    {22876, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x540, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22877, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3FF8, 0xD116, 0x17D0, 0x1110, 0x1FF0, 0x1010, 0x17D0, 0x2450, 0x27D0, 0x4010, 0x8030}},
    {22878, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC906, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {22879, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3E78, 0xD24E, 0x1248, 0x1E78, 0x1248, 0x1248, 0x1E78, 0x1248, 0x2288, 0x2AA8, 0x4510}},
    {22880, 2, {0x820, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22881, 2, {0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0xFFFE, 0x820, 0x3458, 0xC446, 0x840, 0x1040}},
    {22882, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3218, 0xDFD6, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x810}},
    {22883, 2, {0x0, 0x3FF8, 0x200, 0x7FFC, 0x440, 0x1830, 0xE00E, 0x3CF8, 0x2488, 0x24F8, 0x3C88, 0x24F8, 0x2488, 0x3C88, 0x128, 0x210}},
    {22884, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3FF8, 0xC106, 0x200, 0x3FF8, 0x2448, 0x27C8, 0x2448, 0x27C8, 0x2448, 0x3FF8, 0x2008}},
    {22885, 2, {0x200, 0x400, 0x3FF8, 0x2008, 0x2928, 0x2548, 0x2FE8, 0x2548, 0x2928, 0x2008, 0x100, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006}},
    {22886, 2, {0x40, 0x7EFC, 0x1544, 0x838, 0xFE44, 0x1AA2, 0x2CFC, 0x4844, 0xA894, 0x1108, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22887, 2, {0x200, 0x3FF8, 0x2008, 0x20C8, 0x2708, 0x2928, 0x2548, 0x2FE8, 0x2548, 0x2928, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22888, 2, {0x83E, 0x4BC4, 0x2A48, 0x2920, 0x808, 0x1BFE, 0x2908, 0xC888, 0x828, 0x910, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22889, 2, {0x200, 0x7FFC, 0x420, 0x1818, 0xE006, 0x3FF8, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2F78, 0x2948, 0x2F78, 0x2000, 0x3FFC}},
    {22890, 2, {0x200, 0x7FFC, 0x420, 0x1898, 0xE846, 0x1FF8, 0x3080, 0x5FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1020, 0xFFFE, 0x820, 0x460}},
    {22891, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC146, 0x2548, 0x1778, 0x8408, 0x47F8, 0x1408, 0x2778, 0xE548, 0x2548, 0x2808}},
    {22892, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8518, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22893, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x7D7C, 0x1110, 0x7D7C, 0x4544, 0x7D7C, 0x4544, 0x7D7C, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {22894, 2, {0x200, 0x7FFC, 0x420, 0x1818, 0xE906, 0x1FF8, 0x3100, 0x5FF8, 0x9100, 0x1FFC, 0x1000, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {22895, 2, {0x200, 0x7FFC, 0x420, 0x1918, 0xE9E6, 0x900, 0x7FFC, 0x28, 0x3FFE, 0x2020, 0x2FA4, 0x2228, 0x2A90, 0x532A, 0x4646, 0x9882}},
    {22896, 2, {0x0, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0xFEFE, 0xAAAA, 0xAAAA, 0xFEFE, 0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22897, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {22898, 2, {0x1000, 0x10EE, 0xFEAA, 0x28AA, 0x44EE, 0x9000, 0x7AFE, 0x1492, 0xFEFE, 0x1092, 0x7CFE, 0xC410, 0x7DFE, 0x4410, 0x7C10, 0x4410}},
    {22899, 2, {0x200, 0x200, 0x200, 0x200, 0x400, 0xFFFE, 0x420, 0x820, 0x820, 0x1040, 0x1840, 0x680, 0x180, 0x660, 0x1810, 0x6008}},
    {22900, 2, {0x1000, 0x1000, 0x11FC, 0x1084, 0xFC84, 0x2484, 0x2488, 0x2488, 0x2450, 0x4850, 0x2820, 0x1020, 0x2850, 0x4488, 0x8104, 0x202}},
    {22901, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x2420, 0x2420, 0x2420, 0x2420, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {22902, 2, {0x1000, 0x1000, 0x13F8, 0x1088, 0xFC90, 0x2490, 0x24A0, 0x24BC, 0x2484, 0x4884, 0x2904, 0x1104, 0x2904, 0x4604, 0x8228, 0x410}},
    {22903, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x2420, 0x2420, 0x27FE, 0x2420, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x8420, 0x20}},
    {22904, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFE20, 0x2220, 0x2220, 0x23FE, 0x4220, 0x2420, 0x1420, 0x820, 0x1420, 0x2220, 0x4220, 0x8020}},
    {22905, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0xFD2C, 0x2534, 0x2564, 0x27A4, 0x2524, 0x4934, 0x2928, 0x1122, 0x2922, 0x4502, 0x84FE, 0x0}},
    {22906, 2, {0x1080, 0x1080, 0x1080, 0x10F8, 0xFD08, 0x2508, 0x2510, 0x2610, 0x2420, 0x4820, 0x2850, 0x1050, 0x2888, 0x4508, 0x8204, 0x402}},
    {22907, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFEFE, 0x2244, 0x2244, 0x2244, 0x4244, 0x2484, 0x1448, 0x828, 0x1410, 0x2228, 0x4244, 0x8082}},
    {22908, 2, {0x1010, 0x1038, 0x11E0, 0x1020, 0xFE20, 0x2220, 0x2220, 0x223E, 0x43E0, 0x2420, 0x1420, 0x822, 0x1422, 0x2222, 0x421E, 0x8000}},
    {22909, 2, {0x1000, 0x10FC, 0x1004, 0x1008, 0xFC10, 0x2420, 0x2420, 0x25FE, 0x2420, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {22910, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFD24, 0x2524, 0x2524, 0x2524, 0x2524, 0x4924, 0x2924, 0x1124, 0x2924, 0x45FC, 0x8004, 0x0}},
    {22911, 2, {0x2040, 0x2040, 0x2040, 0x23FE, 0xFC88, 0x2488, 0x2508, 0x2490, 0x6450, 0x2420, 0x3450, 0x248A, 0x2502, 0x4402, 0x43FE, 0x8000}},
    {22912, 2, {0x40, 0x40, 0xFC40, 0x440, 0x7FE, 0x488, 0x488, 0x7C88, 0x4088, 0x4108, 0x4090, 0x4450, 0x5820, 0x6050, 0x4088, 0x304}},
    {22913, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0xFD04, 0x2604, 0x2404, 0x2504, 0x2484, 0x4844, 0x2844, 0x1004, 0x2804, 0x4404, 0x8428, 0x10}},
    {22914, 2, {0x1000, 0x1000, 0x1000, 0x107C, 0xFE44, 0x2244, 0x2244, 0x2244, 0x4244, 0x2444, 0x1444, 0x844, 0x1444, 0x227C, 0x4244, 0x8000}},
    {22915, 2, {0x1000, 0x1000, 0x11F8, 0x1008, 0xFC08, 0x2408, 0x2408, 0x25F8, 0x2508, 0x4900, 0x2900, 0x1100, 0x2902, 0x4502, 0x84FE, 0x0}},
    {22916, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1FF8, 0x200, 0x200, 0xFFFE, 0x420, 0x820, 0x1E40, 0x180, 0x660, 0x1818, 0x6004}},
    {22917, 2, {0x1000, 0x1000, 0x11FC, 0x1020, 0xFC20, 0x2420, 0x2420, 0x2420, 0x2420, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {22918, 2, {0x840, 0x840, 0x840, 0x4840, 0x2BFE, 0x2888, 0x888, 0x888, 0x1888, 0x2908, 0xC890, 0x850, 0x820, 0x850, 0x888, 0xB04}},
    {22919, 2, {0x1000, 0x1000, 0x11FC, 0x1004, 0xFC04, 0x2404, 0x2404, 0x25FC, 0x2404, 0x4804, 0x2804, 0x1004, 0x2804, 0x45FC, 0x8004, 0x0}},
    {22920, 2, {0x1000, 0x13F0, 0x1010, 0x1010, 0xFD10, 0x2510, 0x2510, 0x25FC, 0x2404, 0x4804, 0x2804, 0x13F4, 0x2804, 0x4404, 0x8428, 0x10}},
    {22921, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0xFD22, 0x2624, 0x2420, 0x2420, 0x2450, 0x4850, 0x2850, 0x1050, 0x2890, 0x4492, 0x8512, 0x20E}},
    {22922, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x2420, 0x2420, 0x27FE, 0x2420, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x85FC, 0x0}},
    {22923, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x2420, 0x2420, 0x2420, 0x27FE, 0x4820, 0x2850, 0x1050, 0x2888, 0x4488, 0x8504, 0x202}},
    {22924, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFC, 0x2488, 0x2488, 0x2488, 0x27FE, 0x4888, 0x2888, 0x1088, 0x2888, 0x4508, 0x8508, 0x208}},
    {22925, 2, {0x1000, 0x13FE, 0x1088, 0x1088, 0xFC88, 0x2488, 0x2488, 0x27FE, 0x2488, 0x4888, 0x2888, 0x1088, 0x2888, 0x4488, 0x8108, 0x208}},
    {22926, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0xFC88, 0x2504, 0x2602, 0x2488, 0x2488, 0x4888, 0x2888, 0x1088, 0x2888, 0x4508, 0x8108, 0x208}},
    {22927, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC88, 0x2488, 0x2488, 0x2488, 0x2488, 0x4850, 0x2850, 0x1020, 0x2850, 0x4488, 0x8104, 0x602}},
    {22928, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0xFC88, 0x2524, 0x2524, 0x2622, 0x2420, 0x4840, 0x2848, 0x1044, 0x2884, 0x45FE, 0x8482, 0x0}},
    {22929, 2, {0x1000, 0x1000, 0x11FC, 0x1124, 0xFD24, 0x2524, 0x2524, 0x2524, 0x25FC, 0x4900, 0x2900, 0x1100, 0x2902, 0x4502, 0x84FE, 0x0}},
    {22930, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFD02, 0x2502, 0x2502, 0x25FE, 0x2502, 0x4900, 0x2900, 0x1100, 0x2A00, 0x4600, 0x8400, 0x800}},
    {22931, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x2420, 0x2420, 0x25FC, 0x2484, 0x4888, 0x2848, 0x1050, 0x2820, 0x4450, 0x8188, 0x606}},
    {22932, 2, {0x1080, 0x1040, 0x1040, 0x13FC, 0xFC00, 0x2400, 0x25F0, 0x2510, 0x2510, 0x4910, 0x2910, 0x1112, 0x2912, 0x4612, 0x820E, 0x400}},
    {22933, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x2524, 0x2524, 0x2524, 0x2524, 0x49FC, 0x2924, 0x1020, 0x2820, 0x4420, 0x8420, 0x20}},
    {22934, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x2420, 0x25FE, 0x2420, 0x2420, 0x4850, 0x2850, 0x1050, 0x2888, 0x4488, 0x8104, 0x202}},
    {22935, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0xFC88, 0x2544, 0x2622, 0x2420, 0x2400, 0x49FC, 0x2804, 0x1008, 0x2808, 0x4410, 0x8410, 0x20}},
    {22936, 2, {0x1000, 0x1000, 0x10FC, 0x1000, 0xFC00, 0x2400, 0x25FE, 0x2420, 0x2420, 0x4840, 0x2840, 0x1088, 0x2884, 0x45FE, 0x8482, 0x0}},
    {22937, 2, {0x1020, 0x1020, 0x1020, 0x10A8, 0xFCA4, 0x24A2, 0x2522, 0x2520, 0x2624, 0x4824, 0x2828, 0x1008, 0x2810, 0x4420, 0x80C0, 0x300}},
    {22938, 2, {0x1000, 0x1000, 0x11FE, 0x1010, 0xFC10, 0x2420, 0x2420, 0x2468, 0x24A4, 0x4922, 0x2A22, 0x1020, 0x2820, 0x4420, 0x8420, 0x20}},
    {22939, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22940, 2, {0x1040, 0x1040, 0x1040, 0x13F8, 0xFC48, 0x2448, 0x2448, 0x2448, 0x27FE, 0x4840, 0x28A0, 0x10A0, 0x2910, 0x4510, 0x8208, 0x406}},
    {22941, 2, {0x420, 0x2420, 0x2420, 0x2420, 0x25FE, 0x3C44, 0x444, 0x444, 0xFC44, 0x2484, 0x2448, 0x2428, 0x2410, 0x4428, 0x4444, 0x8582}},
    {22942, 2, {0x1000, 0x11F8, 0x1048, 0x1048, 0xFC48, 0x2448, 0x2448, 0x25F8, 0x2488, 0x4888, 0x2888, 0x1088, 0x2888, 0x4488, 0x83FE, 0x0}},
    {22943, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22944, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x2522, 0x2522, 0x2522, 0x2552, 0x494A, 0x298A, 0x1102, 0x2902, 0x4502, 0x850A, 0x104}},
    {22945, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x2500, 0x25FE, 0x2510, 0x2510, 0x4910, 0x2910, 0x1110, 0x2910, 0x4610, 0x8210, 0x410}},
    {22946, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0xFD08, 0x2504, 0x2604, 0x25FA, 0x2488, 0x4888, 0x2888, 0x1088, 0x2908, 0x4508, 0x8228, 0x410}},
    {22947, 2, {0x1010, 0x1110, 0x1110, 0x1112, 0xFD12, 0x2514, 0x25D8, 0x2510, 0x2510, 0x4910, 0x2910, 0x1112, 0x2952, 0x4592, 0x850E, 0x0}},
    {22948, 2, {0x1000, 0x11F8, 0x1008, 0x1050, 0xFC20, 0x2410, 0x27FE, 0x2422, 0x2424, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {22949, 2, {0xFC, 0x7F00, 0x208, 0x2108, 0x1110, 0x1020, 0x200, 0x200, 0xFFFE, 0x420, 0x820, 0x1E40, 0x180, 0x660, 0x1818, 0x6004}},
    {22950, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x2420, 0x2420, 0x25FC, 0x2420, 0x4820, 0x2820, 0x13FE, 0x2820, 0x4420, 0x8420, 0x20}},
    {22951, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x27FE, 0x2490, 0x2490, 0x2490, 0x4890, 0x2890, 0x1092, 0x2912, 0x4512, 0x820E, 0x400}},
    {22952, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFC40, 0x2440, 0x2440, 0x247C, 0x2444, 0x4844, 0x2844, 0x1044, 0x2884, 0x4484, 0x8128, 0x210}},
    {22953, 2, {0x2000, 0x23FC, 0x2040, 0x2040, 0xF840, 0x4840, 0x4FFE, 0x4840, 0x88A0, 0x48A0, 0x30A0, 0x1120, 0x2922, 0x4A22, 0x841E, 0x800}},
    {22954, 2, {0x1000, 0x11FE, 0x1100, 0x1104, 0xFD44, 0x2528, 0x2528, 0x2510, 0x2510, 0x4928, 0x2928, 0x1144, 0x2984, 0x4500, 0x85FE, 0x0}},
    {22955, 2, {0x1020, 0x1120, 0x10A0, 0x10A0, 0xFC20, 0x27FE, 0x2422, 0x2422, 0x2442, 0x4852, 0x284A, 0x108A, 0x2882, 0x4502, 0x8214, 0x408}},
    {22956, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x2440, 0x2440, 0x24FC, 0x2584, 0x4A84, 0x2C84, 0x1084, 0x2884, 0x4484, 0x84FC, 0x84}},
    {22957, 2, {0x2050, 0x2048, 0x2048, 0x2040, 0xFBFE, 0x4880, 0x4880, 0x48FC, 0x8944, 0x4944, 0x3128, 0x1128, 0x2A10, 0x4A28, 0x8444, 0x182}},
    {22958, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x25FC, 0x2500, 0x2540, 0x2544, 0x4948, 0x2970, 0x1140, 0x2942, 0x4642, 0x823E, 0x400}},
    {22959, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x2524, 0x2524, 0x2524, 0x2524, 0x49FC, 0x2924, 0x1124, 0x2924, 0x4524, 0x85FC, 0x104}},
    {22960, 2, {0x1080, 0x1080, 0x1080, 0x10FE, 0xFD40, 0x2540, 0x2640, 0x247C, 0x2440, 0x4840, 0x2840, 0x107C, 0x2840, 0x4440, 0x8440, 0x40}},
    {22961, 2, {0x1000, 0x13FC, 0x1084, 0x1084, 0xFC84, 0x2504, 0x2514, 0x2608, 0x2400, 0x49FC, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {22962, 2, {0x1000, 0x1000, 0x10FC, 0x1084, 0xFC84, 0x2484, 0x24FC, 0x2484, 0x2484, 0x4884, 0x28FC, 0x1084, 0x2800, 0x4400, 0x85FE, 0x0}},
    {22963, 2, {0x1080, 0x1080, 0x1080, 0x10FE, 0xFD02, 0x2504, 0x2620, 0x2420, 0x24A8, 0x48A4, 0x2924, 0x1122, 0x2A22, 0x4420, 0x80A0, 0x40}},
    {22964, 2, {0x1000, 0x10F8, 0x3E88, 0x4288, 0xA4A8, 0x1894, 0x1084, 0x207C, 0xC200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22965, 2, {0x1040, 0x1020, 0x1000, 0x13FE, 0xFC20, 0x2420, 0x2420, 0x2420, 0x25FC, 0x4820, 0x2820, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {22966, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFC20, 0x2420, 0x2444, 0x2484, 0x25F8, 0x4810, 0x2820, 0x1044, 0x2882, 0x45FE, 0x8482, 0x0}},
    {22967, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0xFDFC, 0x2520, 0x2620, 0x2420, 0x25FE, 0x4820, 0x2850, 0x1050, 0x2888, 0x4488, 0x8504, 0x202}},
    {22968, 2, {0x1000, 0x1000, 0x13FE, 0x1008, 0xFC08, 0x25E8, 0x2528, 0x2528, 0x2528, 0x4928, 0x29E8, 0x1128, 0x2808, 0x4408, 0x8428, 0x10}},
    {22969, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x2420, 0x2420, 0x27FE, 0x2470, 0x48A8, 0x28A8, 0x1124, 0x2924, 0x4622, 0x8420, 0x20}},
    {22970, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x2420, 0x2420, 0x25FC, 0x2470, 0x48A8, 0x28A8, 0x1124, 0x2924, 0x4622, 0x8020, 0x20}},
    {22971, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22972, 2, {0x1040, 0x1020, 0x1028, 0x1008, 0xFC08, 0x2448, 0x2454, 0x2552, 0x2552, 0x4960, 0x2A60, 0x1044, 0x28C4, 0x4544, 0x823C, 0x0}},
    {22973, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x2524, 0x2524, 0x25FC, 0x2524, 0x4924, 0x2924, 0x11FC, 0x2924, 0x4420, 0x8020, 0x20}},
    {22974, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x200, 0x200, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0xC70, 0x7008}},
    {22975, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x1244, 0x2244, 0x4A7C, 0x8444, 0x200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22976, 2, {0x101C, 0x11E0, 0x1020, 0x1020, 0xFC20, 0x27FE, 0x2420, 0x2470, 0x2470, 0x48A8, 0x28A8, 0x1124, 0x2A22, 0x4420, 0x8020, 0x20}},
    {22977, 2, {0x1080, 0x1080, 0x1100, 0x11FC, 0xFE04, 0x2404, 0x25E4, 0x2524, 0x2524, 0x4924, 0x29E4, 0x1124, 0x2804, 0x4404, 0x8428, 0x10}},
    {22978, 2, {0x1004, 0x101E, 0x11E0, 0x1000, 0xFC20, 0x2410, 0x25FE, 0x2404, 0x2408, 0x4810, 0x2820, 0x1040, 0x2880, 0x4540, 0x823E, 0x0}},
    {22979, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x2420, 0x2520, 0x2520, 0x253C, 0x4920, 0x2920, 0x1120, 0x2920, 0x4520, 0x87FE, 0x0}},
    {22980, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x25FC, 0x2520, 0x2520, 0x25FE, 0x4920, 0x2920, 0x1110, 0x2912, 0x454A, 0x8586, 0x102}},
    {22981, 2, {0x1020, 0x1020, 0x1124, 0x10A4, 0xFCA8, 0x2420, 0x25FC, 0x2420, 0x2420, 0x4820, 0x2BFE, 0x1020, 0x2820, 0x4420, 0x8420, 0x20}},
    {22982, 2, {0x1000, 0x10FC, 0x1084, 0x10A4, 0xFC94, 0x2484, 0x2484, 0x27FE, 0x2504, 0x4944, 0x2924, 0x1104, 0x29FE, 0x4404, 0x8428, 0x10}},
    {22983, 2, {0x810, 0x810, 0x17FE, 0x3010, 0x5210, 0x9110, 0x1110, 0x1050, 0x1220, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22984, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0xFC88, 0x2524, 0x2612, 0x2410, 0x25FC, 0x4804, 0x2808, 0x1088, 0x2850, 0x4420, 0x8410, 0x10}},
    {22985, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x2420, 0x25FC, 0x2524, 0x2524, 0x4924, 0x2924, 0x1134, 0x2928, 0x4420, 0x8420, 0x20}},
    {22986, 2, {0x1020, 0x1020, 0x102E, 0x11F0, 0xFD20, 0x2520, 0x2520, 0x25FE, 0x2422, 0x4862, 0x28A2, 0x112A, 0x2E24, 0x4420, 0x8420, 0x20}},
    {22987, 2, {0x1020, 0x1020, 0x1020, 0x1040, 0xFC48, 0x2484, 0x25FE, 0x2482, 0x2400, 0x48FC, 0x2884, 0x1084, 0x2884, 0x4484, 0x80FC, 0x84}},
    {22988, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x2524, 0x25FC, 0x2524, 0x2524, 0x4924, 0x2BFE, 0x1104, 0x2904, 0x4504, 0x8514, 0x108}},
    {22989, 2, {0x1000, 0x11FC, 0x1154, 0x1154, 0xFD54, 0x2554, 0x2554, 0x27FE, 0x2554, 0x4954, 0x2954, 0x1154, 0x2954, 0x4544, 0x8514, 0x108}},
    {22990, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x2524, 0x2524, 0x2524, 0x2524, 0x4BFE, 0x2820, 0x1050, 0x2850, 0x4488, 0x8104, 0x202}},
    {22991, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFE, 0x2488, 0x2488, 0x2488, 0x2488, 0x48F8, 0x2888, 0x1088, 0x2888, 0x4488, 0x84F8, 0x88}},
    {22992, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x25F8, 0x2508, 0x2508, 0x2508, 0x49F8, 0x2908, 0x1108, 0x2908, 0x4508, 0x87FE, 0x0}},
    {22993, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x2420, 0x2420, 0x2420, 0x25FC, 0x4904, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {22994, 2, {0x1008, 0x1048, 0x1128, 0x1128, 0xFD28, 0x2508, 0x2508, 0x2508, 0x2508, 0x4908, 0x2948, 0x1194, 0x2914, 0x4422, 0x8442, 0x82}},
    {22995, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0xFDFC, 0x2520, 0x2620, 0x2420, 0x2420, 0x49FC, 0x2820, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {22996, 2, {0x70, 0x1F80, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC206, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22997, 2, {0x880, 0x888, 0x2890, 0x2EA0, 0x28C0, 0x2884, 0x2E84, 0xF07C, 0x200, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {22998, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0xFD00, 0x25FC, 0x2504, 0x2504, 0x2504, 0x4904, 0x29FC, 0x1100, 0x2900, 0x4500, 0x85FE, 0x0}},
    {22999, 2, {0x1000, 0x11DC, 0x1154, 0x1154, 0xFD54, 0x2554, 0x2554, 0x27FE, 0x2554, 0x4954, 0x2954, 0x1154, 0x2954, 0x46D4, 0x8224, 0x44C}},
    {23000, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFDFE, 0x2488, 0x2488, 0x2488, 0x2488, 0x4BFE, 0x2888, 0x1088, 0x2908, 0x4508, 0x8208, 0x408}},
    {23001, 2, {0x1080, 0x1088, 0x109C, 0x1170, 0xFD10, 0x2710, 0x2510, 0x2510, 0x25FE, 0x4910, 0x2910, 0x1110, 0x2910, 0x4510, 0x857C, 0x100}},
    {23002, 2, {0x2090, 0x2090, 0x2090, 0x2292, 0xF994, 0x4898, 0x4890, 0x4998, 0x8A94, 0x4C92, 0x3090, 0x1090, 0x2912, 0x4912, 0x8212, 0x40E}},
    {23003, 2, {0x1000, 0x11FE, 0x1102, 0x1102, 0xFD7A, 0x2502, 0x2502, 0x257A, 0x254A, 0x494A, 0x294A, 0x117A, 0x2902, 0x4502, 0x850A, 0x104}},
    {23004, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x200, 0x400, 0xFFFE, 0x820, 0x1020, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23005, 2, {0x2020, 0x2120, 0x2120, 0x21FC, 0xF920, 0x4A20, 0x4820, 0x4BFE, 0x4870, 0x90A8, 0x50A8, 0x2124, 0x3124, 0x4A22, 0x4820, 0x8020}},
    {23006, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x2420, 0x25FC, 0x2400, 0x2400, 0x49FC, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23007, 2, {0x1020, 0x1010, 0x1010, 0x13FE, 0xFC20, 0x2442, 0x2484, 0x25F8, 0x2410, 0x4822, 0x28C4, 0x1308, 0x2810, 0x4428, 0x80C4, 0x302}},
    {23008, 2, {0x1020, 0x1020, 0x1040, 0x11FE, 0xFD02, 0x2502, 0x257A, 0x254A, 0x254A, 0x494A, 0x294A, 0x117A, 0x2902, 0x4502, 0x850A, 0x104}},
    {23009, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x27FE, 0x2420, 0x2420, 0x2420, 0x49FC, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23010, 2, {0x1020, 0x1020, 0x1048, 0x1084, 0xFDFE, 0x2402, 0x2400, 0x24FC, 0x2484, 0x48FC, 0x2884, 0x10FC, 0x2884, 0x4484, 0x8494, 0x88}},
    {23011, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFC00, 0x2488, 0x2504, 0x2602, 0x2488, 0x4888, 0x2850, 0x1050, 0x2820, 0x4450, 0x8088, 0x306}},
    {23012, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x25FE, 0x2500, 0x2500, 0x257C, 0x4944, 0x2944, 0x1144, 0x2944, 0x467C, 0x8244, 0x400}},
    {23013, 2, {0x1040, 0x1040, 0x1044, 0x11F4, 0xFC48, 0x2450, 0x27FE, 0x2440, 0x2480, 0x4984, 0x2A98, 0x10E0, 0x2882, 0x4482, 0x847E, 0x0}},
    {23014, 2, {0x400, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008, 0x10, 0x1010, 0xFEFE, 0x2224, 0x6464, 0x1818, 0x2434, 0xC2C2}},
    {23015, 2, {0x1000, 0x1000, 0xFE7C, 0x2210, 0x6410, 0x1810, 0x2410, 0xC210, 0x10FE, 0x1010, 0xFE10, 0x2210, 0x6410, 0x1810, 0x2410, 0xC210}},
    {23016, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0xFBF8, 0x4848, 0x4BF8, 0x4A40, 0x8BFC, 0x4844, 0x3054, 0x10A8, 0x28A0, 0x4910, 0x8208, 0x406}},
    {23017, 2, {0x1080, 0x1080, 0x10FE, 0x1110, 0xFE10, 0x2410, 0x24FC, 0x2490, 0x2490, 0x4890, 0x2BFE, 0x1010, 0x2810, 0x4410, 0x8410, 0x10}},
    {23018, 2, {0x1000, 0x11FE, 0x1020, 0x1040, 0xFC88, 0x2504, 0x25FE, 0x2422, 0x2420, 0x4820, 0x29FE, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {23019, 2, {0x1000, 0x11FE, 0x1120, 0x1120, 0xFD20, 0x25FC, 0x2504, 0x2504, 0x2504, 0x4904, 0x29FC, 0x1120, 0x2920, 0x4520, 0x85FE, 0x0}},
    {23020, 2, {0x1000, 0x11FE, 0x1110, 0x1110, 0xFD10, 0x257C, 0x2544, 0x2544, 0x2544, 0x4944, 0x297C, 0x1110, 0x2910, 0x4510, 0x85FE, 0x0}},
    {23021, 2, {0x200, 0x1FF0, 0x410, 0x850, 0x1020, 0x7EFC, 0x1224, 0x2A54, 0x4488, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23022, 2, {0x1000, 0x13FE, 0x1000, 0x1000, 0xFDFC, 0x2504, 0x2504, 0x25FC, 0x2504, 0x4904, 0x29FC, 0x1104, 0x2800, 0x4400, 0x83FE, 0x0}},
    {23023, 2, {0x1040, 0x1044, 0x1244, 0x1148, 0xFD50, 0x2440, 0x27FE, 0x2490, 0x2490, 0x4890, 0x2890, 0x1092, 0x2912, 0x4512, 0x820E, 0x400}},
    {23024, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0xFE04, 0x25E4, 0x2524, 0x2524, 0x25E4, 0x4924, 0x2924, 0x11E4, 0x2904, 0x4404, 0x8428, 0x10}},
    {23025, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x2504, 0x26FA, 0x2400, 0x27FE, 0x4840, 0x2880, 0x11FC, 0x2804, 0x4404, 0x8428, 0x10}},
    {23026, 2, {0x2040, 0x2020, 0x2020, 0x23FE, 0xFA02, 0x4C44, 0x4840, 0x4BFE, 0x4888, 0x9088, 0x5108, 0x20D0, 0x3020, 0x4850, 0x4888, 0x8304}},
    {23027, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0xFE88, 0x2450, 0x2420, 0x2440, 0x2480, 0x49FC, 0x2A84, 0x1084, 0x2884, 0x4484, 0x84FC, 0x84}},
    {23028, 2, {0x4, 0x7FA4, 0x1024, 0x3F24, 0x5124, 0xA04, 0xC04, 0x3014, 0xC208, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23029, 2, {0x2000, 0x27FC, 0x2404, 0x2404, 0xF5F4, 0x5444, 0x5444, 0x55F4, 0x5554, 0x9554, 0x5554, 0x2554, 0x2574, 0x5446, 0x5446, 0x8842}},
    {23030, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x2602, 0x25FC, 0x2400, 0x2400, 0x49FC, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23031, 2, {0x1020, 0x1020, 0x11FE, 0x1040, 0xFC40, 0x24FC, 0x2484, 0x2584, 0x26FC, 0x4884, 0x2884, 0x10FC, 0x2884, 0x4484, 0x8494, 0x88}},
    {23032, 2, {0x1000, 0x11DC, 0x1088, 0x1088, 0xFC88, 0x2488, 0x2488, 0x27DE, 0x2488, 0x4888, 0x2888, 0x1088, 0x2888, 0x4508, 0x8108, 0x208}},
    {23033, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFD02, 0x2604, 0x2418, 0x25E0, 0x2440, 0x4840, 0x287E, 0x13C0, 0x2840, 0x4444, 0x8444, 0x3C}},
    {23034, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x2620, 0x2420, 0x27FE, 0x2490, 0x4890, 0x2890, 0x1090, 0x2912, 0x4512, 0x8212, 0x40E}},
    {23035, 2, {0x1000, 0x11FC, 0x1104, 0x1124, 0xFD24, 0x2524, 0x25FC, 0x2524, 0x2524, 0x4954, 0x294C, 0x118C, 0x2904, 0x4504, 0x81FC, 0x104}},
    {23036, 2, {0x1020, 0x1020, 0x107C, 0x1084, 0xFD48, 0x2430, 0x2420, 0x2448, 0x2590, 0x483E, 0x2842, 0x11A4, 0x2818, 0x4410, 0x8460, 0x180}},
    {23037, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0xFE08, 0x25FE, 0x2500, 0x257C, 0x2544, 0x4944, 0x2954, 0x1148, 0x2942, 0x4542, 0x853E, 0x200}},
    {23038, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x2602, 0x25FC, 0x2420, 0x2420, 0x4820, 0x29FC, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {23039, 2, {0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23040, 2, {0x2028, 0x2024, 0x2024, 0x2020, 0xFBFE, 0x4820, 0x4924, 0x4924, 0x8924, 0x4BA8, 0x3128, 0x1110, 0x2A12, 0x4A2A, 0x8446, 0x82}},
    {23041, 2, {0x28, 0x24, 0x20, 0x3FFE, 0x2020, 0x2020, 0x3FE4, 0x2224, 0x2224, 0x3FA8, 0x24A8, 0x2890, 0x2512, 0x422A, 0x4546, 0x9882}},
    {23042, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFE, 0x2488, 0x2488, 0x2488, 0x2488, 0x4BFE, 0x2800, 0x1088, 0x2884, 0x4504, 0x8202, 0x402}},
    {23043, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x2420, 0x27FE, 0x2400, 0x2420, 0x4820, 0x29FC, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {23044, 2, {0x100, 0x1110, 0x920, 0xFFFE, 0x540, 0x920, 0x3118, 0xC006, 0x200, 0xFFFE, 0x420, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23045, 2, {0x2000, 0x23FC, 0x2090, 0x2090, 0xF890, 0x4892, 0x4C92, 0x4A94, 0x8A94, 0x4A98, 0x3090, 0x1090, 0x2890, 0x4890, 0x87FE, 0x0}},
    {23046, 2, {0x2080, 0x2080, 0x20BC, 0x23C0, 0xF850, 0x4824, 0x48D4, 0x4B0C, 0x8800, 0x4BFE, 0x3090, 0x1090, 0x2912, 0x4912, 0x820E, 0x400}},
    {23047, 2, {0x1008, 0x101C, 0x10E0, 0x1020, 0xFC20, 0x25FE, 0x2450, 0x2488, 0x2504, 0x4A8A, 0x2888, 0x1088, 0x2888, 0x4488, 0x8488, 0x108}},
    {23048, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x1448, 0x2444, 0x4204, 0x200, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0xC70, 0x7008}},
    {23049, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFDFC, 0x2524, 0x25FC, 0x2400, 0x27FE, 0x4880, 0x2900, 0x11FC, 0x2804, 0x4404, 0x8428, 0x10}},
    {23050, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x2508, 0x2508, 0x25F8, 0x2508, 0x4908, 0x29F8, 0x1090, 0x2890, 0x4512, 0x8212, 0x40E}},
    {23051, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0xFC20, 0x25FC, 0x2504, 0x2504, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2904, 0x4504, 0x8514, 0x108}},
    {23052, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x2524, 0x2524, 0x25FC, 0x2420, 0x4820, 0x29FC, 0x1020, 0x2820, 0x4420, 0x87FE, 0x0}},
    {23053, 2, {0x2028, 0x2024, 0x2024, 0x2020, 0xFBFE, 0x4A20, 0x4A24, 0x4A24, 0x8BA4, 0x4AA8, 0x32A8, 0x1290, 0x2A92, 0x4DAA, 0x8446, 0x882}},
    {23054, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23055, 2, {0x1090, 0x1088, 0x1080, 0x13FE, 0xFCA0, 0x24A8, 0x24B0, 0x24A4, 0x24A8, 0x48B0, 0x2924, 0x112A, 0x2932, 0x4622, 0x82DE, 0x400}},
    {23056, 2, {0x1008, 0x103C, 0x11C0, 0x1004, 0xFD44, 0x24A8, 0x2400, 0x25F8, 0x2410, 0x4820, 0x2BFE, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {23057, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x92A, 0x7230, 0x10C0, 0x1700, 0x200, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0xC70, 0x7008}},
    {23058, 2, {0x1080, 0x1080, 0x10FE, 0x1100, 0xFEFC, 0x2484, 0x24A4, 0x2494, 0x25FE, 0x4884, 0x2924, 0x1114, 0x29FE, 0x4404, 0x8428, 0x10}},
    {23059, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x4A00, 0x4A18, 0x4AE0, 0x8A38, 0x4AE0, 0x323C, 0x13E0, 0x2A22, 0x4C22, 0x841E, 0x800}},
    {23060, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x25FC, 0x2504, 0x2504, 0x2504, 0x49FC, 0x2850, 0x1050, 0x2892, 0x4492, 0x8112, 0x20E}},
    {23061, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x25FC, 0x2524, 0x2524, 0x25FC, 0x4820, 0x2870, 0x10A8, 0x2924, 0x4622, 0x8020, 0x20}},
    {23062, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x2504, 0x25FC, 0x2420, 0x2420, 0x4920, 0x293E, 0x1120, 0x2920, 0x46A0, 0x847E, 0x800}},
    {23063, 2, {0x2000, 0x2006, 0x23B8, 0x2088, 0xF888, 0x4908, 0x493E, 0x4B88, 0x8888, 0x4A88, 0x3288, 0x113E, 0x2900, 0x4A80, 0x847E, 0x800}},
    {23064, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0xFD04, 0x25FC, 0x2504, 0x2504, 0x25FC, 0x4920, 0x2922, 0x1114, 0x2908, 0x4544, 0x8582, 0x100}},
    {23065, 2, {0x1000, 0x11FE, 0x1000, 0x1092, 0xFD24, 0x2648, 0x2524, 0x2492, 0x2400, 0x49FE, 0x2820, 0x1020, 0x2820, 0x4420, 0x83FE, 0x0}},
    {23066, 2, {0x11FC, 0x1124, 0x1124, 0x11FC, 0xFD24, 0x2524, 0x25FC, 0x2440, 0x2440, 0x4BFE, 0x2842, 0x1082, 0x2882, 0x4502, 0x8214, 0x408}},
    {23067, 2, {0x2000, 0x20F8, 0x2088, 0x2088, 0xFAF8, 0x4A00, 0x4A00, 0x4BFC, 0x8844, 0x4844, 0x37FE, 0x1040, 0x28A0, 0x4910, 0x8208, 0x406}},
    {23068, 2, {0x2000, 0x2FDE, 0x2252, 0x2252, 0xF254, 0x57D4, 0x5258, 0x5254, 0x5252, 0x9FD2, 0x5252, 0x225A, 0x2254, 0x5450, 0x5550, 0x8890}},
    {23069, 2, {0x1000, 0x13FE, 0x1010, 0x1020, 0xFC68, 0x24A4, 0x2722, 0x2420, 0x2400, 0x49FC, 0x2904, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23070, 2, {0x2008, 0x203C, 0x23C0, 0x2044, 0xFA24, 0x4928, 0x4900, 0x4840, 0x8BFE, 0x4888, 0x3108, 0x1390, 0x2860, 0x4850, 0x8188, 0x604}},
    {23071, 2, {0x2000, 0x21F8, 0x2108, 0x2108, 0xF9F8, 0x4800, 0x4BFC, 0x4A04, 0x4A04, 0x93FC, 0x5204, 0x2204, 0x33FC, 0x4A04, 0x4A14, 0x8208}},
    {23072, 2, {0x2000, 0x23FC, 0x2200, 0x2200, 0xFAF8, 0x4A00, 0x4A00, 0x4BFC, 0x8AA0, 0x4AA4, 0x32A8, 0x1290, 0x2A90, 0x4C88, 0x84A4, 0x8C2}},
    {23073, 2, {0x2020, 0x2020, 0x2020, 0x23FE, 0xF820, 0x4820, 0x4820, 0x49FC, 0x8840, 0x4820, 0x30A4, 0x1282, 0x2A8A, 0x4A8A, 0x8478, 0x0}},
    {23074, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD44, 0x2622, 0x24F8, 0x2408, 0x2410, 0x4820, 0x29FC, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23075, 2, {0x1108, 0x1088, 0x1090, 0x11FC, 0xFC24, 0x2424, 0x25FC, 0x2520, 0x2520, 0x49FE, 0x2862, 0x10A2, 0x292A, 0x4624, 0x8420, 0x20}},
    {23076, 2, {0x2420, 0x2420, 0x2420, 0x3DFC, 0x420, 0xFC20, 0x2420, 0x24F8, 0x4600, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23077, 2, {0x1010, 0x10D8, 0x1394, 0x1094, 0xFC90, 0x27FE, 0x2490, 0x2494, 0x2494, 0x48D8, 0x2B98, 0x1090, 0x28AA, 0x44CA, 0x8486, 0x182}},
    {23078, 2, {0x1010, 0x1038, 0x11C0, 0x1100, 0xFD00, 0x25FC, 0x2510, 0x2510, 0x2510, 0x4910, 0x2BFE, 0x1000, 0x2890, 0x4508, 0x8204, 0x402}},
    {23079, 2, {0x1104, 0x1084, 0x1088, 0x1010, 0xFDFC, 0x2504, 0x2504, 0x2504, 0x25FC, 0x4850, 0x2850, 0x1090, 0x2892, 0x4512, 0x820E, 0x400}},
    {23080, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x25FC, 0x2400, 0x2400, 0x25FC, 0x4820, 0x2820, 0x13FE, 0x2820, 0x4420, 0x8420, 0x20}},
    {23081, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFE10, 0x25FC, 0x2524, 0x2524, 0x2524, 0x49FC, 0x2850, 0x1050, 0x2890, 0x4492, 0x8112, 0x20E}},
    {23082, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0xFDFC, 0x2484, 0x2484, 0x27FE, 0x2400, 0x4800, 0x29FC, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23083, 2, {0x2004, 0x200E, 0x23B8, 0x2088, 0xF888, 0x4928, 0x492E, 0x4BA8, 0x88A8, 0x4AA8, 0x32A8, 0x113E, 0x2900, 0x4A80, 0x847E, 0x800}},
    {23084, 2, {0x2010, 0x2018, 0x23D4, 0x2014, 0xF810, 0x4FFE, 0x4810, 0x4890, 0x8890, 0x4AD0, 0x3290, 0x1290, 0x2A8A, 0x4AEA, 0x8706, 0x202}},
    {23085, 2, {0x2020, 0x2040, 0x2088, 0x2104, 0xFBFE, 0x4882, 0x4880, 0x49FC, 0x8A20, 0x4820, 0x33FE, 0x1020, 0x2850, 0x4888, 0x8104, 0x602}},
    {23086, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFC00, 0x25FC, 0x2400, 0x2400, 0x25FC, 0x4800, 0x2800, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23087, 2, {0x2000, 0x20F8, 0x2088, 0x2088, 0xFAF8, 0x4A00, 0x4A00, 0x4BFC, 0x8804, 0x4804, 0x37FE, 0x1000, 0x28A0, 0x4910, 0x8208, 0x406}},
    {23088, 2, {0x2108, 0x2148, 0x2128, 0x22A8, 0xFAA8, 0x4E88, 0x4A88, 0x4A88, 0x4A88, 0x9288, 0x52A8, 0x22D4, 0x3294, 0x4A22, 0x4A42, 0x8282}},
    {23089, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0xFBF8, 0x4800, 0x4800, 0x4BF8, 0x8840, 0x4840, 0x37FE, 0x1040, 0x28A0, 0x4910, 0x8208, 0x406}},
    {23090, 2, {0x2000, 0x21FC, 0x2104, 0x2104, 0xF904, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x9222, 0x5252, 0x228A, 0x330A, 0x4A02, 0x4A0A, 0x8204}},
    {23091, 2, {0x2042, 0x20E2, 0x2382, 0x208A, 0xF88A, 0x488A, 0x4BEA, 0x488A, 0x498A, 0x91CA, 0x52AA, 0x2282, 0x3482, 0x4882, 0x488A, 0x8084}},
    {23092, 2, {0x1100, 0x10BC, 0x1004, 0x1124, 0xFD24, 0x2524, 0x25FC, 0x2524, 0x2574, 0x49AC, 0x2924, 0x1124, 0x2924, 0x4504, 0x8514, 0x108}},
    {23093, 2, {0x2000, 0x27E0, 0x225E, 0x2252, 0xFA52, 0x4BD2, 0x4A52, 0x4A54, 0x4BD4, 0x9254, 0x5248, 0x22E8, 0x3754, 0x4854, 0x4862, 0x8040}},
    {23094, 2, {0x0, 0x7F7C, 0x2244, 0x3E44, 0x2228, 0x3E10, 0x23A8, 0xFE44, 0x282, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23095, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFDFE, 0x24A8, 0x24A8, 0x24A8, 0x27FE, 0x48A8, 0x28A8, 0x13FE, 0x2820, 0x4420, 0x85FC, 0x0}},
    {23096, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x24F8, 0x2488, 0x2488, 0x24F8, 0x4888, 0x2888, 0x13FE, 0x2800, 0x4488, 0x8504, 0x202}},
    {23097, 2, {0x2010, 0x2388, 0x2088, 0x20FE, 0xF890, 0x4B90, 0x4A24, 0x4A24, 0x4A38, 0x9388, 0x5090, 0x2090, 0x30A4, 0x48BE, 0x4A82, 0x8100}},
    {23098, 2, {0x2000, 0x23DE, 0x2042, 0x214A, 0xF884, 0x494A, 0x4A52, 0x4C20, 0x8800, 0x4BDE, 0x3052, 0x1152, 0x2894, 0x4948, 0x8254, 0x422}},
    {23099, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x2524, 0x25FC, 0x2524, 0x25FC, 0x4820, 0x2870, 0x10A8, 0x2924, 0x4622, 0x8020, 0x20}},
    {23100, 2, {0x10FC, 0x1084, 0x1084, 0x10FC, 0xFC84, 0x2484, 0x24FC, 0x2400, 0x25FE, 0x4902, 0x2902, 0x11FE, 0x2902, 0x4502, 0x85FE, 0x102}},
    {23101, 2, {0x1000, 0x11F8, 0x1008, 0x1008, 0xFDF8, 0x2408, 0x2408, 0x27FE, 0x2420, 0x4A22, 0x2974, 0x10A8, 0x2924, 0x4622, 0x80A0, 0x40}},
    {23102, 2, {0x1000, 0x11FE, 0x1110, 0x1110, 0xFD7C, 0x2510, 0x2510, 0x25FE, 0x2500, 0x4910, 0x2910, 0x117C, 0x2910, 0x4610, 0x82FE, 0x400}},
    {23103, 2, {0x0, 0x7BFE, 0x4808, 0x51E8, 0x4928, 0x45E8, 0x5408, 0x4828, 0x4210, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23104, 2, {0x2000, 0x277E, 0x2504, 0x2504, 0xF574, 0x5554, 0x5654, 0x5554, 0x5554, 0x9554, 0x5574, 0x2554, 0x2604, 0x5404, 0x5414, 0x8408}},
    {23105, 2, {0x100, 0x1FF0, 0x1110, 0xFFFE, 0x1110, 0x1FF0, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23106, 2, {0x2020, 0x2040, 0x21FC, 0x2104, 0xF9FC, 0x4904, 0x49FC, 0x4820, 0x4820, 0x93FE, 0x5222, 0x2222, 0x322A, 0x4A24, 0x4820, 0x8020}},
    {23107, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xF800, 0x49FC, 0x4800, 0x4800, 0x4BFE, 0x9020, 0x5128, 0x2124, 0x3222, 0x4C22, 0x48A0, 0x8040}},
    {23108, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFD08, 0x2490, 0x2400, 0x27FE, 0x2400, 0x4800, 0x29FC, 0x1104, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23109, 2, {0x1100, 0x1100, 0x11FC, 0x1204, 0xFC44, 0x2554, 0x24E4, 0x2444, 0x27FC, 0x4844, 0x28E4, 0x1154, 0x2A44, 0x4444, 0x8414, 0x8}},
    {23110, 2, {0x2020, 0x13FE, 0x9222, 0x4220, 0x4BFC, 0x1284, 0xE448, 0x2430, 0x29CE, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23111, 2, {0x2008, 0x203C, 0x23C0, 0x2044, 0xFA24, 0x4928, 0x4900, 0x4820, 0x8BFE, 0x4870, 0x30A8, 0x10A8, 0x2924, 0x4A22, 0x8420, 0x20}},
    {23112, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xF820, 0x4BFE, 0x4888, 0x4944, 0x8A42, 0x48F8, 0x3188, 0x1250, 0x2820, 0x4850, 0x8188, 0x606}},
    {23113, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xFD04, 0x4900, 0x49DE, 0x4A52, 0x8A52, 0x4B52, 0x349A, 0x1094, 0x2910, 0x4912, 0x8212, 0x40E}},
    {23114, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x25FC, 0x2420, 0x2420, 0x27FE, 0x4840, 0x28A4, 0x11A8, 0x2A90, 0x4488, 0x84C6, 0x80}},
    {23115, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFDFE, 0x2522, 0x2538, 0x25E0, 0x2522, 0x491E, 0x2900, 0x1178, 0x2A48, 0x464A, 0x848A, 0x906}},
    {23116, 2, {0x2100, 0x2100, 0x21DC, 0x2114, 0xF114, 0x5114, 0x57D4, 0x5014, 0x5114, 0x9114, 0x5588, 0x2548, 0x2948, 0x5114, 0x5514, 0x8222}},
    {23117, 2, {0x1020, 0x1020, 0x11FC, 0x1050, 0xFC88, 0x2504, 0x27FE, 0x2408, 0x25E8, 0x4928, 0x2928, 0x11E8, 0x2928, 0x4408, 0x8428, 0x10}},
    {23118, 2, {0x10A0, 0x1090, 0x1090, 0x11FE, 0xFD20, 0x2520, 0x27FC, 0x2520, 0x2520, 0x49FC, 0x2920, 0x1120, 0x2920, 0x45FE, 0x8500, 0x100}},
    {23119, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFE10, 0x25FC, 0x2524, 0x2524, 0x2524, 0x49FC, 0x2850, 0x1058, 0x2894, 0x4492, 0x8112, 0x20E}},
    {23120, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x2524, 0x2524, 0x25FC, 0x2420, 0x4BFE, 0x2870, 0x10A8, 0x2924, 0x4622, 0x8020, 0x20}},
    {23121, 2, {0x201C, 0x21E0, 0x2020, 0x2020, 0xFBFE, 0x48A8, 0x4924, 0x4A42, 0x4840, 0x93FE, 0x5088, 0x2108, 0x30D0, 0x4830, 0x4848, 0x8184}},
    {23122, 2, {0x1020, 0x10A2, 0x10A2, 0x1124, 0xFC50, 0x2488, 0x2704, 0x2422, 0x2420, 0x48A4, 0x28A4, 0x1128, 0x2850, 0x4488, 0x8504, 0x202}},
    {23123, 2, {0x440, 0x440, 0xFC7E, 0x440, 0x7C7C, 0x440, 0xFC7E, 0x440, 0x640, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23124, 2, {0x2090, 0x2090, 0x2090, 0x279E, 0xF890, 0x4890, 0x4890, 0x4B9C, 0x4890, 0x9090, 0x5090, 0x279E, 0x3090, 0x4890, 0x4890, 0x8090}},
    {23125, 2, {0x2020, 0x2020, 0x23FE, 0x2020, 0xF9FC, 0x4824, 0x4BFE, 0x4824, 0x89FC, 0x4820, 0x3120, 0x113E, 0x2920, 0x4AA0, 0x827E, 0x400}},
    {23126, 2, {0x2000, 0x23FC, 0x2040, 0x2040, 0xFBFE, 0x48A0, 0x4910, 0x4A08, 0x4C46, 0x9040, 0x5048, 0x2264, 0x3252, 0x4C52, 0x4940, 0x8080}},
    {23127, 2, {0x2000, 0x2040, 0x239C, 0x2204, 0xFA04, 0x4B9C, 0x4A04, 0x4A04, 0x8BFC, 0x4890, 0x3090, 0x1090, 0x2890, 0x4912, 0x8212, 0x40E}},
    {23128, 2, {0x1020, 0x1124, 0x10A8, 0x1020, 0xFDFC, 0x2440, 0x27FE, 0x2488, 0x2504, 0x4AFA, 0x2C88, 0x1088, 0x28A8, 0x4492, 0x8482, 0x7E}},
    {23129, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFE10, 0x25FC, 0x2424, 0x2424, 0x25FE, 0x4824, 0x2824, 0x11FC, 0x2824, 0x4420, 0x84A0, 0x40}},
    {23130, 2, {0x201C, 0x23E0, 0x2220, 0x23FE, 0xFA20, 0x4A92, 0x4B0A, 0x4A06, 0x49FC, 0x9104, 0x5104, 0x21FC, 0x3104, 0x4904, 0x49FC, 0x8104}},
    {23131, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFC00, 0x25FC, 0x2504, 0x2504, 0x25FC, 0x4820, 0x2928, 0x1124, 0x2A22, 0x4422, 0x80A0, 0x40}},
    {23132, 2, {0x0, 0x7F7C, 0x4844, 0x7F44, 0x4128, 0x7F10, 0x4828, 0x7F44, 0x82, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23133, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xFC04, 0x4800, 0x4BFE, 0x4820, 0x8820, 0x4920, 0x313C, 0x1120, 0x2AA0, 0x4A60, 0x843E, 0x800}},
    {23134, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x27FE, 0x2488, 0x2450, 0x25FC, 0x4820, 0x2820, 0x13FE, 0x2820, 0x4420, 0x8420, 0x20}},
    {23135, 2, {0x2000, 0x23FE, 0x2222, 0x2222, 0xFBFE, 0x4A22, 0x4A22, 0x4AFA, 0x4A8A, 0x928A, 0x528A, 0x22FA, 0x328A, 0x4A02, 0x4BFE, 0x8202}},
    {23136, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x25F8, 0x2508, 0x2508, 0x25F8, 0x4900, 0x2900, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23137, 2, {0x2020, 0x2020, 0x2020, 0x23FE, 0xF820, 0x4924, 0x4924, 0x4924, 0x8AAA, 0x4870, 0x30A8, 0x10A8, 0x2924, 0x4A22, 0x8420, 0x20}},
    {23138, 2, {0x1040, 0x1080, 0x11FC, 0x1124, 0xFD24, 0x25FC, 0x2524, 0x2544, 0x25FC, 0x4890, 0x2910, 0x13FE, 0x2810, 0x4410, 0x8410, 0x10}},
    {23139, 2, {0x2050, 0x2252, 0x2252, 0x2252, 0xFBDE, 0x4A02, 0x4A02, 0x4BFE, 0x8A02, 0x4A02, 0x33DE, 0x1252, 0x2A52, 0x4252, 0x8452, 0x802}},
    {23140, 2, {0x2000, 0x23FE, 0x2222, 0x2222, 0xFAFA, 0x4A22, 0x4A22, 0x4BFE, 0x8A02, 0x4AFA, 0x328A, 0x128A, 0x2AFA, 0x4A02, 0x820A, 0x404}},
    {23141, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFDFC, 0x2504, 0x25FC, 0x2504, 0x25FC, 0x4924, 0x2820, 0x13FE, 0x2820, 0x4420, 0x8420, 0x20}},
    {23142, 2, {0x2000, 0x21FC, 0x2004, 0x20FC, 0xF804, 0x49FC, 0x4800, 0x4BFE, 0x4A22, 0x9020, 0x51FC, 0x2124, 0x3124, 0x4934, 0x4928, 0x8020}},
    {23143, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x2420, 0x27FE, 0x2400, 0x25FC, 0x4904, 0x29FC, 0x1104, 0x29FC, 0x4504, 0x8514, 0x108}},
    {23144, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x26FA, 0x2400, 0x25FC, 0x2554, 0x4954, 0x29FC, 0x1154, 0x2954, 0x4554, 0x8504, 0x10C}},
    {23145, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x2400, 0x25FE, 0x2500, 0x257C, 0x4910, 0x2910, 0x11FE, 0x2910, 0x4610, 0x8210, 0x410}},
    {23146, 2, {0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0xA20, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0x670, 0x3808}},
    {23147, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x2504, 0x2504, 0x25FC, 0x2400, 0x4912, 0x29D4, 0x1118, 0x2910, 0x4552, 0x8592, 0x10E}},
    {23148, 2, {0x1008, 0x103C, 0x11C0, 0x1044, 0xFD24, 0x24A8, 0x2490, 0x241C, 0x25E0, 0x4820, 0x2820, 0x13FE, 0x2820, 0x4420, 0x85FC, 0x0}},
    {23149, 2, {0x1000, 0x11FE, 0x1048, 0x1048, 0xFC48, 0x25CE, 0x2502, 0x2502, 0x2502, 0x49CE, 0x2848, 0x1048, 0x2848, 0x4448, 0x85FE, 0x0}},
    {23150, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x2510, 0x2510, 0x25FE, 0x2510, 0x4910, 0x297C, 0x1144, 0x2944, 0x4544, 0x857C, 0x244}},
    {23151, 2, {0x0, 0xFEFE, 0x0, 0x7C7C, 0x4444, 0x6464, 0x5454, 0x4444, 0x4644, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23152, 2, {0x1050, 0x1050, 0x1050, 0x11FC, 0xFD54, 0x2554, 0x2554, 0x25FC, 0x2554, 0x4954, 0x2954, 0x13FE, 0x2800, 0x4488, 0x8104, 0x202}},
    {23153, 2, {0x40, 0x7C20, 0x5FC, 0x7C40, 0x4088, 0x7DF0, 0x420, 0x448, 0x29FC, 0x1004, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23154, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x2450, 0x2450, 0x2494, 0x2594, 0x4A98, 0x2890, 0x10B2, 0x28D2, 0x4492, 0x848E, 0x80}},
    {23155, 2, {0x2000, 0x23FE, 0x2000, 0x2000, 0xF8F8, 0x48A8, 0x4AAA, 0x4AFA, 0x4AAA, 0x92AA, 0x52AA, 0x22FA, 0x3202, 0x4A02, 0x4BFE, 0x8002}},
    {23156, 2, {0x0, 0x3EF8, 0x2288, 0x2AA8, 0x2AA8, 0x2AA8, 0x1450, 0x6288, 0x200, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0x670, 0x3808}},
    {23157, 2, {0x1104, 0x1088, 0x1050, 0x11FC, 0xFD24, 0x2524, 0x25FC, 0x2524, 0x2524, 0x49FC, 0x2820, 0x1020, 0x2BFE, 0x4420, 0x8420, 0x20}},
    {23158, 2, {0x2040, 0x2020, 0x23FE, 0x2222, 0xF820, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x9124, 0x5124, 0x21FC, 0x3124, 0x4820, 0x4820, 0x8020}},
    {23159, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x2504, 0x25FC, 0x2400, 0x27FE, 0x4A02, 0x29FC, 0x1020, 0x2820, 0x4420, 0x84A0, 0x40}},
    {23160, 2, {0x1000, 0x10FC, 0x1084, 0x10FC, 0xFC84, 0x24FC, 0x2400, 0x25FE, 0x2480, 0x48FE, 0x292A, 0x124A, 0x2892, 0x4522, 0x8454, 0x88}},
    {23161, 2, {0x2000, 0x23FE, 0x2050, 0x2050, 0xFBFE, 0x4A52, 0x4A52, 0x4BFE, 0x4800, 0x9040, 0x57FE, 0x2088, 0x3190, 0x4860, 0x48D8, 0x8304}},
    {23162, 2, {0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x98C6, 0x200, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0x670, 0x3808}},
    {23163, 2, {0x2020, 0x2020, 0x23FE, 0x2020, 0xF820, 0x4BFE, 0x4A8A, 0x4A52, 0x4AFA, 0x9222, 0x5222, 0x22FA, 0x3222, 0x4A22, 0x4A2A, 0x8204}},
    {23164, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCA8, 0x2420, 0x27FE, 0x2440, 0x2480, 0x48FC, 0x2984, 0x1284, 0x2884, 0x4484, 0x84FC, 0x84}},
    {23165, 2, {0x2000, 0x23DE, 0x2242, 0x2242, 0xFA42, 0x4BDE, 0x4A00, 0x4A3E, 0x4BD2, 0x9212, 0x5214, 0x23D4, 0x3208, 0x4A14, 0x4A24, 0x8242}},
    {23166, 2, {0x21C0, 0x20A0, 0x2110, 0x2208, 0xFDF6, 0x4800, 0x4BCA, 0x4A4A, 0x4A54, 0x93D4, 0x5268, 0x2254, 0x33D4, 0x4A4A, 0x4A4A, 0x82C0}},
    {23167, 2, {0x1000, 0x13FE, 0x1022, 0x1120, 0xFD3C, 0x2520, 0x26FE, 0x2400, 0x25FC, 0x4904, 0x29FC, 0x1104, 0x29FC, 0x4504, 0x8114, 0x108}},
    {23168, 2, {0x21FC, 0x2124, 0x2124, 0x21FC, 0xF924, 0x4924, 0x49FC, 0x4820, 0x4820, 0x93FE, 0x5222, 0x222A, 0x32FA, 0x4A0A, 0x4A02, 0x8206}},
    {23169, 2, {0x2040, 0x2040, 0x21FC, 0x2084, 0xFBFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x9020, 0x53FE, 0x2020, 0x3220, 0x4BFE, 0x4820, 0x8020}},
    {23170, 2, {0x2040, 0x2020, 0x21FC, 0x2000, 0xF888, 0x4850, 0x4BFE, 0x4A22, 0x4C24, 0x91FC, 0x5124, 0x2124, 0x3134, 0x4928, 0x4820, 0x8020}},
    {23171, 2, {0x2000, 0x21FC, 0x2048, 0x2030, 0xFBFE, 0x4852, 0x4894, 0x4910, 0x8A30, 0x4820, 0x33FE, 0x1070, 0x28A8, 0x4924, 0x8622, 0x20}},
    {23172, 2, {0x1104, 0x1088, 0x13FE, 0x1020, 0xFC20, 0x25FC, 0x2420, 0x2420, 0x27FE, 0x4800, 0x2820, 0x13FE, 0x2820, 0x4450, 0x8088, 0x306}},
    {23173, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x2488, 0x24F8, 0x2488, 0x2488, 0x4BFE, 0x2900, 0x1148, 0x2984, 0x4500, 0x85FE, 0x0}},
    {23174, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0xFDFC, 0x2554, 0x2554, 0x2554, 0x2554, 0x492C, 0x2820, 0x13FE, 0x2850, 0x4488, 0x8104, 0x202}},
    {23175, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFC88, 0x2450, 0x27FE, 0x2420, 0x2420, 0x49FC, 0x2820, 0x10A8, 0x2924, 0x4622, 0x80A0, 0x40}},
    {23176, 2, {0x2000, 0x23FE, 0x2222, 0x2020, 0xFBFE, 0x4820, 0x49FC, 0x4924, 0x49FC, 0x9124, 0x51FC, 0x2020, 0x33FE, 0x4820, 0x4820, 0x8020}},
    {23177, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x4A00, 0x4BFC, 0x4A40, 0x8A88, 0x4BFC, 0x3224, 0x1220, 0x2DFC, 0x4420, 0x8820, 0x13FE}},
    {23178, 2, {0x2108, 0x2088, 0x2090, 0x27FE, 0xF800, 0x4BC4, 0x4A54, 0x4A54, 0x4BD4, 0x9254, 0x5254, 0x23D4, 0x3254, 0x4A44, 0x4A54, 0x82C8}},
    {23179, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x2440, 0x27FE, 0x2488, 0x2504, 0x4AFA, 0x2888, 0x1088, 0x28F8, 0x4488, 0x8488, 0xF8}},
    {23180, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x2400, 0x25FC, 0x2524, 0x2524, 0x4924, 0x29FC, 0x1124, 0x2924, 0x4524, 0x85FC, 0x104}},
    {23181, 2, {0x1040, 0x1040, 0x10F8, 0x1108, 0xFE10, 0x25FC, 0x2504, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x11FC, 0x2800, 0x4488, 0x8104, 0x202}},
    {23182, 2, {0x1020, 0x1022, 0x11FA, 0x1024, 0xFC28, 0x27FE, 0x2420, 0x2440, 0x24FC, 0x4984, 0x2A84, 0x10FC, 0x2884, 0x4484, 0x84FC, 0x84}},
    {23183, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x2400, 0x27FE, 0x2420, 0x2440, 0x49FC, 0x2954, 0x1154, 0x2954, 0x4554, 0x8554, 0x10C}},
    {23184, 2, {0x0, 0x7F7C, 0x4444, 0x4444, 0x5F7C, 0x5140, 0x5F42, 0x4442, 0x7FBE, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23185, 2, {0x2008, 0x203C, 0x21E0, 0x2020, 0xFBFE, 0x4820, 0x49FC, 0x4924, 0x49FC, 0x9124, 0x51FC, 0x2020, 0x31FC, 0x4820, 0x4BFE, 0x8000}},
    {23186, 2, {0x2110, 0x2110, 0x2110, 0x27FC, 0xF910, 0x29F0, 0x2910, 0x29F0, 0x4840, 0x4FFE, 0x2840, 0x10E0, 0x1150, 0x2A4E, 0x4444, 0x8040}},
    {23187, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x2504, 0x25FC, 0x2400, 0x27FE, 0x4820, 0x2820, 0x11FC, 0x2820, 0x4420, 0x83FE, 0x0}},
    {23188, 2, {0x2000, 0x27FE, 0x2040, 0x2080, 0xFBFC, 0x4A94, 0x4A94, 0x4AF4, 0x4A94, 0x9294, 0x52F4, 0x2294, 0x3294, 0x4A94, 0x4BFC, 0x8204}},
    {23189, 2, {0x1020, 0x1050, 0x1088, 0x1104, 0xFEFA, 0x2400, 0x25FC, 0x2504, 0x2504, 0x49FC, 0x2888, 0x1088, 0x2BFE, 0x4488, 0x8108, 0x208}},
    {23190, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF950, 0x4840, 0x4BF8, 0x4A48, 0x8A48, 0x4A48, 0x37FE, 0x1040, 0x28A0, 0x4910, 0x8208, 0x406}},
    {23191, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xFC04, 0x49FC, 0x4800, 0x49FC, 0x4904, 0x91FC, 0x5104, 0x21FC, 0x3104, 0x4800, 0x4BFE, 0x8000}},
    {23192, 2, {0x1110, 0x1112, 0x11D4, 0x1118, 0xFD52, 0x2592, 0x252E, 0x2440, 0x25FC, 0x4904, 0x2904, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23193, 2, {0x200C, 0x200A, 0x2008, 0x23FE, 0xFA08, 0x4BF8, 0x4A4A, 0x4A4A, 0x8BEA, 0x4AAC, 0x32AC, 0x124C, 0x2AAA, 0x4B0A, 0x8216, 0x422}},
    {23194, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x2500, 0x257C, 0x2544, 0x2544, 0x497C, 0x2944, 0x117C, 0x2944, 0x4644, 0x827C, 0x444}},
    {23195, 2, {0x203C, 0x27C0, 0x2244, 0x2128, 0xF800, 0x4BFC, 0x4880, 0x4880, 0x4FFE, 0x9100, 0x51F8, 0x2288, 0x3250, 0x4C20, 0x48D8, 0x8306}},
    {23196, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFC20, 0x25FC, 0x2504, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x11FC, 0x2800, 0x4488, 0x8504, 0x202}},
    {23197, 2, {0x610, 0x3810, 0x852, 0x7E54, 0x890, 0x1C28, 0x2A28, 0x4844, 0xA82, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23198, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x2504, 0x2504, 0x25FC, 0x2400, 0x4BFE, 0x2820, 0x1120, 0x293C, 0x4520, 0x82A0, 0x47E}},
    {23199, 2, {0x1048, 0x1148, 0x1148, 0x13FE, 0xFD48, 0x2548, 0x2578, 0x2500, 0x25FE, 0x4820, 0x2BFE, 0x1070, 0x28A8, 0x4524, 0x8622, 0x20}},
    {23200, 2, {0x1040, 0x1040, 0x13FE, 0x1080, 0xFDFC, 0x2620, 0x25FE, 0x2400, 0x25FC, 0x4904, 0x29FC, 0x1104, 0x29FC, 0x4504, 0x8514, 0x108}},
    {23201, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x2524, 0x25AC, 0x2574, 0x2524, 0x49FC, 0x2820, 0x1070, 0x28A8, 0x4524, 0x8222, 0x20}},
    {23202, 2, {0x1000, 0x11FE, 0x1102, 0x117A, 0xFD02, 0x257A, 0x2502, 0x2400, 0x24FC, 0x4884, 0x28FC, 0x1084, 0x28FC, 0x4484, 0x84FC, 0x84}},
    {23203, 2, {0x2120, 0x20A0, 0x2478, 0x2228, 0xF8A8, 0x492A, 0x4F2A, 0x4946, 0x49A0, 0x9020, 0x53FE, 0x2070, 0x30A8, 0x4924, 0x4A22, 0x8020}},
    {23204, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x2524, 0x2524, 0x25FC, 0x2420, 0x4810, 0x2854, 0x1142, 0x2942, 0x454A, 0x8238, 0x0}},
    {23205, 2, {0x2040, 0x2020, 0x21FE, 0x2102, 0xF902, 0x49FE, 0x4900, 0x4900, 0x89FE, 0x49AA, 0x31AA, 0x12FE, 0x2AAA, 0x4AAA, 0x84A2, 0x86}},
    {23206, 2, {0x2000, 0x23FE, 0x2222, 0x23FE, 0xFA22, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x91FC, 0x5104, 0x21FC, 0x3104, 0x4904, 0x4914, 0x8108}},
    {23207, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0xFCF4, 0x2494, 0x2494, 0x25FE, 0x2502, 0x497A, 0x294A, 0x114A, 0x297A, 0x4502, 0x850A, 0x104}},
    {23208, 2, {0x2108, 0x2088, 0x2090, 0x27FE, 0xF890, 0x4890, 0x4BFC, 0x4A94, 0x4A94, 0x930C, 0x5204, 0x23FC, 0x3204, 0x4A04, 0x4BFC, 0x8204}},
    {23209, 2, {0x1080, 0x109E, 0x1092, 0x13F2, 0xFC92, 0x249E, 0x2492, 0x25D2, 0x2552, 0x495E, 0x2952, 0x1152, 0x29D2, 0x4422, 0x842A, 0x44}},
    {23210, 2, {0x2000, 0x21F8, 0x2108, 0x2108, 0xF9F8, 0x4908, 0x4908, 0x49F8, 0x8800, 0x4BFC, 0x3294, 0x1294, 0x2A94, 0x4A94, 0x87FE, 0x0}},
    {23211, 2, {0x2040, 0x2040, 0x207E, 0x2040, 0xFBFC, 0x4804, 0x49FC, 0x4804, 0x8BFC, 0x4840, 0x3240, 0x127C, 0x2A40, 0x4540, 0x84FE, 0x800}},
    {23212, 2, {0x2100, 0x217C, 0x2144, 0x2244, 0xFA44, 0x4E7C, 0x4A10, 0x4A10, 0x4AFE, 0x9238, 0x5254, 0x2254, 0x3294, 0x4B12, 0x4A10, 0x8210}},
    {23213, 2, {0xBFE, 0x3020, 0xC1FC, 0x904, 0x3124, 0xC524, 0x924, 0x3050, 0xC088, 0x304, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23214, 2, {0x2040, 0x20A0, 0x2110, 0x2208, 0xFDF6, 0x4800, 0x4BC4, 0x4A54, 0x4A54, 0x93D4, 0x5254, 0x2254, 0x33D4, 0x4A44, 0x4A54, 0x82C8}},
    {23215, 2, {0x2040, 0x2240, 0x2140, 0x27F8, 0xF888, 0x4890, 0x48FC, 0x4904, 0x8908, 0x4BFE, 0x3202, 0x142A, 0x2AAA, 0x4A82, 0x8414, 0x8}},
    {23216, 2, {0x2080, 0x2080, 0x21FC, 0x2244, 0xF954, 0x49F4, 0x4884, 0x4928, 0x4A90, 0x91FC, 0x5244, 0x2554, 0x31F4, 0x4884, 0x4928, 0x8210}},
    {23217, 2, {0x2008, 0x203C, 0x23C0, 0x2004, 0xFA44, 0x4928, 0x49FC, 0x4A20, 0x4820, 0x93FE, 0x5020, 0x2124, 0x3124, 0x4924, 0x49FC, 0x8004}},
    {23218, 2, {0x1020, 0x1040, 0x11FC, 0x1154, 0xFD24, 0x2554, 0x2504, 0x25FC, 0x2400, 0x4912, 0x29D4, 0x1118, 0x2910, 0x4552, 0x8592, 0x10E}},
    {23219, 2, {0x2020, 0x2040, 0x21FC, 0x2104, 0xF9FC, 0x4904, 0x49FC, 0x4904, 0x89FC, 0x4840, 0x3020, 0x10A4, 0x2A8A, 0x4A8A, 0x8478, 0x0}},
    {23220, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xFBFE, 0x4800, 0x49FC, 0x4904, 0x4904, 0x91FC, 0x5052, 0x2094, 0x3188, 0x4EA4, 0x48C2, 0x8080}},
    {23221, 2, {0x20, 0x7924, 0x48A8, 0x4820, 0x49FC, 0x7840, 0x4BFE, 0x4888, 0x4944, 0x7A42, 0x4DFC, 0x4888, 0x49D0, 0x4830, 0x4848, 0x9984}},
    {23222, 2, {0x2108, 0x2108, 0x27FE, 0x2108, 0xF800, 0x4FFE, 0x4908, 0x49F8, 0x4908, 0x91F8, 0x5108, 0x213E, 0x37C8, 0x4808, 0x4808, 0x8008}},
    {23223, 2, {0x2000, 0x23FE, 0x2200, 0x22FC, 0xFA00, 0x4BFE, 0x4AA8, 0x4A90, 0x8AC8, 0x4A86, 0x3208, 0x13FE, 0x2A88, 0x4C48, 0x8408, 0x818}},
    {23224, 2, {0x2020, 0x2124, 0x2124, 0x21FC, 0xF820, 0x4BFE, 0x4800, 0x4820, 0x49FC, 0x9124, 0x5124, 0x21FC, 0x3020, 0x4824, 0x4BFE, 0x8102}},
    {23225, 2, {0x2040, 0x23BE, 0x2212, 0x2292, 0xFA52, 0x4AAA, 0x4B24, 0x4840, 0x4BFE, 0x9222, 0x5222, 0x23FE, 0x3222, 0x4A22, 0x4BFE, 0x8202}},
    {23226, 2, {0x2108, 0x2548, 0x2548, 0x2550, 0xF7DE, 0x5024, 0x5014, 0x5FD4, 0x5014, 0x9794, 0x5494, 0x24A8, 0x24C8, 0x5494, 0x5814, 0x9022}},
    {23227, 2, {0x1078, 0x3E48, 0x2248, 0x2A86, 0xFF78, 0x2248, 0x2A28, 0x4210, 0x4A28, 0x84C6, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23228, 2, {0x1000, 0x10F8, 0x10A8, 0x10A8, 0xFCD8, 0x2488, 0x24F8, 0x2400, 0x2400, 0x49FC, 0x2954, 0x1154, 0x2954, 0x4554, 0x83FE, 0x0}},
    {23229, 2, {0x2000, 0x21FE, 0x2120, 0x21FC, 0xF920, 0x49FC, 0x4920, 0x4920, 0x89FE, 0x4802, 0x32AA, 0x12AA, 0x2AAA, 0x4A02, 0x8014, 0x8}},
    {23230, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFDFC, 0x2488, 0x27FE, 0x2420, 0x25FC, 0x4924, 0x29FC, 0x1124, 0x2BFE, 0x4504, 0x8514, 0x108}},
    {23231, 2, {0x2020, 0x2040, 0x23FC, 0x2224, 0xFA24, 0x4BFC, 0x4A24, 0x4A44, 0x8BFC, 0x4840, 0x30A8, 0x10B4, 0x293C, 0x4922, 0x8222, 0x41E}},
    {23232, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0xFBF8, 0x4880, 0x4FFC, 0x4910, 0x8A68, 0x4DC6, 0x3040, 0x13F8, 0x28E0, 0x4958, 0x8644, 0x40}},
    {23233, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xF800, 0x49FC, 0x4840, 0x48A2, 0x4B34, 0x9058, 0x5094, 0x2334, 0x3052, 0x4890, 0x4B50, 0x8020}},
    {23234, 2, {0x10A0, 0x112C, 0x1124, 0x1124, 0xFDAC, 0x2524, 0x2524, 0x25FC, 0x2420, 0x49FC, 0x2888, 0x1050, 0x2820, 0x4450, 0x8088, 0x306}},
    {23235, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x2504, 0x25FC, 0x2504, 0x25FC, 0x4904, 0x29FC, 0x1104, 0x2BFE, 0x4488, 0x8104, 0x202}},
    {23236, 2, {0x1000, 0x11FE, 0x1110, 0x1120, 0xFD7C, 0x2544, 0x257C, 0x2544, 0x257C, 0x4910, 0x2910, 0x1154, 0x2952, 0x4692, 0x8250, 0x420}},
    {23237, 2, {0x2108, 0x2088, 0x2090, 0x23FC, 0xF840, 0x49F8, 0x4840, 0x4BFE, 0x4880, 0x9100, 0x51FC, 0x2220, 0x3420, 0x4820, 0x4BFE, 0x8000}},
    {23238, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC50, 0x2488, 0x2524, 0x2450, 0x2488, 0x4904, 0x2AFA, 0x1088, 0x2888, 0x4488, 0x84F8, 0x88}},
    {23239, 2, {0x1000, 0x13FE, 0x1202, 0x1000, 0xFDFC, 0x2504, 0x25FC, 0x2504, 0x25FC, 0x4840, 0x2820, 0x13FE, 0x2800, 0x4488, 0x8104, 0x202}},
    {23240, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8204, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23241, 2, {0x2020, 0x2010, 0x21FE, 0x2100, 0xF920, 0x4D20, 0x4B3C, 0x4950, 0x8990, 0x4B10, 0x35FE, 0x1110, 0x2928, 0x4A28, 0x8244, 0x482}},
    {23242, 2, {0x2020, 0x2020, 0x23FE, 0x2020, 0xF9FC, 0x4820, 0x4BFE, 0x4840, 0x4888, 0x91F0, 0x5024, 0x23FE, 0x3022, 0x4924, 0x4AA2, 0x8040}},
    {23243, 2, {0x2000, 0x23DE, 0x2042, 0x2042, 0xFBDE, 0x4A10, 0x4A10, 0x4BDE, 0x4842, 0x9252, 0x514A, 0x214A, 0x3252, 0x4842, 0x4A94, 0x8108}},
    {23244, 2, {0x1088, 0x1050, 0x1000, 0x13FE, 0xFC50, 0x25FC, 0x2454, 0x27FE, 0x2454, 0x49FC, 0x2850, 0x10D8, 0x2954, 0x4652, 0x8450, 0x50}},
    {23245, 2, {0x101E, 0x11E0, 0x1022, 0x1112, 0xFC94, 0x2480, 0x2420, 0x25CE, 0x2502, 0x4902, 0x29CE, 0x1102, 0x2902, 0x4502, 0x85FE, 0x102}},
    {23246, 2, {0x2040, 0x2020, 0x23FC, 0x2108, 0xF890, 0x4BFE, 0x4A02, 0x4C44, 0x4820, 0x93FC, 0x5080, 0x20F8, 0x3088, 0x4908, 0x4928, 0x8210}},
    {23247, 2, {0x2100, 0x209E, 0x23D2, 0x2252, 0xFA54, 0x4BD4, 0x4A58, 0x4A54, 0x4BD2, 0x9212, 0x5292, 0x225A, 0x32D4, 0x4B50, 0x4A10, 0x8010}},
    {23248, 2, {0x2008, 0x27C8, 0x2548, 0x2548, 0xF7FE, 0x5554, 0x5554, 0x57D4, 0x5214, 0x9224, 0x57D4, 0x2248, 0x2248, 0x5254, 0x5554, 0x88A0}},
    {23249, 2, {0x7FFC, 0x300, 0xD70, 0x710C, 0x100, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38}},
    {23250, 2, {0x200E, 0x23F0, 0x2044, 0x2224, 0xF908, 0x4BFE, 0x4A42, 0x4840, 0x4BFE, 0x9080, 0x50FC, 0x2144, 0x3128, 0x4A10, 0x4C68, 0x8186}},
    {23251, 2, {0x1080, 0x10F8, 0x1108, 0x1210, 0xFDFC, 0x2524, 0x2554, 0x258C, 0x2504, 0x4810, 0x2912, 0x11D4, 0x2918, 0x4552, 0x8592, 0x10E}},
    {23252, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xFC14, 0x49E0, 0x4900, 0x4900, 0x49FC, 0x9110, 0x5110, 0x27FE, 0x3000, 0x4890, 0x4908, 0x8204}},
    {23253, 2, {0x1000, 0x11FE, 0x1140, 0x117C, 0xFD90, 0x2510, 0x25FE, 0x2528, 0x2544, 0x49FE, 0x2820, 0x1010, 0x2944, 0x4542, 0x824A, 0x38}},
    {23254, 2, {0x1000, 0x13FE, 0x1050, 0x11FC, 0xFD54, 0x2554, 0x25FC, 0x2400, 0x25FC, 0x4800, 0x2BFE, 0x1020, 0x28A8, 0x4524, 0x82A2, 0x40}},
    {23255, 2, {0x2000, 0x23FE, 0x2200, 0x227C, 0xFA44, 0x4A44, 0x4A7C, 0x4A00, 0x4AEE, 0x92AA, 0x52AA, 0x22AA, 0x32EE, 0x4A00, 0x4BFE, 0x8000}},
    {23256, 2, {0x1000, 0x11FC, 0x1124, 0x11FC, 0xFD24, 0x25FC, 0x2440, 0x2488, 0x25F0, 0x4820, 0x2844, 0x13FE, 0x2822, 0x4524, 0x8222, 0x60}},
    {23257, 2, {0x2420, 0x2220, 0x2220, 0x2F7E, 0xF480, 0x5400, 0x577C, 0x5514, 0x5510, 0x9550, 0x555C, 0x2550, 0x2950, 0x5B50, 0x50BE, 0xA100}},
    {23258, 2, {0x2000, 0x21F8, 0x2108, 0x21F8, 0xF908, 0x49F8, 0x4800, 0x4BFC, 0x4A94, 0x93FC, 0x5000, 0x21F8, 0x3090, 0x4860, 0x4998, 0x8606}},
    {23259, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23260, 2, {0x2020, 0x21FC, 0x2088, 0x2050, 0xFBFE, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x9104, 0x51FC, 0x2020, 0x33FE, 0x4820, 0x4820, 0x8020}},
    {23261, 2, {0x1020, 0x1010, 0x11FE, 0x1110, 0xFD7C, 0x2514, 0x25FE, 0x2514, 0x257C, 0x4910, 0x2992, 0x1154, 0x2938, 0x4654, 0x8292, 0x430}},
    {23262, 2, {0x2020, 0x23FE, 0x2220, 0x22FC, 0xFA24, 0x4BFE, 0x4A24, 0x4AFC, 0x8A20, 0x4AFC, 0x32A4, 0x12FC, 0x2AA4, 0x4CFC, 0x84A4, 0x8AC}},
    {23263, 2, {0x2000, 0x23FE, 0x2248, 0x2248, 0xFBFE, 0x4A48, 0x4A20, 0x4BFE, 0x4A40, 0x92FC, 0x5344, 0x2244, 0x327C, 0x4A00, 0x4BFE, 0x8000}},
    {23264, 2, {0x820, 0x3E7E, 0x8A4, 0x7F28, 0x1C10, 0x2A28, 0x4944, 0x3FFE, 0x2100, 0x2100, 0x3FFE, 0x2210, 0x2720, 0x40C0, 0x4330, 0x8C0C}},
    {23265, 2, {0x2040, 0x2020, 0x23FE, 0x2088, 0xF850, 0x4BFE, 0x4A22, 0x4AFA, 0x4A22, 0x92FA, 0x528A, 0x228A, 0x32FA, 0x4A02, 0x4A0A, 0x8204}},
    {23266, 2, {0x11F8, 0x7D08, 0x11F8, 0x1108, 0xFDF8, 0x1108, 0x29F8, 0x2490, 0x4112, 0x860E, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23267, 2, {0x2000, 0x27FC, 0x2040, 0x2278, 0xFA40, 0x4FFE, 0x4900, 0x49FC, 0x4A00, 0x93FC, 0x5004, 0x2554, 0x3554, 0x4804, 0x4828, 0x8010}},
    {23268, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x2420, 0x25FC, 0x2524, 0x25FC, 0x4820, 0x2BFE, 0x1020, 0x29FC, 0x4420, 0x83FE, 0x0}},
    {23269, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFD24, 0x25FC, 0x2524, 0x25FC, 0x2422, 0x4BFE, 0x2808, 0x13FE, 0x2908, 0x4488, 0x8428, 0x10}},
    {23270, 2, {0x2020, 0x2124, 0x20A8, 0x23FE, 0xFA02, 0x48F8, 0x4888, 0x4888, 0x48F8, 0x9020, 0x51FC, 0x2124, 0x3124, 0x4934, 0x4928, 0x8020}},
    {23271, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC20, 0x27FE, 0x2400, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x11FC, 0x2904, 0x45FC, 0x8488, 0x104}},
    {23272, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x2420, 0x25FC, 0x2524, 0x25FC, 0x4820, 0x29FC, 0x1020, 0x2BFE, 0x4450, 0x8088, 0x306}},
    {23273, 2, {0x2108, 0x2108, 0x27C8, 0x2110, 0xF11E, 0x57E4, 0x5554, 0x5554, 0x57D4, 0x9114, 0x5394, 0x2548, 0x2948, 0x5114, 0x5114, 0x8122}},
    {23274, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xFA94, 0x4CA4, 0x4850, 0x4988, 0x4E26, 0x90C0, 0x5310, 0x2064, 0x3388, 0x4830, 0x48C0, 0x8700}},
    {23275, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF800, 0x4BF8, 0x4A08, 0x4BF8, 0x8A08, 0x4BF8, 0x3040, 0x17FC, 0x28A0, 0x4910, 0x8208, 0x406}},
    {23276, 2, {0x2040, 0x2020, 0x23FE, 0x2200, 0xFA48, 0x4A48, 0x4BFE, 0x4A48, 0x8A48, 0x4A48, 0x3278, 0x1200, 0x2D54, 0x452A, 0x8A2A, 0x1000}},
    {23277, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0xFA42, 0x4A78, 0x4BC4, 0x4A3C, 0x8A08, 0x4AF0, 0x3324, 0x12A8, 0x2DFE, 0x4420, 0x88A0, 0x40}},
    {23278, 2, {0x11FC, 0x1020, 0x13FE, 0x1222, 0xFDAC, 0x2420, 0x25AC, 0x2400, 0x25FC, 0x4800, 0x2BFE, 0x1080, 0x28FC, 0x4404, 0x8428, 0x10}},
    {23279, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x8200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23280, 2, {0x2108, 0x2108, 0x27C8, 0x2108, 0xF11E, 0x57D2, 0x5564, 0x5540, 0x57C8, 0x9108, 0x5388, 0x2548, 0x2954, 0x5114, 0x5124, 0x8142}},
    {23281, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFD24, 0x24A8, 0x27FE, 0x2400, 0x25FC, 0x4904, 0x2974, 0x1154, 0x2974, 0x4504, 0x85FC, 0x104}},
    {23282, 2, {0x2040, 0x2020, 0x27FE, 0x2490, 0xF490, 0x5490, 0x57FC, 0x5490, 0x54D8, 0x95B8, 0x55B4, 0x26D4, 0x2892, 0x5890, 0x5090, 0x8090}},
    {23283, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23284, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFD04, 0x25FC, 0x2420, 0x25FC, 0x2488, 0x4850, 0x2BFE, 0x1020, 0x29FC, 0x4420, 0x8420, 0x20}},
    {23285, 2, {0x2080, 0x2080, 0x21FE, 0x2354, 0xFD54, 0x4954, 0x4BFE, 0x4954, 0x4954, 0x9154, 0x57FE, 0x2000, 0x3154, 0x492A, 0x4A2A, 0x8000}},
    {23286, 2, {0x2090, 0x2088, 0x21FE, 0x2110, 0xFB10, 0x4DFE, 0x4910, 0x4910, 0x89FE, 0x4910, 0x3110, 0x11FE, 0x2900, 0x4AA4, 0x8252, 0x452}},
    {23287, 2, {0x2020, 0x2720, 0x25FE, 0x2540, 0xF57C, 0x5690, 0x557E, 0x5500, 0x557C, 0x9544, 0x557C, 0x2644, 0x247C, 0x5444, 0x5444, 0x844C}},
    {23288, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0xFDFC, 0x2420, 0x27FE, 0x2524, 0x24A8, 0x4BFE, 0x2800, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23289, 2, {0x1088, 0x1088, 0x11FC, 0x1088, 0xFC88, 0x27FE, 0x2420, 0x25FC, 0x2524, 0x49FC, 0x2924, 0x11FC, 0x2800, 0x4488, 0x8104, 0x202}},
    {23290, 2, {0x239C, 0x2294, 0x239C, 0x2294, 0xFB9C, 0x4A04, 0x4AF4, 0x4A94, 0x4AF4, 0x9294, 0x52F4, 0x2294, 0x3294, 0x4B34, 0x4A04, 0x820C}},
    {23291, 2, {0x2000, 0x23DE, 0x2252, 0x23DE, 0xFA52, 0x4BDE, 0x4A22, 0x4A22, 0x4AFA, 0x9222, 0x5272, 0x22AA, 0x3326, 0x4A22, 0x4A2A, 0x8204}},
    {23292, 2, {0x2000, 0x21FC, 0x2124, 0x21AC, 0xF974, 0x4924, 0x49FC, 0x4820, 0x89FC, 0x4820, 0x33FE, 0x1000, 0x2AA4, 0x4A52, 0x8452, 0x0}},
    {23293, 2, {0x2020, 0x2020, 0x23FE, 0x2050, 0xFA8A, 0x4904, 0x4BFE, 0x4D04, 0x49FC, 0x9104, 0x51FC, 0x2020, 0x3124, 0x4A22, 0x4CA2, 0x8040}},
    {23294, 2, {0x2020, 0x2124, 0x20A8, 0x23FE, 0xF8A8, 0x4924, 0x4A22, 0x4904, 0x8904, 0x49DE, 0x3244, 0x1554, 0x289E, 0x4904, 0x8204, 0x404}},
    {23295, 2, {0x2020, 0x21FC, 0x2024, 0x23FE, 0xF824, 0x49FC, 0x4820, 0x4BFE, 0x4800, 0x91FC, 0x5124, 0x21FC, 0x3124, 0x49FC, 0x4800, 0x83FE}},
    {23296, 2, {0x2004, 0x203E, 0x27C0, 0x2244, 0xF128, 0x53F8, 0x5208, 0x53FC, 0x5204, 0x93FE, 0x5402, 0x26AA, 0x2AAA, 0x5A02, 0x5014, 0x8008}},
    {23297, 2, {0x2020, 0x27A4, 0x20A8, 0x2292, 0xF914, 0x4A08, 0x4DF4, 0x4802, 0x8BF8, 0x4A08, 0x3208, 0x13F8, 0x2A08, 0x4110, 0x8FFE, 0x0}},
    {23298, 2, {0x2110, 0x2094, 0x23D2, 0x2012, 0xFA50, 0x4990, 0x4FFE, 0x4810, 0x4BD2, 0x9252, 0x5254, 0x23D4, 0x324A, 0x4A4A, 0x4BD6, 0x8022}},
    {23299, 2, {0xBFE, 0x1040, 0x61FC, 0x904, 0x11FC, 0x6104, 0x5FC, 0x904, 0x31FC, 0xC088, 0x504, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38}},
    {23300, 2, {0x2020, 0x23FE, 0x2020, 0x21FC, 0xF800, 0x4BFE, 0x4A02, 0x49FC, 0x4800, 0x91FC, 0x5104, 0x21FC, 0x3104, 0x4888, 0x4BFE, 0x8000}},
    {23301, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0xFDFC, 0x24A8, 0x24A8, 0x27FE, 0x24A8, 0x48A8, 0x29FC, 0x1020, 0x2BFE, 0x4420, 0x8420, 0x20}},
    {23302, 2, {0x2040, 0x20A0, 0x2118, 0x26E6, 0xF000, 0x53F8, 0x5208, 0x53F8, 0x5000, 0x97BC, 0x5084, 0x24A4, 0x2294, 0x54A4, 0x5294, 0x8108}},
    {23303, 2, {0x1020, 0x11FC, 0x1124, 0x11FC, 0xFC20, 0x27FE, 0x2400, 0x25FC, 0x2504, 0x49FC, 0x2904, 0x11FC, 0x2904, 0x45FC, 0x8488, 0x104}},
    {23304, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xFBFE, 0x4908, 0x4B9C, 0x4908, 0x4988, 0x963E, 0x5000, 0x23FE, 0x3090, 0x4890, 0x4912, 0x820E}},
    {23305, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC00, 0x25FC, 0x2504, 0x25FC, 0x2488, 0x4BFE, 0x2800, 0x11FC, 0x2904, 0x4504, 0x85FC, 0x104}},
    {23306, 2, {0x440, 0xFFFE, 0x440, 0x17D0, 0xF012, 0x17DC, 0x3452, 0xD7D2, 0x100E, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23307, 2, {0x2000, 0x23DE, 0x2252, 0x2252, 0xFBDE, 0x4800, 0x49FC, 0x4924, 0x49FC, 0x9124, 0x51FC, 0x2020, 0x33FE, 0x4820, 0x4820, 0x8020}},
    {23308, 2, {0x1018, 0x11E0, 0x1040, 0x13FE, 0xFC88, 0x2574, 0x2652, 0x2470, 0x2400, 0x49FC, 0x2904, 0x1174, 0x2954, 0x4574, 0x8504, 0x10C}},
    {23309, 2, {0x2108, 0x2548, 0x2548, 0x2550, 0xF7DE, 0x5014, 0x57E4, 0x5014, 0x57D4, 0x9114, 0x57D4, 0x2108, 0x2108, 0x5FD4, 0x5024, 0x8042}},
    {23310, 2, {0x2200, 0x2208, 0x23A8, 0x24A8, 0xF93E, 0x57C8, 0x5548, 0x555C, 0x5548, 0x97C8, 0x52BE, 0x2280, 0x2282, 0x5482, 0x547E, 0x8800}},
    {23311, 2, {0x203C, 0x23E0, 0x2124, 0x20A8, 0xFBFE, 0x48A8, 0x4924, 0x4A02, 0x49FC, 0x9124, 0x5124, 0x21FC, 0x3124, 0x4924, 0x49FC, 0x8104}},
    {23312, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x26FA, 0x2400, 0x2400, 0x25DC, 0x4954, 0x2954, 0x11DC, 0x2888, 0x4488, 0x8554, 0x222}},
    {23313, 2, {0x2040, 0x2020, 0x21FC, 0x2088, 0xF850, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x91FC, 0x5104, 0x21FC, 0x3020, 0x4A94, 0x4A8A, 0x847A}},
    {23314, 2, {0x1020, 0x1050, 0x1088, 0x1174, 0xFE02, 0x25FC, 0x2524, 0x25AC, 0x2524, 0x49FC, 0x2800, 0x10F8, 0x2888, 0x44F8, 0x8488, 0xF8}},
    {23315, 2, {0x2108, 0x2208, 0x27C8, 0x2450, 0xF7DE, 0x5464, 0x57D4, 0x5214, 0x5114, 0x97D4, 0x5214, 0x23C8, 0x2248, 0x5454, 0x5554, 0x88A2}},
    {23316, 2, {0x2200, 0x2208, 0x2388, 0x24A8, 0xF93E, 0x57C8, 0x5548, 0x557E, 0x5548, 0x97C8, 0x52BE, 0x2280, 0x22A2, 0x5492, 0x547E, 0x8800}},
    {23317, 2, {0x1000, 0x11FC, 0x1154, 0x11FC, 0xFC20, 0x25FC, 0x2420, 0x27FE, 0x2488, 0x4850, 0x29FC, 0x1020, 0x2BFE, 0x4420, 0x8420, 0x20}},
    {23318, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23319, 2, {0x2020, 0x23FE, 0x2000, 0x23FC, 0xFA04, 0x4AF4, 0x4A94, 0x4BFC, 0x4800, 0x91F8, 0x5108, 0x21F8, 0x3108, 0x49F8, 0x4800, 0x83FE}},
    {23320, 2, {0x2088, 0x2448, 0x2250, 0x20FC, 0xF820, 0x4850, 0x4E94, 0x4A38, 0x8A50, 0x4A98, 0x3234, 0x1254, 0x2A90, 0x4A20, 0x85FE, 0x800}},
    {23321, 2, {0x2020, 0x2020, 0x23FE, 0x2124, 0xF924, 0x4AAA, 0x4FFE, 0x4800, 0x49FC, 0x9104, 0x5174, 0x2154, 0x3174, 0x4904, 0x49FC, 0x8104}},
    {23322, 2, {0x2020, 0x23FE, 0x2288, 0x2250, 0xFBFE, 0x4A50, 0x4AFC, 0x4A54, 0x8BFE, 0x4A54, 0x32FC, 0x1250, 0x2AD8, 0x4D54, 0x8652, 0x850}},
    {23323, 2, {0x2000, 0x23FC, 0x2294, 0x2294, 0xFBFC, 0x4800, 0x4FFE, 0x4800, 0x4BFC, 0x9204, 0x53FC, 0x20A2, 0x3114, 0x4B08, 0x4D44, 0x8182}},
    {23324, 2, {0x1020, 0x7E20, 0x243E, 0xFF42, 0x94, 0x7E10, 0x4228, 0x7E28, 0x4244, 0x7E82, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23325, 2, {0x2080, 0x23F8, 0x2208, 0x23F8, 0xFA08, 0x4BF8, 0x4A00, 0x4BFC, 0x8A00, 0x4BFC, 0x3044, 0x13FC, 0x28A0, 0x4994, 0x86C8, 0x86}},
    {23326, 2, {0x2088, 0x23FE, 0x2088, 0x21FC, 0xF820, 0x4BFE, 0x4820, 0x49FC, 0x4924, 0x91FC, 0x5124, 0x21FC, 0x3020, 0x4BFE, 0x4820, 0x87FE}},
    {23327, 2, {0x2104, 0x2088, 0x23FE, 0x2020, 0xF9FC, 0x4820, 0x4BFE, 0x4854, 0x4992, 0x9090, 0x53FE, 0x2090, 0x30D4, 0x4B8A, 0x4896, 0x81A2}},
    {23328, 2, {0x10F8, 0x1088, 0x1088, 0x10F8, 0xFC00, 0x25DC, 0x2554, 0x2554, 0x25DC, 0x4820, 0x2BFE, 0x1070, 0x28A8, 0x4524, 0x8222, 0x20}},
    {23329, 2, {0x203C, 0x23C0, 0x2044, 0x2224, 0xF108, 0x57FE, 0x5442, 0x52A4, 0x528A, 0x947A, 0x5080, 0x20F8, 0x2108, 0x5290, 0x5060, 0x879E}},
    {23330, 2, {0x2080, 0x2040, 0x27FE, 0x2110, 0xFA08, 0x4D14, 0x4BF8, 0x4910, 0x4BF8, 0x9110, 0x57FC, 0x2128, 0x3310, 0x4D48, 0x4986, 0x8100}},
    {23331, 2, {0x2040, 0x2020, 0x23FE, 0x2222, 0xF954, 0x494A, 0x4A3A, 0x4800, 0x49FC, 0x9154, 0x53FE, 0x2000, 0x31FC, 0x4820, 0x48A0, 0x8040}},
    {23332, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0xF7DE, 0x5488, 0x55DC, 0x56AA, 0x5488, 0x9440, 0x5488, 0x25F0, 0x2420, 0x5448, 0x59FC, 0x9084}},
    {23333, 2, {0x2000, 0x23DE, 0x2042, 0x2252, 0xF94A, 0x4A52, 0x4890, 0x4908, 0x4BFE, 0x9510, 0x51FE, 0x2110, 0x31FE, 0x4910, 0x49FE, 0x8100}},
    {23334, 2, {0x2040, 0x27FC, 0x2040, 0x23FC, 0xF000, 0x57FE, 0x5002, 0x53F8, 0x5040, 0x97FE, 0x5000, 0x27FE, 0x2008, 0x57FE, 0x54A8, 0x8798}},
    {23335, 2, {0x2040, 0x23F8, 0x2048, 0x27FE, 0xF848, 0x4BF8, 0x4840, 0x4FFE, 0x4800, 0x92A4, 0x5452, 0x23F8, 0x32A8, 0x4AA8, 0x4FFE, 0x8000}},
    {23336, 2, {0x2108, 0x2090, 0x23FE, 0x2108, 0xF908, 0x4A52, 0x4B9C, 0x4908, 0x8A52, 0x4BDE, 0x3040, 0x10A4, 0x2AAA, 0x4A8A, 0x8478, 0x0}},
    {23337, 2, {0x2140, 0x224C, 0x2264, 0x2244, 0xFB4C, 0x4A64, 0x4AA4, 0x4BAC, 0x8AA4, 0x4AA4, 0x37FE, 0x1000, 0x2910, 0x4908, 0x8204, 0x404}},
    {23338, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xF9FC, 0x4948, 0x4A50, 0x49FC, 0x4B04, 0x91FC, 0x5104, 0x21FC, 0x3104, 0x49FC, 0x4888, 0x8104}},
    {23339, 2, {0x2088, 0x22AA, 0x22DC, 0x2488, 0xF954, 0x4A22, 0x4800, 0x4BFE, 0x8A22, 0x4820, 0x33FE, 0x1070, 0x28A8, 0x4924, 0x8622, 0x20}},
    {23340, 2, {0x2000, 0x21FC, 0x2020, 0x23FE, 0xFA22, 0x49AC, 0x4820, 0x49AC, 0x4800, 0x93FE, 0x5020, 0x21FC, 0x3154, 0x4954, 0x4954, 0x810C}},
    {23341, 2, {0x2000, 0x27FC, 0x2248, 0x2444, 0xFFFE, 0x5444, 0x56EC, 0x5554, 0x56EC, 0x9444, 0x56EC, 0x2554, 0x26EC, 0x5444, 0x5454, 0x8408}},
    {23342, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x2080, 0x3FFE, 0x2208, 0x2710, 0x40E0, 0x431C, 0x9C02}},
    {23343, 2, {0x2040, 0x2FFE, 0x2040, 0x27FC, 0xF000, 0x53F8, 0x5208, 0x5FFE, 0x5802, 0x97FC, 0x5110, 0x23F8, 0x2040, 0x57FC, 0x5040, 0x8FFE}},
    {23344, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2288, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {23345, 2, {0x420C, 0x2FB0, 0x8220, 0x4FA0, 0xABE, 0x2FA8, 0x4AA8, 0xCFA8, 0x4228, 0x5FA8, 0x4248, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23346, 2, {0x0, 0xF93E, 0xA92A, 0xA92A, 0xF93E, 0xAFAA, 0xAAAA, 0xFABE, 0x4290, 0x4490, 0xFABE, 0x4912, 0x4912, 0x4A92, 0xACAA, 0x1044}},
    {23347, 2, {0x1088, 0x13FE, 0x10A8, 0x1090, 0xFDFE, 0x2720, 0x25FC, 0x2520, 0x25FC, 0x4920, 0x29FE, 0x1100, 0x2BFC, 0x4488, 0x8070, 0x38E}},
    {23348, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0x7938, 0x4FA8, 0x7AA8, 0x4AB8, 0x792A, 0x4AAA, 0x9C46}},
    {23349, 2, {0x2108, 0x2110, 0x223E, 0x2222, 0xF4BE, 0x5722, 0x513E, 0x5208, 0x54BE, 0x97AA, 0x502A, 0x20AA, 0x256A, 0x552E, 0x5408, 0x8008}},
    {23350, 2, {0x1040, 0x10F8, 0x1088, 0x10F8, 0xFC88, 0x24F8, 0x2400, 0x25FC, 0x2524, 0x49FC, 0x2924, 0x11FC, 0x2800, 0x47FE, 0x8488, 0x108}},
    {23351, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0xF7DE, 0x5488, 0x55DC, 0x56AA, 0x5488, 0x9440, 0x5490, 0x2520, 0x2448, 0x5484, 0x55FC, 0x8884}},
    {23352, 2, {0x2020, 0x23FE, 0x2202, 0x241C, 0xFBE0, 0x4924, 0x48A8, 0x4BFE, 0x4924, 0x9202, 0x55FC, 0x2124, 0x31FC, 0x4924, 0x49FC, 0x8104}},
    {23353, 2, {0x2200, 0x24EC, 0x24A4, 0x24A4, 0xF6EC, 0x54A4, 0x54A4, 0x56EC, 0x54A4, 0x94A4, 0x5FFE, 0x2000, 0x2110, 0x5108, 0x5204, 0x8404}},
    {23354, 2, {0x23FE, 0x2248, 0x23FE, 0x2248, 0xFA20, 0x4BFE, 0x4A40, 0x4AFC, 0x8B44, 0x4A7C, 0x3200, 0x13FE, 0x2840, 0x4AA4, 0x828A, 0x47A}},
    {23355, 2, {0x2020, 0x23FE, 0x2020, 0x21FC, 0xF800, 0x4BFE, 0x4A52, 0x4BFE, 0x4904, 0x91FC, 0x5104, 0x21FC, 0x3104, 0x49FC, 0x4888, 0x8104}},
    {23356, 2, {0x2202, 0x25E2, 0x24A2, 0x26AA, 0xF52A, 0x566A, 0x510A, 0x528A, 0x544A, 0x9BAA, 0x510A, 0x27CA, 0x2122, 0x5542, 0x51EA, 0x8E04}},
    {23357, 2, {0x23FA, 0x22AA, 0x23FA, 0x2002, 0xF7FE, 0x5400, 0x55FC, 0x5554, 0x55FC, 0x9400, 0x57FE, 0x2040, 0x27FE, 0x50A0, 0x5118, 0x8606}},
    {23358, 2, {0x2108, 0x2108, 0x27CE, 0x2112, 0xF124, 0x57DE, 0x5552, 0x555E, 0x57D2, 0x911E, 0x5392, 0x255E, 0x2940, 0x5114, 0x5112, 0x8122}},
    {23359, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF1F0, 0x5208, 0x52EA, 0x56AC, 0x52A8, 0x92AA, 0x56EA, 0x2A06, 0x2000, 0x52A4, 0x5252, 0x8452}},
    {23360, 2, {0x21FC, 0x2020, 0x23FE, 0x2222, 0xF9AC, 0x4820, 0x49AC, 0x4800, 0x489E, 0x93F2, 0x509E, 0x21D2, 0x32BE, 0x4C92, 0x489E, 0x8092}},
    {23361, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3BB8, 0x2AA8, 0x3BB8, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {23362, 2, {0x2210, 0x23DE, 0x2528, 0x2084, 0xFBF8, 0x4A08, 0x4BF8, 0x4A08, 0x8BFA, 0x4A24, 0x3298, 0x1306, 0x2800, 0x4BFC, 0x8294, 0xFFE}},
    {23363, 2, {0x2040, 0x27FC, 0x2000, 0x23B8, 0xFAA8, 0x4BB8, 0x4910, 0x4FFC, 0x4910, 0x97FC, 0x5110, 0x27FE, 0x3128, 0x4B10, 0x4D48, 0x8186}},
    {23364, 2, {0x23DE, 0x2252, 0x23DE, 0x2252, 0xFBDE, 0x4A22, 0x4AFA, 0x4A22, 0x4AFA, 0x92AA, 0x52FA, 0x22AA, 0x32FA, 0x4A72, 0x4AAA, 0x8226}},
    {23365, 2, {0x2148, 0x214C, 0x22AA, 0x2008, 0xF7FE, 0x5148, 0x5368, 0x514A, 0x536A, 0x914C, 0x536C, 0x2148, 0x216A, 0x539A, 0x5026, 0x8042}},
    {23366, 2, {0x23DE, 0x2252, 0x23DE, 0x2252, 0xFBDE, 0x4A52, 0x4BDE, 0x498C, 0x4A52, 0x9040, 0x57FE, 0x2088, 0x3190, 0x4860, 0x48D8, 0x8304}},
    {23367, 2, {0x2108, 0x2294, 0x23DE, 0x26B4, 0xFBDE, 0x4A94, 0x4BDE, 0x4A94, 0x4BDE, 0x9210, 0x57FC, 0x2108, 0x3090, 0x4860, 0x4998, 0x8606}},
    {23368, 2, {0x2222, 0x23FE, 0x2090, 0x21FE, 0xFB10, 0x4DFE, 0x4910, 0x49FE, 0x4910, 0x91FE, 0x5100, 0x23FE, 0x328A, 0x4B76, 0x4A52, 0x8276}},
    {23369, 2, {0x2088, 0x23FE, 0x2088, 0x23DE, 0xFA52, 0x4BDE, 0x48A0, 0x4890, 0x49FE, 0x9120, 0x53FC, 0x2520, 0x31FC, 0x4920, 0x49FE, 0x8100}},
    {23370, 2, {0x2020, 0x23FE, 0x2248, 0x23FE, 0xFA48, 0x4AEC, 0x4B5A, 0x4A48, 0x8A50, 0x4BDE, 0x3250, 0x13DC, 0x2A50, 0x4DDE, 0x8450, 0x850}},
    {23371, 2, {0x23DE, 0x2000, 0x23DE, 0x2252, 0xFB5A, 0x4A52, 0x4820, 0x4BFE, 0x4A50, 0x93FE, 0x5252, 0x23FE, 0x3292, 0x4ADC, 0x4A92, 0x84CE}},
    {23372, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {23373, 2, {0x27BC, 0x24A4, 0x27BC, 0x2000, 0xF7FE, 0x5400, 0x5590, 0x549E, 0x57D4, 0x9564, 0x55D4, 0x2554, 0x25D4, 0x5568, 0x5BC8, 0x9054}},
    {23374, 2, {0x23FE, 0x2202, 0x23FE, 0x2292, 0xFA54, 0x4A92, 0x4AFE, 0x4AAA, 0x4AFE, 0x9322, 0x52FA, 0x22AA, 0x32FA, 0x4A22, 0x4DFA, 0x8004}},
    {23375, 2, {0x2088, 0x23FE, 0x2088, 0x23DE, 0xFA52, 0x4BDE, 0x4A52, 0x4BDE, 0x4A22, 0x93FE, 0x52AA, 0x22FA, 0x3222, 0x4A72, 0x4AAA, 0x8224}},
    {23376, 2, {0x0, 0x7FF8, 0x10, 0x20, 0x40, 0x180, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23377, 2, {0x0, 0x7FF8, 0x10, 0x20, 0x40, 0x184, 0x118, 0x160, 0x380, 0x1D00, 0xE100, 0x4100, 0x100, 0x100, 0x500, 0x200}},
    {23378, 2, {0x0, 0x7FF8, 0x10, 0x20, 0x40, 0x180, 0x100, 0x1FE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23379, 2, {0x0, 0x7FF8, 0x10, 0x20, 0x2040, 0x1180, 0x900, 0x500, 0x300, 0x100, 0x180, 0x140, 0x130, 0x10C, 0x500, 0x200}},
    {23380, 2, {0x40, 0x7E40, 0x240, 0x440, 0x840, 0x840, 0xA40, 0xC40, 0x3840, 0xC840, 0x840, 0x842, 0x842, 0x842, 0x283E, 0x1000}},
    {23381, 2, {0x0, 0x3FF0, 0x810, 0x820, 0x83C, 0x1004, 0x17C4, 0x2054, 0x4088, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23382, 2, {0x0, 0x7EFC, 0x204, 0x408, 0x810, 0x820, 0xA20, 0xDFE, 0x3820, 0xC820, 0x820, 0x820, 0x820, 0x820, 0x28A0, 0x1040}},
    {23383, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FE0, 0x40, 0x80, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23384, 2, {0x400, 0x400, 0xFFFE, 0x800, 0x800, 0x13F8, 0x1010, 0x3020, 0x5040, 0x97FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080}},
    {23385, 2, {0x20, 0x7E20, 0x220, 0x420, 0x820, 0x8A8, 0xAA4, 0xCA4, 0x3922, 0xC922, 0xA22, 0x820, 0x820, 0x820, 0x28A0, 0x1040}},
    {23386, 2, {0xF8, 0x7F00, 0x2210, 0x1110, 0x1120, 0x0, 0x1FE0, 0x40, 0x80, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23387, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0xFE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {23388, 2, {0x40, 0xFC40, 0x440, 0x880, 0x10FE, 0x1108, 0x1688, 0x1888, 0x3088, 0xD050, 0x1050, 0x1020, 0x1050, 0x1088, 0x5104, 0x2202}},
    {23389, 2, {0x200, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x100, 0x200, 0x3FE0, 0x1040, 0x2080, 0x5FFC, 0x8080, 0x80, 0x280, 0x100}},
    {23390, 2, {0x0, 0x1FE0, 0x40, 0x80, 0x100, 0xFFFE, 0x100, 0x100, 0x500, 0x200, 0x910, 0x4908, 0x4824, 0x4824, 0x87E0, 0x0}},
    {23391, 2, {0x0, 0x1FE0, 0x40, 0x180, 0xFFFE, 0x100, 0x100, 0x500, 0x200, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {23392, 2, {0x0, 0x3FF8, 0x8, 0x7FC8, 0x8, 0x3F88, 0x2088, 0x3FA8, 0x10, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23393, 2, {0x20, 0x7E20, 0x220, 0x440, 0x848, 0x884, 0xBFE, 0xC82, 0x3800, 0xC8FC, 0x884, 0x884, 0x884, 0x884, 0x28FC, 0x1084}},
    {23394, 2, {0x80, 0xFC80, 0x5FC, 0x904, 0x1204, 0x15F4, 0x1514, 0x1914, 0x3114, 0xD1F4, 0x1104, 0x1128, 0x1112, 0x1102, 0x50FE, 0x2000}},
    {23395, 2, {0x70, 0x1F80, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xCFE6, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {23396, 2, {0x8, 0xF83C, 0xBD0, 0x1290, 0x2290, 0x2290, 0x2A90, 0x3290, 0x2290, 0xE288, 0x2288, 0x22C8, 0x22A4, 0x24D4, 0xA492, 0x4800}},
    {23397, 2, {0x1000, 0x1000, 0xFEFC, 0x2244, 0x6428, 0x1810, 0x2428, 0xC2C6, 0x0, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23398, 2, {0x2208, 0x1108, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8004, 0x1FE0, 0x40, 0x180, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23399, 2, {0x820, 0x1020, 0x22FC, 0x7C24, 0x824, 0x1244, 0x7F54, 0x88, 0x100, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23400, 2, {0x0, 0x1FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x100, 0x500, 0x7A7C, 0x808, 0x1010, 0x1EFE, 0xF010, 0x1010, 0x5050, 0x2020}},
    {23401, 2, {0x20, 0xFC10, 0x410, 0xBFE, 0x1020, 0x1042, 0x1484, 0x19F8, 0x3010, 0xD022, 0x10C4, 0x1308, 0x1010, 0x1028, 0x50C4, 0x2302}},
    {23402, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x1448, 0x2444, 0x4444, 0x0, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23403, 2, {0xC, 0xFDF0, 0x420, 0x844, 0x1088, 0x11F0, 0x1420, 0x1844, 0x31FE, 0xD022, 0x1020, 0x10A8, 0x1124, 0x1222, 0x50A0, 0x2040}},
    {23404, 2, {0x7FFC, 0x80, 0x360, 0xD18, 0x7104, 0x100, 0x1000, 0x10FC, 0xFE04, 0x2208, 0x2210, 0x64FE, 0x1410, 0x810, 0x3450, 0xC220}},
    {23405, 2, {0x0, 0xFDFC, 0x504, 0x904, 0x11FC, 0x1104, 0x1504, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1000, 0x1088, 0x1084, 0x5102, 0x2202}},
    {23406, 2, {0x40, 0xFC20, 0x7FE, 0xA02, 0x1000, 0x11FC, 0x1400, 0x1800, 0x33FE, 0xD020, 0x1128, 0x1124, 0x1222, 0x1422, 0x50A0, 0x2040}},
    {23407, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x0, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23408, 2, {0x1040, 0x840, 0xFF40, 0x40, 0x7EF8, 0x4248, 0x7E48, 0x148, 0x7EC8, 0x448, 0x868, 0xE5A, 0xF88A, 0x88A, 0x2906, 0x1202}},
    {23409, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x23F0, 0x2020, 0x3FFE, 0x2040, 0x20C0, 0x2FBE, 0x2104, 0x2388, 0x5E7E, 0x4208, 0x8A28, 0x410}},
    {23410, 2, {0x0, 0xFDFE, 0x448, 0x848, 0x1048, 0x11CE, 0x1502, 0x1902, 0x3102, 0xD1CE, 0x1048, 0x1048, 0x1048, 0x1048, 0x53FE, 0x2000}},
    {23411, 2, {0x820, 0x440, 0x7FFC, 0x1248, 0x3CF0, 0x820, 0x1450, 0x3EF8, 0x0, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23412, 2, {0x1FE0, 0x40, 0xFFFE, 0x100, 0x300, 0x7C7C, 0x808, 0x1EFE, 0xF010, 0x1050, 0x3020, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {23413, 2, {0x200E, 0xCEF0, 0x8A22, 0xAA92, 0xAA54, 0xEE00, 0xAA7C, 0xAA08, 0xAA10, 0xAA10, 0xEEFE, 0xA810, 0x2810, 0x4810, 0x4850, 0x8820}},
    {23414, 2, {0x820, 0xFFFE, 0x820, 0x1248, 0x3CF0, 0x820, 0x1450, 0x3EF8, 0x0, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23415, 2, {0x820, 0x3E7E, 0x8A4, 0x7F28, 0x1C10, 0x2A28, 0x4944, 0x3FFE, 0x2000, 0x23F0, 0x2020, 0x2040, 0x2FFE, 0x4040, 0x4140, 0x8080}},
    {23416, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0xBFE4, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23417, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x0, 0x3FE0, 0x80, 0x100, 0xFFFE, 0x100, 0x500, 0x200}},
    {23418, 2, {0x0, 0xFDFC, 0x420, 0xBFE, 0x1222, 0x11AC, 0x1420, 0x19AC, 0x3000, 0xD3FE, 0x1020, 0x11FC, 0x1154, 0x1154, 0x5154, 0x210C}},
    {23419, 2, {0x20, 0xFDFC, 0x424, 0xBFE, 0x1024, 0x11FC, 0x1420, 0x1BFE, 0x3000, 0xD2A4, 0x1452, 0x11FC, 0x1154, 0x1154, 0x57FE, 0x2000}},
    {23420, 2, {0x5210, 0x7EFE, 0x2844, 0x7E28, 0xA2FE, 0x3E10, 0x2010, 0x3E7C, 0x2210, 0x3E10, 0x0, 0x1FE0, 0x40, 0xFFFE, 0x100, 0x300}},
    {23421, 2, {0x820, 0xFFFE, 0x2820, 0x7DFC, 0x4488, 0x7C50, 0x41FE, 0x7C20, 0x45FC, 0x7C20, 0x0, 0x1FE0, 0x40, 0xFFFE, 0x100, 0x300}},
    {23422, 2, {0x3DE, 0xFA52, 0xBDE, 0x1252, 0x23DE, 0x2252, 0x2BDE, 0x318C, 0x2252, 0xE040, 0x27FE, 0x2088, 0x2190, 0x2060, 0xA0D8, 0x4304}},
    {23423, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x0, 0x1FE0, 0x40, 0xFFFE, 0x100, 0x300}},
    {23424, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {23425, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23426, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0xFE0, 0x820, 0x820, 0x820, 0x1020, 0x1022, 0x2022, 0x401E, 0x8000}},
    {23427, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x800, 0x808, 0x830, 0x9C0, 0xE00, 0x800, 0x804, 0x804, 0x804, 0x7FC, 0x0}},
    {23428, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8204, 0x200, 0x1FE0, 0x220, 0x220, 0x420, 0x420, 0x822, 0x1022, 0x201E, 0x4000}},
    {23429, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x20, 0x3F0, 0x3E00, 0x200, 0x3FC, 0xFE00, 0x204, 0x204, 0x204, 0x1FC, 0x0}},
    {23430, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x2000, 0x2000, 0x3FFC, 0x4, 0x4, 0x28, 0x10}},
    {23431, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {23432, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8044, 0x40, 0x40, 0xFFFE, 0x40, 0x40, 0x1040, 0x840, 0x840, 0x40, 0x140, 0x80}},
    {23433, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x4208, 0x200, 0x200, 0xFFFE, 0x420, 0x820, 0x1840, 0x640, 0x180, 0x260, 0xC10, 0x7008}},
    {23434, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8104, 0x120, 0x110, 0x7FFC, 0x100, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {23435, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {23436, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {23437, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x100, 0x80, 0x0, 0xFFFE, 0x0, 0x440, 0x420, 0x810, 0x1008, 0x2004, 0x4004}},
    {23438, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0xF0, 0x3F00, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {23439, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x200, 0x7FFC, 0x400, 0x480, 0x880, 0x900, 0x1100, 0x2220, 0x4410, 0x8FF8, 0x408}},
    {23440, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8404, 0x400, 0x7F0, 0x810, 0x810, 0x1420, 0x2220, 0x140, 0x40, 0x80, 0xFFFE, 0x0}},
    {23441, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x820, 0x3FF8, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {23442, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0xF8, 0x3F00, 0x100, 0x1F8, 0x3F00, 0x100, 0x1FC, 0x7F02, 0x102, 0x102, 0xFE}},
    {23443, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8104, 0x90, 0x890, 0x4824, 0x4842, 0x4882, 0x8902, 0xE10, 0x1810, 0xE7F0, 0x0}},
    {23444, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {23445, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x7FFC, 0x400, 0x400, 0x800, 0x1FF8, 0x2808, 0xC808, 0x808, 0xFF8, 0x808}},
    {23446, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x7E10, 0x210, 0x210, 0x3E10, 0x2020, 0x4028, 0x7E24, 0x244, 0x2FE, 0x1442, 0x800}},
    {23447, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {23448, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1000, 0x1000, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {23449, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {23450, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x3FF8, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {23451, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9004, 0x1EF8, 0x1288, 0x2288, 0x2288, 0x52A8, 0x8C90, 0x482, 0x882, 0x1082, 0x207E, 0x4000}},
    {23452, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {23453, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x120, 0x110, 0xFFFE, 0x0}},
    {23454, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8884, 0x480, 0x480, 0x1080, 0x880, 0x880, 0xFFFE, 0x140, 0x220, 0x410, 0x1808, 0x6004}},
    {23455, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {23456, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8424, 0x410, 0x410, 0xFFFE, 0x480, 0x888, 0x890, 0x10A0, 0x10C2, 0x2082, 0x4182, 0x867E}},
    {23457, 2, {0x200, 0x100, 0x7FFE, 0x4102, 0x8104, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100}},
    {23458, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8804, 0xFF0, 0x1020, 0x2C40, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {23459, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x0, 0xFFFE, 0x0}},
    {23460, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x400, 0x820, 0x1FF0, 0x110, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {23461, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8204, 0x7FF8, 0x400, 0xFF0, 0x1810, 0x2FF0, 0x4810, 0x8FF0, 0x810, 0x810, 0x850, 0x820}},
    {23462, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x2100, 0x2100, 0x3FF0, 0x2010, 0x2010, 0x2010, 0x3FF0, 0x2100, 0x2100, 0x3FFC}},
    {23463, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FFC, 0x2080, 0x2080, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x2080, 0x2080, 0x3FFE, 0x0}},
    {23464, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x440, 0x4448, 0x2450, 0x2460, 0xC40, 0x3460, 0xC852, 0x84A, 0x1042, 0x203E, 0x4000}},
    {23465, 2, {0x200, 0x100, 0x7FFE, 0x4102, 0x9114, 0x920, 0x540, 0x100, 0x7FFC, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {23466, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x440, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {23467, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {23468, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8094, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3E88, 0x2250, 0x2250, 0x2A24, 0x4454, 0x418C, 0x8604}},
    {23469, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x408, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {23470, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {23471, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x810C, 0x1FD0, 0x120, 0xFFFE, 0x100, 0x1FF0, 0x820, 0x30C0, 0xCFFC, 0x80, 0x280, 0x100}},
    {23472, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8104, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {23473, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8884, 0x880, 0x10FE, 0x1140, 0x3240, 0x547C, 0x9040, 0x1040, 0x107E, 0x1040, 0x1040, 0x1040}},
    {23474, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {23475, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {23476, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x3008}},
    {23477, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9114, 0x920, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {23478, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x7FFC, 0x200, 0xD08, 0x7190, 0x2A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {23479, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x80F4, 0x3F00, 0x1110, 0x920, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {23480, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x2000, 0x2FF0, 0x2000, 0x3FFC, 0x2488, 0x2450, 0x4420, 0x4518, 0x8606, 0x400}},
    {23481, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1110, 0x2288, 0x440, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {23482, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x920, 0x920, 0x1120, 0x1122, 0x2122, 0x411E}},
    {23483, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xBFF4, 0x1010, 0x4824, 0x2448, 0x380, 0x1C70, 0xE00E, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23484, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8024, 0x3E20, 0x1FE, 0x44, 0x7E44, 0x14C8, 0x1428, 0x1410, 0x146A, 0x2586, 0x2402, 0x43FE}},
    {23485, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8444, 0x3FF8, 0x440, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1290, 0x482, 0x1882, 0xE07E}},
    {23486, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8074, 0x1F80, 0x1000, 0x1000, 0x1FF8, 0x1080, 0x1080, 0xFFFE, 0x0, 0x840, 0x1020, 0x2010}},
    {23487, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8804, 0xFFC, 0x1040, 0x3080, 0x53F8, 0x9208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {23488, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8014, 0xF8, 0x3F00, 0x1110, 0x820, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {23489, 2, {0x100, 0x7FFE, 0x4002, 0x8104, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x1100, 0x11F8, 0x2900, 0x47FE, 0x8000}},
    {23490, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8804, 0xF7C, 0x844, 0x844, 0xFF44, 0x28, 0x828, 0x2A10, 0x4910, 0x8928, 0x2844, 0x1082}},
    {23491, 2, {0x200, 0x100, 0x7FFE, 0x4802, 0x8FE4, 0x1020, 0x2040, 0x5FF8, 0x1108, 0x1108, 0x1FF8, 0x280, 0x4A0, 0x892, 0x1082, 0x607E}},
    {23492, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8204, 0x3FF8, 0x480, 0x840, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x50, 0x20}},
    {23493, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x440, 0x820, 0x1010}},
    {23494, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8224, 0x940, 0x2888, 0x4B14, 0x1C10, 0xE7F0, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23495, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8024, 0x3E38, 0x20, 0x20, 0x7EF8, 0x1408, 0x1450, 0x1420, 0x1454, 0x2484, 0x2404, 0x43FC}},
    {23496, 2, {0x100, 0x7FFE, 0x4102, 0xBFFC, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {23497, 2, {0x200, 0x100, 0x7FFE, 0x4902, 0x8884, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000}},
    {23498, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x83F4, 0x200, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {23499, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0xBFFC, 0x820, 0x3FF8, 0x820, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0x8A0, 0x848, 0x808, 0x7F8}},
    {23500, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xBFFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {23501, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8204, 0x908, 0x4924, 0x4824, 0x87E0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {23502, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8404, 0x25FC, 0x2420, 0x3C20, 0x5FC, 0x524, 0x7D24, 0x2554, 0x2594, 0x2504, 0x4514, 0x8508}},
    {23503, 2, {0x200, 0x100, 0x7FFE, 0x4802, 0x8FE4, 0x1040, 0x3FF8, 0x5248, 0x1428, 0x1108, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {23504, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8424, 0x2420, 0x25FC, 0x3C20, 0x420, 0x7FE, 0x7C60, 0x24B0, 0x2528, 0x2626, 0x4420, 0x8420}},
    {23505, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8BFC, 0x808, 0x11F8, 0x3008, 0x53F8, 0x9000, 0x17FC, 0x1404, 0x13F8, 0x1110, 0x10E0, 0x171C}},
    {23506, 2, {0x200, 0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1310, 0x2088, 0xC046, 0x600, 0x180, 0x40}},
    {23507, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2148, 0x2FE8, 0x2428, 0x2010}},
    {23508, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1100, 0x11F8, 0x2900, 0x47FE, 0x8000}},
    {23509, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xBFFC, 0x200, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {23510, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xA3FC, 0x1008, 0x11F8, 0x8008, 0x43F8, 0x5000, 0x17FC, 0x2404, 0xE3F8, 0x2110, 0x20E0, 0x271C}},
    {23511, 2, {0x100, 0x7FFE, 0x4202, 0x8144, 0x880, 0x2928, 0x47E4, 0x1800, 0x6100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x2008, 0x2018}},
    {23512, 2, {0x100, 0x7FFE, 0x4102, 0xBFFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x1010}},
    {23513, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8004, 0x61C, 0x38E8, 0x28A8, 0x2CA8, 0x2AA8, 0x2AA8, 0x2AA8, 0x28A4, 0x5554, 0x5D74, 0x8212}},
    {23514, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xBFFC, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x1000, 0x1FF0, 0x2100, 0xFFFE, 0x100, 0x2108, 0x3FF8}},
    {23515, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0xBFFC, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {23516, 2, {0x200, 0x100, 0x7FFE, 0x4102, 0x9494, 0x1428, 0x23E8, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {23517, 2, {0x100, 0x7FFE, 0x4002, 0x8004, 0x13F8, 0x1008, 0x91F8, 0x5008, 0x53F8, 0x1000, 0x37FC, 0x5404, 0x93F8, 0x1110, 0x10E0, 0x171C}},
    {23518, 2, {0x100, 0x7FFE, 0x4002, 0x8444, 0x3FF8, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x440, 0x1830, 0xE00E}},
    {23519, 2, {0x100, 0x7FFE, 0x4002, 0x9084, 0x1EF8, 0x2288, 0x5450, 0x2820, 0x17D8, 0x2006, 0xDFF0, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {23520, 2, {0x200, 0x100, 0x7FFE, 0x4102, 0xBFFC, 0x2108, 0xFFFE, 0x2108, 0x3FF8, 0x1110, 0x1FF0, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {23521, 2, {0x100, 0x7FFE, 0x4002, 0xBFF4, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x37D8, 0xC246, 0xCC0}},
    {23522, 2, {0x100, 0x7FFE, 0x4002, 0x8404, 0x24FC, 0x2404, 0x247C, 0x3C04, 0x4FC, 0x400, 0x7DFE, 0x2502, 0x24FC, 0x2448, 0x4430, 0x85CE}},
    {23523, 2, {0x200, 0x100, 0x7FFE, 0x4042, 0x8424, 0x25FC, 0x2400, 0x3CF8, 0x400, 0x4F8, 0x7C00, 0x24F8, 0x2488, 0x2488, 0x44F8, 0x8488}},
    {23524, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x85FC, 0x2420, 0x24F8, 0x3C48, 0x448, 0x5FE, 0x7C00, 0x24FC, 0x2484, 0x2484, 0x44FC, 0x8484}},
    {23525, 2, {0x100, 0x7FFE, 0x4002, 0xBCFC, 0x2448, 0x1428, 0x2548, 0x6C0, 0x1830, 0xE30E, 0xC40, 0x3180, 0x630, 0x38C0, 0x700, 0x3800}},
    {23526, 2, {0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1110, 0xFFFE, 0x2210, 0x3FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {23527, 2, {0x100, 0x7FFE, 0x4102, 0x9494, 0x14A8, 0x23E8, 0x0, 0x3FF8, 0x2448, 0xFFFE, 0x0, 0x3FF8, 0x100, 0x100, 0x500, 0x200}},
    {23528, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1110, 0x2FE8, 0xC106, 0x920, 0x1110, 0x2508, 0x200}},
    {23529, 2, {0x100, 0x7FFE, 0x4002, 0x80FC, 0x3F00, 0x1110, 0x920, 0xFFFE, 0x920, 0x3018, 0xDFF6, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1010}},
    {23530, 2, {0x100, 0x7FFE, 0x4002, 0x80FC, 0x7F00, 0x2210, 0x1120, 0x1FE0, 0x1020, 0x1FF0, 0x1010, 0x1FFC, 0x2004, 0x2AA4, 0x4AA4, 0x800C}},
    {23531, 2, {0x100, 0x7FFE, 0x4002, 0x8204, 0x1CF0, 0x1010, 0x1EF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2004, 0x4924, 0x2494, 0x2494, 0x4008}},
    {23532, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0xBFFC, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x452, 0x184A, 0xE03E}},
    {23533, 2, {0x100, 0x7FFE, 0x4102, 0xBFFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x204, 0x4908, 0x4824, 0x87E4}},
    {23534, 2, {0x100, 0x7FFE, 0x4002, 0x8104, 0x7FFC, 0x440, 0x2828, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x100, 0x1110, 0x2508, 0x200}},
    {23535, 2, {0x100, 0x7FFE, 0x4902, 0x8884, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3E7C, 0x2244, 0x23D4, 0x2008}},
    {23536, 2, {0x100, 0x7FFE, 0x4002, 0xBFFC, 0x2448, 0x3FF8, 0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x508, 0x1890, 0xEA60, 0xC1C}},
    {23537, 2, {0x200, 0x100, 0x7FFE, 0x4042, 0x89FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x904, 0x9FC, 0xF820, 0x4BFE, 0x48A8, 0x4924, 0x8A22}},
    {23538, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0xA47C, 0x3A04, 0x2228, 0x1E10, 0x20FE, 0x3E12, 0x4850, 0x85E, 0x7F50, 0x14B0, 0x229E, 0x4100}},
    {23539, 2, {0x100, 0x7FFE, 0x4082, 0xBEFC, 0x924, 0x3EA8, 0x924, 0x7EA0, 0x40, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {23540, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8804, 0x7F7C, 0x2244, 0x147C, 0xFF44, 0x87C, 0x7F44, 0x87C, 0x2A28, 0x492A, 0xA94A, 0x1086}},
    {23541, 2, {0x100, 0xFFFE, 0x8842, 0x7F7C, 0x2240, 0x147C, 0xFF04, 0x7C, 0x7F40, 0x417C, 0x7F40, 0x417C, 0x7F40, 0x417C, 0x4542, 0x423E}},
    {23542, 2, {0x100, 0x7FFE, 0x4082, 0xBEFC, 0x920, 0x3EFC, 0x820, 0x7EA4, 0xFC, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {23543, 2, {0x100, 0x7FFE, 0x4822, 0x9D74, 0x4924, 0x5D74, 0x4924, 0x5D74, 0x4924, 0x7FFC, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x820, 0xFFFE}},
    {23544, 2, {0x40, 0x40, 0x40, 0x40, 0xFFFE, 0x40, 0x40, 0x40, 0x1040, 0x840, 0x840, 0x40, 0x40, 0x40, 0x140, 0x80}},
    {23545, 2, {0x10, 0x10, 0x10, 0x7E10, 0x2FE, 0x210, 0x2410, 0x1490, 0x850, 0x850, 0x1410, 0x1210, 0x2210, 0x4010, 0x50, 0x20}},
    {23546, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x20, 0x20, 0x7FFC, 0x820, 0x420, 0x420, 0x20, 0xA0, 0x40}},
    {23547, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x3FF8, 0x20, 0x20, 0xFFFE, 0x820, 0x420, 0x420, 0xA0, 0x40}},
    {23548, 2, {0x0, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2004, 0x2004, 0x1FFC, 0x20, 0x20, 0xFFFE, 0x820, 0x420, 0x420, 0xA0, 0x40}},
    {23549, 2, {0x8, 0x1FC, 0x7E00, 0x208, 0x2108, 0x1110, 0x1040, 0x40, 0xFFFE, 0x40, 0x1040, 0x840, 0x840, 0x40, 0x140, 0x80}},
    {23550, 2, {0x2010, 0x1010, 0x1010, 0x10, 0xFEFE, 0x410, 0x410, 0x4890, 0x2850, 0x1050, 0x1010, 0x2810, 0x2410, 0x4410, 0x8050, 0x20}},
    {23551, 2, {0x200, 0x200, 0x7FFC, 0x200, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x820, 0x1FFC, 0x1020, 0x2420, 0x4220, 0x8220, 0xA0, 0x40}},
    {23552, 2, {0x8, 0x8, 0x7F88, 0x4008, 0x407E, 0x5F08, 0x5108, 0x5148, 0x5128, 0x5128, 0x5F08, 0x5108, 0x4008, 0x7F88, 0x28, 0x10}},
    {23553, 2, {0x808, 0x808, 0x808, 0x7F08, 0x87E, 0x808, 0xFF08, 0x48, 0x828, 0x828, 0x7F08, 0x808, 0x808, 0xF08, 0xF028, 0x4010}},
    {23554, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x20, 0xFFFE, 0x820, 0x420, 0x420, 0xA0, 0x40}},
    {23555, 2, {0x120, 0x110, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {23556, 2, {0x808, 0x1008, 0x3E08, 0x2208, 0x3EFE, 0x2208, 0x3E08, 0x2208, 0x2248, 0xFE28, 0x628, 0xA08, 0x1208, 0x2208, 0x4A28, 0x410}},
    {23557, 2, {0x808, 0x808, 0xFF08, 0x808, 0x8FE, 0x7F08, 0x4148, 0x4128, 0x7F28, 0x1408, 0x1408, 0x1428, 0x2412, 0x2402, 0x43FE, 0x8000}},
    {23558, 2, {0x880, 0x8F8, 0x908, 0x4A10, 0x28A0, 0x2840, 0x890, 0xB10, 0x1810, 0x2BFE, 0xC810, 0x910, 0x890, 0x810, 0x850, 0x820}},
    {23559, 2, {0x440, 0x247C, 0x24C4, 0x2528, 0x2490, 0x3CA0, 0x448, 0x588, 0xFC08, 0x25FE, 0x2408, 0x2488, 0x2448, 0x4408, 0x8428, 0x410}},
    {23560, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0xFFFE, 0x22, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {23561, 2, {0x4, 0x3F84, 0x2084, 0x2084, 0x3FBE, 0x2004, 0x2F84, 0x2024, 0x2014, 0x3FD4, 0x2204, 0x2A84, 0x4A44, 0x5244, 0x8A14, 0x408}},
    {23562, 2, {0x820, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {23563, 2, {0x0, 0x3FF0, 0x10, 0x1FF0, 0x10, 0x3FF0, 0x0, 0x3E7C, 0x844, 0x7F7C, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {23564, 2, {0x808, 0x808, 0xFF88, 0x808, 0x87E, 0x7F08, 0x8, 0x7F48, 0x4128, 0x4128, 0x7F08, 0x8, 0x2208, 0x1788, 0xF828, 0x4010}},
    {23565, 2, {0x1408, 0x1408, 0x5508, 0x3608, 0x147E, 0xFF08, 0x8, 0x2248, 0x1428, 0x7F28, 0x808, 0x3E08, 0x808, 0xF08, 0x7828, 0x2010}},
    {23566, 2, {0x2110, 0x17FE, 0x1080, 0x3F8, 0xF208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE, 0x20, 0xFFFE, 0x820, 0x460}},
    {23567, 2, {0x100, 0x100, 0x100, 0x100, 0x100, 0x1110, 0x1108, 0x1104, 0x2104, 0x2102, 0x4102, 0x8102, 0x100, 0x100, 0x500, 0x200}},
    {23568, 2, {0x100, 0x100, 0x100, 0x920, 0x910, 0x1108, 0x1104, 0x2104, 0x4100, 0x100, 0x0, 0x1800, 0x2400, 0x380, 0x7E, 0x0}},
    {23569, 2, {0x100, 0x100, 0x100, 0x920, 0x910, 0x1108, 0x1104, 0x2114, 0x4110, 0x120, 0x40, 0x80, 0x100, 0x200, 0xC00, 0x7000}},
    {23570, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x100, 0x500, 0x200}},
    {23571, 2, {0x800, 0x800, 0x800, 0x1FFC, 0x1000, 0x2100, 0x4100, 0x8100, 0x100, 0x1110, 0x1108, 0x2104, 0x4102, 0x8102, 0x500, 0x200}},
    {23572, 2, {0x800, 0x800, 0x800, 0x1FFC, 0x1004, 0x2108, 0x4100, 0x8100, 0x100, 0x1110, 0x1108, 0x2104, 0x4102, 0x8102, 0x500, 0x200}},
    {23573, 2, {0x0, 0x3FF0, 0x810, 0x820, 0x83C, 0x1004, 0x1004, 0x2114, 0x4108, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x500, 0x200}},
    {23574, 2, {0x100, 0x100, 0x920, 0x910, 0x1108, 0x2108, 0x0, 0x104, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x830, 0x300E, 0xC004}},
    {23575, 2, {0x200, 0x200, 0x3F8, 0x200, 0x200, 0x200, 0xFFFE, 0x100, 0x100, 0x920, 0x1110, 0x2108, 0x4104, 0x100, 0x500, 0x200}},
    {23576, 2, {0x100, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x0, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {23577, 2, {0x100, 0x1110, 0x1108, 0x2104, 0x4104, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {23578, 2, {0x100, 0x2108, 0x1110, 0x920, 0x100, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {23579, 2, {0x100, 0x100, 0x920, 0x910, 0x1108, 0x2104, 0x4104, 0x1010, 0x1010, 0x1010, 0x5454, 0x5252, 0x9292, 0x1010, 0x5050, 0x2020}},
    {23580, 2, {0x100, 0x920, 0x1110, 0x2108, 0x0, 0x100, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC106, 0x920, 0x910, 0x1108, 0x2508, 0x200}},
    {23581, 2, {0x100, 0x1110, 0x920, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {23582, 2, {0x100, 0x100, 0x7FFC, 0x2288, 0x1450, 0x820, 0x3FF8, 0xD016, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {23583, 2, {0x10, 0x3E10, 0x2210, 0x3E54, 0x2252, 0x3E52, 0x90, 0xFF12, 0x804, 0x808, 0x2F10, 0x2860, 0x2980, 0x5800, 0x4FFE, 0x8000}},
    {23584, 2, {0x2210, 0x2210, 0xFF10, 0x2210, 0x3E54, 0x2252, 0x3E52, 0x2290, 0x2210, 0xFF14, 0x4004, 0x5408, 0x6208, 0x4010, 0x7E20, 0xC0}},
    {23585, 2, {0x1000, 0x11FC, 0x9104, 0x5504, 0x59FC, 0x1104, 0xFD04, 0x29FC, 0x2800, 0x2912, 0x29D4, 0x2918, 0x2B10, 0x4D52, 0x4992, 0x810E}},
    {23586, 2, {0x400, 0x400, 0x400, 0x400, 0xFFFE, 0x480, 0x480, 0x480, 0x480, 0x880, 0x880, 0x1082, 0x1082, 0x2082, 0x407E, 0x8000}},
    {23587, 2, {0x440, 0x420, 0x810, 0x1008, 0x2004, 0x4442, 0x8442, 0x440, 0x440, 0x440, 0x440, 0x842, 0x842, 0x1042, 0x203E, 0x4000}},
    {23588, 2, {0x420, 0x410, 0x410, 0x400, 0xFFFE, 0x480, 0x480, 0x480, 0x480, 0x880, 0x880, 0x1082, 0x1082, 0x2082, 0x407E, 0x8000}},
    {23589, 2, {0x2080, 0x2080, 0x2080, 0x21F8, 0xFD08, 0x2A08, 0x2888, 0x2848, 0x2848, 0x2808, 0x2808, 0x2850, 0x2822, 0x4802, 0x47FE, 0x8000}},
    {23590, 2, {0x40, 0x2440, 0x2240, 0x427C, 0x8084, 0x2504, 0x2444, 0x2424, 0x2424, 0x2404, 0x2404, 0x2428, 0x2412, 0x4402, 0x43FE, 0x8000}},
    {23591, 2, {0x200, 0x200, 0x2FC, 0x7F20, 0x140, 0x84, 0x764, 0x381C, 0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {23592, 2, {0x420, 0x410, 0x410, 0x400, 0xFFFE, 0x490, 0x4A0, 0x4C8, 0x490, 0x8A4, 0x8C8, 0x1092, 0x10E2, 0x2182, 0x4E7E, 0x8000}},
    {23593, 2, {0x0, 0x24FC, 0x2220, 0x4220, 0x8020, 0x24F8, 0x2420, 0x2420, 0x2420, 0x2420, 0x25FC, 0x2400, 0x2402, 0x4402, 0x43FE, 0x8000}},
    {23594, 2, {0x2000, 0x23F8, 0x2040, 0x2040, 0xFC40, 0x2840, 0x29F0, 0x2840, 0x2840, 0x2840, 0x2840, 0x2BF8, 0x2802, 0x4802, 0x47FE, 0x8000}},
    {23595, 2, {0x0, 0x1FC, 0xFE20, 0x2420, 0x2420, 0x25FC, 0x2420, 0x2420, 0x2420, 0x2420, 0x27FE, 0x2400, 0x2400, 0x4402, 0x43FE, 0x8000}},
    {23596, 2, {0x2020, 0x2020, 0x2050, 0x2088, 0xFD04, 0x2A02, 0x2888, 0x2888, 0x2888, 0x2888, 0x2908, 0x2908, 0x2A0A, 0x4802, 0x47FE, 0x8000}},
    {23597, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x920, 0x820, 0xFFFE, 0x820, 0x820, 0x7FFC, 0x440, 0x440, 0x840, 0x1042, 0x2042, 0xC03E}},
    {23598, 2, {0x2000, 0x23F8, 0x2088, 0x2090, 0xFCBC, 0x2904, 0x2914, 0x2A48, 0x2840, 0x2BFC, 0x28E0, 0x2950, 0x2A4A, 0x4842, 0x47FE, 0x8000}},
    {23599, 2, {0x2080, 0x2080, 0x21F8, 0x2208, 0xFDFE, 0x2900, 0x2978, 0x2948, 0x2968, 0x2950, 0x2944, 0x2A44, 0x2C3E, 0x4802, 0x47FE, 0x8000}},
    {23600, 2, {0x203C, 0x21E0, 0x2020, 0x23FE, 0xFC20, 0x29FC, 0x2924, 0x29FC, 0x2924, 0x29FC, 0x2820, 0x29FC, 0x2820, 0x4BFE, 0x4802, 0x87FE}},
    {23601, 2, {0x2040, 0x1050, 0xFE48, 0x48, 0x40, 0x7DFE, 0x4450, 0x4450, 0x4450, 0x7C50, 0x1090, 0x5490, 0x9292, 0x1112, 0x510E, 0x2200}},
    {23602, 2, {0x2088, 0x2050, 0x23FE, 0x2050, 0xFDFC, 0x2854, 0x2BFE, 0x2854, 0x29FC, 0x2850, 0x28D8, 0x2954, 0x2A52, 0x4852, 0x47FE, 0x8000}},
    {23603, 2, {0x20F8, 0x2088, 0x20E8, 0x20A8, 0xFDFC, 0x2904, 0x28F8, 0x2888, 0x28F8, 0x2888, 0x28F8, 0x2888, 0x2888, 0x48AA, 0x4892, 0x87FE}},
    {23604, 2, {0x2090, 0x2290, 0x22BC, 0x22D0, 0xFE88, 0x2A88, 0x2800, 0x2BF8, 0x2AA8, 0x2AA8, 0x2AA8, 0x2FFE, 0x2802, 0x4802, 0x47FE, 0x8000}},
    {23605, 2, {0x2020, 0x21FC, 0x2124, 0x21FC, 0xFC20, 0x2BFE, 0x2904, 0x29FC, 0x2904, 0x29FC, 0x2904, 0x29FC, 0x2888, 0x4906, 0x4802, 0x87FE}},
    {23606, 2, {0x3E8, 0xFE88, 0x2BCE, 0x2A50, 0x2BE4, 0x2A82, 0x2BE0, 0x2800, 0x29FC, 0x2954, 0x2954, 0x2954, 0x2BFE, 0x4802, 0x47FE, 0x8000}},
    {23607, 2, {0x23E8, 0x2288, 0x23CE, 0x2250, 0xFFE4, 0x2A82, 0x2BE0, 0x2800, 0x29FC, 0x2954, 0x2954, 0x2954, 0x2BFE, 0x4802, 0x47FE, 0x8000}},
    {23608, 2, {0x0, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {23609, 2, {0x0, 0x0, 0x3FF0, 0x410, 0x410, 0xFFFE, 0x410, 0x410, 0x3FF0, 0x410, 0x400, 0x800, 0x800, 0x1000, 0x2000, 0x4000}},
    {23610, 2, {0x0, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1080, 0x1080, 0x1080, 0x1040, 0x1040, 0x2020, 0x2010, 0x4008, 0x8006}},
    {23611, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2100, 0x2100, 0x2FE0, 0x2120, 0x2120, 0x2220, 0x2220, 0x4422, 0x4422, 0x881E, 0x1000}},
    {23612, 2, {0x0, 0x1FFC, 0x1004, 0x1004, 0x1004, 0x1FFC, 0x1000, 0x1200, 0x120C, 0x1230, 0x13C0, 0x1202, 0x2202, 0x2202, 0x41FE, 0x8000}},
    {23613, 2, {0x0, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1048, 0x1040, 0x1020, 0x2310, 0x2088, 0x4006, 0x8C00, 0x300, 0x80, 0x40}},
    {23614, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x20F0, 0x2F00, 0x2100, 0x21F0, 0x2F00, 0x2100, 0x21F8, 0x5F02, 0x4102, 0x80FE}},
    {23615, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x2088, 0x2ED0, 0x22E0, 0x22A0, 0x24A0, 0x4490, 0x4888, 0x9284, 0x100}},
    {23616, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2000, 0x2000, 0x3FFC, 0x2004, 0x27E4, 0x2424, 0x2424, 0x47E4, 0x4004, 0x8028, 0x10}},
    {23617, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2000, 0x2840, 0x2840, 0x2844, 0x2F48, 0x2850, 0x2860, 0x4844, 0x4944, 0x8A44, 0xC3C}},
    {23618, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x2FFC, 0x2000, 0x2000, 0x3FFE, 0x2100, 0x2210, 0x4408, 0x4FFC, 0x8404}},
    {23619, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x27FC, 0x2404, 0x2444, 0x2444, 0x2444, 0x24B4, 0x4108, 0x4604, 0x9802}},
    {23620, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2040, 0x2FFE, 0x2802, 0x3124, 0x2120, 0x2210, 0x2210, 0x4408, 0x4804, 0x9002}},
    {23621, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x2080, 0x2FF8, 0x2808, 0x4808, 0x4808, 0x8FF8, 0x808}},
    {23622, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2080, 0x2884, 0x2BE4, 0x2884, 0x2884, 0x2FF4, 0x2804, 0x4804, 0x4FFC, 0x8004}},
    {23623, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x27FC, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x4444, 0x47FC, 0x8404}},
    {23624, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2888, 0x2888, 0x2888, 0x2FF8, 0x2080, 0x3084, 0x5084, 0x5084, 0x9FFC, 0x4}},
    {23625, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2090, 0x2490, 0x2490, 0x2490, 0x3FFC, 0x2490, 0x2490, 0x44F0, 0x4400, 0x8400, 0x7FE}},
    {23626, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x2080, 0x2FF8, 0x2888, 0x2888, 0x2FF8, 0x4888, 0x4888, 0x8FF8, 0x808}},
    {23627, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2FFC, 0x2100, 0x2210, 0x27F8, 0x2088, 0x2080, 0x2FFC, 0x4080, 0x4080, 0xBFFE}},
    {23628, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x27FC, 0x2404, 0x2404, 0x27FC, 0x2040, 0x2FFE, 0x2842, 0x484A, 0x4844, 0x8040}},
    {23629, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x3FFE, 0x2420, 0x2424, 0x27A8, 0x28B0, 0x3520, 0x4222, 0x4422, 0x981E}},
    {23630, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2490, 0x22A0, 0x2080, 0x3FFC, 0x21C0, 0x22A0, 0x2490, 0x4888, 0x5086, 0x8080}},
    {23631, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2410, 0x2220, 0x2FF8, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x4220, 0x4420, 0x8420, 0x820}},
    {23632, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2020, 0x2420, 0x29FE, 0x3020, 0x2220, 0x25FC, 0x2C84, 0x3488, 0x4450, 0x4420, 0x84D8, 0x706}},
    {23633, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2490, 0x22A0, 0x2FF8, 0x2808, 0x2FF8, 0x2808, 0x4FF8, 0x4808, 0x8828, 0x810}},
    {23634, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2FF8, 0x2800, 0x2BF8, 0x2800, 0x2FFC, 0x2A48, 0x2A50, 0x5220, 0x5298, 0xA306, 0x200}},
    {23635, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x27FC, 0x2404, 0x27FC, 0x2404, 0x27FC, 0x2404, 0x27FC, 0x4110, 0x4208, 0x8404}},
    {23636, 2, {0x0, 0x2FE, 0x3C82, 0x2082, 0x2082, 0x20FE, 0x3E80, 0x24A0, 0x24A2, 0x24A4, 0x24B8, 0x2EA0, 0xF122, 0x4122, 0x21E, 0x400}},
    {23637, 2, {0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x3FFC, 0x2440, 0x2444, 0x2428, 0x4510, 0x4608, 0x8406}},
    {23638, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2100, 0x2080, 0x2FF8, 0x2410, 0x2220, 0x3FFC, 0x2080, 0x2080, 0x4FF8, 0x4080, 0x8080, 0x80}},
    {23639, 2, {0x8, 0x3F08, 0x2108, 0x217E, 0x3F08, 0x2048, 0x2328, 0x3C28, 0x2708, 0x3C08, 0x27A8, 0x3C12, 0x2402, 0x4402, 0x43FE, 0x8000}},
    {23640, 2, {0x0, 0x3F7E, 0x2102, 0x2104, 0x3F08, 0x2008, 0x237E, 0x3C08, 0x2708, 0x3C08, 0x2788, 0x3C2A, 0x2412, 0x4402, 0x43FE, 0x8000}},
    {23641, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2EFC, 0x2A08, 0x2AE8, 0x2CA8, 0x2AA8, 0x2AA8, 0x2AE8, 0x4AA8, 0x4E08, 0x8828, 0x810}},
    {23642, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x3FFE, 0x2080, 0x2FFC, 0x2884, 0x2CA4, 0x2A94, 0x2884, 0x4CA4, 0x4A94, 0x8884, 0x80C}},
    {23643, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2108, 0x2210, 0x2F7C, 0x2210, 0x2210, 0x3F7E, 0x2210, 0x2210, 0x4410, 0x4410, 0x8810}},
    {23644, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2228, 0x24A8, 0x2AA8, 0x23FE, 0x24A8, 0x2CA8, 0x34A8, 0x24B8, 0x4480, 0x44FE, 0x8400}},
    {23645, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2120, 0x2120, 0x2F3C, 0x2120, 0x2120, 0x2738, 0x2120, 0x2120, 0x4F3C, 0x4120, 0x8120}},
    {23646, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2FF8, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2080, 0x3FFC, 0x5094, 0x57F4, 0x9014, 0x1008}},
    {23647, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2488, 0x3FFE, 0x2488, 0x24F8, 0x2400, 0x27FC, 0x2040, 0x2FFE, 0x4150, 0x4248, 0x8C46, 0x40}},
    {23648, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x2088, 0x27F0, 0x20A0, 0x3FFE, 0x2080, 0x23F8, 0x2608, 0x5BF8, 0x4208, 0x83F8, 0x208}},
    {23649, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2490, 0x22A0, 0x3FFC, 0x22A0, 0x2490, 0x2888, 0x2100, 0x3FFC, 0x2210, 0x4720, 0x40C0, 0x8330, 0xC08}},
    {23650, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x2FF8, 0x2888, 0x3FFE, 0x2888, 0x2FF8, 0x2490, 0x27F0, 0x2100, 0x5FFE, 0x4410, 0x83E0, 0x1C1C}},
    {23651, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2210, 0x2410, 0x289C, 0x3090, 0x2290, 0x25FE, 0x2C10, 0x3490, 0x249C, 0x4490, 0x4550, 0x863E, 0x400}},
    {23652, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2210, 0x2120, 0x2FFC, 0x2A94, 0x29A4, 0x2FFC, 0x2000, 0x27F8, 0x2408, 0x47F8, 0x4408, 0x87F8, 0x408}},
    {23653, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2100, 0x29FE, 0x3200, 0x21FC, 0x2504, 0x29FC, 0x3904, 0x29FC, 0x2880, 0x49FC, 0x4A88, 0x8870, 0xB8E}},
    {23654, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2524, 0x28A8, 0x31FC, 0x24A8, 0x2924, 0x3840, 0x2BFE, 0x4888, 0x4990, 0x8870, 0xB8C}},
    {23655, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2448, 0x2548, 0x2BFE, 0x3548, 0x2578, 0x2D00, 0x35FE, 0x2420, 0x27FE, 0x44A8, 0x4524, 0x8622, 0x420}},
    {23656, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2420, 0x29FC, 0x3124, 0x23FE, 0x2524, 0x2DFC, 0x34A8, 0x24F8, 0x2440, 0x45FE, 0x4488, 0x8470, 0x58C}},
    {23657, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2200, 0x25FC, 0x2840, 0x33FE, 0x2488, 0x2D74, 0x3652, 0x25FC, 0x2504, 0x4574, 0x4554, 0x8574, 0x50C}},
    {23658, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2040, 0x2FFE, 0x2882, 0x27FC, 0x2110, 0x2BFA, 0x260C, 0x2BFA, 0x2208, 0x43F8, 0x4040, 0x8448, 0x8C4}},
    {23659, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x22FE, 0x2A20, 0x2BFE, 0x2E44, 0x22BA, 0x2328, 0x3EFE, 0x2A82, 0x4ABA, 0x4AAA, 0x92BA, 0x2286}},
    {23660, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2490, 0x22A0, 0x2490, 0x2FF8, 0x2948, 0x2FFC, 0x2484, 0x2FF4, 0x3494, 0x47F4, 0x4094, 0x8FF4, 0x8}},
    {23661, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x2208, 0x2FBE, 0x28A2, 0x2FBE, 0x48A2, 0x4FBE, 0x8514, 0x8A2}},
    {23662, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x108, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {23663, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x102, 0x102, 0x102, 0xFE, 0x0}},
    {23664, 2, {0x1010, 0x810, 0x820, 0xFFFE, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x208, 0x400, 0x400, 0x800, 0x3000, 0xC000}},
    {23665, 2, {0x100, 0x100, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8, 0x0}},
    {23666, 2, {0x8, 0x7C, 0x780, 0x7800, 0x100, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8, 0x0}},
    {23667, 2, {0x600, 0x100, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23668, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x200, 0x200, 0x7FF8, 0x208, 0x408, 0x408, 0x808, 0x1008, 0x2050, 0x4020}},
    {23669, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x3FFC, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {23670, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FF8, 0x408, 0x408, 0x808, 0x808, 0x1008, 0x2050, 0xC020}},
    {23671, 2, {0x1000, 0x1000, 0x13F8, 0x1088, 0x5490, 0x5490, 0x54A0, 0x54BC, 0x5484, 0x5484, 0x5504, 0x5D04, 0x6504, 0x204, 0x228, 0x410}},
    {23672, 2, {0x1000, 0x1000, 0x11FC, 0x1020, 0x5420, 0x5420, 0x5420, 0x5420, 0x5420, 0x5420, 0x5420, 0x5C20, 0x6420, 0x20, 0x3FE, 0x0}},
    {23673, 2, {0x1040, 0x1040, 0x1080, 0x10FE, 0x5500, 0x5600, 0x54FC, 0x5408, 0x5410, 0x5420, 0x5440, 0x5C80, 0x6502, 0x102, 0xFE, 0x0}},
    {23674, 2, {0x1000, 0x1000, 0x11F8, 0x1008, 0x5408, 0x5408, 0x5408, 0x55F8, 0x5508, 0x5500, 0x5500, 0x5D00, 0x6502, 0x102, 0xFE, 0x0}},
    {23675, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0x5444, 0x5544, 0x5544, 0x5544, 0x5544, 0x5644, 0x5444, 0x5C84, 0x6484, 0x104, 0x228, 0x410}},
    {23676, 2, {0x1000, 0x1000, 0x13FC, 0x1090, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5C92, 0x6512, 0x112, 0x20E, 0x400}},
    {23677, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x5420, 0x5420, 0x5420, 0x57FE, 0x5420, 0x5420, 0x5420, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {23678, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x5524, 0x5524, 0x5524, 0x5524, 0x5524, 0x5524, 0x5524, 0x5D24, 0x6524, 0x1FC, 0x4, 0x0}},
    {23679, 2, {0x1040, 0x1040, 0x1040, 0x107E, 0x5440, 0x5480, 0x5480, 0x54FC, 0x5404, 0x5404, 0x5404, 0x5DF4, 0x6404, 0x4, 0x28, 0x10}},
    {23680, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x4104, 0x4104, 0x4104, 0x4104, 0x7FFC, 0x4, 0x0}},
    {23681, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x200, 0x200, 0x7F0, 0x810, 0x1410, 0x2220, 0x140, 0x80, 0x300, 0x1C00, 0xE000}},
    {23682, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x3FF0, 0x2000, 0x2004, 0x2004, 0x1FFC}},
    {23683, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x208, 0x1208, 0x1208, 0x2208, 0x408, 0x808, 0x1050, 0x6020}},
    {23684, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0x5484, 0x54FC, 0x5484, 0x5484, 0x5484, 0x54FC, 0x5484, 0x5C84, 0x6504, 0x104, 0x214, 0x408}},
    {23685, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x5500, 0x55FC, 0x5544, 0x5544, 0x5544, 0x5528, 0x5528, 0x5D10, 0x6510, 0x228, 0x244, 0x482}},
    {23686, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x5420, 0x5420, 0x57FE, 0x5420, 0x5420, 0x5450, 0x5450, 0x5C50, 0x6488, 0x88, 0x104, 0x202}},
    {23687, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0xC00, 0x70FC, 0x4084, 0x4084, 0x4084, 0x4C84, 0x7094, 0x4088, 0x80, 0x80}},
    {23688, 2, {0x1000, 0x11FE, 0x1008, 0x1088, 0x5488, 0x5488, 0x5508, 0x55FE, 0x5418, 0x5428, 0x5448, 0x5C88, 0x6508, 0x208, 0x28, 0x10}},
    {23689, 2, {0x1080, 0x1080, 0x1080, 0x10FC, 0x5554, 0x5654, 0x5454, 0x5494, 0x5494, 0x5524, 0x5624, 0x5C44, 0x6444, 0x84, 0x128, 0x10}},
    {23690, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2002, 0x2002, 0x1FFE, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23691, 2, {0x1000, 0x13FC, 0x1084, 0x1088, 0x5488, 0x5490, 0x549C, 0x5484, 0x5544, 0x5544, 0x5528, 0x5D28, 0x6610, 0x228, 0x444, 0x182}},
    {23692, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF0, 0x820, 0x840, 0xCF8, 0xA08, 0x1110, 0x10A0, 0x2040, 0x20A0, 0x4318, 0x8C06}},
    {23693, 2, {0x1000, 0x11FC, 0x1088, 0x1088, 0x5488, 0x5488, 0x5488, 0x57FE, 0x5488, 0x5488, 0x5488, 0x5C88, 0x6488, 0x108, 0x108, 0x208}},
    {23694, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0x5508, 0x5504, 0x5604, 0x55FA, 0x5488, 0x5488, 0x5488, 0x5C88, 0x6508, 0x108, 0x228, 0x410}},
    {23695, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0x5400, 0x57FE, 0x5490, 0x5490, 0x5490, 0x5490, 0x5490, 0x5C92, 0x6512, 0x112, 0x20E, 0x400}},
    {23696, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5484, 0x5488, 0x5448, 0x5C50, 0x6420, 0x50, 0x188, 0x606}},
    {23697, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0x440, 0xA20, 0x3118, 0xC106, 0x1FE0, 0x20, 0x40, 0x40, 0x80, 0x100}},
    {23698, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0x5488, 0x5544, 0x5622, 0x5420, 0x5400, 0x55FC, 0x5404, 0x5C08, 0x6408, 0x10, 0x10, 0x20}},
    {23699, 2, {0x1008, 0x101C, 0x10E0, 0x1080, 0x5480, 0x5480, 0x54FE, 0x5488, 0x5488, 0x5488, 0x5488, 0x5C88, 0x6508, 0x108, 0x208, 0x408}},
    {23700, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23701, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0x440, 0x820, 0x3458, 0xC446, 0x440, 0x440, 0x840, 0x840, 0x1040}},
    {23702, 2, {0x1000, 0x11FE, 0x1100, 0x1104, 0x5544, 0x5528, 0x5528, 0x5510, 0x5510, 0x5528, 0x5528, 0x5D44, 0x6584, 0x100, 0x1FE, 0x0}},
    {23703, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x2048, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2808, 0x2028, 0x2010}},
    {23704, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x5524, 0x5524, 0x5524, 0x5524, 0x5524, 0x5554, 0x5450, 0x5C90, 0x6490, 0x112, 0x212, 0x40E}},
    {23705, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC106, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23706, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1050, 0x1450, 0x1290, 0x1110, 0x1292, 0x244A, 0x284A, 0x4006, 0x8002}},
    {23707, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1210, 0x1150, 0x1020, 0x1000, 0x1FFC, 0x204, 0x2224, 0x2224, 0x3FE4, 0x4, 0x28, 0x10}},
    {23708, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2002, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {23709, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x800, 0x800, 0x1FFC, 0x2400, 0x4400, 0x87F0, 0x400, 0x400, 0x7F8, 0x400, 0x400, 0x400}},
    {23710, 2, {0x1080, 0x1080, 0x1080, 0x10FE, 0x5540, 0x5540, 0x5640, 0x547C, 0x5440, 0x5440, 0x5440, 0x5C7E, 0x6440, 0x40, 0x40, 0x40}},
    {23711, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5524, 0x5524, 0x5524, 0x5524, 0x5524, 0x57FE, 0x5420, 0x5C50, 0x6450, 0x88, 0x104, 0x202}},
    {23712, 2, {0x1000, 0x10FE, 0x1080, 0x1080, 0x5480, 0x54FC, 0x5484, 0x5484, 0x5484, 0x5484, 0x54FC, 0x5C80, 0x6480, 0x80, 0xFE, 0x0}},
    {23713, 2, {0x0, 0x7FFC, 0x4004, 0x4824, 0x4424, 0x4444, 0x5FF4, 0x4104, 0x4104, 0x4924, 0x4924, 0x4924, 0x4FE4, 0x4004, 0x4014, 0x4008}},
    {23714, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x10, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x10, 0x50, 0x20}},
    {23715, 2, {0x1080, 0x1080, 0x1100, 0x11FC, 0x5604, 0x5404, 0x55E4, 0x5524, 0x5524, 0x5524, 0x55E4, 0x5D24, 0x6404, 0x4, 0x28, 0x10}},
    {23716, 2, {0x2040, 0x2020, 0x2020, 0x23FE, 0xAA02, 0xAC04, 0xA800, 0xA890, 0xA890, 0xA890, 0xA888, 0xB908, 0xC908, 0x104, 0x204, 0x402}},
    {23717, 2, {0x1010, 0x1010, 0x1010, 0x11FE, 0x5512, 0x5514, 0x5510, 0x55FC, 0x5544, 0x5544, 0x5528, 0x5D28, 0x6510, 0x228, 0x244, 0x482}},
    {23718, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x820, 0x440, 0x0, 0xFFFE, 0x0}},
    {23719, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x2020, 0xDFF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {23720, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0x5488, 0x54F8, 0x5488, 0x5488, 0x5488, 0x54F8, 0x5488, 0x5C88, 0x6488, 0x88, 0x3FE, 0x0}},
    {23721, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x400, 0x400, 0xFF8, 0x1808, 0x2808, 0xC808, 0xFF8, 0x808}},
    {23722, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x440, 0x3FF8, 0x448, 0x448, 0x1FF8, 0x1440, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2048, 0x4040}},
    {23723, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x55FC, 0x5524, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5524, 0x5D24, 0x6524, 0x124, 0x1FC, 0x104}},
    {23724, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x5524, 0x55FC, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5524, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {23725, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0x5488, 0x5524, 0x5612, 0x5410, 0x55FC, 0x5404, 0x5408, 0x5C88, 0x6450, 0x20, 0x10, 0x10}},
    {23726, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0x5502, 0x5604, 0x5480, 0x5488, 0x5490, 0x54A0, 0x54C0, 0x5C82, 0x6482, 0x82, 0x7E, 0x0}},
    {23727, 2, {0x1000, 0x11FE, 0x1010, 0x1010, 0x5420, 0x5420, 0x5468, 0x54A4, 0x5522, 0x5622, 0x5420, 0x5C20, 0x6420, 0x0, 0x3FE, 0x0}},
    {23728, 2, {0x2010, 0x2090, 0x2090, 0x2110, 0xA97E, 0xAA52, 0xAB92, 0xA892, 0xA912, 0xA912, 0xAA52, 0xBBD2, 0xC862, 0x22, 0x4A, 0x84}},
    {23729, 2, {0x890, 0x888, 0x1080, 0x309E, 0x57E0, 0x9080, 0x1044, 0x1034, 0x110C, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23730, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x5504, 0x5504, 0x55FC, 0x5450, 0x5450, 0x5450, 0x5450, 0x5C92, 0x6492, 0x112, 0x20E, 0x400}},
    {23731, 2, {0x70, 0x1F80, 0x1000, 0x1000, 0x1FF8, 0x1080, 0x1080, 0x1080, 0xFFFE, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23732, 2, {0x1008, 0x101C, 0x10E0, 0x1080, 0x5480, 0x5480, 0x5480, 0x54FE, 0x5488, 0x5488, 0x5488, 0x5C88, 0x6488, 0x88, 0x1FE, 0x0}},
    {23733, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x55FE, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5504, 0x5504, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23734, 2, {0x1020, 0x1020, 0x1040, 0x11FC, 0x5504, 0x5504, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5504, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23735, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x5504, 0x55FC, 0x5520, 0x5520, 0x55FE, 0x5520, 0x5520, 0x5D10, 0x6512, 0x14A, 0x186, 0x102}},
    {23736, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x4080, 0x4080, 0x8080, 0x80}},
    {23737, 2, {0x1000, 0x13FC, 0x1084, 0x1084, 0x5484, 0x5504, 0x5514, 0x5608, 0x5400, 0x55FC, 0x5504, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23738, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0xC60, 0x3218, 0xC106, 0x1FE0, 0x20, 0x40, 0xC80, 0x300, 0x80, 0x40}},
    {23739, 2, {0x1004, 0x101E, 0x11F0, 0x1110, 0x5510, 0x5510, 0x5510, 0x55FE, 0x5510, 0x5510, 0x5510, 0x5D08, 0x650A, 0x14A, 0x1A6, 0x112}},
    {23740, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x5524, 0x54A4, 0x54A8, 0x5420, 0x57FE, 0x5420, 0x5420, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {23741, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x400, 0x900, 0x1100, 0x3FF8, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {23742, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x543E, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5504, 0x5504, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23743, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x800, 0x4BF8, 0x4808, 0x4808, 0x4808, 0x49F8, 0x4808, 0x4808, 0x1008, 0x13F8, 0x2008}},
    {23744, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {23745, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0xC00, 0x70FC, 0x4084, 0x4484, 0x4484, 0x5C84, 0x6494, 0x4888, 0x880, 0x1080, 0x6080}},
    {23746, 2, {0x1040, 0x1040, 0x107C, 0x1084, 0x5588, 0x5650, 0x5420, 0x5450, 0x5488, 0x5706, 0x5460, 0x5C10, 0x6408, 0xC0, 0x20, 0x10}},
    {23747, 2, {0x208, 0x2108, 0x1110, 0x1020, 0x0, 0x7FFE, 0x4002, 0x8104, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23748, 2, {0x1000, 0x13FC, 0x1104, 0x1088, 0x5450, 0x5420, 0x54D8, 0x5726, 0x5420, 0x55FC, 0x5420, 0x5C20, 0x67FE, 0x20, 0x20, 0x20}},
    {23749, 2, {0x1020, 0x1020, 0x1040, 0x1088, 0x5504, 0x57FE, 0x5402, 0x5488, 0x5488, 0x57FE, 0x5488, 0x5C88, 0x6508, 0x108, 0x208, 0x408}},
    {23750, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0x5504, 0x5602, 0x55FC, 0x5400, 0x5400, 0x55FC, 0x5504, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23751, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {23752, 2, {0x2080, 0x2080, 0x20F8, 0x2108, 0xAB10, 0xACA0, 0xA840, 0xA8A0, 0xA918, 0xAA06, 0xADF8, 0xB908, 0xC908, 0x108, 0x1F8, 0x108}},
    {23753, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x800, 0xFF0, 0x1020, 0x2C40, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {23754, 2, {0x400, 0x800, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23755, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0x5604, 0x55E4, 0x5524, 0x5524, 0x55E4, 0x5524, 0x5524, 0x5DE4, 0x6504, 0x4, 0x28, 0x10}},
    {23756, 2, {0x1000, 0x11FE, 0x1020, 0x1040, 0x5488, 0x5504, 0x55FE, 0x5422, 0x5420, 0x5420, 0x55FE, 0x5C20, 0x6420, 0x20, 0x3FE, 0x0}},
    {23757, 2, {0x2020, 0x2020, 0x21FC, 0x2024, 0xABFE, 0xA824, 0xA9FC, 0xA820, 0xA820, 0xA9FC, 0xA820, 0xB820, 0xCBFE, 0x20, 0x20, 0x20}},
    {23758, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF0, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x2208, 0x2110, 0x20A0, 0x2440, 0x2830, 0x300E}},
    {23759, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0x5440, 0x55FC, 0x5554, 0x5554, 0x5554, 0x5554, 0x5554, 0x5D54, 0x6554, 0x144, 0x114, 0x108}},
    {23760, 2, {0x1020, 0x1010, 0x1010, 0x13FE, 0x5420, 0x5442, 0x5484, 0x55F8, 0x5410, 0x5422, 0x54C4, 0x5F08, 0x6410, 0x28, 0xC4, 0x302}},
    {23761, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x7FFC}},
    {23762, 2, {0x1000, 0x11FE, 0x1102, 0x1102, 0x557A, 0x5502, 0x5502, 0x557A, 0x554A, 0x554A, 0x554A, 0x5D7A, 0x6502, 0x102, 0x10A, 0x104}},
    {23763, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x55FC, 0x5424, 0x55FC, 0x5520, 0x55FE, 0x5422, 0x542A, 0x5C54, 0x6450, 0x88, 0x104, 0x202}},
    {23764, 2, {0x1040, 0x1040, 0x1044, 0x11F4, 0x5448, 0x5450, 0x57FE, 0x5440, 0x5480, 0x5584, 0x5698, 0x5CE0, 0x6482, 0x82, 0x7E, 0x0}},
    {23765, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {23766, 2, {0x2040, 0x2020, 0x2020, 0x23FE, 0xAA02, 0xAC44, 0xA840, 0xABFE, 0xA888, 0xA888, 0xA908, 0xB8D0, 0xC820, 0x50, 0x88, 0x304}},
    {23767, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0x5608, 0x55FE, 0x5500, 0x557C, 0x5544, 0x5544, 0x5554, 0x5D48, 0x6542, 0x242, 0x23E, 0x400}},
    {23768, 2, {0x1000, 0x13FE, 0x1000, 0x1000, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5D04, 0x6400, 0x0, 0x3FE, 0x0}},
    {23769, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5420, 0x5420, 0x57FE, 0x5408, 0x5408, 0x55FE, 0x5408, 0x5C88, 0x6448, 0x8, 0x28, 0x10}},
    {23770, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x200, 0x7FFC, 0x440, 0x1830, 0xE10E, 0x100, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0}},
    {23771, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x4, 0x7F84, 0x824, 0x824, 0x1F24, 0x2124, 0x5124, 0xA24, 0x404, 0x804, 0x3014, 0xC008}},
    {23772, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x9F0, 0x900, 0x900, 0x900, 0x7FFC, 0x0}},
    {23773, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {23774, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x800, 0x1FE0, 0x2020, 0x5FFC, 0x9000, 0x17F0, 0x1410, 0x1454, 0x2424, 0x2404, 0x43FC, 0x8000}},
    {23775, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x400, 0x400, 0x7FFC, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0x850, 0x820}},
    {23776, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0x5420, 0x55FE, 0x5400, 0x5400, 0x55FE, 0x5420, 0x5428, 0x5C24, 0x6422, 0x20, 0x20, 0x20}},
    {23777, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5420, 0x5524, 0x54A4, 0x54A8, 0x5420, 0x57FE, 0x5450, 0x5C50, 0x6488, 0x88, 0x104, 0x202}},
    {23778, 2, {0x2002, 0x2002, 0x27E2, 0x210A, 0xA90A, 0xA9EA, 0xA92A, 0xAA2A, 0xAB2A, 0xAAAA, 0xAC4A, 0xB84A, 0xC882, 0x102, 0x20A, 0x404}},
    {23779, 2, {0x1080, 0x1080, 0x10BC, 0x13C0, 0x5450, 0x5424, 0x54D4, 0x570C, 0x5400, 0x57FE, 0x5490, 0x5C90, 0x6512, 0x112, 0x20E, 0x400}},
    {23780, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x5420, 0x57FE, 0x5450, 0x5488, 0x5504, 0x568A, 0x5488, 0x5C88, 0x6488, 0x108, 0x108, 0x208}},
    {23781, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x5610, 0x55FC, 0x5424, 0x5424, 0x55FE, 0x5424, 0x5424, 0x5DFC, 0x6424, 0x20, 0xA0, 0x40}},
    {23782, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x1448, 0x2444, 0x4444, 0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23783, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x5400, 0x5488, 0x5504, 0x5602, 0x5488, 0x5488, 0x5450, 0x5C50, 0x6420, 0x50, 0x88, 0x306}},
    {23784, 2, {0x1010, 0x10D8, 0x1394, 0x1094, 0x5490, 0x57FE, 0x5490, 0x5494, 0x5494, 0x54D8, 0x5798, 0x5C90, 0x64AA, 0xCA, 0x286, 0x102}},
    {23785, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x640, 0x3850, 0x848, 0xFFFE, 0x840, 0x840, 0xE24, 0x7828, 0x812, 0x82A, 0x28C6, 0x1302}},
    {23786, 2, {0x1000, 0x1050, 0x1048, 0x1084, 0x5524, 0x5420, 0x5450, 0x5488, 0x5506, 0x56F8, 0x5488, 0x5C88, 0x6488, 0x88, 0xF8, 0x88}},
    {23787, 2, {0x2000, 0x27DE, 0x2092, 0x2292, 0xAA94, 0xAA94, 0xABD8, 0xA894, 0xA992, 0xA992, 0xAA92, 0xBA9A, 0xCC94, 0x890, 0x290, 0x110}},
    {23788, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0x5420, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x5524, 0x5524, 0x5DFC, 0x6524, 0x124, 0x124, 0x10C}},
    {23789, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0x5420, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5504, 0x5DFC, 0x6504, 0x104, 0x114, 0x108}},
    {23790, 2, {0x2000, 0x23F8, 0x2088, 0x2088, 0xAFFE, 0xA888, 0xA888, 0xABF8, 0xA900, 0xA900, 0xABF8, 0xBD08, 0xC908, 0x108, 0x1F8, 0x108}},
    {23791, 2, {0x100, 0x4104, 0x7FFC, 0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE10E, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100}},
    {23792, 2, {0x1040, 0x1040, 0x10FC, 0x1088, 0x5550, 0x5420, 0x54D8, 0x5726, 0x54F8, 0x5420, 0x54F8, 0x5C20, 0x67FE, 0x20, 0x20, 0x20}},
    {23793, 2, {0x0, 0x45FC, 0x2824, 0x1024, 0x29FC, 0x4844, 0x8844, 0xBFE, 0x1800, 0x2820, 0x4820, 0x8924, 0x924, 0x924, 0x51FC, 0x2004}},
    {23794, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x608, 0x7808, 0x848, 0x848, 0xFF48, 0x1C48, 0x2A48, 0x4948, 0x8808, 0x828, 0x810}},
    {23795, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1040, 0x1040, 0x247E, 0x2488, 0x6488, 0xA548, 0x2450, 0x2420, 0x2050, 0x2088, 0x2306}},
    {23796, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5C50, 0x6450, 0x92, 0x112, 0x20E}},
    {23797, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xAC44, 0xA840, 0xABFE, 0xA840, 0xA890, 0xA890, 0xA910, 0xB920, 0xCA24, 0x242, 0x4FE, 0x842}},
    {23798, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x204, 0x2224, 0x3FE4, 0x28, 0x10}},
    {23799, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {23800, 2, {0x2028, 0x2024, 0x2024, 0x2020, 0xABFE, 0xAA20, 0xAA24, 0xAA24, 0xABA4, 0xAAA8, 0xAAA8, 0xBA90, 0xCA92, 0x5AA, 0x446, 0x882}},
    {23801, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x3FF8, 0x100, 0x1110, 0x2508, 0x200, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23802, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x5D20, 0x64A0, 0x40, 0xB0, 0x30E}},
    {23803, 2, {0x2040, 0x2040, 0x2088, 0x2104, 0xABFE, 0xA802, 0xA888, 0xA944, 0xAA42, 0xA8F8, 0xA988, 0xBA50, 0xC820, 0x50, 0x188, 0x606}},
    {23804, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0x5520, 0x5620, 0x5420, 0x57FE, 0x5400, 0x5400, 0x55FC, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23805, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x5420, 0x5420, 0x5524, 0x5524, 0x5524, 0x56AA, 0x5420, 0x5C50, 0x6450, 0x88, 0x104, 0x202}},
    {23806, 2, {0x200C, 0x11F0, 0x8100, 0x4900, 0x9FE, 0x1110, 0xE110, 0x2210, 0x2210, 0x2510, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23807, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0x55FC, 0x5484, 0x5484, 0x57FE, 0x5400, 0x5400, 0x55FC, 0x5D04, 0x6504, 0x104, 0x1FC, 0x104}},
    {23808, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1108, 0x1090, 0x1260, 0x1418, 0x1806}},
    {23809, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1080, 0x1080, 0x10FC, 0x7D04, 0x1248, 0x1040, 0x1040, 0x1CA0, 0xE0A0, 0x4110, 0x208, 0xC06}},
    {23810, 2, {0x2088, 0x2088, 0x23FE, 0x2088, 0xA800, 0xABFE, 0xAA02, 0xAC44, 0xA840, 0xABFC, 0xA844, 0xB884, 0xC884, 0x104, 0x228, 0x410}},
    {23811, 2, {0x2040, 0x2040, 0x2040, 0x27FC, 0xA840, 0xAA48, 0xA948, 0xA950, 0xAFFE, 0xA8E0, 0xA950, 0xB950, 0xCA48, 0x444, 0x842, 0x40}},
    {23812, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0x5488, 0x5504, 0x56FA, 0x5400, 0x5444, 0x5424, 0x5524, 0x5CA8, 0x6488, 0x10, 0x3FE, 0x0}},
    {23813, 2, {0x1040, 0x1040, 0x107C, 0x1084, 0x5508, 0x56FE, 0x5492, 0x5492, 0x54FE, 0x5492, 0x5492, 0x5CFE, 0x6492, 0x112, 0x10A, 0x204}},
    {23814, 2, {0x2040, 0x2020, 0x2020, 0x23FE, 0xAA02, 0xAC94, 0xA908, 0xAA04, 0xA800, 0xA9FC, 0xA820, 0xB820, 0xC820, 0x20, 0x7FE, 0x0}},
    {23815, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x7FFC, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {23816, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x7FFC, 0x1110, 0x2108, 0x4504, 0x200, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23817, 2, {0x1020, 0x1022, 0x13B4, 0x10A8, 0x54A8, 0x5524, 0x56A2, 0x5440, 0x55FC, 0x5504, 0x5504, 0x5DFC, 0x6504, 0x104, 0x1FC, 0x104}},
    {23818, 2, {0x2088, 0x2088, 0x2088, 0x2088, 0xABDE, 0xA888, 0xA888, 0xA99C, 0xA9DC, 0xAAAA, 0xAAAA, 0xBCC8, 0xC888, 0x88, 0x88, 0x88}},
    {23819, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x1110, 0x1110, 0xFFFE, 0x1110, 0x1110, 0x7FFC, 0x100, 0xFFFE, 0x100, 0x100}},
    {23820, 2, {0x1000, 0x11FE, 0x1102, 0x1102, 0x55FE, 0x5510, 0x5510, 0x55FE, 0x5510, 0x5510, 0x557E, 0x5D42, 0x6542, 0x242, 0x27E, 0x442}},
    {23821, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x5420, 0x5524, 0x5524, 0x5524, 0x56AA, 0x5470, 0x54A8, 0x5CA8, 0x6524, 0x222, 0x420, 0x20}},
    {23822, 2, {0x1020, 0x1020, 0x11FC, 0x1050, 0x5488, 0x5504, 0x57FE, 0x5408, 0x55E8, 0x5528, 0x5528, 0x5DE8, 0x6528, 0x8, 0x28, 0x10}},
    {23823, 2, {0x201C, 0x23E0, 0x2220, 0x23FE, 0xAA20, 0xAA92, 0xAB0A, 0xAA06, 0xA9FC, 0xA904, 0xA904, 0xB9FC, 0xC904, 0x104, 0x1FC, 0x104}},
    {23824, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5400, 0x5512, 0x55D4, 0x5D18, 0x6510, 0x152, 0x192, 0x10E}},
    {23825, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x2080, 0x208C, 0x3EF0, 0x2082, 0x2682, 0x387E}},
    {23826, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0x7FFC, 0x820, 0x1450, 0x2288, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {23827, 2, {0x2000, 0x23FE, 0x2222, 0x2222, 0xABFE, 0xAA22, 0xAA22, 0xAAFA, 0xAA8A, 0xAA8A, 0xAA8A, 0xBAFA, 0xCA8A, 0x202, 0x3FE, 0x202}},
    {23828, 2, {0x100, 0x4104, 0x7FFC, 0x900, 0x880, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000}},
    {23829, 2, {0x1000, 0x11FE, 0x1110, 0x1110, 0x557C, 0x5510, 0x5510, 0x55FE, 0x5500, 0x5510, 0x5510, 0x5D7C, 0x6510, 0x210, 0x2FE, 0x400}},
    {23830, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FFC, 0x2080, 0x2FF8, 0x2080, 0x3FFC, 0x2000, 0x2080, 0x4FF8, 0x4080, 0x9FFE, 0x0}},
    {23831, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x4824, 0x4444, 0x5FF4, 0x4104, 0x4924, 0x4924, 0x4924, 0x4FE4, 0x4014, 0x4008}},
    {23832, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0x5504, 0x56FA, 0x5400, 0x55FC, 0x5554, 0x5554, 0x55FC, 0x5D54, 0x6554, 0x154, 0x104, 0x10C}},
    {23833, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x2488, 0x2488, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {23834, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xA820, 0xABFE, 0xA888, 0xA944, 0xAA42, 0xA8F8, 0xA988, 0xBA50, 0xC820, 0x50, 0x188, 0x606}},
    {23835, 2, {0x2000, 0x23FE, 0x2202, 0x2202, 0xABFE, 0xAA10, 0xAA92, 0xAA92, 0xAA92, 0xAAFE, 0xAA10, 0xBA92, 0xCA92, 0x492, 0x4FE, 0x802}},
    {23836, 2, {0x2008, 0x203C, 0x21E0, 0x2020, 0xABFE, 0xA8A8, 0xA8A8, 0xA8A8, 0xABFE, 0xA8A8, 0xA8A8, 0xBBFE, 0xC820, 0x20, 0x1FC, 0x0}},
    {23837, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x55FC, 0x5420, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5D04, 0x65FC, 0x104, 0x114, 0x108}},
    {23838, 2, {0x2040, 0x2020, 0x23FE, 0x2000, 0xA9FC, 0xA904, 0xA9FC, 0xA800, 0xA9FC, 0xA808, 0xA810, 0xBBFE, 0xC820, 0x20, 0xA0, 0x40}},
    {23839, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE, 0x0}},
    {23840, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x55FC, 0x5420, 0x5470, 0x5CA8, 0x6124, 0x622, 0x20, 0x20}},
    {23841, 2, {0x2000, 0x21FC, 0x2008, 0x2010, 0xAA22, 0xAAAA, 0xAA72, 0xAA22, 0xAA72, 0xAAAA, 0xAB26, 0xBAA2, 0xCA42, 0x202, 0x3FE, 0x2}},
    {23842, 2, {0x101E, 0x13E0, 0x1122, 0x1094, 0x5400, 0x55FC, 0x5424, 0x5424, 0x57FE, 0x5424, 0x5424, 0x5DFC, 0x6424, 0x20, 0xA0, 0x40}},
    {23843, 2, {0x100, 0x4104, 0x7FFC, 0x70, 0x1F80, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x400, 0x7FFC, 0x820, 0x1C40, 0x380, 0x7C78}},
    {23844, 2, {0x2184, 0x2068, 0x2030, 0x20C8, 0xAB24, 0xA820, 0xABFE, 0xA840, 0xA8FC, 0xA984, 0xAAFC, 0xB884, 0xC8FC, 0x84, 0x94, 0x88}},
    {23845, 2, {0x1040, 0x1080, 0x11FC, 0x1124, 0x5524, 0x55FC, 0x5524, 0x5544, 0x55FC, 0x5490, 0x5510, 0x5FFE, 0x6410, 0x10, 0x10, 0x10}},
    {23846, 2, {0x2040, 0x2080, 0x27FE, 0x2110, 0xAA48, 0xAC46, 0xABF8, 0xAA48, 0xAA48, 0xABF8, 0xAA48, 0xBA48, 0xCBF8, 0x42, 0x42, 0x3E}},
    {23847, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1010, 0x1090, 0x1090, 0xFD08, 0x3148, 0x3A44, 0x5442, 0x5090, 0x9088, 0x1108, 0x13FC, 0x1104}},
    {23848, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x55FC, 0x5424, 0x57FE, 0x5424, 0x55FC, 0x5420, 0x5520, 0x5D3E, 0x6520, 0x2A0, 0x27E, 0x400}},
    {23849, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x4A94, 0x8508}},
    {23850, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x5488, 0x5488, 0x5554, 0x5622, 0x5400, 0x5420, 0x57FE, 0x5C20, 0x6420, 0x20, 0x20, 0x20}},
    {23851, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2080, 0x2490, 0x2490, 0x27F0, 0x2080, 0x4888, 0x4888, 0x8FF8}},
    {23852, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x380, 0xD60, 0x3118, 0xC106}},
    {23853, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1000, 0x100C, 0xFEF0, 0x2080, 0x4880, 0x7EFE, 0x888, 0xE88, 0xF888, 0x4908, 0x908, 0xA08}},
    {23854, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {23855, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0x5504, 0x5602, 0x54F8, 0x5420, 0x5420, 0x55FC, 0x5420, 0x5D24, 0x64A4, 0xA8, 0x3FE, 0x0}},
    {23856, 2, {0x1092, 0x1092, 0x1124, 0x1248, 0x5524, 0x5492, 0x5492, 0x5400, 0x55FE, 0x5522, 0x5522, 0x5DFE, 0x6522, 0x122, 0x1FE, 0x102}},
    {23857, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3F04, 0x2124, 0x3F24, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x1204, 0x2114, 0x4088}},
    {23858, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x55FC, 0x5504, 0x55FC, 0x5400, 0x57FE, 0x5420, 0x5420, 0x5DFC, 0x6420, 0x20, 0x3FE, 0x0}},
    {23859, 2, {0x2040, 0x20A0, 0x2110, 0x2208, 0xADF6, 0xA800, 0xABC4, 0xAA54, 0xAA54, 0xABD4, 0xAA54, 0xBA54, 0xCBD4, 0x244, 0x254, 0x2C8}},
    {23860, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2224, 0x2FA8, 0x2498, 0x2D12, 0x422A, 0x4546, 0x8882}},
    {23861, 2, {0x1000, 0x10FC, 0x1084, 0x10FC, 0x5484, 0x54FC, 0x5400, 0x55FE, 0x5480, 0x54FE, 0x552A, 0x5E4A, 0x6492, 0x122, 0x54, 0x88}},
    {23862, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x800, 0x808, 0x7F08, 0x8FE, 0xFF08, 0x848, 0x828, 0x7F28, 0x808, 0xF08, 0xF028, 0x4010}},
    {23863, 2, {0x1088, 0x1048, 0x1050, 0x13FE, 0x5450, 0x5450, 0x55FC, 0x5554, 0x5554, 0x558C, 0x5504, 0x5DFC, 0x6504, 0x104, 0x1FC, 0x104}},
    {23864, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5D04, 0x65FC, 0x0, 0x88, 0x104}},
    {23865, 2, {0x2040, 0x2020, 0x21FC, 0x2000, 0xA888, 0xA850, 0xABFE, 0xAA22, 0xAC24, 0xA9FC, 0xA924, 0xB924, 0xC934, 0x128, 0x20, 0x20}},
    {23866, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x2040, 0x1040, 0xFEFE, 0x2120, 0x20AC, 0x3CB4, 0x25E4, 0x24AC, 0x24A0, 0x44A2, 0x5482, 0x887E}},
    {23867, 2, {0x2110, 0x2110, 0x2210, 0x247C, 0xA910, 0xA910, 0xAAFE, 0xAE08, 0xAA08, 0xAAFE, 0xAA08, 0xBA48, 0xCA28, 0x208, 0x228, 0x210}},
    {23868, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5400, 0x57FE, 0x5420, 0x5D20, 0x653C, 0x120, 0x2A0, 0x47E}},
    {23869, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {23870, 2, {0x2000, 0x23FE, 0x2050, 0x2050, 0xABFE, 0xAA52, 0xAA52, 0xABFE, 0xA800, 0xA840, 0xAFFE, 0xB888, 0xC990, 0x60, 0xD8, 0x304}},
    {23871, 2, {0x1000, 0x11DC, 0x1154, 0x1154, 0x55DC, 0x5400, 0x55FC, 0x5400, 0x57FE, 0x5480, 0x54FC, 0x5C04, 0x6404, 0x4, 0x28, 0x10}},
    {23872, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x840, 0x820, 0x7EFC, 0x820, 0x1820, 0x1CFC, 0x2A20, 0x2A20, 0x4820, 0x89FE, 0x800}},
    {23873, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x54F8, 0x5488, 0x54F8, 0x5488, 0x5488, 0x57FE, 0x5500, 0x5D48, 0x6584, 0x100, 0x1FE, 0x0}},
    {23874, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x40, 0x13F8, 0x2048, 0x47FE, 0x8848, 0x13F8, 0x3040, 0x53FC, 0x9040, 0x17FE, 0x1040, 0x1040}},
    {23875, 2, {0x1020, 0x1010, 0x11FE, 0x1000, 0x5484, 0x5448, 0x55FE, 0x5508, 0x5510, 0x5564, 0x5508, 0x5D12, 0x6564, 0x208, 0x230, 0x4C0}},
    {23876, 2, {0x2104, 0x2088, 0x23FE, 0x2020, 0xA820, 0xA9FC, 0xA820, 0xA820, 0xABFE, 0xA800, 0xA820, 0xBBFE, 0xC820, 0x50, 0x88, 0x306}},
    {23877, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2024, 0x2FA8, 0x2898, 0x2892, 0x4FAA, 0x4046, 0x8082}},
    {23878, 2, {0xC48, 0x7044, 0x1040, 0x13FE, 0xFC50, 0x3890, 0x5490, 0x9112, 0x1212, 0x110E, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23879, 2, {0x848, 0x1C44, 0xF040, 0x13FE, 0x1050, 0xFC50, 0x1092, 0x3892, 0x350E, 0x5420, 0x5020, 0x9124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {23880, 2, {0x203C, 0x27C0, 0x2244, 0x2128, 0xA800, 0xABFC, 0xA880, 0xA880, 0xAFFE, 0xA900, 0xA9F8, 0xBA88, 0xCA50, 0x420, 0x8D8, 0x306}},
    {23881, 2, {0x2040, 0x2020, 0x23FE, 0x2000, 0xA9FC, 0xA904, 0xA9FC, 0xA800, 0xABFE, 0xAA02, 0xA9FC, 0xB820, 0xC820, 0x20, 0xA0, 0x40}},
    {23882, 2, {0x2038, 0x23C0, 0x2040, 0x27FC, 0xA950, 0xA954, 0xAF58, 0xA954, 0xAB54, 0xAD4C, 0xA8E0, 0xB950, 0xCA48, 0x444, 0x842, 0x40}},
    {23883, 2, {0x1000, 0x11FE, 0x1112, 0x1112, 0x55FE, 0x5500, 0x557E, 0x5542, 0x5542, 0x557E, 0x5542, 0x5D7E, 0x6242, 0x242, 0x47E, 0x842}},
    {23884, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x20, 0x2220, 0x223E, 0xFF42, 0x2294, 0x2210, 0x3E10, 0x2210, 0x2228, 0x3E28, 0x2244, 0x82}},
    {23885, 2, {0x40, 0x7F40, 0x127E, 0xC88, 0xFF48, 0x1948, 0x2A50, 0x4820, 0xA850, 0x118C, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23886, 2, {0x21FC, 0x2124, 0x2124, 0x21FC, 0xA924, 0xA924, 0xA9FC, 0xA820, 0xA820, 0xABFE, 0xAA22, 0xBA2A, 0xCAFA, 0x20A, 0x202, 0x206}},
    {23887, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x1010, 0x2208, 0x47E4, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {23888, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x10D0, 0x1710, 0x1110, 0x17D0, 0x1550, 0x27CA, 0x212A, 0x4FE6, 0x8022}},
    {23889, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x5504, 0x55FC, 0x5480, 0x55FE, 0x5622, 0x5522, 0x5552, 0x5D02, 0x65FA, 0x2, 0x14, 0x8}},
    {23890, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7C7C, 0x4444, 0x4444, 0x7D7C, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23891, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7EFC, 0x4284, 0x4284, 0x4284, 0x7EFC, 0x4284}},
    {23892, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE, 0x1110, 0x10A0, 0x1440, 0x1830, 0x100E}},
    {23893, 2, {0x2000, 0x2088, 0x2252, 0x2222, 0xAA52, 0xAA8A, 0xAA02, 0xABFE, 0xA888, 0xA944, 0xAA7A, 0xB888, 0xC950, 0x20, 0xD8, 0x706}},
    {23894, 2, {0x1020, 0x1020, 0x13FE, 0x1070, 0x54A8, 0x5524, 0x5622, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5DFC, 0x6504, 0x0, 0x3FE, 0x0}},
    {23895, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x48, 0x44, 0x3FFE, 0x2040, 0x2F44, 0x2244, 0x2A28, 0x2AA8, 0x3112, 0x422A, 0x44C6, 0x9B02}},
    {23896, 2, {0x2088, 0x2088, 0x23FE, 0x2088, 0xA800, 0xABFE, 0xAA02, 0xAC24, 0xA820, 0xABFE, 0xA870, 0xB8A8, 0xC8A8, 0x124, 0x222, 0x20}},
    {23897, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0xC10, 0x7090, 0x1050, 0xFC90, 0x3050, 0x381E, 0x55F0, 0x5010, 0x9010, 0x1010, 0x1010}},
    {23898, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1040, 0x1040, 0x3E7C, 0x4084, 0xBD28, 0x1020, 0x1020, 0xFE50, 0x1050, 0x1488, 0x1904, 0x1202}},
    {23899, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x280, 0xC60, 0x37D8, 0xC006, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2228, 0x2610}},
    {23900, 2, {0x100, 0x4104, 0x7FFC, 0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x50, 0x20}},
    {23901, 2, {0x2020, 0x2124, 0x20A8, 0x2020, 0xABFE, 0xA8A8, 0xA924, 0xAA02, 0xA840, 0xAFFE, 0xA888, 0xB908, 0xC990, 0x60, 0x198, 0x604}},
    {23902, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x3FF8, 0x100, 0x1110, 0x2508, 0x200, 0x1010, 0x9292, 0x9292, 0xFEFE, 0x202}},
    {23903, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x3FFC, 0x2280, 0x2240, 0x27FC, 0x2C40, 0x37F8, 0x2440, 0x27F8, 0x2440, 0x4440, 0x47FC, 0x8400}},
    {23904, 2, {0x101E, 0x13E0, 0x1122, 0x1094, 0x5440, 0x5488, 0x55F0, 0x5420, 0x54C4, 0x55FE, 0x5422, 0x5C20, 0x63FE, 0x50, 0x88, 0x306}},
    {23905, 2, {0x100, 0x4104, 0x7FFC, 0x440, 0x820, 0x3218, 0xC446, 0xFE0, 0x20, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x2A54, 0x4488}},
    {23906, 2, {0x1020, 0x1020, 0x1050, 0x1048, 0x54A4, 0x55FE, 0x5684, 0x54FC, 0x5484, 0x54FC, 0x5480, 0x5CFC, 0x6544, 0x144, 0x27C, 0x44}},
    {23907, 2, {0x1020, 0x1010, 0x11FE, 0x1110, 0x557C, 0x5514, 0x55FE, 0x5514, 0x557C, 0x5510, 0x557C, 0x5D44, 0x6544, 0x244, 0x27C, 0x444}},
    {23908, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23909, 2, {0x2104, 0x2104, 0x21DE, 0x2244, 0xAD54, 0xA89E, 0xA904, 0xAA24, 0xAC24, 0xABFE, 0xA870, 0xB8A8, 0xC924, 0x622, 0x20, 0x20}},
    {23910, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x55FC, 0x5400, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x5504, 0x5DFC, 0x6488, 0x50, 0x3FE, 0x0}},
    {23911, 2, {0x2040, 0x23BE, 0x2212, 0x2292, 0xAA52, 0xAAAA, 0xAB24, 0xA840, 0xABFE, 0xAA22, 0xAA22, 0xBBFE, 0xCA22, 0x222, 0x3FE, 0x202}},
    {23912, 2, {0x2040, 0x2080, 0x21FC, 0x2104, 0xA904, 0xA9FC, 0xA900, 0xA9FE, 0xA900, 0xA9FE, 0xA802, 0xBAAA, 0xCAAA, 0x402, 0x14, 0x8}},
    {23913, 2, {0x100, 0x4104, 0x7FFC, 0x200, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {23914, 2, {0x2040, 0x2020, 0x23FC, 0x2000, 0xA9F8, 0xA908, 0xA9F8, 0xA800, 0xABFC, 0xAA04, 0xAAF4, 0xBA94, 0xCAF4, 0x204, 0x214, 0x208}},
    {23915, 2, {0x2084, 0x2044, 0x2048, 0x23FE, 0xA884, 0xA884, 0xA908, 0xA94A, 0xAA52, 0xAB9C, 0xA884, 0xB908, 0xC908, 0x252, 0x3DE, 0x42}},
    {23916, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x400, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x290, 0x4A8, 0x8FA, 0x308A, 0xC07E}},
    {23917, 2, {0x2040, 0x2020, 0x23FC, 0x2108, 0xA890, 0xABFE, 0xAA02, 0xAC44, 0xA820, 0xABFC, 0xA880, 0xB8F8, 0xC888, 0x108, 0x128, 0x210}},
    {23918, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5D04, 0x63FE, 0x88, 0x104, 0x202}},
    {23919, 2, {0x1108, 0x1088, 0x1090, 0x13FC, 0x5440, 0x55F8, 0x5440, 0x57FE, 0x5480, 0x5500, 0x55FC, 0x5E20, 0x6420, 0x820, 0x3FE, 0x0}},
    {23920, 2, {0x1088, 0x1050, 0x1000, 0x11FE, 0x5450, 0x55FC, 0x5454, 0x57FE, 0x5454, 0x55FC, 0x5450, 0x5CD8, 0x6554, 0x252, 0x50, 0x50}},
    {23921, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xAC94, 0xA908, 0xAA64, 0xA890, 0xA908, 0xAA04, 0xADFA, 0xB908, 0xC908, 0x108, 0x1F8, 0x108}},
    {23922, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5420, 0x57FE, 0x5C70, 0x64A8, 0x124, 0x222, 0x20}},
    {23923, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x800, 0x37F0, 0xC080, 0x1FFC, 0x0}},
    {23924, 2, {0x2040, 0x2248, 0x2150, 0x2248, 0xACA4, 0xA918, 0xAE06, 0xABF8, 0xAA08, 0xABF8, 0xAA08, 0xBBF8, 0xCA08, 0x208, 0x228, 0x210}},
    {23925, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x40, 0x7840, 0x4BFC, 0x4840, 0x4FFE, 0x7810, 0x4FFE, 0x4810, 0x4A10, 0x7910, 0x4850, 0x20}},
    {23926, 2, {0x2000, 0x23DE, 0x2042, 0x2042, 0xABDE, 0xAA10, 0xAA10, 0xABDE, 0xA842, 0xAA52, 0xA94A, 0xB94A, 0xCA52, 0x42, 0x294, 0x108}},
    {23927, 2, {0x2124, 0x2124, 0x2224, 0x2424, 0xA954, 0xA94A, 0xAA92, 0xAE10, 0xAA10, 0xAA50, 0xAA5C, 0xBA50, 0xCA50, 0x2B0, 0x29E, 0x300}},
    {23928, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1110, 0x2110, 0x4110, 0x8AA8, 0x1444, 0x3040, 0x5240, 0x927C, 0x1240, 0x1540, 0x14FE, 0x1800}},
    {23929, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x7CF8, 0x488, 0x4F8, 0x7C20, 0x41FC, 0x4124, 0x7D24, 0x5FC, 0x420, 0x424, 0x2BFE, 0x1002}},
    {23930, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xAA94, 0xACA4, 0xA850, 0xA988, 0xAE26, 0xA8C0, 0xAB10, 0xB864, 0xCB88, 0x30, 0xC0, 0x700}},
    {23931, 2, {0x1020, 0x1010, 0x11FE, 0x1110, 0x557C, 0x5514, 0x55FE, 0x5514, 0x557C, 0x5510, 0x5592, 0x5D54, 0x6538, 0x254, 0x292, 0x430}},
    {23932, 2, {0x2040, 0x2020, 0x23FC, 0x2000, 0xA908, 0xA890, 0xABFE, 0xAA20, 0xAAA0, 0xAAFC, 0xAB20, 0xBA20, 0xCAFC, 0x420, 0x420, 0x9FE}},
    {23933, 2, {0x2154, 0x2154, 0x23FE, 0x2154, 0xA954, 0xAA72, 0xAC00, 0xABFE, 0xAA22, 0xA820, 0xA9FC, 0xB924, 0xC924, 0x124, 0x12C, 0x20}},
    {23934, 2, {0x100, 0x4104, 0x7FFC, 0x820, 0x1FF0, 0x1020, 0x2548, 0x7EFC, 0x440, 0x1930, 0xE64E, 0x1990, 0x620, 0x18C0, 0x700, 0x3800}},
    {23935, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x0, 0xFDFC, 0x1040, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x5090, 0x2108}},
    {23936, 2, {0x11FC, 0x1020, 0x13FE, 0x1222, 0x55AC, 0x5420, 0x55AC, 0x5400, 0x55FC, 0x5400, 0x57FE, 0x5C80, 0x64FC, 0x4, 0x28, 0x10}},
    {23937, 2, {0x1020, 0x11FC, 0x1124, 0x13FE, 0x5524, 0x55FC, 0x5420, 0x55FC, 0x5524, 0x55FC, 0x5440, 0x5FFE, 0x6488, 0x1D0, 0x70, 0x38C}},
    {23938, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5C20, 0x67FE, 0x20, 0x20, 0x20}},
    {23939, 2, {0x2200, 0x2202, 0x223C, 0x2FA0, 0xAA20, 0xAFA0, 0xAABE, 0xAFA4, 0xAAA4, 0xAFA4, 0xAA24, 0xBFA4, 0xCA24, 0x244, 0x244, 0x284}},
    {23940, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x806, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x908}},
    {23941, 2, {0x820, 0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x8100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23942, 2, {0x2090, 0x2090, 0x27FE, 0x2090, 0xABFC, 0xAA94, 0xABFC, 0xAA94, 0xABFC, 0xA800, 0xA9F8, 0xB908, 0xC9F8, 0x108, 0x1F8, 0x108}},
    {23943, 2, {0x2000, 0x23FE, 0x2200, 0x227C, 0xAA44, 0xAA44, 0xAA7C, 0xAA00, 0xAAEE, 0xAAAA, 0xAAAA, 0xBAAA, 0xCAEE, 0x200, 0x3FE, 0x0}},
    {23944, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x448, 0x7DFE, 0x2488, 0x2448, 0x2448, 0x4408, 0x8418}},
    {23945, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0xA0, 0x90, 0x79FE, 0x4920, 0x4B20, 0x4DFC, 0x4920, 0x49FC, 0x7920, 0x4920, 0x1FE, 0x100}},
    {23946, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x10A0, 0x1090, 0xF9FE, 0x1120, 0x1B20, 0x15FC, 0x3120, 0xD1FC, 0x1120, 0x1120, 0x51FE, 0x2100}},
    {23947, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0x55FC, 0x5504, 0x55FC, 0x5500, 0x55FE, 0x5500, 0x55FE, 0x5C02, 0x62AA, 0x2AA, 0x202, 0xC}},
    {23948, 2, {0x100, 0x4104, 0x7FFC, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4924, 0x8494, 0x8}},
    {23949, 2, {0x1000, 0x13DE, 0x1042, 0x1252, 0x554A, 0x5652, 0x5442, 0x5420, 0x55FC, 0x5504, 0x5504, 0x5DFC, 0x6504, 0x104, 0x1FC, 0x104}},
    {23950, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3F08, 0x2108, 0x3F08, 0x20FE, 0x2E08, 0x2048, 0x3F28, 0x4428, 0x5508, 0xA4A8, 0xC10}},
    {23951, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x7FFE, 0x4000, 0x5FDC, 0x4504, 0x5FC4, 0x5544, 0x59DC, 0x5050, 0x5FD0, 0x5052, 0x5FD2, 0x904E}},
    {23952, 2, {0x100, 0x4104, 0x7FFC, 0x100, 0x79F0, 0x4A90, 0x5060, 0x6198, 0x5606, 0x49F0, 0x4A40, 0x6BF8, 0x5440, 0x43F8, 0x4040, 0x47FC}},
    {23953, 2, {0x2080, 0x20F8, 0x2108, 0x23FE, 0xAD12, 0xA922, 0xA9FE, 0xA840, 0xA8A2, 0xAB54, 0xA898, 0xBB34, 0xC854, 0x92, 0x350, 0x20}},
    {23954, 2, {0x2104, 0x2088, 0x2000, 0x23FE, 0xAA22, 0xAAAA, 0xAA72, 0xAA22, 0xABFE, 0xA800, 0xA9FC, 0xB904, 0xC9FC, 0x104, 0x1FC, 0x104}},
    {23955, 2, {0x103C, 0x13E0, 0x1124, 0x10A8, 0x55FE, 0x54A8, 0x5524, 0x5602, 0x55FC, 0x5524, 0x5524, 0x5DFC, 0x6524, 0x124, 0x1FC, 0x104}},
    {23956, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x840, 0x1440, 0x227C, 0x5C84, 0x8928, 0x3E20, 0x820, 0x4A50, 0x2850, 0xE88, 0x7104, 0x2202}},
    {23957, 2, {0x2090, 0x2088, 0x21FE, 0x2110, 0xAB10, 0xADFE, 0xA910, 0xA910, 0xA9FE, 0xA910, 0xA910, 0xB9FE, 0xC900, 0x2A4, 0x252, 0x452}},
    {23958, 2, {0x2040, 0x20A0, 0x2118, 0x26E6, 0xA800, 0xABF8, 0xAA08, 0xABF8, 0xA800, 0xAFBC, 0xA884, 0xBCA4, 0xC294, 0x4A4, 0x294, 0x108}},
    {23959, 2, {0x2088, 0x22AA, 0x22DC, 0x2488, 0xA954, 0xAA22, 0xA800, 0xABFE, 0xAA42, 0xAC44, 0xABFC, 0xB844, 0xC844, 0x84, 0x114, 0x208}},
    {23960, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1090, 0x1088, 0x13FE, 0xFC48, 0x1032, 0x31CE, 0x3890, 0x5488, 0x53FE, 0x9048, 0x1032, 0x11CE}},
    {23961, 2, {0x2020, 0x2124, 0x20A8, 0x23FE, 0xA8A8, 0xA924, 0xAA22, 0xA904, 0xA904, 0xA9DE, 0xAA44, 0xBD54, 0xC89E, 0x104, 0x204, 0x404}},
    {23962, 2, {0x100, 0x4104, 0x7FFC, 0x100, 0x7FFC, 0x440, 0x2828, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x100, 0x1110, 0x2508, 0x200}},
    {23963, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0x5554, 0x5488, 0x5504, 0x56FA, 0x5488, 0x54F8, 0x5488, 0x5CF8, 0x6420, 0xA8, 0x124, 0x60}},
    {23964, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x3E7C, 0x810, 0x4890, 0x7EFE, 0x1428, 0x244A, 0x4686, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8}},
    {23965, 2, {0x2020, 0x27A4, 0x20A8, 0x2292, 0xA914, 0xAA08, 0xADF4, 0xA802, 0xABF8, 0xAA08, 0xAA08, 0xBBF8, 0xCA08, 0x110, 0xFFE, 0x0}},
    {23966, 2, {0x40, 0x7BFC, 0x4880, 0x51F8, 0x6220, 0x55FC, 0x4888, 0x48F8, 0x6888, 0x50F8, 0x4088, 0x4100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23967, 2, {0x2088, 0x2050, 0x23FE, 0x2050, 0xA9FC, 0xA954, 0xA98C, 0xA974, 0xA904, 0xA9FC, 0xA808, 0xBBFE, 0xC908, 0x88, 0xA8, 0x10}},
    {23968, 2, {0x1018, 0x11E0, 0x1040, 0x13FE, 0x5488, 0x5574, 0x5652, 0x5470, 0x5400, 0x55FC, 0x5504, 0x5D74, 0x6554, 0x174, 0x104, 0x10C}},
    {23969, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x3FFE, 0x2890, 0x2510, 0x2FBE, 0x2222, 0x2AD4, 0x2A90, 0x2F90, 0x2228, 0x4428, 0x4444, 0x8882}},
    {23970, 2, {0x2020, 0x2020, 0x21FC, 0x2020, 0xABFE, 0xA908, 0xAB9C, 0xA908, 0xA988, 0xAE3E, 0xA800, 0xBBFE, 0xC890, 0x90, 0x112, 0x20E}},
    {23971, 2, {0x100, 0x4104, 0x7FFC, 0x100, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x2488, 0x2244, 0x4244}},
    {23972, 2, {0x100, 0x4104, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x820, 0x3EF8, 0x820, 0x7EFC, 0x0, 0x7FFC, 0x440, 0x1842, 0xE03E}},
    {23973, 2, {0x2000, 0x2FFE, 0x2800, 0x2A28, 0xA948, 0xABEE, 0xA892, 0xA884, 0xAAA0, 0xAAA8, 0xABE8, 0xB888, 0xC894, 0x914, 0x924, 0x1242}},
    {23974, 2, {0x1080, 0x10F8, 0x1108, 0x13FE, 0x5544, 0x5592, 0x55FE, 0x5500, 0x557C, 0x5500, 0x557C, 0x5D00, 0x657C, 0x244, 0x27C, 0x444}},
    {23975, 2, {0x1000, 0x11FC, 0x1154, 0x11FC, 0x5420, 0x55FC, 0x5420, 0x57FE, 0x5488, 0x5450, 0x55FC, 0x5C20, 0x67FE, 0x20, 0x20, 0x20}},
    {23976, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {23977, 2, {0x1050, 0x11FC, 0x1154, 0x11FC, 0x5554, 0x55FC, 0x5400, 0x55FE, 0x5500, 0x557C, 0x5500, 0x5DFE, 0x6550, 0x154, 0x148, 0x266}},
    {23978, 2, {0x100, 0x4104, 0x7FFC, 0x2448, 0x1450, 0xFFFE, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E}},
    {23979, 2, {0x2050, 0x2252, 0x2154, 0x2050, 0xABFE, 0xA888, 0xA850, 0xABFE, 0xA820, 0xA9FC, 0xA820, 0xBBFE, 0xC8A8, 0x124, 0x622, 0x20}},
    {23980, 2, {0x2104, 0x2088, 0x23FE, 0x2020, 0xA9FC, 0xA820, 0xABFE, 0xA854, 0xA992, 0xA890, 0xABFE, 0xB890, 0xC8D4, 0x38A, 0x96, 0x1A2}},
    {23981, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x1040, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x4020, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x4420}},
    {23982, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0x5504, 0x56FA, 0x5400, 0x5400, 0x55DC, 0x5554, 0x5554, 0x5DDC, 0x6488, 0x88, 0x154, 0x222}},
    {23983, 2, {0x4104, 0x7FFC, 0x0, 0x3E28, 0x2224, 0x3E24, 0x20, 0xFFFE, 0x2220, 0x3E28, 0x2228, 0x3E10, 0x2212, 0x2F2A, 0xF246, 0x4282}},
    {23984, 2, {0x2100, 0x21DE, 0x224A, 0x248A, 0xABEA, 0xAAB6, 0xAAA4, 0xABF4, 0xAABE, 0xAAA4, 0xABE4, 0xBABE, 0xCAA4, 0x2A4, 0x224, 0x464}},
    {23985, 2, {0x1088, 0x13FE, 0x1088, 0x11FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5480, 0x55FE, 0x5622, 0x5D52, 0x6502, 0x1FA, 0xA, 0x4}},
    {23986, 2, {0x100, 0x4104, 0x7FFC, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3E7C, 0x2244, 0x23D4, 0x2008}},
    {23987, 2, {0x2108, 0x3FF8, 0x0, 0x7FFC, 0x600, 0x3B08, 0x4B0, 0x19C0, 0x62A0, 0xC98, 0x7326, 0x1020, 0x7CFC, 0x1020, 0x1E20, 0xF1FE}},
    {23988, 2, {0x400, 0x3FF8, 0x2008, 0x2928, 0x2548, 0x2FE8, 0x2548, 0x2928, 0x200, 0xFFFE, 0x820, 0x3118, 0xD116, 0x1110, 0x1FF0, 0x10}},
    {23989, 2, {0x2000, 0x23FC, 0x2294, 0x2294, 0xABFC, 0xA800, 0xA890, 0xAF9E, 0xA890, 0xA890, 0xAB9C, 0xB890, 0xC890, 0x79E, 0x90, 0x90}},
    {23990, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x2220, 0x4AA0, 0x8ABE, 0x1FC4, 0x20A4, 0x6F24, 0xA028, 0x2F28, 0x2910, 0x29A8, 0x2944, 0x3082}},
    {23991, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x247C, 0x3A04, 0x2228, 0x1E10, 0x20FE, 0x3E12, 0x4850, 0x85E, 0x7F50, 0x14B0, 0x229E, 0x4100}},
    {23992, 2, {0x2088, 0x22AA, 0x22DC, 0x2488, 0xA954, 0xAA22, 0xA800, 0xABFE, 0xAA22, 0xA820, 0xABFE, 0xB870, 0xC8A8, 0x124, 0x622, 0x20}},
    {23993, 2, {0x2020, 0x23FE, 0x2020, 0x21FC, 0xA800, 0xABFE, 0xA802, 0xA9FC, 0xA820, 0xABFE, 0xA800, 0xBBFE, 0xC804, 0x3BE, 0x2A4, 0x38C}},
    {23994, 2, {0x100, 0x4104, 0x7FFC, 0x1000, 0x13FC, 0x2840, 0x45F8, 0x9308, 0x9F8, 0xFD08, 0x5F8, 0x908, 0x51F8, 0x2000, 0x1090, 0x108}},
    {23995, 2, {0x100, 0x4104, 0x7FFC, 0x820, 0x7F28, 0x824, 0xFFFE, 0x1220, 0x3FA4, 0x6424, 0xBFA8, 0x2418, 0x3F92, 0x242A, 0x3FC6, 0x2082}},
    {23996, 2, {0x2140, 0x224C, 0x2264, 0x2244, 0xAB4C, 0xAA64, 0xAAA4, 0xABAC, 0xAAA4, 0xAAA4, 0xAFFE, 0xB800, 0xC910, 0x108, 0x204, 0x404}},
    {23997, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x10, 0x5218, 0x2F94, 0x5014, 0x977E, 0x1010, 0x3728, 0x5028, 0x9728, 0x1544, 0x5744, 0x2082}},
    {23998, 2, {0x203E, 0x27C0, 0x2244, 0x2128, 0xABF8, 0xA840, 0xAFFC, 0xA800, 0xABF8, 0xA808, 0xABF8, 0xB808, 0xCBF8, 0x544, 0x52A, 0x8FA}},
    {23999, 2, {0x2000, 0x21FC, 0x2020, 0x23FE, 0xAA22, 0xA9AC, 0xA820, 0xA9AC, 0xA800, 0xABFE, 0xA820, 0xB9FC, 0xC954, 0x154, 0x154, 0x10C}},
    {24000, 2, {0x100, 0x4104, 0x7FFC, 0x800, 0x2A28, 0x4D24, 0x820, 0x142E, 0x3F70, 0x6424, 0xBF24, 0x2428, 0x3F12, 0x242A, 0x3F46, 0x2082}},
    {24001, 2, {0x2000, 0x23FE, 0x2248, 0x23FE, 0xAA48, 0xAAFC, 0xAAA4, 0xAAFC, 0xAAA4, 0xAAFC, 0xAA20, 0xBBFE, 0xC32A, 0x57A, 0x50A, 0x906}},
    {24002, 2, {0x2108, 0x3FF8, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3FFC, 0x2424, 0x2BD4, 0x2244, 0x23CC}},
    {24003, 2, {0x4104, 0x7FFC, 0x840, 0x7F7C, 0x2240, 0x147C, 0xFF04, 0x7C, 0x7F40, 0x417C, 0x7F40, 0x417C, 0x7F40, 0x417C, 0x4542, 0x423E}},
    {24004, 2, {0x2110, 0x2090, 0x23DE, 0x2010, 0xAA5E, 0xA982, 0xABDE, 0xA810, 0xABDE, 0xAA50, 0xABDE, 0xBA50, 0xCBDE, 0x250, 0x252, 0x2CE}},
    {24005, 2, {0x2108, 0x3FF8, 0x800, 0x7EFE, 0x810, 0x3E20, 0x227C, 0x3E44, 0x2254, 0x3E54, 0x2254, 0x3E54, 0x2254, 0xFF28, 0x2444, 0x4282}},
    {24006, 2, {0x2088, 0x22AA, 0x22DC, 0x2488, 0xA954, 0xAA22, 0xABFE, 0xAA02, 0xA8F8, 0xA888, 0xA8F8, 0xB800, 0xC9FC, 0x104, 0x1FC, 0x104}},
    {24007, 2, {0x2088, 0x20CC, 0x208A, 0x23E8, 0xAAA8, 0xAADE, 0xABA8, 0xAA6A, 0xAA0A, 0xABEA, 0xAC0C, 0xBFEC, 0xCA2A, 0x3EA, 0x156, 0x7E2}},
    {24008, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x2900, 0x7DFC, 0x2904, 0x3A44, 0x1154, 0x7CE4, 0x55F4, 0x7C44, 0x10E4, 0xFD54, 0x1054, 0x1008}},
    {24009, 2, {0x21F0, 0x2210, 0x27FC, 0x2204, 0xABFC, 0xAA24, 0xABB8, 0xAA22, 0xA9FE, 0xAA10, 0xAFFC, 0xBA44, 0xCBFC, 0xD0, 0x14A, 0x63E}},
    {24010, 2, {0x23DE, 0x2252, 0x23DE, 0x2252, 0xABDE, 0xAA52, 0xABDE, 0xA98C, 0xAA52, 0xA840, 0xAFFE, 0xB888, 0xC990, 0x60, 0xD8, 0x304}},
    {24011, 2, {0x100, 0x4104, 0x7FFC, 0x1000, 0x7DFC, 0x4404, 0x7CFC, 0x4004, 0x7DFC, 0x4400, 0x7DFE, 0x1122, 0x5CFC, 0x50A4, 0x5EAC, 0xF020}},
    {24012, 2, {0x4104, 0x7FFC, 0x2208, 0x1110, 0x3FFC, 0x2000, 0x2E20, 0x223C, 0x3FA8, 0x2948, 0x2F28, 0x2928, 0x2F10, 0x49D0, 0x7F28, 0x8144}},
    {24013, 2, {0x100, 0x4104, 0x7FFC, 0xC20, 0x7040, 0x11FC, 0xFD24, 0x39FC, 0x5524, 0x93FC, 0xFC24, 0x245A, 0x685E, 0x1090, 0x2912, 0xC60E}},
    {24014, 2, {0x2000, 0x27FE, 0x2040, 0x22EE, 0xAAAA, 0xABEA, 0xAAAE, 0xAAE8, 0xABAA, 0xACE6, 0xAA88, 0xBDF4, 0xCA90, 0x60, 0x198, 0x606}},
    {24015, 2, {0x2088, 0x23FE, 0x2088, 0x23DE, 0xAA52, 0xABDE, 0xA8A0, 0xA890, 0xA9FE, 0xA920, 0xABFC, 0xBD20, 0xC9FC, 0x120, 0x1FE, 0x100}},
    {24016, 2, {0x100, 0x4104, 0x7FFC, 0x800, 0x7F7C, 0x44, 0x3E7C, 0x2244, 0x3E7C, 0x10, 0x7F64, 0x4124, 0x5DFE, 0x5524, 0x5D24, 0x4344}},
    {24017, 2, {0x2528, 0x27BE, 0x2948, 0x27BE, 0xAB18, 0xADAA, 0xA946, 0xABFC, 0xAA04, 0xABFC, 0xAA04, 0xBBFC, 0xCA04, 0x3FC, 0x108, 0x204}},
    {24018, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {24019, 2, {0x100, 0x4104, 0x7FFC, 0x2200, 0x3CFE, 0x2110, 0x1F7C, 0x5244, 0x5E7C, 0x5244, 0x5E7C, 0x5244, 0x7F7C, 0x28, 0x2444, 0x4282}},
    {24020, 2, {0x2108, 0x3FF8, 0x800, 0x7EFE, 0x810, 0x3E20, 0x227C, 0x3E44, 0x227C, 0x3E44, 0x227C, 0x3E44, 0x227C, 0xFF28, 0x2444, 0x4282}},
    {24021, 2, {0x2108, 0x2090, 0x27FE, 0x2108, 0xABFE, 0xAA54, 0xABC8, 0xAA3E, 0xABC8, 0xAA5C, 0xABC8, 0xB840, 0xCBFE, 0x88, 0x70, 0x38C}},
    {24022, 2, {0x100, 0x4104, 0x7FFC, 0x0, 0x1E78, 0x1248, 0x3FFC, 0x2020, 0x2E20, 0x223C, 0x3FC8, 0x2928, 0x2F28, 0x4990, 0x5F28, 0x8144}},
    {24023, 2, {0x23DE, 0x2252, 0x23DE, 0x2000, 0xABFE, 0xAA00, 0xAAC8, 0xAA4E, 0xABEA, 0xAAB2, 0xAAEA, 0xBAAA, 0xCAEA, 0x2B4, 0x5E4, 0x82A}},
    {24024, 2, {0x2108, 0x21EC, 0x210A, 0x2FEA, 0xA928, 0xABC8, 0xA93E, 0xAFE8, 0xA948, 0xABE8, 0xAA28, 0xBB68, 0xCAA8, 0xBF4, 0xAB4, 0x12A2}},
    {24025, 2, {0x2110, 0x27FE, 0x2040, 0x22EE, 0xAAAA, 0xABEA, 0xAAAE, 0xAAE8, 0xABAA, 0xACE6, 0xAA88, 0xBDF4, 0xCA90, 0x60, 0x198, 0x606}},
    {24026, 2, {0x100, 0x4104, 0x7FFC, 0x800, 0xF90, 0x814, 0x7F92, 0x48FE, 0x7E10, 0x4890, 0x5FA8, 0x5128, 0x5F28, 0x5544, 0x5544, 0x9382}},
    {24027, 2, {0x0, 0x0, 0x888, 0x888, 0x888, 0x1110, 0x1110, 0x2220, 0x2220, 0x1110, 0x1110, 0x888, 0x888, 0x888, 0x0, 0x0}},
    {24028, 2, {0x0, 0x410, 0x410, 0x820, 0x820, 0x1040, 0x2080, 0x4100, 0x2080, 0x1040, 0x820, 0x820, 0x410, 0x410, 0x0, 0x0}},
    {24029, 2, {0x1004, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x1084, 0x2084, 0x2084, 0x4004, 0x8004}},
    {24030, 2, {0x1004, 0x1084, 0x1084, 0x1084, 0x1084, 0x54A4, 0x5294, 0x5294, 0x9084, 0x1084, 0x1084, 0x1084, 0x2084, 0x2084, 0x4004, 0x8004}},
    {24031, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0, 0x1110, 0x1110, 0x1110, 0x1110, 0x2110, 0x2112, 0x4112, 0x800E}},
    {24032, 2, {0x0, 0x7FFE, 0x0, 0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {24033, 2, {0x0, 0x2124, 0x1124, 0x1124, 0x248, 0x248, 0xF490, 0x1248, 0x1248, 0x1124, 0x1124, 0x1124, 0x1000, 0x2800, 0x47FE, 0x0}},
    {24034, 2, {0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {24035, 2, {0x2208, 0x1108, 0x1110, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {24036, 2, {0x1084, 0x2108, 0x1084, 0x3FF8, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x0, 0x2948, 0x2528, 0x2948, 0x252A, 0x210A, 0x2946, 0x3182}},
    {24037, 2, {0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0, 0x0}},
    {24038, 2, {0x200, 0x200, 0x200, 0xFFFE, 0x400, 0x400, 0x400, 0x800, 0xFF8, 0x1080, 0x1080, 0x2080, 0x4080, 0x8080, 0x3FFE, 0x0}},
    {24039, 2, {0x0, 0x3FE, 0xFC40, 0x1040, 0x1080, 0x1080, 0x1100, 0x11FC, 0x1004, 0x1004, 0x1E04, 0xE004, 0x4004, 0x4, 0x28, 0x10}},
    {24040, 2, {0x0, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x3FF0, 0x2010, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FFC, 0x0}},
    {24041, 2, {0x0, 0x1F0, 0x7D10, 0x1110, 0x1110, 0x1110, 0x1190, 0x1150, 0x1150, 0x1110, 0x1D10, 0xE112, 0x4212, 0x212, 0x40E, 0x800}},
    {24042, 2, {0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x2000, 0x3FFC, 0x0, 0x7FF8, 0x8, 0x8, 0x8}},
    {24043, 2, {0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1110, 0x1110, 0x1110, 0x2928, 0x2928, 0x4544, 0x8584, 0x100, 0x100, 0xFFFE, 0x0}},
    {24044, 2, {0x0, 0x7DF0, 0x1110, 0x1110, 0x1112, 0x1D12, 0xE20E, 0x4400, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0x600C}},
    {24045, 2, {0x40, 0x7C40, 0x13FC, 0x1044, 0x1044, 0x1C84, 0xE114, 0x4208, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0x600C}},
    {24046, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x200, 0x200, 0xFFFE, 0x400, 0x800, 0x17F8, 0x2080, 0x4080, 0x8080, 0x1FFC}},
    {24047, 2, {0x20, 0x7C10, 0x5FE, 0x820, 0x1048, 0x2884, 0x45FE, 0x8202, 0xA8, 0x7CA8, 0x10A8, 0x10A8, 0x10AA, 0x1D2A, 0xE126, 0x4200}},
    {24048, 2, {0x20, 0xFE10, 0x55FE, 0x5420, 0xA848, 0x5484, 0x55FE, 0x2, 0xFEA8, 0x10A8, 0x10A8, 0x10A8, 0x1EAA, 0xF12A, 0x4126, 0x200}},
    {24049, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x3FF0, 0x2000, 0x2000, 0x2000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {24050, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x2000, 0x2000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {24051, 2, {0x0, 0x3FF0, 0x2010, 0x2010, 0x2010, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x2000, 0x2000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {24052, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x2000, 0x2000, 0x2002, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {24053, 2, {0x7C, 0x3F80, 0x2000, 0x2000, 0x3FFE, 0x2000, 0x2000, 0x2FF8, 0x2888, 0x2888, 0x2888, 0x2FF8, 0x2808, 0x4802, 0x4802, 0x87FE}},
    {24054, 2, {0x0, 0xFEFC, 0x2284, 0x2284, 0x4284, 0x4A84, 0x8484, 0xFC, 0x7E80, 0x4280, 0x4280, 0x4282, 0x4282, 0x7E82, 0x427E, 0x0}},
    {24055, 2, {0x440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x1FF0, 0x2828, 0xC826, 0xFE0, 0x820, 0x808, 0x808, 0x7F8}},
    {24056, 2, {0x0, 0x7F7C, 0x4844, 0x4844, 0x4844, 0x5E44, 0x5244, 0x527C, 0x5240, 0x5240, 0x5E40, 0x4842, 0x4842, 0x4842, 0x7F3E, 0x0}},
    {24057, 2, {0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3118, 0xC506, 0x200, 0x3FF8, 0x0, 0x3FF0, 0x10, 0x3FF0, 0x2004, 0x2004, 0x1FFC}},
    {24058, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x7CF8, 0x4080, 0x4284, 0x4284, 0x3E7C, 0x0, 0x0, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {24059, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC026, 0x20, 0xFE0, 0x800, 0x808, 0x808, 0x7F8}},
    {24060, 2, {0x0, 0x3FF0, 0x2110, 0x2110, 0x3FF0, 0x2000, 0x2004, 0x2004, 0x1FFC, 0x0, 0x3E44, 0x2258, 0x2260, 0x2242, 0x3E42, 0x223E}},
    {24061, 2, {0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4080, 0x4284, 0x3E7C, 0x440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {24062, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100, 0x100}},
    {24063, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100, 0x100}},
    {24064, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100, 0x100}},
    {24065, 2, {0x8, 0x1FC, 0x7F00, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100, 0x100}},
    {24066, 2, {0x200, 0x100, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24067, 2, {0x200, 0x200, 0x200, 0xFFFE, 0x400, 0x900, 0x1100, 0x3FF8, 0x5108, 0x9108, 0x1108, 0x1108, 0x1128, 0x1110, 0x100, 0x100}},
    {24068, 2, {0x800, 0x800, 0x8FE, 0x7F10, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x4910, 0x4B10, 0x810, 0x810, 0x850, 0x820}},
    {24069, 2, {0x820, 0x820, 0x4820, 0x4820, 0x49FC, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x934, 0x1128, 0x1020, 0x2020, 0x4020}},
    {24070, 2, {0x1000, 0x11F0, 0x1110, 0x7D10, 0x5510, 0x5510, 0x5590, 0x5550, 0x5550, 0x5510, 0x5510, 0x5D12, 0x1112, 0x1212, 0x120E, 0x1400}},
    {24071, 2, {0x100, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24072, 2, {0x800, 0xBFE, 0x4820, 0x4820, 0x4820, 0x49FC, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x934, 0x1128, 0x1020, 0x2020, 0x4020}},
    {24073, 2, {0x1010, 0x1090, 0x1090, 0x7C88, 0x5508, 0x5504, 0x5604, 0x55FA, 0x5488, 0x5488, 0x5488, 0x5C88, 0x1108, 0x1108, 0x1228, 0x1410}},
    {24074, 2, {0x1000, 0x1000, 0x11FC, 0x7D24, 0x5524, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5500, 0x5500, 0x5D00, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {24075, 2, {0x78, 0x3F80, 0x2080, 0x2080, 0x3FFC, 0x2040, 0x2822, 0x311A, 0x2106, 0x1FF0, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {24076, 2, {0x1818, 0x660, 0x180, 0x660, 0x3A18, 0x200, 0xFFFE, 0x480, 0x880, 0x1FF8, 0x2888, 0x4888, 0x88A8, 0x890, 0x80, 0x80}},
    {24077, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2040, 0x2040, 0x27FC, 0x2444, 0x2444, 0x2444, 0x4454, 0x4448, 0x8040, 0x40}},
    {24078, 2, {0x1020, 0x1020, 0x1020, 0x7DFE, 0x5522, 0x5624, 0x5420, 0x5420, 0x5450, 0x5450, 0x5450, 0x5C50, 0x1090, 0x1092, 0x1112, 0x120E}},
    {24079, 2, {0x1020, 0x1020, 0x1020, 0x7DFE, 0x5420, 0x5420, 0x54FC, 0x5420, 0x5420, 0x55FE, 0x5422, 0x5C22, 0x102A, 0x1024, 0x1020, 0x1020}},
    {24080, 2, {0x1080, 0x1084, 0x1084, 0x7C88, 0x5490, 0x54A0, 0x5480, 0x57FE, 0x54A0, 0x5490, 0x5490, 0x5C88, 0x1084, 0x10A2, 0x10C0, 0x1080}},
    {24081, 2, {0x1000, 0x1000, 0xFEFC, 0x2244, 0x6428, 0x1810, 0x2428, 0xC346, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24082, 2, {0x8A0, 0x1090, 0x30BE, 0x57C0, 0x9040, 0x1024, 0x1014, 0x110C, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24083, 2, {0x1020, 0x1020, 0x1020, 0x7DFE, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5470, 0x54A8, 0x54A8, 0x5D24, 0x1124, 0x1222, 0x1020, 0x1020}},
    {24084, 2, {0x1010, 0x1010, 0x1010, 0x7DFE, 0x5512, 0x5514, 0x5510, 0x55FC, 0x5544, 0x5544, 0x5528, 0x5D28, 0x1110, 0x1228, 0x1244, 0x1482}},
    {24085, 2, {0x1020, 0x1020, 0x1040, 0x7DFC, 0x5504, 0x5504, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5504, 0x5D04, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24086, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x543E, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5504, 0x5504, 0x5D04, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24087, 2, {0x1050, 0x1048, 0x1048, 0x7C40, 0x55FE, 0x5440, 0x5440, 0x547C, 0x54A4, 0x54A4, 0x54A8, 0x5D28, 0x1110, 0x1228, 0x1044, 0x1082}},
    {24088, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2108, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24089, 2, {0x1020, 0x1020, 0x1120, 0x7D20, 0x55FC, 0x5520, 0x5620, 0x5420, 0x57FE, 0x5420, 0x5450, 0x5C50, 0x1088, 0x1088, 0x1104, 0x1202}},
    {24090, 2, {0x0, 0x1FF0, 0x10, 0xFF0, 0x10, 0x1FF0, 0x0, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1110, 0x1150, 0x1120, 0x100}},
    {24091, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24092, 2, {0x1000, 0x1000, 0x11FC, 0x7D04, 0x5504, 0x5504, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5400, 0x5C90, 0x1088, 0x1104, 0x1202, 0x1402}},
    {24093, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {24094, 2, {0x1000, 0x13FE, 0x1020, 0x7C20, 0x5440, 0x55FC, 0x5504, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5D04, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24095, 2, {0x200, 0x100, 0x7FFC, 0x440, 0x2450, 0x4448, 0x8944, 0x1080, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24096, 2, {0x200, 0x1CF0, 0x1010, 0x1010, 0x1EF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24097, 2, {0x1104, 0x1084, 0x1088, 0x7C00, 0x55FE, 0x5488, 0x5488, 0x5488, 0x5488, 0x57FE, 0x5488, 0x5C88, 0x1108, 0x1108, 0x1208, 0x1408}},
    {24098, 2, {0x1020, 0x1020, 0x1050, 0x7C88, 0x5504, 0x5602, 0x55FC, 0x5400, 0x5400, 0x55FC, 0x5504, 0x5D04, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24099, 2, {0x1110, 0x910, 0x920, 0x7FFC, 0x200, 0x400, 0xFFFE, 0x1110, 0x2108, 0xDFF6, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {24100, 2, {0x1000, 0x1000, 0xFE7C, 0x2244, 0x4244, 0x3444, 0x87C, 0x3400, 0xC200, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100}},
    {24101, 2, {0x1020, 0x2020, 0x7C20, 0x4420, 0x45FC, 0x4524, 0x7D24, 0x4124, 0x4124, 0x7D24, 0x4524, 0x4534, 0x4528, 0x7C20, 0x4420, 0x20}},
    {24102, 2, {0x920, 0x920, 0x7FFC, 0x920, 0x920, 0x0, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {24103, 2, {0x1020, 0x1020, 0x103E, 0x7C20, 0x5420, 0x55FC, 0x5504, 0x5524, 0x5524, 0x5524, 0x5524, 0x5D24, 0x1050, 0x1048, 0x1084, 0x1104}},
    {24104, 2, {0x1104, 0x1084, 0x1088, 0x7C10, 0x55FC, 0x5504, 0x5504, 0x5504, 0x55FC, 0x5450, 0x5450, 0x5C90, 0x1092, 0x1112, 0x120E, 0x1400}},
    {24105, 2, {0x1020, 0x1124, 0x10A4, 0x7CA8, 0x5420, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x5504, 0x5DFC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {24106, 2, {0x1000, 0x11FE, 0x1100, 0x7D00, 0x557C, 0x5500, 0x5500, 0x55FE, 0x5550, 0x5552, 0x5554, 0x5D48, 0x1148, 0x1244, 0x1252, 0x1460}},
    {24107, 2, {0x1000, 0x23FE, 0x7C20, 0x4420, 0x4420, 0x45FC, 0x7D24, 0x4124, 0x4124, 0x7D24, 0x4524, 0x4534, 0x4528, 0x7C20, 0x4420, 0x20}},
    {24108, 2, {0x0, 0x3FF8, 0x208, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0xCFF8, 0x80, 0x1FFC, 0x1084, 0x1094, 0x1088, 0x80}},
    {24109, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x23E0, 0x2080, 0x2FF8, 0x2888, 0x4888, 0x48A8, 0x8890, 0x80}},
    {24110, 2, {0x800, 0x87C, 0xFF48, 0x850, 0x7E48, 0x844, 0xFF54, 0x1048, 0x2040, 0x4100, 0xBFF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100}},
    {24111, 2, {0x920, 0x920, 0x7FFC, 0x920, 0xFE0, 0x0, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {24112, 2, {0x800, 0x9FC, 0x4804, 0x48FC, 0x4804, 0x49FC, 0x4800, 0x4BFE, 0x4A22, 0x4820, 0x49FC, 0x924, 0x1124, 0x1134, 0x2128, 0x4020}},
    {24113, 2, {0x2040, 0x2040, 0x27FC, 0xF840, 0xABF8, 0xA880, 0xAFFE, 0xA890, 0xA910, 0xA9FE, 0xAA10, 0xBA90, 0x2450, 0x2810, 0x2050, 0x2020}},
    {24114, 2, {0x1044, 0x1044, 0x1088, 0x7DDC, 0x5488, 0x5488, 0x5488, 0x5488, 0x55DE, 0x5488, 0x5488, 0x5C88, 0x1088, 0x1088, 0x1088, 0x1108}},
    {24115, 2, {0x1000, 0x10FC, 0x1080, 0x7CF8, 0x5480, 0x54F8, 0x5480, 0x57FE, 0x5540, 0x5524, 0x5528, 0x5D10, 0x1108, 0x1144, 0x1182, 0x1100}},
    {24116, 2, {0x1048, 0x1044, 0x105E, 0x7DE0, 0x5428, 0x5412, 0x546A, 0x5596, 0x5448, 0x545E, 0x55E0, 0x5C24, 0x1028, 0x1012, 0x106A, 0x1186}},
    {24117, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0xAD04, 0xA900, 0xA9DE, 0xAA52, 0xAA52, 0xAB52, 0xAC9A, 0xB894, 0x2110, 0x2112, 0x2212, 0x240E}},
    {24118, 2, {0x1450, 0x1450, 0x7FFC, 0x1450, 0x2452, 0x47CE, 0x8000, 0x7FFC, 0x4104, 0x1FF0, 0x1110, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {24119, 2, {0x1050, 0x1048, 0x1080, 0x7CFE, 0x5590, 0x5690, 0x54FC, 0x5490, 0x5490, 0x54FC, 0x5490, 0x5C90, 0x1090, 0x10FE, 0x1080, 0x1080}},
    {24120, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {24121, 2, {0x2080, 0x2040, 0x23FC, 0xF800, 0xA908, 0xA890, 0xAFFE, 0xA840, 0xA840, 0xAFFE, 0xA888, 0xB908, 0x2090, 0x2060, 0x2198, 0x2604}},
    {24122, 2, {0x1088, 0x1088, 0x11FC, 0x7C88, 0x5488, 0x54F8, 0x5488, 0x5488, 0x54F8, 0x5488, 0x5488, 0x5DFE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {24123, 2, {0x2040, 0x2040, 0x27FC, 0xF840, 0xABF8, 0xA840, 0xAFFE, 0xA800, 0xABF8, 0xAA08, 0xAA48, 0xBA48, 0x2248, 0x20A0, 0x2110, 0x2608}},
    {24124, 2, {0x2000, 0x23FE, 0x2202, 0xFA02, 0xABFE, 0xAA22, 0xAA22, 0xAAFA, 0xAA22, 0xAA32, 0xAA2A, 0xBBFE, 0x2202, 0x2202, 0x23FE, 0x2202}},
    {24125, 2, {0x2000, 0x23FC, 0x2204, 0xFAF4, 0xAA04, 0xAAF4, 0xAA04, 0xA800, 0xA9F8, 0xA908, 0xA9F8, 0xB908, 0x21F8, 0x2108, 0x21F8, 0x2108}},
    {24126, 2, {0x1020, 0x1022, 0x11FA, 0x7C24, 0x5428, 0x57FE, 0x5420, 0x5440, 0x54FC, 0x5584, 0x5684, 0x5CFC, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {24127, 2, {0x2100, 0x2178, 0x2108, 0xFA08, 0xAAFE, 0xAE40, 0xAA40, 0xAA7C, 0xAA90, 0xAA10, 0xAAFE, 0xBA10, 0x2228, 0x2228, 0x2244, 0x2282}},
    {24128, 2, {0x1020, 0x1020, 0x103E, 0x7C20, 0x5420, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5DFC, 0x1000, 0x1088, 0x1104, 0x1202}},
    {24129, 2, {0x1000, 0x13FE, 0x1020, 0x7C40, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5D04, 0x11FC, 0x1000, 0x1088, 0x1104}},
    {24130, 2, {0x0, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0xFFFE, 0x920, 0x3FF8, 0xD116, 0x1150, 0x1120, 0x100}},
    {24131, 2, {0x1040, 0x1040, 0x11FC, 0x7C84, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5420, 0x57FE, 0x5C20, 0x1220, 0x13FE, 0x1020, 0x1020}},
    {24132, 2, {0x2000, 0x23FC, 0x2204, 0xFA04, 0xABFC, 0xAA00, 0xABFC, 0xAA40, 0xAA88, 0xABFC, 0xAA24, 0xBA20, 0x25FC, 0x2420, 0x2820, 0x33FE}},
    {24133, 2, {0x1000, 0x11FE, 0x1000, 0x7CF8, 0x5488, 0x5488, 0x54F8, 0x5400, 0x55FC, 0x5524, 0x5524, 0x5DFC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {24134, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5480, 0x55FE, 0x5622, 0x5522, 0x5552, 0x5D02, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {24135, 2, {0x808, 0x808, 0x7E08, 0x8FE, 0xFE08, 0x848, 0x7E28, 0x808, 0xF28, 0xF110, 0x100, 0x3FF8, 0x2108, 0x2128, 0x2110, 0x100}},
    {24136, 2, {0x2000, 0x23FC, 0x2204, 0xFA04, 0xABFC, 0xAA00, 0xAA84, 0xAA48, 0xAAFC, 0xAA48, 0xAA48, 0xBBFE, 0x2448, 0x2448, 0x2888, 0x3108}},
    {24137, 2, {0x2048, 0x2148, 0x2148, 0xFBFE, 0xA948, 0xA948, 0xA978, 0xA900, 0xA9FE, 0xA820, 0xABFE, 0xB870, 0x20A8, 0x2124, 0x2622, 0x2020}},
    {24138, 2, {0x1000, 0x11FC, 0x1020, 0x7C20, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5D04, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {24139, 2, {0x1078, 0x3E48, 0x2248, 0x2A86, 0xFF78, 0x2248, 0x2A28, 0x4210, 0x4A28, 0x8546, 0x100, 0x3FF8, 0x2108, 0x2128, 0x2110, 0x100}},
    {24140, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5420, 0x5524, 0x54A8, 0x5420, 0x57FE, 0x5C50, 0x1050, 0x1092, 0x1112, 0x120E}},
    {24141, 2, {0x101E, 0x11E0, 0x1022, 0x7D12, 0x5494, 0x5480, 0x5420, 0x55CE, 0x5502, 0x5502, 0x55CE, 0x5D02, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {24142, 2, {0x2000, 0x23FE, 0x2202, 0xF800, 0xA9FC, 0xA904, 0xA9FC, 0xA904, 0xA9FC, 0xA840, 0xA820, 0xBBFE, 0x2000, 0x2088, 0x2104, 0x2202}},
    {24143, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0xA800, 0xA9FC, 0xA840, 0xA8A2, 0xAB34, 0xA858, 0xA894, 0xBB34, 0x2052, 0x2090, 0x2350, 0x2020}},
    {24144, 2, {0x20, 0x7924, 0x48A8, 0x49FC, 0x4840, 0x7BFE, 0x4908, 0x4A24, 0x4C22, 0x79FC, 0x4924, 0x4924, 0x4934, 0x4928, 0x4820, 0x9820}},
    {24145, 2, {0x1210, 0x1A90, 0x2A90, 0x4AA0, 0x8FBE, 0x1044, 0x2FA4, 0x6224, 0xA224, 0x2FA8, 0x2AA8, 0x2A90, 0x2AA8, 0x2BA8, 0x2244, 0x2282}},
    {24146, 2, {0x1020, 0x1040, 0x11FC, 0x7D24, 0x557C, 0x558C, 0x5554, 0x5524, 0x5554, 0x55FC, 0x5440, 0x5C24, 0x12A2, 0x128A, 0x1488, 0x1078}},
    {24147, 2, {0x2040, 0x2088, 0x21FC, 0xF908, 0xAA52, 0xABFE, 0xA850, 0xA988, 0xAE26, 0xA8C0, 0xAB10, 0xB864, 0x2388, 0x2030, 0x20C0, 0x2700}},
    {24148, 2, {0x2000, 0x21F8, 0x2108, 0xF9F8, 0xA908, 0xA9F8, 0xA800, 0xABFC, 0xAA94, 0xABFC, 0xA800, 0xB9F8, 0x2090, 0x2060, 0x2198, 0x2606}},
    {24149, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1110, 0x3FF8, 0xD116, 0x1150, 0x1120, 0x100}},
    {24150, 2, {0x1000, 0x13FE, 0x1050, 0x7DFC, 0x5554, 0x5554, 0x55FC, 0x5400, 0x55FC, 0x5400, 0x57FE, 0x5C20, 0x10A8, 0x1124, 0x12A2, 0x1040}},
    {24151, 2, {0x2000, 0x23FE, 0x221A, 0xFA16, 0xABFE, 0xAA12, 0xAAEA, 0xAAAA, 0xAAAA, 0xAAEA, 0xAA0E, 0xBA6A, 0x239A, 0x2226, 0x23FE, 0x2202}},
    {24152, 2, {0x1020, 0x13FE, 0x1020, 0x7DFC, 0x5420, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5DFC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {24153, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5420, 0x5DFE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {24154, 2, {0x87C, 0xFF48, 0x850, 0x7E48, 0x844, 0xFF54, 0x1048, 0x2240, 0x5FF0, 0x9010, 0x1FF0, 0x1010, 0x3FF8, 0x2108, 0x2118, 0x100}},
    {24155, 2, {0x1020, 0x11FC, 0x1088, 0x7C50, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5C20, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {24156, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5420, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x5DFC, 0x1020, 0x1124, 0x1222, 0x1060}},
    {24157, 2, {0x2000, 0x23DE, 0x2252, 0xFA52, 0xABDE, 0xA800, 0xA9FC, 0xA924, 0xA9FC, 0xA924, 0xA9FC, 0xB820, 0x23FE, 0x2020, 0x2020, 0x2020}},
    {24158, 2, {0x2050, 0x2252, 0x2154, 0xF850, 0xABFE, 0xA888, 0xA850, 0xABFE, 0xA820, 0xA9FC, 0xA820, 0xBBFE, 0x2050, 0x2088, 0x2104, 0x2602}},
    {24159, 2, {0x2110, 0x2094, 0x23D2, 0xF812, 0xAA50, 0xA990, 0xAFFE, 0xA810, 0xABD2, 0xAA52, 0xAA54, 0xBBD4, 0x224A, 0x224A, 0x23D6, 0x2022}},
    {24160, 2, {0x2080, 0x2080, 0x21FE, 0xFB54, 0xAD54, 0xA954, 0xABFE, 0xA954, 0xA954, 0xA954, 0xAFFE, 0xB800, 0x2154, 0x212A, 0x222A, 0x2000}},
    {24161, 2, {0x2078, 0x27C0, 0x2248, 0xF950, 0xABFC, 0xA950, 0xAA48, 0xAC06, 0xABF8, 0xAA48, 0xAA48, 0xBBF8, 0x2248, 0x2248, 0x23F8, 0x2208}},
    {24162, 2, {0x1020, 0x11FC, 0x1088, 0x7C50, 0x57FE, 0x5400, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x55FC, 0x5C20, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {24163, 2, {0x4920, 0x2A20, 0x7F3E, 0x4948, 0x5DA8, 0x6B28, 0x4910, 0x4928, 0x4046, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100}},
    {24164, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x41A8, 0x5D28, 0x5510, 0x5D28, 0x4246, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110, 0x100}},
    {24165, 2, {0x2124, 0x20A8, 0x23FE, 0xFA02, 0xA8F8, 0xA888, 0xA8F8, 0xA80C, 0xA9F0, 0xA820, 0xA9FC, 0xB820, 0x23FE, 0x2020, 0x20A0, 0x2040}},
    {24166, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3D20, 0x100, 0x3FF8, 0x2108, 0x2128, 0x2110, 0x100}},
    {24167, 2, {0x10F8, 0x1088, 0x1088, 0x7CF8, 0x5400, 0x55DC, 0x5554, 0x5554, 0x55DC, 0x5420, 0x57FE, 0x5C70, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {24168, 2, {0x2100, 0x21F8, 0x2208, 0xFFFE, 0xAA88, 0xAB24, 0xABFE, 0xAA00, 0xAAFC, 0xAA00, 0xAAFC, 0xBA00, 0x22FC, 0x2484, 0x24FC, 0x2884}},
    {24169, 2, {0x1020, 0x11FC, 0x1020, 0x7C88, 0x57FE, 0x5488, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5D04, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {24170, 2, {0x2088, 0x23FE, 0x2088, 0xFBFE, 0xAA02, 0xA9FC, 0xA800, 0xABFE, 0xA840, 0xA8A2, 0xAB54, 0xB8B8, 0x2354, 0x2092, 0x2350, 0x2020}},
    {24171, 2, {0x808, 0x3EFE, 0x808, 0x7F48, 0x828, 0x3E08, 0x928, 0x7FF0, 0x2820, 0xFE0, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x2118}},
    {24172, 2, {0x2020, 0x23FE, 0x2020, 0xF9FC, 0xA800, 0xABFE, 0xA802, 0xA9FC, 0xA820, 0xABFE, 0xA800, 0xBBFE, 0x2004, 0x23BE, 0x22A4, 0x238C}},
    {24173, 2, {0x2108, 0x27FE, 0x2108, 0xF800, 0xABFC, 0xAA94, 0xAA94, 0xABFC, 0xA828, 0xABFE, 0xAA20, 0xBB24, 0x22A8, 0x2292, 0x242A, 0x28C6}},
    {24174, 2, {0x2040, 0x2020, 0x23FE, 0xFA44, 0xABF4, 0xAA44, 0xAAEE, 0xAA04, 0xAAF4, 0xAAAC, 0xAAE4, 0xBA14, 0x22A4, 0x2474, 0x2594, 0x2808}},
    {24175, 2, {0x2108, 0x2108, 0x21EE, 0xFA94, 0xA842, 0xA800, 0xABDE, 0xAA52, 0xABD2, 0xAA52, 0xABD2, 0xBA1A, 0x2294, 0x2350, 0x2210, 0x2010}},
    {24176, 2, {0x1020, 0x13FE, 0x1222, 0x7DFC, 0x5420, 0x54F8, 0x5420, 0x55FC, 0x5420, 0x55FC, 0x5554, 0x5DFC, 0x1040, 0x12A4, 0x128A, 0x147A}},
    {24177, 2, {0x23DE, 0x2252, 0x23DE, 0xFA52, 0xABDE, 0xAA22, 0xAAFA, 0xAA22, 0xAAFA, 0xAAAA, 0xAAFA, 0xBAAA, 0x22FA, 0x2272, 0x22AA, 0x2226}},
    {24178, 2, {0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {24179, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x1110, 0x910, 0x920, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {24180, 2, {0x1000, 0x1000, 0x1FFC, 0x2080, 0x2080, 0x4080, 0x1FF8, 0x1080, 0x1080, 0x1080, 0xFFFE, 0x80, 0x80, 0x80, 0x80, 0x80}},
    {24181, 2, {0x0, 0x7C7C, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0xFEFE, 0x1010, 0x1010, 0x1010, 0x1010, 0x2010, 0x2010, 0x4010, 0x8010}},
    {24182, 2, {0x1010, 0x810, 0x820, 0x0, 0x7FFC, 0x820, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {24183, 2, {0x808, 0x808, 0x1010, 0x7C7C, 0x1010, 0x1010, 0x1010, 0x1010, 0xFCFE, 0x1010, 0x1010, 0x1010, 0x2010, 0x2010, 0x4010, 0x8010}},
    {24184, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x820, 0x440, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {24185, 2, {0x1020, 0x1020, 0x1050, 0xFE50, 0x1088, 0x7D04, 0x46FA, 0x7C20, 0x4420, 0x7C20, 0x11FC, 0xFE20, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24186, 2, {0x200, 0x200, 0x400, 0x810, 0x1010, 0x2020, 0x7FC0, 0x2080, 0x100, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008, 0x0}},
    {24187, 2, {0x800, 0x800, 0x11FC, 0x1004, 0x2204, 0x4204, 0xFC04, 0x404, 0x804, 0x1004, 0x2404, 0x4204, 0xFE04, 0x4204, 0x28, 0x10}},
    {24188, 2, {0x1040, 0x1040, 0x2040, 0x2040, 0x45FC, 0x4444, 0xF844, 0x844, 0x1044, 0x1084, 0x2884, 0x4484, 0xFD04, 0x4504, 0x228, 0x410}},
    {24189, 2, {0x100, 0x100, 0x920, 0x920, 0x5144, 0x5554, 0x7DF4, 0x4924, 0x4924, 0x5554, 0x7DF4, 0x4514, 0x4104, 0x4104, 0x7FFC, 0x4}},
    {24190, 2, {0x1110, 0x2120, 0x4544, 0x7978, 0x1110, 0x2524, 0x7D7C, 0x1524, 0x1110, 0xFFFE, 0x1080, 0x1088, 0x2852, 0x2422, 0x405A, 0x8186}},
    {24191, 2, {0x100, 0x80, 0x80, 0x3FFC, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {24192, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2200, 0x2204, 0x2218, 0x2260, 0x2380, 0x2200, 0x2202, 0x2202, 0x4202, 0x41FE, 0x8000}},
    {24193, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2000, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x4040, 0x4040, 0x8140, 0x80}},
    {24194, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2080, 0x2080, 0x2080, 0x2140, 0x2140, 0x2220, 0x4220, 0x4410, 0x8808, 0x1006}},
    {24195, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2080, 0x2100, 0x2100, 0x2100, 0x2210, 0x2208, 0x4408, 0x4FFC, 0x8404, 0x4}},
    {24196, 2, {0x100, 0x80, 0x80, 0x3FFC, 0x2000, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x2080, 0x2080, 0x2080, 0x4080, 0x5FFE, 0x8000}},
    {24197, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2100, 0x2100, 0x2220, 0x2220, 0x2440, 0x2880, 0x2100, 0x2208, 0x4404, 0x4FFE, 0x8402, 0x0}},
    {24198, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x2140, 0x2140, 0x2220, 0x4220, 0x4410, 0x8808, 0x1006}},
    {24199, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2420, 0x2420, 0x2424, 0x2428, 0x2730, 0x2420, 0x2420, 0x2422, 0x4522, 0x4622, 0x841E, 0x0}},
    {24200, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2140, 0x2220, 0x2410, 0x2908, 0x3086, 0x2080, 0x2FF0, 0x2010, 0x4020, 0x4020, 0x8040, 0x80}},
    {24201, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2040, 0x2FFE, 0x2040, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2042, 0x2042, 0x4042, 0x403E, 0x8000}},
    {24202, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x21C0, 0x22A0, 0x22A0, 0x2490, 0x4888, 0x5086, 0x8080, 0x80}},
    {24203, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2080, 0x3FFC, 0x2080, 0x2080, 0x2FF0, 0x2410, 0x2220, 0x2140, 0x4080, 0x4160, 0x8618, 0x3806}},
    {24204, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2FFC, 0x2020, 0x2420, 0x2820, 0x2FFE, 0x20A0, 0x2120, 0x4620, 0x5820, 0x80A0, 0x40}},
    {24205, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x201C, 0x27E0, 0x2400, 0x2400, 0x27FE, 0x2420, 0x2420, 0x2420, 0x4820, 0x4820, 0x9020, 0x2020}},
    {24206, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2D16, 0x2110, 0x2110, 0x2110, 0x4110, 0x4210, 0x8210, 0x410}},
    {24207, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x23F8, 0x2010, 0x20A0, 0x2040, 0x2FFE, 0x2042, 0x2044, 0x2040, 0x4040, 0x4040, 0x8140, 0x80}},
    {24208, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2080, 0x27F8, 0x2408, 0x2408, 0x2408, 0x27F8, 0x2408, 0x2400, 0x4400, 0x4800, 0x8800, 0x1000}},
    {24209, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2FFC, 0x2080, 0x2080, 0x3FFE, 0x2140, 0x2140, 0x2240, 0x4242, 0x4442, 0x883E, 0x1000}},
    {24210, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x2080, 0x2090, 0x2088, 0x4088, 0x4080, 0x9FFE, 0x0}},
    {24211, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2100, 0x3FFC, 0x2200, 0x2480, 0x2880, 0x2FF8, 0x2080, 0x2080, 0x5FFE, 0x4080, 0x8080, 0x80}},
    {24212, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2104, 0x2884, 0x2484, 0x2448, 0x2248, 0x2210, 0x2210, 0x4020, 0x4040, 0x9FFE, 0x0}},
    {24213, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x203C, 0x27C0, 0x2440, 0x2440, 0x2440, 0x27FE, 0x2420, 0x2420, 0x2412, 0x450A, 0x4646, 0x8422}},
    {24214, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2200, 0x27F8, 0x2808, 0x37C8, 0x2448, 0x2448, 0x27C8, 0x2428, 0x2412, 0x4402, 0x43FE, 0x8000}},
    {24215, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2080, 0x2080, 0x20FC, 0x2080, 0x2080, 0x27F8, 0x2408, 0x2408, 0x4408, 0x4408, 0x87F8, 0x408}},
    {24216, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x4040, 0x4040, 0x8040}},
    {24217, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2080, 0x2FF8, 0x2888, 0x2888, 0x2888, 0x2FF8, 0x2888, 0x4888, 0x4888, 0x8FF8, 0x808}},
    {24218, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2FF8, 0x2088, 0x2088, 0x3FFE, 0x2088, 0x2088, 0x2FF8, 0x4140, 0x4220, 0x8C18, 0x3006}},
    {24219, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2110, 0x2110, 0x2512, 0x2514, 0x25D8, 0x2510, 0x2510, 0x2510, 0x2512, 0x45D2, 0x5E12, 0x880E}},
    {24220, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2208, 0x2208, 0x2408, 0x24FE, 0x2C08, 0x3488, 0x2448, 0x2448, 0x4408, 0x4408, 0x8428, 0x410}},
    {24221, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2100, 0x23F8, 0x2610, 0x2920, 0x20C0, 0x2330, 0x3C0E, 0x20C0, 0x4020, 0x4300, 0x80C0, 0x20}},
    {24222, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2210, 0x2200, 0x2FFC, 0x2240, 0x2240, 0x2248, 0x2450, 0x2460, 0x4842, 0x48C2, 0x933E, 0x2000}},
    {24223, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2490, 0x2888, 0x2FFE, 0x2100, 0x21FC, 0x2284, 0x2288, 0x2450, 0x2420, 0x4850, 0x5188, 0x8606}},
    {24224, 2, {0x100, 0x80, 0x3FFE, 0x2410, 0x2220, 0x3FFC, 0x2080, 0x2080, 0x2FF8, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x4080, 0x4080, 0x8080}},
    {24225, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2040, 0x3FFE, 0x2080, 0x2100, 0x2284, 0x2688, 0x2A50, 0x3220, 0x4210, 0x4288, 0x8306, 0x200}},
    {24226, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0x2FFC, 0x2100, 0x2210, 0x27F8, 0x2088, 0x2080, 0x2FFC, 0x4080, 0x4080, 0xBFFE, 0x0}},
    {24227, 2, {0x100, 0x80, 0x3FFE, 0x2120, 0x2120, 0x2924, 0x2528, 0x2130, 0x2120, 0x2330, 0x2528, 0x2924, 0x2220, 0x4222, 0x4422, 0x881E}},
    {24228, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2040, 0x27FC, 0x2040, 0x2040, 0x3FFE, 0x2000, 0x2010, 0x2FFC, 0x4210, 0x4110, 0x8150, 0x20}},
    {24229, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2220, 0x2220, 0x2420, 0x2DFE, 0x3470, 0x24A8, 0x24A8, 0x2524, 0x2622, 0x4420, 0x4420, 0x8420}},
    {24230, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x23E0, 0x2000, 0x2FF0, 0x2410, 0x4220, 0x41C0, 0x8630, 0x380E}},
    {24231, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2490, 0x2490, 0x2490, 0x2AA8, 0x31C4, 0x2080, 0x2FF8, 0x2080, 0x4080, 0x4080, 0xBFFE, 0x0}},
    {24232, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2088, 0x27F0, 0x20A0, 0x3FFE, 0x2080, 0x2FF0, 0x2620, 0x3840, 0x47FC, 0x4040, 0x8140, 0x80}},
    {24233, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2140, 0x2220, 0x2410, 0x3BEC, 0x2080, 0x2080, 0x2FF8, 0x2080, 0x4490, 0x4888, 0x9284, 0x100}},
    {24234, 2, {0x100, 0x80, 0x3FFE, 0x2410, 0x2410, 0x247E, 0x3F10, 0x2410, 0x257C, 0x2644, 0x3C44, 0x2428, 0x4410, 0x4428, 0x9444, 0x882}},
    {24235, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x3FFC, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x2080, 0x5FFE, 0x4080, 0x8080, 0x80}},
    {24236, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2208, 0x3FFE, 0x2248, 0x2250, 0x2264, 0x2248, 0x2252, 0x2264, 0x444A, 0x4472, 0x88C2, 0x133E}},
    {24237, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2E0E, 0x22F0, 0x2410, 0x2410, 0x2EFE, 0x2210, 0x2210, 0x2AFE, 0x4400, 0x4600, 0x89FE, 0x1000}},
    {24238, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FFC, 0x2120, 0x2FFC, 0x2924, 0x2924, 0x2A1C, 0x2C04, 0x2FFC, 0x4804, 0x4804, 0x8FFC, 0x804}},
    {24239, 2, {0x100, 0x80, 0x3FFE, 0x2048, 0x2044, 0x2FFE, 0x2040, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x4444, 0x4454, 0x8408}},
    {24240, 2, {0x100, 0x80, 0x3FFE, 0x2208, 0x2108, 0x2110, 0x27FC, 0x2110, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x4110, 0x4210, 0x8410, 0x810}},
    {24241, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2080, 0x2FFC, 0x2080, 0x3FFE, 0x2408, 0x2904, 0x33F2, 0x2610, 0x4920, 0x40C0, 0x8730, 0x380E}},
    {24242, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2040, 0x2FFE, 0x2040, 0x2248, 0x2248, 0x2554, 0x28E2, 0x2150, 0x4248, 0x4C46, 0x8040, 0x40}},
    {24243, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2100, 0x27FC, 0x2444, 0x27FC, 0x2484, 0x27FC, 0x2120, 0x2220, 0x3FFE, 0x4020, 0x4020, 0x8020}},
    {24244, 2, {0x80, 0x3FFE, 0x2210, 0x2210, 0x2FFC, 0x2210, 0x2210, 0x3FFE, 0x2000, 0x27F8, 0x2408, 0x2408, 0x47F8, 0x4408, 0x8408, 0x7F8}},
    {24245, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0x2FFC, 0x2120, 0x2210, 0x2488, 0x2FF6, 0x3490, 0x27F0, 0x2490, 0x47F2, 0x4082, 0x807E}},
    {24246, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x2220, 0x23E0, 0x2000, 0x4A48, 0x4924, 0x9124, 0x0}},
    {24247, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2880, 0x24C4, 0x22A8, 0x4490, 0x4888, 0x9286, 0x100}},
    {24248, 2, {0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x4888, 0x4FF8, 0x8888, 0x898}},
    {24249, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0x3FFE, 0x2210, 0x23F0, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x4420, 0x4410, 0x8808, 0x1006}},
    {24250, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2488, 0x2924, 0x2210, 0x27F8, 0x2088, 0x2080, 0x2FFC, 0x21C0, 0x42A0, 0x4498, 0x9886, 0x80}},
    {24251, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x23E0, 0x2000, 0x2410, 0x2410, 0x4410, 0x4A28, 0x9144, 0x2082}},
    {24252, 2, {0x100, 0x80, 0x3FFE, 0x2800, 0x29FC, 0x2820, 0x2E40, 0x29FC, 0x2904, 0x2924, 0x2924, 0x2B24, 0x2D24, 0x4850, 0x4088, 0x8304}},
    {24253, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27F8, 0x2488, 0x27F8, 0x2488, 0x27F8, 0x2080, 0x2FFC, 0x2884, 0x48A4, 0x4BF4, 0x8914, 0x808}},
    {24254, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2280, 0x2CB8, 0x2888, 0x2888, 0x2EB8, 0x2888, 0x2888, 0x2FF8, 0x4140, 0x4220, 0x8C18, 0x3006}},
    {24255, 2, {0x100, 0x80, 0x3FFE, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2000, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x4444, 0x47FC, 0x8404}},
    {24256, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2FFE, 0x2842, 0x2248, 0x24B0, 0x2308, 0x2C04, 0x27F8, 0x2208, 0x2110, 0x40E0, 0x4318, 0x9C06}},
    {24257, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2F84, 0x2884, 0x28A4, 0x2FA4, 0x28A4, 0x2FA4, 0x28A4, 0x2FA4, 0x4024, 0x4904, 0x8894, 0x1088}},
    {24258, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2200, 0x227C, 0x3F44, 0x2244, 0x267C, 0x2744, 0x2AC4, 0x2AFC, 0x5244, 0x6244, 0x827C, 0x244}},
    {24259, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2F50, 0x2124, 0x2A18, 0x2408, 0x3BF6, 0x2120, 0x2120, 0x2FFC, 0x2120, 0x4224, 0x4424, 0x881C}},
    {24260, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2F78, 0x2948, 0x2948, 0x2F86, 0x2900, 0x297C, 0x2F44, 0x2844, 0x2A28, 0x4910, 0x4B28, 0x8D46}},
    {24261, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FFC, 0x2080, 0x3FFE, 0x2200, 0x2408, 0x2FFC, 0x2000, 0x2FFC, 0x4924, 0x4924, 0xBFFE, 0x0}},
    {24262, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2100, 0x2FFC, 0x2884, 0x2FFC, 0x2884, 0x2FFC, 0x2148, 0x2154, 0x425C, 0x4446, 0x8842, 0x303E}},
    {24263, 2, {0x80, 0x3FFE, 0x2080, 0x277C, 0x2424, 0x2524, 0x24A4, 0x2554, 0x2648, 0x2080, 0x27FC, 0x2444, 0x47FC, 0x4444, 0x87FC, 0x404}},
    {24264, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FFC, 0x2080, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2200, 0x27F8, 0x4A08, 0x41F0, 0x9E0E}},
    {24265, 2, {0x100, 0x80, 0x3FFE, 0x2208, 0x2110, 0x2FFE, 0x2120, 0x27FC, 0x2124, 0x3FFE, 0x2124, 0x27FC, 0x2330, 0x4528, 0x5926, 0x8120}},
    {24266, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2100, 0x2FBE, 0x28A2, 0x2FA4, 0x28A8, 0x2FA4, 0x2A22, 0x2922, 0x2AAA, 0x4CA4, 0x4820, 0x8020}},
    {24267, 2, {0x100, 0x80, 0x3FFE, 0x2140, 0x265C, 0x2444, 0x275C, 0x2444, 0x27FC, 0x2040, 0x2FFC, 0x2208, 0x2110, 0x40E0, 0x4318, 0x8C06}},
    {24268, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x3FFE, 0x2222, 0x3FFE, 0x2400, 0x27FC, 0x2400, 0x27FE, 0x2002, 0x224A, 0x4926, 0x500A, 0x8004}},
    {24269, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x221E, 0x2F92, 0x2014, 0x2898, 0x2514, 0x3FD2, 0x2012, 0x2F9A, 0x4894, 0x4890, 0x8F90, 0x890}},
    {24270, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x29FE, 0x2820, 0x28FC, 0x2E84, 0x28FC, 0x2884, 0x28FC, 0x2A84, 0x4CFC, 0x4800, 0x8048, 0x84}},
    {24271, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2FBC, 0x28A4, 0x2FA4, 0x28C2, 0x2F80, 0x207E, 0x2922, 0x2A22, 0x2C94, 0x4888, 0x4794, 0x8062}},
    {24272, 2, {0x80, 0x3FFE, 0x2000, 0x2F7E, 0x2910, 0x2910, 0x2F50, 0x2950, 0x297E, 0x2F10, 0x2828, 0x2A28, 0x4928, 0x4B4A, 0x8D4A, 0x886}},
    {24273, 2, {0x80, 0x3FFE, 0x2410, 0x3FFE, 0x2410, 0x27F0, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2080, 0x3FFC, 0x4080, 0x4FF8, 0x8080, 0x3FFE}},
    {24274, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F90, 0x221E, 0x2FA4, 0x2224, 0x3FD4, 0x2414, 0x2794, 0x2488, 0x4488, 0x4894, 0x8AA4, 0x1142}},
    {24275, 2, {0x80, 0x3FFE, 0x2200, 0x3FBC, 0x2024, 0x2FA4, 0x28A8, 0x2FA8, 0x2024, 0x2FA4, 0x2124, 0x23F4, 0x5E28, 0x4220, 0x8A20, 0x420}},
    {24276, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2888, 0x3FFE, 0x2888, 0x2FF8, 0x2490, 0x27F0, 0x2100, 0x3FFE, 0x4410, 0x43E0, 0x9C1C}},
    {24277, 2, {0x100, 0x80, 0x3FFE, 0x2010, 0x2F28, 0x2944, 0x2A92, 0x2C7C, 0x2A08, 0x2910, 0x297E, 0x2D00, 0x2AFE, 0x4820, 0x4844, 0x88FE}},
    {24278, 2, {0x80, 0x3FFE, 0x2000, 0x2F7C, 0x2924, 0x2514, 0x29A4, 0x2160, 0x2618, 0x3886, 0x2320, 0x2C40, 0x4188, 0x4E30, 0x81C0, 0xE00}},
    {24279, 2, {0x100, 0x80, 0x3FFE, 0x22A8, 0x22A8, 0x2FFC, 0x22AA, 0x24E6, 0x2800, 0x2FFE, 0x2842, 0x23F8, 0x4248, 0x4248, 0x8258, 0x40}},
    {24280, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2040, 0x2FFC, 0x2920, 0x2FFC, 0x2924, 0x2FFC, 0x2A20, 0x2A24, 0x4BB8, 0x5224, 0x92A4, 0x231C}},
    {24281, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2110, 0x27FC, 0x2110, 0x2FFE, 0x4000, 0x4208, 0x8404}},
    {24282, 2, {0x100, 0x80, 0x3FFE, 0x2204, 0x3FC4, 0x2204, 0x2FBE, 0x2004, 0x2FA4, 0x2894, 0x2F94, 0x2884, 0x4504, 0x4384, 0x9C14, 0x808}},
    {24283, 2, {0x80, 0x3FFE, 0x2000, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x2080, 0x2FF8, 0x2080, 0x3FFE, 0x4888, 0x53E4, 0x8080, 0x3FFE}},
    {24284, 2, {0x80, 0x3FFE, 0x2000, 0x27FC, 0x2404, 0x27FC, 0x2444, 0x25F8, 0x2450, 0x27FE, 0x2440, 0x24F8, 0x4988, 0x4AF8, 0x9488, 0xF8}},
    {24285, 2, {0x80, 0x3FFE, 0x2900, 0x290C, 0x3FB0, 0x2920, 0x2920, 0x2F3E, 0x2928, 0x2F28, 0x2928, 0x2928, 0x5FA8, 0x4048, 0x8A88, 0x1108}},
    {24286, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2220, 0x2520, 0x28BE, 0x3742, 0x2294, 0x2F90, 0x2210, 0x3290, 0x4A28, 0x43A8, 0x9C44, 0x882}},
    {24287, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2FBC, 0x2224, 0x2FA4, 0x28BC, 0x2FA4, 0x28A4, 0x2FBC, 0x2224, 0x5FA4, 0x4244, 0x8254, 0x288}},
    {24288, 2, {0x80, 0x7FFE, 0x4208, 0x5248, 0x4A50, 0x4A9E, 0x5FE4, 0x5054, 0x5754, 0x5554, 0x5548, 0x5748, 0x5048, 0x5154, 0x5094, 0x8022}},
    {24289, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2400, 0x2FFC, 0x3550, 0x2550, 0x2FFC, 0x2550, 0x2550, 0x3FFE, 0x4000, 0x4A48, 0x8924, 0x1124}},
    {24290, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2F50, 0x2124, 0x2A18, 0x2408, 0x3B76, 0x2150, 0x2F8C, 0x2878, 0x2F48, 0x4130, 0x4548, 0x8284}},
    {24291, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2FF8, 0x2220, 0x3FFC, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x4FF8, 0x4220, 0x8410, 0x808}},
    {24292, 2, {0x80, 0x3FFE, 0x2080, 0x3FFC, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x2080, 0x3FFC, 0x40A4, 0x5F3A, 0x9122, 0x1F1E}},
    {24293, 2, {0x80, 0x3FFE, 0x2140, 0x2630, 0x3BEC, 0x2000, 0x2FF8, 0x2AA8, 0x29C8, 0x2FF8, 0x2000, 0x27F0, 0x4410, 0x47F0, 0x8410, 0x7F0}},
    {24294, 2, {0x100, 0x80, 0x3FFE, 0x2010, 0x2788, 0x24BE, 0x2480, 0x27A2, 0x2494, 0x243E, 0x2F88, 0x2C88, 0x54BE, 0x4788, 0x8488, 0x8}},
    {24295, 2, {0x80, 0x3FFE, 0x2040, 0x27FC, 0x2248, 0x2248, 0x2554, 0x3FFE, 0x2000, 0x27FC, 0x2404, 0x25F4, 0x4514, 0x45F4, 0x8404, 0x7FC}},
    {24296, 2, {0x80, 0x3FFE, 0x2400, 0x2F3C, 0x3114, 0x2224, 0x2FCC, 0x2A80, 0x2FA8, 0x2ABE, 0x2FC8, 0x2A88, 0x4ABE, 0x4A88, 0x8888, 0x1188}},
    {24297, 2, {0x80, 0x3FFE, 0x2040, 0x2FFE, 0x2000, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x27FC, 0x2008, 0x23F0, 0x4040, 0x4FFE, 0x8150, 0xE4E}},
    {24298, 2, {0x80, 0x3FFE, 0x2040, 0x2FFE, 0x2000, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x27FC, 0x2000, 0x27FC, 0x4000, 0x4FFE, 0x8248, 0x4C4}},
    {24299, 2, {0x80, 0x3FFE, 0x2000, 0x2EEE, 0x2A22, 0x2AAA, 0x2A66, 0x2EAA, 0x2A10, 0x2A28, 0x2E54, 0x2AAA, 0x4A50, 0x5224, 0x9248, 0x2610}},
    {24300, 2, {0x80, 0x3FFE, 0x2040, 0x207C, 0x2040, 0x27FE, 0x2442, 0x25F8, 0x2444, 0x24FC, 0x24A8, 0x24F8, 0x44A8, 0x49FC, 0x8954, 0x13FE}},
    {24301, 2, {0x80, 0x3FFE, 0x2010, 0x237C, 0x2E10, 0x227C, 0x2F10, 0x22FE, 0x2644, 0x277C, 0x2AC4, 0x2A7C, 0x5244, 0x427C, 0x8228, 0x244}},
    {24302, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0x28A2, 0x2FBE, 0x28A2, 0x2FBE, 0x28A2, 0x2FBE, 0x2514, 0x28A2, 0x4FFE, 0x4310, 0x80E0, 0xF1C}},
    {24303, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2722, 0x2914, 0x323E, 0x2F88, 0x2A88, 0x2F9C, 0x2A88, 0x2F88, 0x403E, 0x5548, 0x9548, 0x8}},
    {24304, 2, {0x80, 0x7FFE, 0x4020, 0x7E20, 0x55FE, 0x5420, 0x5DFC, 0x5554, 0x5554, 0x5DFC, 0x5420, 0x5610, 0x5C54, 0x7542, 0x454A, 0x8638}},
    {24305, 2, {0x80, 0x3FFE, 0x2000, 0x2AA8, 0x3524, 0x2ABE, 0x2F68, 0x2968, 0x2FBE, 0x2028, 0x2FA8, 0x2ABE, 0x4FA8, 0x4828, 0x88BE, 0x7A0}},
    {24306, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0x2514, 0x271C, 0x2554, 0x2FFE, 0x2920, 0x2FFC, 0x2924, 0x2FFC, 0x4A24, 0x4BB8, 0x9222, 0x239E}},
    {24307, 2, {0x80, 0x7FFE, 0x4010, 0x7F10, 0x52FE, 0x5E10, 0x52FE, 0x5EAA, 0x52AA, 0x7EFE, 0x4200, 0x7EFE, 0x4A24, 0x5E12, 0x4AA4, 0xBE9C}},
    {24308, 2, {0x0, 0x0, 0xF800, 0x800, 0x1000, 0x1000, 0x2000, 0x7800, 0x800, 0x800, 0x4800, 0x3000, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {24309, 2, {0x0, 0x124, 0xF924, 0x924, 0x1248, 0x1248, 0x2490, 0x7A48, 0xA48, 0x924, 0x4924, 0x3124, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {24310, 2, {0x8, 0x3C, 0xF9E0, 0x820, 0x1020, 0x1120, 0x213C, 0x7920, 0x920, 0x920, 0x4920, 0x31FC, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {24311, 2, {0x8, 0x3C, 0xF9E0, 0x820, 0x1020, 0x1020, 0x21FC, 0x7820, 0x820, 0x820, 0x4820, 0x33FE, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {24312, 2, {0x20, 0x20, 0xF820, 0x9FC, 0x1124, 0x1124, 0x2124, 0x79FC, 0x924, 0x924, 0x4924, 0x31FC, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {24313, 2, {0x20, 0x20, 0xF840, 0x9FC, 0x1104, 0x1104, 0x2104, 0x79FC, 0x904, 0x904, 0x4904, 0x31FC, 0x1104, 0x2C00, 0x43FE, 0x8000}},
    {24314, 2, {0x40, 0x40, 0xFBF8, 0x848, 0x17FE, 0x1048, 0x23F8, 0x7840, 0xBF8, 0x840, 0x4840, 0x37FC, 0x1040, 0x2C40, 0x43FE, 0x8000}},
    {24315, 2, {0x0, 0x3FC, 0xFA04, 0xA04, 0x12F4, 0x1294, 0x2294, 0x7AF4, 0xA04, 0xA04, 0x4BFC, 0x3204, 0x1000, 0x2C00, 0x43FE, 0x8000}},
    {24316, 2, {0x0, 0x7FE, 0xF890, 0x890, 0x13FC, 0x1294, 0x2294, 0x7A94, 0xA9C, 0xB04, 0x4A04, 0x33FC, 0x1204, 0x2C00, 0x43FE, 0x8000}},
    {24317, 2, {0x0, 0x3FC, 0xFA94, 0xA94, 0x12F4, 0x1294, 0x2294, 0x7AF4, 0xA94, 0xA94, 0x4A94, 0x33FC, 0x1204, 0x2C00, 0x43FE, 0x8000}},
    {24318, 2, {0x0, 0x0, 0x820, 0x820, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {24319, 2, {0x810, 0x810, 0x810, 0x810, 0x810, 0xFFFE, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0xFF0, 0x810}},
    {24320, 2, {0x0, 0x7FFC, 0x820, 0x820, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {24321, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x1020, 0x2020, 0x4020}},
    {24322, 2, {0x0, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2004, 0x2004, 0x1FFC, 0x0, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24323, 2, {0x100, 0x80, 0x3FFC, 0x200, 0x420, 0x810, 0x1FF8, 0x8, 0x420, 0x420, 0x7FFE, 0x420, 0x420, 0x820, 0x1020, 0x2020}},
    {24324, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24325, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24326, 2, {0x100, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x400, 0x820, 0x1FF0, 0x10, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24327, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24328, 2, {0x200, 0x100, 0x7FFC, 0x440, 0x2450, 0x4448, 0x8944, 0x1080, 0x0, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24329, 2, {0x2420, 0x2420, 0x2420, 0x3DFC, 0x420, 0xFC20, 0x2420, 0x24F8, 0x4400, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24330, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4346, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {24331, 2, {0x220, 0x210, 0x210, 0x200, 0x2FE, 0xFF00, 0x200, 0x200, 0x200, 0x100, 0x100, 0x80, 0x42, 0x22, 0x1A, 0x6}},
    {24332, 2, {0x48, 0x44, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x40, 0x40, 0x40, 0x20, 0x22, 0x7F92, 0xA, 0x6, 0x2}},
    {24333, 2, {0x48, 0x44, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x40, 0x7E40, 0x40, 0x20, 0x22, 0xF12, 0xF00A, 0x4006, 0x2}},
    {24334, 2, {0x48, 0x44, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x7F40, 0x40, 0x40, 0x3E20, 0x22, 0x12, 0x78A, 0xF806, 0x4002}},
    {24335, 2, {0x48, 0x44, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x3E40, 0x840, 0x840, 0x820, 0x822, 0xF12, 0x780A, 0x2006, 0x2}},
    {24336, 2, {0x48, 0x3E44, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x40, 0x7E40, 0x40, 0x20, 0x22, 0xF12, 0xF00A, 0x4006, 0x2}},
    {24337, 2, {0x14, 0x4412, 0x2810, 0x13FE, 0x2810, 0x4410, 0x1010, 0x11F0, 0xFC90, 0x1090, 0x5890, 0x5488, 0x94EA, 0x138A, 0x5106, 0x2002}},
    {24338, 2, {0x4410, 0x2814, 0x1012, 0x2810, 0x45FE, 0x1010, 0x1410, 0x11D0, 0xFC90, 0x3090, 0x3890, 0x5488, 0x50EA, 0x938A, 0x1006, 0x1002}},
    {24339, 2, {0x0, 0x7FF0, 0x10, 0x10, 0x10, 0x3FF0, 0x2000, 0x2000, 0x4000, 0x7FF8, 0x8, 0x8, 0x8, 0x8, 0x50, 0x20}},
    {24340, 2, {0x0, 0x3FF8, 0x108, 0x108, 0x108, 0x1FF8, 0x1100, 0x1100, 0x2100, 0x3FFC, 0x104, 0x104, 0x114, 0x108, 0x100, 0x100}},
    {24341, 2, {0x8, 0x7F08, 0x108, 0x108, 0x108, 0x3F08, 0x2008, 0x2008, 0x4008, 0x7F08, 0x108, 0x108, 0x108, 0x108, 0xA08, 0x408}},
    {24342, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x1FF8, 0x1000, 0x2000, 0x3FFC, 0x4, 0x4, 0x4, 0x28, 0x10, 0x0, 0xFFFE, 0x0}},
    {24343, 2, {0x440, 0x440, 0x440, 0x7FF8, 0x448, 0x448, 0x3FF8, 0x2440, 0x2440, 0x3FFC, 0x444, 0x844, 0x854, 0x1048, 0x2040, 0x4040}},
    {24344, 2, {0x20, 0x7C20, 0x420, 0x420, 0x420, 0x7C40, 0x4040, 0x4040, 0x4080, 0x7C88, 0x484, 0x504, 0x7FE, 0x502, 0x2802, 0x1000}},
    {24345, 2, {0x0, 0x7CFE, 0x410, 0x410, 0x410, 0x7C10, 0x4010, 0x41FE, 0x4010, 0x7C10, 0x410, 0x410, 0x410, 0x410, 0x2850, 0x1020}},
    {24346, 2, {0x820, 0x420, 0x440, 0x7FF8, 0x108, 0x108, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x104, 0x104, 0x114, 0x108, 0x100, 0x100}},
    {24347, 2, {0x20, 0xF820, 0x820, 0x920, 0x92C, 0x7934, 0x4164, 0x43A4, 0x4124, 0x7934, 0x928, 0x922, 0x922, 0x902, 0x50FE, 0x2000}},
    {24348, 2, {0x0, 0x7DFC, 0x404, 0x404, 0x404, 0x7CFC, 0x4080, 0x4080, 0x4100, 0x7DFC, 0x404, 0x404, 0x404, 0x404, 0x2828, 0x1010}},
    {24349, 2, {0x0, 0x7DFC, 0x524, 0x524, 0x524, 0x7D24, 0x4124, 0x41FC, 0x4104, 0x7D00, 0x500, 0x500, 0x502, 0x502, 0x28FE, 0x1000}},
    {24350, 2, {0x40, 0x7C40, 0x440, 0x47C, 0x484, 0x7C88, 0x4120, 0x4020, 0x4020, 0x7C50, 0x450, 0x450, 0x488, 0x488, 0x2904, 0x1202}},
    {24351, 2, {0x820, 0x420, 0x440, 0x7FF8, 0x108, 0x108, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x304, 0x504, 0x914, 0x1108, 0x6100, 0x100}},
    {24352, 2, {0x100, 0xF908, 0x908, 0x910, 0x920, 0x7940, 0x4100, 0x47FE, 0x4140, 0x7920, 0x920, 0x910, 0x908, 0x944, 0x5182, 0x2100}},
    {24353, 2, {0x0, 0x7CFE, 0x480, 0x480, 0x480, 0x7CFC, 0x4084, 0x4084, 0x4084, 0x7C84, 0x4FC, 0x480, 0x480, 0x480, 0x28FE, 0x1000}},
    {24354, 2, {0x20, 0x7D24, 0x524, 0x524, 0x5FC, 0x7C20, 0x4020, 0x43FC, 0x4104, 0x7C88, 0x488, 0x450, 0x420, 0x450, 0x2988, 0x1606}},
    {24355, 2, {0x88, 0xF888, 0x888, 0x908, 0x97E, 0x7B08, 0x4508, 0x4148, 0x4128, 0x7928, 0x908, 0x908, 0x908, 0x908, 0x5128, 0x2110}},
    {24356, 2, {0x4, 0xF81E, 0x9F0, 0x910, 0x910, 0x7910, 0x4110, 0x41FE, 0x4110, 0x7910, 0x910, 0x908, 0x90A, 0x94A, 0x51A6, 0x2112}},
    {24357, 2, {0x80, 0xF880, 0x880, 0x9FE, 0x902, 0x7A04, 0x4420, 0x4020, 0x4128, 0x7924, 0xA24, 0xA22, 0xC22, 0x820, 0x50A0, 0x2040}},
    {24358, 2, {0x40, 0xF820, 0x820, 0xBFE, 0x840, 0x7840, 0x4088, 0x4108, 0x43F0, 0x7820, 0x840, 0x888, 0x904, 0xBFC, 0x5104, 0x2000}},
    {24359, 2, {0x8, 0xF83C, 0xBD0, 0xA90, 0xA90, 0x7A90, 0x4290, 0x4290, 0x4290, 0x7A88, 0xA88, 0xAC8, 0xAA4, 0xCD4, 0x5492, 0x2800}},
    {24360, 2, {0x0, 0xFBFC, 0x884, 0x884, 0x884, 0x7904, 0x4114, 0x4208, 0x4400, 0x79FC, 0x904, 0x904, 0x904, 0x904, 0x51FC, 0x2104}},
    {24361, 2, {0x800, 0x7F7C, 0x1144, 0x3228, 0xC10, 0x1228, 0x6146, 0x0, 0x3FF0, 0x10, 0x3FF0, 0x2000, 0x3FF8, 0x8, 0x50, 0x20}},
    {24362, 2, {0x0, 0x7DFC, 0x408, 0x410, 0x430, 0x7C48, 0x4084, 0x4302, 0x4000, 0x7DFC, 0x420, 0x420, 0x420, 0x420, 0x2BFE, 0x1000}},
    {24363, 2, {0x0, 0x7DFE, 0x520, 0x520, 0x520, 0x7DFC, 0x4104, 0x4104, 0x4104, 0x7D04, 0x5FC, 0x520, 0x520, 0x520, 0x29FE, 0x1000}},
    {24364, 2, {0x0, 0x7DFE, 0x510, 0x510, 0x550, 0x7D5C, 0x4144, 0x4144, 0x4144, 0x7D44, 0x55C, 0x550, 0x510, 0x510, 0x29FE, 0x1000}},
    {24365, 2, {0x0, 0xFBFE, 0x908, 0x908, 0x908, 0x79F8, 0x4108, 0x4108, 0x41F8, 0x7908, 0x908, 0x91E, 0xBE8, 0x808, 0x5008, 0x2008}},
    {24366, 2, {0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1010, 0x2008, 0xDFF6, 0x10, 0xFF0, 0x800, 0x1000, 0x1FF8, 0x8, 0x28, 0x10}},
    {24367, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4444, 0x0, 0x3FF0, 0x10, 0x3FF0, 0x2000, 0x3FF8, 0x8, 0x50, 0x20}},
    {24368, 2, {0x20, 0x7D24, 0x4A4, 0x4A8, 0x420, 0x7DFC, 0x4104, 0x4104, 0x41FC, 0x7D04, 0x504, 0x5FC, 0x504, 0x504, 0x2914, 0x1108}},
    {24369, 2, {0x0, 0x7EFC, 0x204, 0x204, 0x7EFC, 0x4080, 0x4080, 0x7EFC, 0x2244, 0x1224, 0x60C, 0x1A34, 0x62C4, 0x204, 0x1428, 0x810}},
    {24370, 2, {0x0, 0xF9F8, 0x908, 0x908, 0x9F8, 0x7800, 0x43FC, 0x4204, 0x4204, 0x7BFC, 0xA04, 0xA04, 0xBFC, 0xA04, 0x5214, 0x2208}},
    {24371, 2, {0x0, 0xFBFC, 0x800, 0x924, 0xA48, 0x7C90, 0x4248, 0x4124, 0x4000, 0x7BFC, 0x840, 0x840, 0x840, 0x840, 0x57FE, 0x2000}},
    {24372, 2, {0x40, 0xF820, 0xBFE, 0x800, 0x9FC, 0x7904, 0x41FC, 0x4000, 0x41FC, 0x7808, 0x810, 0xBFE, 0x820, 0x820, 0x50A0, 0x2040}},
    {24373, 2, {0x0, 0xF8FC, 0x880, 0x8F8, 0x880, 0x78F8, 0x4080, 0x47FE, 0x4140, 0x7924, 0x928, 0x910, 0x908, 0x944, 0x5182, 0x2100}},
    {24374, 2, {0x40, 0xF820, 0xBFE, 0x800, 0x800, 0x79FC, 0x4104, 0x4104, 0x4104, 0x79FC, 0x820, 0x8A8, 0x924, 0xA22, 0x50A0, 0x2040}},
    {24375, 2, {0x20, 0xF840, 0x888, 0x904, 0xBFE, 0x7822, 0x4020, 0x43FE, 0x4222, 0x7A22, 0xBFE, 0x820, 0x824, 0x822, 0x57FE, 0x2002}},
    {24376, 2, {0x0, 0xFBDE, 0xA52, 0xA52, 0xA52, 0x7BDE, 0x4252, 0x4252, 0x4252, 0x7BDE, 0xA52, 0xA52, 0xA52, 0xA52, 0x5552, 0x28A6}},
    {24377, 2, {0x104, 0xF888, 0x850, 0x9FC, 0x924, 0x7924, 0x41FC, 0x4124, 0x4124, 0x79FC, 0x820, 0x820, 0xBFE, 0x820, 0x5020, 0x2020}},
    {24378, 2, {0x0, 0xF9FC, 0x904, 0x904, 0x79FC, 0x4020, 0x83FE, 0x8222, 0xFA22, 0xA22, 0xBFE, 0x820, 0x824, 0x822, 0x903F, 0x63C1}},
    {24379, 2, {0x0, 0xF01E, 0x17C2, 0x1102, 0x1102, 0xF7DE, 0x8550, 0x8550, 0x8550, 0xF6DE, 0x1442, 0x1442, 0x1442, 0x17C2, 0xA454, 0x4008}},
    {24380, 2, {0x0, 0xF01E, 0x17C2, 0x1102, 0x1202, 0xF7DE, 0x8450, 0x8450, 0x8450, 0xF7DE, 0x1442, 0x1442, 0x1442, 0x17C2, 0xA454, 0x4008}},
    {24381, 2, {0x48, 0xF948, 0x948, 0xBFE, 0x948, 0x7948, 0x4178, 0x4100, 0x41FE, 0x7820, 0xBFE, 0x870, 0x8A8, 0x924, 0x5622, 0x2020}},
    {24382, 2, {0x242, 0xF924, 0x800, 0x9FC, 0x924, 0x7924, 0x41FC, 0x4124, 0x4124, 0x79FC, 0x820, 0x820, 0xBFE, 0x820, 0x5020, 0x2020}},
    {24383, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1FF0, 0x2028, 0xDFE6, 0x1000, 0x1FF0, 0x10, 0xE0}},
    {24384, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0x7C44, 0x428, 0x7C28, 0x4010, 0x7C28, 0x444, 0xD82}},
    {24385, 2, {0x0, 0xFBFE, 0x808, 0x9E8, 0x928, 0x79E8, 0x4000, 0x43FE, 0x4008, 0x79E8, 0x928, 0x928, 0x9E8, 0x808, 0x5028, 0x2010}},
    {24386, 2, {0x440, 0x2448, 0x1450, 0x440, 0xFFFE, 0x0, 0xFC80, 0x4F8, 0x7D20, 0x4020, 0x81FC, 0xFC20, 0x450, 0x450, 0x2888, 0x1104}},
    {24387, 2, {0x0, 0xF9FC, 0x924, 0x9FC, 0x924, 0x79FC, 0x40A8, 0x40A8, 0x43FE, 0x78A8, 0x8A8, 0x9FC, 0x820, 0xBFE, 0x5020, 0x2020}},
    {24388, 2, {0x0, 0xFBFE, 0xA00, 0xA7C, 0xA44, 0x7A44, 0x427C, 0x4200, 0x42EE, 0x7AAA, 0xAAA, 0xAAA, 0xAEE, 0xA00, 0x53FE, 0x2000}},
    {24389, 2, {0x108, 0xF890, 0xFFE, 0x800, 0xBC4, 0x7A54, 0x43D4, 0x4254, 0x43D4, 0x7A44, 0xACC, 0x800, 0xBFC, 0x884, 0x5114, 0x2608}},
    {24390, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x3FF0, 0x10, 0x3FF0, 0x2000, 0x3FF8, 0x8, 0x50, 0x20}},
    {24391, 2, {0x20, 0xF9FC, 0x824, 0xBFE, 0x824, 0x79FC, 0x4020, 0x42AA, 0x43AE, 0x7A22, 0xBFE, 0xA22, 0xBAE, 0xAAA, 0x52AA, 0x2422}},
    {24392, 2, {0x0, 0xFBDE, 0xA52, 0xA52, 0xBDE, 0x7800, 0x41FC, 0x4124, 0x41FC, 0x7924, 0x9FC, 0x820, 0xBFE, 0x820, 0x5020, 0x2020}},
    {24393, 2, {0x88, 0xF888, 0x9FC, 0x888, 0x888, 0x7BFE, 0x4020, 0x41FC, 0x4124, 0x79FC, 0x924, 0x9FC, 0x800, 0x888, 0x5104, 0x2202}},
    {24394, 2, {0x0, 0xFBFE, 0x800, 0x9FC, 0x924, 0x79FC, 0x4124, 0x43FE, 0x4000, 0x79FC, 0x924, 0x9FC, 0x924, 0x9FC, 0x5000, 0x23FE}},
    {24395, 2, {0x0, 0xFBFC, 0xA94, 0xA94, 0xBFC, 0x7800, 0x47FE, 0x4000, 0x43FC, 0x7A04, 0xBFC, 0x8A2, 0x914, 0xB08, 0x5D44, 0x2182}},
    {24396, 2, {0x0, 0xF7FC, 0x1248, 0x1444, 0x1FFE, 0xF444, 0x86EC, 0x8554, 0x86EC, 0xF444, 0x16EC, 0x1554, 0x16EC, 0x1444, 0xA454, 0x4408}},
    {24397, 2, {0x40, 0xF020, 0x17FE, 0x1488, 0x15FC, 0xF488, 0x87FE, 0x8420, 0x85FC, 0xF524, 0x15FC, 0x1524, 0x15FC, 0x1400, 0xA488, 0x4904}},
    {24398, 2, {0x2108, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0x0, 0x3FF0, 0x10, 0x3FF0, 0x2000, 0x3FF8, 0x8, 0x70}},
    {24399, 2, {0x3DE, 0xFA52, 0xBDE, 0xA52, 0xBDE, 0x78A0, 0x41FE, 0x4320, 0x45FC, 0x7920, 0x9FE, 0x800, 0x9FC, 0x888, 0x5070, 0x278E}},
    {24400, 2, {0x0, 0x0, 0x7FF8, 0x8, 0x8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x8, 0x8, 0x7FF8, 0x8, 0x0, 0x0}},
    {24401, 2, {0x400, 0x400, 0x400, 0x800, 0xFF0, 0x810, 0x1010, 0x1020, 0x2020, 0x3FE0, 0x40, 0x40, 0x80, 0x80, 0xFFFE, 0x0}},
    {24402, 2, {0x800, 0x800, 0x4BFC, 0x4804, 0x4804, 0x4804, 0x4804, 0x49FC, 0x4804, 0x4804, 0x4804, 0x804, 0x1004, 0x13FC, 0x2004, 0x4000}},
    {24403, 2, {0x100, 0x2108, 0x1108, 0x910, 0x920, 0x100, 0x7FF8, 0x8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x8, 0x7FF8, 0x8}},
    {24404, 2, {0x800, 0xFE0, 0x820, 0x1020, 0x1FC0, 0x40, 0x40, 0xFFFE, 0x100, 0x2108, 0x1190, 0x560, 0x920, 0x3118, 0xC506, 0x200}},
    {24405, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x1FF0, 0x10, 0x10, 0xFFFE, 0x100, 0x2108, 0x1190, 0x560, 0x920, 0x3118, 0xC506, 0x200}},
    {24406, 2, {0x800, 0xFF0, 0x1010, 0x1FE0, 0x20, 0xFFFE, 0x200, 0xD08, 0x7190, 0x2A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {24407, 2, {0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x0, 0x7FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x7FF8, 0x8}},
    {24408, 2, {0x800, 0xFF0, 0x1010, 0x1FE0, 0x20, 0xFFFE, 0x200, 0x4210, 0x43D2, 0x7514, 0x4118, 0x4FD0, 0x4112, 0x6292, 0x444E, 0x820}},
    {24409, 2, {0x800, 0xFF0, 0x1010, 0x1FE0, 0x20, 0x7FFE, 0x4002, 0x9FF4, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x1110, 0x610C}},
    {24410, 2, {0x0, 0x3FF0, 0x10, 0x1FF0, 0x10, 0x7FFE, 0x4002, 0x9FF4, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x1110, 0x610C}},
    {24411, 2, {0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x9248, 0x5484, 0xFF7A, 0x3828, 0x5448, 0x9298, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {24412, 2, {0x800, 0xFF0, 0x1010, 0x1FE0, 0x20, 0xFFFE, 0x9248, 0x5484, 0xFF7A, 0x3828, 0x5448, 0x9298, 0x820, 0xFFFE, 0x820, 0x1020}},
    {24413, 2, {0x800, 0xFE0, 0x1020, 0x1FC0, 0x40, 0xFFFE, 0x9224, 0x5478, 0xFE12, 0x38FE, 0x5454, 0x9292, 0x820, 0xFFFE, 0x820, 0x1020}},
    {24414, 2, {0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x9224, 0x5478, 0xFE12, 0x38FE, 0x5454, 0x9292, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {24415, 2, {0x44, 0xFDFE, 0x454, 0x448, 0xFCFE, 0x590, 0x4FE, 0xFC90, 0xFE, 0x890, 0xFEFE, 0x880, 0x49FE, 0x2844, 0x838, 0x19C6}},
    {24416, 2, {0x7E44, 0x3FE, 0x3E54, 0x248, 0x7EFE, 0x190, 0xEEFE, 0x4A90, 0x6EFE, 0xC090, 0x4FE, 0xFE80, 0x45FE, 0x2444, 0x2438, 0xDC6}},
    {24417, 2, {0x40, 0x80, 0x100, 0x600, 0x1800, 0x20, 0x40, 0x180, 0x600, 0x1800, 0x10, 0x20, 0x40, 0x180, 0x600, 0x3800}},
    {24418, 2, {0x0, 0x7F84, 0x1204, 0x1208, 0x1210, 0x1222, 0x1202, 0xFFC4, 0x1208, 0x1210, 0x1222, 0x1202, 0x2204, 0x2208, 0x4210, 0x8260}},
    {24419, 2, {0x1004, 0x804, 0x808, 0xFF90, 0x2224, 0x2244, 0x2208, 0x2212, 0x1422, 0x1444, 0x808, 0x810, 0x1420, 0x2300, 0x40FE, 0x8000}},
    {24420, 2, {0x0, 0x3F04, 0x2104, 0x2108, 0x2910, 0x2522, 0x2502, 0x2104, 0xFFC8, 0x2110, 0x2122, 0x2102, 0x2104, 0x4108, 0x4510, 0x8260}},
    {24421, 2, {0x200, 0x100, 0x3FF8, 0x420, 0x3C0, 0x420, 0x3FFC, 0x2020, 0x21C0, 0x2E10, 0x2060, 0x2188, 0x2E10, 0x4060, 0x4380, 0x9C00}},
    {24422, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0x3FFC, 0x2000, 0x2020, 0x21C0, 0x2E10, 0x2060, 0x2188, 0x2E10, 0x4060, 0x4380, 0x9C00}},
    {24423, 2, {0x50, 0x48, 0x40, 0xFFFE, 0x40, 0x50, 0x7C60, 0x45C8, 0x4450, 0x4464, 0x7DA8, 0x12, 0xE32, 0xF0CA, 0x4306, 0x2}},
    {24424, 2, {0x2, 0xFFE2, 0x4, 0x8, 0x7BD0, 0x4A42, 0x4A42, 0x4A44, 0x6B48, 0x5AD0, 0x4A42, 0x4A42, 0x4A44, 0x4A48, 0x4A50, 0x5AE0}},
    {24425, 2, {0x100, 0x784, 0x7804, 0x888, 0x4490, 0x2522, 0x2002, 0x404, 0x7F88, 0xC10, 0x1622, 0x1502, 0x2484, 0x4408, 0x8410, 0x460}},
    {24426, 2, {0x400, 0x784, 0x404, 0x3FC8, 0x2450, 0x2724, 0x3C84, 0x2388, 0x2010, 0x2F24, 0x2904, 0x2908, 0x4912, 0x5122, 0x90FE, 0x2000}},
    {24427, 2, {0x0, 0x3F84, 0x2484, 0x2488, 0x2E90, 0x24A2, 0x3F82, 0x2084, 0x2E88, 0x2A90, 0x2AA2, 0x2E82, 0x2084, 0x4088, 0x4290, 0x8160}},
    {24428, 2, {0x1082, 0x1082, 0x1084, 0x1088, 0xFBD0, 0x1082, 0x3082, 0x39C4, 0x55A8, 0x5290, 0x9282, 0x1482, 0x1084, 0x1088, 0x1090, 0x10A0}},
    {24429, 2, {0x800, 0x804, 0xFF84, 0x808, 0x810, 0x7F22, 0x2, 0x7F04, 0x4108, 0x4110, 0x7F22, 0x2, 0x2204, 0x1788, 0xF810, 0x4060}},
    {24430, 2, {0x1000, 0x804, 0xFF84, 0x8088, 0x1410, 0x2222, 0x4902, 0x1404, 0x2208, 0x4110, 0xBEA2, 0x2202, 0x2204, 0x2208, 0x3E10, 0x2260}},
    {24431, 2, {0x0, 0xFF84, 0x1404, 0x7F08, 0x5510, 0x5522, 0x7F02, 0x4, 0x7F08, 0x10, 0xFFA2, 0x802, 0x2A04, 0x4908, 0xA890, 0x1060}},
    {24432, 2, {0x1000, 0x804, 0x7F04, 0x2208, 0x1410, 0xFFA2, 0x2, 0x7F04, 0x4108, 0x7F10, 0x4122, 0x7F02, 0x804, 0xFF88, 0x810, 0x860}},
    {24433, 2, {0x0, 0x7F04, 0x4104, 0x7F08, 0x4110, 0x7F22, 0x802, 0xFFC4, 0x8, 0x7F10, 0x4122, 0x7F02, 0x804, 0x4908, 0x8890, 0x1860}},
    {24434, 2, {0x7BC2, 0x2, 0x7BC4, 0x4A48, 0x6B50, 0x4A42, 0x402, 0x7FC4, 0x4A08, 0x7FD0, 0x4A42, 0x7FC2, 0x5244, 0x5B88, 0x5250, 0x99E0}},
    {24435, 2, {0x800, 0x800, 0x1000, 0x2000, 0x4800, 0x800, 0x1000, 0x3000, 0x5000, 0x9000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {24436, 2, {0x840, 0x840, 0x1080, 0x20FC, 0x4904, 0x904, 0x1204, 0x3484, 0x5044, 0x9024, 0x1024, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {24437, 2, {0x840, 0x840, 0x1040, 0x2240, 0x4A58, 0xA68, 0x12C8, 0x3748, 0x5248, 0x9268, 0x1250, 0x1244, 0x1244, 0x1204, 0x11FC, 0x1000}},
    {24438, 2, {0x1000, 0x17F8, 0x2108, 0x4110, 0x9110, 0x1120, 0x213C, 0x6104, 0xA104, 0x2288, 0x2288, 0x2250, 0x2420, 0x2450, 0x2888, 0x2306}},
    {24439, 2, {0x840, 0x820, 0x1020, 0x23FE, 0x4880, 0x880, 0x1080, 0x30FC, 0x5084, 0x9084, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {24440, 2, {0x820, 0x920, 0x1120, 0x2110, 0x4910, 0xA48, 0x1248, 0x3444, 0x5882, 0x9080, 0x1110, 0x1108, 0x1208, 0x17FC, 0x1204, 0x1000}},
    {24441, 2, {0x800, 0x9F0, 0x1110, 0x2110, 0x4910, 0xA0E, 0x1400, 0x33F8, 0x5108, 0x9110, 0x1090, 0x10A0, 0x1040, 0x10A0, 0x1318, 0x1C06}},
    {24442, 2, {0x800, 0x800, 0x13FE, 0x2020, 0x4820, 0x820, 0x1020, 0x3020, 0x51FC, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {24443, 2, {0x1000, 0x1200, 0x22FC, 0x4224, 0x9224, 0x1224, 0x23A4, 0x6E24, 0xA224, 0x2224, 0x22A4, 0x2324, 0x2244, 0x2044, 0x2094, 0x2108}},
    {24444, 2, {0x820, 0x820, 0x1020, 0x23FE, 0x4A22, 0xA24, 0x1220, 0x33FC, 0x5284, 0x9288, 0x1248, 0x1250, 0x1220, 0x1450, 0x1488, 0x1906}},
    {24445, 2, {0x808, 0x83C, 0x13E0, 0x2220, 0x4A20, 0xA20, 0x1220, 0x33FE, 0x5220, 0x9210, 0x1210, 0x1212, 0x120A, 0x128A, 0x1326, 0x1212}},
    {24446, 2, {0x840, 0x840, 0x10A0, 0x20A0, 0x4910, 0xA48, 0x1426, 0x3020, 0x53F8, 0x9008, 0x1010, 0x1110, 0x10A0, 0x1040, 0x1020, 0x1020}},
    {24447, 2, {0x890, 0x890, 0x1090, 0x23FC, 0x4894, 0x894, 0x13FC, 0x3290, 0x5290, 0x93FE, 0x1092, 0x1092, 0x111A, 0x1114, 0x1210, 0x1410}},
    {24448, 2, {0x880, 0x840, 0x1000, 0x27FC, 0x4840, 0x840, 0x1040, 0x3040, 0x57FC, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {24449, 2, {0x800, 0x800, 0x13FE, 0x2020, 0x4820, 0x820, 0x1120, 0x3120, 0x513C, 0x9120, 0x1120, 0x1120, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {24450, 2, {0x800, 0x9F8, 0x1108, 0x2108, 0x4908, 0x9F8, 0x1108, 0x3108, 0x5108, 0x91F8, 0x1108, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {24451, 2, {0x820, 0x920, 0x1120, 0x2120, 0x49FC, 0xA20, 0x1220, 0x3420, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {24452, 2, {0x800, 0xBF8, 0x1010, 0x2020, 0x4860, 0x898, 0x1104, 0x3602, 0x5000, 0x93FC, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {24453, 2, {0x840, 0x840, 0x1040, 0x23FC, 0x4840, 0x840, 0x17FE, 0x3010, 0x5010, 0x97FE, 0x1010, 0x1210, 0x1110, 0x1110, 0x1050, 0x1020}},
    {24454, 2, {0x800, 0xFFE, 0x1090, 0x2090, 0x4890, 0xBFC, 0x1294, 0x3294, 0x5294, 0x9294, 0x129C, 0x1304, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {24455, 2, {0x880, 0x880, 0x1100, 0x21FC, 0x4A04, 0xC04, 0x13E4, 0x3224, 0x5224, 0x93E4, 0x1224, 0x1224, 0x13E4, 0x1004, 0x1028, 0x1010}},
    {24456, 2, {0x800, 0xBF8, 0x1208, 0x2208, 0x4BF8, 0xA08, 0x1208, 0x33F8, 0x5244, 0x9248, 0x1230, 0x1220, 0x1210, 0x1288, 0x1306, 0x1200}},
    {24457, 2, {0x904, 0x884, 0x1088, 0x2000, 0x4BFE, 0x820, 0x1020, 0x31FC, 0x5020, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24458, 2, {0x800, 0x800, 0x13FC, 0x2204, 0x4A04, 0xAF4, 0x1294, 0x3294, 0x5294, 0x9294, 0x12F4, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1000}},
    {24459, 2, {0x840, 0x840, 0x13F8, 0x2048, 0x4FFE, 0x848, 0x13F8, 0x3040, 0x5040, 0x93F8, 0x1040, 0x1040, 0x17FC, 0x1040, 0x1040, 0x1040}},
    {24460, 2, {0x820, 0x840, 0x1088, 0x2110, 0x4BE0, 0x840, 0x1088, 0x33FC, 0x5084, 0x9080, 0x11F8, 0x1308, 0x1490, 0x1060, 0x1198, 0x1606}},
    {24461, 2, {0x820, 0x820, 0x11FC, 0x2020, 0x4820, 0x820, 0x17FE, 0x3000, 0x5020, 0x9020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {24462, 2, {0x800, 0x9FC, 0x1104, 0x2104, 0x4904, 0x9FC, 0x1000, 0x3000, 0x53FE, 0x9020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24463, 2, {0x820, 0x820, 0x1120, 0x213C, 0x4920, 0x920, 0x17FE, 0x3000, 0x5020, 0x9124, 0x1124, 0x1228, 0x1410, 0x1020, 0x10C0, 0x1700}},
    {24464, 2, {0x1040, 0x1040, 0x20A0, 0x4110, 0x9208, 0x15F6, 0x2040, 0x6040, 0xA7FC, 0x2040, 0x2250, 0x2248, 0x2444, 0x2844, 0x2140, 0x2080}},
    {24465, 2, {0x800, 0xBFC, 0x1000, 0x2124, 0x4A48, 0xC90, 0x1248, 0x3124, 0x5000, 0x93FC, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {24466, 2, {0x820, 0x820, 0x11FC, 0x2020, 0x4820, 0x820, 0x13FE, 0x3020, 0x5020, 0x9120, 0x113C, 0x1120, 0x12A0, 0x1260, 0x143E, 0x1800}},
    {24467, 2, {0x804, 0x904, 0x1088, 0x2090, 0x4800, 0xBFE, 0x1020, 0x3020, 0x5120, 0x913C, 0x1120, 0x1120, 0x12A0, 0x1260, 0x143E, 0x1800}},
    {24468, 2, {0x888, 0x888, 0x13FE, 0x2088, 0x4888, 0x840, 0x1020, 0x33FC, 0x5008, 0x9010, 0x1020, 0x1040, 0x1080, 0x1100, 0x1280, 0x147E}},
    {24469, 2, {0x1040, 0x1040, 0x2040, 0x47FC, 0x9040, 0x1248, 0x2148, 0x6150, 0xAFFE, 0x20E0, 0x2150, 0x2150, 0x2248, 0x2444, 0x2842, 0x2040}},
    {24470, 2, {0x840, 0x820, 0x13FE, 0x2202, 0x4800, 0x9FC, 0x1000, 0x3000, 0x53FE, 0x9020, 0x1128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {24471, 2, {0x800, 0xBF8, 0x1208, 0x23F8, 0x4A08, 0xBF8, 0x1000, 0x33FC, 0x5010, 0x97FE, 0x1010, 0x1210, 0x1110, 0x1010, 0x1050, 0x1020}},
    {24472, 2, {0x890, 0x890, 0x1090, 0x279E, 0x4890, 0x890, 0x1090, 0x339C, 0x5090, 0x9090, 0x1090, 0x179E, 0x1090, 0x1090, 0x1090, 0x1090}},
    {24473, 2, {0x820, 0x820, 0x1120, 0x213C, 0x4920, 0x920, 0x17FE, 0x3000, 0x5020, 0x9120, 0x113C, 0x1120, 0x12A0, 0x1460, 0x183E, 0x1000}},
    {24474, 2, {0x1040, 0x1040, 0x2FFE, 0x4040, 0x97FC, 0x1444, 0x2554, 0x64E4, 0xA444, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x2040}},
    {24475, 2, {0x840, 0x840, 0x13FC, 0x20A0, 0x4910, 0xA08, 0x17FE, 0x3008, 0x53C8, 0x9248, 0x1248, 0x1248, 0x13C8, 0x1008, 0x1028, 0x1010}},
    {24476, 2, {0x820, 0x924, 0x10A4, 0x20A8, 0x4820, 0xBFE, 0x1202, 0x3202, 0x52FA, 0x928A, 0x128A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {24477, 2, {0x840, 0x840, 0x17FC, 0x2040, 0x4840, 0xBF8, 0x1208, 0x33F8, 0x5208, 0x93F8, 0x1208, 0x13F8, 0x1208, 0x1208, 0x1FFE, 0x1000}},
    {24478, 2, {0x888, 0x888, 0x1088, 0x2088, 0x4954, 0xA22, 0x1442, 0x3020, 0x5020, 0x9120, 0x113C, 0x1120, 0x12A0, 0x1260, 0x143E, 0x1800}},
    {24479, 2, {0x1000, 0x17FC, 0x2444, 0x4444, 0x95F4, 0x1444, 0x2444, 0x67FC, 0xA404, 0x25F4, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x2808}},
    {24480, 2, {0x840, 0x840, 0x1040, 0x27FC, 0x4840, 0xA48, 0x1248, 0x3248, 0x5554, 0x90E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040}},
    {24481, 2, {0x1400, 0x1400, 0x27DE, 0x4912, 0x9112, 0x1112, 0x2FD2, 0x6112, 0xA112, 0x25D2, 0x2512, 0x251A, 0x25D4, 0x3E10, 0x2810, 0x2010}},
    {24482, 2, {0x1040, 0x1040, 0x27FE, 0x4040, 0x93F8, 0x1048, 0x27FE, 0x6048, 0xA3F8, 0x2040, 0x2240, 0x227C, 0x2240, 0x2540, 0x24FE, 0x2800}},
    {24483, 2, {0x1110, 0x1110, 0x2110, 0x47FC, 0x9110, 0x1110, 0x2FFE, 0x6000, 0xA3F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {24484, 2, {0x810, 0x810, 0x177C, 0x2114, 0x49FE, 0xA14, 0x127C, 0x3710, 0x517C, 0x9510, 0x15FE, 0x1210, 0x1210, 0x1500, 0x18FE, 0x1000}},
    {24485, 2, {0x800, 0x9FC, 0x1104, 0x2104, 0x49FC, 0x904, 0x1104, 0x31FC, 0x5000, 0x93FE, 0x1020, 0x1120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {24486, 2, {0x1000, 0x17BC, 0x2484, 0x4484, 0x9484, 0x17BC, 0x2400, 0x6400, 0xA7BC, 0x2424, 0x2424, 0x27A8, 0x2410, 0x2428, 0x2444, 0x2482}},
    {24487, 2, {0x840, 0x820, 0x13FE, 0x2202, 0x4A02, 0xBFE, 0x1200, 0x3200, 0x53FE, 0x9352, 0x1552, 0x15FE, 0x1552, 0x1552, 0x1952, 0x1106}},
    {24488, 2, {0x840, 0x880, 0x13F8, 0x2208, 0x4BF8, 0xA08, 0x13F8, 0x3000, 0x57FC, 0x9040, 0x1040, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x1000}},
    {24489, 2, {0x900, 0x900, 0x11FE, 0x2200, 0x4DFC, 0x904, 0x11FC, 0x3104, 0x51FC, 0x9080, 0x10FC, 0x1108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {24490, 2, {0x103C, 0x17E0, 0x2420, 0x4420, 0x97FE, 0x1420, 0x25FC, 0x6504, 0xA504, 0x25FC, 0x2504, 0x25FC, 0x2904, 0x2904, 0x31FC, 0x2104}},
    {24491, 2, {0x880, 0x880, 0x13F8, 0x2108, 0x4FFE, 0x800, 0x13F8, 0x3208, 0x53F8, 0x9040, 0x17FC, 0x1040, 0x1440, 0x17FE, 0x1040, 0x1040}},
    {24492, 2, {0x840, 0x820, 0x13FC, 0x2108, 0x4890, 0xBFE, 0x1202, 0x3444, 0x5020, 0x93FC, 0x1080, 0x10F8, 0x1088, 0x1108, 0x1128, 0x1210}},
    {24493, 2, {0x808, 0x83C, 0x13C0, 0x2004, 0x4A44, 0x928, 0x11FC, 0x3220, 0x5020, 0x93FE, 0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {24494, 2, {0x1108, 0x1548, 0x2548, 0x4550, 0x97DE, 0x1024, 0x2014, 0x6FD4, 0xA014, 0x2794, 0x2494, 0x24A8, 0x24C8, 0x2494, 0x2814, 0x3022}},
    {24495, 2, {0x83C, 0xFC0, 0x1244, 0x2128, 0x4880, 0x910, 0x13E0, 0x3040, 0x5188, 0x93FC, 0x1044, 0x1040, 0x17FC, 0x10A0, 0x1110, 0x160E}},
    {24496, 2, {0x800, 0xBFE, 0x1020, 0x213C, 0x4920, 0x920, 0x17FE, 0x3000, 0x57FE, 0x9020, 0x1120, 0x113C, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {24497, 2, {0x1000, 0x17FC, 0x20A0, 0x47FC, 0x94A4, 0x14A4, 0x27FC, 0x6000, 0xA7FC, 0x2000, 0x2FFE, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {24498, 2, {0x1000, 0x17FC, 0x2404, 0x47FC, 0x9420, 0x14A8, 0x2470, 0x64A8, 0xA504, 0x24A0, 0x24FC, 0x2520, 0x2420, 0x2BFE, 0x2820, 0x3020}},
    {24499, 2, {0x820, 0x820, 0x17FE, 0x2040, 0x4840, 0xBFC, 0x1294, 0x3294, 0x53FC, 0x9040, 0x1024, 0x1522, 0x150A, 0x1908, 0x10F8, 0x1000}},
    {24500, 2, {0x1210, 0x1A90, 0x2A90, 0x4AA0, 0x8FBE, 0x1044, 0x2024, 0x6FA4, 0xA224, 0x2228, 0x2FA8, 0x2210, 0x2228, 0x2FA8, 0x2044, 0x2082}},
    {24501, 2, {0x1108, 0x1548, 0x2548, 0x4550, 0x97DE, 0x1014, 0x27E4, 0x6014, 0xA7D4, 0x2114, 0x27D4, 0x2108, 0x2108, 0x2FD4, 0x2024, 0x2042}},
    {24502, 2, {0x1108, 0x1108, 0x2548, 0x4390, 0x911E, 0x17D4, 0x2564, 0x6554, 0xA554, 0x2754, 0x25D4, 0x2548, 0x2548, 0x2554, 0x2464, 0x24C2}},
    {24503, 2, {0x1040, 0x1040, 0x2FFE, 0x4040, 0x97FC, 0x14A4, 0x24A4, 0x67FC, 0xA000, 0x2FFE, 0x2040, 0x2024, 0x2522, 0x250A, 0x2908, 0x20F8}},
    {24504, 2, {0x840, 0xBF8, 0x1110, 0x20A0, 0x4BFC, 0x800, 0x13F8, 0x3248, 0x53F8, 0x9248, 0x13F8, 0x1040, 0x13F8, 0x1040, 0x17FE, 0x1000}},
    {24505, 2, {0x1210, 0x1110, 0x27D0, 0x4210, 0x94BE, 0x1FE4, 0x2054, 0x6794, 0xA494, 0x2794, 0x2494, 0x2788, 0x2488, 0x2494, 0x2494, 0x25A2}},
    {24506, 2, {0x820, 0x820, 0x11FC, 0x2020, 0x4BFE, 0x908, 0x139C, 0x3108, 0x5188, 0x963E, 0x1000, 0x13FE, 0x1090, 0x1090, 0x1112, 0x120E}},
    {24507, 2, {0x1040, 0x10A0, 0x2318, 0x4DF6, 0x9000, 0x17FC, 0x2554, 0x64E4, 0xA7FC, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208}},
    {24508, 2, {0x1210, 0x1410, 0x2F90, 0x4890, 0x9FBE, 0x18A4, 0x2FD4, 0x6414, 0xA214, 0x2FD4, 0x2414, 0x2788, 0x2488, 0x2894, 0x2A94, 0x3122}},
    {24509, 2, {0x1210, 0x1A90, 0x2A90, 0x4FA0, 0x803E, 0x1FC4, 0x2424, 0x68A4, 0xAF24, 0x2228, 0x24A8, 0x3FD0, 0x2228, 0x2AA8, 0x3244, 0x2682}},
    {24510, 2, {0x1210, 0x1A90, 0x2A90, 0x4F90, 0x943E, 0x1224, 0x2FD4, 0x6014, 0xAF94, 0x2014, 0x2F94, 0x2008, 0x2F88, 0x2894, 0x2F94, 0x2022}},
    {24511, 2, {0x910, 0x890, 0x13DE, 0x2010, 0x4A5E, 0x982, 0x13DE, 0x3010, 0x53DE, 0x9250, 0x13DE, 0x1250, 0x13DE, 0x1250, 0x1252, 0x12CE}},
    {24512, 2, {0x1040, 0x17FC, 0x2000, 0x43B8, 0x92A8, 0x13B8, 0x2110, 0x67FC, 0xA110, 0x27FC, 0x2110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {24513, 2, {0xBF8, 0xA08, 0x13F8, 0x2208, 0x4BF8, 0xA48, 0x13F8, 0x3444, 0x57FC, 0x9080, 0x1FFE, 0x1248, 0x1554, 0x18E2, 0x1150, 0x16C8}},
    {24514, 2, {0x17BC, 0x14A4, 0x27BC, 0x44A4, 0x97BC, 0x14A4, 0x27BC, 0x6120, 0xA3FE, 0x2220, 0x27FC, 0x2A20, 0x23FC, 0x2220, 0x23FE, 0x2200}},
    {24515, 2, {0x0, 0x200, 0x100, 0x80, 0x80, 0x400, 0x408, 0x2404, 0x2404, 0x2402, 0x4402, 0x4412, 0x8410, 0x410, 0x3F0, 0x0}},
    {24516, 2, {0x1000, 0x1000, 0x1000, 0x1000, 0x1800, 0x5400, 0x5000, 0x5000, 0x9000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {24517, 2, {0x0, 0x400, 0x210, 0x110, 0x120, 0x820, 0x2848, 0x2844, 0x2884, 0x4902, 0x4A02, 0x8C12, 0x810, 0x1810, 0x27F0, 0x4000}},
    {24518, 2, {0x1000, 0x1000, 0x13FC, 0x1008, 0x1810, 0x5420, 0x5040, 0x5080, 0x9080, 0x1100, 0x1100, 0x1202, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {24519, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x19FC, 0x5444, 0x5044, 0x5044, 0x9044, 0x1084, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {24520, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0, 0x100, 0x880, 0x888, 0x4804, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24521, 2, {0x1000, 0x1000, 0x17FC, 0x1084, 0x1884, 0x5484, 0x5084, 0x5084, 0x9084, 0x1084, 0x1104, 0x1104, 0x1204, 0x1204, 0x1428, 0x1810}},
    {24522, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x1820, 0x5420, 0x5020, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {24523, 2, {0x1000, 0x1000, 0x13F8, 0x1008, 0x1808, 0x5408, 0x5008, 0x53F8, 0x9208, 0x1200, 0x1200, 0x1202, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {24524, 2, {0x0, 0x3FF0, 0x10, 0x10, 0x3FF0, 0x2000, 0x2004, 0x2004, 0x1FFC, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24525, 2, {0x0, 0x3FF8, 0x208, 0x1208, 0x1208, 0x2408, 0x808, 0x1050, 0x6020, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24526, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24527, 2, {0x1010, 0x1078, 0x13C0, 0x1040, 0x1840, 0x5440, 0x5040, 0x57FE, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24528, 2, {0x200, 0x200, 0x200, 0x3F8, 0x200, 0x200, 0x200, 0xFFFE, 0x0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24529, 2, {0x0, 0xFFFE, 0x200, 0x240, 0x220, 0x210, 0x208, 0x200, 0x0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24530, 2, {0x48, 0x44, 0x44, 0x40, 0xFFFE, 0x40, 0x840, 0x440, 0x1440, 0x5540, 0x50A0, 0x50A2, 0x9292, 0x120A, 0xE06, 0x2}},
    {24531, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x1820, 0x5420, 0x5020, 0x53FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24532, 2, {0x1080, 0x1080, 0x1100, 0x11FE, 0x1A00, 0x5400, 0x51F8, 0x5008, 0x9010, 0x1060, 0x1080, 0x1100, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {24533, 2, {0x1040, 0x1040, 0x1040, 0x1840, 0x5440, 0x53FE, 0x5040, 0x9040, 0x1040, 0x10A0, 0x10A0, 0x1090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {24534, 2, {0x1010, 0x1010, 0x1010, 0x1010, 0x1BFE, 0x5410, 0x5010, 0x5010, 0x9110, 0x1090, 0x1090, 0x1010, 0x1010, 0x1010, 0x1050, 0x1020}},
    {24535, 2, {0x100, 0x100, 0x108, 0x7FFC, 0x100, 0x100, 0x110, 0x3FF8, 0x0, 0xA10, 0x2908, 0x292C, 0x6824, 0x4824, 0x7E0, 0x0}},
    {24536, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24537, 2, {0x1020, 0x1010, 0x1010, 0x1000, 0x1BFE, 0x5480, 0x5080, 0x5080, 0x9080, 0x1080, 0x1080, 0x1080, 0x1080, 0x1080, 0x10FE, 0x1000}},
    {24538, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0x192C, 0x5534, 0x5164, 0x53A4, 0x9124, 0x1134, 0x1128, 0x1122, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {24539, 2, {0x1000, 0x11F0, 0x1110, 0x1110, 0x1910, 0x5510, 0x5190, 0x5150, 0x9150, 0x1110, 0x1110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {24540, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2100, 0x2080, 0x2240, 0x2200, 0x2200, 0x2A08, 0x2A04, 0x2A14, 0x5210, 0x4210, 0x81F0, 0x0}},
    {24541, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x920, 0x3118, 0xC106, 0x1190, 0x1148, 0x2124, 0x4124, 0x500, 0x200}},
    {24542, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC006, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24543, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1888, 0x5488, 0x5088, 0x5088, 0x9088, 0x1050, 0x1050, 0x1020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {24544, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24545, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x1BFE, 0x5622, 0x5222, 0x5222, 0x9222, 0x1222, 0x13FE, 0x1222, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24546, 2, {0x0, 0x7FFC, 0x200, 0x200, 0x3FF0, 0x410, 0x410, 0x810, 0xFFFE, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24547, 2, {0x1000, 0x13FC, 0x1084, 0x1088, 0x1888, 0x5490, 0x509C, 0x5084, 0x9144, 0x1144, 0x1128, 0x1128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {24548, 2, {0x1080, 0x1080, 0x1080, 0x11FC, 0x1920, 0x5620, 0x5020, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24549, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x810, 0x410, 0x1490, 0x5050, 0x514A, 0x514A, 0x8F06, 0x2}},
    {24550, 2, {0x1040, 0x1040, 0x10A0, 0x18A0, 0x5510, 0x5208, 0x5406, 0x9110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {24551, 2, {0x1090, 0x1088, 0x1088, 0x1080, 0x1BFE, 0x54A0, 0x50A0, 0x50A0, 0x90A0, 0x10A0, 0x1120, 0x1122, 0x1122, 0x1222, 0x121E, 0x1400}},
    {24552, 2, {0x1000, 0x1000, 0x13F8, 0x1800, 0x5400, 0x5000, 0x57FC, 0x9120, 0x1120, 0x1120, 0x1120, 0x1120, 0x1222, 0x1222, 0x141E, 0x1800}},
    {24553, 2, {0x440, 0x440, 0x820, 0x1010, 0x2208, 0xC446, 0x820, 0x1FF0, 0x10, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24554, 2, {0x1020, 0x1120, 0x1120, 0x1910, 0x5510, 0x5248, 0x5248, 0x9444, 0x1882, 0x1080, 0x1110, 0x1108, 0x1208, 0x17FC, 0x1204, 0x1000}},
    {24555, 2, {0x1040, 0x1040, 0x1040, 0x13F8, 0x1848, 0x5448, 0x5048, 0x5048, 0x97FE, 0x1040, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1406}},
    {24556, 2, {0x1000, 0x13F8, 0x1008, 0x1910, 0x54A0, 0x5040, 0x57FE, 0x9042, 0x1044, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080}},
    {24557, 2, {0x1040, 0x1020, 0x1020, 0x1000, 0x1BFE, 0x5440, 0x5040, 0x5060, 0x9050, 0x1048, 0x1044, 0x1044, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24558, 2, {0x1040, 0x1040, 0x1040, 0x17FE, 0x1840, 0x5440, 0x5040, 0x57F8, 0x9208, 0x1208, 0x1110, 0x10A0, 0x1040, 0x10A0, 0x1118, 0x1606}},
    {24559, 2, {0x1008, 0x103C, 0x13E0, 0x1220, 0x1A20, 0x5620, 0x5220, 0x53FE, 0x9220, 0x1210, 0x1210, 0x1212, 0x120A, 0x128A, 0x1306, 0x1202}},
    {24560, 2, {0x1080, 0x1080, 0x13F0, 0x1890, 0x5490, 0x5112, 0x5112, 0x924E, 0x1440, 0x1040, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24561, 2, {0x1040, 0x1040, 0x1040, 0x13FC, 0x1A44, 0x5448, 0x5040, 0x5040, 0x90A0, 0x10A0, 0x10A0, 0x1120, 0x1122, 0x1222, 0x141E, 0x1800}},
    {24562, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x1BFE, 0x5440, 0x5040, 0x50A0, 0x90A0, 0x10A0, 0x1110, 0x1110, 0x1288, 0x1248, 0x1444, 0x1802}},
    {24563, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1820, 0x5524, 0x5124, 0x5124, 0x9124, 0x11FC, 0x1024, 0x1020, 0x1022, 0x1022, 0x101E, 0x1000}},
    {24564, 2, {0x1040, 0x1040, 0x10A0, 0x18A0, 0x5510, 0x5288, 0x5446, 0x9040, 0x1000, 0x13F8, 0x1008, 0x1010, 0x1010, 0x1020, 0x1020, 0x1040}},
    {24565, 2, {0x100, 0x100, 0x280, 0x440, 0xA20, 0x3118, 0xC106, 0xFE0, 0x40, 0x80, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24566, 2, {0x1000, 0x1000, 0x11FC, 0x1800, 0x5400, 0x5000, 0x53FE, 0x9020, 0x1020, 0x1040, 0x1040, 0x1088, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {24567, 2, {0x1000, 0x1008, 0x1108, 0x1088, 0x1A52, 0x5652, 0x5222, 0x5222, 0x9252, 0x1292, 0x130A, 0x120A, 0x1202, 0x13FE, 0x1002, 0x1000}},
    {24568, 2, {0x1000, 0x13F8, 0x1088, 0x1088, 0x1888, 0x5488, 0x5088, 0x57F8, 0x9108, 0x1108, 0x1108, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {24569, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x1820, 0x5420, 0x5020, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {24570, 2, {0x1080, 0x1080, 0x1080, 0x10FC, 0x1904, 0x5508, 0x5240, 0x5440, 0x9040, 0x10A0, 0x10A0, 0x1090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {24571, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x1900, 0x5500, 0x51FE, 0x5110, 0x9110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {24572, 2, {0x1080, 0x1040, 0x1040, 0x17FC, 0x1800, 0x5400, 0x51F0, 0x5110, 0x9110, 0x1110, 0x1110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {24573, 2, {0x800, 0x800, 0x1FF8, 0x1248, 0x2248, 0x4488, 0x888, 0x1108, 0x2228, 0x410, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24574, 2, {0x1100, 0x1100, 0x11FC, 0x1A00, 0x5400, 0x51F8, 0x5000, 0x9000, 0x13F8, 0x1008, 0x1008, 0x1008, 0x100A, 0x100A, 0x1006, 0x1002}},
    {24575, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24576, 2, {0x1000, 0x1000, 0x17FE, 0x1020, 0x1820, 0x5440, 0x5040, 0x50D0, 0x9148, 0x1244, 0x1442, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24577, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x280, 0x440, 0xA20, 0x3118, 0xC006, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24578, 2, {0x820, 0x820, 0x820, 0x820, 0x1450, 0x1250, 0x2288, 0x2088, 0x4104, 0x8202, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24579, 2, {0x1000, 0x13FC, 0x1040, 0x1840, 0x5440, 0x5040, 0x57FE, 0x9040, 0x10A0, 0x10A0, 0x10A0, 0x1120, 0x1122, 0x1222, 0x141E, 0x1800}},
    {24580, 2, {0x1000, 0x13FC, 0x1200, 0x1208, 0x1A88, 0x5650, 0x5250, 0x5220, 0x9220, 0x1250, 0x1250, 0x1288, 0x1308, 0x1200, 0x13FE, 0x1000}},
    {24581, 2, {0x1100, 0x1108, 0x1108, 0x1110, 0x1920, 0x5540, 0x5100, 0x57FE, 0x9140, 0x1120, 0x1120, 0x1110, 0x1108, 0x1144, 0x1182, 0x1100}},
    {24582, 2, {0x1040, 0x1040, 0x10A0, 0x18A0, 0x5510, 0x5208, 0x5406, 0x91F0, 0x1110, 0x1110, 0x1150, 0x1120, 0x1104, 0x1104, 0x10FC, 0x1000}},
    {24583, 2, {0x1000, 0x13FE, 0x1200, 0x1200, 0x1A00, 0x57FC, 0x5204, 0x5204, 0x9204, 0x1204, 0x13FC, 0x1200, 0x1200, 0x1200, 0x13FE, 0x1000}},
    {24584, 2, {0x1048, 0x1248, 0x1248, 0x1248, 0x1A48, 0x57FE, 0x5248, 0x5248, 0x9248, 0x1248, 0x1278, 0x1200, 0x1200, 0x1200, 0x13FE, 0x1000}},
    {24585, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0x1A04, 0x55F4, 0x5114, 0x5114, 0x9114, 0x11F4, 0x1104, 0x1128, 0x1112, 0x1102, 0x10FE, 0x1000}},
    {24586, 2, {0x1000, 0x13FC, 0x1084, 0x1084, 0x1884, 0x5504, 0x5114, 0x5208, 0x9400, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24587, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0x1A04, 0x57FC, 0x5220, 0x5220, 0x93FE, 0x1220, 0x1220, 0x1210, 0x1212, 0x128A, 0x1306, 0x1202}},
    {24588, 2, {0x1000, 0x13FC, 0x1020, 0x1020, 0x1840, 0x5440, 0x50D0, 0x5148, 0x9244, 0x1444, 0x1040, 0x1040, 0x1040, 0x1000, 0x17FE, 0x1000}},
    {24589, 2, {0x1080, 0x1080, 0x1080, 0x18FE, 0x5540, 0x5140, 0x5240, 0x907C, 0x1040, 0x1040, 0x1040, 0x107E, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24590, 2, {0x800, 0x800, 0xFFC, 0x1400, 0x2400, 0x47F0, 0x400, 0x400, 0x7F8, 0x400, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24591, 2, {0x1040, 0x1040, 0x1040, 0x13F8, 0x1A48, 0x5648, 0x5248, 0x5248, 0x9248, 0x17FE, 0x1040, 0x10A0, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {24592, 2, {0x1080, 0x1080, 0x1100, 0x11FC, 0x1A04, 0x5404, 0x51E4, 0x5124, 0x9124, 0x1124, 0x11E4, 0x1124, 0x1004, 0x1004, 0x1028, 0x1010}},
    {24593, 2, {0x2040, 0x2244, 0x2148, 0x3150, 0xA840, 0xA7FC, 0xA040, 0xA040, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040, 0x2040}},
    {24594, 2, {0x1000, 0x1000, 0xFEFC, 0x2284, 0x4488, 0x2850, 0x1020, 0x2858, 0xC586, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24595, 2, {0x2100, 0x2100, 0x213E, 0x2112, 0x37D2, 0xAA52, 0xA252, 0xA254, 0xA254, 0x2494, 0x2288, 0x2108, 0x2294, 0x24A4, 0x2842, 0x2080}},
    {24596, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x1820, 0x5420, 0x5120, 0x5120, 0x913C, 0x1120, 0x1120, 0x1120, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {24597, 2, {0x1020, 0x1020, 0x1040, 0x11FC, 0x1904, 0x5504, 0x5104, 0x5104, 0x91FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24598, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0x1880, 0x54A0, 0x5120, 0x51FC, 0x9324, 0x1524, 0x1124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {24599, 2, {0x1040, 0x1040, 0x1040, 0x187E, 0x5440, 0x5040, 0x5040, 0x9040, 0x13FC, 0x1204, 0x1204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {24600, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24601, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x1BFE, 0x5420, 0x5020, 0x5020, 0x91FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24602, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x1908, 0x55F8, 0x5108, 0x5108, 0x9108, 0x11F8, 0x1108, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {24603, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0x1904, 0x5504, 0x51FC, 0x5104, 0x9104, 0x1104, 0x11FC, 0x1104, 0x1000, 0x1000, 0x13FE, 0x1000}},
    {24604, 2, {0x1040, 0x1040, 0x10A0, 0x18A0, 0x5510, 0x5248, 0x5426, 0x9020, 0x13F8, 0x1008, 0x1010, 0x1110, 0x10A0, 0x1040, 0x1020, 0x1020}},
    {24605, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24606, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x1BFE, 0x5622, 0x5222, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1222, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {24607, 2, {0x1008, 0x103C, 0x13E0, 0x1220, 0x1A20, 0x5620, 0x5220, 0x53FE, 0x9220, 0x1210, 0x1210, 0x1212, 0x120A, 0x128A, 0x1326, 0x1212}},
    {24608, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24609, 2, {0x1020, 0x1020, 0x1040, 0x1840, 0x5488, 0x5104, 0x53FE, 0x9002, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24610, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0x19FC, 0x5520, 0x5220, 0x5020, 0x93FE, 0x1020, 0x1050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {24611, 2, {0x900, 0x900, 0x1FF0, 0x2100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC106, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24612, 2, {0x810, 0x810, 0x17FE, 0x3010, 0x5210, 0x9110, 0x1110, 0x1050, 0x1020, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24613, 2, {0x800, 0xFE0, 0x1020, 0x2040, 0x5FF8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x3FF8, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24614, 2, {0x1000, 0x13F8, 0x1040, 0x1040, 0x1A48, 0x5548, 0x5150, 0x5040, 0x97FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24615, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0x19FC, 0x5520, 0x5220, 0x5020, 0x9020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24616, 2, {0x1000, 0x10F8, 0x3E88, 0x4288, 0xA2A8, 0x1490, 0x884, 0x1084, 0x207C, 0xC000, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24617, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1904, 0x55FC, 0x5100, 0x5140, 0x9144, 0x1148, 0x1170, 0x1140, 0x1142, 0x1242, 0x123E, 0x1400}},
    {24618, 2, {0x1000, 0x17F8, 0x1208, 0x1110, 0x18A0, 0x5440, 0x50A0, 0x5318, 0x9C46, 0x1040, 0x17FC, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000}},
    {24619, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0x1894, 0x5494, 0x53FC, 0x5290, 0x9290, 0x13FE, 0x1092, 0x1092, 0x111A, 0x1114, 0x1210, 0x1410}},
    {24620, 2, {0x1000, 0x1000, 0x13FE, 0x1252, 0x1A52, 0x5652, 0x5252, 0x5252, 0x9252, 0x1292, 0x128E, 0x1302, 0x1202, 0x1202, 0x13FE, 0x1202}},
    {24621, 2, {0x1000, 0x1040, 0x1020, 0x1028, 0x1808, 0x5488, 0x5090, 0x5294, 0x92A2, 0x12A2, 0x14C2, 0x1088, 0x1188, 0x1288, 0x1478, 0x1000}},
    {24622, 2, {0x2220, 0x2220, 0x2220, 0x3420, 0xAD7C, 0xA924, 0xAE24, 0xA224, 0x2424, 0x2424, 0x2924, 0x2F24, 0x2144, 0x2044, 0x2094, 0x2108}},
    {24623, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x1BFC, 0x5440, 0x5040, 0x5040, 0x97FE, 0x1040, 0x1080, 0x1080, 0x1110, 0x1208, 0x17FC, 0x1204}},
    {24624, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1840, 0x5440, 0x5088, 0x5108, 0x93F0, 0x1020, 0x1040, 0x1088, 0x1104, 0x13FC, 0x1104, 0x1000}},
    {24625, 2, {0x800, 0x800, 0x1FF8, 0x1248, 0x2A48, 0x4488, 0xA88, 0x1108, 0x22A8, 0x410, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24626, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0x1820, 0x57FE, 0x5222, 0x5222, 0x9252, 0x124A, 0x128A, 0x1302, 0x1202, 0x1202, 0x120A, 0x1204}},
    {24627, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0x1A04, 0x5604, 0x53FC, 0x5090, 0x9090, 0x1090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E, 0x1800}},
    {24628, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0x1BFE, 0x5620, 0x5224, 0x5224, 0x9224, 0x1228, 0x1228, 0x1290, 0x1312, 0x122A, 0x1046, 0x1082}},
    {24629, 2, {0x1020, 0x1028, 0x1024, 0x1020, 0x1BFE, 0x5420, 0x5070, 0x5070, 0x90A8, 0x10A8, 0x1124, 0x1222, 0x1420, 0x1020, 0x1020, 0x1020}},
    {24630, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1A22, 0x5624, 0x5220, 0x53FC, 0x9284, 0x1288, 0x1248, 0x1250, 0x1220, 0x1450, 0x1488, 0x1906}},
    {24631, 2, {0x110, 0x108, 0x100, 0xFFFE, 0x100, 0x1110, 0x1110, 0x1110, 0x2110, 0xC00E, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24632, 2, {0x120, 0x110, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24633, 2, {0x1040, 0x1040, 0x2240, 0x2278, 0x63C8, 0xAE48, 0x2268, 0x2252, 0x2242, 0x2202, 0x21FE, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24634, 2, {0x1040, 0x1020, 0x1010, 0x19E0, 0x5422, 0x5034, 0x53B8, 0x90B0, 0x10A8, 0x10A8, 0x1128, 0x1124, 0x1224, 0x1422, 0x10A0, 0x1040}},
    {24635, 2, {0x1010, 0x820, 0x440, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24636, 2, {0x8, 0x8, 0x7E08, 0x2FE, 0x2408, 0x1448, 0x828, 0x1408, 0x2228, 0x4010, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24637, 2, {0x1040, 0x1040, 0x1040, 0x13FC, 0x1840, 0x5440, 0x5040, 0x57FE, 0x90E0, 0x1150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040}},
    {24638, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0x1904, 0x5504, 0x5104, 0x5104, 0x91FC, 0x1104, 0x1000, 0x1090, 0x1088, 0x1104, 0x1202, 0x1402}},
    {24639, 2, {0x1000, 0x17F8, 0x1208, 0x1110, 0x18A0, 0x5440, 0x51B0, 0x564E, 0x9040, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x1040, 0x1040, 0x1040}},
    {24640, 2, {0x1040, 0x1040, 0x10FC, 0x1108, 0x1A90, 0x5460, 0x5048, 0x5190, 0x963E, 0x1042, 0x1084, 0x1348, 0x1030, 0x1020, 0x10C0, 0x1700}},
    {24641, 2, {0x838, 0xBC0, 0x1040, 0x3040, 0x57FC, 0x9040, 0x1040, 0x1040, 0x13F8, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24642, 2, {0x1080, 0x1080, 0x1100, 0x11FC, 0x1A04, 0x5404, 0x53E4, 0x5224, 0x9224, 0x13E4, 0x1224, 0x1224, 0x13E4, 0x1004, 0x1028, 0x1010}},
    {24643, 2, {0x1040, 0x1040, 0x1040, 0x1BFC, 0x5440, 0x5040, 0x57FE, 0x9010, 0x1010, 0x17FE, 0x1010, 0x1210, 0x1110, 0x1110, 0x1050, 0x1020}},
    {24644, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1820, 0x5420, 0x51FC, 0x5000, 0x9000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24645, 2, {0x1040, 0x1040, 0x1044, 0x19F4, 0x5448, 0x5050, 0x53FE, 0x9040, 0x1080, 0x1184, 0x1298, 0x14E0, 0x1082, 0x1082, 0x107E, 0x1000}},
    {24646, 2, {0x1000, 0x17FE, 0x1080, 0x1080, 0x1880, 0x54F8, 0x5108, 0x5148, 0x9128, 0x1208, 0x1288, 0x1050, 0x1010, 0x1010, 0x17FE, 0x1000}},
    {24647, 2, {0x1000, 0x13FE, 0x1200, 0x1200, 0x1BFC, 0x5620, 0x5220, 0x52F8, 0x9220, 0x1220, 0x1220, 0x13FC, 0x1200, 0x1200, 0x13FE, 0x1000}},
    {24648, 2, {0x1040, 0x1040, 0x1090, 0x1108, 0x1BFC, 0x5424, 0x5120, 0x5120, 0x91FC, 0x1220, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24649, 2, {0x1100, 0x1104, 0x1138, 0x11C0, 0x1902, 0x5502, 0x50FE, 0x5000, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24650, 2, {0x2040, 0x2040, 0x23FC, 0x3084, 0xA884, 0xA114, 0xA208, 0xA010, 0x2410, 0x2FBE, 0x2492, 0x2492, 0x2492, 0x28A2, 0x2AAA, 0x3144}},
    {24651, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x1448, 0x2444, 0x4444, 0x0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24652, 2, {0x1090, 0x1090, 0x1090, 0x1292, 0x1994, 0x5498, 0x5090, 0x5198, 0x9294, 0x1492, 0x1090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {24653, 2, {0x1040, 0x1044, 0x1244, 0x1948, 0x5550, 0x5040, 0x57FE, 0x9090, 0x1090, 0x1090, 0x1090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {24654, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x1888, 0x5504, 0x53FE, 0x5022, 0x9020, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {24655, 2, {0x1000, 0x10FC, 0xFE04, 0x2208, 0x2210, 0x64FE, 0x1410, 0x810, 0x3450, 0xC220, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24656, 2, {0x0, 0x7DF0, 0x1110, 0x1110, 0x1190, 0x1150, 0x1D12, 0xE212, 0x440E, 0x800, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24657, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0x1A08, 0x55FE, 0x5100, 0x517C, 0x9144, 0x1144, 0x1154, 0x1148, 0x1142, 0x1242, 0x123E, 0x1400}},
    {24658, 2, {0x1000, 0x13FE, 0x1000, 0x1000, 0x19FC, 0x5504, 0x5104, 0x51FC, 0x9104, 0x1104, 0x11FC, 0x1104, 0x1000, 0x1000, 0x13FE, 0x1000}},
    {24659, 2, {0x1000, 0x17FE, 0x1090, 0x1890, 0x5490, 0x53FC, 0x5294, 0x9294, 0x1294, 0x1294, 0x129C, 0x1304, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {24660, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1800, 0x5488, 0x5104, 0x5202, 0x9088, 0x1088, 0x1050, 0x1050, 0x1020, 0x1050, 0x1088, 0x1306}},
    {24661, 2, {0x1000, 0x1000, 0xFE7C, 0x2244, 0x4444, 0x3444, 0x844, 0x147C, 0x2244, 0xC000, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24662, 2, {0x200, 0x400, 0x3FF8, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x3FF8, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24663, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0x1888, 0x5504, 0x52FA, 0x5000, 0x93FE, 0x1040, 0x1080, 0x11FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {24664, 2, {0x1090, 0x1090, 0x1090, 0x1910, 0x55FE, 0x5310, 0x5510, 0x9138, 0x1138, 0x1154, 0x1154, 0x1192, 0x1110, 0x1110, 0x1110, 0x1110}},
    {24665, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24666, 2, {0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24667, 2, {0x1000, 0x1000, 0x13FC, 0x1204, 0x1A04, 0x56F4, 0x5294, 0x5294, 0x9294, 0x1294, 0x12F4, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1000}},
    {24668, 2, {0x1028, 0x1024, 0x1024, 0x1820, 0x57FE, 0x5020, 0x5020, 0x97E0, 0x1120, 0x1110, 0x1110, 0x1110, 0x11CA, 0x170A, 0x1206, 0x1002}},
    {24669, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x824, 0x7F44, 0x854, 0x888, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24670, 2, {0x1040, 0x1040, 0x17FC, 0x1040, 0x1BF8, 0x5448, 0x53F8, 0x5240, 0x93FC, 0x1044, 0x1054, 0x10A8, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {24671, 2, {0x1100, 0x1100, 0x1100, 0x19FE, 0x5602, 0x5202, 0x5512, 0x98A2, 0x124A, 0x12AA, 0x131A, 0x120A, 0x13FA, 0x1002, 0x1014, 0x1008}},
    {24672, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0x1880, 0x5490, 0x5110, 0x5110, 0x937C, 0x1510, 0x1110, 0x1110, 0x1110, 0x11FE, 0x1100, 0x1100}},
    {24673, 2, {0x1008, 0x1190, 0x1060, 0x1098, 0x1B04, 0x5440, 0x5040, 0x57FE, 0x9080, 0x1090, 0x1110, 0x1120, 0x1248, 0x1484, 0x19FC, 0x1084}},
    {24674, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0x1880, 0x5490, 0x5090, 0x5112, 0x9152, 0x1154, 0x1290, 0x1228, 0x1428, 0x1044, 0x1084, 0x1102}},
    {24675, 2, {0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24676, 2, {0x1040, 0x1040, 0x1080, 0x1100, 0x1BFC, 0x5694, 0x5294, 0x5294, 0x9294, 0x1294, 0x1294, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {24677, 2, {0x0, 0xFF20, 0x2410, 0x2410, 0x3C00, 0x2440, 0x2440, 0x3D44, 0x2542, 0x2542, 0x2E42, 0xF448, 0x4448, 0x448, 0x438, 0x400}},
    {24678, 2, {0x1040, 0x1040, 0x1080, 0x13FC, 0x1A04, 0x5604, 0x52F4, 0x5294, 0x9294, 0x1294, 0x1294, 0x12F4, 0x1204, 0x1204, 0x1214, 0x1208}},
    {24679, 2, {0x0, 0xFFFE, 0x200, 0x400, 0x3FF8, 0x2488, 0x2488, 0x2488, 0x24A8, 0x2010, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24680, 2, {0x1000, 0x13F8, 0x1208, 0x1208, 0x1BF8, 0x5608, 0x5208, 0x53F8, 0x9244, 0x1248, 0x1230, 0x1220, 0x1210, 0x1288, 0x1306, 0x1200}},
    {24681, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x2288, 0x2448, 0x2828, 0x3FF8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24682, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x1B10, 0x54A0, 0x5040, 0x50A0, 0x9118, 0x1206, 0x15F8, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {24683, 2, {0x2000, 0x27FC, 0x2404, 0x2404, 0x35F4, 0xAC04, 0xA404, 0xA5F4, 0xA514, 0x2514, 0x2514, 0x25F4, 0x2404, 0x2404, 0x2414, 0x2408}},
    {24684, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x1820, 0x55FE, 0x5020, 0x5020, 0x9020, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24685, 2, {0x440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x1210, 0x2208, 0xC226, 0x1290, 0x1248, 0x2248, 0xA00, 0x400}},
    {24686, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0x1A08, 0x5406, 0x53F8, 0x5040, 0x9040, 0x1040, 0x13F8, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {24687, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24688, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0x1A08, 0x5406, 0x53F8, 0x5000, 0x9000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {24689, 2, {0x1090, 0x1090, 0x1108, 0x1148, 0x1A44, 0x5492, 0x5108, 0x53FC, 0x9094, 0x1090, 0x1090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {24690, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0x1BFE, 0x5488, 0x5088, 0x5088, 0x9088, 0x17FE, 0x1088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {24691, 2, {0x0, 0x3FF0, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x2208, 0x2190, 0x2C60, 0x301C, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24692, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24693, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24694, 2, {0x0, 0x7FFC, 0x440, 0x2448, 0x1448, 0x1450, 0x440, 0xFFFE, 0x0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24695, 2, {0x840, 0x840, 0x1040, 0x37FC, 0x50E0, 0x9150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24696, 2, {0x2010, 0x2010, 0x2790, 0x3010, 0xA87E, 0xA012, 0xAFD2, 0xA212, 0x2212, 0x2422, 0x2522, 0x28A2, 0x2FC2, 0x2042, 0x2094, 0x2108}},
    {24697, 2, {0x1000, 0x13FE, 0x1220, 0x1228, 0x1A24, 0x5620, 0x53FE, 0x5220, 0x9220, 0x1250, 0x1250, 0x1250, 0x1488, 0x1488, 0x1904, 0x1202}},
    {24698, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x19FC, 0x5400, 0x51FC, 0x5004, 0x9004, 0x11FC, 0x1100, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {24699, 2, {0x2004, 0x27C4, 0x2444, 0x2454, 0x3554, 0xAD54, 0xA554, 0xA554, 0xA554, 0x2554, 0x2554, 0x2104, 0x2284, 0x2244, 0x2414, 0x2808}},
    {24700, 2, {0x1040, 0x1020, 0x1020, 0x17FE, 0x1800, 0x5400, 0x5090, 0x5254, 0x9224, 0x1224, 0x1254, 0x1294, 0x1204, 0x1204, 0x13FC, 0x1004}},
    {24701, 2, {0x1000, 0x13FE, 0x1202, 0x1C44, 0x5440, 0x53FC, 0x5080, 0x90A0, 0x1120, 0x11FC, 0x1020, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020}},
    {24702, 2, {0x1090, 0x1090, 0x17FE, 0x1090, 0x1890, 0x5440, 0x5020, 0x57FE, 0x9100, 0x1100, 0x1100, 0x1100, 0x1100, 0x1100, 0x11FC, 0x1000}},
    {24703, 2, {0x3FF0, 0x660, 0x180, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x2018, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24704, 2, {0x1000, 0x13FC, 0x1008, 0x18D0, 0x5420, 0x53FE, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1222, 0x1222, 0x122A, 0x1204}},
    {24705, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x19F8, 0x5400, 0x53FC, 0x5204, 0x9204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1214, 0x1208}},
    {24706, 2, {0x1220, 0x1224, 0x1228, 0x13B0, 0x1A20, 0x5622, 0x52A2, 0x531E, 0x9040, 0x1040, 0x13FC, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {24707, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0x3444, 0xAFFC, 0xA444, 0xA4C4, 0xA4E4, 0x2554, 0x2644, 0x2444, 0x2444, 0x2404, 0x27FC, 0x2404}},
    {24708, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0x1820, 0x55FC, 0x5104, 0x5104, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {24709, 2, {0x1010, 0x1090, 0x1088, 0x1108, 0x1A04, 0x55FA, 0x5108, 0x5108, 0x9108, 0x11F8, 0x1090, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {24710, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x3FF8, 0x1110, 0x2508, 0x200, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24711, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0x1A08, 0x55F6, 0x5040, 0x5040, 0x97FC, 0x1040, 0x1250, 0x1248, 0x1444, 0x1844, 0x1140, 0x1080}},
    {24712, 2, {0x1010, 0x1014, 0x1012, 0x1010, 0x1BFE, 0x5410, 0x5150, 0x5154, 0x97F4, 0x1154, 0x1154, 0x1148, 0x124A, 0x125A, 0x1426, 0x1042}},
    {24713, 2, {0xF8, 0x3F00, 0x1110, 0x920, 0x100, 0xFFFE, 0x920, 0x1110, 0x2108, 0xC006, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24714, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24715, 2, {0x1040, 0x1020, 0x17FE, 0x1108, 0x1890, 0x5460, 0x5060, 0x5198, 0x9606, 0x11F8, 0x1108, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {24716, 2, {0x1108, 0x1088, 0x1090, 0x13FC, 0x1824, 0x5424, 0x53FC, 0x5220, 0x9220, 0x13FE, 0x1062, 0x10A2, 0x112A, 0x1224, 0x1420, 0x1020}},
    {24717, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1904, 0x55FC, 0x5000, 0x5000, 0x91FC, 0x1020, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24718, 2, {0x1020, 0x1120, 0x1120, 0x11FE, 0x1A20, 0x5420, 0x5020, 0x53FE, 0x9000, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24719, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0x1820, 0x5420, 0x5124, 0x5124, 0x9124, 0x12AA, 0x1020, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {24720, 2, {0x4420, 0x2820, 0x10A4, 0x28A4, 0x48A8, 0x9920, 0x2820, 0x4850, 0x8850, 0x888, 0x5104, 0x2202, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24721, 2, {0x1048, 0x1044, 0x13FE, 0x1840, 0x5440, 0x53FC, 0x5244, 0x9244, 0x13FC, 0x1244, 0x1244, 0x13FC, 0x1244, 0x1244, 0x1254, 0x1208}},
    {24722, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x19F8, 0x5400, 0x53FC, 0x5244, 0x9244, 0x1244, 0x13FC, 0x1200, 0x1202, 0x1202, 0x11FE, 0x1000}},
    {24723, 2, {0x1000, 0x13F8, 0x1208, 0x1208, 0x1BF8, 0x5608, 0x5208, 0x53F8, 0x9208, 0x1208, 0x13F8, 0x1120, 0x1120, 0x1222, 0x1422, 0x181E}},
    {24724, 2, {0x1100, 0x1100, 0x11FC, 0x1A00, 0x55F8, 0x5108, 0x5148, 0x9128, 0x17FE, 0x1108, 0x1248, 0x1228, 0x13FC, 0x1008, 0x1050, 0x1020}},
    {24725, 2, {0x2208, 0x2110, 0x20E0, 0x3110, 0xAA08, 0xA080, 0xAFFE, 0xA140, 0x2240, 0x27FC, 0x2A44, 0x2244, 0x2254, 0x2248, 0x2040, 0x2040}},
    {24726, 2, {0x1020, 0x1020, 0x13FE, 0x1820, 0x5420, 0x53FE, 0x5202, 0x9404, 0x11F8, 0x1010, 0x1020, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {24727, 2, {0x1080, 0x1080, 0x11F8, 0x1A08, 0x5410, 0x53FC, 0x5244, 0x9244, 0x1244, 0x13FC, 0x10A0, 0x10A0, 0x1122, 0x1122, 0x121E, 0x1400}},
    {24728, 2, {0x0, 0x3FF8, 0x2400, 0x27F0, 0x2880, 0x3FFC, 0x2140, 0x2220, 0x2410, 0x3FFC, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24729, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x19F8, 0x5508, 0x5108, 0x51F8, 0x9000, 0x13FC, 0x1008, 0x1030, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {24730, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x1820, 0x55FC, 0x5124, 0x5124, 0x91FC, 0x1020, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {24731, 2, {0x1040, 0x1040, 0x1088, 0x1104, 0x1BFE, 0x5402, 0x5088, 0x5144, 0x9242, 0x10F8, 0x1188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {24732, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x1904, 0x55FC, 0x5000, 0x5000, 0x93FE, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24733, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0x37FC, 0xAC44, 0xA444, 0xA7FC, 0xA040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x2000}},
    {24734, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0x1AF8, 0x5600, 0x5200, 0x53FC, 0x9044, 0x1044, 0x17FE, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {24735, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0x19FC, 0x5484, 0x5084, 0x53FE, 0x9000, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24736, 2, {0x1040, 0x1040, 0x247E, 0x6488, 0xA548, 0x2450, 0x2420, 0x2058, 0x2086, 0x2300, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24737, 2, {0x604, 0x7804, 0x844, 0x844, 0xFF44, 0x1C44, 0x2A44, 0x4904, 0x8814, 0x808, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24738, 2, {0x1080, 0x1040, 0x13F8, 0x1208, 0x1A08, 0x57F8, 0x5208, 0x5208, 0x93F8, 0x1240, 0x1244, 0x1228, 0x1210, 0x1288, 0x1306, 0x1200}},
    {24739, 2, {0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24740, 2, {0x200, 0x400, 0x3FF8, 0x2208, 0x23C8, 0x2448, 0x2A88, 0x2108, 0x2288, 0x2408, 0x3FF8, 0x2008, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24741, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24742, 2, {0x1208, 0x1108, 0x1110, 0x1020, 0x1BF8, 0x5608, 0x5208, 0x5208, 0x93F8, 0x10A0, 0x10A0, 0x1120, 0x1122, 0x1222, 0x141E, 0x1800}},
    {24743, 2, {0x1084, 0x11C4, 0x1704, 0x1114, 0x1914, 0x5514, 0x57D4, 0x5114, 0x9314, 0x1394, 0x1554, 0x1504, 0x1904, 0x1104, 0x1114, 0x1108}},
    {24744, 2, {0x900, 0x900, 0x11FC, 0x3204, 0x5448, 0x9950, 0x1148, 0x1244, 0x1444, 0x1140, 0x1080, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24745, 2, {0x1042, 0x1222, 0x1124, 0x1104, 0x1808, 0x5400, 0x520A, 0x528A, 0x9252, 0x1222, 0x1252, 0x128A, 0x130A, 0x1202, 0x13FE, 0x1002}},
    {24746, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x440, 0xFFFE, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24747, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0x7FFE, 0x4002, 0x8FE4, 0x820, 0x1022, 0x2022, 0xC11E, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24748, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x1010, 0x3FF8, 0x1008, 0x100, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24749, 2, {0x1080, 0x12FC, 0x12A4, 0x12A4, 0x1AA8, 0x5690, 0x52A8, 0x50C6, 0x9020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24750, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0x33F8, 0xA800, 0xA000, 0xA7FC, 0xA040, 0x2040, 0x2FFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {24751, 2, {0x1100, 0x10BE, 0x1282, 0x1202, 0x1A42, 0x5622, 0x53FE, 0x5292, 0x9252, 0x1222, 0x1252, 0x128A, 0x130A, 0x1202, 0x120A, 0x1204}},
    {24752, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1800, 0x55FC, 0x5000, 0x5000, 0x93FE, 0x1020, 0x1128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {24753, 2, {0x1090, 0x1090, 0x1090, 0x179E, 0x1890, 0x5490, 0x5090, 0x539C, 0x9090, 0x1090, 0x1090, 0x179E, 0x1090, 0x1090, 0x1090, 0x1090}},
    {24754, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x3C78, 0x440, 0x440, 0x7C7C, 0x440, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24755, 2, {0x80, 0x3FFE, 0x80, 0x7F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2000, 0x3FFE, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24756, 2, {0x1080, 0x1040, 0x13FC, 0x1800, 0x5510, 0x5110, 0x52A8, 0x9444, 0x1000, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24757, 2, {0x1000, 0x10FC, 0x1080, 0x18F8, 0x5480, 0x50F8, 0x5080, 0x97FE, 0x1140, 0x1124, 0x1128, 0x1110, 0x1108, 0x1144, 0x1182, 0x1100}},
    {24758, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4104, 0x4504, 0x5424, 0x5454, 0x6454, 0x43C4, 0x4004, 0x4014, 0x4008}},
    {24759, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0x1A04, 0x57FC, 0x5200, 0x5228, 0x9224, 0x13FE, 0x1220, 0x1250, 0x1250, 0x1488, 0x1504, 0x1A02}},
    {24760, 2, {0x2038, 0x23C0, 0x2040, 0x27FC, 0x30E0, 0xA950, 0xA248, 0xAC46, 0xA3F0, 0x2020, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2140, 0x2080}},
    {24761, 2, {0x100, 0x7FFE, 0x4002, 0x9FE4, 0x1020, 0x1FE0, 0x1000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24762, 2, {0x1040, 0x1020, 0x13FE, 0x1A02, 0x5404, 0x51F8, 0x5108, 0x9108, 0x11F8, 0x1100, 0x1100, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24763, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x1820, 0x57FE, 0x5088, 0x5050, 0x93FE, 0x1020, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {24764, 2, {0x1040, 0x1040, 0x107E, 0x1040, 0x1BFC, 0x5604, 0x53FC, 0x5204, 0x93FC, 0x1244, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1040}},
    {24765, 2, {0x1040, 0x1040, 0x17FC, 0x1040, 0x1BF8, 0x5448, 0x57FE, 0x5048, 0x93F8, 0x1080, 0x17FE, 0x1108, 0x1390, 0x1060, 0x1198, 0x1604}},
    {24766, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x1A02, 0x5494, 0x5108, 0x5204, 0x9000, 0x11FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {24767, 2, {0x1000, 0x13FC, 0x1040, 0x1040, 0x1BFE, 0x54A0, 0x5110, 0x5208, 0x9446, 0x1040, 0x1048, 0x1264, 0x1252, 0x1452, 0x1140, 0x1080}},
    {24768, 2, {0x2040, 0x2040, 0x20A0, 0x3110, 0xAA08, 0xADF6, 0xA000, 0xA7FC, 0x24A4, 0x24A4, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x2414, 0x2408}},
    {24769, 2, {0x80C, 0x8F0, 0x7E80, 0x1880, 0x2CFE, 0x2A88, 0x4888, 0x8888, 0x908, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24770, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x1A10, 0x5400, 0x5040, 0x539C, 0x9204, 0x1204, 0x139C, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {24771, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0x1BFC, 0x5604, 0x5204, 0x53FC, 0x9000, 0x1220, 0x1224, 0x13A8, 0x1230, 0x1222, 0x12A2, 0x131E}},
    {24772, 2, {0x800, 0xF7C, 0x844, 0x844, 0xFF48, 0x828, 0x2A30, 0x4910, 0x8928, 0x2844, 0x1082, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24773, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x51FC, 0x5820, 0x57FF, 0x9400, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x110C}},
    {24774, 2, {0x1000, 0x13FE, 0x1222, 0x1222, 0x1AFA, 0x5622, 0x5222, 0x53FE, 0x9202, 0x12FA, 0x128A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1404}},
    {24775, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x19FC, 0x5504, 0x51FC, 0x5000, 0x91FC, 0x1008, 0x1010, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {24776, 2, {0x1000, 0x13F8, 0x1248, 0x1248, 0x1BF8, 0x5648, 0x5248, 0x53F8, 0x9040, 0x17FC, 0x10E0, 0x1150, 0x1248, 0x1446, 0x1040, 0x1040}},
    {24777, 2, {0x1040, 0x840, 0x407E, 0x2040, 0x840, 0x13FC, 0x7204, 0x1204, 0x13FC, 0x1000, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24778, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x1800, 0x55FC, 0x5104, 0x5104, 0x91FC, 0x1020, 0x1128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {24779, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1D04, 0x5500, 0x51DE, 0x5252, 0x9252, 0x1352, 0x149A, 0x1094, 0x1110, 0x1112, 0x1212, 0x140E}},
    {24780, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x9004, 0x1EF8, 0x2288, 0x24A8, 0x5490, 0x884, 0x1084, 0x607C, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24781, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0x1A08, 0x5406, 0x51F0, 0x5040, 0x9040, 0x13F8, 0x1040, 0x1248, 0x1148, 0x1150, 0x17FC, 0x1000}},
    {24782, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1010, 0x2208, 0xC106, 0x900, 0x2828, 0x2824, 0x47E4, 0x0}},
    {24783, 2, {0x2110, 0x2110, 0x2110, 0x2110, 0x37BC, 0xA910, 0xA110, 0xA338, 0xA3B8, 0x2554, 0x2554, 0x2992, 0x2110, 0x2110, 0x2110, 0x2110}},
    {24784, 2, {0x1014, 0x1012, 0x1010, 0x17FE, 0x1810, 0x5410, 0x53D2, 0x5252, 0x9252, 0x1254, 0x13D4, 0x1008, 0x10EA, 0x171A, 0x1226, 0x1042}},
    {24785, 2, {0x50, 0x48, 0xFFFE, 0x40, 0x3E40, 0x2224, 0x3E28, 0x12, 0xE2A, 0x70C6, 0x302, 0x100, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24786, 2, {0x600, 0x787C, 0x844, 0x844, 0xFF44, 0x1C44, 0x2A44, 0x497C, 0x8844, 0x800, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24787, 2, {0x1020, 0x1124, 0x10A8, 0x1BFE, 0x5440, 0x5080, 0x57FE, 0x9108, 0x1204, 0x15FA, 0x1908, 0x1108, 0x1128, 0x1112, 0x1102, 0x10FE}},
    {24788, 2, {0x1040, 0x1244, 0x1244, 0x1448, 0x18A0, 0x5510, 0x5608, 0x5044, 0x9040, 0x1248, 0x1248, 0x1450, 0x10A0, 0x1110, 0x1208, 0x1C06}},
    {24789, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1904, 0x55FC, 0x5080, 0x5080, 0x91FE, 0x124A, 0x144A, 0x1092, 0x1122, 0x1242, 0x1094, 0x1108}},
    {24790, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FF8, 0x2248, 0x4488, 0x908, 0x1228, 0x2410, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24791, 2, {0x2040, 0x2040, 0x20A0, 0x3110, 0xAA88, 0xA446, 0xA800, 0xA3F0, 0x2010, 0x2020, 0x2040, 0x2124, 0x2522, 0x250A, 0x2908, 0x20F8}},
    {24792, 2, {0x1000, 0x13FE, 0x1202, 0x128A, 0x1A52, 0x57FE, 0x5242, 0x5222, 0x93FE, 0x1282, 0x1282, 0x1282, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {24793, 2, {0x2000, 0x27BC, 0x2084, 0x2294, 0x3108, 0xAA94, 0xA4A4, 0xA840, 0xA000, 0x27BC, 0x20A4, 0x22A4, 0x2128, 0x2290, 0x24A8, 0x2846}},
    {24794, 2, {0x1080, 0x1080, 0x10FC, 0x1154, 0x1A54, 0x54A4, 0x5124, 0x5244, 0x9094, 0x1108, 0x1040, 0x10A4, 0x12AA, 0x128A, 0x1478, 0x1000}},
    {24795, 2, {0x101C, 0x13E0, 0x1220, 0x13FE, 0x1A20, 0x5692, 0x530A, 0x5206, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24796, 2, {0x1088, 0x1088, 0x1088, 0x13FE, 0x1888, 0x5488, 0x57FE, 0x5000, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24797, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0x1820, 0x57FE, 0x5202, 0x5202, 0x92FA, 0x128A, 0x128A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {24798, 2, {0x2010, 0x2090, 0x2710, 0x243E, 0x3422, 0xAC44, 0xA790, 0xA510, 0xA510, 0x2510, 0x2528, 0x2528, 0x2528, 0x2944, 0x2944, 0x3082}},
    {24799, 2, {0x10A0, 0x1090, 0x1080, 0x11FE, 0x1910, 0x5710, 0x55FC, 0x5110, 0x9110, 0x11FC, 0x1110, 0x1110, 0x1110, 0x11FE, 0x1100, 0x1100}},
    {24800, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0xFFFE, 0x102, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24801, 2, {0x0, 0x7FFC, 0x440, 0x3C78, 0x2008, 0x2008, 0x3C78, 0x440, 0xFFFE, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24802, 2, {0x100, 0x80, 0x888, 0x4804, 0x4812, 0x4812, 0x87F0, 0x0, 0x1010, 0x808, 0x2220, 0xA124, 0xA4A2, 0xA4A4, 0x1D1C, 0x0}},
    {24803, 2, {0x1080, 0x5080, 0x7CFC, 0x5154, 0x9254, 0x1CA4, 0xF124, 0x1244, 0x1094, 0x1108, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24804, 2, {0x1010, 0x1388, 0x1088, 0x10FE, 0x1890, 0x5790, 0x5224, 0x5224, 0x9238, 0x1388, 0x1090, 0x1090, 0x10A4, 0x10BE, 0x1282, 0x1100}},
    {24805, 2, {0x100, 0x500, 0x3978, 0x2108, 0x3D78, 0x2108, 0x3FF8, 0x280, 0x440, 0x1830, 0xE10E, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24806, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0x1A20, 0x5620, 0x523E, 0x5220, 0x9220, 0x12FC, 0x1284, 0x1284, 0x1284, 0x1484, 0x14FC, 0x1884}},
    {24807, 2, {0x1000, 0x13F8, 0x1208, 0x1208, 0x1BF8, 0x5608, 0x53F8, 0x5208, 0x93F8, 0x1208, 0x1208, 0x1FFE, 0x1000, 0x1110, 0x1208, 0x1404}},
    {24808, 2, {0x2080, 0x2090, 0x2108, 0x33FC, 0xA840, 0xA7FE, 0xA110, 0xA248, 0x2C86, 0x2310, 0x2020, 0x20C4, 0x2308, 0x2030, 0x20C0, 0x2700}},
    {24809, 2, {0x1000, 0x23FE, 0x4020, 0x8920, 0x113C, 0x3120, 0x5120, 0x9120, 0x17FE, 0x1000, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24810, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24811, 2, {0x800, 0xFF0, 0x1020, 0x2C40, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24812, 2, {0x2000, 0x27FE, 0x2420, 0x2420, 0x35FC, 0xAC20, 0xA524, 0xA4A8, 0xA7FE, 0x2420, 0x2450, 0x2448, 0x2484, 0x2504, 0x27FE, 0x2000}},
    {24813, 2, {0x2100, 0x2102, 0x211C, 0x2FD0, 0x3210, 0xAA10, 0xA51E, 0xA7D4, 0xA114, 0x2114, 0x21D4, 0x2F14, 0x2514, 0x2124, 0x2124, 0x2144}},
    {24814, 2, {0x1104, 0x1088, 0x1050, 0x11FC, 0x1924, 0x5524, 0x51FC, 0x5124, 0x9124, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {24815, 2, {0x11FC, 0x1124, 0x1124, 0x13FE, 0x1924, 0x5524, 0x51FC, 0x5000, 0x91FC, 0x1104, 0x1124, 0x1124, 0x1124, 0x1058, 0x1084, 0x1302}},
    {24816, 2, {0x1040, 0x1040, 0x13FE, 0x1080, 0x19FC, 0x5620, 0x55FE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {24817, 2, {0x1092, 0x1124, 0x1248, 0x1124, 0x1892, 0x5440, 0x5080, 0x53FE, 0x9202, 0x128A, 0x1252, 0x1222, 0x1252, 0x128A, 0x13FE, 0x1202}},
    {24818, 2, {0x1000, 0x13FE, 0x1222, 0x1020, 0x1BFE, 0x5420, 0x51FC, 0x5124, 0x91FC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {24819, 2, {0x800, 0x8F8, 0x888, 0x7E88, 0x8F8, 0x1888, 0x1CF8, 0x2A88, 0x4888, 0x8F8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24820, 2, {0x2040, 0x2444, 0x2444, 0x3444, 0xAFFC, 0xA000, 0xAFFE, 0xA040, 0x2080, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x240C}},
    {24821, 2, {0x1090, 0x1290, 0x1290, 0x17FE, 0x1A90, 0x5690, 0x52F0, 0x5200, 0x93FC, 0x1040, 0x17FE, 0x10E0, 0x1150, 0x1248, 0x1C46, 0x1040}},
    {24822, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0x19FC, 0x5504, 0x51FC, 0x5000, 0x93FE, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24823, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24824, 2, {0x1100, 0x11FE, 0x1202, 0x15F2, 0x1912, 0x55F2, 0x5112, 0x51FA, 0x9004, 0x13F8, 0x1010, 0x1020, 0x17FE, 0x1020, 0x10A0, 0x1040}},
    {24825, 2, {0x820, 0x7FFC, 0x820, 0x200, 0xFFFE, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24826, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1904, 0x55FC, 0x5020, 0x5120, 0x91FE, 0x1220, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24827, 2, {0x2002, 0x27C2, 0x2442, 0x2452, 0x37D2, 0xAC52, 0xA452, 0xA7D2, 0xA452, 0x2452, 0x27D2, 0x2002, 0x2282, 0x2242, 0x244A, 0x2804}},
    {24828, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1A02, 0x57FE, 0x5200, 0x5200, 0x93FE, 0x1352, 0x1552, 0x15FE, 0x1552, 0x1552, 0x1952, 0x1106}},
    {24829, 2, {0x1000, 0x13FC, 0x1204, 0x13FC, 0x1A20, 0x57FE, 0x5210, 0x528A, 0x9306, 0x1000, 0x13FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1204}},
    {24830, 2, {0x1000, 0x1088, 0x1252, 0x1222, 0x1A52, 0x568A, 0x5202, 0x53FE, 0x9088, 0x1144, 0x127A, 0x1088, 0x1150, 0x1020, 0x10D8, 0x1706}},
    {24831, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x19FC, 0x5504, 0x5104, 0x51FC, 0x9000, 0x13FE, 0x1020, 0x1120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {24832, 2, {0x2088, 0x21C8, 0x2708, 0x3108, 0xA92A, 0xA12A, 0xA7AC, 0xA148, 0x2308, 0x2388, 0x2554, 0x2514, 0x2914, 0x2124, 0x2124, 0x2142}},
    {24833, 2, {0xE20, 0x7820, 0x824, 0x8A4, 0xFEA8, 0x1920, 0x2C50, 0x4A50, 0x8888, 0x904, 0xA02, 0x100, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24834, 2, {0x810, 0x7F10, 0x810, 0x7F7E, 0x4112, 0x3E12, 0x412, 0xF12, 0x7822, 0x822, 0x284A, 0x1084, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24835, 2, {0x1040, 0x1020, 0x13FE, 0x1A02, 0x5404, 0x51FC, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1000, 0x13FE, 0x1000}},
    {24836, 2, {0x1000, 0x13F8, 0x1248, 0x1248, 0x1BF8, 0x5648, 0x5248, 0x53F8, 0x9000, 0x17FE, 0x1240, 0x1244, 0x1228, 0x1290, 0x1308, 0x1206}},
    {24837, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0x31F0, 0xA840, 0xA3F8, 0xA248, 0xA248, 0x23F8, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x2040}},
    {24838, 2, {0x1200, 0x213C, 0x4800, 0x9400, 0x217E, 0x6208, 0xAE08, 0x2208, 0x2228, 0x2210, 0x2100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24839, 2, {0x1040, 0x1040, 0x11FC, 0x1084, 0x1BFE, 0x5400, 0x51FC, 0x5104, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1220, 0x13FE, 0x1020, 0x1020}},
    {24840, 2, {0x100, 0x280, 0xC60, 0x37D8, 0xC006, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2618, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24841, 2, {0x1040, 0x10A0, 0x1110, 0x1208, 0x1DF6, 0x5400, 0x53C4, 0x5254, 0x9254, 0x13D4, 0x1254, 0x1254, 0x13D4, 0x1244, 0x1254, 0x12C8}},
    {24842, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0x1904, 0x5504, 0x51FC, 0x5000, 0x93FE, 0x1222, 0x1222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {24843, 2, {0x103C, 0x17C0, 0x1244, 0x1928, 0x5400, 0x53FC, 0x5080, 0x9080, 0x17FE, 0x1100, 0x11F8, 0x1288, 0x1250, 0x1420, 0x18D8, 0x1306}},
    {24844, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0x1A10, 0x55FC, 0x5104, 0x5154, 0x918C, 0x1124, 0x1020, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1602}},
    {24845, 2, {0x20, 0x7F20, 0x413E, 0x7F44, 0x48A4, 0x7F24, 0x4828, 0x4510, 0x5328, 0x6146, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24846, 2, {0x1100, 0x1100, 0x11FE, 0x1A00, 0x55FC, 0x5104, 0x51FC, 0x9104, 0x11FC, 0x1080, 0x10FC, 0x1108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {24847, 2, {0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24848, 2, {0x1000, 0x17FE, 0x1040, 0x1080, 0x1BFC, 0x5694, 0x5294, 0x52F4, 0x9294, 0x1294, 0x12F4, 0x1294, 0x1294, 0x1294, 0x13FC, 0x1204}},
    {24849, 2, {0x1000, 0x13FC, 0x1088, 0x1050, 0x1BFE, 0x5622, 0x53FE, 0x5222, 0x93FE, 0x1222, 0x1040, 0x13FE, 0x1042, 0x1082, 0x1114, 0x1208}},
    {24850, 2, {0x1000, 0x13F8, 0x1208, 0x1BF8, 0x5608, 0x53F8, 0x5100, 0x93FC, 0x1444, 0x1244, 0x12A4, 0x1204, 0x13F4, 0x1004, 0x1028, 0x1010}},
    {24851, 2, {0x2000, 0x23F8, 0x2208, 0x33F8, 0xAA08, 0xA3F8, 0xA000, 0xAFFE, 0x2200, 0x23FC, 0x2494, 0x2924, 0x2244, 0x2484, 0x2128, 0x2210}},
    {24852, 2, {0x2080, 0x2040, 0x27FC, 0x3000, 0xAA08, 0xA110, 0xAFFE, 0xA000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {24853, 2, {0x2000, 0x27BC, 0x24A4, 0x24A4, 0x37BC, 0xA800, 0xA3F8, 0xA000, 0xA7FC, 0x2100, 0x2200, 0x23F8, 0x2008, 0x2008, 0x2050, 0x2020}},
    {24854, 2, {0x1110, 0x1110, 0x17FC, 0x1110, 0x19F0, 0x5510, 0x51F0, 0x5110, 0x9110, 0x17FE, 0x1200, 0x1290, 0x1308, 0x1200, 0x13FC, 0x1000}},
    {24855, 2, {0x40, 0x7F40, 0x127E, 0xC88, 0xFF48, 0x1948, 0x2A50, 0x4820, 0xA850, 0x108C, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24856, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x1880, 0x54F8, 0x5108, 0x5290, 0x9060, 0x1198, 0x1606, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {24857, 2, {0x200, 0x100, 0x7FFE, 0x4802, 0x9FF4, 0x2820, 0x7C0, 0x1830, 0xE00E, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24858, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2148, 0x2FE8, 0x2428, 0x2010, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24859, 2, {0x7C, 0x3F80, 0x1108, 0x890, 0x7FFE, 0x4102, 0x9494, 0x1428, 0x23E8, 0x800, 0xFF0, 0x1420, 0x2240, 0x4180, 0xE60, 0x701C}},
    {24860, 2, {0x2000, 0x27FE, 0x2420, 0x2420, 0x35FC, 0xAC20, 0xA4A8, 0xA4A8, 0xA574, 0x2420, 0x2450, 0x2448, 0x2484, 0x2500, 0x27FE, 0x2000}},
    {24861, 2, {0x1000, 0x13FE, 0x1200, 0x12FC, 0x1A84, 0x56FC, 0x5284, 0x52FC, 0x9220, 0x13FE, 0x1248, 0x12C8, 0x1230, 0x1248, 0x1284, 0x13FE}},
    {24862, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x1BFE, 0x5652, 0x5252, 0x5252, 0x9242, 0x1226, 0x1020, 0x17FE, 0x1050, 0x1088, 0x1104, 0x1602}},
    {24863, 2, {0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2024, 0x2FA8, 0x2898, 0x2892, 0x4FAA, 0x4046, 0x8082, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24864, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x19F8, 0x5508, 0x5108, 0x51F8, 0x9000, 0x13FC, 0x1294, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {24865, 2, {0x1080, 0x1080, 0x10FC, 0x1154, 0x1AD4, 0x54B4, 0x512C, 0x5244, 0x9094, 0x1108, 0x1040, 0x10A4, 0x12AA, 0x128A, 0x1478, 0x1000}},
    {24866, 2, {0x1000, 0x13FC, 0x1244, 0x1244, 0x1BFC, 0x5644, 0x5244, 0x53FC, 0x9000, 0x1040, 0x1024, 0x1522, 0x150A, 0x1908, 0x10F8, 0x1000}},
    {24867, 2, {0x1000, 0x13FC, 0x1294, 0x1A94, 0x5694, 0x53FC, 0x5080, 0x9040, 0x17FE, 0x1080, 0x10F8, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {24868, 2, {0x1020, 0x11FC, 0x1020, 0x1088, 0x1BFE, 0x5488, 0x5000, 0x51FC, 0x9104, 0x1124, 0x1124, 0x1124, 0x1124, 0x1050, 0x1088, 0x1304}},
    {24869, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0x3150, 0xA840, 0xA3F8, 0xA248, 0xA248, 0x2248, 0x2FFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2406}},
    {24870, 2, {0x1020, 0x1020, 0x11FC, 0x1924, 0x55FC, 0x5020, 0x53FE, 0x9000, 0x11FC, 0x1104, 0x1124, 0x1124, 0x1124, 0x1050, 0x1088, 0x1304}},
    {24871, 2, {0x2040, 0x2080, 0x27FC, 0x2444, 0x3444, 0xAFFC, 0xA444, 0xA484, 0xA7FC, 0x2080, 0x2150, 0x2164, 0x227C, 0x2442, 0x2842, 0x303E}},
    {24872, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x7F78, 0x4148, 0x30, 0x3ECE, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24873, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0x1BFE, 0x5400, 0x51FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {24874, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0x19F8, 0x5400, 0x53FC, 0x5204, 0x93FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1090, 0x1108, 0x1204}},
    {24875, 2, {0x1040, 0x1040, 0x17FC, 0x1040, 0x1BF8, 0x5440, 0x57FE, 0x5080, 0x9110, 0x13E0, 0x1048, 0x17FC, 0x1044, 0x1248, 0x1544, 0x1080}},
    {24876, 2, {0x2200, 0x147C, 0xFF44, 0x87C, 0x4944, 0x497C, 0x7F44, 0x844, 0x1094, 0x2108, 0x4100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24877, 2, {0x1040, 0x1044, 0x11F8, 0x1050, 0x1BFE, 0x5440, 0x50F8, 0x5182, 0x927E, 0x1400, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {24878, 2, {0x1008, 0x103C, 0x13C0, 0x1004, 0x1A44, 0x5528, 0x51FC, 0x5220, 0x9020, 0x13FE, 0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {24879, 2, {0x10A0, 0x1090, 0x11FE, 0x1B20, 0x55FC, 0x5120, 0x51FC, 0x9120, 0x11FE, 0x1100, 0x13F8, 0x1108, 0x1090, 0x1060, 0x1198, 0x1606}},
    {24880, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x1904, 0x55FC, 0x5020, 0x5124, 0x90A8, 0x1020, 0x13FE, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {24881, 2, {0x1020, 0x1010, 0x11FE, 0x1900, 0x5520, 0x5120, 0x553C, 0x9350, 0x1190, 0x1310, 0x15FE, 0x1110, 0x1128, 0x1228, 0x1244, 0x1482}},
    {24882, 2, {0x1000, 0x11F8, 0x1108, 0x11E8, 0x1928, 0x57FE, 0x5402, 0x51F8, 0x9108, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x1108, 0x1128, 0x1110}},
    {24883, 2, {0x0, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x0, 0x100, 0x888, 0x4884, 0x4812, 0x4812, 0x87F0, 0x0}},
    {24884, 2, {0x1020, 0x1020, 0x1050, 0x18C8, 0x5524, 0x53FE, 0x5504, 0x91FC, 0x1104, 0x11FC, 0x1100, 0x11FC, 0x1284, 0x1284, 0x14FC, 0x1084}},
    {24885, 2, {0x1000, 0x13DE, 0x1042, 0x1042, 0x1BDE, 0x5610, 0x5210, 0x53DE, 0x9042, 0x1252, 0x114A, 0x114A, 0x1252, 0x1042, 0x1294, 0x1108}},
    {24886, 2, {0x1040, 0x11FC, 0x1044, 0x1094, 0x1908, 0x57DE, 0x514A, 0x514A, 0x9294, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x110C}},
    {24887, 2, {0x1040, 0x1248, 0x1248, 0x1A48, 0x57F8, 0x5000, 0x57FC, 0x9000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1110, 0x10A0, 0x17FE, 0x1000}},
    {24888, 2, {0x820, 0x820, 0x7F7C, 0x2A24, 0x2A24, 0x5D24, 0x824, 0x1C44, 0x2A94, 0x4808, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24889, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0x3494, 0xA908, 0xA264, 0xA090, 0xA108, 0x2204, 0x25FA, 0x2108, 0x2108, 0x2108, 0x21F8, 0x2108}},
    {24890, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x1800, 0x55FC, 0x5104, 0x51FC, 0x9104, 0x11FC, 0x1124, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {24891, 2, {0xC, 0x7DF0, 0x440, 0x888, 0x11F0, 0x1020, 0x1C44, 0xF3FE, 0x1022, 0x1124, 0x5222, 0x2060, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24892, 2, {0x120C, 0x13F2, 0x1202, 0x19FE, 0x5400, 0x51FC, 0x5504, 0x95FC, 0x1504, 0x15FC, 0x1504, 0x15FC, 0x1400, 0x17FE, 0x1104, 0x1202}},
    {24893, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0x19FC, 0x5524, 0x51FC, 0x5124, 0x91FC, 0x1124, 0x1008, 0x13FE, 0x1088, 0x1048, 0x1048, 0x1018}},
    {24894, 2, {0x1100, 0x1100, 0x11FC, 0x1A00, 0x55F8, 0x5000, 0x53F8, 0x9088, 0x12A8, 0x11C8, 0x17F8, 0x108A, 0x11CA, 0x16AA, 0x1086, 0x1082}},
    {24895, 2, {0x0, 0x3FFE, 0x2080, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2250, 0x2448, 0x2944, 0x2080, 0x2044, 0x4A42, 0x4A12, 0x91F0}},
    {24896, 2, {0x103C, 0x17C0, 0x1244, 0x1928, 0x5480, 0x5110, 0x53E0, 0x9040, 0x1188, 0x13FC, 0x1044, 0x1040, 0x17FC, 0x10A0, 0x1110, 0x160E}},
    {24897, 2, {0x3FF8, 0x2008, 0x2FF8, 0x2208, 0x2D28, 0x22C8, 0x2DA8, 0x2298, 0x2D88, 0x2008, 0x3FF8, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24898, 2, {0x2000, 0x17F8, 0x9110, 0x40A0, 0x47FC, 0x1444, 0x27FC, 0xE444, 0x27FC, 0x2444, 0x2404, 0x100, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24899, 2, {0x2120, 0x2110, 0x2208, 0x3486, 0xA910, 0xA3F8, 0xA008, 0xA000, 0x27BC, 0x2084, 0x24A4, 0x2294, 0x24A4, 0x2084, 0x2294, 0x2108}},
    {24900, 2, {0x2000, 0x27FC, 0x20A0, 0x20A0, 0x37FC, 0xACA4, 0xA4A4, 0xA7FC, 0xA040, 0x2040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {24901, 2, {0x1000, 0x13F8, 0x1148, 0x1110, 0x18A0, 0x5440, 0x51B0, 0x564E, 0x9040, 0x13F8, 0x1248, 0x1248, 0x13F8, 0x1044, 0x17FC, 0x1004}},
    {24902, 2, {0x103C, 0x13C0, 0x1044, 0x1224, 0x1928, 0x5500, 0x5040, 0x539C, 0x9204, 0x1204, 0x139C, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {24903, 2, {0x600, 0x3878, 0x2048, 0x3E48, 0x2286, 0x3E00, 0x22FC, 0x3E44, 0x2028, 0x3E10, 0x2228, 0x4A46, 0x8500, 0x4884, 0x4812, 0x87F2}},
    {24904, 2, {0x820, 0x440, 0xFFFE, 0x1020, 0x1020, 0x2448, 0x78F0, 0x1020, 0x2448, 0x7EFC, 0x204, 0x100, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24905, 2, {0x2080, 0x2040, 0x27FC, 0x2080, 0x3108, 0xABF0, 0xA060, 0xA184, 0xA7FE, 0x2002, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x2404}},
    {24906, 2, {0x2110, 0x20A0, 0x2000, 0x27FE, 0x30A0, 0xABF8, 0xA0A8, 0xA7FE, 0xA0A8, 0x23F8, 0x20A0, 0x21B0, 0x22A8, 0x2CA6, 0x20A0, 0x20A0}},
    {24907, 2, {0x1080, 0x2498, 0x7EE0, 0x84, 0x7E7C, 0x4200, 0x7E98, 0x42E0, 0x7E84, 0x4284, 0x467C, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24908, 2, {0x2108, 0x2108, 0x27FE, 0x2148, 0x3020, 0xAFFE, 0xA200, 0xA200, 0xA3FC, 0x2000, 0x2248, 0x2248, 0x2248, 0x244A, 0x244A, 0x2846}},
    {24909, 2, {0x2000, 0x23F8, 0x2248, 0x22A8, 0x3318, 0xAA08, 0xA3F8, 0xA000, 0xA000, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x2FFE, 0x2000}},
    {24910, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x33F8, 0xAA08, 0xA3F8, 0xA208, 0xA3F8, 0x2208, 0x23F8, 0x2208, 0x2FFE, 0x2110, 0x2208, 0x2404}},
    {24911, 2, {0x1000, 0x13FE, 0x1202, 0x1000, 0x19FC, 0x5504, 0x51FC, 0x5104, 0x91FC, 0x1040, 0x1020, 0x13FE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {24912, 2, {0x3FF8, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1110, 0x4884, 0x4812, 0x87F2}},
    {24913, 2, {0x2000, 0x23FC, 0x2108, 0x21F8, 0x3108, 0xA9F8, 0xA10E, 0xA7F8, 0xA008, 0x27BC, 0x20A4, 0x22A4, 0x2128, 0x2290, 0x24A8, 0x2846}},
    {24914, 2, {0x1050, 0x1050, 0x13FE, 0x1050, 0x1BFE, 0x5652, 0x53FE, 0x5252, 0x93FE, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {24915, 2, {0x2000, 0x2FFE, 0x20A0, 0x27FC, 0x34A4, 0xACA4, 0xA7FC, 0xA000, 0xA7FC, 0x2000, 0x2FFE, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {24916, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0x3000, 0xABF8, 0xA208, 0xA3F8, 0xA208, 0x23F8, 0x2040, 0x27FC, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {24917, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1210, 0x2208, 0xCB26, 0x1290, 0x2A90, 0x400}},
    {24918, 2, {0x2000, 0x27FE, 0x242A, 0x2426, 0x37FE, 0xAC22, 0xA7AA, 0xA6AA, 0xA7AA, 0x2432, 0x2596, 0x262E, 0x2446, 0x2482, 0x27FE, 0x2402}},
    {24919, 2, {0x820, 0x820, 0xFFBE, 0x844, 0x7FA4, 0x4924, 0x7F28, 0x1C28, 0x2A10, 0x4928, 0x8844, 0x882, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24920, 2, {0x2080, 0x2110, 0x23F8, 0x2210, 0x34A4, 0xAFBE, 0xA042, 0xA0A0, 0xA318, 0x2C46, 0x2190, 0x2664, 0x2188, 0x2630, 0x21C0, 0x2E00}},
    {24921, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24922, 2, {0x2200, 0x2202, 0x223C, 0x2FA0, 0x3220, 0xAFA0, 0xAABE, 0xAFA4, 0xAAA4, 0x2FA4, 0x2224, 0x2FA4, 0x2224, 0x2244, 0x2244, 0x2284}},
    {24923, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0x1888, 0x55FE, 0x5310, 0x5510, 0x91FE, 0x1110, 0x1110, 0x11FE, 0x1110, 0x1110, 0x11FE, 0x1100}},
    {24924, 2, {0x2020, 0x3FA0, 0x403E, 0xBF44, 0x29A4, 0x2524, 0x7FE8, 0x2928, 0x4510, 0x7FA8, 0x144, 0x282, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24925, 2, {0x3FFC, 0x2420, 0x2FF8, 0x2520, 0x3FFC, 0x2200, 0x27F0, 0x2A10, 0x33F0, 0x2000, 0x3FFC, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24926, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0x1BFE, 0x5400, 0x51FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {24927, 2, {0x21C8, 0x2708, 0x2108, 0x2FE8, 0x311E, 0xAFCA, 0xA54A, 0xA7CA, 0xA54A, 0x27CA, 0x210A, 0x27CA, 0x2112, 0x21D2, 0x2E2A, 0x2444}},
    {24928, 2, {0x2108, 0x2108, 0x2108, 0x27D0, 0x311E, 0xAF94, 0xA124, 0xAFD4, 0xA214, 0x23D4, 0x2254, 0x2248, 0x2248, 0x24D4, 0x2424, 0x2842}},
    {24929, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0x3554, 0xAA48, 0xA554, 0xA040, 0xA554, 0x2248, 0x2554, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {24930, 2, {0x23F8, 0x2208, 0x23F8, 0x3208, 0xABF8, 0xA000, 0xA7FC, 0xA4A4, 0x24A4, 0x27FC, 0x2000, 0x23F8, 0x2110, 0x20E0, 0x2318, 0x2C06}},
    {24931, 2, {0x1000, 0x11FC, 0x1124, 0x13FE, 0x1924, 0x55FC, 0x5000, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {24932, 2, {0x878, 0x7F48, 0x848, 0x3E86, 0x78, 0x7F48, 0x4148, 0x9C28, 0x1510, 0x2628, 0x4046, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24933, 2, {0x1010, 0x1450, 0x127C, 0x1290, 0x1810, 0x54FE, 0x5000, 0x5600, 0x927C, 0x1244, 0x1244, 0x127C, 0x1244, 0x1500, 0x18FE, 0x1000}},
    {24934, 2, {0xA20, 0x920, 0x7FBE, 0x844, 0x6AA4, 0x1C24, 0x2A28, 0x4910, 0x2828, 0x1046, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24935, 2, {0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x3FF8, 0x8, 0x1FF8, 0x8, 0x3FF8, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24936, 2, {0x2000, 0x27BE, 0x2488, 0x2488, 0x37A8, 0xACA8, 0xA4A8, 0xA7BE, 0xA410, 0x2518, 0x2498, 0x25A8, 0x26AA, 0x244A, 0x2046, 0x2080}},
    {24937, 2, {0x2010, 0x2410, 0x22FE, 0x3210, 0xA8FE, 0xA092, 0xA6FE, 0xA292, 0x22FE, 0x2210, 0x22FE, 0x2210, 0x2210, 0x2510, 0x28FE, 0x2000}},
    {24938, 2, {0x1000, 0x13FE, 0x1200, 0x127C, 0x1A44, 0x5644, 0x527C, 0x5200, 0x92EE, 0x12AA, 0x12AA, 0x12AA, 0x12EE, 0x1200, 0x13FE, 0x1000}},
    {24939, 2, {0x1110, 0x2110, 0x4110, 0x8AA8, 0x1444, 0x3240, 0x527C, 0x9240, 0x1540, 0x18FE, 0x1100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24940, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x18F8, 0x5420, 0x51FC, 0x5124, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {24941, 2, {0x814, 0x812, 0x7F10, 0x4AFE, 0x2C10, 0xFF10, 0x1C28, 0x2A28, 0x4944, 0x8882, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24942, 2, {0x100, 0x1FC, 0x100, 0x3FFE, 0x2102, 0x2FF4, 0x2100, 0x27F8, 0x2448, 0x27F8, 0x2448, 0x27F8, 0x2080, 0x4A44, 0x5212, 0x81F0}},
    {24943, 2, {0x1100, 0x11FC, 0x1200, 0x15F8, 0x1908, 0x55F8, 0x5108, 0x51F8, 0x9000, 0x17FE, 0x1100, 0x13FC, 0x14A4, 0x1124, 0x1254, 0x1088}},
    {24944, 2, {0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5FA8, 0x4408, 0x9528, 0x2C90, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24945, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0x1924, 0x55FC, 0x5124, 0x51FC, 0x9022, 0x13FE, 0x1008, 0x13FE, 0x1108, 0x1088, 0x1028, 0x1010}},
    {24946, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x1888, 0x54F8, 0x5020, 0x53FE, 0x9222, 0x1332, 0x12AA, 0x1376, 0x1222, 0x1222, 0x122A, 0x1224}},
    {24947, 2, {0x2000, 0x27DC, 0x2514, 0x27D4, 0x3454, 0xAFC8, 0xA508, 0xA514, 0xA7E2, 0x2000, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x2000}},
    {24948, 2, {0x1000, 0x13DE, 0x1042, 0x1252, 0x194A, 0x5652, 0x5042, 0x5020, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24949, 2, {0x2020, 0x27FE, 0x2420, 0x25FC, 0x3424, 0xAFFE, 0xA424, 0xA5FC, 0xA420, 0x25FC, 0x2524, 0x25FC, 0x2524, 0x29FC, 0x2924, 0x312C}},
    {24950, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x3FFE, 0x2082, 0x2A54, 0x31F2, 0x2200, 0x27F8, 0x4A08, 0x41F0, 0x9E0E}},
    {24951, 2, {0x2040, 0x2020, 0x27FE, 0x2420, 0x35FC, 0xAC24, 0xA7FE, 0xA424, 0xA5FC, 0x2420, 0x2562, 0x24B4, 0x2928, 0x2A26, 0x30A0, 0x2040}},
    {24952, 2, {0x1450, 0x1450, 0x7FFC, 0x2452, 0x47CE, 0x8000, 0x7FFC, 0x4104, 0x1FF0, 0x1110, 0x1150, 0x1120, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24953, 2, {0x1040, 0x7C40, 0x11F8, 0xFE48, 0x4448, 0x28C8, 0x7C68, 0x109A, 0xFE8A, 0x1106, 0x1202, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24954, 2, {0x2040, 0x23F8, 0x2248, 0x2FFE, 0x3248, 0xABF8, 0xA040, 0xA3F8, 0xA248, 0x23F8, 0x2080, 0x2FFE, 0x2110, 0x20E0, 0x2318, 0x2C04}},
    {24955, 2, {0x1040, 0x1248, 0x1150, 0x13F8, 0x1880, 0x57FC, 0x5110, 0x5208, 0x97FC, 0x1A0A, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208}},
    {24956, 2, {0x28, 0x24, 0x3FFE, 0x2220, 0x2324, 0x2224, 0x2FA8, 0x2228, 0x2A92, 0x522A, 0x4646, 0x8102, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {24957, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0x3410, 0xAC90, 0xA4D2, 0xA492, 0xA7F2, 0x2494, 0x2494, 0x25C8, 0x2AAA, 0x289A, 0x31A6, 0x2042}},
    {24958, 2, {0x1420, 0x2220, 0x493E, 0x1442, 0x2294, 0x4110, 0xBE28, 0x2228, 0x2244, 0x3E82, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24959, 2, {0x43FE, 0x2220, 0x23FC, 0xA20, 0xBFC, 0x1220, 0xE3FE, 0x2122, 0x27FA, 0x2122, 0x226, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24960, 2, {0x1000, 0x17BC, 0x1084, 0x14A4, 0x1A94, 0x54A4, 0x5050, 0x5188, 0x9626, 0x10C0, 0x1310, 0x1064, 0x1388, 0x1030, 0x10C0, 0x1700}},
    {24961, 2, {0x1040, 0x1080, 0x13FC, 0x1244, 0x1AF4, 0x5714, 0x52A4, 0x5244, 0x92A4, 0x13FC, 0x1040, 0x1024, 0x1522, 0x150A, 0x1908, 0x10F8}},
    {24962, 2, {0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFE, 0x4102, 0x9494, 0x1428, 0x23E8, 0x800, 0x1FF0, 0x6820, 0x7C0, 0xF83E}},
    {24963, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x3CF8, 0xD016, 0x1CF0, 0x1010, 0x1FF0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24964, 2, {0x1040, 0x27FC, 0x4040, 0x8BF8, 0x1208, 0x33F8, 0x5208, 0x93F8, 0x1208, 0x13F8, 0x1208, 0x17FE, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24965, 2, {0x7F20, 0x820, 0xFF20, 0x8FC, 0x7F24, 0x4924, 0x7F24, 0x4924, 0x7F24, 0x844, 0x7F44, 0x894, 0xFF08, 0x208, 0x5124, 0x9FE4}},
    {24966, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0x1A02, 0x54F8, 0x5088, 0x5088, 0x90F8, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {24967, 2, {0x688, 0x3888, 0x9FE, 0x7E88, 0x888, 0x8F8, 0x3E88, 0x2288, 0x2288, 0x3EF8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24968, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0x1A42, 0x5678, 0x53C4, 0x523C, 0x9200, 0x1250, 0x1254, 0x1354, 0x14D8, 0x1450, 0x19FE, 0x1000}},
    {24969, 2, {0x1082, 0x1082, 0x17F4, 0x1088, 0x1880, 0x57E2, 0x5002, 0x53E4, 0x9228, 0x1220, 0x13E0, 0x1222, 0x1142, 0x1074, 0x1788, 0x1210}},
    {24970, 2, {0x1110, 0x17FC, 0x2110, 0x2FFE, 0x6100, 0xA3F8, 0x2648, 0x2BF8, 0x2248, 0x23F8, 0x2248, 0x2258, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24971, 2, {0x820, 0x4920, 0x2A3E, 0x7F44, 0x49A4, 0x5D28, 0x6B28, 0x4910, 0x4928, 0x4346, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24972, 2, {0x840, 0x1440, 0x227C, 0x5C84, 0x8944, 0x3E2C, 0x814, 0x4A24, 0x28C4, 0xE28, 0x7010, 0x2200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {24973, 2, {0x2038, 0x23C0, 0x2080, 0x27FE, 0x3110, 0xAA08, 0xA5F6, 0xA910, 0xA1F0, 0x2000, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x25F4, 0x240C}},
    {24974, 2, {0x1104, 0x1088, 0x1000, 0x13FE, 0x1A22, 0x56AA, 0x5272, 0x5222, 0x93FE, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {24975, 2, {0x2100, 0x2110, 0x23DC, 0x2254, 0x3554, 0xAA88, 0xA108, 0xA2F4, 0xA402, 0x2000, 0x23FC, 0x2040, 0x2150, 0x2248, 0x2544, 0x2080}},
    {24976, 2, {0x2040, 0x2248, 0x2150, 0x27FC, 0x3150, 0xAA48, 0xA444, 0xA208, 0xA208, 0x23BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x2408, 0x2808}},
    {24977, 2, {0x43FC, 0x2240, 0xBF8, 0xA40, 0x13F8, 0x2240, 0xE3FC, 0x2004, 0x2554, 0x2008, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24978, 2, {0x1040, 0x13F8, 0x1248, 0x13F8, 0x1840, 0x57FE, 0x5000, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1110, 0x1208}},
    {24979, 2, {0x2040, 0x2040, 0x2FFE, 0x2040, 0x37FC, 0xAC44, 0xA7FC, 0xA444, 0xA7FC, 0x2044, 0x2FFE, 0x2042, 0x2024, 0x2522, 0x250A, 0x28F8}},
    {24980, 2, {0x2140, 0x2120, 0x23FE, 0x2220, 0x3620, 0xABFC, 0xA220, 0xA220, 0xA3FC, 0x2220, 0x2220, 0x23FE, 0x2200, 0x2524, 0x2492, 0x2892}},
    {24981, 2, {0x2020, 0x27A4, 0x20A8, 0x2292, 0x3114, 0xAA08, 0xA5F4, 0xA802, 0xA3F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2110, 0x2FFE, 0x2000}},
    {24982, 2, {0x814, 0x812, 0x7F10, 0x2AFE, 0x2A10, 0x5D10, 0x1C28, 0x2A28, 0x4944, 0x8882, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24983, 2, {0x820, 0x820, 0x7F3E, 0x2A44, 0x2AA4, 0x5D24, 0x828, 0x1C10, 0x2A28, 0x4846, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {24984, 2, {0x1020, 0x13FE, 0x1020, 0x19FC, 0x5400, 0x51FC, 0x5104, 0x91FC, 0x1088, 0x17FE, 0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {24985, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0xFFFE, 0x2008, 0x3FF8, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24986, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0x1BDE, 0x5400, 0x51FC, 0x5124, 0x91FC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {24987, 2, {0x1000, 0x13FE, 0x1050, 0x1BFE, 0x5652, 0x53FE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {24988, 2, {0x2020, 0x2720, 0x25FE, 0x2540, 0x357C, 0xAE90, 0xA57E, 0xA500, 0xA57C, 0x2544, 0x257C, 0x2644, 0x247C, 0x2444, 0x2444, 0x244C}},
    {24989, 2, {0x840, 0x7F40, 0x7E, 0x3E88, 0x2348, 0x3E48, 0x50, 0x3E50, 0x420, 0x7F50, 0x888, 0x2904, 0x1100, 0x4884, 0x4812, 0x87F2}},
    {24990, 2, {0x2210, 0x2110, 0x2FD0, 0x2010, 0x37BE, 0xACA4, 0xA7D4, 0xA014, 0xA794, 0x2094, 0x2114, 0x21C8, 0x2F08, 0x2114, 0x2514, 0x2222}},
    {24991, 2, {0x2000, 0x27FC, 0x20A0, 0x20A0, 0x37FC, 0xACA4, 0xA4A4, 0xA7FC, 0xA248, 0x2150, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {24992, 2, {0x3FFE, 0x2890, 0x2510, 0x2FBE, 0x2222, 0x2AD4, 0x2A90, 0x2F90, 0x2228, 0x4428, 0x4444, 0x8882, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24993, 2, {0x1210, 0x1210, 0x13DE, 0x1528, 0x18A4, 0x5420, 0x53FE, 0x5020, 0x93FE, 0x1222, 0x122A, 0x1274, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {24994, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0x1BFE, 0x5508, 0x539C, 0x5108, 0x9188, 0x163E, 0x1000, 0x13FE, 0x1090, 0x1090, 0x1112, 0x120E}},
    {24995, 2, {0x103C, 0x13E0, 0x1124, 0x10A8, 0x1BFE, 0x54A8, 0x5124, 0x5202, 0x91FC, 0x1124, 0x1124, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {24996, 2, {0x1020, 0x11FC, 0x1020, 0x1088, 0x1BFE, 0x5488, 0x51FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {24997, 2, {0x820, 0x2AA8, 0x820, 0x1450, 0x2288, 0x7FFC, 0x4004, 0x200, 0x3FF8, 0x408, 0x850, 0x3020, 0x100, 0x4884, 0x4812, 0x87F2}},
    {24998, 2, {0x1088, 0x12AA, 0x12DC, 0x1488, 0x1954, 0x5622, 0x5000, 0x53FE, 0x9242, 0x1444, 0x13FC, 0x1044, 0x1044, 0x1084, 0x1114, 0x1208}},
    {24999, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0x1BFE, 0x5400, 0x51FC, 0x5124, 0x91FC, 0x1124, 0x11FC, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {25000, 2, {0x7C20, 0x420, 0xFF3E, 0x2244, 0x3EA4, 0x2228, 0x3E28, 0x2390, 0xFE28, 0x4244, 0x282, 0x100, 0x2888, 0x28A4, 0x4824, 0x7E0}},
    {25001, 2, {0x1C10, 0x7020, 0x10FC, 0xFE84, 0x10FC, 0x7C84, 0x44FC, 0x4484, 0x7CFC, 0x4400, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {25002, 2, {0x279E, 0x2492, 0x279E, 0x2492, 0x379E, 0xAC02, 0xA4F2, 0xA492, 0xA4F2, 0x2492, 0x24F2, 0x2492, 0x2492, 0x2532, 0x2402, 0x2406}},
    {25003, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0x34A4, 0xAFBC, 0xA484, 0xA444, 0xA5F4, 0x2424, 0x24A4, 0x2444, 0x24A4, 0x2504, 0x2414, 0x2408}},
    {25004, 2, {0x1000, 0x13F8, 0x1208, 0x1BF8, 0x5608, 0x53F8, 0x5040, 0x97FE, 0x1000, 0x13F8, 0x1208, 0x13F8, 0x1040, 0x1248, 0x1444, 0x10C0}},
    {25005, 2, {0x2040, 0x2040, 0x27FC, 0x20A0, 0x3514, 0xAA08, 0xA7FC, 0xAA0A, 0xA3F8, 0x2208, 0x23F8, 0x2040, 0x2248, 0x2444, 0x2944, 0x2080}},
    {25006, 2, {0x2100, 0x2100, 0x23FC, 0x36A8, 0xAAA8, 0xA2A8, 0xA7FC, 0xA2A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000, 0x22A8, 0x2254, 0x2454, 0x2000}},
    {25007, 2, {0x2000, 0x27BC, 0x2108, 0x2528, 0x37BC, 0xAB18, 0xA5AA, 0xA946, 0xA000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8}},
    {25008, 2, {0x1000, 0x13FC, 0x1048, 0x1030, 0x1BFE, 0x5452, 0x5094, 0x5350, 0x9020, 0x13FE, 0x1252, 0x128A, 0x1376, 0x1252, 0x1272, 0x1206}},
    {25009, 2, {0x2214, 0x2112, 0x2FD2, 0x2010, 0x307E, 0xAF90, 0xA490, 0xA490, 0xA790, 0x2128, 0x25A8, 0x2568, 0x2928, 0x214A, 0x254A, 0x2286}},
    {25010, 2, {0x100, 0x7FFC, 0x4104, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2448, 0x3FF8, 0x100, 0x4884, 0x4812, 0x87F2}},
    {25011, 2, {0x1000, 0x13FC, 0x1004, 0x13FC, 0x1804, 0x57FC, 0x5000, 0x53BC, 0x9124, 0x17BC, 0x1008, 0x13FE, 0x1108, 0x1088, 0x10A8, 0x1010}},
    {25012, 2, {0x2000, 0x27FC, 0x24A4, 0x24A4, 0x37BC, 0xA8A0, 0xA7BC, 0xA4A4, 0xA4A4, 0x27BC, 0x24A4, 0x24A4, 0x27BC, 0x24A2, 0x20A2, 0x207E}},
    {25013, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {25014, 2, {0x1080, 0x1040, 0x13FC, 0x1908, 0x5490, 0x57FE, 0x5000, 0x93FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1040, 0x1524, 0x150A, 0x18FA}},
    {25015, 2, {0x2110, 0x2110, 0x27BC, 0x3110, 0xABB8, 0xA554, 0xA912, 0xA000, 0x27FC, 0x2044, 0x2240, 0x2278, 0x2240, 0x2540, 0x28FE, 0x3000}},
    {25016, 2, {0x2040, 0x2040, 0x20A0, 0x3110, 0xAA08, 0xA5F6, 0xA800, 0xA7BC, 0x24A4, 0x24A4, 0x27BC, 0x2210, 0x2210, 0x2528, 0x2944, 0x3082}},
    {25017, 2, {0x1090, 0x13FC, 0x1294, 0x1BFC, 0x5694, 0x53FC, 0x5000, 0x93FC, 0x1200, 0x12F8, 0x1200, 0x13FE, 0x1520, 0x1514, 0x1548, 0x1986}},
    {25018, 2, {0x1100, 0x11F8, 0x1208, 0x17FE, 0x1A88, 0x5724, 0x53FE, 0x5200, 0x92FC, 0x1200, 0x12FC, 0x1200, 0x12FC, 0x1484, 0x14FC, 0x1884}},
    {25019, 2, {0x1020, 0x17FE, 0x1000, 0x13FE, 0x1A02, 0x56FA, 0x528A, 0x53FE, 0x9000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1000, 0x17FE}},
    {25020, 2, {0x2220, 0xFFA0, 0x223E, 0x2044, 0x7FA4, 0x8128, 0x7D28, 0x4510, 0x7D28, 0xA44, 0x482, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {25021, 2, {0x2110, 0x2FFE, 0x2110, 0x3200, 0xABFC, 0xA4A4, 0xABA4, 0xA164, 0x2254, 0x2494, 0x2108, 0x2040, 0x2024, 0x2522, 0x250A, 0x28F8}},
    {25022, 2, {0x2014, 0x2012, 0x27FE, 0x2410, 0x35D0, 0xAC14, 0xA5D4, 0xA558, 0xA5CA, 0x2416, 0x2442, 0x2824, 0x22A2, 0x228A, 0x2478, 0x2000}},
    {25023, 2, {0x2108, 0x2208, 0x27C8, 0x2450, 0x37DE, 0xAC64, 0xA7D4, 0xA214, 0xA114, 0x27D4, 0x2214, 0x23C8, 0x2248, 0x2454, 0x2554, 0x28A2}},
    {25024, 2, {0x2040, 0x20A0, 0x2318, 0x2DF6, 0x3000, 0xAFFC, 0xA554, 0xA4E4, 0xA7FC, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208}},
    {25025, 2, {0x2000, 0x23FC, 0x2294, 0x3294, 0xABFC, 0xA000, 0xA7FE, 0xA000, 0x23FC, 0x2204, 0x23FC, 0x20A2, 0x2114, 0x2308, 0x2D44, 0x2182}},
    {25026, 2, {0x2110, 0x27FE, 0x2110, 0x23F8, 0x3040, 0xAFFC, 0xA040, 0xA3F8, 0xA248, 0x23F8, 0x2248, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x2FFE}},
    {25027, 2, {0x2220, 0xFFA0, 0x2220, 0x3EFC, 0x824, 0x7F24, 0x4924, 0x7F24, 0x824, 0x7F44, 0x854, 0xFF88, 0x100, 0x4884, 0x4812, 0x87F2}},
    {25028, 2, {0x2288, 0x2288, 0x2FE8, 0x2288, 0x33BE, 0xA90A, 0xA7CA, 0xA54A, 0xA7CA, 0x210A, 0x27CA, 0x210A, 0x27D2, 0x2112, 0x21EA, 0x2E44}},
    {25029, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0x1A42, 0x5678, 0x53C4, 0x523C, 0x9200, 0x13FE, 0x1240, 0x13A4, 0x1258, 0x15B4, 0x1452, 0x19B0}},
    {25030, 2, {0x23F8, 0x2208, 0x2208, 0x33F8, 0xA800, 0xA7BC, 0xA4A4, 0xA4A4, 0x27BC, 0x2040, 0x2FFE, 0x2160, 0x2250, 0x2448, 0x3846, 0x2040}},
    {25031, 2, {0x600, 0xF8FC, 0x4A84, 0x24FC, 0x3884, 0xCCFC, 0x36A2, 0xCA94, 0x12A8, 0x6AC4, 0x482, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {25032, 2, {0x2200, 0x239E, 0x248A, 0x290A, 0x37D2, 0xAD66, 0xA540, 0xA7D4, 0xA55E, 0x2564, 0x27C4, 0x2544, 0x255E, 0x2544, 0x2444, 0x28C4}},
    {25033, 2, {0x100, 0x80, 0x3FFE, 0x2248, 0x24FE, 0x2D90, 0x34FC, 0x2490, 0x24FC, 0x2490, 0x24FE, 0x2480, 0x4244, 0x524A, 0x920A, 0x21F8}},
    {25034, 2, {0x2080, 0x2100, 0x27FC, 0x2444, 0x3554, 0xAC44, 0xA7FC, 0xA4E4, 0xA554, 0x2404, 0x2040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {25035, 2, {0x2788, 0x2088, 0xFABE, 0x2108, 0x77D8, 0x6B5C, 0xA52A, 0x2948, 0x2508, 0x2208, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {25036, 2, {0x2000, 0x27FC, 0x24A4, 0x27FC, 0x3040, 0xABF8, 0xA040, 0xA7FC, 0xA110, 0x20A0, 0x23F8, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x2040}},
    {25037, 2, {0x2040, 0x2FFE, 0x2000, 0x27FC, 0x3404, 0xADF4, 0xA514, 0xA5F4, 0xA404, 0x27FC, 0x2008, 0x23F0, 0x2040, 0x2FFE, 0x2150, 0x2E4E}},
    {25038, 2, {0x2040, 0x2040, 0x27FC, 0x3248, 0xAA48, 0xA554, 0xAFFE, 0xA000, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x25F4, 0x2404, 0x27FC, 0x2404}},
    {25039, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0x19FC, 0x5520, 0x51FC, 0x5120, 0x91FE, 0x1100, 0x13DE, 0x1252, 0x1252, 0x1272, 0x1202, 0x1206}},
    {25040, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x19FC, 0x5554, 0x5154, 0x51FC, 0x9020, 0x13FE, 0x1050, 0x1094, 0x1188, 0x12A8, 0x14C4, 0x1082}},
    {25041, 2, {0x110, 0x4FFE, 0x2110, 0x27FC, 0x80A0, 0x47FC, 0x54A4, 0x14A4, 0xE564, 0x2654, 0x2484, 0x240C, 0x100, 0x4884, 0x4812, 0x87F2}},
    {25042, 2, {0x2210, 0x2210, 0x2F9C, 0x2224, 0x3248, 0xAFBE, 0xAAA2, 0xAAAA, 0xAFAA, 0x222A, 0x272A, 0x2AAA, 0x3288, 0x2214, 0x2222, 0x2242}},
    {25043, 2, {0x103C, 0x13C0, 0x1044, 0x1224, 0x1908, 0x57FE, 0x5442, 0x52A4, 0x928A, 0x147A, 0x1080, 0x10F8, 0x1108, 0x1290, 0x1060, 0x179E}},
    {25044, 2, {0x2040, 0x2FFE, 0x2000, 0x37FC, 0xAC04, 0xA5F4, 0xA514, 0xA5F4, 0x2404, 0x27FC, 0x2000, 0x23F8, 0x2000, 0x2FFE, 0x2248, 0x24C4}},
    {25045, 2, {0x0, 0x3FFE, 0x288C, 0x2F8A, 0x2888, 0x2FBE, 0x2008, 0x2F88, 0x2888, 0x2F94, 0x2894, 0x29A2, 0x4080, 0x4240, 0x8A14, 0x11F2}},
    {25046, 2, {0x80C, 0x1470, 0x2210, 0x5C10, 0x88FE, 0x3E10, 0x810, 0x4A7C, 0x2844, 0xE44, 0x707C, 0x2200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {25047, 2, {0x2110, 0x2110, 0x2110, 0x37BC, 0xA910, 0xA110, 0xAFFE, 0xA2A8, 0x22A8, 0x26EC, 0x2AAA, 0x22AA, 0x22A8, 0x25A8, 0x2448, 0x2898}},
    {25048, 2, {0x42A8, 0x27FC, 0x22AA, 0x84E6, 0x4000, 0x57FC, 0x1444, 0xE3F8, 0x2248, 0x2258, 0x2140, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {25049, 2, {0x1140, 0x124C, 0x1264, 0x1244, 0x1B4C, 0x5664, 0x52A4, 0x53AC, 0x92A4, 0x12A4, 0x17FE, 0x1000, 0x1110, 0x1108, 0x1204, 0x1404}},
    {25050, 2, {0x203E, 0x27C0, 0x2244, 0x2128, 0x33F8, 0xA840, 0xA7FC, 0xA000, 0xA3F8, 0x2008, 0x23F8, 0x2008, 0x23F8, 0x2544, 0x252A, 0x28FA}},
    {25051, 2, {0x2040, 0x2FFE, 0x2040, 0x27FC, 0x3000, 0xABF8, 0xA208, 0xAFFE, 0xA802, 0x27FC, 0x2110, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x2FFE}},
    {25052, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x1BFE, 0x5652, 0x53FE, 0x5000, 0x93FE, 0x1242, 0x10F8, 0x1108, 0x1290, 0x1060, 0x10C0, 0x1700}},
    {25053, 2, {0x2400, 0x257C, 0x2604, 0x34A8, 0xAB90, 0xA07C, 0xA414, 0xA794, 0x2A50, 0x2250, 0x2FDC, 0x2250, 0x2550, 0x24B0, 0x289E, 0x2100}},
    {25054, 2, {0x1088, 0x13FE, 0x1088, 0x13FE, 0x1A02, 0x55FC, 0x5000, 0x53FE, 0x9040, 0x10A2, 0x1354, 0x10B8, 0x1354, 0x1092, 0x1350, 0x1020}},
    {25055, 2, {0x1408, 0x5508, 0x3608, 0x147E, 0x7F08, 0x1448, 0x7F28, 0x828, 0x3E08, 0x808, 0xF28, 0x7810, 0x2100, 0x4884, 0x4812, 0x87F2}},
    {25056, 2, {0x2080, 0x2040, 0x27FC, 0x2110, 0x30A4, 0xAF58, 0xA554, 0xA552, 0xAB58, 0x2000, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2408}},
    {25057, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0x37DE, 0xAC88, 0xA5DC, 0xA6AA, 0xA488, 0x2440, 0x2490, 0x2520, 0x2448, 0x2484, 0x25FC, 0x2884}},
    {25058, 2, {0x2010, 0x27D0, 0x2510, 0x27DE, 0x3450, 0xAFE8, 0xA504, 0xA7C4, 0xA000, 0x23F8, 0x22A8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000}},
    {25059, 2, {0x110, 0x47FC, 0x2110, 0x81F0, 0x4040, 0x17FC, 0x1664, 0xE554, 0x26EC, 0x2444, 0x244C, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {25060, 2, {0x2040, 0x27FC, 0x2040, 0x33FC, 0xA800, 0xA7FE, 0xA002, 0xA3F8, 0x2040, 0x27FE, 0x2000, 0x27FE, 0x2008, 0x27FE, 0x24A8, 0x2798}},
    {25061, 2, {0x2040, 0x27FC, 0x2040, 0x27FE, 0x3402, 0xABFC, 0xA248, 0xA3F8, 0xA248, 0x23F8, 0x2000, 0x27FC, 0x2244, 0x2278, 0x2540, 0x28FE}},
    {25062, 2, {0x2000, 0x27FC, 0x2040, 0x2FFE, 0x3842, 0xAB58, 0xA040, 0xA358, 0xA000, 0x2FFE, 0x2040, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x240C}},
    {25063, 2, {0x2080, 0x2040, 0x27FC, 0x2444, 0x32A8, 0xAA94, 0xA474, 0xA000, 0xA3F8, 0x22A8, 0x2FFE, 0x2000, 0x27FC, 0x2040, 0x2140, 0x2080}},
    {25064, 2, {0x2000, 0x2FFE, 0x2808, 0x2BC8, 0x3A4C, 0xABCA, 0xAA48, 0xABDE, 0xA808, 0x2BC8, 0x2A48, 0x2BD4, 0x2A54, 0x2BD4, 0x2A54, 0x32E2}},
    {25065, 2, {0x2208, 0x2110, 0x27FC, 0x3040, 0xABF8, 0xA080, 0xA7FC, 0xA148, 0x23FC, 0x260A, 0x2BF8, 0x2208, 0x23FA, 0x2234, 0x228C, 0x2302}},
    {25066, 2, {0x23F8, 0x2208, 0x23F8, 0x2208, 0x33F8, 0xA910, 0xA7FC, 0xA110, 0xAFFE, 0x2110, 0x2248, 0x2554, 0x20E0, 0x2150, 0x2248, 0x20C0}},
    {25067, 2, {0x1042, 0x139C, 0x1210, 0x1BDE, 0x5694, 0x5294, 0x5420, 0x93FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1108, 0x1204}},
    {25068, 2, {0x80, 0x3FFE, 0x2220, 0x2FF8, 0x2220, 0x3FFC, 0x2080, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x4220, 0x5494, 0x924A, 0x23FA}},
    {25069, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0x35FC, 0xAC88, 0xA7FE, 0xA420, 0xA5FC, 0x2524, 0x25FC, 0x2524, 0x25FC, 0x2400, 0x2488, 0x2904}},
    {25070, 2, {0x27FC, 0x2080, 0x23F8, 0x3208, 0xABF8, 0xA208, 0xAFFE, 0xA882, 0x2548, 0x2514, 0x28F4, 0x2100, 0x23F8, 0x2508, 0x20F0, 0x2F0E}},
    {25071, 2, {0xA20, 0xFFA0, 0x493E, 0x7F44, 0x49A4, 0x5124, 0x828, 0xFF28, 0x2010, 0x3E28, 0x4244, 0x8682, 0x100, 0x4884, 0x4812, 0x87F2}},
    {25072, 2, {0x2202, 0x25E2, 0x24A2, 0x26AA, 0x352A, 0xAE6A, 0xA10A, 0xA28A, 0xA44A, 0x2BAA, 0x210A, 0x27CA, 0x2122, 0x2542, 0x21EA, 0x2E04}},
    {25073, 2, {0x1108, 0x17FE, 0x1108, 0x1000, 0x1BFC, 0x5694, 0x5294, 0x53FC, 0x9028, 0x13FE, 0x1220, 0x1324, 0x12A8, 0x1292, 0x142A, 0x18C6}},
    {25074, 2, {0x2220, 0x4AA0, 0x8ABE, 0x1FC4, 0x20A4, 0x6F24, 0xA028, 0x2FA8, 0x2210, 0x2FA8, 0x2244, 0x2F82, 0x100, 0x4884, 0x4812, 0x87F2}},
    {25075, 2, {0x1088, 0x13DE, 0x1088, 0x19DC, 0x5488, 0x53DE, 0x5088, 0x93FC, 0x1004, 0x11FC, 0x1004, 0x13FC, 0x1040, 0x12A4, 0x128A, 0x147A}},
    {25076, 2, {0x2114, 0x2112, 0x27D2, 0x3110, 0xA910, 0xA7FE, 0xA290, 0xA6D2, 0x2292, 0x26D2, 0x2294, 0x26D4, 0x228A, 0x22CA, 0x2F16, 0x2022}},
    {25077, 2, {0x1108, 0x17FE, 0x1108, 0x13FC, 0x1A94, 0x57FC, 0x5000, 0x57FE, 0x9402, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x1108, 0x11F8}},
    {25078, 2, {0x2108, 0x2108, 0x27CE, 0x2112, 0x3124, 0xAFDE, 0xA552, 0xA55E, 0xA7D2, 0x211E, 0x2392, 0x255E, 0x2940, 0x2114, 0x2112, 0x2122}},
    {25079, 2, {0x2080, 0x2040, 0x2FFE, 0x2000, 0x37FC, 0xACA4, 0xA7FC, 0xA248, 0xA150, 0x2208, 0x2444, 0x20A8, 0x2110, 0x2308, 0x2D46, 0x2180}},
    {25080, 2, {0x3E0E, 0x22F0, 0x3E20, 0x2244, 0x3EF8, 0x2210, 0x3E24, 0x22FE, 0x7F10, 0x854, 0x2A92, 0x5950, 0x120, 0x4884, 0x4812, 0x87F2}},
    {25081, 2, {0x2040, 0x27FC, 0x2000, 0x23B8, 0x32A8, 0xABB8, 0xA110, 0xA7FC, 0xA110, 0x27FC, 0x2110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {25082, 2, {0x2448, 0x244C, 0x2AAA, 0x3008, 0xAFFE, 0xA288, 0xAEE8, 0xA28A, 0x2EEA, 0x228C, 0x2EEC, 0x2288, 0x22EA, 0x2F1A, 0x2426, 0x2042}},
    {25083, 2, {0x1094, 0x1398, 0x1092, 0x118E, 0x1A80, 0x55FC, 0x5124, 0x51FC, 0x9124, 0x11FC, 0x1088, 0x11FC, 0x1088, 0x13FE, 0x1088, 0x1104}},
    {25084, 2, {0x27BC, 0x24A4, 0x27BC, 0x24A4, 0x37BC, 0xACA4, 0xA7BC, 0xA120, 0xA3FE, 0x2220, 0x27FC, 0x2A20, 0x23FC, 0x2220, 0x23FE, 0x2200}},
    {25085, 2, {0x2110, 0x2FFE, 0x2110, 0x27BC, 0x34A4, 0xAFBC, 0xA140, 0xA120, 0xA3FE, 0x2220, 0x27FC, 0x2A20, 0x23FC, 0x2220, 0x23FE, 0x2200}},
    {25086, 2, {0x13FC, 0x1108, 0x11F8, 0x1908, 0x55F8, 0x510E, 0x57F8, 0x9008, 0x17FE, 0x1294, 0x139C, 0x1294, 0x139C, 0x12D6, 0x17BC, 0x1084}},
    {25087, 2, {0x1020, 0xFF20, 0x10BE, 0x7C44, 0x50, 0xFE90, 0x8390, 0x7CA8, 0xA4, 0x7CC2, 0x4420, 0x7C10, 0x4554, 0x2F42, 0xF14A, 0x4238}},
    {25088, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {25089, 2, {0x2850, 0xFE48, 0x28FE, 0x3990, 0x1290, 0x7CFE, 0x5490, 0x7C90, 0x10FE, 0x7C90, 0x1090, 0xFEFE, 0x2880, 0x4504, 0x2492, 0x43F0}},
    {25090, 2, {0x2020, 0x27FE, 0x2488, 0x27DE, 0x3488, 0xADDC, 0xA6AA, 0xA488, 0xA450, 0x27DE, 0x2450, 0x25DC, 0x2450, 0x2BDE, 0x2850, 0x3050}},
    {25091, 2, {0x2248, 0x2150, 0x27FC, 0x2404, 0x31F0, 0xA910, 0xA7FC, 0xA554, 0xA4E4, 0x27FC, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x22A4, 0x2452}},
    {25092, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0x1BDE, 0x54A0, 0x51FE, 0x5320, 0x95FC, 0x1120, 0x11FE, 0x1000, 0x11FC, 0x1088, 0x1070, 0x178E}},
    {25093, 2, {0x1000, 0xFEF8, 0x2820, 0xFEFC, 0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x1088, 0xFEF8, 0x1050, 0x208, 0x5124, 0x9FE4}},
    {25094, 2, {0x1040, 0xFEF8, 0x2948, 0xFE30, 0x1CE, 0x7CF8, 0x4420, 0x7DFC, 0x4488, 0x7CA8, 0x10A8, 0xFE50, 0x1188, 0x4884, 0x4812, 0x87F2}},
    {25095, 2, {0x1040, 0xFEF8, 0x2948, 0xFE30, 0x1CE, 0x7CF8, 0x4420, 0x7DFC, 0x4488, 0x7CF8, 0x1088, 0xFEF8, 0x1050, 0x288, 0x5124, 0x9FE4}},
    {25096, 2, {0x220, 0x210, 0x210, 0x200, 0x2FE, 0xFF00, 0x208, 0x208, 0x210, 0x120, 0x140, 0x80, 0x342, 0xC22, 0x701A, 0x6}},
    {25097, 2, {0x90, 0x88, 0x88, 0x80, 0x7FFE, 0x4080, 0x4080, 0x4088, 0x4088, 0x4090, 0x4060, 0x4844, 0x50A4, 0x6114, 0x420C, 0x404}},
    {25098, 2, {0x90, 0x88, 0x80, 0x3FFE, 0x2080, 0x2080, 0x2084, 0x2044, 0x2048, 0x2048, 0x2030, 0x2022, 0x2052, 0x408A, 0x4306, 0x8C02}},
    {25099, 2, {0x240, 0x220, 0x220, 0x2FC, 0x7F00, 0x200, 0x200, 0x23E, 0xFFC0, 0x210, 0x120, 0x140, 0x184, 0xE44, 0x7034, 0xC}},
    {25100, 2, {0x50, 0x48, 0x40, 0x3FFE, 0x2040, 0x2040, 0x2044, 0x2044, 0x3E44, 0x2028, 0x2028, 0x2012, 0x2032, 0x404A, 0x4086, 0x8102}},
    {25101, 2, {0x50, 0x48, 0x40, 0x3FFE, 0x2040, 0x2040, 0x2044, 0x3044, 0x2844, 0x2428, 0x2428, 0x2012, 0x2032, 0x404A, 0x4086, 0x8102}},
    {25102, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x40, 0x840, 0x844, 0x844, 0x7F44, 0x828, 0x828, 0x1012, 0x1032, 0x204A, 0x4086, 0x302}},
    {25103, 2, {0x20, 0x28, 0x24, 0x7E24, 0x220, 0x23E, 0x25E0, 0x1424, 0x824, 0x828, 0x1428, 0x1210, 0x2232, 0x404A, 0x86, 0x102}},
    {25104, 2, {0x50, 0x48, 0x40, 0x3FFE, 0x2040, 0x2040, 0x2044, 0x3E44, 0x2244, 0x2228, 0x2228, 0x2212, 0x2A32, 0x444A, 0x4086, 0x8102}},
    {25105, 2, {0x440, 0xE50, 0x7848, 0x848, 0x840, 0xFFFE, 0x840, 0x844, 0xA44, 0xC48, 0x1830, 0x6822, 0x852, 0x88A, 0x2B06, 0x1002}},
    {25106, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x40, 0x1240, 0x1244, 0x1244, 0x7F44, 0x1228, 0x1228, 0x1212, 0x2232, 0x224A, 0x4086, 0x302}},
    {25107, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x44, 0x3E44, 0x2244, 0x2228, 0x2228, 0x2212, 0x3E32, 0x224A, 0x86, 0x302}},
    {25108, 2, {0x240, 0x220, 0x2FC, 0x7F10, 0x160, 0x384, 0x1C64, 0xE01C, 0x240, 0x220, 0x2FC, 0x7F10, 0x160, 0x384, 0x1C64, 0xE01C}},
    {25109, 2, {0x420, 0x2428, 0x2424, 0x2424, 0x2420, 0x3C3E, 0x5E0, 0x424, 0xFC24, 0x2428, 0x2428, 0x2410, 0x2432, 0x444A, 0x8486, 0x502}},
    {25110, 2, {0x48, 0x44, 0x40, 0xFFFE, 0x40, 0x40, 0x3E44, 0x2244, 0x2244, 0x2228, 0x3E28, 0x12, 0x732, 0x784A, 0x2086, 0x302}},
    {25111, 2, {0x828, 0x824, 0x1424, 0x1220, 0x212E, 0x40F0, 0xBE20, 0x2224, 0x2224, 0x2228, 0x2A28, 0x2410, 0x2112, 0x212A, 0x1F46, 0x82}},
    {25112, 2, {0x1020, 0x1028, 0x1024, 0x1E24, 0x1020, 0x103E, 0x11E0, 0x1024, 0x7E24, 0x4228, 0x4228, 0x4210, 0x4232, 0x7E4A, 0x4286, 0x102}},
    {25113, 2, {0x28, 0xFF24, 0x8124, 0x8120, 0xBD2E, 0x81F0, 0x8120, 0xBD24, 0xA524, 0xA528, 0xA528, 0xBD10, 0x8112, 0x812A, 0x8546, 0x8282}},
    {25114, 2, {0x28, 0x24, 0x20, 0x3FFE, 0x2220, 0x23A0, 0x2224, 0x2224, 0x3FE4, 0x2028, 0x2228, 0x2A90, 0x3252, 0x422A, 0x4A46, 0x8482}},
    {25115, 2, {0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x240, 0x3F8, 0x7E20, 0x140, 0x84, 0x764, 0x381C}},
    {25116, 2, {0x50, 0x48, 0x40, 0xFFFE, 0x40, 0x3E40, 0x2244, 0x3E44, 0x44, 0x7F28, 0x828, 0x3E12, 0x832, 0xF4A, 0x7886, 0x2302}},
    {25117, 2, {0x28, 0x7E24, 0x4224, 0x4220, 0x7E2E, 0x42F0, 0x4220, 0x7E24, 0x4224, 0x4228, 0x7E28, 0x10, 0x2412, 0x222A, 0x4146, 0x8182}},
    {25118, 2, {0x7FFC, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFE, 0x4242, 0x83FC, 0x7E20, 0x140, 0x84, 0x764, 0x381C}},
    {25119, 2, {0x820, 0x828, 0xFFA4, 0x824, 0x7F20, 0x413E, 0x7FE0, 0x4124, 0x7F24, 0x4928, 0x828, 0xFF90, 0x832, 0x84A, 0x886, 0x902}},
    {25120, 2, {0x1040, 0x848, 0xFF44, 0x44, 0x4240, 0x2440, 0xFFFE, 0x40, 0x7E44, 0x4244, 0x4228, 0x7E2A, 0x4212, 0x422A, 0x7E46, 0x4282}},
    {25121, 2, {0x2220, 0x2228, 0xFF24, 0x2224, 0x3E20, 0x223E, 0x3FE0, 0x2224, 0x2224, 0xFF28, 0x4028, 0x5410, 0x6232, 0x404A, 0x7E86, 0x102}},
    {25122, 2, {0x3F28, 0x2124, 0x2124, 0x3F20, 0x20, 0xFFFE, 0x2120, 0x3F24, 0x2124, 0x3F28, 0x2128, 0x27D0, 0xF912, 0x412A, 0x146, 0x182}},
    {25123, 2, {0xA28, 0x7424, 0x1524, 0x5220, 0x222E, 0x21F0, 0x4120, 0xBE24, 0x824, 0x828, 0x7F28, 0x810, 0x1412, 0x122A, 0x2246, 0x4082}},
    {25124, 2, {0x10, 0x7E14, 0x2212, 0x2410, 0x3F16, 0x5578, 0x4912, 0x5512, 0xA212, 0x14, 0x7F14, 0x5508, 0x551A, 0x57AA, 0xFC46, 0x82}},
    {25125, 2, {0x20, 0x7E28, 0x4224, 0x7E24, 0x4220, 0x7E3E, 0x9E0, 0x4824, 0x7F24, 0x8828, 0x7E28, 0x810, 0xF32, 0xF04A, 0x4086, 0x102}},
    {25126, 2, {0x90A8, 0x4924, 0x24, 0x7F20, 0x492E, 0x49F0, 0x7F20, 0x4924, 0x4924, 0x7F28, 0x828, 0x810, 0xFF92, 0x82A, 0x846, 0x882}},
    {25127, 2, {0x828, 0xC24, 0x1224, 0x2920, 0x44AE, 0xBFF0, 0x2120, 0x3F24, 0x2124, 0x3F28, 0x2028, 0x3F10, 0x5112, 0x512A, 0x9F46, 0x1182}},
    {25128, 2, {0x28, 0xFF24, 0x224, 0x7A20, 0x4A2E, 0x7AF0, 0x20, 0xFF24, 0x224, 0x7A28, 0x4A28, 0x4A10, 0x7A12, 0x22A, 0xA46, 0x482}},
    {25129, 2, {0x28, 0xFF24, 0x2224, 0x5520, 0xF72E, 0xF0, 0xFF20, 0x24, 0x7E24, 0x4228, 0x4228, 0x7E10, 0x4212, 0x422A, 0x7E46, 0x4282}},
    {25130, 2, {0x820, 0x828, 0x7F24, 0x820, 0xFFFE, 0x1420, 0x2224, 0x7FA4, 0xA424, 0x3F28, 0x2428, 0x3F10, 0x2412, 0x3FAA, 0x2046, 0x2082}},
    {25131, 2, {0x1014, 0x1012, 0xFE10, 0x11FE, 0x2010, 0x3C10, 0x25D2, 0x6552, 0xBD52, 0x2554, 0x25D4, 0x3C08, 0x24CA, 0x271A, 0x2426, 0x2C42}},
    {25132, 2, {0x10, 0x7F94, 0x1212, 0x5290, 0x3316, 0x1278, 0xFFD2, 0x12, 0x3F12, 0x2114, 0x2114, 0x3F08, 0x211A, 0x212A, 0x3F46, 0x2182}},
    {25133, 2, {0x1010, 0x814, 0xFF92, 0x8090, 0x16, 0x7F78, 0x812, 0x7F12, 0x4912, 0x7F14, 0x4914, 0x7F08, 0x1A, 0x222A, 0x4146, 0x8082}},
    {25134, 2, {0x10, 0x7714, 0x1112, 0x5510, 0x3316, 0x5578, 0xC12, 0x3312, 0xC492, 0x1814, 0x6214, 0xC88, 0x711A, 0x62A, 0x1846, 0xE082}},
    {25135, 2, {0x1028, 0x1E24, 0x1024, 0x7F20, 0x512E, 0x5CF0, 0x7220, 0x4E24, 0x4024, 0x4A28, 0x6AA8, 0x5B10, 0x4A12, 0x8F2A, 0x7846, 0x82}},
    {25136, 2, {0x10, 0xF794, 0x9492, 0x9490, 0xF796, 0x78, 0x7F12, 0x4912, 0x7F12, 0x4914, 0x7F14, 0x808, 0xFF9A, 0x82A, 0x846, 0x882}},
    {25137, 2, {0x810, 0xF14, 0x812, 0x7F90, 0x4896, 0x4E78, 0x7912, 0x4712, 0x4012, 0x4A14, 0x6A94, 0xAA88, 0xBB9A, 0xA2A, 0x7F46, 0x2082}},
    {25138, 2, {0x828, 0xF24, 0x824, 0x7F20, 0x492E, 0x4CF0, 0x7920, 0x4724, 0x4024, 0x5F28, 0x4028, 0x5F10, 0x5112, 0x5F2A, 0x8A46, 0x3F82}},
    {25139, 2, {0x20, 0xEE28, 0x2224, 0xAA24, 0x6620, 0xAA3E, 0x1E0, 0x2424, 0x7F24, 0xC828, 0x7E28, 0x4810, 0x7E32, 0x484A, 0x7F86, 0x4102}},
    {25140, 2, {0x820, 0x7F28, 0x824, 0xFFFE, 0x20, 0x7F20, 0x4924, 0x7F24, 0x4924, 0x7F28, 0x2228, 0x7F10, 0x2212, 0xFFAA, 0x2246, 0x4182}},
    {25141, 2, {0x7728, 0x5524, 0x7724, 0x5520, 0x772E, 0x55F0, 0x7720, 0x2224, 0x7FA4, 0xA428, 0x3F28, 0x2410, 0x3F12, 0x242A, 0x3FC6, 0x2082}},
    {25142, 2, {0x10, 0xF8, 0x1F00, 0x1000, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {25143, 2, {0x200, 0x100, 0x100, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000}},
    {25144, 2, {0x0, 0x1FF8, 0x0, 0x0, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {25145, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x2FF0, 0x2020, 0x20C0, 0x2100, 0x4200, 0x4404, 0x8404, 0x3FC}},
    {25146, 2, {0x800, 0x400, 0x47C, 0x3F44, 0x2144, 0x2144, 0x2144, 0x3F44, 0x217C, 0x2040, 0x2040, 0x2040, 0x2042, 0x4042, 0x4042, 0x803E}},
    {25147, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2080, 0x2080, 0x2FFC, 0x2140, 0x2140, 0x4220, 0x4410, 0x8808, 0x1006}},
    {25148, 2, {0x0, 0xFC7E, 0x440, 0x440, 0x440, 0x7C7C, 0x4444, 0x4444, 0x4444, 0x7C7C, 0x444, 0x840, 0x840, 0x1040, 0x2040, 0x4040}},
    {25149, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2220, 0x2120, 0x2420, 0x2220, 0x207E, 0x5FA0, 0x4020, 0x8020, 0x20}},
    {25150, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x20A0, 0x2090, 0x2FFC, 0x2140, 0x2140, 0x4220, 0x4410, 0x8808, 0x1006}},
    {25151, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2100, 0x2080, 0x3FFC, 0x2200, 0x2200, 0x23F8, 0x4408, 0x4408, 0x8828, 0x1010}},
    {25152, 2, {0x208, 0x71C, 0x38E0, 0x2080, 0x2080, 0x3E80, 0x22FE, 0x2288, 0x2288, 0x3E88, 0x2088, 0x2088, 0x4108, 0x4108, 0x8208, 0x408}},
    {25153, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x2FFC, 0x2924, 0x2924, 0x2FFC, 0x4924, 0x4924, 0x8924, 0x80C}},
    {25154, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2080, 0x2080, 0x20FC, 0x2080, 0x2080, 0x2FF8, 0x4808, 0x4808, 0x8FF8, 0x808}},
    {25155, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x2FFC, 0x2804, 0x29E4, 0x2924, 0x4924, 0x49E4, 0x8814, 0x808}},
    {25156, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2100, 0x2200, 0x2FFC, 0x2804, 0x29E4, 0x2924, 0x4924, 0x49E4, 0x8814, 0x808}},
    {25157, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2100, 0x23F0, 0x2C20, 0x22C0, 0x2340, 0x3CFC, 0x2308, 0x4C90, 0x4060, 0x8180, 0x1E00}},
    {25158, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2100, 0x2080, 0x2FFC, 0x2140, 0x2244, 0x2628, 0x4A10, 0x5288, 0x8306, 0x200}},
    {25159, 2, {0x200, 0x100, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2000, 0x3F7C, 0x2104, 0x2924, 0x2514, 0x4924, 0x5144, 0x8514, 0x208}},
    {25160, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2FF8, 0x2888, 0x4FF8, 0x4802, 0x8802, 0x7FE}},
    {25161, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2120, 0x2120, 0x3F3E, 0x2120, 0x2120, 0x2F3C, 0x2120, 0x4120, 0x5F3E, 0x8120, 0x120}},
    {25162, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2488, 0x2490, 0x2960, 0x2218, 0x2C84, 0x348A, 0x2488, 0x4950, 0x4220, 0x8C18, 0x3006}},
    {25163, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {25164, 2, {0x1000, 0x1000, 0x1000, 0x1000, 0xF800, 0x1000, 0x1400, 0x1800, 0x3000, 0xD000, 0x1000, 0x1000, 0x1000, 0x1000, 0x5000, 0x2000}},
    {25165, 2, {0x80, 0x80, 0x80, 0x80, 0xFFFE, 0x180, 0x280, 0x280, 0x480, 0x480, 0x880, 0x1080, 0x2080, 0x4080, 0x280, 0x100}},
    {25166, 2, {0x1080, 0x1080, 0x1080, 0x1080, 0xFE80, 0x1080, 0x1080, 0x1280, 0x1C80, 0x3080, 0xD080, 0x1084, 0x1084, 0x1084, 0x507C, 0x2000}},
    {25167, 2, {0x1080, 0x1080, 0x1080, 0x1080, 0xFDF0, 0x1090, 0x1090, 0x1490, 0x1890, 0x3090, 0xD090, 0x1092, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25168, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0xFDFC, 0x1044, 0x1044, 0x1444, 0x1844, 0x3084, 0xD084, 0x1084, 0x1104, 0x1104, 0x5228, 0x2410}},
    {25169, 2, {0x1080, 0x1080, 0x1080, 0x1080, 0xFE80, 0x10A0, 0x1090, 0x1288, 0x1C84, 0x3084, 0xD080, 0x1080, 0x1080, 0x1080, 0x5080, 0x2080}},
    {25170, 2, {0x1000, 0x1020, 0x1120, 0x1120, 0xFD20, 0x1120, 0x1120, 0x1520, 0x1920, 0x3110, 0xD210, 0x1210, 0x1208, 0x1408, 0x5404, 0x2802}},
    {25171, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x1020, 0x1420, 0x1820, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25172, 2, {0x1000, 0x1000, 0x13F8, 0x1088, 0xFC90, 0x1090, 0x10A0, 0x14BC, 0x1884, 0x3084, 0xD104, 0x1104, 0x1104, 0x1204, 0x5228, 0x2410}},
    {25173, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0xFC40, 0x1080, 0x1080, 0x1480, 0x1900, 0x3110, 0xD108, 0x1208, 0x17FC, 0x1204, 0x5004, 0x2000}},
    {25174, 2, {0x1200, 0x1100, 0x1080, 0x1080, 0xFC40, 0x1040, 0x1040, 0x14A0, 0x18A0, 0x30A0, 0xD110, 0x1110, 0x1208, 0x1208, 0x5404, 0x2802}},
    {25175, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFC20, 0x1020, 0x11FC, 0x1420, 0x1820, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25176, 2, {0x1008, 0x103C, 0x13C0, 0x1040, 0xFC40, 0x1040, 0x1040, 0x147E, 0x1BC0, 0x3040, 0xD040, 0x1042, 0x1042, 0x1042, 0x503E, 0x2000}},
    {25177, 2, {0x1010, 0x1010, 0x1010, 0x1010, 0xFDFE, 0x1010, 0x1010, 0x1510, 0x1910, 0x3090, 0xD090, 0x1050, 0x1020, 0x1050, 0x5188, 0x2606}},
    {25178, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0xFD04, 0x1204, 0x1004, 0x1504, 0x1884, 0x3044, 0xD044, 0x1004, 0x1004, 0x1004, 0x5028, 0x2010}},
    {25179, 2, {0x1000, 0x1000, 0x11FC, 0x1020, 0xFE20, 0x1020, 0x1020, 0x1220, 0x1C20, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25180, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFC20, 0x1020, 0x1420, 0x1BFE, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25181, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x1040, 0x1440, 0x1880, 0x30FC, 0xD004, 0x1004, 0x1004, 0x1004, 0x5028, 0x2010}},
    {25182, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFC20, 0x1020, 0x1420, 0x1BFE, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25183, 2, {0x1000, 0x13F0, 0x1090, 0x1090, 0xFC90, 0x1090, 0x1090, 0x13F0, 0x1890, 0x3090, 0xD090, 0x1092, 0x108A, 0x108A, 0x5086, 0x2082}},
    {25184, 2, {0x1000, 0x17F8, 0x1208, 0x1288, 0xFE48, 0x1150, 0x1110, 0x1510, 0x18A0, 0x30A0, 0xD040, 0x1040, 0x10A0, 0x1110, 0x5208, 0x2C06}},
    {25185, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0xFD2C, 0x1134, 0x1164, 0x17A4, 0x1924, 0x3134, 0xD128, 0x1122, 0x1122, 0x1102, 0x50FE, 0x2000}},
    {25186, 2, {0x1040, 0x1040, 0x1080, 0x10FE, 0xFD00, 0x1200, 0x10FC, 0x1408, 0x1810, 0x3020, 0xD040, 0x1080, 0x1102, 0x1102, 0x50FE, 0x2000}},
    {25187, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x1104, 0x1504, 0x1904, 0x3104, 0xD104, 0x1104, 0x1104, 0x11FC, 0x5104, 0x2000}},
    {25188, 2, {0x1000, 0x1000, 0x13FC, 0x1090, 0xFC90, 0x1090, 0x1090, 0x1490, 0x1890, 0x3090, 0xD090, 0x1092, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25189, 2, {0x1020, 0x1020, 0x1020, 0x103C, 0xFDE0, 0x1020, 0x1020, 0x1420, 0x183E, 0x33E0, 0xD020, 0x1022, 0x1022, 0x1022, 0x501E, 0x2000}},
    {25190, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x1020, 0x1020, 0x17FE, 0x1820, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25191, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0xFDF8, 0x1048, 0x1048, 0x1448, 0x1948, 0x30C8, 0xD048, 0x10A8, 0x10AA, 0x110A, 0x5206, 0x2402}},
    {25192, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0xFC44, 0x1144, 0x1144, 0x1544, 0x1944, 0x3244, 0xD044, 0x1084, 0x1084, 0x1104, 0x5228, 0x2410}},
    {25193, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFD00, 0x1100, 0x1100, 0x1500, 0x1900, 0x3100, 0xD100, 0x1100, 0x1100, 0x1200, 0x5200, 0x2400}},
    {25194, 2, {0x1100, 0x1080, 0x10BC, 0x1204, 0xFA04, 0x1204, 0x1604, 0x1A04, 0x3204, 0xD204, 0x1204, 0x1204, 0x1204, 0x1204, 0x5214, 0x2208}},
    {25195, 2, {0x1000, 0x1000, 0x13FC, 0x1004, 0xFC04, 0x1004, 0x1004, 0x15FC, 0x1804, 0x3004, 0xD004, 0x1004, 0x1004, 0x13FC, 0x5004, 0x2000}},
    {25196, 2, {0x1000, 0x11F8, 0x1010, 0x1020, 0xFC40, 0x1080, 0x11FE, 0x1492, 0x1892, 0x3092, 0xD112, 0x1122, 0x1222, 0x1442, 0x5094, 0x2108}},
    {25197, 2, {0x1000, 0x11FC, 0x1044, 0x1044, 0xFC44, 0x1044, 0x1044, 0x15FC, 0x1884, 0x3084, 0xD084, 0x1084, 0x1084, 0x1084, 0x57FE, 0x2000}},
    {25198, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0xFD08, 0x1104, 0x1204, 0x15FA, 0x1888, 0x3088, 0xD088, 0x1088, 0x1108, 0x1108, 0x5228, 0x2410}},
    {25199, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0xFD20, 0x1120, 0x113C, 0x1520, 0x1920, 0x3120, 0xD120, 0x1120, 0x1120, 0x1120, 0x57FE, 0x2000}},
    {25200, 2, {0x1040, 0x1050, 0x1048, 0x1048, 0xFC40, 0x11FE, 0x1050, 0x1450, 0x1850, 0x3050, 0xD090, 0x1090, 0x1092, 0x1112, 0x510E, 0x2200}},
    {25201, 2, {0x1000, 0x13FC, 0x1084, 0x1088, 0xFC88, 0x1090, 0x109C, 0x1484, 0x1944, 0x3144, 0xD128, 0x1128, 0x1210, 0x1228, 0x5444, 0x2182}},
    {25202, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1288, 0x1446, 0x1040, 0x1800, 0x33F8, 0xD008, 0x1010, 0x1010, 0x1020, 0x5020, 0x2040}},
    {25203, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x11FC, 0x1144, 0x1544, 0x1944, 0x3128, 0xD128, 0x1110, 0x1110, 0x1228, 0x5244, 0x2482}},
    {25204, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0xFC88, 0x1104, 0x1202, 0x1088, 0x1888, 0x3088, 0xD088, 0x1088, 0x1088, 0x1108, 0x5108, 0x2208}},
    {25205, 2, {0x1020, 0x1020, 0x1020, 0x1050, 0xFC50, 0x1088, 0x1504, 0x1A02, 0x3060, 0xD010, 0x1000, 0x10C0, 0x1020, 0x1010, 0x5008, 0x2000}},
    {25206, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x1020, 0x1420, 0x1BFE, 0x3020, 0xD050, 0x1050, 0x1088, 0x1088, 0x5104, 0x2202}},
    {25207, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x1020, 0x13FE, 0x1420, 0x1820, 0x3050, 0xD050, 0x1050, 0x1088, 0x1088, 0x5104, 0x2202}},
    {25208, 2, {0x1010, 0x1110, 0x1110, 0x1110, 0xFD10, 0x11FE, 0x1100, 0x1500, 0x1900, 0x31F8, 0xD108, 0x1108, 0x1108, 0x1208, 0x5208, 0x2408}},
    {25209, 2, {0x1010, 0x1110, 0x1110, 0x1112, 0xFD12, 0x1114, 0x11D8, 0x1510, 0x1910, 0x3110, 0xD110, 0x1112, 0x1152, 0x1192, 0x510E, 0x2000}},
    {25210, 2, {0x1008, 0x103C, 0x13E0, 0x1220, 0xFE20, 0x1220, 0x1220, 0x17FE, 0x1A20, 0x3210, 0xD210, 0x1212, 0x120A, 0x128A, 0x5306, 0x2202}},
    {25211, 2, {0x1080, 0x1080, 0x1080, 0x10FC, 0xFD04, 0x1108, 0x1240, 0x1440, 0x1840, 0x30A0, 0xD0A0, 0x1090, 0x1110, 0x1108, 0x5204, 0x2402}},
    {25212, 2, {0x2000, 0x23FE, 0x2200, 0x2200, 0xFAF8, 0x2288, 0x2288, 0x2A88, 0x3288, 0xE2A8, 0x2290, 0x2282, 0x2482, 0x2482, 0xA87E, 0x5000}},
    {25213, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1124, 0x1124, 0x1524, 0x1924, 0x31FC, 0xD024, 0x1020, 0x1022, 0x1022, 0x501E, 0x2000}},
    {25214, 2, {0x1090, 0x1088, 0x1088, 0x1080, 0xFCBC, 0x13C0, 0x1080, 0x1488, 0x1888, 0x3090, 0xD060, 0x1044, 0x10A4, 0x1114, 0x520C, 0x2004}},
    {25215, 2, {0x1FE0, 0x40, 0x80, 0x104, 0x7928, 0xFF0, 0x920, 0x1110, 0x17D0, 0x2108, 0x2108, 0x4FE4, 0x8102, 0x100, 0x500, 0x200}},
    {25216, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x1020, 0x15FC, 0x1884, 0x3088, 0xD048, 0x1050, 0x1020, 0x1050, 0x5188, 0x2606}},
    {25217, 2, {0x1040, 0x1040, 0x1080, 0x1088, 0xFD04, 0x13FE, 0x1002, 0x1490, 0x1890, 0x3090, 0xD090, 0x1090, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25218, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x1020, 0x1420, 0x19FC, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x57FE, 0x2000}},
    {25219, 2, {0x1040, 0x1020, 0x1020, 0x1000, 0xFDFE, 0x1040, 0x1040, 0x1460, 0x1850, 0x3048, 0xD044, 0x1044, 0x1040, 0x1040, 0x5040, 0x2040}},
    {25220, 2, {0x1020, 0x1020, 0x1020, 0x10A8, 0xFCA4, 0x10A2, 0x1122, 0x1520, 0x1A24, 0x3024, 0xD028, 0x1008, 0x1010, 0x1020, 0x50C0, 0x2300}},
    {25221, 2, {0x1080, 0x1080, 0x1080, 0x10FE, 0xFD02, 0x1242, 0x1042, 0x1482, 0x18A2, 0x3112, 0xD3FA, 0x110A, 0x1002, 0x1002, 0x5014, 0x2008}},
    {25222, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC88, 0x1088, 0x1088, 0x1488, 0x1888, 0x3050, 0xD050, 0x1020, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25223, 2, {0x1000, 0x1000, 0x11FE, 0x1102, 0xFD02, 0x1102, 0x1102, 0x15FA, 0x1902, 0x3102, 0xD102, 0x1102, 0x1102, 0x11FE, 0x5102, 0x2000}},
    {25224, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1104, 0x1504, 0x1904, 0x31FC, 0xD104, 0x1104, 0x1204, 0x1204, 0x5414, 0x2808}},
    {25225, 2, {0x1040, 0x1040, 0x1040, 0x13F8, 0xF848, 0x1048, 0x1448, 0x1848, 0x37FE, 0xD040, 0x10A0, 0x10A0, 0x1110, 0x1110, 0x5208, 0x2406}},
    {25226, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFD24, 0x1124, 0x1124, 0x15FC, 0x1904, 0x3100, 0xD100, 0x1100, 0x1102, 0x1102, 0x50FE, 0x2000}},
    {25227, 2, {0x1000, 0x1040, 0x1020, 0x1010, 0xFC90, 0x1080, 0x1080, 0x1284, 0x1A82, 0x3282, 0xD282, 0x1488, 0x1088, 0x1088, 0x5078, 0x2000}},
    {25228, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0xFD22, 0x1224, 0x1020, 0x1420, 0x1850, 0x3050, 0xD050, 0x1050, 0x1090, 0x1092, 0x5112, 0x220E}},
    {25229, 2, {0x1008, 0x1068, 0x1388, 0x1088, 0xFC88, 0x1088, 0x1488, 0x1BFE, 0x3088, 0xD088, 0x1088, 0x1088, 0x1108, 0x1108, 0x5208, 0x2408}},
    {25230, 2, {0x1000, 0x1000, 0x11FC, 0x1000, 0xFC00, 0x1000, 0x13FE, 0x1420, 0x1820, 0x3040, 0xD040, 0x1088, 0x1104, 0x13FE, 0x5102, 0x2000}},
    {25231, 2, {0x1000, 0x1000, 0x11FC, 0x1000, 0xFC00, 0x1000, 0x13FE, 0x1090, 0x1890, 0x3090, 0xD090, 0x1090, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25232, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x1122, 0x1122, 0x1522, 0x1952, 0x314A, 0xD18A, 0x1102, 0x1102, 0x1102, 0x510A, 0x2104}},
    {25233, 2, {0x1000, 0x1080, 0x133C, 0x1224, 0xFE24, 0x1224, 0x1224, 0x1624, 0x1A24, 0x3224, 0xD2B4, 0x1328, 0x1220, 0x1020, 0x5020, 0x2020}},
    {25234, 2, {0x1000, 0x11F8, 0x1008, 0x1050, 0xFC20, 0x1010, 0x13FE, 0x1422, 0x1824, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25235, 2, {0x1008, 0x101C, 0x11F0, 0x1150, 0xFD50, 0x1150, 0x1150, 0x1550, 0x1950, 0x3148, 0xD148, 0x1148, 0x1144, 0x1244, 0x5242, 0x2400}},
    {25236, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1040, 0x1040, 0x14D0, 0x1948, 0x3244, 0xD444, 0x1040, 0x1040, 0x1040, 0x5040, 0x2040}},
    {25237, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFC88, 0x1106, 0x1200, 0x15FC, 0x1884, 0x3084, 0xD048, 0x1050, 0x1020, 0x1050, 0x5088, 0x2306}},
    {25238, 2, {0x1010, 0x1110, 0x1090, 0x1090, 0xFC10, 0x1110, 0x1090, 0x1490, 0x1810, 0x301E, 0xD3F0, 0x1010, 0x1010, 0x1010, 0x5010, 0x2010}},
    {25239, 2, {0x1080, 0x1040, 0x1040, 0x13FC, 0xFC00, 0x1000, 0x11F0, 0x1510, 0x1910, 0x3110, 0xD110, 0x1112, 0x1112, 0x1212, 0x520E, 0x2400}},
    {25240, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x1100, 0x11FE, 0x1510, 0x1910, 0x3110, 0xD110, 0x1110, 0x1110, 0x1210, 0x5210, 0x2410}},
    {25241, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x11FC, 0x1020, 0x1420, 0x1820, 0x33FE, 0xD020, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25242, 2, {0x1000, 0x13FC, 0x1040, 0x1040, 0xFC40, 0x1040, 0x17FE, 0x1040, 0x18A0, 0x30A0, 0xD0A0, 0x1120, 0x1122, 0x1222, 0x541E, 0x2800}},
    {25243, 2, {0x1210, 0x1210, 0x1210, 0x127C, 0xFA14, 0x1794, 0x1294, 0x1A94, 0x32A4, 0xD2A4, 0x12D4, 0x128A, 0x1482, 0x1482, 0x587E, 0x3000}},
    {25244, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0xFBFE, 0x1080, 0x1480, 0x18FC, 0x3144, 0xD144, 0x1148, 0x1250, 0x1220, 0x1450, 0x5888, 0x2306}},
    {25245, 2, {0x2080, 0x2080, 0x213E, 0x2102, 0xFA42, 0x2242, 0x27C2, 0x2882, 0x3082, 0xE102, 0x2142, 0x2222, 0x27E2, 0x2222, 0xA00A, 0x4004}},
    {25246, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1524, 0x1920, 0x3120, 0xD110, 0x1110, 0x1208, 0x1208, 0x5404, 0x2802}},
    {25247, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1040, 0x11FE, 0x1440, 0x1880, 0x31FC, 0xD004, 0x1088, 0x1050, 0x1020, 0x5010, 0x2010}},
    {25248, 2, {0x1000, 0x11FE, 0x1100, 0x1104, 0xFD44, 0x1128, 0x1128, 0x1510, 0x1910, 0x3128, 0xD128, 0x1144, 0x1184, 0x1100, 0x51FE, 0x2000}},
    {25249, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1208, 0x1406, 0x1110, 0x1920, 0x3140, 0xD180, 0x1100, 0x1104, 0x1104, 0x50FC, 0x2000}},
    {25250, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1208, 0x1406, 0x11F0, 0x1910, 0x3110, 0xD150, 0x1120, 0x1104, 0x1104, 0x50FC, 0x2000}},
    {25251, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0xFD04, 0x1204, 0x1084, 0x1444, 0x1844, 0x3014, 0xD024, 0x1044, 0x1184, 0x1004, 0x5028, 0x2010}},
    {25252, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFD02, 0x1102, 0x1102, 0x15FE, 0x1902, 0x3100, 0xD100, 0x1100, 0x1200, 0x1200, 0x5400, 0x2800}},
    {25253, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD14, 0x1108, 0x1100, 0x15FC, 0x1944, 0x3144, 0xD128, 0x1128, 0x1110, 0x1128, 0x5144, 0x2182}},
    {25254, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x1222, 0x1622, 0x1A52, 0x324A, 0xD28A, 0x1302, 0x1202, 0x1202, 0x520A, 0x2204}},
    {25255, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x1104, 0x1504, 0x19FC, 0x3104, 0xD000, 0x1090, 0x1088, 0x1104, 0x5202, 0x2402}},
    {25256, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFD24, 0x10A4, 0x10A8, 0x1420, 0x1BFE, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25257, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x1524, 0x1924, 0x3124, 0xD7FE, 0x1104, 0x1104, 0x1104, 0x5114, 0x2108}},
    {25258, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0xFC80, 0x10A0, 0x1520, 0x19FC, 0x3324, 0xD524, 0x1124, 0x1124, 0x1134, 0x1128, 0x5020, 0x2020}},
    {25259, 2, {0x1010, 0x1010, 0x1010, 0x11FE, 0xFD12, 0x1114, 0x1110, 0x15FC, 0x1944, 0x3144, 0xD128, 0x1128, 0x1110, 0x1228, 0x5244, 0x2482}},
    {25260, 2, {0x1020, 0x1020, 0x1040, 0x1088, 0xFD04, 0x13FE, 0x1002, 0x1400, 0x19FC, 0x3104, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25261, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1404, 0x1000, 0x18F0, 0x3090, 0xD090, 0x1090, 0x1090, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25262, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0xFA08, 0x1426, 0x1040, 0x1880, 0x3310, 0xD020, 0x1040, 0x1088, 0x1310, 0x1020, 0x50C0, 0x2700}},
    {25263, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x1108, 0x1508, 0x1908, 0x31F8, 0xD108, 0x1108, 0x1108, 0x1108, 0x57FE, 0x2000}},
    {25264, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x1124, 0x1524, 0x1924, 0x33FE, 0xD020, 0x1050, 0x1050, 0x1088, 0x5104, 0x2202}},
    {25265, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0xFA04, 0x15F4, 0x1114, 0x1914, 0x3114, 0xD1F4, 0x1104, 0x1128, 0x1112, 0x1102, 0x50FE, 0x2000}},
    {25266, 2, {0x1000, 0x1000, 0x17FE, 0x1008, 0xFC08, 0x13C8, 0x1248, 0x1648, 0x1A48, 0x3248, 0xD3C8, 0x1248, 0x1008, 0x1008, 0x5028, 0x2010}},
    {25267, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1100, 0x1540, 0x1944, 0x3148, 0xD170, 0x1140, 0x1142, 0x1242, 0x523E, 0x2400}},
    {25268, 2, {0x2048, 0x2248, 0x2248, 0x2248, 0xFA48, 0x27FE, 0x2248, 0x2A48, 0x3248, 0xE248, 0x2278, 0x2200, 0x2200, 0x2200, 0xA3FE, 0x4000}},
    {25269, 2, {0x1008, 0x103C, 0x13E0, 0x1220, 0xFE20, 0x1220, 0x1220, 0x13FE, 0x1A20, 0x3210, 0xD210, 0x1212, 0x120A, 0x128A, 0x5326, 0x2212}},
    {25270, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0xFDFC, 0x1120, 0x1220, 0x1020, 0x1BFE, 0x3020, 0xD050, 0x1050, 0x1088, 0x1088, 0x5104, 0x2202}},
    {25271, 2, {0x1000, 0x13FC, 0x1020, 0x1020, 0xFC40, 0x1040, 0x14D0, 0x1948, 0x3244, 0xD444, 0x1040, 0x1040, 0x1040, 0x1000, 0x57FE, 0x2000}},
    {25272, 2, {0x1008, 0x103C, 0x13C0, 0x1000, 0xFC40, 0x1020, 0x13FC, 0x1408, 0x1810, 0x3020, 0xD040, 0x1080, 0x1100, 0x1280, 0x547E, 0x2000}},
    {25273, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x1020, 0x15FC, 0x1870, 0x30A8, 0xD0A8, 0x1124, 0x1124, 0x1222, 0x5020, 0x2020}},
    {25274, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x1020, 0x13FE, 0x1870, 0x30A8, 0xD0A8, 0x1124, 0x1124, 0x1222, 0x5020, 0x2020}},
    {25275, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFA22, 0x1222, 0x1622, 0x1BFE, 0x3222, 0xD222, 0x1222, 0x13FE, 0x1222, 0x1020, 0x5020, 0x2020}},
    {25276, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1524, 0x1924, 0x31FC, 0xD124, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25277, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x1124, 0x1524, 0x1924, 0x31FC, 0xD124, 0x1124, 0x1124, 0x1124, 0x51FC, 0x2104}},
    {25278, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x1020, 0x1420, 0x1BFE, 0x3020, 0xD040, 0x1040, 0x1088, 0x1104, 0x53FE, 0x2102}},
    {25279, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFE04, 0x13FC, 0x1220, 0x1620, 0x1BFE, 0x3220, 0xD220, 0x1210, 0x1212, 0x128A, 0x5306, 0x2202}},
    {25280, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1450, 0x1850, 0x3050, 0xD050, 0x1092, 0x1092, 0x1112, 0x520E, 0x2400}},
    {25281, 2, {0x2200, 0x2200, 0x2200, 0x221E, 0xF7D2, 0x2252, 0x2252, 0x2252, 0x3252, 0xE252, 0x2252, 0x2252, 0x2452, 0x245E, 0xA952, 0x5080}},
    {25282, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0xFC94, 0x1094, 0x13FC, 0x1690, 0x1A90, 0x33FE, 0xD092, 0x1092, 0x111A, 0x1114, 0x5210, 0x2410}},
    {25283, 2, {0x1080, 0x1080, 0x1080, 0x10FE, 0xFD40, 0x1140, 0x1240, 0x107C, 0x1840, 0x3040, 0xD040, 0x107E, 0x1040, 0x1040, 0x5040, 0x2040}},
    {25284, 2, {0x1040, 0x1020, 0x1000, 0x13FE, 0xFC20, 0x1020, 0x1020, 0x1420, 0x19FC, 0x3020, 0xD020, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25285, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1504, 0x1904, 0x3104, 0xD1FC, 0x1104, 0x1000, 0x1000, 0x53FE, 0x2000}},
    {25286, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x1100, 0x11FE, 0x1510, 0x1910, 0x3130, 0xD118, 0x1114, 0x1112, 0x1210, 0x5210, 0x2410}},
    {25287, 2, {0x1000, 0x11F8, 0x1108, 0x1148, 0xFD28, 0x1108, 0x1108, 0x17FE, 0x1A08, 0x3288, 0xD248, 0x1208, 0x13FE, 0x1008, 0x5050, 0x2020}},
    {25288, 2, {0x1040, 0x1040, 0x1040, 0x107E, 0xFC40, 0x1040, 0x1440, 0x1840, 0x33FC, 0xD204, 0x1204, 0x1204, 0x1204, 0x1204, 0x53FC, 0x2204}},
    {25289, 2, {0x1040, 0x1020, 0x1020, 0x1000, 0xFBFE, 0x1000, 0x1404, 0x1904, 0x3104, 0xD088, 0x1088, 0x1088, 0x1090, 0x1010, 0x57FE, 0x2000}},
    {25290, 2, {0x1088, 0x1088, 0x1088, 0x1108, 0xFD7E, 0x1308, 0x1508, 0x1148, 0x1928, 0x3128, 0xD108, 0x1108, 0x1108, 0x1108, 0x5128, 0x2110}},
    {25291, 2, {0x2210, 0x2210, 0x2210, 0x227C, 0xFA14, 0x27D4, 0x2294, 0x2A94, 0x32A4, 0xE2A4, 0x22D4, 0x228A, 0x2482, 0x2482, 0xA87E, 0x5000}},
    {25292, 2, {0x1020, 0x1020, 0x1124, 0x10A4, 0xFCA8, 0x1020, 0x11FC, 0x1420, 0x1820, 0x3020, 0xD3FE, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25293, 2, {0x1020, 0x1020, 0x1040, 0x11FC, 0xFD04, 0x1104, 0x1104, 0x1504, 0x19FC, 0x3104, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25294, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1248, 0x1426, 0x1020, 0x1BF8, 0x3008, 0xD010, 0x1110, 0x10A0, 0x1040, 0x5020, 0x2020}},
    {25295, 2, {0x1000, 0x1000, 0xFEFC, 0x2244, 0x6428, 0x1810, 0x2428, 0xC2C6, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25296, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1040, 0x1440, 0x1BFC, 0x3044, 0xD044, 0x1084, 0x1084, 0x1104, 0x5228, 0x2410}},
    {25297, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFE, 0x1088, 0x1088, 0x1488, 0x1888, 0x30F8, 0xD088, 0x1088, 0x1088, 0x1088, 0x50F8, 0x2088}},
    {25298, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0xFD00, 0x11FC, 0x1104, 0x1504, 0x1904, 0x3104, 0xD1FC, 0x1100, 0x1100, 0x1100, 0x51FE, 0x2000}},
    {25299, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1040, 0x1040, 0x14FC, 0x1984, 0x3284, 0xD484, 0x1084, 0x1084, 0x1084, 0x50FC, 0x2084}},
    {25300, 2, {0x1050, 0x1048, 0x1048, 0x1040, 0xFBFE, 0x1080, 0x1480, 0x18FC, 0x3144, 0xD144, 0x1128, 0x1128, 0x1210, 0x1228, 0x5444, 0x2182}},
    {25301, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFD02, 0x1204, 0x1080, 0x1488, 0x1890, 0x30A0, 0xD0C0, 0x1082, 0x1082, 0x1082, 0x507E, 0x2000}},
    {25302, 2, {0x1080, 0x1080, 0x10FE, 0x1100, 0xFE20, 0x1120, 0x112C, 0x1574, 0x1BA4, 0x3124, 0xD134, 0x1128, 0x1122, 0x1102, 0x50FE, 0x2000}},
    {25303, 2, {0x2020, 0x2120, 0x2120, 0x2220, 0xFAFC, 0x24A4, 0x2724, 0x2924, 0x3224, 0xE224, 0x24A4, 0x27A4, 0x20C4, 0x2044, 0xA094, 0x4108}},
    {25304, 2, {0x1080, 0x1080, 0x1100, 0x11FC, 0xFA04, 0x1404, 0x11E4, 0x1924, 0x3124, 0xD124, 0x11E4, 0x1124, 0x1004, 0x1004, 0x5028, 0x2010}},
    {25305, 2, {0x1020, 0x1020, 0x1124, 0x1124, 0xFD24, 0x1124, 0x15FC, 0x1824, 0x3020, 0xD020, 0x1222, 0x1222, 0x1222, 0x1222, 0x53FE, 0x2002}},
    {25306, 2, {0x1020, 0x1020, 0x1040, 0x1088, 0xFD04, 0x13FE, 0x1002, 0x1488, 0x1888, 0x33FE, 0xD088, 0x1088, 0x1108, 0x1108, 0x5208, 0x2408}},
    {25307, 2, {0x1000, 0x13FC, 0x1084, 0x1084, 0xFC84, 0x1104, 0x1114, 0x1208, 0x1C00, 0x31FC, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25308, 2, {0x400, 0x1EFC, 0xF020, 0x1020, 0x10FC, 0xFE20, 0x1020, 0x10FC, 0x1020, 0xFE20, 0x11FE, 0x1020, 0x2020, 0x2020, 0x4020, 0x8020}},
    {25309, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x1020, 0x15FC, 0x1820, 0x3020, 0xD3FE, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25310, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x1100, 0x1100, 0x15FE, 0x1910, 0x3110, 0xD110, 0x1110, 0x1110, 0x1110, 0x57FE, 0x2000}},
    {25311, 2, {0x1008, 0x1088, 0x1048, 0x1228, 0xFA28, 0x1208, 0x1208, 0x1A08, 0x3208, 0xD208, 0x1210, 0x1298, 0x1324, 0x1224, 0x5042, 0x2082}},
    {25312, 2, {0x2200, 0x2238, 0x2228, 0xFBA8, 0x22A8, 0x22A8, 0x34A8, 0x26AA, 0xEAAA, 0x22AA, 0x2126, 0x2140, 0x2280, 0x2260, 0xA41E, 0x4800}},
    {25313, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFD00, 0x1110, 0x1110, 0x1510, 0x1910, 0x3120, 0xD128, 0x1124, 0x1244, 0x12FE, 0x5442, 0x2800}},
    {25314, 2, {0x1090, 0x1088, 0x1088, 0x1080, 0xFBFE, 0x10A0, 0x14A0, 0x18A4, 0x30A4, 0xD128, 0x1128, 0x1132, 0x1222, 0x1262, 0x549E, 0x2800}},
    {25315, 2, {0x1040, 0x1040, 0x17FC, 0x1080, 0xFC80, 0x13E0, 0x1120, 0x1220, 0x1BFC, 0x3020, 0xD128, 0x1124, 0x1222, 0x1422, 0x50A0, 0x2040}},
    {25316, 2, {0x1040, 0x1040, 0x1040, 0x107C, 0xFC40, 0x1040, 0x1440, 0x1BFE, 0x3040, 0xD040, 0x1050, 0x1048, 0x1044, 0x1040, 0x5040, 0x2040}},
    {25317, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0xFC44, 0x27FC, 0x2444, 0x2C44, 0x3444, 0xE7FC, 0x2444, 0x2444, 0x2444, 0x2844, 0xA854, 0x5008}},
    {25318, 2, {0x1000, 0x1108, 0x1088, 0x1090, 0xFC00, 0x13FE, 0x1000, 0x1400, 0x1800, 0x31FC, 0xD000, 0x1000, 0x1000, 0x1000, 0x53FE, 0x2000}},
    {25319, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1404, 0x1000, 0x1800, 0x33FE, 0xD020, 0x1020, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25320, 2, {0x1040, 0x1048, 0x1244, 0x1240, 0xFBFE, 0x1080, 0x1480, 0x18FC, 0x3144, 0xD144, 0x1148, 0x1250, 0x1220, 0x1450, 0x5888, 0x2306}},
    {25321, 2, {0x2000, 0x27F8, 0x2208, 0x2110, 0xF0A0, 0x2040, 0x21B0, 0x264E, 0x3040, 0xE3F8, 0x2040, 0x2040, 0x27FC, 0x2040, 0xA040, 0x4040}},
    {25322, 2, {0x1000, 0x17FE, 0x1090, 0x1090, 0xFC90, 0x13FC, 0x1294, 0x1694, 0x1A94, 0x3294, 0xD29C, 0x1304, 0x1204, 0x1204, 0x53FC, 0x2204}},
    {25323, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1108, 0x1108, 0x15F8, 0x1944, 0x3148, 0xD130, 0x1120, 0x1110, 0x1148, 0x5186, 0x2100}},
    {25324, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x13FE, 0x1020, 0x1420, 0x1820, 0x31FC, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25325, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0xFBFE, 0x1020, 0x1420, 0x1BA0, 0x3120, 0xD110, 0x1110, 0x1110, 0x11CA, 0x170A, 0x5206, 0x2002}},
    {25326, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x11FC, 0x1400, 0x1800, 0x31FC, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25327, 2, {0x1000, 0x11FC, 0x1008, 0x1010, 0xFC22, 0x13B2, 0x10B4, 0x14A8, 0x1928, 0x3124, 0xD222, 0x14A0, 0x1040, 0x1000, 0x57FE, 0x2000}},
    {25328, 2, {0x1080, 0x1088, 0x109C, 0x1170, 0xFD10, 0x1310, 0x1510, 0x1110, 0x19FE, 0x3110, 0xD110, 0x1110, 0x1110, 0x1110, 0x517C, 0x2100}},
    {25329, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFE, 0x1088, 0x1088, 0x1488, 0x1888, 0x33FE, 0xD000, 0x1088, 0x1084, 0x1104, 0x5202, 0x2402}},
    {25330, 2, {0x440, 0x440, 0x7FFC, 0x440, 0xFFFE, 0x820, 0x3018, 0xC0E6, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25331, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x10F0, 0x3F08, 0xC106, 0x1FE0, 0x100, 0x3FF8, 0x100, 0x500, 0x200}},
    {25332, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x1202, 0x11FC, 0x1420, 0x1820, 0x3020, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25333, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0xFC80, 0x1080, 0x1178, 0x1508, 0x1B10, 0x3510, 0xD1FE, 0x1110, 0x1110, 0x1110, 0x5150, 0x2120}},
    {25334, 2, {0x1124, 0x1124, 0x1248, 0x1124, 0xFD24, 0x1000, 0x1080, 0x14FC, 0x1904, 0x3184, 0xD248, 0x1428, 0x1010, 0x1020, 0x50C0, 0x2300}},
    {25335, 2, {0x1040, 0x1044, 0x11F4, 0x1048, 0xFC50, 0x13FE, 0x1040, 0x1480, 0x19FE, 0x3240, 0xD480, 0x10FC, 0x1004, 0x1004, 0x5028, 0x2010}},
    {25336, 2, {0x1040, 0x1040, 0x10FC, 0x1108, 0xFA90, 0x1060, 0x1448, 0x1990, 0x363E, 0xD042, 0x1084, 0x1348, 0x1030, 0x1020, 0x50C0, 0x2700}},
    {25337, 2, {0x1040, 0x1040, 0x11FC, 0x1044, 0xFC84, 0x1084, 0x1528, 0x1A10, 0x3108, 0xD3DE, 0x114A, 0x114A, 0x114A, 0x1252, 0x535A, 0x24A4}},
    {25338, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x1124, 0x1524, 0x1924, 0x312C, 0xD070, 0x10A8, 0x1124, 0x1622, 0x5020, 0x2020}},
    {25339, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0xFC80, 0x1090, 0x1090, 0x1512, 0x1952, 0x3154, 0xD290, 0x1228, 0x1428, 0x1044, 0x5084, 0x2102}},
    {25340, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFDFE, 0x1088, 0x1088, 0x1488, 0x1888, 0x33FE, 0xD088, 0x1088, 0x1108, 0x1108, 0x5208, 0x2408}},
    {25341, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1524, 0x19FC, 0x3020, 0xD024, 0x1018, 0x1032, 0x10CA, 0x5306, 0x2002}},
    {25342, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x1202, 0x11FC, 0x1400, 0x1800, 0x31FC, 0xD104, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25343, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25344, 2, {0x1004, 0x101E, 0x11E0, 0x1100, 0xFD06, 0x1178, 0x1150, 0x1550, 0x1952, 0x3154, 0xD148, 0x1148, 0x1144, 0x1254, 0x5262, 0x2440}},
    {25345, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x13FE, 0x1008, 0x1808, 0x33FE, 0xD008, 0x1088, 0x1048, 0x1008, 0x5028, 0x2010}},
    {25346, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x13FE, 0x1000, 0x1820, 0x3020, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25347, 2, {0x1000, 0x11FE, 0x1020, 0x1040, 0xFC88, 0x1104, 0x11FE, 0x1422, 0x1820, 0x3020, 0xD1FE, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25348, 2, {0x1040, 0x1044, 0x1244, 0x1148, 0xFD50, 0x1040, 0x17FE, 0x1090, 0x1890, 0x3090, 0xD090, 0x1092, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25349, 2, {0x1000, 0x13F8, 0x1088, 0x1090, 0xFCBC, 0x1104, 0x1514, 0x1A48, 0x3040, 0xD7FE, 0x1040, 0x10E0, 0x1150, 0x1248, 0x5446, 0x2040}},
    {25350, 2, {0x1000, 0x11F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x124E, 0x1440, 0x1BFC, 0x3040, 0xD0E0, 0x1150, 0x1248, 0x1446, 0x5040, 0x2040}},
    {25351, 2, {0x1100, 0x1104, 0x1138, 0x11C0, 0xFD02, 0x1102, 0x10FE, 0x1400, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25352, 2, {0x800, 0x7F7C, 0x824, 0x3E24, 0x824, 0x7F44, 0x854, 0x8F8, 0x3F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25353, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1444, 0x1040, 0x1BFE, 0x3088, 0xD088, 0x1108, 0x10D0, 0x1020, 0x1050, 0x5088, 0x2304}},
    {25354, 2, {0x1040, 0x1040, 0x107C, 0x1040, 0xFC40, 0x13FE, 0x1000, 0x1000, 0x1BFE, 0x3040, 0xD050, 0x1048, 0x1044, 0x1040, 0x5040, 0x2040}},
    {25355, 2, {0x1000, 0x13FE, 0x1220, 0x1220, 0xFE20, 0x13FC, 0x1204, 0x1604, 0x1A04, 0x3204, 0xD3FC, 0x1220, 0x1220, 0x1220, 0x53FE, 0x2000}},
    {25356, 2, {0x1080, 0x1080, 0x11F8, 0x1208, 0xFD10, 0x10A0, 0x1440, 0x18A0, 0x3318, 0xDC06, 0x13F8, 0x1208, 0x1208, 0x1208, 0x53F8, 0x2208}},
    {25357, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC00, 0x1088, 0x1104, 0x1202, 0x1888, 0x3088, 0xD050, 0x1050, 0x1020, 0x1050, 0x5088, 0x2306}},
    {25358, 2, {0x2040, 0x2040, 0x27FC, 0x20A0, 0xF110, 0x2208, 0x2DF6, 0x2000, 0x37FC, 0xE080, 0x2100, 0x23F8, 0x2008, 0x2008, 0xA050, 0x4020}},
    {25359, 2, {0x1000, 0x11FE, 0x1102, 0x1102, 0xFD7A, 0x1102, 0x1102, 0x157A, 0x194A, 0x314A, 0xD14A, 0x117A, 0x1102, 0x1102, 0x510A, 0x2104}},
    {25360, 2, {0x1000, 0x1000, 0xFE7C, 0x2244, 0x4244, 0x3444, 0x87C, 0x3400, 0xC220, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25361, 2, {0x1090, 0x1090, 0x1090, 0x1292, 0xFD94, 0x1098, 0x1090, 0x1598, 0x1A94, 0x3492, 0xD090, 0x1090, 0x1112, 0x1112, 0x5212, 0x240E}},
    {25362, 2, {0x2002, 0x2002, 0x2FC2, 0x2212, 0xFA12, 0x23D2, 0x2252, 0x2C52, 0x3652, 0xE552, 0x2892, 0x2092, 0x2102, 0x2202, 0xA40A, 0x4804}},
    {25363, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1404, 0x1038, 0x1BC0, 0x3040, 0xD040, 0x107E, 0x17C0, 0x1040, 0x1042, 0x5042, 0x203E}},
    {25364, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC40, 0x1040, 0x10A0, 0x14A2, 0x19A4, 0x3298, 0xD490, 0x1088, 0x1084, 0x10A2, 0x50C0, 0x2080}},
    {25365, 2, {0x1000, 0x13FE, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x1108, 0x1508, 0x19F8, 0x3108, 0xD108, 0x111E, 0x17E8, 0x1008, 0x5008, 0x2008}},
    {25366, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC88, 0x1104, 0x1202, 0x14F8, 0x1810, 0x3020, 0xD040, 0x1080, 0x1102, 0x1102, 0x50FE, 0x2000}},
    {25367, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1024, 0x11FC, 0x1520, 0x19FE, 0x3022, 0xD02A, 0x1054, 0x1050, 0x1088, 0x5104, 0x2202}},
    {25368, 2, {0x1020, 0x1020, 0x1124, 0x1122, 0xFA2A, 0x1010, 0x1460, 0x1B80, 0x3040, 0xD040, 0x13FC, 0x1044, 0x1084, 0x1084, 0x5114, 0x2208}},
    {25369, 2, {0x2208, 0x1108, 0x1110, 0x20, 0xFFFE, 0x820, 0x1010, 0x20E8, 0xDF06, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25370, 2, {0x1080, 0xFBF0, 0x1090, 0x1990, 0xF094, 0x1154, 0x520C, 0x24E4, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25371, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4444, 0xF0, 0x3F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25372, 2, {0x1000, 0x13FC, 0x1090, 0x1090, 0xFC90, 0x1092, 0x1492, 0x1294, 0x1A94, 0x3298, 0xD090, 0x1090, 0x1090, 0x1090, 0x57FE, 0x2000}},
    {25373, 2, {0x2408, 0x2208, 0x2208, 0x2008, 0xF8FE, 0x2E08, 0x2248, 0x2A28, 0x3228, 0xE208, 0x2208, 0x2228, 0x2210, 0x2500, 0xA8FE, 0x4000}},
    {25374, 2, {0x1010, 0x1210, 0x1110, 0x1110, 0xFCFE, 0x1010, 0x1310, 0x1528, 0x1924, 0x3144, 0xD142, 0x1182, 0x1100, 0x1280, 0x547E, 0x2000}},
    {25375, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1124, 0x10A4, 0x14A8, 0x1820, 0x33FE, 0xD050, 0x1050, 0x1088, 0x1088, 0x5104, 0x2202}},
    {25376, 2, {0x1080, 0x1080, 0x10BC, 0x13C0, 0xFC50, 0x1024, 0x10D4, 0x130C, 0x1800, 0x33FE, 0xD090, 0x1090, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25377, 2, {0x1040, 0x1244, 0x1144, 0x1144, 0xFD48, 0x1040, 0x17FC, 0x1004, 0x1804, 0x3004, 0xD3FC, 0x1004, 0x1004, 0x1004, 0x57FC, 0x2004}},
    {25378, 2, {0x2010, 0x2078, 0x23C0, 0x2040, 0xF840, 0x27FE, 0x20A0, 0x2910, 0x3208, 0xE516, 0x2110, 0x2110, 0x2110, 0x2210, 0xA210, 0x4410}},
    {25379, 2, {0x1080, 0x1080, 0x11F0, 0x1210, 0xF820, 0x13FC, 0x1044, 0x1844, 0x37FE, 0xD044, 0x1044, 0x13FC, 0x1044, 0x1040, 0x5140, 0x2080}},
    {25380, 2, {0x1080, 0x1040, 0x17FE, 0x1208, 0xFD10, 0x10A0, 0x1440, 0x19B0, 0x360E, 0xD110, 0x1110, 0x1110, 0x1110, 0x1210, 0x5210, 0x2410}},
    {25381, 2, {0x1000, 0x13FE, 0x1202, 0x1444, 0xFC40, 0x13FC, 0x1080, 0x14A0, 0x1920, 0x31FC, 0xD020, 0x1020, 0x17FE, 0x1020, 0x5020, 0x2020}},
    {25382, 2, {0x1000, 0x13FC, 0x1004, 0x1004, 0xFDFC, 0x1004, 0x1004, 0x17FC, 0x1808, 0x3008, 0xD7FE, 0x1108, 0x1088, 0x1088, 0x5028, 0x2010}},
    {25383, 2, {0x1000, 0x13DE, 0x1042, 0x1042, 0xFA52, 0x114A, 0x154A, 0x1842, 0x30C6, 0xD14A, 0x1252, 0x1042, 0x1042, 0x1042, 0x514A, 0x2084}},
    {25384, 2, {0x2040, 0x2080, 0x2110, 0x2208, 0xF7FC, 0x2104, 0x2100, 0x23F8, 0x3440, 0xE040, 0x2FFE, 0x2040, 0x20A0, 0x2110, 0xA208, 0x4C06}},
    {25385, 2, {0x1010, 0x1090, 0x1088, 0x1108, 0xFA04, 0x15FA, 0x1108, 0x1908, 0x3108, 0xD1F8, 0x1090, 0x1090, 0x1090, 0x1112, 0x5212, 0x240E}},
    {25386, 2, {0x1000, 0x17DE, 0x1252, 0x1252, 0xFA54, 0x17D4, 0x1258, 0x1A54, 0x3252, 0xDFD2, 0x1252, 0x125A, 0x1254, 0x1450, 0x5550, 0x2890}},
    {25387, 2, {0x2040, 0x2040, 0x2248, 0x2248, 0xF248, 0x2554, 0x28E2, 0x2040, 0x3040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2040, 0xAFFE, 0x4000}},
    {25388, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x13FE, 0x1202, 0x1404, 0x19F8, 0x3010, 0xD020, 0x13FE, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25389, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x1124, 0x15FC, 0x1924, 0x3124, 0xD1FC, 0x1120, 0x10A0, 0x1040, 0x50B0, 0x230E}},
    {25390, 2, {0x1108, 0x1088, 0x1090, 0x13FC, 0xFC24, 0x1024, 0x13FC, 0x1620, 0x1A20, 0x33FE, 0xD062, 0x10A2, 0x112A, 0x1224, 0x5420, 0x2020}},
    {25391, 2, {0x1000, 0x13FC, 0x1200, 0x1200, 0xFEF8, 0x1200, 0x1200, 0x17FC, 0x1AA0, 0x32A4, 0xD2A8, 0x1290, 0x1290, 0x1488, 0x54A4, 0x28C2}},
    {25392, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1000, 0x1400, 0x1BFE, 0x3020, 0xD020, 0x11FC, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25393, 2, {0x1010, 0x1210, 0x1110, 0x1110, 0xFC54, 0x1252, 0x1252, 0x1090, 0x1810, 0x3114, 0xD604, 0x1208, 0x1208, 0x1210, 0x5220, 0x20C0}},
    {25394, 2, {0x2040, 0x1248, 0x4444, 0x2014, 0x860, 0x7380, 0x2000, 0xF0, 0x3F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25395, 2, {0x1000, 0x11FE, 0x1000, 0x1092, 0xFD24, 0x1248, 0x1124, 0x1492, 0x1800, 0x31FE, 0xD020, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25396, 2, {0x1100, 0x1100, 0x11FC, 0x1200, 0xFDF8, 0x1108, 0x1148, 0x1528, 0x1BFE, 0x3108, 0xD248, 0x1228, 0x13FC, 0x1008, 0x5050, 0x2020}},
    {25397, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x1420, 0x1BFE, 0x3088, 0xD088, 0x13FE, 0x1088, 0x1088, 0x1108, 0x5108, 0x2208}},
    {25398, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1100, 0x1100, 0x15FE, 0x1902, 0x317A, 0xD14A, 0x114A, 0x127A, 0x1202, 0x5414, 0x2008}},
    {25399, 2, {0x2100, 0x211E, 0x2112, 0x27D2, 0xF914, 0x2114, 0x27D8, 0x2914, 0x3112, 0xE7D2, 0x2112, 0x211A, 0x2114, 0x2210, 0xA210, 0x4410}},
    {25400, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1504, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1050, 0x1050, 0x1092, 0x5112, 0x220E}},
    {25401, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1000, 0x13FC, 0x1244, 0x1A44, 0x3244, 0xD3FC, 0x1200, 0x1202, 0x1202, 0x51FE, 0x2000}},
    {25402, 2, {0x2000, 0x200C, 0x2770, 0x2110, 0xF910, 0x2210, 0x227C, 0x2F10, 0x3110, 0xE510, 0x2510, 0x227C, 0x2200, 0x2500, 0xA8FE, 0x5000}},
    {25403, 2, {0x1004, 0x100E, 0x13B8, 0x1088, 0xFC88, 0x1128, 0x152E, 0x1BA8, 0x30A8, 0xD2A8, 0x12A8, 0x113E, 0x1100, 0x1280, 0x547E, 0x2800}},
    {25404, 2, {0x1008, 0x103C, 0x13C0, 0x1044, 0xFA24, 0x1128, 0x1500, 0x1840, 0x33FE, 0xD088, 0x1108, 0x1390, 0x1060, 0x1050, 0x5188, 0x2604}},
    {25405, 2, {0x1080, 0x1080, 0x11F8, 0x1208, 0xFC10, 0x13FC, 0x1244, 0x1644, 0x1A44, 0x33FC, 0xD0A0, 0x10A0, 0x1122, 0x1122, 0x521E, 0x2400}},
    {25406, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x1124, 0x1524, 0x1924, 0x32AA, 0xD020, 0x1050, 0x1050, 0x1088, 0x5104, 0x2202}},
    {25407, 2, {0x101C, 0x11E0, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x11FC, 0x1524, 0x1924, 0x31FC, 0xD124, 0x1124, 0x11FC, 0x1020, 0x5020, 0x2020}},
    {25408, 2, {0x1100, 0x11F8, 0x1308, 0x1490, 0xFC60, 0x1198, 0x1646, 0x1040, 0x1BF8, 0x3040, 0xD3F8, 0x1040, 0x17FC, 0x1040, 0x5040, 0x2040}},
    {25409, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x1020, 0x13FE, 0x1800, 0x3000, 0xD1FC, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25410, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0xFDFC, 0x1084, 0x1084, 0x17FE, 0x1800, 0x3000, 0xD1FC, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25411, 2, {0x1000, 0x11FC, 0x1044, 0x1044, 0xFBFE, 0x1044, 0x1444, 0x19FC, 0x3080, 0xD080, 0x11FC, 0x1284, 0x1484, 0x1084, 0x50FC, 0x2084}},
    {25412, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0xFDFE, 0x1020, 0x1120, 0x14B2, 0x18B4, 0x3068, 0xD0A8, 0x1124, 0x1222, 0x1020, 0x50A0, 0x2040}},
    {25413, 2, {0x2000, 0x27F8, 0x2010, 0x21A0, 0xF840, 0x27FC, 0x2444, 0x2C44, 0x37FC, 0xE444, 0x2444, 0x27FC, 0x2444, 0x2444, 0xA454, 0x4408}},
    {25414, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0xFC44, 0x27FC, 0x2444, 0x2CC4, 0x34E4, 0xE554, 0x2644, 0x2444, 0x2444, 0x2404, 0xA7FC, 0x4404}},
    {25415, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x13FE, 0x1450, 0x1850, 0x3154, 0xD152, 0x1152, 0x1252, 0x1090, 0x5090, 0x2130}},
    {25416, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x1020, 0x1420, 0x1BFE, 0x3020, 0xD128, 0x1124, 0x1222, 0x1422, 0x50A0, 0x2040}},
    {25417, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1020, 0x1420, 0x1920, 0x313C, 0xD120, 0x1120, 0x12A0, 0x1260, 0x543E, 0x2800}},
    {25418, 2, {0x1008, 0x103C, 0x13C0, 0x1004, 0xFA44, 0x1128, 0x1000, 0x1BF8, 0x3010, 0xD020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25419, 2, {0x1008, 0x103C, 0x13C0, 0x1044, 0xFA24, 0x1128, 0x1500, 0x1810, 0x33FE, 0xD010, 0x1110, 0x1090, 0x1090, 0x1010, 0x5050, 0x2020}},
    {25420, 2, {0x2004, 0x27C4, 0x2444, 0x2454, 0xFC54, 0x27D4, 0x2214, 0x2A14, 0x37D4, 0xE254, 0x2254, 0x2254, 0x2444, 0x2444, 0xA954, 0x5088}},
    {25421, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1000, 0x1400, 0x19FC, 0x3020, 0xD020, 0x13FE, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25422, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0xFC20, 0x11FC, 0x1104, 0x1504, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1104, 0x1104, 0x5114, 0x2108}},
    {25423, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x15FC, 0x1820, 0x3020, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25424, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1000, 0x13FC, 0x1604, 0x1A04, 0x33FC, 0xD204, 0x1204, 0x13FC, 0x1204, 0x5214, 0x2208}},
    {25425, 2, {0x1000, 0x10FC, 0x1084, 0x1284, 0xFEFC, 0x1284, 0x1284, 0x16FC, 0x1A00, 0x33E0, 0xD020, 0x13FE, 0x1020, 0x1050, 0x5088, 0x2306}},
    {25426, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x1124, 0x1524, 0x19FC, 0x3020, 0xD070, 0x10A8, 0x1124, 0x1222, 0x5020, 0x2020}},
    {25427, 2, {0x2000, 0x27DE, 0x2092, 0x2492, 0xF494, 0x2494, 0x27D8, 0x2094, 0x3192, 0xE192, 0x2292, 0x229A, 0x2494, 0x2890, 0xA290, 0x4110}},
    {25428, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0xFA08, 0x15FE, 0x1122, 0x1922, 0x31FE, 0xD122, 0x1122, 0x11FE, 0x1122, 0x1222, 0x522A, 0x2404}},
    {25429, 2, {0x1048, 0x1044, 0x17FE, 0x1040, 0xFC40, 0x13FC, 0x1244, 0x1644, 0x1BFC, 0x3244, 0xD244, 0x13FC, 0x1244, 0x1244, 0x5254, 0x2208}},
    {25430, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11F8, 0x1000, 0x1400, 0x1BFE, 0x3090, 0xD090, 0x1090, 0x1112, 0x1112, 0x520E, 0x2400}},
    {25431, 2, {0x1010, 0x1010, 0x1090, 0x109E, 0xFC90, 0x1090, 0x13FE, 0x1400, 0x1810, 0x3092, 0xD092, 0x1114, 0x1208, 0x1010, 0x5060, 0x2380}},
    {25432, 2, {0x1040, 0x1040, 0x1088, 0x1104, 0xFBFE, 0x1002, 0x1488, 0x1944, 0x3242, 0xD0F8, 0x1188, 0x1250, 0x1020, 0x1050, 0x5188, 0x2606}},
    {25433, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x1124, 0x15FC, 0x1924, 0x3124, 0xD1FC, 0x1020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {25434, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x15FC, 0x1820, 0x3020, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25435, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x1108, 0x11F8, 0x1400, 0x1800, 0x33FC, 0xD204, 0x1204, 0x1204, 0x1204, 0x53FC, 0x2204}},
    {25436, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1524, 0x19FC, 0x3020, 0xD3FC, 0x1088, 0x1050, 0x1020, 0x50D8, 0x2306}},
    {25437, 2, {0x1104, 0x1084, 0x1088, 0x1010, 0xFDFC, 0x1104, 0x1104, 0x1504, 0x19FC, 0x3050, 0xD050, 0x1090, 0x1092, 0x1112, 0x520E, 0x2400}},
    {25438, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC00, 0x13FE, 0x1202, 0x1444, 0x1840, 0x33FC, 0xD044, 0x1084, 0x1084, 0x1104, 0x5228, 0x2410}},
    {25439, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFCF8, 0x1000, 0x11FC, 0x1504, 0x1924, 0x3124, 0xD124, 0x1124, 0x1050, 0x1048, 0x5084, 0x2304}},
    {25440, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x11FE, 0x1110, 0x1510, 0x1910, 0x3110, 0xD7FE, 0x1000, 0x1090, 0x1108, 0x5204, 0x2402}},
    {25441, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1208, 0x15F6, 0x1000, 0x1888, 0x3048, 0xD248, 0x1150, 0x1110, 0x1020, 0x57FE, 0x2000}},
    {25442, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFA10, 0x15FC, 0x1124, 0x1924, 0x3124, 0xD124, 0x17FE, 0x1050, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25443, 2, {0x2040, 0x2080, 0x23F8, 0x2208, 0xFA88, 0x2248, 0x2218, 0x2A00, 0x33FE, 0xE082, 0x2492, 0x2492, 0x27F2, 0x2002, 0xA014, 0x4008}},
    {25444, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x2200, 0x2218, 0x2AE0, 0x3238, 0xE2E0, 0x223C, 0x23E0, 0x2222, 0x2422, 0xA41E, 0x4800}},
    {25445, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFD04, 0x1100, 0x11DE, 0x1252, 0x1A52, 0x3352, 0xD49A, 0x1094, 0x1110, 0x1112, 0x5212, 0x240E}},
    {25446, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x1202, 0x10F8, 0x1420, 0x1820, 0x31FC, 0xD020, 0x1124, 0x10A4, 0x10A8, 0x53FE, 0x2000}},
    {25447, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x13FE, 0x1488, 0x1924, 0x3222, 0xD0F8, 0x1020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {25448, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x1020, 0x1420, 0x1BFE, 0x3020, 0xD020, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25449, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0xFA04, 0x13FC, 0x1600, 0x1A28, 0x3224, 0xD3FE, 0x1220, 0x1250, 0x1250, 0x1488, 0x5504, 0x2A02}},
    {25450, 2, {0x101C, 0x13E0, 0x1220, 0x13FE, 0xFE20, 0x1292, 0x130A, 0x1206, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25451, 2, {0x2000, 0x27BC, 0x24A4, 0x24A4, 0xFFBC, 0x24A4, 0x24A4, 0x2FBC, 0x3404, 0xE404, 0x2404, 0x2404, 0x2404, 0x2404, 0xA414, 0x4408}},
    {25452, 2, {0x2040, 0x2020, 0x27FE, 0x2400, 0xFC88, 0x2488, 0x2508, 0x2D7E, 0x3708, 0xE548, 0x2528, 0x2528, 0x2908, 0x2908, 0xB128, 0x4110}},
    {25453, 2, {0x1040, 0x1080, 0x13FC, 0x1224, 0xFA24, 0x13FC, 0x1624, 0x1A44, 0x33FC, 0xD090, 0x1110, 0x17FE, 0x1010, 0x1010, 0x5010, 0x2010}},
    {25454, 2, {0x11FE, 0x1102, 0x1102, 0xFF02, 0x11FE, 0x1310, 0x1510, 0x19FF, 0x1110, 0x3110, 0x51FE, 0x9282, 0x1282, 0x1482, 0x5482, 0x28FE}},
    {25455, 2, {0x2004, 0x2004, 0x2FC4, 0x2214, 0xFA14, 0x2494, 0x27D4, 0x2914, 0x3114, 0xE7D4, 0x2114, 0x2114, 0x21C4, 0x2E04, 0xA414, 0x4008}},
    {25456, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x15FC, 0x1820, 0x33FE, 0xD070, 0x10A8, 0x1124, 0x1222, 0x5020, 0x2020}},
    {25457, 2, {0x1000, 0x13FE, 0x1210, 0x1210, 0xFEFE, 0x1210, 0x1210, 0x17FE, 0x1A00, 0x3210, 0xD210, 0x12FE, 0x1210, 0x1410, 0x55FE, 0x2800}},
    {25458, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFC40, 0x1080, 0x17FE, 0x1108, 0x1A04, 0x35FA, 0xD908, 0x1108, 0x1128, 0x1112, 0x5102, 0x20FE}},
    {25459, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFDFE, 0x1110, 0x1510, 0x1910, 0x37FE, 0xD000, 0x1020, 0x1222, 0x1222, 0x1222, 0x53FE, 0x2002}},
    {25460, 2, {0x2090, 0x2090, 0x2108, 0x2148, 0xFA44, 0x2492, 0x2108, 0x2BFC, 0x3044, 0xE020, 0x20A4, 0x2282, 0x228A, 0x228A, 0xA478, 0x4000}},
    {25461, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0xFE94, 0x1294, 0x1294, 0x17FC, 0x1A94, 0x3294, 0xD294, 0x17FE, 0x1000, 0x1090, 0x5108, 0x2204}},
    {25462, 2, {0x1010, 0x1078, 0x13C0, 0x1040, 0xF840, 0x17FC, 0x1248, 0x1A48, 0x37FE, 0xD248, 0x1248, 0x17FC, 0x1040, 0x1040, 0x57FC, 0x2000}},
    {25463, 2, {0x2040, 0x2040, 0x27FE, 0x2040, 0xFBF8, 0x2048, 0x27FE, 0x2848, 0x33F8, 0xE040, 0x2240, 0x227C, 0x2240, 0x2540, 0xA4FE, 0x4800}},
    {25464, 2, {0x1020, 0x1020, 0x11FC, 0x1024, 0xFC24, 0x13FE, 0x1024, 0x1424, 0x19FC, 0x3222, 0xD174, 0x10A8, 0x1124, 0x1222, 0x50A0, 0x2040}},
    {25465, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFCA8, 0x1124, 0x12FA, 0x1420, 0x18A8, 0x3088, 0xD3FE, 0x1088, 0x1088, 0x1088, 0x5108, 0x2208}},
    {25466, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x1104, 0x12FA, 0x1000, 0x1800, 0x33FE, 0xD020, 0x1124, 0x1122, 0x1222, 0x50A0, 0x2040}},
    {25467, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD44, 0x1222, 0x11F8, 0x1408, 0x1850, 0x3020, 0xD0A4, 0x1282, 0x128A, 0x128A, 0x5478, 0x2000}},
    {25468, 2, {0x101C, 0x11E0, 0x1020, 0x1020, 0xFBFE, 0x10A8, 0x1524, 0x1A42, 0x3040, 0xD3FE, 0x1088, 0x1108, 0x10D0, 0x1030, 0x5048, 0x2184}},
    {25469, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFC88, 0x1088, 0x1554, 0x1A22, 0x3000, 0xD020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25470, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11F8, 0x1108, 0x1508, 0x19F8, 0x3100, 0xD100, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25471, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1024, 0x13FE, 0x1424, 0x19FC, 0x3040, 0xD3FE, 0x1084, 0x11C8, 0x1030, 0x50CC, 0x2302}},
    {25472, 2, {0x2010, 0x2090, 0x2710, 0x243E, 0xF422, 0x2444, 0x2790, 0x2510, 0x3510, 0xE510, 0x2528, 0x2528, 0x2528, 0x2944, 0xA944, 0x5082}},
    {25473, 2, {0x1000, 0x10FC, 0x1080, 0x10F8, 0xFC80, 0x10F8, 0x1080, 0x17FE, 0x1940, 0x3124, 0xD128, 0x1110, 0x1108, 0x1144, 0x5182, 0x2100}},
    {25474, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0xFE20, 0x1220, 0x123E, 0x1220, 0x1A20, 0x32FC, 0xD284, 0x1284, 0x1284, 0x1484, 0x54FC, 0x2884}},
    {25475, 2, {0x1000, 0x11FC, 0x1004, 0x10FC, 0xFC04, 0x11FC, 0x1000, 0x13FE, 0x1A22, 0x3020, 0xD1FC, 0x1124, 0x1124, 0x1134, 0x5128, 0x2020}},
    {25476, 2, {0x2040, 0x2040, 0x20A0, 0x2110, 0xF208, 0x2DF6, 0x2000, 0x27FC, 0x34A4, 0xE4A4, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0xA414, 0x4408}},
    {25477, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x13FE, 0x1400, 0x19FC, 0x3104, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x5114, 0x2108}},
    {25478, 2, {0x1000, 0x13FE, 0x1202, 0x128A, 0xFE52, 0x13FE, 0x1222, 0x1622, 0x1AAA, 0x32AA, 0xD2AA, 0x12FA, 0x1202, 0x1202, 0x520A, 0x2204}},
    {25479, 2, {0x2000, 0x27BC, 0x2084, 0x2294, 0xF908, 0x2294, 0x24A4, 0x2840, 0x3000, 0xE7BC, 0x20A4, 0x22A4, 0x2128, 0x2290, 0xA4A8, 0x4846}},
    {25480, 2, {0x101C, 0x13E0, 0x1084, 0x1244, 0xFD48, 0x1110, 0x17FE, 0x1402, 0x1800, 0x33F8, 0xD108, 0x1110, 0x10A0, 0x1040, 0x51B0, 0x260E}},
    {25481, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFDFC, 0x1104, 0x15FC, 0x1904, 0x31FC, 0xD124, 0x1020, 0x13FE, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25482, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFD08, 0x1090, 0x1000, 0x13FE, 0x1800, 0x3000, 0xD1FC, 0x1104, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25483, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0xFE1C, 0x12F0, 0x1290, 0x1690, 0x1AFE, 0x3290, 0xD290, 0x1290, 0x12AA, 0x12CA, 0x54A6, 0x2812}},
    {25484, 2, {0x1110, 0x920, 0x3FFC, 0x2004, 0x4FE8, 0x820, 0xFE0, 0xF0, 0x3F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25485, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x15FC, 0x1800, 0x3112, 0xD1D4, 0x1118, 0x1110, 0x1152, 0x5192, 0x210E}},
    {25486, 2, {0x1040, 0x1040, 0x13FC, 0x10A0, 0xFD10, 0x1208, 0x17FE, 0x1008, 0x1BC8, 0x3248, 0xD248, 0x1248, 0x13C8, 0x1008, 0x5028, 0x2010}},
    {25487, 2, {0x1080, 0x1080, 0x10FE, 0x1102, 0xFE82, 0x10F2, 0x1142, 0x1442, 0x1BFA, 0x3042, 0xD152, 0x1152, 0x11F2, 0x1002, 0x5014, 0x2008}},
    {25488, 2, {0x1040, 0x1040, 0x107C, 0x1084, 0xFD08, 0x1200, 0x1020, 0x15CE, 0x1902, 0x3102, 0xD1CE, 0x1102, 0x1102, 0x1102, 0x51FE, 0x2102}},
    {25489, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x10F8, 0x1088, 0x1488, 0x18F8, 0x3088, 0xD088, 0x13FE, 0x1000, 0x1088, 0x5104, 0x2202}},
    {25490, 2, {0x1090, 0x1090, 0x1090, 0x179E, 0xF890, 0x1090, 0x1090, 0x1B9C, 0x3090, 0xD090, 0x1090, 0x179E, 0x1090, 0x1090, 0x5090, 0x2090}},
    {25491, 2, {0x2100, 0x2100, 0x21DC, 0x2114, 0xF914, 0x2114, 0x27D4, 0x2814, 0x3114, 0xE114, 0x2588, 0x2548, 0x2948, 0x2114, 0xA514, 0x4222}},
    {25492, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25493, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x1088, 0x1544, 0x1A42, 0x30F8, 0xD188, 0x1250, 0x1020, 0x1050, 0x5188, 0x2606}},
    {25494, 2, {0x1080, 0x1040, 0x17FE, 0x1120, 0xFD20, 0x123C, 0x1244, 0x1664, 0x1A94, 0x3348, 0xD248, 0x1230, 0x1220, 0x1250, 0x5288, 0x2306}},
    {25495, 2, {0x1000, 0x17FE, 0x1090, 0x1090, 0xFC90, 0x139C, 0x1204, 0x1604, 0x1A04, 0x339C, 0xD090, 0x1090, 0x1090, 0x1090, 0x57FE, 0x2000}},
    {25496, 2, {0x2000, 0x27FC, 0x2404, 0x2404, 0xF7FC, 0x2420, 0x2524, 0x2524, 0x3524, 0xE5FC, 0x2420, 0x2524, 0x2524, 0x2924, 0xA9FC, 0x5004}},
    {25497, 2, {0x103C, 0x17C0, 0x1244, 0x1128, 0xFC00, 0x13F8, 0x1448, 0x1848, 0x37FE, 0xD048, 0x1048, 0x13F8, 0x1048, 0x1040, 0x5140, 0x2080}},
    {25498, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFBFE, 0x1222, 0x1222, 0x1B32, 0x32AA, 0xD2AA, 0x1376, 0x1266, 0x1222, 0x1222, 0x522A, 0x2204}},
    {25499, 2, {0x1088, 0x1088, 0x13E8, 0x1088, 0xFC88, 0x1088, 0x17EC, 0x140A, 0x188A, 0x3088, 0xD3E8, 0x1088, 0x1088, 0x10E8, 0x5388, 0x2108}},
    {25500, 2, {0x1000, 0x1040, 0x139C, 0x1204, 0xFE04, 0x139C, 0x1204, 0x1604, 0x1BFC, 0x3090, 0xD090, 0x1090, 0x1090, 0x1112, 0x5212, 0x240E}},
    {25501, 2, {0x1014, 0x1012, 0x1010, 0x17FE, 0xF810, 0x1010, 0x17D2, 0x1A52, 0x3252, 0xD254, 0x13D4, 0x1008, 0x10EA, 0x171A, 0x5226, 0x2042}},
    {25502, 2, {0x1020, 0x1122, 0x1122, 0x1224, 0xFC50, 0x1088, 0x1304, 0x1422, 0x1820, 0x3124, 0xD124, 0x1228, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25503, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x1000, 0x13FE, 0x1420, 0x1820, 0x3120, 0xD13C, 0x1120, 0x12A0, 0x1260, 0x543E, 0x2800}},
    {25504, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFC00, 0x11FC, 0x1104, 0x1504, 0x19FC, 0x3020, 0xD128, 0x1124, 0x1222, 0x1422, 0x50A0, 0x2040}},
    {25505, 2, {0x1008, 0x103C, 0x13C0, 0x1044, 0xFA24, 0x1128, 0x1500, 0x1820, 0x33FE, 0xD070, 0x10A8, 0x10A8, 0x1124, 0x1222, 0x5420, 0x2020}},
    {25506, 2, {0x2000, 0x27FC, 0x2404, 0x24A4, 0xF910, 0x2208, 0x2040, 0x2840, 0x37FC, 0xE040, 0x20E0, 0x2150, 0x2248, 0x2C46, 0xA040, 0x4040}},
    {25507, 2, {0x2404, 0x3FA4, 0x4424, 0xFFA4, 0x424, 0x3F84, 0x2494, 0x2588, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25508, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFA52, 0x13DE, 0x1252, 0x1A52, 0x3252, 0xD3DE, 0x1252, 0x1252, 0x1252, 0x1252, 0x5552, 0x28A6}},
    {25509, 2, {0x1080, 0x1040, 0x13FC, 0x1000, 0xFD08, 0x1090, 0x17FE, 0x1040, 0x1840, 0x37FE, 0xD088, 0x1108, 0x1090, 0x1060, 0x5198, 0x2604}},
    {25510, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1080, 0x1480, 0x19FE, 0x324A, 0xD44A, 0x1092, 0x1122, 0x1242, 0x5094, 0x2108}},
    {25511, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1494, 0x1108, 0x1A04, 0x3000, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x57FE, 0x2000}},
    {25512, 2, {0x10A0, 0x1090, 0x1080, 0x11FE, 0xFD10, 0x1310, 0x15FC, 0x1110, 0x1910, 0x31FC, 0xD110, 0x1110, 0x1110, 0x11FE, 0x5100, 0x2100}},
    {25513, 2, {0x1040, 0x1080, 0x17FE, 0x1110, 0xFE48, 0x1446, 0x13F8, 0x1648, 0x1A48, 0x33F8, 0xD248, 0x1248, 0x13F8, 0x1042, 0x5042, 0x203E}},
    {25514, 2, {0x1088, 0x1088, 0x1088, 0x13FE, 0xF888, 0x1088, 0x17FE, 0x1800, 0x31FC, 0xD104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25515, 2, {0x2000, 0x2FC0, 0x24BC, 0x24A4, 0xF4A4, 0x27A4, 0x24A4, 0x24A8, 0x37A8, 0xE4A8, 0x2490, 0x25D0, 0x2EA8, 0x20A8, 0xA0C4, 0x4082}},
    {25516, 2, {0x1100, 0x1100, 0x11FC, 0x1204, 0xFC44, 0x1154, 0x14E4, 0x1844, 0x33FC, 0xD044, 0x10E4, 0x1154, 0x1244, 0x1044, 0x5014, 0x2008}},
    {25517, 2, {0x2000, 0x23FC, 0x2040, 0x2040, 0xFBFE, 0x20A0, 0x2110, 0x2A08, 0x3446, 0xE040, 0x2048, 0x2264, 0x2252, 0x2452, 0xA140, 0x4080}},
    {25518, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0xFE04, 0x13FC, 0x1200, 0x12FC, 0x1A84, 0x32FC, 0xD284, 0x12FC, 0x1484, 0x1484, 0x5894, 0x2088}},
    {25519, 2, {0x1020, 0x1120, 0x113C, 0x1120, 0xF920, 0x17FE, 0x1000, 0x1BFC, 0x3204, 0xD3FC, 0x1204, 0x13FC, 0x1204, 0x1204, 0x5214, 0x2208}},
    {25520, 2, {0x1000, 0x3A86, 0xE2B8, 0x2488, 0x2448, 0xF85E, 0x2028, 0x2788, 0x2288, 0xFABE, 0x2288, 0x2488, 0x2488, 0x2A88, 0x5128, 0x8010}},
    {25521, 2, {0xF0, 0x3F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x504, 0x7A78, 0x1010, 0x7CFC, 0x1010, 0xFEFE, 0x1010, 0x5050, 0x2020}},
    {25522, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1080, 0x15FE, 0x1A02, 0x3112, 0xD1EA, 0x110A, 0x10FA, 0x1002, 0x5014, 0x2008}},
    {25523, 2, {0x2040, 0x207C, 0x2040, 0x23FE, 0xFA42, 0x2278, 0x23C4, 0x2A3C, 0x3240, 0xE240, 0x23FC, 0x2244, 0x2484, 0x2484, 0xA914, 0x4208}},
    {25524, 2, {0x2000, 0x23FE, 0x2202, 0x2202, 0xFBFE, 0x2222, 0x2222, 0x2AFA, 0x3222, 0xE232, 0x222A, 0x23FE, 0x2202, 0x2202, 0xA3FE, 0x4202}},
    {25525, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x1000, 0x17DE, 0x1A52, 0x3252, 0xD252, 0x13DA, 0x1254, 0x1010, 0x5010, 0x2010}},
    {25526, 2, {0x1040, 0x1020, 0x13FE, 0x1220, 0xFEFC, 0x1224, 0x1224, 0x17FE, 0x1A24, 0x3224, 0xD2FC, 0x1220, 0x1250, 0x1450, 0x5488, 0x2906}},
    {25527, 2, {0x2440, 0x225E, 0x2292, 0x2012, 0xFFD4, 0x2114, 0x2118, 0x2914, 0x37F2, 0xE112, 0x2112, 0x211A, 0x2294, 0x2250, 0xA450, 0x4810}},
    {25528, 2, {0x1104, 0x1088, 0x1050, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x1524, 0x1924, 0x31FC, 0xD020, 0x1020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {25529, 2, {0x1000, 0x11FC, 0x1008, 0x1010, 0xF820, 0x13FE, 0x1420, 0x18A0, 0x3040, 0xD1FC, 0x1154, 0x1154, 0x1154, 0x1154, 0x57FE, 0x2000}},
    {25530, 2, {0x1080, 0x1090, 0x1108, 0x13FC, 0xFC40, 0x17FE, 0x1110, 0x1248, 0x1C86, 0x3310, 0xD020, 0x10C4, 0x1308, 0x1030, 0x50C0, 0x2700}},
    {25531, 2, {0x1000, 0x11FC, 0x1084, 0x1088, 0xFC50, 0x1020, 0x14D8, 0x1B26, 0x3020, 0xD1FC, 0x1124, 0x1124, 0x11FC, 0x1022, 0x53FE, 0x2002}},
    {25532, 2, {0x11FC, 0x1124, 0x1124, 0x13FE, 0xFD24, 0x1124, 0x11FC, 0x1400, 0x19FC, 0x3104, 0xD124, 0x1124, 0x1124, 0x1058, 0x5084, 0x2302}},
    {25533, 2, {0x1108, 0x1090, 0x1000, 0x13FC, 0xFC90, 0x1090, 0x1092, 0x1492, 0x1A94, 0x3298, 0xD090, 0x1090, 0x1090, 0x1090, 0x57FE, 0x2000}},
    {25534, 2, {0x2100, 0x21F8, 0x2208, 0x23F0, 0xF810, 0x27FE, 0x2080, 0x2944, 0x3668, 0xE0B0, 0x2128, 0x2668, 0x20A4, 0x2122, 0xA6A0, 0x4040}},
    {25535, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCA8, 0x1020, 0x13FE, 0x1440, 0x1880, 0x30FC, 0xD184, 0x1284, 0x1084, 0x1084, 0x50FC, 0x2084}},
    {25536, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x11AC, 0x1574, 0x1924, 0x31FC, 0xD020, 0x1070, 0x10A8, 0x1124, 0x5222, 0x2020}},
    {25537, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFC20, 0x11FC, 0x1104, 0x15FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1000, 0x1088, 0x5104, 0x2202}},
    {25538, 2, {0x2108, 0x2088, 0x2090, 0x27FE, 0xF890, 0x2090, 0x23FC, 0x2A94, 0x3294, 0xE30C, 0x2204, 0x23FC, 0x2204, 0x2204, 0xA3FC, 0x4204}},
    {25539, 2, {0x1108, 0x1088, 0x1090, 0x17FE, 0xF800, 0x13C4, 0x1254, 0x1A54, 0x33D4, 0xD254, 0x1254, 0x13D4, 0x1254, 0x1244, 0x5254, 0x22C8}},
    {25540, 2, {0x2040, 0x20A0, 0x2110, 0x2208, 0xFDF6, 0x2000, 0x23C4, 0x2A54, 0x3254, 0xE3D4, 0x2254, 0x2254, 0x23D4, 0x2244, 0xA254, 0x42C8}},
    {25541, 2, {0xFDFE, 0x2048, 0x4048, 0x7DFE, 0xC448, 0x4488, 0x7D08, 0xE8, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25542, 2, {0x1010, 0x13D4, 0x1058, 0x1252, 0xFD8C, 0x1088, 0x1104, 0x12FA, 0x1820, 0x3020, 0xD3FE, 0x1020, 0x1050, 0x1088, 0x5104, 0x2202}},
    {25543, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x13FE, 0x128A, 0x1652, 0x1AFA, 0x3222, 0xD222, 0x12FA, 0x1222, 0x1222, 0x522A, 0x2204}},
    {25544, 2, {0x1100, 0x1100, 0x11FE, 0x1202, 0xFC82, 0x17F2, 0x1002, 0x13F2, 0x1802, 0x33F2, 0xD002, 0x13F2, 0x1212, 0x13F2, 0x5014, 0x2008}},
    {25545, 2, {0x2000, 0x21FC, 0x2048, 0x2030, 0xFBFE, 0x2052, 0x2094, 0x2910, 0x3230, 0xE020, 0x23FE, 0x2070, 0x20A8, 0x2124, 0xA622, 0x4020}},
    {25546, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1400, 0x1BFE, 0x3222, 0xD222, 0x13FE, 0x1222, 0x1222, 0x53FE, 0x2202}},
    {25547, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x15FC, 0x1800, 0x33FE, 0xD120, 0x1122, 0x1114, 0x1148, 0x5184, 0x2102}},
    {25548, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x1524, 0x19FC, 0x3000, 0xD040, 0x1024, 0x12A2, 0x128A, 0x1488, 0x5078, 0x2000}},
    {25549, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0xFBF8, 0x2080, 0x27FC, 0x2910, 0x3208, 0xE5F6, 0x2040, 0x23F8, 0x2040, 0x20A0, 0xA110, 0x4608}},
    {25550, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11FC, 0x1400, 0x19FC, 0x3104, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x1000, 0x53FE, 0x2000}},
    {25551, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x1000, 0x11FC, 0x1524, 0x1924, 0x3124, 0xD1FC, 0x1124, 0x1124, 0x1124, 0x51FC, 0x2104}},
    {25552, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x15FC, 0x1800, 0x33FE, 0xD020, 0x1120, 0x113C, 0x1120, 0x52A0, 0x247E}},
    {25553, 2, {0x1040, 0x119C, 0x1104, 0x1104, 0xFDDC, 0x1104, 0x1104, 0x15FC, 0x1800, 0x3000, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25554, 2, {0x2008, 0x203C, 0x27C0, 0x2040, 0xF840, 0x2FFE, 0x2040, 0x2940, 0x365C, 0xE444, 0x2444, 0x275C, 0x2444, 0x2444, 0xA7FC, 0x4404}},
    {25555, 2, {0x2210, 0x2110, 0x2110, 0x27BE, 0xFA28, 0x2248, 0x23AA, 0x2AAE, 0x32BA, 0xE2EA, 0x22AE, 0x22A8, 0x24AA, 0x24A2, 0xA9A2, 0x501E}},
    {25556, 2, {0x1080, 0x1080, 0x10FC, 0x1154, 0xFAD4, 0x10B4, 0x152C, 0x1A44, 0x3094, 0xD108, 0x1040, 0x10A4, 0x12AA, 0x128A, 0x5478, 0x2000}},
    {25557, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x1088, 0x10F8, 0x1488, 0x1888, 0x33FE, 0xD100, 0x1148, 0x1184, 0x1100, 0x51FE, 0x2000}},
    {25558, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1000, 0x17FE, 0x1508, 0x19F8, 0x3108, 0xD1F8, 0x1108, 0x113E, 0x17C8, 0x5008, 0x2008}},
    {25559, 2, {0x101C, 0x13E0, 0x1220, 0x1220, 0xFBFE, 0x1220, 0x12FC, 0x1A84, 0x3284, 0xD2FC, 0x1284, 0x12FC, 0x1484, 0x1484, 0x58FC, 0x2084}},
    {25560, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x1400, 0x1BFE, 0x3020, 0xD020, 0x11FC, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25561, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFE02, 0x13FE, 0x1200, 0x1600, 0x1BFE, 0x3352, 0xD552, 0x15FE, 0x1552, 0x1552, 0x5952, 0x2106}},
    {25562, 2, {0x1000, 0x11F8, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x1400, 0x1BFE, 0x3100, 0xD1FC, 0x1254, 0x1494, 0x1124, 0x1244, 0x50A8, 0x2110}},
    {25563, 2, {0x1080, 0x1080, 0x11F8, 0x1208, 0xFC10, 0x13FE, 0x1202, 0x1A52, 0x328A, 0xD222, 0x1020, 0x17FE, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25564, 2, {0x1020, 0x1050, 0x1088, 0x1104, 0xFEFA, 0x1000, 0x15FC, 0x1904, 0x3104, 0xD1FC, 0x1088, 0x1088, 0x13FE, 0x1088, 0x5108, 0x2208}},
    {25565, 2, {0x1088, 0x1088, 0x11EC, 0x112A, 0xFA28, 0x1548, 0x1088, 0x197E, 0x3600, 0xD1FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x5104, 0x21FC}},
    {25566, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFD04, 0x1088, 0x17FE, 0x1000, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25567, 2, {0x1000, 0x13FE, 0x1022, 0x1120, 0xFD3C, 0x1120, 0x12FE, 0x1400, 0x19FC, 0x3104, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x5114, 0x2108}},
    {25568, 2, {0x1000, 0x13FE, 0x1200, 0x12FC, 0xFA84, 0x12FC, 0x1684, 0x1AFC, 0x3220, 0xD3FE, 0x1248, 0x12C8, 0x1230, 0x1248, 0x5284, 0x23FE}},
    {25569, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x2200, 0x23FC, 0x2A40, 0x3288, 0xE3FC, 0x2224, 0x2220, 0x25FC, 0x2420, 0xA820, 0x53FE}},
    {25570, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC80, 0x10F8, 0x1508, 0x1A90, 0x3060, 0xD198, 0x1606, 0x11F8, 0x1108, 0x1108, 0x51F8, 0x2108}},
    {25571, 2, {0x2040, 0x2444, 0x2444, 0x2444, 0xF7FC, 0x2000, 0x2FFE, 0x2040, 0x3080, 0xE7FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0xA4A4, 0x440C}},
    {25572, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFA52, 0x13D2, 0x1252, 0x1A52, 0x3252, 0xD3D2, 0x121A, 0x1294, 0x1250, 0x12B0, 0x5310, 0x2010}},
    {25573, 2, {0x1080, 0x1040, 0x13FC, 0x1000, 0xFD08, 0x1090, 0x17FE, 0x1442, 0x1844, 0x33FC, 0xD244, 0x1244, 0x1254, 0x1248, 0x5040, 0x2040}},
    {25574, 2, {0x1082, 0x1082, 0x17F2, 0x1082, 0xFC8A, 0x13EA, 0x16AA, 0x1AAA, 0x33EA, 0xD08A, 0x11CA, 0x12AA, 0x14A2, 0x1082, 0x508A, 0x2084}},
    {25575, 2, {0x404, 0x7FC4, 0x424, 0x3FA4, 0x24A4, 0x3FA4, 0x1504, 0x2494, 0x4428, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25576, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x11FC, 0x1400, 0x1BFE, 0x3202, 0xD1FC, 0x1020, 0x1020, 0x1020, 0x50A0, 0x2040}},
    {25577, 2, {0x1110, 0x1112, 0x11D4, 0x1118, 0xFD52, 0x1192, 0x112E, 0x1440, 0x19FC, 0x3104, 0xD104, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25578, 2, {0x2088, 0x21C8, 0x2708, 0x2108, 0xF92A, 0x212A, 0x27AC, 0x2948, 0x3308, 0xE388, 0x2554, 0x2514, 0x2914, 0x2124, 0xA124, 0x4142}},
    {25579, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x9E4, 0x1F00, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x500, 0x200}},
    {25580, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC88, 0x1104, 0x1222, 0x1028, 0x1824, 0x33FE, 0xD020, 0x1050, 0x1050, 0x1088, 0x5104, 0x2202}},
    {25581, 2, {0x1000, 0x13F8, 0x1208, 0x13F8, 0xFE08, 0x13F8, 0x1100, 0x13FC, 0x1C44, 0x3244, 0xD2A4, 0x1204, 0x13F4, 0x1004, 0x5028, 0x2010}},
    {25582, 2, {0x1000, 0x13FE, 0x1222, 0x1020, 0xFDFE, 0x1020, 0x11FC, 0x1524, 0x19FC, 0x3124, 0xD1FC, 0x1020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {25583, 2, {0x2200, 0x22FE, 0x2220, 0x2220, 0xFB20, 0x22BC, 0x2644, 0x2A54, 0x324C, 0xE284, 0x22A4, 0x2218, 0x2208, 0x2208, 0xA2FE, 0x4200}},
    {25584, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFBFE, 0x1020, 0x15FC, 0x1924, 0x31FC, 0xD124, 0x11FC, 0x1020, 0x11FC, 0x1020, 0x53FE, 0x2000}},
    {25585, 2, {0x4904, 0x2A24, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x7F24, 0x4104, 0x432C, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25586, 2, {0x1048, 0x1148, 0x1148, 0x13FE, 0xFD48, 0x1148, 0x1178, 0x1500, 0x19FE, 0x3020, 0xD3FE, 0x1070, 0x10A8, 0x1124, 0x5622, 0x2020}},
    {25587, 2, {0x1080, 0x109E, 0x13EA, 0x108A, 0xFDCA, 0x108A, 0x13EA, 0x1492, 0x18A6, 0x3020, 0xD3FE, 0x1020, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25588, 2, {0x203C, 0x27C0, 0x2244, 0x2128, 0xF800, 0x23FC, 0x2080, 0x2880, 0x37FE, 0xE100, 0x21F8, 0x2288, 0x2250, 0x2420, 0xA8D8, 0x4306}},
    {25589, 2, {0x2010, 0x2010, 0x277C, 0x2114, 0xF9FE, 0x2214, 0x227C, 0x2F10, 0x317C, 0xE510, 0x25FE, 0x2210, 0x2210, 0x2500, 0xA8FE, 0x4000}},
    {25590, 2, {0x2000, 0x2FDE, 0x2492, 0x2492, 0xF494, 0x2794, 0x2498, 0x2494, 0x3792, 0xE492, 0x2492, 0x25DA, 0x2E94, 0x2090, 0xA090, 0x4090}},
    {25591, 2, {0x101C, 0x11E0, 0x1020, 0x1020, 0xFDFE, 0x10A0, 0x112C, 0x1524, 0x1924, 0x31AC, 0xD124, 0x1124, 0x11FC, 0x1124, 0x5020, 0x2020}},
    {25592, 2, {0x1020, 0x1020, 0x13FE, 0x1070, 0xFCA8, 0x1124, 0x1222, 0x11FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x1000, 0x57FE, 0x2000}},
    {25593, 2, {0x1050, 0x1052, 0x13DC, 0x1050, 0xFCD2, 0x134E, 0x1000, 0x15FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x1104, 0x5114, 0x2108}},
    {25594, 2, {0x1008, 0x103C, 0x13C0, 0x1004, 0xFA44, 0x1128, 0x15FC, 0x1820, 0x3020, 0xD3FE, 0x1020, 0x1124, 0x1124, 0x1124, 0x51FC, 0x2004}},
    {25595, 2, {0x2014, 0x2012, 0x27FE, 0x2410, 0xFC10, 0x27F0, 0x2492, 0x2C92, 0x37D2, 0xE554, 0x2554, 0x2488, 0x254A, 0x2A1A, 0xA826, 0x5042}},
    {25596, 2, {0x1000, 0x13FE, 0x1040, 0x1080, 0xFDFC, 0x1284, 0x1084, 0x14FC, 0x1820, 0x3022, 0xD3B4, 0x10A8, 0x1128, 0x1224, 0x54A2, 0x2040}},
    {25597, 2, {0x1090, 0x1290, 0x129E, 0x12A8, 0xFAC4, 0x1080, 0x15FC, 0x1904, 0x3124, 0xD124, 0x1124, 0x1154, 0x1050, 0x1090, 0x5112, 0x260E}},
    {25598, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1108, 0x1108, 0x15F8, 0x1800, 0x33FC, 0xD294, 0x1294, 0x1294, 0x1294, 0x57FE, 0x2000}},
    {25599, 2, {0x2210, 0x2210, 0x2210, 0x23BE, 0xF422, 0x2444, 0x2790, 0x2A10, 0x3210, 0xEF90, 0x2228, 0x2228, 0x22A8, 0x2344, 0xA244, 0x4082}},
    {25600, 2, {0x2100, 0x21F0, 0x2210, 0x2420, 0xFBFC, 0x2244, 0x2244, 0x2BFC, 0x30A0, 0xE122, 0x261E, 0x20C0, 0x2030, 0x2180, 0xA060, 0x4010}},
    {25601, 2, {0x2200, 0x217C, 0x2504, 0x2484, 0xF4F4, 0x2514, 0x26A4, 0x2444, 0x34A4, 0xE71C, 0x25F4, 0x2514, 0x2514, 0x25F4, 0xA514, 0x4408}},
    {25602, 2, {0x1020, 0x1124, 0x10A8, 0x1020, 0xFBFE, 0x10A8, 0x1124, 0x1A02, 0x3040, 0xD7FE, 0x1088, 0x1108, 0x1190, 0x1060, 0x5198, 0x2604}},
    {25603, 2, {0x1108, 0x1088, 0x1090, 0x13FC, 0xFE04, 0x1204, 0x1204, 0x13FC, 0x1840, 0x3020, 0xD0A4, 0x1282, 0x128A, 0x128A, 0x5478, 0x2000}},
    {25604, 2, {0x2100, 0x213C, 0x2100, 0x2100, 0xF9BC, 0x2564, 0x2524, 0x2D3C, 0x3524, 0xE124, 0x2124, 0x213C, 0x2124, 0x2100, 0xA17E, 0x4100}},
    {25605, 2, {0x2484, 0x2244, 0x2248, 0x27FE, 0xFC02, 0x2804, 0x23F8, 0x2A08, 0x3248, 0xE248, 0x2248, 0x22A8, 0x20A0, 0x2122, 0xA222, 0x4C1E}},
    {25606, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFDFC, 0x1088, 0x13FE, 0x1420, 0x19FC, 0x3124, 0xD1FC, 0x1124, 0x13FE, 0x1104, 0x5114, 0x2108}},
    {25607, 2, {0x2040, 0x20A0, 0x2118, 0x2686, 0xF840, 0x23F0, 0x2010, 0x28A0, 0x3040, 0xE7FE, 0x2088, 0x2150, 0x2320, 0x2510, 0xA948, 0x4186}},
    {25608, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC50, 0x1088, 0x1124, 0x1450, 0x1888, 0x3104, 0xD2FA, 0x1088, 0x1088, 0x1088, 0x50F8, 0x2088}},
    {25609, 2, {0x1040, 0x1040, 0x13FE, 0x12A2, 0xFC90, 0x11FE, 0x1510, 0x1B10, 0x35FE, 0xD110, 0x1110, 0x11FE, 0x1110, 0x1110, 0x51FE, 0x2100}},
    {25610, 2, {0x1080, 0x1080, 0x11FC, 0x1244, 0xFD54, 0x11F4, 0x1084, 0x1528, 0x1A90, 0x31FC, 0xD244, 0x1554, 0x11F4, 0x1084, 0x5128, 0x2210}},
    {25611, 2, {0x1004, 0x101E, 0x13F0, 0x121E, 0xFE10, 0x12FE, 0x1292, 0x1698, 0x1AF2, 0x328E, 0xD280, 0x12B8, 0x12A8, 0x14AA, 0x554A, 0x2A86}},
    {25612, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x2248, 0x2248, 0x2AFC, 0x3248, 0xE248, 0x23FE, 0x22A4, 0x22A8, 0x2490, 0xA4C8, 0x4886}},
    {25613, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1000, 0x13FC, 0x1604, 0x1BFC, 0x3204, 0xD3FC, 0x1204, 0x13FC, 0x1090, 0x5108, 0x2204}},
    {25614, 2, {0x2006, 0x2F78, 0x2110, 0x2120, 0xFA44, 0x2278, 0x2310, 0x2A24, 0x367E, 0xEA12, 0x2210, 0x2254, 0x2252, 0x2292, 0xAA50, 0x4420}},
    {25615, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x11FC, 0x1524, 0x19FC, 0x3124, 0xD008, 0x13FE, 0x1088, 0x1048, 0x5048, 0x2018}},
    {25616, 2, {0x2080, 0x2040, 0x27FC, 0x2080, 0xF908, 0x23F0, 0x2060, 0x2984, 0x37FE, 0xE002, 0x27FC, 0x2444, 0x27FC, 0x2444, 0xA7FC, 0x4404}},
    {25617, 2, {0x1108, 0x1108, 0x17FE, 0x1108, 0xF800, 0x17FE, 0x1108, 0x19F8, 0x3108, 0xD1F8, 0x1108, 0x113E, 0x17C8, 0x1008, 0x5008, 0x2008}},
    {25618, 2, {0x1040, 0x1020, 0x13FC, 0x1108, 0xFC90, 0x13FE, 0x1202, 0x1444, 0x1820, 0x33FC, 0xD080, 0x10F8, 0x1088, 0x1108, 0x5128, 0x2210}},
    {25619, 2, {0x1108, 0x1088, 0x1090, 0x13FC, 0xFC40, 0x11F8, 0x1040, 0x17FE, 0x1880, 0x3100, 0xD1FC, 0x1220, 0x1420, 0x1820, 0x53FE, 0x2000}},
    {25620, 2, {0x2000, 0x23F8, 0x2148, 0x2110, 0xF8A0, 0x2040, 0x21B0, 0x2E4E, 0x3040, 0xE3F8, 0x2248, 0x2248, 0x23F8, 0x2044, 0xA7FC, 0x4004}},
    {25621, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x1080, 0x1504, 0x1BFE, 0x3002, 0xD3FC, 0x1294, 0x1294, 0x1294, 0x57FE, 0x2000}},
    {25622, 2, {0x1040, 0x1080, 0x11FC, 0x1248, 0xFD30, 0x10C0, 0x1300, 0x11FC, 0x1A20, 0x3020, 0xD3FE, 0x1020, 0x1124, 0x1124, 0x51FC, 0x2004}},
    {25623, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x1500, 0x19FE, 0x3100, 0xD1FE, 0x1042, 0x124A, 0x13FA, 0x5014, 0x2008}},
    {25624, 2, {0x1040, 0x1044, 0x11F8, 0x1050, 0xFBFE, 0x1040, 0x14F8, 0x1982, 0x327E, 0xD400, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x51FC, 0x2104}},
    {25625, 2, {0x1000, 0x13FE, 0x1200, 0x12FC, 0xFE00, 0x13FE, 0x12A8, 0x1690, 0x1AC8, 0x3286, 0xD208, 0x13FE, 0x1288, 0x1448, 0x5408, 0x2818}},
    {25626, 2, {0x1040, 0x11FC, 0x1044, 0x1094, 0xFD08, 0x13DE, 0x114A, 0x154A, 0x1A94, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x11FC, 0x5104, 0x210C}},
    {25627, 2, {0x1110, 0x10A0, 0x1000, 0x17FE, 0xF8A0, 0x13F8, 0x10A8, 0x1FFE, 0x30A8, 0xD3F8, 0x10A0, 0x11B0, 0x12A8, 0x14A6, 0x50A0, 0x20A0}},
    {25628, 2, {0x10A0, 0x112C, 0x1124, 0x1124, 0xFDAC, 0x1124, 0x1124, 0x15FC, 0x1820, 0x31FC, 0xD088, 0x1050, 0x1020, 0x1050, 0x5088, 0x2306}},
    {25629, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x1504, 0x19FC, 0x3028, 0xD024, 0x13FE, 0x1050, 0x1088, 0x5104, 0x2202}},
    {25630, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x11FC, 0x1400, 0x1BFE, 0x3202, 0xD2FA, 0x128A, 0x12FA, 0x1202, 0x520A, 0x2204}},
    {25631, 2, {0x2100, 0x2100, 0x213E, 0x27C8, 0xF908, 0x27C8, 0x2548, 0x2FC8, 0x357E, 0xE7C8, 0x2108, 0x2FE8, 0x2108, 0x2108, 0xA108, 0x4108}},
    {25632, 2, {0x2440, 0x225E, 0x2292, 0x2012, 0xF7D2, 0x211E, 0x2552, 0x2552, 0x3552, 0xE7DE, 0x2152, 0x2112, 0x2212, 0x2222, 0xA42A, 0x4844}},
    {25633, 2, {0x23F0, 0x2110, 0x20E0, 0x2318, 0xF800, 0x27BC, 0x24A4, 0x2B18, 0x34A4, 0xE040, 0x27FC, 0x20E0, 0x2150, 0x2248, 0xAC46, 0x4040}},
    {25634, 2, {0x1000, 0x13FC, 0x1090, 0x1294, 0xFD98, 0x1090, 0x17FE, 0x1800, 0x31F8, 0xD108, 0x1108, 0x11F8, 0x1108, 0x1108, 0x51F8, 0x2108}},
    {25635, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0xFC10, 0x2410, 0x27F2, 0x2C92, 0x3492, 0xE5B4, 0x26D4, 0x2488, 0x254A, 0x293A, 0xAA26, 0x5042}},
    {25636, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFBFE, 0x1000, 0x1488, 0x1904, 0x3202, 0xD1FC, 0x1154, 0x1154, 0x1154, 0x1154, 0x57FE, 0x2000}},
    {25637, 2, {0x1010, 0x1220, 0x117C, 0x1144, 0xFC44, 0x107C, 0x1340, 0x1540, 0x197C, 0x3144, 0xD144, 0x117C, 0x1144, 0x1280, 0x547E, 0x2000}},
    {25638, 2, {0x2000, 0x27BC, 0x2084, 0x2084, 0xF7BC, 0x2420, 0x2420, 0x27BC, 0x3084, 0xE4A4, 0x2294, 0x2294, 0x24A4, 0x2084, 0xA528, 0x4210}},
    {25639, 2, {0x2080, 0x2040, 0x27FE, 0x2402, 0xFC02, 0x27FE, 0x2400, 0x2FDE, 0x3442, 0xE652, 0x254A, 0x24C6, 0x2B5A, 0x2842, 0xB14A, 0x4084}},
    {25640, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1000, 0x1400, 0x1BDE, 0x3042, 0xD252, 0x114A, 0x1252, 0x1042, 0x514A, 0x2084}},
    {25641, 2, {0x2208, 0x2208, 0x23BE, 0xF488, 0x2AA8, 0x213E, 0x2208, 0x3448, 0xE848, 0x27FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0xA040, 0x4040}},
    {25642, 2, {0x2040, 0x2020, 0x27FE, 0x2420, 0xFDFC, 0x2424, 0x27FE, 0x2C24, 0x35FC, 0xE420, 0x25FC, 0x2504, 0x2504, 0x2904, 0xA9FC, 0x5104}},
    {25643, 2, {0x878, 0x3E48, 0x228E, 0x2A00, 0x7EF8, 0x2248, 0x2A30, 0x46CC, 0x8020, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25644, 2, {0x2100, 0x223C, 0x27A4, 0x24A4, 0xF6A4, 0x25A4, 0x24C2, 0x2F80, 0x34BC, 0xE4A4, 0x26A4, 0x25A4, 0x2494, 0x2488, 0xA494, 0x49A2}},
    {25645, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF840, 0x20A0, 0x2110, 0x2A08, 0x35F6, 0xE000, 0x2000, 0x23F8, 0x2208, 0x2208, 0xA3F8, 0x4208}},
    {25646, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFBFE, 0x1252, 0x1652, 0x1BFE, 0x3020, 0xD020, 0x13FE, 0x1070, 0x10A8, 0x1124, 0x5622, 0x2020}},
    {25647, 2, {0x103C, 0x13C0, 0x1044, 0x1224, 0xF928, 0x1100, 0x1440, 0x1B9C, 0x3204, 0xD204, 0x139C, 0x1204, 0x1204, 0x1204, 0x53FC, 0x2204}},
    {25648, 2, {0x1000, 0x11FC, 0x1104, 0x11E4, 0xFD24, 0x13FE, 0x1202, 0x11FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x1104, 0x5114, 0x2108}},
    {25649, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFBFC, 0x1240, 0x1220, 0x1BFC, 0x3288, 0xD250, 0x13FE, 0x1220, 0x15FC, 0x1420, 0x5820, 0x2020}},
    {25650, 2, {0x2080, 0x2040, 0x27FC, 0x2404, 0xF910, 0x2208, 0x203C, 0x2BD0, 0x3290, 0xE290, 0x2290, 0x2288, 0x2288, 0x24A4, 0xA4D4, 0x4892}},
    {25651, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC24, 0x11FC, 0x1020, 0x15FC, 0x1820, 0x33FE, 0xD020, 0x11FC, 0x1104, 0x1104, 0x51FC, 0x2104}},
    {25652, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x17D0, 0x2108, 0xDFF6, 0x100, 0x7FF8, 0x100, 0x300}},
    {25653, 2, {0x2000, 0x23F8, 0x2248, 0x22A8, 0xFB18, 0x2208, 0x23F8, 0x2800, 0x3000, 0xE7FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0xAFFE, 0x4000}},
    {25654, 2, {0x1020, 0x1020, 0x1050, 0x10C8, 0xFD24, 0x13FE, 0x1504, 0x11FC, 0x1904, 0x31FC, 0xD100, 0x11FC, 0x1284, 0x1284, 0x54FC, 0x2084}},
    {25655, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1104, 0x11FC, 0x1504, 0x19FC, 0x3104, 0xD1FC, 0x1104, 0x13FE, 0x1088, 0x5104, 0x2202}},
    {25656, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x13FE, 0x1488, 0x1934, 0x32E2, 0xD020, 0x11FC, 0x1070, 0x10AC, 0x5322, 0x2020}},
    {25657, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x1400, 0x1BFE, 0x328A, 0xD252, 0x13FE, 0x1222, 0x1222, 0x522A, 0x2204}},
    {25658, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x11FC, 0x1520, 0x19FE, 0x3100, 0xD3FC, 0x1088, 0x109E, 0x1102, 0x520A, 0x2404}},
    {25659, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x40C, 0x78F0, 0x1010, 0x7CFC, 0x1010, 0xFEFE, 0x1010, 0x3030}},
    {25660, 2, {0x1040, 0x1248, 0x1150, 0x13F8, 0xF880, 0x17FC, 0x1110, 0x1A28, 0x35C4, 0xD842, 0x13F8, 0x1040, 0x17FC, 0x1040, 0x5140, 0x2080}},
    {25661, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCA8, 0x1050, 0x1088, 0x1524, 0x1A22, 0x31FC, 0xD020, 0x10A8, 0x1124, 0x1222, 0x50A0, 0x2040}},
    {25662, 2, {0x1040, 0x1020, 0x13FE, 0x128A, 0xFD04, 0x1080, 0x10FE, 0x1540, 0x1A40, 0x307C, 0xD040, 0x1040, 0x107E, 0x1040, 0x5040, 0x2040}},
    {25663, 2, {0x1100, 0x3906, 0xE2B8, 0x2288, 0x2448, 0xF85E, 0x2388, 0x2008, 0x2008, 0xFBBE, 0x2288, 0x2288, 0x2288, 0x2388, 0x42A8, 0x8010}},
    {25664, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x1100, 0x15FE, 0x1900, 0x31FE, 0xD002, 0x12AA, 0x12AA, 0x1402, 0x5014, 0x2008}},
    {25665, 2, {0x2000, 0x27FC, 0x2444, 0x2444, 0xF7FC, 0x2444, 0x24A4, 0x2514, 0x3404, 0xE7FC, 0x2040, 0x2024, 0x2522, 0x250A, 0xA908, 0x40F8}},
    {25666, 2, {0x1000, 0x13FC, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x110E, 0x17F8, 0x1808, 0x3008, 0xD204, 0x1108, 0x1090, 0x1108, 0x5204, 0x2402}},
    {25667, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFBFE, 0x1000, 0x15FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x5104, 0x2202}},
    {25668, 2, {0x1000, 0x13FC, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x110E, 0x17F8, 0x1808, 0x37BC, 0xD0A4, 0x12A4, 0x1128, 0x1290, 0x54A8, 0x2846}},
    {25669, 2, {0x2040, 0x207C, 0x2040, 0x27FE, 0xF442, 0x2470, 0x27C4, 0x2444, 0x343C, 0xE420, 0x2410, 0x2454, 0x2542, 0x2942, 0xA94A, 0x5238}},
    {25670, 2, {0x1000, 0x13FE, 0x1252, 0x1252, 0xFBFE, 0x1020, 0x1420, 0x19FC, 0x3020, 0xD020, 0x17FE, 0x1040, 0x1088, 0x1104, 0x53FE, 0x2102}},
    {25671, 2, {0x1008, 0x103C, 0x13C0, 0x1004, 0xFA44, 0x1128, 0x11FC, 0x1A20, 0x3020, 0xD3FE, 0x1020, 0x1124, 0x1124, 0x1124, 0x51FC, 0x2004}},
    {25672, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC14, 0x11E0, 0x1100, 0x1500, 0x19FC, 0x3110, 0xD110, 0x17FE, 0x1000, 0x1090, 0x5108, 0x2204}},
    {25673, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFCA8, 0x1124, 0x1450, 0x1888, 0x3346, 0xD0F8, 0x1108, 0x1288, 0x1050, 0x1020, 0x5040, 0x2380}},
    {25674, 2, {0x2028, 0x2024, 0x2740, 0x217E, 0xF1C8, 0x2948, 0x257E, 0x2548, 0x3248, 0xE27E, 0x2548, 0x2548, 0x2948, 0x207E, 0xA040, 0x4040}},
    {25675, 2, {0x2000, 0x245C, 0x2294, 0x2114, 0xFA94, 0x2454, 0x2126, 0x2900, 0x37DC, 0xE114, 0x2594, 0x2554, 0x2948, 0x2108, 0xA514, 0x4222}},
    {25676, 2, {0x1040, 0x1020, 0x13FC, 0x1000, 0xFD08, 0x1090, 0x17FE, 0x1A20, 0x32A0, 0xD2FC, 0x1320, 0x1220, 0x12FC, 0x1420, 0x5420, 0x29FE}},
    {25677, 2, {0x1040, 0x1020, 0x17FE, 0x1402, 0xFD00, 0x11FE, 0x1210, 0x1620, 0x1A7C, 0x3244, 0xD244, 0x127C, 0x1244, 0x1244, 0x527C, 0x2244}},
    {25678, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xFA94, 0x24A4, 0x2050, 0x2988, 0x3626, 0xE0C0, 0x2310, 0x2064, 0x2388, 0x2030, 0xA0C0, 0x4700}},
    {25679, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x13FE, 0x1488, 0x1904, 0x3242, 0xD59C, 0x1104, 0x11DC, 0x1104, 0x51FC, 0x2104}},
    {25680, 2, {0x2124, 0x2124, 0x2224, 0xF424, 0x2954, 0x214A, 0x2292, 0x3610, 0xEA10, 0x2250, 0x225C, 0x2250, 0x2250, 0x22B0, 0xA29E, 0x4300}},
    {25681, 2, {0x2000, 0x27FE, 0x242A, 0x2426, 0xF7FE, 0x2422, 0x27AA, 0x26AA, 0x37AA, 0xE432, 0x2596, 0x262E, 0x2446, 0x2482, 0xA7FE, 0x4402}},
    {25682, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0xFBFC, 0x2200, 0x2284, 0x2A48, 0x32FC, 0xE248, 0x2248, 0x23FE, 0x2448, 0x2448, 0xA888, 0x5108}},
    {25683, 2, {0x1020, 0x147C, 0x1284, 0x1148, 0xFC30, 0x10CE, 0x1610, 0x1A7C, 0x3210, 0xD27C, 0x1210, 0x12FE, 0x1210, 0x1210, 0x55FE, 0x2800}},
    {25684, 2, {0x1040, 0x1020, 0x13FE, 0x1020, 0xFC48, 0x12F2, 0x1124, 0x1652, 0x18F8, 0x3008, 0xD020, 0x13FE, 0x1020, 0x1020, 0x5020, 0x2020}},
    {25685, 2, {0x22A8, 0x22A8, 0x27FC, 0x22A8, 0xFAAA, 0x24E6, 0x2800, 0x2FFC, 0x3444, 0xE040, 0x23F8, 0x2248, 0x2248, 0x2248, 0xA258, 0x4040}},
    {25686, 2, {0x1100, 0x1110, 0x13DC, 0x1254, 0xFD54, 0x1288, 0x1108, 0x1AF4, 0x3402, 0xD000, 0x13FC, 0x1040, 0x1150, 0x1248, 0x5544, 0x2080}},
    {25687, 2, {0x2108, 0x2108, 0x27C8, 0x2108, 0xF91E, 0x27D2, 0x2564, 0x2D40, 0x37C8, 0xE108, 0x2388, 0x2548, 0x2954, 0x2114, 0xA124, 0x4142}},
    {25688, 2, {0x2080, 0x2040, 0x27FC, 0x2110, 0xF8A0, 0x27FC, 0x2444, 0x2DF4, 0x3444, 0xE5F4, 0x2514, 0x2514, 0x25F4, 0x2404, 0xA414, 0x4408}},
    {25689, 2, {0x2010, 0x2410, 0x22FE, 0x2210, 0xF8FE, 0x2092, 0x26FE, 0x2A92, 0x32FE, 0xE210, 0x22FE, 0x2210, 0x2210, 0x2510, 0xA8FE, 0x4000}},
    {25690, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFA02, 0x10F8, 0x1488, 0x1888, 0x30F8, 0xD020, 0x1020, 0x11FC, 0x1020, 0x1020, 0x53FE, 0x2000}},
    {25691, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFD54, 0x1124, 0x1154, 0x15FC, 0x1820, 0x33FE, 0xD242, 0x1292, 0x12FA, 0x120A, 0x5202, 0x2206}},
    {25692, 2, {0x1000, 0x11FC, 0x1124, 0x13FE, 0xFD24, 0x11FC, 0x1000, 0x15FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x11FC, 0x5088, 0x2104}},
    {25693, 2, {0x1040, 0x1020, 0x13FE, 0x1250, 0xFE50, 0x13FE, 0x1252, 0x1652, 0x1BFE, 0x3200, 0xD292, 0x12D4, 0x1298, 0x1492, 0x54D2, 0x288E}},
    {25694, 2, {0x1000, 0x13FC, 0x1244, 0x13FC, 0xFE44, 0x13FC, 0x1080, 0x1510, 0x1BE0, 0x3048, 0xD184, 0x13FE, 0x1022, 0x1128, 0x52A4, 0x2442}},
    {25695, 2, {0x1020, 0x11FC, 0x1124, 0x13FE, 0xFD24, 0x11FC, 0x1020, 0x15FC, 0x1924, 0x31FC, 0xD040, 0x13FE, 0x1088, 0x11D0, 0x5070, 0x238C}},
    {25696, 2, {0x1040, 0x1080, 0x13FC, 0x1244, 0xFAF4, 0x1314, 0x12A4, 0x1A44, 0x32A4, 0xD3FC, 0x1040, 0x1024, 0x1522, 0x150A, 0x5908, 0x20F8}},
    {25697, 2, {0x2000, 0x27BE, 0x2488, 0x2488, 0xF7A8, 0x24A8, 0x24A8, 0x27BE, 0x3410, 0xE518, 0x2498, 0x25A8, 0x26AA, 0x244A, 0xA046, 0x4080}},
    {25698, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFA42, 0x1278, 0x13C4, 0x1A3C, 0x3208, 0xD2F0, 0x1324, 0x12A8, 0x15FE, 0x1420, 0x58A0, 0x2040}},
    {25699, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFA42, 0x1278, 0x13C4, 0x1A3C, 0x3200, 0xD2FC, 0x1284, 0x12FC, 0x1284, 0x14FC, 0x5484, 0x29FE}},
    {25700, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFAAA, 0x1124, 0x12AA, 0x1820, 0x32AA, 0xD124, 0x12AA, 0x1050, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25701, 2, {0x1100, 0x11FC, 0x1200, 0x15F8, 0xF908, 0x11F8, 0x1508, 0x19F8, 0x3000, 0xD7FE, 0x1100, 0x13FC, 0x14A4, 0x1124, 0x5254, 0x2088}},
    {25702, 2, {0x2200, 0x220E, 0x2F74, 0xF254, 0x2554, 0x28D4, 0x2754, 0x3054, 0xEFD4, 0x2454, 0x2754, 0x2154, 0x2152, 0x2154, 0xA55C, 0x4284}},
    {25703, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0xFC88, 0x11FE, 0x1310, 0x1510, 0x19FE, 0x3110, 0xD110, 0x11FE, 0x1110, 0x1110, 0x51FE, 0x2100}},
    {25704, 2, {0x1000, 0x13FE, 0x1202, 0x13FE, 0xFE10, 0x1254, 0x1238, 0x1654, 0x1A82, 0x3250, 0xD27C, 0x1290, 0x1210, 0x15FE, 0x5410, 0x2810}},
    {25705, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2038, 0x27C0, 0x2040, 0x27F8, 0x2040, 0x2FFE, 0x4040, 0x4140, 0x8080}},
    {25706, 2, {0x2090, 0x2290, 0x229E, 0x22A2, 0xFAD4, 0x23A8, 0x2094, 0x28A4, 0x37FE, 0xE284, 0x22A4, 0x2294, 0x2294, 0x2284, 0xA494, 0x4888}},
    {25707, 2, {0x2200, 0x227C, 0x2244, 0x2744, 0xFA7C, 0x2244, 0x2244, 0x2F7C, 0x3244, 0xE244, 0x227C, 0x2228, 0x2528, 0x254A, 0xA84A, 0x5086}},
    {25708, 2, {0x1080, 0x1040, 0x13F8, 0x1110, 0xF8A0, 0x17FE, 0x1000, 0x1BF8, 0x3208, 0xD3F8, 0x1208, 0x13F8, 0x1120, 0x1122, 0x5222, 0x2C1E}},
    {25709, 2, {0x2040, 0x2020, 0x23FE, 0x2200, 0xFA48, 0x2248, 0x23FE, 0x2A48, 0x3248, 0xE248, 0x2278, 0x2200, 0x2554, 0x252A, 0xAA2A, 0x5000}},
    {25710, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2264, 0x5FC2, 0x8100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25711, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x8A2, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25712, 2, {0x820, 0x3E20, 0x8F8, 0x7F28, 0x2A68, 0x5D2A, 0x856, 0x3E82, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25713, 2, {0x1000, 0x11F8, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x1000, 0x17FC, 0x1A94, 0x33FC, 0xD000, 0x11F8, 0x1090, 0x1060, 0x5198, 0x2606}},
    {25714, 2, {0x2200, 0x2202, 0x223C, 0x2FA0, 0xF220, 0x2FA0, 0x2ABE, 0x2FA4, 0x3AA4, 0xEFA4, 0x2224, 0x2FA4, 0x2224, 0x2244, 0xA244, 0x4284}},
    {25715, 2, {0x1000, 0x13FE, 0x1200, 0x127C, 0xFE44, 0x1244, 0x127C, 0x1600, 0x1AEE, 0x32AA, 0xD2AA, 0x12AA, 0x12EE, 0x1200, 0x53FE, 0x2000}},
    {25716, 2, {0x11FC, 0x1020, 0x13FE, 0x1222, 0xFDAC, 0x1020, 0x11AC, 0x1400, 0x19FC, 0x3000, 0xD3FE, 0x1080, 0x10FC, 0x1004, 0x5028, 0x2010}},
    {25717, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0xFC10, 0x2490, 0x24D2, 0x2C92, 0x37F2, 0xE494, 0x2494, 0x25C8, 0x2AAA, 0x289A, 0xB1A6, 0x4042}},
    {25718, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x1124, 0x15FC, 0x1822, 0x33FE, 0xD008, 0x13FE, 0x1108, 0x1088, 0x5028, 0x2010}},
    {25719, 2, {0x1124, 0x1248, 0x1124, 0x1000, 0xFDFC, 0x1124, 0x11FC, 0x1524, 0x19FC, 0x3020, 0xD3FE, 0x1070, 0x10A8, 0x1124, 0x5622, 0x2020}},
    {25720, 2, {0x1110, 0x1110, 0x17FC, 0x1110, 0xFC00, 0x13F8, 0x1208, 0x17F8, 0x1A08, 0x33F8, 0xD040, 0x17FC, 0x10A0, 0x1110, 0x5208, 0x2406}},
    {25721, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1050, 0x2788, 0xC106, 0x1FF0, 0x100, 0x300}},
    {25722, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xFA94, 0x24A4, 0x2084, 0x2840, 0x33F8, 0xE208, 0x2208, 0x23F8, 0x2208, 0x2208, 0xA3F8, 0x4208}},
    {25723, 2, {0x1040, 0x1088, 0x11FC, 0x1108, 0xFA52, 0x13FE, 0x1450, 0x1988, 0x3626, 0xD0C0, 0x1310, 0x1064, 0x1388, 0x1030, 0x50C0, 0x2700}},
    {25724, 2, {0x2000, 0x27DC, 0x2514, 0x27D4, 0xFC54, 0x27C8, 0x2508, 0x2D14, 0x37E2, 0xE000, 0x2040, 0x27FC, 0x2040, 0x2040, 0xAFFE, 0x4000}},
    {25725, 2, {0x1000, 0x13FE, 0x1050, 0x11FC, 0xFD54, 0x1154, 0x11FC, 0x1400, 0x19FC, 0x3000, 0xD3FE, 0x1020, 0x10A8, 0x1124, 0x52A2, 0x2040}},
    {25726, 2, {0x2000, 0x277C, 0x2144, 0x2144, 0xF97C, 0x2710, 0x2410, 0x2C7C, 0x3454, 0xE754, 0x2154, 0x217C, 0x2110, 0x2114, 0xAAFE, 0x4402}},
    {25727, 2, {0x1020, 0x1050, 0x1088, 0x1346, 0xFC20, 0x11FC, 0x1088, 0x1450, 0x1BFE, 0x3000, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x51FC, 0x2104}},
    {25728, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4068, 0x9FC6, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25729, 2, {0x1020, 0x13FE, 0x1048, 0x10F0, 0xFC20, 0x13FE, 0x1242, 0x1090, 0x19F8, 0x3028, 0xD120, 0x11FC, 0x1220, 0x17FE, 0x5020, 0x2020}},
    {25730, 2, {0x2000, 0x27FC, 0x2444, 0x27FC, 0xFC44, 0x27FC, 0x2100, 0x2BF8, 0x3510, 0xE0E0, 0x2318, 0x2C06, 0x23F8, 0x2208, 0xA3F8, 0x4208}},
    {25731, 2, {0x1078, 0xFE48, 0x1048, 0x7C86, 0x5478, 0x7C48, 0x5430, 0xFECE, 0x1020, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25732, 2, {0x2000, 0x2FBE, 0x28A2, 0x2AAA, 0xFAAA, 0x2AAA, 0x2514, 0x28A2, 0x3080, 0xEFFE, 0x2110, 0x2210, 0x23A0, 0x2060, 0xA198, 0x4E04}},
    {25733, 2, {0x2000, 0x2FFE, 0x2800, 0x2A28, 0xF948, 0x2BEE, 0x2892, 0x2884, 0x3AA0, 0xEAA8, 0x2BE8, 0x2888, 0x2894, 0x2914, 0xA924, 0x5242}},
    {25734, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25735, 2, {0x2108, 0x2108, 0x2548, 0x2390, 0xF91E, 0x27D4, 0x2564, 0x2D54, 0x3554, 0xE754, 0x25D4, 0x2548, 0x2548, 0x2554, 0xA464, 0x44C2}},
    {25736, 2, {0x2088, 0x22AA, 0x22DC, 0x2488, 0xF954, 0x2222, 0x2000, 0x2BFE, 0x3242, 0xE444, 0x23FC, 0x2044, 0x2044, 0x2084, 0xA114, 0x4208}},
    {25737, 2, {0x840, 0x7F40, 0x227E, 0x3E88, 0x148, 0x3E48, 0x450, 0x7E20, 0x850, 0x288C, 0x10F0, 0x3F00, 0x100, 0x7FFC, 0x100, 0x300}},
    {25738, 2, {0x279E, 0x2492, 0x279E, 0x2492, 0xFF9E, 0x2402, 0x24F2, 0x2C92, 0x34F2, 0xE492, 0x24F2, 0x2492, 0x2492, 0x2532, 0xA402, 0x4406}},
    {25739, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0xFCA4, 0x27BC, 0x2404, 0x2DF4, 0x3444, 0xE4E4, 0x2444, 0x2444, 0x25F4, 0x2404, 0xA414, 0x4408}},
    {25740, 2, {0x1040, 0x13F8, 0x1248, 0x13F8, 0xF840, 0x17FE, 0x1000, 0x1BF8, 0x3208, 0xD3F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x5110, 0x2208}},
    {25741, 2, {0x2000, 0x27BC, 0x2108, 0x2528, 0xF7BC, 0x2318, 0x25AA, 0x2946, 0x3000, 0xE3F8, 0x2208, 0x2208, 0x23F8, 0x2208, 0xA208, 0x43F8}},
    {25742, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC00, 0x13FE, 0x1202, 0x15FC, 0x1800, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x1088, 0x53FE, 0x2000}},
    {25743, 2, {0x1000, 0x13FC, 0x1004, 0x13FC, 0xFC04, 0x13FC, 0x1000, 0x17BC, 0x1924, 0x37BC, 0xD008, 0x13FE, 0x1108, 0x1088, 0x50A8, 0x2010}},
    {25744, 2, {0x1124, 0x10A8, 0x13FE, 0x1202, 0xFCF8, 0x1088, 0x10F8, 0x1400, 0x19FC, 0x3010, 0xD110, 0x11FE, 0x1050, 0x1090, 0x5310, 0x2030}},
    {25745, 2, {0x1124, 0x10A8, 0x13FE, 0x1202, 0xFCF8, 0x1088, 0x10F8, 0x140C, 0x19F0, 0x3020, 0xD1FC, 0x1020, 0x13FE, 0x1020, 0x50A0, 0x2040}},
    {25746, 2, {0x2288, 0x2288, 0x27C8, 0x2290, 0xFA9E, 0x27D4, 0x2024, 0x2FD4, 0x3454, 0xE7D4, 0x2454, 0x27C8, 0x2448, 0x2454, 0xA464, 0x44C2}},
    {25747, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFBFE, 0x1108, 0x179C, 0x1908, 0x3188, 0xD63E, 0x1000, 0x13FE, 0x1090, 0x1090, 0x5112, 0x220E}},
    {25748, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1420, 0x1BFE, 0x3000, 0xD1FC, 0x1104, 0x11FC, 0x1020, 0x1124, 0x5222, 0x2060}},
    {25749, 2, {0x2480, 0x2482, 0x249C, 0x2FD0, 0xF490, 0x2790, 0x249E, 0x2794, 0x3494, 0xE494, 0x2FD4, 0x2014, 0x2514, 0x28A4, 0xB024, 0x4044}},
    {25750, 2, {0x2008, 0x2788, 0x2088, 0x2110, 0xFFDE, 0x2494, 0x24A4, 0x2794, 0x3494, 0xE794, 0x2494, 0x24C8, 0x2788, 0x2C94, 0xA0A4, 0x40C2}},
    {25751, 2, {0x1088, 0x1088, 0x11FC, 0x1088, 0xFC88, 0x13FE, 0x1020, 0x15FC, 0x1924, 0x31FC, 0xD124, 0x11FC, 0x1000, 0x1088, 0x5104, 0x2202}},
    {25752, 2, {0x1210, 0x1210, 0x13DE, 0x1228, 0xFD44, 0x10A0, 0x1510, 0x1A08, 0x35F6, 0xD000, 0x1000, 0x13F8, 0x1208, 0x1208, 0x53F8, 0x2208}},
    {25753, 2, {0x2108, 0x2090, 0x27FE, 0x2090, 0xFBFC, 0x2294, 0x231C, 0x2A04, 0x33FC, 0xE204, 0x23FC, 0x2008, 0x27FE, 0x2108, 0xA0A8, 0x4010}},
    {25754, 2, {0x2110, 0x2114, 0x21D2, 0x2250, 0xFA7E, 0x2550, 0x2090, 0x2AA8, 0x3128, 0xE144, 0x2244, 0x2482, 0x2008, 0x22A4, 0xA252, 0x4452}},
    {25755, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFCA8, 0x1124, 0x1222, 0x1104, 0x1904, 0x31DE, 0xD244, 0x1554, 0x109E, 0x1104, 0x5204, 0x2404}},
    {25756, 2, {0x2020, 0x27A4, 0x20A8, 0x2292, 0xF914, 0x2208, 0x25F4, 0x2802, 0x33F8, 0xE208, 0x2208, 0x23F8, 0x2208, 0x2110, 0xAFFE, 0x4000}},
    {25757, 2, {0x2004, 0x203E, 0x27C0, 0x2244, 0xF928, 0x23F8, 0x2208, 0x2BFC, 0x3204, 0xE3FE, 0x2402, 0x26AA, 0x2AAA, 0x2A02, 0xB014, 0x4008}},
    {25758, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0xFDFE, 0x1000, 0x11FC, 0x1524, 0x19FC, 0x3124, 0xD1FC, 0x1020, 0x11FC, 0x1020, 0x53FE, 0x2000}},
    {25759, 2, {0x101C, 0x11E0, 0x1040, 0x13FE, 0xFC88, 0x1104, 0x12FA, 0x1488, 0x18F8, 0x3000, 0xD3FE, 0x1202, 0x12FA, 0x128A, 0x52FA, 0x2206}},
    {25760, 2, {0x2110, 0x2114, 0x2112, 0x27D2, 0xF910, 0x27DE, 0x2470, 0x2FD2, 0x3452, 0xE7D4, 0x2114, 0x2FD8, 0x210A, 0x211A, 0xA126, 0x4142}},
    {25761, 2, {0x2080, 0x2110, 0x23F8, 0x2210, 0xFCA4, 0x27BC, 0x2040, 0x28A0, 0x3358, 0xEC46, 0x2048, 0x2264, 0x2252, 0x2452, 0xA140, 0x4080}},
    {25762, 2, {0x1000, 0x13FE, 0x1050, 0x13FE, 0xFE52, 0x13FE, 0x1000, 0x15FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1020, 0x13FE, 0x5020, 0x2020}},
    {25763, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFBDE, 0x1000, 0x15FC, 0x1924, 0x31FC, 0xD124, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {25764, 2, {0x2208, 0x2108, 0x27C8, 0x2210, 0xF49E, 0x2FD4, 0x2064, 0x2794, 0x3494, 0xE794, 0x2494, 0x2788, 0x2488, 0x2494, 0xA4A4, 0x45C2}},
    {25765, 2, {0x2024, 0x23A8, 0x2092, 0x2514, 0xF208, 0x2C04, 0x27BA, 0x20A8, 0x37C6, 0xE400, 0x27BC, 0x2084, 0x20A8, 0x2090, 0xA528, 0x4244}},
    {25766, 2, {0x1040, 0x1040, 0x17FC, 0x1110, 0xFA48, 0x1446, 0x13F8, 0x1850, 0x37FC, 0xD080, 0x13F8, 0x1D08, 0x11F8, 0x1108, 0x51F8, 0x2108}},
    {25767, 2, {0x2110, 0x2110, 0x221C, 0x22A4, 0xF4C8, 0x2F3E, 0x212A, 0x222A, 0x34AA, 0xEFBE, 0x20A0, 0x2022, 0x2AA2, 0x2AA2, 0xA81E, 0x4000}},
    {25768, 2, {0x1090, 0x1088, 0x11FE, 0x1110, 0xFB10, 0x15FE, 0x1110, 0x1910, 0x31FE, 0xD110, 0x1110, 0x11FE, 0x1100, 0x12A4, 0x5252, 0x2452}},
    {25769, 2, {0x2040, 0x2040, 0x27FC, 0x20A0, 0xF514, 0x2208, 0x27FC, 0x2A0A, 0x33F8, 0xE208, 0x23F8, 0x2040, 0x2248, 0x2444, 0xA944, 0x4080}},
    {25770, 2, {0x2080, 0x2110, 0x23F8, 0x2080, 0xF7FE, 0x2248, 0x25F4, 0x2842, 0x33F8, 0xE248, 0x23F8, 0x2248, 0x23F8, 0x2040, 0xA7FC, 0x4040}},
    {25771, 2, {0x1100, 0x1100, 0x13FC, 0x16A8, 0xFAA8, 0x12A8, 0x17FC, 0x1AA8, 0x32A8, 0xD2A8, 0x1FFE, 0x1000, 0x12A8, 0x1254, 0x5454, 0x2000}},
    {25772, 2, {0x1038, 0x13C0, 0x1078, 0x13C0, 0xFC78, 0x13C4, 0x103C, 0x1400, 0x18C6, 0x3738, 0xD1CE, 0x1738, 0x11CE, 0x1738, 0x514A, 0x2186}},
    {25773, 2, {0x2078, 0x27C0, 0x2248, 0x2150, 0xFBFC, 0x2150, 0x2248, 0x2C06, 0x33F8, 0xE248, 0x2248, 0x23F8, 0x2248, 0x2248, 0xA3F8, 0x4208}},
    {25774, 2, {0x2000, 0x23F8, 0x2208, 0x23F8, 0xFA08, 0x23F8, 0x2000, 0x2FFE, 0x3480, 0xE7BC, 0x2494, 0x2794, 0x24D4, 0x2F88, 0xA094, 0x40A2}},
    {25775, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x13FE, 0x1480, 0x19FC, 0x3284, 0xD4FC, 0x1084, 0x10FC, 0x1084, 0x50FC, 0x2084}},
    {25776, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0xFC20, 0x24A4, 0x239C, 0x2800, 0x3110, 0xE7FC, 0x2110, 0x2110, 0x2FFE, 0x2110, 0xA208, 0x4404}},
    {25777, 2, {0x2020, 0x2720, 0x25FE, 0x2540, 0xFD7C, 0x2690, 0x257E, 0x2D00, 0x357C, 0xE544, 0x257C, 0x2644, 0x247C, 0x2444, 0xA444, 0x444C}},
    {25778, 2, {0x1050, 0x1252, 0x1154, 0x1050, 0xFBFE, 0x1088, 0x1450, 0x1BFE, 0x3020, 0xD1FC, 0x1020, 0x13FE, 0x1050, 0x1088, 0x5104, 0x2602}},
    {25779, 2, {0x2210, 0x2210, 0x2510, 0xF4BE, 0x2822, 0x2744, 0x2210, 0x3210, 0xEF10, 0x2210, 0x2AA8, 0x2728, 0x2228, 0x2344, 0xAC44, 0x4082}},
    {25780, 2, {0x2210, 0x2110, 0x2FD0, 0x2010, 0xF7BE, 0x24A4, 0x27D4, 0x2014, 0x3794, 0xE094, 0x2114, 0x21C8, 0x2F08, 0x2114, 0xA514, 0x4222}},
    {25781, 2, {0x1088, 0x1088, 0x13DE, 0x1088, 0xFBDE, 0x1088, 0x1154, 0x1A22, 0x33FC, 0xD080, 0x1120, 0x13FC, 0x1020, 0x17FE, 0x5020, 0x2020}},
    {25782, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0xFDFC, 0x10A8, 0x10A8, 0x17FE, 0x18A8, 0x30A8, 0xD1FC, 0x1020, 0x13FE, 0x1020, 0x5020, 0x2020}},
    {25783, 2, {0x2200, 0x22FE, 0x2210, 0x2FA0, 0xF27C, 0x2244, 0x2754, 0x2054, 0x3054, 0xE754, 0x2554, 0x2554, 0x2528, 0x2724, 0xA542, 0x4082}},
    {25784, 2, {0x2100, 0x23F0, 0x2410, 0x2FFC, 0xF444, 0x27FC, 0x2444, 0x27FC, 0x3000, 0xEFFE, 0x2000, 0x23F8, 0x2208, 0x23F8, 0xA208, 0x43F8}},
    {25785, 2, {0x1244, 0x1124, 0x1128, 0x13FE, 0xFA02, 0x11F8, 0x1508, 0x19F8, 0x3108, 0xD1F8, 0x1108, 0x11F8, 0x1090, 0x1090, 0x5112, 0x260E}},
    {25786, 2, {0x2080, 0x2040, 0x27FC, 0x2514, 0xFA48, 0x2444, 0x23F8, 0x2A48, 0x33F8, 0xE040, 0x27FC, 0x2444, 0x27FC, 0x2444, 0xA040, 0x4040}},
    {25787, 2, {0x2010, 0x2410, 0x227C, 0x2010, 0xF8FE, 0x2044, 0x2628, 0x2AFE, 0x3210, 0xE27C, 0x2210, 0x22FE, 0x2210, 0x2210, 0xA5FE, 0x4800}},
    {25788, 2, {0x2014, 0x2012, 0x27FE, 0x2410, 0xF5D0, 0x2414, 0x25D4, 0x2558, 0x35CA, 0xE416, 0x2442, 0x2824, 0x22A2, 0x228A, 0xA478, 0x4000}},
    {25789, 2, {0x2108, 0x2208, 0x27C8, 0x2450, 0xF7DE, 0x2464, 0x27D4, 0x2214, 0x3114, 0xE7D4, 0x2214, 0x23C8, 0x2248, 0x2454, 0xA554, 0x48A2}},
    {25790, 2, {0x1000, 0x127C, 0x1144, 0x1174, 0xFC54, 0x1054, 0x16FE, 0x1282, 0x1ABA, 0x32AA, 0xD2BA, 0x1282, 0x128A, 0x1284, 0x5500, 0x28FE}},
    {25791, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x1000, 0x1400, 0x19DC, 0x3154, 0xD154, 0x11DC, 0x1088, 0x1088, 0x5154, 0x2222}},
    {25792, 2, {0x2108, 0x2108, 0x2114, 0x2FD4, 0xF122, 0x27C0, 0x245C, 0x27C8, 0x3448, 0xE7C8, 0x213E, 0x2FC8, 0x2108, 0x2108, 0xA108, 0x4108}},
    {25793, 2, {0x2040, 0x2020, 0x27FE, 0x2028, 0xFA24, 0x227E, 0x24C8, 0x2F48, 0x317E, 0xE248, 0x24C8, 0x2F7E, 0x2148, 0x2248, 0xA47E, 0x4840}},
    {25794, 2, {0x2000, 0x27FC, 0x2040, 0x2FFE, 0xF842, 0x2358, 0x2040, 0x2B58, 0x3000, 0xE7FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0xA444, 0x47FC}},
    {25795, 2, {0x1090, 0x13FC, 0x1294, 0x13FC, 0xFE94, 0x13FC, 0x1000, 0x17FC, 0x1A00, 0x32F8, 0xD200, 0x13FE, 0x1520, 0x1514, 0x5548, 0x2986}},
    {25796, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFE42, 0x13F8, 0x1244, 0x12FC, 0x1AA8, 0x32F8, 0xD2A8, 0x12F8, 0x1220, 0x15FC, 0x5444, 0x298C}},
    {25797, 2, {0x1020, 0x13FF, 0x1000, 0xFDFE, 0x1102, 0x117A, 0x114A, 0x157A, 0x1902, 0x31FE, 0xD000, 0x10FC, 0x1084, 0x10FC, 0x5084, 0x33FF}},
    {25798, 2, {0x1108, 0x1108, 0x17FE, 0x1108, 0xF844, 0x13F4, 0x1048, 0x1850, 0x37FE, 0xD080, 0x11FC, 0x1304, 0x15FC, 0x1904, 0x51FC, 0x2104}},
    {25799, 2, {0x1000, 0x13FE, 0x1252, 0x13FE, 0xFC20, 0x11FC, 0x1420, 0x1BFE, 0x3088, 0xD050, 0x11FC, 0x1020, 0x17FE, 0x1020, 0x5020, 0x2020}},
    {25800, 2, {0x1050, 0x1252, 0x1154, 0x1050, 0xFBFE, 0x1088, 0x1450, 0x1BFE, 0x3020, 0xD1FC, 0x1020, 0x13FE, 0x10A8, 0x1124, 0x5622, 0x2020}},
    {25801, 2, {0x1000, 0x13F8, 0x12A8, 0x12A8, 0xFBF8, 0x1100, 0x13FC, 0x1C84, 0x33E4, 0xD2A4, 0x12A4, 0x13E4, 0x1094, 0x17F4, 0x5014, 0x2008}},
    {25802, 2, {0x83C, 0x7F24, 0x2A24, 0x3E42, 0x2A3C, 0xFFA4, 0x4918, 0x7F66, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25803, 2, {0x2040, 0x2248, 0x2150, 0x27FE, 0xF402, 0x2804, 0x23F8, 0x2208, 0x33F8, 0xE000, 0x27FC, 0x2444, 0x27FC, 0x2444, 0xA7FC, 0x4404}},
    {25804, 2, {0x2210, 0x2210, 0x23DE, 0x2528, 0xF8C4, 0x2040, 0x27FE, 0x2840, 0x3248, 0xE248, 0x2554, 0x28A2, 0x20A0, 0x2110, 0xA208, 0x4C06}},
    {25805, 2, {0x23F8, 0x2208, 0x2208, 0x23F8, 0xF800, 0x27BC, 0x24A4, 0x2CA4, 0x37BC, 0xE040, 0x2FFE, 0x2160, 0x2250, 0x2448, 0xB846, 0x4040}},
    {25806, 2, {0x2420, 0xFF20, 0x247E, 0x7EC4, 0x8228, 0x7A10, 0x4A28, 0x7AC6, 0x420, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25807, 2, {0x2288, 0x2288, 0x27E8, 0x2290, 0xFC1E, 0x27E4, 0x2454, 0x2854, 0x3754, 0xE554, 0x2554, 0x2748, 0x2548, 0x2054, 0xA294, 0x4122}},
    {25808, 2, {0x1000, 0x13FC, 0x1294, 0x1294, 0xFBFC, 0x1000, 0x17FE, 0x1800, 0x33FC, 0xD204, 0x13FC, 0x10A2, 0x1114, 0x1308, 0x5D44, 0x2182}},
    {25809, 2, {0x1010, 0x13D4, 0x1252, 0x1252, 0xFBD0, 0x1010, 0x17FE, 0x1A50, 0x33D2, 0xD252, 0x13D4, 0x1254, 0x12EA, 0x174A, 0x5056, 0x2062}},
    {25810, 2, {0x2040, 0x20A0, 0x2110, 0x2248, 0xFDF6, 0x20A0, 0x2248, 0x2AA8, 0x33F8, 0xE040, 0x27FC, 0x2484, 0x2524, 0x25F4, 0xA414, 0x440C}},
    {25811, 2, {0x2000, 0x27FE, 0x2428, 0x2624, 0xFD7E, 0x2448, 0x24C8, 0x2C7E, 0x3648, 0xE57E, 0x2448, 0x2448, 0x257E, 0x2640, 0xA7FE, 0x4000}},
    {25812, 2, {0x1080, 0x10F8, 0x1108, 0x13FE, 0xFD44, 0x1192, 0x11FE, 0x1500, 0x197C, 0x3100, 0xD17C, 0x1100, 0x117C, 0x1244, 0x527C, 0x2444}},
    {25813, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x11FC, 0x1520, 0x19FE, 0x3100, 0xD3DE, 0x1252, 0x1252, 0x1272, 0x5202, 0x2206}},
    {25814, 2, {0x1088, 0x13FE, 0x1088, 0x11FC, 0xFD04, 0x11FC, 0x1104, 0x15FC, 0x1880, 0x31FE, 0xD222, 0x1552, 0x1102, 0x11FA, 0x500A, 0x2004}},
    {25815, 2, {0x2010, 0x2008, 0x2788, 0x24BE, 0xF480, 0x2494, 0x2788, 0x247E, 0x3408, 0xE788, 0x26BE, 0x2A88, 0x2A88, 0x2B88, 0xB288, 0x4008}},
    {25816, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25817, 2, {0x1040, 0x1080, 0x13FE, 0x1222, 0xFEAA, 0x1222, 0x13FE, 0x1672, 0x1AAA, 0x3202, 0xD020, 0x13FE, 0x1020, 0x1050, 0x5088, 0x2306}},
    {25818, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFE42, 0x1278, 0x13C4, 0x163C, 0x1A00, 0x33FE, 0xD240, 0x13A4, 0x1258, 0x15B4, 0x5452, 0x29B0}},
    {25819, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0xFD48, 0x13FE, 0x1148, 0x1578, 0x1900, 0x31FC, 0xD020, 0x13FE, 0x1070, 0x10A8, 0x5326, 0x2020}},
    {25820, 2, {0x2108, 0x2108, 0x27CC, 0x210A, 0xF908, 0x27DE, 0x2448, 0x2EC8, 0x3548, 0xE7C8, 0x2554, 0x27D4, 0x2554, 0x2554, 0xA454, 0x44E2}},
    {25821, 2, {0x2000, 0x27BC, 0x24A4, 0x24BC, 0xF7A4, 0x24BC, 0x24A4, 0x27A4, 0x3024, 0xE04C, 0x23F8, 0x22A8, 0x22A8, 0x22A8, 0xAFFE, 0x4000}},
    {25822, 2, {0x2210, 0x2A90, 0x2710, 0x2220, 0xFFBE, 0x2244, 0x27A4, 0x2AA4, 0x3228, 0xEFA8, 0x2490, 0x2890, 0x2528, 0x2228, 0xA544, 0x4882}},
    {25823, 2, {0x2000, 0x27FC, 0x2248, 0xF444, 0x2FFE, 0x2444, 0x26EC, 0x3554, 0xE6EC, 0x2444, 0x26EC, 0x2554, 0x26EC, 0x2444, 0xA454, 0x4408}},
    {25824, 2, {0x2080, 0x2040, 0x27FC, 0x2110, 0xF0A4, 0x2F58, 0x2554, 0x2552, 0x3B58, 0xE000, 0x2208, 0x23F8, 0x2208, 0x23F8, 0xA208, 0x4408}},
    {25825, 2, {0x2040, 0x2FFE, 0x2040, 0x27FC, 0xF000, 0x23F8, 0x2208, 0x2FFE, 0x3802, 0xE7FC, 0x2110, 0x23F8, 0x2040, 0x27FC, 0xA040, 0x4FFE}},
    {25826, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xFA94, 0x24A4, 0x2120, 0x2A10, 0x37FC, 0xEA20, 0x23FC, 0x2220, 0x23FC, 0x2220, 0xA3FE, 0x4200}},
    {25827, 2, {0x2040, 0x27FC, 0x2040, 0x23FC, 0xF800, 0x27FE, 0x2002, 0x2BF8, 0x3040, 0xE7FE, 0x2000, 0x27FE, 0x2008, 0x27FE, 0xA4A8, 0x4798}},
    {25828, 2, {0x2080, 0x23F8, 0x2208, 0x23F8, 0xFA08, 0x23F8, 0x2000, 0x2FFC, 0x3444, 0xE7FC, 0x2444, 0x27FC, 0x2000, 0x2FFE, 0xA110, 0x4210}},
    {25829, 2, {0x0, 0x7E40, 0x487E, 0x7E90, 0x4308, 0x7EFE, 0x48AA, 0x7EFE, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25830, 2, {0x2080, 0x2040, 0x27FE, 0x2402, 0xFA24, 0x23BC, 0x24A4, 0x2AA8, 0x3510, 0xE2E8, 0x2404, 0x2BFA, 0x2040, 0x2248, 0xA444, 0x40C0}},
    {25831, 2, {0xA00, 0x33B8, 0x2208, 0x3BB8, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x820, 0x37D8, 0xC106, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x300}},
    {25832, 2, {0x2008, 0x2068, 0x2F88, 0x28EE, 0xF88A, 0x2BF2, 0x2AA4, 0x2AC4, 0x3B84, 0xEAA4, 0x2A64, 0x2A0A, 0x2AEA, 0x2AAA, 0xAAAA, 0x5530}},
    {25833, 2, {0x2000, 0x27FC, 0x2040, 0x2FFE, 0xF842, 0x2358, 0x2040, 0x2B58, 0x3000, 0xEFFE, 0x2040, 0x27FC, 0x24A4, 0x24A4, 0xA4A4, 0x440C}},
    {25834, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x27F8, 0x2040, 0x27FC, 0x2040, 0x5FFE, 0x4040, 0x80C0}},
    {25835, 2, {0x2000, 0x2FFE, 0x2808, 0x2BC8, 0xFA4C, 0x2BCA, 0x2A48, 0x2BDE, 0x3808, 0xEBC8, 0x2A48, 0x2BD4, 0x2A54, 0x2BD4, 0xAA54, 0x52E2}},
    {25836, 2, {0x2400, 0x257C, 0x2604, 0xF4A8, 0x2390, 0x207C, 0x2414, 0x3794, 0xEA50, 0x2250, 0x2FDC, 0x2250, 0x2550, 0x24B0, 0xA89E, 0x4100}},
    {25837, 2, {0x1088, 0x13FE, 0x10A8, 0x1090, 0xFDFE, 0x1320, 0x11FC, 0x1520, 0x19FC, 0x3120, 0xD1FE, 0x1100, 0x13FC, 0x1088, 0x5070, 0x238E}},
    {25838, 2, {0x2110, 0x2114, 0x27D2, 0x2110, 0xFFFE, 0x2290, 0x2450, 0x2FF4, 0x3494, 0xE7F4, 0x2494, 0x27E8, 0x248A, 0x27FA, 0xA426, 0x4042}},
    {25839, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFDFC, 0x1148, 0x1650, 0x19FC, 0x3304, 0xD1FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x5088, 0x2104}},
    {25840, 2, {0x1040, 0x1020, 0x13FE, 0x1222, 0xFD54, 0x114A, 0x123A, 0x1000, 0x19FC, 0x3154, 0xD3FE, 0x1000, 0x11FC, 0x1020, 0x50A0, 0x2040}},
    {25841, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0xFCA4, 0x27BC, 0x2484, 0x2DF4, 0x3694, 0xE4E4, 0x271C, 0x24E4, 0x24A4, 0x24A4, 0xA4E4, 0x440C}},
    {25842, 2, {0x2440, 0x228E, 0x2FEA, 0xF28A, 0x2FEA, 0x2AAA, 0x2C6C, 0x3BAA, 0xE82A, 0x2FEA, 0x210A, 0x2FEA, 0x228C, 0x2448, 0xA828, 0x4008}},
    {25843, 2, {0x1108, 0x1108, 0x11EE, 0x1294, 0xFC42, 0x1000, 0x13DE, 0x1652, 0x1BD2, 0x3252, 0xD3D2, 0x121A, 0x1294, 0x1350, 0x5210, 0x2010}},
    {25844, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0xFDFC, 0x2488, 0x27FE, 0x2C20, 0x35FC, 0xE524, 0x25FC, 0x2524, 0x25FC, 0x2400, 0xA488, 0x4904}},
    {25845, 2, {0x2020, 0x27FE, 0x2488, 0x27DE, 0xF488, 0x25DC, 0x26AA, 0x248C, 0x35F0, 0xE420, 0x25FC, 0x2420, 0x27FE, 0x2820, 0xA8A0, 0x5040}},
    {25846, 2, {0x1210, 0x13DE, 0x1528, 0x1084, 0xFD08, 0x1090, 0x17FE, 0x1000, 0x1BD4, 0x3254, 0xD3D4, 0x1254, 0x13D4, 0x1244, 0x5254, 0x22C8}},
    {25847, 2, {0x2200, 0x227E, 0x2210, 0x2FA0, 0xF27C, 0x2244, 0x2744, 0x207C, 0x3044, 0xE77C, 0x2544, 0x2544, 0x257C, 0x2700, 0xA528, 0x4044}},
    {25848, 2, {0x1124, 0x1248, 0x1124, 0x1000, 0xFBFC, 0x1294, 0x1664, 0x1A94, 0x33FC, 0xD248, 0x1368, 0x1248, 0x136A, 0x124A, 0x5246, 0x2362}},
    {25849, 2, {0x2290, 0x2290, 0x27D0, 0x229E, 0xFB92, 0x2122, 0x27C8, 0x2D48, 0x37C8, 0xE108, 0x27C8, 0x2114, 0x2FD4, 0x2294, 0xA454, 0x4822}},
    {25850, 2, {0x2000, 0x27FC, 0x24A4, 0x24A4, 0xF7FC, 0x2220, 0x24A4, 0x27B8, 0x3022, 0xEF9E, 0x2880, 0x2FA4, 0x28B8, 0x2FA2, 0xA8A2, 0x499E}},
    {25851, 2, {0x2108, 0x27C8, 0x2548, 0x2FF0, 0xF55E, 0x27E4, 0x2114, 0x27D4, 0x3554, 0xE7D4, 0x2214, 0x2FC8, 0x2448, 0x2694, 0xA194, 0x4662}},
    {25852, 2, {0x2100, 0x23F0, 0x2410, 0x2FFC, 0xF444, 0x27FC, 0x2444, 0x27FC, 0x3524, 0xE892, 0x23F8, 0x2208, 0x23F8, 0x2208, 0xA3F8, 0x4208}},
    {25853, 2, {0x2444, 0x24E4, 0x28A8, 0x2AAA, 0xFEEE, 0x24A4, 0x2AAA, 0x2EEE, 0x3242, 0xE040, 0x2FFE, 0x20E0, 0x2150, 0x2248, 0xAC46, 0x4040}},
    {25854, 2, {0x27FC, 0x2080, 0x23F8, 0x2208, 0xF3F8, 0x2208, 0x2FFE, 0x2882, 0x3548, 0xE514, 0x28F4, 0x2100, 0x23F8, 0x2508, 0xA0F0, 0x4F0E}},
    {25855, 2, {0x2020, 0x2010, 0x24FE, 0x2244, 0xFA28, 0x20FE, 0x2092, 0x2EFE, 0x3292, 0xE2BA, 0x22AA, 0x22BA, 0x2286, 0x2500, 0xA8FE, 0x4000}},
    {25856, 2, {0x2288, 0x2108, 0xFABE, 0x2008, 0x729C, 0xA92A, 0x2288, 0xFFFE, 0x820, 0x37D8, 0xC106, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x300}},
    {25857, 2, {0x2208, 0x2110, 0x27FC, 0x2040, 0xFBF8, 0x2080, 0x27FC, 0x2948, 0x33FC, 0xE60A, 0x2BF8, 0x2208, 0x23FA, 0x2234, 0xA28C, 0x4302}},
    {25858, 2, {0x23F8, 0x2248, 0x2248, 0x23F8, 0xF248, 0x2248, 0x23F8, 0x2000, 0x3FBE, 0xEAAA, 0x2AAA, 0x2FBE, 0x2AAA, 0x2AAA, 0xAFBE, 0x48A2}},
    {25859, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFD10, 0x11DE, 0x1252, 0x1554, 0x1A88, 0x3174, 0xD202, 0x11FC, 0x1020, 0x1124, 0x5222, 0x2060}},
    {25860, 2, {0x2020, 0x203E, 0x2020, 0x27FE, 0xFC22, 0x25F8, 0x2422, 0x2DFE, 0x3524, 0xE5FC, 0x2524, 0x25FC, 0x2840, 0x2AA4, 0xB28A, 0x447A}},
    {25861, 2, {0x1088, 0x11DC, 0x1088, 0x13DE, 0xFC88, 0x1154, 0x1222, 0x11FC, 0x1904, 0x31FC, 0xD104, 0x11FC, 0x1104, 0x11FC, 0x5088, 0x2104}},
    {25862, 2, {0x1088, 0x13DE, 0x1088, 0x13DE, 0xFD54, 0x1222, 0x11FC, 0x1420, 0x19FC, 0x3124, 0xD1FC, 0x1124, 0x11FC, 0x1020, 0x53FE, 0x2020}},
    {25863, 2, {0x2040, 0x2FFE, 0x2842, 0x27FC, 0xF840, 0x23F8, 0x2040, 0x2FFC, 0x3040, 0xE7FC, 0x24A4, 0x27FC, 0x2040, 0x2524, 0xA52A, 0x48FA}},
    {25864, 2, {0x1020, 0x13FE, 0x1250, 0x13FE, 0xFA52, 0x13FE, 0x1294, 0x1AD8, 0x3292, 0xD2CE, 0x1208, 0x12F0, 0x1220, 0x15FC, 0x54A8, 0x2B26}},
    {25865, 2, {0x27FC, 0x2040, 0x27FE, 0x2442, 0xFB5C, 0x2040, 0x275C, 0x2920, 0x33FC, 0xE620, 0x2BFC, 0x2220, 0x23FC, 0x2220, 0xA3FE, 0x4200}},
    {25866, 2, {0x2000, 0x27FE, 0x2444, 0x2598, 0xFC88, 0x27DE, 0x2488, 0x2DDC, 0x36AA, 0xE488, 0x2420, 0x2520, 0x253C, 0x2920, 0xA920, 0x57FE}},
    {25867, 2, {0x2108, 0x2108, 0x27CE, 0x2112, 0xF924, 0x27DE, 0x2552, 0x2D5E, 0x37D2, 0xE11E, 0x2392, 0x255E, 0x2940, 0x2114, 0xA112, 0x4122}},
    {25868, 2, {0x2FFE, 0x2802, 0x2BFA, 0x2AAA, 0xFBFA, 0x2802, 0x2BFA, 0x2802, 0x39F2, 0xE912, 0x29F2, 0x28AA, 0x2992, 0x2ACA, 0xA882, 0x4FFE}},
    {25869, 2, {0x4080, 0x7FFC, 0x4800, 0x4FF8, 0xE000, 0x4FF8, 0x4808, 0x4FF8, 0x6080, 0xDC9C, 0x57D4, 0x5D54, 0x555C, 0x5C96, 0x5556, 0xEE22}},
    {25870, 2, {0x1040, 0x107C, 0x1040, 0x13FC, 0xFE44, 0x13F0, 0x1244, 0x16FC, 0x1AA8, 0x32F8, 0xD2A8, 0x12F8, 0x1200, 0x15FC, 0x5554, 0x2BFE}},
    {25871, 2, {0x1110, 0x1090, 0x13DE, 0x1010, 0xFA5E, 0x1182, 0x17DE, 0x1810, 0x33DE, 0xD250, 0x13DE, 0x1250, 0x13DE, 0x1250, 0x5252, 0x22CE}},
    {25872, 2, {0x1040, 0x17FE, 0x1492, 0x13FC, 0xFC90, 0x13FC, 0x1090, 0x17FE, 0x1948, 0x3224, 0xD5FA, 0x10A4, 0x11A8, 0x1290, 0x54C8, 0x2086}},
    {25873, 2, {0x1140, 0x126C, 0x1244, 0x136C, 0xFEA4, 0x13AC, 0x12A4, 0x17FE, 0x1908, 0x3244, 0xD5F2, 0x1040, 0x17FC, 0x1040, 0x5040, 0x2040}},
    {25874, 2, {0x2108, 0x2528, 0x27BE, 0x2948, 0xF7BE, 0x2318, 0x25AA, 0x2946, 0x3000, 0xE3F8, 0x2208, 0x2248, 0x2248, 0x20B0, 0xA108, 0x4604}},
    {25875, 2, {0x2040, 0x27FE, 0x2492, 0x23FC, 0xF890, 0x23FC, 0x2090, 0x2FFE, 0x3108, 0xE3FC, 0x250A, 0x21F8, 0x2240, 0x227C, 0xA540, 0x48FE}},
    {25876, 2, {0x27BC, 0x24A4, 0x27BC, 0x24A4, 0xFFBC, 0x2444, 0x25F4, 0x2C44, 0x35F4, 0xE554, 0x25F4, 0x2554, 0x25F4, 0x24E4, 0xA554, 0x444C}},
    {25877, 2, {0x2448, 0x244C, 0x2AAA, 0xF008, 0x2FFE, 0x2288, 0x2EE8, 0x328A, 0xEEEA, 0x228C, 0x2EEC, 0x2288, 0x22EA, 0x2F1A, 0xA426, 0x4042}},
    {25878, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x1252, 0x13DE, 0x198C, 0x3252, 0xD040, 0x17FE, 0x1088, 0x1190, 0x1060, 0x50D8, 0x2304}},
    {25879, 2, {0x2040, 0x27FE, 0x24A0, 0x27FC, 0xFCA4, 0x27FC, 0x2514, 0x2DD8, 0x3512, 0xE54E, 0x25A4, 0x24A8, 0x27FE, 0x28A8, 0xA924, 0x5222}},
    {25880, 2, {0x2040, 0x27FC, 0x2000, 0x23B8, 0xFAA8, 0x23B8, 0x2110, 0x2FFC, 0x3110, 0xE7FC, 0x2110, 0x2FFE, 0x2128, 0x2310, 0xAD48, 0x4186}},
    {25881, 2, {0x11F0, 0x1210, 0x17FC, 0x1204, 0xFBFC, 0x1224, 0x13B8, 0x1A22, 0x31FE, 0xD210, 0x17FC, 0x1244, 0x13FC, 0x10D0, 0x514A, 0x263E}},
    {25882, 2, {0x1088, 0x12AA, 0x12DC, 0x1488, 0xF954, 0x1222, 0x17FE, 0x1A02, 0x30F8, 0xD088, 0x10F8, 0x1000, 0x11FC, 0x1104, 0x51FC, 0x2104}},
    {25883, 2, {0x1040, 0x17FE, 0x1492, 0x1108, 0xFC80, 0x131C, 0x1204, 0x179C, 0x1A04, 0x33FC, 0xD248, 0x1368, 0x1248, 0x136A, 0x5246, 0x2362}},
    {25884, 2, {0x1222, 0x13FE, 0x1090, 0x11FE, 0xFB10, 0x15FE, 0x1110, 0x19FE, 0x3110, 0xD1FE, 0x1100, 0x13FE, 0x128A, 0x1376, 0x5252, 0x2276}},
    {25885, 2, {0x13FC, 0x1108, 0x11F8, 0x1108, 0xFDF8, 0x110E, 0x17F8, 0x1808, 0x37FE, 0xD294, 0x139C, 0x1294, 0x139C, 0x12D6, 0x57BC, 0x2084}},
    {25886, 2, {0x2000, 0x27FC, 0x24A4, 0xF4A4, 0x27FC, 0x2210, 0x22A8, 0x34BE, 0xEF68, 0x223C, 0x24A8, 0x2FBC, 0x2028, 0x2AA8, 0xAABE, 0x4020}},
    {25887, 2, {0x2040, 0x27FE, 0x24A0, 0x27FC, 0xFCA4, 0x27FC, 0x2400, 0x2DFC, 0x3504, 0xE574, 0x2524, 0x25FC, 0x2574, 0x29AC, 0xA924, 0x51FC}},
    {25888, 2, {0x2020, 0x27FE, 0x2488, 0x27DE, 0xFC88, 0x25DC, 0x26AA, 0x2C88, 0x3450, 0xE7DE, 0x2450, 0x25DC, 0x2450, 0x2BDE, 0xA850, 0x5050}},
    {25889, 2, {0x2418, 0x2214, 0x2F90, 0x253E, 0xFAA8, 0x2DE8, 0x28BE, 0x2FA8, 0x3228, 0xEFBE, 0x2AA8, 0x2AA8, 0x2BA8, 0x28BE, 0xAAA0, 0x4920}},
    {25890, 2, {0x2528, 0x27BE, 0x2948, 0x27BE, 0xFB18, 0x25AA, 0x2946, 0x2BFC, 0x3204, 0xE3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0xA108, 0x4204}},
    {25891, 2, {0x2108, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0x20, 0x1FC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {25892, 2, {0x2518, 0x2514, 0x2F90, 0x253E, 0xF728, 0x2268, 0x2FBE, 0x2AA8, 0x3FA8, 0xE23E, 0x2FA8, 0x2228, 0x3FE8, 0x253E, 0xA8A0, 0x5060}},
    {25893, 2, {0x2210, 0x23DE, 0x2528, 0x2084, 0xFBFC, 0x22A4, 0x2254, 0x2BFC, 0x3080, 0xE7FE, 0x22A8, 0x25E4, 0x2852, 0x23F8, 0xA150, 0x42C8}},
    {25894, 2, {0x2FBE, 0x2000, 0x27BC, 0x24A4, 0xF6B4, 0x24A4, 0x2040, 0x27FE, 0x34A0, 0xE7FC, 0x24A4, 0x27FC, 0x2512, 0x29DC, 0xA912, 0x51CE}},
    {25895, 2, {0x2200, 0x221E, 0x2FC4, 0x2208, 0xF79E, 0x2492, 0x2792, 0x249E, 0x3792, 0xE49E, 0x2792, 0x2492, 0x2FDE, 0x2500, 0xA48C, 0x4892}},
    {25896, 2, {0x2040, 0x2FFE, 0x2A0A, 0x2434, 0xF3C0, 0x2290, 0x2288, 0x22A4, 0x34D0, 0xE18E, 0x2E74, 0x2B54, 0x2AD4, 0x2A52, 0xAB5A, 0x5294}},
    {25897, 2, {0x2248, 0x2150, 0x27FC, 0x2404, 0xF9F0, 0x2110, 0x27FC, 0x2D54, 0x34E4, 0xE7FC, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0xA2A4, 0x4452}},
    {25898, 2, {0x22AC, 0x2444, 0x26AC, 0x2444, 0xF6AC, 0x2404, 0x2FFE, 0x2A0A, 0x33F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x20A0, 0xA124, 0x461C}},
    {25899, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x10A0, 0x11FE, 0x1B20, 0x35FC, 0xD120, 0x11FE, 0x1000, 0x11FC, 0x1088, 0x5070, 0x278E}},
    {25900, 2, {0x2790, 0x251E, 0x27A8, 0x2484, 0xFFBE, 0x252A, 0x27BE, 0x2800, 0x33F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x20A0, 0xA124, 0x461C}},
    {25901, 2, {0x21F0, 0x2210, 0x23E0, 0x2020, 0xF7FC, 0x2188, 0x2650, 0x21A8, 0x3664, 0xE1A0, 0x2248, 0x2FBE, 0x2AAA, 0x2FBE, 0xA28A, 0x4FBE}},
    {25902, 2, {0x2040, 0x27FC, 0x2248, 0x23F8, 0xF040, 0x2FFE, 0x2AAA, 0x23B8, 0x3110, 0xE7FC, 0x2110, 0x2FFE, 0x2128, 0x2310, 0xAD48, 0x4186}},
    {25903, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF0, 0x1010, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {25904, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x10FC, 0x1024, 0xFC24, 0x4424, 0x4424, 0x2844, 0x2844, 0x1094, 0x1908, 0x2400, 0x43FE, 0x8000}},
    {25905, 2, {0x2020, 0x1020, 0x1020, 0xFDFE, 0x20, 0x820, 0x8820, 0x89FC, 0x4884, 0x4888, 0x5048, 0x5050, 0x1C20, 0xE050, 0x4188, 0x606}},
    {25906, 2, {0x1010, 0x1010, 0xFE10, 0x10FE, 0x2810, 0x4410, 0xFEFC, 0x444, 0x7444, 0x5444, 0x5428, 0x7428, 0x5410, 0x428, 0x1444, 0x982}},
    {25907, 2, {0x11FC, 0x1004, 0x10FC, 0xFE04, 0x11FC, 0x1000, 0xFDDE, 0x4492, 0x47DE, 0x2808, 0x2BFE, 0x1108, 0x18A8, 0x2410, 0x43FE, 0x8000}},
    {25908, 2, {0x100, 0x100, 0x100, 0x1FC, 0x100, 0x100, 0x100, 0x3FF0, 0x1010, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {25909, 2, {0x400, 0x400, 0x400, 0xFFE, 0x810, 0x1010, 0x3020, 0x4820, 0x8440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {25910, 2, {0x840, 0x840, 0x4840, 0x4880, 0x48FE, 0x4908, 0x4A88, 0x4888, 0x4888, 0x5850, 0x6850, 0x4820, 0x850, 0x888, 0x904, 0xA02}},
    {25911, 2, {0x40, 0x40, 0xFE40, 0x2080, 0x20FE, 0x2108, 0x4288, 0x7C88, 0x488, 0x450, 0x450, 0x420, 0x450, 0x488, 0x2904, 0x1202}},
    {25912, 2, {0x820, 0x820, 0x820, 0x103E, 0x1244, 0x3244, 0x5244, 0x92A4, 0x1228, 0x1228, 0x1210, 0x1210, 0x1228, 0x1048, 0x1084, 0x1102}},
    {25913, 2, {0x40, 0x40, 0xFC80, 0x4FE, 0x508, 0x508, 0x688, 0x7C88, 0x4088, 0x4050, 0x4050, 0x4420, 0x5850, 0x6088, 0x4104, 0x202}},
    {25914, 2, {0x40, 0x40, 0x7C40, 0x4480, 0x44FE, 0x4508, 0x4688, 0x4488, 0x7C88, 0x4050, 0x4050, 0x4020, 0x4450, 0x5888, 0x6104, 0x202}},
    {25915, 2, {0x40, 0x40, 0x40, 0xFE80, 0x10FE, 0x1108, 0x1288, 0x1088, 0x1088, 0x1050, 0x1050, 0x1E20, 0xF050, 0x4088, 0x104, 0x202}},
    {25916, 2, {0x40, 0x7C40, 0x1040, 0x1080, 0x10FE, 0x1108, 0x1288, 0xFE88, 0x1088, 0x1050, 0x1050, 0x1020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {25917, 2, {0x20, 0x1420, 0x1220, 0x123E, 0x2144, 0x2144, 0x4044, 0xBEA4, 0x1228, 0x1228, 0x1210, 0x1210, 0x2228, 0x2248, 0x4A84, 0x8502}},
    {25918, 2, {0x2040, 0x1040, 0x40, 0xFE80, 0x20FE, 0x2108, 0x3E88, 0x2488, 0x2488, 0x2450, 0x2450, 0x2420, 0x4450, 0x5488, 0x8904, 0x202}},
    {25919, 2, {0x20, 0x20, 0xFF20, 0x83E, 0x844, 0x844, 0x4844, 0x4EA4, 0x4828, 0x4828, 0x4810, 0x4810, 0x4F28, 0xF048, 0x84, 0x102}},
    {25920, 2, {0x1040, 0x1040, 0x2040, 0x7C80, 0x44FE, 0x4508, 0x4688, 0x4488, 0x7C88, 0x4450, 0x4450, 0x4420, 0x4450, 0x7C88, 0x4504, 0x202}},
    {25921, 2, {0x1020, 0x1020, 0x1020, 0x1E3E, 0x1020, 0x1020, 0x1020, 0x7DFC, 0x4484, 0x4488, 0x4448, 0x4450, 0x7C20, 0x4450, 0x188, 0x606}},
    {25922, 2, {0x1020, 0x1020, 0x2020, 0x3F3E, 0x4144, 0x8144, 0x3D44, 0x25A4, 0x2528, 0x2528, 0x3D10, 0x2510, 0x128, 0x148, 0xA84, 0x502}},
    {25923, 2, {0x20, 0xFE20, 0x8220, 0x823E, 0x8244, 0xFE44, 0x9044, 0x90A4, 0xFF28, 0x9028, 0x9010, 0x8810, 0x8928, 0xA528, 0xC344, 0x8182}},
    {25924, 2, {0x40, 0xFC40, 0x440, 0x6880, 0x10FE, 0x908, 0xFE88, 0x1288, 0x3488, 0x3050, 0x5050, 0x5020, 0x9050, 0x1088, 0x5104, 0x2202}},
    {25925, 2, {0x1040, 0x1040, 0x1040, 0x1080, 0xFEFE, 0x1108, 0x1288, 0x1088, 0x7C88, 0x4450, 0x4450, 0x4420, 0x4450, 0x7C88, 0x4504, 0x202}},
    {25926, 2, {0x810, 0x810, 0x141E, 0x2210, 0x4110, 0xBEFC, 0x44, 0x44, 0x3E44, 0x2228, 0x2228, 0x2210, 0x2210, 0x3E28, 0x2244, 0x82}},
    {25927, 2, {0x810, 0x810, 0xFF90, 0x820, 0x83E, 0x7F44, 0x4924, 0x4924, 0x4D24, 0x4A28, 0x1C28, 0x2A10, 0x4928, 0x88A8, 0x844, 0x882}},
    {25928, 2, {0x1020, 0x820, 0x20, 0xFF3E, 0x44, 0x2444, 0x4244, 0x81A4, 0x2428, 0x1428, 0x810, 0x1410, 0x2228, 0x4248, 0x8084, 0x102}},
    {25929, 2, {0x1040, 0x1040, 0x9440, 0x5480, 0x58FE, 0x1108, 0xFE88, 0x3088, 0x3888, 0x5450, 0x5450, 0x9020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {25930, 2, {0x810, 0x810, 0xE10, 0x81E, 0x810, 0x810, 0xFEFC, 0x44, 0x844, 0x844, 0x4A28, 0x4928, 0x8910, 0x828, 0x2844, 0x1082}},
    {25931, 2, {0x2040, 0x2040, 0x3C40, 0x4480, 0xC4FE, 0x2908, 0x1288, 0x2888, 0x4488, 0x8250, 0x7C50, 0x4420, 0x4450, 0x4488, 0x7D04, 0x4602}},
    {25932, 2, {0x420, 0xE20, 0x3820, 0x83E, 0x844, 0xFF44, 0x844, 0x8A4, 0x3E28, 0x2228, 0x2210, 0x2210, 0x2228, 0x3E48, 0x2284, 0x102}},
    {25933, 2, {0x810, 0x810, 0x1410, 0x221E, 0x4110, 0xBE10, 0x8FC, 0x844, 0x7F44, 0x844, 0x2A28, 0x2928, 0x4910, 0x8828, 0x2844, 0x1082}},
    {25934, 2, {0x240, 0x6440, 0x1840, 0x2480, 0xD2FE, 0x1108, 0xFE88, 0x2088, 0xFE88, 0x4450, 0x8850, 0xF20, 0xF850, 0x888, 0x2904, 0x1202}},
    {25935, 2, {0x4020, 0x4020, 0x7F20, 0x803E, 0x7E44, 0x4244, 0x5244, 0x4AA4, 0xFFA8, 0x4228, 0x9210, 0x8A10, 0xFF28, 0x248, 0x1484, 0x902}},
    {25936, 2, {0x20, 0x7F20, 0x4020, 0x403E, 0x5E44, 0x4044, 0x4044, 0x7FA4, 0x5028, 0x5128, 0x5A10, 0x5410, 0x5228, 0x9128, 0x9844, 0x1082}},
    {25937, 2, {0x1040, 0x1440, 0x1240, 0x1080, 0xFEFE, 0x1108, 0x1088, 0x9288, 0x5488, 0x1050, 0x3850, 0x5420, 0x9250, 0x1088, 0x5104, 0x2202}},
    {25938, 2, {0x1210, 0x1210, 0x1210, 0x2FA0, 0x2ABE, 0x6AC4, 0x6AA4, 0xAFA4, 0x2AA4, 0x2AA8, 0x2AA8, 0x2F90, 0x2228, 0x2228, 0x2244, 0x2282}},
    {25939, 2, {0x40, 0x2840, 0x2440, 0x4280, 0x80FE, 0x7D08, 0x4688, 0x4488, 0x4488, 0x7C50, 0x2850, 0x2820, 0x2850, 0x2A88, 0x4D04, 0x8A02}},
    {25940, 2, {0x20, 0xFF20, 0x1020, 0x103E, 0x7E44, 0x2244, 0x2244, 0xFFA4, 0x28, 0x28, 0x7E10, 0x4210, 0x4228, 0x4228, 0x7E44, 0x4282}},
    {25941, 2, {0x820, 0x820, 0xFFA0, 0x83E, 0x844, 0x7F44, 0x4944, 0x49A4, 0x7F28, 0x828, 0x1C10, 0x2A10, 0x4928, 0x8848, 0x884, 0x902}},
    {25942, 2, {0x820, 0x820, 0x820, 0x7F3E, 0x844, 0x3E44, 0x844, 0xFFA4, 0x1028, 0x1E28, 0x1210, 0x1210, 0x2228, 0x2A48, 0x4484, 0x8102}},
    {25943, 2, {0x40, 0x7C40, 0x4480, 0x44FE, 0x7D08, 0x4508, 0x4688, 0x7C88, 0x4488, 0x4450, 0x7C50, 0x20, 0x2850, 0x2488, 0x4504, 0x8202}},
    {25944, 2, {0x820, 0x820, 0x1420, 0x223E, 0x4144, 0xBE44, 0x844, 0x8A4, 0x7F28, 0x828, 0x2A10, 0x2910, 0x4928, 0x8848, 0x2884, 0x1102}},
    {25945, 2, {0x820, 0x820, 0x7EA0, 0x93E, 0xA44, 0xFF44, 0x844, 0x7EA4, 0x2428, 0x4828, 0x8F10, 0x7810, 0x828, 0x848, 0x2884, 0x1102}},
    {25946, 2, {0x4440, 0x2440, 0x2840, 0x80, 0x7CFE, 0x4508, 0x4688, 0x4488, 0x7C88, 0x2850, 0x2850, 0x2820, 0x2A50, 0x4C88, 0x8904, 0x202}},
    {25947, 2, {0x820, 0x820, 0x1420, 0x123E, 0x2144, 0x40C4, 0xBE44, 0xA4, 0x1128, 0x928, 0x4910, 0x2210, 0x2228, 0x748, 0x7884, 0x2102}},
    {25948, 2, {0x820, 0x820, 0x1420, 0x223E, 0x5144, 0x8844, 0x7E44, 0x2A4, 0x1428, 0x828, 0x2A10, 0xA110, 0xA128, 0xA528, 0x1C44, 0x82}},
    {25949, 2, {0x810, 0x4910, 0x2910, 0x2A20, 0x83E, 0x7F44, 0x4924, 0x4924, 0x5D24, 0x6B28, 0x4928, 0x4910, 0x4928, 0x4128, 0x4544, 0x4282}},
    {25950, 2, {0x820, 0x4920, 0x2920, 0x2A3E, 0x844, 0x7F44, 0x4144, 0x41A4, 0x5D28, 0x5528, 0x5510, 0x5510, 0x5D28, 0x4128, 0x4544, 0x4282}},
    {25951, 2, {0x1420, 0x1420, 0x1420, 0x7F3E, 0x5544, 0x5544, 0x5544, 0x7FA4, 0x5528, 0x5528, 0x5510, 0xFF90, 0x2428, 0x2248, 0x4284, 0x8102}},
    {25952, 2, {0x10, 0xF710, 0x1110, 0x551E, 0x2210, 0x5510, 0x89FC, 0x1044, 0x44, 0xF744, 0x1128, 0x5528, 0x2210, 0x5528, 0x8944, 0x1082}},
    {25953, 2, {0x10, 0x7E10, 0x4210, 0x421E, 0x7E10, 0x4210, 0x42FC, 0x7E44, 0x2044, 0x7F44, 0x9528, 0x2528, 0x4910, 0x9128, 0x2544, 0x4282}},
    {25954, 2, {0x20, 0x7E20, 0x220, 0x43E, 0xFF44, 0x2244, 0x2244, 0x3EA4, 0x2228, 0x3E28, 0x2210, 0x2210, 0x2F28, 0xF228, 0x4244, 0x282}},
    {25955, 2, {0x2420, 0x2420, 0x7E20, 0x243E, 0x2444, 0xFF44, 0x44, 0x7EA4, 0x4228, 0x7E28, 0x4210, 0x7E10, 0x4228, 0x4228, 0x4A44, 0x4482}},
    {25956, 2, {0x10, 0x7F10, 0x4910, 0x491E, 0x7F10, 0x4910, 0x49FC, 0x7F44, 0x844, 0xFF44, 0x1C28, 0x2A28, 0x4910, 0x8828, 0x844, 0x882}},
    {25957, 2, {0x810, 0x910, 0x4910, 0x4A1E, 0x8810, 0x1410, 0x22FC, 0x4944, 0x8844, 0x4944, 0x4928, 0x8A28, 0x1410, 0x2228, 0x4144, 0x8082}},
    {25958, 2, {0x1020, 0x820, 0xFF20, 0x3E, 0x7E44, 0x4244, 0x7E44, 0xA4, 0x7E28, 0x428, 0x810, 0xF10, 0xF828, 0x848, 0x2884, 0x1102}},
    {25959, 2, {0x810, 0x810, 0x7F1E, 0x1410, 0x2210, 0x41FC, 0xFF44, 0x244, 0x7A44, 0x4A28, 0x4A28, 0x7A10, 0x4A10, 0x228, 0xA44, 0x482}},
    {25960, 2, {0x1020, 0x820, 0x7F20, 0x3E, 0x2244, 0x1444, 0xFF44, 0xA4, 0x28, 0x3E28, 0x2210, 0x2210, 0x2228, 0x3E28, 0x2244, 0x82}},
    {25961, 2, {0x5110, 0x2910, 0x2A10, 0x20, 0xFFBE, 0x80C4, 0x124, 0x7E24, 0x424, 0x828, 0xF28, 0xF810, 0x828, 0x828, 0x2844, 0x1082}},
    {25962, 2, {0x20, 0xF720, 0x1120, 0x553E, 0x2244, 0x5544, 0x9944, 0xA4, 0xF728, 0x1128, 0x5510, 0x2210, 0x5528, 0x8948, 0x1084, 0x102}},
    {25963, 2, {0x820, 0x1020, 0x7E20, 0x423E, 0x7E44, 0x4244, 0x7E44, 0x10A4, 0x828, 0xFF28, 0x2010, 0x3E10, 0x2228, 0x4248, 0x4A84, 0x8502}},
    {25964, 2, {0x2220, 0x2220, 0xFFA0, 0x223E, 0x4044, 0x7F44, 0x4144, 0x81A4, 0x7D28, 0x4528, 0x4510, 0x7D10, 0x4528, 0x128, 0xA44, 0x482}},
    {25965, 2, {0x20, 0x7E20, 0x4220, 0x7E3E, 0x4244, 0x7E44, 0x44, 0xFFA4, 0x4028, 0x7F28, 0x9510, 0x2510, 0x4928, 0x9148, 0x2A84, 0x4502}},
    {25966, 2, {0x210, 0xF10, 0x7810, 0x81E, 0x810, 0xFF10, 0x28FC, 0x4B44, 0x4944, 0x4944, 0x6B28, 0x4928, 0x4910, 0x7F28, 0x4144, 0x82}},
    {25967, 2, {0x10, 0xFE10, 0x8210, 0xFE1E, 0x8810, 0xFF10, 0x88FC, 0xA544, 0xC344, 0x44, 0xFE28, 0x8228, 0xFE10, 0x8228, 0xFE44, 0x8282}},
    {25968, 2, {0x820, 0x4920, 0x2A20, 0x83E, 0xFF44, 0x2A44, 0x4944, 0x88A4, 0x1028, 0xFE28, 0x2210, 0x4210, 0x6428, 0x1828, 0x3444, 0xC282}},
    {25969, 2, {0x810, 0x4910, 0x4910, 0x7F1E, 0x10, 0xFF10, 0xFC, 0x7F44, 0x4144, 0x4144, 0x7F28, 0x4228, 0x2410, 0xF28, 0xF044, 0x4082}},
    {25970, 2, {0x1010, 0x810, 0xFF10, 0x1E, 0x7E10, 0x4210, 0x7EFC, 0x44, 0xFF44, 0x8144, 0xBD28, 0xA528, 0xBD10, 0x8128, 0x8544, 0x8282}},
    {25971, 2, {0x820, 0x4920, 0x4920, 0x7F3E, 0x44, 0xFF44, 0x44, 0x7FA4, 0x4128, 0x4128, 0x7F10, 0x4210, 0x2428, 0xF48, 0xF084, 0x4102}},
    {25972, 2, {0x820, 0x820, 0xFFBE, 0x844, 0x7FA4, 0x4924, 0x7F28, 0x2A10, 0x4928, 0x8846, 0x0, 0x7FFC, 0x100, 0x11F8, 0x1100, 0xFFFE}},
    {25973, 2, {0x1020, 0x820, 0x7F20, 0x223E, 0x1444, 0x7F44, 0x4144, 0x49A4, 0x7F28, 0x4928, 0x5D10, 0x5510, 0x5528, 0x5D28, 0x4144, 0x4382}},
    {25974, 2, {0x210, 0xE210, 0xBFD0, 0xA220, 0xAFBE, 0xCAC4, 0xAFA4, 0xAAA4, 0xAFA4, 0xA228, 0xA728, 0xCA90, 0x9268, 0x8228, 0x8244, 0x8282}},
    {25975, 2, {0xA20, 0x920, 0xFFA0, 0x83E, 0x7F44, 0x4944, 0x7F44, 0x49A4, 0x7F28, 0x4928, 0x1010, 0xFF90, 0x2028, 0x3E28, 0x4244, 0x8682}},
    {25976, 2, {0x810, 0x7F10, 0x4910, 0xFFA0, 0x493E, 0x7F44, 0x824, 0x7F24, 0x4924, 0x7F28, 0x1028, 0xFF10, 0x2128, 0x7228, 0xC44, 0xF382}},
    {25977, 2, {0x820, 0x820, 0xF20, 0x83E, 0xFF44, 0x8144, 0x644, 0x78A4, 0x4928, 0x2A28, 0xFF10, 0x1810, 0x2C28, 0x4A48, 0x8884, 0x902}},
    {25978, 2, {0x10, 0xFF90, 0x801E, 0x9F10, 0x9110, 0x91FC, 0x9F44, 0x8044, 0xBBC4, 0xAAA8, 0xAAA8, 0xAA90, 0xBB90, 0x8028, 0xFFC4, 0x82}},
    {25979, 2, {0x1000, 0x1FE0, 0x2020, 0x7FF8, 0xA308, 0x2488, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FF8, 0x2820, 0x7C0, 0xF83E}},
    {25980, 2, {0x810, 0xFF90, 0x810, 0x7F1E, 0x10, 0x7F10, 0x41FC, 0x7F44, 0x2244, 0xFFC4, 0x28, 0x7F28, 0x4110, 0x4128, 0x7F44, 0x4182}},
    {25981, 2, {0x610, 0x7810, 0x1010, 0xFF9E, 0x2210, 0x5D10, 0x94FC, 0x1C44, 0x44, 0x7F44, 0x4128, 0x5D28, 0x5510, 0x5D28, 0x4144, 0x4382}},
    {25982, 2, {0x2210, 0x1410, 0xFF90, 0x820, 0x7F3E, 0x844, 0xFFA4, 0x4924, 0x2A24, 0xFFA8, 0x28, 0x7F10, 0x4128, 0x4128, 0x7F44, 0x4182}},
    {25983, 2, {0x610, 0x7810, 0x1010, 0xFFA0, 0x223E, 0x5D44, 0x94A4, 0x1C24, 0x24, 0x7F28, 0x4128, 0x5D10, 0x5528, 0x5D28, 0x4144, 0x4382}},
    {25984, 2, {0x10, 0x7F10, 0x551E, 0x5510, 0x7F10, 0x20FC, 0x7F44, 0x9144, 0x7D44, 0x5528, 0x5528, 0x7D10, 0x1110, 0x7D28, 0x544, 0x282}},
    {25985, 2, {0x20, 0x7F20, 0x5520, 0x7F3E, 0x844, 0x7F44, 0x844, 0xFFA4, 0x2228, 0x1428, 0x7F10, 0x810, 0xFFA8, 0x848, 0x884, 0x902}},
    {25986, 2, {0x820, 0x820, 0x1420, 0x223E, 0x4144, 0xBE44, 0x44, 0xA4, 0x7728, 0x5528, 0x5510, 0x7710, 0x2228, 0x2228, 0x5544, 0x8882}},
    {25987, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x0, 0xFFFE, 0x1080, 0x3E98, 0x42E0, 0x2484, 0x1884, 0xE07C}},
    {25988, 2, {0x820, 0x3E3E, 0x844, 0x7FA8, 0x2A10, 0x4928, 0x844, 0x3FFE, 0x2080, 0x2FFC, 0x2490, 0x2490, 0x2AA8, 0x51C4, 0x46B0, 0x988E}},
    {25989, 2, {0x3410, 0x4B10, 0x5510, 0x631E, 0x5510, 0x6B10, 0x557C, 0xFFA4, 0x80A4, 0x7E24, 0x428, 0xF28, 0xF810, 0x828, 0x2844, 0x1082}},
    {25990, 2, {0x3410, 0x4B10, 0x5510, 0x6320, 0x553E, 0x6B44, 0x5524, 0xFFA4, 0x80A4, 0x7E28, 0x428, 0xF10, 0xF828, 0x828, 0x2844, 0x1082}},
    {25991, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x1010, 0x1010, 0x820, 0x820, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {25992, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE00E, 0xFE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {25993, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xF01E, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x2010, 0x4010}},
    {25994, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE00E, 0x3FF8, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20}},
    {25995, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x200, 0x400, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {25996, 2, {0x10, 0x2014, 0x11D2, 0x12, 0xFC10, 0xBFE, 0x810, 0x4890, 0x2890, 0x12D0, 0x1290, 0x2A90, 0x2A8A, 0x4AEA, 0x8706, 0x202}},
    {25997, 2, {0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x842, 0xF03E}},
    {25998, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE00E, 0x17D0, 0x1010, 0x17D0, 0x1110, 0x1590, 0x2550, 0x2910, 0x4010}},
    {25999, 2, {0x40, 0x2020, 0x11FC, 0x104, 0xFD04, 0x9FC, 0x904, 0x4904, 0x29FC, 0x1120, 0x1122, 0x2914, 0x2908, 0x4944, 0x8182, 0x100}},
    {26000, 2, {0x440, 0x7C7C, 0x440, 0x440, 0x3C78, 0x440, 0x440, 0x7C7C, 0x440, 0x100, 0xFFFE, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {26001, 2, {0x0, 0x200, 0xF93E, 0x2108, 0x2FC8, 0x2088, 0x2488, 0xFA88, 0x22BE, 0x2108, 0x2108, 0x3A88, 0xE288, 0x4448, 0x87E, 0x0}},
    {26002, 2, {0x40, 0x2020, 0x11FE, 0x102, 0xFD02, 0x9FE, 0x900, 0x4900, 0x29FE, 0x11AA, 0x11AA, 0x2AFE, 0x2AAA, 0x4AAA, 0x84A2, 0x86}},
    {26003, 2, {0x200, 0x417C, 0x2504, 0x444, 0xFDF4, 0x1444, 0x15F4, 0x1554, 0x9574, 0x55D4, 0x2554, 0x25F4, 0x54E4, 0x5554, 0x8444, 0x44C}},
    {26004, 2, {0x2020, 0x1020, 0xFEA0, 0x52C, 0x6924, 0x1124, 0x29AC, 0xD524, 0x1124, 0x7DFC, 0x5524, 0x7C50, 0x1050, 0x1C88, 0xE504, 0x4202}},
    {26005, 2, {0x3DE, 0x2252, 0x13DE, 0x252, 0xFFDE, 0xA22, 0xAFA, 0x4A22, 0x2AFA, 0x12AA, 0x12FA, 0x2AAA, 0x2AFA, 0x4A72, 0x82AA, 0x226}},
    {26006, 2, {0x100, 0x7FFC, 0x820, 0x7C0, 0x1830, 0xE7CE, 0x2448, 0x37D8, 0x2448, 0xFFFE, 0x9012, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {26007, 2, {0x40, 0x440, 0x240, 0x240, 0x1040, 0x840, 0x840, 0x40, 0xFE, 0x7F40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {26008, 2, {0x1010, 0x1010, 0x1090, 0x1E50, 0x1210, 0x2290, 0x2250, 0x5210, 0x8A1E, 0x5F0, 0x410, 0x810, 0x810, 0x1010, 0x2010, 0x4010}},
    {26009, 2, {0x808, 0x888, 0x4A48, 0x2A48, 0x2C08, 0x888, 0xFE48, 0x1848, 0x1C08, 0x2A0E, 0x2AF8, 0x4808, 0x8808, 0x808, 0x808, 0x808}},
    {26010, 2, {0x1020, 0x1020, 0x2448, 0x7EFC, 0x2244, 0x0, 0x7FFE, 0x4442, 0x8244, 0x1040, 0x840, 0x7E, 0xFFC0, 0x40, 0x40, 0x40}},
    {26011, 2, {0x1008, 0x1088, 0x3C48, 0x2448, 0x4808, 0xBE88, 0x2A48, 0x2A48, 0x3E08, 0x2A0E, 0x2AF8, 0x3E08, 0x2A08, 0x4A08, 0x4208, 0x8608}},
    {26012, 2, {0x808, 0x848, 0x1428, 0x2228, 0x4108, 0xBE48, 0x828, 0x828, 0x7F0E, 0x878, 0x2A08, 0x2908, 0x4908, 0x8808, 0x2808, 0x1008}},
    {26013, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x0, 0x7FFE, 0x4442, 0x8244, 0x1040, 0x840, 0x7E, 0xFFC0, 0x40, 0x40, 0x40}},
    {26014, 2, {0x808, 0x2848, 0x4B28, 0x4928, 0x4908, 0x6B48, 0x4928, 0x4928, 0x7F0E, 0x4978, 0x808, 0x1408, 0x1208, 0x2108, 0x4108, 0x8008}},
    {26015, 2, {0x2208, 0x2248, 0xFF28, 0x2228, 0x3E08, 0x2248, 0x3E28, 0x2228, 0x220E, 0xFF78, 0x4008, 0x5408, 0x6208, 0x4008, 0x7E08, 0x8}},
    {26016, 2, {0x2208, 0x2248, 0xFFA8, 0x2228, 0x7F08, 0x2248, 0xFFA8, 0x828, 0x7F0E, 0x4978, 0x7F08, 0x4908, 0xFFC8, 0x4108, 0x4508, 0x4208}},
    {26017, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1104, 0x7C12, 0x4490, 0x7C50, 0x4510, 0x7C90, 0x101E, 0xFEF0, 0x1010, 0x1010, 0x1010, 0x1010}},
    {26018, 2, {0x2208, 0x2248, 0x7F28, 0x2228, 0x2208, 0xFFC8, 0x828, 0x7F28, 0x490E, 0x7F78, 0x4908, 0x7F08, 0x8, 0x2208, 0x4108, 0x8088}},
    {26019, 2, {0x8, 0x7F48, 0x5528, 0x5528, 0x7F08, 0x2048, 0x7FA8, 0x90A8, 0x7C8E, 0x54F8, 0x5488, 0x7C88, 0x1288, 0xFE88, 0x288, 0x108}},
    {26020, 2, {0x10, 0xF8, 0x1F00, 0x1000, 0x1000, 0x1000, 0x1FFE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x2040, 0x2040, 0x4040, 0x8040}},
    {26021, 2, {0x10, 0xF8, 0x1F00, 0x1000, 0x1000, 0x1000, 0x1FFE, 0x1040, 0x1040, 0x11C0, 0x1060, 0x1050, 0x2048, 0x2040, 0x4040, 0x8040}},
    {26022, 2, {0x200, 0x704, 0x3878, 0x2040, 0x2040, 0x2040, 0x3F7E, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x4488, 0x4488, 0x8108}},
    {26023, 2, {0x440, 0x820, 0x1850, 0x2488, 0x300, 0xCC0, 0x3030, 0xC0EE, 0x1F00, 0x1000, 0x1000, 0x1FF8, 0x1080, 0x1080, 0x2080, 0x4080}},
    {26024, 2, {0x408, 0x241C, 0x24E0, 0x2480, 0x2480, 0x3C80, 0x4FE, 0x488, 0xFC88, 0x2488, 0x2488, 0x2488, 0x2508, 0x4508, 0x4608, 0x8408}},
    {26025, 2, {0x2008, 0x201C, 0x21E0, 0xFD00, 0x4100, 0x5100, 0x91FE, 0xFD10, 0x1110, 0x1110, 0x1D10, 0xF110, 0x5110, 0x1210, 0x1210, 0x1410}},
    {26026, 2, {0x2000, 0x2004, 0x4078, 0x7E40, 0x8240, 0x240, 0x7A7E, 0x4A48, 0x4A48, 0x4A48, 0x7A48, 0x4A48, 0x248, 0x288, 0x1488, 0x908}},
    {26027, 2, {0x8, 0x1C, 0xFCE0, 0x1080, 0x1080, 0x2080, 0x3CFE, 0x6488, 0x6488, 0xA488, 0x2488, 0x2488, 0x3D08, 0x2508, 0x2208, 0x408}},
    {26028, 2, {0x1008, 0x101C, 0x10E0, 0xFE80, 0x1080, 0x7C80, 0x54FE, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE90, 0x1090, 0x1110, 0x1110, 0x1210}},
    {26029, 2, {0x400, 0x404, 0x5578, 0x4E40, 0x4440, 0x7F40, 0x447E, 0x4E48, 0x5548, 0x6548, 0x4448, 0x4448, 0x4048, 0x7F88, 0x88, 0x108}},
    {26030, 2, {0x2400, 0x2404, 0x7E78, 0x2440, 0x2440, 0xFF40, 0x7E, 0x7E48, 0x4248, 0x4248, 0x7E48, 0x4248, 0x4248, 0x7E88, 0x4288, 0x108}},
    {26031, 2, {0x2200, 0x2204, 0x7F78, 0x2240, 0x2240, 0x3E40, 0x227E, 0x2248, 0x3E48, 0x2248, 0x2248, 0xFF48, 0x448, 0x2288, 0x4188, 0x8108}},
    {26032, 2, {0x1000, 0x804, 0x7F78, 0x40, 0x2240, 0x1440, 0xFF7E, 0x848, 0x848, 0x7F48, 0x848, 0x2A48, 0x4948, 0x8888, 0x2888, 0x1108}},
    {26033, 2, {0x1000, 0x1004, 0x7D78, 0x1240, 0x1440, 0xFF40, 0x107E, 0x2048, 0x7E48, 0xA248, 0x2248, 0x3E48, 0x2248, 0x2288, 0x3E88, 0x2308}},
    {26034, 2, {0x0, 0x7704, 0x5578, 0x5540, 0x7740, 0x1440, 0x147E, 0x7748, 0x4148, 0x4148, 0x7748, 0x1448, 0x1448, 0x1F88, 0xF088, 0x4108}},
    {26035, 2, {0x2200, 0x2202, 0xFFBC, 0x2220, 0x3E20, 0x820, 0x7F3E, 0x4924, 0x7F24, 0x824, 0x7F24, 0x824, 0x7F24, 0x844, 0xF44, 0xF084}},
    {26036, 2, {0x800, 0x4902, 0x2A3C, 0xFFA0, 0x2A20, 0x4920, 0x88BE, 0x4124, 0x4124, 0x77A4, 0x9124, 0x5524, 0x27A4, 0x4144, 0x8144, 0x184}},
    {26037, 2, {0x1000, 0x6F04, 0x4578, 0x5540, 0x6940, 0x4B40, 0x107E, 0xFFC8, 0x1448, 0x7748, 0x4148, 0x7748, 0x1448, 0x1788, 0xF888, 0x2108}},
    {26038, 2, {0x0, 0xDFC, 0x7154, 0x4154, 0x41FC, 0x4080, 0x7DFC, 0x4A44, 0x49F4, 0x4954, 0x4954, 0x49F4, 0x4844, 0x49F4, 0x4814, 0x8808}},
    {26039, 2, {0x882, 0x509C, 0x5D50, 0x49D0, 0x5490, 0x5DDE, 0x4454, 0x7FD4, 0x4894, 0x5094, 0x5D54, 0x49D4, 0x5494, 0x5DD4, 0x4464, 0x7FC4}},
    {26040, 2, {0x7FC2, 0x405C, 0x7FD0, 0x5250, 0x4A90, 0x525E, 0x5FD4, 0x5554, 0x5FD4, 0x6454, 0x5F54, 0x5554, 0x5F54, 0x4454, 0xBF64, 0x84}},
    {26041, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x400, 0x400, 0x400, 0x7F0, 0x410, 0x410, 0x410, 0x810, 0x810, 0x1010, 0x20A0, 0x4040}},
    {26042, 2, {0x2020, 0x1020, 0x50, 0xFE50, 0x2088, 0x2104, 0x3E02, 0x2488, 0x2488, 0x2488, 0x2488, 0x2488, 0x4488, 0x5508, 0x8908, 0x208}},
    {26043, 2, {0x2080, 0x1080, 0x80, 0xFEFC, 0x2100, 0x2200, 0x3CF0, 0x2490, 0x2490, 0x2490, 0x2490, 0x2492, 0x4492, 0x5512, 0x890E, 0x200}},
    {26044, 2, {0x2020, 0x1020, 0x20, 0xFE50, 0x2050, 0x2088, 0x3D04, 0x2602, 0x2460, 0x2410, 0x2400, 0x24C0, 0x4420, 0x5410, 0x8808, 0x0}},
    {26045, 2, {0x2080, 0x1080, 0x10FE, 0x120, 0xFE20, 0x2120, 0x212C, 0x3D74, 0x27A4, 0x2524, 0x2534, 0x252A, 0x2522, 0x4502, 0x54FE, 0x8800}},
    {26046, 2, {0x2080, 0x1080, 0x80, 0xFEFE, 0x2120, 0x2220, 0x3C20, 0x25FC, 0x2524, 0x2524, 0x2524, 0x2524, 0x4534, 0x5528, 0x8820, 0x20}},
    {26047, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE00, 0x20FC, 0x2004, 0x3C08, 0x2410, 0x25FE, 0x2410, 0x2410, 0x2410, 0x4410, 0x5450, 0x8820}},
    {26048, 2, {0x2020, 0x1020, 0x48, 0xFE84, 0x21FE, 0x2002, 0x3C20, 0x2420, 0x24A8, 0x24A4, 0x2524, 0x2522, 0x4622, 0x5420, 0x88A0, 0x40}},
    {26049, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0x7FFE, 0x4202, 0x8104, 0x3FF8, 0x400, 0x400, 0x7F0, 0x810, 0x810, 0x1050, 0x2020}},
    {26050, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE0C, 0x20F0, 0x2080, 0x3C80, 0x2480, 0x24FE, 0x2488, 0x2488, 0x2488, 0x4508, 0x5508, 0x8A08}},
    {26051, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE00, 0x20FC, 0x2084, 0x3CA4, 0x2494, 0x2484, 0x25FE, 0x2484, 0x2484, 0x4504, 0x5514, 0x8A08}},
    {26052, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE1C, 0x21E0, 0x2020, 0x3C3C, 0x25E0, 0x2420, 0x243E, 0x27E0, 0x2420, 0x4422, 0x5422, 0x881E}},
    {26053, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE0C, 0x20F0, 0x2090, 0x3C90, 0x2492, 0x2494, 0x2488, 0x2488, 0x2484, 0x44A4, 0x54C2, 0x8880}},
    {26054, 2, {0x2080, 0x1080, 0x10FE, 0x120, 0xFE20, 0x21FE, 0x2020, 0x3C20, 0x25FC, 0x2524, 0x2524, 0x2524, 0x2534, 0x4528, 0x5420, 0x8820}},
    {26055, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE10, 0x2010, 0x20FE, 0x3C92, 0x2490, 0x24FC, 0x24A4, 0x24A4, 0x24A8, 0x4510, 0x5528, 0x8A46}},
    {26056, 2, {0x2020, 0x1010, 0x1FE, 0xFE20, 0x2048, 0x2084, 0x3DFE, 0x2402, 0x24A8, 0x24A8, 0x24A8, 0x24A8, 0x44A8, 0x54AA, 0x892A, 0x206}},
    {26057, 2, {0x110, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x80, 0xFFFE, 0x800, 0xFF8, 0x1008, 0x2028, 0xC010}},
    {26058, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE00, 0x20FC, 0x2040, 0x3C40, 0x2478, 0x2448, 0x24A8, 0x2498, 0x2488, 0x44AA, 0x54CA, 0x888E}},
    {26059, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE00, 0x21FC, 0x2024, 0x3C24, 0x24A0, 0x24A0, 0x24BC, 0x24A0, 0x24A0, 0x4560, 0x553E, 0x8A00}},
    {26060, 2, {0x2080, 0x1080, 0x10FE, 0x110, 0xFE10, 0x2090, 0x20FE, 0x3C90, 0x2510, 0x2410, 0x24FE, 0x2410, 0x2410, 0x4410, 0x55FE, 0x8800}},
    {26061, 2, {0x2080, 0x1080, 0x10FE, 0x120, 0xFE20, 0x2050, 0x2088, 0x3D24, 0x2612, 0x2410, 0x25FC, 0x2404, 0x2488, 0x4450, 0x5420, 0x8810}},
    {26062, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE00, 0x20FC, 0x2084, 0x3C84, 0x24FC, 0x2480, 0x24A4, 0x24A8, 0x24B0, 0x4522, 0x5522, 0x8A1E}},
    {26063, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE80, 0x2080, 0x20FC, 0x3CA0, 0x2520, 0x2420, 0x27FE, 0x2420, 0x2450, 0x4450, 0x5488, 0x8906}},
    {26064, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFC50, 0x2050, 0x2254, 0x3D58, 0x2450, 0x24D8, 0x2554, 0x2652, 0x2450, 0x4450, 0x5492, 0x890E}},
    {26065, 2, {0x2020, 0x1020, 0x11FC, 0x50, 0xFC88, 0x2104, 0x23FE, 0x3C08, 0x25E8, 0x2528, 0x2528, 0x25E8, 0x2528, 0x4408, 0x5428, 0x8810}},
    {26066, 2, {0x2100, 0x1100, 0x11FE, 0x240, 0xFC20, 0x21FE, 0x2040, 0x3C84, 0x25FE, 0x2402, 0x24A8, 0x24A8, 0x24A8, 0x44AA, 0x552A, 0x8A26}},
    {26067, 2, {0x2100, 0x1100, 0x11FE, 0x220, 0xFD24, 0x20A8, 0x2020, 0x3DFC, 0x2504, 0x25FC, 0x2504, 0x25FC, 0x2504, 0x4504, 0x5514, 0x8908}},
    {26068, 2, {0x4010, 0x2010, 0x7C, 0xF714, 0x41FE, 0x4114, 0x727C, 0x5210, 0x577C, 0x5110, 0x51FE, 0x5510, 0x5210, 0x5300, 0xB4FE, 0x800}},
    {26069, 2, {0x1010, 0x810, 0xFF28, 0x2044, 0x3E82, 0x2210, 0x2208, 0x4220, 0x4A10, 0x8408, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {26070, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFC20, 0x21FC, 0x2050, 0x3C88, 0x27FE, 0x2408, 0x25E8, 0x2528, 0x25E8, 0x4408, 0x5428, 0x8810}},
    {26071, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFC88, 0x21FC, 0x2088, 0x3CF8, 0x2488, 0x24F8, 0x2488, 0x2488, 0x27FE, 0x4450, 0x5488, 0x8904}},
    {26072, 2, {0x2110, 0x1094, 0x13D2, 0x12, 0xFA50, 0x2190, 0x27FE, 0x3810, 0x2BD2, 0x2A52, 0x2A54, 0x2BD4, 0x2A4A, 0x4A4A, 0x4BD6, 0x9822}},
    {26073, 2, {0x203C, 0x13E0, 0x1124, 0xA8, 0xFDFE, 0x20A8, 0x2124, 0x3E02, 0x25FC, 0x2524, 0x2524, 0x25FC, 0x2524, 0x4524, 0x55FC, 0x8904}},
    {26074, 2, {0x2100, 0x11FE, 0x1200, 0x1FE, 0xFC50, 0x21FC, 0x2154, 0x3DFC, 0x2400, 0x25FC, 0x2400, 0x27FE, 0x2420, 0x44A8, 0x5524, 0x8A64}},
    {26075, 2, {0x2100, 0x1100, 0x11FE, 0x200, 0xFC1C, 0x23E0, 0x2124, 0x3CA8, 0x27FE, 0x2524, 0x2602, 0x25FC, 0x2524, 0x45FC, 0x5524, 0x89FC}},
    {26076, 2, {0x2100, 0x11FE, 0x1220, 0x4FC, 0xFC00, 0x21FE, 0x2102, 0x3D7A, 0x254A, 0x25FE, 0x2484, 0x24FC, 0x2484, 0x44FC, 0x5400, 0x89FE}},
    {26077, 2, {0x2080, 0x10FE, 0x1130, 0x248, 0xFCB4, 0x2102, 0x21FE, 0x3DAA, 0x2572, 0x25FE, 0x2400, 0x24FC, 0x2484, 0x44FC, 0x5484, 0x88FC}},
    {26078, 2, {0x2080, 0x1080, 0x10FE, 0x100, 0xFE44, 0x2028, 0x22FE, 0x3D20, 0x2454, 0x2728, 0x255C, 0x252A, 0x254A, 0x4528, 0x5690, 0x8C7E}},
    {26079, 2, {0x2100, 0x1100, 0x1FE, 0xFE00, 0x2540, 0x224C, 0x3A64, 0x2B4C, 0x2A64, 0x2AA4, 0x2BAC, 0x2AA4, 0x4FFE, 0x6890, 0x9108, 0x204}},
    {26080, 2, {0x0, 0x3FF0, 0x200, 0x200, 0x200, 0x200, 0x7FFC, 0x480, 0x480, 0x480, 0x880, 0x880, 0x1084, 0x2084, 0x407C, 0x8000}},
    {26081, 2, {0x0, 0x3FF0, 0x200, 0x200, 0x4200, 0x4200, 0x7FFC, 0x480, 0x480, 0x480, 0x880, 0x880, 0x1084, 0x2084, 0x407C, 0x8000}},
    {26082, 2, {0x0, 0x7DFC, 0x4420, 0x4420, 0x7D20, 0x4520, 0x45FE, 0x7C20, 0x4020, 0x4830, 0x4450, 0x4A50, 0x5292, 0x6092, 0x410E, 0x200}},
    {26083, 2, {0x1000, 0x21FC, 0x7C20, 0x4420, 0x7D20, 0x4520, 0x7DFE, 0x20, 0x4020, 0x4030, 0x7C50, 0x4050, 0x4092, 0x4C92, 0x710E, 0x200}},
    {26084, 2, {0x0, 0x7DFC, 0x4420, 0x4420, 0x7520, 0x5520, 0x55FE, 0xFE20, 0x8220, 0xBA30, 0xAA50, 0xAA50, 0xBA92, 0x8292, 0x8B0E, 0x8600}},
    {26085, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26086, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0xFFFE, 0x0}},
    {26087, 2, {0x2000, 0x23FC, 0x2204, 0x2204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x2204, 0x2204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x2204, 0x2000}},
    {26088, 2, {0x2000, 0x2070, 0x2380, 0x3C08, 0x2008, 0x2008, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26089, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {26090, 2, {0x20, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {26091, 2, {0x0, 0x0, 0x7DFC, 0x4444, 0x4444, 0x4444, 0x4444, 0x7C44, 0x4444, 0x4444, 0x4444, 0x4484, 0x7C84, 0x4504, 0x228, 0x410}},
    {26092, 2, {0x800, 0x800, 0xFFC, 0x1004, 0x1004, 0x2FC4, 0x4844, 0x8844, 0xFC4, 0x844, 0x844, 0xFC4, 0x844, 0x4, 0x28, 0x10}},
    {26093, 2, {0x2000, 0x20F8, 0x2088, 0x2088, 0xFC88, 0x24F8, 0x2488, 0x2488, 0x2488, 0x2488, 0x24F8, 0x248A, 0x2402, 0x4402, 0x43FE, 0x8000}},
    {26094, 2, {0x200, 0x200, 0x3FE0, 0x420, 0x822, 0x3022, 0xC01E, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26095, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x200, 0x7FE0, 0x420, 0x420, 0x822, 0x1022, 0x601E}},
    {26096, 2, {0x0, 0xFC, 0x7C20, 0x4420, 0x4420, 0x4420, 0x4420, 0x7DFE, 0x4420, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {26097, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {26098, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {26099, 2, {0x40, 0x40, 0x7C80, 0x44FC, 0x4504, 0x4604, 0x4404, 0x7D04, 0x4484, 0x4444, 0x4444, 0x4404, 0x7C04, 0x4404, 0x28, 0x10}},
    {26100, 2, {0x0, 0x1FC, 0x7C20, 0x4420, 0x4420, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0xA0, 0x40}},
    {26101, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {26102, 2, {0x8, 0x8, 0x7C08, 0x4408, 0x45FE, 0x4408, 0x4408, 0x7C08, 0x4488, 0x4448, 0x4448, 0x4408, 0x7C08, 0x4408, 0x28, 0x10}},
    {26103, 2, {0x40, 0x20, 0x7820, 0x49FE, 0x4900, 0x4900, 0x7900, 0x4900, 0x4900, 0x4900, 0x4900, 0x7900, 0x4900, 0x200, 0x200, 0x400}},
    {26104, 2, {0x0, 0x1F8, 0x7C10, 0x4420, 0x4440, 0x4480, 0x45FE, 0x7C92, 0x4492, 0x4492, 0x4512, 0x4522, 0x7E22, 0x4442, 0x94, 0x108}},
    {26105, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26106, 2, {0x0, 0x7FC, 0x7840, 0x4840, 0x4840, 0x4840, 0x7840, 0x4BFC, 0x4840, 0x4840, 0x4840, 0x7840, 0x4840, 0x40, 0xFFE, 0x0}},
    {26107, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x440, 0x1830, 0xE00E}},
    {26108, 2, {0x40, 0x20, 0x7C20, 0x47FE, 0x4488, 0x4488, 0x4488, 0x7C88, 0x4488, 0x4450, 0x4450, 0x4420, 0x7C50, 0x4488, 0x104, 0x602}},
    {26109, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4924, 0x4924, 0x7924, 0x4924, 0x49FC, 0x4824, 0x4820, 0x7822, 0x4822, 0x1E, 0x0}},
    {26110, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x2108, 0x3FF8, 0x102, 0x102, 0xFE, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {26111, 2, {0x80, 0x80, 0x7C80, 0x45FC, 0x4520, 0x4620, 0x4420, 0x7C20, 0x47FE, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {26112, 2, {0x40, 0x40, 0x7C80, 0x44FC, 0x4504, 0x4604, 0x4484, 0x7C44, 0x4444, 0x4414, 0x4424, 0x4444, 0x7D84, 0x4404, 0x28, 0x10}},
    {26113, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x49FC, 0x7924, 0x4924, 0x4924, 0x4924, 0x4934, 0x7928, 0x4820, 0x20, 0x20}},
    {26114, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0xC00, 0x70FC, 0x4084, 0x4084, 0x4084, 0x4C84, 0x7094, 0x4088, 0x80, 0x80}},
    {26115, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FFC, 0x2080, 0x2080, 0x2140, 0x2140, 0x2220, 0x4410, 0x4808, 0x9006}},
    {26116, 2, {0x8, 0x3C, 0x7BC0, 0x4A00, 0x4A00, 0x4A00, 0x7BF8, 0x4A88, 0x4A88, 0x4A88, 0x4A50, 0x7A50, 0x4A20, 0x450, 0x488, 0x906}},
    {26117, 2, {0x0, 0x3FC, 0x7884, 0x4888, 0x4888, 0x4890, 0x489C, 0x7884, 0x4944, 0x4944, 0x4928, 0x4928, 0x7A10, 0x4A28, 0x444, 0x182}},
    {26118, 2, {0x0, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x0, 0x2080, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {26119, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x720, 0x7820, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x1020, 0x2020}},
    {26120, 2, {0x20, 0x10, 0x7C10, 0x44FE, 0x4482, 0x4482, 0x4482, 0x7CFE, 0x4482, 0x4480, 0x4480, 0x4480, 0x7C80, 0x4500, 0x100, 0x200}},
    {26121, 2, {0x40, 0x20, 0x7C20, 0x47FE, 0x4480, 0x4480, 0x4480, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x4484, 0x7D04, 0x4504, 0x228, 0x410}},
    {26122, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {26123, 2, {0x3FF8, 0x100, 0x200, 0xFFFE, 0x440, 0x820, 0x3018, 0xC006, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26124, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {26125, 2, {0x0, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x4504, 0x4504, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x104, 0x0}},
    {26126, 2, {0x0, 0xFC, 0x7C84, 0x4484, 0x4484, 0x44FC, 0x7C84, 0x4484, 0x4484, 0x44FC, 0x7C84, 0x4484, 0x104, 0x104, 0x214, 0x408}},
    {26127, 2, {0x78, 0x3F80, 0x2080, 0x2080, 0x3FFC, 0x2040, 0x2822, 0x301A, 0x2006, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0}},
    {26128, 2, {0x10, 0x90, 0x7C90, 0x4488, 0x4508, 0x4504, 0x4604, 0x7DFA, 0x4488, 0x4488, 0x4488, 0x4488, 0x7D08, 0x4508, 0x228, 0x410}},
    {26129, 2, {0x20, 0x20, 0x7C50, 0x4450, 0x4488, 0x4544, 0x4622, 0x7C20, 0x4400, 0x45FC, 0x4404, 0x4408, 0x7C08, 0x4410, 0x10, 0x20}},
    {26130, 2, {0x100, 0x100, 0x7900, 0x49FC, 0x4AA4, 0x4CA4, 0x48A4, 0x7924, 0x4924, 0x4A44, 0x4C44, 0x4884, 0x7904, 0x4A04, 0x428, 0x10}},
    {26131, 2, {0xFF0, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0xFF0, 0x400, 0x800, 0x1FFC, 0x2244, 0x4244, 0x484, 0x884, 0x1128, 0x210}},
    {26132, 2, {0x440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26133, 2, {0x4, 0xE, 0x7CF0, 0x4480, 0x4480, 0x4480, 0x44FE, 0x7C88, 0x4488, 0x4488, 0x4488, 0x4488, 0x7C88, 0x4508, 0x108, 0x208}},
    {26134, 2, {0x10, 0x90, 0x7890, 0x4888, 0x4888, 0x4924, 0x4924, 0x7A22, 0x4C40, 0x4840, 0x4888, 0x4884, 0x7904, 0x4BFE, 0x102, 0x0}},
    {26135, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x100, 0x7FFC, 0x0, 0x440, 0x820, 0x1010, 0x2008}},
    {26136, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x100, 0xFFFE, 0x400, 0x7F0, 0x410, 0x810, 0x1010, 0x20A0, 0x4040}},
    {26137, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {26138, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x2288, 0x1450, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26139, 2, {0x0, 0x1FE, 0x7D00, 0x4500, 0x4500, 0x45FC, 0x4504, 0x7D04, 0x4504, 0x4504, 0x45FC, 0x4500, 0x7D00, 0x4500, 0x1FE, 0x0}},
    {26140, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2248, 0x4448, 0x888, 0x1108, 0x2228, 0x410}},
    {26141, 2, {0x1040, 0x1040, 0x3E50, 0x2248, 0x5444, 0x840, 0x17FE, 0x2000, 0xDFF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26142, 2, {0x0, 0x3FE, 0x7820, 0x4820, 0x4820, 0x4BFE, 0x7A22, 0x4A22, 0x4A52, 0x4A4A, 0x4A8A, 0x7B02, 0x4A02, 0x202, 0x20A, 0x204}},
    {26143, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0}},
    {26144, 2, {0x20, 0x20, 0x7820, 0x49FC, 0x4924, 0x4924, 0x7924, 0x4924, 0x4924, 0x4BFE, 0x4820, 0x7850, 0x4850, 0x88, 0x104, 0x202}},
    {26145, 2, {0x20, 0x10, 0x7C10, 0x45FE, 0x4420, 0x4420, 0x4444, 0x7C84, 0x45F8, 0x4410, 0x4420, 0x4444, 0x7C82, 0x45FE, 0x82, 0x0}},
    {26146, 2, {0x20, 0x20, 0x7D24, 0x4524, 0x4524, 0x4524, 0x45FC, 0x7C20, 0x4420, 0x4524, 0x4524, 0x4524, 0x7D24, 0x4524, 0x1FC, 0x4}},
    {26147, 2, {0x20, 0x20, 0x7C50, 0x4488, 0x4504, 0x4612, 0x4420, 0x7C40, 0x4588, 0x4410, 0x4420, 0x4444, 0x7D88, 0x4410, 0x60, 0x380}},
    {26148, 2, {0x20, 0x20, 0x7C50, 0x4450, 0x4488, 0x4524, 0x4612, 0x7C10, 0x45FC, 0x4404, 0x4408, 0x4488, 0x7C50, 0x4420, 0x10, 0x10}},
    {26149, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1010, 0x2FE8, 0xC826, 0x820, 0xFE0, 0x820, 0x820, 0xFE0}},
    {26150, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x440, 0x820, 0x3458, 0xC446, 0x440, 0x440, 0x840, 0x1040}},
    {26151, 2, {0x20, 0x20, 0x7820, 0x49FC, 0x4820, 0x4820, 0x4820, 0x7BFE, 0x4870, 0x48A8, 0x48A8, 0x4924, 0x7924, 0x4A22, 0x20, 0x20}},
    {26152, 2, {0x100, 0x100, 0x7900, 0x49FE, 0x4A80, 0x4A80, 0x4C80, 0x78F8, 0x4880, 0x4880, 0x4880, 0x48FC, 0x7880, 0x4880, 0x80, 0x80}},
    {26153, 2, {0x40, 0x40, 0x7840, 0x4FFE, 0x4840, 0x4840, 0x4840, 0x7BFC, 0x4840, 0x48E0, 0x4950, 0x4A48, 0x7C46, 0x4840, 0x40, 0x40}},
    {26154, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010, 0x3FF8, 0x828, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {26155, 2, {0x80, 0x80, 0x7900, 0x49FC, 0x4A04, 0x4C04, 0x49E4, 0x7924, 0x4924, 0x4924, 0x49E4, 0x4924, 0x7804, 0x4804, 0x28, 0x10}},
    {26156, 2, {0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2080, 0x3FFC, 0x2040, 0x2C24, 0x3014, 0x200C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26157, 2, {0x0, 0x3FC, 0x7884, 0x4884, 0x4884, 0x4904, 0x4914, 0x7A08, 0x4C00, 0x49FC, 0x4904, 0x4904, 0x7904, 0x4904, 0x1FC, 0x104}},
    {26158, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x40, 0x40, 0x7DFC, 0x1044, 0x1044, 0x1084, 0x1C84, 0xE104, 0x4228, 0x410}},
    {26159, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x100, 0x1100, 0x11F8, 0x1100, 0x2900, 0x4500, 0x83FE}},
    {26160, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {26161, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x820, 0x440, 0xFFFE}},
    {26162, 2, {0x90, 0x90, 0x7890, 0x4BFC, 0x4894, 0x4894, 0x7BFC, 0x4A90, 0x4A90, 0x4BFE, 0x4892, 0x7892, 0x491A, 0x114, 0x210, 0x410}},
    {26163, 2, {0x40, 0x40, 0x7A40, 0x4A40, 0x4BFC, 0x4A40, 0x4C40, 0x7840, 0x4FFE, 0x4840, 0x48A0, 0x48A0, 0x7910, 0x4A08, 0x404, 0x802}},
    {26164, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xC00, 0x70FC, 0x4484, 0x4484, 0x5C84, 0x6494, 0x4888, 0x1080, 0x6080}},
    {26165, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x4904, 0x49FC, 0x4900, 0x7940, 0x4944, 0x4948, 0x4970, 0x4940, 0x7942, 0x4A42, 0x23E, 0x400}},
    {26166, 2, {0x800, 0x400, 0x27C, 0x3C44, 0x444, 0x4C4, 0x757C, 0x1644, 0x1644, 0x1544, 0x257C, 0x2480, 0x4440, 0x8430, 0x140E, 0x800}},
    {26167, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {26168, 2, {0x40, 0x40, 0x7C7C, 0x4484, 0x4588, 0x4650, 0x4420, 0x7C50, 0x4488, 0x4706, 0x4460, 0x4410, 0x7C08, 0x44C0, 0x20, 0x10}},
    {26169, 2, {0x40, 0x20, 0x7810, 0x49E0, 0x4822, 0x4834, 0x7BB8, 0x48B0, 0x48A8, 0x48A8, 0x4928, 0x7924, 0x4A24, 0x422, 0xA0, 0x40}},
    {26170, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x2018}},
    {26171, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7CFC, 0x1084, 0x1084, 0x1084, 0x1084, 0x1E94, 0xF088, 0x4080, 0x80}},
    {26172, 2, {0x0, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1040, 0x1020, 0x2010, 0x2FE8, 0x4826, 0x8FE0, 0x820, 0xFE0, 0x0, 0x7FFC, 0x0}},
    {26173, 2, {0x90, 0x88, 0x7888, 0x4880, 0x4BFE, 0x48A0, 0x48A0, 0x78A4, 0x48A4, 0x4928, 0x4928, 0x4932, 0x7A22, 0x4A62, 0x49E, 0x800}},
    {26174, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x440, 0x4444, 0x2444, 0x1448, 0x1450, 0x440, 0xFFFE, 0x0}},
    {26175, 2, {0x20, 0x10, 0x7810, 0x49FE, 0x4900, 0x4910, 0x7910, 0x4910, 0x4910, 0x4920, 0x4928, 0x7924, 0x4A44, 0x2FE, 0x442, 0x800}},
    {26176, 2, {0x90, 0x90, 0x7890, 0x4A92, 0x4994, 0x4898, 0x4890, 0x7998, 0x4A94, 0x4C92, 0x4890, 0x4890, 0x7912, 0x4912, 0x212, 0x40E}},
    {26177, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x2448, 0x1450, 0xC60, 0x1450, 0x6448, 0x842, 0x1042, 0x603E}},
    {26178, 2, {0x20, 0x20, 0x7C20, 0x45FC, 0x4420, 0x4420, 0x47FE, 0x7C08, 0x4408, 0x45FE, 0x4408, 0x4488, 0x7C48, 0x4408, 0x28, 0x10}},
    {26179, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x1110, 0x920, 0x100, 0xFFFE, 0x440, 0x440, 0x842, 0x1042, 0x603E}},
    {26180, 2, {0x40, 0x44, 0x7A44, 0x4948, 0x4950, 0x4840, 0x4FFE, 0x7890, 0x4890, 0x4890, 0x4890, 0x4892, 0x7912, 0x4912, 0x20E, 0x400}},
    {26181, 2, {0x0, 0x3FE, 0x7800, 0x4800, 0x49FC, 0x4904, 0x4904, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x4904, 0x7800, 0x4800, 0x3FE, 0x0}},
    {26182, 2, {0x20, 0x20, 0x7DFC, 0x4420, 0x4420, 0x4420, 0x47FE, 0x7C00, 0x4420, 0x4420, 0x45FC, 0x4420, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {26183, 2, {0x20, 0x20, 0x7BFE, 0x4850, 0x4888, 0x4904, 0x7AFA, 0x4800, 0x4BFE, 0x4840, 0x4880, 0x79FC, 0x4804, 0x4, 0x28, 0x10}},
    {26184, 2, {0x40, 0x20, 0x7C20, 0x47FE, 0x4400, 0x4488, 0x4504, 0x7E02, 0x4488, 0x4488, 0x4450, 0x4450, 0x7C20, 0x4450, 0x88, 0x306}},
    {26185, 2, {0x0, 0x7FFC, 0x810, 0x1020, 0x2448, 0x7EFC, 0x204, 0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26186, 2, {0x0, 0x3FE, 0x7840, 0x4880, 0x4910, 0x4A08, 0x4BFC, 0x7844, 0x4840, 0x4840, 0x4BFC, 0x4840, 0x7840, 0x4840, 0x7FE, 0x0}},
    {26187, 2, {0x0, 0x7FFC, 0x440, 0x2448, 0x1450, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26188, 2, {0x40, 0x40, 0x7880, 0x4BFC, 0x4A04, 0x4A04, 0x4AF4, 0x7A94, 0x4A94, 0x4A94, 0x4A94, 0x4AF4, 0x7A04, 0x4A04, 0x214, 0x208}},
    {26189, 2, {0x0, 0x1FE, 0x7D02, 0x4502, 0x457A, 0x4502, 0x4502, 0x7D7A, 0x454A, 0x454A, 0x454A, 0x457A, 0x7D02, 0x4502, 0x10A, 0x104}},
    {26190, 2, {0x110, 0x110, 0x7910, 0x4910, 0x4BFC, 0x4910, 0x7910, 0x4910, 0x4910, 0x4FFE, 0x4800, 0x7910, 0x4908, 0x208, 0x404, 0x804}},
    {26191, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x7FFC, 0x4204, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {26192, 2, {0x20, 0x10, 0x7810, 0x4BFE, 0x4820, 0x4842, 0x4884, 0x79F8, 0x4810, 0x4822, 0x48C4, 0x4B08, 0x7810, 0x4828, 0xC4, 0x302}},
    {26193, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x400, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2028, 0x2010}},
    {26194, 2, {0x0, 0x7FE, 0x7890, 0x4890, 0x4890, 0x4BFC, 0x4A94, 0x7A94, 0x4A94, 0x4A94, 0x4A9C, 0x4B04, 0x7A04, 0x4A04, 0x3FC, 0x204}},
    {26195, 2, {0x80, 0x80, 0x78BC, 0x4BC0, 0x4850, 0x4824, 0x48D4, 0x7B0C, 0x4800, 0x4BFE, 0x4890, 0x4890, 0x7912, 0x4912, 0x20E, 0x400}},
    {26196, 2, {0x90, 0x94, 0x7914, 0x4918, 0x4B10, 0x4D32, 0x4952, 0x790E, 0x4920, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20}},
    {26197, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4202, 0x9FF4, 0x400, 0x900, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100}},
    {26198, 2, {0x0, 0x3FE, 0x7A02, 0x4C44, 0x4840, 0x4BFC, 0x4880, 0x78A0, 0x4920, 0x49FC, 0x4820, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {26199, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A88, 0x4C46, 0x49F0, 0x7810, 0x4820, 0x4840, 0x4BF8, 0x4A08, 0x7A08, 0x4A08, 0x3F8, 0x208}},
    {26200, 2, {0x0, 0x1FC, 0x7D04, 0x45FC, 0x4504, 0x45FC, 0x4400, 0x7C00, 0x45FC, 0x4420, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x20, 0x20}},
    {26201, 2, {0x40, 0x40, 0x7888, 0x4904, 0x4BFE, 0x4802, 0x4888, 0x7944, 0x4A42, 0x48F8, 0x4988, 0x4A50, 0x7820, 0x4850, 0x188, 0x606}},
    {26202, 2, {0x80, 0x80, 0x79F8, 0x4A08, 0x4C10, 0x4BFC, 0x4A44, 0x7A44, 0x4A44, 0x4BFC, 0x48A0, 0x48A0, 0x7922, 0x4922, 0x21E, 0x400}},
    {26203, 2, {0x0, 0x3F8, 0x7A08, 0x4A08, 0x4BF8, 0x4A08, 0x7A08, 0x4BF8, 0x4A08, 0x4A08, 0x4BF8, 0x7920, 0x4920, 0x222, 0x422, 0x181E}},
    {26204, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7FF8, 0x108, 0x3FF8, 0x2100, 0x3FFC, 0x304, 0x514, 0x1908, 0x6100}},
    {26205, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0xFFFE}},
    {26206, 2, {0x184, 0x68, 0x7830, 0x48C8, 0x4B04, 0x4840, 0x7BFE, 0x48A0, 0x4920, 0x4BFC, 0x4D24, 0x7924, 0x4934, 0x128, 0x20, 0x20}},
    {26207, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3E88, 0x2250, 0x2250, 0x2A24, 0x4454, 0x418C, 0x8604}},
    {26208, 2, {0x28, 0x24, 0x7824, 0x4820, 0x4BFE, 0x4A20, 0x7A24, 0x4A24, 0x4BA4, 0x4AA8, 0x4AA8, 0x7A90, 0x4A92, 0x5AA, 0x446, 0x882}},
    {26209, 2, {0x48, 0x44, 0x7BFE, 0x4840, 0x4840, 0x4BFC, 0x4A44, 0x7A44, 0x4BFC, 0x4A44, 0x4A44, 0x4BFC, 0x7A44, 0x4A44, 0x254, 0x208}},
    {26210, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26211, 2, {0x104, 0x10E, 0x7930, 0x4920, 0x4FE0, 0x4920, 0x793E, 0x4964, 0x49A4, 0x4F24, 0x4924, 0x7924, 0x4924, 0x124, 0x544, 0x284}},
    {26212, 2, {0x0, 0x3FE, 0x7840, 0x4840, 0x49FC, 0x4884, 0x4884, 0x7BFE, 0x4800, 0x4800, 0x49FC, 0x4904, 0x7904, 0x4904, 0x1FC, 0x104}},
    {26213, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x49F8, 0x7800, 0x4800, 0x4BFE, 0x4890, 0x4890, 0x7890, 0x4912, 0x112, 0x20E, 0x400}},
    {26214, 2, {0x100, 0x100, 0x79FC, 0x4A00, 0x4DF8, 0x4908, 0x4948, 0x7928, 0x4FFE, 0x4908, 0x4A48, 0x4A28, 0x7BFC, 0x4808, 0x50, 0x20}},
    {26215, 2, {0x20, 0x120, 0x7920, 0x49FC, 0x4920, 0x4A20, 0x7820, 0x4BFE, 0x4800, 0x4800, 0x49FC, 0x7904, 0x4904, 0x104, 0x1FC, 0x104}},
    {26216, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2908, 0x2890, 0x4860, 0x4A18, 0x8C06}},
    {26217, 2, {0x80, 0x80, 0x79F8, 0x4A08, 0x4C10, 0x4BFC, 0x4A44, 0x7A44, 0x4A44, 0x4BFC, 0x48A0, 0x48A0, 0x7922, 0x4922, 0x21E, 0x400}},
    {26218, 2, {0x90, 0x90, 0x7890, 0x4BFC, 0x4A94, 0x4A94, 0x4A94, 0x7BFC, 0x4A94, 0x4A94, 0x4A94, 0x4FFE, 0x7800, 0x4890, 0x108, 0x204}},
    {26219, 2, {0x40, 0x40, 0x787E, 0x4840, 0x4BFC, 0x4A04, 0x4BFC, 0x7A04, 0x4BFC, 0x4A44, 0x4840, 0x4FFE, 0x7840, 0x4840, 0x40, 0x40}},
    {26220, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x4888, 0x4888, 0x4954, 0x7A22, 0x4800, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20}},
    {26221, 2, {0x0, 0x3FE, 0x7A22, 0x4A22, 0x4AFA, 0x4A22, 0x7A22, 0x4BFE, 0x4A02, 0x4AFA, 0x4A8A, 0x7A8A, 0x4AFA, 0x202, 0x20A, 0x404}},
    {26222, 2, {0x820, 0x440, 0x7FFC, 0x440, 0x2448, 0x1450, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26223, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x2108, 0x4504, 0x200}},
    {26224, 2, {0x200, 0x20C, 0xF270, 0x9240, 0x9F40, 0x9240, 0x927E, 0xF648, 0x9748, 0x9AC8, 0x9A48, 0x9248, 0xF248, 0x9248, 0x248, 0x288}},
    {26225, 2, {0x20, 0x122, 0x7922, 0x4A24, 0x4850, 0x4888, 0x7B04, 0x4822, 0x4820, 0x4924, 0x4924, 0x7A28, 0x4850, 0x88, 0x104, 0x602}},
    {26226, 2, {0x0, 0x40, 0x7B9C, 0x4A04, 0x4A04, 0x4B9C, 0x7A04, 0x4A04, 0x4BFC, 0x4890, 0x4890, 0x7890, 0x4890, 0x112, 0x212, 0x40E}},
    {26227, 2, {0x80C, 0x8F0, 0x7E80, 0x1880, 0x2CFE, 0x2A88, 0x4888, 0x8888, 0x908, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0}},
    {26228, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x7800, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x114, 0x108}},
    {26229, 2, {0x1040, 0x840, 0x3E7E, 0x2288, 0x2350, 0x3E20, 0x2058, 0x4186, 0x9FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26230, 2, {0xFE0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0xFE0, 0x0, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x4284}},
    {26231, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x840, 0x1F60, 0x2250, 0x5448, 0x840, 0x37FE, 0xC000, 0x1FF0, 0x1010, 0x1010, 0x1FF0}},
    {26232, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x20, 0x7F20, 0x43E, 0x444, 0x27A4, 0x2428, 0x2428, 0x2410, 0x2F28, 0xF044, 0x4082}},
    {26233, 2, {0x1FC, 0x104, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x7800, 0x4900, 0x49FE, 0x4A4A, 0x4C4A, 0x7892, 0x4912, 0x222, 0x4A, 0x184}},
    {26234, 2, {0x2000, 0x3E7C, 0x4844, 0x844, 0xFF44, 0x1444, 0x227C, 0x4000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26235, 2, {0x40, 0x80, 0x7BFC, 0x4910, 0x4A48, 0x4C46, 0x4BF8, 0x7A48, 0x4A48, 0x4BF8, 0x4A48, 0x4A48, 0x7BF8, 0x4842, 0x42, 0x3E}},
    {26236, 2, {0x80, 0x40, 0xF7FE, 0x9402, 0x9A04, 0x9200, 0x93BC, 0xF4A4, 0x94A4, 0x96A4, 0x9934, 0x9128, 0xF222, 0x9222, 0x422, 0x81E}},
    {26237, 2, {0x110, 0x110, 0xF110, 0x9110, 0x97BC, 0x9110, 0x9110, 0xF338, 0x93B8, 0x9554, 0x9554, 0x9992, 0xF110, 0x9110, 0x110, 0x110}},
    {26238, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x4800, 0x49FC, 0x4904, 0x7904, 0x4904, 0x49FC, 0x4820, 0x48A8, 0x7924, 0x4A22, 0xA0, 0x40}},
    {26239, 2, {0xFC, 0x84, 0x7C84, 0x44FC, 0x4484, 0x4484, 0x44FC, 0x7C00, 0x45FE, 0x4502, 0x4502, 0x45FE, 0x7D02, 0x4502, 0x1FE, 0x102}},
    {26240, 2, {0x90, 0x88, 0x7900, 0x4A7E, 0x4C90, 0x4890, 0x4910, 0x7B10, 0x4D7E, 0x4910, 0x4910, 0x4910, 0x7910, 0x4910, 0x17E, 0x100}},
    {26241, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x48A8, 0x4888, 0x7BFE, 0x4888, 0x4800, 0x4BFE, 0x4890, 0x7890, 0x4912, 0x112, 0x20E, 0x400}},
    {26242, 2, {0x100C, 0xFEF0, 0x2080, 0x4880, 0x7EFE, 0x888, 0xE88, 0xF908, 0x4A08, 0x1FF8, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0}},
    {26243, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7C7C, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0xFC7E, 0x440, 0x440}},
    {26244, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x49FC, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4800, 0x3FE, 0x0}},
    {26245, 2, {0x100, 0x13C, 0xF100, 0x9100, 0x91BC, 0x9564, 0x9524, 0xF53C, 0x9524, 0x9124, 0x9124, 0x913C, 0xF124, 0x9100, 0x17E, 0x100}},
    {26246, 2, {0x210, 0x110, 0xF110, 0x97BE, 0x9228, 0x9248, 0x93AA, 0xF2AE, 0x92BA, 0x92EA, 0x92AE, 0x92A8, 0xF4AA, 0x94A2, 0x9A2, 0x101E}},
    {26247, 2, {0x0, 0x7BC, 0xF484, 0x9484, 0x9484, 0x97BC, 0x9400, 0xF400, 0x97BC, 0x9424, 0x9424, 0x97A8, 0xF410, 0x9428, 0x444, 0x482}},
    {26248, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4104, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {26249, 2, {0x0, 0x3FE, 0x7A22, 0x4820, 0x4BFE, 0x4820, 0x79FC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x7820, 0x4BFE, 0x20, 0x20, 0x20}},
    {26250, 2, {0x0, 0x3FE, 0x7C20, 0x4440, 0x45FC, 0x4504, 0x4504, 0x7DFC, 0x4504, 0x45FC, 0x4504, 0x4504, 0x7DFC, 0x4400, 0x88, 0x104}},
    {26251, 2, {0x20, 0x7F20, 0x413E, 0x7F44, 0x48A4, 0x7F24, 0x4828, 0x4510, 0x5328, 0x6146, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26252, 2, {0x10, 0x3D4, 0x7858, 0x4A52, 0x498C, 0x4888, 0x4904, 0x7AFA, 0x4820, 0x4820, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x202}},
    {26253, 2, {0x0, 0x1FC, 0x7D04, 0x45FC, 0x4504, 0x45FC, 0x4480, 0x7DFE, 0x4622, 0x4522, 0x4552, 0x4502, 0x7DFA, 0x4402, 0x14, 0x8}},
    {26254, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48A8, 0x4820, 0x49FC, 0x7924, 0x4924, 0x4924, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x202}},
    {26255, 2, {0x40, 0x44, 0x7BF4, 0x4848, 0x4850, 0x4FFE, 0x4840, 0x7880, 0x49F8, 0x4B08, 0x4D08, 0x49F8, 0x7908, 0x4908, 0x1F8, 0x108}},
    {26256, 2, {0x40, 0x40, 0x79FC, 0x4884, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4820, 0x4BFE, 0x4820, 0x7A20, 0x4BFE, 0x20, 0x20}},
    {26257, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x1FD0, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x810}},
    {26258, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4920, 0x49FE, 0x4A20, 0x4820, 0x79FC, 0x4820, 0x20, 0x3FE, 0x0}},
    {26259, 2, {0x40, 0x7F40, 0x127E, 0xC88, 0xFF48, 0x1948, 0x2A50, 0x4820, 0xA850, 0x108C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26260, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x4820, 0x4BFE, 0x7A8A, 0x4A52, 0x4AFA, 0x4A22, 0x4A22, 0x7AFA, 0x4A22, 0x222, 0x22A, 0x204}},
    {26261, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x4924, 0x79AC, 0x4974, 0x4924, 0x49FC, 0x4820, 0x7870, 0x48A8, 0x124, 0x222, 0x20}},
    {26262, 2, {0x3C, 0x7C0, 0x7A44, 0x4928, 0x4800, 0x4BFC, 0x7880, 0x4880, 0x4FFE, 0x4900, 0x49F8, 0x7A88, 0x4A50, 0x420, 0x8D8, 0x306}},
    {26263, 2, {0x80, 0x40, 0x7BFC, 0x4800, 0x4908, 0x4890, 0x4FFE, 0x7800, 0x49F8, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x1F8, 0x108}},
    {26264, 2, {0x0, 0x1F8, 0x7908, 0x49F8, 0x4908, 0x49F8, 0x7800, 0x4BFE, 0x4900, 0x49FC, 0x4A54, 0x7C94, 0x4924, 0x244, 0xA8, 0x110}},
    {26265, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x93F8, 0x9080, 0x97FC, 0xF110, 0x9208, 0x95F6, 0x9110, 0x9110, 0xF1F0, 0x9110, 0x110, 0x1F0}},
    {26266, 2, {0x8, 0x3C, 0x7BC0, 0x4804, 0x4A44, 0x4928, 0x79FC, 0x4A20, 0x4820, 0x4BFE, 0x4820, 0x7924, 0x4924, 0x124, 0x1FC, 0x4}},
    {26267, 2, {0x108, 0x88, 0x7890, 0x4BFC, 0x4840, 0x49F8, 0x4840, 0x7BFE, 0x4880, 0x4900, 0x49FC, 0x4A20, 0x7C20, 0x4820, 0x3FE, 0x0}},
    {26268, 2, {0x2020, 0x1010, 0xFEFE, 0x0, 0x4444, 0x2424, 0x2828, 0xEFE, 0xF000, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26269, 2, {0x0, 0x3FE, 0x7A02, 0x4800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4840, 0x4820, 0x4BFE, 0x7800, 0x4888, 0x104, 0x202}},
    {26270, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4820, 0x4BFE, 0x7870, 0x48A8, 0x124, 0x222, 0x20}},
    {26271, 2, {0x20, 0x124, 0x7924, 0x4924, 0x49FC, 0x4800, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x4904, 0x79FC, 0x4888, 0x50, 0x3FE, 0x0}},
    {26272, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {26273, 2, {0x90, 0x88, 0x7904, 0x4A42, 0x4C88, 0x49FC, 0x7804, 0x4800, 0x4BDE, 0x4842, 0x4A52, 0x794A, 0x4A52, 0x42, 0x14A, 0x84}},
    {26274, 2, {0x1000, 0x11F8, 0x1108, 0x7DF8, 0x5508, 0x55F8, 0x7C00, 0x57FE, 0x5500, 0x55FC, 0x7E54, 0x5494, 0x1124, 0x1244, 0x10A8, 0x1110}},
    {26275, 2, {0x100, 0x100, 0x79FC, 0x4A00, 0x4DF8, 0x4800, 0x7BF8, 0x4888, 0x4AA8, 0x49C8, 0x4FF8, 0x788A, 0x49CA, 0x6AA, 0x86, 0x82}},
    {26276, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4840, 0x4BFE, 0x4888, 0x4924, 0x7A22, 0x4DFC, 0x20, 0x20, 0x20}},
    {26277, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4BFE, 0x4A02, 0x4840, 0x4BFE, 0x7888, 0x4990, 0x60, 0x98, 0x304}},
    {26278, 2, {0x0, 0x3FFE, 0x2210, 0x2210, 0x3FBE, 0x2718, 0x2AB4, 0x3252, 0x2210, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x4408, 0x47F8, 0x8408}},
    {26279, 2, {0xE, 0x3F0, 0x7844, 0x4A24, 0x4908, 0x4BFE, 0x7A42, 0x4840, 0x4BFE, 0x4880, 0x48FC, 0x7944, 0x4928, 0x210, 0x468, 0x186}},
    {26280, 2, {0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x0}},
    {26281, 2, {0x100, 0x110, 0x7BDC, 0x4A54, 0x4D54, 0x4A88, 0x7908, 0x4AF4, 0x4C02, 0x4800, 0x4BFC, 0x7840, 0x4950, 0x248, 0x544, 0x80}},
    {26282, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4888, 0x48F8, 0x7820, 0x4BFE, 0x4A22, 0x4B32, 0x4AAA, 0x7B76, 0x4A22, 0x222, 0x22A, 0x224}},
    {26283, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {26284, 2, {0x820, 0x3E20, 0x8F8, 0x7F28, 0x2228, 0x496A, 0x3E2A, 0x856, 0x7E82, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26285, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4924, 0x48A8, 0x4924, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {26286, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x820, 0xFE0}},
    {26287, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4820, 0x7BFE, 0x4850, 0x88, 0x104, 0x202}},
    {26288, 2, {0x124, 0x124, 0xF224, 0x9424, 0x9954, 0x914A, 0x9292, 0xF610, 0x9A10, 0x9250, 0x925C, 0x9250, 0xF250, 0x92B0, 0x29E, 0x300}},
    {26289, 2, {0x0, 0x3FE, 0x7A48, 0x4A48, 0x4BFE, 0x4A48, 0x7A20, 0x4BFE, 0x4A40, 0x4AFC, 0x4B44, 0x7A44, 0x4A7C, 0x200, 0x3FE, 0x0}},
    {26290, 2, {0x20, 0x1FC, 0x7888, 0x4850, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {26291, 2, {0x88, 0x3DE, 0x7888, 0x49DC, 0x4888, 0x4BDE, 0x7888, 0x4800, 0x4BFC, 0x4804, 0x4804, 0x79FC, 0x4804, 0x4, 0x3FC, 0x4}},
    {26292, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x7FFC, 0x820, 0xFFFE, 0x820, 0x3118, 0xC926, 0x540, 0x920, 0x1510, 0x2208}},
    {26293, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48F8, 0x4820, 0x79FC, 0x4924, 0x49FC, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x50, 0x88, 0x306}},
    {26294, 2, {0x210, 0x110, 0xF110, 0x97BE, 0x9240, 0x9200, 0x93BE, 0xF28A, 0x9288, 0x92A8, 0x92AE, 0x92A8, 0xF4A8, 0x95A8, 0x85E, 0x1080}},
    {26295, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4822, 0x4BFE, 0x4808, 0x7BFE, 0x4908, 0x88, 0x28, 0x10}},
    {26296, 2, {0x40, 0x40, 0xF7FC, 0x90A0, 0x9514, 0x9208, 0x97FC, 0xFA0A, 0x93F8, 0x9208, 0x93F8, 0x9040, 0xF248, 0x9444, 0x944, 0x80}},
    {26297, 2, {0x7F8, 0x4408, 0x27F8, 0x2408, 0x7F8, 0x240, 0xE7FC, 0x2C40, 0x37F8, 0x2440, 0x27F8, 0x2440, 0x27FC, 0x5400, 0x8FFE, 0x0}},
    {26298, 2, {0x0, 0x3DE, 0x7A52, 0x4A52, 0x4BDE, 0x4800, 0x79FC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x7820, 0x4BFE, 0x20, 0x20, 0x20}},
    {26299, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x124, 0x222, 0x60}},
    {26300, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26301, 2, {0x20, 0x124, 0x78A8, 0x4BFE, 0x48A8, 0x4924, 0x4A22, 0x7904, 0x4904, 0x49DE, 0x4A44, 0x4D54, 0x789E, 0x4904, 0x204, 0x404}},
    {26302, 2, {0x208, 0x108, 0xF7C8, 0x9010, 0x979E, 0x9494, 0x97A4, 0xF014, 0x9794, 0x9094, 0x9114, 0x91C8, 0xF708, 0x9114, 0x524, 0x242}},
    {26303, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4888, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x104, 0x1FC, 0x104}},
    {26304, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4800, 0x4BFE, 0x7A02, 0x49FC, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x88, 0x3FE, 0x0}},
    {26305, 2, {0x1000, 0x7EFC, 0x4210, 0x7E90, 0x42FE, 0x7E30, 0x4050, 0x7C92, 0x410E, 0x5FF0, 0x7010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE}},
    {26306, 2, {0x88, 0x88, 0x79FC, 0x4888, 0x4888, 0x4BFE, 0x7820, 0x49FC, 0x4924, 0x49FC, 0x4924, 0x79FC, 0x4800, 0x88, 0x104, 0x202}},
    {26307, 2, {0x10, 0x210, 0x797C, 0x4814, 0x48FE, 0x4814, 0x4B7C, 0x7992, 0x4954, 0x4938, 0x4954, 0x4992, 0x7930, 0x4900, 0x2FE, 0x400}},
    {26308, 2, {0x88, 0x3FE, 0x7888, 0x4800, 0x49FC, 0x48A8, 0x48A8, 0x7BFE, 0x48A8, 0x48A8, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {26309, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7FFC, 0x440, 0x3FF8, 0x920, 0x7FFC, 0x920, 0x3FF8, 0x100, 0x7FFC, 0x100}},
    {26310, 2, {0x3FFE, 0x2000, 0x218C, 0x2E70, 0x2210, 0x3FBE, 0x2718, 0x2AB4, 0x3252, 0x2210, 0x27F8, 0x2408, 0x47F8, 0x4408, 0x87F8, 0x408}},
    {26311, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1010, 0x3FF8, 0x8}},
    {26312, 2, {0x20, 0x1FC, 0x7888, 0x4850, 0x4BFE, 0x4800, 0x79FC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x7820, 0x49FC, 0x20, 0x3FE, 0x0}},
    {26313, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4BFE, 0x4908, 0x7B9C, 0x4908, 0x4988, 0x4E3E, 0x4800, 0x7BFE, 0x4890, 0x90, 0x112, 0x20E}},
    {26314, 2, {0x90, 0x7FC, 0x7894, 0x4BFC, 0x4A90, 0x4BFE, 0x7912, 0x4BFA, 0x4D0E, 0x49F8, 0x4908, 0x79F8, 0x4908, 0x1F8, 0x90, 0x108}},
    {26315, 2, {0x0, 0x3FE, 0x7850, 0x4BFE, 0x4A52, 0x4BFE, 0x7800, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x3FE, 0x20, 0x20}},
    {26316, 2, {0xF8, 0x7C88, 0x44F8, 0x7C88, 0x44F8, 0x7C88, 0x2A8, 0x110, 0x7FFC, 0x4824, 0x1010, 0x2FE8, 0x100, 0x100, 0x7FFC, 0x0}},
    {26317, 2, {0x40, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4820, 0x4924, 0x48A8, 0x7924, 0x4820, 0x3FE, 0x20, 0x20}},
    {26318, 2, {0x0, 0x1FC, 0x7954, 0x49FC, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x4888, 0x4850, 0x49FC, 0x7820, 0x4BFE, 0x20, 0x20, 0x20}},
    {26319, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2100, 0x27DE, 0x5452, 0x97D4, 0xE458, 0x2454, 0x57D2, 0x9512, 0xF49A, 0x2554, 0x4610}},
    {26320, 2, {0xFE0, 0x820, 0xFE0, 0x820, 0xFE0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x1100, 0x1FF8, 0x2100, 0x5FF0, 0x100, 0xFFFE}},
    {26321, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x6C0, 0x1930, 0xE64E, 0x180, 0xE20, 0xC0, 0x1F00}},
    {26322, 2, {0x108, 0x208, 0xF7C8, 0x9450, 0x97DE, 0x9464, 0x97D4, 0xF214, 0x9114, 0x97D4, 0x9214, 0x93C8, 0xF248, 0x9454, 0x554, 0x8A2}},
    {26323, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x1110, 0x1FF0, 0x4104, 0x7FFC, 0x200, 0xFFFE, 0x920, 0x3FF8, 0xC546, 0xBA0, 0x1110}},
    {26324, 2, {0x288, 0x288, 0xF7E8, 0x9290, 0x941E, 0x97E4, 0x9454, 0xF854, 0x9754, 0x9554, 0x9554, 0x9748, 0xF548, 0x9054, 0x294, 0x122}},
    {26325, 2, {0x100, 0x1F8, 0x7A08, 0x4FFE, 0x4A88, 0x4B24, 0x7BFE, 0x4A00, 0x4AFC, 0x4A00, 0x4AFC, 0x7A00, 0x4AFC, 0x484, 0x4FC, 0x884}},
    {26326, 2, {0x3C, 0x3C0, 0xF044, 0x9224, 0x9108, 0x97FE, 0x9442, 0xF2A4, 0x928A, 0x947A, 0x9080, 0x90F8, 0xF108, 0x9290, 0x60, 0x79E}},
    {26327, 2, {0x50, 0x252, 0x7954, 0x4850, 0x4BFE, 0x4888, 0x7850, 0x4BFE, 0x4820, 0x49FC, 0x4820, 0x7BFE, 0x48A8, 0x124, 0x622, 0x20}},
    {26328, 2, {0x0, 0x1FC, 0x7820, 0x4BFE, 0x4A22, 0x49AC, 0x7820, 0x49AC, 0x4800, 0x4BFE, 0x4820, 0x79FC, 0x4954, 0x154, 0x154, 0x10C}},
    {26329, 2, {0x0, 0x3FE, 0x7A52, 0x4A52, 0x4BFE, 0x4824, 0x49FC, 0x7828, 0x4BFE, 0x4820, 0x48FC, 0x4984, 0x7AFC, 0x4884, 0xFC, 0x84}},
    {26330, 2, {0x110, 0x7FC, 0xF110, 0x97FC, 0x9404, 0x93F8, 0x9000, 0xF7FC, 0x9080, 0x9144, 0x96A8, 0x9170, 0xF6A8, 0x9126, 0x6A0, 0x40}},
    {26331, 2, {0x3C, 0x7C0, 0xF040, 0x9FFE, 0x9040, 0x97FC, 0x9554, 0xF4E4, 0x97FC, 0x9040, 0x97FC, 0x9040, 0xFFFE, 0x9000, 0x524, 0x892}},
    {26332, 2, {0x0, 0x7BC, 0xF084, 0x94A4, 0x9294, 0x94A4, 0x9120, 0xF210, 0x97FC, 0x9A20, 0x93FC, 0x9220, 0xF3FC, 0x9220, 0x3FE, 0x200}},
    {26333, 2, {0x3F8, 0x208, 0xF3F8, 0x9208, 0x93F8, 0x9110, 0x97FC, 0xF110, 0x9FFE, 0x9110, 0x9248, 0x9554, 0xF0E0, 0x9150, 0x248, 0xC0}},
    {26334, 2, {0x0, 0x3FE, 0x7A48, 0x4BFE, 0x4A48, 0x4AFC, 0x7AA4, 0x4AFC, 0x4AA4, 0x4AFC, 0x4A20, 0x7BFE, 0x4B2A, 0x57A, 0x50A, 0x906}},
    {26335, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x4430, 0x860E}},
    {26336, 2, {0x40, 0x20, 0xF7FE, 0x9488, 0x95FC, 0x9488, 0x97FE, 0xF420, 0x95FC, 0x9524, 0x95FC, 0x9524, 0xF5FC, 0x9400, 0x488, 0x904}},
    {26337, 2, {0xFE0, 0x820, 0xFE0, 0x820, 0xFE0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0xFFFE, 0x9012, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {26338, 2, {0x40, 0x20, 0x7BFE, 0x4A22, 0x49FC, 0x4850, 0x7A8A, 0x49FC, 0x4A8A, 0x48F8, 0x4888, 0x78F8, 0x4820, 0x124, 0x2A2, 0x40}},
    {26339, 2, {0x110, 0x110, 0xF7FC, 0x9110, 0x91F0, 0x9208, 0x92EA, 0xF6AC, 0x92A8, 0x92AA, 0x96EA, 0x9A06, 0xF000, 0x92A4, 0x252, 0x452}},
    {26340, 2, {0x7FC, 0x40, 0xF7FE, 0x9442, 0x9B5C, 0x9040, 0x975C, 0xF120, 0x93FC, 0x9620, 0x9BFC, 0x9220, 0xF3FC, 0x9220, 0x3FE, 0x200}},
    {26341, 2, {0x40, 0x7C, 0x7840, 0x4BFC, 0x4A44, 0x4BF0, 0x7A44, 0x4AFC, 0x4AA8, 0x4AF8, 0x4AA8, 0x7AF8, 0x4A00, 0x5FC, 0x554, 0xBFE}},
    {26342, 2, {0x110, 0x7FC, 0xF040, 0x93F8, 0x9040, 0x97FC, 0x9000, 0xF3D4, 0x9112, 0x97FE, 0x9350, 0x9534, 0xF38C, 0x908A, 0x296, 0x102}},
    {26343, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x10, 0x7F10, 0xFE, 0x3E92, 0x2292, 0x3EFE, 0x10, 0x7F14, 0x5512, 0x7FFE, 0x4902}},
    {26344, 2, {0x210, 0x110, 0xF7DE, 0x9010, 0x945E, 0x9282, 0x97DE, 0xF010, 0x97DE, 0x9450, 0x97DE, 0x9450, 0xF7DE, 0x9450, 0x452, 0x4CE}},
    {26345, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x2288, 0x3EF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x490, 0x1C60, 0x671C}},
    {26346, 2, {0x0, 0x3FE, 0xF252, 0x9252, 0x93FE, 0x9108, 0x9154, 0xF25E, 0x97B4, 0x911E, 0x9254, 0x97DE, 0xF014, 0x9554, 0x55E, 0x10}},
    {26347, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {26348, 2, {0xFBE, 0x0, 0xF7BC, 0x94A4, 0x96B4, 0x94A4, 0x9040, 0xF7FE, 0x94A0, 0x97FC, 0x94A4, 0x97FC, 0xF512, 0x99DC, 0x912, 0x11CE}},
    {26349, 2, {0x248, 0x150, 0xF7FC, 0x9404, 0x91F0, 0x9110, 0x97FC, 0xF554, 0x94E4, 0x97FC, 0x9040, 0x97FC, 0xF040, 0x9FFE, 0x2A4, 0x452}},
    {26350, 2, {0x7BC, 0x4A4, 0xF7BC, 0x9000, 0x97FE, 0x9400, 0x9590, 0xF49E, 0x97D4, 0x9564, 0x95D4, 0x9554, 0xF5D4, 0x9568, 0xBC8, 0x1054}},
    {26351, 2, {0x3FE, 0x202, 0x7BFE, 0x4A92, 0x4A54, 0x4A92, 0x7AFE, 0x4AAA, 0x4AFE, 0x4B22, 0x4AFA, 0x7AAA, 0x4AFA, 0x222, 0x5FA, 0x4}},
    {26352, 2, {0x0, 0x0, 0x7FFC, 0x4004, 0x4004, 0x4004, 0x4004, 0x7FE4, 0x4004, 0x4004, 0x4004, 0x4004, 0x7FFC, 0x4004, 0x0, 0x0}},
    {26353, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {26354, 2, {0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x4444, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x4444, 0x7FFC, 0x4004}},
    {26355, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x90, 0xA0, 0x42, 0x1A2, 0xE1A, 0xF006}},
    {26356, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x1100, 0xA00, 0x600, 0x19C0, 0xE03E}},
    {26357, 2, {0x120, 0x110, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x90, 0xA0, 0x42, 0x1A2, 0xE1A, 0xF006}},
    {26358, 2, {0x800, 0x800, 0x1FF8, 0x2248, 0x4448, 0x888, 0x1128, 0x2210, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26359, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x800, 0x1FFC, 0x2104, 0x5284, 0x9444, 0x1004, 0x1F84, 0x28, 0x10}},
    {26360, 2, {0x100, 0x3FF0, 0x110, 0xFFFE, 0x110, 0x3FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {26361, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26362, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26363, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x1010, 0x1010, 0x3EFC, 0x4210, 0xA490, 0x1490, 0x8FE, 0x1010, 0x2010, 0x4010}},
    {26364, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF0, 0x820, 0x7C0, 0x1830, 0xE00E}},
    {26365, 2, {0x820, 0x440, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26366, 2, {0x820, 0x440, 0x3FF8, 0x2108, 0x2928, 0x2548, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26367, 2, {0x820, 0x820, 0x7EFC, 0x820, 0xFEFE, 0x1450, 0x2288, 0x4106, 0x9FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26368, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x2200, 0x3EF8, 0x2288, 0x3E90, 0x2250, 0x2F20, 0xF250, 0x4288, 0x306}},
    {26369, 2, {0x3E7C, 0x810, 0x4890, 0x7EFC, 0x1428, 0x244A, 0x4686, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26370, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x2220, 0x3E20, 0x22FC, 0x3E24, 0x2224, 0x2F24, 0xF244, 0x4254, 0x288}},
    {26371, 2, {0x100, 0x280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {26372, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5524, 0x7DAC, 0x5574, 0x5524, 0x55FC, 0x7C20, 0x5470, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {26373, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x1104, 0x11FC, 0x1080, 0xFDFE, 0x1222, 0x1122, 0x2152, 0x2902, 0x45FA, 0xFC02, 0x4414, 0x8}},
    {26374, 2, {0x4440, 0x2440, 0x2840, 0xFE7E, 0x92AA, 0xD72A, 0xBA2A, 0x924A, 0xFE4A, 0x92, 0x7D12, 0x4422, 0x7C22, 0x4442, 0x7C94, 0x4408}},
    {26375, 2, {0x1010, 0x2820, 0x44FE, 0xBA92, 0x92, 0xFEFE, 0x9292, 0xD692, 0x92FE, 0xFE28, 0x48, 0x7DFE, 0x4408, 0x7C08, 0x4408, 0x7C08}},
    {26376, 2, {0x7FC, 0x404, 0x404, 0x404, 0x404, 0x7FC, 0x404, 0x404, 0x404, 0x7FC, 0x404, 0x404, 0x804, 0x804, 0x1014, 0x2008}},
    {26377, 2, {0x200, 0x200, 0xFFFE, 0x400, 0x400, 0xFF0, 0x810, 0x1810, 0x2FF0, 0x4810, 0x8810, 0xFF0, 0x810, 0x810, 0x850, 0x820}},
    {26378, 2, {0x0, 0x7800, 0x49FC, 0x4800, 0x4800, 0x7800, 0x4BFE, 0x4890, 0x4890, 0x7890, 0x4890, 0x4890, 0x4892, 0x4912, 0x4912, 0x9A0E}},
    {26379, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x2244, 0x4284, 0x4A94, 0x8508}},
    {26380, 2, {0x10, 0x3C50, 0x2450, 0x2450, 0x2488, 0x3C88, 0x2504, 0x26FA, 0x2448, 0x3C48, 0x2448, 0x2448, 0x2488, 0x4488, 0x5528, 0x8A10}},
    {26381, 2, {0x0, 0x3EFC, 0x2284, 0x2284, 0x2294, 0x3E88, 0x2280, 0x22FC, 0x22A4, 0x3EA4, 0x22A8, 0x22A8, 0x2290, 0x42A8, 0x4AC4, 0x8482}},
    {26382, 2, {0x20, 0x3C20, 0x2450, 0x2450, 0x2488, 0x3D24, 0x2612, 0x2410, 0x25FC, 0x3C04, 0x2408, 0x2488, 0x2450, 0x4420, 0x5410, 0x8810}},
    {26383, 2, {0x20, 0x3C20, 0x2524, 0x2524, 0x2524, 0x3D24, 0x25FC, 0x2420, 0x2420, 0x3D24, 0x2524, 0x2524, 0x2524, 0x4524, 0x55FC, 0x8804}},
    {26384, 2, {0x80, 0x3C80, 0x2500, 0x25FC, 0x2604, 0x3C04, 0x25E4, 0x2524, 0x2524, 0x3D24, 0x25E4, 0x2524, 0x2404, 0x4404, 0x5428, 0x8810}},
    {26385, 2, {0x48, 0x7A48, 0x4A48, 0x4A48, 0x4A48, 0x7FFE, 0x4A48, 0x4A48, 0x4A48, 0x7A48, 0x4A78, 0x4A00, 0x4A00, 0x4A00, 0x4BFE, 0x9800}},
    {26386, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2524, 0x3D24, 0x2554, 0x258C, 0x2524, 0x3D24, 0x2554, 0x258C, 0x2504, 0x4504, 0x5514, 0x8908}},
    {26387, 2, {0x50, 0x3C50, 0x2450, 0x2552, 0x24D4, 0x3C58, 0x2450, 0x2458, 0x24D4, 0x3D52, 0x2650, 0x2450, 0x2492, 0x4492, 0x5512, 0x8A0E}},
    {26388, 2, {0x4200, 0x227C, 0x2444, 0xFF44, 0x844, 0x87C, 0x4944, 0x4944, 0x4944, 0x7F7C, 0x944, 0x1044, 0x1084, 0x2084, 0x4114, 0x8208}},
    {26389, 2, {0x88, 0x3C48, 0x2450, 0x2400, 0x25FC, 0x3C20, 0x2420, 0x2420, 0x27FE, 0x3C20, 0x2450, 0x2450, 0x2488, 0x4488, 0x5504, 0x8A02}},
    {26390, 2, {0x40, 0x3C20, 0x25FC, 0x2504, 0x2504, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x3D20, 0x2522, 0x2514, 0x2508, 0x4544, 0x5582, 0x8900}},
    {26391, 2, {0x1000, 0x87C, 0x7E44, 0x4244, 0x4244, 0x7E7C, 0x4244, 0x4244, 0x7E44, 0x407C, 0x4844, 0x4444, 0x5A84, 0x6284, 0x4114, 0x208}},
    {26392, 2, {0x40, 0x7840, 0x4888, 0x4904, 0x4BFE, 0x7802, 0x4888, 0x4944, 0x4A42, 0x78F8, 0x4988, 0x4A50, 0x4820, 0x4850, 0x4988, 0x9E06}},
    {26393, 2, {0x0, 0xFE7C, 0x8244, 0xAA44, 0xA644, 0xC67C, 0x8244, 0xBA44, 0xAA44, 0xAA7C, 0xAA44, 0xAA44, 0xFE84, 0x8284, 0x114, 0x208}},
    {26394, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1FF8, 0x0, 0x7CFC, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7C84, 0x114, 0x208}},
    {26395, 2, {0x1000, 0x87C, 0xFF44, 0x207C, 0x2044, 0x267C, 0x3844, 0x2094, 0x108, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {26396, 2, {0x40, 0x3C20, 0x27FE, 0x2400, 0x25FC, 0x3D04, 0x25FC, 0x2400, 0x25FC, 0x3C08, 0x2410, 0x27FE, 0x2420, 0x4420, 0x54A0, 0x8840}},
    {26397, 2, {0x800, 0x83E, 0xFFA2, 0x822, 0x7F22, 0x413E, 0x7F22, 0x4122, 0x7F22, 0x493E, 0x822, 0xFFA2, 0x842, 0x842, 0x88A, 0x904}},
    {26398, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x820, 0xFE0, 0x820, 0x860}},
    {26399, 2, {0x2200, 0x227C, 0x7F44, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x3E44, 0x227C, 0x2244, 0xFF44, 0x484, 0x2284, 0x4114, 0x8208}},
    {26400, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x48A8, 0x7820, 0x49FC, 0x4924, 0x4924, 0x7924, 0x4BFE, 0x4820, 0x4850, 0x4888, 0x4904, 0x9A02}},
    {26401, 2, {0x0, 0x7888, 0x4A52, 0x4A22, 0x4A52, 0x7A8A, 0x4A02, 0x4BFE, 0x4888, 0x7944, 0x4A7A, 0x4888, 0x4950, 0x4820, 0x48D8, 0x9B06}},
    {26402, 2, {0x0, 0x7F7C, 0x4844, 0x7E7C, 0x4244, 0x7E7C, 0x4844, 0x7F54, 0x88, 0x7FFC, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {26403, 2, {0x20, 0x3DFC, 0x2488, 0x2450, 0x27FE, 0x3C00, 0x25FC, 0x2524, 0x25FC, 0x3D24, 0x25FC, 0x2420, 0x25FC, 0x4420, 0x57FE, 0x8800}},
    {26404, 2, {0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x4A94, 0x8508, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x4A94, 0x8508}},
    {26405, 2, {0x88, 0x7AAA, 0x4ADC, 0x4C88, 0x4954, 0x7A22, 0x4800, 0x4BFE, 0x4A42, 0x7C44, 0x4BFC, 0x4844, 0x4844, 0x4884, 0x4914, 0x9A08}},
    {26406, 2, {0x88, 0x7BFE, 0x4888, 0x4BFE, 0x4A02, 0x79FC, 0x4800, 0x4BFE, 0x4840, 0x78A2, 0x4B54, 0x48B8, 0x4B54, 0x4892, 0x4B50, 0x9820}},
    {26407, 2, {0x110, 0x7890, 0x4BDE, 0x4810, 0x4A5E, 0x7982, 0x4BDE, 0x4810, 0x4BDE, 0x7A50, 0x4BDE, 0x4A50, 0x4BDE, 0x4A50, 0x4A52, 0x9ACE}},
    {26408, 2, {0x100, 0x100, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x540, 0x920, 0x1110, 0x2108, 0x4104, 0x8102, 0x100, 0x100, 0x100}},
    {26409, 2, {0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x920, 0x910, 0x1108, 0x1108, 0x2104, 0x4104, 0x8104, 0x500, 0x200}},
    {26410, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {26411, 2, {0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {26412, 2, {0x100, 0x100, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x540, 0x920, 0x1110, 0x2108, 0x4FE4, 0x8102, 0x100, 0x100, 0x100}},
    {26413, 2, {0x880, 0x880, 0x880, 0x880, 0xFE80, 0x880, 0x1880, 0x1C80, 0x2A80, 0x2A80, 0x4880, 0x8884, 0x884, 0x884, 0x87C, 0x800}},
    {26414, 2, {0x100, 0x120, 0x110, 0x100, 0xFFFE, 0x100, 0x920, 0x920, 0x920, 0x920, 0x1120, 0x1120, 0x2120, 0x411E, 0x8100, 0x100}},
    {26415, 2, {0x100, 0x120, 0x110, 0x110, 0x7FFC, 0x380, 0x540, 0x540, 0x920, 0x1110, 0x2108, 0x4104, 0x8102, 0x100, 0x100, 0x100}},
    {26416, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x3FF8, 0x10, 0x60, 0x380, 0x1C04, 0x2004, 0x1FFC, 0x0}},
    {26417, 2, {0x100, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {26418, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0xFC40, 0x1040, 0x3040, 0x3840, 0x54A0, 0x54A0, 0x90A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802}},
    {26419, 2, {0x1000, 0x1020, 0x1120, 0x1120, 0xFD20, 0x1120, 0x3120, 0x3920, 0x5520, 0x5110, 0x9110, 0x1210, 0x1208, 0x1408, 0x1804, 0x1002}},
    {26420, 2, {0x840, 0x840, 0x840, 0x840, 0xFE40, 0x850, 0x1848, 0x1C44, 0x2A44, 0x2A40, 0x4840, 0x8840, 0x840, 0x840, 0x840, 0x840}},
    {26421, 2, {0x0, 0xFE0, 0x820, 0x820, 0x820, 0x1020, 0x201E, 0xC100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26422, 2, {0x0, 0x3FF0, 0x810, 0x820, 0x83C, 0x1004, 0x1004, 0x2014, 0x4108, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26423, 2, {0x800, 0x800, 0x9FC, 0x844, 0xFE44, 0x844, 0x1844, 0x1C44, 0x2A44, 0x2A44, 0x4844, 0x8884, 0x884, 0x904, 0x928, 0xA10}},
    {26424, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0xFDFC, 0x1044, 0x3044, 0x3844, 0x5444, 0x5484, 0x9084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {26425, 2, {0x1080, 0x1080, 0x1080, 0x1080, 0xFDF0, 0x1090, 0x3090, 0x3890, 0x5490, 0x5490, 0x9090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {26426, 2, {0x1000, 0x11F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x3110, 0x3910, 0x5510, 0x5510, 0x9110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {26427, 2, {0x804, 0x804, 0x884, 0x884, 0xFE84, 0x884, 0x1884, 0x1C84, 0x2A8C, 0x2AB4, 0x48C4, 0x8884, 0x804, 0x804, 0x804, 0x804}},
    {26428, 2, {0x880, 0x880, 0x880, 0x884, 0xFE88, 0x890, 0x18A0, 0x1CC0, 0x2A80, 0x2A80, 0x4880, 0x8884, 0x884, 0x884, 0x87C, 0x800}},
    {26429, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0xFC80, 0x1080, 0x3100, 0x39FC, 0x5404, 0x5004, 0x9004, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26430, 2, {0x800, 0x800, 0xBFE, 0x820, 0xFE20, 0x820, 0x1820, 0x1C20, 0x2A20, 0x2A20, 0x4820, 0x8820, 0x820, 0x820, 0x8A0, 0x840}},
    {26431, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2128, 0x2390, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26432, 2, {0x8, 0x1810, 0x660, 0x180, 0x660, 0x1818, 0x6104, 0x100, 0xFFFE, 0x100, 0x1120, 0x1110, 0x2108, 0x2104, 0x4504, 0x200}},
    {26433, 2, {0x880, 0x840, 0x840, 0x820, 0xFE20, 0x820, 0x1820, 0x1C50, 0x2A50, 0x2A50, 0x4848, 0x8888, 0x888, 0x884, 0x904, 0xA02}},
    {26434, 2, {0x200, 0x200, 0x3FE0, 0x420, 0x422, 0x822, 0x111E, 0x6100, 0x100, 0xFFFE, 0x100, 0x920, 0x1110, 0x2108, 0x4504, 0x200}},
    {26435, 2, {0x1000, 0x13FC, 0x1104, 0x1104, 0xFD04, 0x1088, 0x3088, 0x3888, 0x5450, 0x5450, 0x9020, 0x1020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {26436, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x1020, 0x3020, 0x3BFE, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26437, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFC20, 0x1020, 0x3020, 0x3BFE, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26438, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFC20, 0x1020, 0x3020, 0x3BFE, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26439, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x3040, 0x3840, 0x5480, 0x54FC, 0x9004, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26440, 2, {0x1000, 0x13FC, 0x1104, 0x1144, 0xFD24, 0x10A8, 0x3088, 0x3888, 0x5450, 0x5450, 0x9020, 0x1020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {26441, 2, {0x804, 0x804, 0x808, 0x810, 0xFE20, 0x844, 0x1804, 0x1C08, 0x2A10, 0x2A22, 0x4842, 0x8804, 0x808, 0x810, 0x820, 0x8C0}},
    {26442, 2, {0x1104, 0x1124, 0x1124, 0x1124, 0xFD24, 0x1124, 0x3124, 0x3924, 0x5524, 0x5124, 0x9124, 0x1124, 0x1124, 0x1224, 0x1204, 0x1404}},
    {26443, 2, {0x1000, 0x11F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x3190, 0x3950, 0x5550, 0x5510, 0x9110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {26444, 2, {0x1000, 0x1000, 0x13FC, 0x1090, 0xFC90, 0x1090, 0x3090, 0x3890, 0x5490, 0x5490, 0x9090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {26445, 2, {0x1000, 0x11FC, 0x1004, 0x1008, 0xFC10, 0x1020, 0x3020, 0x3BFE, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26446, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0xFE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {26447, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26448, 2, {0x808, 0x808, 0x808, 0x808, 0xFEFE, 0x808, 0x1818, 0x1C18, 0x2A28, 0x2A28, 0x4848, 0x8888, 0x808, 0x808, 0x828, 0x810}},
    {26449, 2, {0x1010, 0x1010, 0x1010, 0x1010, 0xFDFE, 0x1010, 0x3010, 0x3810, 0x5510, 0x5490, 0x9090, 0x1010, 0x1010, 0x1010, 0x1050, 0x1020}},
    {26450, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0xFC44, 0x1144, 0x3944, 0x3544, 0x5244, 0x5044, 0x9084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {26451, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0xFD04, 0x1204, 0x3004, 0x3904, 0x5484, 0x5444, 0x9044, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26452, 2, {0x1008, 0x103C, 0x13C0, 0x1040, 0xFC40, 0x1040, 0x3840, 0x347E, 0x53C0, 0x5040, 0x9040, 0x1042, 0x1042, 0x1042, 0x103E, 0x1000}},
    {26453, 2, {0x820, 0x820, 0x820, 0x820, 0xFE20, 0x9FE, 0x1820, 0x1C20, 0x2A20, 0x2A50, 0x4850, 0x8850, 0x888, 0x888, 0x904, 0xA02}},
    {26454, 2, {0x1010, 0x1010, 0x1010, 0x1010, 0xFDFE, 0x1010, 0x3010, 0x3910, 0x5510, 0x5490, 0x9090, 0x1050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {26455, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26456, 2, {0x0, 0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2000, 0x2080, 0x2080, 0x3FFE, 0x21C0, 0x22A0, 0x2490, 0x2888, 0x5086, 0x4080, 0x8080}},
    {26457, 2, {0x848, 0x844, 0x844, 0x840, 0xFE5E, 0x9E0, 0x1840, 0x1C40, 0x2A40, 0x2A40, 0x4820, 0x8822, 0x812, 0x80A, 0x806, 0x802}},
    {26458, 2, {0x1040, 0x1040, 0x1080, 0x10FE, 0xFD00, 0x1200, 0x30FC, 0x3808, 0x5410, 0x5420, 0x9040, 0x1080, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {26459, 2, {0x1000, 0x11FC, 0x1004, 0x1004, 0xFC04, 0x10FC, 0x3080, 0x3880, 0x5500, 0x55FC, 0x9004, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26460, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFC20, 0x1020, 0x33FE, 0x3820, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {26461, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0xFD2C, 0x1134, 0x3164, 0x3BA4, 0x5524, 0x5534, 0x9128, 0x1122, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {26462, 2, {0x1000, 0x1000, 0x11F8, 0x1008, 0xFC08, 0x1008, 0x3008, 0x39F8, 0x5508, 0x5500, 0x9100, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {26463, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2388, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26464, 2, {0x1000, 0x1000, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x3020, 0x3820, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26465, 2, {0x400, 0x400, 0xFF0, 0x1820, 0x6440, 0x380, 0x1D70, 0xE10E, 0x100, 0x7FFC, 0x100, 0x920, 0x1110, 0x2108, 0x4504, 0x200}},
    {26466, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {26467, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFD24, 0x1124, 0x3124, 0x3924, 0x5524, 0x5524, 0x9124, 0x1124, 0x1124, 0x11FC, 0x1004, 0x1000}},
    {26468, 2, {0x1000, 0x1000, 0x13FE, 0x1080, 0xFC80, 0x1080, 0x30FC, 0x3884, 0x5484, 0x5484, 0x9104, 0x1104, 0x1204, 0x1204, 0x1428, 0x1810}},
    {26469, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x1110, 0x910, 0x920, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100, 0x100}},
    {26470, 2, {0x1080, 0x1080, 0x1080, 0x10F8, 0xFD08, 0x1108, 0x3110, 0x3A10, 0x5420, 0x5020, 0x9050, 0x1050, 0x1088, 0x1108, 0x1204, 0x1402}},
    {26471, 2, {0x1020, 0x1010, 0x1010, 0x1000, 0xFDFE, 0x1080, 0x3080, 0x3880, 0x5480, 0x5480, 0x9080, 0x1080, 0x1080, 0x1080, 0x10FC, 0x1000}},
    {26472, 2, {0x1000, 0x11F8, 0x1010, 0x1020, 0xFC40, 0x1080, 0x31FE, 0x3892, 0x5492, 0x5492, 0x9112, 0x1122, 0x1222, 0x1442, 0x1094, 0x1108}},
    {26473, 2, {0x1000, 0x13F0, 0x1010, 0x1010, 0xFD10, 0x1110, 0x3110, 0x39FC, 0x5404, 0x5404, 0x9004, 0x13F4, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26474, 2, {0x1020, 0x1020, 0x1020, 0x10A8, 0xFCA4, 0x10A2, 0x3122, 0x3920, 0x5624, 0x5424, 0x9028, 0x1008, 0x1010, 0x1020, 0x10C0, 0x1300}},
    {26475, 2, {0x810, 0x810, 0x810, 0x890, 0xFE90, 0x890, 0x189E, 0x1C90, 0x2A90, 0x2A90, 0x4890, 0x8890, 0x890, 0x890, 0xBFE, 0x800}},
    {26476, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x3090, 0x3890, 0x5490, 0x5490, 0x9090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {26477, 2, {0x1080, 0x1040, 0x1040, 0x13FC, 0xFC00, 0x1000, 0x31F0, 0x3910, 0x5510, 0x5510, 0x9110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {26478, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x31FC, 0x3924, 0x5524, 0x5524, 0x9124, 0x1134, 0x1128, 0x1020, 0x1020, 0x1020}},
    {26479, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1040, 0x3840, 0x34D0, 0x5148, 0x5244, 0x9444, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040}},
    {26480, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {26481, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2388, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26482, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26483, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {26484, 2, {0x840, 0x840, 0x840, 0x87C, 0xFE84, 0x888, 0x1920, 0x1C20, 0x2A20, 0x2A50, 0x4850, 0x8850, 0x888, 0x888, 0x904, 0xA02}},
    {26485, 2, {0x1080, 0x1080, 0x1080, 0x11FC, 0xFD20, 0x1220, 0x3020, 0x3820, 0x57FE, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26486, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1124, 0x3124, 0x3924, 0x5524, 0x55FC, 0x9024, 0x1020, 0x1022, 0x1022, 0x101E, 0x1000}},
    {26487, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFD24, 0x1124, 0x3124, 0x39FC, 0x5504, 0x5500, 0x9100, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {26488, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFC88, 0x1106, 0x3200, 0x39FC, 0x5484, 0x5484, 0x9048, 0x1050, 0x1020, 0x1050, 0x1088, 0x1306}},
    {26489, 2, {0x10A0, 0x10A0, 0x10A0, 0x1124, 0xFD24, 0x1328, 0x3530, 0x3920, 0x5560, 0x51A0, 0x9120, 0x1122, 0x1122, 0x1122, 0x111E, 0x1100}},
    {26490, 2, {0x1000, 0x1040, 0x1020, 0x1010, 0xFC90, 0x1080, 0x3080, 0x3A84, 0x5682, 0x5282, 0x9282, 0x1488, 0x1088, 0x1088, 0x1078, 0x1000}},
    {26491, 2, {0x800, 0x9F8, 0x848, 0x848, 0xFE48, 0x848, 0x1848, 0x1DF8, 0x2A88, 0x2A88, 0x4888, 0x8888, 0x888, 0x888, 0xBFE, 0x800}},
    {26492, 2, {0x1000, 0x11F8, 0x1008, 0x1050, 0xFC20, 0x1010, 0x33FE, 0x3822, 0x5424, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26493, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x11FC, 0x3820, 0x3420, 0x5020, 0x53FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26494, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0xFC88, 0x1124, 0x3124, 0x3A22, 0x5440, 0x5440, 0x9088, 0x1084, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {26495, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x11FC, 0x3144, 0x3944, 0x5544, 0x5528, 0x9128, 0x1110, 0x1110, 0x1228, 0x1244, 0x1482}},
    {26496, 2, {0x440, 0x440, 0x820, 0x1010, 0x2208, 0xC446, 0x820, 0x1FF0, 0x110, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26497, 2, {0x1000, 0x13FC, 0x1084, 0x1088, 0xFC88, 0x1090, 0x309C, 0x3884, 0x5544, 0x5544, 0x9128, 0x1128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {26498, 2, {0x800, 0x8FC, 0x884, 0x884, 0xFE84, 0x8FC, 0x1884, 0x1C84, 0x2A84, 0x2AFC, 0x4884, 0x8884, 0x884, 0x884, 0x914, 0xA08}},
    {26499, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0xFD04, 0x1204, 0x3084, 0x3844, 0x5444, 0x5414, 0x9024, 0x1044, 0x1184, 0x1004, 0x1028, 0x1010}},
    {26500, 2, {0x1080, 0x1080, 0x1080, 0x10FC, 0xFD04, 0x1104, 0x3244, 0x3844, 0x5484, 0x54A4, 0x9114, 0x11F4, 0x1014, 0x1004, 0x1028, 0x1010}},
    {26501, 2, {0x1000, 0x11FC, 0x1088, 0x1088, 0xFC88, 0x1088, 0x3088, 0x3BFE, 0x5488, 0x5088, 0x9088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1208}},
    {26502, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x103C, 0x31E0, 0x3820, 0x5420, 0x503E, 0x93E0, 0x1020, 0x1022, 0x1022, 0x1022, 0x101E}},
    {26503, 2, {0x1010, 0x1110, 0x1110, 0x1112, 0xFD12, 0x1114, 0x31D8, 0x3910, 0x5510, 0x5510, 0x9110, 0x1112, 0x1152, 0x1192, 0x110E, 0x1000}},
    {26504, 2, {0x2080, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C, 0x2100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26505, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x3020, 0x3820, 0x55FC, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {26506, 2, {0x1000, 0x1040, 0x119E, 0x1112, 0xFD12, 0x1112, 0x3112, 0x3912, 0x5512, 0x5512, 0x915A, 0x1194, 0x1110, 0x1010, 0x1010, 0x1010}},
    {26507, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFC80, 0x1080, 0x3080, 0x38FC, 0x5484, 0x5484, 0x9084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {26508, 2, {0x1010, 0x1090, 0x1090, 0x1088, 0xFD08, 0x1104, 0x3A04, 0x35FA, 0x5088, 0x5088, 0x9088, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {26509, 2, {0x1090, 0x1090, 0x1108, 0x1108, 0xFA04, 0x1402, 0x33FC, 0x3880, 0x5500, 0x51F8, 0x9008, 0x1008, 0x1008, 0x1008, 0x1050, 0x1020}},
    {26510, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x3020, 0x3820, 0x57FE, 0x5420, 0x9050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {26511, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x13FE, 0x3108, 0x3908, 0x5508, 0x53FE, 0x9108, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {26512, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x1100, 0x31FE, 0x3910, 0x5510, 0x5110, 0x9110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {26513, 2, {0x1000, 0x17FE, 0x1080, 0x1080, 0xFC80, 0x10F8, 0x3908, 0x3508, 0x5108, 0x5208, 0x93F8, 0x1010, 0x1010, 0x1010, 0x17FE, 0x1000}},
    {26514, 2, {0x800, 0x9FE, 0x808, 0x888, 0xFE88, 0x888, 0x1908, 0x1DFE, 0x2A18, 0x2A28, 0x4848, 0x8888, 0x908, 0xA08, 0x828, 0x810}},
    {26515, 2, {0x808, 0x888, 0x848, 0x848, 0xFE08, 0x888, 0x1848, 0x1C48, 0x2A08, 0x2A0E, 0x49F8, 0x8808, 0x808, 0x808, 0x808, 0x808}},
    {26516, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0xFC88, 0x1144, 0x3222, 0x3820, 0x5400, 0x55FC, 0x9004, 0x1008, 0x1008, 0x1010, 0x1010, 0x1020}},
    {26517, 2, {0x1040, 0x1040, 0x1040, 0x13FC, 0xFA44, 0x1448, 0x3040, 0x3840, 0x54A0, 0x50A0, 0x90A0, 0x1120, 0x1122, 0x1222, 0x141E, 0x1800}},
    {26518, 2, {0x808, 0x83C, 0x9E0, 0x820, 0xFE20, 0x820, 0x19FE, 0x1C20, 0x2A20, 0x2A50, 0x4850, 0x8850, 0x888, 0x888, 0x904, 0xA02}},
    {26519, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x3070, 0x3870, 0x54A8, 0x50A8, 0x9124, 0x1222, 0x1420, 0x1020, 0x1020, 0x1020}},
    {26520, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0xFD22, 0x1122, 0x3122, 0x3922, 0x5552, 0x554A, 0x918A, 0x1102, 0x1102, 0x1102, 0x110A, 0x1104}},
    {26521, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0xFD78, 0x1148, 0x3148, 0x3948, 0x5548, 0x5568, 0x9150, 0x1142, 0x1142, 0x1142, 0x123E, 0x1400}},
    {26522, 2, {0x1040, 0x1040, 0x1040, 0x1080, 0xFCFE, 0x1108, 0x3288, 0x3888, 0x5488, 0x5450, 0x9050, 0x1020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26523, 2, {0x1008, 0x101C, 0x11F0, 0x1150, 0xFD50, 0x1150, 0x3150, 0x3950, 0x5550, 0x5548, 0x9148, 0x1148, 0x1144, 0x1244, 0x1242, 0x1400}},
    {26524, 2, {0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26525, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3020, 0x39FC, 0x5484, 0x5488, 0x9048, 0x1050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {26526, 2, {0x1110, 0x1110, 0x1110, 0x1110, 0xFD10, 0x1110, 0x3110, 0x3910, 0x5510, 0x52A8, 0x9268, 0x1228, 0x1444, 0x1444, 0x1884, 0x1102}},
    {26527, 2, {0x1000, 0x1000, 0x11FC, 0x1000, 0xFC00, 0x1000, 0x33FE, 0x3820, 0x5420, 0x5440, 0x9040, 0x1088, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {26528, 2, {0x1040, 0x1040, 0x11F8, 0x1048, 0xFC88, 0x108A, 0x310A, 0x3A26, 0x5420, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26529, 2, {0x1008, 0x1068, 0x1388, 0x1088, 0xFC88, 0x1088, 0x3088, 0x3BFE, 0x5488, 0x5488, 0x9088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {26530, 2, {0x1000, 0x13FC, 0x1200, 0x1208, 0xFE88, 0x1250, 0x3250, 0x3A20, 0x5620, 0x5250, 0x9250, 0x1288, 0x1308, 0x1200, 0x13FE, 0x1000}},
    {26531, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2388, 0x2548, 0x920, 0x3018, 0xC606, 0x180, 0x40, 0xE00, 0x180, 0x40}},
    {26532, 2, {0x820, 0x828, 0x824, 0x824, 0xFE20, 0x9FE, 0x1820, 0x1C20, 0x2A20, 0x2A50, 0x4850, 0x8850, 0x888, 0x888, 0x904, 0xA02}},
    {26533, 2, {0x2000, 0x23FE, 0x2200, 0x2220, 0xFA20, 0x2220, 0x22FC, 0x7224, 0x6A24, 0xA224, 0xA244, 0x2244, 0x2484, 0x2484, 0x2928, 0x3210}},
    {26534, 2, {0x1040, 0x1020, 0x1020, 0x11FC, 0xFD04, 0x1104, 0x3104, 0x39FC, 0x5504, 0x5500, 0x9100, 0x1100, 0x1100, 0x1200, 0x1200, 0x1400}},
    {26535, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD24, 0x1124, 0x3124, 0x3924, 0x5524, 0x5554, 0x9050, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {26536, 2, {0x1080, 0x1084, 0x1084, 0x1088, 0xFC90, 0x10A0, 0x3080, 0x3BFE, 0x54A0, 0x5490, 0x9090, 0x1088, 0x1084, 0x10A2, 0x10C0, 0x1080}},
    {26537, 2, {0x100, 0x100, 0x7FFC, 0x520, 0x910, 0x3108, 0xC544, 0x440, 0x820, 0x1110, 0x2108, 0xC206, 0x440, 0x820, 0x1FF0, 0x10}},
    {26538, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1208, 0x3406, 0x39F0, 0x5510, 0x5110, 0x9150, 0x1120, 0x1104, 0x1104, 0x10FC, 0x1000}},
    {26539, 2, {0x2000, 0x23F8, 0x2208, 0x2208, 0xFB28, 0x22A8, 0x22A8, 0x7248, 0x6A48, 0xA2A8, 0xA2A8, 0x232A, 0x240A, 0x240A, 0x2806, 0x3002}},
    {26540, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD48, 0x1128, 0x3928, 0x3508, 0x53FE, 0x5108, 0x9108, 0x1108, 0x1108, 0x1208, 0x1228, 0x1410}},
    {26541, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1210, 0x1150, 0x1020, 0x1FFC, 0x104, 0x104, 0x7FF4, 0x388, 0x540, 0x1930, 0xE10E, 0x100}},
    {26542, 2, {0x810, 0x810, 0x810, 0x810, 0xFE1E, 0x810, 0x1810, 0x1C10, 0x2AFE, 0x2A82, 0x4882, 0x8882, 0x882, 0x882, 0x8FE, 0x882}},
    {26543, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x3020, 0x3820, 0x55FC, 0x5504, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26544, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFD24, 0x10A4, 0x30A8, 0x3820, 0x57FE, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26545, 2, {0x820, 0x820, 0x820, 0x840, 0xFE48, 0x884, 0x19FE, 0x1C82, 0x2A00, 0x2AFC, 0x4884, 0x8884, 0x884, 0x884, 0x8FC, 0x884}},
    {26546, 2, {0x400, 0x800, 0x1010, 0x3FF8, 0x8, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26547, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x3104, 0x3904, 0x55FC, 0x5504, 0x9000, 0x1090, 0x1088, 0x1104, 0x1202, 0x1402}},
    {26548, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x3000, 0x3800, 0x57FC, 0x5444, 0x9044, 0x1084, 0x1084, 0x1104, 0x1228, 0x1410}},
    {26549, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x3000, 0x3BFE, 0x5480, 0x5500, 0x91F8, 0x1008, 0x1008, 0x1008, 0x1050, 0x1020}},
    {26550, 2, {0x800, 0x800, 0x7F7C, 0x944, 0x1144, 0x1144, 0x257C, 0x4200, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26551, 2, {0x2200, 0x2200, 0x2200, 0xFA1E, 0x27D2, 0x2252, 0x7252, 0x6A52, 0xAA52, 0xA252, 0x2252, 0x2252, 0x2452, 0x245E, 0x2952, 0x3080}},
    {26552, 2, {0x1080, 0x1080, 0x1080, 0x11FC, 0xFD04, 0x1204, 0x31E4, 0x3924, 0x5524, 0x5524, 0x9124, 0x11E4, 0x1124, 0x1004, 0x1028, 0x1010}},
    {26553, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0xFA04, 0x15F4, 0x3114, 0x3914, 0x5514, 0x51F4, 0x9104, 0x1128, 0x1112, 0x1102, 0x10FE, 0x1000}},
    {26554, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3020, 0x39FC, 0x5470, 0x54A8, 0x90A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {26555, 2, {0x1048, 0x1148, 0x1148, 0x1148, 0xFD48, 0x13FE, 0x3148, 0x3948, 0x5548, 0x5548, 0x9178, 0x1100, 0x1100, 0x1100, 0x11FE, 0x1000}},
    {26556, 2, {0x1220, 0x1220, 0xFFFE, 0x1220, 0x1220, 0x13E0, 0x1000, 0x1FF8, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26557, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1010, 0xFFFE, 0x1010, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26558, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3124, 0x3924, 0x5524, 0x55FC, 0x9124, 0x1124, 0x1224, 0x1224, 0x1424, 0x1020}},
    {26559, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x38A8, 0x3488, 0x5088, 0x57FE, 0x9088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {26560, 2, {0x1010, 0x1010, 0x1010, 0x11FE, 0xFD12, 0x1114, 0x3110, 0x39FC, 0x5544, 0x5544, 0x9128, 0x1128, 0x1110, 0x1228, 0x1244, 0x1482}},
    {26561, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFD02, 0x1204, 0x3080, 0x3888, 0x5490, 0x54A0, 0x90C0, 0x1082, 0x1082, 0x1082, 0x107E, 0x1000}},
    {26562, 2, {0x1080, 0x1080, 0x10FE, 0x1100, 0xFA20, 0x1120, 0x312C, 0x3974, 0x57A4, 0x5124, 0x9134, 0x1128, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {26563, 2, {0x1020, 0x1020, 0x1050, 0x1050, 0xFC88, 0x1124, 0x3212, 0x3810, 0x55FC, 0x5404, 0x9008, 0x1088, 0x1050, 0x1020, 0x1010, 0x1010}},
    {26564, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x3222, 0x3A22, 0x5652, 0x524A, 0x928A, 0x1302, 0x1202, 0x1202, 0x120A, 0x1204}},
    {26565, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x3100, 0x3940, 0x5544, 0x5548, 0x9170, 0x1140, 0x1142, 0x1242, 0x123E, 0x1400}},
    {26566, 2, {0x1040, 0x1020, 0x1020, 0x1000, 0xFDFE, 0x1000, 0x3004, 0x3884, 0x5484, 0x5448, 0x9048, 0x1048, 0x1050, 0x1010, 0x13FE, 0x1000}},
    {26567, 2, {0x1038, 0x11C0, 0x1040, 0x1040, 0xFC40, 0x13FC, 0x3040, 0x38E0, 0x54E0, 0x5550, 0x9150, 0x1248, 0x1444, 0x1842, 0x1040, 0x1040}},
    {26568, 2, {0x1020, 0x1020, 0x1124, 0x10A4, 0xFCA8, 0x1020, 0x31FC, 0x3820, 0x5420, 0x5420, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26569, 2, {0x804, 0x81E, 0x9E0, 0x800, 0xFE20, 0x810, 0x19FE, 0x1C04, 0x2A08, 0x2A10, 0x4820, 0x8840, 0x880, 0x940, 0xA3E, 0x800}},
    {26570, 2, {0x1040, 0x1040, 0x107C, 0x1084, 0xFD88, 0x1250, 0x3020, 0x3850, 0x5488, 0x5306, 0x9060, 0x1010, 0x1008, 0x10C0, 0x1020, 0x1010}},
    {26571, 2, {0x890, 0x888, 0x1080, 0x309E, 0x57E0, 0x9080, 0x1044, 0x1034, 0x110C, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26572, 2, {0x1000, 0x11FC, 0x1004, 0x1004, 0xFDF4, 0x1004, 0x3004, 0x39F4, 0x5514, 0x5514, 0x91F4, 0x1114, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26573, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x3124, 0x3924, 0x5524, 0x53FE, 0x9020, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26574, 2, {0x1088, 0x1088, 0x1088, 0x1108, 0xFD7E, 0x1308, 0x3508, 0x3948, 0x5528, 0x5528, 0x9108, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {26575, 2, {0x1020, 0x1020, 0x1040, 0x11FC, 0xFD04, 0x1104, 0x3104, 0x3904, 0x55FC, 0x5504, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26576, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0xFE0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26577, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFE, 0x1088, 0x3088, 0x3888, 0x5488, 0x54F8, 0x9088, 0x1088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {26578, 2, {0x2080, 0x1080, 0x9C, 0x47E0, 0x2080, 0x884, 0x1084, 0x607C, 0x2100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26579, 2, {0x2080, 0x1080, 0x3F0, 0x4090, 0x2110, 0x912, 0x1212, 0x640E, 0x2900, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26580, 2, {0x0, 0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26581, 2, {0x1000, 0x11F8, 0x1008, 0x1050, 0xFC20, 0x1010, 0x33FE, 0x3862, 0x54A4, 0x54A0, 0x9120, 0x1220, 0x1420, 0x1020, 0x10A0, 0x1040}},
    {26582, 2, {0x1000, 0x13FC, 0x1084, 0x1084, 0xFC84, 0x1104, 0x3114, 0x3A08, 0x5400, 0x51FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26583, 2, {0x1010, 0x1090, 0x1090, 0x1108, 0xFD08, 0x1204, 0x3402, 0x39F8, 0x5508, 0x5508, 0x9108, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {26584, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1040, 0x3040, 0x38FC, 0x5584, 0x5284, 0x9484, 0x1084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {26585, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x3124, 0x3924, 0x5524, 0x55FC, 0x9124, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26586, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x3124, 0x3924, 0x5524, 0x55FC, 0x9124, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {26587, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x3124, 0x39FC, 0x5524, 0x5524, 0x9124, 0x11FC, 0x1124, 0x1020, 0x1020, 0x1020}},
    {26588, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0xFD00, 0x11FC, 0x3104, 0x3904, 0x5504, 0x5504, 0x91FC, 0x1100, 0x1100, 0x1100, 0x11FE, 0x1000}},
    {26589, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x1100, 0x31FE, 0x3910, 0x5510, 0x5530, 0x9118, 0x1114, 0x1112, 0x1210, 0x1210, 0x1410}},
    {26590, 2, {0x1080, 0x1080, 0x1080, 0x10FE, 0xFD40, 0x1140, 0x3240, 0x387C, 0x5440, 0x5040, 0x9040, 0x107E, 0x1040, 0x1040, 0x1040, 0x1040}},
    {26591, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x31FC, 0x3924, 0x5524, 0x5524, 0x93FE, 0x1104, 0x1104, 0x1104, 0x1114, 0x1108}},
    {26592, 2, {0x2040, 0x2020, 0x2020, 0x23FE, 0xFA02, 0x2404, 0x2000, 0x7000, 0x6BFE, 0xA020, 0xA020, 0x2020, 0x2020, 0x2020, 0x20A0, 0x2040}},
    {26593, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0xFBFE, 0x1000, 0x3020, 0x3822, 0x57B2, 0x50B4, 0x90A8, 0x1128, 0x1124, 0x1224, 0x14A2, 0x1040}},
    {26594, 2, {0x1008, 0x103C, 0x13E0, 0x1220, 0xFE20, 0x1220, 0x3220, 0x3BFE, 0x5620, 0x5210, 0x9210, 0x1212, 0x120A, 0x128A, 0x1326, 0x1212}},
    {26595, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0xFDFC, 0x1120, 0x3220, 0x3820, 0x57FE, 0x5420, 0x9050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {26596, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x3108, 0x3908, 0x5508, 0x55F8, 0x9108, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {26597, 2, {0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x1110, 0x3FF8, 0xD016, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x0}},
    {26598, 2, {0x800, 0x800, 0x8FC, 0x884, 0xFE84, 0x884, 0x18FC, 0x1C84, 0x2A84, 0x2A84, 0x48FC, 0x8884, 0x800, 0x800, 0x9FE, 0x800}},
    {26599, 2, {0x1008, 0x101C, 0x11F0, 0x1150, 0xFD50, 0x1150, 0x3150, 0x3950, 0x5550, 0x5548, 0x9148, 0x1168, 0x1154, 0x1274, 0x1252, 0x1400}},
    {26600, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0xFC80, 0x10A0, 0x3920, 0x35FC, 0x5324, 0x5524, 0x9124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {26601, 2, {0x1000, 0x13FC, 0x1200, 0x1240, 0xFE40, 0x1278, 0x3288, 0x3A90, 0x5710, 0x5220, 0x9250, 0x1288, 0x1304, 0x1200, 0x13FE, 0x1000}},
    {26602, 2, {0x1010, 0x1090, 0x1090, 0x1110, 0xFD7E, 0x1252, 0x3392, 0x3892, 0x5512, 0x5112, 0x9252, 0x13D2, 0x1062, 0x1022, 0x104A, 0x1084}},
    {26603, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0xFC94, 0x1094, 0x33FC, 0x3A90, 0x5690, 0x53FE, 0x9092, 0x1092, 0x111A, 0x1114, 0x1210, 0x1410}},
    {26604, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2928, 0x2548, 0x3FF8, 0x2388, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26605, 2, {0x1050, 0x1048, 0x1048, 0x1040, 0xFBFE, 0x1080, 0x3080, 0x38FC, 0x5544, 0x5144, 0x9128, 0x1128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {26606, 2, {0x1020, 0x1020, 0x1124, 0x1124, 0xFD24, 0x1124, 0x31FC, 0x3820, 0x5420, 0x5524, 0x9124, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {26607, 2, {0x1000, 0x1000, 0x17FE, 0x1008, 0xFC08, 0x13C8, 0x3248, 0x3A48, 0x5648, 0x5248, 0x93C8, 0x1248, 0x1008, 0x1008, 0x1028, 0x1010}},
    {26608, 2, {0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xCFE6, 0x0, 0x0, 0x7FFC, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {26609, 2, {0x1040, 0x1020, 0x1000, 0x13FE, 0xFC20, 0x1020, 0x3020, 0x3820, 0x55FC, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26610, 2, {0x1000, 0x1040, 0x1020, 0x1028, 0xFC08, 0x1088, 0x3090, 0x3A94, 0x56A2, 0x52A2, 0x94C2, 0x1088, 0x1188, 0x1288, 0x1478, 0x1000}},
    {26611, 2, {0x1000, 0x1080, 0x131E, 0x1252, 0xFE52, 0x1252, 0x3252, 0x3A52, 0x5652, 0x52D2, 0x935A, 0x1254, 0x1090, 0x1090, 0x1110, 0x1210}},
    {26612, 2, {0x880, 0x888, 0x2890, 0x2EA0, 0x28C0, 0x2884, 0x2E84, 0xF07C, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26613, 2, {0x1000, 0x11FC, 0x1154, 0x1154, 0xFD54, 0x1154, 0x3154, 0x3BFE, 0x5554, 0x5554, 0x9154, 0x1154, 0x1154, 0x1144, 0x1114, 0x1108}},
    {26614, 2, {0x1000, 0x1000, 0x13FE, 0x1252, 0xFE52, 0x1252, 0x3252, 0x3A52, 0x5652, 0x528E, 0x9282, 0x1302, 0x1202, 0x13FE, 0x1202, 0x1000}},
    {26615, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x1104, 0x31FC, 0x3850, 0x5450, 0x5450, 0x9050, 0x1092, 0x1092, 0x1112, 0x120E, 0x1400}},
    {26616, 2, {0x1000, 0x11FE, 0x1010, 0x1010, 0xFC20, 0x1020, 0x3068, 0x38A4, 0x5522, 0x5222, 0x9020, 0x1020, 0x1020, 0x1000, 0x13FE, 0x1000}},
    {26617, 2, {0x1020, 0x1020, 0x102E, 0x11F0, 0xFD20, 0x1120, 0x3120, 0x39FE, 0x5422, 0x5462, 0x90A2, 0x112A, 0x1624, 0x1020, 0x1020, 0x1020}},
    {26618, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x3040, 0x3840, 0x55FC, 0x5444, 0x9044, 0x1084, 0x1084, 0x1104, 0x1228, 0x1410}},
    {26619, 2, {0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {26620, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1404, 0x3000, 0x3890, 0x5490, 0x5090, 0x9088, 0x1108, 0x1108, 0x1104, 0x1204, 0x1402}},
    {26621, 2, {0x1000, 0x11FC, 0x1084, 0x1088, 0xFC50, 0x1020, 0x3050, 0x3888, 0x5726, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26622, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x3120, 0x3920, 0x553C, 0x5520, 0x9120, 0x1120, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {26623, 2, {0x1040, 0x1020, 0x1000, 0x13FE, 0xFC20, 0x1020, 0x31FC, 0x3924, 0x5524, 0x5524, 0x9124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {26624, 2, {0x101C, 0x11E0, 0x1100, 0x1100, 0xFDFE, 0x1100, 0x3100, 0x397C, 0x5544, 0x5544, 0x9154, 0x1148, 0x1142, 0x1242, 0x123E, 0x1400}},
    {26625, 2, {0x2100, 0x2100, 0x211E, 0x21D2, 0xFA52, 0x2252, 0x2252, 0x7552, 0x6892, 0xA092, 0xA11A, 0x2114, 0x2210, 0x2410, 0x2810, 0x2010}},
    {26626, 2, {0x1000, 0x10FC, 0x1084, 0x10A4, 0xFC94, 0x1084, 0x3084, 0x3BFE, 0x5504, 0x5544, 0x9124, 0x1104, 0x11FE, 0x1004, 0x1028, 0x1010}},
    {26627, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x11FE, 0x3120, 0x3920, 0x553C, 0x5524, 0x9124, 0x1124, 0x1124, 0x1244, 0x1254, 0x1488}},
    {26628, 2, {0x2208, 0x1108, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8104, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {26629, 2, {0x1000, 0x11DC, 0x1154, 0x1154, 0xFD54, 0x1154, 0x3154, 0x3BFE, 0x5554, 0x5154, 0x9154, 0x1154, 0x1154, 0x12D4, 0x1224, 0x144C}},
    {26630, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x280, 0x440, 0x820, 0x3318, 0xC0C6, 0x600, 0x180, 0x40}},
    {26631, 2, {0x1000, 0x11FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x3020, 0x3820, 0x5528, 0x5524, 0x9224, 0x1222, 0x1422, 0x1020, 0x10A0, 0x1040}},
    {26632, 2, {0x1050, 0x1048, 0x1040, 0x105C, 0xFDE0, 0x1040, 0x305E, 0x39E0, 0x5444, 0x5048, 0x9030, 0x1022, 0x1052, 0x108A, 0x1306, 0x1002}},
    {26633, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x1000, 0x33FE, 0x3842, 0x5442, 0x5442, 0x9042, 0x104A, 0x1044, 0x1040, 0x1040, 0x1040}},
    {26634, 2, {0x1090, 0x1088, 0x1088, 0x1080, 0xFBFE, 0x10A0, 0x30A0, 0x38A4, 0x54A4, 0x5128, 0x9128, 0x1132, 0x1222, 0x1262, 0x149E, 0x1800}},
    {26635, 2, {0x1040, 0x1040, 0x1040, 0x17FE, 0xF880, 0x1120, 0x3120, 0x3A20, 0x57FC, 0x5020, 0x9128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {26636, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFC20, 0x11FC, 0x3104, 0x3904, 0x5504, 0x55FC, 0x9104, 0x1100, 0x1100, 0x1200, 0x1200, 0x1400}},
    {26637, 2, {0x1020, 0x1020, 0x1120, 0x1120, 0xFDFC, 0x1120, 0x3220, 0x3820, 0x5420, 0x55FC, 0x9020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26638, 2, {0x1008, 0x103C, 0x13C0, 0x1200, 0xFE20, 0x1220, 0x3220, 0x3BFE, 0x5420, 0x5420, 0x9128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {26639, 2, {0x1000, 0x1108, 0x1088, 0x1090, 0xFC00, 0x13FE, 0x3000, 0x3800, 0x5400, 0x55FC, 0x9000, 0x1000, 0x1000, 0x1000, 0x13FE, 0x1000}},
    {26640, 2, {0x1040, 0x1020, 0x1010, 0x11E0, 0xFC22, 0x1034, 0x33B8, 0x38B0, 0x54A8, 0x54A8, 0x9128, 0x1124, 0x1224, 0x1422, 0x10A0, 0x1040}},
    {26641, 2, {0x2004, 0x2004, 0x2784, 0x2084, 0xF8BE, 0x2484, 0x2284, 0x72A4, 0x6914, 0xA114, 0xA284, 0x2284, 0x2484, 0x2804, 0x2014, 0x2008}},
    {26642, 2, {0x1080, 0x1080, 0x11FC, 0x1104, 0xFE04, 0x11E4, 0x3124, 0x3924, 0x55E4, 0x5524, 0x9124, 0x11E4, 0x1104, 0x1004, 0x1028, 0x1010}},
    {26643, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x1202, 0x31FC, 0x3820, 0x5420, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26644, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x900, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26645, 2, {0x1000, 0x11FE, 0x1120, 0x1120, 0xFD20, 0x11FC, 0x3104, 0x3904, 0x5504, 0x5504, 0x91FC, 0x1120, 0x1120, 0x1120, 0x11FE, 0x1000}},
    {26646, 2, {0x1000, 0x17FE, 0x1090, 0x1090, 0xFC90, 0x13FC, 0x3294, 0x3A94, 0x5694, 0x5294, 0x929C, 0x1304, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {26647, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26648, 2, {0x1020, 0x1020, 0x107C, 0x1084, 0xFD48, 0x1030, 0x3020, 0x3848, 0x5590, 0x543E, 0x9042, 0x11A4, 0x1018, 0x1010, 0x1060, 0x1180}},
    {26649, 2, {0x2080, 0x2080, 0x20FC, 0x2108, 0xFA90, 0x2060, 0x2198, 0x7626, 0x6820, 0xA1FC, 0xA020, 0x2220, 0x23FE, 0x2020, 0x2020, 0x2020}},
    {26650, 2, {0x1088, 0x1048, 0x1050, 0x1000, 0xFDFC, 0x1020, 0x3020, 0x3820, 0x55FE, 0x5420, 0x9050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {26651, 2, {0x1040, 0x1040, 0x11FC, 0x1044, 0xFC84, 0x1084, 0x3128, 0x3A10, 0x5508, 0x53DE, 0x914A, 0x114A, 0x114A, 0x1252, 0x135A, 0x14A4}},
    {26652, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x3124, 0x3924, 0x5524, 0x552C, 0x9070, 0x10A8, 0x1124, 0x1622, 0x1020, 0x1020}},
    {26653, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0xFC20, 0x13FE, 0x3020, 0x3820, 0x5420, 0x55FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26654, 2, {0x0, 0x7EFC, 0x820, 0x820, 0xFEFE, 0x820, 0x1020, 0x2020, 0x4120, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26655, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFDFE, 0x1088, 0x3088, 0x3888, 0x5488, 0x53FE, 0x9088, 0x1088, 0x1108, 0x1108, 0x1208, 0x1408}},
    {26656, 2, {0x838, 0x13C0, 0x3040, 0x5040, 0x97FC, 0x1040, 0x1040, 0x13F8, 0x0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26657, 2, {0x1040, 0x1020, 0x1020, 0x11FE, 0xFC00, 0x1088, 0x3104, 0x3A02, 0x5488, 0x5088, 0x9050, 0x1050, 0x1020, 0x1050, 0x1088, 0x1306}},
    {26658, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFC40, 0x11FC, 0x3104, 0x3904, 0x5504, 0x55FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26659, 2, {0x1080, 0x1088, 0x109C, 0x1170, 0xFD10, 0x1310, 0x3510, 0x3910, 0x55FE, 0x5510, 0x9110, 0x1110, 0x1110, 0x1110, 0x117C, 0x1100}},
    {26660, 2, {0x2020, 0x2420, 0x2220, 0x2220, 0xF832, 0x21B2, 0x20B4, 0x72A8, 0x6AA8, 0xA4A8, 0xA524, 0x2924, 0x2A22, 0x2020, 0x20A0, 0x2040}},
    {26661, 2, {0x100, 0x4100, 0x21FC, 0xA44, 0x1448, 0xE0A0, 0x2110, 0x2608, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26662, 2, {0x1082, 0x1092, 0x1092, 0x1092, 0xFC92, 0x1092, 0x32DA, 0x3AB6, 0x5492, 0x5092, 0x9092, 0x1092, 0x1092, 0x1112, 0x1102, 0x1202}},
    {26663, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x3124, 0x3924, 0x55FC, 0x5420, 0x9024, 0x1018, 0x1032, 0x10CA, 0x1306, 0x1002}},
    {26664, 2, {0x1010, 0x1010, 0x1210, 0x113E, 0xFD22, 0x1044, 0x3810, 0x3490, 0x5090, 0x5310, 0x9128, 0x1128, 0x1128, 0x1144, 0x1044, 0x1082}},
    {26665, 2, {0x1000, 0x13DE, 0x1042, 0x1042, 0xFE52, 0x114A, 0x314A, 0x3842, 0x54C6, 0x514A, 0x9252, 0x1042, 0x1042, 0x1042, 0x114A, 0x1084}},
    {26666, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x3020, 0x3BFE, 0x5470, 0x54A8, 0x90A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {26667, 2, {0x1040, 0x1040, 0x1040, 0x17FE, 0xF880, 0x1080, 0x317C, 0x3908, 0x5710, 0x5510, 0x91FE, 0x1110, 0x1110, 0x1110, 0x1150, 0x1120}},
    {26668, 2, {0x1040, 0x1040, 0x10F8, 0x1108, 0xFA10, 0x11FC, 0x3124, 0x3924, 0x5524, 0x51FC, 0x9100, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {26669, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFC40, 0x11FC, 0x3154, 0x3954, 0x5554, 0x5554, 0x9154, 0x1154, 0x1154, 0x1144, 0x1114, 0x1108}},
    {26670, 2, {0x1000, 0x13FE, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x3108, 0x3908, 0x55F8, 0x5508, 0x9108, 0x111E, 0x17E8, 0x1008, 0x1008, 0x1008}},
    {26671, 2, {0x1040, 0x1040, 0x17FE, 0x1080, 0xFD00, 0x13FC, 0x3504, 0x3904, 0x55FC, 0x5504, 0x9104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {26672, 2, {0x10A8, 0x10A4, 0x10A4, 0x1120, 0xFD2E, 0x13F0, 0x3520, 0x3924, 0x5524, 0x5524, 0x9128, 0x1110, 0x1132, 0x114A, 0x1186, 0x1102}},
    {26673, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0xFDFE, 0x1088, 0x3088, 0x3888, 0x5488, 0x53FE, 0x9000, 0x1088, 0x1084, 0x1104, 0x1202, 0x1402}},
    {26674, 2, {0x1040, 0x1044, 0x11F4, 0x1048, 0xFC50, 0x13FE, 0x3040, 0x3880, 0x55FE, 0x5240, 0x9480, 0x10FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26675, 2, {0x1040, 0x1040, 0x1044, 0x11F4, 0xFC48, 0x1050, 0x33FE, 0x3840, 0x5480, 0x5584, 0x9298, 0x14E0, 0x1082, 0x1082, 0x107E, 0x1000}},
    {26676, 2, {0x1080, 0x1080, 0x10FE, 0x1100, 0xFA00, 0x10FC, 0x3084, 0x38A4, 0x5494, 0x5084, 0x93FE, 0x1084, 0x1084, 0x1104, 0x1114, 0x1208}},
    {26677, 2, {0x1002, 0x1002, 0x17E2, 0x110A, 0xF90A, 0x11EA, 0x312A, 0x3A2A, 0x572A, 0x52AA, 0x944A, 0x104A, 0x1082, 0x1102, 0x120A, 0x1404}},
    {26678, 2, {0x1000, 0x13FC, 0x1204, 0x1244, 0xFE44, 0x1244, 0x33FC, 0x3A44, 0x5644, 0x52A4, 0x9294, 0x1314, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {26679, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFBFE, 0x1020, 0x3020, 0x39FC, 0x5420, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26680, 2, {0x1020, 0x1010, 0x1010, 0x13FE, 0xFC20, 0x1042, 0x3084, 0x39F8, 0x5410, 0x5422, 0x90C4, 0x1308, 0x1010, 0x1028, 0x10C4, 0x1302}},
    {26681, 2, {0x1000, 0x13F8, 0x1208, 0x1208, 0xFBF8, 0x1208, 0x3208, 0x3BF8, 0x5644, 0x5248, 0x9230, 0x1220, 0x1210, 0x1288, 0x1306, 0x1200}},
    {26682, 2, {0x1100, 0x1104, 0x1138, 0x11C0, 0xFD02, 0x1102, 0x30FE, 0x3800, 0x55FC, 0x5504, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26683, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0xFBFE, 0x1020, 0x3020, 0x3BA0, 0x5520, 0x5110, 0x9110, 0x1110, 0x11CA, 0x170A, 0x1206, 0x1002}},
    {26684, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFB10, 0x14A0, 0x3040, 0x38A0, 0x5518, 0x5206, 0x95F8, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {26685, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x40, 0x844, 0x844, 0xFF44, 0x1828, 0x2C2A, 0x2A12, 0x492A, 0x8846, 0x882}},
    {26686, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x1448, 0x2444, 0x4444, 0x0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26687, 2, {0x1090, 0x1094, 0x1092, 0x1112, 0xFD10, 0x137E, 0x3110, 0x3910, 0x5510, 0x5528, 0x9128, 0x1128, 0x1128, 0x1144, 0x1144, 0x1182}},
    {26688, 2, {0x1010, 0x1010, 0x3EFC, 0x4210, 0xA490, 0x1490, 0x8FE, 0x1010, 0x2010, 0x4110, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26689, 2, {0x1080, 0x1080, 0x113C, 0x1200, 0xFC80, 0x1080, 0x317E, 0x3B08, 0x5508, 0x5508, 0x9108, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {26690, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x33FE, 0x3800, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26691, 2, {0x1090, 0x1090, 0x1090, 0x1292, 0xFD94, 0x1098, 0x3090, 0x3998, 0x5694, 0x5492, 0x9090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {26692, 2, {0x1020, 0x1020, 0x1124, 0x10A4, 0xFCA8, 0x1020, 0x33FE, 0x3890, 0x5490, 0x5490, 0x9090, 0x1092, 0x1112, 0x1112, 0x120E, 0x1400}},
    {26693, 2, {0x1040, 0x1040, 0x10FC, 0x1104, 0xFE08, 0x11FE, 0x3100, 0x397C, 0x5544, 0x5144, 0x9154, 0x1148, 0x1142, 0x1242, 0x123E, 0x1400}},
    {26694, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0xFDFE, 0x1110, 0x3110, 0x397C, 0x5510, 0x5510, 0x9110, 0x11FE, 0x1100, 0x1100, 0x11FE, 0x1000}},
    {26695, 2, {0x1000, 0x1000, 0xFE7C, 0x2244, 0x4244, 0x3444, 0x87C, 0x3400, 0xC300, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26696, 2, {0x100, 0x7FFC, 0x4004, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26697, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFE02, 0x1444, 0x3040, 0x3BFE, 0x5488, 0x5488, 0x9108, 0x10D0, 0x1020, 0x1050, 0x1088, 0x1304}},
    {26698, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x820, 0x1110, 0x2108, 0xDFF6, 0x300, 0x5C0, 0x1930, 0x6108, 0x100}},
    {26699, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1024, 0x31FC, 0x3920, 0x55FE, 0x5422, 0x902A, 0x1054, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26700, 2, {0x100, 0x100, 0x1FC, 0x100, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26701, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x1104, 0x32FA, 0x3800, 0x57FE, 0x5440, 0x9080, 0x11FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26702, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0xFC88, 0x1104, 0x33FE, 0x3822, 0x5420, 0x5420, 0x93FE, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {26703, 2, {0x1000, 0x101E, 0x13F2, 0x1092, 0xFC94, 0x1094, 0x3098, 0x3894, 0x5492, 0x5492, 0x90F2, 0x171A, 0x1214, 0x1010, 0x1010, 0x1010}},
    {26704, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFEF4, 0x1204, 0x3204, 0x3AF4, 0x5694, 0x5294, 0x9294, 0x12F4, 0x1204, 0x1204, 0x1214, 0x1208}},
    {26705, 2, {0x1FE0, 0x640, 0x180, 0xE40, 0x0, 0x7CF8, 0x2448, 0x1830, 0x66CC, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26706, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x920, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26707, 2, {0x1000, 0x13FE, 0x1000, 0x1000, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1104, 0x1000, 0x1000, 0x13FE, 0x1000}},
    {26708, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x31FC, 0x3800, 0x5400, 0x55FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26709, 2, {0x1000, 0x1040, 0x119C, 0x1104, 0xFD04, 0x1104, 0x3104, 0x39DC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104, 0x1000}},
    {26710, 2, {0x1020, 0x1020, 0x1040, 0x1080, 0xFDFC, 0x1154, 0x3154, 0x3954, 0x5554, 0x5554, 0x9154, 0x1154, 0x1154, 0x1154, 0x13FE, 0x1000}},
    {26711, 2, {0x1000, 0x13F8, 0x1088, 0x1090, 0xFCBC, 0x1104, 0x3914, 0x3648, 0x5040, 0x57FE, 0x9040, 0x10E0, 0x1150, 0x1248, 0x1446, 0x1040}},
    {26712, 2, {0x1040, 0x1080, 0x11F8, 0x1108, 0xFD08, 0x1108, 0x31F8, 0x3900, 0x5500, 0x5500, 0x91FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26713, 2, {0x1040, 0x1040, 0x1090, 0x1108, 0xFBFC, 0x1024, 0x3120, 0x3920, 0x55FC, 0x5220, 0x9020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26714, 2, {0x1124, 0x1124, 0x1248, 0x1124, 0xFD24, 0x1000, 0x3080, 0x38FC, 0x5504, 0x5184, 0x9248, 0x1428, 0x1010, 0x1020, 0x10C0, 0x1300}},
    {26715, 2, {0x820, 0x820, 0x83E, 0x820, 0xFE20, 0x9FE, 0x1800, 0x1C00, 0x2BFE, 0x2A20, 0x4828, 0x8824, 0x822, 0x820, 0x820, 0x820}},
    {26716, 2, {0x1040, 0x1024, 0x1224, 0x1108, 0xFD08, 0x1010, 0x3040, 0x3840, 0x57FE, 0x5488, 0x9108, 0x1190, 0x1060, 0x1050, 0x1188, 0x1604}},
    {26717, 2, {0x2208, 0x2208, 0x2208, 0x2388, 0xFABE, 0x2288, 0x24A8, 0x74A8, 0x6AA8, 0xA128, 0xA13E, 0x2108, 0x2208, 0x2208, 0x2408, 0x2808}},
    {26718, 2, {0x2100, 0x2100, 0x211E, 0x21D2, 0xFA54, 0x2254, 0x2258, 0x7554, 0x6892, 0xA092, 0xA112, 0x211A, 0x2214, 0x2410, 0x2810, 0x2010}},
    {26719, 2, {0x1050, 0x1048, 0x107E, 0x13C0, 0xFC40, 0x107C, 0x3BC0, 0x3440, 0x507E, 0x53C0, 0x9044, 0x1028, 0x1032, 0x10CA, 0x1706, 0x1002}},
    {26720, 2, {0x1000, 0x11FE, 0x1048, 0x1048, 0xFC48, 0x1048, 0x3248, 0x394A, 0x554A, 0x554C, 0x9048, 0x1048, 0x1048, 0x1048, 0x13FE, 0x1000}},
    {26721, 2, {0x1080, 0x1080, 0x10BC, 0x13C0, 0xFC50, 0x1024, 0x30D4, 0x3B0C, 0x5400, 0x53FE, 0x9090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {26722, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFC20, 0x11FC, 0x3104, 0x3924, 0x5524, 0x5524, 0x9124, 0x1124, 0x1050, 0x1048, 0x1084, 0x1104}},
    {26723, 2, {0x1020, 0x1124, 0x10A4, 0x10A4, 0xFCA8, 0x1020, 0x33FC, 0x3804, 0x5404, 0x5004, 0x91FC, 0x1004, 0x1004, 0x1004, 0x13FC, 0x1004}},
    {26724, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1000, 0x31FC, 0x3804, 0x5404, 0x55FC, 0x9100, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {26725, 2, {0x2010, 0x2078, 0x23C0, 0x2040, 0xF840, 0x27FE, 0x20A0, 0x7110, 0x6A08, 0xA516, 0xA110, 0x2110, 0x2110, 0x2210, 0x2210, 0x2410}},
    {26726, 2, {0x1090, 0x1094, 0x1114, 0x1118, 0xFB10, 0x1532, 0x3152, 0x390E, 0x5520, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26727, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x1202, 0x38F8, 0x3400, 0x5000, 0x53FE, 0x9020, 0x1040, 0x1088, 0x1104, 0x13FE, 0x1102}},
    {26728, 2, {0x840, 0x4840, 0x28FC, 0x904, 0x1A88, 0x2850, 0x4860, 0x8980, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26729, 2, {0x1020, 0x1010, 0x1010, 0x11FE, 0xFD00, 0x1110, 0x3110, 0x3910, 0x557C, 0x5510, 0x9110, 0x1110, 0x1110, 0x1210, 0x12FE, 0x1400}},
    {26730, 2, {0x1000, 0x13FC, 0x1004, 0x1004, 0xFDFC, 0x1004, 0x3004, 0x3BFC, 0x5408, 0x5008, 0x97FE, 0x1108, 0x1088, 0x1088, 0x1028, 0x1010}},
    {26731, 2, {0x2010, 0x2210, 0x2110, 0x2110, 0xF454, 0x2252, 0x2252, 0x7090, 0x6810, 0xA114, 0xA604, 0x2208, 0x2208, 0x2210, 0x2220, 0x20C0}},
    {26732, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x922, 0x7228, 0x1030, 0x10C0, 0x700, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26733, 2, {0x1000, 0x13FC, 0x1200, 0x1200, 0xFEF8, 0x1200, 0x3200, 0x3BFC, 0x56A0, 0x52A4, 0x92A8, 0x1290, 0x1290, 0x1488, 0x14A4, 0x18C2}},
    {26734, 2, {0x1000, 0x13FE, 0x1010, 0x1020, 0xFC68, 0x10A4, 0x3322, 0x3820, 0x5400, 0x55FC, 0x9104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26735, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x3000, 0x3800, 0x57FE, 0x5420, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26736, 2, {0x101C, 0x13E0, 0x1220, 0x1220, 0xFBFE, 0x1220, 0x3290, 0x3B0A, 0x5606, 0x51FA, 0x9108, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {26737, 2, {0x1000, 0x13FC, 0x1000, 0x1124, 0xFA48, 0x1490, 0x3248, 0x3924, 0x5400, 0x53FC, 0x9040, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {26738, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x13FE, 0x3A02, 0x3404, 0x51F8, 0x5010, 0x9020, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26739, 2, {0x1040, 0x1080, 0x1104, 0x13FE, 0xFC22, 0x1020, 0x33FE, 0x3850, 0x5488, 0x5124, 0x9222, 0x11FC, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26740, 2, {0x1008, 0x103C, 0x11C0, 0x1004, 0xFD44, 0x10A8, 0x3000, 0x39F8, 0x5410, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26741, 2, {0x2008, 0x203C, 0x23C0, 0x2044, 0xFA24, 0x2128, 0x2100, 0x7040, 0x6BFE, 0xA088, 0xA108, 0x2390, 0x2060, 0x2050, 0x2188, 0x2604}},
    {26742, 2, {0x2000, 0x27F8, 0x2010, 0x21A0, 0xF840, 0x27FC, 0x2444, 0x7444, 0x6FFC, 0xA444, 0xA444, 0x27FC, 0x2444, 0x2444, 0x2454, 0x2408}},
    {26743, 2, {0x1080, 0x1080, 0x10F8, 0x1108, 0xFA10, 0x15FC, 0x3124, 0x3924, 0x55FC, 0x5124, 0x9124, 0x11FC, 0x1124, 0x1224, 0x1214, 0x1408}},
    {26744, 2, {0x1184, 0x1068, 0x1030, 0x10C8, 0xFB04, 0x1040, 0x33FE, 0x38A0, 0x5520, 0x53FC, 0x9524, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {26745, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x3104, 0x3904, 0x55FC, 0x5520, 0x9122, 0x1114, 0x1108, 0x1144, 0x1182, 0x1100}},
    {26746, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFC50, 0x13DE, 0x3252, 0x3A52, 0x5652, 0x5252, 0x93DE, 0x1050, 0x1090, 0x1090, 0x1110, 0x1210}},
    {26747, 2, {0x2100, 0x21F8, 0x2308, 0x2490, 0xF860, 0x2198, 0x2646, 0x7040, 0x6BF8, 0xA040, 0xA3F8, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2040}},
    {26748, 2, {0x100, 0x100, 0x7FFC, 0x920, 0x1110, 0x2108, 0xC284, 0x440, 0x920, 0x3118, 0xCB26, 0x5C0, 0x1930, 0x6108, 0x500, 0x200}},
    {26749, 2, {0x100, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x820, 0x920, 0x1550, 0x2388, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {26750, 2, {0x1000, 0x11FC, 0x1044, 0x1044, 0xFBFE, 0x1044, 0x3044, 0x39FC, 0x5480, 0x5080, 0x91FC, 0x1284, 0x1484, 0x1084, 0x10FC, 0x1084}},
    {26751, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3000, 0x3800, 0x55FC, 0x5420, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26752, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x3124, 0x3924, 0x55FC, 0x5420, 0x9070, 0x10A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {26753, 2, {0x0, 0x23F0, 0x1090, 0x8294, 0x5292, 0x2492, 0xE110, 0x2250, 0x2520, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26754, 2, {0x1028, 0x1024, 0x1024, 0x1020, 0xFDFE, 0x1020, 0x3120, 0x38B2, 0x54B4, 0x5468, 0x90A8, 0x1124, 0x1222, 0x1020, 0x10A0, 0x1040}},
    {26755, 2, {0x2000, 0x200C, 0x2770, 0x2110, 0xF910, 0x2210, 0x327C, 0x6F10, 0x6110, 0xA510, 0xA510, 0x227C, 0x2200, 0x2500, 0x28FE, 0x3000}},
    {26756, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFDFC, 0x1154, 0x3154, 0x3954, 0x558C, 0x5504, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26757, 2, {0x1100, 0x1100, 0x11FC, 0x1200, 0xFDF8, 0x1108, 0x3948, 0x3528, 0x57FE, 0x5108, 0x9248, 0x1228, 0x13FC, 0x1008, 0x1050, 0x1020}},
    {26758, 2, {0x2100, 0x211E, 0x2112, 0x2FD2, 0xF114, 0x2114, 0x27D8, 0x7114, 0x6912, 0xAFD2, 0xA112, 0x211A, 0x2214, 0x2210, 0x2410, 0x2810}},
    {26759, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x3020, 0x3BFE, 0x5488, 0x5088, 0x93FE, 0x1088, 0x1088, 0x1108, 0x1108, 0x1208}},
    {26760, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDF8, 0x1108, 0x3108, 0x39F8, 0x5400, 0x53FC, 0x9008, 0x1030, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26761, 2, {0x1048, 0x1148, 0x1148, 0x1148, 0xFD48, 0x11C8, 0x307E, 0x3848, 0x57C8, 0x5148, 0x9148, 0x1148, 0x1148, 0x1148, 0x125C, 0x1440}},
    {26762, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26763, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1000, 0x33FC, 0x3A04, 0x5604, 0x53FC, 0x9204, 0x1204, 0x13FC, 0x1204, 0x1214, 0x1208}},
    {26764, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x3020, 0x3820, 0x57FE, 0x5420, 0x9128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {26765, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x3104, 0x3904, 0x55FC, 0x5440, 0x9040, 0x107E, 0x17C0, 0x1042, 0x1042, 0x103E}},
    {26766, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3800, 0x5500, 0x550C, 0x91F0, 0x1100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {26767, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x3020, 0x3BFE, 0x5400, 0x5400, 0x91FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26768, 2, {0x1220, 0x1224, 0x1228, 0x13B0, 0xFE20, 0x1222, 0x32A2, 0x3B1E, 0x5440, 0x5040, 0x93FC, 0x1040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {26769, 2, {0x2010, 0x2890, 0x2510, 0x2212, 0xF552, 0x2954, 0x2158, 0x7190, 0x6B10, 0xA528, 0xA928, 0x2128, 0x2148, 0x2144, 0x2A84, 0x2502}},
    {26770, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD44, 0x1222, 0x30F8, 0x3808, 0x5410, 0x5420, 0x91FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26771, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFC88, 0x1050, 0x33FE, 0x3820, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26772, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFDFE, 0x1100, 0x317C, 0x3954, 0x5554, 0x5554, 0x9154, 0x117C, 0x1140, 0x1242, 0x1242, 0x143E}},
    {26773, 2, {0x1000, 0x13FC, 0x1044, 0x1144, 0xFD44, 0x1284, 0x3894, 0x3508, 0x5240, 0x5020, 0x90A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {26774, 2, {0x800, 0x8FC, 0x884, 0x884, 0xFEFC, 0x884, 0x1884, 0x1CFC, 0x2A84, 0x2A84, 0x48FC, 0x8800, 0x848, 0x844, 0x882, 0x902}},
    {26775, 2, {0x1000, 0x13FE, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5524, 0x5524, 0x91FC, 0x1120, 0x10A0, 0x1040, 0x10B0, 0x130E}},
    {26776, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1050, 0x1050, 0x1092, 0x1112, 0x120E}},
    {26777, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x3020, 0x3820, 0x57FE, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1020, 0x1020, 0x1020}},
    {26778, 2, {0x1080, 0x1080, 0x11F8, 0x1208, 0xFC10, 0x13FC, 0x3244, 0x3A44, 0x5644, 0x53FC, 0x90A0, 0x10A0, 0x1122, 0x1122, 0x121E, 0x1400}},
    {26779, 2, {0x2000, 0x2FDE, 0x2252, 0x2252, 0xFA54, 0x27D4, 0x3258, 0x6A54, 0x6252, 0xAFD2, 0xA252, 0x225A, 0x2254, 0x2450, 0x2550, 0x2890}},
    {26780, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3924, 0x3524, 0x5124, 0x52AA, 0x9020, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26781, 2, {0x840, 0x840, 0x10FC, 0x1088, 0x3550, 0x5420, 0x94D8, 0x1726, 0x1420, 0x15FC, 0x1420, 0x14A8, 0x1524, 0x1222, 0x10A0, 0x1040}},
    {26782, 2, {0x1000, 0x11F8, 0x1008, 0x1008, 0xFDF8, 0x1100, 0x3104, 0x3904, 0x54FC, 0x5440, 0x9020, 0x10A4, 0x128A, 0x128A, 0x1478, 0x1000}},
    {26783, 2, {0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x104, 0x3FF4, 0x548, 0x1930, 0xE10E, 0x100}},
    {26784, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFD08, 0x1108, 0x31F8, 0x3800, 0x5400, 0x53FC, 0x9204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {26785, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11F8, 0x3800, 0x3400, 0x53FE, 0x5090, 0x9090, 0x1090, 0x1112, 0x1112, 0x120E, 0x1400}},
    {26786, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0xFC20, 0x11FC, 0x3104, 0x3904, 0x55FC, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {26787, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1020, 0x3050, 0x3888, 0x5544, 0x5222, 0x9020, 0x11F8, 0x1008, 0x1010, 0x1010, 0x1020}},
    {26788, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1000, 0x3088, 0x3888, 0x5504, 0x52FA, 0x9048, 0x1048, 0x1088, 0x1088, 0x1128, 0x1210}},
    {26789, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9214, 0x2448, 0x820, 0x1FF0, 0x110, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26790, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x400, 0x7F0, 0x810, 0x1420, 0x2240, 0x180, 0x200, 0xC00, 0x7000}},
    {26791, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0xFDFC, 0x1084, 0x3084, 0x3BFE, 0x5400, 0x5000, 0x91FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26792, 2, {0x308, 0x3C08, 0x448, 0x7F48, 0xC48, 0x1648, 0x2508, 0x4428, 0x510, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26793, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {26794, 2, {0x1000, 0x13FC, 0x1000, 0x1000, 0xFDF8, 0x1108, 0x3108, 0x3908, 0x55F8, 0x5400, 0x9108, 0x1088, 0x1090, 0x1000, 0x17FE, 0x1000}},
    {26795, 2, {0x1000, 0x11FC, 0x1004, 0x10FC, 0xFC04, 0x11FC, 0x3800, 0x37FE, 0x5202, 0x51F8, 0x9088, 0x1088, 0x1050, 0x1020, 0x10D8, 0x1306}},
    {26796, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x31FC, 0x3800, 0x57FE, 0x5480, 0x9100, 0x11FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {26797, 2, {0x1040, 0x1040, 0x1088, 0x1104, 0xFBFE, 0x1002, 0x3088, 0x3944, 0x5642, 0x50F8, 0x9188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {26798, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1100, 0x3100, 0x39FE, 0x5502, 0x557A, 0x914A, 0x114A, 0x127A, 0x1202, 0x1414, 0x1008}},
    {26799, 2, {0x1108, 0x1088, 0x1090, 0x11FC, 0xFC24, 0x1024, 0x31FC, 0x3920, 0x5520, 0x51FE, 0x9062, 0x10A2, 0x112A, 0x1224, 0x1420, 0x1020}},
    {26800, 2, {0x1010, 0x1014, 0x1012, 0x1010, 0xFBFE, 0x1010, 0x3150, 0x3954, 0x57F4, 0x5154, 0x9154, 0x1148, 0x124A, 0x125A, 0x1426, 0x1042}},
    {26801, 2, {0x1000, 0x13FE, 0x1222, 0x1222, 0xFE22, 0x13FE, 0x3222, 0x3A62, 0x5672, 0x52AA, 0x9322, 0x1222, 0x1222, 0x1202, 0x13FE, 0x1202}},
    {26802, 2, {0x1010, 0x1090, 0x1088, 0x1108, 0xFA04, 0x15FA, 0x3108, 0x3908, 0x5508, 0x51F8, 0x9090, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {26803, 2, {0x1040, 0x1020, 0x13FC, 0x1040, 0xFC90, 0x1108, 0x33FC, 0x3804, 0x5550, 0x5150, 0x9150, 0x1150, 0x1250, 0x1252, 0x1452, 0x180E}},
    {26804, 2, {0x1004, 0x100E, 0x13B8, 0x1088, 0xFC88, 0x1128, 0x312E, 0x3BA8, 0x54A8, 0x52A8, 0x92A8, 0x113E, 0x1100, 0x1280, 0x147E, 0x1800}},
    {26805, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x0, 0xFE0, 0x820, 0x820, 0xA20, 0x920, 0x1122, 0x2022, 0x401E}},
    {26806, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFBFC, 0x1200, 0x3218, 0x3AE0, 0x5638, 0x52E0, 0x923C, 0x13E0, 0x1222, 0x1422, 0x141E, 0x1800}},
    {26807, 2, {0x1000, 0x10FC, 0x1E04, 0x1048, 0x7E28, 0x4210, 0x4228, 0x7EC4, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26808, 2, {0x2084, 0x21C4, 0x2704, 0x2114, 0xF914, 0x2114, 0x27D4, 0x7114, 0x6B14, 0xA394, 0xA554, 0x2504, 0x2904, 0x2104, 0x2114, 0x2108}},
    {26809, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0xFD00, 0x11FE, 0x3110, 0x3910, 0x5510, 0x5110, 0x97FE, 0x1000, 0x1090, 0x1108, 0x1204, 0x1402}},
    {26810, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x0, 0xFFFE, 0x200, 0x240, 0x220, 0x210, 0x208, 0x200, 0x200}},
    {26811, 2, {0x2250, 0x2250, 0x2250, 0x25FC, 0xF454, 0x2C54, 0x35FC, 0x7550, 0x6D50, 0xA5FE, 0xA452, 0x2452, 0x245A, 0x2494, 0x2490, 0x2510}},
    {26812, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x33FE, 0x3848, 0x5488, 0x54FE, 0x9108, 0x1148, 0x1228, 0x1408, 0x1028, 0x1010}},
    {26813, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3020, 0x39FC, 0x5440, 0x5420, 0x90A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {26814, 2, {0x1020, 0x1020, 0x1020, 0x13FE, 0xFC20, 0x1124, 0x30A4, 0x38A8, 0x55FE, 0x5470, 0x90A8, 0x10A8, 0x1124, 0x1222, 0x1420, 0x1020}},
    {26815, 2, {0x1010, 0x1210, 0x1110, 0x117E, 0xFC20, 0x1028, 0x3748, 0x397E, 0x5508, 0x5508, 0x91FE, 0x1108, 0x1108, 0x1288, 0x147E, 0x1000}},
    {26816, 2, {0x1040, 0x1040, 0x10A0, 0x10A0, 0xFD10, 0x1208, 0x35F6, 0x3800, 0x5488, 0x5048, 0x9248, 0x1150, 0x1110, 0x1020, 0x17FE, 0x1000}},
    {26817, 2, {0x1104, 0x1084, 0x1088, 0x1010, 0xFDFC, 0x1104, 0x3104, 0x3904, 0x55FC, 0x5450, 0x9050, 0x1090, 0x1092, 0x1112, 0x120E, 0x1400}},
    {26818, 2, {0x1000, 0x13F8, 0x1008, 0x1008, 0xFDF8, 0x1008, 0x3008, 0x3BF8, 0x5440, 0x5148, 0x9148, 0x1250, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {26819, 2, {0xC80, 0x70FC, 0x1124, 0xFD24, 0x1224, 0x3844, 0x5484, 0x9128, 0x1010, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26820, 2, {0x200, 0x100, 0x7FFC, 0x1010, 0x3FF8, 0x108, 0x1110, 0xFFFE, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {26821, 2, {0x2078, 0x23C0, 0x2040, 0x27FC, 0xF840, 0x23F8, 0x2048, 0x77FE, 0x6848, 0xA3F8, 0xA0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x2040}},
    {26822, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x3000, 0x39FC, 0x5554, 0x5554, 0x91FC, 0x1154, 0x1154, 0x1154, 0x1104, 0x110C}},
    {26823, 2, {0x1090, 0x1090, 0x1108, 0x1148, 0xFA44, 0x1492, 0x3108, 0x3BFC, 0x5444, 0x5020, 0x90A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {26824, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3800, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {26825, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3820, 0x5420, 0x53FE, 0x9222, 0x1222, 0x122A, 0x1224, 0x1020, 0x1020}},
    {26826, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1110, 0x2108, 0xDFF6, 0x300, 0x5C0, 0x1930, 0x6108, 0x100}},
    {26827, 2, {0x1108, 0x1108, 0x13FC, 0x1108, 0xFD08, 0x11F8, 0x3108, 0x3908, 0x55F8, 0x5108, 0x9108, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {26828, 2, {0x2008, 0x203C, 0x23C0, 0x2044, 0xFA24, 0x2128, 0x2100, 0x7020, 0x6BFE, 0xA070, 0xA0A8, 0x20A8, 0x2124, 0x2222, 0x2420, 0x2020}},
    {26829, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5400, 0x5512, 0x91D4, 0x1118, 0x1110, 0x1152, 0x1192, 0x110E}},
    {26830, 2, {0x1000, 0x13FE, 0x1202, 0x1252, 0xFC88, 0x1104, 0x3020, 0x3820, 0x57FE, 0x5420, 0x9070, 0x10A8, 0x1124, 0x1622, 0x1020, 0x1020}},
    {26831, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3000, 0x39FE, 0x5408, 0x53FE, 0x9008, 0x1108, 0x1088, 0x1008, 0x1028, 0x1010}},
    {26832, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x540, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26833, 2, {0x1090, 0x1090, 0x1090, 0x179E, 0xF890, 0x1090, 0x3090, 0x3B9C, 0x5490, 0x5090, 0x9090, 0x179E, 0x1090, 0x1090, 0x1090, 0x1090}},
    {26834, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x33FE, 0x3888, 0x5524, 0x5222, 0x90F8, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {26835, 2, {0x1080, 0x1040, 0x11FC, 0x1000, 0xFD08, 0x1088, 0x3090, 0x3BFE, 0x5400, 0x5400, 0x91FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26836, 2, {0x101C, 0x13E0, 0x1220, 0x13FE, 0xFE20, 0x1292, 0x330A, 0x3A06, 0x55FC, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26837, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC00, 0x11FC, 0x3000, 0x3800, 0x57FE, 0x5420, 0x9128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {26838, 2, {0x1000, 0x10FC, 0x1080, 0x10F8, 0xFC80, 0x10F8, 0x3080, 0x3BFE, 0x5540, 0x5524, 0x9128, 0x1110, 0x1108, 0x1144, 0x1182, 0x1100}},
    {26839, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x2388, 0xD60, 0x3010, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x2388, 0x560, 0x1918, 0xE106}},
    {26840, 2, {0x1020, 0x1020, 0xFDFC, 0x1020, 0x1020, 0x7CF8, 0x54A8, 0x54A8, 0x54A8, 0x5CB8, 0x1020, 0x3870, 0x54A8, 0x9526, 0x1020, 0x1020}},
    {26841, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0xFE04, 0x13FC, 0x3200, 0x3A28, 0x5624, 0x53FE, 0x9220, 0x1250, 0x1250, 0x1488, 0x1504, 0x1A02}},
    {26842, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFE52, 0x13DE, 0x3252, 0x3A52, 0x5652, 0x53DE, 0x9252, 0x1252, 0x1252, 0x1252, 0x1552, 0x18A6}},
    {26843, 2, {0x1040, 0x1020, 0x13FE, 0x1080, 0xFD04, 0x13FE, 0x3002, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {26844, 2, {0x2108, 0x2088, 0x2008, 0x27D4, 0xFA14, 0x2222, 0x2200, 0x73C8, 0x6A44, 0xA244, 0xA240, 0x2250, 0x2248, 0x2444, 0x2544, 0x2880}},
    {26845, 2, {0x1000, 0x11FC, 0x1104, 0x1124, 0xFD24, 0x11FC, 0x3124, 0x3924, 0x5574, 0x5554, 0x9154, 0x1174, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26846, 2, {0x1000, 0x13FE, 0x1202, 0x121A, 0xFEE2, 0x1222, 0x3222, 0x3BFE, 0x5622, 0x5272, 0x92AA, 0x1326, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {26847, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x31FC, 0x3924, 0x55FC, 0x5420, 0x9070, 0x10A8, 0x1124, 0x1622, 0x1020, 0x1020}},
    {26848, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26849, 2, {0x1000, 0x13FE, 0x1202, 0x128A, 0xFE52, 0x13FE, 0x3222, 0x3A22, 0x56AA, 0x52AA, 0x92AA, 0x12FA, 0x1202, 0x1202, 0x120A, 0x1204}},
    {26850, 2, {0x1000, 0x13FE, 0x1202, 0x128A, 0xFE52, 0x13FE, 0x3242, 0x3A22, 0x57FE, 0x5282, 0x9282, 0x1282, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {26851, 2, {0x1020, 0x1020, 0x11FC, 0x1024, 0xFC24, 0x13FE, 0x3824, 0x3424, 0x51FC, 0x5222, 0x9174, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {26852, 2, {0x1088, 0x1088, 0x1088, 0x13FE, 0xFC88, 0x1088, 0x37FE, 0x3800, 0x55FC, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26853, 2, {0x2088, 0x2488, 0x2288, 0x2108, 0xF93E, 0x2288, 0x2498, 0x701C, 0x68AC, 0xA2AA, 0xA148, 0x2108, 0x2288, 0x2288, 0x2488, 0x2808}},
    {26854, 2, {0x1080, 0x1080, 0x11F0, 0x1210, 0xFC20, 0x13FC, 0x3044, 0x3844, 0x57FE, 0x5444, 0x9044, 0x13FC, 0x1044, 0x1040, 0x1140, 0x1080}},
    {26855, 2, {0x1048, 0x1044, 0x105E, 0x11E0, 0xFC28, 0x1012, 0x306A, 0x3996, 0x5448, 0x545E, 0x91E0, 0x1024, 0x1028, 0x1012, 0x106A, 0x1186}},
    {26856, 2, {0x1020, 0x820, 0x3E3E, 0x2244, 0x2284, 0x3E28, 0x2210, 0x4028, 0x8144, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26857, 2, {0x1050, 0x1252, 0x1252, 0x1252, 0xFBDE, 0x1202, 0x3202, 0x3BFE, 0x5602, 0x5202, 0x93DE, 0x1252, 0x1252, 0x1252, 0x1452, 0x1802}},
    {26858, 2, {0x1020, 0x1122, 0x1122, 0x1224, 0xFC50, 0x1088, 0x3304, 0x3822, 0x5420, 0x5524, 0x9124, 0x1228, 0x1050, 0x1088, 0x1104, 0x1602}},
    {26859, 2, {0x1014, 0x1012, 0x1010, 0x17FE, 0xF810, 0x1010, 0x33D2, 0x3A52, 0x5652, 0x5254, 0x93D4, 0x1008, 0x10EA, 0x171A, 0x1226, 0x1042}},
    {26860, 2, {0x1020, 0x1124, 0x10A8, 0x1020, 0xFDFC, 0x1040, 0x33FE, 0x3888, 0x5504, 0x52FA, 0x9488, 0x1088, 0x10A8, 0x1092, 0x1082, 0x107E}},
    {26861, 2, {0x2080, 0x2040, 0x27FE, 0x2120, 0xF920, 0x223C, 0x2244, 0x7664, 0x6A94, 0xA348, 0xA248, 0x2230, 0x2220, 0x2250, 0x2288, 0x2306}},
    {26862, 2, {0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0x820}},
    {26863, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD44, 0x1222, 0x31F8, 0x3808, 0x5450, 0x5420, 0x90A4, 0x1282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {26864, 2, {0x2010, 0x2078, 0x23C0, 0x2040, 0xF840, 0x27FC, 0x2248, 0x7248, 0x6FFE, 0xA248, 0xA248, 0x27FC, 0x2040, 0x2040, 0x27FC, 0x2000}},
    {26865, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x3088, 0x3944, 0x5642, 0x54F8, 0x9188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {26866, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1024, 0x33FE, 0x3824, 0x55FC, 0x5040, 0x93FE, 0x1084, 0x11C8, 0x1030, 0x10CC, 0x1302}},
    {26867, 2, {0x1000, 0x13DE, 0x1042, 0x114A, 0xFC84, 0x114A, 0x3A52, 0x3420, 0x5000, 0x53DE, 0x9052, 0x1152, 0x1094, 0x1148, 0x1254, 0x1422}},
    {26868, 2, {0x2000, 0x27BE, 0x24A2, 0x24A2, 0xFCAA, 0x27A4, 0x24A0, 0x74BE, 0x6CAA, 0xA7AA, 0xA4AA, 0x24AA, 0x24A4, 0x24A4, 0x24AA, 0x29B0}},
    {26869, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5420, 0x53FE, 0x9070, 0x10A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {26870, 2, {0x1040, 0x1040, 0x1040, 0x17FC, 0xF840, 0x1248, 0x3248, 0x3A48, 0x5554, 0x50E0, 0x9150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040}},
    {26871, 2, {0x2000, 0x2FC0, 0x24BC, 0x24A4, 0xF4A4, 0x27A4, 0x24A4, 0x74A8, 0x6FA8, 0xA4A8, 0xA490, 0x25D0, 0x2EA8, 0x20A8, 0x20C4, 0x2082}},
    {26872, 2, {0x0, 0x7F00, 0x227C, 0x3E04, 0x2228, 0x3E10, 0x23A8, 0xFE44, 0x200, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26873, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5524, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26874, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11F8, 0x3108, 0x3908, 0x55F8, 0x5500, 0x9100, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26875, 2, {0x820, 0xFFFE, 0x820, 0x460, 0x810, 0x37E8, 0xC224, 0x4A0, 0x1840, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26876, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC824, 0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040}},
    {26877, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC924, 0x280, 0xE60, 0x3118, 0xC106, 0x1FE0, 0x20, 0x40, 0x80, 0x100}},
    {26878, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x0, 0x7FFE, 0x4002, 0x8FE4, 0x820, 0x820, 0x1022, 0x2022, 0xC01E}},
    {26879, 2, {0x1000, 0x1040, 0x139C, 0x1204, 0xFE04, 0x139C, 0x3204, 0x3A04, 0x57FC, 0x5490, 0x9090, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {26880, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFD04, 0x1100, 0x31DE, 0x3A52, 0x5652, 0x5352, 0x949A, 0x1094, 0x1110, 0x1112, 0x1212, 0x140E}},
    {26881, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x31FC, 0x3800, 0x55FC, 0x5408, 0x9010, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26882, 2, {0x1000, 0x11F8, 0x1008, 0x1008, 0xFDF8, 0x1008, 0x3008, 0x3BFE, 0x5420, 0x5222, 0x9174, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {26883, 2, {0x1040, 0x1040, 0x107C, 0x1040, 0xFBFC, 0x1244, 0x3270, 0x3BC0, 0x5644, 0x523C, 0x9200, 0x12F0, 0x1490, 0x1492, 0x1912, 0x120E}},
    {26884, 2, {0x1080, 0x1040, 0x13FC, 0x1000, 0xFD08, 0x1090, 0x37FE, 0x3840, 0x5440, 0x57FE, 0x9088, 0x1108, 0x1090, 0x1060, 0x1198, 0x1604}},
    {26885, 2, {0x2040, 0x2040, 0x23FC, 0x20A0, 0xF910, 0x2208, 0x27FE, 0x7008, 0x6BC8, 0xA248, 0xA248, 0x2248, 0x23C8, 0x2008, 0x2028, 0x2010}},
    {26886, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFD74, 0x1124, 0x31FC, 0x3904, 0x5574, 0x5154, 0x9154, 0x1174, 0x1104, 0x1204, 0x1214, 0x1408}},
    {26887, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1108, 0x31F8, 0x3908, 0x55F8, 0x5108, 0x9108, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {26888, 2, {0x1100, 0x1100, 0x11FC, 0x1204, 0xFC84, 0x1494, 0x32A4, 0x3884, 0x57FC, 0x5084, 0x91C4, 0x12A4, 0x1494, 0x1084, 0x10A8, 0x1010}},
    {26889, 2, {0x200, 0x100, 0xFFFE, 0x1010, 0x3EFC, 0x4210, 0x1490, 0x18FE, 0xE010, 0x110, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26890, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFC88, 0x1088, 0x3154, 0x3A22, 0x5400, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {26891, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFC00, 0x11FC, 0x3104, 0x3904, 0x55FC, 0x5420, 0x9128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {26892, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0xFA02, 0x1494, 0x3108, 0x3A04, 0x5400, 0x51FC, 0x9020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {26893, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x3104, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1104, 0x17FE, 0x1000}},
    {26894, 2, {0x10A0, 0x1090, 0x1090, 0x11FE, 0xFD10, 0x1310, 0x35FE, 0x3910, 0x5510, 0x55FE, 0x9110, 0x1110, 0x1110, 0x11FE, 0x1100, 0x1100}},
    {26895, 2, {0x1000, 0x17FE, 0x1090, 0x1090, 0xFC90, 0x139C, 0x3204, 0x3A04, 0x5604, 0x539C, 0x9090, 0x1090, 0x1090, 0x1090, 0x17FE, 0x1000}},
    {26896, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFBFC, 0x1220, 0x3220, 0x3BFE, 0x5620, 0x5220, 0x92FC, 0x1284, 0x1284, 0x1484, 0x14FC, 0x1884}},
    {26897, 2, {0x1040, 0x1080, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x3124, 0x3944, 0x55FC, 0x5490, 0x9110, 0x13FE, 0x1010, 0x1010, 0x1010, 0x1010}},
    {26898, 2, {0x2100, 0x2100, 0x21DC, 0x2114, 0xF914, 0x2114, 0x27D4, 0x7014, 0x6914, 0xA114, 0xA588, 0x2548, 0x2948, 0x2114, 0x2514, 0x2222}},
    {26899, 2, {0x1000, 0x13FE, 0x1040, 0x1080, 0xFD44, 0x1224, 0x3068, 0x3AB0, 0x5530, 0x52A8, 0x9068, 0x10A4, 0x1122, 0x1220, 0x10A0, 0x1040}},
    {26900, 2, {0x1092, 0x1092, 0x1124, 0x1248, 0xFD24, 0x1092, 0x3092, 0x3800, 0x55FE, 0x5522, 0x9122, 0x11FE, 0x1122, 0x1122, 0x11FE, 0x1102}},
    {26901, 2, {0x2008, 0x2088, 0x2288, 0x2288, 0xFABE, 0x2448, 0x2018, 0x779C, 0x6AAC, 0xA2AA, 0xA2C8, 0x2288, 0x2288, 0x2488, 0x2488, 0x2988}},
    {26902, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x3120, 0x3910, 0x55FE, 0x5520, 0x9120, 0x113C, 0x1124, 0x1244, 0x1254, 0x1488}},
    {26903, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x1000, 0x33FE, 0x3820, 0x5420, 0x5520, 0x913C, 0x1120, 0x12A0, 0x1260, 0x143E, 0x1800}},
    {26904, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC824, 0x200, 0x100, 0x7FF8, 0x60, 0x180, 0xE00, 0x3000, 0x4800, 0x87FE}},
    {26905, 2, {0x11F8, 0x1108, 0x1108, 0x11F8, 0xFD08, 0x1108, 0x31F8, 0x3800, 0x57FC, 0x5204, 0x9204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {26906, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFBDE, 0x1252, 0x3252, 0x3BDE, 0x5602, 0x5202, 0x9202, 0x1202, 0x1202, 0x1202, 0x120A, 0x1204}},
    {26907, 2, {0x1108, 0x1108, 0x17FE, 0x1108, 0xFD08, 0x1000, 0x3890, 0x3492, 0x5114, 0x5318, 0x9510, 0x1132, 0x1152, 0x1112, 0x110E, 0x1100}},
    {26908, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x3020, 0x3820, 0x55FC, 0x5524, 0x9124, 0x11FC, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26909, 2, {0x1000, 0x11FC, 0x7D04, 0x1124, 0x1124, 0xFC50, 0x1090, 0x2912, 0x460E, 0x8100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26910, 2, {0x80C, 0x8F0, 0x7E80, 0x1880, 0x2CFE, 0x2A88, 0x4888, 0x8888, 0x908, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26911, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x3002, 0x3894, 0x5450, 0x5510, 0x9090, 0x13FE, 0x1028, 0x1044, 0x1082, 0x1302}},
    {26912, 2, {0x100C, 0xFEF0, 0x2080, 0x4880, 0x7EFE, 0x888, 0xE88, 0xF908, 0x4A08, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26913, 2, {0x1002, 0x1002, 0x17E2, 0x110A, 0xFD0A, 0x124A, 0x33EA, 0x388A, 0x548A, 0x53EA, 0x908A, 0x108A, 0x10E2, 0x1702, 0x120A, 0x1004}},
    {26914, 2, {0x1000, 0x13FE, 0x1202, 0x1202, 0xFBFE, 0x1222, 0x3222, 0x3AFA, 0x5622, 0x5232, 0x922A, 0x13FE, 0x1202, 0x1202, 0x13FE, 0x1202}},
    {26915, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0xFE94, 0x1294, 0x3294, 0x3BFC, 0x5694, 0x5294, 0x9294, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {26916, 2, {0x1000, 0x13FE, 0x1252, 0x1252, 0xFE52, 0x13FE, 0x3020, 0x3840, 0x54FC, 0x5104, 0x9288, 0x1050, 0x1020, 0x1040, 0x1180, 0x1600}},
    {26917, 2, {0x1100, 0x1100, 0x1100, 0x13DE, 0xFA92, 0x1492, 0x3092, 0x3892, 0x57F2, 0x5092, 0x9092, 0x1152, 0x115E, 0x1220, 0x1420, 0x1800}},
    {26918, 2, {0x1020, 0x1124, 0x10A8, 0x1020, 0xFDFC, 0x1040, 0x33FE, 0x3888, 0x5544, 0x5242, 0x95F8, 0x1048, 0x1088, 0x1088, 0x1128, 0x1210}},
    {26919, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x3000, 0x3BDE, 0x5652, 0x5252, 0x9252, 0x13DA, 0x1254, 0x1010, 0x1010, 0x1010}},
    {26920, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0xFE44, 0x1244, 0x3284, 0x3ABE, 0x5784, 0x52A4, 0x9294, 0x1294, 0x1284, 0x1284, 0x1494, 0x1088}},
    {26921, 2, {0x1040, 0x1020, 0x13FE, 0x1220, 0xFEFC, 0x1224, 0x3224, 0x3BFE, 0x5624, 0x5224, 0x92FC, 0x1220, 0x1250, 0x1450, 0x1488, 0x1906}},
    {26922, 2, {0x1084, 0x1048, 0x1000, 0x11FE, 0xFC48, 0x1048, 0x3048, 0x3A48, 0x554A, 0x554C, 0x9048, 0x1048, 0x1048, 0x1048, 0x13FE, 0x1000}},
    {26923, 2, {0x1104, 0x1088, 0x1050, 0x11FC, 0xFD24, 0x1124, 0x31FC, 0x3924, 0x5524, 0x55FC, 0x9020, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {26924, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11F8, 0x3108, 0x3908, 0x55F8, 0x5508, 0x9108, 0x11F8, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {26925, 2, {0x2010, 0x2790, 0x2490, 0x257E, 0xFD20, 0x2620, 0x257C, 0x74A4, 0x6CA4, 0xA4BC, 0xA6A4, 0x253C, 0x2424, 0x2424, 0x2424, 0x242C}},
    {26926, 2, {0x1040, 0x1048, 0x1084, 0x11FE, 0xFC20, 0x13FE, 0x3088, 0x3924, 0x5642, 0x5188, 0x9010, 0x1062, 0x1184, 0x1018, 0x1060, 0x1380}},
    {26927, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1000, 0x33FE, 0x3820, 0x5440, 0x55FC, 0x9154, 0x1154, 0x1154, 0x1154, 0x1154, 0x110C}},
    {26928, 2, {0x2000, 0x2FDE, 0x2492, 0x2492, 0xF494, 0x2794, 0x2498, 0x7494, 0x6F92, 0xA492, 0xA492, 0x25DA, 0x2E94, 0x2090, 0x2090, 0x2090}},
    {26929, 2, {0x1100, 0x1100, 0x11FE, 0x1200, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5480, 0x90FC, 0x1108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {26930, 2, {0x1040, 0x1040, 0x11FC, 0x1084, 0xFBFE, 0x1000, 0x31FC, 0x3904, 0x55FC, 0x5020, 0x93FE, 0x1020, 0x1220, 0x13FE, 0x1020, 0x1020}},
    {26931, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5400, 0x53FE, 0x9120, 0x1122, 0x1114, 0x1148, 0x1184, 0x1102}},
    {26932, 2, {0x1020, 0x11DC, 0x1114, 0x1114, 0xFD14, 0x11D4, 0x3126, 0x3900, 0x55DC, 0x5514, 0x9114, 0x11D4, 0x1708, 0x1108, 0x1114, 0x1122}},
    {26933, 2, {0x1000, 0x13DE, 0x1242, 0x1242, 0xFE42, 0x13DE, 0x3200, 0x3A3E, 0x57D2, 0x5212, 0x9214, 0x13D4, 0x1208, 0x1214, 0x1224, 0x1242}},
    {26934, 2, {0x1000, 0x1088, 0x1252, 0x1222, 0xFE52, 0x128A, 0x3202, 0x3BFE, 0x5488, 0x5144, 0x927A, 0x1088, 0x1150, 0x1020, 0x10D8, 0x1706}},
    {26935, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0xFC10, 0x2410, 0x25D0, 0x7412, 0x6C12, 0xA5D4, 0xA554, 0x2548, 0x25DA, 0x242A, 0x2846, 0x3082}},
    {26936, 2, {0x2210, 0x2110, 0x2110, 0x27BE, 0xFA28, 0x2248, 0x23AA, 0x72AE, 0x6ABA, 0xA2EA, 0xA2AE, 0x22A8, 0x24AA, 0x24A2, 0x29A2, 0x301E}},
    {26937, 2, {0x1110, 0x1110, 0x17FC, 0x1110, 0xFDF0, 0x1110, 0x39F0, 0x3510, 0x5110, 0x57FE, 0x9200, 0x1290, 0x1308, 0x1200, 0x13FC, 0x1000}},
    {26938, 2, {0x2100, 0x217C, 0x2144, 0x2244, 0xFA44, 0x267C, 0x2A10, 0x7210, 0x6AFE, 0xA238, 0xA254, 0x2254, 0x2294, 0x2312, 0x2210, 0x2210}},
    {26939, 2, {0x1000, 0x13FE, 0x1200, 0x12FC, 0xFE84, 0x12FC, 0x3284, 0x3AFC, 0x5620, 0x53FE, 0x9248, 0x12C8, 0x1230, 0x1248, 0x1284, 0x13FE}},
    {26940, 2, {0x1080, 0x108E, 0x1120, 0x1210, 0xFCC0, 0x10AE, 0x3104, 0x3B14, 0x5524, 0x5164, 0x9124, 0x1124, 0x1124, 0x1104, 0x1114, 0x1108}},
    {26941, 2, {0x2100, 0x21F8, 0x2208, 0x23F0, 0xF810, 0x27FE, 0x2080, 0x7144, 0x6E68, 0xA0B0, 0xA128, 0x2668, 0x20A4, 0x2122, 0x26A0, 0x2040}},
    {26942, 2, {0x1108, 0x1088, 0x1090, 0x17FE, 0xF800, 0x13C4, 0x3254, 0x3A54, 0x57D4, 0x5254, 0x9254, 0x13D4, 0x1254, 0x1244, 0x1254, 0x12C8}},
    {26943, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x33FE, 0x3888, 0x5504, 0x52FA, 0x9088, 0x1088, 0x10F8, 0x1088, 0x1088, 0x10F8}},
    {26944, 2, {0x1004, 0x101E, 0x13E0, 0x1020, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5420, 0x53FE, 0x9222, 0x122A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {26945, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC80, 0x10F8, 0x3108, 0x3A90, 0x5460, 0x5198, 0x9606, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {26946, 2, {0x1020, 0x1020, 0x13FE, 0x1070, 0xFCA8, 0x1124, 0x3222, 0x39FC, 0x5504, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x1000, 0x13FE, 0x1000}},
    {26947, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFBFC, 0x1200, 0x33FC, 0x3A40, 0x5688, 0x53FC, 0x9224, 0x1220, 0x15FC, 0x1420, 0x1820, 0x13FE}},
    {26948, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFE02, 0x13FE, 0x3200, 0x3A00, 0x57FE, 0x5352, 0x9552, 0x15FE, 0x1552, 0x1552, 0x1952, 0x1106}},
    {26949, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x1104, 0x31FC, 0x3800, 0x57FE, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {26950, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFBFE, 0x1252, 0x3252, 0x3BFE, 0x5400, 0x5040, 0x97FE, 0x1088, 0x1190, 0x1060, 0x10D8, 0x1304}},
    {26951, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0xFDE4, 0x1124, 0x3124, 0x3BFE, 0x5602, 0x52FA, 0x928A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {26952, 2, {0x1000, 0x13FE, 0x1022, 0x1120, 0xFD3C, 0x1120, 0x3AFE, 0x3400, 0x51FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {26953, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCA8, 0x1020, 0x33FE, 0x3840, 0x5480, 0x54FC, 0x9184, 0x1284, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {26954, 2, {0x1000, 0x11F8, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x3000, 0x3BFE, 0x5500, 0x51FC, 0x9254, 0x1494, 0x1124, 0x1244, 0x10A8, 0x1110}},
    {26955, 2, {0x1082, 0x1082, 0x17F2, 0x1082, 0xFC8A, 0x13EA, 0x32AA, 0x3AAA, 0x57EA, 0x508A, 0x91CA, 0x12AA, 0x14A2, 0x1082, 0x108A, 0x1084}},
    {26956, 2, {0x1040, 0x1020, 0x13FC, 0x1000, 0xFD08, 0x1090, 0x33FE, 0x3A10, 0x5620, 0x52C8, 0x9210, 0x1224, 0x12C8, 0x1410, 0x1460, 0x1980}},
    {26957, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x0, 0x7BBC, 0x4AA4, 0x4AA4, 0x4AA4, 0x4AA4, 0x7BBC, 0x4AA4}},
    {26958, 2, {0x1000, 0x13FE, 0x1222, 0x1020, 0xFDFE, 0x1020, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {26959, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x1326, 0x31FC, 0x3820, 0x5420, 0x53FE, 0x9000, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE}},
    {26960, 2, {0x11FC, 0x1124, 0x1124, 0x11FC, 0xFD24, 0x1124, 0x31FC, 0x3850, 0x5488, 0x5104, 0x928A, 0x1088, 0x1088, 0x1088, 0x1108, 0x1208}},
    {26961, 2, {0x1010, 0x13D4, 0x1058, 0x1252, 0xFD8C, 0x1088, 0x3104, 0x3AFA, 0x5420, 0x5020, 0x93FE, 0x1020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26962, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5400, 0x5440, 0x9024, 0x12A2, 0x128A, 0x1488, 0x1078, 0x1000}},
    {26963, 2, {0x2000, 0x27FC, 0x2404, 0x2434, 0xFDC4, 0x2444, 0x25F4, 0x7554, 0x6D54, 0xA5F4, 0xA444, 0x2454, 0x25F4, 0x2494, 0x2806, 0x3002}},
    {26964, 2, {0x2100, 0x213C, 0x27D4, 0x2114, 0xFB94, 0x2114, 0x27D4, 0x7124, 0x694C, 0xA040, 0xA7FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {26965, 2, {0x1040, 0x1040, 0x13FE, 0x1080, 0xFDFC, 0x1220, 0x35FE, 0x3800, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {26966, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFE52, 0x13D2, 0x3252, 0x3A52, 0x5652, 0x53D2, 0x921A, 0x1294, 0x1250, 0x12B0, 0x1310, 0x1010}},
    {26967, 2, {0x2020, 0x2020, 0x20FC, 0x2E24, 0xFBFE, 0x2224, 0x24FC, 0x7420, 0x6EFC, 0xA220, 0xA220, 0x2BFE, 0x2420, 0x2620, 0x29FE, 0x3000}},
    {26968, 2, {0x20, 0x7E20, 0x247E, 0x1884, 0xFF28, 0x2910, 0x4A28, 0x9944, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26969, 2, {0x2008, 0x2788, 0x2088, 0x2288, 0xF93E, 0x2108, 0x27D8, 0x715C, 0x6B2C, 0xA32A, 0xA548, 0x2508, 0x2908, 0x2108, 0x2508, 0x2208}},
    {26970, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC824, 0x0, 0x7FFC, 0x104, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {26971, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCA8, 0x1020, 0x33FE, 0x3820, 0x5420, 0x5420, 0x91FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {26972, 2, {0x2100, 0x211E, 0x2112, 0x27D2, 0xF912, 0x211E, 0x2112, 0x77D2, 0x6C52, 0xA45E, 0xA452, 0x27D2, 0x2452, 0x2022, 0x202A, 0x2044}},
    {26973, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x31AC, 0x3974, 0x5524, 0x55FC, 0x9020, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {26974, 2, {0x1000, 0x13FC, 0x1294, 0x1294, 0xFE94, 0x13FC, 0x3080, 0x3840, 0x57FE, 0x5480, 0x90F8, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {26975, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x39FC, 0x3400, 0x53FE, 0x5202, 0x91FC, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {26976, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFC20, 0x13FE, 0x328A, 0x3A52, 0x56FA, 0x5222, 0x9222, 0x12FA, 0x1222, 0x1222, 0x122A, 0x1204}},
    {26977, 2, {0x11C0, 0x10A0, 0x1110, 0x1208, 0xFDF6, 0x1000, 0x33CA, 0x3A4A, 0x5654, 0x53D4, 0x9268, 0x1254, 0x13D4, 0x124A, 0x124A, 0x12C0}},
    {26978, 2, {0x1108, 0x1088, 0x1090, 0x17FE, 0xF890, 0x1090, 0x33FC, 0x3A94, 0x5694, 0x530C, 0x9204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {26979, 2, {0x1000, 0x13FC, 0x1224, 0x1224, 0xFBFC, 0x1200, 0x32FC, 0x3A84, 0x5684, 0x52FC, 0x9284, 0x12FC, 0x1284, 0x1284, 0x14FC, 0x1884}},
    {26980, 2, {0x1080, 0x1080, 0x10FC, 0x1154, 0xFAD4, 0x10B4, 0x312C, 0x3A44, 0x5494, 0x5108, 0x9040, 0x10A4, 0x12AA, 0x128A, 0x1478, 0x1000}},
    {26981, 2, {0x103C, 0x17C0, 0x1244, 0x1128, 0xFC00, 0x13FC, 0x3080, 0x3880, 0x57FE, 0x5500, 0x91F8, 0x1288, 0x1250, 0x1420, 0x18D8, 0x1306}},
    {26982, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC04, 0x11FC, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1000, 0x13FE, 0x1000}},
    {26983, 2, {0x2110, 0x2110, 0x27FC, 0x2110, 0xF950, 0x2040, 0x23F8, 0x7248, 0x6A48, 0xA248, 0xA7FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2406}},
    {26984, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFC20, 0x11FC, 0x3104, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1000, 0x1088, 0x1104, 0x1202}},
    {26985, 2, {0x2100, 0x21FE, 0x2110, 0x2210, 0xFAFE, 0x2292, 0x2692, 0x72FE, 0x6A92, 0xA292, 0xA2FE, 0x2290, 0x2250, 0x2220, 0x2258, 0x2286}},
    {26986, 2, {0x1048, 0x1148, 0x1148, 0x13FE, 0xFD48, 0x1148, 0x3178, 0x3900, 0x55FE, 0x5420, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {26987, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1000, 0x37FE, 0x3908, 0x55F8, 0x5508, 0x91F8, 0x1108, 0x113E, 0x17C8, 0x1008, 0x1008}},
    {26988, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3880, 0x35FE, 0x5222, 0x5122, 0x9152, 0x1102, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {26989, 2, {0x440, 0x2448, 0x1450, 0xFFFE, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {26990, 2, {0x1020, 0x1022, 0x11FA, 0x1024, 0xFC28, 0x13FE, 0x3020, 0x3840, 0x54FC, 0x5184, 0x9284, 0x10FC, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {26991, 2, {0x201C, 0x23E0, 0x2220, 0x2220, 0xFBFE, 0x2220, 0x22FC, 0x7284, 0x6A84, 0xA2FC, 0xA284, 0x22FC, 0x2484, 0x2484, 0x28FC, 0x2084}},
    {26992, 2, {0x1020, 0x1020, 0x10A0, 0x112C, 0xFD24, 0x1124, 0x31AC, 0x3924, 0x5524, 0x55FC, 0x9124, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {26993, 2, {0x2040, 0x2040, 0x27FC, 0x2040, 0xFBF8, 0x2080, 0x27FC, 0x7110, 0x6A08, 0xA5F6, 0xA040, 0x23F8, 0x2040, 0x20A0, 0x2110, 0x2608}},
    {26994, 2, {0x2014, 0x2012, 0x27FE, 0x2410, 0xFC10, 0x27F0, 0x2492, 0x7492, 0x6FD2, 0xA554, 0xA554, 0x2488, 0x254A, 0x2A1A, 0x2826, 0x3042}},
    {26995, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x10F8, 0x3088, 0x3888, 0x54F8, 0x5420, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {26996, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFC88, 0x1050, 0x33FE, 0x3A22, 0x5424, 0x55FC, 0x9124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {26997, 2, {0x1000, 0x13FC, 0x1008, 0x1010, 0xFC20, 0x13AE, 0x32A2, 0x3AAA, 0x56A4, 0x52A4, 0x93AA, 0x1032, 0x10A0, 0x1040, 0x17FE, 0x1000}},
    {26998, 2, {0x4100, 0x21FC, 0xA44, 0x1048, 0xE0A0, 0x2110, 0x2608, 0x1FF4, 0x1010, 0x1FF0, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {26999, 2, {0x1110, 0x1112, 0x11D4, 0x1118, 0xFD52, 0x1192, 0x312E, 0x3840, 0x55FC, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {27000, 2, {0x2088, 0x21C8, 0x2708, 0x2108, 0xF92A, 0x212A, 0x27AC, 0x7148, 0x6B08, 0xA388, 0xA554, 0x2514, 0x2914, 0x2124, 0x2124, 0x2142}},
    {27001, 2, {0x1000, 0x13F8, 0x1108, 0x1110, 0xFD1C, 0x12E4, 0x32A4, 0x3A54, 0x54A8, 0x5100, 0x93FC, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {27002, 2, {0x1000, 0x10FC, 0x1048, 0x1030, 0xFDFE, 0x1052, 0x3894, 0x3510, 0x5230, 0x5020, 0x91FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {27003, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x39FC, 0x3400, 0x53FE, 0x5020, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {27004, 2, {0x2020, 0x2124, 0x20A8, 0x2020, 0xFBFE, 0x20A8, 0x2124, 0x7202, 0x6840, 0xA7FE, 0xA088, 0x2108, 0x2190, 0x2060, 0x2198, 0x2604}},
    {27005, 2, {0x100, 0x4204, 0x27C8, 0x1450, 0x440, 0x17D0, 0x2448, 0xC444, 0x7C0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27006, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3820, 0x5432, 0x53B4, 0x90A8, 0x1128, 0x1124, 0x1224, 0x14A2, 0x1040}},
    {27007, 2, {0x101C, 0x11E0, 0x1020, 0x1020, 0xFBFE, 0x10A8, 0x3124, 0x3A22, 0x55FC, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {27008, 2, {0x1000, 0x10F8, 0x1088, 0x1088, 0xFC88, 0x1088, 0x30F8, 0x3888, 0x5400, 0x53DE, 0x9252, 0x1252, 0x1252, 0x1252, 0x13DE, 0x1252}},
    {27009, 2, {0x1040, 0x1020, 0x11FE, 0x1102, 0xFE04, 0x11FC, 0x3040, 0x3888, 0x55FC, 0x5424, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {27010, 2, {0x2000, 0x27BE, 0x2488, 0x2488, 0xFFA8, 0x24A8, 0x24A8, 0x77BE, 0x6C10, 0xA518, 0xA498, 0x25A8, 0x26AA, 0x244A, 0x2046, 0x2080}},
    {27011, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {27012, 2, {0x1090, 0x1290, 0x129E, 0x12A8, 0xFEC4, 0x1080, 0x31FC, 0x3904, 0x5524, 0x5524, 0x9124, 0x1154, 0x1050, 0x1090, 0x1112, 0x160E}},
    {27013, 2, {0x1000, 0x11F8, 0x1108, 0x1108, 0xFDF8, 0x1108, 0x3108, 0x39F8, 0x5400, 0x53FC, 0x9294, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {27014, 2, {0x1040, 0x10A0, 0x1110, 0x1208, 0xFDF6, 0x1000, 0x33C4, 0x3A54, 0x5654, 0x53D4, 0x9254, 0x1254, 0x13D4, 0x1244, 0x1254, 0x12C8}},
    {27015, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0xFC88, 0x1050, 0x33FE, 0x3820, 0x5420, 0x55FC, 0x9020, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {27016, 2, {0x1100, 0x10BC, 0x1284, 0x1204, 0xFEF4, 0x1294, 0x3294, 0x3AF4, 0x5604, 0x52F4, 0x9294, 0x1294, 0x12F4, 0x1204, 0x1214, 0x1208}},
    {27017, 2, {0x1044, 0x1124, 0x1088, 0x1010, 0xFBFE, 0x1088, 0x3124, 0x3A22, 0x54F8, 0x5020, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {27018, 2, {0x1208, 0x1108, 0x1108, 0x103E, 0xFBAA, 0x10AA, 0x30BE, 0x392A, 0x55AA, 0x536A, 0x953E, 0x112A, 0x1108, 0x1108, 0x1108, 0x1108}},
    {27019, 2, {0x1000, 0x11FC, 0x1004, 0x10FC, 0xFC04, 0x11FC, 0x3004, 0x3800, 0x57DE, 0x5442, 0x9042, 0x13DE, 0x1042, 0x1042, 0x13DE, 0x1042}},
    {27020, 2, {0x1000, 0x13FA, 0x100C, 0x104A, 0xFB88, 0x112A, 0x3126, 0x3922, 0x57F8, 0x512A, 0x912C, 0x112A, 0x1128, 0x122A, 0x1226, 0x1422}},
    {27021, 2, {0x2000, 0x23FE, 0x2202, 0x2202, 0xFBFE, 0x2210, 0x2292, 0x7254, 0x6AFE, 0xA282, 0xA2FE, 0x2282, 0x22FE, 0x2482, 0x248A, 0x2884}},
    {27022, 2, {0x13FE, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5480, 0x90FC, 0x1108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {27023, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFBFE, 0x1000, 0x3088, 0x3904, 0x5602, 0x51FC, 0x9154, 0x1154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {27024, 2, {0x1000, 0x13FC, 0x1204, 0x1204, 0xFBFC, 0x1248, 0x3248, 0x3AFC, 0x5648, 0x5248, 0x93FE, 0x12A4, 0x12A8, 0x1490, 0x14C8, 0x1886}},
    {27025, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x9008, 0x13FE, 0x1088, 0x1048, 0x1048, 0x1018}},
    {27026, 2, {0x1000, 0x13DE, 0x1042, 0x1042, 0xFBDE, 0x1210, 0x3210, 0x3BDE, 0x5442, 0x5252, 0x914A, 0x114A, 0x1252, 0x1042, 0x1294, 0x1108}},
    {27027, 2, {0x2040, 0x2028, 0x2090, 0x22A4, 0xFAC2, 0x248A, 0x2188, 0x7678, 0x6800, 0xA3FC, 0xA294, 0x2294, 0x2294, 0x2294, 0x27FE, 0x2000}},
    {27028, 2, {0x1100, 0x109E, 0x13D2, 0x1252, 0xFE54, 0x13D4, 0x3258, 0x3A54, 0x57D2, 0x5212, 0x9292, 0x125A, 0x12D4, 0x1350, 0x1210, 0x1010}},
    {27029, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC50, 0x1088, 0x3124, 0x3850, 0x5488, 0x5104, 0x92FA, 0x1088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {27030, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x48, 0x7F86, 0x4100, 0xBEFC, 0x44, 0x844, 0x7E28, 0x1828, 0x2C10, 0x4A28, 0x8844, 0x982}},
    {27031, 2, {0x1000, 0x13FC, 0x1090, 0x1294, 0xFD98, 0x1090, 0x37FE, 0x3800, 0x55F8, 0x5508, 0x9108, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {27032, 2, {0x2000, 0x20FC, 0x3E80, 0x48F8, 0x8888, 0x7EF8, 0x1480, 0x22FC, 0x4000, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27033, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC20, 0x1050, 0x3888, 0x3504, 0x52FA, 0x5000, 0x9000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {27034, 2, {0x1104, 0x1084, 0x1088, 0x1000, 0xFBFE, 0x1020, 0x3020, 0x39FC, 0x5420, 0x5020, 0x93FE, 0x1000, 0x12A4, 0x1252, 0x1452, 0x1000}},
    {27035, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x3BFE, 0x3488, 0x5134, 0x52E2, 0x9020, 0x11FC, 0x1070, 0x10AC, 0x1322, 0x1020}},
    {27036, 2, {0x1040, 0x1020, 0x13FC, 0x1108, 0xFC90, 0x13FE, 0x3202, 0x3C44, 0x5420, 0x53FC, 0x9080, 0x10F8, 0x1088, 0x1108, 0x1128, 0x1210}},
    {27037, 2, {0x2000, 0x245C, 0x2294, 0x2114, 0xFA94, 0x2454, 0x2126, 0x7100, 0x6FDC, 0xA114, 0xA594, 0x2554, 0x2948, 0x2108, 0x2514, 0x2222}},
    {27038, 2, {0x1000, 0x11FE, 0x1110, 0x1120, 0xFD7C, 0x1144, 0x317C, 0x3944, 0x557C, 0x5510, 0x9110, 0x1154, 0x1152, 0x1292, 0x1250, 0x1420}},
    {27039, 2, {0x1040, 0x1020, 0x13FE, 0x1242, 0xFC20, 0x11FC, 0x3088, 0x3850, 0x57FE, 0x5020, 0x9020, 0x11FC, 0x1020, 0x1020, 0x1020, 0x1020}},
    {27040, 2, {0x1000, 0x13FE, 0x1202, 0x1000, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5440, 0x9020, 0x13FE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {27041, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3840, 0x5488, 0x55F0, 0x9024, 0x13FE, 0x1022, 0x1124, 0x12A2, 0x1040}},
    {27042, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC00, 0x11FC, 0x3040, 0x38A2, 0x5734, 0x5458, 0x9094, 0x1334, 0x1052, 0x1090, 0x1350, 0x1020}},
    {27043, 2, {0x1008, 0x103C, 0x13C0, 0x1004, 0xFA44, 0x1128, 0x31FC, 0x3A20, 0x5420, 0x53FE, 0x9020, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {27044, 2, {0x2208, 0x2208, 0x23BE, 0x2488, 0xFAA8, 0x213E, 0x2208, 0x7448, 0x6848, 0xA7FC, 0xA0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x2040}},
    {27045, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3020, 0x3924, 0x54A8, 0x5020, 0x93FE, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {27046, 2, {0x1020, 0x1020, 0x1050, 0xFE50, 0x1088, 0x7D04, 0x4622, 0x7C20, 0x45FC, 0x7C60, 0x1070, 0xFEA8, 0x10A4, 0x1124, 0x1220, 0x1020}},
    {27047, 2, {0x1000, 0x13FE, 0x1228, 0x1228, 0xFEEE, 0x1228, 0x3228, 0x3AEE, 0x5628, 0x5228, 0x92EE, 0x1228, 0x1228, 0x1228, 0x13FE, 0x1000}},
    {27048, 2, {0x1040, 0x1020, 0x13FE, 0x128A, 0xFD04, 0x1080, 0x30FE, 0x3940, 0x5640, 0x507C, 0x9040, 0x1040, 0x107E, 0x1040, 0x1040, 0x1040}},
    {27049, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFE42, 0x1278, 0x33C4, 0x3A3C, 0x5640, 0x5220, 0x93FE, 0x1288, 0x1250, 0x1420, 0x1458, 0x1986}},
    {27050, 2, {0x1000, 0x11FE, 0x1120, 0x11FC, 0xFD20, 0x11FC, 0x3120, 0x3920, 0x55FE, 0x5002, 0x92AA, 0x12AA, 0x12AA, 0x1202, 0x1014, 0x1008}},
    {27051, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x31FC, 0x3920, 0x55FE, 0x5500, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {27052, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFBFE, 0x1000, 0x31FC, 0x3904, 0x5504, 0x51FC, 0x9052, 0x1094, 0x1188, 0x16A4, 0x10C2, 0x1080}},
    {27053, 2, {0x2084, 0x2104, 0x23C4, 0x2244, 0xFBDE, 0x2244, 0x33C4, 0x6A64, 0x6254, 0xA7D4, 0xA0C4, 0x2144, 0x2244, 0x2444, 0x2154, 0x2088}},
    {27054, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27055, 2, {0x2010, 0x2010, 0x2710, 0x257C, 0xFD10, 0x2510, 0x25FE, 0x7708, 0x6D08, 0xA5FE, 0xA508, 0x2548, 0x2728, 0x2008, 0x2028, 0x2010}},
    {27056, 2, {0x1040, 0x1044, 0x11F8, 0x1050, 0xFBFE, 0x1040, 0x30F8, 0x3982, 0x567E, 0x5400, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {27057, 2, {0x2080, 0x2040, 0x27FE, 0x2000, 0xFBF8, 0x2208, 0x2FFE, 0x7208, 0x6BF8, 0xA0A4, 0xA128, 0x2310, 0x2510, 0x2948, 0x2186, 0x2100}},
    {27058, 2, {0x2000, 0x23F8, 0x2248, 0x22A8, 0xFB18, 0x2208, 0x23F8, 0x7000, 0x6800, 0xA7FC, 0xA4A4, 0x24A4, 0x24A4, 0x24A4, 0x2FFE, 0x2000}},
    {27059, 2, {0x2040, 0x2020, 0x23FE, 0x2202, 0xFAEC, 0x2224, 0x2244, 0x7244, 0x6AEE, 0xA224, 0xA2A4, 0x22A4, 0x224E, 0x2460, 0x249E, 0x2900}},
    {27060, 2, {0x1040, 0x13BE, 0x1212, 0x1292, 0xFE52, 0x12AA, 0x3324, 0x3840, 0x57FE, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {27061, 2, {0x1108, 0x1108, 0x17FE, 0x1108, 0xF800, 0x17FE, 0x3108, 0x39F8, 0x5508, 0x51F8, 0x9108, 0x113E, 0x17C8, 0x1008, 0x1008, 0x1008}},
    {27062, 2, {0x1020, 0x1010, 0x13FE, 0x1220, 0xFAFC, 0x1224, 0x33FE, 0x3A24, 0x56FC, 0x5220, 0x92FC, 0x1284, 0x1284, 0x1484, 0x14FC, 0x1884}},
    {27063, 2, {0x1080, 0x1080, 0x17FE, 0x1522, 0xFD10, 0x13FC, 0x3620, 0x3A20, 0x57FC, 0x5220, 0x9220, 0x13FC, 0x1220, 0x1220, 0x13FE, 0x1200}},
    {27064, 2, {0x1100, 0x113E, 0x112A, 0x112A, 0xF93E, 0x17AA, 0x312A, 0x393E, 0x5508, 0x5108, 0x913E, 0x1188, 0x1708, 0x1208, 0x107E, 0x1000}},
    {27065, 2, {0x1004, 0x101E, 0x13F0, 0x121E, 0xFE10, 0x12FE, 0x3292, 0x3A98, 0x56F2, 0x528E, 0x9280, 0x12B8, 0x12A8, 0x14AA, 0x154A, 0x1A86}},
    {27066, 2, {0x20, 0x3D24, 0x24A8, 0x2420, 0x25FC, 0x3C40, 0x27FE, 0x2488, 0x2524, 0x3E22, 0x25FC, 0x2470, 0x24A8, 0x4524, 0x5624, 0x8820}},
    {27067, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3800, 0x3400, 0x53DE, 0x5042, 0x9252, 0x114A, 0x1252, 0x1042, 0x114A, 0x1084}},
    {27068, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x3080, 0x3904, 0x57FE, 0x5402, 0x93FC, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {27069, 2, {0x103C, 0x17C0, 0x1244, 0x1128, 0xFC80, 0x1110, 0x33E0, 0x3840, 0x5588, 0x53FC, 0x9044, 0x1040, 0x17FC, 0x10A0, 0x1110, 0x160E}},
    {27070, 2, {0x1000, 0x11FC, 0x1104, 0x11E4, 0xFD24, 0x13FE, 0x3202, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {27071, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0xFDFC, 0x1000, 0x33FE, 0x3800, 0x55FC, 0x5504, 0x9104, 0x11FC, 0x1088, 0x1050, 0x13FE, 0x1000}},
    {27072, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FEC, 0x100, 0xFFFE, 0x920, 0x711C}},
    {27073, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x31FC, 0x3800, 0x57FE, 0x5202, 0x92FA, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {27074, 2, {0x2006, 0x2F78, 0x2110, 0x2120, 0xFA44, 0x2278, 0x3310, 0x6A24, 0x667E, 0xAA12, 0xA210, 0x2254, 0x2252, 0x2292, 0x2A50, 0x2420}},
    {27075, 2, {0x1000, 0x3E78, 0x2248, 0x2A48, 0xFE86, 0x2378, 0x2A48, 0x4230, 0x4A48, 0x8484, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27076, 2, {0x103C, 0x13C0, 0x1044, 0x1224, 0xFD28, 0x1100, 0x3040, 0x3B9C, 0x5604, 0x5204, 0x939C, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {27077, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x1104, 0x31FC, 0x3800, 0x57FE, 0x528A, 0x9252, 0x13FE, 0x1222, 0x1222, 0x122A, 0x1204}},
    {27078, 2, {0x1110, 0x1110, 0x17FE, 0x1110, 0xFA00, 0x13FC, 0x3404, 0x3BE4, 0x5624, 0x53E4, 0x9224, 0x13E4, 0x1224, 0x1004, 0x1028, 0x1010}},
    {27079, 2, {0x1118, 0x11E0, 0x1102, 0x10FE, 0xFC00, 0x117C, 0x3144, 0x397C, 0x5544, 0x557C, 0x9144, 0x117C, 0x1100, 0x11FE, 0x1088, 0x1104}},
    {27080, 2, {0x1000, 0x13FE, 0x1200, 0x12FC, 0xFE00, 0x13FE, 0x32A8, 0x3A90, 0x56C8, 0x5286, 0x9208, 0x13FE, 0x1288, 0x1448, 0x1408, 0x1818}},
    {27081, 2, {0x1020, 0x1010, 0x11FE, 0x1100, 0xF920, 0x1520, 0x333C, 0x3950, 0x5590, 0x5310, 0x95FE, 0x1110, 0x1128, 0x1228, 0x1244, 0x1482}},
    {27082, 2, {0x2200, 0x147C, 0xFF44, 0x87C, 0x4944, 0x497C, 0x7F44, 0x844, 0x1094, 0x2108, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27083, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFDFC, 0x1088, 0x33FE, 0x3820, 0x55FC, 0x5524, 0x91FC, 0x1124, 0x13FE, 0x1104, 0x1114, 0x1108}},
    {27084, 2, {0x1010, 0x1220, 0x117C, 0x1144, 0xFC44, 0x107C, 0x3340, 0x3940, 0x557C, 0x5544, 0x9144, 0x117C, 0x1144, 0x1280, 0x147E, 0x1000}},
    {27085, 2, {0x1020, 0x1020, 0x1050, 0x10C8, 0xFD24, 0x13FE, 0x3504, 0x39FC, 0x5504, 0x51FC, 0x9100, 0x11FC, 0x1284, 0x1284, 0x14FC, 0x1084}},
    {27086, 2, {0x1108, 0x1088, 0x1090, 0x13FC, 0xFC40, 0x11F8, 0x3040, 0x3BFE, 0x5480, 0x5500, 0x91FC, 0x1220, 0x1420, 0x1820, 0x13FE, 0x1000}},
    {27087, 2, {0x1088, 0x1050, 0x1000, 0x13FE, 0xFC50, 0x11FC, 0x3054, 0x3BFE, 0x5454, 0x55FC, 0x9050, 0x10D8, 0x1154, 0x1252, 0x1050, 0x1050}},
    {27088, 2, {0x1020, 0x1040, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x3124, 0x3944, 0x55FC, 0x5440, 0x90B0, 0x10A4, 0x113C, 0x1122, 0x1222, 0x141E}},
    {27089, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7CF8, 0x1020, 0x1020, 0xFDFC, 0x3070, 0x38A8, 0x54A8, 0x5124, 0x9222, 0x1020, 0x1020}},
    {27090, 2, {0x1040, 0x1020, 0x13FE, 0x1040, 0xFC88, 0x11F0, 0x3020, 0x3844, 0x55FE, 0x5402, 0x91FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1104}},
    {27091, 2, {0x1000, 0x11FC, 0x1020, 0x1020, 0xFBFE, 0x1000, 0x31FC, 0x3904, 0x55FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {27092, 2, {0x1040, 0x1080, 0x13F8, 0x1208, 0xFBF8, 0x1208, 0x33F8, 0x3880, 0x57FC, 0x5110, 0x9248, 0x1444, 0x1BFA, 0x1040, 0x1040, 0x1040}},
    {27093, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {27094, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0xBFF4, 0x800, 0x1FF0, 0x6810, 0xFF0, 0x100, 0xFFFE, 0x920, 0x711C}},
    {27095, 2, {0x1040, 0x1040, 0x17FE, 0x1080, 0xFDF8, 0x1288, 0x34F8, 0x3800, 0x57FE, 0x5202, 0x92FA, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {27096, 2, {0x1104, 0x1088, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3820, 0x5524, 0x54A8, 0x9070, 0x10A8, 0x1124, 0x1622, 0x10A0, 0x1040}},
    {27097, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x13FE, 0x1088, 0x1104, 0x1202}},
    {27098, 2, {0x1000, 0x13FE, 0x1050, 0x13FE, 0xFE52, 0x1252, 0x33FE, 0x3800, 0x55FC, 0x5504, 0x9124, 0x1124, 0x1124, 0x1050, 0x1088, 0x1304}},
    {27099, 2, {0x1048, 0x1148, 0x114E, 0x1150, 0xFD68, 0x1144, 0x3144, 0x3840, 0x5400, 0x55FC, 0x9154, 0x1154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {27100, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x31FC, 0x3920, 0x55FE, 0x5100, 0x93FC, 0x1088, 0x109E, 0x1102, 0x120A, 0x1404}},
    {27101, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3900, 0x55FE, 0x5500, 0x91FE, 0x1042, 0x124A, 0x13FA, 0x1014, 0x1008}},
    {27102, 2, {0x1080, 0x1040, 0x17FC, 0x1110, 0xF8A0, 0x17FE, 0x3000, 0x3BF8, 0x5648, 0x53F8, 0x9248, 0x13F8, 0x124A, 0x1042, 0x103E, 0x1000}},
    {27103, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC14, 0x11E0, 0x3100, 0x3900, 0x55FC, 0x5110, 0x9110, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {27104, 2, {0x2020, 0x2420, 0x22FA, 0x2224, 0xF828, 0x21FE, 0x2E20, 0x7240, 0x6AFC, 0xA344, 0xA244, 0x227C, 0x22C4, 0x2344, 0x227C, 0x2044}},
    {27105, 2, {0x11F8, 0x1088, 0x1070, 0x118C, 0xFC00, 0x13DE, 0x3252, 0x398C, 0x5652, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {27106, 2, {0x1000, 0x13DE, 0x1042, 0x1252, 0xFD4A, 0x1252, 0x3042, 0x3820, 0x55FC, 0x5504, 0x9104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {27107, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0xFC20, 0x11FC, 0x3050, 0x3888, 0x57FE, 0x5408, 0x91E8, 0x1128, 0x11E8, 0x1008, 0x1028, 0x1010}},
    {27108, 2, {0x1010, 0x1010, 0x12FE, 0x1110, 0xFD7C, 0x1054, 0x307C, 0x3B54, 0x557C, 0x5110, 0x91FE, 0x1110, 0x1110, 0x1290, 0x147E, 0x1000}},
    {27109, 2, {0x1088, 0x13DE, 0x1088, 0x11DC, 0xFC88, 0x13DE, 0x3888, 0x3400, 0x53FC, 0x5004, 0x9004, 0x11FC, 0x1004, 0x1004, 0x13FC, 0x1004}},
    {27110, 2, {0x2020, 0x27FE, 0x2420, 0x25FC, 0xFC24, 0x27FE, 0x2424, 0x75FC, 0x6C20, 0xA5FC, 0xA524, 0x25FC, 0x2524, 0x29FC, 0x2924, 0x312C}},
    {27111, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {27112, 2, {0x2200, 0x211E, 0x27D2, 0x2012, 0xF7D4, 0x2454, 0x27D8, 0x7014, 0x6FD2, 0xA092, 0xA112, 0x21DA, 0x2714, 0x2110, 0x2510, 0x2210}},
    {27113, 2, {0x0, 0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x650E, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27114, 2, {0x2100, 0x223E, 0x2788, 0x2488, 0xFCA8, 0x27A8, 0x24BE, 0x7488, 0x6F88, 0xA010, 0xA498, 0x2718, 0x2428, 0x24AA, 0x23CA, 0x2086}},
    {27115, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x3124, 0x39FC, 0x5422, 0x53FE, 0x9008, 0x13FE, 0x1108, 0x1088, 0x1028, 0x1010}},
    {27116, 2, {0x2200, 0x220E, 0x2F74, 0x2254, 0xF554, 0x28D4, 0x2754, 0x7054, 0x6FD4, 0xA454, 0xA754, 0x2154, 0x2152, 0x2154, 0x255C, 0x2284}},
    {27117, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0xFC10, 0x2490, 0x24D2, 0x7492, 0x6FF2, 0xA494, 0xA494, 0x25C8, 0x2AAA, 0x289A, 0x31A6, 0x2042}},
    {27118, 2, {0x1040, 0x1088, 0x11FC, 0x1108, 0xFA52, 0x13FE, 0x3050, 0x3988, 0x5626, 0x50C0, 0x9310, 0x1064, 0x1388, 0x1030, 0x10C0, 0x1700}},
    {27119, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0xFC88, 0x11FE, 0x3B10, 0x3510, 0x51FE, 0x5110, 0x9110, 0x11FE, 0x1110, 0x1110, 0x11FE, 0x1100}},
    {27120, 2, {0x2020, 0x247C, 0x2284, 0x2148, 0xF830, 0x20CE, 0x2610, 0x727C, 0x6A10, 0xA27C, 0xA210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x2800}},
    {27121, 2, {0x1000, 0x11FE, 0x1048, 0x1048, 0xFDFE, 0x114A, 0x314A, 0x398E, 0x5502, 0x55FE, 0x9102, 0x11FE, 0x1000, 0x1154, 0x112A, 0x122A}},
    {27122, 2, {0x2204, 0x2204, 0x2784, 0x2494, 0xF90C, 0x27C4, 0x2564, 0x7554, 0x6FC4, 0xA546, 0xA57C, 0x27C4, 0x2544, 0x2544, 0x2444, 0x28C4}},
    {27123, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8518, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27124, 2, {0x1040, 0x1020, 0x13FC, 0x1204, 0xFBFC, 0x1200, 0x32F8, 0x3A88, 0x56F8, 0x5200, 0x93FC, 0x1324, 0x15FC, 0x1502, 0x1902, 0x10FE}},
    {27125, 2, {0x1020, 0x1020, 0x11FC, 0x1124, 0xFDFC, 0x1020, 0x33FE, 0x3A22, 0x57FE, 0x5020, 0x9040, 0x1024, 0x12A2, 0x128A, 0x1478, 0x1000}},
    {27126, 2, {0x2000, 0x27FE, 0x242A, 0x2426, 0xF7FE, 0x2422, 0x27AA, 0x76AA, 0x6FAA, 0xA432, 0xA596, 0x262E, 0x2446, 0x2482, 0x27FE, 0x2402}},
    {27127, 2, {0x820, 0x3E20, 0x8F8, 0x7F28, 0x2228, 0x496A, 0x3E2A, 0x856, 0x7E82, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27128, 2, {0x2110, 0x2110, 0x27D0, 0x2110, 0xF93C, 0x27D4, 0x2014, 0x72B4, 0x6D54, 0xA11C, 0xA7D4, 0x2114, 0x2114, 0x21E4, 0x2726, 0x2242}},
    {27129, 2, {0x1020, 0x1040, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3820, 0x5524, 0x54A8, 0x9124, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {27130, 2, {0x2040, 0x2020, 0x27FE, 0x2420, 0xFDFC, 0x2424, 0x27FE, 0x7424, 0x6DFC, 0xA420, 0xA562, 0x24B4, 0x2928, 0x2A26, 0x30A0, 0x2040}},
    {27131, 2, {0x2200, 0x227C, 0x2244, 0x2744, 0xFA7C, 0x2244, 0x2244, 0x777C, 0x6A44, 0xA244, 0xA27C, 0x2228, 0x2528, 0x254A, 0x284A, 0x3086}},
    {27132, 2, {0x11F8, 0x7D08, 0x11F8, 0x1108, 0xFDF8, 0x1108, 0x29F8, 0x2490, 0x4112, 0x860E, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27133, 2, {0x1090, 0x1090, 0x17FE, 0x1090, 0xFBFC, 0x1294, 0x33FC, 0x3A94, 0x57FC, 0x5000, 0x91F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x1108}},
    {27134, 2, {0x1000, 0x11F8, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x3000, 0x3BFC, 0x5694, 0x53FC, 0x9000, 0x11F8, 0x1090, 0x1060, 0x1198, 0x1606}},
    {27135, 2, {0x1088, 0x1088, 0x17FE, 0x1088, 0xFCF8, 0x1020, 0x31FC, 0x3924, 0x55FC, 0x5020, 0x93FE, 0x1020, 0x11FC, 0x1020, 0x17FE, 0x1000}},
    {27136, 2, {0x1040, 0x1020, 0x13FE, 0x1088, 0xFC50, 0x13FE, 0x3222, 0x3AFA, 0x5622, 0x52FA, 0x928A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {27137, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1040, 0x33FE, 0x3888, 0x5504, 0x5242, 0x959C, 0x1104, 0x11DC, 0x1104, 0x11FC, 0x1104}},
    {27138, 2, {0x2104, 0x2204, 0x47C8, 0x9452, 0xE45C, 0x47C8, 0x9452, 0xF45E, 0x17C2, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27139, 2, {0x1100, 0x109E, 0x13D2, 0x1252, 0xFE52, 0x13DE, 0x3252, 0x3A52, 0x57D2, 0x521E, 0x9292, 0x1252, 0x12D2, 0x1352, 0x122A, 0x1044}},
    {27140, 2, {0x2010, 0x2710, 0x25FE, 0x2510, 0xFD7C, 0x2654, 0x257C, 0x7554, 0x6D7C, 0xA510, 0xA538, 0x2654, 0x2492, 0x2410, 0x2410, 0x2410}},
    {27141, 2, {0x2124, 0x2124, 0x2224, 0x2424, 0xF954, 0x214A, 0x2292, 0x7610, 0x6A10, 0xA250, 0xA25C, 0x2250, 0x2250, 0x22B0, 0x229E, 0x2300}},
    {27142, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0xFD54, 0x1124, 0x3154, 0x39FC, 0x5420, 0x53FE, 0x9242, 0x1292, 0x12FA, 0x120A, 0x1202, 0x1206}},
    {27143, 2, {0x2120, 0x213C, 0x2144, 0x22A8, 0xFA10, 0x26A8, 0x2AC6, 0x72BC, 0x6AA4, 0xA2BC, 0xA2A4, 0x22BC, 0x2224, 0x2224, 0x2224, 0x222C}},
    {27144, 2, {0x2080, 0x2040, 0x23F8, 0x2110, 0xF8A0, 0x27FE, 0x2000, 0x73F8, 0x6A08, 0xA3F8, 0xA208, 0x23F8, 0x2120, 0x2122, 0x2222, 0x2C1E}},
    {27145, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFAAA, 0x1124, 0x32AA, 0x3820, 0x56AA, 0x5124, 0x92AA, 0x1050, 0x1050, 0x1088, 0x1104, 0x1602}},
    {27146, 2, {0x2448, 0x2288, 0xF93E, 0x2288, 0x6458, 0x729C, 0xA92A, 0x2288, 0x2548, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {27147, 2, {0x20FE, 0x2404, 0x2228, 0x2210, 0xF8FE, 0x2092, 0x2E92, 0x72FE, 0x6A92, 0xA2FE, 0xA292, 0x2292, 0x2296, 0x2500, 0x28FE, 0x2000}},
    {27148, 2, {0x1000, 0x11FC, 0x1124, 0x13FE, 0xFD24, 0x11FC, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27149, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC20, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27150, 2, {0x1020, 0x1010, 0x13FE, 0x1202, 0xF880, 0x10FE, 0x3110, 0x3B20, 0x557C, 0x5144, 0x9144, 0x117C, 0x1144, 0x1144, 0x117C, 0x1144}},
    {27151, 2, {0x1000, 0x11FC, 0x1124, 0x11FC, 0xFD24, 0x11FC, 0x3040, 0x3888, 0x55F0, 0x5420, 0x9044, 0x13FE, 0x1022, 0x1124, 0x1222, 0x1060}},
    {27152, 2, {0x1020, 0x1020, 0x103E, 0x1020, 0xFC20, 0x13FE, 0x3202, 0x3B26, 0x568A, 0x5252, 0x9326, 0x1252, 0x128A, 0x1326, 0x13FE, 0x1202}},
    {27153, 2, {0x2278, 0x2128, 0x246C, 0x22AA, 0xF82A, 0x2148, 0x2298, 0x7040, 0x6840, 0xA7FC, 0xA0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x2040}},
    {27154, 2, {0x1040, 0x1020, 0x13FE, 0x1242, 0xFC28, 0x1294, 0x3AAA, 0x34CA, 0x50F8, 0x5300, 0x9020, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {27155, 2, {0x1020, 0x11FC, 0x1124, 0x13FE, 0xFD24, 0x11FC, 0x3020, 0x39FC, 0x5524, 0x55FC, 0x9040, 0x13FE, 0x1088, 0x11D0, 0x1070, 0x138C}},
    {27156, 2, {0x1124, 0x1248, 0x1124, 0x1000, 0xFDFC, 0x1124, 0x31FC, 0x3924, 0x55FC, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {27157, 2, {0x2108, 0x2108, 0x27C8, 0x2108, 0xF91E, 0x27D2, 0x2564, 0x7540, 0x6FC8, 0xA108, 0xA388, 0x2548, 0x2954, 0x2114, 0x2124, 0x2142}},
    {27158, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x13FE, 0xFC08, 0x11E8, 0x3928, 0x55E8, 0x9008, 0x1018}},
    {27159, 2, {0x11FC, 0x1020, 0x13FE, 0x1222, 0xFDAC, 0x1020, 0x31AC, 0x3800, 0x55FC, 0x5000, 0x93FE, 0x1080, 0x10FC, 0x1004, 0x1028, 0x1010}},
    {27160, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFE02, 0x10F8, 0x3088, 0x3888, 0x54F8, 0x5420, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {27161, 2, {0x1000, 0x13FE, 0x1050, 0x11FC, 0xFD54, 0x1154, 0x31FC, 0x3800, 0x55FC, 0x5400, 0x93FE, 0x1020, 0x10A8, 0x1124, 0x12A2, 0x1040}},
    {27162, 2, {0x1040, 0x1020, 0x13FE, 0x1250, 0xFE50, 0x13FE, 0x3252, 0x3A52, 0x57FE, 0x5200, 0x9292, 0x12D4, 0x1298, 0x1492, 0x14D2, 0x188E}},
    {27163, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xFA94, 0x24A4, 0x2050, 0x7188, 0x6E26, 0xA0C0, 0xA310, 0x2064, 0x2388, 0x2030, 0x20C0, 0x2700}},
    {27164, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0xFE48, 0x1248, 0x33FE, 0x3A48, 0x5648, 0x5248, 0x9278, 0x1200, 0x1554, 0x152A, 0x1A2A, 0x1000}},
    {27165, 2, {0x1040, 0x107C, 0x1040, 0x13FE, 0xFE42, 0x1278, 0x33C4, 0x3A3C, 0x5600, 0x52FC, 0x9284, 0x12FC, 0x1284, 0x14FC, 0x1484, 0x19FE}},
    {27166, 2, {0x1000, 0x13FE, 0x1200, 0x127C, 0xFE44, 0x1244, 0x327C, 0x3A00, 0x56EE, 0x52AA, 0x92AA, 0x12AA, 0x12EE, 0x1200, 0x13FE, 0x1000}},
    {27167, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0xFBFE, 0x1000, 0x31FC, 0x3904, 0x55FC, 0x5104, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {27168, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x10F8, 0x3020, 0x3BFE, 0x5622, 0x5332, 0x92AA, 0x1376, 0x1222, 0x1222, 0x122A, 0x1224}},
    {27169, 2, {0x1110, 0x1110, 0x17FC, 0x1110, 0xFC00, 0x13F8, 0x3208, 0x3BF8, 0x5608, 0x53F8, 0x9040, 0x17FC, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {27170, 2, {0x1040, 0x1080, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x31FC, 0x3900, 0x55FE, 0x5500, 0x91FE, 0x1002, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {27171, 2, {0x1104, 0x1088, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3820, 0x5410, 0x54E0, 0x9022, 0x13B4, 0x10A8, 0x1124, 0x16A2, 0x1040}},
    {27172, 2, {0x2110, 0x211E, 0x2122, 0x2254, 0xF208, 0x2694, 0x2AE2, 0x7288, 0x6A88, 0xA2BE, 0xA288, 0x22AC, 0x222A, 0x224A, 0x2228, 0x2210}},
    {27173, 2, {0x1088, 0x13FE, 0x1088, 0x1080, 0xFDF8, 0x1288, 0x3070, 0x3B8E, 0x5420, 0x55FC, 0x9020, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {27174, 2, {0x1210, 0x1210, 0x13DE, 0x1528, 0xF884, 0x1020, 0x3010, 0x3BFE, 0x5420, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {27175, 2, {0x2000, 0x245C, 0x2294, 0x2114, 0xFA94, 0x2414, 0x20A6, 0x7140, 0x691C, 0xA7D4, 0xA114, 0x2394, 0x2548, 0x2908, 0x2114, 0x2122}},
    {27176, 2, {0x2000, 0x23FE, 0x2202, 0x23FE, 0xFA10, 0x2254, 0x2238, 0x7254, 0x6A82, 0xA250, 0xA27C, 0x2290, 0x2210, 0x25FE, 0x2410, 0x2810}},
    {27177, 2, {0x1080, 0x1080, 0x11FE, 0x1220, 0xFC40, 0x13FE, 0x3890, 0x3508, 0x53FE, 0x5510, 0x91FE, 0x1110, 0x11FE, 0x1110, 0x11FE, 0x1100}},
    {27178, 2, {0x1110, 0x1110, 0x13FC, 0xFD10, 0x1110, 0x17FE, 0x3040, 0x3BF8, 0x5648, 0x53F8, 0x9248, 0x13F8, 0x1000, 0x1110, 0x1208, 0x1404}},
    {27179, 2, {0x2000, 0x27DC, 0x2514, 0x27D4, 0xFC54, 0x27C8, 0x2508, 0x7514, 0x6FE2, 0xA000, 0xA040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x2000}},
    {27180, 2, {0x1040, 0x1080, 0x13FC, 0x1244, 0xFEF4, 0x1314, 0x32A4, 0x3A44, 0x56A4, 0x53FC, 0x9040, 0x1024, 0x1522, 0x150A, 0x1908, 0x10F8}},
    {27181, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x1088, 0x30F8, 0x3888, 0x57FE, 0x5088, 0x9124, 0x12FA, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {27182, 2, {0x1000, 0x13FE, 0x1050, 0x1050, 0xFBFE, 0x1252, 0x3252, 0x3BFE, 0x5420, 0x5124, 0x9124, 0x1228, 0x1050, 0x1088, 0x1104, 0x1202}},
    {27183, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFD24, 0x10A8, 0x33FE, 0x3800, 0x55FC, 0x5504, 0x9174, 0x1154, 0x1174, 0x1104, 0x11FC, 0x1104}},
    {27184, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0xFE22, 0x11AC, 0x3020, 0x39AC, 0x5400, 0x55FC, 0x9004, 0x11FC, 0x1004, 0x1004, 0x11FC, 0x1004}},
    {27185, 2, {0x2000, 0x2FBE, 0x28A2, 0x2AAA, 0xFAAA, 0x2AAA, 0x2514, 0x78A2, 0x6880, 0xAFFE, 0xA110, 0x2210, 0x23A0, 0x2060, 0x2198, 0x2E04}},
    {27186, 2, {0x1014, 0x17FE, 0x1010, 0x13D0, 0xFC10, 0x13D0, 0x3010, 0x3BD0, 0x5650, 0x53D0, 0x9250, 0x13C8, 0x124A, 0x13CA, 0x1186, 0x1242}},
    {27187, 2, {0x1000, 0x13FC, 0x1004, 0x13FC, 0xFC04, 0x13FC, 0x3000, 0x3BBC, 0x5524, 0x57BC, 0x9008, 0x13FE, 0x1108, 0x1088, 0x10A8, 0x1010}},
    {27188, 2, {0x1110, 0x1094, 0x13D2, 0x1012, 0xFA50, 0x1190, 0x37FE, 0x3810, 0x57D2, 0x5252, 0x9254, 0x13D4, 0x124A, 0x124A, 0x13D6, 0x1022}},
    {27189, 2, {0x1090, 0x1088, 0x11FE, 0x1110, 0xFB10, 0x15FE, 0x3110, 0x3910, 0x55FE, 0x5110, 0x9110, 0x11FE, 0x1100, 0x12A4, 0x1252, 0x1452}},
    {27190, 2, {0x1000, 0x11F8, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x3800, 0x37FE, 0x5240, 0x53DC, 0x9254, 0x13D4, 0x1268, 0x17C8, 0x1054, 0x1062}},
    {27191, 2, {0x820, 0x820, 0x7EFC, 0x1830, 0x2C68, 0x4AA4, 0x8820, 0x7F7C, 0x2244, 0x3E44, 0x2228, 0x3E28, 0x2210, 0x27A8, 0xFA44, 0x282}},
    {27192, 2, {0x1050, 0x1252, 0x1154, 0x1050, 0xFBFE, 0x1088, 0x3050, 0x3BFE, 0x5420, 0x51FC, 0x9020, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1602}},
    {27193, 2, {0x2104, 0x2104, 0x2FE4, 0x2104, 0xF93E, 0x27C4, 0x2004, 0x77E4, 0x6C54, 0xA454, 0xA7C4, 0x2444, 0x2284, 0x20E4, 0x2F14, 0x2408}},
    {27194, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0xFDFC, 0x10A8, 0x30A8, 0x3BFE, 0x54A8, 0x54A8, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {27195, 2, {0x1020, 0x11FC, 0x1124, 0x11FC, 0xFC20, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27196, 2, {0x27FC, 0x2404, 0x27FC, 0x2400, 0xFDF8, 0x2410, 0x27FE, 0x7420, 0x6C60, 0xA79C, 0xA484, 0x2508, 0x29DE, 0x2F08, 0x3108, 0x2318}},
    {27197, 2, {0x1088, 0x1050, 0x13FE, 0x1050, 0xFDFC, 0x1154, 0x318C, 0x3974, 0x5504, 0x55FC, 0x9008, 0x13FE, 0x1108, 0x1088, 0x10A8, 0x1010}},
    {27198, 2, {0x2114, 0x2112, 0x2112, 0x2790, 0xF97E, 0x2150, 0x2FD4, 0x7154, 0x6954, 0xA548, 0xA5EA, 0x2556, 0x2722, 0x2500, 0x24FE, 0x2800}},
    {27199, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0xFBDE, 0x1000, 0x31FC, 0x3924, 0x55FC, 0x5124, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {27200, 2, {0x1040, 0x1020, 0x13FE, 0x1040, 0xFC84, 0x11FE, 0x30A8, 0x38AA, 0x5526, 0x5200, 0x91FC, 0x1154, 0x1154, 0x1154, 0x13FE, 0x1000}},
    {27201, 2, {0x1110, 0x1110, 0x11DE, 0x12A8, 0xFC44, 0x1100, 0x31FC, 0x3A04, 0x55F4, 0x5114, 0x91F4, 0x1114, 0x11F4, 0x1114, 0x1028, 0x1010}},
    {27202, 2, {0x1088, 0x1050, 0x13FE, 0x1050, 0xFDFC, 0x1154, 0x318C, 0x3974, 0x5504, 0x55FC, 0x9020, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {27203, 2, {0x1024, 0x13A8, 0x1092, 0x1514, 0xFA08, 0x1404, 0x33BA, 0x38A8, 0x54C6, 0x5380, 0x923C, 0x1384, 0x10A8, 0x1090, 0x12A8, 0x1144}},
    {27204, 2, {0x2008, 0x2788, 0x2088, 0x2110, 0xFFDE, 0x2494, 0x24A4, 0x7794, 0x6C94, 0xA794, 0xA494, 0x24C8, 0x2788, 0x2C94, 0x20A4, 0x20C2}},
    {27205, 2, {0x1080, 0x1080, 0x11FE, 0x1354, 0xFD54, 0x1154, 0x33FE, 0x3954, 0x5554, 0x5154, 0x97FE, 0x1000, 0x1154, 0x112A, 0x122A, 0x1000}},
    {27206, 2, {0x2000, 0x3FFC, 0x5290, 0x9290, 0x7FFC, 0x1290, 0x1290, 0xFFFE, 0x820, 0x820, 0x7EFC, 0x1820, 0x2C70, 0x4AA8, 0x8924, 0x820}},
    {27207, 2, {0x1038, 0x13C0, 0x1078, 0x13C0, 0xFC78, 0x13C4, 0x383C, 0x3400, 0x50C6, 0x5738, 0x91CE, 0x1738, 0x11CE, 0x1738, 0x114A, 0x1186}},
    {27208, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0xFBFE, 0x1108, 0x339C, 0x3908, 0x5588, 0x563E, 0x9000, 0x13FE, 0x1090, 0x1090, 0x1112, 0x120E}},
    {27209, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFCA8, 0x1124, 0x3222, 0x3904, 0x5504, 0x51DE, 0x9244, 0x1554, 0x109E, 0x1104, 0x1204, 0x1404}},
    {27210, 2, {0x1000, 0x13FE, 0x1050, 0x13DE, 0xFA52, 0x1252, 0x33DE, 0x3850, 0x57FE, 0x5222, 0x9222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {27211, 2, {0x1018, 0x11E0, 0x1040, 0x13FE, 0xFC88, 0x1174, 0x3252, 0x3870, 0x5400, 0x55FC, 0x9104, 0x1174, 0x1154, 0x1174, 0x1104, 0x110C}},
    {27212, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x1202, 0x32F2, 0x3A92, 0x56F2, 0x5292, 0x92F2, 0x1292, 0x1292, 0x1332, 0x1202, 0x1206}},
    {27213, 2, {0x1000, 0x13DE, 0x1252, 0x13DE, 0xFE52, 0x13DE, 0x3202, 0x3AFA, 0x5622, 0x5272, 0x9222, 0x1222, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {27214, 2, {0x103C, 0x13E0, 0x1124, 0x10A8, 0xFBFE, 0x10A8, 0x3124, 0x3A02, 0x55FC, 0x5124, 0x9124, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {27215, 2, {0x1088, 0x1050, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3924, 0x54A8, 0x53FE, 0x9000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {27216, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x8002, 0x3FF8, 0x400, 0x1FF0, 0x6810, 0xFF0, 0x100, 0xFFFE, 0x920, 0x711C}},
    {27217, 2, {0x2040, 0x2040, 0x27FC, 0xF8A0, 0x2514, 0x2208, 0x77FC, 0x6A0A, 0xABF8, 0xA208, 0x23F8, 0x2040, 0x2248, 0x2444, 0x2944, 0x2080}},
    {27218, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0xFE22, 0x11AC, 0x3020, 0x39AC, 0x5400, 0x55FC, 0x9000, 0x13FE, 0x1040, 0x1088, 0x11FC, 0x1084}},
    {27219, 2, {0x203C, 0x27C0, 0x2044, 0xFA24, 0x2108, 0x27FE, 0x7402, 0x6A08, 0xAA08, 0xA3BE, 0x2488, 0x2AA8, 0x213E, 0x2208, 0x2408, 0x2808}},
    {27220, 2, {0x2210, 0x2110, 0x2FD0, 0x2010, 0xF7BE, 0x24A4, 0x27D4, 0x7014, 0x6F94, 0xA094, 0xA114, 0x21C8, 0x2F08, 0x2114, 0x2514, 0x2222}},
    {27221, 2, {0x1124, 0x10A8, 0x13FE, 0x1202, 0xFCF8, 0x1088, 0x30F8, 0x3800, 0x55FC, 0x5410, 0x9110, 0x11FE, 0x1050, 0x1090, 0x1310, 0x1030}},
    {27222, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFE02, 0x10F8, 0x3088, 0x3888, 0x54F8, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {27223, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC00, 0x13DE, 0x3252, 0x3A52, 0x57DE, 0x5252, 0x9252, 0x13DE, 0x1252, 0x1022, 0x102A, 0x1044}},
    {27224, 2, {0x1000, 0x13FC, 0x1048, 0x1030, 0xFBFE, 0x1052, 0x3094, 0x3B50, 0x5420, 0x53FE, 0x9252, 0x128A, 0x1376, 0x1252, 0x1272, 0x1206}},
    {27225, 2, {0x1020, 0x17A4, 0x10A8, 0x1292, 0xF914, 0x1208, 0x35F4, 0x3802, 0x57F8, 0x5208, 0x9208, 0x13F8, 0x1208, 0x1110, 0x1FFE, 0x1000}},
    {27226, 2, {0x1020, 0x11FC, 0x1024, 0x13FE, 0xFC24, 0x11FC, 0x3020, 0x3AAA, 0x57AE, 0x5222, 0x93FE, 0x1222, 0x13AE, 0x12AA, 0x12AA, 0x1422}},
    {27227, 2, {0x2000, 0x2FFE, 0x2800, 0x2A28, 0xF948, 0x2BEE, 0x2892, 0x7884, 0x6AA0, 0xAAA8, 0xABE8, 0x2888, 0x2894, 0x2914, 0x2924, 0x3242}},
    {27228, 2, {0x3FFE, 0x2890, 0x2510, 0x3FDE, 0x2222, 0x2A88, 0x2F88, 0x2294, 0x2422, 0x2880, 0x3FFE, 0x21C0, 0x22A0, 0x4498, 0x5886, 0x8080}},
    {27229, 2, {0x1000, 0x13FE, 0x1050, 0x13FE, 0xFE52, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {27230, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x31FC, 0x3924, 0x55FC, 0x5022, 0x93FE, 0x1042, 0x1024, 0x12A2, 0x128A, 0x1478}},
    {27231, 2, {0x1124, 0x1124, 0x12AA, 0x13AE, 0xF924, 0x12AA, 0x33AE, 0x3924, 0x57FE, 0x5110, 0x9114, 0x1114, 0x128A, 0x124A, 0x1216, 0x1422}},
    {27232, 2, {0x1020, 0x13FE, 0x1040, 0x10F8, 0xFB48, 0x1030, 0x33D0, 0x383C, 0x55C4, 0x5048, 0x9030, 0x13C4, 0x10A8, 0x1190, 0x16A8, 0x10C6}},
    {27233, 2, {0x1080, 0x10F8, 0x1108, 0x13FE, 0xFD12, 0x1122, 0x39FE, 0x3440, 0x50A2, 0x5354, 0x9098, 0x1334, 0x1054, 0x1092, 0x1350, 0x1020}},
    {27234, 2, {0x2020, 0x2720, 0x25FE, 0x2540, 0xFD7C, 0x2690, 0x257E, 0x7500, 0x6D7C, 0xA544, 0xA57C, 0x2644, 0x247C, 0x2444, 0x2444, 0x244C}},
    {27235, 2, {0x2080, 0x2040, 0x2FFE, 0x2882, 0xF548, 0x2514, 0x28F4, 0x7000, 0x6FFC, 0xA444, 0xA7FC, 0x2444, 0x27FC, 0x2444, 0x2454, 0x2408}},
    {27236, 2, {0x100, 0x2488, 0x2424, 0x43E4, 0x0, 0x2424, 0x1212, 0xA0A0, 0xA4A4, 0x1C1C, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27237, 2, {0x2822, 0x11FC, 0x2828, 0x4BFE, 0x9840, 0x29FC, 0x4E84, 0x88FC, 0x2884, 0x10FC, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27238, 2, {0x1020, 0x11FC, 0x1088, 0x1050, 0xFBFE, 0x1000, 0x31FC, 0x3924, 0x55FC, 0x5124, 0x91FC, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {27239, 2, {0x1104, 0x1088, 0x1000, 0x13FE, 0xFE22, 0x12AA, 0x3272, 0x3A22, 0x57FE, 0x5400, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {27240, 2, {0x1020, 0x11FC, 0x1020, 0x1088, 0xFBFE, 0x1088, 0x31FC, 0x3904, 0x55FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {27241, 2, {0x1088, 0x12AA, 0x12DC, 0x1488, 0xF954, 0x1222, 0x33FE, 0x3A02, 0x5400, 0x50F8, 0x9088, 0x1088, 0x108A, 0x110A, 0x1206, 0x1400}},
    {27242, 2, {0x1110, 0x1114, 0x11D2, 0x1250, 0xFA7E, 0x1550, 0x3090, 0x3AA8, 0x5528, 0x5144, 0x9244, 0x1482, 0x1008, 0x12A4, 0x1252, 0x1452}},
    {27243, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFCF8, 0x1000, 0x33FE, 0x3820, 0x55FC, 0x5524, 0x91FC, 0x1124, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {27244, 2, {0x1000, 0x13DE, 0x1088, 0x12A8, 0xFBFE, 0x1188, 0x32DA, 0x3CA6, 0x55F8, 0x5108, 0x9108, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {27245, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFD04, 0x11FC, 0x3020, 0x39FC, 0x5488, 0x5050, 0x93FE, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1020}},
    {27246, 2, {0x1040, 0x13BE, 0x1292, 0x1252, 0xFEAA, 0x1324, 0x3040, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27247, 2, {0x2088, 0x22AA, 0x22DC, 0x2488, 0xF954, 0x2222, 0x2000, 0x73FE, 0x6A42, 0xA444, 0xA3FC, 0x2044, 0x2044, 0x2084, 0x2114, 0x2208}},
    {27248, 2, {0x1040, 0x11FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3104, 0x39FC, 0x5420, 0x5524, 0x90A8, 0x1124, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {27249, 2, {0x2000, 0x27FE, 0x2400, 0x2404, 0xF7E4, 0x2404, 0x25DE, 0x7544, 0x6D44, 0xA5D4, 0xA40C, 0x2544, 0x2484, 0x24E4, 0x2B14, 0x2008}},
    {27250, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC00, 0x11FC, 0x3904, 0x35FC, 0x5088, 0x53FE, 0x9000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {27251, 2, {0x2010, 0x2754, 0x2538, 0x2510, 0xFD7C, 0x2710, 0x25FE, 0x7528, 0x6D24, 0xA752, 0xA5FC, 0x2514, 0x2524, 0x2524, 0x2B54, 0x2088}},
    {27252, 2, {0x1090, 0x1092, 0x12D4, 0x129A, 0xFAD2, 0x174E, 0x3088, 0x39F0, 0x5420, 0x5044, 0x93FE, 0x1022, 0x1128, 0x1224, 0x14A2, 0x1040}},
    {27253, 2, {0x2288, 0x2288, 0x27C8, 0x2290, 0xFA9E, 0x27D4, 0x2024, 0x77D4, 0x6C54, 0xA7D4, 0xA454, 0x27C8, 0x2448, 0x2454, 0x2464, 0x24C2}},
    {27254, 2, {0x2110, 0x2114, 0x2112, 0x27D2, 0xF910, 0x27DE, 0x2470, 0x77D2, 0x6C52, 0xA7D4, 0xA114, 0x2FD8, 0x210A, 0x211A, 0x2126, 0x2142}},
    {27255, 2, {0x2040, 0x2288, 0x22EA, 0x2EAA, 0xF2AC, 0x22E8, 0x22AA, 0x76AA, 0x6AE6, 0xA000, 0xA120, 0x2120, 0x2120, 0x2222, 0x2422, 0x281E}},
    {27256, 2, {0x10F8, 0x1088, 0x1088, 0x10F8, 0xFC88, 0x1088, 0x30F8, 0x3800, 0x57DE, 0x5252, 0x9252, 0x13DE, 0x1252, 0x1252, 0x13DE, 0x1252}},
    {27257, 2, {0x1080, 0x10F8, 0x1110, 0x13FC, 0xFD24, 0x11FC, 0x3924, 0x35FC, 0x5000, 0x53FE, 0x9000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC}},
    {27258, 2, {0x1000, 0x13DE, 0x1252, 0x13DE, 0xFE52, 0x13DE, 0x3202, 0x3AFA, 0x568A, 0x52FA, 0x928A, 0x12FA, 0x128A, 0x1202, 0x120A, 0x1204}},
    {27259, 2, {0x2000, 0x23F8, 0x2048, 0x2048, 0xFBF8, 0x2088, 0x2088, 0x77FC, 0x6800, 0xA7BC, 0xA294, 0x2294, 0x27BC, 0x2294, 0x22D4, 0x2F3E}},
    {27260, 2, {0x2220, 0x223C, 0x2444, 0x2578, 0xF908, 0x2EFE, 0x2220, 0x74D2, 0x681A, 0xAF2C, 0xA0CC, 0x201A, 0x232A, 0x2CC8, 0x2028, 0x2010}},
    {27261, 2, {0x1010, 0x1210, 0x117C, 0x1010, 0xFCFE, 0x1044, 0x3328, 0x39FE, 0x5510, 0x557C, 0x9110, 0x11FE, 0x1110, 0x1110, 0x12FE, 0x1400}},
    {27262, 2, {0x11FC, 0x1104, 0x1104, 0x11FC, 0xFC00, 0x13DE, 0x3252, 0x3A52, 0x57DE, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {27263, 2, {0x1000, 0x13FE, 0x1000, 0x11FC, 0xFD24, 0x11FC, 0x3124, 0x3BFE, 0x5400, 0x55FC, 0x9124, 0x11FC, 0x1124, 0x11FC, 0x1000, 0x13FE}},
    {27264, 2, {0x2040, 0x2FFE, 0x2000, 0x27FC, 0xFC04, 0x25F4, 0x2514, 0x77FC, 0x6800, 0xA3F8, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2000, 0x2FFE}},
    {27265, 2, {0x1020, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x1174, 0x3154, 0x3974, 0x5504, 0x55FC, 0x9008, 0x10F0, 0x1020, 0x13FE, 0x10A8, 0x1326}},
    {27266, 2, {0x1090, 0x13FC, 0x1294, 0x13FC, 0xFE94, 0x13FC, 0x3000, 0x3BFC, 0x5600, 0x52F8, 0x9200, 0x13FE, 0x1520, 0x1514, 0x1548, 0x1986}},
    {27267, 2, {0x1E, 0x7BE4, 0x4928, 0x53FC, 0x6020, 0x53FE, 0x4800, 0x4BFC, 0x6804, 0x53FC, 0x4004, 0x43FC, 0x100, 0xFFFE, 0x920, 0x711C}},
    {27268, 2, {0x2108, 0x2208, 0x27C8, 0x2450, 0xFFDE, 0x2464, 0x27D4, 0x7214, 0x6914, 0xA7D4, 0xA214, 0x23C8, 0x2248, 0x2454, 0x2554, 0x28A2}},
    {27269, 2, {0x1020, 0x113C, 0x1120, 0x13FE, 0xFC10, 0x1014, 0x33FE, 0x3A10, 0x57F0, 0x5254, 0x9254, 0x12D4, 0x1368, 0x144A, 0x1896, 0x1322}},
    {27270, 2, {0x2108, 0x2108, 0x2148, 0x2548, 0xF58E, 0x2908, 0x2288, 0x7448, 0x693E, 0xA162, 0xA5A2, 0x2522, 0x2922, 0x22A2, 0x247E, 0x2822}},
    {27271, 2, {0x10A0, 0x1090, 0x11FE, 0x1320, 0xFDFC, 0x1120, 0x31FC, 0x3920, 0x55FE, 0x5100, 0x93DE, 0x1252, 0x1252, 0x1272, 0x1202, 0x1206}},
    {27272, 2, {0x1000, 0x13FC, 0x1294, 0x1294, 0xFBFC, 0x1000, 0x37FE, 0x3800, 0x57FC, 0x5204, 0x93FC, 0x10A2, 0x1114, 0x1308, 0x1D44, 0x1182}},
    {27273, 2, {0x1000, 0x17E0, 0x125E, 0x13D2, 0xFE52, 0x13D2, 0x3A72, 0x37DE, 0x5040, 0x53FE, 0x9020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {27274, 2, {0x2108, 0x2108, 0x2114, 0x2FD4, 0xF122, 0x27C0, 0x245C, 0x77C8, 0x6C48, 0xA7C8, 0xA13E, 0x2FC8, 0x2108, 0x2108, 0x2108, 0x2108}},
    {27275, 2, {0x1088, 0x13FE, 0x1088, 0x13FE, 0xFC88, 0x1124, 0x32FA, 0x3820, 0x55FC, 0x5524, 0x91FC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1020}},
    {27276, 2, {0x2000, 0x23FC, 0x2294, 0x2294, 0xFBFC, 0x2000, 0x2090, 0x779E, 0x6890, 0xA090, 0xA39C, 0x2090, 0x2090, 0x279E, 0x2090, 0x2090}},
    {27277, 2, {0x1040, 0x1020, 0x11FC, 0x1088, 0xFC50, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1020, 0x1294, 0x128A, 0x147A}},
    {27278, 2, {0x1020, 0x1050, 0x1088, 0x1124, 0xFEFA, 0x1050, 0x3124, 0x3954, 0x55FC, 0x5020, 0x93FE, 0x1242, 0x1292, 0x12FA, 0x120A, 0x1206}},
    {27279, 2, {0x1050, 0x1252, 0x1154, 0x1050, 0xFBFE, 0x1088, 0x3050, 0x3BFE, 0x5420, 0x51FC, 0x9020, 0x13FE, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {27280, 2, {0x1100, 0x11F8, 0x1208, 0x17FE, 0xFA88, 0x1324, 0x33FE, 0x3A00, 0x56FC, 0x5200, 0x92FC, 0x1200, 0x12FC, 0x1484, 0x14FC, 0x1884}},
    {27281, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0xFA22, 0x11AC, 0x3020, 0x39AC, 0x5400, 0x51FC, 0x9124, 0x1124, 0x11FC, 0x1124, 0x1124, 0x11FC}},
    {27282, 2, {0x820, 0x820, 0x7EFC, 0x1830, 0x2C68, 0x4AA6, 0x9FF0, 0x10D0, 0x1710, 0x1110, 0x17D0, 0x1550, 0x27CA, 0x212A, 0x4FE6, 0x8022}},
    {27283, 2, {0x2100, 0x26DC, 0x2454, 0x2454, 0xF6D4, 0x2454, 0x2466, 0x77C0, 0x681C, 0xA014, 0xA7D4, 0x2114, 0x2108, 0x21C8, 0x2714, 0x2222}},
    {27284, 2, {0x1020, 0x1124, 0x10A8, 0x13FE, 0xFE02, 0x10F8, 0x3088, 0x3888, 0x54F8, 0x5400, 0x91FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1104}},
    {27285, 2, {0x800, 0xFF78, 0x848, 0x7F48, 0x4986, 0x7F00, 0x4978, 0x7F48, 0x828, 0xFF90, 0x4928, 0x7F44, 0x100, 0xFFFE, 0x920, 0x711C}},
    {27286, 2, {0x1044, 0x1224, 0x1128, 0x107E, 0xFC10, 0x1028, 0x334A, 0x391C, 0x5528, 0x554C, 0x911A, 0x112A, 0x1148, 0x1110, 0x12FE, 0x1400}},
    {27287, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27288, 2, {0x2010, 0x2008, 0x2788, 0x24BE, 0xFC80, 0x2494, 0x2788, 0x747E, 0x6C08, 0xA788, 0xA6BE, 0x2A88, 0x2A88, 0x2B88, 0x3288, 0x2008}},
    {27289, 2, {0x21F8, 0x2108, 0x21F8, 0x250A, 0xF5FA, 0x250A, 0x25FA, 0x7402, 0x6F9E, 0xA090, 0xA090, 0x2F9E, 0x2492, 0x2492, 0x2492, 0x2892}},
    {27290, 2, {0x2110, 0x2110, 0x27BC, 0x2110, 0xFBB8, 0x2554, 0x2912, 0x7000, 0x6FFC, 0xA044, 0xA240, 0x2278, 0x2240, 0x2540, 0x28FE, 0x3000}},
    {27291, 2, {0x2000, 0x227C, 0x2144, 0x2174, 0xF854, 0x2054, 0x26FE, 0x7282, 0x6ABA, 0xA2AA, 0xA2BA, 0x2282, 0x228A, 0x2284, 0x2500, 0x28FE}},
    {27292, 2, {0x2040, 0x20A0, 0x2318, 0x2DF6, 0xF000, 0x27FC, 0x2554, 0x74E4, 0x6FFC, 0xA000, 0xA3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208}},
    {27293, 2, {0x1010, 0x13D4, 0x1252, 0x1252, 0xFBD0, 0x1010, 0x37FE, 0x3A50, 0x57D2, 0x5252, 0x93D4, 0x1254, 0x12EA, 0x174A, 0x1056, 0x1062}},
    {27294, 2, {0x2200, 0x239E, 0x248A, 0x290A, 0xF7D2, 0x2566, 0x2540, 0x77D4, 0x6D5E, 0xA564, 0xA7C4, 0x2544, 0x255E, 0x2544, 0x2444, 0x28C4}},
    {27295, 2, {0x1000, 0x13FE, 0x1050, 0x13FE, 0xFE52, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27296, 2, {0x2420, 0xFF20, 0x243E, 0x7E44, 0x8284, 0x7A28, 0x4A10, 0x7A28, 0x444, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {27297, 2, {0x1000, 0x11FC, 0x1154, 0x11FC, 0xFC20, 0x11FC, 0x3020, 0x3BFE, 0x5488, 0x5450, 0x91FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {27298, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0xFD04, 0x12FA, 0x3000, 0x3800, 0x55DC, 0x5554, 0x9154, 0x11DC, 0x1088, 0x1088, 0x1154, 0x1222}},
    {27299, 2, {0x2040, 0x2040, 0x27FC, 0x2248, 0xFA48, 0x2554, 0x2FFE, 0x7000, 0x6BF8, 0xA208, 0xA2E8, 0x22A8, 0x22E8, 0x2208, 0x23F8, 0x2208}},
    {27300, 2, {0x1044, 0x1028, 0x1200, 0x11FE, 0xFD20, 0x107C, 0x3044, 0x3B7C, 0x5544, 0x517C, 0x9144, 0x117C, 0x1144, 0x1280, 0x147E, 0x1000}},
    {27301, 2, {0x1104, 0x1088, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x3BFE, 0x3454, 0x5192, 0x5090, 0x93FE, 0x1090, 0x10D4, 0x138A, 0x1096, 0x11A2}},
    {27302, 2, {0x2002, 0x2FE2, 0x2282, 0x2FEA, 0xFAAA, 0x2AAA, 0x2FEA, 0x700A, 0x6BCA, 0xA00A, 0xAFEA, 0x210A, 0x2542, 0x2522, 0x292A, 0x2304}},
    {27303, 2, {0x1088, 0x13FE, 0x1088, 0x1100, 0xFBFC, 0x1554, 0x30D4, 0x3B24, 0x5454, 0x5094, 0x9348, 0x1020, 0x10A4, 0x128A, 0x128A, 0x1478}},
    {27304, 2, {0x1104, 0x1088, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x33FE, 0x3800, 0x5620, 0x513E, 0x9042, 0x1114, 0x1610, 0x1228, 0x1244, 0x1082}},
    {27305, 2, {0x1020, 0x13FE, 0x1000, 0x11FC, 0xFD04, 0x1174, 0x3154, 0x3974, 0x5504, 0x55FC, 0x9000, 0x11FC, 0x1000, 0x13FE, 0x1124, 0x1262}},
    {27306, 2, {0x1020, 0x1040, 0x10F8, 0x128A, 0xFD8C, 0x10F8, 0x318C, 0x3A8A, 0x54F8, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {27307, 2, {0x2080, 0x2040, 0x27FE, 0x2402, 0xFA24, 0x23BC, 0x34A4, 0x6AA8, 0x6510, 0xA2E8, 0xA404, 0x2BFA, 0x2040, 0x2248, 0x2444, 0x20C0}},
    {27308, 2, {0x2110, 0x27FC, 0x2110, 0x27FC, 0xF404, 0x23F8, 0x2000, 0x77FC, 0x6880, 0xA144, 0xA6A8, 0x2170, 0x26A8, 0x2126, 0x26A0, 0x2040}},
    {27309, 2, {0x2200, 0x227C, 0x2544, 0x24C4, 0xF87C, 0x2744, 0x2244, 0x727C, 0x6F52, 0xA252, 0xAA54, 0x2748, 0x2248, 0x2354, 0x2C62, 0x2040}},
    {27310, 2, {0x2040, 0x27FC, 0x2040, 0x23FC, 0xF800, 0x27FE, 0x2002, 0x73F8, 0x6840, 0xA7FE, 0xA000, 0x27FE, 0x2008, 0x27FE, 0x24A8, 0x2798}},
    {27311, 2, {0x2040, 0x2FFE, 0x2040, 0x27FC, 0xF800, 0x23F8, 0x3208, 0x6FFE, 0x6802, 0xA7FC, 0xA110, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x2FFE}},
    {27312, 2, {0x2108, 0x2110, 0x223E, 0x2222, 0xF4BE, 0x2722, 0x213E, 0x7208, 0x6CBE, 0xA7AA, 0xA02A, 0x20AA, 0x256A, 0x252E, 0x2408, 0x2008}},
    {27313, 2, {0x1108, 0x11EE, 0x1294, 0x1042, 0xFC88, 0x1088, 0x33FE, 0x3888, 0x54F8, 0x5488, 0x90F8, 0x1088, 0x13FE, 0x1088, 0x1104, 0x1202}},
    {27314, 2, {0x13FE, 0x1222, 0x13FE, 0x1222, 0xFEFA, 0x12AA, 0x32FA, 0x3AAA, 0x56FA, 0x5226, 0x93FE, 0x120A, 0x13FE, 0x128A, 0x125A, 0x13FE}},
    {27315, 2, {0x2080, 0x2040, 0x27FE, 0x2402, 0xFBFC, 0x2148, 0x2250, 0x77FC, 0x6A04, 0xA7FC, 0xAA04, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {27316, 2, {0x1088, 0x13FE, 0x10A8, 0x1090, 0xFDFE, 0x1320, 0x31FC, 0x3920, 0x55FC, 0x5120, 0x91FE, 0x1100, 0x13FC, 0x1088, 0x1070, 0x138E}},
    {27317, 2, {0x1044, 0x1284, 0x12EA, 0x124E, 0xFEA4, 0x12EE, 0x3222, 0x3BFE, 0x5644, 0x5284, 0x92EA, 0x124E, 0x12A4, 0x12EE, 0x1222, 0x13FE}},
    {27318, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0xFC88, 0x13FE, 0x3200, 0x3A7C, 0x5644, 0x527C, 0x9200, 0x12EE, 0x12AA, 0x12EE, 0x1200, 0x13FE}},
    {27319, 2, {0x2000, 0x27FC, 0x2248, 0x2444, 0xFFFE, 0x2444, 0x26EC, 0x7554, 0x6EEC, 0xA444, 0xA6EC, 0x2554, 0x26EC, 0x2444, 0x2454, 0x2408}},
    {27320, 2, {0x1040, 0x1020, 0x13FE, 0x1222, 0xFD54, 0x114A, 0x323A, 0x3800, 0x55FC, 0x5154, 0x93FE, 0x1000, 0x11FC, 0x1020, 0x10A0, 0x1040}},
    {27321, 2, {0x2420, 0x2220, 0x223E, 0x2F40, 0xF490, 0x247E, 0x2724, 0x7542, 0x6DFE, 0xA504, 0xA574, 0x2554, 0x2974, 0x2904, 0x3314, 0x2008}},
    {27322, 2, {0x2040, 0x27FE, 0x2000, 0x21F8, 0xF908, 0x21F8, 0x2000, 0x77FE, 0x6C02, 0xA3F8, 0xA180, 0x2644, 0x21B8, 0x2668, 0x21A6, 0x2660}},
    {27323, 2, {0x1008, 0x13E8, 0x1288, 0x13EE, 0xFE28, 0x13F4, 0x3282, 0x3BE2, 0x5400, 0x51FC, 0x9154, 0x1154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {27324, 2, {0x203E, 0x27C0, 0x2244, 0x2128, 0xFBF8, 0x2040, 0x27FC, 0x7000, 0x6BF8, 0xA008, 0xA3F8, 0x2008, 0x23F8, 0x2544, 0x252A, 0x28FA}},
    {27325, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0xFA22, 0x11AC, 0x3020, 0x39AC, 0x5400, 0x53FE, 0x9020, 0x11FC, 0x1154, 0x1154, 0x1154, 0x110C}},
    {27326, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8824, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0x820}},
    {27327, 2, {0x3FFE, 0x2894, 0x2F92, 0x2890, 0x2FFE, 0x2010, 0x2F90, 0x28A8, 0x2FA4, 0x28C4, 0x2980, 0x2040, 0x4FFE, 0x4160, 0x8658, 0x1846}},
    {27328, 2, {0x1108, 0x17FE, 0x1108, 0x11F8, 0xFD08, 0x11F8, 0x3908, 0x37FE, 0x5108, 0x5294, 0x95E2, 0x1048, 0x13FC, 0x1040, 0x1248, 0x14C4}},
    {27329, 2, {0x1040, 0x1020, 0x13FE, 0x1242, 0xFC28, 0x1294, 0x3AAA, 0x34CA, 0x5378, 0x5020, 0x91FC, 0x1124, 0x11FC, 0x1022, 0x13FE, 0x1002}},
    {27330, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xF294, 0x24A4, 0x2120, 0x7210, 0x6FFC, 0xAA20, 0xA3FC, 0x2220, 0x23FC, 0x2220, 0x23FE, 0x2200}},
    {27331, 2, {0x27FE, 0x2420, 0x25FC, 0x2524, 0xFDFC, 0x2420, 0x27FE, 0x7504, 0x6DFC, 0xA504, 0xA5FC, 0x2504, 0x25FC, 0x2488, 0x2504, 0x27FE}},
    {27332, 2, {0x101C, 0x11E0, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x31AC, 0x3974, 0x55FC, 0x5420, 0x91FC, 0x1020, 0x13FE, 0x1000, 0x1154, 0x122A}},
    {27333, 2, {0x2080, 0x2040, 0x27FC, 0x2110, 0xF0A4, 0x2F58, 0x2554, 0x7552, 0x6B58, 0xA000, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2408}},
    {27334, 2, {0x2104, 0x2214, 0x27CC, 0x2544, 0xFD54, 0x27CC, 0x2546, 0x755C, 0x6FC4, 0xA104, 0xA2A4, 0x22D4, 0x24F0, 0x2482, 0x287E, 0x2000}},
    {27335, 2, {0x2080, 0x209E, 0x2084, 0x27E8, 0xFCBE, 0x2492, 0x2492, 0x77DE, 0x6E52, 0xA55E, 0xA552, 0x2492, 0x289E, 0x2940, 0x314C, 0x2252}},
    {27336, 2, {0x2024, 0x27A8, 0x2092, 0x2514, 0xF2E8, 0x2404, 0x23FA, 0x7208, 0x6BF8, 0xA110, 0xA7FC, 0x2000, 0x21F0, 0x2110, 0x2212, 0x240E}},
    {27337, 2, {0x2000, 0x27FE, 0x2484, 0x2484, 0xF7E4, 0x249E, 0x25C4, 0x7404, 0x6DD4, 0xA54C, 0xA5C4, 0x2544, 0x2884, 0x28C4, 0x3314, 0x2008}},
    {27338, 2, {0x2000, 0x27BC, 0x24A4, 0x27BC, 0xFCA4, 0x27BC, 0x2484, 0x75F4, 0x6E94, 0xA4E4, 0xA71C, 0x24E4, 0x24A4, 0x24A4, 0x24E4, 0x240C}},
    {27339, 2, {0x1040, 0x11FC, 0x1124, 0x1194, 0xFD4C, 0x1124, 0x39FC, 0x3420, 0x53FE, 0x528A, 0x9124, 0x13FE, 0x1080, 0x10FC, 0x1104, 0x120C}},
    {27340, 2, {0x23FC, 0x2040, 0x21F8, 0x2108, 0xF9F8, 0x2108, 0x27FE, 0x7442, 0x6AA4, 0xA28A, 0xA47A, 0x2080, 0x21F8, 0x2288, 0x2070, 0x238E}},
    {27341, 2, {0x1042, 0x139C, 0x1210, 0x13DE, 0xFE94, 0x1294, 0x3420, 0x3BFC, 0x5604, 0x53FC, 0x9204, 0x13FC, 0x1204, 0x13FC, 0x1108, 0x1204}},
    {27342, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0xFDFC, 0x2488, 0x27FE, 0x7420, 0x6DFC, 0xA524, 0xA5FC, 0x2524, 0x25FC, 0x2400, 0x2488, 0x2904}},
    {27343, 2, {0x2000, 0x20FE, 0x2428, 0x22FE, 0xFAAA, 0x20FE, 0x2010, 0x71FE, 0x6E44, 0xA2FE, 0xA344, 0x227C, 0x2242, 0x223E, 0x2500, 0x28FE}},
    {27344, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0x3EF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {27345, 2, {0x23F8, 0x2248, 0x2248, 0x23F8, 0xFA48, 0x2248, 0x23F8, 0x7000, 0x6FBE, 0xAAAA, 0xAAAA, 0x2FBE, 0x2AAA, 0x2AAA, 0x2FBE, 0x28A2}},
    {27346, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFD10, 0x11DE, 0x3A52, 0x3554, 0x5288, 0x5174, 0x9202, 0x11FC, 0x1020, 0x1124, 0x1222, 0x1060}},
    {27347, 2, {0x1080, 0x11F8, 0x1208, 0x15FC, 0xFD24, 0x11FC, 0x3924, 0x35FC, 0x5154, 0x522A, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {27348, 2, {0x1000, 0x13FE, 0x1248, 0x13FE, 0xFE48, 0x12FC, 0x32A4, 0x3AFC, 0x56A4, 0x52FC, 0x9220, 0x13FE, 0x132A, 0x157A, 0x150A, 0x1906}},
    {27349, 2, {0x1088, 0x11DC, 0x1088, 0x13DE, 0xFC88, 0x1154, 0x3222, 0x39FC, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27350, 2, {0x2020, 0x203E, 0x2020, 0x27FE, 0xFC22, 0x25F8, 0x2422, 0x75FE, 0x6D24, 0xA5FC, 0xA524, 0x25FC, 0x2840, 0x2AA4, 0x328A, 0x247A}},
    {27351, 2, {0x1108, 0x17FE, 0x1108, 0x1000, 0xFBFC, 0x1294, 0x3294, 0x3BFC, 0x5428, 0x53FE, 0x9220, 0x1324, 0x12A8, 0x1292, 0x142A, 0x18C6}},
    {27352, 2, {0x1088, 0x13DE, 0x1088, 0x11DC, 0xFC88, 0x13DE, 0x3088, 0x3BFC, 0x5404, 0x51FC, 0x9004, 0x13FC, 0x1040, 0x12A4, 0x128A, 0x147A}},
    {27353, 2, {0x1088, 0x1088, 0x17FE, 0x1088, 0xFBFE, 0x1200, 0x327C, 0x3A44, 0x567C, 0x5200, 0x92EE, 0x12AA, 0x12AA, 0x12EE, 0x1200, 0x13FE}},
    {27354, 2, {0x1000, 0x13DE, 0x1252, 0x13DE, 0xFE52, 0x13DE, 0x3202, 0x3AFA, 0x568A, 0x52FA, 0x9202, 0x12FA, 0x128A, 0x12FA, 0x1202, 0x1206}},
    {27355, 2, {0x1108, 0x1108, 0x11EE, 0x1294, 0xFC42, 0x1000, 0x33DE, 0x3A52, 0x57D2, 0x5252, 0x93D2, 0x121A, 0x1294, 0x1350, 0x1210, 0x1010}},
    {27356, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x8842, 0x1F60, 0x3250, 0x4C48, 0x37FE, 0xC100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {27357, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0xFC00, 0x13FE, 0x3252, 0x3BFE, 0x5504, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {27358, 2, {0x2120, 0x213C, 0x2144, 0x2278, 0xFA88, 0x24FE, 0x2F20, 0x7152, 0x6A1A, 0xA4AC, 0xAFCC, 0x209A, 0x202A, 0x2AC8, 0x2AA8, 0x2810}},
    {27359, 2, {0x2444, 0x24E4, 0x28A8, 0x2AAA, 0xFEEE, 0x24A4, 0x2AAA, 0x7EEE, 0x6A42, 0xA040, 0xAFFE, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {27360, 2, {0x2020, 0x27FE, 0x2410, 0x27D4, 0xFD4A, 0x2484, 0x2502, 0x77DC, 0x6C54, 0xA5E6, 0xA500, 0x25DC, 0x2844, 0x2854, 0x3148, 0x2094}},
    {27361, 2, {0x2210, 0x2210, 0x23DE, 0x2528, 0xF884, 0x2044, 0x23F8, 0x7050, 0x6FFC, 0xA080, 0xA3F8, 0x2D08, 0x21F8, 0x2108, 0x21F8, 0x2108}},
    {27362, 2, {0x2108, 0x27C8, 0x2548, 0x2FF0, 0xF55E, 0x27E4, 0x2114, 0x77D4, 0x6D54, 0xA7D4, 0xA214, 0x2FC8, 0x2448, 0x2694, 0x2194, 0x2662}},
    {27363, 2, {0x1108, 0x17FE, 0x1108, 0x1410, 0xFAFE, 0x1210, 0x307C, 0x3854, 0x567C, 0x5254, 0x927C, 0x1210, 0x12FE, 0x1210, 0x1510, 0x18FE}},
    {27364, 2, {0x1210, 0x13DE, 0x1528, 0x1084, 0xFD08, 0x1090, 0x37FE, 0x3800, 0x57D4, 0x5254, 0x93D4, 0x1254, 0x13D4, 0x1244, 0x1254, 0x12C8}},
    {27365, 2, {0x1040, 0x1020, 0x13FE, 0x1244, 0xFBF4, 0x1244, 0x32EE, 0x3A04, 0x56F4, 0x52AC, 0x92E4, 0x1214, 0x12A4, 0x1474, 0x1594, 0x1808}},
    {27366, 2, {0x1020, 0x13FE, 0x1250, 0x13FE, 0xFE52, 0x13FE, 0x3200, 0x3BFE, 0x5692, 0x5348, 0x923A, 0x1240, 0x12FC, 0x1548, 0x1430, 0x19CE}},
    {27367, 2, {0x2410, 0x2210, 0x207A, 0x2712, 0xF814, 0x277E, 0x2008, 0x7710, 0x683C, 0xA764, 0xA5A4, 0x253C, 0x2524, 0x2724, 0x253C, 0x2024}},
    {27368, 2, {0x1040, 0x107C, 0x1040, 0x13FC, 0xFE44, 0x13F0, 0x3244, 0x3AFC, 0x56A8, 0x52F8, 0x92A8, 0x12F8, 0x1200, 0x15FC, 0x1554, 0x1BFE}},
    {27369, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x1242, 0x327A, 0x3A8A, 0x5712, 0x52DA, 0x928A, 0x12DA, 0x128A, 0x12FA, 0x120A, 0x1204}},
    {27370, 2, {0x2000, 0x27FE, 0x2444, 0x2598, 0xFC88, 0x27DE, 0x2488, 0x75DC, 0x6EAA, 0xA488, 0xA420, 0x2520, 0x253C, 0x2920, 0x2920, 0x37FE}},
    {27371, 2, {0x20, 0x7F24, 0x10FC, 0x2828, 0x4DFE, 0x1620, 0x247C, 0x4EC4, 0x157C, 0x6444, 0xC7C, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {27372, 2, {0x2200, 0x213C, 0x27A4, 0x2024, 0xF4BC, 0x2324, 0x27A4, 0x713C, 0x6924, 0xA7A4, 0xA13C, 0x2590, 0x2558, 0x2968, 0x252A, 0x2246}},
    {27373, 2, {0x2F7E, 0x2910, 0x2F50, 0x2950, 0xFF7E, 0x2A28, 0x2B28, 0x7D4A, 0x6886, 0xA7F8, 0xA408, 0x27F8, 0x2408, 0x27F8, 0x2000, 0x2FFE}},
    {27374, 2, {0x1000, 0x13FE, 0x1020, 0x13AE, 0xFEAA, 0x12AA, 0x33AE, 0x3820, 0x57FE, 0x5020, 0x93AE, 0x12AA, 0x12AA, 0x13AE, 0x1020, 0x13FE}},
    {27375, 2, {0x2206, 0x2238, 0x2788, 0x2488, 0xF908, 0x27DE, 0x2548, 0x7548, 0x6FDC, 0xA55A, 0xA568, 0x27E8, 0x2008, 0x2AA8, 0x2AA8, 0x2008}},
    {27376, 2, {0x2080, 0x2040, 0x2FFE, 0x2000, 0xF7FC, 0x24A4, 0x27FC, 0x7248, 0x6950, 0xA208, 0xA444, 0x20A8, 0x2110, 0x2308, 0x2D46, 0x2180}},
    {27377, 2, {0x5210, 0x7EFE, 0x2844, 0x7E28, 0xA2FE, 0x3E10, 0x2010, 0x3E7C, 0x2210, 0x3E10, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27378, 2, {0x2020, 0x2F3C, 0x2144, 0x25BE, 0xFA2A, 0x212A, 0x2FBE, 0x7290, 0x6B2A, 0xA25A, 0xA22C, 0x225C, 0x222A, 0x2248, 0x2A28, 0x2410}},
    {27379, 2, {0x2210, 0x2110, 0x27DE, 0x2010, 0xF45E, 0x2282, 0x27DE, 0x7010, 0x6FDE, 0xA450, 0xA7DE, 0x2450, 0x27DE, 0x2450, 0x2452, 0x24CE}},
    {27380, 2, {0x2108, 0x2108, 0x27CE, 0x2112, 0xF924, 0x27DE, 0x2552, 0x755E, 0x6FD2, 0xA11E, 0xA392, 0x255E, 0x2940, 0x2114, 0x2112, 0x2122}},
    {27381, 2, {0x1088, 0x1088, 0x13FE, 0x10A8, 0xFC90, 0x11FE, 0x3120, 0x3BFC, 0x5520, 0x51FC, 0x9120, 0x11FE, 0x1100, 0x12A4, 0x1252, 0x1452}},
    {27382, 2, {0x1020, 0x13FE, 0x1222, 0x11FC, 0xFC20, 0x10F8, 0x3020, 0x39FC, 0x5420, 0x55FC, 0x9154, 0x11FC, 0x1040, 0x12A4, 0x128A, 0x147A}},
    {27383, 2, {0x2080, 0x27FC, 0x2404, 0x27BC, 0xF8A0, 0x26BE, 0x22A2, 0x77B6, 0x6AAA, 0xA6AA, 0xA0B6, 0x26A2, 0x22BE, 0x27A0, 0x22A2, 0x267E}},
    {27384, 2, {0x2140, 0x226C, 0x2244, 0x236C, 0xFAA4, 0x23AC, 0x22A4, 0x77FE, 0x6908, 0xA244, 0xA5F2, 0x2040, 0x27FC, 0x2040, 0x2040, 0x2040}},
    {27385, 2, {0x2110, 0x27FE, 0x2150, 0x23F8, 0xF848, 0x27FE, 0x2048, 0x73F8, 0x6D54, 0xA75C, 0xA444, 0x27FC, 0x2444, 0x275C, 0x2554, 0x2954}},
    {27386, 2, {0x2000, 0x27FC, 0x2040, 0x2FFE, 0xF842, 0x2358, 0x2040, 0x7358, 0x6800, 0xA000, 0xAEEE, 0x2AAA, 0x2AAA, 0x2AAA, 0x2EEE, 0x2AAA}},
    {27387, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x1252, 0x33DE, 0x398C, 0x5652, 0x5040, 0x97FE, 0x1088, 0x1190, 0x1060, 0x10D8, 0x1304}},
    {27388, 2, {0x2148, 0x214C, 0x22AA, 0xF808, 0x27FE, 0x2148, 0x7368, 0x694A, 0xAB6A, 0xA14C, 0x236C, 0x2148, 0x216A, 0x239A, 0x2026, 0x2042}},
    {27389, 2, {0x18, 0x7BE0, 0x4950, 0x53F8, 0x6040, 0x53F8, 0x4808, 0x49F8, 0x6840, 0x5294, 0x44F4, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {27390, 2, {0x2182, 0x2E3C, 0x2088, 0x2A90, 0xF522, 0x243C, 0x2788, 0x7A12, 0x6A3E, 0xAF88, 0xA208, 0x2AAC, 0x2AAA, 0x2BCA, 0x2EA8, 0x2090}},
    {27391, 2, {0x1088, 0x12AA, 0x12DC, 0x1488, 0xF954, 0x1222, 0x33FE, 0x3A02, 0x54F8, 0x5088, 0x90F8, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {27392, 2, {0x2040, 0x27FC, 0x2000, 0x23B8, 0xFAA8, 0x23B8, 0x2110, 0x77FC, 0x6910, 0xA7FC, 0xA110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {27393, 2, {0x2288, 0x2FE8, 0x2AA8, 0x2FE8, 0xFABE, 0x2FE8, 0x3008, 0x6FDC, 0x681A, 0xABE8, 0xA828, 0x2FC8, 0x2AA8, 0x2A48, 0x2B28, 0x3228}},
    {27394, 2, {0x1088, 0x13FE, 0x1088, 0x1412, 0xFAFE, 0x1010, 0x34FE, 0x3A92, 0x56FE, 0x5092, 0x91FE, 0x1292, 0x1604, 0x12FE, 0x1244, 0x122C}},
    {27395, 2, {0x11F0, 0x1210, 0x17FC, 0x1204, 0xFBFC, 0x1224, 0x33B8, 0x3A22, 0x55FE, 0x5210, 0x97FC, 0x1244, 0x13FC, 0x10D0, 0x114A, 0x163E}},
    {27396, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x1222, 0x32FA, 0x3A22, 0x56FA, 0x52AA, 0x92FA, 0x12AA, 0x12FA, 0x1272, 0x12AA, 0x1226}},
    {27397, 2, {0x2140, 0x226C, 0x2244, 0x236C, 0xFAA4, 0x23AC, 0x22A4, 0x77FE, 0x6908, 0xA2F4, 0xA442, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x20C0}},
    {27398, 2, {0x2108, 0x2294, 0x23DE, 0x26B4, 0xFBDE, 0x2294, 0x23DE, 0x7294, 0x6BDE, 0xA210, 0xA7FC, 0x2108, 0x2090, 0x2060, 0x2198, 0x2606}},
    {27399, 2, {0x13FC, 0x1108, 0x11F8, 0x1108, 0xFDF8, 0x110E, 0x37F8, 0x3808, 0x57FE, 0x5294, 0x939C, 0x1294, 0x139C, 0x12D6, 0x17BC, 0x1084}},
    {27400, 2, {0x1222, 0x13FE, 0x1090, 0x11FE, 0xFB10, 0x15FE, 0x3110, 0x39FE, 0x5510, 0x51FE, 0x9100, 0x13FE, 0x128A, 0x1376, 0x1252, 0x1276}},
    {27401, 2, {0x2492, 0x2294, 0x2FFE, 0x2110, 0xF7FC, 0x2040, 0x23F8, 0x7040, 0x6FFE, 0xA480, 0xA7BC, 0x2494, 0x2794, 0x24C8, 0x2F94, 0x20A2}},
    {27402, 2, {0x1088, 0x13FE, 0x1088, 0x13DE, 0xFE52, 0x13DE, 0x30A0, 0x3890, 0x55FE, 0x5120, 0x93FC, 0x1520, 0x11FC, 0x1120, 0x11FE, 0x1100}},
    {27403, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x1252, 0x33DE, 0x3890, 0x55FE, 0x5110, 0x93FE, 0x1510, 0x11FE, 0x1110, 0x11FE, 0x1100}},
    {27404, 2, {0x2090, 0x2090, 0x27FE, 0x2094, 0xF80A, 0x21FE, 0x2508, 0x77E8, 0x694A, 0xA1EA, 0xAF2A, 0x25EA, 0x254C, 0x29EA, 0x2216, 0x2462}},
    {27405, 2, {0x2108, 0x27FE, 0x2128, 0x23FE, 0xFA20, 0x27FC, 0x2A20, 0x73FC, 0x6A20, 0xA3FE, 0xA080, 0x231C, 0x2204, 0x239C, 0x2204, 0x23FC}},
    {27406, 2, {0x2288, 0xF93E, 0x2288, 0x729C, 0xA92A, 0x2288, 0xFFFE, 0x8002, 0x3E08, 0x22FE, 0x3E08, 0x2248, 0x3E28, 0x2808, 0x2428, 0x3A10}},
    {27407, 2, {0x2000, 0x27FC, 0x24A4, 0x24A4, 0xF7FC, 0x2210, 0x22A8, 0x74BE, 0x6F68, 0xA23C, 0xA4A8, 0x2FBC, 0x2028, 0x2AA8, 0x2ABE, 0x2020}},
    {27408, 2, {0x13DE, 0x1000, 0x13DE, 0x1252, 0xFB5A, 0x1252, 0x3020, 0x3BFE, 0x5650, 0x53FE, 0x9252, 0x13FE, 0x1292, 0x12DC, 0x1292, 0x14CE}},
    {27409, 2, {0x2528, 0x27BE, 0x2948, 0x27BE, 0xFB18, 0x25AA, 0x2946, 0x73FC, 0x6A04, 0xA3FC, 0xA204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {27410, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {27411, 2, {0x2248, 0x2150, 0x27FC, 0x2404, 0xF9F0, 0x2110, 0x27FC, 0x7554, 0x6CE4, 0xA7FC, 0xA040, 0x27FC, 0x2040, 0x2FFE, 0x22A4, 0x2452}},
    {27412, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0xFBDE, 0x10A0, 0x31FE, 0x3B20, 0x55FC, 0x5120, 0x91FE, 0x1000, 0x11FC, 0x1088, 0x1070, 0x178E}},
    {27413, 2, {0x27BC, 0x24A4, 0x27BC, 0x2000, 0xF7FE, 0x2400, 0x2590, 0x749E, 0x6FD4, 0xA564, 0xA5D4, 0x2554, 0x25D4, 0x2568, 0x2BC8, 0x3054}},
    {27414, 2, {0x2790, 0x251E, 0x27A8, 0x2484, 0xF7BE, 0x252A, 0x27BE, 0x7000, 0x6BF8, 0xA208, 0xA3F8, 0x2208, 0x23F8, 0x20A0, 0x2124, 0x261C}},
    {27415, 2, {0x1088, 0x13FE, 0x1088, 0x13DE, 0xFE52, 0x13DE, 0x3252, 0x3BDE, 0x5622, 0x53FE, 0x92AA, 0x12FA, 0x1222, 0x1272, 0x12AA, 0x1224}},
    {27416, 2, {0x13FE, 0x1202, 0x13FE, 0x1292, 0xFE54, 0x1292, 0x32FE, 0x3AAA, 0x56FE, 0x5322, 0x92FA, 0x12AA, 0x14FA, 0x1422, 0x19FA, 0x1004}},
    {27417, 2, {0x27FC, 0x2444, 0x27FC, 0x2444, 0xFFFE, 0x2AAA, 0x2FBE, 0x7AAA, 0x6FBE, 0xA220, 0xA7C0, 0x2084, 0x2FFE, 0x2042, 0x2548, 0x2884}},
    {27418, 2, {0x21F0, 0x2210, 0x23E0, 0x2020, 0xF7FC, 0x2188, 0x2650, 0x71A8, 0x6E64, 0xA1A0, 0xA248, 0x2FBE, 0x2AAA, 0x2FBE, 0x228A, 0x2FBE}},
    {27419, 2, {0x23FE, 0x2020, 0x27FE, 0x2422, 0xF9AC, 0x2280, 0x27DE, 0x7292, 0x6B92, 0xA11E, 0xA7D2, 0x2552, 0x27DE, 0x2112, 0x2FD2, 0x2126}},
    {27420, 2, {0x2040, 0x27FC, 0x2248, 0x23F8, 0xF040, 0x2FFE, 0x2AAA, 0x73B8, 0x6910, 0xA7FC, 0xA110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {27421, 2, {0x2288, 0xF93E, 0x2288, 0x729C, 0xA92A, 0x2288, 0x7FFC, 0x4444, 0x7FFC, 0x2208, 0x3EFE, 0x2248, 0x3E28, 0x2808, 0x2428, 0x3A10}},
    {27422, 2, {0x27FC, 0x2040, 0x2FFE, 0x2842, 0xFB58, 0x2040, 0x2358, 0x7000, 0x6EEE, 0xAAAA, 0xAEEE, 0x2000, 0x2FFE, 0x2248, 0x2554, 0x2FFE}},
    {27423, 2, {0x2280, 0x27DE, 0x2292, 0x27D2, 0xFD5E, 0x27D2, 0x3112, 0x6C9E, 0x67D2, 0xAD12, 0xB7DE, 0x2508, 0x27CC, 0x2514, 0x27D4, 0x2422}},
    {27424, 2, {0x800, 0x800, 0x800, 0xFFC, 0x1004, 0x1108, 0x2110, 0x4100, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {27425, 2, {0x80, 0x4080, 0x2080, 0x20FC, 0x104, 0x908, 0xA40, 0x1440, 0x1040, 0xE0A0, 0x20A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {27426, 2, {0x80, 0x80, 0xFC80, 0x4FC, 0x504, 0x4908, 0x2A40, 0x1440, 0x1040, 0x28A0, 0x24A0, 0x4510, 0x8110, 0x208, 0x404, 0x802}},
    {27427, 2, {0x220, 0x720, 0x3820, 0x207E, 0x2042, 0x2084, 0x3F10, 0x2410, 0x2410, 0x2410, 0x2428, 0x2428, 0x2448, 0x4444, 0x4484, 0x8102}},
    {27428, 2, {0x2020, 0x2020, 0x2020, 0x3F7E, 0x2042, 0x4084, 0x4010, 0x7E10, 0x210, 0x210, 0x1A28, 0xE228, 0x4248, 0x244, 0x1484, 0x902}},
    {27429, 2, {0x40, 0x40, 0x7C40, 0x447C, 0x4484, 0x4488, 0x4520, 0x7C20, 0x4420, 0x4450, 0x4450, 0x4450, 0x7C88, 0x4488, 0x104, 0x202}},
    {27430, 2, {0x840, 0x840, 0x1440, 0x147C, 0x2284, 0x5288, 0x8920, 0x820, 0x20, 0x7E50, 0x250, 0x450, 0x488, 0x888, 0x904, 0x202}},
    {27431, 2, {0x40, 0x7E40, 0x4040, 0x447C, 0x6484, 0x5488, 0x4920, 0x4820, 0x5420, 0x5450, 0x6450, 0x4050, 0x4088, 0x7E88, 0x104, 0x202}},
    {27432, 2, {0x2040, 0x2040, 0x4040, 0x7E7C, 0x8284, 0x288, 0x7B20, 0x4A20, 0x4A20, 0x4A50, 0x7A50, 0x4A50, 0x288, 0x288, 0x1504, 0xA02}},
    {27433, 2, {0x20, 0x7F20, 0x1120, 0x117E, 0x2142, 0x2584, 0x4210, 0x8010, 0x3F10, 0x2110, 0x2128, 0x2128, 0x2128, 0x3F44, 0x2144, 0x82}},
    {27434, 2, {0x1020, 0x1020, 0x9220, 0x927E, 0x9242, 0x9284, 0xFE10, 0x1010, 0x1010, 0x9210, 0x9228, 0x9228, 0x9E48, 0xF244, 0x84, 0x102}},
    {27435, 2, {0x2020, 0x2020, 0x3F20, 0x487E, 0x8842, 0x884, 0x7F10, 0x810, 0x810, 0x2E10, 0x2828, 0x2828, 0x2F48, 0xF044, 0x4084, 0x102}},
    {27436, 2, {0x1020, 0x820, 0x820, 0xFF7E, 0x842, 0x1084, 0x2210, 0x7C10, 0x910, 0x1210, 0x2428, 0xC828, 0x1448, 0x2244, 0xC084, 0x102}},
    {27437, 2, {0x20, 0xFE20, 0x8220, 0x927E, 0x9242, 0x9284, 0xFE10, 0x9210, 0x9210, 0xAA10, 0xA628, 0xC228, 0x8248, 0xFE44, 0x8284, 0x102}},
    {27438, 2, {0x4220, 0x2220, 0x2420, 0xFF7E, 0x842, 0x884, 0x4910, 0x4910, 0x4910, 0x7F10, 0x928, 0x1028, 0x1048, 0x2044, 0x4084, 0x8102}},
    {27439, 2, {0x1040, 0x1040, 0x1040, 0xFE7C, 0x1084, 0x1088, 0x7D20, 0x20, 0x20, 0x7C50, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {27440, 2, {0x820, 0x820, 0x1020, 0x7F7E, 0x5542, 0x5584, 0x5510, 0x5510, 0x5510, 0x5510, 0x5528, 0x5528, 0x57A8, 0xFC44, 0x44, 0x82}},
    {27441, 2, {0x1040, 0x1040, 0x2840, 0x247C, 0x4284, 0x8088, 0x7D20, 0x20, 0x20, 0x7C50, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {27442, 2, {0x1420, 0x1220, 0x2120, 0x497E, 0x842, 0x1484, 0x2210, 0x4110, 0xBE10, 0x2210, 0x2228, 0x2228, 0x2248, 0x3E44, 0x2284, 0x102}},
    {27443, 2, {0x820, 0x820, 0xF20, 0x87E, 0x842, 0x7F84, 0x5510, 0x5510, 0x5510, 0x7F10, 0x5528, 0x5528, 0x5528, 0x5144, 0x4544, 0x4282}},
    {27444, 2, {0x1040, 0x840, 0x7E40, 0x427C, 0x4284, 0x7E88, 0x4320, 0x4220, 0x7E20, 0x4050, 0x4850, 0x4450, 0x4A88, 0x5288, 0x6104, 0x4202}},
    {27445, 2, {0x4040, 0x4C40, 0x7040, 0x447C, 0x4484, 0x3C88, 0x4120, 0x7C20, 0x9020, 0x1050, 0xFE50, 0x1050, 0x2888, 0x2488, 0x4504, 0x8202}},
    {27446, 2, {0x820, 0x820, 0xFFA0, 0x87E, 0x842, 0x7F84, 0x4910, 0x4910, 0x7F10, 0x810, 0x1C28, 0x2A28, 0x4948, 0x8844, 0x884, 0x902}},
    {27447, 2, {0x4220, 0x2420, 0x1820, 0x247E, 0x5242, 0x1084, 0xFF10, 0x2010, 0x2810, 0x7E10, 0xAA28, 0x2A28, 0x2A48, 0x2E44, 0x884, 0x902}},
    {27448, 2, {0x1020, 0x1020, 0x2020, 0x447E, 0xFE42, 0x284, 0x4010, 0x7C10, 0x9010, 0x1010, 0xFE28, 0x1028, 0x1048, 0x2844, 0x4484, 0x8502}},
    {27449, 2, {0x820, 0x820, 0x7F20, 0x143E, 0x2222, 0x4144, 0xFF50, 0x290, 0x7A10, 0x4A10, 0x4A10, 0x7A28, 0x4A28, 0x244, 0xA44, 0x482}},
    {27450, 2, {0x2220, 0x2220, 0xFFA0, 0x227E, 0x2242, 0x3E84, 0x2210, 0x2210, 0x3E10, 0x2210, 0x2228, 0xFFA8, 0x2448, 0x2244, 0x4284, 0x8102}},
    {27451, 2, {0x1040, 0x5240, 0x5440, 0x907C, 0x2884, 0x4488, 0x8320, 0x1020, 0x1020, 0x5250, 0x5450, 0x9050, 0x2888, 0x2488, 0x4304, 0x8202}},
    {27452, 2, {0x20, 0xF720, 0x1120, 0x557E, 0x2242, 0x5584, 0x8910, 0x1010, 0x10, 0xF710, 0x1128, 0x5528, 0x2228, 0x5544, 0x8944, 0x1082}},
    {27453, 2, {0x1040, 0x1040, 0x2840, 0x247C, 0x4284, 0xBC88, 0x1120, 0x1020, 0xFE20, 0x1050, 0x9450, 0x5850, 0x5088, 0x1E88, 0xF104, 0x4202}},
    {27454, 2, {0x820, 0x820, 0xFF20, 0x87E, 0x842, 0x7E84, 0x10, 0x7E10, 0x10, 0xFF10, 0x828, 0x4A28, 0x4948, 0x8944, 0x2884, 0x1102}},
    {27455, 2, {0x1020, 0x1020, 0x3E20, 0x227E, 0x4442, 0x8084, 0x1010, 0x6610, 0x4210, 0x4210, 0x6628, 0x4228, 0x4248, 0x7E44, 0x4284, 0x102}},
    {27456, 2, {0x820, 0x820, 0x7F20, 0x1C7E, 0x2A42, 0x4984, 0x10, 0x7E10, 0x10, 0xFF10, 0x828, 0x4A28, 0x4948, 0x8944, 0x2884, 0x1102}},
    {27457, 2, {0x2220, 0x2220, 0xFF20, 0x227E, 0x3E42, 0x2284, 0x3E10, 0x2210, 0x2210, 0xFF10, 0x4028, 0x5428, 0x6248, 0x4044, 0x7E84, 0x102}},
    {27458, 2, {0x1020, 0x9220, 0x9220, 0x927E, 0xFE42, 0x84, 0xFE10, 0x1010, 0x2010, 0xFE10, 0xAA28, 0xAA28, 0xAA28, 0xAA44, 0xAA44, 0x8682}},
    {27459, 2, {0x220, 0xF20, 0x7820, 0x87E, 0x842, 0xFF84, 0x2810, 0x4B10, 0x4910, 0x4910, 0x6B28, 0x4928, 0x4948, 0x7F44, 0x4184, 0x102}},
    {27460, 2, {0x20, 0x3E20, 0x2220, 0x227E, 0x3A42, 0x2A84, 0x2A10, 0x7F10, 0x4110, 0x5D10, 0x5528, 0x5528, 0x5D28, 0x4144, 0x4544, 0x4282}},
    {27461, 2, {0x40, 0xFE40, 0x2840, 0x287C, 0xFE84, 0xAA88, 0xAB20, 0xFE20, 0x1020, 0x1050, 0xFE50, 0x1050, 0x1088, 0x1E88, 0xF104, 0x4202}},
    {27462, 2, {0x1020, 0x820, 0xFF20, 0x7E, 0x4242, 0x2484, 0xFF10, 0x10, 0x7E10, 0x4210, 0x4228, 0x7E28, 0x4248, 0x4244, 0x7E84, 0x4302}},
    {27463, 2, {0x20, 0x7E20, 0x4220, 0x7E7E, 0x4242, 0x7E84, 0x2010, 0x7F10, 0x8910, 0x4910, 0x5528, 0x4128, 0x7D28, 0x144, 0xA44, 0x482}},
    {27464, 2, {0x1020, 0x2820, 0x4420, 0x827E, 0x7D42, 0x84, 0xF110, 0x9510, 0x9510, 0xF510, 0x9528, 0x9528, 0xF528, 0x9144, 0x9544, 0xB282}},
    {27465, 2, {0x2220, 0x1420, 0xFFA0, 0x147E, 0x1442, 0x7F84, 0x1510, 0xFF90, 0x1510, 0x7F10, 0x1428, 0x3628, 0x5528, 0x94C4, 0x1444, 0x1482}},
    {27466, 2, {0x1020, 0x820, 0xFF20, 0x7E, 0x7E42, 0x4284, 0x7E10, 0x10, 0xFF10, 0x8110, 0xBD28, 0xA528, 0xBD28, 0x8144, 0x8544, 0x8282}},
    {27467, 2, {0x320, 0x7C20, 0x4720, 0x447E, 0x5F42, 0x5584, 0x5610, 0x5C10, 0x5510, 0x5310, 0x5028, 0x5728, 0x5548, 0x5544, 0xA984, 0x102}},
    {27468, 2, {0x20, 0xFF20, 0x220, 0x7A7E, 0x4A42, 0x7A84, 0x10, 0xFF10, 0x210, 0x7A10, 0x4A28, 0x4A28, 0x7A28, 0x244, 0xA44, 0x482}},
    {27469, 2, {0x1020, 0x2020, 0x7F20, 0x417E, 0x4142, 0x7F84, 0x4010, 0x7FD0, 0x4010, 0x7F90, 0xA8, 0xAAA8, 0xAAA8, 0xC4, 0x544, 0x282}},
    {27470, 2, {0x2220, 0x2220, 0xFF20, 0x227E, 0x3E42, 0x884, 0x7F10, 0x4910, 0x7F10, 0x810, 0x7F28, 0x828, 0xFF48, 0x1444, 0x2284, 0x4102}},
    {27471, 2, {0x2220, 0x2220, 0xFFA0, 0x227E, 0x3E42, 0x884, 0x7F10, 0x4910, 0x7F10, 0x810, 0x7F28, 0x828, 0x7F28, 0x844, 0xF44, 0xF082}},
    {27472, 2, {0x20, 0xFFA0, 0x8020, 0x9F7E, 0x9142, 0x9184, 0x9F10, 0x8010, 0xBB90, 0xAA90, 0xAAA8, 0xAAA8, 0xBBA8, 0x8044, 0xFFC4, 0x82}},
    {27473, 2, {0x410, 0x790, 0x410, 0x3FDE, 0x2452, 0x2724, 0x3C88, 0x2388, 0x2088, 0x2F08, 0x3248, 0x2A94, 0x5FD4, 0x4214, 0x8A14, 0x422}},
    {27474, 2, {0x2020, 0x1020, 0xFE20, 0x447E, 0x2842, 0xFE84, 0x8210, 0x9210, 0xFE10, 0x9210, 0xBA28, 0xAA28, 0xAA28, 0xBA44, 0x8244, 0x8682}},
    {27475, 2, {0x2020, 0x2020, 0x3F20, 0x483E, 0x8822, 0x7F44, 0x1450, 0x2290, 0x7F10, 0xA410, 0x3F10, 0x2428, 0x3F28, 0x2444, 0x3F44, 0x2082}},
    {27476, 2, {0x1020, 0x1E20, 0x1020, 0x7F3E, 0x5122, 0x5C44, 0x7250, 0x4E90, 0x4010, 0x4A10, 0x6A90, 0x5B28, 0x4A28, 0x8F44, 0x7844, 0x82}},
    {27477, 2, {0x820, 0x7F20, 0x820, 0x223E, 0xFFA2, 0x2244, 0x7F50, 0x4190, 0x7F10, 0x4110, 0x7F10, 0x4128, 0x7F28, 0x2244, 0x4144, 0x8082}},
    {27478, 2, {0x820, 0xFF20, 0x820, 0x7E7E, 0x42, 0x7E84, 0x4210, 0x7E10, 0x2410, 0xFF10, 0x28, 0x7E28, 0x4228, 0x4244, 0x7E44, 0x4282}},
    {27479, 2, {0x410, 0x3F90, 0x490, 0x7FDE, 0x492, 0x3FA4, 0x408, 0x5548, 0x75C8, 0x4448, 0x7FC8, 0x4454, 0x75D4, 0x5554, 0x5554, 0x8462}},
    {27480, 2, {0x820, 0x2920, 0x2920, 0x4A7E, 0x1442, 0x2284, 0x4110, 0x10, 0x2210, 0x2A90, 0xB728, 0x6228, 0x2228, 0x5244, 0x4544, 0x8882}},
    {27481, 2, {0x820, 0x1420, 0x2220, 0x5D3E, 0x80A2, 0x3E44, 0x2250, 0x3E90, 0x10, 0x7710, 0x1110, 0x5528, 0x3328, 0x5544, 0x1144, 0x3382}},
    {27482, 2, {0x1110, 0xA10, 0x7FD0, 0x41E, 0x3F92, 0x424, 0x7FC8, 0x2488, 0x1508, 0x7FC8, 0x8, 0x3F94, 0x2094, 0x2094, 0x3F94, 0x20A2}},
    {27483, 2, {0x820, 0x820, 0x1420, 0x227E, 0x4142, 0xBE84, 0x10, 0x10, 0x7710, 0x5510, 0x5528, 0x7728, 0x2228, 0x2244, 0x5544, 0x8882}},
    {27484, 2, {0x20, 0x7F20, 0x5520, 0x557E, 0x7F42, 0x2084, 0x7F10, 0x9110, 0x7D10, 0x5510, 0x5528, 0x7D28, 0x1128, 0x7D44, 0x544, 0x282}},
    {27485, 2, {0x20, 0x7F20, 0x5520, 0x7F7E, 0x842, 0x7F84, 0x810, 0xFF90, 0x2210, 0x1410, 0x7F28, 0x828, 0xFFA8, 0x844, 0x844, 0x882}},
    {27486, 2, {0x20, 0x7F20, 0x4120, 0x7F7E, 0x4142, 0x7F84, 0x2210, 0x5510, 0x7710, 0x2210, 0x5528, 0x7728, 0x28, 0x5544, 0x5544, 0x82}},
    {27487, 2, {0x2820, 0x4B20, 0x4D20, 0x497E, 0x6B42, 0x4D84, 0x5510, 0x7710, 0x5510, 0x5510, 0xFFA8, 0x28, 0x2228, 0x2144, 0x4144, 0x8082}},
    {27488, 2, {0x7720, 0x5520, 0x2220, 0x557E, 0xF742, 0x5584, 0x2210, 0x5510, 0xFF90, 0x1410, 0x7F28, 0x5528, 0x6328, 0x5D44, 0x4144, 0x7F82}},
    {27489, 2, {0x2220, 0x2220, 0xFFA0, 0x227E, 0x7742, 0x5584, 0x7710, 0x1210, 0x3F10, 0x6410, 0xBF28, 0x2428, 0x3F28, 0x2444, 0x3F44, 0x2082}},
    {27490, 2, {0x100, 0x100, 0x100, 0x100, 0x1100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x1100, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {27491, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {27492, 2, {0x440, 0x440, 0x440, 0x2444, 0x2448, 0x2750, 0x2460, 0x2440, 0x2440, 0x2440, 0x2440, 0x2442, 0x2F42, 0xF042, 0x403E, 0x0}},
    {27493, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x100, 0x100, 0x1108, 0x1110, 0x2120, 0x40C0, 0x300, 0x1C00, 0xE000}},
    {27494, 2, {0x40, 0x50, 0x3F48, 0x48, 0x40, 0xFFFE, 0x40, 0x440, 0x440, 0x2740, 0x2420, 0x2422, 0x2412, 0x278A, 0xF806, 0x4002}},
    {27495, 2, {0x820, 0x820, 0x820, 0xBFE, 0x4820, 0x4E20, 0x4820, 0x49FC, 0x4884, 0x4888, 0x4848, 0x4E50, 0x5820, 0xE050, 0x188, 0x606}},
    {27496, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {27497, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x100, 0x110, 0x1108, 0x1114, 0x2124, 0x40C0, 0x300, 0x1C00, 0xE000}},
    {27498, 2, {0x0, 0x7FFC, 0x100, 0x300, 0x560, 0x1918, 0x6104, 0x0, 0x7FFC, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {27499, 2, {0x800, 0x8FE, 0x880, 0x880, 0x4880, 0x4EFC, 0x4884, 0x4884, 0x4884, 0x4884, 0x48FC, 0x4E80, 0x5880, 0xE080, 0xFE, 0x0}},
    {27500, 2, {0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x100, 0x200, 0x1FF0, 0x1010, 0x1210, 0x1110, 0xFFFE, 0x1210, 0x2110, 0x4050, 0x8020}},
    {27501, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5020, 0x5C20, 0x53FE, 0x5008, 0x5008, 0x51FE, 0x5008, 0x5C88, 0x7048, 0xC008, 0x28, 0x10}},
    {27502, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0, 0x810, 0x810, 0x4890, 0x4E9C, 0x4890, 0x4890, 0x4E90, 0xF1FE, 0x0}},
    {27503, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0, 0x2928, 0x2548, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x2008, 0x3FF8, 0x8}},
    {27504, 2, {0x0, 0x7EFC, 0x1224, 0x52A4, 0x9324, 0x2244, 0x2A54, 0x4488, 0x810, 0x810, 0x4E9C, 0x4890, 0x4890, 0x4E90, 0xF1FE, 0x0}},
    {27505, 2, {0x1008, 0x103C, 0x11E0, 0x1020, 0x53FE, 0x5C20, 0x51FC, 0x5124, 0x51FC, 0x5124, 0x51FC, 0x5C20, 0x71FC, 0xC020, 0x3FE, 0x0}},
    {27506, 2, {0x80, 0x8FC, 0x880, 0x880, 0x7FFE, 0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2224, 0x2A18, 0x2A90, 0x512A, 0x4646, 0x9882}},
    {27507, 2, {0x80, 0x8FC, 0x880, 0x880, 0x7FFE, 0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2224, 0x2B18, 0x2A92, 0x52AA, 0x4A46, 0x8482}},
    {27508, 2, {0x3FFE, 0x2210, 0x2210, 0x3FBE, 0x2610, 0x2738, 0x2AD4, 0x3292, 0x2210, 0x2250, 0x2040, 0x227C, 0x4240, 0x4240, 0x9FFE, 0x0}},
    {27509, 2, {0x1020, 0x11FE, 0x1020, 0x11FC, 0x5020, 0x5DFE, 0x5000, 0x51FC, 0x5104, 0x51FC, 0x5104, 0x5DFC, 0x7104, 0xC1FC, 0x88, 0x104}},
    {27510, 2, {0x844, 0x2F78, 0x2842, 0x2F3E, 0xF000, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2148, 0x2FE8, 0x2428, 0x2018}},
    {27511, 2, {0x3FFE, 0x2000, 0x230C, 0x3C70, 0x2410, 0x3F7E, 0x2410, 0x2E38, 0x3554, 0x2492, 0x2080, 0x24F8, 0x4480, 0x4480, 0xBFFE, 0x0}},
    {27512, 2, {0x1000, 0x7DFC, 0x4404, 0x7CFC, 0x4004, 0x7DFC, 0x4400, 0x7FFE, 0x1222, 0x1020, 0x5DFC, 0x5124, 0x5124, 0x5D34, 0xE128, 0x20}},
    {27513, 2, {0x0, 0xFFFE, 0x200, 0x200, 0x400, 0x7F8, 0x808, 0x1010, 0x2410, 0x4220, 0x140, 0x80, 0x100, 0x200, 0xC00, 0x7000}},
    {27514, 2, {0x80, 0x80, 0xFE, 0x880, 0x880, 0xFF8, 0x1008, 0x1410, 0x2210, 0x4120, 0x140, 0x80, 0x100, 0x200, 0xC00, 0x7000}},
    {27515, 2, {0x0, 0xFFFC, 0x1080, 0x1080, 0x1084, 0x1E88, 0x2290, 0x22A0, 0x52C0, 0x8C80, 0x480, 0x884, 0x884, 0x1084, 0x207C, 0x4000}},
    {27516, 2, {0x8, 0x3C, 0xFDE0, 0x2020, 0x2020, 0x3C20, 0x4420, 0x47FE, 0x6420, 0x9420, 0x820, 0x820, 0x1020, 0x2020, 0x4020, 0x8020}},
    {27517, 2, {0x8, 0x1C, 0xFDE0, 0x2100, 0x2100, 0x3D00, 0x45FE, 0x4510, 0x6510, 0x9510, 0x910, 0x910, 0x1110, 0x2210, 0x4210, 0x8410}},
    {27518, 2, {0x80, 0x80, 0xFC80, 0x20FC, 0x2154, 0x3E54, 0x4454, 0x4494, 0x6494, 0x9524, 0xA24, 0x844, 0x1044, 0x2084, 0x4128, 0x8010}},
    {27519, 2, {0x80, 0x80, 0xFCFC, 0x2104, 0x2104, 0x3E28, 0x4410, 0x47FC, 0x6504, 0x9488, 0x888, 0x850, 0x1020, 0x2050, 0x4188, 0x8606}},
    {27520, 2, {0x8, 0x3C, 0xFDE0, 0x2020, 0x2020, 0x3C20, 0x45FE, 0x4420, 0x6420, 0x9450, 0x850, 0x850, 0x1088, 0x2088, 0x4104, 0x8202}},
    {27521, 2, {0x0, 0xF8, 0xFC88, 0x2088, 0x2088, 0x3D06, 0x4600, 0x45FC, 0x6484, 0x9484, 0x848, 0x850, 0x1020, 0x2050, 0x4088, 0x8306}},
    {27522, 2, {0x0, 0x1F8, 0xFD08, 0x2108, 0x2108, 0x3DF8, 0x4508, 0x4508, 0x6508, 0x95F8, 0x908, 0x908, 0x1108, 0x2108, 0x47FE, 0x8000}},
    {27523, 2, {0x20, 0x20, 0xFC20, 0x21FC, 0x2124, 0x3D24, 0x4524, 0x4524, 0x6524, 0x97FE, 0x820, 0x850, 0x1050, 0x2088, 0x4104, 0x8202}},
    {27524, 2, {0x20, 0x20, 0xFC50, 0x2088, 0x2104, 0x3E12, 0x4420, 0x4440, 0x6588, 0x9410, 0x820, 0x844, 0x1188, 0x2010, 0x4060, 0x8380}},
    {27525, 2, {0x20, 0x20, 0xFD20, 0x2120, 0x21FC, 0x3D20, 0x4620, 0x4420, 0x6420, 0x95FC, 0x820, 0x820, 0x1020, 0x2020, 0x43FE, 0x8000}},
    {27526, 2, {0x20, 0xFC20, 0x2020, 0x2040, 0x2048, 0x3C84, 0x45FE, 0x4482, 0x6400, 0x94FC, 0x884, 0x884, 0x1084, 0x2084, 0x40FC, 0x8084}},
    {27527, 2, {0x80, 0x80, 0xFCFE, 0x2100, 0x22F8, 0x3C10, 0x4420, 0x4440, 0x65FC, 0x9494, 0x894, 0x924, 0x1224, 0x2444, 0x40A8, 0x8110}},
    {27528, 2, {0x20, 0x20, 0xFC40, 0x2080, 0x21FC, 0x3D54, 0x4554, 0x4554, 0x6554, 0x9554, 0x954, 0x954, 0x1154, 0x2154, 0x43FE, 0x8000}},
    {27529, 2, {0x80, 0xFE80, 0x20FC, 0x2104, 0x2104, 0x3FF4, 0x4514, 0x4514, 0xA5F4, 0x1914, 0x914, 0x11F4, 0x1114, 0x2004, 0x4014, 0x8008}},
    {27530, 2, {0x20, 0x120, 0xFD20, 0x21FC, 0x2120, 0x3E20, 0x4420, 0x47FE, 0x6470, 0x94A8, 0x8A8, 0x924, 0x1124, 0x2222, 0x4020, 0x8020}},
    {27531, 2, {0x50, 0x48, 0xFC40, 0x205C, 0x21E0, 0x3C40, 0x445E, 0x47E0, 0x6444, 0x9448, 0x830, 0x822, 0x1052, 0x208A, 0x4306, 0x8002}},
    {27532, 2, {0x0, 0x3FC, 0xF800, 0x2124, 0x2248, 0x3C90, 0x4A48, 0x4924, 0x6800, 0x9BFC, 0x840, 0x1040, 0x1040, 0x2040, 0x47FE, 0x8000}},
    {27533, 2, {0x8, 0x3C, 0xFDC0, 0x2004, 0x2144, 0x3CA8, 0x4400, 0x45F8, 0x6410, 0x9420, 0xBFE, 0x820, 0x1020, 0x2020, 0x40A0, 0x8040}},
    {27534, 2, {0x20, 0x20, 0xFC20, 0x23FE, 0x2020, 0x3C20, 0x4524, 0x4524, 0x6524, 0x96AA, 0x820, 0x850, 0x1050, 0x2088, 0x4104, 0x8202}},
    {27535, 2, {0x28, 0x24, 0xFC24, 0x2020, 0x23FE, 0x3C20, 0x4520, 0x44B2, 0x64B4, 0x9468, 0x8A8, 0x924, 0x1222, 0x2020, 0x40A0, 0x8040}},
    {27536, 2, {0x20, 0x20, 0xFDFE, 0x2020, 0x2020, 0x3DFC, 0x4524, 0x4524, 0x65FC, 0x9420, 0x870, 0x8A8, 0x1124, 0x2222, 0x4020, 0x8020}},
    {27537, 2, {0x20, 0xFC20, 0x23FE, 0x2020, 0x2020, 0x3DFC, 0x4504, 0x4504, 0x6504, 0x95FC, 0x850, 0x850, 0x1092, 0x2092, 0x4112, 0x820E}},
    {27538, 2, {0x0, 0xF8, 0xFC88, 0x2088, 0x20F8, 0x3C00, 0x45FC, 0x4504, 0x6524, 0x9524, 0x924, 0x924, 0x1050, 0x2048, 0x4084, 0x8304}},
    {27539, 2, {0x20, 0x20, 0xFC50, 0x2050, 0x2088, 0x3D04, 0x46FA, 0x4400, 0x6444, 0x9424, 0x924, 0x8A8, 0x1088, 0x2010, 0x43FE, 0x8000}},
    {27540, 2, {0x20, 0x20, 0xF9FC, 0x2024, 0x2024, 0x3BFE, 0x4824, 0x4824, 0x69FC, 0x9A22, 0x974, 0x10A8, 0x1124, 0x2222, 0x40A0, 0x8040}},
    {27541, 2, {0x40, 0x20, 0xFDFC, 0x2000, 0x2108, 0x3C90, 0x4400, 0x47FE, 0x6400, 0x9400, 0x9FC, 0x904, 0x1104, 0x2104, 0x41FC, 0x8104}},
    {27542, 2, {0x20, 0xFC20, 0x23FE, 0x2020, 0x2020, 0x3DFC, 0x4504, 0x45FC, 0x6504, 0x95FC, 0x904, 0x9FC, 0x1104, 0x2104, 0x47FE, 0x8000}},
    {27543, 2, {0x40, 0x80, 0xFBFC, 0x2110, 0x2248, 0x3C46, 0x4BF8, 0x4A48, 0x6A48, 0x9BF8, 0xA48, 0x1248, 0x13F8, 0x2042, 0x4042, 0x803E}},
    {27544, 2, {0x90, 0x88, 0xFCBC, 0x23C0, 0x2050, 0x3C24, 0x44D4, 0x472C, 0x6490, 0x94BC, 0xBC0, 0x848, 0x1050, 0x2024, 0x40D4, 0x830C}},
    {27545, 2, {0x1C, 0x3E0, 0xFA20, 0x23FE, 0x2220, 0x3A92, 0x4B0A, 0x4A06, 0x69FC, 0x9904, 0x904, 0x11FC, 0x1104, 0x2104, 0x41FC, 0x8104}},
    {27546, 2, {0x104, 0x88, 0xFC50, 0x21FC, 0x2124, 0x3D24, 0x45FC, 0x4524, 0x6524, 0x95FC, 0x820, 0x820, 0x13FE, 0x2020, 0x4020, 0x8020}},
    {27547, 2, {0x0, 0x3FC, 0xF808, 0x2010, 0x2020, 0x3BAE, 0x4AA2, 0x4AAA, 0x6AA4, 0x9AA4, 0xBAA, 0x1032, 0x10A0, 0x2040, 0x47FE, 0x8000}},
    {27548, 2, {0x48, 0x148, 0xFD48, 0x23FE, 0x2148, 0x3D48, 0x4578, 0x4500, 0x65FE, 0x9420, 0xBFE, 0x870, 0x10A8, 0x2124, 0x4622, 0x8020}},
    {27549, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x3840, 0x4BFE, 0x4888, 0x6934, 0x9AE2, 0x820, 0x11FC, 0x1070, 0x20AC, 0x4322, 0x8020}},
    {27550, 2, {0x0, 0x1F8, 0xF908, 0x2108, 0x21F8, 0x3800, 0x4BFC, 0x4A04, 0x6BFC, 0x9A04, 0xBFC, 0x1204, 0x13FC, 0x2090, 0x4108, 0x8204}},
    {27551, 2, {0x0, 0x1F8, 0xFD08, 0x2108, 0x21F8, 0x3D08, 0x4508, 0x45F8, 0x6400, 0x97FC, 0xA94, 0xA94, 0x1294, 0x2294, 0x47FE, 0x8000}},
    {27552, 2, {0x20, 0x40, 0xFDFC, 0x2104, 0x21FC, 0x3D04, 0x45FC, 0x4504, 0x65FC, 0x9428, 0x824, 0xBFE, 0x1050, 0x2088, 0x4104, 0x8602}},
    {27553, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2414, 0x39E0, 0x4900, 0x4900, 0x69FC, 0x9910, 0x910, 0x17FE, 0x1000, 0x2090, 0x4108, 0x8204}},
    {27554, 2, {0x154, 0x154, 0xFBFE, 0x2154, 0x2154, 0x3A72, 0x4C00, 0x4BFE, 0x6A22, 0x9820, 0x9FC, 0x1124, 0x1124, 0x2124, 0x412C, 0x8020}},
    {27555, 2, {0x88, 0xFC88, 0x23FE, 0x2088, 0x20F8, 0x3C20, 0x45FC, 0x4524, 0x65FC, 0x9420, 0xBFE, 0x820, 0x11FC, 0x2020, 0x43FE, 0x8000}},
    {27556, 2, {0x80, 0xFE, 0xFD00, 0x22FC, 0x2084, 0x3CFC, 0x4484, 0x44FC, 0x6400, 0x97FE, 0x880, 0x9FE, 0x1252, 0x2092, 0x412A, 0x8044}},
    {27557, 2, {0x40, 0xFC20, 0x23FE, 0x2202, 0x2000, 0x3DFC, 0x4420, 0x45FC, 0x6524, 0x95FC, 0x924, 0x9FC, 0x1000, 0x2088, 0x4104, 0x8202}},
    {27558, 2, {0x40, 0x80, 0xFDFC, 0x2104, 0x21FC, 0x3D04, 0x45FC, 0x4500, 0x65FE, 0x9500, 0x9FE, 0x802, 0x12AA, 0x22AA, 0x4202, 0x800C}},
    {27559, 2, {0x214, 0x112, 0xFBD2, 0x2010, 0x203E, 0x3BD0, 0x4A50, 0x4A50, 0x6BD0, 0x9928, 0xBA8, 0x1368, 0x1528, 0x212A, 0x452A, 0x8246}},
    {27560, 2, {0x20, 0x1FC, 0xFD24, 0x21FC, 0x2020, 0x3FFE, 0x4400, 0x45FC, 0x6504, 0x95FC, 0x904, 0x9FC, 0x1104, 0x21FC, 0x4088, 0x8104}},
    {27561, 2, {0x80, 0xDC, 0xFA84, 0x23D4, 0x2448, 0x3A94, 0x4B24, 0x4C40, 0x6A48, 0x9950, 0xFFC, 0x10E0, 0x1150, 0x2248, 0x4C46, 0x8040}},
    {27562, 2, {0x20, 0x3FE, 0xF820, 0x21FC, 0x2000, 0x3BFE, 0x4A02, 0x49FC, 0x6800, 0x99FC, 0x904, 0x11FC, 0x1104, 0x2088, 0x43FE, 0x8000}},
    {27563, 2, {0x0, 0x3DE, 0xFA52, 0x2252, 0x23DE, 0x3800, 0x49FC, 0x4924, 0x69FC, 0x9924, 0x9FC, 0x1020, 0x13FE, 0x2020, 0x4020, 0x8020}},
    {27564, 2, {0x0, 0x1FC, 0xFD54, 0x21FC, 0x2020, 0x3DFC, 0x4420, 0x47FE, 0x6488, 0x9450, 0x9FC, 0x820, 0x13FE, 0x2020, 0x4020, 0x8020}},
    {27565, 2, {0x0, 0x3FE, 0xFC00, 0x21FC, 0x2124, 0x3DFC, 0x4524, 0x47FE, 0x6400, 0x95FC, 0x924, 0x9FC, 0x1124, 0x21FC, 0x4000, 0x83FE}},
    {27566, 2, {0x20, 0x20, 0xFC50, 0x2088, 0x2104, 0x3EFA, 0x4400, 0x4400, 0x65DC, 0x9554, 0x954, 0x9DC, 0x1088, 0x2088, 0x4154, 0x8222}},
    {27567, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x21FC, 0x3948, 0x4A50, 0x49FC, 0x6B04, 0x99FC, 0x904, 0x11FC, 0x1104, 0x21FC, 0x4088, 0x8104}},
    {27568, 2, {0x20, 0x3FE, 0xF820, 0x21FC, 0x2000, 0x3BFE, 0x4A52, 0x4BFE, 0x6904, 0x99FC, 0x904, 0x11FC, 0x1104, 0x21FC, 0x4088, 0x8104}},
    {27569, 2, {0x88, 0x8C, 0xFBEA, 0x2088, 0x2088, 0x3BFE, 0x4948, 0x4B6A, 0x694A, 0x9B6A, 0x94C, 0x136C, 0x114A, 0x216A, 0x4796, 0x8022}},
    {27570, 2, {0x148, 0x14C, 0xFAAA, 0x2008, 0x27FE, 0x3948, 0x4B68, 0x494A, 0x6B6A, 0x994C, 0xB6C, 0x1148, 0x116A, 0x239A, 0x4026, 0x8042}},
    {27571, 2, {0x0, 0xFE0, 0x820, 0x820, 0x820, 0x1020, 0x201E, 0xC000, 0x3FF0, 0x1010, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC006}},
    {27572, 2, {0x0, 0x7EF8, 0x4088, 0x4488, 0x6488, 0x5486, 0x4900, 0x4800, 0x55FC, 0x5484, 0x6488, 0x4050, 0x4020, 0x7E50, 0x88, 0x306}},
    {27573, 2, {0x600, 0x38F8, 0x2088, 0x2088, 0x2088, 0x3C86, 0x2100, 0x2000, 0x3DFC, 0x2084, 0x2088, 0x2C50, 0xF020, 0x2050, 0x2088, 0x2306}},
    {27574, 2, {0x1000, 0x878, 0x48, 0x7F48, 0x848, 0x886, 0x800, 0x7EFC, 0x844, 0x844, 0x828, 0x828, 0xF10, 0xF028, 0x4044, 0x182}},
    {27575, 2, {0x600, 0x3878, 0x2048, 0x3E48, 0x2248, 0x3E86, 0x2200, 0x3EFC, 0x2044, 0x3E44, 0x2228, 0x2228, 0x4210, 0x4A28, 0x8444, 0x182}},
    {27576, 2, {0x800, 0x878, 0x7F48, 0x848, 0x848, 0x3E86, 0x0, 0x3EFC, 0x2A44, 0x2A44, 0x3E28, 0x2028, 0x2010, 0x4028, 0x4044, 0x8182}},
    {27577, 2, {0x0, 0x7F78, 0x4848, 0x4848, 0x5F48, 0x6486, 0x4400, 0x44FC, 0x7F44, 0x4444, 0x4A28, 0x5128, 0x6010, 0x4028, 0x7E44, 0x182}},
    {27578, 2, {0x400, 0x4478, 0x2848, 0x1048, 0x2848, 0x4486, 0x9200, 0x10FC, 0xFE44, 0x1044, 0x5828, 0x5428, 0x9210, 0x1228, 0x5044, 0x2182}},
    {27579, 2, {0x800, 0x878, 0x7F48, 0x848, 0x3E48, 0x86, 0x7F00, 0x41FC, 0x8244, 0x3C44, 0x2428, 0x2428, 0x2510, 0x2628, 0x4444, 0x8182}},
    {27580, 2, {0x800, 0x878, 0x7F48, 0x848, 0x3E48, 0x86, 0x7F00, 0x41FC, 0xBE44, 0x44, 0x3C28, 0x2428, 0x2510, 0x2628, 0x4444, 0x8182}},
    {27581, 2, {0x4200, 0x2478, 0x1848, 0x2448, 0x5248, 0x1086, 0xFF00, 0x20FC, 0x3E44, 0x6244, 0xBE28, 0x2228, 0x3E10, 0x2228, 0x2A44, 0x2582}},
    {27582, 2, {0x1000, 0x1078, 0x7D48, 0x1248, 0x1448, 0xFF86, 0x1000, 0x20FC, 0x7E44, 0xA244, 0x2228, 0x3E28, 0x2210, 0x2228, 0x3E44, 0x2382}},
    {27583, 2, {0x0, 0x3F78, 0x2148, 0x2148, 0x3F48, 0x2086, 0x2A00, 0x2AFC, 0x3F44, 0x2A44, 0x2A28, 0x3F28, 0x2010, 0x4A28, 0x9144, 0x2182}},
    {27584, 2, {0x2000, 0xCE78, 0x8248, 0x8248, 0xEE48, 0x8286, 0x8200, 0xFEFC, 0x1044, 0x1044, 0xFE28, 0x1028, 0x1010, 0x1E28, 0xF044, 0x4182}},
    {27585, 2, {0x2000, 0xCE78, 0x8248, 0x8248, 0xEE48, 0x8286, 0x8200, 0xFEFC, 0x44, 0x44, 0xFE28, 0x1028, 0x1010, 0x1E28, 0xF044, 0x4182}},
    {27586, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x10FC, 0x7C44, 0x2044, 0x5028, 0x7C28, 0x1010, 0xFE28, 0x1044, 0x1182}},
    {27587, 2, {0x2000, 0x1078, 0xFE48, 0x48, 0x7C48, 0x4486, 0x7C00, 0xFC, 0xFE44, 0x8244, 0xBA28, 0xAA28, 0xBA10, 0x8228, 0x8A44, 0x8582}},
    {27588, 2, {0x1000, 0x7E78, 0x1048, 0x7E48, 0x5248, 0x7E86, 0x5200, 0x7EFC, 0x1044, 0xFF44, 0x1028, 0x7E28, 0x4210, 0x4228, 0x7E44, 0x4382}},
    {27589, 2, {0x1000, 0x878, 0x7F48, 0x2248, 0x1448, 0xFF86, 0x1000, 0x29FC, 0xCE44, 0x1444, 0x2628, 0xCD28, 0x1510, 0x2428, 0xD444, 0x982}},
    {27590, 2, {0x0, 0xFFB8, 0x8028, 0x9F28, 0x9128, 0x9146, 0x9F00, 0x807C, 0xBBA4, 0xAAA4, 0xAAA8, 0xAAA8, 0xBB90, 0x8028, 0xFFC4, 0x82}},
    {27591, 2, {0x1000, 0x6738, 0x4128, 0x7728, 0x4128, 0x4146, 0x7F00, 0x87C, 0x4924, 0x2A24, 0xFFA8, 0x1C28, 0x2A10, 0x4928, 0x8844, 0x882}},
    {27592, 2, {0x2040, 0xCE9C, 0x8A94, 0xAA94, 0xAAD4, 0xEE94, 0xAAA6, 0xAAC0, 0xAA9C, 0xAA94, 0xEE94, 0xA8D4, 0x2B88, 0x4888, 0x4894, 0x88A2}},
    {27593, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0xFFFE, 0x0}},
    {27594, 2, {0x878, 0xFF48, 0x886, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x9FF4, 0x200, 0xFFFE, 0x820, 0x3FF8, 0xD016, 0x1390, 0x1290, 0x13B0}},
    {27595, 2, {0x0, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x2110, 0x2210, 0x2210, 0x2210, 0x3FFC, 0x410, 0x810, 0x10A0, 0x2040}},
    {27596, 2, {0x0, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x2210, 0x2210, 0x2210, 0x2210, 0x2210, 0x3FF0, 0x10, 0x10}},
    {27597, 2, {0x0, 0x1FF0, 0x1010, 0x1210, 0x1110, 0x1110, 0xFFFE, 0x2010, 0x2210, 0x2110, 0x2110, 0x3FFC, 0x10, 0x10, 0xA0, 0x40}},
    {27598, 2, {0x1000, 0x1FFC, 0x2000, 0x2000, 0x5FF0, 0x9110, 0x1110, 0x1110, 0xFFFE, 0x2210, 0x2210, 0x2210, 0x3FFC, 0x10, 0xA0, 0x40}},
    {27599, 2, {0x1000, 0x1FFC, 0x2000, 0x2000, 0x5FF0, 0x9010, 0x1210, 0x1110, 0xFFFE, 0x2010, 0x2210, 0x2110, 0x3FFC, 0x10, 0xA0, 0x40}},
    {27600, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x1FF0, 0x1110, 0x1110, 0xFFFE, 0x2110, 0x2210, 0x3FFC, 0x410, 0x850, 0x1020}},
    {27601, 2, {0x10, 0x3E50, 0x2250, 0x325C, 0x2A74, 0x22D4, 0x2254, 0xFF54, 0x4254, 0x525C, 0x4A50, 0x4250, 0x7F42, 0x242, 0x143E, 0x800}},
    {27602, 2, {0x100, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1210, 0x1110, 0xFFFE, 0x2210, 0x2110, 0x3FFC, 0x10, 0x60}},
    {27603, 2, {0x4020, 0x4010, 0x7DFE, 0x8020, 0x7C48, 0x4484, 0x65FE, 0x5402, 0xFEA8, 0x44A8, 0xA4A8, 0x94A8, 0xFEAA, 0x52A, 0x2926, 0x1200}},
    {27604, 2, {0x80, 0x2080, 0x2080, 0x2084, 0x2088, 0x2090, 0x3EA0, 0x20C0, 0x2080, 0x2080, 0x2080, 0x2082, 0x2682, 0x3882, 0x207E, 0x0}},
    {27605, 2, {0x2080, 0x2088, 0x20B0, 0x3EC0, 0x2080, 0x2084, 0x2684, 0x387C, 0x2100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {27606, 2, {0x2080, 0x2088, 0x20B0, 0x3EC0, 0x2080, 0x2084, 0x2684, 0x387C, 0x100, 0x90, 0x48A4, 0x4842, 0x8982, 0xE10, 0x1810, 0xE7F0}},
    {27607, 2, {0x10, 0x110, 0x7D10, 0x5512, 0x5512, 0x5514, 0x55D8, 0x7D10, 0x5510, 0x5510, 0x5510, 0x5512, 0x7D52, 0x4592, 0x10E, 0x0}},
    {27608, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x2080, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {27609, 2, {0x4100, 0x4118, 0x7DE0, 0x4104, 0x4D04, 0x70FC, 0x0, 0xFFFE, 0x1080, 0x1088, 0x3E90, 0x42E0, 0x1480, 0x884, 0x3084, 0xC07C}},
    {27610, 2, {0x1FC0, 0x2080, 0x7FE0, 0xA020, 0x3FE0, 0x2088, 0x3CF4, 0x2084, 0x1FFC, 0x2040, 0x7FF0, 0xA210, 0x3FF0, 0x544, 0x1924, 0xE0FC}},
    {27611, 2, {0x10, 0x1F8, 0x7E00, 0x200, 0x200, 0x3F0, 0x3E00, 0x200, 0x200, 0x3FC, 0xFE00, 0x200, 0x204, 0x204, 0x204, 0x1FC}},
    {27612, 2, {0x820, 0x1C20, 0x7020, 0x1020, 0x10A8, 0x1CA4, 0x7124, 0x1122, 0x1222, 0x1C20, 0xF020, 0x10A2, 0x1042, 0x1002, 0xFFE, 0x0}},
    {27613, 2, {0x804, 0x1C08, 0x7030, 0x11C0, 0x1004, 0x1C08, 0x7030, 0x11C0, 0x1004, 0x1C08, 0xF030, 0x11C2, 0x1002, 0x1002, 0xFFE, 0x0}},
    {27614, 2, {0x2080, 0x2088, 0x20B0, 0x3EC0, 0x2084, 0x2084, 0x267C, 0x3800, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {27615, 2, {0x100, 0x1110, 0x1108, 0x2124, 0x4144, 0x180, 0xE00, 0xF000, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {27616, 2, {0x800, 0x800, 0x7F7C, 0x944, 0x1144, 0x1144, 0x257C, 0x4200, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {27617, 2, {0x840, 0x1C40, 0x7040, 0x107C, 0x1040, 0x1C40, 0x7040, 0x11F8, 0x1108, 0x1D08, 0xF108, 0x11FA, 0x110A, 0x1002, 0xFFE, 0x0}},
    {27618, 2, {0x1000, 0x3BF8, 0xE0A0, 0x20A0, 0x23F8, 0x3AA8, 0xE2A8, 0x22A8, 0x22B8, 0x3B08, 0xE208, 0x23F8, 0x220A, 0x2002, 0x2002, 0x1FFE}},
    {27619, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x60C, 0x1A34, 0x62C4, 0x0, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {27620, 2, {0x2818, 0x24E0, 0x4220, 0x9238, 0x10E0, 0x2020, 0x443C, 0xFEE0, 0x222, 0x2422, 0x241E, 0x2400, 0x2402, 0x4402, 0x43FE, 0x8000}},
    {27621, 2, {0x1100, 0x3900, 0xE1FC, 0x2204, 0x2404, 0x3BE4, 0xE224, 0x23E4, 0x2224, 0x3A24, 0xE3E4, 0x2004, 0x202A, 0x2012, 0x2002, 0x1FFE}},
    {27622, 2, {0x8, 0xFF1C, 0x44E0, 0x4420, 0x4420, 0x7C3C, 0x44E0, 0x4420, 0x7C20, 0x443C, 0x44E0, 0x4F22, 0xF422, 0x422, 0x41E, 0x400}},
    {27623, 2, {0x1050, 0x3848, 0xE040, 0x2FFC, 0x2040, 0x3A48, 0xE248, 0x2F50, 0x2250, 0x3A20, 0xE228, 0x2468, 0x289A, 0x210A, 0x2002, 0x1FFE}},
    {27624, 2, {0x820, 0x1CA0, 0x70FC, 0x1120, 0x1020, 0x1DFE, 0x7050, 0x1050, 0x1090, 0x1C92, 0xF112, 0x120E, 0x1000, 0x1002, 0xFFE, 0x0}},
    {27625, 2, {0x820, 0x1C20, 0x7124, 0x10A4, 0x10A8, 0x1C20, 0x73FE, 0x1020, 0x1070, 0x1CA8, 0xF124, 0x1222, 0x1020, 0x1002, 0xFFE, 0x0}},
    {27626, 2, {0x1040, 0x3840, 0xE090, 0x2108, 0x23FC, 0x3824, 0xE120, 0x21FC, 0x2120, 0x3A20, 0xE7FE, 0x2020, 0x2022, 0x2022, 0x2022, 0x1FFE}},
    {27627, 2, {0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x81F4, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE}},
    {27628, 2, {0x820, 0x1C28, 0x7024, 0x13FE, 0x1020, 0x1D24, 0x70A8, 0x1070, 0x10A8, 0x1D24, 0xF222, 0x10A0, 0x1040, 0x1002, 0xFFE, 0x0}},
    {27629, 2, {0x8, 0xFF1C, 0xE0, 0x20, 0x7E20, 0x423C, 0x42E0, 0x4220, 0x7E20, 0x3C, 0x42E0, 0x2222, 0x2422, 0xF22, 0xF01E, 0x4000}},
    {27630, 2, {0x1020, 0x1020, 0x1124, 0x1122, 0xFA2A, 0x1010, 0x1460, 0x1B80, 0x303C, 0xD3C0, 0x107C, 0x13C0, 0x107C, 0x17C2, 0x5042, 0x203E}},
    {27631, 2, {0x1040, 0x3A44, 0xE244, 0x2448, 0x20A0, 0x3910, 0xE608, 0x2044, 0x2040, 0x3A48, 0xE248, 0x2450, 0x20A2, 0x2112, 0x220A, 0x1FFE}},
    {27632, 2, {0x840, 0x1C20, 0x71FC, 0x1000, 0x1088, 0x1C50, 0x73FE, 0x1000, 0x10F8, 0x1C88, 0xF088, 0x108A, 0x10FA, 0x1002, 0xFFE, 0x0}},
    {27633, 2, {0x1100, 0x3900, 0xE1FC, 0x2204, 0x2484, 0x3C94, 0xE2A4, 0x27FC, 0x2084, 0x39C4, 0xE2A4, 0x2494, 0x20AA, 0x2092, 0x2002, 0x1FFE}},
    {27634, 2, {0x4, 0xF70E, 0x1170, 0x5510, 0x2210, 0x551C, 0x8970, 0x1010, 0x10, 0xF71E, 0x1170, 0x5510, 0x2212, 0x5512, 0x8912, 0x100E}},
    {27635, 2, {0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE, 0x0, 0x1830, 0x71C0, 0x1C78, 0x71C0, 0x1C78, 0x71C4, 0x1444, 0x183C}},
    {27636, 2, {0x1090, 0x3890, 0xE090, 0x279E, 0x2090, 0x3890, 0xE39C, 0x2090, 0x2090, 0x3890, 0xE79E, 0x2090, 0x2092, 0x2092, 0x2002, 0x1FFE}},
    {27637, 2, {0x808, 0x101C, 0x24E0, 0x7E20, 0x1020, 0xFF3C, 0x24E0, 0x4220, 0x8920, 0x303C, 0xC4E0, 0x822, 0x3222, 0xC422, 0x381E, 0xC000}},
    {27638, 2, {0x840, 0x1C88, 0x71FC, 0x1024, 0x13FE, 0x1C88, 0x7134, 0x12C2, 0x1018, 0x1CE0, 0xF00C, 0x1030, 0x11C0, 0x1002, 0xFFE, 0x0}},
    {27639, 2, {0x4, 0x7F0E, 0x4170, 0x5D10, 0x4110, 0x5D1C, 0x4170, 0x10, 0x3E10, 0x221E, 0x3E70, 0x2210, 0x3E12, 0x2212, 0x3E12, 0x220E}},
    {27640, 2, {0x800, 0x1DFC, 0x7124, 0x11FC, 0x1124, 0x1DFC, 0x7020, 0x1014, 0x1152, 0x1D42, 0xF248, 0x103A, 0x1002, 0x1002, 0xFFE, 0x0}},
    {27641, 2, {0x804, 0x140E, 0x2270, 0x4110, 0xBE90, 0x1C, 0x78F0, 0x4A90, 0x4A90, 0x7A9E, 0x4AF0, 0x4A90, 0x7A92, 0x4892, 0x4A92, 0x590E}},
    {27642, 2, {0x1020, 0x3850, 0xE088, 0x2104, 0x26FA, 0x3800, 0xE3C4, 0x2254, 0x23D4, 0x3A54, 0xE3D4, 0x2244, 0x2256, 0x22CA, 0x2002, 0x1FFE}},
    {27643, 2, {0x1008, 0x101C, 0xFEE0, 0x2020, 0x7E20, 0x883C, 0x7EE0, 0x20, 0x7E20, 0x423C, 0x7EE0, 0x4222, 0x7E22, 0x4222, 0x4A1E, 0x4400}},
    {27644, 2, {0x8, 0x7E1C, 0x42E0, 0x7E20, 0x4220, 0x7E3C, 0x20E0, 0x7F20, 0x9120, 0x513C, 0x69E0, 0x4122, 0x7D22, 0x122, 0xA1E, 0x400}},
    {27645, 2, {0x1010, 0x3810, 0xE77C, 0x2114, 0x22FE, 0x3A14, 0xE77C, 0x2110, 0x217C, 0x3910, 0xE6FE, 0x2210, 0x2512, 0x28FE, 0x2002, 0x1FFE}},
    {27646, 2, {0x4, 0x7F0E, 0x4170, 0x7F10, 0x4110, 0x7F1C, 0x70, 0xF790, 0x1090, 0x949E, 0x52F0, 0x3190, 0xD692, 0x1092, 0x5292, 0x210E}},
    {27647, 2, {0x1004, 0x220E, 0x7F70, 0x4210, 0x9490, 0xF79C, 0x870, 0x3610, 0xC910, 0x309E, 0xC470, 0x1910, 0x6212, 0xC12, 0x3012, 0xC00E}},
    {27648, 2, {0x804, 0x7F0E, 0x4970, 0xFF90, 0x4910, 0x7F1C, 0x870, 0x7F10, 0x4910, 0x7F1E, 0x1070, 0xFF10, 0x2112, 0x7212, 0xC12, 0xF30E}},
    {27649, 2, {0x1110, 0x3910, 0xE7FC, 0x2110, 0x23F8, 0x3A08, 0xE3F8, 0x2208, 0x23F8, 0x3840, 0xE7FC, 0x20A0, 0x2112, 0x220A, 0x2002, 0x1FFE}},
    {27650, 2, {0x820, 0x3E7E, 0x8A4, 0x7F28, 0x1C10, 0x2A28, 0x4944, 0x3FFE, 0x2000, 0x2078, 0x2F80, 0x20F0, 0x2780, 0x40FA, 0x4F82, 0x807E}},
    {27651, 2, {0x1004, 0x80E, 0x7F70, 0x2210, 0xFF10, 0x1C, 0x7F70, 0x4910, 0x7F10, 0x491E, 0x7F70, 0x810, 0x7F12, 0x812, 0xF12, 0xF00E}},
    {27652, 2, {0x7E04, 0x140E, 0x870, 0xFF10, 0x2910, 0x4A1C, 0xA870, 0x1010, 0x7F10, 0x551E, 0x6370, 0x5D10, 0x5512, 0x5D12, 0x4112, 0x430E}},
    {27653, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x41A8, 0x5D28, 0x5510, 0x5D28, 0x42F6, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {27654, 2, {0x1108, 0x3890, 0xE3FC, 0x2090, 0x2294, 0x3998, 0xE090, 0x27FE, 0x2000, 0x39F8, 0xE108, 0x21F8, 0x210A, 0x21FA, 0x2002, 0x1FFE}},
    {27655, 2, {0x1100, 0x3BF0, 0xE410, 0x2BF8, 0x2248, 0x3BF8, 0xE248, 0x27FC, 0x2000, 0x3BF8, 0xE208, 0x23F8, 0x220A, 0x23FA, 0x2002, 0x1FFE}},
    {27656, 2, {0x1008, 0xFE1C, 0xE0, 0xFE20, 0x8220, 0xBA3C, 0xAAE0, 0xFE20, 0x20, 0x7C3C, 0x44E0, 0x7C22, 0x4422, 0x7C22, 0x1E, 0xFE00}},
    {27657, 2, {0x4, 0x7F0E, 0x4170, 0x7F10, 0x10, 0xF79C, 0x94F0, 0xF790, 0x810, 0x81E, 0xFF70, 0x1C10, 0x2A12, 0x4912, 0x8892, 0x80E}},
    {27658, 2, {0x1020, 0x3BFE, 0xE000, 0x23FC, 0x2204, 0x3AF4, 0xE294, 0x23FC, 0x2108, 0x39F8, 0xE108, 0x21F8, 0x2002, 0x27FE, 0x2002, 0x1FFE}},
    {27659, 2, {0x2404, 0xFF0E, 0x2470, 0xFF10, 0x8110, 0x7E1C, 0x70, 0xFF10, 0x1010, 0x289E, 0xD570, 0x2E10, 0xD512, 0x2492, 0xD412, 0x80E}},
    {27660, 2, {0x1080, 0x39F8, 0xE208, 0x25FC, 0x2124, 0x39FC, 0xE124, 0x21FC, 0x2154, 0x3AFA, 0xE088, 0x20F8, 0x208A, 0x20FA, 0x2002, 0x1FFE}},
    {27661, 2, {0x7704, 0x550E, 0x7770, 0x5510, 0x7710, 0x551C, 0x7770, 0x2210, 0x7F90, 0xA41E, 0x3F70, 0x2410, 0x3F12, 0x2412, 0x3F92, 0x200E}},
    {27662, 2, {0x3F84, 0x248E, 0x3FF0, 0x2490, 0x7FD0, 0x555C, 0x7FF0, 0x5550, 0x7FD0, 0x41E, 0x7FF0, 0x4850, 0x1F12, 0x2A12, 0x412, 0x7FCE}},
    {27663, 2, {0x10, 0xF8, 0x3F80, 0x2080, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x2080, 0x2040, 0x2040, 0x2022, 0x2012, 0x280A, 0x3006, 0x2002}},
    {27664, 2, {0x10, 0xF8, 0x3F80, 0x2080, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x2080, 0x2040, 0x2040, 0x2022, 0x2112, 0x288A, 0x3046, 0x2042}},
    {27665, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x3FFC, 0x2080, 0x2040, 0x2040, 0x2424, 0x2814, 0x300C, 0x2004}},
    {27666, 2, {0xF0, 0x3F00, 0x2100, 0x2100, 0x3FFC, 0x2080, 0x2040, 0x2824, 0x3114, 0x210C, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {27667, 2, {0x2000, 0x13FC, 0x1204, 0x204, 0xFA04, 0x43FC, 0x4220, 0x4220, 0x43FE, 0x4220, 0x4220, 0x4210, 0x7A12, 0x28A, 0x306, 0x202}},
    {27668, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x10, 0x10, 0x10, 0x10, 0xA, 0xA, 0x6, 0x2}},
    {27669, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x410, 0x410, 0x410, 0x410, 0x80A, 0x80A, 0x1006, 0x2002}},
    {27670, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x3F10, 0x1110, 0x1210, 0x17D0, 0x204A, 0x204A, 0x4286, 0x8102}},
    {27671, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x90, 0x3090, 0xD10, 0x210, 0x50A, 0x188A, 0x6046, 0x2}},
    {27672, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x10, 0x1110, 0x1110, 0x1110, 0x210A, 0x210A, 0x4106, 0x8102}},
    {27673, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x410, 0x410, 0x4450, 0x4450, 0x444A, 0x7FCA, 0x46, 0x2}},
    {27674, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x10, 0x2490, 0x2490, 0x2490, 0x2490, 0x248A, 0x448A, 0x4486, 0x8082}},
    {27675, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x1210, 0x2110, 0x4090, 0xBF50, 0x1112, 0x110A, 0x110A, 0x2506, 0x4202}},
    {27676, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x3F90, 0x2090, 0x2090, 0x3F90, 0x2092, 0x208A, 0x208A, 0x3F86, 0x2082}},
    {27677, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x410, 0x410, 0x7FD0, 0x4450, 0x4A50, 0x514A, 0x604A, 0x4146, 0x4082}},
    {27678, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x7FD0, 0x410, 0x7FD0, 0x4450, 0x4A52, 0x514A, 0x604A, 0x4146, 0x4082}},
    {27679, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x1210, 0x7F90, 0x1290, 0x7F90, 0x5210, 0x7FD2, 0x124A, 0x134A, 0x2286, 0x4202}},
    {27680, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x410, 0x7FD0, 0x4450, 0x7FD0, 0x4452, 0x7FCA, 0x444A, 0x406, 0x402}},
    {27681, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x1010, 0x1F90, 0x2090, 0x5110, 0xE10, 0x3192, 0xCC6A, 0x30A, 0x1C06, 0x302}},
    {27682, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x3FD0, 0x310, 0xCD0, 0x7030, 0x1FD2, 0x20A, 0x20A, 0x7FF6, 0x2}},
    {27683, 2, {0x1000, 0x1000, 0x3FFC, 0x2000, 0x4FF0, 0x8000, 0x3FF0, 0x410, 0x2490, 0x1510, 0x7FD0, 0xE10, 0x150A, 0x248A, 0x4446, 0x402}},
    {27684, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x7FD0, 0x4450, 0x4450, 0x5F50, 0x4452, 0x4A4A, 0x514A, 0x7FC6, 0x4042}},
    {27685, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x7FD0, 0xA10, 0x3F90, 0x2A90, 0x2A92, 0x338A, 0x208A, 0x3F86, 0x2082}},
    {27686, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x7FD0, 0x810, 0x1110, 0x3E10, 0x492, 0x190A, 0x620A, 0xD06, 0x3082}},
    {27687, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x1110, 0xA10, 0x7FD0, 0x410, 0x3F90, 0x412, 0xFFEA, 0x40A, 0x406, 0x402}},
    {27688, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x7FD0, 0x4050, 0x810, 0xFFD0, 0x1112, 0x320A, 0xC0A, 0x1B06, 0x6082}},
    {27689, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x7F90, 0x1210, 0x1250, 0x9290, 0x5312, 0x520A, 0x120A, 0xFFC6, 0x2}},
    {27690, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x7FD0, 0x410, 0x3F90, 0x2090, 0x3F92, 0xA0A, 0x124A, 0x2246, 0x41C2}},
    {27691, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x7FD0, 0x2490, 0x4910, 0x2490, 0x7FD2, 0x40A, 0x40A, 0xFFE6, 0x2}},
    {27692, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0xFFD0, 0x1210, 0x7390, 0x4090, 0x4092, 0x738A, 0x120A, 0x1206, 0xFFE2}},
    {27693, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x7FD0, 0x410, 0x3F90, 0x2490, 0x3F92, 0x248A, 0x3F8A, 0x1506, 0x2482}},
    {27694, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x2490, 0xB10, 0x1090, 0x6050, 0x412, 0x248A, 0xB0A, 0x1086, 0x6042}},
    {27695, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x7F90, 0x90, 0x3F90, 0x90, 0xFFF2, 0x248A, 0x150A, 0x2486, 0x4C42}},
    {27696, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x410, 0x7FD0, 0x410, 0x3F90, 0x410, 0xFFF2, 0x208A, 0x3F8A, 0x2086, 0x2182}},
    {27697, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x1090, 0x1F90, 0x1090, 0x3FF0, 0x1010, 0x3FD2, 0x4A4A, 0x124A, 0x2546, 0x882}},
    {27698, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x3F90, 0x2090, 0x3F90, 0x2090, 0x3F92, 0xA, 0x7FCA, 0x4A46, 0xFFE2}},
    {27699, 2, {0x2000, 0x3FFC, 0x4000, 0x9FF0, 0x0, 0x7FF0, 0x10, 0x3F90, 0x2490, 0x2A90, 0x3F90, 0x12, 0x7FCA, 0x4A4A, 0x4A46, 0xFFE2}},
    {27700, 2, {0x100, 0x100, 0x100, 0x108, 0x108, 0x7D90, 0x5A0, 0x940, 0x940, 0x1120, 0x1110, 0x2108, 0x4106, 0x8100, 0x500, 0x200}},
    {27701, 2, {0x0, 0x2000, 0x1000, 0x1000, 0x8000, 0x4000, 0x4800, 0x800, 0x1000, 0x1000, 0xE000, 0x2000, 0x2000, 0x2000, 0x2000, 0x0}},
    {27702, 2, {0x0, 0x3FF0, 0x20, 0x40, 0x84, 0x108, 0x7D90, 0x5A0, 0x540, 0x940, 0x920, 0x1110, 0x2108, 0xC106, 0x500, 0x200}},
    {27703, 2, {0x100, 0x100, 0x1100, 0x908, 0x908, 0x190, 0x7DA0, 0x540, 0x940, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x500, 0x200}},
    {27704, 2, {0x200, 0x100, 0x80, 0x1F00, 0x104, 0x108, 0x7D90, 0x5A0, 0x540, 0x940, 0x920, 0x1110, 0x2108, 0xC106, 0x500, 0x200}},
    {27705, 2, {0x40, 0x40, 0xF840, 0x844, 0xF48, 0x950, 0x1160, 0x1250, 0x2248, 0x2444, 0x4844, 0x4140, 0x4082, 0x4002, 0x3FFE, 0x0}},
    {27706, 2, {0x100, 0x100, 0x100, 0x4108, 0x2108, 0x1110, 0x1120, 0x180, 0x540, 0x920, 0x1110, 0xE108, 0x4104, 0x100, 0x500, 0x200}},
    {27707, 2, {0x80, 0x2080, 0x1080, 0x1080, 0x87FC, 0x4084, 0x4884, 0x884, 0x1084, 0x1084, 0xE104, 0x2104, 0x2104, 0x2204, 0x2228, 0x410}},
    {27708, 2, {0x100, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200, 0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {27709, 2, {0x100, 0x280, 0x440, 0x820, 0x3118, 0xC106, 0x100, 0x3D84, 0x588, 0x950, 0x920, 0x1110, 0x2108, 0xC106, 0x500, 0x200}},
    {27710, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x4208, 0x4A08, 0xA08, 0x1250, 0x1220, 0xE200, 0x2202, 0x2202, 0x2202, 0x21FE, 0x0}},
    {27711, 2, {0x100, 0x2100, 0x1100, 0x1100, 0x87F0, 0x4110, 0x4910, 0x910, 0x1110, 0x1110, 0xE110, 0x2212, 0x2212, 0x2412, 0x280E, 0x1000}},
    {27712, 2, {0x0, 0x2000, 0x13FE, 0x1020, 0x8020, 0x4020, 0x4820, 0x820, 0x1020, 0x1020, 0xE020, 0x2020, 0x2020, 0x2020, 0x20A0, 0x40}},
    {27713, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8040, 0x4040, 0x47FE, 0x1040, 0x1040, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {27714, 2, {0x120, 0x110, 0x100, 0x7FFC, 0x100, 0x2108, 0x1110, 0x9A0, 0x340, 0x520, 0x910, 0x1108, 0x6106, 0x100, 0x500, 0x200}},
    {27715, 2, {0x0, 0x2020, 0x1120, 0x1120, 0x8120, 0x4120, 0x4920, 0x920, 0x1120, 0x1110, 0xE110, 0x2210, 0x2208, 0x2408, 0x2804, 0x1002}},
    {27716, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8040, 0x4040, 0x4840, 0x840, 0x10A0, 0x10A0, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {27717, 2, {0x0, 0x27FE, 0x1080, 0x1080, 0x8100, 0x4100, 0x4A00, 0xBF8, 0x1008, 0x1008, 0xE008, 0x2008, 0x2008, 0x2008, 0x2050, 0x20}},
    {27718, 2, {0x600, 0x100, 0x280, 0xC60, 0x3118, 0xC106, 0x100, 0x3D84, 0x588, 0x950, 0x920, 0x1110, 0x2108, 0xC106, 0x500, 0x200}},
    {27719, 2, {0x0, 0x23FE, 0x1200, 0x1200, 0x8200, 0x4200, 0x4A00, 0xA00, 0x1200, 0x1200, 0xE200, 0x2200, 0x2200, 0x2200, 0x23FE, 0x0}},
    {27720, 2, {0x0, 0x27FC, 0x1004, 0x1004, 0x8004, 0x4014, 0x4824, 0x844, 0x1084, 0x1104, 0xE604, 0x2004, 0x2004, 0x2004, 0x2028, 0x10}},
    {27721, 2, {0x0, 0x27F8, 0x1208, 0x1208, 0x8208, 0x4110, 0x4910, 0x910, 0x10A0, 0x10A0, 0xE040, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {27722, 2, {0x0, 0x27F8, 0x1208, 0x1288, 0x8248, 0x4150, 0x4910, 0x910, 0x10A0, 0x10A0, 0xE040, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {27723, 2, {0x80, 0x2080, 0x1100, 0x11FC, 0x8204, 0x4204, 0x5404, 0x1904, 0x1084, 0x2044, 0xE044, 0x2004, 0x2004, 0x2004, 0x2028, 0x10}},
    {27724, 2, {0x204, 0x2244, 0x1244, 0x1244, 0x8244, 0x4244, 0x4244, 0x1244, 0x1244, 0x2244, 0xE244, 0x2244, 0x2244, 0x2444, 0x2404, 0x804}},
    {27725, 2, {0x80, 0x2080, 0x1080, 0x1080, 0x87F0, 0x4090, 0x4890, 0xA90, 0x1190, 0x1090, 0xE150, 0x2152, 0x220A, 0x220A, 0x2406, 0x802}},
    {27726, 2, {0x0, 0x23F0, 0x1210, 0x1210, 0x8210, 0x4310, 0x5290, 0x1250, 0x1250, 0x2210, 0xE210, 0x2212, 0x2412, 0x2412, 0x280E, 0x1000}},
    {27727, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8040, 0x47FE, 0x4840, 0x840, 0x10A0, 0x10A0, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {27728, 2, {0x80, 0x2080, 0x1080, 0x10FC, 0x8104, 0x4104, 0x4284, 0x1448, 0x1028, 0x2010, 0xE010, 0x2020, 0x2040, 0x2080, 0x2100, 0x600}},
    {27729, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x8040, 0x4040, 0x4840, 0x87E, 0x17C0, 0x1040, 0xE040, 0x2042, 0x2042, 0x2042, 0x203E, 0x0}},
    {27730, 2, {0x40, 0x2020, 0x1020, 0x1000, 0x87FE, 0x4100, 0x4900, 0x900, 0x1100, 0x1100, 0xE100, 0x2100, 0x2100, 0x2100, 0x21FC, 0x0}},
    {27731, 2, {0x0, 0x23F8, 0x1008, 0x1010, 0x8020, 0x4040, 0x4840, 0xFFE, 0x1040, 0x1040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2140, 0x80}},
    {27732, 2, {0x80, 0x2080, 0x1100, 0x11FE, 0x8200, 0x4400, 0x41F8, 0x1008, 0x1010, 0x2060, 0xE080, 0x2100, 0x2202, 0x2202, 0x21FE, 0x0}},
    {27733, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8444, 0x4444, 0x4444, 0x1444, 0x1444, 0x2444, 0xE444, 0x2444, 0x2444, 0x27FC, 0x2004, 0x0}},
    {27734, 2, {0x100, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200}},
    {27735, 2, {0x0, 0x2000, 0x13FC, 0x1040, 0x8040, 0x4040, 0x4840, 0x840, 0x17FE, 0x1040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {27736, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x8040, 0x4040, 0x4840, 0xFFE, 0x1040, 0x1040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {27737, 2, {0x0, 0x21FC, 0x1020, 0x1020, 0x8020, 0x4020, 0x4820, 0xBFE, 0x1020, 0x1020, 0xE020, 0x2020, 0x2020, 0x2020, 0x20A0, 0x40}},
    {27738, 2, {0x0, 0x23F8, 0x1040, 0x1040, 0x8040, 0x47FE, 0x4880, 0x880, 0x1100, 0x11F8, 0xE008, 0x2008, 0x2008, 0x2008, 0x2050, 0x20}},
    {27739, 2, {0x0, 0x27F0, 0x1110, 0x1110, 0x8110, 0x4110, 0x4110, 0x17D0, 0x1110, 0x2110, 0xE110, 0x2112, 0x210A, 0x210A, 0x2106, 0x102}},
    {27740, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x4208, 0x4A08, 0xBF8, 0x1200, 0x1200, 0xE200, 0x2202, 0x2202, 0x2202, 0x21FE, 0x0}},
    {27741, 2, {0x80, 0x2080, 0x1080, 0x1080, 0x8FFE, 0x4108, 0x4108, 0x1208, 0x1210, 0x2610, 0xE1A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC04}},
    {27742, 2, {0x0, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200}},
    {27743, 2, {0x0, 0x2000, 0x17FC, 0x1040, 0x8040, 0x4040, 0x4840, 0x840, 0x1040, 0x1040, 0xE040, 0x2040, 0x2040, 0x2FFE, 0x2000, 0x0}},
    {27744, 2, {0x40, 0x2040, 0x1240, 0x1258, 0x8268, 0x42C8, 0x5348, 0x1E48, 0x1248, 0x2248, 0xE258, 0x2242, 0x2202, 0x2202, 0x21FE, 0x0}},
    {27745, 2, {0x0, 0x23F8, 0x1000, 0x1000, 0x8000, 0x47FE, 0x4080, 0x1080, 0x1100, 0x21F8, 0xE008, 0x2008, 0x2008, 0x2008, 0x2050, 0x20}},
    {27746, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8040, 0x4040, 0x47FC, 0x1040, 0x1040, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {27747, 2, {0x80, 0x2080, 0x1080, 0x10F8, 0x8108, 0x4108, 0x4910, 0xA10, 0x1020, 0x1020, 0xE050, 0x2050, 0x2088, 0x2108, 0x2204, 0x402}},
    {27748, 2, {0x0, 0x23F0, 0x1020, 0x1040, 0x8080, 0x4100, 0x43FC, 0x1124, 0x1124, 0x2124, 0xE224, 0x2244, 0x2444, 0x2884, 0x2128, 0x210}},
    {27749, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x8040, 0x4040, 0x4040, 0x17FC, 0x1204, 0x2208, 0xE110, 0x20A0, 0x2040, 0x20A0, 0x2318, 0xC06}},
    {27750, 2, {0x8, 0x203C, 0x13E0, 0x1220, 0x8220, 0x4220, 0x4A20, 0xBFE, 0x1220, 0x1210, 0xE210, 0x2212, 0x220A, 0x228A, 0x2306, 0x202}},
    {27751, 2, {0x0, 0x27FC, 0x1110, 0x1110, 0x8110, 0x4110, 0x4910, 0xFFE, 0x1110, 0x1110, 0xE110, 0x2110, 0x2110, 0x2110, 0x2210, 0x410}},
    {27752, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x8204, 0x4204, 0x4A04, 0xBFC, 0x1204, 0x1204, 0xE204, 0x2204, 0x2204, 0x23FC, 0x2204, 0x0}},
    {27753, 2, {0x0, 0x2000, 0x17FC, 0x1404, 0x8404, 0x4404, 0x4404, 0x17F4, 0x1404, 0x2404, 0xE404, 0x2404, 0x2404, 0x27FC, 0x2404, 0x0}},
    {27754, 2, {0x0, 0x27FC, 0x1040, 0x1040, 0x8040, 0x4040, 0x4040, 0x13FC, 0x1040, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {27755, 2, {0x110, 0x2110, 0x1110, 0x1110, 0x87FC, 0x4110, 0x4110, 0x1110, 0x1110, 0x2FFE, 0xE110, 0x2110, 0x2110, 0x2210, 0x2210, 0x410}},
    {27756, 2, {0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFFFE, 0x1020, 0x2020, 0x4120, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {27757, 2, {0x20, 0x2020, 0x1020, 0x1020, 0x83FE, 0x4222, 0x4A22, 0xA22, 0x1252, 0x124A, 0xE28A, 0x2302, 0x2202, 0x2202, 0x220A, 0x204}},
    {27758, 2, {0x100, 0x4100, 0x21FC, 0x2204, 0x204, 0x8504, 0x4884, 0x4084, 0x1014, 0x2024, 0xE044, 0x2384, 0x2104, 0x2004, 0x2028, 0x10}},
    {27759, 2, {0x80, 0x4080, 0x2080, 0x2FFE, 0x80, 0x8080, 0x4120, 0x5120, 0x1120, 0x2240, 0xE240, 0x2488, 0x2488, 0x29FC, 0x2084, 0x4}},
    {27760, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FE, 0x4040, 0x4840, 0x8A0, 0x10A0, 0x10A0, 0xE110, 0x2110, 0x2288, 0x2248, 0x2444, 0x802}},
    {27761, 2, {0x40, 0x2050, 0x1048, 0x1048, 0x8040, 0x47FE, 0x4840, 0x840, 0x10A0, 0x10A0, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {27762, 2, {0x0, 0x27F8, 0x1108, 0x1110, 0x8110, 0x4120, 0x493C, 0x904, 0x1104, 0x1288, 0xE288, 0x2250, 0x2420, 0x2450, 0x2888, 0x306}},
    {27763, 2, {0x8, 0x203C, 0x13C0, 0x1200, 0x8200, 0x4200, 0x4BF8, 0xA88, 0x1288, 0x1288, 0xE250, 0x2250, 0x2220, 0x2450, 0x2488, 0x906}},
    {27764, 2, {0x80, 0x2040, 0x1040, 0x1000, 0x87FE, 0x4080, 0x4880, 0x8C0, 0x10A0, 0x1090, 0xE088, 0x2088, 0x2080, 0x2080, 0x2080, 0x80}},
    {27765, 2, {0x40, 0x2040, 0x10A0, 0x10A0, 0x8110, 0x4288, 0x4C46, 0x840, 0x1000, 0x13F8, 0xE008, 0x2010, 0x2010, 0x2020, 0x2020, 0x40}},
    {27766, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x8110, 0x4110, 0x4910, 0x910, 0x1110, 0x10A0, 0xE0A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {27767, 2, {0x80, 0x2080, 0x1080, 0x1080, 0x81F8, 0x4108, 0x4A08, 0xD10, 0x1110, 0x10A0, 0xE0A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0x406}},
    {27768, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x8100, 0x4100, 0x4900, 0x9F8, 0x1108, 0x1108, 0xE108, 0x2108, 0x2208, 0x2208, 0x2450, 0x820}},
    {27769, 2, {0x0, 0x2008, 0x1108, 0x1088, 0x8252, 0x4252, 0x4A22, 0xA22, 0x1252, 0x1292, 0xE30A, 0x220A, 0x2202, 0x23FE, 0x2002, 0x0}},
    {27770, 2, {0x80, 0x2080, 0x1080, 0x17F8, 0x8088, 0x4088, 0x4888, 0x888, 0x17FE, 0x1080, 0xE140, 0x2140, 0x2220, 0x2410, 0x2808, 0x1006}},
    {27771, 2, {0x80, 0x2080, 0x1080, 0x11FC, 0x8120, 0x4120, 0x4A20, 0x820, 0x1020, 0x17FE, 0xE020, 0x2020, 0x2020, 0x2020, 0x2020, 0x20}},
    {27772, 2, {0x20, 0x2020, 0x1120, 0x1120, 0x81FC, 0x4220, 0x4220, 0x1420, 0x1020, 0x27FE, 0xE020, 0x2020, 0x2020, 0x2020, 0x2020, 0x20}},
    {27773, 2, {0x100, 0x2100, 0x11FC, 0x1200, 0x8400, 0x41F8, 0x4800, 0x800, 0x13F8, 0x1008, 0xE008, 0x2008, 0x200A, 0x200A, 0x2006, 0x2}},
    {27774, 2, {0x10, 0x2110, 0x1110, 0x1208, 0x8208, 0x4404, 0x5802, 0x13F8, 0x1108, 0x2108, 0xE108, 0x2108, 0x2208, 0x2208, 0x2450, 0x820}},
    {27775, 2, {0x0, 0x23F8, 0x1008, 0x1110, 0x80A0, 0x4040, 0x47FE, 0x1042, 0x1044, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2140, 0x80}},
    {27776, 2, {0x800, 0x87C, 0x804, 0x928, 0x910, 0xEA08, 0x2CFE, 0x2A12, 0x2914, 0x4910, 0x4890, 0x8890, 0x810, 0x810, 0x2850, 0x1020}},
    {27777, 2, {0x0, 0x2080, 0x1040, 0x1020, 0x8120, 0x4100, 0x4100, 0x1504, 0x1502, 0x2502, 0xE502, 0x2908, 0x2108, 0x2108, 0x20F8, 0x0}},
    {27778, 2, {0x8, 0x203C, 0x13C0, 0x1200, 0x8200, 0x4200, 0x43FE, 0x1210, 0x1210, 0x2210, 0xE210, 0x2210, 0x2210, 0x2410, 0x2410, 0x810}},
    {27779, 2, {0x10, 0x2078, 0x17C0, 0x1040, 0x8040, 0x4040, 0x4040, 0x17FE, 0x1040, 0x20A0, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {27780, 2, {0x0, 0x2000, 0x13FC, 0x1000, 0x8000, 0x4000, 0x47FE, 0x1040, 0x1040, 0x2080, 0xE080, 0x2108, 0x2204, 0x27FE, 0x2202, 0x0}},
    {27781, 2, {0x0, 0x2000, 0x13F8, 0x1000, 0x8000, 0x4000, 0x47FC, 0x1120, 0x1120, 0x2120, 0xE120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {27782, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x8000, 0x4000, 0x49F0, 0x910, 0x1110, 0x1110, 0xE110, 0x2112, 0x2212, 0x2212, 0x240E, 0x800}},
    {27783, 2, {0x80, 0x2080, 0x1100, 0x1110, 0x8208, 0x47FC, 0x4004, 0x1120, 0x1120, 0x2120, 0xE120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {27784, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x8444, 0x4848, 0x4040, 0x10A0, 0x10A0, 0x20A0, 0xE120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {27785, 2, {0x0, 0x2000, 0x13FE, 0x1202, 0x8404, 0x4000, 0x41F0, 0x1110, 0x1110, 0x2110, 0xE110, 0x2112, 0x2212, 0x2212, 0x240E, 0x800}},
    {27786, 2, {0x1020, 0x1020, 0x1020, 0x7E24, 0x53A8, 0x94B0, 0x10A8, 0x1124, 0x2922, 0x2A22, 0x2C20, 0x28A2, 0x2842, 0x4802, 0x47FE, 0x8000}},
    {27787, 2, {0x90, 0x2088, 0x1088, 0x1080, 0x87FE, 0x40A0, 0x48A0, 0x8A0, 0x10A0, 0x10A0, 0xE120, 0x2122, 0x2122, 0x2222, 0x221E, 0x400}},
    {27788, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8040, 0x4248, 0x4A48, 0xA48, 0x1248, 0x13F8, 0xE048, 0x2040, 0x2042, 0x2042, 0x203E, 0x0}},
    {27789, 2, {0x0, 0x27FE, 0x1080, 0x1080, 0x8080, 0x40F8, 0x4908, 0x908, 0x1108, 0x1208, 0xE3F8, 0x2010, 0x2010, 0x2010, 0x27FE, 0x0}},
    {27790, 2, {0x120, 0x2120, 0x1120, 0x1224, 0x8224, 0x4628, 0x5630, 0x1A20, 0x1260, 0x22A0, 0xE220, 0x2222, 0x2222, 0x2222, 0x221E, 0x200}},
    {27791, 2, {0x0, 0x2200, 0x12FC, 0x1224, 0x8224, 0x4224, 0x53A4, 0x1E24, 0x1224, 0x2224, 0xE2A4, 0x2324, 0x2244, 0x2044, 0x2094, 0x108}},
    {27792, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FE, 0x4040, 0x48E0, 0x8E0, 0x1150, 0x1150, 0xE248, 0x2444, 0x2842, 0x2040, 0x2040, 0x40}},
    {27793, 2, {0x0, 0x23F8, 0x1088, 0x1088, 0x8088, 0x4088, 0x4888, 0xBF8, 0x1108, 0x1108, 0xE108, 0x2108, 0x2108, 0x2108, 0x2FFE, 0x0}},
    {27794, 2, {0x80, 0x2080, 0x10FC, 0x1104, 0x8104, 0x4228, 0x4810, 0xBFC, 0x1104, 0x1088, 0xE088, 0x2050, 0x2020, 0x2050, 0x2188, 0x606}},
    {27795, 2, {0x100, 0x108, 0x7D10, 0x5A0, 0x960, 0x3118, 0xC506, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {27796, 2, {0x0, 0x27FE, 0x1040, 0x1040, 0x8240, 0x427C, 0x4A44, 0xA44, 0x1244, 0x13FC, 0xE004, 0x2004, 0x2004, 0x2004, 0x2028, 0x10}},
    {27797, 2, {0x100, 0x2100, 0x1100, 0x11FC, 0x82A4, 0x44A4, 0x40A4, 0x1124, 0x1124, 0x2244, 0xE444, 0x2084, 0x2104, 0x2204, 0x2428, 0x10}},
    {27798, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FC, 0x4444, 0x4444, 0x1444, 0x1444, 0x2444, 0xE7FC, 0x2444, 0x2040, 0x2040, 0x2040, 0x40}},
    {27799, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE10E, 0x1110, 0xB20, 0x5C0, 0x1930, 0x6108, 0x500, 0x200}},
    {27800, 2, {0x20, 0x2220, 0x1220, 0x1222, 0x8222, 0x4224, 0x4BA8, 0xA30, 0x1220, 0x1220, 0xE220, 0x2222, 0x22A2, 0x2322, 0x221E, 0x0}},
    {27801, 2, {0x40, 0x2040, 0x1040, 0x1148, 0x8144, 0x4242, 0x4242, 0x1448, 0x1048, 0x2048, 0xE010, 0x2010, 0x2020, 0x2040, 0x2180, 0x600}},
    {27802, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x8240, 0x4240, 0x4A7C, 0xA40, 0x1240, 0x1240, 0xE240, 0x2240, 0x2240, 0x2240, 0x2FFE, 0x0}},
    {27803, 2, {0x40, 0x4040, 0x2040, 0x2FFE, 0x40, 0x8040, 0x47FC, 0x5444, 0x1444, 0x2444, 0xE444, 0x2454, 0x2448, 0x2040, 0x2040, 0x40}},
    {27804, 2, {0x10, 0x2210, 0x1210, 0x1210, 0x8210, 0x43FE, 0x4A00, 0xA00, 0x1200, 0x13F8, 0xE208, 0x2208, 0x2208, 0x2408, 0x2408, 0x808}},
    {27805, 2, {0x820, 0x820, 0x820, 0xA20, 0xA32, 0xEDB2, 0x28B4, 0x2CA8, 0x2AA8, 0x2AA8, 0x4924, 0x4924, 0x8A22, 0x820, 0x28A0, 0x1040}},
    {27806, 2, {0x0, 0x4FFE, 0x2040, 0x2040, 0x40, 0x87FC, 0x4444, 0x5444, 0x1444, 0x2444, 0xE444, 0x2454, 0x2448, 0x2040, 0x2040, 0x40}},
    {27807, 2, {0x100, 0x2100, 0x1100, 0x11FC, 0x8204, 0x4484, 0x4084, 0x1104, 0x1144, 0x2224, 0xE7F4, 0x2214, 0x2004, 0x2004, 0x2028, 0x10}},
    {27808, 2, {0x8, 0x203C, 0x17D0, 0x1490, 0x8490, 0x4490, 0x4490, 0x1490, 0x1490, 0x2488, 0xE488, 0x2488, 0x2484, 0x2884, 0x2882, 0x1000}},
    {27809, 2, {0x0, 0x21F0, 0x1110, 0x1110, 0x8110, 0x4210, 0x540E, 0x1800, 0x13F8, 0x2208, 0xE108, 0x2110, 0x20A0, 0x2040, 0x21B0, 0xE0E}},
    {27810, 2, {0x0, 0x21FC, 0x1104, 0x1104, 0x8104, 0x4104, 0x49FC, 0x924, 0x1120, 0x1120, 0xE110, 0x2110, 0x2208, 0x2208, 0x2404, 0x802}},
    {27811, 2, {0x20, 0x2020, 0x1020, 0x13FE, 0x8020, 0x4020, 0x4820, 0x9FC, 0x1020, 0x1020, 0xE020, 0x27FE, 0x2020, 0x2020, 0x2020, 0x20}},
    {27812, 2, {0x0, 0x23FC, 0x1200, 0x1208, 0x8288, 0x4250, 0x4A50, 0xA20, 0x1220, 0x1250, 0xE250, 0x2288, 0x2308, 0x2200, 0x23FE, 0x0}},
    {27813, 2, {0x0, 0x23FE, 0x1200, 0x1220, 0x8220, 0x4220, 0x4AFC, 0xA24, 0x1224, 0x1224, 0xE244, 0x2244, 0x2484, 0x2484, 0x2928, 0x1210}},
    {27814, 2, {0x40, 0x2040, 0x10A0, 0x10A0, 0x8110, 0x4208, 0x4406, 0x1110, 0x1120, 0x2140, 0xE180, 0x2100, 0x2104, 0x2104, 0x20FC, 0x0}},
    {27815, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x208, 0x8404, 0x4BF2, 0x4210, 0x1210, 0x2210, 0xE250, 0x2220, 0x2204, 0x2204, 0x21FC, 0x0}},
    {27816, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8328, 0x42A8, 0x4AA8, 0xA48, 0x1248, 0x12A8, 0xE2A8, 0x232A, 0x240A, 0x240A, 0x2806, 0x1002}},
    {27817, 2, {0x40, 0x2240, 0x1140, 0x1140, 0x8040, 0x47FC, 0x4044, 0x1044, 0x1084, 0x20A4, 0xE094, 0x2114, 0x2104, 0x2204, 0x2428, 0x810}},
    {27818, 2, {0x80, 0x2040, 0x1040, 0x13FC, 0x8204, 0x4204, 0x4204, 0x13FC, 0x1204, 0x2200, 0xE200, 0x2200, 0x2400, 0x2400, 0x2800, 0x1000}},
    {27819, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x8040, 0x4040, 0x4840, 0xBFC, 0x1040, 0x10E0, 0xE150, 0x2248, 0x2446, 0x2840, 0x2040, 0x40}},
    {27820, 2, {0x40, 0x2040, 0x1040, 0x13FC, 0x8040, 0x4040, 0x4840, 0xFFE, 0x10E0, 0x1150, 0xE150, 0x2248, 0x2444, 0x2842, 0x2040, 0x40}},
    {27821, 2, {0x40, 0x2050, 0x1048, 0x1040, 0x87FC, 0x4040, 0x40E0, 0x10E0, 0x1150, 0x2150, 0xE248, 0x2444, 0x2842, 0x2040, 0x2040, 0x40}},
    {27822, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x43F8, 0x4A08, 0xA08, 0x1208, 0x13F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x2FFE, 0x0}},
    {27823, 2, {0x100, 0x108, 0x7D10, 0x5A0, 0x960, 0x3118, 0xC506, 0x200, 0x7FFC, 0x400, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {27824, 2, {0x0, 0x2000, 0x17FE, 0x1040, 0x8040, 0x4080, 0x4880, 0x9F8, 0x1308, 0x1508, 0xE908, 0x2108, 0x2108, 0x2108, 0x21F8, 0x108}},
    {27825, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x8404, 0x4808, 0x4200, 0x1210, 0x1220, 0x2240, 0xE380, 0x2204, 0x2204, 0x2204, 0x21FC, 0x0}},
    {27826, 2, {0x100, 0x2100, 0x13FC, 0x1200, 0x8440, 0x4840, 0x4278, 0x12C8, 0x1348, 0x2E48, 0xE248, 0x225A, 0x2242, 0x2202, 0x21FE, 0x0}},
    {27827, 2, {0x0, 0x2000, 0x17FE, 0x1008, 0x8008, 0x43C8, 0x4248, 0x1248, 0x1248, 0x2248, 0xE3C8, 0x2248, 0x2008, 0x2008, 0x2028, 0x10}},
    {27828, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x4426, 0x4040, 0x1080, 0x1310, 0x2020, 0xE040, 0x2088, 0x2310, 0x2020, 0x20C0, 0x700}},
    {27829, 2, {0x80, 0x2080, 0x1080, 0x11FE, 0x8102, 0x4204, 0x5420, 0x1020, 0x1128, 0x2124, 0xE224, 0x2222, 0x2422, 0x2020, 0x20A0, 0x40}},
    {27830, 2, {0x0, 0x23FC, 0x1000, 0x1000, 0x8000, 0x4000, 0x4FFE, 0x1040, 0x1040, 0x2248, 0xE244, 0x2444, 0x2442, 0x2842, 0x2140, 0x80}},
    {27831, 2, {0x90, 0x2088, 0x1088, 0x1080, 0x87FE, 0x4080, 0x4900, 0x9FC, 0x1144, 0x1244, 0xE248, 0x2450, 0x2420, 0x2850, 0x2088, 0x306}},
    {27832, 2, {0x120, 0x2120, 0x1120, 0x17F8, 0x8128, 0x4128, 0x47F8, 0x1520, 0x1520, 0x27FC, 0xE124, 0x2124, 0x2234, 0x2228, 0x2420, 0x820}},
    {27833, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FC, 0x4444, 0x4444, 0x1444, 0x1444, 0x27FC, 0xE444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x404}},
    {27834, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x8444, 0x4444, 0x4444, 0x17FC, 0x1444, 0x2444, 0xE444, 0x2444, 0x2444, 0x27FC, 0x2404, 0x0}},
    {27835, 2, {0x40, 0x2040, 0x1080, 0x1110, 0x8208, 0x47FC, 0x4004, 0x1000, 0x13F8, 0x2208, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27836, 2, {0x0, 0x27FC, 0x1084, 0x1084, 0x8084, 0x4104, 0x4114, 0x1208, 0x1400, 0x23FC, 0xE204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {27837, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FE, 0x4040, 0x4840, 0x840, 0x13F8, 0x1208, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27838, 2, {0x40, 0x2040, 0x1040, 0x107E, 0x8040, 0x4040, 0x4840, 0x840, 0x13FC, 0x1204, 0xE204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {27839, 2, {0x0, 0x21F0, 0x1110, 0x1110, 0x8110, 0x4110, 0x420E, 0x1400, 0x1000, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27840, 2, {0x0, 0x23FC, 0x1004, 0x1004, 0x87F4, 0x4004, 0x4804, 0xBE4, 0x1224, 0x1224, 0xE224, 0x23E4, 0x2004, 0x2004, 0x2014, 0x8}},
    {27841, 2, {0x0, 0x27F8, 0x1408, 0x1408, 0x8408, 0x4408, 0x47F8, 0x1120, 0x1120, 0x2120, 0xE120, 0x2220, 0x2222, 0x2422, 0x281E, 0x1000}},
    {27842, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x8204, 0x42F4, 0x4A94, 0xA94, 0x1294, 0x1294, 0xE2F4, 0x2294, 0x2204, 0x2204, 0x2214, 0x208}},
    {27843, 2, {0x100, 0x2100, 0x1200, 0x13FC, 0x8404, 0x4804, 0x43E4, 0x1224, 0x1224, 0x2224, 0xE224, 0x23E4, 0x2224, 0x2004, 0x2028, 0x10}},
    {27844, 2, {0x48, 0x2248, 0x1248, 0x1248, 0x8248, 0x4FFE, 0x4248, 0x1248, 0x1248, 0x2248, 0xE278, 0x2200, 0x2200, 0x2200, 0x23FE, 0x0}},
    {27845, 2, {0x0, 0x27FC, 0x1404, 0x1444, 0x8444, 0x4444, 0x4444, 0x1444, 0x1444, 0x24A4, 0xE494, 0x2514, 0x2604, 0x2404, 0x27FC, 0x404}},
    {27846, 2, {0x40, 0x2040, 0x1240, 0x1240, 0x83FC, 0x4240, 0x4440, 0x1040, 0x17FE, 0x2040, 0xE0A0, 0x20A0, 0x2110, 0x2208, 0x2404, 0x802}},
    {27847, 2, {0x200, 0x2200, 0x1200, 0x121E, 0x87D2, 0x4252, 0x4A52, 0xA52, 0x1252, 0x1252, 0xE252, 0x2252, 0x2452, 0x245E, 0x2952, 0x1080}},
    {27848, 2, {0x80, 0x2080, 0x10F8, 0x1108, 0x8310, 0x44A0, 0x4040, 0x10A0, 0x1318, 0x2C06, 0xE0C0, 0x2020, 0x2010, 0x2180, 0x2060, 0x10}},
    {27849, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {27850, 2, {0x40, 0x2040, 0x1080, 0x13FC, 0x8204, 0x4204, 0x4A04, 0xA04, 0x13FC, 0x1204, 0xE204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {27851, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8040, 0x4040, 0x4150, 0x1110, 0x1110, 0x2FFE, 0xE110, 0x2110, 0x2210, 0x2210, 0x2410, 0x810}},
    {27852, 2, {0x0, 0x2080, 0x1040, 0x1028, 0x8128, 0x4108, 0x4110, 0x1514, 0x1522, 0x2522, 0xE542, 0x2988, 0x2108, 0x2308, 0x24F8, 0x800}},
    {27853, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FC, 0x4040, 0x48E0, 0x8E0, 0x1150, 0x1150, 0xE248, 0x25F4, 0x2842, 0x2040, 0x2040, 0x40}},
    {27854, 2, {0x100, 0x2100, 0x1100, 0x11FE, 0x8280, 0x4280, 0x4480, 0x10F8, 0x1080, 0x2080, 0xE080, 0x20FC, 0x2080, 0x2080, 0x2080, 0x80}},
    {27855, 2, {0x40, 0x2040, 0x1248, 0x1248, 0x8248, 0x4248, 0x4BF8, 0x848, 0x1040, 0x1040, 0xE444, 0x2444, 0x2444, 0x2444, 0x27FC, 0x4}},
    {27856, 2, {0x10, 0x4790, 0x2490, 0x2510, 0x57E, 0x8612, 0x4512, 0x5492, 0x1492, 0x2492, 0xE692, 0x2512, 0x2422, 0x2422, 0x244A, 0x484}},
    {27857, 2, {0x220, 0x4220, 0x2220, 0x2420, 0x57C, 0x8924, 0x4E24, 0x4224, 0x1424, 0x2424, 0xE924, 0x2F24, 0x2144, 0x2044, 0x2094, 0x108}},
    {27858, 2, {0x8, 0x203C, 0x13D0, 0x1290, 0x8290, 0x4290, 0x4A90, 0xA90, 0x1290, 0x1288, 0xE288, 0x22C8, 0x22A4, 0x24D4, 0x2492, 0x800}},
    {27859, 2, {0x8, 0x2788, 0x1088, 0x1088, 0x8088, 0x4788, 0x4408, 0x1410, 0x1410, 0x2790, 0xE094, 0x20A4, 0x20A2, 0x20BE, 0x2502, 0x200}},
    {27860, 2, {0x108, 0x2108, 0x1108, 0x1108, 0x87FE, 0x4108, 0x4108, 0x1108, 0x1108, 0x21F8, 0xE108, 0x2108, 0x2108, 0x2108, 0x21F8, 0x108}},
    {27861, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x87FC, 0x4040, 0x4040, 0x1040, 0x1FFE, 0x2040, 0xE080, 0x2100, 0x2210, 0x2408, 0x2FFC, 0x404}},
    {27862, 2, {0x0, 0x2100, 0x163C, 0x14A4, 0x84A4, 0x44A4, 0x44A4, 0x14A4, 0x14A4, 0x25A4, 0xE6B4, 0x24A8, 0x2120, 0x2120, 0x2220, 0x420}},
    {27863, 2, {0x0, 0x2000, 0x17FC, 0x14A4, 0x84A4, 0x44A4, 0x44A4, 0x14A4, 0x14A4, 0x2524, 0xE51C, 0x2604, 0x2404, 0x2404, 0x27FC, 0x404}},
    {27864, 2, {0x10, 0x2078, 0x17C0, 0x1040, 0x8444, 0x4244, 0x4248, 0x1040, 0x17FE, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2140, 0x80}},
    {27865, 2, {0x0, 0x23F8, 0x1040, 0x1040, 0x8248, 0x4148, 0x4950, 0x840, 0x17FE, 0x1040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {27866, 2, {0x120, 0x2120, 0x1120, 0x1524, 0x8524, 0x45E8, 0x4530, 0x1520, 0x1520, 0x2520, 0xE520, 0x2522, 0x25E2, 0x2E22, 0x241E, 0x0}},
    {27867, 2, {0x10, 0x2078, 0x1780, 0x1000, 0x8080, 0x4040, 0x47F8, 0x1010, 0x1020, 0x2040, 0xE080, 0x2100, 0x2200, 0x2500, 0x28FC, 0x0}},
    {27868, 2, {0x8, 0x203C, 0x13E0, 0x1220, 0x8220, 0x4220, 0x4A20, 0xBFE, 0x1220, 0x1210, 0xE210, 0x2212, 0x220A, 0x228A, 0x2326, 0x212}},
    {27869, 2, {0x8, 0x203C, 0x13C0, 0x1200, 0x8200, 0x4200, 0x4BFE, 0xA10, 0x1210, 0x1270, 0xE210, 0x2218, 0x2214, 0x2410, 0x2410, 0x810}},
    {27870, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x8404, 0x4808, 0x4000, 0x1000, 0x17FC, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2140, 0x80}},
    {27871, 2, {0x0, 0x27FC, 0x1040, 0x1040, 0x8040, 0x4040, 0x4A40, 0xA78, 0x1240, 0x1240, 0xE240, 0x2240, 0x2240, 0x2240, 0x2FFE, 0x0}},
    {27872, 2, {0x40, 0x2040, 0x10A0, 0x10A0, 0x8110, 0x4248, 0x4426, 0x1020, 0x13F8, 0x2008, 0xE010, 0x2110, 0x20A0, 0x2040, 0x2020, 0x20}},
    {27873, 2, {0x100, 0x2100, 0x11FC, 0x1204, 0x8204, 0x47E4, 0x4A24, 0x1224, 0x1224, 0x23E4, 0xE214, 0x2208, 0x2202, 0x2202, 0x21FE, 0x0}},
    {27874, 2, {0x20, 0x2020, 0x1020, 0x13FE, 0x8222, 0x4224, 0x4A20, 0xBFC, 0x1284, 0x1288, 0xE248, 0x2250, 0x2220, 0x2450, 0x2488, 0x906}},
    {27875, 2, {0x80, 0x2040, 0x1040, 0x1000, 0x87FC, 0x4000, 0x4008, 0x1208, 0x1208, 0x2110, 0xE110, 0x2110, 0x2120, 0x2020, 0x2FFE, 0x0}},
    {27876, 2, {0x8, 0x2108, 0x1088, 0x1448, 0x8448, 0x4408, 0x4408, 0x1408, 0x1410, 0x2410, 0xE410, 0x2528, 0x2624, 0x2444, 0x2082, 0x102}},
    {27877, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x8204, 0x43FC, 0x4A00, 0xA40, 0x1244, 0x1248, 0xE270, 0x2240, 0x2442, 0x2442, 0x283E, 0x1000}},
    {27878, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4240, 0x4A40, 0xA40, 0x13F8, 0x1248, 0xE248, 0x2248, 0x2488, 0x248A, 0x290A, 0x206}},
    {27879, 2, {0x28, 0x2024, 0x1024, 0x1020, 0x83FE, 0x4220, 0x4A24, 0xA24, 0x1224, 0x1228, 0xE228, 0x2290, 0x2312, 0x222A, 0x2046, 0x82}},
    {27880, 2, {0x80, 0x2040, 0x1000, 0x17FC, 0x8040, 0x4040, 0x4040, 0x1040, 0x13FC, 0x2040, 0xE040, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {27881, 2, {0x40, 0x2240, 0x1240, 0x1240, 0x83FC, 0x4440, 0x5440, 0x1840, 0x1040, 0x23FC, 0xE040, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {27882, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x8204, 0x43FC, 0x4A04, 0xA04, 0x1204, 0x13FC, 0xE204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {27883, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x8040, 0x4080, 0x4108, 0x1210, 0x17E0, 0x2040, 0xE080, 0x2108, 0x2204, 0x27FE, 0x2202, 0x0}},
    {27884, 2, {0x40, 0x2020, 0x1020, 0x13FE, 0x8202, 0x4404, 0x4000, 0x1090, 0x1090, 0x2090, 0xE088, 0x2108, 0x2108, 0x2104, 0x2204, 0x402}},
    {27885, 2, {0x108, 0x2108, 0x1108, 0x1208, 0x82FE, 0x4608, 0x4A08, 0x1208, 0x1288, 0x2248, 0xE248, 0x2208, 0x2208, 0x2208, 0x2228, 0x210}},
    {27886, 2, {0x40, 0x2040, 0x1248, 0x1148, 0x8150, 0x4040, 0x4BF8, 0x840, 0x1040, 0x1040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {27887, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x8204, 0x43FC, 0x4A20, 0xA20, 0x13FE, 0x1220, 0xE220, 0x2210, 0x2212, 0x228A, 0x2306, 0x202}},
    {27888, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x440, 0x920, 0x3118, 0xCB26, 0x5C0, 0x1930, 0x6108, 0x500, 0x200}},
    {27889, 2, {0x40, 0x2040, 0x1040, 0x13F8, 0x8248, 0x4248, 0x4A48, 0xA48, 0x1248, 0x17FE, 0xE040, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x406}},
    {27890, 2, {0x40, 0x2040, 0x105C, 0x13E0, 0x8240, 0x4240, 0x4A40, 0xBFC, 0x1044, 0x10C4, 0xE144, 0x2254, 0x2C48, 0x2040, 0x2040, 0x40}},
    {27891, 2, {0x80, 0x2040, 0x1020, 0x13C0, 0x8044, 0x4044, 0x4F48, 0x1150, 0x1160, 0x2160, 0xE250, 0x2448, 0x2844, 0x3042, 0x2140, 0x80}},
    {27892, 2, {0x100, 0x108, 0x7D10, 0x5A0, 0x960, 0x3118, 0xC506, 0x200, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {27893, 2, {0x0, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x100, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {27894, 2, {0x2208, 0x1108, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8004, 0x100, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200}},
    {27895, 2, {0x120, 0x2110, 0x1110, 0x1100, 0x87FE, 0x4140, 0x4940, 0x948, 0x1148, 0x1250, 0xE250, 0x2262, 0x2442, 0x24C2, 0x293E, 0x1000}},
    {27896, 2, {0x40, 0x2040, 0x107E, 0x1040, 0x8040, 0x43FC, 0x4A04, 0xA04, 0x1204, 0x13FC, 0xE204, 0x2200, 0x2200, 0x2400, 0x2400, 0x800}},
    {27897, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x4208, 0x4BF8, 0xA08, 0x1208, 0x1208, 0xE208, 0x23F8, 0x2000, 0x2000, 0x2FFE, 0x0}},
    {27898, 2, {0x8, 0x203C, 0x17C0, 0x1400, 0x8440, 0x4440, 0x4440, 0x17FE, 0x1040, 0x2040, 0xE248, 0x2244, 0x2442, 0x2842, 0x2140, 0x80}},
    {27899, 2, {0x0, 0x27FE, 0x1402, 0x1904, 0x8100, 0x41FC, 0x4900, 0xA00, 0x13FC, 0x1004, 0xE004, 0x27F4, 0x2004, 0x2004, 0x2028, 0x10}},
    {27900, 2, {0x40, 0x2048, 0x1244, 0x1240, 0x83FE, 0x4080, 0x4880, 0x8FC, 0x1144, 0x1144, 0xE148, 0x2250, 0x2220, 0x2450, 0x2888, 0x306}},
    {27901, 2, {0x0, 0x27F8, 0x1208, 0x1110, 0x80A0, 0x4040, 0x41B0, 0x164E, 0x1040, 0x23F8, 0xE040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x40}},
    {27902, 2, {0x0, 0x27F8, 0x1010, 0x1020, 0x8060, 0x4098, 0x4104, 0x1602, 0x1000, 0x23FC, 0xE040, 0x2040, 0x2040, 0x2040, 0x27FE, 0x0}},
    {27903, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4208, 0x4A08, 0xBF8, 0x1244, 0x1248, 0xE230, 0x2220, 0x2210, 0x2288, 0x2306, 0x200}},
    {27904, 2, {0x80, 0x2100, 0x13F8, 0x1208, 0x8288, 0x4248, 0x4248, 0x1FFE, 0x1208, 0x2288, 0xE248, 0x2248, 0x2208, 0x2208, 0x2428, 0x810}},
    {27905, 2, {0x40, 0x4040, 0x2040, 0x2FFE, 0x40, 0x8040, 0x47FC, 0x5000, 0x1000, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27906, 2, {0x40, 0x2020, 0x1020, 0x17FE, 0x8090, 0x4090, 0x4890, 0xA94, 0x1292, 0x1292, 0xE492, 0x2090, 0x2110, 0x2110, 0x2250, 0x420}},
    {27907, 2, {0x80, 0x2080, 0x1080, 0x17FE, 0x8100, 0x4120, 0x4A20, 0xAA4, 0x14A4, 0x1528, 0xEA50, 0x2050, 0x2088, 0x2108, 0x2204, 0x402}},
    {27908, 2, {0x0, 0x27FE, 0x1402, 0x1402, 0x84F2, 0x4492, 0x4492, 0x1492, 0x1492, 0x2492, 0xE4F2, 0x2492, 0x2402, 0x2402, 0x27FE, 0x402}},
    {27909, 2, {0x0, 0x27FC, 0x1040, 0x1040, 0x83F8, 0x4248, 0x4A48, 0xBF8, 0x1248, 0x1248, 0xEFFE, 0x2208, 0x2208, 0x2208, 0x2228, 0x210}},
    {27910, 2, {0x0, 0x23F8, 0x1010, 0x1020, 0x8044, 0x4764, 0x4168, 0x1150, 0x1250, 0x2248, 0xE444, 0x2942, 0x2080, 0x2000, 0x2FFE, 0x0}},
    {27911, 2, {0x0, 0x27FC, 0x1404, 0x1444, 0x8444, 0x4444, 0x47FC, 0x1444, 0x1444, 0x24A4, 0xE494, 0x2514, 0x2604, 0x2404, 0x27FC, 0x404}},
    {27912, 2, {0x100, 0x2100, 0x13F8, 0x1208, 0x8410, 0x4BFE, 0x4200, 0x12F8, 0x1288, 0x2288, 0xE2A8, 0x2290, 0x2282, 0x2482, 0x247E, 0x800}},
    {27913, 2, {0x8, 0x203C, 0x13C0, 0x1200, 0x8200, 0x43FE, 0x4A00, 0xA00, 0x1200, 0x12FC, 0xE284, 0x2284, 0x2484, 0x2484, 0x28FC, 0x84}},
    {27914, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x8080, 0x4080, 0x497C, 0x908, 0x1310, 0x1510, 0xE1FE, 0x2110, 0x2110, 0x2110, 0x2150, 0x120}},
    {27915, 2, {0x208, 0x2108, 0x1110, 0x1000, 0x87FC, 0x4040, 0x4040, 0x13F8, 0x1040, 0x2040, 0xE7FE, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {27916, 2, {0x2, 0x4002, 0x2FC2, 0x2212, 0x212, 0x83D2, 0x4252, 0x5452, 0x1652, 0x2552, 0xE892, 0x2092, 0x2102, 0x2202, 0x240A, 0x804}},
    {27917, 2, {0x0, 0x27FE, 0x1420, 0x1420, 0x8420, 0x44F8, 0x4488, 0x1488, 0x1488, 0x2488, 0xE4F8, 0x2420, 0x2420, 0x2420, 0x27FE, 0x0}},
    {27918, 2, {0x40, 0x2080, 0x13FC, 0x1204, 0x8204, 0x4204, 0x4BFC, 0xA04, 0x1204, 0x1204, 0xE3FC, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {27919, 2, {0x0, 0x2FFE, 0x1040, 0x1040, 0x8080, 0x47FC, 0x44A4, 0x14A4, 0x14A4, 0x24A4, 0xE4A4, 0x24A4, 0x24A4, 0x2484, 0x2414, 0x408}},
    {27920, 2, {0x80, 0x2080, 0x113C, 0x1200, 0x8480, 0x4080, 0x497E, 0xB08, 0x1508, 0x1108, 0xE108, 0x2108, 0x2108, 0x2108, 0x2128, 0x110}},
    {27921, 2, {0x120, 0x4128, 0x2124, 0x2224, 0x220, 0x8620, 0x4AFC, 0x4220, 0x1220, 0x2250, 0xE250, 0x2250, 0x2288, 0x2288, 0x2304, 0x202}},
    {27922, 2, {0x0, 0x2FFE, 0x10A0, 0x10A0, 0x87FC, 0x44A4, 0x44A4, 0x14A4, 0x14A4, 0x24A4, 0xE51C, 0x2604, 0x2404, 0x2404, 0x27FC, 0x404}},
    {27923, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8040, 0x43F8, 0x4A48, 0xA48, 0x1248, 0x1258, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {27924, 2, {0x40, 0x2040, 0x1040, 0x13FC, 0x8040, 0x4040, 0x47FE, 0x1010, 0x1010, 0x27FE, 0xE010, 0x2210, 0x2110, 0x2110, 0x2050, 0x20}},
    {27925, 2, {0x10, 0x2110, 0x1108, 0x1208, 0x8404, 0x4BFA, 0x4208, 0x1208, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {27926, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4000, 0x4000, 0x17FC, 0x1040, 0x2040, 0xEFFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {27927, 2, {0x40, 0x2240, 0x1240, 0x13FC, 0x8440, 0x4840, 0x4040, 0x17FE, 0x1120, 0x2120, 0xE120, 0x2120, 0x2222, 0x2222, 0x241E, 0x800}},
    {27928, 2, {0x80, 0x2088, 0x13E8, 0x1090, 0x80A0, 0x47FE, 0x4880, 0x900, 0x13FC, 0x1480, 0xE900, 0x21F8, 0x2008, 0x2008, 0x2050, 0x20}},
    {27929, 2, {0x40, 0x2240, 0x1240, 0x13FC, 0x8440, 0x4040, 0x4040, 0x17FE, 0x10E0, 0x2150, 0xE150, 0x2248, 0x2444, 0x2842, 0x2040, 0x40}},
    {27930, 2, {0x100, 0x2100, 0x11F8, 0x1310, 0x84A0, 0x4040, 0x51B0, 0x164E, 0x1040, 0x23FC, 0xE040, 0x2440, 0x27FE, 0x2040, 0x2040, 0x40}},
    {27931, 2, {0x80, 0x2080, 0x11F8, 0x1208, 0x8510, 0x40A0, 0x4040, 0x10A0, 0x1318, 0x2C06, 0xE3F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27932, 2, {0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE00E, 0x1FF0, 0x1010, 0x1FF0, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {27933, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x8404, 0x4888, 0x4080, 0x17FC, 0x1110, 0x2110, 0xE210, 0x21A0, 0x2040, 0x20A0, 0x2110, 0x608}},
    {27934, 2, {0x0, 0x27FC, 0x1404, 0x1404, 0x85F4, 0x4404, 0x4404, 0x15F4, 0x1514, 0x2514, 0xE514, 0x25F4, 0x2404, 0x2404, 0x2414, 0x408}},
    {27935, 2, {0x40, 0x2040, 0x17FE, 0x1040, 0x87FC, 0x4044, 0x43FC, 0x1240, 0x1440, 0x27FE, 0xE042, 0x20AA, 0x20A4, 0x2110, 0x2208, 0xC06}},
    {27936, 2, {0x40, 0x2040, 0x1090, 0x1108, 0x83FC, 0x4024, 0x4920, 0x920, 0x11FC, 0x1220, 0xE020, 0x27FE, 0x2020, 0x2020, 0x2020, 0x20}},
    {27937, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8040, 0x43F8, 0x4840, 0x840, 0x17FE, 0x10E0, 0xE150, 0x2150, 0x2248, 0x2444, 0x2842, 0x40}},
    {27938, 2, {0x100, 0x217C, 0x1124, 0x1224, 0x8224, 0x46FE, 0x4A24, 0x1224, 0x1224, 0x227C, 0xE224, 0x2220, 0x2220, 0x2220, 0x2240, 0x280}},
    {27939, 2, {0x40, 0x2040, 0x1248, 0x1148, 0x8150, 0x4040, 0x47FE, 0x1040, 0x10E0, 0x2150, 0xE150, 0x2248, 0x2444, 0x2842, 0x2040, 0x40}},
    {27940, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x4406, 0x43F8, 0x1040, 0x1040, 0x2040, 0xE3F8, 0x2040, 0x2040, 0x2040, 0x27FE, 0x0}},
    {27941, 2, {0x40, 0x2040, 0x13F8, 0x1048, 0x87FE, 0x4048, 0x4BF8, 0x840, 0x1040, 0x13F8, 0xE040, 0x2040, 0x27FC, 0x2040, 0x2040, 0x40}},
    {27942, 2, {0x0, 0x27FE, 0x1020, 0x1020, 0x8040, 0x43FC, 0x4A04, 0xA04, 0x1204, 0x13FC, 0xE204, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {27943, 2, {0x40, 0x2040, 0x17FE, 0x1080, 0x8100, 0x41FC, 0x4304, 0x1504, 0x19FC, 0x2104, 0xE104, 0x21FC, 0x2104, 0x2104, 0x2114, 0x108}},
    {27944, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x8000, 0x4110, 0x4208, 0x1404, 0x1110, 0x2110, 0xE0A0, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {27945, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8444, 0x4444, 0x47FC, 0x1444, 0x1444, 0x27FC, 0xE028, 0x2012, 0x2032, 0x204A, 0x2186, 0xE02}},
    {27946, 2, {0x110, 0x2110, 0x1110, 0x1110, 0x87FC, 0x4110, 0x4110, 0x1110, 0x1110, 0x2FFE, 0xE000, 0x2110, 0x2108, 0x2208, 0x2404, 0x804}},
    {27947, 2, {0x40, 0x2040, 0x1080, 0x1100, 0x83FC, 0x4294, 0x4A94, 0xA94, 0x1294, 0x1294, 0xE294, 0x2294, 0x2294, 0x2294, 0x2FFE, 0x0}},
    {27948, 2, {0x0, 0x27F8, 0x1408, 0x1408, 0x87F8, 0x4488, 0x4488, 0x14E8, 0x1528, 0x2528, 0xE6A8, 0x244A, 0x244A, 0x288A, 0x2906, 0x1202}},
    {27949, 2, {0x0, 0x23FE, 0x1200, 0x1200, 0x83FC, 0x4220, 0x4A20, 0xAF8, 0x1220, 0x1220, 0xE220, 0x23FC, 0x2200, 0x2200, 0x23FE, 0x0}},
    {27950, 2, {0x90, 0x2090, 0x1090, 0x1492, 0x8294, 0x4098, 0x4090, 0x1198, 0x1294, 0x2492, 0xE090, 0x2090, 0x2112, 0x2112, 0x2212, 0x40E}},
    {27951, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x900, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {27952, 2, {0x0, 0x23FE, 0x1200, 0x1200, 0x8200, 0x43FC, 0x4A04, 0xA04, 0x1204, 0x1204, 0xE3FC, 0x2200, 0x2200, 0x2200, 0x23FE, 0x0}},
    {27953, 2, {0x0, 0x27FC, 0x1208, 0x1208, 0x83F8, 0x4208, 0x4208, 0x13F8, 0x1208, 0x2208, 0xE20E, 0x23F8, 0x2E08, 0x2008, 0x2008, 0x8}},
    {27954, 2, {0x104, 0x2124, 0x1124, 0x1124, 0x8124, 0x4124, 0x45B4, 0x156C, 0x1924, 0x2124, 0xE124, 0x2124, 0x2124, 0x2224, 0x2204, 0x404}},
    {27955, 2, {0x100, 0x2100, 0x1100, 0x111E, 0x87D2, 0x4252, 0x4252, 0x1252, 0x1252, 0x2492, 0xE292, 0x2112, 0x2292, 0x225E, 0x2452, 0x800}},
    {27956, 2, {0x208, 0x2108, 0x1110, 0x1000, 0x87FC, 0x4110, 0x4110, 0x1110, 0x1110, 0x2FFE, 0xE110, 0x2110, 0x2210, 0x2210, 0x2410, 0x810}},
    {27957, 2, {0x80, 0x2080, 0x1100, 0x11FC, 0x8204, 0x4404, 0x43E4, 0x1224, 0x1224, 0x23E4, 0xE224, 0x2224, 0x23E4, 0x2004, 0x2028, 0x10}},
    {27958, 2, {0x200, 0x2200, 0x1200, 0x13FC, 0x8404, 0x4404, 0x4A24, 0x1144, 0x1494, 0x2554, 0xE634, 0x2414, 0x27F4, 0x2004, 0x2028, 0x10}},
    {27959, 2, {0x0, 0x27FE, 0x1040, 0x1080, 0x8110, 0x4208, 0x47FC, 0x1044, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {27960, 2, {0x80, 0x2088, 0x1488, 0x1290, 0x82A0, 0x4080, 0x4FFC, 0x1120, 0x1120, 0x2120, 0xE120, 0x2124, 0x2224, 0x2224, 0x241C, 0x800}},
    {27961, 2, {0x0, 0x27FC, 0x1000, 0x1000, 0x83F8, 0x4208, 0x4208, 0x13F8, 0x1208, 0x2208, 0xE3F8, 0x2208, 0x2000, 0x2000, 0x2FFE, 0x0}},
    {27962, 2, {0x80, 0x2080, 0x10FC, 0x1104, 0x8288, 0x4450, 0x4020, 0x1040, 0x11FC, 0x2304, 0xED04, 0x2104, 0x2104, 0x2104, 0x21FC, 0x104}},
    {27963, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x8040, 0x47FE, 0x4040, 0x1040, 0x1040, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27964, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x8040, 0x4040, 0x47FC, 0x1000, 0x1040, 0x2040, 0xE3F8, 0x2040, 0x2040, 0x2040, 0x27FC, 0x0}},
    {27965, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x4406, 0x43F8, 0x1000, 0x1000, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27966, 2, {0x8, 0x203C, 0x17C0, 0x1400, 0x841C, 0x45E0, 0x4520, 0x1522, 0x1524, 0x2528, 0xE510, 0x2510, 0x2508, 0x2944, 0x2982, 0x1100}},
    {27967, 2, {0x40, 0x2040, 0x17FC, 0x10A0, 0x8110, 0x4208, 0x4DF6, 0x1000, 0x17FC, 0x2080, 0xE100, 0x23F8, 0x2008, 0x2008, 0x2050, 0x20}},
    {27968, 2, {0x50, 0x2050, 0x1050, 0x1050, 0x83FE, 0x4252, 0x4A52, 0xA52, 0x1252, 0x13FE, 0xE252, 0x2252, 0x2252, 0x2252, 0x23FE, 0x202}},
    {27969, 2, {0x80, 0x2040, 0x17FE, 0x1080, 0x8110, 0x4208, 0x47FC, 0x1004, 0x1000, 0x2248, 0xE248, 0x2248, 0x2248, 0x224A, 0x244A, 0x846}},
    {27970, 2, {0x208, 0x2108, 0x1110, 0x1000, 0x83FC, 0x4040, 0x4840, 0x840, 0x17FE, 0x1040, 0xE040, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x406}},
    {27971, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8040, 0x4248, 0x5150, 0x1040, 0x1FFE, 0x20A0, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {27972, 2, {0x100, 0x2100, 0x13F0, 0x1210, 0x8420, 0x4BFC, 0x4044, 0x1044, 0x17FE, 0x2044, 0xE044, 0x23FC, 0x2044, 0x2040, 0x2140, 0x80}},
    {27973, 2, {0xA0, 0x2090, 0x1080, 0x10B8, 0x87C0, 0x4080, 0x40BC, 0x17C0, 0x1088, 0x2090, 0xE060, 0x2044, 0x20A4, 0x2114, 0x260C, 0x4}},
    {27974, 2, {0x840, 0x4840, 0x28FC, 0x904, 0x1A88, 0x2850, 0x4860, 0x8980, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200}},
    {27975, 2, {0x80, 0x2080, 0x10BC, 0x13C0, 0x8050, 0x4024, 0x48D4, 0xB0C, 0x1000, 0x17FE, 0xE090, 0x2090, 0x2112, 0x2112, 0x220E, 0x400}},
    {27976, 2, {0x40, 0x2040, 0x107E, 0x1040, 0x8040, 0x43F8, 0x4A08, 0xA48, 0x1248, 0x1248, 0xE248, 0x2248, 0x20A0, 0x2110, 0x2208, 0x404}},
    {27977, 2, {0x100, 0x21FE, 0x1110, 0x1510, 0x857C, 0x4554, 0x4554, 0x1554, 0x1554, 0x2554, 0xE554, 0x2154, 0x225C, 0x2210, 0x2410, 0x810}},
    {27978, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8444, 0x4444, 0x4444, 0x1444, 0x17FC, 0x2444, 0xE040, 0x2048, 0x2044, 0x2FFE, 0x2402, 0x0}},
    {27979, 2, {0x4, 0x27C4, 0x1444, 0x1454, 0x8554, 0x4554, 0x4554, 0x1554, 0x1554, 0x2554, 0xE554, 0x2104, 0x2284, 0x2244, 0x2414, 0x808}},
    {27980, 2, {0x120, 0x2128, 0x1124, 0x1224, 0x8220, 0x463E, 0x4BE0, 0x1224, 0x1224, 0x2228, 0xE228, 0x2210, 0x2212, 0x222A, 0x2246, 0x282}},
    {27981, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x4406, 0x41F0, 0x1000, 0x1000, 0x27FC, 0xE040, 0x2080, 0x2110, 0x2208, 0x27FC, 0x204}},
    {27982, 2, {0x80, 0x2040, 0x17FE, 0x1208, 0x8110, 0x40A0, 0x4040, 0x11B0, 0x160E, 0x2110, 0xE110, 0x2110, 0x2110, 0x2210, 0x2210, 0x410}},
    {27983, 2, {0x204, 0x4104, 0x2104, 0x2FE4, 0x94, 0x8494, 0x4294, 0x5294, 0x1114, 0x2114, 0xE294, 0x2294, 0x2444, 0x2844, 0x3014, 0x8}},
    {27984, 2, {0x40, 0x2020, 0x13FC, 0x1000, 0x8108, 0x4090, 0x4BFE, 0xA00, 0x1200, 0x1200, 0xE200, 0x2200, 0x2200, 0x2400, 0x2400, 0x800}},
    {27985, 2, {0x0, 0x23FE, 0x1202, 0x1444, 0x8040, 0x43FC, 0x4880, 0x8A0, 0x1120, 0x11FC, 0xE020, 0x2020, 0x27FE, 0x2020, 0x2020, 0x20}},
    {27986, 2, {0x40, 0x4440, 0x2240, 0x227C, 0x50, 0x8090, 0x4E10, 0x5210, 0x12FE, 0x2210, 0xE210, 0x2290, 0x2310, 0x2210, 0x2010, 0x10}},
    {27987, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8484, 0x4888, 0x4140, 0x1144, 0x1348, 0x2530, 0xE920, 0x2110, 0x2108, 0x2144, 0x2182, 0x100}},
    {27988, 2, {0x0, 0x27F8, 0x1008, 0x1008, 0x83F8, 0x4008, 0x4008, 0x17F8, 0x1010, 0x2010, 0xEFFE, 0x2210, 0x2110, 0x2110, 0x2050, 0x20}},
    {27989, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4220, 0x4A20, 0xA10, 0x1208, 0x14C6, 0xE420, 0x2800, 0x21C0, 0x2030, 0x2008, 0x0}},
    {27990, 2, {0x10, 0x2078, 0x1780, 0x1088, 0x8448, 0x4250, 0x4200, 0x1020, 0x17FC, 0x2020, 0xE220, 0x2120, 0x2120, 0x2020, 0x20A0, 0x40}},
    {27991, 2, {0x50, 0x2048, 0x1048, 0x1040, 0x87FE, 0x4040, 0x4A44, 0x964, 0x1168, 0x10D0, 0xE150, 0x2248, 0x2444, 0x2042, 0x2140, 0x80}},
    {27992, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4200, 0x4A18, 0xAE0, 0x1238, 0x12E0, 0xE23C, 0x23E0, 0x2222, 0x2422, 0x241E, 0x800}},
    {27993, 2, {0x208, 0x421C, 0x2260, 0x2240, 0xFC0, 0x8240, 0x427E, 0x52C8, 0x1348, 0x2E48, 0xE248, 0x2248, 0x2248, 0x2248, 0x2A88, 0x508}},
    {27994, 2, {0x80, 0x2080, 0x1110, 0x1208, 0x87FC, 0x4004, 0x4110, 0x1288, 0x1484, 0x21F0, 0xE310, 0x24A0, 0x2040, 0x20A0, 0x2310, 0xC0C}},
    {27995, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8288, 0x4446, 0x41F0, 0x1010, 0x1020, 0x2040, 0xE3F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {27996, 2, {0x10, 0x2038, 0x13C0, 0x1200, 0x8200, 0x43FC, 0x4220, 0x1220, 0x1220, 0x2220, 0xEFFE, 0x2000, 0x2120, 0x2210, 0x2408, 0x804}},
    {27997, 2, {0x110, 0x2108, 0x1100, 0x17FE, 0x8148, 0x4150, 0x4964, 0x948, 0x1150, 0x1264, 0xE248, 0x2252, 0x2462, 0x2442, 0x29BE, 0x1000}},
    {27998, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x4208, 0x43F8, 0x1040, 0x1040, 0x2240, 0xE27C, 0x2240, 0x2240, 0x2540, 0x28FE, 0x1000}},
    {27999, 2, {0x210, 0x2210, 0x1210, 0x1420, 0x853E, 0x4D44, 0x4DA4, 0x1524, 0x1528, 0x2528, 0xE510, 0x2510, 0x2428, 0x2428, 0x2444, 0x482}},
    {28000, 2, {0x208, 0x2110, 0x10E0, 0x1110, 0x8208, 0x4080, 0x4FFE, 0x1140, 0x1240, 0x27FC, 0xEA44, 0x2244, 0x2254, 0x2248, 0x2040, 0x40}},
    {28001, 2, {0x40, 0x2040, 0x13FC, 0x1040, 0x8040, 0x47FE, 0x5402, 0x1804, 0x11F0, 0x2010, 0xE020, 0x2FFE, 0x2040, 0x2040, 0x2140, 0x80}},
    {28002, 2, {0x0, 0x27FC, 0x1000, 0x1000, 0x83F8, 0x4208, 0x4A08, 0xA08, 0x1208, 0x13F8, 0xE000, 0x2208, 0x2108, 0x2110, 0x2FFE, 0x0}},
    {28003, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8808, 0x43F0, 0x4000, 0x1000, 0x17FC, 0x2120, 0xE120, 0x2120, 0x2224, 0x2224, 0x241C, 0x800}},
    {28004, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8444, 0x4040, 0x4BFE, 0x840, 0x1090, 0x1090, 0xE110, 0x2120, 0x2224, 0x2242, 0x24FE, 0x842}},
    {28005, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4000, 0x47FC, 0x1444, 0x1444, 0x2444, 0xE7FC, 0x2400, 0x2402, 0x2402, 0x23FE, 0x0}},
    {28006, 2, {0x48, 0x4044, 0x2FFE, 0x2040, 0x40, 0x87FC, 0x4444, 0x5444, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x2454, 0x408}},
    {28007, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x43F8, 0x4800, 0x800, 0x17FC, 0x1040, 0xE040, 0x23F8, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28008, 2, {0x80, 0x2040, 0x17FE, 0x1402, 0x8844, 0x4040, 0x4040, 0x17FC, 0x10E0, 0x2150, 0xE150, 0x2248, 0x2444, 0x2842, 0x2040, 0x40}},
    {28009, 2, {0x40, 0x2240, 0x1240, 0x13FC, 0x8440, 0x4840, 0x4040, 0x17FE, 0x1000, 0x2000, 0xE3F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28010, 2, {0x80, 0x2040, 0x13F8, 0x1208, 0x8208, 0x43F8, 0x4A08, 0xA08, 0x13F8, 0x1240, 0xE244, 0x2228, 0x2210, 0x2288, 0x2306, 0x200}},
    {28011, 2, {0x0, 0x47FE, 0x2402, 0x2914, 0x208, 0x8404, 0x43F8, 0x5040, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {28012, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x87FC, 0x4444, 0x4444, 0x17FC, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28013, 2, {0x0, 0x2FFE, 0x1040, 0x17FC, 0x8444, 0x4444, 0x47FC, 0x1444, 0x1444, 0x27FC, 0xE240, 0x2140, 0x2080, 0x2140, 0x2230, 0xC0E}},
    {28014, 2, {0x8, 0x203C, 0x13C0, 0x1004, 0x8244, 0x4128, 0x4800, 0xBF8, 0x1010, 0x1020, 0xE7FE, 0x2020, 0x2020, 0x2020, 0x20A0, 0x40}},
    {28015, 2, {0x0, 0x27FC, 0x1040, 0x1040, 0x83F8, 0x4088, 0x4088, 0x17FE, 0x1000, 0x2000, 0xE3F8, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28016, 2, {0x84, 0x21C4, 0x1704, 0x1114, 0x8114, 0x4114, 0x4FD4, 0x1114, 0x1314, 0x2394, 0xE554, 0x2504, 0x2904, 0x2104, 0x2114, 0x108}},
    {28017, 2, {0x0, 0x23FC, 0x1200, 0x1200, 0x82F8, 0x4200, 0x4A00, 0xBFC, 0x12A0, 0x12A4, 0xE2A8, 0x2290, 0x2290, 0x2488, 0x24A4, 0x8C2}},
    {28018, 2, {0x100, 0x21F8, 0x1308, 0x1490, 0x8060, 0x4198, 0x4646, 0x1040, 0x13F8, 0x2040, 0xE3F8, 0x2040, 0x27FC, 0x2040, 0x2040, 0x40}},
    {28019, 2, {0x88, 0x2084, 0x10BE, 0x17C0, 0x8032, 0x400E, 0x4800, 0xBFC, 0x1204, 0x13FC, 0xE204, 0x23FC, 0x2204, 0x2204, 0x2214, 0x208}},
    {28020, 2, {0x110, 0x2108, 0x1204, 0x1444, 0x8040, 0x40A0, 0x4110, 0x1208, 0x1406, 0x23F8, 0xE208, 0x2208, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28021, 2, {0x2, 0x2782, 0x1484, 0x1488, 0x8690, 0x4582, 0x4482, 0x1484, 0x1FC8, 0x2490, 0xE482, 0x2482, 0x2484, 0x2488, 0x2890, 0x11A0}},
    {28022, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8424, 0x4120, 0x49FC, 0x920, 0x1220, 0x1020, 0xE7FE, 0x2020, 0x2020, 0x2020, 0x2020, 0x20}},
    {28023, 2, {0x100, 0x2100, 0x11FC, 0x1200, 0x85F8, 0x4108, 0x4948, 0x928, 0x17FE, 0x1108, 0xE248, 0x2228, 0x23FC, 0x2008, 0x2050, 0x20}},
    {28024, 2, {0x0, 0x23F8, 0x1008, 0x11F8, 0x8008, 0x43F8, 0x4000, 0x17FC, 0x1404, 0x23F0, 0xE110, 0x2110, 0x20A0, 0x2040, 0x21B0, 0x60E}},
    {28025, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x8040, 0x4248, 0x4248, 0x1248, 0x1554, 0x28A2, 0xE0A0, 0x2110, 0x2110, 0x2208, 0x2404, 0x802}},
    {28026, 2, {0x108, 0x2108, 0x1108, 0x113E, 0x81AA, 0x456A, 0x452A, 0x152A, 0x192A, 0x213E, 0xE12A, 0x2108, 0x2108, 0x2108, 0x2108, 0x108}},
    {28027, 2, {0x0, 0x27FC, 0x1404, 0x14A4, 0x8514, 0x460C, 0x5404, 0x15F4, 0x1514, 0x2514, 0xE514, 0x25F4, 0x2514, 0x2404, 0x2414, 0x408}},
    {28028, 2, {0x100, 0x2100, 0x11F0, 0x1210, 0x8220, 0x47FC, 0x4A44, 0x1244, 0x1244, 0x23FC, 0xE2A4, 0x20A0, 0x2120, 0x2222, 0x2422, 0x81E}},
    {28029, 2, {0x8, 0x203C, 0x13C0, 0x1044, 0x8224, 0x4128, 0x4900, 0x840, 0x13FE, 0x1088, 0xE108, 0x2390, 0x2060, 0x2050, 0x2188, 0x604}},
    {28030, 2, {0x40, 0x2040, 0x13FC, 0x1040, 0x8040, 0x4040, 0x47FE, 0x1090, 0x1294, 0x2292, 0xE492, 0x2912, 0x2110, 0x2210, 0x2450, 0x820}},
    {28031, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4204, 0x4A04, 0xBFC, 0x1204, 0x1204, 0xE3FC, 0x2000, 0x2110, 0x2108, 0x2204, 0x404}},
    {28032, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4208, 0x4208, 0x13F8, 0x1208, 0x2208, 0xE3F8, 0x2120, 0x2120, 0x2222, 0x2422, 0x181E}},
    {28033, 2, {0x102, 0x2102, 0x1104, 0x1108, 0x8FD0, 0x4102, 0x4302, 0x1384, 0x1548, 0x2550, 0xE902, 0x2102, 0x2104, 0x2108, 0x2110, 0x160}},
    {28034, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x45F6, 0x4040, 0x1040, 0x17FC, 0x2040, 0xE250, 0x2248, 0x2444, 0x2844, 0x2140, 0x80}},
    {28035, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x8444, 0x47FC, 0x4444, 0x14C4, 0x14E4, 0x2554, 0xE644, 0x2444, 0x2444, 0x2404, 0x27FC, 0x404}},
    {28036, 2, {0x40, 0x2040, 0x17FC, 0x1444, 0x87FC, 0x4444, 0x47FC, 0x1000, 0x1FFE, 0x2100, 0xE200, 0x23FC, 0x2004, 0x2004, 0x2028, 0x10}},
    {28037, 2, {0x0, 0x27FC, 0x1404, 0x1404, 0x87FC, 0x4404, 0x4404, 0x17FC, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28038, 2, {0x0, 0x203E, 0x1788, 0x1488, 0x8488, 0x4488, 0x4488, 0x17BE, 0x1488, 0x2488, 0xE488, 0x2488, 0x2788, 0x2488, 0x2008, 0x8}},
    {28039, 2, {0x0, 0x23FC, 0x1000, 0x1124, 0x8248, 0x4490, 0x5248, 0x1124, 0x1000, 0x23FC, 0xE040, 0x2040, 0x2040, 0x2040, 0x27FE, 0x0}},
    {28040, 2, {0x40, 0x2248, 0x1148, 0x1150, 0x8040, 0x43F8, 0x4A08, 0xA08, 0x13F8, 0x1208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {28041, 2, {0x20, 0x2020, 0x1120, 0x113C, 0x8120, 0x4120, 0x47FE, 0x1000, 0x1020, 0x2124, 0xE124, 0x2228, 0x2410, 0x2020, 0x20C0, 0x700}},
    {28042, 2, {0x0, 0x23FC, 0x1044, 0x1144, 0x8144, 0x4284, 0x4894, 0x908, 0x1240, 0x1020, 0xE0A4, 0x2282, 0x228A, 0x228A, 0x2478, 0x0}},
    {28043, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8404, 0x4050, 0x4048, 0x1040, 0x17FE, 0x2040, 0xE0A0, 0x20A0, 0x2110, 0x2110, 0x2208, 0x406}},
    {28044, 2, {0x0, 0x27F8, 0x1010, 0x11A0, 0x8040, 0x47FC, 0x4444, 0x1444, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x2454, 0x408}},
    {28045, 2, {0x40, 0x2044, 0x13F4, 0x1048, 0x8048, 0x47FE, 0x4020, 0x1040, 0x13F8, 0x2110, 0xE220, 0x25FE, 0x2820, 0x2020, 0x20A0, 0x40}},
    {28046, 2, {0x8, 0x201C, 0x1770, 0x1110, 0x8110, 0x4250, 0x425C, 0x1750, 0x1150, 0x2550, 0xE550, 0x227C, 0x2200, 0x2500, 0x28FE, 0x1000}},
    {28047, 2, {0x0, 0x200C, 0x1770, 0x1110, 0x8110, 0x4210, 0x427C, 0x1710, 0x1110, 0x2510, 0xE510, 0x227C, 0x2200, 0x2500, 0x28FE, 0x1000}},
    {28048, 2, {0x120, 0x43A8, 0x2E24, 0x2224, 0x220, 0x9FFE, 0x4220, 0x5224, 0x12A4, 0x2328, 0xE228, 0x2E10, 0x2232, 0x224A, 0x2A86, 0x402}},
    {28049, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x40, 0x87FC, 0x4444, 0x5444, 0x17FC, 0x2040, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {28050, 2, {0x0, 0x23F8, 0x1088, 0x1088, 0x87FE, 0x4088, 0x4888, 0xBF8, 0x1100, 0x1100, 0xE3F8, 0x2508, 0x2908, 0x2108, 0x21F8, 0x108}},
    {28051, 2, {0x0, 0x21F8, 0x1108, 0x1108, 0x81F8, 0x4000, 0x4BFC, 0xA04, 0x1204, 0x13FC, 0xE204, 0x2204, 0x23FC, 0x2204, 0x2214, 0x208}},
    {28052, 2, {0x40, 0x2248, 0x1248, 0x1248, 0x83F8, 0x4040, 0x48A0, 0x910, 0x1288, 0x1446, 0xE040, 0x23F0, 0x2010, 0x2020, 0x2020, 0x40}},
    {28053, 2, {0x210, 0x2110, 0x1120, 0x17F8, 0x8048, 0x4048, 0x47F8, 0x1440, 0x1440, 0x27FC, 0xE0C4, 0x2144, 0x2254, 0x2448, 0x2840, 0x40}},
    {28054, 2, {0x120, 0x2110, 0x1110, 0x1200, 0x82FE, 0x4600, 0x5604, 0x1A84, 0x1284, 0x2248, 0xE248, 0x2248, 0x2250, 0x2210, 0x22FE, 0x200}},
    {28055, 2, {0x10, 0x2090, 0x1088, 0x1108, 0x8204, 0x45FA, 0x4108, 0x1108, 0x1108, 0x21F8, 0xE090, 0x2090, 0x2090, 0x2112, 0x2212, 0x40E}},
    {28056, 2, {0x40, 0x2080, 0x1110, 0x1208, 0x87FC, 0x4104, 0x4100, 0x13F8, 0x1440, 0x2040, 0xEFFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28057, 2, {0x40, 0x2020, 0x13FC, 0x1204, 0x8204, 0x43FC, 0x4A00, 0xA20, 0x1220, 0x13FE, 0xE220, 0x2250, 0x2250, 0x2488, 0x2504, 0xA02}},
    {28058, 2, {0x208, 0x2108, 0x1110, 0x1020, 0x83F8, 0x4208, 0x4A08, 0xA08, 0x13F8, 0x10A0, 0xE0A0, 0x2120, 0x2122, 0x2222, 0x241E, 0x800}},
    {28059, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FE, 0x1090, 0x1110, 0x21FE, 0xE210, 0x2290, 0x2450, 0x2810, 0x2050, 0x20}},
    {28060, 2, {0x20, 0x2020, 0x13FE, 0x1020, 0x81FC, 0x4000, 0x4BFE, 0xA02, 0x1404, 0x1090, 0xE090, 0x2090, 0x2090, 0x2112, 0x2112, 0x20E}},
    {28061, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8000, 0x47FE, 0x5402, 0x1884, 0x1080, 0x27FC, 0xE084, 0x2104, 0x2104, 0x2204, 0x2428, 0x810}},
    {28062, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8040, 0x4248, 0x4148, 0x1150, 0x17FE, 0x20E0, 0xE150, 0x2150, 0x2248, 0x2444, 0x2842, 0x40}},
    {28063, 2, {0x20, 0x2420, 0x1220, 0x12FE, 0x8040, 0x4050, 0x4E90, 0x12FE, 0x1210, 0x2210, 0xE3FE, 0x2210, 0x2210, 0x2510, 0x28FE, 0x0}},
    {28064, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x85F4, 0x4444, 0x44E4, 0x1444, 0x15F4, 0x2454, 0xE454, 0x2474, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28065, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x8208, 0x43F8, 0x4040, 0x1040, 0x17FC, 0x2444, 0xE4A4, 0x2514, 0x2614, 0x2404, 0x2414, 0x408}},
    {28066, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4000, 0x4000, 0x17FC, 0x1404, 0x2444, 0xE444, 0x2444, 0x24B4, 0x2108, 0x2604, 0x1802}},
    {28067, 2, {0x100, 0x2100, 0x11F0, 0x1210, 0x8420, 0x4BF8, 0x4248, 0x1248, 0x1248, 0x2248, 0xEFFE, 0x20A0, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28068, 2, {0x80, 0x2080, 0x11F8, 0x1308, 0x8490, 0x4060, 0x4198, 0x1646, 0x1040, 0x27FC, 0xE040, 0x2248, 0x2244, 0x2444, 0x2140, 0x80}},
    {28069, 2, {0x80, 0x4040, 0x2FFE, 0x2000, 0x3F8, 0x8208, 0x4208, 0x53F8, 0x1000, 0x23F8, 0xE010, 0x2060, 0x2040, 0x2040, 0x2140, 0x80}},
    {28070, 2, {0x200, 0x217C, 0x1504, 0x1404, 0x85F4, 0x4444, 0x4444, 0x1444, 0x15F4, 0x2444, 0xE444, 0x2444, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28071, 2, {0x200, 0x217C, 0x1504, 0x1404, 0x8404, 0x45F4, 0x4514, 0x1514, 0x15F4, 0x2514, 0xE514, 0x25F4, 0x2404, 0x2404, 0x2414, 0x408}},
    {28072, 2, {0x40, 0x4F44, 0x2144, 0x2148, 0x148, 0x8F50, 0x4840, 0x58FE, 0x1850, 0x2F50, 0xE148, 0x2148, 0x2144, 0x2144, 0x2A62, 0x440}},
    {28073, 2, {0x0, 0x23FC, 0x1044, 0x1244, 0x8244, 0x4484, 0x4094, 0x1108, 0x1220, 0x2020, 0xE13C, 0x2120, 0x2120, 0x2120, 0x27FE, 0x0}},
    {28074, 2, {0x80, 0x2040, 0x13FC, 0x1000, 0x8208, 0x4108, 0x4110, 0x17FE, 0x1000, 0x2000, 0xE3FC, 0x2204, 0x2204, 0x2204, 0x23FC, 0x204}},
    {28075, 2, {0x80, 0x4040, 0x27FE, 0x2402, 0x804, 0x83F8, 0x4208, 0x5208, 0x13F8, 0x2200, 0xE200, 0x23FC, 0x2204, 0x2204, 0x23FC, 0x204}},
    {28076, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8040, 0x4FFE, 0x4110, 0x10A0, 0x17FC, 0x2040, 0xE040, 0x2FFE, 0x2040, 0x2040, 0x2040, 0x40}},
    {28077, 2, {0x1C, 0x23E0, 0x1084, 0x1244, 0x8148, 0x4110, 0x57FE, 0x1402, 0x1000, 0x23F8, 0xE108, 0x2110, 0x20A0, 0x2040, 0x21B0, 0x60E}},
    {28078, 2, {0x2, 0x47E2, 0x2422, 0x242A, 0x7EA, 0x848A, 0x448A, 0x548A, 0x17EA, 0x26AA, 0xE6AA, 0x2AAA, 0x2AA2, 0x32E2, 0x208A, 0x84}},
    {28079, 2, {0x0, 0x27FE, 0x1420, 0x1420, 0x85FC, 0x4420, 0x4420, 0x17FE, 0x1400, 0x2420, 0xE420, 0x25FC, 0x2420, 0x2820, 0x2BFE, 0x1000}},
    {28080, 2, {0x0, 0x27BC, 0x1084, 0x1294, 0x8108, 0x4294, 0x54A4, 0x1840, 0x1000, 0x27BC, 0xE0A4, 0x22A4, 0x2128, 0x2290, 0x24A8, 0x846}},
    {28081, 2, {0x0, 0x21FC, 0x1100, 0x11F8, 0x8100, 0x41F8, 0x5100, 0x1FFE, 0x1280, 0x2248, 0xE250, 0x2220, 0x2210, 0x2288, 0x2306, 0x200}},
    {28082, 2, {0x80, 0x2040, 0x17FE, 0x1120, 0x8120, 0x423C, 0x4244, 0x1664, 0x1294, 0x2348, 0xE248, 0x2230, 0x2220, 0x2250, 0x2288, 0x306}},
    {28083, 2, {0x40, 0x2020, 0x1020, 0x13FE, 0x8202, 0x4494, 0x4108, 0x1204, 0x1000, 0x21FC, 0xE020, 0x2020, 0x2020, 0x2020, 0x27FE, 0x0}},
    {28084, 2, {0x80, 0x4040, 0x27FE, 0x2402, 0xA04, 0x8200, 0x43BC, 0x54A4, 0x14A4, 0x26A4, 0xE934, 0x2128, 0x2222, 0x2222, 0x2422, 0x81E}},
    {28085, 2, {0x0, 0x23F8, 0x1010, 0x1020, 0x8444, 0x4554, 0x44E4, 0x1444, 0x14E4, 0x2554, 0xE64C, 0x2544, 0x2484, 0x2404, 0x27FC, 0x4}},
    {28086, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x8040, 0x47FC, 0x4248, 0x1248, 0x1FFE, 0x2248, 0xE248, 0x27FC, 0x2040, 0x2040, 0x27FC, 0x0}},
    {28087, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4248, 0x4BF8, 0xA48, 0x13F8, 0x1040, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {28088, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x87FC, 0x4444, 0x4444, 0x15F4, 0x1514, 0x2514, 0xE514, 0x25F4, 0x2514, 0x2404, 0x27FC, 0x404}},
    {28089, 2, {0x38, 0x23C0, 0x1040, 0x1040, 0x87FC, 0x4150, 0x4248, 0x1486, 0x1080, 0x27FC, 0xE110, 0x2210, 0x21A0, 0x2060, 0x2090, 0x308}},
    {28090, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4220, 0x4A20, 0xBFE, 0x1220, 0x1220, 0xE2FC, 0x2284, 0x2284, 0x2484, 0x24FC, 0x884}},
    {28091, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x45F6, 0x4040, 0x1040, 0x17FC, 0x2040, 0xE040, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28092, 2, {0x80, 0x2040, 0x17FE, 0x1000, 0x8000, 0x43F8, 0x4A08, 0xA08, 0x13F8, 0x1040, 0xE250, 0x2248, 0x2444, 0x2844, 0x2140, 0x80}},
    {28093, 2, {0x1C, 0x23E0, 0x1220, 0x13FE, 0x8220, 0x4292, 0x4B0A, 0xA06, 0x11FC, 0x1104, 0xE104, 0x21FC, 0x2104, 0x2104, 0x21FC, 0x104}},
    {28094, 2, {0x40, 0x2044, 0x1768, 0x1150, 0x8150, 0x4248, 0x4544, 0x1080, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28095, 2, {0x0, 0x4FFE, 0x2080, 0x2184, 0x244, 0x8468, 0x4AB0, 0x4130, 0x12A8, 0x2468, 0xE8A4, 0x2124, 0x2222, 0x2420, 0x20A0, 0x40}},
    {28096, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8404, 0x4000, 0x43FE, 0x1020, 0x1020, 0x2120, 0xE13C, 0x2120, 0x22A0, 0x2260, 0x243E, 0x800}},
    {28097, 2, {0x80, 0x2040, 0x13FC, 0x1000, 0x8108, 0x4090, 0x47FE, 0x1040, 0x1040, 0x27FE, 0xE088, 0x2108, 0x2090, 0x2060, 0x2198, 0x604}},
    {28098, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4800, 0xBFC, 0x1010, 0x17FE, 0xE010, 0x2210, 0x2110, 0x2010, 0x2050, 0x20}},
    {28099, 2, {0x40, 0x2248, 0x1150, 0x17FC, 0x8080, 0x4100, 0x4FFE, 0x1210, 0x1408, 0x2BF4, 0xF212, 0x2210, 0x2250, 0x2224, 0x2204, 0x1FC}},
    {28100, 2, {0x124, 0x2124, 0x1248, 0x1490, 0x8248, 0x4124, 0x4124, 0x1000, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28101, 2, {0x200, 0x420C, 0x2270, 0x2240, 0xF40, 0x8240, 0x427E, 0x5648, 0x1748, 0x2AC8, 0xEA48, 0x3248, 0x2248, 0x2248, 0x2248, 0x288}},
    {28102, 2, {0x210, 0x2120, 0x10C0, 0x1330, 0x8C88, 0x4080, 0x4FFE, 0x1100, 0x13F8, 0x2608, 0xEBF8, 0x2208, 0x23F8, 0x2208, 0x2228, 0x210}},
    {28103, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8108, 0x41F8, 0x4908, 0x908, 0x11F8, 0x1108, 0xE108, 0x27FE, 0x2000, 0x2090, 0x2108, 0x204}},
    {28104, 2, {0x0, 0x23FE, 0x1202, 0x1202, 0x83FE, 0x4210, 0x4A92, 0xA92, 0x1292, 0x12FE, 0xE210, 0x2292, 0x2292, 0x2492, 0x24FE, 0x802}},
    {28105, 2, {0x0, 0x23F8, 0x1248, 0x1248, 0x83F8, 0x4248, 0x4A48, 0xBF8, 0x1040, 0x17FC, 0xE0E0, 0x2150, 0x2248, 0x2446, 0x2040, 0x40}},
    {28106, 2, {0x100, 0x2100, 0x11F8, 0x1208, 0x8410, 0x4820, 0x4180, 0x163C, 0x1404, 0x2404, 0xE7BC, 0x2404, 0x2404, 0x2404, 0x27FC, 0x404}},
    {28107, 2, {0x110, 0x2110, 0x1110, 0x1110, 0x87BC, 0x4110, 0x4110, 0x1338, 0x13B8, 0x2554, 0xE554, 0x2992, 0x2110, 0x2110, 0x2110, 0x110}},
    {28108, 2, {0x40, 0x2248, 0x1148, 0x1150, 0x8040, 0x47FC, 0x4404, 0x1404, 0x15F4, 0x2514, 0xE514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28109, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x85F4, 0x4444, 0x4444, 0x17FC, 0x1404, 0x25F4, 0xE514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x808}},
    {28110, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FC, 0x1110, 0x1248, 0x2446, 0xE1F0, 0x2040, 0x27FC, 0x2040, 0x2040, 0x40}},
    {28111, 2, {0x0, 0x23FC, 0x1204, 0x13FC, 0x8204, 0x43FC, 0x4800, 0xBFC, 0x1040, 0x1040, 0xE7FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x406}},
    {28112, 2, {0x3F8, 0x2208, 0x1208, 0x13F8, 0x8208, 0x4208, 0x43F8, 0x1000, 0x17FC, 0x2404, 0xE404, 0x27FC, 0x2404, 0x2404, 0x27FC, 0x404}},
    {28113, 2, {0x100, 0x2100, 0x11DC, 0x1114, 0x8114, 0x4114, 0x47D4, 0x1014, 0x1114, 0x2114, 0xE588, 0x2548, 0x2948, 0x2114, 0x2514, 0x222}},
    {28114, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4048, 0x47FE, 0x1048, 0x13F8, 0x2080, 0xE7FE, 0x2108, 0x2390, 0x2060, 0x2198, 0x604}},
    {28115, 2, {0x208, 0x4208, 0x2FFE, 0x2208, 0x288, 0x8040, 0x4FFE, 0x4100, 0x1100, 0x21F8, 0xE108, 0x2108, 0x2208, 0x2208, 0x2428, 0x810}},
    {28116, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8040, 0x43F8, 0x4A08, 0xBF8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2FFE, 0x0}},
    {28117, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x8040, 0x47FC, 0x4110, 0x1208, 0x1444, 0x2040, 0xE3F8, 0x2040, 0x2040, 0x2040, 0x27FC, 0x0}},
    {28118, 2, {0x40, 0x2040, 0x107E, 0x1040, 0x83FC, 0x4204, 0x4BFC, 0xA04, 0x13FC, 0x1244, 0xE040, 0x27FE, 0x2040, 0x2040, 0x2040, 0x40}},
    {28119, 2, {0x100, 0x2100, 0x11FC, 0x1204, 0x8484, 0x4494, 0x42A4, 0x1084, 0x17FC, 0x2084, 0xE1C4, 0x22A4, 0x2494, 0x2084, 0x20A8, 0x10}},
    {28120, 2, {0x100, 0x2100, 0x11FC, 0x1204, 0x8504, 0x41E4, 0x4284, 0x1084, 0x17F4, 0x2084, 0xE2A4, 0x22A4, 0x23E4, 0x2004, 0x2028, 0x10}},
    {28121, 2, {0x80, 0x4040, 0x27FE, 0x2402, 0x804, 0x83F0, 0x4000, 0x5000, 0x17FC, 0x2040, 0xE250, 0x2248, 0x2444, 0x2844, 0x2140, 0x80}},
    {28122, 2, {0x40, 0x2020, 0x13FC, 0x1204, 0x8204, 0x43FC, 0x4A00, 0xA28, 0x1224, 0x13FE, 0xE220, 0x2250, 0x2250, 0x2488, 0x2504, 0xA02}},
    {28123, 2, {0x104, 0x2104, 0x1504, 0x17D4, 0x8914, 0x4114, 0x4FF4, 0x1114, 0x17D4, 0x2554, 0xE554, 0x2554, 0x2544, 0x25C4, 0x2114, 0x108}},
    {28124, 2, {0x0, 0x27BC, 0x14A4, 0x14A4, 0x84A4, 0x47BC, 0x44A4, 0x14A4, 0x14A4, 0x27BC, 0xE4A4, 0x24A4, 0x24A4, 0x24A4, 0x2AA4, 0x114C}},
    {28125, 2, {0x0, 0x27BE, 0x14AA, 0x14AA, 0x84AA, 0x47AA, 0x44AA, 0x14AA, 0x14BE, 0x27A0, 0xE4A0, 0x24A0, 0x24A2, 0x28A2, 0x2AA2, 0x111E}},
    {28126, 2, {0x208, 0x4228, 0x2228, 0x2228, 0xFA4, 0x8244, 0x4254, 0x5692, 0x1710, 0x2AA0, 0xEA20, 0x3228, 0x2244, 0x22FC, 0x2244, 0x200}},
    {28127, 2, {0x90, 0x2090, 0x1090, 0x13FC, 0x8294, 0x4294, 0x4A94, 0xBFC, 0x1294, 0x1294, 0xE294, 0x27FE, 0x2000, 0x2090, 0x2108, 0x204}},
    {28128, 2, {0x0, 0x27FC, 0x1444, 0x1444, 0x87FC, 0x4444, 0x4444, 0x17FC, 0x1000, 0x2000, 0xEFFE, 0x2208, 0x2208, 0x2208, 0x2408, 0x808}},
    {28129, 2, {0x40, 0x2244, 0x1244, 0x1448, 0x80A0, 0x4110, 0x4608, 0x1044, 0x1040, 0x2248, 0xE248, 0x2450, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28130, 2, {0x14, 0x2012, 0x1010, 0x17FE, 0x8010, 0x4010, 0x4BD2, 0xA52, 0x1252, 0x1254, 0xE3D4, 0x2008, 0x20EA, 0x271A, 0x2226, 0x42}},
    {28131, 2, {0x0, 0x2040, 0x139C, 0x1204, 0x8204, 0x439C, 0x4A04, 0xA04, 0x13FC, 0x1090, 0xE090, 0x2090, 0x2090, 0x2112, 0x2212, 0x40E}},
    {28132, 2, {0x210, 0x4110, 0x2010, 0x2FA8, 0x428, 0x8444, 0x4402, 0x5790, 0x1488, 0x2488, 0xE480, 0x24A0, 0x2490, 0x2888, 0x2A88, 0x1100}},
    {28133, 2, {0x80, 0x20F8, 0x1088, 0x1108, 0x81F0, 0x4010, 0x4810, 0xBFE, 0x1020, 0x1222, 0xE174, 0x20A8, 0x2124, 0x2222, 0x20A0, 0x40}},
    {28134, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x4406, 0x41F0, 0x1040, 0x1040, 0x23F8, 0xE040, 0x2248, 0x2148, 0x2150, 0x2FFE, 0x0}},
    {28135, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8444, 0x4020, 0x4828, 0x888, 0x1090, 0x12A4, 0xE2C2, 0x2482, 0x2188, 0x2288, 0x2478, 0x0}},
    {28136, 2, {0x3C, 0x27C0, 0x1244, 0x1128, 0x8000, 0x43F8, 0x4848, 0x848, 0x17FE, 0x1048, 0xE048, 0x23F8, 0x2048, 0x2040, 0x2140, 0x80}},
    {28137, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8040, 0x4FFE, 0x4208, 0x1484, 0x1882, 0x21F8, 0xE208, 0x2510, 0x20A0, 0x2040, 0x21B0, 0xE0E}},
    {28138, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x4DF6, 0x4000, 0x17FC, 0x14A4, 0x24A4, 0xE7FC, 0x24A4, 0x24A4, 0x24A4, 0x2414, 0x408}},
    {28139, 2, {0x8, 0x203C, 0x17C0, 0x1084, 0x8444, 0x4248, 0x4210, 0x103C, 0x17C0, 0x2040, 0xE040, 0x2FFE, 0x2040, 0x2040, 0x27FC, 0x0}},
    {28140, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x8110, 0x4110, 0x42A8, 0x1444, 0x1000, 0x2040, 0xE7FE, 0x2040, 0x2040, 0x2040, 0x2040, 0x40}},
    {28141, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8404, 0x47FC, 0x4440, 0x1440, 0x17FE, 0x24E0, 0xE950, 0x2950, 0x3248, 0x2444, 0x2842, 0x40}},
    {28142, 2, {0x140, 0x2120, 0x1120, 0x13FE, 0x8220, 0x4620, 0x4BFC, 0x1220, 0x1220, 0x23FC, 0xE220, 0x2220, 0x2220, 0x23FE, 0x2200, 0x200}},
    {28143, 2, {0x80, 0x2040, 0x17FE, 0x1100, 0x8208, 0x47FC, 0x4004, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {28144, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8288, 0x4446, 0x4000, 0x13F0, 0x1010, 0x2020, 0xE040, 0x2124, 0x2522, 0x250A, 0x2908, 0xF8}},
    {28145, 2, {0x0, 0x27FC, 0x1404, 0x14A4, 0x8110, 0x4208, 0x4040, 0x1040, 0x17FC, 0x2040, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {28146, 2, {0x40, 0x2040, 0x107C, 0x1040, 0x83FC, 0x4244, 0x4A70, 0xBC0, 0x1244, 0x123C, 0xE200, 0x22F0, 0x2490, 0x2492, 0x2912, 0x120E}},
    {28147, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x83F8, 0x4208, 0x4BF8, 0x800, 0x13F8, 0x1010, 0xE020, 0x27FC, 0x2040, 0x2040, 0x2140, 0x80}},
    {28148, 2, {0x100, 0x2100, 0x13FC, 0x1494, 0x8894, 0x4124, 0x4244, 0x1484, 0x1128, 0x2210, 0xE040, 0x2124, 0x252A, 0x250A, 0x28F8, 0x0}},
    {28149, 2, {0xA0, 0x24A4, 0x14A4, 0x14A4, 0x87BC, 0x4404, 0x4404, 0x17FC, 0x1404, 0x2404, 0xE7BC, 0x24A4, 0x24A4, 0x24A4, 0x28A4, 0x1004}},
    {28150, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x8040, 0x4248, 0x4248, 0x1248, 0x1554, 0x20E0, 0xE150, 0x2150, 0x2248, 0x2444, 0x2842, 0x40}},
    {28151, 2, {0x0, 0x27FC, 0x1404, 0x1404, 0x87FC, 0x4404, 0x4404, 0x17FC, 0x1000, 0x2420, 0xE422, 0x27AC, 0x2430, 0x2422, 0x25A2, 0x61E}},
    {28152, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FE, 0x1000, 0x13F8, 0x2248, 0xE248, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {28153, 2, {0x40, 0x2080, 0x17FE, 0x1110, 0x8248, 0x4446, 0x43F8, 0x1248, 0x1248, 0x23F8, 0xE248, 0x2248, 0x23F8, 0x2042, 0x2042, 0x3E}},
    {28154, 2, {0xA0, 0x2090, 0x10BC, 0x17C8, 0x8050, 0x4064, 0x5194, 0x1E0C, 0x10A0, 0x2090, 0xE0BC, 0x27C8, 0x2050, 0x2064, 0x2194, 0xE0C}},
    {28155, 2, {0x0, 0x23FC, 0x1040, 0x1040, 0x87FE, 0x40A0, 0x4110, 0x1208, 0x1446, 0x2040, 0xE048, 0x2264, 0x2252, 0x2452, 0x2140, 0x80}},
    {28156, 2, {0x100, 0x108, 0x7D10, 0x5A0, 0x960, 0x3118, 0xC506, 0x200, 0x810, 0xA14, 0x6AD4, 0x2C58, 0x2A54, 0x4A94, 0xA952, 0x1020}},
    {28157, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8108, 0x4020, 0x4820, 0x920, 0x1120, 0x113C, 0xE120, 0x2120, 0x2120, 0x2120, 0x2FFE, 0x0}},
    {28158, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE, 0x0, 0x108, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {28159, 2, {0x40, 0x2080, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x4BF8, 0x840, 0x1040, 0x17FC, 0xE444, 0x2444, 0x2454, 0x2448, 0x2040, 0x40}},
    {28160, 2, {0x40, 0x2040, 0x17FE, 0x10A0, 0x8150, 0x4248, 0x45F6, 0x1040, 0x1150, 0x2110, 0xE7FE, 0x2110, 0x2110, 0x2110, 0x2210, 0x410}},
    {28161, 2, {0x40, 0x4744, 0x2168, 0x2150, 0x248, 0x8444, 0x4940, 0x4080, 0x1000, 0x2040, 0xE744, 0x2168, 0x2250, 0x2448, 0x2946, 0x80}},
    {28162, 2, {0x3F8, 0x2208, 0x1208, 0x13F8, 0x8208, 0x4208, 0x4BF8, 0x880, 0x1040, 0x17FC, 0xE110, 0x2110, 0x20A0, 0x2040, 0x21B0, 0xE0E}},
    {28163, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8148, 0x4040, 0x47FE, 0x1080, 0x1100, 0x23FC, 0xE504, 0x2904, 0x2104, 0x2104, 0x21FC, 0x104}},
    {28164, 2, {0x90, 0x2090, 0x1090, 0x179E, 0x8090, 0x4090, 0x4890, 0xB9C, 0x1090, 0x1090, 0xE090, 0x279E, 0x2090, 0x2090, 0x2090, 0x90}},
    {28165, 2, {0x2040, 0x1040, 0x17FC, 0x40, 0x83F8, 0x4040, 0x4FFE, 0x800, 0x13F8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2208, 0x2228, 0x210}},
    {28166, 2, {0x844, 0x4484, 0x2004, 0x2FD4, 0x494, 0x8494, 0x4494, 0x5494, 0x1FF4, 0x2494, 0xE494, 0x2494, 0x2484, 0x2884, 0x2894, 0x1088}},
    {28167, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4900, 0xBFC, 0x1404, 0x1224, 0xE3D4, 0x2214, 0x21F4, 0x2004, 0x2028, 0x10}},
    {28168, 2, {0x40, 0x2020, 0x17FE, 0x1108, 0x8090, 0x4060, 0x4198, 0x1606, 0x1108, 0x21F8, 0xE108, 0x21F8, 0x2108, 0x2108, 0x2208, 0x408}},
    {28169, 2, {0x20, 0x2020, 0x1120, 0x113C, 0x8120, 0x4120, 0x47FE, 0x1000, 0x1028, 0x2124, 0xE122, 0x222A, 0x2410, 0x2020, 0x20C0, 0x700}},
    {28170, 2, {0x444, 0x2444, 0x1654, 0x1554, 0x8564, 0x4444, 0x47FC, 0x1444, 0x14C4, 0x24E4, 0xE554, 0x2554, 0x2644, 0x2844, 0x2844, 0x1044}},
    {28171, 2, {0x20, 0x2020, 0x1120, 0x113C, 0x8120, 0x4120, 0x47FE, 0x1000, 0x1000, 0x2204, 0xE108, 0x2090, 0x2108, 0x2204, 0x2402, 0x0}},
    {28172, 2, {0x0, 0x23F8, 0x1008, 0x1008, 0x83F8, 0x4008, 0x4008, 0x17FE, 0x1040, 0x2444, 0xE2E8, 0x2150, 0x2248, 0x2446, 0x2140, 0x80}},
    {28173, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FE, 0x1000, 0x13F8, 0x2208, 0xE248, 0x2248, 0x2248, 0x20A0, 0x2110, 0x608}},
    {28174, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x8040, 0x47FC, 0x4004, 0x1128, 0x10A0, 0x2220, 0xE120, 0x27FC, 0x2050, 0x2088, 0x2104, 0x604}},
    {28175, 2, {0x40, 0x2040, 0x13FC, 0x10A0, 0x8110, 0x4208, 0x57FE, 0x1008, 0x13C8, 0x2248, 0xE248, 0x2248, 0x23C8, 0x2008, 0x2028, 0x10}},
    {28176, 2, {0x100, 0x4102, 0x211C, 0x2FD0, 0x210, 0x8210, 0x451E, 0x57D4, 0x1114, 0x2114, 0xE1D4, 0x2F14, 0x2514, 0x2124, 0x2124, 0x144}},
    {28177, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4040, 0x47FC, 0x1444, 0x1444, 0x27FC, 0xE444, 0x2444, 0x27FC, 0x2042, 0x2042, 0x3E}},
    {28178, 2, {0x40, 0x2080, 0x13FC, 0x1224, 0x8224, 0x43FC, 0x4A24, 0xA44, 0x13FC, 0x1090, 0xE110, 0x27FE, 0x2010, 0x2010, 0x2010, 0x10}},
    {28179, 2, {0x3C, 0x27C0, 0x1084, 0x1444, 0x8248, 0x4210, 0x4840, 0x840, 0x13F8, 0x1040, 0xE040, 0x27FC, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28180, 2, {0x100, 0x4100, 0x23F0, 0x2210, 0x420, 0x8FFC, 0x5444, 0x5444, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2000, 0x2000, 0x2FFE, 0x0}},
    {28181, 2, {0x444, 0x2244, 0x1284, 0x17D4, 0x8114, 0x4114, 0x5114, 0x1FF4, 0x1114, 0x2114, 0xE294, 0x2254, 0x2444, 0x2444, 0x2814, 0x8}},
    {28182, 2, {0x80, 0x2040, 0x17FC, 0x1444, 0x8040, 0x43F8, 0x4248, 0x1248, 0x13F8, 0x2248, 0xE248, 0x23F8, 0x2248, 0x2040, 0x2040, 0x40}},
    {28183, 2, {0x80, 0x2090, 0x1108, 0x13FC, 0x8040, 0x47FE, 0x4110, 0x1248, 0x1C86, 0x2310, 0xE020, 0x20C4, 0x2308, 0x2030, 0x20C0, 0x700}},
    {28184, 2, {0x0, 0x23F8, 0x1090, 0x1060, 0x87FC, 0x40A4, 0x4128, 0x1220, 0x1460, 0x2040, 0xE7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28185, 2, {0x100, 0x2100, 0x13F0, 0x1410, 0x8820, 0x47FC, 0x4404, 0x14A4, 0x1514, 0x2444, 0xE040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28186, 2, {0x80, 0x2088, 0x17E8, 0x1090, 0x80A0, 0x4FFE, 0x4080, 0x1100, 0x13F8, 0x2608, 0xEA08, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28187, 2, {0x14, 0x2012, 0x1010, 0x17FE, 0x8410, 0x4410, 0x45D0, 0x1412, 0x1412, 0x25D4, 0xE554, 0x2548, 0x25DA, 0x242A, 0x2846, 0x1082}},
    {28188, 2, {0x0, 0x27FC, 0x1040, 0x1080, 0x87FC, 0x44A4, 0x44A4, 0x14A4, 0x1484, 0x244C, 0xE040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28189, 2, {0x40, 0x20A0, 0x1110, 0x1208, 0x85F6, 0x4800, 0x47C4, 0x1454, 0x1454, 0x27D4, 0xE454, 0x2454, 0x27D4, 0x2444, 0x2554, 0x488}},
    {28190, 2, {0x208, 0x2110, 0x17FE, 0x1040, 0x8080, 0x43F8, 0x4A08, 0xA08, 0x13F8, 0x1208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28191, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x83F8, 0x4208, 0x43F8, 0x1000, 0x1FFE, 0x2802, 0xE3F8, 0x2040, 0x2040, 0x2040, 0x2140, 0x80}},
    {28192, 2, {0x23FC, 0x1200, 0x83F8, 0x4A08, 0xA08, 0x13F8, 0xE200, 0x23FC, 0x2100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {28193, 2, {0x40, 0x2020, 0x17FE, 0x1488, 0x8488, 0x47FE, 0x4488, 0x14F8, 0x1400, 0x25FC, 0xE484, 0x2488, 0x2450, 0x2820, 0x28D8, 0x1306}},
    {28194, 2, {0x4000, 0x2FF8, 0x2808, 0x8868, 0x4B88, 0x4888, 0x888, 0x2BE8, 0x2AA8, 0x4AA8, 0xCBE8, 0x488A, 0x48AA, 0x50EA, 0x5726, 0x2002}},
    {28195, 2, {0x40, 0x2040, 0x17FC, 0x10E0, 0x8150, 0x4248, 0x4446, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2000, 0x2FFE, 0x0}},
    {28196, 2, {0x210, 0x8210, 0x5FD0, 0x210, 0x823E, 0x5FD2, 0x5052, 0x2092, 0x2F12, 0x4112, 0xC212, 0x4392, 0x5E22, 0x4222, 0x4A4A, 0x484}},
    {28197, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4200, 0x4BFC, 0xA40, 0x1288, 0x13FC, 0xE224, 0x2220, 0x25FC, 0x2420, 0x2820, 0x13FE}},
    {28198, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83C8, 0x4248, 0x4248, 0x17FC, 0x1404, 0x25F4, 0xE514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28199, 2, {0x80, 0x2040, 0x13FC, 0x1000, 0x8108, 0x4090, 0x47FE, 0x1442, 0x1844, 0x23FC, 0xE244, 0x2244, 0x2254, 0x2248, 0x2040, 0x40}},
    {28200, 2, {0x0, 0x23F8, 0x1248, 0x1248, 0x83F8, 0x4248, 0x4A48, 0xBF8, 0x1000, 0x17FE, 0xE240, 0x2244, 0x2228, 0x2290, 0x2308, 0x206}},
    {28201, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4208, 0x4208, 0x13F8, 0x1000, 0x27FC, 0xE4A4, 0x24A4, 0x24A4, 0x24A4, 0x2FFE, 0x0}},
    {28202, 2, {0x8, 0x203C, 0x17C0, 0x1040, 0x83F8, 0x4248, 0x4A48, 0xBF8, 0x1040, 0x17FC, 0xE444, 0x2454, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28203, 2, {0x90, 0x2290, 0x1290, 0x17FE, 0x8290, 0x4290, 0x42F0, 0x1200, 0x13FC, 0x2040, 0xE7FE, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28204, 2, {0x2, 0x27C2, 0x1442, 0x1452, 0x87D2, 0x4452, 0x4452, 0x17D2, 0x1452, 0x2452, 0xE7D2, 0x2002, 0x2282, 0x2242, 0x244A, 0x804}},
    {28205, 2, {0x0, 0x27FC, 0x1444, 0x17FC, 0x8444, 0x47FC, 0x4000, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {28206, 2, {0x110, 0x2110, 0x17FE, 0x1110, 0x8000, 0x47FE, 0x4008, 0x1008, 0x13C8, 0x2248, 0xE248, 0x23C8, 0x2248, 0x2008, 0x2028, 0x10}},
    {28207, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8110, 0x4FFE, 0x4110, 0x1208, 0x17F4, 0x2A12, 0xE210, 0x23F0, 0x2204, 0x2204, 0x21FC, 0x0}},
    {28208, 2, {0x40, 0x20A0, 0x1110, 0x1208, 0x8DF6, 0x4000, 0x43F8, 0x1208, 0x1208, 0x23F8, 0xE110, 0x2110, 0x27FC, 0x2110, 0x2210, 0x410}},
    {28209, 2, {0x100, 0x2100, 0x113E, 0x17C8, 0x8548, 0x4548, 0x4548, 0x1548, 0x17C8, 0x2508, 0xE108, 0x2148, 0x21C8, 0x2E48, 0x243E, 0x0}},
    {28210, 2, {0x80, 0x4040, 0x27FC, 0x2404, 0x808, 0x83F8, 0x4000, 0x53F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2000, 0x27FC, 0x0}},
    {28211, 2, {0x0, 0x277E, 0x1124, 0x1124, 0x813C, 0x4724, 0x4424, 0x143C, 0x1424, 0x2724, 0xE126, 0x217C, 0x2104, 0x2104, 0x2A04, 0x404}},
    {28212, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4900, 0xBFC, 0x1444, 0x1244, 0xE2A4, 0x2204, 0x23F4, 0x2004, 0x2028, 0x10}},
    {28213, 2, {0x208, 0x4208, 0x2FFE, 0x2208, 0x208, 0x8000, 0x47FC, 0x5444, 0x1444, 0x2444, 0xE7FC, 0x2444, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28214, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8150, 0x4040, 0x43F8, 0x1248, 0x1248, 0x2248, 0xEFFE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x406}},
    {28215, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x8208, 0x4404, 0x43F8, 0x1208, 0x1208, 0x23F8, 0xE110, 0x2110, 0x2112, 0x2212, 0x240E, 0x800}},
    {28216, 2, {0x210, 0x2110, 0x1110, 0x17BE, 0x8220, 0x4240, 0x43BC, 0x1284, 0x1288, 0x2288, 0xE2BE, 0x2288, 0x2488, 0x2488, 0x29A8, 0x1010}},
    {28217, 2, {0x200, 0x2200, 0x13FE, 0x1402, 0x8882, 0x47FA, 0x4002, 0x17F2, 0x1002, 0x27F2, 0xE002, 0x27F2, 0x2412, 0x27F2, 0x2014, 0x8}},
    {28218, 2, {0x4010, 0x2010, 0x2F10, 0x8910, 0x4914, 0x4F52, 0x952, 0x2992, 0x2F10, 0x4914, 0xC904, 0x4F08, 0x4910, 0x4020, 0x4040, 0x180}},
    {28219, 2, {0x40, 0x2040, 0x1248, 0x1244, 0x8454, 0x4060, 0x4980, 0xE00, 0x13F8, 0x1208, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x208}},
    {28220, 2, {0x208, 0x2110, 0x17FE, 0x1040, 0x8040, 0x43FC, 0x4840, 0x840, 0x17FE, 0x1000, 0xE040, 0x27FC, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28221, 2, {0x128, 0x2124, 0x17A4, 0x1120, 0x8120, 0x4FFE, 0x4020, 0x1024, 0x17A4, 0x24A8, 0xE4A8, 0x2490, 0x2792, 0x24AA, 0x2046, 0x82}},
    {28222, 2, {0x0, 0x27FC, 0x1444, 0x1040, 0x87FE, 0x4040, 0x4BF8, 0xA48, 0x13F8, 0x1248, 0xE3F8, 0x2040, 0x27FE, 0x2040, 0x2040, 0x40}},
    {28223, 2, {0x40, 0x2040, 0x17FC, 0x10E0, 0x8150, 0x4248, 0x4C46, 0x13F8, 0x1000, 0x2000, 0xE7FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x80}},
    {28224, 2, {0x20, 0x27A8, 0x10B0, 0x14A4, 0x8318, 0x4110, 0x4208, 0x15F4, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x20A0, 0x2110, 0x2208, 0x404}},
    {28225, 2, {0x210, 0x2210, 0x1228, 0x1228, 0x8F44, 0x4282, 0x427C, 0x1300, 0x1200, 0x2E7C, 0xE244, 0x2244, 0x2244, 0x2244, 0x2A7C, 0x444}},
    {28226, 2, {0x0, 0x27BC, 0x14A4, 0x14A4, 0x87BC, 0x4000, 0x43F8, 0x1000, 0x1FFE, 0x2100, 0xE200, 0x23F8, 0x2008, 0x2008, 0x2050, 0x20}},
    {28227, 2, {0x100, 0x43BC, 0x2E10, 0x2210, 0x23C, 0x8F90, 0x4210, 0x523C, 0x1210, 0x2F90, 0xE27E, 0x2210, 0x2210, 0x2210, 0x2410, 0x810}},
    {28228, 2, {0x0, 0x23FC, 0x1224, 0x1224, 0x83FC, 0x4200, 0x4AFC, 0xA84, 0x1284, 0x12FC, 0xE284, 0x22FC, 0x2484, 0x2484, 0x28FC, 0x1084}},
    {28229, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x7FC, 0x8444, 0x4554, 0x54E4, 0x1444, 0x27FC, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {28230, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x8208, 0x4110, 0x4FFE, 0x1000, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28231, 2, {0x80, 0x2040, 0x17FC, 0x1208, 0x8110, 0x4FFE, 0x4000, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {28232, 2, {0x208, 0x4208, 0x2FFE, 0x2208, 0x208, 0x83F8, 0x4208, 0x5208, 0x13F8, 0x2040, 0xEFFE, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28233, 2, {0x208, 0x421C, 0x2270, 0x2210, 0x310, 0x8AFE, 0x4A10, 0x4A10, 0x1210, 0x227C, 0xE244, 0x2244, 0x2244, 0x2244, 0x227C, 0x244}},
    {28234, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FC, 0x1110, 0x1208, 0x25F6, 0xE040, 0x23F8, 0x2040, 0x20A0, 0x2110, 0x608}},
    {28235, 2, {0x80, 0x2080, 0x13F8, 0x1108, 0x87FE, 0x4000, 0x4BF8, 0xA08, 0x13F8, 0x1040, 0xE7FC, 0x2040, 0x2440, 0x27FE, 0x2040, 0x40}},
    {28236, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8288, 0x4446, 0x4BF8, 0xA08, 0x13F8, 0x1208, 0xE3F8, 0x2244, 0x2228, 0x2290, 0x2308, 0x204}},
    {28237, 2, {0x40, 0x2444, 0x1444, 0x1444, 0x87FC, 0x4000, 0x4FFE, 0x1040, 0x1080, 0x27FC, 0xE4A4, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x40C}},
    {28238, 2, {0x0, 0x27FE, 0x1040, 0x1080, 0x87FE, 0x4492, 0x4492, 0x14F2, 0x1492, 0x2492, 0xE4F2, 0x2492, 0x2492, 0x2492, 0x27FE, 0x402}},
    {28239, 2, {0x0, 0x27FE, 0x1040, 0x1080, 0x83FC, 0x4204, 0x4A04, 0xBFC, 0x1204, 0x13FC, 0xE204, 0x2204, 0x23FC, 0x2000, 0x2108, 0x204}},
    {28240, 2, {0x210, 0x4210, 0x2220, 0x227C, 0xFC4, 0x8244, 0x4244, 0x5744, 0x16FC, 0x2A44, 0xEA44, 0x3244, 0x2244, 0x2244, 0x227C, 0x244}},
    {28241, 2, {0x0, 0x27FC, 0x1044, 0x1240, 0x8278, 0x4240, 0x55FE, 0x1800, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x2228, 0x210}},
    {28242, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4000, 0x4FFE, 0x1208, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x227E, 0x2F88, 0x2008, 0x8}},
    {28243, 2, {0x10, 0x2110, 0x1108, 0x1204, 0x85FA, 0x4088, 0x4108, 0x1228, 0x1410, 0x23F8, 0xE2A8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x0}},
    {28244, 2, {0x208, 0x4108, 0x2110, 0x2FFE, 0x0, 0x87C4, 0x4454, 0x5454, 0x17D4, 0x2454, 0xE454, 0x27D4, 0x2454, 0x2444, 0x2554, 0x488}},
    {28245, 2, {0x10, 0x2010, 0x177C, 0x1114, 0x81FE, 0x4214, 0x427C, 0x1710, 0x117C, 0x2510, 0xE5FE, 0x2210, 0x2210, 0x2500, 0x28FE, 0x0}},
    {28246, 2, {0x100, 0x211E, 0x1112, 0x17D2, 0x8112, 0x411E, 0x4112, 0x17D2, 0x1452, 0x245E, 0xE452, 0x27D2, 0x2452, 0x2022, 0x202A, 0x44}},
    {28247, 2, {0x104, 0x2104, 0x1104, 0x17C4, 0x813E, 0x4104, 0x47C4, 0x1024, 0x1114, 0x2114, 0xE7C4, 0x2104, 0x2104, 0x21C4, 0x2E14, 0x408}},
    {28248, 2, {0x200, 0x4200, 0x227C, 0x2244, 0xF44, 0x8244, 0x427C, 0x5744, 0x16C4, 0x2A44, 0xF27C, 0x2244, 0x2244, 0x2244, 0x227C, 0x244}},
    {28249, 2, {0x80, 0x2040, 0x17FE, 0x1090, 0x8298, 0x4494, 0x4114, 0x1230, 0x1440, 0x2040, 0xE7FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28250, 2, {0x40, 0x2488, 0x1528, 0x15E8, 0x8448, 0x44A8, 0x45E8, 0x1408, 0x15E8, 0x2528, 0xE5E8, 0x2528, 0x25EA, 0x292A, 0x292A, 0x1166}},
    {28251, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x81F8, 0x4108, 0x49F8, 0x908, 0x1108, 0x17FE, 0xE250, 0x2288, 0x2304, 0x2200, 0x23FC, 0x0}},
    {28252, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4208, 0x4A08, 0xBF8, 0x1000, 0x17FE, 0xE040, 0x2240, 0x2278, 0x2240, 0x2540, 0x8FE}},
    {28253, 2, {0x220, 0x2224, 0x13A8, 0x1230, 0x82A2, 0x4322, 0x4A5E, 0x880, 0x13FC, 0x1204, 0xE204, 0x23FC, 0x2204, 0x2204, 0x23FC, 0x204}},
    {28254, 2, {0x40, 0x207E, 0x1040, 0x1040, 0x87FC, 0x4404, 0x4404, 0x17FC, 0x1404, 0x27FC, 0xE404, 0x2404, 0x27FC, 0x2110, 0x2208, 0x404}},
    {28255, 2, {0x40, 0x2080, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x4BF8, 0x800, 0x17FC, 0x1040, 0xE040, 0x23F8, 0x2040, 0x2040, 0x27FC, 0x0}},
    {28256, 2, {0x40, 0x2444, 0x1444, 0x1444, 0x87FC, 0x4100, 0x4100, 0x17FE, 0x1120, 0x2220, 0xE2A4, 0x22A8, 0x2550, 0x2450, 0x2888, 0x106}},
    {28257, 2, {0x3F8, 0x2248, 0x1248, 0x13F8, 0x8248, 0x4248, 0x4BF8, 0x840, 0x1040, 0x17FC, 0xE444, 0x2454, 0x25F4, 0x2414, 0x2404, 0x40C}},
    {28258, 2, {0x0, 0x27FC, 0x1000, 0x13F8, 0x8208, 0x4208, 0x43F8, 0x1000, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28259, 2, {0x0, 0x23FC, 0x1204, 0x13FC, 0x8220, 0x43FE, 0x4A10, 0xA8A, 0x1306, 0x1000, 0xE3FC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x204}},
    {28260, 2, {0x210, 0x2110, 0x1110, 0x17BE, 0x8228, 0x4248, 0x4BAA, 0xAAE, 0x12BA, 0x12EA, 0xE2AE, 0x22A8, 0x24AA, 0x24A2, 0x29A2, 0x101E}},
    {28261, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8110, 0x4208, 0x4444, 0x1050, 0x1048, 0x27FE, 0xE040, 0x20A0, 0x20A0, 0x2110, 0x2208, 0x406}},
    {28262, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4840, 0xA40, 0x13FC, 0x1440, 0xE040, 0x23F8, 0x2040, 0x2040, 0x27FE, 0x0}},
    {28263, 2, {0x0, 0x27F8, 0x1110, 0x10A0, 0x87FC, 0x4444, 0x47FC, 0x1444, 0x17FC, 0x2444, 0xE080, 0x27FC, 0x2084, 0x2104, 0x2228, 0x410}},
    {28264, 2, {0x0, 0x23FC, 0x1204, 0x13FC, 0x8204, 0x43FC, 0x4A04, 0xBFC, 0x1050, 0x1048, 0xE7FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28265, 2, {0x8, 0x203C, 0x17C0, 0x1040, 0x8FFE, 0x4040, 0x47FC, 0x1444, 0x17FC, 0x2444, 0xE7FC, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x0}},
    {28266, 2, {0x100, 0x21F8, 0x1208, 0x13F0, 0x8010, 0x47FE, 0x4080, 0x1144, 0x1668, 0x20B0, 0xE128, 0x2668, 0x20A4, 0x2122, 0x26A0, 0x40}},
    {28267, 2, {0x4390, 0x2E10, 0x2210, 0x8212, 0x4252, 0x4FD4, 0x258, 0x2690, 0x2710, 0x4AA8, 0xCA28, 0x5228, 0x4248, 0x4244, 0x4284, 0x302}},
    {28268, 2, {0x610, 0x3810, 0x852, 0x7E54, 0x890, 0x1C28, 0x2A44, 0x4882, 0x900, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {28269, 2, {0x108, 0x2088, 0x1090, 0x17FE, 0x8090, 0x4090, 0x4BFC, 0xA94, 0x1294, 0x130C, 0xE204, 0x23FC, 0x2204, 0x2204, 0x23FC, 0x204}},
    {28270, 2, {0x0, 0x27FC, 0x10A0, 0x10A0, 0x87FC, 0x44A4, 0x44A4, 0x14A4, 0x17FC, 0x2040, 0xE040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28271, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x5000, 0x1FFE, 0x1200, 0x23FC, 0xE494, 0x2924, 0x2244, 0x2484, 0x2128, 0x210}},
    {28272, 2, {0x80, 0x21F8, 0x1208, 0x1510, 0x80E0, 0x4318, 0x4C06, 0x13F8, 0x1040, 0x2240, 0xE3FC, 0x2440, 0x23F8, 0x2040, 0x2FFE, 0x0}},
    {28273, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FC, 0x1040, 0x17FC, 0x2080, 0xE100, 0x23F8, 0x2D08, 0x2108, 0x21F8, 0x108}},
    {28274, 2, {0x3C, 0x27C0, 0x1244, 0x1128, 0x8000, 0x43FC, 0x4880, 0x880, 0x17FE, 0x1100, 0xE1F8, 0x2288, 0x2250, 0x2420, 0x28D8, 0x306}},
    {28275, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x40, 0x87FC, 0x4514, 0x54A4, 0x15F4, 0x2444, 0xE444, 0x25F4, 0x2444, 0x2444, 0x2454, 0x408}},
    {28276, 2, {0x108, 0x2090, 0x1000, 0x13FC, 0x8090, 0x4090, 0x4092, 0x1492, 0x1294, 0x2298, 0xE090, 0x2090, 0x2090, 0x2090, 0x27FE, 0x0}},
    {28277, 2, {0x208, 0x2110, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FC, 0x1100, 0x1100, 0x23F8, 0xE410, 0x2820, 0x2040, 0x20A0, 0x2318, 0xC06}},
    {28278, 2, {0x80, 0x2100, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x43F8, 0x1040, 0x1040, 0x2764, 0xE168, 0x2150, 0x2248, 0x2444, 0x2942, 0x80}},
    {28279, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FE, 0x1110, 0x1208, 0x27FC, 0xEA0A, 0x2208, 0x23F8, 0x2208, 0x2208, 0x3F8}},
    {28280, 2, {0x80, 0x2040, 0x17FE, 0x1000, 0x83F8, 0x4208, 0x43F8, 0x1000, 0x17FE, 0x2402, 0xE9F4, 0x2110, 0x2110, 0x2212, 0x2412, 0x80E}},
    {28281, 2, {0x0, 0x27FE, 0x1400, 0x15FC, 0x8524, 0x4524, 0x45FC, 0x1524, 0x1524, 0x25FC, 0xE420, 0x2420, 0x25FC, 0x2820, 0x2820, 0x13FE}},
    {28282, 2, {0x100, 0x217C, 0x1144, 0x1244, 0x8244, 0x467C, 0x4A10, 0x1210, 0x12FE, 0x2238, 0xE254, 0x2254, 0x2294, 0x2312, 0x2210, 0x210}},
    {28283, 2, {0x40, 0x2020, 0x17FE, 0x1000, 0x81F8, 0x4108, 0x49F8, 0x800, 0x13FC, 0x1204, 0xE204, 0x23FC, 0x2204, 0x2204, 0x23FC, 0x204}},
    {28284, 2, {0x100, 0x261C, 0x1404, 0x1404, 0x871C, 0x4404, 0x4404, 0x17FC, 0x1000, 0x2000, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28285, 2, {0x248, 0x4248, 0x2490, 0x2FFE, 0x490, 0x8248, 0x4248, 0x5000, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28286, 2, {0x100, 0x2080, 0x1FFC, 0x1120, 0x8528, 0x4924, 0x4000, 0x17F0, 0x1010, 0x27F0, 0xE400, 0x27F8, 0x2008, 0x2008, 0x2050, 0x20}},
    {28287, 2, {0x0, 0x27F8, 0x1408, 0x1408, 0x87F8, 0x4408, 0x4408, 0x17F8, 0x1120, 0x2120, 0xE924, 0x2528, 0x2330, 0x2120, 0x2FFE, 0x0}},
    {28288, 2, {0x110, 0x4110, 0x2FFE, 0x2110, 0x0, 0x87FC, 0x4040, 0x5040, 0x17FC, 0x2444, 0xE554, 0x2554, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28289, 2, {0x88, 0x2088, 0x13FE, 0x1088, 0x8000, 0x43FE, 0x4202, 0x1424, 0x1020, 0x23FE, 0xE070, 0x20A8, 0x20A8, 0x2124, 0x2222, 0x20}},
    {28290, 2, {0x102, 0x2102, 0x17C2, 0x1112, 0x8112, 0x47D2, 0x4552, 0x1552, 0x17D2, 0x2112, 0xE392, 0x2542, 0x2922, 0x2102, 0x210A, 0x104}},
    {28291, 2, {0x40, 0x2040, 0x13F8, 0x1248, 0x83F8, 0x4040, 0x4FFE, 0x1000, 0x13F8, 0x2208, 0xE248, 0x2248, 0x2248, 0x20A0, 0x2110, 0x608}},
    {28292, 2, {0x40, 0x2444, 0x1444, 0x17FC, 0x8080, 0x41F8, 0x4208, 0x1510, 0x10E0, 0x2318, 0xEC46, 0x23F8, 0x2040, 0x2FFE, 0x2040, 0x40}},
    {28293, 2, {0x10, 0x4F94, 0x2892, 0x2890, 0xA9E, 0x8AF0, 0x4A90, 0x5A9E, 0x1AF0, 0x2A94, 0xEA94, 0x2208, 0x250A, 0x249A, 0x2826, 0x1042}},
    {28294, 2, {0x4200, 0x2200, 0x257C, 0x84A4, 0x4824, 0x5724, 0x224, 0x2228, 0x2FA8, 0x4228, 0xCB10, 0x4A90, 0x52A8, 0x4228, 0x4A44, 0x482}},
    {28295, 2, {0x40, 0x2248, 0x1150, 0x1040, 0x87FC, 0x4150, 0x4248, 0x1404, 0x1080, 0x2FFE, 0xE110, 0x2210, 0x2320, 0x20C0, 0x2330, 0xC08}},
    {28296, 2, {0x40, 0x2240, 0x1140, 0x17F8, 0x8088, 0x4090, 0x48FC, 0x904, 0x1108, 0x13FE, 0xE202, 0x242A, 0x2AAA, 0x2282, 0x2414, 0x8}},
    {28297, 2, {0x0, 0x27BE, 0x1488, 0x1488, 0x87A8, 0x44A8, 0x44A8, 0x17BE, 0x1410, 0x2518, 0xE498, 0x25A8, 0x26AA, 0x244A, 0x2046, 0x80}},
    {28298, 2, {0x0, 0x27BC, 0x1484, 0x1484, 0x8484, 0x47BC, 0x4400, 0x1400, 0x17BC, 0x2424, 0xE424, 0x27A8, 0x2410, 0x2428, 0x2444, 0x482}},
    {28299, 2, {0x0, 0x23F8, 0x1108, 0x1110, 0x811C, 0x42E4, 0x42A4, 0x1254, 0x14A8, 0x2100, 0xE3FC, 0x2294, 0x2294, 0x2294, 0x27FE, 0x0}},
    {28300, 2, {0x10, 0x23D4, 0x1058, 0x1452, 0x828C, 0x4108, 0x4204, 0x15FA, 0x1090, 0x2090, 0xE7FE, 0x2090, 0x2110, 0x2112, 0x2212, 0x40E}},
    {28301, 2, {0x0, 0x27FC, 0x10A0, 0x14A4, 0x82A8, 0x40A0, 0x4FFE, 0x1000, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28302, 2, {0x210, 0x2210, 0x127C, 0x1210, 0x8F90, 0x4210, 0x42FE, 0x1700, 0x1690, 0x2A10, 0xEA7C, 0x3210, 0x2210, 0x2210, 0x22FE, 0x200}},
    {28303, 2, {0x40, 0x2020, 0x17FE, 0x1420, 0x85FC, 0x4424, 0x47FE, 0x1424, 0x15FC, 0x2420, 0xE5FC, 0x2504, 0x2504, 0x2904, 0x29FC, 0x1104}},
    {28304, 2, {0x0, 0x27FE, 0x1420, 0x1440, 0x85FC, 0x4504, 0x45FC, 0x1504, 0x15FC, 0x2524, 0xE420, 0x24A8, 0x2924, 0x2A22, 0x30A0, 0x40}},
    {28305, 2, {0x40, 0x2244, 0x1148, 0x1040, 0x83FC, 0x4204, 0x4A04, 0xBFC, 0x1204, 0x13FC, 0xE204, 0x2204, 0x23FC, 0x2000, 0x2108, 0x204}},
    {28306, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x87FE, 0x4000, 0x4BF8, 0xA08, 0x1208, 0x13F8, 0xE0A4, 0x2128, 0x2310, 0x2D48, 0x2186, 0x100}},
    {28307, 2, {0x110, 0x20A0, 0x1000, 0x17FE, 0x80A0, 0x43F8, 0x40A8, 0x17FE, 0x10A8, 0x23F8, 0xE0A0, 0x21B0, 0x22A8, 0x2CA6, 0x20A0, 0xA0}},
    {28308, 2, {0x208, 0x2108, 0x1110, 0x1000, 0x87FC, 0x4040, 0x4840, 0xBF8, 0x1040, 0x1040, 0xE7FE, 0x2000, 0x2548, 0x24A4, 0x28A4, 0x0}},
    {28309, 2, {0x0, 0x47FE, 0x2402, 0x29F4, 0x0, 0x83F8, 0x4080, 0x5144, 0x1668, 0x20B0, 0xE128, 0x2668, 0x20A4, 0x2122, 0x26A0, 0x40}},
    {28310, 2, {0x2140, 0x1120, 0x83FC, 0x5240, 0x17F8, 0x2A40, 0xE3F8, 0x2240, 0x23FC, 0x2200, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {28311, 2, {0x38, 0x23C0, 0x1040, 0x17FC, 0x8150, 0x4154, 0x4758, 0x1154, 0x1354, 0x254C, 0xE0E0, 0x2150, 0x2248, 0x2444, 0x2842, 0x40}},
    {28312, 2, {0x40, 0x2040, 0x13FC, 0x1040, 0x8040, 0x4FFE, 0x4080, 0x1110, 0x13F8, 0x2008, 0xE7FC, 0x24A4, 0x24A4, 0x24A4, 0x2FFE, 0x0}},
    {28313, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FC, 0x1110, 0x1248, 0x2446, 0xE248, 0x2150, 0x2248, 0x2444, 0x2140, 0x80}},
    {28314, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8040, 0x40A0, 0x4910, 0xA08, 0x15F6, 0x1000, 0xE000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28315, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8110, 0x4208, 0x483C, 0xBD0, 0x1290, 0x1290, 0xE290, 0x2288, 0x2288, 0x24A4, 0x24D4, 0x892}},
    {28316, 2, {0x80, 0x277C, 0x1424, 0x1524, 0x84A4, 0x4554, 0x4648, 0x1080, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28317, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x83FC, 0x4110, 0x4FFE, 0x1040, 0x13FC, 0x2244, 0xE3FC, 0x2244, 0x2FFE, 0x2204, 0x2214, 0x208}},
    {28318, 2, {0x0, 0x23F8, 0x1148, 0x1110, 0x80A0, 0x4040, 0x41B0, 0x164E, 0x1040, 0x23F8, 0xE248, 0x2248, 0x23F8, 0x2044, 0x27FC, 0x4}},
    {28319, 2, {0x0, 0x27FC, 0x1404, 0x1000, 0x83F8, 0x4208, 0x4BF8, 0xA08, 0x13F8, 0x1080, 0xE040, 0x27FC, 0x2000, 0x2110, 0x2208, 0x404}},
    {28320, 2, {0x108, 0x2088, 0x1090, 0x13FC, 0x8040, 0x41F8, 0x4840, 0xBFE, 0x1080, 0x1100, 0xE1FC, 0x2220, 0x2420, 0x2820, 0x23FE, 0x0}},
    {28321, 2, {0x10, 0x2010, 0x1710, 0x157C, 0x8510, 0x4510, 0x45FE, 0x1708, 0x1508, 0x25FE, 0xE508, 0x2548, 0x2728, 0x2008, 0x2028, 0x10}},
    {28322, 2, {0x208, 0x2108, 0x1110, 0x1000, 0x87FC, 0x4000, 0x4110, 0x1208, 0x1404, 0x23F8, 0xE2A8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x0}},
    {28323, 2, {0x220, 0x4220, 0x2250, 0x2488, 0x504, 0x8EFA, 0x5400, 0x55FC, 0x1554, 0x2554, 0xE5FC, 0x2554, 0x2554, 0x2554, 0x2504, 0x50C}},
    {28324, 2, {0x0, 0x23FC, 0x1240, 0x13F8, 0x8240, 0x43F8, 0x4A40, 0xA40, 0x13FC, 0x1004, 0xE554, 0x2554, 0x2554, 0x2404, 0x2028, 0x10}},
    {28325, 2, {0x50, 0x4048, 0x2FFE, 0x2040, 0x7FC, 0x8444, 0x47FC, 0x5444, 0x17FC, 0x2444, 0xE008, 0x2FFE, 0x2208, 0x2108, 0x2128, 0x10}},
    {28326, 2, {0x108, 0x2548, 0x1548, 0x1550, 0x87DE, 0x4024, 0x4014, 0x1FD4, 0x1014, 0x2794, 0xE494, 0x24A8, 0x24C8, 0x2494, 0x2814, 0x1022}},
    {28327, 2, {0x0, 0x27FC, 0x10A0, 0x10A0, 0x87FC, 0x44A4, 0x44A4, 0x17FC, 0x1040, 0x2040, 0xE7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28328, 2, {0x110, 0x2114, 0x17D2, 0x1110, 0x8110, 0x4FFE, 0x4010, 0x1114, 0x17D4, 0x2114, 0xE394, 0x2548, 0x290A, 0x211A, 0x2126, 0x142}},
    {28329, 2, {0x80, 0x2100, 0x13F8, 0x1208, 0x8208, 0x43F8, 0x4A00, 0xBFE, 0x1200, 0x13FC, 0xE004, 0x2554, 0x2554, 0x2804, 0x2028, 0x10}},
    {28330, 2, {0x3C, 0x27C0, 0x1244, 0x1128, 0x8080, 0x4110, 0x43E0, 0x1040, 0x1188, 0x23FC, 0xE044, 0x2040, 0x27FC, 0x20A0, 0x2110, 0x60E}},
    {28331, 2, {0x4000, 0x23F8, 0x2248, 0xA48, 0x8AA8, 0x5308, 0x53F8, 0x2208, 0x2000, 0xC7FC, 0x44A4, 0x44A4, 0x44A4, 0x44A4, 0x4FFE, 0x0}},
    {28332, 2, {0x208, 0x2110, 0x17FC, 0x1040, 0x8040, 0x43F8, 0x4840, 0x880, 0x17FC, 0x10A0, 0xE128, 0x2134, 0x223C, 0x2222, 0x2422, 0x81E}},
    {28333, 2, {0x0, 0x27FC, 0x1444, 0x17FC, 0x8444, 0x47FC, 0x4110, 0x1208, 0x1484, 0x21F8, 0xE208, 0x2508, 0x2090, 0x2060, 0x2198, 0xE06}},
    {28334, 2, {0x200, 0x24FE, 0x1710, 0x1510, 0x857C, 0x4554, 0x4754, 0x1454, 0x1454, 0x2754, 0xE554, 0x255C, 0x2510, 0x2710, 0x2510, 0x10}},
    {28335, 2, {0x440, 0x425E, 0x2292, 0x2012, 0x7D2, 0x811E, 0x4552, 0x5552, 0x1552, 0x27DE, 0xE152, 0x2112, 0x2212, 0x2222, 0x242A, 0x844}},
    {28336, 2, {0x40, 0x2248, 0x1248, 0x1248, 0x83F8, 0x4000, 0x47FC, 0x1000, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2110, 0x20A0, 0x27FE, 0x0}},
    {28337, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FC, 0x1110, 0x1268, 0x25C6, 0xE040, 0x23F8, 0x20E0, 0x2158, 0x2644, 0x40}},
    {28338, 2, {0x40, 0x2140, 0x165C, 0x1444, 0x8444, 0x475C, 0x4444, 0x1444, 0x17FC, 0x2040, 0xE7F8, 0x2110, 0x20A0, 0x2060, 0x2398, 0x1C06}},
    {28339, 2, {0x0, 0x21F8, 0x1108, 0x1108, 0x81F8, 0x4000, 0x4BFC, 0xA04, 0x13FC, 0x1204, 0xE3FC, 0x2204, 0x23FC, 0x2090, 0x2108, 0x204}},
    {28340, 2, {0x40, 0x2080, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x43F8, 0x1208, 0x13F8, 0x2050, 0xE048, 0x27FE, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28341, 2, {0x0, 0x21B8, 0x1628, 0x1428, 0x87A8, 0x44A8, 0x47CE, 0x1480, 0x17B8, 0x2428, 0xE7A8, 0x24A8, 0x2490, 0x2890, 0x2AA8, 0x1146}},
    {28342, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x8494, 0x4108, 0x4A64, 0x890, 0x1108, 0x1204, 0xE5FA, 0x2108, 0x2108, 0x2108, 0x21F8, 0x108}},
    {28343, 2, {0x0, 0x27FE, 0x1402, 0x15FA, 0x8442, 0x44C2, 0x452A, 0x146A, 0x14B2, 0x252A, 0xE46A, 0x24AA, 0x2522, 0x2442, 0x27FE, 0x402}},
    {28344, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FE, 0x1080, 0x1110, 0x23E0, 0xE048, 0x27FC, 0x2044, 0x2248, 0x2544, 0x80}},
    {28345, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8FFE, 0x4882, 0x4110, 0x13E0, 0x1040, 0x2088, 0xE7FC, 0x2044, 0x2248, 0x2444, 0x2942, 0x80}},
    {28346, 2, {0x0, 0x27BC, 0x1084, 0x1084, 0x87BC, 0x4420, 0x4420, 0x17BC, 0x1084, 0x24A4, 0xE294, 0x2294, 0x24A4, 0x2084, 0x2528, 0x210}},
    {28347, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4000, 0x17BC, 0x1084, 0x24A4, 0xE294, 0x218C, 0x26B4, 0x2084, 0x2294, 0x108}},
    {28348, 2, {0x0, 0x2FFE, 0x1210, 0x1210, 0x84A4, 0x4738, 0x4210, 0x14A4, 0x17BC, 0x2040, 0xE040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28349, 2, {0x0, 0x23FE, 0x1200, 0x12FC, 0x8200, 0x43FE, 0x4AA8, 0xA90, 0x12C8, 0x1286, 0xE208, 0x23FE, 0x2288, 0x2448, 0x2408, 0x818}},
    {28350, 2, {0x80, 0x2100, 0x17FC, 0x1444, 0x8444, 0x47FC, 0x4444, 0x1484, 0x17FC, 0x20C8, 0xE150, 0x2168, 0x227C, 0x2242, 0x2442, 0x83E}},
    {28351, 2, {0x10, 0x8094, 0x5F54, 0x1554, 0x9558, 0x5510, 0x557C, 0x1F10, 0x3510, 0x55FE, 0xD510, 0x5510, 0x5F10, 0x5110, 0x4010, 0x10}},
    {28352, 2, {0x80, 0x2040, 0x17FC, 0x1080, 0x8108, 0x43F0, 0x4860, 0x984, 0x17FE, 0x1002, 0xE7FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x404}},
    {28353, 2, {0x4010, 0x2F10, 0x2928, 0x8A28, 0x4A44, 0x4CBA, 0xA10, 0x2910, 0x29FE, 0x4910, 0xCD58, 0x4A54, 0x4892, 0x4912, 0x4850, 0x820}},
    {28354, 2, {0x80, 0x4040, 0x27FC, 0x2210, 0x120, 0x8FFE, 0x4882, 0x5044, 0x17FC, 0x2100, 0xE100, 0x21F8, 0x2108, 0x2208, 0x2228, 0x410}},
    {28355, 2, {0x120, 0x2110, 0x1208, 0x1486, 0x8910, 0x43F8, 0x4008, 0x1000, 0x17BC, 0x2084, 0xE4A4, 0x2294, 0x24A4, 0x2084, 0x2294, 0x108}},
    {28356, 2, {0x40, 0x2040, 0x10A0, 0x1190, 0x8248, 0x47FE, 0x4A08, 0x13F8, 0x1208, 0x23F8, 0xE200, 0x23F8, 0x2508, 0x2508, 0x29F8, 0x108}},
    {28357, 2, {0x14, 0x2012, 0x1010, 0x17FE, 0x8410, 0x4410, 0x47F2, 0x1492, 0x1492, 0x25B4, 0xE6D4, 0x2488, 0x254A, 0x293A, 0x2A26, 0x1042}},
    {28358, 2, {0x0, 0x2FFE, 0x1000, 0x13F8, 0x8208, 0x4208, 0x43F8, 0x1000, 0x17FC, 0x2514, 0xE4A4, 0x27FC, 0x2444, 0x2444, 0x2454, 0x408}},
    {28359, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x83F8, 0x4208, 0x43F8, 0x1208, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x2FFE, 0x2110, 0x2208, 0x404}},
    {28360, 2, {0x80, 0x2040, 0x17FE, 0x1000, 0x83F8, 0x4208, 0x43F8, 0x1000, 0x17FC, 0x2404, 0xE5F4, 0x2514, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28361, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4840, 0xA48, 0x1150, 0x1040, 0xE7FC, 0x2120, 0x2120, 0x2224, 0x2424, 0x81C}},
    {28362, 2, {0x100, 0x2100, 0x11FC, 0x1200, 0x85F8, 0x4000, 0x4BF8, 0x888, 0x12A8, 0x11C8, 0xE7F8, 0x208A, 0x21CA, 0x26AA, 0x2086, 0x82}},
    {28363, 2, {0x108, 0x2088, 0x1090, 0x17FE, 0x8108, 0x4108, 0x4210, 0x1294, 0x14A4, 0x2738, 0xE108, 0x2210, 0x2210, 0x24A4, 0x27BC, 0x84}},
    {28364, 2, {0x4420, 0x2420, 0x247C, 0x8844, 0x4AA8, 0x4A10, 0x1A28, 0x2AD6, 0x2A10, 0x4AFE, 0xCA10, 0x4A54, 0x4892, 0x4912, 0x4850, 0x820}},
    {28365, 2, {0x40, 0x2244, 0x1244, 0x13FC, 0x8040, 0x4FFE, 0x4000, 0x1040, 0x13FC, 0x2244, 0xE244, 0x23FC, 0x2040, 0x2044, 0x27FE, 0x202}},
    {28366, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {28367, 2, {0x110, 0x4208, 0x2514, 0x20A0, 0xE0, 0x8318, 0x4C06, 0x53F8, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2248, 0x2150, 0x2FFE, 0x0}},
    {28368, 2, {0x208, 0x2208, 0x13BE, 0x1488, 0x8AA8, 0x413E, 0x4208, 0x1448, 0x1848, 0x27FC, 0xE0E0, 0x2150, 0x2248, 0x2C46, 0x2040, 0x40}},
    {28369, 2, {0x0, 0x23F8, 0x1208, 0x13C8, 0x8248, 0x4FFE, 0x4802, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2208, 0x2228, 0x210}},
    {28370, 2, {0x0, 0x27FC, 0x1008, 0x13C8, 0x8248, 0x4248, 0x43C8, 0x1000, 0x1FFE, 0x2008, 0xE3C8, 0x2248, 0x2248, 0x23C8, 0x2028, 0x10}},
    {28371, 2, {0x80, 0x2040, 0x17FC, 0x1484, 0x8040, 0x43F8, 0x4110, 0x10A0, 0x17FC, 0x2040, 0xE040, 0x23F8, 0x2040, 0x2040, 0x2040, 0x40}},
    {28372, 2, {0x3E, 0x4FC0, 0x2084, 0x2444, 0x248, 0x8210, 0x4100, 0x561C, 0x1404, 0x2404, 0xE71C, 0x2404, 0x2404, 0x2404, 0x27FC, 0x404}},
    {28373, 2, {0x20, 0x7924, 0x48A8, 0x49FC, 0x4820, 0x7840, 0x4BFE, 0x4888, 0x4924, 0x7A22, 0x4D24, 0x48A8, 0x4924, 0x4A22, 0x48A0, 0x9840}},
    {28374, 2, {0x80, 0x2040, 0x17FE, 0x1000, 0x83F8, 0x4208, 0x4FFE, 0x1208, 0x13F8, 0x20A4, 0xE128, 0x2310, 0x2510, 0x2948, 0x2186, 0x100}},
    {28375, 2, {0x210, 0x2210, 0x13DE, 0x1528, 0x8884, 0x4078, 0x4780, 0x1080, 0x10F8, 0x2780, 0xE080, 0x20F8, 0x2F84, 0x2084, 0x2084, 0x7C}},
    {28376, 2, {0x0, 0x47FE, 0x2402, 0x2914, 0x208, 0x85F4, 0x4040, 0x5040, 0x17FC, 0x2000, 0xE000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28377, 2, {0x0, 0x27FE, 0x1450, 0x1488, 0x85FE, 0x4690, 0x44FC, 0x1490, 0x14FC, 0x2490, 0xE490, 0x24FE, 0x2480, 0x2400, 0x27FE, 0x0}},
    {28378, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x8248, 0x4484, 0x4120, 0x13F0, 0x1010, 0x2088, 0xE150, 0x2320, 0x2510, 0x2948, 0x2186, 0x100}},
    {28379, 2, {0x4, 0x203E, 0x17C0, 0x1244, 0x8124, 0x4128, 0x4200, 0x13FC, 0x1440, 0x2040, 0xEFFE, 0x2040, 0x2444, 0x2444, 0x27FC, 0x4}},
    {28380, 2, {0x40, 0x2080, 0x13F8, 0x1208, 0x83F8, 0x4208, 0x4BF8, 0x880, 0x17FC, 0x1110, 0xE248, 0x2444, 0x2BFA, 0x2040, 0x2040, 0x40}},
    {28381, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x80A0, 0x47FE, 0x4800, 0xBF8, 0x1248, 0x13F8, 0xE248, 0x23F8, 0x224A, 0x2042, 0x203E, 0x0}},
    {28382, 2, {0x150, 0x2150, 0x17FC, 0x1150, 0x8150, 0x4000, 0x47FE, 0x1442, 0x1844, 0x23F8, 0xE248, 0x2248, 0x2248, 0x2258, 0x2040, 0x40}},
    {28383, 2, {0x420, 0x8420, 0x4478, 0x1F48, 0x8490, 0x447C, 0x4454, 0x1E54, 0x2454, 0x447C, 0xC440, 0x5F40, 0x4442, 0x4442, 0x443E, 0x400}},
    {28384, 2, {0x0, 0x23FC, 0x1108, 0x11F8, 0x8108, 0x41F8, 0x410E, 0x17F8, 0x1008, 0x27BC, 0xE0A4, 0x22A4, 0x2128, 0x2290, 0x24A8, 0x846}},
    {28385, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x8000, 0x47FE, 0x4090, 0x1090, 0x17FE, 0x2492, 0xE492, 0x256A, 0x2646, 0x2402, 0x240A, 0x404}},
    {28386, 2, {0x110, 0x2110, 0x17FE, 0x1110, 0x8000, 0x4FFE, 0x4802, 0x1000, 0x17FC, 0x2040, 0xE040, 0x23F8, 0x2050, 0x2048, 0x2FFE, 0x0}},
    {28387, 2, {0x0, 0x23FE, 0x1200, 0x12FC, 0x8200, 0x43FE, 0x4AA8, 0xA90, 0x12C8, 0x1286, 0xE200, 0x22FC, 0x2284, 0x2484, 0x24FC, 0x884}},
    {28388, 2, {0x40, 0x207C, 0x1040, 0x17FE, 0x8442, 0x4470, 0x47C4, 0x1444, 0x143C, 0x2420, 0xE410, 0x2454, 0x2542, 0x2942, 0x294A, 0x1238}},
    {28389, 2, {0x90, 0x2490, 0x149E, 0x1490, 0x84A8, 0x44A4, 0x44C4, 0x1080, 0x1000, 0x27FC, 0xE4A4, 0x24A4, 0x24A4, 0x24A4, 0x2FFE, 0x0}},
    {28390, 2, {0x80, 0x4040, 0x2FFE, 0x20A0, 0x2A8, 0x82A4, 0x44A4, 0x5040, 0x1040, 0x27FE, 0xE0E0, 0x2150, 0x2248, 0x2444, 0x2842, 0x40}},
    {28391, 2, {0x4410, 0x2210, 0x2210, 0x8FA0, 0x403E, 0x4544, 0x4A4, 0x28A4, 0x3128, 0x4528, 0xC210, 0x4210, 0x4528, 0x4928, 0x5044, 0x82}},
    {28392, 2, {0x80, 0x2040, 0x17FE, 0x1402, 0x8814, 0x43E0, 0x4A00, 0xA00, 0x13FC, 0x1210, 0xE210, 0x2FFE, 0x2000, 0x2110, 0x2208, 0x404}},
    {28393, 2, {0x50, 0x4048, 0x2E80, 0x22FE, 0x390, 0x9290, 0x4AFC, 0x5A90, 0x1490, 0x24FC, 0xEA90, 0x2A90, 0x3290, 0x20FE, 0x2080, 0x80}},
    {28394, 2, {0x0, 0x4FBE, 0x2088, 0x2510, 0x23E, 0x8122, 0x4FAA, 0x52AA, 0x122A, 0x222A, 0xE22A, 0x222A, 0x2210, 0x2214, 0x2A22, 0x442}},
    {28395, 2, {0x4440, 0x2440, 0x24FC, 0x8888, 0x4B50, 0x5A20, 0x1AD8, 0x2B06, 0x2AF8, 0x4A88, 0xCAF8, 0x4A88, 0x4AF8, 0x4888, 0x4888, 0x898}},
    {28396, 2, {0x40, 0x2020, 0x13FC, 0x1204, 0x83FC, 0x4200, 0x4AF8, 0xA88, 0x12F8, 0x1200, 0xE3FC, 0x2324, 0x25FC, 0x2502, 0x2902, 0xFE}},
    {28397, 2, {0x0, 0x27FC, 0x1444, 0x17FC, 0x8444, 0x47FC, 0x5248, 0x1248, 0x17FE, 0x2248, 0xE248, 0x27FC, 0x2040, 0x2FFE, 0x2040, 0x40}},
    {28398, 2, {0x4200, 0x23C2, 0x2202, 0x8FE4, 0x4A28, 0x4B92, 0xE42, 0x29C4, 0x2808, 0x4B92, 0xCA82, 0x4A84, 0x4A8A, 0x4C92, 0x547E, 0x800}},
    {28399, 2, {0x2A8, 0x22A8, 0x17FC, 0x12A8, 0x82AA, 0x44E6, 0x4800, 0x17FC, 0x1444, 0x2040, 0xE3F8, 0x2248, 0x2248, 0x2248, 0x2258, 0x40}},
    {28400, 2, {0x80, 0x2040, 0x13F8, 0x1110, 0x80A0, 0x47FE, 0x4000, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2120, 0x2122, 0x2222, 0xC1E}},
    {28401, 2, {0x4080, 0x2040, 0x2FFE, 0x8802, 0x5004, 0x4F90, 0x1C, 0x2010, 0x2F90, 0x457C, 0xC504, 0x4528, 0x4910, 0x492A, 0x5142, 0x20FE}},
    {28402, 2, {0x80, 0x2110, 0x13F8, 0x1210, 0x84A4, 0x4FBE, 0x4042, 0x10A0, 0x1318, 0x2C46, 0xE190, 0x2664, 0x2188, 0x2630, 0x21C0, 0xE00}},
    {28403, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x80A0, 0x47FC, 0x44A4, 0x1514, 0x160C, 0x25F4, 0xE514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28404, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x80A0, 0x47FC, 0x4444, 0x15F4, 0x1444, 0x25F4, 0xE514, 0x2514, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28405, 2, {0x80, 0x4040, 0x27FE, 0x2402, 0x894, 0x8050, 0x4524, 0x554A, 0x18FA, 0x2300, 0xE040, 0x2444, 0x2444, 0x2444, 0x27FC, 0x4}},
    {28406, 2, {0x108, 0x2108, 0x1108, 0x17D0, 0x811E, 0x4794, 0x5124, 0x1FD4, 0x1214, 0x23D4, 0xE254, 0x2248, 0x2248, 0x24D4, 0x2424, 0x842}},
    {28407, 2, {0x40, 0x2040, 0x107E, 0x1040, 0x8040, 0x47FC, 0x4404, 0x164C, 0x1514, 0x24A4, 0xE64C, 0x24A4, 0x2514, 0x264C, 0x27FC, 0x404}},
    {28408, 2, {0x210, 0x4110, 0x2110, 0x2FDE, 0x28, 0x8788, 0x4008, 0x5788, 0x103E, 0x2788, 0xE488, 0x2488, 0x2488, 0x2788, 0x2488, 0x8}},
    {28409, 2, {0x40, 0x207C, 0x1040, 0x17FE, 0x8442, 0x4478, 0x45C4, 0x143C, 0x1408, 0x24F0, 0xE524, 0x24A8, 0x27FE, 0x2820, 0x28A0, 0x1040}},
    {28410, 2, {0x220, 0x4220, 0x243E, 0x2548, 0xDA8, 0x9510, 0x4528, 0x5446, 0x1480, 0x2040, 0xE124, 0x2522, 0x250A, 0x250A, 0x28F8, 0x0}},
    {28411, 2, {0x40, 0x2020, 0x13FC, 0x1000, 0x8108, 0x4090, 0x4BFE, 0xA20, 0x12A0, 0x12FC, 0xE320, 0x2220, 0x22FC, 0x2420, 0x2420, 0x9FE}},
    {28412, 2, {0x110, 0x2110, 0x17BE, 0x1110, 0x83B8, 0x4554, 0x4912, 0x1000, 0x13F0, 0x2210, 0xE290, 0x2250, 0x2252, 0x2212, 0x240E, 0x800}},
    {28413, 2, {0x20, 0x27FE, 0x1420, 0x15FC, 0x8424, 0x47FE, 0x4424, 0x15FC, 0x1420, 0x25FC, 0xE524, 0x25FC, 0x2524, 0x29FC, 0x2924, 0x112C}},
    {28414, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x8208, 0x45F4, 0x4110, 0x1110, 0x11F0, 0x2088, 0xE150, 0x2320, 0x2510, 0x2948, 0x2186, 0x100}},
    {28415, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8110, 0x41F0, 0x4040, 0x17FC, 0x1444, 0x2664, 0xE554, 0x26EC, 0x2444, 0x2444, 0x2454, 0x448}},
    {28416, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x8084, 0x108, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {28417, 2, {0x100, 0x2100, 0x13F0, 0x1210, 0x8420, 0x4FFC, 0x5444, 0x1444, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2000, 0x2524, 0x2492, 0x892}},
    {28418, 2, {0x0, 0x4FFE, 0x20A0, 0x27FC, 0x4A4, 0x84A4, 0x47FC, 0x5000, 0x17FC, 0x2000, 0xEFFE, 0x2040, 0x2248, 0x2444, 0x2942, 0x80}},
    {28419, 2, {0x80, 0x4040, 0x2FFE, 0x2802, 0x1204, 0x83BC, 0x4224, 0x5224, 0x1FA4, 0x2028, 0xE228, 0x2B10, 0x3290, 0x2228, 0x2A44, 0x482}},
    {28420, 2, {0x20, 0x2222, 0x1222, 0x13FE, 0x8000, 0x43FE, 0x4A20, 0xBFC, 0x1220, 0x13FE, 0xE200, 0x2220, 0x25FC, 0x2420, 0x2BFE, 0x0}},
    {28421, 2, {0x248, 0x2490, 0x1248, 0x1000, 0x83F8, 0x4248, 0x4BF8, 0xA48, 0x13F8, 0x1040, 0xE7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28422, 2, {0x40, 0x2040, 0x17FC, 0x10E0, 0x8158, 0x4644, 0x40A0, 0x1110, 0x1248, 0x2C46, 0xE248, 0x2150, 0x2248, 0x2444, 0x2140, 0x80}},
    {28423, 2, {0x210, 0x4210, 0x2450, 0x295C, 0x150, 0x8250, 0x46FE, 0x5A00, 0x1210, 0x2250, 0xE25C, 0x2250, 0x22B0, 0x2290, 0x230E, 0x200}},
    {28424, 2, {0x100, 0x2110, 0x13DC, 0x1254, 0x8554, 0x4288, 0x4108, 0x12F4, 0x1402, 0x2000, 0xE3FC, 0x2040, 0x2150, 0x2248, 0x2544, 0x80}},
    {28425, 2, {0x80, 0x2040, 0x17FC, 0x14A0, 0x84A0, 0x47FC, 0x44A4, 0x14A4, 0x17FC, 0x2400, 0xE524, 0x25A8, 0x2530, 0x2924, 0x29A4, 0x111C}},
    {28426, 2, {0x40, 0x43F8, 0x2248, 0x2FFE, 0x248, 0x83F8, 0x4040, 0x53F8, 0x1248, 0x23F8, 0xE080, 0x2FFE, 0x2110, 0x20E0, 0x2318, 0xC04}},
    {28427, 2, {0x20, 0x4F3C, 0x2944, 0x2AA8, 0xA10, 0x8C28, 0x4AC6, 0x4938, 0x1910, 0x2950, 0xED7E, 0x2A90, 0x287C, 0x2810, 0x28FE, 0x800}},
    {28428, 2, {0x110, 0x4110, 0x2FFE, 0x2110, 0x1F0, 0x8040, 0x47FC, 0x5444, 0x17FC, 0x2040, 0xEFFE, 0x2040, 0x27FC, 0x2040, 0x2FFE, 0x0}},
    {28429, 2, {0x0, 0x4FFE, 0x282A, 0x2826, 0xFFE, 0x8822, 0x4BAA, 0x4AAA, 0x1BAA, 0x2832, 0xE996, 0x2E2E, 0x2846, 0x2882, 0x2FFE, 0x802}},
    {28430, 2, {0x124, 0x4124, 0x2224, 0x2424, 0x954, 0x814A, 0x4292, 0x5610, 0x1A10, 0x2250, 0xE25C, 0x2250, 0x2250, 0x22B0, 0x229E, 0x300}},
    {28431, 2, {0x0, 0x47FE, 0x2402, 0x2402, 0x7FE, 0x8400, 0x47FE, 0x5420, 0x17FE, 0x2622, 0xE6AA, 0x2A22, 0x2AAA, 0x3222, 0x222A, 0x204}},
    {28432, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x982, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {28433, 2, {0x100, 0x423E, 0x2788, 0x2488, 0x4A8, 0x87A8, 0x44BE, 0x5488, 0x1788, 0x2010, 0xE498, 0x2718, 0x2428, 0x24AA, 0x23CA, 0x86}},
    {28434, 2, {0x0, 0x277C, 0x1144, 0x1144, 0x817C, 0x4710, 0x4410, 0x147C, 0x1454, 0x2754, 0xE154, 0x217C, 0x2110, 0x2114, 0x2AFE, 0x402}},
    {28435, 2, {0x80, 0x2040, 0x17FE, 0x1000, 0x82A8, 0x4248, 0x42A8, 0x13F8, 0x1040, 0x27FC, 0xE484, 0x2524, 0x25F4, 0x2414, 0x2404, 0x40C}},
    {28436, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x8000, 0x43F8, 0x4040, 0x13F8, 0x1248, 0x23F8, 0xE248, 0x23F8, 0x2000, 0x2110, 0x2208, 0x404}},
    {28437, 2, {0xA0, 0x20A0, 0x17FE, 0x10A0, 0x87FC, 0x44A4, 0x47FC, 0x14A4, 0x17FC, 0x2000, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x208}},
    {28438, 2, {0x208, 0x2248, 0x1748, 0x1290, 0x829E, 0x4FE4, 0x4114, 0x1214, 0x1FD4, 0x2494, 0xE914, 0x21C8, 0x2F08, 0x2114, 0x2514, 0x222}},
    {28439, 2, {0x40, 0x2080, 0x13FC, 0x1244, 0x82F4, 0x4314, 0x4AA4, 0xA44, 0x12A4, 0x13FC, 0xE040, 0x2024, 0x2522, 0x250A, 0x2908, 0xF8}},
    {28440, 2, {0x7FE, 0x2400, 0x15FC, 0x1400, 0x87FE, 0x4514, 0x4548, 0x1586, 0x1400, 0x25FC, 0xE504, 0x25FC, 0x2504, 0x29FC, 0x2904, 0x110C}},
    {28441, 2, {0x40, 0x4FFE, 0x2040, 0x27FC, 0x444, 0x87FC, 0x4444, 0x57FC, 0x1042, 0x2FFE, 0xE010, 0x2FFE, 0x2210, 0x2110, 0x2150, 0x20}},
    {28442, 2, {0x0, 0x27FC, 0x1400, 0x14F8, 0x8488, 0x4488, 0x44F8, 0x1400, 0x15DC, 0x2554, 0xE554, 0x2554, 0x25DC, 0x2400, 0x27FE, 0x0}},
    {28443, 2, {0x40, 0x2248, 0x1150, 0x17FC, 0x8040, 0x4080, 0x4FFE, 0x1110, 0x1248, 0x2446, 0xEA48, 0x2150, 0x2248, 0x2444, 0x2140, 0x80}},
    {28444, 2, {0x110, 0x2110, 0x17BC, 0x1110, 0x83B8, 0x4554, 0x4912, 0x1110, 0x1040, 0x2040, 0xE7FC, 0x2040, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28445, 2, {0x0, 0x27BC, 0x1084, 0x14A4, 0x8294, 0x44A4, 0x4084, 0x1040, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28446, 2, {0xE, 0x27F0, 0x1044, 0x1224, 0x8108, 0x43FC, 0x4A04, 0xBFC, 0x1204, 0x13FC, 0xE204, 0x23FC, 0x2090, 0x2110, 0x2212, 0xC0E}},
    {28447, 2, {0x40, 0x4444, 0x2248, 0x2FFE, 0x802, 0x83F8, 0x4208, 0x5208, 0x13F8, 0x2040, 0xE040, 0x27FC, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28448, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8000, 0x43F8, 0x4A08, 0xBF8, 0x1208, 0x13F8, 0xE040, 0x27FC, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28449, 2, {0x100, 0x21FC, 0x1200, 0x15F8, 0x8108, 0x41F8, 0x4908, 0x9F8, 0x1000, 0x17FE, 0xE100, 0x23FC, 0x24A4, 0x2124, 0x2254, 0x88}},
    {28450, 2, {0x110, 0x2110, 0x17FE, 0x1110, 0x81F0, 0x4040, 0x4BF8, 0xA48, 0x13F8, 0x1040, 0xE7FC, 0x2040, 0x2FFE, 0x20A0, 0x2110, 0x60C}},
    {28451, 2, {0x20, 0x4820, 0x25FE, 0x2420, 0x1FC, 0x8124, 0x4DFC, 0x5524, 0x15FC, 0x2420, 0xE5FE, 0x2420, 0x2420, 0x2A20, 0x31FE, 0x0}},
    {28452, 2, {0x110, 0x4110, 0x27BC, 0x2110, 0x3B8, 0x8554, 0x4912, 0x4080, 0x1080, 0x2FFC, 0xE110, 0x2210, 0x2120, 0x20C0, 0x2330, 0xC08}},
    {28453, 2, {0x80, 0x4040, 0x2FFE, 0x2912, 0x248, 0x8444, 0x43F8, 0x5040, 0x1040, 0x2FFE, 0xE000, 0x2040, 0x23F8, 0x2040, 0x2040, 0xFFE}},
    {28454, 2, {0x820, 0x3E3E, 0x844, 0x7FA8, 0x2A10, 0x4928, 0x844, 0x3FFE, 0x2000, 0x2088, 0x2ED0, 0x22A0, 0x2490, 0x4888, 0x5286, 0x8100}},
    {28455, 2, {0x110, 0x4110, 0x2110, 0x2FDE, 0x120, 0x87C0, 0x445C, 0x57C4, 0x1448, 0x27C8, 0xE110, 0x2FD0, 0x2122, 0x2122, 0x211E, 0x100}},
    {28456, 2, {0x20, 0x247C, 0x1284, 0x1148, 0x8030, 0x40CE, 0x4610, 0x127C, 0x1210, 0x227C, 0xE210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x800}},
    {28457, 2, {0x420, 0x4220, 0x2220, 0x2F7E, 0x480, 0x8400, 0x477C, 0x5514, 0x1510, 0x2550, 0xE55C, 0x2550, 0x2950, 0x2B50, 0x30BE, 0x2100}},
    {28458, 2, {0x20, 0x4920, 0x25FC, 0x2220, 0x650, 0x8A88, 0x43FE, 0x5208, 0x16E8, 0x2AA8, 0xE2A8, 0x22E8, 0x22A8, 0x2208, 0x2A28, 0x410}},
    {28459, 2, {0x3F8, 0x2208, 0x13F8, 0x1208, 0x83F8, 0x4000, 0x47FC, 0x14A4, 0x14A4, 0x27FC, 0xE000, 0x23F8, 0x2110, 0x20E0, 0x2318, 0xC06}},
    {28460, 2, {0x40, 0x27FC, 0x1040, 0x13F8, 0x8040, 0x47FE, 0x4800, 0xBF8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {28461, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8048, 0x47FC, 0x40A0, 0x1110, 0x1208, 0x2514, 0xE110, 0x27FE, 0x2110, 0x2110, 0x2210, 0x410}},
    {28462, 2, {0x40, 0x2020, 0x17FE, 0x1420, 0x85FC, 0x4424, 0x47FE, 0x1424, 0x15FC, 0x2420, 0xE562, 0x24B4, 0x2928, 0x2A26, 0x30A0, 0x40}},
    {28463, 2, {0x0, 0x27FC, 0x1444, 0x17FC, 0x8444, 0x47FC, 0x4100, 0x1210, 0x17E0, 0x20C8, 0xE304, 0x2FFE, 0x2042, 0x2248, 0x2544, 0x882}},
    {28464, 2, {0x40, 0x2444, 0x1444, 0x17FC, 0x8000, 0x47BC, 0x44A4, 0x14A4, 0x17BC, 0x24A4, 0xE4A4, 0x27BC, 0x24A4, 0x24A4, 0x2AB4, 0x1148}},
    {28465, 2, {0x108, 0x2108, 0x17C8, 0x1108, 0x811E, 0x47D2, 0x4564, 0x1540, 0x17C8, 0x2108, 0xE388, 0x2548, 0x2954, 0x2114, 0x2124, 0x142}},
    {28466, 2, {0x0, 0x4F3E, 0x2120, 0x213C, 0x120, 0x8F3C, 0x4820, 0x48FE, 0x1850, 0x2F52, 0xE154, 0x2148, 0x2148, 0x2144, 0x2A62, 0x440}},
    {28467, 2, {0x40, 0x27FC, 0x1110, 0x10A0, 0x8FFE, 0x4000, 0x43F8, 0x1208, 0x13F8, 0x2208, 0xE3F8, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x40}},
    {28468, 2, {0x40, 0x2444, 0x1444, 0x17FC, 0x8080, 0x4040, 0x47FE, 0x1402, 0x1BF4, 0x2000, 0xE7FC, 0x2040, 0x2250, 0x2448, 0x2944, 0x80}},
    {28469, 2, {0x108, 0x2108, 0x1288, 0x1250, 0x841E, 0x4B94, 0x4124, 0x1114, 0x17D4, 0x2114, 0xE594, 0x2548, 0x2948, 0x2114, 0x2524, 0x242}},
    {28470, 2, {0x40, 0x2040, 0x13F8, 0x1248, 0x83F8, 0x4040, 0x47FC, 0x1444, 0x17FC, 0x2040, 0xE080, 0x2048, 0x2544, 0x2514, 0x28F0, 0x0}},
    {28471, 2, {0x200, 0x211E, 0x17D2, 0x1012, 0x87D4, 0x4454, 0x47D8, 0x1014, 0x17D2, 0x2092, 0xE112, 0x21DA, 0x2714, 0x2110, 0x2510, 0x210}},
    {28472, 2, {0x200, 0x4202, 0x223C, 0x2FA0, 0x220, 0x8FA0, 0x4ABE, 0x4FA4, 0x1AA4, 0x2FA4, 0xE224, 0x2FA4, 0x2224, 0x2244, 0x2244, 0x284}},
    {28473, 2, {0x0, 0x27FC, 0x1040, 0x1278, 0x8240, 0x4FFE, 0x4100, 0x11FC, 0x1200, 0x23FC, 0xE004, 0x2554, 0x2554, 0x2804, 0x2028, 0x10}},
    {28474, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x8554, 0x4248, 0x4554, 0x1040, 0x1554, 0x2248, 0xE554, 0x20A0, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28475, 2, {0x0, 0x27BC, 0x1084, 0x14A4, 0x8294, 0x44A4, 0x4050, 0x1188, 0x1626, 0x20C0, 0xE310, 0x2064, 0x2388, 0x2030, 0x20C0, 0x700}},
    {28476, 2, {0x40, 0x2444, 0x17FC, 0x1120, 0x8110, 0x43FE, 0x5220, 0x1620, 0x1BFC, 0x2220, 0xE220, 0x23FC, 0x2220, 0x2220, 0x23FE, 0x200}},
    {28477, 2, {0x0, 0x27FC, 0x1404, 0x17FC, 0x8420, 0x44A8, 0x4470, 0x14A8, 0x1504, 0x24A0, 0xE4FC, 0x2520, 0x2420, 0x2BFE, 0x2820, 0x1020}},
    {28478, 2, {0x208, 0x2110, 0x17FC, 0x1040, 0x83F8, 0x4040, 0x47FC, 0x1040, 0x1020, 0x21C0, 0xE044, 0x2768, 0x2150, 0x2248, 0x2D46, 0x80}},
    {28479, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8518, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {28480, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x87FC, 0x4000, 0x401C, 0x17E0, 0x1248, 0x2248, 0xE248, 0x2554, 0x2554, 0x2862, 0x2040, 0x40}},
    {28481, 2, {0x2000, 0x24FE, 0x3810, 0x2220, 0x22FC, 0x1E84, 0x84, 0x8FC, 0xA84, 0xECFC, 0x2884, 0x2C84, 0x4AFC, 0x4A48, 0xA884, 0x1102}},
    {28482, 2, {0x40, 0x2444, 0x1444, 0x17FC, 0x8000, 0x4208, 0x4110, 0x17FC, 0x1110, 0x2110, 0xE514, 0x2318, 0x2110, 0x2110, 0x2FFE, 0x0}},
    {28483, 2, {0x240, 0x4440, 0x28FC, 0x3088, 0x350, 0x8420, 0x4CD8, 0x5706, 0x14F8, 0x2488, 0xE4F8, 0x2488, 0x24F8, 0x2488, 0x2488, 0x498}},
    {28484, 2, {0x108, 0x2108, 0x17C8, 0x1110, 0x811E, 0x47E4, 0x4554, 0x1554, 0x17D4, 0x2114, 0xE394, 0x2548, 0x2948, 0x2114, 0x2114, 0x122}},
    {28485, 2, {0x100, 0x2100, 0x13FC, 0x1440, 0x8080, 0x4FFE, 0x4120, 0x1210, 0x17FC, 0x2A20, 0xE3FC, 0x2220, 0x23FC, 0x2220, 0x23FE, 0x200}},
    {28486, 2, {0x110, 0x4FFE, 0x2110, 0x2000, 0xFFE, 0x8882, 0x5114, 0x43E0, 0x1040, 0x2088, 0xE7FC, 0x2044, 0x2250, 0x2448, 0x2944, 0x80}},
    {28487, 2, {0x110, 0x4110, 0x2FFE, 0x2110, 0x40, 0x87FC, 0x4044, 0x5FFE, 0x1044, 0x27FC, 0xE040, 0x2564, 0x2554, 0x2654, 0x2844, 0x1044}},
    {28488, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x84A4, 0x47FC, 0x4080, 0x1144, 0x1668, 0x20B0, 0xE128, 0x2668, 0x20A4, 0x2122, 0x26A0, 0x40}},
    {28489, 2, {0x40, 0x2444, 0x1444, 0x17FC, 0x8000, 0x47FC, 0x4404, 0x17FC, 0x1404, 0x27FC, 0xE000, 0x2424, 0x27A8, 0x2432, 0x25A2, 0x61E}},
    {28490, 2, {0x4210, 0x2210, 0x2510, 0x849E, 0x4810, 0x5710, 0x27C, 0x2224, 0x2FA4, 0x4224, 0xCB28, 0x4AA8, 0x5290, 0x4228, 0x4A44, 0x482}},
    {28491, 2, {0x4210, 0x2210, 0x2510, 0x84A0, 0x483E, 0x5744, 0x24, 0x24A4, 0x32A4, 0x4AA8, 0xC8A8, 0x4110, 0x43A8, 0x5C28, 0x4044, 0x82}},
    {28492, 2, {0x80, 0x2040, 0x17FC, 0x1514, 0x8208, 0x47FC, 0x4080, 0x1110, 0x13F8, 0x2048, 0xE040, 0x23F8, 0x2040, 0x2040, 0x2FFE, 0x0}},
    {28493, 2, {0x450, 0x8448, 0x4480, 0x8FE, 0x8B90, 0x5E90, 0x44FC, 0x490, 0x2890, 0x5EFC, 0xC890, 0x4090, 0x4690, 0x58FE, 0x4080, 0x80}},
    {28494, 2, {0x108, 0x2108, 0x1548, 0x1390, 0x811E, 0x47D4, 0x4564, 0x1554, 0x1554, 0x2754, 0xE5D4, 0x2548, 0x2548, 0x2554, 0x2464, 0x4C2}},
    {28495, 2, {0x0, 0x27F8, 0x1090, 0x1060, 0x87FC, 0x40A4, 0x4128, 0x16A0, 0x1040, 0x27FC, 0xE4A4, 0x2514, 0x26EC, 0x24A4, 0x24E4, 0x40C}},
    {28496, 2, {0x140, 0x2120, 0x13FE, 0x1220, 0x8620, 0x4BFC, 0x4220, 0x1220, 0x13FC, 0x2220, 0xE220, 0x23FE, 0x2200, 0x2524, 0x2492, 0x892}},
    {28497, 2, {0x24, 0x23A8, 0x1092, 0x1514, 0x8208, 0x4C04, 0x47BA, 0x10A8, 0x17C6, 0x2400, 0xE7BC, 0x2084, 0x20A8, 0x2090, 0x2528, 0x244}},
    {28498, 2, {0x100, 0x41F0, 0x2210, 0x27FC, 0xA24, 0x8244, 0x43FC, 0x5080, 0x1144, 0x26A8, 0xE130, 0x2668, 0x20A8, 0x2124, 0x26A2, 0x40}},
    {28499, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x7FC, 0x8444, 0x47FC, 0x5444, 0x17FC, 0x2044, 0xEFFE, 0x2042, 0x2024, 0x2522, 0x250A, 0x8F8}},
    {28500, 2, {0x100, 0x27BC, 0x1114, 0x1394, 0x8114, 0x47A4, 0x414C, 0x1080, 0x1110, 0x23E0, 0xE048, 0x27FC, 0x2044, 0x2248, 0x2544, 0x80}},
    {28501, 2, {0x100, 0x2100, 0x13FC, 0x16A8, 0x8AA8, 0x42A8, 0x47FC, 0x12A8, 0x12A8, 0x22A8, 0xEFFE, 0x2000, 0x22A8, 0x2254, 0x2454, 0x0}},
    {28502, 2, {0x0, 0x4FBE, 0x2208, 0x27BC, 0x208, 0x8FBE, 0x4000, 0x57FC, 0x1444, 0x2444, 0xE7FC, 0x2400, 0x2402, 0x2402, 0x23FE, 0x0}},
    {28503, 2, {0x140, 0x2120, 0x13FE, 0x1620, 0x8BFC, 0x4220, 0x43FC, 0x1220, 0x13FE, 0x2240, 0xE040, 0x2FFE, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28504, 2, {0x78, 0x27C0, 0x1248, 0x1150, 0x87FC, 0x4150, 0x4248, 0x1406, 0x13F8, 0x2248, 0xE248, 0x23F8, 0x2248, 0x2248, 0x23F8, 0x208}},
    {28505, 2, {0x4, 0x203E, 0x17C0, 0x1244, 0x8128, 0x43F8, 0x4A08, 0xBFC, 0x1204, 0x13FE, 0xE402, 0x26AA, 0x2AAA, 0x2A02, 0x3014, 0x8}},
    {28506, 2, {0x40, 0x23F8, 0x1048, 0x17FE, 0x8048, 0x43F8, 0x4040, 0x1554, 0x175C, 0x2444, 0xE7FC, 0x2444, 0x275C, 0x2554, 0x2554, 0x844}},
    {28507, 2, {0x0, 0x27BC, 0x1108, 0x1528, 0x87BC, 0x4318, 0x55AA, 0x1946, 0x1000, 0x23F8, 0xE208, 0x2208, 0x23F8, 0x2208, 0x2208, 0x3F8}},
    {28508, 2, {0x110, 0x4110, 0x27BC, 0x2110, 0xFFE, 0x8110, 0x42A8, 0x5444, 0x13FA, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28509, 2, {0x40, 0x20A0, 0x1118, 0x16E6, 0x8000, 0x43F8, 0x4A08, 0xBF8, 0x1000, 0x17BC, 0xE084, 0x24A4, 0x2294, 0x24A4, 0x2294, 0x108}},
    {28510, 2, {0x10, 0x2790, 0x14BC, 0x14A4, 0x84C4, 0x47A8, 0x4110, 0x1128, 0x1146, 0x25C0, 0xE53C, 0x2524, 0x2524, 0x25A4, 0x2E3C, 0x24}},
    {28511, 2, {0x100, 0x2638, 0x1408, 0x1408, 0x8738, 0x4408, 0x4408, 0x17F8, 0x1200, 0x27FC, 0xE804, 0x2554, 0x2554, 0x2804, 0x2028, 0x10}},
    {28512, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x8420, 0x44A4, 0x439C, 0x1000, 0x1110, 0x27FC, 0xE110, 0x2110, 0x2FFE, 0x2110, 0x2208, 0x404}},
    {28513, 2, {0x210, 0x4110, 0x2FD0, 0x2010, 0x7BE, 0x84A4, 0x47D4, 0x5014, 0x1794, 0x2094, 0xE114, 0x21C8, 0x2F08, 0x2114, 0x2514, 0x222}},
    {28514, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x8110, 0x4FFE, 0x4040, 0x17FC, 0x1444, 0x27FC, 0xE444, 0x27FC, 0x2000, 0x2110, 0x2208, 0x404}},
    {28515, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x84A4, 0x47BC, 0x4484, 0x1444, 0x15F4, 0x2424, 0xE4A4, 0x2444, 0x24A4, 0x2504, 0x2414, 0x408}},
    {28516, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x84A4, 0x47BC, 0x4404, 0x15F4, 0x1444, 0x24E4, 0xE444, 0x2444, 0x25F4, 0x2404, 0x2414, 0x408}},
    {28517, 2, {0x0, 0x27FC, 0x10A0, 0x10A0, 0x87FC, 0x44A4, 0x44A4, 0x17FC, 0x1248, 0x2150, 0xE7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28518, 2, {0x40, 0x2040, 0x17FC, 0x10A0, 0x8514, 0x4208, 0x47FC, 0x1A0A, 0x13F8, 0x2208, 0xE3F8, 0x2040, 0x2248, 0x2444, 0x2944, 0x80}},
    {28519, 2, {0x208, 0x2110, 0x1000, 0x17FC, 0x8444, 0x4554, 0x44E4, 0x1444, 0x17FC, 0x2000, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x208}},
    {28520, 2, {0x80, 0x2100, 0x17FC, 0x14A4, 0x84A4, 0x4FFE, 0x4000, 0x103C, 0x17C0, 0x2144, 0xE244, 0x24E8, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28521, 2, {0x0, 0x27FC, 0x1444, 0x17FC, 0x8444, 0x47FC, 0x4110, 0x1110, 0x17FC, 0x2110, 0xE110, 0x2FFE, 0x2000, 0x2110, 0x2208, 0x404}},
    {28522, 2, {0x200, 0x43DE, 0x2512, 0x2112, 0xFD2, 0x8292, 0x425E, 0x5400, 0x13FC, 0x2204, 0xE204, 0x23FC, 0x2204, 0x2204, 0x23FC, 0x204}},
    {28523, 2, {0x40, 0x2248, 0x1150, 0x13F8, 0x8080, 0x47FC, 0x4110, 0x1288, 0x1516, 0x21E0, 0xE048, 0x23FC, 0x2044, 0x2248, 0x2444, 0xC0}},
    {28524, 2, {0x0, 0x27BC, 0x14A4, 0x14A4, 0x87BC, 0x4000, 0x4BF8, 0xA48, 0x13F8, 0x1248, 0xE3F8, 0x2040, 0x27FC, 0x2040, 0x2040, 0x40}},
    {28525, 2, {0x0, 0x27FC, 0x10A0, 0x17FC, 0x84A4, 0x47FC, 0x4000, 0x13F8, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x40}},
    {28526, 2, {0x100, 0x411E, 0x27D2, 0x2112, 0x112, 0x87DE, 0x4452, 0x57D2, 0x1452, 0x27DE, 0xE112, 0x2112, 0x2FD2, 0x2122, 0x212A, 0x144}},
    {28527, 2, {0x0, 0x23FC, 0x1004, 0x13FC, 0x8004, 0x43FC, 0x4800, 0xBBC, 0x1124, 0x17BC, 0xE008, 0x23FE, 0x2108, 0x2088, 0x20A8, 0x10}},
    {28528, 2, {0x40, 0x23F8, 0x1248, 0x13F8, 0x8040, 0x47FE, 0x4800, 0xBF8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {28529, 2, {0x40, 0x27FC, 0x1040, 0x13F8, 0x8000, 0x47FC, 0x4404, 0x13F8, 0x1000, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x2110, 0x27FC, 0x0}},
    {28530, 2, {0x110, 0x4E10, 0x2254, 0x2238, 0x210, 0x8F7C, 0x4244, 0x5344, 0x16FC, 0x2644, 0xEA44, 0x227C, 0x2244, 0x2244, 0x2254, 0x248}},
    {28531, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x83FC, 0x4222, 0x4AFC, 0xA28, 0x13FE, 0x1220, 0xE2FC, 0x2384, 0x22FC, 0x2484, 0x24FC, 0x884}},
    {28532, 2, {0x20, 0x4920, 0x25FA, 0x2224, 0x628, 0x8BFE, 0x4220, 0x5240, 0x16FC, 0x2B44, 0xE244, 0x227C, 0x2244, 0x2244, 0x2A7C, 0x444}},
    {28533, 2, {0x288, 0x2288, 0x17C8, 0x1290, 0x829E, 0x4FD4, 0x4024, 0x17D4, 0x1454, 0x27D4, 0xE454, 0x27C8, 0x2448, 0x2454, 0x2464, 0x4C2}},
    {28534, 2, {0x0, 0x27FC, 0x1444, 0x1554, 0x84E4, 0x4444, 0x47FC, 0x1040, 0x17FC, 0x2040, 0xE040, 0x2FFE, 0x2000, 0x2524, 0x2492, 0x892}},
    {28535, 2, {0x210, 0x23DE, 0x1528, 0x1084, 0x8040, 0x43F8, 0x4048, 0x17FE, 0x1048, 0x23F8, 0xE040, 0x23FC, 0x2040, 0x27FE, 0x2040, 0x40}},
    {28536, 2, {0x420, 0x8420, 0x5FFC, 0x420, 0x9568, 0x55AA, 0x6526, 0x0, 0x2FF8, 0x4808, 0xCFF8, 0x4808, 0x4FF8, 0x4808, 0x4828, 0x810}},
    {28537, 2, {0x40, 0x2040, 0x17FC, 0x10E0, 0x8150, 0x464C, 0x4040, 0x1110, 0x1110, 0x27BC, 0xE110, 0x23B8, 0x2554, 0x2912, 0x2110, 0x110}},
    {28538, 2, {0x7FC, 0x2404, 0x17FC, 0x1400, 0x85F8, 0x4410, 0x47FE, 0x1420, 0x1460, 0x279C, 0xE484, 0x2508, 0x29DE, 0x2F08, 0x3108, 0x318}},
    {28539, 2, {0x38, 0x23C0, 0x1040, 0x17FC, 0x8150, 0x4248, 0x44A4, 0x1110, 0x1248, 0x2C46, 0xE248, 0x2150, 0x2248, 0x2444, 0x2140, 0x80}},
    {28540, 2, {0x40, 0x23F8, 0x1110, 0x10A0, 0x87FC, 0x4000, 0x4BF8, 0xA48, 0x13F8, 0x1248, 0xE3F8, 0x2040, 0x23F8, 0x2040, 0x27FE, 0x0}},
    {28541, 2, {0x208, 0x2110, 0x17FC, 0x10A0, 0x84A4, 0x42A8, 0x4FFE, 0x1000, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28542, 2, {0x40, 0x2248, 0x1150, 0x17FC, 0x8150, 0x4248, 0x4444, 0x1208, 0x1208, 0x23BE, 0xE488, 0x2AA8, 0x213E, 0x2208, 0x2408, 0x808}},
    {28543, 2, {0x0, 0x27FC, 0x1444, 0x15F4, 0x8494, 0x47FC, 0x4404, 0x15F4, 0x1514, 0x25F4, 0xE444, 0x24F4, 0x2544, 0x25FC, 0x2444, 0x7FC}},
    {28544, 2, {0x0, 0x4FBC, 0x2294, 0x26B4, 0xAD4, 0x84A4, 0x45AC, 0x5840, 0x1108, 0x2108, 0xE5A8, 0x252C, 0x2528, 0x25A8, 0x2E7E, 0x0}},
    {28545, 2, {0x4040, 0x2040, 0x227C, 0x8240, 0x4240, 0x4FFE, 0x0, 0x2208, 0x2208, 0x4228, 0xCBAE, 0x4A28, 0x4A28, 0x4BA8, 0x5E7E, 0x0}},
    {28546, 2, {0x108, 0x2548, 0x1548, 0x1550, 0x87DE, 0x4014, 0x47E4, 0x1014, 0x17D4, 0x2114, 0xE7D4, 0x2108, 0x2108, 0x2FD4, 0x2024, 0x42}},
    {28547, 2, {0xBFE, 0x1040, 0x61FC, 0x904, 0x11FC, 0x6104, 0x5FC, 0x904, 0x31FC, 0xC088, 0x10C, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {28548, 2, {0x20, 0x27A4, 0x10A8, 0x1292, 0x8114, 0x4208, 0x55F4, 0x1802, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2110, 0x2FFE, 0x0}},
    {28549, 2, {0x40, 0x23F8, 0x1048, 0x17FE, 0x8048, 0x43F8, 0x4040, 0x17FE, 0x1000, 0x23F8, 0xE248, 0x23F8, 0x2248, 0x23F8, 0x2000, 0x7FE}},
    {28550, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0xFFE, 0x8110, 0x47BC, 0x5110, 0x1190, 0x263E, 0xE000, 0x2FFE, 0x2120, 0x2122, 0x2222, 0xC1E}},
    {28551, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4000, 0x57FC, 0x1484, 0x2888, 0xE7F8, 0x2088, 0x2088, 0x2108, 0x2228, 0x410}},
    {28552, 2, {0x208, 0x2108, 0x17C8, 0x1210, 0x849E, 0x4FD4, 0x4064, 0x1794, 0x1494, 0x2794, 0xE494, 0x2788, 0x2488, 0x2494, 0x24A4, 0x5C2}},
    {28553, 2, {0x8, 0x4788, 0x2088, 0x2110, 0xFDE, 0x8494, 0x44A4, 0x5794, 0x1494, 0x2794, 0xE494, 0x24C8, 0x2788, 0x2C94, 0x20A4, 0xC2}},
    {28554, 2, {0x108, 0x2090, 0x17FE, 0x1090, 0x83FC, 0x4294, 0x4B1C, 0xA04, 0x13FC, 0x1204, 0xE3FC, 0x2008, 0x27FE, 0x2108, 0x20A8, 0x10}},
    {28555, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4040, 0x17FE, 0x1000, 0x23F8, 0xE208, 0x23F8, 0x2040, 0x2248, 0x2444, 0xC0}},
    {28556, 2, {0x480, 0x4482, 0x249C, 0x2FD0, 0x490, 0x8790, 0x449E, 0x5794, 0x1494, 0x2494, 0xEFD4, 0x2014, 0x2514, 0x28A4, 0x3024, 0x44}},
    {28557, 2, {0x104, 0x4104, 0x2FE4, 0x2104, 0x13E, 0x87C4, 0x4004, 0x57E4, 0x1454, 0x2454, 0xE7C4, 0x2444, 0x2284, 0x20E4, 0x2F14, 0x408}},
    {28558, 2, {0x104, 0x4104, 0x2FE8, 0x2110, 0x100, 0x87C4, 0x4004, 0x57C8, 0x1450, 0x2442, 0xE7C2, 0x2444, 0x2284, 0x20E8, 0x2F10, 0x420}},
    {28559, 2, {0x0, 0x23F8, 0x1208, 0x1208, 0x83F8, 0x4000, 0x47BC, 0x14A4, 0x14A4, 0x27BC, 0xE040, 0x2444, 0x2444, 0x2444, 0x27FC, 0x4}},
    {28560, 2, {0x0, 0x47FC, 0x2040, 0x2FFE, 0x842, 0x8358, 0x4040, 0x5358, 0x1000, 0x27FC, 0xE000, 0x2FFE, 0x2100, 0x2208, 0x27FC, 0x204}},
    {28561, 2, {0x0, 0x27FE, 0x10A0, 0x17BC, 0x84A4, 0x44A4, 0x47BC, 0x10A0, 0x17FC, 0x2444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x404}},
    {28562, 2, {0x0, 0x20FE, 0x1010, 0x1720, 0x827C, 0x4244, 0x4244, 0x127C, 0x1244, 0x227C, 0xE3C4, 0x2E44, 0x207C, 0x2028, 0x2044, 0x82}},
    {28563, 2, {0x140, 0x2140, 0x127E, 0x1480, 0x897C, 0x4144, 0x427C, 0x1644, 0x1A7C, 0x2220, 0xE23C, 0x2244, 0x22A8, 0x2210, 0x2228, 0x2C6}},
    {28564, 2, {0x108, 0x2128, 0x1228, 0x17BE, 0x84A8, 0x44C8, 0x4488, 0x14FE, 0x1780, 0x2480, 0xE4BE, 0x24A2, 0x24A2, 0x27A2, 0x24BE, 0x22}},
    {28565, 2, {0x110, 0x2FFE, 0x1110, 0x1000, 0x87FC, 0x4150, 0x5150, 0x1FFE, 0x1150, 0x2150, 0xE7FC, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x40}},
    {28566, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x84A4, 0x47BC, 0x4444, 0x1444, 0x15F4, 0x2444, 0xE4E4, 0x2554, 0x264C, 0x2444, 0x2454, 0x408}},
    {28567, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x84A4, 0x47BC, 0x4404, 0x15F4, 0x1514, 0x25F4, 0xE514, 0x25F4, 0x2514, 0x2404, 0x2414, 0x408}},
    {28568, 2, {0x110, 0x2110, 0x17BC, 0x1110, 0x83B8, 0x4554, 0x4912, 0x1000, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28569, 2, {0x0, 0x2080, 0x173C, 0x1404, 0x87BC, 0x4404, 0x47FC, 0x1200, 0x13FC, 0x2400, 0xE7FE, 0x2002, 0x2552, 0x24AA, 0x28AA, 0x4}},
    {28570, 2, {0x100, 0x27FC, 0x1444, 0x1554, 0x8444, 0x47FC, 0x44E4, 0x1554, 0x1000, 0x2FFE, 0xE200, 0x23FC, 0x2004, 0x2004, 0x2028, 0x10}},
    {28571, 2, {0x100, 0x43F0, 0x2410, 0x2FFC, 0x1444, 0x87FC, 0x4444, 0x57FC, 0x1000, 0x2FFE, 0xE000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x3F8}},
    {28572, 2, {0x200, 0x217C, 0x1504, 0x1444, 0x85F4, 0x4444, 0x45F4, 0x1554, 0x1574, 0x25D4, 0xE554, 0x25F4, 0x24E4, 0x2554, 0x2444, 0x44C}},
    {28573, 2, {0x80, 0x4040, 0x2FFE, 0x2882, 0x548, 0x8514, 0x48F4, 0x4000, 0x17FC, 0x2444, 0xE7FC, 0x2444, 0x27FC, 0x2444, 0x2454, 0x408}},
    {28574, 2, {0x40, 0x207C, 0x1040, 0x17FE, 0x8442, 0x45F8, 0x4444, 0x15FC, 0x1508, 0x25F8, 0xE400, 0x25FE, 0x2420, 0x2BFE, 0x2888, 0x1306}},
    {28575, 2, {0x40, 0x4FFE, 0x2000, 0x27FC, 0x404, 0x85F4, 0x4514, 0x55F4, 0x1404, 0x27FC, 0xE008, 0x23F0, 0x2040, 0x2FFE, 0x2150, 0xE4E}},
    {28576, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x87BC, 0x40A0, 0x47BC, 0x14A4, 0x14A4, 0x27BC, 0xE4A4, 0x24A4, 0x27BC, 0x24A2, 0x20A2, 0x7E}},
    {28577, 2, {0x3F8, 0x4208, 0x2208, 0x23F8, 0x0, 0x87BC, 0x44A4, 0x54A4, 0x17BC, 0x2040, 0xEFFE, 0x2160, 0x2250, 0x2448, 0x3846, 0x40}},
    {28578, 2, {0x40, 0x2248, 0x1150, 0x17FE, 0x8402, 0x4804, 0x43F8, 0x1208, 0x13F8, 0x2000, 0xE7FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x404}},
    {28579, 2, {0x108, 0x4108, 0x2114, 0x2FD4, 0x122, 0x87C0, 0x445C, 0x57C8, 0x1448, 0x27C8, 0xE13E, 0x2FC8, 0x2108, 0x2108, 0x2108, 0x108}},
    {28580, 2, {0x0, 0x27FC, 0x14A4, 0x17FC, 0x8040, 0x43F8, 0x4040, 0x17FC, 0x1110, 0x20A0, 0xE3F8, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x40}},
    {28581, 2, {0x4400, 0x273C, 0x2914, 0x9214, 0x4FA4, 0x4ACC, 0xA80, 0x2FA8, 0x2ABE, 0x4AC8, 0xCF88, 0x4A88, 0x4ABE, 0x4A88, 0x4888, 0x1188}},
    {28582, 2, {0x0, 0x4FBE, 0x2088, 0x2510, 0x23E, 0x8122, 0x4FA2, 0x52BE, 0x1222, 0x223E, 0xE222, 0x2222, 0x223E, 0x2200, 0x2A14, 0x422}},
    {28583, 2, {0xA0, 0x20A0, 0x17FC, 0x14A4, 0x87FC, 0x44A4, 0x47FC, 0x1000, 0x17FC, 0x2000, 0xE3F8, 0x2208, 0x23F8, 0x2110, 0x20A0, 0xFFE}},
    {28584, 2, {0x210, 0x2210, 0x13DE, 0x1528, 0x8884, 0x4000, 0x47FC, 0x1040, 0x1248, 0x2248, 0xE248, 0x2554, 0x28E2, 0x2040, 0x2FFE, 0x0}},
    {28585, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4102, 0x8104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {28586, 2, {0x0, 0x47FC, 0x2040, 0x2FFE, 0x842, 0x8358, 0x4040, 0x5358, 0x10A0, 0x2398, 0xEC46, 0x23F0, 0x2010, 0x2320, 0x20C0, 0x20}},
    {28587, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x83F8, 0x4248, 0x4BF8, 0xA48, 0x13F8, 0x1040, 0xE7FC, 0x2444, 0x2454, 0x25F4, 0x2414, 0x40C}},
    {28588, 2, {0x440, 0x227C, 0x1084, 0x1128, 0x8620, 0x4250, 0x488C, 0xBF8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {28589, 2, {0x40, 0x2020, 0x17FE, 0x1028, 0x8224, 0x427E, 0x44C8, 0x1748, 0x117E, 0x2248, 0xE4C8, 0x2F7E, 0x2148, 0x2248, 0x247E, 0x840}},
    {28590, 2, {0x40, 0x40A0, 0x2318, 0x2DF6, 0x0, 0x87FC, 0x4554, 0x54E4, 0x17FC, 0x2000, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x208}},
    {28591, 2, {0x80, 0x40DC, 0x2284, 0x23D4, 0x448, 0x8A94, 0x4324, 0x4C40, 0x1248, 0x2150, 0xE7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28592, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x8208, 0x45F6, 0x4000, 0x17BC, 0x14A4, 0x24A4, 0xE7BC, 0x2210, 0x2210, 0x2528, 0x2944, 0x1082}},
    {28593, 2, {0x4000, 0x2FDC, 0x2854, 0x8854, 0x4FD4, 0x4A94, 0xAA6, 0x2FC0, 0x2A9C, 0x4A94, 0xCFD4, 0x4814, 0x5288, 0x5248, 0x6454, 0x822}},
    {28594, 2, {0xA0, 0x24A4, 0x12A8, 0x10A0, 0x87FC, 0x4110, 0x40A0, 0x17FC, 0x1040, 0x23F8, 0xE040, 0x27FC, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28595, 2, {0x80, 0x2100, 0x17FC, 0x1444, 0x8554, 0x4444, 0x47FC, 0x14E4, 0x1554, 0x2404, 0xE040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28596, 2, {0x0, 0x23FC, 0x1294, 0x1294, 0x83FC, 0x4000, 0x47FE, 0x1000, 0x13FC, 0x2204, 0xE3FC, 0x20A2, 0x2114, 0x2308, 0x2D44, 0x182}},
    {28597, 2, {0x200, 0x2102, 0x17DC, 0x1010, 0x8450, 0x4290, 0x47DE, 0x1114, 0x1114, 0x27D4, 0xE114, 0x2594, 0x2554, 0x2924, 0x2524, 0x244}},
    {28598, 2, {0x40, 0x4FFE, 0x2000, 0x27FC, 0x404, 0x85F4, 0x4514, 0x57FC, 0x1000, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2000, 0xFFE}},
    {28599, 2, {0x3F8, 0x2208, 0x13F8, 0x1208, 0x83F8, 0x4000, 0x47FC, 0x14A4, 0x17FC, 0x2040, 0xE7FE, 0x2100, 0x21F8, 0x2108, 0x2228, 0x410}},
    {28600, 2, {0x14, 0x2012, 0x17FE, 0x1410, 0x85D0, 0x4414, 0x45D4, 0x1558, 0x15CA, 0x2416, 0xE442, 0x2824, 0x22A2, 0x228A, 0x2478, 0x0}},
    {28601, 2, {0x200, 0x43F8, 0x2410, 0x27FE, 0xD48, 0x9624, 0x47FE, 0x5400, 0x15FC, 0x2400, 0xE5FC, 0x2400, 0x25FC, 0x2904, 0x29FC, 0x1104}},
    {28602, 2, {0x80, 0x2040, 0x13FC, 0x1108, 0x8090, 0x47FE, 0x4000, 0x13FC, 0x1204, 0x23FC, 0xE204, 0x23FC, 0x2040, 0x2524, 0x250A, 0x8FA}},
    {28603, 2, {0x88, 0x2448, 0x1250, 0x10FC, 0x8020, 0x4050, 0x4694, 0x1238, 0x1250, 0x2298, 0xE234, 0x2254, 0x2290, 0x2220, 0x25FE, 0x800}},
    {28604, 2, {0x10, 0x2008, 0x1788, 0x14BE, 0x8480, 0x4494, 0x4788, 0x147E, 0x1408, 0x2788, 0xE6BE, 0x2A88, 0x2A88, 0x2B88, 0x3288, 0x8}},
    {28605, 2, {0x40, 0x207C, 0x1040, 0x13FE, 0x8242, 0x4278, 0x4BC4, 0xA3C, 0x1200, 0x13FE, 0xE240, 0x23A4, 0x2258, 0x25B4, 0x2452, 0x9B0}},
    {28606, 2, {0x10, 0x2410, 0x127C, 0x1010, 0x80FE, 0x4044, 0x4628, 0x12FE, 0x1210, 0x227C, 0xE210, 0x22FE, 0x2210, 0x2210, 0x25FE, 0x800}},
    {28607, 2, {0x110, 0x2110, 0x17BC, 0x1110, 0x83B8, 0x4554, 0x4912, 0x1000, 0x13F8, 0x2000, 0xE7FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x80}},
    {28608, 2, {0x210, 0x4410, 0x2F90, 0x2890, 0xFBE, 0x88A4, 0x4FD4, 0x5414, 0x1214, 0x2FD4, 0xE414, 0x2788, 0x2488, 0x2894, 0x2A94, 0x1122}},
    {28609, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x87FC, 0x4200, 0x53FE, 0x1442, 0x1BFA, 0x224A, 0xE3FA, 0x2052, 0x207A, 0x278A, 0x2214, 0x8}},
    {28610, 2, {0x20, 0x27FE, 0x1508, 0x1490, 0x87FC, 0x4450, 0x45FC, 0x1454, 0x17FE, 0x2454, 0xE5FC, 0x2450, 0x24D8, 0x2954, 0x2A52, 0x1050}},
    {28611, 2, {0x90, 0x23FC, 0x1294, 0x13FC, 0x8294, 0x43FC, 0x4800, 0xBFC, 0x1200, 0x12F8, 0xE200, 0x23FE, 0x2520, 0x2514, 0x2548, 0x986}},
    {28612, 2, {0x0, 0x44F8, 0x2288, 0x22E8, 0xA8, 0x80A8, 0x4DFC, 0x4504, 0x1574, 0x2554, 0xE574, 0x2504, 0x2514, 0x2508, 0x2A00, 0x11FE}},
    {28613, 2, {0x40, 0x4FFE, 0x2802, 0x23F8, 0x8, 0x83F8, 0x4008, 0x57FC, 0x1444, 0x23F8, 0xE248, 0x2040, 0x27F8, 0x2110, 0x20E0, 0xF1E}},
    {28614, 2, {0x40, 0x27FC, 0x1040, 0x1110, 0x8FFE, 0x4110, 0x43F8, 0x1208, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x2110, 0x2208, 0x404}},
    {28615, 2, {0x40, 0x2040, 0x17FC, 0x1248, 0x8248, 0x4554, 0x4FFE, 0x1000, 0x17FC, 0x2404, 0xE5F4, 0x2514, 0x25F4, 0x2404, 0x27FC, 0x404}},
    {28616, 2, {0x3D4, 0x2252, 0x1252, 0x13D0, 0x8010, 0x4FFE, 0x4250, 0x13D4, 0x1254, 0x23D4, 0xE254, 0x22F8, 0x2F4A, 0x244A, 0x2056, 0x62}},
    {28617, 2, {0x28, 0x4024, 0x2F40, 0x297E, 0x9C8, 0x8F48, 0x497E, 0x5948, 0x1F48, 0x297E, 0xE948, 0x2948, 0x2F48, 0x297E, 0x2040, 0x40}},
    {28618, 2, {0x20, 0x213C, 0x1120, 0x17FE, 0x8010, 0x4014, 0x4BFE, 0xA10, 0x13F0, 0x1254, 0xE254, 0x22D4, 0x2368, 0x244A, 0x2896, 0x322}},
    {28619, 2, {0x110, 0x2110, 0x17BC, 0x1110, 0x83B8, 0x4554, 0x4912, 0x1000, 0x17FC, 0x2044, 0xE240, 0x2278, 0x2240, 0x2540, 0x28FE, 0x1000}},
    {28620, 2, {0x1C20, 0x7022, 0x11B4, 0xFCA8, 0x10A8, 0x7D24, 0x56A2, 0x7C40, 0x55FC, 0x7D04, 0x1104, 0x7DFC, 0x1104, 0x1F04, 0xF1FC, 0x4104}},
    {28621, 2, {0x110, 0x4FFE, 0x2110, 0x2200, 0x3FC, 0x84A4, 0x4BA4, 0x4164, 0x1254, 0x2494, 0xE108, 0x2040, 0x2024, 0x2522, 0x250A, 0x8F8}},
    {28622, 2, {0x1F8, 0x2108, 0x11F8, 0x150A, 0x85FA, 0x450A, 0x45FA, 0x1402, 0x179E, 0x2090, 0xE090, 0x2F9E, 0x2492, 0x2492, 0x2492, 0x892}},
    {28623, 2, {0x0, 0x27BE, 0x1108, 0x1108, 0x87BE, 0x4108, 0x4108, 0x17BE, 0x1000, 0x2080, 0xE044, 0x2512, 0x2522, 0x29C0, 0x2304, 0xCFC}},
    {28624, 2, {0x3FC, 0x2204, 0x13FC, 0x1204, 0x83FC, 0x4044, 0x4BF8, 0x850, 0x17FE, 0x1040, 0xE1FC, 0x2704, 0x21FC, 0x2104, 0x21FC, 0x104}},
    {28625, 2, {0x210, 0x4210, 0x2F9C, 0x2224, 0x248, 0x8FBE, 0x4AA2, 0x5AAA, 0x1FAA, 0x222A, 0xE72A, 0x2AAA, 0x3288, 0x2214, 0x2222, 0x242}},
    {28626, 2, {0x4200, 0x227E, 0x2A10, 0x8BA0, 0x4A7C, 0x4A44, 0x1FD4, 0x2054, 0x2254, 0x4AD4, 0xCAD4, 0x4B54, 0x5120, 0x4228, 0x4444, 0x1882}},
    {28627, 2, {0x40, 0x27FC, 0x1110, 0x10A0, 0x87FE, 0x44A0, 0x47F8, 0x14A8, 0x17FE, 0x24A8, 0xE7F8, 0x24A0, 0x25B0, 0x2AA8, 0x2CA6, 0x10A0}},
    {28628, 2, {0x0, 0x27FC, 0x1248, 0x1444, 0x8FFE, 0x4444, 0x46EC, 0x1554, 0x16EC, 0x2444, 0xE6EC, 0x2554, 0x26EC, 0x2444, 0x2454, 0x408}},
    {28629, 2, {0x0, 0x27FC, 0x1404, 0x17FC, 0x8404, 0x47FC, 0x4210, 0x14A4, 0x1738, 0x2210, 0xE4A4, 0x27BC, 0x2000, 0x2548, 0x24A4, 0x8A4}},
    {28630, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x87FC, 0x4048, 0x43F8, 0x1050, 0x17FE, 0x2040, 0xE1F8, 0x2308, 0x25F8, 0x2108, 0x21F8, 0x108}},
    {28631, 2, {0x110, 0x2FFE, 0x1110, 0x13F8, 0x8208, 0x43F8, 0x4208, 0x13F8, 0x1080, 0x2FFE, 0xE248, 0x27FC, 0x2A4A, 0x2248, 0x2258, 0x40}},
    {28632, 2, {0x80, 0x2040, 0x17FC, 0x1444, 0x82A8, 0x4294, 0x4474, 0x1000, 0x13F8, 0x22A8, 0xEFFE, 0x2000, 0x27FC, 0x2040, 0x2140, 0x80}},
    {28633, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4FFE, 0x4802, 0x1040, 0x2248, 0xE248, 0x2450, 0x20A0, 0x2110, 0x2208, 0xC06}},
    {28634, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4000, 0x57FC, 0x1444, 0x2040, 0xE7FC, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28635, 2, {0x110, 0x27FC, 0x1110, 0x17FC, 0x8404, 0x43F8, 0x4000, 0x17FC, 0x1080, 0x2144, 0xE6A8, 0x2170, 0x26A8, 0x2126, 0x26A0, 0x40}},
    {28636, 2, {0x40, 0x23F8, 0x1048, 0x17FE, 0x8048, 0x43F8, 0x4040, 0x17FE, 0x1000, 0x22A4, 0xE452, 0x23F8, 0x22A8, 0x22A8, 0x2FFE, 0x0}},
    {28637, 2, {0x108, 0x27FE, 0x1108, 0x11F8, 0x8108, 0x41F8, 0x4108, 0x17FE, 0x1108, 0x2294, 0xE5E2, 0x2048, 0x23FC, 0x2040, 0x2248, 0x4C4}},
    {28638, 2, {0x80, 0x23F8, 0x1208, 0x13F8, 0x8208, 0x43F8, 0x4000, 0x17FC, 0x1444, 0x27FC, 0xE444, 0x27FC, 0x2000, 0x2FFE, 0x2110, 0x210}},
    {28639, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0x80A4, 0x4F58, 0x4554, 0x1552, 0x1B58, 0x2000, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x408}},
    {28640, 2, {0x40, 0x27FE, 0x1000, 0x11F8, 0x8108, 0x41F8, 0x4000, 0x17FE, 0x1402, 0x23F8, 0xE180, 0x2644, 0x21B8, 0x2668, 0x21A6, 0xE60}},
    {28641, 2, {0x0, 0x47FC, 0x2040, 0x2FFE, 0x842, 0x8358, 0x4040, 0x5358, 0x1000, 0x2FFE, 0xE040, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x40C}},
    {28642, 2, {0x7BC, 0x2084, 0x14A4, 0x1294, 0x84A4, 0x4040, 0x47FC, 0x1110, 0x1110, 0x22A8, 0xE444, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x40}},
    {28643, 2, {0x4210, 0x2210, 0x2228, 0x9FA8, 0x4244, 0x4F8A, 0x8A8, 0x2F98, 0x28C8, 0x4FA8, 0xC20E, 0x5FB8, 0x4208, 0x4208, 0x4208, 0x208}},
    {28644, 2, {0x40, 0x27FC, 0x1040, 0x13FC, 0x8000, 0x47FE, 0x4002, 0x13F8, 0x1040, 0x27FE, 0xE000, 0x27FE, 0x2008, 0x27FE, 0x24A8, 0x798}},
    {28645, 2, {0x80, 0x41F8, 0x2690, 0x2060, 0x1A0, 0x8FFE, 0x4802, 0x47FC, 0x1040, 0x27FC, 0xE444, 0x27FC, 0x2444, 0x27FC, 0x2208, 0x404}},
    {28646, 2, {0x3E, 0x27C0, 0x1244, 0x1128, 0x83F8, 0x4040, 0x47FC, 0x1000, 0x13F8, 0x2008, 0xE3F8, 0x2008, 0x23F8, 0x2544, 0x252A, 0x8FA}},
    {28647, 2, {0x284, 0x4284, 0x2AA4, 0x26C4, 0x29E, 0x8FE4, 0x4004, 0x5454, 0x128C, 0x2FE4, 0xE104, 0x27C4, 0x2104, 0x21C4, 0x2E14, 0x8}},
    {28648, 2, {0x210, 0x2120, 0x17FC, 0x1210, 0x8210, 0x44A4, 0x4738, 0x1210, 0x14A4, 0x27BC, 0xE080, 0x2148, 0x2554, 0x2514, 0x28F0, 0x0}},
    {28649, 2, {0x108, 0x27FE, 0x1148, 0x1220, 0x87FE, 0x4A20, 0x43FC, 0x1220, 0x13FC, 0x2220, 0xE3FE, 0x2200, 0x27F8, 0x2110, 0x20E0, 0xF1E}},
    {28650, 2, {0x210, 0x4210, 0x2F9C, 0x2224, 0xFC8, 0x823E, 0x4FAA, 0x502A, 0x1FAA, 0x28BE, 0xEFA0, 0x28A0, 0x2FA2, 0x28A2, 0x2AA2, 0x91E}},
    {28651, 2, {0x10, 0x27D0, 0x1510, 0x17DE, 0x8450, 0x47E8, 0x4504, 0x17C4, 0x1000, 0x23F8, 0xE2A8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x0}},
    {28652, 2, {0x40, 0x207C, 0x1040, 0x17FE, 0x8402, 0x43FC, 0x4248, 0x14A4, 0x1110, 0x23F8, 0xE60C, 0x2BFA, 0x2208, 0x23F8, 0x2208, 0x3F8}},
    {28653, 2, {0x110, 0x27FC, 0x1110, 0x1040, 0x83F8, 0x4040, 0x47FC, 0x1080, 0x1110, 0x23F8, 0xE000, 0x23F8, 0x22A8, 0x22A8, 0x2FFE, 0x0}},
    {28654, 2, {0x128, 0x21AA, 0x116C, 0x1228, 0x82FE, 0x4644, 0x5628, 0x1AFE, 0x1210, 0x227C, 0xE210, 0x22FE, 0x2210, 0x2228, 0x2244, 0x282}},
    {28655, 2, {0x0, 0x27BC, 0x1084, 0x14A4, 0x8294, 0x44A4, 0x4120, 0x1210, 0x17FC, 0x2A20, 0xE3FC, 0x2220, 0x23FC, 0x2220, 0x23FE, 0x200}},
    {28656, 2, {0x4228, 0x2224, 0x2440, 0x857E, 0x49C8, 0x5F48, 0x27E, 0x2448, 0x2948, 0x5F7E, 0xC148, 0x4048, 0x5548, 0x557E, 0x5040, 0x40}},
    {28657, 2, {0x80, 0x4040, 0x27FE, 0x2402, 0xBFC, 0x8148, 0x4250, 0x57FC, 0x1204, 0x27FC, 0xEA04, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x204}},
    {28658, 2, {0x200, 0x421C, 0x2F94, 0x2214, 0x714, 0x8026, 0x4F80, 0x48BC, 0x1714, 0x2014, 0xE214, 0x2F88, 0x2708, 0x2A94, 0x3224, 0x242}},
    {28659, 2, {0x108, 0x4528, 0x27BE, 0x2948, 0x7BE, 0x8318, 0x45AA, 0x5946, 0x13F8, 0x2208, 0xE208, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x208}},
    {28660, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4FFE, 0x4842, 0x1040, 0x2764, 0xE168, 0x2150, 0x2248, 0x2444, 0x2942, 0x80}},
    {28661, 2, {0x20, 0x27FE, 0x1402, 0x13FC, 0x8224, 0x43FC, 0x4A00, 0xAFC, 0x1284, 0x12FC, 0xE284, 0x22FC, 0x2284, 0x24FC, 0x2448, 0x884}},
    {28662, 2, {0x7BC, 0x24A4, 0x17BC, 0x14A4, 0x87BC, 0x4404, 0x4434, 0x15C4, 0x1444, 0x27FC, 0xE444, 0x25F4, 0x2514, 0x25F4, 0x2414, 0x408}},
    {28663, 2, {0x108, 0x7990, 0x960, 0x3518, 0xC206, 0x110, 0x47FC, 0x2110, 0x81F0, 0x4040, 0x17FC, 0x1664, 0xE554, 0x26EC, 0x2444, 0x244C}},
    {28664, 2, {0x110, 0x2FFE, 0x1110, 0x1040, 0x81B0, 0x4E4E, 0x43F8, 0x1208, 0x13F8, 0x2208, 0xE3F8, 0x2200, 0x25FC, 0x2504, 0x29FC, 0x104}},
    {28665, 2, {0x0, 0x27FC, 0x1444, 0x17FC, 0x8444, 0x47FC, 0x4040, 0x17FC, 0x1040, 0x2FFE, 0xE524, 0x2892, 0x2040, 0x27FC, 0x2040, 0xFFE}},
    {28666, 2, {0x14, 0x2792, 0x149E, 0x14F0, 0x8794, 0x448A, 0x4496, 0x17E2, 0x1494, 0x2492, 0xE79E, 0x2070, 0x2514, 0x248A, 0x2896, 0x62}},
    {28667, 2, {0x10, 0x277C, 0x1554, 0x157C, 0x8510, 0x46FE, 0x4500, 0x157C, 0x1544, 0x257C, 0xE544, 0x267C, 0x2444, 0x247C, 0x2428, 0x444}},
    {28668, 2, {0x444, 0x44E4, 0x28A8, 0x2AAA, 0xEEE, 0x84A4, 0x4AAA, 0x4EEE, 0x1242, 0x2040, 0xEFFE, 0x20E0, 0x2150, 0x2248, 0x2C46, 0x40}},
    {28669, 2, {0x208, 0x4FBE, 0x2208, 0x2FBE, 0x208, 0x8514, 0x48A2, 0x57FC, 0x1404, 0x27FC, 0xE404, 0x27FC, 0x2404, 0x27FC, 0x2110, 0x208}},
    {28670, 2, {0x20, 0x203E, 0x1020, 0x17FE, 0x8422, 0x45F8, 0x4422, 0x15FE, 0x1524, 0x25FC, 0xE524, 0x25FC, 0x2840, 0x2AA4, 0x328A, 0x47A}},
    {28671, 2, {0x0, 0x27FE, 0x1488, 0x17FE, 0x8488, 0x45FC, 0x4524, 0x15FC, 0x1524, 0x25FC, 0xE420, 0x27FE, 0x262A, 0x2AFA, 0x2A0A, 0x1206}},
    {28672, 2, {0x7FC, 0x4080, 0x23F8, 0x2208, 0x3F8, 0x8208, 0x4FFE, 0x5882, 0x1548, 0x2514, 0xE8F4, 0x2100, 0x23F8, 0x2508, 0x20F0, 0xF0E}},
    {28673, 2, {0x208, 0x2110, 0x17FC, 0x1040, 0x83F8, 0x4080, 0x47FC, 0x1148, 0x13FC, 0x260A, 0xEBF8, 0x2208, 0x23FA, 0x2234, 0x228C, 0x302}},
    {28674, 2, {0x100, 0x43F0, 0x2410, 0x2FFC, 0x1444, 0x87FC, 0x4444, 0x57FC, 0x1524, 0x2892, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x208}},
    {28675, 2, {0x4000, 0x2F7C, 0x2944, 0x897C, 0x4F44, 0x4944, 0x97C, 0x2F20, 0x297E, 0x49AA, 0xCF2A, 0x404A, 0x4A92, 0x4922, 0x494A, 0x1104}},
    {28676, 2, {0x210, 0x2210, 0x13DE, 0x1528, 0x8084, 0x4000, 0x47BC, 0x14A4, 0x17A4, 0x24A4, 0xE7A4, 0x2434, 0x2528, 0x26A0, 0x2420, 0x20}},
    {28677, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4FFE, 0x4802, 0x13F8, 0x2040, 0xE040, 0x23F8, 0x2050, 0x2048, 0x2FFE, 0x0}},
    {28678, 2, {0x40, 0x4FFE, 0x2040, 0x27FC, 0x0, 0x87FC, 0x44A4, 0x57FC, 0x1208, 0x23F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2110, 0x208}},
    {28679, 2, {0x40, 0x2020, 0x17FE, 0x1488, 0x85FC, 0x4488, 0x47FE, 0x1420, 0x15FC, 0x2524, 0xE5FC, 0x2524, 0x29FC, 0x2888, 0x3104, 0x202}},
    {28680, 2, {0x7BC, 0x24A4, 0x1294, 0x14A4, 0x8FFE, 0x4842, 0x47FC, 0x1040, 0x13F8, 0x2248, 0xE3F8, 0x2248, 0x23F8, 0x2040, 0x27FC, 0x40}},
    {28681, 2, {0x40, 0x4FFE, 0x2882, 0x2338, 0x208, 0x83B8, 0x4208, 0x53F8, 0x1200, 0x27FC, 0xE804, 0x2554, 0x2554, 0x2804, 0x2028, 0x10}},
    {28682, 2, {0x100, 0x23DC, 0x1254, 0x1366, 0x8240, 0x47DC, 0x4254, 0x1348, 0x1254, 0x24A2, 0xE000, 0x23FC, 0x2294, 0x2294, 0x27FE, 0x0}},
    {28683, 2, {0x40, 0x47FC, 0x2404, 0x2838, 0x7C0, 0x8248, 0x4150, 0x57FC, 0x1248, 0x2404, 0xEBFA, 0x2248, 0x23F8, 0x2248, 0x23F8, 0x208}},
    {28684, 2, {0x80, 0x2040, 0x17FC, 0x14A0, 0x87FC, 0x44A4, 0x47FC, 0x1400, 0x1524, 0x25A8, 0xE530, 0x25A4, 0x251C, 0x2800, 0x2AA4, 0x1452}},
    {28685, 2, {0x20, 0x27FE, 0x1400, 0x15FC, 0x8524, 0x45FC, 0x4524, 0x15FC, 0x1420, 0x25FC, 0xE420, 0x27FE, 0x2524, 0x2AFA, 0x2820, 0x13FE}},
    {28686, 2, {0x108, 0x27FE, 0x1108, 0x1000, 0x83FC, 0x4294, 0x4A94, 0xBFC, 0x1028, 0x13FE, 0xE220, 0x2324, 0x22A8, 0x2292, 0x242A, 0x8C6}},
    {28687, 2, {0x202, 0x25E2, 0x14A2, 0x16AA, 0x852A, 0x466A, 0x410A, 0x128A, 0x144A, 0x2BAA, 0xE10A, 0x27CA, 0x2122, 0x2542, 0x21EA, 0xE04}},
    {28688, 2, {0x114, 0x2112, 0x17D2, 0x1110, 0x8110, 0x4FFE, 0x4290, 0x16D2, 0x1292, 0x26D2, 0xE294, 0x26D4, 0x228A, 0x22CA, 0x2F16, 0x22}},
    {28689, 2, {0x3F8, 0x2208, 0x13F8, 0x1208, 0x83F8, 0x4110, 0x47FC, 0x1110, 0x1FFE, 0x2110, 0xE248, 0x2554, 0x20E0, 0x2150, 0x2248, 0xC0}},
    {28690, 2, {0x40, 0x4FFE, 0x2248, 0x2554, 0x8E2, 0x8150, 0x4248, 0x4DF6, 0x1000, 0x27FC, 0xE404, 0x25F4, 0x2514, 0x25F4, 0x2404, 0x7FC}},
    {28691, 2, {0x288, 0x42A8, 0x26A8, 0x2BF0, 0x1E, 0x8224, 0x45D4, 0x5C14, 0x15D4, 0x2494, 0xE494, 0x25C8, 0x2488, 0x2494, 0x25D4, 0x422}},
    {28692, 2, {0x200, 0x9FDC, 0x4214, 0xF94, 0x8014, 0x5FD4, 0x5066, 0xF80, 0x201C, 0x4F94, 0xC214, 0x5FD4, 0x4708, 0x4A88, 0x5254, 0x222}},
    {28693, 2, {0x100, 0x213E, 0x1508, 0x15D0, 0x853E, 0x4522, 0x4FE2, 0x103E, 0x1122, 0x257E, 0xE562, 0x25A2, 0x28BE, 0x2100, 0x2214, 0xC22}},
    {28694, 2, {0x7FC, 0x2040, 0x17FE, 0x1442, 0x8B5C, 0x4040, 0x475C, 0x1120, 0x13FC, 0x2620, 0xEBFC, 0x2220, 0x23FC, 0x2220, 0x23FE, 0x200}},
    {28695, 2, {0x40, 0x4FFE, 0x2842, 0x27FC, 0x40, 0x83F8, 0x4040, 0x57FC, 0x1040, 0x27FC, 0xE4A4, 0x27FC, 0x2040, 0x2524, 0x252A, 0x8FA}},
    {28696, 2, {0x40, 0x207C, 0x1040, 0x13FC, 0x8244, 0x43F0, 0x4A44, 0xAFC, 0x12A8, 0x12F8, 0xE2A8, 0x22F8, 0x2200, 0x25FC, 0x2554, 0xBFE}},
    {28697, 2, {0x4400, 0x223C, 0x2FA4, 0x8024, 0x48BC, 0x4524, 0xFA4, 0x223C, 0x2224, 0x4FA4, 0xC23C, 0x4B10, 0x4A98, 0x52A8, 0x4A2A, 0x446}},
    {28698, 2, {0x410, 0x8410, 0x4428, 0x1F44, 0x8482, 0x5F00, 0x51EE, 0x1F22, 0x31AA, 0x5F66, 0xC422, 0x5F66, 0x44AA, 0x4422, 0x44AA, 0x444}},
    {28699, 2, {0x80, 0xBFFC, 0x4800, 0xFF8, 0x8000, 0x4FF8, 0x4808, 0xFF8, 0x2080, 0x5C9C, 0xD7D4, 0x5D54, 0x555C, 0x5C96, 0x5556, 0x2E22}},
    {28700, 2, {0x8, 0x4F88, 0x2008, 0x2708, 0x53E, 0x852A, 0x472A, 0x502A, 0x1FAA, 0x2DBE, 0xEAA8, 0x2F88, 0x2A8A, 0x2A8E, 0x2ABA, 0x990}},
    {28701, 2, {0x0, 0x27FE, 0x1444, 0x1598, 0x8488, 0x47DE, 0x4488, 0x15DC, 0x16AA, 0x2488, 0xE420, 0x2520, 0x253C, 0x2920, 0x2920, 0x17FE}},
    {28702, 2, {0x220, 0x4220, 0x2FB8, 0x2248, 0xF90, 0x827C, 0x4F94, 0x5014, 0x1FFE, 0x2894, 0xEF94, 0x28FC, 0x2F94, 0x2890, 0x2A90, 0x930}},
    {28703, 2, {0x110, 0x27FE, 0x1150, 0x13F8, 0x8048, 0x47FE, 0x4048, 0x13F8, 0x1554, 0x275C, 0xE444, 0x27FC, 0x2444, 0x275C, 0x2554, 0x954}},
    {28704, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x47FC, 0x5484, 0x1110, 0x23E0, 0xE048, 0x27FC, 0x2044, 0x2248, 0x2544, 0x80}},
    {28705, 2, {0x10, 0x4E7E, 0x2B20, 0x2ABC, 0xAC8, 0x8CBE, 0x4A00, 0x4BBC, 0x1AA4, 0x2ABC, 0xEAA4, 0x2CBC, 0x28A4, 0x28AC, 0x2940, 0xA3E}},
    {28706, 2, {0x20, 0x49FC, 0x2524, 0x21FC, 0x20, 0x83FE, 0x4D04, 0x45FC, 0x1504, 0x25FC, 0xE504, 0x25FC, 0x2488, 0x2504, 0x2BFE, 0x1000}},
    {28707, 2, {0x100, 0x41BC, 0x2524, 0x27A8, 0x490, 0x8AA8, 0x4146, 0x52A0, 0x17BC, 0x20A0, 0xE3B8, 0x20A0, 0x27BC, 0x20A0, 0x2FFE, 0x0}},
    {28708, 2, {0x80, 0x4040, 0x2FFE, 0x2000, 0x7FC, 0x84A4, 0x47FC, 0x5248, 0x1150, 0x2208, 0xE444, 0x20A8, 0x2110, 0x2308, 0x2D46, 0x180}},
    {28709, 2, {0x7BC, 0x24A4, 0x1294, 0x14A4, 0x8040, 0x47FC, 0x4000, 0x11F0, 0x1110, 0x21F0, 0xE000, 0x27FC, 0x2404, 0x24E4, 0x24A4, 0x4EC}},
    {28710, 2, {0x10, 0x9F90, 0x427A, 0x412, 0x8C14, 0x52FE, 0x4708, 0xA10, 0x333C, 0x46E4, 0xCAA4, 0x523C, 0x4224, 0x4224, 0x4A3C, 0x424}},
    {28711, 2, {0x210, 0x2110, 0x17DE, 0x1010, 0x845E, 0x4282, 0x47DE, 0x1010, 0x17DE, 0x2450, 0xE7DE, 0x2450, 0x27DE, 0x2450, 0x2452, 0x4CE}},
    {28712, 2, {0x108, 0x2108, 0x17CE, 0x1112, 0x8124, 0x47DE, 0x4552, 0x155E, 0x17D2, 0x211E, 0xE392, 0x255E, 0x2940, 0x2114, 0x2112, 0x122}},
    {28713, 2, {0x180, 0x4E3E, 0x2208, 0x3F90, 0x23E, 0x8722, 0x4AA2, 0x527E, 0x1022, 0x273E, 0xE522, 0x2522, 0x257E, 0x2980, 0x2914, 0x1022}},
    {28714, 2, {0x2020, 0x7F3E, 0xAA44, 0x7FA8, 0x2A10, 0x3F28, 0x444, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x184, 0x3D48, 0x930, 0x730E}},
    {28715, 2, {0x200, 0x9FDC, 0x4214, 0xF94, 0x8014, 0x5FD4, 0x5066, 0xF80, 0x241C, 0x4894, 0xCF14, 0x4294, 0x5FC8, 0x4A88, 0x5254, 0x622}},
    {28716, 2, {0x4200, 0x223E, 0x2F88, 0x8210, 0x423E, 0x4FA2, 0xAA2, 0x2ABE, 0x2FA2, 0x423E, 0xC722, 0x46A2, 0x4ABE, 0x5200, 0x4214, 0x222}},
    {28717, 2, {0x410, 0x8410, 0x4428, 0x1F44, 0x8482, 0x5F7C, 0x5500, 0x1FE2, 0x35AA, 0x5FAA, 0xC4EA, 0x5FAA, 0x44AA, 0x44E2, 0x44AA, 0x4A4}},
    {28718, 2, {0x7FC, 0x4040, 0x2FFE, 0x2842, 0x358, 0x8040, 0x4358, 0x5000, 0x1110, 0x2FBE, 0xE110, 0x23B8, 0x2554, 0x2912, 0x2110, 0x110}},
    {28719, 2, {0x88, 0x22AA, 0x12DC, 0x1488, 0x8154, 0x4222, 0x47FE, 0x1402, 0x11F8, 0x2108, 0xE1F8, 0x2000, 0x23FC, 0x2204, 0x23FC, 0x204}},
    {28720, 2, {0x4000, 0x2EFE, 0x2254, 0x8292, 0x42FE, 0x4E92, 0x8D6, 0x28BA, 0x28D6, 0x4E92, 0xC2D6, 0x42BA, 0x42D6, 0x4292, 0x4A9A, 0x484}},
    {28721, 2, {0x4000, 0x2FFE, 0x2912, 0x8FFE, 0x4000, 0x4FFE, 0x880, 0x2AA2, 0x288A, 0x494A, 0xCA2A, 0x488A, 0x4AAA, 0x5082, 0x514A, 0x2224}},
    {28722, 2, {0x208, 0x8208, 0x4508, 0x890, 0x905E, 0x6FA4, 0x4014, 0x14, 0x3DD4, 0x5554, 0xD554, 0x5DC8, 0x4888, 0x4894, 0x5554, 0x2222}},
    {28723, 2, {0x110, 0x27FE, 0x1110, 0x1040, 0x87FE, 0x44A0, 0x47FC, 0x14A4, 0x17FC, 0x2500, 0xE5FC, 0x2500, 0x25FC, 0x2804, 0x3554, 0x8}},
    {28724, 2, {0x7BC, 0x24A4, 0x17BC, 0x14A4, 0x87BC, 0x44A4, 0x47BC, 0x1318, 0x14A4, 0x2080, 0xEFFC, 0x2110, 0x2320, 0x20C0, 0x21B0, 0x608}},
    {28725, 2, {0x248, 0x2150, 0x17FC, 0x1150, 0x864C, 0x43F8, 0x4248, 0x13F8, 0x1248, 0x23F8, 0xE110, 0x27FC, 0x2110, 0x2FFE, 0x2208, 0x404}},
    {28726, 2, {0x40, 0x4F40, 0x2A7E, 0x2A80, 0xA00, 0x8F7C, 0x4944, 0x4944, 0x197C, 0x2F00, 0xEAEE, 0x2AAA, 0x2AAA, 0x2AAA, 0x2FEE, 0xAA}},
    {28727, 2, {0x0, 0x27BC, 0x14A4, 0x1294, 0x84A4, 0x43F8, 0x4248, 0x13F8, 0x1248, 0x23F8, 0xE110, 0x27FC, 0x2110, 0x2FFE, 0x2208, 0x404}},
    {28728, 2, {0x448, 0x444C, 0x2AAA, 0x2008, 0xFFE, 0x8288, 0x4EE8, 0x528A, 0x1EEA, 0x228C, 0xEEEC, 0x2288, 0x22EA, 0x2F1A, 0x2426, 0x42}},
    {28729, 2, {0x4040, 0x20A0, 0x2110, 0x8208, 0x4DF6, 0x4000, 0xEEE, 0x2AAA, 0x2EEE, 0x4000, 0xCFFE, 0x4922, 0x4FFE, 0x4922, 0x492A, 0x804}},
    {28730, 2, {0x1F0, 0x2210, 0x17FC, 0x1204, 0x83FC, 0x4224, 0x4BB8, 0xA22, 0x11FE, 0x1210, 0xE7FC, 0x2244, 0x23FC, 0x20D0, 0x214A, 0x63E}},
    {28731, 2, {0x114, 0x47D2, 0x2110, 0x2FFE, 0x10, 0x87D0, 0x4552, 0x57D2, 0x1552, 0x27D4, 0xE294, 0x27C8, 0x228A, 0x2FFA, 0x22A6, 0x442}},
    {28732, 2, {0x40, 0x27FC, 0x1000, 0x13B8, 0x82A8, 0x43B8, 0x4110, 0x17FC, 0x1110, 0x27FC, 0xE110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x186}},
    {28733, 2, {0x40, 0x27FE, 0x1492, 0x13FC, 0x8090, 0x43FC, 0x4090, 0x17FE, 0x1108, 0x23FC, 0xE50A, 0x21F8, 0x2240, 0x227C, 0x2540, 0x8FE}},
    {28734, 2, {0x7BC, 0x24A4, 0x17BC, 0x14A4, 0x87BC, 0x4444, 0x45F4, 0x1444, 0x15F4, 0x2554, 0xE5F4, 0x2554, 0x25F4, 0x24E4, 0x2554, 0x44C}},
    {28735, 2, {0x210, 0x43D0, 0x241E, 0x2FE4, 0x554, 0x8FF4, 0x4548, 0x57D4, 0x11B2, 0x2220, 0xE7C0, 0x2088, 0x27FC, 0x2044, 0x2350, 0x488}},
    {28736, 2, {0x7FC, 0x4040, 0x2FFE, 0x2842, 0x358, 0x8040, 0x4358, 0x5000, 0x127C, 0x2FC4, 0xE27C, 0x2744, 0x2AFC, 0x3244, 0x227C, 0x244}},
    {28737, 2, {0x4FBE, 0x28A2, 0x2FBE, 0x88A2, 0x4FBE, 0x4842, 0x9F2, 0x2912, 0x29F2, 0x4912, 0xC9F2, 0x4846, 0x4F6A, 0x4952, 0x4A4A, 0xCC6}},
    {28738, 2, {0x3C, 0x27C0, 0x1244, 0x1128, 0x87FC, 0x44A4, 0x47FC, 0x1008, 0x1788, 0x24FE, 0xE788, 0x24A8, 0x2798, 0x2508, 0x25A8, 0x690}},
    {28739, 2, {0x150, 0x23F8, 0x1554, 0x17FC, 0x8554, 0x47FC, 0x4554, 0x17FC, 0x1000, 0x27FC, 0xE000, 0x23F8, 0x2208, 0x23F8, 0x2110, 0xFFE}},
    {28740, 2, {0x3FC, 0x2108, 0x11F8, 0x1108, 0x81F8, 0x410E, 0x47F8, 0x1008, 0x17FE, 0x2294, 0xE39C, 0x2294, 0x239C, 0x22D6, 0x27BC, 0x84}},
    {28741, 2, {0x7FC, 0x4444, 0x27FC, 0x2444, 0x7FC, 0x8000, 0x4FBE, 0x5AAA, 0x1FBE, 0x2AAA, 0xEFBE, 0x2040, 0x2040, 0x27FC, 0x2040, 0xFFE}},
    {28742, 2, {0x90, 0x27FE, 0x1090, 0x13E8, 0x8288, 0x43CE, 0x4A50, 0xBD4, 0x12A2, 0x13E2, 0xE000, 0x23FC, 0x2294, 0x2294, 0x27FE, 0x0}},
    {28743, 2, {0x492, 0x4294, 0x2FFE, 0x2110, 0x7FC, 0x8040, 0x43F8, 0x5040, 0x1FFE, 0x2480, 0xE7BC, 0x2494, 0x2794, 0x24C8, 0x2F94, 0xA2}},
    {28744, 2, {0x7BC, 0x24A4, 0x17BC, 0x14A4, 0x87BC, 0x44A4, 0x47BC, 0x1120, 0x13FE, 0x2220, 0xE7FC, 0x2A20, 0x23FC, 0x2220, 0x23FE, 0x200}},
    {28745, 2, {0xA98, 0x9514, 0x5490, 0xA3E, 0x9F28, 0x5168, 0x5F3E, 0x28, 0x3F28, 0x553E, 0xD528, 0x5F28, 0x50A8, 0x50BE, 0x4FA0, 0x20}},
    {28746, 2, {0xFBE, 0x4528, 0x2FBE, 0x2308, 0x59A, 0x8926, 0x47FC, 0x5000, 0x13F8, 0x2208, 0xE3F8, 0x2000, 0x27FC, 0x2494, 0x25F4, 0x44C}},
    {28747, 2, {0x40, 0x4FFE, 0x2920, 0x2FFC, 0x924, 0x8FFC, 0x4900, 0x4BFE, 0x1802, 0x2AAA, 0xE844, 0x2BF8, 0x2840, 0x2FFC, 0x2910, 0x13F8}},
    {28748, 2, {0x110, 0x4FFE, 0x2110, 0x27BC, 0x4A4, 0x87BC, 0x4140, 0x5120, 0x13FE, 0x2220, 0xE7FC, 0x2A20, 0x23FC, 0x2220, 0x23FE, 0x200}},
    {28749, 2, {0xFBE, 0x48A2, 0x2FBE, 0x28A2, 0xFBE, 0x8C42, 0x4A92, 0x4FDE, 0x1926, 0x2D5A, 0xEFD2, 0x2912, 0x292A, 0x2A46, 0x2C0A, 0x804}},
    {28750, 2, {0x508, 0x9FC8, 0x555C, 0x1FC8, 0x9548, 0x5FFE, 0x4008, 0x1FD0, 0x2014, 0x4FBE, 0xC880, 0x4FBE, 0x48AA, 0x452A, 0x43AA, 0x1C7E}},
    {28751, 2, {0x4000, 0x2FBE, 0x2888, 0x8F90, 0x48BE, 0x4FA2, 0x222, 0x3FEA, 0x202A, 0x4FAA, 0xC8AA, 0x4FAA, 0x4208, 0x4A94, 0x5252, 0x622}},
    {28752, 2, {0x110, 0x4554, 0x25B8, 0x2910, 0x2A8, 0x8444, 0x4FFE, 0x4842, 0x10A0, 0x2318, 0xEDF6, 0x2040, 0x27FC, 0x2248, 0x2150, 0xFFE}},
    {28753, 2, {0xFBE, 0x4000, 0x27BC, 0x24A4, 0x6B4, 0x84A4, 0x4040, 0x57FE, 0x14A0, 0x27FC, 0xE4A4, 0x27FC, 0x2512, 0x29DC, 0x2912, 0x11CE}},
    {28754, 2, {0x528, 0x47BE, 0x2948, 0x27BE, 0x318, 0x85AA, 0x4946, 0x53FC, 0x1204, 0x23FC, 0xE204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x204}},
    {28755, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x108, 0x7D90, 0x960, 0x1120, 0x2518, 0xC206}},
    {28756, 2, {0x510, 0x9FD0, 0x555C, 0x1FE4, 0x9548, 0x5FFE, 0x402A, 0x1FEA, 0x202A, 0x4FBE, 0xC8A0, 0x4FA0, 0x48A2, 0x4522, 0x43A2, 0x1C1E}},
    {28757, 2, {0x40C, 0x820A, 0x5FC8, 0x51E, 0x9254, 0x5574, 0x505E, 0x1FD4, 0x2214, 0x5FDE, 0xD454, 0x5554, 0x5754, 0x505E, 0x5150, 0x1090}},
    {28758, 2, {0x20, 0x27FE, 0x1488, 0x17DE, 0x8488, 0x45DC, 0x46AA, 0x1488, 0x1450, 0x27DE, 0xE450, 0x25DC, 0x2450, 0x2BDE, 0x2850, 0x1050}},
    {28759, 2, {0x810, 0x88FE, 0x4800, 0x3EFE, 0x2A82, 0x2ABA, 0xAAAA, 0x6AFE, 0x3E00, 0x287C, 0x844, 0x4A7C, 0x8E44, 0x3A7C, 0x1000, 0xFE}},
    {28760, 2, {0x4518, 0x2514, 0x2F90, 0x853E, 0x4728, 0x4268, 0xFBE, 0x2AA8, 0x2FA8, 0x423E, 0xCFA8, 0x4228, 0x5FE8, 0x453E, 0x48A0, 0x1060}},
    {28761, 2, {0x248, 0x2150, 0x17FC, 0x1404, 0x81F0, 0x4110, 0x47FC, 0x1554, 0x14E4, 0x27FC, 0xE040, 0x27FC, 0x2040, 0x2FFE, 0x22A4, 0x452}},
    {28762, 2, {0x2AC, 0x4444, 0x26AC, 0x2444, 0x6AC, 0x8404, 0x4FFE, 0x4A0A, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x20A0, 0x2124, 0x61C}},
    {28763, 2, {0x4F1E, 0x2912, 0x2F1E, 0x8912, 0x4F1E, 0x4BBA, 0xAAA, 0x2BBA, 0x29F2, 0x4952, 0xC9F2, 0x4952, 0x49F2, 0x4842, 0x4BFA, 0x846}},
    {28764, 2, {0x80, 0xBFFC, 0x4800, 0xFF8, 0x8000, 0x4FF8, 0x4808, 0xFF8, 0x2000, 0x5DDC, 0xD554, 0x5DD4, 0x555C, 0x5DD6, 0x5496, 0x2D62}},
    {28765, 2, {0x4000, 0x2FBE, 0x2888, 0x8F90, 0x48BE, 0x4FA2, 0x222, 0x3FFE, 0x2022, 0x4FBE, 0xC8A2, 0x4FA2, 0x423E, 0x4A80, 0x5254, 0x622}},
    {28766, 2, {0x47FC, 0x2040, 0x2FFE, 0x8842, 0x4358, 0x4500, 0xFBC, 0x2524, 0x2724, 0x423C, 0xCFA4, 0x4AA4, 0x4FBC, 0x4224, 0x5FA4, 0x24C}},
    {28767, 2, {0x7FC, 0x2404, 0x17FC, 0x1524, 0x84A8, 0x4524, 0x45FC, 0x1554, 0x15FC, 0x2644, 0xE5F4, 0x2554, 0x29F4, 0x2844, 0x33F4, 0x8}},
    {28768, 2, {0x790, 0x251E, 0x17A8, 0x1484, 0x87BE, 0x452A, 0x47BE, 0x1000, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x20A0, 0x2124, 0x61C}},
    {28769, 2, {0x110, 0x27FE, 0x1110, 0x17BC, 0x84A4, 0x47BC, 0x44A4, 0x17BC, 0x1444, 0x27FC, 0xE554, 0x25F4, 0x2444, 0x24E4, 0x2554, 0x448}},
    {28770, 2, {0x40, 0x27FC, 0x1248, 0x13F8, 0x8040, 0x4FFE, 0x4AAA, 0x13B8, 0x1110, 0x27FC, 0xE110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x186}},
    {28771, 2, {0x4444, 0x28E8, 0x2E0E, 0x84E4, 0x4A0A, 0x4EEE, 0xA0, 0x2AEA, 0x2000, 0x47F8, 0xC008, 0x47F8, 0x4400, 0x47FC, 0x4004, 0x38}},
    {28772, 2, {0x4444, 0x24E4, 0x2A0A, 0x8EEE, 0x4404, 0x4AEA, 0xE0E, 0x20E0, 0x2AAA, 0x4AEA, 0xC040, 0x4FFE, 0x4150, 0x4248, 0x4C46, 0x40}},
    {28773, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x920, 0x1310, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x9292, 0x5454, 0xB2B2}},
    {28774, 2, {0x0, 0x9FDE, 0x5044, 0x1FC8, 0x905E, 0x5FD2, 0x4892, 0x155E, 0x3DD2, 0x489E, 0xD552, 0x5DD2, 0x401E, 0x5540, 0x554C, 0x12}},
    {28775, 2, {0x4210, 0x2A90, 0x2F9C, 0x8AA4, 0x4FC8, 0x4ABE, 0xFAA, 0x202A, 0x3FEA, 0x403E, 0xCFA0, 0x48A0, 0x4FA2, 0x48A2, 0x4522, 0x1F9E}},
    {28776, 2, {0x410, 0x9F9C, 0x4934, 0x608, 0x9FB6, 0x401C, 0x4F88, 0x8BE, 0x2F80, 0x48BE, 0xCFA2, 0x423E, 0x5FA2, 0x423E, 0x4214, 0x222}},
    {28777, 2, {0x4208, 0x2A88, 0x2F9C, 0x8A88, 0x4F88, 0x4ABE, 0xF88, 0x2010, 0x3FD4, 0x403E, 0xCF80, 0x48BE, 0x4FAA, 0x48AA, 0x452A, 0x1FFE}},
    {28778, 2, {0x4288, 0x22E8, 0x2F5E, 0x82EC, 0x475A, 0x4AE8, 0x208, 0x2FFE, 0x2942, 0x5494, 0xC558, 0x47F2, 0x4204, 0x43E8, 0x4202, 0x3E4}},
    {28779, 2, {0x100, 0x100, 0x100, 0x1108, 0x1108, 0x1110, 0x2120, 0x2100, 0x4280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {28780, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {28781, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x1108, 0x1108, 0x1110, 0x2120, 0x4280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {28782, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xE00E, 0x440, 0x440, 0x440, 0x440, 0x840, 0x1042, 0x2042, 0xC03E, 0x0}},
    {28783, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x5420, 0x5820, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x2820, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {28784, 2, {0x400, 0x400, 0x400, 0xFFFE, 0x480, 0x880, 0x888, 0xA90, 0x12A0, 0x1480, 0x2140, 0x4140, 0x8220, 0x210, 0x40E, 0x1804}},
    {28785, 2, {0x1000, 0x1000, 0x11FC, 0x1044, 0x5444, 0x5844, 0x5044, 0x9044, 0x1044, 0x1044, 0x1044, 0x2884, 0x2484, 0x4504, 0x4228, 0x8410}},
    {28786, 2, {0x804, 0x804, 0x804, 0x944, 0x4944, 0x4A44, 0x4C44, 0x8844, 0x844, 0x844, 0x1444, 0x1244, 0x2104, 0x2104, 0x4014, 0x8008}},
    {28787, 2, {0x800, 0x800, 0x800, 0x1FFC, 0x1004, 0x2204, 0x4244, 0x9244, 0x1284, 0x2204, 0x504, 0x484, 0x844, 0x1004, 0x6028, 0x10}},
    {28788, 2, {0x1000, 0x1000, 0x11FC, 0x1020, 0x5420, 0x5820, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {28789, 2, {0x0, 0x3FF8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x3FF8, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28790, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x5420, 0x5820, 0x51FC, 0x9020, 0x1020, 0x1020, 0x1020, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {28791, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xE00E, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {28792, 2, {0x400, 0x400, 0xFE0, 0x1040, 0x2080, 0x360, 0xC18, 0x7004, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28793, 2, {0x1010, 0x1078, 0x13C0, 0x1040, 0x5440, 0x5840, 0x5040, 0x907E, 0x17C0, 0x1040, 0x1040, 0x2842, 0x2442, 0x4442, 0x403E, 0x8000}},
    {28794, 2, {0x1020, 0x1020, 0x1020, 0x1120, 0x552C, 0x5934, 0x5164, 0x93A4, 0x1124, 0x1134, 0x1128, 0x2922, 0x2522, 0x4502, 0x40FE, 0x8000}},
    {28795, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28796, 2, {0x1040, 0x1040, 0x1080, 0x10FC, 0x5504, 0x5A04, 0x5004, 0x9104, 0x1084, 0x1044, 0x1044, 0x2804, 0x2404, 0x4404, 0x4028, 0x8010}},
    {28797, 2, {0x1110, 0x1110, 0x2220, 0x4440, 0x2220, 0x1110, 0x1110, 0x0, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28798, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x100, 0x1108, 0x1108, 0x1110, 0x2120, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {28799, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x5524, 0x5924, 0x5124, 0x9124, 0x1124, 0x1124, 0x1124, 0x2924, 0x2524, 0x45FC, 0x4004, 0x8000}},
    {28800, 2, {0x1000, 0x11F8, 0x1010, 0x1420, 0x5840, 0x5080, 0x51FE, 0x9092, 0x1092, 0x1092, 0x1112, 0x2922, 0x2622, 0x4442, 0x4094, 0x8108}},
    {28801, 2, {0x0, 0x7FF8, 0x200, 0x4200, 0x4200, 0x7FFC, 0x480, 0x880, 0x1084, 0x2084, 0xC07C, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {28802, 2, {0x1010, 0x1090, 0x1090, 0x1488, 0x5888, 0x5124, 0x5124, 0x9222, 0x1440, 0x1040, 0x1088, 0x2884, 0x2504, 0x43FE, 0x4102, 0x8000}},
    {28803, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {28804, 2, {0x1000, 0x13F8, 0x1088, 0x1488, 0x5888, 0x5088, 0x5088, 0x97F8, 0x1108, 0x1108, 0x1108, 0x2908, 0x2508, 0x4108, 0x47FE, 0x8000}},
    {28805, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x1108, 0x1110, 0x2280, 0x440, 0x820, 0x3018, 0xC006}},
    {28806, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x5488, 0x5888, 0x5088, 0x9088, 0x1088, 0x1050, 0x1050, 0x2820, 0x2450, 0x4488, 0x4104, 0x8602}},
    {28807, 2, {0x1040, 0x1040, 0x1040, 0x1480, 0x58FE, 0x5108, 0x5288, 0x9088, 0x1088, 0x1050, 0x1050, 0x2820, 0x2450, 0x4488, 0x4104, 0x8202}},
    {28808, 2, {0x1000, 0x10F8, 0x1088, 0x1488, 0x5888, 0x5106, 0x5200, 0x91FC, 0x1084, 0x1084, 0x1048, 0x2850, 0x2420, 0x4450, 0x4088, 0x8306}},
    {28809, 2, {0x1080, 0x1040, 0x1040, 0x15FC, 0x5904, 0x5104, 0x5104, 0x91FC, 0x1104, 0x1100, 0x1100, 0x2900, 0x2600, 0x4200, 0x4400, 0x8800}},
    {28810, 2, {0x1040, 0x1040, 0x1040, 0x127C, 0x5284, 0x5488, 0x5920, 0x9020, 0x1020, 0x1050, 0x2850, 0x2850, 0x2488, 0x4488, 0x4104, 0x8202}},
    {28811, 2, {0x1000, 0x1000, 0x11FE, 0x1010, 0x5410, 0x5820, 0x5020, 0x9068, 0x10A4, 0x1122, 0x1222, 0x2820, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28812, 2, {0x1020, 0x1020, 0x1050, 0x1450, 0x5888, 0x5104, 0x5202, 0x9088, 0x1088, 0x1088, 0x1088, 0x2888, 0x2488, 0x4508, 0x4108, 0x8208}},
    {28813, 2, {0x1008, 0x101C, 0x11E0, 0x1500, 0x5900, 0x51FC, 0x5144, 0x9144, 0x1144, 0x1128, 0x1128, 0x2910, 0x2510, 0x4228, 0x4244, 0x8482}},
    {28814, 2, {0x100, 0x1108, 0x1110, 0x2120, 0x2C0, 0x430, 0x1808, 0x6004, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28815, 2, {0x1040, 0x1040, 0x1040, 0x1440, 0x5944, 0x5144, 0x5148, 0x9250, 0x1040, 0x10A0, 0x10A0, 0x2890, 0x2510, 0x4108, 0x4204, 0x8402}},
    {28816, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5420, 0x5820, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28817, 2, {0x1020, 0x1020, 0x1020, 0x1420, 0x5BFE, 0x5020, 0x5070, 0x9070, 0x10A8, 0x10A8, 0x1124, 0x2A22, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28818, 2, {0x1020, 0x1020, 0x1020, 0x10A8, 0x54A4, 0x58A2, 0x5122, 0x9120, 0x1224, 0x1024, 0x1028, 0x2808, 0x2410, 0x4420, 0x40C0, 0x8300}},
    {28819, 2, {0x1010, 0x1110, 0x1090, 0x1090, 0x5410, 0x5910, 0x5090, 0x9090, 0x1010, 0x101E, 0x13F0, 0x2810, 0x2410, 0x4410, 0x4010, 0x8010}},
    {28820, 2, {0x1040, 0x1040, 0x1040, 0x17F8, 0x5848, 0x5048, 0x5048, 0x9048, 0x17FE, 0x1040, 0x10A0, 0x28A0, 0x2510, 0x4110, 0x4208, 0x8406}},
    {28821, 2, {0x1080, 0x1040, 0x1040, 0x13FC, 0x5400, 0x5800, 0x51F0, 0x9110, 0x1110, 0x1110, 0x1110, 0x2912, 0x2512, 0x4612, 0x420E, 0x8400}},
    {28822, 2, {0x1020, 0x1020, 0x1020, 0x11FE, 0x5420, 0x5924, 0x5124, 0x9124, 0x1124, 0x11FC, 0x1024, 0x2820, 0x2422, 0x4422, 0x401E, 0x8000}},
    {28823, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x820, 0xFE0, 0x820, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28824, 2, {0x1008, 0x101C, 0x11E0, 0x1500, 0x5900, 0x5100, 0x51FE, 0x9110, 0x1110, 0x1110, 0x1110, 0x2910, 0x2510, 0x4210, 0x4210, 0x8410}},
    {28825, 2, {0x200, 0x400, 0xFF8, 0x1210, 0x6920, 0x4C0, 0x300, 0xC00, 0x7100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28826, 2, {0x20, 0x20, 0x7C20, 0x4424, 0x44A4, 0x44A4, 0x44A8, 0x7D20, 0x4420, 0x4450, 0x4450, 0x4450, 0x7C88, 0x4488, 0x104, 0x202}},
    {28827, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xE00E, 0x880, 0x888, 0x1090, 0x30E0, 0x5180, 0x9684, 0x1084, 0x107C, 0x1000}},
    {28828, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0x5440, 0x5840, 0x51FC, 0x9040, 0x1040, 0x13FE, 0x1042, 0x2842, 0x244A, 0x4444, 0x4040, 0x8040}},
    {28829, 2, {0x1040, 0x1040, 0x10A0, 0x14A0, 0x5910, 0x5208, 0x5406, 0x91F0, 0x1110, 0x1110, 0x1150, 0x2920, 0x2504, 0x4504, 0x40FC, 0x8000}},
    {28830, 2, {0x1040, 0x1020, 0x1020, 0x1400, 0x5BFE, 0x5040, 0x5040, 0x9060, 0x1050, 0x1048, 0x1044, 0x2844, 0x2440, 0x4440, 0x4040, 0x8040}},
    {28831, 2, {0x1000, 0x1000, 0x11FC, 0x1504, 0x5904, 0x5104, 0x51FC, 0x9104, 0x1104, 0x1104, 0x11FC, 0x2904, 0x2400, 0x4000, 0x43FE, 0x8000}},
    {28832, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x5524, 0x59FC, 0x5124, 0x9124, 0x1124, 0x11FC, 0x1124, 0x2820, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28833, 2, {0x1000, 0x1000, 0x13FE, 0x1420, 0x5820, 0x5020, 0x5120, 0x9120, 0x113C, 0x1120, 0x1120, 0x2920, 0x2520, 0x4120, 0x47FE, 0x8000}},
    {28834, 2, {0x1020, 0x1028, 0x1024, 0x1020, 0x55FE, 0x5820, 0x5070, 0x9070, 0x10A8, 0x10A8, 0x1124, 0x2A22, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28835, 2, {0x1000, 0x1000, 0x13FE, 0x1008, 0x5408, 0x59E8, 0x5128, 0x9128, 0x1128, 0x1128, 0x11E8, 0x2928, 0x2408, 0x4408, 0x4028, 0x8010}},
    {28836, 2, {0x1000, 0x13FC, 0x1084, 0x1084, 0x5484, 0x5904, 0x5114, 0x9208, 0x1400, 0x11FC, 0x1104, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28837, 2, {0x1090, 0x1090, 0x1090, 0x13FC, 0x5494, 0x5894, 0x53FC, 0x9290, 0x1290, 0x13FE, 0x1092, 0x2892, 0x251A, 0x4514, 0x4210, 0x8410}},
    {28838, 2, {0x1050, 0x1048, 0x1048, 0x1440, 0x5BFE, 0x5080, 0x5080, 0x90FC, 0x1144, 0x1144, 0x1128, 0x2928, 0x2610, 0x4228, 0x4444, 0x8182}},
    {28839, 2, {0x1080, 0x1080, 0x10FE, 0x1500, 0x5A20, 0x5120, 0x512C, 0x9174, 0x13A4, 0x1124, 0x1134, 0x2928, 0x2522, 0x4502, 0x40FE, 0x8000}},
    {28840, 2, {0x1020, 0x1010, 0x1010, 0x15FE, 0x5902, 0x5204, 0x5080, 0x9088, 0x1090, 0x10A0, 0x10C0, 0x2882, 0x2482, 0x4482, 0x407E, 0x8000}},
    {28841, 2, {0x1040, 0x1040, 0x10A0, 0x14A0, 0x5910, 0x5248, 0x5426, 0x9020, 0x13F8, 0x1008, 0x1010, 0x2910, 0x24A0, 0x4440, 0x4020, 0x8020}},
    {28842, 2, {0x1020, 0x1020, 0x1124, 0x1524, 0x5924, 0x5124, 0x51FC, 0x9020, 0x1020, 0x1124, 0x1124, 0x2924, 0x2524, 0x4524, 0x41FC, 0x8004}},
    {28843, 2, {0x1020, 0x1010, 0x1010, 0x15FE, 0x5820, 0x5020, 0x5044, 0x9084, 0x11F8, 0x1010, 0x1020, 0x2844, 0x2482, 0x45FE, 0x4082, 0x8000}},
    {28844, 2, {0x1000, 0x11FE, 0x1100, 0x1100, 0x5500, 0x59FC, 0x5104, 0x9104, 0x1104, 0x1104, 0x11FC, 0x2900, 0x2500, 0x4500, 0x41FE, 0x8000}},
    {28845, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x800, 0x800, 0xFFFE, 0x1080, 0x1088, 0x1488, 0x2490, 0x2940, 0x4140, 0x8220, 0xC18, 0x3006}},
    {28846, 2, {0x1080, 0x1080, 0x11FC, 0x1504, 0x5A04, 0x55F4, 0x5114, 0x9114, 0x1114, 0x11F4, 0x1104, 0x2928, 0x2512, 0x4502, 0x40FE, 0x8000}},
    {28847, 2, {0x1000, 0x11FE, 0x1102, 0x1502, 0x5902, 0x517A, 0x514A, 0x914A, 0x114A, 0x114A, 0x117A, 0x294A, 0x2502, 0x4502, 0x410A, 0x8104}},
    {28848, 2, {0x800, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1090, 0x1F90, 0x1050, 0x1024, 0x1004, 0xFFC, 0x0, 0x4888, 0x4444, 0x8444}},
    {28849, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {28850, 2, {0x1020, 0x1020, 0x1040, 0x1088, 0x5504, 0x5BFE, 0x5002, 0x9000, 0x11FC, 0x1104, 0x1104, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28851, 2, {0x1000, 0x13FE, 0x1020, 0x1420, 0x5820, 0x53FE, 0x5222, 0x9222, 0x1252, 0x124A, 0x128A, 0x2B02, 0x2602, 0x4202, 0x420A, 0x8204}},
    {28852, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5524, 0x5924, 0x5124, 0x9124, 0x1124, 0x13FE, 0x1020, 0x2850, 0x2450, 0x4488, 0x4104, 0x8202}},
    {28853, 2, {0x1040, 0x1040, 0x107C, 0x1084, 0x5588, 0x5A50, 0x5020, 0x9050, 0x1088, 0x1306, 0x1060, 0x2810, 0x2408, 0x44C0, 0x4020, 0x8010}},
    {28854, 2, {0x1020, 0x1020, 0x1020, 0x1020, 0x543E, 0x5820, 0x5020, 0x9020, 0x11FC, 0x1104, 0x1104, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28855, 2, {0x1040, 0x1020, 0x1000, 0x15FE, 0x5820, 0x5020, 0x5020, 0x9020, 0x11FC, 0x1020, 0x1020, 0x2820, 0x2420, 0x4020, 0x43FE, 0x8000}},
    {28856, 2, {0x1100, 0x1100, 0x1100, 0x15FE, 0x5A80, 0x5280, 0x5480, 0x90F8, 0x1080, 0x1080, 0x1080, 0x28FC, 0x2480, 0x4480, 0x4080, 0x8080}},
    {28857, 2, {0x200, 0x200, 0x200, 0x3FC, 0x200, 0x200, 0x3FF0, 0x2010, 0x2010, 0x2010, 0x3FF0, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {28858, 2, {0x200, 0x1200, 0xA00, 0x7FF0, 0x420, 0x440, 0xFF8, 0x808, 0x1010, 0x1FFC, 0x2004, 0x2AA4, 0x4AA4, 0x9004, 0x28, 0x10}},
    {28859, 2, {0x1000, 0x1000, 0x13FE, 0x1020, 0x5420, 0x5840, 0x5040, 0x90FC, 0x1184, 0x1284, 0x1484, 0x2884, 0x2484, 0x4484, 0x40FC, 0x8084}},
    {28860, 2, {0x1040, 0x1040, 0x17FC, 0x1080, 0x5480, 0x5BE0, 0x5120, 0x9220, 0x13FC, 0x1020, 0x1128, 0x2924, 0x2622, 0x4422, 0x40A0, 0x8040}},
    {28861, 2, {0x1000, 0x1000, 0x11FC, 0x1104, 0x5504, 0x5904, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1000, 0x2890, 0x2488, 0x4504, 0x4202, 0x8402}},
    {28862, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x5504, 0x5904, 0x51FC, 0x9050, 0x1050, 0x1050, 0x1050, 0x2892, 0x2492, 0x4512, 0x420E, 0x8400}},
    {28863, 2, {0x1020, 0x1040, 0x11F8, 0x1508, 0x5908, 0x5148, 0x5128, 0x9128, 0x1108, 0x17FE, 0x1108, 0x2908, 0x2508, 0x4108, 0x4228, 0x8410}},
    {28864, 2, {0x1008, 0x103C, 0x13E0, 0x1020, 0x5524, 0x58A4, 0x50A8, 0x9020, 0x13FE, 0x1020, 0x1020, 0x2820, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {28865, 2, {0x1008, 0x101C, 0x11E0, 0x1100, 0x5520, 0x5920, 0x5120, 0x91FE, 0x1020, 0x1020, 0x10A8, 0x28A4, 0x2522, 0x4622, 0x40A0, 0x8040}},
    {28866, 2, {0x1000, 0x1108, 0x1088, 0x1490, 0x5800, 0x53FE, 0x5000, 0x9000, 0x1000, 0x11FC, 0x1000, 0x2800, 0x2400, 0x4000, 0x43FE, 0x8000}},
    {28867, 2, {0x1000, 0x11FC, 0x1008, 0x1010, 0x5430, 0x5848, 0x5084, 0x9302, 0x1000, 0x11FC, 0x1020, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {28868, 2, {0x1040, 0x1020, 0x1020, 0x13FE, 0x5400, 0x5888, 0x5104, 0x9202, 0x1088, 0x1088, 0x1050, 0x2850, 0x2420, 0x4450, 0x4088, 0x8306}},
    {28869, 2, {0x1020, 0x1020, 0x1040, 0x1480, 0x59FC, 0x5154, 0x5154, 0x9154, 0x1154, 0x1154, 0x1154, 0x2954, 0x2554, 0x4154, 0x43FE, 0x8000}},
    {28870, 2, {0x1080, 0x1080, 0x113C, 0x1200, 0x5480, 0x5880, 0x517E, 0x9308, 0x1508, 0x1108, 0x1108, 0x2908, 0x2508, 0x4508, 0x4128, 0x8110}},
    {28871, 2, {0x1020, 0x1020, 0x1050, 0x1488, 0x5904, 0x5202, 0x51FC, 0x9020, 0x1020, 0x1020, 0x11FC, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {28872, 2, {0x4, 0x7F84, 0x824, 0x824, 0x1F24, 0x2124, 0x5124, 0xA24, 0x404, 0x814, 0x3008, 0xC000, 0x2488, 0x2244, 0x4244, 0x8004}},
    {28873, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2008, 0x2288, 0x2448, 0x2828, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {28874, 2, {0x1104, 0x1084, 0x1088, 0x1400, 0x5BFE, 0x5020, 0x5020, 0x91FC, 0x1020, 0x1020, 0x13FE, 0x2820, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28875, 2, {0x840, 0x840, 0x1040, 0x37FC, 0x50E0, 0x9150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {28876, 2, {0x1090, 0x1090, 0x1090, 0x1510, 0x59FE, 0x5310, 0x5510, 0x9138, 0x1138, 0x1154, 0x1154, 0x2992, 0x2510, 0x4510, 0x4110, 0x8110}},
    {28877, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0x5520, 0x5A20, 0x5020, 0x93FE, 0x1090, 0x1090, 0x1090, 0x2890, 0x2512, 0x4512, 0x4212, 0x840E}},
    {28878, 2, {0x0, 0x3FF8, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4120, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28879, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1000, 0x1FFE, 0x1000, 0x1FFC, 0x4, 0x924, 0x2494, 0x2494, 0x4028, 0x10}},
    {28880, 2, {0x1020, 0x1040, 0x11F8, 0x1508, 0x5948, 0x5128, 0x5128, 0x93FE, 0x1108, 0x1148, 0x1128, 0x2928, 0x2508, 0x4208, 0x4228, 0x8410}},
    {28881, 2, {0x1090, 0x1090, 0x1090, 0x1492, 0x5A94, 0x5098, 0x5090, 0x9198, 0x1294, 0x1492, 0x1090, 0x2890, 0x2512, 0x4112, 0x4212, 0x840E}},
    {28882, 2, {0x1028, 0x1024, 0x1024, 0x1420, 0x5BFE, 0x5020, 0x5020, 0x93A0, 0x1120, 0x1110, 0x1110, 0x2910, 0x25CA, 0x470A, 0x4206, 0x8002}},
    {28883, 2, {0x1020, 0x1020, 0x11FC, 0x1420, 0x5820, 0x5020, 0x53FE, 0x9000, 0x1020, 0x1020, 0x11FC, 0x2820, 0x2420, 0x4020, 0x43FE, 0x8000}},
    {28884, 2, {0x1000, 0x11FE, 0x1102, 0x1102, 0x557A, 0x5902, 0x5102, 0x917A, 0x114A, 0x114A, 0x114A, 0x297A, 0x2502, 0x4502, 0x410A, 0x8104}},
    {28885, 2, {0x28, 0x24, 0x20, 0x3FFE, 0x2020, 0x2020, 0x2FA4, 0x2224, 0x2264, 0x2AA8, 0x3228, 0x2510, 0x2492, 0x48AA, 0x5046, 0x8082}},
    {28886, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x40, 0x844, 0x2944, 0x2944, 0x2A28, 0x482A, 0x1412, 0x122A, 0x2246, 0x4082}},
    {28887, 2, {0x1020, 0x1010, 0x1010, 0x13FE, 0x5420, 0x5842, 0x5084, 0x91F8, 0x1010, 0x1022, 0x10C4, 0x2B08, 0x2410, 0x4428, 0x40C4, 0x8302}},
    {28888, 2, {0x1088, 0x1088, 0x1088, 0x1088, 0x55FE, 0x5888, 0x5088, 0x9088, 0x1088, 0x13FE, 0x1000, 0x2888, 0x2484, 0x4504, 0x4202, 0x8402}},
    {28889, 2, {0x1080, 0x1080, 0x10F8, 0x1508, 0x5B10, 0x54A0, 0x5040, 0x90A0, 0x1118, 0x1206, 0x15F8, 0x2908, 0x2508, 0x4508, 0x41F8, 0x8108}},
    {28890, 2, {0x1020, 0x1020, 0x1050, 0x1088, 0x5504, 0x5A02, 0x51FC, 0x9000, 0x1000, 0x11FC, 0x1104, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28891, 2, {0x1020, 0x1020, 0x1020, 0x11FC, 0x5524, 0x5924, 0x5124, 0x9124, 0x11FC, 0x1124, 0x1020, 0x2828, 0x2424, 0x45FE, 0x4082, 0x8000}},
    {28892, 2, {0x1000, 0x13FE, 0x1000, 0x1400, 0x59FC, 0x5104, 0x5104, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x2904, 0x2400, 0x4000, 0x43FE, 0x8000}},
    {28893, 2, {0x0, 0x1FE0, 0x40, 0x80, 0x7D04, 0x588, 0x950, 0x1120, 0x2118, 0x4506, 0x8200, 0x1FF0, 0x0, 0x4888, 0x4444, 0x8444}},
    {28894, 2, {0x1088, 0x1088, 0x1088, 0x1488, 0x5BE8, 0x5088, 0x518C, 0x91CA, 0x12AA, 0x1288, 0x1488, 0x2888, 0x2488, 0x4488, 0x4088, 0x8088}},
    {28895, 2, {0x1000, 0x13FE, 0x1202, 0x1622, 0x5A22, 0x5222, 0x53FE, 0x9222, 0x1222, 0x1252, 0x124A, 0x2A8A, 0x2702, 0x4202, 0x43FE, 0x8202}},
    {28896, 2, {0x1040, 0x1040, 0x17FE, 0x1080, 0x5500, 0x59FC, 0x5304, 0x9504, 0x19FC, 0x1104, 0x1104, 0x29FC, 0x2504, 0x4504, 0x4114, 0x8108}},
    {28897, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xE44E, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {28898, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x5400, 0x5838, 0x53C0, 0x9040, 0x1040, 0x107C, 0x13C0, 0x2840, 0x2442, 0x4442, 0x403E, 0x8000}},
    {28899, 2, {0x1040, 0x1040, 0x1040, 0x13FE, 0x5480, 0x5890, 0x5090, 0x9112, 0x1152, 0x1154, 0x1290, 0x2A28, 0x2428, 0x4444, 0x4084, 0x8102}},
    {28900, 2, {0x1040, 0x1044, 0x11F4, 0x1048, 0x5450, 0x5BFE, 0x5040, 0x9080, 0x11FE, 0x1240, 0x1480, 0x28FC, 0x2404, 0x4404, 0x4028, 0x8010}},
    {28901, 2, {0x1000, 0x11FE, 0x1120, 0x1120, 0x5520, 0x59FC, 0x5104, 0x9104, 0x1104, 0x1104, 0x11FC, 0x2920, 0x2520, 0x4520, 0x41FE, 0x8000}},
    {28902, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x55FC, 0x5904, 0x5104, 0x9124, 0x1124, 0x1124, 0x1124, 0x2944, 0x2450, 0x4488, 0x4104, 0x8204}},
    {28903, 2, {0x2080, 0x2080, 0x20BC, 0x23C0, 0xA850, 0xB024, 0xA0D4, 0xA30C, 0x2000, 0x27FE, 0x2090, 0x2090, 0x5112, 0x4912, 0x420E, 0x8400}},
    {28904, 2, {0x1090, 0x1094, 0x1114, 0x1518, 0x5B10, 0x5532, 0x5152, 0x910E, 0x1120, 0x1020, 0x13FE, 0x2820, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28905, 2, {0x1020, 0x1020, 0x1050, 0x1488, 0x5904, 0x5202, 0x50F8, 0x9000, 0x1000, 0x13FE, 0x1020, 0x2840, 0x2488, 0x4104, 0x43FE, 0x8102}},
    {28906, 2, {0x1088, 0x1048, 0x1050, 0x1000, 0x55FC, 0x5820, 0x5020, 0x9020, 0x13FE, 0x1020, 0x1050, 0x2850, 0x2488, 0x4488, 0x4104, 0x8202}},
    {28907, 2, {0x23F0, 0x1020, 0x80C0, 0x43FC, 0x1124, 0x2124, 0xE244, 0x2484, 0x2928, 0x2010, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {28908, 2, {0x1000, 0x11FC, 0x1104, 0x1504, 0x59FC, 0x5110, 0x5110, 0x9108, 0x1104, 0x1262, 0x1410, 0x2800, 0x24E0, 0x4418, 0x4004, 0x8000}},
    {28909, 2, {0x1040, 0x1040, 0x1040, 0xFDF8, 0x1048, 0x1048, 0x1CC8, 0x3048, 0xD0AA, 0x10AA, 0x5106, 0x2202, 0x0, 0x4888, 0x4444, 0x8444}},
    {28910, 2, {0x7F84, 0x824, 0x1024, 0x3F24, 0x5124, 0xA24, 0x404, 0x804, 0x3014, 0xC108, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {28911, 2, {0x1184, 0x1068, 0x1030, 0x14C8, 0x5B04, 0x5040, 0x53FE, 0x90A0, 0x1120, 0x13FC, 0x1524, 0x2924, 0x2534, 0x4128, 0x4020, 0x8020}},
    {28912, 2, {0x1008, 0x103C, 0x11C0, 0x1004, 0x5544, 0x58A8, 0x5000, 0x91F8, 0x1010, 0x1020, 0x13FE, 0x2820, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {28913, 2, {0x1000, 0x13FE, 0x1202, 0x1652, 0x5A8A, 0x5306, 0x5202, 0x92FA, 0x128A, 0x128A, 0x128A, 0x2AFA, 0x268A, 0x4202, 0x420A, 0x8204}},
    {28914, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {28915, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0x5420, 0x59FC, 0x5124, 0x9124, 0x11FC, 0x1124, 0x1124, 0x29FC, 0x2524, 0x4524, 0x4124, 0x810C}},
    {28916, 2, {0x1000, 0x11FE, 0x1000, 0x1492, 0x5924, 0x5248, 0x5124, 0x9092, 0x1000, 0x11FE, 0x1020, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {28917, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x5400, 0x5880, 0x50FC, 0x9104, 0x1204, 0x1484, 0x1044, 0x2844, 0x2404, 0x4404, 0x4028, 0x8010}},
    {28918, 2, {0x1000, 0x1006, 0x13B8, 0x1488, 0x5888, 0x5108, 0x513E, 0x9388, 0x1088, 0x1288, 0x1288, 0x293E, 0x2500, 0x4280, 0x447E, 0x8800}},
    {28919, 2, {0x1040, 0x1020, 0x13FE, 0x1602, 0x5C04, 0x51F8, 0x5000, 0x9000, 0x13FE, 0x1090, 0x1090, 0x2890, 0x2512, 0x4112, 0x420E, 0x8400}},
    {28920, 2, {0x1080, 0x1080, 0x10FE, 0x1500, 0x5AFC, 0x5084, 0x50A4, 0x9094, 0x13FE, 0x1084, 0x1124, 0x2914, 0x25FE, 0x4404, 0x4028, 0x8010}},
    {28921, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF0, 0x60, 0x180, 0x100, 0x500, 0x200, 0x4888, 0x4444, 0x8444}},
    {28922, 2, {0x1040, 0x1020, 0x11FC, 0x1104, 0x5504, 0x59FC, 0x5104, 0x9104, 0x11FC, 0x1120, 0x1122, 0x2914, 0x2508, 0x4544, 0x4182, 0x8100}},
    {28923, 2, {0x1004, 0x100E, 0x13B8, 0x1488, 0x5888, 0x5128, 0x512E, 0x93A8, 0x10A8, 0x12A8, 0x12A8, 0x293E, 0x2500, 0x4280, 0x447E, 0x8800}},
    {28924, 2, {0x1000, 0x13FE, 0x1040, 0x1480, 0x5944, 0x5224, 0x5068, 0x90B0, 0x1130, 0x1228, 0x1068, 0x28A4, 0x2522, 0x4220, 0x40A0, 0x8040}},
    {28925, 2, {0x1040, 0x1040, 0x10FC, 0x1088, 0x5550, 0x5820, 0x50D8, 0x9326, 0x10F8, 0x1020, 0x10F8, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {28926, 2, {0x100, 0x910, 0x920, 0x12C0, 0x430, 0x1908, 0x6914, 0x910, 0x12A0, 0x440, 0x1830, 0x610C, 0x1FF0, 0x100, 0x100, 0x7FFC}},
    {28927, 2, {0x1002, 0x13C2, 0x1244, 0x1648, 0x5B50, 0x52C2, 0x5242, 0x9244, 0x1FE8, 0x1250, 0x1242, 0x2A42, 0x2644, 0x4248, 0x4450, 0x88E0}},
    {28928, 2, {0x1000, 0x1050, 0x1048, 0x1084, 0x5524, 0x5820, 0x5050, 0x9088, 0x1106, 0x12F8, 0x1088, 0x2888, 0x2488, 0x4488, 0x40F8, 0x8088}},
    {28929, 2, {0x80, 0x80, 0x78FC, 0x4904, 0x4A48, 0x4840, 0x78B0, 0x30C, 0xC02, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {28930, 2, {0x1040, 0x1040, 0x24FE, 0x6488, 0xA550, 0x2420, 0x2450, 0x2088, 0x104, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28931, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0x5420, 0x5820, 0x53FE, 0x9048, 0x1048, 0x114C, 0x114A, 0x2A4A, 0x2488, 0x4488, 0x4128, 0x8210}},
    {28932, 2, {0x0, 0x3FF8, 0x408, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8FF8, 0x808, 0x0, 0x4888, 0x4444, 0x8444}},
    {28933, 2, {0x1020, 0x1120, 0x1120, 0x11FC, 0x5520, 0x5A20, 0x5020, 0x93FE, 0x1000, 0x1000, 0x11FC, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28934, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0x54FC, 0x5800, 0x51FE, 0x9102, 0x1102, 0x11FE, 0x1102, 0x2902, 0x25FE, 0x4502, 0x410A, 0x8104}},
    {28935, 2, {0x1020, 0x1124, 0x10A4, 0x10A8, 0x5420, 0x59FC, 0x5104, 0x9104, 0x11FC, 0x1104, 0x1104, 0x29FC, 0x2504, 0x4504, 0x4114, 0x8108}},
    {28936, 2, {0x1000, 0x87C, 0x3E44, 0x2244, 0x227C, 0x3E40, 0x2242, 0x2042, 0x413E, 0x9110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28937, 2, {0x0, 0x7FFC, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x800, 0xFF8, 0x1000, 0x1FFC, 0x4, 0x2924, 0x2494, 0x4028, 0x10}},
    {28938, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x5504, 0x59FC, 0x5000, 0x9000, 0x11FC, 0x1020, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28939, 2, {0x2420, 0x2420, 0x2420, 0x3DFC, 0x420, 0xFC20, 0x2420, 0x24F8, 0x4400, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {28940, 2, {0x1040, 0x1040, 0x1088, 0x1504, 0x5BFE, 0x5002, 0x5088, 0x9144, 0x1242, 0x10F8, 0x1188, 0x2A50, 0x2420, 0x4050, 0x4188, 0x8606}},
    {28941, 2, {0x1108, 0x1088, 0x1090, 0x15FC, 0x5824, 0x5024, 0x51FC, 0x9120, 0x1120, 0x11FE, 0x1062, 0x28A2, 0x252A, 0x4224, 0x4420, 0x8020}},
    {28942, 2, {0x1000, 0x100C, 0x11F0, 0xFD00, 0x1100, 0x11FE, 0x1D10, 0xF110, 0x1110, 0x1210, 0x5210, 0x2410, 0x0, 0x4888, 0x4444, 0x8444}},
    {28943, 2, {0x0, 0x3FF0, 0x20, 0x40, 0x180, 0x793C, 0x4944, 0x4928, 0x4910, 0x7928, 0x144, 0x500, 0x200, 0x4888, 0x4444, 0x8444}},
    {28944, 2, {0x1000, 0x13FE, 0x1040, 0x1040, 0x55FC, 0x5884, 0x5084, 0x93FE, 0x1000, 0x1000, 0x11FC, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28945, 2, {0x1000, 0x13FE, 0x1222, 0x1622, 0x5A22, 0x53FE, 0x5222, 0x9262, 0x1272, 0x12AA, 0x1322, 0x2A22, 0x2622, 0x4202, 0x43FE, 0x8202}},
    {28946, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0x5484, 0x5884, 0x50FC, 0x9000, 0x1000, 0x11FE, 0x1102, 0x2902, 0x2502, 0x4502, 0x41FE, 0x8102}},
    {28947, 2, {0x1020, 0x1020, 0x1050, 0x1488, 0x5944, 0x5222, 0x50F8, 0x9008, 0x1010, 0x1020, 0x11FC, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28948, 2, {0x1040, 0x1040, 0x107C, 0x1484, 0x5908, 0x5210, 0x513E, 0x9122, 0x1122, 0x1122, 0x113E, 0x2922, 0x2522, 0x4522, 0x413E, 0x8122}},
    {28949, 2, {0x1080, 0x1080, 0x10F8, 0x1508, 0x5A10, 0x55FC, 0x5124, 0x9124, 0x1124, 0x1124, 0x13FE, 0x2850, 0x2450, 0x4488, 0x4104, 0x8602}},
    {28950, 2, {0x1100, 0x10BE, 0x1282, 0x1602, 0x5A22, 0x5212, 0x5252, 0x9242, 0x12CA, 0x1356, 0x1252, 0x2A32, 0x2602, 0x4202, 0x420A, 0x8204}},
    {28951, 2, {0x1000, 0x11FC, 0x1104, 0x1504, 0x59FC, 0x5100, 0x5100, 0x91FE, 0x1102, 0x117A, 0x114A, 0x294A, 0x267A, 0x4202, 0x4414, 0x8008}},
    {28952, 2, {0x200, 0x200, 0x7FFC, 0x200, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x1FFC, 0x2020, 0x4420, 0x8220, 0x2A0, 0x40, 0x2488, 0x4244}},
    {28953, 2, {0x1040, 0x1020, 0x11FC, 0x1000, 0x5508, 0x5890, 0x5000, 0x93FE, 0x1000, 0x1000, 0x11FC, 0x2904, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28954, 2, {0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0x920, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28955, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4104, 0x4104, 0x4924, 0x4924, 0x4944, 0x5184, 0x4244, 0x4224, 0x4424, 0x4804, 0x500C}},
    {28956, 2, {0x1000, 0x11FC, 0x1104, 0x1104, 0x55FC, 0x5904, 0x5104, 0x91FC, 0x1000, 0x1112, 0x11D4, 0x2918, 0x2510, 0x4552, 0x4192, 0x810E}},
    {28957, 2, {0x101C, 0x13E0, 0x1220, 0x13FE, 0x5620, 0x5A92, 0x530A, 0x9206, 0x11FC, 0x1104, 0x1104, 0x29FC, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28958, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x55FC, 0x5904, 0x51FC, 0x9000, 0x11FC, 0x1008, 0x1010, 0x2BFE, 0x2420, 0x4420, 0x40A0, 0x8040}},
    {28959, 2, {0x1088, 0x1088, 0x1088, 0x17FE, 0x5888, 0x5088, 0x57FE, 0x9000, 0x11FC, 0x1104, 0x1104, 0x29FC, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {28960, 2, {0x1040, 0x1020, 0x13FE, 0x1000, 0x5488, 0x5888, 0x5154, 0x9222, 0x1000, 0x1020, 0x13FE, 0x2820, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28961, 2, {0x1000, 0x1000, 0x1FFC, 0x2AA0, 0x4AA0, 0xAA0, 0x7FFC, 0xAA0, 0xAA0, 0xAA0, 0x7FFC, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {28962, 2, {0x1020, 0x1010, 0x1010, 0x15FE, 0x5902, 0x5254, 0x5088, 0x9104, 0x1000, 0x10FC, 0x1020, 0x2820, 0x2420, 0x4420, 0x41FE, 0x8000}},
    {28963, 2, {0x0, 0x7F00, 0x22FC, 0x3E44, 0x2244, 0x3E44, 0x2228, 0x2228, 0x2F10, 0xF228, 0x4244, 0x282, 0x0, 0x4888, 0x4444, 0x8444}},
    {28964, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0x25FE, 0x2C90, 0x3450, 0x4410, 0x4450, 0x8520, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {28965, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x5504, 0x5900, 0x51DE, 0x9252, 0x1252, 0x1352, 0x149A, 0x2894, 0x2510, 0x4512, 0x4212, 0x840E}},
    {28966, 2, {0x900, 0x880, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000, 0x4888, 0x4444, 0x8444}},
    {28967, 2, {0x1090, 0x1090, 0x1108, 0x1548, 0x5A44, 0x5492, 0x5108, 0x93FC, 0x1044, 0x1020, 0x10A4, 0x2A82, 0x268A, 0x428A, 0x4478, 0x8000}},
    {28968, 2, {0x1000, 0x13FC, 0x1204, 0x12F4, 0x5604, 0x5AF4, 0x5000, 0x91F8, 0x1108, 0x11F8, 0x1108, 0x29F8, 0x2508, 0x4508, 0x4128, 0x8110}},
    {28969, 2, {0x1000, 0x13DE, 0x1252, 0x1652, 0x5A52, 0x53DE, 0x5252, 0x9252, 0x1252, 0x13DE, 0x1252, 0x2A52, 0x2652, 0x4252, 0x4552, 0x88A6}},
    {28970, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x5488, 0x5904, 0x5202, 0x91FC, 0x1004, 0x11FC, 0x1100, 0x29FC, 0x2404, 0x4404, 0x4028, 0x8010}},
    {28971, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x54A8, 0x5820, 0x51FC, 0x9124, 0x1124, 0x1154, 0x114C, 0x2984, 0x2504, 0x4504, 0x4114, 0x8108}},
    {28972, 2, {0x1000, 0x11F8, 0x1108, 0x1508, 0x59F8, 0x5108, 0x5108, 0x91F8, 0x1080, 0x11FC, 0x1254, 0x2894, 0x2524, 0x4244, 0x4094, 0x8108}},
    {28973, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0xFE0, 0x820, 0x820, 0x822, 0x1022, 0x201E, 0xC000}},
    {28974, 2, {0x2010, 0x2090, 0x2710, 0x243E, 0xAC22, 0xB444, 0xA790, 0xA510, 0x2510, 0x2510, 0x2528, 0x2528, 0x5528, 0x4944, 0x4944, 0x9082}},
    {28975, 2, {0x1020, 0x1020, 0x103E, 0x1420, 0x59FC, 0x5104, 0x51FC, 0x9104, 0x11FC, 0x1124, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28976, 2, {0x1040, 0x1040, 0x107C, 0x1484, 0x5908, 0x5200, 0x5020, 0x91CE, 0x1102, 0x1102, 0x11CE, 0x2902, 0x2502, 0x4502, 0x41FE, 0x8102}},
    {28977, 2, {0x100, 0x1108, 0x1108, 0x2290, 0x440, 0x1830, 0xE00E, 0x810, 0xA10, 0x2A54, 0x2C54, 0x4898, 0x1410, 0x2228, 0x4244, 0x8082}},
    {28978, 2, {0x2080, 0x2040, 0x27FE, 0x2120, 0xA920, 0xB23C, 0xA244, 0xA664, 0x2294, 0x2348, 0x2248, 0x2230, 0x5220, 0x4A50, 0x4288, 0x8306}},
    {28979, 2, {0x10A0, 0x1090, 0x1080, 0x15FE, 0x5910, 0x5310, 0x55FC, 0x9110, 0x1110, 0x11FC, 0x1110, 0x2910, 0x2510, 0x45FE, 0x4100, 0x8100}},
    {28980, 2, {0x1040, 0x1020, 0x13FE, 0x1080, 0x5504, 0x5BFE, 0x5002, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x29FC, 0x2504, 0x4504, 0x4114, 0x8108}},
    {28981, 2, {0x1000, 0x13FE, 0x1202, 0x168A, 0x5A52, 0x53FE, 0x5222, 0x9222, 0x12AA, 0x12AA, 0x12AA, 0x2AFA, 0x2602, 0x4202, 0x420A, 0x8204}},
    {28982, 2, {0x1020, 0x1028, 0x1E24, 0x2220, 0x33FE, 0x4A20, 0xA450, 0x1450, 0x888, 0x1088, 0x2104, 0x4202, 0x0, 0x4888, 0x4444, 0x8444}},
    {28983, 2, {0x1040, 0x1080, 0x11FC, 0x1124, 0x5524, 0x59FC, 0x5124, 0x9144, 0x11FC, 0x1090, 0x1110, 0x2BFE, 0x2410, 0x4410, 0x4010, 0x8010}},
    {28984, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {28985, 2, {0x1000, 0x13FE, 0x1202, 0x168A, 0x5A52, 0x53FE, 0x5242, 0x9222, 0x13FE, 0x1282, 0x1282, 0x2A82, 0x26FA, 0x4202, 0x420A, 0x8204}},
    {28986, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x5504, 0x59FC, 0x5000, 0x9028, 0x11C8, 0x1088, 0x1088, 0x2BFE, 0x2488, 0x4488, 0x4108, 0x8208}},
    {28987, 2, {0x11F8, 0x1108, 0x1108, 0x15F8, 0x5908, 0x5108, 0x51F8, 0x9000, 0x13FC, 0x1204, 0x1204, 0x2BFC, 0x2604, 0x4204, 0x43FC, 0x8204}},
    {28988, 2, {0x1020, 0x1020, 0x11FC, 0x1020, 0x54A8, 0x5888, 0x53FE, 0x9088, 0x1000, 0x13FE, 0x1090, 0x2890, 0x2512, 0x4512, 0x420E, 0x8400}},
    {28989, 2, {0x0, 0x7CFC, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7C84, 0x114, 0x208, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {28990, 2, {0x1020, 0x1020, 0x1050, 0x1488, 0x5944, 0x5222, 0x51F8, 0x9008, 0x1050, 0x1020, 0x10A4, 0x2A82, 0x268A, 0x428A, 0x4478, 0x8000}},
    {28991, 2, {0x1040, 0x1020, 0x13FE, 0x1620, 0x5AFC, 0x5224, 0x5224, 0x93FE, 0x1224, 0x1224, 0x12FC, 0x2A20, 0x2650, 0x4450, 0x4488, 0x8906}},
    {28992, 2, {0x1000, 0x13FE, 0x1202, 0x1602, 0x5BFE, 0x5210, 0x5292, 0x9292, 0x1292, 0x12FE, 0x1210, 0x2A92, 0x2692, 0x4492, 0x44FE, 0x8802}},
    {28993, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x54F8, 0x5888, 0x50F8, 0x9088, 0x1088, 0x13FE, 0x1100, 0x2948, 0x2584, 0x4500, 0x41FE, 0x8000}},
    {28994, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x54F8, 0x5820, 0x51FC, 0x9124, 0x1124, 0x11FC, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x4020, 0x8020}},
    {28995, 2, {0x1020, 0x1020, 0x13FE, 0x1050, 0x5488, 0x5B26, 0x51FC, 0x9020, 0x1020, 0x13FE, 0x1000, 0x2820, 0x25FC, 0x4420, 0x4020, 0x83FE}},
    {28996, 2, {0x1008, 0x103C, 0x11E0, 0x1420, 0x5BFE, 0x5020, 0x51FC, 0x9124, 0x11FC, 0x1124, 0x11FC, 0x2820, 0x25FC, 0x4420, 0x43FE, 0x8000}},
    {28997, 2, {0x1020, 0x11DC, 0x1114, 0x1114, 0x5514, 0x59D4, 0x5126, 0x9100, 0x11DC, 0x1114, 0x1114, 0x29D4, 0x2708, 0x4508, 0x4114, 0x8122}},
    {28998, 2, {0x1000, 0x13DE, 0x1242, 0x1642, 0x5A42, 0x53DE, 0x5200, 0x923E, 0x13D2, 0x1212, 0x1214, 0x2BD4, 0x2608, 0x4214, 0x4224, 0x8242}},
    {28999, 2, {0x1000, 0x13FE, 0x1222, 0x1020, 0x55FE, 0x5820, 0x51FC, 0x9124, 0x11FC, 0x1124, 0x11FC, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {29000, 2, {0x2000, 0x27FC, 0x2404, 0x2434, 0xADC4, 0xB444, 0xA5F4, 0xA554, 0x2554, 0x25F4, 0x2444, 0x2454, 0x55F4, 0x4C94, 0x4806, 0x9002}},
    {29001, 2, {0x1020, 0x1020, 0x13FE, 0x1020, 0x55FC, 0x5924, 0x51AC, 0x9174, 0x1124, 0x11FC, 0x1020, 0x2870, 0x24A8, 0x4524, 0x4222, 0x8020}},
    {29002, 2, {0x1040, 0x1020, 0x13FE, 0x1202, 0x5404, 0x59FC, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x29FC, 0x2504, 0x4400, 0x43FE, 0x8000}},
    {29003, 2, {0x1000, 0x11FC, 0x1104, 0x15FC, 0x5904, 0x51FC, 0x5020, 0x9120, 0x11FE, 0x1220, 0x1020, 0x29FC, 0x2420, 0x4020, 0x43FE, 0x8000}},
    {29004, 2, {0x1020, 0x1040, 0x11FC, 0x1504, 0x59FC, 0x5104, 0x51FC, 0x9000, 0x13FE, 0x1020, 0x1020, 0x29FC, 0x2420, 0x4020, 0x43FE, 0x8000}},
    {29005, 2, {0x2088, 0x21C8, 0x2708, 0x2108, 0xA92A, 0xB12A, 0xA7AC, 0xA148, 0x2308, 0x2388, 0x2554, 0x2514, 0x5914, 0x4924, 0x4124, 0x8142}},
    {29006, 2, {0x1010, 0x820, 0xFFFE, 0x0, 0x3E48, 0x2248, 0x3E48, 0x2248, 0x3E08, 0x2208, 0x2A28, 0x2410, 0x0, 0x4888, 0x4444, 0x8444}},
    {29007, 2, {0x1000, 0x13FE, 0x1000, 0x15FC, 0x5904, 0x5104, 0x51FC, 0x9000, 0x13FE, 0x1222, 0x1222, 0x2BFE, 0x2622, 0x4222, 0x43FE, 0x8202}},
    {29008, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x54A8, 0x5820, 0x51FC, 0x9124, 0x1124, 0x1124, 0x13FE, 0x2820, 0x2450, 0x4488, 0x4104, 0x8202}},
    {29009, 2, {0x108, 0x1FD0, 0x120, 0x7FFC, 0x200, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {29010, 2, {0x2040, 0x2040, 0x21FC, 0x2084, 0xABFE, 0xB000, 0xA1FC, 0xA104, 0x21FC, 0x2020, 0x23FE, 0x2020, 0x5220, 0x4BFE, 0x4020, 0x8020}},
    {29011, 2, {0x1020, 0x1124, 0x1124, 0x1124, 0x55FC, 0x5800, 0x53FE, 0x9020, 0x1040, 0x11FC, 0x1154, 0x2954, 0x2554, 0x4554, 0x4154, 0x810C}},
    {29012, 2, {0x1020, 0x5220, 0x5420, 0x9020, 0x283C, 0x4420, 0x8220, 0x1020, 0x10FC, 0x5284, 0x5484, 0x9084, 0x2884, 0x2484, 0x42FC, 0x8084}},
    {29013, 2, {0x0, 0x7F7C, 0x4444, 0x5444, 0x5744, 0x517C, 0x5140, 0x5740, 0x5442, 0x4442, 0x7F3E, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {29014, 2, {0x101E, 0x13E0, 0x1044, 0x1524, 0x5888, 0x51FC, 0x5040, 0x9040, 0x13FE, 0x1080, 0x10FC, 0x2944, 0x2528, 0x4210, 0x4068, 0x8186}},
    {29015, 2, {0x1000, 0x13FE, 0x1020, 0x1040, 0x55FC, 0x5954, 0x5154, 0x9154, 0x1154, 0x112C, 0x1020, 0x2BFE, 0x2450, 0x4488, 0x4104, 0x8202}},
    {29016, 2, {0x2014, 0x2012, 0x2010, 0x27FE, 0xAC10, 0xB410, 0xA5D0, 0xA412, 0x2412, 0x25D4, 0x2554, 0x2548, 0x55DA, 0x4C2A, 0x4846, 0x9082}},
    {29017, 2, {0x1000, 0x13FE, 0x1050, 0x1450, 0x5BFE, 0x5252, 0x5252, 0x9252, 0x13FE, 0x1020, 0x1020, 0x2BFE, 0x2420, 0x4020, 0x47FE, 0x8000}},
    {29018, 2, {0x0, 0x7CFE, 0x4480, 0x44FC, 0x7C84, 0x44FC, 0x4480, 0x7CFE, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {29019, 2, {0x0, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {29020, 2, {0x1000, 0x11FC, 0x1104, 0x15FC, 0x5904, 0x51FC, 0x5104, 0x9040, 0x1020, 0x13FE, 0x1000, 0x2904, 0x2488, 0x4000, 0x47FE, 0x8000}},
    {29021, 2, {0x1000, 0x13FC, 0x1224, 0x1624, 0x5BFC, 0x5200, 0x52FC, 0x9284, 0x1284, 0x12FC, 0x1284, 0x2AFC, 0x2684, 0x4284, 0x44FC, 0x8884}},
    {29022, 2, {0x2020, 0x2020, 0x3C40, 0x447E, 0x8884, 0x7F44, 0x244, 0x228, 0x7E28, 0x210, 0x228, 0xFE44, 0x82, 0x4888, 0x4444, 0x8444}},
    {29023, 2, {0x1000, 0x13FE, 0x1222, 0x13FE, 0x5622, 0x5BFE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x29FC, 0x2504, 0x4504, 0x4114, 0x8108}},
    {29024, 2, {0x1048, 0x1148, 0x1148, 0x17FE, 0x5948, 0x5148, 0x5178, 0x9100, 0x11FE, 0x1020, 0x13FE, 0x2870, 0x24A8, 0x4124, 0x4622, 0x8020}},
    {29025, 2, {0x2010, 0x2010, 0x207C, 0x2714, 0xA9FE, 0xB114, 0xA27C, 0xA210, 0x277C, 0x2110, 0x21FE, 0x2510, 0x5210, 0x4B00, 0x44FE, 0x8800}},
    {29026, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0x1FE0, 0x420, 0x420, 0x3FA0, 0x420, 0x414, 0x40C, 0x404}},
    {29027, 2, {0x1000, 0x11FC, 0x1048, 0x1430, 0x5BFE, 0x5052, 0x5094, 0x9110, 0x1230, 0x1020, 0x13FE, 0x2870, 0x24A8, 0x4124, 0x4622, 0x8020}},
    {29028, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x5488, 0x58F8, 0x5088, 0x9088, 0x10F8, 0x1020, 0x13FE, 0x2870, 0x24A8, 0x4524, 0x4222, 0x8020}},
    {29029, 2, {0x2080, 0x2080, 0x20F8, 0x2108, 0xAA10, 0xB5FC, 0xA104, 0xA154, 0x218C, 0x2124, 0x2020, 0x23FE, 0x5050, 0x4888, 0x4104, 0x8602}},
    {29030, 2, {0x100, 0x7900, 0x49FC, 0x4A04, 0x4C04, 0x79E4, 0x4924, 0x4924, 0x49E4, 0x4804, 0x7828, 0x10, 0x2488, 0x2244, 0x4244, 0x8004}},
    {29031, 2, {0x0, 0x7DFC, 0x4444, 0x4444, 0x4444, 0x4494, 0x7D08, 0x44FC, 0x4484, 0x4484, 0x4484, 0x7CFC, 0x0, 0x4888, 0x4444, 0x8444}},
    {29032, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x55FC, 0x5924, 0x5124, 0x91FC, 0x1000, 0x13FE, 0x1120, 0x2922, 0x2514, 0x4548, 0x4184, 0x8102}},
    {29033, 2, {0x1000, 0x13FE, 0x1020, 0x1440, 0x59FC, 0x5104, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x2904, 0x25FC, 0x4400, 0x4088, 0x8104}},
    {29034, 2, {0x1108, 0x1088, 0x1090, 0x17FE, 0x5890, 0x5090, 0x53FC, 0x9294, 0x1294, 0x130C, 0x1204, 0x2BFC, 0x2604, 0x4204, 0x43FC, 0x8204}},
    {29035, 2, {0x2108, 0x2090, 0x2000, 0x2BFE, 0xB080, 0xA184, 0xA244, 0xA4A8, 0x2130, 0x2270, 0x24A8, 0x2128, 0x5224, 0x4C22, 0x48A0, 0x8040}},
    {29036, 2, {0x2000, 0x23F8, 0x2208, 0x2BF8, 0xB208, 0xA3F8, 0xA000, 0xA7FE, 0x2200, 0x23FC, 0x2494, 0x2924, 0x5244, 0x4C84, 0x4928, 0x8210}},
    {29037, 2, {0x2484, 0x4904, 0x24A4, 0x24, 0x7FA4, 0x824, 0x1F24, 0x6124, 0x1224, 0xC04, 0x3014, 0xC008, 0x0, 0x4888, 0x4444, 0x8444}},
    {29038, 2, {0x200, 0x208, 0x3FD0, 0x220, 0xFFFE, 0x100, 0xFF0, 0x3810, 0xCFF0, 0x810, 0x810, 0xFF0, 0x0, 0x4888, 0x4444, 0x8444}},
    {29039, 2, {0x1110, 0x1112, 0x11D4, 0x1118, 0x5552, 0x5992, 0x512E, 0x9040, 0x11FC, 0x1104, 0x1104, 0x29FC, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {29040, 2, {0x1000, 0x10F8, 0x1088, 0x1488, 0x5888, 0x5088, 0x50F8, 0x9088, 0x1000, 0x13DE, 0x1252, 0x2A52, 0x2652, 0x4252, 0x43DE, 0x8252}},
    {29041, 2, {0x1000, 0x11FC, 0x1104, 0x1504, 0x59E4, 0x5124, 0x5124, 0x93FE, 0x1202, 0x12FA, 0x128A, 0x2A8A, 0x26FA, 0x4202, 0x420A, 0x8204}},
    {29042, 2, {0x800, 0xBF8, 0x1208, 0x33F8, 0x5040, 0x97FC, 0x1150, 0x1248, 0x1444, 0x1040, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {29043, 2, {0x1100, 0x113C, 0x1124, 0x17A4, 0x5924, 0x513C, 0x5124, 0x93A4, 0x12A4, 0x12BC, 0x12A4, 0x2AA4, 0x27A4, 0x4444, 0x4054, 0x8088}},
    {29044, 2, {0x1000, 0x11F8, 0x1108, 0x1508, 0x59F8, 0x5108, 0x5108, 0x91F8, 0x1000, 0x13FC, 0x1294, 0x2A94, 0x2694, 0x4294, 0x47FE, 0x8000}},
    {29045, 2, {0x1020, 0x1020, 0x13FE, 0x1420, 0x5820, 0x53FE, 0x528A, 0x9252, 0x12FA, 0x1222, 0x1222, 0x2AFA, 0x2622, 0x4222, 0x422A, 0x8204}},
    {29046, 2, {0x1000, 0x11FC, 0x1104, 0x1504, 0x59FC, 0x5104, 0x5104, 0x91FC, 0x1000, 0x13FE, 0x1020, 0x2920, 0x253C, 0x4120, 0x42A0, 0x847E}},
    {29047, 2, {0x2080, 0x2040, 0x27FE, 0x2000, 0xABF8, 0xB208, 0xA3F8, 0xA000, 0x27FE, 0x2402, 0x29F4, 0x2110, 0x5110, 0x4A12, 0x4412, 0x880E}},
    {29048, 2, {0x1040, 0x1020, 0x11FE, 0x1102, 0x5502, 0x59FE, 0x5100, 0x9100, 0x11FE, 0x11AA, 0x12AA, 0x2AFE, 0x24AA, 0x44AA, 0x40A2, 0x8086}},
    {29049, 2, {0x1088, 0x1088, 0x13FE, 0x1488, 0x59FC, 0x5088, 0x53FE, 0x9020, 0x11FC, 0x1124, 0x11FC, 0x2924, 0x27FE, 0x4104, 0x4114, 0x8108}},
    {29050, 2, {0x2000, 0x24F8, 0x2288, 0x2288, 0xA8F8, 0xB088, 0xAE88, 0xA2F8, 0x22A2, 0x2294, 0x2288, 0x22A8, 0x52C4, 0x4A84, 0x4500, 0x88FE}},
    {29051, 2, {0x1020, 0x1010, 0x13FE, 0x1620, 0x5AFC, 0x5224, 0x53FE, 0x9224, 0x12FC, 0x1220, 0x12FC, 0x2A84, 0x2684, 0x4484, 0x44FC, 0x8884}},
    {29052, 2, {0x2080, 0x2080, 0x21FC, 0x2A44, 0xB154, 0xA1F4, 0xA084, 0xA128, 0x2290, 0x21FC, 0x2244, 0x2554, 0x51F4, 0x4884, 0x4928, 0x8210}},
    {29053, 2, {0x1020, 0x1010, 0x11FE, 0x1502, 0x5902, 0x51FE, 0x5100, 0x91EE, 0x1122, 0x11AA, 0x1166, 0x2922, 0x2566, 0x42AA, 0x4222, 0x8466}},
    {29054, 2, {0x1000, 0x11FC, 0x1124, 0x1524, 0x59FC, 0x5124, 0x5154, 0x918C, 0x1104, 0x11FC, 0x1040, 0x2820, 0x24A4, 0x428A, 0x428A, 0x8478}},
    {29055, 2, {0x1028, 0x1024, 0x13FE, 0x1020, 0x55FC, 0x5924, 0x51FC, 0x9124, 0x11FC, 0x1124, 0x1008, 0x2BFE, 0x2488, 0x4448, 0x4048, 0x8018}},
    {29056, 2, {0x1000, 0x11FC, 0x1104, 0x11FC, 0x5504, 0x59FC, 0x5020, 0x9124, 0x10A8, 0x1020, 0x13FE, 0x2890, 0x2490, 0x4512, 0x4212, 0x840E}},
    {29057, 2, {0x1040, 0x11FC, 0x1044, 0x1094, 0x5508, 0x5BDE, 0x514A, 0x914A, 0x1294, 0x11FC, 0x1104, 0x29FC, 0x2504, 0x45FC, 0x4104, 0x810C}},
    {29058, 2, {0x2100, 0x2100, 0x21FC, 0x2A00, 0xB5F8, 0xA000, 0xA3F8, 0xA088, 0x22A8, 0x21C8, 0x27F8, 0x208A, 0x51CA, 0x4EAA, 0x4886, 0x8082}},
    {29059, 2, {0x40, 0x7EFC, 0x1544, 0x838, 0xFE44, 0x1AA2, 0x2CFC, 0x4844, 0xA894, 0x1108, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {29060, 2, {0x1020, 0x1040, 0x11FC, 0x1504, 0x59FC, 0x5104, 0x51FC, 0x9104, 0x11FC, 0x1040, 0x1020, 0x28A4, 0x268A, 0x428A, 0x4478, 0x8000}},
    {29061, 2, {0x1000, 0x13F8, 0x1248, 0x12A8, 0x5718, 0x5A08, 0x53F8, 0x9000, 0x1000, 0x17FC, 0x14A4, 0x2CA4, 0x24A4, 0x44A4, 0x4FFE, 0x8000}},
    {29062, 2, {0x1020, 0x1020, 0x11FC, 0x1420, 0x5820, 0x53FE, 0x5080, 0x9104, 0x13FE, 0x1002, 0x13FC, 0x2A94, 0x2694, 0x4294, 0x47FE, 0x8000}},
    {29063, 2, {0x1040, 0x1020, 0x13FE, 0x1400, 0x59FC, 0x5104, 0x51FC, 0x9000, 0x13FE, 0x1202, 0x12FA, 0x2A8A, 0x26FA, 0x4202, 0x420A, 0x8204}},
    {29064, 2, {0x0, 0x2FBC, 0x2A24, 0x2A24, 0x2FA4, 0x28BC, 0x28A0, 0x2FA0, 0x2A22, 0x4A22, 0x4F9E, 0x8000, 0x2488, 0x2244, 0x4244, 0x8004}},
    {29065, 2, {0x1000, 0x10FC, 0x1084, 0x1084, 0x54FC, 0x5800, 0x51FE, 0x9102, 0x11FE, 0x1102, 0x11FE, 0x2902, 0x25FE, 0x4448, 0x4084, 0x8102}},
    {29066, 2, {0x1080, 0x2498, 0x42E0, 0x7E84, 0x84, 0x7E7C, 0x4280, 0x7E98, 0x42E0, 0x7E84, 0x4284, 0x467C, 0x0, 0x4888, 0x4444, 0x8444}},
    {29067, 2, {0x1080, 0x2498, 0x7EE0, 0x84, 0x7E7C, 0x4200, 0x7E98, 0x42E0, 0x7E84, 0x4284, 0x467C, 0x100, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {29068, 2, {0x1000, 0x13DE, 0x1252, 0x17DE, 0x5A52, 0x53DE, 0x5202, 0x9222, 0x1222, 0x1222, 0x1252, 0x2A4A, 0x268A, 0x4202, 0x420A, 0x8204}},
    {29069, 2, {0x1020, 0x1010, 0x11FE, 0x1102, 0x5400, 0x58FC, 0x5084, 0x9084, 0x10FC, 0x1000, 0x11FE, 0x2902, 0x2502, 0x4502, 0x41FE, 0x8102}},
    {29070, 2, {0x1008, 0x103C, 0x13C0, 0x1004, 0x5644, 0x5928, 0x51FC, 0x9220, 0x1020, 0x13FE, 0x1020, 0x2924, 0x2524, 0x4524, 0x41FC, 0x8004}},
    {29071, 2, {0xF8, 0x3F00, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x4888, 0x4444, 0x8444}},
    {29072, 2, {0x1000, 0x13FE, 0x1202, 0x1000, 0x55FC, 0x5904, 0x51FC, 0x9104, 0x11FC, 0x1040, 0x1020, 0x2BFE, 0x2400, 0x4488, 0x4104, 0x8202}},
    {29073, 2, {0x1088, 0x1050, 0x1000, 0x15FE, 0x5850, 0x51FC, 0x5054, 0x93FE, 0x1054, 0x11FC, 0x1050, 0x28D8, 0x2554, 0x4252, 0x4050, 0x8050}},
    {29074, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8104, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {29075, 2, {0x2080, 0x2100, 0x23F8, 0x2208, 0xAA08, 0xB3F8, 0xA200, 0xA3FE, 0x2200, 0x23FC, 0x2004, 0x2554, 0x5554, 0x4804, 0x4028, 0x8010}},
    {29076, 2, {0x1040, 0x1020, 0x13FE, 0x1602, 0x5850, 0x5088, 0x5124, 0x9050, 0x1088, 0x1104, 0x12FA, 0x2888, 0x2488, 0x4488, 0x40F8, 0x8088}},
    {29077, 2, {0x1000, 0x11FC, 0x1020, 0x1420, 0x5BFE, 0x5000, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x2904, 0x25FC, 0x4488, 0x4104, 0x8202}},
    {29078, 2, {0x101E, 0x11E0, 0x1022, 0x1112, 0x5494, 0x5880, 0x5020, 0x91CE, 0x1102, 0x1102, 0x11CE, 0x2902, 0x2502, 0x4502, 0x41FE, 0x8102}},
    {29079, 2, {0x2040, 0x2040, 0x20A0, 0x2990, 0xB248, 0xA7FE, 0xAA08, 0xA3F8, 0x2208, 0x23F8, 0x2200, 0x23F8, 0x5508, 0x4D08, 0x49F8, 0x8108}},
    {29080, 2, {0x1040, 0x13BE, 0x1212, 0x1692, 0x5A52, 0x52AA, 0x5324, 0x9040, 0x13FE, 0x1222, 0x1222, 0x2BFE, 0x2622, 0x4222, 0x43FE, 0x8202}},
    {29081, 2, {0x0, 0x7F7C, 0x4444, 0x4444, 0x5F44, 0x517C, 0x5140, 0x5F40, 0x4442, 0x4442, 0x7FBE, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {29082, 2, {0x1000, 0x11FC, 0x1124, 0x15FC, 0x5924, 0x51FC, 0x50A8, 0x90A8, 0x13FE, 0x10A8, 0x10A8, 0x29FC, 0x2420, 0x43FE, 0x4020, 0x8020}},
    {29083, 2, {0x1000, 0x13FE, 0x1050, 0x11FC, 0x5554, 0x5954, 0x51FC, 0x9000, 0x11FC, 0x1000, 0x13FE, 0x2820, 0x24A8, 0x4524, 0x42A2, 0x8040}},
    {29084, 2, {0x1020, 0x1040, 0x11FC, 0x1524, 0x597C, 0x518C, 0x5154, 0x9124, 0x1154, 0x11FC, 0x1020, 0x2810, 0x2554, 0x4542, 0x414A, 0x8238}},
    {29085, 2, {0x1040, 0x1020, 0x13FE, 0x1250, 0x5650, 0x5BFE, 0x5252, 0x9252, 0x13FE, 0x1200, 0x1292, 0x2AD4, 0x2698, 0x4492, 0x44D2, 0x888E}},
    {29086, 2, {0x1000, 0x13DC, 0x1294, 0x17D4, 0x5A54, 0x53C8, 0x5288, 0x9294, 0x13E2, 0x1000, 0x1020, 0x29FC, 0x2420, 0x4020, 0x43FE, 0x8000}},
    {29087, 2, {0x1040, 0xFE40, 0x40, 0x7DF8, 0x4448, 0x7C48, 0xC8, 0x7C48, 0x8AA, 0xFEAA, 0x1106, 0x5202, 0x2000, 0x4888, 0x4444, 0x8444}},
    {29088, 2, {0x1000, 0x13DE, 0x1042, 0x1252, 0x554A, 0x5A52, 0x5042, 0x9020, 0x11FC, 0x1104, 0x1104, 0x29FC, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {29089, 2, {0x2040, 0x23F8, 0x2248, 0x2FFE, 0xAA48, 0xB3F8, 0xA040, 0xA3F8, 0x2248, 0x23F8, 0x2080, 0x2FFE, 0x5110, 0x48E0, 0x4318, 0x8C04}},
    {29090, 2, {0x1020, 0x147C, 0x1284, 0x1548, 0x5830, 0x50CE, 0x5610, 0x927C, 0x1210, 0x127C, 0x1210, 0x2AFE, 0x2610, 0x4210, 0x45FE, 0x8800}},
    {29091, 2, {0x1020, 0x1222, 0x13FE, 0x1090, 0x5488, 0x59FE, 0x5310, 0x9510, 0x11FE, 0x1110, 0x1110, 0x29FE, 0x2510, 0x4510, 0x41FE, 0x8100}},
    {29092, 2, {0x1000, 0x13DE, 0x1042, 0x114A, 0x54C6, 0x594A, 0x5252, 0x9002, 0x1040, 0x1020, 0x11FC, 0x2800, 0x2508, 0x4488, 0x4090, 0x83FE}},
    {29093, 2, {0x20FE, 0x2404, 0x2228, 0x2210, 0xA8FE, 0xB092, 0xAE92, 0xA2FE, 0x2292, 0x22FE, 0x2292, 0x2292, 0x5296, 0x4D00, 0x48FE, 0x8000}},
    {29094, 2, {0x1020, 0x1124, 0x1122, 0x152A, 0x5A32, 0x5060, 0x5190, 0x97FE, 0x1120, 0x11FC, 0x1120, 0x29FC, 0x2520, 0x4520, 0x41FE, 0x8100}},
    {29095, 2, {0x1110, 0x2110, 0xCAA8, 0x1444, 0x3240, 0xD27C, 0x1540, 0x14FE, 0x800, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {29096, 2, {0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5FA8, 0x4408, 0x9528, 0x2C90, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {29097, 2, {0x1040, 0x1020, 0x13FC, 0x1604, 0x5BFC, 0x5200, 0x52F8, 0x9288, 0x12F8, 0x1200, 0x13FC, 0x2B24, 0x25FC, 0x4502, 0x4902, 0x80FE}},
    {29098, 2, {0x2010, 0x2190, 0x261E, 0x2222, 0xAA54, 0xB208, 0xAF90, 0xA264, 0x2208, 0x271E, 0x2AA2, 0x2242, 0x5214, 0x4A08, 0x4210, 0x8260}},
    {29099, 2, {0x1040, 0x1020, 0x13FE, 0x1200, 0x5648, 0x5A48, 0x53FE, 0x9248, 0x1248, 0x1248, 0x1278, 0x2A00, 0x2554, 0x452A, 0x4A2A, 0x9000}},
    {29100, 2, {0x820, 0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x8000, 0x2488, 0x2244, 0x4244, 0x8004}},
    {29101, 2, {0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x3FF8, 0x8, 0x1FF8, 0x8, 0x3FF8, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {29102, 2, {0x2000, 0x27BC, 0x2084, 0x24A4, 0xAA94, 0xB4A4, 0xA050, 0xA188, 0x2626, 0x20C0, 0x2310, 0x2064, 0x5388, 0x4830, 0x40C0, 0x8700}},
    {29103, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0x54F8, 0x5820, 0x51FC, 0x9124, 0x11FC, 0x1020, 0x11FC, 0x2820, 0x27FE, 0x4450, 0x4088, 0x8306}},
    {29104, 2, {0x1000, 0x13FE, 0x1200, 0x167C, 0x5A44, 0x5244, 0x527C, 0x9200, 0x12EE, 0x12AA, 0x12AA, 0x2AAA, 0x26EE, 0x4200, 0x43FE, 0x8000}},
    {29105, 2, {0x1040, 0x1040, 0x7C40, 0x10F8, 0xFE48, 0x4448, 0x92C8, 0x7C48, 0x10AA, 0x1CAA, 0xE106, 0x4202, 0x0, 0x4888, 0x4444, 0x8444}},
    {29106, 2, {0x4000, 0x44FE, 0x7810, 0x4220, 0x42FC, 0x3E84, 0x84, 0x10FC, 0x5284, 0x52FC, 0x5484, 0x9084, 0x28FC, 0x2448, 0x4484, 0x8102}},
    {29107, 2, {0x1000, 0x11F8, 0x1108, 0x11F8, 0x5508, 0x59F8, 0x5000, 0x93FC, 0x1294, 0x13FC, 0x1000, 0x29F8, 0x2490, 0x4460, 0x4198, 0x8606}},
    {29108, 2, {0x2040, 0x2444, 0x2444, 0x27FC, 0xA800, 0xB7FC, 0xA404, 0xA7FC, 0x2404, 0x27FC, 0x2000, 0x2424, 0x57A8, 0x4C32, 0x45A2, 0x861E}},
    {29109, 2, {0x1040, 0x1020, 0x13FE, 0x1488, 0x5850, 0x53FE, 0x5252, 0x928A, 0x1306, 0x12FA, 0x128A, 0x2A8A, 0x26FA, 0x4202, 0x420A, 0x8204}},
    {29110, 2, {0x2000, 0x23F8, 0x2208, 0x23F8, 0xAA08, 0xB3F8, 0xA000, 0xAFFE, 0x2480, 0x27BC, 0x2494, 0x2794, 0x54D4, 0x4F88, 0x4094, 0x80A2}},
    {29111, 2, {0x1104, 0x1088, 0x1000, 0x13FE, 0x5622, 0x5AAA, 0x5272, 0x9222, 0x13FE, 0x1000, 0x11FC, 0x2904, 0x25FC, 0x4504, 0x41FC, 0x8104}},
    {29112, 2, {0x2000, 0x27BC, 0x2108, 0x2528, 0xAFBC, 0xB318, 0xA5AA, 0xA946, 0x2000, 0x23F8, 0x2208, 0x2208, 0x53F8, 0x4A08, 0x4208, 0x83F8}},
    {29113, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2488, 0x4244}},
    {29114, 2, {0x1020, 0x13FE, 0x1020, 0x11FC, 0x5400, 0x59FC, 0x5104, 0x91FC, 0x1088, 0x13FE, 0x1000, 0x29FC, 0x2504, 0x4504, 0x41FC, 0x8104}},
    {29115, 2, {0x2040, 0x20A0, 0x2118, 0x2EE6, 0xB000, 0xA3F8, 0xA208, 0xA3F8, 0x2000, 0x27BC, 0x2084, 0x24A4, 0x5294, 0x4CA4, 0x4A94, 0x8108}},
    {29116, 2, {0x2000, 0x27FC, 0x2444, 0x27FC, 0xAC44, 0xB7FC, 0xA110, 0xA110, 0x27FC, 0x2110, 0x2110, 0x2FFE, 0x5000, 0x4910, 0x4208, 0x8404}},
    {29117, 2, {0x1020, 0x11FC, 0x1024, 0x17FE, 0x5824, 0x51FC, 0x5020, 0x92AA, 0x13AE, 0x1222, 0x13FE, 0x2A22, 0x27AE, 0x42AA, 0x42AA, 0x8422}},
    {29118, 2, {0x1110, 0x1094, 0x13D2, 0x1412, 0x5A50, 0x5190, 0x57FE, 0x9010, 0x13D2, 0x1252, 0x1254, 0x2BD4, 0x264A, 0x424A, 0x43D6, 0x8022}},
    {29119, 2, {0x1088, 0x1088, 0x11FC, 0x1088, 0x5488, 0x5BFE, 0x5020, 0x91FC, 0x1124, 0x11FC, 0x1124, 0x29FC, 0x2400, 0x4488, 0x4104, 0x8202}},
    {29120, 2, {0x1000, 0x13DE, 0x1252, 0x1252, 0x57DE, 0x5800, 0x51FC, 0x9124, 0x11FC, 0x1124, 0x11FC, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {29121, 2, {0x1088, 0x13FE, 0x1088, 0x1000, 0x55FC, 0x58A8, 0x50A8, 0x93FE, 0x10A8, 0x10A8, 0x11FC, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {29122, 2, {0x1000, 0x13FE, 0x1050, 0x13FE, 0x5652, 0x5BFE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x29FC, 0x2420, 0x47FE, 0x4020, 0x8020}},
    {29123, 2, {0x1110, 0x1114, 0x11D2, 0x1650, 0x5A7E, 0x5550, 0x5090, 0x92A8, 0x1128, 0x1144, 0x1244, 0x2C82, 0x2408, 0x42A4, 0x4252, 0x8452}},
    {29124, 2, {0x1010, 0x1052, 0x3E54, 0x2290, 0x4428, 0x8044, 0x1082, 0x6610, 0x4252, 0x4254, 0x6698, 0x4210, 0x4228, 0x7E28, 0x4244, 0x82}},
    {29125, 2, {0x810, 0x852, 0x7E54, 0x890, 0x828, 0xFF44, 0x2482, 0x4210, 0x8952, 0x854, 0x7E98, 0x810, 0x828, 0xF28, 0xF844, 0x4082}},
    {29126, 2, {0x101C, 0x11E0, 0x1040, 0x17FE, 0x5888, 0x5104, 0x52FA, 0x9488, 0x10F8, 0x1000, 0x13FE, 0x2A02, 0x26FA, 0x428A, 0x42FA, 0x8206}},
    {29127, 2, {0x1088, 0x1050, 0x13FE, 0x1050, 0x55FC, 0x5954, 0x518C, 0x9174, 0x1104, 0x11FC, 0x1008, 0x2BFE, 0x2508, 0x4488, 0x40A8, 0x8010}},
    {29128, 2, {0x2020, 0x27A4, 0x20A8, 0x2A92, 0xB114, 0xA208, 0xA5F4, 0xA802, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x5208, 0x4910, 0x4FFE, 0x8000}},
    {29129, 2, {0x2208, 0x2108, 0x27C8, 0x2010, 0xAF9E, 0xB494, 0xA7A4, 0xA014, 0x2794, 0x2094, 0x2114, 0x21C8, 0x5708, 0x4914, 0x4524, 0x8242}},
    {29130, 2, {0x1108, 0x2110, 0x2E0, 0xC18, 0x7004, 0x10, 0x5252, 0x9494, 0x2828, 0xC4C4, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {29131, 2, {0x1090, 0x1088, 0x11FE, 0x1510, 0x5B10, 0x55FE, 0x5110, 0x9110, 0x11FE, 0x1110, 0x1110, 0x29FE, 0x2500, 0x42A4, 0x4252, 0x8452}},
    {29132, 2, {0x1020, 0x11FC, 0x1020, 0x1488, 0x5BFE, 0x5088, 0x51FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x2904, 0x25FC, 0x4488, 0x4104, 0x8202}},
    {29133, 2, {0x2240, 0x2242, 0x225C, 0x27F0, 0xAA50, 0xB3D0, 0xA25E, 0xA3D4, 0x2254, 0x2254, 0x27F4, 0x2014, 0x5294, 0x4C64, 0x4824, 0x8044}},
    {29134, 2, {0x2040, 0x2040, 0x27FC, 0x20A0, 0xAD14, 0xB208, 0xA7FC, 0xAA0A, 0x23F8, 0x2208, 0x23F8, 0x2040, 0x5248, 0x4C44, 0x4944, 0x8080}},
    {29135, 2, {0x1000, 0x13FC, 0x1048, 0x1430, 0x5BFE, 0x5052, 0x5094, 0x9350, 0x1020, 0x13FE, 0x1252, 0x2A8A, 0x2776, 0x4252, 0x4272, 0x8206}},
    {29136, 2, {0x2040, 0x2248, 0x2150, 0x27FC, 0xA950, 0xB248, 0xA444, 0xA210, 0x2390, 0x24BC, 0x2A90, 0x2150, 0x517E, 0x4A10, 0x4210, 0x8410}},
    {29137, 2, {0x1020, 0x11FC, 0x1088, 0x1450, 0x5BFE, 0x5000, 0x51FC, 0x9124, 0x11FC, 0x1124, 0x11FC, 0x2820, 0x25FC, 0x4420, 0x43FE, 0x8000}},
    {29138, 2, {0x1020, 0x1020, 0x11FC, 0x1420, 0x5BFE, 0x5108, 0x539C, 0x9108, 0x1188, 0x163E, 0x1000, 0x2BFE, 0x2490, 0x4490, 0x4112, 0x820E}},
    {29139, 2, {0x2448, 0x2288, 0xF93E, 0x2288, 0x6458, 0x729C, 0xA92A, 0x2288, 0x2448, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {29140, 2, {0x103C, 0x13E0, 0x1124, 0x10A8, 0x55FE, 0x58A8, 0x5124, 0x9202, 0x11FC, 0x1124, 0x1124, 0x29FC, 0x2524, 0x4524, 0x41FC, 0x8104}},
    {29141, 2, {0x440, 0xFFFE, 0x440, 0x440, 0x17D0, 0x1012, 0xF01C, 0x17D0, 0x1450, 0x3452, 0xD7D2, 0x100E, 0x0, 0x4888, 0x4444, 0x8444}},
    {29142, 2, {0x1000, 0x13FC, 0x1004, 0x13FC, 0x5404, 0x5BFC, 0x5000, 0x93BC, 0x1124, 0x17BC, 0x1008, 0x2BFE, 0x2508, 0x4488, 0x40A8, 0x8010}},
    {29143, 2, {0x279E, 0x2492, 0x279E, 0x2492, 0xAF9E, 0xB402, 0xA4F2, 0xA492, 0x24F2, 0x2492, 0x24F2, 0x2492, 0x5492, 0x4D32, 0x4402, 0x8406}},
    {29144, 2, {0x1000, 0x13DE, 0x1252, 0x17DE, 0x5A52, 0x53DE, 0x5242, 0x9222, 0x12FA, 0x1212, 0x1252, 0x2A22, 0x2652, 0x4282, 0x420A, 0x8204}},
    {29145, 2, {0x23F8, 0x1208, 0x93F8, 0x4208, 0x43F8, 0x1000, 0x2FFE, 0xE200, 0x27FC, 0x2924, 0x2254, 0x588, 0x1110, 0x22A0, 0xC60, 0x701C}},
    {29146, 2, {0x1010, 0x5252, 0x5454, 0x9090, 0x2828, 0x4444, 0x8282, 0x1010, 0x1052, 0x5254, 0x5498, 0x9010, 0x2828, 0x2428, 0x4244, 0x8082}},
    {29147, 2, {0x0, 0xFEFE, 0x1290, 0x7EFC, 0x4284, 0x7EFC, 0x1290, 0xFEFE, 0x100, 0x1110, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {29148, 2, {0x2000, 0x279E, 0x2492, 0x279E, 0xAC92, 0xB79E, 0xA442, 0xA422, 0x240A, 0x2546, 0x2542, 0x2552, 0x5632, 0x4C02, 0x440A, 0x8404}},
    {29149, 2, {0x1000, 0x11FC, 0x1104, 0x15FC, 0x5904, 0x51FC, 0x5020, 0x93FE, 0x1000, 0x11FC, 0x1104, 0x29FC, 0x2420, 0x4524, 0x4222, 0x8060}},
    {29150, 2, {0x1420, 0x1220, 0x3F3E, 0x2442, 0x6494, 0xBF10, 0x2410, 0x3F10, 0x2428, 0x2428, 0x3F44, 0x2082, 0x0, 0x4888, 0x4444, 0x8444}},
    {29151, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {29152, 2, {0x2080, 0x2100, 0x27FC, 0x2444, 0xAD54, 0xB444, 0xA7FC, 0xA4E4, 0x2554, 0x2404, 0x2040, 0x27FE, 0x50A0, 0x4910, 0x4208, 0x8C06}},
    {29153, 2, {0x1000, 0x13FE, 0x1252, 0x13FE, 0x5420, 0x59FC, 0x5020, 0x93FE, 0x1088, 0x1050, 0x11FC, 0x2820, 0x27FE, 0x4420, 0x4020, 0x8020}},
    {29154, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4102, 0x9114, 0x1110, 0x22A0, 0x440, 0x820, 0x3018, 0xC006}},
    {29155, 2, {0x1020, 0x13FE, 0x1000, 0x11FC, 0x5504, 0x5974, 0x5154, 0x9174, 0x1104, 0x11FC, 0x1008, 0x28F0, 0x2420, 0x47FE, 0x40A8, 0x8326}},
    {29156, 2, {0x1088, 0x1088, 0x13FE, 0x1488, 0x59FC, 0x5124, 0x51FC, 0x9124, 0x11FC, 0x1020, 0x13FE, 0x2A22, 0x262A, 0x42FA, 0x420A, 0x8206}},
    {29157, 2, {0x10F8, 0x1088, 0x1088, 0x14F8, 0x5800, 0x51DC, 0x5154, 0x9154, 0x11DC, 0x1020, 0x13FE, 0x2870, 0x24A8, 0x4124, 0x4222, 0x8020}},
    {29158, 2, {0x2080, 0x20DC, 0x2284, 0x2BD4, 0xB448, 0xAA94, 0xA324, 0xAC40, 0x2248, 0x2150, 0x27FC, 0x20E0, 0x5150, 0x4A48, 0x4C46, 0x8040}},
    {29159, 2, {0x1044, 0x1224, 0x1128, 0x147E, 0x5810, 0x5028, 0x534A, 0x911C, 0x1128, 0x114C, 0x111A, 0x292A, 0x2548, 0x4110, 0x42FE, 0x8400}},
    {29160, 2, {0x2208, 0x2110, 0x2FFE, 0x2840, 0xB7FC, 0xA040, 0xAFFE, 0xA054, 0x2792, 0x2110, 0x2FFE, 0x2110, 0x51D4, 0x4F0A, 0x4916, 0x8322}},
    {29161, 2, {0x2108, 0x2208, 0x27C8, 0x2450, 0xAFDE, 0xB464, 0xA7D4, 0xA214, 0x2114, 0x27D4, 0x2214, 0x23C8, 0x5248, 0x4C54, 0x4554, 0x88A2}},
    {29162, 2, {0x1088, 0x13FE, 0x1088, 0x1500, 0x5BFC, 0x5554, 0x50D4, 0x9324, 0x1054, 0x1094, 0x1348, 0x2820, 0x24A4, 0x428A, 0x428A, 0x8478}},
    {29163, 2, {0x1020, 0x13FE, 0x1288, 0x1250, 0x57FE, 0x5A50, 0x52FC, 0x9254, 0x13FE, 0x1254, 0x12FC, 0x2A50, 0x26D8, 0x4554, 0x4652, 0x8850}},
    {29164, 2, {0x2100, 0x26DC, 0x2454, 0x2454, 0xAED4, 0xB454, 0xA466, 0xA7C0, 0x201C, 0x2014, 0x27D4, 0x2114, 0x5108, 0x49C8, 0x4714, 0x8222}},
    {29165, 2, {0x2000, 0x23F8, 0x22A8, 0x2AA8, 0xB3F8, 0xA100, 0xA3FC, 0xA484, 0x23E4, 0x22A4, 0x22A4, 0x23E4, 0x5094, 0x4FF4, 0x4814, 0x8008}},
    {29166, 2, {0x2108, 0x27C8, 0x2008, 0x6B9A, 0xB02C, 0x2388, 0x2008, 0x5394, 0x4A92, 0x8BA2, 0x0, 0x3FF0, 0x820, 0x4C0, 0x780, 0xF87E}},
    {29167, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0x5622, 0x59AC, 0x5020, 0x91AC, 0x1050, 0x10C8, 0x1326, 0x28F8, 0x2408, 0x44D0, 0x4020, 0x8010}},
    {29168, 2, {0x203C, 0x23C0, 0x2044, 0x2224, 0xA908, 0xB7FE, 0xA442, 0xA2A4, 0x228A, 0x247A, 0x2080, 0x20F8, 0x5108, 0x4A90, 0x4060, 0x879E}},
    {29169, 2, {0x1040, 0x1020, 0x11FC, 0x1488, 0x5850, 0x53FE, 0x5000, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x29FC, 0x2420, 0x4294, 0x428A, 0x847A}},
    {29170, 2, {0x1040, 0x13FC, 0x1044, 0x1094, 0x5508, 0x5BDE, 0x514A, 0x92D6, 0x11FC, 0x1104, 0x11FC, 0x2904, 0x25FC, 0x4504, 0x41FC, 0x8088}},
    {29171, 2, {0x2000, 0x27BE, 0x2492, 0x2492, 0xACAA, 0xB7A4, 0xA4C0, 0xA4BE, 0x24A2, 0x24A2, 0x27BE, 0x24A2, 0x5000, 0x4AA4, 0x4252, 0x8452}},
    {29172, 2, {0x2040, 0x20A0, 0x2318, 0x2DF6, 0xB000, 0xA7FC, 0xA554, 0xA4E4, 0x27FC, 0x2000, 0x23F8, 0x2208, 0x53F8, 0x4A08, 0x4BF8, 0x8208}},
    {29173, 2, {0x1010, 0x1410, 0x127C, 0x1410, 0x58FE, 0x5044, 0x5628, 0x92FE, 0x1210, 0x127C, 0x1210, 0x2AFE, 0x2610, 0x4210, 0x45FE, 0x8800}},
    {29174, 2, {0x1090, 0x13FC, 0x1294, 0x17FC, 0x5A94, 0x53FC, 0x5000, 0x93FC, 0x1200, 0x12F8, 0x1200, 0x2BFE, 0x2520, 0x4514, 0x4548, 0x8986}},
    {29175, 2, {0x1020, 0x13FE, 0x1000, 0x15FC, 0x5904, 0x5174, 0x5154, 0x9174, 0x1104, 0x11FC, 0x1000, 0x29FC, 0x2400, 0x43FE, 0x4124, 0x8262}},
    {29176, 2, {0x1000, 0x11FC, 0x1020, 0x13FE, 0x5622, 0x59AC, 0x5020, 0x91AC, 0x1000, 0x13FE, 0x1020, 0x29FC, 0x2554, 0x4554, 0x4154, 0x810C}},
    {29177, 2, {0xFEFE, 0x2020, 0x5050, 0x9A9A, 0x2C2C, 0x4848, 0x9C9C, 0x2A2A, 0xC8C8, 0x1818, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {29178, 2, {0x1020, 0x13FE, 0x1000, 0x14F8, 0x5888, 0x50F8, 0x5000, 0x93FE, 0x1202, 0x1272, 0x1252, 0x2A76, 0x2420, 0x43FE, 0x40A8, 0x8326}},
    {29179, 2, {0x101C, 0x11E0, 0x1020, 0x13FE, 0x5420, 0x59FC, 0x51AC, 0x9174, 0x11FC, 0x1020, 0x11FC, 0x2820, 0x27FE, 0x4400, 0x4154, 0x822A}},
    {29180, 2, {0x2040, 0x23F8, 0x2048, 0x27FE, 0xA848, 0xB3F8, 0xA040, 0xA7FE, 0x2000, 0x22A4, 0x2452, 0x23F8, 0x52A8, 0x4AA8, 0x4FFE, 0x8000}},
    {29181, 2, {0x2040, 0x27FC, 0x2040, 0x23FC, 0xA800, 0xB7FE, 0xA002, 0xA3F8, 0x2040, 0x27FE, 0x2000, 0x27FE, 0x5008, 0x4FFE, 0x44A8, 0x8798}},
    {29182, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x0, 0xFFFE, 0x2, 0x3FF0, 0x100, 0xFFFE, 0x10, 0x7DFC, 0x4490, 0x7C30, 0x2488, 0x4244}},
    {29183, 2, {0x1000, 0x13DE, 0x1042, 0x1252, 0x554A, 0x5A52, 0x5090, 0x9108, 0x13FE, 0x1510, 0x11FE, 0x2910, 0x25FE, 0x4510, 0x41FE, 0x8100}},
    {29184, 2, {0x2110, 0x2110, 0x2110, 0x27BC, 0xA910, 0xB110, 0xAFFE, 0xA2A8, 0x22A8, 0x26EC, 0x2AAA, 0x22AA, 0x52A8, 0x4DA8, 0x4448, 0x8898}},
    {29185, 2, {0x1008, 0x13E8, 0x1288, 0x17EE, 0x5A28, 0x53F4, 0x5282, 0x93E2, 0x1000, 0x11FC, 0x1154, 0x2954, 0x2554, 0x4154, 0x47FE, 0x8000}},
    {29186, 2, {0x2FEC, 0x4824, 0x4BA4, 0x682C, 0x4BA4, 0x6AAC, 0x4BA4, 0x4824, 0xFFFE, 0x8102, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {29187, 2, {0x1088, 0x12AA, 0x12DC, 0x1488, 0x5954, 0x5222, 0x5000, 0x93FE, 0x1222, 0x1020, 0x13FE, 0x2870, 0x24A8, 0x4124, 0x4622, 0x8020}},
    {29188, 2, {0x1000, 0x13FE, 0x1248, 0x17FE, 0x5A48, 0x52FC, 0x52A4, 0x92FC, 0x12A4, 0x12FC, 0x1220, 0x2BFE, 0x272A, 0x457A, 0x450A, 0x8906}},
    {29189, 2, {0x11FC, 0x1124, 0x11AC, 0x1524, 0x59FC, 0x5020, 0x51FC, 0x9020, 0x13FE, 0x1154, 0x122A, 0x2820, 0x25FC, 0x4020, 0x43FE, 0x8000}},
    {29190, 2, {0x11FC, 0x1104, 0x11FC, 0x1504, 0x59FC, 0x5088, 0x53FE, 0x9088, 0x17FE, 0x1088, 0x1124, 0x2AAA, 0x2470, 0x40A8, 0x4124, 0x8060}},
    {29191, 2, {0x820, 0xFFFE, 0x820, 0x1040, 0x7C40, 0x11F8, 0xFE48, 0x44C8, 0x9248, 0x7CAA, 0x10AA, 0x1D06, 0xE202, 0x4888, 0x4444, 0x8444}},
    {29192, 2, {0x2020, 0x203E, 0x2020, 0x27FE, 0xAC22, 0xB5F8, 0xA422, 0xA5FE, 0x2524, 0x25FC, 0x2524, 0x25FC, 0x5840, 0x4AA4, 0x528A, 0x847A}},
    {29193, 2, {0x1124, 0x1248, 0x1124, 0x1400, 0x5BFC, 0x5294, 0x5264, 0x9294, 0x13FC, 0x1248, 0x1368, 0x2A48, 0x276A, 0x424A, 0x4246, 0x8362}},
    {29194, 2, {0x2080, 0x2040, 0x27FC, 0x24A0, 0xAFFC, 0xB4A4, 0xA7FC, 0xA400, 0x2524, 0x25A8, 0x2530, 0x25A4, 0x551C, 0x4800, 0x4AA4, 0x9452}},
    {29195, 2, {0x21C8, 0x2708, 0x2108, 0x27C8, 0xA93E, 0xB7CA, 0xA54A, 0xA7CA, 0x254A, 0x27CA, 0x210A, 0x27CA, 0x510A, 0x4FEA, 0x4AAA, 0x9156}},
    {29196, 2, {0x2040, 0x2020, 0x27FE, 0x2488, 0xADFC, 0xB488, 0xA7FE, 0xA420, 0x25FC, 0x2524, 0x25FC, 0x2524, 0x55FC, 0x4C00, 0x4488, 0x8904}},
    {29197, 2, {0x2444, 0x24E4, 0x28A8, 0x2AAA, 0xAEEE, 0xB4A4, 0xAAAA, 0xAEEE, 0x2242, 0x2040, 0x2FFE, 0x20E0, 0x5150, 0x4A48, 0x4C46, 0x8040}},
    {29198, 2, {0x1040, 0x1020, 0x13FE, 0x1222, 0x55FC, 0x5850, 0x528A, 0x91FC, 0x128A, 0x10F8, 0x1088, 0x28F8, 0x2420, 0x4524, 0x42A2, 0x8040}},
    {29199, 2, {0x2000, 0x27FE, 0x2444, 0x2598, 0xAC88, 0xB7DE, 0xA488, 0xA5DC, 0x26AA, 0x2488, 0x2420, 0x2520, 0x553C, 0x4920, 0x4920, 0x97FE}},
    {29200, 2, {0x1040, 0x107C, 0x1040, 0x13FC, 0x5644, 0x5BF0, 0x5244, 0x92FC, 0x12A8, 0x12F8, 0x12A8, 0x2AF8, 0x2600, 0x45FC, 0x4554, 0x8BFE}},
    {29201, 2, {0x1088, 0x1088, 0x13FE, 0x14A8, 0x5890, 0x51FE, 0x5120, 0x93FC, 0x1520, 0x11FC, 0x1120, 0x29FE, 0x2500, 0x42A4, 0x4252, 0x8452}},
    {29202, 2, {0x1010, 0x1010, 0x1EFE, 0x2210, 0x54AA, 0xA844, 0x1082, 0x287C, 0xCA44, 0x2A7C, 0x2C44, 0x487C, 0x1410, 0x1254, 0x2292, 0x4030}},
    {29203, 2, {0x13DE, 0x1252, 0x13DE, 0x1652, 0x5BDE, 0x5242, 0x527A, 0x928A, 0x1312, 0x12DA, 0x128A, 0x2ADA, 0x268A, 0x42FA, 0x420A, 0x8204}},
    {29204, 2, {0x2110, 0x27FC, 0x2040, 0x23F8, 0xA840, 0xB7FC, 0xA000, 0xA3D4, 0x2112, 0x27FE, 0x2350, 0x2534, 0x538C, 0x488A, 0x4296, 0x8102}},
    {29205, 2, {0x2108, 0x27C8, 0x2008, 0x6B9A, 0xB02C, 0x2388, 0x2008, 0x5394, 0x4A92, 0x8BA2, 0x100, 0x1110, 0x22A0, 0x440, 0x1830, 0xE00E}},
    {29206, 2, {0x2210, 0x2110, 0x27DE, 0x2010, 0xAC5E, 0xB282, 0xA7DE, 0xA010, 0x27DE, 0x2450, 0x27DE, 0x2450, 0x57DE, 0x4C50, 0x4452, 0x84CE}},
    {29207, 2, {0x11FC, 0x1104, 0x11FC, 0x1504, 0x59FC, 0x5088, 0x53FE, 0x9088, 0x11FC, 0x10A8, 0x13FE, 0x28A8, 0x25FC, 0x4420, 0x43FE, 0x8020}},
    {29208, 2, {0x1080, 0x10DE, 0x1282, 0x13D4, 0x5448, 0x5AB4, 0x514A, 0x96A4, 0x11FA, 0x1288, 0x10F8, 0x2888, 0x24FA, 0x4494, 0x40A8, 0x80C4}},
    {29209, 2, {0x1020, 0x13FE, 0x1000, 0x15DC, 0x5954, 0x51DC, 0x5088, 0x93FE, 0x1088, 0x13FE, 0x1088, 0x2BFE, 0x2494, 0x4188, 0x42A4, 0x80C2}},
    {29210, 2, {0x2040, 0x20A0, 0x2110, 0x2208, 0xADF6, 0xB000, 0xAEEE, 0xAAAA, 0x2EEE, 0x2000, 0x2FFE, 0x2922, 0x5FFE, 0x4922, 0x492A, 0x8804}},
    {29211, 2, {0x13DE, 0x1252, 0x13DE, 0x1652, 0x5BDE, 0x5222, 0x52FA, 0x9222, 0x12FA, 0x12AA, 0x12FA, 0x2AAA, 0x26FA, 0x4272, 0x42AA, 0x8226}},
    {29212, 2, {0x2492, 0x2294, 0x2FFE, 0x2910, 0xB7FC, 0xA040, 0xA3F8, 0xA040, 0x2FFE, 0x2480, 0x27BC, 0x2494, 0x5794, 0x4CC8, 0x4F94, 0x80A2}},
    {29213, 2, {0x101E, 0x13E0, 0x1122, 0x1494, 0x5BFE, 0x5252, 0x53FE, 0x9004, 0x13C4, 0x127E, 0x13C4, 0x2A54, 0x27CC, 0x4284, 0x42D4, 0x8348}},
    {29214, 2, {0x2040, 0x2040, 0x23F8, 0x2248, 0xABF8, 0xB044, 0xA7FC, 0xA004, 0x2208, 0x2FBE, 0x2AAA, 0x2FBE, 0x5208, 0x4A8A, 0x4FBE, 0x8082}},
    {29215, 2, {0x2088, 0x23FE, 0x2088, 0x23DE, 0xAA52, 0xB3DE, 0xA0A0, 0xA090, 0x21FE, 0x2120, 0x23FC, 0x2520, 0x51FC, 0x4920, 0x41FE, 0x8100}},
    {29216, 2, {0x27BC, 0x24A4, 0x27BC, 0x24A4, 0xAFBC, 0xB4A4, 0xA7BC, 0xA120, 0x23FE, 0x2220, 0x27FC, 0x2A20, 0x53FC, 0x4A20, 0x43FE, 0x8200}},
    {29217, 2, {0x1088, 0x13FE, 0x1088, 0x11FC, 0x5524, 0x59FC, 0x5124, 0x93FE, 0x122A, 0x12FA, 0x1226, 0x29FC, 0x2524, 0x45FC, 0x4022, 0x83FE}},
    {29218, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2120, 0x2F3C, 0x2120, 0x2738, 0x2120, 0x2F3C, 0x4120, 0x5244, 0xA122}},
    {29219, 2, {0x2248, 0x2150, 0x27FC, 0x2404, 0xA9F0, 0xB110, 0xA7FC, 0xA554, 0x24E4, 0x27FC, 0x2040, 0x27FC, 0x5040, 0x4FFE, 0x42A4, 0x8452}},
    {29220, 2, {0x2110, 0x27FE, 0x2110, 0x27BC, 0xACA4, 0xB7BC, 0xA4A4, 0xA7BC, 0x2444, 0x27FC, 0x2554, 0x25F4, 0x5444, 0x4CE4, 0x4554, 0x8448}},
    {29221, 2, {0x13FE, 0x1202, 0x13FE, 0x1692, 0x5A54, 0x5292, 0x52FE, 0x92AA, 0x12FE, 0x1322, 0x12FA, 0x2AAA, 0x24FA, 0x4422, 0x49FA, 0x8004}},
    {29222, 2, {0x2790, 0x251E, 0x27A8, 0x2484, 0xAFBE, 0xB52A, 0xA7BE, 0xA000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x53F8, 0x48A0, 0x4124, 0x861C}},
    {29223, 2, {0x27FC, 0x2040, 0x2FFE, 0x2842, 0xAB58, 0xB040, 0xA358, 0xA000, 0x2EEE, 0x2AAA, 0x2EEE, 0x2000, 0x5FFE, 0x4A48, 0x4554, 0x8FFE}},
    {29224, 2, {0x2FEC, 0x4824, 0x6BAC, 0x4824, 0x6BAC, 0x4AA4, 0xFFFE, 0x8822, 0x7EFC, 0x2C68, 0x4AA4, 0xFFFE, 0x1110, 0x2928, 0xC6C6, 0x1830}},
    {29225, 2, {0x2288, 0x22E8, 0x2F5E, 0x22EC, 0xAF5A, 0xBAE8, 0xA208, 0xAFFE, 0x2942, 0x3494, 0x2558, 0x27F2, 0x5204, 0x4BE8, 0x4202, 0x83E4}},
    {29226, 2, {0x10, 0xF8, 0x1F20, 0x1220, 0x1220, 0x1220, 0x1220, 0x1220, 0x1220, 0x1210, 0x1210, 0x2210, 0x2208, 0x4208, 0x4204, 0x8002}},
    {29227, 2, {0x0, 0x8, 0x7C, 0x780, 0x7A10, 0x2110, 0x1120, 0x1000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {29228, 2, {0xE00, 0x7400, 0x55FC, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5500, 0x5500, 0x5502, 0x5502, 0x52FE, 0x5200, 0x5100, 0x90FE, 0x1000}},
    {29229, 2, {0x8, 0x1FC, 0x7E10, 0x2210, 0x1120, 0x3FF8, 0x108, 0x108, 0xFFFE, 0x108, 0x108, 0x3FF8, 0x108, 0x100, 0x500, 0x200}},
    {29230, 2, {0x240, 0x740, 0x3A7C, 0x2A84, 0x2B74, 0x2A54, 0x2A54, 0x2A74, 0x2A48, 0x2942, 0x2942, 0x28BE, 0x2840, 0x4830, 0x480E, 0x8800}},
    {29231, 2, {0x8, 0x1FC, 0x7E10, 0x2210, 0x1120, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0xFFFE, 0x2008, 0x2028, 0x2010}},
    {29232, 2, {0x8, 0x1FC, 0x7E10, 0x2210, 0x1120, 0x3FF8, 0x200, 0x200, 0x7FFC, 0x400, 0xFE0, 0x1420, 0x2240, 0x4180, 0x8660, 0x381C}},
    {29233, 2, {0x8, 0x1FC, 0x7E10, 0x2210, 0x1120, 0x7FFE, 0x4202, 0x8204, 0x7FF8, 0x400, 0x7F0, 0xA10, 0x1120, 0x20C0, 0x4330, 0x1C0E}},
    {29234, 2, {0xFC, 0x7F00, 0x1210, 0x920, 0x1FF0, 0x1010, 0x1010, 0x1FF8, 0x1008, 0x1008, 0x1FFE, 0x2002, 0x2AA2, 0x4952, 0x5154, 0x8008}},
    {29235, 2, {0xF8, 0x3F20, 0x940, 0x1FE0, 0x1020, 0x1FF0, 0x1010, 0x1FFC, 0x2004, 0x2AA4, 0x400C, 0x9FE0, 0x40, 0x180, 0x500, 0x200}},
    {29236, 2, {0x4, 0xFF0E, 0x8D74, 0x8B54, 0xFF54, 0x8954, 0xE954, 0xA954, 0xEB54, 0x8554, 0xE554, 0x8B54, 0x9154, 0xFF92, 0x92, 0x100}},
    {29237, 2, {0xFC, 0x7F00, 0x2208, 0x1110, 0x7FFC, 0x4444, 0x7FFC, 0x8, 0x7E08, 0x42FE, 0x7E08, 0x4248, 0x7E28, 0x4808, 0x7428, 0x4210}},
    {29238, 2, {0x440, 0x820, 0x1010, 0x2008, 0x4024, 0x820, 0x820, 0x440, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {29239, 2, {0x440, 0x820, 0x1850, 0x2488, 0x300, 0xCC0, 0x3030, 0xC00E, 0x1FF0, 0x210, 0x210, 0x210, 0x250, 0x220, 0x200, 0x200}},
    {29240, 2, {0x440, 0x820, 0x1850, 0x2488, 0x300, 0xCC0, 0x3030, 0xC00E, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010, 0x1004, 0x1004, 0xFFC}},
    {29241, 2, {0x440, 0x820, 0x1850, 0x2488, 0x300, 0xCC0, 0x3230, 0xC7CE, 0x1A40, 0x1A0, 0x1E78, 0x188, 0xE90, 0x60, 0x380, 0x3C00}},
    {29242, 2, {0x840, 0x1020, 0x2850, 0x4488, 0x300, 0x1CE0, 0xE01E, 0x0, 0x7F78, 0x2248, 0x3E50, 0x2248, 0x3E44, 0x23D4, 0xFE48, 0x240}},
    {29243, 2, {0x8, 0x1810, 0x660, 0x180, 0x660, 0x1818, 0x6004, 0x20, 0x820, 0x440, 0x280, 0x100, 0x280, 0xC40, 0x3030, 0xC00E}},
    {29244, 2, {0x400, 0x45F8, 0x2908, 0x1108, 0x2908, 0x45F8, 0x8108, 0x508, 0x4508, 0x29F8, 0x1108, 0x2908, 0x4508, 0x8108, 0x7FE, 0x0}},
    {29245, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x2928, 0x1110, 0x2928, 0x100, 0x2928, 0x1110, 0x2AA8, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {29246, 2, {0x0, 0x7FFC, 0x2108, 0x4104, 0xBFFA, 0x2108, 0x3558, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3558, 0x2108, 0x2128, 0x2010}},
    {29247, 2, {0x10, 0x810, 0x810, 0x810, 0xFF0, 0x10, 0x10, 0x10, 0x7FF0, 0x810, 0x810, 0x810, 0x1010, 0x1010, 0x2010, 0x4010}},
    {29248, 2, {0x420, 0x2420, 0x2420, 0x2420, 0x25FE, 0x3C20, 0x470, 0x470, 0xFCA8, 0x24A8, 0x2524, 0x2524, 0x2622, 0x4420, 0x4420, 0x8420}},
    {29249, 2, {0x400, 0x2400, 0x25FE, 0x2408, 0x2408, 0x3DE8, 0x528, 0x528, 0xFD28, 0x2528, 0x25E8, 0x2528, 0x2408, 0x4408, 0x4428, 0x8410}},
    {29250, 2, {0x484, 0x2444, 0x2448, 0x2400, 0x25FE, 0x3C20, 0x420, 0x4FC, 0xFC20, 0x2420, 0x25FE, 0x2420, 0x2420, 0x4420, 0x4420, 0x8420}},
    {29251, 2, {0x448, 0x2548, 0x2548, 0x27FE, 0x2548, 0x3D48, 0x578, 0x500, 0xFDFE, 0x2420, 0x27FE, 0x2470, 0x24A8, 0x4524, 0x4622, 0x8420}},
    {29252, 2, {0x420, 0x2420, 0x2450, 0x2448, 0x24A4, 0x3DFE, 0x684, 0x4FC, 0xFC84, 0x24FC, 0x2480, 0x24FC, 0x2544, 0x4544, 0x467C, 0x8444}},
    {29253, 2, {0x1020, 0x57FE, 0x5420, 0x55FC, 0x5424, 0x77FE, 0x1424, 0x15FC, 0xF420, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x892C}},
    {29254, 2, {0x820, 0x2820, 0x2BFE, 0x2924, 0x2924, 0x3AAA, 0xFFE, 0x800, 0xF9FC, 0x2904, 0x2974, 0x2954, 0x2974, 0x4904, 0x49FC, 0x8904}},
    {29255, 2, {0x40, 0x1040, 0x1040, 0x1040, 0x1040, 0x1FFC, 0x1000, 0x1000, 0x1000, 0x1FE0, 0x1020, 0x1020, 0x1020, 0x2020, 0x2020, 0x4020}},
    {29256, 2, {0x808, 0x481C, 0x49E0, 0x4900, 0x4900, 0x7DFC, 0x4144, 0x4144, 0x7944, 0x4928, 0x4928, 0x4910, 0x4910, 0x4A28, 0x4A44, 0x8C82}},
    {29257, 2, {0x420, 0x2420, 0x2524, 0x24A4, 0x24A8, 0x3E20, 0x21FC, 0x2020, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x2420, 0x4420, 0x4420, 0x8420}},
    {29258, 2, {0x400, 0x25FC, 0x2444, 0x2444, 0x2444, 0x3E44, 0x2094, 0x2088, 0x3D00, 0x24FC, 0x2484, 0x2484, 0x2484, 0x4484, 0x44FC, 0x8484}},
    {29259, 2, {0x448, 0x2444, 0x245E, 0x25E0, 0x2428, 0x3E12, 0x206A, 0x2196, 0x3C48, 0x245E, 0x25E0, 0x2424, 0x2428, 0x4412, 0x446A, 0x8586}},
    {29260, 2, {0x840, 0x4880, 0x49FC, 0x4924, 0x4924, 0x7DFC, 0x4124, 0x4144, 0x79FC, 0x4890, 0x4910, 0x4BFE, 0x4810, 0x4810, 0x4810, 0x8810}},
    {29261, 2, {0x820, 0x4820, 0x48FC, 0x4820, 0x4820, 0x7DFE, 0x4002, 0x4094, 0x7850, 0x4890, 0x4850, 0x49FE, 0x4828, 0x4844, 0x4882, 0x8902}},
    {29262, 2, {0x880, 0x4880, 0x48FC, 0x4954, 0x4AD4, 0x7CB4, 0x412C, 0x4244, 0x7894, 0x4908, 0x4840, 0x48A4, 0x4AAA, 0x4A8A, 0x4C78, 0x8800}},
    {29263, 2, {0x820, 0x4850, 0x4888, 0x4904, 0x4AFA, 0x7C00, 0x41E2, 0x412A, 0x792A, 0x49EA, 0x492A, 0x492A, 0x49EA, 0x4922, 0x492A, 0x8964}},
    {29264, 2, {0x808, 0x483C, 0x49E0, 0x4820, 0x4820, 0x7DFE, 0x4020, 0x40A0, 0x792C, 0x4924, 0x4924, 0x49AC, 0x4924, 0x4924, 0x49FC, 0x8904}},
    {29265, 2, {0x840, 0x4820, 0x49FE, 0x4902, 0x4902, 0x7DFE, 0x4100, 0x4100, 0x79FE, 0x49AA, 0x4AAA, 0x4AFE, 0x4CAA, 0x48AA, 0x48A2, 0x8886}},
    {29266, 2, {0x848, 0x4948, 0x4948, 0x4BFE, 0x4948, 0x7D48, 0x4178, 0x4100, 0x79FE, 0x4820, 0x4BFE, 0x4870, 0x48A8, 0x4924, 0x4E22, 0x8820}},
    {29267, 2, {0x820, 0x4810, 0x49FE, 0x4884, 0x4848, 0x7DFE, 0x4102, 0x4224, 0x7810, 0x49FE, 0x4840, 0x487C, 0x4844, 0x4884, 0x4894, 0x8908}},
    {29268, 2, {0x828, 0x4824, 0x4BFE, 0x4820, 0x49FC, 0x7D24, 0x41FC, 0x4124, 0x79FC, 0x4924, 0x4808, 0x4BFE, 0x4888, 0x4848, 0x4848, 0x8818}},
    {29269, 2, {0x820, 0x4840, 0x49FC, 0x4924, 0x497C, 0x7D8C, 0x4154, 0x4124, 0x7954, 0x49FC, 0x4840, 0x4824, 0x4AA2, 0x4A8A, 0x4C88, 0x8878}},
    {29270, 2, {0x840, 0x4820, 0x4BFE, 0x4A02, 0x4BFE, 0x7E14, 0x43FE, 0x4210, 0x7AFE, 0x4A92, 0x4AFE, 0x4A92, 0x4AFE, 0x4A92, 0x4C92, 0x8896}},
    {29271, 2, {0x810, 0x49FE, 0x4910, 0x497C, 0x4914, 0x7DFE, 0x4114, 0x417C, 0x7910, 0x497C, 0x4954, 0x497C, 0x4954, 0x4A7C, 0x4A54, 0x8C4C}},
    {29272, 2, {0x820, 0x4BFE, 0x4820, 0x49FC, 0x4800, 0x7FFE, 0x4252, 0x43FE, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4888, 0x8904}},
    {29273, 2, {0x0, 0x3FFC, 0x40, 0x40, 0x1040, 0x1040, 0x2040, 0x3FFE, 0x140, 0x240, 0x440, 0x840, 0x1040, 0x6040, 0x140, 0x80}},
    {29274, 2, {0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1040, 0x2040, 0x3FFE, 0x240, 0x440, 0x1940, 0x6080}},
    {29275, 2, {0x100, 0x1100, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {29276, 2, {0x800, 0x800, 0x4800, 0x4800, 0x7E00, 0x4800, 0x8800, 0x800, 0xE00, 0xF800, 0x4800, 0x800, 0x800, 0x800, 0x800, 0x800}},
    {29277, 2, {0x880, 0x880, 0x4880, 0x4884, 0x7E88, 0x4890, 0x88A0, 0x8C0, 0xE80, 0xF880, 0x4880, 0x884, 0x884, 0x884, 0x87C, 0x800}},
    {29278, 2, {0x1040, 0x1040, 0x5040, 0x5040, 0x7DFC, 0x5044, 0x9044, 0x1044, 0x1C44, 0xF084, 0x5084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {29279, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {29280, 2, {0x1020, 0x1020, 0x5020, 0x5120, 0x7D2C, 0x5134, 0x9164, 0x13A4, 0x1D24, 0xF134, 0x5128, 0x1122, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {29281, 2, {0x820, 0x820, 0x4820, 0x4820, 0x7E20, 0x4820, 0x89FC, 0x820, 0xE20, 0xF820, 0x4820, 0x820, 0x820, 0x820, 0xBFE, 0x800}},
    {29282, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8104, 0x1100, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {29283, 2, {0x1000, 0x1000, 0x51FC, 0x5044, 0x7C44, 0x5144, 0x9144, 0x1144, 0x1D44, 0xF244, 0x5044, 0x1084, 0x1084, 0x1104, 0x1228, 0x1410}},
    {29284, 2, {0x1020, 0x1010, 0x5010, 0x5000, 0x7DFE, 0x5080, 0x9080, 0x1080, 0x1C80, 0xF080, 0x5080, 0x1080, 0x1080, 0x1080, 0x10FC, 0x1000}},
    {29285, 2, {0x1040, 0x1020, 0x5020, 0x53FE, 0x7C80, 0x5080, 0x9080, 0x10FC, 0x1C84, 0xF084, 0x5084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {29286, 2, {0x808, 0x83C, 0x49E0, 0x4820, 0x7E20, 0x483C, 0x89E0, 0x820, 0xE20, 0xF83E, 0x4BE0, 0x820, 0x822, 0x822, 0x822, 0x81E}},
    {29287, 2, {0x840, 0x840, 0x4840, 0x4880, 0x7EFE, 0x4908, 0x8A88, 0x888, 0xE88, 0xF850, 0x4850, 0x820, 0x850, 0x888, 0x904, 0xA02}},
    {29288, 2, {0x1080, 0x1040, 0x5040, 0x53FC, 0x7C00, 0x5000, 0x91F0, 0x1110, 0x1D10, 0xF110, 0x5110, 0x1112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {29289, 2, {0x1080, 0x1080, 0x5080, 0x50FC, 0x7D54, 0x5254, 0x9054, 0x1094, 0x1C94, 0xF124, 0x5224, 0x1044, 0x1044, 0x1084, 0x1128, 0x1010}},
    {29290, 2, {0x810, 0x810, 0x4890, 0x4890, 0x7EFE, 0x4890, 0x8910, 0x810, 0xE10, 0xF9FE, 0x4810, 0x810, 0x810, 0x810, 0x810, 0x810}},
    {29291, 2, {0x850, 0x848, 0x4848, 0x4840, 0x7E5E, 0x49E0, 0x8840, 0x844, 0xE44, 0xF848, 0x4830, 0x822, 0x852, 0x88A, 0xB06, 0x802}},
    {29292, 2, {0x1020, 0x1020, 0x502E, 0x51F0, 0x7D20, 0x5120, 0x9120, 0x11FE, 0x1C22, 0xF062, 0x50A2, 0x112A, 0x1624, 0x1020, 0x1020, 0x1020}},
    {29293, 2, {0x1000, 0x1000, 0x53FC, 0x5294, 0x7E94, 0x5294, 0x9294, 0x1294, 0x1E94, 0xF294, 0x529C, 0x1304, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {29294, 2, {0x8A0, 0x1090, 0x30BE, 0x57C0, 0x9040, 0x1024, 0x1014, 0x110C, 0x900, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {29295, 2, {0x1020, 0x1020, 0x5020, 0x5020, 0x7DFE, 0x5020, 0x9020, 0x1020, 0x1DFC, 0xF104, 0x5104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {29296, 2, {0x1020, 0x1020, 0x5020, 0x5020, 0x7DFC, 0x5124, 0x9124, 0x1124, 0x1D24, 0xF1FC, 0x5124, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {29297, 2, {0x1000, 0x1000, 0x53FE, 0x5008, 0x7C08, 0x51E8, 0x9128, 0x1128, 0x1D28, 0xF128, 0x51E8, 0x1128, 0x1008, 0x1008, 0x1028, 0x1010}},
    {29298, 2, {0x1020, 0x1020, 0x5120, 0x5120, 0x7DFC, 0x5120, 0x9220, 0x1020, 0x1C20, 0xF1FC, 0x5020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {29299, 2, {0x1000, 0x10FC, 0x5084, 0x50A4, 0x7C94, 0x5084, 0x9084, 0x13FE, 0x1D04, 0xF144, 0x5124, 0x1104, 0x11FE, 0x1004, 0x1028, 0x1010}},
    {29300, 2, {0x1004, 0x101E, 0x51F0, 0x5110, 0x7D10, 0x5110, 0x9110, 0x11FE, 0x1D10, 0xF110, 0x5110, 0x1108, 0x110A, 0x114A, 0x11A6, 0x1112}},
    {29301, 2, {0x100, 0x100, 0x3FF8, 0x280, 0x440, 0x820, 0x7FFE, 0x4102, 0x8904, 0xFF0, 0x1100, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {29302, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1110, 0x2908, 0xCFF6, 0x1100, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {29303, 2, {0x1020, 0x1020, 0x5050, 0x5088, 0x7D04, 0x5202, 0x91FC, 0x1020, 0x1C20, 0xF020, 0x51FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {29304, 2, {0x1040, 0x1020, 0x53FE, 0x5202, 0x7C04, 0x51F8, 0x9008, 0x1010, 0x1C20, 0xF3FE, 0x5020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {29305, 2, {0x1020, 0x1020, 0x5020, 0x51FC, 0x7C20, 0x5020, 0x93FE, 0x1008, 0x1C08, 0xF1FE, 0x5008, 0x1088, 0x1048, 0x1008, 0x1028, 0x1010}},
    {29306, 2, {0x1000, 0x17FE, 0x5090, 0x5090, 0x7C90, 0x53FC, 0x9294, 0x1294, 0x1E94, 0xF294, 0x529C, 0x1304, 0x1204, 0x1204, 0x13FC, 0x1204}},
    {29307, 2, {0x1090, 0x1088, 0x5080, 0x51FE, 0x7CA0, 0x50A8, 0x90B0, 0x10A4, 0x1CA8, 0xF0B0, 0x5124, 0x112A, 0x1132, 0x1222, 0x12DE, 0x1400}},
    {29308, 2, {0x1000, 0x11FE, 0x5000, 0x5092, 0x7D24, 0x5248, 0x9124, 0x1092, 0x1C00, 0xF1FE, 0x5020, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {29309, 2, {0x100, 0x7FFC, 0x820, 0x1FC0, 0x200, 0x7FFE, 0x4822, 0x9FF4, 0x110, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {29310, 2, {0x1000, 0x13FE, 0x5040, 0x5040, 0x7DFC, 0x5084, 0x9084, 0x13FE, 0x1C00, 0xF000, 0x51FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {29311, 2, {0x1020, 0x1120, 0x5120, 0x51FC, 0x7D20, 0x5220, 0x9020, 0x13FE, 0x1C00, 0xF000, 0x51FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {29312, 2, {0x0, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2888, 0x26B0, 0x2808, 0x2480, 0x27F8, 0x2880, 0x2080, 0x5FFE, 0x4080, 0x8080, 0x80}},
    {29313, 2, {0x308, 0x3C08, 0x448, 0x7F48, 0xC48, 0x1648, 0x2508, 0x4428, 0x510, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {29314, 2, {0xC80, 0x70FC, 0x1124, 0xFD24, 0x1244, 0x3884, 0x5528, 0x9010, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {29315, 2, {0x1040, 0x1020, 0x51FC, 0x5000, 0x7D08, 0x5090, 0x9000, 0x13FE, 0x1C00, 0xF000, 0x51FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {29316, 2, {0x1020, 0x1020, 0x51FC, 0x5050, 0x7C88, 0x5104, 0x93FE, 0x1008, 0x1DE8, 0xF128, 0x5128, 0x11E8, 0x1128, 0x1008, 0x1028, 0x1010}},
    {29317, 2, {0x1000, 0x13FE, 0x5202, 0x528A, 0x7E52, 0x53FE, 0x9222, 0x1222, 0x1EAA, 0xF2AA, 0x52AA, 0x12FA, 0x1202, 0x1202, 0x120A, 0x1204}},
    {29318, 2, {0x1020, 0x1020, 0x53FE, 0x5020, 0x7C20, 0x51FC, 0x9104, 0x11FC, 0x1D04, 0xF1FC, 0x5104, 0x11FC, 0x1104, 0x1104, 0x17FE, 0x1000}},
    {29319, 2, {0x100, 0x900, 0xFF0, 0x1100, 0x100, 0x7FFC, 0x100, 0x1110, 0x5150, 0x7C7C, 0x9090, 0x1E10, 0xF0FE, 0x5010, 0x1010, 0x1010}},
    {29320, 2, {0x1020, 0x1124, 0x50A8, 0x5020, 0x7DFC, 0x5040, 0x93FE, 0x1088, 0x1D04, 0xF2FA, 0x5488, 0x1088, 0x10A8, 0x1092, 0x1082, 0x107E}},
    {29321, 2, {0x1040, 0x1020, 0x53FE, 0x5000, 0x7DFC, 0x5104, 0x91FC, 0x1000, 0x1DFC, 0xF008, 0x5010, 0x13FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {29322, 2, {0x1020, 0x1020, 0x51FC, 0x5020, 0x7C20, 0x53FE, 0x9002, 0x1094, 0x1C50, 0xF110, 0x5090, 0x13FE, 0x1028, 0x1044, 0x1082, 0x1302}},
    {29323, 2, {0x1000, 0x11F8, 0x5108, 0x5108, 0x7DF8, 0x5108, 0x91F8, 0x1108, 0x1DF8, 0xF108, 0x5108, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {29324, 2, {0x1000, 0x13DE, 0x5242, 0x5242, 0x7A42, 0x53DE, 0x9200, 0x123E, 0x1BD2, 0x3212, 0xD214, 0x13D4, 0x1208, 0x1214, 0x1224, 0x1242}},
    {29325, 2, {0x1010, 0x1010, 0x577C, 0x5114, 0x7DFE, 0x5214, 0x927C, 0x1710, 0x1D7C, 0xF510, 0x55FE, 0x1210, 0x1210, 0x1500, 0x18FE, 0x1000}},
    {29326, 2, {0x808, 0x7E08, 0x8FE, 0xFE08, 0x848, 0x7E28, 0x808, 0xF28, 0xF110, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {29327, 2, {0x1080, 0x1040, 0x53FC, 0x5204, 0x7E04, 0x53FC, 0x9200, 0x1200, 0x1FFC, 0xF354, 0x5354, 0x15FC, 0x1554, 0x1554, 0x1944, 0x110C}},
    {29328, 2, {0x10C4, 0x1704, 0x5124, 0x5114, 0x7914, 0x5784, 0x9124, 0x1114, 0x1B94, 0x3346, 0xD53C, 0x1504, 0x1104, 0x1104, 0x1104, 0x1104}},
    {29329, 2, {0x1000, 0x11FC, 0x5104, 0x51FC, 0x7D04, 0x51FC, 0x9104, 0x11FC, 0x1C28, 0xF024, 0x53FE, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {29330, 2, {0x1040, 0x1020, 0x53FE, 0x5000, 0x7DFC, 0x5104, 0x91FC, 0x1000, 0x1FFE, 0xF202, 0x52FA, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {29331, 2, {0x1080, 0x1080, 0x51FC, 0x5244, 0x7954, 0x51F4, 0x9084, 0x1128, 0x1A90, 0x31FC, 0xD244, 0x1554, 0x11F4, 0x1084, 0x1128, 0x1210}},
    {29332, 2, {0x1100, 0x1100, 0x51FC, 0x5200, 0x7DF8, 0x5000, 0x93F8, 0x1088, 0x1AA8, 0x31C8, 0xD7F8, 0x108A, 0x11CA, 0x16AA, 0x1086, 0x1082}},
    {29333, 2, {0x1088, 0x1088, 0x53FE, 0x5088, 0x7888, 0x57FE, 0x9080, 0x11FC, 0x1B24, 0x35FC, 0xD124, 0x11FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {29334, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4102, 0x8904, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {29335, 2, {0x1040, 0x1020, 0x53FE, 0x5202, 0x7C24, 0x51FC, 0x9020, 0x11FC, 0x1C20, 0xF3FE, 0x5020, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {29336, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2638, 0x2B54, 0x3292, 0x2040, 0x2240, 0x23FC, 0x2440, 0x2040, 0x5FFE, 0x4040, 0x8040}},
    {29337, 2, {0x1040, 0x1088, 0x51FC, 0x5108, 0x7A52, 0x53FE, 0x9050, 0x1188, 0x1E26, 0x30C0, 0xD310, 0x1064, 0x1388, 0x1030, 0x10C0, 0x1700}},
    {29338, 2, {0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5FA8, 0x4408, 0x9528, 0x2C90, 0x1100, 0x1FF8, 0x2100, 0xFFFE, 0x100, 0x100}},
    {29339, 2, {0x820, 0x3E7E, 0x8A4, 0x7F28, 0x1C10, 0x2A28, 0x4944, 0x3FFE, 0x2480, 0x2480, 0x27F8, 0x2880, 0x2080, 0x5FFC, 0x4080, 0x8080}},
    {29340, 2, {0x1108, 0x1088, 0x53C8, 0x5010, 0x7BDE, 0x5264, 0x93D4, 0x1014, 0x1BD4, 0x3054, 0xD094, 0x10E8, 0x1388, 0x1094, 0x1294, 0x1122}},
    {29341, 2, {0x1020, 0x11FC, 0x5088, 0x5050, 0x7DFE, 0x5000, 0x91FC, 0x1124, 0x1DFC, 0xF124, 0x51FC, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {29342, 2, {0x1018, 0x11E0, 0x5040, 0x53FE, 0x7C88, 0x5174, 0x9252, 0x1070, 0x1C00, 0xF1FC, 0x5104, 0x1174, 0x1154, 0x1174, 0x1104, 0x110C}},
    {29343, 2, {0x7CF8, 0x488, 0x4F8, 0x7C20, 0x41FC, 0x7D24, 0x5FC, 0x428, 0x2BFC, 0x1004, 0x1100, 0x1FF8, 0x2100, 0xFFFE, 0x100, 0x100}},
    {29344, 2, {0x1104, 0x1088, 0x53FE, 0x5020, 0x7DFC, 0x5020, 0x93FE, 0x1054, 0x1D92, 0xF090, 0x53FE, 0x1090, 0x10D4, 0x138A, 0x1096, 0x11A2}},
    {29345, 2, {0x1000, 0x13FE, 0x5248, 0x53FE, 0x7E48, 0x52FC, 0x92A4, 0x12FC, 0x1EA4, 0xF2FC, 0x5220, 0x13FE, 0x132A, 0x157A, 0x150A, 0x1906}},
    {29346, 2, {0x1020, 0x13FE, 0x5020, 0x51FC, 0x7C00, 0x53FE, 0x9252, 0x13FE, 0x1D04, 0xF1FC, 0x5104, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {29347, 2, {0x1124, 0x1248, 0x5124, 0x5000, 0x7BFC, 0x5294, 0x9264, 0x1294, 0x1BFC, 0x3248, 0xD368, 0x1248, 0x136A, 0x124A, 0x1246, 0x1362}},
    {29348, 2, {0x1000, 0x13FC, 0x5294, 0x5294, 0x7BFC, 0x5110, 0x9252, 0x13DC, 0x1812, 0x33CE, 0xD240, 0x13D2, 0x125C, 0x13D0, 0x1252, 0x12CE}},
    {29349, 2, {0x1040, 0x1020, 0x53FE, 0x5250, 0x7BFE, 0x5252, 0x93FE, 0x1200, 0x1A94, 0x32D8, 0xD290, 0x12D2, 0x128E, 0x1400, 0x1554, 0x1A2A}},
    {29350, 2, {0x13F8, 0x1208, 0x53F8, 0x5208, 0x7BF8, 0x5110, 0x97FC, 0x1110, 0x1FFE, 0x3110, 0xD248, 0x1554, 0x10E0, 0x1150, 0x1248, 0x10C0}},
    {29351, 2, {0x1110, 0x17FC, 0x5040, 0x53F8, 0x7840, 0x57FC, 0x9000, 0x13D4, 0x1912, 0x37FE, 0xD350, 0x1534, 0x138C, 0x108A, 0x1296, 0x1102}},
    {29352, 2, {0x810, 0x2448, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x2448, 0x3F7E, 0x2040, 0x1100, 0x1FF8, 0x2100, 0xFFFE, 0x100, 0x100}},
    {29353, 2, {0x7C40, 0x11FC, 0xFD24, 0x39FC, 0x5524, 0xFDFC, 0x2468, 0x18A2, 0x651E, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100}},
    {29354, 2, {0x2110, 0x27FE, 0xA040, 0xA2EE, 0xFAAA, 0xABEA, 0xAAAE, 0x2AE8, 0x3BAA, 0xECE6, 0x2288, 0x25F4, 0x2290, 0x2060, 0x2198, 0x2606}},
    {29355, 2, {0x5118, 0x4B94, 0x7C3E, 0xD3A8, 0x7C7E, 0x53A8, 0x7C3E, 0x53A8, 0x7EBE, 0x43A0, 0x1100, 0x1FF8, 0x2100, 0xFFFE, 0x100, 0x100}},
    {29356, 2, {0x100, 0x120, 0x110, 0x110, 0x100, 0xFFFE, 0x100, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {29357, 2, {0x0, 0x4400, 0x2800, 0x1000, 0x2800, 0x4800, 0x8800, 0x800, 0x1800, 0x2800, 0x4800, 0x8800, 0x800, 0x800, 0x5000, 0x2000}},
    {29358, 2, {0x110, 0x108, 0x108, 0x100, 0x7FFE, 0x200, 0x400, 0x7F8, 0xA08, 0xA08, 0x1110, 0x20A0, 0x4040, 0x80A0, 0x318, 0xC06}},
    {29359, 2, {0x0, 0x4400, 0x29F8, 0x1108, 0x2908, 0x4908, 0x8908, 0x908, 0x1928, 0x2910, 0x4900, 0x8902, 0x902, 0x902, 0x50FE, 0x2000}},
    {29360, 2, {0x80, 0x4480, 0x2880, 0x1080, 0x2BF0, 0x4890, 0x8890, 0x890, 0x1890, 0x2890, 0x4890, 0x8892, 0x912, 0x912, 0x520E, 0x2400}},
    {29361, 2, {0x40, 0x4440, 0x2840, 0x1040, 0x29F8, 0x4848, 0x8848, 0x848, 0x1948, 0x28C8, 0x4848, 0x88A8, 0x8AA, 0x90A, 0x5206, 0x2402}},
    {29362, 2, {0x10, 0x4410, 0x2810, 0x1010, 0x2BFE, 0x4830, 0x8830, 0x850, 0x1850, 0x2890, 0x4890, 0x8910, 0xA10, 0x810, 0x5050, 0x2020}},
    {29363, 2, {0x40, 0x4440, 0x2880, 0x10FC, 0x2904, 0x4A04, 0x8804, 0x904, 0x1884, 0x2844, 0x4844, 0x8804, 0x804, 0x804, 0x5028, 0x2010}},
    {29364, 2, {0x0, 0x45FC, 0x2820, 0x1020, 0x2820, 0x4820, 0x8820, 0xBFE, 0x1820, 0x2820, 0x4820, 0x8820, 0x820, 0x820, 0x5020, 0x2020}},
    {29365, 2, {0x80, 0x4480, 0x2900, 0x11FE, 0x2A00, 0x4C00, 0x89F8, 0x808, 0x1810, 0x2860, 0x4880, 0x8900, 0xA02, 0xA02, 0x51FE, 0x2000}},
    {29366, 2, {0x840, 0x848, 0x844, 0x4844, 0x2840, 0x2FFE, 0x840, 0x840, 0x1840, 0x28A0, 0xC8A0, 0x890, 0x910, 0x908, 0xA04, 0xC02}},
    {29367, 2, {0x20, 0x4410, 0x2810, 0x11FE, 0x2900, 0x4900, 0x8900, 0x900, 0x1900, 0x2900, 0x4900, 0x8900, 0x900, 0xA00, 0x5200, 0x2400}},
    {29368, 2, {0x0, 0x47F0, 0x2810, 0x1010, 0x2910, 0x4910, 0x8910, 0x9FC, 0x1804, 0x2804, 0x4804, 0x8BF4, 0x804, 0x804, 0x5028, 0x2010}},
    {29369, 2, {0x90, 0x4488, 0x2888, 0x1080, 0x2BFE, 0x48A0, 0x88A0, 0x8A0, 0x18A0, 0x28A0, 0x4920, 0x8922, 0x922, 0xA22, 0x521E, 0x2400}},
    {29370, 2, {0x80, 0x4440, 0x2840, 0x13FC, 0x2800, 0x4800, 0x89F0, 0x910, 0x1910, 0x2910, 0x4910, 0x8912, 0x912, 0xA12, 0x520E, 0x2400}},
    {29371, 2, {0x20, 0x4420, 0x2820, 0x13FE, 0x2820, 0x4820, 0x89FC, 0x924, 0x1924, 0x2924, 0x4924, 0x8934, 0x928, 0x820, 0x5020, 0x2020}},
    {29372, 2, {0x10, 0x8FD0, 0x5050, 0x2090, 0x5110, 0x9110, 0x1150, 0x1190, 0x3110, 0x5710, 0x9110, 0x1112, 0x1112, 0x1112, 0xA50E, 0x4200}},
    {29373, 2, {0x0, 0x45FC, 0x2810, 0x1110, 0x2910, 0x4910, 0x8A10, 0xBFE, 0x1830, 0x2850, 0x4850, 0x8890, 0x910, 0xA10, 0x5050, 0x2020}},
    {29374, 2, {0x40, 0x8850, 0x5048, 0x2048, 0x5040, 0x97FE, 0x1040, 0x1040, 0x30A0, 0x50A0, 0x90A0, 0x1110, 0x1110, 0x1208, 0xA404, 0x4802}},
    {29375, 2, {0x40, 0x4420, 0x2820, 0x1000, 0x2BFE, 0x4840, 0x8840, 0x860, 0x1850, 0x2848, 0x4844, 0x8844, 0x840, 0x840, 0x5040, 0x2040}},
    {29376, 2, {0x420, 0x2428, 0x2424, 0x2424, 0x2420, 0x3DFE, 0x420, 0x420, 0xFC20, 0x2450, 0x2450, 0x2450, 0x2488, 0x4488, 0x4504, 0x8602}},
    {29377, 2, {0x40, 0x4440, 0x2880, 0x1088, 0x2904, 0x4BFE, 0x8802, 0x890, 0x1890, 0x2890, 0x4890, 0x8890, 0x912, 0x912, 0x520E, 0x2400}},
    {29378, 2, {0x0, 0x4400, 0x29FC, 0x1020, 0x2820, 0x4820, 0x8820, 0x820, 0x19FC, 0x2820, 0x4820, 0x8820, 0x820, 0x820, 0x53FE, 0x2000}},
    {29379, 2, {0x0, 0x45FC, 0x2844, 0x1044, 0x2844, 0x4844, 0x8844, 0xBFC, 0x1884, 0x2884, 0x4884, 0x8884, 0x884, 0x884, 0x53FE, 0x2000}},
    {29380, 2, {0x40, 0x4440, 0x2840, 0x1040, 0x2944, 0x4944, 0x8948, 0xA50, 0x1840, 0x28A0, 0x48A0, 0x8890, 0x910, 0x908, 0x5204, 0x2402}},
    {29381, 2, {0x8, 0x443C, 0x29E0, 0x1020, 0x2820, 0x4820, 0x8820, 0xBFE, 0x1820, 0x2820, 0x4820, 0x8820, 0x820, 0x820, 0x51FC, 0x2000}},
    {29382, 2, {0x20, 0x4420, 0x2820, 0x1020, 0x29FC, 0x4924, 0x8924, 0x924, 0x1924, 0x29FC, 0x4924, 0x8820, 0x820, 0x820, 0x5020, 0x2020}},
    {29383, 2, {0x40, 0x4440, 0x2840, 0x1040, 0x2BFE, 0x4840, 0x88E0, 0x8E0, 0x1950, 0x2950, 0x4A48, 0x8C44, 0x842, 0x840, 0x5040, 0x2040}},
    {29384, 2, {0x0, 0x45FC, 0x2904, 0x1104, 0x2924, 0x4924, 0x8924, 0x924, 0x1924, 0x2924, 0x4924, 0x8850, 0x848, 0x884, 0x5102, 0x2602}},
    {29385, 2, {0x0, 0x45FE, 0x2810, 0x1010, 0x2820, 0x4820, 0x8868, 0x8A4, 0x1922, 0x2A22, 0x4820, 0x8820, 0x820, 0x800, 0x53FE, 0x2000}},
    {29386, 2, {0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x110, 0x108, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {29387, 2, {0x0, 0x45FC, 0x2800, 0x1000, 0x2800, 0x4BFE, 0x8820, 0x820, 0x1928, 0x2924, 0x4A24, 0x8A22, 0xC22, 0x820, 0x50A0, 0x2040}},
    {29388, 2, {0x20, 0x4420, 0x2920, 0x1120, 0x29FC, 0x4920, 0x8A20, 0x820, 0x1820, 0x29FC, 0x4820, 0x8820, 0x820, 0x820, 0x53FE, 0x2000}},
    {29389, 2, {0x80, 0x4480, 0x29FC, 0x1104, 0x2A04, 0x4DF4, 0x8914, 0x914, 0x1914, 0x29F4, 0x4904, 0x8928, 0x912, 0x902, 0x50FE, 0x2000}},
    {29390, 2, {0x0, 0x45FC, 0x2924, 0x1124, 0x2924, 0x49FC, 0x8924, 0x924, 0x1924, 0x29FC, 0x4924, 0x8820, 0x820, 0x820, 0x5020, 0x2020}},
    {29391, 2, {0x80, 0x4480, 0x28FE, 0x1100, 0x2A20, 0x4920, 0x892C, 0x974, 0x1BA4, 0x2924, 0x4934, 0x8928, 0x922, 0x902, 0x50FE, 0x2000}},
    {29392, 2, {0x8, 0x883C, 0x53D0, 0x2290, 0x5290, 0x9290, 0x1290, 0x1290, 0x3290, 0x5288, 0x9288, 0x12C8, 0x12A4, 0x14D4, 0xA492, 0x4800}},
    {29393, 2, {0x40, 0x4440, 0x28A0, 0x10A0, 0x2910, 0x4A48, 0x8C26, 0x820, 0x1BF8, 0x2808, 0x4810, 0x8910, 0x8A0, 0x840, 0x5020, 0x2020}},
    {29394, 2, {0x90, 0x4490, 0x2890, 0x13FC, 0x2894, 0x4894, 0x8BFC, 0xA90, 0x1A90, 0x2BFE, 0x4892, 0x8892, 0x91A, 0x914, 0x5210, 0x2410}},
    {29395, 2, {0x10, 0x4410, 0x2810, 0x11FE, 0x2912, 0x4914, 0x8910, 0x9FC, 0x1944, 0x2944, 0x4928, 0x8928, 0x910, 0xA28, 0x5244, 0x2482}},
    {29396, 2, {0x0, 0x45FC, 0x2904, 0x1104, 0x2904, 0x49FC, 0x8900, 0x940, 0x1944, 0x2948, 0x4970, 0x8940, 0x942, 0xA42, 0x523E, 0x2400}},
    {29397, 2, {0x20, 0x8920, 0x5120, 0x2220, 0x52FC, 0x94A4, 0x1724, 0x1124, 0x3224, 0x5224, 0x94A4, 0x17A4, 0x10C4, 0x1044, 0xA094, 0x4108}},
    {29398, 2, {0x40, 0x4420, 0x2820, 0x13FE, 0x2A02, 0x4C04, 0x8800, 0x890, 0x1890, 0x2890, 0x4888, 0x8908, 0x908, 0x904, 0x5204, 0x2402}},
    {29399, 2, {0x80, 0x4480, 0x2900, 0x11FC, 0x2A04, 0x4C04, 0x89E4, 0x924, 0x1924, 0x2924, 0x49E4, 0x8924, 0x804, 0x804, 0x5028, 0x2010}},
    {29400, 2, {0x28, 0x4424, 0x2824, 0x1020, 0x2BFE, 0x4A20, 0x8A24, 0xA24, 0x1A24, 0x2A28, 0x4A28, 0x8A90, 0xB12, 0xA2A, 0x5046, 0x2082}},
    {29401, 2, {0x0, 0x45F8, 0x2908, 0x1108, 0x2908, 0x49F8, 0x8908, 0x908, 0x1908, 0x29F8, 0x4908, 0x8908, 0x908, 0x908, 0x57FE, 0x2000}},
    {29402, 2, {0x0, 0x4400, 0x29FC, 0x1104, 0x2904, 0x4904, 0x89FC, 0x904, 0x1904, 0x2904, 0x49FC, 0x8904, 0x800, 0x800, 0x53FE, 0x2000}},
    {29403, 2, {0x20, 0x4420, 0x2840, 0x11FC, 0x2904, 0x4904, 0x8904, 0x904, 0x19FC, 0x2904, 0x4904, 0x8904, 0x904, 0x904, 0x51FC, 0x2104}},
    {29404, 2, {0x20, 0x4420, 0x2820, 0x1020, 0x2BFE, 0x4820, 0x8820, 0x820, 0x19FC, 0x2904, 0x4904, 0x8904, 0x904, 0x904, 0x51FC, 0x2104}},
    {29405, 2, {0x80, 0x4480, 0x2880, 0x11FE, 0x2902, 0x4A04, 0x8C20, 0x820, 0x1928, 0x2924, 0x4A24, 0x8A22, 0xC22, 0x820, 0x50A0, 0x2040}},
    {29406, 2, {0x40, 0x4420, 0x2820, 0x13FE, 0x2A02, 0x4C04, 0x8800, 0x800, 0x1BFE, 0x2820, 0x4820, 0x8820, 0x820, 0x820, 0x50A0, 0x2040}},
    {29407, 2, {0x0, 0x45FE, 0x2800, 0x1000, 0x29FC, 0x4904, 0x8904, 0x9FC, 0x1904, 0x2904, 0x49FC, 0x8904, 0x800, 0x800, 0x53FE, 0x2000}},
    {29408, 2, {0x0, 0x45F8, 0x2908, 0x1108, 0x29F8, 0x4908, 0x8908, 0x9F8, 0x1944, 0x2948, 0x4930, 0x8920, 0x910, 0x948, 0x5186, 0x2100}},
    {29409, 2, {0x40, 0x4420, 0x2820, 0x13FE, 0x2800, 0x4888, 0x8904, 0xA02, 0x1888, 0x2888, 0x4850, 0x8850, 0x820, 0x850, 0x5088, 0x2306}},
    {29410, 2, {0x80, 0x4480, 0x28F8, 0x1108, 0x2B10, 0x4CA0, 0x8840, 0x8A0, 0x1918, 0x2A06, 0x4DF8, 0x8908, 0x908, 0x908, 0x51F8, 0x2108}},
    {29411, 2, {0x90, 0x4490, 0x2890, 0x1292, 0x2994, 0x4898, 0x8890, 0x998, 0x1A94, 0x2C92, 0x4890, 0x8890, 0x912, 0x912, 0x5212, 0x240E}},
    {29412, 2, {0x20, 0x4420, 0x2820, 0x13FE, 0x2820, 0x4820, 0x89FC, 0x800, 0x1800, 0x29FC, 0x4904, 0x8904, 0x904, 0x904, 0x51FC, 0x2104}},
    {29413, 2, {0x80, 0x4480, 0x29FC, 0x1104, 0x2A04, 0x49E4, 0x8924, 0x924, 0x19E4, 0x2924, 0x4924, 0x89E4, 0x904, 0x804, 0x5028, 0x2010}},
    {29414, 2, {0x0, 0x8BF8, 0x52A8, 0x22A8, 0x52A8, 0x9FFE, 0x12A8, 0x12A8, 0x32A8, 0x5FFE, 0x92A8, 0x12A8, 0x12A8, 0x12A8, 0xA208, 0x4218}},
    {29415, 2, {0x8, 0x443C, 0x29E0, 0x1020, 0x2820, 0x4BFE, 0x8820, 0x820, 0x1820, 0x29FC, 0x4904, 0x8904, 0x904, 0x904, 0x51FC, 0x2104}},
    {29416, 2, {0x28, 0x4424, 0x2824, 0x1020, 0x2BFE, 0x4820, 0x8924, 0x924, 0x1924, 0x2BA8, 0x4928, 0x8910, 0xA12, 0xA2A, 0x5446, 0x2082}},
    {29417, 2, {0x40, 0x8820, 0x53FE, 0x2202, 0x5414, 0x9010, 0x1010, 0x13FE, 0x3010, 0x5110, 0x9090, 0x1090, 0x1010, 0x1010, 0xA050, 0x4020}},
    {29418, 2, {0x0, 0x45FE, 0x2902, 0x1102, 0x297A, 0x4902, 0x8902, 0x97A, 0x194A, 0x294A, 0x494A, 0x897A, 0x902, 0x902, 0x510A, 0x2104}},
    {29419, 2, {0x40, 0x4440, 0x2844, 0x13F4, 0x2848, 0x4850, 0x8FFE, 0x840, 0x1880, 0x2984, 0x4A98, 0x8CE0, 0x882, 0x882, 0x507E, 0x2000}},
    {29420, 2, {0x20, 0x4420, 0x2820, 0x11FC, 0x2924, 0x4924, 0x8924, 0x924, 0x19FC, 0x2924, 0x4820, 0x8828, 0x824, 0xBFE, 0x5102, 0x2000}},
    {29421, 2, {0x20, 0x4420, 0x2820, 0x11FC, 0x2820, 0x4924, 0x88A4, 0x8A8, 0x1820, 0x2BFE, 0x4850, 0x8850, 0x888, 0x888, 0x5104, 0x2202}},
    {29422, 2, {0x100, 0x89FE, 0x5110, 0x2510, 0x557C, 0x9554, 0x1554, 0x1554, 0x3554, 0x5554, 0x9554, 0x1154, 0x125C, 0x1210, 0xA410, 0x4810}},
    {29423, 2, {0x40, 0x8840, 0x50A0, 0x2110, 0x5208, 0x9406, 0x11F0, 0x1000, 0x3000, 0x57FC, 0x9040, 0x1080, 0x1110, 0x1208, 0xA7FC, 0x4204}},
    {29424, 2, {0x80, 0x4480, 0x29F0, 0x1210, 0x2820, 0x4BFC, 0x8844, 0x844, 0x1FFE, 0x2844, 0x4844, 0x8BFC, 0x844, 0x840, 0x5140, 0x2080}},
    {29425, 2, {0x20, 0x9428, 0x5224, 0x2224, 0x5020, 0x9E20, 0x12FC, 0x1220, 0x3220, 0x5250, 0x92D0, 0x1350, 0x1288, 0x1088, 0xA104, 0x4202}},
    {29426, 2, {0x0, 0x8810, 0x5790, 0x2090, 0x5110, 0x9210, 0x12D8, 0x1354, 0x3254, 0x5692, 0x9A92, 0x1312, 0x1210, 0x1210, 0xA250, 0x4620}},
    {29427, 2, {0x40, 0x8840, 0x50A0, 0x2110, 0x5208, 0x95F6, 0x1040, 0x1040, 0x37FC, 0x5040, 0x9250, 0x1248, 0x1444, 0x1844, 0xA140, 0x4080}},
    {29428, 2, {0x220, 0x8A24, 0x5228, 0x23B0, 0x5220, 0x9222, 0x12A2, 0x131E, 0x3040, 0x5040, 0x93FC, 0x1040, 0x1040, 0x1040, 0xA7FE, 0x4000}},
    {29429, 2, {0x90, 0x4488, 0x2880, 0x13FE, 0x28A0, 0x48A8, 0x88B0, 0x8A4, 0x18A8, 0x28B0, 0x4924, 0x892A, 0x932, 0xA22, 0x52DE, 0x2400}},
    {29430, 2, {0x184, 0x4468, 0x2830, 0x10C8, 0x2B04, 0x4840, 0x8BFE, 0x8A0, 0x1920, 0x2BFC, 0x4D24, 0x8924, 0x934, 0x928, 0x5020, 0x2020}},
    {29431, 2, {0x0, 0x45F8, 0x2908, 0x1108, 0x29F8, 0x4800, 0x8BFC, 0xA04, 0x1A04, 0x2BFC, 0x4A04, 0x8A04, 0xBFC, 0xA04, 0x5214, 0x2208}},
    {29432, 2, {0x0, 0x45FC, 0x2924, 0x1124, 0x29FC, 0x4924, 0x8924, 0x9FC, 0x1820, 0x2820, 0x49FC, 0x8820, 0x820, 0x820, 0x53FE, 0x2000}},
    {29433, 2, {0x20, 0x4420, 0x2820, 0x13FE, 0x2820, 0x4820, 0x8924, 0x924, 0x1924, 0x2AAA, 0x4820, 0x8850, 0x850, 0x888, 0x5104, 0x2202}},
    {29434, 2, {0x40, 0x4420, 0x2BFE, 0x1000, 0x2800, 0x49FC, 0x8800, 0x800, 0x19FC, 0x2800, 0x4800, 0x89FC, 0x904, 0x904, 0x51FC, 0x2104}},
    {29435, 2, {0x40, 0x4440, 0x2888, 0x1104, 0x2BFE, 0x4802, 0x8888, 0x944, 0x1A42, 0x28F8, 0x4988, 0x8A50, 0x820, 0x850, 0x5188, 0x2606}},
    {29436, 2, {0x40, 0x4420, 0x29FC, 0x1104, 0x2904, 0x49FC, 0x8904, 0x904, 0x19FC, 0x2920, 0x4922, 0x8914, 0x908, 0x944, 0x5182, 0x2100}},
    {29437, 2, {0x0, 0x45FC, 0x2904, 0x1104, 0x29FC, 0x4904, 0x8904, 0x9FC, 0x1904, 0x2904, 0x49FC, 0x8800, 0x888, 0x884, 0x5102, 0x2202}},
    {29438, 2, {0x104, 0x890E, 0x5130, 0x2120, 0x57E0, 0x9120, 0x113E, 0x1164, 0x31A4, 0x5724, 0x9124, 0x1124, 0x1124, 0x1124, 0xA544, 0x4284}},
    {29439, 2, {0x4, 0x440E, 0x2BB8, 0x1088, 0x2888, 0x4928, 0x892E, 0xBA8, 0x18A8, 0x2AA8, 0x4AA8, 0x893E, 0x900, 0xA80, 0x547E, 0x2800}},
    {29440, 2, {0x10, 0x8A10, 0x5110, 0x2110, 0x5454, 0x9252, 0x1252, 0x1090, 0x3010, 0x5114, 0x9604, 0x1208, 0x1208, 0x1210, 0xA220, 0x40C0}},
    {29441, 2, {0x84, 0x89C4, 0x5704, 0x2114, 0x5114, 0x9114, 0x1FD4, 0x1114, 0x3314, 0x5394, 0x9554, 0x1504, 0x1904, 0x1104, 0xA114, 0x4108}},
    {29442, 2, {0x0, 0x45FC, 0x2904, 0x11FC, 0x2904, 0x49FC, 0x8800, 0x800, 0x19FC, 0x2820, 0x4820, 0x8BFE, 0x820, 0x820, 0x5020, 0x2020}},
    {29443, 2, {0x40, 0x8840, 0x50A0, 0x20A0, 0x5110, 0x9208, 0x15F6, 0x1000, 0x3088, 0x5048, 0x9248, 0x1150, 0x1110, 0x1020, 0xA7FE, 0x4000}},
    {29444, 2, {0x40, 0x4420, 0x2BFE, 0x1000, 0x2800, 0x49FC, 0x8904, 0x904, 0x1904, 0x29FC, 0x4820, 0x88A8, 0x924, 0xA22, 0x50A0, 0x2040}},
    {29445, 2, {0x90, 0x8890, 0x5090, 0x279E, 0x5090, 0x9090, 0x1090, 0x139C, 0x3090, 0x5090, 0x9090, 0x179E, 0x1090, 0x1090, 0xA090, 0x4090}},
    {29446, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440, 0x120, 0x110, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {29447, 2, {0x20, 0x4420, 0x283E, 0x1020, 0x29FE, 0x4922, 0x8938, 0x9E0, 0x1922, 0x291E, 0x4900, 0x8978, 0xA48, 0xA4A, 0x548A, 0x2906}},
    {29448, 2, {0x40, 0x4480, 0x29FC, 0x1124, 0x2924, 0x49FC, 0x8924, 0x944, 0x19FC, 0x2890, 0x4910, 0x8BFE, 0x810, 0x810, 0x5010, 0x2010}},
    {29449, 2, {0x88, 0x4488, 0x2BFE, 0x1088, 0x2888, 0x48F8, 0x8888, 0x888, 0x18F8, 0x2888, 0x4888, 0x8BFE, 0x800, 0x888, 0x5104, 0x2202}},
    {29450, 2, {0x0, 0x9040, 0x539C, 0x2204, 0x5204, 0x939C, 0x1204, 0x1204, 0x33FC, 0x5090, 0x9090, 0x1090, 0x1090, 0x1112, 0xA212, 0x440E}},
    {29451, 2, {0x110, 0x108, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006, 0x1428, 0x1224, 0xFDFE, 0x1020, 0x2850, 0x2450, 0x4488, 0x8106}},
    {29452, 2, {0x810, 0x814, 0x812, 0xFF12, 0x810, 0x2AFE, 0x2A10, 0x2A10, 0x5D10, 0x8828, 0x1C28, 0x2A28, 0x4944, 0x8844, 0x884, 0x902}},
    {29453, 2, {0x40, 0x8840, 0x5040, 0x27FC, 0x5040, 0x9248, 0x1248, 0x1248, 0x3554, 0x50E0, 0x9150, 0x1150, 0x1248, 0x1444, 0xA842, 0x4040}},
    {29454, 2, {0x88, 0x4488, 0x2888, 0x13FE, 0x2888, 0x4888, 0x8FFE, 0x800, 0x19FC, 0x2904, 0x4904, 0x89FC, 0x904, 0x904, 0x51FC, 0x2104}},
    {29455, 2, {0x40, 0x8820, 0x53FC, 0x2204, 0x5204, 0x93FC, 0x1200, 0x12FC, 0x3284, 0x52FC, 0x9284, 0x12FC, 0x1484, 0x1484, 0xA894, 0x4088}},
    {29456, 2, {0x208, 0x8908, 0x5110, 0x27FC, 0x5040, 0x9040, 0x13F8, 0x1040, 0x3080, 0x5FFE, 0x9120, 0x1120, 0x1220, 0x1222, 0xA422, 0x481E}},
    {29457, 2, {0x0, 0x45FC, 0x2904, 0x1104, 0x29FC, 0x4904, 0x8904, 0x9FC, 0x1800, 0x2912, 0x49D4, 0x8918, 0x910, 0x952, 0x5192, 0x210E}},
    {29458, 2, {0x10, 0x7E14, 0x4212, 0x7E12, 0x4210, 0x7EFE, 0x10, 0x7E10, 0x4210, 0x7E28, 0x4228, 0x7E28, 0x4244, 0x4244, 0x4A84, 0x4502}},
    {29459, 2, {0x0, 0x8BF8, 0x5248, 0x2248, 0x53F8, 0x9248, 0x1248, 0x13F8, 0x3040, 0x57FC, 0x90E0, 0x1150, 0x1248, 0x1446, 0xA040, 0x4040}},
    {29460, 2, {0x40, 0x4420, 0x2BFE, 0x1202, 0x2800, 0x49FC, 0x8800, 0x800, 0x1BFE, 0x2820, 0x4928, 0x8924, 0xA22, 0xC22, 0x50A0, 0x2040}},
    {29461, 2, {0x20, 0x9720, 0x5120, 0x213E, 0x5142, 0x9744, 0x1490, 0x1410, 0x3410, 0x5754, 0x9152, 0x1152, 0x1192, 0x1110, 0xA550, 0x4220}},
    {29462, 2, {0x1F8, 0x4508, 0x2908, 0x11F8, 0x2908, 0x4908, 0x89F8, 0x800, 0x1BFC, 0x2A04, 0x4A04, 0x8BFC, 0xA04, 0xA04, 0x53FC, 0x2204}},
    {29463, 2, {0x40, 0x8840, 0x53FC, 0x20A0, 0x5110, 0x9208, 0x17FE, 0x1008, 0x33C8, 0x5248, 0x9248, 0x1248, 0x13C8, 0x1008, 0xA028, 0x4010}},
    {29464, 2, {0x82, 0x8882, 0x5282, 0x23EA, 0x548A, 0x908A, 0x17FA, 0x108A, 0x33EA, 0x52AA, 0x92AA, 0x12AA, 0x12A2, 0x12E2, 0xA08A, 0x4084}},
    {29465, 2, {0x3C, 0x47C0, 0x2A44, 0x1128, 0x2800, 0x4BF8, 0x8848, 0x848, 0x1FFE, 0x2848, 0x4848, 0x8BF8, 0x848, 0x840, 0x5140, 0x2080}},
    {29466, 2, {0xA0, 0x4490, 0x2880, 0x11FE, 0x2910, 0x4B10, 0x8DFC, 0x910, 0x1910, 0x29FC, 0x4910, 0x8910, 0x910, 0x9FE, 0x5100, 0x2100}},
    {29467, 2, {0x0, 0x45FC, 0x2808, 0x1010, 0x2820, 0x4BFE, 0x8820, 0x8A0, 0x1840, 0x29FC, 0x4954, 0x8954, 0x954, 0x954, 0x57FE, 0x2000}},
    {29468, 2, {0x20, 0x4420, 0x2BFE, 0x1020, 0x29FC, 0x4820, 0x8BFE, 0x800, 0x19FC, 0x2904, 0x49FC, 0x8904, 0x9FC, 0x904, 0x5114, 0x2108}},
    {29469, 2, {0x80, 0x8840, 0x57FC, 0x2000, 0x5110, 0x9110, 0x12A8, 0x1444, 0x3000, 0x5040, 0x97FE, 0x1040, 0x1040, 0x1040, 0xA040, 0x4040}},
    {29470, 2, {0x40, 0x8840, 0x50A0, 0x2110, 0x5208, 0x95F6, 0x1040, 0x1040, 0x37FC, 0x5040, 0x9040, 0x13F8, 0x1208, 0x1208, 0xA3F8, 0x4208}},
    {29471, 2, {0x0, 0x9484, 0x5244, 0x2248, 0x5000, 0x93F8, 0x1248, 0x1248, 0x33F8, 0x5248, 0x9248, 0x13F8, 0x124A, 0x144A, 0xA446, 0x4802}},
    {29472, 2, {0x90, 0x4490, 0x2890, 0x13FC, 0x2A94, 0x4A94, 0x8A94, 0xBFC, 0x1A94, 0x2A94, 0x4A94, 0x8FFE, 0x800, 0x890, 0x5108, 0x2204}},
    {29473, 2, {0x0, 0x8BFE, 0x5252, 0x2252, 0x5252, 0x93FE, 0x1020, 0x1040, 0x30FC, 0x5104, 0x9288, 0x1050, 0x1020, 0x1040, 0xA180, 0x4600}},
    {29474, 2, {0x100, 0x891E, 0x5112, 0x27D2, 0x5112, 0x911E, 0x1112, 0x17D2, 0x3452, 0x545E, 0x9452, 0x17D2, 0x1452, 0x1022, 0xA02A, 0x4044}},
    {29475, 2, {0x0, 0x4488, 0x2A52, 0x1222, 0x2A52, 0x4A8A, 0x8A02, 0xBFE, 0x1888, 0x2944, 0x4A7A, 0x8888, 0x950, 0x820, 0x50D8, 0x2706}},
    {29476, 2, {0x10, 0x45D4, 0x2858, 0x1252, 0x298C, 0x4888, 0x8904, 0xAFA, 0x1820, 0x2820, 0x49FE, 0x8820, 0x850, 0x888, 0x5104, 0x2202}},
    {29477, 2, {0x0, 0x8BF8, 0x5248, 0x2248, 0x53F8, 0x9248, 0x1248, 0x13F8, 0x3000, 0x57FE, 0x9240, 0x1244, 0x1228, 0x1290, 0xA308, 0x4206}},
    {29478, 2, {0x0, 0x97FC, 0x5404, 0x2434, 0x65C4, 0xA444, 0x25F4, 0x2554, 0x6554, 0xA5F4, 0x2444, 0x2454, 0x25F4, 0x2494, 0xA806, 0x5002}},
    {29479, 2, {0x0, 0x45FC, 0x2904, 0x1104, 0x29E4, 0x4924, 0x8924, 0xBFE, 0x1A02, 0x2AFA, 0x4A8A, 0x8A8A, 0xAFA, 0xA02, 0x520A, 0x2204}},
    {29480, 2, {0x3C, 0x97C0, 0x5244, 0x2128, 0x5000, 0x93FC, 0x1080, 0x1080, 0x37FE, 0x5100, 0x91F8, 0x1288, 0x1250, 0x1420, 0xA8D8, 0x4306}},
    {29481, 2, {0x0, 0x8BF8, 0x5208, 0x23F8, 0x5208, 0x93F8, 0x1040, 0x1240, 0x33FC, 0x5440, 0x9040, 0x13F8, 0x1040, 0x1040, 0xA7FE, 0x4000}},
    {29482, 2, {0x40, 0x8844, 0x53F4, 0x2048, 0x5050, 0x97FE, 0x1040, 0x1080, 0x31F8, 0x5308, 0x9508, 0x11F8, 0x1108, 0x1108, 0xA1F8, 0x4108}},
    {29483, 2, {0x88, 0x4488, 0x2BFE, 0x1088, 0x2888, 0x4800, 0x89FC, 0x924, 0x1924, 0x2924, 0x49FC, 0x8924, 0x924, 0x924, 0x51FC, 0x2104}},
    {29484, 2, {0x0, 0x97FC, 0x5444, 0x27FC, 0x5444, 0x97FC, 0x1000, 0x13F8, 0x3208, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x1208, 0xA228, 0x4210}},
    {29485, 2, {0x100, 0x89F8, 0x5208, 0x23F0, 0x5010, 0x97FE, 0x1080, 0x1144, 0x3668, 0x50B0, 0x9128, 0x1668, 0x10A4, 0x1122, 0xA6A0, 0x4040}},
    {29486, 2, {0x810, 0x814, 0xFF92, 0x812, 0x810, 0x7F7E, 0x4110, 0x5510, 0x4910, 0x5D10, 0x4928, 0x7F28, 0x4928, 0x4944, 0x4544, 0x4282}},
    {29487, 2, {0x40, 0x9444, 0x5444, 0x2444, 0x57FC, 0x9000, 0x1FFE, 0x1040, 0x3080, 0x57FC, 0x94A4, 0x14A4, 0x14A4, 0x14A4, 0xA4A4, 0x440C}},
    {29488, 2, {0x80, 0x449E, 0x2BEA, 0x108A, 0x29CA, 0x488A, 0x8BEA, 0x892, 0x18A6, 0x2820, 0x4BFE, 0x8820, 0x850, 0x888, 0x5104, 0x2602}},
    {29489, 2, {0x0, 0x8BF8, 0x5090, 0x2060, 0x57FC, 0x90A4, 0x1128, 0x1220, 0x3460, 0x5040, 0x97FC, 0x10E0, 0x1150, 0x1248, 0xAC46, 0x4040}},
    {29490, 2, {0x0, 0x45FC, 0x2904, 0x11FC, 0x2904, 0x49FC, 0x8880, 0x9FE, 0x1A22, 0x2922, 0x4952, 0x8902, 0x9FA, 0x802, 0x5014, 0x2008}},
    {29491, 2, {0x0, 0x97BC, 0x5484, 0x2484, 0x5484, 0x97BC, 0x1400, 0x1400, 0x37BC, 0x5424, 0x9424, 0x17A8, 0x1410, 0x1428, 0xA444, 0x4482}},
    {29492, 2, {0x100, 0x8978, 0x5108, 0x2208, 0x52FE, 0x9640, 0x1A40, 0x127C, 0x3290, 0x5210, 0x92FE, 0x1210, 0x1228, 0x1228, 0xA244, 0x4282}},
    {29493, 2, {0x80, 0x8840, 0x53FC, 0x2204, 0x5204, 0x93FC, 0x1200, 0x1200, 0x33FC, 0x5354, 0x9354, 0x15FC, 0x1554, 0x1554, 0xA944, 0x410C}},
    {29494, 2, {0x108, 0x4488, 0x2890, 0x17FE, 0x2890, 0x4890, 0x8BFC, 0xA94, 0x1A94, 0x2B0C, 0x4A04, 0x8BFC, 0xA04, 0xA04, 0x53FC, 0x2204}},
    {29495, 2, {0x2210, 0x1214, 0x1412, 0xFF92, 0x1410, 0x14FE, 0x7F10, 0x5510, 0x5510, 0x6328, 0x4128, 0x7F28, 0x4144, 0x4144, 0x7F84, 0x4102}},
    {29496, 2, {0x0, 0x8BFC, 0x5224, 0x2224, 0x53FC, 0x9200, 0x12FC, 0x1284, 0x3284, 0x52FC, 0x9284, 0x12FC, 0x1284, 0x1284, 0xA4FC, 0x4884}},
    {29497, 2, {0x40, 0x9040, 0x57FC, 0x20E0, 0x5150, 0x9248, 0x1446, 0x13F8, 0x3208, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x1000, 0xA7FE, 0x4000}},
    {29498, 2, {0x8, 0x443C, 0x2BC0, 0x1004, 0x2A44, 0x4928, 0x89FC, 0xA20, 0x1820, 0x2BFE, 0x4820, 0x8924, 0x924, 0x924, 0x51FC, 0x2004}},
    {29499, 2, {0x6, 0x8F78, 0x5110, 0x2120, 0x5244, 0x9278, 0x1310, 0x1224, 0x367E, 0x5A12, 0x9210, 0x1254, 0x1252, 0x1292, 0xA250, 0x4620}},
    {29500, 2, {0x28, 0x4424, 0x2BFE, 0x1020, 0x29FC, 0x4924, 0x89FC, 0x924, 0x19FC, 0x2924, 0x4808, 0x8BFE, 0x888, 0x848, 0x5048, 0x2018}},
    {29501, 2, {0x0, 0x47FE, 0x2A02, 0x1000, 0x29FC, 0x4904, 0x89FC, 0x904, 0x19FC, 0x2840, 0x4820, 0x8BFE, 0x800, 0x888, 0x5104, 0x2202}},
    {29502, 2, {0x0, 0x8BF8, 0x5208, 0x23C8, 0x5248, 0x9FFE, 0x1802, 0x13F8, 0x3208, 0x53F8, 0x9208, 0x13F8, 0x1208, 0x1208, 0xA228, 0x4210}},
    {29503, 2, {0x40, 0x9040, 0x53F8, 0x2040, 0x57FE, 0x9000, 0x13F8, 0x1208, 0x3208, 0x53F8, 0x90A4, 0x1128, 0x1310, 0x1D48, 0xA186, 0x4100}},
    {29504, 2, {0x140, 0x965C, 0x5444, 0x2444, 0x575C, 0x9444, 0x1444, 0x17FC, 0x3040, 0x57FC, 0x9208, 0x1110, 0x10A0, 0x1040, 0xA1B0, 0x460E}},
    {29505, 2, {0x0, 0x8BFC, 0x5240, 0x23F8, 0x5240, 0x93F8, 0x1240, 0x1240, 0x33FC, 0x5004, 0x9554, 0x1554, 0x1554, 0x1404, 0xA028, 0x4010}},
    {29506, 2, {0x0, 0x45FE, 0x2910, 0x1120, 0x297C, 0x4944, 0x897C, 0x944, 0x197C, 0x2910, 0x4910, 0x8954, 0x952, 0xA92, 0x5250, 0x2420}},
    {29507, 2, {0x810, 0x4914, 0x4912, 0x4912, 0x7F10, 0xFE, 0x7F10, 0x10, 0x7F10, 0x4128, 0x7F28, 0x4128, 0x2244, 0xF44, 0xF084, 0x4102}},
    {29508, 2, {0x210, 0x9114, 0x5112, 0x27D2, 0x5010, 0x9790, 0x103E, 0x1790, 0x3010, 0x5790, 0x94A8, 0x14A8, 0x14A8, 0x17C4, 0xA044, 0x4082}},
    {29509, 2, {0x200, 0x94FE, 0x5710, 0x2510, 0x557C, 0x9554, 0x1754, 0x1454, 0x3454, 0x5754, 0x9554, 0x155C, 0x1510, 0x1710, 0xA510, 0x4010}},
    {29510, 2, {0x20, 0x4440, 0x29FC, 0x1104, 0x29FC, 0x4904, 0x89FC, 0x840, 0x1BFE, 0x2888, 0x4924, 0x8A22, 0xDFC, 0x820, 0x5020, 0x2020}},
    {29511, 2, {0x208, 0x8910, 0x57FC, 0x2040, 0x5040, 0x93F8, 0x1040, 0x1080, 0x37FC, 0x50A0, 0x9128, 0x1134, 0x123C, 0x1222, 0xA422, 0x481E}},
    {29512, 2, {0x208, 0x8908, 0x5110, 0x2000, 0x57FC, 0x9000, 0x1110, 0x1208, 0x3404, 0x53F8, 0x92A8, 0x12A8, 0x12A8, 0x12A8, 0xAFFE, 0x4000}},
    {29513, 2, {0x40, 0x8840, 0x57FC, 0x2040, 0x53F8, 0x9080, 0x17FC, 0x1110, 0x3268, 0x55C6, 0x9040, 0x13F8, 0x10E0, 0x1158, 0xA644, 0x4040}},
    {29514, 2, {0x40, 0x8840, 0x50A0, 0x2190, 0x5248, 0x97FE, 0x1A08, 0x13F8, 0x3208, 0x53F8, 0x9200, 0x13F8, 0x1508, 0x1508, 0xA9F8, 0x4108}},
    {29515, 2, {0x40, 0x45FC, 0x2904, 0x11FC, 0x2904, 0x49FC, 0x8904, 0x9FC, 0x1820, 0x2924, 0x48A8, 0x8924, 0x820, 0xBFE, 0x5020, 0x2020}},
    {29516, 2, {0x0, 0x45F8, 0x2908, 0x11F8, 0x2908, 0x49F8, 0x8800, 0xBFC, 0x1A94, 0x2BFC, 0x4800, 0x89F8, 0x890, 0x860, 0x5198, 0x2606}},
    {29517, 2, {0x80, 0x8840, 0x53F8, 0x2110, 0x50A0, 0x97FE, 0x1000, 0x13F8, 0x3208, 0x53F8, 0x9208, 0x13F8, 0x1120, 0x1122, 0xA222, 0x4C1E}},
    {29518, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8418, 0x120, 0x110, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {29519, 2, {0x88, 0x4488, 0x2BFE, 0x1088, 0x2800, 0x49FC, 0x8904, 0x9FC, 0x1904, 0x29FC, 0x4820, 0x8BFE, 0x850, 0x888, 0x5104, 0x2202}},
    {29520, 2, {0x20, 0x45FC, 0x2888, 0x1050, 0x2BFE, 0x4800, 0x89FC, 0x904, 0x19FC, 0x2904, 0x49FC, 0x8820, 0xBFE, 0x820, 0x5020, 0x2020}},
    {29521, 2, {0x100, 0x8902, 0x511C, 0x27D0, 0x5110, 0x97D0, 0x155E, 0x17D4, 0x3554, 0x57D4, 0x9114, 0x17D4, 0x1114, 0x1124, 0xA124, 0x4144}},
    {29522, 2, {0x820, 0x7F20, 0x83E, 0x7E44, 0x884, 0xFF28, 0x2010, 0x3E28, 0x4244, 0x8722, 0x110, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {29523, 2, {0x108, 0x9108, 0x5108, 0x27D0, 0x511E, 0x9794, 0x1124, 0x1FD4, 0x3214, 0x53D4, 0x9254, 0x1248, 0x1248, 0x14D4, 0xA424, 0x4842}},
    {29524, 2, {0x20, 0x4440, 0x29FC, 0x1104, 0x29FC, 0x4904, 0x89FC, 0x820, 0x1924, 0x28A8, 0x4924, 0x8820, 0xBFE, 0x820, 0x5020, 0x2020}},
    {29525, 2, {0x20, 0x4622, 0x2BFE, 0x1090, 0x2888, 0x49FE, 0x8B10, 0xD10, 0x19FE, 0x2910, 0x4910, 0x89FE, 0x910, 0x910, 0x51FE, 0x2100}},
    {29526, 2, {0x20, 0x45FC, 0x2820, 0x1088, 0x2BFE, 0x4888, 0x89FC, 0x904, 0x19FC, 0x2904, 0x49FC, 0x8904, 0x9FC, 0x888, 0x5104, 0x2202}},
    {29527, 2, {0x0, 0x9FFE, 0x5800, 0x2A28, 0x6948, 0xABEE, 0x2892, 0x2884, 0x6AA0, 0xAAA8, 0x2BE8, 0x2888, 0x2894, 0x2914, 0xA924, 0x5242}},
    {29528, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4144, 0x4020, 0x110, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006}},
    {29529, 2, {0x108, 0x8908, 0x5548, 0x2390, 0x511E, 0x97D4, 0x1564, 0x1554, 0x3554, 0x5754, 0x95D4, 0x1548, 0x1548, 0x1554, 0xA464, 0x44C2}},
    {29530, 2, {0x88, 0x4488, 0x29FC, 0x1088, 0x2888, 0x4BFE, 0x8820, 0x9FC, 0x1924, 0x29FC, 0x4924, 0x89FC, 0x800, 0x888, 0x5104, 0x2202}},
    {29531, 2, {0xA0, 0x94A4, 0x52A8, 0x20A0, 0x57FC, 0x9110, 0x10A0, 0x17FC, 0x3040, 0x53F8, 0x9040, 0x17FC, 0x10A0, 0x1110, 0xA208, 0x4C06}},
    {29532, 2, {0x20, 0x4524, 0x28A8, 0x13FE, 0x28A8, 0x4924, 0x8A22, 0x904, 0x1904, 0x29DE, 0x4A44, 0x8D54, 0x89E, 0x904, 0x5204, 0x2404}},
    {29533, 2, {0x0, 0x45FC, 0x2848, 0x1030, 0x2BFE, 0x4852, 0x8894, 0xB50, 0x1820, 0x2BFE, 0x4A52, 0x8A8A, 0xB76, 0xA52, 0x5272, 0x2206}},
    {29534, 2, {0x20, 0x45FC, 0x2888, 0x1050, 0x2BFE, 0x4800, 0x89FC, 0x924, 0x19FC, 0x2924, 0x49FC, 0x8820, 0x9FC, 0x820, 0x53FE, 0x2000}},
    {29535, 2, {0x20, 0x4420, 0x29FC, 0x1020, 0x2BFE, 0x4908, 0x8B9C, 0x908, 0x1988, 0x2E3E, 0x4800, 0x8BFE, 0x890, 0x890, 0x5112, 0x220E}},
    {29536, 2, {0x40, 0x9040, 0x57FC, 0x20A0, 0x5514, 0x9208, 0x17FC, 0x1A0A, 0x33F8, 0x5208, 0x93F8, 0x1040, 0x1248, 0x1444, 0xA944, 0x4080}},
    {29537, 2, {0x80, 0x8B1C, 0x5204, 0x2204, 0x539C, 0x9204, 0x1204, 0x13FC, 0x3100, 0x53FE, 0x9402, 0x12AA, 0x12AA, 0x1402, 0xA014, 0x4008}},
    {29538, 2, {0x38, 0x8BC0, 0x5080, 0x27FE, 0x5110, 0x9208, 0x15F6, 0x1910, 0x31F0, 0x5000, 0x97FC, 0x1404, 0x15F4, 0x1514, 0xA5F4, 0x440C}},
    {29539, 2, {0x9090, 0x4914, 0x212, 0x7F12, 0x4910, 0x7FFE, 0x4910, 0x7F10, 0x10, 0xFFA8, 0x28, 0x7F28, 0x4144, 0x4144, 0x7F84, 0x4102}},
    {29540, 2, {0x208, 0x8908, 0x57C8, 0x2010, 0x579E, 0x9494, 0x17A4, 0x1014, 0x3794, 0x5094, 0x9114, 0x11C8, 0x1708, 0x1114, 0xA524, 0x4242}},
    {29541, 2, {0x108, 0x8A08, 0x57C8, 0x2450, 0x57DE, 0x9464, 0x17D4, 0x1214, 0x3114, 0x57D4, 0x9214, 0x13C8, 0x1248, 0x1454, 0xA554, 0x48A2}},
    {29542, 2, {0x88, 0x47FE, 0x2888, 0x11FC, 0x2904, 0x49FC, 0x8904, 0x9FC, 0x1880, 0x29FE, 0x4A22, 0x8D52, 0x902, 0x9FA, 0x500A, 0x2004}},
    {29543, 2, {0x0, 0x8BFC, 0x5294, 0x2294, 0x53FC, 0x9000, 0x17FE, 0x1000, 0x33FC, 0x5204, 0x93FC, 0x10A2, 0x1114, 0x1308, 0xAD44, 0x4182}},
    {29544, 2, {0x0, 0x45FC, 0x2954, 0x1154, 0x29FC, 0x4880, 0x89FE, 0xA42, 0x19F2, 0x2952, 0x4952, 0x89F2, 0x84A, 0xBFA, 0x500A, 0x2004}},
    {29545, 2, {0x20, 0x893C, 0x5120, 0x27FE, 0x5010, 0x9014, 0x13FE, 0x1210, 0x33F0, 0x5254, 0x9254, 0x12D4, 0x1368, 0x144A, 0xA896, 0x4322}},
    {29546, 2, {0x40, 0x88A0, 0x5318, 0x2DF6, 0x5000, 0x97FC, 0x1554, 0x14E4, 0x37FC, 0x5000, 0x93F8, 0x1208, 0x13F8, 0x1208, 0xA3F8, 0x4208}},
    {29547, 2, {0x40, 0x4440, 0x28A0, 0x1110, 0x2A08, 0x4DF6, 0x8800, 0x800, 0x1BB8, 0x2AA8, 0x4AA8, 0x8BB8, 0x910, 0x910, 0x52A8, 0x2444}},
    {29548, 2, {0x200, 0x939E, 0x548A, 0x290A, 0x67D2, 0xA566, 0x2540, 0x27D4, 0x655E, 0xA564, 0x27C4, 0x2544, 0x255E, 0x2544, 0xA444, 0x48C4}},
    {29549, 2, {0x210, 0x9210, 0x5F9C, 0x2224, 0x6248, 0xAFBE, 0x2AA2, 0x2AAA, 0x6FAA, 0xA22A, 0x272A, 0x2AAA, 0x3288, 0x2214, 0xA222, 0x4242}},
    {29550, 2, {0x0, 0x97FC, 0x5248, 0x2444, 0x6FFE, 0xA444, 0x26EC, 0x2554, 0x66EC, 0xA444, 0x26EC, 0x2554, 0x26EC, 0x2444, 0xA454, 0x4408}},
    {29551, 2, {0x3C, 0x97C0, 0x5040, 0x2FFE, 0x5040, 0x97FC, 0x1554, 0x14E4, 0x37FC, 0x5040, 0x97FC, 0x1040, 0x1FFE, 0x1000, 0xA524, 0x4892}},
    {29552, 2, {0x80, 0x8840, 0x57FC, 0x2444, 0x52A8, 0x9294, 0x1474, 0x1000, 0x33F8, 0x52A8, 0x9FFE, 0x1000, 0x17FC, 0x1040, 0xA140, 0x4080}},
    {29553, 2, {0x40, 0x4420, 0x2BFE, 0x1202, 0x29FC, 0x4948, 0x8A50, 0x9FC, 0x1B04, 0x29FC, 0x4904, 0x89FC, 0x904, 0x9FC, 0x5088, 0x2104}},
    {29554, 2, {0x88, 0x47FE, 0x28A8, 0x1090, 0x29FE, 0x4B20, 0x89FC, 0x920, 0x19FC, 0x2920, 0x49FE, 0x8900, 0xBFC, 0x888, 0x5070, 0x238E}},
    {29555, 2, {0x0, 0x45FC, 0x2820, 0x13FE, 0x2A22, 0x49AC, 0x8820, 0x9AC, 0x1800, 0x2BFE, 0x4820, 0x89FC, 0x954, 0x954, 0x5154, 0x210C}},
    {29556, 2, {0x110, 0x97FC, 0x5110, 0x27FC, 0x5404, 0x93F8, 0x1000, 0x17FC, 0x3080, 0x5144, 0x96A8, 0x1170, 0x16A8, 0x1126, 0xA6A0, 0x4040}},
    {29557, 2, {0x124, 0x8A48, 0x5124, 0x2000, 0x53FC, 0x9294, 0x1264, 0x1294, 0x33FC, 0x5248, 0x9368, 0x1248, 0x136A, 0x124A, 0xA246, 0x4362}},
    {29558, 2, {0x3FC, 0x8840, 0x51F8, 0x2108, 0x51F8, 0x9108, 0x17FE, 0x1442, 0x32A4, 0x528A, 0x947A, 0x1080, 0x11F8, 0x1288, 0xA070, 0x438E}},
    {29559, 2, {0x40, 0x8820, 0x57FE, 0x2488, 0x55FC, 0x9488, 0x17FE, 0x1420, 0x35FC, 0x5524, 0x95FC, 0x1524, 0x15FC, 0x1400, 0xA488, 0x4904}},
    {29560, 2, {0x7710, 0x5514, 0x7712, 0x12, 0x7F10, 0x49FE, 0x7F10, 0x4910, 0x7F10, 0x28, 0xFFA8, 0x28, 0x7F44, 0x4144, 0x7F84, 0x4102}},
    {29561, 2, {0x40, 0x887C, 0x5040, 0x23FC, 0x5244, 0x93F0, 0x1244, 0x12FC, 0x32A8, 0x52F8, 0x92A8, 0x12F8, 0x1200, 0x15FC, 0xA554, 0x4BFE}},
    {29562, 2, {0x108, 0x8908, 0x57CE, 0x2112, 0x5124, 0x97DE, 0x1552, 0x155E, 0x37D2, 0x511E, 0x9392, 0x155E, 0x1940, 0x1114, 0xA112, 0x4122}},
    {29563, 2, {0x810, 0xF14, 0x812, 0x7F12, 0x4910, 0x5EFE, 0x4910, 0x7F10, 0x4A10, 0x5F28, 0x5128, 0x5B28, 0x5544, 0x5F44, 0x5584, 0x9502}},
    {29564, 2, {0x0, 0x9EFE, 0x5254, 0x2292, 0x62FE, 0xAE92, 0x28D6, 0x28BA, 0x68D6, 0xAE92, 0x22D6, 0x22BA, 0x22D6, 0x2292, 0xAA9A, 0x4484}},
    {29565, 2, {0x40, 0x97FC, 0x5000, 0x23B8, 0x52A8, 0x93B8, 0x1110, 0x17FC, 0x3110, 0x57FC, 0x9110, 0x1FFE, 0x1128, 0x1310, 0xAD48, 0x4186}},
    {29566, 2, {0x110, 0x8FFE, 0x5110, 0x27BC, 0x54A4, 0x97BC, 0x1140, 0x1120, 0x33FE, 0x5220, 0x97FC, 0x1A20, 0x13FC, 0x1220, 0xA3FE, 0x4200}},
    {29567, 2, {0x0, 0x97FE, 0x5040, 0x22EE, 0x62AA, 0xABEA, 0x2AAE, 0x2AE8, 0x6BAA, 0xACE6, 0x2288, 0x25F4, 0x2290, 0x2060, 0xA198, 0x4606}},
    {29568, 2, {0x0, 0x97FC, 0x54A4, 0x24A4, 0x67FC, 0xA210, 0x22A8, 0x24BE, 0x6F68, 0xA23C, 0x24A8, 0x2FBC, 0x2028, 0x2AA8, 0xAABE, 0x4020}},
    {29569, 2, {0x7BC, 0x94A4, 0x57BC, 0x2000, 0x57FE, 0x9400, 0x1590, 0x149E, 0x37D4, 0x5564, 0x95D4, 0x1554, 0x15D4, 0x1568, 0xABC8, 0x5054}},
    {29570, 2, {0x240, 0x97E2, 0x525C, 0x2EF0, 0x6AB0, 0xAEF0, 0x201E, 0x27D4, 0x6554, 0xA7D4, 0x2554, 0x27D4, 0x2114, 0x2FE4, 0xA124, 0x4144}},
    {29571, 2, {0x3DE, 0x4652, 0x2BDE, 0x1252, 0x2BDE, 0x48A0, 0x89FE, 0xB20, 0x1DFC, 0x2920, 0x49FE, 0x8800, 0x9FC, 0x888, 0x5070, 0x278E}},
    {29572, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x200, 0x420, 0x820, 0x1040, 0x3F80, 0x1100, 0x200, 0x420, 0x810, 0x3FF8, 0x1008, 0x8}},
    {29573, 2, {0x1010, 0x810, 0x810, 0xFE10, 0x1054, 0x1052, 0x2052, 0x4290, 0xFC10, 0x4814, 0x1004, 0x2008, 0x4408, 0xFE10, 0x4220, 0xC0}},
    {29574, 2, {0x1010, 0x808, 0x808, 0xFEFE, 0x1010, 0x1010, 0x2020, 0x4242, 0xFCFC, 0x4848, 0x1010, 0x2020, 0x4444, 0xFEFE, 0x4242, 0x0}},
    {29575, 2, {0x200, 0x100, 0x7FFC, 0x200, 0x4444, 0x2F88, 0x1110, 0x2248, 0x4FE4, 0x20, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {29576, 2, {0x1040, 0x840, 0x87E, 0xFE80, 0x1100, 0x101C, 0x20E0, 0x42A0, 0xFCA2, 0x48A4, 0x10A8, 0x2090, 0x4490, 0xFEA8, 0x42C4, 0x82}},
    {29577, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x120, 0x110, 0x110, 0x100, 0xFFFE, 0x0}},
    {29578, 2, {0x0, 0x0, 0x7FFC, 0x100, 0x120, 0x110, 0x110, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {29579, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {29580, 2, {0x40, 0x40, 0xFF40, 0x840, 0x840, 0x840, 0x840, 0x7E40, 0x840, 0x840, 0x840, 0x842, 0xF42, 0xF042, 0x403E, 0x0}},
    {29581, 2, {0x1000, 0x1000, 0x1000, 0x3FFC, 0x2100, 0x4100, 0x8100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {29582, 2, {0x0, 0x0, 0xFEFE, 0x1010, 0x1010, 0x1010, 0x7C10, 0x1010, 0x1010, 0x1010, 0x1010, 0x1E10, 0xF010, 0x4010, 0x50, 0x20}},
    {29583, 2, {0x40, 0x40, 0xFE40, 0x1040, 0x11FC, 0x1044, 0x7C44, 0x1044, 0x1044, 0x1084, 0x1084, 0x1E84, 0xF104, 0x4104, 0x228, 0x410}},
    {29584, 2, {0x0, 0x20, 0xFD20, 0x1120, 0x1120, 0x1120, 0x7D20, 0x1120, 0x1120, 0x1110, 0x1110, 0x1E10, 0xE208, 0x4408, 0x804, 0x1002}},
    {29585, 2, {0x0, 0x1F0, 0xFD10, 0x1110, 0x1110, 0x1110, 0x1110, 0x7D10, 0x1110, 0x1110, 0x1110, 0x1112, 0x1D12, 0xE212, 0x420E, 0x400}},
    {29586, 2, {0x0, 0x0, 0xFDFC, 0x1020, 0x1020, 0x1020, 0x7C20, 0x1020, 0x1020, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x3FE, 0x0}},
    {29587, 2, {0x40, 0x40, 0xFE80, 0x10FC, 0x1104, 0x1204, 0x7C04, 0x1104, 0x1084, 0x1044, 0x1044, 0x1E04, 0xF004, 0x4004, 0x28, 0x10}},
    {29588, 2, {0x104, 0x124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x7D24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D24, 0xE124, 0x4224, 0x204, 0x404}},
    {29589, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x1020, 0x1020, 0x7C20, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {29590, 2, {0x80, 0x80, 0xFC80, 0x10F8, 0x1108, 0x1108, 0x7D10, 0x1210, 0x1020, 0x1020, 0x1050, 0x1C50, 0xE088, 0x4108, 0x204, 0x402}},
    {29591, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x1020, 0x1020, 0x7C20, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0xA0, 0x40}},
    {29592, 2, {0x0, 0x0, 0xFDF8, 0x1008, 0x1008, 0x1008, 0x1008, 0x7DF8, 0x1108, 0x1100, 0x1100, 0x1100, 0x1D02, 0xE102, 0x40FE, 0x0}},
    {29593, 2, {0x80, 0x80, 0xFC80, 0x10FE, 0x1080, 0x1100, 0x1100, 0x7DFC, 0x1004, 0x1004, 0x1004, 0x13F4, 0x1C04, 0xE004, 0x4028, 0x10}},
    {29594, 2, {0x0, 0x1F8, 0xFC10, 0x1020, 0x1040, 0x1080, 0x7DFE, 0x1092, 0x1092, 0x1092, 0x1112, 0x1D22, 0xE222, 0x4442, 0x94, 0x108}},
    {29595, 2, {0x0, 0x3F8, 0xFC08, 0x1008, 0x1108, 0x1108, 0x1108, 0x7DFE, 0x1002, 0x1002, 0x1002, 0x13FA, 0x1C02, 0xE002, 0x4014, 0x8}},
    {29596, 2, {0x10, 0x90, 0xFC90, 0x1088, 0x1088, 0x1124, 0x7D24, 0x1222, 0x1440, 0x1040, 0x1088, 0x1C84, 0xE104, 0x43FE, 0x102, 0x0}},
    {29597, 2, {0x80, 0x80, 0xFC80, 0x11FC, 0x1120, 0x1220, 0x7C20, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {29598, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x11FC, 0x1020, 0x7C20, 0x1020, 0x13FE, 0x1020, 0x1050, 0x1C50, 0xE088, 0x4088, 0x104, 0x202}},
    {29599, 2, {0x40, 0x20, 0xFC20, 0x13FE, 0x1088, 0x1088, 0x1088, 0x7C88, 0x1088, 0x1050, 0x1050, 0x1020, 0x1C50, 0xE088, 0x4104, 0x602}},
    {29600, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x1104, 0x7E02, 0x1088, 0x1088, 0x1088, 0x1088, 0x1C88, 0xE088, 0x4108, 0x108, 0x208}},
    {29601, 2, {0x0, 0x1FE, 0xFE08, 0x1088, 0x1088, 0x1088, 0x7D08, 0x11FE, 0x1018, 0x1028, 0x1048, 0x1E88, 0xF108, 0x4208, 0x28, 0x10}},
    {29602, 2, {0x10, 0x90, 0xFC90, 0x1088, 0x1108, 0x1104, 0x1204, 0x7DFA, 0x1088, 0x1088, 0x1088, 0x1088, 0x1D08, 0xE108, 0x4228, 0x410}},
    {29603, 2, {0x40, 0x20, 0xFE20, 0x1000, 0x13FE, 0x1040, 0x7C40, 0x1060, 0x1050, 0x1048, 0x1044, 0x1E44, 0xF040, 0x4040, 0x40, 0x40}},
    {29604, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x1020, 0x1020, 0x7DFC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20}},
    {29605, 2, {0x0, 0xFC, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x7C84, 0x1084, 0x1084, 0x10FC, 0x1084, 0x1E84, 0xF104, 0x4104, 0x214, 0x408}},
    {29606, 2, {0x40, 0x40, 0xFC40, 0x13F8, 0x1048, 0x1048, 0x7C48, 0x1048, 0x17FE, 0x1040, 0x10A0, 0x1CA0, 0xE110, 0x4110, 0x208, 0x406}},
    {29607, 2, {0x40, 0x40, 0xFC80, 0x1088, 0x1104, 0x13FE, 0x7C02, 0x1090, 0x1090, 0x1090, 0x1090, 0x1C90, 0xE112, 0x4112, 0x20E, 0x400}},
    {29608, 2, {0x0, 0x0, 0xFDFC, 0x1020, 0x1020, 0x1020, 0x7C20, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x3FE, 0x0}},
    {29609, 2, {0x0, 0x0, 0xFDFC, 0x1000, 0x1000, 0x1000, 0x13FE, 0x7C90, 0x1090, 0x1090, 0x1090, 0x1090, 0x1D12, 0xE112, 0x420E, 0x400}},
    {29610, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x1144, 0x1222, 0x7C20, 0x1000, 0x11FC, 0x1004, 0x1008, 0x1C08, 0xE010, 0x4010, 0x20}},
    {29611, 2, {0x40, 0x40, 0xFE40, 0x1080, 0x10FE, 0x1108, 0x7E88, 0x1088, 0x1088, 0x1050, 0x1050, 0x1E20, 0xF050, 0x4088, 0x104, 0x202}},
    {29612, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x1148, 0x1128, 0x7D28, 0x1108, 0x17FE, 0x1108, 0x1108, 0x1D08, 0xE108, 0x4208, 0x228, 0x410}},
    {29613, 2, {0x10, 0x110, 0xFD10, 0x1112, 0x1112, 0x1114, 0x7DD8, 0x1110, 0x1110, 0x1110, 0x1110, 0x1D12, 0xE152, 0x4192, 0x10E, 0x0}},
    {29614, 2, {0x40, 0x40, 0xFC40, 0x13FE, 0x1040, 0x1040, 0x11FC, 0x7C40, 0x1040, 0x13FE, 0x1042, 0x1042, 0x1C4A, 0xE044, 0x4040, 0x40}},
    {29615, 2, {0x0, 0x0, 0xFDFE, 0x1010, 0x1010, 0x1020, 0x1020, 0x7C68, 0x10A4, 0x1122, 0x1222, 0x1020, 0x1C20, 0xE020, 0x4020, 0x20}},
    {29616, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1124, 0x1124, 0x1124, 0x7D24, 0x1124, 0x1154, 0x1050, 0x1090, 0x1C90, 0xE112, 0x4212, 0x40E}},
    {29617, 2, {0x20, 0x20, 0xFE50, 0x1050, 0x1088, 0x1104, 0x7E02, 0x10F8, 0x1088, 0x1088, 0x10A8, 0x1E90, 0xF082, 0x4082, 0x7E, 0x0}},
    {29618, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x1124, 0x1212, 0x7C10, 0x11FC, 0x1004, 0x1008, 0x1088, 0x1C50, 0xE020, 0x4010, 0x10}},
    {29619, 2, {0x50, 0x54, 0xFC52, 0x1092, 0x1090, 0x119E, 0x11F0, 0x7E90, 0x1090, 0x1090, 0x1090, 0x1090, 0x1C8A, 0xE08A, 0x4086, 0x82}},
    {29620, 2, {0x48, 0x148, 0xFD48, 0x1148, 0x1148, 0x13FE, 0x7D48, 0x1148, 0x1148, 0x1148, 0x1178, 0x1D00, 0xE100, 0x4100, 0x1FE, 0x0}},
    {29621, 2, {0x88, 0x88, 0xFE88, 0x1088, 0x13FE, 0x1088, 0x7C88, 0x1088, 0x1088, 0x10F8, 0x1088, 0x1E88, 0xF088, 0x4088, 0xF8, 0x88}},
    {29622, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x1124, 0x10A4, 0x7CA8, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {29623, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x103E, 0x1020, 0x1020, 0x7C20, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {29624, 2, {0x80, 0x80, 0xFDFC, 0x1104, 0x1204, 0x15F4, 0x1114, 0x7D14, 0x1114, 0x11F4, 0x1104, 0x1128, 0x1D12, 0xE102, 0x40FE, 0x0}},
    {29625, 2, {0x40, 0x20, 0xFE20, 0x11FE, 0x1020, 0x1020, 0x7C44, 0x1084, 0x11F8, 0x1010, 0x1020, 0x1E44, 0xF082, 0x41FE, 0x82, 0x0}},
    {29626, 2, {0x800, 0x800, 0x1FFC, 0x2104, 0x4108, 0x920, 0x1110, 0x2508, 0x200, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {29627, 2, {0x20, 0x20, 0xF820, 0x23FE, 0x2222, 0x2224, 0x2220, 0xFBFC, 0x2284, 0x2288, 0x2248, 0x3A50, 0xE220, 0x4450, 0x488, 0x906}},
    {29628, 2, {0x90, 0x90, 0xF890, 0x2290, 0x2292, 0x22D4, 0xFA98, 0x2290, 0x2290, 0x2290, 0x2290, 0x3A92, 0xE3D2, 0x4E12, 0x40E, 0x0}},
    {29629, 2, {0x80, 0x80, 0xFD00, 0x11FC, 0x1204, 0x1404, 0x7DE4, 0x1124, 0x1124, 0x1124, 0x11E4, 0x1D24, 0xE004, 0x4004, 0x28, 0x10}},
    {29630, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x1124, 0x11FC, 0x7D24, 0x1124, 0x1124, 0x11FC, 0x1124, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {29631, 2, {0x0, 0x1FC, 0xFE44, 0x1044, 0x1044, 0x1044, 0x7C94, 0x1088, 0x1100, 0x10FC, 0x1084, 0x1E84, 0xF084, 0x4084, 0xFC, 0x84}},
    {29632, 2, {0x20, 0x20, 0xFC40, 0x11FC, 0x1104, 0x1104, 0x1104, 0x7D04, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {29633, 2, {0x0, 0x0, 0xFBFC, 0x2100, 0x2100, 0x21F0, 0x2110, 0xF910, 0x2190, 0x2250, 0x2250, 0x3A10, 0xE212, 0x4292, 0x312, 0x20E}},
    {29634, 2, {0x0, 0x0, 0xFDFE, 0x1008, 0x1008, 0x11E8, 0x1128, 0x7D28, 0x1128, 0x1128, 0x11E8, 0x1128, 0x1C08, 0xE008, 0x4028, 0x10}},
    {29635, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x7D24, 0x1124, 0x1124, 0x13FE, 0x1104, 0x1D04, 0xE104, 0x4114, 0x108}},
    {29636, 2, {0x20, 0x20, 0xFD20, 0x1120, 0x11FC, 0x1120, 0x1220, 0x7C20, 0x1020, 0x11FC, 0x1020, 0x1020, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {29637, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1124, 0x1124, 0x7D24, 0x11FC, 0x1124, 0x1124, 0x1124, 0x1DFC, 0xE124, 0x4020, 0x20, 0x20}},
    {29638, 2, {0x20, 0x20, 0xFC40, 0x1088, 0x1104, 0x13FE, 0x7C02, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1D04, 0xE104, 0x4104, 0x1FC, 0x104}},
    {29639, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x1108, 0x11F8, 0x7D08, 0x1108, 0x1108, 0x11F8, 0x1108, 0x1D08, 0xE108, 0x4108, 0x7FE, 0x0}},
    {29640, 2, {0x200, 0x200, 0xFA00, 0x221E, 0x27D2, 0x2252, 0x2252, 0xFA52, 0x2252, 0x2252, 0x2252, 0x3A52, 0xE452, 0x445E, 0x952, 0x1080}},
    {29641, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1104, 0x11FC, 0x1120, 0x7D20, 0x11FE, 0x1120, 0x1120, 0x1110, 0x1D12, 0xE14A, 0x4186, 0x102}},
    {29642, 2, {0x0, 0xF7BC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0xF4A4, 0x2FFE, 0x24A4, 0x24A4, 0x24A4, 0x34A4, 0xE4A4, 0x45A4, 0x854, 0x1088}},
    {29643, 2, {0x0, 0x80, 0xFB1E, 0x2252, 0x2252, 0x2252, 0x2252, 0xFA52, 0x2252, 0x22D2, 0x235A, 0x3A54, 0xE090, 0x4090, 0x110, 0x210}},
    {29644, 2, {0x0, 0x40, 0xF820, 0x2028, 0x2008, 0x2088, 0x2090, 0xFA94, 0x22A2, 0x22A2, 0x24C2, 0x3888, 0xE188, 0x4288, 0x478, 0x0}},
    {29645, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x1212, 0x7C20, 0x1040, 0x1188, 0x1010, 0x1020, 0x1C44, 0xE188, 0x4010, 0x60, 0x380}},
    {29646, 2, {0x80, 0x80, 0xFC80, 0x10FE, 0x1102, 0x1104, 0x7E20, 0x1020, 0x10A8, 0x10A4, 0x1124, 0x1D22, 0xE222, 0x4020, 0xA0, 0x40}},
    {29647, 2, {0x0, 0x0, 0xFDFC, 0x1020, 0x1020, 0x1020, 0x1020, 0x7C20, 0x11FC, 0x1020, 0x1028, 0x1024, 0x1C24, 0xE020, 0x43FE, 0x0}},
    {29648, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1020, 0x1020, 0x1020, 0x7DFE, 0x1020, 0x1020, 0x1040, 0x1048, 0x1C84, 0xE1FE, 0x4082, 0x0}},
    {29649, 2, {0x90, 0x88, 0xFC88, 0x1080, 0x13FE, 0x10A0, 0x10A0, 0x7CA4, 0x10A4, 0x10A8, 0x1128, 0x1132, 0x1D22, 0xE262, 0x429E, 0x400}},
    {29650, 2, {0x20, 0x20, 0xFDFC, 0x1024, 0x13FE, 0x1024, 0x7DFC, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {29651, 2, {0x40, 0x20, 0xFC20, 0x13FE, 0x1000, 0x1088, 0x7D04, 0x1202, 0x1088, 0x1088, 0x1050, 0x1C50, 0xE020, 0x4050, 0x88, 0x306}},
    {29652, 2, {0x40, 0x40, 0xF840, 0x27FE, 0x2080, 0x2080, 0xF97C, 0x2108, 0x2310, 0x2510, 0x21FE, 0x3910, 0xE110, 0x4110, 0x150, 0x120}},
    {29653, 2, {0x40, 0x40, 0xFDFC, 0x1044, 0x1084, 0x1084, 0x7D28, 0x1210, 0x1108, 0x13DE, 0x114A, 0x1D4A, 0xE14A, 0x4252, 0x35A, 0x4A4}},
    {29654, 2, {0x20, 0x20, 0xFD24, 0x10A4, 0x10A8, 0x1020, 0x13FE, 0x7C90, 0x1090, 0x1090, 0x1090, 0x1092, 0x1D12, 0xE112, 0x420E, 0x400}},
    {29655, 2, {0x20, 0x120, 0xFD20, 0x11FC, 0x1120, 0x1220, 0x7C20, 0x13FE, 0x1090, 0x1090, 0x1090, 0x1C90, 0xE112, 0x4112, 0x212, 0x40E}},
    {29656, 2, {0x20, 0x40, 0xFDF8, 0x1108, 0x1148, 0x1128, 0x7D28, 0x13FE, 0x1108, 0x1148, 0x1128, 0x1D28, 0xE108, 0x4208, 0x228, 0x410}},
    {29657, 2, {0x88, 0x88, 0xFC88, 0x1088, 0x13FE, 0x1088, 0x1088, 0x7C88, 0x1088, 0x17FE, 0x1000, 0x1088, 0x1C84, 0xE104, 0x4202, 0x402}},
    {29658, 2, {0x0, 0x1FC, 0xFD04, 0x1124, 0x1124, 0x1124, 0x7DFC, 0x1124, 0x1124, 0x1154, 0x114C, 0x1D8C, 0xE104, 0x4104, 0x1FC, 0x104}},
    {29659, 2, {0x20, 0x20, 0xFDFE, 0x1040, 0x1040, 0x10FC, 0x7C84, 0x1184, 0x12FC, 0x1084, 0x1084, 0x1CFC, 0xE084, 0x4084, 0x94, 0x88}},
    {29660, 2, {0x104, 0x84, 0xFC88, 0x1000, 0x13FE, 0x1020, 0x7C20, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {29661, 2, {0x0, 0x3DE, 0xF842, 0x2042, 0x2252, 0x214A, 0xF94A, 0x2042, 0x20C6, 0x214A, 0x2252, 0x3842, 0xE042, 0x4042, 0x14A, 0x84}},
    {29662, 2, {0x80, 0x80, 0xF8F8, 0x2108, 0x2310, 0x24A0, 0x2040, 0xF8A0, 0x2118, 0x2206, 0x25F8, 0x2108, 0x3908, 0xE108, 0x1F8, 0x108}},
    {29663, 2, {0x0, 0x7F8, 0xF408, 0x2408, 0x27F8, 0x2488, 0x2488, 0xF4E8, 0x2528, 0x2528, 0x26A8, 0x244A, 0x344A, 0xC88A, 0x906, 0x1202}},
    {29664, 2, {0x20, 0x120, 0xFD20, 0x11FC, 0x1120, 0x1220, 0x7C20, 0x13FE, 0x1070, 0x10A8, 0x10A8, 0x1D24, 0xE124, 0x4222, 0x20, 0x20}},
    {29665, 2, {0x0, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x100, 0x100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {29666, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x11F8, 0x1108, 0x7D08, 0x11F8, 0x1144, 0x1148, 0x1130, 0x1D20, 0xE110, 0x4148, 0x186, 0x100}},
    {29667, 2, {0x80, 0x80, 0xFDFC, 0x1104, 0x1204, 0x11E4, 0x7D24, 0x1124, 0x11E4, 0x1124, 0x1124, 0x1DE4, 0xE104, 0x4004, 0x28, 0x10}},
    {29668, 2, {0x80, 0x80, 0xFC80, 0x11FC, 0x1120, 0x1220, 0x7C20, 0x1020, 0x17FE, 0x1020, 0x1124, 0x1D24, 0xE124, 0x4124, 0x1FC, 0x4}},
    {29669, 2, {0x0, 0x3FE, 0xFD08, 0x1108, 0x1108, 0x11F8, 0x7D08, 0x1108, 0x11F8, 0x1108, 0x1108, 0x1D1E, 0xE3E8, 0x4008, 0x8, 0x8}},
    {29670, 2, {0x20, 0x20, 0xFC40, 0x11FE, 0x1102, 0x1102, 0x7D7A, 0x114A, 0x114A, 0x114A, 0x114A, 0x1D7A, 0xE102, 0x4102, 0x10A, 0x104}},
    {29671, 2, {0x90, 0x90, 0xF890, 0x2292, 0x2194, 0x2098, 0x2090, 0xF998, 0x2294, 0x2492, 0x2090, 0x2090, 0x3912, 0xE112, 0x212, 0x40E}},
    {29672, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x1202, 0x11FC, 0x7C00, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {29673, 2, {0x80, 0x80, 0xF93C, 0x2200, 0x2480, 0x2080, 0x217E, 0xFB08, 0x2508, 0x2108, 0x2108, 0x3908, 0xE108, 0x4108, 0x128, 0x110}},
    {29674, 2, {0x20, 0x20, 0xFDFC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x7C00, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {29675, 2, {0x40, 0x20, 0xFC20, 0x13FE, 0x1040, 0x1088, 0x7D04, 0x13FE, 0x1092, 0x1090, 0x1090, 0x1C90, 0xE112, 0x4112, 0x20E, 0x400}},
    {29676, 2, {0x28, 0x24, 0xF824, 0x2020, 0x23FE, 0x2220, 0x2224, 0xFA24, 0x2224, 0x23A8, 0x2228, 0x3A10, 0xE212, 0x442A, 0x446, 0x882}},
    {29677, 2, {0x80, 0x80, 0xF8BE, 0x2288, 0x2288, 0x2288, 0x2288, 0xFABE, 0x2288, 0x2488, 0x2088, 0x3908, 0xE108, 0x423E, 0x400, 0x800}},
    {29678, 2, {0x0, 0x7FC, 0xF404, 0x2404, 0x25F4, 0x2444, 0x2444, 0xF5F4, 0x2554, 0x2554, 0x2554, 0x2554, 0x3574, 0xC446, 0x446, 0x842}},
    {29679, 2, {0x40, 0x40, 0xFC44, 0x11F4, 0x1048, 0x1050, 0x13FE, 0x7C40, 0x1080, 0x1184, 0x1298, 0x10E0, 0x1C82, 0xE082, 0x407E, 0x0}},
    {29680, 2, {0x20, 0x124, 0xFCA4, 0x10A4, 0x10A8, 0x1020, 0x11FC, 0x7C04, 0x1004, 0x1004, 0x11FC, 0x1004, 0x1C04, 0xE004, 0x41FC, 0x4}},
    {29681, 2, {0x40, 0x24, 0xFA24, 0x2108, 0x2108, 0x2010, 0x2040, 0xF840, 0x23FE, 0x2088, 0x2108, 0x3990, 0xE060, 0x4050, 0x188, 0x604}},
    {29682, 2, {0x0, 0x3FE, 0xFA02, 0x2444, 0x2040, 0x23FC, 0x2080, 0xF8A0, 0x2120, 0x21FC, 0x2020, 0x3820, 0xE3FE, 0x4020, 0x20, 0x20}},
    {29683, 2, {0x204, 0x104, 0xF908, 0x2FE8, 0x2294, 0x22A4, 0x2288, 0xFA8A, 0x2292, 0x2124, 0x2104, 0x3908, 0xE290, 0x4240, 0x43E, 0x800}},
    {29684, 2, {0x10, 0xD8, 0xFB94, 0x2094, 0x2090, 0x23FE, 0xF890, 0x2094, 0x2094, 0x20D8, 0x2398, 0x3890, 0xE0AA, 0x40CA, 0x286, 0x102}},
    {29685, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1104, 0x11FC, 0x7C00, 0x1000, 0x13FE, 0x1020, 0x1020, 0x1DFC, 0xE020, 0x4020, 0x3FE, 0x0}},
    {29686, 2, {0x108, 0x88, 0xFC90, 0x11FC, 0x1024, 0x1024, 0x11FC, 0x7D20, 0x1120, 0x11FE, 0x1062, 0x10A2, 0x1D2A, 0xE224, 0x4420, 0x20}},
    {29687, 2, {0x10, 0x18, 0xFBD4, 0x2014, 0x2010, 0x27FE, 0x2010, 0xF890, 0x2090, 0x22D0, 0x2290, 0x3A90, 0xE28A, 0x42EA, 0x706, 0x202}},
    {29688, 2, {0x0, 0x3FE, 0xF840, 0x2040, 0x21FC, 0x2084, 0x2084, 0xFBFE, 0x2000, 0x2000, 0x21FC, 0x3904, 0xE104, 0x4104, 0x1FC, 0x104}},
    {29689, 2, {0x28, 0x24, 0xF824, 0x2020, 0x23FE, 0x2220, 0x2224, 0xFA24, 0x23A4, 0x22A8, 0x22A8, 0x3A90, 0xE292, 0x45AA, 0x446, 0x882}},
    {29690, 2, {0x0, 0x1FC, 0xFC44, 0x1044, 0x13FE, 0x1044, 0x7C44, 0x11FC, 0x1080, 0x1080, 0x11FC, 0x1E84, 0xE484, 0x4084, 0xFC, 0x84}},
    {29691, 2, {0x80, 0x80, 0xFCFE, 0x1100, 0x12FC, 0x1084, 0x7CA4, 0x1094, 0x13FE, 0x1084, 0x1124, 0x1D14, 0xE1FE, 0x4004, 0x28, 0x10}},
    {29692, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x7D04, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1C00, 0xE088, 0x4084, 0x102, 0x202}},
    {29693, 2, {0x0, 0x6, 0xFBB8, 0x2088, 0x2088, 0x2108, 0x213E, 0xFB88, 0x2088, 0x2288, 0x2288, 0x393E, 0xE100, 0x4280, 0x47E, 0x800}},
    {29694, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x7DFC, 0x1104, 0x1104, 0x11FC, 0x1050, 0x1C90, 0xE092, 0x4112, 0x20E}},
    {29695, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1104, 0x11FC, 0x7C20, 0x1020, 0x1120, 0x113C, 0x1120, 0x1D20, 0xE2A0, 0x4260, 0x43E, 0x800}},
    {29696, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1144, 0x1222, 0x10F8, 0x7C08, 0x1010, 0x1020, 0x11FC, 0x1104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {29697, 2, {0x100, 0x100, 0xFDFE, 0x1200, 0x1400, 0x11FE, 0x7C22, 0x1024, 0x1120, 0x113C, 0x1120, 0x1D20, 0xE120, 0x42A0, 0x47E, 0x800}},
    {29698, 2, {0x20, 0x10, 0xFDFE, 0x1000, 0x1000, 0x10FC, 0x1000, 0x7C00, 0x10FC, 0x1000, 0x1000, 0x10FC, 0x1C84, 0xE084, 0x40FC, 0x84}},
    {29699, 2, {0x28, 0x24, 0xFC24, 0x1020, 0x13FE, 0x1020, 0x1120, 0x7CB2, 0x10B4, 0x1068, 0x10A8, 0x1124, 0x1E22, 0xE020, 0x40A0, 0x40}},
    {29700, 2, {0x0, 0xFC, 0xFC84, 0x1084, 0x10FC, 0x1000, 0x7DFE, 0x1102, 0x1102, 0x11FE, 0x1102, 0x1D02, 0xE1FE, 0x4102, 0x10A, 0x104}},
    {29701, 2, {0x40, 0x20, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x7D04, 0x11FC, 0x1120, 0x1122, 0x1114, 0x1D08, 0xE144, 0x4182, 0x100}},
    {29702, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x11FC, 0x1124, 0x1124, 0x7DFC, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {29703, 2, {0x1C, 0x1E0, 0xFC20, 0x1020, 0x13FE, 0x10A8, 0x7D24, 0x1222, 0x11F8, 0x1088, 0x1090, 0x1CBE, 0xE102, 0x4102, 0x214, 0x408}},
    {29704, 2, {0x8, 0x3C, 0xFDC0, 0x1004, 0x1144, 0x10A8, 0x1000, 0x7DF8, 0x1010, 0x1020, 0x13FE, 0x1020, 0x1C20, 0xE020, 0x40A0, 0x40}},
    {29705, 2, {0x40, 0x20, 0xFBFC, 0x2040, 0x2090, 0x2108, 0x23FC, 0xF804, 0x2150, 0x2150, 0x2150, 0x3950, 0xE250, 0x4252, 0x452, 0x80E}},
    {29706, 2, {0x0, 0x7DE, 0xF892, 0x2492, 0x2494, 0x2494, 0x27D8, 0xF894, 0x2192, 0x2192, 0x2292, 0x3A9A, 0xE494, 0x4890, 0x290, 0x110}},
    {29707, 2, {0x184, 0x68, 0xF830, 0x20C8, 0x2304, 0x2040, 0xFBFE, 0x20A0, 0x2120, 0x23FC, 0x2524, 0x3924, 0xE134, 0x4128, 0x20, 0x20}},
    {29708, 2, {0x20, 0x124, 0xFD24, 0x1124, 0x11FC, 0x1020, 0x7C50, 0x1088, 0x1144, 0x1222, 0x1020, 0x1DF8, 0xE008, 0x4010, 0x10, 0x20}},
    {29709, 2, {0x42, 0xE2, 0xFB82, 0x208A, 0x208A, 0x208A, 0x27EA, 0xF88A, 0x218A, 0x21CA, 0x22AA, 0x3A82, 0xE482, 0x4082, 0x8A, 0x84}},
    {29710, 2, {0x24, 0x224, 0xFD24, 0x117E, 0x1024, 0x1024, 0x7C24, 0x137E, 0x1124, 0x1124, 0x1124, 0x1D24, 0xE144, 0x4100, 0x2FE, 0x400}},
    {29711, 2, {0x20, 0x420, 0xFA20, 0x22FE, 0x2040, 0x2050, 0x2E90, 0xFAFE, 0x2210, 0x2210, 0x23FE, 0x3A10, 0xE210, 0x4510, 0x8FE, 0x0}},
    {29712, 2, {0x20, 0x124, 0xFCA4, 0x10A8, 0x1020, 0x11FC, 0x1104, 0x7D24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1C50, 0xE088, 0x4104, 0x202}},
    {29713, 2, {0x20, 0x124, 0xFCA4, 0x10A8, 0x1020, 0x11FC, 0x7D04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1DFC, 0xE104, 0x4104, 0x114, 0x108}},
    {29714, 2, {0x40, 0x40, 0xFCFC, 0x1088, 0x1150, 0x1020, 0x7CD8, 0x1326, 0x10F8, 0x1020, 0x10F8, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {29715, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2404, 0x21F8, 0x2000, 0xF800, 0x23FE, 0x2090, 0x2090, 0x3890, 0xE112, 0x4112, 0x20E, 0x400}},
    {29716, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2404, 0x2000, 0x23FE, 0xF820, 0x2020, 0x2120, 0x213C, 0x3920, 0xE2A0, 0x4260, 0x43E, 0x800}},
    {29717, 2, {0x40, 0x80, 0xFDFC, 0x1124, 0x1124, 0x11FC, 0x1124, 0x7D44, 0x11FC, 0x1090, 0x1110, 0x13FE, 0x1C10, 0xE010, 0x4010, 0x10}},
    {29718, 2, {0x48, 0x44, 0xFC5E, 0x11E0, 0x1028, 0x1012, 0x7C6A, 0x1196, 0x1048, 0x105E, 0x11E0, 0x1C24, 0xE028, 0x4012, 0x6A, 0x186}},
    {29719, 2, {0x40, 0x20, 0xFDFE, 0x1000, 0x1088, 0x1088, 0x7D54, 0x1222, 0x1000, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20, 0x20}},
    {29720, 2, {0x1C, 0x3E0, 0xFA20, 0x23FE, 0x2220, 0x2292, 0x230A, 0xFA06, 0x21FC, 0x2104, 0x2104, 0x39FC, 0xE104, 0x4104, 0x1FC, 0x104}},
    {29721, 2, {0x14, 0x12, 0xF810, 0x27FE, 0x2010, 0x2010, 0xFBD2, 0x2252, 0x2252, 0x2254, 0x23D4, 0x3808, 0xE0EA, 0x471A, 0x226, 0x42}},
    {29722, 2, {0x0, 0x3FC, 0xFA04, 0x2204, 0x23FC, 0x2220, 0x2220, 0xFBFE, 0x2220, 0x2220, 0x22FC, 0x3A84, 0xE284, 0x4484, 0x4FC, 0x884}},
    {29723, 2, {0x0, 0x7FC, 0xFC04, 0x24A4, 0x2110, 0x2208, 0x2040, 0xF840, 0x27FC, 0x2040, 0x20E0, 0x3950, 0xE248, 0x4C46, 0x40, 0x40}},
    {29724, 2, {0x40, 0x40, 0xF840, 0x27FC, 0x2040, 0x2248, 0x2248, 0xFA48, 0x2554, 0x20E0, 0x2150, 0x3950, 0xE248, 0x4444, 0x842, 0x40}},
    {29725, 2, {0x1FC, 0x104, 0xFD04, 0x11FC, 0x1104, 0x1104, 0x7DFC, 0x1040, 0x1020, 0x13FE, 0x1088, 0x1C88, 0xE050, 0x4020, 0xD8, 0x706}},
    {29726, 2, {0x7CFC, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7C84, 0x114, 0x208, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {29727, 2, {0xA0, 0x90, 0xFC90, 0x11FE, 0x1110, 0x1310, 0x7DFE, 0x1110, 0x1110, 0x11FE, 0x1110, 0x1D10, 0xE110, 0x41FE, 0x100, 0x100}},
    {29728, 2, {0x50, 0x50, 0xFC50, 0x11FC, 0x1154, 0x1154, 0x1154, 0x7DFC, 0x1154, 0x1154, 0x1154, 0x13FE, 0x1C00, 0xE088, 0x4104, 0x202}},
    {29729, 2, {0x100, 0x100, 0xF9DC, 0x2114, 0x2114, 0x2114, 0x27D4, 0xF814, 0x2114, 0x2114, 0x2588, 0x3D48, 0xE948, 0x4114, 0x514, 0x222}},
    {29730, 2, {0x0, 0x3FE, 0xF840, 0x2080, 0x2144, 0x2224, 0x2068, 0xFAB0, 0x2130, 0x22A8, 0x2068, 0x38A4, 0xE122, 0x4220, 0xA0, 0x40}},
    {29731, 2, {0x40, 0x20, 0xF9FC, 0x2000, 0x2108, 0x2090, 0x2000, 0xFBFE, 0x2000, 0x2000, 0x21FC, 0x3904, 0xE104, 0x4104, 0x1FC, 0x104}},
    {29732, 2, {0x80, 0x80, 0xFDF8, 0x1108, 0x1210, 0x15FC, 0x7C24, 0x1024, 0x13FE, 0x1024, 0x1024, 0x1DFC, 0xE024, 0x4020, 0xA0, 0x40}},
    {29733, 2, {0x40, 0x40, 0xF87C, 0x2040, 0x23FC, 0x2244, 0x2270, 0xFBC0, 0x2244, 0x223C, 0x2200, 0x3AF0, 0xE490, 0x4492, 0x912, 0x120E}},
    {29734, 2, {0x40, 0x40, 0xFBFC, 0x20A0, 0x2110, 0x2208, 0x27FE, 0xF808, 0x23C8, 0x2248, 0x2248, 0x3A48, 0xE3C8, 0x4008, 0x28, 0x10}},
    {29735, 2, {0x0, 0x7FFC, 0x440, 0x3C78, 0x2008, 0x3C78, 0x440, 0xFFFE, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {29736, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x1104, 0x7DFC, 0x1000, 0x1112, 0x11D4, 0x1118, 0x1D10, 0xE152, 0x4192, 0x10E}},
    {29737, 2, {0xFC, 0x84, 0xFC84, 0x10FC, 0x1084, 0x1084, 0x7CFC, 0x1000, 0x11FE, 0x1102, 0x1102, 0x1DFE, 0xE102, 0x4102, 0x1FE, 0x102}},
    {29738, 2, {0x108, 0x108, 0xFBFC, 0x2108, 0x2108, 0x21F8, 0x2108, 0xF908, 0x21F8, 0x2108, 0x2108, 0x3BFE, 0xE000, 0x4090, 0x108, 0x204}},
    {29739, 2, {0x20, 0x20, 0xFDFC, 0x1040, 0x10F8, 0x1040, 0x13FE, 0x7C88, 0x1124, 0x1222, 0x10F8, 0x1020, 0x1DFC, 0xE020, 0x4020, 0x20}},
    {29740, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2504, 0x2100, 0x21DE, 0xFA52, 0x2252, 0x2352, 0x249A, 0x3894, 0xE110, 0x4112, 0x212, 0x40E}},
    {29741, 2, {0x0, 0x1F8, 0xFC08, 0x1008, 0x11F8, 0x1008, 0x7C08, 0x13FE, 0x1020, 0x1222, 0x1174, 0x1CA8, 0xE124, 0x4222, 0xA0, 0x40}},
    {29742, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2000, 0x21FC, 0x2000, 0xF800, 0x23FE, 0x2020, 0x2128, 0x3924, 0xE222, 0x4422, 0xA0, 0x40}},
    {29743, 2, {0x20, 0x10, 0xFDFE, 0x1102, 0x1204, 0x10F8, 0x7C88, 0x1088, 0x10F8, 0x1080, 0x1080, 0x1CFC, 0xE084, 0x4084, 0xFC, 0x84}},
    {29744, 2, {0x20, 0x122, 0xF922, 0x2224, 0x2050, 0x2088, 0x2304, 0xF822, 0x2020, 0x2124, 0x2124, 0x2228, 0x3850, 0xE088, 0x104, 0x602}},
    {29745, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x1174, 0x1124, 0x7DFC, 0x1104, 0x1174, 0x1154, 0x1154, 0x1D74, 0xE104, 0x4204, 0x214, 0x408}},
    {29746, 2, {0x90, 0x90, 0xF890, 0x239E, 0x2090, 0x2090, 0x2090, 0xFB9C, 0x2090, 0x2090, 0x2090, 0x239E, 0x3890, 0xE090, 0x90, 0x90}},
    {29747, 2, {0x110, 0x110, 0xF910, 0x2110, 0x27BC, 0x2110, 0x2110, 0xFB38, 0x23B8, 0x2554, 0x2554, 0x3992, 0xE110, 0x4110, 0x110, 0x110}},
    {29748, 2, {0x0, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x100, 0x6C0, 0x1A30, 0xE10E, 0x1FE0, 0x40, 0x80, 0x100}},
    {29749, 2, {0x0, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x0, 0x2080, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {29750, 2, {0x0, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2002, 0x2002, 0x1FFE}},
    {29751, 2, {0x108, 0x88, 0xF890, 0x23FC, 0x2040, 0x2040, 0x21F8, 0xF840, 0x2040, 0x27FE, 0x20A0, 0x38A0, 0xE120, 0x4122, 0x222, 0x41E}},
    {29752, 2, {0x20, 0x20, 0xFC3E, 0x1020, 0x11FC, 0x1104, 0x11FC, 0x7D04, 0x11FC, 0x1124, 0x1020, 0x13FE, 0x1C20, 0xE020, 0x4020, 0x20}},
    {29753, 2, {0x0, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {29754, 2, {0x210, 0x110, 0xF910, 0x247C, 0x2210, 0x2210, 0x2010, 0xF97E, 0x2110, 0x2210, 0x2620, 0x2224, 0x3A42, 0xE2FE, 0x242, 0x0}},
    {29755, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x1202, 0x10F8, 0x7C20, 0x1020, 0x11FC, 0x1020, 0x1124, 0x1CA4, 0xE0A8, 0x43FE, 0x0}},
    {29756, 2, {0x40, 0x20, 0xFBFE, 0x2000, 0x2000, 0x21FC, 0xF904, 0x2104, 0x21FC, 0x2020, 0x2128, 0x3924, 0xE222, 0x4422, 0xA0, 0x40}},
    {29757, 2, {0x20, 0x22, 0xFDFA, 0x1024, 0x1028, 0x13FE, 0x7C20, 0x1040, 0x10FC, 0x1184, 0x1284, 0x1CFC, 0xE084, 0x4084, 0xFC, 0x84}},
    {29758, 2, {0x3F8, 0x248, 0xFA48, 0x23F8, 0x2248, 0x2248, 0x23F8, 0xF8A0, 0x2110, 0x2208, 0x2516, 0x2110, 0x3910, 0xE110, 0x210, 0x410}},
    {29759, 2, {0x0, 0x3FE, 0xFA22, 0x2020, 0x23FE, 0x2020, 0xF9FC, 0x2124, 0x21FC, 0x2124, 0x21FC, 0x3820, 0xE3FE, 0x4020, 0x20, 0x20}},
    {29760, 2, {0x4, 0x1E, 0xFBE0, 0x2020, 0x21FC, 0x2124, 0x2124, 0xF9FC, 0x2020, 0x23FE, 0x2222, 0x3A2A, 0xE2FA, 0x4202, 0x20A, 0x204}},
    {29761, 2, {0x0, 0x1FE, 0xFD02, 0x117A, 0x1102, 0x117A, 0x1102, 0x7C00, 0x10FC, 0x1084, 0x10FC, 0x1084, 0x1CFC, 0xE084, 0x40FC, 0x84}},
    {29762, 2, {0x0, 0x1FE, 0xFD12, 0x1112, 0x11FE, 0x1100, 0x7D7E, 0x1142, 0x1142, 0x117E, 0x1142, 0x1D7E, 0xE242, 0x4242, 0x47E, 0x842}},
    {29763, 2, {0x40, 0x40, 0xFBFC, 0x2040, 0x21F8, 0x2080, 0x23FC, 0xF910, 0x2208, 0x25F6, 0x2110, 0x3910, 0xE1F0, 0x4110, 0x110, 0x1F0}},
    {29764, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2404, 0x21FC, 0x2000, 0xF9FC, 0x2104, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x4000, 0x3FE, 0x0}},
    {29765, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x7D04, 0x11FC, 0x1000, 0x13FE, 0x1020, 0x1D20, 0xE13C, 0x4120, 0x2A0, 0x47E}},
    {29766, 2, {0x0, 0x1FC, 0xFD04, 0x11FC, 0x1104, 0x11FC, 0x1020, 0x7D20, 0x11FE, 0x1220, 0x1020, 0x11FC, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {29767, 2, {0x20, 0x1FE, 0xFC20, 0x10FC, 0x1020, 0x11FE, 0x7C00, 0x10FC, 0x10A4, 0x1094, 0x11FE, 0x1CA4, 0xE114, 0x41FE, 0x4, 0x18}},
    {29768, 2, {0x0, 0x1FC, 0xFC48, 0x1030, 0x13FE, 0x1052, 0x7C94, 0x1110, 0x1230, 0x1020, 0x13FE, 0x1C70, 0xE0A8, 0x4124, 0x622, 0x20}},
    {29769, 2, {0x0, 0x1FC, 0xFD04, 0x11FC, 0x1110, 0x11FE, 0x7D10, 0x114A, 0x1186, 0x1000, 0x11FC, 0x1D04, 0xE1FC, 0x4104, 0x1FC, 0x104}},
    {29770, 2, {0x8, 0xC, 0xF80A, 0x23FE, 0x2208, 0x2208, 0x22E8, 0xFA0A, 0x220A, 0x22EC, 0x22AC, 0x3AA8, 0xE2EA, 0x421A, 0x226, 0x442}},
    {29771, 2, {0x40, 0x40, 0xFDFC, 0x1084, 0x13FE, 0x1000, 0x7DFC, 0x1104, 0x11FC, 0x1020, 0x13FE, 0x1C20, 0xE220, 0x43FE, 0x20, 0x20}},
    {29772, 2, {0x0, 0x3FE, 0xFC20, 0x1040, 0x11FC, 0x1154, 0x1154, 0x7D54, 0x1154, 0x112C, 0x1020, 0x13FE, 0x1C50, 0xE088, 0x4104, 0x202}},
    {29773, 2, {0x80, 0x80, 0xF8F8, 0x2108, 0x2210, 0x25FC, 0xF904, 0x2154, 0x218C, 0x2124, 0x2020, 0x3BFE, 0xE050, 0x4088, 0x104, 0x602}},
    {29774, 2, {0x110, 0x112, 0xFDD4, 0x1118, 0x1152, 0x1192, 0x7D2E, 0x1040, 0x11FC, 0x1104, 0x1104, 0x1DFC, 0xE104, 0x4104, 0x1FC, 0x104}},
    {29775, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2088, 0x2104, 0x21FC, 0xF890, 0x2110, 0x23FE, 0x2050, 0x3890, 0xE110, 0x4610, 0x50, 0x20}},
    {29776, 2, {0x108, 0x88, 0xF890, 0x27FE, 0x2000, 0x23C4, 0x2254, 0xFA54, 0x23D4, 0x2254, 0x2254, 0x3BD4, 0xE254, 0x4244, 0x254, 0x2C8}},
    {29777, 2, {0x80, 0xFC, 0xFD04, 0x11F8, 0x1008, 0x13FE, 0x1040, 0x7CA2, 0x1334, 0x1058, 0x1094, 0x1334, 0x1C52, 0xE090, 0x4350, 0x20}},
    {29778, 2, {0x0, 0xFC, 0xFC84, 0x10FC, 0x1084, 0x10FC, 0x1000, 0x7DFE, 0x1080, 0x10FE, 0x112A, 0x124A, 0x1C92, 0xE122, 0x4054, 0x88}},
    {29779, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x2124, 0xF9AC, 0x2174, 0x2124, 0x21FC, 0x2020, 0x3870, 0xE0A8, 0x4124, 0x222, 0x20}},
    {29780, 2, {0x40, 0x80, 0xFDFC, 0x1104, 0x11FC, 0x1104, 0x7DFC, 0x1020, 0x1032, 0x13B4, 0x10A8, 0x1D28, 0xE124, 0x4224, 0x4A2, 0x40}},
    {29781, 2, {0x0, 0x3DE, 0xFA42, 0x2242, 0x2242, 0x23DE, 0x2200, 0xFA3E, 0x23D2, 0x2212, 0x2214, 0x3BD4, 0xE208, 0x4214, 0x224, 0x242}},
    {29782, 2, {0x80, 0x338, 0xFA28, 0x2228, 0x23A8, 0x2246, 0x2200, 0xFA7C, 0x23A4, 0x2224, 0x2228, 0x3BA8, 0xE610, 0x4228, 0x244, 0x282}},
    {29783, 2, {0x3C, 0x7C0, 0xFA44, 0x2128, 0x2000, 0x23FC, 0x2080, 0xF880, 0x27FE, 0x2100, 0x21F8, 0x3A88, 0xE250, 0x4420, 0x8D8, 0x306}},
    {29784, 2, {0x0, 0xFDE, 0xF492, 0x2492, 0x2494, 0x2794, 0x2498, 0xF494, 0x2792, 0x2492, 0x2492, 0x25DA, 0x3E94, 0xC090, 0x90, 0x90}},
    {29785, 2, {0x92, 0x124, 0xFA48, 0x2124, 0x2092, 0x2040, 0x2080, 0xFBFE, 0x2202, 0x228A, 0x2252, 0x3A22, 0xE252, 0x428A, 0x3FE, 0x202}},
    {29786, 2, {0x100, 0x13C, 0xF924, 0x27E4, 0x2124, 0x213C, 0x2124, 0xFBA4, 0x22A4, 0x22BC, 0x22A4, 0x3AA4, 0xE3A4, 0x4044, 0x54, 0x88}},
    {29787, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20A8, 0x2020, 0x21FC, 0xF924, 0x2124, 0x2124, 0x23FE, 0x3820, 0xE050, 0x4088, 0x104, 0x202}},
    {29788, 2, {0x40, 0xA0, 0xF910, 0x2208, 0x25F6, 0x2000, 0x23C4, 0xFA54, 0x2254, 0x23D4, 0x2254, 0x3A54, 0xE3D4, 0x4244, 0x254, 0x2C8}},
    {29789, 2, {0x20, 0x40, 0xFDFC, 0x1104, 0x11FC, 0x1104, 0x7DFC, 0x1000, 0x13FE, 0x1020, 0x1020, 0x1DFC, 0xE020, 0x4020, 0x3FE, 0x0}},
    {29790, 2, {0x20, 0x222, 0xFA22, 0x2222, 0x23FE, 0x2000, 0x27FE, 0xF820, 0x2040, 0x23FE, 0x2252, 0x3A52, 0xE252, 0x4252, 0x252, 0x206}},
    {29791, 2, {0x0, 0xFEFE, 0x1010, 0x1010, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x110, 0x8A0, 0x4844, 0x4982, 0x8E12, 0x1810, 0xE7F0, 0x0}},
    {29792, 2, {0x40, 0x3BE, 0xFA12, 0x2292, 0x2252, 0x22AA, 0x2324, 0xF840, 0x23FE, 0x2222, 0x2222, 0x3BFE, 0xE222, 0x4222, 0x3FE, 0x202}},
    {29793, 2, {0x100, 0x27E, 0xFB88, 0x2288, 0x22BE, 0x22AA, 0x23AA, 0xFA2A, 0x222A, 0x23AA, 0x22AA, 0x3AAE, 0xE288, 0x4388, 0x288, 0x8}},
    {29794, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2494, 0x2108, 0x2264, 0xF890, 0x2108, 0x2204, 0x25FA, 0x3908, 0xE108, 0x4108, 0x1F8, 0x108}},
    {29795, 2, {0x20, 0x124, 0xFCA8, 0x1020, 0x11FC, 0x1104, 0x7D04, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1D04, 0xE1FC, 0x4000, 0x88, 0x104}},
    {29796, 2, {0x40, 0x80, 0xFDFC, 0x1248, 0x1130, 0x10C0, 0x1300, 0x7DFC, 0x1220, 0x1020, 0x13FE, 0x1020, 0x1D24, 0xE124, 0x41FC, 0x4}},
    {29797, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x11F8, 0x1108, 0x7D08, 0x11F8, 0x1000, 0x13FC, 0x1294, 0x1E94, 0xE294, 0x4294, 0x7FE, 0x0}},
    {29798, 2, {0x40, 0x80, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x7D00, 0x11FE, 0x1100, 0x11FE, 0x1002, 0x1EAA, 0xE2AA, 0x4402, 0x14, 0x8}},
    {29799, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x2040, 0x23FE, 0xF888, 0x2134, 0x22E2, 0x2020, 0x39FC, 0xE070, 0x40AC, 0x322, 0x20}},
    {29800, 2, {0x0, 0x3FC, 0xF890, 0x2294, 0x2198, 0x2090, 0x27FE, 0xF800, 0x21F8, 0x2108, 0x2108, 0x21F8, 0x3908, 0xE108, 0x1F8, 0x108}},
    {29801, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0x3FFC, 0x100, 0x100, 0x1FF8, 0x100, 0x120, 0x110, 0x7FFE}},
    {29802, 2, {0x0, 0x1FE, 0xFD20, 0x11FC, 0x1120, 0x11FC, 0x7D20, 0x1120, 0x11FE, 0x1002, 0x12AA, 0x1EAA, 0xE2AA, 0x4202, 0x14, 0x8}},
    {29803, 2, {0x1E, 0x1E0, 0xFC22, 0x1112, 0x1094, 0x1080, 0x7C20, 0x11CE, 0x1102, 0x1102, 0x11CE, 0x1D02, 0xE102, 0x4102, 0x1FE, 0x102}},
    {29804, 2, {0x2040, 0x17FE, 0x8090, 0x4908, 0x13FC, 0xE150, 0x2252, 0x244E, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {29805, 2, {0x20, 0x10, 0xFBFE, 0x2220, 0x22FC, 0x2224, 0x23FE, 0xFA24, 0x22FC, 0x2220, 0x22FC, 0x3A84, 0xE284, 0x4484, 0x4FC, 0x884}},
    {29806, 2, {0x0, 0x3FE, 0xF850, 0x2050, 0x23FE, 0x2252, 0x2252, 0xFBFE, 0x2020, 0x2020, 0x23FE, 0x3870, 0xE0A8, 0x4124, 0x622, 0x20}},
    {29807, 2, {0x100, 0x9E, 0xFBD2, 0x2252, 0x2254, 0x23D4, 0x2258, 0xFA54, 0x23D2, 0x2212, 0x2292, 0x3A5A, 0xE2D4, 0x4350, 0x210, 0x10}},
    {29808, 2, {0x20, 0x40, 0xFBFC, 0x2224, 0x2224, 0x23FC, 0x2224, 0xFA44, 0x23FC, 0x2040, 0x20A8, 0x38B4, 0xE13C, 0x4122, 0x222, 0x41E}},
    {29809, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x2104, 0x21FC, 0xF904, 0x21FC, 0x2104, 0x21FC, 0x3904, 0xE3FE, 0x4088, 0x104, 0x202}},
    {29810, 2, {0x40, 0x40, 0xF8A0, 0x2190, 0x2248, 0x27FE, 0xFA08, 0x23F8, 0x2208, 0x23F8, 0x2200, 0x3BF8, 0xE508, 0x4508, 0x9F8, 0x108}},
    {29811, 2, {0x84, 0x44, 0xFC48, 0x11FE, 0x1020, 0x10FC, 0x7C20, 0x11FE, 0x1040, 0x1080, 0x10FE, 0x1D10, 0xE210, 0x4410, 0x1FE, 0x0}},
    {29812, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0xFE44, 0x1028, 0x7C28, 0x1010, 0x1E28, 0xF044, 0x4182}},
    {29813, 2, {0x0, 0x1FC, 0xFCA4, 0x1088, 0x1050, 0x1020, 0x7CD8, 0x1326, 0x1020, 0x11FC, 0x1124, 0x1D24, 0xE1FC, 0x4022, 0x3FE, 0x2}},
    {29814, 2, {0x8, 0x3C, 0xFBC0, 0x2004, 0x2244, 0x2128, 0xF9FC, 0x2220, 0x2020, 0x23FE, 0x2020, 0x3924, 0xE124, 0x4124, 0x1FC, 0x4}},
    {29815, 2, {0xE, 0x3F0, 0xF844, 0x2224, 0x2108, 0x23FE, 0x2242, 0xF840, 0x23FE, 0x2080, 0x20FC, 0x3944, 0xE128, 0x4210, 0x468, 0x186}},
    {29816, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2414, 0x21E0, 0x2100, 0xF900, 0x21FC, 0x2110, 0x2110, 0x27FE, 0x3800, 0xE090, 0x108, 0x204}},
    {29817, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20A8, 0x2050, 0x2088, 0xF974, 0x2222, 0x2020, 0x21FC, 0x3820, 0xE0A8, 0x4124, 0x2A2, 0x40}},
    {29818, 2, {0x20, 0x124, 0xFCA8, 0x13FE, 0x1202, 0x10F8, 0x7C88, 0x1088, 0x10F8, 0x1020, 0x11FC, 0x1D24, 0xE124, 0x4134, 0x128, 0x20}},
    {29819, 2, {0x0, 0x1FC, 0xFD24, 0x13FE, 0x1124, 0x11FC, 0x7C00, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE104, 0x41FC, 0x88, 0x104}},
    {29820, 2, {0x20, 0x3FE, 0xFC20, 0x11FC, 0x1124, 0x11FC, 0x1124, 0x7DFC, 0x1022, 0x13FE, 0x1008, 0x13FE, 0x1D08, 0xE088, 0x4028, 0x10}},
    {29821, 2, {0x124, 0x124, 0xFA24, 0x2424, 0x2954, 0x214A, 0xFA92, 0x2610, 0x2A10, 0x2250, 0x225C, 0x3A50, 0xE250, 0x42B0, 0x29E, 0x300}},
    {29822, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20F8, 0x2020, 0x21FC, 0xF924, 0x21FC, 0x2020, 0x23FE, 0x3820, 0xE1FC, 0x4020, 0x3FE, 0x0}},
    {29823, 2, {0x0, 0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0xFFFE}},
    {29824, 2, {0x20, 0x222, 0xFBFE, 0x2090, 0x2088, 0x21FE, 0x2310, 0xFD10, 0x21FE, 0x2110, 0x2110, 0x39FE, 0xE110, 0x4110, 0x1FE, 0x100}},
    {29825, 2, {0x40, 0x80, 0xFBFC, 0x2244, 0x22F4, 0x2314, 0x22A4, 0xFA44, 0x22A4, 0x23FC, 0x2040, 0x3824, 0xE522, 0x450A, 0x908, 0xF8}},
    {29826, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20F8, 0x2088, 0x20F8, 0xF888, 0x23FE, 0x2088, 0x2124, 0x3AFA, 0xE020, 0x4020, 0x3FE, 0x0}},
    {29827, 2, {0x40, 0x20, 0xFBFE, 0x2000, 0x2154, 0x2124, 0x2154, 0xF9FC, 0x2020, 0x23FE, 0x2242, 0x3A92, 0xE2FA, 0x420A, 0x202, 0x206}},
    {29828, 2, {0x40, 0x20, 0xFDFC, 0x1088, 0x1050, 0x13FE, 0x7C00, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE090, 0x4092, 0x112, 0x60E}},
    {29829, 2, {0x124, 0x248, 0xFD24, 0x1000, 0x11FC, 0x1124, 0x7DFC, 0x1124, 0x11FC, 0x1020, 0x13FE, 0x1C70, 0xE0A8, 0x4124, 0x622, 0x20}},
    {29830, 2, {0x0, 0x7BC, 0xF884, 0x24A4, 0x2294, 0x24A4, 0x2050, 0xF988, 0x2626, 0x20C0, 0x2310, 0x3864, 0xE388, 0x4030, 0xC0, 0x700}},
    {29831, 2, {0x210, 0x110, 0xF910, 0x27BE, 0x2240, 0x2200, 0x23BE, 0xFA8A, 0x2288, 0x22A8, 0x22AE, 0x3AA8, 0xE4A8, 0x45A8, 0x85E, 0x1080}},
    {29832, 2, {0x108, 0x108, 0xF908, 0x27D0, 0x211E, 0x2794, 0xF924, 0x2FD4, 0x2214, 0x23D4, 0x2254, 0x3A48, 0xE248, 0x44D4, 0x424, 0x842}},
    {29833, 2, {0x10, 0x410, 0xFAFE, 0x2210, 0x20FE, 0x2092, 0x26FE, 0xFA92, 0x22FE, 0x2210, 0x22FE, 0x3A10, 0xE210, 0x4510, 0x8FE, 0x0}},
    {29834, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x2088, 0x20F8, 0x2020, 0xFBFE, 0x2222, 0x2332, 0x22AA, 0x3B76, 0xE222, 0x4222, 0x22A, 0x224}},
    {29835, 2, {0x20, 0x1FC, 0xFC88, 0x1050, 0x13FE, 0x1000, 0x7DFC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1C20, 0xE3FE, 0x4020, 0x20, 0x20}},
    {29836, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2000, 0x21FC, 0x2020, 0xF9FC, 0x2124, 0x21FC, 0x2124, 0x39FC, 0xE000, 0x4088, 0x104, 0x202}},
    {29837, 2, {0x88, 0x3FE, 0xF888, 0x2000, 0x21FC, 0x20A8, 0x20A8, 0xFBFE, 0x20A8, 0x20A8, 0x21FC, 0x3820, 0xE3FE, 0x4020, 0x20, 0x20}},
    {29838, 2, {0x0, 0xFBE, 0xF8A2, 0x2AAA, 0x2AAA, 0x2AAA, 0x2514, 0xF8A2, 0x2080, 0x2FFE, 0x2110, 0x3A10, 0xE3A0, 0x4060, 0x198, 0xE04}},
    {29839, 2, {0x100, 0x1F8, 0xFA08, 0x23F0, 0x2010, 0x2FFE, 0x2080, 0xFC88, 0x24EA, 0x274A, 0x244C, 0x3DE8, 0xE44A, 0x464A, 0x4A6, 0x110}},
    {29840, 2, {0x10, 0x790, 0xFCBC, 0x24A4, 0x24C4, 0x27A8, 0x2110, 0xF928, 0x2146, 0x25C0, 0x253C, 0x3D24, 0xE524, 0x45A4, 0xE3C, 0x24}},
    {29841, 2, {0x80, 0x80, 0xF9FE, 0x2354, 0x2554, 0x2154, 0x23FE, 0xF954, 0x2154, 0x2154, 0x27FE, 0x2000, 0x3954, 0xE12A, 0x22A, 0x0}},
    {29842, 2, {0x20, 0x7A4, 0xF8A8, 0x2292, 0x2114, 0x2208, 0x25F4, 0xF802, 0x23F8, 0x2208, 0x2208, 0x3BF8, 0xE208, 0x4110, 0xFFE, 0x0}},
    {29843, 2, {0x88, 0x3FE, 0xFC88, 0x101C, 0x11E0, 0x1020, 0x13FE, 0x7CA8, 0x1124, 0x1222, 0x11F8, 0x1088, 0x1C9E, 0xE082, 0x410A, 0x204}},
    {29844, 2, {0x104, 0x88, 0xF800, 0x23FE, 0x2222, 0x22AA, 0x2272, 0xFA22, 0x23FE, 0x2000, 0x21FC, 0x3904, 0xE1FC, 0x4104, 0x1FC, 0x104}},
    {29845, 2, {0x0, 0x3FC, 0xF804, 0x23FC, 0x2004, 0x23FC, 0x2000, 0xFBBC, 0x2124, 0x27BC, 0x2008, 0x3BFE, 0xE108, 0x4088, 0xA8, 0x10}},
    {29846, 2, {0x200, 0x17E, 0xF840, 0x247C, 0x2244, 0x20FC, 0xF940, 0x267E, 0x2220, 0x2020, 0x23FE, 0x3870, 0xE0A8, 0x4124, 0x222, 0x20}},
    {29847, 2, {0x21F8, 0x1108, 0x41F8, 0x2108, 0xBFE, 0x900, 0x73FC, 0x14A4, 0x1154, 0x1288, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0xFFFE}},
    {29848, 2, {0x20, 0x124, 0xF8A8, 0x23FE, 0x20A8, 0x2124, 0x2222, 0xF904, 0x2104, 0x21DE, 0x2244, 0x2554, 0x389E, 0xE104, 0x204, 0x404}},
    {29849, 2, {0x20, 0x20, 0xFBFE, 0x2050, 0x228A, 0x2104, 0xFBFE, 0x2504, 0x21FC, 0x2104, 0x21FC, 0x3820, 0xE124, 0x4222, 0x4A2, 0x40}},
    {29850, 2, {0x0, 0x3FC, 0xF848, 0x2030, 0x23FE, 0x2052, 0x2094, 0xFB50, 0x2020, 0x23FE, 0x2252, 0x3A8A, 0xE376, 0x4252, 0x272, 0x206}},
    {29851, 2, {0x40, 0x3F8, 0xF848, 0x27FE, 0x2048, 0x23F8, 0xF840, 0x2554, 0x275C, 0x2444, 0x27FC, 0x3C44, 0xE75C, 0x4554, 0x554, 0x844}},
    {29852, 2, {0x88, 0x88, 0xFDFC, 0x1088, 0x1088, 0x13FE, 0x1020, 0x7DFC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1C00, 0xE088, 0x4104, 0x202}},
    {29853, 2, {0x20, 0x1FC, 0xFD24, 0x11FC, 0x1020, 0x13FE, 0x7C00, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE104, 0x41FC, 0x88, 0x104}},
    {29854, 2, {0x50, 0x252, 0xF954, 0x2050, 0x23FE, 0x2088, 0x2050, 0xFBFE, 0x2020, 0x21FC, 0x2020, 0x3BFE, 0xE050, 0x4088, 0x104, 0x602}},
    {29855, 2, {0x0, 0x1FC, 0xFD04, 0x11FC, 0x1104, 0x11FC, 0x7C20, 0x13FE, 0x1000, 0x11FC, 0x1104, 0x1DFC, 0xE020, 0x4124, 0x222, 0x60}},
    {29856, 2, {0x3C, 0x3E0, 0xFD24, 0x10A8, 0x13FE, 0x10A8, 0x7D24, 0x1202, 0x11FC, 0x1124, 0x1124, 0x1DFC, 0xE124, 0x4124, 0x1FC, 0x104}},
    {29857, 2, {0x28, 0x424, 0xFA7E, 0x22C8, 0x2148, 0x207E, 0x2648, 0xFA48, 0x227E, 0x2248, 0x2248, 0x3A7E, 0xE240, 0x4500, 0x8FE, 0x0}},
    {29858, 2, {0x0, 0x3FE, 0xF850, 0x23DE, 0x2252, 0x2252, 0x23DE, 0xF850, 0x23FE, 0x2222, 0x2222, 0x3BFE, 0xE222, 0x4222, 0x3FE, 0x202}},
    {29859, 2, {0x124, 0x124, 0xFAAA, 0x23AE, 0x2124, 0x22AA, 0x23AE, 0xF924, 0x27FE, 0x2110, 0x2114, 0x3914, 0xE28A, 0x424A, 0x216, 0x422}},
    {29860, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x2124, 0xF9FC, 0x2124, 0x21FC, 0x2022, 0x23FE, 0x3842, 0xE024, 0x42A2, 0x28A, 0x478}},
    {29861, 2, {0x288, 0x288, 0xF7E8, 0x2290, 0x241E, 0x27E4, 0x2454, 0xF854, 0x2754, 0x2554, 0x2554, 0x3748, 0xE548, 0x4054, 0x294, 0x122}},
    {29862, 2, {0x3C, 0x3C0, 0xF844, 0x2224, 0x2108, 0x27FE, 0x2442, 0xFAA4, 0x228A, 0x247A, 0x2080, 0x38F8, 0xE108, 0x4290, 0x60, 0x79E}},
    {29863, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x0}},
    {29864, 2, {0x80, 0xDC, 0xFA84, 0x23D4, 0x2448, 0x2A94, 0x2324, 0xFC40, 0x2248, 0x2150, 0x27FC, 0x38E0, 0xE150, 0x4248, 0xC46, 0x40}},
    {29865, 2, {0x40, 0x7C, 0xF840, 0x23FE, 0x2242, 0x2278, 0x23C4, 0xFA3C, 0x2200, 0x23FE, 0x2240, 0x3BA4, 0xE258, 0x45B4, 0x452, 0x9B0}},
    {29866, 2, {0x1FC, 0x104, 0xF904, 0x21FC, 0x2000, 0x23DE, 0x2252, 0xFA52, 0x23DE, 0x2020, 0x23FE, 0x3870, 0xE0A8, 0x4124, 0x622, 0x20}},
    {29867, 2, {0x20, 0x124, 0xFCA8, 0x13FE, 0x1202, 0x10F8, 0x1088, 0x7C88, 0x10F8, 0x1000, 0x11FC, 0x1124, 0x1DFC, 0xE124, 0x41FC, 0x104}},
    {29868, 2, {0x108, 0x208, 0xF7C8, 0x2450, 0x27DE, 0x2464, 0x27D4, 0xF214, 0x2114, 0x27D4, 0x2214, 0x23C8, 0x3248, 0xC454, 0x554, 0x8A2}},
    {29869, 2, {0x0, 0x4FE, 0xFA92, 0x227C, 0x2010, 0x207C, 0x2654, 0xFA7C, 0x2254, 0x227C, 0x2210, 0x22FE, 0x3A10, 0xE510, 0x8FE, 0x0}},
    {29870, 2, {0x20, 0x7FE, 0xF800, 0x23FE, 0x2202, 0x22FA, 0x228A, 0xFBFE, 0x2000, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x41FC, 0x0, 0x7FE}},
    {29871, 2, {0x20, 0x50, 0xFC88, 0x1174, 0x1202, 0x11FC, 0x7D24, 0x11AC, 0x1124, 0x11FC, 0x1000, 0x1CF8, 0xE088, 0x40F8, 0x88, 0xF8}},
    {29872, 2, {0x0, 0x3FC, 0xFA94, 0x2294, 0x23FC, 0x2000, 0xFBFE, 0x2000, 0x23FC, 0x2204, 0x23FC, 0x38A2, 0xE114, 0x4308, 0xD44, 0x182}},
    {29873, 2, {0x0, 0x7BE, 0xF108, 0x2108, 0x27BE, 0x2108, 0x2108, 0xF7BE, 0x2000, 0x2080, 0x2044, 0x2512, 0x3522, 0xC9C0, 0x304, 0xCFC}},
    {29874, 2, {0x88, 0x448, 0xFA50, 0x20FC, 0x2020, 0x2050, 0x2694, 0xFA38, 0x2250, 0x2298, 0x2234, 0x3A54, 0xE290, 0x4220, 0x5FE, 0x800}},
    {29875, 2, {0xA0, 0xFC, 0xFD20, 0x13FE, 0x10A8, 0x1124, 0x7E22, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE104, 0x41FC, 0x88, 0x104}},
    {29876, 2, {0x88, 0x88, 0xFBDE, 0x2088, 0x21DC, 0x22AA, 0x2488, 0xF800, 0x23FE, 0x2022, 0x2120, 0x393C, 0xE120, 0x42A0, 0x47E, 0x800}},
    {29877, 2, {0x140, 0x24C, 0xFA64, 0x2244, 0x234C, 0x2264, 0x22A4, 0xFBAC, 0x22A4, 0x22A4, 0x27FE, 0x3800, 0xE110, 0x4108, 0x204, 0x404}},
    {29878, 2, {0x40, 0x3F8, 0xF848, 0x27FE, 0x2048, 0x23F8, 0xF840, 0x27FE, 0x2000, 0x22A4, 0x2452, 0x3BF8, 0xE2A8, 0x42A8, 0xFFE, 0x0}},
    {29879, 2, {0x148, 0x7C8, 0xF908, 0x27D0, 0x255E, 0x27E4, 0xFD54, 0x27D4, 0x2554, 0x2114, 0x27D4, 0x3A08, 0xE3C8, 0x4254, 0x554, 0x8A2}},
    {29880, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x21FC, 0x2148, 0xFA50, 0x21FC, 0x2304, 0x21FC, 0x2104, 0x39FC, 0xE104, 0x41FC, 0x88, 0x104}},
    {29881, 2, {0x20, 0x3FE, 0xFC20, 0x11FC, 0x1000, 0x13FE, 0x1002, 0x7DFC, 0x1020, 0x13FE, 0x1000, 0x13FE, 0x1C04, 0xE3BE, 0x42A4, 0x38C}},
    {29882, 2, {0x2FEC, 0x4824, 0x4BA4, 0x682C, 0x4BA4, 0x6AAC, 0x4BA4, 0x4824, 0xFFFE, 0x8002, 0x7FFC, 0x100, 0x3FF8, 0x120, 0x110, 0xFFFE}},
    {29883, 2, {0x3DE, 0x42, 0xFA52, 0x214A, 0x2252, 0x2020, 0x23FE, 0xF888, 0x2088, 0x2154, 0x2222, 0x3820, 0xE7FE, 0x4020, 0x20, 0x20}},
    {29884, 2, {0x8, 0x3E8, 0xFA88, 0x23EE, 0x2228, 0x23F4, 0x2282, 0xFBE2, 0x2000, 0x21FC, 0x2154, 0x3954, 0xE154, 0x4154, 0x7FE, 0x0}},
    {29885, 2, {0x7FFC, 0x2108, 0x7FFC, 0xB55A, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3558, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0xFFFE}},
    {29886, 2, {0x40, 0x20, 0xFBFE, 0x2088, 0x2052, 0x27AC, 0x22AA, 0xFAA8, 0x25AC, 0x2000, 0x2104, 0x39FC, 0xE104, 0x41FC, 0x104, 0x204}},
    {29887, 2, {0x40, 0x7C, 0xF840, 0x27FE, 0x2402, 0x23FC, 0xFA48, 0x24A4, 0x2110, 0x23F8, 0x260C, 0x3BFA, 0xE208, 0x43F8, 0x208, 0x3F8}},
    {29888, 2, {0x0, 0x1FC, 0xFC20, 0x13FE, 0x1222, 0x11AC, 0x7C20, 0x11AC, 0x1000, 0x13FE, 0x1020, 0x1DFC, 0xE154, 0x4154, 0x154, 0x10C}},
    {29889, 2, {0x88, 0x3FE, 0xFCA8, 0x1090, 0x11FE, 0x1320, 0x7DFC, 0x1120, 0x11FC, 0x1120, 0x11FE, 0x1D00, 0xE3FC, 0x4088, 0x70, 0x38E}},
    {29890, 2, {0x88, 0x3FE, 0xFC88, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x7C40, 0x1088, 0x11FC, 0x1000, 0x11FC, 0x1D54, 0xE154, 0x47FE, 0x0}},
    {29891, 2, {0x3F8, 0x248, 0xFA48, 0x23F8, 0x2248, 0x2248, 0xFBF8, 0x2000, 0x2FBE, 0x2AAA, 0x2AAA, 0x3FBE, 0xEAAA, 0x4AAA, 0xFBE, 0x8A2}},
    {29892, 2, {0x20, 0x3FE, 0xFC20, 0x11FC, 0x1000, 0x13FE, 0x7E52, 0x13FE, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE104, 0x41FC, 0x88, 0x104}},
    {29893, 2, {0x444, 0x4E4, 0xF8A8, 0x2AAA, 0x2EEE, 0x24A4, 0x2AAA, 0xFEEE, 0x2242, 0x2040, 0x2FFE, 0x20E0, 0x3150, 0xC248, 0xC46, 0x40}},
    {29894, 2, {0x42, 0x39C, 0xFA10, 0x23DE, 0x2294, 0x2294, 0x2420, 0xFBFC, 0x2204, 0x23FC, 0x2204, 0x3BFC, 0xE204, 0x43FC, 0x108, 0x204}},
    {29895, 2, {0x3FC, 0x40, 0xF9F8, 0x2108, 0x21F8, 0x2108, 0x27FE, 0xFC42, 0x22A4, 0x228A, 0x247A, 0x3880, 0xE1F8, 0x4288, 0x70, 0x38E}},
    {29896, 2, {0x50, 0x190, 0xFC9E, 0x13EA, 0x108A, 0x11D2, 0x128A, 0x7CA4, 0x1050, 0x1088, 0x1326, 0x10A8, 0x1C70, 0xE0A8, 0x4124, 0x60}},
    {29897, 2, {0x88, 0x1DC, 0xFC88, 0x13DE, 0x1088, 0x1154, 0x7E22, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1DFC, 0xE104, 0x41FC, 0x88, 0x104}},
    {29898, 2, {0x100, 0x1F8, 0xFA08, 0x27FE, 0x2262, 0x2292, 0xF9FC, 0x2104, 0x21FC, 0x2104, 0x21FC, 0x3880, 0xE1F8, 0x4288, 0x70, 0x38E}},
    {29899, 2, {0x20, 0x10, 0xF4FE, 0x2244, 0x2228, 0x20FE, 0x2092, 0xF6FE, 0x2292, 0x22BA, 0x22AA, 0x22BA, 0x3286, 0xC500, 0x8FE, 0x0}},
    {29900, 2, {0x40, 0x20, 0xFDFE, 0x1000, 0x11FC, 0x1154, 0x11FC, 0x7C20, 0x1122, 0x108C, 0x1122, 0x1054, 0x1CC8, 0xE144, 0x4252, 0x60}},
    {29901, 2, {0x8, 0x73E, 0xF590, 0x255E, 0x2564, 0x265E, 0x2500, 0xF5DE, 0x2552, 0x255E, 0x2552, 0x265E, 0x3452, 0xC456, 0x4A0, 0x51E}},
    {29902, 2, {0x108, 0x108, 0xF7CE, 0x2112, 0x2124, 0x27DE, 0x2552, 0xF55E, 0x27D2, 0x211E, 0x2392, 0x255E, 0x3940, 0xC114, 0x112, 0x122}},
    {29903, 2, {0x110, 0x90, 0xFBDE, 0x2010, 0x225E, 0x2182, 0xFBDE, 0x2010, 0x23DE, 0x2250, 0x23DE, 0x3A50, 0xE3DE, 0x4250, 0x252, 0x2CE}},
    {29904, 2, {0x40, 0x7C, 0xF840, 0x23FC, 0x2244, 0x23F0, 0xFA44, 0x22FC, 0x22A8, 0x22F8, 0x22A8, 0x3AF8, 0xE200, 0x45FC, 0x554, 0xBFE}},
    {29905, 2, {0x0, 0x7FE, 0xF444, 0x2598, 0x2488, 0x27DE, 0x2488, 0xF5DC, 0x26AA, 0x2488, 0x2420, 0x2520, 0x353C, 0xC920, 0x920, 0x17FE}},
    {29906, 2, {0x104, 0x514, 0xF7DE, 0x2924, 0x2FFF, 0x210C, 0xF355, 0x2DA7, 0x23FC, 0x2204, 0x2224, 0x3A24, 0xE244, 0xB0, 0x108, 0x604}},
    {29907, 2, {0x3DE, 0x252, 0xFBDE, 0x2252, 0x23DE, 0x2222, 0x22FA, 0xFA22, 0x22FA, 0x22AA, 0x22FA, 0x3AAA, 0xE2FA, 0x4272, 0x2AA, 0x226}},
    {29908, 2, {0x3DE, 0x252, 0xFBDE, 0x2252, 0x23DE, 0x2252, 0x23DE, 0xF98C, 0x2252, 0x2040, 0x27FE, 0x3888, 0xE190, 0x4060, 0xD8, 0x304}},
    {29909, 2, {0xF7FC, 0x1248, 0x17FC, 0xFC46, 0x86EC, 0xF554, 0x16EC, 0x1554, 0x56EC, 0x2444, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x120, 0xFFFE}},
    {29910, 2, {0x20, 0x3FE, 0xF800, 0x21DC, 0x2154, 0x21DC, 0x2088, 0xFBFE, 0x2088, 0x23FE, 0x2088, 0x27FE, 0x3894, 0xE188, 0x6A4, 0xC2}},
    {29911, 2, {0x222, 0x3FE, 0xF890, 0x21FE, 0x2310, 0x25FE, 0x2110, 0xF9FE, 0x2110, 0x21FE, 0x2100, 0x23FE, 0x3A8A, 0xE376, 0x252, 0x276}},
    {29912, 2, {0x88, 0x3FE, 0xF888, 0x23DE, 0x2252, 0x23DE, 0x20A0, 0xF890, 0x21FE, 0x2120, 0x23FC, 0x3D20, 0xE1FC, 0x4120, 0x1FE, 0x100}},
    {29913, 2, {0x40, 0x3FC, 0xF840, 0x21F8, 0x2000, 0x23FC, 0xF804, 0x21F0, 0x2040, 0x23FC, 0x2008, 0x3BBC, 0xE2A8, 0x4398, 0x524, 0x892}},
    {29914, 2, {0x528, 0x7BE, 0xF948, 0x27BE, 0x2318, 0x25AA, 0xF946, 0x23FC, 0x2204, 0x23FC, 0x2204, 0x3BFC, 0xE204, 0x43FC, 0x108, 0x204}},
    {29915, 2, {0x108, 0x1EC, 0xF10A, 0x2FEA, 0x2928, 0x2BC8, 0x293E, 0xFFE8, 0x2948, 0x2BE8, 0x2A28, 0x2B68, 0x3AA8, 0xCBF4, 0xAB4, 0x12A2}},
    {29916, 2, {0x10, 0xF8, 0x1F20, 0x1220, 0x1220, 0x1220, 0x1220, 0x1220, 0x1220, 0x1210, 0x1290, 0x2250, 0x22A8, 0x4328, 0x4204, 0x8002}},
    {29917, 2, {0x220, 0x720, 0x383C, 0x2A44, 0x2A84, 0x2A24, 0x2A14, 0x2A14, 0x2A04, 0x2904, 0x2928, 0x2890, 0x2840, 0x4A30, 0x4D0E, 0x8900}},
    {29918, 2, {0x210, 0x710, 0x3850, 0x2A7C, 0x2A90, 0x2B10, 0x2AFE, 0x2A10, 0x2A28, 0x2924, 0x2942, 0x2882, 0x2840, 0x4A30, 0x4D0E, 0x8900}},
    {29919, 2, {0x240, 0x740, 0x387C, 0x2A84, 0x2B74, 0x2A54, 0x2A54, 0x2A74, 0x2A48, 0x2942, 0x2942, 0x28BE, 0x2840, 0x4A30, 0x4D0E, 0x8900}},
    {29920, 2, {0x1008, 0x101C, 0xFEE8, 0x28A8, 0x44A8, 0x82A8, 0x7CA8, 0xA8, 0xFEA8, 0x20A4, 0x40A4, 0x7CA4, 0x4A2, 0x528, 0x2934, 0x1224}},
    {29921, 2, {0x1008, 0x101C, 0x7CE8, 0x10A8, 0x10A8, 0xFEA8, 0x44A8, 0x28A8, 0x7CA8, 0x10A4, 0x10A4, 0xFEA4, 0x10A2, 0x1128, 0x1134, 0x1224}},
    {29922, 2, {0x8, 0xFE1C, 0x28E8, 0xFEA8, 0xAAA8, 0xAAA8, 0xFEA8, 0xA8, 0x7CA8, 0xA4, 0xFEA4, 0x10A4, 0x54A2, 0x9328, 0x5134, 0x2224}},
    {29923, 2, {0x4090, 0x21C8, 0x2708, 0xFD3E, 0x580, 0x5554, 0x2548, 0xFD7E, 0x2548, 0x2548, 0x755C, 0x2528, 0x2508, 0x4948, 0x49A8, 0x9128}},
    {29924, 2, {0x1008, 0xFE1C, 0xE8, 0xEEA8, 0xAAA8, 0xEEA8, 0x44A8, 0xFEA8, 0x44A8, 0xFEA4, 0x44A4, 0xFEA4, 0x50A2, 0xCB28, 0x6534, 0x4224}},
    {29925, 2, {0x6FE, 0x3820, 0x2C54, 0x2AB4, 0x2A58, 0x29B4, 0x5454, 0x5D92, 0x9030, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {29926, 2, {0x0, 0xFFFE, 0x400, 0x400, 0x400, 0x7E0, 0x820, 0xA20, 0x920, 0x920, 0x820, 0x1022, 0x1322, 0x1C22, 0x101E, 0x0}},
    {29927, 2, {0x20, 0xFC20, 0x2020, 0x2020, 0x2020, 0x3DFC, 0x2420, 0x2420, 0x5420, 0x4C20, 0x4420, 0x4420, 0x5422, 0x6402, 0x43FE, 0x0}},
    {29928, 2, {0x0, 0x1FC, 0x7C80, 0x1080, 0x1080, 0x10F8, 0x1088, 0x1088, 0x1148, 0x1128, 0x1128, 0x1D08, 0xE108, 0x414A, 0x18A, 0x106}},
    {29929, 2, {0x10, 0xFC38, 0x20E0, 0x2020, 0x2020, 0x3C20, 0x25FC, 0x2420, 0x5420, 0x4C20, 0x4420, 0x4420, 0x5422, 0x6402, 0x43FE, 0x0}},
    {29930, 2, {0x4, 0xFC0E, 0x20F0, 0x2080, 0x2080, 0x3CFC, 0x2484, 0x24C4, 0x54A8, 0x4C90, 0x4528, 0x4544, 0x5602, 0x6402, 0x43FE, 0x0}},
    {29931, 2, {0x440, 0x820, 0x3018, 0xDFE6, 0x420, 0x420, 0x820, 0x10A0, 0x2040, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29932, 2, {0x2000, 0x11FC, 0x80, 0xFE80, 0x2080, 0x20F8, 0x3C88, 0x2488, 0x2548, 0x2528, 0x2528, 0x2508, 0x4508, 0x554A, 0x898A, 0x106}},
    {29933, 2, {0x0, 0x11FC, 0x1080, 0x1080, 0xFE80, 0x92F8, 0x1088, 0x1088, 0x2948, 0x2928, 0x2928, 0x2908, 0x4B08, 0x4D4A, 0x898A, 0x106}},
    {29934, 2, {0x440, 0x820, 0x1010, 0x2208, 0xC446, 0x820, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29935, 2, {0x0, 0x7EFC, 0x4040, 0x4440, 0x6440, 0x5478, 0x4848, 0x4848, 0x54A8, 0x5498, 0x6488, 0x4088, 0x40AA, 0x7ECA, 0x8A, 0x6}},
    {29936, 2, {0x10, 0xFC50, 0x2048, 0x2088, 0x2084, 0x3D02, 0x26F8, 0x2448, 0x5448, 0x4C88, 0x4488, 0x4528, 0x5612, 0x6402, 0x43FE, 0x0}},
    {29937, 2, {0x20, 0xFC70, 0x21C0, 0x2040, 0x2070, 0x3DC0, 0x2440, 0x2478, 0x57C0, 0x4C48, 0x4448, 0x4448, 0x543A, 0x6402, 0x43FE, 0x0}},
    {29938, 2, {0x20, 0xFC20, 0x21FE, 0x2020, 0x2124, 0x3D24, 0x2524, 0x25FC, 0x5420, 0x4C24, 0x4424, 0x441C, 0x5402, 0x6402, 0x43FE, 0x0}},
    {29939, 2, {0x1000, 0x11FC, 0x1080, 0x1080, 0xFE80, 0x10F8, 0x1088, 0x1088, 0x7D48, 0x4528, 0x4528, 0x4508, 0x4508, 0x7D4A, 0x458A, 0x106}},
    {29940, 2, {0x1000, 0x10FC, 0x2840, 0x2440, 0x4240, 0x9178, 0x848, 0x848, 0xFEA8, 0x298, 0x4488, 0x2888, 0x10AA, 0x8CA, 0x88A, 0x6}},
    {29941, 2, {0x1000, 0x11FC, 0x2080, 0x2880, 0x4480, 0xFEF8, 0x4288, 0x88, 0x7D48, 0x4528, 0x4528, 0x4508, 0x4508, 0x7D4A, 0x458A, 0x106}},
    {29942, 2, {0x4200, 0x24FC, 0x40, 0x7E40, 0x2440, 0x2478, 0x2448, 0x2448, 0xFEA8, 0x2498, 0x2488, 0x2488, 0x24AA, 0x44CA, 0x448A, 0x8406}},
    {29943, 2, {0x100, 0x4100, 0x21FC, 0xA44, 0x1448, 0xE0A0, 0x2110, 0x260C, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29944, 2, {0x0, 0xFDFC, 0x2020, 0x2040, 0x20F8, 0x3C88, 0x2488, 0x24F8, 0x5488, 0x4C88, 0x44F8, 0x4488, 0x5402, 0x6402, 0x43FE, 0x0}},
    {29945, 2, {0x0, 0x7EFC, 0x4240, 0x4240, 0x7E40, 0x78, 0x7E48, 0x4248, 0x42A8, 0x7E98, 0x4288, 0x4288, 0x7EAA, 0x42CA, 0x4A8A, 0x4406}},
    {29946, 2, {0x0, 0x3EFC, 0x2040, 0x2040, 0x3C40, 0x2078, 0x2048, 0x3C48, 0x20A8, 0x2098, 0xFE88, 0x2088, 0x24AA, 0x42CA, 0xFE8A, 0x4206}},
    {29947, 2, {0x4200, 0x24FC, 0x1840, 0x2440, 0x5240, 0x1078, 0xFF48, 0x2048, 0x28A8, 0x7E98, 0xAA88, 0x2A88, 0x2AAA, 0x2ECA, 0x88A, 0x806}},
    {29948, 2, {0x0, 0xFDFC, 0x2124, 0x2124, 0x21FC, 0x3D24, 0x2524, 0x25FC, 0x5420, 0x4DFC, 0x4420, 0x4420, 0x55FE, 0x6402, 0x43FE, 0x0}},
    {29949, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29950, 2, {0xC00, 0x71FC, 0x1080, 0xFE80, 0x3880, 0x54F8, 0x9288, 0x88, 0x1148, 0xFD28, 0x2528, 0x4508, 0x2908, 0x114A, 0x298A, 0xC506}},
    {29951, 2, {0x1000, 0x8FC, 0x7F40, 0x40, 0x2240, 0x1478, 0xFF48, 0x48, 0xA8, 0x3E98, 0x2288, 0x2288, 0x22AA, 0x3ECA, 0x228A, 0x6}},
    {29952, 2, {0x200, 0xF7C, 0x7820, 0x820, 0xFFA0, 0x2A38, 0x2A28, 0xFFA8, 0x2A58, 0x2A48, 0xFFC8, 0x848, 0x85A, 0xF6A, 0xF04A, 0x4006}},
    {29953, 2, {0x2200, 0x22FC, 0x4440, 0xEE40, 0x4440, 0x4478, 0x4448, 0x4448, 0xEFA8, 0x4498, 0x4488, 0x4488, 0x44AA, 0x44CA, 0x448A, 0x8406}},
    {29954, 2, {0x1000, 0x87C, 0x7E20, 0x4220, 0x4220, 0x7E38, 0x4228, 0x4028, 0x7F58, 0x5548, 0x5548, 0x7F48, 0xD55A, 0x556A, 0x554A, 0x4306}},
    {29955, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x904, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29956, 2, {0x0, 0xFEFC, 0x2840, 0x2840, 0xFE40, 0xAA78, 0xAA48, 0xFE48, 0x10A8, 0x1098, 0xFE88, 0x1088, 0x10AA, 0x1ECA, 0xF08A, 0x4006}},
    {29957, 2, {0x0, 0xFDFE, 0x2100, 0x217C, 0x2154, 0x3D7C, 0x2554, 0x257C, 0x5510, 0x4D7C, 0x4510, 0x46FE, 0x5402, 0x6402, 0x43FE, 0x0}},
    {29958, 2, {0x820, 0x440, 0x7FFC, 0x1248, 0x3CF0, 0x820, 0x1450, 0x3EF8, 0x0, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29959, 2, {0x820, 0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29960, 2, {0x800, 0x10FC, 0x7E40, 0x4240, 0x7E40, 0x4278, 0x7E48, 0x4248, 0x7EA8, 0x898, 0xFF88, 0x1888, 0x2CAA, 0x4ACA, 0x888A, 0x806}},
    {29961, 2, {0x2200, 0x147C, 0xFFA0, 0x1420, 0x1420, 0x7F38, 0x1528, 0xFFA8, 0x1558, 0x7F48, 0x1448, 0x3648, 0x555A, 0x94EA, 0x144A, 0x1406}},
    {29962, 2, {0x800, 0x7F7C, 0x4920, 0xFFA0, 0x4920, 0x7F38, 0x828, 0x7F28, 0x4958, 0x7F48, 0x1048, 0xFF48, 0x215A, 0x726A, 0xC4A, 0xF306}},
    {29963, 2, {0x2000, 0x10FC, 0xFE40, 0x4440, 0x2840, 0xFE78, 0x8248, 0x9248, 0xFEA8, 0x9298, 0xBA88, 0xAA88, 0xAAAA, 0xBACA, 0x828A, 0x8606}},
    {29964, 2, {0x0, 0xFEFC, 0x8040, 0x8040, 0xBC40, 0xA478, 0xA448, 0xBC48, 0x80A8, 0xEE98, 0xAA88, 0xAA88, 0xEEAA, 0x80CA, 0xFE8A, 0x6}},
    {29965, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFE, 0x4002, 0xBFF4, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29966, 2, {0x1000, 0x10FC, 0xFE40, 0x1040, 0x7C40, 0x5478, 0x7C48, 0x5448, 0x7CA8, 0x1298, 0xFE88, 0x888, 0xFEAA, 0x48CA, 0x288A, 0x1806}},
    {29967, 2, {0x804, 0x7F08, 0x810, 0x3E64, 0x8, 0x3E10, 0x2264, 0x3E08, 0x1410, 0x7F60, 0x0, 0x7FFC, 0x800, 0xFE0, 0x1124, 0x1C9C}},
    {29968, 2, {0x800, 0x497C, 0x2A20, 0x820, 0xFF20, 0x2A38, 0x4928, 0x8828, 0x4258, 0x7748, 0x9248, 0x5A48, 0x2F5A, 0x226A, 0x424A, 0x8206}},
    {29969, 2, {0x4400, 0x24FC, 0x2840, 0xFE40, 0x9240, 0xD678, 0xBA48, 0x9248, 0xFEA8, 0x98, 0x7C88, 0x4488, 0x7CAA, 0x44CA, 0x7C8A, 0x4406}},
    {29970, 2, {0x2000, 0x207C, 0x7FA0, 0xD520, 0x5520, 0x5538, 0xFFA8, 0x5528, 0x5558, 0x5548, 0xFFC8, 0x48, 0x6A5A, 0x556A, 0x954A, 0x6}},
    {29971, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x0, 0x7FFC, 0x800, 0xFE0, 0x1124, 0x1C9C}},
    {29972, 2, {0x2000, 0x3E7C, 0x4220, 0xFFA0, 0x5120, 0x64B8, 0x7FA8, 0x4028, 0x5F58, 0x4048, 0x5F48, 0x4048, 0x5F5A, 0x516A, 0x5F4A, 0x9106}},
    {29973, 2, {0x100, 0xFFFE, 0x1048, 0x24FE, 0x7990, 0x10FC, 0x2290, 0x7CFC, 0x890, 0x30FE, 0xC080, 0x7FFC, 0x800, 0xFE0, 0x1124, 0x1C9C}},
    {29974, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2288, 0x7FFC, 0x800, 0xFE0, 0x920, 0x10A4, 0x1624, 0x181C}},
    {29975, 2, {0x800, 0xF7C, 0x820, 0x7F20, 0x4920, 0x5E38, 0x4928, 0x7F28, 0x4A58, 0x5F48, 0x5148, 0x5B48, 0x555A, 0x5F6A, 0x554A, 0x9506}},
    {29976, 2, {0x810, 0x810, 0x810, 0x810, 0x7FFE, 0x810, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0x810, 0x810, 0xFF0, 0x810}},
    {29977, 2, {0x50, 0x48, 0x40, 0xFFFE, 0x40, 0x2240, 0x2240, 0xFF40, 0x2240, 0x2220, 0x3E20, 0x2212, 0x2212, 0x3E0A, 0x2206, 0x2}},
    {29978, 2, {0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0xFFFE, 0x1220, 0x1410, 0x1000, 0x1FF8}},
    {29979, 2, {0x4404, 0x441E, 0x44F0, 0x4410, 0xFE10, 0x44FE, 0x4410, 0x4410, 0x4410, 0x7CFE, 0x4482, 0x4482, 0x4482, 0x4482, 0x7CFE, 0x4482}},
    {29980, 2, {0x888, 0x1C88, 0x7088, 0x1088, 0x11FE, 0xFC88, 0x1088, 0x1088, 0x7C88, 0x44F8, 0x4488, 0x4488, 0x4488, 0x7C88, 0x44F8, 0x88}},
    {29981, 2, {0x848, 0xF48, 0x848, 0x7F48, 0x49FE, 0x4E48, 0x7848, 0x4948, 0x4778, 0x4048, 0x5E48, 0x5248, 0x527A, 0x5202, 0x91FE, 0x2000}},
    {29982, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x0, 0x1010, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {29983, 2, {0x100, 0x1100, 0x1100, 0x1100, 0x3FFC, 0x2100, 0x4100, 0x8100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {29984, 2, {0x20, 0x20, 0x7D20, 0x4520, 0x45FC, 0x4520, 0x4620, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x4420, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {29985, 2, {0x810, 0x810, 0x4890, 0x4890, 0x7EFE, 0x4890, 0x8910, 0x810, 0x810, 0x7EFE, 0x810, 0x810, 0x810, 0xE10, 0xF1FE, 0x4000}},
    {29986, 2, {0x100, 0x80, 0x3FFE, 0x220, 0x1C0, 0x220, 0x3FFE, 0x2080, 0x2480, 0x27FC, 0x2880, 0x3080, 0x27F8, 0x4080, 0x4080, 0x9FFE}},
    {29987, 2, {0x100, 0x80, 0x3FFE, 0x410, 0x220, 0x3FFE, 0x2080, 0x2480, 0x27FC, 0x2880, 0x3080, 0x27F8, 0x4080, 0x4080, 0x9FFE, 0x0}},
    {29988, 2, {0x10, 0x10, 0xFF50, 0x87C, 0x1050, 0x3090, 0x4A10, 0x9C7C, 0x2810, 0x4C10, 0x9A10, 0x29FE, 0x4880, 0x8860, 0x281E, 0x1000}},
    {29989, 2, {0x11FC, 0x1124, 0x5124, 0x51FC, 0x7D24, 0x5124, 0x91FC, 0x1040, 0x7C40, 0x13FE, 0x1042, 0x1082, 0x1C82, 0xE102, 0x4214, 0x408}},
    {29990, 2, {0x10, 0xFF10, 0x1050, 0x107C, 0xFE50, 0x9290, 0x9210, 0xFE7C, 0x9210, 0x9210, 0xFE10, 0x90FE, 0x5000, 0x2000, 0x5FFE, 0x8000}},
    {29991, 2, {0x1010, 0x5050, 0x7E7E, 0x9090, 0x7C7C, 0x1010, 0x1EFE, 0xE000, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {29992, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x4108, 0x4128, 0x8010}},
    {29993, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x4112, 0x4102, 0x80FE}},
    {29994, 2, {0x400, 0x800, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x4108, 0x4128, 0x8010}},
    {29995, 2, {0x120, 0x110, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2010}},
    {29996, 2, {0x0, 0x3FF0, 0x20, 0x640, 0x180, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2010}},
    {29997, 2, {0x7FFC, 0x100, 0x770, 0x190C, 0xE102, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x4108, 0x4128, 0x8110}},
    {29998, 2, {0x800, 0x1FF8, 0x2248, 0x4488, 0x950, 0x1220, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x4108, 0x4128, 0x8110}},
    {29999, 2, {0x200, 0x100, 0x7FFE, 0x4102, 0x9494, 0x1428, 0x23E8, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x2128, 0x2010}},
    {30000, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x0}},
    {30001, 2, {0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30002, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {30003, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100}},
    {30004, 2, {0x100, 0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {30005, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x210A, 0x102, 0x102, 0xFE}},
    {30006, 2, {0x200, 0x200, 0x400, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x0}},
    {30007, 2, {0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x7FFC, 0x104, 0x104, 0x204, 0x404, 0x1814, 0x6008}},
    {30008, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2008, 0x4008, 0xBF88, 0x2488, 0x2488, 0x3F88, 0x2488, 0x2488, 0x3F88, 0x2088, 0x50, 0x20}},
    {30009, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x800, 0x1000, 0x1FF0, 0x10, 0x10, 0xA0, 0x40}},
    {30010, 2, {0x0, 0x0, 0x7DFE, 0x5410, 0x5410, 0x5410, 0x5410, 0x7C10, 0x5410, 0x5410, 0x5410, 0x5410, 0x7C10, 0x4410, 0x50, 0x20}},
    {30011, 2, {0x0, 0xFFFE, 0x0, 0x0, 0x4FE4, 0x4924, 0x4924, 0x4FE4, 0x4924, 0x4924, 0x4924, 0x4FE4, 0x4004, 0x4004, 0x7FFC, 0x4}},
    {30012, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {30013, 2, {0x82, 0x92, 0x7C92, 0x5492, 0x5492, 0x5492, 0x5492, 0x7C92, 0x5492, 0x5492, 0x5492, 0x5492, 0x7C92, 0x4512, 0x102, 0x202}},
    {30014, 2, {0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30015, 2, {0x20, 0x10, 0x7C10, 0x5400, 0x55FE, 0x5480, 0x5480, 0x7C80, 0x5480, 0x5480, 0x5480, 0x5480, 0x7C80, 0x4480, 0xFC, 0x0}},
    {30016, 2, {0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x0, 0x0, 0xFFFE, 0x820, 0x820, 0x820, 0x1020, 0x2020}},
    {30017, 2, {0x100, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x0, 0xFFFE, 0x820, 0x820, 0x820, 0x1020, 0x2020}},
    {30018, 2, {0x80, 0x80, 0x7C80, 0x54F8, 0x5508, 0x5508, 0x5510, 0x7E10, 0x5420, 0x5420, 0x5450, 0x5450, 0x7C88, 0x4508, 0x204, 0x402}},
    {30019, 2, {0x1040, 0x1040, 0x1040, 0xFE7C, 0x9284, 0x9284, 0xFF04, 0x9244, 0x9224, 0xFE24, 0x9204, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {30020, 2, {0x408, 0x4208, 0x2110, 0x1110, 0x1020, 0x40, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30021, 2, {0x1000, 0x11F8, 0x1010, 0x7C20, 0x5440, 0x5480, 0x7DFC, 0x5454, 0x5454, 0x5494, 0x7D24, 0x5424, 0x1044, 0x1084, 0x1128, 0x1010}},
    {30022, 2, {0x1000, 0x808, 0x808, 0xFF88, 0x8, 0x8, 0x7F10, 0x4910, 0x4910, 0x4920, 0x7F24, 0x4924, 0x4942, 0x49FE, 0x7F42, 0x4102}},
    {30023, 2, {0x40, 0x40, 0x7C80, 0x54FC, 0x5504, 0x5604, 0x5484, 0x7C44, 0x5444, 0x5414, 0x5424, 0x5444, 0x7D84, 0x4404, 0x28, 0x10}},
    {30024, 2, {0x8, 0x1C, 0x7DE0, 0x5500, 0x5500, 0x55FC, 0x5544, 0x7D44, 0x5544, 0x5528, 0x5528, 0x5510, 0x7D10, 0x4628, 0x244, 0x482}},
    {30025, 2, {0x20, 0x20, 0x7C20, 0x5420, 0x55FC, 0x5420, 0x5420, 0x7C20, 0x57FE, 0x5420, 0x5450, 0x5450, 0x7C88, 0x4488, 0x104, 0x202}},
    {30026, 2, {0x88, 0x88, 0x7C88, 0x5488, 0x55FE, 0x5488, 0x5488, 0x7C88, 0x57FE, 0x5488, 0x5488, 0x5488, 0x7C88, 0x4508, 0x108, 0x208}},
    {30027, 2, {0x40, 0x40, 0x7C40, 0x5480, 0x54FE, 0x5508, 0x5688, 0x7C88, 0x5488, 0x5450, 0x5450, 0x5420, 0x7C50, 0x4488, 0x104, 0x202}},
    {30028, 2, {0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x280, 0xC60, 0x3458, 0xC446, 0x440, 0x840, 0x840, 0x1040}},
    {30029, 2, {0x20, 0x20, 0x7C50, 0x5450, 0x5488, 0x5504, 0x5602, 0x7C88, 0x5488, 0x5488, 0x5488, 0x5488, 0x7C88, 0x4508, 0x108, 0x208}},
    {30030, 2, {0x20, 0x28, 0x7C24, 0x5424, 0x5420, 0x55FE, 0x5420, 0x7C20, 0x5420, 0x5450, 0x5450, 0x5450, 0x7C88, 0x4488, 0x104, 0x202}},
    {30031, 2, {0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x0, 0xFFFE, 0x1110, 0x10A0, 0x1040, 0x1430, 0x180E, 0x1000}},
    {30032, 2, {0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30033, 2, {0x1000, 0x11FC, 0x1124, 0x1124, 0x5524, 0x5924, 0x5124, 0x91FC, 0x1124, 0x1124, 0x1124, 0x2924, 0x2524, 0x45FC, 0x4104, 0x8000}},
    {30034, 2, {0x1010, 0x810, 0x810, 0xFF90, 0x10, 0x10, 0x7F10, 0x4910, 0x4910, 0x4928, 0x7F28, 0x4928, 0x4944, 0x4944, 0x7F84, 0x4102}},
    {30035, 2, {0x100, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30036, 2, {0x20, 0x20, 0x7D24, 0x54A4, 0x54A8, 0x5420, 0x55FC, 0x7C20, 0x5420, 0x5420, 0x57FE, 0x5420, 0x7C20, 0x4420, 0x20, 0x20}},
    {30037, 2, {0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30038, 2, {0x8, 0x1C, 0x7DF0, 0x5550, 0x5550, 0x5550, 0x5550, 0x7D50, 0x5550, 0x5548, 0x5548, 0x5568, 0x7D54, 0x4674, 0x252, 0x400}},
    {30039, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30040, 2, {0x20, 0x20, 0x7C20, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x7D24, 0x5524, 0x5524, 0x57FE, 0x5504, 0x7D04, 0x4504, 0x114, 0x108}},
    {30041, 2, {0x600, 0x78FC, 0x4044, 0x4844, 0x4444, 0x5A94, 0x6108, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30042, 2, {0x200, 0x440, 0x820, 0x1FF0, 0x200, 0x7FFC, 0x820, 0x1010, 0x3FF8, 0xD116, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {30043, 2, {0x20, 0x20, 0x7C50, 0x5488, 0x5504, 0x5612, 0x5420, 0x7C40, 0x5588, 0x5410, 0x5420, 0x5444, 0x7D88, 0x4410, 0x60, 0x380}},
    {30044, 2, {0x200, 0x100, 0x7FFC, 0x400, 0x820, 0x1FC0, 0x310, 0xC08, 0x3FFC, 0x4, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {30045, 2, {0x1020, 0x820, 0x820, 0xFFBC, 0x24, 0x44, 0x7F44, 0x4988, 0x4908, 0x4910, 0x7F10, 0x4928, 0x4928, 0x4944, 0x7F84, 0x4102}},
    {30046, 2, {0x100, 0x100, 0xFFFE, 0x120, 0x120, 0x7D20, 0x5538, 0x5548, 0x5548, 0x7D88, 0x5510, 0x5510, 0x5528, 0x7D48, 0x4584, 0x102}},
    {30047, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x1210, 0x2208, 0x47E4, 0x8822, 0x1440, 0x2280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {30048, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30049, 2, {0x20, 0x10, 0x7C10, 0x55FE, 0x5420, 0x5424, 0x5444, 0x7CF8, 0x5412, 0x5422, 0x5444, 0x5588, 0x7C10, 0x4428, 0x44, 0x182}},
    {30050, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x920, 0x920, 0x7FFC, 0x920, 0x920, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x100}},
    {30051, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {30052, 2, {0x20, 0x20, 0x7C20, 0x55FC, 0x5420, 0x5420, 0x57FE, 0x7C08, 0x5408, 0x55FE, 0x5408, 0x5488, 0x7C48, 0x4408, 0x28, 0x10}},
    {30053, 2, {0x80, 0x80, 0xF8F8, 0xA908, 0xAB10, 0xACA0, 0xF840, 0xA8A0, 0xA918, 0xAA06, 0xADF8, 0xF908, 0x8908, 0x108, 0x1F8, 0x108}},
    {30054, 2, {0x20, 0x20, 0x7DFC, 0x5420, 0x5420, 0x5420, 0x57FE, 0x7C00, 0x5420, 0x5420, 0x55FC, 0x5420, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {30055, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE00E, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30056, 2, {0x100, 0x1110, 0x920, 0xFFFE, 0x540, 0x920, 0x3118, 0xC006, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {30057, 2, {0x40, 0x20, 0x7C20, 0x55FE, 0x5440, 0x5440, 0x54A0, 0x7CA2, 0x55A4, 0x5698, 0x5490, 0x5488, 0x7C88, 0x44A4, 0xC2, 0x80}},
    {30058, 2, {0xF8, 0x3F00, 0x1110, 0x920, 0x100, 0xFFFE, 0x920, 0x1010, 0x3FF8, 0xD116, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010}},
    {30059, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE}},
    {30060, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x3FF8, 0x1110, 0x2508, 0x200, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {30061, 2, {0x20, 0x20, 0x7C50, 0x5488, 0x5504, 0x56FA, 0x5420, 0x7C20, 0x57FE, 0x5420, 0x5528, 0x5524, 0x7E22, 0x4422, 0xA0, 0x40}},
    {30062, 2, {0x80, 0x80, 0x7CFE, 0x5500, 0x56FC, 0x5484, 0x54A4, 0x7C94, 0x57FE, 0x5484, 0x5524, 0x5514, 0x7DFE, 0x4404, 0x28, 0x10}},
    {30063, 2, {0x40, 0x40, 0xF888, 0xA904, 0xABFE, 0xA802, 0xF888, 0xA944, 0xAA42, 0xA8F8, 0xA988, 0xFA50, 0x8820, 0x50, 0x188, 0x606}},
    {30064, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008}},
    {30065, 2, {0x0, 0x7FFC, 0x280, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x280, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {30066, 2, {0x100, 0x280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x920, 0x1110, 0x2508, 0x200, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {30067, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x8002, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {30068, 2, {0x20, 0x20, 0xFBFE, 0xA820, 0xA9FC, 0xA840, 0xFBFE, 0xA848, 0xA888, 0xA8FE, 0xA908, 0xF948, 0x8A28, 0x408, 0x28, 0x10}},
    {30069, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x5114, 0x5FF4, 0x5114, 0x5FF4, 0x4004, 0x7FFC}},
    {30070, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {30071, 2, {0x0, 0x3DE, 0xF842, 0xA94A, 0xA884, 0xA94A, 0xFA52, 0xAC20, 0xA800, 0xABDE, 0xA852, 0xF952, 0x8894, 0x148, 0x254, 0x422}},
    {30072, 2, {0x20, 0x20, 0x7DFC, 0x5450, 0x5488, 0x5504, 0x57FE, 0x7C08, 0x55E8, 0x5528, 0x5528, 0x55E8, 0x7D28, 0x4408, 0x28, 0x10}},
    {30073, 2, {0x40, 0x20, 0xFBFE, 0xAA02, 0xAD04, 0xA900, 0xF9DE, 0xAA52, 0xAA52, 0xAB52, 0xAC9A, 0xF894, 0x8910, 0x112, 0x212, 0x40E}},
    {30074, 2, {0x0, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE}},
    {30075, 2, {0x20, 0x124, 0x7CA8, 0x5420, 0x55FC, 0x5440, 0x57FE, 0x7C88, 0x5524, 0x5622, 0x5420, 0x55FC, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {30076, 2, {0x0, 0xFC, 0x7C84, 0x54FC, 0x5484, 0x54FC, 0x5400, 0x7DFE, 0x5480, 0x54FE, 0x552A, 0x564A, 0x7C92, 0x4522, 0x54, 0x88}},
    {30077, 2, {0x8, 0x3C, 0x7DE0, 0x5420, 0x57FE, 0x5420, 0x55FC, 0x7D24, 0x55FC, 0x5524, 0x55FC, 0x5420, 0x7DFC, 0x4420, 0x3FE, 0x0}},
    {30078, 2, {0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x0, 0xFEFE, 0x9292, 0x9292, 0xFEFE, 0x9292, 0x9292, 0xFEFE, 0x8282}},
    {30079, 2, {0x1110, 0x2120, 0x4544, 0x7978, 0x1110, 0x2524, 0x7D7C, 0x524, 0x110, 0xFFFE, 0x80, 0x7C88, 0x5452, 0x7C22, 0x545A, 0x7D86}},
    {30080, 2, {0x20, 0x5420, 0x55FE, 0xA820, 0x55FC, 0x5424, 0x3FE, 0x7C24, 0x55FC, 0x5420, 0x7CA0, 0x54BC, 0x54A0, 0x7D60, 0x453E, 0x200}},
    {30081, 2, {0x0, 0x7BC, 0xF884, 0xACA4, 0xAA94, 0xACA4, 0xF850, 0xA988, 0xAE26, 0xA8C0, 0xAB10, 0xF864, 0x8B88, 0x30, 0xC0, 0x700}},
    {30082, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x1830, 0x6C0, 0x1830, 0x7FFE, 0x5012, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {30083, 2, {0x20, 0x1FC, 0x7C88, 0x5450, 0x57FE, 0x5400, 0x55FC, 0x7D24, 0x55FC, 0x5524, 0x55FC, 0x5420, 0x7DFC, 0x4420, 0x3FE, 0x0}},
    {30084, 2, {0x20, 0x124, 0xF8A8, 0xABFE, 0xA8A8, 0xA924, 0xFA22, 0xA904, 0xA904, 0xA9DE, 0xAA44, 0xFD54, 0x889E, 0x104, 0x204, 0x404}},
    {30085, 2, {0x0, 0x3FE, 0xF800, 0xA9FC, 0xA924, 0xA9FC, 0xF924, 0xABFE, 0xA800, 0xA9FC, 0xA924, 0xF9FC, 0x8924, 0x1FC, 0x0, 0x3FE}},
    {30086, 2, {0x0, 0xFBFE, 0x800, 0x9FC, 0xF924, 0x81FC, 0x8124, 0xFBFE, 0x4800, 0x49FC, 0xE924, 0x49FC, 0x6924, 0x89FC, 0x5000, 0x23FE}},
    {30087, 2, {0x20, 0x3FE, 0xF820, 0xA9FC, 0xA800, 0xABFE, 0xF802, 0xA9FC, 0xA820, 0xABFE, 0xA800, 0xFBFE, 0x8804, 0x3BE, 0x2A4, 0x38C}},
    {30088, 2, {0x80, 0xF8BE, 0x80, 0x709C, 0x5294, 0x729C, 0x280, 0xFABE, 0xAAAA, 0xACAA, 0xF8BE, 0xA8AA, 0xA8AA, 0xF93E, 0x8922, 0x200}},
    {30089, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0xFFFE, 0x8802, 0xFF0, 0x1010, 0x1FE0, 0x20, 0xFFFE, 0x0}},
    {30090, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0xFFFE, 0x9012, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x0}},
    {30091, 2, {0x0, 0x0, 0x7FFC, 0x104, 0x108, 0x100, 0x1100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x2900, 0x2500, 0x43FE, 0x8000}},
    {30092, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {30093, 2, {0x0, 0x7FFC, 0x104, 0x1100, 0x11F0, 0x2900, 0x47FE, 0x8000, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x0}},
    {30094, 2, {0x20, 0x20, 0x7DFE, 0x420, 0x820, 0x11FC, 0x5124, 0x5124, 0x5DFC, 0x5020, 0x5070, 0x50A8, 0x5D24, 0x7222, 0xC020, 0x20}},
    {30095, 2, {0x20, 0x10, 0x7DFE, 0x420, 0x848, 0x1084, 0x51FE, 0x5002, 0x5CA8, 0x50A8, 0x50A8, 0x50A8, 0x5D2A, 0x712A, 0xC22A, 0x406}},
    {30096, 2, {0x100, 0x3FF8, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0x7FFC, 0x104, 0x11F0, 0x2900, 0xC7FE}},
    {30097, 2, {0x4000, 0x4DFC, 0x7008, 0x4450, 0x4420, 0x3DFE, 0x4024, 0x7C28, 0x90A0, 0x10B8, 0xFEA0, 0x10A0, 0x28E0, 0x2520, 0x451E, 0x8200}},
    {30098, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x5000, 0x5000, 0x1000, 0x3000, 0x5000, 0x9000, 0x1000, 0x2000, 0x2000, 0x4000, 0x8000}},
    {30099, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57F8, 0x5108, 0x1110, 0x3120, 0x513C, 0x9104, 0x1104, 0x2204, 0x2204, 0x4428, 0x8810}},
    {30100, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5040, 0x1040, 0x3040, 0x5040, 0x9040, 0x1040, 0x2040, 0x2040, 0x4140, 0x8080}},
    {30101, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x5200, 0x5208, 0x1210, 0x3260, 0x5380, 0x9200, 0x1204, 0x2204, 0x2204, 0x41FC, 0x8000}},
    {30102, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5084, 0x1084, 0x3084, 0x5084, 0x9094, 0x1088, 0x2080, 0x2080, 0x4080, 0x8080}},
    {30103, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5008, 0x1010, 0x3020, 0x5040, 0x9040, 0x1040, 0x2040, 0x2040, 0x4140, 0x8080}},
    {30104, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x5000, 0x57FC, 0x1040, 0x3040, 0x5040, 0x9040, 0x1040, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {30105, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x9100, 0x53FC, 0x5400, 0x1800, 0x33F8, 0x5010, 0x9060, 0x1080, 0x2100, 0x2202, 0x4202, 0x81FE}},
    {30106, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x9100, 0x5100, 0x53F0, 0x1210, 0x3410, 0x5820, 0x9020, 0x1040, 0x20A0, 0x2110, 0x4608, 0x9806}},
    {30107, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9010, 0x5010, 0x57FE, 0x1010, 0x3010, 0x5210, 0x9110, 0x1110, 0x2010, 0x2010, 0x4050, 0x8020}},
    {30108, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x5000, 0x57FE, 0x1040, 0x3040, 0x5050, 0x9048, 0x1044, 0x2040, 0x2040, 0x4040, 0x8040}},
    {30109, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x5040, 0x1444, 0x3444, 0x5444, 0x9444, 0x1444, 0x2444, 0x27FC, 0x4004, 0x8000}},
    {30110, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53F8, 0x5000, 0x1000, 0x37FE, 0x5080, 0x9100, 0x13F8, 0x2008, 0x2008, 0x4050, 0x8020}},
    {30111, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53FC, 0x5200, 0x1200, 0x3200, 0x5FFE, 0x9200, 0x1200, 0x2200, 0x23FC, 0x4000, 0x8000}},
    {30112, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5100, 0x1100, 0x31F8, 0x5108, 0x9108, 0x1208, 0x2208, 0x2408, 0x4850, 0x9020}},
    {30113, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F0, 0x5020, 0x50C0, 0x1100, 0x33FC, 0x5124, 0x9124, 0x1224, 0x2244, 0x2444, 0x48A8, 0x8110}},
    {30114, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x5244, 0x1244, 0x3248, 0x5450, 0x90A0, 0x10A0, 0x2110, 0x2208, 0x4404, 0x8802}},
    {30115, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9090, 0x5088, 0x5080, 0x17FE, 0x30A0, 0x5120, 0x9120, 0x1120, 0x2222, 0x2222, 0x4422, 0x881E}},
    {30116, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53F8, 0x5248, 0x1248, 0x3248, 0x5248, 0x93F8, 0x1200, 0x2202, 0x2202, 0x41FE, 0x8000}},
    {30117, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x50A0, 0x5110, 0x1208, 0x3D16, 0x5110, 0x9110, 0x1110, 0x2110, 0x2210, 0x4210, 0x8410}},
    {30118, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x53F8, 0x1048, 0x3048, 0x5048, 0x97FE, 0x10A0, 0x20A0, 0x2110, 0x4208, 0x8C06}},
    {30119, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x903C, 0x57C0, 0x5440, 0x1440, 0x3440, 0x57FE, 0x9420, 0x1420, 0x2412, 0x250A, 0x4606, 0x8402}},
    {30120, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5010, 0x1210, 0x3410, 0x57FE, 0x9050, 0x1090, 0x2110, 0x2210, 0x4C50, 0x8020}},
    {30121, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9080, 0x57F0, 0x5110, 0x1112, 0x3212, 0x544E, 0x9040, 0x17FE, 0x2040, 0x2040, 0x4040, 0x8040}},
    {30122, 2, {0x80, 0x40, 0x1FFE, 0x1020, 0x9220, 0x5220, 0x5222, 0x1224, 0x33A8, 0x5230, 0x9220, 0x1222, 0x22A2, 0x2322, 0x421E, 0x8000}},
    {30123, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F0, 0x5210, 0x5210, 0x140E, 0x3800, 0x57F8, 0x9208, 0x1110, 0x20A0, 0x2040, 0x41B0, 0x8E0E}},
    {30124, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93FE, 0x5200, 0x5220, 0x1220, 0x32FC, 0x5224, 0x9224, 0x1244, 0x2244, 0x2484, 0x44A8, 0x8910}},
    {30125, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9110, 0x5110, 0x5110, 0x1110, 0x3110, 0x52A8, 0x9268, 0x1228, 0x2444, 0x2444, 0x4884, 0x9102}},
    {30126, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x90A0, 0x5110, 0x5208, 0x1406, 0x3BF0, 0x5210, 0x9210, 0x1250, 0x2220, 0x2204, 0x4204, 0x81FC}},
    {30127, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53F8, 0x5208, 0x1328, 0x32A8, 0x5248, 0x9248, 0x12A8, 0x252A, 0x240A, 0x4806, 0x9002}},
    {30128, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x5000, 0x57FC, 0x1040, 0x3040, 0x5040, 0x93F8, 0x1040, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {30129, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9200, 0x53FC, 0x5404, 0x1BE4, 0x3224, 0x5224, 0x93E4, 0x1214, 0x2208, 0x2202, 0x4202, 0x81FE}},
    {30130, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x57FC, 0x5444, 0x1440, 0x37F8, 0x5508, 0x9510, 0x14A0, 0x2440, 0x28A0, 0x4918, 0x9606}},
    {30131, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9208, 0x5208, 0x5FFE, 0x1208, 0x3208, 0x5208, 0x93F8, 0x1208, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {30132, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FE, 0x5008, 0x1008, 0x33C8, 0x5248, 0x9248, 0x13C8, 0x2248, 0x2008, 0x4028, 0x8010}},
    {30133, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9110, 0x5110, 0x5512, 0x1514, 0x35D8, 0x5510, 0x9510, 0x1510, 0x2512, 0x25D2, 0x5E12, 0x880E}},
    {30134, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9048, 0x5248, 0x5248, 0x1248, 0x3FFE, 0x5248, 0x9248, 0x1278, 0x2200, 0x2200, 0x43FE, 0x8000}},
    {30135, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x903C, 0x57C0, 0x5440, 0x1440, 0x3440, 0x57FE, 0x9420, 0x1420, 0x2412, 0x250A, 0x4646, 0x8422}},
    {30136, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2000, 0xA7F8, 0x6408, 0x2408, 0x27F8, 0x6408, 0xA408, 0x27F8, 0x2000, 0x4000, 0x5FFE, 0x8000}},
    {30137, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x90A0, 0x5110, 0x5208, 0x1426, 0x30C0, 0x5310, 0x9020, 0x10C0, 0x2308, 0x2010, 0x40E0, 0x8700}},
    {30138, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9078, 0x5780, 0x5040, 0x1020, 0x37FC, 0x5008, 0x9030, 0x10C0, 0x2300, 0x2480, 0x487E, 0x9000}},
    {30139, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53FC, 0x5204, 0x1204, 0x3204, 0x5204, 0x93FC, 0x1204, 0x2090, 0x2108, 0x4204, 0x8402}},
    {30140, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x9100, 0x53F8, 0x5610, 0x1920, 0x30C0, 0x5330, 0x9C0E, 0x10C0, 0x2020, 0x2300, 0x40C0, 0x8020}},
    {30141, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53F8, 0x5208, 0x1208, 0x33F8, 0x5208, 0x9208, 0x13F8, 0x2208, 0x2208, 0x4FFE, 0x8000}},
    {30142, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9200, 0x53FC, 0x5240, 0x1440, 0x3040, 0x5FFE, 0x9040, 0x10A0, 0x20A0, 0x2110, 0x4208, 0x8C06}},
    {30143, 2, {0x100, 0x80, 0x3FFE, 0x2120, 0x3FFC, 0xA124, 0x6124, 0x2FFC, 0x2920, 0x6920, 0xAFFE, 0x2122, 0x2222, 0x422A, 0x4424, 0x8820}},
    {30144, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9200, 0x53FC, 0x5404, 0x1804, 0x33E4, 0x5224, 0x9224, 0x13E4, 0x2224, 0x2004, 0x4028, 0x8010}},
    {30145, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9080, 0x5080, 0x50FC, 0x1080, 0x3080, 0x57F8, 0x9408, 0x1408, 0x2408, 0x2408, 0x47F8, 0x8408}},
    {30146, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2400, 0xA400, 0x6FBC, 0x24A4, 0x24A4, 0x64A4, 0xA4A4, 0x24A4, 0x28A4, 0x48BC, 0x52A4, 0xA100}},
    {30147, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x5FFE, 0x5080, 0x1108, 0x3210, 0x57E0, 0x9040, 0x1080, 0x2110, 0x2208, 0x47FC, 0x8204}},
    {30148, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9200, 0x53FE, 0x5500, 0x1900, 0x31F8, 0x5100, 0x9100, 0x11FC, 0x2100, 0x2100, 0x4100, 0x8100}},
    {30149, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5040, 0x1040, 0x37FC, 0x5444, 0x94A4, 0x1514, 0x2614, 0x2404, 0x4414, 0x8408}},
    {30150, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93FC, 0x5204, 0x5204, 0x13FC, 0x3200, 0x5240, 0x9244, 0x1248, 0x2270, 0x2442, 0x4442, 0x883E}},
    {30151, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x5FFC, 0x5040, 0x1040, 0x3240, 0x527C, 0x9240, 0x1240, 0x2240, 0x2240, 0x5FFE, 0x8000}},
    {30152, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5444, 0x1444, 0x37FC, 0x5444, 0x9444, 0x17FC, 0x2444, 0x2844, 0x4854, 0x9008}},
    {30153, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97F8, 0x5010, 0x5020, 0x10D8, 0x3304, 0x5C02, 0x93F8, 0x1040, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {30154, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x90A0, 0x5110, 0x5208, 0x1DF6, 0x3040, 0x5040, 0x93F8, 0x1040, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {30155, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x57FC, 0x1444, 0x3444, 0x5444, 0x97FC, 0x1040, 0x2048, 0x207C, 0x4F82, 0x8402}},
    {30156, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5404, 0x15F4, 0x3404, 0x55F4, 0x9514, 0x1514, 0x25F4, 0x2404, 0x4414, 0x8408}},
    {30157, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9FFE, 0x5040, 0x5FFC, 0x1044, 0x37FC, 0x5440, 0x97FE, 0x1042, 0x20AA, 0x2114, 0x4208, 0x8C06}},
    {30158, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x5FFE, 0x5080, 0x1108, 0x33F0, 0x5020, 0x9044, 0x1188, 0x2610, 0x2068, 0x4184, 0x8602}},
    {30159, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x57FE, 0x5080, 0x1100, 0x33FC, 0x5504, 0x99FC, 0x1104, 0x21FC, 0x2104, 0x4114, 0x8108}},
    {30160, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5404, 0x1404, 0x35F4, 0x5514, 0x9514, 0x15F4, 0x2404, 0x2404, 0x47FC, 0x8404}},
    {30161, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x91F8, 0x5210, 0x5D20, 0x10C0, 0x31A0, 0x567C, 0x9084, 0x1108, 0x2690, 0x2060, 0x40C0, 0x8700}},
    {30162, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x90A0, 0x57FC, 0x5040, 0x1040, 0x33F8, 0x5040, 0x9040, 0x17FC, 0x2040, 0x2040, 0x4040, 0x8040}},
    {30163, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5080, 0x5110, 0x1208, 0x37FC, 0x5044, 0x9040, 0x17FC, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {30164, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x57FC, 0x5040, 0x1040, 0x3FFE, 0x5010, 0x9010, 0x1FFE, 0x2210, 0x2110, 0x4150, 0x8020}},
    {30165, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97F8, 0x5408, 0x57F8, 0x1408, 0x37F8, 0x5484, 0x9488, 0x1450, 0x2420, 0x2518, 0x4606, 0x8400}},
    {30166, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x53FC, 0x5090, 0x1090, 0x3492, 0x5292, 0x9294, 0x1098, 0x2090, 0x2090, 0x47FE, 0x8000}},
    {30167, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9200, 0x53FE, 0x5400, 0x1BF8, 0x3288, 0x5248, 0x9FFE, 0x1488, 0x2448, 0x27FC, 0x4008, 0x8030}},
    {30168, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9000, 0x57FC, 0x5000, 0x13F8, 0x3208, 0x5208, 0x93F8, 0x1000, 0x2110, 0x20A0, 0x4FFE, 0x8000}},
    {30169, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5000, 0x5248, 0x1490, 0x3248, 0x5000, 0x97FC, 0x1040, 0x2040, 0x2040, 0x4FFE, 0x8000}},
    {30170, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2088, 0xA7F0, 0x60A0, 0x3FFE, 0x2080, 0x6FF0, 0xA620, 0x3840, 0x27FC, 0x4040, 0x4140, 0x8080}},
    {30171, 2, {0x100, 0x80, 0x1FFE, 0x1000, 0x97F8, 0x50A0, 0x5040, 0x17FC, 0x3444, 0x57FC, 0x9444, 0x17FC, 0x2444, 0x2444, 0x4454, 0x8448}},
    {30172, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9038, 0x53C0, 0x5040, 0x17FC, 0x3150, 0x5248, 0x9446, 0x11F0, 0x2110, 0x2110, 0x4212, 0x8C0E}},
    {30173, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2208, 0xBFFE, 0x6248, 0x2250, 0x2264, 0x6248, 0xA252, 0x2264, 0x244A, 0x4472, 0x48C2, 0x933E}},
    {30174, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5020, 0x5040, 0x11D0, 0x3648, 0x5044, 0x9000, 0x13F8, 0x2208, 0x2208, 0x4208, 0x83F8}},
    {30175, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9248, 0x5150, 0x5040, 0x13F8, 0x3208, 0x53F8, 0x9208, 0x13F8, 0x2208, 0x2208, 0x4228, 0x8210}},
    {30176, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9108, 0x5204, 0x57FE, 0x110A, 0x3284, 0x5482, 0x91F8, 0x1308, 0x2490, 0x2060, 0x4198, 0x8E06}},
    {30177, 2, {0x80, 0x40, 0x1FFE, 0x1050, 0x9048, 0x5FFE, 0x5040, 0x17FC, 0x3444, 0x57FC, 0x9444, 0x17FC, 0x2444, 0x2444, 0x4454, 0x8448}},
    {30178, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x90C4, 0x5704, 0x5114, 0x1114, 0x3FD4, 0x5314, 0x9394, 0x1554, 0x2504, 0x2904, 0x4114, 0x8108}},
    {30179, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x5040, 0x57FC, 0x1040, 0x3040, 0x53FC, 0x9000, 0x1040, 0x2124, 0x252A, 0x450A, 0x88F8}},
    {30180, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9040, 0x5248, 0x5248, 0x1554, 0x38E2, 0x5040, 0x9040, 0x17FC, 0x2040, 0x2040, 0x4040, 0x8FFE}},
    {30181, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9210, 0x5120, 0x57F8, 0x1408, 0x3408, 0x5408, 0x97F8, 0x1120, 0x2120, 0x2222, 0x4422, 0x981E}},
    {30182, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5080, 0x53F8, 0x1088, 0x3108, 0x5FFE, 0x9000, 0x13F8, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {30183, 2, {0x80, 0x40, 0x1FFE, 0x1010, 0x9210, 0x5110, 0x5154, 0x1452, 0x3292, 0x5210, 0x9114, 0x1604, 0x2208, 0x2210, 0x4220, 0x80C0}},
    {30184, 2, {0x80, 0x40, 0x1FFE, 0x1108, 0x9108, 0x57FE, 0x5108, 0x1000, 0x37FE, 0x5402, 0x9040, 0x17FC, 0x2084, 0x2104, 0x4228, 0x8C10}},
    {30185, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x97FC, 0x5444, 0x57FC, 0x1444, 0x37FC, 0x5040, 0x97F8, 0x1208, 0x2110, 0x20E0, 0x4318, 0x8C06}},
    {30186, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x91F0, 0x5210, 0x5420, 0x1BF8, 0x3248, 0x5248, 0x9248, 0x1FFE, 0x20A0, 0x2110, 0x4208, 0x8C06}},
    {30187, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9200, 0x517C, 0x5404, 0x1444, 0x35F4, 0x5444, 0x94E4, 0x1554, 0x2644, 0x2444, 0x4454, 0x8408}},
    {30188, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5208, 0x53F8, 0x1208, 0x33F8, 0x5100, 0x93FC, 0x1494, 0x2924, 0x2244, 0x44A8, 0x8110}},
    {30189, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97BC, 0x54A4, 0x54A4, 0x17BC, 0x34A4, 0x54A4, 0x97BC, 0x14A4, 0x24A4, 0x24A4, 0x4AA4, 0x914C}},
    {30190, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93FC, 0x5200, 0x53F8, 0x1200, 0x33F8, 0x5200, 0x9FFE, 0x1488, 0x2450, 0x2520, 0x4618, 0x8406}},
    {30191, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x57FC, 0x5404, 0x13F0, 0x3210, 0x5210, 0x93F0, 0x1200, 0x23F8, 0x2208, 0x4208, 0x83F8}},
    {30192, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9248, 0x5250, 0x54A0, 0x1110, 0x3208, 0x5C44, 0x9248, 0x1250, 0x24A0, 0x2110, 0x4208, 0x8C06}},
    {30193, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9090, 0x5090, 0x579E, 0x1090, 0x3090, 0x539C, 0x9090, 0x1090, 0x279E, 0x2090, 0x4090, 0x8090}},
    {30194, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x9110, 0x5110, 0x5FFE, 0x1110, 0x3554, 0x5554, 0x9554, 0x1554, 0x2554, 0x2956, 0x4110, 0x8110}},
    {30195, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0xA210, 0x7FBE, 0x2210, 0x2610, 0x6738, 0xAAB4, 0x2A54, 0x3292, 0x4210, 0x4210, 0x8210}},
    {30196, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x9200, 0x53BC, 0x5524, 0x1924, 0x3124, 0x5FE4, 0x9124, 0x1124, 0x22A4, 0x22BC, 0x4424, 0x8800}},
    {30197, 2, {0x80, 0x1FFE, 0x1000, 0x1038, 0x93C0, 0x5040, 0x17FC, 0x3150, 0x5248, 0x9C46, 0x11F0, 0x1020, 0x2FFE, 0x2040, 0x4140, 0x8080}},
    {30198, 2, {0x80, 0x40, 0x1FFE, 0x10A0, 0x90A0, 0x57FC, 0x54A4, 0x14A4, 0x37FC, 0x54A4, 0x94A4, 0x1FFE, 0x2000, 0x2110, 0x4208, 0x8404}},
    {30199, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2080, 0xAFFC, 0x6120, 0x2210, 0x2488, 0x6FF6, 0xB490, 0x27F0, 0x2490, 0x47F2, 0x4082, 0x807E}},
    {30200, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2504, 0xA7C4, 0x6914, 0x3FF4, 0x2114, 0x67D4, 0xA554, 0x2554, 0x2554, 0x45C4, 0x4114, 0x8108}},
    {30201, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5248, 0x53F8, 0x1248, 0x33F8, 0x5000, 0x9FFE, 0x1110, 0x2110, 0x2110, 0x4210, 0x8410}},
    {30202, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9100, 0x57FC, 0x5444, 0x17FC, 0x3484, 0x57FC, 0x9140, 0x1240, 0x3FFE, 0x2040, 0x4040, 0x8040}},
    {30203, 2, {0x80, 0x3FFE, 0x2000, 0x2038, 0xAFC0, 0x6840, 0x2FFC, 0x2820, 0x6B1A, 0xAC06, 0x27F8, 0x2408, 0x27F8, 0x4408, 0x47F8, 0x8408}},
    {30204, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5444, 0x5444, 0x17FC, 0x3444, 0x55F4, 0x9514, 0x1514, 0x25F4, 0x2404, 0x47FC, 0x8404}},
    {30205, 2, {0x80, 0x40, 0x1FFE, 0x1120, 0x9110, 0x53FE, 0x5220, 0x1620, 0x3BFC, 0x5220, 0x9220, 0x13FC, 0x2220, 0x2220, 0x43FE, 0x8200}},
    {30206, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x977E, 0x5504, 0x5574, 0x1654, 0x3554, 0x5554, 0x9554, 0x1574, 0x2754, 0x2404, 0x4414, 0x8408}},
    {30207, 2, {0x80, 0x40, 0x1FFE, 0x1038, 0x97C0, 0x5040, 0x5FFE, 0x1150, 0x3248, 0x5C46, 0x9080, 0x1FFE, 0x2108, 0x2310, 0x40E0, 0x8F1C}},
    {30208, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2410, 0xA210, 0x6FA8, 0x2428, 0x2444, 0x6792, 0xA488, 0x2488, 0x28A0, 0x4890, 0x5288, 0xA108}},
    {30209, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9040, 0x57FC, 0x5110, 0x1110, 0x32A8, 0x5444, 0x9040, 0x1FFE, 0x2040, 0x2040, 0x4040, 0x8040}},
    {30210, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x9FFC, 0x5120, 0x5120, 0x1738, 0x3408, 0x5408, 0x9738, 0x1120, 0x2120, 0x2120, 0x5FFE, 0x8000}},
    {30211, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FFC, 0xA100, 0x6308, 0x2488, 0x2D50, 0x6260, 0xA5E0, 0x2950, 0x2250, 0x4448, 0x5944, 0x8080}},
    {30212, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2FFC, 0xA210, 0x6210, 0x3FFE, 0x2000, 0x67F8, 0xA408, 0x2408, 0x27F8, 0x4408, 0x4408, 0x87F8}},
    {30213, 2, {0x80, 0x40, 0x1FFE, 0x1208, 0x9110, 0x57FC, 0x5444, 0x17FC, 0x3444, 0x57FC, 0x9444, 0x1040, 0x2FFE, 0x2040, 0x4040, 0x8040}},
    {30214, 2, {0x80, 0x40, 0x1FFE, 0x1090, 0x9108, 0x53FC, 0x5040, 0x17FE, 0x3110, 0x5268, 0x9D86, 0x1030, 0x21C0, 0x2018, 0x4060, 0x8380}},
    {30215, 2, {0x80, 0x3FFE, 0x2008, 0x27F0, 0xA040, 0x6FFE, 0x2040, 0x27FC, 0x6444, 0xA7FC, 0x2444, 0x27FC, 0x2040, 0x47FC, 0x4040, 0x8FFE}},
    {30216, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x97BC, 0x5114, 0x5794, 0x1114, 0x37A4, 0x514C, 0x9040, 0x1FFE, 0x20A0, 0x2110, 0x4208, 0x8C06}},
    {30217, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x90A0, 0x5318, 0x5DF6, 0x1000, 0x3784, 0x54A4, 0x97A4, 0x14A4, 0x27A4, 0x2484, 0x4494, 0x8588}},
    {30218, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x22F8, 0xA208, 0x6408, 0x2DFE, 0x3480, 0x64F8, 0xA520, 0x2420, 0x25FE, 0x4450, 0x4488, 0x8506}},
    {30219, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FF8, 0xA808, 0x6868, 0x2B88, 0x2888, 0x6BE8, 0xAAA8, 0x2BE8, 0x288A, 0x50EA, 0x5726, 0xA002}},
    {30220, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x9104, 0x57C4, 0x5114, 0x17D4, 0x3554, 0x57D4, 0x9114, 0x1394, 0x2554, 0x2944, 0x4114, 0x8108}},
    {30221, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5208, 0x53F8, 0x1208, 0x33F8, 0x5000, 0x9FFE, 0x1200, 0x27FC, 0x2924, 0x4254, 0x8488}},
    {30222, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x97FC, 0x5110, 0x51F0, 0x1110, 0x31F0, 0x5110, 0x97FE, 0x1200, 0x2290, 0x2308, 0x4200, 0x83FC}},
    {30223, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9044, 0x53F4, 0x5048, 0x1050, 0x37FE, 0x5080, 0x91F8, 0x1308, 0x25F8, 0x2908, 0x41F8, 0x8108}},
    {30224, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2280, 0xACB8, 0x6888, 0x2888, 0x2EB8, 0x6888, 0xA888, 0x2FF8, 0x2140, 0x4220, 0x4C18, 0xB006}},
    {30225, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x23F0, 0xA210, 0x63D0, 0x2250, 0x2250, 0x6FFC, 0xA804, 0x29E4, 0x2924, 0x49E4, 0x4814, 0x8808}},
    {30226, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FE, 0x5442, 0x5BFC, 0x1040, 0x33F8, 0x5248, 0x93F8, 0x1248, 0x23F8, 0x2040, 0x4FFE, 0x8040}},
    {30227, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x23F0, 0xA420, 0x6FFC, 0x34A4, 0x2514, 0x6444, 0xA040, 0x2FFE, 0x20A0, 0x4110, 0x4208, 0x8C06}},
    {30228, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x9110, 0x5FFE, 0x5110, 0x1040, 0x3FFE, 0x5040, 0x9040, 0x17FC, 0x2404, 0x2404, 0x47FC, 0x8404}},
    {30229, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97BC, 0x5484, 0x5484, 0x17BC, 0x3400, 0x547C, 0x97A4, 0x1424, 0x2428, 0x2790, 0x4428, 0x8446}},
    {30230, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x9040, 0x57FC, 0x5210, 0x1120, 0x3FFE, 0x5000, 0x93F8, 0x1208, 0x23F8, 0x2208, 0x43F8, 0x8208}},
    {30231, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0xA490, 0x62A0, 0x3FFE, 0x2140, 0x6220, 0xAC98, 0x3086, 0x2FF8, 0x4080, 0x4080, 0x9FFC}},
    {30232, 2, {0x80, 0x40, 0x1FFE, 0x1248, 0x9150, 0x5FFE, 0x5150, 0x1248, 0x3444, 0x5080, 0x9FFE, 0x1108, 0x2390, 0x2060, 0x4198, 0x8604}},
    {30233, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x5248, 0x5110, 0x10E0, 0x3318, 0x5C46, 0x93F8, 0x1248, 0x23F8, 0x2044, 0x47FC, 0x8004}},
    {30234, 2, {0x100, 0x80, 0x3FFE, 0x2890, 0x2510, 0xAFBE, 0x6222, 0x2244, 0x2A90, 0x6A90, 0xAF90, 0x2210, 0x2228, 0x4428, 0x4444, 0x8882}},
    {30235, 2, {0x80, 0x40, 0x1FFE, 0x1100, 0x97BC, 0x5114, 0x5394, 0x1114, 0x37A4, 0x514C, 0x9100, 0x1040, 0x2024, 0x2522, 0x450A, 0x88F8}},
    {30236, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x27F8, 0xA408, 0x67F8, 0x2408, 0x27F8, 0x6408, 0xA7F8, 0x2080, 0x2044, 0x4A42, 0x4A12, 0x91F0}},
    {30237, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x54A4, 0x54A4, 0x17FC, 0x3040, 0x5150, 0x9248, 0x1444, 0x2150, 0x2248, 0x4444, 0x8040}},
    {30238, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x27FC, 0xA248, 0x6248, 0x2554, 0x28A2, 0x6318, 0xAC46, 0x23F8, 0x2040, 0x4040, 0x4FFE, 0x8000}},
    {30239, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x27F8, 0xA408, 0x67F8, 0x2408, 0x27F8, 0x6000, 0xAFFC, 0x2924, 0x2924, 0x4924, 0x7FFE, 0x8000}},
    {30240, 2, {0x80, 0x40, 0x1FFE, 0x1444, 0x9248, 0x54A4, 0x5110, 0x1208, 0x37FC, 0x5A0A, 0x93F8, 0x1208, 0x23F8, 0x2208, 0x4228, 0x8210}},
    {30241, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x91B0, 0x5E4E, 0x53F8, 0x1208, 0x33F8, 0x5208, 0x93F8, 0x1200, 0x25FC, 0x2504, 0x49FC, 0x8104}},
    {30242, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2F78, 0xA948, 0x6D48, 0x2B86, 0x3F00, 0x6978, 0xA948, 0x2D48, 0x2B28, 0x4910, 0x4928, 0x9346}},
    {30243, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2100, 0xAFFC, 0x6884, 0x2FFC, 0x2884, 0x6FFC, 0xA148, 0x2154, 0x225C, 0x4446, 0x4842, 0xB03E}},
    {30244, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x977C, 0x5424, 0x55A4, 0x1554, 0x3648, 0x5080, 0x97FC, 0x1444, 0x27FC, 0x2444, 0x47FC, 0x8404}},
    {30245, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x90A0, 0x57FC, 0x5040, 0x13F8, 0x3080, 0x57FC, 0x9100, 0x13F8, 0x2440, 0x2840, 0x47FE, 0x8000}},
    {30246, 2, {0x80, 0x40, 0x1FFE, 0x1140, 0x965C, 0x5444, 0x575C, 0x1444, 0x37FC, 0x5040, 0x97FC, 0x1208, 0x2110, 0x20E0, 0x4318, 0x8C06}},
    {30247, 2, {0x80, 0x3FFE, 0x2080, 0x20FC, 0xA080, 0x6FFE, 0x2882, 0x28F0, 0x6F84, 0xA87C, 0x2800, 0x29FC, 0x2900, 0x4BFE, 0x4900, 0x91FC}},
    {30248, 2, {0x80, 0x1FFE, 0x1040, 0x17FC, 0x9040, 0x53F8, 0x1208, 0x33F8, 0x5208, 0x93F8, 0x1208, 0x13F8, 0x2208, 0x2FFE, 0x4110, 0x8208}},
    {30249, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x9110, 0x57FC, 0x5150, 0x10A0, 0x3110, 0x52E8, 0x9C06, 0x13F8, 0x2208, 0x2208, 0x43F8, 0x8208}},
    {30250, 2, {0x80, 0x3FFE, 0x2100, 0x27F8, 0xA408, 0x67F8, 0x2408, 0x27F8, 0x6408, 0xA7F8, 0x2120, 0x2618, 0x3A66, 0x4388, 0x4208, 0x81F8}},
    {30251, 2, {0x100, 0x80, 0x3FFE, 0x2050, 0x2048, 0xBEFE, 0x7290, 0x2A90, 0x2BFE, 0x6490, 0xA490, 0x2AFE, 0x2A90, 0x5290, 0x40FE, 0x8080}},
    {30252, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x973C, 0x5120, 0x513C, 0x1720, 0x343C, 0x5420, 0x977E, 0x1128, 0x212A, 0x2124, 0x4524, 0x8232}},
    {30253, 2, {0x80, 0x3FFE, 0x2000, 0x3FFE, 0xA120, 0x6FFC, 0x2924, 0x2FFC, 0x6000, 0xA7F8, 0x2000, 0x2FFC, 0x2248, 0x4444, 0x4942, 0x8080}},
    {30254, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x93F8, 0x5210, 0x54A4, 0x17FC, 0x30A0, 0x5318, 0x9C46, 0x1390, 0x2064, 0x2398, 0x4060, 0x8380}},
    {30255, 2, {0x80, 0x40, 0x1FFE, 0x1420, 0x9220, 0x5F7E, 0x5440, 0x14A0, 0x373C, 0x5550, 0x9510, 0x15FE, 0x2510, 0x2928, 0x4B44, 0x9082}},
    {30256, 2, {0x80, 0x3FFE, 0x2000, 0x27FC, 0xA444, 0x67FC, 0x2444, 0x27FC, 0x6100, 0xA208, 0x27F0, 0x20C4, 0x2FFE, 0x4448, 0x4944, 0x9082}},
    {30257, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2FFE, 0xA248, 0x6248, 0x2554, 0x28A2, 0x6318, 0xAC06, 0x2080, 0x2A44, 0x4A4A, 0x520A, 0x81F8}},
    {30258, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x2448, 0xA848, 0x7054, 0x22A4, 0x2500, 0x6C10, 0xB490, 0x249C, 0x2490, 0x4550, 0x463E, 0x8400}},
    {30259, 2, {0x80, 0x3FFE, 0x2000, 0x2F7C, 0xA924, 0x6514, 0x29A4, 0x2160, 0x6618, 0xB886, 0x2320, 0x2C40, 0x2188, 0x4E30, 0x41C0, 0x8E00}},
    {30260, 2, {0x80, 0x3FFE, 0x2040, 0x27FC, 0xA110, 0x6FFE, 0x2000, 0x23F8, 0x6208, 0xA3F8, 0x2208, 0x23F8, 0x2040, 0x4FFE, 0x4040, 0x8040}},
    {30261, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x23BC, 0xA4A4, 0x6AA8, 0x2510, 0x2208, 0x65F6, 0xA800, 0x27FC, 0x2250, 0x4448, 0x4944, 0x8080}},
    {30262, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2220, 0xAFA0, 0x623E, 0x2FC2, 0x2A94, 0x6A90, 0xAF90, 0x2210, 0x2728, 0x4AA8, 0x5244, 0x8282}},
    {30263, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0xAF90, 0x623E, 0x2FA4, 0x2AD4, 0x6A94, 0xAF94, 0x2208, 0x2708, 0x4A94, 0x5224, 0x8242}},
    {30264, 2, {0x80, 0x40, 0x1FFE, 0x1200, 0x97BC, 0x52A4, 0x54A4, 0x15BC, 0x3840, 0x57FC, 0x94A4, 0x1554, 0x2444, 0x24A4, 0x4514, 0x840C}},
    {30265, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2520, 0xA8BC, 0x7744, 0x2204, 0x2224, 0x6F94, 0xA214, 0x2A84, 0x2704, 0x43C4, 0x5C28, 0x8010}},
    {30266, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FFC, 0xA804, 0x6FFC, 0x2800, 0x2FFE, 0x6820, 0xABFE, 0x2B32, 0x2AAA, 0x4B32, 0x4AAA, 0x9206}},
    {30267, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0xA888, 0x7FFE, 0x2888, 0x2FF8, 0x6490, 0xA7F0, 0x2100, 0x3FFE, 0x4410, 0x43E0, 0x9C1C}},
    {30268, 2, {0x80, 0x40, 0x1FFE, 0x1110, 0x97FC, 0x5110, 0x53F8, 0x1208, 0x33F8, 0x5208, 0x93F8, 0x1040, 0x2FFE, 0x20A0, 0x4318, 0x8C06}},
    {30269, 2, {0x80, 0x3FFE, 0x2208, 0x3FFE, 0xA208, 0x63F8, 0x2040, 0x27FC, 0x6444, 0xA7FC, 0x2040, 0x2FFE, 0x2040, 0x47FC, 0x4040, 0x9FFE}},
    {30270, 2, {0x80, 0x3FFE, 0x2040, 0x2E78, 0xAA88, 0x6B10, 0x2CFC, 0x2A04, 0x6AFC, 0xAA04, 0x2BFC, 0x2E20, 0x2814, 0x4952, 0x494A, 0x8A38}},
    {30271, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0xA8A2, 0x6AAA, 0x2AAA, 0x2AAA, 0x6514, 0xA8A2, 0x3FFE, 0x2208, 0x2710, 0x40E0, 0x431C, 0x9C02}},
    {30272, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2FF8, 0xA220, 0x7FFC, 0x2080, 0x2FF8, 0x6888, 0xAFF8, 0x2888, 0x2FF8, 0x4220, 0x4410, 0x8808}},
    {30273, 2, {0x80, 0x40, 0x1FFE, 0x1080, 0x94FC, 0x5500, 0x58FC, 0x1284, 0x32FC, 0x5484, 0x9CFC, 0x1440, 0x24FC, 0x2548, 0x4430, 0x85CE}},
    {30274, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x97FC, 0x50A0, 0x5514, 0x13F8, 0x360C, 0x5BFA, 0x9208, 0x13F8, 0x2040, 0x2444, 0x4942, 0x8080}},
    {30275, 2, {0x80, 0x3FFE, 0x2020, 0x2F3C, 0xA944, 0x6AA8, 0x2C10, 0x2A28, 0x6946, 0xA938, 0x2D50, 0x2A7E, 0x2890, 0x487C, 0x4810, 0x88FE}},
    {30276, 2, {0x80, 0x40, 0x1FFE, 0x1120, 0x9110, 0x53FE, 0x5620, 0x1BFC, 0x3220, 0x53FC, 0x9220, 0x13FE, 0x2200, 0x2524, 0x4492, 0x8892}},
    {30277, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FE, 0x50A0, 0x57BC, 0x14A4, 0x37BC, 0x50A0, 0x97FC, 0x1444, 0x27FC, 0x2444, 0x47FC, 0x8404}},
    {30278, 2, {0x100, 0x80, 0x3FFE, 0x2110, 0x2554, 0xA5B8, 0x6910, 0x22A8, 0x2444, 0x6FFE, 0xA882, 0x3084, 0x2FF8, 0x4108, 0x4228, 0x8C10}},
    {30279, 2, {0x80, 0x3FFE, 0x2000, 0x2F3C, 0xA924, 0x6F3C, 0x2924, 0x2F3C, 0x6804, 0xA9E4, 0x2924, 0x29E4, 0x2924, 0x49E4, 0x4924, 0x8A6C}},
    {30280, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2F50, 0xA124, 0x6A18, 0x2408, 0x3B76, 0x6150, 0xAF8C, 0x2878, 0x2F48, 0x4130, 0x4548, 0x8284}},
    {30281, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0xA8A2, 0x6FBE, 0x2000, 0x27FC, 0x6444, 0xA7FC, 0x2444, 0x27FC, 0x2040, 0x7FFE, 0x4040, 0x8040}},
    {30282, 2, {0x100, 0x80, 0x3FFE, 0x2010, 0x2F28, 0xA944, 0x6A92, 0x2C7C, 0x2A08, 0x6910, 0xA97E, 0x2D00, 0x2AFE, 0x4820, 0x4844, 0x88FE}},
    {30283, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97FC, 0x50A0, 0x53B8, 0x1208, 0x33B8, 0x50A0, 0x9FFE, 0x1040, 0x2124, 0x252A, 0x450A, 0x88F8}},
    {30284, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x93F8, 0x5208, 0x53F8, 0x1000, 0x37BC, 0x54A4, 0x97BC, 0x1040, 0x2444, 0x2444, 0x47FC, 0x8004}},
    {30285, 2, {0x100, 0x80, 0x3FFE, 0x2080, 0x2E4E, 0xA444, 0x65F4, 0x2424, 0x2524, 0x6EAE, 0xA444, 0x2444, 0x24A4, 0x46A4, 0x590E, 0x8200}},
    {30286, 2, {0x80, 0x40, 0x1FFE, 0x1000, 0x97BC, 0x54A4, 0x57BC, 0x14A4, 0x37BC, 0x5404, 0x95F4, 0x1514, 0x25F4, 0x2514, 0x45F4, 0x840C}},
    {30287, 2, {0x80, 0x1FFE, 0x1000, 0x17FC, 0x94A4, 0x57FC, 0x1000, 0x3FFE, 0x5000, 0x93F8, 0x1208, 0x13F8, 0x2144, 0x2328, 0x4D10, 0x818C}},
    {30288, 2, {0x80, 0x1FFE, 0x10A0, 0x1318, 0x9DF6, 0x5000, 0x17FC, 0x3554, 0x54E4, 0x97FC, 0x1000, 0x13F8, 0x2208, 0x23F8, 0x4208, 0x83F8}},
    {30289, 2, {0x80, 0x3FFE, 0x2120, 0x27F8, 0xA528, 0x67F8, 0x2528, 0x2FFE, 0x6800, 0xABFC, 0x2800, 0x2FFE, 0x2A44, 0x4A28, 0x5298, 0xA306}},
    {30290, 2, {0x80, 0x3FFE, 0x2040, 0x21A0, 0xA618, 0x7BF6, 0x2000, 0x27A8, 0x64A8, 0xA7A8, 0x24A8, 0x2788, 0x24A8, 0x4590, 0x4A54, 0x91F2}},
    {30291, 2, {0x100, 0x80, 0x3FFE, 0x2510, 0x2950, 0xB550, 0x67DE, 0x2824, 0x3B94, 0x6814, 0xAB94, 0x2A94, 0x2A88, 0x4AC8, 0x4A94, 0x8C22}},
    {30292, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x97FC, 0x5110, 0x5FFE, 0x1000, 0x33F8, 0x5208, 0x93F8, 0x1208, 0x23F8, 0x2544, 0x452A, 0x88FA}},
    {30293, 2, {0x80, 0x40, 0x1FFE, 0x1040, 0x9020, 0x57FE, 0x5228, 0x127E, 0x34C8, 0x577E, 0x9248, 0x14C8, 0x277E, 0x2248, 0x447E, 0x8840}},
    {30294, 2, {0x100, 0x80, 0x3FFE, 0x2010, 0x2788, 0xA4BE, 0x6480, 0x27A2, 0x2494, 0x643E, 0xAF88, 0x2C88, 0x34BE, 0x4788, 0x4488, 0x8008}},
    {30295, 2, {0x80, 0x3FFE, 0x2000, 0x27FC, 0xA040, 0x6FFE, 0x2842, 0x2358, 0x6040, 0xA358, 0x2000, 0x27FC, 0x2444, 0x47FC, 0x4444, 0x87FC}},
    {30296, 2, {0x80, 0x1FFE, 0x1110, 0x1FFE, 0x9110, 0x53F8, 0x1248, 0x33F8, 0x5248, 0x93F8, 0x1040, 0x17FC, 0x2454, 0x25F4, 0x4414, 0x840C}},
    {30297, 2, {0x100, 0x80, 0x3FFE, 0x2100, 0x2E78, 0xA808, 0x6F78, 0x2808, 0x2FF8, 0x6888, 0xACC8, 0x2AA8, 0x2CCA, 0x4AAA, 0x4CC6, 0x8882}},
    {30298, 2, {0x80, 0x3FFE, 0x2200, 0x23F8, 0xA410, 0x6FFE, 0x3524, 0x26FA, 0x6400, 0xA5FC, 0x2400, 0x25FC, 0x2400, 0x49FC, 0x4904, 0x91FC}},
    {30299, 2, {0x80, 0x3FFE, 0x2040, 0x2FFE, 0xA000, 0x67FC, 0x2404, 0x25F4, 0x6514, 0xA7FC, 0x2008, 0x23F0, 0x2040, 0x4FFE, 0x4150, 0x8E4E}},
    {30300, 2, {0x100, 0x80, 0x3FFE, 0x2000, 0x2FDC, 0xA854, 0x6FD4, 0x2AA2, 0x2A80, 0x6FDC, 0xAA94, 0x2FD4, 0x2808, 0x5288, 0x5454, 0xA862}},
    {30301, 2, {0x80, 0x3FFE, 0x2040, 0x2FFE, 0xA000, 0x67FC, 0x2404, 0x25F4, 0x6514, 0xA7FC, 0x2000, 0x27FC, 0x2000, 0x4FFE, 0x4248, 0x84C4}},
    {30302, 2, {0x80, 0x3FFE, 0x2210, 0x2F9C, 0xA224, 0x6248, 0x2FBE, 0x2AA2, 0x6AAA, 0xAFAA, 0x222A, 0x272A, 0x2AAA, 0x5294, 0x4222, 0x8242}},
    {30303, 2, {0x80, 0x3FFE, 0x2100, 0x27F8, 0xA648, 0x6528, 0x2498, 0x27F8, 0x6210, 0xA5E8, 0x3806, 0x27F8, 0x2548, 0x47F8, 0x4548, 0x8558}},
    {30304, 2, {0x100, 0x80, 0x3FFE, 0x2040, 0x2FFE, 0xA0A0, 0x6F5E, 0x2554, 0x2554, 0x6B5A, 0xA208, 0x23F8, 0x2208, 0x43F8, 0x4408, 0x8808}},
    {30305, 2, {0x100, 0x80, 0x3FFE, 0x2480, 0x273C, 0xA484, 0x6394, 0x2408, 0x277E, 0x6A0A, 0xA228, 0x2FAE, 0x2228, 0x4558, 0x484E, 0x9080}},
    {30306, 2, {0x80, 0x1FFE, 0x1208, 0x1110, 0x97FC, 0x5040, 0x13F8, 0x3080, 0x5FFE, 0x9210, 0x17F8, 0x1A16, 0x23F0, 0x2264, 0x4318, 0x8204}},
    {30307, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2744, 0xA928, 0x727E, 0x2F90, 0x2A90, 0x6FBC, 0xAA90, 0x2F90, 0x207E, 0x4390, 0x5C10, 0x8010}},
    {30308, 2, {0x80, 0x40, 0x1FFE, 0x1210, 0x93DE, 0x5528, 0x5884, 0x17BC, 0x34A4, 0x57A4, 0x94A4, 0x17A4, 0x2434, 0x2528, 0x46A0, 0x8420}},
    {30309, 2, {0x100, 0x80, 0x3FFE, 0x2510, 0x2950, 0xB550, 0x67DE, 0x2824, 0x3B94, 0x6814, 0xAFD4, 0x2914, 0x2B88, 0x4908, 0x4FD4, 0x8822}},
    {30310, 2, {0x80, 0x3FFE, 0x2000, 0x27FC, 0xA554, 0x64E4, 0x27FC, 0x2040, 0x67FC, 0xA040, 0x2FFE, 0x24A4, 0x2852, 0x47FC, 0x4040, 0x8FFE}},
    {30311, 2, {0x80, 0x3FFE, 0x2000, 0x2FFC, 0xA888, 0x6B30, 0x2910, 0x2FBC, 0x6BB8, 0xAD54, 0x2912, 0x2840, 0x2978, 0x4940, 0x4940, 0x97FC}},
    {30312, 2, {0x80, 0x3FFE, 0x2000, 0x27FC, 0xA040, 0x6FFE, 0x2842, 0x2358, 0x6220, 0xA7FE, 0x2C20, 0x37FC, 0x2420, 0x47FC, 0x4420, 0x87FE}},
    {30313, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x221E, 0xBFA2, 0x6244, 0x2FBE, 0x2AA2, 0x6ABE, 0xAFA2, 0x223E, 0x2722, 0x4ABE, 0x5294, 0x8222}},
    {30314, 2, {0x80, 0x3FFE, 0x2010, 0x237C, 0xAE10, 0x627C, 0x2F10, 0x22FE, 0x6644, 0xA77C, 0x2AC4, 0x2A7C, 0x3244, 0x427C, 0x4228, 0x8244}},
    {30315, 2, {0x80, 0x3FFE, 0x2400, 0x3FBE, 0xA408, 0x6F10, 0x293E, 0x2F22, 0x692A, 0xAF2A, 0x292A, 0x2F2A, 0x292A, 0x7F94, 0x4912, 0x90A2}},
    {30316, 2, {0x100, 0x80, 0x3FFE, 0x2400, 0x2722, 0xA914, 0x723E, 0x2F88, 0x2A88, 0x6F9C, 0xAA88, 0x2F88, 0x203E, 0x5548, 0x5548, 0x8008}},
    {30317, 2, {0x80, 0x3FFE, 0x2000, 0x2FBE, 0xA8A2, 0x6FBE, 0x28A2, 0x2FBE, 0x68A2, 0xAFBE, 0x2514, 0x28A2, 0x2FFE, 0x4310, 0x40E0, 0x8F1C}},
    {30318, 2, {0x80, 0x3FFE, 0x2004, 0x2EF8, 0xAA54, 0x6AFE, 0x2C10, 0x2AFE, 0x6A04, 0xAAFC, 0x2A04, 0x2EFC, 0x2820, 0x4954, 0x4A4A, 0x8838}},
    {30319, 2, {0x80, 0x1FFE, 0x14A4, 0x17BC, 0x94A4, 0x57BC, 0x14A4, 0x37BC, 0x5120, 0x93FE, 0x1220, 0x17FC, 0x2A20, 0x23FC, 0x4220, 0x83FE}},
    {30320, 2, {0x80, 0x3FFE, 0x2000, 0x2AA8, 0xB524, 0x6ABE, 0x2F68, 0x2968, 0x6FBE, 0xA028, 0x2FA8, 0x2ABE, 0x2FA8, 0x4828, 0x48BE, 0x87A0}},
    {30321, 2, {0x80, 0x3FFE, 0x2518, 0x2F94, 0xA53E, 0x6728, 0x2268, 0x2FBE, 0x6AA8, 0xAFA8, 0x223E, 0x2FA8, 0x2228, 0x5FBE, 0x4520, 0x88A0}},
    {30322, 2, {0x80, 0x3FFE, 0x2200, 0x3FBE, 0xA208, 0x6F90, 0x28BE, 0x2FA2, 0x68BE, 0xAFA2, 0x28BE, 0x2FA2, 0x28BE, 0x5FC0, 0x4914, 0x90A2}},
    {30323, 2, {0x80, 0x3FFE, 0x2444, 0x27FC, 0xA444, 0x6FFE, 0x2AAA, 0x2FBE, 0x6AAA, 0xAFBE, 0x2208, 0x27F0, 0x20C4, 0x4FFE, 0x4248, 0x84C4}},
    {30324, 2, {0x80, 0x3FFE, 0x2444, 0x2AEA, 0xAE0E, 0x64E4, 0x2AAA, 0x2EEE, 0x6000, 0xA7FC, 0x2040, 0x27FC, 0x2040, 0x4FFE, 0x4040, 0x80C0}},
    {30325, 2, {0x80, 0x3FFE, 0x2444, 0x28E8, 0xAE0E, 0x64E4, 0x2A0A, 0x2EEE, 0x60A0, 0xAAEA, 0x2040, 0x2FFE, 0x28A2, 0x4912, 0x4842, 0x88A6}},
    {30326, 2, {0x88, 0x7C90, 0x4A0, 0x4C0, 0x4482, 0x2444, 0x2448, 0x850, 0x820, 0x1020, 0x1010, 0x2008, 0x4004, 0x8002, 0x0, 0x0}},
    {30327, 2, {0x90, 0x7EA0, 0x244, 0x4448, 0x2830, 0x1010, 0x2008, 0xDFF6, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {30328, 2, {0x90, 0x7EA0, 0x244, 0x4448, 0x2830, 0x1010, 0x2008, 0xCFE6, 0x100, 0x100, 0x7FFC, 0x100, 0x2C0, 0x420, 0x1810, 0x6008}},
    {30329, 2, {0x90, 0x7EA0, 0x244, 0x4448, 0x2830, 0x1010, 0x2FE8, 0xC826, 0x820, 0x101C, 0x2FE0, 0x20, 0xC40, 0x380, 0xC70, 0x7008}},
    {30330, 2, {0x90, 0x7EA0, 0x244, 0x4448, 0x2830, 0x1010, 0x2008, 0xDFF6, 0x440, 0x440, 0x7FFC, 0x440, 0x840, 0x844, 0x1044, 0x603C}},
    {30331, 2, {0x80, 0x3E90, 0x2A4, 0x2448, 0x1450, 0x820, 0x17D0, 0x2008, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x820, 0x440, 0x7FFC, 0x0}},
    {30332, 2, {0x90, 0x7EA0, 0x444, 0x2828, 0x1010, 0x3CF8, 0xC496, 0x490, 0x3D0C, 0x2000, 0x41F8, 0x7C08, 0x4D0, 0x420, 0x2850, 0x1188}},
    {30333, 2, {0x100, 0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {30334, 2, {0x0, 0xFFFE, 0x100, 0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {30335, 2, {0x820, 0x820, 0x1020, 0x3F20, 0x2120, 0x2120, 0x2120, 0x2120, 0x3F20, 0x2120, 0x2120, 0x2122, 0x2122, 0x3F22, 0x211E, 0x0}},
    {30336, 2, {0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x0, 0x2038, 0x23C0, 0x3C04, 0x2004, 0x2004, 0x1FFC}},
    {30337, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {30338, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x200, 0x200, 0x3FC, 0xFE00, 0x204, 0x204, 0x1FC}},
    {30339, 2, {0x400, 0x800, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {30340, 2, {0x1040, 0x1040, 0x2040, 0x7E7C, 0x4284, 0x4284, 0x4304, 0x4244, 0x7E24, 0x4224, 0x4204, 0x4204, 0x4204, 0x7E04, 0x4228, 0x10}},
    {30341, 2, {0x1000, 0x1000, 0x21FC, 0x7D24, 0x4524, 0x4524, 0x4524, 0x4524, 0x7DFC, 0x4500, 0x4500, 0x4500, 0x4502, 0x7D02, 0x44FE, 0x0}},
    {30342, 2, {0x2080, 0x2088, 0x3CB0, 0x20C0, 0x2084, 0x2C84, 0x317C, 0x200, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {30343, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {30344, 2, {0x1004, 0x100E, 0x20F0, 0x7C80, 0x4480, 0x44FC, 0x44A4, 0x44A4, 0x7CA4, 0x44A8, 0x44A8, 0x4490, 0x4510, 0x7D28, 0x4644, 0x82}},
    {30345, 2, {0x1090, 0x1090, 0x2090, 0x7A90, 0x4A92, 0x4AD4, 0x4A98, 0x4A90, 0x7A90, 0x4A90, 0x4A90, 0x4A92, 0x4AD2, 0x7F12, 0x4A0E, 0x0}},
    {30346, 2, {0x1020, 0x1020, 0x2050, 0x7C50, 0x4488, 0x4524, 0x4612, 0x4410, 0x7DFC, 0x4404, 0x4408, 0x4488, 0x4450, 0x7C20, 0x4410, 0x10}},
    {30347, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x440, 0x920, 0x3118, 0xCFE6, 0x100, 0x100, 0x100}},
    {30348, 2, {0x1020, 0x1020, 0x2020, 0x7DFE, 0x4420, 0x4420, 0x4420, 0x45FC, 0x7C70, 0x44A8, 0x44A8, 0x4524, 0x4524, 0x7E22, 0x4420, 0x20}},
    {30349, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2000, 0x20F8, 0x3E88, 0x2088, 0x2088, 0x26A8, 0x3890, 0x80, 0x80}},
    {30350, 2, {0x1040, 0x1020, 0x2020, 0x7BFE, 0x4800, 0x4888, 0x4904, 0x4A02, 0x7888, 0x4888, 0x4850, 0x4820, 0x4850, 0x7888, 0x4904, 0x602}},
    {30351, 2, {0x1084, 0x1048, 0x2000, 0x7CFC, 0x4448, 0x4448, 0x4448, 0x4448, 0x7DFE, 0x4448, 0x4448, 0x4448, 0x4448, 0x7C88, 0x4488, 0x108}},
    {30352, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2108, 0x1110, 0x920, 0x1110, 0x2108, 0xFFFE, 0x100, 0x100, 0x100}},
    {30353, 2, {0x1020, 0x1124, 0x2124, 0x7D24, 0x45FC, 0x4400, 0x45FC, 0x4404, 0x7C04, 0x45FC, 0x4500, 0x4500, 0x4502, 0x7D02, 0x44FE, 0x0}},
    {30354, 2, {0x1010, 0x10D8, 0x2394, 0x7894, 0x4890, 0x4BFE, 0x4890, 0x4894, 0x7894, 0x48D8, 0x4B98, 0x4890, 0x48AA, 0x78CA, 0x4A86, 0x102}},
    {30355, 2, {0x1020, 0x1120, 0x2120, 0x7DFC, 0x4520, 0x4620, 0x4420, 0x47FE, 0x7C00, 0x4400, 0x45FC, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x104}},
    {30356, 2, {0x1000, 0x11FC, 0x2104, 0x7DFC, 0x4504, 0x45FC, 0x4400, 0x4400, 0x7DFC, 0x4420, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x4420, 0x20}},
    {30357, 2, {0x0, 0x0, 0xFEFE, 0x1020, 0x1020, 0x2040, 0x7EFC, 0x4284, 0x4284, 0x4284, 0x7EFC, 0x4284, 0x4284, 0x4284, 0x7EFC, 0x4284}},
    {30358, 2, {0x1040, 0x1020, 0x23FE, 0x7A02, 0x4C04, 0x49F8, 0x4800, 0x4800, 0x7BFE, 0x4890, 0x4890, 0x4890, 0x4912, 0x7912, 0x4A0E, 0x400}},
    {30359, 2, {0x1000, 0x11FC, 0x2124, 0x7924, 0x4974, 0x4924, 0x49FC, 0x4904, 0x7974, 0x4954, 0x4954, 0x4974, 0x4904, 0x7A04, 0x4A14, 0x408}},
    {30360, 2, {0x1020, 0x1020, 0x23FE, 0x7C20, 0x45FC, 0x4420, 0x47FE, 0x4400, 0x7DFC, 0x4504, 0x45FC, 0x4504, 0x45FC, 0x7D04, 0x4514, 0x108}},
    {30361, 2, {0x80C, 0x8F0, 0x7E80, 0x1880, 0x2CFE, 0x2A88, 0x4888, 0x8888, 0x908, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30362, 2, {0x1020, 0x1124, 0x2124, 0x7924, 0x49FC, 0x4800, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x4888, 0x7850, 0x4BFE, 0x0}},
    {30363, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x810, 0x1020, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x4284, 0x4284, 0x7EFC}},
    {30364, 2, {0x1020, 0x1010, 0x21FE, 0x7C00, 0x44FC, 0x4484, 0x44FC, 0x4400, 0x7DFE, 0x4502, 0x457A, 0x454A, 0x457A, 0x7D02, 0x450A, 0x104}},
    {30365, 2, {0x1020, 0x2020, 0x7D24, 0x44A4, 0x7CA8, 0x4420, 0x7DFE, 0x50, 0xFE50, 0x1050, 0x1050, 0x7C52, 0x1052, 0x1E92, 0xF08E, 0x4100}},
    {30366, 2, {0x1020, 0x1040, 0x21FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4840, 0x7BFE, 0x4888, 0x4924, 0x4A22, 0x4DFC, 0x7820, 0x4820, 0x20}},
    {30367, 2, {0x1020, 0x13FE, 0x2020, 0x7DFC, 0x4420, 0x47FE, 0x4400, 0x45FC, 0x7D04, 0x45FC, 0x4504, 0x45FC, 0x4504, 0x7DFC, 0x4488, 0x104}},
    {30368, 2, {0x1020, 0x1222, 0x23FE, 0x7890, 0x4888, 0x49FE, 0x4B10, 0x4D10, 0x79FE, 0x4910, 0x4910, 0x49FE, 0x4910, 0x7910, 0x49FE, 0x100}},
    {30369, 2, {0x1020, 0x1040, 0x21FC, 0x7D04, 0x45FC, 0x4504, 0x45FC, 0x4420, 0x7D24, 0x44A8, 0x4524, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x20}},
    {30370, 2, {0x1020, 0x1020, 0x21FC, 0x7820, 0x4BFE, 0x4908, 0x4B9C, 0x4908, 0x7988, 0x4E3E, 0x4800, 0x4BFE, 0x4890, 0x7890, 0x4912, 0x20E}},
    {30371, 2, {0x1088, 0x13FE, 0x2088, 0x7800, 0x49FC, 0x48A8, 0x48A8, 0x4BFE, 0x78A8, 0x48A8, 0x49FC, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20}},
    {30372, 2, {0x103C, 0x13E0, 0x2124, 0x78A8, 0x4BFE, 0x48A8, 0x4924, 0x4A02, 0x79FC, 0x4924, 0x4924, 0x49FC, 0x4924, 0x7924, 0x49FC, 0x104}},
    {30373, 2, {0x1040, 0x11FC, 0x2104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4924, 0x48A8, 0x4924, 0x4820, 0x7BFE, 0x4820, 0x20}},
    {30374, 2, {0x1088, 0x1108, 0x23C8, 0x7A50, 0x4BDE, 0x4A64, 0x4BD4, 0x4914, 0x7894, 0x4FF4, 0x4914, 0x49C8, 0x4948, 0x7A54, 0x4B54, 0x4A2}},
    {30375, 2, {0x103C, 0x13C0, 0x2044, 0x7A24, 0x4908, 0x4FFE, 0x4C42, 0x4AA4, 0x7A8A, 0x4C7A, 0x4880, 0x48F8, 0x4908, 0x7A90, 0x4860, 0x79E}},
    {30376, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2040, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {30377, 2, {0x1000, 0x21FC, 0x7D04, 0x45FC, 0x7D04, 0x45FC, 0x7C20, 0x124, 0xFCA8, 0x1020, 0x13FE, 0x7C90, 0x1090, 0x1D12, 0xE212, 0x440E}},
    {30378, 2, {0x1124, 0x1174, 0x2258, 0x7ADA, 0x4BFE, 0x4954, 0x4ADA, 0x4BFE, 0x78A2, 0x4820, 0x4BFE, 0x4870, 0x48A8, 0x7924, 0x4A22, 0x20}},
    {30379, 2, {0x1040, 0x1020, 0x23FE, 0x7A50, 0x4BFE, 0x4A52, 0x4BFE, 0x4A00, 0x7A94, 0x4AD8, 0x4A90, 0x4AD2, 0x4A8E, 0x7C00, 0x4954, 0x22A}},
    {30380, 2, {0x11FC, 0x1020, 0x23FE, 0x7A22, 0x4DAC, 0x4820, 0x49AC, 0x4890, 0x79FE, 0x4B20, 0x4DFC, 0x4920, 0x49FC, 0x7920, 0x49FE, 0x100}},
    {30381, 2, {0x101E, 0x13E0, 0x2122, 0x7894, 0x4BFE, 0x4A52, 0x4BFE, 0x4804, 0x7BC4, 0x4A7E, 0x4BC4, 0x4A54, 0x4BCC, 0x7A84, 0x4AD4, 0x348}},
    {30382, 2, {0x80, 0x80, 0x80, 0x3FFC, 0x2084, 0x2088, 0x2080, 0x2FF0, 0x2410, 0x2410, 0x2220, 0x2140, 0x2080, 0x4140, 0x4630, 0x980E}},
    {30383, 2, {0x800, 0x87C, 0x810, 0x7F10, 0x4910, 0x4A10, 0x4810, 0x7EFE, 0x4210, 0x4210, 0x5210, 0x4C10, 0x4410, 0x4A10, 0x9210, 0x2010}},
    {30384, 2, {0x820, 0x820, 0x87E, 0x7F42, 0x4982, 0x4A7A, 0x484A, 0x7E4A, 0x427A, 0x4242, 0x524A, 0x4C44, 0x4442, 0x4A42, 0x923E, 0x2000}},
    {30385, 2, {0x2010, 0x2010, 0x3E10, 0x42FE, 0x8492, 0x894, 0xFE90, 0x2FC, 0x2A4, 0x7EA4, 0x2A8, 0x2A8, 0x290, 0xFF28, 0x244, 0x82}},
    {30386, 2, {0x10, 0xFF10, 0x9110, 0x10FE, 0xFE92, 0x2094, 0x2890, 0x48FC, 0x7EA4, 0x8A4, 0x8A8, 0xEA8, 0xF890, 0x4928, 0x944, 0xA82}},
    {30387, 2, {0x1010, 0x1410, 0x1210, 0x10FE, 0xFE92, 0x1094, 0x1090, 0x92FC, 0x54A4, 0x10A4, 0x38A8, 0x54A8, 0x9290, 0x1128, 0x5144, 0x2282}},
    {30388, 2, {0x1010, 0x1010, 0x2010, 0x44FE, 0xFE92, 0x2894, 0x4490, 0xA2FC, 0x3CA4, 0x44A4, 0x44A8, 0xA8A8, 0x1090, 0x2928, 0x4544, 0x8282}},
    {30389, 2, {0x2410, 0x2410, 0x7E10, 0x24FE, 0x2492, 0xFF94, 0x90, 0x7EFC, 0x42A4, 0x42A4, 0x7EA8, 0x42A8, 0x4290, 0x7F28, 0x4244, 0x82}},
    {30390, 2, {0x1010, 0x1010, 0xFE10, 0x38FE, 0x5492, 0x9294, 0x1090, 0x7CFC, 0x44A4, 0x7CA4, 0x44A8, 0x7CA8, 0x90, 0x1D28, 0xE144, 0x4282}},
    {30391, 2, {0x810, 0x810, 0xFF10, 0x8FE, 0x892, 0x7E94, 0x90, 0x7EFC, 0x42A4, 0x42A4, 0x7EA8, 0x42A8, 0x2490, 0xF28, 0xF144, 0x4282}},
    {30392, 2, {0x10, 0xFE10, 0x9210, 0x10FE, 0xFE92, 0x1094, 0x7C90, 0x54FC, 0x7CA4, 0x54A4, 0x7CA8, 0x10A8, 0xFE90, 0x1128, 0x1144, 0x1282}},
    {30393, 2, {0x800, 0x8FE, 0x892, 0x7F10, 0x49FE, 0x4A10, 0x487C, 0x7E54, 0x427C, 0x4254, 0x527C, 0x4C10, 0x44FE, 0x4A10, 0x9210, 0x2010}},
    {30394, 2, {0x4010, 0x4010, 0xFE10, 0x22FE, 0xAA92, 0xFA94, 0x4290, 0x94FC, 0x48A4, 0xFEA4, 0x22A8, 0xAAA8, 0xFA90, 0x4328, 0x9444, 0x882}},
    {30395, 2, {0x1010, 0x1E10, 0x1010, 0x7EFE, 0x5292, 0x5C94, 0x7290, 0x4EFC, 0x40A4, 0x5CA4, 0x54A8, 0x5CA8, 0x5490, 0x5D28, 0x9644, 0x3882}},
    {30396, 2, {0x1010, 0xFE10, 0x1010, 0x7CFE, 0x92, 0xFE94, 0x8290, 0x7CFC, 0xA4, 0x7CA4, 0x44A8, 0x7CA8, 0x4490, 0x2928, 0xF44, 0xF282}},
    {30397, 2, {0x1010, 0xFE10, 0x10, 0xFEFE, 0x8292, 0xBA94, 0xAA90, 0xFEFC, 0xA4, 0x7CA4, 0x44A8, 0x7CA8, 0x4490, 0x7D28, 0x44, 0xFE82}},
    {30398, 2, {0x1010, 0xFE10, 0x1010, 0x7CFE, 0x92, 0xFE94, 0xAA90, 0xFEFC, 0x44A4, 0x7CA4, 0x44A8, 0x7CA8, 0x4490, 0x7D28, 0x2844, 0x4482}},
    {30399, 2, {0x0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0, 0x0}},
    {30400, 2, {0x100, 0x100, 0x180, 0x140, 0x120, 0x110, 0x108, 0x100, 0x100, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30401, 2, {0x0, 0x3FF0, 0x810, 0x820, 0x83C, 0x1004, 0x1004, 0x2014, 0x4008, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30402, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x500, 0x200, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30403, 2, {0x0, 0x7FFC, 0x100, 0x300, 0x570, 0x90C, 0x3102, 0xC100, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30404, 2, {0x0, 0x7FF8, 0x108, 0x3FF8, 0x2100, 0x3FFC, 0x104, 0x114, 0x108, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30405, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30406, 2, {0x440, 0x820, 0x1010, 0x2FE8, 0xC426, 0x420, 0x820, 0x10A0, 0x6040, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30407, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0xA20, 0x3118, 0xC006, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30408, 2, {0x0, 0x3FF0, 0x810, 0xFA0, 0x8BC, 0x1284, 0x1104, 0x2294, 0x4448, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30409, 2, {0x70, 0x1F80, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC006, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30410, 2, {0x1010, 0x820, 0x0, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008, 0x4004, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30411, 2, {0x240, 0x220, 0x7FFC, 0x400, 0xBF0, 0x1120, 0x20C0, 0x4330, 0x1C0E, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30412, 2, {0x1000, 0x10F8, 0x3E88, 0x4288, 0xA2A8, 0x1490, 0x884, 0x1084, 0x207C, 0xC000, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30413, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x400, 0x820, 0x1FF0, 0x10, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30414, 2, {0x100, 0x100, 0x1FF0, 0x1110, 0x1110, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30415, 2, {0x240, 0x220, 0x3F8, 0x7E00, 0x3FC, 0xFE10, 0x160, 0x382, 0x1C62, 0xE01E, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE}},
    {30416, 2, {0x820, 0x820, 0x7F20, 0x828, 0x824, 0xF22, 0xF020, 0x4020, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30417, 2, {0x440, 0x2440, 0x247C, 0x2440, 0x2490, 0x2488, 0x2508, 0x400, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30418, 2, {0x100, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30419, 2, {0x1000, 0x9FC, 0x4020, 0x2020, 0xBFE, 0x1020, 0x7020, 0x1020, 0x10A0, 0x40, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30420, 2, {0x400, 0x400, 0x7FFC, 0x840, 0x1248, 0x1250, 0x24A0, 0x4110, 0x8208, 0xC06, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE}},
    {30421, 2, {0x1000, 0x9F0, 0x4110, 0x2590, 0x550, 0x950, 0x7112, 0x1112, 0x120E, 0x1400, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30422, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE}},
    {30423, 2, {0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30424, 2, {0x200, 0x400, 0x1FF0, 0x1110, 0x1090, 0xFFFE, 0x1010, 0x1210, 0x2150, 0x4020, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30425, 2, {0x120, 0x110, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30426, 2, {0x110, 0x108, 0x7FFC, 0x100, 0x2118, 0x11A0, 0x740, 0x1920, 0xE118, 0x4506, 0x200, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30427, 2, {0x90, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3E48, 0x2250, 0x2224, 0x4A54, 0x448C, 0x8104, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30428, 2, {0x80, 0x2080, 0x11FC, 0x8204, 0x5448, 0x2040, 0xE0A0, 0x2318, 0x2C06, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30429, 2, {0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x1110, 0x9A0, 0x540, 0x1930, 0xE50E, 0x200, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30430, 2, {0x210, 0x2FC, 0x7F40, 0x84, 0x764, 0x781C, 0x210, 0x2FC, 0x7F40, 0x84, 0x764, 0x781C, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {30431, 2, {0x0, 0x7CFC, 0x4484, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7C84, 0x114, 0x208, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30432, 2, {0x800, 0xFE0, 0x1020, 0x1FC0, 0x40, 0x7FFC, 0x600, 0x1920, 0x6C0, 0x39B0, 0x68C, 0x3980, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {30433, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7FFC, 0x2488, 0x4244, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30434, 2, {0x20, 0x7C28, 0x5424, 0x55FE, 0x7C20, 0x5450, 0x5450, 0x7C88, 0x104, 0x202, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30435, 2, {0x0, 0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30436, 2, {0x1078, 0x3E48, 0x2248, 0x2A86, 0xFF78, 0x2248, 0x2A28, 0x4210, 0x4A28, 0x8446, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30437, 2, {0x1100, 0x611C, 0x4D44, 0x4544, 0x759C, 0x4984, 0x4944, 0x755C, 0x4204, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30438, 2, {0x100, 0x6C0, 0x1930, 0xE7CE, 0x80, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x37B8, 0x2008, 0x3FF8, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {30439, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2490, 0x27F0, 0x2490, 0x27F0, 0x2000, 0x4FF8, 0x4948, 0xBFFE}},
    {30440, 2, {0xBFE, 0x1040, 0x61FC, 0x904, 0x11FC, 0x6104, 0x5FC, 0x904, 0x31FC, 0xC088, 0x104, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30441, 2, {0x820, 0x7F20, 0x83E, 0xFF44, 0x22A4, 0x7F28, 0x828, 0xFF10, 0x828, 0x846, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30442, 2, {0x23F8, 0x1208, 0x93F8, 0x4208, 0x43F8, 0x1000, 0x2FFE, 0xE200, 0x27FC, 0x2924, 0x2254, 0x488, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {30443, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x7FFC, 0x440, 0x1FF0, 0x1450, 0x1BB0, 0x1010, 0x3FF8, 0x2448, 0xFFFE}},
    {30444, 2, {0x40, 0x7E7C, 0x4890, 0x7E10, 0x42FE, 0x7E10, 0x487C, 0x4844, 0x7E7C, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {30445, 2, {0x1020, 0x2220, 0x7C3E, 0xA44, 0x7FA4, 0x824, 0x7F28, 0x1428, 0x7F10, 0x828, 0x7F44, 0x882, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {30446, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {30447, 2, {0x0, 0x0, 0x7BFE, 0x4820, 0x4820, 0x7820, 0x4820, 0x4820, 0x7820, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0xA0, 0x40}},
    {30448, 2, {0x0, 0x1FC, 0x7C20, 0x4420, 0x4420, 0x7C20, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {30449, 2, {0x0, 0x1FC, 0x7C20, 0x4420, 0x4420, 0x7C20, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0xA0, 0x40}},
    {30450, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30451, 2, {0x20, 0x10, 0x7C10, 0x4400, 0x45FE, 0x7C80, 0x4480, 0x4480, 0x7C80, 0x4480, 0x4480, 0x4480, 0x7C80, 0x4480, 0xFC, 0x0}},
    {30452, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {30453, 2, {0x40, 0x40, 0x7C80, 0x44FE, 0x4500, 0x7E00, 0x44FC, 0x4408, 0x7C10, 0x4420, 0x4440, 0x4480, 0x7D02, 0x4502, 0xFE, 0x0}},
    {30454, 2, {0x0, 0x0, 0x7DFC, 0x4400, 0x4400, 0x7C00, 0x47FE, 0x4490, 0x7C90, 0x4490, 0x4490, 0x4490, 0x7D12, 0x4512, 0x20E, 0x400}},
    {30455, 2, {0x40, 0x40, 0x7C80, 0x44FC, 0x4504, 0x7E04, 0x4484, 0x4444, 0x7C44, 0x4414, 0x4424, 0x4444, 0x7D84, 0x4404, 0x28, 0x10}},
    {30456, 2, {0x800, 0x8FC, 0x884, 0x884, 0xFE84, 0x8FC, 0x1884, 0x1C84, 0x2A84, 0x2AFC, 0x4884, 0x8884, 0x884, 0x884, 0x8FC, 0x884}},
    {30457, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x7924, 0x4924, 0x4924, 0x7924, 0x49FC, 0x4824, 0x4820, 0x7822, 0x4822, 0x1E, 0x0}},
    {30458, 2, {0x8, 0x1C, 0x7CE0, 0x4480, 0x4480, 0x7C80, 0x44FE, 0x4488, 0x7C88, 0x4488, 0x4488, 0x4488, 0x7D08, 0x4508, 0x208, 0x408}},
    {30459, 2, {0x90, 0x90, 0x7908, 0x4908, 0x4A04, 0x7C02, 0x4BFC, 0x4880, 0x7900, 0x49F8, 0x4808, 0x4808, 0x7808, 0x4808, 0x50, 0x20}},
    {30460, 2, {0x10, 0x90, 0x7890, 0x4888, 0x4908, 0x7904, 0x4A04, 0x4DFA, 0x7888, 0x4888, 0x4888, 0x4888, 0x7908, 0x4908, 0x228, 0x410}},
    {30461, 2, {0x20, 0x20, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x4420, 0x45FC, 0x7C20, 0x4420, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x20, 0x20}},
    {30462, 2, {0x7C, 0x3F80, 0x2080, 0x3FFC, 0x2080, 0x2FF8, 0x2808, 0x2808, 0x2FF8, 0x2808, 0x2808, 0x2FF8, 0x4808, 0x4808, 0x8FF8, 0x808}},
    {30463, 2, {0x40, 0x20, 0x7C20, 0x47FE, 0x4488, 0x7C88, 0x4488, 0x4488, 0x7C88, 0x4450, 0x4450, 0x4420, 0x7C50, 0x4488, 0x104, 0x602}},
    {30464, 2, {0x0, 0xFC, 0x7C84, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7C84, 0x44FC, 0x4484, 0x4484, 0x7D04, 0x4504, 0x214, 0x408}},
    {30465, 2, {0x100, 0x1110, 0x1108, 0x2124, 0x40C0, 0x300, 0xC00, 0x3FF8, 0xD008, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {30466, 2, {0x4, 0x1E, 0x7DF0, 0x4510, 0x4510, 0x7D10, 0x4510, 0x45FE, 0x7D10, 0x4510, 0x4510, 0x4508, 0x7D0A, 0x454A, 0x186, 0x102}},
    {30467, 2, {0x0, 0x0, 0x79FC, 0x4800, 0x4800, 0x7800, 0x4BFE, 0x4820, 0x7820, 0x4840, 0x4840, 0x4888, 0x7904, 0x4BFE, 0x102, 0x0}},
    {30468, 2, {0x0, 0x3FE, 0x7C20, 0x4420, 0x4520, 0x7D3E, 0x4522, 0x4522, 0x7D22, 0x45FE, 0x4402, 0x4402, 0x7C02, 0x4402, 0x14, 0x8}},
    {30469, 2, {0x8, 0x1C, 0x79E0, 0x4900, 0x4900, 0x79FC, 0x4944, 0x4944, 0x7944, 0x4928, 0x4928, 0x4910, 0x7910, 0x4A28, 0x244, 0x482}},
    {30470, 2, {0x40, 0x20, 0x7C20, 0x45FE, 0x4440, 0x7C40, 0x4440, 0x447C, 0x7C44, 0x4444, 0x4444, 0x4444, 0x7C84, 0x4484, 0x128, 0x210}},
    {30471, 2, {0x40, 0x40, 0x7840, 0x4948, 0x4944, 0x7A42, 0x4A42, 0x4C48, 0x7848, 0x4848, 0x4810, 0x4810, 0x7820, 0x4840, 0x180, 0x600}},
    {30472, 2, {0x40, 0x40, 0x7840, 0x4BFC, 0x4A44, 0x7C48, 0x4840, 0x4840, 0x78A0, 0x48A0, 0x48A0, 0x4920, 0x7922, 0x4A22, 0x41E, 0x800}},
    {30473, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2FF8, 0x2808, 0x2808, 0x2FF8, 0x2808, 0x2FF8, 0x4808, 0x4808, 0x8FF8, 0x808}},
    {30474, 2, {0x8, 0x3C, 0x7DE0, 0x4420, 0x4420, 0x7C3C, 0x45E0, 0x4420, 0x7C20, 0x443E, 0x47E0, 0x4420, 0x7C22, 0x4422, 0x22, 0x1E}},
    {30475, 2, {0xF8, 0x7F00, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0xFF0, 0x810}},
    {30476, 2, {0x0, 0xFF0, 0x4810, 0x4FF0, 0x4810, 0x4FF0, 0x4810, 0x4FF0, 0x4000, 0x7FFC, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {30477, 2, {0x0, 0x1FE, 0x7D00, 0x4504, 0x4544, 0x7D28, 0x4528, 0x4510, 0x7D10, 0x4528, 0x4528, 0x4544, 0x7D84, 0x4500, 0x1FE, 0x0}},
    {30478, 2, {0x0, 0x1FC, 0x7800, 0x4800, 0x4800, 0x7BFE, 0x4820, 0x4820, 0x7928, 0x4924, 0x4A24, 0x4A22, 0x7C22, 0x4820, 0xA0, 0x40}},
    {30479, 2, {0x20, 0x20, 0x7C20, 0x45FC, 0x4524, 0x7D24, 0x4524, 0x4524, 0x7D24, 0x47FE, 0x4420, 0x4450, 0x7C50, 0x4488, 0x104, 0x202}},
    {30480, 2, {0x0, 0x0, 0x7BFE, 0x4820, 0x4820, 0x7820, 0x4920, 0x4920, 0x793C, 0x4920, 0x4920, 0x4920, 0x7920, 0x4920, 0x7FE, 0x0}},
    {30481, 2, {0x10, 0x90, 0x7890, 0x4910, 0x497E, 0x7A52, 0x4B92, 0x4892, 0x7912, 0x4912, 0x4A52, 0x4BD2, 0x7862, 0x4822, 0x4A, 0x84}},
    {30482, 2, {0x20, 0x20, 0x7C20, 0x45FC, 0x4524, 0x7D24, 0x4524, 0x45FC, 0x7D24, 0x4524, 0x4524, 0x45FC, 0x7D24, 0x4420, 0x20, 0x20}},
    {30483, 2, {0x28, 0x24, 0x7824, 0x4820, 0x4BFE, 0x7A20, 0x4A24, 0x4A24, 0x7A24, 0x4A28, 0x4A28, 0x4A90, 0x7B12, 0x4A2A, 0x46, 0x82}},
    {30484, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x920, 0x1110, 0x2108, 0x4104, 0x920, 0x1110, 0x2108, 0x4104, 0x100, 0x100}},
    {30485, 2, {0x20, 0x20, 0x7C50, 0x4488, 0x4504, 0x7E12, 0x4420, 0x4440, 0x7D88, 0x4410, 0x4420, 0x4444, 0x7D88, 0x4410, 0x60, 0x380}},
    {30486, 2, {0x0, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x4450, 0x7C50, 0x4450, 0x4450, 0x4492, 0x7C92, 0x4512, 0x20E, 0x400}},
    {30487, 2, {0x80, 0x80, 0x7D00, 0x45FC, 0x4604, 0x7C04, 0x45E4, 0x4524, 0x7D24, 0x4524, 0x45E4, 0x4524, 0x7C04, 0x4404, 0x28, 0x10}},
    {30488, 2, {0x100, 0x100, 0xFFFE, 0x100, 0x2288, 0x1450, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30489, 2, {0x20, 0x20, 0x7C20, 0x4440, 0x4448, 0x7C84, 0x45FE, 0x4482, 0x7C00, 0x44FC, 0x4484, 0x4484, 0x7C84, 0x4484, 0xFC, 0x84}},
    {30490, 2, {0x100, 0x1100, 0x1FF8, 0x2100, 0x5FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30491, 2, {0x20, 0x20, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x4420, 0x47FE, 0x7C70, 0x44A8, 0x44A8, 0x4524, 0x7D24, 0x4622, 0x20, 0x20}},
    {30492, 2, {0x20, 0x20, 0x7C20, 0x47FE, 0x4420, 0x7C20, 0x4420, 0x45FC, 0x7C70, 0x44A8, 0x44A8, 0x4524, 0x7D24, 0x4622, 0x20, 0x20}},
    {30493, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4A02, 0x7C04, 0x4800, 0x4800, 0x7BFE, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0xA0, 0x40}},
    {30494, 2, {0x2070, 0x3F84, 0x2004, 0x1FFC, 0x0, 0xFF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4808, 0x4FF8, 0x4000, 0x7FFE, 0x1010, 0x2008}},
    {30495, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {30496, 2, {0x0, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x7DFC, 0x4520, 0x4520, 0x7DFE, 0x4520, 0x4520, 0x4510, 0x7D12, 0x454A, 0x186, 0x102}},
    {30497, 2, {0x4, 0x1E, 0x7DF0, 0x4510, 0x4510, 0x7D10, 0x4510, 0x45FE, 0x7D10, 0x4510, 0x4510, 0x4508, 0x7D0A, 0x454A, 0x1A6, 0x112}},
    {30498, 2, {0x1000, 0x1EF8, 0x2288, 0x62A8, 0x9492, 0x882, 0x307E, 0xC000, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30499, 2, {0x20, 0x20, 0x7D20, 0x4520, 0x45FC, 0x7D20, 0x4620, 0x4420, 0x7DFE, 0x4420, 0x4450, 0x4450, 0x7C88, 0x4488, 0x104, 0x202}},
    {30500, 2, {0x0, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x7DFC, 0x4500, 0x4540, 0x7D44, 0x4548, 0x4570, 0x4540, 0x7D42, 0x4642, 0x23E, 0x400}},
    {30501, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30502, 2, {0x90, 0x90, 0x7890, 0x4A90, 0x4A92, 0x7AD4, 0x4A98, 0x4A90, 0x7A90, 0x4A90, 0x4A90, 0x4A92, 0x7AD2, 0x4F12, 0x20E, 0x0}},
    {30503, 2, {0x0, 0x1FC, 0x7C44, 0x4444, 0x4444, 0x7C44, 0x4494, 0x4488, 0x7D00, 0x44FC, 0x4484, 0x4484, 0x7C84, 0x4484, 0xFC, 0x84}},
    {30504, 2, {0x8, 0x3C, 0x7BC0, 0x4800, 0x4840, 0x7820, 0x4BFC, 0x4808, 0x7810, 0x4820, 0x4840, 0x4880, 0x7900, 0x4A80, 0x47E, 0x0}},
    {30505, 2, {0x40, 0x20, 0x7820, 0x4BFE, 0x4840, 0x7840, 0x4888, 0x4908, 0x7BF0, 0x4820, 0x4840, 0x4888, 0x7904, 0x4BFC, 0x104, 0x0}},
    {30506, 2, {0x0, 0x1FE, 0x7C20, 0x4420, 0x4420, 0x7DFE, 0x4522, 0x4522, 0x7D52, 0x454A, 0x458A, 0x4502, 0x7D02, 0x4502, 0x10A, 0x104}},
    {30507, 2, {0x20, 0x20, 0x7D24, 0x44A4, 0x44A8, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {30508, 2, {0x90, 0x88, 0x7888, 0x4880, 0x4BFE, 0x78A0, 0x48A0, 0x48A4, 0x78A4, 0x4928, 0x4928, 0x4932, 0x7A22, 0x4A62, 0x49E, 0x800}},
    {30509, 2, {0x20, 0x20, 0x7DFC, 0x4420, 0x4420, 0x7C20, 0x47FE, 0x4400, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {30510, 2, {0x0, 0x1FE, 0x7D02, 0x4502, 0x457A, 0x7D02, 0x4502, 0x457A, 0x7D4A, 0x454A, 0x454A, 0x457A, 0x7D02, 0x4502, 0x10A, 0x104}},
    {30511, 2, {0x20, 0x20, 0x7924, 0x48A4, 0x48A8, 0x7820, 0x4820, 0x4BFE, 0x7870, 0x48A8, 0x48A8, 0x4924, 0x7924, 0x4A22, 0x20, 0x20}},
    {30512, 2, {0x0, 0x1FE, 0x7C20, 0x4440, 0x4488, 0x7D04, 0x45FE, 0x4422, 0x7C20, 0x4420, 0x45FE, 0x4420, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {30513, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x7824, 0x49FC, 0x4920, 0x79FE, 0x4822, 0x482A, 0x4854, 0x7850, 0x4888, 0x104, 0x202}},
    {30514, 2, {0x0, 0x3FE, 0x7C88, 0x4488, 0x44F8, 0x7C88, 0x4488, 0x44F8, 0x7C88, 0x4488, 0x448E, 0x47F8, 0x7C08, 0x4408, 0x8, 0x8}},
    {30515, 2, {0x40, 0x40, 0x7CFC, 0x4504, 0x4688, 0x7C50, 0x4420, 0x4440, 0x7C80, 0x45FC, 0x4684, 0x4484, 0x7C84, 0x4484, 0xFC, 0x84}},
    {30516, 2, {0x80, 0x80, 0x79FC, 0x4904, 0x4A04, 0x79E4, 0x4924, 0x4924, 0x79E4, 0x4924, 0x4924, 0x49E4, 0x7904, 0x4804, 0x28, 0x10}},
    {30517, 2, {0x20, 0x20, 0x7C7C, 0x4484, 0x4548, 0x7C30, 0x4420, 0x4448, 0x7D90, 0x443E, 0x4442, 0x45A4, 0x7C18, 0x4410, 0x60, 0x180}},
    {30518, 2, {0x0, 0x1FE, 0x7D00, 0x4500, 0x45FE, 0x7D10, 0x4510, 0x457C, 0x7D10, 0x4510, 0x4510, 0x45FE, 0x7D00, 0x4500, 0x1FE, 0x0}},
    {30519, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x820, 0xFE0, 0x820, 0xFE0, 0x820}},
    {30520, 2, {0x40, 0x40, 0x7890, 0x4908, 0x4BFC, 0x7824, 0x4920, 0x4920, 0x79FC, 0x4A20, 0x4820, 0x4FFE, 0x7820, 0x4820, 0x20, 0x20}},
    {30521, 2, {0x88, 0x48, 0x7C50, 0x4400, 0x45FC, 0x7C20, 0x4420, 0x4420, 0x7FFE, 0x4420, 0x4450, 0x4450, 0x7C88, 0x4488, 0x104, 0x202}},
    {30522, 2, {0x90, 0x90, 0x7890, 0x4A92, 0x4994, 0x7898, 0x4890, 0x4998, 0x7A94, 0x4C92, 0x4890, 0x4890, 0x7912, 0x4912, 0x212, 0x40E}},
    {30523, 2, {0x104, 0x84, 0x7C88, 0x4400, 0x47FE, 0x7C20, 0x4420, 0x45FC, 0x7C20, 0x4420, 0x47FE, 0x4420, 0x7C20, 0x4420, 0x20, 0x20}},
    {30524, 2, {0x0, 0x3F8, 0x7A08, 0x4A08, 0x4BF8, 0x7A08, 0x4A08, 0x4BF8, 0x7A44, 0x4A48, 0x4A30, 0x4A20, 0x7A10, 0x4A88, 0x306, 0x200}},
    {30525, 2, {0x4, 0x1E, 0x7DE0, 0x4500, 0x4506, 0x7D78, 0x4550, 0x4550, 0x7D52, 0x4554, 0x4548, 0x4548, 0x7D44, 0x4654, 0x262, 0x440}},
    {30526, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0xFC, 0x7F00, 0x1110, 0x1110, 0x1110, 0x2928, 0x2528, 0x4544, 0x8182, 0x100}},
    {30527, 2, {0x40, 0x20, 0x7810, 0x49E0, 0x4822, 0x7834, 0x4BB8, 0x48B0, 0x78A8, 0x48A8, 0x4928, 0x4924, 0x7A24, 0x4C22, 0xA0, 0x40}},
    {30528, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x400, 0xFF0, 0x1810, 0x2FF0, 0xC810, 0xFF0, 0x810, 0xFF0, 0x810}},
    {30529, 2, {0x100, 0x100, 0x79F0, 0x4A10, 0x4C20, 0x7BF8, 0x4848, 0x4848, 0x7BFE, 0x4848, 0x4848, 0x4BF8, 0x7848, 0x4840, 0x140, 0x80}},
    {30530, 2, {0x208, 0x508, 0x888, 0x1248, 0x508, 0x888, 0x3FF8, 0x2000, 0x2FF8, 0x2808, 0x2FF8, 0x2808, 0x2FF8, 0x4808, 0x4FF8, 0x8808}},
    {30531, 2, {0x40, 0x40, 0x7888, 0x4904, 0x4BFE, 0x7802, 0x4888, 0x4944, 0x7A42, 0x48F8, 0x4988, 0x4A50, 0x7820, 0x4850, 0x188, 0x606}},
    {30532, 2, {0x20, 0x124, 0x7CA4, 0x44A8, 0x4420, 0x7DFC, 0x4504, 0x4504, 0x7DFC, 0x4504, 0x4504, 0x45FC, 0x7D04, 0x4504, 0x114, 0x108}},
    {30533, 2, {0x0, 0x1FC, 0x7D04, 0x45FC, 0x4504, 0x7DFC, 0x4400, 0x4400, 0x7DFC, 0x4420, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x20, 0x20}},
    {30534, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x79F8, 0x4800, 0x4800, 0x7BFE, 0x4890, 0x4890, 0x4890, 0x7912, 0x4912, 0x20E, 0x400}},
    {30535, 2, {0x108, 0x88, 0x7890, 0x4BFC, 0x4824, 0x7824, 0x4BFC, 0x4A20, 0x7A20, 0x4BFE, 0x4862, 0x48A2, 0x792A, 0x4A24, 0x420, 0x20}},
    {30536, 2, {0x0, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x7DFC, 0x4400, 0x4400, 0x7DFE, 0x4420, 0x4420, 0x45FC, 0x7C20, 0x4420, 0x3FE, 0x0}},
    {30537, 2, {0x20, 0x20, 0x7924, 0x4924, 0x4924, 0x7924, 0x4AAA, 0x4C72, 0x7820, 0x4820, 0x49FC, 0x4820, 0x7820, 0x4820, 0x3FE, 0x0}},
    {30538, 2, {0x0, 0xFC, 0x7C84, 0x4484, 0x44FC, 0x7C00, 0x45FE, 0x4502, 0x7D02, 0x45FE, 0x4502, 0x4502, 0x7DFE, 0x4502, 0x10A, 0x104}},
    {30539, 2, {0x10, 0xD8, 0x7B94, 0x4894, 0x4890, 0x7BFE, 0x4890, 0x4894, 0x7894, 0x48D8, 0x4B98, 0x4890, 0x78AA, 0x48CA, 0x286, 0x102}},
    {30540, 2, {0x80, 0x80, 0x79F8, 0x4A08, 0x4C10, 0x7BFC, 0x4A44, 0x4A44, 0x7A44, 0x4BFC, 0x48A0, 0x48A0, 0x7922, 0x4922, 0x21E, 0x400}},
    {30541, 2, {0x0, 0x3F8, 0x7A08, 0x4A08, 0x4BF8, 0x7A08, 0x4A08, 0x4BF8, 0x7A08, 0x4A08, 0x4BF8, 0x4920, 0x7920, 0x4A22, 0x422, 0x181E}},
    {30542, 2, {0x184, 0x68, 0x7830, 0x48C8, 0x4B04, 0x7840, 0x4BFE, 0x48A0, 0x7920, 0x4BFC, 0x4D24, 0x4924, 0x7934, 0x4928, 0x20, 0x20}},
    {30543, 2, {0x0, 0x3FE, 0x7A22, 0x4A22, 0x4A22, 0x7BFE, 0x4A22, 0x4A62, 0x7A72, 0x4AAA, 0x4B22, 0x4A22, 0x7A22, 0x4A02, 0x3FE, 0x202}},
    {30544, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x7924, 0x48A4, 0x48A8, 0x7BFE, 0x4870, 0x48A8, 0x48A8, 0x7924, 0x4A22, 0x420, 0x20}},
    {30545, 2, {0x40, 0x40, 0x78A0, 0x48A0, 0x4910, 0x7A08, 0x4DF6, 0x4800, 0x7888, 0x4848, 0x4A48, 0x4950, 0x7910, 0x4820, 0x7FE, 0x0}},
    {30546, 2, {0x20, 0x122, 0x7922, 0x4A24, 0x4850, 0x7888, 0x4B04, 0x4822, 0x7820, 0x4924, 0x4924, 0x4A28, 0x7850, 0x4888, 0x104, 0x602}},
    {30547, 2, {0x90, 0x90, 0x7890, 0x4BFC, 0x4A94, 0x7A94, 0x4A94, 0x4BFC, 0x7A94, 0x4A94, 0x4A94, 0x4FFE, 0x7800, 0x4890, 0x108, 0x204}},
    {30548, 2, {0x20, 0x20, 0x7850, 0x4888, 0x4904, 0x7AFA, 0x4800, 0x49FC, 0x7954, 0x4954, 0x49FC, 0x4954, 0x7954, 0x4954, 0x104, 0x10C}},
    {30549, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4D04, 0x7900, 0x49DE, 0x4A52, 0x7A52, 0x4B52, 0x4C9A, 0x4894, 0x7910, 0x4912, 0x212, 0x40E}},
    {30550, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x7BFE, 0x4888, 0x4944, 0x7A42, 0x48F8, 0x4988, 0x4A50, 0x7820, 0x4850, 0x188, 0x606}},
    {30551, 2, {0x0, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x49F8, 0x7880, 0x49FC, 0x4A54, 0x4C94, 0x7924, 0x4A44, 0x94, 0x108}},
    {30552, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x304, 0xC88, 0x1870, 0xEA18, 0xC06}},
    {30553, 2, {0x40, 0x20, 0x7BFC, 0x4A04, 0x4A04, 0x7BFC, 0x4A00, 0x4A28, 0x7A24, 0x4BFE, 0x4A20, 0x4A50, 0x7A50, 0x4C88, 0x504, 0xA02}},
    {30554, 2, {0x0, 0x3FE, 0x7A10, 0x4A10, 0x4AFE, 0x7A10, 0x4A10, 0x4BFE, 0x7A00, 0x4A10, 0x4A10, 0x4AFE, 0x7A10, 0x4C10, 0x5FE, 0x800}},
    {30555, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x114, 0x108}},
    {30556, 2, {0x3C, 0x7C0, 0x7A44, 0x4928, 0x4800, 0x7BF8, 0x4848, 0x4848, 0x7FFE, 0x4848, 0x4848, 0x4BF8, 0x7848, 0x4840, 0x140, 0x80}},
    {30557, 2, {0xC80, 0x70FC, 0x1124, 0xFE24, 0x3844, 0x54A8, 0x9110, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30558, 2, {0x20, 0x20, 0x7820, 0x4BFE, 0x4820, 0x7924, 0x4924, 0x4924, 0x7AAA, 0x4870, 0x48A8, 0x48A8, 0x7924, 0x4A22, 0x420, 0x20}},
    {30559, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x4888, 0x7888, 0x4954, 0x4A22, 0x7800, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20, 0x20}},
    {30560, 2, {0x20, 0x124, 0x78A8, 0x4820, 0x49FC, 0x7840, 0x4BFE, 0x4888, 0x7904, 0x4AFA, 0x4C88, 0x4888, 0x78A8, 0x4892, 0x82, 0x7E}},
    {30561, 2, {0x10, 0x78, 0xF3C0, 0x9040, 0x9040, 0xF7FC, 0x9248, 0x9248, 0xF7FE, 0x9248, 0x9248, 0x97FC, 0xF040, 0x9040, 0x7FC, 0x0}},
    {30562, 2, {0xA0, 0x90, 0x7880, 0x49FE, 0x4910, 0x7B10, 0x4DFC, 0x4910, 0x7910, 0x49FC, 0x4910, 0x4910, 0x7910, 0x49FE, 0x100, 0x100}},
    {30563, 2, {0x800, 0xEFC, 0x844, 0xFF44, 0x828, 0x2A10, 0x4928, 0x844, 0x1FF2, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30564, 2, {0x0, 0x1FC, 0x7924, 0x4924, 0x49FC, 0x7924, 0x4924, 0x49FC, 0x7800, 0x4800, 0x4BFE, 0x4888, 0x7888, 0x4888, 0x108, 0x208}},
    {30565, 2, {0x40, 0x80, 0x7BFC, 0x4A24, 0x4A24, 0x7BFC, 0x4A24, 0x4A44, 0x7BFC, 0x4890, 0x4910, 0x4FFE, 0x7810, 0x4810, 0x10, 0x10}},
    {30566, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x7BFE, 0x4888, 0x4904, 0x7A22, 0x4820, 0x49FC, 0x4820, 0x7820, 0x4820, 0x3FE, 0x0}},
    {30567, 2, {0x1C, 0x3E0, 0x7A20, 0x4BFE, 0x4A20, 0x7A92, 0x4B0A, 0x4A06, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30568, 2, {0x0, 0x40, 0x7B9C, 0x4A04, 0x4A04, 0x7B9C, 0x4A04, 0x4A04, 0x7BFC, 0x4890, 0x4890, 0x4890, 0x7890, 0x4912, 0x212, 0x40E}},
    {30569, 2, {0x0, 0x1F8, 0x7808, 0x4808, 0x49F8, 0x7808, 0x4808, 0x4BFE, 0x7820, 0x4A22, 0x4974, 0x48A8, 0x7924, 0x4A22, 0xA0, 0x40}},
    {30570, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x820, 0x440, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {30571, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x7824, 0x4BFE, 0x4824, 0x79FC, 0x4820, 0x4920, 0x493E, 0x7920, 0x4AA0, 0x27E, 0x400}},
    {30572, 2, {0x8, 0x3C, 0x7BC0, 0x4844, 0x4A24, 0x7928, 0x4900, 0x4820, 0x7BFE, 0x4870, 0x48A8, 0x48A8, 0x7924, 0x4A22, 0x420, 0x20}},
    {30573, 2, {0x0, 0x1FC, 0x7924, 0x4924, 0x4974, 0x7924, 0x49FC, 0x4904, 0x7974, 0x4954, 0x4954, 0x4974, 0x7904, 0x4A04, 0x214, 0x408}},
    {30574, 2, {0x40, 0xA0, 0x7910, 0x4A08, 0x4DF6, 0x7800, 0x4BC4, 0x4A54, 0x7A54, 0x4BD4, 0x4A54, 0x4A54, 0x7BD4, 0x4A44, 0x254, 0x2C8}},
    {30575, 2, {0x20, 0x7E20, 0x427E, 0x7EA4, 0x4824, 0x7F28, 0x4810, 0x6528, 0x4346, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30576, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x48A8, 0x7820, 0x4BFE, 0x4840, 0x7880, 0x48FC, 0x4984, 0x4A84, 0x7884, 0x4884, 0xFC, 0x84}},
    {30577, 2, {0x0, 0x3DE, 0x7A42, 0x4A42, 0x4A42, 0x7BDE, 0x4A00, 0x4A3E, 0x7BD2, 0x4A12, 0x4A14, 0x4BD4, 0x7A08, 0x4A14, 0x224, 0x242}},
    {30578, 2, {0x0, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x4920, 0x79FE, 0x4A20, 0x4820, 0x49FC, 0x7820, 0x4820, 0x3FE, 0x0}},
    {30579, 2, {0x20, 0x20, 0x7BFE, 0x4850, 0x4888, 0x7B26, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x4800, 0x4820, 0x79FC, 0x4820, 0x20, 0x3FE}},
    {30580, 2, {0x0, 0x3FE, 0x7A22, 0x4820, 0x4BFE, 0x7820, 0x49FC, 0x4924, 0x79FC, 0x4924, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {30581, 2, {0x128, 0x124, 0x7BA4, 0x4920, 0x4920, 0x7FFE, 0x4820, 0x4824, 0x7BA4, 0x4AA8, 0x4AA8, 0x4A90, 0x7B92, 0x482A, 0x46, 0x82}},
    {30582, 2, {0x40, 0x40, 0x7BFC, 0x4840, 0x49F8, 0x7880, 0x4BFC, 0x4910, 0x7A08, 0x4DF6, 0x4910, 0x4910, 0x79F0, 0x4910, 0x110, 0x1F0}},
    {30583, 2, {0x20, 0x20, 0xF0FC, 0x9E24, 0x93FE, 0xF224, 0x94FC, 0x9420, 0xFEFC, 0x9220, 0x9220, 0x9BFE, 0xF420, 0x9620, 0x9FE, 0x1000}},
    {30584, 2, {0x0, 0x3FC, 0x7A24, 0x4A24, 0x4BFC, 0x7A00, 0x4AFC, 0x4A84, 0x7A84, 0x4AFC, 0x4A84, 0x4AFC, 0x7A84, 0x4A84, 0x4FC, 0x884}},
    {30585, 2, {0x40, 0x44, 0x7BF4, 0x4848, 0x4850, 0x7FFE, 0x4840, 0x4880, 0x79FC, 0x4B04, 0x4D04, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30586, 2, {0x80, 0xBC, 0x7884, 0x4908, 0x49FE, 0x7B20, 0x4D20, 0x497C, 0x7990, 0x4910, 0x49FE, 0x4910, 0x7928, 0x4928, 0x144, 0x182}},
    {30587, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C04, 0x79FC, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4800, 0x3FE, 0x0}},
    {30588, 2, {0x0, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x7904, 0x4904, 0x49FC, 0x7800, 0x4BFE, 0x4820, 0x4920, 0x793C, 0x4920, 0x2A0, 0x47E}},
    {30589, 2, {0x10, 0x3D4, 0x7858, 0x4A52, 0x498C, 0x7888, 0x4904, 0x4AFA, 0x7820, 0x4820, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x202}},
    {30590, 2, {0x400, 0x800, 0x7FFC, 0x4444, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x820, 0x440, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {30591, 2, {0x100, 0x1F8, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x920, 0x3290, 0xC60, 0x3FF8, 0xD016, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30592, 2, {0x7E20, 0x243E, 0x1848, 0xFFA8, 0x2910, 0x4A28, 0x9846, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {30593, 2, {0x0, 0x3FC, 0x7A04, 0x4BFC, 0x4A04, 0x7BFC, 0x4A04, 0x4BFC, 0x7850, 0x4848, 0x4FFE, 0x4840, 0x78A0, 0x4910, 0x208, 0xC06}},
    {30594, 2, {0x228, 0x724, 0x3C24, 0x2420, 0x24FE, 0x3F20, 0x2420, 0x2F3C, 0x2934, 0x2F54, 0x2954, 0x2F54, 0x2988, 0x4988, 0x4F14, 0x8922}},
    {30595, 2, {0x8, 0x1C, 0x79F0, 0x4910, 0x4910, 0x79FE, 0x4910, 0x497C, 0x7944, 0x497C, 0x4944, 0x497C, 0x7944, 0x4A44, 0x27C, 0x444}},
    {30596, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4888, 0x7800, 0x49FC, 0x4924, 0x7924, 0x4924, 0x49FC, 0x4924, 0x7924, 0x4924, 0x1FC, 0x104}},
    {30597, 2, {0x88, 0x1C8, 0xF708, 0x9108, 0x912A, 0xF12A, 0x97AC, 0x9148, 0xF308, 0x9388, 0x9554, 0x9514, 0xF914, 0x9124, 0x124, 0x142}},
    {30598, 2, {0x20, 0x20, 0x79FC, 0x4924, 0x49FC, 0x7820, 0x4FFE, 0x4800, 0x79FC, 0x4904, 0x4924, 0x4924, 0x7924, 0x4850, 0x88, 0x304}},
    {30599, 2, {0x10, 0x92, 0x7A52, 0x4954, 0x4910, 0x787E, 0x4810, 0x4B38, 0x7954, 0x4992, 0x4910, 0x4910, 0x7910, 0x4A80, 0x47E, 0x0}},
    {30600, 2, {0x90, 0x88, 0x7904, 0x4A42, 0x4C88, 0x79FC, 0x4804, 0x4800, 0x7BDE, 0x4842, 0x4A52, 0x494A, 0x7A52, 0x4842, 0x14A, 0x84}},
    {30601, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0x7C44, 0x4428, 0x7C28, 0x4410, 0x7C28, 0x4444, 0x7D82}},
    {30602, 2, {0x20, 0x10, 0x7BFE, 0x4A20, 0x4AFC, 0x7A24, 0x4BFE, 0x4A24, 0x7AFC, 0x4A20, 0x4AFC, 0x4A84, 0x7A84, 0x4C84, 0x4FC, 0x884}},
    {30603, 2, {0x20, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x7BFE, 0x4888, 0x104, 0x202}},
    {30604, 2, {0x20, 0x20, 0x79FC, 0x4820, 0x4820, 0x7BFE, 0x4880, 0x4904, 0x7BFE, 0x4802, 0x4BFC, 0x4A94, 0x7A94, 0x4A94, 0x7FE, 0x0}},
    {30605, 2, {0x140, 0x65C, 0xF444, 0x9444, 0x975C, 0xF444, 0x9444, 0x97FC, 0xF040, 0x97FC, 0x9208, 0x9110, 0xF0A0, 0x9040, 0x1B0, 0xE0E}},
    {30606, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4C24, 0x79FC, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30607, 2, {0x7FFC, 0x4444, 0x7FFC, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {30608, 2, {0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x0, 0x7E7E, 0x4242, 0x7E7E, 0x4242, 0x7E7E, 0x4242, 0x7E7E, 0x4242}},
    {30609, 2, {0x0, 0x3FE, 0x7A02, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4840, 0x4820, 0x4BFE, 0x7800, 0x4888, 0x104, 0x202}},
    {30610, 2, {0x108, 0x108, 0xF7FE, 0x9108, 0x9000, 0xF7FE, 0x9090, 0x9090, 0xF7FE, 0x9492, 0x9492, 0x956A, 0xF646, 0x9402, 0x40A, 0x404}},
    {30611, 2, {0x202, 0x122, 0x782A, 0x4FEA, 0x482A, 0x7BAA, 0x482A, 0x4BAA, 0x782A, 0x4BAA, 0x4AAA, 0x4AAA, 0x7BAA, 0x4AAA, 0x42, 0x82}},
    {30612, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4820, 0x7BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x88, 0x104}},
    {30613, 2, {0x20, 0x1FC, 0x7888, 0x4850, 0x4BFE, 0x7800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {30614, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30615, 2, {0x40, 0x80, 0x79FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4900, 0x79FE, 0x4900, 0x49FE, 0x4802, 0x7AAA, 0x4AAA, 0x202, 0xC}},
    {30616, 2, {0x0, 0x3FE, 0x7A00, 0x4A7C, 0x4A44, 0x7A44, 0x4A7C, 0x4A00, 0x7AEE, 0x4AAA, 0x4AAA, 0x4AAA, 0x7AEE, 0x4A00, 0x3FE, 0x0}},
    {30617, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4820, 0x4BFE, 0x7850, 0x4888, 0x104, 0x202}},
    {30618, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x4800, 0x79FC, 0x4820, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x49FC, 0x7800, 0x4888, 0x104, 0x202}},
    {30619, 2, {0x124, 0x124, 0xF224, 0x9424, 0x9954, 0xF14A, 0x9292, 0x9610, 0xFA10, 0x9250, 0x925C, 0x9250, 0xF250, 0x92B0, 0x29E, 0x300}},
    {30620, 2, {0x20, 0x1FC, 0x7924, 0x4BFE, 0x4924, 0x79FC, 0x4820, 0x49FC, 0x7924, 0x49FC, 0x4840, 0x4BFE, 0x7888, 0x49D0, 0x70, 0x38C}},
    {30621, 2, {0x40, 0x20, 0x7BFE, 0x4800, 0x4954, 0x7924, 0x4954, 0x49FC, 0x7820, 0x4BFE, 0x4A42, 0x4A92, 0x7AFA, 0x4A0A, 0x202, 0x206}},
    {30622, 2, {0x88, 0x88, 0x7BFE, 0x4888, 0x4888, 0x78F8, 0x4820, 0x4BFE, 0x7A22, 0x4B32, 0x4AAA, 0x4B76, 0x7A22, 0x4A22, 0x22A, 0x224}},
    {30623, 2, {0x0, 0x3FE, 0x7850, 0x49FC, 0x4954, 0x7954, 0x49FC, 0x4800, 0x79FC, 0x4800, 0x4BFE, 0x4820, 0x78A8, 0x4924, 0x2A2, 0x40}},
    {30624, 2, {0x20, 0x124, 0x78A8, 0x4BFE, 0x4A02, 0x78F8, 0x4888, 0x4888, 0x78F8, 0x4820, 0x4820, 0x49FC, 0x7820, 0x4820, 0x3FE, 0x0}},
    {30625, 2, {0x200, 0x27C, 0xF244, 0x9744, 0x927C, 0xF244, 0x9244, 0x977C, 0xF244, 0x9244, 0x927C, 0x9228, 0xF528, 0x954A, 0x84A, 0x1086}},
    {30626, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30627, 2, {0x20, 0x20, 0x79FC, 0x4924, 0x49FC, 0x7820, 0x4BFE, 0x4A22, 0x7BFE, 0x4820, 0x4840, 0x4824, 0x7AA2, 0x4A8A, 0x478, 0x0}},
    {30628, 2, {0x0, 0x7BDE, 0x4A52, 0x4BDE, 0x4A52, 0x7BDE, 0x4A02, 0x4AFA, 0x7A22, 0x4A72, 0x4A22, 0x4A22, 0x7AFA, 0x4A02, 0x20A, 0x204}},
    {30629, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4346, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30630, 2, {0x20, 0x3FE, 0x7820, 0x49FC, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x7888, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30631, 2, {0xA0, 0x90, 0x79FE, 0x4910, 0x4B10, 0x7DFE, 0x4910, 0x4910, 0x79FE, 0x4910, 0x4910, 0x49FE, 0x7900, 0x4AA4, 0x252, 0x452}},
    {30632, 2, {0xA0, 0x4A4, 0xF2A8, 0x90A0, 0x97FC, 0xF110, 0x90A0, 0x97FC, 0xF040, 0x93F8, 0x9040, 0x97FC, 0xF0A0, 0x9110, 0x208, 0xC06}},
    {30633, 2, {0x0, 0x7FC, 0xF404, 0x97FC, 0x9400, 0xF5FC, 0x9420, 0x95FC, 0xF524, 0x95FC, 0x9420, 0x97FE, 0xF62A, 0x96FA, 0xA0A, 0x1206}},
    {30634, 2, {0x20, 0x7A4, 0xF0A8, 0x9292, 0x9114, 0xF208, 0x95F4, 0x9802, 0xF3F8, 0x9208, 0x9208, 0x93F8, 0xF208, 0x9110, 0xFFE, 0x0}},
    {30635, 2, {0x0, 0x3FE, 0x7850, 0x4BFE, 0x4A52, 0x7BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7820, 0x4BFE, 0x20, 0x20}},
    {30636, 2, {0x3C, 0x7C0, 0xF044, 0x9224, 0x9108, 0xF7FE, 0x9402, 0x9A08, 0xF208, 0x93BE, 0x9488, 0x9AA8, 0xF13E, 0x9208, 0x408, 0x808}},
    {30637, 2, {0x40, 0x40, 0xF7FC, 0x90A0, 0x9514, 0xF208, 0x97FC, 0x9A0A, 0xF3F8, 0x9208, 0x93F8, 0x9040, 0xF248, 0x9444, 0x944, 0x80}},
    {30638, 2, {0x208, 0x108, 0xF7C8, 0x9210, 0x949E, 0xFFD4, 0x9064, 0x9794, 0xF494, 0x9794, 0x9494, 0x9788, 0xF488, 0x9494, 0x4A4, 0x5C2}},
    {30639, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A02, 0x4AF2, 0x4A92, 0x7AF2, 0x4A92, 0x4AF2, 0x4A92, 0x7A92, 0x4B32, 0x202, 0x206}},
    {30640, 2, {0x8, 0x788, 0xF088, 0x9110, 0x9FDE, 0xF494, 0x94A4, 0x9794, 0xF494, 0x9794, 0x9494, 0x94C8, 0xF788, 0x9C94, 0xA4, 0xC2}},
    {30641, 2, {0x88, 0x3FE, 0x7888, 0x4800, 0x49FC, 0x78A8, 0x48A8, 0x4BFE, 0x78A8, 0x48A8, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {30642, 2, {0x0, 0x3FC, 0x7848, 0x4830, 0x4BFE, 0x7852, 0x4894, 0x4B50, 0x7820, 0x4BFE, 0x4A52, 0x4A8A, 0x7B76, 0x4A52, 0x272, 0x206}},
    {30643, 2, {0x20, 0x1FC, 0x7888, 0x4850, 0x4BFE, 0x7800, 0x49FC, 0x4924, 0x79FC, 0x4924, 0x49FC, 0x4820, 0x79FC, 0x4820, 0x3FE, 0x0}},
    {30644, 2, {0x80, 0x80, 0x79FE, 0x4B54, 0x4D54, 0x7954, 0x4BFE, 0x4954, 0x7954, 0x4954, 0x4FFE, 0x4800, 0x7954, 0x492A, 0x22A, 0x0}},
    {30645, 2, {0x40, 0x248, 0xF150, 0x97FC, 0x9150, 0xF248, 0x9444, 0x9208, 0xF208, 0x93BE, 0x9488, 0x9AA8, 0xF13E, 0x9208, 0x408, 0x808}},
    {30646, 2, {0x20, 0x1FC, 0x7924, 0x49FC, 0x4820, 0x7BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x88, 0x104}},
    {30647, 2, {0x0, 0x7BDE, 0x4A52, 0x4BDE, 0x4A52, 0x7BDE, 0x4A02, 0x4AFA, 0x7A8A, 0x4AFA, 0x4A8A, 0x4AFA, 0x7A8A, 0x4A02, 0x20A, 0x204}},
    {30648, 2, {0x88, 0x3FE, 0x7888, 0x4800, 0x4948, 0x7BFE, 0x4948, 0x4978, 0x7900, 0x49FC, 0x4820, 0x4BFE, 0x7870, 0x48A8, 0x326, 0x20}},
    {30649, 2, {0x3C, 0x3C0, 0xF044, 0x9224, 0x9108, 0xF7FE, 0x9442, 0x92A4, 0xF28A, 0x947A, 0x9080, 0x90F8, 0xF108, 0x9290, 0x60, 0x79E}},
    {30650, 2, {0x40, 0xA0, 0xF318, 0x9DF6, 0x9000, 0xF7FC, 0x9554, 0x94E4, 0xF7FC, 0x9000, 0x93F8, 0x9208, 0xF3F8, 0x9208, 0x3F8, 0x208}},
    {30651, 2, {0x100, 0x1F8, 0x7A08, 0x4FFE, 0x4A88, 0x7B24, 0x4BFE, 0x4A00, 0x7AFC, 0x4A00, 0x4AFC, 0x4A00, 0x7AFC, 0x4C84, 0x4FC, 0x884}},
    {30652, 2, {0x40, 0x40, 0x78A0, 0x4910, 0x4A08, 0x7DF6, 0x4800, 0x4800, 0x7BB8, 0x4AA8, 0x4AA8, 0x4BB8, 0x7910, 0x4910, 0x2A8, 0x444}},
    {30653, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E28, 0x2418, 0xFEE6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30654, 2, {0x0, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x100, 0x7FFC, 0x4824, 0x1010, 0x2FE8, 0x100, 0x100, 0x7FFC}},
    {30655, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {30656, 2, {0x288, 0x2A8, 0xF6A8, 0x9BF0, 0x901E, 0xF224, 0x95D4, 0x9C14, 0xF414, 0x95D4, 0x9554, 0x9548, 0xF568, 0x9554, 0x614, 0x422}},
    {30657, 2, {0xC8, 0x708, 0xF12A, 0x912C, 0x9FC8, 0xF108, 0x9388, 0x9554, 0xF914, 0x9122, 0x9040, 0x9024, 0xF522, 0x950A, 0x908, 0xF8}},
    {30658, 2, {0x1FC, 0x104, 0x7904, 0x49FC, 0x4800, 0x7BDE, 0x4A52, 0x4A52, 0x7BDE, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x4924, 0x622, 0x20}},
    {30659, 2, {0x40, 0x20, 0x7BFE, 0x4A22, 0x4954, 0x794A, 0x4A3A, 0x4800, 0x79FC, 0x4954, 0x4BFE, 0x4800, 0x79FC, 0x4820, 0xA0, 0x40}},
    {30660, 2, {0x1C, 0x1E0, 0x7820, 0x4BFE, 0x4820, 0x79FC, 0x49AC, 0x4974, 0x79FC, 0x4820, 0x49FC, 0x4820, 0x7BFE, 0x4800, 0x154, 0x22A}},
    {30661, 2, {0x0, 0x3DE, 0x7842, 0x4A52, 0x494A, 0x7A52, 0x4890, 0x4908, 0x7BFE, 0x4D10, 0x49FE, 0x4910, 0x79FE, 0x4910, 0x1FE, 0x100}},
    {30662, 2, {0x88, 0x3FE, 0x78A8, 0x4890, 0x49FE, 0x7B20, 0x49FC, 0x4920, 0x79FC, 0x4920, 0x49FE, 0x4900, 0x7BFC, 0x4888, 0x70, 0x38E}},
    {30663, 2, {0x88, 0x3FE, 0x7888, 0x4BFE, 0x4A02, 0x79FC, 0x4800, 0x4BFE, 0x7840, 0x48A2, 0x4B54, 0x48B8, 0x7B54, 0x4892, 0x350, 0x20}},
    {30664, 2, {0x40, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x4BFE, 0x7A8A, 0x4924, 0x4BFE, 0x4880, 0x78FC, 0x4884, 0x114, 0x208}},
    {30665, 2, {0x40, 0x20, 0x7BFE, 0x4A02, 0x49FC, 0x7948, 0x4A50, 0x49FC, 0x7B04, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x88, 0x104}},
    {30666, 2, {0x108, 0x110, 0xF23E, 0x9222, 0x94BE, 0xF722, 0x913E, 0x9208, 0xF4BE, 0x97AA, 0x902A, 0x90AA, 0xF56A, 0x952E, 0x408, 0x8}},
    {30667, 2, {0x0, 0x3FE, 0x7A48, 0x4BFE, 0x4A48, 0x7AFC, 0x4AA4, 0x4AFC, 0x7AA4, 0x4AFC, 0x4A20, 0x4BFE, 0x7B2A, 0x4D7A, 0x50A, 0x906}},
    {30668, 2, {0x20, 0x10, 0x7BFE, 0x4A44, 0x4AFE, 0x7A44, 0x4BFE, 0x4A10, 0x7AFE, 0x4A92, 0x4AFE, 0x4A92, 0x7AFE, 0x4A00, 0x244, 0x482}},
    {30669, 2, {0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x900, 0x1FF8, 0x3100, 0x5FF0, 0x9100, 0x1FFC, 0x1000, 0x3FF0, 0x820, 0x7C0, 0xF83E}},
    {30670, 2, {0x100, 0x1F8, 0x7A08, 0x4FFE, 0x4A62, 0x7A92, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4880, 0x79F8, 0x4A88, 0x70, 0x38E}},
    {30671, 2, {0x40, 0x1FC, 0x7924, 0x4994, 0x494C, 0x7924, 0x49FC, 0x4820, 0x7BFE, 0x4A8A, 0x4924, 0x4BFE, 0x7880, 0x48FC, 0x104, 0x20C}},
    {30672, 2, {0x1FC, 0x7820, 0x4BFE, 0x4A22, 0x4DAC, 0x7820, 0x49AC, 0x4890, 0x79FE, 0x4B20, 0x4DFC, 0x4920, 0x79FC, 0x4920, 0x1FE, 0x100}},
    {30673, 2, {0x40, 0x7C, 0x7840, 0x4BFC, 0x4A44, 0x7BF0, 0x4A44, 0x4AFC, 0x7AA8, 0x4AF8, 0x4AA8, 0x4AF8, 0x7A00, 0x4DFC, 0x554, 0xBFE}},
    {30674, 2, {0x108, 0x7FE, 0x7908, 0x4BFC, 0x4A94, 0x7BFC, 0x4800, 0x4FFE, 0x7C02, 0x49F8, 0x4908, 0x49F8, 0x7908, 0x49F8, 0x108, 0x1F8}},
    {30675, 2, {0x110, 0x7890, 0x4BDE, 0x4810, 0x4A5E, 0x7982, 0x4BDE, 0x4810, 0x7BDE, 0x4A50, 0x4BDE, 0x4A50, 0x7BDE, 0x4A50, 0x252, 0x2CE}},
    {30676, 2, {0x88, 0x7BFE, 0x4888, 0x4BDE, 0x4A52, 0x7BDE, 0x48A0, 0x4890, 0x79FE, 0x4920, 0x4BFC, 0x4D20, 0x79FC, 0x4920, 0x1FE, 0x100}},
    {30677, 2, {0x2108, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {30678, 2, {0x3DE, 0x7800, 0x4BDE, 0x4A52, 0x4B5A, 0x7A52, 0x4820, 0x4BFE, 0x7A50, 0x4BFE, 0x4A52, 0x4BFE, 0x7A92, 0x4ADC, 0x292, 0x4CE}},
    {30679, 2, {0x100, 0x3FF8, 0x100, 0xFE0, 0x820, 0x920, 0x920, 0x7FFC, 0x820, 0x7EFC, 0x820, 0x3EF8, 0x2288, 0x2AA8, 0x2AA8, 0xFFFE}},
    {30680, 2, {0x248, 0x150, 0xF7FC, 0x9404, 0x91F0, 0xF110, 0x97FC, 0x9554, 0xF4E4, 0x97FC, 0x9040, 0x97FC, 0xF040, 0x9FFE, 0x2A4, 0x452}},
    {30681, 2, {0xFBE, 0x8A2, 0xEFBE, 0xA8A2, 0xAFBE, 0xE802, 0xAB92, 0xA892, 0xEFDE, 0xAAAA, 0xABCA, 0xAAAA, 0xEB92, 0xAAD2, 0xFAA, 0x8C4}},
    {30682, 2, {0x3FE, 0x202, 0x7BFE, 0x4A92, 0x4A54, 0x7A92, 0x4AFE, 0x4AAA, 0x7AFE, 0x4B22, 0x4AFA, 0x4AAA, 0x7AFA, 0x4A22, 0x5FA, 0x4}},
    {30683, 2, {0x0, 0x1FF0, 0x10, 0x220, 0x140, 0x80, 0x7FFE, 0x182, 0x284, 0x480, 0x880, 0x1080, 0x2080, 0xC080, 0x280, 0x100}},
    {30684, 2, {0x20, 0x7E20, 0x250, 0x2450, 0x1888, 0x944, 0xFE22, 0xA20, 0x1800, 0x19FC, 0x2804, 0x4808, 0x8808, 0x810, 0x2810, 0x1020}},
    {30685, 2, {0x20, 0x7C20, 0x450, 0x2850, 0x1088, 0x924, 0xFE12, 0x1A10, 0x29FC, 0x2804, 0x4808, 0x4888, 0x8850, 0x820, 0x2810, 0x1010}},
    {30686, 2, {0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x100, 0x3FF8, 0x2448, 0x2828, 0x37D8, 0x2448, 0x2448, 0x27C8, 0x2018}},
    {30687, 2, {0x10, 0x7C92, 0x452, 0x2854, 0x1010, 0x8FE, 0xFE82, 0x1A82, 0x28FE, 0x2882, 0x4882, 0x48FE, 0x8882, 0x882, 0x288A, 0x1084}},
    {30688, 2, {0x48, 0x7C48, 0x448, 0x29FE, 0x1048, 0x848, 0xFFFE, 0x1A00, 0x28FC, 0x2884, 0x4884, 0x48FC, 0x8884, 0x884, 0x28FC, 0x1084}},
    {30689, 2, {0x3DE, 0xFA52, 0xBDE, 0x5252, 0x23DE, 0x10A0, 0xFDFE, 0x3720, 0x35FC, 0x5120, 0x51FE, 0x5000, 0x91FC, 0x1088, 0x5070, 0x278E}},
    {30690, 2, {0x800, 0x800, 0x800, 0x1FF8, 0x1100, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {30691, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x808, 0x800, 0x1FF0, 0x2100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {30692, 2, {0x40, 0x7C40, 0x480, 0x4FC, 0x520, 0x7E20, 0x4020, 0x4020, 0x43FE, 0x7C20, 0x450, 0x450, 0x488, 0x488, 0x2904, 0x1202}},
    {30693, 2, {0x2000, 0x2000, 0x207C, 0x7E44, 0x4844, 0x8844, 0x844, 0x844, 0xFF44, 0x844, 0x1444, 0x1444, 0x227C, 0x2244, 0x4200, 0x8000}},
    {30694, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFFC, 0x2400, 0x2400, 0x2FF8, 0x3080, 0x2080, 0x3FFC, 0x4140, 0x4220, 0x8410, 0x180C}},
    {30695, 2, {0x2004, 0x21E4, 0x3C24, 0x5024, 0x9024, 0x11E4, 0x1104, 0xFF04, 0x1104, 0x11E4, 0x1024, 0x2824, 0x2424, 0x4424, 0x4144, 0x8084}},
    {30696, 2, {0x2008, 0x203C, 0x21E0, 0x7C20, 0x5020, 0x9020, 0x13FE, 0xFC20, 0x1020, 0x1050, 0x1050, 0x2850, 0x2488, 0x4488, 0x4104, 0x8202}},
    {30697, 2, {0x2000, 0x21FE, 0x3D00, 0x5100, 0x9100, 0x11FC, 0x1104, 0xFF04, 0x1104, 0x1104, 0x11FC, 0x2900, 0x2500, 0x4500, 0x41FE, 0x8000}},
    {30698, 2, {0x2010, 0x2020, 0x20FC, 0x7C84, 0x50A4, 0x9094, 0x1094, 0xFDFE, 0x1084, 0x10A4, 0x1094, 0x2894, 0x2484, 0x4504, 0x4114, 0x8208}},
    {30699, 2, {0x2008, 0x203C, 0x21E0, 0x7C20, 0x5020, 0x93FE, 0x1050, 0xFC88, 0x1104, 0x128A, 0x1088, 0x2888, 0x2488, 0x4508, 0x4108, 0x8208}},
    {30700, 2, {0x2020, 0x2020, 0x2124, 0x7D24, 0x5124, 0x9124, 0x12AA, 0xFC72, 0x1020, 0x1020, 0x11FC, 0x2820, 0x2420, 0x4420, 0x43FE, 0x8000}},
    {30701, 2, {0x2000, 0x21FE, 0x3C00, 0x5000, 0x90FC, 0x1084, 0x1084, 0xFE84, 0x10FC, 0x1000, 0x1084, 0x2844, 0x2448, 0x4400, 0x41FE, 0x8000}},
    {30702, 2, {0x201C, 0x21E0, 0x3C20, 0x5020, 0x93FE, 0x10A8, 0x1124, 0xFE42, 0x1040, 0x13FE, 0x1088, 0x2908, 0x24D0, 0x4430, 0x4048, 0x8184}},
    {30703, 2, {0x2018, 0x21E0, 0x3C40, 0x53FE, 0x9088, 0x1174, 0x1252, 0xFE70, 0x1000, 0x11FC, 0x1104, 0x2974, 0x2554, 0x4574, 0x4104, 0x810C}},
    {30704, 2, {0x2104, 0x2088, 0x3C00, 0x53FE, 0x9222, 0x12AA, 0x1272, 0xFE22, 0x13FE, 0x1000, 0x11FC, 0x2904, 0x25FC, 0x4504, 0x41FC, 0x8104}},
    {30705, 2, {0x2088, 0x23FE, 0x20A8, 0x7C90, 0x51FE, 0x9320, 0x11FC, 0xFD20, 0x11FC, 0x1120, 0x11FE, 0x2900, 0x27FC, 0x4488, 0x4070, 0x838E}},
    {30706, 2, {0x2000, 0x23FE, 0x4252, 0x7A52, 0xA3FE, 0x2110, 0x2252, 0x23DC, 0xF812, 0x23CE, 0x2240, 0x23D2, 0x525C, 0x4BD0, 0x4A52, 0x82CE}},
    {30707, 2, {0x0, 0xFFFE, 0x200, 0x200, 0x400, 0x400, 0x800, 0xFF8, 0x1808, 0x2808, 0x4808, 0x8808, 0x808, 0x808, 0xFF8, 0x808}},
    {30708, 2, {0x0, 0x0, 0xFEFE, 0x1010, 0x1010, 0x2010, 0x3E10, 0x6210, 0x6210, 0xA210, 0x2210, 0x2210, 0x3E10, 0x2210, 0x50, 0x20}},
    {30709, 2, {0x4, 0x4, 0xFE04, 0x1044, 0x1044, 0x2044, 0x3E44, 0x6244, 0x6244, 0xA244, 0x2244, 0x2204, 0x3E04, 0x2204, 0x14, 0x8}},
    {30710, 2, {0x0, 0x1F0, 0xF910, 0x2110, 0x2110, 0x4110, 0x7910, 0xC910, 0x4910, 0x4910, 0x4910, 0x4912, 0x7A12, 0x4212, 0x40E, 0x800}},
    {30711, 2, {0x0, 0x1FC, 0xFC04, 0x1008, 0x1010, 0x2020, 0x3C20, 0x67FE, 0x6420, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x20A0, 0x40}},
    {30712, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x1020, 0x2020, 0x3C20, 0x67FE, 0x6420, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x2020, 0x20}},
    {30713, 2, {0x0, 0x0, 0xFDFC, 0x1090, 0x1090, 0x2090, 0x3C90, 0x6490, 0x6490, 0xA490, 0x2490, 0x2492, 0x3C92, 0x2492, 0x210E, 0x200}},
    {30714, 2, {0x8, 0x3C, 0xFDE0, 0x1020, 0x1020, 0x2020, 0x3C20, 0x643E, 0x65E0, 0xA420, 0x2420, 0x2422, 0x3C22, 0x2422, 0x201E, 0x0}},
    {30715, 2, {0x40, 0x40, 0xFE80, 0x10FE, 0x1100, 0x2200, 0x3CF8, 0x6410, 0xA420, 0x2440, 0x2440, 0x2482, 0x3D02, 0x2502, 0xFE, 0x0}},
    {30716, 2, {0x0, 0x0, 0xFDFC, 0x1020, 0x1020, 0x2020, 0x3C20, 0x6420, 0x6420, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {30717, 2, {0x40, 0x40, 0xFC40, 0x107E, 0x1082, 0x2082, 0x3D42, 0x6624, 0x6414, 0xA408, 0x2408, 0x2410, 0x3C20, 0x2440, 0x2080, 0x300}},
    {30718, 2, {0x0, 0x1F0, 0xF910, 0x2110, 0x2110, 0x4110, 0x7990, 0xC950, 0x4950, 0x4910, 0x4910, 0x4912, 0x7A12, 0x4212, 0x40E, 0x800}},
    {30719, 2, {0x20, 0x10, 0xFC10, 0x10FE, 0x1080, 0x2080, 0x3C80, 0x6480, 0x6480, 0xA480, 0x2480, 0x2480, 0x3C80, 0x2500, 0x2100, 0x200}},
    {30720, 2, {0x0, 0x1F8, 0xFC10, 0x1020, 0x1040, 0x2080, 0x3DFE, 0x6492, 0x6492, 0xA492, 0x2512, 0x2522, 0x3E22, 0x2442, 0x2094, 0x108}},
    {30721, 2, {0x0, 0x1F8, 0xFC08, 0x1008, 0x1088, 0x2088, 0x3C88, 0x64FE, 0x6402, 0xA402, 0x2402, 0x25FA, 0x3C02, 0x2402, 0x2014, 0x8}},
    {30722, 2, {0x20, 0x20, 0xFC20, 0x10A8, 0x10A4, 0x20A2, 0x3D22, 0x6520, 0x6624, 0xA424, 0x2428, 0x2408, 0x3C10, 0x2420, 0x20C0, 0x300}},
    {30723, 2, {0x0, 0xFC, 0xFC84, 0x1084, 0x10A4, 0x2094, 0x3C94, 0x6484, 0x65FE, 0xA484, 0x2484, 0x2484, 0x3C84, 0x2504, 0x2114, 0x208}},
    {30724, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1024, 0x2024, 0x3C24, 0x6424, 0x65FE, 0xA420, 0x2450, 0x2450, 0x3C88, 0x2488, 0x2104, 0x202}},
    {30725, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x1032, 0x21B2, 0x3CB4, 0x64A8, 0x64A8, 0xA4A8, 0x2524, 0x2524, 0x3E22, 0x2420, 0x20A0, 0x40}},
    {30726, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x11FC, 0x2020, 0x3C20, 0x6420, 0x67FE, 0xA420, 0x2450, 0x2450, 0x3C88, 0x2488, 0x2104, 0x202}},
    {30727, 2, {0x40, 0x20, 0xFC20, 0x13FE, 0x1088, 0x2088, 0x3C88, 0x6488, 0x6488, 0xA450, 0x2450, 0x2420, 0x3C50, 0x2488, 0x2104, 0x602}},
    {30728, 2, {0x0, 0x1FE, 0xFD00, 0x1100, 0x1178, 0x2148, 0x3D48, 0x6548, 0x6548, 0xA568, 0x2550, 0x2542, 0x3D42, 0x2542, 0x223E, 0x400}},
    {30729, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x7FFC, 0x400, 0x800, 0x3FF0, 0xC810, 0x810, 0xFF0, 0x810}},
    {30730, 2, {0x40, 0x20, 0xFC00, 0x11FC, 0x1000, 0x2000, 0x3CF0, 0x6490, 0x6490, 0xA490, 0x2490, 0x2492, 0x3C92, 0x2512, 0x210E, 0x200}},
    {30731, 2, {0x10, 0x10, 0xFC10, 0x1090, 0x1090, 0x2090, 0x3C9E, 0x6490, 0x6490, 0xA490, 0x2490, 0x2490, 0x3C90, 0x2490, 0x23FE, 0x0}},
    {30732, 2, {0x0, 0x100, 0xF97E, 0x2112, 0x2112, 0x4112, 0x79D2, 0x4F12, 0xC912, 0x4912, 0x4952, 0x4992, 0x7922, 0x4822, 0x4A, 0x84}},
    {30733, 2, {0x40, 0x40, 0xFC40, 0x107C, 0x1084, 0x2088, 0x3D20, 0x6420, 0x6420, 0xA450, 0x2450, 0x2450, 0x3C88, 0x2488, 0x2104, 0x202}},
    {30734, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x2104, 0x3E02, 0x6488, 0x6488, 0xA488, 0x2488, 0x2488, 0x3C88, 0x2508, 0x2108, 0x208}},
    {30735, 2, {0x10, 0x50, 0xFC50, 0x1050, 0x1088, 0x2088, 0x3D04, 0x66FA, 0x6448, 0xA448, 0x2448, 0x2448, 0x3C88, 0x2488, 0x2128, 0x210}},
    {30736, 2, {0x0, 0x3FC, 0xF884, 0x2088, 0x2088, 0x4090, 0x789C, 0xC884, 0x4944, 0x4944, 0x4928, 0x4928, 0x7A10, 0x4228, 0x444, 0x182}},
    {30737, 2, {0x0, 0x3FC, 0xF810, 0x2110, 0x2110, 0x4110, 0x7A10, 0xCBFE, 0x4830, 0x4850, 0x4890, 0x4910, 0x7A10, 0x4410, 0x50, 0x20}},
    {30738, 2, {0x10, 0x110, 0xFD10, 0x1112, 0x1112, 0x2114, 0x3DD8, 0x6510, 0x6510, 0xA510, 0x2510, 0x2512, 0x3D52, 0x2592, 0x210E, 0x0}},
    {30739, 2, {0x0, 0xF8, 0xFC88, 0x1088, 0x1088, 0x2106, 0x3E00, 0x65FC, 0x6484, 0xA484, 0x2448, 0x2450, 0x3C20, 0x2450, 0x2088, 0x306}},
    {30740, 2, {0x0, 0x1FC, 0xFC88, 0x1088, 0x1088, 0x2088, 0x3C88, 0x67FE, 0x6488, 0xA488, 0x2488, 0x2488, 0x3C88, 0x2508, 0x2108, 0x208}},
    {30741, 2, {0x40, 0x40, 0xFDF8, 0x1048, 0x1088, 0x208A, 0x3D0A, 0x6626, 0x6420, 0xA420, 0x27FE, 0x2420, 0x3C20, 0x2420, 0x2020, 0x20}},
    {30742, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1020, 0x2040, 0x3DFE, 0x6440, 0x6480, 0xA5FC, 0x2404, 0x2488, 0x3C50, 0x2420, 0x2010, 0x10}},
    {30743, 2, {0x40, 0x40, 0xFC40, 0x13FE, 0x1080, 0x20A0, 0x3D20, 0x65FC, 0x6420, 0xA420, 0x2420, 0x27FE, 0x3C20, 0x2420, 0x2020, 0x20}},
    {30744, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x2124, 0x3D24, 0x6524, 0x6524, 0xA5FC, 0x2424, 0x2420, 0x3C22, 0x2422, 0x201E, 0x0}},
    {30745, 2, {0x0, 0x0, 0xFBFC, 0x2100, 0x2100, 0x41F0, 0x7910, 0x4910, 0xC990, 0x4A50, 0x4A50, 0x4A10, 0x7A12, 0x4A92, 0x312, 0x20E}},
    {30746, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x1124, 0x2124, 0x3D24, 0x6524, 0x6524, 0xA554, 0x2450, 0x2490, 0x3C90, 0x2512, 0x2212, 0x40E}},
    {30747, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x2144, 0x3E22, 0x6420, 0x6400, 0xA5FC, 0x2404, 0x2408, 0x3C08, 0x2410, 0x2010, 0x20}},
    {30748, 2, {0x0, 0x3F8, 0xFA08, 0x2208, 0x2328, 0x42A8, 0x7AA8, 0xCA48, 0x4A48, 0x4AA8, 0x4AA8, 0x4B2A, 0x7C0A, 0x440A, 0x806, 0x1002}},
    {30749, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1020, 0x2020, 0x3C20, 0x65FE, 0x6420, 0xA420, 0x2440, 0x2448, 0x3C84, 0x25FE, 0x2082, 0x0}},
    {30750, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x2020, 0x3C20, 0x65FC, 0x6470, 0xA4A8, 0x24A8, 0x2524, 0x3D24, 0x2622, 0x2020, 0x20}},
    {30751, 2, {0x80, 0x80, 0xFC80, 0x10FE, 0x1140, 0x2140, 0x3E40, 0x647C, 0x6440, 0xA440, 0x2440, 0x247E, 0x3C40, 0x2440, 0x2040, 0x40}},
    {30752, 2, {0x0, 0xF8, 0xFC88, 0x1088, 0x1088, 0x20F8, 0x3C88, 0x6488, 0x6488, 0xA4F8, 0x2488, 0x2488, 0x3C88, 0x2488, 0x23FE, 0x0}},
    {30753, 2, {0x0, 0x0, 0xFDFC, 0x1020, 0x1020, 0x2020, 0x3C20, 0x6420, 0x65FC, 0xA420, 0x2428, 0x2424, 0x3C24, 0x2420, 0x23FE, 0x0}},
    {30754, 2, {0x0, 0x0, 0xFDFE, 0x1008, 0x1008, 0x21E8, 0x3D28, 0x6528, 0x6528, 0xA528, 0x25E8, 0x2528, 0x3C08, 0x2408, 0x2028, 0x10}},
    {30755, 2, {0x20, 0x10, 0xFC10, 0x11FE, 0x1102, 0x2204, 0x3C80, 0x6488, 0x6490, 0xA4A0, 0x24C0, 0x2482, 0x3C82, 0x2482, 0x207E, 0x0}},
    {30756, 2, {0x80, 0x80, 0xFCFE, 0x1100, 0x1220, 0x2120, 0x3D2C, 0x6574, 0x67A4, 0xA524, 0x2534, 0x2528, 0x3D22, 0x2502, 0x20FE, 0x0}},
    {30757, 2, {0x4, 0x1E, 0xFDF0, 0x1110, 0x1110, 0x2110, 0x3D10, 0x65FE, 0x6510, 0xA510, 0x2510, 0x2508, 0x3D0A, 0x254A, 0x21A6, 0x112}},
    {30758, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C, 0x0, 0x7FFC, 0x400, 0x800, 0x3FF0, 0xC810, 0x810, 0xFF0, 0x810}},
    {30759, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x103E, 0x2020, 0x3C20, 0x6420, 0x65FC, 0xA504, 0x2504, 0x2504, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30760, 2, {0x40, 0x20, 0xFDFC, 0x1104, 0x1104, 0x21FC, 0x3D00, 0x6500, 0x657C, 0xA508, 0x2510, 0x2520, 0x3D42, 0x2682, 0x227E, 0x400}},
    {30761, 2, {0x90, 0x90, 0xF890, 0x23FC, 0x2094, 0x4094, 0x7BFC, 0x4A90, 0xCA90, 0x4BFE, 0x4892, 0x4892, 0x791A, 0x4914, 0x210, 0x410}},
    {30762, 2, {0x0, 0xFC, 0xFC84, 0x10A4, 0x1094, 0x2084, 0x3C84, 0x67FE, 0x6504, 0xA544, 0x2524, 0x2504, 0x3DFE, 0x2404, 0x2028, 0x10}},
    {30763, 2, {0x40, 0x20, 0xFC00, 0x13FE, 0x1020, 0x2020, 0x3C20, 0x6420, 0x65FC, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {30764, 2, {0x40, 0x20, 0xFC20, 0x1000, 0x11FE, 0x2000, 0x3C04, 0x6484, 0x6484, 0xA448, 0x2448, 0x2448, 0x3C50, 0x2410, 0x23FE, 0x0}},
    {30765, 2, {0x4, 0x1E, 0xFDE0, 0x1000, 0x1020, 0x2010, 0x3DFE, 0x6404, 0x6408, 0xA410, 0x2420, 0x2440, 0x3C80, 0x2540, 0x223E, 0x0}},
    {30766, 2, {0x1000, 0x1000, 0xFEFC, 0x2244, 0x6428, 0x1810, 0x2428, 0xC2C6, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30767, 2, {0x20, 0x20, 0xFD20, 0x10A0, 0x1032, 0x21B2, 0x3CB4, 0x64A8, 0x64A8, 0xA4A8, 0x2524, 0x2524, 0x3E22, 0x2420, 0x20A0, 0x40}},
    {30768, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x1124, 0x20A4, 0x3CA8, 0x6420, 0x67FE, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x2020, 0x20}},
    {30769, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x2124, 0x3E12, 0x6410, 0x65FC, 0xA404, 0x2408, 0x2488, 0x3C50, 0x2420, 0x2010, 0x10}},
    {30770, 2, {0x80, 0x80, 0xFDFC, 0x1104, 0x1204, 0x25F4, 0x3D14, 0x6514, 0x6514, 0xA5F4, 0x2504, 0x2528, 0x3D12, 0x2502, 0x20FE, 0x0}},
    {30771, 2, {0x0, 0x0, 0xFEFE, 0x1020, 0x1020, 0x2040, 0x3E7C, 0x62C4, 0x62C4, 0xA344, 0x2244, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x0}},
    {30772, 2, {0x10, 0x10, 0xFC10, 0x11FE, 0x1112, 0x2114, 0x3D10, 0x65FC, 0x6544, 0xA544, 0x2528, 0x2528, 0x3D10, 0x2628, 0x2244, 0x482}},
    {30773, 2, {0x20, 0x20, 0xFC20, 0x1020, 0x11FE, 0x2020, 0x3C70, 0x6470, 0x64A8, 0xA4A8, 0x2524, 0x26FA, 0x3C20, 0x2420, 0x2020, 0x20}},
    {30774, 2, {0x20, 0x20, 0xFC40, 0x11FC, 0x1104, 0x2104, 0x3D04, 0x6504, 0x65FC, 0xA504, 0x2504, 0x2504, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30775, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1124, 0x2124, 0x3D24, 0x65FC, 0x6524, 0xA524, 0x2524, 0x25FC, 0x3D24, 0x2420, 0x2020, 0x20}},
    {30776, 2, {0x0, 0x3FE, 0xFA10, 0x2210, 0x2210, 0x42FE, 0x7A92, 0x4A92, 0xCA92, 0x4A92, 0x4A9A, 0x4A94, 0x7A10, 0x4A10, 0x3FE, 0x0}},
    {30777, 2, {0x110, 0x110, 0xFBFC, 0x2110, 0x2110, 0x4000, 0x7A08, 0xCA08, 0x4910, 0x4910, 0x48A0, 0x4840, 0x78A0, 0x4110, 0x208, 0xC06}},
    {30778, 2, {0x0, 0x1FE, 0xFD00, 0x1100, 0x1100, 0x21FE, 0x3D20, 0x6520, 0x653C, 0xA524, 0x2524, 0x2524, 0x3D24, 0x2244, 0x254, 0x488}},
    {30779, 2, {0x440, 0x420, 0xFFFC, 0x910, 0x1160, 0x2184, 0x4F04, 0x80FC, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30780, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x2104, 0x3E02, 0x6400, 0x65FC, 0xA420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {30781, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x1124, 0x21FC, 0x3D24, 0x6524, 0x6524, 0xA5FC, 0x2524, 0x2524, 0x3D24, 0x2524, 0x2104, 0x20C}},
    {30782, 2, {0x8, 0x1C, 0xFDE0, 0x1100, 0x1120, 0x2120, 0x3D20, 0x65FE, 0x6420, 0xA420, 0x24A8, 0x24A4, 0x3D22, 0x2622, 0x20A0, 0x40}},
    {30783, 2, {0x20, 0x10, 0xF810, 0x21FE, 0x2100, 0x4110, 0x7910, 0x4910, 0xC910, 0x4920, 0x4928, 0x4924, 0x7A44, 0x4AFE, 0x442, 0x800}},
    {30784, 2, {0x20, 0x20, 0xFD24, 0x1124, 0x1124, 0x2124, 0x3DFC, 0x6420, 0x6420, 0xA524, 0x2524, 0x2524, 0x3D24, 0x2524, 0x21FC, 0x4}},
    {30785, 2, {0x0, 0x1FC, 0xFC08, 0x1010, 0x1030, 0x2048, 0x3C84, 0x6702, 0x6400, 0xA5FC, 0x2420, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {30786, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x2202, 0x3DFC, 0x6420, 0x6420, 0xA420, 0x25FC, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {30787, 2, {0x20, 0x120, 0xFD20, 0x11FC, 0x1120, 0x2220, 0x3C20, 0x67FE, 0x6470, 0xA4A8, 0x24A8, 0x2524, 0x3D24, 0x2622, 0x2020, 0x20}},
    {30788, 2, {0x20, 0x20, 0xFD24, 0x10A4, 0x10A8, 0x2020, 0x3DFE, 0x6490, 0x6490, 0xA490, 0x2490, 0x2492, 0x3D12, 0x2512, 0x220E, 0x400}},
    {30789, 2, {0x20, 0x20, 0xF9FC, 0x2020, 0x2020, 0x4020, 0x7BFE, 0x4800, 0xC820, 0x4820, 0x49FC, 0x4820, 0x7820, 0x4820, 0x3FE, 0x0}},
    {30790, 2, {0x20, 0x20, 0xFC50, 0x1050, 0x1088, 0x2104, 0x3EFA, 0x6400, 0x6400, 0xA4F8, 0x2488, 0x2488, 0x3C88, 0x2488, 0x20F8, 0x88}},
    {30791, 2, {0x20, 0x20, 0xFC40, 0x11FC, 0x1104, 0x2104, 0x3D94, 0x6554, 0x6524, 0xA524, 0x2554, 0x2594, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30792, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x2020, 0x3DFC, 0x6400, 0x6400, 0xA5FC, 0x2504, 0x2504, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30793, 2, {0x20, 0x20, 0xFDFC, 0x1024, 0x13FE, 0x2024, 0x21FC, 0x7C20, 0x6420, 0xA5FC, 0x2420, 0x2420, 0x27FE, 0x3C20, 0x2420, 0x20}},
    {30794, 2, {0x40, 0x40, 0xFCFC, 0x1104, 0x1208, 0x21FE, 0x3D00, 0x657C, 0x6544, 0xA544, 0x2554, 0x2548, 0x3D42, 0x2542, 0x213E, 0x200}},
    {30795, 2, {0x20, 0x10, 0xFC10, 0x13FE, 0x1020, 0x2042, 0x3C84, 0x65F8, 0x6410, 0xA422, 0x24C4, 0x2708, 0x3C10, 0x2428, 0x20C4, 0x302}},
    {30796, 2, {0x40, 0x40, 0xFC78, 0x1088, 0x1150, 0x2020, 0x3C50, 0x6488, 0x6506, 0xA4F8, 0x2488, 0x2488, 0x3C88, 0x2488, 0x20F8, 0x88}},
    {30797, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x11F8, 0x2108, 0x3D08, 0x65F8, 0x6544, 0xA548, 0x2530, 0x2520, 0x3D10, 0x2548, 0x2186, 0x100}},
    {30798, 2, {0x2, 0x7E2, 0xFA42, 0x224A, 0x224A, 0x424A, 0x7A4A, 0x4FEA, 0xCA4A, 0x4A4A, 0x4A4A, 0x4A4A, 0x7A42, 0x4C42, 0x44A, 0x844}},
    {30799, 2, {0x0, 0x1DC, 0xFC88, 0x1088, 0x1088, 0x2088, 0x3C88, 0x67DE, 0x6488, 0xA488, 0x2488, 0x2488, 0x3C88, 0x2508, 0x2108, 0x208}},
    {30800, 2, {0x0, 0x1FE, 0xFD02, 0x1102, 0x117A, 0x2102, 0x3D02, 0x657A, 0x654A, 0xA54A, 0x254A, 0x257A, 0x3D02, 0x2502, 0x210A, 0x104}},
    {30801, 2, {0x104, 0x84, 0xFC88, 0x1000, 0x13FE, 0x2088, 0x3C88, 0x6488, 0x6488, 0xA7FE, 0x2488, 0x2488, 0x3D08, 0x2508, 0x2208, 0x408}},
    {30802, 2, {0x0, 0x3FE, 0xFC50, 0x1050, 0x11FC, 0x2154, 0x3D54, 0x6554, 0x6554, 0xA554, 0x258C, 0x2504, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30803, 2, {0x40, 0x40, 0xFC44, 0x11F4, 0x1048, 0x2050, 0x23FE, 0x7C40, 0x6480, 0xA584, 0x2698, 0x24E0, 0x2482, 0x3C82, 0x247E, 0x0}},
    {30804, 2, {0x88, 0x88, 0xFC88, 0x1088, 0x13FE, 0x2088, 0x3C88, 0x6488, 0x6488, 0xA7FE, 0x2400, 0x2488, 0x3C84, 0x2504, 0x2202, 0x402}},
    {30805, 2, {0x0, 0x3FE, 0xFC20, 0x1040, 0x11FC, 0x2104, 0x3D04, 0x6524, 0x6524, 0xA524, 0x2524, 0x2544, 0x3C50, 0x2488, 0x2104, 0x204}},
    {30806, 2, {0x20, 0x20, 0xFC20, 0x11FC, 0x1020, 0x2124, 0x3CA4, 0x64A8, 0x6420, 0xA7FE, 0x2450, 0x2450, 0x3C88, 0x2488, 0x2104, 0x202}},
    {30807, 2, {0x80, 0x80, 0xF8BC, 0x23C0, 0x2050, 0x4024, 0x78D4, 0x4B0C, 0xC800, 0x4BFE, 0x4890, 0x4890, 0x7912, 0x4912, 0x20E, 0x400}},
    {30808, 2, {0x0, 0x0, 0xFDFC, 0x1104, 0x1104, 0x2174, 0x3D54, 0x6554, 0x6554, 0xA554, 0x2574, 0x2504, 0x3D04, 0x25FC, 0x2104, 0x0}},
    {30809, 2, {0x20, 0x124, 0xFD24, 0x1124, 0x11FC, 0x2000, 0x3DFC, 0x6404, 0x6404, 0xA5FC, 0x2500, 0x2500, 0x3D02, 0x2502, 0x20FE, 0x0}},
    {30810, 2, {0x8, 0x3C, 0xFDE0, 0x1020, 0x1020, 0x23FE, 0x3C50, 0x6488, 0x6504, 0xA68A, 0x2488, 0x2488, 0x3C88, 0x2508, 0x2108, 0x208}},
    {30811, 2, {0x40, 0x20, 0xF820, 0x27FE, 0x2090, 0x4090, 0x7890, 0x4A94, 0xCA92, 0x4A92, 0x4C92, 0x4890, 0x7910, 0x4910, 0x250, 0x420}},
    {30812, 2, {0x0, 0x1FE, 0xFC00, 0x1092, 0x1124, 0x2248, 0x3D24, 0x6492, 0x6400, 0xA5FE, 0x2420, 0x2420, 0x3C20, 0x2420, 0x23FE, 0x0}},
    {30813, 2, {0x20, 0x124, 0xFCA4, 0x10A8, 0x1020, 0x21FC, 0x3D04, 0x6504, 0x65FC, 0xA504, 0x2504, 0x25FC, 0x3D04, 0x2504, 0x2114, 0x108}},
    {30814, 2, {0x20, 0x120, 0xFD20, 0x11FC, 0x1120, 0x2220, 0x3C20, 0x67FE, 0x6400, 0xA400, 0x25FC, 0x2504, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30815, 2, {0x4, 0xE, 0xFBB8, 0x2088, 0x2088, 0x4128, 0x792E, 0x4BA8, 0xC8A8, 0x4AA8, 0x4AA8, 0x493E, 0x7900, 0x4A80, 0x47E, 0x800}},
    {30816, 2, {0x40, 0x20, 0xFDFC, 0x1104, 0x1104, 0x21FC, 0x3D04, 0x6504, 0x65FC, 0xA520, 0x2522, 0x2514, 0x3D08, 0x2544, 0x2182, 0x100}},
    {30817, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2444, 0x4040, 0x7BFE, 0x4840, 0xC890, 0x4890, 0x4910, 0x4920, 0x7A24, 0x4A42, 0x4FE, 0x842}},
    {30818, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x22FA, 0x3C20, 0x6420, 0x67FE, 0xA420, 0x2528, 0x2524, 0x3E22, 0x2422, 0x20A0, 0x40}},
    {30819, 2, {0x20, 0x22, 0xFDFA, 0x1024, 0x1024, 0x23FE, 0x3C10, 0x6420, 0x65FC, 0xA488, 0x2510, 0x26FE, 0x3C10, 0x2410, 0x2050, 0x20}},
    {30820, 2, {0x20, 0x20, 0xFC20, 0x13FE, 0x1020, 0x2020, 0x3D24, 0x6524, 0x6524, 0xA6AA, 0x2420, 0x2450, 0x3C50, 0x2488, 0x2104, 0x202}},
    {30821, 2, {0x90, 0x88, 0xFC80, 0x13FE, 0x10A0, 0x20A8, 0x3CB0, 0x64A4, 0x64A8, 0xA4B0, 0x2524, 0x252A, 0x3D32, 0x2222, 0x2DE, 0x400}},
    {30822, 2, {0x0, 0x3FE, 0xFC20, 0x1020, 0x11FC, 0x2020, 0x3C20, 0x67FE, 0x6488, 0xA488, 0x27FE, 0x2488, 0x3C88, 0x2508, 0x2108, 0x208}},
    {30823, 2, {0x0, 0x1F8, 0xFC08, 0x10D0, 0x1020, 0x21FC, 0x3D24, 0x6524, 0x65FC, 0xA524, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x2124, 0x10C}},
    {30824, 2, {0x20, 0x20, 0xFDFE, 0x1020, 0x11FC, 0x2124, 0x3D24, 0x65FC, 0x6524, 0xA524, 0x25FC, 0x2420, 0x3DFE, 0x2420, 0x2020, 0x20}},
    {30825, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30826, 2, {0x10, 0xD8, 0xFB94, 0x2094, 0x2090, 0x43FE, 0x7890, 0x4894, 0xC894, 0x48D8, 0x4B98, 0x4890, 0x78AA, 0x48CA, 0x286, 0x102}},
    {30827, 2, {0x40, 0x20, 0xFBFC, 0x2040, 0x2090, 0x4108, 0x7BFC, 0x4804, 0xC950, 0x4950, 0x4950, 0x4950, 0x7A50, 0x4A52, 0x452, 0x80E}},
    {30828, 2, {0x0, 0x1FE, 0xFC20, 0x1020, 0x11FC, 0x2124, 0x3D24, 0x65FC, 0x6524, 0xA524, 0x25FC, 0x2520, 0x3CA0, 0x2440, 0x20B0, 0x30E}},
    {30829, 2, {0x48, 0x48, 0xFDFE, 0x1048, 0x1048, 0x2020, 0x3C10, 0x65FE, 0x6480, 0xA480, 0x2480, 0x2480, 0x3C80, 0x2480, 0x20FC, 0x0}},
    {30830, 2, {0x40, 0x40, 0xFC7C, 0x1084, 0x1108, 0x22FE, 0x3C92, 0x6492, 0x64FE, 0xA492, 0x2492, 0x24FE, 0x3C92, 0x2512, 0x210A, 0x204}},
    {30831, 2, {0x0, 0x1F8, 0xFD08, 0x1108, 0x11F8, 0x2108, 0x3D08, 0x65F8, 0x6508, 0xA508, 0x25F8, 0x2490, 0x3C90, 0x2112, 0x212, 0x40E}},
    {30832, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x92A, 0x7230, 0x10C0, 0x1700, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30833, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x1124, 0x21FC, 0x3D24, 0x6564, 0x6574, 0xA5AC, 0x2524, 0x2524, 0x3D24, 0x2504, 0x21FC, 0x104}},
    {30834, 2, {0x0, 0x50, 0xFC48, 0x1084, 0x1124, 0x2020, 0x3C50, 0x6488, 0x6506, 0xA6F8, 0x2488, 0x2488, 0x3C88, 0x2488, 0x20F8, 0x88}},
    {30835, 2, {0x20, 0x20, 0xF9FC, 0x2020, 0x2020, 0x4020, 0x7BFE, 0x4848, 0xC848, 0x494C, 0x494A, 0x4A4A, 0x7888, 0x4888, 0x128, 0x210}},
    {30836, 2, {0x88, 0x88, 0xFDFE, 0x1088, 0x1088, 0x2050, 0x3C50, 0x6494, 0x6594, 0xA698, 0x2490, 0x24B2, 0x3CD2, 0x2492, 0x208E, 0x80}},
    {30837, 2, {0x20, 0x20, 0xFC3E, 0x1020, 0x1020, 0x21FE, 0x3D02, 0x654A, 0x652A, 0xA512, 0x2512, 0x252A, 0x3D4A, 0x2502, 0x21FE, 0x102}},
    {30838, 2, {0x20, 0x124, 0xFD24, 0x1124, 0x11FC, 0x2020, 0x3C50, 0x6488, 0x6544, 0xA622, 0x2420, 0x25F8, 0x3C08, 0x2410, 0x2010, 0x20}},
    {30839, 2, {0x40, 0x40, 0xF8A0, 0x20A0, 0x2110, 0x4208, 0x7DF6, 0x4800, 0xC888, 0x4848, 0x4A48, 0x4950, 0x7910, 0x4820, 0x7FE, 0x0}},
    {30840, 2, {0x20, 0x124, 0xFD24, 0x1124, 0x11FC, 0x2000, 0x3DFE, 0x6500, 0x657C, 0xA510, 0x2510, 0x25FE, 0x3D10, 0x2210, 0x210, 0x410}},
    {30841, 2, {0x208, 0x228, 0xF228, 0x2228, 0x2FA4, 0x4244, 0x7254, 0xD692, 0x5710, 0x5AA0, 0x5A20, 0x5228, 0x7244, 0x52FC, 0x244, 0x200}},
    {30842, 2, {0x0, 0x1FE, 0xFC20, 0x1040, 0x10A2, 0x2112, 0x3C34, 0x6558, 0x6498, 0xA554, 0x2434, 0x2452, 0x3C90, 0x2510, 0x2050, 0x20}},
    {30843, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30844, 2, {0x0, 0x3DE, 0xFA52, 0x2252, 0x2252, 0x43DE, 0x7A52, 0x4A52, 0xCA52, 0x4BDE, 0x4A52, 0x4A52, 0x7A52, 0x4A52, 0x552, 0x8A6}},
    {30845, 2, {0x40, 0x80, 0xFBFC, 0x2110, 0x2248, 0x4446, 0x7BF8, 0x4A48, 0xCA48, 0x4BF8, 0x4A48, 0x4A48, 0x7BF8, 0x4842, 0x42, 0x3E}},
    {30846, 2, {0x8, 0x3C, 0xFDE0, 0x1020, 0x13FE, 0x20A8, 0x3CA8, 0x64A8, 0x67FE, 0xA4A8, 0x24A8, 0x27FE, 0x3C20, 0x2420, 0x21FC, 0x0}},
    {30847, 2, {0x40, 0x20, 0xF820, 0x23FE, 0x2202, 0x4494, 0x7908, 0x4A04, 0xC800, 0x49FC, 0x4820, 0x4820, 0x7820, 0x4820, 0x7FE, 0x0}},
    {30848, 2, {0x40, 0x40, 0xFCF8, 0x1108, 0x1010, 0x21FC, 0x3C24, 0x6424, 0x67FE, 0xA424, 0x2424, 0x25FC, 0x3C24, 0x2420, 0x20A0, 0x40}},
    {30849, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1010, 0x3FF8, 0xC406, 0xFE0, 0x1820, 0x6820, 0xFE0, 0x820}},
    {30850, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2000, 0x41FC, 0x7800, 0x4800, 0xCBFE, 0x4820, 0x4928, 0x4924, 0x7A22, 0x4C22, 0xA0, 0x40}},
    {30851, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x4020, 0x7BFE, 0x4800, 0xC9FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x114, 0x108}},
    {30852, 2, {0x110, 0x110, 0xF910, 0x2110, 0x27BC, 0x4110, 0x7910, 0x4B38, 0xCBB8, 0x4D54, 0x4D54, 0x4992, 0x7910, 0x4910, 0x110, 0x110}},
    {30853, 2, {0x0, 0x3FE, 0xFA02, 0x221A, 0x22E2, 0x4222, 0x7A22, 0x4BFE, 0xCA22, 0x4A72, 0x4AAA, 0x4B26, 0x7A22, 0x4A22, 0x3FE, 0x202}},
    {30854, 2, {0x2020, 0x13FC, 0x8224, 0x4BF8, 0x1288, 0x2450, 0xE470, 0x298C, 0x2000, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30855, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2404, 0x4000, 0x7BFE, 0x4820, 0xC820, 0x4920, 0x493C, 0x4920, 0x7AA0, 0x4A60, 0x43E, 0x800}},
    {30856, 2, {0x1C, 0x3E0, 0xFA20, 0x23FE, 0x2220, 0x4292, 0x7B0A, 0x4A06, 0xC9FC, 0x4904, 0x4904, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30857, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x1174, 0x2124, 0x3DFC, 0x6504, 0x6574, 0xA554, 0x2554, 0x2574, 0x3D04, 0x2204, 0x214, 0x408}},
    {30858, 2, {0x48, 0x44, 0xFC5E, 0x11E0, 0x1028, 0x2012, 0x3C6A, 0x6596, 0x6448, 0xA45E, 0x25E0, 0x2424, 0x3C28, 0x2412, 0x206A, 0x186}},
    {30859, 2, {0x20, 0x124, 0xFD24, 0x1124, 0x11FC, 0x2000, 0x3DFE, 0x6408, 0x65E8, 0xA528, 0x2528, 0x25E8, 0x3D28, 0x2408, 0x2028, 0x10}},
    {30860, 2, {0x0, 0x1F8, 0xF808, 0x2008, 0x21F8, 0x4008, 0x7808, 0x4BFE, 0xC820, 0x4A22, 0x4974, 0x48A8, 0x7924, 0x4A22, 0xA0, 0x40}},
    {30861, 2, {0x0, 0x1F8, 0xFD08, 0x11F8, 0x1108, 0x21F8, 0x3C00, 0x65FC, 0x6410, 0xA7FE, 0x2410, 0x2510, 0x3C90, 0x2410, 0x2050, 0x20}},
    {30862, 2, {0x40, 0x20, 0xFDFE, 0x1000, 0x1088, 0x2088, 0x3D54, 0x6622, 0x6400, 0xA420, 0x27FE, 0x2420, 0x3C20, 0x2420, 0x2020, 0x20}},
    {30863, 2, {0x48, 0x48, 0xFC48, 0x11FE, 0x1048, 0x2048, 0x3DFE, 0x6400, 0x64FC, 0xA484, 0x2484, 0x24FC, 0x3C84, 0x2484, 0x20FC, 0x84}},
    {30864, 2, {0x20, 0x20, 0xF9FC, 0x2020, 0x2020, 0x43FE, 0x7888, 0x4944, 0xCA42, 0x48F8, 0x4988, 0x4A50, 0x7820, 0x4850, 0x188, 0x606}},
    {30865, 2, {0x40, 0x80, 0xFDFC, 0x1124, 0x1124, 0x21FC, 0x3D24, 0x6544, 0x65FC, 0xA490, 0x2510, 0x27FE, 0x3C10, 0x2410, 0x2010, 0x10}},
    {30866, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x2202, 0x3CF8, 0x6420, 0x6420, 0xA5FC, 0x2420, 0x2524, 0x3CA4, 0x24A8, 0x23FE, 0x0}},
    {30867, 2, {0x50, 0x48, 0xFC80, 0x10FE, 0x1190, 0x2290, 0x3CFC, 0x6490, 0x6490, 0xA4FC, 0x2490, 0x2490, 0x3C90, 0x24FE, 0x2080, 0x80}},
    {30868, 2, {0x8, 0xC, 0xFCEA, 0x100A, 0x1008, 0x21FE, 0x3C48, 0x6448, 0x6548, 0xA568, 0x2548, 0x2548, 0x3D4A, 0x256A, 0x2386, 0x102}},
    {30869, 2, {0x20, 0x20, 0xFDFC, 0x1050, 0x1088, 0x2104, 0x23FE, 0x7C08, 0x65E8, 0xA528, 0x2528, 0x25E8, 0x2528, 0x3C08, 0x2428, 0x10}},
    {30870, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x22FA, 0x3C00, 0x65FC, 0x6554, 0xA554, 0x25FC, 0x2554, 0x3D54, 0x2554, 0x2104, 0x10C}},
    {30871, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2504, 0x4100, 0x79DE, 0x4A52, 0xCA52, 0x4B52, 0x4C9A, 0x4894, 0x7910, 0x4912, 0x212, 0x40E}},
    {30872, 2, {0x50, 0x50, 0xFC50, 0x11FC, 0x1154, 0x2154, 0x3D54, 0x65FC, 0x6554, 0xA554, 0x2554, 0x27FE, 0x3C00, 0x2488, 0x2104, 0x202}},
    {30873, 2, {0x0, 0x3FE, 0xFA02, 0x228A, 0x2252, 0x43FE, 0x7A22, 0x4A22, 0xCAAA, 0x4AAA, 0x4AAA, 0x4AFA, 0x7A02, 0x4A02, 0x20A, 0x204}},
    {30874, 2, {0x40, 0x20, 0xFDFC, 0x1000, 0x1108, 0x2090, 0x3C00, 0x67FE, 0x6400, 0xA400, 0x25FC, 0x2504, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30875, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x4020, 0x7BFE, 0x4800, 0xC9FC, 0x4904, 0x4924, 0x4924, 0x7924, 0x4850, 0x88, 0x304}},
    {30876, 2, {0x80, 0x90, 0xF908, 0x23FC, 0x2040, 0x47FE, 0x7910, 0x4A48, 0xCC86, 0x4B10, 0x4820, 0x48C4, 0x7B08, 0x4830, 0xC0, 0x700}},
    {30877, 2, {0x0, 0x3FE, 0xFC20, 0x1040, 0x11FC, 0x2154, 0x3D54, 0x6554, 0x6554, 0xA52C, 0x2420, 0x27FE, 0x3C50, 0x2488, 0x2104, 0x202}},
    {30878, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30879, 2, {0x90, 0x290, 0xFA90, 0x27FE, 0x2290, 0x4290, 0x7AF0, 0xCA00, 0x4BFC, 0x4840, 0x4FFE, 0x48E0, 0x7950, 0x4248, 0xC46, 0x40}},
    {30880, 2, {0x40, 0x20, 0xFBFE, 0x2000, 0x21FC, 0x4104, 0x79FC, 0x4800, 0xCBFE, 0x4A02, 0x49FC, 0x4820, 0x7820, 0x4820, 0xA0, 0x40}},
    {30881, 2, {0x20, 0x1FE, 0xFC20, 0x10FC, 0x1020, 0x21FE, 0x3C00, 0x64FC, 0x64A4, 0xA494, 0x25FE, 0x24A4, 0x3D14, 0x25FE, 0x2004, 0x18}},
    {30882, 2, {0x0, 0xFC, 0xFC84, 0x1084, 0x10F4, 0x2094, 0x3C94, 0x65FE, 0x6502, 0xA57A, 0x254A, 0x254A, 0x3D7A, 0x2502, 0x210A, 0x104}},
    {30883, 2, {0x0, 0x1FC, 0xFD04, 0x11FC, 0x1104, 0x21FC, 0x3C80, 0x65FE, 0x6622, 0xA522, 0x2552, 0x2502, 0x3DFA, 0x2402, 0x2014, 0x8}},
    {30884, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20A8, 0x4020, 0x79FC, 0x4924, 0xC924, 0x4924, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x202}},
    {30885, 2, {0x80, 0x40, 0xFBFC, 0x2204, 0x2204, 0x43FC, 0x7A00, 0xCA00, 0x4BFC, 0x4B54, 0x4B54, 0x4DFC, 0x7D54, 0x4554, 0x944, 0x10C}},
    {30886, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2080, 0x40F8, 0x7908, 0x4A90, 0xC860, 0x4998, 0x4E06, 0x49F8, 0x7908, 0x4908, 0x1F8, 0x108}},
    {30887, 2, {0x20, 0xFE40, 0x10FC, 0x7C84, 0x10FC, 0x1E84, 0xF0FC, 0x0, 0x7FFC, 0x400, 0x800, 0x3FF0, 0xC810, 0x810, 0xFF0, 0x810}},
    {30888, 2, {0x0, 0x1FC, 0xFD24, 0x1124, 0x11FC, 0x2124, 0x3D24, 0x65FC, 0x6400, 0xA7FE, 0x2520, 0x2522, 0x3D14, 0x2548, 0x2184, 0x102}},
    {30889, 2, {0x0, 0x3FE, 0xFC20, 0x1040, 0x11FC, 0x2104, 0x3D04, 0x65FC, 0x6504, 0xA5FC, 0x2504, 0x2504, 0x3DFC, 0x2400, 0x2088, 0x104}},
    {30890, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x20F8, 0x4088, 0x78F8, 0x4888, 0xC888, 0x4BFE, 0x4900, 0x4948, 0x7984, 0x4900, 0x1FE, 0x0}},
    {30891, 2, {0x80, 0x338, 0xFA28, 0x2228, 0x23A8, 0x4246, 0x7A00, 0x4A7C, 0xCBA4, 0x4A24, 0x4A28, 0x4BA8, 0x7E10, 0x4A28, 0x244, 0x282}},
    {30892, 2, {0x0, 0x3DE, 0xFA42, 0x2242, 0x2242, 0x43DE, 0x7A00, 0x4A3E, 0xCBD2, 0x4A12, 0x4A14, 0x4BD4, 0x7A08, 0x4A14, 0x224, 0x242}},
    {30893, 2, {0x0, 0x1F8, 0xF908, 0x21F8, 0x2108, 0x41F8, 0x7800, 0x4BFE, 0xC900, 0x49FC, 0x4A54, 0x4C94, 0x7924, 0x4A44, 0xA8, 0x110}},
    {30894, 2, {0x0, 0x1FC, 0xFD04, 0x1104, 0x11FC, 0x2104, 0x3D04, 0x65FC, 0x6400, 0xA7FE, 0x2420, 0x2520, 0x3D3C, 0x2520, 0x22A0, 0x47E}},
    {30895, 2, {0x92, 0x124, 0xFA48, 0x2124, 0x2092, 0x4040, 0x7880, 0x4BFE, 0xCA02, 0x4A8A, 0x4A52, 0x4A22, 0x7A52, 0x4A8A, 0x3FE, 0x202}},
    {30896, 2, {0x108, 0x90, 0xF800, 0x23FC, 0x2090, 0x4090, 0x7892, 0x4C92, 0xCA94, 0x4A98, 0x4890, 0x4890, 0x7890, 0x4890, 0x7FE, 0x0}},
    {30897, 2, {0x14, 0x12, 0xF010, 0x27FE, 0x2410, 0x4410, 0x75D0, 0xD412, 0x5412, 0x55D4, 0x5554, 0x5548, 0x75DA, 0x542A, 0x846, 0x1082}},
    {30898, 2, {0x40, 0x20, 0xF9FC, 0x2000, 0x2088, 0x4050, 0x7BFE, 0x4A22, 0xCC24, 0x49FC, 0x4924, 0x4924, 0x7934, 0x4928, 0x20, 0x20}},
    {30899, 2, {0x20, 0x124, 0xF924, 0x2124, 0x21FC, 0x4080, 0x7880, 0x4BFE, 0xC890, 0x4910, 0x4952, 0x4A54, 0x7AA8, 0x4C28, 0x844, 0x82}},
    {30900, 2, {0x20, 0x20, 0xFBFE, 0x2070, 0x20A8, 0x4124, 0x7A22, 0x49FC, 0xC904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4800, 0x3FE, 0x0}},
    {30901, 2, {0x20, 0x20, 0xFC3E, 0x1020, 0x1020, 0x21FC, 0x3D04, 0x65FC, 0x6504, 0xA5FC, 0x2504, 0x25FC, 0x3C00, 0x2488, 0x2104, 0x202}},
    {30902, 2, {0x80, 0x9E, 0xFBEA, 0x208A, 0x21CA, 0x408A, 0x7BEA, 0x4892, 0xC8A6, 0x4820, 0x4BFE, 0x4820, 0x7850, 0x4888, 0x104, 0x602}},
    {30903, 2, {0x8, 0x1C, 0xFDF0, 0x1110, 0x1110, 0x21FE, 0x3D10, 0x657C, 0x6544, 0xA57C, 0x2544, 0x257C, 0x3D44, 0x2244, 0x27C, 0x444}},
    {30904, 2, {0x0, 0x7FC, 0xF404, 0x2434, 0x25C4, 0x4444, 0x75F4, 0xD554, 0x5554, 0x55F4, 0x5444, 0x5454, 0x75F4, 0x5494, 0x806, 0x1002}},
    {30905, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2404, 0x41FC, 0x7800, 0x49FC, 0xC904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4800, 0x3FE, 0x0}},
    {30906, 2, {0x40, 0x40, 0xFBFE, 0x22A2, 0x2090, 0x41FE, 0x7910, 0x4B10, 0xCDFE, 0x4910, 0x4910, 0x49FE, 0x7910, 0x4910, 0x1FE, 0x100}},
    {30907, 2, {0x20, 0x10, 0xFDFE, 0x1000, 0x10FC, 0x2084, 0x3CFC, 0x6400, 0x65FE, 0xA502, 0x257A, 0x254A, 0x3D7A, 0x2502, 0x210A, 0x104}},
    {30908, 2, {0x0, 0x1FE, 0xF920, 0x21FC, 0x2120, 0x41FC, 0x7920, 0x4920, 0xC9FE, 0x4802, 0x4AAA, 0x4AAA, 0x7AAA, 0x4A02, 0x14, 0x8}},
    {30909, 2, {0x0, 0x1FC, 0xFC20, 0x1020, 0x13FE, 0x2000, 0x3DFC, 0x6504, 0x65FC, 0xA504, 0x25FC, 0x2504, 0x3DFC, 0x2488, 0x2104, 0x202}},
    {30910, 2, {0x0, 0x1FE, 0xFD02, 0x1102, 0x11FE, 0x2124, 0x3D24, 0x657E, 0x6524, 0xA524, 0x25FE, 0x2552, 0x3D54, 0x2648, 0x2264, 0x442}},
    {30911, 2, {0x0, 0x3FE, 0xFA02, 0x2202, 0x23FE, 0x4210, 0x7A92, 0xCA54, 0x4AFE, 0x4A82, 0x4AFE, 0x4A82, 0x7AFE, 0x4482, 0x48A, 0x884}},
    {30912, 2, {0x244, 0x124, 0xF928, 0x207E, 0x2010, 0x4354, 0x7954, 0x4954, 0xC97C, 0x4914, 0x4920, 0x4920, 0x7940, 0x4A80, 0x47E, 0x0}},
    {30913, 2, {0x108, 0x88, 0xF090, 0x27FE, 0x2108, 0x4108, 0x7210, 0xD294, 0x54A4, 0x5738, 0x5108, 0x5210, 0x7210, 0x54A4, 0x7BC, 0x84}},
    {30914, 2, {0x40, 0x3BE, 0xFA12, 0x2292, 0x2252, 0x42AA, 0x7B24, 0x4840, 0xCBFE, 0x4A22, 0x4A22, 0x4BFE, 0x7A22, 0x4A22, 0x3FE, 0x202}},
    {30915, 2, {0x4, 0x1E, 0xFBF0, 0x221E, 0x2210, 0x42FE, 0x7A92, 0xCA98, 0x4AF2, 0x4A8E, 0x4A80, 0x4AB8, 0x7AA8, 0x44AA, 0x54A, 0xA86}},
    {30916, 2, {0x20, 0x10, 0xFDFE, 0x1110, 0x117C, 0x2114, 0x3DFE, 0x6514, 0x657C, 0xA510, 0x257C, 0x2544, 0x3D44, 0x2244, 0x27C, 0x444}},
    {30917, 2, {0x40, 0x20, 0xFBFC, 0x2108, 0x2090, 0x43FE, 0x7A02, 0x4C44, 0xC820, 0x4BFC, 0x4880, 0x48F8, 0x7888, 0x4908, 0x128, 0x210}},
    {30918, 2, {0x0, 0x1FC, 0xFD04, 0x11E4, 0x1124, 0x23FE, 0x2202, 0x7DFC, 0x6504, 0xA5FC, 0x2504, 0x25FC, 0x2504, 0x3D04, 0x2514, 0x108}},
    {30919, 2, {0x20, 0x40, 0xFDFC, 0x1104, 0x1154, 0x2124, 0x3D54, 0x6504, 0x65FC, 0xA400, 0x2512, 0x25D4, 0x3D18, 0x2552, 0x2192, 0x10E}},
    {30920, 2, {0x20, 0x40, 0xFBFC, 0x2224, 0x2224, 0x43FC, 0x7A24, 0x4A44, 0xCBFC, 0x4840, 0x48A8, 0x48B4, 0x793C, 0x4922, 0x222, 0x41E}},
    {30921, 2, {0x3F0, 0x110, 0xF0E0, 0x2318, 0x2000, 0x47BC, 0x74A4, 0xD318, 0x54A4, 0x5040, 0x57FC, 0x50E0, 0x7150, 0x5248, 0xC46, 0x40}},
    {30922, 2, {0x7FFC, 0x400, 0x800, 0x1FF0, 0x2810, 0x4810, 0x8FF0, 0x0, 0x7EFE, 0x1020, 0x2040, 0x7EFC, 0xA344, 0x2244, 0x3E7C, 0x2244}},
    {30923, 2, {0x84, 0x44, 0xFC48, 0x11FE, 0x1020, 0x20FC, 0x3C20, 0x65FE, 0x6440, 0xA480, 0x24FE, 0x2510, 0x3E10, 0x2410, 0x21FE, 0x0}},
    {30924, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x21FC, 0x4104, 0x79FC, 0x4904, 0xC9FC, 0x4904, 0x49FC, 0x4904, 0x7BFE, 0x4888, 0x104, 0x202}},
    {30925, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x2424, 0x41FC, 0x7820, 0x49FC, 0xC820, 0x4BFE, 0x4820, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30926, 2, {0x1E, 0x3E0, 0xFD22, 0x1094, 0x1040, 0x2088, 0x21F0, 0x7C20, 0x64C4, 0xA5FE, 0x2422, 0x2420, 0x27FE, 0x3C50, 0x2488, 0x306}},
    {30927, 2, {0x88, 0x50, 0xFC00, 0x13FE, 0x1050, 0x21FC, 0x3C54, 0x67FE, 0x6454, 0xA5FC, 0x2450, 0x24D8, 0x3D54, 0x2652, 0x2050, 0x50}},
    {30928, 2, {0x878, 0x3E48, 0x228E, 0x2A00, 0x7EF8, 0x2248, 0x2A30, 0x46CC, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30929, 2, {0x20, 0x124, 0xFD24, 0x1124, 0x11FC, 0x2000, 0x3FFE, 0x6400, 0x65FC, 0xA504, 0x2504, 0x25FC, 0x3C88, 0x2450, 0x23FE, 0x0}},
    {30930, 2, {0x0, 0xFC, 0xFC84, 0x1084, 0x10FC, 0x2000, 0x3DFE, 0x6502, 0x65FE, 0xA502, 0x25FE, 0x2502, 0x3DFE, 0x2448, 0x2084, 0x102}},
    {30931, 2, {0x10, 0x220, 0xF97C, 0x2144, 0x2044, 0x407C, 0x7B40, 0x4940, 0xC97C, 0x4944, 0x4944, 0x497C, 0x7944, 0x4A80, 0x47E, 0x0}},
    {30932, 2, {0x208, 0x208, 0xF3BE, 0x2488, 0x2AA8, 0x413E, 0x7208, 0xD448, 0x5848, 0x57FC, 0x50E0, 0x5150, 0x7248, 0x5C46, 0x40, 0x40}},
    {30933, 2, {0x20, 0x20, 0xFCFC, 0x1020, 0x1020, 0x21FE, 0x3C40, 0x6484, 0x65FE, 0xA402, 0x25FC, 0x2554, 0x3D54, 0x2554, 0x23FE, 0x0}},
    {30934, 2, {0x0, 0x1DE, 0xFC42, 0x1152, 0x10CA, 0x2152, 0x3C42, 0x6420, 0x64FC, 0xA484, 0x2484, 0x24FC, 0x3C84, 0x2484, 0x20FC, 0x84}},
    {30935, 2, {0x28, 0x24, 0xFBFE, 0x2020, 0x21FC, 0x4124, 0x79FC, 0x4924, 0xC9FC, 0x4924, 0x4808, 0x4BFE, 0x7888, 0x4848, 0x48, 0x18}},
    {30936, 2, {0x8, 0x3C, 0xFBC0, 0x2004, 0x2244, 0x4128, 0x79FC, 0x4A20, 0xC820, 0x4BFE, 0x4820, 0x4924, 0x7924, 0x4924, 0x1FC, 0x4}},
    {30937, 2, {0x40, 0x20, 0xFBFE, 0x2088, 0x2124, 0x4242, 0x7890, 0x49F8, 0xC808, 0x4844, 0x48A8, 0x4990, 0x7A88, 0x4CA4, 0xC2, 0x80}},
    {30938, 2, {0x20, 0x3FE, 0xFC20, 0x11FC, 0x1124, 0x21FC, 0x3D24, 0x65FC, 0x6422, 0xA7FE, 0x2408, 0x27FE, 0x3D08, 0x2488, 0x2028, 0x10}},
    {30939, 2, {0x804, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {30940, 2, {0x100, 0x110, 0xFBDC, 0x2254, 0x2554, 0x4288, 0x7908, 0x4AF4, 0xCC02, 0x4800, 0x4BFC, 0x4840, 0x7950, 0x4A48, 0x544, 0x80}},
    {30941, 2, {0x88, 0x88, 0xF888, 0x23D0, 0x209E, 0x43E4, 0x7894, 0x4FD4, 0xC914, 0x49D4, 0x4954, 0x4948, 0x7948, 0x4A54, 0x2D4, 0x422}},
    {30942, 2, {0x20, 0x222, 0xFA22, 0x23FE, 0x2000, 0x43DE, 0x7A52, 0x4A52, 0xCBDE, 0x4A52, 0x4A52, 0x4BDE, 0x7A52, 0x4A52, 0x55A, 0x8A4}},
    {30943, 2, {0x0, 0x7BC, 0xF884, 0x24A4, 0x2294, 0x44A4, 0x7850, 0x4988, 0xCE26, 0x48C0, 0x4B10, 0x4864, 0x7B88, 0x4830, 0xC0, 0x700}},
    {30944, 2, {0x20, 0x20, 0xF83E, 0x2020, 0x2020, 0x43FE, 0x7A02, 0x4B26, 0xCA8A, 0x4A52, 0x4B26, 0x4A52, 0x7A8A, 0x4B26, 0x3FE, 0x202}},
    {30945, 2, {0x248, 0x248, 0xF7E8, 0x2248, 0x23DE, 0x424A, 0x73CA, 0xD24A, 0x524A, 0x5FEA, 0x540A, 0x554A, 0x7632, 0x5412, 0x7EA, 0x44}},
    {30946, 2, {0x20, 0x20, 0xFBFE, 0x2020, 0x22AA, 0x4124, 0x7AAA, 0x4820, 0xCAAA, 0x4924, 0x4AAA, 0x4850, 0x7850, 0x4888, 0x104, 0x602}},
    {30947, 2, {0x40, 0x88, 0xF9FC, 0x2108, 0x2252, 0x43FE, 0x7850, 0x4988, 0xCE26, 0x48C0, 0x4B10, 0x4864, 0x7B88, 0x4830, 0xC0, 0x700}},
    {30948, 2, {0x0, 0xDC, 0xFB14, 0x2214, 0x23D4, 0x4254, 0x7BE6, 0xCA40, 0x4BDC, 0x4A14, 0x4BD4, 0x4A54, 0x7A48, 0x4448, 0x554, 0x8A2}},
    {30949, 2, {0x0, 0x1FC, 0xFD24, 0x11FC, 0x1124, 0x21FC, 0x3C40, 0x6488, 0x65F0, 0xA420, 0x2444, 0x27FE, 0x3C22, 0x2524, 0x2222, 0x60}},
    {30950, 2, {0x0, 0x3FE, 0xFC50, 0x11FC, 0x1154, 0x2154, 0x3DFC, 0x6400, 0x65FC, 0xA400, 0x27FE, 0x2420, 0x3CA8, 0x2524, 0x22A2, 0x40}},
    {30951, 2, {0x20, 0x3FE, 0xFC20, 0x11FC, 0x1020, 0x23FE, 0x3C00, 0x65FC, 0x6504, 0xA5FC, 0x2504, 0x25FC, 0x3D04, 0x25FC, 0x2088, 0x104}},
    {30952, 2, {0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2210, 0x2738, 0x2AD4, 0x2210, 0x2FFC, 0x2100, 0x2200, 0x27F8, 0x4A08, 0x5208, 0x83F8, 0x208}},
    {30953, 2, {0xC, 0xA, 0xF808, 0x23FE, 0x2208, 0x4248, 0x7A6A, 0xCA4A, 0x4BFA, 0x4A4C, 0x4A4C, 0x4AEC, 0x7B5A, 0x444A, 0x4D6, 0x822}},
    {30954, 2, {0x20, 0x222, 0xFBFE, 0x2090, 0x2088, 0x41FE, 0x7B10, 0x4D10, 0xC9FE, 0x4910, 0x4910, 0x49FE, 0x7910, 0x4910, 0x1FE, 0x100}},
    {30955, 2, {0x124, 0x124, 0xF224, 0x2424, 0x2954, 0x414A, 0x7292, 0xD610, 0x5A10, 0x5250, 0x525C, 0x5250, 0x7250, 0x52B0, 0x29E, 0x300}},
    {30956, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30957, 2, {0x3FE, 0x200, 0xFAFC, 0x2200, 0x23FE, 0x4294, 0x7AA8, 0xCAC4, 0x4A00, 0x4AFC, 0x4A84, 0x4AFC, 0x7A84, 0x44FC, 0x484, 0x88C}},
    {30958, 2, {0x20, 0x124, 0xFD24, 0x11FC, 0x1020, 0x2050, 0x3C88, 0x6574, 0x6602, 0xA5FC, 0x2554, 0x2554, 0x3DFC, 0x2554, 0x2154, 0x10C}},
    {30959, 2, {0x124, 0x124, 0xFAAA, 0x23AE, 0x2124, 0x42AA, 0x7BAE, 0x4924, 0xCFFE, 0x4910, 0x4914, 0x4914, 0x7A8A, 0x4A4A, 0x216, 0x422}},
    {30960, 2, {0x88, 0x50, 0xFBFE, 0x2020, 0x21FC, 0x4020, 0x7BFE, 0x4924, 0xC8A8, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x4904, 0x1FC, 0x104}},
    {30961, 2, {0x88, 0x2AA, 0xFADC, 0x2488, 0x2154, 0x4222, 0x7800, 0x4BFE, 0xCA42, 0x4C44, 0x4BFC, 0x4844, 0x7844, 0x4884, 0x114, 0x208}},
    {30962, 2, {0x200, 0x17E, 0xF840, 0x247C, 0x2244, 0x40FC, 0x7940, 0x4E7E, 0xCA20, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x4924, 0x222, 0x20}},
    {30963, 2, {0x104, 0x88, 0xF800, 0x23FE, 0x2222, 0x42AA, 0x7A72, 0x4A22, 0xCBFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x1FC, 0x104}},
    {30964, 2, {0x20, 0x7A4, 0xF0A8, 0x2292, 0x2114, 0x4208, 0x75F4, 0xD802, 0x53F8, 0x5208, 0x5208, 0x53F8, 0x7208, 0x5110, 0xFFE, 0x0}},
    {30965, 2, {0x0, 0x3DE, 0xFA52, 0x23DE, 0x2252, 0x43DE, 0x7A02, 0x4AFA, 0xCA8A, 0x4AFA, 0x4A8A, 0x4AFA, 0x7A8A, 0x4A02, 0x20A, 0x204}},
    {30966, 2, {0x80, 0x31C, 0xFA04, 0x2204, 0x239C, 0x4204, 0x7A04, 0xCBFC, 0x4900, 0x4BFE, 0x4C02, 0x4AAA, 0x7AAA, 0x4402, 0x14, 0x8}},
    {30967, 2, {0x20, 0x124, 0xF8A8, 0x23FE, 0x20A8, 0x4124, 0x7A22, 0x4904, 0xC904, 0x49DE, 0x4A44, 0x4D54, 0x789E, 0x4904, 0x204, 0x404}},
    {30968, 2, {0x88, 0x50, 0xFBFE, 0x2050, 0x21FC, 0x4154, 0x798C, 0x4974, 0xC904, 0x49FC, 0x4820, 0x4BFE, 0x7850, 0x4888, 0x104, 0x202}},
    {30969, 2, {0x0, 0x3FE, 0xFC50, 0x13FE, 0x1252, 0x23FE, 0x3C00, 0x65FC, 0x6504, 0xA5FC, 0x2504, 0x25FC, 0x3C20, 0x23FE, 0x20, 0x20}},
    {30970, 2, {0x88, 0x88, 0xFDFC, 0x1088, 0x1088, 0x23FE, 0x3C20, 0x65FC, 0x6524, 0xA5FC, 0x2524, 0x25FC, 0x3C00, 0x2488, 0x2104, 0x202}},
    {30971, 2, {0x3C, 0x3E0, 0xFD24, 0x10A8, 0x13FE, 0x20A8, 0x3D24, 0x6602, 0x65FC, 0xA524, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x21FC, 0x104}},
    {30972, 2, {0xA0, 0x90, 0xFDFE, 0x1320, 0x11FC, 0x2120, 0x3DFC, 0x6520, 0x65FE, 0xA500, 0x2420, 0x27FE, 0x3C70, 0x24A8, 0x2326, 0x20}},
    {30973, 2, {0x20, 0x20, 0xF9FC, 0x2020, 0x23FE, 0x4108, 0x7B9C, 0x4908, 0xC988, 0x4E3E, 0x4800, 0x4BFE, 0x7890, 0x4890, 0x112, 0x20E}},
    {30974, 2, {0x0, 0x3DE, 0xFA52, 0x2252, 0x23DE, 0x4000, 0x79FC, 0x4924, 0xC9FC, 0x4924, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20, 0x20}},
    {30975, 2, {0x3FFE, 0x2104, 0x2E38, 0x2210, 0x2FBC, 0x2718, 0x2AB4, 0x3252, 0x2000, 0x2FFC, 0x2100, 0x23F8, 0x2608, 0x5A08, 0x43F8, 0x8208}},
    {30976, 2, {0x3DE, 0x252, 0xFBDE, 0x2252, 0x23DE, 0x4202, 0x7AF2, 0x4A92, 0xCAF2, 0x4A92, 0x4AF2, 0x4A92, 0x7A92, 0x4B32, 0x202, 0x206}},
    {30977, 2, {0xA0, 0x90, 0xF9FE, 0x2110, 0x2310, 0x45FE, 0x7910, 0x4910, 0xC9FE, 0x4910, 0x4910, 0x49FE, 0x7900, 0x4AA4, 0x252, 0x452}},
    {30978, 2, {0x20, 0x3FE, 0xFC20, 0x11FC, 0x1000, 0x21FC, 0x3D04, 0x65FC, 0x6488, 0xA7FE, 0x2400, 0x25FC, 0x3D04, 0x2504, 0x21FC, 0x104}},
    {30979, 2, {0x124, 0xA8, 0xFBFE, 0x2202, 0x20F8, 0x4088, 0x78F8, 0x480C, 0xC9F0, 0x4820, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0xA0, 0x40}},
    {30980, 2, {0x18, 0x1E0, 0xFC40, 0x13FE, 0x1088, 0x2174, 0x3E52, 0x6470, 0x6400, 0xA5FC, 0x2504, 0x2574, 0x3D54, 0x2574, 0x2104, 0x10C}},
    {30981, 2, {0x208, 0x108, 0xF7C8, 0x2010, 0x279E, 0x4494, 0x77A4, 0xD014, 0x5794, 0x5094, 0x5114, 0x51C8, 0x7708, 0x5114, 0x524, 0x242}},
    {30982, 2, {0x20, 0x20, 0xFC50, 0x1088, 0x1104, 0x22FA, 0x3C00, 0x6400, 0x65DC, 0xA554, 0x2554, 0x25DC, 0x3C88, 0x2488, 0x2154, 0x222}},
    {30983, 2, {0x40, 0x80, 0xFBFE, 0x2222, 0x22AA, 0x4222, 0x7BFE, 0x4A72, 0xCAAA, 0x4A02, 0x4820, 0x4BFE, 0x7850, 0x4888, 0x104, 0x602}},
    {30984, 2, {0x44, 0x224, 0xF928, 0x207E, 0x2010, 0x4028, 0x7B4A, 0x491C, 0xC928, 0x494C, 0x491A, 0x492A, 0x7948, 0x4910, 0x2FE, 0x400}},
    {30985, 2, {0x108, 0x208, 0xF7C8, 0x2450, 0x27DE, 0x4464, 0x77D4, 0xD214, 0x5114, 0x57D4, 0x5214, 0x53C8, 0x7248, 0x5454, 0x554, 0x8A2}},
    {30986, 2, {0x878, 0xFFC8, 0x848, 0x3E86, 0x2B78, 0x3E48, 0x2A28, 0xFF90, 0x4928, 0x7F46, 0x0, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {30987, 2, {0x0, 0x1FC, 0xFD54, 0x11FC, 0x1020, 0x21FC, 0x3C20, 0x67FE, 0x6488, 0xA450, 0x25FC, 0x2420, 0x3FFE, 0x2420, 0x2020, 0x20}},
    {30988, 2, {0x0, 0x1FC, 0xFC20, 0x13FE, 0x1222, 0x21AC, 0x3C20, 0x65AC, 0x6400, 0xA5FC, 0x2524, 0x2524, 0x3DFC, 0x2524, 0x2124, 0x1FC}},
    {30989, 2, {0x88, 0x3FE, 0xFC88, 0x11FC, 0x1104, 0x21FC, 0x3D04, 0x65FC, 0x6480, 0xA5FE, 0x2622, 0x2552, 0x3D02, 0x25FA, 0x200A, 0x4}},
    {30990, 2, {0x88, 0x88, 0xFBDE, 0x2088, 0x21DC, 0x42AA, 0x7C88, 0x4800, 0xCBFE, 0x4822, 0x4920, 0x493C, 0x7920, 0x4AA0, 0x47E, 0x800}},
    {30991, 2, {0x50, 0x252, 0xF954, 0x2050, 0x23FE, 0x4088, 0x7850, 0x4BFE, 0xC820, 0x49FC, 0x4820, 0x4BFE, 0x78A8, 0x4924, 0x622, 0x20}},
    {30992, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4002, 0xBFF4, 0x400, 0xFF0, 0x1810, 0x6810, 0xFF0, 0x810}},
    {30993, 2, {0x20, 0x124, 0xFCA8, 0x13FE, 0x1202, 0x20F8, 0x3C88, 0x6488, 0x64F8, 0xA400, 0x25FC, 0x2524, 0x3DFC, 0x2524, 0x21FC, 0x104}},
    {30994, 2, {0x104, 0x88, 0xFBFE, 0x2020, 0x21FC, 0x4020, 0x7BFE, 0x4854, 0xC992, 0x4890, 0x4BFE, 0x4890, 0x78D4, 0x4B8A, 0x96, 0x1A2}},
    {30995, 2, {0x0, 0x3FE, 0xFC00, 0x11FC, 0x1124, 0x21FC, 0x3D24, 0x67FE, 0x6400, 0xA5FC, 0x2524, 0x25FC, 0x3D24, 0x25FC, 0x2000, 0x3FE}},
    {30996, 2, {0x10, 0x3C8, 0xFA7E, 0x2240, 0x2262, 0x43D4, 0x7A00, 0x4A3E, 0xCBC8, 0x4A48, 0x4E7E, 0x4A48, 0x7A48, 0x4BC8, 0x248, 0x8}},
    {30997, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x0, 0xFFFE, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {30998, 2, {0x140, 0x24C, 0xFA64, 0x2244, 0x234C, 0x4264, 0x7AA4, 0x4BAC, 0xCAA4, 0x4AA4, 0x4FFE, 0x4800, 0x7910, 0x4908, 0x204, 0x404}},
    {30999, 2, {0x40, 0x20, 0xFBFE, 0x2202, 0x21FC, 0x4148, 0x7A50, 0x49FC, 0xCB04, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x88, 0x104}},
    {31000, 2, {0x0, 0x3FE, 0xFA02, 0x23FE, 0x2202, 0x43FE, 0x7908, 0x4A52, 0xCB9C, 0x4908, 0x4A52, 0x4BDE, 0x7800, 0x4AA4, 0x252, 0x452}},
    {31001, 2, {0x200, 0x2BE, 0xFB02, 0x2254, 0x21C8, 0x403E, 0x7A0A, 0x4BCA, 0xCD28, 0x4928, 0x4FEE, 0x4928, 0x7AA8, 0x4A58, 0x44E, 0x80}},
    {31002, 2, {0x88, 0x3FE, 0xF888, 0x2020, 0x21FC, 0x4020, 0x7BFE, 0x4840, 0xC888, 0x49FC, 0x4800, 0x49FC, 0x7954, 0x4954, 0x7FE, 0x0}},
    {31003, 2, {0x8, 0x3E8, 0xFA88, 0x23EE, 0x2228, 0x43F4, 0x7A82, 0x4BE2, 0xC800, 0x49FC, 0x4954, 0x4954, 0x7954, 0x4954, 0x7FE, 0x0}},
    {31004, 2, {0xA00, 0x33B8, 0x2208, 0x3BB8, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x1010, 0x3FF8, 0xC406, 0xFF0, 0x1810, 0x6810, 0xFF0, 0x810}},
    {31005, 2, {0x0, 0x1FC, 0xFC20, 0x13FE, 0x1222, 0x21AC, 0x3C20, 0x65AC, 0x6400, 0xA7FE, 0x2420, 0x25FC, 0x3D54, 0x2554, 0x2154, 0x10C}},
    {31006, 2, {0x88, 0x3FE, 0xF888, 0x23FE, 0x2202, 0x41FC, 0x7800, 0x4BFE, 0xC840, 0x48A2, 0x4B54, 0x48B8, 0x7B54, 0x4892, 0x350, 0x20}},
    {31007, 2, {0x0, 0x7C4, 0xF514, 0x27C8, 0x2508, 0x47D4, 0x7514, 0xD500, 0x57D4, 0x5054, 0x5554, 0x5548, 0x7548, 0x5454, 0x54, 0x1A2}},
    {31008, 2, {0x108, 0x90, 0xFBFE, 0x2108, 0x2108, 0x4252, 0x7B9C, 0x4908, 0xCA52, 0x4BDE, 0x4840, 0x48A4, 0x7AAA, 0x4A8A, 0x478, 0x0}},
    {31009, 2, {0x88, 0x3FE, 0xFC88, 0x1024, 0x13FE, 0x2020, 0x3DFC, 0x6524, 0x65FC, 0xA524, 0x25FC, 0x2524, 0x3C08, 0x25FE, 0x2088, 0x58}},
    {31010, 2, {0x108, 0x90, 0xFBFC, 0x2040, 0x21F8, 0x4080, 0x7BFC, 0x4948, 0xC9FC, 0x4B0A, 0x4DF8, 0x4908, 0x79FA, 0x4934, 0x18C, 0x102}},
    {31011, 2, {0x48, 0x1FE, 0xFC48, 0x1000, 0x11FE, 0x214A, 0x3D4A, 0x65FE, 0x6414, 0xA5FE, 0x2510, 0x2592, 0x3D4C, 0x254A, 0x2116, 0x222}},
    {31012, 2, {0x88, 0x88, 0xFBFE, 0x2088, 0x2110, 0x41DE, 0x7A52, 0x4D54, 0xCA88, 0x4974, 0x4A02, 0x49FC, 0x7820, 0x4924, 0x222, 0x60}},
    {31013, 2, {0x7DE, 0x512, 0xF7D2, 0x2454, 0x27C8, 0x4514, 0x77E2, 0xD000, 0x53FC, 0x5204, 0x53FC, 0x5204, 0x73FC, 0x5204, 0x3FC, 0x108}},
    {31014, 2, {0x20, 0x10, 0xFBFE, 0x2244, 0x22FE, 0x4244, 0x7BFE, 0x4A10, 0xCAFE, 0x4A92, 0x4AFE, 0x4A92, 0x7AFE, 0x4A00, 0x244, 0x482}},
    {31015, 2, {0x0, 0x1FC, 0xF924, 0x21FC, 0x2124, 0x41FC, 0x7904, 0x4800, 0xCBFE, 0x4AAA, 0x4AAA, 0x4BFE, 0x7AAA, 0x4AAA, 0x3FE, 0x222}},
    {31016, 2, {0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x0, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {31017, 2, {0x42, 0x39C, 0xFA10, 0x23DE, 0x2294, 0x4294, 0x7C20, 0x4BFC, 0xCA04, 0x4BFC, 0x4A04, 0x4BFC, 0x7A04, 0x4BFC, 0x108, 0x204}},
    {31018, 2, {0x0, 0x1FE, 0xFD28, 0x11FE, 0x1128, 0x217C, 0x3D54, 0x657C, 0x6554, 0xA57C, 0x2510, 0x25FE, 0x3D92, 0x22BA, 0x28A, 0x484}},
    {31019, 2, {0x124, 0x174, 0xFA58, 0x22DA, 0x23FE, 0x4154, 0x7ADA, 0x4BFE, 0xC8A2, 0x4820, 0x4BFE, 0x4870, 0x78A8, 0x4924, 0x222, 0x20}},
    {31020, 2, {0x2288, 0x2108, 0xFABE, 0x2008, 0x729C, 0xA92A, 0x2288, 0xFFFE, 0x820, 0x3FF8, 0xC406, 0xFF0, 0x1810, 0x6810, 0xFF0, 0x810}},
    {31021, 2, {0x1FC, 0x20, 0xFBFE, 0x2222, 0x25AC, 0x4020, 0x79AC, 0x4890, 0xC9FE, 0x4B20, 0x4DFC, 0x4920, 0x79FC, 0x4920, 0x1FE, 0x100}},
    {31022, 2, {0x10, 0x7C8, 0xF53E, 0x27C0, 0x2514, 0x47D2, 0x7522, 0xD500, 0x57D4, 0x5054, 0x5554, 0x5548, 0x7548, 0x5454, 0x54, 0x1A2}},
    {31023, 2, {0x88, 0x2AA, 0xFADC, 0x2488, 0x2154, 0x4222, 0x7BFE, 0x4A42, 0xC888, 0x49F0, 0x4824, 0x4BFE, 0x7822, 0x4924, 0x2A2, 0x40}},
    {31024, 2, {0x0, 0x7FE, 0xF444, 0x2598, 0x2488, 0x47DE, 0x7488, 0xD5DC, 0x56AA, 0x5488, 0x5420, 0x5420, 0x74BC, 0x54A0, 0x4A0, 0xBFE}},
    {31025, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E, 0x0, 0x7FFC, 0x800, 0x1FF8, 0x6808, 0xFF8}},
    {31026, 2, {0x110, 0x90, 0xFBDE, 0x2010, 0x225E, 0x4182, 0x7BDE, 0x4810, 0xCBDE, 0x4A50, 0x4BDE, 0x4A50, 0x7BDE, 0x4A50, 0x252, 0x2CE}},
    {31027, 2, {0x40, 0x20, 0xFBFE, 0x2248, 0x23FE, 0x4248, 0x7AEC, 0x4B5A, 0xCA48, 0x4BFE, 0x4A20, 0x4A7C, 0x7AC4, 0x4B44, 0x47C, 0x844}},
    {31028, 2, {0x88, 0x3FE, 0xF088, 0x2412, 0x22FE, 0x4010, 0x74FE, 0xD292, 0x52FE, 0x5092, 0x51FE, 0x5292, 0x7604, 0x52FE, 0x244, 0x22C}},
    {31029, 2, {0x1FC, 0x20, 0xFBFE, 0x2222, 0x21AC, 0x4020, 0x79AC, 0x4800, 0xC89E, 0x4BF2, 0x489E, 0x49D2, 0x7ABE, 0x4C92, 0x9E, 0x92}},
    {31030, 2, {0x88, 0x3FE, 0xF888, 0x23DE, 0x2252, 0x43DE, 0x78A0, 0x4890, 0xC9FE, 0x4920, 0x4BFC, 0x4D20, 0x79FC, 0x4920, 0x1FE, 0x100}},
    {31031, 2, {0x90, 0x7FE, 0xF890, 0x23E8, 0x2288, 0x43CE, 0x7A50, 0x4BD4, 0xCAA2, 0x4BE2, 0x4800, 0x4BFC, 0x7A94, 0x4A94, 0x7FE, 0x0}},
    {31032, 2, {0x528, 0x7BE, 0xF948, 0x27BE, 0x2318, 0x45AA, 0x7946, 0x4BFC, 0xCA04, 0x4BFC, 0x4A04, 0x4BFC, 0x7A04, 0x4BFC, 0x108, 0x204}},
    {31033, 2, {0x7BC, 0x4A4, 0xF7BC, 0x2000, 0x27FE, 0x4400, 0x7590, 0xD49E, 0x57D4, 0x5564, 0x55D4, 0x5554, 0x75D4, 0x5568, 0xBC8, 0x1054}},
    {31034, 2, {0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x100, 0x100, 0x1110, 0x1108, 0x2104, 0x4102, 0x8102, 0x500, 0x200}},
    {31035, 2, {0x2000, 0x1000, 0x1000, 0xF800, 0x800, 0x1000, 0x1000, 0x3800, 0x5400, 0x9400, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {31036, 2, {0x2080, 0x1080, 0x1080, 0xF880, 0x880, 0x1080, 0x1080, 0x3880, 0x5480, 0x9480, 0x1080, 0x1084, 0x1084, 0x1084, 0x107C, 0x1000}},
    {31037, 2, {0x2000, 0x1000, 0x13F8, 0xF888, 0x890, 0x1090, 0x10A0, 0x38BC, 0x5484, 0x9484, 0x1104, 0x1104, 0x1104, 0x1204, 0x1228, 0x1410}},
    {31038, 2, {0x2020, 0x1020, 0x1020, 0xF820, 0x820, 0x1020, 0x13FE, 0x3820, 0x5420, 0x9420, 0x1020, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {31039, 2, {0x2040, 0x1040, 0x1080, 0xF8FC, 0x904, 0x1204, 0x1004, 0x3904, 0x5484, 0x9444, 0x1044, 0x1004, 0x1004, 0x1004, 0x1028, 0x1010}},
    {31040, 2, {0x2000, 0x1000, 0x11FC, 0xF904, 0x904, 0x1104, 0x1104, 0x39FC, 0x5504, 0x9500, 0x1100, 0x1102, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {31041, 2, {0x2000, 0x10FC, 0x1084, 0xFC88, 0x488, 0x890, 0x1088, 0x3888, 0x5484, 0x9284, 0x10C4, 0x10A8, 0x1090, 0x1080, 0x1080, 0x1080}},
    {31042, 2, {0x2020, 0x1020, 0x1020, 0xF920, 0x92C, 0x1134, 0x1164, 0x3BA4, 0x5524, 0x9534, 0x1128, 0x1122, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {31043, 2, {0x2000, 0x13F0, 0x1010, 0xF810, 0x910, 0x1110, 0x1110, 0x39FC, 0x5404, 0x9404, 0x1004, 0x13F4, 0x1004, 0x1004, 0x1028, 0x1010}},
    {31044, 2, {0x2020, 0x1020, 0x1050, 0xF850, 0x888, 0x1104, 0x1202, 0x3888, 0x5488, 0x9488, 0x1088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1208}},
    {31045, 2, {0x2008, 0x103C, 0x11E0, 0xF820, 0x820, 0x1020, 0x13FE, 0x3820, 0x5420, 0x9450, 0x1050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {31046, 2, {0x2000, 0x1000, 0x11FC, 0xF820, 0x820, 0x1020, 0x1020, 0x3BFE, 0x5420, 0x9420, 0x1050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {31047, 2, {0x2004, 0x101E, 0x11F0, 0xF910, 0x910, 0x1110, 0x1110, 0x39FE, 0x5510, 0x9510, 0x1110, 0x1108, 0x110A, 0x114A, 0x1186, 0x1102}},
    {31048, 2, {0x2008, 0x101C, 0x11E0, 0xF900, 0x900, 0x1100, 0x11FE, 0x3910, 0x5510, 0x9510, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {31049, 2, {0x2020, 0x1020, 0x1020, 0xF920, 0x920, 0x1120, 0x113C, 0x3920, 0x5520, 0x9520, 0x1120, 0x1120, 0x1120, 0x1120, 0x17FE, 0x1000}},
    {31050, 2, {0x2040, 0x1020, 0x1020, 0xFBFE, 0x880, 0x1080, 0x1080, 0x38FC, 0x5484, 0x9484, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410}},
    {31051, 2, {0x2000, 0x10F8, 0x1088, 0xF888, 0x888, 0x1106, 0x1200, 0x39FC, 0x5484, 0x9484, 0x1048, 0x1050, 0x1020, 0x1050, 0x1088, 0x1306}},
    {31052, 2, {0x2020, 0x1020, 0x1020, 0xF820, 0x9FC, 0x1124, 0x1124, 0x3924, 0x5524, 0x95FC, 0x1124, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31053, 2, {0x2008, 0x103C, 0x11E0, 0xF820, 0x820, 0x1020, 0x1020, 0x3BFE, 0x5420, 0x9420, 0x1020, 0x1020, 0x1020, 0x1020, 0x11FC, 0x1000}},
    {31054, 2, {0x2020, 0x1020, 0x1020, 0xFBFE, 0x820, 0x1020, 0x19FE, 0x3420, 0x5020, 0x93FE, 0x1022, 0x1022, 0x102A, 0x1024, 0x1020, 0x1020}},
    {31055, 2, {0x2000, 0x1000, 0x13FE, 0xF820, 0x820, 0x1040, 0x1840, 0x34FC, 0x5184, 0x9284, 0x1484, 0x1084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {31056, 2, {0x2020, 0x1020, 0x1020, 0xF820, 0xBFE, 0x1040, 0x1040, 0x3880, 0x54FC, 0x9184, 0x1284, 0x1084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {31057, 2, {0x2020, 0x1020, 0x1120, 0xF920, 0x9FC, 0x1120, 0x1220, 0x3820, 0x57FE, 0x9420, 0x1050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {31058, 2, {0x2000, 0x13FC, 0x1084, 0xF884, 0x884, 0x1104, 0x1114, 0x3A08, 0x5400, 0x95FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31059, 2, {0x2050, 0x1048, 0x1048, 0xF840, 0xBFE, 0x1080, 0x1080, 0x38FC, 0x5544, 0x9544, 0x1128, 0x1128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {31060, 2, {0x2088, 0x1088, 0x1088, 0xF908, 0x97E, 0x1308, 0x1508, 0x3948, 0x5528, 0x9528, 0x1108, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {31061, 2, {0x2000, 0x1040, 0x1020, 0xF828, 0x808, 0x1088, 0x1090, 0x3A94, 0x56A2, 0x92A2, 0x14C2, 0x1088, 0x1188, 0x1288, 0x1478, 0x1000}},
    {31062, 2, {0x2000, 0x11F8, 0x1108, 0xF908, 0x908, 0x11F8, 0x1108, 0x3908, 0x5508, 0x95F8, 0x1108, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {31063, 2, {0x4008, 0x203C, 0x23E0, 0xFA20, 0x1220, 0x1220, 0x2220, 0x33FE, 0x6A20, 0xAA10, 0x2210, 0x2212, 0x220A, 0x228A, 0x2326, 0x2212}},
    {31064, 2, {0x0, 0x3E7C, 0x0, 0x0, 0x0, 0x7EFE, 0x810, 0x810, 0x2A54, 0x2952, 0x4952, 0x8890, 0x810, 0x810, 0x2850, 0x1020}},
    {31065, 2, {0x2020, 0x1020, 0x1020, 0xF9FC, 0x820, 0x1020, 0x1020, 0x3BFE, 0x5470, 0x94A8, 0x10A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31066, 2, {0x2080, 0x1080, 0x1080, 0xF8FE, 0x940, 0x1140, 0x1240, 0x387C, 0x5440, 0x9440, 0x1040, 0x107E, 0x1040, 0x1040, 0x1040, 0x1040}},
    {31067, 2, {0x2020, 0x1020, 0x1020, 0xF9FC, 0x820, 0x1020, 0x1020, 0x3BFE, 0x5420, 0x9420, 0x1040, 0x1048, 0x1084, 0x11FE, 0x1082, 0x1000}},
    {31068, 2, {0x2020, 0x1020, 0x1020, 0xF820, 0xBFE, 0x1020, 0x1020, 0x3820, 0x55FC, 0x9504, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31069, 2, {0x2000, 0x13FC, 0x1204, 0xFA04, 0xA04, 0x1204, 0x13FC, 0x3890, 0x5490, 0x9490, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E, 0x1800}},
    {31070, 2, {0x2020, 0x1020, 0x1020, 0xF9FC, 0x924, 0x1124, 0x1124, 0x39FC, 0x5524, 0x9524, 0x1124, 0x11FC, 0x1124, 0x1020, 0x1020, 0x1020}},
    {31071, 2, {0x100, 0x1110, 0x1110, 0x1FF0, 0x4104, 0x4104, 0x7FFC, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {31072, 2, {0x2000, 0x11FC, 0x1004, 0xF804, 0xBF4, 0x1004, 0x1004, 0x39E4, 0x5524, 0x9524, 0x1124, 0x11E4, 0x1004, 0x1004, 0x1014, 0x1008}},
    {31073, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {31074, 2, {0x2080, 0x1080, 0x1080, 0xF8FE, 0x902, 0x1104, 0x1220, 0x3820, 0x54A8, 0x94A4, 0x1124, 0x1122, 0x1222, 0x1020, 0x10A0, 0x1040}},
    {31075, 2, {0x2080, 0x1080, 0x10FE, 0xF900, 0xA00, 0x101C, 0x11E0, 0x3920, 0x5522, 0x9524, 0x1128, 0x1110, 0x1110, 0x1148, 0x1184, 0x1102}},
    {31076, 2, {0x2000, 0x13DE, 0x1042, 0xF842, 0xA52, 0x114A, 0x114A, 0x3842, 0x54C6, 0x914A, 0x1252, 0x1042, 0x1042, 0x1042, 0x114A, 0x1084}},
    {31077, 2, {0x2104, 0x1084, 0x1088, 0xF800, 0xBFE, 0x1020, 0x1020, 0x39FC, 0x5420, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31078, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x9FC, 0x1000, 0x1000, 0x39FC, 0x5420, 0x9020, 0x13FE, 0x1020, 0x1050, 0x1088, 0x1104, 0x1202}},
    {31079, 2, {0x2090, 0x1090, 0x1090, 0xFA92, 0x994, 0x1098, 0x1890, 0x3598, 0x5294, 0x9492, 0x1090, 0x1090, 0x1112, 0x1112, 0x1212, 0x140E}},
    {31080, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {31081, 2, {0x2020, 0x1120, 0x1120, 0xF9FC, 0x920, 0x1220, 0x1020, 0x3BFE, 0x5470, 0x94A8, 0x10A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31082, 2, {0x2040, 0x1040, 0x10FC, 0xF904, 0xA08, 0x11FE, 0x1100, 0x397C, 0x5544, 0x9544, 0x1154, 0x1148, 0x1142, 0x1242, 0x123E, 0x1400}},
    {31083, 2, {0x2020, 0x1020, 0x1050, 0xF888, 0x904, 0x1202, 0x11FC, 0x3800, 0x5400, 0x95FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31084, 2, {0x2000, 0x13FE, 0x1020, 0xF840, 0x888, 0x1104, 0x13FE, 0x3822, 0x5420, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1000}},
    {31085, 2, {0x1080, 0x1080, 0x3EFC, 0x6244, 0x9448, 0x4830, 0x3010, 0x27C8, 0xC006, 0x0, 0x1FF0, 0x100, 0x920, 0x1110, 0x2508, 0x200}},
    {31086, 2, {0x2020, 0x1020, 0x1020, 0xFBFE, 0x820, 0x1020, 0x11FC, 0x3800, 0x5400, 0x95FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31087, 2, {0x2020, 0x1020, 0x103E, 0xF820, 0x820, 0x11FC, 0x1104, 0x3924, 0x5524, 0x9524, 0x1124, 0x1124, 0x1050, 0x1048, 0x1084, 0x1104}},
    {31088, 2, {0x2020, 0x1120, 0x1120, 0xF9FC, 0x920, 0x1220, 0x1020, 0x3BFE, 0x5400, 0x9400, 0x11FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31089, 2, {0x2104, 0x1084, 0x1088, 0xF810, 0x9FC, 0x1104, 0x1104, 0x3904, 0x55FC, 0x9450, 0x1050, 0x1090, 0x1092, 0x1112, 0x120E, 0x1400}},
    {31090, 2, {0x2000, 0x11FC, 0x1004, 0xF8FC, 0x804, 0x11FC, 0x1000, 0x3BFE, 0x5602, 0x91F8, 0x1088, 0x1088, 0x1050, 0x1020, 0x10D8, 0x1306}},
    {31091, 2, {0x2000, 0x13FC, 0x1200, 0xFA00, 0xAF8, 0x1200, 0x1200, 0x3BFC, 0x56A0, 0x92A4, 0x12A8, 0x1290, 0x1290, 0x1488, 0x14A4, 0x18C2}},
    {31092, 2, {0x2010, 0x1014, 0x1012, 0xF810, 0xBFE, 0x1010, 0x1150, 0x3954, 0x57F4, 0x9554, 0x1154, 0x1148, 0x124A, 0x125A, 0x1426, 0x1042}},
    {31093, 2, {0x4000, 0x23FE, 0x2222, 0xFA22, 0x1222, 0x13FE, 0x2222, 0x3262, 0x6A72, 0xAAAA, 0x2322, 0x2222, 0x2222, 0x2202, 0x23FE, 0x2202}},
    {31094, 2, {0x4108, 0x2088, 0x2090, 0xFBFC, 0x1024, 0x1024, 0x23FC, 0x3220, 0x6A20, 0xABFE, 0x2062, 0x20A2, 0x212A, 0x2224, 0x2420, 0x2020}},
    {31095, 2, {0x2020, 0x1020, 0x13FE, 0xF820, 0x9FC, 0x1040, 0x13FE, 0x3848, 0x5488, 0x94FE, 0x1108, 0x1148, 0x1228, 0x1408, 0x1028, 0x1010}},
    {31096, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x904, 0x11FC, 0x1820, 0x3420, 0x53FE, 0x9222, 0x1252, 0x128A, 0x130A, 0x1202, 0x120A, 0x1204}},
    {31097, 2, {0x2080, 0x1080, 0x10FE, 0xF902, 0xA82, 0x10F2, 0x1142, 0x3842, 0x57FA, 0x9442, 0x1152, 0x1152, 0x11F2, 0x1002, 0x1014, 0x1008}},
    {31098, 2, {0x2108, 0x1108, 0x13FC, 0xF908, 0x908, 0x11F8, 0x1108, 0x3908, 0x55F8, 0x9508, 0x1108, 0x17FE, 0x1000, 0x1090, 0x1108, 0x1204}},
    {31099, 2, {0x2000, 0x11FC, 0x1104, 0xF924, 0x924, 0x11FC, 0x1124, 0x3924, 0x5574, 0x9554, 0x1154, 0x1174, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31100, 2, {0x2000, 0x11FC, 0x1124, 0xF924, 0x9FC, 0x1124, 0x1924, 0x35FC, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31101, 2, {0x2040, 0x1020, 0x13FE, 0xF800, 0x888, 0x1088, 0x1154, 0x3A22, 0x5400, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31102, 2, {0x2020, 0x1020, 0x11FC, 0xF820, 0x820, 0x13FE, 0x1888, 0x3544, 0x5242, 0x90F8, 0x1188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {31103, 2, {0x2080, 0x10F8, 0x1088, 0xF908, 0x9F0, 0x1010, 0x1810, 0x37FE, 0x5020, 0x9222, 0x1174, 0x10A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {31104, 2, {0x100, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1110, 0x2308}},
    {31105, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC826, 0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x100, 0x1110, 0x2508, 0x4204}},
    {31106, 2, {0x2000, 0x11FC, 0x1124, 0xF924, 0x974, 0x1124, 0x11FC, 0x3904, 0x5574, 0x9554, 0x1154, 0x1174, 0x1104, 0x1204, 0x1214, 0x1408}},
    {31107, 2, {0x4040, 0x2040, 0x27FC, 0xF840, 0x1040, 0x13F8, 0x2208, 0x33F8, 0x6A08, 0xABF8, 0x2208, 0x23F8, 0x2208, 0x2208, 0x2FFE, 0x2000}},
    {31108, 2, {0x4000, 0x21F8, 0x2008, 0xF808, 0x11F8, 0x1008, 0x2008, 0x33FE, 0x6820, 0xAA22, 0x2174, 0x20A8, 0x2124, 0x2222, 0x20A0, 0x2040}},
    {31109, 2, {0x2104, 0x1088, 0x1050, 0xF9FC, 0x924, 0x1124, 0x11FC, 0x3924, 0x5524, 0x95FC, 0x1020, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {31110, 2, {0x2040, 0x1080, 0x13FC, 0xFA24, 0xA24, 0x13FC, 0x1A24, 0x3644, 0x53FC, 0x9090, 0x1110, 0x17FE, 0x1010, 0x1010, 0x1010, 0x1010}},
    {31111, 2, {0x2040, 0x1044, 0x13F4, 0xF848, 0x850, 0x17FE, 0x1040, 0x3880, 0x55F8, 0x9308, 0x1508, 0x11F8, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {31112, 2, {0x2000, 0x13FE, 0x1222, 0xF820, 0xBFE, 0x1020, 0x11FC, 0x3924, 0x55FC, 0x9524, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {31113, 2, {0x2088, 0x1048, 0x1050, 0xFBFE, 0x850, 0x1050, 0x11FC, 0x3954, 0x5554, 0x958C, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31114, 2, {0x4100, 0x213C, 0x27D4, 0xF914, 0x1394, 0x1114, 0x27D4, 0x3124, 0x694C, 0xA840, 0x27FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {31115, 2, {0x2000, 0x13FE, 0x1050, 0xF850, 0x9FC, 0x1154, 0x1154, 0x3954, 0x55FC, 0x9420, 0x1020, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {31116, 2, {0x4084, 0x2044, 0x2048, 0xFBFE, 0x1084, 0x1084, 0x2108, 0x314A, 0x6A52, 0xAB9C, 0x2084, 0x2108, 0x2108, 0x2252, 0x23DE, 0x2042}},
    {31117, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x9E4, 0x1124, 0x1124, 0x3BFE, 0x5602, 0x92FA, 0x128A, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {31118, 2, {0x2020, 0x1020, 0x103E, 0xF820, 0x820, 0x11FC, 0x1104, 0x39FC, 0x5504, 0x95FC, 0x1104, 0x11FC, 0x1000, 0x1088, 0x1104, 0x1202}},
    {31119, 2, {0x4000, 0x23FE, 0x2000, 0xF9FC, 0x1104, 0x1104, 0x21FC, 0x3000, 0x6BFE, 0xAA22, 0x2222, 0x23FE, 0x2222, 0x2222, 0x23FE, 0x2202}},
    {31120, 2, {0x203C, 0x17C0, 0x1244, 0xF928, 0x800, 0x13FC, 0x1080, 0x3880, 0x57FE, 0x9500, 0x11F8, 0x1288, 0x1250, 0x1420, 0x18D8, 0x1306}},
    {31121, 2, {0x21FC, 0x1124, 0x1124, 0xF9FC, 0x924, 0x1124, 0x19FC, 0x3420, 0x5020, 0x93FE, 0x1222, 0x122A, 0x12FA, 0x120A, 0x1202, 0x1206}},
    {31122, 2, {0x2100, 0x11F8, 0x1208, 0xFBF0, 0x810, 0x17FE, 0x1080, 0x3944, 0x5668, 0x94B0, 0x1128, 0x1668, 0x10A4, 0x1122, 0x16A0, 0x1040}},
    {31123, 2, {0x2000, 0x11F8, 0x1108, 0xF9F8, 0x908, 0x11F8, 0x1000, 0x3BFE, 0x5500, 0x95FC, 0x1254, 0x1494, 0x1124, 0x1244, 0x10A8, 0x1110}},
    {31124, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x9FC, 0x1104, 0x1904, 0x35FC, 0x5000, 0x93FE, 0x1020, 0x1120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {31125, 2, {0x2040, 0x1040, 0x11FC, 0xF884, 0xBFE, 0x1000, 0x11FC, 0x3904, 0x55FC, 0x9020, 0x13FE, 0x1020, 0x1220, 0x13FE, 0x1020, 0x1020}},
    {31126, 2, {0x2088, 0x1088, 0x13FE, 0xF888, 0x888, 0x10F8, 0x1088, 0x3888, 0x54F8, 0x9020, 0x13FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {31127, 2, {0x2000, 0x11FC, 0x1124, 0xF924, 0x9FC, 0x1124, 0x1124, 0x39FC, 0x5400, 0x9440, 0x1024, 0x12A2, 0x128A, 0x1488, 0x1078, 0x1000}},
    {31128, 2, {0x2040, 0x1020, 0x11FC, 0xF800, 0x888, 0x1050, 0x13FE, 0x3A22, 0x5424, 0x95FC, 0x1124, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {31129, 2, {0x2050, 0x1052, 0x13DC, 0xF850, 0x8D2, 0x134E, 0x1000, 0x39FC, 0x5504, 0x95FC, 0x1104, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {31130, 2, {0x2104, 0x1084, 0x1088, 0xF800, 0xBFE, 0x1020, 0x1820, 0x35FC, 0x5020, 0x9020, 0x13FE, 0x1000, 0x12A4, 0x1252, 0x1452, 0x1000}},
    {31131, 2, {0x2020, 0x1020, 0x13FE, 0xF820, 0x9FC, 0x1104, 0x11FC, 0x3904, 0x55FC, 0x9504, 0x11FC, 0x1104, 0x13FE, 0x1088, 0x1104, 0x1202}},
    {31132, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x0, 0xFFFE, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {31133, 2, {0x2000, 0x11FC, 0x1124, 0xF9FC, 0x924, 0x11FC, 0x1888, 0x3544, 0x5242, 0x90F8, 0x1188, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {31134, 2, {0x2040, 0x1020, 0x13FE, 0xF800, 0x9FC, 0x1104, 0x19FC, 0x3400, 0x53FE, 0x9202, 0x12FA, 0x128A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {31135, 2, {0x2020, 0x1010, 0x11FE, 0xF910, 0x97C, 0x1114, 0x11FE, 0x3914, 0x557C, 0x9510, 0x117C, 0x1144, 0x1144, 0x1244, 0x127C, 0x1444}},
    {31136, 2, {0x4004, 0x201E, 0x23F0, 0xFA1E, 0x1210, 0x12FE, 0x2292, 0x3298, 0x6AF2, 0xAA8E, 0x2280, 0x22B8, 0x22A8, 0x24AA, 0x254A, 0x2A86}},
    {31137, 2, {0x2000, 0x11FE, 0x1120, 0xF9FC, 0x920, 0x11FC, 0x1120, 0x3920, 0x55FE, 0x9002, 0x12AA, 0x12AA, 0x12AA, 0x1202, 0x1014, 0x1008}},
    {31138, 2, {0x2000, 0x11FC, 0x1104, 0xF9FC, 0x904, 0x11FC, 0x1800, 0x3400, 0x53DE, 0x9042, 0x1252, 0x114A, 0x1252, 0x1042, 0x114A, 0x1084}},
    {31139, 2, {0x2028, 0x1024, 0x13FE, 0xF820, 0x9FC, 0x1124, 0x11FC, 0x3924, 0x55FC, 0x9524, 0x1008, 0x13FE, 0x1088, 0x1048, 0x1048, 0x1018}},
    {31140, 2, {0x2000, 0x13FE, 0x1252, 0xFA52, 0xBFE, 0x1000, 0x13DE, 0x3842, 0x5652, 0x954A, 0x10C6, 0x135A, 0x1042, 0x1042, 0x114A, 0x1084}},
    {31141, 2, {0x2088, 0x1088, 0x13FE, 0xF888, 0x8F8, 0x1088, 0x10F8, 0x3888, 0x57FE, 0x9488, 0x1124, 0x12FA, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {31142, 2, {0x200, 0x27DE, 0x4912, 0x97D2, 0x2112, 0x65DA, 0xA514, 0x2FD0, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x100, 0x1110, 0x2508, 0x200}},
    {31143, 2, {0x4040, 0x27FC, 0x2040, 0xFBF8, 0x1000, 0x13F8, 0x2208, 0x33F8, 0x6910, 0xAFFE, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {31144, 2, {0x4124, 0x2124, 0x22AA, 0xFBAE, 0x1124, 0x12AA, 0x23AE, 0x3124, 0x6FFE, 0xA910, 0x2114, 0x2114, 0x228A, 0x224A, 0x2216, 0x2422}},
    {31145, 2, {0x4000, 0x23FE, 0x2222, 0xFBFE, 0x1222, 0x13FE, 0x2088, 0x3088, 0x6BFE, 0xA888, 0x2088, 0x27FE, 0x2000, 0x2088, 0x2104, 0x2202}},
    {31146, 2, {0x2000, 0x13DE, 0x1252, 0xFA52, 0xBDE, 0x1000, 0x11FC, 0x3924, 0x55FC, 0x9524, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {31147, 2, {0x2000, 0x13FE, 0x1050, 0xFBFE, 0xA52, 0x13FE, 0x1000, 0x39FC, 0x5504, 0x95FC, 0x1104, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {31148, 2, {0x2020, 0x1050, 0x1088, 0xF974, 0xA02, 0x11FC, 0x1124, 0x39AC, 0x5524, 0x95FC, 0x1000, 0x10F8, 0x1088, 0x10F8, 0x1088, 0x10F8}},
    {31149, 2, {0x4088, 0x2448, 0x2250, 0xF8FC, 0x1020, 0x1050, 0x2694, 0x3238, 0x6A50, 0xAA98, 0x2234, 0x2254, 0x2290, 0x2220, 0x25FE, 0x2800}},
    {31150, 2, {0x4090, 0x23FC, 0x2294, 0xFBFC, 0x1294, 0x13FC, 0x2000, 0x33FC, 0x6800, 0xABFC, 0x2204, 0x23FC, 0x2108, 0x2090, 0x27FE, 0x2000}},
    {31151, 2, {0x4090, 0x23FC, 0x2294, 0xFBFC, 0x1294, 0x13FC, 0x2000, 0x33FC, 0x6A00, 0xAAF8, 0x2200, 0x23FE, 0x2520, 0x2514, 0x2548, 0x2986}},
    {31152, 2, {0x4000, 0x23FE, 0x2124, 0xFA22, 0x17FE, 0x1222, 0x2376, 0x32AA, 0x6B76, 0xAA22, 0x2376, 0x22AA, 0x2376, 0x2222, 0x222A, 0x2204}},
    {31153, 2, {0x4040, 0x27FC, 0x2040, 0xFBFC, 0x1000, 0x17FE, 0x2002, 0x33F8, 0x6840, 0xAFFE, 0x2000, 0x27FE, 0x2008, 0x27FE, 0x24A8, 0x2798}},
    {31154, 2, {0x2000, 0x13FE, 0x1248, 0xFBFE, 0xA48, 0x12FC, 0x12A4, 0x3AFC, 0x56A4, 0x92FC, 0x1220, 0x13FE, 0x132A, 0x157A, 0x150A, 0x1906}},
    {31155, 2, {0x2020, 0x13FE, 0x1000, 0xF9DC, 0x954, 0x11DC, 0x1888, 0x37FE, 0x5088, 0x93FE, 0x1088, 0x17FE, 0x1094, 0x1188, 0x16A4, 0x10C2}},
    {31156, 2, {0x4040, 0x20A0, 0x2110, 0x208, 0xF5F6, 0x1000, 0x2EEE, 0x2AAA, 0x6EEE, 0xB000, 0x2FFE, 0x2922, 0x2FFE, 0x2922, 0x292A, 0x2804}},
    {31157, 2, {0x4000, 0x23BE, 0x2288, 0xFB90, 0x12BE, 0x13A2, 0x203E, 0x37E2, 0x693E, 0xA922, 0x25BE, 0x2514, 0x2522, 0x2500, 0x2BFE, 0x3000}},
    {31158, 2, {0x4528, 0x27BE, 0x2948, 0xFFBE, 0x1318, 0x15AA, 0x2946, 0x33FC, 0x6A04, 0xABFC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {31159, 2, {0x4100, 0x255E, 0x2384, 0xF908, 0x17DE, 0x1112, 0x2392, 0x355E, 0x6912, 0xA81E, 0x2152, 0x2FF2, 0x211E, 0x2280, 0x244C, 0x2812}},
    {31160, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x4204, 0x4204, 0x4404, 0x4444, 0x4824, 0x5FF4, 0x4814, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {31161, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x4104, 0x4124, 0x5FF4, 0x4814, 0x4004, 0x400C}},
    {31162, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x7FFC, 0x4104, 0x4124, 0x5FE4, 0x4824, 0x4014, 0x4008}},
    {31163, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1450, 0x1390, 0x1450, 0x1FF0, 0x100, 0x7FFC, 0x4204, 0x4444, 0x4FE4, 0x4424, 0x4014, 0x4008}},
    {31164, 2, {0x100, 0x1FC, 0x100, 0x1FF0, 0x1450, 0x1390, 0x1450, 0x1FF0, 0x100, 0x7FFC, 0x4204, 0x4444, 0x4FE4, 0x4424, 0x4014, 0x4008}},
    {31165, 2, {0x100, 0x280, 0xC60, 0x3118, 0xCFE6, 0x440, 0x1390, 0x1450, 0x1FF0, 0x100, 0x7FFC, 0x4204, 0x4444, 0x4FE4, 0x4424, 0x400C}},
    {31166, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {31167, 2, {0x70, 0x1F80, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC006, 0x440, 0x440, 0x440, 0x440, 0x842, 0x1042, 0x603E}},
    {31168, 2, {0x70, 0x1F80, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xCFE6, 0x420, 0x440, 0x478, 0x808, 0x808, 0x1050, 0x6020}},
    {31169, 2, {0x420, 0xE20, 0x7820, 0x820, 0x820, 0xFF40, 0x840, 0x1840, 0x1C40, 0x2A48, 0x2A84, 0x4884, 0x89FE, 0x882, 0x802, 0x800}},
    {31170, 2, {0xF8, 0x3F00, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x1110, 0x2008, 0xC104, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31171, 2, {0x70, 0x1F80, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC006, 0xFC0, 0x840, 0x840, 0x844, 0x1044, 0x203C, 0x4000}},
    {31172, 2, {0x400, 0xEFC, 0x7804, 0x808, 0x810, 0xFF20, 0x820, 0x19FE, 0x1C20, 0x2A20, 0x2A20, 0x4820, 0x8820, 0x820, 0x8A0, 0x840}},
    {31173, 2, {0x808, 0x1C3C, 0xF3C0, 0x1040, 0x1040, 0xFC40, 0x1040, 0x307E, 0x3BC0, 0x5440, 0x5040, 0x9042, 0x1042, 0x1042, 0x103E, 0x1000}},
    {31174, 2, {0x800, 0x1DFC, 0xF020, 0x1020, 0x1020, 0xFC20, 0x1020, 0x33FE, 0x3820, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31175, 2, {0x840, 0x1C40, 0xF040, 0x1040, 0x11F8, 0xFC48, 0x1048, 0x3048, 0x3948, 0x54C8, 0x5448, 0x90A8, 0x10AA, 0x110A, 0x1206, 0x1402}},
    {31176, 2, {0x820, 0x1C20, 0xF020, 0x1020, 0x1124, 0xFD24, 0x1124, 0x3124, 0x3924, 0x5524, 0x5524, 0x9124, 0x1124, 0x11FC, 0x1004, 0x1000}},
    {31177, 2, {0xF8, 0x3F00, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31178, 2, {0x70, 0x1F80, 0x100, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC0E6, 0x1F00, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {31179, 2, {0x840, 0x1C40, 0xF040, 0x1040, 0x1144, 0xFD44, 0x1148, 0x3250, 0x3840, 0x54A0, 0x54A0, 0x9090, 0x1110, 0x1108, 0x1204, 0x1402}},
    {31180, 2, {0x1038, 0x11C0, 0x1040, 0x1440, 0x5840, 0x53FC, 0x5040, 0x90E0, 0x10E0, 0x1150, 0x1150, 0x2A48, 0x2444, 0x4842, 0x4040, 0x8040}},
    {31181, 2, {0x820, 0x1C20, 0xF020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x3124, 0x3924, 0x55FC, 0x5524, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31182, 2, {0x810, 0x1C90, 0xF090, 0x1088, 0x1108, 0xFD04, 0x1204, 0x35FA, 0x3888, 0x5488, 0x5488, 0x9088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {31183, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x1020, 0xFC3C, 0x11E0, 0x3820, 0x3420, 0x503E, 0x53E0, 0x9020, 0x1022, 0x1022, 0x1022, 0x101E}},
    {31184, 2, {0x800, 0x1C00, 0xF1FC, 0x1000, 0x1000, 0xFC00, 0x13FE, 0x3020, 0x3820, 0x5440, 0x5440, 0x9088, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {31185, 2, {0x810, 0x1D10, 0xF090, 0x1090, 0x1010, 0xFD10, 0x1090, 0x3890, 0x3410, 0x501E, 0x53F0, 0x9010, 0x1010, 0x1010, 0x1010, 0x1010}},
    {31186, 2, {0x820, 0x1C20, 0xF020, 0x10A8, 0x10A4, 0xFCA2, 0x1122, 0x3120, 0x3A24, 0x5424, 0x5428, 0x9008, 0x1010, 0x1020, 0x10C0, 0x1300}},
    {31187, 2, {0x820, 0x1C20, 0xF020, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x31FC, 0x3884, 0x5488, 0x5448, 0x9050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {31188, 2, {0x880, 0x1C40, 0xF040, 0x13FC, 0x1000, 0xFC00, 0x11F0, 0x3110, 0x3910, 0x5510, 0x5510, 0x9112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {31189, 2, {0x810, 0x1D10, 0xF110, 0x1112, 0x1112, 0xFD14, 0x11D8, 0x3110, 0x3910, 0x5510, 0x5510, 0x9112, 0x1152, 0x1192, 0x110E, 0x1000}},
    {31190, 2, {0x804, 0x1C1E, 0xF1F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x31FE, 0x3910, 0x5510, 0x5510, 0x9108, 0x110A, 0x114A, 0x1186, 0x1102}},
    {31191, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x3020, 0x3820, 0x5450, 0x5450, 0x9050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {31192, 2, {0x800, 0x1C40, 0xF020, 0x1028, 0x1008, 0xFC88, 0x1090, 0x3A94, 0x36A2, 0x52A2, 0x54C2, 0x9088, 0x1188, 0x1288, 0x1478, 0x1000}},
    {31193, 2, {0x820, 0x1C20, 0xF020, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x3020, 0x39FC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31194, 2, {0x820, 0x1C20, 0xF124, 0x10A4, 0x10A8, 0xFC20, 0x11FC, 0x3820, 0x3420, 0x5020, 0x53FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31195, 2, {0x810, 0x1C10, 0xF010, 0x11FE, 0x1112, 0xFD14, 0x1110, 0x31FC, 0x3944, 0x5544, 0x5528, 0x9128, 0x1110, 0x1228, 0x1244, 0x1482}},
    {31196, 2, {0x800, 0x1DFC, 0xF104, 0x1104, 0x1104, 0xFDFC, 0x1100, 0x3140, 0x3944, 0x5548, 0x5570, 0x9140, 0x1142, 0x1242, 0x123E, 0x1400}},
    {31197, 2, {0x838, 0x1DE0, 0xF020, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x3070, 0x3870, 0x54A8, 0x54A8, 0x9124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31198, 2, {0x820, 0x1C20, 0xF020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x3124, 0x3924, 0x55FC, 0x5524, 0x9124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {31199, 2, {0x800, 0x1DF8, 0xF108, 0x1108, 0x1108, 0xFDF8, 0x1108, 0x3108, 0x3908, 0x55F8, 0x5508, 0x9108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {31200, 2, {0x800, 0x1DFE, 0xF010, 0x1010, 0x1020, 0xFC20, 0x1068, 0x38A4, 0x3522, 0x5222, 0x5020, 0x9020, 0x1020, 0x1000, 0x13FE, 0x1000}},
    {31201, 2, {0x850, 0x1C48, 0xF048, 0x1040, 0x13FE, 0xFC80, 0x1080, 0x30FC, 0x3944, 0x5544, 0x5528, 0x9128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {31202, 2, {0x820, 0x1C20, 0xF050, 0x1050, 0x1088, 0xFD24, 0x1212, 0x3010, 0x39FC, 0x5404, 0x5408, 0x9088, 0x1050, 0x1020, 0x1010, 0x1010}},
    {31203, 2, {0x820, 0x1C20, 0xF020, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x31FC, 0x3870, 0x54A8, 0x54A8, 0x9124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31204, 2, {0x800, 0x1DFC, 0xF020, 0x1020, 0x1124, 0xFCA4, 0x10A8, 0x3020, 0x3BFE, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31205, 2, {0x820, 0x1C20, 0xF020, 0x1020, 0x103E, 0xFC20, 0x1020, 0x3020, 0x39FC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31206, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x10D0, 0x2F08, 0xC106, 0x1FF0, 0x540, 0x920, 0x1110, 0x6108, 0x100}},
    {31207, 2, {0x820, 0x1C20, 0xF020, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x3124, 0x3924, 0x57FE, 0x5420, 0x9050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {31208, 2, {0x880, 0x1C80, 0xF080, 0x10FE, 0x1140, 0xFD40, 0x1240, 0x307C, 0x3840, 0x5440, 0x5440, 0x907C, 0x1040, 0x1040, 0x1040, 0x1040}},
    {31209, 2, {0x820, 0x1C20, 0xF120, 0x1120, 0x11FC, 0xFD20, 0x1220, 0x3020, 0x39FE, 0x5420, 0x5450, 0x9050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {31210, 2, {0x804, 0x1C1E, 0xF1F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x31FE, 0x3910, 0x5510, 0x5510, 0x9108, 0x110A, 0x114A, 0x11A6, 0x1112}},
    {31211, 2, {0x840, 0x1C50, 0xF048, 0x1040, 0x17FC, 0xFC40, 0x10E0, 0x38E0, 0x3550, 0x5150, 0x5248, 0x9444, 0x1842, 0x1040, 0x1040, 0x1040}},
    {31212, 2, {0x800, 0x1DFE, 0xF100, 0x1100, 0x1100, 0xFDFC, 0x1104, 0x3104, 0x3904, 0x5504, 0x55FC, 0x9100, 0x1100, 0x1100, 0x11FE, 0x1000}},
    {31213, 2, {0x820, 0x1C20, 0xF02E, 0x11F0, 0x1120, 0xFD20, 0x1120, 0x39FE, 0x3422, 0x5462, 0x50A2, 0x912A, 0x1624, 0x1020, 0x1020, 0x1020}},
    {31214, 2, {0x820, 0x1C20, 0xF040, 0x1040, 0x1088, 0xFD04, 0x13FE, 0x3002, 0x3800, 0x55FC, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31215, 2, {0x400, 0xE00, 0x78FC, 0x884, 0x884, 0xFF84, 0x884, 0x1884, 0x1CFC, 0x2A84, 0x2A00, 0x4848, 0x8844, 0x884, 0x882, 0x902}},
    {31216, 2, {0x880, 0x1C80, 0xF080, 0x10FE, 0x1102, 0xFD04, 0x1220, 0x3020, 0x38A8, 0x54A4, 0x5124, 0x9122, 0x1222, 0x1020, 0x10A0, 0x1040}},
    {31217, 2, {0x800, 0x1DFE, 0xF102, 0x1102, 0x117A, 0xFD02, 0x1102, 0x317A, 0x394A, 0x554A, 0x554A, 0x917A, 0x1102, 0x1102, 0x110A, 0x1104}},
    {31218, 2, {0x820, 0x1C20, 0xF020, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x3008, 0x3808, 0x55FE, 0x5408, 0x9088, 0x1048, 0x1008, 0x1028, 0x1010}},
    {31219, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x3020, 0x3820, 0x55FC, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31220, 2, {0x820, 0x1C20, 0xF050, 0x1088, 0x1104, 0xFE02, 0x11FC, 0x3000, 0x3800, 0x55FC, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31221, 2, {0x800, 0x1DFC, 0xF104, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x3124, 0x3924, 0x5554, 0x554C, 0x918C, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31222, 2, {0x100, 0x4100, 0x21FC, 0xA44, 0x1448, 0xE0A0, 0x2110, 0x2608, 0xE0, 0x1F00, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {31223, 2, {0x800, 0x1DFE, 0xF020, 0x1040, 0x1088, 0xFD04, 0x11FE, 0x3022, 0x3820, 0x5420, 0x55FE, 0x9020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {31224, 2, {0x820, 0x1C20, 0xF020, 0x13FE, 0x1020, 0xFC20, 0x11FC, 0x3000, 0x3800, 0x55FC, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31225, 2, {0x80, 0x1888, 0x709C, 0x1170, 0x1110, 0xFB10, 0x1510, 0x3110, 0x39FE, 0x5510, 0x5110, 0x9110, 0x1110, 0x1110, 0x117C, 0x1100}},
    {31226, 2, {0x840, 0x1C20, 0xF3FE, 0x1202, 0x1000, 0xFC38, 0x13C0, 0x3840, 0x3440, 0x507C, 0x53C0, 0x9040, 0x1042, 0x1042, 0x103E, 0x1000}},
    {31227, 2, {0x820, 0x1C20, 0xF07C, 0x1084, 0x1148, 0xFC30, 0x1020, 0x3048, 0x3990, 0x543E, 0x5442, 0x91A4, 0x1018, 0x1010, 0x1060, 0x1180}},
    {31228, 2, {0x820, 0x1D20, 0xF120, 0x11FC, 0x1120, 0xFE20, 0x1020, 0x33FE, 0x3870, 0x54A8, 0x50A8, 0x9124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31229, 2, {0x820, 0x1D24, 0xF124, 0x1124, 0x11FC, 0xFC40, 0x1040, 0x38FC, 0x3504, 0x5284, 0x5048, 0x9050, 0x1020, 0x1040, 0x1080, 0x1300}},
    {31230, 2, {0x820, 0x1C20, 0xF020, 0x13FE, 0x1242, 0xFC44, 0x10A0, 0x38A2, 0x35A4, 0x5298, 0x5490, 0x9088, 0x1084, 0x10A2, 0x10C0, 0x1080}},
    {31231, 2, {0x828, 0x1C24, 0xF3FE, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x3124, 0x39FC, 0x5524, 0x5524, 0x91FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {31232, 2, {0x984, 0x1C68, 0xF030, 0x10C8, 0x1304, 0xFC40, 0x13FE, 0x38A0, 0x3520, 0x53FC, 0x5524, 0x9124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {31233, 2, {0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x8074, 0x1F80, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E}},
    {31234, 2, {0x840, 0x1C20, 0xF1FC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5520, 0x5522, 0x9114, 0x1108, 0x1144, 0x1182, 0x1100}},
    {31235, 2, {0x804, 0x1C1E, 0xF1E0, 0x1002, 0x1122, 0xFC94, 0x1000, 0x39FC, 0x3408, 0x5010, 0x53FE, 0x9010, 0x1010, 0x1010, 0x1050, 0x1020}},
    {31236, 2, {0x840, 0x1C40, 0xF088, 0x1104, 0x13FE, 0xFC02, 0x1088, 0x3144, 0x3A42, 0x54F8, 0x5188, 0x9250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {31237, 2, {0x810, 0x1C90, 0xF088, 0x1108, 0x1204, 0xFDFA, 0x1108, 0x3108, 0x3908, 0x55F8, 0x5490, 0x9090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {31238, 2, {0x800, 0x1CFC, 0xF084, 0x1084, 0x1084, 0xFC84, 0x10FC, 0x3800, 0x3400, 0x55FE, 0x5102, 0x9102, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {31239, 2, {0x800, 0x1DFC, 0xF124, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3164, 0x3974, 0x55AC, 0x5524, 0x9124, 0x1124, 0x1104, 0x11FC, 0x1104}},
    {31240, 2, {0x800, 0x1DFC, 0xF104, 0x11FC, 0x1104, 0xFDFC, 0x1000, 0x3800, 0x35FC, 0x5420, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31241, 2, {0x800, 0x1DFE, 0xF020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x31FC, 0x3924, 0x5524, 0x55FC, 0x9120, 0x10A0, 0x1040, 0x10B0, 0x130E}},
    {31242, 2, {0x908, 0x1C88, 0xF090, 0x11FC, 0x1024, 0xFC24, 0x11FC, 0x3120, 0x3920, 0x55FE, 0x5462, 0x90A2, 0x112A, 0x1224, 0x1420, 0x1020}},
    {31243, 2, {0x0, 0xDFC, 0xF104, 0x1104, 0x1104, 0xFDFC, 0x1000, 0x3800, 0x35FC, 0x5420, 0x5020, 0x91FC, 0x1020, 0x1020, 0x1020, 0x17FE}},
    {31244, 2, {0x820, 0x1C20, 0xF050, 0x1088, 0x1104, 0xFEFA, 0x1020, 0x3020, 0x3BFE, 0x5420, 0x5528, 0x9124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {31245, 2, {0x820, 0x1D24, 0xF0A4, 0x10A8, 0x1020, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {31246, 2, {0x904, 0x1C84, 0xF088, 0x1010, 0x11FC, 0xFD04, 0x1104, 0x3104, 0x39FC, 0x5450, 0x5450, 0x9090, 0x1092, 0x1112, 0x120E, 0x1400}},
    {31247, 2, {0x800, 0x1DFE, 0xF048, 0x1048, 0x1048, 0xFDCE, 0x1102, 0x3102, 0x3902, 0x55CE, 0x5048, 0x9048, 0x1048, 0x1048, 0x13FE, 0x1000}},
    {31248, 2, {0x820, 0x1C20, 0xF050, 0x1088, 0x1104, 0xFEFA, 0x1000, 0x31FC, 0x3954, 0x5554, 0x55FC, 0x9154, 0x1154, 0x1154, 0x1104, 0x110C}},
    {31249, 2, {0x820, 0x1C20, 0xF1FC, 0x1020, 0x1020, 0xFDFE, 0x1088, 0x3104, 0x3A22, 0x5420, 0x51FC, 0x9020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {31250, 2, {0x800, 0x1DFC, 0xF104, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3124, 0x3974, 0x5554, 0x5554, 0x9174, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31251, 2, {0x888, 0x1C88, 0xF088, 0x13FE, 0x1088, 0xFC88, 0x17FE, 0x3000, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31252, 2, {0x820, 0x1C20, 0xF050, 0x1088, 0x1144, 0xFE22, 0x11F8, 0x3808, 0x3450, 0x5420, 0x50A4, 0x9282, 0x128A, 0x128A, 0x1478, 0x1000}},
    {31253, 2, {0x840, 0x1C20, 0xF3FE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x3800, 0x35FC, 0x5408, 0x5010, 0x93FE, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {31254, 2, {0x880, 0x1C40, 0xF1FC, 0x1000, 0x1108, 0xFC88, 0x1090, 0x33FE, 0x3800, 0x5400, 0x55FC, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31255, 2, {0x840, 0x1C80, 0xF1FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x3944, 0x35FC, 0x5490, 0x5110, 0x93FE, 0x1010, 0x1010, 0x1010, 0x1010}},
    {31256, 2, {0x888, 0x1C88, 0xF3FE, 0x1088, 0x1088, 0xFCF8, 0x1088, 0x3888, 0x34F8, 0x5488, 0x5088, 0x93FE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {31257, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x1020, 0xFDFC, 0x1104, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x1104, 0x17FE, 0x1000}},
    {31258, 2, {0x8A0, 0x1C90, 0xF090, 0x11FE, 0x1110, 0xFF10, 0x15FE, 0x3110, 0x3910, 0x55FE, 0x5510, 0x9110, 0x1110, 0x11FE, 0x1100, 0x1100}},
    {31259, 2, {0x1000, 0x3BFE, 0xE202, 0x221A, 0x22E2, 0xFA22, 0x2222, 0x33FE, 0x6A22, 0x6272, 0xA2AA, 0x2326, 0x2222, 0x2222, 0x23FE, 0x2202}},
    {31260, 2, {0x1020, 0x3820, 0xE1FC, 0x2020, 0x2020, 0xFBFE, 0x2088, 0x3144, 0x6A42, 0x60F8, 0xA188, 0x2250, 0x2020, 0x2050, 0x2188, 0x2606}},
    {31261, 2, {0x1000, 0x3BDE, 0xE252, 0x2252, 0x2252, 0xFBDE, 0x2252, 0x3252, 0x6A52, 0x63DE, 0xA252, 0x2252, 0x2252, 0x2252, 0x2552, 0x28A6}},
    {31262, 2, {0x800, 0x1DFC, 0xF124, 0x1124, 0x11FC, 0xFD24, 0x1124, 0x31FC, 0x3820, 0x57FE, 0x5470, 0x90A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31263, 2, {0x100, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x0, 0x1FF0, 0x100, 0xFFFE, 0x920, 0x711C}},
    {31264, 2, {0x800, 0x1DFC, 0xF124, 0x1124, 0x1174, 0xFD24, 0x11FC, 0x3104, 0x3974, 0x5554, 0x5554, 0x9174, 0x1104, 0x1204, 0x1214, 0x1408}},
    {31265, 2, {0x840, 0x1C20, 0xF3FE, 0x1000, 0x1088, 0xFC88, 0x1154, 0x3222, 0x3800, 0x5420, 0x53FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31266, 2, {0x1014, 0x3812, 0xE010, 0x27FE, 0x2010, 0xF810, 0x23D2, 0x3252, 0x6A52, 0x6254, 0xA3D4, 0x2008, 0x20EA, 0x271A, 0x2226, 0x2042}},
    {31267, 2, {0x2008, 0x203C, 0x7DE0, 0x4420, 0x8820, 0x7C20, 0x55FE, 0x5420, 0x7C70, 0x5468, 0x54A8, 0x7CA4, 0x122, 0x1C20, 0xE020, 0x4020}},
    {31268, 2, {0x840, 0x1C20, 0xF3FE, 0x1000, 0x1000, 0xFDFC, 0x1104, 0x3104, 0x39FC, 0x5420, 0x5528, 0x9124, 0x1222, 0x1422, 0x10A0, 0x1040}},
    {31269, 2, {0x70, 0x1F80, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC006, 0x1490, 0x2248, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {31270, 2, {0x840, 0x1C40, 0xF1FC, 0x1084, 0x13FE, 0xFC00, 0x11FC, 0x3904, 0x35FC, 0x5020, 0x53FE, 0x9020, 0x1220, 0x13FE, 0x1020, 0x1020}},
    {31271, 2, {0x880, 0x1C9E, 0xF3EA, 0x108A, 0x11CA, 0xFC8A, 0x13EA, 0x3892, 0x34A6, 0x5020, 0x53FE, 0x9020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {31272, 2, {0x840, 0x1C20, 0xF1FE, 0x1102, 0x1102, 0xFDFE, 0x1100, 0x3900, 0x35FE, 0x55AA, 0x51AA, 0x92FE, 0x12AA, 0x12AA, 0x14A2, 0x1086}},
    {31273, 2, {0x800, 0x1DFE, 0xF122, 0x11FE, 0x1122, 0xFDFE, 0x1000, 0x30FC, 0x3884, 0x54FC, 0x5484, 0x90FC, 0x1084, 0x1084, 0x1094, 0x1088}},
    {31274, 2, {0x900, 0x1D00, 0xF1FE, 0x1200, 0x15FC, 0xFD04, 0x11FC, 0x3104, 0x39FC, 0x5480, 0x54FC, 0x9108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {31275, 2, {0x800, 0x1DFE, 0xF000, 0x10FC, 0x1084, 0xFC84, 0x10FC, 0x3000, 0x39FE, 0x5522, 0x5522, 0x91FE, 0x1122, 0x1122, 0x11FE, 0x1102}},
    {31276, 2, {0x800, 0x1DFC, 0xF020, 0x1040, 0x11FC, 0xFD54, 0x1154, 0x3954, 0x3544, 0x552C, 0x5020, 0x93FE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {31277, 2, {0x910, 0x1D12, 0xF1D4, 0x1118, 0x1152, 0xFD92, 0x112E, 0x3040, 0x39FC, 0x5504, 0x5504, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31278, 2, {0x808, 0x1C3C, 0xF1E0, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x3124, 0x39FC, 0x5524, 0x55FC, 0x9020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {31279, 2, {0x800, 0x1C88, 0xF252, 0x1222, 0x1252, 0xFE8A, 0x1202, 0x3BFE, 0x3488, 0x5144, 0x527A, 0x9088, 0x1150, 0x1020, 0x10D8, 0x1706}},
    {31280, 2, {0x800, 0x1DFE, 0xF012, 0x1090, 0x109C, 0xFC90, 0x117E, 0x3200, 0x38FC, 0x5484, 0x50FC, 0x9084, 0x10FC, 0x1084, 0x1094, 0x1088}},
    {31281, 2, {0x81E, 0x1DE0, 0xF044, 0x1124, 0x1088, 0xFC20, 0x11FC, 0x3924, 0x3524, 0x55FC, 0x5124, 0x9124, 0x13FE, 0x1104, 0x1114, 0x1108}},
    {31282, 2, {0x804, 0x1C1E, 0xF1E0, 0x1022, 0x1112, 0xFC94, 0x1080, 0x313E, 0x3922, 0x5522, 0x553E, 0x9122, 0x1122, 0x1122, 0x113E, 0x1122}},
    {31283, 2, {0x840, 0x1CF8, 0xF108, 0x1210, 0x11FC, 0xFC04, 0x1004, 0x39FC, 0x3404, 0x5404, 0x51FC, 0x9040, 0x1024, 0x12A2, 0x128A, 0x1478}},
    {31284, 2, {0x888, 0x1C50, 0xF000, 0x11FE, 0x1050, 0xFDFC, 0x1054, 0x33FE, 0x3854, 0x55FC, 0x5450, 0x90D8, 0x1154, 0x1252, 0x1050, 0x1050}},
    {31285, 2, {0x884, 0x1C44, 0xF048, 0x13FE, 0x1084, 0xFC84, 0x1108, 0x314A, 0x3A52, 0x579C, 0x5084, 0x9108, 0x1108, 0x1252, 0x13DE, 0x1042}},
    {31286, 2, {0x814, 0x1C12, 0xF010, 0x13FE, 0x1014, 0xFC18, 0x1390, 0x32B4, 0x3A98, 0x5690, 0x53B2, 0x900C, 0x10CA, 0x171A, 0x1066, 0x1002}},
    {31287, 2, {0x800, 0x1DFC, 0xF124, 0x11FC, 0x1124, 0xFDFC, 0x1088, 0x3144, 0x3A42, 0x54F8, 0x5188, 0x9250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {31288, 2, {0x840, 0x1C20, 0xF3FE, 0x1040, 0x1088, 0xFDF0, 0x1020, 0x3044, 0x39FE, 0x5402, 0x55FC, 0x9124, 0x11FC, 0x1124, 0x11FC, 0x1104}},
    {31289, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x11FC, 0xFD04, 0x11FC, 0x3904, 0x35FC, 0x5504, 0x51FC, 0x9104, 0x17FE, 0x1088, 0x1104, 0x1202}},
    {31290, 2, {0x1000, 0x3BFC, 0xE204, 0x2204, 0x23FC, 0xFA40, 0x2220, 0x33FC, 0x6A88, 0x6250, 0xA3FE, 0x2220, 0x25FC, 0x2420, 0x2820, 0x2020}},
    {31291, 2, {0x81E, 0x1DE0, 0xF022, 0x1112, 0x1094, 0xFC80, 0x1020, 0x31CE, 0x3902, 0x5502, 0x55CE, 0x9102, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {31292, 2, {0x840, 0x1C20, 0xF3FE, 0x1202, 0x1000, 0xFDFC, 0x1040, 0x30A2, 0x3B34, 0x5458, 0x5094, 0x9334, 0x1052, 0x1090, 0x1350, 0x1020}},
    {31293, 2, {0x850, 0x1C48, 0xF3FC, 0x10A0, 0x1122, 0xFE1E, 0x1110, 0x31E0, 0x3904, 0x54FC, 0x5400, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC}},
    {31294, 2, {0x100, 0x7FFC, 0x820, 0xFE0, 0x0, 0x7FFC, 0x4824, 0x4FE4, 0x4004, 0x5FEC, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {31295, 2, {0x840, 0x1C20, 0xF3FE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x3800, 0x37FE, 0x5202, 0x52FA, 0x928A, 0x12FA, 0x1202, 0x120A, 0x1204}},
    {31296, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0x7C44, 0x1028, 0xFE28, 0x3810, 0x5428, 0x9244, 0x1182}},
    {31297, 2, {0x888, 0x1C88, 0xF1FE, 0x1088, 0x1000, 0xFDFE, 0x1088, 0x3088, 0x38F8, 0x5488, 0x54F8, 0x9088, 0x109E, 0x11E8, 0x1008, 0x1008}},
    {31298, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x3924, 0x35FC, 0x5440, 0x5024, 0x92A2, 0x128A, 0x1488, 0x1078, 0x1000}},
    {31299, 2, {0x840, 0x1C20, 0xF3FE, 0x1202, 0x1050, 0xFC88, 0x1124, 0x3050, 0x3888, 0x5504, 0x52FA, 0x9088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {31300, 2, {0x1100, 0x3910, 0xE3DC, 0x2254, 0x2554, 0xFA88, 0x2108, 0x32F4, 0x6C02, 0x6000, 0xA3FC, 0x2040, 0x2150, 0x2248, 0x2544, 0x2080}},
    {31301, 2, {0x820, 0x1C10, 0xF1FE, 0x1110, 0x117C, 0xFD14, 0x11FE, 0x3914, 0x357C, 0x5510, 0x5192, 0x9154, 0x1138, 0x1254, 0x1292, 0x1430}},
    {31302, 2, {0x820, 0x1C40, 0xF1FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x3820, 0x3524, 0x5632, 0x50C0, 0x9318, 0x1060, 0x118C, 0x1030, 0x13C0}},
    {31303, 2, {0x1040, 0x3888, 0xE1FC, 0x2108, 0x2252, 0xFBFE, 0x2050, 0x3188, 0x6E26, 0x60C0, 0xA310, 0x2064, 0x2388, 0x2030, 0x20C0, 0x2700}},
    {31304, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2638, 0x2B54, 0x3292, 0x2038, 0x27C0, 0x2040, 0x2FFE, 0x2150, 0x4248, 0x4C46, 0x8040}},
    {31305, 2, {0x800, 0x1DFE, 0xF102, 0x11FE, 0x1110, 0xFD54, 0x1138, 0x3954, 0x3582, 0x5550, 0x517C, 0x9190, 0x1110, 0x12FE, 0x1210, 0x1410}},
    {31306, 2, {0x1000, 0xE7BE, 0x2488, 0x2488, 0x27A8, 0xFCA8, 0x24A8, 0x37BE, 0x6C10, 0x6518, 0xA498, 0x25A8, 0x26AA, 0x244A, 0x2046, 0x2080}},
    {31307, 2, {0x1000, 0x3BDE, 0xE042, 0x2252, 0x214A, 0xFA52, 0x2028, 0x30C4, 0x6B12, 0x6060, 0xA188, 0x2032, 0x21C4, 0x2018, 0x2060, 0x2380}},
    {31308, 2, {0x2008, 0x203C, 0x79E0, 0x4820, 0x9020, 0x7C20, 0x55FE, 0x5420, 0x7C70, 0x5468, 0x54A8, 0x7CA4, 0x122, 0xAA20, 0xAA20, 0x20}},
    {31309, 2, {0x820, 0x1DFE, 0xF020, 0x11FC, 0x1020, 0xFDFE, 0x1000, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x91FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {31310, 2, {0x4000, 0x4CFE, 0x7010, 0x4420, 0x44FC, 0x3C84, 0x84, 0xCFC, 0x7084, 0x10FC, 0xFE84, 0x1084, 0x38FC, 0x5448, 0x9084, 0x1102}},
    {31311, 2, {0x808, 0x1C3C, 0xF1C0, 0x1044, 0x1124, 0xFCA8, 0x11FC, 0x3004, 0x38FC, 0x5404, 0x55FC, 0x9040, 0x10A4, 0x12AA, 0x128A, 0x1478}},
    {31312, 2, {0x1100, 0xE1F0, 0x2220, 0x27F8, 0x2A48, 0xFBF8, 0x2248, 0x33F8, 0x6840, 0x67FC, 0xA444, 0x27FC, 0x2444, 0x27FE, 0x2042, 0x203E}},
    {31313, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x1124, 0xFCA8, 0x13FE, 0x3000, 0x39FC, 0x5504, 0x5574, 0x9154, 0x1174, 0x1104, 0x11FC, 0x1104}},
    {31314, 2, {0x70, 0x1F80, 0x100, 0xFFFE, 0x540, 0x1930, 0xE20E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {31315, 2, {0x1000, 0x3BFE, 0xE222, 0x23FE, 0x2222, 0xFBFE, 0x2088, 0x3088, 0x6BFE, 0x6088, 0xA088, 0x27FE, 0x2000, 0x2088, 0x2104, 0x2202}},
    {31316, 2, {0x88, 0x1888, 0x71FC, 0x1088, 0x1088, 0xFBFE, 0x1020, 0x31FC, 0x3924, 0x55FC, 0x5124, 0x91FC, 0x1000, 0x1088, 0x1104, 0x1202}},
    {31317, 2, {0x8A0, 0x1C90, 0xF1FE, 0x1320, 0x11FC, 0xFD20, 0x11FC, 0x3920, 0x35FE, 0x5500, 0x5020, 0x93FE, 0x1070, 0x10A8, 0x1326, 0x1020}},
    {31318, 2, {0x124, 0x1924, 0x72AA, 0x13AE, 0x1124, 0xFAAA, 0x13AE, 0x3124, 0x3BFE, 0x5510, 0x5114, 0x9114, 0x128A, 0x124A, 0x1216, 0x1422}},
    {31319, 2, {0x820, 0x1C20, 0xF3FE, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x3924, 0x35FC, 0x5022, 0x53FE, 0x9042, 0x1024, 0x12A2, 0x128A, 0x1478}},
    {31320, 2, {0x820, 0x1C20, 0xF1FC, 0x1020, 0x13FE, 0xFD08, 0x139C, 0x3108, 0x3988, 0x563E, 0x5000, 0x93FE, 0x1090, 0x1090, 0x1112, 0x120E}},
    {31321, 2, {0x850, 0x1E52, 0xF154, 0x1050, 0x13FE, 0xFC88, 0x1050, 0x33FE, 0x3820, 0x55FC, 0x5020, 0x93FE, 0x1050, 0x1088, 0x1104, 0x1602}},
    {31322, 2, {0x818, 0x1DE0, 0xF040, 0x13FE, 0x1088, 0xFD74, 0x1252, 0x3070, 0x3800, 0x55FC, 0x5504, 0x9174, 0x1154, 0x1174, 0x1104, 0x110C}},
    {31323, 2, {0x890, 0x1C88, 0xF1FE, 0x1110, 0x1310, 0xFDFE, 0x1110, 0x3910, 0x35FE, 0x5510, 0x5110, 0x91FE, 0x1100, 0x12A4, 0x1252, 0x1452}},
    {31324, 2, {0x820, 0x1DFC, 0xF088, 0x1050, 0x13FE, 0xFC00, 0x11FC, 0x3124, 0x39FC, 0x5524, 0x55FC, 0x9020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {31325, 2, {0x1000, 0x39F8, 0xE108, 0x21F8, 0x2108, 0xF9F8, 0x2000, 0x37FE, 0x6A40, 0x63DC, 0xA254, 0x23D4, 0x2268, 0x27C8, 0x2054, 0x2062}},
    {31326, 2, {0x880, 0x1CF8, 0xF110, 0x13FC, 0x1524, 0xFDFC, 0x1124, 0x31FC, 0x3800, 0x57FE, 0x5400, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC}},
    {31327, 2, {0x1088, 0xE448, 0x2250, 0x20FC, 0x2020, 0xF850, 0x2694, 0x3238, 0x6A50, 0x6298, 0xA234, 0x2254, 0x2290, 0x2220, 0x25FE, 0x2800}},
    {31328, 2, {0x850, 0x1DFC, 0xF154, 0x11FC, 0x1154, 0xFDFC, 0x1000, 0x31FE, 0x3900, 0x557C, 0x5500, 0x91FE, 0x1150, 0x1154, 0x1148, 0x1266}},
    {31329, 2, {0x20, 0x1820, 0x73FE, 0x1124, 0x1124, 0xFAAA, 0x17FE, 0x3000, 0x39FC, 0x5504, 0x5174, 0x9154, 0x1174, 0x1104, 0x11FC, 0x1104}},
    {31330, 2, {0x820, 0x1D3C, 0xF120, 0x17FE, 0x1010, 0xFC14, 0x13FE, 0x3210, 0x3BF0, 0x5654, 0x5254, 0x92D4, 0x1368, 0x144A, 0x1896, 0x1322}},
    {31331, 2, {0x1080, 0x3840, 0xE7FE, 0x2110, 0x2208, 0xFD14, 0x23F8, 0x3110, 0x6BF8, 0x6110, 0xA7FC, 0x2128, 0x2310, 0x2D48, 0x2186, 0x2100}},
    {31332, 2, {0x800, 0x1DFC, 0xF020, 0x13FE, 0x1222, 0xFDAC, 0x1020, 0x31AC, 0x3800, 0x57FE, 0x5020, 0x91FC, 0x1154, 0x1154, 0x1154, 0x110C}},
    {31333, 2, {0x1140, 0x3A4C, 0xE264, 0x2244, 0x234C, 0xFA64, 0x22A4, 0x33AC, 0x6AA4, 0x62A4, 0xA7FE, 0x2000, 0x2110, 0x2108, 0x2204, 0x2404}},
    {31334, 2, {0x840, 0x1C20, 0xF3FE, 0x1202, 0x11FC, 0xFD48, 0x1250, 0x39FC, 0x3704, 0x55FC, 0x5104, 0x91FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {31335, 2, {0x1080, 0x3840, 0xE7FC, 0x2110, 0x20A4, 0xFF58, 0x2554, 0x3552, 0x6B58, 0x6000, 0xA208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2408}},
    {31336, 2, {0xC20, 0x71FC, 0x1124, 0xFDFC, 0x1020, 0x3BFE, 0x5400, 0x91FC, 0x104, 0x39FC, 0x2904, 0x29FC, 0x2904, 0x4DFC, 0x4888, 0x8104}},
    {31337, 2, {0x103E, 0xE7C0, 0x2244, 0x2128, 0x23F8, 0xF840, 0x27FC, 0x3000, 0x6BF8, 0x6008, 0xA3F8, 0x2008, 0x23F8, 0x2544, 0x252A, 0x28FA}},
    {31338, 2, {0x1000, 0x3BFE, 0xE124, 0x2222, 0x27FE, 0xFA22, 0x2376, 0x32AA, 0x6B76, 0x6222, 0xA376, 0x22AA, 0x2376, 0x2222, 0x222A, 0x2204}},
    {31339, 2, {0x88, 0x1BFE, 0x70A8, 0x1090, 0x11FE, 0xFB20, 0x11FC, 0x3120, 0x39FC, 0x5520, 0x51FE, 0x9100, 0x13FC, 0x1088, 0x1070, 0x138E}},
    {31340, 2, {0x1040, 0x3820, 0xE7FE, 0x2488, 0x25FC, 0xFC88, 0x27FE, 0x3420, 0x6DFC, 0x6524, 0xA5FC, 0x2524, 0x25FC, 0x2400, 0x2488, 0x2904}},
    {31341, 2, {0x880, 0x1DF8, 0xF208, 0x15FC, 0x1124, 0xFDFC, 0x1124, 0x31FC, 0x3954, 0x562A, 0x55FC, 0x9104, 0x11FC, 0x1104, 0x11FC, 0x1104}},
    {31342, 2, {0x840, 0x1C20, 0xF3FE, 0x1250, 0x13FE, 0xFE52, 0x13FE, 0x3200, 0x3A94, 0x56D8, 0x5290, 0x92D2, 0x128E, 0x1400, 0x1554, 0x1A2A}},
    {31343, 2, {0x20, 0x1BFE, 0x7020, 0x13AE, 0x12AA, 0xFBAE, 0x1020, 0x33FE, 0x3800, 0x57FE, 0x5202, 0x92FA, 0x128A, 0x12FA, 0x1202, 0x13FE}},
    {31344, 2, {0x1040, 0xE7FC, 0x2000, 0x23B8, 0x22A8, 0xFBB8, 0x2110, 0x37FC, 0x6910, 0x67FC, 0xA110, 0x2FFE, 0x2128, 0x2310, 0x2D48, 0x2186}},
    {31345, 2, {0x1E, 0x1BE0, 0x7122, 0x1094, 0x13FE, 0xFA52, 0x13FE, 0x3004, 0x3BC4, 0x567E, 0x53C4, 0x9254, 0x13CC, 0x1284, 0x12D4, 0x1348}},
    {31346, 2, {0x13DE, 0x3800, 0xE3DE, 0x2252, 0x235A, 0xFA52, 0x2020, 0x33FE, 0x6A50, 0x63FE, 0xA252, 0x23FE, 0x2292, 0x22DC, 0x2292, 0x24CE}},
    {31347, 2, {0x1528, 0xE7BE, 0x2948, 0x27BE, 0x2318, 0xFDAA, 0x2946, 0x33FC, 0x6A04, 0x63FC, 0xA204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {31348, 2, {0x200, 0x100, 0x100, 0x7FFE, 0x4002, 0x8044, 0x440, 0x440, 0x420, 0x820, 0x820, 0x1010, 0x1010, 0x2008, 0x4004, 0x8002}},
    {31349, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2008, 0x1FE0, 0xC0, 0x300, 0x400, 0x800, 0x1008, 0x1008, 0xFF8, 0x0}},
    {31350, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1210, 0x2208, 0x200, 0x3FE0, 0x220, 0x420, 0x420, 0x822, 0x1022, 0x201E, 0x4000}},
    {31351, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2208, 0x200, 0x7FF0, 0x210, 0x410, 0x410, 0x810, 0x1010, 0x20A0, 0x4040}},
    {31352, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2208, 0x400, 0xFF0, 0x1010, 0x6820, 0x440, 0x280, 0x300, 0x1C00, 0xE000}},
    {31353, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2FF8, 0x10, 0x10, 0x1FF0, 0x1000, 0x2000, 0x3FF8, 0x8, 0x8, 0x50, 0x20}},
    {31354, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2008, 0x0, 0x1FF0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x7FFC, 0x0}},
    {31355, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2FE8, 0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {31356, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2108, 0x100, 0x7FFC, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {31357, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2828, 0x820, 0x3FF8, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x1020, 0x2020}},
    {31358, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0xF0, 0x3F00, 0x100, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31359, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x3FFC, 0x40, 0x1040, 0x2040, 0x3FFE, 0x140, 0x640, 0x1840, 0x6140, 0x80}},
    {31360, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x2108, 0xFFFE, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x10A, 0x102, 0x102, 0xFE}},
    {31361, 2, {0x200, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1110, 0x2148, 0x120, 0x100, 0x7FFC, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31362, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2108, 0x1100, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {31363, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2000, 0x21FC, 0x2C44, 0xF044, 0x2044, 0x2044, 0x2484, 0x2884, 0x3104, 0x2228, 0x410}},
    {31364, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2808, 0xFF8, 0x1400, 0x2400, 0x47F0, 0x400, 0x400, 0x7F8, 0x400, 0x400, 0x400}},
    {31365, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31366, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x1F0, 0x3E00, 0x200, 0x100, 0x7FF8, 0x60, 0x180, 0xE00, 0x3000, 0x4800, 0x87FE}},
    {31367, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1090, 0x1F90, 0x1050, 0x1024, 0x1004, 0xFFC}},
    {31368, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x840, 0x840, 0x12F8, 0x2248, 0x7C48, 0x848, 0x1448, 0x2288, 0x7E88, 0x2128, 0x210}},
    {31369, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x1FF0, 0x100, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2808, 0x2028, 0x2010}},
    {31370, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2078, 0x1FA0, 0x1220, 0x1220, 0x1220, 0x1210, 0x1290, 0x2248, 0x22A8, 0x4324, 0x8202}},
    {31371, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2108, 0x1110, 0x1110, 0x1110, 0x1FF0, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {31372, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0xC00, 0x70FC, 0x4084, 0x4484, 0x4484, 0x5C84, 0x6494, 0x4888, 0x880, 0x1080}},
    {31373, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x0, 0x7DFE, 0x1040, 0x1080, 0x11FC, 0x1004, 0x1E04, 0xE004, 0x4028, 0x10}},
    {31374, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x2208, 0x1FF0, 0x1010, 0x1210, 0x1150, 0x1020, 0x1FFC, 0x4, 0x4, 0x7FD4, 0x8}},
    {31375, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x1000, 0x9FC, 0x4820, 0x2020, 0x2020, 0xBFE, 0x1020, 0x7020, 0x1020, 0x10A0, 0x1040}},
    {31376, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {31377, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x800, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {31378, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x7FFC, 0x400, 0x820, 0x1FF0, 0x110, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {31379, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9214, 0x2408, 0x820, 0x1010, 0x3FF8, 0x8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {31380, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9214, 0x100, 0x7FFC, 0x820, 0x1010, 0x2448, 0x4444, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {31381, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x440, 0x4448, 0x2450, 0x2460, 0xC40, 0x3460, 0xC852, 0x84A, 0x1042, 0x203E, 0x4000}},
    {31382, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x900, 0xFF0, 0x1100, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31383, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9214, 0x400, 0x3FF8, 0x2208, 0x23C8, 0x2448, 0x2A88, 0x2108, 0x2288, 0x2408, 0x3FF8, 0x2008}},
    {31384, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x3FF8, 0x408, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {31385, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x911C, 0xFD0, 0x120, 0xFFFE, 0x100, 0x1FF0, 0x820, 0x30C0, 0xCFFC, 0x80, 0x280, 0x100}},
    {31386, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9094, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3E88, 0x2250, 0x2250, 0x2A24, 0x4454, 0x418C, 0x8604}},
    {31387, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2008, 0x3E20, 0x1FE, 0x24, 0x7E44, 0x14E8, 0x1410, 0x146A, 0x2586, 0x2402, 0x43FE}},
    {31388, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x7FFC, 0x4104, 0x7FFC, 0x4104, 0x100, 0x100}},
    {31389, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2FE8, 0x820, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2008, 0x2018}},
    {31390, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9414, 0x27E8, 0x820, 0x1040, 0x2678, 0x1808, 0x1008, 0x1E78, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {31391, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x3FF8, 0x2008, 0x3FF8, 0x2080, 0x2490, 0x2490, 0x27F0, 0x2080, 0x4888, 0x4888, 0x8FF8}},
    {31392, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x3FF8, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {31393, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x3EF8, 0x208, 0x14D0, 0x820, 0x1450, 0x6288, 0x3EF8, 0x288, 0x1450, 0x820, 0x3458, 0xC286}},
    {31394, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x50, 0x48, 0xFFFE, 0x40, 0x3E40, 0x2224, 0x3E28, 0x12, 0xE2A, 0x70C6, 0x302}},
    {31395, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9214, 0x100, 0x7FFC, 0x820, 0x1450, 0x2288, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {31396, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x3FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x2080, 0x208C, 0x3EF0, 0x2082, 0x2682, 0x387E}},
    {31397, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x11FC, 0x7D04, 0x1124, 0x1124, 0xFD24, 0x1154, 0x2850, 0x2490, 0x4112, 0x860E}},
    {31398, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x2FE8, 0x100, 0x7FFC, 0x484, 0x1288, 0x900, 0xFFFE, 0x260, 0x410, 0x1808, 0x6004}},
    {31399, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x1FC, 0x100, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {31400, 2, {0x200, 0x100, 0x7FFE, 0x4A22, 0x9114, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {31401, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x3FF8, 0x1010, 0x1F90, 0x1090, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x400C}},
    {31402, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x40, 0x23F8, 0x1040, 0x8040, 0x47FC, 0x1000, 0x2040, 0xE3F8, 0x2040, 0x2040, 0x27FC}},
    {31403, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x200, 0xFFFE, 0x440, 0x1830, 0xE00E}},
    {31404, 2, {0x200, 0x100, 0x7FFE, 0x4922, 0x9294, 0xC60, 0x37D8, 0xC006, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2228, 0x2610}},
    {31405, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9114, 0x920, 0x7FFC, 0x540, 0x1930, 0x610C, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38}},
    {31406, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x3F7C, 0x2104, 0x3F04, 0x217C, 0x3F40, 0x2140, 0x7F7C, 0x504, 0x1904, 0xE528, 0x210}},
    {31407, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2448, 0x280, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x2488, 0x2244, 0x4244}},
    {31408, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x1F8, 0x7E00, 0x2208, 0x1110, 0x2000, 0x3FF8, 0x4100, 0x100, 0xFFFE, 0x100, 0x2108, 0x3FF8}},
    {31409, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9094, 0x11F8, 0x2288, 0x6870, 0xABAE, 0x2820, 0x29FC, 0x2820, 0x28A8, 0x2924, 0x22A4, 0x2040}},
    {31410, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x4948, 0x2A28, 0x808, 0xFF48, 0x182E, 0x2CF8, 0x4A08, 0x8808, 0x808, 0x808}},
    {31411, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x61C, 0x38E8, 0x28A8, 0x2CA8, 0x2AA8, 0x2AA8, 0x2AA8, 0x28A4, 0x5554, 0x5D74, 0x8212}},
    {31412, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x1010, 0x2008}},
    {31413, 2, {0x100, 0x7FFE, 0x4822, 0x9414, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4924, 0x8494, 0x8}},
    {31414, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x3FF8, 0x2108, 0xFFFE, 0x2108, 0x3FF8, 0x1110, 0x1FF0, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {31415, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x40, 0xFD9E, 0x4952, 0x4952, 0x7952, 0x4952, 0x7952, 0x49DA, 0x5D54, 0xE890, 0x890, 0x910}},
    {31416, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0xF8, 0x3F00, 0x1110, 0x920, 0xFFFE, 0x920, 0x3118, 0xC206, 0x900, 0x2908, 0x2824, 0x47E4}},
    {31417, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x5FC, 0x2420, 0x24F8, 0x3C48, 0x448, 0x5FE, 0x7C00, 0x24FC, 0x2484, 0x2484, 0x44FC, 0x8484}},
    {31418, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x0, 0x11F8, 0x7D08, 0x11F8, 0x1108, 0xFDF8, 0x1108, 0x29F8, 0x2490, 0x4112, 0x860E}},
    {31419, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x200, 0x3FF8, 0x2208, 0x27E8, 0x2A48, 0x2188, 0x2648, 0x3FF8, 0x100, 0x4884, 0x4812, 0x87F2}},
    {31420, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x2220, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {31421, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x2220, 0x3C20, 0x213E, 0x1F42, 0x2094, 0x3F10, 0x4810, 0xFFA8, 0x1428, 0x2244, 0x4182}},
    {31422, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x820, 0x7F20, 0x83E, 0x3E42, 0x94, 0x3E10, 0x10, 0x7F10, 0x828, 0x2A28, 0x4944, 0x1882}},
    {31423, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x79F0, 0x4A90, 0x5060, 0x6198, 0x5606, 0x49F0, 0x4A40, 0x6BF8, 0x5440, 0x43F8, 0x4040, 0x47FC}},
    {31424, 2, {0x200, 0x100, 0x7FFE, 0x4822, 0x9014, 0x1F8, 0xFD08, 0x11F8, 0x1108, 0x51F8, 0x5D08, 0x51F8, 0x5090, 0x5C90, 0xF112, 0x60E}},
    {31425, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE, 0xC20, 0x703C, 0x1DE2, 0xF022, 0x1C1E}},
    {31426, 2, {0x100, 0x7FFE, 0x4822, 0x9114, 0x7FFC, 0x440, 0x2828, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x100, 0x1110, 0x2508, 0x200}},
    {31427, 2, {0x100, 0x7FFE, 0x4922, 0x97D4, 0x100, 0x7FFC, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {31428, 2, {0x100, 0x7FFE, 0x4822, 0x9614, 0x38F8, 0x2008, 0x3CF8, 0x2008, 0x3FF8, 0x0, 0x2948, 0x2528, 0x2948, 0x252A, 0x2946, 0x3182}},
    {31429, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x3E20, 0x2220, 0x3E3E, 0x2244, 0x3EA4, 0x824, 0x7F28, 0x1028, 0x1E10, 0x2228, 0x2A44, 0x4482}},
    {31430, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x800, 0x3E7C, 0x2244, 0x3E44, 0x227C, 0x3E00, 0x22FE, 0xFE82, 0x1282, 0x2282, 0xCAFE, 0x482}},
    {31431, 2, {0x100, 0x7FFE, 0x4922, 0x97D4, 0x100, 0x3FF8, 0x2448, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {31432, 2, {0x100, 0x7FFE, 0x4922, 0x97D4, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EFA, 0x282, 0x1FE}},
    {31433, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0x840, 0x7E7C, 0x2440, 0xFF7C, 0x4, 0x7E7C, 0x4240, 0x7E7C, 0x4240, 0x7E7C, 0x4242, 0x463E}},
    {31434, 2, {0x100, 0x7FFE, 0x4822, 0x9014, 0xC40, 0x707C, 0x1040, 0x55FC, 0x3924, 0x1154, 0xFD04, 0x33FE, 0x3A22, 0x564A, 0x92FA, 0x1206}},
    {31435, 2, {0x200, 0x100, 0x100, 0x0, 0x7FFC, 0x0, 0x10, 0x1010, 0x820, 0x820, 0x440, 0x440, 0x480, 0x0, 0xFFFE, 0x0}},
    {31436, 2, {0x1000, 0x800, 0x8F0, 0x7E90, 0x90, 0x490, 0x4490, 0x4490, 0x2490, 0x2890, 0x2890, 0xE92, 0xF092, 0x4112, 0x10E, 0x200}},
    {31437, 2, {0x1020, 0x820, 0x820, 0x7E20, 0x20, 0x420, 0x47FE, 0x4420, 0x2420, 0x2820, 0x2820, 0xE20, 0xF020, 0x4020, 0x20, 0x20}},
    {31438, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0, 0x440, 0x440, 0x7FFC, 0x440, 0x840, 0x840, 0x1040, 0x2040}},
    {31439, 2, {0x1008, 0x83C, 0x9E0, 0x7E20, 0x20, 0x420, 0x4420, 0x47FE, 0x2420, 0x2820, 0x2820, 0xE20, 0xF020, 0x4020, 0x20, 0x20}},
    {31440, 2, {0x1000, 0x8F8, 0x888, 0x7E88, 0x88, 0x506, 0x4600, 0x45FC, 0x2484, 0x2884, 0x2848, 0xE50, 0xF020, 0x4050, 0x88, 0x306}},
    {31441, 2, {0x2040, 0x1040, 0x1040, 0xFC40, 0x3FE, 0x840, 0x8880, 0x8890, 0x4890, 0x4920, 0x5120, 0x5248, 0x1E44, 0xE484, 0x41FE, 0x82}},
    {31442, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x10, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x10, 0x50, 0x20}},
    {31443, 2, {0x1008, 0x81C, 0x8E0, 0x7E20, 0x20, 0x43C, 0x44E0, 0x4420, 0x2420, 0x283E, 0x29E0, 0xE20, 0xF022, 0x4022, 0x22, 0x1E}},
    {31444, 2, {0x2008, 0x1068, 0x1388, 0xFC88, 0x88, 0x888, 0x8888, 0x8BFE, 0x4888, 0x4888, 0x5088, 0x5088, 0x1D08, 0xE108, 0x4208, 0x408}},
    {31445, 2, {0x2010, 0x1090, 0x1090, 0xFC88, 0x108, 0x904, 0x8A04, 0x8DFA, 0x4888, 0x4888, 0x5088, 0x5088, 0x1D08, 0xE108, 0x4228, 0x410}},
    {31446, 2, {0x800, 0x4BF8, 0x4A08, 0x4910, 0x48A0, 0x4840, 0x48B0, 0x4B0C, 0xA00, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {31447, 2, {0x1020, 0x820, 0x820, 0x7EA8, 0xA4, 0x4A2, 0x4522, 0x4520, 0x2624, 0x2824, 0x2828, 0xE08, 0xF010, 0x4020, 0xC0, 0x300}},
    {31448, 2, {0x1040, 0x840, 0x880, 0x7EFE, 0x102, 0x602, 0x44F2, 0x4492, 0x2492, 0x2892, 0x28F2, 0xE92, 0xF002, 0x4002, 0x14, 0x8}},
    {31449, 2, {0x1020, 0x820, 0x820, 0x7E20, 0x3E, 0x420, 0x4420, 0x4420, 0x25FC, 0x2904, 0x2904, 0xF04, 0xF104, 0x4104, 0x1FC, 0x104}},
    {31450, 2, {0x2040, 0x1020, 0x1020, 0xFDFE, 0x102, 0xA04, 0x8800, 0x8800, 0x49FE, 0x4820, 0x5020, 0x5020, 0x1C20, 0xE020, 0x40A0, 0x40}},
    {31451, 2, {0x1020, 0x820, 0x850, 0x7E50, 0x88, 0x524, 0x4612, 0x4410, 0x25FC, 0x2804, 0x2808, 0xE88, 0xF050, 0x4020, 0x10, 0x10}},
    {31452, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1112, 0x102, 0xFE, 0x0}},
    {31453, 2, {0x1020, 0x810, 0x810, 0x7EFE, 0x0, 0x400, 0x4484, 0x4484, 0x2444, 0x2848, 0x2848, 0xE10, 0xF010, 0x41FE, 0x0, 0x0}},
    {31454, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {31455, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x842, 0x1042, 0x603E}},
    {31456, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {31457, 2, {0x1000, 0x9FE, 0x810, 0x7E10, 0x20, 0x4FC, 0x4484, 0x4484, 0x2484, 0x28FC, 0x2884, 0xE84, 0xF084, 0x4084, 0xFC, 0x84}},
    {31458, 2, {0x2020, 0x1040, 0x1088, 0xFD04, 0x1FE, 0x802, 0x8880, 0x88FC, 0x4920, 0x4820, 0x53FE, 0x5020, 0x1C50, 0xE088, 0x4104, 0x202}},
    {31459, 2, {0x2040, 0x1040, 0x1088, 0xFD04, 0x3FE, 0x802, 0x8888, 0x8944, 0x4A42, 0x48F8, 0x5188, 0x5250, 0x1C20, 0xE050, 0x4188, 0x606}},
    {31460, 2, {0x2040, 0x1020, 0x13FE, 0x202, 0xFC44, 0x40, 0xBFE, 0x8840, 0x8890, 0x4890, 0x5110, 0x5120, 0x1A24, 0xE242, 0x4FE, 0x842}},
    {31461, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {31462, 2, {0x2020, 0x1020, 0x13FE, 0xFC20, 0x20, 0x9FC, 0x8924, 0x8924, 0x49FC, 0x4820, 0x5070, 0x50A8, 0x1D24, 0xE222, 0x4020, 0x20}},
    {31463, 2, {0x1020, 0x840, 0x8F8, 0x7E88, 0xF8, 0x488, 0x44F8, 0x448A, 0x248C, 0x29F8, 0x2818, 0xE28, 0xF048, 0x4088, 0x328, 0x10}},
    {31464, 2, {0x2020, 0x1020, 0x103E, 0xFC20, 0x1FC, 0x904, 0x89FC, 0x8904, 0x49FC, 0x4924, 0x5020, 0x53FE, 0x1C20, 0xE020, 0x4020, 0x20}},
    {31465, 2, {0x2020, 0x1010, 0x11FE, 0xFD02, 0x204, 0x8F8, 0x8888, 0x8888, 0x48F8, 0x4888, 0x5088, 0x50F8, 0x1C88, 0xE088, 0x43FE, 0x0}},
    {31466, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {31467, 2, {0x2040, 0x1040, 0x10F8, 0xFD08, 0x10, 0x9FC, 0x8824, 0x8824, 0x4BFE, 0x4824, 0x5024, 0x51FC, 0x1C24, 0xE020, 0x40A0, 0x40}},
    {31468, 2, {0x2004, 0x101E, 0x13E0, 0xFC20, 0x1FC, 0x924, 0x8924, 0x89FC, 0x4820, 0x4BFE, 0x5222, 0x522A, 0x1EFA, 0xE202, 0x420A, 0x204}},
    {31469, 2, {0x2000, 0x11FC, 0x1104, 0xFDFC, 0x104, 0x9FC, 0x8880, 0x89FE, 0x4A22, 0x4922, 0x5152, 0x5102, 0x1DFA, 0xE002, 0x4014, 0x8}},
    {31470, 2, {0x2000, 0x13FC, 0x1204, 0xFE04, 0x3FC, 0xA00, 0x8A84, 0x8A48, 0x4AFC, 0x4A48, 0x5248, 0x53FE, 0x1C48, 0xE448, 0x4888, 0x1108}},
    {31471, 2, {0x2020, 0x1124, 0x1124, 0xFD24, 0x1FC, 0x800, 0x8BFE, 0x8820, 0x4840, 0x49FC, 0x5154, 0x5154, 0x1D54, 0xE154, 0x4154, 0x10C}},
    {31472, 2, {0x1000, 0x9FE, 0x900, 0x7F7C, 0x154, 0x554, 0x457C, 0x4554, 0x2554, 0x297C, 0x2910, 0xF10, 0xF17C, 0x4210, 0x210, 0x4FE}},
    {31473, 2, {0x2020, 0x13FE, 0x1020, 0xFDFC, 0x124, 0x9FC, 0x8924, 0x89FC, 0x4822, 0x4BFE, 0x5008, 0x53FE, 0x1D08, 0xE088, 0x4028, 0x10}},
    {31474, 2, {0x2104, 0x1088, 0x1000, 0x3FE, 0xFA22, 0x2AA, 0xA72, 0x8A22, 0x8BFE, 0x4800, 0x51FC, 0x5104, 0x19FC, 0xE104, 0x1FC, 0x104}},
    {31475, 2, {0x2010, 0x13D2, 0x1054, 0xFD48, 0x8A, 0x904, 0x8AFA, 0x8C00, 0x49FC, 0x4904, 0x5104, 0x51FC, 0x1D04, 0xE088, 0x47FE, 0x0}},
    {31476, 2, {0x2088, 0x1050, 0x13FE, 0xFC50, 0x1FC, 0x954, 0x898C, 0x8974, 0x4904, 0x49FC, 0x5008, 0x53FE, 0x1D08, 0xE088, 0x40A8, 0x10}},
    {31477, 2, {0x2000, 0x13FE, 0x1008, 0x3BE, 0xFA88, 0x2A8, 0xABE, 0x8B88, 0x8808, 0x4BFE, 0x528A, 0x5252, 0x1BFE, 0xE222, 0x22A, 0x224}},
    {31478, 2, {0x2020, 0x1010, 0xFEFE, 0x4444, 0x2828, 0xFEFE, 0x0, 0x7EFC, 0x4284, 0x4284, 0x7EFC, 0x2850, 0x2A50, 0x2C92, 0x4912, 0x820E}},
    {31479, 2, {0x1020, 0xFE7C, 0x44A4, 0x2818, 0xFEE6, 0x10, 0x7CFE, 0x4410, 0x7C90, 0x44FE, 0x7C10, 0x1020, 0xFE7C, 0x10C8, 0x1030, 0x11CE}},
    {31480, 2, {0x2020, 0x1010, 0xFEFE, 0x4444, 0x2828, 0xFEFE, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x2828, 0x2A28, 0x4C4A, 0x8886}},
    {31481, 2, {0x1040, 0x1040, 0x1040, 0x1040, 0x3EFE, 0x2890, 0x4910, 0x8A10, 0x810, 0x810, 0x810, 0x810, 0x810, 0x810, 0x850, 0x820}},
    {31482, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8000, 0x0, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFFFE, 0x0}},
    {31483, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8200, 0x200, 0x7FF8, 0x208, 0x408, 0x408, 0x808, 0x1008, 0x2050, 0x4020}},
    {31484, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8000, 0xFE0, 0x820, 0x820, 0xA20, 0x920, 0x922, 0x1022, 0x2022, 0x401E}},
    {31485, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {31486, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8100, 0x1170, 0x1390, 0x1D10, 0xF110, 0x1150, 0x1122, 0x1102, 0x1002, 0xFFE}},
    {31487, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {31488, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x200, 0x100, 0xFFFE, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800, 0xFFC, 0x0}},
    {31489, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8000, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {31490, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8200, 0x200, 0x7FE0, 0x220, 0xA20, 0x420, 0x620, 0x922, 0x1022, 0x201E, 0x4000}},
    {31491, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0xBFF0, 0x10, 0x1010, 0x1010, 0x1FFC, 0x4, 0x4, 0x7FE4, 0x4, 0x28, 0x10}},
    {31492, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FFC, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {31493, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9810, 0x660, 0x180, 0x660, 0x1818, 0x6C24, 0x240, 0x180, 0x660, 0x1818, 0xE006}},
    {31494, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2002, 0x2002, 0x2002, 0x1FFE}},
    {31495, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8100, 0x80, 0x0, 0x7FFE, 0x100, 0x120, 0x110, 0x108, 0x100, 0x100, 0x100}},
    {31496, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x3FE0, 0x820, 0x840, 0xC78, 0xA08, 0x1210, 0x1120, 0x20C0, 0x4330, 0x8C0E}},
    {31497, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9508, 0x1000, 0x1FF8, 0x2008, 0x4C08, 0x8208, 0x68, 0x188, 0x1E08, 0x808, 0x50, 0x20}},
    {31498, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8010, 0xF8, 0x1F20, 0x1220, 0x1220, 0x1220, 0x1210, 0x2210, 0x2208, 0x4204, 0x8202}},
    {31499, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FF0, 0x210, 0x210, 0xFFFE, 0x210, 0x410, 0x3FF0, 0x810, 0x1000, 0x2000, 0xC000}},
    {31500, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FFC, 0x20, 0x1020, 0x2020, 0x3FFE, 0x120, 0x220, 0xC20, 0x3020, 0xC0A0, 0x40}},
    {31501, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x2008, 0x2028, 0x2010}},
    {31502, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x444, 0x844, 0x1044, 0x603C}},
    {31503, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9508, 0x1000, 0x1FF8, 0x2248, 0x4248, 0x8488, 0x888, 0x1108, 0x2208, 0x4408, 0x850, 0x1020}},
    {31504, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x7FFC, 0x0, 0x0, 0xFE0, 0x820, 0x820, 0x822, 0x1022, 0x201E, 0xC000}},
    {31505, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x80F0, 0x3F00, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31506, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8280, 0x440, 0xA20, 0x3118, 0xC106, 0x1FE0, 0x20, 0x40, 0x40, 0x80, 0x100}},
    {31507, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x2080, 0x2088, 0x3C90, 0x20A0, 0x20C0, 0x2084, 0x2484, 0x2884, 0x307C, 0x0}},
    {31508, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0xF8, 0x3F00, 0x100, 0x1F8, 0x3F00, 0x100, 0x1FC, 0x7F02, 0x102, 0x102, 0xFE}},
    {31509, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1290, 0x1490, 0x882, 0x3082, 0xC07E}},
    {31510, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x2410, 0x2210, 0x2110, 0x2110, 0x2010, 0x2420, 0x2830, 0x3048, 0x2184, 0x602}},
    {31511, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE30E, 0xC0, 0xE00, 0x180, 0x40}},
    {31512, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {31513, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9100, 0x1100, 0x3FFC, 0x4100, 0x8100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE}},
    {31514, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x100, 0x100}},
    {31515, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {31516, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0x4104, 0x4104, 0x7FFC, 0x4}},
    {31517, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2008, 0x2FC8, 0x2048, 0x2688, 0x2108, 0x2288, 0x2C48, 0x2008, 0x3FF8, 0x2008}},
    {31518, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x1010, 0x3FF8, 0x8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31519, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8078, 0x1FA0, 0x1220, 0x1220, 0x1220, 0x1210, 0x1290, 0x2248, 0x22A8, 0x4324, 0x8202}},
    {31520, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8200, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x820, 0x440, 0x0, 0xFFFE}},
    {31521, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {31522, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x3FFC, 0x2040, 0x2024, 0x2C14, 0x300C}},
    {31523, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1090, 0x1F90, 0x1050, 0x1024, 0x1004, 0xFFC, 0x0}},
    {31524, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x2020, 0xDFF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {31525, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x3FF8, 0x8, 0x7FE8, 0x8, 0x1F88, 0x1088, 0x1088, 0x1F88, 0x10A8, 0x10}},
    {31526, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x810, 0x810, 0x13FE, 0x3010, 0x5110, 0x9090, 0x1090, 0x1010, 0x1010, 0x1050, 0x1020}},
    {31527, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x1E78, 0x1248, 0x1248, 0x1248, 0xFFFE, 0x1248, 0x1248, 0x1248, 0x1288, 0x26A8, 0x4110}},
    {31528, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2FE8, 0xC106, 0x100, 0x100, 0x100}},
    {31529, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8850, 0x848, 0x1040, 0x307C, 0x57C0, 0x9040, 0x1020, 0x1020, 0x1014, 0x100C, 0x1004}},
    {31530, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x0}},
    {31531, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8170, 0x1F80, 0x1100, 0x2100, 0x3FFC, 0x304, 0x504, 0x904, 0x3128, 0xC110, 0x100}},
    {31532, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x108, 0x108, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x304, 0x504, 0x1928, 0xE110, 0x100}},
    {31533, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8280, 0x440, 0xA20, 0x3118, 0xC106, 0x1FE0, 0x20, 0x40, 0xC80, 0x300, 0x80}},
    {31534, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8800, 0x800, 0x1FFC, 0x2400, 0x4400, 0x87F0, 0x400, 0x400, 0x7F8, 0x400, 0x400}},
    {31535, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9000, 0x11F8, 0xFE88, 0x1288, 0x2288, 0x2450, 0x6450, 0x1820, 0x1850, 0x2488, 0xC306}},
    {31536, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x440, 0x3FF8, 0x448, 0x448, 0x1FF8, 0x1440, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2048, 0x4040}},
    {31537, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9508, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x9F88, 0x1088, 0x1088, 0x1F88, 0x8, 0x50, 0x20}},
    {31538, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8820, 0x1010, 0x3FF8, 0x8, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020}},
    {31539, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x107C, 0x7E44, 0x1244, 0x1244, 0x1244, 0x1244, 0x2244, 0x2244, 0x4A7C, 0x8444}},
    {31540, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0xFFFE, 0x10, 0x10, 0x1F90, 0x1090, 0x1090, 0x1F90, 0x10, 0x50, 0x20}},
    {31541, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x9F8, 0x4908, 0x2108, 0x2508, 0x528, 0x910, 0x7102, 0x1102, 0x1102, 0x10FE}},
    {31542, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8800, 0x800, 0x1FF0, 0x2100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31543, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0xC00, 0x70FC, 0x4084, 0x4484, 0x4484, 0x5C84, 0x6494, 0x4888, 0x880, 0x1080, 0x6080}},
    {31544, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x3FFC, 0x2000, 0x2000, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x2000, 0x2000, 0x3FFE}},
    {31545, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8220, 0x1220, 0x1220, 0xFFFE, 0x1220, 0x1220, 0x13E0, 0x1000, 0x1000, 0x1FFC, 0x0}},
    {31546, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x240, 0x220, 0x3F8, 0x7E00, 0x200, 0x3FC, 0xFE10, 0x120, 0xC0, 0x744, 0x3834, 0xC}},
    {31547, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x7CFC, 0x1084, 0x1084, 0x1084, 0x1084, 0x1E94, 0xF088, 0x4080, 0x80, 0x80}},
    {31548, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8040, 0x420, 0x7FFC, 0x480, 0x488, 0x890, 0x8A0, 0x10C2, 0x2382, 0x4C82, 0x807E}},
    {31549, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x2488, 0x4508, 0x8000, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {31550, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x1080, 0x17F8, 0x88, 0x7088, 0x1108, 0x1108, 0x1228, 0x1410, 0x2800, 0x47FE}},
    {31551, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0xFF0, 0x1020, 0x2C40, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31552, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {31553, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4548, 0x8440, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2008}},
    {31554, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8548, 0x2040, 0x1258, 0x9268, 0x43C8, 0x4E48, 0xA48, 0x125A, 0xE242, 0x2202, 0x21FE, 0x2000}},
    {31555, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x2008}},
    {31556, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x440, 0x4448, 0x2450, 0x2460, 0xC40, 0x3460, 0xC852, 0x84A, 0x1042, 0x203E, 0x4000}},
    {31557, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1100, 0x1100, 0x3FF8, 0x4100, 0x100, 0xFFFE, 0x440, 0x840, 0x1042, 0x2042, 0xC03E}},
    {31558, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FF0, 0x110, 0xFFFE, 0x110, 0x3FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {31559, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x7EFC, 0x884, 0x888, 0x890, 0x888, 0x884, 0xE84, 0xF0A8, 0x4090, 0x80}},
    {31560, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xF0, 0x1F00, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31561, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x20, 0x7FFC, 0x820, 0x420, 0x4A0, 0x40}},
    {31562, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x7FFC, 0x820, 0x1010, 0x2448, 0x4444, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {31563, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3E20, 0x2220, 0x22FC, 0x3E24, 0x2224, 0x2224, 0x3E24, 0x2244, 0x4244, 0x4A94, 0x8508}},
    {31564, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8280, 0x440, 0x820, 0x3018, 0xCFE6, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x7FFC}},
    {31565, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0x1000, 0x1FF8, 0x2008, 0x5FC8, 0x9048, 0x1FC8, 0x1048, 0x1FC8, 0x1048, 0x8, 0x50, 0x20}},
    {31566, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9508, 0x1000, 0xFE7C, 0x1244, 0x2244, 0x2244, 0x6444, 0x1444, 0x844, 0x147C, 0x2244, 0xC000}},
    {31567, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x88A0, 0x890, 0x1080, 0x30FC, 0x5780, 0x9048, 0x1050, 0x1024, 0x1054, 0x118C, 0x1604}},
    {31568, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FFE, 0x2000, 0x2FFC, 0x2080, 0x2080, 0x27F8, 0x2080, 0x2080, 0x2FFC, 0x2000, 0x3FFE}},
    {31569, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7DF0, 0x1110, 0x1110, 0x1190, 0x1150, 0x1D12, 0xE212, 0x4212, 0x40E, 0x800}},
    {31570, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0xBFF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {31571, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x7EFC, 0x820, 0x820, 0x820, 0xFEFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {31572, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31573, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x21FC, 0x4000, 0x8800, 0x13FE, 0x3020, 0x5020, 0x9020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {31574, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2388, 0x2558, 0x920, 0x1110, 0x610C, 0x100}},
    {31575, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x1040, 0x27FC, 0x6444, 0xA444, 0x2444, 0x27FC, 0x2444, 0x2040, 0x2040, 0x2040}},
    {31576, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8800, 0x800, 0x7EFC, 0x884, 0x884, 0xE84, 0x7884, 0x884, 0x8FC, 0x2884, 0x1000}},
    {31577, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31578, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x2088, 0x3EB0, 0x20C4, 0x2084, 0x267C, 0x3900, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {31579, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x4BFE, 0x4820, 0x4820, 0x49FC, 0x4924, 0x4924, 0x4924, 0x1134, 0x1128, 0x2020, 0x4020}},
    {31580, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9110, 0x920, 0x100, 0x7FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x7FF8, 0x8}},
    {31581, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8800, 0x1FF0, 0x2020, 0x5FF8, 0x108, 0xFFFE, 0x108, 0x1FF8, 0x108, 0x500, 0x200}},
    {31582, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x3FFC, 0x2004, 0x4108, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31583, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xF8, 0x7F00, 0x2210, 0x1120, 0x3FE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {31584, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9080, 0x10FC, 0x7D04, 0x1204, 0x1084, 0x1054, 0x1C24, 0xE044, 0x4184, 0x28, 0x10}},
    {31585, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A88, 0x440, 0x820, 0x37D8, 0xC106, 0x100, 0x3FF8, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {31586, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x11FC, 0xFD24, 0x1124, 0x1124, 0x1DFC, 0xF100, 0x1100, 0x1102, 0x5102, 0x20FE}},
    {31587, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8808, 0x608, 0x7808, 0x848, 0x848, 0xFF48, 0x1C48, 0x2A48, 0x4948, 0x8808, 0x828, 0x810}},
    {31588, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1108, 0x1090, 0x1260, 0x1418, 0x1806}},
    {31589, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {31590, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x3FFC, 0x2004, 0x5FE8, 0x0, 0x7FFC, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {31591, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x842, 0x1042, 0x603E}},
    {31592, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x280, 0xC60, 0x3118, 0xCFE6, 0x40, 0x80, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31593, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FE0, 0x440, 0x280, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x2128, 0x2010}},
    {31594, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0xBFFC, 0x2000, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x2080, 0x2080, 0x3FFE, 0x0}},
    {31595, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8000, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {31596, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8090, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3E48, 0x2250, 0x2224, 0x4A54, 0x448C, 0x8104}},
    {31597, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {31598, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7FFC, 0x100, 0x1110, 0x1110, 0x1110, 0x2928, 0x4544, 0x8584, 0x100, 0xFFFE}},
    {31599, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8528, 0x3E20, 0x2220, 0x22FC, 0x3E24, 0x2224, 0x3E24, 0x2224, 0x2244, 0x2F44, 0xF094, 0x4108}},
    {31600, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8880, 0x880, 0x10FE, 0x1140, 0x3240, 0x547C, 0x9040, 0x1040, 0x107E, 0x1040, 0x1040}},
    {31601, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x1040, 0x247E, 0x2488, 0x6488, 0xA548, 0x2450, 0x2420, 0x2050, 0x2088, 0x2306}},
    {31602, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1110, 0x920, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {31603, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x38, 0x7BC0, 0x840, 0x1040, 0x23F8, 0x7840, 0x840, 0x2BFC, 0x1000, 0x2800, 0x47FE}},
    {31604, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31605, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x18, 0x7BE0, 0x840, 0x1040, 0x2278, 0x7A40, 0xA40, 0x2BFC, 0x1000, 0x2800, 0x47FE, 0x8000}},
    {31606, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31607, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x11F8, 0x1848, 0x5448, 0x5048, 0x53FE, 0x9040, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {31608, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {31609, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFC, 0x200, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x1FFC, 0x2420, 0x4220, 0x80A0, 0x40}},
    {31610, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FFE, 0x2000, 0x2FFC, 0x2080, 0x2080, 0x27F8, 0x20A0, 0x2090, 0x2FFC, 0x2000, 0x3FFE}},
    {31611, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2100, 0x1A00, 0x600, 0x19C0, 0x603E}},
    {31612, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2108, 0x2208, 0x4E0, 0x1818, 0xE004}},
    {31613, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {31614, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8280, 0x440, 0x820, 0x37D8, 0xC006, 0x210, 0x1110, 0x920, 0x840, 0x7FFC, 0x0}},
    {31615, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8FE0, 0x1440, 0x380, 0x1C70, 0xE10E, 0x100, 0x3FF8, 0x100, 0x1110, 0x2508, 0x200}},
    {31616, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0xBF8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2018}},
    {31617, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31618, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0x380, 0x540, 0x920, 0x3118, 0xC106}},
    {31619, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7F00, 0x22FC, 0x3E44, 0x2244, 0x3E44, 0x2228, 0x2228, 0x2F10, 0xF228, 0x4244, 0x282}},
    {31620, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x400, 0x3FF8, 0x2108, 0x3FF8, 0x2208, 0x3FF8, 0x880, 0x1080, 0xFFFE, 0x80, 0x80}},
    {31621, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020}},
    {31622, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x0, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {31623, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8, 0x2008}},
    {31624, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x1110, 0x1208, 0x87FC, 0x4204, 0x5000, 0x13F8, 0x2208, 0xE208, 0x2208, 0x23F8, 0x2208}},
    {31625, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x11F8, 0xFD08, 0x1108, 0x15F8, 0x1840, 0x33FC, 0xD044, 0x1084, 0x1084, 0x5114, 0x2208}},
    {31626, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9528, 0x820, 0x7E50, 0x1050, 0x1088, 0x1F06, 0x1260, 0x1210, 0x2200, 0x22C0, 0x4A30, 0x8408}},
    {31627, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x210, 0x2FC, 0x7F40, 0x84, 0x764, 0x781C, 0x210, 0x2FC, 0x7F40, 0x84, 0x764, 0x781C}},
    {31628, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7F84, 0x1004, 0x2224, 0x7F24, 0x924, 0x824, 0x7F24, 0x824, 0x824, 0xF04, 0xF014, 0x4008}},
    {31629, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x13FE, 0x1220, 0xFE20, 0x12FC, 0x12A4, 0x1EA4, 0xF2A4, 0x12B4, 0x12A8, 0x5220, 0x23FE}},
    {31630, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x27E0, 0x4422, 0x4822, 0xB01E}},
    {31631, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xFC, 0x7F00, 0x2208, 0x1110, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x108, 0x500, 0x200}},
    {31632, 2, {0x2040, 0x3F7E, 0x4990, 0xBFF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {31633, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x1100, 0x11F8, 0x2900, 0x47FE, 0x8000}},
    {31634, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x10, 0xFF0, 0x10, 0x1FF0, 0x0, 0x7FFE, 0x4102, 0x9FF4, 0x1110, 0x1150, 0x1120}},
    {31635, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x1110, 0x9A0, 0x540, 0x1930, 0xE10E, 0x500, 0x200}},
    {31636, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8528, 0x1040, 0x9FC, 0x4904, 0x2104, 0x2504, 0x5FC, 0x904, 0x7104, 0x1104, 0x11FC, 0x1104}},
    {31637, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {31638, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x820, 0x7EFC, 0x820, 0xC70, 0x1A68, 0x2AA4, 0x4922, 0x8820, 0x820, 0x820}},
    {31639, 2, {0x2040, 0x3E7E, 0x4890, 0x8000, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {31640, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2008, 0x20C8, 0x2708, 0x2108, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x2108, 0x3FF8, 0x2008}},
    {31641, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3EFC, 0x2284, 0x2294, 0x3E88, 0x2280, 0x22FC, 0x3EA4, 0x22A4, 0x22A8, 0x2290, 0x4AA8, 0x84C6}},
    {31642, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8C04, 0x1224, 0x2124, 0x5EA4, 0x8024, 0x3F24, 0x2124, 0x2124, 0x2104, 0x3F14, 0x2108}},
    {31643, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1C, 0x7DF0, 0x950, 0x1150, 0x1150, 0x1D50, 0x3148, 0xD148, 0x1144, 0x1254, 0x525A, 0x2468}},
    {31644, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2008, 0x1FF0, 0x100, 0x100, 0x100, 0x7FFC, 0x0}},
    {31645, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x1088, 0x1088, 0xFBFE, 0x1088, 0x1088, 0x1CF8, 0xF088, 0x1088, 0x1088, 0x50F8, 0x2088}},
    {31646, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC826, 0x868, 0x808, 0x7F8}},
    {31647, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x9FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x2080, 0x208C, 0x3EF0, 0x2082, 0x2682, 0x387E}},
    {31648, 2, {0x2080, 0x3EFE, 0x4910, 0x80F8, 0x3F00, 0x100, 0x7FFC, 0x920, 0x920, 0xFFFE, 0x920, 0x920, 0x7FFC, 0x100, 0x100, 0x3FF8}},
    {31649, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x7FFE, 0x4002, 0x9FE4, 0x1020, 0x1FE0, 0x1000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31650, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x7FFE, 0x4002, 0x9004, 0x1EF8, 0x2288, 0x24A8, 0x5490, 0x884, 0x1084, 0x607C}},
    {31651, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8808, 0x804, 0xFFA4, 0x824, 0x7F24, 0x4924, 0x4B24, 0x1C24, 0x2A24, 0x4904, 0x8814, 0x808}},
    {31652, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x100, 0x7FFC, 0x820, 0x820, 0x1450, 0x2288, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {31653, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2020, 0x1020, 0x13FC, 0x8224, 0x4220, 0x4BF8, 0xA88, 0x1288, 0xE250, 0x2420, 0x2450, 0x298C}},
    {31654, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1110, 0x2C0, 0xC30, 0x7008}},
    {31655, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x3FFC, 0x2080, 0x2080, 0x2FF8, 0x2490, 0x22A0, 0x2FFC, 0x2140, 0x2220, 0x2410, 0x3FFE}},
    {31656, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x11FC, 0x1088, 0xFC70, 0x138E, 0x1020, 0x1DFC, 0x3020, 0xD3FE, 0x1020, 0x5020, 0x2020}},
    {31657, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x200, 0x7F0, 0x810, 0x3420, 0x240, 0x180, 0xE00, 0x7000}},
    {31658, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x440, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {31659, 2, {0x1040, 0x1F7C, 0x28A0, 0x4510, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x2548, 0x2528, 0x2528, 0x2928, 0x2108, 0x4108}},
    {31660, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x7FFC, 0x820, 0x200, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x810}},
    {31661, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1010, 0x820, 0xFFFE, 0x0, 0x3E48, 0x2248, 0x3E48, 0x2248, 0x3E08, 0x2208, 0x2A28, 0x2410}},
    {31662, 2, {0x2040, 0x3F7E, 0x5090, 0x8A08, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE}},
    {31663, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x17FC, 0x3040, 0x53F8, 0x9248, 0x13F8, 0x1248, 0x13F8, 0x1140, 0x10C0, 0x1330, 0x1C0E}},
    {31664, 2, {0x2040, 0x3E7E, 0x4890, 0x9FF8, 0x2008, 0x5FC8, 0x9048, 0x1FC8, 0x1068, 0x1FD0, 0x0, 0x1FE0, 0x40, 0xFFFE, 0x100, 0x300}},
    {31665, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8800, 0x8FC, 0x7E84, 0x884, 0x18FC, 0x1C84, 0x2A84, 0x2AFC, 0x4884, 0x8884, 0x8FC}},
    {31666, 2, {0x2040, 0x3F7E, 0x4890, 0x8528, 0x1120, 0x920, 0x9FC, 0x4220, 0x2020, 0x2BFE, 0x890, 0x1090, 0x7090, 0x1112, 0x1112, 0x120E}},
    {31667, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2008, 0x3FF8, 0x2410, 0x2220, 0x2FF8, 0x2220, 0x2220, 0x2FFC, 0x4220, 0x4420, 0x8820}},
    {31668, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4528, 0x8024, 0x3FFE, 0x2020, 0x2FA4, 0x2024, 0x2FA8, 0x2898, 0x2892, 0x4FAA, 0x4046, 0x8082}},
    {31669, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1110, 0x2168, 0x4384, 0x1C00, 0xFFF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {31670, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8808, 0x87C, 0x7F44, 0x844, 0x87C, 0x3E44, 0x2244, 0x227C, 0x2244, 0x3E84, 0x2294, 0x108}},
    {31671, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0xA548, 0x1040, 0xFEFE, 0x2120, 0x20AC, 0x3CB4, 0x25E4, 0x24AC, 0x24A0, 0x44A2, 0x5482, 0x887E}},
    {31672, 2, {0x1040, 0x1040, 0x3F7E, 0x2890, 0x4508, 0x8200, 0x1FD0, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x810}},
    {31673, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x1440, 0x247C, 0x7884, 0x1104, 0x2444, 0x7E24, 0x24, 0x5404, 0x4A04, 0x8A28, 0x10}},
    {31674, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x820}},
    {31675, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x40, 0x13F8, 0x2048, 0x47FE, 0x8848, 0x13F8, 0x3040, 0x53FC, 0x9040, 0x17FE, 0x1040, 0x1040}},
    {31676, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2FF8, 0x2200, 0x2410, 0x2FF8, 0x2080, 0x4FF8, 0x4080, 0xBFFE}},
    {31677, 2, {0x2040, 0x3E7E, 0x4890, 0x9FF0, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {31678, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x4904, 0x2A24, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x4104, 0x4514, 0x4208}},
    {31679, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x103E, 0xFFD0, 0x10}},
    {31680, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7E7C, 0x4244, 0x7E44, 0x4244, 0x7E44, 0x4844, 0x4454, 0x5A48, 0x6240, 0x4040}},
    {31681, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A08, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {31682, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2100, 0x3FF8, 0x4100, 0x9FF0, 0x100, 0xFFFE, 0x0}},
    {31683, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2108, 0x3FF8, 0x2000, 0x2FF8, 0x2808, 0x2FF8, 0x2808, 0x4FF8, 0x4808, 0x8FF8, 0x808}},
    {31684, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x7F7C, 0x844, 0x7F44, 0x4944, 0x7F44, 0x4954, 0x7F48, 0x842, 0xFFC2, 0x842, 0x83E}},
    {31685, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x200, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {31686, 2, {0x2040, 0x3E7E, 0x4890, 0x8800, 0xFE0, 0x1020, 0x1FC0, 0x40, 0xFFFE, 0x600, 0x3B08, 0x4B0, 0x19C0, 0x62A0, 0xC98, 0x7306}},
    {31687, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2FFC, 0x2924, 0x2924, 0x2FFC, 0x4924, 0x4924, 0x880C}},
    {31688, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x800, 0x808, 0x7F08, 0x8FE, 0xFF08, 0x848, 0x828, 0x7F28, 0x808, 0xF08, 0xF028, 0x4010}},
    {31689, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {31690, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x2090, 0x1090, 0x13FC, 0x8090, 0x4890, 0x4890, 0x17FE, 0xE000, 0x2090, 0x2108, 0x2204}},
    {31691, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x3FFC, 0x2080, 0x2FF8, 0x2080, 0x2490, 0x2490, 0x2AA8, 0x3144, 0x2220, 0x2410, 0x3FFE}},
    {31692, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x9F0, 0x810, 0x1010, 0x37FC, 0x5100, 0x91F8, 0x1240, 0x1040, 0x17FC, 0x10A0, 0x1110, 0x120C}},
    {31693, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xC20, 0x7020, 0x10A4, 0x10A4, 0xFCA8, 0x3120, 0x3820, 0x5450, 0x5050, 0x9088, 0x1104, 0x1202}},
    {31694, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x20, 0x7C20, 0x44A8, 0x7CA4, 0x4522, 0x7D22, 0x4628, 0x4410, 0x7C60, 0x180, 0x600}},
    {31695, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8528, 0x2220, 0x223E, 0xFF42, 0x2294, 0x2210, 0x3E10, 0x2210, 0x2228, 0x3E28, 0x2244, 0x82}},
    {31696, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x13FE, 0xFE20, 0x1220, 0x32FC, 0x3AA4, 0x56A4, 0x92A4, 0x12B4, 0x12A8, 0x1220, 0x13FE}},
    {31697, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1110, 0x2C0, 0xC30, 0x7008}},
    {31698, 2, {0x2040, 0x2040, 0x7EFE, 0x5110, 0x8A88, 0xE60, 0x3118, 0xDFE6, 0x1020, 0x1FE0, 0x1020, 0x1FE0, 0x1008, 0x1190, 0x1460, 0x1810}},
    {31699, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0x610C, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38}},
    {31700, 2, {0x2080, 0x3EFE, 0x4910, 0x9FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010}},
    {31701, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x800, 0x1010, 0x3FF8, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {31702, 2, {0x2080, 0x3EFE, 0x4910, 0x8080, 0x3FFC, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2080, 0x4FF8, 0x4808, 0x8FF8, 0x808}},
    {31703, 2, {0x2080, 0x3EFE, 0x4910, 0x8880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3FF0, 0xC60, 0x380, 0x7C7C}},
    {31704, 2, {0x2080, 0x3EFE, 0x4910, 0x9000, 0x1FFC, 0x2104, 0x5114, 0x1FF4, 0x208, 0x1400, 0x1FFC, 0x2104, 0x5114, 0x1FF4, 0x414, 0x3808}},
    {31705, 2, {0x2040, 0x3E7E, 0x4890, 0x8200, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {31706, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FFC, 0x2240, 0x2240, 0x3E78, 0x2240, 0x2E70, 0x2240, 0x3E78, 0x2240, 0x2240, 0x3FFC, 0x0}},
    {31707, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x7EFC, 0x204, 0x7EFC, 0x4080, 0x7EFC, 0x204, 0x2244, 0x1A34, 0x62C4, 0x1428, 0x810}},
    {31708, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3118, 0xC506, 0x200, 0x3FF8, 0x0, 0x2488, 0x4244}},
    {31709, 2, {0x2040, 0x3F7E, 0x4890, 0x8548, 0x7FFC, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x1010, 0x1050, 0x1020}},
    {31710, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2000, 0x17FC, 0x1404, 0x87FC, 0x4404, 0x57FC, 0x1040, 0x27FC, 0xE040, 0x2040, 0x2FFE, 0x2000}},
    {31711, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x840, 0xFFC, 0x1040, 0x33F8, 0x5040, 0x97FE, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1218}},
    {31712, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1080, 0x11F8, 0x2288, 0x6870, 0xABAE, 0x2820, 0x29FC, 0x2820, 0x28A8, 0x2924, 0x22A4, 0x2040}},
    {31713, 2, {0x2040, 0x3F7E, 0x4890, 0x9FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0x7FFC, 0x440, 0x1A30, 0xE44E, 0xFE0}},
    {31714, 2, {0x2080, 0x3EFE, 0x4910, 0xBFF8, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {31715, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x820, 0x440, 0x7FFE, 0x4202, 0x8104, 0x3FF8, 0x400, 0x7F0, 0x810, 0x1050, 0x2020}},
    {31716, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x0, 0x1FF8, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x4, 0x2924, 0x4494, 0x8}},
    {31717, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x100}},
    {31718, 2, {0x2040, 0x3E7E, 0x4890, 0x8200, 0x3FF8, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x0, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {31719, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x400, 0x7FFC, 0x4504, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {31720, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7840, 0x48A0, 0x5110, 0x6208, 0x55F6, 0x4840, 0x4BFC, 0x6840, 0x5150, 0x4248, 0x4544, 0x4080}},
    {31721, 2, {0x2040, 0x3E7E, 0x4890, 0x9000, 0x23FE, 0x7C20, 0x4420, 0x45FC, 0x7D24, 0x4124, 0x7D24, 0x4524, 0x4534, 0x7D28, 0x4420, 0x20}},
    {31722, 2, {0x2040, 0x3F7E, 0x4890, 0x857C, 0x3F80, 0x20F8, 0x2080, 0x2FFC, 0x2884, 0x28F0, 0x2B88, 0x2878, 0x4800, 0x51E0, 0x9224, 0x241C}},
    {31723, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7DF0, 0x1110, 0x1190, 0x1D52, 0xE20E, 0x5FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x300}},
    {31724, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x6C0, 0x1A30, 0xE10E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x2FF8, 0x4808, 0x8FF8}},
    {31725, 2, {0x2040, 0x3F7E, 0x5090, 0x8A08, 0x100, 0x3FF8, 0x820, 0xFFFE, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {31726, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x480, 0x2480, 0x24FC, 0x24A0, 0x2510, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {31727, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9528, 0x1024, 0x3C3E, 0x41E0, 0xBC20, 0x102E, 0x13F4, 0xFC28, 0x1012, 0x106A, 0x1586, 0x1802}},
    {31728, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x87C, 0x7F44, 0x2248, 0x1450, 0xFFC8, 0x44, 0x3E44, 0x2254, 0x2248, 0x3E40, 0x2240}},
    {31729, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0xFFFE, 0x440, 0x1390, 0x1450, 0x1FF0, 0x100, 0x7FFC, 0x4444, 0x4FE4, 0x4424, 0x400C}},
    {31730, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x3FF8, 0x8, 0x1FF8, 0x8, 0x3FF8}},
    {31731, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x920, 0xFFFE, 0x920, 0x3FF8, 0x100, 0x7FFC, 0x100}},
    {31732, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x23FC, 0x1080, 0x1144, 0x668, 0x70B0, 0x1128, 0x1668, 0x10A4, 0x1124, 0x16A0, 0x2840, 0x47FE}},
    {31733, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1110, 0x2110, 0x4110, 0x8AA8, 0x1444, 0x3040, 0x5240, 0x927C, 0x1240, 0x1540, 0x14FE, 0x1800}},
    {31734, 2, {0x2040, 0x3E7E, 0x4890, 0xFFFC, 0x100, 0x11F8, 0x1100, 0xFFFE, 0x1000, 0x1FF8, 0x2000, 0x3FFC, 0x4, 0x4924, 0x8494, 0x8}},
    {31735, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x23F0, 0x1510, 0x10E0, 0x71C, 0x7040, 0x13F8, 0x1040, 0x11F0, 0x1040, 0x17FC, 0x2840, 0x47FE}},
    {31736, 2, {0x2080, 0x3EFE, 0x4910, 0x8820, 0x1FF0, 0x1020, 0x2548, 0x7EFC, 0x440, 0x1930, 0xE64E, 0x1990, 0x620, 0x18C0, 0x700, 0x3800}},
    {31737, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1490, 0x1250, 0x1FF0, 0x200, 0xFFFE, 0x820, 0x3FD8, 0xC846, 0x8C0, 0x808, 0x7F8}},
    {31738, 2, {0x1040, 0x1040, 0x3F7E, 0x4890, 0x9528, 0x1040, 0x11FC, 0x7D24, 0x11FC, 0x1144, 0x1DFC, 0x7050, 0x1090, 0x13FE, 0x5010, 0x2010}},
    {31739, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1110, 0x2508, 0x200}},
    {31740, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1200, 0x67DC, 0x4444, 0x47C4, 0x4444, 0x57DC, 0x6000, 0x440, 0x440, 0x842, 0x1042, 0x603E}},
    {31741, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1400, 0x27DE, 0x4912, 0x8112, 0x17D2, 0x2112, 0x6112, 0xA5D2, 0x251A, 0x25D4, 0x2E10, 0x2410}},
    {31742, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x90, 0x3FFC, 0x2080, 0x3048, 0x2850, 0x2424, 0x40D4, 0x830C}},
    {31743, 2, {0x2080, 0x3EFE, 0x4910, 0xFFFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFC, 0x20, 0xFFFE, 0x820, 0x460}},
    {31744, 2, {0x2080, 0x3EFE, 0x4910, 0xBFF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {31745, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1020, 0x213C, 0x4120, 0x8920, 0x17FE, 0x3000, 0x5020, 0x9120, 0x113E, 0x1120, 0x12A0, 0x147E}},
    {31746, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x40A4, 0x7FFC, 0x4084, 0x5E94, 0x5294, 0x5E64, 0x4054, 0x46B4, 0x5914, 0x7FFC, 0x4004}},
    {31747, 2, {0x2040, 0x3F7E, 0x4890, 0x8528, 0xC40, 0x70FC, 0x1308, 0x1090, 0xFC60, 0x3090, 0x3B3E, 0x5442, 0x55A4, 0x9018, 0x1060, 0x1380}},
    {31748, 2, {0x2080, 0x3EFE, 0x4910, 0xBFF8, 0x2008, 0x3FF8, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2FF8, 0x2888, 0x4FF8, 0x4802, 0x87FE}},
    {31749, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x3FFC, 0x410, 0x220, 0x3FFE, 0x2080, 0x2480, 0x2FFC, 0x3080, 0x47F8, 0x4080, 0x9FFE}},
    {31750, 2, {0x2040, 0x3F7E, 0x5090, 0x8A08, 0x100, 0x7FFE, 0x4022, 0xBE3C, 0x20, 0x7EF8, 0x1408, 0x1450, 0x1424, 0x24D4, 0x4404, 0x83FC}},
    {31751, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2040, 0x1040, 0xFEFE, 0x2120, 0x2040, 0x3CFC, 0x2520, 0x2420, 0x25FE, 0x4450, 0x5488, 0x8906}},
    {31752, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FFC, 0x2004, 0x3FFC, 0x2208, 0x2410, 0x2F78, 0x2410, 0x2410, 0x3F7C, 0x4410, 0x4410, 0x8810}},
    {31753, 2, {0x2040, 0x3F7E, 0x4890, 0x8548, 0x2240, 0x13F8, 0x1440, 0x40, 0x77FC, 0x1000, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE, 0x0}},
    {31754, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1110, 0x2FE8, 0xC106, 0x3FF8}},
    {31755, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF4, 0x1088, 0x1470, 0x180E, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {31756, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0xFFA0, 0x83E, 0x7F42, 0x4994, 0x4910, 0x7F10, 0x1C10, 0x2A28, 0x4928, 0x8844, 0x882}},
    {31757, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF8, 0x2108, 0xFFFE, 0x2108, 0x3FF8, 0x1110, 0x1FF0, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {31758, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1090, 0x1090, 0x13FC, 0xFC90, 0x13FE, 0x1000, 0x1DF8, 0xF108, 0x11F8, 0x1108, 0x51F8, 0x2108}},
    {31759, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2848, 0x2F70, 0x4844, 0x4B44, 0x8C3C}},
    {31760, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x7F20, 0x83E, 0x7F42, 0x4984, 0x7F10, 0x4910, 0x7F10, 0x828, 0xFFA8, 0x844, 0x882}},
    {31761, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0xFFFE, 0x1010, 0x1FF0, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {31762, 2, {0x2080, 0x3EFE, 0x4910, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x200, 0xFFFE, 0x820, 0x3518, 0xCA26, 0x17F0}},
    {31763, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x4, 0x7F08, 0x4910, 0x5D60, 0x4904, 0x7F08, 0x4110, 0x5D62, 0x5504, 0x5D08, 0x4510, 0x8260}},
    {31764, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x0, 0x3FF8, 0x1010, 0xFFFE, 0x1010, 0x3FF8, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {31765, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2220, 0x7F20, 0x2220, 0x3EFC, 0x824, 0x7F24, 0x4924, 0x7F24, 0x844, 0xFF44, 0x894, 0x908}},
    {31766, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x40C, 0x5570, 0x4E40, 0x4440, 0x5F7E, 0x4448, 0x4E48, 0x5548, 0x4448, 0x7F88, 0x88, 0x108}},
    {31767, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0xBF0, 0x4090, 0x2A94, 0x1292, 0x7512, 0x1250, 0x1420, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {31768, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FF0, 0x110, 0xFFFC, 0x110, 0x3FF0, 0x920, 0x2548, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x4108}},
    {31769, 2, {0x2080, 0x3EFE, 0x4950, 0x8048, 0x27FC, 0x2040, 0x23F8, 0xFA48, 0x23F8, 0x2248, 0x23F8, 0x2248, 0x2010, 0x2FFE, 0x2110, 0x20B0}},
    {31770, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x28, 0xFFFE, 0x820, 0x7F24, 0x4928, 0x4D10, 0x4A2A, 0x846}},
    {31771, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x220C, 0x2270, 0xFF40, 0x2240, 0x3E40, 0x227E, 0x3E48, 0x2248, 0xFF48, 0x2248, 0x4148, 0x8088}},
    {31772, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x23F8, 0x1208, 0x93F8, 0x4208, 0x43F8, 0x1000, 0x2FFE, 0xE200, 0x27FC, 0x2924, 0x2254, 0x488}},
    {31773, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x440, 0x2828, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x100, 0x1110, 0x2508, 0x200}},
    {31774, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3EF8, 0x2288, 0x3EF8, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {31775, 2, {0x2040, 0x3E7E, 0x4890, 0xFFFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {31776, 2, {0x2040, 0x3F7E, 0x4890, 0x8528, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {31777, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {31778, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3F7C, 0x2084, 0x2FF4, 0x2044, 0x2244, 0x2184, 0x2654, 0x2008}},
    {31779, 2, {0x2080, 0x3EFE, 0x4910, 0x9FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {31780, 2, {0x2040, 0x3E7E, 0x4890, 0xBFF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {31781, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x70, 0x1F80, 0x200, 0xFFFE, 0x820, 0x37D8, 0xC446, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27D8}},
    {31782, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x90, 0x7EA0, 0x444, 0x2828, 0x17D0, 0x2008, 0xDFF6, 0x1010, 0x1FF0, 0x820, 0x440, 0x7FFC}},
    {31783, 2, {0x2040, 0x3E7E, 0x4890, 0x8440, 0x3FF8, 0x440, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {31784, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7CF8, 0x4488, 0x7CF8, 0x4284, 0x3E7C, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {31785, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x2AA8, 0x820, 0x1450, 0x2288, 0x7FFC, 0x4004, 0x200, 0x3FF8, 0x408, 0x850, 0x3020}},
    {31786, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3E7C, 0x810, 0x4890, 0x7EFE, 0x1428, 0x244A, 0x4686, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8}},
    {31787, 2, {0x2080, 0x3EFE, 0x4910, 0xBFF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x2548, 0x3D78, 0x2108, 0x3FF8, 0x2108, 0x3D78, 0x2548, 0x4548}},
    {31788, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x80, 0x78F8, 0x4908, 0x4A90, 0x7860, 0x1198, 0x1206, 0x5DF8, 0x5108, 0x5108, 0x5DF8, 0xE108}},
    {31789, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x1110, 0x2928, 0x4544, 0xFFFE, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {31790, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x3EF8, 0x820, 0x7EFC, 0x1450, 0x2288, 0x5FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {31791, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1088, 0x1088, 0x13FE, 0xFC88, 0x10F8, 0x3888, 0x34F8, 0x5088, 0x53FE, 0x9088, 0x1104, 0x1202}},
    {31792, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x4840, 0x49FC, 0x7D24, 0x41FC, 0x4144, 0x79FC, 0x48A0, 0x4920, 0x4BFE, 0x4820, 0x8820}},
    {31793, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1048, 0x848, 0x7EFC, 0x1048, 0x1078, 0x1E48, 0x1278, 0x1248, 0x2248, 0x23FE, 0x4A48, 0x8484}},
    {31794, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x20, 0x3C40, 0x25FC, 0x2524, 0x3DFC, 0x2544, 0x25FC, 0x3C50, 0x2490, 0x27FE, 0x5410, 0x8810}},
    {31795, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x810, 0x7F28, 0x844, 0x3E82, 0x227C, 0x3E10, 0x2210, 0x3EFE, 0x810, 0x7F10, 0x810, 0x810}},
    {31796, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2924, 0x2528, 0x5FFE, 0x4210, 0x8408}},
    {31797, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0xFE78, 0x1088, 0x7D50, 0x5420, 0x7CD8, 0x5706, 0x7CF8, 0x1088, 0xFE88, 0x10F8, 0x1088}},
    {31798, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2000, 0x11F8, 0xFC08, 0x8F8, 0x1008, 0x3BFE, 0x5420, 0x9534, 0x10A8, 0x1124, 0x12A2, 0x1040}},
    {31799, 2, {0x2040, 0x3F7E, 0x4890, 0x9FF8, 0x2020, 0x7FFE, 0xA488, 0x2BF4, 0x2000, 0x27F8, 0x2000, 0x27F8, 0x2000, 0x47F8, 0x4408, 0x87F8}},
    {31800, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2410, 0x2410, 0xFEFE, 0x2492, 0x3C90, 0x24FC, 0x3CA4, 0x24A4, 0xFEA8, 0x2910, 0x4528, 0x8246}},
    {31801, 2, {0x2040, 0x3E7E, 0x4890, 0x9110, 0x920, 0x3FFC, 0x2004, 0x4FE8, 0x820, 0xFE0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {31802, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1110, 0x2FE8, 0xC106, 0x1FF0}},
    {31803, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x21F8, 0x1108, 0x11E8, 0x128, 0x73FC, 0x1204, 0x12F4, 0x1294, 0x12F4, 0x1214, 0x2A08, 0x47FE}},
    {31804, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2110, 0x27FC, 0xF910, 0x27FC, 0x2910, 0x37FE, 0xE248, 0x23F8, 0x2248, 0x2FFE, 0xA208, 0x4218}},
    {31805, 2, {0x2040, 0x3F7E, 0x5090, 0x8908, 0x280, 0x440, 0x1830, 0x67CC, 0x0, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x820, 0x1450, 0x2288}},
    {31806, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3FFC, 0x2220, 0x3FFC, 0x2240, 0x2FF8, 0x2248, 0x3FFC, 0x2248, 0x2FF8, 0x4660, 0x4A50, 0xB24C}},
    {31807, 2, {0x2040, 0x3F7E, 0x4890, 0x8548, 0x27FC, 0x1040, 0x13F8, 0x8248, 0x43F8, 0x4A48, 0xBF8, 0x1248, 0xF010, 0x27FC, 0x2210, 0x2130}},
    {31808, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x277C, 0xF4A4, 0x25A4, 0x3654, 0x2088, 0xE7FC, 0x2444, 0x27FC, 0x2444, 0xA7FC, 0x4404}},
    {31809, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x7F3C, 0x844, 0x3E88, 0x2A7C, 0x2A44, 0x3E54, 0x854, 0x1C54, 0x2A28, 0x4844, 0x882}},
    {31810, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x840, 0x1440, 0x2AFE, 0x7F10, 0xA210, 0x3EFE, 0x2292, 0x3E92, 0x289A, 0x2494, 0x2A10, 0x3010}},
    {31811, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {31812, 2, {0x2040, 0x3F7E, 0x4890, 0xBFFC, 0x2080, 0x27F0, 0x2490, 0x27F0, 0x2080, 0x3FFC, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x2220, 0x3FFC}},
    {31813, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0xA00, 0x33B8, 0x2208, 0x3A38, 0x2388, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {31814, 2, {0x2080, 0x3EFE, 0x4910, 0x8820, 0x7FFC, 0x920, 0x1FF8, 0x3100, 0x5FF0, 0x9100, 0x1FFC, 0x1000, 0x3FF0, 0x820, 0x7C0, 0xF83E}},
    {31815, 2, {0x2080, 0x3EFE, 0x4910, 0x8080, 0x3FFC, 0x0, 0xFF0, 0x810, 0x7FFE, 0x4002, 0x9FF4, 0x100, 0x698, 0x39E0, 0x690, 0x398C}},
    {31816, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7F7C, 0x1420, 0x7F20, 0x5538, 0x5528, 0x7F48, 0x868, 0x858, 0x7F4A, 0x84A, 0xF6A, 0xF046}},
    {31817, 2, {0x2080, 0x3EFE, 0x4910, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x7FFE, 0x4822, 0x9FF4, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {31818, 2, {0x2080, 0x3EFE, 0x4910, 0xBEF8, 0x1248, 0xA28, 0x1248, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {31819, 2, {0x2040, 0x3E7E, 0x4890, 0x8000, 0x7FFC, 0x2108, 0x4104, 0xBFFA, 0x2108, 0x3558, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3558}},
    {31820, 2, {0x2080, 0x3EFE, 0x4910, 0xFFFC, 0x100, 0x3FF8, 0x0, 0xFFFE, 0x2, 0x3FF0, 0x100, 0xFFFE, 0x10, 0x7DFC, 0x4490, 0x7C30}},
    {31821, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1048, 0x7CFC, 0x1048, 0x7C48, 0x11FE, 0xFE00, 0x10FC, 0x3884, 0x54FC, 0x9484, 0x10FC, 0x1084}},
    {31822, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x247C, 0x3A04, 0x2228, 0x1E10, 0x20FE, 0x3E12, 0x4850, 0x85E, 0x7F50, 0x14B0, 0x229E, 0x4100}},
    {31823, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2080, 0x11FE, 0xFE88, 0x21FE, 0x2088, 0x3CF8, 0x2488, 0x24F8, 0x2488, 0x45FE, 0x5488, 0x8904}},
    {31824, 2, {0x2040, 0x3E7E, 0x4890, 0x8020, 0x3D24, 0x24A8, 0x25FC, 0x3C40, 0x27FE, 0x2488, 0x3D24, 0x26AA, 0x2470, 0x44A8, 0x5524, 0x8A62}},
    {31825, 2, {0x2040, 0x3E7E, 0x4890, 0x9FF0, 0x1490, 0x1250, 0x1FF0, 0x200, 0xFFFE, 0x920, 0x3FF8, 0xCA26, 0x920, 0xFE8, 0x8D0, 0xE38}},
    {31826, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x27FE, 0x20A0, 0xF7BC, 0x24A4, 0x37BC, 0x20A0, 0x67FC, 0xA444, 0x27FC, 0x2444, 0xA7FC, 0x4404}},
    {31827, 2, {0x2080, 0x3EFE, 0x4910, 0xA038, 0x17C0, 0x1248, 0x8150, 0x47FC, 0x5248, 0x1404, 0x2BFA, 0xE248, 0x23F8, 0x2248, 0x23F8, 0x2208}},
    {31828, 2, {0x2040, 0x3E7E, 0x4890, 0x8800, 0x7F20, 0x4920, 0xFFBE, 0x4944, 0x7FA4, 0x2A24, 0x3E28, 0x1028, 0xFF10, 0x2228, 0x1C44, 0x6282}},
    {31829, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x11BC, 0x1114, 0xFD54, 0x11AC, 0x3040, 0x39FC, 0x5524, 0x51FC, 0x9124, 0x11FC, 0x1104}},
    {31830, 2, {0x2040, 0x3E7E, 0x4890, 0x8420, 0x3F28, 0x424, 0xFFFE, 0x1220, 0x73A4, 0x1224, 0x73A8, 0x1228, 0x7392, 0x121A, 0x17A6, 0xF842}},
    {31831, 2, {0x2040, 0x3E7E, 0x4890, 0xBFF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC}},
    {31832, 2, {0x2040, 0x3E7E, 0x4890, 0x8124, 0x3CA8, 0x25FC, 0x2440, 0x3DFE, 0x2488, 0x2544, 0x3E8A, 0x24F0, 0x2424, 0x45FC, 0x54A8, 0x8964}},
    {31833, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1000, 0x29F8, 0x4408, 0xB8F8, 0x1008, 0x7DFE, 0x1020, 0x9534, 0x50A8, 0x1D24, 0xE2A2, 0x4040}},
    {31834, 2, {0x2080, 0x3EFE, 0x4910, 0x81F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2490, 0x27F0, 0x2490, 0x4FF8, 0x4948, 0x9FFC}},
    {31835, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1090, 0x1088, 0x2BFE, 0x4448, 0xB832, 0x11CE, 0x7C90, 0x1088, 0x97FE, 0x5048, 0x1C32, 0xE1CE}},
    {31836, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x11FC, 0x1154, 0xFDFC, 0x1020, 0x15FC, 0x1820, 0x33FE, 0xD088, 0x11FC, 0x1020, 0x53FE, 0x2020}},
    {31837, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x1000, 0x1FF8, 0x820, 0xFE0, 0x7938, 0x4FA8, 0x7AA8, 0x4AB8, 0x792A, 0x4AAA, 0x9C46}},
    {31838, 2, {0x2040, 0x3E7E, 0x4A90, 0xA7DE, 0x4912, 0x97D2, 0x2112, 0x65DA, 0xA514, 0x2FD0, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1110, 0x2308}},
    {31839, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0x1080, 0xFEF8, 0x1108, 0x7EFC, 0x5484, 0x54FC, 0x7C84, 0x10FC, 0x3884, 0x54FC, 0x9048, 0x1084}},
    {31840, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0x840, 0x7E7C, 0x2440, 0xFF7C, 0x4, 0x7E7C, 0x4240, 0x7E7C, 0x4240, 0x7E7C, 0x4242, 0x463E}},
    {31841, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x1040, 0x10A0, 0xFD10, 0x12E8, 0x1406, 0x1BB8, 0x32A8, 0xD3B8, 0x1110, 0x1110, 0x52A8, 0x2444}},
    {31842, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x1420, 0x223E, 0x5D44, 0x80A4, 0x7724, 0x5528, 0x7728, 0x2210, 0x2228, 0x5544, 0x8882}},
    {31843, 2, {0x2040, 0x3F7E, 0x4890, 0xFD7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x5FF4, 0x4104, 0x5FF4, 0x5554, 0x5FF4, 0x4544, 0x592C}},
    {31844, 2, {0x2040, 0x3E7E, 0x4890, 0x9120, 0x1128, 0x2AA4, 0xFFFE, 0x1220, 0x73A4, 0x1224, 0x73A8, 0x1228, 0x7392, 0x121A, 0x17A6, 0xF842}},
    {31845, 2, {0x2080, 0x3EFE, 0x4910, 0x8280, 0xC60, 0x37D8, 0xC006, 0x3BB8, 0x2AA8, 0x3BB8, 0x0, 0x3FF8, 0x2488, 0x3FF8, 0x2488, 0x24B8}},
    {31846, 2, {0x2040, 0x3F7E, 0x4890, 0x85FC, 0x1020, 0x2BFE, 0x4420, 0xB9FC, 0x1124, 0x7DFC, 0x1124, 0x95FC, 0x5020, 0x1DFE, 0xE020, 0x47FE}},
    {31847, 2, {0x2040, 0x3F7E, 0x4890, 0x85F8, 0x4080, 0x2FFC, 0x2884, 0xBF0, 0xE888, 0x2BF8, 0x2940, 0x2AA8, 0x3170, 0x22A8, 0x5160, 0x8FFE}},
    {31848, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x820, 0x1420, 0x223E, 0x5D42, 0x8084, 0x7710, 0x5510, 0x7710, 0x2228, 0x2228, 0x5544, 0x8882}},
    {31849, 2, {0x2040, 0x3E7E, 0x4890, 0x87F8, 0x4548, 0x24A8, 0x27F8, 0x40, 0xEFFC, 0x2A14, 0x2488, 0x2BF4, 0x2100, 0x21F0, 0x5210, 0x8FFE}},
    {31850, 2, {0x2080, 0x3EFE, 0x4910, 0x8882, 0x511C, 0x5DD0, 0x4890, 0x5DD0, 0x445E, 0x7FF4, 0x5114, 0x5DD4, 0x4894, 0x5DD4, 0x4454, 0x7FE4}},
    {31851, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x2850, 0x3E7C, 0x4890, 0xFEFE, 0x1528, 0x264A, 0x5FF6, 0x1490, 0x1250, 0x1FF0, 0x820, 0x1010}},
    {31852, 2, {0x2040, 0x3F7E, 0x4890, 0x9508, 0x828, 0x7F24, 0x147E, 0x4948, 0x55C8, 0x7F7E, 0x848, 0x7F7E, 0x5148, 0x5548, 0x5D7E, 0x4340}},
    {31853, 2, {0x2040, 0x3E7E, 0x4890, 0xBF7E, 0x1224, 0x1B34, 0x12A4, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {31854, 2, {0x2040, 0x3E7E, 0x4890, 0xFFFC, 0x4444, 0x7FFC, 0x10A0, 0x2090, 0x45FE, 0x7920, 0x13FC, 0x2520, 0x7DFC, 0x120, 0x55FE, 0x8100}},
    {31855, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x7FFC, 0x1000, 0x1FF8, 0x820, 0xFE0, 0x7BB8, 0x4AA8, 0x7BA8, 0x4AB8, 0x7BAA, 0x492A, 0x9AC6}},
    {31856, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3FF8, 0x3100, 0x5FF0, 0x9100, 0x1FFC, 0x820, 0x7C0, 0xF83E}},
    {31857, 2, {0x2040, 0x3E7E, 0x4890, 0xBFF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x4890, 0x7EFC, 0xC990, 0x7EFC, 0x4890, 0x7EFC}},
    {31858, 2, {0x2040, 0x3F7E, 0x4890, 0x8508, 0x18FE, 0x2410, 0x7E7C, 0x8144, 0x7F7C, 0x5544, 0x7F7C, 0x44, 0x7F7C, 0x5528, 0x7F44, 0x5582}},
    {31859, 2, {0x100, 0x2108, 0x1108, 0x910, 0x920, 0x100, 0x7FFC, 0x380, 0x540, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100, 0x100}},
    {31860, 2, {0x600, 0x100, 0x280, 0x440, 0x1830, 0xE10E, 0x1110, 0x920, 0x540, 0x7FFC, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {31861, 2, {0x810, 0x810, 0x4A10, 0x2A10, 0x2C10, 0x810, 0xFEFE, 0x1810, 0x1C10, 0x2A10, 0x2A10, 0x4810, 0x8810, 0x810, 0x810, 0x810}},
    {31862, 2, {0x800, 0x800, 0x4AF0, 0x2A90, 0x2C90, 0x890, 0xFE90, 0x1890, 0x1C90, 0x2A90, 0x2A90, 0x4892, 0x8892, 0x912, 0x90E, 0xA00}},
    {31863, 2, {0x808, 0x81C, 0x4AE0, 0x2A20, 0x2C20, 0x820, 0xFE20, 0x183E, 0x1DE0, 0x2A20, 0x2A20, 0x4822, 0x8822, 0x822, 0x81E, 0x800}},
    {31864, 2, {0x1000, 0x13F0, 0x9490, 0x5490, 0x5890, 0x1090, 0xFC90, 0x33F0, 0x3890, 0x5490, 0x5490, 0x9092, 0x108A, 0x108A, 0x1086, 0x1082}},
    {31865, 2, {0x1040, 0x1040, 0x9440, 0x5440, 0x5BFE, 0x1088, 0xFC88, 0x3088, 0x3888, 0x5508, 0x5490, 0x9050, 0x1020, 0x1050, 0x1088, 0x1304}},
    {31866, 2, {0x840, 0x840, 0x4A40, 0x2AFE, 0x2C80, 0x900, 0xFE7C, 0x1808, 0x1C10, 0x2A20, 0x2A40, 0x4880, 0x8902, 0x902, 0x8FE, 0x800}},
    {31867, 2, {0x100, 0x1110, 0x920, 0x100, 0x7FFC, 0x540, 0x920, 0x1110, 0x6008, 0x100, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {31868, 2, {0x1020, 0x1020, 0x9420, 0x5420, 0x5924, 0x1124, 0xFD24, 0x3124, 0x3924, 0x5524, 0x5524, 0x9124, 0x1124, 0x11FC, 0x1004, 0x1000}},
    {31869, 2, {0x1000, 0x11FC, 0x9404, 0x5408, 0x5810, 0x1020, 0xFC20, 0x33FE, 0x3820, 0x5420, 0x5420, 0x9020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {31870, 2, {0x1000, 0x1000, 0x95FC, 0x5444, 0x5844, 0x1144, 0xFD44, 0x3144, 0x3944, 0x5644, 0x5444, 0x9084, 0x1084, 0x1104, 0x1228, 0x1410}},
    {31871, 2, {0x808, 0x808, 0x4A08, 0x2A08, 0x2CFE, 0x808, 0xFE08, 0x1808, 0x1C88, 0x2A48, 0x2A48, 0x4808, 0x8808, 0x808, 0x828, 0x810}},
    {31872, 2, {0x1010, 0x1010, 0x9410, 0x5410, 0x5BFE, 0x1010, 0xFC10, 0x3110, 0x3910, 0x5490, 0x5490, 0x9050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {31873, 2, {0x808, 0x81C, 0x4AE0, 0x2A20, 0x2C20, 0x820, 0xFE20, 0x19FE, 0x1C20, 0x2A20, 0x2A20, 0x4820, 0x8820, 0x820, 0x820, 0x820}},
    {31874, 2, {0x400, 0x400, 0xFE0, 0x1040, 0x2080, 0x360, 0xC18, 0x7004, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31875, 2, {0x1010, 0x1110, 0x9510, 0x5512, 0x5912, 0x1114, 0xFDD8, 0x3110, 0x3910, 0x5510, 0x5510, 0x9112, 0x1152, 0x1192, 0x110E, 0x1000}},
    {31876, 2, {0x1008, 0x101C, 0x95E0, 0x5500, 0x5900, 0x11FC, 0xFD44, 0x3144, 0x3944, 0x5528, 0x5528, 0x9110, 0x1110, 0x1228, 0x1244, 0x1482}},
    {31877, 2, {0x1080, 0x1080, 0x9480, 0x54FC, 0x5954, 0x1254, 0xFC54, 0x3094, 0x3894, 0x5524, 0x5624, 0x9044, 0x1044, 0x1084, 0x1128, 0x1010}},
    {31878, 2, {0x1020, 0x1020, 0x9420, 0x54A8, 0x58A4, 0x10A2, 0xFD22, 0x3120, 0x3A24, 0x5424, 0x5428, 0x9008, 0x1010, 0x1020, 0x10C0, 0x1300}},
    {31879, 2, {0x1080, 0x1040, 0x9440, 0x57FC, 0x5800, 0x1000, 0xFDF0, 0x3110, 0x3910, 0x5510, 0x5510, 0x9112, 0x1112, 0x1212, 0x120E, 0x1400}},
    {31880, 2, {0x1000, 0x11F8, 0x9448, 0x5448, 0x5848, 0x1048, 0xFC48, 0x31F8, 0x3888, 0x5488, 0x5488, 0x9088, 0x1088, 0x1088, 0x13FE, 0x1000}},
    {31881, 2, {0x1010, 0x1050, 0x9450, 0x5450, 0x5888, 0x1088, 0xFD04, 0x32FA, 0x3848, 0x5448, 0x5448, 0x9048, 0x1088, 0x1088, 0x1128, 0x1210}},
    {31882, 2, {0x2080, 0x2088, 0x3CB0, 0x20C0, 0x2084, 0x2C84, 0x307C, 0x0, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31883, 2, {0x1040, 0x1040, 0x95F8, 0x5448, 0x5888, 0x108A, 0xFD0A, 0x3226, 0x3820, 0x5420, 0x57FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31884, 2, {0x1004, 0x11E4, 0x9424, 0x5424, 0x5824, 0x11E4, 0xFD04, 0x3104, 0x3904, 0x55E4, 0x5424, 0x9024, 0x1024, 0x1024, 0x1144, 0x1084}},
    {31885, 2, {0x1008, 0x103C, 0x95E0, 0x5420, 0x5820, 0x103C, 0xFDE0, 0x3020, 0x3820, 0x543E, 0x57E0, 0x9020, 0x1022, 0x1022, 0x1022, 0x101E}},
    {31886, 2, {0x800, 0x8FC, 0x4A84, 0x2A84, 0x2C84, 0x884, 0xFEFC, 0x18A0, 0x1CA0, 0x2AA0, 0x2A90, 0x4890, 0x8908, 0x908, 0xA04, 0xC02}},
    {31887, 2, {0x820, 0x820, 0x4A20, 0x2A20, 0x2C20, 0x9FE, 0xFE20, 0x1820, 0x1C20, 0x2A50, 0x2A50, 0x4850, 0x8888, 0x8C8, 0x924, 0xA02}},
    {31888, 2, {0x820, 0x810, 0x4A10, 0x2AFE, 0x2C82, 0x882, 0xFE82, 0x18FE, 0x1C82, 0x2A80, 0x2A80, 0x4880, 0x8880, 0x900, 0x900, 0xA00}},
    {31889, 2, {0x1000, 0x1000, 0x95FC, 0x5524, 0x5924, 0x1124, 0xFD24, 0x3124, 0x39FC, 0x5500, 0x5500, 0x9100, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {31890, 2, {0x820, 0x810, 0x4A10, 0x2A00, 0x2DFE, 0x800, 0xFE04, 0x1884, 0x1C84, 0x2A48, 0x2A48, 0x4848, 0x8850, 0x810, 0x9FE, 0x800}},
    {31891, 2, {0x888, 0x888, 0x4A88, 0x2A88, 0x2DFE, 0x888, 0xFE88, 0x1888, 0x1C88, 0x2AF8, 0x2A88, 0x4888, 0x8888, 0x888, 0x8F8, 0x888}},
    {31892, 2, {0x800, 0x8FE, 0x4A80, 0x2A80, 0x2C80, 0x8FC, 0xFE84, 0x1884, 0x1C84, 0x2A84, 0x2AFC, 0x4880, 0x8880, 0x880, 0x8FE, 0x800}},
    {31893, 2, {0x1020, 0x1020, 0x9440, 0x55FC, 0x5904, 0x1104, 0xFD04, 0x3104, 0x39FC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31894, 2, {0x1020, 0x1020, 0x9420, 0x55FE, 0x5820, 0x1020, 0xFC20, 0x31FC, 0x3870, 0x54A8, 0x54A8, 0x9124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31895, 2, {0x1000, 0x11F8, 0x9508, 0x5508, 0x5908, 0x11F8, 0xFD08, 0x3108, 0x3908, 0x55F8, 0x5508, 0x9108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {31896, 2, {0x1020, 0x1020, 0x9420, 0x5420, 0x583E, 0x1020, 0xFC20, 0x3020, 0x39FC, 0x5504, 0x5504, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31897, 2, {0x1020, 0x1020, 0x9420, 0x5420, 0x59FC, 0x1124, 0xFD24, 0x3124, 0x3924, 0x55FC, 0x5524, 0x9124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {31898, 2, {0x1080, 0x1080, 0x94FE, 0x5500, 0x5A20, 0x1120, 0xFD2C, 0x3174, 0x3BA4, 0x5524, 0x5534, 0x9128, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {31899, 2, {0x100, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x2928, 0x2548, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x4108, 0x8108}},
    {31900, 2, {0x100, 0x1110, 0x1110, 0x1FF0, 0x4104, 0x4104, 0x7FFC, 0x0, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31901, 2, {0x1000, 0x11FE, 0x9500, 0x5500, 0x5900, 0x11FE, 0xFD20, 0x3120, 0x393C, 0x5524, 0x5524, 0x9124, 0x1124, 0x1244, 0x1254, 0x1488}},
    {31902, 2, {0x1000, 0x13FE, 0x9450, 0x5450, 0x59FC, 0x1154, 0xFD54, 0x3154, 0x3954, 0x5554, 0x558C, 0x9104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31903, 2, {0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31904, 2, {0x1088, 0x1088, 0x9488, 0x5488, 0x59FE, 0x1088, 0xFC88, 0x3088, 0x3888, 0x57FE, 0x5400, 0x9088, 0x1084, 0x1104, 0x1202, 0x1402}},
    {31905, 2, {0x1000, 0x11FE, 0x9502, 0x5502, 0x597A, 0x1102, 0xFD02, 0x317A, 0x394A, 0x554A, 0x554A, 0x917A, 0x1102, 0x1102, 0x110A, 0x1104}},
    {31906, 2, {0x100, 0x4100, 0x21FC, 0xA44, 0x1448, 0xE0A0, 0x2110, 0x2608, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31907, 2, {0x1000, 0x11DC, 0x9554, 0x5554, 0x5954, 0x1154, 0xFD54, 0x33FE, 0x3954, 0x5554, 0x5554, 0x9154, 0x1154, 0x12D4, 0x1224, 0x144C}},
    {31908, 2, {0x200, 0x3FF8, 0x2928, 0x2548, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x3FF8, 0x0, 0xFFFE, 0x1000, 0x1FF8, 0x8, 0x50, 0x20}},
    {31909, 2, {0x100, 0xF93C, 0x904, 0xD44, 0xB84, 0x793C, 0x4120, 0x47E0, 0x4120, 0x793C, 0xB84, 0xD44, 0x904, 0x904, 0x5128, 0x2110}},
    {31910, 2, {0x100, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x1010, 0x3EFC, 0x4210, 0xA290, 0x1490, 0x8FE, 0x3010, 0xC010}},
    {31911, 2, {0x1020, 0x1010, 0x9410, 0x55FE, 0x5900, 0x1110, 0xFD10, 0x3110, 0x397C, 0x5510, 0x5510, 0x9110, 0x1110, 0x1210, 0x12FE, 0x1400}},
    {31912, 2, {0x800, 0x9FE, 0x4A20, 0x2A20, 0x2C40, 0x8FC, 0xFE84, 0x1884, 0x1C84, 0x2AFC, 0x2A84, 0x4884, 0x8884, 0x884, 0x8FC, 0x884}},
    {31913, 2, {0x1040, 0x1040, 0x9444, 0x55F4, 0x5848, 0x1050, 0xFDFE, 0x3040, 0x3880, 0x5584, 0x5698, 0x90E0, 0x1082, 0x1082, 0x107E, 0x1000}},
    {31914, 2, {0x100, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC446, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {31915, 2, {0x1000, 0x13FE, 0x9420, 0x5420, 0x5840, 0x11FC, 0xFD54, 0x3154, 0x3954, 0x5554, 0x5554, 0x9154, 0x1154, 0x1144, 0x1114, 0x1108}},
    {31916, 2, {0x1050, 0x1050, 0x9450, 0x5450, 0x59FC, 0x1154, 0xFD54, 0x3154, 0x3954, 0x55FC, 0x5554, 0x9154, 0x1154, 0x1154, 0x11FC, 0x1104}},
    {31917, 2, {0x820, 0x820, 0x4A50, 0x2A48, 0x2C84, 0x902, 0xFEFC, 0x1800, 0x1C00, 0x2AFC, 0x2A84, 0x4884, 0x8884, 0x884, 0x8FC, 0x884}},
    {31918, 2, {0x1040, 0x1020, 0x95FC, 0x5504, 0x5904, 0x11FC, 0xFD04, 0x3104, 0x39FC, 0x5520, 0x5522, 0x9114, 0x1108, 0x1144, 0x1182, 0x1100}},
    {31919, 2, {0x1000, 0x11F8, 0x9508, 0x5508, 0x59F8, 0x1108, 0xFD08, 0x31F8, 0x3908, 0x5508, 0x55F8, 0x9090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {31920, 2, {0x1008, 0x103C, 0x95C0, 0x5404, 0x5944, 0x10A8, 0xFC00, 0x31F8, 0x3810, 0x5420, 0x57FE, 0x9020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {31921, 2, {0x0, 0x23F0, 0x9090, 0x4294, 0x1292, 0xE512, 0x2250, 0x2420, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {31922, 2, {0x800, 0xF7C, 0x2804, 0x3F44, 0x6128, 0x9210, 0xC68, 0x3184, 0xC000, 0x1110, 0x920, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106}},
    {31923, 2, {0x1000, 0x11FE, 0x9420, 0x5420, 0x59FC, 0x1124, 0xFD24, 0x31FC, 0x3924, 0x5524, 0x55FC, 0x9120, 0x10A0, 0x1040, 0x10B0, 0x130E}},
    {31924, 2, {0x1000, 0x11FC, 0x9524, 0x5524, 0x59FC, 0x1124, 0xFD24, 0x31FC, 0x3820, 0x5420, 0x55FC, 0x9020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {31925, 2, {0x400, 0x3FF8, 0x20C8, 0x2F28, 0x2548, 0x3FF8, 0x2548, 0x2928, 0x3FF8, 0x0, 0xFFFE, 0x1000, 0x1FF8, 0x8, 0x50, 0x20}},
    {31926, 2, {0x1000, 0x11F8, 0x9408, 0x5408, 0x59F8, 0x1008, 0xFC08, 0x33FE, 0x3820, 0x5622, 0x5574, 0x90A8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {31927, 2, {0x1080, 0x1080, 0x94FE, 0x5502, 0x5A42, 0x124A, 0xFD52, 0x3042, 0x3BFE, 0x5442, 0x54E2, 0x9152, 0x124A, 0x1042, 0x1054, 0x1008}},
    {31928, 2, {0x1088, 0x1088, 0x97FE, 0x5488, 0x5888, 0x10F8, 0xFC88, 0x3088, 0x38F8, 0x5488, 0x5488, 0x93FE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {31929, 2, {0x1040, 0x1020, 0x95FE, 0x5400, 0x5888, 0x1088, 0xFD54, 0x3222, 0x3800, 0x5420, 0x55FE, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {31930, 2, {0x1040, 0x1080, 0x95FC, 0x5524, 0x5924, 0x11FC, 0xFD24, 0x3144, 0x39FC, 0x5490, 0x5510, 0x93FE, 0x1010, 0x1010, 0x1010, 0x1010}},
    {31931, 2, {0x1000, 0x10FC, 0x9480, 0x54F8, 0x5880, 0x10F8, 0xFC80, 0x33FE, 0x3940, 0x5524, 0x5528, 0x9110, 0x1108, 0x1144, 0x1182, 0x1100}},
    {31932, 2, {0x800, 0x4912, 0x2A12, 0x812, 0xFF24, 0x2A24, 0x4948, 0x8890, 0x4248, 0x7748, 0x9224, 0x5A24, 0x2F92, 0x2212, 0x4212, 0x8200}},
    {31933, 2, {0x2040, 0x2020, 0x23FE, 0xAA02, 0x7000, 0x21FC, 0xF800, 0x2000, 0x73FE, 0x6820, 0xA928, 0x2124, 0x2222, 0x2422, 0x20A0, 0x2040}},
    {31934, 2, {0x1020, 0x1020, 0x95FE, 0x5420, 0x58FC, 0x1020, 0xFDFE, 0x3000, 0x38FC, 0x5484, 0x54FC, 0x9084, 0x10FC, 0x1084, 0x1094, 0x1088}},
    {31935, 2, {0x1000, 0x11FC, 0x9524, 0x5524, 0x59FC, 0x1124, 0xFD24, 0x31FC, 0x3820, 0x57FE, 0x5470, 0x90A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {31936, 2, {0x1088, 0x1088, 0x95FE, 0x5488, 0x5888, 0x1050, 0xFC50, 0x3094, 0x3994, 0x5698, 0x5490, 0x90B2, 0x10D2, 0x1092, 0x108E, 0x1080}},
    {31937, 2, {0x1040, 0x1048, 0x9484, 0x55FE, 0x5820, 0x13FE, 0xFC88, 0x3124, 0x3A42, 0x5588, 0x5410, 0x9062, 0x1184, 0x1018, 0x1060, 0x1380}},
    {31938, 2, {0x1088, 0x1088, 0x97FE, 0x5488, 0x58F8, 0x1088, 0xFCF8, 0x3088, 0x3888, 0x57FE, 0x5500, 0x9148, 0x1184, 0x1100, 0x11FE, 0x1000}},
    {31939, 2, {0x1000, 0x10FC, 0x9484, 0x54FC, 0x5884, 0x10FC, 0xFC00, 0x31FE, 0x3880, 0x54FE, 0x552A, 0x924A, 0x1092, 0x1122, 0x1054, 0x1088}},
    {31940, 2, {0x1040, 0x1020, 0x95FE, 0x5502, 0x5902, 0x11FE, 0xFD00, 0x3100, 0x39FE, 0x55AA, 0x55AA, 0x92FE, 0x12AA, 0x12AA, 0x14A2, 0x1086}},
    {31941, 2, {0x2000, 0x21FC, 0x2048, 0xA830, 0x73FE, 0x2052, 0xF894, 0x2110, 0x7230, 0x6820, 0xABFE, 0x2070, 0x20A8, 0x2124, 0x2622, 0x2020}},
    {31942, 2, {0x1000, 0x13FE, 0x9420, 0x5440, 0x59FC, 0x1154, 0xFD54, 0x3174, 0x3954, 0x5554, 0x5574, 0x9154, 0x1154, 0x1154, 0x11FC, 0x1104}},
    {31943, 2, {0x1080, 0x10BC, 0x9484, 0x5508, 0x59FE, 0x1320, 0xFD20, 0x317C, 0x3990, 0x5510, 0x55FE, 0x9110, 0x1128, 0x1128, 0x1144, 0x1182}},
    {31944, 2, {0x2000, 0x23FE, 0x2022, 0xA920, 0x713C, 0x2120, 0xFAFE, 0x2400, 0x71FC, 0x6904, 0xA9FC, 0x2104, 0x21FC, 0x2104, 0x2114, 0x2108}},
    {31945, 2, {0x2000, 0x2088, 0x2252, 0xAA22, 0x7252, 0x228A, 0xFA02, 0x23FE, 0x7088, 0x6944, 0xAA7A, 0x2088, 0x2150, 0x2020, 0x20D8, 0x2706}},
    {31946, 2, {0x1080, 0x109E, 0x9492, 0x55D2, 0x5892, 0x109E, 0xFC92, 0x31D2, 0x3952, 0x555E, 0x5552, 0x9152, 0x11D2, 0x1022, 0x102A, 0x1044}},
    {31947, 2, {0x2108, 0x2088, 0x2090, 0xABFE, 0x7000, 0x23C4, 0xFA54, 0x2254, 0x73D4, 0x6A54, 0xAA54, 0x23D4, 0x2254, 0x2244, 0x2254, 0x22C8}},
    {31948, 2, {0x2088, 0x2088, 0x21EC, 0xA92A, 0x7228, 0x2548, 0xF888, 0x217E, 0x7600, 0x69FC, 0xA904, 0x2104, 0x21FC, 0x2104, 0x2104, 0x21FC}},
    {31949, 2, {0x2108, 0x2088, 0x2090, 0xAFFE, 0x7108, 0x2108, 0xFA10, 0x2294, 0x74A4, 0x6F38, 0xA908, 0x2210, 0x2210, 0x24A4, 0x27BC, 0x2084}},
    {31950, 2, {0x1000, 0x11FE, 0x9500, 0x557C, 0x5954, 0x1154, 0xFD7C, 0x3154, 0x3954, 0x557C, 0x5510, 0x9110, 0x117C, 0x1210, 0x1210, 0x14FE}},
    {31951, 2, {0x2000, 0x23FE, 0x2202, 0xAA02, 0x73FE, 0x2210, 0xFA92, 0x2254, 0x72FE, 0x6A82, 0xAAFE, 0x2282, 0x22FE, 0x2482, 0x248A, 0x2884}},
    {31952, 2, {0x1028, 0x1024, 0x97FE, 0x5420, 0x59FC, 0x1124, 0xFDFC, 0x3124, 0x39FC, 0x5524, 0x5408, 0x93FE, 0x1088, 0x1048, 0x1048, 0x1018}},
    {31953, 2, {0x2000, 0x23DE, 0x2042, 0xA842, 0x73DE, 0x2210, 0xFA10, 0x23DE, 0x7042, 0x6A52, 0xA94A, 0x214A, 0x2252, 0x2042, 0x2294, 0x2108}},
    {31954, 2, {0x1088, 0x1088, 0x97FE, 0x5488, 0x5888, 0x13FE, 0xFC80, 0x31FC, 0x3B24, 0x55FC, 0x5524, 0x91FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {31955, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x1044, 0x5444, 0x3828, 0xFE28, 0x1010, 0x3828, 0x5444, 0x9182}},
    {31956, 2, {0x1000, 0x11FC, 0x94A4, 0x5488, 0x5850, 0x1020, 0xFCD8, 0x3326, 0x3820, 0x55FC, 0x5524, 0x9124, 0x11FC, 0x1022, 0x13FE, 0x1002}},
    {31957, 2, {0x1104, 0x1084, 0x9488, 0x5400, 0x5BFE, 0x1020, 0xFC20, 0x31FC, 0x3820, 0x5420, 0x57FE, 0x9000, 0x12A4, 0x1252, 0x1452, 0x1000}},
    {31958, 2, {0x1020, 0x1010, 0x95FE, 0x5510, 0x597C, 0x1114, 0xFDFE, 0x3114, 0x397C, 0x5510, 0x557C, 0x9144, 0x1144, 0x1244, 0x127C, 0x1444}},
    {31959, 2, {0x2040, 0x2080, 0x23F8, 0xAA08, 0x73F8, 0x2208, 0xFBF8, 0x2208, 0x73F8, 0x6850, 0xA848, 0x27FE, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {31960, 2, {0x1040, 0x1020, 0x95FE, 0x5502, 0x5800, 0x10FC, 0xFC40, 0x30A2, 0x3B34, 0x5458, 0x5494, 0x9334, 0x1052, 0x1090, 0x1350, 0x1020}},
    {31961, 2, {0x1010, 0x1050, 0x9650, 0x557C, 0x5890, 0x1010, 0xFCFE, 0x3300, 0x397C, 0x5544, 0x5544, 0x917C, 0x1144, 0x1100, 0x12FE, 0x1400}},
    {31962, 2, {0x1048, 0x1048, 0x95FE, 0x5448, 0x5800, 0x1048, 0xFD48, 0x3148, 0x39FE, 0x5448, 0x5448, 0x93C8, 0x1148, 0x1148, 0x117E, 0x1240}},
    {31963, 2, {0x1020, 0x1124, 0x94A8, 0x57FE, 0x5A02, 0x10F8, 0xFC88, 0x3088, 0x38F8, 0x5420, 0x5420, 0x91FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {31964, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2638, 0x2B54, 0x3292, 0x2080, 0x2490, 0x22A0, 0x3FFC, 0x21C0, 0x42A0, 0x4498, 0x9886}},
    {31965, 2, {0x2040, 0x2088, 0x21FC, 0xA908, 0x7252, 0x23FE, 0xF850, 0x2188, 0x7626, 0x68C0, 0xAB10, 0x2064, 0x2388, 0x2030, 0x20C0, 0x2700}},
    {31966, 2, {0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0x7FFC, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010}},
    {31967, 2, {0x2090, 0x2090, 0x27FE, 0xA890, 0x73FC, 0x2294, 0xFBFC, 0x2294, 0x73FC, 0x6800, 0xA9F8, 0x2108, 0x21F8, 0x2108, 0x21F8, 0x2108}},
    {31968, 2, {0x1020, 0x1010, 0x95FE, 0x5510, 0x597C, 0x1114, 0xFDFE, 0x3114, 0x397C, 0x5510, 0x5592, 0x9154, 0x1138, 0x1254, 0x1292, 0x1430}},
    {31969, 2, {0x1040, 0x1020, 0x95FC, 0x5488, 0x5850, 0x13FE, 0xFC00, 0x31FC, 0x3904, 0x55FC, 0x5504, 0x91FC, 0x1090, 0x1092, 0x1112, 0x160E}},
    {31970, 2, {0x1088, 0x1088, 0x97FE, 0x5488, 0x5800, 0x11FC, 0xFD04, 0x31FC, 0x3904, 0x55FC, 0x5420, 0x93FE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {31971, 2, {0x2000, 0x27BC, 0x2108, 0xAD28, 0x77BC, 0x2318, 0xFDAA, 0x2946, 0x7000, 0x6BF8, 0xAA08, 0x2208, 0x23F8, 0x2208, 0x2208, 0x23F8}},
    {31972, 2, {0x2288, 0x2288, 0x27C8, 0xAA90, 0x729E, 0x27D4, 0xF824, 0x27D4, 0x7454, 0x6FD4, 0xAC54, 0x27C8, 0x2448, 0x2454, 0x2464, 0x24C2}},
    {31973, 2, {0x1000, 0x11FE, 0x9420, 0x55FE, 0x5952, 0x1152, 0xFD52, 0x3106, 0x3800, 0x55FE, 0x5420, 0x91FE, 0x1152, 0x1152, 0x1152, 0x1106}},
    {31974, 2, {0x1020, 0x13FE, 0x9420, 0x55FC, 0x5800, 0x11FC, 0xFD04, 0x31FC, 0x3888, 0x57FE, 0x5400, 0x91FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {31975, 2, {0x1000, 0x11FC, 0x9504, 0x55FC, 0x5904, 0x13FE, 0xFC00, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x91FC, 0x1020, 0x11FC, 0x1020, 0x13FE}},
    {31976, 2, {0x2000, 0x23BE, 0x20A2, 0xA8A2, 0x70BE, 0x2388, 0xFA08, 0x223E, 0x722A, 0x6BAA, 0xA8AA, 0x20BE, 0x2088, 0x208A, 0x257E, 0x2202}},
    {31977, 2, {0x1020, 0x1050, 0x9488, 0x5574, 0x5A02, 0x11FC, 0xFD24, 0x31AC, 0x3924, 0x55FC, 0x5400, 0x90F8, 0x1088, 0x10F8, 0x1088, 0x10F8}},
    {31978, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3C20, 0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {31979, 2, {0x2000, 0x23FC, 0x2294, 0xAA94, 0x73FC, 0x2000, 0xFBFE, 0x2000, 0x73FC, 0x6A04, 0xABFC, 0x20A2, 0x2114, 0x2308, 0x2D44, 0x2182}},
    {31980, 2, {0x2000, 0x27FC, 0x24A4, 0xACA4, 0x77FC, 0x2048, 0xFBF8, 0x2050, 0x77FE, 0x6840, 0xA9F8, 0x2308, 0x25F8, 0x2108, 0x21F8, 0x2108}},
    {31981, 2, {0x2020, 0x2088, 0x2252, 0xAB26, 0x7252, 0x228A, 0xFA22, 0x23FE, 0x7088, 0x6944, 0xAA7A, 0x2088, 0x2150, 0x2020, 0x20D8, 0x2706}},
    {31982, 2, {0x2010, 0x27D0, 0x2510, 0xAFDE, 0x7450, 0x27E8, 0xFD04, 0x27C4, 0x7000, 0x6BF8, 0xAAA8, 0x22A8, 0x22A8, 0x22A8, 0x2FFE, 0x2000}},
    {31983, 2, {0x1000, 0x11FC, 0x9420, 0x57FE, 0x5A22, 0x11AC, 0xFC20, 0x31AC, 0x3800, 0x57FE, 0x5420, 0x91FC, 0x1154, 0x1154, 0x1154, 0x110C}},
    {31984, 2, {0x23FE, 0x2222, 0x23FE, 0xAA22, 0x72FA, 0x22AA, 0xFAFA, 0x22AA, 0x72FA, 0x6A26, 0xABFE, 0x220A, 0x23FE, 0x228A, 0x225A, 0x23FE}},
    {31985, 2, {0x5210, 0x7EFE, 0x2844, 0x7E28, 0xA2FE, 0x3E10, 0x2010, 0x3E7C, 0x2210, 0x3E10, 0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {31986, 2, {0x2000, 0x23FE, 0x2248, 0xABFE, 0x7248, 0x22FC, 0xFAA4, 0x22FC, 0x72A4, 0x6AFC, 0xAA20, 0x23FE, 0x232A, 0x257A, 0x250A, 0x2906}},
    {31987, 2, {0x553C, 0x1424, 0xFFA4, 0x2242, 0x7F00, 0x83C, 0xFF84, 0x828, 0x6B10, 0x4928, 0x7F44, 0x1010, 0x920, 0xFFFE, 0x920, 0x711C}},
    {31988, 2, {0x2000, 0x11DC, 0x2844, 0x4554, 0x90CC, 0x1154, 0x5400, 0x3848, 0x10FE, 0xFD90, 0x12FC, 0x3890, 0x54FC, 0x9490, 0x10FE, 0x1080}},
    {31989, 2, {0x820, 0xFFFE, 0x2820, 0x7DFC, 0x4488, 0x7C50, 0x41FE, 0x7C20, 0x45FC, 0x7C20, 0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {31990, 2, {0x5400, 0x55DC, 0x7C44, 0x1154, 0x54CC, 0x5554, 0x7C00, 0x48, 0x54FE, 0x3990, 0x12FC, 0xFC90, 0x38FC, 0x5490, 0x92FE, 0x1080}},
    {31991, 2, {0x2088, 0x23FE, 0x2088, 0xABDE, 0x7252, 0x23DE, 0xFA52, 0x23DE, 0x7222, 0x6BFE, 0xAAAA, 0x22FA, 0x2222, 0x2272, 0x22AA, 0x2224}},
    {31992, 2, {0x200, 0x400, 0x820, 0x1040, 0x3F80, 0x100, 0x610, 0x1808, 0x7FFC, 0x104, 0x920, 0x910, 0x1108, 0x2104, 0x4504, 0x200}},
    {31993, 2, {0x800, 0x800, 0x1000, 0x1200, 0x2200, 0x7C00, 0x800, 0x1000, 0x2400, 0x7E00, 0x200, 0x0, 0x5200, 0x4900, 0x8900, 0x0}},
    {31994, 2, {0x840, 0x840, 0x1040, 0x1240, 0x2240, 0x7C40, 0x840, 0x1040, 0x2440, 0x7E40, 0x240, 0x44, 0x5244, 0x4944, 0x893C, 0x0}},
    {31995, 2, {0xF8, 0x3F00, 0x400, 0x820, 0x1040, 0x3F80, 0x100, 0x610, 0x1808, 0x7FFC, 0x104, 0x920, 0x1110, 0x2108, 0x4504, 0x200}},
    {31996, 2, {0x1040, 0x1040, 0x2040, 0x2440, 0x45FC, 0xF844, 0x1044, 0x2044, 0x4844, 0xFC84, 0x484, 0x84, 0x5504, 0x5504, 0x8228, 0x410}},
    {31997, 2, {0x800, 0x800, 0x10FE, 0x1210, 0x2210, 0x7C10, 0x810, 0x1010, 0x2410, 0x7E10, 0x210, 0x10, 0x5210, 0x4910, 0x8950, 0x20}},
    {31998, 2, {0x804, 0x804, 0x1044, 0x1244, 0x2244, 0x7C44, 0x844, 0x1044, 0x244C, 0x7E54, 0x264, 0x44, 0x5204, 0x4904, 0x8904, 0x4}},
    {31999, 2, {0x1000, 0x1000, 0x21FC, 0x2444, 0x4444, 0xF844, 0x1044, 0x2044, 0x4844, 0xFC44, 0x444, 0x84, 0x5484, 0x5504, 0x8228, 0x410}},
    {32000, 2, {0x1000, 0x1000, 0x21F8, 0x2408, 0x4408, 0xF808, 0x1008, 0x21F8, 0x4908, 0xFD00, 0x500, 0x100, 0x5502, 0x5502, 0x80FE, 0x0}},
    {32001, 2, {0x1000, 0x13FC, 0x2104, 0x2544, 0x4524, 0xF8A8, 0x1088, 0x2088, 0x4850, 0xFC50, 0x420, 0x20, 0x5450, 0x5488, 0x8104, 0x602}},
    {32002, 2, {0x808, 0x808, 0x1008, 0x1208, 0x22FE, 0x7C08, 0x808, 0x1008, 0x2448, 0x7E28, 0x228, 0x8, 0x5208, 0x4908, 0x8928, 0x10}},
    {32003, 2, {0x1082, 0x1092, 0x2092, 0x2492, 0x4492, 0xF892, 0x1092, 0x2092, 0x4892, 0xFC92, 0x492, 0x92, 0x5492, 0x5512, 0x8102, 0x202}},
    {32004, 2, {0x1040, 0x1040, 0x2080, 0x24FC, 0x4504, 0xFA04, 0x1004, 0x2104, 0x4884, 0xFC44, 0x444, 0x4, 0x5404, 0x5404, 0x8028, 0x10}},
    {32005, 2, {0x1000, 0x1000, 0x21FC, 0x2420, 0x4420, 0xF820, 0x1020, 0x2020, 0x4820, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32006, 2, {0x1000, 0x11FC, 0x2020, 0x2420, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4820, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x80A0, 0x40}},
    {32007, 2, {0x1040, 0x1040, 0x2080, 0x24FE, 0x4500, 0xFA00, 0x10FC, 0x2008, 0x4810, 0xFC20, 0x440, 0x80, 0x5502, 0x5502, 0x80FE, 0x0}},
    {32008, 2, {0x1040, 0x1040, 0x2040, 0x2440, 0x45F8, 0xF848, 0x1048, 0x2048, 0x4948, 0xFCC8, 0x448, 0xA8, 0x54AA, 0x550A, 0x8206, 0x402}},
    {32009, 2, {0x1000, 0x1000, 0x21FC, 0x2444, 0x4444, 0xF944, 0x1144, 0x2144, 0x4944, 0xFE44, 0x444, 0x84, 0x5484, 0x5504, 0x8228, 0x410}},
    {32010, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x7C0, 0x1830, 0xE20E, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32011, 2, {0x1040, 0x1020, 0x2020, 0x27FE, 0x4488, 0xF888, 0x1088, 0x2088, 0x4888, 0xFC50, 0x450, 0x20, 0x5450, 0x5488, 0x8104, 0x602}},
    {32012, 2, {0x1040, 0x1050, 0x2048, 0x2448, 0x4440, 0xF9FE, 0x1050, 0x2050, 0x4850, 0xFC50, 0x490, 0x90, 0x5492, 0x5512, 0x810E, 0x200}},
    {32013, 2, {0x1020, 0x1020, 0x2020, 0x25FE, 0x4522, 0xF922, 0x1122, 0x2122, 0x4952, 0xFD4A, 0x58A, 0x102, 0x5502, 0x5502, 0x810A, 0x104}},
    {32014, 2, {0x1040, 0x1048, 0x2044, 0x2444, 0x4440, 0xFBFE, 0x1040, 0x2040, 0x4840, 0xFCA0, 0x4A0, 0x90, 0x5510, 0x5508, 0x8204, 0x402}},
    {32015, 2, {0x1008, 0x1088, 0x2048, 0x2448, 0x4408, 0xF888, 0x1048, 0x2048, 0x4808, 0xFC0E, 0x5F8, 0x8, 0x5408, 0x5408, 0x8008, 0x8}},
    {32016, 2, {0x1000, 0x11F8, 0x2048, 0x2448, 0x4448, 0xF848, 0x1048, 0x21F8, 0x4888, 0xFC88, 0x488, 0x88, 0x5488, 0x5488, 0x83FE, 0x0}},
    {32017, 2, {0x1000, 0x1000, 0x21FE, 0x2410, 0x4410, 0xF820, 0x1020, 0x2068, 0x48A4, 0xFD22, 0x622, 0x20, 0x5420, 0x5420, 0x8020, 0x20}},
    {32018, 2, {0x1020, 0x1020, 0x2050, 0x2450, 0x4488, 0xF904, 0x1202, 0x2088, 0x4888, 0xFC88, 0x488, 0x88, 0x5488, 0x5508, 0x8108, 0x208}},
    {32019, 2, {0x1000, 0x11F8, 0x2008, 0x2450, 0x4420, 0xF810, 0x13FE, 0x2022, 0x4824, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x80A0, 0x40}},
    {32020, 2, {0x1020, 0x1020, 0x2020, 0x2BFE, 0x4820, 0xF124, 0x1124, 0x2124, 0x4924, 0xFDFC, 0x424, 0x20, 0x5422, 0x5422, 0x801E, 0x0}},
    {32021, 2, {0x1010, 0x1110, 0x2110, 0x2512, 0x4512, 0xF914, 0x11D8, 0x2110, 0x4910, 0xFD10, 0x510, 0x112, 0x5552, 0x5592, 0x810E, 0x0}},
    {32022, 2, {0x1004, 0x11E4, 0x2024, 0x2424, 0x4424, 0xF9E4, 0x1104, 0x2104, 0x4904, 0xFDE4, 0x424, 0x24, 0x5424, 0x5424, 0x8144, 0x84}},
    {32023, 2, {0x1020, 0x1020, 0x2020, 0x24A8, 0x44A4, 0xF8A2, 0x1122, 0x2120, 0x4A24, 0xFC24, 0x428, 0x8, 0x5410, 0x5420, 0x80C0, 0x300}},
    {32024, 2, {0x1040, 0x1040, 0x2040, 0x2840, 0x4BFE, 0xF040, 0x1080, 0x2090, 0x4890, 0xF920, 0x920, 0x248, 0xAA44, 0xAC84, 0x81FE, 0x82}},
    {32025, 2, {0x1004, 0x101E, 0x21F0, 0x2510, 0x4510, 0xF910, 0x1110, 0x21FE, 0x4910, 0xFD10, 0x510, 0x108, 0x550A, 0x554A, 0x8186, 0x102}},
    {32026, 2, {0x1000, 0x13FC, 0x2084, 0x2488, 0x4488, 0xF890, 0x109C, 0x2084, 0x4944, 0xFD44, 0x528, 0x128, 0xAA10, 0xAA28, 0x8444, 0x182}},
    {32027, 2, {0x1010, 0x1090, 0x2090, 0x2488, 0x4508, 0xF904, 0x1204, 0x25FA, 0x4888, 0xFC88, 0x488, 0x88, 0x5508, 0x5508, 0x8228, 0x410}},
    {32028, 2, {0x1000, 0x1000, 0x20FC, 0x2400, 0x4400, 0xF800, 0x11FE, 0x2020, 0x4820, 0xFC40, 0x440, 0x88, 0x5484, 0x55FE, 0x8082, 0x0}},
    {32029, 2, {0x1008, 0x103C, 0x21E0, 0x2420, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4820, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x81FC, 0x0}},
    {32030, 2, {0x1020, 0x1020, 0x2020, 0x25FE, 0x4522, 0xFA24, 0x1020, 0x2020, 0x4850, 0xFC50, 0x450, 0x50, 0x5490, 0x5492, 0x8112, 0x20E}},
    {32031, 2, {0x1020, 0x1020, 0x2050, 0x2450, 0x4488, 0xF944, 0x1222, 0x2020, 0x4800, 0xFDFC, 0x404, 0x8, 0x5408, 0x5410, 0x8010, 0x20}},
    {32032, 2, {0x100, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32033, 2, {0x1040, 0x1020, 0x2020, 0x25FE, 0x4440, 0xF840, 0x1040, 0x207C, 0x4844, 0xFC44, 0x444, 0x44, 0x5484, 0x5484, 0x8128, 0x210}},
    {32034, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x7FFE, 0x4202, 0x8424, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32035, 2, {0x1040, 0x1040, 0x21F8, 0x2848, 0x4888, 0xF08A, 0x110A, 0x2226, 0x4820, 0xF820, 0xBFE, 0x20, 0xA820, 0xA820, 0x8020, 0x20}},
    {32036, 2, {0x1004, 0x100E, 0x20F0, 0x2480, 0x4480, 0xF880, 0x10FE, 0x2088, 0x4888, 0xFC88, 0x488, 0x88, 0x5488, 0x5508, 0x8108, 0x208}},
    {32037, 2, {0x840, 0x840, 0x7F40, 0x840, 0xE40, 0x7844, 0x844, 0x283C, 0x1440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32038, 2, {0x1000, 0x1000, 0x21FC, 0x2524, 0x4524, 0xF924, 0x1124, 0x2124, 0x49FC, 0xFD00, 0x500, 0x100, 0x5502, 0x5502, 0x80FE, 0x0}},
    {32039, 2, {0x400, 0x25FC, 0x2488, 0x2450, 0x2420, 0x2450, 0x28C, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32040, 2, {0x1088, 0x1088, 0x2088, 0x2908, 0x497E, 0xF308, 0x1508, 0x2148, 0x4928, 0xFD28, 0x508, 0x108, 0x5508, 0x5508, 0x8128, 0x110}},
    {32041, 2, {0x1020, 0x1020, 0x2120, 0x2520, 0x45FC, 0xF920, 0x1220, 0x2020, 0x4BFE, 0xFC20, 0x450, 0x50, 0x5488, 0x5488, 0x8104, 0x202}},
    {32042, 2, {0x1090, 0x1090, 0x2090, 0x2A90, 0x4A92, 0xF2D4, 0x1298, 0x2290, 0x4A90, 0xFA90, 0xA90, 0x292, 0xAAD2, 0xAF12, 0x820E, 0x0}},
    {32043, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF07C, 0x200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32044, 2, {0x1020, 0x1020, 0x2020, 0x2420, 0x45FC, 0xF924, 0x1124, 0x2124, 0x4924, 0xFDFC, 0x524, 0x124, 0x5524, 0x5524, 0x81FC, 0x104}},
    {32045, 2, {0x2008, 0x2788, 0x4088, 0x4088, 0x9088, 0xE788, 0x2408, 0x4410, 0x9410, 0xF790, 0x94, 0x10A4, 0xA8A2, 0xA0BE, 0x8502, 0x200}},
    {32046, 2, {0x420, 0x420, 0x7FA0, 0xE22, 0x1522, 0x649E, 0x200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32047, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32048, 2, {0x1000, 0x1000, 0x21FC, 0x2524, 0x4524, 0xF924, 0x1124, 0x2124, 0x49FC, 0xFD24, 0x524, 0x124, 0x5524, 0x5524, 0x81FC, 0x104}},
    {32049, 2, {0x1050, 0x1048, 0x2048, 0x2440, 0x47FE, 0xF880, 0x1080, 0x20FC, 0x4944, 0xFD44, 0x528, 0x128, 0xAA10, 0xAA28, 0x8444, 0x182}},
    {32050, 2, {0x1048, 0x1148, 0x2148, 0x2548, 0x4548, 0xFBFE, 0x1148, 0x2148, 0x4948, 0xFD48, 0x578, 0x100, 0x5500, 0x5500, 0x81FE, 0x0}},
    {32051, 2, {0x1020, 0x1020, 0x2020, 0x25FC, 0x4524, 0xF924, 0x1124, 0x21FC, 0x4924, 0xFD24, 0x524, 0x1FC, 0x5524, 0x5420, 0x8020, 0x20}},
    {32052, 2, {0x1010, 0x1010, 0x2010, 0x25FE, 0x4512, 0xF914, 0x1110, 0x21FC, 0x4944, 0xFD44, 0x528, 0x128, 0xA910, 0xAA28, 0x8244, 0x482}},
    {32053, 2, {0x1040, 0x1020, 0x2020, 0x2BFE, 0x4A02, 0xF404, 0x1000, 0x2000, 0x4BFE, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x80A0, 0x40}},
    {32054, 2, {0x1020, 0x1020, 0x2020, 0x25FC, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4820, 0xFC20, 0x440, 0x48, 0x5484, 0x55FE, 0x8082, 0x0}},
    {32055, 2, {0x1020, 0x1020, 0x2050, 0x2450, 0x4488, 0xF924, 0x1212, 0x2010, 0x49FC, 0xFC04, 0x408, 0x88, 0x5450, 0x5420, 0x8010, 0x10}},
    {32056, 2, {0x1040, 0x1020, 0x2000, 0x25FE, 0x4420, 0xF820, 0x1020, 0x2020, 0x49FC, 0xFC20, 0x420, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32057, 2, {0x1000, 0x11FC, 0x2044, 0x2444, 0x4444, 0xF844, 0x1094, 0x2088, 0x4900, 0xFCFC, 0x484, 0x84, 0x5484, 0x5484, 0x80FC, 0x84}},
    {32058, 2, {0x1088, 0x1088, 0x2088, 0x2488, 0x47FE, 0xF888, 0x1088, 0x2088, 0x4888, 0xFCF8, 0x488, 0x88, 0x5488, 0x5488, 0x80F8, 0x88}},
    {32059, 2, {0x1020, 0x1020, 0x2020, 0x25FC, 0x4524, 0xF924, 0x1124, 0x2124, 0x4924, 0xFFFE, 0x420, 0x50, 0x5450, 0x5488, 0x8104, 0x202}},
    {32060, 2, {0x1090, 0x1090, 0x2090, 0x2BFC, 0x4894, 0xF094, 0x13FC, 0x2290, 0x4A90, 0xFBFE, 0x892, 0x92, 0xA91A, 0xA914, 0x8210, 0x410}},
    {32061, 2, {0x1020, 0x1010, 0x2010, 0x25FE, 0x4502, 0xFA04, 0x1080, 0x2088, 0x4890, 0xFCA0, 0x4C0, 0x82, 0x5482, 0x5482, 0x807E, 0x0}},
    {32062, 2, {0x1020, 0x1020, 0x2050, 0x2488, 0x4504, 0xFA12, 0x1020, 0x2040, 0x4988, 0xFC10, 0x420, 0x44, 0x5588, 0x5410, 0x8060, 0x380}},
    {32063, 2, {0x1020, 0x1020, 0x2040, 0x2440, 0x4488, 0xF904, 0x13FE, 0x2002, 0x4800, 0xFDFC, 0x504, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32064, 2, {0x1020, 0x1020, 0x2124, 0x2524, 0x4524, 0xF924, 0x11FC, 0x2020, 0x4820, 0xFD24, 0x524, 0x124, 0x5524, 0x5524, 0x81FC, 0x4}},
    {32065, 2, {0x1080, 0x1080, 0x20FE, 0x2500, 0x4620, 0xF920, 0x112C, 0x2174, 0x4BA4, 0xFD24, 0x534, 0x128, 0x5522, 0x5502, 0x80FE, 0x0}},
    {32066, 2, {0x1040, 0x1040, 0x207C, 0x2884, 0x4988, 0xF250, 0x1020, 0x2050, 0x4888, 0xFB06, 0x860, 0x10, 0xA808, 0xA8C0, 0x8020, 0x10}},
    {32067, 2, {0x1020, 0x1010, 0x2010, 0x25FE, 0x4420, 0xF820, 0x1044, 0x2084, 0x49F8, 0xFC10, 0x420, 0x44, 0x5482, 0x55FE, 0x8082, 0x0}},
    {32068, 2, {0x1000, 0x11F8, 0x2108, 0x2508, 0x4508, 0xF9F8, 0x1108, 0x2108, 0x4908, 0xFDF8, 0x508, 0x108, 0xA908, 0xA908, 0x87FE, 0x0}},
    {32069, 2, {0x1000, 0x11FE, 0x2102, 0x2502, 0x4502, 0xF97A, 0x114A, 0x214A, 0x494A, 0xFD4A, 0x57A, 0x14A, 0x5502, 0x5502, 0x810A, 0x104}},
    {32070, 2, {0x1020, 0x1020, 0x2124, 0x24A4, 0x44A8, 0xF820, 0x11FC, 0x2020, 0x4820, 0xFC20, 0x7FE, 0x20, 0x5420, 0x5420, 0x8020, 0x20}},
    {32071, 2, {0x1080, 0x1080, 0x2100, 0x29FC, 0x4A04, 0xF404, 0x11E4, 0x2124, 0x4924, 0xFD24, 0x5E4, 0x124, 0x5404, 0x5404, 0x8028, 0x10}},
    {32072, 2, {0x1020, 0x1020, 0x2040, 0x25FC, 0x4504, 0xF904, 0x1104, 0x2104, 0x49FC, 0xFD04, 0x504, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32073, 2, {0x1020, 0x1028, 0x2024, 0x2420, 0x45FE, 0xF820, 0x1070, 0x2070, 0x48A8, 0xFCA8, 0x524, 0x222, 0x5420, 0x5420, 0x8020, 0x20}},
    {32074, 2, {0x1020, 0x1020, 0x2020, 0x2420, 0x45FE, 0xF870, 0x10A8, 0x20A8, 0x4924, 0xFD24, 0x622, 0xF8, 0x5420, 0x5420, 0x8020, 0x20}},
    {32075, 2, {0x1020, 0x1010, 0x2010, 0x25FE, 0x4500, 0xF910, 0x1110, 0x2110, 0x4910, 0xFD20, 0x528, 0x124, 0xAA44, 0xAAFE, 0x8442, 0x800}},
    {32076, 2, {0x1000, 0x11FC, 0x2084, 0x2488, 0x4450, 0xF820, 0x1050, 0x2088, 0x4B26, 0xFC20, 0x5FC, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32077, 2, {0x1080, 0x1088, 0x209C, 0x2570, 0x4510, 0xFB10, 0x1510, 0x2110, 0x49FE, 0xFD10, 0x510, 0x110, 0x5510, 0x5510, 0x817C, 0x100}},
    {32078, 2, {0x1040, 0x1040, 0x209C, 0x2500, 0x4640, 0xF840, 0x10BE, 0x2188, 0x4A88, 0xFC88, 0x488, 0x88, 0x5488, 0x5488, 0x80A8, 0x90}},
    {32079, 2, {0x1020, 0x1020, 0x21FC, 0x2524, 0x4524, 0xF9FC, 0x1124, 0x2124, 0x49FC, 0xFC20, 0x424, 0x18, 0x5432, 0x54CA, 0x8306, 0x2}},
    {32080, 2, {0x1020, 0x1020, 0x2020, 0x2BFE, 0x4820, 0xF020, 0x11FC, 0x2000, 0x4800, 0xFDFC, 0x504, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32081, 2, {0x1020, 0x1120, 0x2120, 0x25FC, 0x4520, 0xFA20, 0x1020, 0x23FE, 0x4870, 0xFCA8, 0x4A8, 0x124, 0x5524, 0x5622, 0x8020, 0x20}},
    {32082, 2, {0x1104, 0x1124, 0x2124, 0x2924, 0x4924, 0xF124, 0x15B4, 0x256C, 0x4924, 0xF924, 0x924, 0x124, 0xA924, 0xAA24, 0x8204, 0x404}},
    {32083, 2, {0x1020, 0x1020, 0x21FC, 0x2420, 0x4420, 0xF820, 0x13FE, 0x2000, 0x4820, 0xFC20, 0x5FC, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32084, 2, {0x1000, 0x13FE, 0x2020, 0x2420, 0x4440, 0xF9FC, 0x1104, 0x2104, 0x4904, 0xFDFC, 0x504, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32085, 2, {0x1000, 0x13FC, 0x2044, 0x2444, 0x4484, 0xF928, 0x1210, 0x21FC, 0x4924, 0xFD24, 0x5FC, 0x100, 0x5502, 0x5502, 0x80FE, 0x0}},
    {32086, 2, {0x1040, 0x1044, 0x2244, 0x2948, 0x4950, 0xF040, 0x17FE, 0x2090, 0x4890, 0xFC90, 0x490, 0x92, 0x5512, 0x5512, 0x820E, 0x400}},
    {32087, 2, {0x1000, 0x1000, 0x21FE, 0x2502, 0x4502, 0xF97A, 0x114A, 0x214A, 0x494A, 0xFD4A, 0x57A, 0x102, 0x5502, 0x55FE, 0x8102, 0x0}},
    {32088, 2, {0x1010, 0x1010, 0x2210, 0x293E, 0x4922, 0xF044, 0x1010, 0x2090, 0x4890, 0xFB10, 0x928, 0x128, 0xA928, 0xA944, 0x8044, 0x82}},
    {32089, 2, {0x1000, 0x13FE, 0x2000, 0x2400, 0x45FC, 0xF904, 0x1104, 0x21FC, 0x4904, 0xFD04, 0x5FC, 0x104, 0x5400, 0x5400, 0x83FE, 0x0}},
    {32090, 2, {0x1000, 0x13FE, 0x2040, 0x2440, 0x4440, 0xF87C, 0x1084, 0x20A4, 0x4894, 0xFD04, 0x544, 0x28, 0x5408, 0x5408, 0x83FE, 0x0}},
    {32091, 2, {0x1080, 0x10F8, 0x1108, 0x2290, 0x2060, 0x6598, 0x6646, 0xA488, 0x25F0, 0x2420, 0x2448, 0x25FC, 0x2020, 0x2124, 0x22A2, 0x2040}},
    {32092, 2, {0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0xA88, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32093, 2, {0x1020, 0x1020, 0x23FE, 0x2450, 0x4488, 0xF904, 0x12FA, 0x2000, 0x4BFE, 0xFC40, 0x480, 0x1FC, 0x5404, 0x5404, 0x8028, 0x10}},
    {32094, 2, {0x1040, 0x1020, 0x2020, 0x25FE, 0x4400, 0xF888, 0x1104, 0x2202, 0x4888, 0xFC88, 0x450, 0x50, 0x5420, 0x5450, 0x8088, 0x306}},
    {32095, 2, {0x1020, 0x1020, 0x2050, 0x2488, 0x4504, 0xFA02, 0x11FC, 0x2020, 0x4820, 0xFC20, 0x5FC, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32096, 2, {0x1020, 0x1020, 0x21FE, 0x2440, 0x4440, 0xF8FC, 0x1084, 0x2184, 0x4AFC, 0xFC84, 0x484, 0xFC, 0x5484, 0x5484, 0x8094, 0x88}},
    {32097, 2, {0x1040, 0x1040, 0x2078, 0x2488, 0x4550, 0xF820, 0x1050, 0x2088, 0x4906, 0xFCF8, 0x488, 0x88, 0x5488, 0x5488, 0x80F8, 0x88}},
    {32098, 2, {0x1080, 0x1080, 0x21FC, 0x2504, 0x4604, 0xF9E4, 0x1124, 0x2124, 0x49E4, 0xFD24, 0x524, 0x1E4, 0x5504, 0x5404, 0x8028, 0x10}},
    {32099, 2, {0x1104, 0x1084, 0x2088, 0x2800, 0x49FE, 0xF088, 0x1088, 0x2088, 0x4888, 0xFBFE, 0x888, 0x88, 0xA908, 0xA908, 0x8208, 0x408}},
    {32100, 2, {0x1000, 0x13FE, 0x2050, 0x2450, 0x45FC, 0xF954, 0x1154, 0x2154, 0x4954, 0xFD54, 0x58C, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32101, 2, {0x1090, 0x1094, 0x2092, 0x2912, 0x4910, 0xF37E, 0x1510, 0x2110, 0x4910, 0xFD28, 0x528, 0x128, 0x5528, 0x5544, 0x8144, 0x182}},
    {32102, 2, {0x1020, 0x1020, 0x2050, 0x2488, 0x4504, 0xFA02, 0x11FC, 0x2000, 0x4800, 0xFDFC, 0x504, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32103, 2, {0x1000, 0x11FE, 0x2102, 0x2502, 0x457A, 0xF902, 0x1102, 0x217A, 0x494A, 0xFD4A, 0x54A, 0x17A, 0x5502, 0x5502, 0x810A, 0x104}},
    {32104, 2, {0x1028, 0x1024, 0x2024, 0x2820, 0x4BFE, 0xF020, 0x1124, 0x2124, 0x4924, 0xFBA8, 0x928, 0x110, 0xAA12, 0xAA2A, 0x8446, 0x82}},
    {32105, 2, {0x1090, 0x1090, 0x2090, 0x2A92, 0x4994, 0xF098, 0x1090, 0x2198, 0x4A94, 0xFC92, 0x890, 0x90, 0xA912, 0xA912, 0x8212, 0x40E}},
    {32106, 2, {0x1000, 0x11FC, 0x2104, 0x2524, 0x4524, 0xF924, 0x11FC, 0x2124, 0x4924, 0xFD54, 0x54C, 0x18C, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32107, 2, {0x400, 0x820, 0x1FF0, 0x1020, 0x2448, 0x7EFC, 0x200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32108, 2, {0x1008, 0x103C, 0x21E0, 0x2420, 0x4420, 0xFBFE, 0x1020, 0x2020, 0x4820, 0xFDFC, 0x504, 0x104, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32109, 2, {0x1110, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x3218, 0xC426, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32110, 2, {0x800, 0x7F7C, 0x1144, 0x3244, 0xC44, 0x327C, 0xC200, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32111, 2, {0x1020, 0x1010, 0x2010, 0x25FE, 0x4420, 0xF824, 0x1044, 0x20F8, 0x4812, 0xFC22, 0x444, 0x188, 0x5410, 0x5428, 0x8044, 0x182}},
    {32112, 2, {0x1000, 0x11FE, 0x2020, 0x2440, 0x4488, 0xF904, 0x11FE, 0x2022, 0x4820, 0xFC20, 0x5FE, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32113, 2, {0x1040, 0x1020, 0x2020, 0x2BFE, 0x4840, 0xF088, 0x1104, 0x23FE, 0x4892, 0xFC90, 0x490, 0x90, 0x5512, 0x5512, 0x820E, 0x400}},
    {32114, 2, {0x1020, 0x1020, 0x2048, 0x2488, 0x45F0, 0xF820, 0x1040, 0x2088, 0x49FC, 0xFC24, 0x420, 0xA8, 0x5524, 0x5622, 0x80A0, 0x40}},
    {32115, 2, {0x1080, 0x1080, 0x20FC, 0x2908, 0x4A90, 0xF060, 0x1198, 0x2626, 0x4820, 0xF9FC, 0x820, 0x220, 0xABFE, 0xA820, 0x8020, 0x20}},
    {32116, 2, {0x1104, 0x1084, 0x2088, 0x2800, 0x4BFE, 0xF020, 0x1020, 0x21FC, 0x4820, 0xFC20, 0x7FE, 0x20, 0x5420, 0x5420, 0x8020, 0x20}},
    {32117, 2, {0x1020, 0x1020, 0x2050, 0x2888, 0x4904, 0xF202, 0x10F8, 0x2000, 0x4800, 0xFBFE, 0x820, 0x40, 0xA888, 0xA904, 0x83FE, 0x102}},
    {32118, 2, {0x1040, 0x1040, 0x20F8, 0x2508, 0x4610, 0xF9FC, 0x1124, 0x2124, 0x4924, 0xFDFC, 0x500, 0x100, 0x5502, 0x5502, 0x80FE, 0x0}},
    {32119, 2, {0x1080, 0xFBF0, 0x1090, 0x1990, 0xF094, 0x1154, 0x520C, 0x2424, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32120, 2, {0x1000, 0x11F8, 0x2108, 0x2508, 0x45F8, 0xF908, 0x1108, 0x21F8, 0x4908, 0xFD08, 0x5F8, 0x90, 0x5490, 0x5512, 0x8212, 0x40E}},
    {32121, 2, {0x1000, 0x10F8, 0x2088, 0x2488, 0x44F8, 0xF800, 0x11FC, 0x2104, 0x4904, 0xFDFC, 0x504, 0x104, 0x55FC, 0x5504, 0x8114, 0x108}},
    {32122, 2, {0x1184, 0x1068, 0x2030, 0x28C8, 0x4B04, 0xF040, 0x13FE, 0x20A0, 0x4920, 0xFBFC, 0xD24, 0x124, 0xA934, 0xA928, 0x8020, 0x20}},
    {32123, 2, {0x1080, 0x1080, 0x20F8, 0x2508, 0x4610, 0xF9FC, 0x1124, 0x2124, 0x4924, 0xFDFC, 0x450, 0x50, 0x5490, 0x5492, 0x8112, 0x20E}},
    {32124, 2, {0x1008, 0x103C, 0x21E0, 0x2496, 0x4448, 0xF830, 0x1048, 0x2194, 0x4824, 0xFC4C, 0x594, 0x24, 0x5444, 0x5584, 0x8028, 0x10}},
    {32125, 2, {0x1000, 0x10FC, 0x2084, 0x2484, 0x4484, 0xF884, 0x10FC, 0x2000, 0x4800, 0xFDFE, 0x502, 0x102, 0x5502, 0x5502, 0x81FE, 0x102}},
    {32126, 2, {0x1028, 0x1024, 0x2024, 0x2820, 0x4BFE, 0xF220, 0x1224, 0x2224, 0x4BA4, 0xFAA8, 0xAA8, 0x290, 0xAA92, 0xADAA, 0x8446, 0x882}},
    {32127, 2, {0x1028, 0x1024, 0x2024, 0x2420, 0x45FE, 0xF820, 0x1120, 0x20B2, 0x48B4, 0xFC68, 0x4A8, 0x124, 0x5622, 0x5420, 0x80A0, 0x40}},
    {32128, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x4420, 0xF9FC, 0x1124, 0x2124, 0x49FC, 0xFC20, 0x470, 0xA8, 0x5524, 0x5622, 0x8020, 0x20}},
    {32129, 2, {0x2100, 0x211E, 0x4112, 0x47D2, 0x9114, 0xE114, 0x27D8, 0x4114, 0x9112, 0xF7D2, 0x112, 0x111A, 0xA914, 0xA210, 0x8210, 0x410}},
    {32130, 2, {0x1040, 0x1020, 0x23FE, 0x2400, 0x4400, 0xF9FC, 0x1104, 0x2104, 0x4904, 0xFDFC, 0x450, 0x50, 0x5490, 0x5492, 0x8112, 0x20E}},
    {32131, 2, {0x1020, 0x1124, 0x20A4, 0x24A8, 0x4420, 0xF9FC, 0x1104, 0x2104, 0x49FC, 0xFD04, 0x504, 0x1FC, 0x5504, 0x5504, 0x8114, 0x108}},
    {32132, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4C04, 0xF1F8, 0x1000, 0x2000, 0x4BFE, 0xFC90, 0x490, 0x90, 0x5512, 0x5512, 0x820E, 0x400}},
    {32133, 2, {0x1000, 0x11FC, 0x2004, 0x24FC, 0x4404, 0xF9FC, 0x1000, 0x23FE, 0x4A02, 0xFDF8, 0x488, 0x88, 0x5450, 0x5420, 0x80D8, 0x306}},
    {32134, 2, {0x1000, 0x13FE, 0x2020, 0x2420, 0x45FC, 0xF924, 0x1124, 0x21FC, 0x4924, 0xFD24, 0x5FC, 0x120, 0x54A0, 0x5440, 0x80B0, 0x30E}},
    {32135, 2, {0x1000, 0x13FE, 0x2050, 0x2450, 0x45FC, 0xF954, 0x1154, 0x2154, 0x498C, 0xFD04, 0x504, 0x1FC, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32136, 2, {0x1108, 0x1088, 0x2090, 0x25FC, 0x4424, 0xF824, 0x11FC, 0x2120, 0x4920, 0xFDFE, 0x462, 0xA2, 0x552A, 0x5624, 0x8420, 0x20}},
    {32137, 2, {0x101C, 0x11E0, 0x2020, 0x2420, 0x47FE, 0xF8A8, 0x1124, 0x2222, 0x49F8, 0xFC88, 0x490, 0xBE, 0x5502, 0x5502, 0x8214, 0x408}},
    {32138, 2, {0x1020, 0x1020, 0x2020, 0x2BFE, 0x4820, 0xF020, 0x1124, 0x2124, 0x4924, 0xFAAA, 0x820, 0x50, 0xA850, 0xA888, 0x8104, 0x202}},
    {32139, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4C44, 0xF040, 0x13FE, 0x2040, 0x4890, 0xF890, 0x910, 0x120, 0xAA24, 0xAA42, 0x84FE, 0x842}},
    {32140, 2, {0x1000, 0x1050, 0x2048, 0x2484, 0x4524, 0xF820, 0x1050, 0x2088, 0x4906, 0xFEF8, 0x488, 0x88, 0x5488, 0x5488, 0x80F8, 0x88}},
    {32141, 2, {0x1020, 0x1020, 0x23FE, 0x2820, 0x4820, 0xF3FE, 0x1202, 0x2404, 0x49F8, 0xFC10, 0x420, 0x3FE, 0x5420, 0x5420, 0x80A0, 0x40}},
    {32142, 2, {0x1000, 0x1006, 0x23B8, 0x2888, 0x4888, 0xF108, 0x113E, 0x2388, 0x4888, 0xFA88, 0xA88, 0x13E, 0xA900, 0xAA80, 0x847E, 0x800}},
    {32143, 2, {0x1008, 0x103C, 0x23C0, 0x2844, 0x4A24, 0xF128, 0x1100, 0x2040, 0x4BFE, 0xF888, 0x908, 0x390, 0xA860, 0xA850, 0x8188, 0x604}},
    {32144, 2, {0x1104, 0x1084, 0x2088, 0x2410, 0x45FC, 0xF904, 0x1104, 0x2104, 0x49FC, 0xFC50, 0x450, 0x90, 0x5492, 0x5512, 0x820E, 0x400}},
    {32145, 2, {0x1000, 0x13FE, 0x2222, 0x2A22, 0x4A22, 0xF3FE, 0x1222, 0x2262, 0x4A72, 0xFAAA, 0xB22, 0x222, 0xAA22, 0xAA02, 0x83FE, 0x202}},
    {32146, 2, {0x1008, 0x103C, 0x21C0, 0x2404, 0x4544, 0xF8A8, 0x1000, 0x21F8, 0x4810, 0xFC20, 0x7FE, 0x20, 0x5420, 0x5420, 0x80A0, 0x40}},
    {32147, 2, {0x1000, 0x11FE, 0x2000, 0x2892, 0x4924, 0xF248, 0x1124, 0x2092, 0x4800, 0xFDFE, 0x420, 0x20, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32148, 2, {0x1000, 0x100C, 0xFEF0, 0x2820, 0x4444, 0x82F8, 0x7C10, 0x24, 0xFEFE, 0x2012, 0x4050, 0x7C54, 0x492, 0x512, 0x2850, 0x1020}},
    {32149, 2, {0x1020, 0x1020, 0x2020, 0x2BFE, 0x4820, 0xF020, 0x1020, 0x21FC, 0x4840, 0xF820, 0x8A4, 0x282, 0xAA8A, 0xAA8A, 0x8478, 0x0}},
    {32150, 2, {0x1004, 0x100E, 0x23B8, 0x2888, 0x4888, 0xF128, 0x112E, 0x23A8, 0x48A8, 0xFAA8, 0xAA8, 0x13E, 0xA900, 0xAA80, 0x847E, 0x800}},
    {32151, 2, {0x1000, 0x13FE, 0x2202, 0x2A52, 0x4A8A, 0xF306, 0x1202, 0x22FA, 0x4A8A, 0xFA8A, 0xA8A, 0x2FA, 0xAA8A, 0xAA02, 0x820A, 0x204}},
    {32152, 2, {0x1040, 0x1040, 0x20FC, 0x2488, 0x4550, 0xF820, 0x10D8, 0x2326, 0x48F8, 0xFC20, 0x4F8, 0x20, 0x57FE, 0x5420, 0x8020, 0x20}},
    {32153, 2, {0x1010, 0x1110, 0x2110, 0x2592, 0x4554, 0xF910, 0x11FE, 0x2110, 0x4938, 0xFD54, 0x592, 0x110, 0x5510, 0x5510, 0x81FE, 0x0}},
    {32154, 2, {0x1020, 0x1020, 0x23FE, 0x2820, 0x49FC, 0xF000, 0x13FE, 0x2202, 0x4C04, 0xF890, 0x890, 0x90, 0xA890, 0xA912, 0x8112, 0x20E}},
    {32155, 2, {0x1000, 0x13FC, 0x2044, 0x2944, 0x4944, 0xF284, 0x1094, 0x2108, 0x4A40, 0xF820, 0x8A4, 0x282, 0xAA8A, 0xAA8A, 0x8478, 0x0}},
    {32156, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4800, 0xF1FC, 0x1000, 0x2000, 0x4BFE, 0xF820, 0x928, 0x124, 0xAA22, 0xAC22, 0x80A0, 0x40}},
    {32157, 2, {0x1088, 0x1088, 0x2088, 0x2888, 0x4BDE, 0xF088, 0x1088, 0x219C, 0x49DC, 0xFAAA, 0xAAA, 0x4C8, 0xA888, 0xA888, 0x8088, 0x88}},
    {32158, 2, {0x1008, 0x103C, 0x21E0, 0x2820, 0x4BFE, 0xF0A8, 0x10A8, 0x20A8, 0x4BFE, 0xFCA8, 0x4A8, 0x3FE, 0x5420, 0x5420, 0x81FC, 0x0}},
    {32159, 2, {0x1040, 0x1020, 0x23FC, 0x2A04, 0x4A04, 0xF3FC, 0x1200, 0x2228, 0x4A24, 0xFBFE, 0xA20, 0x250, 0xAA50, 0xAC88, 0x8504, 0xA02}},
    {32160, 2, {0x1080, 0x10F8, 0x2088, 0x2908, 0x49F0, 0xF010, 0x1010, 0x23FE, 0x4820, 0xFA22, 0x974, 0xA8, 0xA924, 0xAA22, 0x80A0, 0x40}},
    {32161, 2, {0x1040, 0x1020, 0x23FE, 0x2400, 0x4400, 0xF9FC, 0x1104, 0x2104, 0x4904, 0xFDFC, 0x420, 0xA8, 0x5524, 0x5622, 0x80A0, 0x40}},
    {32162, 2, {0x1000, 0x11FC, 0x2124, 0x2524, 0x4574, 0xF924, 0x11FC, 0x2104, 0x4974, 0xFD54, 0x554, 0x174, 0x5504, 0x5604, 0x8214, 0x408}},
    {32163, 2, {0x1020, 0x1124, 0x20A8, 0x2420, 0x45FC, 0xF840, 0x13FE, 0x2088, 0x4904, 0xFEFA, 0x488, 0x88, 0x54A8, 0x5492, 0x8082, 0x7E}},
    {32164, 2, {0x7E7C, 0x1244, 0x2A48, 0x4444, 0xBE44, 0x2254, 0x2248, 0x3E00, 0x440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32165, 2, {0x1000, 0x11FC, 0x2124, 0x2924, 0x49FC, 0xF124, 0x1124, 0x21FC, 0x4800, 0xF800, 0xBFE, 0x88, 0xA888, 0xA888, 0x8108, 0x208}},
    {32166, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1450, 0x2888, 0xDF06, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32167, 2, {0x1040, 0x1020, 0x23FE, 0x2400, 0x45FC, 0xF904, 0x11FC, 0x2000, 0x49FC, 0xFC08, 0x410, 0x3FE, 0x5420, 0x5420, 0x80A0, 0x40}},
    {32168, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x4488, 0xF8F8, 0x1088, 0x2088, 0x48F8, 0xFC88, 0x488, 0x3FE, 0x5400, 0x5488, 0x8104, 0x202}},
    {32169, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4D04, 0xF100, 0x11DE, 0x2252, 0x4A52, 0xFB52, 0xC9A, 0x94, 0xA910, 0xA912, 0x8212, 0x40E}},
    {32170, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF820, 0x13FE, 0x2000, 0x49FC, 0xFD04, 0x5FC, 0x104, 0x55FC, 0x5504, 0x8114, 0x108}},
    {32171, 2, {0x1048, 0x1044, 0x205E, 0x25E0, 0x4428, 0xF812, 0x106A, 0x2196, 0x4848, 0xFC5E, 0x5E0, 0x24, 0x5428, 0x5412, 0x806A, 0x186}},
    {32172, 2, {0x101C, 0x13E0, 0x2084, 0x2A44, 0x4948, 0xF110, 0x17FE, 0x2402, 0x4800, 0xFBF8, 0x908, 0x110, 0xA8A0, 0xA840, 0x81B0, 0x60E}},
    {32173, 2, {0x10A0, 0x1090, 0x2080, 0x25FE, 0x4510, 0xFB10, 0x15FC, 0x2110, 0x4910, 0xFDFC, 0x510, 0x110, 0x5510, 0x55FE, 0x8100, 0x100}},
    {32174, 2, {0x1040, 0x840, 0x3E7E, 0x2288, 0x2350, 0x3E20, 0x2058, 0x4186, 0x8440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32175, 2, {0x1080, 0x1080, 0x21FC, 0x2504, 0x4684, 0xF8F4, 0x1144, 0x2044, 0x4BFC, 0xFC44, 0x554, 0x154, 0x55F4, 0x5404, 0x8028, 0x10}},
    {32176, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4C04, 0xF1F8, 0x1108, 0x2108, 0x49F8, 0xFD00, 0x500, 0x1FC, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32177, 2, {0x1000, 0x13FE, 0x2202, 0x2A8A, 0x4A52, 0xF3FE, 0x1222, 0x2222, 0x4AAA, 0xFAAA, 0xAAA, 0x2FA, 0xAA02, 0xAA02, 0x820A, 0x204}},
    {32178, 2, {0x1000, 0x11FE, 0x2102, 0x218A, 0x4952, 0xF9FE, 0x1122, 0x2112, 0x21FE, 0x4942, 0xF542, 0x17A, 0x5502, 0x5502, 0x550A, 0x8104}},
    {32179, 2, {0x1000, 0x13DE, 0x2252, 0x2A52, 0x4A52, 0xF3DE, 0x1252, 0x2252, 0x4A52, 0xFBDE, 0xA52, 0x252, 0xAA52, 0xAA52, 0x8552, 0x8A6}},
    {32180, 2, {0x1000, 0x13DE, 0x2042, 0x294A, 0x4884, 0xF14A, 0x1252, 0x2420, 0x4800, 0xFBDE, 0x852, 0x152, 0xA894, 0xA948, 0x8254, 0x422}},
    {32181, 2, {0x1008, 0x103C, 0x23C0, 0x2844, 0x4A24, 0xF128, 0x1100, 0x2020, 0x4BFE, 0xF870, 0x8A8, 0xA8, 0xA924, 0xAA22, 0x8420, 0x20}},
    {32182, 2, {0x1000, 0x11FC, 0x2124, 0x2924, 0x49FC, 0xF124, 0x1124, 0x21FC, 0x4820, 0xFBFE, 0x870, 0xA8, 0xA924, 0xAA22, 0x8020, 0x20}},
    {32183, 2, {0x1040, 0x1020, 0x21FE, 0x2400, 0x4488, 0xF888, 0x1154, 0x2222, 0x4800, 0xFC20, 0x5FE, 0x20, 0x5420, 0x5420, 0x8020, 0x20}},
    {32184, 2, {0x1020, 0x1020, 0x2050, 0x2488, 0x4504, 0xFAFA, 0x1000, 0x21FC, 0x4954, 0xFD54, 0x5FC, 0x154, 0x5554, 0x5554, 0x8104, 0x10C}},
    {32185, 2, {0x1088, 0x1088, 0x20E8, 0x252C, 0x452A, 0xFAA8, 0x1048, 0x20A0, 0x491E, 0xFE00, 0x4FC, 0x84, 0x5484, 0x5484, 0x80FC, 0x84}},
    {32186, 2, {0x1020, 0x1020, 0x21FC, 0x2450, 0x4488, 0xF904, 0x13FE, 0x2008, 0x49E8, 0xFD28, 0x528, 0x1E8, 0x5528, 0x5408, 0x8028, 0x10}},
    {32187, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4C04, 0xF000, 0x13FE, 0x2020, 0x4820, 0xF920, 0x93C, 0x120, 0xAAA0, 0xAA60, 0x843E, 0x800}},
    {32188, 2, {0x1040, 0x1080, 0x21FC, 0x2524, 0x4524, 0xF9FC, 0x1124, 0x2144, 0x49FC, 0xFC90, 0x510, 0x3FE, 0x5410, 0x5410, 0x8010, 0x10}},
    {32189, 2, {0x1020, 0x1020, 0x203E, 0x2420, 0x45FC, 0xF904, 0x11FC, 0x2104, 0x49FC, 0xFD24, 0x420, 0x3FE, 0x5420, 0x5420, 0x8020, 0x20}},
    {32190, 2, {0x1020, 0x1020, 0x21FC, 0x2820, 0x4820, 0xF3FE, 0x1088, 0x2144, 0x4A42, 0xF8F8, 0x988, 0x250, 0xA820, 0xA850, 0x8188, 0x606}},
    {32191, 2, {0x1020, 0x1040, 0x21FC, 0x2904, 0x49FC, 0xF104, 0x11FC, 0x2020, 0x4820, 0xFBFE, 0xA22, 0x222, 0xAA2A, 0xAA24, 0x8020, 0x20}},
    {32192, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF824, 0x13FE, 0x2024, 0x49FC, 0xFC40, 0x7FE, 0x84, 0x55C8, 0x5430, 0x80CC, 0x302}},
    {32193, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF824, 0x13FE, 0x2024, 0x49FC, 0xFC20, 0x520, 0x13E, 0xA920, 0xAAA0, 0x827E, 0x400}},
    {32194, 2, {0x1020, 0x1122, 0x2122, 0x2A24, 0x4850, 0xF088, 0x1304, 0x2022, 0x4820, 0xF924, 0x924, 0x228, 0xA850, 0xA888, 0x8104, 0x602}},
    {32195, 2, {0x1088, 0x1088, 0x2088, 0x2488, 0x4554, 0xFA22, 0x1442, 0x2020, 0x4820, 0xFD20, 0x53C, 0x120, 0xAAA0, 0xAA60, 0x843E, 0x800}},
    {32196, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x45FC, 0xF904, 0x1104, 0x21FC, 0x4800, 0xFD12, 0x5D4, 0x118, 0x5510, 0x5552, 0x8192, 0x10E}},
    {32197, 2, {0x2000, 0x2FC0, 0x44BC, 0x44A4, 0x94A4, 0xE7A4, 0x24A4, 0x44A8, 0x97A8, 0xF4A8, 0x490, 0x15D0, 0xAEA8, 0xA0A8, 0x80C4, 0x82}},
    {32198, 2, {0x1000, 0x11FC, 0x2104, 0x25FC, 0x4504, 0xF9FC, 0x1080, 0x2080, 0x49FE, 0xFE4A, 0x44A, 0x92, 0xA922, 0xAA42, 0x8094, 0x108}},
    {32199, 2, {0x1092, 0x1092, 0x2124, 0x2A48, 0x4924, 0xF092, 0x1092, 0x2000, 0x49FE, 0xFD22, 0x522, 0x1FE, 0x5522, 0x5522, 0x81FE, 0x102}},
    {32200, 2, {0x1020, 0x1020, 0x21FC, 0x2420, 0x4420, 0xFBFE, 0x1088, 0x2050, 0x49FC, 0xFC20, 0x420, 0x3FE, 0x5420, 0x5420, 0x8020, 0x20}},
    {32201, 2, {0x1000, 0x13FE, 0x2020, 0x2820, 0x4BFE, 0xF222, 0x1222, 0x2332, 0x4AAA, 0xFAAA, 0xB76, 0x266, 0xAA22, 0xAA22, 0x822A, 0x204}},
    {32202, 2, {0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x420, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32203, 2, {0x1090, 0x1090, 0x2090, 0x2B9E, 0x4890, 0xF090, 0x1090, 0x239C, 0x4890, 0xF890, 0x890, 0x39E, 0xA890, 0xA890, 0x8090, 0x90}},
    {32204, 2, {0x101C, 0x11E0, 0x2020, 0x2820, 0x4BFE, 0xF0A8, 0x1124, 0x2242, 0x4840, 0xFBFE, 0x888, 0x108, 0xA8D0, 0xA830, 0x8048, 0x184}},
    {32205, 2, {0x101C, 0x13E0, 0x2220, 0x2BFE, 0x4A20, 0xF292, 0x130A, 0x2206, 0x49FC, 0xFD04, 0x504, 0x1FC, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32206, 2, {0x1014, 0x1012, 0x2010, 0x2BFE, 0x4810, 0xF010, 0x13D2, 0x2252, 0x4A52, 0xFA54, 0xBD4, 0xC, 0xA86A, 0xAB8A, 0x8116, 0x22}},
    {32207, 2, {0x1090, 0x1090, 0x2108, 0x2948, 0x4A44, 0xF492, 0x1108, 0x23FC, 0x4844, 0xF820, 0x8A4, 0x282, 0xAA8A, 0xAA8A, 0x8478, 0x0}},
    {32208, 2, {0x4000, 0x400C, 0x7EF0, 0x8020, 0x7C44, 0x44F8, 0x6410, 0x5424, 0xFEFE, 0x4412, 0xA450, 0x9454, 0xFE92, 0x512, 0x2850, 0x1020}},
    {32209, 2, {0x1000, 0x11F8, 0x2008, 0x2808, 0x49F8, 0xF008, 0x1008, 0x23FE, 0x4820, 0xFA22, 0x974, 0xA8, 0xA924, 0xAA22, 0x80A0, 0x40}},
    {32210, 2, {0x1020, 0x1022, 0x21FA, 0x2424, 0x4428, 0xFBFE, 0x1020, 0x2040, 0x48FC, 0xFD84, 0x684, 0xFC, 0x5484, 0x5484, 0x80FC, 0x84}},
    {32211, 2, {0x1088, 0x1088, 0x23FE, 0x2888, 0x48A8, 0xF020, 0x11FC, 0x2124, 0x4924, 0xF924, 0xBFE, 0x20, 0xA850, 0xA888, 0x8104, 0x202}},
    {32212, 2, {0x1020, 0x1124, 0x20A4, 0x28A8, 0x4820, 0xF3FE, 0x1202, 0x2202, 0x4AFA, 0xFA8A, 0xA8A, 0x28A, 0xAAFA, 0xAA02, 0x820A, 0x204}},
    {32213, 2, {0x1040, 0x1020, 0x23FE, 0x2488, 0x4450, 0xF820, 0x10D8, 0x2306, 0x4888, 0xFCF8, 0x488, 0x88, 0x54F8, 0x5508, 0x8108, 0x208}},
    {32214, 2, {0x1020, 0x1022, 0x21FA, 0x2424, 0x4428, 0xFBFE, 0x1024, 0x2042, 0x48FC, 0xFD84, 0x684, 0xFC, 0x5484, 0x5484, 0x80FC, 0x84}},
    {32215, 2, {0x1100, 0x1100, 0x213C, 0x2924, 0x4BA4, 0xF124, 0x113C, 0x23A4, 0x4B64, 0xFD24, 0x93C, 0x124, 0xA924, 0xA924, 0x813C, 0x124}},
    {32216, 2, {0x1008, 0x100C, 0x200A, 0x2BFE, 0x4A08, 0xF208, 0x12E8, 0x220A, 0x4A0A, 0xFAEC, 0xAAC, 0x2A8, 0xAAEA, 0xAA1A, 0x8226, 0x442}},
    {32217, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x44F8, 0xF820, 0x11FC, 0x2124, 0x4924, 0xFDFC, 0x420, 0x3FE, 0x5420, 0x5420, 0x8020, 0x20}},
    {32218, 2, {0x1040, 0x1080, 0x21FC, 0x2904, 0x49FC, 0xF104, 0x11FC, 0x2020, 0x4832, 0xFBB4, 0x8A8, 0x128, 0xA924, 0xAA24, 0x84A2, 0x40}},
    {32219, 2, {0x1000, 0x13FE, 0x2020, 0x2440, 0x45FC, 0xF954, 0x1154, 0x2154, 0x4954, 0xFD2C, 0x420, 0x3FE, 0x5450, 0x5488, 0x8104, 0x202}},
    {32220, 2, {0x1000, 0x200C, 0x7CF0, 0x4420, 0x7C44, 0x44F8, 0x7C10, 0x1024, 0x10FE, 0xFE12, 0x9250, 0x9254, 0x9292, 0x9712, 0x1050, 0x1020}},
    {32221, 2, {0x1000, 0x11F8, 0x2108, 0x2908, 0x49F8, 0xF000, 0x17FE, 0x2108, 0x49F8, 0xF908, 0x9F8, 0x108, 0xA93E, 0xAFC8, 0x8008, 0x8}},
    {32222, 2, {0x1080, 0x1338, 0x2228, 0x2A28, 0x4BA8, 0xF246, 0x1200, 0x227C, 0x4BA4, 0xFA24, 0xA28, 0x3A8, 0xAE10, 0xAA28, 0x8244, 0x282}},
    {32223, 2, {0x1008, 0x103C, 0x21E0, 0x2820, 0x4BFE, 0xF020, 0x11FC, 0x2124, 0x49FC, 0xFD24, 0x5FC, 0x20, 0x55FC, 0x5420, 0x83FE, 0x0}},
    {32224, 2, {0x1080, 0x1040, 0x23FC, 0x2800, 0x4908, 0xF090, 0x17FE, 0x2442, 0x4844, 0xFBFC, 0xA44, 0x244, 0xAA54, 0xAA48, 0x8040, 0x40}},
    {32225, 2, {0x1000, 0x11FC, 0x2104, 0x25FC, 0x4510, 0xF9FE, 0x1110, 0x214A, 0x4986, 0xFC00, 0x5FC, 0x104, 0x55FC, 0x5504, 0x81FC, 0x104}},
    {32226, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x4488, 0xF800, 0x11FC, 0x2124, 0x4924, 0xFD24, 0x5FC, 0x124, 0x5524, 0x5524, 0x81FC, 0x104}},
    {32227, 2, {0x1080, 0x10FC, 0x2104, 0x29F8, 0x4808, 0xF3FE, 0x1040, 0x20A2, 0x4B34, 0xF858, 0x894, 0x334, 0xA852, 0xA890, 0x8350, 0x20}},
    {32228, 2, {0x1048, 0x1148, 0x2148, 0x2BFE, 0x4948, 0xF148, 0x1178, 0x2100, 0x49FE, 0xF820, 0xBFE, 0x70, 0xA8A8, 0xA924, 0x8622, 0x20}},
    {32229, 2, {0x1100, 0x117C, 0x2144, 0x2A44, 0x4A44, 0xF67C, 0x1A10, 0x2210, 0x4AFE, 0xFA38, 0xA54, 0x254, 0xAA94, 0xAB12, 0x8210, 0x210}},
    {32230, 2, {0x1000, 0x11FC, 0x2124, 0x2524, 0x45FC, 0xF924, 0x1124, 0x21FC, 0x4820, 0xFC10, 0x454, 0x142, 0x5542, 0x554A, 0x8238, 0x0}},
    {32231, 2, {0x1088, 0x1048, 0x2050, 0x2BFE, 0x4850, 0xF050, 0x11FC, 0x2154, 0x4954, 0xFD8C, 0x504, 0x1FC, 0x5504, 0x5504, 0x81FC, 0x104}},
    {32232, 2, {0x1040, 0x1020, 0x21FE, 0x2502, 0x4502, 0xF9FE, 0x1100, 0x2100, 0x49FE, 0xFDAA, 0x5AA, 0x2FE, 0xAAAA, 0xAAAA, 0x84A2, 0x86}},
    {32233, 2, {0x101E, 0x13E0, 0x2044, 0x2524, 0x4488, 0xF9FC, 0x1040, 0x2040, 0x4BFE, 0xFC80, 0x4FC, 0x144, 0x5528, 0x5610, 0x8068, 0x186}},
    {32234, 2, {0x1100, 0x117E, 0x2110, 0x2510, 0x4590, 0xF95E, 0x1322, 0x252A, 0x4926, 0xFD42, 0x552, 0x10C, 0x5504, 0x5504, 0x817E, 0x100}},
    {32235, 2, {0x1080, 0x1080, 0x20FC, 0x2954, 0x4AD4, 0xF0B4, 0x112C, 0x2244, 0x4894, 0xF908, 0x840, 0xA4, 0xAAAA, 0xAA8A, 0x8478, 0x0}},
    {32236, 2, {0x1000, 0x17FE, 0x2040, 0x2880, 0x4BFC, 0xF294, 0x1294, 0x22F4, 0x4A94, 0xFA94, 0xAF4, 0x294, 0xAA94, 0xAA94, 0x83FC, 0x204}},
    {32237, 2, {0x1000, 0x13FE, 0x2222, 0x2BFE, 0x4A22, 0xF3FE, 0x1000, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x5504, 0x8114, 0x108}},
    {32238, 2, {0x1100, 0x1100, 0x21FE, 0x2A00, 0x4DFC, 0xF104, 0x11FC, 0x2104, 0x49FC, 0xFC80, 0x4FC, 0x108, 0x5690, 0x5460, 0x8198, 0x606}},
    {32239, 2, {0x1040, 0x1040, 0x21FC, 0x2884, 0x4BFE, 0xF000, 0x11FC, 0x2104, 0x49FC, 0xF820, 0xBFE, 0x20, 0xAA20, 0xABFE, 0x8020, 0x20}},
    {32240, 2, {0x1040, 0x10A0, 0x2110, 0x2A08, 0x4DF6, 0xF000, 0x13C4, 0x2254, 0x4A54, 0xFBD4, 0xA54, 0x254, 0xABD4, 0xAA44, 0x8254, 0x2C8}},
    {32241, 2, {0x1100, 0x1178, 0x2108, 0x2A08, 0x4AFE, 0xF640, 0x1A40, 0x227C, 0x4A90, 0xFA10, 0xAFE, 0x210, 0xAA28, 0xAA28, 0x8244, 0x282}},
    {32242, 2, {0x1010, 0x1010, 0x2390, 0x2A90, 0x4A94, 0xF3B2, 0x12B2, 0x22D2, 0x4B90, 0xFA94, 0xA84, 0x388, 0xAA90, 0xA820, 0x8040, 0x180}},
    {32243, 2, {0x100, 0x80, 0x3FFE, 0x2200, 0x3FBC, 0x2214, 0x2F94, 0x2214, 0x3FA4, 0x220C, 0x2420, 0x2FC0, 0x4108, 0x5FFC, 0x8888, 0x1184}},
    {32244, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF924, 0x11AC, 0x2174, 0x4924, 0xFDFC, 0x420, 0x70, 0x54A8, 0x5524, 0x8222, 0x20}},
    {32245, 2, {0x1000, 0x1088, 0x2252, 0x2A22, 0x4A52, 0xF28A, 0x1202, 0x23FE, 0x4888, 0xF944, 0xA7A, 0x88, 0xA950, 0xA820, 0x80D8, 0x306}},
    {32246, 2, {0x1100, 0x11FE, 0x2110, 0x2A10, 0x4AFE, 0xF292, 0x1692, 0x22FE, 0x4A92, 0xFA92, 0xAFE, 0x290, 0xAA50, 0xAA20, 0x8258, 0x286}},
    {32247, 2, {0x1000, 0x13FE, 0x2222, 0x2420, 0x45FE, 0xF820, 0x11FC, 0x2124, 0x49FC, 0xFD24, 0x5FC, 0x20, 0x57FE, 0x5420, 0x8020, 0x20}},
    {32248, 2, {0x1000, 0x13FE, 0x2050, 0x2450, 0x45FC, 0xF954, 0x1154, 0x2154, 0x49FC, 0xFC20, 0x420, 0x1FC, 0x5420, 0x5420, 0x83FE, 0x0}},
    {32249, 2, {0x1000, 0x11FC, 0x2104, 0x2904, 0x49FC, 0xF104, 0x1104, 0x21FC, 0x4800, 0xFBFE, 0x820, 0x120, 0xA93C, 0xA920, 0x82A0, 0x47E}},
    {32250, 2, {0x1000, 0x10FC, 0x2084, 0x2484, 0x44F4, 0xF894, 0x1094, 0x21FE, 0x4902, 0xFD7A, 0x54A, 0x14A, 0x557A, 0x5502, 0x810A, 0x104}},
    {32251, 2, {0x2008, 0x2008, 0x47C8, 0x4210, 0x921E, 0xE4A4, 0x27D4, 0x4114, 0x9114, 0xF7D4, 0x114, 0x1108, 0xA9C8, 0xAE14, 0x8414, 0x22}},
    {32252, 2, {0x1000, 0x10F8, 0x2088, 0x2488, 0x44F8, 0xF888, 0x1088, 0x20F8, 0x4800, 0xFDFC, 0x554, 0x154, 0x5554, 0x5554, 0x83FE, 0x0}},
    {32253, 2, {0x1020, 0x1020, 0x203E, 0x2420, 0x4420, 0xF9FC, 0x1104, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5400, 0x5488, 0x8104, 0x202}},
    {32254, 2, {0x1000, 0x11FE, 0x2100, 0x257C, 0x4554, 0xF954, 0x117C, 0x2154, 0x4954, 0xFD7C, 0x510, 0x110, 0x557C, 0x5610, 0x8210, 0x4FE}},
    {32255, 2, {0x1040, 0x1090, 0x2108, 0x2BFC, 0x4804, 0xF000, 0x11F8, 0x2108, 0x4908, 0xF9F8, 0x840, 0x24, 0xAAA2, 0xAA8A, 0x8488, 0x78}},
    {32256, 2, {0x1000, 0x13DE, 0x2242, 0x2A42, 0x4A42, 0xF3DE, 0x1200, 0x223E, 0x4BD2, 0xFA12, 0xA14, 0x3D4, 0xAA08, 0xAA14, 0x8224, 0x242}},
    {32257, 2, {0x1000, 0x11FC, 0x2004, 0x28FC, 0x4804, 0xF3FE, 0x1040, 0x20A2, 0x4B34, 0xF858, 0x894, 0x334, 0xA852, 0xA890, 0x8350, 0x20}},
    {32258, 2, {0x1088, 0x1048, 0x2050, 0x25FC, 0x4504, 0xF904, 0x1104, 0x21FC, 0x4820, 0xFC10, 0x454, 0x142, 0x5542, 0x554A, 0x8238, 0x0}},
    {32259, 2, {0x1000, 0x13FE, 0x2022, 0x2520, 0x453C, 0xF920, 0x12FE, 0x2400, 0x49FC, 0xFD04, 0x5FC, 0x104, 0x55FC, 0x5504, 0x8114, 0x108}},
    {32260, 2, {0x1000, 0x11FC, 0x2124, 0x25FC, 0x4524, 0xF9FC, 0x1020, 0x21FC, 0x4924, 0xFDFC, 0x524, 0x1FC, 0x5420, 0x5422, 0x8022, 0x1E}},
    {32261, 2, {0x100C, 0x100A, 0x2008, 0x2BFE, 0x4A08, 0xF3F8, 0x124A, 0x224A, 0x4BEA, 0xFAAC, 0xAAC, 0x24C, 0xAAAA, 0xAB0A, 0x8216, 0x422}},
    {32262, 2, {0x2100, 0x213C, 0x4100, 0x4100, 0x91BC, 0xE564, 0x2524, 0x453C, 0x9524, 0xF124, 0x124, 0x113C, 0xA924, 0xA100, 0x817E, 0x100}},
    {32263, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4C04, 0xF1FC, 0x1000, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x5400, 0x83FE, 0x0}},
    {32264, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4202, 0x8424, 0x1FC0, 0x180, 0x610, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {32265, 2, {0x1000, 0x13FC, 0x2090, 0x2A94, 0x4998, 0xF090, 0x17FE, 0x2000, 0x49F8, 0xFD08, 0x508, 0x1F8, 0x5508, 0x5508, 0x81F8, 0x108}},
    {32266, 2, {0x1104, 0x1084, 0x2088, 0x2800, 0x4BFE, 0xF000, 0x1088, 0x2104, 0x4A02, 0xF9FC, 0x954, 0x154, 0xA954, 0xA954, 0x87FE, 0x0}},
    {32267, 2, {0x1010, 0x1220, 0x217C, 0x2544, 0x4444, 0xF87C, 0x1340, 0x2140, 0x497C, 0xFD44, 0x544, 0x17C, 0xA944, 0xAA80, 0x847E, 0x0}},
    {32268, 2, {0x1244, 0x1124, 0x2128, 0x247E, 0x4410, 0xFB54, 0x1154, 0x2154, 0x497C, 0xFD14, 0x520, 0x120, 0xA940, 0xAA80, 0x847E, 0x0}},
    {32269, 2, {0x1020, 0x1010, 0x21FE, 0x2484, 0x4448, 0xF9FE, 0x1102, 0x2224, 0x4810, 0xFDFE, 0x440, 0x7C, 0x5444, 0x5484, 0x8094, 0x108}},
    {32270, 2, {0x1000, 0x11FC, 0x2104, 0x25E4, 0x4524, 0xFBFE, 0x1202, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x5504, 0x8114, 0x108}},
    {32271, 2, {0x878, 0x3E48, 0x228E, 0x2A00, 0x7EF8, 0x2248, 0x2A30, 0x46CC, 0x8440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32272, 2, {0x1080, 0x1080, 0x21FC, 0x2A44, 0x4954, 0xF1F4, 0x1084, 0x2128, 0x4A90, 0xF9FC, 0xA44, 0x554, 0xA9F4, 0xA884, 0x8128, 0x210}},
    {32273, 2, {0x1088, 0x1050, 0x2000, 0x25FE, 0x4450, 0xF9FC, 0x1054, 0x23FE, 0x4854, 0xFDFC, 0x450, 0xD8, 0x5554, 0x5652, 0x8050, 0x50}},
    {32274, 2, {0x1084, 0x1044, 0x2048, 0x25FE, 0x4420, 0xF8FC, 0x1020, 0x21FE, 0x4840, 0xFC80, 0x4FE, 0x110, 0x5610, 0x5410, 0x81FE, 0x0}},
    {32275, 2, {0x1000, 0x11FE, 0x2110, 0x2520, 0x457C, 0xF944, 0x117C, 0x2144, 0x497C, 0xFD10, 0x510, 0x154, 0xA952, 0xAA92, 0x8250, 0x420}},
    {32276, 2, {0x11F8, 0x1088, 0x2070, 0x298C, 0x4800, 0xF3DE, 0x1252, 0x218C, 0x4A52, 0xF820, 0xBFE, 0x70, 0xA8A8, 0xA924, 0x8622, 0x20}},
    {32277, 2, {0x1000, 0x10F8, 0x20A8, 0x24A8, 0x44D8, 0xF888, 0x10F8, 0x2000, 0x4800, 0xFDFC, 0x554, 0x154, 0x5554, 0x5554, 0x83FE, 0x0}},
    {32278, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4C24, 0xF1FC, 0x1020, 0x21FC, 0x4820, 0xFBFE, 0x820, 0x1FC, 0xA904, 0xA904, 0x81FC, 0x104}},
    {32279, 2, {0x1040, 0x1020, 0x23FE, 0x2800, 0x49FC, 0xF104, 0x17FE, 0x2104, 0x49FC, 0xFC52, 0x494, 0x188, 0x5688, 0x54A4, 0x80C2, 0x80}},
    {32280, 2, {0x101E, 0x13E0, 0x2122, 0x2894, 0x4840, 0xF088, 0x11F0, 0x2020, 0x48C4, 0xF9FE, 0x822, 0x20, 0xABFE, 0xA850, 0x8088, 0x306}},
    {32281, 2, {0x1088, 0x1088, 0x21FE, 0x2488, 0x4400, 0xF9FE, 0x1088, 0x2088, 0x48F8, 0xFC88, 0x4F8, 0x88, 0x549E, 0x55E8, 0x8008, 0x8}},
    {32282, 2, {0x101E, 0x11E0, 0x2022, 0x2512, 0x4494, 0xF880, 0x1020, 0x21CE, 0x4902, 0xFD02, 0x5CE, 0x102, 0x5502, 0x5502, 0x81FE, 0x102}},
    {32283, 2, {0x1028, 0x1024, 0x23FE, 0x2420, 0x45FC, 0xF924, 0x11FC, 0x2124, 0x49FC, 0xFD24, 0x408, 0x3FE, 0x5488, 0x5448, 0x8048, 0x18}},
    {32284, 2, {0x1000, 0x10FC, 0x2084, 0x2484, 0x44FC, 0xF800, 0x11FE, 0x2102, 0x49FE, 0xFD02, 0x5FE, 0x102, 0x55FE, 0x5448, 0x8084, 0x102}},
    {32285, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF904, 0x11FC, 0x2104, 0x49FC, 0xFD04, 0x5FC, 0x104, 0x57FE, 0x5488, 0x8104, 0x202}},
    {32286, 2, {0x1040, 0x1020, 0x23FC, 0x2800, 0x49F8, 0xF108, 0x11F8, 0x2000, 0x4BFC, 0xFA04, 0xAF4, 0x294, 0xAAF4, 0xAA04, 0x8214, 0x208}},
    {32287, 2, {0x1000, 0x13FE, 0x2200, 0x22FC, 0x4A00, 0xFBFE, 0x12A8, 0x2290, 0x4AC8, 0xFA86, 0xA08, 0x13FE, 0xAA88, 0xA448, 0x8408, 0x818}},
    {32288, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x2044, 0x4444, 0x7828, 0x1428, 0xFE10, 0x5428, 0x9244, 0x3182}},
    {32289, 2, {0x1040, 0x1020, 0x23FE, 0x2A42, 0x4820, 0xF1FC, 0x1088, 0x2050, 0x4BFE, 0xFC20, 0x420, 0x1FC, 0x5420, 0x5420, 0x8020, 0x20}},
    {32290, 2, {0x20, 0x7924, 0x48A8, 0x49FC, 0x4840, 0x7BFE, 0x4888, 0x4944, 0x4A8A, 0x78F0, 0x4824, 0x49FC, 0x4820, 0x4924, 0x4A22, 0x9860}},
    {32291, 2, {0x0, 0x7C0C, 0x44F0, 0x7C20, 0x4444, 0x7CF8, 0x4410, 0x7C24, 0x44FE, 0xFE12, 0x1050, 0x5454, 0x5292, 0x9312, 0x5050, 0x2020}},
    {32292, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF820, 0x13FE, 0x2040, 0x4888, 0xFDF0, 0x424, 0x3FE, 0x5422, 0x5524, 0x82A2, 0x40}},
    {32293, 2, {0x1020, 0x1020, 0x23FE, 0x2420, 0x45FC, 0xF840, 0x13FE, 0x2088, 0x4934, 0xFEE2, 0x420, 0x1FC, 0x5470, 0x54AC, 0x8322, 0x20}},
    {32294, 2, {0x1080, 0x10A2, 0x2112, 0x2514, 0x4680, 0xF8FE, 0x1108, 0x2108, 0x4B28, 0xFD2E, 0x528, 0x128, 0x5528, 0x5558, 0x814E, 0x180}},
    {32295, 2, {0x1110, 0x111E, 0x2122, 0x2A54, 0x4A08, 0xF694, 0x1AE2, 0x2288, 0x4A88, 0xFABE, 0xA88, 0x2AC, 0xAA2A, 0xAA4A, 0x8228, 0x210}},
    {32296, 2, {0x1000, 0x11FC, 0x2104, 0x25FC, 0x4504, 0xF9FC, 0x1020, 0x2124, 0x48A8, 0xFC20, 0x7FE, 0x90, 0x5490, 0x5512, 0x8212, 0x40E}},
    {32297, 2, {0x1100, 0x1110, 0x23DC, 0x2A54, 0x4D54, 0xF288, 0x1108, 0x22F4, 0x4C02, 0xF800, 0xBFC, 0x40, 0xA950, 0xAA48, 0x8544, 0x80}},
    {32298, 2, {0x1000, 0x11FC, 0x2124, 0x25FC, 0x4524, 0xF9FC, 0x10A8, 0x20A8, 0x4BFE, 0xFCA8, 0x4A8, 0x1FC, 0xA820, 0xABFE, 0x8020, 0x20}},
    {32299, 2, {0x1020, 0x147C, 0x2284, 0x2948, 0x4830, 0xF0CE, 0x1610, 0x227C, 0x4A10, 0xFA7C, 0xA10, 0x2FE, 0xAA10, 0xAA10, 0x85FE, 0x800}},
    {32300, 2, {0x100C, 0x100A, 0x2008, 0x23FE, 0x4A08, 0xFA48, 0x126A, 0x224A, 0x4BFA, 0xFA4C, 0xA4C, 0x12EC, 0xAB5A, 0xA44A, 0x84D6, 0x822}},
    {32301, 2, {0x1040, 0x1020, 0x23FE, 0x2800, 0x4954, 0xF124, 0x1154, 0x21FC, 0x4820, 0xFBFE, 0xA42, 0x292, 0xAAFA, 0xAA0A, 0x8202, 0x206}},
    {32302, 2, {0x1020, 0x1010, 0x23FE, 0x2A02, 0x4880, 0xF0FE, 0x1110, 0x2320, 0x4D7C, 0xF944, 0x944, 0x17C, 0xA944, 0xA944, 0x817C, 0x144}},
    {32303, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x4800, 0xF1FC, 0x1020, 0x21FC, 0x4924, 0xFDFC, 0x524, 0x1FC, 0x5400, 0x5488, 0x8104, 0x202}},
    {32304, 2, {0x1108, 0x1108, 0x2228, 0x24AE, 0x44A8, 0xF928, 0x137E, 0x2500, 0x4908, 0xFD28, 0x52E, 0x128, 0x5558, 0x5548, 0x8186, 0x100}},
    {32305, 2, {0x1124, 0x1124, 0x2224, 0x2C24, 0x4954, 0xF14A, 0x1292, 0x2610, 0x4A10, 0xFA50, 0xA5C, 0x250, 0xAA50, 0xAAB0, 0x829E, 0x300}},
    {32306, 2, {0x1000, 0x11FC, 0x2124, 0x25FC, 0x4524, 0xF9FC, 0x1040, 0x2088, 0x49F0, 0xFC20, 0x444, 0x3FE, 0x5422, 0x5524, 0x8222, 0x60}},
    {32307, 2, {0x1020, 0x13FE, 0x2020, 0x25FC, 0x4524, 0xF9FC, 0x1124, 0x21FC, 0x4822, 0xFDFE, 0x408, 0x1FE, 0x5508, 0x5488, 0x8028, 0x10}},
    {32308, 2, {0x1020, 0x13FE, 0x2048, 0x20F0, 0x4820, 0xFBFE, 0x1242, 0x2090, 0x49F8, 0xF828, 0x920, 0x11FC, 0xAA20, 0xA7FE, 0x8020, 0x20}},
    {32309, 2, {0x1000, 0x11F8, 0x2108, 0x29F8, 0x4908, 0xF1F8, 0x1000, 0x23FC, 0x4A94, 0xFBFC, 0x800, 0x1F8, 0xA890, 0xA860, 0x8198, 0x606}},
    {32310, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0xA82, 0x420, 0xFC0, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32311, 2, {0x1020, 0x11FC, 0x2124, 0x2BFE, 0x4924, 0xF1FC, 0x1020, 0x21FC, 0x4924, 0xFDFC, 0x440, 0x3FE, 0x5488, 0x55D0, 0x8070, 0x38C}},
    {32312, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x4400, 0xF9FC, 0x1104, 0x21FC, 0x4904, 0xFDFC, 0x420, 0x3FE, 0x5450, 0x5488, 0x8104, 0x202}},
    {32313, 2, {0x1000, 0x13FE, 0x2050, 0x25FC, 0x4554, 0xF954, 0x11FC, 0x2000, 0x49FC, 0xFC00, 0x7FE, 0x20, 0x54A8, 0x5524, 0x82A2, 0x40}},
    {32314, 2, {0x1010, 0x1010, 0x22FE, 0x2510, 0x457C, 0xF854, 0x107C, 0x2354, 0x497C, 0xFD10, 0x5FE, 0x110, 0xA910, 0xAA90, 0x847E, 0x0}},
    {32315, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2638, 0x2B54, 0x3292, 0x2108, 0x27F0, 0x2060, 0x2384, 0x2FFE, 0x4042, 0x4548, 0x8884}},
    {32316, 2, {0x1210, 0x1110, 0x2110, 0x27BE, 0x4A40, 0xFA00, 0x13BE, 0x228A, 0x4A88, 0xFAA8, 0xAAE, 0x12A8, 0xACA8, 0xA5A8, 0x885E, 0x1080}},
    {32317, 2, {0x1020, 0x1040, 0x21FC, 0x2524, 0x457C, 0xF98C, 0x1154, 0x2124, 0x4954, 0xFDFC, 0x420, 0x10, 0x5554, 0x5542, 0x814A, 0x238}},
    {32318, 2, {0x1020, 0x13FE, 0x2020, 0x25FC, 0x4420, 0xFBFE, 0x1000, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x55FC, 0x8088, 0x104}},
    {32319, 2, {0x1040, 0x1088, 0x21FC, 0x2908, 0x4A52, 0xF3FE, 0x1050, 0x2188, 0x4E26, 0xF8C0, 0xB10, 0x64, 0xAB88, 0xA830, 0x80C0, 0x700}},
    {32320, 2, {0x1020, 0x1222, 0x23FE, 0x2890, 0x4888, 0xF1FE, 0x1310, 0x2510, 0x49FE, 0xFD10, 0x510, 0x1FE, 0x5510, 0x5510, 0x81FE, 0x100}},
    {32321, 2, {0x2020, 0x3F20, 0x403E, 0xBE48, 0x2AA8, 0xFF28, 0x4A10, 0x7F28, 0x446, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32322, 2, {0x1040, 0x1020, 0x23FE, 0x2820, 0x4848, 0xF2F2, 0x1124, 0x2252, 0x48F8, 0xFC08, 0x420, 0x3FE, 0x5420, 0x5420, 0x8020, 0x20}},
    {32323, 2, {0x1020, 0x1222, 0x2222, 0x2BFE, 0x4800, 0xF3DE, 0x1252, 0x2252, 0x4BDE, 0xFA52, 0xA52, 0x3DE, 0xAA52, 0xAA52, 0x855A, 0x8A4}},
    {32324, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32325, 2, {0x1124, 0x1248, 0x2124, 0x2800, 0x49FC, 0xF124, 0x11FC, 0x2124, 0x49FC, 0xF820, 0xBFE, 0x70, 0xA8A8, 0xA924, 0x8622, 0x20}},
    {32326, 2, {0x1000, 0x13DE, 0x2042, 0x2A52, 0x494A, 0xF252, 0x1028, 0x20C4, 0x4B12, 0xF860, 0x988, 0x32, 0xA9C4, 0xA818, 0x8060, 0x380}},
    {32327, 2, {0x0, 0xE0C, 0xF0F0, 0x2220, 0x9244, 0x44F8, 0x2010, 0x3C24, 0x50FE, 0x1012, 0xFE50, 0x1054, 0x5492, 0x5512, 0x5C50, 0x6420}},
    {32328, 2, {0x1008, 0x1388, 0x2090, 0x28A4, 0x48BE, 0xF38A, 0x1208, 0x223E, 0x4A2A, 0xFBAA, 0x8AA, 0xBE, 0xA888, 0xA88A, 0x82BE, 0x102}},
    {32329, 2, {0x11FC, 0x1104, 0x21FC, 0x2504, 0x45FC, 0xF904, 0x11FC, 0x2050, 0x4954, 0xFCD8, 0x4D8, 0x154, 0x5652, 0x5490, 0x8092, 0x10E}},
    {32330, 2, {0x1088, 0x108C, 0x23EA, 0x2888, 0x4888, 0xF7FE, 0x1008, 0x214A, 0x494A, 0xFD5A, 0xB6C, 0x14C, 0xA94A, 0xA9EA, 0x8716, 0x22}},
    {32331, 2, {0x1078, 0xFE48, 0x1048, 0x7C86, 0x5478, 0x7C48, 0x5430, 0xFECE, 0x1440, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32332, 2, {0x1040, 0x1080, 0x23FC, 0x2A94, 0x4A64, 0xF294, 0x13FC, 0x2090, 0x4908, 0xFA84, 0xCFA, 0x108, 0xAA90, 0xA860, 0x8198, 0x606}},
    {32333, 2, {0x1020, 0x1020, 0x21FC, 0x2424, 0x47FE, 0xF824, 0x11FC, 0x2020, 0x49AC, 0xFD74, 0x524, 0x1FC, 0xA964, 0xAAB4, 0x832C, 0x424}},
    {32334, 2, {0x1110, 0x1114, 0x21D2, 0x2A50, 0x4A7E, 0xF550, 0x1090, 0x22A8, 0x4928, 0xF944, 0xA44, 0x482, 0xA808, 0xAAA4, 0x8252, 0x452}},
    {32335, 2, {0x1000, 0x13DE, 0x2252, 0x2BDE, 0x4A10, 0xF252, 0x11CE, 0x2000, 0x4888, 0xFBFE, 0x888, 0x88, 0xABFE, 0xA888, 0x8104, 0x202}},
    {32336, 2, {0x1020, 0x1020, 0x23FE, 0x2820, 0x49FC, 0xF124, 0x11FC, 0x2124, 0x49FC, 0xF822, 0xBFE, 0x42, 0xA824, 0xAAA2, 0x828A, 0x478}},
    {32337, 2, {0x1018, 0x11E0, 0x2040, 0x2BFE, 0x4888, 0xF174, 0x1252, 0x2070, 0x4800, 0xFDFC, 0x504, 0x174, 0x5554, 0x5574, 0x8104, 0x10C}},
    {32338, 2, {0x1104, 0x1088, 0x2000, 0x2BFE, 0x4A22, 0xF2AA, 0x1272, 0x2222, 0x4BFE, 0xF800, 0x9FC, 0x104, 0xA9FC, 0xA904, 0x81FC, 0x104}},
    {32339, 2, {0x1000, 0x11F8, 0x2108, 0x21F8, 0x4908, 0xF9F8, 0x1000, 0x27FE, 0x4A40, 0xFBDC, 0xA54, 0x13D4, 0xAA68, 0xA7C8, 0x8054, 0x62}},
    {32340, 2, {0x1110, 0x1094, 0x23D2, 0x2812, 0x4A50, 0xF190, 0x17FE, 0x2010, 0x4BD2, 0xFA52, 0xA54, 0x3D4, 0xAA4A, 0xAA4A, 0x83D6, 0x22}},
    {32341, 2, {0x1088, 0x1050, 0x23FE, 0x2820, 0x49FC, 0xF020, 0x13FE, 0x2124, 0x48A8, 0xFBFE, 0x800, 0x1FC, 0xA904, 0xA904, 0x81FC, 0x104}},
    {32342, 2, {0x2288, 0x2288, 0x47C8, 0x4290, 0x929E, 0xEFD4, 0x2024, 0x47D4, 0x9454, 0xF7D4, 0x454, 0x17C8, 0xAC48, 0xA454, 0x8464, 0x4C2}},
    {32343, 2, {0x1020, 0x1124, 0x20A8, 0x2BFE, 0x48A8, 0xF124, 0x1222, 0x2104, 0x4904, 0xF9DE, 0xA44, 0x554, 0xA89E, 0xA904, 0x8204, 0x404}},
    {32344, 2, {0x1000, 0x13FC, 0x2048, 0x2830, 0x4BFE, 0xF052, 0x1094, 0x2350, 0x4820, 0xFBFE, 0xA52, 0x28A, 0xAB76, 0xAA52, 0x8272, 0x206}},
    {32345, 2, {0x103C, 0x13E0, 0x2124, 0x28A8, 0x4BFE, 0xF0A8, 0x1124, 0x2202, 0x49FC, 0xFD24, 0x524, 0x1FC, 0x5524, 0x5524, 0x81FC, 0x104}},
    {32346, 2, {0x1020, 0x1020, 0x21FC, 0x2420, 0x4554, 0xF888, 0x1104, 0x22FA, 0x4888, 0xFCF8, 0x488, 0xF8, 0x5420, 0x54A8, 0x8124, 0x60}},
    {32347, 2, {0x810, 0x810, 0x7F1E, 0x810, 0x3E7E, 0x842, 0xFF7E, 0x2242, 0x7C7E, 0x852, 0x1210, 0x7F7E, 0x810, 0x2A10, 0x4910, 0x1810}},
    {32348, 2, {0x1088, 0x1050, 0x23FE, 0x2450, 0x45FC, 0xF954, 0x118C, 0x2174, 0x4904, 0xFDFC, 0x408, 0x3FE, 0x5508, 0x5488, 0x80A8, 0x10}},
    {32349, 2, {0x13DE, 0x1252, 0x23DE, 0x2A52, 0x4BDE, 0xF202, 0x12F2, 0x2292, 0x4AF2, 0xFA92, 0xAF2, 0x292, 0xAA92, 0xAB32, 0x8202, 0x206}},
    {32350, 2, {0x1020, 0x1020, 0x21FC, 0x2820, 0x4BFE, 0xF108, 0x139C, 0x2108, 0x4988, 0xFE3E, 0x800, 0x3FE, 0xA890, 0xA890, 0x8112, 0x20E}},
    {32351, 2, {0x1000, 0x13DE, 0x2252, 0x2A52, 0x4BDE, 0xF000, 0x11FC, 0x2124, 0x49FC, 0xFD24, 0x5FC, 0x20, 0x57FE, 0x5420, 0x8020, 0x20}},
    {32352, 2, {0x100, 0x2488, 0x2424, 0x43E4, 0x0, 0x2424, 0x1212, 0xA0A0, 0xA4A4, 0x1C1C, 0x420, 0x1FC0, 0x210, 0x3FF8, 0x920, 0x1310}},
    {32353, 2, {0x2040, 0x23F8, 0x4048, 0x47FE, 0x9048, 0xE3F8, 0x2040, 0x4554, 0x975C, 0xF444, 0x7FC, 0x1444, 0xAF5C, 0xA554, 0x8554, 0x844}},
    {32354, 2, {0x1020, 0x11FC, 0x2124, 0x25FC, 0x4420, 0xFBFE, 0x1000, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x55FC, 0x8088, 0x104}},
    {32355, 2, {0x1020, 0x11FC, 0x2024, 0x2BFE, 0x4824, 0xF1FC, 0x1020, 0x23FE, 0x4800, 0xFDFC, 0x524, 0x1FC, 0x5524, 0x55FC, 0x8000, 0x3FE}},
    {32356, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1490, 0x1250, 0x1130, 0x1FF0, 0x400, 0xFFFE, 0x1250, 0x2F88, 0xC226, 0x1FF0, 0x948, 0x1320}},
    {32357, 2, {0x1020, 0x13FE, 0x2020, 0x29FC, 0x4800, 0xF1FC, 0x1104, 0x21FC, 0x4888, 0xFBFE, 0x800, 0x1FC, 0xA904, 0xA904, 0x81FC, 0x104}},
    {32358, 2, {0x2000, 0x277C, 0x4144, 0x4144, 0x917C, 0xE710, 0x2410, 0x447C, 0x9454, 0xF754, 0x154, 0x117C, 0xA910, 0xA114, 0x8AFE, 0x402}},
    {32359, 2, {0x1000, 0x11FC, 0x2020, 0x2BFE, 0x4A22, 0xF1AC, 0x1020, 0x21AC, 0x4800, 0xFDFC, 0x400, 0x3FE, 0x5440, 0x5488, 0x81FC, 0x84}},
    {32360, 2, {0x1010, 0x1210, 0x217C, 0x2410, 0x44FE, 0xF844, 0x1328, 0x21FE, 0x4910, 0xFD7C, 0x510, 0x1FE, 0x5510, 0x5510, 0x82FE, 0x400}},
    {32361, 2, {0x1000, 0x11FC, 0x2154, 0x2554, 0x45DC, 0xF850, 0x11DC, 0x2154, 0x4954, 0xFDDC, 0x554, 0x154, 0x55DC, 0x5550, 0x8052, 0x3E}},
    {32362, 2, {0x1020, 0x1050, 0x2088, 0x2974, 0x4A02, 0xF1FC, 0x1124, 0x21AC, 0x4924, 0xFDFC, 0x400, 0xF8, 0x5488, 0x54F8, 0x8088, 0xF8}},
    {32363, 2, {0x878, 0xFFC8, 0x848, 0x3E86, 0x2B78, 0x3E48, 0x2A28, 0xFF90, 0x4928, 0x7F46, 0x420, 0x1FC0, 0x210, 0x3FF8, 0x920, 0x1310}},
    {32364, 2, {0x1020, 0x1020, 0x23FE, 0x2924, 0x4924, 0xF2AA, 0x17FE, 0x2000, 0x49FC, 0xFD04, 0x574, 0x154, 0x5574, 0x5504, 0x81FC, 0x104}},
    {32365, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x4924, 0x4924, 0x55FC, 0x5DAC, 0x49FC, 0x5524, 0x5D3C, 0x41CC, 0x6B04, 0x4114, 0x4008}},
    {32366, 2, {0x1000, 0x13FE, 0x2000, 0x25FC, 0x4524, 0xF9FC, 0x1124, 0x23FE, 0x4800, 0xFDFC, 0x524, 0x1FC, 0x5524, 0x55FC, 0x8000, 0x3FE}},
    {32367, 2, {0x1000, 0x13FC, 0x2294, 0x2A94, 0x4BFC, 0xF000, 0x17FE, 0x2000, 0x4BFC, 0xFA04, 0xBFC, 0xA2, 0xA914, 0xAB08, 0x8D44, 0x182}},
    {32368, 2, {0x11FC, 0x1104, 0x2104, 0x29FC, 0x4800, 0xF3DE, 0x1252, 0x2252, 0x4BDE, 0xF820, 0xBFE, 0x70, 0xA8A8, 0xA924, 0x8622, 0x20}},
    {32369, 2, {0x1088, 0x13FE, 0x2088, 0x2900, 0x4BFC, 0xF554, 0x10D4, 0x2324, 0x4854, 0xF894, 0xB48, 0x20, 0xA8A4, 0xAA8A, 0x828A, 0x478}},
    {32370, 2, {0x1100, 0x11DE, 0x224A, 0x2C8A, 0x4BEA, 0xF2B6, 0x12A4, 0x23F4, 0x4ABE, 0xFAA4, 0xBE4, 0x2BE, 0xAAA4, 0xAAA4, 0x8224, 0x464}},
    {32371, 2, {0x2108, 0x2208, 0x47C8, 0x4450, 0x97DE, 0xE464, 0x27D4, 0x4214, 0x9114, 0xF7D4, 0x214, 0x13C8, 0xAA48, 0xA454, 0x8554, 0x8A2}},
    {32372, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x450, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32373, 2, {0x1020, 0x17FE, 0x2000, 0x2BFE, 0x4A02, 0xF2FA, 0x128A, 0x23FE, 0x4800, 0xF9FC, 0x904, 0x1FC, 0xA904, 0xA9FC, 0x8000, 0x7FE}},
    {32374, 2, {0x1040, 0x1020, 0x21FC, 0x2488, 0x4450, 0xFBFE, 0x1000, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0xA820, 0xAA94, 0x828A, 0x47A}},
    {32375, 2, {0x1090, 0x13FC, 0x2294, 0x23FC, 0x4A94, 0xFBFC, 0x1000, 0x23FC, 0x4A00, 0xFAF8, 0xA00, 0x13FE, 0xAD20, 0xA514, 0x8548, 0x986}},
    {32376, 2, {0x1088, 0x1448, 0x2250, 0x28FC, 0x4820, 0xF050, 0x1694, 0x2238, 0x4A50, 0xFA98, 0xA34, 0x254, 0xAA90, 0xAA20, 0x85FE, 0x800}},
    {32377, 2, {0x1000, 0x11FC, 0x2154, 0x25FC, 0x4420, 0xF9FC, 0x1020, 0x23FE, 0x4888, 0xFC50, 0x5FC, 0x20, 0x57FE, 0x5420, 0x8020, 0x20}},
    {32378, 2, {0x1210, 0x1210, 0x23D0, 0x2C5E, 0x48A4, 0xF3D4, 0x1054, 0x2054, 0x4BC8, 0xF848, 0x854, 0x7E2, 0xA800, 0xAAA4, 0x8252, 0x452}},
    {32379, 2, {0x1000, 0x11FC, 0x2020, 0x2BFE, 0x4A22, 0xF1AC, 0x1020, 0x21AC, 0x4800, 0xFBFE, 0x820, 0x1FC, 0xA954, 0xA954, 0x8154, 0x10C}},
    {32380, 2, {0x1044, 0x1284, 0x22EA, 0x2A4E, 0x4AA4, 0xF2EE, 0x1222, 0x23FE, 0x4A44, 0xFA84, 0xAEA, 0x24E, 0xAAA4, 0xAAEE, 0x8222, 0x3FE}},
    {32381, 2, {0x1040, 0x1020, 0x23FE, 0x2A02, 0x49FC, 0xF148, 0x1250, 0x21FC, 0x4B04, 0xF9FC, 0x904, 0x1FC, 0xA904, 0xA9FC, 0x8088, 0x104}},
    {32382, 2, {0x1010, 0x107C, 0x2254, 0x257C, 0x4510, 0xF8FE, 0x1000, 0x237C, 0x4944, 0xFD7C, 0x540, 0x17C, 0x5544, 0x557C, 0x8280, 0x47E}},
    {32383, 2, {0x1008, 0x13E8, 0x2288, 0x2BEE, 0x4A28, 0xF3F4, 0x1282, 0x23E2, 0x4800, 0xF9FC, 0x954, 0x154, 0xA954, 0xA954, 0x87FE, 0x0}},
    {32384, 2, {0x1128, 0x11AA, 0x216C, 0x2A28, 0x4AFE, 0xF644, 0x1628, 0x2AFE, 0x4A10, 0xFA7C, 0xA10, 0x2FE, 0xAA10, 0xAA28, 0x8244, 0x282}},
    {32385, 2, {0x101C, 0x11E0, 0x2020, 0x2BFE, 0x4820, 0xF1FC, 0x11AC, 0x2174, 0x49FC, 0xFC20, 0x5FC, 0x20, 0x57FE, 0x5400, 0x8154, 0x22A}},
    {32386, 2, {0x2040, 0x3E7E, 0x4890, 0xBFF8, 0x2488, 0x2248, 0x3FF8, 0x200, 0xFFFE, 0x920, 0x1250, 0x2F88, 0xC226, 0x1FF0, 0x948, 0x1320}},
    {32387, 2, {0x2080, 0x2040, 0x47FC, 0x4110, 0x90A4, 0xEF58, 0x2554, 0x4552, 0x9B58, 0xF000, 0x208, 0x13F8, 0xAA08, 0xA3F8, 0x8208, 0x408}},
    {32388, 2, {0x1088, 0x13FE, 0x2088, 0x2820, 0x4A3C, 0xF144, 0x1038, 0x20D6, 0x4B38, 0xF910, 0x97C, 0x110, 0xA97E, 0xA910, 0x8310, 0x4FE}},
    {32389, 2, {0x2444, 0x24E4, 0x48A8, 0x4AAA, 0x9EEE, 0xE4A4, 0x2AAA, 0x4EEE, 0x9242, 0xF040, 0xFFE, 0x10E0, 0xA950, 0xA248, 0x8C46, 0x40}},
    {32390, 2, {0x11FC, 0x1124, 0x21AC, 0x2524, 0x45FC, 0xF820, 0x11FC, 0x2020, 0x4BFE, 0xFD54, 0x62A, 0x20, 0x55FC, 0x5420, 0x83FE, 0x0}},
    {32391, 2, {0x9200, 0x54FE, 0xFE10, 0x3020, 0x54FC, 0x9284, 0x2084, 0x44FC, 0x7884, 0x10FC, 0x2484, 0xFE84, 0x10FC, 0x5448, 0x9284, 0x3102}},
    {32392, 2, {0x1100, 0x113E, 0x2108, 0x2FD0, 0x493E, 0xF122, 0x13A2, 0x203E, 0x4822, 0xFBBE, 0xAA2, 0x2A2, 0xAABE, 0xAB80, 0x8294, 0x22}},
    {32393, 2, {0x1088, 0x11DC, 0x2088, 0x2BDE, 0x4888, 0xF154, 0x1222, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x55FC, 0x8088, 0x104}},
    {32394, 2, {0x1020, 0x1010, 0x21FE, 0x2528, 0x457C, 0xF928, 0x11FE, 0x2110, 0x497C, 0xFD54, 0x57C, 0x154, 0xA97C, 0xAA00, 0x8244, 0x482}},
    {32395, 2, {0x13FC, 0x1040, 0x21F8, 0x2908, 0x49F8, 0xF108, 0x17FE, 0x2442, 0x4AA4, 0xFA8A, 0xC7A, 0x80, 0xA9F8, 0xAA88, 0x8070, 0x38E}},
    {32396, 2, {0x1020, 0x13FE, 0x2020, 0x29FC, 0x4800, 0xF3FE, 0x1252, 0x23FE, 0x4904, 0xFDFC, 0x504, 0x1FC, 0x5504, 0x55FC, 0x8088, 0x104}},
    {32397, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {32398, 2, {0x1088, 0x108C, 0x23EA, 0x2888, 0x4888, 0xF3FE, 0x1148, 0x236A, 0x494A, 0xFB6A, 0x94C, 0x36C, 0xA94A, 0xA96A, 0x8796, 0x22}},
    {32399, 2, {0x2020, 0x27FE, 0x4400, 0x45FC, 0x9524, 0xE5FC, 0x2524, 0x45FC, 0x9420, 0xF5FC, 0x420, 0x17FE, 0xAD24, 0xAAFA, 0x8820, 0x13FE}},
    {32400, 2, {0x1200, 0x113E, 0x27C8, 0x2810, 0x4ABE, 0xF262, 0x1422, 0x20BE, 0x4AA2, 0xFABE, 0x922, 0x122, 0xAABE, 0xAA80, 0x8414, 0x22}},
    {32401, 2, {0x1040, 0x107C, 0x2040, 0x2BFC, 0x4A44, 0xF3F0, 0x1244, 0x22FC, 0x4AA8, 0xFAF8, 0xAA8, 0x2F8, 0xAA00, 0xADFC, 0x8554, 0xBFE}},
    {32402, 2, {0x13FE, 0x1200, 0x22FE, 0x2292, 0x4AFE, 0xFA92, 0x12FE, 0x2210, 0x4AFE, 0xFA10, 0xBFE, 0x12AA, 0xAA10, 0xA2FE, 0x8410, 0x9FE}},
    {32403, 2, {0x13DE, 0x1252, 0x23DE, 0x2A52, 0x4BDE, 0xF252, 0x13DE, 0x218C, 0x4A52, 0xF840, 0xFFE, 0x88, 0xA990, 0xA860, 0x80D8, 0x304}},
    {32404, 2, {0x11F0, 0x1210, 0x27FC, 0x2A04, 0x4BFC, 0xF224, 0x13B8, 0x2222, 0x49FE, 0xFA10, 0xFFC, 0x244, 0xABFC, 0xA8D0, 0x814A, 0x63E}},
    {32405, 2, {0x1020, 0x13FE, 0x2000, 0x29DC, 0x4954, 0xF1DC, 0x1088, 0x23FE, 0x4888, 0xFBFE, 0x888, 0x7FE, 0xA894, 0xA988, 0x86A4, 0xC2}},
    {32406, 2, {0x1148, 0x114C, 0x22AA, 0x2808, 0x4FFE, 0xF148, 0x1368, 0x214A, 0x4B6A, 0xF94C, 0xB6C, 0x148, 0xA96A, 0xAB9A, 0x8026, 0x42}},
    {32407, 2, {0x1222, 0x13FE, 0x2090, 0x29FE, 0x4B10, 0xF5FE, 0x1110, 0x21FE, 0x4910, 0xF9FE, 0x900, 0x3FE, 0xAA8A, 0xAB76, 0x8252, 0x276}},
    {32408, 2, {0x2528, 0x27BE, 0x4948, 0x47BE, 0x9318, 0xE5AA, 0x2946, 0x43FC, 0x9204, 0xF3FC, 0x204, 0x13FC, 0xAA04, 0xA3FC, 0x8108, 0x204}},
    {32409, 2, {0x2000, 0x23FE, 0x4252, 0x4252, 0x93FE, 0xE108, 0x2154, 0x425E, 0x97B4, 0xF11E, 0x254, 0x17DE, 0xA814, 0xA554, 0x855E, 0x10}},
    {32410, 2, {0x13DE, 0x1000, 0x23DE, 0x2A52, 0x4B5A, 0xF252, 0x1020, 0x23FE, 0x4A50, 0xFBFE, 0xA52, 0x3FE, 0xAA92, 0xAADC, 0x8292, 0x4CE}},
    {32411, 2, {0x100, 0xFFFE, 0x100, 0x1FF0, 0x1210, 0xFFFE, 0x1090, 0x1FF0, 0x1C, 0x3EE0, 0x2224, 0x2AF8, 0x2A24, 0xFFFE, 0x2A54, 0x59B2}},
    {32412, 2, {0x1790, 0x151E, 0x27A8, 0x2C84, 0x4FBE, 0xF52A, 0x17BE, 0x2000, 0x4BF8, 0xFA08, 0xBF8, 0x208, 0xABF8, 0xA8A0, 0x8124, 0x61C}},
    {32413, 2, {0x11FC, 0x1124, 0x21FC, 0x2924, 0x4BFE, 0xF2AA, 0x13FE, 0x22AA, 0x4BFE, 0xF888, 0x9F0, 0x64, 0xABFE, 0xA924, 0x8222, 0x60}},
    {32414, 2, {0x4444, 0x44E4, 0x4A0A, 0x4EEE, 0xA404, 0xAAEA, 0xCE0E, 0x40E0, 0x4AAA, 0x8AEA, 0xA040, 0xE024, 0x522, 0xA50A, 0xA908, 0xF8}},
    {32415, 2, {0x1000, 0x1000, 0x2000, 0x2400, 0x4400, 0xF800, 0x1000, 0x2000, 0x4000, 0xFC00, 0x4000, 0x0, 0x1C00, 0xE000, 0x4000, 0x0}},
    {32416, 2, {0x1004, 0x1004, 0x2084, 0x2484, 0x4484, 0xF884, 0x1084, 0x2084, 0x408C, 0xFCB4, 0x40C4, 0x84, 0x1C04, 0xE004, 0x4004, 0x4}},
    {32417, 2, {0x1000, 0x11FC, 0x2020, 0x2420, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4020, 0xFC20, 0x4020, 0x20, 0x1C20, 0xE020, 0x40A0, 0x40}},
    {32418, 2, {0x1000, 0x1000, 0x21FC, 0x2420, 0x4420, 0xF820, 0x1020, 0x2020, 0x4020, 0xFC20, 0x4020, 0x20, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {32419, 2, {0x1008, 0x1008, 0x2008, 0x2408, 0x45FE, 0xF808, 0x1008, 0x2008, 0x4088, 0xFC48, 0x4048, 0x8, 0x1C08, 0xE008, 0x4028, 0x10}},
    {32420, 2, {0x1008, 0x103C, 0x21E0, 0x2420, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4020, 0xFC20, 0x4020, 0x20, 0x1C20, 0xE020, 0x4020, 0x20}},
    {32421, 2, {0x1040, 0x1040, 0x2080, 0x24FE, 0x4500, 0xFA00, 0x10FC, 0x2008, 0x4010, 0xFC20, 0x4040, 0x80, 0x1D02, 0xE102, 0x40FE, 0x0}},
    {32422, 2, {0x1080, 0x1080, 0x2080, 0x2104, 0x45FE, 0xFA04, 0x1404, 0x2084, 0x4044, 0xFC24, 0x24, 0x4, 0x1C04, 0xE044, 0x4028, 0x10}},
    {32423, 2, {0x1000, 0x13FC, 0x2084, 0x2088, 0x4888, 0xF890, 0x109C, 0x2084, 0x4144, 0xF944, 0x4128, 0x128, 0x1A10, 0xE228, 0x4444, 0x182}},
    {32424, 2, {0x1040, 0x1040, 0x2040, 0x2440, 0x45F8, 0xF848, 0x1048, 0x2048, 0x4148, 0xFCC8, 0x4048, 0xA8, 0x1CAA, 0xE10A, 0x4206, 0x402}},
    {32425, 2, {0x1020, 0x1010, 0x2010, 0x25FE, 0x4500, 0xF900, 0x1100, 0x2100, 0x4100, 0xFD00, 0x4100, 0x100, 0x1D00, 0xE200, 0x4200, 0x400}},
    {32426, 2, {0x1000, 0x1000, 0x23F8, 0x2008, 0x4808, 0xF008, 0x1008, 0x23F8, 0x4208, 0xFA00, 0x4200, 0x202, 0x1A02, 0xE202, 0x41FE, 0x0}},
    {32427, 2, {0x1000, 0x1000, 0x23FC, 0x2044, 0x4844, 0xFA44, 0x1244, 0x2244, 0x4484, 0xF884, 0x4084, 0x104, 0x1904, 0xE204, 0x4428, 0x810}},
    {32428, 2, {0x1020, 0x1020, 0x2020, 0x25FE, 0x4420, 0xF820, 0x10FC, 0x2020, 0x4020, 0xFDFE, 0x4022, 0x22, 0x1C2A, 0xE024, 0x4020, 0x20}},
    {32429, 2, {0x1000, 0x1000, 0x21FC, 0x2400, 0x4400, 0xF800, 0x13FE, 0x2020, 0x4020, 0xFC40, 0x4040, 0x88, 0x1D04, 0xE3FE, 0x4102, 0x0}},
    {32430, 2, {0x1040, 0x1040, 0x2040, 0x2040, 0x4BFE, 0xF840, 0x1080, 0x2090, 0x4090, 0xF920, 0x4120, 0x248, 0x1A44, 0xE484, 0x41FE, 0x82}},
    {32431, 2, {0x1040, 0x1040, 0x2040, 0x23FC, 0x4840, 0xFA48, 0x1248, 0x2248, 0x4248, 0xFBF8, 0x4048, 0x40, 0x1842, 0xE042, 0x403E, 0x0}},
    {32432, 2, {0x1010, 0x1110, 0x2110, 0x2512, 0x4512, 0xF914, 0x11D8, 0x2110, 0x4110, 0xFD10, 0x4110, 0x112, 0x1D52, 0xE192, 0x410E, 0x0}},
    {32433, 2, {0x1040, 0x1040, 0x2040, 0x2148, 0x4944, 0xFA42, 0x1242, 0x2448, 0x4048, 0xF848, 0x4010, 0x10, 0x1820, 0xE040, 0x4180, 0x600}},
    {32434, 2, {0x1000, 0x13FC, 0x2204, 0x2204, 0x4A94, 0xFA54, 0x1224, 0x2224, 0x4254, 0xFA54, 0x4294, 0x304, 0x1A04, 0xE204, 0x4214, 0x208}},
    {32435, 2, {0x1020, 0x1020, 0x2020, 0x2020, 0x4BFE, 0xF222, 0x1222, 0x2222, 0x4252, 0xFA4A, 0x428A, 0x302, 0x1A02, 0xE202, 0x420A, 0x204}},
    {32436, 2, {0x1008, 0x103C, 0x21E0, 0x2420, 0x4420, 0xF820, 0x1020, 0x23FE, 0x4020, 0xFC20, 0x4020, 0x20, 0x1C20, 0xE020, 0x41FC, 0x0}},
    {32437, 2, {0x1110, 0x1110, 0x2110, 0x2110, 0x4910, 0xF910, 0x1110, 0x2110, 0x4110, 0xFAA8, 0x4268, 0x228, 0x1C44, 0xE444, 0x4884, 0x1102}},
    {32438, 2, {0x1040, 0x1040, 0x20A0, 0x20A0, 0x4910, 0xFA08, 0x1406, 0x2110, 0x4120, 0xF940, 0x4180, 0x100, 0x1904, 0xE104, 0x40FC, 0x0}},
    {32439, 2, {0x1010, 0x1090, 0x2090, 0x2088, 0x4908, 0xF904, 0x1204, 0x25FA, 0x4088, 0xF888, 0x4088, 0x88, 0x1908, 0xE108, 0x4228, 0x410}},
    {32440, 2, {0x1008, 0x103C, 0x23E0, 0x2220, 0x4A20, 0xFA20, 0x1220, 0x23FE, 0x4220, 0xFA10, 0x4210, 0x212, 0x1A0A, 0xE28A, 0x4306, 0x202}},
    {32441, 2, {0x1080, 0x1040, 0x2040, 0x27FC, 0x4910, 0xF110, 0x1110, 0x2110, 0x4110, 0xF8A0, 0x40A0, 0x40, 0x18A0, 0xE110, 0x4208, 0xC06}},
    {32442, 2, {0x1040, 0x1020, 0x2020, 0x27FE, 0x4480, 0xF880, 0x1080, 0x20FC, 0x4084, 0xFC84, 0x4084, 0x84, 0x1D04, 0xE104, 0x4228, 0x410}},
    {32443, 2, {0x1040, 0x1020, 0x2020, 0x2400, 0x45FE, 0xF902, 0x1204, 0x2000, 0x4000, 0xFC00, 0x4000, 0x0, 0x1C00, 0xE1FE, 0x4000, 0x0}},
    {32444, 2, {0x1004, 0x11E4, 0x2024, 0x2424, 0x4424, 0xF9E4, 0x1104, 0x2104, 0x4104, 0xFDE4, 0x4024, 0x24, 0x1C24, 0xE024, 0x4144, 0x84}},
    {32445, 2, {0x1000, 0x11FC, 0x2044, 0x2444, 0x4444, 0xF844, 0x1044, 0x21FC, 0x4084, 0xFC84, 0x4084, 0x84, 0x1C84, 0xE084, 0x47FE, 0x0}},
    {32446, 2, {0x1000, 0x11F8, 0x2008, 0x2450, 0x4420, 0xF810, 0x13FE, 0x2022, 0x4024, 0xFC20, 0x4020, 0x20, 0x1C20, 0xE020, 0x40A0, 0x40}},
    {32447, 2, {0x1050, 0x1048, 0x2040, 0x245C, 0x45E0, 0xF840, 0x105E, 0x23E0, 0x4044, 0xFC48, 0x4030, 0x22, 0x1C52, 0xE08A, 0x4306, 0x2}},
    {32448, 2, {0x1108, 0x1108, 0x2108, 0x2508, 0x47FE, 0xF908, 0x1108, 0x2108, 0x4108, 0xFDF8, 0x4108, 0x108, 0x1D08, 0xE108, 0x41F8, 0x108}},
    {32449, 2, {0x1048, 0x1248, 0x2248, 0x2248, 0x4A48, 0xF7FE, 0x1248, 0x2248, 0x4248, 0xFA48, 0x4278, 0x200, 0x1A00, 0xE200, 0x43FE, 0x0}},
    {32450, 2, {0x1050, 0x1048, 0x2048, 0x2040, 0x4BFE, 0xF880, 0x1080, 0x20FC, 0x4144, 0xF944, 0x4128, 0x128, 0x1A10, 0xE228, 0x4444, 0x182}},
    {32451, 2, {0x1040, 0x1040, 0x27FC, 0x2080, 0x4880, 0xFBE0, 0x1120, 0x2220, 0x43FC, 0xF820, 0x4128, 0x124, 0x1A22, 0xE422, 0x40A0, 0x40}},
    {32452, 2, {0x1000, 0x11F8, 0x2108, 0x2108, 0x4908, 0xF9F8, 0x1108, 0x2108, 0x4108, 0xF9F8, 0x4108, 0x108, 0x1908, 0xE108, 0x47FE, 0x0}},
    {32453, 2, {0x1020, 0x1020, 0x2020, 0x25FC, 0x4524, 0xF924, 0x1124, 0x21FC, 0x4124, 0xFD24, 0x4124, 0x1FC, 0x1D24, 0xE020, 0x4020, 0x20}},
    {32454, 2, {0x1000, 0x1000, 0x21FC, 0x2524, 0x4524, 0xF924, 0x1124, 0x2124, 0x41FC, 0xFD24, 0x4124, 0x124, 0x1D24, 0xE124, 0x41FC, 0x104}},
    {32455, 2, {0x1000, 0x1000, 0x21FC, 0x2504, 0x4504, 0xF904, 0x1104, 0x2104, 0x41FC, 0xFD04, 0x4000, 0x90, 0x1C88, 0xE104, 0x4202, 0x402}},
    {32456, 2, {0x1080, 0x1080, 0x20F8, 0x2108, 0x4B10, 0xFCA0, 0x1040, 0x20A0, 0x4318, 0xFC06, 0x40C0, 0x20, 0x1810, 0xE180, 0x4060, 0x10}},
    {32457, 2, {0x1080, 0x1080, 0x20F8, 0x2108, 0x4A10, 0xFC20, 0x13FC, 0x2004, 0x4004, 0xF804, 0x41FC, 0x4, 0x1804, 0xE004, 0x43FC, 0x4}},
    {32458, 2, {0x1020, 0x1020, 0x2124, 0x24A4, 0x44A8, 0xF820, 0x11FC, 0x2020, 0x4020, 0xFC20, 0x43FE, 0x20, 0x1C20, 0xE020, 0x4020, 0x20}},
    {32459, 2, {0x1090, 0x1090, 0x2090, 0x23FC, 0x4894, 0xF894, 0x13FC, 0x2290, 0x4290, 0xFBFE, 0x4092, 0x92, 0x191A, 0xE114, 0x4210, 0x410}},
    {32460, 2, {0x1020, 0x1020, 0x2124, 0x2524, 0x4524, 0xF924, 0x11FC, 0x2020, 0x4020, 0xFD24, 0x4124, 0x124, 0x1D24, 0xE124, 0x41FC, 0x4}},
    {32461, 2, {0x1000, 0x13FC, 0x2084, 0x2084, 0x4884, 0xF904, 0x1114, 0x2208, 0x4400, 0xF9FC, 0x4104, 0x104, 0x1904, 0xE104, 0x41FC, 0x104}},
    {32462, 2, {0x1000, 0x13FC, 0x2104, 0x2088, 0x4850, 0xF820, 0x10D8, 0x2326, 0x4020, 0xF9FC, 0x4020, 0x20, 0x1BFE, 0xE020, 0x4020, 0x20}},
    {32463, 2, {0x1000, 0x11FC, 0x2008, 0x2410, 0x4430, 0xF848, 0x1084, 0x2302, 0x4000, 0xFDFC, 0x4020, 0x20, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {32464, 2, {0x1020, 0x1020, 0x2040, 0x2488, 0x4504, 0xFBFE, 0x1002, 0x2000, 0x41FC, 0xFD04, 0x4104, 0x104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {32465, 2, {0x2100, 0x211E, 0x2112, 0x47D2, 0x5114, 0xF114, 0x27D8, 0x2114, 0x4112, 0xF7D2, 0x4112, 0x11A, 0x3114, 0xC210, 0x210, 0x410}},
    {32466, 2, {0x1028, 0x1024, 0x2024, 0x2020, 0x4BFE, 0xF820, 0x1124, 0x2124, 0x4124, 0xFBA8, 0x4128, 0x110, 0x1A12, 0xE22A, 0x4446, 0x82}},
    {32467, 2, {0x1020, 0x1020, 0x2020, 0x27FE, 0x4420, 0xF820, 0x11FC, 0x2000, 0x4000, 0xFDFC, 0x4104, 0x104, 0x1D04, 0xE104, 0x41FC, 0x104}},
    {32468, 2, {0x1040, 0x1040, 0x27FC, 0x20A0, 0x4910, 0xFA08, 0x1DF6, 0x2000, 0x47FC, 0xF880, 0x4100, 0x3F8, 0x1808, 0xE008, 0x4050, 0x20}},
    {32469, 2, {0x1080, 0x1080, 0x20BC, 0x23C0, 0x4850, 0xF824, 0x10D4, 0x230C, 0x4000, 0xFBFE, 0x4090, 0x90, 0x1912, 0xE112, 0x420E, 0x400}},
    {32470, 2, {0x1000, 0x11FE, 0x2020, 0x2440, 0x4488, 0xF904, 0x11FE, 0x2022, 0x4020, 0xFC20, 0x41FE, 0x20, 0x1C20, 0xE020, 0x43FE, 0x0}},
    {32471, 2, {0x1080, 0x1080, 0x213C, 0x2200, 0x4C80, 0xF080, 0x117E, 0x2308, 0x4508, 0xF908, 0x4108, 0x108, 0x1908, 0xE108, 0x4128, 0x110}},
    {32472, 2, {0x1040, 0x1040, 0x20A0, 0x2110, 0x4A08, 0xF406, 0x11F0, 0x2000, 0x4000, 0xFBFC, 0x4040, 0x80, 0x1910, 0xE208, 0x47FC, 0x204}},
    {32473, 2, {0x1040, 0x1040, 0x20A0, 0x2110, 0x4A08, 0xF406, 0x13F8, 0x2000, 0x4000, 0xFBF8, 0x4208, 0x208, 0x1A08, 0xE208, 0x43F8, 0x208}},
    {32474, 2, {0x1080, 0x1080, 0x2100, 0x21FC, 0x4A04, 0xF404, 0x13E4, 0x2224, 0x4224, 0xFBE4, 0x4224, 0x224, 0x1BE4, 0xE004, 0x4028, 0x10}},
    {32475, 2, {0x1080, 0x1080, 0x20FC, 0x2108, 0x4A90, 0xF860, 0x1198, 0x2626, 0x4020, 0xF9FC, 0x4020, 0x220, 0x1BFE, 0xE020, 0x4020, 0x20}},
    {32476, 2, {0x1080, 0x1080, 0x20F8, 0x2108, 0x4B10, 0xFCA0, 0x1040, 0x20A0, 0x4118, 0xFA06, 0x45F8, 0x108, 0x1908, 0xE108, 0x41F8, 0x108}},
    {32477, 2, {0x1080, 0x1080, 0x20F8, 0x2108, 0x4A10, 0xF5FC, 0x1124, 0x2124, 0x4124, 0xF9FC, 0x4100, 0x100, 0x1902, 0xE102, 0x40FE, 0x0}},
    {32478, 2, {0x1040, 0x1020, 0x2020, 0x25FE, 0x4400, 0xF888, 0x1104, 0x2202, 0x4088, 0xFC88, 0x4050, 0x50, 0x1C20, 0xE050, 0x4088, 0x306}},
    {32479, 2, {0x1040, 0x1020, 0x2020, 0x23FE, 0x4840, 0xF888, 0x1104, 0x23FE, 0x4092, 0xF890, 0x4090, 0x90, 0x1912, 0xE112, 0x420E, 0x400}},
    {32480, 2, {0x1000, 0x13FE, 0x2020, 0x2420, 0x45FC, 0xF924, 0x1124, 0x21FC, 0x4124, 0xFD24, 0x41FC, 0x120, 0x1CA0, 0xE040, 0x40B0, 0x30E}},
    {32481, 2, {0x1020, 0x1124, 0x20A4, 0x24A8, 0x4420, 0xF9FC, 0x1104, 0x2104, 0x41FC, 0xFD04, 0x4104, 0x1FC, 0x1D04, 0xE104, 0x4114, 0x108}},
    {32482, 2, {0x1000, 0x11F8, 0x2108, 0x2108, 0x49F8, 0xF800, 0x13FC, 0x2204, 0x4204, 0xFBFC, 0x4204, 0x204, 0x1BFC, 0xE204, 0x4214, 0x208}},
    {32483, 2, {0x1038, 0x13C0, 0x2040, 0x2040, 0x4FFC, 0xF950, 0x1248, 0x2446, 0x43F0, 0xF910, 0x4120, 0x17C, 0x1A04, 0xE204, 0x4428, 0x810}},
    {32484, 2, {0x1000, 0x1050, 0x2048, 0x2484, 0x4524, 0xF820, 0x1050, 0x2088, 0x4106, 0xFCF8, 0x4088, 0x88, 0x1C88, 0xE088, 0x40F8, 0x88}},
    {32485, 2, {0x1008, 0x103C, 0x23C0, 0x2044, 0x4A24, 0xF928, 0x1100, 0x2040, 0x43FE, 0xF888, 0x4108, 0x390, 0x1860, 0xE050, 0x4188, 0x604}},
    {32486, 2, {0x1080, 0x1080, 0x21F8, 0x2308, 0x4C90, 0xF060, 0x1198, 0x2646, 0x4040, 0xFBFC, 0x4040, 0x248, 0x1A44, 0xE444, 0x4140, 0x80}},
    {32487, 2, {0x1020, 0x1220, 0x2220, 0x2324, 0x4AA8, 0xF220, 0x13FC, 0x2220, 0x4270, 0xFAA8, 0x4324, 0x220, 0x1A20, 0xE220, 0x43FE, 0x0}},
    {32488, 2, {0x1108, 0x1088, 0x2090, 0x23FC, 0x4824, 0xF824, 0x13FC, 0x2220, 0x4220, 0xFBFE, 0x4062, 0xA2, 0x192A, 0xE224, 0x4420, 0x20}},
    {32489, 2, {0x1040, 0x1040, 0x27FC, 0x2040, 0x4BF8, 0xF040, 0x17FE, 0x2000, 0x43F8, 0xFA08, 0x4248, 0x248, 0x1A48, 0xE0A0, 0x4110, 0x608}},
    {32490, 2, {0x1040, 0x1044, 0x23F4, 0x2048, 0x4850, 0xF7FE, 0x1040, 0x2080, 0x41F8, 0xFB08, 0x4508, 0x1F8, 0x1908, 0xE108, 0x41F8, 0x108}},
    {32491, 2, {0x1020, 0x1020, 0x21FC, 0x2020, 0x4820, 0xFBFE, 0x1088, 0x2144, 0x4242, 0xF8F8, 0x4188, 0x250, 0x1820, 0xE050, 0x4188, 0x606}},
    {32492, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x44A8, 0xF820, 0x11FC, 0x2124, 0x4124, 0xFD24, 0x43FE, 0x20, 0x1C50, 0xE088, 0x4104, 0x202}},
    {32493, 2, {0x1020, 0x1020, 0x21FC, 0x2420, 0x4420, 0xFBFE, 0x1002, 0x2094, 0x4050, 0xFD10, 0x4090, 0x3FE, 0x1C28, 0xE044, 0x4082, 0x302}},
    {32494, 2, {0x1020, 0x1020, 0x21FC, 0x2450, 0x4488, 0xF904, 0x13FE, 0x2008, 0x41E8, 0xFD28, 0x4128, 0x1E8, 0x1D28, 0xE008, 0x4028, 0x10}},
    {32495, 2, {0x1090, 0x1090, 0x2090, 0x279E, 0x4890, 0xF890, 0x1090, 0x239C, 0x4090, 0xF890, 0x4090, 0x79E, 0x1890, 0xE090, 0x4090, 0x90}},
    {32496, 2, {0x1040, 0x1040, 0x207E, 0x2040, 0x4BFC, 0xFA04, 0x13FC, 0x2204, 0x43FC, 0xFA44, 0x4040, 0x7FE, 0x1840, 0xE040, 0x4040, 0x40}},
    {32497, 2, {0x1020, 0x1124, 0x20A4, 0x20A8, 0x4820, 0xFBFE, 0x1202, 0x2202, 0x42FA, 0xFA8A, 0x428A, 0x28A, 0x1AFA, 0xE202, 0x420A, 0x204}},
    {32498, 2, {0x1000, 0x11FC, 0x2104, 0x2504, 0x45FC, 0xF904, 0x1104, 0x21FC, 0x4000, 0xFD12, 0x41D4, 0x118, 0x1D10, 0xE152, 0x4192, 0x10E}},
    {32499, 2, {0x1000, 0x11F0, 0x2110, 0x2110, 0x49F0, 0xF840, 0x13F8, 0x2248, 0x4248, 0xFBF8, 0x4248, 0x248, 0x1BFA, 0xE042, 0x4042, 0x3E}},
    {32500, 2, {0x10A0, 0x1090, 0x2080, 0x25FE, 0x4510, 0xFB10, 0x15FC, 0x2110, 0x4110, 0xFDFC, 0x4110, 0x110, 0x1D10, 0xE1FE, 0x4100, 0x100}},
    {32501, 2, {0x1020, 0x1040, 0x21FC, 0x2104, 0x49FC, 0xF104, 0x11FC, 0x2020, 0x4020, 0xFBFE, 0x4222, 0x222, 0x1A2A, 0xE224, 0x4020, 0x20}},
    {32502, 2, {0x101C, 0x13E0, 0x2084, 0x2244, 0x4948, 0xF910, 0x17FE, 0x2402, 0x4000, 0xFBF8, 0x4108, 0x110, 0x18A0, 0xE040, 0x41B0, 0x60E}},
    {32503, 2, {0x2000, 0x27BC, 0x24A4, 0x44A4, 0x54A4, 0xF7BC, 0x24A4, 0x24A4, 0x44A4, 0xF7BC, 0x44A4, 0x4A4, 0x34A4, 0xC4A4, 0xAA4, 0x114C}},
    {32504, 2, {0x1000, 0x13FE, 0x2222, 0x2222, 0x4AFA, 0xFA22, 0x1222, 0x23FE, 0x4202, 0xFAFA, 0x428A, 0x28A, 0x1AFA, 0xE202, 0x420A, 0x404}},
    {32505, 2, {0x1080, 0x1080, 0x20FE, 0x2502, 0x4682, 0xF8F2, 0x1142, 0x2042, 0x43FA, 0xFC42, 0x4152, 0x152, 0x1DF2, 0xE002, 0x4014, 0x8}},
    {32506, 2, {0x1088, 0x1088, 0x21EC, 0x212A, 0x4A28, 0xF548, 0x1088, 0x217E, 0x4600, 0xF800, 0x41FC, 0x104, 0x1904, 0xE104, 0x41FC, 0x104}},
    {32507, 2, {0x1040, 0x1248, 0x2150, 0x2040, 0x4BF8, 0xF880, 0x17FC, 0x2110, 0x4208, 0xFDF4, 0x4912, 0x110, 0x1950, 0xE124, 0x4104, 0xFC}},
    {32508, 2, {0x1040, 0x1020, 0x23FE, 0x2202, 0x4800, 0xF9FC, 0x1000, 0x2000, 0x43FE, 0xF820, 0x4128, 0x124, 0x1A22, 0xE422, 0x40A0, 0x40}},
    {32509, 2, {0x1040, 0x1020, 0x23FE, 0x2202, 0x4C04, 0xF800, 0x13FE, 0x2020, 0x4020, 0xF920, 0x413C, 0x120, 0x1AA0, 0xE260, 0x443E, 0x800}},
    {32510, 2, {0x1040, 0x1020, 0x23FE, 0x2202, 0x4C04, 0xF9F8, 0x1108, 0x2108, 0x41F8, 0xF900, 0x4100, 0x1FC, 0x1904, 0xE104, 0x41FC, 0x104}},
    {32511, 2, {0x1000, 0x11F8, 0x2008, 0x2008, 0x49F8, 0xF808, 0x1008, 0x23FE, 0x4020, 0xFA22, 0x4174, 0xA8, 0x1924, 0xE222, 0x40A0, 0x40}},
    {32512, 2, {0x2000, 0x27BC, 0x2084, 0x4294, 0x5108, 0xF294, 0x24A4, 0x2840, 0x4000, 0xF7BC, 0x40A4, 0x2A4, 0x3128, 0xC290, 0x4A8, 0x846}},
    {32513, 2, {0x1124, 0x1124, 0x2248, 0x2490, 0x4A48, 0xF924, 0x1124, 0x2000, 0x43FC, 0xFA44, 0x4244, 0x3FC, 0x1A44, 0xE244, 0x43FC, 0x204}},
    {32514, 2, {0x1088, 0x1088, 0x23FE, 0x2488, 0x44F8, 0xF820, 0x11FC, 0x2124, 0x4124, 0xFDFC, 0x4020, 0x3FE, 0x1C20, 0xE020, 0x4020, 0x20}},
    {32515, 2, {0x1100, 0x1100, 0x213E, 0x2122, 0x4FA2, 0xF122, 0x113E, 0x23A2, 0x4362, 0xFD22, 0x493E, 0x122, 0x1922, 0xE122, 0x413E, 0x122}},
    {32516, 2, {0x2014, 0x2012, 0x2010, 0x47FE, 0x5410, 0xF410, 0x25D0, 0x2412, 0x4412, 0xF5D4, 0x4554, 0x548, 0x35DA, 0xC42A, 0x846, 0x1082}},
    {32517, 2, {0x1000, 0x17FE, 0x2040, 0x2080, 0x4BFC, 0xF294, 0x1294, 0x22F4, 0x4294, 0xFA94, 0x42F4, 0x294, 0x1A94, 0xE294, 0x43FC, 0x204}},
    {32518, 2, {0x1090, 0x1290, 0x229E, 0x22A8, 0x4AC4, 0xF880, 0x11FC, 0x2104, 0x4124, 0xF924, 0x4124, 0x154, 0x1850, 0xE090, 0x4112, 0x60E}},
    {32519, 2, {0x1000, 0x11FC, 0x2104, 0x2104, 0x49FC, 0xF904, 0x1104, 0x21FC, 0x4000, 0xFBFE, 0x4020, 0x120, 0x193C, 0xE120, 0x42A0, 0x47E}},
    {32520, 2, {0x2010, 0x2010, 0x2710, 0x4510, 0x5514, 0xF752, 0x2552, 0x2592, 0x4710, 0xF514, 0x4508, 0x710, 0x3520, 0xC040, 0x80, 0x300}},
    {32521, 2, {0x1000, 0x11F8, 0x2108, 0x2108, 0x49F8, 0xF000, 0x17FE, 0x2108, 0x41F8, 0xF908, 0x41F8, 0x108, 0x193E, 0xE7C8, 0x4008, 0x8}},
    {32522, 2, {0x1000, 0x11F8, 0x2108, 0x2108, 0x49F8, 0xF908, 0x1108, 0x21F8, 0x4000, 0xFBFC, 0x4294, 0x294, 0x1A94, 0xE294, 0x47FE, 0x0}},
    {32523, 2, {0x1020, 0x1020, 0x21FC, 0x2524, 0x45FC, 0xF820, 0x13FE, 0x2000, 0x41FC, 0xFD04, 0x4124, 0x124, 0x1D24, 0xE050, 0x4088, 0x304}},
    {32524, 2, {0x1000, 0x11FC, 0x2124, 0x2124, 0x49FC, 0xF924, 0x1124, 0x21FC, 0x4000, 0xF840, 0x4024, 0x2A2, 0x1A8A, 0xE488, 0x4078, 0x0}},
    {32525, 2, {0x1008, 0x103C, 0x21E0, 0x2020, 0x4820, 0xFBFE, 0x1124, 0x2124, 0x47FE, 0xF924, 0x4124, 0x3FE, 0x1820, 0xE020, 0x43FE, 0x0}},
    {32526, 2, {0x1080, 0x1338, 0x2228, 0x2228, 0x4BA8, 0xFA46, 0x1200, 0x227C, 0x43A4, 0xFA24, 0x4228, 0x3A8, 0x1E10, 0xE228, 0x4244, 0x282}},
    {32527, 2, {0x1100, 0x11FE, 0x2110, 0x2210, 0x4AFE, 0xF292, 0x1692, 0x22FE, 0x4292, 0xFA92, 0x42FE, 0x290, 0x1A50, 0xE220, 0x4258, 0x286}},
    {32528, 2, {0x1040, 0x1080, 0x21FC, 0x2104, 0x49FC, 0xF904, 0x11FC, 0x2020, 0x4032, 0xFBB4, 0x40A8, 0x128, 0x1924, 0xE224, 0x44A2, 0x40}},
    {32529, 2, {0x1100, 0x1178, 0x2108, 0x2208, 0x4AFE, 0xF640, 0x1A40, 0x227C, 0x4290, 0xFA10, 0x42FE, 0x210, 0x1A28, 0xE228, 0x4244, 0x282}},
    {32530, 2, {0x1010, 0x1220, 0x217C, 0x2544, 0x4444, 0xF87C, 0x1340, 0x2140, 0x417C, 0xFD44, 0x4144, 0x17C, 0x1D44, 0xE280, 0x447E, 0x0}},
    {32531, 2, {0x103C, 0x17C0, 0x2244, 0x2128, 0x4800, 0xFBFC, 0x1080, 0x2080, 0x47FE, 0xF900, 0x41F8, 0x288, 0x1A50, 0xE420, 0x48D8, 0x306}},
    {32532, 2, {0x1040, 0x1020, 0x21FC, 0x2400, 0x4488, 0xF850, 0x13FE, 0x2222, 0x4424, 0xFDFC, 0x4124, 0x124, 0x1D34, 0xE128, 0x4020, 0x20}},
    {32533, 2, {0x1020, 0x1124, 0x20A8, 0x2020, 0x4BFE, 0xF8A8, 0x1124, 0x2202, 0x4040, 0xFBFE, 0x4088, 0x108, 0x1990, 0xE060, 0x4198, 0x604}},
    {32534, 2, {0x1080, 0x1040, 0x23FC, 0x2204, 0x4A04, 0xFBFC, 0x1200, 0x2200, 0x43FC, 0xFB54, 0x4354, 0x5FC, 0x1D54, 0xE554, 0x4944, 0x10C}},
    {32535, 2, {0x1000, 0x13FC, 0x2204, 0x23FC, 0x4A20, 0xFBFE, 0x1210, 0x228A, 0x4306, 0xF800, 0x43FC, 0x204, 0x1BFC, 0xE204, 0x43FC, 0x204}},
    {32536, 2, {0x1080, 0x10FC, 0x2104, 0x21F8, 0x4808, 0xFBFE, 0x1040, 0x20A2, 0x4334, 0xF858, 0x4094, 0x334, 0x1852, 0xE090, 0x4350, 0x20}},
    {32537, 2, {0x1000, 0x13FC, 0x2090, 0x2294, 0x4998, 0xF890, 0x17FE, 0x2000, 0x41F8, 0xF908, 0x4108, 0x1F8, 0x1908, 0xE108, 0x41F8, 0x108}},
    {32538, 2, {0x1050, 0x1048, 0x27FE, 0x2040, 0x4BFC, 0xFA44, 0x13FC, 0x2244, 0x43FC, 0xFA44, 0x4008, 0x3FE, 0x1908, 0xE088, 0x40A8, 0x10}},
    {32539, 2, {0x1000, 0x13FE, 0x2200, 0x22FC, 0x4A00, 0xFBFE, 0x12A8, 0x2290, 0x42C8, 0xFA86, 0x4208, 0x3FE, 0x1A88, 0xE448, 0x4408, 0x818}},
    {32540, 2, {0x1040, 0x1040, 0x27FC, 0x2040, 0x4BF8, 0xFA08, 0x13F8, 0x2208, 0x43F8, 0xFA08, 0x43F8, 0x208, 0x1FFE, 0xE110, 0x4208, 0x404}},
    {32541, 2, {0x1020, 0x147C, 0x2284, 0x2148, 0x4830, 0xF8CE, 0x1610, 0x227C, 0x4210, 0xFA7C, 0x4210, 0x2FE, 0x1A10, 0xE210, 0x45FE, 0x800}},
    {32542, 2, {0x1040, 0x1020, 0x23FE, 0x2000, 0x49FC, 0xF904, 0x17FE, 0x2104, 0x41FC, 0xF852, 0x4094, 0x188, 0x1A88, 0xE4A4, 0x40C2, 0x80}},
    {32543, 2, {0x1040, 0x1020, 0x23FE, 0x2000, 0x49FC, 0xF904, 0x11FC, 0x2000, 0x43FE, 0xFA02, 0x42FA, 0x28A, 0x1AFA, 0xE202, 0x420A, 0x204}},
    {32544, 2, {0x2040, 0x2020, 0x27FE, 0x4400, 0x55FC, 0xF524, 0x2524, 0x25FC, 0x4524, 0xF524, 0x45FC, 0x420, 0x35FC, 0xC820, 0x820, 0x13FE}},
    {32545, 2, {0x1040, 0x1020, 0x23FE, 0x2000, 0x4954, 0xF924, 0x1154, 0x21FC, 0x4020, 0xFBFE, 0x4242, 0x292, 0x1AFA, 0xE20A, 0x4202, 0x206}},
    {32546, 2, {0x1104, 0x1084, 0x2088, 0x2000, 0x4BFE, 0xF800, 0x1088, 0x2104, 0x4202, 0xF9FC, 0x4154, 0x154, 0x1954, 0xE154, 0x47FE, 0x0}},
    {32547, 2, {0x1088, 0x1050, 0x2000, 0x25FE, 0x4450, 0xF9FC, 0x1054, 0x23FE, 0x4054, 0xFDFC, 0x4050, 0xD8, 0x1D54, 0xE252, 0x4050, 0x50}},
    {32548, 2, {0x1040, 0x1020, 0x23FE, 0x2202, 0x4C14, 0xF9E0, 0x1100, 0x2100, 0x41FC, 0xF910, 0x4110, 0x7FE, 0x1800, 0xE090, 0x4108, 0x204}},
    {32549, 2, {0x1000, 0x13FE, 0x2050, 0x25FC, 0x4554, 0xF954, 0x11FC, 0x2000, 0x41FC, 0xFC00, 0x43FE, 0x20, 0x1CA8, 0xE124, 0x42A2, 0x40}},
    {32550, 2, {0x1000, 0x11F8, 0x2108, 0x21F8, 0x4908, 0xF9F8, 0x1000, 0x23FC, 0x4294, 0xFBFC, 0x4000, 0x1F8, 0x1890, 0xE060, 0x4198, 0x606}},
    {32551, 2, {0x1000, 0x11FC, 0x2124, 0x25FC, 0x4524, 0xF9FC, 0x1040, 0x2088, 0x41F0, 0xFC20, 0x4044, 0x3FE, 0x1C22, 0xE124, 0x4222, 0x60}},
    {32552, 2, {0x2000, 0x2FBE, 0x28A2, 0x4AAA, 0x5AAA, 0xFAAA, 0x2514, 0x28A2, 0x4080, 0xF7FE, 0x4110, 0x210, 0x33A0, 0xC060, 0x198, 0xE04}},
    {32553, 2, {0x2040, 0x2020, 0x27FE, 0x4402, 0x5100, 0xF1FE, 0x2210, 0x2620, 0x4A7C, 0xF244, 0x4244, 0x27C, 0x3244, 0xC244, 0x27C, 0x244}},
    {32554, 2, {0x1000, 0x17BC, 0x2084, 0x24A4, 0x4A94, 0xF4A4, 0x1050, 0x2188, 0x4626, 0xF8C0, 0x4310, 0x64, 0x1B88, 0xE030, 0x40C0, 0x700}},
    {32555, 2, {0x1124, 0x1248, 0x2124, 0x2400, 0x45FC, 0xF924, 0x11FC, 0x2124, 0x41FC, 0xFC20, 0x43FE, 0x70, 0x1CA8, 0xE124, 0x4622, 0x20}},
    {32556, 2, {0x2200, 0x22FE, 0x2210, 0x4FA0, 0x527C, 0xF244, 0x2754, 0x2054, 0x4054, 0xF754, 0x4554, 0x554, 0x3528, 0xC724, 0x542, 0x82}},
    {32557, 2, {0x1020, 0x1020, 0x23FE, 0x2050, 0x4A8A, 0xF904, 0x13FE, 0x2504, 0x41FC, 0xF904, 0x41FC, 0x20, 0x1924, 0xE222, 0x44A2, 0x40}},
    {32558, 2, {0x1088, 0x1050, 0x23FE, 0x2020, 0x49FC, 0xF820, 0x13FE, 0x2124, 0x40A8, 0xFBFE, 0x4000, 0x1FC, 0x1904, 0xE104, 0x41FC, 0x104}},
    {32559, 2, {0x1104, 0x1088, 0x2000, 0x23FE, 0x4A22, 0xFAAA, 0x1272, 0x2222, 0x43FE, 0xF800, 0x41FC, 0x104, 0x19FC, 0xE104, 0x41FC, 0x104}},
    {32560, 2, {0x1000, 0x13FE, 0x2000, 0x25FC, 0x4524, 0xF9FC, 0x1124, 0x23FE, 0x4000, 0xFDFC, 0x4124, 0x1FC, 0x1D24, 0xE1FC, 0x4000, 0x3FE}},
    {32561, 2, {0x1010, 0x14FE, 0x2292, 0x22FE, 0x4810, 0xF9FE, 0x1000, 0x2EFE, 0x4282, 0xFAFE, 0x4280, 0x2FE, 0x1A82, 0xE2FE, 0x4500, 0x8FE}},
    {32562, 2, {0x11FC, 0x1104, 0x2104, 0x21FC, 0x4800, 0xFBDE, 0x1252, 0x2252, 0x43DE, 0xF820, 0x47FE, 0x70, 0x18A8, 0xE124, 0x4622, 0x20}},
    {32563, 2, {0x1000, 0x13FC, 0x2294, 0x2294, 0x4BFC, 0xF800, 0x17FE, 0x2000, 0x43FC, 0xFA04, 0x43FC, 0xA2, 0x1914, 0xE308, 0x4D44, 0x182}},
    {32564, 2, {0x2108, 0x2208, 0x27C8, 0x4450, 0x57DE, 0xF464, 0x27D4, 0x2214, 0x4114, 0xF7D4, 0x4214, 0x3C8, 0x3248, 0xC454, 0x554, 0x8A2}},
    {32565, 2, {0x2108, 0x2528, 0x27BE, 0x4948, 0x57BE, 0xF318, 0x25AA, 0x2946, 0x4000, 0xF3F8, 0x4208, 0x248, 0x3248, 0xC0B0, 0x108, 0x604}},
    {32566, 2, {0x800, 0x800, 0x1000, 0x1FF8, 0x2100, 0x4100, 0x100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {32567, 2, {0x2000, 0x2000, 0x3F3E, 0x4822, 0x8822, 0x822, 0xFFA2, 0x822, 0x822, 0x4922, 0x492A, 0x4924, 0x4F20, 0x7920, 0x20, 0x20}},
    {32568, 2, {0x2000, 0x2000, 0x3F7E, 0x4810, 0x8810, 0x810, 0xFF10, 0x810, 0x810, 0x4910, 0x4910, 0x4910, 0x4F10, 0x79FE, 0x100, 0x0}},
    {32569, 2, {0x800, 0x800, 0x1FF8, 0x2100, 0x4100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x4888, 0x4444, 0x8444}},
    {32570, 2, {0x2020, 0x2020, 0x3C20, 0x51FC, 0x9024, 0x1024, 0xFE24, 0x1024, 0x13FE, 0x5420, 0x5450, 0x5450, 0x5C88, 0x6488, 0x504, 0x202}},
    {32571, 2, {0x2000, 0x2000, 0x3DFE, 0x5040, 0x9040, 0x1078, 0xFE48, 0x1048, 0x1068, 0x5498, 0x5498, 0x5488, 0x5C8A, 0x64AA, 0x4CA, 0x86}},
    {32572, 2, {0x2040, 0x2040, 0x3C40, 0x507C, 0x9044, 0x1088, 0xFEA0, 0x1120, 0x1020, 0x5450, 0x5450, 0x5448, 0x5C88, 0x6484, 0x504, 0x202}},
    {32573, 2, {0x2020, 0x2020, 0x3C20, 0x5020, 0x91FE, 0x1070, 0xFEA8, 0x10A8, 0x1124, 0x5524, 0x5622, 0x54F8, 0x5C20, 0x6420, 0x420, 0x20}},
    {32574, 2, {0x2084, 0x2048, 0x3C00, 0x50FC, 0x9048, 0x1048, 0xFE48, 0x1048, 0x11FE, 0x5448, 0x5448, 0x5448, 0x5C48, 0x6488, 0x488, 0x108}},
    {32575, 2, {0x2008, 0x201C, 0x3CE0, 0x5080, 0x9080, 0x10FE, 0xFE80, 0x1080, 0x10BC, 0x54A4, 0x54A4, 0x54A4, 0x5CA4, 0x653C, 0x524, 0x200}},
    {32576, 2, {0x2050, 0x2050, 0x3C50, 0x5152, 0x90D4, 0x1058, 0xFE50, 0x1058, 0x10D4, 0x5552, 0x5650, 0x5450, 0x5C92, 0x6492, 0x512, 0x20E}},
    {32577, 2, {0x4000, 0x43FE, 0x7A02, 0xA28A, 0x2252, 0x23FE, 0xFA22, 0x2222, 0x22AA, 0xAAAA, 0xAAAA, 0xAAFA, 0xBA02, 0xCA02, 0x20A, 0x204}},
    {32578, 2, {0x3EF8, 0x2288, 0x2AA8, 0x2AA8, 0x2AA8, 0x1450, 0x2288, 0x4904, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x2108, 0x3FF8, 0x8}},
    {32579, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8804, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x3FF8}},
    {32580, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x9086, 0x1FF0, 0x2100, 0xFFFE, 0x100, 0x2108, 0x3FF8}},
    {32581, 2, {0x4040, 0x407C, 0x7840, 0xA3FE, 0x2242, 0x2278, 0xFBC4, 0x223C, 0x2208, 0xAAF0, 0xAB24, 0xAAA8, 0xBDFE, 0xCC20, 0x8A0, 0x40}},
    {32582, 2, {0x2000, 0x21FC, 0x3D24, 0x53FE, 0x9124, 0x11FC, 0xFE00, 0x11FC, 0x1104, 0x55FC, 0x5504, 0x55FC, 0x5D04, 0x65FC, 0x488, 0x104}},
    {32583, 2, {0x4108, 0x4090, 0x7BFE, 0xA090, 0x23FC, 0x2294, 0xFB1C, 0x2204, 0x23FC, 0xAA04, 0xABFC, 0xA808, 0xBBFE, 0xC908, 0xA8, 0x10}},
    {32584, 2, {0x4000, 0x43FE, 0x7850, 0xA3FE, 0x2252, 0x23FE, 0xF800, 0x21FC, 0x2104, 0xA9FC, 0xA904, 0xA9FC, 0xB820, 0xCBFE, 0x20, 0x20}},
    {32585, 2, {0x4124, 0x40A8, 0x7BFE, 0xA202, 0x20F8, 0x2088, 0xF8F8, 0x200C, 0x21F0, 0xA820, 0xA9FC, 0xA820, 0xBBFE, 0xC820, 0xA0, 0x40}},
    {32586, 2, {0x878, 0x7F48, 0x848, 0x3E86, 0x2B78, 0x3E48, 0x2A28, 0xFF90, 0x4928, 0x7F46, 0x2000, 0x3FF8, 0x4100, 0xFFFE, 0x2108, 0x3FF8}},
    {32587, 2, {0x100, 0xFFFE, 0x1048, 0x24FE, 0x7990, 0x10FC, 0x2290, 0x7CFC, 0x890, 0x30FE, 0xC880, 0x3FF8, 0x4100, 0xFFFE, 0x2108, 0x3FF8}},
    {32588, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2288, 0x1000, 0x1FF0, 0x2100, 0xFFFE, 0x100, 0x2108, 0x3FF8}},
    {32589, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x1000, 0x1FF0, 0x2100, 0xFFFE, 0x100, 0x2108, 0x3FF8}},
    {32590, 2, {0x41FC, 0x4104, 0x79FC, 0xA104, 0x23FE, 0x2020, 0xFBFE, 0x2222, 0x21AC, 0xA820, 0xA9FC, 0xA800, 0xBBFE, 0xC840, 0x88, 0x1FC}},
    {32591, 2, {0x4040, 0x407C, 0x7840, 0xA3FC, 0x2244, 0x23F0, 0xFA44, 0x22FC, 0x22A8, 0xAAF8, 0xAAA8, 0xAAF8, 0xBA00, 0xCDFC, 0x554, 0xBFE}},
    {32592, 2, {0x4088, 0x43FE, 0x7888, 0xA3DE, 0x2252, 0x23DE, 0xF8A0, 0x2090, 0x21FE, 0xA920, 0xABFC, 0xAD20, 0xB9FC, 0xC920, 0x1FE, 0x100}},
    {32593, 2, {0x0, 0x7FFC, 0x4004, 0x4004, 0x4214, 0x5294, 0x4A54, 0x4424, 0x4424, 0x4A54, 0x4A54, 0x5294, 0x6104, 0x4004, 0x4014, 0x4008}},
    {32594, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {32595, 2, {0x0, 0x3FF8, 0x2008, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2008, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {32596, 2, {0x0, 0x7FFC, 0x4004, 0x4824, 0x4444, 0x5FF4, 0x4104, 0x4084, 0x5FF4, 0x4404, 0x4404, 0x4404, 0x47E4, 0x4004, 0x4014, 0x4008}},
    {32597, 2, {0x0, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2008, 0x1FF0, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {32598, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC206, 0x100, 0x80, 0x80, 0x0}},
    {32599, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x200, 0x400, 0xFF0, 0x1010, 0x6820, 0x440, 0x280, 0x300, 0x1C00, 0xE000}},
    {32600, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x100, 0x300, 0x570, 0x90C, 0x3102, 0xC100, 0x100, 0x100}},
    {32601, 2, {0x0, 0x7FFE, 0x4002, 0x8824, 0x1010, 0x2108, 0x100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {32602, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x2008, 0x1088, 0x88, 0xF088, 0x1088, 0x1088, 0x1488, 0x1808, 0x1028, 0x10}},
    {32603, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x78, 0x1FA0, 0x1220, 0x1220, 0x1210, 0x1290, 0x1248, 0x22A8, 0x4324, 0x8202}},
    {32604, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x200, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {32605, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {32606, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF0, 0x20, 0x640, 0x180, 0xFFFE, 0x282, 0x484, 0x880, 0x3080, 0xC280, 0x100}},
    {32607, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {32608, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x3FFC, 0x2040, 0x2024, 0x2C14, 0x300C}},
    {32609, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x0, 0x7FFC, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {32610, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {32611, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {32612, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x820, 0x440, 0x7FF8, 0x108, 0x3FF8, 0x2100, 0x3FFC, 0x304, 0x514, 0x1908, 0x6100}},
    {32613, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x2028, 0x2010}},
    {32614, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0xF8, 0x7F00, 0x2210, 0x1120, 0x3FE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {32615, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x820, 0x820, 0x7EFC, 0x820, 0x1C30, 0x1A70, 0x2A68, 0x48A4, 0x8922, 0x820, 0x820}},
    {32616, 2, {0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x200, 0x7FFC, 0x440, 0x920, 0x3FF8, 0xD116, 0x1FF0, 0x1110, 0x1FF4, 0x104, 0x104, 0xFC}},
    {32617, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x100, 0x1FC, 0x100, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {32618, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x440, 0x7C7C, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0xFC7E, 0x440, 0x440}},
    {32619, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x820, 0x7F20, 0x820, 0xFF30, 0x828, 0x824, 0x7F24, 0x820, 0xF20, 0xF020}},
    {32620, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x14D0, 0x820, 0x1450, 0x6288, 0x3EF8, 0x288, 0x1450, 0x820, 0x3458, 0xC286}},
    {32621, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x50, 0x48, 0xFFFE, 0x40, 0x3E40, 0x2224, 0x3E28, 0x12, 0xE2A, 0x70C6, 0x302}},
    {32622, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0xFFFE, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {32623, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {32624, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x800, 0x404, 0x7FA4, 0x24, 0x3F24, 0x24, 0x3F24, 0x24, 0x3F24, 0x2104, 0x3F14, 0x2108}},
    {32625, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2448, 0x2288, 0x2FE8, 0x2108, 0x3FF8, 0x2108, 0x2128, 0x2010}},
    {32626, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x200, 0x1FD0, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x810}},
    {32627, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x4884, 0x4812, 0x87F2}},
    {32628, 2, {0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x400, 0x820, 0x1FF0, 0x10, 0x4888, 0x4444, 0x8444}},
    {32629, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x1FF8, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x4, 0x2924, 0x4494, 0x8}},
    {32630, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x400, 0x79FC, 0x4044, 0x4844, 0x5454, 0x6288, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {32631, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x1080, 0x2498, 0x42E0, 0x7E84, 0x84, 0x7E7C, 0x4280, 0x7E98, 0x42E0, 0x7E84, 0x4284, 0x467C}},
    {32632, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x1008, 0x808, 0xFF08, 0xFE, 0x7E08, 0x88, 0x7E48, 0x48, 0x7E08, 0x4208, 0x7E28, 0x4210}},
    {32633, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x10A0, 0x1090, 0x19FE, 0x5520, 0x53FC, 0x9520, 0x11FC, 0x1120, 0x1120, 0x11FE, 0x1100}},
    {32634, 2, {0x7FFC, 0x4444, 0x7FFC, 0x1110, 0x2220, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {32635, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x3F08, 0x2108, 0x3F08, 0x20FE, 0x2E08, 0x2048, 0x3F28, 0x4428, 0x5508, 0xA4A8, 0xC10}},
    {32636, 2, {0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x920, 0xFFFE, 0x920, 0x3FF8, 0x100, 0x7FFC, 0x100}},
    {32637, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x7FFE, 0x4200, 0x4A44, 0x52A4, 0x4524, 0x48A4, 0x5224, 0x4AA4, 0x5304, 0x4494, 0x9848}},
    {32638, 2, {0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x820, 0x440, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {32639, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x100, 0x3FF8, 0x440, 0xFFFE, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {32640, 2, {0x3FF8, 0x2AA8, 0x2448, 0x2AA8, 0x0, 0xFFFE, 0x280, 0x3EF8, 0x2288, 0x3EF8, 0x280, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {32641, 2, {0x0, 0x7FFC, 0x4444, 0x7FFC, 0x2200, 0x7F78, 0x2208, 0x3E48, 0x848, 0x7F7C, 0x4904, 0x7F04, 0x8F4, 0xFF04, 0x814, 0x808}},
    {32642, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x1000, 0x11FC, 0x2504, 0x7DFC, 0x800, 0x11FC, 0x2504, 0x7DFC, 0x104, 0x55FC, 0x5504, 0x810C}},
    {32643, 2, {0x7FFC, 0x4444, 0x7FFC, 0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x920, 0x3FF8, 0xD116, 0x1130, 0x100}},
    {32644, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x1020, 0x103C, 0x2420, 0x7DFC, 0x904, 0x11FC, 0x2504, 0x7DFC, 0x20, 0x55FE, 0x5420, 0x8020}},
    {32645, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x1050, 0x1488, 0x25FE, 0x7A90, 0x10FC, 0x2490, 0x7EFC, 0x90, 0x5490, 0x4AFE, 0x8A80}},
    {32646, 2, {0x7FFC, 0x4444, 0x7FFC, 0x1080, 0x2498, 0x7EE0, 0x84, 0x7E7C, 0x4200, 0x7E98, 0x42E0, 0x7E84, 0x4284, 0x467C, 0x2488, 0x4244}},
    {32647, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x2820, 0xFDFC, 0x2820, 0x3850, 0x1088, 0x7DFE, 0x5408, 0x7DE8, 0x1128, 0xFDE8, 0x1008, 0x1018}},
    {32648, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x2800, 0xFDFE, 0x2920, 0x39FC, 0x1120, 0x7DFC, 0x5520, 0x7DFE, 0x1002, 0xFEAA, 0x1004}},
    {32649, 2, {0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x2108, 0x2390, 0x4824, 0x73B8, 0x2010, 0x4BA4, 0x783E, 0x388, 0xAAAC, 0xABCA, 0x8298}},
    {32650, 2, {0x820, 0x420, 0x440, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {32651, 2, {0x920, 0x920, 0x920, 0xF93E, 0x920, 0x920, 0x820, 0x100, 0x7FFC, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {32652, 2, {0x820, 0x420, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x200, 0x200, 0xFFFE, 0x480, 0x480, 0x880, 0x1082, 0x2082, 0xC07E}},
    {32653, 2, {0x200, 0x200, 0x7FFC, 0x440, 0x820, 0x3018, 0xC826, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {32654, 2, {0x1010, 0x810, 0x420, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {32655, 2, {0x2204, 0x1204, 0x1408, 0xFF90, 0x820, 0x844, 0x7F04, 0x808, 0x810, 0xFFA2, 0x842, 0x804, 0x1008, 0x1010, 0x2060, 0xC180}},
    {32656, 2, {0x820, 0x820, 0xF83E, 0x820, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x400, 0xFE0, 0x1040, 0x6080, 0x340, 0x1C30, 0xE00E}},
    {32657, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x400, 0x7E0, 0x820, 0x1040, 0x6080, 0x140, 0x620, 0x1818, 0xE006}},
    {32658, 2, {0x4410, 0x2450, 0x2850, 0xFE50, 0x1088, 0x1088, 0x7D04, 0x12FA, 0x1048, 0xFE48, 0x1048, 0x1048, 0x2088, 0x2088, 0x4128, 0x8210}},
    {32659, 2, {0x4400, 0x2400, 0x29FC, 0xFD24, 0x1124, 0x1124, 0x7D24, 0x1124, 0x11FC, 0xFD00, 0x1100, 0x1100, 0x2102, 0x2102, 0x40FE, 0x8000}},
    {32660, 2, {0x820, 0x420, 0x440, 0x7FFC, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0, 0x4888, 0x4444, 0x8444}},
    {32661, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x80, 0x1F00, 0x104, 0x7D88, 0x950, 0x1120, 0x2518, 0xC206}},
    {32662, 2, {0x4400, 0x24F8, 0x2888, 0xFE88, 0x1088, 0x1106, 0x7E00, 0x11FC, 0x1084, 0xFE84, 0x1048, 0x1050, 0x2020, 0x2050, 0x4088, 0x8306}},
    {32663, 2, {0x820, 0x420, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x200, 0x200, 0xFFFE, 0x490, 0x4A4, 0x8BC, 0x1082, 0x2082, 0xC07E}},
    {32664, 2, {0x104, 0x3C84, 0x2488, 0x2400, 0x27FE, 0x3C20, 0x2420, 0x25FC, 0x2420, 0x3C20, 0x27FE, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32665, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x1110, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {32666, 2, {0x4420, 0x2420, 0x2850, 0xFC50, 0x1088, 0x1124, 0x7E12, 0x1010, 0x11FC, 0xFC04, 0x1008, 0x1088, 0x2050, 0x2020, 0x4010, 0x8010}},
    {32667, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x440, 0x7FF8, 0x448, 0x3FF8, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2048}},
    {32668, 2, {0x4440, 0x2420, 0x2820, 0xFDFE, 0x1102, 0x1204, 0x7C00, 0x1000, 0x11FE, 0xFC20, 0x1020, 0x1020, 0x2020, 0x2020, 0x40A0, 0x8040}},
    {32669, 2, {0x4404, 0x241E, 0x29F0, 0xFD10, 0x1110, 0x1110, 0x7D10, 0x11FE, 0x1110, 0xFD10, 0x1110, 0x1108, 0x210A, 0x214A, 0x41A6, 0x8112}},
    {32670, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x400, 0xFF0, 0x1110, 0x2110, 0xCFF0, 0x210, 0x210, 0x3FFE, 0x0}},
    {32671, 2, {0x4400, 0x25FC, 0x2808, 0xFC10, 0x1030, 0x1048, 0x7C84, 0x1302, 0x1000, 0xFDFC, 0x1020, 0x1020, 0x2020, 0x2020, 0x43FE, 0x8000}},
    {32672, 2, {0x4420, 0x2420, 0x2BFE, 0xFC20, 0x11FC, 0x1024, 0x7DFC, 0x1120, 0x11FE, 0xFC22, 0x102A, 0x1054, 0x2050, 0x2088, 0x4104, 0x8202}},
    {32673, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x4080, 0x2080, 0x9FC, 0x1204, 0x2448, 0xE040, 0x20A0, 0x2318, 0x2C06}},
    {32674, 2, {0x4414, 0x2412, 0x2810, 0xFDFE, 0x1010, 0x1090, 0x7C92, 0x1092, 0x13F2, 0xFC94, 0x1094, 0x1088, 0x208A, 0x211A, 0x4126, 0x8242}},
    {32675, 2, {0x3FF8, 0x208, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0xCFF8, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100}},
    {32676, 2, {0x82, 0x7E44, 0x1228, 0x12FE, 0xFF10, 0x1210, 0x1210, 0x7E7C, 0x2010, 0x2010, 0x7E10, 0x62FE, 0xA210, 0x2210, 0x3E10, 0x2210}},
    {32677, 2, {0x4400, 0x25FE, 0x2800, 0xFC92, 0x1124, 0x1248, 0x7D24, 0x1092, 0x1000, 0xFDFE, 0x1020, 0x1020, 0x2020, 0x2020, 0x43FE, 0x8000}},
    {32678, 2, {0x4420, 0x2410, 0x29FE, 0xFD02, 0x1204, 0x10F8, 0x7C00, 0x1000, 0x11FE, 0xFC50, 0x1050, 0x1050, 0x2092, 0x2092, 0x410E, 0x8200}},
    {32679, 2, {0x4440, 0x2440, 0x2888, 0xFD04, 0x13FE, 0x1002, 0x7C88, 0x1144, 0x1242, 0xFCF8, 0x1188, 0x1250, 0x2020, 0x2050, 0x4188, 0x8606}},
    {32680, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x80, 0x2080, 0x11FC, 0x8204, 0x5448, 0x2040, 0xE0A0, 0x2318, 0x2C06}},
    {32681, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x240, 0x3C50, 0x848, 0xFFFE, 0x848, 0xE50, 0x7824, 0x854, 0x198C}},
    {32682, 2, {0x4440, 0x2420, 0x29FC, 0xFD04, 0x1104, 0x11FC, 0x7D04, 0x1104, 0x11FC, 0xFD20, 0x1122, 0x1114, 0x2108, 0x2144, 0x4182, 0x8100}},
    {32683, 2, {0x4440, 0x2420, 0x29FE, 0xFD02, 0x1204, 0x1050, 0x7C88, 0x1104, 0x1000, 0xFDFC, 0x1020, 0x1020, 0x2020, 0x2020, 0x43FE, 0x8000}},
    {32684, 2, {0x8808, 0x480C, 0x500A, 0xFBFE, 0x2208, 0x2208, 0xFAE8, 0x220A, 0x220A, 0xFAEC, 0x22AC, 0x22A8, 0x22EA, 0x421A, 0x4226, 0x8442}},
    {32685, 2, {0x8840, 0x48A0, 0x5110, 0xFA08, 0x25F6, 0x2000, 0xFBC4, 0x2254, 0x2254, 0xFBD4, 0x2254, 0x2254, 0x23D4, 0x4244, 0x4254, 0x82C8}},
    {32686, 2, {0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x2488, 0x4244, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x280, 0xC60, 0x701C}},
    {32687, 2, {0x4400, 0x25FC, 0x2904, 0xFDFC, 0x1104, 0x11FC, 0x7C80, 0x11FE, 0x1222, 0xFD22, 0x1152, 0x1102, 0x21FA, 0x2002, 0x4014, 0x8008}},
    {32688, 2, {0x4420, 0x2524, 0x2924, 0xFD24, 0x11FC, 0x1080, 0x7C80, 0x13FE, 0x1090, 0xFD10, 0x1152, 0x1254, 0x22A8, 0x2428, 0x4844, 0x8082}},
    {32689, 2, {0x4400, 0x25FE, 0x2910, 0xFD20, 0x117C, 0x1144, 0x7D7C, 0x1144, 0x117C, 0xFD10, 0x1110, 0x1154, 0x2152, 0x2292, 0x4250, 0x8420}},
    {32690, 2, {0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x28, 0x3E24, 0x824, 0xFFFE, 0x2A20, 0x7D24, 0x90A8, 0x1E12, 0x22A, 0xC46}},
    {32691, 2, {0x443C, 0x27E0, 0x2924, 0xFCA8, 0x13FE, 0x10A8, 0x7D24, 0x1202, 0x11FC, 0xFD24, 0x1124, 0x11FC, 0x2124, 0x2124, 0x41FC, 0x8104}},
    {32692, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x4544, 0x2928, 0x7CFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x2010}},
    {32693, 2, {0x4420, 0x25FC, 0x2820, 0xFC88, 0x13FE, 0x1088, 0x7DFC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x1104, 0x21FC, 0x2088, 0x4104, 0x8202}},
    {32694, 2, {0x8820, 0x4BFE, 0x5000, 0xFBFC, 0x2204, 0x22F4, 0xFA94, 0x23FC, 0x2000, 0xF9F8, 0x2108, 0x21F8, 0x2108, 0x41F8, 0x4000, 0x83FE}},
    {32695, 2, {0x4420, 0x2420, 0x2850, 0xFC88, 0x1104, 0x12FA, 0x7C00, 0x1000, 0x11DC, 0xFD54, 0x1154, 0x11DC, 0x2088, 0x2088, 0x4154, 0x8222}},
    {32696, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x440, 0x729C, 0x57D4, 0x7114, 0x539C, 0x7116, 0x57D6, 0xB122}},
    {32697, 2, {0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x4824, 0x8282, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x280, 0xC60, 0x701C}},
    {32698, 2, {0x8800, 0x49FC, 0x5020, 0xFBFE, 0x2222, 0x21AC, 0xF820, 0x21AC, 0x2000, 0xFBFE, 0x2020, 0x21FC, 0x2154, 0x4154, 0x4154, 0x810C}},
    {32699, 2, {0x8820, 0x4BFE, 0x5250, 0xFBFE, 0x2252, 0x23FE, 0xFA00, 0x23FE, 0x2292, 0xFB48, 0x223A, 0x2240, 0x22FC, 0x4548, 0x4430, 0x89CE}},
    {32700, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2220, 0x2FF8, 0x2080, 0x27F0, 0x2080, 0x3FFC, 0x2AA8, 0x3F7C, 0x2410, 0x5F7C, 0x4410, 0xBF7E, 0x810}},
    {32701, 2, {0x0, 0x7EFC, 0x204, 0x204, 0x2244, 0x1224, 0x1224, 0x204, 0xA14, 0x1224, 0x62C4, 0x2244, 0x204, 0x204, 0x1428, 0x810}},
    {32702, 2, {0x0, 0xEE00, 0x227C, 0x2210, 0xAA10, 0x6610, 0x2210, 0x2210, 0x2610, 0x6A10, 0xB210, 0x2210, 0x2210, 0x2210, 0xAAFE, 0x4400}},
    {32703, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x60C, 0x1A34, 0x62C4, 0x204, 0x420, 0x420, 0x7FFE, 0x420, 0x420, 0x820, 0x1020, 0x2020}},
    {32704, 2, {0x10, 0xEE10, 0x2210, 0x2210, 0xAAFE, 0x6692, 0x2292, 0x2292, 0x2692, 0x6AFE, 0xB292, 0x2210, 0x2210, 0x2210, 0xAA10, 0x4410}},
    {32705, 2, {0x440, 0x820, 0x1210, 0x2408, 0xC826, 0x1FF0, 0x10, 0x0, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x2244, 0x4A94, 0x408}},
    {32706, 2, {0x8, 0xEE28, 0x2228, 0x2228, 0xAA44, 0x6644, 0x2282, 0x237C, 0x2624, 0x6A24, 0xB224, 0x2224, 0x2244, 0x2244, 0xAA94, 0x4508}},
    {32707, 2, {0x800, 0x8EE, 0x822, 0x7E22, 0x10AA, 0x1066, 0x2022, 0x2822, 0x4826, 0x486A, 0x90B2, 0x1422, 0x2222, 0x7E22, 0x22AA, 0x44}},
    {32708, 2, {0x10, 0xEE10, 0x2210, 0x22FE, 0xAA10, 0x6610, 0x22FC, 0x2244, 0x2644, 0x6A44, 0xB228, 0x2228, 0x2210, 0x2228, 0xAA44, 0x4482}},
    {32709, 2, {0x1000, 0x11DC, 0x1044, 0xFE44, 0x1154, 0x10CC, 0xFC44, 0x44CC, 0x4554, 0x2844, 0x2844, 0x1154, 0x1888, 0x2400, 0x43FE, 0x8000}},
    {32710, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x400, 0x3FC0, 0x440, 0x844, 0x113C, 0x2100, 0xFFFE, 0x100, 0x100, 0x100}},
    {32711, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x224, 0x110, 0x7FFE, 0x200, 0x7F0, 0xA10, 0x1120, 0x60C0, 0x330, 0x1C0E}},
    {32712, 2, {0x0, 0x7DDC, 0x5444, 0x5444, 0x5554, 0x7CCC, 0x5444, 0x5444, 0x544C, 0x7CD4, 0x5564, 0x1044, 0x1044, 0x1044, 0x1154, 0x1088}},
    {32713, 2, {0x1000, 0x11DC, 0x1044, 0xFC44, 0x1154, 0x30CC, 0x3844, 0x5444, 0x524C, 0x90D4, 0x1164, 0x7C44, 0x1044, 0x1044, 0x1154, 0x1088}},
    {32714, 2, {0x2000, 0x13DE, 0x1042, 0xFC42, 0x252, 0x94A, 0x894A, 0x8842, 0x48C6, 0x494A, 0x5252, 0x5042, 0x1C42, 0xE042, 0x414A, 0x84}},
    {32715, 2, {0x20, 0xEE10, 0x2210, 0x22FE, 0xAA00, 0x6604, 0x2244, 0x2244, 0x2624, 0x6A24, 0xB228, 0x2228, 0x2208, 0x2210, 0xAAFE, 0x4400}},
    {32716, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x60C, 0x1A34, 0x62C4, 0x100, 0x80, 0x7FFC, 0x0, 0x1010, 0x820, 0x440, 0xFFFE, 0x0}},
    {32717, 2, {0x10, 0xEE10, 0x2210, 0x22FE, 0xAA92, 0x6694, 0x2290, 0x22FC, 0x26A4, 0x6AA4, 0xB2A8, 0x22A8, 0x2290, 0x2328, 0xAA44, 0x4482}},
    {32718, 2, {0x1000, 0x11DC, 0x2844, 0x2444, 0x4354, 0x90CC, 0x844, 0x44, 0xFC4C, 0x4D4, 0x964, 0x5044, 0x2044, 0x1044, 0x1154, 0x88}},
    {32719, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x304, 0x6C0, 0x1830, 0xE30E, 0xC40, 0x3180, 0x630, 0x38C0, 0x700, 0x3800}},
    {32720, 2, {0x10, 0xEE10, 0x2250, 0x2250, 0xAA7C, 0x6690, 0x2310, 0x2210, 0x26FE, 0x6A10, 0xB228, 0x2228, 0x2228, 0x2244, 0xAA44, 0x4482}},
    {32721, 2, {0x40, 0xEE40, 0x2280, 0x22FC, 0xAB04, 0x6604, 0x22F4, 0x2294, 0x2694, 0x6A94, 0xB2F4, 0x2294, 0x2204, 0x2204, 0xAA28, 0x4410}},
    {32722, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x104, 0x200, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {32723, 2, {0x1000, 0x11DC, 0x1044, 0xFE44, 0x1154, 0x10CC, 0x7C44, 0x44, 0x4C, 0x7CD4, 0x4564, 0x4444, 0x4444, 0x7C44, 0x4554, 0x88}},
    {32724, 2, {0x4400, 0x25DC, 0x2844, 0xFE44, 0x1154, 0x10CC, 0x7C44, 0x1044, 0x104C, 0xFED4, 0x1164, 0x1044, 0x2044, 0x2044, 0x4154, 0x8088}},
    {32725, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x6AD4, 0x408}},
    {32726, 2, {0x1000, 0x11DC, 0x2844, 0x2444, 0x4354, 0x80CC, 0x7C44, 0x44, 0x4C, 0x7CD4, 0x4564, 0x4444, 0x4444, 0x7C44, 0x4554, 0x88}},
    {32727, 2, {0x1000, 0x10EE, 0x3E22, 0x4222, 0xA4AA, 0x1866, 0x1022, 0x2422, 0xC826, 0x1F6A, 0x21B2, 0xD222, 0xC22, 0x822, 0x30AA, 0xC044}},
    {32728, 2, {0x2000, 0x21DC, 0xFE44, 0x2444, 0x2954, 0x12CC, 0x2A44, 0xC6CC, 0x154, 0xFC44, 0x2844, 0x2954, 0x288A, 0x4802, 0x47FE, 0x8000}},
    {32729, 2, {0x1000, 0x92EE, 0x9222, 0x9222, 0xFEAA, 0x2066, 0x2022, 0x3E22, 0x4226, 0x426A, 0xA4B2, 0x1422, 0x822, 0x1022, 0x20AA, 0xC044}},
    {32730, 2, {0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x0, 0x7FFE, 0x4202, 0x9FF4, 0x400, 0x900, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100}},
    {32731, 2, {0x1080, 0x10F8, 0x1108, 0x2290, 0x2860, 0x6998, 0x6E06, 0xABB8, 0x2888, 0x2AA8, 0x2998, 0x2998, 0x2AA8, 0x2888, 0x22A8, 0x2110}},
    {32732, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x204, 0x100, 0x7FFC, 0x920, 0x920, 0x1550, 0x2288, 0x440, 0x1830, 0xE00E}},
    {32733, 2, {0x1000, 0x8EE, 0x7F22, 0x4122, 0x92AA, 0x1066, 0x7F22, 0x1022, 0x1426, 0x246A, 0x24B2, 0x2822, 0x4A22, 0x5122, 0xBFAA, 0x1144}},
    {32734, 2, {0x20, 0xEE10, 0x22FE, 0x2200, 0xAA00, 0x667C, 0x2244, 0x2244, 0x267C, 0x6A10, 0xB254, 0x2252, 0x2292, 0x2210, 0xAA50, 0x4420}},
    {32735, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x204, 0x900, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {32736, 2, {0x0, 0x7EFC, 0x2244, 0x1224, 0x1A34, 0x62C4, 0x204, 0x100, 0x7FFC, 0x820, 0x1450, 0x2288, 0x100, 0xFFFE, 0x100, 0x100}},
    {32737, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440, 0x0, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x6AD4, 0x408}},
    {32738, 2, {0x0, 0x7F6E, 0x4922, 0x4922, 0x5DAA, 0x4966, 0x7F22, 0x4122, 0x5D26, 0x556A, 0x55B2, 0x5D22, 0x4122, 0x4122, 0x452A, 0x8264}},
    {32739, 2, {0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x204, 0x100, 0x3FF8, 0x820, 0xFFFE, 0x400, 0x7FFC, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {32740, 2, {0x0, 0x7CEE, 0x4422, 0x4422, 0x7CAA, 0x66, 0xFE22, 0xAA22, 0xAA26, 0xAA6A, 0xFEB2, 0xAA22, 0xAA22, 0xAA22, 0xA2AA, 0x8644}},
    {32741, 2, {0x108, 0x1FD0, 0x120, 0x7FFC, 0x200, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0, 0x0, 0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x60C}},
    {32742, 2, {0x820, 0xFFFE, 0x0, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2618, 0x0, 0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x60C}},
    {32743, 2, {0x2000, 0x11DC, 0xFE44, 0x8244, 0x154, 0x7CCC, 0x44, 0x7C44, 0x444C, 0x7CD4, 0x4564, 0x7C44, 0x44, 0x1C44, 0xE154, 0x4088}},
    {32744, 2, {0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x204, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1100, 0x11F8, 0x2900, 0x47FE}},
    {32745, 2, {0x1000, 0x8EE, 0x7E22, 0x4222, 0x42AA, 0x7E66, 0x4222, 0x4022, 0x7F26, 0x556A, 0x55B2, 0x7F22, 0xD522, 0x5522, 0x55AA, 0x4344}},
    {32746, 2, {0x28, 0xEE92, 0x22AA, 0x2282, 0xAAFE, 0x6628, 0x2244, 0x22A2, 0x263C, 0x6A44, 0xB244, 0x22A8, 0x2210, 0x2228, 0xAA44, 0x4482}},
    {32747, 2, {0x0, 0xEE7C, 0x2200, 0xAA00, 0x6600, 0xAAFE, 0x2250, 0x1050, 0x7E50, 0x4250, 0x4250, 0x7E52, 0x4252, 0x4292, 0x7E8E, 0x4300}},
    {32748, 2, {0x7CF8, 0x2448, 0x1428, 0x2448, 0x7FFC, 0x4104, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {32749, 2, {0x40, 0xEE5C, 0x2244, 0x2284, 0xAABE, 0x6690, 0x2390, 0x229E, 0x26A8, 0x6A88, 0xB2BE, 0x2288, 0x2288, 0x2294, 0xAA94, 0x44A2}},
    {32750, 2, {0x0, 0xFEEE, 0x22, 0x7C22, 0x44AA, 0x7C66, 0x22, 0xFE22, 0x8226, 0xAA6A, 0x92B2, 0xFE22, 0x9222, 0x9222, 0x92AA, 0x8644}},
    {32751, 2, {0x7CF8, 0x2448, 0x1428, 0x2548, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {32752, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1106, 0x7C00, 0x45DC, 0x7C44, 0x4554, 0x7CCC, 0x1044, 0xFECC, 0x1154, 0x1044, 0x1154, 0x1088}},
    {32753, 2, {0x1000, 0x21DC, 0x7C44, 0x4444, 0x7D54, 0x44CC, 0x7C44, 0x1044, 0xFE4C, 0x28D4, 0x5564, 0x9244, 0x7C44, 0x1044, 0x1154, 0x1088}},
    {32754, 2, {0x0, 0xFEEE, 0x2822, 0xFE22, 0xAAAA, 0xAA66, 0xFE22, 0x22, 0x7C26, 0x6A, 0xFEB2, 0x1022, 0x5422, 0x9222, 0x50AA, 0x2044}},
    {32755, 2, {0x0, 0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4448, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x60C}},
    {32756, 2, {0x10, 0xEC10, 0x26FE, 0x2510, 0xB57C, 0x6C54, 0x247C, 0x2754, 0x257C, 0x6D10, 0xB5FE, 0x2510, 0x2510, 0x2690, 0xB47E, 0x4800}},
    {32757, 2, {0x10, 0xEE20, 0x227C, 0x2244, 0xAA7C, 0x6644, 0x227C, 0x2240, 0x267E, 0x6A40, 0xB27E, 0x2202, 0x22AA, 0x22AA, 0xAB02, 0x440C}},
    {32758, 2, {0x1000, 0x21DC, 0x7C44, 0x4444, 0x7D54, 0x44CC, 0x7C44, 0x1044, 0x544C, 0x38D4, 0x5564, 0x1044, 0xFE44, 0x1044, 0x1154, 0x1088}},
    {32759, 2, {0x800, 0x49EE, 0x2A22, 0x822, 0xFFAA, 0x2A66, 0x4922, 0x8822, 0x4226, 0x776A, 0x92B2, 0x5A22, 0x2F22, 0x2222, 0x42AA, 0x8244}},
    {32760, 2, {0x1000, 0x7EEE, 0x1022, 0x2422, 0xFEAA, 0x2466, 0x7E22, 0x4222, 0x7E26, 0x426A, 0x7EB2, 0x4222, 0x7E22, 0x2422, 0x42AA, 0x8144}},
    {32761, 2, {0x1000, 0x7DDC, 0x1044, 0xFE44, 0x4554, 0xEECC, 0x4444, 0xEECC, 0x154, 0xFC44, 0x2844, 0x2954, 0x288A, 0x4802, 0x47FE, 0x8000}},
    {32762, 2, {0x2000, 0xFEEE, 0x8222, 0xFE22, 0x82AA, 0xFE66, 0x8222, 0xFE22, 0x1026, 0x926A, 0x54B2, 0x9222, 0x1022, 0xFE22, 0x10AA, 0x1044}},
    {32763, 2, {0xE00, 0xF1DC, 0x9244, 0x5444, 0xFF54, 0x38CC, 0x5444, 0x8244, 0x7C4C, 0x54D4, 0x5564, 0x7C44, 0x5444, 0x5444, 0x7D54, 0x4488}},
    {32764, 2, {0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x204, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010}},
    {32765, 2, {0x800, 0x4E76, 0x4812, 0xFF12, 0x45A, 0x536, 0x7F92, 0x4412, 0x7C12, 0x5536, 0x555A, 0x7512, 0x5A12, 0x9292, 0x25DA, 0xC8A4}},
    {32766, 2, {0x0, 0xFEEE, 0xAA22, 0xAA22, 0xFEAA, 0x66, 0xFE22, 0x22, 0x7C26, 0x446A, 0x7CB2, 0x2A22, 0x6422, 0xA222, 0x31AA, 0x2044}},
    {32767, 2, {0x1000, 0xFEEE, 0x1022, 0x7C22, 0xAA, 0xFE66, 0x222, 0x7C22, 0x1026, 0xFE6A, 0xB2, 0xFF22, 0x222, 0xFF22, 0xAAAA, 0xE644}},
    {32768, 2, {0x1000, 0x11DC, 0x9044, 0x5554, 0x58CC, 0x1154, 0xFC00, 0x2848, 0x28FE, 0x2990, 0x2AFC, 0x2890, 0x2AFC, 0x4C90, 0x48FE, 0x8080}},
    {32769, 2, {0x200, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x100, 0x200, 0xC10, 0x18E0, 0x2F00, 0x4808, 0x8808, 0x808, 0x7F8, 0x0}},
    {32770, 2, {0x200, 0x208, 0x3FC8, 0x210, 0x220, 0xFFFE, 0x80, 0x100, 0x200, 0xC00, 0x3000, 0xC000, 0x0, 0x0, 0x0, 0x0}},
    {32771, 2, {0x200, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x100, 0x200, 0xFF8, 0x1200, 0x2400, 0x47F0, 0x8010, 0x10, 0xA0, 0x40}},
    {32772, 2, {0x220, 0x3FC0, 0x280, 0xFFFE, 0x220, 0xFC0, 0x3408, 0xC3F8, 0x0, 0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE}},
    {32773, 2, {0x200, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x100, 0x200, 0xFF0, 0x1810, 0x2810, 0x4FF0, 0x8810, 0x810, 0xFF0, 0x810}},
    {32774, 2, {0x208, 0x3FD0, 0x220, 0xFFFE, 0x200, 0xC70, 0x3F84, 0xC804, 0x7FC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {32775, 2, {0x200, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x300, 0x1C00, 0xEFF8, 0x1008, 0x2FC8, 0xC848, 0xFC8, 0x8, 0x50, 0x20}},
    {32776, 2, {0x220, 0x3FC0, 0x280, 0xFFFE, 0x220, 0xFC0, 0x3408, 0xC3F8, 0x1000, 0x3FF8, 0x4008, 0x9F88, 0x1088, 0x1F88, 0x50, 0x20}},
    {32777, 2, {0x200, 0x210, 0x3FE0, 0x240, 0xFFFE, 0x100, 0x7F8, 0x1C00, 0xE7F8, 0x8, 0x1F88, 0x1088, 0x1088, 0x1F88, 0x50, 0x20}},
    {32778, 2, {0x200, 0x210, 0x3FE0, 0x240, 0xFFFE, 0x100, 0x600, 0x1FF0, 0xE400, 0x820, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x0}},
    {32779, 2, {0x220, 0x3FC0, 0x280, 0xFFFE, 0x220, 0xFC0, 0x3408, 0xC3F8, 0x0, 0x7FFC, 0x820, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC}},
    {32780, 2, {0x0, 0x7FFC, 0x100, 0x100, 0x200, 0x3FF8, 0x2488, 0x2488, 0x2488, 0x2488, 0x2488, 0x2488, 0x2488, 0x2488, 0x2028, 0x2010}},
    {32781, 2, {0x0, 0xFFFE, 0x200, 0x400, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2210, 0x200, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {32782, 2, {0x0, 0xFFFE, 0x200, 0x400, 0x3FF8, 0x2488, 0x2488, 0x2488, 0x24A8, 0x2110, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {32783, 2, {0x2, 0xFFE2, 0x404, 0x408, 0x810, 0x7FC2, 0x4A42, 0x4A44, 0x4A48, 0x4A50, 0x4A42, 0x4A42, 0x4A44, 0x4848, 0x4150, 0x40A0}},
    {32784, 2, {0x8, 0xFF88, 0x808, 0x808, 0x10FE, 0x7F08, 0x5508, 0x5508, 0x5548, 0x5528, 0x5528, 0x5508, 0x5508, 0x5508, 0x4128, 0x4310}},
    {32785, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x200, 0x400, 0x3FF8, 0x2488, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {32786, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100, 0x100}},
    {32787, 2, {0x1000, 0x1000, 0x11FE, 0xFE10, 0x1010, 0x7C10, 0x1010, 0xFE10, 0x1010, 0x3810, 0x3410, 0x5410, 0x5010, 0x9010, 0x1050, 0x1020}},
    {32788, 2, {0x1000, 0x10FC, 0x1004, 0xFC08, 0x1008, 0x7810, 0x1010, 0xFDFE, 0x1010, 0x3810, 0x3410, 0x5410, 0x9010, 0x1010, 0x1050, 0x1020}},
    {32789, 2, {0x1088, 0x1088, 0x1088, 0xFE88, 0x11FC, 0x7C88, 0x1088, 0xFE88, 0x13FE, 0x3888, 0x3488, 0x5488, 0x5088, 0x9108, 0x1108, 0x1208}},
    {32790, 2, {0x1020, 0x1020, 0x1020, 0xFCA8, 0x10A4, 0x78A2, 0x1122, 0xFD20, 0x1224, 0x3824, 0x3428, 0x5408, 0x9010, 0x1020, 0x10C0, 0x1300}},
    {32791, 2, {0x1008, 0x101C, 0x10E0, 0xFE20, 0x1020, 0x7C3C, 0x10E0, 0xFE20, 0x1020, 0x383E, 0x35E0, 0x5420, 0x5022, 0x9022, 0x1022, 0x101E}},
    {32792, 2, {0x1000, 0x10FC, 0x1000, 0xFE00, 0x1000, 0x7C00, 0x11FE, 0xFE20, 0x1020, 0x3840, 0x3448, 0x5484, 0x5084, 0x91FE, 0x1082, 0x1002}},
    {32793, 2, {0x1000, 0x1000, 0x11FC, 0xFD24, 0x1124, 0x7924, 0x1124, 0xFD24, 0x11FC, 0x3900, 0x3500, 0x5500, 0x9102, 0x1102, 0x10FE, 0x1000}},
    {32794, 2, {0x1010, 0x1010, 0x1010, 0xFDFE, 0x1112, 0x7914, 0x1110, 0xFDFC, 0x1144, 0x3944, 0x3528, 0x5528, 0x9110, 0x1228, 0x1244, 0x1482}},
    {32795, 2, {0x1020, 0x1020, 0x1020, 0xFE40, 0x1048, 0x7C84, 0x11FE, 0xFE82, 0x1000, 0x38FC, 0x3484, 0x5484, 0x5084, 0x9084, 0x10FC, 0x1084}},
    {32796, 2, {0x1000, 0x11F8, 0x1108, 0xFD08, 0x1108, 0x7908, 0x11F8, 0xFD00, 0x1100, 0x39FC, 0x3504, 0x5504, 0x9104, 0x1104, 0x11FC, 0x1104}},
    {32797, 2, {0x1000, 0x10F8, 0x1088, 0xFE88, 0x1088, 0x7CF8, 0x1088, 0xFE88, 0x1088, 0x38F8, 0x3488, 0x5488, 0x5088, 0x9088, 0x13FE, 0x1000}},
    {32798, 2, {0x1100, 0x1100, 0x1100, 0xFD1C, 0x13D4, 0x7954, 0x1154, 0xFD54, 0x1154, 0x3954, 0x3554, 0x5554, 0x9154, 0x125C, 0x1354, 0x1480}},
    {32799, 2, {0x1000, 0x11FE, 0x1100, 0xFF00, 0x1100, 0x7DFC, 0x1104, 0xFF04, 0x1104, 0x3904, 0x35FC, 0x5500, 0x5100, 0x9100, 0x11FE, 0x1000}},
    {32800, 2, {0x1020, 0x1020, 0x1050, 0xFC88, 0x1104, 0x7A02, 0x11FC, 0xFC00, 0x1000, 0x39FC, 0x3504, 0x5504, 0x9104, 0x1104, 0x11FC, 0x1104}},
    {32801, 2, {0x1008, 0x1008, 0x13C8, 0xFE48, 0x125E, 0x7A4A, 0x13CA, 0xFE4A, 0x124A, 0x3BCA, 0x364A, 0x524A, 0x526A, 0x93D2, 0x1612, 0x1026}},
    {32802, 2, {0x1088, 0x1088, 0x13FE, 0xFC88, 0x1000, 0x7BFE, 0x1202, 0xFC44, 0x1040, 0x3BFC, 0x3444, 0x5484, 0x9084, 0x1104, 0x1228, 0x1410}},
    {32803, 2, {0x1020, 0x1020, 0x1050, 0xFC88, 0x1104, 0x7AFA, 0x1000, 0xFDFC, 0x1154, 0x3954, 0x35FC, 0x5554, 0x9154, 0x1154, 0x1104, 0x110C}},
    {32804, 2, {0x1048, 0x1048, 0x1048, 0xFDFE, 0x1048, 0x7848, 0x11FE, 0xFC00, 0x10FC, 0x3884, 0x3484, 0x54FC, 0x9084, 0x1084, 0x10FC, 0x1084}},
    {32805, 2, {0x1020, 0x1124, 0x10A4, 0xFCA8, 0x1020, 0x7BFE, 0x1202, 0xFE02, 0x12FA, 0x3A8A, 0x368A, 0x528A, 0x52FA, 0x9202, 0x120A, 0x1204}},
    {32806, 2, {0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x7924, 0x11FC, 0xFC20, 0x1020, 0x3BFE, 0x3622, 0x522A, 0x52FA, 0x920A, 0x1202, 0x1206}},
    {32807, 2, {0x1020, 0x1124, 0x10A8, 0xFC20, 0x13FE, 0x78A8, 0x1124, 0xFE02, 0x1040, 0x3BFE, 0x3488, 0x5108, 0x5190, 0x9060, 0x1198, 0x1604}},
    {32808, 2, {0x1000, 0x11FE, 0x1100, 0xFD7E, 0x1100, 0x79FE, 0x1154, 0xFD48, 0x1164, 0x3942, 0x3508, 0x51FE, 0x5148, 0x9228, 0x1208, 0x1418}},
    {32809, 2, {0x1088, 0x1088, 0x13FE, 0xFC88, 0x11FC, 0x7888, 0x13FE, 0xFC20, 0x11FC, 0x3924, 0x35FC, 0x5124, 0x53FE, 0x9104, 0x1114, 0x1108}},
    {32810, 2, {0x1040, 0x1020, 0x13FC, 0xFD08, 0x1090, 0x7BFE, 0x1202, 0xFC44, 0x1020, 0x3BFC, 0x3480, 0x54F8, 0x9088, 0x1108, 0x1128, 0x1210}},
    {32811, 2, {0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1020, 0x7BFE, 0x1000, 0xFDFC, 0x1104, 0x39FC, 0x3504, 0x55FC, 0x9104, 0x11FC, 0x1088, 0x1104}},
    {32812, 2, {0x1020, 0x11FC, 0x1124, 0xFFFE, 0x1124, 0x79FC, 0x1020, 0xFDFC, 0x1124, 0x39FC, 0x3440, 0x55FE, 0x9088, 0x11D0, 0x1070, 0x138C}},
    {32813, 2, {0x1124, 0x1124, 0x12AA, 0xFFAE, 0x1124, 0x7AAA, 0x13AE, 0xFD24, 0x13FE, 0x3910, 0x3514, 0x5114, 0x528A, 0x924A, 0x1216, 0x1422}},
    {32814, 2, {0x1048, 0x116A, 0x10DC, 0xFE48, 0x10B4, 0x7D22, 0x11FE, 0xFF02, 0x1020, 0x3820, 0x34FC, 0x5424, 0x5044, 0x9044, 0x1094, 0x1108}},
    {32815, 2, {0x1088, 0x13FE, 0x10A8, 0xFC90, 0x11FE, 0x7B20, 0x11FC, 0xFD20, 0x11FC, 0x3920, 0x35FE, 0x5500, 0x93FC, 0x1088, 0x1070, 0x138E}},
    {32816, 2, {0x13FC, 0x1040, 0x11F8, 0xFD08, 0x11F8, 0x7908, 0x17FE, 0xFC42, 0x12A4, 0x3A8A, 0x347A, 0x5080, 0x51F8, 0x9288, 0x1070, 0x138E}},
    {32817, 2, {0x1040, 0x1020, 0x13FE, 0xFE48, 0x13FE, 0x7A48, 0x12EC, 0xFF5A, 0x1248, 0x3BFE, 0x3620, 0x527C, 0x52C4, 0x9344, 0x147C, 0x1844}},
    {32818, 2, {0x1040, 0x1020, 0x13FE, 0xFC00, 0x11FC, 0x7954, 0x11FC, 0xFC20, 0x1122, 0x388C, 0x3522, 0x5454, 0x90C8, 0x1144, 0x1252, 0x1060}},
    {32819, 2, {0x0, 0x7FFC, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x103E, 0x17D0, 0xF810, 0x10, 0x10, 0x10}},
    {32820, 2, {0x20, 0xFFA0, 0x2220, 0x2220, 0x3E20, 0x2220, 0x2220, 0x3E20, 0x2220, 0x2220, 0x27A0, 0xFA22, 0x4222, 0x222, 0x21E, 0x200}},
    {32821, 2, {0x0, 0xFF80, 0x227E, 0x2208, 0x3E08, 0x2208, 0x2208, 0x3E08, 0x2208, 0x2208, 0x2788, 0xFA08, 0x4208, 0x208, 0x228, 0x210}},
    {32822, 2, {0x0, 0xFFBE, 0x2222, 0x2224, 0x3E24, 0x2228, 0x2224, 0x3E24, 0x2222, 0x2222, 0x27A2, 0xFA34, 0x4228, 0x220, 0x220, 0x220}},
    {32823, 2, {0x100, 0x100, 0x7FFC, 0x280, 0x440, 0x820, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE0, 0x820, 0x9F8, 0x7E20, 0x20, 0x20}},
    {32824, 2, {0x820, 0x820, 0x1450, 0x2288, 0x4104, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x103E, 0xFFD0, 0x10, 0x10}},
    {32825, 2, {0x20, 0xFF20, 0x2420, 0x2450, 0x3C50, 0x2488, 0x2544, 0x3E22, 0x2420, 0x24F8, 0x2E08, 0xF408, 0x4410, 0x410, 0x420, 0x420}},
    {32826, 2, {0x0, 0xFF00, 0x24FC, 0x2400, 0x3C00, 0x2400, 0x25FE, 0x3C20, 0x2420, 0x2440, 0x2E40, 0xF488, 0x4484, 0x5FE, 0x482, 0x400}},
    {32827, 2, {0x10, 0xFF10, 0x2410, 0x2490, 0x3C90, 0x2490, 0x249E, 0x3C90, 0x2490, 0x2490, 0x2E90, 0xF490, 0x4490, 0x490, 0x5FE, 0x400}},
    {32828, 2, {0x0, 0xFF7C, 0x2444, 0x2444, 0x3C44, 0x24FE, 0x2444, 0x3C44, 0x2444, 0x24FE, 0x2E44, 0xF444, 0x4444, 0x444, 0x454, 0x448}},
    {32829, 2, {0x40, 0xFC40, 0x4840, 0x4BFC, 0x7A44, 0x4C48, 0x4840, 0x7840, 0x48A0, 0x48A0, 0x4CA0, 0x7920, 0xC922, 0xA22, 0xC1E, 0x800}},
    {32830, 2, {0x40, 0xFF40, 0x2440, 0x2440, 0x3DFE, 0x2440, 0x2440, 0x3C50, 0x2490, 0x2490, 0x2EA0, 0xF528, 0x4544, 0x644, 0x4FE, 0x442}},
    {32831, 2, {0x20, 0xFF20, 0x2420, 0x2424, 0x3CA4, 0x24A4, 0x24A8, 0x3D20, 0x2420, 0x2450, 0x2E50, 0xF450, 0x4488, 0x488, 0x504, 0x602}},
    {32832, 2, {0x48, 0xFF44, 0x2444, 0x2440, 0x3C5E, 0x25E0, 0x2440, 0x3C44, 0x2444, 0x2448, 0x2E30, 0xF422, 0x4452, 0x48A, 0x506, 0x402}},
    {32833, 2, {0x10, 0xFC90, 0x4890, 0x4888, 0x7908, 0x4904, 0x4A04, 0x7DFA, 0x4888, 0x4888, 0x4C88, 0x7888, 0xC908, 0x908, 0xA28, 0xC10}},
    {32834, 2, {0x0, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x3EF8, 0x288, 0x2288, 0x1450, 0x820, 0x34D8, 0xC306}},
    {32835, 2, {0x20, 0xFF20, 0x2420, 0x25FC, 0x3D24, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x2524, 0x2FFE, 0xF504, 0x4504, 0x504, 0x514, 0x508}},
    {32836, 2, {0x20, 0xFC20, 0x4850, 0x4888, 0x7904, 0x4A12, 0x4820, 0x7840, 0x4988, 0x4810, 0x4C20, 0x7844, 0xC988, 0x810, 0x860, 0xB80}},
    {32837, 2, {0x40, 0xFC40, 0x4880, 0x48FC, 0x7920, 0x4A20, 0x4820, 0x7820, 0x4BFE, 0x4820, 0x4C50, 0x7850, 0xC888, 0x888, 0x904, 0xA02}},
    {32838, 2, {0x20, 0xFF20, 0x2450, 0x2450, 0x3C88, 0x2524, 0x2612, 0x3C10, 0x25FC, 0x2404, 0x2E08, 0xF488, 0x4450, 0x420, 0x410, 0x410}},
    {32839, 2, {0x0, 0xFF00, 0x24FE, 0x2410, 0x3C10, 0x2410, 0x2490, 0x3C90, 0x249E, 0x2490, 0x2E90, 0xF490, 0x4490, 0x490, 0x7FE, 0x400}},
    {32840, 2, {0x10, 0xFC90, 0x4890, 0x4910, 0x797E, 0x4A52, 0x4B92, 0x7892, 0x4912, 0x4912, 0x4E52, 0x7BD2, 0xC862, 0x822, 0x84A, 0x884}},
    {32841, 2, {0x10, 0xFF10, 0x2492, 0x2492, 0x3C92, 0x2492, 0x24FE, 0x3C10, 0x2410, 0x2492, 0x2E92, 0xF492, 0x4492, 0x492, 0x4FE, 0x402}},
    {32842, 2, {0x0, 0xFC80, 0x4B1E, 0x4A52, 0x7A52, 0x4A52, 0x4A52, 0x7A52, 0x4A52, 0x4AD2, 0x4F5A, 0x7A54, 0xC890, 0x890, 0x910, 0xA10}},
    {32843, 2, {0x480, 0x440, 0xFFFE, 0x940, 0x1184, 0x66FC, 0x0, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32844, 2, {0x0, 0xFF00, 0x24FC, 0x2484, 0x3C84, 0x2484, 0x2484, 0x3C84, 0x24FC, 0x2484, 0x2E00, 0xF448, 0x4444, 0x484, 0x482, 0x502}},
    {32845, 2, {0x40, 0xFC20, 0x4820, 0x4BFE, 0x7A02, 0x4C04, 0x4800, 0x7800, 0x4BFE, 0x4820, 0x4C20, 0x7820, 0xC820, 0x820, 0x8A0, 0x840}},
    {32846, 2, {0x50, 0xFF50, 0x2450, 0x2552, 0x3CD4, 0x2458, 0x2450, 0x3C58, 0x24D4, 0x2552, 0x2E50, 0xF450, 0x4492, 0x492, 0x512, 0x60E}},
    {32847, 2, {0x0, 0xFDFC, 0x4820, 0x4820, 0x7840, 0x49FC, 0x4954, 0x7954, 0x4954, 0x4954, 0x4D54, 0x7954, 0xC954, 0x944, 0x914, 0x908}},
    {32848, 2, {0x20, 0xFC20, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x49FC, 0x7800, 0x4800, 0x49FC, 0x4D04, 0x7904, 0xC904, 0x904, 0x9FC, 0x904}},
    {32849, 2, {0x0, 0xFDFE, 0x4888, 0x4888, 0x7888, 0x48F8, 0x4888, 0x7888, 0x48F8, 0x4888, 0x4C88, 0x789E, 0xCBE8, 0x808, 0x808, 0x808}},
    {32850, 2, {0x8, 0xFF1C, 0x24E0, 0x2420, 0x3C20, 0x25FE, 0x2420, 0x3C20, 0x2420, 0x24FC, 0x2E84, 0xF484, 0x4484, 0x484, 0x4FC, 0x484}},
    {32851, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1D52, 0xF20E, 0x400, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32852, 2, {0x88, 0xFC48, 0x4850, 0x4800, 0x79FC, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x4820, 0x4C50, 0x7850, 0xC888, 0x888, 0x904, 0xA02}},
    {32853, 2, {0x20, 0xFD20, 0x4920, 0x49FC, 0x7920, 0x4A20, 0x4820, 0x7BFE, 0x4800, 0x4800, 0x4DFC, 0x7904, 0xC904, 0x904, 0x9FC, 0x904}},
    {32854, 2, {0x0, 0x7F00, 0x227C, 0x3E44, 0x2244, 0x3E44, 0x23C4, 0xFE7C, 0x244, 0x200, 0x7FFC, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {32855, 2, {0x20, 0xFC20, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x4924, 0x7924, 0x4924, 0x4AAA, 0x4C20, 0x7850, 0xC850, 0x888, 0x904, 0xA02}},
    {32856, 2, {0x20, 0xFC20, 0x49FC, 0x4924, 0x79FC, 0x4924, 0x49FC, 0x7800, 0x4BFE, 0x4880, 0x4D00, 0x79FC, 0xC804, 0x804, 0x828, 0x810}},
    {32857, 2, {0x20, 0xFC20, 0x4BFE, 0x4820, 0x79FC, 0x4820, 0x4BFE, 0x7800, 0x49FC, 0x4904, 0x4DFC, 0x7904, 0xC9FC, 0x904, 0x914, 0x908}},
    {32858, 2, {0x7F00, 0x227C, 0x3E04, 0x2228, 0x3E10, 0x23A8, 0xFE44, 0x200, 0xF8, 0x7F00, 0x904, 0x1188, 0x6350, 0x520, 0x1918, 0xE106}},
    {32859, 2, {0x20, 0xFF40, 0x24FE, 0x2492, 0x3C92, 0x24FE, 0x2492, 0x3CA2, 0x24FE, 0x2448, 0x2E88, 0xF5FE, 0x4408, 0x408, 0x408, 0x408}},
    {32860, 2, {0x20, 0xFF10, 0x2410, 0x25FE, 0x3D02, 0x2654, 0x2488, 0x3D04, 0x2400, 0x24FC, 0x2E20, 0xF420, 0x4420, 0x420, 0x5FE, 0x400}},
    {32861, 2, {0x14, 0xFC12, 0x4810, 0x49FE, 0x7810, 0x4810, 0x49D2, 0x7952, 0x4952, 0x4954, 0x4DD4, 0x7808, 0xC8CA, 0xB1A, 0x826, 0x842}},
    {32862, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4444, 0x47C4, 0x4444, 0x47C4, 0x4474, 0x5FC4, 0x4054, 0x4048}},
    {32863, 2, {0x2000, 0x3E7C, 0x4844, 0xFF44, 0x1444, 0x227C, 0x4000, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32864, 2, {0x84, 0xFF48, 0x2400, 0x24FC, 0x3C48, 0x2448, 0x2448, 0x3C48, 0x25FE, 0x2448, 0x2E48, 0xF448, 0x4448, 0x488, 0x488, 0x508}},
    {32865, 2, {0x90, 0xFC90, 0x4908, 0x4948, 0x7A44, 0x4C92, 0x4908, 0x7BFC, 0x4844, 0x4820, 0x4CA4, 0x7A82, 0xCA8A, 0xA8A, 0xC78, 0x800}},
    {32866, 2, {0x20, 0xFF10, 0x24FE, 0x2482, 0x3D04, 0x2400, 0x24FE, 0x3C10, 0x2410, 0x2490, 0x2E9E, 0xF490, 0x4490, 0x4D0, 0x53E, 0x600}},
    {32867, 2, {0x0, 0xFF20, 0x24CE, 0x2482, 0x3C82, 0x24EE, 0x2482, 0x3C82, 0x24FE, 0x2428, 0x2E28, 0xF428, 0x4448, 0x44A, 0x48A, 0x506}},
    {32868, 2, {0x40, 0xFC20, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x7800, 0x4BFE, 0x4A02, 0x4DFC, 0x7820, 0xC820, 0x820, 0x8A0, 0x840}},
    {32869, 2, {0x4, 0xFC1E, 0x4BE0, 0x4820, 0x79FC, 0x4924, 0x4924, 0x79FC, 0x4820, 0x4BFE, 0x4E22, 0x7A2A, 0xCAFA, 0xA02, 0xA0A, 0xA04}},
    {32870, 2, {0x80, 0xFC80, 0x48FC, 0x4954, 0x7AD4, 0x48B4, 0x492C, 0x7A44, 0x4894, 0x4908, 0x4C40, 0x78A4, 0xCAAA, 0xA8A, 0xC78, 0x800}},
    {32871, 2, {0x10, 0xFDD4, 0x4858, 0x4A52, 0x798C, 0x4888, 0x4904, 0x7AFA, 0x4820, 0x4820, 0x4DFE, 0x7820, 0xC850, 0x888, 0x904, 0xA02}},
    {32872, 2, {0x108, 0xFD4A, 0x4A52, 0x4B9C, 0x7908, 0x4A52, 0x4BDE, 0x7842, 0x4888, 0x4888, 0x4FFE, 0x7888, 0xC888, 0x908, 0x908, 0xA08}},
    {32873, 2, {0x20, 0xFC20, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x4BFE, 0x7800, 0x49FC, 0x4904, 0x4D24, 0x7924, 0xC924, 0x850, 0x888, 0xB04}},
    {32874, 2, {0x88, 0xFC48, 0x4850, 0x49FC, 0x7904, 0x4904, 0x4904, 0x79FC, 0x4820, 0x4810, 0x4C54, 0x7942, 0xC942, 0x94A, 0xA38, 0x800}},
    {32875, 2, {0x108, 0xFD4A, 0x4A52, 0x4B9C, 0x7908, 0x4A52, 0x4BDE, 0x7842, 0x4820, 0x4820, 0x4FFE, 0x7820, 0xC850, 0x888, 0x904, 0xA02}},
    {32876, 2, {0x90, 0xFC88, 0x4904, 0x4A42, 0x7C88, 0x49FC, 0x4804, 0x7800, 0x4BDE, 0x4842, 0x4E52, 0x794A, 0xCA52, 0x842, 0x94A, 0x884}},
    {32877, 2, {0x10, 0xFF20, 0x24FE, 0x2492, 0x3C92, 0x24FE, 0x2492, 0x3C92, 0x24FE, 0x2420, 0x2E24, 0xF45A, 0x445E, 0x490, 0x492, 0x50E}},
    {32878, 2, {0x84, 0xFC84, 0x4908, 0x494A, 0x7A52, 0x4B9C, 0x4884, 0x7908, 0x494A, 0x4BDE, 0x4C42, 0x7800, 0xC954, 0x92A, 0xA2A, 0x800}},
    {32879, 2, {0x108, 0xFD4A, 0x4A52, 0x4B9C, 0x7908, 0x4A52, 0x4BDE, 0x7842, 0x4890, 0x4A92, 0x4E92, 0x7B9E, 0xC890, 0x890, 0x890, 0x910}},
    {32880, 2, {0x20, 0xFC40, 0x49FC, 0x4924, 0x797C, 0x498C, 0x4954, 0x7924, 0x4954, 0x49FC, 0x4C40, 0x7824, 0xCAA2, 0xA8A, 0xC88, 0x878}},
    {32881, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x7FFE, 0x8820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32882, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0xFFFE, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32883, 2, {0x1110, 0x2110, 0xCAA8, 0x1444, 0x3278, 0xD240, 0x15FC, 0x1800, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20}},
    {32884, 2, {0x20, 0xFC20, 0x4BFE, 0x4840, 0x7840, 0x49FC, 0x4954, 0x7954, 0x49FC, 0x4840, 0x4C24, 0x7AA2, 0xCA8A, 0xC88, 0x878, 0x800}},
    {32885, 2, {0x20, 0xFDFC, 0x4924, 0x49FC, 0x7820, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x4D04, 0x79FC, 0xC904, 0x9FC, 0x888, 0x904}},
    {32886, 2, {0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x7EFC, 0x2448, 0x3C78, 0x2448, 0x3C78, 0x264E, 0xFDF8, 0x408}},
    {32887, 2, {0x110, 0xFC94, 0x4BD2, 0x4812, 0x7A50, 0x4990, 0x4FFE, 0x7810, 0x4BD2, 0x4A52, 0x4E54, 0x7BD4, 0xCA4A, 0xA4A, 0xBD6, 0x822}},
    {32888, 2, {0x80, 0xFCF8, 0x4908, 0x4BFE, 0x7944, 0x4992, 0x49FE, 0x7900, 0x497C, 0x4900, 0x4D7C, 0x7900, 0xC97C, 0xA44, 0xA7C, 0xC44}},
    {32889, 2, {0x40, 0xFC20, 0x4BFE, 0x4A22, 0x7954, 0x494A, 0x4A3A, 0x7800, 0x49FC, 0x4954, 0x4FFE, 0x7800, 0xC9FC, 0x820, 0x8A0, 0x840}},
    {32890, 2, {0x40, 0xFC20, 0x4BFE, 0x4A02, 0x7910, 0x49DE, 0x4A52, 0x7D54, 0x4A88, 0x4974, 0x4E02, 0x79FC, 0xC820, 0x924, 0xA22, 0x860}},
    {32891, 2, {0x4204, 0x3FD8, 0x8A90, 0x4F9E, 0x2A94, 0x5FD4, 0x8224, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32892, 2, {0x20, 0x7820, 0xBFE, 0x1020, 0xFDFC, 0x4954, 0x4954, 0x79FC, 0x4800, 0x7BFE, 0x4840, 0x5C24, 0xEAA2, 0x4A8A, 0xC88, 0x878}},
    {32893, 2, {0x20, 0xFC20, 0x4BFE, 0x7820, 0x49FC, 0x7954, 0x4954, 0xF9FC, 0x800, 0xFBFE, 0x2840, 0x7824, 0x2AA2, 0xFA8A, 0xC88, 0x878}},
    {32894, 2, {0x840, 0x3E7C, 0x1440, 0xFF78, 0x2208, 0x3E78, 0x2242, 0x263E, 0x0, 0x7FFC, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20, 0x20}},
    {32895, 2, {0x100, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {32896, 2, {0x100, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x108, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {32897, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2080, 0x4FFC, 0x4080, 0xBFFE, 0x80}},
    {32898, 2, {0x20, 0x20, 0xFDFC, 0x2024, 0x23FE, 0x3C24, 0x45FC, 0x4420, 0x6420, 0x95FC, 0x820, 0x820, 0x13FE, 0x2020, 0x4020, 0x8020}},
    {32899, 2, {0x100, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x2108, 0x2548, 0x2528, 0x2528, 0x2928, 0x4108, 0x4108, 0x8108}},
    {32900, 2, {0x4020, 0x4C20, 0x70FC, 0x4424, 0x45FE, 0x3C24, 0x40FC, 0x7C20, 0x9020, 0x11FC, 0xFE20, 0x1020, 0x29FE, 0x2420, 0x4420, 0x8020}},
    {32901, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x540, 0x2548, 0x3D78, 0x2108, 0x3FF8, 0x2108, 0x3D78, 0x2548, 0x4548, 0x8108}},
    {32902, 2, {0x20, 0x3E20, 0x20FC, 0x2024, 0x3DFE, 0x2024, 0x20FC, 0x3C20, 0x2020, 0x21FC, 0xFE20, 0x2020, 0x25FE, 0x4220, 0xFE20, 0x4220}},
    {32903, 2, {0x840, 0x3E7E, 0x22A8, 0x3E10, 0x206E, 0x4100, 0x9FF0, 0x110, 0xFFFE, 0x110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100}},
    {32904, 2, {0x824, 0x3E3E, 0x22E8, 0x3E12, 0x202A, 0x4146, 0x9FF0, 0x110, 0xFFFE, 0x110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100}},
    {32905, 2, {0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x4284, 0x4444, 0x4924, 0x5104, 0x4104, 0x4284, 0x4244, 0x4444, 0x4804, 0x4014, 0x4008}},
    {32906, 2, {0x0, 0x3E00, 0x22FC, 0x2204, 0x2208, 0x3E10, 0x2220, 0x2220, 0x2240, 0x3E40, 0x2280, 0x2282, 0x2282, 0x4282, 0x4A7E, 0x8400}},
    {32907, 2, {0x20, 0x3E20, 0x2220, 0x2220, 0x22FE, 0x3E22, 0x2222, 0x2222, 0x2222, 0x3E22, 0x2242, 0x2242, 0x2242, 0x4282, 0x4A94, 0x8508}},
    {32908, 2, {0x0, 0x3CF0, 0x2490, 0x2490, 0x2490, 0x3C90, 0x2490, 0x2490, 0x2490, 0x3C90, 0x2490, 0x2492, 0x2492, 0x4512, 0x550E, 0x8A00}},
    {32909, 2, {0x40, 0x3C40, 0x2440, 0x2440, 0x25F8, 0x3C48, 0x2448, 0x2448, 0x2448, 0x3C48, 0x2448, 0x244A, 0x248A, 0x448A, 0x5506, 0x8A00}},
    {32910, 2, {0x0, 0x7FFE, 0x4002, 0x8004, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32911, 2, {0x600, 0x100, 0x280, 0xC60, 0x3118, 0xC106, 0x3FF8, 0x2288, 0x2448, 0x2928, 0x2108, 0x2288, 0x2448, 0x2008, 0x2028, 0x2010}},
    {32912, 2, {0x40, 0x3C40, 0x2480, 0x24FE, 0x2500, 0x3E00, 0x24FC, 0x2408, 0x2410, 0x3C20, 0x2440, 0x2480, 0x2502, 0x4502, 0x54FE, 0x8800}},
    {32913, 2, {0x40, 0x3C40, 0x2480, 0x24FC, 0x2504, 0x3E04, 0x2404, 0x2504, 0x2484, 0x3C44, 0x2444, 0x2404, 0x2404, 0x4404, 0x5428, 0x8810}},
    {32914, 2, {0x40, 0x3C40, 0x2440, 0x2440, 0x25F8, 0x3C48, 0x2448, 0x2548, 0x24C8, 0x3C48, 0x2468, 0x245A, 0x248A, 0x448A, 0x5506, 0x8A02}},
    {32915, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32916, 2, {0x20, 0x3C20, 0x2420, 0x2520, 0x252C, 0x3D34, 0x2564, 0x27A4, 0x2524, 0x3D34, 0x2528, 0x2522, 0x2522, 0x4502, 0x54FE, 0x8800}},
    {32917, 2, {0x0, 0x3C00, 0x24FC, 0x2424, 0x2424, 0x3CA4, 0x24A4, 0x24A4, 0x24A4, 0x3D24, 0x2424, 0x2444, 0x2444, 0x4484, 0x5528, 0x8A10}},
    {32918, 2, {0x100, 0x2108, 0x1110, 0x920, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2028, 0x2010}},
    {32919, 2, {0x40, 0x3C40, 0x2440, 0x2440, 0x27FE, 0x3C88, 0x2488, 0x2488, 0x2488, 0x3D08, 0x2490, 0x2450, 0x2420, 0x4450, 0x5488, 0x8B04}},
    {32920, 2, {0x8, 0x3C08, 0x2408, 0x2408, 0x25FE, 0x3C08, 0x2408, 0x2408, 0x2488, 0x3C48, 0x2448, 0x2408, 0x2408, 0x4408, 0x5428, 0x8810}},
    {32921, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2028, 0x2010}},
    {32922, 2, {0x20, 0x3C20, 0x2420, 0x2420, 0x2420, 0x3C20, 0x25FC, 0x2420, 0x2420, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {32923, 2, {0x0, 0x3C00, 0x25FC, 0x2420, 0x2420, 0x3C20, 0x2420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {32924, 2, {0x4, 0x3E04, 0x2208, 0x2210, 0x2220, 0x3E44, 0x2204, 0x2208, 0x2210, 0x3E22, 0x2242, 0x2204, 0x2208, 0x4210, 0x4A20, 0x84C0}},
    {32925, 2, {0x0, 0x3DFC, 0x2420, 0x2420, 0x2420, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32926, 2, {0x40, 0x3C24, 0x2524, 0x2504, 0x2504, 0x3C88, 0x2488, 0x2488, 0x2450, 0x3C50, 0x2420, 0x2420, 0x2450, 0x4488, 0x5504, 0x8A02}},
    {32927, 2, {0x0, 0x3CFC, 0x2400, 0x2400, 0x2400, 0x3DFE, 0x2440, 0x2440, 0x2480, 0x3CFC, 0x2404, 0x2404, 0x2404, 0x4404, 0x5428, 0x8810}},
    {32928, 2, {0x0, 0x3DF8, 0x2410, 0x2420, 0x2440, 0x3C80, 0x25FE, 0x2492, 0x2492, 0x3C92, 0x2512, 0x2522, 0x2622, 0x4442, 0x5494, 0x8908}},
    {32929, 2, {0x0, 0x3CF8, 0x2488, 0x2488, 0x2488, 0x3D06, 0x2600, 0x25FC, 0x2484, 0x3C84, 0x2448, 0x2450, 0x2420, 0x4450, 0x5488, 0x8B06}},
    {32930, 2, {0x20, 0x3C20, 0x2420, 0x25FE, 0x2420, 0x3C20, 0x2420, 0x25FC, 0x2484, 0x3C88, 0x2448, 0x2450, 0x2420, 0x4450, 0x5588, 0x8E06}},
    {32931, 2, {0x20, 0x3C20, 0x2450, 0x2450, 0x2488, 0x3D44, 0x2622, 0x2420, 0x2400, 0x3DFC, 0x2404, 0x2408, 0x2408, 0x4410, 0x5410, 0x8820}},
    {32932, 2, {0x20, 0x3C20, 0x2420, 0x2420, 0x25FC, 0x3C20, 0x2420, 0x2420, 0x27FE, 0x3C20, 0x2450, 0x2450, 0x2488, 0x4488, 0x5504, 0x8A02}},
    {32933, 2, {0x0, 0x3DFC, 0x2524, 0x2524, 0x2524, 0x3D24, 0x2524, 0x25FC, 0x2504, 0x3D00, 0x2500, 0x2500, 0x2502, 0x4502, 0x54FE, 0x8800}},
    {32934, 2, {0x10, 0x3C50, 0x2450, 0x2450, 0x3488, 0x2C88, 0x2504, 0x26FA, 0x2C48, 0x3448, 0x2448, 0x2448, 0x2488, 0x4488, 0x5528, 0x8A10}},
    {32935, 2, {0x0, 0x3C00, 0x25FE, 0x2410, 0x2410, 0x3C20, 0x2420, 0x2468, 0x24A4, 0x3D22, 0x2622, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32936, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2420, 0x3C20, 0x2420, 0x25FC, 0x2420, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32937, 2, {0x100, 0x80, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1000, 0x17F8, 0x1408, 0x17F8, 0x1408, 0x27F8, 0x2408, 0x4408, 0x8428, 0x410}},
    {32938, 2, {0x40, 0x3C20, 0x2420, 0x25FE, 0x2440, 0x3C40, 0x2440, 0x247C, 0x2444, 0x3C44, 0x2444, 0x2444, 0x2484, 0x4484, 0x5528, 0x8A10}},
    {32939, 2, {0x40, 0x7840, 0x4840, 0x4BFC, 0x4840, 0x7A48, 0x4A48, 0x4A48, 0x4A48, 0x7BF8, 0x4848, 0x4840, 0x4842, 0x4842, 0x483E, 0x9800}},
    {32940, 2, {0x40, 0x3C50, 0x2448, 0x2448, 0x2440, 0x3DFE, 0x2450, 0x2450, 0x2450, 0x3C50, 0x2490, 0x2490, 0x2492, 0x4512, 0x550E, 0x8A00}},
    {32941, 2, {0x20, 0x3C20, 0x2420, 0x25FE, 0x2522, 0x3D22, 0x2522, 0x2522, 0x2552, 0x3D4A, 0x258A, 0x2502, 0x2502, 0x4502, 0x550A, 0x8904}},
    {32942, 2, {0x40, 0x3C20, 0x2400, 0x25FC, 0x2400, 0x3C00, 0x24F0, 0x2490, 0x2490, 0x3C90, 0x2490, 0x2492, 0x2492, 0x4512, 0x550E, 0x8A00}},
    {32943, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32944, 2, {0x20, 0x3C28, 0x2424, 0x2424, 0x2420, 0x3DFE, 0x2420, 0x2420, 0x2420, 0x3C50, 0x2450, 0x2450, 0x2488, 0x4488, 0x5504, 0x8A02}},
    {32945, 2, {0x40, 0x7840, 0x4840, 0x4840, 0x4BFE, 0x7840, 0x4880, 0x4890, 0x4890, 0x7920, 0x4920, 0x4A48, 0x4A44, 0x4C84, 0x49FE, 0x9882}},
    {32946, 2, {0x200, 0x100, 0xFFFE, 0x800, 0x1010, 0x3FF8, 0x8, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32947, 2, {0x80, 0x3C80, 0x2480, 0x24FC, 0x2554, 0x3E54, 0x2454, 0x2494, 0x2494, 0x3D24, 0x2624, 0x2444, 0x2444, 0x4484, 0x5528, 0x8810}},
    {32948, 2, {0x1818, 0x660, 0x180, 0x660, 0x3A18, 0x200, 0xFFFE, 0x400, 0xFF8, 0x1808, 0x2FF8, 0x4808, 0x8FF8, 0x808, 0x828, 0x810}},
    {32949, 2, {0x4, 0x3C0E, 0x24F0, 0x2480, 0x2480, 0x3C80, 0x24FE, 0x2488, 0x2488, 0x3C88, 0x2488, 0x2488, 0x2488, 0x4508, 0x5508, 0x8A08}},
    {32950, 2, {0x10, 0x3D10, 0x2510, 0x2512, 0x2512, 0x3D14, 0x25D8, 0x2510, 0x2510, 0x3D10, 0x2510, 0x2512, 0x2552, 0x4592, 0x550E, 0x8800}},
    {32951, 2, {0x40, 0x3C40, 0x2440, 0x247C, 0x2484, 0x3C88, 0x2520, 0x2420, 0x2420, 0x3C50, 0x2450, 0x2450, 0x2488, 0x4488, 0x5504, 0x8A02}},
    {32952, 2, {0x10, 0x3C90, 0x2490, 0x2488, 0x2508, 0x3D04, 0x2624, 0x2422, 0x2420, 0x3DFE, 0x2420, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32953, 2, {0x48, 0x3C48, 0x2484, 0x2484, 0x2502, 0x3E00, 0x25FE, 0x2440, 0x2480, 0x3CFC, 0x2404, 0x2404, 0x2404, 0x4404, 0x5428, 0x8810}},
    {32954, 2, {0x20, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x3C20, 0x25FC, 0x2524, 0x2524, 0x3D24, 0x2524, 0x2534, 0x2528, 0x4420, 0x5420, 0x8820}},
    {32955, 2, {0x100, 0x1F8, 0x100, 0x100, 0x3FFC, 0x2004, 0x4008, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32956, 2, {0x88, 0x7888, 0x4888, 0x4888, 0x4BFE, 0x7888, 0x4888, 0x4888, 0x4FFE, 0x7888, 0x4888, 0x4888, 0x4888, 0x4908, 0x4908, 0x9A08}},
    {32957, 2, {0x20, 0x3C20, 0x2420, 0x2420, 0x2420, 0x3DFE, 0x2420, 0x2420, 0x2420, 0x3C50, 0x2450, 0x2450, 0x2488, 0x44C8, 0x5524, 0x8A02}},
    {32958, 2, {0x400, 0x25F8, 0x2488, 0x2450, 0x2420, 0x24D8, 0x2706, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {32959, 2, {0x20, 0x3C20, 0x2420, 0x2420, 0x25FC, 0x3D24, 0x2524, 0x2524, 0x2524, 0x3DFC, 0x2524, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32960, 2, {0x80, 0x3C84, 0x2484, 0x2488, 0x2490, 0x3CA0, 0x2480, 0x27FE, 0x24A0, 0x3C90, 0x2490, 0x2488, 0x2484, 0x44A2, 0x54C0, 0x8880}},
    {32961, 2, {0x80, 0x7880, 0x4880, 0x4880, 0x4BF0, 0x7890, 0x4890, 0x4A98, 0x4A94, 0x7C92, 0x4892, 0x4890, 0x4910, 0x4910, 0x4A50, 0x9C20}},
    {32962, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2524, 0x3D24, 0x2524, 0x25FC, 0x2524, 0x3D24, 0x2524, 0x25FC, 0x2524, 0x4420, 0x5420, 0x8820}},
    {32963, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32964, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {32965, 2, {0x20, 0x3C20, 0x2520, 0x2520, 0x25FC, 0x3D20, 0x2620, 0x2420, 0x25FE, 0x3C20, 0x2450, 0x2450, 0x2488, 0x4488, 0x5504, 0x8A02}},
    {32966, 2, {0x0, 0x3C00, 0x24FC, 0x2484, 0x2484, 0x3C84, 0x24FC, 0x2484, 0x2484, 0x3C84, 0x24FC, 0x2484, 0x2400, 0x4400, 0x55FE, 0x8800}},
    {32967, 2, {0x50, 0x3C50, 0x2450, 0x25FC, 0x2454, 0x3C54, 0x25FC, 0x2550, 0x2550, 0x3DFE, 0x2452, 0x2452, 0x245A, 0x4494, 0x5490, 0x8910}},
    {32968, 2, {0x50, 0x3C48, 0x2448, 0x2440, 0x25FE, 0x3C40, 0x2440, 0x247C, 0x24A4, 0x3CA4, 0x24A8, 0x2528, 0x2510, 0x4628, 0x5444, 0x8882}},
    {32969, 2, {0x10, 0x3C10, 0x2420, 0x24FC, 0x2484, 0x3C84, 0x2484, 0x2484, 0x24FC, 0x3C84, 0x2484, 0x2484, 0x2484, 0x4484, 0x54FC, 0x8884}},
    {32970, 2, {0x80, 0x3C80, 0x2500, 0x25FC, 0x3604, 0x2C04, 0x25E4, 0x2524, 0x2D24, 0x3524, 0x25E4, 0x2524, 0x2404, 0x4404, 0x5428, 0x8810}},
    {32971, 2, {0x20, 0x3C20, 0x2420, 0x2420, 0x243E, 0x3C20, 0x2420, 0x2420, 0x25FC, 0x3D04, 0x2504, 0x2504, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {32972, 2, {0x480, 0x498, 0x7CE0, 0x484, 0x1C84, 0xE47C, 0x4000, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {32973, 2, {0x8, 0x781C, 0x49F0, 0x4950, 0x4950, 0x7950, 0x4950, 0x4950, 0x4950, 0x7948, 0x4948, 0x4968, 0x4954, 0x4A74, 0x4A52, 0x9C00}},
    {32974, 2, {0x20, 0x3C20, 0x2420, 0x2440, 0x2448, 0x3C84, 0x25FE, 0x2482, 0x2400, 0x3CFC, 0x2484, 0x2484, 0x2484, 0x4484, 0x54FC, 0x8884}},
    {32975, 2, {0x20, 0x3C20, 0x242E, 0x25F0, 0x2520, 0x3D20, 0x2520, 0x25FE, 0x2422, 0x3C62, 0x24A2, 0x252A, 0x2624, 0x4420, 0x5420, 0x8820}},
    {32976, 2, {0x20, 0x3C20, 0x2524, 0x2524, 0x3524, 0x2D24, 0x25FC, 0x2420, 0x2C20, 0x3524, 0x2524, 0x2524, 0x2524, 0x4524, 0x55FC, 0x8804}},
    {32977, 2, {0x0, 0x3C00, 0x24FC, 0x2484, 0x2484, 0x3C84, 0x2484, 0x2484, 0x24FC, 0x3C84, 0x2400, 0x2448, 0x2444, 0x4484, 0x5482, 0x8902}},
    {32978, 2, {0x0, 0x3CFC, 0x2484, 0x2484, 0x2484, 0x3CFC, 0x2480, 0x24A0, 0x24A2, 0x3CA4, 0x24B8, 0x24A0, 0x24A2, 0x4522, 0x551E, 0x8A00}},
    {32979, 2, {0x0, 0x3DFC, 0x2420, 0x2420, 0x2524, 0x3CA4, 0x24A8, 0x2420, 0x27FE, 0x3C20, 0x2420, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32980, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C, 0x100, 0x3FF8, 0x2288, 0x2448, 0x2928, 0x2108, 0x2288, 0x2448, 0x2018}},
    {32981, 2, {0x44, 0x3C44, 0x2444, 0x2484, 0x24BE, 0x3D84, 0x2684, 0x24A4, 0x2494, 0x3C94, 0x2484, 0x2484, 0x2484, 0x4484, 0x5494, 0x8888}},
    {32982, 2, {0x20, 0x3C20, 0x2524, 0x24A4, 0x24A8, 0x3C20, 0x25FC, 0x2420, 0x2420, 0x3C20, 0x27FE, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32983, 2, {0x20, 0x3C20, 0x2450, 0x2488, 0x2504, 0x3E12, 0x2420, 0x2440, 0x2588, 0x3C10, 0x2420, 0x2444, 0x2588, 0x4410, 0x5460, 0x8B80}},
    {32984, 2, {0x20, 0x3C10, 0x2410, 0x25FE, 0x2420, 0x3C20, 0x2444, 0x2484, 0x25F8, 0x3C10, 0x2420, 0x2444, 0x2482, 0x45FE, 0x5482, 0x8800}},
    {32985, 2, {0x80, 0x3C80, 0x2480, 0x24FE, 0x2540, 0x3D40, 0x2640, 0x247C, 0x2440, 0x3C40, 0x2440, 0x247E, 0x2440, 0x4440, 0x5440, 0x8840}},
    {32986, 2, {0x0, 0x7BFC, 0x4820, 0x4820, 0x4840, 0x7840, 0x48D0, 0x4948, 0x4A44, 0x7C44, 0x4840, 0x4840, 0x4840, 0x4800, 0x4FFE, 0x9800}},
    {32987, 2, {0x0, 0x3DFC, 0x2524, 0x2524, 0x2524, 0x3DFC, 0x2524, 0x2524, 0x2524, 0x3DFC, 0x2524, 0x2420, 0x2420, 0x4420, 0x5420, 0x8820}},
    {32988, 2, {0x20, 0x7820, 0x4920, 0x4920, 0x49FC, 0x7920, 0x4A20, 0x4820, 0x4820, 0x79FC, 0x4820, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x9800}},
    {32989, 2, {0x4, 0x3C1E, 0x25F0, 0x2510, 0x2510, 0x3D10, 0x2510, 0x25FE, 0x2510, 0x3D10, 0x2510, 0x2508, 0x250A, 0x454A, 0x55A6, 0x8912}},
    {32990, 2, {0x80, 0x7880, 0x49FC, 0x4904, 0x4A04, 0x7DF4, 0x4914, 0x4914, 0x4914, 0x79F4, 0x4904, 0x4928, 0x4912, 0x4902, 0x48FE, 0x9800}},
    {32991, 2, {0x0, 0x3CFC, 0x2484, 0x24A4, 0x2494, 0x3C84, 0x2484, 0x27FE, 0x2504, 0x3D44, 0x2524, 0x2504, 0x25FE, 0x4404, 0x5428, 0x8810}},
    {32992, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2420, 0x3C20, 0x2420, 0x27FE, 0x2420, 0x3C20, 0x2440, 0x2448, 0x2484, 0x45FE, 0x5482, 0x8800}},
    {32993, 2, {0x800, 0x87C, 0x844, 0x844, 0x7F44, 0x87C, 0x844, 0x3E44, 0x2244, 0x227C, 0x2244, 0x2244, 0x3E84, 0x2284, 0x114, 0x208}},
    {32994, 2, {0x0, 0x3C00, 0x25FE, 0x2408, 0x2408, 0x3DE8, 0x2528, 0x2528, 0x2528, 0x3D28, 0x25E8, 0x2528, 0x2408, 0x4408, 0x5428, 0x8810}},
    {32995, 2, {0x80, 0x7880, 0x48FE, 0x4900, 0x4A20, 0x7920, 0x492C, 0x4974, 0x4BA4, 0x7924, 0x4934, 0x4928, 0x4922, 0x4902, 0x48FE, 0x9800}},
    {32996, 2, {0x100, 0x2208, 0x2448, 0x2788, 0x2108, 0x2248, 0x27C8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2448, 0x27CA, 0x444A, 0x444A, 0x84C6}},
    {32997, 2, {0x7FFC, 0x104, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {32998, 2, {0x20, 0x3C20, 0x2420, 0x25FC, 0x2524, 0x3D24, 0x2524, 0x2524, 0x2524, 0x3FFE, 0x2420, 0x2450, 0x2450, 0x4488, 0x5504, 0x8A02}},
    {32999, 2, {0x90, 0x7888, 0x4888, 0x4880, 0x4BFE, 0x78A0, 0x48A0, 0x48A4, 0x48A4, 0x78A8, 0x4928, 0x4932, 0x4922, 0x4A62, 0x4A9E, 0x9C00}},
    {33000, 2, {0x40, 0x7840, 0x4840, 0x4FFE, 0x4880, 0x7920, 0x4920, 0x4A20, 0x4BFC, 0x7820, 0x4928, 0x4924, 0x4A22, 0x4C22, 0x48A0, 0x9840}},
    {33001, 2, {0x40, 0x3C40, 0x2440, 0x247C, 0x2440, 0x3C40, 0x2440, 0x27FE, 0x2440, 0x3C40, 0x2450, 0x2448, 0x2444, 0x4440, 0x5440, 0x8840}},
    {33002, 2, {0x20, 0x3C20, 0x243C, 0x2420, 0x2420, 0x3CFC, 0x2484, 0x2484, 0x2484, 0x3CFC, 0x2484, 0x2480, 0x2480, 0x4500, 0x5500, 0x8A00}},
    {33003, 2, {0x0, 0x79FC, 0x4808, 0x4810, 0x4830, 0x7848, 0x4884, 0x4B02, 0x4800, 0x79FC, 0x4820, 0x4820, 0x4820, 0x4820, 0x4BFE, 0x9800}},
    {33004, 2, {0x1000, 0x1000, 0xFEFC, 0x2244, 0x6428, 0x1810, 0x2428, 0xC2C6, 0x100, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2288, 0x2448, 0x2018}},
    {33005, 2, {0x0, 0x3DFC, 0x2504, 0x2524, 0x2524, 0x3D24, 0x25FC, 0x2524, 0x2524, 0x3D54, 0x254C, 0x258C, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {33006, 2, {0x80, 0x7880, 0x48FC, 0x4908, 0x4A90, 0x7860, 0x4998, 0x4E26, 0x4820, 0x79FC, 0x4820, 0x4A20, 0x4BFE, 0x4820, 0x4820, 0x9820}},
    {33007, 2, {0x20, 0x3C20, 0x25FC, 0x2450, 0x2488, 0x3D04, 0x26FA, 0x2400, 0x27FE, 0x3C40, 0x2480, 0x25FC, 0x2404, 0x4404, 0x5428, 0x8810}},
    {33008, 2, {0x40, 0x7840, 0x4FFC, 0x4840, 0x4BF8, 0x7848, 0x4BF8, 0x4A40, 0x4BFC, 0x7844, 0x4854, 0x48A8, 0x48A0, 0x4910, 0x4A08, 0x9C06}},
    {33009, 2, {0x40, 0x7844, 0x4A44, 0x4948, 0x4950, 0x7840, 0x4FFE, 0x4890, 0x4890, 0x7890, 0x4890, 0x4892, 0x4912, 0x4912, 0x4A0E, 0x9C00}},
    {33010, 2, {0x40, 0x7820, 0x4BFE, 0x4840, 0x4890, 0x7890, 0x4924, 0x4BE4, 0x4844, 0x7848, 0x4888, 0x4910, 0x4A30, 0x4848, 0x4884, 0x9B04}},
    {33011, 2, {0x40, 0x3C40, 0x2478, 0x2488, 0x2550, 0x3C20, 0x2450, 0x2488, 0x2506, 0x3CF8, 0x2488, 0x2488, 0x2488, 0x4488, 0x54F8, 0x8888}},
    {33012, 2, {0x0, 0x7BFC, 0x4A04, 0x4A04, 0x4AF4, 0x7A04, 0x4A04, 0x4AF4, 0x4A94, 0x7A94, 0x4A94, 0x4AF4, 0x4A04, 0x4A04, 0x4A14, 0x9A08}},
    {33013, 2, {0x0, 0x3DFE, 0x2420, 0x2440, 0x2488, 0x3D04, 0x25FE, 0x2422, 0x2420, 0x3C20, 0x25FE, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {33014, 2, {0x40, 0x7820, 0x4820, 0x4BFE, 0x4800, 0x7888, 0x4904, 0x4A02, 0x4888, 0x7888, 0x4850, 0x4850, 0x4820, 0x4850, 0x4888, 0x9B06}},
    {33015, 2, {0x1000, 0x1FFC, 0x2004, 0x5454, 0x9394, 0x1454, 0x1FF4, 0x8, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {33016, 2, {0x100, 0x7900, 0x4900, 0x49FE, 0x4A02, 0x7A02, 0x4D12, 0x48A2, 0x4A4A, 0x7AAA, 0x4B1A, 0x4A0A, 0x4BFA, 0x4802, 0x4814, 0x9808}},
    {33017, 2, {0x0, 0x3DFC, 0x2420, 0x2420, 0x2440, 0x3DFC, 0x2554, 0x2554, 0x2554, 0x3D54, 0x2554, 0x2554, 0x2554, 0x4544, 0x5514, 0x8908}},
    {33018, 2, {0x40, 0x7820, 0x4820, 0x4BFE, 0x4A02, 0x7C44, 0x4840, 0x4BFE, 0x4888, 0x7888, 0x4908, 0x48D0, 0x4820, 0x4850, 0x4888, 0x9B04}},
    {33019, 2, {0x80, 0x7880, 0x493C, 0x4A00, 0x4C80, 0x7880, 0x497E, 0x4B08, 0x4D08, 0x7908, 0x4908, 0x4908, 0x4908, 0x4908, 0x4928, 0x9910}},
    {33020, 2, {0x84, 0x3C48, 0x2400, 0x24FC, 0x2448, 0x3C48, 0x2448, 0x2448, 0x25FE, 0x3C48, 0x2448, 0x2448, 0x2448, 0x4488, 0x5488, 0x8908}},
    {33021, 2, {0x1040, 0x2444, 0x4248, 0xFF70, 0x140, 0x42, 0x7E42, 0x423E, 0x4200, 0x7E44, 0x4248, 0x4270, 0x7E40, 0x4242, 0x4A42, 0x443E}},
    {33022, 2, {0x820, 0x828, 0x7F24, 0x820, 0xFFFE, 0x20, 0x824, 0x7F24, 0x4924, 0x5528, 0x6B28, 0x4910, 0x5512, 0x632A, 0x4546, 0x4282}},
    {33023, 2, {0x20, 0x3C20, 0x25FC, 0x2420, 0x2420, 0x3C20, 0x27FE, 0x2400, 0x2420, 0x3C20, 0x25FC, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {33024, 2, {0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3518, 0xC206, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33025, 2, {0x50, 0x3C50, 0x2450, 0x2552, 0x34D4, 0x2C58, 0x2450, 0x2458, 0x2CD4, 0x3552, 0x2450, 0x2450, 0x2492, 0x4492, 0x5512, 0x8A0E}},
    {33026, 2, {0x100, 0x3D04, 0x2538, 0x25C0, 0x2502, 0x3D02, 0x24FE, 0x2400, 0x25FC, 0x3D04, 0x2504, 0x25FC, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {33027, 2, {0x80, 0x3C80, 0x24F8, 0x2508, 0x2610, 0x3DFC, 0x2524, 0x2524, 0x2524, 0x3DFC, 0x2500, 0x2500, 0x2502, 0x4502, 0x54FE, 0x8800}},
    {33028, 2, {0x40, 0x7840, 0x4840, 0x4BFE, 0x4880, 0x7890, 0x4890, 0x4912, 0x4952, 0x7954, 0x4A90, 0x4A28, 0x4C28, 0x4844, 0x4884, 0x9902}},
    {33029, 2, {0x200, 0x1FF0, 0x410, 0x850, 0x1020, 0x7EFC, 0x1224, 0x2A54, 0x4488, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33030, 2, {0x40, 0x3C40, 0x24FC, 0x2504, 0x2608, 0x3DFE, 0x2500, 0x257C, 0x2544, 0x3D44, 0x2554, 0x2548, 0x2542, 0x4542, 0x553E, 0x8A00}},
    {33031, 2, {0x40, 0x7840, 0x49FC, 0x4844, 0x4884, 0x7884, 0x4928, 0x4A10, 0x4908, 0x7BDE, 0x494A, 0x494A, 0x494A, 0x4A52, 0x4B5A, 0x9CA4}},
    {33032, 2, {0x4, 0x781E, 0x49E0, 0x4900, 0x4906, 0x7978, 0x4950, 0x4950, 0x4952, 0x7954, 0x4948, 0x4948, 0x4944, 0x4A54, 0x4A62, 0x9C40}},
    {33033, 2, {0x40, 0x7820, 0x4810, 0x49E0, 0x4822, 0x7834, 0x4BB8, 0x48B0, 0x48A8, 0x78A8, 0x4928, 0x4924, 0x4A24, 0x4C22, 0x48A0, 0x9840}},
    {33034, 2, {0x2108, 0x1110, 0x920, 0x1290, 0x2448, 0x820, 0x3FF8, 0xC826, 0x820, 0xFE0, 0x820, 0x820, 0xFE0, 0x820, 0x8A0, 0x840}},
    {33035, 2, {0x0, 0x1FF0, 0x410, 0x450, 0x820, 0x7EFC, 0x1244, 0x2A94, 0x4508, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33036, 2, {0x80, 0x3C80, 0x24FE, 0x2510, 0x2610, 0x3C10, 0x24FC, 0x2490, 0x2490, 0x3C90, 0x27FE, 0x2410, 0x2410, 0x4410, 0x5410, 0x8810}},
    {33037, 2, {0x20, 0x7820, 0x4850, 0x4888, 0x4904, 0x7A02, 0x48F8, 0x4800, 0x4800, 0x7BFE, 0x4820, 0x4840, 0x4888, 0x4904, 0x4BFE, 0x9902}},
    {33038, 2, {0x4, 0x7988, 0x4850, 0x4820, 0x4850, 0x7888, 0x4924, 0x4820, 0x4BFE, 0x7820, 0x48A8, 0x48A4, 0x4922, 0x4A22, 0x48A0, 0x9840}},
    {33039, 2, {0x20, 0x7810, 0x4810, 0x49FE, 0x4900, 0x7910, 0x4910, 0x4910, 0x497C, 0x7910, 0x4910, 0x4910, 0x4910, 0x4A10, 0x4AFE, 0x9C00}},
    {33040, 2, {0x40, 0x7820, 0x4BFE, 0x4904, 0x4888, 0x7850, 0x4820, 0x48D8, 0x4B06, 0x7888, 0x4888, 0x4888, 0x4888, 0x4908, 0x4908, 0x9A08}},
    {33041, 2, {0x40, 0x7820, 0x4820, 0x4FFE, 0x4800, 0x7800, 0x4890, 0x4A54, 0x4A24, 0x7A24, 0x4A54, 0x4A94, 0x4A04, 0x4A04, 0x4BFC, 0x9804}},
    {33042, 2, {0x20, 0x3C20, 0x2524, 0x24A4, 0x24A8, 0x3C20, 0x2420, 0x27FE, 0x2470, 0x3CA8, 0x24A8, 0x2524, 0x2524, 0x4622, 0x5420, 0x8820}},
    {33043, 2, {0x20, 0x7820, 0x4820, 0x4BFE, 0x4A42, 0x7C44, 0x48A0, 0x48A2, 0x49A4, 0x7A98, 0x4C90, 0x4888, 0x4884, 0x48A2, 0x48C0, 0x9880}},
    {33044, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2548, 0x4104, 0x3FF8, 0x2288, 0x2448, 0x2928, 0x2108, 0x2288, 0x2448, 0x2008, 0x2018}},
    {33045, 2, {0x80, 0x3C80, 0x24F8, 0x2508, 0x2610, 0x3DFC, 0x2524, 0x2524, 0x2524, 0x3DFC, 0x2450, 0x2450, 0x2490, 0x4492, 0x5512, 0x8A0E}},
    {33046, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x7BFE, 0x4A02, 0x4C04, 0x49F8, 0x7810, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x48A0, 0x9840}},
    {33047, 2, {0x80, 0x3C80, 0x24FC, 0x2554, 0x2654, 0x3CA4, 0x2524, 0x2644, 0x2494, 0x3D08, 0x25FC, 0x2504, 0x2504, 0x4504, 0x55FC, 0x8904}},
    {33048, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4C04, 0x79F8, 0x4800, 0x4800, 0x4BFE, 0x7890, 0x4890, 0x4890, 0x4912, 0x4912, 0x4A0E, 0x9C00}},
    {33049, 2, {0x28, 0x3C24, 0x2424, 0x2420, 0x27FE, 0x3C20, 0x2520, 0x24B2, 0x24B4, 0x3C68, 0x24A8, 0x2524, 0x2622, 0x4420, 0x54A0, 0x8840}},
    {33050, 2, {0x100, 0x7900, 0x491E, 0x4B92, 0x4912, 0x7912, 0x4912, 0x4FD2, 0x4912, 0x7912, 0x4A1A, 0x4A94, 0x4FD0, 0x4A50, 0x4810, 0x9810}},
    {33051, 2, {0x0, 0x7BFC, 0x4800, 0x4924, 0x4A48, 0x7C90, 0x4A48, 0x4924, 0x4800, 0x7BFC, 0x4840, 0x4840, 0x4840, 0x4840, 0x4FFE, 0x9800}},
    {33052, 2, {0x0, 0x3DFE, 0x2410, 0x2420, 0x24FC, 0x3C84, 0x2484, 0x24FC, 0x2484, 0x3C84, 0x24FC, 0x2484, 0x2484, 0x4484, 0x54FC, 0x8884}},
    {33053, 2, {0x40, 0x7820, 0x4BFE, 0x4800, 0x49F8, 0x7908, 0x4908, 0x49F8, 0x4800, 0x7BFC, 0x4808, 0x4830, 0x4820, 0x4820, 0x48A0, 0x9840}},
    {33054, 2, {0x20, 0x3C20, 0x24A8, 0x24A8, 0x24A8, 0x3D74, 0x2622, 0x2420, 0x2420, 0x3DFC, 0x2420, 0x2420, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {33055, 2, {0x4, 0x3C1E, 0x25E0, 0x2422, 0x2512, 0x3C94, 0x2480, 0x2408, 0x25FE, 0x3C08, 0x2488, 0x2448, 0x2448, 0x4408, 0x5428, 0x8810}},
    {33056, 2, {0x4, 0x780E, 0x4BB8, 0x4888, 0x4888, 0x7928, 0x492E, 0x4BA8, 0x48A8, 0x7AA8, 0x4AA8, 0x493E, 0x4900, 0x4A80, 0x4C7E, 0x9800}},
    {33057, 2, {0x0, 0x7806, 0x4BB8, 0x4888, 0x4888, 0x7908, 0x493E, 0x4B88, 0x4888, 0x7A88, 0x4A88, 0x493E, 0x4900, 0x4A80, 0x4C7E, 0x9800}},
    {33058, 2, {0x80, 0x3C80, 0x24FE, 0x2500, 0x26FC, 0x3C84, 0x24A4, 0x2494, 0x27FE, 0x3C84, 0x2524, 0x2514, 0x25FE, 0x4404, 0x5428, 0x8810}},
    {33059, 2, {0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x2430, 0x260E, 0x2FF8, 0x2808, 0x2FF8, 0x4808, 0x4FF8, 0x8808, 0x828, 0x810}},
    {33060, 2, {0x0, 0x79FE, 0x4900, 0x4900, 0x497C, 0x7900, 0x4900, 0x49FE, 0x4950, 0x7952, 0x4954, 0x4948, 0x4948, 0x4A44, 0x4A52, 0x9C60}},
    {33061, 2, {0x20, 0x7820, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4924, 0x4924, 0x4924, 0x7AAA, 0x4820, 0x4850, 0x4850, 0x4888, 0x4904, 0x9A02}},
    {33062, 2, {0x10, 0x7814, 0x4812, 0x4810, 0x4BFE, 0x7810, 0x4890, 0x4850, 0x4848, 0x7808, 0x4888, 0x4AA8, 0x4A9A, 0x4AAA, 0x4C66, 0x9802}},
    {33063, 2, {0x40, 0x3C40, 0x2488, 0x2504, 0x37FE, 0x2C02, 0x2488, 0x2544, 0x2E42, 0x34F8, 0x2588, 0x2650, 0x2420, 0x4450, 0x5588, 0x8E06}},
    {33064, 2, {0x20, 0x3C20, 0x27FE, 0x2420, 0x2420, 0x3DFC, 0x2524, 0x2524, 0x25FC, 0x3C20, 0x2470, 0x24A8, 0x2524, 0x4622, 0x5420, 0x8820}},
    {33065, 2, {0x1080, 0x1080, 0x11F8, 0x2108, 0x2A90, 0x6C60, 0xA998, 0x2E06, 0x29F8, 0x2908, 0x29F8, 0x2908, 0x29F8, 0x2108, 0x2128, 0x2110}},
    {33066, 2, {0xC2, 0x3C34, 0x2418, 0x2464, 0x2582, 0x3C20, 0x25FE, 0x2450, 0x2490, 0x3DFE, 0x2692, 0x2492, 0x249A, 0x4494, 0x5410, 0x8810}},
    {33067, 2, {0x10, 0x7890, 0x4888, 0x4908, 0x4A04, 0x7DFA, 0x4908, 0x4908, 0x4908, 0x79F8, 0x4890, 0x4890, 0x4890, 0x4912, 0x4A12, 0x9C0E}},
    {33068, 2, {0x8, 0x783C, 0x4BC0, 0x4804, 0x4A44, 0x7928, 0x4800, 0x4BF8, 0x4810, 0x7820, 0x4FFE, 0x4820, 0x4820, 0x4820, 0x48A0, 0x9840}},
    {33069, 2, {0x0, 0x79FC, 0x4904, 0x4904, 0x4904, 0x79FC, 0x4800, 0x4800, 0x4BFE, 0x7820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x9800}},
    {33070, 2, {0x8, 0x783C, 0x4BC0, 0x4844, 0x4A24, 0x7928, 0x4900, 0x4840, 0x4BFE, 0x7888, 0x4908, 0x4B90, 0x4860, 0x4850, 0x4988, 0x9E04}},
    {33071, 2, {0x48, 0x7844, 0x4BFE, 0x4840, 0x4840, 0x7BFC, 0x4A44, 0x4A44, 0x4BFC, 0x7A44, 0x4A44, 0x4BFC, 0x4A44, 0x4A44, 0x4A54, 0x9A08}},
    {33072, 2, {0x0, 0x3DFE, 0x2400, 0x2400, 0x24FC, 0x3C84, 0x2484, 0x2484, 0x24FC, 0x3C00, 0x2484, 0x2444, 0x2448, 0x4400, 0x55FE, 0x8800}},
    {33073, 2, {0x104, 0x7884, 0x4888, 0x4810, 0x49FC, 0x7904, 0x4904, 0x4904, 0x49FC, 0x7850, 0x4850, 0x4890, 0x4892, 0x4912, 0x4A0E, 0x9C00}},
    {33074, 2, {0x0, 0x7BFC, 0x4A04, 0x4A04, 0x4BFC, 0x7A00, 0x4A20, 0x4A22, 0x4BB2, 0x7AB4, 0x4AA8, 0x4AA8, 0x4D24, 0x4D24, 0x4AA2, 0x9840}},
    {33075, 2, {0x42, 0x7A22, 0x4924, 0x4904, 0x4808, 0x7800, 0x4A0A, 0x4A8A, 0x4A52, 0x7A22, 0x4A52, 0x4A8A, 0x4B0A, 0x4A02, 0x4BFE, 0x9802}},
    {33076, 2, {0x0, 0x3DFE, 0x2410, 0x2420, 0x2468, 0x3CA4, 0x2522, 0x2420, 0x2400, 0x3CFC, 0x2484, 0x2484, 0x2484, 0x4484, 0x54FC, 0x8884}},
    {33077, 2, {0x0, 0x79FC, 0x4904, 0x4904, 0x49FC, 0x7800, 0x4800, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x4820, 0x4850, 0x4888, 0x4904, 0x9A02}},
    {33078, 2, {0x0, 0x79FC, 0x4904, 0x4904, 0x4904, 0x79FC, 0x4820, 0x4820, 0x4BFE, 0x7A22, 0x4A52, 0x4A8A, 0x4B0A, 0x4A02, 0x4A0A, 0x9A04}},
    {33079, 2, {0x42, 0x78E2, 0x4B82, 0x488A, 0x488A, 0x788A, 0x4FEA, 0x488A, 0x498A, 0x79CA, 0x4AAA, 0x4A82, 0x4C82, 0x4882, 0x488A, 0x9884}},
    {33080, 2, {0x20, 0x7820, 0x4850, 0x4850, 0x4888, 0x7904, 0x4AFA, 0x4800, 0x4844, 0x7824, 0x4924, 0x48A8, 0x4888, 0x4810, 0x4BFE, 0x9800}},
    {33081, 2, {0x0, 0x3CFC, 0x2480, 0x24F8, 0x2480, 0x3CF8, 0x2480, 0x27FE, 0x2540, 0x3D24, 0x2528, 0x2510, 0x2508, 0x4544, 0x5582, 0x8900}},
    {33082, 2, {0x40, 0x7820, 0x4BFE, 0x4800, 0x4888, 0x7888, 0x4954, 0x4A22, 0x4800, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x4820, 0x9820}},
    {33083, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7824, 0x4BFE, 0x4824, 0x49FC, 0x7820, 0x4920, 0x493E, 0x4920, 0x4AA0, 0x4A7E, 0x9C00}},
    {33084, 2, {0x0, 0x7BFE, 0x4820, 0x4820, 0x4BFE, 0x7A22, 0x4A22, 0x4B32, 0x4AAA, 0x7AAA, 0x4B76, 0x4A66, 0x4A22, 0x4A22, 0x4A2A, 0x9A04}},
    {33085, 2, {0x50, 0x3C48, 0x2480, 0x24FE, 0x2590, 0x3E90, 0x24FC, 0x2490, 0x2490, 0x3CFC, 0x2490, 0x2490, 0x2490, 0x44FE, 0x5480, 0x8880}},
    {33086, 2, {0x40, 0x7880, 0x4BFC, 0x4A24, 0x4A24, 0x7BFC, 0x4A24, 0x4A44, 0x4BFC, 0x7890, 0x4910, 0x4FFE, 0x4810, 0x4810, 0x4810, 0x9810}},
    {33087, 2, {0x20, 0x3C20, 0x27FE, 0x2420, 0x2420, 0x3DFC, 0x2420, 0x2420, 0x27FE, 0x3C40, 0x24A4, 0x25A8, 0x2690, 0x4488, 0x54C6, 0x8880}},
    {33088, 2, {0x20, 0x3C20, 0x2450, 0x2488, 0x2504, 0x3EFA, 0x2400, 0x25FC, 0x2554, 0x3D54, 0x25FC, 0x2554, 0x2554, 0x4554, 0x5504, 0x890C}},
    {33089, 2, {0x84, 0x788C, 0x4908, 0x4FBE, 0x4908, 0x7908, 0x4908, 0x4908, 0x4FBE, 0x7908, 0x4908, 0x4908, 0x4908, 0x4A08, 0xAA08, 0x9408}},
    {33090, 2, {0x0, 0x3DFC, 0x2524, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x25FC, 0x2420, 0x3FFE, 0x2470, 0x24A8, 0x2524, 0x4622, 0x5420, 0x8820}},
    {33091, 2, {0x20, 0x3D24, 0x24A8, 0x2420, 0x25FC, 0x3C40, 0x27FE, 0x2488, 0x2504, 0x3EFA, 0x2488, 0x2488, 0x24A8, 0x4492, 0x5482, 0x887E}},
    {33092, 2, {0x8, 0x3C3C, 0x25E0, 0x2420, 0x27FE, 0x3CA8, 0x24A8, 0x24A8, 0x27FE, 0x3CA8, 0x24A8, 0x27FE, 0x2420, 0x4420, 0x55FC, 0x8800}},
    {33093, 2, {0x20, 0x3CA2, 0x24A2, 0x2524, 0x2450, 0x3C88, 0x2704, 0x2422, 0x2420, 0x3CA4, 0x24A4, 0x2528, 0x2450, 0x4488, 0x5504, 0x8A02}},
    {33094, 2, {0x50, 0x7850, 0x4850, 0x49FC, 0x4954, 0x7954, 0x4954, 0x49FC, 0x4954, 0x7954, 0x4954, 0x4BFE, 0x4800, 0x4888, 0x4904, 0x9A02}},
    {33095, 2, {0x1C, 0x79E0, 0x4820, 0x4820, 0x4BFE, 0x78A8, 0x4924, 0x4A42, 0x4840, 0x7BFE, 0x4888, 0x4908, 0x48D0, 0x4830, 0x4848, 0x9984}},
    {33096, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x4914, 0x9908}},
    {33097, 2, {0x0, 0x7840, 0x4B9C, 0x4A04, 0x4A04, 0x7B9C, 0x4A04, 0x4A04, 0x4BFC, 0x7890, 0x4890, 0x4890, 0x4890, 0x4912, 0x4A12, 0x9C0E}},
    {33098, 2, {0x88, 0x7888, 0x4888, 0x4BFE, 0x4888, 0x7888, 0x4FFE, 0x4800, 0x49FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x9904}},
    {33099, 2, {0x80, 0x7840, 0x4FFE, 0x4920, 0x4920, 0x7A3C, 0x4A44, 0x4E64, 0x4A94, 0x7B48, 0x4A48, 0x4A30, 0x4A20, 0x4A50, 0x4A88, 0x9B06}},
    {33100, 2, {0x40, 0x7880, 0x4BFC, 0x4910, 0x4A48, 0x7C46, 0x4BF8, 0x4A48, 0x4A48, 0x7BF8, 0x4A48, 0x4A48, 0x4BF8, 0x4842, 0x4842, 0x983E}},
    {33101, 2, {0x20, 0x7820, 0x4850, 0x4888, 0x4944, 0x7A22, 0x49F8, 0x4808, 0x4850, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x4A8A, 0x4C78, 0x9800}},
    {33102, 2, {0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {33103, 2, {0x0, 0x7BDE, 0x4842, 0x494A, 0x4884, 0x794A, 0x4A52, 0x4C20, 0x4800, 0x7BDE, 0x4852, 0x4952, 0x4894, 0x4948, 0x4A54, 0x9C22}},
    {33104, 2, {0x80, 0x3FFE, 0x2208, 0x2208, 0x25FE, 0x2C88, 0x3448, 0x2458, 0x2080, 0x2FFC, 0x2944, 0x2AA4, 0x4884, 0x4944, 0x8A24, 0x80C}},
    {33105, 2, {0x20, 0x7810, 0x4BFE, 0x4A00, 0x4A44, 0x7A44, 0x4A84, 0x4ABE, 0x4B84, 0x7AA4, 0x4A94, 0x4A94, 0x4C84, 0x4C84, 0x4894, 0x9888}},
    {33106, 2, {0x0, 0x3DFC, 0x2504, 0x2504, 0x25FC, 0x3D10, 0x2510, 0x25FE, 0x2510, 0x3D10, 0x257C, 0x2544, 0x2544, 0x4544, 0x557C, 0x8A44}},
    {33107, 2, {0x90, 0x7890, 0x4890, 0x4F9E, 0x4890, 0x7890, 0x4890, 0x4B9C, 0x4890, 0x7890, 0x4890, 0x4F9E, 0x4890, 0x4890, 0x4890, 0x9890}},
    {33108, 2, {0x40, 0x7820, 0x4820, 0x4BFE, 0x4A02, 0x7C94, 0x4908, 0x4A04, 0x4800, 0x79FC, 0x4820, 0x4820, 0x4820, 0x4820, 0x4FFE, 0x9800}},
    {33109, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4D04, 0x7900, 0x49DE, 0x4A52, 0x4A52, 0x7B52, 0x4C9A, 0x4894, 0x4910, 0x4912, 0x4A12, 0x9C0E}},
    {33110, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x49FC, 0x7820, 0x4870, 0x48A8, 0x4924, 0x4E22, 0x4820, 0x9820}},
    {33111, 2, {0x20, 0x3C20, 0x25FC, 0x2524, 0x2524, 0x3DFC, 0x2524, 0x2524, 0x25FC, 0x3C00, 0x27FE, 0x2488, 0x2488, 0x4488, 0x5508, 0x8A08}},
    {33112, 2, {0x0, 0x7BFE, 0x4A02, 0x4A02, 0x4BFE, 0x7A22, 0x4A22, 0x4AFA, 0x4A22, 0x7A32, 0x4A2A, 0x4BFE, 0x4A02, 0x4A02, 0x4BFE, 0x9A02}},
    {33113, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4800, 0x79FC, 0x4800, 0x4800, 0x4BFE, 0x7820, 0x4928, 0x4924, 0x4A22, 0x4C22, 0x48A0, 0x9840}},
    {33114, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4C04, 0x7800, 0x4BFE, 0x4820, 0x4820, 0x7920, 0x493C, 0x4920, 0x4AA0, 0x4A60, 0x4C3E, 0x9800}},
    {33115, 2, {0x0, 0x3DFE, 0x2502, 0x2502, 0x25FE, 0x3D00, 0x25FC, 0x2520, 0x2544, 0x3DFE, 0x2512, 0x2510, 0x25FC, 0x4510, 0x5610, 0x89FE}},
    {33116, 2, {0x88, 0x3C88, 0x27FE, 0x2488, 0x2488, 0x3CF8, 0x2488, 0x2488, 0x24F8, 0x3C20, 0x27FE, 0x2470, 0x24A8, 0x4524, 0x5622, 0x8820}},
    {33117, 2, {0x0, 0x3DFC, 0x2420, 0x2440, 0x25FC, 0x3D54, 0x2554, 0x2554, 0x2544, 0x3D2C, 0x2420, 0x27FE, 0x2450, 0x4488, 0x5504, 0x8A02}},
    {33118, 2, {0x40, 0x3C7C, 0x2484, 0x24F8, 0x2408, 0x3DFE, 0x2440, 0x24A4, 0x2538, 0x3C58, 0x2494, 0x2534, 0x2454, 0x4492, 0x5550, 0x8820}},
    {33119, 2, {0x40, 0x3C20, 0x25FE, 0x2502, 0x2604, 0x3DFC, 0x2440, 0x2488, 0x25FC, 0x3C24, 0x2420, 0x25FC, 0x2420, 0x4420, 0x57FE, 0x8800}},
    {33120, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7840, 0x4BFE, 0x4888, 0x4904, 0x7AFA, 0x4820, 0x49FC, 0x4820, 0x4850, 0x4888, 0x9B04}},
    {33121, 2, {0x0, 0x79FC, 0x4904, 0x4904, 0x49E4, 0x7924, 0x4924, 0x4BFE, 0x4A02, 0x7AFA, 0x4A8A, 0x4A8A, 0x4AFA, 0x4A02, 0x4A0A, 0x9A04}},
    {33122, 2, {0x1FC, 0x7924, 0x4924, 0x49FC, 0x4924, 0x7924, 0x49FC, 0x4820, 0x4820, 0x7BFE, 0x4A22, 0x4A2A, 0x4AFA, 0x4A0A, 0x4A02, 0x9A06}},
    {33123, 2, {0x40, 0x7820, 0x49FC, 0x4800, 0x4888, 0x7850, 0x4BFE, 0x4A22, 0x4C24, 0x79FC, 0x4924, 0x4924, 0x4934, 0x4928, 0x4820, 0x9820}},
    {33124, 2, {0x20, 0x3C10, 0x25FE, 0x2400, 0x2484, 0x3C48, 0x25FE, 0x2400, 0x24FC, 0x3C84, 0x2484, 0x24FC, 0x2484, 0x4484, 0x54FC, 0x8884}},
    {33125, 2, {0x0, 0x7BF8, 0x4A08, 0x4BF8, 0x4A08, 0x7BF8, 0x4840, 0x4A40, 0x4BFC, 0x7C40, 0x4840, 0x4BF8, 0x4840, 0x4840, 0x4FFE, 0x9800}},
    {33126, 2, {0x92, 0x7924, 0x4A48, 0x4924, 0x4892, 0x7840, 0x4880, 0x4BFE, 0x4A02, 0x7A8A, 0x4A52, 0x4A22, 0x4A52, 0x4A8A, 0x4BFE, 0x9A02}},
    {33127, 2, {0x40, 0x78A0, 0x4910, 0x4A08, 0x4DF6, 0x7800, 0x4BC4, 0x4A54, 0x4A54, 0x7BD4, 0x4A54, 0x4A54, 0x4BD4, 0x4A44, 0x4A54, 0x9AC8}},
    {33128, 2, {0x20, 0x3D24, 0x2524, 0x2524, 0x25FC, 0x3C00, 0x27FE, 0x2420, 0x2440, 0x3DFC, 0x2554, 0x2554, 0x2554, 0x4554, 0x5554, 0x890C}},
    {33129, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x7BFE, 0x4A8A, 0x4A52, 0x4AFA, 0x7A22, 0x4A22, 0x4AFA, 0x4A22, 0x4A22, 0x4A2A, 0x9A04}},
    {33130, 2, {0x0, 0x7BFE, 0x4A22, 0x4820, 0x4BFE, 0x7820, 0x49FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x9820}},
    {33131, 2, {0x8, 0x3C3C, 0x25E0, 0x2420, 0x27FE, 0x3C20, 0x25FC, 0x2524, 0x25FC, 0x3D24, 0x25FC, 0x2420, 0x25FC, 0x4420, 0x57FE, 0x8800}},
    {33132, 2, {0x0, 0x3CFC, 0x2448, 0x2430, 0x25FE, 0x3C52, 0x2494, 0x2510, 0x2630, 0x3C20, 0x25FE, 0x2470, 0x24A8, 0x4524, 0x5622, 0x8820}},
    {33133, 2, {0x0, 0x3DDC, 0x2554, 0x2554, 0x25DC, 0x3C00, 0x25FC, 0x2400, 0x25FE, 0x3C80, 0x24FC, 0x2404, 0x2404, 0x4404, 0x5428, 0x8810}},
    {33134, 2, {0x0, 0x79FC, 0x4924, 0x4924, 0x49FC, 0x7924, 0x4924, 0x49FC, 0x4800, 0x7840, 0x4824, 0x4AA2, 0x4A8A, 0x4C88, 0x4878, 0x9800}},
    {33135, 2, {0x4, 0x3C1E, 0x25F0, 0x2510, 0x2510, 0x3DFE, 0x2510, 0x257C, 0x2544, 0x3D7C, 0x2544, 0x257C, 0x2544, 0x4544, 0x557C, 0x8A44}},
    {33136, 2, {0x0, 0x7BFE, 0x4850, 0x4850, 0x4BFE, 0x7A52, 0x4A52, 0x4BFE, 0x4800, 0x7840, 0x4FFE, 0x4888, 0x4990, 0x4860, 0x48D8, 0x9B04}},
    {33137, 2, {0x10, 0x7810, 0x487C, 0x4B14, 0x49FE, 0x7914, 0x4A7C, 0x4A10, 0x4F7C, 0x7910, 0x49FE, 0x4D10, 0x4A10, 0x4B00, 0x4CFE, 0x9800}},
    {33138, 2, {0x0, 0x3DFC, 0x2524, 0x2524, 0x25FC, 0x3D24, 0x2524, 0x25FC, 0x2400, 0x3FFE, 0x2520, 0x2522, 0x2514, 0x4548, 0x5584, 0x8902}},
    {33139, 2, {0x0, 0x7A9E, 0x4A52, 0x4A52, 0x4D52, 0x7912, 0x4A92, 0x4A52, 0x4C32, 0x7BD2, 0x4A5A, 0x4A54, 0x4A50, 0x4A50, 0x4BD0, 0x9A50}},
    {33140, 2, {0x20, 0x3C20, 0x24A0, 0x252C, 0x2524, 0x3D24, 0x25AC, 0x2524, 0x2524, 0x3DFC, 0x2524, 0x2450, 0x2450, 0x4488, 0x5504, 0x8A02}},
    {33141, 2, {0x0, 0x7BDE, 0x4A42, 0x4A42, 0x4A42, 0x7BDE, 0x4A00, 0x4A3E, 0x4BD2, 0x7A12, 0x4A14, 0x4BD4, 0x4A08, 0x4A14, 0x4A24, 0x9A42}},
    {33142, 2, {0x20, 0x3DDC, 0x2514, 0x2514, 0x2514, 0x3DD4, 0x2526, 0x2500, 0x25DC, 0x3D14, 0x2514, 0x25D4, 0x2708, 0x4508, 0x5514, 0x8922}},
    {33143, 2, {0x0, 0x3DFC, 0x2400, 0x24F8, 0x2488, 0x3C88, 0x24F8, 0x2400, 0x25FC, 0x3D24, 0x2524, 0x25FC, 0x2524, 0x4524, 0x55FC, 0x8904}},
    {33144, 2, {0x0, 0x3CFC, 0x2484, 0x24FC, 0x2484, 0x3CFC, 0x2400, 0x25FE, 0x2480, 0x3CFE, 0x252A, 0x264A, 0x2492, 0x4522, 0x5454, 0x8888}},
    {33145, 2, {0x40, 0x3C40, 0x24FE, 0x2480, 0x25FC, 0x3E84, 0x24FC, 0x2484, 0x24FC, 0x3C40, 0x247C, 0x24C4, 0x2528, 0x4410, 0x5468, 0x8986}},
    {33146, 2, {0x40, 0x7880, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4820, 0x4832, 0x7BB4, 0x48A8, 0x4928, 0x4924, 0x4A24, 0x4CA2, 0x9840}},
    {33147, 2, {0x18, 0x7814, 0x4FFE, 0x4810, 0x4BD0, 0x7810, 0x4BF0, 0x4810, 0x4BE8, 0x7A28, 0x4AA8, 0x4AA8, 0x4AAA, 0x494A, 0x4A26, 0x9C22}},
    {33148, 2, {0x0, 0x3DFE, 0x2420, 0x2440, 0x25FC, 0x3D54, 0x2554, 0x2574, 0x2554, 0x3D54, 0x2574, 0x2554, 0x2554, 0x4554, 0x55FC, 0x8904}},
    {33149, 2, {0x0, 0x3CF8, 0x2488, 0x2488, 0x24F8, 0x3C88, 0x2488, 0x24F8, 0x2400, 0x3DFC, 0x2554, 0x2554, 0x2554, 0x4554, 0x57FE, 0x8800}},
    {33150, 2, {0x248, 0x7948, 0x4950, 0x4FFC, 0x4840, 0x7FFE, 0x4910, 0x4A08, 0x4DF6, 0x7810, 0x4910, 0x49FC, 0x4804, 0x4BE4, 0x4814, 0x9808}},
    {33151, 2, {0x0, 0x7A7C, 0x4944, 0x4944, 0x487C, 0x7844, 0x4F44, 0x497C, 0x4950, 0x794A, 0x4944, 0x4954, 0x4962, 0x4942, 0x4A80, 0x9C7E}},
    {33152, 2, {0x40, 0x7820, 0x4BFC, 0x4908, 0x4890, 0x7BFE, 0x4A02, 0x4C44, 0x4820, 0x7BFC, 0x4880, 0x48F8, 0x4888, 0x4908, 0x4928, 0x9A10}},
    {33153, 2, {0x88, 0x3C50, 0x2400, 0x25FE, 0x2450, 0x3DFC, 0x2454, 0x27FE, 0x2454, 0x3DFC, 0x2450, 0x24D8, 0x2554, 0x4652, 0x5450, 0x8850}},
    {33154, 2, {0x2040, 0xFEFE, 0x2100, 0x3C1C, 0x24E0, 0x4494, 0x54A8, 0x88C6, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {33155, 2, {0x0, 0x3CF8, 0x24A8, 0x24A8, 0x24D8, 0x3C88, 0x24F8, 0x2400, 0x2400, 0x3DFC, 0x2554, 0x2554, 0x2554, 0x4554, 0x57FE, 0x8800}},
    {33156, 2, {0xA0, 0x3D2C, 0x2524, 0x2524, 0x25AC, 0x3D24, 0x2524, 0x25FC, 0x2420, 0x3DFC, 0x2488, 0x2450, 0x2420, 0x4450, 0x5488, 0x8B06}},
    {33157, 2, {0x20, 0x3C10, 0x25FE, 0x2510, 0x257C, 0x3D14, 0x25FE, 0x2514, 0x257C, 0x3D10, 0x257C, 0x2544, 0x2544, 0x4544, 0x557C, 0x8A44}},
    {33158, 2, {0x20, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x4840, 0x4888, 0x79F0, 0x4824, 0x4BFE, 0x4822, 0x4924, 0x4AA2, 0x9840}},
    {33159, 2, {0x10, 0x7A20, 0x497C, 0x4944, 0x4844, 0x787C, 0x4B40, 0x4940, 0x497C, 0x7944, 0x4944, 0x497C, 0x4944, 0x4A80, 0x4C7E, 0x9800}},
    {33160, 2, {0x0, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x4800, 0x4BFE, 0x7A8A, 0x4A52, 0x4BFE, 0x4A22, 0x4A22, 0x4A2A, 0x9A04}},
    {33161, 2, {0x104, 0x3C84, 0x2488, 0x2400, 0x27FE, 0x3C00, 0x2488, 0x2504, 0x2602, 0x3DFC, 0x2554, 0x2554, 0x2554, 0x4554, 0x57FE, 0x8800}},
    {33162, 2, {0x28, 0x7824, 0x4BFE, 0x4820, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x49FC, 0x7924, 0x4808, 0x4BFE, 0x4888, 0x4848, 0x4848, 0x9818}},
    {33163, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {33164, 2, {0x20, 0x3D24, 0x24A8, 0x2554, 0x2488, 0x3D04, 0x26FA, 0x2488, 0x2488, 0x3CF8, 0x2488, 0x2488, 0x24F8, 0x4488, 0x5488, 0x8898}},
    {33165, 2, {0x20, 0x3C40, 0x25FC, 0x2504, 0x2554, 0x3D24, 0x2554, 0x2504, 0x25FC, 0x3C00, 0x2512, 0x25D4, 0x2518, 0x4552, 0x5592, 0x890E}},
    {33166, 2, {0x1E, 0x7BE0, 0x4922, 0x4894, 0x4840, 0x7888, 0x49F0, 0x4820, 0x48C4, 0x79FE, 0x4822, 0x4820, 0x4BFE, 0x4850, 0x4888, 0x9B06}},
    {33167, 2, {0x100, 0x7FFC, 0x820, 0xFE0, 0x0, 0x7FFE, 0x4822, 0x8FE4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33168, 2, {0x2040, 0xFEFE, 0x2100, 0x3C1C, 0x24E0, 0x4494, 0x54A8, 0x88C6, 0x100, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2288, 0x2448, 0x2018}},
    {33169, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4C14, 0x79E0, 0x4900, 0x4900, 0x49FC, 0x7910, 0x4910, 0x4FFE, 0x4800, 0x4890, 0x4908, 0x9A04}},
    {33170, 2, {0x0, 0x7BFE, 0x4A00, 0x4A7C, 0x4A44, 0x7A44, 0x4A7C, 0x4A00, 0x4AEE, 0x7AAA, 0x4AAA, 0x4AAA, 0x4AEE, 0x4A00, 0x4BFE, 0x9800}},
    {33171, 2, {0x100, 0x79FC, 0x4A00, 0x4DF8, 0x4908, 0x79F8, 0x4908, 0x49F8, 0x4800, 0x7FFE, 0x4900, 0x4BFC, 0x4CA4, 0x4924, 0x4A54, 0x9888}},
    {33172, 2, {0x40, 0x7820, 0x4BFE, 0x4A50, 0x4A50, 0x7BFE, 0x4A52, 0x4A52, 0x4BFE, 0x7A00, 0x4A92, 0x4AD4, 0x4A98, 0x4A92, 0x4AD2, 0x9C8E}},
    {33173, 2, {0x0, 0x77FE, 0x542A, 0x5426, 0x57FE, 0x7422, 0x57AA, 0x56AA, 0x57AA, 0x7432, 0x5596, 0x562E, 0x5446, 0x5482, 0x57FE, 0xB402}},
    {33174, 2, {0x20, 0x7A20, 0x497C, 0x48A4, 0x4818, 0x78E6, 0x4B10, 0x497E, 0x4910, 0x797C, 0x4910, 0x49FE, 0x4910, 0x4910, 0x4AFE, 0x9C00}},
    {33175, 2, {0x20, 0x7A22, 0x4BFE, 0x4890, 0x4888, 0x79FE, 0x4B10, 0x4D10, 0x49FE, 0x7910, 0x4910, 0x49FE, 0x4910, 0x4910, 0x49FE, 0x9900}},
    {33176, 2, {0x0, 0x7BFE, 0x4850, 0x49FC, 0x4954, 0x7954, 0x49FC, 0x4800, 0x49FC, 0x7800, 0x4BFE, 0x4820, 0x48A8, 0x4924, 0x4AA2, 0x9840}},
    {33177, 2, {0x0, 0x7BBE, 0x48A2, 0x48A2, 0x48BE, 0x7B88, 0x4A08, 0x4A3E, 0x4A2A, 0x7BAA, 0x48AA, 0x48BE, 0x4888, 0x488A, 0x4D7E, 0x9A02}},
    {33178, 2, {0x100, 0x1FC, 0x100, 0x3FFE, 0x2002, 0x2FFC, 0x2884, 0x2FFC, 0x2884, 0x2FFC, 0x2408, 0x27F8, 0x2408, 0x47F8, 0x4408, 0x8418}},
    {33179, 2, {0x20, 0x7924, 0x48A8, 0x4BFE, 0x4A02, 0x78F8, 0x4888, 0x4888, 0x48F8, 0x7820, 0x4820, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x9800}},
    {33180, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4820, 0x4BFE, 0x4850, 0x4888, 0x4904, 0x9A02}},
    {33181, 2, {0x20, 0x7820, 0x4BFE, 0x4870, 0x48AC, 0x7B22, 0x4850, 0x4888, 0x4924, 0x7A22, 0x4924, 0x48A8, 0x4924, 0x4A22, 0x48A0, 0x9840}},
    {33182, 2, {0x20, 0x3DFE, 0x2420, 0x25FC, 0x2524, 0x3DFC, 0x2524, 0x25FC, 0x2422, 0x3DFE, 0x2408, 0x25FE, 0x2508, 0x4488, 0x5428, 0x8810}},
    {33183, 2, {0x40, 0x7820, 0x4BFE, 0x4820, 0x4848, 0x7AF2, 0x4924, 0x4A52, 0x48F8, 0x7808, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x9820}},
    {33184, 2, {0x0, 0x7BDE, 0x4842, 0x4A52, 0x494A, 0x7A52, 0x4828, 0x48C4, 0x4B12, 0x7860, 0x4988, 0x4832, 0x49C4, 0x4818, 0x4860, 0x9B80}},
    {33185, 2, {0x20, 0x7924, 0x48A8, 0x49FC, 0x4840, 0x7BFE, 0x4888, 0x4904, 0x4BFE, 0x7D04, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x9904}},
    {33186, 2, {0x20, 0x79FC, 0x4924, 0x4BFE, 0x4924, 0x79FC, 0x4820, 0x49FC, 0x4924, 0x79FC, 0x4840, 0x4BFE, 0x4888, 0x49D0, 0x4870, 0x9B8C}},
    {33187, 2, {0x40, 0x7820, 0x4BFE, 0x4A8A, 0x4904, 0x7BFE, 0x4840, 0x4888, 0x49FC, 0x7824, 0x4820, 0x49FC, 0x4820, 0x4820, 0x4BFE, 0x9800}},
    {33188, 2, {0x0, 0x79FC, 0x4820, 0x4BFE, 0x4A22, 0x79AC, 0x4820, 0x49AC, 0x4800, 0x79FC, 0x4804, 0x49FC, 0x4804, 0x4804, 0x49FC, 0x9804}},
    {33189, 2, {0x1014, 0x1010, 0xFDFE, 0x1110, 0x7DD4, 0x3948, 0x554A, 0x9296, 0x1122, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2288, 0x2448, 0x2018}},
    {33190, 2, {0x20, 0x7924, 0x48A8, 0x4BFE, 0x48A8, 0x7924, 0x4A22, 0x4904, 0x4904, 0x79DE, 0x4A44, 0x4D54, 0x489E, 0x4904, 0x4A04, 0x9C04}},
    {33191, 2, {0x20, 0x3DFC, 0x2488, 0x2450, 0x37FE, 0x2C00, 0x25FC, 0x2524, 0x2DFC, 0x3524, 0x25FC, 0x2420, 0x25FC, 0x4420, 0x57FE, 0x8800}},
    {33192, 2, {0x82, 0x7882, 0x4FF4, 0x4888, 0x4880, 0x7BE2, 0x4802, 0x4BE4, 0x4A28, 0x7A20, 0x4BE0, 0x4A22, 0x4942, 0x4874, 0x4F88, 0x9A10}},
    {33193, 2, {0x14, 0x7BFE, 0x4810, 0x4BD0, 0x4810, 0x7BD0, 0x4810, 0x4BD0, 0x4A50, 0x7BD0, 0x4A50, 0x4BC8, 0x4A4A, 0x4BCA, 0x4986, 0x9A42}},
    {33194, 2, {0x40, 0x7820, 0x49FC, 0x4888, 0x4850, 0x7BFE, 0x4A22, 0x4C24, 0x49FC, 0x7924, 0x492C, 0x4820, 0x49FC, 0x4904, 0x49FC, 0x9904}},
    {33195, 2, {0x20, 0x3C20, 0x25FC, 0x2420, 0x2554, 0x3C88, 0x2504, 0x26FA, 0x2488, 0x3CF8, 0x2488, 0x24F8, 0x2420, 0x44A8, 0x5524, 0x8860}},
    {33196, 2, {0x38, 0x7BC0, 0x4878, 0x4BC0, 0x4878, 0x7BC4, 0x483C, 0x4800, 0x48C6, 0x7B38, 0x49CE, 0x4F38, 0x49CE, 0x4F38, 0x494A, 0x9986}},
    {33197, 2, {0x20, 0x3DFC, 0x2524, 0x25FC, 0x2420, 0x3FFE, 0x2400, 0x25FC, 0x2504, 0x3DFC, 0x2504, 0x25FC, 0x2504, 0x45FC, 0x5488, 0x8904}},
    {33198, 2, {0x20, 0x7820, 0x49FC, 0x4820, 0x4BFE, 0x7908, 0x4B9C, 0x4908, 0x4988, 0x7E3E, 0x4800, 0x4BFE, 0x4890, 0x4890, 0x4912, 0x9A0E}},
    {33199, 2, {0x10, 0x7BD2, 0x4854, 0x4948, 0x488A, 0x7904, 0x4AFA, 0x4C00, 0x49FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x4888, 0x4FFE, 0x9800}},
    {33200, 2, {0x3C, 0x7BE0, 0x4924, 0x48A8, 0x4BFE, 0x78A8, 0x4924, 0x4A02, 0x49FC, 0x7924, 0x4924, 0x49FC, 0x4924, 0x4924, 0x49FC, 0x9904}},
    {33201, 2, {0x110, 0x7894, 0x4BD2, 0x4812, 0x4A50, 0x7990, 0x4FFE, 0x4810, 0x4BD2, 0x7A52, 0x4A54, 0x4BD4, 0x4A4A, 0x4A4A, 0x4BD6, 0x9822}},
    {33202, 2, {0xA0, 0x7890, 0x49FE, 0x4910, 0x4B10, 0x7DFE, 0x4910, 0x4910, 0x49FE, 0x7910, 0x4910, 0x49FE, 0x4900, 0x4AA4, 0x4A52, 0x9C52}},
    {33203, 2, {0x88, 0x7850, 0x4BFE, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x4924, 0x48A8, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x9904}},
    {33204, 2, {0x80, 0x3C80, 0x25FE, 0x2754, 0x2554, 0x3D54, 0x27FE, 0x2554, 0x2554, 0x3D54, 0x27FE, 0x2400, 0x2554, 0x452A, 0x562A, 0x8800}},
    {33205, 2, {0x88, 0x3C88, 0x27FE, 0x2488, 0x2420, 0x3FFE, 0x2488, 0x2488, 0x2554, 0x3E22, 0x2420, 0x27FE, 0x2420, 0x4420, 0x5420, 0x8820}},
    {33206, 2, {0x0, 0x7BDE, 0x4A52, 0x4BDE, 0x4A52, 0x7BDE, 0x4A02, 0x4AFA, 0x4A22, 0x7A72, 0x4A22, 0x4A22, 0x4AFA, 0x4A02, 0x4A0A, 0x9A04}},
    {33207, 2, {0x280, 0x724E, 0x54EA, 0x55AA, 0x59AA, 0x7EEA, 0x52AC, 0x54AA, 0x55EA, 0x7F8A, 0x528A, 0x52AA, 0x54AC, 0x54E8, 0x58A8, 0xB008}},
    {33208, 2, {0x20, 0x79FE, 0x4C40, 0x4AFC, 0x4910, 0x7AFE, 0x4800, 0x4EFC, 0x4A84, 0x7AFC, 0x4A84, 0x4AFC, 0x4A84, 0x4A8C, 0x4D00, 0x98FE}},
    {33209, 2, {0x20, 0x3DFC, 0x2420, 0x2488, 0x27FE, 0x3C88, 0x25FC, 0x2504, 0x25FC, 0x3D04, 0x25FC, 0x2504, 0x25FC, 0x4488, 0x5504, 0x8A02}},
    {33210, 2, {0x80, 0x3FFE, 0x2250, 0x24FC, 0x2D90, 0x36FC, 0x2490, 0x24FC, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x4408, 0x47F8, 0x8408, 0x418}},
    {33211, 2, {0x20, 0x7BFE, 0x4800, 0x4BFC, 0x4A04, 0x7AF4, 0x4A94, 0x4BFC, 0x4800, 0x79F8, 0x4908, 0x49F8, 0x4908, 0x49F8, 0x4800, 0x9BFE}},
    {33212, 2, {0x0, 0x7A7C, 0x4944, 0x4974, 0x4854, 0x7854, 0x4EFE, 0x4A82, 0x4ABA, 0x7AAA, 0x4ABA, 0x4A82, 0x4A8A, 0x4A84, 0x4D00, 0x98FE}},
    {33213, 2, {0x80, 0x3CF8, 0x2508, 0x27FE, 0x2544, 0x3D92, 0x25FE, 0x2500, 0x257C, 0x3D00, 0x257C, 0x2500, 0x257C, 0x4544, 0x557C, 0x8A44}},
    {33214, 2, {0x20, 0x3C50, 0x2488, 0x2574, 0x2602, 0x3DFC, 0x2524, 0x25AC, 0x2524, 0x3DFC, 0x2400, 0x24F8, 0x2488, 0x44F8, 0x5488, 0x88F8}},
    {33215, 2, {0x50, 0x3DFC, 0x2554, 0x25FC, 0x2554, 0x3DFC, 0x2400, 0x25FE, 0x2500, 0x3D7C, 0x2500, 0x25FE, 0x2550, 0x4554, 0x5548, 0x8A66}},
    {33216, 2, {0x7E78, 0x4248, 0x7E48, 0x5486, 0x7E78, 0x5448, 0x7E30, 0x94CE, 0x6200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33217, 2, {0x20, 0x7BFE, 0x4A88, 0x4A50, 0x4BFE, 0x7A50, 0x4AFC, 0x4A54, 0x4BFE, 0x7A54, 0x4AFC, 0x4A50, 0x4AD8, 0x4D54, 0x4E52, 0x9850}},
    {33218, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33219, 2, {0x40, 0x7820, 0x4FFE, 0x4828, 0x4A24, 0x7A7E, 0x4CC8, 0x4F48, 0x497E, 0x7A48, 0x4CC8, 0x4F7E, 0x4948, 0x4A48, 0x4C7E, 0x9840}},
    {33220, 2, {0x40, 0x787C, 0x4840, 0x4BFE, 0x4A42, 0x7A78, 0x4BC4, 0x4A3C, 0x4A00, 0x7BFE, 0x4A40, 0x4BA4, 0x4A58, 0x4DB4, 0x4C52, 0x99B0}},
    {33221, 2, {0x0, 0x3DFC, 0x2554, 0x2554, 0x25FC, 0x3C80, 0x25FE, 0x2642, 0x25F2, 0x3D52, 0x2552, 0x25F2, 0x244A, 0x47FA, 0x540A, 0x8804}},
    {33222, 2, {0x40, 0x7820, 0x49FC, 0x4888, 0x4850, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4820, 0x4A94, 0x4A8A, 0x9C7A}},
    {33223, 2, {0xA0, 0x7890, 0x49FE, 0x4B20, 0x49FC, 0x7920, 0x49FC, 0x4920, 0x49FE, 0x7900, 0x4BDE, 0x4A52, 0x4A52, 0x4A72, 0x4A02, 0x9A06}},
    {33224, 2, {0x88, 0x7BFE, 0x4888, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4880, 0x79FE, 0x4A22, 0x4D52, 0x4902, 0x49FA, 0x480A, 0x9804}},
    {33225, 2, {0x20, 0x3C20, 0x2450, 0x2488, 0x2504, 0x3EFA, 0x2400, 0x2400, 0x25DC, 0x3D54, 0x2554, 0x25DC, 0x2488, 0x4488, 0x5554, 0x8A22}},
    {33226, 2, {0xF8, 0x3C88, 0x2488, 0x24F8, 0x2400, 0x3DDC, 0x2554, 0x2554, 0x25DC, 0x3C20, 0x27FE, 0x2470, 0x24A8, 0x4524, 0x5622, 0x8820}},
    {33227, 2, {0x7E78, 0x4248, 0x7E48, 0x5486, 0x7E78, 0x5448, 0x7E30, 0x94CE, 0x6300, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2288, 0x2448, 0x2018}},
    {33228, 2, {0x88, 0x7888, 0x4FC8, 0x48BE, 0x4888, 0x7BC8, 0x481C, 0x4BD4, 0x4A54, 0x7A54, 0x4BD4, 0x4A48, 0x4988, 0x48D4, 0x4F14, 0x9A22}},
    {33229, 2, {0x80, 0x7040, 0x57FC, 0x5110, 0x50A4, 0x7F58, 0x5554, 0x5552, 0x5B58, 0x7000, 0x5208, 0x53F8, 0x5208, 0x53F8, 0x5208, 0xB408}},
    {33230, 2, {0x3DE, 0x7842, 0x4A52, 0x494A, 0x4A52, 0x7820, 0x4BFE, 0x4888, 0x4888, 0x7954, 0x4A22, 0x4820, 0x4BFE, 0x4820, 0x4820, 0x9820}},
    {33231, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x49FC, 0x7948, 0x4A50, 0x49FC, 0x4B04, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4888, 0x9904}},
    {33232, 2, {0x1C, 0x3DE0, 0x2420, 0x27FE, 0x2420, 0x3DFC, 0x25AC, 0x2574, 0x25FC, 0x3C20, 0x25FC, 0x2420, 0x27FE, 0x4400, 0x5554, 0x8A2A}},
    {33233, 2, {0x0, 0x79FC, 0x4820, 0x4BFE, 0x4A22, 0x79AC, 0x4820, 0x49AC, 0x4800, 0x7BFE, 0x4820, 0x49FC, 0x4954, 0x4954, 0x4954, 0x990C}},
    {33234, 2, {0x88, 0x7BFE, 0x48A8, 0x4890, 0x49FE, 0x7B20, 0x49FC, 0x4920, 0x49FC, 0x7920, 0x49FE, 0x4900, 0x4BFC, 0x4888, 0x4870, 0x9B8E}},
    {33235, 2, {0x90, 0x7090, 0x57FE, 0x5094, 0x500A, 0x77FE, 0x5408, 0x55E8, 0x554A, 0x75EA, 0x552A, 0x55EA, 0x554C, 0x55EA, 0x5416, 0xB862}},
    {33236, 2, {0x3DE, 0x7A92, 0x4BD2, 0x4A54, 0x4BC8, 0x7A94, 0x4BE2, 0x4800, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x9888}},
    {33237, 2, {0x40, 0x7820, 0x4BFE, 0x4A50, 0x4BFE, 0x7A52, 0x4BFE, 0x4A00, 0x4A94, 0x7AD8, 0x4A90, 0x4AD2, 0x4A8E, 0x4C00, 0x4954, 0x9A2A}},
    {33238, 2, {0x2FEC, 0x4824, 0x4BA4, 0x682C, 0x4BA4, 0x6AAC, 0x4BA4, 0xFFFE, 0x1010, 0x3FF8, 0xD016, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33239, 2, {0x20, 0x7BFE, 0x4A8A, 0x49FC, 0x4888, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x4890, 0x4894, 0x4912, 0x9A0E}},
    {33240, 2, {0x124, 0x7A48, 0x4924, 0x4800, 0x4BFC, 0x7A94, 0x4A64, 0x4A94, 0x4BFC, 0x7A48, 0x4B68, 0x4A48, 0x4B6A, 0x4A4A, 0x4A46, 0x9B62}},
    {33241, 2, {0x110, 0x7110, 0x57FC, 0x5110, 0x51F0, 0x7208, 0x52EA, 0x56AC, 0x52A8, 0x72AA, 0x56EA, 0x5A06, 0x5000, 0x52A4, 0x5252, 0xB452}},
    {33242, 2, {0x40, 0x787C, 0x4840, 0x4BFC, 0x4A44, 0x7BF0, 0x4A44, 0x4AFC, 0x4AA8, 0x7AF8, 0x4AA8, 0x4AF8, 0x4A00, 0x4DFC, 0x4954, 0x9BFE}},
    {33243, 2, {0x1FC, 0x7820, 0x4BFE, 0x4A22, 0x4DAC, 0x7820, 0x49AC, 0x4890, 0x49FE, 0x7B20, 0x4DFC, 0x4920, 0x49FC, 0x4920, 0x49FE, 0x9900}},
    {33244, 2, {0x108, 0x7528, 0x57BE, 0x5948, 0x57BE, 0x7318, 0x55AA, 0x5946, 0x5000, 0x73F8, 0x5208, 0x5248, 0x5248, 0x50B0, 0x5108, 0xB604}},
    {33245, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x540, 0x77DC, 0x5554, 0x77D4, 0x511C, 0x7FF6, 0x5556, 0xB922}},
    {33246, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A52, 0x4BDE, 0x4890, 0x49FE, 0x7910, 0x4BFE, 0x4D10, 0x49FE, 0x4910, 0x49FE, 0x9900}},
    {33247, 2, {0x90, 0x7090, 0x57FE, 0x5094, 0x500A, 0x71FE, 0x5508, 0x57E8, 0x514A, 0x71EA, 0x5F2A, 0x55EA, 0x554C, 0x59EA, 0x5216, 0xB462}},
    {33248, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x3FF8, 0x2288, 0x2448, 0x2108, 0x2298}},
    {33249, 2, {0x2450, 0xFF48, 0x24FE, 0x3D90, 0x8FC, 0x7E90, 0x52FC, 0x7E90, 0x24FE, 0x4280, 0xBFF8, 0x2288, 0x2448, 0x2928, 0x2288, 0x2458}},
    {33250, 2, {0x528, 0x77BE, 0x5948, 0x57BE, 0x5318, 0x75AA, 0x5946, 0x53FC, 0x5204, 0x73FC, 0x5204, 0x53FC, 0x5204, 0x53FC, 0x5108, 0xB204}},
    {33251, 2, {0x0, 0x3FFC, 0x2100, 0x2100, 0x2100, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2100, 0x2100, 0x2100, 0x3FFC, 0x0}},
    {33252, 2, {0x0, 0x7E00, 0x48FC, 0x4844, 0x4844, 0x7E44, 0x4244, 0x4228, 0x4228, 0x7E28, 0x4810, 0x4810, 0x4828, 0x4848, 0x7E84, 0x102}},
    {33253, 2, {0x20, 0x7E20, 0x4820, 0x4820, 0x4820, 0x7E20, 0x4220, 0x4220, 0x4250, 0x7E50, 0x4850, 0x4850, 0x4848, 0x4888, 0x7E84, 0x102}},
    {33254, 2, {0x0, 0xFE7E, 0x1248, 0x1248, 0x1248, 0x7E7E, 0x4242, 0x4242, 0x4242, 0x4242, 0x7E7E, 0x1248, 0x1248, 0x1248, 0xFE7E, 0x0}},
    {33255, 2, {0x28, 0x24, 0x5FFE, 0x5020, 0x57A0, 0x7520, 0x1520, 0x17A4, 0xF4A4, 0x54A8, 0x57A8, 0x5510, 0x9512, 0x27AA, 0x2046, 0x4082}},
    {33256, 2, {0x40, 0x7E40, 0x487E, 0x4880, 0x4900, 0x7E7C, 0x4244, 0x4244, 0x427C, 0x7E00, 0x48EE, 0x48AA, 0x48AA, 0x48AA, 0x7EEE, 0xAA}},
    {33257, 2, {0x0, 0xFEFE, 0x1290, 0x7EFC, 0x4284, 0x7EFC, 0x1290, 0xFEFE, 0x200, 0x7FFC, 0x440, 0x1830, 0xE44E, 0x440, 0x840, 0x1040}},
    {33258, 2, {0x100, 0x200, 0x400, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33259, 2, {0x820, 0x820, 0x1020, 0x7F20, 0x4120, 0x4120, 0x7F20, 0x4120, 0x4120, 0x7F20, 0x4120, 0x4122, 0x4122, 0x7F22, 0x411E, 0x0}},
    {33260, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33261, 2, {0x200, 0x400, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x120, 0x110, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {33262, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFC, 0x7F00, 0x1110, 0x1110, 0x2928, 0x4544, 0x100}},
    {33263, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2108, 0x1110, 0x920, 0x1110, 0x2108, 0xFFFE, 0x100, 0x100}},
    {33264, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE, 0x840, 0x1F48, 0x6250, 0x1462, 0x1842, 0xE03E}},
    {33265, 2, {0x200, 0x1FF0, 0x1490, 0x1250, 0x1130, 0x1FF0, 0x100, 0x7FFE, 0x4A22, 0x9114, 0x3FF8, 0x400, 0x7F0, 0x810, 0x1050, 0x2020}},
    {33266, 2, {0x1020, 0x2020, 0x7E7C, 0x4284, 0x7F08, 0x42FE, 0x7E80, 0x42BC, 0x7EA4, 0x10A4, 0xFEB4, 0x18A8, 0x34A2, 0x5322, 0x911E, 0x1200}},
    {33267, 2, {0x0, 0x7FFC, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {33268, 2, {0x40, 0xFE40, 0x1040, 0x2080, 0x28FE, 0x4508, 0xFE88, 0x1288, 0x1088, 0x1050, 0x7C50, 0x1020, 0x1050, 0x1E88, 0xF104, 0x4202}},
    {33269, 2, {0x40, 0xFE40, 0x1078, 0x2088, 0x2950, 0x4420, 0xFE50, 0x1288, 0x1106, 0x10F8, 0x7C88, 0x1088, 0x1088, 0x1E88, 0xF0F8, 0x4088}},
    {33270, 2, {0x20, 0xFE20, 0x1020, 0x21FE, 0x2840, 0x4440, 0xFEBC, 0x1284, 0x1188, 0x1288, 0x7CFE, 0x1088, 0x1088, 0x1E88, 0xF0A8, 0x4090}},
    {33271, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x40, 0x7F44, 0x1044, 0x2244, 0x7F28, 0x92A, 0x3E12, 0x82A, 0xF46, 0xF082}},
    {33272, 2, {0x0, 0xFEFE, 0x1010, 0x2020, 0x2844, 0x4482, 0xFEFE, 0x1212, 0x1010, 0x1010, 0x7CFE, 0x1010, 0x1010, 0x1E10, 0xF1FE, 0x4000}},
    {33273, 2, {0x28, 0xF824, 0x2024, 0x4020, 0x53FE, 0x8A20, 0xFA24, 0x2A24, 0x23A4, 0xFAA8, 0x22A8, 0x2290, 0x3A92, 0xE5AA, 0x4446, 0x882}},
    {33274, 2, {0x100, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x7FFE, 0x4002, 0x9FF4, 0x820, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {33275, 2, {0x20, 0xF820, 0x23FE, 0x4020, 0x51FC, 0x8840, 0xFBFE, 0x2888, 0x2134, 0xFAE2, 0x2020, 0x21FC, 0x3870, 0xE0AC, 0x4322, 0x20}},
    {33276, 2, {0x0, 0x400, 0x3878, 0x2008, 0x2008, 0x2008, 0x2008, 0x3C78, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x0}},
    {33277, 2, {0x400, 0x400, 0xFF0, 0x1010, 0x2020, 0x4040, 0x8C00, 0x3078, 0x2008, 0x2008, 0x3C78, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33278, 2, {0x100, 0x500, 0x3978, 0x2108, 0x2108, 0x3D78, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33279, 2, {0x10, 0x1F8, 0x3F00, 0x100, 0xFFFE, 0x100, 0xD00, 0x3178, 0x2108, 0x2108, 0x3D78, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {33280, 2, {0xFC, 0x7F00, 0x208, 0x2108, 0x1110, 0x1020, 0xC00, 0x3078, 0x2008, 0x2008, 0x3C78, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33281, 2, {0x600, 0x38F8, 0x2008, 0x2008, 0x3EF8, 0x2008, 0x2008, 0x3FF8, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {33282, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x440, 0x820, 0x3418, 0xD876, 0x1010, 0x1C70, 0x1010, 0x1FF0, 0x1010}},
    {33283, 2, {0x600, 0x38F8, 0x2008, 0x2008, 0x3EF8, 0x2008, 0x2008, 0x3FF8, 0x1000, 0x1FF8, 0x2000, 0x3FFC, 0x4, 0x4924, 0x8494, 0x8}},
    {33284, 2, {0x600, 0x38F8, 0x2008, 0x2008, 0x3EF8, 0x2008, 0x2008, 0x3FF8, 0x1000, 0x3FFC, 0x4004, 0x8924, 0x2494, 0x2494, 0x4028, 0x10}},
    {33285, 2, {0x3E78, 0x2008, 0x3C78, 0x2008, 0x3FF8, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x200, 0x7FF8, 0x408, 0x1828, 0x6010}},
    {33286, 2, {0x1100, 0x611C, 0x47C4, 0x4204, 0x751C, 0x47C4, 0x4104, 0x77DC, 0x4104, 0x4104, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008, 0x4004}},
    {33287, 2, {0xA00, 0x3238, 0x2388, 0x2208, 0x3A38, 0x2388, 0x2088, 0x3AB8, 0x2288, 0x2288, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008, 0x4004}},
    {33288, 2, {0x0, 0x2FEC, 0x4824, 0x4BA4, 0x682C, 0x4BA4, 0x4AA4, 0x6AAC, 0x4BA4, 0x4824, 0xFFFE, 0x0, 0x820, 0x1010, 0x2008, 0x4004}},
    {33289, 2, {0xA00, 0x33B8, 0x2208, 0x3BB8, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x820, 0x1110, 0x2FE8, 0xC106, 0x1FF0, 0x100, 0x100, 0x100}},
    {33290, 2, {0x820, 0xFFFE, 0x920, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x200, 0x1C78, 0x1008, 0x1E78, 0x1008, 0x1FF8}},
    {33291, 2, {0x2FEC, 0x4824, 0x4BA4, 0x682C, 0x4BA4, 0x6AAC, 0x4BA4, 0xFFFE, 0x8002, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE}},
    {33292, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33293, 2, {0x100, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33294, 2, {0x100, 0x280, 0x440, 0x920, 0x3118, 0xCFE6, 0x100, 0x100, 0x3FF8, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33295, 2, {0x404, 0xE04, 0x7844, 0x844, 0x844, 0xFF44, 0x844, 0x844, 0x3E4C, 0x2254, 0x2264, 0x2244, 0x2204, 0x3E04, 0x2204, 0x4}},
    {33296, 2, {0x804, 0x1C1E, 0x71F0, 0x1110, 0x1110, 0xFD10, 0x1110, 0x11FE, 0x7D10, 0x4510, 0x4510, 0x4508, 0x450A, 0x7D4A, 0x4586, 0x102}},
    {33297, 2, {0x820, 0x1C20, 0x7020, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x7D24, 0x4524, 0x47FE, 0x4504, 0x4504, 0x7D04, 0x4514, 0x108}},
    {33298, 2, {0x1000, 0x11F8, 0x2808, 0x4450, 0x8220, 0x7C10, 0x13FE, 0x1022, 0xFE24, 0x1020, 0x1020, 0x7C20, 0x4420, 0x4420, 0x7CA0, 0x4440}},
    {33299, 2, {0x800, 0x1DFC, 0x7104, 0x11FC, 0x1104, 0xFDFC, 0x1080, 0x1080, 0x7DFE, 0x464A, 0x444A, 0x4492, 0x4522, 0x7E42, 0x4494, 0x108}},
    {33300, 2, {0x800, 0x1DFC, 0x7020, 0x1020, 0x11FE, 0xFC50, 0x1050, 0x1088, 0x7D24, 0x4622, 0x4424, 0x44B2, 0x44AA, 0x7D2A, 0x44A0, 0x40}},
    {33301, 2, {0x820, 0x1CA2, 0x70A2, 0x1124, 0x1050, 0xFC88, 0x1304, 0x1022, 0x7C20, 0x44A4, 0x44A4, 0x4528, 0x4450, 0x7C88, 0x4504, 0x202}},
    {33302, 2, {0x1028, 0x1024, 0x2BFE, 0x4420, 0x8220, 0x7DFC, 0x1124, 0x1124, 0xFFFC, 0x1124, 0x1124, 0x7DFC, 0x4524, 0x4524, 0x7D24, 0x450C}},
    {33303, 2, {0x1028, 0x1024, 0x2BFE, 0x5420, 0x9220, 0x7DFC, 0x1124, 0x1124, 0xFDFC, 0x124, 0x7D24, 0x45FC, 0x4524, 0x4524, 0x7D24, 0x450C}},
    {33304, 2, {0x1020, 0x1010, 0x29FE, 0x4502, 0x8200, 0x7CF8, 0x1088, 0x1088, 0xFEF8, 0x1080, 0x1080, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484}},
    {33305, 2, {0xF8, 0x3F00, 0x100, 0xFFFE, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0xC0C, 0x7070, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x4444, 0x7C7C}},
    {33306, 2, {0x880, 0x1CF8, 0x7108, 0x13FE, 0x1144, 0xFD92, 0x11FE, 0x1100, 0x7D7C, 0x4500, 0x457C, 0x4500, 0x457C, 0x7D44, 0x457C, 0x244}},
    {33307, 2, {0x1010, 0x1010, 0x1010, 0x1E10, 0x12FE, 0x2210, 0x2290, 0x5290, 0x8A90, 0x490, 0x4FE, 0x810, 0x810, 0x1010, 0x2010, 0x4010}},
    {33308, 2, {0x7C, 0x3F80, 0x108, 0x1088, 0x890, 0x7FFE, 0x4812, 0x810, 0x1EFC, 0x2210, 0x5290, 0xC90, 0x8FE, 0x1010, 0x2010, 0x4010}},
    {33309, 2, {0x200, 0x3FF0, 0x410, 0xFFFE, 0x1110, 0x2220, 0x1110, 0x7FFE, 0x4402, 0x8824, 0x1FF0, 0x100, 0x1FF0, 0x2100, 0x3FF8, 0x100}},
    {33310, 2, {0x1000, 0x3FFC, 0x4AA0, 0xAA0, 0x7FFC, 0xAA0, 0xAA0, 0x7FFC, 0x1010, 0x1010, 0x3EFC, 0x4210, 0x1490, 0x8FE, 0x3010, 0xC010}},
    {33311, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1210, 0x1110, 0x1110, 0xFFFE, 0x1010, 0x1210, 0x1110, 0x1110, 0x2010, 0x2010, 0x4050, 0x8020}},
    {33312, 2, {0x800, 0x1000, 0x3EFC, 0x2224, 0x3224, 0x2A24, 0x2A24, 0xFE24, 0x2224, 0x3224, 0x2A24, 0x2A44, 0x2244, 0x4284, 0x4B28, 0x8610}},
    {33313, 2, {0x800, 0x1000, 0x3E7C, 0x2210, 0x3210, 0x2A10, 0x2A10, 0xFE10, 0x2210, 0x3210, 0x2A10, 0x2A10, 0x2210, 0x4210, 0x4AFE, 0x8400}},
    {33314, 2, {0x1020, 0x2020, 0x7C20, 0x4420, 0x6524, 0x5524, 0x5524, 0xFD24, 0x4524, 0x6524, 0x5524, 0x5524, 0x4524, 0x45FC, 0x5404, 0x8800}},
    {33315, 2, {0x1040, 0x2024, 0x7D24, 0x4504, 0x6504, 0x5488, 0x5488, 0xFC88, 0x4450, 0x6450, 0x5420, 0x5420, 0x4450, 0x4488, 0x5504, 0x8A02}},
    {33316, 2, {0x800, 0x1000, 0x3E78, 0x2248, 0x3248, 0x2A48, 0x2A48, 0xFE68, 0x2258, 0x3248, 0x2A48, 0x2A4A, 0x224A, 0x428A, 0x4A86, 0x8500}},
    {33317, 2, {0x1000, 0x2000, 0x7DFC, 0x4524, 0x6524, 0x5524, 0x5524, 0xFD24, 0x45FC, 0x6500, 0x5500, 0x5500, 0x4502, 0x4502, 0x54FE, 0x8800}},
    {33318, 2, {0x1020, 0x2020, 0x7C20, 0x4420, 0x6420, 0x55FE, 0x5420, 0xFC20, 0x4420, 0x6450, 0x5450, 0x5450, 0x4488, 0x44C8, 0x5524, 0x8A02}},
    {33319, 2, {0x810, 0x1020, 0x3E78, 0x2248, 0x3248, 0x2A48, 0x2A48, 0xFE68, 0x2258, 0x3248, 0x2A48, 0x2A4A, 0x224A, 0x428A, 0x4A86, 0x8500}},
    {33320, 2, {0x1004, 0x200E, 0x7CF0, 0x4480, 0x6480, 0x54FC, 0x54A4, 0xFCA4, 0x44A4, 0x64A8, 0x54A8, 0x5490, 0x4510, 0x4528, 0x5644, 0x8882}},
    {33321, 2, {0x1010, 0x2090, 0x7C90, 0x4488, 0x6488, 0x5524, 0x5524, 0xFE22, 0x4420, 0x6440, 0x5448, 0x5444, 0x4484, 0x45FE, 0x5482, 0x8800}},
    {33322, 2, {0x1040, 0x2020, 0x7C00, 0x45FC, 0x6400, 0x5400, 0x54F0, 0xFC90, 0x4490, 0x6490, 0x5490, 0x5492, 0x4492, 0x4512, 0x550E, 0x8A00}},
    {33323, 2, {0x1040, 0x2020, 0x7C20, 0x45FE, 0x6440, 0x5440, 0x5440, 0xFC7C, 0x4444, 0x6444, 0x5444, 0x5444, 0x4484, 0x4484, 0x5528, 0x8A10}},
    {33324, 2, {0x800, 0x1078, 0x3E48, 0x2248, 0x3248, 0x2A86, 0x2A00, 0xFEFC, 0x2244, 0x3244, 0x2A28, 0x2A28, 0x2210, 0x4228, 0x4A44, 0x8582}},
    {33325, 2, {0x1010, 0x2110, 0x7D10, 0x4512, 0x6512, 0x5514, 0x55D8, 0xFD10, 0x4510, 0x6510, 0x5510, 0x5512, 0x4552, 0x4592, 0x550E, 0x8800}},
    {33326, 2, {0x1020, 0x2010, 0x7C10, 0x44FE, 0x6482, 0x5482, 0x5482, 0xFCFE, 0x4482, 0x6480, 0x5480, 0x5480, 0x4480, 0x4500, 0x5500, 0x8A00}},
    {33327, 2, {0x1020, 0x2020, 0x7C20, 0x4420, 0x65FC, 0x5524, 0x5524, 0xFD24, 0x4524, 0x65FC, 0x5524, 0x5420, 0x4420, 0x4420, 0x5420, 0x8820}},
    {33328, 2, {0x1000, 0x21FC, 0x7D04, 0x4504, 0x6524, 0x5524, 0x5524, 0xFD24, 0x4524, 0x6554, 0x5450, 0x5450, 0x4490, 0x4492, 0x5512, 0x8A0E}},
    {33329, 2, {0x1020, 0x2020, 0x7C50, 0x4450, 0x6488, 0x5504, 0x5602, 0xFCF8, 0x4488, 0x6488, 0x54A8, 0x5490, 0x4482, 0x4482, 0x547E, 0x8800}},
    {33330, 2, {0x820, 0x1020, 0x3E50, 0x2250, 0x3288, 0x2B24, 0x2A12, 0xFE10, 0x23FC, 0x3204, 0x2A08, 0x2A88, 0x2250, 0x4220, 0x4A10, 0x8410}},
    {33331, 2, {0x1020, 0x2020, 0x7C20, 0x4420, 0x65FC, 0x5524, 0x5524, 0xFD24, 0x4524, 0x65FC, 0x5524, 0x5524, 0x4524, 0x4524, 0x55FC, 0x8904}},
    {33332, 2, {0x1080, 0x2080, 0x7C80, 0x44FE, 0x6540, 0x5540, 0x5640, 0xFC7C, 0x4440, 0x6440, 0x5440, 0x547E, 0x4440, 0x4440, 0x5440, 0x8840}},
    {33333, 2, {0x820, 0x1010, 0x3E10, 0x22FE, 0x3282, 0x2A84, 0x2A40, 0xFE44, 0x2248, 0x3250, 0x2A60, 0x2A42, 0x2242, 0x4242, 0x4A3E, 0x8400}},
    {33334, 2, {0x810, 0x1010, 0x3E20, 0x22FC, 0x3284, 0x2A84, 0x2A84, 0xFE84, 0x22FC, 0x3284, 0x2A84, 0x2A84, 0x2284, 0x4284, 0x4AFC, 0x8484}},
    {33335, 2, {0x1020, 0x2010, 0x7C10, 0x45FE, 0x6420, 0x5420, 0x5444, 0xFC84, 0x45F8, 0x6410, 0x5420, 0x5444, 0x4482, 0x45FE, 0x5482, 0x8800}},
    {33336, 2, {0x1000, 0x2000, 0x7DFE, 0x4408, 0x6408, 0x55E8, 0x5528, 0xFD28, 0x4528, 0x6528, 0x55E8, 0x5528, 0x4408, 0x4408, 0x5428, 0x8810}},
    {33337, 2, {0x800, 0x1078, 0x3E48, 0x2248, 0x3248, 0x2A48, 0x2A8E, 0xFF00, 0x2200, 0x32FC, 0x2A84, 0x2A84, 0x2284, 0x4284, 0x4AFC, 0x8484}},
    {33338, 2, {0x1000, 0x21FC, 0x7D24, 0x4524, 0x6524, 0x55FC, 0x5524, 0xFD24, 0x4524, 0x65FC, 0x5524, 0x5420, 0x4420, 0x4420, 0x5420, 0x8820}},
    {33339, 2, {0x1020, 0x2020, 0x7C3C, 0x4420, 0x6420, 0x54FC, 0x5484, 0xFC84, 0x4484, 0x64FC, 0x5484, 0x5480, 0x4480, 0x4500, 0x5500, 0x8A00}},
    {33340, 2, {0x1048, 0x2048, 0x7C48, 0x4448, 0x65FE, 0x5448, 0x5448, 0xFC48, 0x4448, 0x65FE, 0x5400, 0x5448, 0x4444, 0x4484, 0x5482, 0x8902}},
    {33341, 2, {0x820, 0x1020, 0x3E3C, 0x2244, 0x32A8, 0x2A10, 0x2A28, 0xFED6, 0x2210, 0x32FC, 0x2A10, 0x2A90, 0x22FE, 0x4210, 0x4A10, 0x8410}},
    {33342, 2, {0x1000, 0x23FE, 0x7C50, 0x4450, 0x65FC, 0x5554, 0x5554, 0xFD54, 0x4554, 0x6554, 0x558C, 0x5504, 0x4504, 0x4504, 0x55FC, 0x8904}},
    {33343, 2, {0x1020, 0x2020, 0x7BFE, 0x4850, 0x6888, 0x5904, 0x4AFA, 0xF800, 0x4BFE, 0x4840, 0x6880, 0x59FC, 0x4804, 0x4804, 0x4828, 0x9810}},
    {33344, 2, {0x1008, 0x203C, 0x7DC0, 0x4404, 0x6544, 0x54A8, 0x5400, 0xFDF8, 0x4410, 0x6420, 0x57FE, 0x5420, 0x4420, 0x4420, 0x54A0, 0x8840}},
    {33345, 2, {0x810, 0x1050, 0x3E50, 0x227C, 0x3250, 0x2A90, 0x2A10, 0xFEFE, 0x2200, 0x3200, 0x2A7C, 0x2A44, 0x2244, 0x4244, 0x4A7C, 0x8444}},
    {33346, 2, {0x1040, 0x2040, 0x7CFC, 0x4488, 0x6550, 0x5420, 0x54D8, 0xFF26, 0x44F8, 0x6420, 0x54F8, 0x5420, 0x47FE, 0x4420, 0x5420, 0x8820}},
    {33347, 2, {0x1000, 0x21FC, 0x7D24, 0x4524, 0x65FC, 0x5524, 0x5524, 0xFDFC, 0x4420, 0x6420, 0x55FC, 0x5420, 0x4420, 0x4420, 0x57FE, 0x8800}},
    {33348, 2, {0x1020, 0x2124, 0x7CA4, 0x44A8, 0x6420, 0x55FC, 0x5504, 0xFD04, 0x45FC, 0x6504, 0x5504, 0x55FC, 0x4504, 0x4504, 0x5514, 0x8908}},
    {33349, 2, {0x1020, 0x2020, 0x7C50, 0x4488, 0x6504, 0x56FA, 0x5420, 0xFC20, 0x47FE, 0x6420, 0x5528, 0x5524, 0x4622, 0x4422, 0x54A0, 0x8840}},
    {33350, 2, {0x820, 0x1010, 0x3EFC, 0x2284, 0x3284, 0x2AFC, 0x2A84, 0xFE84, 0x22FC, 0x32A2, 0x2AA4, 0x2A98, 0x2290, 0x4288, 0x4AC4, 0x8482}},
    {33351, 2, {0x1000, 0x2006, 0x7BB8, 0x4888, 0x6888, 0x5908, 0x493E, 0xFB88, 0x4888, 0x4A88, 0x6A88, 0x593E, 0x4900, 0x4A80, 0x4C7E, 0x9800}},
    {33352, 2, {0x1020, 0x2010, 0x7DFE, 0x4420, 0x6448, 0x5484, 0x55FE, 0xFC02, 0x44A8, 0x64A8, 0x54A8, 0x54A8, 0x44A8, 0x44AA, 0x552A, 0x8A06}},
    {33353, 2, {0x1000, 0x21FC, 0x7D04, 0x4504, 0x65FC, 0x5500, 0x551C, 0xFD70, 0x451C, 0x6570, 0x551E, 0x55F0, 0x4510, 0x4512, 0x5512, 0x8A0E}},
    {33354, 2, {0x810, 0x1020, 0x3E7C, 0x2244, 0x327C, 0x2A44, 0x2A7C, 0xFE10, 0x2210, 0x32FE, 0x2A92, 0x2A92, 0x2292, 0x4296, 0x4A10, 0x8410}},
    {33355, 2, {0x1000, 0x21FC, 0x7C08, 0x4410, 0x6420, 0x57FE, 0x5420, 0xFCA0, 0x4440, 0x65FC, 0x5554, 0x5554, 0x4554, 0x4554, 0x57FE, 0x8800}},
    {33356, 2, {0x1020, 0x2020, 0x7850, 0x4888, 0x6944, 0x5A22, 0x49F8, 0xF808, 0x4850, 0x4820, 0x68A4, 0x5A82, 0x4A8A, 0x4A8A, 0x4C78, 0x9800}},
    {33357, 2, {0x1000, 0x21FC, 0x7D04, 0x4504, 0x65FC, 0x5510, 0x5510, 0xFDFE, 0x4510, 0x6510, 0x557C, 0x5544, 0x4544, 0x4544, 0x557C, 0x8A44}},
    {33358, 2, {0x1020, 0x2040, 0x7DFC, 0x4504, 0x65FC, 0x5504, 0x55FC, 0xFC00, 0x47FE, 0x6420, 0x5420, 0x55FC, 0x4420, 0x4420, 0x57FE, 0x8800}},
    {33359, 2, {0x1084, 0x2048, 0x7DFE, 0x4420, 0x6440, 0x54FC, 0x5484, 0xFC84, 0x44FC, 0x6484, 0x5484, 0x54FC, 0x4484, 0x4484, 0x54FC, 0x8884}},
    {33360, 2, {0x1050, 0x2124, 0x7D54, 0x4504, 0x65FC, 0x5450, 0x5488, 0xFD44, 0x4478, 0x6488, 0x5488, 0x5550, 0x4420, 0x4450, 0x5488, 0x8906}},
    {33361, 2, {0x1020, 0x2010, 0x7DFE, 0x4502, 0x6502, 0x55FE, 0x5500, 0xFD00, 0x45FE, 0x65AA, 0x56AA, 0x54FE, 0x44AA, 0x44AA, 0x54AA, 0x8886}},
    {33362, 2, {0x1000, 0x21FE, 0x7D02, 0x457A, 0x6502, 0x557A, 0x5502, 0xFC00, 0x44FC, 0x6484, 0x54FC, 0x5484, 0x44FC, 0x4484, 0x54FC, 0x8884}},
    {33363, 2, {0x1048, 0x2148, 0x7D48, 0x47FE, 0x6548, 0x5548, 0x5578, 0xFD00, 0x45FE, 0x6420, 0x57FE, 0x5470, 0x44A8, 0x4524, 0x5622, 0x8820}},
    {33364, 2, {0x1040, 0x2020, 0x7DFE, 0x4548, 0x6548, 0x55FE, 0x5548, 0xFD78, 0x4500, 0x65FC, 0x5544, 0x5548, 0x4528, 0x4510, 0x5528, 0x8AC6}},
    {33365, 2, {0x1020, 0x2010, 0x7DFE, 0x4484, 0x6448, 0x55FE, 0x5502, 0xFE24, 0x4410, 0x65FE, 0x5440, 0x547C, 0x4444, 0x4484, 0x5494, 0x8908}},
    {33366, 2, {0x1084, 0x2044, 0x7C48, 0x45FE, 0x6420, 0x54FC, 0x5420, 0xFDFE, 0x4440, 0x6480, 0x54FE, 0x5510, 0x4610, 0x4410, 0x55FE, 0x8800}},
    {33367, 2, {0x1104, 0x2084, 0x7C88, 0x4400, 0x67FE, 0x5400, 0x5488, 0xFD04, 0x4602, 0x65FC, 0x5554, 0x5554, 0x4554, 0x4554, 0x57FE, 0x8800}},
    {33368, 2, {0x10A0, 0x212C, 0x7D24, 0x4524, 0x65AC, 0x5524, 0x5524, 0xFDFC, 0x4420, 0x65FC, 0x5488, 0x5450, 0x4420, 0x4450, 0x5488, 0x8B06}},
    {33369, 2, {0x1020, 0x2020, 0x7C50, 0x4448, 0x64A4, 0x55FE, 0x5684, 0xFCFC, 0x4484, 0x64FC, 0x5480, 0x54FC, 0x4544, 0x4544, 0x567C, 0x8844}},
    {33370, 2, {0x1050, 0x2050, 0x7DFE, 0x4450, 0x65FC, 0x5554, 0x55FC, 0xFD54, 0x45FC, 0x6400, 0x54F8, 0x5488, 0x44F8, 0x4488, 0x54F8, 0x8888}},
    {33371, 2, {0x1020, 0x21FC, 0x7D24, 0x47FE, 0x6524, 0x55FC, 0x5420, 0xFDFC, 0x4524, 0x65FC, 0x5440, 0x57FE, 0x4488, 0x45D0, 0x5470, 0x8B8C}},
    {33372, 2, {0x1154, 0x2154, 0x7BFE, 0x4954, 0x6954, 0x5A72, 0x4C00, 0xFBFE, 0x4A22, 0x4820, 0x69FC, 0x5924, 0x4924, 0x4924, 0x492C, 0x9820}},
    {33373, 2, {0x1000, 0x21FC, 0x7820, 0x4BFE, 0x6A22, 0x59AC, 0x4820, 0xF9AC, 0x4800, 0x49FC, 0x6804, 0x59FC, 0x4804, 0x4804, 0x49FC, 0x9804}},
    {33374, 2, {0x1110, 0x2110, 0x7DDE, 0x46A8, 0x6444, 0x5450, 0x5554, 0xFCD8, 0x4450, 0x64D8, 0x5554, 0x5652, 0x4490, 0x4492, 0x5512, 0x8A0E}},
    {33375, 2, {0x1020, 0x21FC, 0x7C88, 0x4450, 0x67FE, 0x5400, 0x55FC, 0xFD24, 0x45FC, 0x6524, 0x55FC, 0x5420, 0x45FC, 0x4420, 0x57FE, 0x8800}},
    {33376, 2, {0x1010, 0x23D2, 0x7854, 0x4948, 0x688A, 0x5904, 0x4AFA, 0xFC00, 0x49FC, 0x4904, 0x6904, 0x59FC, 0x4904, 0x4888, 0x4FFE, 0x9800}},
    {33377, 2, {0x1020, 0x2124, 0x78A8, 0x4BFE, 0x6A02, 0x58F8, 0x4888, 0xF888, 0x48F8, 0x4800, 0x69FC, 0x5924, 0x49FC, 0x4924, 0x49FC, 0x9904}},
    {33378, 2, {0x1020, 0x2020, 0x7BFE, 0x4924, 0x6924, 0x5AAA, 0x4FFE, 0xF800, 0x49FC, 0x4904, 0x6974, 0x5954, 0x4974, 0x4904, 0x49FC, 0x9904}},
    {33379, 2, {0x1040, 0x207C, 0x7840, 0x4BFE, 0x6A42, 0x5BF8, 0x4A44, 0xFAFC, 0x4AA8, 0x4AF8, 0x6AA8, 0x5AF8, 0x4A20, 0x4DFC, 0x4844, 0x998C}},
    {33380, 2, {0x1104, 0x2088, 0x7BFE, 0x4820, 0x69FC, 0x5820, 0x4BFE, 0xF854, 0x4992, 0x4890, 0x6BFE, 0x5890, 0x48D4, 0x4B8A, 0x4896, 0x99A2}},
    {33381, 2, {0x1010, 0x23D4, 0x7A52, 0x4A52, 0x6BD0, 0x5810, 0x4FFE, 0xFA50, 0x4BD2, 0x4A52, 0x6BD4, 0x5A54, 0x4AEA, 0x4F4A, 0x4856, 0x9862}},
    {33382, 2, {0x1008, 0x23E8, 0x7A88, 0x4BEE, 0x6A28, 0x5BF4, 0x4A82, 0xFBE2, 0x4800, 0x49FC, 0x6954, 0x5954, 0x4954, 0x4954, 0x4FFE, 0x9800}},
    {33383, 2, {0x1088, 0x23FE, 0x7CA8, 0x4490, 0x65FE, 0x5720, 0x55FC, 0xFD20, 0x45FC, 0x6520, 0x55FE, 0x5500, 0x47FC, 0x4488, 0x5470, 0x8B8E}},
    {33384, 2, {0x1088, 0x23FE, 0x7888, 0x4BFE, 0x6A02, 0x59FC, 0x4800, 0xFBFE, 0x4840, 0x48A2, 0x6B54, 0x58B8, 0x4B54, 0x4892, 0x4B50, 0x9820}},
    {33385, 2, {0x1040, 0x2020, 0x7BFE, 0x4888, 0x6852, 0x5FAC, 0x4AAA, 0xFAA8, 0x4DAC, 0x4800, 0x6904, 0x59FC, 0x4904, 0x49FC, 0x4904, 0x9A04}},
    {33386, 2, {0x1080, 0x21F8, 0x7E08, 0x45FC, 0x6524, 0x55FC, 0x5524, 0xFDFC, 0x4554, 0x662A, 0x55FC, 0x5504, 0x45FC, 0x4504, 0x55FC, 0x8904}},
    {33387, 2, {0x1040, 0x207C, 0x7840, 0x4BFC, 0x6A44, 0x5BF0, 0x4A44, 0xFAFC, 0x4AA8, 0x4AF8, 0x6AA8, 0x5AF8, 0x4A00, 0x4DFC, 0x4954, 0x9BFE}},
    {33388, 2, {0x11F0, 0x2210, 0x7BFC, 0x4E04, 0x6BFC, 0x5A24, 0x4BB8, 0xFA22, 0x49FE, 0x4A10, 0x6FFC, 0x5A44, 0x4BFC, 0x48D0, 0x494A, 0x9A3E}},
    {33389, 2, {0x1108, 0x2294, 0x7BDE, 0x4EB4, 0x6BDE, 0x5A94, 0x4BDE, 0xFA94, 0x4BDE, 0x4A10, 0x6FFC, 0x5908, 0x4890, 0x4860, 0x4998, 0x9E06}},
    {33390, 2, {0x0, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2200, 0x2208, 0x2110, 0x20A0, 0x2440, 0x2820, 0x3018, 0x2006}},
    {33391, 2, {0x200, 0x100, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2208, 0x2110, 0x20A0, 0x2040, 0x2420, 0x2818, 0x3006}},
    {33392, 2, {0x0, 0x1F8, 0xFD08, 0x508, 0x45F8, 0x4508, 0x2908, 0x29F8, 0x1144, 0x1148, 0x2930, 0x2920, 0x4510, 0x4548, 0x8186, 0x100}},
    {33393, 2, {0x2200, 0x227C, 0xFF44, 0x2244, 0x3E7C, 0x844, 0x7F44, 0x497C, 0x7F52, 0x852, 0x7F54, 0x848, 0xFF48, 0x1444, 0x2262, 0x4140}},
    {33394, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2002, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {33395, 2, {0x1080, 0x1080, 0x10F8, 0xFD08, 0x1210, 0x15FC, 0x1124, 0x7D24, 0x1124, 0x11FC, 0x1100, 0xFD00, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {33396, 2, {0x2820, 0x2820, 0x287C, 0xFE44, 0x2A88, 0x2A7C, 0xFE54, 0xA854, 0xA854, 0xFF7C, 0x2940, 0x2940, 0x2D42, 0x4A42, 0x483E, 0x8800}},
    {33397, 2, {0x4240, 0x2440, 0xFC, 0x7E88, 0x2510, 0x24FC, 0x24A4, 0x24A4, 0xFFA4, 0x24FC, 0x2480, 0x2480, 0x2482, 0x4482, 0x447E, 0x8400}},
    {33398, 2, {0x2420, 0xFF20, 0xA57C, 0xFF44, 0xA588, 0xFF7C, 0x54, 0xFF54, 0x54, 0xFF7C, 0x8140, 0xFF40, 0x4242, 0x2442, 0x1F3E, 0xE000}},
    {33399, 2, {0x2A10, 0x7F10, 0xAA9C, 0xFFA4, 0xAAC8, 0xFFBE, 0xAAAA, 0xFFAA, 0x2A, 0xFFBE, 0x20, 0x7F20, 0x4122, 0x7F22, 0x2222, 0xFF9E}},
    {33400, 2, {0x1010, 0x1010, 0x1010, 0x9292, 0x9292, 0x9292, 0x9292, 0x9292, 0xFEFE, 0x1212, 0x1010, 0x1010, 0x2010, 0x2010, 0x4010, 0x8010}},
    {33401, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {33402, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FE0, 0xC0, 0x300, 0x400, 0x800, 0x1000, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {33403, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x200, 0x200, 0x7FF8, 0x208, 0x408, 0x408, 0x808, 0x808, 0x1008, 0x2050, 0x4020}},
    {33404, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {33405, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x400, 0x400, 0x7FE0, 0x420, 0x420, 0x820, 0x820, 0x1022, 0x2022, 0x401E, 0x8000}},
    {33406, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x820, 0x820, 0x440, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33407, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x408, 0x410, 0x420, 0x47C, 0x804, 0x804, 0x1004, 0x2044, 0x4028, 0x8010}},
    {33408, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x208, 0x208, 0x208, 0x208, 0x408, 0x408, 0x808, 0x1050, 0x6020}},
    {33409, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0xFE0, 0x820, 0x820, 0x820, 0x820, 0x1022, 0x1022, 0x2022, 0x401E, 0x8000}},
    {33410, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x7FF8, 0x208, 0x208, 0x208, 0x208, 0x250, 0x220, 0x200, 0x200, 0x200}},
    {33411, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0xFE0, 0x820, 0x820, 0xA20, 0x920, 0x922, 0x1022, 0x1022, 0x201E, 0x4000}},
    {33412, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x200, 0x200, 0x7FE0, 0x220, 0xA20, 0x420, 0x620, 0x922, 0x1022, 0x201E, 0x4000}},
    {33413, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x210, 0x210, 0x2FE, 0xFF00, 0x200, 0x100, 0x100, 0x80, 0x42, 0x32, 0xE}},
    {33414, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF0, 0x1210, 0x1120, 0x820, 0x440, 0x280, 0x100, 0x280, 0x440, 0x1830, 0xE00E}},
    {33415, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100, 0x100}},
    {33416, 2, {0x920, 0x920, 0xF93E, 0x920, 0x920, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {33417, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {33418, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0xF8, 0x3F00, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {33419, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {33420, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x400, 0x800, 0xFF0, 0x10, 0x10, 0x10, 0xA0, 0x40}},
    {33421, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x8808, 0x408, 0x208, 0x208, 0x8, 0x50, 0x20}},
    {33422, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x2000, 0x2000, 0x3FFC, 0x4, 0x4, 0x28, 0x10}},
    {33423, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {33424, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0xFFFE, 0x200, 0x200, 0x240, 0x220, 0x210, 0x208, 0x200, 0x200, 0x200}},
    {33425, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x3FF0, 0x2000, 0x2004, 0x2004, 0x2004, 0x1FFC, 0x0}},
    {33426, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x200, 0x100, 0xFFFE, 0x800, 0x800, 0x800, 0x800, 0x800, 0x800, 0xFFC, 0x0}},
    {33427, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x1FE0, 0x40, 0x80, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {33428, 2, {0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x1110, 0x1010, 0x9292, 0x9292, 0x9292, 0xFEFE, 0x2212, 0x2010, 0x4010, 0x8010}},
    {33429, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x200, 0x7F8, 0x808, 0x1010, 0x6420, 0x240, 0x280, 0x100, 0x600, 0x1800, 0xE000}},
    {33430, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x280, 0x440, 0x820, 0x1010, 0x2008, 0xC006}},
    {33431, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x400, 0x810, 0x1FE0, 0xC0, 0x304, 0xC08, 0x3FF0, 0x1020, 0x40, 0x180, 0xE00, 0x7000}},
    {33432, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x2080, 0x2080, 0x2088, 0x3C90, 0x20A0, 0x20C0, 0x2084, 0x2484, 0x2884, 0x307C, 0x0}},
    {33433, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33434, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0xFFFE, 0x100, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x10A, 0x102, 0x102, 0xFE}},
    {33435, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF0, 0x210, 0x210, 0xFFFE, 0x210, 0x410, 0x3FF0, 0x810, 0x1000, 0x2000, 0xC000}},
    {33436, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF0, 0x200, 0x200, 0x7FFC, 0x280, 0x480, 0x480, 0x884, 0x1084, 0x207C, 0xC000}},
    {33437, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x100, 0x7FF8, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x4800, 0x87FE, 0x0}},
    {33438, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1FFC, 0x2000, 0x5FF8, 0x8000, 0x3FF0, 0x10, 0x10, 0x10, 0xA, 0x6, 0x2}},
    {33439, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0xFE0, 0x820, 0x1020, 0x601C, 0x0, 0x1FE0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {33440, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x200, 0x100, 0x7FFC, 0x820, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {33441, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x800, 0xFFC, 0x1004, 0x2108, 0x4100, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33442, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x800, 0x800, 0x13FC, 0x3000, 0x5000, 0x9000, 0x1000, 0x1000, 0x17FE, 0x1000, 0x1000}},
    {33443, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x7FFC, 0x80, 0x100, 0x300, 0x560, 0x918, 0x3104, 0xC102, 0x100, 0x100}},
    {33444, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x7F40, 0x240, 0x440, 0x840, 0xF40, 0xF840, 0x4842, 0x842, 0x842, 0x283E, 0x1000}},
    {33445, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0x440, 0x820, 0x3458, 0xC446, 0x440, 0x440, 0x840, 0x840, 0x1040, 0x2040}},
    {33446, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x80, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2000, 0x2000, 0x4000, 0x4000, 0x8000}},
    {33447, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF0, 0x20, 0x640, 0x180, 0xFFFE, 0x82, 0x84, 0x80, 0x80, 0x280, 0x100}},
    {33448, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF0, 0x820, 0x840, 0xCF8, 0xA08, 0x1110, 0x10A0, 0x2040, 0x20A0, 0x4318, 0x8C06}},
    {33449, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0x440, 0xA20, 0x3118, 0xC106, 0x1FE0, 0x20, 0x40, 0x40, 0x80, 0x100}},
    {33450, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x78, 0x3F80, 0x2080, 0x2080, 0x2080, 0x3FFE, 0x2040, 0x2040, 0x2024, 0x2814, 0x300C, 0x2004}},
    {33451, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x844, 0x844, 0x1044, 0x603C}},
    {33452, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x40, 0x440, 0x820, 0x1010, 0x2008, 0xDFE6, 0x420, 0x420, 0x820, 0x1020, 0x2140, 0x4080}},
    {33453, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2000, 0x2002, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {33454, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x2008, 0x2028, 0x2010}},
    {33455, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x200, 0x100, 0x80, 0x408, 0x2404, 0x2402, 0x4402, 0x8410, 0x410, 0x3F0, 0x0}},
    {33456, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF0, 0x820, 0x440, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {33457, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x888, 0x890, 0x10A0, 0x30C0, 0x5080, 0x9180, 0x1282, 0x1482, 0x1082, 0x107E, 0x1000}},
    {33458, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0xC60, 0x3018, 0xC806, 0x8E0, 0xF00, 0x7808, 0x808, 0x808, 0x7F8, 0x0}},
    {33459, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0xFFFE, 0x400, 0x400, 0x7F0, 0x410, 0x810, 0x810, 0x1010, 0x20A0, 0x4040}},
    {33460, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x1000, 0x1FF8, 0x2248, 0x4248, 0x8488, 0x888, 0x1108, 0x2208, 0x4408, 0x850, 0x1020}},
    {33461, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF0, 0x110, 0x110, 0x110, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33462, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x1FF8, 0x2008, 0x4408, 0x8408, 0x908, 0x1088, 0x3FC8, 0x1048, 0x8, 0x50, 0x20}},
    {33463, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0}},
    {33464, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {33465, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x38, 0x1FC0, 0x1000, 0x1000, 0x1FFC, 0x1040, 0x1040, 0x1040, 0x2040, 0x2040, 0x4040, 0x8040}},
    {33466, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0xF0, 0x3F00, 0x100, 0x100, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33467, 2, {0x800, 0x800, 0x1FF8, 0x2208, 0x5248, 0x1FC8, 0x408, 0x850, 0x3020, 0x1FF8, 0x2208, 0x5248, 0x9FC8, 0x408, 0x850, 0x3020}},
    {33468, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0xF8, 0x3F00, 0x100, 0x1F8, 0x3F00, 0x100, 0x1FC, 0x7F02, 0x102, 0x102, 0xFE}},
    {33469, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FFC, 0x20, 0x1020, 0x2020, 0x3FFE, 0x120, 0x220, 0xC20, 0x3020, 0xC0A0, 0x40}},
    {33470, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0xFFFE, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110, 0x100}},
    {33471, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x17F8, 0x2108, 0x2110, 0x6120, 0xA13C, 0x2104, 0x2104, 0x2204, 0x2204, 0x2428, 0x2810}},
    {33472, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x7FFC, 0x0, 0x0, 0xFE0, 0x820, 0x820, 0x822, 0x1022, 0x201E, 0xC000}},
    {33473, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x820, 0x820, 0x820, 0x820, 0x820, 0x1450, 0x1250, 0x2288, 0x4104, 0x8202}},
    {33474, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x1108, 0x1108, 0x1110, 0x2120, 0x280, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33475, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x200, 0xFFFE, 0x400, 0x400, 0xFF0, 0x810, 0x1420, 0x2240, 0x4180, 0x8670, 0x380E}},
    {33476, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x100, 0xFFFE, 0x200, 0x200, 0x240, 0x220, 0x210, 0x200, 0x200, 0x200}},
    {33477, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x204, 0x4244, 0x2444, 0x1444, 0x844, 0x844, 0x1444, 0x2404, 0x4204, 0x8214, 0x8}},
    {33478, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x2000, 0x21FC, 0x2C44, 0xF044, 0x2044, 0x2044, 0x2484, 0x2884, 0x3104, 0x2228, 0x410}},
    {33479, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x104, 0x104, 0x114, 0x108}},
    {33480, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FFC, 0x2100, 0x2100, 0x2FF8, 0x2108, 0x2108, 0x2208, 0x2208, 0x4408, 0x4850, 0x9020}},
    {33481, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FFC, 0x2240, 0x2240, 0x2240, 0x2244, 0x2444, 0x243C, 0x2800, 0x2000, 0x3FFC, 0x0}},
    {33482, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FFC, 0x2000, 0x2000, 0x27E0, 0x2420, 0x2420, 0x24A0, 0x2444, 0x4404, 0x4404, 0x83FC}},
    {33483, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1290, 0x280, 0x482, 0x1882, 0x607E}},
    {33484, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x830, 0x9C0, 0xE00, 0x800, 0xFFFE, 0x900, 0x880, 0x840, 0xA30, 0xC0E, 0x800}},
    {33485, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x820, 0x820, 0x8A8, 0x848, 0x808, 0x7F8}},
    {33486, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x0, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0}},
    {33487, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x200, 0x7FF0, 0x210, 0x214, 0x2212, 0x2212, 0x2412, 0x4410, 0x810, 0x10A0, 0x2040}},
    {33488, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x108, 0x108, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x304, 0x504, 0x1928, 0xE110, 0x100}},
    {33489, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1EF8, 0x1288, 0x2288, 0x2288, 0x52A8, 0x8C90, 0x482, 0x882, 0x1082, 0x207E, 0x4000}},
    {33490, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0xFFFE, 0x2008, 0x2028, 0x2010}},
    {33491, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0xA20, 0x3118, 0xC106, 0x1FE0, 0x20, 0x40, 0xC80, 0x300, 0x80, 0x40}},
    {33492, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33493, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FF8, 0x408, 0x808, 0x1050, 0x2020, 0xDFF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {33494, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {33495, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008}},
    {33496, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {33497, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x0, 0x7FFC, 0x10, 0x1010, 0x820, 0x440, 0x480, 0x0, 0xFFFE, 0x0}},
    {33498, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x4108, 0x4128, 0x8010}},
    {33499, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x10, 0x10, 0x1F10, 0x1110, 0x1110, 0x1F10, 0x1110, 0x10, 0x50, 0x20}},
    {33500, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33501, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x440, 0x444, 0x7C48, 0x470, 0x440, 0x442, 0xC42, 0x3442, 0xC43E, 0x400}},
    {33502, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1090, 0x1F90, 0x1050, 0x1024, 0x1004, 0xFFC, 0x0}},
    {33503, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x9F88, 0x1088, 0x1088, 0x1F88, 0x1088, 0x8, 0x50, 0x20}},
    {33504, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2080, 0x2080, 0x3FFC, 0x2040, 0x2024, 0x2414, 0x280C, 0x3004}},
    {33505, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x2410, 0x2210, 0x2110, 0x2110, 0x2010, 0x2020, 0x2420, 0x2850, 0x3088, 0x2104, 0x602}},
    {33506, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1000, 0x1000, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008}},
    {33507, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF0, 0x2010, 0x2010, 0x3FF0, 0x2000, 0x2000, 0x2000, 0x3FFC}},
    {33508, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x80, 0x100, 0x360, 0x510, 0x1908, 0x6104, 0x100, 0x100, 0xFFFE, 0x0}},
    {33509, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x200, 0xFFFE, 0x400, 0x800, 0x1FF0, 0x2810, 0x4810, 0x8810, 0x810, 0xFF0, 0x810}},
    {33510, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33511, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {33512, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1000, 0x1200, 0x1238, 0x13C0, 0x2204, 0x2204, 0x41FC, 0x8000}},
    {33513, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x200, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33514, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x2028, 0x2010}},
    {33515, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x1FE, 0x100, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33516, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x2008}},
    {33517, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x1040, 0x21FC, 0x2444, 0x4444, 0xF844, 0x1044, 0x2044, 0x4884, 0xFC84, 0x4528, 0x210}},
    {33518, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x820, 0x820, 0x1020, 0x3124, 0x5124, 0x9124, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {33519, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2FE8, 0xC106, 0x100, 0x100, 0x100}},
    {33520, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7E10, 0x210, 0x210, 0x3E10, 0x2020, 0x4028, 0x7E24, 0x244, 0x2FE, 0x1442, 0x800}},
    {33521, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1110, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33522, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x800, 0x1FFC, 0x2400, 0x4400, 0x87F0, 0x400, 0x400, 0x7F8, 0x400, 0x400, 0x400}},
    {33523, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE30E, 0xC0, 0xE00, 0x180, 0x40}},
    {33524, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE, 0x0}},
    {33525, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33526, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0x440, 0x920, 0x3118, 0xC106, 0x100, 0x920, 0x1110, 0x2108, 0x500, 0x200}},
    {33527, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x1010, 0xFFFE, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33528, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x10, 0xF8, 0x3F00, 0x1110, 0x920, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x500, 0x200}},
    {33529, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x1110, 0x910, 0x920, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {33530, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1210, 0x1110, 0xFFFE, 0x1010, 0x2210, 0x2110, 0x3FFC, 0x10, 0xA0, 0x40}},
    {33531, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x810, 0x810, 0x1010, 0x13FE, 0x3010, 0x5110, 0x9090, 0x1090, 0x1010, 0x1010, 0x1050, 0x1020}},
    {33532, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x1100, 0x1100, 0x3FFC, 0x4100, 0x8100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {33533, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x78, 0x1FA0, 0x1220, 0x1220, 0x1220, 0x1210, 0x1290, 0x2248, 0x22A8, 0x4324, 0x8202}},
    {33534, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x90, 0x890, 0x4824, 0x4842, 0x4882, 0x8902, 0xE10, 0x1810, 0xE7F0, 0x0}},
    {33535, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33536, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x3FF8, 0x448, 0x448, 0x1FF8, 0x1440, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2048, 0x4040}},
    {33537, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x108, 0x100, 0x4104, 0x4104, 0x4104, 0x7FFC, 0x4}},
    {33538, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x90, 0x88, 0x3FFC, 0x2080, 0x2088, 0x2088, 0x2050, 0x2050, 0x2024, 0x4054, 0x418C, 0x8604}},
    {33539, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x9F8, 0x4908, 0x2108, 0x2508, 0x528, 0x910, 0x7102, 0x1102, 0x1102, 0x10FE, 0x0}},
    {33540, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1000, 0x107C, 0x7E44, 0x1244, 0x1244, 0x1244, 0x1244, 0x2244, 0x2244, 0x4A7C, 0x8444}},
    {33541, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF0, 0x20, 0x640, 0x180, 0xFFFE, 0x282, 0x484, 0x880, 0x3080, 0xC280, 0x100}},
    {33542, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0xC00, 0x70FC, 0x4084, 0x4484, 0x4484, 0x5C84, 0x6494, 0x4888, 0x880, 0x1080, 0x6080}},
    {33543, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x210, 0x208, 0xFFFE, 0x400, 0x400, 0xFF0, 0x810, 0x1420, 0x2240, 0x4180, 0x8670, 0x380E}},
    {33544, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x440, 0x2444, 0x2448, 0x2750, 0x2460, 0x2440, 0x2442, 0x2442, 0x2F42, 0xF03E, 0x4000}},
    {33545, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33546, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x10, 0x78, 0x1F80, 0x1000, 0x1000, 0x1FFC, 0x1040, 0x1040, 0x1040, 0x1040, 0xFFFE, 0x0}},
    {33547, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x878, 0x3F80, 0x2080, 0x2080, 0x2080, 0x3FFE, 0x2040, 0x2040, 0x2224, 0x2914, 0x308C, 0x2084}},
    {33548, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x840, 0x1040, 0x1040, 0x37FE, 0x5040, 0x9040, 0x1040, 0x1040, 0x1040, 0x13FC, 0x1000}},
    {33549, 2, {0x440, 0x440, 0xFC7E, 0x440, 0x1440, 0x1000, 0x1FF8, 0x2008, 0x4008, 0x9F88, 0x1088, 0x1088, 0x1F88, 0x8, 0x50, 0x20}},
    {33550, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF0, 0x60, 0x180, 0xE70, 0x700C, 0x1FF0, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {33551, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x40, 0x420, 0x7FFC, 0x480, 0x488, 0x890, 0x8A0, 0x10C2, 0x2382, 0x4C82, 0x807E}},
    {33552, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1000, 0x1FF8, 0x2248, 0x4A48, 0x8488, 0xA88, 0x1108, 0x2288, 0x4408, 0x850, 0x1020}},
    {33553, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x200, 0x1FF0, 0x1210, 0x1110, 0x1150, 0x1020, 0x1FFC, 0x4, 0x7FE4, 0x4, 0x28, 0x10}},
    {33554, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1E7C, 0x1244, 0x2248, 0x2250, 0x5248, 0x8C44, 0x444, 0x854, 0x1048, 0x2040, 0x4040}},
    {33555, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x440, 0x440, 0x820, 0x820, 0x1010, 0x2008, 0xC006}},
    {33556, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8104, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {33557, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8004, 0x1FE0, 0x420, 0x420, 0x3FA0, 0x420, 0x414, 0x40C, 0x404}},
    {33558, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0xFF0, 0x1020, 0x2C40, 0x380, 0x1C70, 0xE00E, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33559, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x200, 0x7F8, 0x810, 0x3420, 0x240, 0x180, 0xFF8, 0x1808, 0xE808, 0x808, 0xFF8, 0x808}},
    {33560, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x408, 0x850, 0x3020, 0x0, 0x7EFC, 0x1224, 0x1224, 0x2244, 0x4A94, 0x8508}},
    {33561, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x90, 0x88, 0xFFFE, 0x80, 0x1088, 0x1088, 0x7E50, 0x1050, 0x1024, 0x2054, 0x418C, 0x8604}},
    {33562, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x600, 0x78FC, 0x4084, 0x4084, 0x7E84, 0x4084, 0x4084, 0x4094, 0x4E88, 0x7080, 0x80, 0x80}},
    {33563, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF0, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x2208, 0x2110, 0x20A0, 0x2440, 0x2830, 0x300E}},
    {33564, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0xFFFE, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2848, 0x3038, 0x2008, 0x3FF8, 0x2008}},
    {33565, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FFC, 0x2080, 0x2080, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x2080, 0x2080, 0x3FFE, 0x0}},
    {33566, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FFC, 0x2100, 0x2100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2100, 0x2100, 0x3FFC, 0x0}},
    {33567, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF0, 0x110, 0xFFFE, 0x110, 0x3FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {33568, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x840, 0x840, 0x1040, 0x37FC, 0x50E0, 0x9150, 0x1150, 0x1248, 0x1444, 0x1842, 0x1040}},
    {33569, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x4004, 0x8008, 0x1FE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {33570, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4, 0x7F84, 0x824, 0x824, 0x1F24, 0x2124, 0x5124, 0xA24, 0x404, 0x804, 0x3014, 0xC008}},
    {33571, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {33572, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x7E0, 0x1840, 0x6480, 0x320, 0xC40, 0x71F8, 0x608, 0x1A10, 0x120, 0xC0, 0x700, 0x7800}},
    {33573, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {33574, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2390, 0x560, 0x1918, 0xE106, 0x100}},
    {33575, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x100, 0x110, 0x1F8, 0x7F04, 0x2004}},
    {33576, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4080, 0x2080, 0x20FC, 0x904, 0x1248, 0x1440, 0x2040, 0xE0A0, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {33577, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x78, 0x1F80, 0x1000, 0x1000, 0x1FFC, 0x1000, 0x1000, 0x17F8, 0x2408, 0x2408, 0x47F8, 0x8408}},
    {33578, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x1110, 0x910, 0x920, 0xFFFE, 0x440, 0x440, 0x440, 0x842, 0x1042, 0x203E, 0xC000}},
    {33579, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x2040, 0x1020, 0x87FE, 0x4900, 0x900, 0x1100, 0xE100, 0x2100, 0x2100, 0x21FC, 0x0}},
    {33580, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x200, 0x400, 0xFFFE, 0x800, 0x1040, 0x3040, 0x53FC, 0x9040, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {33581, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x7FFC, 0x820, 0x1010, 0x2448, 0x4444, 0x280, 0x100, 0x6C0, 0x1830, 0xE00E}},
    {33582, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x200, 0x3F8, 0x200, 0x200, 0xFFFE, 0x0, 0x100, 0x1110, 0x1108, 0x2104, 0x4504, 0x200}},
    {33583, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x850, 0x848, 0x1040, 0x37FE, 0x5040, 0x90A0, 0x10A0, 0x1110, 0x1110, 0x1208, 0x1406}},
    {33584, 2, {0x820, 0xFFFE, 0x920, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x1100, 0xA00, 0x600, 0x19C0, 0xE03E}},
    {33585, 2, {0x440, 0x440, 0xFFFE, 0x540, 0x1100, 0x1100, 0x3FF8, 0x4100, 0x100, 0xFFFE, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33586, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x2040, 0x2448, 0x4488, 0x78F0, 0x810, 0x1020, 0x1428, 0x2244, 0x7EFC, 0x204}},
    {33587, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x2000, 0x17FC, 0x9040, 0x4040, 0x4840, 0x840, 0x1040, 0xE040, 0x2040, 0x2FFE, 0x2000}},
    {33588, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33589, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2FE8, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x2008}},
    {33590, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0x440, 0x920, 0x3118, 0xDFF6, 0x100, 0x920, 0x1110, 0x2108, 0x500, 0x200}},
    {33591, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x8A0, 0x890, 0x1080, 0x30FC, 0x5780, 0x9048, 0x1050, 0x1024, 0x1054, 0x118C, 0x1604}},
    {33592, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x103E, 0xFFD0, 0x10, 0x10, 0x10}},
    {33593, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x1000, 0xFE7C, 0x1244, 0x2244, 0x2244, 0x6444, 0x1444, 0x844, 0x147C, 0x2244, 0xC000}},
    {33594, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x448, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {33595, 2, {0x1010, 0x9292, 0x9292, 0x9292, 0xFEFE, 0x2010, 0x4010, 0x8000, 0x1010, 0x9292, 0x9292, 0x9292, 0xFEFE, 0x2212, 0x2010, 0x4010}},
    {33596, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2008, 0x2FE8, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {33597, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1040, 0x1040, 0x27FC, 0x6444, 0xA444, 0x2444, 0x27FC, 0x2444, 0x2040, 0x2040, 0x2040}},
    {33598, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x7FFC, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {33599, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x0, 0x7DF0, 0x1110, 0x1110, 0x1190, 0x1150, 0x1D12, 0xE212, 0x4212, 0x40E, 0x800}},
    {33600, 2, {0x440, 0x440, 0xFFFE, 0x1440, 0x1000, 0x1FF8, 0x2008, 0x5FC8, 0x9048, 0x1FC8, 0x1048, 0x1FC8, 0x1048, 0x8, 0x50, 0x20}},
    {33601, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x0, 0xFFFE, 0x0}},
    {33602, 2, {0x820, 0xFFFE, 0xA20, 0x200, 0x7FFC, 0x440, 0x820, 0x37D8, 0xC006, 0x3FF8, 0x800, 0xFE0, 0x20, 0x20, 0x140, 0x80}},
    {33603, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0x820, 0x3018, 0xCFE6, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0}},
    {33604, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xA20, 0x100, 0x7FFC, 0x400, 0x840, 0x1F80, 0x110, 0x620, 0x3840, 0x1A0, 0xE10, 0x7008}},
    {33605, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0xC60, 0x3018, 0xCFE6, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33606, 2, {0x1204, 0x1204, 0xFFC4, 0x1224, 0x24, 0x7FA4, 0x1224, 0x1224, 0x1224, 0xFFE4, 0x1224, 0x1224, 0x2204, 0x2204, 0x4214, 0x8208}},
    {33607, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x21FC, 0x4000, 0x8800, 0x13FE, 0x3020, 0x5020, 0x9020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {33608, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x1010, 0x3E10, 0x22FC, 0x4210, 0xA490, 0x1490, 0x8FE, 0x1010, 0x2010, 0x4010, 0x8010}},
    {33609, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {33610, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4, 0x7F84, 0x1224, 0x1224, 0x1224, 0xFFE4, 0x1224, 0x1224, 0x2204, 0x2204, 0x4214, 0x8208}},
    {33611, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x200, 0x3FF8, 0x2488, 0x2488, 0x2488, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {33612, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8204, 0x200, 0xFFFE, 0x420, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {33613, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x4440, 0x4480, 0x44FE, 0x4508, 0x4688, 0x5C90, 0x6450, 0x4420, 0x450, 0x488, 0x506}},
    {33614, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x400, 0x820, 0x1FF0, 0x110, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {33615, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x810, 0x878, 0x13C0, 0x3040, 0x5040, 0x9FFE, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000}},
    {33616, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x200, 0x7FFC, 0x400, 0x9F8, 0x810, 0x1820, 0x2BFE, 0x4820, 0x8820, 0x820, 0x8A0, 0x840}},
    {33617, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x100, 0x7FF8, 0x108, 0x3FF8, 0x2100, 0x3FFC, 0x284, 0x44C, 0x1830, 0xE00E}},
    {33618, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1FF8, 0x0, 0x1110, 0x1110, 0x1110, 0x2112, 0x4112, 0x810E}},
    {33619, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x820, 0x7FFC, 0x820, 0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020}},
    {33620, 2, {0x820, 0xFFFE, 0xA20, 0x200, 0x3FF0, 0x410, 0x18A0, 0x6040, 0x1020, 0x1020, 0x7EFC, 0x1224, 0x2244, 0x2244, 0x4A94, 0x8508}},
    {33621, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3E20, 0x2220, 0x22FC, 0x3E24, 0x2224, 0x2224, 0x3E24, 0x2244, 0x4244, 0x4A94, 0x8508}},
    {33622, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x100, 0x610, 0x18E0, 0x2F00, 0xC808, 0x808, 0x7F8}},
    {33623, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x90, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3088, 0x2850, 0x2450, 0x2024, 0x4054, 0x418C, 0x8604}},
    {33624, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x820, 0x4820, 0x2820, 0x2820, 0xBFE, 0x820, 0x1820, 0x2820, 0xC820, 0x9FC, 0x800}},
    {33625, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1040, 0x1040, 0x7FC, 0x7040, 0x10A0, 0x1090, 0x1108, 0x1204, 0x1404, 0x2800, 0x47FE}},
    {33626, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x100, 0x3FF8, 0x100, 0x1110, 0x920, 0xFFFE, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33627, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x2FC, 0x7F20, 0x140, 0x84, 0x764, 0x381C, 0x0, 0x7FFC, 0x440, 0x842, 0x1042, 0x603E}},
    {33628, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2080, 0x2088, 0x3EB0, 0x20C4, 0x2084, 0x267C, 0x3900, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {33629, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4, 0x7F04, 0x4124, 0x4924, 0x4924, 0x4924, 0x4924, 0x4924, 0x1424, 0x2204, 0x4114, 0x8088}},
    {33630, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xF0, 0x3F00, 0x200, 0xFFFE, 0x440, 0x820, 0x3458, 0xC446, 0x440, 0x440, 0x840, 0x1040}},
    {33631, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0x820, 0x37D8, 0xC006, 0x0, 0x3FF8, 0x200, 0x440, 0x820, 0x1FF0, 0x810}},
    {33632, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC446, 0x440, 0x440, 0x840, 0x1040}},
    {33633, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x23F0, 0x1020, 0x1040, 0x8080, 0x4BFC, 0x48A4, 0x1124, 0xE244, 0x2484, 0x2128, 0x2210}},
    {33634, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {33635, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8104, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33636, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4202, 0x8204, 0x3FF8, 0x900, 0x1100, 0x1FF0, 0x100, 0x100, 0xFFFE, 0x100, 0x100}},
    {33637, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8104, 0x104, 0x7D88, 0x550, 0x920, 0x1110, 0x2108, 0xC506, 0x200}},
    {33638, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8104, 0x1100, 0x1FF8, 0x2100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {33639, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8104, 0x1110, 0x1110, 0x22A0, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33640, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3FF0, 0x10, 0x1FF0, 0x10, 0x3FF0, 0x20, 0xFFFE, 0x820, 0x420, 0x420, 0xA0, 0x40}},
    {33641, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1080, 0x1040, 0x1620, 0x2110, 0x2088, 0x4606, 0x8180, 0x40}},
    {33642, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x7C20, 0x420, 0x820, 0x10A8, 0x10A4, 0x1D24, 0xF122, 0x1222, 0x1020, 0x50A0, 0x2040}},
    {33643, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x7CFC, 0x4484, 0x4884, 0x50FC, 0x4884, 0x4484, 0x44FC, 0x5484, 0x4884, 0x4114, 0x4208}},
    {33644, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x7FFC, 0x404, 0x288, 0x1280, 0x880, 0x900, 0xFFFE, 0x240, 0x420, 0x1810, 0x6008}},
    {33645, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x1000, 0x21FC, 0x4420, 0xF820, 0x1020, 0x2020, 0xFC20, 0x4020, 0x1C20, 0xE3FE, 0x4000}},
    {33646, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1008, 0x1008, 0x2008, 0x45FE, 0xF808, 0x1088, 0x2048, 0xFC48, 0x4008, 0x1C08, 0xE028, 0x4010}},
    {33647, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1080, 0x1080, 0x20FC, 0x4504, 0xFA04, 0x1084, 0x2044, 0xFC44, 0x4, 0x1C04, 0xE028, 0x4010}},
    {33648, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x1020, 0xFC20, 0x31FC, 0x3820, 0x5420, 0x5020, 0x9020, 0x1020, 0x13FE, 0x1000}},
    {33649, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3FF8, 0x2008, 0x3FF8, 0x2000, 0x2070, 0x2F80, 0x20F0, 0x2F80, 0x20F8, 0x5F82, 0x4082, 0x807E}},
    {33650, 2, {0x440, 0xFFFE, 0x440, 0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE}},
    {33651, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {33652, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x1420, 0x1820, 0x33FE, 0xD050, 0x1050, 0x1088, 0x5104, 0x2202}},
    {33653, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FF8, 0x208, 0x1208, 0x2428, 0x810, 0x3100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {33654, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x40, 0x40, 0x78A0, 0x4910, 0x4A88, 0x4C46, 0x4840, 0x4BF0, 0x7810, 0x4820, 0x20, 0x40}},
    {33655, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x17FE, 0x2008, 0x6008, 0xA7C8, 0x2448, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {33656, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x3FF8, 0x100, 0x7FFE, 0x4002, 0x9FE4, 0x40, 0x180, 0xFFFE, 0x100, 0x500, 0x200}},
    {33657, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x100, 0x908, 0x1110, 0x6060, 0x180, 0xE00, 0xF000}},
    {33658, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x1080, 0x10FC, 0x7D04, 0x1204, 0x1084, 0x1054, 0x1C24, 0xE044, 0x4184, 0x28, 0x10}},
    {33659, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x4420, 0x2820, 0x10A4, 0x28A4, 0x48A8, 0x9920, 0x2820, 0x4850, 0x8850, 0x888, 0x5104, 0x2202}},
    {33660, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0x820, 0x37D8, 0xC106, 0x100, 0x3FF8, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {33661, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xF8, 0x7F00, 0x2210, 0x1120, 0x200, 0xFFFE, 0x420, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {33662, 2, {0x820, 0xFFFE, 0x820, 0x400, 0x820, 0x1010, 0x3FF8, 0x448, 0x820, 0x1210, 0x67E8, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {33663, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x90, 0x88, 0x3FFC, 0x2080, 0x2088, 0x3E88, 0x2250, 0x2250, 0x2A24, 0x4454, 0x418C, 0x8604}},
    {33664, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2000, 0x2FF0, 0x2000, 0x3FFC, 0x2488, 0x2450, 0x4420, 0x4518, 0x8606, 0x400}},
    {33665, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x1110, 0x1110, 0x1110, 0x2928, 0x4544, 0x8584, 0x100, 0xFFFE, 0x0}},
    {33666, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3F04, 0x2104, 0x2124, 0x3F24, 0x824, 0x824, 0x7FA4, 0x8A4, 0x10A4, 0x1084, 0x2294, 0x4108}},
    {33667, 2, {0x440, 0xFFFE, 0x440, 0x3010, 0xCE0, 0x300, 0xCC0, 0x3230, 0xFFFE, 0x800, 0x1100, 0x3FF0, 0xD110, 0x1150, 0x1120, 0x100}},
    {33668, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2100, 0x1A00, 0x600, 0x19C0, 0x603E}},
    {33669, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x880, 0x840, 0x1040, 0x33FC, 0x5000, 0x9008, 0x1208, 0x1110, 0x1110, 0x1020, 0x17FE, 0x1000}},
    {33670, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x110, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x2128, 0x2110}},
    {33671, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x3E20, 0x2220, 0x22FC, 0x3E24, 0x2224, 0x3E24, 0x2224, 0x2244, 0x2F44, 0xF094, 0x4108}},
    {33672, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x21F0, 0x1110, 0x1110, 0x820C, 0x4400, 0x53F0, 0x1110, 0x2110, 0xE0A0, 0x2040, 0x21B0, 0x2E0E}},
    {33673, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x608, 0x7808, 0x848, 0x848, 0xFF48, 0x1C48, 0x2A48, 0x4948, 0x8808, 0x828, 0x810}},
    {33674, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x2420, 0x2420, 0x2420, 0x3C20, 0x7FE, 0x420, 0xFC20, 0x2420, 0x2420, 0x25FC, 0x4400}},
    {33675, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x880, 0x880, 0x10FE, 0x1140, 0x3240, 0x547C, 0x9040, 0x1040, 0x107E, 0x1040, 0x1040, 0x1040}},
    {33676, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x842, 0x1042, 0x603E}},
    {33677, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x110, 0x108, 0x7FFC, 0x100, 0x2118, 0x11A0, 0x740, 0x1920, 0xE118, 0x4106, 0x500, 0x200}},
    {33678, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1020, 0x820, 0x40A8, 0x24A4, 0x522, 0x922, 0x7228, 0x1010, 0x1060, 0x1180, 0x600}},
    {33679, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x1020, 0xFC20, 0x10A8, 0x14A4, 0x1922, 0x3122, 0xD228, 0x1010, 0x1060, 0x5180, 0x2600}},
    {33680, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1040, 0x13FC, 0x8244, 0x4448, 0x10A0, 0x10A0, 0xE120, 0x2120, 0x2222, 0x2422, 0x281E}},
    {33681, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE10E, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100}},
    {33682, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33683, 2, {0x440, 0x440, 0xFFFE, 0x1440, 0x1000, 0x3FFC, 0x4000, 0x9FF0, 0x1210, 0x1110, 0xFFFE, 0x2210, 0x2110, 0x3FFC, 0x10, 0x60}},
    {33684, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FF8, 0x2008, 0x2448, 0x2828, 0x3018, 0x27C8, 0x2448, 0x2448, 0x2448, 0x3FF8, 0x2008}},
    {33685, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33686, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x888, 0x1110, 0x2220, 0x1110, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {33687, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100}},
    {33688, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100}},
    {33689, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x408, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0x4808, 0x8FF8, 0x808}},
    {33690, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x18, 0x7BE0, 0x840, 0x1040, 0x2278, 0x7A40, 0xA40, 0x2BFC, 0x1000, 0x2800, 0x47FE, 0x8000}},
    {33691, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x38, 0x7BC0, 0x840, 0x1040, 0x23F8, 0x7840, 0x840, 0x2BFC, 0x1000, 0x2800, 0x47FE, 0x8000}},
    {33692, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1040, 0x1040, 0x247E, 0x2488, 0x6488, 0xA548, 0x2450, 0x2420, 0x2050, 0x2088, 0x2306}},
    {33693, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x1110, 0x1110, 0x1110, 0x2928, 0x4544, 0x8100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0}},
    {33694, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x3FFC, 0x2004, 0x4008, 0x1FE0, 0x0, 0x7FFC, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {33695, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0x920, 0x3098, 0xCFE6, 0x40, 0x80, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33696, 2, {0x440, 0x440, 0xFFFE, 0x470, 0x1F80, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xCFC6, 0x440, 0x4F0, 0x810, 0x1050, 0x6020}},
    {33697, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {33698, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x100, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33699, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x7FFC, 0x800, 0x800, 0xFF8, 0x0, 0x100, 0x884, 0x4892, 0x4812, 0x87F0, 0x0}},
    {33700, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0xFFFE, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2838, 0x3008, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {33701, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4400, 0x2BF8, 0x1048, 0x2848, 0x4848, 0x9BF8, 0x2848, 0x4888, 0x8888, 0x888, 0x57FE, 0x2000}},
    {33702, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x1110, 0x920, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {33703, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x842, 0x1042, 0x603E}},
    {33704, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1104, 0x1088, 0x1050, 0x1220, 0x1418, 0x1806}},
    {33705, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xF8, 0x7F00, 0x2210, 0x1120, 0x3FE0, 0x40, 0x80, 0xFFFE, 0x100, 0x100, 0x500, 0x200}},
    {33706, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x640, 0x3850, 0x848, 0xFFFE, 0x840, 0x840, 0xE24, 0x7828, 0x812, 0x82A, 0x28C6, 0x1302}},
    {33707, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x7FFC, 0x280, 0x440, 0x1830, 0xE00E}},
    {33708, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x7E0, 0x820, 0x1040, 0x3FF8, 0x5108, 0x1108, 0x1FF8, 0x280, 0x482, 0x1882, 0x607E}},
    {33709, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7E7C, 0x4244, 0x7E44, 0x4244, 0x7E44, 0x4044, 0x5054, 0x4848, 0x5440, 0x6240, 0x4040}},
    {33710, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x200, 0x7FFC, 0x404, 0x804, 0x1028, 0x6010}},
    {33711, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x2040, 0x1040, 0x97FE, 0x40E0, 0x4150, 0x1150, 0x2248, 0xE444, 0x2842, 0x2040, 0x2040}},
    {33712, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1080, 0x1080, 0x10FC, 0x7D04, 0x1248, 0x1040, 0x1040, 0x1CA0, 0xE0A0, 0x4110, 0x208, 0xC06}},
    {33713, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x3FF8, 0x100, 0x1110, 0x920, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33714, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2080, 0x17FC, 0x1140, 0x240, 0x73FC, 0x1040, 0x1040, 0x17FE, 0x1040, 0x2840, 0x47FE, 0x0}},
    {33715, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x10, 0x7C10, 0x4410, 0x45FE, 0x4410, 0x7D10, 0x4490, 0x4490, 0x4410, 0x7C10, 0x4450, 0x20}},
    {33716, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x7FFC, 0x4104, 0x4284, 0x4444, 0x4844, 0x4014, 0x4008}},
    {33717, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x200, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2288, 0x490, 0x888, 0x108A, 0x2082, 0xC07E}},
    {33718, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0x440, 0x820, 0x37D8, 0xC006, 0x210, 0x1110, 0x920, 0x840, 0x7FFC, 0x0}},
    {33719, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4440, 0x2850, 0x1048, 0x2840, 0x4FFE, 0x9840, 0x28A0, 0x48A0, 0x8910, 0x910, 0x5208, 0x2406}},
    {33720, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x4490, 0x2888, 0x1088, 0x2BFE, 0x48A0, 0x98A0, 0x28A0, 0x4920, 0x8920, 0xA22, 0x5222, 0x241E}},
    {33721, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4002, 0x8004, 0x3FF8, 0x100, 0x100, 0x1FF0, 0x100, 0x140, 0x120, 0x7FFC}},
    {33722, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4202, 0x9FF4, 0x1010, 0x1210, 0x1150, 0x1020, 0x1FFC, 0x4, 0x4, 0x7FD4, 0x8}},
    {33723, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x3FE0, 0x420, 0x47C, 0x804, 0x3014, 0xC008, 0x3E44, 0x2258, 0x2262, 0x3E42, 0x223E}},
    {33724, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x23FE, 0x4420, 0x7924, 0x1124, 0x2124, 0x7DFC, 0x22, 0x1C22, 0xE022, 0x401E}},
    {33725, 2, {0x820, 0xFFFE, 0x820, 0x240, 0x220, 0x7FFC, 0x440, 0x820, 0x3018, 0xC826, 0x820, 0x7FFC, 0x820, 0x1020, 0x2020, 0x4020}},
    {33726, 2, {0x820, 0xFFFE, 0xA20, 0x200, 0x7FFC, 0x440, 0x920, 0x37D8, 0xC106, 0x920, 0x820, 0x7FFC, 0x820, 0x820, 0x1020, 0x2020}},
    {33727, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x804, 0xFFA4, 0x824, 0x7F24, 0x4924, 0x4B24, 0x1C24, 0x2A24, 0x4904, 0x8814, 0x808}},
    {33728, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x9004, 0x1EF8, 0x2288, 0x2488, 0x54A8, 0x890, 0x884, 0x1084, 0x607C}},
    {33729, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33730, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x2040, 0x7E7C, 0x4284, 0x4304, 0x4244, 0x7E24, 0x4224, 0x4204, 0x4204, 0x7E28, 0x4210}},
    {33731, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2000, 0x13FE, 0x9200, 0x4200, 0x4BFC, 0xA04, 0x1204, 0xE3FC, 0x2200, 0x2200, 0x23FE, 0x2000}},
    {33732, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x380, 0xD60, 0x3118, 0xC106}},
    {33733, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x7FFE, 0x4002, 0x9FE4, 0x1020, 0x1FE0, 0x1000, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33734, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7F00, 0x22FC, 0x2244, 0x3E44, 0x2244, 0x2228, 0x3E28, 0x2210, 0x2790, 0xFA28, 0x4244, 0x282}},
    {33735, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x1010, 0xFEFE, 0x1210, 0x2210, 0x227C, 0x6444, 0x1444, 0x844, 0x1444, 0x227C, 0xC044}},
    {33736, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x1020, 0xFC20, 0x11FC, 0x1400, 0x1808, 0x3108, 0xD090, 0x1090, 0x1020, 0x53FE, 0x2000}},
    {33737, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3FF0, 0x10, 0x1FF0, 0x10, 0xFFFE, 0x1110, 0x9A0, 0x540, 0x1930, 0xE10E, 0x500, 0x200}},
    {33738, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1FFC, 0x2204, 0x5244, 0x8A84, 0x204, 0x7FF4, 0x704, 0xA84, 0x1244, 0x6214, 0x208}},
    {33739, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x40, 0x40, 0x7BF8, 0x4840, 0x4840, 0x4FFC, 0x48E0, 0x7950, 0x4A48, 0x444, 0x842, 0x40}},
    {33740, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2008, 0x20C8, 0x2708, 0x2108, 0x2FE8, 0x2388, 0x2548, 0x2928, 0x2108, 0x3FF8, 0x2008}},
    {33741, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0x920, 0x3098, 0xCFE6, 0x40, 0x280, 0x100, 0x4884, 0x4812, 0x87F2, 0x0}},
    {33742, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x2080, 0x208C, 0x3EF0, 0x2082, 0x2682, 0x387E}},
    {33743, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2000, 0x17FE, 0x1008, 0x8008, 0x4BC8, 0x4A48, 0x1248, 0xE3C8, 0x2248, 0x2008, 0x2028, 0x10}},
    {33744, 2, {0x440, 0x2448, 0x1450, 0x440, 0xFFFE, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x280, 0x440, 0x1830, 0xE00E}},
    {33745, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {33746, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {33747, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {33748, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3EFC, 0x2284, 0x2294, 0x3E88, 0x2280, 0x22FC, 0x3EA4, 0x22A4, 0x22A8, 0x2290, 0x4AA8, 0x84C6}},
    {33749, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x2488, 0x2488, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {33750, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {33751, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x1124, 0x1DFC, 0xF124, 0x1124, 0x1124, 0x51FC, 0x2104}},
    {33752, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x1090, 0x1090, 0xFD08, 0x3148, 0x3A44, 0x5442, 0x5090, 0x9088, 0x1108, 0x13FC, 0x1104}},
    {33753, 2, {0x820, 0xFFFE, 0x820, 0xF8, 0x3F00, 0x100, 0x7FFC, 0x920, 0x920, 0xFFFE, 0x920, 0x920, 0x7FFC, 0x100, 0x100, 0x3FF8}},
    {33754, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x210, 0x2FC, 0x7F40, 0x84, 0x764, 0x781C, 0x210, 0x2FC, 0x7F40, 0x84, 0x764, 0x781C}},
    {33755, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {33756, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x10, 0xF8, 0x3F00, 0x1110, 0x820, 0x100, 0x7FFC, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33757, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1048, 0x1044, 0xFDFE, 0x1040, 0x1440, 0x18FC, 0x30A4, 0xD128, 0x1128, 0x1210, 0x5428, 0x20C6}},
    {33758, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xC80, 0x7080, 0x10FC, 0x1124, 0xFD24, 0x3224, 0x3844, 0x5444, 0x5084, 0x9104, 0x1228, 0x1010}},
    {33759, 2, {0x440, 0xFFFE, 0x440, 0x800, 0xFE0, 0x1020, 0x2040, 0x5FF8, 0x1108, 0x1108, 0x1FF8, 0x280, 0x4A0, 0x892, 0x1082, 0x607E}},
    {33760, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2020, 0x1020, 0x13FC, 0x8224, 0x4220, 0x4BF8, 0xA88, 0x1288, 0xE250, 0x2420, 0x2450, 0x298C}},
    {33761, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x1FE0, 0x4044, 0x5194, 0x4924, 0x43C4, 0x4544, 0x4924, 0x5514, 0x4204, 0x7FFC, 0x4}},
    {33762, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1080, 0x1080, 0xFDFC, 0x1204, 0x15F4, 0x1114, 0x1D14, 0xF1F4, 0x110C, 0x1102, 0x5102, 0x20FE}},
    {33763, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7F7C, 0x4844, 0x4844, 0x7F44, 0x4128, 0x4128, 0x7F10, 0x4810, 0x4828, 0x7F44, 0x82}},
    {33764, 2, {0x440, 0x440, 0xFFFE, 0x540, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1010, 0x2FE8, 0xC826, 0x868, 0x808, 0x7F8}},
    {33765, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x100C, 0x11F0, 0xFD00, 0x1100, 0x31FE, 0x3910, 0x5510, 0x5110, 0x9210, 0x1210, 0x1410}},
    {33766, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2018, 0x13E0, 0x1200, 0x200, 0x73FC, 0x1220, 0x1220, 0x1220, 0x1420, 0x2820, 0x47FE, 0x0}},
    {33767, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFE, 0x2000, 0x2078, 0x2F80, 0x2880, 0x2880, 0x2FFC, 0x2840, 0x4824, 0x4A94, 0x8C4C}},
    {33768, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x400, 0x7FFC, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {33769, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {33770, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x3FF0, 0x400, 0x800, 0x1FF0, 0x2810, 0x4810, 0x810, 0xFF0, 0x810}},
    {33771, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0}},
    {33772, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23FC, 0x1084, 0x1084, 0x8084, 0x4914, 0x4A08, 0x15FC, 0xE104, 0x2104, 0x2104, 0x21FC, 0x2104}},
    {33773, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2080, 0x1110, 0x1208, 0x87FC, 0x4204, 0x5000, 0x13F8, 0x2208, 0xE208, 0x2208, 0x23F8, 0x2208}},
    {33774, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2140, 0x4220, 0x4C18, 0xB006}},
    {33775, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x7FFC, 0x1110, 0x1110, 0xFFFE, 0x1110, 0x1110, 0x7FFC, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {33776, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1C, 0x7DF0, 0x950, 0x1150, 0x1150, 0x1D50, 0x3148, 0xD148, 0x1144, 0x1254, 0x525A, 0x2468}},
    {33777, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1010, 0x2208, 0x47E4, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {33778, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x440, 0x7C7C, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0xFC7E, 0x440, 0x440, 0x440}},
    {33779, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x920, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE, 0x0}},
    {33780, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x200, 0x7FFC, 0x440, 0x920, 0x3FF8, 0xD116, 0x1FF0, 0x1110, 0x1FF4, 0x104, 0x104, 0xFC}},
    {33781, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x7FFC, 0x4824, 0x4444, 0x5FF4, 0x4104, 0x4084, 0x5FF4, 0x4404, 0x4404, 0x47E4, 0x4014, 0x4008}},
    {33782, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x920, 0x17D0, 0x2108, 0xDFF6, 0x100, 0x100}},
    {33783, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x10, 0xFF0, 0x10, 0x1FF0, 0x0, 0x7FFE, 0x4102, 0x9FF4, 0x1110, 0x1150, 0x1120, 0x100}},
    {33784, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x820, 0x7E50, 0x1050, 0x1088, 0x1F06, 0x1260, 0x1210, 0x2200, 0x22C0, 0x4A30, 0x8408}},
    {33785, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x9F8, 0x908, 0x4108, 0x21F8, 0x2908, 0x908, 0x11F8, 0x7108, 0x1108, 0x17FE, 0x1000}},
    {33786, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x1FFC, 0x1004, 0x1FFC, 0x1000, 0x17F8, 0x1408, 0x17F8, 0x2408, 0x27F8, 0x4408, 0x8428, 0x410}},
    {33787, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x820, 0x820, 0x7EFC, 0x820, 0x1C30, 0x1A70, 0x2A68, 0x48A4, 0x8922, 0x820, 0x820}},
    {33788, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x910, 0x920, 0x12C0, 0x430, 0x1908, 0x6914, 0x910, 0x12A0, 0x440, 0x1830, 0x600C}},
    {33789, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0xF7C, 0x844, 0x844, 0xFF44, 0x28, 0x828, 0x2A10, 0x4910, 0x8928, 0x2844, 0x1082}},
    {33790, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x100, 0x200, 0xFFFE, 0x820, 0x3118, 0xC106, 0x990, 0x1148, 0x2148, 0x500, 0x200}},
    {33791, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7F84, 0x1004, 0x2224, 0x7F24, 0x924, 0x824, 0x7F24, 0x824, 0x824, 0xF04, 0xF014, 0x4008}},
    {33792, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x27E0, 0x4422, 0x4822, 0xB01E}},
    {33793, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {33794, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xE00, 0x787C, 0x844, 0x844, 0xFF44, 0x1844, 0x1C44, 0x2A44, 0x4A44, 0x887C, 0x844, 0x800}},
    {33795, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x820, 0x1450, 0x2288, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {33796, 2, {0x820, 0xFFFE, 0x820, 0x2000, 0x3FFC, 0x4804, 0x9FE4, 0x2204, 0x204, 0xFFFC, 0x204, 0x2224, 0x2224, 0x3FE4, 0x28, 0x10}},
    {33797, 2, {0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x100, 0x2108, 0x3FFA, 0x102, 0xFE, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8}},
    {33798, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x400, 0x3FF8, 0x2108, 0x3FF8, 0x2208, 0x3FF8, 0x880, 0x1080, 0xFFFE, 0x80, 0x80, 0x80}},
    {33799, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xFF8, 0x800, 0xFF0, 0x800, 0xFF0, 0x800, 0xFFFE, 0x1110, 0x10A0, 0x1440, 0x1830, 0x100E}},
    {33800, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x450, 0x84A, 0x1042, 0x603E}},
    {33801, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3EFC, 0x22A4, 0x22A4, 0x3EA4, 0x22FC, 0x2280, 0x3E80, 0x2282, 0x2282, 0x4A7E, 0x8400}},
    {33802, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x7FFC, 0x1110, 0x1110, 0x2928, 0x4544, 0x380, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {33803, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x7C38}},
    {33804, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0xFC, 0x7C84, 0x4484, 0x44FC, 0x7C84, 0x4484, 0x44FC, 0x7C84, 0x4484, 0x114, 0x208}},
    {33805, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2000, 0x17FC, 0x9040, 0x4248, 0x4950, 0x840, 0x17FE, 0xE040, 0x2040, 0x2040, 0x2040, 0x40}},
    {33806, 2, {0x820, 0xFFFE, 0x820, 0x70, 0x1F80, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x400, 0x7FFC, 0x820, 0x1C40, 0x380, 0x7C78}},
    {33807, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x1FE0, 0x2020, 0x4440, 0xB878, 0x2008, 0x2008, 0x3C78, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33808, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x1100, 0x11F8, 0x2900, 0x47FE, 0x8000}},
    {33809, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x880, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000}},
    {33810, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x1210, 0x2448, 0x820, 0x1FF0, 0x440, 0x440, 0x844, 0x1044, 0x603C}},
    {33811, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0xFFFE}},
    {33812, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x11FC, 0xFC44, 0x1044, 0x1494, 0x1908, 0x32FC, 0xD084, 0x1084, 0x1084, 0x50FC, 0x2084}},
    {33813, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE82E, 0xFE0, 0x820, 0xFE0, 0x820, 0x1020, 0x2020}},
    {33814, 2, {0x820, 0xFFFE, 0x820, 0x600, 0x38F8, 0x2008, 0x2008, 0x3CF8, 0x2008, 0x2008, 0x3FF8, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {33815, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x1110, 0x2108, 0x4504, 0x200}},
    {33816, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x200, 0x7FFC, 0x440, 0x820, 0x37D8, 0xC006, 0x3FF8, 0x100, 0x920, 0x1110, 0x2508, 0x200}},
    {33817, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0xFDFE, 0x1122, 0x3224, 0x3850, 0x5450, 0x5490, 0x9090, 0x1112, 0x1212, 0x140E}},
    {33818, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x11FC, 0x1088, 0xFC70, 0x138E, 0x1020, 0x1DFC, 0x3020, 0xD3FE, 0x1020, 0x5020, 0x2020}},
    {33819, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {33820, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0xFE3E, 0x9220, 0x9220, 0x92FC, 0x9284, 0x9A84, 0x9484, 0x1084, 0x10FC, 0x1084}},
    {33821, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x200, 0x7F0, 0x810, 0x3420, 0x240, 0x180, 0xE00, 0x7000}},
    {33822, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2288, 0x2448, 0x0, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {33823, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x640, 0x3840, 0x9F8, 0x848, 0x7E48, 0x18C8, 0x1C48, 0x2A68, 0x289A, 0x488A, 0x906, 0xA02}},
    {33824, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x4A94, 0x8508}},
    {33825, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x1040, 0x9FC, 0x4904, 0x2104, 0x2504, 0x5FC, 0x904, 0x7104, 0x1104, 0x11FC, 0x1104}},
    {33826, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2100, 0x1100, 0x13F8, 0x8408, 0x4BE8, 0x5228, 0x1228, 0x23E8, 0xE218, 0x2204, 0x2204, 0x21FC}},
    {33827, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x100, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {33828, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x108, 0x7FFC, 0x4}},
    {33829, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {33830, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4402, 0x8844, 0x1F80, 0x320, 0xC10, 0x3FF8, 0x108, 0x1120, 0x2510, 0x4208}},
    {33831, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x2548, 0x2528, 0x2928, 0x4108, 0x8108}},
    {33832, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x7DFC, 0x4488, 0x4850, 0x51FE, 0x4900, 0x4500, 0x4500, 0x5500, 0x4900, 0x4200, 0x4400}},
    {33833, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xC20, 0x7020, 0x10A4, 0x10A4, 0xFCA8, 0x3120, 0x3820, 0x5450, 0x5050, 0x9088, 0x1104, 0x1202}},
    {33834, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xC10, 0x7090, 0x1050, 0x1010, 0xFC90, 0x3050, 0x381E, 0x55F0, 0x5010, 0x9010, 0x1010, 0x1010}},
    {33835, 2, {0x820, 0xFFFE, 0x820, 0xF8, 0x3F00, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {33836, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2148, 0x2FE8, 0x2428, 0x2010}},
    {33837, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xF0, 0x3F00, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFC, 0x4124, 0x41F4, 0x5E14, 0x4008}},
    {33838, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0xC60, 0x37D8, 0xC006, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2228, 0x2610}},
    {33839, 2, {0x820, 0xFFFE, 0x820, 0x400, 0xFE0, 0x1020, 0x3FF8, 0x5008, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x420, 0x810, 0x1008}},
    {33840, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x380, 0x560, 0x1918, 0xE106, 0x100}},
    {33841, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0xFFFE}},
    {33842, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xFC, 0x7F00, 0x2208, 0x1110, 0x3FF8, 0x200, 0xFFFE, 0x400, 0xFF0, 0x1220, 0x21C0, 0xCE3C}},
    {33843, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2448, 0x2288, 0x2FE8, 0x2108, 0x3FF8, 0x2108, 0x2128, 0x2010}},
    {33844, 2, {0x820, 0xFFFE, 0x820, 0x3F04, 0x2104, 0x2124, 0x3F24, 0x2124, 0x3F24, 0x2124, 0x2124, 0x3F24, 0x24, 0x1204, 0x2114, 0x4088}},
    {33845, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1F90, 0x1090, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {33846, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x820}},
    {33847, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x4904, 0x2A24, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x7F24, 0x4124, 0x4104, 0x4514, 0x4208}},
    {33848, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x500, 0x3978, 0x2108, 0x2108, 0x3D78, 0x2108, 0x2108, 0x3FF8, 0x280, 0x440, 0x1830, 0xE00E}},
    {33849, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2000, 0x2FFC, 0x2924, 0x2924, 0x2FFC, 0x4924, 0x4924, 0x880C}},
    {33850, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4004, 0x5FF4, 0x4004, 0x5FF4, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {33851, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {33852, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3EF8, 0x2288, 0x3EF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x800, 0xFF0, 0x10, 0xA0, 0x40}},
    {33853, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2080, 0x10F8, 0x1108, 0x8290, 0x4860, 0x4998, 0x1606, 0xE1F8, 0x2108, 0x2108, 0x21F8, 0x2108}},
    {33854, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FE0, 0x820, 0xBF8, 0xA88, 0x1128, 0x2290, 0x5FF8, 0x1248, 0x1248, 0x1248, 0xFFFE, 0x0}},
    {33855, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2038, 0x13C0, 0x1040, 0x8040, 0x47FC, 0x5040, 0x1040, 0x23F8, 0xE208, 0x2208, 0x23F8, 0x2208}},
    {33856, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1038, 0x13C0, 0xFC40, 0x1040, 0x17FC, 0x1840, 0x3040, 0xD3F8, 0x1208, 0x1208, 0x53F8, 0x2208}},
    {33857, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x200, 0xFFFE, 0x820, 0x1C40, 0x3C0, 0x3C38}},
    {33858, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2020, 0x3E20, 0x4420, 0x88FC, 0x7F24, 0x4924, 0x4924, 0x7F44, 0x1454, 0x248A, 0x4502, 0x83FE}},
    {33859, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x80, 0x7C80, 0x44FE, 0x4540, 0x4640, 0x7C7C, 0x4440, 0x4440, 0x447E, 0x7C40, 0x4440, 0x40}},
    {33860, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x40, 0x3E40, 0x227E, 0x22A0, 0x3EA0, 0x233C, 0x2220, 0x3E20, 0x223E, 0x2220, 0x4A20, 0x8420}},
    {33861, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x11F8, 0x1108, 0x2908, 0x45F8, 0x9108, 0x1108, 0x29F8, 0x2508, 0x4108, 0x87FE, 0x0}},
    {33862, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0xBF8, 0x1208, 0x3208, 0x53F8, 0x9040, 0x17FE, 0x10E0, 0x1150, 0x1248, 0x1C46, 0x1040}},
    {33863, 2, {0x440, 0xFFFE, 0x440, 0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {33864, 2, {0x440, 0x440, 0xFFFE, 0x1440, 0x2008, 0x7FFC, 0x4, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {33865, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1220, 0x7FFC, 0x1220, 0x13E0, 0x1000, 0x1FF8, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {33866, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {33867, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x80, 0x3C80, 0x24FC, 0x2504, 0x3EF4, 0x2494, 0x2494, 0x3CF4, 0x2494, 0x2404, 0x5428, 0x8810}},
    {33868, 2, {0x820, 0xFFFE, 0x820, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38, 0x1010, 0xFEFE, 0x2224, 0x6464, 0x1818, 0x2434, 0xC2C2}},
    {33869, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {33870, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x40, 0x13F8, 0x2048, 0x47FE, 0x8848, 0x13F8, 0x3040, 0x53FC, 0x9040, 0x17FE, 0x1040, 0x1040}},
    {33871, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x13F8, 0x9048, 0x4FFE, 0x4048, 0x13F8, 0x2040, 0xE3FC, 0x2040, 0x2FFE, 0x2040, 0x2040}},
    {33872, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x820, 0x3018, 0xCFE6, 0x420, 0x18A0, 0x6040, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE}},
    {33873, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x808, 0x7F08, 0x8FE, 0xFF08, 0x848, 0x828, 0x7F28, 0x808, 0xF08, 0xF028, 0x4010}},
    {33874, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x1000, 0x25FC, 0x7C20, 0x820, 0x1020, 0x2420, 0x7E20, 0x20, 0x5420, 0x4BFE, 0x8A00}},
    {33875, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x2090, 0x1090, 0x13FC, 0x8090, 0x4890, 0x4890, 0x17FE, 0xE000, 0x2090, 0x2108, 0x2204}},
    {33876, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x9F0, 0x810, 0x1010, 0x37FC, 0x5100, 0x91F8, 0x1240, 0x1040, 0x17FC, 0x10A0, 0x1110, 0x120C}},
    {33877, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1200, 0x213C, 0x4900, 0x9400, 0x247E, 0x6108, 0xA208, 0x2E08, 0x2208, 0x2208, 0x2228, 0x2010}},
    {33878, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4822, 0x9114, 0x2128, 0x110, 0x7FFC, 0x280, 0x440, 0x820, 0x3018, 0xC006}},
    {33879, 2, {0x820, 0xFFFE, 0x820, 0x208, 0x3FD0, 0x220, 0x240, 0xFFFE, 0x300, 0xC00, 0x3FF0, 0xC810, 0xFF0, 0x810, 0xFF0, 0x810}},
    {33880, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x888, 0x1110, 0xFFFE, 0x1110, 0x888, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {33881, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x8FC, 0x7E84, 0x884, 0x18FC, 0x1C84, 0x2A84, 0x2AFC, 0x4884, 0x8884, 0x8FC, 0x884}},
    {33882, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1240, 0x1420, 0x1000, 0x1FF8}},
    {33883, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2104, 0x5284, 0x9444, 0x1004, 0x1FA8, 0x10}},
    {33884, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x100, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006}},
    {33885, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x7F20, 0x2A20, 0x54FC, 0x2A24, 0x24, 0x7E24, 0x824, 0x844, 0xE44, 0xF094, 0x4108}},
    {33886, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7DFC, 0x488, 0x488, 0x7CF8, 0x4088, 0x4088, 0x7CF8, 0x488, 0x49E, 0x5E8, 0x2808, 0x1008}},
    {33887, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {33888, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xBF8, 0x808, 0x11F8, 0x3008, 0x53F8, 0x9000, 0x17FC, 0x1404, 0x13F8, 0x1110, 0x10E0, 0x171C}},
    {33889, 2, {0x440, 0x440, 0xFFFE, 0x2440, 0x3FFC, 0x4224, 0x7FF4, 0x8204, 0x3FE4, 0x2224, 0x3FE4, 0x2224, 0x3FE4, 0x2224, 0x22A4, 0x204C}},
    {33890, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x100, 0x7FFC, 0x480, 0xA60, 0x3118, 0xC006, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {33891, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {33892, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1008, 0x1008, 0x2408, 0x7DFE, 0x808, 0x1088, 0x2448, 0x7E48, 0x8, 0x5408, 0x4A28, 0x8A10}},
    {33893, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x820, 0xFFFE, 0x0, 0x3E48, 0x2248, 0x3E48, 0x2248, 0x3E08, 0x2208, 0x2A28, 0x2410}},
    {33894, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x3FF0, 0x410, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x1FF0, 0x2100, 0x3FF8, 0x100}},
    {33895, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x810, 0x7F10, 0x810, 0x7F7E, 0x4112, 0x3E12, 0x412, 0xF12, 0x7822, 0x822, 0x284A, 0x1084}},
    {33896, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE, 0x1110, 0x10A0, 0x1440, 0x1830, 0x100E}},
    {33897, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x21FC, 0x7D24, 0x4524, 0x4524, 0x4524, 0x7DFC, 0x4500, 0x4500, 0x4502, 0x7D02, 0x44FE}},
    {33898, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1004, 0x1F04, 0x2224, 0x7FA4, 0xA4A4, 0x3FA4, 0x24A4, 0x3FA4, 0x24A4, 0x2484, 0x4294, 0x8108}},
    {33899, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x87C, 0x7F44, 0x844, 0x87C, 0x3E44, 0x2244, 0x227C, 0x2244, 0x3E84, 0x2294, 0x108}},
    {33900, 2, {0x820, 0xFFFE, 0x820, 0xFFFE, 0x1080, 0x3E98, 0x42E0, 0x2484, 0x1884, 0xE07C, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {33901, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7CF8, 0x4408, 0x4408, 0x7CF8, 0x4000, 0x41F8, 0x7C88, 0x4050, 0x4020, 0x7C50, 0x4088, 0x4306}},
    {33902, 2, {0x820, 0xFFFE, 0x820, 0x600, 0x38F8, 0x2088, 0x2088, 0x3D06, 0x2200, 0x21F8, 0x3C88, 0x2088, 0x2C50, 0xF020, 0x20D8, 0x2706}},
    {33903, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x1440, 0x247C, 0x7884, 0x1104, 0x2444, 0x7E24, 0x24, 0x5404, 0x4A04, 0x8A28, 0x10}},
    {33904, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x910, 0xA08, 0x17FC, 0x3110, 0x5208, 0x9484, 0x11F8, 0x1308, 0x1490, 0x1060, 0x1198, 0x1606}},
    {33905, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x1FF8, 0x2A48, 0x4448, 0xA88, 0x1108, 0x22A8, 0x410, 0x100, 0x4884, 0x4812, 0x87F2}},
    {33906, 2, {0x820, 0xFFFE, 0xA20, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x104, 0x7D88, 0x950, 0x1120, 0x2118, 0xC506, 0x200}},
    {33907, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2224, 0x2FA8, 0x2498, 0x2D12, 0x422A, 0x4546, 0x8882}},
    {33908, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x28, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2024, 0x2FA8, 0x2898, 0x2892, 0x4FAA, 0x4046, 0x8082}},
    {33909, 2, {0x820, 0xFFFE, 0x820, 0x90, 0x7EA0, 0x444, 0x2828, 0x1010, 0x2FE8, 0xC106, 0x100, 0x7FFC, 0x100, 0x2C0, 0xC30, 0x7008}},
    {33910, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x100, 0x100, 0x500, 0x200}},
    {33911, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4104, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {33912, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x200, 0x5104, 0x5112, 0x9012, 0xFF0}},
    {33913, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1040, 0xFEFE, 0x2120, 0x20AC, 0x3CB4, 0x25E4, 0x24AC, 0x24A0, 0x44A2, 0x5482, 0x887E}},
    {33914, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x103E, 0xFFD0, 0x10, 0x10}},
    {33915, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x0, 0x1FF0, 0x1010, 0x10D0, 0x1710, 0x1110, 0x17D0, 0x1550, 0x27CA, 0x212A, 0x4FE6, 0x8022}},
    {33916, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x1010, 0x2208, 0x47E4, 0x820, 0x1440, 0x380, 0x1C70, 0xE00E}},
    {33917, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x3008}},
    {33918, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1000, 0x1EF8, 0x2288, 0x62A8, 0x9492, 0x882, 0x307E, 0xC100, 0x884, 0x4892, 0x4812, 0x87F0}},
    {33919, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x3FF8, 0x2108, 0x3FF8, 0x2000, 0x2FF8, 0x2808, 0x2FF8, 0x2808, 0x4FF8, 0x4808, 0x8FF8, 0x808}},
    {33920, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {33921, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2050, 0x1048, 0x17FC, 0xC0, 0x160, 0xF150, 0x1248, 0x1444, 0x1040, 0x2800, 0x47FE, 0x0}},
    {33922, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x3FF8, 0x820, 0x440, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1150, 0x1120, 0x100, 0x100}},
    {33923, 2, {0x440, 0xFFFE, 0x440, 0x800, 0xFE0, 0x1020, 0x1FC0, 0x40, 0xFFFE, 0x600, 0x3B08, 0x4B0, 0x19C0, 0x62A0, 0xC98, 0x7306}},
    {33924, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0x8004, 0x3E08, 0x8, 0x7EFE, 0x1448, 0x1428, 0x140A, 0x242A, 0x2412, 0x43FE}},
    {33925, 2, {0x440, 0xFFFE, 0x440, 0x1080, 0x880, 0x43F0, 0x2890, 0x1112, 0x7212, 0x140E, 0x1100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {33926, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7DF0, 0x1110, 0x1190, 0x1D52, 0xF20E, 0x400, 0xFFFE, 0x820, 0x1C40, 0x380, 0xC70, 0x7008}},
    {33927, 2, {0x440, 0x440, 0xFFFE, 0x448, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x28A4, 0x2AA8, 0x2AA8, 0x2A90, 0x2A92, 0x452A, 0x4946, 0x9082}},
    {33928, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2088, 0x3C90, 0x20E4, 0x2C84, 0x307C, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {33929, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1110, 0x2C0, 0xC30, 0x7008}},
    {33930, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x880, 0x1098, 0x30E0, 0x5784, 0x9084, 0x107C, 0x0, 0x7C88, 0x44B0, 0x44C4, 0x7C84, 0x447C}},
    {33931, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x80, 0x8F8, 0x4910, 0x2AA0, 0x2850, 0x890, 0x1BFE, 0x2910, 0x4890, 0x8890, 0x810, 0x830}},
    {33932, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0x610C, 0x400, 0xFFFE, 0x820, 0x1E40, 0x3C0, 0x3C38}},
    {33933, 2, {0x820, 0xFFFE, 0x820, 0x2400, 0x1400, 0x7FE0, 0x420, 0x820, 0xFF8, 0x808, 0x1008, 0x1FFE, 0x2002, 0x2552, 0x4954, 0x8008}},
    {33934, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x201C, 0x13E0, 0x920E, 0x42F0, 0x4A90, 0xA90, 0x1292, 0xE294, 0x2288, 0x2488, 0x24A4, 0x28C2}},
    {33935, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0xBFF4, 0x480, 0x3FF8, 0x2488, 0x2878, 0x3008, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {33936, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x400, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x290, 0x2A8, 0x4B8, 0x88A, 0x3082, 0xC07E}},
    {33937, 2, {0x820, 0xFFFE, 0x820, 0x600, 0x3878, 0x2048, 0x3E48, 0x2286, 0x3F00, 0x22FC, 0x3E44, 0x2044, 0x3E28, 0x2210, 0x4A28, 0x84C6}},
    {33938, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x23FE, 0x7C20, 0x4420, 0x45FC, 0x7D24, 0x4124, 0x7D24, 0x4524, 0x4534, 0x7D28, 0x4420, 0x20}},
    {33939, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x25FE, 0x7C20, 0x924, 0x1124, 0x2524, 0x7DFC, 0x22, 0x5422, 0x5422, 0x801E}},
    {33940, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x40, 0x7840, 0x4BFC, 0x4840, 0x4FFE, 0x7810, 0x4FFE, 0x4810, 0x4A10, 0x7910, 0x4850, 0x20}},
    {33941, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1110, 0x1290, 0x1450, 0x1FF0, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {33942, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x1010}},
    {33943, 2, {0x820, 0x820, 0xFFFE, 0x8A0, 0x2040, 0x13F8, 0x9208, 0x43F8, 0x4A08, 0xBF8, 0x1242, 0xE244, 0x2228, 0x2290, 0x2308, 0x2206}},
    {33944, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x7F7C, 0x1144, 0x3244, 0xC7C, 0x3220, 0xDFC0, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x300}},
    {33945, 2, {0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0x8FE4, 0x0, 0x7FFC, 0x600, 0x3B08, 0x4B0, 0x19C0, 0x62A0, 0xC98, 0x7286, 0x100}},
    {33946, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4824, 0x4444, 0x5FF4, 0x4104, 0x410C}},
    {33947, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2020, 0x2020, 0x3DFC, 0x5024, 0x1024, 0xFC24, 0x11FE, 0x5420, 0x5450, 0x5C50, 0x7488, 0x506}},
    {33948, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7C7C, 0x0, 0x0, 0xFEFE, 0x1010, 0x1010, 0x5454, 0x5252, 0x9292, 0x5050, 0x2020}},
    {33949, 2, {0x440, 0xFFFE, 0x440, 0x0, 0x3FFC, 0x2080, 0x2100, 0x27F0, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x4080, 0x4490, 0x8A88, 0x1104}},
    {33950, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2140, 0x1120, 0x1200, 0x82FC, 0x4600, 0x5A88, 0x1248, 0x2248, 0xE250, 0x2200, 0x22FC, 0x2200}},
    {33951, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2080, 0x1080, 0x10FC, 0xFD04, 0x2F4, 0x4894, 0x4894, 0x48F4, 0x1094, 0x1E04, 0xF028, 0x4010}},
    {33952, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0}},
    {33953, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x820, 0x440, 0x7FFE, 0x4202, 0x8104, 0x3FF8, 0x400, 0x7F0, 0x810, 0x1050, 0x2020}},
    {33954, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7840, 0x48A0, 0x5110, 0x6208, 0x55F6, 0x4840, 0x4BFC, 0x6840, 0x5150, 0x4248, 0x4544, 0x4080}},
    {33955, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x20A0, 0x4110, 0x8A08, 0x15F6, 0x3040, 0x53FC, 0x9040, 0x1150, 0x1248, 0x1544, 0x1080}},
    {33956, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x10A0, 0x9110, 0x4208, 0x4DF6, 0x1040, 0x27FC, 0xE040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {33957, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x79FC, 0x4044, 0x4844, 0x5454, 0x6288, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008}},
    {33958, 2, {0x440, 0xFFFE, 0x440, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3FF0, 0xC60, 0x380, 0x7C7C}},
    {33959, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x28, 0x24, 0x3FFE, 0x2220, 0x23A4, 0x2224, 0x2FA8, 0x2898, 0x2892, 0x4FAA, 0x4046, 0x8082}},
    {33960, 2, {0x440, 0xFFFE, 0x440, 0x13F8, 0x1040, 0x23F8, 0x2040, 0x6FFE, 0xA000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2218}},
    {33961, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xC00, 0x71F8, 0x1108, 0x1108, 0xFDF8, 0x3108, 0x3908, 0x55F8, 0x5108, 0x9108, 0x17FE, 0x1000}},
    {33962, 2, {0x820, 0xFFFE, 0x920, 0x110, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x20, 0xFFFE, 0x820, 0x460}},
    {33963, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x800, 0x17F8, 0x2080, 0x4080, 0x9FFC, 0x0}},
    {33964, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4402, 0x8FE4, 0x1020, 0x3FF8, 0x5108, 0x1108, 0x1FF8, 0x280, 0x492, 0x188A, 0xE07E}},
    {33965, 2, {0x820, 0xFFFE, 0x820, 0x1000, 0x1FFC, 0x2104, 0x5114, 0x1FF4, 0x208, 0x1400, 0x1FFC, 0x2104, 0x5114, 0x1FF4, 0x414, 0x3808}},
    {33966, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x7FFC, 0x4504, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {33967, 2, {0x1102, 0x1102, 0xFFE2, 0x110A, 0xA, 0x7BCA, 0x4A4A, 0x4A4A, 0x7BCA, 0x4A4A, 0x4A4A, 0x7BCA, 0x4A42, 0x4A42, 0xAB4A, 0x1484}},
    {33968, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1078, 0x3E48, 0x2248, 0x2A86, 0x2300, 0xFEFC, 0x2244, 0x2A44, 0x2228, 0x4210, 0x4A28, 0x84C6}},
    {33969, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1028, 0x1024, 0xFDFE, 0x1020, 0x15FC, 0x1924, 0x31FC, 0xD124, 0x11FC, 0x1124, 0x5124, 0x210C}},
    {33970, 2, {0x820, 0x820, 0xFFFE, 0x828, 0x4044, 0x27FE, 0x2040, 0x87FC, 0x4444, 0x57FC, 0x1444, 0xE7FC, 0x2444, 0x2444, 0x2454, 0x408}},
    {33971, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x25FC, 0x7D24, 0x924, 0x1154, 0x254C, 0x7D84, 0x104, 0x5504, 0x5514, 0x8108}},
    {33972, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4100, 0x227C, 0xFFC4, 0x844, 0x497C, 0x4944, 0x4944, 0x7F7C, 0x1144, 0x2044, 0x4094, 0x8108}},
    {33973, 2, {0x440, 0xFFFE, 0x440, 0x38, 0x7FC0, 0x2208, 0x1510, 0x840, 0x1F80, 0x120, 0x1FF0, 0x210, 0x7FFC, 0x440, 0x1830, 0xE00E}},
    {33974, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1010, 0x1090, 0x2488, 0x7D04, 0xA02, 0x11F8, 0x2448, 0x7C48, 0x88, 0x5488, 0x5528, 0x8210}},
    {33975, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010}},
    {33976, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1FC0, 0x44, 0x7D88, 0x550, 0x920, 0x3118, 0xC506, 0x200, 0x3FF8, 0x0, 0x2488, 0x4244}},
    {33977, 2, {0x820, 0xFFFE, 0x820, 0x1010, 0x820, 0xFFFE, 0x440, 0x3FF8, 0x448, 0xFFFE, 0x448, 0x3FF8, 0xC60, 0x3458, 0xC446, 0x440}},
    {33978, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x1FFC, 0x1000, 0x9200, 0x53F8, 0x5440, 0x1040, 0x3FFE, 0x5040, 0x90A0, 0x2110, 0x2208, 0x4C06}},
    {33979, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7EFC, 0x204, 0x7EFC, 0x4080, 0x7EFC, 0x204, 0x2244, 0x1A34, 0x62C4, 0x1428, 0x810}},
    {33980, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x6C0, 0x1A30, 0xE10E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x2FF8, 0x4808, 0x8FF8}},
    {33981, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2108, 0x2FE8, 0x2288, 0x2448, 0x2828, 0x3FF8, 0x100, 0x884, 0x4892, 0x4812, 0x87F0}},
    {33982, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1248, 0x1150, 0x40, 0xF7FC, 0x10E0, 0x1150, 0x1248, 0x1444, 0x1040, 0x2800, 0x47FE}},
    {33983, 2, {0x820, 0xFFFE, 0x820, 0x200, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {33984, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xC, 0x7DF0, 0x440, 0x888, 0x11F0, 0x1020, 0x1C44, 0xF3FE, 0x1022, 0x1124, 0x5222, 0x2060}},
    {33985, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x17D0, 0x2108, 0xDFF6, 0x540, 0x920, 0x1110}},
    {33986, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {33987, 2, {0x820, 0xFFFE, 0x820, 0x500, 0x3978, 0x2108, 0x3D78, 0x2108, 0x3FF8, 0x100, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {33988, 2, {0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x820, 0x1FC0, 0x310, 0xC08, 0x3FFC, 0x4, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {33989, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2020, 0x17FE, 0x1080, 0x8108, 0x43FC, 0x5004, 0x1150, 0x2150, 0xE150, 0x2252, 0x2252, 0x244E}},
    {33990, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x23E0, 0x2080, 0x2FF8, 0x4888, 0x48A8, 0x8890, 0x80}},
    {33991, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1F90, 0x1090, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1030}},
    {33992, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x400, 0x3F7C, 0x2144, 0x3F48, 0x2150, 0x3F48, 0x2844, 0x2444, 0x2A54, 0x3248, 0x2040, 0x40}},
    {33993, 2, {0x820, 0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8924, 0x1290, 0x2448, 0x820, 0x3FF8, 0xC826, 0x820, 0xFE0, 0x820}},
    {33994, 2, {0x820, 0xFFFE, 0x820, 0x440, 0x820, 0x3218, 0xC446, 0xFE0, 0x20, 0x7EFC, 0x2244, 0x1224, 0xA14, 0x1224, 0x2A54, 0x4488}},
    {33995, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x100, 0x7FFC, 0x800, 0x1010, 0x3FF8, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {33996, 2, {0x440, 0xFFFE, 0x440, 0x1110, 0x1110, 0x2928, 0x7FFC, 0x100, 0xFFFE, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE00E}},
    {33997, 2, {0x820, 0xFFFE, 0xA20, 0x3FC0, 0x280, 0xFFFE, 0x220, 0xFC0, 0x3408, 0xC3F8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {33998, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2088, 0x2FF8, 0x2080, 0x4FF8, 0x4808, 0x8FF8, 0x808}},
    {33999, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x61C, 0x38E8, 0x28A8, 0x2CA8, 0x2AA8, 0x2AA8, 0x2AA8, 0x28A4, 0x5554, 0x5D74, 0x8212}},
    {34000, 2, {0x820, 0xFFFE, 0x820, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x2430, 0x260E, 0x2020, 0x5FFC, 0x4420, 0x82A0, 0x40}},
    {34001, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0xFFFE, 0x1010, 0x1FF0, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {34002, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x7F7C, 0x810, 0x7F10, 0x4910, 0x7F10, 0x49FE, 0x7F10, 0x810, 0xFF90, 0x810, 0x810}},
    {34003, 2, {0x820, 0x820, 0xFFFE, 0x8A0, 0x1040, 0x17FC, 0x2000, 0x6208, 0xA110, 0x2FFE, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2208}},
    {34004, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x440, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x2488, 0x2244, 0x4244}},
    {34005, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x11FC, 0xFC20, 0x1020, 0x33FE, 0x3800, 0x5420, 0x51FC, 0x9020, 0x1020, 0x13FE}},
    {34006, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x3FF8, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x0, 0x2084, 0x3E98, 0x20E0, 0x2084, 0x2684, 0x387C}},
    {34007, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x23F8, 0x4040, 0x8840, 0x17FC, 0x3040, 0x5240, 0x9278, 0x1240, 0x1540, 0x18FE, 0x1000}},
    {34008, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x1210, 0x2448, 0x4824, 0x1FF0, 0x510, 0x888, 0x1850, 0x2920, 0xCA18, 0xC06}},
    {34009, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFE, 0x2080, 0x2490, 0x2490, 0x2AA8, 0x31C4, 0x2080, 0x2FF8, 0x4080, 0x4080, 0x9FFE}},
    {34010, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1080, 0x11FC, 0x2288, 0x2870, 0x6B8E, 0xA820, 0x2848, 0x2990, 0x2864, 0x2988, 0x2030, 0x21C0}},
    {34011, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x820, 0xFFBE, 0x844, 0x7FA4, 0x4924, 0x4B28, 0x1C28, 0x2A10, 0x4928, 0x8844, 0x882}},
    {34012, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7F7C, 0x1404, 0x1404, 0x7F04, 0x557C, 0x5740, 0x6140, 0x7F40, 0x4142, 0x4142, 0x7F3E, 0x4100}},
    {34013, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x480, 0x2480, 0x24FC, 0x24A0, 0x2510, 0x0, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {34014, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xFC, 0x7F00, 0x2208, 0x1110, 0xC20, 0x3078, 0x2008, 0x3C78, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {34015, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1004, 0x1F04, 0x2224, 0x7FA4, 0xA4A4, 0x3FA4, 0x24A4, 0x3FA4, 0x4, 0x784, 0x7814, 0x2008}},
    {34016, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0xFFFE, 0x440, 0x1390, 0x1450, 0x1FF0, 0x100, 0x7FFC, 0x4444, 0x4FE4, 0x4424, 0x400C}},
    {34017, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23F8, 0x1008, 0x11F8, 0x8008, 0x43F8, 0x5000, 0x17FC, 0x2404, 0xE3F8, 0x2110, 0x20E0, 0x271C}},
    {34018, 2, {0x440, 0x440, 0xFFFE, 0x1440, 0x800, 0x3E7C, 0x2244, 0x3E44, 0x227C, 0x3E44, 0x2244, 0x287C, 0x2444, 0x2A44, 0x3294, 0x2108}},
    {34019, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7DFC, 0x420, 0x2840, 0x11FC, 0xFD04, 0x1524, 0x1124, 0x1124, 0x1124, 0x1050, 0x5088, 0x2304}},
    {34020, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x4040, 0x23F8, 0x2040, 0xFFE, 0x1110, 0x1288, 0x24F4, 0xE110, 0x22A0, 0x2040, 0x21B0, 0x2E0E}},
    {34021, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7FFE, 0x4102, 0x8284, 0xC60, 0x3018, 0xDFF6, 0x100, 0x3FF8, 0x1110, 0x920, 0x7FFC}},
    {34022, 2, {0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x1010, 0x2FC8, 0xC046, 0x840, 0xFF8, 0x8, 0x7FA8, 0x10}},
    {34023, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1080, 0x11F8, 0x2288, 0x6870, 0xABAE, 0x2820, 0x29FC, 0x2820, 0x28A8, 0x2924, 0x22A4, 0x2040}},
    {34024, 2, {0x440, 0xFFFE, 0x440, 0x1080, 0x11FC, 0x2288, 0x2870, 0x6B8E, 0xA800, 0x29FC, 0x2904, 0x29FC, 0x2904, 0x29FC, 0x2104, 0x210C}},
    {34025, 2, {0x820, 0xFFFE, 0x820, 0x80, 0x7CF8, 0x508, 0x2A90, 0x1060, 0xFD98, 0x1606, 0x3040, 0x53F8, 0x9048, 0x1088, 0x5150, 0x2620}},
    {34026, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x47F8, 0x2110, 0x20A0, 0x7FC, 0xE444, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x2454, 0x5408, 0x8FFE}},
    {34027, 2, {0x820, 0xFFFE, 0x820, 0x23FC, 0x1080, 0x1144, 0x668, 0x70B0, 0x1128, 0x1668, 0x10A4, 0x1124, 0x16A0, 0x2840, 0x47FE, 0x0}},
    {34028, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x23F0, 0x1510, 0x10E0, 0x71C, 0x7040, 0x13F8, 0x1040, 0x11F0, 0x1040, 0x17FC, 0x2840, 0x47FE}},
    {34029, 2, {0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2080, 0x3FFC, 0x2220, 0x2490, 0x2FF8, 0x3494, 0x27F0, 0x2490, 0x47F2, 0x4082, 0x807E}},
    {34030, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x27FC, 0x1040, 0x13F8, 0x248, 0x73F8, 0x1248, 0x13F8, 0x1040, 0x17FC, 0x2840, 0x47FE, 0x0}},
    {34031, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1110, 0x2110, 0x4110, 0x8AA8, 0x1444, 0x3040, 0x5240, 0x927C, 0x1240, 0x1540, 0x14FE, 0x1800}},
    {34032, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x213C, 0x4120, 0x8920, 0x17FE, 0x3000, 0x5020, 0x9120, 0x113E, 0x1120, 0x12A0, 0x147E}},
    {34033, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2108, 0x1090, 0x93FC, 0x4090, 0x4890, 0x890, 0x17FE, 0xE090, 0x2090, 0x2110, 0x2110, 0x210}},
    {34034, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2F78, 0x2948, 0x2F78, 0x2000, 0x3FFC, 0x0}},
    {34035, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x7FFC, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE, 0x0}},
    {34036, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFC, 0x20, 0xFFFE, 0x820, 0x460}},
    {34037, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1024, 0x13FE, 0x1C24, 0xF1FC, 0x1020, 0x113C, 0x5120, 0x22FE}},
    {34038, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xA0, 0x90, 0x79FE, 0x4920, 0x4B20, 0x4DFC, 0x4920, 0x49FC, 0x7920, 0x4920, 0x1FE, 0x100}},
    {34039, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x10A0, 0x1090, 0xF9FE, 0x1120, 0x1B20, 0x15FC, 0x3120, 0xD1FC, 0x1120, 0x1120, 0x51FE, 0x2100}},
    {34040, 2, {0x440, 0xFFFE, 0x440, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {34041, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1400, 0x27DE, 0x4912, 0x8112, 0x17D2, 0x2112, 0x6112, 0xA5D2, 0x251A, 0x25D4, 0x2E10, 0x2410}},
    {34042, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x1040, 0x7C40, 0x10F8, 0xFE48, 0x4448, 0x92C8, 0x7C48, 0x10AA, 0x1CAA, 0xE106, 0x4202}},
    {34043, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x1040, 0x7C40, 0x11F8, 0xFE48, 0x4448, 0x28C8, 0x7C48, 0x10AA, 0xFEAA, 0x1106, 0x1202}},
    {34044, 2, {0x820, 0xFFFE, 0x820, 0x3CF8, 0x2448, 0x1428, 0x2548, 0x6C0, 0x1830, 0xE30E, 0xC40, 0x3180, 0x630, 0x38C0, 0x700, 0x3800}},
    {34045, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x920, 0xFFFE, 0x920, 0x3FF8, 0x100, 0x7FFC, 0x100, 0x100}},
    {34046, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x1FC, 0x100, 0x3FF8, 0x2928, 0x2448, 0x2288, 0x2928, 0x2288, 0x2448, 0x2928, 0x3FF8, 0x2008}},
    {34047, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFE, 0x4802, 0x8BF4, 0x1080, 0x33F8, 0x5208, 0x9208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {34048, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x87C, 0x7F44, 0x2248, 0x1450, 0xFFC8, 0x44, 0x3E44, 0x2254, 0x2248, 0x3E40, 0x2240}},
    {34049, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {34050, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x400, 0x820, 0x1FC0, 0x310, 0x3FF8, 0x1120, 0x2510, 0x4208}},
    {34051, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7CF8, 0x488, 0x4F8, 0x7C20, 0x41FC, 0x4124, 0x7D24, 0x5FC, 0x420, 0x424, 0x2BFE, 0x1002}},
    {34052, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {34053, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4FF4, 0x4244, 0x4644, 0x4184, 0x4244, 0x4C24, 0x400C}},
    {34054, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x13F8, 0x1040, 0x87FE, 0x4110, 0x5288, 0x14F4, 0x2110, 0xE2A0, 0x2040, 0x21B0, 0x2E0E}},
    {34055, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x7DFC, 0x4420, 0x7D20, 0x4520, 0x45FE, 0x7C50, 0x5050, 0x4892, 0x5492, 0x610E, 0x4200}},
    {34056, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1110, 0x2508, 0x200}},
    {34057, 2, {0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x1010, 0x2008, 0x5FF4, 0x1010, 0x1010, 0x1FF0, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {34058, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x11FC, 0x1104, 0x15FC, 0x5504, 0x59FC, 0x5000, 0x91FC, 0x1020, 0x2BFE, 0x2420, 0x4420, 0x8020}},
    {34059, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2200, 0x13BE, 0x9212, 0x4212, 0x5FD2, 0x1014, 0x2214, 0xEA88, 0x3248, 0x2214, 0x2A24, 0x2442}},
    {34060, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x820, 0xFFA0, 0x83E, 0x7F42, 0x4994, 0x4910, 0x7F10, 0x1C10, 0x2A28, 0x4928, 0x8844, 0x882}},
    {34061, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2848, 0x2F70, 0x4844, 0x4B44, 0x8C3C}},
    {34062, 2, {0x440, 0x440, 0xFFFE, 0x2440, 0x10F8, 0xFC88, 0x88, 0x7906, 0x200, 0x79F8, 0x88, 0x7888, 0x4850, 0x4820, 0x78D8, 0x4B06}},
    {34063, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x440, 0x7FFC, 0x4444, 0x4824, 0x57D4, 0x644C, 0x4444, 0x47C4, 0x4014, 0x4008}},
    {34064, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x820, 0x440, 0x7FFC, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x400C}},
    {34065, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x90, 0x3FFC, 0x2080, 0x3048, 0x2850, 0x2424, 0x40D4, 0x830C}},
    {34066, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FF8, 0x208, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0xCFF8, 0x0, 0x2444, 0x4222}},
    {34067, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x3FF0, 0x820, 0x7C0, 0x1830, 0xE00E}},
    {34068, 2, {0x820, 0xFFFE, 0x2820, 0x3FFC, 0x4004, 0xBFE4, 0x4, 0x1FC4, 0x1044, 0x1FC4, 0x4, 0x3FE4, 0x2224, 0x3FE4, 0x2224, 0x3FEC}},
    {34069, 2, {0x820, 0xFFFE, 0x820, 0x1450, 0x1450, 0x7FFC, 0x2452, 0x47CE, 0x8000, 0x7FFC, 0x4104, 0x1FF0, 0x1110, 0x1150, 0x1120, 0x100}},
    {34070, 2, {0x820, 0xFFFE, 0x920, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2410, 0x27F0, 0x2410, 0x47F0, 0x4410, 0x9FFC}},
    {34071, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2220, 0x2220, 0x3FFC, 0x2220, 0x2220, 0x23E0, 0x2000, 0x4A48, 0x4924, 0x9124}},
    {34072, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x1FF0, 0x1020, 0x2548, 0x7EFC, 0x440, 0x1930, 0xE64E, 0x1990, 0x620, 0x18C0, 0x700, 0x3800}},
    {34073, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1040, 0xFEFE, 0x2100, 0x2000, 0x3DFE, 0x2422, 0x2420, 0x24BC, 0x44A0, 0x5560, 0x8A3E}},
    {34074, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x0, 0x3F08, 0x2108, 0x3F08, 0x20FE, 0x2E08, 0x2048, 0x3F28, 0x4428, 0x5508, 0xA4A8, 0xC10}},
    {34075, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1004, 0x1F24, 0x2214, 0x7F84, 0xA4A4, 0x3F94, 0x2486, 0x3FBC, 0x2484, 0x2484, 0x4284, 0x8104}},
    {34076, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x820, 0x7F3E, 0x844, 0x3EA4, 0x824, 0xFF28, 0x1028, 0x1E10, 0x2228, 0x4A44, 0x8482}},
    {34077, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x7924, 0x48A8, 0x4820, 0x7BFE, 0x4870, 0x78A8, 0x48A8, 0x4924, 0x7A22, 0x4820, 0x20}},
    {34078, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x2108, 0xFFFE, 0x2108, 0x3FF8, 0x1110, 0x1FF0, 0x400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {34079, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2040, 0x1040, 0xFEFE, 0x2120, 0x2040, 0x3CFC, 0x2520, 0x2420, 0x25FE, 0x4450, 0x5488, 0x8906}},
    {34080, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x10FC, 0x2488, 0x7D50, 0x820, 0x10D8, 0x2706, 0x7C60, 0x10, 0x54C0, 0x5430, 0x8008}},
    {34081, 2, {0x440, 0xFFFE, 0x440, 0x1080, 0x1EF8, 0x2288, 0x5450, 0x2820, 0x17D8, 0x2006, 0xDFF0, 0x100, 0x1110, 0x2108, 0x4504, 0x200}},
    {34082, 2, {0x440, 0xFFFE, 0x440, 0x2020, 0x13FC, 0x8224, 0x4BF8, 0x1288, 0x2450, 0xE470, 0x298C, 0x2400, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {34083, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x448, 0x7DFE, 0x2488, 0x2448, 0x2448, 0x4408, 0x8418}},
    {34084, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4202, 0x8924, 0x2848, 0x4994, 0xE10, 0x77F0, 0x100, 0x2108, 0x2108, 0x3FF8, 0x8}},
    {34085, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x200, 0x3FF8, 0x2208, 0x27E8, 0x2A48, 0x2188, 0x2648, 0x3FF8, 0x100, 0x4884, 0x4812, 0x87F2}},
    {34086, 2, {0x820, 0xFFFE, 0x820, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4924, 0x8494, 0x8}},
    {34087, 2, {0x820, 0xFFFE, 0x820, 0x1010, 0xFEFE, 0x1010, 0x7C7C, 0x1010, 0xFEFE, 0x1010, 0x3FF8, 0x8, 0x1FF8, 0x8, 0x3FF8, 0x8}},
    {34088, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4544, 0x5FF4, 0x4204, 0x7FFC, 0x4824, 0x57D4, 0x644C, 0x44C4, 0x4424, 0x43E4, 0x4004, 0x7FFC}},
    {34089, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0xBFFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x1010, 0x2008, 0x4004}},
    {34090, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x806, 0x7F78, 0x840, 0x7F40, 0x497E, 0x7F48, 0x4948, 0x7F48, 0x848, 0x7F48, 0x888, 0x908}},
    {34091, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x100, 0x11F8, 0x1100, 0xFFFE, 0x1000, 0x1FF8, 0x2000, 0x3FFC, 0x4, 0x4924, 0x8494, 0x8}},
    {34092, 2, {0x820, 0xFFFE, 0x820, 0x10, 0x7DFE, 0x420, 0x1848, 0x51FE, 0x5002, 0x5CA8, 0x50A8, 0x50A8, 0x5D2A, 0x712A, 0xC22A, 0x406}},
    {34093, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x78A0, 0x4910, 0x5248, 0x65F4, 0x5020, 0x4840, 0x4BF8, 0x6800, 0x57FC, 0x4080, 0x4110, 0x43F8}},
    {34094, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x7FFC, 0x40A4, 0x7FFC, 0x4084, 0x5E94, 0x5294, 0x5E64, 0x4054, 0x46B4, 0x5914, 0x7FFC, 0x4004}},
    {34095, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20, 0x7BFE, 0x4820, 0x51FC, 0x6124, 0x51FC, 0x4924, 0x49FC, 0x6870, 0x50A8, 0x4124, 0x4622}},
    {34096, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x2008, 0x3FF8, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2FF8, 0x2888, 0x4FF8, 0x4802, 0x87FE}},
    {34097, 2, {0x820, 0xFFFE, 0x820, 0x2200, 0x1478, 0x848, 0x1448, 0x2A86, 0x900, 0x7EFC, 0x844, 0x2A44, 0x4928, 0x8910, 0x2828, 0x10C6}},
    {34098, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8044, 0x3E40, 0x7C, 0x7E88, 0x1550, 0x1420, 0x1452, 0x248A, 0x43FE, 0x8000}},
    {34099, 2, {0x440, 0xFFFE, 0x440, 0x23F8, 0x1040, 0x13F8, 0x8040, 0x47FE, 0x5000, 0x13F8, 0x2208, 0xE3F8, 0x2208, 0x23F8, 0x2208, 0x2218}},
    {34100, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2420, 0x2420, 0x3F7C, 0x2420, 0x2E70, 0x2D68, 0x34A4, 0x4522, 0x4420, 0x8420}},
    {34101, 2, {0x440, 0x440, 0xFFFE, 0x448, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2A24, 0x2FA8, 0x28A8, 0x2F90, 0x2A12, 0x4FAA, 0x4046, 0x8082}},
    {34102, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {34103, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8}},
    {34104, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x1200, 0x67DC, 0x4444, 0x47C4, 0x4444, 0x57DC, 0x6000, 0x440, 0x440, 0x842, 0x1042, 0x603E}},
    {34105, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x820, 0x143E, 0x2244, 0x5DA4, 0x8024, 0x1128, 0x928, 0x4A10, 0x27A8, 0xF844, 0x4082}},
    {34106, 2, {0x820, 0xFFFE, 0x820, 0x2000, 0x17FC, 0x504, 0x4484, 0x4FF4, 0x5904, 0x6FE4, 0x4904, 0x4FE4, 0x4904, 0x4FF4, 0x4814, 0x4008}},
    {34107, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4002, 0x8024, 0x3E38, 0x20, 0x7EF8, 0x1408, 0x1450, 0x1424, 0x2454, 0x2484, 0x43FC}},
    {34108, 2, {0x820, 0xFFFE, 0x820, 0x43F8, 0x2208, 0x23F8, 0x208, 0xE3F8, 0x2100, 0x23FC, 0x2444, 0x22A4, 0x2A04, 0x33F4, 0x2014, 0x8}},
    {34109, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x4920, 0x2A3E, 0x7F44, 0x49A4, 0x4924, 0x5D28, 0x6B28, 0x4910, 0x4128, 0x4544, 0x4282}},
    {34110, 2, {0x820, 0xFFFE, 0x820, 0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4908, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106}},
    {34111, 2, {0x820, 0xFFFE, 0x820, 0xF8, 0x7F00, 0x2210, 0x1120, 0x1FE0, 0x1020, 0x1FF0, 0x1010, 0x1FFC, 0x2004, 0x2AA4, 0x4AA4, 0x800C}},
    {34112, 2, {0x440, 0xFFFE, 0x440, 0x1010, 0x1010, 0xFEFE, 0x1010, 0xFEFE, 0x9292, 0x9292, 0x9696, 0x3838, 0x5454, 0x9292, 0x1010, 0x1010}},
    {34113, 2, {0x820, 0xFFFE, 0x820, 0x3FF0, 0x10, 0x1FF0, 0x10, 0x3FF0, 0x0, 0x3E7C, 0x844, 0x7F7C, 0x20, 0xFFFE, 0x820, 0x460}},
    {34114, 2, {0x440, 0xFFFE, 0x440, 0x20, 0x3D24, 0x24A8, 0x25FC, 0x3C40, 0x27FE, 0x2488, 0x3D24, 0x26FA, 0x2448, 0x4448, 0x54A8, 0x8910}},
    {34115, 2, {0x820, 0xFFFE, 0x820, 0xF8, 0x3F00, 0x1110, 0x920, 0xFFFE, 0x920, 0x3018, 0xDFF6, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1010}},
    {34116, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4084, 0x4AA4, 0x4A14, 0x4A54, 0x51C4, 0x4014, 0x4008}},
    {34117, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23F8, 0x1248, 0x93F8, 0x4248, 0x4BF8, 0x840, 0x17FC, 0xE444, 0x2454, 0x25F4, 0x2414, 0x40C}},
    {34118, 2, {0x440, 0x440, 0xFFFE, 0x448, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x28A4, 0x2FA8, 0x28A8, 0x2F90, 0x2892, 0x4FAA, 0x4546, 0x8882}},
    {34119, 2, {0x820, 0xFFFE, 0x820, 0x3EF8, 0x2288, 0x3EF8, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {34120, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {34121, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x2488, 0x2244, 0x4244}},
    {34122, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x100, 0x2488, 0x2424, 0x43E4, 0x0, 0x2424, 0x1212, 0xA0A0, 0xA4A4, 0x1C1C, 0x0}},
    {34123, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x810, 0x810, 0x4E9C, 0x4890, 0x4890, 0x4E90, 0xF1FE, 0x0}},
    {34124, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x4284}},
    {34125, 2, {0x440, 0xFFFE, 0x440, 0x20A0, 0x1110, 0x1208, 0x85F6, 0x4000, 0x5784, 0x14A4, 0x27A4, 0xE4A4, 0x27A4, 0x2484, 0x2494, 0x2588}},
    {34126, 2, {0x820, 0xFFFE, 0x820, 0x70, 0x1F80, 0x200, 0xFFFE, 0x820, 0x37D8, 0xC446, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2018}},
    {34127, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x44, 0x53F4, 0x2048, 0x5050, 0x97FE, 0x1080, 0x31FC, 0x5304, 0x95FC, 0x1904, 0x51FC, 0x2104}},
    {34128, 2, {0x440, 0xFFFE, 0x440, 0x7FFC, 0x1110, 0x2928, 0x4544, 0x1110, 0x2928, 0x4544, 0x7FFC, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {34129, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x47C4, 0x4444, 0x47C4, 0x4444, 0x47C4, 0x4454, 0x48C8}},
    {34130, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {34131, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1010, 0x3FF8}},
    {34132, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x10FC, 0x7C84, 0x1094, 0xFE88, 0x4480, 0x28FC, 0x7CA4, 0x10A4, 0xFEA8, 0x1090, 0x10A8, 0x10C6}},
    {34133, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2288, 0x1450, 0xBFE, 0x1450, 0x25FC, 0x4D54, 0x158C, 0x2504, 0x45FC, 0x8504, 0x29FC, 0x1104}},
    {34134, 2, {0x820, 0xFFFE, 0x820, 0x23FC, 0x1200, 0x43F8, 0x2A08, 0xBF8, 0x7200, 0x13FC, 0x1100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {34135, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x80, 0x78F8, 0x4908, 0x4A90, 0x7860, 0x1198, 0x1206, 0x5DF8, 0x5108, 0x5108, 0x5DF8, 0xE108}},
    {34136, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x820, 0x3EF8, 0x820, 0x7EFC, 0x0, 0x7FFC, 0x440, 0x1842, 0xE03E}},
    {34137, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x204, 0x4908, 0x4824, 0x87E4}},
    {34138, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x7C7C, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1000, 0x1FF8, 0x8, 0x70}},
    {34139, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xC88, 0x7050, 0x11FC, 0x1024, 0xFDFC, 0x3120, 0x39FE, 0x5462, 0x50AA, 0x9124, 0x1220, 0x1020}},
    {34140, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x200, 0x908, 0x4924, 0x4824, 0x87E0}},
    {34141, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1080, 0x10F8, 0x2508, 0x7E10, 0x9FC, 0x1124, 0x2524, 0x7DFC, 0x100, 0x5502, 0x5502, 0x80FE}},
    {34142, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x2200, 0x3EFC, 0x2244, 0x3E44, 0x2228, 0x2F10, 0xF228, 0x42C6}},
    {34143, 2, {0x820, 0xFFFE, 0x8A0, 0x3E90, 0x464, 0x2828, 0x1010, 0x3EF8, 0xC296, 0x1E90, 0x110C, 0x20F0, 0x3E90, 0x260, 0x1490, 0x908}},
    {34144, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x7F7C, 0x1144, 0x3244, 0xC44, 0x367C, 0xC880, 0x1F00, 0x210, 0x3FF8, 0x108, 0x1520, 0x2210}},
    {34145, 2, {0x440, 0xFFFE, 0x540, 0x7FFC, 0x920, 0xFFFE, 0x820, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010}},
    {34146, 2, {0x820, 0xFFFE, 0x920, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {34147, 2, {0x820, 0xFFFE, 0x820, 0xFC, 0x7F00, 0x2208, 0x1110, 0x7FFE, 0x4002, 0x9024, 0x3DF8, 0x4420, 0xA920, 0x11FC, 0x2020, 0xC020}},
    {34148, 2, {0x820, 0xFFFE, 0x820, 0x10, 0xFF50, 0x107C, 0x2990, 0x4E10, 0x967C, 0x2510, 0xCD10, 0x14FE, 0x2440, 0xC430, 0x140E, 0x800}},
    {34149, 2, {0x820, 0xFFFE, 0x820, 0x50, 0x7E48, 0x4FE, 0x2490, 0x2590, 0x44FE, 0x7E90, 0xC90, 0x14FE, 0x2490, 0x4490, 0x94FE, 0x880}},
    {34150, 2, {0x820, 0xFFFE, 0x820, 0x0, 0xBFE, 0x1020, 0x21FC, 0x4904, 0x11FC, 0x2104, 0x45FC, 0x904, 0x11FC, 0x2000, 0xC088, 0x104}},
    {34151, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x13FC, 0x2400, 0x43F8, 0x8A08, 0x13F8, 0x3208, 0x53F8, 0x9100, 0x13F8, 0x1510, 0x10E0, 0x171C}},
    {34152, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3FFE, 0x2890, 0x2510, 0x2FBE, 0x2222, 0x2AD4, 0x2A90, 0x2F90, 0x2228, 0x4428, 0x4444, 0x8882}},
    {34153, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23F8, 0x1208, 0x93F8, 0x4208, 0x43F8, 0x1000, 0x2FFE, 0xE200, 0x27FC, 0x2924, 0x2254, 0x488}},
    {34154, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x3FFC, 0x4AA0, 0xAA0, 0x7FFC, 0xAA0, 0xAA0, 0x7FFC, 0x0, 0x2488, 0x2244, 0x4244}},
    {34155, 2, {0x820, 0xFFFE, 0x920, 0x1FF0, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {34156, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1020, 0x1020, 0x2440, 0x7C88, 0x9F0, 0x1020, 0x2448, 0x7DFC, 0x20, 0x54A8, 0x5524, 0x8060}},
    {34157, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x2548, 0x3D78, 0x2108, 0x3FF8, 0x2108, 0x3D78, 0x2548, 0x4548}},
    {34158, 2, {0x820, 0xFFFE, 0x820, 0x200, 0x1CF0, 0x1010, 0x1EF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2004, 0x4924, 0x2494, 0x2494, 0x4008}},
    {34159, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x79F0, 0x4A90, 0x5060, 0x6198, 0x5606, 0x49F0, 0x4A40, 0x6BF8, 0x5440, 0x43F8, 0x4040, 0x47FC}},
    {34160, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x23F8, 0x1208, 0x93F8, 0x4208, 0x4BF8, 0x800, 0x17FC, 0xE524, 0x2524, 0x2524, 0x2FFE, 0x0}},
    {34161, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xC20, 0x7124, 0x10A8, 0x1020, 0xFDFC, 0x3104, 0x39FC, 0x5504, 0x51FC, 0x9104, 0x1114, 0x1108}},
    {34162, 2, {0x820, 0xFFFE, 0x820, 0x220C, 0x1470, 0x7F40, 0x4940, 0x7F40, 0x497E, 0x7F48, 0x848, 0xFF48, 0x848, 0x888, 0x888, 0x908}},
    {34163, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {34164, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x21F8, 0x2108, 0x45F8, 0x8908, 0xF1F8, 0x2000, 0x43FC, 0xFA94, 0x294, 0x1A94, 0xE7FE, 0x4000}},
    {34165, 2, {0x440, 0xFFFE, 0x440, 0x2040, 0x20A0, 0x2158, 0x3A26, 0x4DF8, 0x4908, 0x49F8, 0xA908, 0x11F8, 0x1144, 0x2128, 0x4150, 0x8188}},
    {34166, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x6C0, 0x1930, 0xEFEE, 0x40, 0x680, 0x100}},
    {34167, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7DFE, 0x420, 0x28FC, 0x1084, 0xFEFC, 0x1284, 0x14FC, 0x1084, 0x10FC, 0x1000, 0x5048, 0x2084}},
    {34168, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x47BC, 0x2484, 0x27BC, 0x400, 0xE7BC, 0x2404, 0x27A8, 0x2410, 0x2428, 0x5444, 0x8FFE, 0x0}},
    {34169, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0xFFFE, 0x10A0, 0x2090, 0x45FE, 0x7B20, 0x11FC, 0x2520, 0x7DFC, 0x920, 0x1120, 0x21FE, 0xC100}},
    {34170, 2, {0x440, 0xFFFE, 0x440, 0x3E28, 0x2224, 0x3E24, 0x20, 0xFFFE, 0x2220, 0x3E28, 0x2228, 0x3E10, 0x2212, 0x2F2A, 0xF246, 0x4282}},
    {34171, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x48, 0x3E48, 0x2048, 0x3CFC, 0x2048, 0x3C48, 0x2048, 0xFEFE, 0x2000, 0x4448, 0xFE84, 0x4302}},
    {34172, 2, {0x440, 0xFFFE, 0x440, 0x20, 0x3EFC, 0x2024, 0x3DFE, 0x2024, 0x3CFC, 0x2020, 0xFEFC, 0x1020, 0x25FE, 0x7E20, 0x2220, 0x20}},
    {34173, 2, {0x820, 0xFFFE, 0x820, 0x240, 0x1FF8, 0x1248, 0x1FF8, 0x1248, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2448, 0x4530, 0x860E}},
    {34174, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {34175, 2, {0x440, 0xFFFE, 0x440, 0x101E, 0x13E0, 0x1522, 0x5494, 0x5BFC, 0x5040, 0x97FE, 0x1080, 0x11FC, 0x2A48, 0x2430, 0x4448, 0x8186}},
    {34176, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x23F8, 0x1248, 0x9248, 0x42A8, 0x4BF8, 0x800, 0x17FC, 0xE4A4, 0x24A4, 0x24A4, 0x2FFE, 0x0}},
    {34177, 2, {0x820, 0xFFFE, 0x820, 0x400, 0x3FF8, 0x2928, 0x2548, 0x2FE8, 0x2548, 0x2828, 0x100, 0xFFFE, 0x280, 0xC60, 0x3018, 0xC006}},
    {34178, 2, {0x820, 0xFFFE, 0x820, 0x1000, 0x3E20, 0x2220, 0x3E3E, 0x2244, 0x3EA4, 0x824, 0x7F28, 0x1028, 0x1E10, 0x2228, 0x2A44, 0x4482}},
    {34179, 2, {0x440, 0xFFFE, 0x440, 0x2020, 0x17FE, 0x1000, 0x81F8, 0x4108, 0x51F8, 0x1000, 0x27FE, 0xE402, 0x25FA, 0x250A, 0x25FA, 0x2406}},
    {34180, 2, {0x820, 0xFFFE, 0x820, 0x48, 0x27FC, 0x1040, 0x13F8, 0x8248, 0x43F8, 0x4A48, 0xBF8, 0x1248, 0xF010, 0x27FC, 0x2210, 0x2130}},
    {34181, 2, {0x820, 0xFFFE, 0x820, 0x23FE, 0x2200, 0xFAFC, 0x4A00, 0x4BFE, 0x4AA4, 0x9298, 0x52C6, 0x2208, 0x53FE, 0x4A88, 0x8448, 0x818}},
    {34182, 2, {0x820, 0xFFFE, 0x820, 0xF8, 0x7F00, 0x2210, 0x1120, 0x7FFC, 0x4204, 0x2928, 0x47E4, 0x800, 0x1FF0, 0x2820, 0x7C0, 0xF83E}},
    {34183, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2220, 0x4AA0, 0x8ABE, 0x1FC4, 0x20A4, 0x6F24, 0xA028, 0x2F28, 0x2910, 0x29A8, 0x2944, 0x3082}},
    {34184, 2, {0x820, 0xFFFE, 0x920, 0x280, 0xC60, 0x37D8, 0xC006, 0x3FF8, 0x2928, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {34185, 2, {0x820, 0xFFFE, 0x920, 0x11F8, 0x1100, 0xFFFE, 0x24, 0x3FFE, 0x2020, 0x2FA4, 0x2224, 0x2A18, 0x2A90, 0x512A, 0x4646, 0x9882}},
    {34186, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1004, 0x1F04, 0x2224, 0x7FA4, 0xA4A4, 0x3FA4, 0x24A4, 0x3FA4, 0x4, 0x2A84, 0x4554, 0x8}},
    {34187, 2, {0x820, 0xFFFE, 0x4920, 0x21FC, 0xA44, 0xF0B0, 0x230E, 0x3FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {34188, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2100, 0x27DE, 0x5452, 0x97D4, 0xE458, 0x2454, 0x57D2, 0x9512, 0xF49A, 0x2554, 0x4610, 0x8410}},
    {34189, 2, {0x440, 0xFFFE, 0x440, 0x7F10, 0x2A10, 0x7F10, 0x1210, 0xC10, 0x7F90, 0x4A90, 0x5E90, 0x4090, 0x5E92, 0x4A92, 0x4492, 0x4B8E}},
    {34190, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x7FFE, 0x4922, 0x9114, 0x31FC, 0x5F10, 0x90A0, 0x1044, 0x11A4, 0x161C}},
    {34191, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FF8, 0x820, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x4904, 0x4892, 0x87F2}},
    {34192, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xC20, 0x71FC, 0x1020, 0x13FE, 0xFC88, 0x3144, 0x3A7A, 0x5488, 0x5150, 0x9020, 0x10D8, 0x1706}},
    {34193, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE}},
    {34194, 2, {0x440, 0xFFFE, 0x440, 0x800, 0xF78, 0x2808, 0x7F28, 0xA110, 0x1628, 0x38C4, 0xD110, 0x920, 0x7FFC, 0x540, 0x1930, 0x610C}},
    {34195, 2, {0x820, 0xFFFE, 0x820, 0x23F8, 0x1008, 0x13F8, 0x8008, 0x47FC, 0x5444, 0x13F8, 0x2248, 0xE040, 0x27F8, 0x2110, 0x20E0, 0x2F1E}},
    {34196, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x1110, 0x2928, 0x4544, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x27C8, 0x2448, 0x27C8, 0x2008, 0x3FF8}},
    {34197, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x3FFC, 0x2220, 0x3FFC, 0x2240, 0x2FF8, 0x2248, 0x3FFC, 0x2248, 0x2FF8, 0x4660, 0x4A50, 0xB24C}},
    {34198, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x21F8, 0x1108, 0x11E8, 0x128, 0x73FC, 0x1204, 0x12F4, 0x1294, 0x12F4, 0x1214, 0x2A08, 0x47FE}},
    {34199, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4694, 0x4A64, 0x5214, 0x7FFC}},
    {34200, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x23F8, 0x1040, 0x17FC, 0x110, 0xF7FC, 0x1040, 0x13F8, 0x1040, 0x17FC, 0x1040, 0x2840, 0x47FE}},
    {34201, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20A0, 0x2090, 0x7DFE, 0x9120, 0x1320, 0x11FC, 0xFD20, 0x11FC, 0x2920, 0x2520, 0x41FE, 0x8100}},
    {34202, 2, {0x820, 0xFFFE, 0x820, 0xFC, 0x7E84, 0x10FC, 0x1084, 0x3CFC, 0x2400, 0x47FE, 0xA480, 0x19FE, 0xA52, 0x1092, 0x212A, 0x4044}},
    {34203, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1040, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x4020, 0x7C20, 0x45FC, 0x4420, 0x7C20, 0x4420}},
    {34204, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x3E20, 0x22FC, 0x2288, 0x3E50, 0x21FE, 0x2020, 0x3E20, 0x52FC, 0x5220, 0x9E20, 0x1220, 0x20}},
    {34205, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x2020, 0x7FFE, 0xA488, 0x2BF4, 0x2000, 0x27F8, 0x2000, 0x27F8, 0x2000, 0x47F8, 0x4408, 0x87F8}},
    {34206, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x7EA0, 0x1198, 0x1246, 0x3DF8, 0x2508, 0x45F8, 0xA508, 0x19F8, 0x944, 0x1128, 0x2190, 0x4108}},
    {34207, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x280, 0x440, 0x1830, 0x67CC, 0x0, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x820, 0x1450, 0x2288}},
    {34208, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x13FC, 0x1040, 0x15F8, 0x5508, 0x59F8, 0x5108, 0x91F8, 0x1108, 0x29F8, 0x2490, 0x4508, 0x8204}},
    {34209, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x5FF4, 0x5014, 0x5FF4, 0x4004, 0x7C7C, 0x440, 0xFC7C, 0x2444, 0x4444}},
    {34210, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x20FC, 0x3C24, 0x4444, 0xBE8C, 0x2A50, 0x3E7C, 0x2A90, 0x3E10, 0x2AFE, 0x2A10, 0x4210, 0x8610}},
    {34211, 2, {0x440, 0xFFFE, 0x440, 0x810, 0xFF10, 0x8FE, 0x7E10, 0x10, 0x7EFE, 0x4242, 0x7E44, 0x4228, 0x2410, 0xE28, 0xF044, 0x4182}},
    {34212, 2, {0x820, 0xFFFE, 0x820, 0x50, 0x7E50, 0x11DC, 0x1050, 0x3C50, 0x25DC, 0x6450, 0x9450, 0x9DC, 0x850, 0x1050, 0x23FE, 0x4000}},
    {34213, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x800, 0x1FFC, 0x2204, 0xDFC4, 0x1244, 0x1FC4, 0x224, 0x3FE4, 0x1028, 0x10}},
    {34214, 2, {0x820, 0xFFFE, 0x920, 0x3FFC, 0x2440, 0x3FF8, 0x2448, 0x3FF8, 0x2400, 0x27FC, 0x2800, 0x2FFC, 0x2004, 0x4AA4, 0x5154, 0x8008}},
    {34215, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x7FFE, 0x4002, 0xBFF4, 0x840, 0x1F48, 0x6250, 0x1462, 0x1842, 0xE03E}},
    {34216, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4444, 0x7FFC, 0x0, 0x7FFE, 0x4002, 0xBFF4, 0x840, 0x1F48, 0x6250, 0x1462, 0x1842, 0xE03E}},
    {34217, 2, {0x440, 0xFFFE, 0x440, 0x20, 0x79FC, 0x4888, 0x5050, 0x63FE, 0x5220, 0x4AA0, 0x4AFC, 0x6B20, 0x52FC, 0x4420, 0x4420, 0x49FE}},
    {34218, 2, {0x820, 0xFFFE, 0x820, 0x1000, 0x804, 0x7F78, 0x2240, 0x1440, 0xFF7E, 0x848, 0x7F48, 0x848, 0x2A48, 0x4988, 0xA888, 0x1108}},
    {34219, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0xFFFE, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x2488, 0x4244}},
    {34220, 2, {0x820, 0xFFFE, 0x920, 0x200, 0x47C4, 0x2448, 0x17D0, 0x2448, 0x47C4, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {34221, 2, {0x440, 0xFFFE, 0x440, 0x20, 0xC40, 0x71FC, 0x1124, 0x11FC, 0xFD44, 0x31FC, 0x3850, 0x5490, 0x51FE, 0x9010, 0x1010, 0x1010}},
    {34222, 2, {0x440, 0xFFFE, 0x440, 0x0, 0x4920, 0x2A20, 0xFF3E, 0x2A44, 0x49A4, 0x1024, 0xFE28, 0x2228, 0x6410, 0x1828, 0x2444, 0xC282}},
    {34223, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x200, 0x1FD0, 0x220, 0xFFFE, 0x300, 0xFF0, 0x3810, 0xCFF0, 0x810, 0xFF0}},
    {34224, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2928, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x2448, 0x4224}},
    {34225, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1408, 0x5508, 0x3608, 0x147E, 0x7F08, 0x1448, 0x7F28, 0x828, 0x3E08, 0x808, 0xF28, 0x7810}},
    {34226, 2, {0x820, 0xFFFE, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x920, 0x11C0, 0x600, 0x1FF8, 0x6808, 0xFF8, 0x808, 0xFF8, 0x410, 0x808}},
    {34227, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x23F8, 0x1040, 0x17FC, 0x0, 0x73F8, 0x1208, 0x13F8, 0x10C4, 0x11A8, 0x1290, 0x2C88, 0x47FE}},
    {34228, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x4204, 0xA928, 0x47E4, 0x0, 0x3FF8, 0x2448, 0xFFFE, 0x0, 0x7FFC, 0x100, 0x500, 0x200}},
    {34229, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x100, 0x3FF8, 0x0, 0xFFFE, 0x2, 0x3FF0, 0x100, 0xFFFE, 0x10, 0x7DFC, 0x4490, 0x7C30}},
    {34230, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x31FC, 0xCB24, 0x2D24, 0x31FC, 0xC924, 0x1524, 0x25FC, 0xCC20, 0x15FC, 0x2420, 0xD420, 0xBFE}},
    {34231, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x7FFC, 0x200, 0x3FF8, 0x2448, 0x2458}},
    {34232, 2, {0x820, 0xFFFE, 0x820, 0x2000, 0x17FC, 0x10A0, 0x87FC, 0x44A4, 0x57FC, 0x1000, 0x27FC, 0xE000, 0x2FFE, 0x2040, 0x2248, 0x24C4}},
    {34233, 2, {0x820, 0xFFFE, 0x920, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x7FFE, 0x4822, 0x9FF4, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {34234, 2, {0x820, 0xFFFE, 0x820, 0x100, 0xFFFE, 0x288, 0x7D70, 0x2548, 0x4D66, 0x9010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x2010, 0x4010}},
    {34235, 2, {0x820, 0xFFFE, 0x820, 0x2248, 0x1490, 0x1248, 0x83F8, 0x4248, 0x53F8, 0x1248, 0x23F8, 0xE040, 0x2FFE, 0x2150, 0x2248, 0x2444}},
    {34236, 2, {0x820, 0xFFFE, 0x920, 0x3FFC, 0x2440, 0x3FF8, 0x2448, 0x3FF8, 0x2840, 0x2F78, 0x2844, 0x2F3C, 0x2080, 0x4FF8, 0x4080, 0x9FFC}},
    {34237, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7F7C, 0x1420, 0x7F20, 0x5538, 0x5528, 0x7F48, 0x868, 0x858, 0x7F4A, 0x84A, 0xF6A, 0xF046}},
    {34238, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x1110, 0x2108, 0x7FFC, 0xB55A, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3558, 0x2008, 0x2018}},
    {34239, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x247C, 0x3A04, 0x2228, 0x1E10, 0x20FE, 0x3E12, 0x4850, 0x85E, 0x7F50, 0x14B0, 0x229E, 0x4100}},
    {34240, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4102, 0x9114, 0x1120, 0x2280, 0x440, 0x1830, 0xE00E}},
    {34241, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x820, 0xFE0, 0x0, 0x7FFC, 0x4824, 0x4FEC, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {34242, 2, {0x820, 0xFFFE, 0x820, 0x7F00, 0x227C, 0x3E04, 0x2228, 0x3E10, 0x23A8, 0xFE44, 0x3F8, 0x7F04, 0x1188, 0x6370, 0xD18, 0xF106}},
    {34243, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x7F20, 0x20, 0x3E3E, 0x2242, 0x3E84, 0x10, 0x7F10, 0x4110, 0x5D28, 0x5528, 0x5D44, 0x4382}},
    {34244, 2, {0x440, 0xFFFE, 0x440, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1450, 0x2F88, 0x4124, 0x1FF0, 0x110, 0x920, 0x1310}},
    {34245, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4444, 0x7FFC, 0x800, 0xFF84, 0x24, 0x7F24, 0x24, 0x7F24, 0x24, 0x7F04, 0x4114, 0x7F08}},
    {34246, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4442, 0xBFFC, 0x440, 0xFFFE, 0x820, 0x31D8, 0xCF06, 0x100, 0x7FFC, 0x100, 0x300}},
    {34247, 2, {0x440, 0x440, 0xFFFE, 0x440, 0xA00, 0x33B8, 0x2208, 0x3A38, 0x2388, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {34248, 2, {0x440, 0xFFFE, 0x440, 0x20, 0xF7A4, 0x90A8, 0x9292, 0xF114, 0x9208, 0x95F4, 0xF842, 0x93F8, 0x9040, 0xF0A0, 0x9110, 0x608}},
    {34249, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x1048, 0x7CFC, 0x1048, 0x7C48, 0x11FE, 0xFE00, 0x10FC, 0x3884, 0x54FC, 0x9484, 0x10FC, 0x1084}},
    {34250, 2, {0x440, 0xFFFE, 0x440, 0x20, 0xDFE, 0x7102, 0x1102, 0x11FE, 0xFD00, 0x31FE, 0x39AA, 0x56AA, 0x52FE, 0x94AA, 0x10A2, 0x1086}},
    {34251, 2, {0x820, 0xFFFE, 0x820, 0x3EF8, 0x1248, 0xA28, 0x1248, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {34252, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFC, 0x4244, 0x2B90, 0x4C28, 0x37E0, 0xC100, 0x1FF0, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34253, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x0, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {34254, 2, {0x820, 0xFFFE, 0x920, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7FFC, 0x2488, 0x4244, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {34255, 2, {0x440, 0xFFFE, 0x448, 0x24, 0x1FFE, 0x5020, 0x57A0, 0x7524, 0x17A4, 0xF4A8, 0x57A8, 0x5510, 0x9512, 0x27AA, 0x2046, 0x4082}},
    {34256, 2, {0x820, 0xFFFE, 0x820, 0x810, 0x3020, 0xCAFC, 0x2C84, 0x3084, 0xC8FC, 0x1484, 0x2484, 0xCCFC, 0x1448, 0x244A, 0xD48A, 0x906}},
    {34257, 2, {0x440, 0xFFFE, 0x1440, 0x1FE0, 0x2020, 0x7FF8, 0xA308, 0x2488, 0x1FF0, 0x1110, 0x1FF0, 0x800, 0x1FF0, 0x2820, 0x7C0, 0xF83E}},
    {34258, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xDFC, 0x7104, 0x11FC, 0x1104, 0xFDFC, 0x3080, 0x39FE, 0x5622, 0x5552, 0x9102, 0x11FA, 0x1006}},
    {34259, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2044, 0x3E28, 0x44FE, 0xFF10, 0x4910, 0x7F7C, 0x4910, 0x7F10, 0xFE, 0xF10, 0xF010, 0x4010}},
    {34260, 2, {0x820, 0xFFFE, 0xA20, 0x7FFC, 0x4104, 0xBFF8, 0x440, 0x2828, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x100, 0x1510, 0x2208}},
    {34261, 2, {0x820, 0xFFFE, 0x820, 0x1000, 0x11FC, 0x7D24, 0x11FC, 0x7D24, 0x11FC, 0xFC20, 0x13FE, 0x3A22, 0x562A, 0x92FA, 0x120A, 0x1206}},
    {34262, 2, {0x820, 0xFFFE, 0x820, 0x3F7C, 0x2444, 0x3F28, 0x2110, 0x3F28, 0x2446, 0x3FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {34263, 2, {0x440, 0xFFFE, 0x440, 0x4210, 0x2F90, 0x2210, 0xF9E, 0xAA2, 0xEA88, 0x2F88, 0x2208, 0x2714, 0x2A94, 0x3222, 0x5242, 0x8FFE}},
    {34264, 2, {0x820, 0xFFFE, 0x920, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2490, 0x27F0, 0x2490, 0x47F4, 0x4A4A, 0x93FA}},
    {34265, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x878, 0x7F48, 0x2248, 0x1486, 0x7F00, 0x10FC, 0x6944, 0x1644, 0x6D28, 0x1490, 0x6428, 0x18C6}},
    {34266, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x100, 0x3FF8, 0x2448, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {34267, 2, {0x820, 0xFFFE, 0xA20, 0x100, 0x7FFE, 0x4202, 0x9CF4, 0x1010, 0x1EF0, 0x1010, 0x1FF0, 0x800, 0x1FFC, 0x2004, 0xAA4, 0x12AC}},
    {34268, 2, {0x820, 0xFFFE, 0x820, 0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4908, 0x2C0, 0x1D30, 0xE92E, 0x540, 0x920, 0x1310}},
    {34269, 2, {0x820, 0xFFFE, 0x820, 0x3E20, 0x8F8, 0x7F28, 0x2A28, 0x5D6A, 0x82A, 0x3E56, 0x82, 0x3FF8, 0x0, 0xFFFE, 0x1010, 0x3FF8}},
    {34270, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7FFC, 0x800, 0x1FF0, 0x2810, 0xCFF0, 0x0, 0x7EFE, 0x2040, 0x7EFC, 0xA344, 0x3E7C, 0x2244}},
    {34271, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFEFE, 0x9292, 0xFEFE, 0x9292, 0xFEFE, 0x8282}},
    {34272, 2, {0x820, 0xFFFE, 0x920, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x2040, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x4284}},
    {34273, 2, {0x820, 0xFFFE, 0x8A0, 0x4040, 0x27FC, 0x2110, 0xA0, 0x7FC, 0xE444, 0x27FC, 0x2444, 0x25F4, 0x2514, 0x25F4, 0x540C, 0x8FFE}},
    {34274, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2220, 0x4AA0, 0x8ABE, 0x1FC4, 0x20A4, 0x6F24, 0xA028, 0x2FA8, 0x2210, 0x2FA8, 0x2244, 0x2F82}},
    {34275, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4444, 0x7FFC, 0x2498, 0x7EE0, 0x84, 0x7E7C, 0x4200, 0x7E98, 0x42E0, 0x7E84, 0x4284, 0x467C}},
    {34276, 2, {0x440, 0xFFFE, 0x440, 0x20, 0x3D24, 0x24A8, 0x25FC, 0x3C40, 0x27FE, 0x2488, 0x3D24, 0x26AA, 0x2470, 0x44A8, 0x5524, 0x8A62}},
    {34277, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2110, 0x4BA4, 0x72B8, 0x2390, 0x4AA4, 0x7BBC, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E, 0x100}},
    {34278, 2, {0x820, 0xFFFE, 0x8A0, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2FF8, 0x2080, 0x27F8, 0x2080, 0x5FFE, 0x4080, 0x8180}},
    {34279, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x103C, 0x13C0, 0x2544, 0x7CA8, 0x9FC, 0x1040, 0x27FE, 0x7C80, 0xF8, 0x5548, 0x5630, 0x80CE}},
    {34280, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x3FFC, 0x2440, 0x3FF8, 0x2448, 0x3FF8, 0x2840, 0x2F78, 0x2844, 0x2F3C, 0x4000, 0x4A48, 0x9124}},
    {34281, 2, {0x820, 0xFFFE, 0x820, 0x2038, 0x17C0, 0x1248, 0x8150, 0x47FC, 0x5248, 0x1404, 0x2BFA, 0xE248, 0x23F8, 0x2248, 0x23F8, 0x2208}},
    {34282, 2, {0x440, 0xFFFE, 0x440, 0x800, 0x7F20, 0x4920, 0xFFBE, 0x4944, 0x7FA4, 0x2A24, 0x3E28, 0x1028, 0xFF10, 0x2228, 0x1C44, 0x6282}},
    {34283, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x27FC, 0x10A0, 0x17FC, 0x84A4, 0x47FC, 0x5208, 0x13F8, 0x2208, 0xE3F8, 0x2040, 0x27FC, 0x2040}},
    {34284, 2, {0x820, 0xFFFE, 0x820, 0x7DFC, 0x4524, 0x49FC, 0x4820, 0x53FE, 0x4904, 0x45FC, 0x4504, 0x55FC, 0x4904, 0x41FC, 0x4088, 0x4104}},
    {34285, 2, {0x820, 0xFFFE, 0x920, 0x7FFE, 0x5022, 0xA814, 0x3EF8, 0x2208, 0x3E08, 0x22F8, 0x3E80, 0x22FC, 0xFE04, 0x1204, 0x2228, 0xCE10}},
    {34286, 2, {0x440, 0xFFFE, 0x440, 0x10A0, 0x1090, 0x11FE, 0xFD20, 0x13FC, 0x3120, 0x39FC, 0x5520, 0x51FE, 0x9000, 0x12A4, 0x1252, 0x1452}},
    {34287, 2, {0x820, 0xFFFE, 0x820, 0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5FA8, 0x4408, 0x9528, 0x2C90, 0x100, 0x4894, 0x87F2}},
    {34288, 2, {0x820, 0xFFFE, 0x820, 0x7784, 0x4284, 0x62A4, 0x54A4, 0x69A4, 0x1424, 0x2224, 0xDD24, 0x824, 0x7F24, 0x4A04, 0x2F94, 0xF008}},
    {34289, 2, {0x820, 0x820, 0xFFFE, 0x920, 0x80, 0x3FFC, 0x2100, 0xAFF8, 0x6888, 0x2FF8, 0x6888, 0xAFF8, 0x2150, 0x42F8, 0x4C8A, 0xB07E}},
    {34290, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x1000, 0x13FE, 0x1200, 0xFE7C, 0x1244, 0x327C, 0x3A00, 0x56EE, 0x52AA, 0x92EE, 0x1200, 0x13FE}},
    {34291, 2, {0x440, 0xFFFE, 0x540, 0x7FFC, 0x820, 0xFE0, 0x0, 0x7FFC, 0x4824, 0x4FEC, 0x0, 0x3FF8, 0x100, 0xFFFE, 0x920, 0x711C}},
    {34292, 2, {0x440, 0x440, 0xFFFE, 0x1440, 0x10F8, 0x2488, 0x7CF8, 0x888, 0x10F8, 0x2400, 0x7DFC, 0x154, 0x5554, 0x5554, 0x83FE, 0x0}},
    {34293, 2, {0x820, 0xFFFE, 0x820, 0xBF8, 0x1208, 0x33F8, 0x5040, 0x97FE, 0x1150, 0x1248, 0x100, 0xFFFE, 0x680, 0x1C48, 0xE530, 0x60E}},
    {34294, 2, {0x820, 0xFFFE, 0x820, 0x3FFE, 0x2104, 0x2E38, 0x2210, 0x2FBC, 0x2718, 0x2AB4, 0x3252, 0x2040, 0x247C, 0x4440, 0x4440, 0x9FFE}},
    {34295, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2020, 0x1022, 0xFDFC, 0x28, 0x7BFE, 0x40, 0x79FC, 0x684, 0x78FC, 0x4884, 0x78FC, 0x4884}},
    {34296, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x22, 0xFCFA, 0x2024, 0x6028, 0x95FE, 0x3440, 0x58FC, 0x9484, 0x35FC, 0x5684, 0x90FC, 0x2084}},
    {34297, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2000, 0x11FC, 0xFD04, 0x1FC, 0x7904, 0x1FC, 0x7880, 0x1FE, 0x7A22, 0x4D52, 0x79FA, 0x4806}},
    {34298, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4484, 0x4FE4, 0x5904, 0x4FC4, 0x4904, 0x4FC4, 0x4904, 0x4FEC}},
    {34299, 2, {0x820, 0xFFFE, 0x820, 0x23F8, 0x1208, 0x13F8, 0x8000, 0x47BC, 0x54A4, 0x17BC, 0x2040, 0xEFFE, 0x2150, 0x2248, 0x2C46, 0x2040}},
    {34300, 2, {0x820, 0xFFFE, 0x920, 0x7FFC, 0x4104, 0x3FF8, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x2448, 0x3FF8, 0x100, 0x4884, 0x4892, 0x87F2}},
    {34301, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x800, 0x7F7C, 0x2244, 0x147C, 0xFF44, 0x87C, 0x7F44, 0x87C, 0x2A28, 0x492A, 0xA94A, 0x1086}},
    {34302, 2, {0x440, 0x440, 0xFFFE, 0x1440, 0x1080, 0xFEF8, 0x1108, 0x7EFC, 0x5484, 0x54FC, 0x7C84, 0x10FC, 0x3884, 0x54FC, 0x9048, 0x1084}},
    {34303, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC}},
    {34304, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x11FC, 0x1154, 0xFDFC, 0x1020, 0x15FC, 0x1820, 0x33FE, 0xD088, 0x11FC, 0x1020, 0x53FE, 0x2020}},
    {34305, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x100, 0x3D78, 0x2548, 0x3D78, 0x100, 0x7FFC, 0x100, 0x3D78, 0x2548, 0x3D78, 0x100, 0xFFFE}},
    {34306, 2, {0x820, 0xFFFE, 0x920, 0x2488, 0x2424, 0x43E4, 0x2428, 0x1214, 0xA0A0, 0xA4A4, 0x1C1C, 0x100, 0xFFFE, 0x540, 0x1930, 0xE10E}},
    {34307, 2, {0x820, 0xFFFE, 0x920, 0x11F8, 0x1100, 0xFFFE, 0x810, 0x2E5C, 0x2850, 0x2E50, 0xF0FE, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {34308, 2, {0x820, 0xFFFE, 0x820, 0x770C, 0x5570, 0x7740, 0x40, 0x7F7E, 0x4948, 0x7F48, 0x4948, 0x7F48, 0x848, 0xFF88, 0x888, 0x908}},
    {34309, 2, {0x820, 0xFFFE, 0x820, 0x2200, 0x47DC, 0x8880, 0x17C0, 0x257E, 0x67C8, 0xA548, 0x27C8, 0x2108, 0x2FE8, 0x2108, 0x22A8, 0x2450}},
    {34310, 2, {0x820, 0xFFFE, 0x920, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2490, 0x27F0, 0x2490, 0x4FF8, 0x4948, 0x9FFC}},
    {34311, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x200C, 0x3E70, 0x4410, 0xFF10, 0x49FE, 0x7F10, 0x4938, 0x7F34, 0x54, 0x5552, 0x8A90, 0x10}},
    {34312, 2, {0x820, 0xFFFE, 0x820, 0x600, 0x38FE, 0x810, 0x7F7C, 0x1C44, 0x2A7C, 0x4944, 0x1C7C, 0x1444, 0x157C, 0x2628, 0x2444, 0x4082}},
    {34313, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0xFFFE, 0x9012, 0x13F0, 0x2010, 0x2FFC, 0x6804, 0xA7F0, 0x2210, 0x2120, 0x20C0, 0x2738}},
    {34314, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x10F8, 0x24A8, 0x7CA8, 0x8D8, 0x10F8, 0x2400, 0x7DFC, 0x154, 0x5554, 0x5554, 0x83FE, 0x0}},
    {34315, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x9FE, 0x2820, 0x2EFC, 0x2884, 0xFEFC, 0x884, 0x28FC, 0x2A84, 0x44FC, 0x800, 0x3048, 0xC084}},
    {34316, 2, {0x820, 0xFFFE, 0xA20, 0x27DE, 0x4912, 0x97D2, 0x2112, 0x65DA, 0xA514, 0x2FD0, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1110, 0x2308}},
    {34317, 2, {0x820, 0xFFFE, 0x8A0, 0x3F10, 0x410, 0x7FD0, 0x43E, 0x7FD2, 0x5552, 0x7FD2, 0x412, 0x7FD2, 0x412, 0xFFE2, 0x2AAA, 0x4544}},
    {34318, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3EF8, 0x22A8, 0x3EA8, 0x22F8, 0x3E84, 0x4284, 0x867C}},
    {34319, 2, {0x820, 0xFFFE, 0x820, 0x4000, 0x49FE, 0x7420, 0x44FC, 0x3C84, 0xFC, 0x7C84, 0xFC, 0xFE84, 0x10FC, 0x5400, 0xD248, 0x2084}},
    {34320, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x203C, 0x13C0, 0xFD44, 0xA8, 0x79FC, 0x40, 0x7BFE, 0x80, 0x78F8, 0x4948, 0x7A30, 0x48CE}},
    {34321, 2, {0x820, 0xFFFE, 0x8A0, 0x3FFE, 0x2210, 0x2FBC, 0x2318, 0x26B4, 0x2A52, 0x2000, 0x3FFC, 0x4400, 0x4FF8, 0x9408, 0x27F8, 0x408}},
    {34322, 2, {0x820, 0xFFFE, 0x8A0, 0x1CF0, 0xF110, 0x12F8, 0x10A8, 0xFCF8, 0x10A8, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x91FC, 0x1022, 0x101E}},
    {34323, 2, {0x820, 0x820, 0xFFFE, 0x820, 0xC80, 0x70F8, 0x1108, 0x1210, 0xFDFC, 0x3124, 0x39FC, 0x5524, 0x51FC, 0x9000, 0x1154, 0x122A}},
    {34324, 2, {0x440, 0xFFFE, 0x440, 0x4800, 0x75FE, 0x4420, 0x3CFC, 0x84, 0xCFC, 0x7084, 0x10FC, 0xFE84, 0x10FC, 0x3800, 0x5448, 0x9084}},
    {34325, 2, {0x440, 0xFFFE, 0x440, 0x2020, 0x247C, 0x4AA4, 0xF818, 0x10E6, 0x2610, 0x4AFE, 0xFA10, 0x27C, 0xAA10, 0xAAFE, 0x510, 0x8FE}},
    {34326, 2, {0x440, 0xFFFE, 0x440, 0x2020, 0x7DFC, 0x4488, 0x7C50, 0x41FE, 0x7C20, 0x45FC, 0x7C20, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {34327, 2, {0x820, 0xFFFE, 0x820, 0x40, 0x7C20, 0x45FC, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x3C20, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {34328, 2, {0x820, 0xFFFE, 0x820, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x2288, 0x3EF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x490, 0x1C60, 0x671C}},
    {34329, 2, {0x820, 0xFFFE, 0x820, 0x7F78, 0x4848, 0x5E86, 0x6478, 0x5F48, 0x4A30, 0x7F4C, 0x0, 0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x60C}},
    {34330, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2044, 0x3E28, 0x44FE, 0xFF10, 0x4910, 0x7F7C, 0x4910, 0x7F10, 0xFE, 0x5510, 0x4A90, 0x8A90}},
    {34331, 2, {0x820, 0xFFFE, 0x820, 0x2410, 0x24FE, 0xFF20, 0x2444, 0x3CFE, 0x2402, 0x3C7C, 0x2444, 0xFF7C, 0x5444, 0x627C, 0x4044, 0x7E4C}},
    {34332, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2900, 0x7DFC, 0x2904, 0x3A44, 0x1154, 0x7CE4, 0x55F4, 0x7C44, 0x10E4, 0xFD54, 0x1054, 0x1008}},
    {34333, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x1420, 0x223E, 0x5D42, 0x8084, 0x7710, 0x5510, 0x7710, 0x2228, 0x2228, 0x5544, 0x8882}},
    {34334, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x820, 0x1420, 0x223E, 0x5D44, 0x80A4, 0x7724, 0x5528, 0x7728, 0x2210, 0x2228, 0x5544, 0x8882}},
    {34335, 2, {0x440, 0xFFFE, 0x440, 0x1FC, 0x7CA8, 0x45FC, 0x4820, 0x51FC, 0x4804, 0x45FC, 0x4404, 0x55FC, 0x4840, 0x42A4, 0x428A, 0x447A}},
    {34336, 2, {0x820, 0xFFFE, 0x820, 0x1040, 0x57FC, 0x5248, 0x5554, 0x7FFE, 0x1000, 0x17FC, 0xF404, 0x55F4, 0x5514, 0x55F4, 0x9404, 0x17FC}},
    {34337, 2, {0x820, 0xFFFE, 0x820, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2288, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {34338, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x840, 0x7E7C, 0x2440, 0xFF7C, 0x4, 0x7E7C, 0x4240, 0x7E7C, 0x4240, 0x7E7C, 0x4242, 0x463E}},
    {34339, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x2220, 0x7F10, 0x227E, 0xFF90, 0x810, 0x7F10, 0x497C, 0x7F10, 0x4910, 0x7F10, 0x22FE, 0x4100}},
    {34340, 2, {0x440, 0xFFFE, 0x440, 0x103C, 0x27C0, 0x7A44, 0x4928, 0x4BF8, 0x4A08, 0x7BFC, 0x4A04, 0x4BFE, 0x7A02, 0x4CAA, 0x4AA, 0x906}},
    {34341, 2, {0x820, 0xFFFE, 0x920, 0x280, 0xC60, 0x37D8, 0xC006, 0x3BB8, 0x2AA8, 0x3BB8, 0x0, 0x3FF8, 0x2488, 0x3FF8, 0x2488, 0x24B8}},
    {34342, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3BB8, 0x2AA8, 0x3BB8, 0x2AA8}},
    {34343, 2, {0x820, 0xFFFE, 0x8A0, 0xF8, 0x4080, 0x2FFC, 0x2884, 0xBF0, 0xE888, 0x2BF8, 0x2940, 0x2AA8, 0x3170, 0x22A8, 0x5160, 0x8FFE}},
    {34344, 2, {0x820, 0xFFFE, 0x820, 0x70C, 0xF870, 0x5110, 0x2A24, 0x2078, 0x3E10, 0x4824, 0xFF7E, 0x812, 0x4954, 0x4F52, 0x7192, 0x30}},
    {34345, 2, {0x820, 0xFFFE, 0x2820, 0x3F20, 0x403E, 0xBE48, 0x2AA8, 0xFF28, 0x4A10, 0x7F28, 0x846, 0x1F80, 0x210, 0x3FF8, 0x1110, 0x2308}},
    {34346, 2, {0x820, 0xFFFE, 0x920, 0x3FFC, 0x2440, 0x3FF8, 0x2448, 0x3FF8, 0x2840, 0x2F78, 0x2844, 0x2F3C, 0x4490, 0x5FFC, 0x82A0, 0x1C9C}},
    {34347, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x27D0, 0x1510, 0x179E, 0x84A0, 0x47A8, 0x5544, 0x17C4, 0x2000, 0xE7F8, 0x2528, 0x2528, 0x2FFE}},
    {34348, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x7DFC, 0x4404, 0x7CFC, 0x4004, 0x7DFC, 0x4400, 0x7DFE, 0x1122, 0x5CFC, 0x50A4, 0x5EAC, 0xF020}},
    {34349, 2, {0x820, 0xFFFE, 0x820, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x5FF4, 0x4104, 0x5FF4, 0x5554, 0x5FF4, 0x4544, 0x592C}},
    {34350, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x3FFE, 0x2204, 0x2A94, 0x2514, 0x2894, 0x2254, 0x4A84, 0x4514, 0x8888}},
    {34351, 2, {0x820, 0xFFFE, 0x820, 0x21F8, 0x1108, 0x41F8, 0x2108, 0xBFE, 0x900, 0x73FC, 0x14A4, 0x1154, 0x1288, 0x3FF8, 0x2448, 0xFFFE}},
    {34352, 2, {0x440, 0x440, 0xFFFE, 0x440, 0x21FC, 0x2104, 0x49FC, 0xF904, 0x13FE, 0x2252, 0x4BFE, 0xF800, 0x1FC, 0xA888, 0xA870, 0x38E}},
    {34353, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x93FE, 0x5420, 0xFEFC, 0x3884, 0x54FC, 0x9284, 0xFC, 0x1484, 0xFEFC, 0x2800, 0x4448, 0x8084}},
    {34354, 2, {0x440, 0xFFFE, 0x440, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x100, 0x3FF8, 0x100, 0xFFFE}},
    {34355, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x2210, 0x7F7C, 0x2210, 0xFF90, 0x87E, 0x7F00, 0x4910, 0x7F10, 0x497C, 0x7F10, 0x227E, 0x4100}},
    {34356, 2, {0x820, 0xFFFE, 0x820, 0x1D70, 0x4924, 0x5D74, 0x4924, 0x5D74, 0x4924, 0x7FFC, 0x0, 0x7FFC, 0x1010, 0x1FF0, 0x820, 0xFFFE}},
    {34357, 2, {0x820, 0xFFFE, 0x820, 0x94, 0xFDD2, 0x4812, 0x4950, 0x7BFE, 0x4810, 0x49D4, 0x7954, 0x49D4, 0x4F48, 0xF9CA, 0x956, 0x822}},
    {34358, 2, {0x440, 0xFFFE, 0x440, 0xC20, 0x7040, 0x11FC, 0xFD24, 0x39FC, 0x5524, 0x93FC, 0xFC24, 0x245A, 0x685E, 0x1090, 0x2912, 0xC60E}},
    {34359, 2, {0x440, 0xFFFE, 0x440, 0xFFFE, 0x1100, 0x13DE, 0x5A52, 0x53DE, 0x5250, 0x5BD2, 0xE40E, 0xFE0, 0x5444, 0x8382, 0x1C70, 0xE00E}},
    {34360, 2, {0x820, 0xFFFE, 0x820, 0x50, 0xFE48, 0x10FE, 0x1190, 0x7CFC, 0x5490, 0x54FC, 0x6C90, 0x44FE, 0x7C80, 0x4554, 0x7D2A, 0x462A}},
    {34361, 2, {0x820, 0xFFFE, 0x8A0, 0x1040, 0x17FC, 0x1000, 0x5BF8, 0x56A8, 0x53F8, 0x9244, 0x1158, 0x1244, 0x10A8, 0x1190, 0x16A8, 0x10C6}},
    {34362, 2, {0x440, 0xFFFE, 0x440, 0x1000, 0x828, 0x7F24, 0x147E, 0x4948, 0x55C8, 0x7F7E, 0x848, 0x7F7E, 0x5148, 0x5548, 0x5D7E, 0x4340}},
    {34363, 2, {0x820, 0xFFFE, 0x820, 0x7F3C, 0x2A24, 0x3E42, 0x2A3C, 0xFFA4, 0x4918, 0x7F66, 0x840, 0x1F80, 0x210, 0x3FF8, 0x1110, 0x2308}},
    {34364, 2, {0x820, 0xFFFE, 0x8A0, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2120, 0x3F3E, 0x2120, 0x2F3C, 0x4120, 0x5F3E, 0x8120}},
    {34365, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x3FF8, 0x2AA8, 0x3EF8, 0x2AA8, 0x3EF8, 0x100, 0x7FFC, 0x920, 0x711C}},
    {34366, 2, {0x820, 0xFFFE, 0x8A0, 0x2040, 0x27FC, 0x2000, 0x23F8, 0xFAA8, 0x23F8, 0x2244, 0x2118, 0x2244, 0x38A8, 0xE190, 0x46A8, 0xC6}},
    {34367, 2, {0x820, 0xFFFE, 0x820, 0x7FFC, 0x4444, 0x7FFC, 0x10A0, 0x2090, 0x45FE, 0x7920, 0x13FC, 0x2520, 0x7DFC, 0x120, 0x55FE, 0x8100}},
    {34368, 2, {0x820, 0xFFFE, 0x920, 0xFFFE, 0x288, 0x7D70, 0x2548, 0x4D66, 0x8280, 0x7EFC, 0x280, 0x3EF8, 0x280, 0x7EFC, 0x280, 0xFFFE}},
    {34369, 2, {0x440, 0xFFFE, 0x440, 0xFFFE, 0x1100, 0x13DE, 0x5A52, 0x53DE, 0x5250, 0x5BD2, 0xE64E, 0xFE0, 0x5444, 0x8382, 0x1C70, 0xE00E}},
    {34370, 2, {0x820, 0xFFFE, 0x820, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x40, 0x3EFC, 0x2348, 0x3E30, 0x9CE, 0x2E78, 0x2848, 0x7E78}},
    {34371, 2, {0x820, 0xFFFE, 0x820, 0x2294, 0x1292, 0x1550, 0x8FFE, 0x4290, 0x5EF4, 0x1294, 0x2EF4, 0xE294, 0x2EE8, 0x228A, 0x22F6, 0x2F22}},
    {34372, 2, {0x820, 0xFFFE, 0x8A0, 0x21F8, 0x2288, 0x2070, 0xFBAE, 0xA8F8, 0xA850, 0xFBFE, 0x22A0, 0x2AFC, 0x3B20, 0xCA78, 0x420, 0x9FE}},
    {34373, 2, {0x820, 0xFFFE, 0x820, 0x124, 0x3CA8, 0x25FC, 0x2440, 0x3FFE, 0x2488, 0x25FC, 0x3EA2, 0x24F8, 0x24A0, 0x44FC, 0x5404, 0x8AAC}},
    {34374, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x7FFC, 0x4AA4, 0x7EFC, 0x4AA4, 0x7EFC, 0x2008, 0x7FFC, 0x1110, 0x2308}},
    {34375, 2, {0x820, 0xFFFE, 0x8A8, 0x7BFE, 0x888, 0xBDE, 0x7A52, 0x43DE, 0x40A0, 0x79FE, 0x920, 0xBFC, 0xD20, 0x9FC, 0x5120, 0x21FE}},
    {34376, 2, {0x820, 0xFFFE, 0x820, 0x0, 0x7BDE, 0x4892, 0x4912, 0x4BD2, 0x7A5E, 0x3C0, 0x7A5E, 0x4BD2, 0x4A52, 0x4BD2, 0x799E, 0x4A52}},
    {34377, 2, {0x820, 0xFFFE, 0x820, 0x7F10, 0x7C, 0x3E44, 0x227C, 0x3E44, 0x7C, 0x7F40, 0x557E, 0x4940, 0x7F7E, 0x4902, 0x49AA, 0x4306}},
    {34378, 2, {0x820, 0xFFFE, 0x820, 0x2108, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0x100, 0x7FFC, 0x540, 0x1930, 0xE10E}},
    {34379, 2, {0x820, 0xFFFE, 0x820, 0x77DC, 0x4444, 0x77DC, 0x4444, 0xFFFE, 0x9292, 0x1C70, 0x1010, 0x1FF0, 0x820, 0x1FF0, 0x642C, 0x860}},
    {34380, 2, {0x820, 0xFFFE, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x5FF6, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x228A, 0x3DFE}},
    {34381, 2, {0x100, 0x100, 0x1F8, 0x100, 0x100, 0x3FFC, 0x2004, 0x2108, 0x2100, 0x2118, 0x21E0, 0x2F00, 0x2104, 0x4104, 0x40FC, 0x8000}},
    {34382, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2000, 0x23E0, 0x2220, 0x2220, 0x4424, 0x4824, 0x901C}},
    {34383, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2100, 0x2FF8, 0x2108, 0x2108, 0x4208, 0x4450, 0x9820}},
    {34384, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2000, 0x27F8, 0x2400, 0x3FFE, 0x4400, 0x47F8, 0x8000}},
    {34385, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2000, 0x2080, 0x2044, 0x2A42, 0x4A12, 0x5210, 0x81F0}},
    {34386, 2, {0x8, 0x7C, 0x3F80, 0x2080, 0x20F8, 0x2080, 0x2FFE, 0x2882, 0x28F0, 0x2F84, 0x287C, 0x2800, 0x4BE0, 0x5222, 0x9422, 0x281E}},
    {34387, 2, {0x2020, 0x203E, 0x2020, 0x21FE, 0xF922, 0x2938, 0x29E0, 0x2922, 0x291E, 0x2900, 0x2B78, 0x2D48, 0x4948, 0x424A, 0x824A, 0x486}},
    {34388, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2100, 0x2080, 0x2FF8, 0x2220, 0x2140, 0x4080, 0x4360, 0x8C1C}},
    {34389, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2400, 0x2438, 0x27A8, 0x28AA, 0x352A, 0x4246, 0x4580, 0x987E}},
    {34390, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2010, 0x27E0, 0x2248, 0x2150, 0x2FFE, 0x4040, 0x4140, 0x8080}},
    {34391, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x201C, 0x27E0, 0x2400, 0x27FE, 0x4420, 0x4420, 0x9FFE}},
    {34392, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x47F8, 0x4408, 0xBFFE}},
    {34393, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2100, 0x2088, 0x2A24, 0x2A44, 0x5380, 0x4608, 0x99F8}},
    {34394, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2160, 0x2F88, 0x2108, 0x20F8, 0x2240, 0x3248, 0x2A48, 0x2A50, 0x4240, 0x5FFC, 0x8000}},
    {34395, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2120, 0x2924, 0x2924, 0x2F3C, 0x2120, 0x4120, 0x5FFE, 0x8000}},
    {34396, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2490, 0x27F0, 0x2490, 0x27F0, 0x4100, 0x5FF8, 0x8408, 0x818}},
    {34397, 2, {0x820, 0xF20, 0x820, 0x7F7E, 0x496A, 0x4EAA, 0x782A, 0x492A, 0x474A, 0x4092, 0x4E22, 0x4A54, 0x4A0A, 0x5202, 0x91FE, 0x2000}},
    {34398, 2, {0x100, 0x1F8, 0x100, 0x3FFC, 0x2104, 0x2FE0, 0x2108, 0x27F8, 0x2410, 0x27F0, 0x2000, 0x2FF8, 0x2080, 0x5FFC, 0x4220, 0x9C1C}},
    {34399, 2, {0x20, 0x783E, 0x4820, 0x49FE, 0x4922, 0x7938, 0x1E0, 0xFD22, 0x211E, 0x4100, 0x7978, 0x948, 0x948, 0xA4A, 0x524A, 0x2486}},
    {34400, 2, {0x2010, 0x101E, 0x10, 0xFEFE, 0x92, 0x289C, 0x44F0, 0x8292, 0x48E, 0x4480, 0x28B8, 0x10A8, 0x28A8, 0x452A, 0x852A, 0x246}},
    {34401, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x2120, 0x2924, 0x2528, 0x2120, 0x5FFE, 0x4210, 0x8408}},
    {34402, 2, {0x410, 0x1E1E, 0xE010, 0x22FE, 0x9292, 0x549C, 0x40F0, 0x492, 0xFF8E, 0x480, 0x44B8, 0x24A8, 0x24A8, 0x52A, 0x152A, 0xA46}},
    {34403, 2, {0x810, 0xC1E, 0xEA10, 0xAFE, 0x892, 0xFE9C, 0x28F0, 0x2892, 0xB88E, 0xA880, 0xAAB8, 0xAAA8, 0xB6A8, 0xC32A, 0x812A, 0x246}},
    {34404, 2, {0x1010, 0x1E1E, 0x1010, 0x7EFE, 0x5292, 0x5C9C, 0x70F0, 0x5292, 0x4E8E, 0x4080, 0x5CB8, 0x54A8, 0x54A8, 0x972A, 0x952A, 0x2246}},
    {34405, 2, {0x1028, 0x1E24, 0x102E, 0x7EF0, 0x5228, 0x5C12, 0x702A, 0x52D6, 0x4E28, 0x402E, 0x5CF0, 0x5424, 0x5428, 0x5612, 0x942A, 0x20C6}},
    {34406, 2, {0x2810, 0x241E, 0x2E10, 0xF0FE, 0x2892, 0x129C, 0x2AF0, 0xD692, 0x288E, 0x2E80, 0xF0B8, 0x24A8, 0x28A8, 0x12AA, 0x2B2A, 0xC646}},
    {34407, 2, {0x800, 0xFBC, 0x800, 0x7FC0, 0x4840, 0x4F3E, 0x7890, 0x4790, 0x5220, 0x5FBE, 0x7202, 0x5F82, 0x5202, 0x5F82, 0x5214, 0x9F88}},
    {34408, 2, {0x802, 0xF82, 0x804, 0x7FC8, 0x4850, 0x4F02, 0x7882, 0x4784, 0x4008, 0x4890, 0x7DE2, 0x4882, 0x5DC4, 0x6AA8, 0x4890, 0x88A0}},
    {34409, 2, {0x1020, 0x923E, 0x5220, 0x55FE, 0x1122, 0x7D38, 0x45E0, 0x7D22, 0x451E, 0x7D00, 0x1178, 0x5948, 0x5548, 0x964A, 0x524A, 0x2486}},
    {34410, 2, {0x2120, 0x213C, 0x3964, 0x2318, 0xFA66, 0xAABE, 0xB2AA, 0xE6BE, 0xAAAA, 0x9ABE, 0x8288, 0xBABE, 0xAA88, 0xAE7E, 0x2A00, 0x42AA}},
    {34411, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x100, 0x110, 0x108, 0x1FC, 0x7E04, 0x0}},
    {34412, 2, {0x840, 0x840, 0x840, 0x7F40, 0x4940, 0x4940, 0x4940, 0x4940, 0x7F40, 0x4840, 0x840, 0xA44, 0xF44, 0xF144, 0x403C, 0x0}},
    {34413, 2, {0x800, 0x800, 0x8FE, 0x7F22, 0x4922, 0x4922, 0x4922, 0x4922, 0x7F22, 0x4822, 0x822, 0xA42, 0xF42, 0xF182, 0x4114, 0x208}},
    {34414, 2, {0x1000, 0x1000, 0x10F0, 0x7C90, 0x5490, 0x5490, 0x5490, 0x5490, 0x7C90, 0x5090, 0x1090, 0x1492, 0x1C92, 0xE512, 0x410E, 0x200}},
    {34415, 2, {0x804, 0x804, 0x844, 0x7F44, 0x4944, 0x4944, 0x4944, 0x4944, 0x7F44, 0x484C, 0x874, 0xA44, 0xF04, 0xF104, 0x4004, 0x4}},
    {34416, 2, {0x800, 0x800, 0x8FE, 0x7F10, 0x4910, 0x4910, 0x4910, 0x4910, 0x7F10, 0x4810, 0x810, 0xA10, 0xF10, 0xF110, 0x4050, 0x20}},
    {34417, 2, {0x0, 0x7FF0, 0x90, 0x3D0, 0x7C10, 0x410, 0x3F90, 0x2490, 0x2490, 0x2490, 0x3F90, 0x412, 0x48A, 0x7CA, 0x7846, 0x2002}},
    {34418, 2, {0x820, 0x820, 0x820, 0x7F20, 0x4920, 0x4930, 0x4928, 0x4924, 0x7F22, 0x4822, 0x820, 0xA20, 0xF20, 0xF120, 0x4020, 0x20}},
    {34419, 2, {0x820, 0x820, 0x820, 0x7F7C, 0x4944, 0x4984, 0x4904, 0x4944, 0x7F24, 0x4814, 0x814, 0xA04, 0xF04, 0xF104, 0x4028, 0x10}},
    {34420, 2, {0x1008, 0x101C, 0x10E0, 0x7C20, 0x5420, 0x5420, 0x5420, 0x543E, 0x7DE0, 0x5020, 0x1020, 0x1422, 0x1E22, 0xE222, 0x401E, 0x0}},
    {34421, 2, {0x1020, 0x1020, 0x1020, 0x7D20, 0x552C, 0x5534, 0x5564, 0x57A4, 0x7D24, 0x5134, 0x1128, 0x1522, 0x1D22, 0xE502, 0x40FE, 0x0}},
    {34422, 2, {0x1000, 0x11FC, 0x1020, 0x7C20, 0x5420, 0x5420, 0x5420, 0x57FE, 0x7C20, 0x5020, 0x1020, 0x1420, 0x1E20, 0xE220, 0x40A0, 0x40}},
    {34423, 2, {0x800, 0x87C, 0x810, 0x7F10, 0x4910, 0x4910, 0x4910, 0x49FE, 0x7F10, 0x4810, 0x810, 0xA10, 0xF10, 0xF110, 0x4010, 0x10}},
    {34424, 2, {0x1000, 0x11FC, 0x1004, 0x7C08, 0x5410, 0x5420, 0x5420, 0x57FE, 0x7C20, 0x5020, 0x1020, 0x1420, 0x1E20, 0xE220, 0x40A0, 0x40}},
    {34425, 2, {0x1000, 0x1000, 0x11FC, 0x7C20, 0x5420, 0x5420, 0x5420, 0x5420, 0x7C20, 0x5020, 0x1020, 0x1420, 0x1E20, 0xE220, 0x43FE, 0x0}},
    {34426, 2, {0x20, 0x20, 0xFE20, 0x25FC, 0x2524, 0x2524, 0x2524, 0x25FC, 0x2420, 0x2420, 0x2428, 0x25FC, 0x2404, 0x4402, 0x43FE, 0x8000}},
    {34427, 2, {0x1020, 0x1010, 0x1010, 0x7C00, 0x55FE, 0x5480, 0x5480, 0x5480, 0x7C80, 0x5080, 0x1080, 0x1480, 0x1E80, 0xE280, 0x40FC, 0x0}},
    {34428, 2, {0x1040, 0x1040, 0x1080, 0x7CFE, 0x5500, 0x5600, 0x54FC, 0x5408, 0x7C10, 0x5020, 0x1040, 0x1480, 0x1D02, 0xE502, 0x40FE, 0x0}},
    {34429, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x120, 0x110, 0x1F8, 0x7F04, 0x2004}},
    {34430, 2, {0x1000, 0x13FE, 0x1040, 0x7C40, 0x5440, 0x5440, 0x5450, 0x5448, 0x7C44, 0x5044, 0x1040, 0x1440, 0x1E40, 0xE240, 0x4040, 0x40}},
    {34431, 2, {0x0, 0xFFFE, 0x400, 0x7F0, 0x810, 0x1010, 0x20A0, 0x4140, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x108, 0x7FFC, 0x2004}},
    {34432, 2, {0x2020, 0x2020, 0x2020, 0x3DFC, 0x4524, 0x4924, 0xA124, 0x2124, 0x21FC, 0x2124, 0x2020, 0x2028, 0x2824, 0x33FE, 0x2102, 0x0}},
    {34433, 2, {0x1040, 0x1024, 0x1124, 0x7D04, 0x5504, 0x5488, 0x5488, 0x5488, 0x7C50, 0x5050, 0x1020, 0x1420, 0x1C50, 0xE488, 0x4104, 0x602}},
    {34434, 2, {0x1000, 0x11F8, 0x1008, 0x7C08, 0x5488, 0x5488, 0x5488, 0x54FE, 0x7C02, 0x5002, 0x1002, 0x15FA, 0x1C02, 0xE402, 0x4014, 0x8}},
    {34435, 2, {0x400, 0x840, 0x1F80, 0x308, 0xC10, 0x3FE0, 0xC0, 0x700, 0x7900, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34436, 2, {0x1040, 0x1020, 0x1020, 0x7DFE, 0x5440, 0x5440, 0x5440, 0x547C, 0x7C44, 0x5044, 0x1044, 0x1444, 0x1E84, 0xE284, 0x4128, 0x210}},
    {34437, 2, {0x1000, 0x11FE, 0x1100, 0x7D00, 0x5578, 0x5548, 0x5548, 0x5548, 0x7D48, 0x5168, 0x1150, 0x1542, 0x1D42, 0xE542, 0x423E, 0x400}},
    {34438, 2, {0x1000, 0x1000, 0x11FC, 0x7D24, 0x5524, 0x5524, 0x5524, 0x5524, 0x7DFC, 0x5100, 0x1100, 0x1500, 0x1D02, 0xE502, 0x40FE, 0x0}},
    {34439, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x5484, 0x5484, 0x54FC, 0x54A0, 0x7CA0, 0x50A0, 0x1090, 0x1490, 0x1D08, 0xE508, 0x4204, 0x402}},
    {34440, 2, {0x1000, 0x11FC, 0x1088, 0x7C88, 0x5488, 0x5488, 0x5488, 0x57FE, 0x7C88, 0x5088, 0x1088, 0x1488, 0x1C88, 0xE508, 0x4108, 0x208}},
    {34441, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE10E, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34442, 2, {0x1040, 0x1020, 0x1020, 0x7DFE, 0x5488, 0x5488, 0x5488, 0x5488, 0x7C88, 0x5050, 0x1050, 0x1420, 0x1E20, 0xE250, 0x4088, 0x106}},
    {34443, 2, {0x1020, 0x1020, 0x1020, 0x7DFE, 0x5522, 0x5522, 0x5522, 0x5522, 0x7D52, 0x514A, 0x118A, 0x1502, 0x1D02, 0xE502, 0x410A, 0x104}},
    {34444, 2, {0x1020, 0x1020, 0x1020, 0x7DFC, 0x5420, 0x5420, 0x5420, 0x55FC, 0x7C20, 0x5020, 0x1020, 0x17FE, 0x1C20, 0xE420, 0x4020, 0x20}},
    {34445, 2, {0x1010, 0x1110, 0x1110, 0x7D12, 0x5512, 0x5514, 0x55D8, 0x5510, 0x7D10, 0x5110, 0x1110, 0x1512, 0x1D52, 0xE592, 0x410E, 0x0}},
    {34446, 2, {0x1000, 0x1000, 0x10FC, 0x7C84, 0x5484, 0x5484, 0x5484, 0x5484, 0x7CFC, 0x5084, 0x1084, 0x1484, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {34447, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x5484, 0x54FC, 0x5484, 0x5484, 0x7C84, 0x50FC, 0x1084, 0x1484, 0x1E84, 0xE284, 0x4114, 0x208}},
    {34448, 2, {0x1040, 0x1040, 0x1080, 0x7CFC, 0x5504, 0x5604, 0x5484, 0x5444, 0x7C44, 0x5014, 0x1024, 0x1444, 0x1D84, 0xE404, 0x4028, 0x10}},
    {34449, 2, {0x1020, 0x1020, 0x1020, 0x7DFE, 0x5420, 0x5420, 0x5420, 0x55FC, 0x7C84, 0x5088, 0x1048, 0x1450, 0x1E20, 0xE250, 0x4188, 0x606}},
    {34450, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x54A4, 0x5494, 0x5494, 0x5484, 0x7DFE, 0x5084, 0x1084, 0x1484, 0x1C84, 0xE504, 0x4114, 0x208}},
    {34451, 2, {0x1004, 0x11E4, 0x1024, 0x7C24, 0x5424, 0x55E4, 0x5504, 0x5504, 0x7D04, 0x51E4, 0x1024, 0x1424, 0x1C24, 0xE424, 0x4144, 0x84}},
    {34452, 2, {0x1004, 0x101E, 0x11F0, 0x7D10, 0x5510, 0x5510, 0x5510, 0x55FE, 0x7D10, 0x5110, 0x1110, 0x1508, 0x1D0A, 0xE54A, 0x4186, 0x102}},
    {34453, 2, {0x0, 0x3FF8, 0x100, 0x200, 0xFFFE, 0x440, 0x820, 0x3118, 0xDFF6, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34454, 2, {0x1000, 0x1000, 0x11FC, 0x7C00, 0x5400, 0x5400, 0x57FE, 0x5490, 0x7C90, 0x5090, 0x1090, 0x1490, 0x1D12, 0xE512, 0x420E, 0x400}},
    {34455, 2, {0x1020, 0x1020, 0x1020, 0x7DFC, 0x5424, 0x5424, 0x5424, 0x5424, 0x7DFE, 0x5020, 0x1050, 0x1450, 0x1E88, 0xE288, 0x4104, 0x202}},
    {34456, 2, {0x1048, 0x1044, 0x1044, 0x7C40, 0x57FE, 0x5450, 0x5450, 0x5450, 0x7C50, 0x5050, 0x1090, 0x1492, 0x1C92, 0xE512, 0x410E, 0x200}},
    {34457, 2, {0x1020, 0x1020, 0x1050, 0x7C50, 0x5488, 0x5544, 0x5622, 0x5420, 0x7C00, 0x51FC, 0x1004, 0x1408, 0x1E08, 0xE210, 0x4010, 0x20}},
    {34458, 2, {0x1008, 0x101C, 0x10E0, 0x7C80, 0x5480, 0x5480, 0x54FE, 0x5488, 0x7C88, 0x5088, 0x1088, 0x1488, 0x1D08, 0xE508, 0x4208, 0x408}},
    {34459, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x55FC, 0x5524, 0x5524, 0x5524, 0x7D24, 0x51FC, 0x1124, 0x1420, 0x1E20, 0xE220, 0x4020, 0x20}},
    {34460, 2, {0x1000, 0x11FE, 0x1008, 0x7C88, 0x5488, 0x5488, 0x5508, 0x55FE, 0x7C18, 0x5028, 0x1048, 0x1488, 0x1D08, 0xE608, 0x4028, 0x10}},
    {34461, 2, {0x1008, 0x103C, 0x11E0, 0x7C20, 0x5420, 0x543C, 0x55E0, 0x5420, 0x7C20, 0x503E, 0x13E0, 0x1420, 0x1E22, 0xE222, 0x4022, 0x1E}},
    {34462, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x55FE, 0x5420, 0x5470, 0x5470, 0x7CA8, 0x50A8, 0x1124, 0x1524, 0x1E22, 0xE420, 0x4020, 0x20}},
    {34463, 2, {0x1000, 0x1000, 0x11FC, 0x7C20, 0x5420, 0x5420, 0x5420, 0x5420, 0x7DFC, 0x5020, 0x1020, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34464, 2, {0x440, 0x820, 0x3018, 0xDFE6, 0x420, 0x420, 0x8A0, 0x1040, 0x2100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34465, 2, {0x1010, 0x1090, 0x1090, 0x7C88, 0x5508, 0x5504, 0x5604, 0x55FA, 0x7C88, 0x5088, 0x1088, 0x1488, 0x1D08, 0xE508, 0x4228, 0x410}},
    {34466, 2, {0x1040, 0x1020, 0x1000, 0x7DFC, 0x5400, 0x5400, 0x54F0, 0x5490, 0x7C90, 0x5090, 0x1090, 0x1492, 0x1C92, 0xE512, 0x410E, 0x200}},
    {34467, 2, {0x1008, 0x1048, 0x1048, 0x7C48, 0x5444, 0x5484, 0x54A4, 0x5522, 0x7C20, 0x5040, 0x1040, 0x1448, 0x1C84, 0xE5FE, 0x4082, 0x0}},
    {34468, 2, {0x0, 0x3FF0, 0x1210, 0x920, 0x440, 0x380, 0x1C70, 0xE10E, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34469, 2, {0x1010, 0x1088, 0x1084, 0x7C84, 0x5502, 0x5602, 0x5488, 0x5488, 0x7C88, 0x5050, 0x1050, 0x1420, 0x1E50, 0xE288, 0x4104, 0x202}},
    {34470, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x5484, 0x55FE, 0x5484, 0x5484, 0x7C84, 0x51FE, 0x1084, 0x1484, 0x1E84, 0xE284, 0x4094, 0x88}},
    {34471, 2, {0x1020, 0x1020, 0x1050, 0x7C50, 0x5488, 0x5504, 0x5602, 0x5488, 0x7C88, 0x5088, 0x1088, 0x1488, 0x1C88, 0xE508, 0x4108, 0x208}},
    {34472, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x55FC, 0x5420, 0x5420, 0x5420, 0x7DFE, 0x5020, 0x1050, 0x1450, 0x1E88, 0xE288, 0x4104, 0x202}},
    {34473, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x108, 0x7FFC, 0x2004}},
    {34474, 2, {0x1010, 0x1110, 0x1090, 0x7C90, 0x5410, 0x5510, 0x5490, 0x5490, 0x7C10, 0x501E, 0x13F0, 0x1410, 0x1E10, 0xE210, 0x4010, 0x10}},
    {34475, 2, {0x1040, 0x1040, 0x10FC, 0x7C84, 0x5504, 0x56F4, 0x5494, 0x5494, 0x7CF4, 0x5084, 0x1094, 0x1488, 0x1E82, 0xE282, 0x407E, 0x0}},
    {34476, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x5524, 0x5524, 0x5524, 0x5524, 0x7D24, 0x5154, 0x1050, 0x1490, 0x1C90, 0xE512, 0x4212, 0x40E}},
    {34477, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x5504, 0x55FC, 0x5500, 0x5540, 0x7D44, 0x5148, 0x1170, 0x1540, 0x1D42, 0xE642, 0x423E, 0x400}},
    {34478, 2, {0x1050, 0x1054, 0x1052, 0x7C92, 0x5490, 0x559E, 0x55F0, 0x5690, 0x7C90, 0x5090, 0x1090, 0x1490, 0x1E8A, 0xE28A, 0x4086, 0x82}},
    {34479, 2, {0x1004, 0x100E, 0x10F0, 0x7C80, 0x5480, 0x5480, 0x5480, 0x54FE, 0x7C88, 0x5088, 0x1088, 0x1488, 0x1C88, 0xE488, 0x43FE, 0x0}},
    {34480, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x55FC, 0x5524, 0x5524, 0x5524, 0x7D24, 0x51FC, 0x1124, 0x1524, 0x1D24, 0xE524, 0x41FC, 0x104}},
    {34481, 2, {0x1080, 0x1080, 0x1080, 0x7CFE, 0x5540, 0x5540, 0x5640, 0x547C, 0x7C40, 0x5040, 0x1040, 0x147E, 0x1E40, 0xE240, 0x4040, 0x40}},
    {34482, 2, {0x1000, 0x11FC, 0x1020, 0x7C20, 0x5524, 0x54A4, 0x54A8, 0x5420, 0x7FFE, 0x5020, 0x1020, 0x1420, 0x1E20, 0xE220, 0x4020, 0x20}},
    {34483, 2, {0x1004, 0x101E, 0x11F0, 0x7D10, 0x5510, 0x5510, 0x5510, 0x55FE, 0x7D10, 0x5110, 0x1110, 0x1508, 0x1D0A, 0xE54A, 0x41A6, 0x112}},
    {34484, 2, {0x2010, 0x2090, 0x2090, 0xF910, 0xA97E, 0xAA52, 0xAB92, 0xA892, 0xF912, 0xA112, 0x2252, 0x2BD2, 0x3862, 0xE822, 0x404A, 0x84}},
    {34485, 2, {0x1000, 0x1000, 0x13FE, 0x7C08, 0x5408, 0x55E8, 0x5528, 0x5528, 0x7D28, 0x5128, 0x11E8, 0x1528, 0x1C08, 0xE408, 0x4028, 0x10}},
    {34486, 2, {0x1088, 0x1088, 0x1088, 0x7C88, 0x55FE, 0x5488, 0x5488, 0x5488, 0x7C88, 0x50F8, 0x1088, 0x1488, 0x1E88, 0xE288, 0x40F8, 0x88}},
    {34487, 2, {0x1000, 0x10FE, 0x1080, 0x7C80, 0x5480, 0x54FC, 0x5484, 0x5484, 0x7C84, 0x5084, 0x10FC, 0x1480, 0x1E80, 0xE280, 0x40FE, 0x0}},
    {34488, 2, {0x1008, 0x101C, 0x11E0, 0x7D00, 0x5500, 0x5500, 0x55FE, 0x5510, 0x7D10, 0x5130, 0x1118, 0x1514, 0x1D12, 0xE610, 0x4210, 0x410}},
    {34489, 2, {0x1044, 0x1044, 0x1044, 0x7C84, 0x54BE, 0x5584, 0x5684, 0x54A4, 0x7C94, 0x5094, 0x1084, 0x1484, 0x1E84, 0xE284, 0x4094, 0x88}},
    {34490, 2, {0x1020, 0x1020, 0x1020, 0x7DFC, 0x5524, 0x5524, 0x55FC, 0x5524, 0x7D24, 0x5124, 0x17FE, 0x1504, 0x1D04, 0xE504, 0x4114, 0x108}},
    {34491, 2, {0x420, 0x420, 0x7FA0, 0xE20, 0x1522, 0x24A2, 0x441E, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34492, 2, {0x1080, 0x1080, 0x1080, 0x7DFC, 0x5504, 0x5604, 0x55E4, 0x5524, 0x7D24, 0x5124, 0x1124, 0x15E4, 0x1D24, 0xE404, 0x4028, 0x10}},
    {34493, 2, {0x1000, 0x11FE, 0x1010, 0x7C10, 0x5420, 0x5420, 0x5468, 0x54A4, 0x7D22, 0x5222, 0x1020, 0x1420, 0x1C20, 0xE400, 0x43FE, 0x0}},
    {34494, 2, {0x1010, 0x1010, 0x1010, 0x7DFE, 0x5512, 0x5514, 0x5510, 0x55FC, 0x7D44, 0x5144, 0x1128, 0x1528, 0x1D10, 0xE628, 0x4244, 0x482}},
    {34495, 2, {0x1020, 0x1010, 0x1010, 0x7DFE, 0x5420, 0x5420, 0x5444, 0x5484, 0x7DF8, 0x5010, 0x1020, 0x1444, 0x1C82, 0xE5FE, 0x4082, 0x0}},
    {34496, 2, {0x1040, 0x1020, 0x1000, 0x7DFE, 0x5420, 0x5420, 0x5420, 0x5420, 0x7DFC, 0x5020, 0x1020, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34497, 2, {0x1000, 0x11FC, 0x1044, 0x7C44, 0x5444, 0x5444, 0x5494, 0x5488, 0x7D00, 0x50FC, 0x1084, 0x1484, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {34498, 2, {0x1050, 0x1048, 0x1048, 0x7C40, 0x55FE, 0x5440, 0x5440, 0x547C, 0x7CA4, 0x50A4, 0x10A8, 0x1528, 0x1D10, 0xE628, 0x4044, 0x82}},
    {34499, 2, {0x1000, 0x11FE, 0x1020, 0x7C20, 0x5420, 0x55FE, 0x5522, 0x5522, 0x7D52, 0x514A, 0x118A, 0x1502, 0x1D02, 0xE502, 0x410A, 0x104}},
    {34500, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x55FE, 0x5420, 0x5420, 0x5420, 0x7DFC, 0x5104, 0x1104, 0x1504, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34501, 2, {0x1020, 0x1020, 0x1020, 0x7C20, 0x543E, 0x5420, 0x5420, 0x5420, 0x7DFC, 0x5104, 0x1104, 0x1504, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34502, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x5484, 0x54FC, 0x5484, 0x5484, 0x7C84, 0x50FC, 0x1084, 0x1484, 0x1C84, 0xE484, 0x43FE, 0x0}},
    {34503, 2, {0x1020, 0x1010, 0x1010, 0x7DFE, 0x5502, 0x5604, 0x5480, 0x5488, 0x7C90, 0x50A0, 0x10C0, 0x1482, 0x1E82, 0xE282, 0x407E, 0x0}},
    {34504, 2, {0x1020, 0x1020, 0x1120, 0x7D20, 0x55FC, 0x5520, 0x5620, 0x5420, 0x7DFE, 0x5020, 0x1050, 0x1450, 0x1E88, 0xE288, 0x4104, 0x202}},
    {34505, 2, {0x1020, 0x1020, 0x1050, 0x7C50, 0x5488, 0x5524, 0x5612, 0x5410, 0x7DFC, 0x5004, 0x1008, 0x1488, 0x1E50, 0xE220, 0x4010, 0x10}},
    {34506, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x108, 0x7FFC, 0x4, 0x3FF8, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {34507, 2, {0x7FFC, 0x104, 0x1100, 0x11F8, 0x1100, 0x2900, 0x47FE, 0x8100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34508, 2, {0x1008, 0x101C, 0x11F0, 0x7D50, 0x5550, 0x5550, 0x5550, 0x5550, 0x7D50, 0x5148, 0x1148, 0x1568, 0x1D54, 0xE674, 0x4252, 0x400}},
    {34509, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8104, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x120, 0x110, 0x1F8, 0x7F04, 0x2004}},
    {34510, 2, {0x1000, 0x11FE, 0x1100, 0x7D00, 0x5500, 0x55FE, 0x5520, 0x5520, 0x7D3C, 0x5124, 0x1124, 0x1524, 0x1D24, 0xE644, 0x4254, 0x488}},
    {34511, 2, {0x1000, 0x11FC, 0x1084, 0x7C88, 0x5450, 0x5420, 0x5450, 0x5488, 0x7F26, 0x5020, 0x11FC, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34512, 2, {0x1050, 0x1050, 0x1050, 0x7C50, 0x55FC, 0x5554, 0x5554, 0x5554, 0x7D54, 0x51FC, 0x1154, 0x1554, 0x1D54, 0xE554, 0x41FC, 0x104}},
    {34513, 2, {0x1020, 0x1020, 0x1048, 0x7C84, 0x55FE, 0x5412, 0x5490, 0x5490, 0x7CFE, 0x5110, 0x1010, 0x17FE, 0x1C10, 0xE410, 0x4010, 0x10}},
    {34514, 2, {0x1040, 0x1040, 0x1078, 0x7C88, 0x5550, 0x5420, 0x5450, 0x5488, 0x7D06, 0x50F8, 0x1088, 0x1488, 0x1E88, 0xE288, 0x40F8, 0x88}},
    {34515, 2, {0x820, 0x828, 0x7F24, 0x820, 0xFFFE, 0x20, 0x824, 0x7F24, 0x4924, 0x4928, 0x7F28, 0x4810, 0x912, 0xFAA, 0xF046, 0x4082}},
    {34516, 2, {0x1000, 0x1000, 0x11FE, 0x7D02, 0x5502, 0x557A, 0x554A, 0x554A, 0x7D4A, 0x514A, 0x117A, 0x1502, 0x1D02, 0xE5FE, 0x4102, 0x0}},
    {34517, 2, {0x1020, 0x1020, 0x11FE, 0x7C40, 0x5440, 0x54FC, 0x5484, 0x5584, 0x7EFC, 0x5084, 0x1084, 0x14FC, 0x1E84, 0xE284, 0x4094, 0x88}},
    {34518, 2, {0x1090, 0x1088, 0x1080, 0x7DFE, 0x54A0, 0x54A8, 0x54B0, 0x54A4, 0x7CA8, 0x50B0, 0x1124, 0x152A, 0x1D32, 0xE622, 0x42DE, 0x400}},
    {34519, 2, {0x400, 0x800, 0x3FF0, 0x2010, 0x3FF0, 0x2000, 0x3FF8, 0x2008, 0x3FF8, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34520, 2, {0x1104, 0x1084, 0x1088, 0x7C00, 0x57FE, 0x5420, 0x5420, 0x55FC, 0x7C20, 0x5020, 0x13FE, 0x1420, 0x1E20, 0xE220, 0x4020, 0x20}},
    {34521, 2, {0x1020, 0x1020, 0x11FC, 0x7C20, 0x5420, 0x5420, 0x57FE, 0x5400, 0x7C20, 0x5020, 0x11FC, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34522, 2, {0x2002, 0x2002, 0x23E2, 0xF90A, 0xA90A, 0xA9EA, 0xA92A, 0xAA2A, 0xFB2A, 0xA2AA, 0x244A, 0x284A, 0x3882, 0xE902, 0x420A, 0x404}},
    {34523, 2, {0x1020, 0x1120, 0x1120, 0x7DFC, 0x5520, 0x5620, 0x5420, 0x57FE, 0x7C70, 0x50A8, 0x10A8, 0x1524, 0x1D24, 0xE622, 0x4020, 0x20}},
    {34524, 2, {0x2100, 0x217C, 0x2124, 0xFA24, 0xAA24, 0xAEFE, 0xAA24, 0xAA24, 0xFA24, 0xA27C, 0x2224, 0x2A20, 0x3A20, 0xEA20, 0x4240, 0x280}},
    {34525, 2, {0x1000, 0x11F8, 0x1108, 0x7D08, 0x55F8, 0x5508, 0x5508, 0x55F8, 0x7D44, 0x5148, 0x1130, 0x1520, 0x1D10, 0xE548, 0x4186, 0x100}},
    {34526, 2, {0x1008, 0x103C, 0x11E0, 0x7C20, 0x5420, 0x57FE, 0x5420, 0x5420, 0x7C20, 0x51FC, 0x1104, 0x1504, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34527, 2, {0x1040, 0x1020, 0x1020, 0x7DFE, 0x5400, 0x5488, 0x5504, 0x5602, 0x7C88, 0x5088, 0x1050, 0x1450, 0x1C20, 0xE450, 0x4088, 0x306}},
    {34528, 2, {0x2040, 0x2040, 0x21FC, 0xF844, 0xA884, 0xA884, 0xA928, 0xAA10, 0xF908, 0xA3DE, 0x214A, 0x294A, 0x394A, 0xEA52, 0x435A, 0x4A4}},
    {34529, 2, {0x1000, 0x13DE, 0x1042, 0x7C42, 0x5652, 0x554A, 0x554A, 0x5442, 0x7CC6, 0x514A, 0x1252, 0x1442, 0x1C42, 0xE442, 0x414A, 0x84}},
    {34530, 2, {0x1104, 0x1084, 0x1088, 0x7C00, 0x55FE, 0x5488, 0x5488, 0x5488, 0x7C88, 0x53FE, 0x1088, 0x1488, 0x1D08, 0xE508, 0x4208, 0x408}},
    {34531, 2, {0x1020, 0x1020, 0x1020, 0x7DFE, 0x5420, 0x5420, 0x54FC, 0x5400, 0x7C00, 0x50FC, 0x1084, 0x1484, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {34532, 2, {0x1020, 0x1020, 0x1050, 0x7C50, 0x5488, 0x5504, 0x56FA, 0x5400, 0x7C00, 0x50F8, 0x1088, 0x1488, 0x1E88, 0xE288, 0x40F8, 0x88}},
    {34533, 2, {0x1020, 0x1020, 0x107C, 0x7C84, 0x5548, 0x5430, 0x5420, 0x5448, 0x7D90, 0x503E, 0x1042, 0x15A4, 0x1E18, 0xE210, 0x4060, 0x180}},
    {34534, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5424, 0x55FC, 0x5520, 0x7DFE, 0x5022, 0x102A, 0x1454, 0x1E50, 0xE288, 0x4104, 0x202}},
    {34535, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x5504, 0x55AC, 0x5554, 0x5554, 0x7D54, 0x5154, 0x11AC, 0x1504, 0x1D04, 0xE504, 0x4114, 0x108}},
    {34536, 2, {0x1000, 0x11FE, 0x1020, 0x7C20, 0x5440, 0x54FC, 0x5484, 0x5484, 0x7C84, 0x50FC, 0x1084, 0x1484, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {34537, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4500, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34538, 2, {0x800, 0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0x900, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34539, 2, {0x1040, 0x1040, 0x10FC, 0x7D04, 0x5608, 0x55FE, 0x5500, 0x557C, 0x7D44, 0x5144, 0x1154, 0x1548, 0x1D42, 0xE542, 0x413E, 0x200}},
    {34540, 2, {0x440, 0x440, 0x7FFC, 0x440, 0x440, 0xFFFE, 0x820, 0x1010, 0x2108, 0xDFF6, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34541, 2, {0x1000, 0x11FE, 0x1020, 0x7C40, 0x5488, 0x5504, 0x55FE, 0x5422, 0x7C20, 0x5020, 0x11FE, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34542, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4444, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34543, 2, {0x1040, 0x1040, 0x1044, 0x7DF4, 0x5448, 0x5450, 0x57FE, 0x5440, 0x7C80, 0x5184, 0x1298, 0x14E0, 0x1C82, 0xE482, 0x407E, 0x0}},
    {34544, 2, {0x1080, 0x1080, 0xFBF0, 0x1090, 0x1990, 0xF094, 0x1154, 0x520C, 0x2504, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34545, 2, {0x1020, 0x1020, 0x1020, 0x7DFC, 0x5420, 0x5524, 0x54A4, 0x54A8, 0x7C20, 0x53FE, 0x1050, 0x1450, 0x1E88, 0xE288, 0x4104, 0x202}},
    {34546, 2, {0x2080, 0x2080, 0x20BC, 0xFBC0, 0xA850, 0xA824, 0xA8D4, 0xAB0C, 0xF800, 0xA3FE, 0x2090, 0x2890, 0x3912, 0xE912, 0x420E, 0x400}},
    {34547, 2, {0x2080, 0x20BE, 0x2088, 0xFA88, 0xAABE, 0xAAAA, 0xAAAA, 0xAAAA, 0xFAAA, 0xA2AA, 0x22AA, 0x28AA, 0x392E, 0xE908, 0x4208, 0x408}},
    {34548, 2, {0x1040, 0x1020, 0x13FE, 0x7C88, 0x5450, 0x5420, 0x5450, 0x5488, 0x7D06, 0x5088, 0x1088, 0x1488, 0x1C88, 0xE508, 0x4108, 0x208}},
    {34549, 2, {0x1000, 0x11FE, 0x1000, 0x7C92, 0x5524, 0x5648, 0x5524, 0x5492, 0x7C00, 0x51FE, 0x1020, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34550, 2, {0x1004, 0x101E, 0x11E0, 0x7C22, 0x5512, 0x5494, 0x5480, 0x5408, 0x7DFE, 0x5008, 0x1088, 0x1448, 0x1E48, 0xE208, 0x4028, 0x10}},
    {34551, 2, {0x1028, 0x1024, 0x1024, 0x7C20, 0x55FE, 0x5420, 0x5520, 0x54B2, 0x7CB4, 0x5068, 0x10A8, 0x1524, 0x1E22, 0xE420, 0x40A0, 0x40}},
    {34552, 2, {0x1020, 0x1124, 0x10A4, 0x7CA8, 0x5420, 0x55FC, 0x5504, 0x5504, 0x7DFC, 0x5104, 0x1104, 0x15FC, 0x1D04, 0xE504, 0x4114, 0x108}},
    {34553, 2, {0x1000, 0x11F8, 0x1008, 0x7CD0, 0x5420, 0x55FC, 0x5524, 0x5524, 0x7DFC, 0x5124, 0x1124, 0x15FC, 0x1D24, 0xE524, 0x4124, 0x10C}},
    {34554, 2, {0x1020, 0x1020, 0x1020, 0x7DFC, 0x5420, 0x54A8, 0x54A8, 0x54A8, 0x7D74, 0x5222, 0x1050, 0x1450, 0x1E88, 0xE288, 0x4104, 0x202}},
    {34555, 2, {0x1010, 0x1090, 0x1088, 0x7D08, 0x5604, 0x55FA, 0x5508, 0x5508, 0x7D08, 0x51F8, 0x1090, 0x1490, 0x1C90, 0xE512, 0x4212, 0x40E}},
    {34556, 2, {0x1020, 0x1020, 0x11FE, 0x7C20, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x7D24, 0x5124, 0x11FC, 0x1420, 0x1DFE, 0xE420, 0x4020, 0x20}},
    {34557, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x54FC, 0x5484, 0x5484, 0x54FC, 0x7C84, 0x5084, 0x10FC, 0x1400, 0x1E48, 0xE244, 0x4082, 0x102}},
    {34558, 2, {0x2010, 0x20D8, 0x2394, 0xF894, 0xA890, 0xABFE, 0xA890, 0xA894, 0xF894, 0xA0D8, 0x2398, 0x2890, 0x38AA, 0xE8CA, 0x4286, 0x102}},
    {34559, 2, {0x1020, 0x1020, 0x1050, 0x7C88, 0x5544, 0x5622, 0x54F8, 0x5408, 0x7C10, 0x5020, 0x11FC, 0x1504, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34560, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x1000, 0x1FFC, 0x2204, 0x5FC4, 0x9244, 0x1244, 0x1FC4, 0x224, 0x3F4, 0x7C14, 0x2008}},
    {34561, 2, {0x2100, 0x2100, 0x21FE, 0xFA00, 0xAC00, 0xA9FE, 0xA822, 0xA824, 0xF920, 0xA13C, 0x2120, 0x2920, 0x3920, 0xEAA0, 0x447E, 0x800}},
    {34562, 2, {0x1040, 0x1040, 0x10FC, 0x7C88, 0x5550, 0x5420, 0x54D8, 0x5726, 0x7CF8, 0x5020, 0x10F8, 0x1420, 0x1FFE, 0xE420, 0x4020, 0x20}},
    {34563, 2, {0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x2430, 0x260E, 0x2080, 0x2FF8, 0x2888, 0x4888, 0x4FF8, 0x8084, 0x1FFE, 0x802}},
    {34564, 2, {0x1000, 0x11FE, 0x1100, 0x7D00, 0x557C, 0x5500, 0x5500, 0x55FE, 0x7D50, 0x5152, 0x1154, 0x1548, 0x1D48, 0xE644, 0x4252, 0x460}},
    {34565, 2, {0x1028, 0x1024, 0x13FE, 0x7C20, 0x5420, 0x55FC, 0x5524, 0x5524, 0x7DFC, 0x5124, 0x1124, 0x15FC, 0x1D24, 0xE524, 0x4124, 0x10C}},
    {34566, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x7D04, 0x5104, 0x11FC, 0x1450, 0x1C50, 0xE492, 0x4112, 0x20E}},
    {34567, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34568, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x55FC, 0x5400, 0x5400, 0x55FC, 0x7C20, 0x5020, 0x13FE, 0x1420, 0x1C50, 0xE488, 0x4104, 0x202}},
    {34569, 2, {0x1008, 0x103C, 0x11C0, 0x7C04, 0x5544, 0x54A8, 0x5400, 0x55F8, 0x7C10, 0x5020, 0x13FE, 0x1420, 0x1E20, 0xE220, 0x40A0, 0x40}},
    {34570, 2, {0x2084, 0x21C4, 0x2704, 0xF914, 0xA914, 0xA914, 0xAFD4, 0xA914, 0xFB14, 0xA394, 0x2554, 0x2D04, 0x3904, 0xE904, 0x4114, 0x108}},
    {34571, 2, {0x1040, 0x1020, 0x11FC, 0x7D04, 0x5504, 0x55FC, 0x5504, 0x5504, 0x7DFC, 0x5120, 0x1122, 0x1514, 0x1D08, 0xE544, 0x4182, 0x100}},
    {34572, 2, {0x1110, 0x1112, 0x1114, 0x7DD8, 0x5510, 0x5512, 0x5552, 0x558E, 0x7D20, 0x5020, 0x11FC, 0x1420, 0x1C20, 0xE420, 0x43FE, 0x0}},
    {34573, 2, {0x1020, 0x1020, 0x1050, 0x7C88, 0x5504, 0x5602, 0x54F8, 0x5420, 0x7C20, 0x53FE, 0x1020, 0x1524, 0x1D22, 0xE622, 0x40A0, 0x40}},
    {34574, 2, {0x1000, 0x107C, 0x1044, 0x7C44, 0x547C, 0x5400, 0x54FE, 0x5482, 0x7C82, 0x50FE, 0x1082, 0x1482, 0x1EFE, 0xE282, 0x408A, 0x84}},
    {34575, 2, {0x101C, 0x11E0, 0x1020, 0x7C20, 0x57FE, 0x54A8, 0x5524, 0x5622, 0x7DF8, 0x5088, 0x1090, 0x14BE, 0x1D02, 0xE502, 0x4214, 0x408}},
    {34576, 2, {0x1088, 0x1088, 0x1088, 0x7DC8, 0x549E, 0x548A, 0x548A, 0x57EA, 0x7C8A, 0x508A, 0x110A, 0x154A, 0x1FEA, 0xE512, 0x402A, 0x44}},
    {34577, 2, {0x18, 0x7BE0, 0x1040, 0x2278, 0x7A40, 0xA40, 0x2BFC, 0x1000, 0x2FFE, 0x4100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34578, 2, {0x2004, 0x200E, 0x23B8, 0xF888, 0xA888, 0xA928, 0xA92E, 0xABA8, 0xF8A8, 0xA2A8, 0x22A8, 0x293E, 0x3900, 0xEA80, 0x447E, 0x800}},
    {34579, 2, {0x2000, 0x2006, 0x23B8, 0xF888, 0xA888, 0xA908, 0xA93E, 0xAB88, 0xF888, 0xA288, 0x2288, 0x293E, 0x3900, 0xEA80, 0x447E, 0x800}},
    {34580, 2, {0x1080, 0x1080, 0x10FE, 0x7D02, 0x5502, 0x56FA, 0x54AA, 0x54AA, 0x7CFA, 0x50AA, 0x10AA, 0x14FA, 0x1E8A, 0xE202, 0x4014, 0x8}},
    {34581, 2, {0x1104, 0x1084, 0x1088, 0x7C10, 0x55FC, 0x5504, 0x5504, 0x5504, 0x7DFC, 0x5050, 0x1050, 0x1490, 0x1C92, 0xE512, 0x420E, 0x400}},
    {34582, 2, {0x1000, 0x11FE, 0x114A, 0x7D4A, 0x554A, 0x557A, 0x554A, 0x554A, 0x7D4A, 0x517A, 0x114A, 0x154A, 0x1D4A, 0xE5FE, 0x4102, 0x0}},
    {34583, 2, {0x2000, 0x21FC, 0x2104, 0xF904, 0xA904, 0xA9FC, 0xA820, 0xA820, 0xFBFE, 0xA222, 0x2252, 0x2A8A, 0x3B0A, 0xEA02, 0x420A, 0x204}},
    {34584, 2, {0x2100, 0x2100, 0x2100, 0xFBDE, 0xAA92, 0xAC92, 0xA892, 0xA892, 0xFBF2, 0xA092, 0x2092, 0x2952, 0x395E, 0xEA20, 0x4420, 0x800}},
    {34585, 2, {0x2104, 0x2114, 0x2114, 0xF914, 0xAFD2, 0xA922, 0xA92A, 0xAB48, 0xFB88, 0xA550, 0x2510, 0x2914, 0x3922, 0xE97E, 0x4122, 0x100}},
    {34586, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34587, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x55FC, 0x5510, 0x5510, 0x55FE, 0x7D10, 0x5110, 0x117C, 0x1544, 0x1D44, 0xE544, 0x417C, 0x244}},
    {34588, 2, {0x200, 0x100, 0x7FFE, 0x4202, 0x8924, 0x2848, 0x4994, 0xE10, 0x77F0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34589, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1010, 0x2108, 0xDFF6, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x4}},
    {34590, 2, {0x1088, 0x1088, 0x11FC, 0x7C88, 0x5488, 0x54F8, 0x5488, 0x5488, 0x7CF8, 0x5088, 0x1088, 0x15FE, 0x1C00, 0xE488, 0x4104, 0x202}},
    {34591, 2, {0x1020, 0x1010, 0x11FE, 0x7C40, 0x5484, 0x55FE, 0x5402, 0x54FC, 0x7C84, 0x50FC, 0x1084, 0x14FC, 0x1E84, 0xE284, 0x4094, 0x88}},
    {34592, 2, {0x2000, 0x23FE, 0x2202, 0xFA1A, 0xAAE2, 0xAA22, 0xAA22, 0xABFE, 0xFA22, 0xA272, 0x22AA, 0x2B26, 0x3A22, 0xEA22, 0x43FE, 0x202}},
    {34593, 2, {0x1048, 0x1048, 0x1048, 0x7DFE, 0x5448, 0x5448, 0x55FE, 0x5400, 0x7CFC, 0x5084, 0x1084, 0x14FC, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {34594, 2, {0x1000, 0x11FC, 0x1008, 0x7C10, 0x5420, 0x57FE, 0x5420, 0x54A0, 0x7C40, 0x51FC, 0x1154, 0x1554, 0x1D54, 0xE554, 0x43FE, 0x0}},
    {34595, 2, {0x1088, 0x1048, 0x1050, 0x7DFC, 0x5420, 0x5420, 0x55FC, 0x5420, 0x7C20, 0x53FE, 0x1050, 0x1450, 0x1C90, 0xE492, 0x4112, 0x20E}},
    {34596, 2, {0x80C, 0x8F0, 0x7E80, 0x1880, 0x2CFE, 0x2A88, 0x4888, 0x8888, 0x908, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34597, 2, {0x2100, 0x2106, 0x2138, 0xF920, 0xAFA0, 0xA920, 0xA93E, 0xAB24, 0xFBA4, 0xA564, 0x2524, 0x2924, 0x3924, 0xE924, 0x4124, 0x144}},
    {34598, 2, {0x1020, 0x1020, 0x1050, 0x7C88, 0x5504, 0x56FA, 0x5400, 0x55FC, 0x7D54, 0x5154, 0x11FC, 0x1554, 0x1D54, 0xE554, 0x4104, 0x10C}},
    {34599, 2, {0x2040, 0x2020, 0x23FC, 0xFA04, 0xAA04, 0xABFC, 0xAA00, 0xAA28, 0xFA24, 0xA3FE, 0x2220, 0x2A50, 0x3A50, 0xEC88, 0x4504, 0xA02}},
    {34600, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5424, 0x57FE, 0x5424, 0x7DFC, 0x5020, 0x1120, 0x153E, 0x1D20, 0xE6A0, 0x427E, 0x400}},
    {34601, 2, {0x1000, 0x11FC, 0x1124, 0x7D24, 0x5574, 0x5524, 0x55FC, 0x5504, 0x7D74, 0x5154, 0x1154, 0x1574, 0x1D04, 0xE604, 0x4214, 0x408}},
    {34602, 2, {0x1080, 0x1080, 0x10FE, 0x7D02, 0x5682, 0x54F2, 0x5542, 0x5442, 0x7FFA, 0x5042, 0x1152, 0x1552, 0x1DF2, 0xE402, 0x4014, 0x8}},
    {34603, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x7C00, 0x5112, 0x11D4, 0x1518, 0x1D10, 0xE552, 0x4192, 0x10E}},
    {34604, 2, {0x2000, 0x21FC, 0x2008, 0xF810, 0xAA22, 0xAAAA, 0xAA72, 0xAA22, 0xFA72, 0xA2AA, 0x2326, 0x2AA2, 0x3A42, 0xEA02, 0x43FE, 0x2}},
    {34605, 2, {0x1040, 0x1040, 0x107C, 0x7C84, 0x5508, 0x5600, 0x5420, 0x55CE, 0x7D02, 0x5102, 0x11CE, 0x1502, 0x1D02, 0xE502, 0x41FE, 0x102}},
    {34606, 2, {0x2014, 0x2012, 0x2010, 0xFBFE, 0xA810, 0xA810, 0xABD2, 0xAA52, 0xFA52, 0xA254, 0x23D4, 0x280C, 0x386A, 0xEB8A, 0x4116, 0x22}},
    {34607, 2, {0x1020, 0x1020, 0x11FE, 0x7C20, 0x54FC, 0x5440, 0x55FE, 0x5488, 0x7D24, 0x5222, 0x10F8, 0x1420, 0x1DFE, 0xE420, 0x4020, 0x20}},
    {34608, 2, {0x0, 0x3EFC, 0x22A4, 0x3EA4, 0x22FC, 0x3E80, 0x2282, 0x2282, 0x4A7E, 0x8500, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34609, 2, {0x1040, 0x1080, 0x11FC, 0x7D24, 0x5524, 0x55FC, 0x5524, 0x5544, 0x7DFC, 0x5090, 0x1110, 0x17FE, 0x1C10, 0xE410, 0x4010, 0x10}},
    {34610, 2, {0x101C, 0x11E0, 0x1020, 0x7C20, 0x57FE, 0x54A8, 0x5524, 0x5642, 0x7C40, 0x53FE, 0x1088, 0x1508, 0x1CD0, 0xE430, 0x4048, 0x184}},
    {34611, 2, {0x1020, 0x1010, 0x11FE, 0x7C00, 0x54FC, 0x5484, 0x54FC, 0x5400, 0x7CFC, 0x5008, 0x1010, 0x15FE, 0x1E10, 0xE210, 0x4050, 0x20}},
    {34612, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x54FC, 0x5484, 0x5484, 0x54FC, 0x7C40, 0x50FE, 0x112A, 0x144A, 0x1C92, 0xE522, 0x404A, 0x84}},
    {34613, 2, {0x1050, 0x1154, 0x1154, 0x7D54, 0x55DC, 0x5504, 0x5504, 0x55FC, 0x7D04, 0x5104, 0x11DC, 0x1554, 0x1D54, 0xE554, 0x4254, 0x404}},
    {34614, 2, {0x1040, 0x1020, 0x11FE, 0x7C00, 0x5488, 0x5488, 0x5554, 0x5622, 0x7C00, 0x5020, 0x13FE, 0x1420, 0x1E20, 0xE220, 0x4020, 0x20}},
    {34615, 2, {0x1020, 0x1124, 0x10A8, 0x7C20, 0x55FC, 0x5440, 0x57FE, 0x5488, 0x7D04, 0x52FA, 0x1488, 0x1488, 0x1CA8, 0xE492, 0x4082, 0x7E}},
    {34616, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC6, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34617, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x54A8, 0x5420, 0x55FC, 0x5524, 0x7D24, 0x5154, 0x114C, 0x1584, 0x1D04, 0xE504, 0x4114, 0x108}},
    {34618, 2, {0x2000, 0x2040, 0x239C, 0xFA04, 0xAA04, 0xAB9C, 0xAA04, 0xAA04, 0xFBFC, 0xA090, 0x2090, 0x2890, 0x3890, 0xE912, 0x4212, 0x40E}},
    {34619, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5420, 0x57FE, 0x5400, 0x7DFC, 0x5104, 0x11FC, 0x1504, 0x1DFC, 0xE504, 0x4114, 0x108}},
    {34620, 2, {0x10A0, 0x1090, 0x1080, 0x7DFE, 0x5510, 0x5710, 0x55FC, 0x5510, 0x7D10, 0x51FC, 0x1110, 0x1510, 0x1D10, 0xE5FE, 0x4100, 0x100}},
    {34621, 2, {0x2000, 0x23FE, 0x2020, 0xF820, 0xABFE, 0xAA22, 0xAA22, 0xAB32, 0xFAAA, 0xA2AA, 0x2376, 0x2A66, 0x3A22, 0xEA22, 0x422A, 0x204}},
    {34622, 2, {0x1000, 0x11FC, 0x1124, 0x7D24, 0x55FC, 0x5524, 0x5524, 0x55FC, 0x7C20, 0x53FE, 0x1070, 0x14A8, 0x1D24, 0xE622, 0x4020, 0x20}},
    {34623, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0xAD04, 0xA900, 0xA9DE, 0xAA52, 0xFA52, 0xA352, 0x249A, 0x2894, 0x3910, 0xE912, 0x4212, 0x40E}},
    {34624, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x7DFC, 0x5020, 0x1070, 0x14A8, 0x1D24, 0xE622, 0x4020, 0x20}},
    {34625, 2, {0x0, 0x7FFC, 0x440, 0x3C78, 0x2008, 0x3C78, 0x440, 0xFFFE, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34626, 2, {0x2080, 0x2286, 0x22B8, 0xFAA0, 0xABE0, 0xAA3E, 0xAA2A, 0xAA2A, 0xFBAA, 0xA2AA, 0x22A4, 0x2AA4, 0x3AA4, 0xECCA, 0x448A, 0x892}},
    {34627, 2, {0x1000, 0x13DE, 0x1042, 0x7D4A, 0x5484, 0x554A, 0x5652, 0x5420, 0x7C00, 0x53DE, 0x1052, 0x1552, 0x1C94, 0xE548, 0x4254, 0x422}},
    {34628, 2, {0x2000, 0x23FE, 0x2202, 0xFA8A, 0xAA52, 0xABFE, 0xAA42, 0xAA22, 0xFBFE, 0xA282, 0x2282, 0x2A82, 0x3AFA, 0xEA02, 0x420A, 0x204}},
    {34629, 2, {0x3EF8, 0x820, 0x820, 0x7EFC, 0x820, 0x1450, 0x2288, 0x4104, 0x9FF2, 0x1110, 0x1110, 0x1FF0, 0x1100, 0x108, 0x7FFC, 0x2004}},
    {34630, 2, {0x10A8, 0x10A8, 0x13AE, 0x7CA8, 0x54A8, 0x5420, 0x55FC, 0x5420, 0x7C20, 0x5020, 0x13FE, 0x1420, 0x1E20, 0xE220, 0x4020, 0x20}},
    {34631, 2, {0x2000, 0x21F0, 0x2110, 0xF910, 0xA9F0, 0xA840, 0xABF8, 0xAA48, 0xFA48, 0xA3F8, 0x2248, 0x2A48, 0x3BFA, 0xE842, 0x4042, 0x3E}},
    {34632, 2, {0x2000, 0x23FE, 0x2202, 0xFA02, 0xABFE, 0xAA22, 0xAA22, 0xAAFA, 0xFA22, 0xA232, 0x222A, 0x2BFE, 0x3A02, 0xEA02, 0x43FE, 0x202}},
    {34633, 2, {0x1104, 0x1088, 0x1050, 0x7DFC, 0x5524, 0x5524, 0x55FC, 0x5524, 0x7D24, 0x51FC, 0x1020, 0x1420, 0x1DFE, 0xE420, 0x4020, 0x20}},
    {34634, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0xAC04, 0xA800, 0xABFE, 0xA820, 0xF820, 0xA120, 0x213C, 0x2920, 0x3AA0, 0xEA60, 0x443E, 0x800}},
    {34635, 2, {0x2000, 0x2484, 0x2244, 0xFA48, 0xA800, 0xABF8, 0xAA48, 0xAA48, 0xFBF8, 0xA248, 0x2248, 0x2BF8, 0x3A4A, 0xEC4A, 0x4446, 0x802}},
    {34636, 2, {0x20C4, 0x2704, 0x2124, 0xF914, 0xA914, 0xAF84, 0xA924, 0xA914, 0xFB94, 0xA346, 0x253C, 0x2D04, 0x3904, 0xE904, 0x4104, 0x104}},
    {34637, 2, {0x2000, 0x23DE, 0x2252, 0xFA52, 0xAA52, 0xABD2, 0xAA52, 0xAA52, 0xFA52, 0xA3D2, 0x221A, 0x2A94, 0x3A50, 0xEAB0, 0x4310, 0x10}},
    {34638, 2, {0x1000, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5480, 0x55FE, 0x7E22, 0x5122, 0x1152, 0x1502, 0x1DFA, 0xE402, 0x4014, 0x8}},
    {34639, 2, {0x2040, 0x2020, 0x23FE, 0xF800, 0xA9FC, 0xA904, 0xA9FC, 0xA800, 0xFBFE, 0xA202, 0x21FC, 0x2820, 0x3820, 0xE820, 0x40A0, 0x40}},
    {34640, 2, {0x1000, 0x11FE, 0x1102, 0x7D7A, 0x5502, 0x557A, 0x5502, 0x5400, 0x7CFC, 0x5084, 0x10FC, 0x1484, 0x1EFC, 0xE284, 0x40FC, 0x84}},
    {34641, 2, {0x1000, 0x13FE, 0x1022, 0x7D20, 0x553C, 0x5520, 0x56FE, 0x5400, 0x7DFC, 0x5104, 0x11FC, 0x1504, 0x1DFC, 0xE504, 0x4114, 0x108}},
    {34642, 2, {0x1000, 0x13FE, 0x1020, 0x7C40, 0x55FC, 0x5554, 0x5554, 0x5574, 0x7D54, 0x5154, 0x1174, 0x1554, 0x1D54, 0xE554, 0x41FC, 0x104}},
    {34643, 2, {0x2040, 0x20A0, 0x2110, 0xFA08, 0xADF6, 0xA800, 0xABC4, 0xAA54, 0xFA54, 0xA3D4, 0x2254, 0x2A54, 0x3BD4, 0xEA44, 0x4254, 0x2C8}},
    {34644, 2, {0x1110, 0x1112, 0x11D4, 0x7D18, 0x5552, 0x5592, 0x552E, 0x5440, 0x7DFC, 0x5104, 0x1104, 0x15FC, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34645, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x5324, 0xFD24, 0x4524, 0x7D24, 0x45FC, 0x7D24, 0x4020, 0x5028, 0x4824, 0x55FE, 0x6482, 0x4000}},
    {34646, 2, {0x1040, 0x1020, 0x13FE, 0x7E02, 0x5404, 0x55FC, 0x5400, 0x55FC, 0x7D04, 0x51FC, 0x1104, 0x15FC, 0x1D04, 0xE400, 0x43FE, 0x0}},
    {34647, 2, {0x2020, 0x2040, 0x21FC, 0xF904, 0xA9FC, 0xA904, 0xA9FC, 0xA800, 0xFBFE, 0xA020, 0x2020, 0x29FC, 0x3820, 0xE820, 0x43FE, 0x0}},
    {34648, 2, {0x1000, 0x11FE, 0x1100, 0x7D7C, 0x5544, 0x557C, 0x5544, 0x557C, 0x7D10, 0x51FE, 0x1124, 0x1564, 0x1D18, 0xE524, 0x4142, 0x1FE}},
    {34649, 2, {0x1020, 0x1010, 0x11FE, 0x7D02, 0x5502, 0x55FE, 0x5500, 0x5500, 0x7DFE, 0x51AA, 0x12AA, 0x14FE, 0x1CAA, 0xE4AA, 0x40AA, 0x86}},
    {34650, 2, {0x1000, 0x10FC, 0x1048, 0x7C30, 0x55FE, 0x5452, 0x5494, 0x5510, 0x7E30, 0x5020, 0x11FE, 0x1470, 0x1CA8, 0xE524, 0x4222, 0x20}},
    {34651, 2, {0x200C, 0x200A, 0x2008, 0xFBFE, 0xAA08, 0xABF8, 0xAA4A, 0xAA4A, 0xFBEA, 0xA2AC, 0x22AC, 0x2A4C, 0x3AAA, 0xEB0A, 0x4216, 0x422}},
    {34652, 2, {0x1040, 0x1040, 0x10F8, 0x7D08, 0x5610, 0x55FC, 0x5504, 0x55FC, 0x7D04, 0x51FC, 0x1104, 0x15FC, 0x1C00, 0xE488, 0x4104, 0x202}},
    {34653, 2, {0x1040, 0x107C, 0x1084, 0x7CF8, 0x5408, 0x55FE, 0x5440, 0x54A4, 0x7D38, 0x5058, 0x1094, 0x1534, 0x1C54, 0xE492, 0x4150, 0x20}},
    {34654, 2, {0x1000, 0x11FC, 0x1124, 0x7D24, 0x55FC, 0x5500, 0x557C, 0x5544, 0x7D44, 0x517C, 0x1144, 0x157C, 0x1D44, 0xE644, 0x427C, 0x444}},
    {34655, 2, {0x1000, 0x11FE, 0x1122, 0x7DFE, 0x5522, 0x55FE, 0x5400, 0x54FC, 0x7C84, 0x50FC, 0x1084, 0x14FC, 0x1E84, 0xE284, 0x4094, 0x88}},
    {34656, 2, {0x1000, 0x11FE, 0x1000, 0x7CFC, 0x5484, 0x5484, 0x54FC, 0x5400, 0x7DFE, 0x5122, 0x1122, 0x15FE, 0x1D22, 0xE522, 0x41FE, 0x102}},
    {34657, 2, {0x1000, 0x13FE, 0x1020, 0x7C40, 0x55FC, 0x5554, 0x5554, 0x5554, 0x7D54, 0x512C, 0x1020, 0x15FE, 0x1C50, 0xE488, 0x4104, 0x202}},
    {34658, 2, {0x1000, 0x13FE, 0x1020, 0x7C40, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x7D04, 0x51FC, 0x1104, 0x1504, 0x1DFC, 0xE400, 0x4088, 0x104}},
    {34659, 2, {0x2108, 0x2088, 0x2088, 0xFBDE, 0xA910, 0xA920, 0xA9DE, 0xA942, 0xF944, 0xA144, 0x215E, 0x2944, 0x3A44, 0xEA44, 0x44D4, 0x808}},
    {34660, 2, {0x1088, 0x1048, 0x1050, 0x7DFE, 0x5450, 0x5450, 0x55FC, 0x5554, 0x7D54, 0x518C, 0x1104, 0x15FC, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34661, 2, {0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x98C6, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34662, 2, {0x2000, 0x23DE, 0x2242, 0xFA42, 0xAA42, 0xABDE, 0xAA00, 0xAA3E, 0xFBD2, 0xA212, 0x2214, 0x2BD4, 0x3A08, 0xEA14, 0x4224, 0x242}},
    {34663, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x54A8, 0x5420, 0x55FC, 0x5524, 0x7D24, 0x5124, 0x13FE, 0x1420, 0x1E50, 0xE288, 0x4104, 0x202}},
    {34664, 2, {0x0, 0x3FF0, 0x210, 0x210, 0x7FEA, 0x206, 0x202, 0x820, 0x7EFC, 0x4AA4, 0x4AA4, 0x7EFC, 0x820, 0xA24, 0xFEFE, 0x202}},
    {34665, 2, {0x1008, 0x103C, 0x11E0, 0x7C20, 0x57FE, 0x5420, 0x55FC, 0x5524, 0x7DFC, 0x5124, 0x11FC, 0x1420, 0x1DFC, 0xE420, 0x43FE, 0x0}},
    {34666, 2, {0x1000, 0x10FC, 0x1084, 0x7CFC, 0x5484, 0x54FC, 0x5400, 0x55FE, 0x7C80, 0x50FE, 0x112A, 0x164A, 0x1C92, 0xE522, 0x4054, 0x88}},
    {34667, 2, {0x1020, 0x1022, 0x11FA, 0x7C24, 0x5428, 0x57FE, 0x5420, 0x5440, 0x7CFC, 0x5184, 0x1284, 0x14FC, 0x1E84, 0xE284, 0x40FC, 0x84}},
    {34668, 2, {0x2000, 0x2088, 0x2252, 0xFA22, 0xAA52, 0xAA8A, 0xAA02, 0xABFE, 0xF888, 0xA144, 0x227A, 0x2888, 0x3950, 0xE820, 0x40D8, 0x706}},
    {34669, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x7C00, 0x53FE, 0x1020, 0x1520, 0x1D3C, 0xE520, 0x42A0, 0x47E}},
    {34670, 2, {0x1040, 0x1040, 0x10FE, 0x7C80, 0x55FC, 0x5684, 0x54FC, 0x5484, 0x7CFC, 0x5040, 0x107C, 0x14C4, 0x1D28, 0xE410, 0x4068, 0x186}},
    {34671, 2, {0x101E, 0x13E0, 0x1044, 0x7D24, 0x5488, 0x55FC, 0x5440, 0x5440, 0x7DFE, 0x5080, 0x10FC, 0x1544, 0x1D28, 0xE610, 0x4068, 0x186}},
    {34672, 2, {0x1020, 0x1020, 0x13FE, 0x7C50, 0x5488, 0x5726, 0x55FC, 0x5420, 0x7C20, 0x53FE, 0x1000, 0x1420, 0x1DFC, 0xE420, 0x4020, 0x3FE}},
    {34673, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1FF8, 0x0, 0x820, 0x7EFC, 0x4AA4, 0x4AA4, 0x7EFC, 0x820, 0xA24, 0xFEFE, 0x202}},
    {34674, 2, {0x2082, 0x2082, 0x27F2, 0xF882, 0xA88A, 0xABEA, 0xAAAA, 0xAAAA, 0xFBEA, 0xA08A, 0x21CA, 0x2AAA, 0x3CA2, 0xE882, 0x408A, 0x84}},
    {34675, 2, {0x1020, 0x11FE, 0x1020, 0x7CFC, 0x5420, 0x55FE, 0x5400, 0x54FC, 0x7CA4, 0x5094, 0x11FE, 0x14A4, 0x1D14, 0xE5FE, 0x4004, 0x18}},
    {34676, 2, {0x2100, 0x213C, 0x2124, 0xFBA4, 0xA924, 0xA93C, 0xA924, 0xABA4, 0xFAA4, 0xA2BC, 0x22A4, 0x2AA4, 0x3BA4, 0xE844, 0x4054, 0x88}},
    {34677, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x904, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34678, 2, {0x2048, 0x2148, 0x2148, 0xFBFE, 0xA948, 0xA948, 0xA978, 0xA900, 0xF9FE, 0xA020, 0x23FE, 0x2870, 0x38A8, 0xE924, 0x4622, 0x20}},
    {34679, 2, {0x1020, 0x1010, 0x11FE, 0x7D00, 0x5544, 0x5528, 0x55FE, 0x5510, 0x7D54, 0x5154, 0x1154, 0x157C, 0x1D14, 0xE520, 0x4240, 0x80}},
    {34680, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x54F4, 0x5494, 0x5494, 0x55FE, 0x7D02, 0x517A, 0x114A, 0x154A, 0x1D7A, 0xE502, 0x410A, 0x104}},
    {34681, 2, {0x1000, 0x10F8, 0x1088, 0x7C88, 0x54F8, 0x5488, 0x5488, 0x54F8, 0x7C00, 0x51FC, 0x1154, 0x1554, 0x1D54, 0xE554, 0x43FE, 0x0}},
    {34682, 2, {0x2004, 0x201E, 0x23E0, 0xF820, 0xA9FC, 0xA924, 0xA924, 0xA9FC, 0xF820, 0xA3FE, 0x2222, 0x2A2A, 0x3AFA, 0xEA02, 0x420A, 0x204}},
    {34683, 2, {0x2020, 0x2020, 0x23FE, 0xF820, 0xA820, 0xABFE, 0xAA8A, 0xAA52, 0xFAFA, 0xA222, 0x2222, 0x2AFA, 0x3A22, 0xEA22, 0x422A, 0x204}},
    {34684, 2, {0x2020, 0x2124, 0x20A8, 0xF820, 0xABFE, 0xA8A8, 0xA924, 0xAA02, 0xF840, 0xA7FE, 0x2088, 0x2908, 0x3990, 0xE860, 0x4198, 0x604}},
    {34685, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5440, 0x57FE, 0x5488, 0x7D04, 0x52FA, 0x1088, 0x1488, 0x1CF8, 0xE488, 0x4088, 0xF8}},
    {34686, 2, {0x2088, 0x2088, 0x23FE, 0xF888, 0xA800, 0xABFE, 0xAA02, 0xAC24, 0xF820, 0xA3FE, 0x2070, 0x28A8, 0x38A8, 0xE924, 0x4222, 0x20}},
    {34687, 2, {0x1000, 0x11FC, 0x1124, 0x7DFC, 0x5524, 0x55FC, 0x5420, 0x55FC, 0x7D24, 0x51FC, 0x1124, 0x15FC, 0x1C20, 0xE422, 0x4022, 0x1E}},
    {34688, 2, {0x880, 0x49F8, 0x2A90, 0x2860, 0x890, 0x1FFE, 0x2A10, 0xC950, 0x920, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34689, 2, {0x1000, 0x127C, 0x1144, 0x7D44, 0x547C, 0x5444, 0x5744, 0x557C, 0x7D50, 0x514A, 0x1144, 0x1554, 0x1D62, 0xE542, 0x4280, 0x47E}},
    {34690, 2, {0x2100, 0x209E, 0x23D2, 0xFA52, 0xAA54, 0xABD4, 0xAA58, 0xAA54, 0xFBD2, 0xA212, 0x2292, 0x2A5A, 0x3AD4, 0xEB50, 0x4210, 0x10}},
    {34691, 2, {0x2040, 0x2020, 0x23FC, 0xF908, 0xA890, 0xABFE, 0xAA02, 0xAC44, 0xF820, 0xA3FC, 0x2080, 0x28F8, 0x3888, 0xE908, 0x4128, 0x210}},
    {34692, 2, {0x2100, 0x227E, 0x2388, 0xFA88, 0xAABE, 0xAAAA, 0xABAA, 0xAA2A, 0xFA2A, 0xA3AA, 0x22AA, 0x2AAE, 0x3A88, 0xEB88, 0x4288, 0x8}},
    {34693, 2, {0x2020, 0x2040, 0x21FC, 0xF904, 0xA9FC, 0xA904, 0xA9FC, 0xA904, 0xF9FC, 0xA040, 0x2020, 0x28A4, 0x3A8A, 0xEA8A, 0x4478, 0x0}},
    {34694, 2, {0x2084, 0x2044, 0x2048, 0xFBFE, 0xA884, 0xA884, 0xA908, 0xA94A, 0xFA52, 0xA39C, 0x2084, 0x2908, 0x3908, 0xEA52, 0x43DE, 0x42}},
    {34695, 2, {0x101E, 0x13E0, 0x1122, 0x7C94, 0x5440, 0x5488, 0x55F0, 0x5420, 0x7CC4, 0x51FE, 0x1022, 0x1420, 0x1FFE, 0xE450, 0x4088, 0x306}},
    {34696, 2, {0x1000, 0x11FE, 0x1110, 0x7D20, 0x557C, 0x5544, 0x557C, 0x5544, 0x7D7C, 0x5110, 0x1110, 0x1554, 0x1F52, 0xE292, 0x4250, 0x420}},
    {34697, 2, {0x2090, 0x2088, 0x2104, 0xFA42, 0xAC88, 0xA9FC, 0xA804, 0xA800, 0xFBDE, 0xA042, 0x2252, 0x294A, 0x3A52, 0xE842, 0x414A, 0x84}},
    {34698, 2, {0x1088, 0x1050, 0x1000, 0x7DFE, 0x5450, 0x55FC, 0x5454, 0x57FE, 0x7C54, 0x51FC, 0x1050, 0x14D8, 0x1D54, 0xE652, 0x4050, 0x50}},
    {34699, 2, {0x10A0, 0x112C, 0x1124, 0x7D24, 0x55AC, 0x5524, 0x5524, 0x55FC, 0x7C20, 0x51FC, 0x1088, 0x1450, 0x1E20, 0xE250, 0x4088, 0x306}},
    {34700, 2, {0x1078, 0x3E48, 0x2248, 0x2A86, 0xFF78, 0x2248, 0x2A28, 0x4210, 0x4A28, 0x8546, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34701, 2, {0x20, 0xFE20, 0x20, 0x7CFC, 0x44A4, 0x7CA4, 0xA4, 0xFEA4, 0x82FC, 0xAAA4, 0x9220, 0xFE28, 0x9224, 0x92FE, 0x9242, 0x8600}},
    {34702, 2, {0x2000, 0x23FE, 0x2000, 0xF9FC, 0xA904, 0xA904, 0xA9FC, 0xA800, 0xFBFE, 0xA28A, 0x2252, 0x2BFE, 0x3A22, 0xEA22, 0x422A, 0x204}},
    {34703, 2, {0x2020, 0x2010, 0x21FE, 0xF900, 0xA920, 0xAD20, 0xAB3C, 0xA950, 0xF990, 0xA310, 0x25FE, 0x2910, 0x3928, 0xEA28, 0x4244, 0x482}},
    {34704, 2, {0x2040, 0x2080, 0x21FC, 0xF904, 0xA904, 0xA9FC, 0xA900, 0xA9FE, 0xF900, 0xA1FE, 0x2002, 0x2AAA, 0x3AAA, 0xEC02, 0x4014, 0x8}},
    {34705, 2, {0x1020, 0x1040, 0x11FC, 0x7D04, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x7DFC, 0x5028, 0x1024, 0x15FE, 0x1C50, 0xE488, 0x4104, 0x202}},
    {34706, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1126, 0x7C20, 0x45FC, 0x7D24, 0x4524, 0x7D24, 0x11FC, 0xFE20, 0x1028, 0x103C, 0x13C4, 0x1104}},
    {34707, 2, {0x2040, 0x2040, 0x27FC, 0xF840, 0xABF8, 0xA880, 0xAFFC, 0xA910, 0xFA68, 0xA5C6, 0x2040, 0x2BF8, 0x38E0, 0xE958, 0x4644, 0x40}},
    {34708, 2, {0x2004, 0x201E, 0x23F0, 0xFA1E, 0xAA10, 0xAAFE, 0xAA92, 0xAA98, 0xFAF2, 0xA28E, 0x2280, 0x2AB8, 0x3AA8, 0xECAA, 0x454A, 0xA86}},
    {34709, 2, {0x1020, 0x1040, 0x11FC, 0x7D04, 0x5554, 0x5524, 0x5554, 0x5504, 0x7DFC, 0x5000, 0x1112, 0x15D4, 0x1D18, 0xE552, 0x4192, 0x10E}},
    {34710, 2, {0x1000, 0x10FC, 0x1084, 0x7CF4, 0x5494, 0x55FE, 0x5502, 0x54FC, 0x7C84, 0x50FC, 0x1084, 0x14FC, 0x1E84, 0xE284, 0x4094, 0x88}},
    {34711, 2, {0x1020, 0x1010, 0x11FE, 0x7D10, 0x557C, 0x5514, 0x55FE, 0x5514, 0x7D7C, 0x5110, 0x117C, 0x1544, 0x1D44, 0xE644, 0x427C, 0x444}},
    {34712, 2, {0x1020, 0x1124, 0x1124, 0x7D24, 0x55FC, 0x5400, 0x57FE, 0x5400, 0x7DFC, 0x5104, 0x1104, 0x15FC, 0x1C88, 0xE450, 0x43FE, 0x0}},
    {34713, 2, {0x100, 0x100, 0x7FFC, 0x380, 0x540, 0x1930, 0xE10E, 0x820, 0x7EFC, 0x4AA4, 0x4AA4, 0x7EFC, 0x820, 0xA24, 0xFEFE, 0x202}},
    {34714, 2, {0x2044, 0x4258, 0xFF62, 0x142, 0x7E3E, 0x4200, 0x7E4C, 0x4272, 0x7E42, 0x423E, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34715, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0xAC24, 0xA9FC, 0xA820, 0xA9FC, 0xF820, 0xA3FE, 0x2020, 0x29FC, 0x3904, 0xE904, 0x41FC, 0x104}},
    {34716, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x7F78, 0x4148, 0x30, 0x3ECE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34717, 2, {0x2020, 0x2040, 0x23FC, 0xFA24, 0xAA24, 0xABFC, 0xAA24, 0xAA44, 0xFBFC, 0xA040, 0x20A8, 0x28B4, 0x393C, 0xE922, 0x4222, 0x41E}},
    {34718, 2, {0x2000, 0x21FE, 0x2120, 0xF9FC, 0xA920, 0xA9FC, 0xA920, 0xA920, 0xF9FE, 0xA002, 0x22AA, 0x2AAA, 0x3AAA, 0xEA02, 0x4014, 0x8}},
    {34719, 2, {0x2000, 0x23FE, 0x2202, 0xF800, 0xA9FC, 0xA904, 0xA9FC, 0xA904, 0xF9FC, 0xA040, 0x2020, 0x2BFE, 0x3800, 0xE888, 0x4104, 0x202}},
    {34720, 2, {0x2104, 0x2084, 0x2088, 0xF800, 0xABFE, 0xA800, 0xA888, 0xA904, 0xFA02, 0xA1FC, 0x2154, 0x2954, 0x3954, 0xE954, 0x47FE, 0x0}},
    {34721, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x7C0, 0x1830, 0xE00E, 0x820, 0x7EFC, 0x4AA4, 0x4AA4, 0x7EFC, 0x820, 0xA24, 0xFEFE, 0x202}},
    {34722, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4102, 0x8104, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0x108, 0x7FFC, 0x4}},
    {34723, 2, {0x20, 0x3D24, 0x24A8, 0x2420, 0x25FC, 0x3C40, 0x27FE, 0x24A8, 0x2524, 0x3EFA, 0x24A8, 0x24A8, 0x24F8, 0x4424, 0x55FC, 0x8804}},
    {34724, 2, {0x2020, 0x1020, 0xFE20, 0x5FC, 0x6924, 0x1124, 0x29FC, 0xD524, 0x1124, 0x7DFC, 0x5524, 0x7C50, 0x1050, 0x1C88, 0xE504, 0x4202}},
    {34725, 2, {0x1020, 0x1020, 0x1050, 0x7C48, 0x54A4, 0x55FE, 0x5684, 0x54FC, 0x7C84, 0x50FC, 0x1080, 0x14FC, 0x1D44, 0xE544, 0x427C, 0x44}},
    {34726, 2, {0x1020, 0x1020, 0x13FE, 0x7C20, 0x55FC, 0x5420, 0x57FE, 0x5440, 0x7C88, 0x51F0, 0x1024, 0x17FE, 0x1C22, 0xE524, 0x42A2, 0x40}},
    {34727, 2, {0x1040, 0x1044, 0x11F8, 0x7C50, 0x57FE, 0x5440, 0x54F8, 0x5582, 0x7E7E, 0x5400, 0x11FC, 0x1504, 0x1DFC, 0xE504, 0x41FC, 0x104}},
    {34728, 2, {0x1048, 0x1048, 0x11FE, 0x7C48, 0x5400, 0x55FE, 0x5448, 0x5448, 0x7DFE, 0x514A, 0x114A, 0x15B6, 0x1D22, 0xE502, 0x410A, 0x104}},
    {34729, 2, {0x2110, 0x211E, 0x2122, 0xFA54, 0xAA08, 0xAE94, 0xAAE2, 0xAA88, 0xFA88, 0xA2BE, 0x2288, 0x2AAC, 0x3A2A, 0xEA4A, 0x4228, 0x210}},
    {34730, 2, {0x2040, 0x2020, 0x23FE, 0xF888, 0xA850, 0xABFE, 0xAA52, 0xAA8A, 0xFB06, 0xA2FA, 0x228A, 0x2A8A, 0x3AFA, 0xEA02, 0x420A, 0x204}},
    {34731, 2, {0x820, 0x7F20, 0x83E, 0xFF44, 0x14A8, 0x5610, 0xA528, 0x4C46, 0x8100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34732, 2, {0x2090, 0x2090, 0x27FE, 0xF890, 0xABFC, 0xAA94, 0xABFC, 0xAA94, 0xFBFC, 0xA000, 0x21F8, 0x2908, 0x39F8, 0xE908, 0x41F8, 0x108}},
    {34733, 2, {0x2040, 0x2020, 0x23FE, 0xF800, 0xA954, 0xA924, 0xA954, 0xA9FC, 0xF820, 0xA3FE, 0x2242, 0x2A92, 0x3AFA, 0xEA0A, 0x4202, 0x206}},
    {34734, 2, {0x2154, 0x2154, 0x23FE, 0xF954, 0xA954, 0xAA72, 0xAC00, 0xABFE, 0xFA22, 0xA020, 0x21FC, 0x2924, 0x3924, 0xE924, 0x412C, 0x20}},
    {34735, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x8100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34736, 2, {0x2040, 0x2020, 0x23FE, 0xFA50, 0xAA50, 0xABFE, 0xAA52, 0xAA52, 0xFBFE, 0xA200, 0x2292, 0x2AD4, 0x3A98, 0xEC92, 0x44D2, 0x88E}},
    {34737, 2, {0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5FA8, 0x4408, 0x9528, 0x2C90, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34738, 2, {0x2040, 0x2020, 0x23FE, 0xFA8A, 0xA904, 0xABFE, 0xA840, 0xA888, 0xF9FC, 0xA024, 0x2020, 0x29FC, 0x3820, 0xE820, 0x43FE, 0x0}},
    {34739, 2, {0x2020, 0x2124, 0x20A8, 0xFBFE, 0xAA02, 0xA8F8, 0xA888, 0xA888, 0xF8F8, 0xA020, 0x2020, 0x29FC, 0x3820, 0xE820, 0x43FE, 0x0}},
    {34740, 2, {0x20, 0x7BFE, 0x4820, 0x51FC, 0x6124, 0x51FC, 0x4924, 0x49FC, 0x6870, 0x50A8, 0x4124, 0x3FFA, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34741, 2, {0x1000, 0x13FE, 0x1050, 0x7DFC, 0x5554, 0x5554, 0x55FC, 0x5400, 0x7DFC, 0x5000, 0x13FE, 0x1420, 0x1CA8, 0xE524, 0x42A2, 0x40}},
    {34742, 2, {0x0, 0x3FFC, 0x2000, 0x3FF8, 0x2008, 0x3FF8, 0x2000, 0x3FFC, 0x1020, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {34743, 2, {0x80, 0x3FFE, 0x2200, 0x2FF8, 0x2888, 0x2FF8, 0x2908, 0x2FF8, 0x2220, 0x3FFC, 0x20A0, 0x2FF8, 0x4888, 0x4FF8, 0x8084, 0x3FFE}},
    {34744, 2, {0x1420, 0x2220, 0x493E, 0x1442, 0x2294, 0x4110, 0xBE28, 0x2228, 0x2244, 0x3E82, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34745, 2, {0x2080, 0x2080, 0x208E, 0xFBE8, 0xA888, 0xABE8, 0xAAAE, 0xABEA, 0xFAAA, 0xA3EA, 0x208A, 0x2BEA, 0x388A, 0xE892, 0x4092, 0xA2}},
    {34746, 2, {0x1000, 0x11FC, 0x1124, 0x7DFC, 0x5524, 0x55FC, 0x5440, 0x5488, 0x7DF0, 0x5020, 0x1044, 0x17FE, 0x1C22, 0xE524, 0x4222, 0x60}},
    {34747, 2, {0x2020, 0x21FC, 0x2124, 0xFBFE, 0xA924, 0xA9FC, 0xA820, 0xA9FC, 0xF924, 0xA1FC, 0x2040, 0x2BFE, 0x3888, 0xE9D0, 0x4070, 0x38C}},
    {34748, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x54F8, 0x5420, 0x55FC, 0x5524, 0x7DFC, 0x5020, 0x13FE, 0x1420, 0x1DFC, 0xE420, 0x43FE, 0x0}},
    {34749, 2, {0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE18E, 0xC40, 0x300, 0x10A0, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {34750, 2, {0x1040, 0x1020, 0x13FE, 0x7E02, 0x5400, 0x55FC, 0x5420, 0x55FC, 0x7D24, 0x51FC, 0x1124, 0x15FC, 0x1C00, 0xE488, 0x4104, 0x202}},
    {34751, 2, {0x440, 0x247C, 0x24A8, 0x3D50, 0x428, 0x7DFE, 0x2488, 0x4448, 0x8418, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34752, 2, {0x2040, 0x2020, 0x23FE, 0xF820, 0xA848, 0xAAF2, 0xA924, 0xAA52, 0xF8F8, 0xA008, 0x2020, 0x2BFE, 0x3820, 0xE820, 0x4020, 0x20}},
    {34753, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2100, 0x3FFC, 0x2080, 0x2C64, 0x301C, 0x1020, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {34754, 2, {0x1040, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5500, 0x55FE, 0x7D00, 0x51FE, 0x1022, 0x15FA, 0x1C76, 0xE4A8, 0x4326, 0x20}},
    {34755, 2, {0x2000, 0x21F8, 0x2108, 0xF9F8, 0xA908, 0xA9F8, 0xA800, 0xABFC, 0xFA94, 0xA3FC, 0x2000, 0x29F8, 0x3890, 0xE860, 0x4198, 0x606}},
    {34756, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x882, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34757, 2, {0x1040, 0x1020, 0x11FE, 0x7D00, 0x5548, 0x5548, 0x55FE, 0x5548, 0x7D48, 0x1148, 0x1578, 0x1D00, 0xE6D4, 0x42AA, 0x52A, 0x800}},
    {34758, 2, {0x1088, 0x1088, 0x13FE, 0x7C88, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x7D04, 0x51FC, 0x1020, 0x17FE, 0x1C50, 0xE488, 0x4104, 0x202}},
    {34759, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1110, 0x3FF8, 0xD116, 0x1FF0, 0x108, 0x7FFC}},
    {34760, 2, {0x2000, 0x23FE, 0x221A, 0xFA16, 0xABFE, 0xAA12, 0xAAEA, 0xAAAA, 0xFAAA, 0xA2EA, 0x220E, 0x2A6A, 0x3B9A, 0xEA26, 0x43FE, 0x202}},
    {34761, 2, {0x2000, 0x23DE, 0x2042, 0xFA52, 0xA94A, 0xAA52, 0xA828, 0xA8C4, 0xFB12, 0xA060, 0x2188, 0x2832, 0x39C4, 0xE818, 0x4060, 0x380}},
    {34762, 2, {0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x100, 0x1020, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {34763, 2, {0x201C, 0x21E0, 0x2020, 0xF924, 0xA8A8, 0xABFE, 0xA870, 0xA8A8, 0xF924, 0xA222, 0x2040, 0x2824, 0x3AA2, 0xEA8A, 0x4488, 0x78}},
    {34764, 2, {0x1020, 0x1040, 0x11FC, 0x7D24, 0x557C, 0x558C, 0x5554, 0x5524, 0x7D54, 0x51FC, 0x1020, 0x1410, 0x1D54, 0xE542, 0x414A, 0x238}},
    {34765, 2, {0x2062, 0x238A, 0x208A, 0xFFEA, 0xA88A, 0xA9CA, 0xAAA2, 0xAC8A, 0xF8A4, 0xA020, 0x23FE, 0x2870, 0x38A8, 0xE924, 0x4222, 0x20}},
    {34766, 2, {0x2088, 0x2088, 0x23FE, 0xF888, 0xA888, 0xA8F8, 0xA820, 0xABFE, 0xFA22, 0xA332, 0x22AA, 0x2B76, 0x3A22, 0xEA22, 0x422A, 0x224}},
    {34767, 2, {0x2088, 0x2088, 0x23FE, 0xF888, 0xA820, 0xABFC, 0xA824, 0xAFFE, 0xF824, 0xA3FC, 0x2020, 0x2AB4, 0x3AAC, 0xEB2C, 0x4224, 0x424}},
    {34768, 2, {0x2020, 0x2124, 0x20A8, 0xFBFE, 0xAA02, 0xA8F8, 0xA888, 0xA888, 0xF8F8, 0xA020, 0x21FC, 0x2924, 0x3924, 0xE934, 0x4128, 0x20}},
    {34769, 2, {0x1020, 0x11FC, 0x1088, 0x7C50, 0x57FE, 0x5400, 0x55FC, 0x5504, 0x7DFC, 0x5104, 0x11FC, 0x1420, 0x1FFE, 0xE420, 0x4020, 0x20}},
    {34770, 2, {0x2088, 0x2088, 0x23FE, 0xF888, 0xA824, 0xABFE, 0xA850, 0xA888, 0xF904, 0xA28A, 0x2088, 0x2BFE, 0x3888, 0xE888, 0x4108, 0x208}},
    {34771, 2, {0x2080, 0x20F8, 0x2108, 0xFBFE, 0xAD12, 0xA922, 0xA9FE, 0xA840, 0xF8A2, 0xA354, 0x2098, 0x2B34, 0x3854, 0xE892, 0x4350, 0x20}},
    {34772, 2, {0x2000, 0x21FC, 0x2124, 0xF9AC, 0xA974, 0xA924, 0xA9FC, 0xA820, 0xF9FC, 0xA020, 0x23FE, 0x2800, 0x3AA4, 0xEA52, 0x4452, 0x0}},
    {34773, 2, {0x2090, 0x2294, 0x22D8, 0xFA92, 0xAAD2, 0xAF0E, 0xA9F0, 0xAA10, 0xFFFC, 0xAA44, 0x23FC, 0x2A44, 0x3BFC, 0xEC44, 0x4454, 0x808}},
    {34774, 2, {0x2140, 0x2144, 0x2158, 0xFBF0, 0xA950, 0xA9D0, 0xA95E, 0xA9D4, 0xF954, 0xA154, 0x23F4, 0x2814, 0x3954, 0xEA34, 0x4414, 0x24}},
    {34775, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x9112, 0x1FF0, 0x4104, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34776, 2, {0x1054, 0x1092, 0x119E, 0x7EF0, 0x548A, 0x5486, 0x5402, 0x54FC, 0x7C84, 0x50FC, 0x1084, 0x14FC, 0x1E84, 0xE2FC, 0x4048, 0x84}},
    {34777, 2, {0x2110, 0x2094, 0x23D2, 0xF812, 0xAA50, 0xA990, 0xAFFE, 0xA810, 0xFBD2, 0xA252, 0x2254, 0x2BD4, 0x3A4A, 0xEA4A, 0x43D6, 0x22}},
    {34778, 2, {0x804, 0x7F08, 0x810, 0x3E64, 0x8, 0x3E10, 0x2264, 0x3E08, 0x1410, 0x7F60, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34779, 2, {0x2082, 0x2082, 0x27F4, 0xF888, 0xA880, 0xABE2, 0xA802, 0xABE4, 0xFA28, 0xA220, 0x23E0, 0x2A22, 0x3942, 0xE874, 0x4788, 0x210}},
    {34780, 2, {0x1018, 0x11E0, 0x1040, 0x7FFE, 0x5488, 0x5574, 0x5652, 0x5470, 0x7C00, 0x51FC, 0x1104, 0x1574, 0x1D54, 0xE574, 0x4104, 0x10C}},
    {34781, 2, {0x2200, 0x217E, 0x2040, 0xFC7C, 0xAA44, 0xA8FC, 0xA940, 0xAE7E, 0xFA20, 0xA020, 0x23FE, 0x2870, 0x38A8, 0xE924, 0x4222, 0x20}},
    {34782, 2, {0x820, 0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34783, 2, {0x1020, 0x1020, 0x11FC, 0x7C20, 0x5554, 0x5488, 0x5504, 0x56FA, 0x7C88, 0x50F8, 0x1088, 0x14F8, 0x1E20, 0xE2A8, 0x4124, 0x60}},
    {34784, 2, {0x203C, 0x23E0, 0x2124, 0xF8A8, 0xABFE, 0xA8A8, 0xA924, 0xAA02, 0xF9FC, 0xA124, 0x2124, 0x29FC, 0x3924, 0xE924, 0x41FC, 0x104}},
    {34785, 2, {0x200E, 0x23F0, 0x2044, 0xFA24, 0xA908, 0xABF8, 0xAA08, 0xABFC, 0xFA04, 0xA3FE, 0x2202, 0x2AAA, 0x3AAA, 0xED02, 0x4414, 0x808}},
    {34786, 2, {0x1020, 0x13FE, 0x1020, 0x7DFC, 0x5400, 0x55FC, 0x5504, 0x55FC, 0x7C88, 0x53FE, 0x1000, 0x15FC, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34787, 2, {0x2124, 0x2124, 0x22AA, 0xFBAE, 0xA924, 0xAAAA, 0xABAE, 0xA924, 0xFBFE, 0xA110, 0x2114, 0x2914, 0x3A8A, 0xEA4A, 0x4216, 0x422}},
    {34788, 2, {0x2000, 0x23DE, 0x2252, 0xFBDE, 0xAA10, 0xAA52, 0xA9CE, 0xA800, 0xF888, 0xA3FE, 0x2088, 0x2888, 0x3BFE, 0xE888, 0x4104, 0x202}},
    {34789, 2, {0x1088, 0x1088, 0x11FC, 0x7C88, 0x5488, 0x57FE, 0x5420, 0x55FC, 0x7D24, 0x51FC, 0x1124, 0x15FC, 0x1C00, 0xE488, 0x4104, 0x202}},
    {34790, 2, {0x1020, 0x11FC, 0x1020, 0x7C88, 0x57FE, 0x5488, 0x55FC, 0x5504, 0x7DFC, 0x5104, 0x11FC, 0x1504, 0x1DFC, 0xE488, 0x4104, 0x202}},
    {34791, 2, {0x2088, 0x22AA, 0x22DC, 0xFC88, 0xA954, 0xAA22, 0xA800, 0xABFE, 0xFA42, 0xA444, 0x23FC, 0x2844, 0x3844, 0xE884, 0x4114, 0x208}},
    {34792, 2, {0x3FFE, 0x2890, 0x2510, 0x3FDE, 0x2222, 0x2A88, 0x2F88, 0x2294, 0x2422, 0x2840, 0x27FC, 0x2444, 0x27FC, 0x4044, 0x4FFE, 0x8002}},
    {34793, 2, {0x2000, 0x2FFE, 0x2800, 0xFA28, 0xA948, 0xABEE, 0xA892, 0xA884, 0xFAA0, 0xAAA8, 0x2BE8, 0x2888, 0x3894, 0xE914, 0x4924, 0x1242}},
    {34794, 2, {0x2020, 0x2020, 0x23FE, 0xF820, 0xA9FC, 0xA924, 0xA9FC, 0xA924, 0xF9FC, 0xA022, 0x23FE, 0x2842, 0x3824, 0xEAA2, 0x428A, 0x478}},
    {34795, 2, {0x2000, 0x23FE, 0x2050, 0xFBFE, 0xAA52, 0xABFE, 0xA800, 0xA9FC, 0xF904, 0xA1FC, 0x2104, 0x29FC, 0x3820, 0xEBFE, 0x4020, 0x20}},
    {34796, 2, {0x2000, 0x23DE, 0x2252, 0xFA52, 0xABDE, 0xA800, 0xA9FC, 0xA924, 0xF9FC, 0xA124, 0x21FC, 0x2820, 0x3BFE, 0xE820, 0x4020, 0x20}},
    {34797, 2, {0x1090, 0x1088, 0x11FE, 0x7D10, 0x5710, 0x55FE, 0x5510, 0x5510, 0x7DFE, 0x1110, 0x1510, 0x1DFE, 0xE500, 0x42A4, 0x252, 0x452}},
    {34798, 2, {0x1088, 0x1050, 0x13FE, 0x7C20, 0x55FC, 0x5420, 0x57FE, 0x5524, 0x7CA8, 0x53FE, 0x1000, 0x15FC, 0x1D04, 0xE504, 0x41FC, 0x104}},
    {34799, 2, {0x2020, 0x2020, 0x21FC, 0xF820, 0xABFE, 0xA908, 0xAB9C, 0xA908, 0xF988, 0xA63E, 0x2000, 0x2BFE, 0x3890, 0xE890, 0x4112, 0x20E}},
    {34800, 2, {0x2020, 0x21FC, 0x2024, 0xFBFE, 0xA824, 0xA9FC, 0xA820, 0xAAAA, 0xFBAE, 0xA222, 0x23FE, 0x2A22, 0x3BAE, 0xEAAA, 0x42AA, 0x422}},
    {34801, 2, {0x2080, 0x2080, 0x21FE, 0xFB54, 0xAD54, 0xA954, 0xABFE, 0xA954, 0xF954, 0xA154, 0x27FE, 0x2800, 0x3954, 0xE92A, 0x422A, 0x0}},
    {34802, 2, {0x100, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4, 0x1020, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {34803, 2, {0x1000, 0x11FC, 0x1004, 0x7CFC, 0x5404, 0x55FC, 0x5400, 0x55DC, 0x7C94, 0x53DC, 0x1008, 0x17FE, 0x1D08, 0xE488, 0x40A8, 0x10}},
    {34804, 2, {0x240C, 0xFF70, 0x2440, 0x3C40, 0x247E, 0x3C48, 0x2448, 0xFF48, 0x2448, 0x4288, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34805, 2, {0x2000, 0x23FE, 0x2200, 0xFA04, 0xABF4, 0xAA04, 0xAAEE, 0xAAA4, 0xFAA4, 0xA2EC, 0x2204, 0x2AA4, 0x3A44, 0xEA74, 0x4594, 0x8}},
    {34806, 2, {0x2000, 0x27E0, 0x225E, 0xFBD2, 0xAA52, 0xABD2, 0xAA72, 0xAFDE, 0xF840, 0xA3FE, 0x2020, 0x29FC, 0x3820, 0xE820, 0x43FE, 0x0}},
    {34807, 2, {0x2020, 0x2124, 0x20A8, 0xFBFE, 0xAA02, 0xA8F8, 0xA888, 0xA888, 0xF8F8, 0xA000, 0x21FC, 0x2924, 0x39FC, 0xE924, 0x41FC, 0x104}},
    {34808, 2, {0x7FFC, 0x600, 0x3B08, 0x4B0, 0x19C0, 0x62A0, 0xC98, 0x7286, 0x1120, 0x7EFC, 0x52A4, 0x7EFC, 0x1020, 0x1424, 0xFEFE, 0x202}},
    {34809, 2, {0x2000, 0x3EFC, 0x4424, 0xBE54, 0x2A98, 0x3E50, 0x2AFC, 0x3E10, 0x2AFE, 0x2A10, 0x4510, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34810, 2, {0x1010, 0x11FE, 0x1000, 0x7DFE, 0x5502, 0x557A, 0x554A, 0x55FE, 0x7C00, 0x50FC, 0x1084, 0x14FC, 0x1E84, 0xE2FC, 0x4000, 0x1FE}},
    {34811, 2, {0x2104, 0x2088, 0x23FE, 0xF820, 0xA9FC, 0xA820, 0xABFE, 0xA854, 0xF992, 0xA090, 0x23FE, 0x2890, 0x38D4, 0xEB8A, 0x4096, 0x1A2}},
    {34812, 2, {0x2420, 0xFF20, 0x247E, 0x7EC4, 0x8228, 0x7A10, 0x4A28, 0x7AC6, 0x500, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x2004}},
    {34813, 2, {0x1010, 0x1210, 0x117C, 0x7C10, 0x54FE, 0x5444, 0x5728, 0x55FE, 0x7D10, 0x517C, 0x1110, 0x15FE, 0x1D10, 0xE510, 0x42FE, 0x400}},
    {34814, 2, {0x1080, 0x10F8, 0x1108, 0x7FFE, 0x5544, 0x5592, 0x55FE, 0x5500, 0x7D7C, 0x5100, 0x117C, 0x1500, 0x1D7C, 0xE544, 0x417C, 0x244}},
    {34815, 2, {0x878, 0xFFC8, 0x848, 0x3E86, 0x2B78, 0x3E48, 0x2A28, 0xFF90, 0x4928, 0x7F46, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34816, 2, {0x1220, 0x113E, 0x1042, 0x7C94, 0x5710, 0x5528, 0x5446, 0x55FC, 0x7D04, 0x51FC, 0x1104, 0x15FC, 0x1D04, 0xE5FC, 0x4088, 0x104}},
    {34817, 2, {0x1100, 0x2BDE, 0x4A52, 0x73D4, 0x2A52, 0x4BD2, 0x7A1A, 0x1294, 0x2350, 0xC240, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34818, 2, {0x1088, 0x13FE, 0x1088, 0x7C00, 0x5548, 0x57FE, 0x5548, 0x5578, 0x7D00, 0x51FC, 0x1020, 0x17FE, 0x1C70, 0xE4A8, 0x4326, 0x20}},
    {34819, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x711C, 0x57D4, 0x7554, 0x57DC, 0x7156, 0x57D6, 0xB062}},
    {34820, 2, {0x2020, 0x2050, 0x2088, 0xF924, 0xAEFA, 0xA850, 0xA924, 0xA954, 0xF9FC, 0xA020, 0x23FE, 0x2A42, 0x3A92, 0xEAFA, 0x420A, 0x206}},
    {34821, 2, {0x1000, 0x11FC, 0x1154, 0x7D54, 0x55DC, 0x5450, 0x55DC, 0x5554, 0x7D54, 0x51DC, 0x1154, 0x1554, 0x1DDC, 0xE550, 0x4052, 0x3E}},
    {34822, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x7FFC, 0x4124, 0x5FEC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34823, 2, {0x2088, 0x2088, 0x23FE, 0xF888, 0xA9FC, 0xA924, 0xA9FC, 0xA924, 0xF9FC, 0xA020, 0x23FE, 0x2A22, 0x3A2A, 0xEAFA, 0x420A, 0x206}},
    {34824, 2, {0x10, 0x7C14, 0x45FE, 0x7C90, 0x4494, 0x7DD4, 0x448A, 0x7C8A, 0x2916, 0x4522, 0x3FF8, 0x2108, 0x3FF8, 0x108, 0x7FFC, 0x4}},
    {34825, 2, {0x1000, 0x11FC, 0x1154, 0x7D54, 0x55FC, 0x5400, 0x57FE, 0x5400, 0x7DFC, 0x5104, 0x11FC, 0x1462, 0x1C94, 0xE588, 0x42A4, 0xC2}},
    {34826, 2, {0x2020, 0x23FE, 0x2288, 0xFA50, 0xABFE, 0xAA50, 0xAAFC, 0xAA54, 0xFBFE, 0xA254, 0x22FC, 0x2A50, 0x3AD8, 0xED54, 0x4652, 0x850}},
    {34827, 2, {0x1000, 0x11FC, 0x1154, 0x7D54, 0x55FC, 0x5480, 0x55FC, 0x5644, 0x7DF4, 0x5154, 0x1154, 0x15F4, 0x1C44, 0xE5F4, 0x4014, 0x8}},
    {34828, 2, {0x1000, 0x11FC, 0x1154, 0x7DFC, 0x5420, 0x55FC, 0x5420, 0x57FE, 0x7C88, 0x5050, 0x11FC, 0x1420, 0x1FFE, 0xE420, 0x4020, 0x20}},
    {34829, 2, {0x2008, 0x23C8, 0x2248, 0xFBCE, 0xAA4A, 0xABD2, 0xA904, 0xABE4, 0xFCA4, 0xA2A4, 0x2364, 0x2A2A, 0x3BEA, 0xE82A, 0x414A, 0x90}},
    {34830, 2, {0x1088, 0x13FE, 0x1088, 0x7C20, 0x57FE, 0x5450, 0x5488, 0x5726, 0x7CF8, 0x5020, 0x1088, 0x15FE, 0x1C88, 0xE488, 0x4108, 0x208}},
    {34831, 2, {0x2100, 0x21DE, 0x224A, 0xFC8A, 0xABEA, 0xAAB6, 0xAAA4, 0xABF4, 0xFABE, 0xA2A4, 0x23E4, 0x2ABE, 0x3AA4, 0xEAA4, 0x4224, 0x464}},
    {34832, 2, {0x2040, 0x2020, 0x23FE, 0xF888, 0xA852, 0xAFAC, 0xAAAA, 0xAAA8, 0xFDAC, 0xA000, 0x2104, 0x29FC, 0x3904, 0xE9FC, 0x4104, 0x204}},
    {34833, 2, {0x2088, 0x22AA, 0x22DC, 0xFC88, 0xA954, 0xAA22, 0xA800, 0xABFE, 0xFA22, 0xA020, 0x23FE, 0x2870, 0x38A8, 0xE924, 0x4622, 0x20}},
    {34834, 2, {0x7FFC, 0x2108, 0x7FFC, 0xB55A, 0x2928, 0x3558, 0x2108, 0x3558, 0x2928, 0x3558, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34835, 2, {0x2088, 0x23FE, 0x2088, 0xFBFE, 0xAA02, 0xA9FC, 0xA800, 0xABFE, 0xF840, 0xA0A2, 0x2354, 0x28B8, 0x3B54, 0xE892, 0x4350, 0x20}},
    {34836, 2, {0x2020, 0x23FE, 0x2000, 0xF9FC, 0xA904, 0xA9FC, 0xA800, 0xABFE, 0xFA02, 0xA1FC, 0x2040, 0x2BA4, 0x38D8, 0xEB34, 0x40D2, 0x330}},
    {34837, 2, {0x2000, 0x21FC, 0x2020, 0xFBFE, 0xAA22, 0xA9AC, 0xA820, 0xA9AC, 0xF800, 0xA3FE, 0x2020, 0x29FC, 0x3954, 0xE954, 0x4154, 0x10C}},
    {34838, 2, {0x1088, 0x13FE, 0x10A8, 0x7C90, 0x55FE, 0x5720, 0x55FC, 0x5520, 0x7DFC, 0x5120, 0x11FE, 0x1500, 0x1FFC, 0xE488, 0x4070, 0x38E}},
    {34839, 2, {0x2000, 0x23DE, 0x2042, 0xFA52, 0xA94A, 0xAA52, 0xA890, 0xA908, 0xFBFE, 0xA510, 0x21FE, 0x2910, 0x39FE, 0xE910, 0x41FE, 0x100}},
    {34840, 2, {0x2110, 0x2114, 0x23F2, 0xF910, 0xAFFE, 0xAA90, 0xAA50, 0xAFF4, 0xFA94, 0xA3F4, 0x2294, 0x2BE8, 0x3A8A, 0xEBFA, 0x4226, 0x42}},
    {34841, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0xA9FC, 0xA948, 0xAA50, 0xA9FC, 0xFB04, 0xA1FC, 0x2104, 0x29FC, 0x3904, 0xE9FC, 0x4088, 0x104}},
    {34842, 2, {0x820, 0x7FFC, 0x820, 0x200, 0xFFFE, 0x800, 0x1FF0, 0x2810, 0xC810, 0xFF0, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34843, 2, {0x1048, 0x11FE, 0x1048, 0x7C00, 0x55FE, 0x554A, 0x554A, 0x55FE, 0x7C14, 0x51FE, 0x1110, 0x1592, 0x1D4C, 0xE54A, 0x4116, 0x222}},
    {34844, 2, {0x2288, 0x2108, 0xFABE, 0x2008, 0x729C, 0xA92A, 0x2288, 0xFFFE, 0x820, 0x3018, 0xC106, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34845, 2, {0x2000, 0x21FC, 0x2124, 0xF9FC, 0xA924, 0xA9FC, 0xA904, 0xA800, 0xFBFE, 0xA2AA, 0x22AA, 0x2BFE, 0x3AAA, 0xEAAA, 0x43FE, 0x222}},
    {34846, 2, {0x2040, 0x3F7E, 0x4890, 0xBF7C, 0x2144, 0x3F44, 0x2144, 0x3F54, 0x2448, 0x3A40, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34847, 2, {0x2124, 0x2248, 0x2124, 0xF800, 0xABFC, 0xAA94, 0xAA64, 0xAA94, 0xFBFC, 0xA248, 0x2368, 0x2A48, 0x3B6A, 0xEA4A, 0x4246, 0x362}},
    {34848, 2, {0x0, 0x7FFC, 0x200, 0x3FF8, 0x2448, 0x27C8, 0x2448, 0x27C8, 0x2448, 0x3FF8, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34849, 2, {0xFE0, 0x1020, 0x1FC0, 0x40, 0xFFFE, 0xC10, 0x7320, 0xDC0, 0x72B0, 0xC8E, 0xF320, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34850, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x3FF8, 0xC826, 0xFE0, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34851, 2, {0x1000, 0x11FE, 0x1128, 0x7DFE, 0x5528, 0x557C, 0x5554, 0x557C, 0x7D54, 0x117C, 0x1510, 0x1DFE, 0xE592, 0x42BA, 0x28A, 0x484}},
    {34852, 2, {0x820, 0x440, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34853, 2, {0x5210, 0x7EFE, 0x2844, 0x7E28, 0xA2FE, 0x3E10, 0x2010, 0x3E7C, 0x2210, 0x3E10, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34854, 2, {0x2040, 0x207C, 0x2040, 0xFBFC, 0xAA44, 0xABF0, 0xAA44, 0xAAFC, 0xFAA8, 0xA2F8, 0x22A8, 0x2AF8, 0x3A00, 0xEDFC, 0x4554, 0xBFE}},
    {34855, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x0, 0x7FFE, 0x4802, 0x9FF4, 0x2810, 0xCFF0, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34856, 2, {0x2088, 0x23FE, 0x20A8, 0xF9FC, 0xA824, 0xABFE, 0xA824, 0xA9FC, 0xFAAA, 0xA3AE, 0x2222, 0x2BFE, 0x3A22, 0xEBAE, 0x42AA, 0x4AA}},
    {34857, 2, {0x2210, 0x2110, 0x203A, 0xFB92, 0xA814, 0xABBE, 0xA808, 0xAB90, 0xF83C, 0xA3E4, 0x22A4, 0x2ABC, 0x3AA4, 0xEBA4, 0x42BC, 0x24}},
    {34858, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34859, 2, {0x704, 0x7824, 0x2524, 0x1E24, 0x6424, 0x1A24, 0x6524, 0x1904, 0x6514, 0x208, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34860, 2, {0x2110, 0x2090, 0x23DE, 0xF810, 0xAA5E, 0xA982, 0xABDE, 0xA810, 0xFBDE, 0xA250, 0x23DE, 0x2A50, 0x3BDE, 0xEA50, 0x4252, 0x2CE}},
    {34861, 2, {0x2100, 0x13F0, 0x1510, 0xE0, 0x775C, 0x11F0, 0x1040, 0x13FC, 0x2840, 0x47FE, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34862, 2, {0x23EE, 0x228A, 0x22EA, 0xFB50, 0xABEE, 0xAA4A, 0xAAA4, 0xABEA, 0xF800, 0xA3DE, 0x2042, 0x2A52, 0x394A, 0xEA52, 0x414A, 0x84}},
    {34863, 2, {0x80, 0x7FFE, 0x4200, 0x4FF8, 0x4888, 0x4FF8, 0x4908, 0x4FF8, 0x4280, 0x7FFC, 0x4490, 0x5F7C, 0x5554, 0x5F7C, 0x4492, 0xBFFE}},
    {34864, 2, {0x1020, 0x13FE, 0x1000, 0x7DDC, 0x5554, 0x55DC, 0x5488, 0x57FE, 0x7C88, 0x53FE, 0x1088, 0x17FE, 0x1C94, 0xE588, 0x42A4, 0xC2}},
    {34865, 2, {0x100, 0x1FF0, 0x1110, 0x1FF0, 0x108, 0x7FFC, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {34866, 2, {0x4400, 0x24FE, 0x28AA, 0xAA, 0xFEFE, 0x40, 0x28FE, 0x4522, 0x82FA, 0xAA, 0xFEAA, 0xAAFA, 0xAA22, 0xAAFA, 0xAF0A, 0xF804}},
    {34867, 2, {0x23DE, 0x2252, 0x23DE, 0xFA52, 0xABDE, 0xAA52, 0xABDE, 0xA98C, 0xFA52, 0xA040, 0x27FE, 0x2888, 0x3990, 0xE860, 0x40D8, 0x304}},
    {34868, 2, {0x1088, 0x13FE, 0x1088, 0x7DFC, 0x5554, 0x55FC, 0x5422, 0x55FC, 0x7C28, 0x53FE, 0x1040, 0x15FC, 0x1E84, 0xE4FC, 0x4084, 0xFC}},
    {34869, 2, {0x2222, 0x23FE, 0x2090, 0xF9FE, 0xAB10, 0xADFE, 0xA910, 0xA9FE, 0xF910, 0xA1FE, 0x2100, 0x2BFE, 0x3A8A, 0xEB76, 0x4252, 0x276}},
    {34870, 2, {0x3E7C, 0x4890, 0x7EFC, 0x142A, 0x66C6, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34871, 2, {0x23DE, 0x2252, 0x23DE, 0xFA52, 0xABDE, 0xAA52, 0xABDE, 0xA890, 0xF9FE, 0xA110, 0x23FE, 0x2D10, 0x39FE, 0xE910, 0x41FE, 0x100}},
    {34872, 2, {0x2088, 0x23FE, 0x2088, 0xFBDE, 0xAA52, 0xABDE, 0xA8A0, 0xA890, 0xF9FE, 0xA120, 0x23FC, 0x2D20, 0x39FC, 0xE920, 0x41FE, 0x100}},
    {34873, 2, {0x100, 0x7FFC, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x8802, 0x1FF0, 0x6810, 0xFF0, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34874, 2, {0x7CF8, 0x1020, 0xFDFC, 0x2850, 0x4488, 0xBFF6, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34875, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x104, 0x7FFE}},
    {34876, 2, {0x23DE, 0x2252, 0x23DE, 0xFA52, 0xABDE, 0xA8A0, 0xA9FE, 0xAB20, 0xFDFC, 0xA120, 0x21FE, 0x2800, 0x39FC, 0xE888, 0x4070, 0x78E}},
    {34877, 2, {0x828, 0x2A24, 0x4D2E, 0x870, 0x1424, 0x3F24, 0x6428, 0xBF12, 0x242A, 0x3F46, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34878, 2, {0x23FE, 0x2202, 0x23FE, 0xFA92, 0xAA54, 0xAA92, 0xAAFE, 0xAAAA, 0xFAFE, 0xA322, 0x22FA, 0x2AAA, 0x3AFA, 0xEA22, 0x45FA, 0x4}},
    {34879, 2, {0x2490, 0x79E4, 0x1454, 0x7DFC, 0x4, 0xFFFC, 0x2494, 0x79E4, 0x1454, 0xFFFC, 0x1020, 0x7CF8, 0x54A8, 0x7CF8, 0x1224, 0xFEFC}},
    {34880, 2, {0x100, 0x100, 0x200, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {34881, 2, {0x200, 0x100, 0xFFFE, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x100, 0x200, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0xFFFE, 0x0}},
    {34882, 2, {0x800, 0x800, 0x107E, 0x7F12, 0x5512, 0x5552, 0x5552, 0x5552, 0x5552, 0x5592, 0x5512, 0x5522, 0x57A2, 0xFC42, 0x94, 0x108}},
    {34883, 2, {0x800, 0x800, 0x10FE, 0x7F10, 0x5510, 0x5520, 0x5530, 0x5558, 0x5554, 0x5592, 0x5512, 0x5510, 0x5790, 0xFC10, 0x10, 0x10}},
    {34884, 2, {0x800, 0x8FC, 0x1024, 0x7F24, 0x5524, 0x5524, 0x5524, 0x55FC, 0x5544, 0x5544, 0x5544, 0x5544, 0x57C4, 0xFC44, 0x1FE, 0x0}},
    {34885, 2, {0x810, 0x810, 0x1094, 0x7F54, 0x5558, 0x5510, 0x55FC, 0x5510, 0x5510, 0x5510, 0x55FE, 0x5510, 0x5790, 0xFC10, 0x10, 0x10}},
    {34886, 2, {0x400, 0x800, 0x3FF8, 0x2448, 0x2448, 0xFFFE, 0x0, 0xF8, 0x7F00, 0x904, 0x1188, 0x6350, 0x520, 0x1918, 0xE106, 0x100}},
    {34887, 2, {0x804, 0x80E, 0x1070, 0x7F40, 0x5546, 0x5558, 0x5554, 0x5554, 0x5556, 0x5554, 0x5554, 0x5554, 0x57D2, 0xFC92, 0x98, 0x110}},
    {34888, 2, {0x800, 0x87E, 0x1024, 0x7F24, 0x553C, 0x5524, 0x5524, 0x553C, 0x5524, 0x5524, 0x5526, 0x557C, 0x5784, 0xFC04, 0x4, 0x4}},
    {34889, 2, {0x820, 0x820, 0x103C, 0x7F44, 0x55A8, 0x5510, 0x5528, 0x5544, 0x5582, 0x557C, 0x5544, 0x5544, 0x57C4, 0xFC44, 0x7C, 0x44}},
    {34890, 2, {0x1024, 0x10FE, 0x2024, 0xFE00, 0xAAFE, 0xAAAA, 0xAAAA, 0xAAFE, 0xAA14, 0xAAFE, 0xAA90, 0xAAD2, 0xAFAC, 0xF8AA, 0x96, 0x122}},
    {34891, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7D7C, 0x1110, 0x7D7C, 0x4444, 0x7C7C, 0x4544, 0x7FFC, 0x2448, 0xFFFE}},
    {34892, 2, {0x800, 0x9FC, 0x1000, 0x2000, 0x4800, 0x800, 0x13FE, 0x3020, 0x5020, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {34893, 2, {0x1000, 0x1200, 0x211C, 0x4100, 0x8000, 0x1400, 0x223E, 0x6288, 0xA088, 0x2108, 0x2708, 0x2108, 0x2108, 0x2108, 0x2128, 0x2010}},
    {34894, 2, {0x1000, 0x1000, 0x27DC, 0x4100, 0x8100, 0x1100, 0x213E, 0x6108, 0xAFE8, 0x2108, 0x2108, 0x2108, 0x2108, 0x2108, 0x2128, 0x2110}},
    {34895, 2, {0x1000, 0x1000, 0x2F9C, 0x4000, 0x8000, 0x1FC0, 0x253E, 0x6508, 0xA508, 0x2508, 0x2548, 0x2988, 0x2908, 0x3008, 0x2028, 0x2010}},
    {34896, 2, {0x1000, 0x1000, 0x27DC, 0x4400, 0x8400, 0x1780, 0x24BE, 0x6488, 0xA488, 0x2788, 0x2408, 0x2408, 0x27C8, 0x2008, 0x2028, 0x2010}},
    {34897, 2, {0x1100, 0x1100, 0x229C, 0x4240, 0x8420, 0x1900, 0x20BE, 0x6008, 0xAFC8, 0x2048, 0x2088, 0x2508, 0x2208, 0x2108, 0x2128, 0x2010}},
    {34898, 2, {0x1200, 0x1100, 0x201C, 0x4FC0, 0x8200, 0x1400, 0x24BE, 0x6888, 0xAF08, 0x2108, 0x2208, 0x2208, 0x2488, 0x2F88, 0x20A8, 0x2010}},
    {34899, 2, {0x1100, 0x1180, 0x215C, 0x4140, 0x8100, 0x17C0, 0x213E, 0x6308, 0xA388, 0x2348, 0x2548, 0x2508, 0x2908, 0x2108, 0x2128, 0x2110}},
    {34900, 2, {0x1200, 0x1200, 0x239C, 0x4200, 0x8400, 0x1780, 0x2A3E, 0x6208, 0xAF88, 0x2208, 0x2208, 0x2208, 0x2288, 0x2308, 0x2228, 0x2010}},
    {34901, 2, {0x2000, 0x2FE0, 0x482E, 0x4820, 0x8BA0, 0x1820, 0x283E, 0x6BA4, 0xAAA4, 0x2AA4, 0x2AA4, 0x2BA4, 0x2824, 0x2824, 0x28B4, 0x2848}},
    {34902, 2, {0x1480, 0x1480, 0x249C, 0x4480, 0x9FE0, 0x1480, 0x24BE, 0x6488, 0xA488, 0x2FE8, 0x2008, 0x2488, 0x2448, 0x2848, 0x3008, 0x2018}},
    {34903, 2, {0x1100, 0x1100, 0x27DC, 0x4100, 0x8100, 0x17C0, 0x203E, 0x6108, 0xA108, 0x27C8, 0x2108, 0x2108, 0x21C8, 0x2E08, 0x2428, 0x2010}},
    {34904, 2, {0x1400, 0x1400, 0x27EE, 0x4900, 0x9100, 0x1100, 0x2FDE, 0x6104, 0xA104, 0x25C4, 0x2504, 0x2504, 0x25C4, 0x3E04, 0x2814, 0x2008}},
    {34905, 2, {0x1000, 0x1FC0, 0x221C, 0x4200, 0x8F80, 0x1480, 0x24BE, 0x6FC8, 0xA008, 0x2008, 0x2788, 0x2488, 0x2488, 0x2488, 0x27A8, 0x2490}},
    {34906, 2, {0x1400, 0x14E0, 0x24AE, 0x4EA0, 0x84A0, 0x14E0, 0x24BE, 0x6EA4, 0xAAA4, 0x2AE4, 0x2AA4, 0x2AA4, 0x2EA4, 0x2124, 0x2134, 0x2268}},
    {34907, 2, {0x1200, 0x1200, 0x2FCE, 0x4440, 0x9FE0, 0x1000, 0x2FDE, 0x6844, 0xAFC4, 0x2104, 0x2FC4, 0x2104, 0x2904, 0x2FE4, 0x2114, 0x2108}},
    {34908, 2, {0x1440, 0x1280, 0x2FEE, 0x4100, 0x8200, 0x17C0, 0x245E, 0x6444, 0xA7C4, 0x2444, 0x2444, 0x27C4, 0x2444, 0x2444, 0x27D4, 0x2448}},
    {34909, 2, {0x1080, 0x11C0, 0x271C, 0x4100, 0x87C0, 0x1100, 0x27FE, 0x6548, 0xA7C8, 0x2548, 0x27C8, 0x2108, 0x27C8, 0x2108, 0x2FE8, 0x2010}},
    {34910, 2, {0x1100, 0x17C0, 0x224E, 0x4FE0, 0x8000, 0x17C0, 0x245E, 0x67C4, 0xA004, 0x2FE4, 0x2104, 0x2FE4, 0x2924, 0x2924, 0x2974, 0x2108}},
    {34911, 2, {0x1540, 0x1540, 0x2A8E, 0x4540, 0x8540, 0x1000, 0x2FDE, 0x6104, 0xA7C4, 0x2444, 0x27C4, 0x2444, 0x27C4, 0x2444, 0x27D4, 0x2448}},
    {34912, 2, {0x1100, 0x1100, 0x2FEE, 0x4100, 0x87C0, 0x1440, 0x27DE, 0x6444, 0xA7C4, 0x2444, 0x27C4, 0x2444, 0x2FE4, 0x2284, 0x2454, 0x2828}},
    {34913, 2, {0x1200, 0x13C0, 0x244E, 0x4880, 0x87C0, 0x1540, 0x255E, 0x67C4, 0xA544, 0x2544, 0x27C4, 0x2104, 0x2FE4, 0x2104, 0x2294, 0x2448}},
    {34914, 2, {0x2EE0, 0x2AA0, 0x4EEE, 0x4AA0, 0x8EE0, 0x1AA0, 0x2EFE, 0x6444, 0xAFF4, 0x3484, 0x27E4, 0x2484, 0x27E4, 0x2484, 0x27F4, 0x2408}},
    {34915, 2, {0x200, 0x100, 0x100, 0xFFFE, 0x200, 0x200, 0x500, 0x508, 0x890, 0x18A0, 0x2840, 0x4820, 0x8810, 0xA08, 0xC06, 0x800}},
    {34916, 2, {0x2000, 0x1000, 0x0, 0xFC00, 0x800, 0x1000, 0x1000, 0x3400, 0x5800, 0x9400, 0x1400, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000}},
    {34917, 2, {0x1040, 0x840, 0x40, 0x7E40, 0x240, 0x460, 0x850, 0x1A48, 0x2C44, 0x4A44, 0xA40, 0x840, 0x840, 0x840, 0x840, 0x840}},
    {34918, 2, {0x2000, 0x11FC, 0x20, 0xF820, 0x820, 0x1020, 0x1420, 0x3BFE, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {34919, 2, {0x2000, 0x11FC, 0x20, 0xF820, 0x820, 0x1020, 0x1420, 0x3BFE, 0x5420, 0x9020, 0x1020, 0x1020, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {34920, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x500, 0x888, 0x1850, 0x2820, 0xC918, 0xA06, 0xC00}},
    {34921, 2, {0x2000, 0x13FC, 0x104, 0xF944, 0x924, 0x10A8, 0x1088, 0x3488, 0x5850, 0x9450, 0x1420, 0x1020, 0x1050, 0x1088, 0x1104, 0x1602}},
    {34922, 2, {0x2020, 0x1020, 0x20, 0xF920, 0x92C, 0x1134, 0x1164, 0x37A4, 0x5924, 0x9534, 0x1528, 0x1122, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {34923, 2, {0x1004, 0x804, 0x8, 0x7E10, 0x220, 0x444, 0x804, 0x1A08, 0x2C10, 0x4A22, 0xA42, 0x804, 0x808, 0x810, 0x820, 0x8C0}},
    {34924, 2, {0x2010, 0x1010, 0x10, 0xFC10, 0x9FE, 0x1010, 0x1010, 0x3410, 0x5910, 0x9490, 0x1490, 0x1010, 0x1010, 0x1010, 0x1050, 0x1020}},
    {34925, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x9FC, 0x1020, 0x1020, 0x3420, 0x5BFE, 0x9420, 0x1450, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {34926, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x820, 0x1210, 0x2448, 0x4824, 0x1FF0, 0x290, 0x444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {34927, 2, {0x2010, 0x1090, 0x90, 0xFC88, 0x908, 0x1104, 0x1204, 0x35FA, 0x5888, 0x9488, 0x1488, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410}},
    {34928, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0xFFFE, 0x1010, 0x1FF0, 0x280, 0x444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {34929, 2, {0x2000, 0x13FC, 0x84, 0xF888, 0x888, 0x1090, 0x109C, 0x3484, 0x5944, 0x9544, 0x1528, 0x1128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {34930, 2, {0x2020, 0x1020, 0x20, 0xFDFE, 0x922, 0x1122, 0x1122, 0x3522, 0x5952, 0x954A, 0x158A, 0x1102, 0x1102, 0x1102, 0x110A, 0x1104}},
    {34931, 2, {0x2010, 0x1090, 0x90, 0xF888, 0x888, 0x1124, 0x1524, 0x3A22, 0x5440, 0x9040, 0x1088, 0x1084, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {34932, 2, {0x2020, 0x1020, 0x20, 0xFDFE, 0x922, 0x1224, 0x1020, 0x3420, 0x5850, 0x9450, 0x1450, 0x1050, 0x1090, 0x1092, 0x1112, 0x120E}},
    {34933, 2, {0x2000, 0x1000, 0x1FC, 0xFD04, 0x904, 0x1104, 0x1104, 0x3504, 0x59FC, 0x9504, 0x1504, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {34934, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x9FC, 0x1124, 0x1124, 0x3524, 0x5924, 0x95FC, 0x1524, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {34935, 2, {0x200, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x280, 0x444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {34936, 2, {0x2040, 0x1040, 0xA0, 0xF8A0, 0x910, 0x1208, 0x1406, 0x3110, 0x5510, 0x9910, 0x1510, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410}},
    {34937, 2, {0x2004, 0x101E, 0x1F0, 0xFD10, 0x910, 0x1110, 0x1110, 0x35FE, 0x5910, 0x9510, 0x1510, 0x1108, 0x110A, 0x114A, 0x1186, 0x1102}},
    {34938, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x7FFC, 0x2040, 0x3FFE, 0x240, 0xD40, 0x7080, 0x308, 0x490, 0xC60, 0x3518, 0xC606, 0x400}},
    {34939, 2, {0x2000, 0x10FC, 0x84, 0xFC84, 0x884, 0x11FE, 0x1084, 0x3484, 0x5884, 0x95FE, 0x1484, 0x1084, 0x1084, 0x1084, 0x1094, 0x1088}},
    {34940, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x820, 0x1020, 0x1020, 0x35FC, 0x5884, 0x9488, 0x1448, 0x1050, 0x1020, 0x1050, 0x1188, 0x1606}},
    {34941, 2, {0x2008, 0x103C, 0x1E0, 0xFC20, 0x820, 0x1020, 0x1020, 0x35FE, 0x5820, 0x9420, 0x1420, 0x1020, 0x1020, 0x1020, 0x11FC, 0x1000}},
    {34942, 2, {0x100, 0x280, 0x440, 0x920, 0x3098, 0xCFE6, 0x40, 0x280, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {34943, 2, {0x2020, 0x1020, 0x50, 0xFC50, 0x888, 0x1144, 0x1222, 0x3420, 0x5800, 0x95FC, 0x1404, 0x1008, 0x1008, 0x1010, 0x1010, 0x1020}},
    {34944, 2, {0x2040, 0x1040, 0x80, 0xFCFC, 0x904, 0x1204, 0x1084, 0x3444, 0x5844, 0x9414, 0x1424, 0x1044, 0x1184, 0x1004, 0x1028, 0x1010}},
    {34945, 2, {0x100, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x504, 0x888, 0x3850, 0xCA20, 0xC18, 0x806}},
    {34946, 2, {0x2020, 0x1020, 0x20, 0xF9FC, 0x824, 0x1024, 0x1424, 0x3824, 0x55FE, 0x9020, 0x1050, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {34947, 2, {0x2000, 0x21FC, 0x2C44, 0xF044, 0x2084, 0x2484, 0x2914, 0x3208, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {34948, 2, {0x2008, 0x103C, 0x1E0, 0xFC20, 0x820, 0x1020, 0x11FE, 0x3420, 0x5820, 0x9450, 0x1450, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {34949, 2, {0x400, 0x1FE0, 0x1420, 0x12A0, 0x1040, 0x1FFC, 0x4, 0x214, 0x108, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {34950, 2, {0x2020, 0x1020, 0x20, 0xFDFE, 0x820, 0x1020, 0x10FC, 0x3420, 0x5820, 0x95FE, 0x1422, 0x1022, 0x102A, 0x1024, 0x1020, 0x1020}},
    {34951, 2, {0x2000, 0x10FC, 0x84, 0xFC84, 0x8A4, 0x1094, 0x1094, 0x3484, 0x59FE, 0x9484, 0x1484, 0x1084, 0x1084, 0x1104, 0x1114, 0x1208}},
    {34952, 2, {0x800, 0x800, 0x7F7C, 0x944, 0x1144, 0x1144, 0x257C, 0x4200, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {34953, 2, {0x2020, 0x1010, 0x10, 0xF9FE, 0x902, 0x1204, 0x1080, 0x3488, 0x5890, 0x94A0, 0x14C0, 0x1082, 0x1082, 0x1082, 0x107E, 0x1000}},
    {34954, 2, {0x2020, 0x1020, 0x50, 0xF850, 0x888, 0x1124, 0x1212, 0x3010, 0x55FC, 0x9804, 0x1408, 0x1088, 0x1050, 0x1020, 0x1010, 0x1010}},
    {34955, 2, {0x8A0, 0x1090, 0x30BE, 0x57C0, 0x9040, 0x1024, 0x1014, 0x120C, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {34956, 2, {0x100, 0xFFFE, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x10D0, 0x1FA0, 0x1004, 0xFFC, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {34957, 2, {0x2080, 0x1080, 0x1FC, 0xF904, 0xA04, 0x15F4, 0x1114, 0x3514, 0x5914, 0x95F4, 0x1504, 0x1128, 0x1112, 0x1102, 0x10FE, 0x1000}},
    {34958, 2, {0x2020, 0x1120, 0x120, 0xF220, 0x12FC, 0x24A4, 0x2724, 0x6924, 0xB224, 0x2A24, 0x24A4, 0x27A4, 0x20C4, 0x2044, 0x2094, 0x2108}},
    {34959, 2, {0x2040, 0x1040, 0x40, 0xFC40, 0xBFE, 0x1040, 0x1080, 0x3480, 0x58FE, 0x9510, 0x1510, 0x1210, 0x1210, 0x1410, 0x11FE, 0x1000}},
    {34960, 2, {0x2000, 0x1040, 0x20, 0xF828, 0x808, 0x1088, 0x1490, 0x3A94, 0x56A2, 0x92A2, 0x14C2, 0x1088, 0x1188, 0x1288, 0x1478, 0x1000}},
    {34961, 2, {0x2000, 0x13FC, 0x84, 0xF884, 0x884, 0x1104, 0x1514, 0x3A08, 0x5400, 0x91FC, 0x1104, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {34962, 2, {0x2000, 0x1000, 0x1FC, 0xFD04, 0x904, 0x1104, 0x11FC, 0x3504, 0x5904, 0x9504, 0x15FC, 0x1104, 0x1000, 0x1000, 0x13FE, 0x1000}},
    {34963, 2, {0x2000, 0x11F8, 0x108, 0xFD08, 0x908, 0x11F8, 0x1108, 0x3508, 0x5908, 0x95F8, 0x1508, 0x1108, 0x1108, 0x1108, 0x17FE, 0x1000}},
    {34964, 2, {0x2000, 0x1000, 0x3FE, 0xF808, 0x808, 0x11E8, 0x1128, 0x3528, 0x5928, 0x9528, 0x15E8, 0x1128, 0x1008, 0x1008, 0x1028, 0x1010}},
    {34965, 2, {0x2040, 0x1020, 0x20, 0xFBFE, 0xA02, 0x1404, 0x1000, 0x3490, 0x5890, 0x9490, 0x1488, 0x1108, 0x1108, 0x1104, 0x1204, 0x1402}},
    {34966, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x9FC, 0x1124, 0x1124, 0x3524, 0x5924, 0x95FC, 0x1524, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {34967, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x904, 0x1212, 0x1020, 0x3440, 0x5988, 0x9410, 0x1420, 0x1044, 0x1188, 0x1010, 0x1060, 0x1380}},
    {34968, 2, {0x2080, 0x1080, 0xFE, 0xFD00, 0xA20, 0x1120, 0x112C, 0x3574, 0x5BA4, 0x9524, 0x1534, 0x1128, 0x1122, 0x1102, 0x10FE, 0x1000}},
    {34969, 2, {0x2020, 0x1020, 0x40, 0xFDFC, 0x904, 0x1104, 0x1104, 0x3504, 0x59FC, 0x9504, 0x1504, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {34970, 2, {0x2050, 0x1048, 0x48, 0xFC40, 0xBFE, 0x1080, 0x1080, 0x34FC, 0x5944, 0x9544, 0x1528, 0x1128, 0x1210, 0x1228, 0x1444, 0x1182}},
    {34971, 2, {0x2004, 0x101E, 0x1F0, 0xFD10, 0x910, 0x1110, 0x1110, 0x35FE, 0x5910, 0x9510, 0x1510, 0x1108, 0x110A, 0x114A, 0x11A6, 0x1112}},
    {34972, 2, {0x2020, 0x1020, 0x20, 0xFDFE, 0x820, 0x1020, 0x1020, 0x35FC, 0x5870, 0x94A8, 0x14A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {34973, 2, {0x2088, 0x1088, 0x88, 0xF908, 0x97E, 0x1308, 0x1508, 0x3148, 0x5528, 0x9928, 0x1508, 0x1108, 0x1108, 0x1108, 0x1128, 0x1110}},
    {34974, 2, {0x200, 0x100, 0xFFFE, 0x820, 0x1010, 0x2008, 0x5FF4, 0x1010, 0x1010, 0x1FF0, 0x444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {34975, 2, {0x2020, 0x1020, 0x120, 0xF920, 0x9FC, 0x1120, 0x1220, 0x3020, 0x55FE, 0x9820, 0x1450, 0x1050, 0x1088, 0x1088, 0x1104, 0x1202}},
    {34976, 2, {0x200, 0x100, 0xFFFE, 0x1100, 0x1FF8, 0x2200, 0x7FFC, 0x480, 0x860, 0x3298, 0xC444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {34977, 2, {0x2020, 0x1020, 0x20, 0xF9FC, 0x924, 0x1124, 0x15FC, 0x3924, 0x5524, 0x9124, 0x13FE, 0x1104, 0x1104, 0x1104, 0x1114, 0x1108}},
    {34978, 2, {0x2020, 0x1020, 0x124, 0xF8A4, 0x8A8, 0x1020, 0x15FC, 0x3820, 0x5420, 0x9020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020}},
    {34979, 2, {0x2048, 0x1148, 0x148, 0xF948, 0x948, 0x13FE, 0x1148, 0x3548, 0x5948, 0x9548, 0x1578, 0x1100, 0x1100, 0x1100, 0x11FE, 0x1000}},
    {34980, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {34981, 2, {0x2000, 0x1000, 0x3FE, 0xF820, 0x820, 0x1040, 0x1440, 0x38FC, 0x5584, 0x9284, 0x1484, 0x1084, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {34982, 2, {0x2020, 0x1020, 0x124, 0xFD24, 0x924, 0x1124, 0x11FC, 0x3420, 0x5820, 0x9524, 0x1524, 0x1124, 0x1124, 0x1124, 0x11FC, 0x1004}},
    {34983, 2, {0x2080, 0x1080, 0x100, 0xF9FC, 0xA04, 0x1404, 0x11E4, 0x3524, 0x5924, 0x9524, 0x15E4, 0x1124, 0x1004, 0x1004, 0x1028, 0x1010}},
    {34984, 2, {0x2040, 0x1020, 0x20, 0xFBFE, 0x840, 0x1040, 0x1088, 0x3508, 0x5BF0, 0x9420, 0x1440, 0x1088, 0x1104, 0x13FC, 0x1104, 0x1000}},
    {34985, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x83E, 0x1020, 0x1020, 0x3420, 0x59FC, 0x9504, 0x1504, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {34986, 2, {0x2020, 0x1020, 0x20, 0xF9FC, 0x820, 0x1020, 0x1420, 0x3BFE, 0x5420, 0x9040, 0x1040, 0x1088, 0x1104, 0x13FE, 0x1102, 0x1000}},
    {34987, 2, {0x2010, 0x1010, 0x10, 0xF9FE, 0x912, 0x1114, 0x1510, 0x39FC, 0x5544, 0x9144, 0x1128, 0x1128, 0x1110, 0x1228, 0x1244, 0x1482}},
    {34988, 2, {0x200, 0x100, 0xFFFE, 0x800, 0x1010, 0x3FF8, 0x8, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x488, 0x850, 0x3820, 0xCB18, 0xC06}},
    {34989, 2, {0x440, 0x420, 0xFFFE, 0x910, 0x1160, 0x2184, 0x4F04, 0x82FC, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {34990, 2, {0x2080, 0x1080, 0x80, 0xF9FE, 0x900, 0x1200, 0x1420, 0x3020, 0x5528, 0x9924, 0x1624, 0x1222, 0x1422, 0x1020, 0x10A0, 0x1040}},
    {34991, 2, {0x2040, 0x1048, 0x244, 0xFA40, 0xBFE, 0x1080, 0x1480, 0x38FC, 0x5544, 0x9144, 0x1148, 0x1250, 0x1220, 0x1450, 0x1888, 0x1306}},
    {34992, 2, {0x1FF0, 0x1210, 0x1110, 0xFFFE, 0x1210, 0x2110, 0x3FFC, 0x10, 0x2A0, 0x140, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {34993, 2, {0x2090, 0x1094, 0x92, 0xF912, 0x910, 0x137E, 0x1510, 0x3110, 0x5510, 0x9928, 0x1528, 0x1128, 0x1128, 0x1144, 0x1144, 0x1182}},
    {34994, 2, {0x100, 0xFFFE, 0x200, 0xFE0, 0x34C0, 0x700, 0x3880, 0x3F8, 0x1D30, 0x1C0, 0x3F00, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {34995, 2, {0x2020, 0x1020, 0x7C, 0xFC84, 0x948, 0x1030, 0x1020, 0x3448, 0x5990, 0x943E, 0x1442, 0x11A4, 0x1018, 0x1010, 0x1060, 0x1180}},
    {34996, 2, {0x2020, 0x1020, 0x3FE, 0xFC50, 0x888, 0x1104, 0x12FA, 0x3400, 0x5BFE, 0x9440, 0x1480, 0x11FC, 0x1004, 0x1004, 0x1028, 0x1010}},
    {34997, 2, {0x2080, 0x1088, 0x9C, 0xF970, 0x910, 0x1310, 0x1510, 0x3110, 0x55FE, 0x9910, 0x1510, 0x1110, 0x1110, 0x1110, 0x117C, 0x1100}},
    {34998, 2, {0x2080, 0x1080, 0xFC, 0xF908, 0xA90, 0x1060, 0x1198, 0x3626, 0x5820, 0x95FC, 0x1420, 0x1220, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {34999, 2, {0x2020, 0x1020, 0x50, 0xF888, 0x904, 0x1202, 0x11FC, 0x3400, 0x5800, 0x95FC, 0x1504, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {35000, 2, {0x2040, 0x1040, 0x40, 0xFBFE, 0x880, 0x1080, 0x1178, 0x3508, 0x5B10, 0x9510, 0x15FE, 0x1110, 0x1110, 0x1110, 0x1150, 0x1120}},
    {35001, 2, {0x2000, 0x13FE, 0x20, 0xFC20, 0x840, 0x11FC, 0x1104, 0x3504, 0x5904, 0x95FC, 0x1504, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {35002, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x820, 0x1020, 0x11FC, 0x3400, 0x5800, 0x95FC, 0x1504, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {35003, 2, {0x2000, 0x13FE, 0x20, 0xFC20, 0x840, 0x11FC, 0x1154, 0x3554, 0x5954, 0x9554, 0x1554, 0x1154, 0x1154, 0x1144, 0x1114, 0x1108}},
    {35004, 2, {0x2080, 0x1080, 0xF8, 0xF908, 0xB10, 0x14A0, 0x1040, 0x34A0, 0x5918, 0x9606, 0x15F8, 0x1108, 0x1108, 0x1108, 0x11F8, 0x1108}},
    {35005, 2, {0x4100, 0x2100, 0x2100, 0x11E, 0xF7D2, 0x1252, 0x2252, 0x6A52, 0xB252, 0x2C92, 0x2292, 0x2112, 0x2292, 0x225E, 0x2452, 0x2800}},
    {35006, 2, {0x2020, 0x1120, 0x120, 0xF9FC, 0x920, 0x1220, 0x1020, 0x33FE, 0x5470, 0x98A8, 0x14A8, 0x1124, 0x1124, 0x1222, 0x1020, 0x1020}},
    {35007, 2, {0x2020, 0x1020, 0x1FC, 0xF820, 0x820, 0x1020, 0x13FE, 0x3000, 0x5420, 0x9820, 0x15FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {35008, 2, {0x2000, 0x11FE, 0x102, 0xF922, 0x922, 0x1122, 0x11FE, 0x3522, 0x5922, 0x9552, 0x154A, 0x118A, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {35009, 2, {0x840, 0x848, 0x7F44, 0x844, 0x840, 0xFFFE, 0x1040, 0x844, 0xFF44, 0x1044, 0x2928, 0x662A, 0xA212, 0x292A, 0x3046, 0x2082}},
    {35010, 2, {0x4, 0x7F84, 0x824, 0x1F24, 0x2124, 0x5224, 0xC24, 0x3004, 0xC20C, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35011, 2, {0x2040, 0x1040, 0x7C, 0xFC40, 0x840, 0x13FE, 0x1000, 0x3400, 0x5BFE, 0x9440, 0x1450, 0x1048, 0x1044, 0x1040, 0x1040, 0x1040}},
    {35012, 2, {0x2040, 0x1040, 0x9C, 0xF900, 0xA40, 0x1040, 0x10BE, 0x3588, 0x5A88, 0x9488, 0x1488, 0x1088, 0x1088, 0x1088, 0x10A8, 0x1090}},
    {35013, 2, {0x820, 0x4820, 0x2BFE, 0x820, 0x1820, 0x2820, 0xC9FC, 0xA00, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35014, 2, {0x2020, 0x1124, 0xA4, 0xFCA4, 0x8A8, 0x1020, 0x13FC, 0x3404, 0x5804, 0x9404, 0x15FC, 0x1004, 0x1004, 0x1004, 0x13FC, 0x1004}},
    {35015, 2, {0x2020, 0x1020, 0x40, 0xF880, 0x9FC, 0x1154, 0x1154, 0x3554, 0x5954, 0x9554, 0x1554, 0x1154, 0x1154, 0x1154, 0x13FE, 0x1000}},
    {35016, 2, {0x2000, 0x13FE, 0x202, 0xFC44, 0x840, 0x13FC, 0x1080, 0x34A0, 0x5920, 0x95FC, 0x1420, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {35017, 2, {0x2000, 0x11F8, 0x108, 0xF908, 0x9F8, 0x1108, 0x1108, 0x35F8, 0x5944, 0x9548, 0x1530, 0x1120, 0x1110, 0x1148, 0x1186, 0x1100}},
    {35018, 2, {0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x104, 0x3FEC, 0x480, 0x1C48, 0x6530, 0x60E}},
    {35019, 2, {0x2000, 0x13FC, 0x0, 0xFC00, 0x9F8, 0x1108, 0x1108, 0x3508, 0x59F8, 0x9400, 0x1508, 0x1088, 0x1090, 0x1000, 0x17FE, 0x1000}},
    {35020, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x820, 0x1020, 0x1124, 0x3524, 0x5924, 0x96AA, 0x1420, 0x1050, 0x1050, 0x1088, 0x1104, 0x1202}},
    {35021, 2, {0x2000, 0x13FE, 0x222, 0xFA22, 0xA22, 0x13FE, 0x1622, 0x3A62, 0x5672, 0x92AA, 0x1322, 0x1222, 0x1222, 0x1202, 0x13FE, 0x1202}},
    {35022, 2, {0x2000, 0x11FC, 0x104, 0xF904, 0x904, 0x11FC, 0x1000, 0x3400, 0x59FE, 0x9420, 0x1420, 0x11FC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {35023, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC, 0x488, 0xC50, 0x3530, 0xC60E}},
    {35024, 2, {0x2000, 0x10FC, 0x84, 0xFC84, 0x8FC, 0x1000, 0x11FE, 0x3502, 0x5902, 0x95FE, 0x1502, 0x1102, 0x11FE, 0x1102, 0x110A, 0x1104}},
    {35025, 2, {0x2040, 0x1080, 0x1F8, 0xF908, 0x9F8, 0x1108, 0x15FA, 0x390A, 0x550C, 0x93F8, 0x1018, 0x1028, 0x1048, 0x1188, 0x1628, 0x1010}},
    {35026, 2, {0x200, 0x100, 0xFFFE, 0x400, 0x3878, 0x2008, 0x3C78, 0x2008, 0x3FF8, 0x280, 0x444, 0xC28, 0x3410, 0xC508, 0x606, 0x400}},
    {35027, 2, {0x2010, 0x1014, 0x12, 0xF810, 0xBFE, 0x1010, 0x1550, 0x3954, 0x57F4, 0x9154, 0x1154, 0x1148, 0x124A, 0x125A, 0x1426, 0x1042}},
    {35028, 2, {0x200, 0x100, 0xFFFE, 0x488, 0x850, 0x3A20, 0xCC18, 0x806, 0x3FF8, 0x2448, 0x2828, 0x37D8, 0x2448, 0x27C8, 0x2028, 0x2010}},
    {35029, 2, {0x2000, 0x1050, 0x48, 0xFC84, 0x924, 0x1020, 0x1050, 0x3488, 0x5906, 0x94F8, 0x1488, 0x1088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {35030, 2, {0x2000, 0x11FE, 0x100, 0xFD00, 0x97C, 0x1100, 0x1100, 0x35FE, 0x5950, 0x9552, 0x1554, 0x1148, 0x1148, 0x1244, 0x1252, 0x1460}},
    {35031, 2, {0x2040, 0x1020, 0x3FC, 0xF840, 0x890, 0x1108, 0x13FC, 0x3004, 0x5550, 0x9950, 0x1550, 0x1150, 0x1250, 0x1252, 0x1452, 0x180E}},
    {35032, 2, {0x120, 0x110, 0x7FFC, 0x1110, 0x9A0, 0x540, 0x1930, 0x6208, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35033, 2, {0x2000, 0x11FC, 0x44, 0xF844, 0xBFE, 0x1044, 0x1044, 0x35FC, 0x5880, 0x9480, 0x15FC, 0x1284, 0x1484, 0x1084, 0x10FC, 0x1084}},
    {35034, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35035, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2002, 0x1FFE, 0x488, 0x1850, 0xEB30, 0xC0E}},
    {35036, 2, {0x2028, 0x1024, 0x3FE, 0xF820, 0x820, 0x11FC, 0x1124, 0x3524, 0x59FC, 0x9524, 0x1524, 0x11FC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {35037, 2, {0x2420, 0x2420, 0x3C20, 0x5FC, 0xFC20, 0x2420, 0x24F8, 0x4600, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35038, 2, {0x2104, 0x1084, 0x88, 0xFC10, 0x9FC, 0x1104, 0x1104, 0x3504, 0x59FC, 0x9450, 0x1450, 0x1090, 0x1092, 0x1112, 0x120E, 0x1400}},
    {35039, 2, {0x1020, 0x820, 0x40A8, 0x24A4, 0x922, 0x7228, 0x1030, 0x10C0, 0x700, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35040, 2, {0x3FF8, 0x208, 0xFFFE, 0x408, 0x3FF8, 0x800, 0x1FF8, 0x2808, 0xCFF8, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35041, 2, {0x2000, 0x11FC, 0x124, 0xFD24, 0x9FC, 0x1124, 0x1124, 0x35FC, 0x5820, 0x9420, 0x15FC, 0x1020, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {35042, 2, {0x2010, 0x1210, 0x110, 0xF97E, 0x820, 0x1028, 0x1748, 0x357E, 0x5908, 0x9508, 0x15FE, 0x1108, 0x1108, 0x1288, 0x147E, 0x1000}},
    {35043, 2, {0x2020, 0x1020, 0x50, 0xF888, 0x904, 0x1202, 0x11FC, 0x3400, 0x5844, 0x9424, 0x1124, 0x10A8, 0x1088, 0x1010, 0x13FE, 0x1000}},
    {35044, 2, {0x2010, 0x1008, 0x1FE, 0xFD10, 0x910, 0x11FE, 0x1120, 0x3528, 0x5948, 0x957E, 0x1508, 0x1108, 0x11FE, 0x1208, 0x1208, 0x1408}},
    {35045, 2, {0x2100, 0x10BE, 0x282, 0xF202, 0x1202, 0x22FA, 0x228A, 0x6A8A, 0xB2FA, 0x2A8A, 0x228A, 0x22FA, 0x2202, 0x2202, 0x220A, 0x2204}},
    {35046, 2, {0x100, 0xFFFE, 0x0, 0x19F8, 0x6090, 0x4060, 0x7BFC, 0x40A4, 0x4128, 0x7A60, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35047, 2, {0x2020, 0x10A2, 0xA2, 0xF924, 0x850, 0x1088, 0x1304, 0x3022, 0x5420, 0x98A4, 0x14A4, 0x1128, 0x1050, 0x1088, 0x1104, 0x1202}},
    {35048, 2, {0x2040, 0x1080, 0x3FC, 0xFA24, 0xA24, 0x13FC, 0x1624, 0x3A44, 0x57FC, 0x9090, 0x1110, 0x17FE, 0x1010, 0x1010, 0x1010, 0x1010}},
    {35049, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x9FC, 0x1104, 0x1104, 0x35FC, 0x5800, 0x9512, 0x15D4, 0x1118, 0x1110, 0x1152, 0x1192, 0x110E}},
    {35050, 2, {0x2080, 0x1080, 0xFE, 0xFD02, 0xA82, 0x10F2, 0x1142, 0x3442, 0x5BFA, 0x9442, 0x1552, 0x1152, 0x11F2, 0x1002, 0x1014, 0x1008}},
    {35051, 2, {0x2050, 0x1252, 0x252, 0xF252, 0x13DE, 0x2202, 0x2202, 0x6BFE, 0xB202, 0x2A02, 0x23DE, 0x2252, 0x2252, 0x2252, 0x2452, 0x2802}},
    {35052, 2, {0x2020, 0x1020, 0x1FC, 0xFC20, 0x820, 0x13FE, 0x1088, 0x3544, 0x5A42, 0x94F8, 0x1588, 0x1250, 0x1020, 0x1050, 0x1188, 0x1606}},
    {35053, 2, {0x2020, 0x1020, 0x3E, 0xFC20, 0x9FE, 0x1122, 0x1138, 0x35E0, 0x5922, 0x951E, 0x1500, 0x1178, 0x1248, 0x124A, 0x148A, 0x1906}},
    {35054, 2, {0x20FC, 0x1084, 0x84, 0xFCFC, 0x884, 0x1084, 0x10FC, 0x3400, 0x59FE, 0x9502, 0x1502, 0x11FE, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {35055, 2, {0x2000, 0x11FC, 0x124, 0xFD24, 0x974, 0x1124, 0x11FC, 0x3504, 0x5974, 0x9554, 0x1554, 0x1174, 0x1104, 0x1204, 0x1214, 0x1408}},
    {35056, 2, {0x2000, 0x17BC, 0x84, 0xF294, 0x1108, 0x2294, 0x24A4, 0x6840, 0xB000, 0x2FBC, 0x20A4, 0x22A4, 0x2128, 0x2290, 0x24A8, 0x2846}},
    {35057, 2, {0x2020, 0x1020, 0x3FE, 0xF820, 0x820, 0x11FC, 0x1020, 0x3420, 0x5BFE, 0x9440, 0x14A4, 0x11A8, 0x1290, 0x1488, 0x10C6, 0x1080}},
    {35058, 2, {0x2000, 0x13FE, 0x20, 0xF020, 0x13FE, 0x2222, 0x2222, 0x6B32, 0xB2AA, 0x2AAA, 0x2376, 0x2266, 0x2222, 0x2222, 0x222A, 0x2204}},
    {35059, 2, {0x100, 0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35060, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x640, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35061, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x308, 0x490, 0xC60, 0x3518, 0xC606, 0x400}},
    {35062, 2, {0x2090, 0x1090, 0x90, 0xF79E, 0x1090, 0x2090, 0x2090, 0x6B9C, 0xB090, 0x2890, 0x2090, 0x279E, 0x2090, 0x2090, 0x2090, 0x2090}},
    {35063, 2, {0x2020, 0x1124, 0xA8, 0xF820, 0x9FC, 0x1040, 0x17FE, 0x3888, 0x5504, 0x92FA, 0x1488, 0x1088, 0x10A8, 0x1092, 0x1082, 0x107E}},
    {35064, 2, {0x2000, 0x11FC, 0x124, 0xF924, 0x9FC, 0x1124, 0x1524, 0x39FC, 0x5420, 0x93FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {35065, 2, {0x100, 0xFFFE, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x1310, 0x6588, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35066, 2, {0x2020, 0x1020, 0x3FE, 0xF850, 0x888, 0x1124, 0x1222, 0x31FC, 0x5524, 0x99FC, 0x1524, 0x11FC, 0x1022, 0x1022, 0x101E, 0x1000}},
    {35067, 2, {0x800, 0xE7C, 0x804, 0x7F44, 0x828, 0x2A10, 0x4928, 0x1A44, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35068, 2, {0x2000, 0x11F8, 0x108, 0xF908, 0x9F8, 0x1108, 0x1508, 0x39F8, 0x5480, 0x91FC, 0x1254, 0x1494, 0x1124, 0x1244, 0x1094, 0x1108}},
    {35069, 2, {0x2404, 0x3FA4, 0x4424, 0xFFA4, 0x424, 0x3F84, 0x2494, 0x2588, 0x200, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35070, 2, {0x2000, 0x13FC, 0x204, 0xF204, 0x13FC, 0x2220, 0x2220, 0x6BFE, 0xB220, 0x2A20, 0x22FC, 0x2284, 0x2284, 0x2484, 0x24FC, 0x2884}},
    {35071, 2, {0x2020, 0x1020, 0x1FC, 0xFC50, 0x888, 0x1104, 0x13FE, 0x3408, 0x59E8, 0x9528, 0x1528, 0x11E8, 0x1128, 0x1008, 0x1028, 0x1010}},
    {35072, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0x888, 0x10F8, 0x1088, 0x3488, 0x58F8, 0x9488, 0x1488, 0x13FE, 0x1000, 0x1088, 0x1104, 0x1202}},
    {35073, 2, {0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x920, 0x1110, 0x6308, 0x490, 0xC60, 0x3518, 0xC606, 0x400}},
    {35074, 2, {0x2110, 0x1110, 0x7D0, 0xF110, 0x1110, 0x2118, 0x2FF4, 0x6812, 0xB112, 0x2910, 0x27D0, 0x2110, 0x2110, 0x21D0, 0x2E10, 0x2410}},
    {35075, 2, {0x2020, 0x1120, 0x13C, 0xFD20, 0x920, 0x13FE, 0x1000, 0x35FC, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {35076, 2, {0x2020, 0x1020, 0x3FE, 0xF820, 0x9FC, 0x1024, 0x13FE, 0x3424, 0x59FC, 0x9440, 0x17FE, 0x1084, 0x11C8, 0x1030, 0x10CC, 0x1302}},
    {35077, 2, {0x2040, 0x1020, 0x1FC, 0xF800, 0x888, 0x1050, 0x13FE, 0x3222, 0x5424, 0x99FC, 0x1524, 0x1124, 0x1134, 0x1128, 0x1020, 0x1020}},
    {35078, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x9FC, 0x1104, 0x1104, 0x35FC, 0x5800, 0x97FE, 0x1420, 0x1120, 0x113C, 0x1120, 0x12A0, 0x147E}},
    {35079, 2, {0x2100, 0x1100, 0x1FE, 0xFA00, 0xDFC, 0x1104, 0x11FC, 0x3504, 0x59FC, 0x9480, 0x14FC, 0x1108, 0x1290, 0x1060, 0x1198, 0x1606}},
    {35080, 2, {0x2008, 0x103C, 0x1E0, 0xF820, 0xBFE, 0x1020, 0x11FC, 0x3524, 0x59FC, 0x9524, 0x15FC, 0x1020, 0x11FC, 0x1020, 0x13FE, 0x1000}},
    {35081, 2, {0x2100, 0x113C, 0x7D4, 0xF114, 0x1394, 0x2114, 0x27D4, 0x6924, 0xB14C, 0x2840, 0x27FE, 0x2040, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {35082, 2, {0x2080, 0x1040, 0x3FC, 0xF204, 0x1204, 0x23FC, 0x2200, 0x6A00, 0xB3FC, 0x2B54, 0x2354, 0x25FC, 0x2554, 0x2554, 0x2944, 0x210C}},
    {35083, 2, {0x2048, 0x1148, 0x148, 0xFBFE, 0x948, 0x1148, 0x1578, 0x3900, 0x55FE, 0x9020, 0x13FE, 0x1070, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {35084, 2, {0x2000, 0x13FE, 0x222, 0xF820, 0xBFE, 0x1020, 0x11FC, 0x3524, 0x59FC, 0x9524, 0x15FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {35085, 2, {0x2020, 0x1124, 0x124, 0xFD24, 0x9FC, 0x1000, 0x13FE, 0x3420, 0x5840, 0x95FC, 0x1554, 0x1154, 0x1154, 0x1154, 0x1154, 0x110C}},
    {35086, 2, {0x200, 0x100, 0xFFFE, 0x18, 0xCE0, 0x3020, 0x21FC, 0x3CA8, 0x2124, 0x3C20, 0x308, 0x490, 0xC60, 0x3518, 0xC606, 0x400}},
    {35087, 2, {0x100, 0xFFFE, 0x20, 0xC20, 0x31FC, 0x2124, 0x3DFC, 0x2124, 0x2124, 0x3DFC, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35088, 2, {0x2000, 0x13F8, 0x208, 0xF3F8, 0x1208, 0x23F8, 0x2100, 0x6BFC, 0xB444, 0x2A44, 0x22A4, 0x2204, 0x23F4, 0x2004, 0x2028, 0x2010}},
    {35089, 2, {0x201E, 0x13E0, 0x44, 0xFD24, 0x888, 0x11FC, 0x1040, 0x3440, 0x5BFE, 0x9480, 0x14FC, 0x1144, 0x1128, 0x1210, 0x1068, 0x1186}},
    {35090, 2, {0x100, 0xFFFE, 0x800, 0xBF8, 0x1208, 0x33F8, 0x5040, 0x97FE, 0x1150, 0x1248, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35091, 2, {0x2100, 0x117C, 0x144, 0xF244, 0x1244, 0x267C, 0x2A10, 0x6A10, 0xB2FE, 0x2A38, 0x2254, 0x2254, 0x2294, 0x2312, 0x2210, 0x2210}},
    {35092, 2, {0x2000, 0x13FE, 0x0, 0xF9FC, 0x904, 0x1104, 0x15FC, 0x3800, 0x57FE, 0x9222, 0x1222, 0x13FE, 0x1222, 0x1222, 0x13FE, 0x1202}},
    {35093, 2, {0x2040, 0x10A0, 0x110, 0xF208, 0x15F6, 0x2000, 0x23C4, 0x6A54, 0xB254, 0x2BD4, 0x2254, 0x2254, 0x23D4, 0x2244, 0x2254, 0x22C8}},
    {35094, 2, {0x2080, 0x10FC, 0x104, 0xFDF8, 0x808, 0x13FE, 0x1040, 0x34A2, 0x5B34, 0x9458, 0x1494, 0x1334, 0x1052, 0x1090, 0x1350, 0x1020}},
    {35095, 2, {0x2000, 0x13FE, 0x200, 0xFAFC, 0xA84, 0x12FC, 0x1684, 0x3AFC, 0x5620, 0x93FE, 0x1248, 0x12C8, 0x1230, 0x1248, 0x1284, 0x13FE}},
    {35096, 2, {0x2040, 0x1040, 0x1FC, 0xF884, 0xBFE, 0x1000, 0x15FC, 0x3904, 0x55FC, 0x9020, 0x13FE, 0x1020, 0x1220, 0x13FE, 0x1020, 0x1020}},
    {35097, 2, {0x2050, 0x1052, 0x3DC, 0xF850, 0x8D2, 0x134E, 0x1000, 0x35FC, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x1104, 0x1114, 0x1108}},
    {35098, 2, {0x2020, 0x1022, 0x1FA, 0xF824, 0x828, 0x13FE, 0x1420, 0x3840, 0x54FC, 0x9184, 0x1284, 0x10FC, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {35099, 2, {0x2020, 0x1124, 0xA8, 0xF020, 0x13FE, 0x20A8, 0x2124, 0x6A02, 0xB040, 0x2BFE, 0x2088, 0x2108, 0x2190, 0x2060, 0x2198, 0x2604}},
    {35100, 2, {0x80, 0x79F8, 0x4A08, 0x7DE8, 0x4928, 0x79E8, 0x4914, 0x4904, 0x9AFC, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35101, 2, {0x2242, 0x1124, 0x0, 0xFDFC, 0x924, 0x1124, 0x11FC, 0x3524, 0x5924, 0x95FC, 0x1420, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {35102, 2, {0x2000, 0x11F8, 0x108, 0xF908, 0x9F8, 0x1108, 0x1508, 0x39F8, 0x5400, 0x93FC, 0x1294, 0x1294, 0x1294, 0x1294, 0x17FE, 0x1000}},
    {35103, 2, {0x2000, 0x11FC, 0x104, 0xF9FC, 0x904, 0x11FC, 0x1400, 0x3BDE, 0x5442, 0x9252, 0x114A, 0x10C6, 0x135A, 0x1042, 0x114A, 0x1084}},
    {35104, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0x9FC, 0x1088, 0x13FE, 0x3420, 0x59FC, 0x9524, 0x15FC, 0x1124, 0x13FE, 0x1104, 0x1114, 0x1108}},
    {35105, 2, {0x2088, 0x1088, 0x3FE, 0xF888, 0x820, 0x1050, 0x1488, 0x3904, 0x56FA, 0x9000, 0x1000, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {35106, 2, {0x100, 0xFFFE, 0x200, 0x1FF0, 0x1110, 0x1FF0, 0x1210, 0x1FF0, 0x4A8, 0x8BA, 0x3382, 0xC4FE, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35107, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0x850, 0x1088, 0x1524, 0x3850, 0x5488, 0x9104, 0x12FA, 0x1088, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {35108, 2, {0x2020, 0x1020, 0x1FC, 0xF820, 0xBFE, 0x1000, 0x11FC, 0x3504, 0x5904, 0x95FC, 0x1452, 0x1094, 0x1188, 0x16A4, 0x10C2, 0x1080}},
    {35109, 2, {0x2000, 0x13FE, 0x200, 0xF2FC, 0x1200, 0x23FE, 0x22A8, 0x6A90, 0xB2C8, 0x2A86, 0x2208, 0x23FE, 0x2288, 0x2448, 0x2408, 0x2818}},
    {35110, 2, {0x2110, 0x1212, 0x454, 0xF7D8, 0x1010, 0x2012, 0x27D2, 0x6C4E, 0xB440, 0x2FD2, 0x2454, 0x2458, 0x27D0, 0x2452, 0x2552, 0x248E}},
    {35111, 2, {0x7E10, 0x2452, 0x3C54, 0x2490, 0x3C28, 0x2744, 0xFC82, 0x600, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35112, 2, {0x2108, 0x1088, 0x90, 0xFBFC, 0x840, 0x11F8, 0x1440, 0x3BFE, 0x5480, 0x9100, 0x11FC, 0x1220, 0x1420, 0x1820, 0x13FE, 0x1000}},
    {35113, 2, {0x1000, 0x3E78, 0x2248, 0x2A48, 0xFE86, 0x2378, 0x2A48, 0x4230, 0x4A48, 0x8684, 0x100, 0xFFFE, 0x680, 0x1C48, 0xE530, 0x60E}},
    {35114, 2, {0x2000, 0x127C, 0x144, 0xF944, 0x87C, 0x1044, 0x1744, 0x357C, 0x5950, 0x954A, 0x1544, 0x1154, 0x1162, 0x1142, 0x1280, 0x147E}},
    {35115, 2, {0x2004, 0x101E, 0x3F0, 0xF21E, 0x1210, 0x22FE, 0x2292, 0x6A98, 0xB2F2, 0x2A8E, 0x2280, 0x22B8, 0x22A8, 0x24AA, 0x254A, 0x2A86}},
    {35116, 2, {0x21F8, 0x1088, 0x70, 0xF18C, 0x1000, 0x23DE, 0x2252, 0x698C, 0xB252, 0x2820, 0x23FE, 0x2070, 0x20A8, 0x2124, 0x2622, 0x2020}},
    {35117, 2, {0x100, 0xFFFE, 0x0, 0x3FFC, 0x2100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35118, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x8204, 0x100, 0xFFFE, 0x500, 0xC88, 0x3450, 0xC530, 0x60E, 0x400}},
    {35119, 2, {0x2040, 0x1020, 0x7FE, 0xF488, 0x1488, 0x27FE, 0x2488, 0x6CF8, 0xB420, 0x2DFC, 0x2524, 0x2524, 0x2934, 0x2928, 0x3020, 0x2020}},
    {35120, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0xA20, 0x1110, 0x2FE8, 0xC28E, 0x450, 0x1C20, 0x6518, 0x606}},
    {35121, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x2448, 0x3FF8, 0x1110, 0x920, 0x1110, 0x2308, 0x490, 0xC60, 0x3518, 0xC606, 0x400}},
    {35122, 2, {0x2020, 0x1010, 0x1FE, 0xFD10, 0x97C, 0x1110, 0x117C, 0x3554, 0x597C, 0x9554, 0x157C, 0x1110, 0x11FE, 0x1210, 0x1210, 0x1410}},
    {35123, 2, {0x2010, 0x1010, 0x2FE, 0xF910, 0x97C, 0x1054, 0x147C, 0x3B54, 0x557C, 0x9110, 0x11FE, 0x1110, 0x1110, 0x1290, 0x147E, 0x1000}},
    {35124, 2, {0x2048, 0x1148, 0x14E, 0xF950, 0x968, 0x1144, 0x1144, 0x3440, 0x5800, 0x95FC, 0x1554, 0x1154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {35125, 2, {0x2040, 0x1020, 0x3FE, 0xF800, 0x954, 0x1124, 0x1554, 0x39FC, 0x5420, 0x93FE, 0x1242, 0x1292, 0x12FA, 0x120A, 0x1202, 0x1206}},
    {35126, 2, {0x2000, 0x13DE, 0x42, 0xFA52, 0x94A, 0x1252, 0x1042, 0x3420, 0x59FC, 0x9504, 0x1504, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {35127, 2, {0x2210, 0x1210, 0x450, 0xF95C, 0x1150, 0x2250, 0x26FE, 0x6A00, 0xB210, 0x2A50, 0x225C, 0x2250, 0x22B0, 0x2290, 0x230E, 0x2200}},
    {35128, 2, {0x2020, 0x11FC, 0x124, 0xFBFE, 0x924, 0x11FC, 0x1420, 0x39FC, 0x5524, 0x91FC, 0x1040, 0x13FE, 0x1088, 0x11D0, 0x1070, 0x138C}},
    {35129, 2, {0x2110, 0x1110, 0x7D0, 0xF110, 0x113C, 0x27D4, 0x2014, 0x6AB4, 0xB554, 0x291C, 0x27D4, 0x2114, 0x2114, 0x21E4, 0x2726, 0x2242}},
    {35130, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0xA82, 0x100, 0xFFFE, 0x488, 0xC50, 0x3420, 0xC518, 0x606}},
    {35131, 2, {0x100, 0xFFFE, 0x820, 0x3E20, 0x8F8, 0x7F28, 0x2A68, 0x5D2A, 0x856, 0x3E82, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35132, 2, {0x4000, 0x23FE, 0x2050, 0x3FE, 0xF252, 0x13FE, 0x2020, 0x6FFE, 0xB088, 0x2904, 0x23FA, 0x2508, 0x21F8, 0x2102, 0x2102, 0x20FE}},
    {35133, 2, {0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5FA8, 0x4408, 0x9528, 0x2E90, 0x100, 0xFFFE, 0x680, 0x1C48, 0xE530, 0x60E}},
    {35134, 2, {0x2000, 0x13FE, 0x50, 0xFDFC, 0x954, 0x1154, 0x11FC, 0x3400, 0x59FC, 0x9400, 0x17FE, 0x1020, 0x10A8, 0x1124, 0x12A2, 0x1040}},
    {35135, 2, {0x2090, 0x1090, 0x7FE, 0xF890, 0xBFC, 0x1294, 0x13FC, 0x3694, 0x5BFC, 0x9400, 0x15F8, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x1108}},
    {35136, 2, {0x2020, 0x13FE, 0x20, 0xFDFC, 0x820, 0x13FE, 0x1000, 0x35FC, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {35137, 2, {0x2000, 0x13BE, 0xA2, 0xF0A2, 0x10BE, 0x2388, 0x2208, 0x6A3E, 0xB22A, 0x2BAA, 0x20AA, 0x20BE, 0x2088, 0x208A, 0x257E, 0x2202}},
    {35138, 2, {0x2040, 0x1088, 0x1FC, 0xF908, 0xA52, 0x13FE, 0x1450, 0x3988, 0x5626, 0x90C0, 0x1310, 0x1064, 0x1388, 0x1030, 0x10C0, 0x1700}},
    {35139, 2, {0x100, 0xFFFE, 0x0, 0xCF8, 0x3088, 0x20F8, 0x3C20, 0x21FC, 0x20A8, 0x3D24, 0x300, 0x488, 0xC50, 0x3530, 0xC60E, 0x400}},
    {35140, 2, {0x100, 0xFFFE, 0x0, 0x3EF8, 0x2288, 0x3EF8, 0x440, 0x7FFC, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x444, 0x1C28, 0xE518, 0x606}},
    {35141, 2, {0x2000, 0x11FC, 0x124, 0xFDFC, 0x924, 0x11FC, 0x10A8, 0x34A8, 0x5BFE, 0x94A8, 0x14A8, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {35142, 2, {0x2050, 0x1252, 0x154, 0xF850, 0xBFE, 0x1088, 0x1450, 0x3BFE, 0x5420, 0x91FC, 0x1020, 0x13FE, 0x1050, 0x1088, 0x1104, 0x1602}},
    {35143, 2, {0x2000, 0x13DE, 0x252, 0xFBDE, 0xA52, 0x13DE, 0x1602, 0x3AFA, 0x568A, 0x92FA, 0x128A, 0x12FA, 0x128A, 0x1202, 0x120A, 0x1204}},
    {35144, 2, {0x2000, 0x13DE, 0x252, 0xFBDE, 0xA10, 0x1252, 0x11CE, 0x3400, 0x5888, 0x97FE, 0x1488, 0x1088, 0x17FE, 0x1088, 0x1104, 0x1202}},
    {35145, 2, {0x279E, 0x1492, 0x79E, 0xF492, 0x179E, 0x2402, 0x24F2, 0x6C92, 0xB4F2, 0x2C92, 0x24F2, 0x2492, 0x2492, 0x2532, 0x2402, 0x2406}},
    {35146, 2, {0x2000, 0x13F8, 0x208, 0xF3F8, 0x1208, 0x23F8, 0x2000, 0x6FFE, 0xB480, 0x2FBC, 0x2494, 0x2794, 0x24D4, 0x2F88, 0x2094, 0x20A2}},
    {35147, 2, {0x4208, 0x2208, 0xFBE, 0xE208, 0x2208, 0x2FBE, 0x4AAA, 0x5AAA, 0xEAAA, 0x5BAE, 0x4208, 0x471C, 0x4AAA, 0x5248, 0x4208, 0x4208}},
    {35148, 2, {0x2000, 0x13DE, 0x252, 0xFA52, 0xBDE, 0x1000, 0x11FC, 0x3524, 0x59FC, 0x9524, 0x15FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {35149, 2, {0x20A0, 0x1090, 0x1FE, 0xFB20, 0x9FC, 0x1120, 0x11FC, 0x3520, 0x59FE, 0x9500, 0x1420, 0x13FE, 0x1070, 0x10A8, 0x1326, 0x1020}},
    {35150, 2, {0x203C, 0x13E0, 0x124, 0xF8A8, 0xBFE, 0x10A8, 0x1124, 0x3602, 0x59FC, 0x9524, 0x1524, 0x11FC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {35151, 2, {0x2024, 0x13A8, 0x92, 0xF514, 0x1208, 0x2404, 0x23BA, 0x68A8, 0xB0C6, 0x2B80, 0x223C, 0x2384, 0x20A8, 0x2090, 0x22A8, 0x2144}},
    {35152, 2, {0x2080, 0x10F8, 0x108, 0xFBFE, 0xD12, 0x1122, 0x15FE, 0x3840, 0x54A2, 0x9354, 0x1098, 0x1334, 0x1054, 0x1092, 0x1350, 0x1020}},
    {35153, 2, {0x2000, 0x11FC, 0x4, 0xFCFC, 0x804, 0x11FC, 0x1000, 0x35DC, 0x5894, 0x97DC, 0x1408, 0x13FE, 0x1108, 0x1088, 0x10A8, 0x1010}},
    {35154, 2, {0x2108, 0x1108, 0x548, 0xF390, 0x111E, 0x27D4, 0x2564, 0x6D54, 0xB554, 0x2F54, 0x25D4, 0x2548, 0x2548, 0x2554, 0x2464, 0x24C2}},
    {35155, 2, {0x2020, 0x1020, 0x1FC, 0xF820, 0xBFE, 0x1108, 0x139C, 0x3508, 0x5988, 0x963E, 0x1400, 0x13FE, 0x1090, 0x1090, 0x1112, 0x120E}},
    {35156, 2, {0x2088, 0x1088, 0x3FE, 0xF888, 0x888, 0x10F8, 0x1420, 0x3BFE, 0x5622, 0x9332, 0x12AA, 0x1376, 0x1222, 0x1222, 0x122A, 0x1224}},
    {35157, 2, {0x2100, 0x10BE, 0x282, 0xF222, 0x12FA, 0x2222, 0x22FA, 0x6AAA, 0xB2BA, 0x2AEA, 0x22AA, 0x22FA, 0x2272, 0x22AA, 0x2222, 0x2226}},
    {35158, 2, {0x2080, 0x1100, 0x7FC, 0xF444, 0x1554, 0x2444, 0x27FC, 0x6CE4, 0xB554, 0x2C04, 0x2040, 0x2FFE, 0x20A0, 0x2110, 0x2208, 0x2C06}},
    {35159, 2, {0x2000, 0x11FC, 0x154, 0xF9FC, 0x820, 0x11FC, 0x1420, 0x3BFE, 0x5488, 0x9050, 0x11FC, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {35160, 2, {0x2020, 0x1050, 0x88, 0xFD74, 0xA02, 0x11FC, 0x1124, 0x35AC, 0x5924, 0x95FC, 0x1400, 0x10F8, 0x1088, 0x10F8, 0x1088, 0x10F8}},
    {35161, 2, {0x20F8, 0x1088, 0x88, 0xFCF8, 0x800, 0x11DC, 0x1154, 0x3554, 0x59DC, 0x9420, 0x15FE, 0x1070, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {35162, 2, {0x2044, 0x1224, 0x128, 0xFC7E, 0x810, 0x1028, 0x134A, 0x351C, 0x5928, 0x954C, 0x151A, 0x112A, 0x1148, 0x1110, 0x12FE, 0x1400}},
    {35163, 2, {0x2090, 0x13FC, 0x294, 0xFBFC, 0xA94, 0x13FC, 0x1400, 0x3BFC, 0x5600, 0x92F8, 0x1200, 0x13FE, 0x1520, 0x1514, 0x1548, 0x1986}},
    {35164, 2, {0x2080, 0x10F8, 0x108, 0xFBFE, 0x944, 0x1192, 0x11FE, 0x3500, 0x597C, 0x9500, 0x157C, 0x1100, 0x117C, 0x1244, 0x127C, 0x1444}},
    {35165, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x904, 0x12FA, 0x1000, 0x3400, 0x59DC, 0x9554, 0x1554, 0x11DC, 0x1088, 0x1088, 0x1154, 0x1222}},
    {35166, 2, {0x40, 0x7C20, 0x45FC, 0x4488, 0x7C50, 0x43FE, 0x7C20, 0xA5FC, 0x2420, 0x3E20, 0x100, 0xFFFE, 0x680, 0x1C48, 0xE530, 0x60E}},
    {35167, 2, {0x2110, 0x1110, 0x7BC, 0xF110, 0x13B8, 0x2554, 0x2912, 0x6800, 0xB3F8, 0x2800, 0x27FC, 0x2040, 0x2248, 0x2444, 0x2942, 0x2080}},
    {35168, 2, {0x2020, 0x1124, 0xA8, 0xFBFE, 0xA02, 0x10F8, 0x1088, 0x3488, 0x58F8, 0x9400, 0x15FC, 0x1124, 0x11FC, 0x1124, 0x11FC, 0x1104}},
    {35169, 2, {0x2000, 0x13F8, 0x2A8, 0xF2A8, 0x13F8, 0x2100, 0x23FC, 0x6C84, 0xB3E4, 0x2AA4, 0x22A4, 0x23E4, 0x2094, 0x27F4, 0x2014, 0x2008}},
    {35170, 2, {0x2020, 0x13FE, 0x0, 0xFBFC, 0xA04, 0x12F4, 0x1294, 0x37FC, 0x5800, 0x95F8, 0x1508, 0x11F8, 0x1108, 0x11F8, 0x1000, 0x13FE}},
    {35171, 2, {0x2040, 0x10F8, 0x88, 0xFCF8, 0x888, 0x10F8, 0x1000, 0x35FC, 0x5924, 0x95FC, 0x1524, 0x11FC, 0x1000, 0x13FE, 0x1088, 0x1108}},
    {35172, 2, {0x2008, 0x13E8, 0x288, 0xFBEE, 0xA28, 0x13F4, 0x1682, 0x3BE2, 0x5400, 0x91FC, 0x1154, 0x1154, 0x1154, 0x1154, 0x17FE, 0x1000}},
    {35173, 2, {0x2128, 0x11AA, 0x16C, 0xF228, 0x12FE, 0x2644, 0x2628, 0x6AFE, 0xB210, 0x2A7C, 0x2210, 0x22FE, 0x2210, 0x2228, 0x2244, 0x2282}},
    {35174, 2, {0x2000, 0x11FC, 0x20, 0xFBFE, 0xA22, 0x11AC, 0x1420, 0x39AC, 0x5400, 0x93FE, 0x1020, 0x11FC, 0x1154, 0x1154, 0x1154, 0x110C}},
    {35175, 2, {0x2000, 0x17FC, 0x248, 0xF444, 0x1FFE, 0x2444, 0x26EC, 0x6D54, 0xB6EC, 0x2C44, 0x26EC, 0x2554, 0x26EC, 0x2444, 0x2454, 0x2408}},
    {35176, 2, {0x2144, 0x1144, 0x554, 0xF364, 0x115E, 0x27E4, 0x2004, 0x6A34, 0xB14C, 0x2BE4, 0x2084, 0x23E4, 0x2084, 0x20E4, 0x2714, 0x2008}},
    {35177, 2, {0x2020, 0x13FE, 0x20, 0xFDFC, 0x800, 0x13FE, 0x1252, 0x37FE, 0x5904, 0x95FC, 0x1504, 0x11FC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {35178, 2, {0x2108, 0x17FE, 0x108, 0xF000, 0x13FC, 0x2294, 0x2294, 0x6BFC, 0xB028, 0x2BFE, 0x2220, 0x2324, 0x22A8, 0x2292, 0x242A, 0x28C6}},
    {35179, 2, {0x2020, 0x1020, 0x7FE, 0xF020, 0x13AE, 0x2124, 0x23AE, 0x6AAA, 0xB3AE, 0x2AAA, 0x23AE, 0x2050, 0x2050, 0x2088, 0x2104, 0x2602}},
    {35180, 2, {0x2000, 0x13FC, 0x294, 0xF294, 0x13FC, 0x2110, 0x2252, 0x6BDC, 0xB012, 0x2BCE, 0x2240, 0x23D2, 0x225C, 0x23D0, 0x2252, 0x22CE}},
    {35181, 2, {0x2100, 0x113E, 0x108, 0xF7D0, 0x113E, 0x2122, 0x23A2, 0x683E, 0xB022, 0x2BBE, 0x22A2, 0x22A2, 0x22BE, 0x2380, 0x2294, 0x2022}},
    {35182, 2, {0x21FC, 0x1104, 0x1FC, 0xFD04, 0x9FC, 0x1088, 0x11FC, 0x3488, 0x5BFE, 0x9488, 0x1524, 0x12AA, 0x1070, 0x10A8, 0x1124, 0x1060}},
    {35183, 2, {0x4200, 0x213C, 0x27A4, 0x24, 0xF4BC, 0x1324, 0x27A4, 0x693C, 0xB124, 0x2FA4, 0x213C, 0x2590, 0x2558, 0x2968, 0x252A, 0x2246}},
    {35184, 2, {0x2108, 0x1108, 0x7CE, 0xF112, 0x1124, 0x27DE, 0x2552, 0x6D5E, 0xB7D2, 0x291E, 0x2392, 0x255E, 0x2940, 0x2114, 0x2112, 0x2122}},
    {35185, 2, {0x2210, 0x1110, 0x7DE, 0xF010, 0x145E, 0x2282, 0x27DE, 0x6810, 0xB7DE, 0x2C50, 0x27DE, 0x2450, 0x27DE, 0x2450, 0x2452, 0x24CE}},
    {35186, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E, 0x100, 0xFFFE, 0x680, 0x1C48, 0xE530, 0x60E}},
    {35187, 2, {0x4148, 0x214C, 0x22AA, 0x8, 0xF7FE, 0x1148, 0x2368, 0x694A, 0xB36A, 0x294C, 0x236C, 0x2148, 0x216A, 0x239A, 0x2026, 0x2042}},
    {35188, 2, {0x27BC, 0x14A4, 0x7BC, 0xF4A4, 0x17BC, 0x2444, 0x25F4, 0x6C44, 0xB5F4, 0x2D54, 0x25F4, 0x2554, 0x25F4, 0x24E4, 0x2554, 0x244C}},
    {35189, 2, {0x23FC, 0x1108, 0x1F8, 0xF108, 0x11F8, 0x210E, 0x27F8, 0x6808, 0xB7FE, 0x2A94, 0x239C, 0x2294, 0x239C, 0x22D6, 0x27BC, 0x2084}},
    {35190, 2, {0x4114, 0x27D2, 0x2110, 0xFFE, 0xF010, 0x17D0, 0x2552, 0x6FD2, 0xB552, 0x2FD4, 0x2294, 0x27C8, 0x228A, 0x2FFA, 0x22A6, 0x2442}},
    {35191, 2, {0x2140, 0x126C, 0x244, 0xF36C, 0x12A4, 0x23AC, 0x22A4, 0x6FFE, 0xB108, 0x2AF4, 0x2442, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x20C0}},
    {35192, 2, {0x2528, 0x17BE, 0x948, 0xF7BE, 0x1318, 0x25AA, 0x2946, 0x6BFC, 0xB204, 0x2BFC, 0x2204, 0x23FC, 0x2204, 0x23FC, 0x2108, 0x2204}},
    {35193, 2, {0x27BE, 0x1000, 0x7BC, 0xF4A4, 0x16B4, 0x24A4, 0x2040, 0x6FFE, 0xB4A0, 0x2FFC, 0x24A4, 0x27FC, 0x2512, 0x29DC, 0x2912, 0x31CE}},
    {35194, 2, {0x2110, 0x17FE, 0x110, 0xF000, 0x17FE, 0x254A, 0x254A, 0x6EFE, 0xB7EA, 0x2D6A, 0x267E, 0x27CA, 0x244E, 0x2572, 0x2542, 0x2406}},
    {35195, 2, {0x22A8, 0x1248, 0x7BE, 0xF2A8, 0x175C, 0x2AAA, 0x2288, 0x6FFE, 0xB218, 0x2DE4, 0x2842, 0x23F8, 0x2040, 0x27FC, 0x2040, 0x20C0}},
    {35196, 2, {0x2108, 0x17FE, 0x108, 0xF390, 0x113C, 0x27D4, 0x22B4, 0x6D5C, 0xB396, 0x2922, 0x23FC, 0x2000, 0x27FE, 0x2080, 0x2104, 0x23FE}},
    {35197, 2, {0x2110, 0x17FE, 0x110, 0xF7BC, 0x14A4, 0x27BC, 0x24A4, 0x6FBC, 0xB444, 0x2FFC, 0x2554, 0x25F4, 0x2444, 0x24E4, 0x2554, 0x2448}},
    {35198, 2, {0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x7FFC, 0x4444, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4004, 0x4004, 0x4004, 0x4004}},
    {35199, 2, {0x0, 0xFFFE, 0x440, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x2448, 0x2448, 0x2838, 0x3008, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {35200, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {35201, 2, {0x0, 0x7FFC, 0x440, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x200, 0x7FFC, 0x420, 0x840, 0x1C80, 0x380, 0x670, 0x3808}},
    {35202, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1F0, 0x3E00, 0x100, 0x7FF8, 0x60, 0x180, 0x600, 0x1800, 0xE7FE, 0x0}},
    {35203, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {35204, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x850, 0x1048, 0x37FE, 0x5040, 0x90A0, 0x10A0, 0x1110, 0x1208, 0x1406}},
    {35205, 2, {0x40, 0xFE40, 0x2840, 0x287E, 0xFEAA, 0xAB2A, 0xAA2A, 0xFE4A, 0x104A, 0xFE92, 0x2512, 0x4422, 0x2822, 0x1042, 0x2894, 0xC408}},
    {35206, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x1200, 0x27FC, 0x4208, 0x8BF8, 0x1208, 0x33F8, 0x5100, 0x93F8, 0x1510, 0x10E0, 0x171E}},
    {35207, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2200, 0x7F7C, 0x2244, 0x3E44, 0x87C, 0x7F44, 0x4944, 0x7F7C, 0x844, 0xFF54, 0x888}},
    {35208, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x1020, 0x7E20, 0x523E, 0x4A44, 0x7EA4, 0x1028, 0xFF28, 0x2010, 0x3E28, 0x4244, 0x8682}},
    {35209, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2820, 0x7DFC, 0x2850, 0x3888, 0x13FE, 0x7C08, 0x55E8, 0x7D28, 0x11E8, 0xFE08, 0x1018}},
    {35210, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2800, 0xFDFE, 0x2920, 0x39FC, 0x1120, 0x7DFC, 0x5520, 0x7DFE, 0x1002, 0xFEAA, 0x1004}},
    {35211, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {35212, 2, {0x0, 0x1F8, 0xFD08, 0x508, 0x5F8, 0x4908, 0x2908, 0x11F8, 0x1108, 0x2908, 0x25F8, 0x4490, 0x8090, 0x112, 0x212, 0xC0E}},
    {35213, 2, {0x100, 0x920, 0x1110, 0x2108, 0x4104, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x842, 0x1042, 0x603E}},
    {35214, 2, {0x0, 0x7E7C, 0x4204, 0x4208, 0x7E10, 0x4210, 0x42FE, 0x7E10, 0x4210, 0x4210, 0x7E10, 0x2450, 0x2422, 0x2402, 0x43FE, 0x8000}},
    {35215, 2, {0x1000, 0x10FC, 0x1084, 0x7C84, 0x10FC, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x1084, 0x10FC, 0x2850, 0x2450, 0x4492, 0x8092, 0x10E}},
    {35216, 2, {0x0, 0x7E88, 0x4250, 0x4220, 0x7E50, 0x4288, 0x4200, 0x7E88, 0x4250, 0x4220, 0x7E50, 0x2488, 0x2402, 0x2402, 0x43FE, 0x8000}},
    {35217, 2, {0x800, 0x48FC, 0x4884, 0x4884, 0x48FC, 0x7E84, 0x4084, 0x40FC, 0x4084, 0x7C84, 0x44FC, 0x4450, 0x4450, 0x4492, 0x4492, 0x850E}},
    {35218, 2, {0x400, 0xEFC, 0x7084, 0x1084, 0x10FC, 0x1C84, 0x7084, 0x10FC, 0x1084, 0x1E84, 0xF0FC, 0x1050, 0x1450, 0x1892, 0x1092, 0x10E}},
    {35219, 2, {0x8, 0x1FC, 0x7E10, 0x2210, 0x1120, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {35220, 2, {0x0, 0x7FFC, 0x100, 0x360, 0xD18, 0x7104, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {35221, 2, {0x1000, 0x87C, 0xA44, 0x244, 0x127C, 0x1344, 0x54C4, 0x54FC, 0x5444, 0x9844, 0x197C, 0x1128, 0x3128, 0x4F4A, 0x804A, 0x86}},
    {35222, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x9FC, 0x1104, 0x1104, 0x39FC, 0x5504, 0x9504, 0x11FC, 0x1050, 0x1090, 0x1092, 0x1112, 0x120E}},
    {35223, 2, {0x0, 0x7F7C, 0x144, 0x144, 0x7D7C, 0x144, 0x144, 0x7D7C, 0x4544, 0x4544, 0x7D7C, 0x4528, 0x128, 0x14A, 0xA4A, 0x486}},
    {35224, 2, {0x1000, 0x10FC, 0x1084, 0x1E84, 0x10FC, 0x1084, 0x1084, 0x10FC, 0x7E84, 0x4284, 0x42FC, 0x4250, 0x4250, 0x7E92, 0x4292, 0x10E}},
    {35225, 2, {0x2000, 0x20FC, 0x2084, 0x7E84, 0x42FC, 0x4484, 0x9084, 0x10FC, 0x5484, 0x5484, 0x52FC, 0x5250, 0x9250, 0x1092, 0x5092, 0x210E}},
    {35226, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x842, 0x1042, 0x603E}},
    {35227, 2, {0x200, 0x77C, 0x3844, 0x2044, 0x237C, 0x2C44, 0x2844, 0x28FC, 0x2D44, 0x2A44, 0x2A7C, 0x2928, 0x2928, 0x494A, 0x4C4A, 0x8886}},
    {35228, 2, {0x2800, 0x28FC, 0x2884, 0x2884, 0xAAFC, 0x6C84, 0x2884, 0x28FC, 0x6C84, 0xAA84, 0x28FC, 0x2850, 0x2A50, 0x4C92, 0x4892, 0x810E}},
    {35229, 2, {0x0, 0x7CFC, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x1084, 0xFC, 0x1084, 0x5284, 0x52FC, 0x9450, 0x2850, 0x2492, 0x4492, 0x810E}},
    {35230, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x2850, 0x2A50, 0x2C92, 0x4892, 0x810E}},
    {35231, 2, {0x800, 0x87C, 0x844, 0xFF44, 0x87C, 0x844, 0x844, 0x7E7C, 0x1044, 0x844, 0x2A7C, 0xA128, 0xA128, 0xA54A, 0x1C4A, 0x86}},
    {35232, 2, {0x0, 0x7E7C, 0x1244, 0x1244, 0xFF7C, 0x1244, 0x1244, 0x7E7C, 0x2044, 0x2044, 0x7E7C, 0x6228, 0xA228, 0x224A, 0x3E4A, 0x2286}},
    {35233, 2, {0x0, 0x7C, 0x7F44, 0x844, 0x87C, 0x2A44, 0x2A44, 0x2A7C, 0x5D44, 0x4944, 0x887C, 0x828, 0xF28, 0xF04A, 0x404A, 0x86}},
    {35234, 2, {0x1000, 0x52FC, 0x5484, 0x9084, 0x28FC, 0x4484, 0x8284, 0x10FC, 0x1084, 0x5284, 0x54FC, 0x9050, 0x2850, 0x2492, 0x4292, 0x810E}},
    {35235, 2, {0xC00, 0x70FC, 0x1084, 0xFE84, 0x38FC, 0x5484, 0x9284, 0xFC, 0x1084, 0xFC84, 0x24FC, 0x4450, 0x2850, 0x1092, 0x2892, 0xC50E}},
    {35236, 2, {0x1000, 0x1EFC, 0x1084, 0x7E84, 0x52FC, 0x5C84, 0x7084, 0x52FC, 0x4E84, 0x4084, 0x5CFC, 0x5450, 0x5450, 0x9692, 0x9492, 0x210E}},
    {35237, 2, {0x1400, 0x147C, 0x1444, 0x7F44, 0x557C, 0x5544, 0x5544, 0x7F7C, 0x5544, 0x5544, 0x557C, 0xFFA8, 0x2428, 0x224A, 0x424A, 0x8086}},
    {35238, 2, {0x1000, 0x287C, 0x4444, 0x8244, 0x7D7C, 0x44, 0xF144, 0x957C, 0x9544, 0xF544, 0x957C, 0x9528, 0xF528, 0x914A, 0x954A, 0xB286}},
    {35239, 2, {0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3FF8, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {35240, 2, {0x0, 0x777C, 0x5544, 0x5544, 0x777C, 0x44, 0x7F44, 0x7C, 0xFFC4, 0x2044, 0x3F7C, 0x128, 0x128, 0x14A, 0xA4A, 0x486}},
    {35241, 2, {0x1000, 0x107C, 0x7D44, 0x1244, 0x147C, 0xFF44, 0x1044, 0x207C, 0x7E44, 0xA244, 0x227C, 0x3E28, 0x2228, 0x224A, 0x3E4A, 0x2286}},
    {35242, 2, {0x1000, 0x87C, 0x7F44, 0x44, 0x227C, 0x1444, 0xFF44, 0x87C, 0x844, 0x7F44, 0x87C, 0x2A28, 0x4928, 0x88CA, 0x284A, 0x1086}},
    {35243, 2, {0x1000, 0x87C, 0x7F44, 0x44, 0x227C, 0x1444, 0x7F44, 0x517C, 0x844, 0x7F44, 0x107C, 0x1E28, 0x1228, 0x224A, 0x2A4A, 0x4486}},
    {35244, 2, {0x800, 0x497C, 0x4944, 0x7F44, 0x7C, 0xFF44, 0x44, 0x7F7C, 0x4144, 0x4144, 0x7F7C, 0x4228, 0x2428, 0xF4A, 0xF04A, 0x4086}},
    {35245, 2, {0x0, 0xFF7C, 0x8144, 0x44, 0x7E7C, 0x4244, 0x7E44, 0x427C, 0x7E44, 0x1044, 0x87C, 0xFF28, 0x28, 0x244A, 0x424A, 0x8186}},
    {35246, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x842, 0xF03E}},
    {35247, 2, {0x2200, 0x227C, 0xFFC4, 0x2244, 0x7F7C, 0x2244, 0xFFC4, 0x87C, 0x7F44, 0x4944, 0x7F7C, 0x4928, 0xFFE8, 0x414A, 0x454A, 0x4286}},
    {35248, 2, {0x1000, 0x1E7C, 0x1044, 0x7F44, 0x517C, 0x5C44, 0x7244, 0x4E7C, 0x4044, 0x5E44, 0x527C, 0x5E28, 0x5228, 0x5E4A, 0x524A, 0xBF86}},
    {35249, 2, {0x804, 0x7F78, 0x840, 0x3E40, 0x2A7E, 0x3E48, 0x2A48, 0xFF48, 0x888, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x783E}},
    {35250, 2, {0x2200, 0x227C, 0xFFC4, 0x2244, 0x3E7C, 0x844, 0x7F44, 0x497C, 0x7F44, 0x844, 0x7F7C, 0x828, 0x7F28, 0x84A, 0xF4A, 0xF086}},
    {35251, 2, {0x2000, 0x207C, 0x3F44, 0x4844, 0x887C, 0x7F44, 0x1444, 0x227C, 0x7F44, 0xA444, 0x3F7C, 0x2428, 0x3F28, 0x244A, 0x3F4A, 0x2086}},
    {35252, 2, {0xA00, 0x747C, 0x1544, 0x5244, 0x227C, 0x2144, 0x5D44, 0x807C, 0x7E44, 0x4244, 0x427C, 0x7E28, 0x4228, 0x244A, 0xF4A, 0xF086}},
    {35253, 2, {0xE700, 0xA57C, 0xE744, 0xA544, 0xE77C, 0x8144, 0xBD44, 0xA57C, 0xBD44, 0xA544, 0xBD7C, 0xA528, 0xA528, 0xCD4A, 0x814A, 0x8386}},
    {35254, 2, {0xF00, 0xF07C, 0x4944, 0x2244, 0xFE7C, 0x1444, 0x844, 0xFF7C, 0x9544, 0xBD44, 0x837C, 0xBD28, 0x9528, 0x894A, 0x954A, 0xA386}},
    {35255, 2, {0x1000, 0x1E7C, 0x1044, 0x7F44, 0x517C, 0x5C44, 0x7244, 0x4E7C, 0x4044, 0x4A44, 0x6AFC, 0x5B28, 0x4A28, 0x8F4A, 0x784A, 0x86}},
    {35256, 2, {0x0, 0x777C, 0x5544, 0x7744, 0x557C, 0x7744, 0x4144, 0x5D7C, 0x5544, 0x5D44, 0x557C, 0x5D28, 0x5528, 0x414A, 0x454A, 0x4286}},
    {35257, 2, {0x144, 0xF954, 0x8B54, 0x8DF8, 0xF80E, 0x8912, 0x8AEA, 0xFE0A, 0x8A0A, 0x8AEA, 0xFAAA, 0x52A4, 0x52B4, 0x56AA, 0x5B0A, 0x9210}},
    {35258, 2, {0x3AB8, 0x2108, 0x3AB8, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x5012, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {35259, 2, {0x800, 0xF7C, 0x844, 0x7F44, 0x497C, 0x4C44, 0x7944, 0x477C, 0x4044, 0x5F44, 0x407C, 0x5F28, 0x5128, 0x5F4A, 0x8A4A, 0x3F86}},
    {35260, 2, {0x0, 0xFF7C, 0x5244, 0x9144, 0xFFFC, 0x9144, 0xDB44, 0xB57C, 0xDB44, 0x9144, 0xDB7C, 0xB528, 0xDB28, 0x914A, 0x954A, 0x8286}},
    {35261, 2, {0x7E40, 0x487E, 0x7E90, 0x4308, 0x7EFE, 0x48AA, 0x7FFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x442, 0x1842, 0xE03E}},
    {35262, 2, {0x1000, 0x87C, 0x7F44, 0x4144, 0x867C, 0x7844, 0x4944, 0x2A7C, 0xFF44, 0x2A44, 0x417C, 0xBE28, 0x2A28, 0x3E4A, 0x2A4A, 0x3E86}},
    {35263, 2, {0x1000, 0xFEFC, 0x1084, 0x7C84, 0xFC, 0xFE84, 0xAA84, 0xFEFC, 0x4484, 0x7C84, 0x44FC, 0x7C50, 0x4450, 0x7C92, 0x2892, 0x450E}},
    {35264, 2, {0x2200, 0x227C, 0xFFC4, 0x2244, 0x777C, 0x5544, 0x7744, 0x127C, 0x3F44, 0x6444, 0xBF7C, 0x2428, 0x3F28, 0x244A, 0x3F4A, 0x2086}},
    {35265, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1290, 0x1290, 0x480, 0x480, 0x882, 0x3082, 0xC07E}},
    {35266, 2, {0x0, 0x1FC, 0x104, 0xFD04, 0x524, 0x524, 0x4924, 0x2924, 0x1124, 0x1154, 0x2850, 0x2490, 0x4490, 0x8112, 0x212, 0x40E}},
    {35267, 2, {0x0, 0x3E7C, 0x2204, 0x2208, 0x2A10, 0x2A10, 0x2AFE, 0x2A10, 0x2A10, 0x2A10, 0x810, 0x1450, 0x1422, 0x2402, 0x43FE, 0x8000}},
    {35268, 2, {0x1000, 0x11FC, 0x1104, 0x7D04, 0x1124, 0x1124, 0x1124, 0xFF24, 0x1124, 0x1154, 0x1050, 0x2890, 0x2490, 0x4512, 0x8212, 0x40E}},
    {35269, 2, {0x1F8, 0x7E00, 0x1210, 0x910, 0x920, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1290, 0x280, 0x482, 0x1882, 0x607E}},
    {35270, 2, {0x2000, 0x11FC, 0x1104, 0xF904, 0x924, 0x1124, 0x1124, 0x3924, 0x5524, 0x9554, 0x1050, 0x1090, 0x1090, 0x1112, 0x1212, 0x140E}},
    {35271, 2, {0x1000, 0x11FC, 0x1104, 0x1F04, 0x1124, 0x1124, 0x1124, 0x1124, 0x7D24, 0x4554, 0x4450, 0x4490, 0x4490, 0x7D12, 0x4612, 0x40E}},
    {35272, 2, {0x440, 0x2440, 0x247C, 0x2490, 0x2508, 0x400, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1290, 0x280, 0x484, 0x1884, 0x607C}},
    {35273, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1290, 0x280, 0x484, 0x1884, 0x607C}},
    {35274, 2, {0x800, 0x497C, 0x4944, 0x4944, 0x7F54, 0x54, 0x7F54, 0x154, 0x154, 0x3F54, 0x2010, 0x2028, 0x2328, 0x2C4A, 0x308A, 0x106}},
    {35275, 2, {0x0, 0x7C, 0x7F44, 0x844, 0x854, 0x2A54, 0x2A54, 0x2A54, 0x5D54, 0x4954, 0x8810, 0x828, 0xF28, 0xF04A, 0x408A, 0x106}},
    {35276, 2, {0x800, 0x87C, 0x7E44, 0x844, 0x854, 0xFF54, 0x154, 0x2A54, 0x1854, 0x4854, 0x2810, 0xFF28, 0x1428, 0x224A, 0x414A, 0x8186}},
    {35277, 2, {0x1400, 0x147C, 0x1444, 0x7F44, 0x5554, 0x5554, 0x5554, 0x7F54, 0x5554, 0x5554, 0x5510, 0xFFA8, 0x2428, 0x224A, 0x428A, 0x8106}},
    {35278, 2, {0x1000, 0x287C, 0x4444, 0x8244, 0x7D54, 0x54, 0xF154, 0x9554, 0x9554, 0xF554, 0x9510, 0x9528, 0xF528, 0x914A, 0x954A, 0xB286}},
    {35279, 2, {0x2400, 0x247C, 0xFF44, 0x2444, 0x7E54, 0x2454, 0xFF54, 0x854, 0x7E54, 0x4A54, 0x7E10, 0x4A28, 0xFF28, 0x424A, 0x4A4A, 0x4486}},
    {35280, 2, {0x2200, 0x227C, 0xFFC4, 0x2244, 0x3E54, 0x854, 0x7F54, 0x4954, 0x7F54, 0x854, 0x7F10, 0x828, 0x7F28, 0x84A, 0xF4A, 0xF086}},
    {35281, 2, {0x1000, 0x1E7C, 0x1044, 0x7F44, 0x5154, 0x5C54, 0x7254, 0x4E54, 0x4054, 0x4A54, 0x6A90, 0x5B28, 0x4A28, 0x8F4A, 0x784A, 0x86}},
    {35282, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x4108, 0x4128, 0x8010}},
    {35283, 2, {0x1004, 0x1004, 0x1F24, 0x2124, 0x4224, 0xBFA4, 0x24A4, 0x24A4, 0x3FA4, 0x24AC, 0x24B4, 0x3FA4, 0x2484, 0x4484, 0x4284, 0x8104}},
    {35284, 2, {0x1010, 0x1010, 0x1F10, 0x2110, 0x427E, 0xBF92, 0x2492, 0x2492, 0x3F92, 0x2492, 0x2492, 0x3F92, 0x2492, 0x44A2, 0x42AA, 0x8144}},
    {35285, 2, {0x1080, 0x1080, 0x50F8, 0x5108, 0x7E10, 0x55FC, 0x9124, 0x1124, 0x1DFC, 0xF124, 0x5124, 0x11FC, 0x1124, 0x1224, 0x1214, 0x1408}},
    {35286, 2, {0x2020, 0x2020, 0x7820, 0x49FC, 0x9024, 0x7C24, 0x5424, 0x5424, 0x7DFE, 0x5420, 0x5450, 0x7C50, 0x5488, 0x5488, 0x4504, 0x8E02}},
    {35287, 2, {0x1004, 0x101E, 0x3CF0, 0x2490, 0x4890, 0xBE90, 0x2A90, 0x2AFE, 0x3E90, 0x2A90, 0x2A90, 0x3E88, 0x2A8A, 0x4AAA, 0x42C6, 0x8682}},
    {35288, 2, {0x1010, 0x1010, 0x3C10, 0x2410, 0x4854, 0xBE52, 0x2A52, 0x2A90, 0x3E10, 0x2A14, 0x2A04, 0x3E08, 0x2A08, 0x4A10, 0x4220, 0x86C0}},
    {35289, 2, {0x1000, 0x10FC, 0x3C44, 0x2448, 0x4848, 0xBE50, 0x2A5C, 0x2A44, 0x3E44, 0x2A44, 0x2A68, 0x3EA8, 0x2A90, 0x4A98, 0x4324, 0x8642}},
    {35290, 2, {0x2008, 0x201C, 0x79F0, 0x4950, 0x9150, 0x7D50, 0x5550, 0x5550, 0x7D50, 0x5548, 0x5548, 0x7D68, 0x5554, 0x5674, 0x4652, 0x8C00}},
    {35291, 2, {0x1000, 0x1000, 0x3CFC, 0x2484, 0x4884, 0xBE84, 0x2AFC, 0x2A84, 0x3E84, 0x2A84, 0x2AFC, 0x3E84, 0x2A00, 0x4A00, 0x43FE, 0x8600}},
    {35292, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF07C, 0x800, 0xFC0, 0x1080, 0x3FF8, 0x5108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x2118}},
    {35293, 2, {0x1004, 0x101E, 0x3CF0, 0x2490, 0x4890, 0xBE90, 0x2A90, 0x2AFE, 0x3E90, 0x2A90, 0x2A90, 0x3E88, 0x2A8A, 0x4AAA, 0x42D6, 0x868A}},
    {35294, 2, {0x1040, 0x1040, 0x3C7E, 0x2480, 0x497C, 0xBE08, 0x2A10, 0x2A20, 0x3EFE, 0x2A4A, 0x2A4A, 0x3E92, 0x2B12, 0x4A22, 0x4254, 0x8688}},
    {35295, 2, {0x1010, 0x1010, 0x3C7C, 0x2410, 0x4810, 0xBE10, 0x2AFE, 0x2A00, 0x3E10, 0x2A10, 0x2A7C, 0x3E10, 0x2A10, 0x4A10, 0x42FE, 0x8600}},
    {35296, 2, {0x1110, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x3818, 0xCFC6, 0x1080, 0x3FF8, 0x5108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x2118}},
    {35297, 2, {0x1040, 0x1040, 0x3C78, 0x2488, 0x4950, 0xBE20, 0x2A50, 0x2A88, 0x3F06, 0x2AF8, 0x2A88, 0x3E88, 0x2A88, 0x4A88, 0x42F8, 0x8688}},
    {35298, 2, {0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x888, 0xFC0, 0x1080, 0x3FF8, 0x5108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x2118}},
    {35299, 2, {0x1000, 0x10FC, 0x3C24, 0x2424, 0x4854, 0xBE88, 0x2A10, 0x2A50, 0x3E7C, 0x2A90, 0x2A10, 0x3EFE, 0x2A10, 0x4A10, 0x4210, 0x8610}},
    {35300, 2, {0x2040, 0x2040, 0x78FC, 0x4904, 0x9208, 0x7DFE, 0x5500, 0x557C, 0x7D44, 0x5544, 0x5554, 0x7D48, 0x5542, 0x5542, 0x453E, 0x8E00}},
    {35301, 2, {0x2020, 0x2020, 0x7924, 0x48A4, 0x90A8, 0x7C20, 0x55FE, 0x5490, 0x7C90, 0x5490, 0x5490, 0x7C92, 0x5512, 0x5512, 0x460E, 0x8C00}},
    {35302, 2, {0x2020, 0x2020, 0x7820, 0x49FC, 0x9124, 0x7D24, 0x5524, 0x5524, 0x7DFC, 0x5524, 0x5420, 0x7C28, 0x5424, 0x57FE, 0x4502, 0x8C00}},
    {35303, 2, {0x1088, 0x1048, 0x3C50, 0x2400, 0x49FC, 0xBE20, 0x2A20, 0x2AFC, 0x3E20, 0x2A20, 0x2BFE, 0x3E20, 0x2A20, 0x4A20, 0x4220, 0x8620}},
    {35304, 2, {0x2000, 0x21FC, 0x7820, 0x4820, 0x93FE, 0x7C00, 0x5420, 0x5422, 0x7FB2, 0x54B4, 0x54A8, 0x7D28, 0x5524, 0x5624, 0x44A2, 0x8C40}},
    {35305, 2, {0x2028, 0x2024, 0x7824, 0x4820, 0x93FE, 0x7C20, 0x5520, 0x54B2, 0x7CB4, 0x5468, 0x54A8, 0x7D24, 0x5622, 0x5420, 0x44A0, 0x8C40}},
    {35306, 2, {0x1040, 0x1020, 0x3CFC, 0x2400, 0x4888, 0xBE50, 0x2BFE, 0x2A20, 0x3E20, 0x2A20, 0x2AFC, 0x3E20, 0x2A20, 0x4A20, 0x4220, 0x8620}},
    {35307, 2, {0x2020, 0x2020, 0x79FE, 0x4820, 0x9020, 0x7DFC, 0x5524, 0x5524, 0x7DFC, 0x5420, 0x5470, 0x7CA8, 0x5524, 0x5622, 0x4420, 0x8C20}},
    {35308, 2, {0x2000, 0x2040, 0x799C, 0x4904, 0x9104, 0x7DDC, 0x5504, 0x5504, 0x7DFC, 0x5450, 0x5450, 0x7C50, 0x5490, 0x5492, 0x4512, 0x8E0E}},
    {35309, 2, {0x2020, 0x2020, 0x79FC, 0x4850, 0x9088, 0x7D04, 0x57FE, 0x5408, 0x7DE8, 0x5528, 0x5528, 0x7DE8, 0x5528, 0x5408, 0x4428, 0x8C10}},
    {35310, 2, {0x1000, 0x10FC, 0x3C04, 0x2404, 0x487C, 0xBE04, 0x2A04, 0x2AFE, 0x3E10, 0x2A92, 0x2A54, 0x3E38, 0x2A54, 0x4A92, 0x4250, 0x8620}},
    {35311, 2, {0x2104, 0x2088, 0x7850, 0x49FC, 0x9124, 0x7D24, 0x55FC, 0x5524, 0x7D24, 0x55FC, 0x5420, 0x7C20, 0x57FE, 0x5420, 0x4420, 0x8C20}},
    {35312, 2, {0x1020, 0x1020, 0x3CFA, 0x2424, 0x4828, 0xBFFE, 0x2A20, 0x2A40, 0x3EFC, 0x2B44, 0x2A44, 0x3E7C, 0x2A44, 0x4A44, 0x427C, 0x8644}},
    {35313, 2, {0x28, 0x3FFE, 0x2020, 0x2FA4, 0x2028, 0x2F90, 0x48AA, 0x4FC6, 0x9082, 0x3FF8, 0x5108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x2118}},
    {35314, 2, {0x2088, 0x2050, 0x7BFE, 0x4820, 0x91FC, 0x7C20, 0x57FE, 0x5400, 0x7CA0, 0x54FC, 0x5520, 0x7C20, 0x57FE, 0x5420, 0x4420, 0x8C20}},
    {35315, 2, {0x800, 0xFFB8, 0x828, 0x7F28, 0x28, 0xFFC6, 0xA080, 0x3E7C, 0x4224, 0xFF24, 0x4928, 0x7F28, 0x4910, 0x7F28, 0x4944, 0x8B82}},
    {35316, 2, {0x1040, 0x107E, 0x3C80, 0x257C, 0x4844, 0xBE7C, 0x2A44, 0x2A7C, 0x3E00, 0x2AFE, 0x2A40, 0x3EFE, 0x2B2A, 0x4A4A, 0x4292, 0x8626}},
    {35317, 2, {0x2048, 0x2048, 0x79FC, 0x4848, 0x9048, 0x7DFE, 0x5420, 0x55FC, 0x7D24, 0x55FC, 0x5524, 0x7DFC, 0x5400, 0x5448, 0x4484, 0x8D02}},
    {35318, 2, {0x2000, 0x21DC, 0x7954, 0x4954, 0x91DC, 0x7C00, 0x55FC, 0x5524, 0x7DFC, 0x5524, 0x55FC, 0x7C20, 0x57FE, 0x5420, 0x4420, 0x8C20}},
    {35319, 2, {0xA80, 0x3138, 0x2288, 0x3938, 0x2288, 0xFFFE, 0x8802, 0xFC0, 0x1080, 0x3FF8, 0x5108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x2118}},
    {35320, 2, {0x2000, 0x21FC, 0x7954, 0x4954, 0x91FC, 0x7C80, 0x55FC, 0x5644, 0x7DF4, 0x5554, 0x5554, 0x7DF4, 0x5444, 0x55F4, 0x4414, 0x8C08}},
    {35321, 2, {0x1050, 0x1048, 0x3CFE, 0x2590, 0x48FC, 0xBE90, 0x2AFC, 0x2A90, 0x3EFE, 0x2A00, 0x2AEE, 0x3EAA, 0x2AAA, 0x4ABA, 0x4282, 0x8686}},
    {35322, 2, {0x48F8, 0x7050, 0x4420, 0x3DFC, 0x2024, 0x7CB8, 0x90A0, 0x7D7E, 0x2A00, 0x4FF0, 0x1020, 0x2FF8, 0x4888, 0xFF8, 0x888, 0x1098}},
    {35323, 2, {0x2124, 0x2174, 0x7A58, 0x4ADA, 0x93FE, 0x7D54, 0x56DA, 0x57FE, 0x7CA2, 0x5420, 0x57FE, 0x7C70, 0x54A8, 0x5524, 0x4622, 0x8C20}},
    {35324, 2, {0x2080, 0x20F8, 0x7908, 0x4BFE, 0x9122, 0x7D52, 0x54FC, 0x5484, 0x7CFC, 0x5484, 0x54FC, 0x7C40, 0x54FC, 0x5544, 0x4438, 0x8CEE}},
    {35325, 2, {0x2010, 0x2112, 0x79FE, 0x4848, 0x90FE, 0x7D90, 0x56FE, 0x5490, 0x7CFE, 0x5490, 0x54FE, 0x7C80, 0x55DE, 0x5552, 0x4572, 0x8D06}},
    {35326, 2, {0x2088, 0x2088, 0x7BFE, 0x4888, 0x90F8, 0x7D04, 0x5574, 0x5756, 0x7D54, 0x5554, 0x5774, 0x7D06, 0x5400, 0x5554, 0x452A, 0x8E2A}},
    {35327, 2, {0x2222, 0x23FE, 0x7890, 0x49FE, 0x9310, 0x7DFE, 0x5510, 0x55FE, 0x7D10, 0x55FE, 0x5400, 0x7DFE, 0x554A, 0x55F6, 0x4552, 0x8D76}},
    {35328, 2, {0x200, 0x100, 0xFFFE, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {35329, 2, {0x2000, 0x1000, 0x1000, 0xFE00, 0x0, 0x7C00, 0x0, 0x7C00, 0x0, 0x7C00, 0x4400, 0x4400, 0x4400, 0x7C00, 0x4400, 0x0}},
    {35330, 2, {0x2000, 0x1000, 0x11FE, 0xFE20, 0x20, 0x7C20, 0x20, 0x7C20, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x44A0, 0x40}},
    {35331, 2, {0x2040, 0x1040, 0x1040, 0xFE40, 0x40, 0x7C50, 0x48, 0x7C44, 0x44, 0x7C40, 0x4440, 0x4440, 0x4440, 0x7C40, 0x4440, 0x40}},
    {35332, 2, {0x2040, 0x2020, 0x23FC, 0x2000, 0xFDF8, 0x2400, 0x25F8, 0x2400, 0x25F8, 0x2508, 0x2508, 0x25FA, 0x250A, 0x4402, 0x43FE, 0x8000}},
    {35333, 2, {0x2040, 0x1040, 0x1040, 0xFE40, 0x1F8, 0x7C48, 0x48, 0x7C48, 0x48, 0x7C48, 0x4448, 0x444A, 0x448A, 0x7C8A, 0x4506, 0x200}},
    {35334, 2, {0x2004, 0x1004, 0x1084, 0xFE84, 0x84, 0x7C84, 0x84, 0x7C84, 0x8C, 0x7C94, 0x44E4, 0x4484, 0x4404, 0x7C04, 0x4404, 0x4}},
    {35335, 2, {0x1000, 0x1000, 0x1FFC, 0x2004, 0x4204, 0xBFE4, 0x4, 0x1FC4, 0x4, 0x1FC4, 0x4, 0x1FC4, 0x1044, 0x1FC4, 0x1054, 0x8}},
    {35336, 2, {0x2020, 0x1020, 0x1020, 0xFE20, 0x20, 0x7C20, 0x3FE, 0x7C20, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {35337, 2, {0x2000, 0x11F0, 0x110, 0xFD10, 0x110, 0x7910, 0x190, 0x7950, 0x150, 0x7910, 0x4910, 0x4912, 0x4912, 0x7A12, 0x4A0E, 0x400}},
    {35338, 2, {0x2000, 0x13F0, 0x90, 0xFC90, 0x90, 0x7890, 0x90, 0x7BF0, 0x90, 0x7890, 0x4890, 0x4892, 0x488A, 0x788A, 0x4886, 0x82}},
    {35339, 2, {0x2040, 0x1040, 0x1040, 0xFEFC, 0x84, 0x7D04, 0x4, 0x7D04, 0x84, 0x7C44, 0x4444, 0x4404, 0x4404, 0x7C04, 0x4428, 0x10}},
    {35340, 2, {0x2000, 0x1000, 0x10FC, 0xFE20, 0x20, 0x7C20, 0x20, 0x7C20, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x45FE, 0x0}},
    {35341, 2, {0x2000, 0x11F8, 0x1088, 0xFEC8, 0xA8, 0x7CA8, 0x88, 0x7C50, 0x50, 0x7C50, 0x4420, 0x4420, 0x4450, 0x7C88, 0x4504, 0x202}},
    {35342, 2, {0x2010, 0x1010, 0x1010, 0xFE10, 0x1FE, 0x7C10, 0x10, 0x7C10, 0x110, 0x7C90, 0x4490, 0x4410, 0x4410, 0x7C10, 0x4450, 0x20}},
    {35343, 2, {0x2000, 0x10FC, 0x1020, 0xFE20, 0x20, 0x7C20, 0x20, 0x7DFE, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x44A0, 0x40}},
    {35344, 2, {0x2000, 0x10FC, 0x1020, 0xFE20, 0x20, 0x7C20, 0x20, 0x7DFE, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {35345, 2, {0x2020, 0x1020, 0x1020, 0xFEA0, 0xAC, 0x7CB4, 0xE4, 0x7DA4, 0xA4, 0x7CB4, 0x44A8, 0x44A2, 0x44A2, 0x7C82, 0x447E, 0x0}},
    {35346, 2, {0x2000, 0x1000, 0x10FC, 0xFE24, 0x24, 0x7CA4, 0xA4, 0x7CA4, 0xA4, 0x7D24, 0x4424, 0x4444, 0x4444, 0x7C84, 0x4528, 0x210}},
    {35347, 2, {0x2104, 0x1124, 0x124, 0xFD24, 0x124, 0x7924, 0x124, 0x7924, 0x124, 0x7924, 0x4924, 0x4924, 0x4924, 0x7A24, 0x4A04, 0x404}},
    {35348, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x200, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {35349, 2, {0x2020, 0x1020, 0x1020, 0xFE20, 0x124, 0x7D24, 0x124, 0x7D24, 0x124, 0x7D24, 0x4524, 0x4524, 0x4524, 0x7DFC, 0x4404, 0x0}},
    {35350, 2, {0x2040, 0x1040, 0x1080, 0xFEFE, 0x100, 0x7E00, 0xFC, 0x7C08, 0x10, 0x7C20, 0x4440, 0x4480, 0x4502, 0x7D02, 0x44FE, 0x0}},
    {35351, 2, {0x2008, 0x103C, 0x11E0, 0xFE20, 0x20, 0x7C20, 0x20, 0x7C3E, 0x1E0, 0x7C20, 0x4420, 0x4422, 0x4422, 0x7C22, 0x441E, 0x0}},
    {35352, 2, {0x2000, 0x1000, 0x10FC, 0xFE04, 0x4, 0x7C04, 0x4, 0x7CFC, 0x84, 0x7C80, 0x4480, 0x4480, 0x4482, 0x7C82, 0x447E, 0x0}},
    {35353, 2, {0x2040, 0x1040, 0x1040, 0xFDF8, 0x48, 0x7C48, 0x48, 0x7D48, 0xC8, 0x7C48, 0x44A8, 0x44A8, 0x448A, 0x7D0A, 0x4506, 0x202}},
    {35354, 2, {0x2000, 0x17FC, 0x204, 0x4104, 0x5FF4, 0x4004, 0x4FE4, 0x4004, 0x4FE4, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4824, 0x400C}},
    {35355, 2, {0x20A0, 0x10A0, 0xA0, 0xFD24, 0x124, 0x7B28, 0x530, 0x7920, 0x160, 0x79A0, 0x4920, 0x4922, 0x4922, 0x7922, 0x491E, 0x100}},
    {35356, 2, {0x2010, 0x1050, 0x1050, 0xFE50, 0x88, 0x7C88, 0x104, 0x7EFA, 0x48, 0x7C48, 0x4448, 0x4448, 0x4488, 0x7C88, 0x4528, 0x210}},
    {35357, 2, {0x2000, 0x11FC, 0x1010, 0xFE90, 0x90, 0x7C90, 0x110, 0x7DFE, 0x30, 0x7C50, 0x4450, 0x4490, 0x4510, 0x7E10, 0x4450, 0x20}},
    {35358, 2, {0x2008, 0x103C, 0x11E0, 0xFE20, 0x20, 0x7C20, 0x3FE, 0x7C20, 0x20, 0x7C50, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {35359, 2, {0x2008, 0x1048, 0x1048, 0xFE48, 0x44, 0x7C84, 0xA4, 0x7D22, 0x20, 0x7C40, 0x4440, 0x4448, 0x4484, 0x7DFE, 0x4482, 0x0}},
    {35360, 2, {0x2004, 0x11E4, 0x24, 0xFC24, 0x24, 0x79E4, 0x104, 0x7904, 0x104, 0x79E4, 0x4824, 0x4824, 0x4824, 0x7824, 0x4944, 0x84}},
    {35361, 2, {0x2020, 0x1020, 0x1050, 0xFE50, 0x88, 0x7D44, 0x222, 0x7C20, 0x0, 0x7DFC, 0x4404, 0x4408, 0x4408, 0x7C10, 0x4410, 0x20}},
    {35362, 2, {0x2004, 0x100E, 0x10F0, 0xFE80, 0x80, 0x7C80, 0xFE, 0x7C88, 0x88, 0x7C88, 0x4488, 0x4488, 0x4488, 0x7D08, 0x4508, 0x208}},
    {35363, 2, {0x2020, 0x1020, 0x1020, 0xFDFC, 0x24, 0x7C24, 0x24, 0x7C24, 0x3FE, 0x7C20, 0x4450, 0x4450, 0x4488, 0x7C88, 0x4504, 0x202}},
    {35364, 2, {0x2008, 0x1088, 0x1050, 0xFE20, 0x50, 0x7C88, 0x100, 0x7C08, 0x88, 0x7C88, 0x4450, 0x4450, 0x4420, 0x7C50, 0x4488, 0x306}},
    {35365, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x3FE, 0x7A22, 0x222, 0x7A22, 0x252, 0x7A4A, 0x4A8A, 0x4B02, 0x4A02, 0x7A02, 0x4A0A, 0x204}},
    {35366, 2, {0x2020, 0x1020, 0x20, 0xFDFE, 0x122, 0x7A24, 0x20, 0x7820, 0x50, 0x7850, 0x4850, 0x4850, 0x4890, 0x7892, 0x4912, 0x20E}},
    {35367, 2, {0x2040, 0x1050, 0x1048, 0xFE48, 0x40, 0x7DFE, 0x50, 0x7C50, 0x50, 0x7C50, 0x4490, 0x4490, 0x4492, 0x7D12, 0x450E, 0x200}},
    {35368, 2, {0x2010, 0x1010, 0x1010, 0xFE90, 0x90, 0x7C90, 0x9E, 0x7C90, 0x90, 0x7C90, 0x4490, 0x4490, 0x4490, 0x7C90, 0x45FE, 0x0}},
    {35369, 2, {0x2000, 0x1008, 0x108, 0xFC88, 0x252, 0x7A52, 0x222, 0x7A22, 0x252, 0x7A92, 0x4B0A, 0x4A0A, 0x4A02, 0x7BFE, 0x4802, 0x0}},
    {35370, 2, {0x2040, 0x1020, 0x1020, 0xFDFE, 0x40, 0x7C40, 0x40, 0x7C7C, 0x44, 0x7C44, 0x4444, 0x4444, 0x4484, 0x7C84, 0x4528, 0x210}},
    {35371, 2, {0x2000, 0x1040, 0x20, 0xFC10, 0x90, 0x7880, 0x80, 0x7A84, 0x282, 0x7A82, 0x4A82, 0x4C88, 0x4888, 0x7888, 0x4878, 0x0}},
    {35372, 2, {0x2020, 0x1020, 0x1020, 0xFEA8, 0xA4, 0x7CA2, 0x122, 0x7D20, 0x224, 0x7C24, 0x4428, 0x4408, 0x4410, 0x7C20, 0x44C0, 0x300}},
    {35373, 2, {0x2000, 0x10F8, 0x1088, 0xFE88, 0x88, 0x7D06, 0x200, 0x7DFC, 0x84, 0x7C84, 0x4448, 0x4450, 0x4420, 0x7C50, 0x4488, 0x306}},
    {35374, 2, {0x2000, 0x11FC, 0x88, 0xFC88, 0x88, 0x7888, 0x88, 0x7BFE, 0x88, 0x7888, 0x4888, 0x4888, 0x4888, 0x7908, 0x4908, 0x208}},
    {35375, 2, {0x2000, 0x13FC, 0x84, 0xFC88, 0x88, 0x7890, 0x9C, 0x7884, 0x144, 0x7944, 0x4928, 0x4928, 0x4A10, 0x7A28, 0x4444, 0x182}},
    {35376, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x20, 0x7924, 0x124, 0x7924, 0x124, 0x79FC, 0x4824, 0x4820, 0x4822, 0x7822, 0x481E, 0x0}},
    {35377, 2, {0x2080, 0x1080, 0x80, 0xFDFC, 0x120, 0x7A20, 0x20, 0x7820, 0x3FE, 0x7820, 0x4820, 0x4820, 0x4820, 0x7820, 0x4820, 0x20}},
    {35378, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x1FC, 0x7924, 0x124, 0x7924, 0x124, 0x79FC, 0x4924, 0x4820, 0x4820, 0x7820, 0x4820, 0x20}},
    {35379, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x104, 0x7904, 0x1FC, 0x7924, 0x120, 0x7920, 0x4910, 0x4910, 0x4A08, 0x7A08, 0x4404, 0x802}},
    {35380, 2, {0x2008, 0x101C, 0x1E0, 0xFD00, 0x100, 0x7900, 0x1FE, 0x7910, 0x110, 0x7930, 0x4918, 0x4914, 0x4912, 0x7A10, 0x4A10, 0x410}},
    {35381, 2, {0x2000, 0x1000, 0x3FC, 0xFE94, 0x294, 0x7A94, 0x294, 0x7A94, 0x294, 0x7A94, 0x4A9C, 0x4B04, 0x4A04, 0x7A04, 0x4BFC, 0x204}},
    {35382, 2, {0x2000, 0x1000, 0x3FE, 0xFC08, 0x8, 0x79E8, 0x128, 0x7928, 0x128, 0x7928, 0x49E8, 0x4928, 0x4808, 0x7808, 0x4828, 0x10}},
    {35383, 2, {0x2020, 0x1020, 0x20, 0xFDFC, 0x124, 0x7924, 0x124, 0x79FC, 0x124, 0x7924, 0x4924, 0x49FC, 0x4924, 0x7820, 0x4820, 0x20}},
    {35384, 2, {0x201C, 0x10E0, 0x1020, 0xFE20, 0x20, 0x7DFE, 0x20, 0x7C70, 0x70, 0x7CA8, 0x44A8, 0x4524, 0x4622, 0x7C20, 0x4420, 0x20}},
    {35385, 2, {0x2020, 0x1028, 0x1024, 0xFC20, 0x3FE, 0x7C20, 0x70, 0x7C70, 0xA8, 0x7CA8, 0x4524, 0x4622, 0x4420, 0x7C20, 0x4420, 0x20}},
    {35386, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x104, 0x7A12, 0x20, 0x7840, 0x188, 0x7810, 0x4820, 0x4844, 0x4988, 0x7810, 0x4860, 0x380}},
    {35387, 2, {0x2040, 0x1020, 0x0, 0xFDFE, 0x20, 0x7820, 0x20, 0x7820, 0x1FC, 0x7820, 0x4820, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35388, 2, {0x2000, 0x1000, 0x3FE, 0xFC20, 0x20, 0x7820, 0x120, 0x7920, 0x13C, 0x7920, 0x4920, 0x4920, 0x4920, 0x7920, 0x47FE, 0x0}},
    {35389, 2, {0x2080, 0x1080, 0x100, 0xFDFC, 0x204, 0x7C04, 0x1E4, 0x7924, 0x124, 0x7924, 0x49E4, 0x4924, 0x4804, 0x7804, 0x4828, 0x10}},
    {35390, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF07C, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35391, 2, {0x2090, 0x1090, 0x90, 0xFA90, 0x292, 0x7AD4, 0x298, 0x7A90, 0x290, 0x7A90, 0x4A90, 0x4A92, 0x4AD2, 0x7F12, 0x4A0E, 0x0}},
    {35392, 2, {0x2020, 0x1020, 0x1020, 0xFE20, 0x3E, 0x7C20, 0x20, 0x7C20, 0x1FC, 0x7D04, 0x4504, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x104}},
    {35393, 2, {0x2020, 0x1020, 0x20, 0xFC20, 0x3FE, 0x7820, 0x20, 0x7820, 0x1FC, 0x7904, 0x4904, 0x4904, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35394, 2, {0x2088, 0x1088, 0x88, 0xFD08, 0x17E, 0x7B08, 0x508, 0x7948, 0x128, 0x7928, 0x4908, 0x4908, 0x4908, 0x7908, 0x4928, 0x110}},
    {35395, 2, {0x2020, 0x1010, 0x1010, 0xFDFE, 0x20, 0x7C20, 0x44, 0x7C84, 0x1F8, 0x7C10, 0x4420, 0x4444, 0x4482, 0x7DFE, 0x4482, 0x0}},
    {35396, 2, {0x2020, 0x1020, 0x120, 0xFD20, 0x1FC, 0x7920, 0x220, 0x7820, 0x3FE, 0x7820, 0x4850, 0x4850, 0x4888, 0x7888, 0x4904, 0x202}},
    {35397, 2, {0x2020, 0x1020, 0x1050, 0xFE50, 0x88, 0x7D24, 0x212, 0x7C10, 0x1FC, 0x7C04, 0x4408, 0x4488, 0x4450, 0x7C20, 0x4410, 0x10}},
    {35398, 2, {0x2004, 0x101E, 0x11F0, 0xFF10, 0x110, 0x7D10, 0x110, 0x7DFE, 0x110, 0x7D10, 0x4510, 0x4508, 0x450A, 0x7D4A, 0x45A6, 0x112}},
    {35399, 2, {0x2020, 0x1020, 0x20, 0xFDFC, 0x124, 0x7924, 0x124, 0x7924, 0x124, 0x7BFE, 0x4820, 0x4850, 0x4850, 0x7888, 0x4904, 0x202}},
    {35400, 2, {0x0, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {35401, 2, {0x4100, 0x2100, 0x13E, 0xF912, 0x7D2, 0x7252, 0x252, 0x7254, 0x254, 0x7494, 0x5288, 0x5108, 0x5294, 0x74A4, 0x5842, 0x80}},
    {35402, 2, {0x2020, 0x1020, 0x1124, 0xFEA4, 0xA8, 0x7C20, 0x1FC, 0x7C20, 0x20, 0x7C20, 0x47FE, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {35403, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x104, 0x7904, 0x1FC, 0x7850, 0x50, 0x7850, 0x4850, 0x4892, 0x4892, 0x7912, 0x4A0E, 0x400}},
    {35404, 2, {0x2088, 0x1088, 0x1088, 0xFE88, 0x1FE, 0x7C88, 0x88, 0x7C88, 0x88, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7C88, 0x44F8, 0x88}},
    {35405, 2, {0x2048, 0x1148, 0x148, 0xFD48, 0x148, 0x7BFE, 0x148, 0x7948, 0x148, 0x7948, 0x4978, 0x4900, 0x4900, 0x7900, 0x49FE, 0x0}},
    {35406, 2, {0x2000, 0x10FE, 0x1080, 0xFE80, 0x80, 0x7CFC, 0x84, 0x7C84, 0x84, 0x7C84, 0x44FC, 0x4480, 0x4480, 0x7C80, 0x44FE, 0x0}},
    {35407, 2, {0x2010, 0x1090, 0x90, 0xFD10, 0x17E, 0x7A52, 0x392, 0x7892, 0x112, 0x7912, 0x4A52, 0x4BD2, 0x4862, 0x7822, 0x484A, 0x84}},
    {35408, 2, {0x2080, 0x1080, 0x1080, 0xFCFE, 0x140, 0x7D40, 0x240, 0x7C7C, 0x40, 0x7C40, 0x4440, 0x447E, 0x4440, 0x7C40, 0x4440, 0x40}},
    {35409, 2, {0x2020, 0x1010, 0x1010, 0xFEFE, 0x82, 0x7D04, 0x40, 0x7C48, 0x50, 0x7C60, 0x4440, 0x4442, 0x4442, 0x7C42, 0x443E, 0x0}},
    {35410, 2, {0x2020, 0x1020, 0x1020, 0xFE40, 0x48, 0x7C84, 0x1FE, 0x7C82, 0x0, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {35411, 2, {0x2020, 0x1020, 0x1020, 0xFDFC, 0x20, 0x7C20, 0x20, 0x7DFE, 0x20, 0x7C20, 0x4440, 0x4448, 0x4484, 0x7DFE, 0x4482, 0x0}},
    {35412, 2, {0x2000, 0x11FC, 0x1044, 0xFE44, 0x44, 0x7C44, 0x94, 0x7C88, 0x100, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {35413, 2, {0x2000, 0x11FC, 0x1020, 0xFE20, 0x124, 0x7CA4, 0xA8, 0x7C20, 0x3FE, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {35414, 2, {0x2010, 0x1010, 0x10, 0xFDFE, 0x112, 0x7914, 0x110, 0x79FC, 0x144, 0x7944, 0x4928, 0x4928, 0x4910, 0x7A28, 0x4A44, 0x482}},
    {35415, 2, {0x2000, 0x11FE, 0x1102, 0xFF02, 0x102, 0x7D7A, 0x14A, 0x7D4A, 0x14A, 0x7D4A, 0x457A, 0x454A, 0x4502, 0x7D02, 0x450A, 0x104}},
    {35416, 2, {0x2020, 0x1020, 0x124, 0xFD24, 0x124, 0x7924, 0x1FC, 0x7820, 0x20, 0x7924, 0x4924, 0x4924, 0x4924, 0x7924, 0x49FC, 0x4}},
    {35417, 2, {0x2050, 0x1048, 0x48, 0xFC40, 0x3FE, 0x7880, 0x80, 0x78FC, 0x144, 0x7944, 0x4928, 0x4928, 0x4A10, 0x7A28, 0x4444, 0x182}},
    {35418, 2, {0x2000, 0x1000, 0x10FC, 0xFE84, 0x84, 0x7C84, 0xFC, 0x7C84, 0x84, 0x7C84, 0x44FC, 0x4484, 0x4400, 0x7C00, 0x45FE, 0x0}},
    {35419, 2, {0x2000, 0x10F8, 0x1088, 0xFE88, 0x88, 0x7CF8, 0x88, 0x7C88, 0x88, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7C88, 0x45FE, 0x0}},
    {35420, 2, {0x2020, 0x1124, 0x124, 0xFD24, 0x1FC, 0x7820, 0x20, 0x7BFC, 0x104, 0x7888, 0x4888, 0x4850, 0x4820, 0x7850, 0x4988, 0x606}},
    {35421, 2, {0x2040, 0x1020, 0x20, 0xFDFE, 0x102, 0x7A04, 0x0, 0x7800, 0x1FE, 0x7820, 0x4820, 0x4820, 0x4820, 0x7820, 0x48A0, 0x40}},
    {35422, 2, {0x2000, 0x11FC, 0x1004, 0xFE04, 0x1F4, 0x7C04, 0x4, 0x7DF4, 0x114, 0x7D14, 0x45F4, 0x4514, 0x4404, 0x7C04, 0x4428, 0x10}},
    {35423, 2, {0x480, 0x440, 0xFFFE, 0x940, 0x1184, 0x66FC, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35424, 2, {0x2040, 0x1020, 0x10, 0xFDE0, 0x22, 0x7834, 0x3B8, 0x78B0, 0xA8, 0x78A8, 0x4928, 0x4924, 0x4A24, 0x7C22, 0x48A0, 0x40}},
    {35425, 2, {0x2000, 0x13DE, 0x42, 0xFC42, 0x252, 0x794A, 0x14A, 0x7842, 0xC6, 0x794A, 0x4A52, 0x4842, 0x4842, 0x7842, 0x494A, 0x84}},
    {35426, 2, {0x2080, 0x1080, 0x1FC, 0xFD04, 0x204, 0x79E4, 0x124, 0x7924, 0x1E4, 0x7924, 0x4924, 0x49E4, 0x4904, 0x7804, 0x4828, 0x10}},
    {35427, 2, {0x2080, 0x1084, 0x1098, 0xFEE0, 0x82, 0x7C82, 0x7E, 0x7C00, 0xFC, 0x7C84, 0x4484, 0x44FC, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {35428, 2, {0x2010, 0x1110, 0x1110, 0xFF28, 0x144, 0x7D00, 0x1FE, 0x7C00, 0xA8, 0x7CA8, 0x44A8, 0x44A8, 0x44A8, 0x7CAA, 0x452A, 0x206}},
    {35429, 2, {0x2020, 0x1020, 0x1050, 0xFE50, 0x88, 0x7D04, 0x2FA, 0x7C00, 0x0, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7C88, 0x44F8, 0x88}},
    {35430, 2, {0x2010, 0x1014, 0x12, 0xFC10, 0x3FE, 0x7810, 0x10, 0x79D0, 0x90, 0x7890, 0x4890, 0x4888, 0x48EA, 0x7B8A, 0x4806, 0x2}},
    {35431, 2, {0x1080, 0x1080, 0x3EFC, 0x6244, 0x9448, 0x4830, 0x3110, 0x2FE8, 0xC006, 0x1FF0, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35432, 2, {0x2040, 0x1020, 0x20, 0xFDFE, 0x0, 0x7888, 0x104, 0x7A02, 0x88, 0x7888, 0x4850, 0x4850, 0x4820, 0x7850, 0x4888, 0x306}},
    {35433, 2, {0x2020, 0x1020, 0x20, 0xFDFC, 0x20, 0x7820, 0x3FE, 0x7808, 0x8, 0x79FE, 0x4808, 0x4888, 0x4848, 0x7808, 0x4828, 0x10}},
    {35434, 2, {0x2000, 0x11F8, 0x108, 0xFD08, 0x1F8, 0x7908, 0x108, 0x79F8, 0x144, 0x7948, 0x4930, 0x4920, 0x4910, 0x7948, 0x4986, 0x100}},
    {35435, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0x0, 0x7838, 0x3C0, 0x7840, 0x40, 0x787C, 0x4BC0, 0x4840, 0x4842, 0x7842, 0x483E, 0x0}},
    {35436, 2, {0x2008, 0x101C, 0x1E0, 0xFD00, 0x100, 0x79FE, 0x100, 0x7900, 0x17C, 0x7944, 0x4944, 0x4944, 0x4944, 0x7A7C, 0x4A44, 0x400}},
    {35437, 2, {0x2040, 0x1040, 0xFC, 0xFD04, 0x208, 0x79FE, 0x100, 0x797C, 0x144, 0x7944, 0x4954, 0x4948, 0x4942, 0x7A42, 0x4A3E, 0x400}},
    {35438, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x104, 0x7A02, 0x1FC, 0x7820, 0x20, 0x7820, 0x49FC, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35439, 2, {0x2020, 0x1020, 0x40, 0xFDFC, 0x104, 0x7904, 0x104, 0x79FC, 0x104, 0x7904, 0x49FC, 0x4904, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35440, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x20, 0x7820, 0x1FC, 0x7800, 0x0, 0x79FC, 0x4904, 0x4904, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35441, 2, {0x2008, 0x103C, 0x11E0, 0xFE20, 0x20, 0x7DFE, 0x20, 0x7C20, 0x20, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x104}},
    {35442, 2, {0x2020, 0x1010, 0x1010, 0xFDFE, 0x20, 0x7C24, 0x44, 0x7CF8, 0x12, 0x7C22, 0x4444, 0x4588, 0x4410, 0x7C28, 0x4444, 0x182}},
    {35443, 2, {0x2104, 0x1084, 0x88, 0xFC00, 0x3FE, 0x7820, 0x20, 0x79FC, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x7820, 0x4820, 0x20}},
    {35444, 2, {0x2020, 0x1020, 0x11FE, 0xFE40, 0x40, 0x7CFC, 0x84, 0x7D84, 0x2FC, 0x7C84, 0x4484, 0x44FC, 0x4484, 0x7C84, 0x4494, 0x88}},
    {35445, 2, {0x2020, 0x1120, 0x120, 0xFDFC, 0x120, 0x7A20, 0x20, 0x7BFE, 0x90, 0x7890, 0x4890, 0x4890, 0x4912, 0x7912, 0x4A12, 0x40E}},
    {35446, 2, {0x2082, 0x1092, 0x92, 0xFC92, 0x92, 0x7892, 0x2DA, 0x7AB6, 0x492, 0x7892, 0x4892, 0x4892, 0x4892, 0x7912, 0x4902, 0x202}},
    {35447, 2, {0x2000, 0x11FE, 0x1102, 0xFF02, 0x17A, 0x7D02, 0x102, 0x7D7A, 0x14A, 0x7D4A, 0x454A, 0x457A, 0x4502, 0x7D02, 0x450A, 0x104}},
    {35448, 2, {0x2020, 0x1020, 0x124, 0xFCA4, 0xA8, 0x7820, 0x20, 0x7BFE, 0x70, 0x78A8, 0x48A8, 0x4924, 0x4924, 0x7A22, 0x4820, 0x20}},
    {35449, 2, {0x800, 0xFF0, 0x1020, 0x3FFE, 0x6508, 0xA884, 0x3FFE, 0x2000, 0x27F8, 0x2000, 0x27F8, 0x2000, 0x27F8, 0x4408, 0x47F8, 0x8408}},
    {35450, 2, {0x2020, 0x1020, 0x107E, 0xFE82, 0x144, 0x7C28, 0x10, 0x7C20, 0x40, 0x7CFE, 0x4542, 0x4442, 0x4442, 0x7C42, 0x447E, 0x42}},
    {35451, 2, {0x2080, 0x1080, 0xF8, 0xFD08, 0x310, 0x7CA0, 0x40, 0x78A0, 0x118, 0x7A06, 0x4DF8, 0x4908, 0x4908, 0x7908, 0x49F8, 0x108}},
    {35452, 2, {0x2040, 0x1040, 0x40, 0xFDFE, 0x80, 0x7890, 0x90, 0x7912, 0x152, 0x7954, 0x4A90, 0x4A28, 0x4C28, 0x7844, 0x4884, 0x102}},
    {35453, 2, {0x2000, 0x11DC, 0x88, 0xFC88, 0x88, 0x7888, 0x88, 0x7BDE, 0x88, 0x7888, 0x4888, 0x4888, 0x4888, 0x7908, 0x4908, 0x208}},
    {35454, 2, {0x2100, 0x1100, 0x100, 0xFDFE, 0x202, 0x7A02, 0x512, 0x78A2, 0x24A, 0x7AAA, 0x4B1A, 0x4A0A, 0x4BFA, 0x7802, 0x4814, 0x8}},
    {35455, 2, {0x2020, 0x1020, 0x1FC, 0xFC20, 0x20, 0x7820, 0x3FE, 0x7800, 0x20, 0x7820, 0x49FC, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35456, 2, {0x2000, 0x13FE, 0x1088, 0xFE88, 0xF8, 0x7C88, 0x88, 0x7CF8, 0x88, 0x7C88, 0x448E, 0x47F8, 0x4408, 0x7C08, 0x4408, 0x8}},
    {35457, 2, {0x2104, 0x1084, 0x88, 0xFC00, 0x1FE, 0x7888, 0x88, 0x7888, 0x88, 0x7BFE, 0x4888, 0x4888, 0x4908, 0x7908, 0x4A08, 0x408}},
    {35458, 2, {0x2050, 0x1050, 0x1050, 0xFD52, 0xD4, 0x7C58, 0x50, 0x7C58, 0xD4, 0x7D52, 0x4650, 0x4450, 0x4492, 0x7C92, 0x4512, 0x20E}},
    {35459, 2, {0x2020, 0x1020, 0x107C, 0xFE84, 0x148, 0x7C30, 0x20, 0x7C48, 0x190, 0x7C3E, 0x4442, 0x45A4, 0x4418, 0x7C10, 0x4460, 0x180}},
    {35460, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x20, 0x79FC, 0x20, 0x7820, 0x3FE, 0x7870, 0x48A8, 0x48A8, 0x4924, 0x7A22, 0x4420, 0x20}},
    {35461, 2, {0x2020, 0x1120, 0x120, 0xFDFC, 0x120, 0x7A20, 0x20, 0x7BFE, 0x70, 0x78A8, 0x48A8, 0x4924, 0x4924, 0x7A22, 0x4820, 0x20}},
    {35462, 2, {0x2000, 0x13FE, 0x200, 0xFE00, 0x3FC, 0x7A20, 0x220, 0x7AF8, 0x220, 0x7A20, 0x4A20, 0x4BFC, 0x4A00, 0x7A00, 0x4BFE, 0x0}},
    {35463, 2, {0x2020, 0x1020, 0x3FE, 0xFC50, 0x88, 0x7904, 0x2FA, 0x7800, 0x3FE, 0x7840, 0x4880, 0x49FC, 0x4804, 0x7804, 0x4828, 0x10}},
    {35464, 2, {0x2000, 0x11FE, 0x20, 0xFC40, 0x88, 0x7904, 0x1FE, 0x7822, 0x20, 0x7820, 0x49FE, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35465, 2, {0x2208, 0x1108, 0x1110, 0x20, 0xFFFE, 0x820, 0x1110, 0x2FE8, 0xC006, 0x1FF0, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35466, 2, {0x1110, 0x910, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x920, 0x3FF8, 0xC006, 0x1FF0, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35467, 2, {0x2000, 0x11F8, 0x8, 0xFC08, 0x1F8, 0x7900, 0x104, 0x7904, 0xFC, 0x7840, 0x4820, 0x48A4, 0x4A8A, 0x7A8A, 0x4478, 0x0}},
    {35468, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x20, 0x7820, 0x20, 0x79FC, 0x40, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x7A8A, 0x4C78, 0x0}},
    {35469, 2, {0x2000, 0x13FC, 0x44, 0xFD44, 0x144, 0x7A84, 0x94, 0x7908, 0x240, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x7A8A, 0x4478, 0x0}},
    {35470, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x20, 0x79FC, 0x124, 0x7924, 0x1FC, 0x7820, 0x4870, 0x48A8, 0x4924, 0x7A22, 0x4820, 0x20}},
    {35471, 2, {0x2020, 0x1010, 0x10FC, 0xFE84, 0x84, 0x7CFC, 0x84, 0x7C84, 0xFC, 0x7CA2, 0x44A4, 0x4498, 0x4490, 0x7C88, 0x44C4, 0x82}},
    {35472, 2, {0x2010, 0x10D8, 0x394, 0xFC94, 0x90, 0x7BFE, 0x90, 0x7894, 0x94, 0x78D8, 0x4B98, 0x4890, 0x48AA, 0x78CA, 0x4A86, 0x102}},
    {35473, 2, {0x4000, 0x2440, 0x2BE, 0xF908, 0x288, 0x7488, 0x88, 0x70BE, 0x188, 0x7288, 0x5488, 0x5088, 0x5088, 0x7088, 0x553E, 0x200}},
    {35474, 2, {0x2020, 0x1040, 0x88, 0xFD04, 0x3FE, 0x7882, 0x80, 0x79FC, 0x220, 0x7820, 0x4BFE, 0x4820, 0x4850, 0x7888, 0x4904, 0x602}},
    {35475, 2, {0x80C, 0x7EF0, 0x880, 0xEFE, 0x7888, 0xA88, 0x1908, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35476, 2, {0x2000, 0x1006, 0x3B8, 0xFC88, 0x88, 0x7908, 0x13E, 0x7B88, 0x88, 0x7A88, 0x4A88, 0x493E, 0x4900, 0x7A80, 0x447E, 0x800}},
    {35477, 2, {0x2004, 0x100E, 0x3B8, 0xFC88, 0x88, 0x7928, 0x12E, 0x7BA8, 0xA8, 0x7AA8, 0x4AA8, 0x493E, 0x4900, 0x7A80, 0x447E, 0x800}},
    {35478, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x20, 0x7BFE, 0x202, 0x7C04, 0x1F8, 0x7810, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x48A0, 0x40}},
    {35479, 2, {0x2042, 0x10E2, 0x382, 0xFC8A, 0x8A, 0x788A, 0x3EA, 0x788A, 0x18A, 0x79CA, 0x4AAA, 0x4A82, 0x4C82, 0x7882, 0x488A, 0x84}},
    {35480, 2, {0x201C, 0x11E0, 0x20, 0xFC20, 0x3FE, 0x78A8, 0x124, 0x7A22, 0x1F8, 0x7888, 0x4890, 0x48BE, 0x4902, 0x7902, 0x4A14, 0x408}},
    {35481, 2, {0x2000, 0x11FE, 0x0, 0xFC92, 0x124, 0x7A48, 0x124, 0x7892, 0x0, 0x79FE, 0x4820, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35482, 2, {0x2020, 0x1124, 0x10A4, 0xFEA8, 0x20, 0x7DFC, 0x104, 0x7D04, 0x1FC, 0x7D04, 0x4504, 0x45FC, 0x4504, 0x7D04, 0x4514, 0x108}},
    {35483, 2, {0x2000, 0x11FC, 0x4, 0xFCFC, 0x4, 0x79FC, 0x0, 0x7BFE, 0x202, 0x79F8, 0x4888, 0x4888, 0x4850, 0x7820, 0x48D8, 0x306}},
    {35484, 2, {0x2040, 0x1040, 0x88, 0xFD04, 0x3FE, 0x7802, 0x88, 0x7944, 0x242, 0x78F8, 0x4988, 0x4A50, 0x4820, 0x7850, 0x4988, 0x606}},
    {35485, 2, {0x2020, 0x1020, 0x1050, 0xFE88, 0x144, 0x7E22, 0xF8, 0x7C08, 0x10, 0x7C20, 0x45FC, 0x4504, 0x4504, 0x7D04, 0x45FC, 0x104}},
    {35486, 2, {0x2000, 0x13FC, 0x40, 0xFC40, 0x1F8, 0x7888, 0x88, 0x7BFE, 0x0, 0x7800, 0x49F8, 0x4908, 0x4908, 0x7908, 0x49F8, 0x108}},
    {35487, 2, {0x2020, 0x1022, 0x11FA, 0xFE24, 0x24, 0x7DFE, 0x10, 0x7C20, 0x1FC, 0x7C88, 0x4510, 0x46FE, 0x4410, 0x7C10, 0x4450, 0x20}},
    {35488, 2, {0x2028, 0x1024, 0x24, 0xFC20, 0x3FE, 0x7A20, 0x224, 0x7A24, 0x3A4, 0x7AA8, 0x4AA8, 0x4A90, 0x4A92, 0x7DAA, 0x4446, 0x882}},
    {35489, 2, {0x2010, 0x1014, 0x12, 0xFC10, 0x3FE, 0x7810, 0x150, 0x7954, 0x7F4, 0x7954, 0x4954, 0x4948, 0x4A4A, 0x7A5A, 0x4426, 0x42}},
    {35490, 2, {0x2000, 0x11F8, 0x108, 0xFD08, 0x1F8, 0x7908, 0x108, 0x79F8, 0x108, 0x7908, 0x49F8, 0x4890, 0x4890, 0x7912, 0x4A12, 0x40E}},
    {35491, 2, {0x2000, 0x11FC, 0x20, 0xFC20, 0x124, 0x7924, 0x124, 0x7924, 0x2AA, 0x7AAA, 0x4CB2, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35492, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x1FC, 0x7800, 0x0, 0x79FC, 0x20, 0x7820, 0x4BFE, 0x4820, 0x4850, 0x7888, 0x4904, 0x202}},
    {35493, 2, {0x2020, 0x1120, 0x120, 0xFDFC, 0x120, 0x7A20, 0x20, 0x7BFE, 0x0, 0x7800, 0x49FC, 0x4904, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35494, 2, {0x2000, 0x11F8, 0x1008, 0xFED0, 0x20, 0x7DFC, 0x124, 0x7D24, 0x1FC, 0x7D24, 0x4524, 0x45FC, 0x4524, 0x7D24, 0x4524, 0x10C}},
    {35495, 2, {0x2028, 0x1024, 0x11FE, 0xFE20, 0x20, 0x7DFC, 0x124, 0x7D24, 0x1FC, 0x7D24, 0x4524, 0x45FC, 0x4524, 0x7D24, 0x4524, 0x10C}},
    {35496, 2, {0x2080, 0x1080, 0x10FE, 0xFD00, 0x2FC, 0x7C84, 0xA4, 0x7C94, 0x3FE, 0x7C84, 0x4524, 0x4514, 0x45FE, 0x7C04, 0x4428, 0x10}},
    {35497, 2, {0x2020, 0x1010, 0x1010, 0xFEFE, 0x0, 0x7C7C, 0x0, 0x7C7C, 0x0, 0x7C7C, 0x4444, 0x4444, 0x4444, 0x7C7C, 0x4444, 0x0}},
    {35498, 2, {0x2010, 0x1090, 0x88, 0xFD08, 0x204, 0x7DFA, 0x108, 0x7908, 0x108, 0x79F8, 0x4890, 0x4890, 0x4890, 0x7912, 0x4A12, 0x40E}},
    {35499, 2, {0x2000, 0x11FE, 0x100, 0xFD00, 0x17C, 0x7900, 0x100, 0x79FE, 0x150, 0x7952, 0x4954, 0x4948, 0x4948, 0x7A44, 0x4A52, 0x460}},
    {35500, 2, {0x2104, 0x1084, 0x88, 0xFC10, 0x1FC, 0x7904, 0x104, 0x7904, 0x1FC, 0x7850, 0x4850, 0x4890, 0x4892, 0x7912, 0x4A0E, 0x400}},
    {35501, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7800, 0x3FE, 0x7A02, 0x404, 0x7890, 0x4890, 0x4890, 0x4890, 0x7912, 0x4912, 0x20E}},
    {35502, 2, {0x2088, 0x1088, 0x13FE, 0xFC88, 0x88, 0x7C50, 0x50, 0x7C94, 0x194, 0x7E98, 0x4490, 0x44B2, 0x44D2, 0x7C92, 0x448E, 0x80}},
    {35503, 2, {0x20FC, 0x1084, 0x1084, 0xFEFC, 0x84, 0x7C84, 0xFC, 0x7C00, 0x1FE, 0x7D02, 0x4502, 0x45FE, 0x4502, 0x7D02, 0x45FE, 0x102}},
    {35504, 2, {0x20A0, 0x1090, 0x80, 0xFDFE, 0x110, 0x7B10, 0x5FC, 0x7910, 0x110, 0x79FC, 0x4910, 0x4910, 0x4910, 0x79FE, 0x4900, 0x100}},
    {35505, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7824, 0x3FE, 0x7824, 0x1FC, 0x7820, 0x4920, 0x493E, 0x4920, 0x7AA0, 0x4A7E, 0x400}},
    {35506, 2, {0x2000, 0x11FC, 0x124, 0xFD24, 0x1FC, 0x7924, 0x124, 0x79FC, 0x20, 0x7BFE, 0x4870, 0x48A8, 0x4924, 0x7A22, 0x4820, 0x20}},
    {35507, 2, {0x4000, 0x23FE, 0x202, 0xFA02, 0x3FE, 0x7210, 0x292, 0x7292, 0x292, 0x72FE, 0x5210, 0x5292, 0x5292, 0x7492, 0x54FE, 0x802}},
    {35508, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0x0, 0x79FC, 0x0, 0x7800, 0x3FE, 0x7820, 0x4928, 0x4924, 0x4A22, 0x7C22, 0x48A0, 0x40}},
    {35509, 2, {0x2184, 0x1068, 0x30, 0xFCC8, 0x324, 0x7820, 0x3FE, 0x7840, 0xFC, 0x7984, 0x4AFC, 0x4884, 0x48FC, 0x7884, 0x4894, 0x88}},
    {35510, 2, {0x2040, 0x1020, 0x3FE, 0xFC00, 0x88, 0x7888, 0x154, 0x7A22, 0x0, 0x7820, 0x4BFE, 0x4820, 0x4820, 0x7820, 0x4820, 0x20}},
    {35511, 2, {0x2000, 0x13FE, 0x202, 0xFE8A, 0x252, 0x7BFE, 0x242, 0x7A22, 0x3FE, 0x7A82, 0x4A82, 0x4A82, 0x4AFA, 0x7A02, 0x4A0A, 0x204}},
    {35512, 2, {0x2010, 0x1388, 0x88, 0xFCFE, 0x90, 0x7B90, 0x224, 0x7A24, 0x238, 0x7B88, 0x4890, 0x4890, 0x48A4, 0x78BE, 0x4A82, 0x100}},
    {35513, 2, {0x2050, 0x1050, 0x50, 0xFBDE, 0x50, 0x7850, 0x50, 0x79DC, 0x50, 0x7850, 0x4850, 0x4850, 0x4BDE, 0x7850, 0x4850, 0x50}},
    {35514, 2, {0x2020, 0x1020, 0x20, 0xFBFE, 0x20, 0x7924, 0x124, 0x7924, 0x2AA, 0x7870, 0x48A8, 0x48A8, 0x4924, 0x7A22, 0x4C20, 0x20}},
    {35515, 2, {0x2020, 0x1022, 0x3B4, 0xFCA8, 0xA8, 0x7924, 0x2A2, 0x7840, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35516, 2, {0x2020, 0x1010, 0x1FE, 0xFD02, 0x204, 0x78F8, 0x88, 0x7888, 0xF8, 0x7888, 0x4888, 0x48F8, 0x4888, 0x7888, 0x4BFE, 0x0}},
    {35517, 2, {0x2000, 0x1040, 0x39C, 0xFA04, 0x204, 0x7B9C, 0x204, 0x7A04, 0x3FC, 0x7890, 0x4890, 0x4890, 0x4890, 0x7912, 0x4A12, 0x40E}},
    {35518, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x5FF4, 0x4004, 0x4FE4, 0x4004, 0x4FE4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {35519, 2, {0x2000, 0x11FC, 0x124, 0xFD24, 0x174, 0x7924, 0x1FC, 0x7904, 0x174, 0x7954, 0x4954, 0x4974, 0x4904, 0x7A04, 0x4A14, 0x408}},
    {35520, 2, {0x2040, 0x1080, 0x1FC, 0xFD24, 0x124, 0x79FC, 0x124, 0x7944, 0x1FC, 0x7890, 0x4910, 0x4BFE, 0x4810, 0x7810, 0x4810, 0x10}},
    {35521, 2, {0x2000, 0x13DE, 0x42, 0xFD4A, 0x84, 0x794A, 0x252, 0x7C20, 0x0, 0x7BDE, 0x4852, 0x4952, 0x4894, 0x7948, 0x4A54, 0x422}},
    {35522, 2, {0x2040, 0x1040, 0x7C, 0xFC84, 0x108, 0x7A00, 0x20, 0x79CE, 0x102, 0x7902, 0x49CE, 0x4902, 0x4902, 0x7902, 0x49FE, 0x102}},
    {35523, 2, {0x2088, 0x1088, 0x88, 0xFC88, 0x3DE, 0x7888, 0x88, 0x799C, 0x1DC, 0x7AAA, 0x4AAA, 0x4CC8, 0x4888, 0x7888, 0x4888, 0x88}},
    {35524, 2, {0x2040, 0x1020, 0x3FE, 0xFC00, 0x1FC, 0x7904, 0x1FC, 0x7800, 0x1FC, 0x7808, 0x4810, 0x4BFE, 0x4820, 0x7820, 0x48A0, 0x40}},
    {35525, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0xFFFE, 0x1110, 0x3FF8, 0xC006, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35526, 2, {0x2088, 0x1088, 0x13FE, 0xFC88, 0x88, 0x7CF8, 0x88, 0x7C88, 0xF8, 0x7C88, 0x4488, 0x47FE, 0x4400, 0x7C88, 0x4504, 0x202}},
    {35527, 2, {0x2020, 0x10A2, 0xA2, 0xFD24, 0x50, 0x7888, 0x304, 0x7822, 0x20, 0x78A4, 0x48A4, 0x4928, 0x4850, 0x7888, 0x4904, 0x202}},
    {35528, 2, {0x2008, 0x103C, 0x1E0, 0xFC20, 0x3FE, 0x78A8, 0xA8, 0x78A8, 0x3FE, 0x78A8, 0x48A8, 0x4BFE, 0x4820, 0x7820, 0x49FC, 0x0}},
    {35529, 2, {0x201C, 0x11E0, 0x20, 0xFC20, 0x3FE, 0x78A8, 0x124, 0x7A42, 0x40, 0x7BFE, 0x4888, 0x4908, 0x48D0, 0x7830, 0x4848, 0x184}},
    {35530, 2, {0x2080, 0x1080, 0xFE, 0xFD02, 0x242, 0x7A4A, 0x152, 0x7842, 0x3FE, 0x7842, 0x48E2, 0x4952, 0x4A4A, 0x7842, 0x4854, 0x8}},
    {35531, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7820, 0x3FE, 0x7800, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4914, 0x108}},
    {35532, 2, {0x2020, 0x1020, 0x13FE, 0xFC20, 0x1FC, 0x7D24, 0x1FC, 0x7D24, 0x1FC, 0x7C20, 0x4470, 0x44A8, 0x4524, 0x7E22, 0x4420, 0x20}},
    {35533, 2, {0x2080, 0x1080, 0x1F8, 0xFD08, 0x210, 0x7DFC, 0x24, 0x7824, 0x3FE, 0x7824, 0x4824, 0x49FC, 0x4824, 0x7820, 0x48A0, 0x40}},
    {35534, 2, {0x2048, 0x1048, 0x1048, 0xFDFE, 0x48, 0x7C48, 0x1FE, 0x7C00, 0xFC, 0x7C84, 0x4484, 0x44FC, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {35535, 2, {0x2000, 0x17E0, 0x25E, 0xFA52, 0x252, 0x7BD2, 0x252, 0x7A54, 0x3D4, 0x7A54, 0x4A48, 0x4AE8, 0x4F54, 0x7854, 0x4862, 0x40}},
    {35536, 2, {0x17F8, 0x2408, 0x67F8, 0xA2A0, 0x24A4, 0x289C, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35537, 2, {0x2000, 0x13FE, 0x40, 0xFC80, 0x144, 0x7A24, 0x68, 0x7AB0, 0x130, 0x7AA8, 0x4868, 0x48A4, 0x4922, 0x7A20, 0x48A0, 0x40}},
    {35538, 2, {0x2040, 0x1020, 0x3FE, 0xFC00, 0x0, 0x79FC, 0x104, 0x7904, 0x104, 0x79FC, 0x4820, 0x48A8, 0x4924, 0x7A22, 0x48A0, 0x40}},
    {35539, 2, {0x2048, 0x1044, 0x105E, 0xFDE0, 0x28, 0x7C12, 0x6A, 0x7D96, 0x48, 0x7C5E, 0x45E0, 0x4424, 0x4428, 0x7C12, 0x446A, 0x186}},
    {35540, 2, {0x4100, 0x2100, 0x1DC, 0xF914, 0x114, 0x7114, 0x7D4, 0x7014, 0x114, 0x7114, 0x5588, 0x5548, 0x5948, 0x7114, 0x5514, 0x222}},
    {35541, 2, {0x2020, 0x1020, 0x3E, 0xFC20, 0x1FE, 0x7922, 0x138, 0x79E0, 0x122, 0x791E, 0x4900, 0x4978, 0x4A48, 0x7A4A, 0x448A, 0x906}},
    {35542, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x104, 0x7AFA, 0x0, 0x79FC, 0x154, 0x7954, 0x49FC, 0x4954, 0x4954, 0x7954, 0x4904, 0x10C}},
    {35543, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x144, 0x7A22, 0x1F8, 0x7808, 0x50, 0x7820, 0x48A4, 0x4A82, 0x4A8A, 0x7A8A, 0x4478, 0x0}},
    {35544, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x20, 0x79FC, 0x20, 0x7820, 0x3FE, 0x7840, 0x48A4, 0x49A8, 0x4A90, 0x7C88, 0x48C6, 0x80}},
    {35545, 2, {0x201C, 0x13E0, 0x220, 0xFBFE, 0x220, 0x7A92, 0x30A, 0x7A06, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35546, 2, {0x2020, 0x1010, 0x10FE, 0xFE82, 0x104, 0x7C00, 0xFE, 0x7C10, 0x10, 0x7C90, 0x449E, 0x4490, 0x4490, 0x7CD0, 0x453E, 0x200}},
    {35547, 2, {0x2020, 0x1020, 0xA0, 0xFD2C, 0x124, 0x7924, 0x1AC, 0x7924, 0x124, 0x79FC, 0x4924, 0x4850, 0x4850, 0x7888, 0x4904, 0x202}},
    {35548, 2, {0x2048, 0x1148, 0x148, 0xFBFE, 0x148, 0x7948, 0x178, 0x7900, 0x1FE, 0x7820, 0x4BFE, 0x4870, 0x48A8, 0x7924, 0x4E22, 0x20}},
    {35549, 2, {0x2000, 0x13FE, 0x22, 0xFD20, 0x13C, 0x7920, 0x2FE, 0x7C00, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x4914, 0x108}},
    {35550, 2, {0x2040, 0x1020, 0x1FE, 0xFD02, 0x102, 0x79FE, 0x100, 0x7900, 0x1FE, 0x79AA, 0x49AA, 0x4AFE, 0x4AAA, 0x7AAA, 0x44A2, 0x86}},
    {35551, 2, {0x2000, 0x11FC, 0x104, 0xFD04, 0x1FC, 0x7904, 0x104, 0x79FC, 0x0, 0x7BFE, 0x4820, 0x4920, 0x493C, 0x7920, 0x4AA0, 0x47E}},
    {35552, 2, {0x2040, 0x1020, 0x3FE, 0xFA02, 0x404, 0x79FC, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x7800, 0x4BFE, 0x0}},
    {35553, 2, {0x2088, 0x1104, 0x2FA, 0xFC40, 0x80, 0x78F8, 0x8, 0x7850, 0x20, 0x79FC, 0x4954, 0x4954, 0x4954, 0x7954, 0x4BFE, 0x0}},
    {35554, 2, {0x2000, 0x13FE, 0x222, 0xFC20, 0x3FE, 0x7820, 0x1FC, 0x7924, 0x1FC, 0x7924, 0x49FC, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20}},
    {35555, 2, {0x2000, 0x10FC, 0x1084, 0xFE84, 0xF4, 0x7C94, 0x94, 0x7DFE, 0x102, 0x7D7A, 0x454A, 0x454A, 0x457A, 0x7D02, 0x450A, 0x104}},
    {35556, 2, {0x2000, 0x13DE, 0x252, 0xFE52, 0x3DE, 0x7800, 0x1FC, 0x7800, 0x3FE, 0x7880, 0x4900, 0x49FC, 0x4804, 0x7804, 0x4828, 0x10}},
    {35557, 2, {0x2008, 0x103C, 0x1E0, 0xFC20, 0x3FE, 0x7820, 0x1FC, 0x7924, 0x1FC, 0x7924, 0x49FC, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x0}},
    {35558, 2, {0x2040, 0x1020, 0x1FC, 0xFC00, 0x88, 0x7850, 0x3FE, 0x7A22, 0x424, 0x79FC, 0x4924, 0x4924, 0x4934, 0x7928, 0x4820, 0x20}},
    {35559, 2, {0x2110, 0x1112, 0x1D4, 0xFD18, 0x152, 0x7992, 0x12E, 0x7840, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35560, 2, {0x2000, 0x13FE, 0x0, 0xFDFC, 0x104, 0x7904, 0x1FC, 0x7800, 0x3FE, 0x7A22, 0x4A22, 0x4BFE, 0x4A22, 0x7A22, 0x4BFE, 0x202}},
    {35561, 2, {0x2084, 0x1048, 0x0, 0xFDFE, 0x48, 0x7848, 0x48, 0x7A48, 0x14A, 0x794C, 0x4848, 0x4848, 0x4848, 0x7848, 0x4BFE, 0x0}},
    {35562, 2, {0x2040, 0x1020, 0x3FE, 0xFC00, 0x1FC, 0x7904, 0x1FC, 0x7800, 0x3FE, 0x7A02, 0x49FC, 0x4820, 0x4820, 0x7820, 0x48A0, 0x40}},
    {35563, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7924, 0x1AC, 0x7974, 0x124, 0x79FC, 0x4820, 0x4870, 0x48A8, 0x7924, 0x4A22, 0x20}},
    {35564, 2, {0xC28, 0x7024, 0x11FE, 0xFE50, 0x3852, 0x548E, 0x9100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35565, 2, {0x2040, 0x10A0, 0x110, 0xFA08, 0x5F6, 0x7800, 0x3C4, 0x7A54, 0x254, 0x7BD4, 0x4A54, 0x4A54, 0x4BD4, 0x7A44, 0x4A54, 0x2C8}},
    {35566, 2, {0x2020, 0x1120, 0xBE, 0xFC42, 0x84, 0x7B10, 0x110, 0x7928, 0x144, 0x7882, 0x48FC, 0x4884, 0x4884, 0x7884, 0x48FC, 0x84}},
    {35567, 2, {0x2020, 0x1124, 0x124, 0xFD24, 0x1FC, 0x7800, 0x3FE, 0x7820, 0x40, 0x79FC, 0x4954, 0x4954, 0x4954, 0x7954, 0x4954, 0x10C}},
    {35568, 2, {0x2000, 0x11FC, 0x124, 0xFD24, 0x1FC, 0x7924, 0x124, 0x79FC, 0x0, 0x7840, 0x4824, 0x4AA2, 0x4A8A, 0x7C88, 0x4878, 0x0}},
    {35569, 2, {0x2040, 0x1040, 0x1FC, 0xFC84, 0x3FE, 0x7800, 0x1FC, 0x7904, 0x1FC, 0x7820, 0x4BFE, 0x4820, 0x4A20, 0x7BFE, 0x4820, 0x20}},
    {35570, 2, {0x2000, 0x13FE, 0x50, 0xFC50, 0x1FC, 0x7954, 0x154, 0x7954, 0x1FC, 0x7820, 0x4820, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35571, 2, {0x2020, 0x1010, 0x11FE, 0xFE00, 0x84, 0x7C48, 0x1FE, 0x7C00, 0xFC, 0x7C84, 0x4484, 0x44FC, 0x4484, 0x7C84, 0x44FC, 0x84}},
    {35572, 2, {0x2008, 0x100C, 0xA, 0xFBFE, 0x208, 0x7A08, 0x2E8, 0x7A0A, 0x20A, 0x7AEC, 0x4AAC, 0x4AA8, 0x4AEA, 0x7A1A, 0x4A26, 0x442}},
    {35573, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x20, 0x7BFE, 0x28A, 0x7A52, 0x2FA, 0x7A22, 0x4A22, 0x4AFA, 0x4A22, 0x7A22, 0x4A2A, 0x204}},
    {35574, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0xF8, 0x7888, 0xF8, 0x7888, 0x88, 0x7BFE, 0x4900, 0x4948, 0x4984, 0x7900, 0x49FE, 0x0}},
    {35575, 2, {0x4000, 0x27FC, 0x404, 0xFC34, 0x5C4, 0x7444, 0x5F4, 0x7554, 0x554, 0x75F4, 0x5444, 0x5454, 0x55F4, 0x7494, 0x5806, 0x1002}},
    {35576, 2, {0x2020, 0x1022, 0x1FA, 0xFC24, 0x28, 0x7BFE, 0x20, 0x7840, 0xFC, 0x7984, 0x4A84, 0x48FC, 0x4884, 0x7884, 0x48FC, 0x84}},
    {35577, 2, {0x2000, 0x11F8, 0x108, 0xFDF8, 0x108, 0x79F8, 0x0, 0x7BFE, 0x100, 0x79FC, 0x4A54, 0x4C94, 0x4924, 0x7A44, 0x48A8, 0x110}},
    {35578, 2, {0x2020, 0x1010, 0x1FE, 0xFC00, 0x84, 0x7848, 0x1FE, 0x7908, 0x110, 0x7964, 0x4908, 0x4912, 0x4964, 0x7A08, 0x4A30, 0x4C0}},
    {35579, 2, {0x2020, 0x1040, 0x1FC, 0xFD04, 0x1FC, 0x7904, 0x1FC, 0x7800, 0x3FE, 0x7820, 0x4820, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35580, 2, {0x201E, 0x13E0, 0x44, 0xFD24, 0x88, 0x79FC, 0x40, 0x7840, 0x3FE, 0x7880, 0x48FC, 0x4944, 0x4928, 0x7A10, 0x4868, 0x186}},
    {35581, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0xF8, 0x7820, 0x1FC, 0x7924, 0x124, 0x79FC, 0x4820, 0x4BFE, 0x4820, 0x7820, 0x4820, 0x20}},
    {35582, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0xA8, 0x7820, 0x3FE, 0x7840, 0x80, 0x78FC, 0x4984, 0x4A84, 0x4884, 0x7884, 0x48FC, 0x84}},
    {35583, 2, {0x2000, 0x10F8, 0x1088, 0xFE88, 0xF8, 0x7C00, 0x1FE, 0x7C88, 0xF8, 0x7C88, 0x44F8, 0x4488, 0x449E, 0x7DE8, 0x4408, 0x8}},
    {35584, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0x88, 0x78F8, 0x88, 0x7888, 0xF8, 0x7820, 0x4BFE, 0x4870, 0x48A8, 0x7924, 0x4A22, 0x20}},
    {35585, 2, {0x2000, 0x11FC, 0x1104, 0xFDFC, 0x104, 0x7DFC, 0x80, 0x7DFE, 0x222, 0x7D22, 0x4552, 0x4502, 0x45FA, 0x7C02, 0x4414, 0x8}},
    {35586, 2, {0x2000, 0x13FE, 0x222, 0xFBFE, 0x222, 0x7BFE, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x7904, 0x4914, 0x108}},
    {35587, 2, {0x2000, 0x11FC, 0x104, 0xFDFC, 0x104, 0x79FC, 0x20, 0x7920, 0x1FE, 0x7A20, 0x4820, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35588, 2, {0x248, 0x7950, 0x4BF8, 0x4880, 0x4FFC, 0x7910, 0x4A48, 0x4FFC, 0x4802, 0x7BF8, 0x4800, 0x4BF8, 0x4800, 0x4BF8, 0x4A08, 0x9BF8}},
    {35589, 2, {0x2040, 0x1040, 0x10FE, 0xFD22, 0xAA, 0x7CFA, 0x42, 0x7C94, 0x148, 0x7CFE, 0x4522, 0x46AA, 0x44FA, 0x7C42, 0x4494, 0x108}},
    {35590, 2, {0x2020, 0x1010, 0x1FE, 0xFD02, 0x102, 0x79FE, 0x100, 0x79EE, 0x122, 0x79AA, 0x4966, 0x4922, 0x4966, 0x7AAA, 0x4A22, 0x466}},
    {35591, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x920, 0x1FF0, 0x2008, 0xDFF6, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35592, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x7FFC, 0x820, 0xFFFE, 0xA20, 0x3118, 0xDFF6, 0x0, 0xFE0, 0x820, 0xFE0}},
    {35593, 2, {0x2020, 0x1040, 0x3FC, 0xFE24, 0x224, 0x7BFC, 0x224, 0x7A44, 0x3FC, 0x7840, 0x48A8, 0x48B4, 0x493C, 0x7922, 0x4A22, 0x41E}},
    {35594, 2, {0x2088, 0x1088, 0x3FE, 0xFCA8, 0x10, 0x7BFE, 0x80, 0x7880, 0xFC, 0x7800, 0x48A8, 0x48A8, 0x48A8, 0x78AA, 0x492A, 0x206}},
    {35595, 2, {0x4208, 0x2208, 0x3BE, 0xF488, 0xAA8, 0x713E, 0x208, 0x7448, 0x848, 0x77FC, 0x50E0, 0x5150, 0x5248, 0x7C46, 0x5040, 0x40}},
    {35596, 2, {0x2000, 0x13FE, 0x8, 0xFDE8, 0x128, 0x79E8, 0x0, 0x7BFE, 0x8, 0x79E8, 0x4928, 0x4928, 0x49E8, 0x7808, 0x4828, 0x10}},
    {35597, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4102, 0xBFF4, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35598, 2, {0x2010, 0x1094, 0x254, 0xFD58, 0x110, 0x787E, 0x10, 0x7B30, 0x138, 0x7954, 0x4994, 0x4910, 0x4910, 0x7A80, 0x4C7E, 0x0}},
    {35599, 2, {0x20A0, 0x112C, 0x1124, 0xFD24, 0x1AC, 0x7D24, 0x124, 0x7DFC, 0x20, 0x7DFC, 0x4488, 0x4450, 0x4420, 0x7C50, 0x4488, 0x306}},
    {35600, 2, {0x4040, 0x2028, 0x90, 0xFAA4, 0x2C2, 0x748A, 0x188, 0x7678, 0x0, 0x73FC, 0x5294, 0x5294, 0x5294, 0x7294, 0x57FE, 0x0}},
    {35601, 2, {0x201E, 0x13E0, 0x122, 0xFC94, 0x40, 0x7888, 0x1F0, 0x7820, 0xC4, 0x79FE, 0x4822, 0x4820, 0x4BFE, 0x7850, 0x4888, 0x306}},
    {35602, 2, {0x2020, 0x1020, 0x50, 0xFC48, 0xA4, 0x79FE, 0x284, 0x78FC, 0x84, 0x78FC, 0x4880, 0x48FC, 0x4944, 0x7944, 0x4A7C, 0x44}},
    {35603, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7904, 0x1FC, 0x7904, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x4BFE, 0x7888, 0x4904, 0x202}},
    {35604, 2, {0x4040, 0x207C, 0x40, 0xFBFE, 0x242, 0x7278, 0x3C0, 0x7244, 0x23C, 0x7200, 0x52FC, 0x5280, 0x53FE, 0x7480, 0x5480, 0x8FC}},
    {35605, 2, {0x4004, 0x201E, 0x3F0, 0xFA1E, 0x210, 0x72FE, 0x292, 0x7298, 0x2F2, 0x728E, 0x5280, 0x52B8, 0x52A8, 0x74AA, 0x554A, 0xA86}},
    {35606, 2, {0x2000, 0x11FC, 0x124, 0xFDFC, 0x124, 0x79FC, 0x88, 0x7944, 0x242, 0x78F8, 0x4988, 0x4A50, 0x4820, 0x7850, 0x4988, 0x606}},
    {35607, 2, {0x2040, 0x1020, 0x3FC, 0xFD08, 0x90, 0x7BFE, 0x202, 0x7C44, 0x20, 0x7BFC, 0x4880, 0x48F8, 0x4888, 0x7908, 0x4928, 0x210}},
    {35608, 2, {0x2000, 0x13FC, 0x204, 0xFA04, 0x3FC, 0x7A40, 0x220, 0x7BFC, 0x288, 0x7A50, 0x4BFE, 0x4A20, 0x4DFC, 0x7C20, 0x4820, 0x20}},
    {35609, 2, {0x2088, 0x1050, 0x0, 0xFDFE, 0x50, 0x79FC, 0x54, 0x7BFE, 0x54, 0x79FC, 0x4850, 0x48D8, 0x4954, 0x7A52, 0x4850, 0x50}},
    {35610, 2, {0x2104, 0x1084, 0x88, 0xFC00, 0x3FE, 0x7800, 0x88, 0x7904, 0x202, 0x79FC, 0x4954, 0x4954, 0x4954, 0x7954, 0x47FE, 0x0}},
    {35611, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0x1FC, 0x7888, 0x3FE, 0x7820, 0x1FC, 0x7924, 0x49FC, 0x4924, 0x4BFE, 0x7904, 0x4914, 0x108}},
    {35612, 2, {0x2000, 0x11FE, 0x110, 0xFD20, 0x17C, 0x7944, 0x17C, 0x7944, 0x17C, 0x7910, 0x4910, 0x4954, 0x4952, 0x7A92, 0x4A50, 0x420}},
    {35613, 2, {0x4084, 0x2104, 0x3C4, 0xFA44, 0x3DE, 0x7244, 0x3C4, 0x7264, 0x254, 0x77D4, 0x50C4, 0x5144, 0x5244, 0x7444, 0x5154, 0x88}},
    {35614, 2, {0x2040, 0x1020, 0x3FE, 0xFC00, 0x1FC, 0x7904, 0x1FC, 0x7800, 0x3FE, 0x7A02, 0x4AFA, 0x4A8A, 0x4AFA, 0x7A02, 0x4A0A, 0x204}},
    {35615, 2, {0x203C, 0x13C0, 0x44, 0xFA24, 0x128, 0x7900, 0x40, 0x7B9C, 0x204, 0x7A04, 0x4B9C, 0x4A04, 0x4A04, 0x7A04, 0x4BFC, 0x204}},
    {35616, 2, {0x2040, 0x1080, 0x1FC, 0xFA48, 0x130, 0x78C0, 0x300, 0x79FC, 0x220, 0x7820, 0x4BFE, 0x4820, 0x4924, 0x7924, 0x49FC, 0x4}},
    {35617, 2, {0x2008, 0x103C, 0x3C0, 0xF804, 0x244, 0x7928, 0x1FC, 0x7A20, 0x20, 0x7BFE, 0x4820, 0x4924, 0x4924, 0x7924, 0x49FC, 0x4}},
    {35618, 2, {0x20A0, 0x1090, 0x1FE, 0xFB20, 0x1FC, 0x7920, 0x1FC, 0x7920, 0x1FE, 0x7900, 0x4BF8, 0x4908, 0x4890, 0x7860, 0x4998, 0x606}},
    {35619, 2, {0x21FC, 0x1020, 0x3FE, 0xFE22, 0x1AC, 0x7820, 0x1AC, 0x7800, 0x1FC, 0x7800, 0x4BFE, 0x4880, 0x48FC, 0x7804, 0x4828, 0x10}},
    {35620, 2, {0x2000, 0x13FE, 0x50, 0xFDFC, 0x154, 0x7954, 0x1FC, 0x7800, 0x1FC, 0x7800, 0x4BFE, 0x4820, 0x48A8, 0x7924, 0x4AA2, 0x40}},
    {35621, 2, {0x2020, 0x1040, 0x1FC, 0xFD24, 0x17C, 0x798C, 0x154, 0x7924, 0x154, 0x79FC, 0x4820, 0x4810, 0x4954, 0x7942, 0x494A, 0x238}},
    {35622, 2, {0x878, 0xFF48, 0x886, 0x3E78, 0x2A48, 0x3E30, 0x4148, 0xFFFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35623, 2, {0x2040, 0x1020, 0x3FE, 0xFC00, 0x154, 0x7924, 0x154, 0x79FC, 0x20, 0x7BFE, 0x4A42, 0x4A92, 0x4AFA, 0x7A0A, 0x4A02, 0x206}},
    {35624, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0x0, 0x79FC, 0x104, 0x79FC, 0x104, 0x79FC, 0x4820, 0x4BFE, 0x4850, 0x7888, 0x4904, 0x202}},
    {35625, 2, {0x440, 0x7FFC, 0x440, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x400, 0xFFFE, 0x1110, 0x2FE8, 0xC006, 0xFE0, 0x820, 0xFE0}},
    {35626, 2, {0x2040, 0x1020, 0x3FE, 0xFC88, 0x50, 0x7BFE, 0x252, 0x7A8A, 0x306, 0x7AFA, 0x4A8A, 0x4A8A, 0x4AFA, 0x7A02, 0x4A0A, 0x204}},
    {35627, 2, {0x2040, 0x1020, 0x3FE, 0xFC88, 0x50, 0x7BFE, 0x222, 0x7AFA, 0x222, 0x7AFA, 0x4A8A, 0x4A8A, 0x4AFA, 0x7A02, 0x4A0A, 0x204}},
    {35628, 2, {0x2000, 0x13DE, 0x42, 0xFA52, 0x14A, 0x7A52, 0x28, 0x78C4, 0x312, 0x7860, 0x4988, 0x4832, 0x49C4, 0x7818, 0x4860, 0x380}},
    {35629, 2, {0x2108, 0x1090, 0x7FE, 0xF800, 0x3C4, 0x7A54, 0x3D4, 0x7A54, 0x3D4, 0x7A44, 0x4ACC, 0x4800, 0x4BFC, 0x7884, 0x4914, 0x608}},
    {35630, 2, {0x2020, 0x13FE, 0x20, 0xFDFC, 0x20, 0x7BFE, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {35631, 2, {0x2020, 0x103C, 0x20, 0xFDFE, 0x122, 0x7938, 0x1E4, 0x791C, 0x100, 0x797C, 0x4944, 0x497C, 0x4944, 0x797C, 0x4944, 0x2FE}},
    {35632, 2, {0x2010, 0x1010, 0x2FE, 0xFD10, 0x17C, 0x7854, 0x7C, 0x7B54, 0x17C, 0x7910, 0x49FE, 0x4910, 0x4910, 0x7A90, 0x447E, 0x0}},
    {35633, 2, {0x2020, 0x11FC, 0x124, 0xFBFE, 0x124, 0x79FC, 0x20, 0x79FC, 0x124, 0x79FC, 0x4840, 0x4BFE, 0x4888, 0x79D0, 0x4870, 0x38C}},
    {35634, 2, {0x2040, 0x1088, 0x1FC, 0xFD08, 0x252, 0x7BFE, 0x50, 0x7988, 0x626, 0x78C0, 0x4B10, 0x4864, 0x4B88, 0x7830, 0x48C0, 0x700}},
    {35635, 2, {0x2000, 0x13FE, 0x200, 0xFE7C, 0x244, 0x7A44, 0x27C, 0x7A00, 0x2EE, 0x7AAA, 0x4AAA, 0x4AAA, 0x4AEE, 0x7A00, 0x4BFE, 0x0}},
    {35636, 2, {0x2000, 0x11FC, 0x124, 0xFBFE, 0x124, 0x79FC, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {35637, 2, {0x2000, 0x13DE, 0x42, 0xFA52, 0x14A, 0x7A52, 0x42, 0x7820, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35638, 2, {0x2040, 0x1020, 0x1FE, 0xFD00, 0x148, 0x7948, 0x1FE, 0x7948, 0x148, 0x7948, 0x4978, 0x4900, 0x4AD4, 0x7AAA, 0x452A, 0x800}},
    {35639, 2, {0x820, 0x7E20, 0x87E, 0xFEC4, 0x1028, 0x1E10, 0x2228, 0x46C6, 0x8100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35640, 2, {0x4108, 0x2108, 0x108, 0xF7D0, 0x11E, 0x7794, 0x124, 0x7FD4, 0x214, 0x73D4, 0x5254, 0x5248, 0x5248, 0x74D4, 0x5424, 0x842}},
    {35641, 2, {0x2088, 0x1088, 0x3FE, 0xFC88, 0xF8, 0x7820, 0x1FC, 0x7924, 0x1FC, 0x7820, 0x4BFE, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x0}},
    {35642, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x982, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35643, 2, {0x4010, 0x2190, 0x61E, 0xFA22, 0x254, 0x7208, 0xF90, 0x7264, 0x208, 0x771E, 0x5AA2, 0x5242, 0x5214, 0x7208, 0x5210, 0x260}},
    {35644, 2, {0x2040, 0x107C, 0x40, 0xFDFC, 0x144, 0x7978, 0x1C4, 0x793C, 0x104, 0x7978, 0x4994, 0x4958, 0x4AFE, 0x7A10, 0x4450, 0x20}},
    {35645, 2, {0x7CF8, 0x488, 0x7CF8, 0x4020, 0x7DFC, 0x524, 0x15FC, 0xA22, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {35646, 2, {0x2000, 0x11F8, 0x108, 0xFDF8, 0x108, 0x79F8, 0x0, 0x7BFC, 0x294, 0x7BFC, 0x4800, 0x49F8, 0x4890, 0x7860, 0x4998, 0x606}},
    {35647, 2, {0x4100, 0x211E, 0x7D2, 0xF912, 0x112, 0x77DE, 0x452, 0x77D2, 0x452, 0x77DE, 0x5112, 0x5112, 0x5FD2, 0x7122, 0x512A, 0x144}},
    {35648, 2, {0x2008, 0x1388, 0x88, 0xF910, 0x7DE, 0x7A94, 0x2A4, 0x7B94, 0x294, 0x7B94, 0x4A94, 0x4AC8, 0x4B88, 0x7E94, 0x48A4, 0xC2}},
    {35649, 2, {0x2088, 0x13FE, 0x88, 0xFC00, 0x1FC, 0x78A8, 0xA8, 0x7BFE, 0xA8, 0x78A8, 0x49FC, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20}},
    {35650, 2, {0x2000, 0x13DE, 0x252, 0xFE52, 0x3DE, 0x7800, 0x1FC, 0x7924, 0x1FC, 0x7924, 0x49FC, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20}},
    {35651, 2, {0x4040, 0x207C, 0x40, 0xFBFE, 0x242, 0x7278, 0x3C4, 0x723C, 0x200, 0x7250, 0x5254, 0x5354, 0x54D8, 0x7450, 0x59FE, 0x0}},
    {35652, 2, {0x2104, 0x1088, 0x0, 0xFBFE, 0x222, 0x7AAA, 0x272, 0x7A22, 0x3FE, 0x7800, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x104}},
    {35653, 2, {0x4000, 0x27DE, 0x14A, 0xFB5A, 0x56A, 0x7252, 0x2D6, 0x7420, 0x84, 0x7084, 0x52D4, 0x5296, 0x5294, 0x72D4, 0x573E, 0x0}},
    {35654, 2, {0x2020, 0x13FE, 0x20, 0xFDFC, 0x0, 0x79FC, 0x104, 0x79FC, 0x88, 0x7BFE, 0x4800, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35655, 2, {0x2020, 0x1020, 0x3FE, 0xFC88, 0x124, 0x7A22, 0x1FC, 0x7828, 0x3FE, 0x7840, 0x49FC, 0x4E84, 0x48FC, 0x7884, 0x48FC, 0x84}},
    {35656, 2, {0x4208, 0x2108, 0x7C8, 0xF810, 0x79E, 0x7494, 0x7A4, 0x7014, 0x794, 0x7094, 0x5114, 0x51C8, 0x5708, 0x7114, 0x5524, 0x242}},
    {35657, 2, {0x2020, 0x17A4, 0xA8, 0xFA92, 0x114, 0x7A08, 0x5F4, 0x7802, 0x3F8, 0x7A08, 0x4A08, 0x4BF8, 0x4A08, 0x7910, 0x4FFE, 0x0}},
    {35658, 2, {0x2020, 0x1020, 0x1FC, 0xFC20, 0x3FE, 0x7908, 0x39C, 0x7908, 0x188, 0x7E3E, 0x4800, 0x4BFE, 0x4890, 0x7890, 0x4912, 0x20E}},
    {35659, 2, {0x239C, 0x1294, 0x39C, 0xFE94, 0x39C, 0x7A04, 0x2F4, 0x7A94, 0x2F4, 0x7A94, 0x4AF4, 0x4A94, 0x4A94, 0x7B34, 0x4A04, 0x20C}},
    {35660, 2, {0x400E, 0x23F0, 0x44, 0xFA24, 0x108, 0x73F8, 0x208, 0x73FC, 0x204, 0x73FE, 0x5202, 0x52AA, 0x52AA, 0x7502, 0x5414, 0x808}},
    {35661, 2, {0x80, 0x3FFE, 0x2290, 0x25FE, 0x2E90, 0x34FE, 0x2490, 0x24FE, 0x2440, 0x2FFE, 0x2000, 0x27FC, 0x4000, 0x47FC, 0x8404, 0x7FC}},
    {35662, 2, {0x2000, 0x13FC, 0x48, 0xFC30, 0x3FE, 0x7852, 0x94, 0x7B50, 0x20, 0x7BFE, 0x4A52, 0x4A8A, 0x4B76, 0x7A52, 0x4A72, 0x206}},
    {35663, 2, {0x2124, 0x1124, 0x2AA, 0xFBAE, 0x124, 0x7AAA, 0x3AE, 0x7924, 0x7FE, 0x7910, 0x4914, 0x4914, 0x4A8A, 0x7A4A, 0x4A16, 0x422}},
    {35664, 2, {0x2088, 0x1050, 0x3FE, 0xFC50, 0x1FC, 0x7954, 0x18C, 0x7974, 0x104, 0x79FC, 0x4808, 0x4BFE, 0x4908, 0x7888, 0x48A8, 0x10}},
    {35665, 2, {0x2018, 0x11E0, 0x40, 0xFBFE, 0x88, 0x7974, 0x252, 0x7870, 0x0, 0x79FC, 0x4904, 0x4974, 0x4954, 0x7974, 0x4904, 0x10C}},
    {35666, 2, {0x203C, 0x13E0, 0x124, 0xFCA8, 0x3FE, 0x78A8, 0x124, 0x7A02, 0x1FC, 0x7924, 0x4924, 0x49FC, 0x4924, 0x7924, 0x49FC, 0x104}},
    {35667, 2, {0x2020, 0x1020, 0x3FE, 0xFC20, 0x1FC, 0x7924, 0x1FC, 0x7924, 0x1FC, 0x7822, 0x4BFE, 0x4842, 0x4824, 0x7AA2, 0x4A8A, 0x478}},
    {35668, 2, {0x2000, 0x13DE, 0x252, 0xFBDE, 0x210, 0x7A52, 0x1CE, 0x7800, 0x88, 0x7BFE, 0x4888, 0x4888, 0x4BFE, 0x7888, 0x4904, 0x202}},
    {35669, 2, {0x2080, 0x1080, 0x1FE, 0xFB54, 0x554, 0x7954, 0x3FE, 0x7954, 0x154, 0x7954, 0x4FFE, 0x4800, 0x4954, 0x792A, 0x4A2A, 0x0}},
    {35670, 2, {0x2000, 0x13DE, 0x88, 0xFAA8, 0x3FE, 0x7988, 0x2DA, 0x7CA6, 0x1F8, 0x7908, 0x4908, 0x49F8, 0x4908, 0x7908, 0x49F8, 0x108}},
    {35671, 2, {0x2210, 0x1210, 0x3DE, 0xFA28, 0x544, 0x78A0, 0x110, 0x7A08, 0x5F6, 0x7800, 0x4800, 0x4BF8, 0x4A08, 0x7A08, 0x4BF8, 0x208}},
    {35672, 2, {0x2110, 0x1094, 0x3D2, 0xF812, 0x250, 0x7990, 0x7FE, 0x7810, 0x3D2, 0x7A52, 0x4A54, 0x4BD4, 0x4A4A, 0x7A4A, 0x4BD6, 0x22}},
    {35673, 2, {0x2090, 0x1088, 0x1FE, 0xFD10, 0x310, 0x7DFE, 0x110, 0x7910, 0x1FE, 0x7910, 0x4910, 0x49FE, 0x4900, 0x7AA4, 0x4A52, 0x452}},
    {35674, 2, {0x2000, 0x13FE, 0x50, 0xFBFE, 0x252, 0x7BFE, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4820, 0x7BFE, 0x4820, 0x20}},
    {35675, 2, {0x2088, 0x1088, 0x3DE, 0xF888, 0x7FE, 0x7888, 0x154, 0x7A22, 0x1FC, 0x7904, 0x4904, 0x49FC, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35676, 2, {0x2108, 0x1090, 0x3FC, 0xF890, 0x294, 0x7998, 0x90, 0x7BFE, 0x0, 0x79F8, 0x4908, 0x4908, 0x49F8, 0x7908, 0x4908, 0x1F8}},
    {35677, 2, {0x4000, 0x27FC, 0x4A4, 0xFCA4, 0x7BC, 0x70A0, 0x7BC, 0x74A4, 0x4A4, 0x77BC, 0x54A4, 0x54A4, 0x57BC, 0x74A2, 0x50A2, 0x7E}},
    {35678, 2, {0x2000, 0x11FC, 0x154, 0xFD54, 0x1FC, 0x7800, 0x3FE, 0x7800, 0x1FC, 0x7904, 0x49FC, 0x4862, 0x4894, 0x7988, 0x4AA4, 0xC2}},
    {35679, 2, {0x20F8, 0x1088, 0x88, 0xFCF8, 0x0, 0x79DC, 0x154, 0x7954, 0x1DC, 0x7820, 0x4BFE, 0x4870, 0x48A8, 0x7924, 0x4A22, 0x20}},
    {35680, 2, {0x2020, 0x17FE, 0x0, 0xFBFE, 0x202, 0x7AFA, 0x28A, 0x7BFE, 0x0, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4800, 0x7FE}},
    {35681, 2, {0x2020, 0x1124, 0xA8, 0xFBFE, 0x202, 0x78F8, 0x88, 0x7888, 0xF8, 0x7800, 0x49FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x104}},
    {35682, 2, {0x2044, 0x1224, 0x128, 0xFC7E, 0x10, 0x7828, 0x34A, 0x791C, 0x128, 0x794C, 0x491A, 0x492A, 0x4948, 0x7910, 0x4AFE, 0x400}},
    {35683, 2, {0x2020, 0x1020, 0x50, 0xFC88, 0x104, 0x7AFA, 0x0, 0x7800, 0x1DC, 0x7954, 0x4954, 0x49DC, 0x4888, 0x7888, 0x4954, 0x222}},
    {35684, 2, {0x4108, 0x2208, 0x7C8, 0xF450, 0x7DE, 0x7464, 0x7D4, 0x7214, 0x114, 0x77D4, 0x5214, 0x53C8, 0x5248, 0x7454, 0x5554, 0x8A2}},
    {35685, 2, {0x1020, 0x7E20, 0x4A7E, 0x7EC4, 0x1028, 0xFF10, 0x2028, 0x3EC6, 0x4300, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35686, 2, {0x2420, 0xFF20, 0x247E, 0x7EC4, 0x8228, 0x7A10, 0x4A28, 0x7AC6, 0x500, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35687, 2, {0x4020, 0x23FE, 0x288, 0xFA50, 0x3FE, 0x7250, 0x2FC, 0x7254, 0x3FE, 0x7254, 0x52FC, 0x5250, 0x52D8, 0x7554, 0x5652, 0x850}},
    {35688, 2, {0x4090, 0x23FC, 0x294, 0xFBFC, 0x294, 0x73FC, 0x0, 0x73FC, 0x200, 0x72F8, 0x5200, 0x53FE, 0x5520, 0x7514, 0x5548, 0x986}},
    {35689, 2, {0x2040, 0x1020, 0x1FC, 0xFC88, 0x50, 0x7BFE, 0x0, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4820, 0x7A94, 0x4A8A, 0x47A}},
    {35690, 2, {0x2088, 0x13FE, 0x88, 0xFDFC, 0x104, 0x79FC, 0x104, 0x79FC, 0x80, 0x79FE, 0x4A22, 0x4D52, 0x4902, 0x79FA, 0x480A, 0x4}},
    {35691, 2, {0x2080, 0x10F8, 0x108, 0xFBFE, 0x144, 0x7992, 0x1FE, 0x7900, 0x17C, 0x7900, 0x497C, 0x4900, 0x497C, 0x7A44, 0x4A7C, 0x444}},
    {35692, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35693, 2, {0x4100, 0x26DC, 0x454, 0xFC54, 0x6D4, 0x7454, 0x466, 0x77C0, 0x1C, 0x7014, 0x57D4, 0x5114, 0x5108, 0x71C8, 0x5714, 0x222}},
    {35694, 2, {0x2020, 0x1050, 0x1088, 0xFD74, 0x202, 0x7DFC, 0x124, 0x7DAC, 0x124, 0x7DFC, 0x4400, 0x44F8, 0x4488, 0x7CF8, 0x4488, 0xF8}},
    {35695, 2, {0x2000, 0x11FC, 0x154, 0xFDFC, 0x20, 0x79FC, 0x20, 0x7BFE, 0x88, 0x7850, 0x49FC, 0x4820, 0x4BFE, 0x7820, 0x4820, 0x20}},
    {35696, 2, {0x2104, 0x1088, 0x3FE, 0xFC20, 0x1FC, 0x7820, 0x3FE, 0x7854, 0x192, 0x7890, 0x4BFE, 0x4890, 0x48D4, 0x7B8A, 0x4896, 0x1A2}},
    {35697, 2, {0x820, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1110, 0x7D7C, 0x100, 0x7D7C, 0x100, 0x7D7C, 0x100, 0x7D7C, 0x4544, 0x7D7C}},
    {35698, 2, {0x4080, 0x2040, 0x7FE, 0xF910, 0x208, 0x7514, 0x3F8, 0x7110, 0x3F8, 0x7110, 0x57FC, 0x5128, 0x5310, 0x7D48, 0x5186, 0x100}},
    {35699, 2, {0x2000, 0x11FC, 0x20, 0xFBFE, 0x222, 0x79AC, 0x20, 0x79AC, 0x0, 0x7BFE, 0x4820, 0x49FC, 0x4954, 0x7954, 0x4954, 0x10C}},
    {35700, 2, {0x2010, 0x107C, 0x254, 0xFD7C, 0x110, 0x78FE, 0x0, 0x7B7C, 0x144, 0x797C, 0x4940, 0x497C, 0x4944, 0x797C, 0x4A80, 0x47E}},
    {35701, 2, {0x4144, 0x2144, 0x554, 0xFB64, 0x15E, 0x77E4, 0x4, 0x7234, 0x14C, 0x77E4, 0x5084, 0x53E4, 0x5084, 0x70E4, 0x5714, 0x8}},
    {35702, 2, {0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x810, 0xFEFE, 0x0, 0x7CFC, 0x0, 0x7CFC, 0x4484, 0x7CFC}},
    {35703, 2, {0x2088, 0x13FE, 0xA8, 0xFC90, 0x1FE, 0x7B20, 0x1FC, 0x7920, 0x1FC, 0x7920, 0x49FE, 0x4900, 0x4BFC, 0x7888, 0x4870, 0x38E}},
    {35704, 2, {0x2020, 0x13FE, 0x20, 0xFDFC, 0x0, 0x7BFE, 0x2, 0x79FC, 0x20, 0x7BFE, 0x4800, 0x4BFE, 0x4804, 0x7BBE, 0x4AA4, 0x38C}},
    {35705, 2, {0x2020, 0x13FE, 0x0, 0xFDFC, 0x104, 0x79FC, 0x0, 0x7BFE, 0x202, 0x79FC, 0x4840, 0x4BA4, 0x48D8, 0x7B34, 0x48D2, 0x330}},
    {35706, 2, {0x2200, 0x12BE, 0x302, 0xFA54, 0x1C8, 0x783E, 0x20A, 0x7BCA, 0x528, 0x7928, 0x4FEE, 0x4928, 0x4AA8, 0x7A58, 0x4C4E, 0x80}},
    {35707, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2288, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35708, 2, {0x7E40, 0x487E, 0x7E90, 0x4308, 0x7EFE, 0x48AA, 0x7EFE, 0x200, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35709, 2, {0xA00, 0x33B8, 0x2208, 0x3BB8, 0x2088, 0x3AB8, 0x2288, 0xFFFE, 0x1110, 0x3FF8, 0x4004, 0x9FF2, 0x0, 0x1FF0, 0x1010, 0x1FF0}},
    {35710, 2, {0x2088, 0x13FE, 0x0, 0xFDE2, 0x12A, 0x79EA, 0x12A, 0x79EA, 0x122, 0x7966, 0x4800, 0x4BDE, 0x4A52, 0x794A, 0x4A52, 0xC6}},
    {35711, 2, {0x2088, 0x13DE, 0x88, 0xFDDC, 0x88, 0x7BDE, 0x88, 0x7BFC, 0x4, 0x79FC, 0x4804, 0x4BFC, 0x4840, 0x7AA4, 0x4A8A, 0x47A}},
    {35712, 2, {0x2020, 0x13FE, 0x20, 0xFDFC, 0x0, 0x7BFE, 0x252, 0x7BFE, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {35713, 2, {0x4020, 0x2010, 0x4FE, 0xFA44, 0x228, 0x70FE, 0x92, 0x76FE, 0x292, 0x72BA, 0x52AA, 0x52BA, 0x5286, 0x7500, 0x58FE, 0x0}},
    {35714, 2, {0x2100, 0x11F8, 0x208, 0xFFFE, 0x262, 0x7A92, 0x1FC, 0x7904, 0x1FC, 0x7904, 0x49FC, 0x4880, 0x49F8, 0x7A88, 0x4870, 0x38E}},
    {35715, 2, {0x2088, 0x11DC, 0x88, 0xFBDE, 0x88, 0x7954, 0x222, 0x79FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {35716, 2, {0x83F8, 0x4248, 0x248, 0xE3F8, 0x248, 0xE248, 0x3F8, 0xE000, 0xFBE, 0xEAAA, 0xAAAA, 0xAFBE, 0xAAAA, 0xEAAA, 0xAFBE, 0x8A2}},
    {35717, 2, {0x2020, 0x13FE, 0x202, 0xFC1C, 0x3E0, 0x7924, 0xA8, 0x7BFE, 0x124, 0x7A02, 0x4DFC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x104}},
    {35718, 2, {0x2200, 0x4F9C, 0x9480, 0x2FDE, 0x6108, 0xAFC8, 0x2508, 0x27D8, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35719, 2, {0x23DE, 0x1252, 0x3DE, 0xFE52, 0x3DE, 0x7A42, 0x27A, 0x7A8A, 0x312, 0x7ADA, 0x4A8A, 0x4ADA, 0x4A8A, 0x7AFA, 0x4A0A, 0x204}},
    {35720, 2, {0x4000, 0x27FE, 0x444, 0xFD98, 0x488, 0x75FE, 0x488, 0x75DC, 0x6AA, 0x7488, 0x5420, 0x5520, 0x553C, 0x7520, 0x5520, 0xBFE}},
    {35721, 2, {0x2010, 0x107C, 0x254, 0xFD7C, 0x110, 0x78FE, 0x44, 0x7B7C, 0x144, 0x797C, 0x4944, 0x497C, 0x4928, 0x7944, 0x4A80, 0x47E}},
    {35722, 2, {0x2104, 0x47C8, 0x8812, 0xF3BC, 0x2008, 0x4B92, 0xF83E, 0x380, 0xAAAA, 0xABAA, 0x800, 0x1FF8, 0x2820, 0x7C0, 0x1830, 0xE00E}},
    {35723, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2342, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35724, 2, {0x4110, 0x2110, 0x7FC, 0xF910, 0x1F0, 0x7208, 0x2EA, 0x76AC, 0x2A8, 0x72AA, 0x56EA, 0x5A06, 0x5000, 0x72A4, 0x5252, 0x452}},
    {35725, 2, {0x2000, 0x13FE, 0x20, 0xFBAE, 0x2AA, 0x7AAA, 0x3AE, 0x7820, 0x3FE, 0x7820, 0x4BAE, 0x4AAA, 0x4AAA, 0x7BAE, 0x4820, 0x3FE}},
    {35726, 2, {0x3218, 0x2914, 0x4020, 0x7BBE, 0xD068, 0x53A8, 0x783E, 0x53A8, 0x5028, 0x7BBE, 0x52A8, 0x52A8, 0x52A8, 0x7BBE, 0x42A0, 0x4020}},
    {35727, 2, {0x2200, 0x4F9C, 0x9480, 0x2FDE, 0x6208, 0xAFC8, 0x2A48, 0x2AD8, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35728, 2, {0x4890, 0x7EFC, 0xC990, 0x7EFC, 0x4890, 0x7EFC, 0x4890, 0x7EFC, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {35729, 2, {0x8040, 0x40A0, 0x110, 0xE208, 0xDF6, 0xE000, 0xEEE, 0xEAAA, 0xEEE, 0xE000, 0xAFFE, 0xA922, 0xAFFE, 0xE922, 0xA92A, 0x804}},
    {35730, 2, {0x21F0, 0x1210, 0x7FC, 0xFA04, 0x3FC, 0x7A24, 0x3B8, 0x7A22, 0x1FE, 0x7A10, 0x4FFC, 0x4A44, 0x4BFC, 0x78D0, 0x494A, 0x63E}},
    {35731, 2, {0x2020, 0x13FE, 0x0, 0xFDDC, 0x154, 0x79DC, 0x88, 0x7BFE, 0x88, 0x7BFE, 0x4888, 0x4BFE, 0x4894, 0x7988, 0x4AA4, 0xC2}},
    {35732, 2, {0x400E, 0x2770, 0x52A, 0xFD14, 0x53E, 0x7608, 0x57E, 0x7500, 0x57E, 0x7502, 0x553E, 0x5602, 0x547E, 0x7414, 0x54AA, 0x4BA}},
    {35733, 2, {0x23DE, 0x1252, 0x3DE, 0xFE52, 0x3DE, 0x7A22, 0x2FA, 0x7A22, 0x2FA, 0x7AAA, 0x4AFA, 0x4AAA, 0x4AFA, 0x7A72, 0x4AAA, 0x226}},
    {35734, 2, {0x2148, 0x114C, 0x2AA, 0xFC08, 0x3FE, 0x7948, 0x368, 0x794A, 0x36A, 0x794A, 0x4B6C, 0x494C, 0x496A, 0x7B8A, 0x4816, 0x22}},
    {35735, 2, {0x2222, 0x13FE, 0x90, 0xF9FE, 0x310, 0x7DFE, 0x110, 0x79FE, 0x110, 0x79FE, 0x4900, 0x4BFE, 0x4A8A, 0x7B76, 0x4A52, 0x276}},
    {35736, 2, {0x43FC, 0x2108, 0x1F8, 0xF908, 0x1F8, 0x710E, 0x7F8, 0x7008, 0x7FE, 0x7294, 0x539C, 0x5294, 0x539C, 0x72D6, 0x57BC, 0x84}},
    {35737, 2, {0x2088, 0x13FE, 0x88, 0xFBDE, 0x252, 0x7BDE, 0xA0, 0x7890, 0x1FE, 0x7920, 0x4BFC, 0x4D20, 0x49FC, 0x7920, 0x49FE, 0x100}},
    {35738, 2, {0x4528, 0x27BE, 0x948, 0xF7BE, 0x318, 0x75AA, 0x946, 0x73FC, 0x204, 0x73FC, 0x5204, 0x53FC, 0x5204, 0x73FC, 0x5108, 0x204}},
    {35739, 2, {0x4108, 0x27FE, 0x108, 0xFB90, 0x13C, 0x77D4, 0x2B4, 0x755C, 0x396, 0x7122, 0x53FC, 0x5000, 0x57FE, 0x7080, 0x5104, 0x3FE}},
    {35740, 2, {0x4248, 0x2150, 0x7FC, 0xF404, 0x1F0, 0x7110, 0x7FC, 0x7554, 0x4E4, 0x77FC, 0x5040, 0x57FC, 0x5040, 0x7FFE, 0x52A4, 0x452}},
    {35741, 2, {0x23DE, 0x1252, 0x3DE, 0xFC00, 0x3FE, 0x7A00, 0x2C8, 0x7A4E, 0x3EA, 0x7AB2, 0x4AEA, 0x4AAA, 0x4AEA, 0x7AB4, 0x45E4, 0x82A}},
    {35742, 2, {0x8108, 0x41EC, 0x10A, 0xEFEA, 0x928, 0xEBC8, 0x93E, 0xEFE8, 0x948, 0xEBE8, 0xAA28, 0xAB68, 0xAAA8, 0xEBF4, 0xAAB4, 0x12A2}},
    {35743, 2, {0x8108, 0x4FE4, 0x100, 0xE7CE, 0x0, 0xEFEE, 0xAA0, 0xEFEE, 0x440, 0xE7CE, 0xA44A, 0xA7CA, 0xA44A, 0xE7CE, 0xA28A, 0x440}},
    {35744, 2, {0x0, 0x2000, 0x1000, 0x1000, 0x0, 0x0, 0xF000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1400, 0x1800, 0x1000, 0x0}},
    {35745, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x40, 0x40, 0xF7FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1440, 0x1840, 0x1040, 0x40}},
    {35746, 2, {0x0, 0x2000, 0x13FE, 0x1020, 0x20, 0x20, 0xF020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1420, 0x1820, 0x10A0, 0x40}},
    {35747, 2, {0x80, 0x2080, 0x1080, 0x1080, 0x80, 0xA0, 0xF090, 0x1088, 0x1084, 0x1084, 0x1080, 0x1080, 0x1480, 0x1880, 0x1080, 0x80}},
    {35748, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x40, 0x40, 0xF040, 0x1040, 0x10A0, 0x10A0, 0x10A0, 0x1510, 0x1910, 0x1208, 0x404, 0x802}},
    {35749, 2, {0x0, 0x21F0, 0x1110, 0x1110, 0x110, 0x110, 0xF110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1512, 0x1912, 0x1212, 0x20E, 0x400}},
    {35750, 2, {0x0, 0x2000, 0x13FC, 0x1040, 0x40, 0x40, 0xF040, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1040, 0x1440, 0x1840, 0x1040, 0x40}},
    {35751, 2, {0x0, 0x2000, 0x13FE, 0x1020, 0x20, 0x20, 0xF020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1420, 0x1820, 0x1020, 0x7FE, 0x0}},
    {35752, 2, {0x10, 0x2010, 0x1010, 0x1010, 0x3FE, 0x10, 0xF010, 0x1010, 0x1110, 0x1090, 0x1090, 0x1010, 0x1410, 0x1810, 0x1050, 0x20}},
    {35753, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x40, 0x40, 0xF07C, 0x1040, 0x1040, 0x1040, 0x1040, 0x1440, 0x1840, 0x1040, 0xFFE, 0x0}},
    {35754, 2, {0x20, 0x2020, 0x1020, 0x1020, 0x124, 0x124, 0xF124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1524, 0x19FC, 0x1004, 0x0}},
    {35755, 2, {0x80, 0x2080, 0x1100, 0x11FE, 0x200, 0x400, 0xF1F8, 0x1008, 0x1010, 0x1060, 0x1080, 0x1500, 0x1A02, 0x1202, 0x1FE, 0x0}},
    {35756, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x40, 0x40, 0xF040, 0x107E, 0x17C0, 0x1040, 0x1040, 0x1442, 0x1842, 0x1042, 0x3E, 0x0}},
    {35757, 2, {0x104, 0x2124, 0x1124, 0x1124, 0x124, 0x124, 0xF124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1524, 0x1924, 0x1224, 0x204, 0x404}},
    {35758, 2, {0x80, 0x2048, 0x1248, 0x1208, 0x208, 0x110, 0xF110, 0x1110, 0x10A0, 0x10A0, 0x1040, 0x1440, 0x18A0, 0x1110, 0x208, 0xC06}},
    {35759, 2, {0x0, 0x27F0, 0x1110, 0x1110, 0x110, 0x110, 0xF110, 0x17D0, 0x1110, 0x1110, 0x1110, 0x1112, 0x150A, 0x190A, 0x1106, 0x102}},
    {35760, 2, {0x0, 0x2000, 0x11F8, 0x1008, 0x8, 0x8, 0xF008, 0x11F8, 0x1108, 0x1100, 0x1100, 0x1100, 0x1502, 0x1902, 0x10FE, 0x0}},
    {35761, 2, {0x0, 0x2000, 0x13FC, 0x1044, 0x44, 0x244, 0xF244, 0x1244, 0x1484, 0x1084, 0x1084, 0x1504, 0x1904, 0x1204, 0x428, 0x810}},
    {35762, 2, {0x110, 0x2110, 0x1110, 0x1110, 0x7FC, 0x110, 0xF110, 0x1110, 0x1FFE, 0x1110, 0x1110, 0x1510, 0x1910, 0x1210, 0x210, 0x410}},
    {35763, 2, {0x40, 0x2040, 0x1040, 0x17FE, 0x40, 0x40, 0xF3FC, 0x1040, 0x1040, 0x17FE, 0x1042, 0x1042, 0x144A, 0x1844, 0x1040, 0x40}},
    {35764, 2, {0x0, 0x23FC, 0x1200, 0x1208, 0x288, 0x250, 0xF250, 0x1220, 0x1220, 0x1250, 0x1250, 0x1288, 0x1708, 0x1A00, 0x13FE, 0x0}},
    {35765, 2, {0x0, 0x23FE, 0x1200, 0x1200, 0x200, 0x3FC, 0xF204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1200, 0x1600, 0x1A00, 0x13FE, 0x0}},
    {35766, 2, {0x0, 0x23FC, 0x1010, 0x1110, 0x110, 0x110, 0xF210, 0x13FE, 0x1030, 0x1050, 0x1090, 0x1510, 0x1A10, 0x1410, 0x50, 0x20}},
    {35767, 2, {0x20, 0x2020, 0x1020, 0x1020, 0x3FE, 0x222, 0xF222, 0x1222, 0x1252, 0x124A, 0x128A, 0x1302, 0x1602, 0x1A02, 0x120A, 0x204}},
    {35768, 2, {0x80, 0x2080, 0x1080, 0x11FC, 0x120, 0x220, 0xF020, 0x1020, 0x13FE, 0x1020, 0x1020, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {35769, 2, {0xA0, 0x20A0, 0x10A0, 0x1124, 0x124, 0x328, 0xF530, 0x1120, 0x1160, 0x11A0, 0x1120, 0x1122, 0x1522, 0x1922, 0x111E, 0x100}},
    {35770, 2, {0x40, 0x2040, 0x10A0, 0x10A0, 0x110, 0x208, 0xF406, 0x1110, 0x1120, 0x1140, 0x1180, 0x1100, 0x1504, 0x1904, 0x10FC, 0x0}},
    {35771, 2, {0x0, 0x2008, 0x1108, 0x1088, 0x252, 0x252, 0xF222, 0x1222, 0x1252, 0x1292, 0x130A, 0x160A, 0x1A02, 0x13FE, 0x2, 0x0}},
    {35772, 2, {0x20, 0x2120, 0x1120, 0x1110, 0x110, 0x248, 0xF248, 0x1444, 0x1882, 0x1080, 0x1110, 0x1508, 0x1A08, 0x17FC, 0x204, 0x0}},
    {35773, 2, {0x0, 0x43F8, 0x2208, 0x2208, 0x328, 0x2A8, 0xE2A8, 0x2248, 0x2248, 0x22A8, 0x22A8, 0x2B2A, 0x340A, 0x240A, 0x806, 0x1002}},
    {35774, 2, {0x0, 0x21F0, 0x1110, 0x1110, 0x110, 0x20E, 0xF400, 0x13F8, 0x1108, 0x1110, 0x1090, 0x14A0, 0x1840, 0x10A0, 0x318, 0xC06}},
    {35775, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x100, 0x100, 0xF100, 0x11F8, 0x1108, 0x1108, 0x1108, 0x1508, 0x1A08, 0x1208, 0x450, 0x820}},
    {35776, 2, {0x40, 0x2040, 0x1040, 0x13FC, 0x44, 0x44, 0xF044, 0x1044, 0x17FE, 0x1040, 0x10A0, 0x14A0, 0x1910, 0x1208, 0x404, 0x802}},
    {35777, 2, {0x0, 0x2000, 0x13FE, 0x1020, 0x20, 0x20, 0xF120, 0x1120, 0x113C, 0x1120, 0x1120, 0x1520, 0x1920, 0x1120, 0x7FE, 0x0}},
    {35778, 2, {0x20, 0x2020, 0x1020, 0x1020, 0x3FE, 0x20, 0xF020, 0x1020, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {35779, 2, {0x0, 0x2000, 0x17FE, 0x1008, 0x8, 0x3C8, 0xF248, 0x1248, 0x1248, 0x1248, 0x13C8, 0x1248, 0x1408, 0x1808, 0x1028, 0x10}},
    {35780, 2, {0x0, 0x23F8, 0x1040, 0x1040, 0x248, 0x148, 0xF150, 0x1040, 0x17FE, 0x1040, 0x1040, 0x1440, 0x1840, 0x1040, 0x40, 0x40}},
    {35781, 2, {0x0, 0x21F8, 0x1108, 0x1108, 0x108, 0x1F8, 0xF108, 0x1108, 0x1108, 0x11F8, 0x1108, 0x1508, 0x1908, 0x1108, 0x7FE, 0x0}},
    {35782, 2, {0x0, 0x2000, 0x11FC, 0x1104, 0x104, 0x104, 0xF104, 0x1104, 0x11FC, 0x1104, 0x1000, 0x1490, 0x1888, 0x1104, 0x202, 0x402}},
    {35783, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x204, 0x2F4, 0xF294, 0x1294, 0x1294, 0x1294, 0x12F4, 0x1694, 0x1A04, 0x1204, 0x214, 0x208}},
    {35784, 2, {0x100, 0x2100, 0x1100, 0x11FE, 0x280, 0x280, 0xF480, 0x10F8, 0x1080, 0x1080, 0x1080, 0x14FC, 0x1880, 0x1080, 0x80, 0x80}},
    {35785, 2, {0x8, 0x201C, 0x11E0, 0x1100, 0x100, 0x100, 0xF1FE, 0x1110, 0x1110, 0x1130, 0x1118, 0x1514, 0x1912, 0x1210, 0x210, 0x410}},
    {35786, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x208, 0x426, 0xE040, 0x2080, 0x2310, 0x2020, 0x2040, 0x2888, 0x3310, 0x2020, 0xC0, 0x700}},
    {35787, 2, {0x8, 0x203C, 0x13E0, 0x1220, 0x220, 0x220, 0xF220, 0x13FE, 0x1220, 0x1210, 0x1210, 0x1212, 0x160A, 0x1A8A, 0x1326, 0x212}},
    {35788, 2, {0x80, 0x2080, 0x10F8, 0x1108, 0x210, 0x420, 0xF3FC, 0x1004, 0x1004, 0x1004, 0x11FC, 0x1004, 0x1404, 0x1804, 0x13FC, 0x4}},
    {35789, 2, {0x0, 0x23FC, 0x1004, 0x1004, 0x7F4, 0x4, 0xF004, 0x13E4, 0x1224, 0x1224, 0x1224, 0x13E4, 0x1404, 0x1804, 0x1014, 0x8}},
    {35790, 2, {0x20, 0x2020, 0x1124, 0x1124, 0x124, 0x124, 0xF1FC, 0x1020, 0x1020, 0x1124, 0x1124, 0x1124, 0x1524, 0x1924, 0x11FC, 0x4}},
    {35791, 2, {0x0, 0x23FC, 0x1084, 0x1084, 0x84, 0x104, 0xF114, 0x1208, 0x1400, 0x11FC, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {35792, 2, {0x20, 0x2020, 0x1020, 0x13FE, 0x222, 0x224, 0xF220, 0x13FC, 0x1284, 0x1288, 0x1248, 0x1650, 0x1A20, 0x1450, 0x488, 0x906}},
    {35793, 2, {0x0, 0x23FC, 0x1104, 0x1088, 0x50, 0x20, 0xF0D8, 0x1326, 0x1020, 0x11FC, 0x1020, 0x1420, 0x1BFE, 0x1020, 0x20, 0x20}},
    {35794, 2, {0x20, 0x2020, 0x1040, 0x1088, 0x104, 0x3FE, 0xF002, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {35795, 2, {0x0, 0x23FE, 0x1200, 0x1200, 0x3FC, 0x220, 0xF220, 0x12F8, 0x1220, 0x1220, 0x1220, 0x13FC, 0x1600, 0x1A00, 0x13FE, 0x0}},
    {35796, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x40, 0x3F8, 0xE040, 0x2040, 0x27FE, 0x20E0, 0x2150, 0x2950, 0x3248, 0x2444, 0x842, 0x40}},
    {35797, 2, {0x28, 0x2024, 0x1024, 0x1020, 0x7FE, 0x20, 0xF020, 0x17E0, 0x1120, 0x1110, 0x1110, 0x1510, 0x19CA, 0x170A, 0x206, 0x2}},
    {35798, 2, {0x20, 0x2020, 0x11FC, 0x1020, 0x20, 0x20, 0xF3FE, 0x1000, 0x1020, 0x1020, 0x11FC, 0x1020, 0x1420, 0x1820, 0x13FE, 0x0}},
    {35799, 2, {0x40, 0x2040, 0x1040, 0x13FC, 0x40, 0x40, 0xF7FE, 0x1010, 0x1010, 0x17FE, 0x1010, 0x1210, 0x1510, 0x1910, 0x1050, 0x20}},
    {35800, 2, {0x20, 0x2020, 0x1020, 0x17FE, 0x20, 0x20, 0xF3FE, 0x1000, 0x1000, 0x11FC, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {35801, 2, {0x80, 0x2080, 0x1080, 0x17FE, 0x100, 0x120, 0xF220, 0x12A4, 0x14A4, 0x1528, 0x1A50, 0x1050, 0x1488, 0x1908, 0x1204, 0x402}},
    {35802, 2, {0x28, 0x4024, 0x2020, 0x27FE, 0x420, 0x420, 0xE424, 0x27A4, 0x24A4, 0x24A8, 0x24A8, 0x2C90, 0x3692, 0x292A, 0x846, 0x1082}},
    {35803, 2, {0x20, 0x2120, 0x1120, 0x11FC, 0x120, 0x220, 0xF020, 0x13FE, 0x1070, 0x10A8, 0x10A8, 0x1524, 0x1924, 0x1222, 0x20, 0x20}},
    {35804, 2, {0x20, 0x2120, 0x1120, 0x11FC, 0x120, 0x220, 0xF020, 0x13FE, 0x1090, 0x1090, 0x1090, 0x1090, 0x1512, 0x1912, 0x1212, 0x40E}},
    {35805, 2, {0x8, 0x203C, 0x11E0, 0x1020, 0x20, 0x3FE, 0xF020, 0x1020, 0x1020, 0x11FC, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {35806, 2, {0x8, 0x401C, 0x2F70, 0x2110, 0x210, 0x450, 0xEF5C, 0x2150, 0x2950, 0x2950, 0x2550, 0x2A7C, 0x3300, 0x2480, 0x87E, 0x1000}},
    {35807, 2, {0x8, 0x403C, 0x23C0, 0x2200, 0x200, 0x3FE, 0xE200, 0x2200, 0x2200, 0x22FC, 0x2284, 0x2A84, 0x3484, 0x2484, 0x8FC, 0x84}},
    {35808, 2, {0x40, 0x2040, 0x10A0, 0x1110, 0x208, 0x406, 0xF3F8, 0x1040, 0x1040, 0x1040, 0x13F8, 0x1040, 0x1440, 0x1840, 0x17FE, 0x0}},
    {35809, 2, {0x100, 0x4100, 0x23F8, 0x2208, 0x410, 0xBFE, 0xE200, 0x22F8, 0x2288, 0x2288, 0x22A8, 0x2290, 0x2A82, 0x3482, 0x247E, 0x800}},
    {35810, 2, {0x200, 0x4200, 0x23FC, 0x2404, 0x404, 0xBE4, 0xE224, 0x2224, 0x23E4, 0x2224, 0x2224, 0x23E4, 0x2A24, 0x3004, 0x2014, 0x8}},
    {35811, 2, {0x100, 0x2104, 0x1138, 0x11C0, 0x102, 0x102, 0xF0FE, 0x1000, 0x11FC, 0x1104, 0x1104, 0x15FC, 0x1904, 0x1104, 0x1FC, 0x104}},
    {35812, 2, {0x80, 0x2080, 0x11F0, 0x1210, 0x20, 0x3FC, 0xF044, 0x1044, 0x17FE, 0x1044, 0x1044, 0x13FC, 0x1444, 0x1840, 0x1140, 0x80}},
    {35813, 2, {0x80, 0x4040, 0x2FFE, 0x2080, 0x110, 0x110, 0xE224, 0x27E4, 0x2048, 0x2088, 0x2110, 0x2A20, 0x3450, 0x2088, 0x304, 0xC04}},
    {35814, 2, {0x208, 0x2108, 0x1110, 0x1000, 0x7FC, 0x40, 0xF040, 0x13F8, 0x1040, 0x1040, 0x17FE, 0x1040, 0x1440, 0x1840, 0x1040, 0x40}},
    {35815, 2, {0x40, 0x2020, 0x1020, 0x13FE, 0x202, 0x404, 0xF038, 0x13C0, 0x1040, 0x1040, 0x107E, 0x17C0, 0x1440, 0x1842, 0x1042, 0x3E}},
    {35816, 2, {0x0, 0x23FE, 0x1202, 0x1444, 0x40, 0x3FC, 0xF080, 0x10A0, 0x1120, 0x11FC, 0x1020, 0x1420, 0x1BFE, 0x1020, 0x20, 0x20}},
    {35817, 2, {0x0, 0x47BC, 0x2084, 0x2084, 0x4A4, 0x294, 0xE294, 0x2084, 0x218C, 0x2294, 0x24A4, 0x2084, 0x2884, 0x3084, 0x2294, 0x108}},
    {35818, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x3F8, 0x80, 0xE7FE, 0x2090, 0x2110, 0x21FE, 0x2210, 0x2A90, 0x3450, 0x2810, 0x50, 0x20}},
    {35819, 2, {0x14, 0x4012, 0x2010, 0x2FFE, 0x10, 0x290, 0xE292, 0x2292, 0x2FD2, 0x2294, 0x2294, 0x2A88, 0x328A, 0x249A, 0x426, 0x842}},
    {35820, 2, {0x0, 0x47FC, 0x2040, 0x2040, 0x248, 0x248, 0xE248, 0x2554, 0x24D2, 0x2862, 0x2040, 0x2840, 0x3040, 0x2040, 0xFFE, 0x0}},
    {35821, 2, {0x0, 0x47FC, 0x2040, 0x2040, 0x3F8, 0x88, 0xE088, 0x27FE, 0x2000, 0x2000, 0x23F8, 0x2208, 0x2A08, 0x3208, 0x23F8, 0x208}},
    {35822, 2, {0x20, 0x2124, 0x10A4, 0x10A8, 0x20, 0x1FC, 0xF104, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1504, 0x1904, 0x1114, 0x108}},
    {35823, 2, {0x0, 0x43F8, 0x2208, 0x2208, 0x3F8, 0x0, 0xE7FC, 0x2040, 0x2040, 0x2FFE, 0x2040, 0x20A0, 0x28A0, 0x3110, 0x2208, 0x406}},
    {35824, 2, {0x20, 0x2120, 0x1120, 0x11FE, 0x220, 0x420, 0xF020, 0x13FE, 0x1000, 0x1000, 0x11FC, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {35825, 2, {0x1C, 0x47E0, 0x2040, 0x2040, 0xFFE, 0x150, 0xE248, 0x2C46, 0x23F0, 0x2110, 0x2120, 0x217C, 0x2A04, 0x3204, 0x2428, 0x810}},
    {35826, 2, {0x100, 0x2100, 0x11FC, 0x1200, 0x5F8, 0x108, 0xF148, 0x1128, 0x17FE, 0x1108, 0x1248, 0x1228, 0x17FC, 0x1808, 0x1050, 0x20}},
    {35827, 2, {0x0, 0x4440, 0x22BE, 0x2108, 0x288, 0x488, 0xE088, 0x20BE, 0x2188, 0x2288, 0x2488, 0x2088, 0x2888, 0x3088, 0x253E, 0x200}},
    {35828, 2, {0x208, 0x2108, 0x1110, 0x1020, 0x3F8, 0x208, 0xF208, 0x1208, 0x13F8, 0x10A0, 0x10A0, 0x1520, 0x1922, 0x1222, 0x41E, 0x800}},
    {35829, 2, {0x0, 0x47F8, 0x2010, 0x21A0, 0x40, 0x7FC, 0xE444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2C44, 0x3444, 0x2454, 0x408}},
    {35830, 2, {0x40, 0x4080, 0x2110, 0x2208, 0x7FC, 0x104, 0xE100, 0x23F8, 0x2440, 0x2040, 0x2FFE, 0x2040, 0x28A0, 0x3110, 0x2208, 0xC06}},
    {35831, 2, {0x40, 0x4040, 0x27FC, 0x2040, 0x3F8, 0x40, 0xE7FE, 0x2000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x2BF8, 0x3208, 0x2228, 0x210}},
    {35832, 2, {0x40, 0x4044, 0x23F4, 0x2048, 0x50, 0x7FE, 0xE040, 0x2080, 0x21F8, 0x2308, 0x2508, 0x21F8, 0x2908, 0x3108, 0x21F8, 0x108}},
    {35833, 2, {0x0, 0x4FC0, 0x24BC, 0x24A4, 0x4A4, 0x7A4, 0xE4A4, 0x24A8, 0x27A8, 0x24A8, 0x2490, 0x25D0, 0x2EA8, 0x30A8, 0x20C4, 0x82}},
    {35834, 2, {0x108, 0x4108, 0x27FE, 0x2108, 0x148, 0x40, 0xE7FE, 0x2080, 0x2100, 0x23FC, 0x2504, 0x2904, 0x2904, 0x3104, 0x21FC, 0x104}},
    {35835, 2, {0x20, 0x2020, 0x11FC, 0x1020, 0x20, 0x3FE, 0xF002, 0x1094, 0x1050, 0x1110, 0x1090, 0x13FE, 0x1428, 0x1844, 0x1082, 0x302}},
    {35836, 2, {0x0, 0x4FFE, 0x2080, 0x2184, 0x244, 0x468, 0xEAB0, 0x2130, 0x22A8, 0x2468, 0x28A4, 0x2124, 0x2A22, 0x3420, 0x20A0, 0x40}},
    {35837, 2, {0x90, 0x2090, 0x1090, 0x179E, 0x90, 0x90, 0xF090, 0x139C, 0x1090, 0x1090, 0x1090, 0x179E, 0x1890, 0x1090, 0x90, 0x90}},
    {35838, 2, {0x0, 0x43F8, 0x2248, 0x2248, 0x3F8, 0x248, 0xE248, 0x23F8, 0x2040, 0x27FC, 0x20E0, 0x2950, 0x3248, 0x2446, 0x40, 0x40}},
    {35839, 2, {0x38, 0x43C0, 0x2040, 0x2040, 0x7FC, 0x150, 0xE248, 0x2486, 0x2080, 0x27FC, 0x2110, 0x2A10, 0x31A0, 0x2060, 0x90, 0x308}},
    {35840, 2, {0x40, 0x4140, 0x265C, 0x2444, 0x444, 0x75C, 0xE444, 0x2444, 0x27FC, 0x2444, 0x20A0, 0x20A0, 0x2910, 0x3208, 0x2404, 0x802}},
    {35841, 2, {0x140, 0x4120, 0x2120, 0x23FE, 0x220, 0x620, 0xEBFC, 0x2220, 0x2220, 0x23FC, 0x2220, 0x2220, 0x2A20, 0x33FE, 0x2200, 0x200}},
    {35842, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x288, 0x446, 0xE800, 0x23F0, 0x2010, 0x2020, 0x2040, 0x2924, 0x3522, 0x250A, 0x908, 0xF8}},
    {35843, 2, {0x0, 0x47FC, 0x2444, 0x2444, 0x5F4, 0x444, 0xE444, 0x27FC, 0x2404, 0x25F4, 0x2514, 0x2D14, 0x35F4, 0x2404, 0x414, 0x808}},
    {35844, 2, {0x80, 0x4080, 0x20F8, 0x2108, 0x210, 0x400, 0xE040, 0x239C, 0x2204, 0x2204, 0x239C, 0x2204, 0x2A04, 0x3204, 0x23FC, 0x204}},
    {35845, 2, {0x40, 0x2020, 0x13FE, 0x1000, 0x0, 0x1FC, 0xF104, 0x1104, 0x1104, 0x11FC, 0x1020, 0x14A8, 0x1924, 0x1222, 0xA0, 0x40}},
    {35846, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x3F8, 0x208, 0xF3F8, 0x1000, 0x13F8, 0x1010, 0x1020, 0x17FC, 0x1840, 0x1040, 0x140, 0x80}},
    {35847, 2, {0x80, 0x4040, 0x27FC, 0x2000, 0x110, 0x110, 0xE2A8, 0x2444, 0x2000, 0x2040, 0x27FE, 0x2040, 0x2840, 0x3040, 0x2040, 0x40}},
    {35848, 2, {0x40, 0x2244, 0x1244, 0x1448, 0xA0, 0x110, 0xF608, 0x1044, 0x1040, 0x1248, 0x1248, 0x1450, 0x18A0, 0x1110, 0x208, 0xC06}},
    {35849, 2, {0x40, 0x2020, 0x13FE, 0x1222, 0x20, 0x1FC, 0xF124, 0x1124, 0x11FC, 0x1124, 0x1124, 0x15FC, 0x1924, 0x1020, 0x20, 0x20}},
    {35850, 2, {0x40, 0x2020, 0x13FE, 0x1202, 0x404, 0x1F8, 0xF108, 0x1108, 0x11F8, 0x1108, 0x1108, 0x15F8, 0x1908, 0x1108, 0x7FE, 0x0}},
    {35851, 2, {0x208, 0x4208, 0x2FFE, 0x2208, 0x208, 0x3F8, 0xE208, 0x2208, 0x23F8, 0x2040, 0x2FFE, 0x20E0, 0x2950, 0x3248, 0x2C46, 0x40}},
    {35852, 2, {0x110, 0x4110, 0x27FC, 0x2110, 0x1F0, 0x110, 0xE1F0, 0x2110, 0x2110, 0x27FE, 0x2200, 0x2290, 0x2B08, 0x3200, 0x23FC, 0x0}},
    {35853, 2, {0x90, 0x4290, 0x2290, 0x27FE, 0x290, 0x290, 0xE2F0, 0x2200, 0x23FC, 0x2040, 0x27FE, 0x20E0, 0x2950, 0x3248, 0x2C46, 0x40}},
    {35854, 2, {0x108, 0x4108, 0x27FE, 0x2148, 0x20, 0x7FE, 0xE200, 0x2200, 0x23FC, 0x2000, 0x2248, 0x2248, 0x2A48, 0x344A, 0x244A, 0x846}},
    {35855, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x7FC, 0x444, 0xE554, 0x24E4, 0x2444, 0x27FC, 0x20E0, 0x2950, 0x3248, 0x2446, 0x40, 0x40}},
    {35856, 2, {0x220, 0x2224, 0x13A8, 0x1230, 0x2A2, 0x322, 0xF25E, 0x1080, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1604, 0x1A04, 0x13FC, 0x204}},
    {35857, 2, {0x40, 0x207C, 0x1040, 0x13FE, 0x242, 0x278, 0xF3C0, 0x1244, 0x123C, 0x1200, 0x12FC, 0x1680, 0x1BFE, 0x1480, 0x480, 0x8FC}},
    {35858, 2, {0x0, 0x43F8, 0x2208, 0x23F8, 0x208, 0x3F8, 0xE100, 0x23FC, 0x2444, 0x2244, 0x22A4, 0x2204, 0x2BF4, 0x3004, 0x2028, 0x10}},
    {35859, 2, {0x0, 0x47FC, 0x2444, 0x27FC, 0x444, 0x7FC, 0xE000, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2A08, 0x3208, 0x2228, 0x210}},
    {35860, 2, {0x0, 0x47BC, 0x24A4, 0x24A4, 0x7BC, 0x0, 0xE3F8, 0x2000, 0x2FFE, 0x2100, 0x2200, 0x23F8, 0x2808, 0x3008, 0x2050, 0x20}},
    {35861, 2, {0x40, 0x20A0, 0x1110, 0x1208, 0x5F6, 0x0, 0xF3C4, 0x1254, 0x1254, 0x13D4, 0x1254, 0x1254, 0x17D4, 0x1A44, 0x1254, 0x2C8}},
    {35862, 2, {0x3C, 0x47C0, 0x2244, 0x2128, 0x0, 0x3FC, 0xE080, 0x2080, 0x27FE, 0x2100, 0x21F8, 0x2A88, 0x3250, 0x2420, 0x8D8, 0x306}},
    {35863, 2, {0x100, 0x21F0, 0x1210, 0x1420, 0x3FC, 0x244, 0xF244, 0x13FC, 0x10A0, 0x1122, 0x161E, 0x10C0, 0x1430, 0x1980, 0x1060, 0x10}},
    {35864, 2, {0x40, 0x2240, 0x117C, 0x1084, 0x108, 0x620, 0xF220, 0x1250, 0x1288, 0x1106, 0x11F8, 0x1108, 0x1508, 0x1908, 0x11F8, 0x108}},
    {35865, 2, {0x80, 0x4040, 0x27FC, 0x2000, 0x208, 0x110, 0xEFFE, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2A08, 0x3208, 0x23F8, 0x208}},
    {35866, 2, {0x40, 0x2020, 0x13FC, 0x1000, 0x108, 0x90, 0xF3FE, 0x1210, 0x1220, 0x12C8, 0x1210, 0x1624, 0x1AC8, 0x1410, 0x460, 0x980}},
    {35867, 2, {0x40, 0x2020, 0x11FC, 0x1000, 0x88, 0x50, 0xF3FE, 0x1222, 0x1424, 0x11FC, 0x1124, 0x1124, 0x1534, 0x1928, 0x1020, 0x20}},
    {35868, 2, {0x10, 0x4092, 0x2452, 0x2254, 0x210, 0xFE, 0xE010, 0x2638, 0x2254, 0x2292, 0x2310, 0x2210, 0x2A10, 0x3500, 0x28FE, 0x0}},
    {35869, 2, {0x80, 0x4040, 0x27FC, 0x2404, 0x404, 0x7FC, 0xE400, 0x2400, 0x27FC, 0x26A4, 0x2AA4, 0x2BFC, 0x3AA4, 0x2AA4, 0x12A4, 0x20C}},
    {35870, 2, {0x0, 0x23FE, 0x1022, 0x1120, 0x13C, 0x120, 0xF2FE, 0x1400, 0x11FC, 0x1104, 0x11FC, 0x1504, 0x19FC, 0x1104, 0x114, 0x108}},
    {35871, 2, {0x110, 0x4110, 0x27FC, 0x2110, 0x0, 0x3F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2040, 0x27FC, 0x28A0, 0x3110, 0x2208, 0xC06}},
    {35872, 2, {0x40, 0x4248, 0x2150, 0x2040, 0x7FE, 0x402, 0xE804, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x20A0, 0x2920, 0x3124, 0x2224, 0x41C}},
    {35873, 2, {0x0, 0x47FC, 0x2444, 0x27FC, 0x444, 0x7FC, 0xE110, 0x2208, 0x2484, 0x21F8, 0x2208, 0x2508, 0x3090, 0x2060, 0x198, 0xE06}},
    {35874, 2, {0x108, 0x4208, 0x2788, 0x2488, 0x7BE, 0x488, 0xE788, 0x24C8, 0x24A8, 0x2FA8, 0x2188, 0x2A88, 0x3488, 0x2888, 0x2A8, 0x110}},
    {35875, 2, {0x8, 0x203C, 0x13C0, 0x1004, 0x244, 0x128, 0xF1FC, 0x1220, 0x1020, 0x13FE, 0x1020, 0x1124, 0x1524, 0x1924, 0x11FC, 0x4}},
    {35876, 2, {0x40, 0x2020, 0x13FC, 0x1108, 0x90, 0x3FE, 0xF202, 0x1444, 0x1020, 0x13FC, 0x1080, 0x10F8, 0x1488, 0x1908, 0x1128, 0x210}},
    {35877, 2, {0x208, 0x4108, 0x2110, 0x2000, 0x7FC, 0x0, 0xE110, 0x2208, 0x2404, 0x23F8, 0x22A8, 0x22A8, 0x2AA8, 0x32A8, 0x2FFE, 0x0}},
    {35878, 2, {0x208, 0x4110, 0x2000, 0x27FE, 0xA0, 0x7FC, 0xE0A4, 0x2FFE, 0x20A4, 0x27FC, 0x20A0, 0x29B0, 0x32A8, 0x2CA6, 0xA0, 0xA0}},
    {35879, 2, {0x40, 0x4028, 0x2090, 0x22A4, 0x2C2, 0x48A, 0xE188, 0x2678, 0x2000, 0x23FC, 0x2294, 0x2294, 0x2A94, 0x3294, 0x27FE, 0x0}},
    {35880, 2, {0x110, 0x4110, 0x2FFE, 0x2110, 0x1F0, 0x40, 0xE7FC, 0x2444, 0x27FC, 0x2040, 0x2FFE, 0x2040, 0x2FFC, 0x3040, 0x2FFE, 0x0}},
    {35881, 2, {0x3F8, 0x4208, 0x23F8, 0x2208, 0x3F8, 0x0, 0xE7FC, 0x24A4, 0x24A4, 0x27FC, 0x2000, 0x23F8, 0x2910, 0x30E0, 0x2318, 0xC06}},
    {35882, 2, {0x80, 0x4040, 0x27FC, 0x2110, 0xA0, 0x7FC, 0xE444, 0x25F4, 0x2444, 0x25F4, 0x2514, 0x2514, 0x2DF4, 0x3404, 0x2414, 0x408}},
    {35883, 2, {0x108, 0x2090, 0x17FE, 0x1000, 0x3C4, 0x254, 0xF3D4, 0x1254, 0x13D4, 0x1244, 0x12CC, 0x1400, 0x1BFC, 0x1084, 0x114, 0x608}},
    {35884, 2, {0x0, 0x47BC, 0x2084, 0x24A4, 0x294, 0x4A4, 0xE050, 0x2188, 0x2626, 0x20C0, 0x2310, 0x2064, 0x2B88, 0x3030, 0x20C0, 0x700}},
    {35885, 2, {0x0, 0x23FE, 0x1050, 0x13FE, 0x252, 0x3FE, 0xF000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x15FC, 0x1820, 0x13FE, 0x20, 0x20}},
    {35886, 2, {0x0, 0x47BC, 0x2108, 0x2528, 0x7BC, 0x318, 0xE5AA, 0x2946, 0x2000, 0x23F8, 0x2208, 0x2208, 0x2BF8, 0x3208, 0x2208, 0x3F8}},
    {35887, 2, {0x140, 0x4120, 0x23FE, 0x2220, 0x620, 0xBFC, 0xE220, 0x2220, 0x23FC, 0x2220, 0x2220, 0x2BFE, 0x3200, 0x2524, 0x492, 0x892}},
    {35888, 2, {0x200, 0x417C, 0x2504, 0x2444, 0x5F4, 0x444, 0xE5F4, 0x2554, 0x2574, 0x25D4, 0x2554, 0x25F4, 0x2CE4, 0x3554, 0x2444, 0x44C}},
    {35889, 2, {0x208, 0x4110, 0x27FC, 0x20A0, 0x4A4, 0x2A8, 0xEFFE, 0x2000, 0x23F8, 0x2208, 0x2208, 0x23F8, 0x2A08, 0x3208, 0x23F8, 0x208}},
    {35890, 2, {0x0, 0x47F8, 0x2090, 0x2060, 0x7FC, 0xA4, 0xE128, 0x26A0, 0x2040, 0x27FC, 0x24A4, 0x2514, 0x2EEC, 0x34A4, 0x24E4, 0x40C}},
    {35891, 2, {0x108, 0x4108, 0x27CC, 0x210A, 0x108, 0x7DE, 0xE448, 0x26C8, 0x2548, 0x27C8, 0x2554, 0x27D4, 0x2D54, 0x3554, 0x2454, 0x4E2}},
    {35892, 2, {0x10, 0x44FE, 0x2292, 0x22FE, 0x10, 0x1FE, 0xE000, 0x2EFE, 0x2282, 0x22FE, 0x2280, 0x22FE, 0x2A82, 0x32FE, 0x2500, 0x8FE}},
    {35893, 2, {0x200, 0x43F8, 0x2410, 0x27FE, 0xD48, 0x1624, 0xE7FE, 0x2400, 0x25FC, 0x2400, 0x25FC, 0x2C00, 0x35FC, 0x2904, 0x9FC, 0x1104}},
    {35894, 2, {0x448, 0x444C, 0x2AAA, 0x2008, 0xFFE, 0x288, 0xEEE8, 0x228A, 0x2EEA, 0x228C, 0x2EEC, 0x2A88, 0x32EA, 0x2F1A, 0x426, 0x42}},
    {35895, 2, {0x820, 0x810, 0x1108, 0x2104, 0x4280, 0x440, 0x820, 0x3018, 0xC006, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {35896, 2, {0x888, 0x1C84, 0xF102, 0x1222, 0x1020, 0x1050, 0xFC88, 0x1104, 0x1202, 0x11FC, 0x1104, 0x1104, 0x2104, 0x2104, 0x41FC, 0x8104}},
    {35897, 2, {0x40, 0x4840, 0x4440, 0x9440, 0x13FE, 0x2840, 0x4480, 0x8090, 0x7890, 0x4920, 0x4920, 0x4A48, 0x4A44, 0x7C84, 0x49FE, 0x82}},
    {35898, 2, {0x0, 0x24FC, 0x2208, 0x4A48, 0x8848, 0x1448, 0x2288, 0x41FE, 0xBC18, 0x2428, 0x2428, 0x2448, 0x2488, 0x3D08, 0x2428, 0x10}},
    {35899, 2, {0x0, 0x25F8, 0x2248, 0x4A48, 0x8848, 0x15F8, 0x2248, 0x4148, 0xBC48, 0x25F8, 0x2448, 0x2448, 0x248A, 0x3C8A, 0x250A, 0x206}},
    {35900, 2, {0x48, 0x2448, 0x2248, 0x4A48, 0x89FE, 0x1448, 0x2248, 0x4148, 0xBC48, 0x25FE, 0x2400, 0x2448, 0x2444, 0x3C84, 0x2482, 0x102}},
    {35901, 2, {0x20, 0x4820, 0x4450, 0x9488, 0x1144, 0x2A22, 0x44F8, 0x8008, 0x7810, 0x4820, 0x49FC, 0x4904, 0x4904, 0x7904, 0x49FC, 0x104}},
    {35902, 2, {0x40, 0x4820, 0x45FE, 0x9502, 0x1204, 0x2850, 0x4488, 0x8104, 0x7800, 0x49FC, 0x4820, 0x4820, 0x4820, 0x7820, 0x4BFE, 0x0}},
    {35903, 2, {0xE00, 0xF050, 0x2248, 0x9284, 0x4524, 0x2020, 0x4450, 0xF888, 0x1106, 0x24F8, 0xFE88, 0x1088, 0xFE88, 0x2888, 0x44F8, 0x8288}},
    {35904, 2, {0x1E, 0x4BE0, 0x4522, 0x9494, 0x1040, 0x2888, 0x45F0, 0x8020, 0x78C4, 0x49FE, 0x4822, 0x4820, 0x4BFE, 0x7850, 0x4888, 0x306}},
    {35905, 2, {0x2000, 0x1050, 0xFE48, 0x8284, 0x1124, 0x7C20, 0x1050, 0x7C88, 0x1104, 0xFEFA, 0x1088, 0x7C88, 0x4488, 0x4488, 0x7CF8, 0x4488}},
    {35906, 2, {0x0, 0x24EE, 0x2222, 0x4AAA, 0x8866, 0x14AA, 0x2210, 0x4128, 0xBC44, 0x2592, 0x2420, 0x24C8, 0x2432, 0x3CC4, 0x2418, 0xE0}},
    {35907, 2, {0x8, 0x5388, 0x4888, 0x8910, 0x27DE, 0x3294, 0x4AA4, 0x8794, 0x7A94, 0x4B94, 0x4A94, 0x4AC8, 0x4B88, 0x7E94, 0x48A4, 0xC2}},
    {35908, 2, {0x20, 0x4BFE, 0x4420, 0x95FC, 0x1000, 0x2BFE, 0x4652, 0x83FE, 0x7904, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4888, 0x104}},
    {35909, 2, {0x110, 0x5090, 0x4BDE, 0x8810, 0x225E, 0x3182, 0x4BDE, 0x8410, 0x7BDE, 0x4A50, 0x4BDE, 0x4A50, 0x4BDE, 0x7A50, 0x4A52, 0x2CE}},
    {35910, 2, {0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x1010, 0x820, 0x440, 0xFFFE, 0x0}},
    {35911, 2, {0x0, 0xFE00, 0x1FC, 0x20, 0x7C20, 0x4420, 0x4420, 0x4420, 0x7C20, 0x20, 0x4420, 0x2820, 0x2E20, 0xF020, 0x43FE, 0x0}},
    {35912, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {35913, 2, {0x20, 0xFE20, 0x20, 0x1FE, 0x7C20, 0x4420, 0x4420, 0x45FC, 0x7C84, 0x88, 0x4448, 0x2850, 0x2E20, 0xF050, 0x4188, 0x606}},
    {35914, 2, {0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE, 0x0}},
    {35915, 2, {0x1080, 0x1080, 0x3EFC, 0x6244, 0x9448, 0x4830, 0x3010, 0x2FE8, 0xC006, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE, 0x0}},
    {35916, 2, {0x40, 0xFC20, 0x3FE, 0x202, 0x7D04, 0x4900, 0x49DE, 0x4A52, 0x7A52, 0x352, 0x8C9A, 0x4894, 0x5110, 0x1912, 0xE212, 0x40E}},
    {35917, 2, {0x20, 0xFE40, 0xFE, 0x92, 0x7C92, 0x44FE, 0x4492, 0x44A2, 0x7CFE, 0x48, 0x4488, 0x29FE, 0x2E08, 0xF008, 0x4008, 0x8}},
    {35918, 2, {0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7E44, 0x82, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x440, 0xFFFE}},
    {35919, 2, {0x88, 0xFE50, 0x0, 0x1FE, 0x7C50, 0x45FC, 0x4454, 0x47FE, 0x7C54, 0x1FC, 0x4450, 0x28D8, 0x2F54, 0xF252, 0x4050, 0x50}},
    {35920, 2, {0x820, 0x1D70, 0x4924, 0x5D74, 0x4924, 0x5D74, 0x4924, 0x7FFC, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7FFC}},
    {35921, 2, {0x2844, 0xFE28, 0xAA00, 0xFEFC, 0xAA14, 0xFE14, 0xFC, 0xFE90, 0x90, 0x7CFE, 0x4432, 0x7C32, 0x445A, 0x2854, 0x1E90, 0xE110}},
    {35922, 2, {0x2A22, 0x7F14, 0xAABE, 0xFF8A, 0xAA8A, 0xFFBE, 0xAAA8, 0xFFA8, 0x3E, 0xFF9A, 0x1A, 0x7F2A, 0x412E, 0x7F48, 0x2208, 0xFF88}},
    {35923, 2, {0x2A10, 0x7F10, 0xAA90, 0xFFBC, 0xAA90, 0xFFA8, 0xAAA8, 0xFFA4, 0x54, 0xFF82, 0x7C, 0x7F54, 0x4154, 0x7F54, 0x2254, 0xFFFE}},
    {35924, 2, {0x2A08, 0x7F08, 0xAA9C, 0xFF88, 0xAA88, 0xFFBE, 0xAA88, 0xFF90, 0x14, 0xFFBE, 0x0, 0x7F3E, 0x412A, 0x7F2A, 0x222A, 0xFFFE}},
    {35925, 2, {0x0, 0x7FFC, 0x200, 0x600, 0x908, 0x1190, 0x62A0, 0x4C0, 0x8C0, 0x11A0, 0x6290, 0x488, 0x886, 0x3080, 0xC280, 0x100}},
    {35926, 2, {0x0, 0x7FFC, 0x200, 0x600, 0x908, 0x1190, 0x62A0, 0x14C0, 0x8C0, 0x15A0, 0x6290, 0x488, 0x886, 0x3080, 0xC280, 0x100}},
    {35927, 2, {0x0, 0xFDFE, 0x2820, 0x2840, 0x28A2, 0x2B34, 0x2858, 0x2894, 0x2B34, 0x2852, 0x2892, 0x2B10, 0x2850, 0x4822, 0x47FE, 0x8000}},
    {35928, 2, {0x20, 0xFC20, 0x1020, 0x23FE, 0x6020, 0x9524, 0x3924, 0x5124, 0x9924, 0x35FC, 0x5424, 0x9020, 0x1022, 0x1022, 0x501E, 0x2000}},
    {35929, 2, {0x200, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x200, 0xD08, 0x7190, 0x2A0, 0xCC0, 0x71A0, 0x698, 0x1886, 0xE280, 0x100}},
    {35930, 2, {0x0, 0x7BFE, 0x4840, 0x4880, 0x4944, 0x7A24, 0x4868, 0x48B0, 0x4930, 0x7A28, 0x4868, 0x48A4, 0x4922, 0x4A20, 0x48A0, 0x9840}},
    {35931, 2, {0x0, 0xFCF8, 0x1088, 0x2088, 0x6088, 0x9506, 0x3A00, 0x51FC, 0x9884, 0x3484, 0x5448, 0x9050, 0x1020, 0x1050, 0x5088, 0x2306}},
    {35932, 2, {0x0, 0xFDFE, 0x1048, 0x2048, 0x6048, 0x9448, 0x3848, 0x53FE, 0x9848, 0x3448, 0x5448, 0x9048, 0x1088, 0x1088, 0x5108, 0x2208}},
    {35933, 2, {0x0, 0xFDFC, 0x1124, 0x2124, 0x6124, 0x9524, 0x3924, 0x51FC, 0x9904, 0x3500, 0x5500, 0x9100, 0x1102, 0x1102, 0x50FE, 0x2000}},
    {35934, 2, {0x40, 0xFE40, 0x880, 0x10FC, 0x3104, 0x4A04, 0x9CF4, 0x2894, 0x4C94, 0x9A94, 0x2AF4, 0x4894, 0x8804, 0x804, 0x2828, 0x1010}},
    {35935, 2, {0x40, 0xFC20, 0x11FC, 0x2104, 0x6104, 0x95FC, 0x3900, 0x5100, 0x997C, 0x3508, 0x5510, 0x9120, 0x1142, 0x1282, 0x527E, 0x2400}},
    {35936, 2, {0x0, 0xFEF8, 0x888, 0x1088, 0x3088, 0x4AF8, 0x9C88, 0x2888, 0x4C88, 0x9AF8, 0x2A88, 0x4888, 0x8888, 0x888, 0x2BFE, 0x1000}},
    {35937, 2, {0x800, 0x1FE0, 0x2020, 0x7FF8, 0xA108, 0x2208, 0x3FF8, 0x600, 0x1910, 0x62A0, 0xCC0, 0x71A0, 0x690, 0x1888, 0xE286, 0x100}},
    {35938, 2, {0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1010, 0x2FC8, 0xC606, 0x1910, 0x62A0, 0xCC0, 0x71B0, 0x688, 0x1884, 0xE280, 0x100}},
    {35939, 2, {0x0, 0xFDDC, 0x1088, 0x2088, 0x6088, 0x9488, 0x3888, 0x53DE, 0x9888, 0x3488, 0x5488, 0x9088, 0x1088, 0x1108, 0x5108, 0x2208}},
    {35940, 2, {0x0, 0xFDF8, 0x1108, 0x2108, 0x61F8, 0x9508, 0x3908, 0x51F8, 0x9944, 0x3548, 0x5530, 0x9120, 0x1110, 0x1148, 0x5186, 0x2100}},
    {35941, 2, {0x20, 0xFE10, 0x810, 0x11FE, 0x3020, 0x4A24, 0x9C44, 0x28F8, 0x4C12, 0x9A22, 0x2A44, 0x4988, 0x8810, 0x828, 0x2844, 0x1182}},
    {35942, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x2FF8, 0x2300, 0x2C90, 0x2360, 0x5CD8, 0x4346, 0x9CC0}},
    {35943, 2, {0x28, 0xFC24, 0x13FE, 0x2020, 0x6020, 0x95FC, 0x3924, 0x5124, 0x99FC, 0x3524, 0x5524, 0x91FC, 0x1124, 0x1124, 0x5124, 0x210C}},
    {35944, 2, {0xC2, 0xFC34, 0x1018, 0x2064, 0x6182, 0x9420, 0x39FE, 0x5050, 0x9890, 0x35FE, 0x5692, 0x9092, 0x109A, 0x1094, 0x5010, 0x2010}},
    {35945, 2, {0x0, 0xFDFE, 0x1020, 0x2040, 0x60A2, 0x9512, 0x3834, 0x5058, 0x9898, 0x3514, 0x5434, 0x9052, 0x1090, 0x1110, 0x5050, 0x2020}},
    {35946, 2, {0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x600, 0x1920, 0x6C0, 0x39B0, 0x68C, 0x3980}},
    {35947, 2, {0x80, 0xF8F8, 0x908, 0x53FE, 0x2512, 0x1122, 0xFDFE, 0x2440, 0x28A2, 0x2354, 0x2098, 0x2334, 0x2054, 0x2092, 0xA350, 0x4020}},
    {35948, 2, {0x20, 0xFE20, 0x8FA, 0x1024, 0x3028, 0x4BFE, 0x9C20, 0x2840, 0x4CFC, 0x9B44, 0x2A44, 0x487C, 0x8844, 0x844, 0x287C, 0x1044}},
    {35949, 2, {0x0, 0xFDDC, 0x1144, 0x2144, 0x6144, 0x95DC, 0x3900, 0x5100, 0x99DC, 0x3514, 0x5514, 0x91D4, 0x1108, 0x1108, 0x5114, 0x2122}},
    {35950, 2, {0x20, 0xFDFC, 0x1020, 0x2088, 0x63FE, 0x9488, 0x3800, 0x51FC, 0x9904, 0x3524, 0x5524, 0x9124, 0x1124, 0x1050, 0x5088, 0x2304}},
    {35951, 2, {0xE, 0xFEF0, 0x822, 0x1092, 0x3044, 0x4A20, 0x9C44, 0x28F8, 0x4C10, 0x9A24, 0x2AFE, 0x4810, 0x88FE, 0x828, 0x2844, 0x1082}},
    {35952, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x8200, 0x7CFC, 0x44, 0xFE44, 0x2028, 0x5428, 0x2810, 0xDC28, 0x2A44, 0xD982}},
    {35953, 2, {0x0, 0xFCF8, 0x1088, 0x2088, 0x60F8, 0x9488, 0x3888, 0x50F8, 0x9800, 0x35FC, 0x5554, 0x9154, 0x1154, 0x1154, 0x53FE, 0x2000}},
    {35954, 2, {0x0, 0xFDFE, 0x1110, 0x2120, 0x617C, 0x9544, 0x397C, 0x5144, 0x997C, 0x3510, 0x5510, 0x9154, 0x1152, 0x1292, 0x5250, 0x2420}},
    {35955, 2, {0x100, 0x4104, 0x7DF4, 0x4924, 0x5144, 0x6BAC, 0x4D34, 0x5964, 0x6DB4, 0x4B2C, 0x5964, 0x69A4, 0x5964, 0x4104, 0x7FFC, 0x4}},
    {35956, 2, {0x20, 0xFE10, 0x8FE, 0x1044, 0x3028, 0x4AFE, 0x9C82, 0x2892, 0x4CFE, 0x9A92, 0x2ABA, 0x48AA, 0x88AA, 0x8BA, 0x2882, 0x1086}},
    {35957, 2, {0x124, 0xF924, 0x2224, 0x44A4, 0xA154, 0x154A, 0x3B92, 0x5508, 0x9908, 0x3528, 0x512E, 0x9128, 0x1128, 0x1158, 0x514E, 0x2180}},
    {35958, 2, {0x20, 0xFDFC, 0x1020, 0x2088, 0x63FE, 0x9488, 0x39FC, 0x5104, 0x99FC, 0x3504, 0x55FC, 0x9104, 0x11FC, 0x1088, 0x5104, 0x2202}},
    {35959, 2, {0x20, 0xFBFE, 0x2020, 0x41FC, 0xA000, 0x17FE, 0x3A02, 0x51FC, 0x9800, 0x35FC, 0x5104, 0x91FC, 0x1104, 0x1088, 0x53FE, 0x2000}},
    {35960, 2, {0xE0, 0x700, 0x7A0C, 0x1130, 0x8C0, 0x780, 0x7840, 0xA0, 0x330, 0xC50, 0x7090, 0x110, 0x610, 0x1810, 0xE0A0, 0x40}},
    {35961, 2, {0xC40, 0x3040, 0xC280, 0x14FC, 0x4904, 0x3204, 0xC804, 0x1504, 0x2484, 0xCC44, 0x1444, 0x2404, 0xC404, 0x404, 0x2828, 0x1010}},
    {35962, 2, {0xC10, 0x3010, 0xC210, 0x1410, 0x49FE, 0x3010, 0xC830, 0x1430, 0x2450, 0xCC50, 0x1490, 0x2510, 0xC410, 0x410, 0x2850, 0x1020}},
    {35963, 2, {0xC00, 0x30FC, 0xC220, 0x1420, 0x4820, 0x3020, 0xC820, 0x15FE, 0x2420, 0xCC20, 0x1420, 0x2420, 0xC420, 0x420, 0x2820, 0x1020}},
    {35964, 2, {0xC10, 0x3110, 0xC310, 0x1512, 0x4912, 0x3114, 0xC9D8, 0x1510, 0x2510, 0xCD10, 0x1510, 0x2512, 0xC552, 0x592, 0x290E, 0x1000}},
    {35965, 2, {0xC20, 0x3020, 0xC220, 0x15FE, 0x4922, 0x3122, 0xC922, 0x1522, 0x2552, 0xCD4A, 0x158A, 0x2502, 0xC502, 0x502, 0x290A, 0x1104}},
    {35966, 2, {0xC00, 0x31FE, 0xC210, 0x1410, 0x4820, 0x3020, 0xC868, 0x14A4, 0x2522, 0xCE22, 0x1420, 0x2420, 0xC420, 0x400, 0x2BFE, 0x1000}},
    {35967, 2, {0xC80, 0x3080, 0xC280, 0x15FC, 0x4904, 0x3204, 0xC9E4, 0x1524, 0x2524, 0xCD24, 0x1524, 0x25E4, 0xC524, 0x404, 0x2828, 0x1010}},
    {35968, 2, {0xC20, 0x3020, 0xC324, 0x1524, 0x4924, 0x3124, 0xC9FC, 0x1420, 0x2420, 0xCD24, 0x1524, 0x2524, 0xC524, 0x524, 0x29FC, 0x1004}},
    {35969, 2, {0x1040, 0x2020, 0xD420, 0xBFE, 0x5202, 0x2404, 0xD000, 0x1890, 0x2890, 0xC890, 0x1888, 0x2908, 0xC908, 0x904, 0x5204, 0x2402}},
    {35970, 2, {0xC00, 0x31FC, 0xC244, 0x1444, 0x4844, 0x3044, 0xC894, 0x1488, 0x2500, 0xCCFC, 0x1484, 0x2484, 0xC484, 0x484, 0x28FC, 0x1084}},
    {35971, 2, {0xC20, 0x3020, 0xC240, 0x15FC, 0x4904, 0x3104, 0xC904, 0x1504, 0x25FC, 0xCD04, 0x1504, 0x2504, 0xC504, 0x504, 0x29FC, 0x1104}},
    {35972, 2, {0xC20, 0x3020, 0xC3FC, 0x1424, 0x4BFE, 0x3024, 0xC9FC, 0x1420, 0x2420, 0xCDFC, 0x1420, 0x2420, 0xC7FE, 0x420, 0x2820, 0x1020}},
    {35973, 2, {0x1090, 0x2090, 0xD490, 0x910, 0x51FE, 0x2310, 0xD510, 0x1938, 0x2938, 0xC954, 0x1954, 0x2992, 0xC910, 0x910, 0x5110, 0x2110}},
    {35974, 2, {0xC00, 0x31FE, 0xC200, 0x1400, 0x48FC, 0x3084, 0xC884, 0x14FC, 0x2484, 0xCC84, 0x14FC, 0x2484, 0xC400, 0x400, 0x29FE, 0x1000}},
    {35975, 2, {0xC00, 0x31F8, 0xC308, 0x1508, 0x49F8, 0x3108, 0xC908, 0x15F8, 0x2544, 0xCD48, 0x1530, 0x2520, 0xC510, 0x548, 0x2986, 0x1100}},
    {35976, 2, {0xC10, 0x3020, 0xC2FC, 0x1484, 0x48A4, 0x3094, 0xC894, 0x15FE, 0x2484, 0xCCA4, 0x1494, 0x2494, 0xC484, 0x504, 0x2914, 0x1208}},
    {35977, 2, {0x1080, 0x2080, 0xD4F8, 0x908, 0x5310, 0x24A0, 0xD040, 0x18A0, 0x2918, 0xCA06, 0x1DF8, 0x2908, 0xC908, 0x908, 0x51F8, 0x2108}},
    {35978, 2, {0xC00, 0x30FE, 0xC220, 0x1420, 0x4840, 0x30FC, 0xC884, 0x1484, 0x2484, 0xCCFC, 0x1484, 0x2484, 0xC484, 0x484, 0x28FC, 0x1084}},
    {35979, 2, {0xC00, 0x31FC, 0xC304, 0x15FC, 0x4904, 0x31FC, 0xC800, 0x1400, 0x25FC, 0xCC20, 0x1420, 0x27FE, 0xC420, 0x420, 0x2820, 0x1020}},
    {35980, 2, {0xC20, 0x3040, 0xC3FC, 0x1504, 0x4904, 0x31FC, 0xC904, 0x1504, 0x25FC, 0xCC50, 0x1450, 0x2450, 0xC492, 0x492, 0x290E, 0x1200}},
    {35981, 2, {0xC00, 0x31FC, 0xC324, 0x1524, 0x49FC, 0x3124, 0xC924, 0x15FC, 0x2420, 0xCC20, 0x15FC, 0x2420, 0xC420, 0x420, 0x2BFE, 0x1000}},
    {35982, 2, {0x1000, 0x2040, 0xD79C, 0xA04, 0x5204, 0x239C, 0xD204, 0x1A04, 0x2BFC, 0xC890, 0x1890, 0x2890, 0xC890, 0x912, 0x5212, 0x240E}},
    {35983, 2, {0x1040, 0x2080, 0xD5FC, 0x924, 0x5124, 0x21FC, 0xD124, 0x1944, 0x29FC, 0xC890, 0x1910, 0x2BFE, 0xC810, 0x810, 0x5010, 0x2010}},
    {35984, 2, {0x1040, 0x20A0, 0xD510, 0xA08, 0x55F6, 0x2000, 0xD3C4, 0x1A54, 0x2A54, 0xCBD4, 0x1A54, 0x2A54, 0xCBD4, 0xA44, 0x5254, 0x22C8}},
    {35985, 2, {0x1000, 0x23DE, 0xD642, 0xA42, 0x5242, 0x23DE, 0xD200, 0x1A3E, 0x2BD2, 0xCA12, 0x1A14, 0x2BD4, 0xCA08, 0xA14, 0x5224, 0x2242}},
    {35986, 2, {0xC20, 0x3124, 0xC324, 0x1524, 0x49FC, 0x3000, 0xCBFE, 0x1420, 0x2440, 0xCDFC, 0x1554, 0x2554, 0xC554, 0x554, 0x2954, 0x110C}},
    {35987, 2, {0xC88, 0x3088, 0xC3FE, 0x1488, 0x4888, 0x3000, 0xC9FC, 0x1524, 0x2524, 0xCD24, 0x15FC, 0x2524, 0xC524, 0x524, 0x29FC, 0x1104}},
    {35988, 2, {0xC20, 0x3040, 0xC3FC, 0x1504, 0x4954, 0x3124, 0xC954, 0x1504, 0x25FC, 0xCC00, 0x1512, 0x25D4, 0xC518, 0x552, 0x2992, 0x110E}},
    {35989, 2, {0x101E, 0x23E0, 0xD522, 0x894, 0x5040, 0x2088, 0xD1F0, 0x1820, 0x28C4, 0xC9FE, 0x1822, 0x2820, 0xCBFE, 0x850, 0x5088, 0x2306}},
    {35990, 2, {0x1104, 0x2084, 0xD488, 0x800, 0x53FE, 0x2000, 0xD088, 0x1904, 0x2A02, 0xC9FC, 0x1954, 0x2954, 0xC954, 0x954, 0x57FE, 0x2000}},
    {35991, 2, {0x1020, 0x21FC, 0xD524, 0xBFE, 0x5124, 0x21FC, 0xD020, 0x19FC, 0x2924, 0xC9FC, 0x1840, 0x2BFE, 0xC888, 0x9D0, 0x5070, 0x238C}},
    {35992, 2, {0xC88, 0x3088, 0xC3FE, 0x1488, 0x4800, 0x31FC, 0xC904, 0x15FC, 0x2504, 0xCDFC, 0x1420, 0x25FE, 0xC450, 0x488, 0x2904, 0x1202}},
    {35993, 2, {0x1000, 0x23FE, 0xD600, 0xA7C, 0x5244, 0x2244, 0xD27C, 0x1A00, 0x2AEE, 0xCAAA, 0x1AAA, 0x2AAA, 0xCAEE, 0xA00, 0x53FE, 0x2000}},
    {35994, 2, {0x1000, 0x23DE, 0xD652, 0xA52, 0x53DE, 0x2000, 0xD1FC, 0x1924, 0x29FC, 0xC924, 0x19FC, 0x2820, 0xCBFE, 0x820, 0x5020, 0x2020}},
    {35995, 2, {0x1088, 0x23FE, 0xD488, 0xBDE, 0x5252, 0x23DE, 0xD0A0, 0x1890, 0x29FE, 0xC920, 0x1BFC, 0x2D20, 0xC9FC, 0x920, 0x51FE, 0x2100}},
    {35996, 2, {0x13DE, 0x2252, 0xD7DE, 0xA52, 0x53DE, 0x20A0, 0xD1FE, 0x1B20, 0x2DFC, 0xC920, 0x19FE, 0x2800, 0xC9FC, 0x888, 0x5070, 0x278E}},
    {35997, 2, {0x0, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x0, 0x820, 0x1010, 0x2008, 0x4004}},
    {35998, 2, {0x100, 0x1FC, 0x100, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {35999, 2, {0x800, 0x1010, 0x2008, 0x7FFC, 0x2004, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {36000, 2, {0x800, 0x800, 0x1FE0, 0x2020, 0x4040, 0xBFF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x460, 0x1818, 0x6004}},
    {36001, 2, {0x10, 0x7C10, 0x4410, 0x4410, 0x7DFE, 0x4410, 0x4430, 0x7C30, 0x4450, 0x4450, 0x7C90, 0x110, 0x2810, 0x2410, 0x4450, 0x8020}},
    {36002, 2, {0x0, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {36003, 2, {0x420, 0x210, 0x13E, 0xFFC0, 0x22, 0x1A, 0x3FF6, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x2010, 0x3FF0, 0x840, 0x1020, 0x2010}},
    {36004, 2, {0x20, 0x7820, 0x4820, 0x4920, 0x792C, 0x4934, 0x4964, 0x7BA4, 0x4924, 0x4934, 0x7928, 0x122, 0x5122, 0x4902, 0x88FE, 0x0}},
    {36005, 2, {0x40, 0x7C20, 0x4400, 0x45FC, 0x7C00, 0x4400, 0x44F0, 0x7C90, 0x4490, 0x4490, 0x7C90, 0x92, 0x2892, 0x2512, 0x450E, 0x8200}},
    {36006, 2, {0x0, 0x7C00, 0x45FC, 0x4400, 0x7C00, 0x4400, 0x47FE, 0x7C90, 0x4490, 0x4490, 0x7C90, 0x90, 0x2892, 0x2512, 0x4512, 0x820E}},
    {36007, 2, {0x440, 0x820, 0x3018, 0xCFE6, 0x420, 0x18A0, 0x6040, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36008, 2, {0x880, 0x1098, 0x30E0, 0x5784, 0x9084, 0x107C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36009, 2, {0x8, 0x781C, 0x49E0, 0x4900, 0x7900, 0x49FC, 0x4944, 0x7944, 0x4944, 0x4928, 0x7928, 0x110, 0x5110, 0x4A28, 0x8A44, 0x482}},
    {36010, 2, {0x100, 0x280, 0xC60, 0x3118, 0xCFE6, 0x40, 0x80, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36011, 2, {0x0, 0x1FF0, 0x1110, 0xFFFE, 0x1110, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36012, 2, {0x100, 0x3FF8, 0x100, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36013, 2, {0x8, 0x7C, 0x3F80, 0x2080, 0x2080, 0x3FFE, 0x2080, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x47F8, 0x4210, 0x8408}},
    {36014, 2, {0x20, 0x3F28, 0x24, 0x20, 0xFFFE, 0x20, 0x7FA0, 0x40A0, 0x7FA0, 0x4090, 0x7F90, 0x4090, 0x7F8A, 0x120A, 0x2106, 0x4082}},
    {36015, 2, {0x40, 0x7C20, 0x4420, 0x45FE, 0x7D02, 0x4604, 0x4400, 0x7C00, 0x45FE, 0x4420, 0x7C20, 0x20, 0x2820, 0x2420, 0x44A0, 0x8040}},
    {36016, 2, {0x1220, 0x7FFC, 0x1220, 0x13E0, 0x1000, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36017, 2, {0x10, 0x7810, 0x4810, 0x49FE, 0x7912, 0x4914, 0x4910, 0x79FC, 0x4944, 0x4944, 0x7928, 0x128, 0x5110, 0x4A28, 0x8A44, 0x482}},
    {36018, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF07C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36019, 2, {0x24, 0xFFFE, 0x20, 0x3FA0, 0x20, 0x7FA0, 0x20, 0x3FA0, 0x20A0, 0x3F90, 0x2090, 0x3F90, 0x208A, 0x3F8A, 0x1106, 0x2082}},
    {36020, 2, {0x100, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36021, 2, {0x400, 0x820, 0x1FF0, 0x210, 0xFFFE, 0x820, 0x3018, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36022, 2, {0x4, 0x7C1E, 0x45E0, 0x4400, 0x7C20, 0x4410, 0x45FE, 0x7C04, 0x4408, 0x4410, 0x7C20, 0x40, 0x2880, 0x2540, 0x463E, 0x8000}},
    {36023, 2, {0x0, 0x7FFC, 0x4444, 0x4444, 0x7FFC, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {36024, 2, {0x890, 0x1088, 0x307E, 0xD7C0, 0x1022, 0x101A, 0x6, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36025, 2, {0x10, 0x7C10, 0x4490, 0x4490, 0x7CFE, 0x4490, 0x4510, 0x7C10, 0x4410, 0x44FE, 0x7C10, 0x10, 0x2810, 0x2410, 0x45FE, 0x8000}},
    {36026, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x7D04, 0x4504, 0x45FC, 0x7C50, 0x4450, 0x4450, 0x7C50, 0x50, 0x2892, 0x2492, 0x450E, 0x8200}},
    {36027, 2, {0x440, 0x7FF8, 0x448, 0x3FF8, 0x2440, 0x3FFC, 0x844, 0x1FF4, 0x301C, 0x5FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36028, 2, {0x20, 0x7C20, 0x4420, 0x4420, 0x7C3E, 0x4420, 0x4420, 0x7C20, 0x45FC, 0x4504, 0x7D04, 0x104, 0x2904, 0x2504, 0x45FC, 0x8104}},
    {36029, 2, {0x20, 0x7C20, 0x4420, 0x4440, 0x7C48, 0x4484, 0x45FE, 0x7C82, 0x4400, 0x44FC, 0x7C84, 0x84, 0x2884, 0x2484, 0x44FC, 0x8084}},
    {36030, 2, {0x4, 0x7C1E, 0x45F0, 0x4510, 0x7D10, 0x4510, 0x4510, 0x7DFE, 0x4510, 0x4510, 0x7D10, 0x108, 0x290A, 0x254A, 0x45A6, 0x8112}},
    {36031, 2, {0x600, 0x78FC, 0x4044, 0x4844, 0x4444, 0x5A94, 0x6108, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36032, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x2244, 0x2A7C, 0x4400, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36033, 2, {0x100, 0x3FF8, 0x100, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36034, 2, {0x80, 0x7880, 0x48F8, 0x4908, 0x7B10, 0x4CA0, 0x4840, 0x78A0, 0x4918, 0x4A06, 0x7DF8, 0x108, 0x5108, 0x4908, 0x89F8, 0x108}},
    {36035, 2, {0x818, 0x13E0, 0x3040, 0x57FC, 0x9040, 0x13F8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36036, 2, {0x20, 0x7C20, 0x45FE, 0x4440, 0x7C40, 0x44FC, 0x4484, 0x7D84, 0x46FC, 0x4484, 0x7C84, 0xFC, 0x2884, 0x2484, 0x4494, 0x8088}},
    {36037, 2, {0x20, 0x7C10, 0x4410, 0x45FE, 0x7C20, 0x4424, 0x4444, 0x7CF8, 0x4412, 0x4422, 0x7C44, 0x188, 0x2810, 0x2428, 0x4444, 0x8182}},
    {36038, 2, {0x82, 0x7C42, 0x4444, 0x4400, 0x7CFE, 0x4444, 0x4444, 0x7C44, 0x4444, 0x45FE, 0x7C44, 0x44, 0x2884, 0x2484, 0x4504, 0x8204}},
    {36039, 2, {0x4100, 0x21FC, 0xA44, 0x1048, 0xE0A0, 0x2118, 0x2606, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36040, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36041, 2, {0x20, 0x7C20, 0x4440, 0x4480, 0x7DFC, 0x4554, 0x4554, 0x7D54, 0x4554, 0x4554, 0x7D54, 0x154, 0x2954, 0x2554, 0x47FE, 0x8000}},
    {36042, 2, {0x14, 0x7C12, 0x4410, 0x45FE, 0x7C10, 0x4490, 0x4492, 0x7C92, 0x47F2, 0x4494, 0x7C94, 0x88, 0x288A, 0x251A, 0x4526, 0x8242}},
    {36043, 2, {0x40, 0x7C20, 0x4420, 0x45FE, 0x7C00, 0x4488, 0x4504, 0x7E02, 0x4488, 0x4488, 0x7C50, 0x50, 0x2820, 0x2450, 0x4488, 0x8306}},
    {36044, 2, {0x100, 0xFFFE, 0x840, 0x1F90, 0x620, 0x38D0, 0x708, 0x7FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36045, 2, {0x20, 0x7810, 0x4810, 0x49FE, 0x7900, 0x4910, 0x4910, 0x7910, 0x497C, 0x4910, 0x7910, 0x110, 0x5110, 0x4A10, 0x8AFE, 0x400}},
    {36046, 2, {0x28, 0x7C24, 0x443E, 0x45E0, 0x7C20, 0x443C, 0x45E0, 0x7C20, 0x443E, 0x45E0, 0x7C22, 0x24, 0x2818, 0x246A, 0x4586, 0x8002}},
    {36047, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x0, 0x2848, 0x2444, 0x4482, 0x8102}},
    {36048, 2, {0x40, 0x7840, 0x4888, 0x4904, 0x7BFE, 0x4802, 0x4888, 0x7944, 0x4A42, 0x48F8, 0x7988, 0x250, 0x5020, 0x4850, 0x8988, 0x606}},
    {36049, 2, {0x0, 0x79FE, 0x4900, 0x4900, 0x797C, 0x4900, 0x4900, 0x79FE, 0x4950, 0x4952, 0x7954, 0x148, 0x5148, 0x4A44, 0x8A52, 0x460}},
    {36050, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x7D04, 0x46FA, 0x4400, 0x7C00, 0x47FE, 0x4420, 0x7D28, 0x124, 0x2A22, 0x2422, 0x44A0, 0x8040}},
    {36051, 2, {0x100, 0x7FFE, 0x4002, 0xBFFC, 0x910, 0x3160, 0x380, 0x3FF8, 0xE008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010}},
    {36052, 2, {0x100, 0x7FFE, 0x4002, 0xBFFC, 0x2108, 0x3FF8, 0x2000, 0x2FF8, 0x2808, 0x2FF8, 0x2808, 0x2FF8, 0x2808, 0x4FF8, 0x4410, 0x8808}},
    {36053, 2, {0x28, 0x7C24, 0x4424, 0x4420, 0x7DFE, 0x4420, 0x4520, 0x7CB2, 0x44B4, 0x4468, 0x7CA8, 0x124, 0x2A22, 0x2420, 0x44A0, 0x8040}},
    {36054, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x7D04, 0x46FA, 0x4420, 0x7C20, 0x47FE, 0x4420, 0x7D28, 0x124, 0x2A22, 0x2422, 0x44A0, 0x8040}},
    {36055, 2, {0x20, 0x7820, 0x49FC, 0x4924, 0x7924, 0x49FC, 0x4820, 0x7820, 0x4BFE, 0x4A22, 0x7A22, 0x3FE, 0x5222, 0x4820, 0x8820, 0x20}},
    {36056, 2, {0x20, 0x7810, 0x4810, 0x49FE, 0x7900, 0x4910, 0x4910, 0x7910, 0x497C, 0x4910, 0x7914, 0x112, 0x5112, 0x4A10, 0x8AFE, 0x400}},
    {36057, 2, {0x0, 0x7CFE, 0x4492, 0x4492, 0x7CBA, 0x4492, 0x44FE, 0x7C82, 0x44BA, 0x44AA, 0x7CAA, 0xBA, 0x2882, 0x2502, 0x450A, 0x8204}},
    {36058, 2, {0x100, 0x7FFC, 0x920, 0x1550, 0x2388, 0x540, 0x920, 0x3FF8, 0xD016, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36059, 2, {0x820, 0x820, 0x7EFC, 0x820, 0xFEFE, 0x1450, 0x2288, 0x5FF6, 0x9010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36060, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x4440, 0x44FE, 0x7D2A, 0x24A, 0x2892, 0x2522, 0x444A, 0x8084}},
    {36061, 2, {0x0, 0x7BFE, 0x4A02, 0x4A52, 0x7888, 0x4904, 0x4820, 0x7820, 0x4BFE, 0x4820, 0x7870, 0xA8, 0x5124, 0x4E22, 0x8820, 0x20}},
    {36062, 2, {0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x2008}},
    {36063, 2, {0x50, 0x7C50, 0x4450, 0x45FC, 0x7D54, 0x4554, 0x4554, 0x7DFC, 0x4554, 0x4554, 0x7D54, 0x3FE, 0x2800, 0x2488, 0x4504, 0x8202}},
    {36064, 2, {0x40, 0x7C20, 0x45FC, 0x4400, 0x7D08, 0x4490, 0x4400, 0x7DFE, 0x4400, 0x4400, 0x7DFC, 0x104, 0x2904, 0x2504, 0x45FC, 0x8104}},
    {36065, 2, {0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2108, 0x2FF8, 0x2410, 0x2FF8, 0x3416, 0x27F0, 0x4410, 0x47F0, 0x8220, 0x410}},
    {36066, 2, {0x0, 0x7EFC, 0x4844, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7FF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36067, 2, {0x100, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x7FFC, 0x4444, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36068, 2, {0x48, 0x7C44, 0x445E, 0x45E0, 0x7C28, 0x4412, 0x446A, 0x7D96, 0x4448, 0x445E, 0x7DE0, 0x24, 0x2828, 0x2412, 0x446A, 0x8186}},
    {36069, 2, {0x40, 0x7C20, 0x45FE, 0x4400, 0x7C88, 0x4488, 0x4554, 0x7E22, 0x4400, 0x4420, 0x7DFE, 0x20, 0x2820, 0x2420, 0x4420, 0x8020}},
    {36070, 2, {0x8, 0x7C0C, 0x44EA, 0x440A, 0x7C08, 0x45FE, 0x4448, 0x7C48, 0x4548, 0x4568, 0x7D48, 0x148, 0x294A, 0x256A, 0x4786, 0x8102}},
    {36071, 2, {0x20, 0x7CA2, 0x44A2, 0x4524, 0x7C50, 0x4488, 0x4704, 0x7C22, 0x4420, 0x44A4, 0x7CA4, 0x128, 0x2850, 0x2488, 0x4504, 0x8202}},
    {36072, 2, {0x100, 0x7FFC, 0x4004, 0x1FF0, 0x0, 0x7FFC, 0x1110, 0x2508, 0x4204, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36073, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x7800, 0x49FC, 0x4800, 0x7800, 0x4BFE, 0x4820, 0x7928, 0x124, 0x5222, 0x4C22, 0x88A0, 0x40}},
    {36074, 2, {0x60C, 0x3870, 0x2040, 0x3F7E, 0x2448, 0x2448, 0x4488, 0x9FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36075, 2, {0x100, 0x7FFC, 0x920, 0x920, 0x1550, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36076, 2, {0x0, 0x7CFC, 0x4480, 0x44F8, 0x7C80, 0x44F8, 0x4480, 0x7FFE, 0x4540, 0x4524, 0x7D28, 0x110, 0x2908, 0x2544, 0x4582, 0x8100}},
    {36077, 2, {0x20, 0x7C22, 0x45FA, 0x4424, 0x7C28, 0x47FE, 0x4420, 0x7C40, 0x44FC, 0x4584, 0x7E84, 0xFC, 0x2884, 0x2484, 0x44FC, 0x8084}},
    {36078, 2, {0x100, 0x3FF8, 0x108, 0xFFFE, 0x108, 0x3FF8, 0x100, 0x7FFC, 0x2488, 0x5FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36079, 2, {0x100, 0x79FE, 0x4A02, 0x4DF2, 0x7912, 0x49F2, 0x4912, 0x79FA, 0x4804, 0x4BF8, 0x7810, 0x20, 0x57FE, 0x4820, 0x88A0, 0x40}},
    {36080, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x7DFC, 0x4440, 0x47FE, 0x7C88, 0x4504, 0x46FA, 0x7C88, 0x88, 0x28F8, 0x2488, 0x4488, 0x80F8}},
    {36081, 2, {0x0, 0x7BFE, 0x4A22, 0x4820, 0x7BFE, 0x4820, 0x49FC, 0x7924, 0x49FC, 0x4924, 0x79FC, 0x20, 0x53FE, 0x4820, 0x8820, 0x20}},
    {36082, 2, {0x13F8, 0x1208, 0x23F8, 0x6040, 0xA7FC, 0x2150, 0x2248, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36083, 2, {0x128, 0x7924, 0x4BA4, 0x4920, 0x7920, 0x4FFE, 0x4820, 0x7824, 0x4BA4, 0x4AA8, 0x7AA8, 0x290, 0x5392, 0x482A, 0x8846, 0x82}},
    {36084, 2, {0x1020, 0x1020, 0xFE78, 0x1088, 0x1110, 0xFEFC, 0x9284, 0x92FC, 0xFE84, 0x10FC, 0x3884, 0x54FC, 0x9200, 0x1048, 0x1084, 0x1102}},
    {36085, 2, {0x0, 0x7DFE, 0x4502, 0x457A, 0x7D02, 0x457A, 0x4502, 0x7C00, 0x44FC, 0x4484, 0x7CFC, 0x84, 0x28FC, 0x2484, 0x44FC, 0x8084}},
    {36086, 2, {0x20, 0x7C20, 0x4450, 0x4448, 0x7CA4, 0x45FE, 0x4684, 0x7CFC, 0x4484, 0x44FC, 0x7C80, 0xFC, 0x2944, 0x2544, 0x467C, 0x8044}},
    {36087, 2, {0x100, 0x7FFC, 0x100, 0x3D78, 0x2548, 0x3D78, 0x100, 0x7FFE, 0x5012, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36088, 2, {0x20, 0x3D24, 0x24A8, 0x25FC, 0x2440, 0x3FFE, 0x2488, 0x25FC, 0x268A, 0x3CF8, 0x2488, 0x24F8, 0x2488, 0x44F8, 0x5450, 0x8888}},
    {36089, 2, {0x104, 0x7884, 0x4888, 0x4800, 0x7BFE, 0x4800, 0x4888, 0x7904, 0x4A02, 0x49FC, 0x7954, 0x154, 0x5154, 0x4954, 0x8FFE, 0x0}},
    {36090, 2, {0x110, 0x78A0, 0x4800, 0x4BFC, 0x78A0, 0x4BF8, 0x48A8, 0x7BFE, 0x48A8, 0x4BF8, 0x78A0, 0x1B0, 0x52A8, 0x4CA6, 0x88A0, 0xA0}},
    {36091, 2, {0x28, 0x7C24, 0x47FE, 0x4420, 0x7DFC, 0x4524, 0x45FC, 0x7D24, 0x45FC, 0x4524, 0x7C08, 0x3FE, 0x2888, 0x2448, 0x4448, 0x8018}},
    {36092, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x79FC, 0x4888, 0x4BFE, 0x7820, 0x49FC, 0x4924, 0x79FC, 0x124, 0x53FE, 0x4904, 0x8914, 0x108}},
    {36093, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFE0, 0x440, 0x820}},
    {36094, 2, {0x20, 0x7DFE, 0x4820, 0x48FC, 0x4820, 0x5DFE, 0x5400, 0x54FC, 0x5484, 0x5CFC, 0x4884, 0x48FC, 0x4884, 0x7EFC, 0x48, 0x84}},
    {36095, 2, {0x0, 0x7BDE, 0x4842, 0x4A52, 0x794A, 0x4A52, 0x4828, 0x78C4, 0x4B12, 0x4860, 0x7988, 0x32, 0x51C4, 0x4818, 0x8860, 0x380}},
    {36096, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36097, 2, {0x7E20, 0x4220, 0x7E20, 0x423E, 0x7E44, 0x4244, 0x7E44, 0x24A4, 0x4228, 0x7E28, 0x4210, 0x7E10, 0x4228, 0x7E48, 0x2484, 0x4302}},
    {36098, 2, {0x40, 0x7888, 0x49FC, 0x4908, 0x7A52, 0x4BFE, 0x4850, 0x7988, 0x4E26, 0x48C0, 0x7B10, 0x64, 0x5388, 0x4830, 0x88C0, 0x700}},
    {36099, 2, {0x20, 0x7820, 0x49FC, 0x4924, 0x79FC, 0x4820, 0x4BFE, 0x7A22, 0x4BFE, 0x4820, 0x7840, 0x24, 0x52A2, 0x4A8A, 0x8C78, 0x0}},
    {36100, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x882, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36101, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x5FF2, 0x9010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36102, 2, {0x28, 0x7824, 0x4BFE, 0x4850, 0x7888, 0x4904, 0x4A02, 0x7998, 0x4954, 0x4910, 0x7BBE, 0x110, 0x52A8, 0x4AA8, 0x8C44, 0x882}},
    {36103, 2, {0x41D4, 0x2010, 0xFBFE, 0x1090, 0x52C8, 0x228A, 0x57E6, 0x8002, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820}},
    {36104, 2, {0x104, 0x7888, 0x4800, 0x4BFE, 0x7A22, 0x4AAA, 0x4A72, 0x7A22, 0x4BFE, 0x4800, 0x79FC, 0x104, 0x51FC, 0x4904, 0x89FC, 0x104}},
    {36105, 2, {0x0, 0x7BFE, 0x4850, 0x4BFE, 0x7A52, 0x4BFE, 0x4800, 0x79FC, 0x4904, 0x49FC, 0x7904, 0x1FC, 0x5020, 0x4BFE, 0x8820, 0x20}},
    {36106, 2, {0x2850, 0x3E7C, 0x4890, 0xFEFE, 0x1528, 0x264A, 0x4486, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36107, 2, {0x3FFE, 0x2250, 0x2488, 0x2DFE, 0x3690, 0x24FC, 0x2490, 0x24FE, 0x2000, 0x27FC, 0x2404, 0x27FC, 0x4404, 0x47FC, 0x8208, 0x404}},
    {36108, 2, {0x50, 0x7A52, 0x4954, 0x4850, 0x7BFE, 0x4888, 0x4850, 0x7BFE, 0x4820, 0x49FC, 0x7820, 0x3FE, 0x5050, 0x4888, 0x8904, 0x602}},
    {36109, 2, {0x80, 0x78F8, 0x4908, 0x4BFE, 0x7944, 0x4992, 0x49FE, 0x7900, 0x497C, 0x4900, 0x797C, 0x100, 0x517C, 0x4A44, 0x8A7C, 0x444}},
    {36110, 2, {0x88, 0x7888, 0x4BFE, 0x4888, 0x79FC, 0x4924, 0x49FC, 0x7924, 0x49FC, 0x4820, 0x7BFE, 0x222, 0x522A, 0x4AFA, 0x8A0A, 0x206}},
    {36111, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x77DC, 0x5454, 0x77D4, 0x545C, 0x77D6, 0x5296, 0xB462}},
    {36112, 2, {0x20, 0x79FC, 0x4824, 0x4BFE, 0x7824, 0x49FC, 0x4820, 0x7BFE, 0x4800, 0x4AA4, 0x7C52, 0x1FC, 0x5154, 0x4954, 0x8FFE, 0x0}},
    {36113, 2, {0x1000, 0xFEFC, 0x4420, 0x2820, 0xFEFE, 0x0, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x1084, 0xFEFC, 0x1048, 0x1084, 0x1102}},
    {36114, 2, {0x7E20, 0x51FC, 0x7D24, 0x45FC, 0x7C24, 0x514A, 0x7EF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820}},
    {36115, 2, {0xC, 0x700A, 0x51FE, 0x5508, 0x75E8, 0x5548, 0x574A, 0x71EA, 0x512A, 0x5FEC, 0x754C, 0x54C, 0x55EA, 0x4A0A, 0x8216, 0x422}},
    {36116, 2, {0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x2448, 0x4284}},
    {36117, 2, {0x2010, 0xCEFE, 0x8A10, 0xAA7C, 0xAA00, 0xEEFE, 0xAAAA, 0xAAFE, 0xAA44, 0xAA7C, 0xEE44, 0xA87C, 0x2844, 0x487C, 0x4828, 0x8844}},
    {36118, 2, {0x20, 0x7BFE, 0x4820, 0x49FC, 0x7800, 0x4BFE, 0x4A52, 0x7BFE, 0x4904, 0x49FC, 0x7904, 0x1FC, 0x5104, 0x49FC, 0x8888, 0x104}},
    {36119, 2, {0x3FFE, 0x2240, 0x25F8, 0x2D08, 0x35F8, 0x2508, 0x25FE, 0x26AA, 0x2406, 0x27F8, 0x2408, 0x27F8, 0x4408, 0x47F8, 0x8210, 0x408}},
    {36120, 2, {0x124, 0x78A8, 0x4BFE, 0x4A02, 0x78F8, 0x4888, 0x49FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x104, 0x51FC, 0x4800, 0x8888, 0x104}},
    {36121, 2, {0xE1C, 0x810, 0x7EFC, 0x4A94, 0x5CB8, 0x4A94, 0x5CB8, 0x952A, 0x2246, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x1010}},
    {36122, 2, {0x110, 0x7890, 0x4BDE, 0x4810, 0x7A5E, 0x4982, 0x4BDE, 0x7810, 0x4BDE, 0x4A50, 0x7BDE, 0x250, 0x53DE, 0x4A50, 0x8A52, 0x2CE}},
    {36123, 2, {0x1040, 0xFE7C, 0x44C8, 0x2830, 0xFECE, 0x78, 0x7C20, 0x45FC, 0x7C88, 0x44F8, 0x7C88, 0x10F8, 0xFE88, 0x10F8, 0x1050, 0x1088}},
    {36124, 2, {0x90, 0x7090, 0x57FE, 0x5094, 0x700A, 0x51FE, 0x5508, 0x77E8, 0x514A, 0x51EA, 0x7F2A, 0x5EA, 0x554C, 0x49EA, 0x8216, 0x462}},
    {36125, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x280, 0x440, 0x820, 0x1010, 0x2008}},
    {36126, 2, {0x100, 0x100, 0x1FC, 0x100, 0x100, 0x3FF8, 0x2008, 0x2108, 0x2108, 0x2108, 0x2208, 0x2208, 0x4C0, 0x830, 0x3008, 0xC004}},
    {36127, 2, {0x400, 0x400, 0xFE0, 0x1020, 0x2040, 0x5FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1210, 0x2C0, 0x420, 0x1810, 0x6008}},
    {36128, 2, {0x200, 0x400, 0x820, 0x1010, 0x3FF8, 0x8, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1210, 0x4C0, 0x1830, 0x6008}},
    {36129, 2, {0x0, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1210, 0x4C0, 0x1830, 0x6008}},
    {36130, 2, {0x10, 0x7C10, 0x4410, 0x5410, 0x55FE, 0x5410, 0x5430, 0x5430, 0x5450, 0x5450, 0x5490, 0x1110, 0x2810, 0x2410, 0x4450, 0x8020}},
    {36131, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36132, 2, {0x400, 0x25F8, 0x2488, 0x2450, 0x2420, 0x24D8, 0x2706, 0x400, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36133, 2, {0x40, 0x7C40, 0x4440, 0x5480, 0x54FE, 0x5508, 0x5688, 0x5488, 0x5488, 0x5450, 0x5450, 0x1020, 0x2850, 0x2488, 0x4504, 0x8202}},
    {36134, 2, {0x80, 0x7C84, 0x4484, 0x5488, 0x5490, 0x54A0, 0x5480, 0x55FE, 0x54A0, 0x54A0, 0x5490, 0x1090, 0x2888, 0x24A4, 0x44C2, 0x8080}},
    {36135, 2, {0x880, 0x880, 0x1098, 0x30E0, 0x5784, 0x9084, 0x107C, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36136, 2, {0x7C, 0x3FC0, 0x2040, 0x2040, 0x3FFE, 0x2040, 0x2040, 0x27FC, 0x2404, 0x2444, 0x2444, 0x2444, 0x24B4, 0x4108, 0x4604, 0x9802}},
    {36137, 2, {0x4, 0x7C0E, 0x44F0, 0x4480, 0x5480, 0x54FC, 0x54A4, 0x54A4, 0x54A4, 0x54A8, 0x54A8, 0x1090, 0x2910, 0x2528, 0x4244, 0x8082}},
    {36138, 2, {0x100, 0x280, 0x440, 0x920, 0x3098, 0xCFE6, 0x40, 0x80, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36139, 2, {0x440, 0x820, 0x3018, 0xDFE6, 0x420, 0x420, 0x8A0, 0x1040, 0x3FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36140, 2, {0x8, 0x7C3C, 0x45C0, 0x4400, 0x5420, 0x5410, 0x55FC, 0x5404, 0x5408, 0x5410, 0x5420, 0x1040, 0x2880, 0x2540, 0x423E, 0x8000}},
    {36141, 2, {0x80, 0x7C80, 0x4480, 0x54FC, 0x5504, 0x5504, 0x5644, 0x5444, 0x5484, 0x54A4, 0x5514, 0x11F4, 0x2814, 0x2404, 0x4428, 0x8010}},
    {36142, 2, {0x40, 0x7C20, 0x4420, 0x5400, 0x55FE, 0x5502, 0x5604, 0x5400, 0x5400, 0x5400, 0x5400, 0x1000, 0x2800, 0x25FE, 0x4400, 0x8000}},
    {36143, 2, {0x1FF0, 0x1110, 0x1110, 0xFFFE, 0x1110, 0x1110, 0x1FF0, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36144, 2, {0x28, 0x24, 0xFFFE, 0x20, 0x3FA0, 0x20, 0x7FA0, 0x20, 0x3FA0, 0x2090, 0x2490, 0x2490, 0x248A, 0xA0A, 0x1106, 0x6082}},
    {36145, 2, {0x50, 0x7C48, 0x4440, 0x445C, 0x55E0, 0x5440, 0x545E, 0x57E0, 0x5444, 0x5448, 0x5430, 0x1022, 0x2852, 0x248A, 0x4306, 0x8002}},
    {36146, 2, {0x100, 0x3FF8, 0x100, 0x820, 0xFFFE, 0x820, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1210, 0x4C0, 0x1830, 0x6008}},
    {36147, 2, {0x1220, 0x1220, 0xFFFE, 0x1220, 0x13E0, 0x1000, 0x1FF8, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36148, 2, {0x20, 0x7C20, 0x4420, 0x5420, 0x543E, 0x5420, 0x5420, 0x5420, 0x55FC, 0x5504, 0x5504, 0x1104, 0x2904, 0x2504, 0x45FC, 0x8104}},
    {36149, 2, {0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36150, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x5504, 0x5504, 0x55FC, 0x5450, 0x5450, 0x5450, 0x5450, 0x1092, 0x2892, 0x2512, 0x420E, 0x8400}},
    {36151, 2, {0x890, 0x1088, 0x307E, 0xD7C0, 0x1022, 0x101A, 0x6, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1210, 0x4C0, 0x1830, 0x6008}},
    {36152, 2, {0x600, 0x78FC, 0x4044, 0x4844, 0x4444, 0x5A94, 0x6108, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36153, 2, {0x440, 0x7FF8, 0x448, 0x3FF8, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2008, 0x5FF0, 0x1010, 0x1110, 0x1110, 0x2C0, 0xC30, 0x3008}},
    {36154, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x2244, 0x2A7C, 0x4400, 0x9FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1210, 0x4C0, 0x1830, 0x6008}},
    {36155, 2, {0x20, 0x7C20, 0x4420, 0x5440, 0x5448, 0x5484, 0x55FE, 0x5482, 0x5400, 0x54FC, 0x5484, 0x1084, 0x2884, 0x2484, 0x44FC, 0x8084}},
    {36156, 2, {0x14, 0x7C12, 0x4410, 0x45FE, 0x5410, 0x5490, 0x5492, 0x5492, 0x57F2, 0x5494, 0x5494, 0x1088, 0x288A, 0x251A, 0x4126, 0x8242}},
    {36157, 2, {0x1080, 0x1080, 0xFBF0, 0x1090, 0x1990, 0xF094, 0x1154, 0x520C, 0x2404, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x2C0, 0xC30, 0x3008}},
    {36158, 2, {0x0, 0x7FFC, 0x440, 0x3FF8, 0x2448, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36159, 2, {0x20, 0x7C20, 0x45FE, 0x5440, 0x5440, 0x54FC, 0x5484, 0x5584, 0x56FC, 0x5484, 0x5484, 0x10FC, 0x2884, 0x2484, 0x4494, 0x8088}},
    {36160, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36161, 2, {0x838, 0x13C0, 0x3040, 0x5040, 0x97FC, 0x1040, 0x1040, 0x13F8, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x2C0, 0xC30, 0x3008}},
    {36162, 2, {0x40, 0x7C40, 0x4478, 0x5488, 0x5550, 0x5420, 0x5450, 0x5488, 0x5506, 0x54F8, 0x5488, 0x1088, 0x2888, 0x2488, 0x44F8, 0x8088}},
    {36163, 2, {0x20, 0x7C10, 0x4410, 0x45FE, 0x5500, 0x5510, 0x5510, 0x5510, 0x557C, 0x5510, 0x5510, 0x1110, 0x2910, 0x2610, 0x42FE, 0x8400}},
    {36164, 2, {0x100, 0x4100, 0x21FC, 0xA44, 0x1448, 0xE0A0, 0x2110, 0x260C, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36165, 2, {0x20, 0x7C10, 0x4410, 0x47FE, 0x5420, 0x5442, 0x5484, 0x55F8, 0x5410, 0x5422, 0x54C4, 0x1308, 0x2810, 0x2428, 0x40C4, 0x8302}},
    {36166, 2, {0x0, 0x7CFC, 0x4484, 0x5484, 0x54FC, 0x5490, 0x5490, 0x5488, 0x5484, 0x5532, 0x5508, 0x1200, 0x2860, 0x2418, 0x4404, 0x8000}},
    {36167, 2, {0x28, 0x7C24, 0x4424, 0x5420, 0x55FE, 0x5420, 0x5520, 0x54B2, 0x54B4, 0x5468, 0x54A8, 0x1124, 0x2A22, 0x2420, 0x44A0, 0x8040}},
    {36168, 2, {0x0, 0x7DFE, 0x4500, 0x4500, 0x557C, 0x5500, 0x5500, 0x55FE, 0x5550, 0x5552, 0x5554, 0x1148, 0x2948, 0x2644, 0x4252, 0x8460}},
    {36169, 2, {0x100, 0x100, 0x7FFC, 0x1110, 0x920, 0xFFFE, 0x540, 0x920, 0x3018, 0xDFF6, 0x1010, 0x1110, 0x1110, 0x2C0, 0xC30, 0x3008}},
    {36170, 2, {0x20, 0x7C20, 0x4450, 0x5488, 0x5504, 0x56FA, 0x5400, 0x5400, 0x57FE, 0x5420, 0x5528, 0x1124, 0x2A22, 0x2422, 0x44A0, 0x8040}},
    {36171, 2, {0x8, 0x7C0C, 0x44EA, 0x440A, 0x5408, 0x55FE, 0x5448, 0x5448, 0x5548, 0x5568, 0x5548, 0x1148, 0x294A, 0x256A, 0x4386, 0x8102}},
    {36172, 2, {0x40, 0x40, 0x7C44, 0x45FC, 0x4448, 0x5450, 0x57FE, 0x5440, 0x54F8, 0x5588, 0x5688, 0x54F8, 0x2888, 0x2488, 0x44F8, 0x8088}},
    {36173, 2, {0x100, 0x7FFC, 0x920, 0x920, 0x1550, 0x7FFE, 0x4002, 0x8004, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36174, 2, {0x20, 0x7C20, 0x45FC, 0x4420, 0x5420, 0x57FE, 0x5402, 0x5494, 0x5450, 0x5510, 0x5490, 0x13FE, 0x2828, 0x2444, 0x4082, 0x8302}},
    {36175, 2, {0x1110, 0x920, 0x3FFC, 0x2004, 0x4FE8, 0x820, 0xFE0, 0x0, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36176, 2, {0x0, 0x7CFC, 0x4484, 0x5484, 0x54FC, 0x5484, 0x5484, 0x54FC, 0x5440, 0x54FE, 0x552A, 0x124A, 0x2892, 0x2522, 0x444A, 0x8084}},
    {36177, 2, {0x1FF0, 0x1010, 0x1110, 0x1110, 0x2C0, 0xC30, 0x7008, 0x0, 0x3EF8, 0x2288, 0x2AA8, 0x2AA8, 0x2AA8, 0x1450, 0x2288, 0x4104}},
    {36178, 2, {0x0, 0x7DFC, 0x4524, 0x4524, 0x5574, 0x5524, 0x55FC, 0x5504, 0x5574, 0x5554, 0x5554, 0x1174, 0x2904, 0x2604, 0x4214, 0x8408}},
    {36179, 2, {0x80, 0x3FFE, 0x2080, 0x2FF8, 0x2088, 0x3FFE, 0x2108, 0x2FF8, 0x2410, 0x2FF8, 0x3416, 0x2490, 0x4490, 0x4160, 0x8210, 0xC08}},
    {36180, 2, {0x40, 0x7C20, 0x45FC, 0x5400, 0x5508, 0x5490, 0x5400, 0x57FE, 0x5400, 0x5400, 0x55FC, 0x1104, 0x2904, 0x2504, 0x45FC, 0x8104}},
    {36181, 2, {0x20, 0x7CA2, 0x44A2, 0x4524, 0x5450, 0x5488, 0x5704, 0x5422, 0x5420, 0x54A4, 0x54A4, 0x1128, 0x2850, 0x2488, 0x4104, 0x8202}},
    {36182, 2, {0x1080, 0x1080, 0xFEF8, 0x1108, 0x1210, 0x7DFC, 0x5504, 0x5524, 0x5524, 0x7D24, 0x1124, 0x3924, 0x5450, 0x9088, 0x1104, 0x1202}},
    {36183, 2, {0x0, 0x7DFE, 0x4502, 0x457A, 0x5502, 0x557A, 0x5502, 0x5400, 0x54FC, 0x5484, 0x54FC, 0x1084, 0x28FC, 0x2484, 0x40FC, 0x8084}},
    {36184, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x4682, 0x9FF0, 0x1010, 0x1110, 0x2C0, 0xC30, 0x7008}},
    {36185, 2, {0x28, 0x7C24, 0x47FE, 0x5420, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x55FC, 0x5524, 0x5408, 0x13FE, 0x2888, 0x2448, 0x4448, 0x8018}},
    {36186, 2, {0x88, 0x7C50, 0x4400, 0x55FE, 0x5450, 0x55FC, 0x5454, 0x57FE, 0x5454, 0x55FC, 0x5450, 0x10D8, 0x2954, 0x2652, 0x4450, 0x8050}},
    {36187, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1FF0, 0x2928, 0xC926, 0x920, 0xAA0, 0x440, 0x1820}},
    {36188, 2, {0x20, 0x7DFE, 0x4820, 0x48FC, 0x4820, 0x5DFE, 0x5400, 0x55FC, 0x5504, 0x5D24, 0x4924, 0x4924, 0x4924, 0x7E50, 0x88, 0x304}},
    {36189, 2, {0x3FFE, 0x2250, 0x24FE, 0x2D90, 0x36FC, 0x2490, 0x24FC, 0x2490, 0x24FE, 0x2000, 0x27FC, 0x2444, 0x4444, 0x40B0, 0x830C, 0x1C02}},
    {36190, 2, {0x810, 0x2850, 0x3E7C, 0x4890, 0x7EFE, 0x1428, 0x2548, 0x468A, 0x8106, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x2C0, 0xC30, 0x3008}},
    {36191, 2, {0x41D4, 0x2010, 0xFBFE, 0x1090, 0x52C8, 0x228A, 0x57E6, 0x8002, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x260, 0xC18, 0x7004}},
    {36192, 2, {0x88, 0x7C48, 0x4450, 0x55FC, 0x5524, 0x55AC, 0x5574, 0x5524, 0x55FC, 0x5400, 0x54F8, 0x1088, 0x28F8, 0x2488, 0x44F8, 0x8088}},
    {36193, 2, {0x80, 0x7CF8, 0x4508, 0x47FE, 0x5544, 0x5592, 0x55FE, 0x5500, 0x557C, 0x5500, 0x557C, 0x1100, 0x297C, 0x2644, 0x427C, 0x8444}},
    {36194, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x77DC, 0x5454, 0x7554, 0x555C, 0x7556, 0x5296, 0xB462}},
    {36195, 2, {0x1040, 0xFE78, 0x44D0, 0x2920, 0xFE58, 0x186, 0x7CF8, 0x4420, 0x7DFC, 0x4400, 0x7CF8, 0x1088, 0xFEA8, 0x10A8, 0x1050, 0x1188}},
    {36196, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x440, 0x440, 0x2448, 0x2444, 0x4442, 0x8842, 0x840, 0x1140, 0x2080}},
    {36197, 2, {0x810, 0x810, 0x810, 0x7E10, 0x83E, 0x822, 0xFF44, 0x1490, 0x1410, 0x5610, 0x5510, 0x9528, 0x2428, 0x2424, 0x5444, 0x8882}},
    {36198, 2, {0x820, 0x820, 0x820, 0x7E40, 0x87E, 0x884, 0xFF44, 0x1444, 0x1444, 0x5628, 0x5528, 0x9510, 0x2410, 0x2428, 0x5444, 0x8882}},
    {36199, 2, {0x800, 0x8FC, 0x884, 0x7E84, 0x894, 0x888, 0xFF80, 0x14FC, 0x14A4, 0x56A4, 0x55A8, 0x94A8, 0x2490, 0x24A8, 0x54C4, 0x8882}},
    {36200, 2, {0x810, 0x810, 0x810, 0x7EFE, 0x892, 0x892, 0xFF92, 0x1492, 0x14FE, 0x5692, 0x5510, 0x9514, 0x2412, 0x241E, 0x54F2, 0x8800}},
    {36201, 2, {0x820, 0x820, 0x878, 0x7E48, 0x890, 0x87C, 0xFF54, 0x1454, 0x1454, 0x567C, 0x5540, 0x9540, 0x2442, 0x2442, 0x543E, 0x8800}},
    {36202, 2, {0x810, 0x810, 0x81E, 0x7E10, 0x810, 0x8FE, 0xFF82, 0x1492, 0x1492, 0x5692, 0x5592, 0x9492, 0x2428, 0x2424, 0x5442, 0x8882}},
    {36203, 2, {0x820, 0x820, 0x820, 0x7EF8, 0x820, 0x820, 0xFEFE, 0x1450, 0x1450, 0x56D8, 0x5554, 0x9452, 0x2450, 0x2490, 0x5490, 0x8930}},
    {36204, 2, {0x820, 0x820, 0x83E, 0x7E20, 0x820, 0x8FC, 0xFF84, 0x14FC, 0x1484, 0x56FC, 0x5584, 0x94FC, 0x2400, 0x2448, 0x5484, 0x8902}},
    {36205, 2, {0x820, 0x824, 0x8F4, 0x7E28, 0x828, 0x8FE, 0xFF10, 0x1420, 0x147C, 0x56C4, 0x5544, 0x957C, 0x2444, 0x2444, 0x547C, 0x8844}},
    {36206, 2, {0x1000, 0x11DC, 0x1144, 0x7D44, 0x1144, 0x11DC, 0xFF00, 0x2900, 0x29DC, 0x6D14, 0xAB14, 0x29D4, 0x2908, 0x4908, 0x4914, 0x9922}},
    {36207, 2, {0x1020, 0x1010, 0x11FE, 0x7D10, 0x117C, 0x1114, 0xFDFE, 0x2914, 0x297C, 0x6D10, 0xAB7C, 0x2944, 0x2944, 0x4A44, 0x4A7C, 0x9C44}},
    {36208, 2, {0x100, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {36209, 2, {0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0, 0x2010, 0x1020, 0xC0, 0x300, 0xC00, 0x1300, 0x20FE, 0x4000}},
    {36210, 2, {0x820, 0x820, 0x820, 0x7E20, 0x8FC, 0x824, 0xFF24, 0x824, 0x2824, 0x2844, 0x2E44, 0x2894, 0x2908, 0x5800, 0x4FFE, 0x8000}},
    {36211, 2, {0x804, 0x804, 0x844, 0x7E44, 0x844, 0x844, 0xFF44, 0x84C, 0x2854, 0x2864, 0x2E04, 0x2804, 0x2804, 0x5800, 0x4FFE, 0x8000}},
    {36212, 2, {0x420, 0x420, 0x420, 0x3F20, 0x430, 0x428, 0x7FA4, 0x424, 0x420, 0x2420, 0x27A0, 0x2420, 0x2420, 0x5400, 0x4FFE, 0x8000}},
    {36213, 2, {0x800, 0x804, 0x884, 0x7E44, 0x828, 0x828, 0xFE10, 0x810, 0x2828, 0x2828, 0x2E44, 0x2884, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {36214, 2, {0x800, 0x8FC, 0x820, 0x7E20, 0x820, 0x820, 0xFE20, 0x9FE, 0x2820, 0x2820, 0x2E20, 0x2820, 0x2820, 0x5820, 0x4FFE, 0x8000}},
    {36215, 2, {0x800, 0x800, 0x8F8, 0x7E08, 0x808, 0x808, 0xFEF8, 0x888, 0x2880, 0x2880, 0x2E84, 0x2884, 0x287C, 0x5800, 0x4FFE, 0x8000}},
    {36216, 2, {0x810, 0x810, 0x810, 0x7E10, 0x892, 0x892, 0xFE92, 0x892, 0x2892, 0x2892, 0x2E92, 0x28FE, 0x2802, 0x5800, 0x4FFE, 0x8000}},
    {36217, 2, {0x820, 0x820, 0x840, 0x7EFE, 0x902, 0x842, 0xFE22, 0x822, 0x280A, 0x2832, 0x2EC2, 0x2802, 0x2814, 0x5808, 0x4FFE, 0x8000}},
    {36218, 2, {0x820, 0x820, 0x850, 0x7E50, 0x888, 0x944, 0xFE22, 0x8F8, 0x2808, 0x2810, 0x2E10, 0x2820, 0x2820, 0x5800, 0x4FFE, 0x8000}},
    {36219, 2, {0x810, 0x810, 0x810, 0x7E54, 0x852, 0x852, 0xFE90, 0x812, 0x2804, 0x2808, 0x2E10, 0x2860, 0x2980, 0x5800, 0x4FFE, 0x8000}},
    {36220, 2, {0x420, 0x420, 0x420, 0x3F3E, 0x442, 0x404, 0x7F90, 0x410, 0x410, 0x2428, 0x27A4, 0x2442, 0x2482, 0x5400, 0x4FFE, 0x8000}},
    {36221, 2, {0x820, 0x820, 0x820, 0x7EFC, 0x824, 0x824, 0xFE24, 0x9FE, 0x2820, 0x2850, 0x2E48, 0x2884, 0x2904, 0x5800, 0x4FFE, 0x8000}},
    {36222, 2, {0x800, 0x80C, 0x870, 0x7E40, 0x840, 0x87E, 0xFF48, 0x848, 0x2848, 0x2848, 0x2E88, 0x2888, 0x2908, 0x5800, 0x4FFE, 0x8000}},
    {36223, 2, {0x1004, 0x101E, 0x11F0, 0x7D10, 0x1110, 0x1110, 0xFDFE, 0x1110, 0x1110, 0x5110, 0x5D10, 0x514A, 0x5186, 0x7102, 0x5000, 0x8FFE}},
    {36224, 2, {0x810, 0x810, 0x816, 0x7EF8, 0x890, 0x890, 0xFEFE, 0x812, 0x2832, 0x2852, 0x2E96, 0x2910, 0x2810, 0x5810, 0x4FFE, 0x8000}},
    {36225, 2, {0x820, 0x820, 0x850, 0x7E50, 0x888, 0x924, 0xFE42, 0x890, 0x2820, 0x2844, 0x2E88, 0x2810, 0x2820, 0x5840, 0x4FFE, 0x8000}},
    {36226, 2, {0x840, 0x840, 0x840, 0x7EFE, 0x882, 0x924, 0xFE20, 0x8A8, 0x28A4, 0x2922, 0x2E22, 0x28A0, 0x2840, 0x5800, 0x4FFE, 0x8000}},
    {36227, 2, {0x820, 0x8A0, 0x8A0, 0x7EFC, 0x8A0, 0x920, 0xFE20, 0x9FE, 0x2820, 0x2850, 0x2E48, 0x2884, 0x2904, 0x5800, 0x4FFE, 0x8000}},
    {36228, 2, {0x800, 0x8F8, 0x888, 0x7E88, 0x8F8, 0x888, 0xFF88, 0x8F8, 0x2888, 0x2888, 0x2E88, 0x29FE, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {36229, 2, {0x800, 0x9FC, 0x844, 0x7E44, 0x844, 0x894, 0xFF08, 0x8FC, 0x2884, 0x2884, 0x2E84, 0x28FC, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {36230, 2, {0x800, 0x81C, 0x8F0, 0x7E90, 0x890, 0x8FE, 0xFE90, 0x890, 0x2890, 0x28AA, 0x2ECA, 0x28A6, 0x2812, 0x5800, 0x4FFE, 0x8000}},
    {36231, 2, {0x820, 0x810, 0x810, 0x7EFE, 0x804, 0x844, 0xFF44, 0x828, 0x2828, 0x2828, 0x2E10, 0x28FE, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {36232, 2, {0x820, 0x820, 0x820, 0x7E3E, 0x820, 0x820, 0xFF20, 0x8FC, 0x2884, 0x2884, 0x2E84, 0x28FC, 0x2884, 0x5800, 0x4FFE, 0x8000}},
    {36233, 2, {0x810, 0x810, 0x892, 0x7E92, 0x892, 0x8FE, 0xFF10, 0x892, 0x2892, 0x2892, 0x2E92, 0x28FE, 0x2802, 0x5800, 0x4FFE, 0x8000}},
    {36234, 2, {0x1028, 0x1024, 0x1020, 0x7DFC, 0x1120, 0x1124, 0xFD24, 0x1128, 0x1128, 0x5110, 0x5D94, 0x512C, 0x7044, 0x5000, 0x4FFE, 0x8000}},
    {36235, 2, {0x840, 0x840, 0x878, 0x7E88, 0x910, 0x8FC, 0xFE04, 0x804, 0x28FC, 0x2804, 0x2E04, 0x29FC, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {36236, 2, {0x810, 0x810, 0x810, 0x7EFE, 0x810, 0x87C, 0xFF00, 0x87C, 0x2844, 0x2844, 0x2E44, 0x287C, 0x2844, 0x5800, 0x4FFE, 0x8000}},
    {36237, 2, {0x1020, 0x1020, 0x107C, 0x7C88, 0x1350, 0x1024, 0xFCC8, 0x131E, 0x1022, 0x50E4, 0x5C18, 0x5010, 0x5060, 0x7380, 0x5000, 0x8FFE}},
    {36238, 2, {0x810, 0x850, 0x850, 0x7EFC, 0x910, 0x810, 0xFEFE, 0x830, 0x2838, 0x2854, 0x2E94, 0x2910, 0x2810, 0x5800, 0x4FFE, 0x8000}},
    {36239, 2, {0x81C, 0x8E0, 0x820, 0x7E20, 0x9FE, 0x820, 0xFE20, 0x8FC, 0x2884, 0x2884, 0x2E84, 0x28FC, 0x2884, 0x5800, 0x4FFE, 0x8000}},
    {36240, 2, {0x800, 0x9DC, 0x844, 0x7E44, 0x954, 0x8CC, 0xFE44, 0x8CC, 0x2954, 0x2844, 0x2E44, 0x2854, 0x28C8, 0x5800, 0x4FFE, 0x8000}},
    {36241, 2, {0x1020, 0x1220, 0x1120, 0x7D3E, 0x1042, 0x1084, 0xFC90, 0x1090, 0x1310, 0x5110, 0x5D28, 0x5124, 0x5142, 0x7082, 0x5000, 0x8FFE}},
    {36242, 2, {0x810, 0x850, 0x850, 0x7E50, 0x954, 0x8D8, 0xFE50, 0x8D8, 0x2954, 0x2852, 0x2E92, 0x2892, 0x290E, 0x5A00, 0x4FFE, 0x8000}},
    {36243, 2, {0x1000, 0x10F8, 0x1088, 0x7C88, 0x1088, 0x1088, 0xFD26, 0x1220, 0x11FE, 0x5020, 0x5C70, 0x50A8, 0x5124, 0x7222, 0x5020, 0x8FFE}},
    {36244, 2, {0x1002, 0x13F2, 0x1082, 0x7C8A, 0x108A, 0x11EA, 0xFD2A, 0x112A, 0x12AA, 0x504A, 0x5C4A, 0x5082, 0x510A, 0x7204, 0x5000, 0x8FFE}},
    {36245, 2, {0x800, 0x87C, 0x844, 0x7E7C, 0x844, 0x87C, 0xFF00, 0x87C, 0x2810, 0x2810, 0x2EFE, 0x2810, 0x2810, 0x5810, 0x4FFE, 0x8000}},
    {36246, 2, {0x810, 0x810, 0x854, 0x7E54, 0x854, 0x8BA, 0xFF10, 0x810, 0x287C, 0x2810, 0x2E10, 0x28FE, 0x2800, 0x5800, 0x4FFE, 0x8000}},
    {36247, 2, {0x800, 0x8F8, 0x888, 0x7E88, 0x888, 0x8F8, 0xFE20, 0x8A0, 0x28BC, 0x28A0, 0x2EA0, 0x2960, 0x2A3C, 0x5800, 0x4FFE, 0x8000}},
    {36248, 2, {0x800, 0x8FE, 0x882, 0x7EFE, 0x880, 0x88C, 0xFEF0, 0x89C, 0x28F0, 0x289E, 0x2EF0, 0x2912, 0x290E, 0x5A00, 0x4FFE, 0x8000}},
    {36249, 2, {0x810, 0x892, 0x854, 0x7E10, 0x8FE, 0x882, 0xFEFE, 0x882, 0x28FE, 0x2882, 0x2E82, 0x288A, 0x2884, 0x5800, 0x4FFE, 0x8000}},
    {36250, 2, {0x820, 0x820, 0x9FE, 0x7E20, 0x820, 0x9FC, 0xFF24, 0x924, 0x29FC, 0x2870, 0x2EA8, 0x2924, 0x2A22, 0x5820, 0x4FFE, 0x8000}},
    {36251, 2, {0x820, 0x820, 0x850, 0x7E88, 0x904, 0xAFA, 0xFE20, 0x820, 0x29FC, 0x2820, 0x2EA8, 0x2870, 0x2BFE, 0x5800, 0x4FFE, 0x8000}},
    {36252, 2, {0x880, 0x880, 0x9FE, 0x7E22, 0x8AA, 0x872, 0xFE22, 0x9FE, 0x2822, 0x2872, 0x2EAA, 0x2922, 0x282A, 0x5804, 0x4FFE, 0x8000}},
    {36253, 2, {0x1020, 0x1020, 0x1050, 0x7C88, 0x1144, 0x1222, 0xFDF8, 0x1008, 0x1050, 0x5020, 0x5CA4, 0x528A, 0x528A, 0x7478, 0x5000, 0x8FFE}},
    {36254, 2, {0x848, 0x848, 0x8FC, 0x7E48, 0x848, 0x9FE, 0xFE00, 0x8FC, 0x2884, 0x28FC, 0x2E84, 0x28FC, 0x2884, 0x5800, 0x4FFE, 0x8000}},
    {36255, 2, {0x1020, 0x1124, 0x10A4, 0x7CA8, 0x11FC, 0x1104, 0xFD04, 0x1174, 0x1154, 0x5154, 0x5D74, 0x5104, 0x710C, 0x5000, 0x4FFE, 0x8000}},
    {36256, 2, {0x1020, 0x1020, 0x103E, 0x7C20, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x5124, 0x5C20, 0x53FE, 0x5020, 0x7020, 0x5020, 0x8FFE}},
    {36257, 2, {0x850, 0x848, 0x880, 0x7EFE, 0x990, 0xA90, 0xFEFC, 0x890, 0x2890, 0x28FC, 0x2E90, 0x2890, 0x28FE, 0x5880, 0x4FFE, 0x8000}},
    {36258, 2, {0x11F8, 0x1008, 0x10F8, 0x7C08, 0x13FE, 0x1020, 0xFC24, 0x1124, 0x10A8, 0x5070, 0x5CA8, 0x5324, 0x5024, 0x70A0, 0x5040, 0x8FFE}},
    {36259, 2, {0x1000, 0x13E0, 0x115E, 0x7D42, 0x11D2, 0x114A, 0xFD4A, 0x11C4, 0x1144, 0x5164, 0x5DCA, 0x534A, 0x7052, 0x5040, 0x4FFE, 0x8000}},
    {36260, 2, {0x820, 0x810, 0x8FE, 0x7E82, 0x800, 0x8FE, 0xFF20, 0x840, 0x28FC, 0x2944, 0x2E44, 0x287C, 0x2844, 0x5800, 0x4FFE, 0x8000}},
    {36261, 2, {0x1088, 0x1050, 0x13FE, 0x7C50, 0x1050, 0x11FC, 0xFD54, 0x118C, 0x1104, 0x51FC, 0x5D04, 0x5104, 0x51FC, 0x7104, 0x5000, 0x8FFE}},
    {36262, 2, {0x1220, 0x113E, 0x1042, 0x7C94, 0x1310, 0x1128, 0xFC46, 0x1000, 0x11FC, 0x5104, 0x5D04, 0x5104, 0x51FC, 0x7104, 0x5000, 0x8FFE}},
    {36263, 2, {0x800, 0x8FC, 0x884, 0x7EFC, 0x884, 0x8FC, 0xFE00, 0x9FE, 0x2810, 0x289C, 0x2E90, 0x2890, 0x297E, 0x5A00, 0x4FFE, 0x8000}},
    {36264, 2, {0x1080, 0x11FC, 0x1244, 0x7D54, 0x11F4, 0x1084, 0xFD18, 0x1080, 0x11FC, 0x5244, 0x5D54, 0x51F4, 0x5084, 0x7118, 0x5000, 0x8FFE}},
    {36265, 2, {0x1000, 0x11FC, 0x1124, 0x7DFC, 0x1124, 0x11FC, 0xFC88, 0x1088, 0x11FC, 0x5088, 0x5C88, 0x53FE, 0x5088, 0x7104, 0x5000, 0x8FFE}},
    {36266, 2, {0x1088, 0x1088, 0x11FC, 0x7C88, 0x1088, 0x13FE, 0xFC20, 0x11FC, 0x1124, 0x51FC, 0x5D24, 0x51FC, 0x5000, 0x7088, 0x5104, 0x8FFE}},
    {36267, 2, {0x103C, 0x13C0, 0x1040, 0x7BFE, 0x1090, 0x1108, 0xFAF4, 0x1492, 0x13FC, 0x5204, 0x5AF4, 0x5294, 0x52F4, 0x720C, 0x5000, 0x8FFE}},
    {36268, 2, {0x810, 0x87C, 0x810, 0x7EFE, 0x844, 0x8EE, 0xFE44, 0x8EE, 0x2800, 0x28FE, 0x2E28, 0x282A, 0x2846, 0x5880, 0x4FFE, 0x8000}},
    {36269, 2, {0x10A0, 0x1090, 0x11FE, 0x7D20, 0x13FC, 0x1120, 0xFDFC, 0x1120, 0x1120, 0x51FE, 0x5C00, 0x52A4, 0x5252, 0x7452, 0x5000, 0x8FFE}},
    {36270, 2, {0x800, 0x87C, 0x844, 0x7E7C, 0x800, 0x8EE, 0xFEAA, 0x8EE, 0x2810, 0x28FE, 0x2E38, 0x2854, 0x2892, 0x5810, 0x4FFE, 0x8000}},
    {36271, 2, {0x800, 0x9EE, 0x8AA, 0x7E66, 0x8AA, 0x850, 0xFEFE, 0x990, 0x28FC, 0x2890, 0x2EFC, 0x2890, 0x28FE, 0x5880, 0x4FFE, 0x8000}},
    {36272, 2, {0x1000, 0x13FE, 0x1124, 0x7A22, 0x17FE, 0x1376, 0xFAAA, 0x1376, 0x1222, 0x5376, 0x5AAA, 0x5376, 0x5222, 0x7206, 0x5000, 0x8FFE}},
    {36273, 2, {0x1108, 0x1528, 0x17BE, 0x7948, 0x17BE, 0x1318, 0xFDAA, 0x1946, 0x13F8, 0x5208, 0x5A48, 0x5248, 0x50A0, 0x7110, 0x5208, 0x8FFE}},
    {36274, 2, {0x1528, 0x17BE, 0x1948, 0x7FBE, 0x1318, 0x15AA, 0xF946, 0x13FC, 0x1204, 0x53FC, 0x5A04, 0x53FC, 0x5204, 0x73FC, 0x5108, 0x8FFE}},
    {36275, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x1100, 0x11F8, 0x1100, 0x2900, 0x2500, 0x43FE, 0x8000}},
    {36276, 2, {0x0, 0x7C20, 0x4520, 0x4520, 0x4520, 0x7D20, 0x1120, 0x1120, 0x5D20, 0x5110, 0x5110, 0x5210, 0x5E08, 0xE408, 0x804, 0x1002}},
    {36277, 2, {0x40, 0x7C40, 0x4440, 0x44FC, 0x4484, 0x7D04, 0x1004, 0x1104, 0x5C84, 0x5044, 0x5044, 0x5004, 0x5C04, 0xE004, 0x28, 0x10}},
    {36278, 2, {0x0, 0x7DFC, 0x4420, 0x4420, 0x4420, 0x7C20, 0x1020, 0x13FE, 0x5C20, 0x5020, 0x5020, 0x5020, 0x5C20, 0xE020, 0xA0, 0x40}},
    {36279, 2, {0x40, 0x7C40, 0x4480, 0x44FE, 0x4500, 0x7E00, 0x10FC, 0x1008, 0x5C10, 0x5020, 0x5040, 0x5080, 0x5D02, 0xE102, 0xFE, 0x0}},
    {36280, 2, {0x0, 0xFFFE, 0x400, 0x7F0, 0x810, 0x10A0, 0x6040, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x1100, 0x2900, 0x47FE}},
    {36281, 2, {0x20, 0x7C20, 0x4420, 0x45FC, 0x4424, 0x7C24, 0x1024, 0x1024, 0x5DFE, 0x5020, 0x5050, 0x5050, 0x5C88, 0xE088, 0x104, 0x202}},
    {36282, 2, {0x20, 0x7C20, 0x4420, 0x4420, 0x45FC, 0x7C20, 0x1020, 0x1020, 0x5DFE, 0x5020, 0x5050, 0x5050, 0x5C88, 0xE088, 0x104, 0x202}},
    {36283, 2, {0x20, 0x7C20, 0x4450, 0x4450, 0x4488, 0x7D44, 0x1222, 0x1020, 0x5C00, 0x51FC, 0x5004, 0x5008, 0x5C08, 0xE010, 0x10, 0x20}},
    {36284, 2, {0x0, 0x7DFE, 0x4448, 0x4448, 0x4448, 0x7C48, 0x1048, 0x13FE, 0x5C48, 0x5048, 0x5048, 0x5048, 0x5C88, 0xE088, 0x108, 0x208}},
    {36285, 2, {0x40, 0x7C20, 0x4420, 0x47FE, 0x4480, 0x7C80, 0x1080, 0x10FC, 0x5C84, 0x5084, 0x5084, 0x5084, 0x5D04, 0xE104, 0x228, 0x410}},
    {36286, 2, {0x20, 0x7C20, 0x4420, 0x4520, 0x4520, 0x7D20, 0x113C, 0x1120, 0x5D20, 0x5120, 0x5120, 0x5120, 0x5D20, 0xE120, 0x7FE, 0x0}},
    {36287, 2, {0x0, 0x7BFC, 0x4884, 0x4888, 0x4888, 0x7890, 0x109C, 0x1084, 0x5144, 0x5D44, 0x5128, 0x5128, 0x5A10, 0xE228, 0x444, 0x182}},
    {36288, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x4484, 0x7CFC, 0x1084, 0x1084, 0x5C84, 0x50FC, 0x5084, 0x5084, 0x5D04, 0xE104, 0x214, 0x408}},
    {36289, 2, {0x0, 0x7DFC, 0x4524, 0x4524, 0x4524, 0x7D24, 0x1124, 0x11FC, 0x1104, 0x5D00, 0x5100, 0x5100, 0x5102, 0x5D02, 0xE0FE, 0x0}},
    {36290, 2, {0x20, 0x7C20, 0x4420, 0x45FE, 0x4420, 0x7C20, 0x1020, 0x11FC, 0x5C84, 0x5088, 0x5048, 0x5050, 0x5C20, 0xE050, 0x188, 0x606}},
    {36291, 2, {0x8, 0x7C3C, 0x45E0, 0x4420, 0x4420, 0x7C20, 0x13FE, 0x1020, 0x5C20, 0x5050, 0x5050, 0x5050, 0x5C88, 0xE088, 0x104, 0x202}},
    {36292, 2, {0x20, 0x7C20, 0x4450, 0x4450, 0x4488, 0x7D04, 0x1202, 0x10F8, 0x5C88, 0x5088, 0x50A8, 0x5090, 0x5C82, 0xE082, 0x7E, 0x0}},
    {36293, 2, {0x8, 0x7C1C, 0x45E0, 0x4500, 0x4500, 0x7D00, 0x11FE, 0x1110, 0x5D10, 0x5130, 0x5118, 0x5114, 0x5D12, 0xE210, 0x210, 0x410}},
    {36294, 2, {0x20, 0x7C20, 0x4440, 0x4440, 0x4488, 0x7D04, 0x13FE, 0x1002, 0x1000, 0x5DFC, 0x5104, 0x5104, 0x5104, 0x5D04, 0xE1FC, 0x104}},
    {36295, 2, {0x48, 0x7D48, 0x4548, 0x4548, 0x4548, 0x7FFE, 0x1148, 0x1148, 0x5D48, 0x5148, 0x5178, 0x5100, 0x5D00, 0xE100, 0x1FE, 0x0}},
    {36296, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x4504, 0x7E12, 0x1020, 0x1040, 0x1188, 0x5C10, 0x5020, 0x5044, 0x5188, 0x5C10, 0xE060, 0x380}},
    {36297, 2, {0x20, 0x7C20, 0x4450, 0x4450, 0x4488, 0x7D24, 0x1212, 0x1010, 0x5DFC, 0x5004, 0x5008, 0x5088, 0x5C50, 0xE020, 0x10, 0x10}},
    {36298, 2, {0x20, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x7C20, 0x1020, 0x13FE, 0x5C70, 0x50A8, 0x50A8, 0x5124, 0x5D24, 0xE222, 0x20, 0x20}},
    {36299, 2, {0x50, 0x7C48, 0x4448, 0x4440, 0x47FE, 0x7C80, 0x1080, 0x10FC, 0x5D44, 0x5144, 0x5128, 0x5128, 0x5E10, 0xE228, 0x444, 0x182}},
    {36300, 2, {0x20, 0x7C20, 0x4520, 0x4520, 0x45FC, 0x7D20, 0x1220, 0x1020, 0x11FE, 0x5C20, 0x5050, 0x5050, 0x5088, 0x5C88, 0xE104, 0x202}},
    {36301, 2, {0x20, 0x7C20, 0x4420, 0x4420, 0x47FE, 0x7C20, 0x1020, 0x1020, 0x11FC, 0x5D04, 0x5104, 0x5104, 0x5104, 0x5D04, 0xE1FC, 0x104}},
    {36302, 2, {0x20, 0x7C10, 0x4410, 0x45FE, 0x4502, 0x7E04, 0x1080, 0x1088, 0x5C90, 0x50A0, 0x50C0, 0x5082, 0x5C82, 0xE082, 0x7E, 0x0}},
    {36303, 2, {0x100, 0x7900, 0x4900, 0x491C, 0x4BD4, 0x7954, 0x1154, 0x1154, 0x5154, 0x5D54, 0x5154, 0x5154, 0x5954, 0xE25C, 0x354, 0x480}},
    {36304, 2, {0x48, 0x7C48, 0x4448, 0x4548, 0x454A, 0x7D6C, 0x1148, 0x1148, 0x5D48, 0x5148, 0x5148, 0x514A, 0x5D6A, 0xE38A, 0x106, 0x0}},
    {36305, 2, {0x80, 0x7C80, 0x45FC, 0x4504, 0x4604, 0x7DF4, 0x1114, 0x1114, 0x5D14, 0x51F4, 0x5104, 0x5128, 0x5D12, 0xE102, 0xFE, 0x0}},
    {36306, 2, {0x0, 0x7C00, 0x45FE, 0x4408, 0x4408, 0x7DE8, 0x1128, 0x1128, 0x1128, 0x5D28, 0x51E8, 0x5128, 0x5008, 0x5C08, 0xE028, 0x10}},
    {36307, 2, {0x40, 0x7C20, 0x4400, 0x47FE, 0x4420, 0x7C20, 0x1020, 0x1020, 0x5DFC, 0x5020, 0x5020, 0x5020, 0x5C20, 0xE020, 0x3FE, 0x0}},
    {36308, 2, {0x80, 0x7C80, 0x4500, 0x45FC, 0x4604, 0x7C04, 0x11E4, 0x1124, 0x5D24, 0x5124, 0x51E4, 0x5124, 0x5C04, 0xE004, 0x28, 0x10}},
    {36309, 2, {0x20, 0x7C20, 0x4420, 0x4420, 0x443E, 0x7C20, 0x1020, 0x1020, 0x11FC, 0x5D04, 0x5104, 0x5104, 0x5104, 0x5D04, 0xE1FC, 0x104}},
    {36310, 2, {0x0, 0x7C00, 0x45FE, 0x4420, 0x4420, 0x7C40, 0x1040, 0x10FC, 0x1084, 0x5D84, 0x5284, 0x5084, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {36311, 2, {0x88, 0x7888, 0x4888, 0x4908, 0x497E, 0x7B08, 0x1508, 0x1148, 0x5128, 0x5D28, 0x5108, 0x5108, 0x5908, 0xE108, 0x128, 0x110}},
    {36312, 2, {0x20, 0x7C20, 0x4524, 0x44A4, 0x44A8, 0x7C20, 0x11FC, 0x1020, 0x5C20, 0x5020, 0x53FE, 0x5020, 0x5C20, 0xE020, 0x20, 0x20}},
    {36313, 2, {0x0, 0x7DF8, 0x4508, 0x4508, 0x4508, 0x7DF8, 0x1108, 0x1108, 0x5D08, 0x51F8, 0x5108, 0x5108, 0x5D08, 0xE108, 0x7FE, 0x0}},
    {36314, 2, {0x0, 0x7DDC, 0x4554, 0x4554, 0x4554, 0x7D54, 0x1154, 0x13FE, 0x5D54, 0x5154, 0x5154, 0x5154, 0x5D54, 0xE2D4, 0x224, 0x44C}},
    {36315, 2, {0x10, 0x7C10, 0x4410, 0x45FE, 0x4512, 0x7D14, 0x1110, 0x11FC, 0x5D44, 0x5144, 0x5128, 0x5128, 0x5D10, 0xE228, 0x244, 0x482}},
    {36316, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x7DFC, 0x1100, 0x1140, 0x5D44, 0x5148, 0x5170, 0x5140, 0x5D42, 0xE242, 0x23E, 0x400}},
    {36317, 2, {0x0, 0x7DFE, 0x4500, 0x4500, 0x4500, 0x7DFC, 0x1104, 0x1104, 0x5D04, 0x5104, 0x51FC, 0x5100, 0x5D00, 0xE100, 0x1FE, 0x0}},
    {36318, 2, {0x8, 0x7C1C, 0x45E0, 0x4500, 0x4520, 0x7D20, 0x1120, 0x11FE, 0x5C20, 0x5020, 0x50A8, 0x50A4, 0x5D22, 0xE222, 0xA0, 0x40}},
    {36319, 2, {0x0, 0x7DF8, 0x4508, 0x4508, 0x45F8, 0x7D08, 0x1108, 0x11F8, 0x5D44, 0x5148, 0x5130, 0x5120, 0x5D10, 0xE148, 0x186, 0x100}},
    {36320, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x45FC, 0x7C24, 0x11FC, 0x1120, 0x11FE, 0x5C22, 0x502A, 0x5054, 0x5050, 0x5C88, 0xE104, 0x202}},
    {36321, 2, {0x40, 0x7C20, 0x4420, 0x47FE, 0x4490, 0x7C90, 0x1090, 0x1294, 0x5E92, 0x5292, 0x5492, 0x5090, 0x5D10, 0xE110, 0x250, 0x420}},
    {36322, 2, {0x20, 0x7C20, 0x447C, 0x4484, 0x4548, 0x7C30, 0x1020, 0x1048, 0x5D90, 0x503E, 0x5042, 0x51A4, 0x5C18, 0xE010, 0x60, 0x180}},
    {36323, 2, {0x20, 0x7920, 0x4920, 0x49FC, 0x4920, 0x7A20, 0x1020, 0x13FE, 0x5090, 0x5C90, 0x5090, 0x5090, 0x5912, 0xE112, 0x212, 0x40E}},
    {36324, 2, {0x40, 0x7C20, 0x4420, 0x45FE, 0x4400, 0x7C88, 0x1104, 0x1202, 0x1088, 0x5C88, 0x5050, 0x5050, 0x5020, 0x5C50, 0xE088, 0x306}},
    {36325, 2, {0x0, 0x7DFC, 0x4444, 0x4448, 0x445E, 0x7C82, 0x108A, 0x1124, 0x5C20, 0x53FE, 0x5020, 0x5070, 0x5CA8, 0xE124, 0x222, 0x20}},
    {36326, 2, {0x20, 0x7D20, 0x4520, 0x45FC, 0x4520, 0x7E20, 0x1020, 0x13FE, 0x5C70, 0x50A8, 0x50A8, 0x5124, 0x5D24, 0xE222, 0x20, 0x20}},
    {36327, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x4504, 0x7E02, 0x11FC, 0x1020, 0x5C20, 0x5020, 0x51FC, 0x5020, 0x5C20, 0xE020, 0x3FE, 0x0}},
    {36328, 2, {0x20, 0x7820, 0x4BFE, 0x4850, 0x4888, 0x7904, 0x16FA, 0x1000, 0x53FE, 0x5C40, 0x5080, 0x51FC, 0x5804, 0xE004, 0x28, 0x10}},
    {36329, 2, {0x20, 0x7C20, 0x45FC, 0x4524, 0x4524, 0x7DFC, 0x1124, 0x1124, 0x5DFC, 0x5020, 0x5024, 0x5018, 0x5C32, 0xE0CA, 0x306, 0x2}},
    {36330, 2, {0x40, 0x7C40, 0x44FC, 0x4504, 0x4608, 0x7DFE, 0x1100, 0x117C, 0x5D44, 0x5144, 0x5154, 0x5148, 0x5D42, 0xE242, 0x23E, 0x400}},
    {36331, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4400, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x1100, 0x2900, 0x47FE}},
    {36332, 2, {0x20, 0x7C20, 0x45FC, 0x4420, 0x4420, 0x7C20, 0x13FE, 0x1000, 0x5C20, 0x5020, 0x51FC, 0x5020, 0x5C20, 0xE020, 0x3FE, 0x0}},
    {36333, 2, {0x80, 0x7880, 0x48FC, 0x4908, 0x4A90, 0x7860, 0x1198, 0x1626, 0x1020, 0x5DFC, 0x5020, 0x5220, 0x53FE, 0x5C20, 0xE020, 0x20}},
    {36334, 2, {0x0, 0x7DFE, 0x4420, 0x4440, 0x4488, 0x7D04, 0x11FE, 0x1022, 0x5C20, 0x5020, 0x51FE, 0x5020, 0x5C20, 0xE020, 0x3FE, 0x0}},
    {36335, 2, {0x40, 0x7C40, 0x4478, 0x4488, 0x4550, 0x7C20, 0x1050, 0x1088, 0x1106, 0x5CF8, 0x5088, 0x5088, 0x5088, 0x5C88, 0xE0F8, 0x88}},
    {36336, 2, {0x104, 0x7C84, 0x4488, 0x4400, 0x45FE, 0x7C88, 0x1088, 0x1088, 0x5C88, 0x53FE, 0x5088, 0x5088, 0x5D08, 0xE108, 0x208, 0x408}},
    {36337, 2, {0x20, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x7C20, 0x13FE, 0x1008, 0x1008, 0x5DFE, 0x5008, 0x5088, 0x5048, 0x5C08, 0xE028, 0x10}},
    {36338, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x4504, 0x7E02, 0x11FC, 0x1000, 0x1000, 0x5DFC, 0x5104, 0x5104, 0x5104, 0x5D04, 0xE1FC, 0x104}},
    {36339, 2, {0x50, 0x7C50, 0x4450, 0x4552, 0x44D4, 0x7C58, 0x1050, 0x1058, 0x10D4, 0x5D52, 0x5050, 0x5050, 0x5092, 0x5C92, 0xE112, 0x20E}},
    {36340, 2, {0x0, 0x7BFE, 0x4850, 0x4850, 0x4850, 0x79FC, 0x1154, 0x1154, 0x1D54, 0x5154, 0x515C, 0x5184, 0x5104, 0x5D04, 0xE1FC, 0x104}},
    {36341, 2, {0x50, 0x7C48, 0x4440, 0x445C, 0x45E0, 0x7C40, 0x105E, 0x11E0, 0x5C44, 0x5048, 0x5030, 0x5022, 0x5C52, 0xE08A, 0x306, 0x2}},
    {36342, 2, {0x10, 0x7A10, 0x4910, 0x4910, 0x48FE, 0x7810, 0x1310, 0x1128, 0x5124, 0x5D44, 0x5142, 0x5182, 0x5900, 0xE280, 0x47E, 0x0}},
    {36343, 2, {0x80, 0x7C80, 0x44BC, 0x47C0, 0x4450, 0x7C24, 0x10D4, 0x130C, 0x5C00, 0x53FE, 0x5090, 0x5090, 0x5D12, 0xE112, 0x20E, 0x400}},
    {36344, 2, {0x110, 0x7D12, 0x4514, 0x45D8, 0x4510, 0x7D12, 0x1152, 0x118E, 0x1120, 0x5C20, 0x53FE, 0x5020, 0x5020, 0x5C20, 0xE020, 0x20}},
    {36345, 2, {0x0, 0x7A0C, 0x4970, 0x4910, 0x4810, 0x7810, 0x10FE, 0x1710, 0x5110, 0x5D10, 0x5110, 0x5110, 0x5910, 0xE280, 0x47E, 0x0}},
    {36346, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x4488, 0x7C88, 0x1126, 0x1220, 0x5DFE, 0x5020, 0x5070, 0x50A8, 0x5D24, 0xE222, 0x20, 0x20}},
    {36347, 2, {0x40, 0x7C20, 0x45FE, 0x4488, 0x4450, 0x7C20, 0x1050, 0x1088, 0x1106, 0x5C88, 0x5088, 0x5088, 0x5088, 0x5D08, 0xE108, 0x208}},
    {36348, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x45FC, 0x7D00, 0x1100, 0x11FE, 0x5D02, 0x517A, 0x514A, 0x514A, 0x5E7A, 0xE202, 0x414, 0x8}},
    {36349, 2, {0x0, 0x7DF8, 0x4408, 0x4408, 0x45F8, 0x7D00, 0x1104, 0x1104, 0x5CFC, 0x5040, 0x5020, 0x50A4, 0x5E8A, 0xE28A, 0x478, 0x0}},
    {36350, 2, {0x1040, 0x1040, 0x24FE, 0x6488, 0xA550, 0x2420, 0x2450, 0x2088, 0x1FF4, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x1100, 0x2900, 0x47FE}},
    {36351, 2, {0x20, 0x7C20, 0x45FC, 0x4420, 0x4420, 0x7C20, 0x13FE, 0x1020, 0x5C20, 0x5120, 0x513C, 0x5120, 0x5EA0, 0xE260, 0x43E, 0x800}},
    {36352, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x4504, 0x7DFC, 0x1020, 0x1020, 0x5D20, 0x513C, 0x5120, 0x5120, 0x5EA0, 0xE260, 0x43E, 0x800}},
    {36353, 2, {0x0, 0x7DFE, 0x4400, 0x4492, 0x4524, 0x7E48, 0x1124, 0x1092, 0x1000, 0x5DFE, 0x5020, 0x5020, 0x5020, 0x5C20, 0xE3FE, 0x0}},
    {36354, 2, {0x0, 0x7BFE, 0x4920, 0x4928, 0x4928, 0x79E8, 0x1128, 0x1128, 0x5128, 0x5DE8, 0x5128, 0x512A, 0x593A, 0xE3EA, 0x26, 0x20}},
    {36355, 2, {0x20, 0x7D24, 0x44A4, 0x44A8, 0x4420, 0x7DFC, 0x1104, 0x1104, 0x5DFC, 0x5104, 0x5104, 0x51FC, 0x5D04, 0xE104, 0x114, 0x108}},
    {36356, 2, {0x10, 0x7C10, 0x4490, 0x449E, 0x4490, 0x7C90, 0x13FE, 0x1000, 0x5C10, 0x5092, 0x5092, 0x5114, 0x5E08, 0xE010, 0x60, 0x380}},
    {36357, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x1208, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36358, 2, {0x40, 0x7C40, 0x4488, 0x4504, 0x47FE, 0x7C02, 0x1088, 0x1144, 0x1242, 0x5CF8, 0x5188, 0x5250, 0x5020, 0x5C50, 0xE188, 0x606}},
    {36359, 2, {0x80, 0x7C80, 0x44FE, 0x4500, 0x46FC, 0x7C84, 0x10A4, 0x1094, 0x5DFE, 0x5084, 0x5124, 0x5114, 0x5DFE, 0xE004, 0x28, 0x10}},
    {36360, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x4420, 0x7DFC, 0x1124, 0x1124, 0x5DFC, 0x5020, 0x5070, 0x50A8, 0x5D24, 0xE222, 0x20, 0x20}},
    {36361, 2, {0x40, 0x7C20, 0x45FC, 0x4504, 0x4504, 0x7DFC, 0x1104, 0x1104, 0x11FC, 0x5D20, 0x5122, 0x5114, 0x5108, 0x5D44, 0xE182, 0x100}},
    {36362, 2, {0x0, 0x7DF8, 0x4408, 0x44D0, 0x4420, 0x7DFC, 0x1124, 0x1124, 0x5DFC, 0x5124, 0x5124, 0x51FC, 0x5D24, 0xE124, 0x124, 0x10C}},
    {36363, 2, {0x100, 0xF100, 0x911E, 0x9392, 0x9112, 0xF112, 0x2112, 0x27D2, 0xA112, 0xB912, 0xA21A, 0xA294, 0xA7D0, 0xB250, 0xC010, 0x10}},
    {36364, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x45FC, 0x7C40, 0x13FE, 0x1048, 0x5C88, 0x50FE, 0x5108, 0x5148, 0x5E28, 0xE408, 0x28, 0x10}},
    {36365, 2, {0x20, 0x7C22, 0x45FA, 0x4424, 0x4424, 0x7DFE, 0x1010, 0x1020, 0x11FC, 0x5C88, 0x5110, 0x52FE, 0x5410, 0x5C10, 0xE050, 0x20}},
    {36366, 2, {0x0, 0x7DFE, 0x4410, 0x4420, 0x4468, 0x7CA4, 0x1122, 0x1020, 0x1000, 0x5CFC, 0x5084, 0x5084, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {36367, 2, {0x20, 0x7822, 0x4BB4, 0x48A8, 0x48A8, 0x7924, 0x12A2, 0x1040, 0x51FC, 0x5D04, 0x5104, 0x51FC, 0x5904, 0xE104, 0x1FC, 0x104}},
    {36368, 2, {0x48, 0x7C44, 0x445E, 0x45E0, 0x4428, 0x7C12, 0x106A, 0x1196, 0x1048, 0x5C5E, 0x51E0, 0x5024, 0x5028, 0x5C12, 0xE06A, 0x186}},
    {36369, 2, {0x88, 0x7C88, 0x47FE, 0x4488, 0x4488, 0x7CF8, 0x1088, 0x1088, 0x5CF8, 0x5088, 0x5088, 0x53FE, 0x5C00, 0xE088, 0x104, 0x202}},
    {36370, 2, {0x1C, 0x7DE0, 0x4420, 0x4420, 0x47FE, 0x7CA8, 0x1124, 0x1242, 0x5C40, 0x53FE, 0x5088, 0x5108, 0x5CD0, 0xE030, 0x48, 0x184}},
    {36371, 2, {0xA0, 0x7C90, 0x4490, 0x45FE, 0x4520, 0x7D20, 0x13FC, 0x1120, 0x1120, 0x5DFC, 0x5120, 0x5120, 0x5120, 0x5DFE, 0xE100, 0x100}},
    {36372, 2, {0x20, 0x7C20, 0x443E, 0x4420, 0x45FC, 0x7D04, 0x11FC, 0x1104, 0x5DFC, 0x5124, 0x5020, 0x53FE, 0x5C20, 0xE020, 0x20, 0x20}},
    {36373, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x45FC, 0x7C24, 0x13FE, 0x1024, 0x5DFC, 0x5020, 0x5120, 0x513E, 0x5D20, 0xE2A0, 0x27E, 0x400}},
    {36374, 2, {0x48, 0x7C48, 0x4448, 0x45FE, 0x4448, 0x7C48, 0x11FE, 0x1000, 0x10FC, 0x5C84, 0x5084, 0x50FC, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {36375, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x4544, 0x7E22, 0x11F8, 0x1008, 0x5C50, 0x5020, 0x50A4, 0x5282, 0x5E8A, 0xE28A, 0x478, 0x0}},
    {36376, 2, {0x100, 0x7D00, 0x45FC, 0x4604, 0x4444, 0x7D54, 0x10E4, 0x1044, 0x13FC, 0x5C44, 0x50E4, 0x5154, 0x5244, 0x5C44, 0xE014, 0x8}},
    {36377, 2, {0x0, 0x7BE0, 0x495C, 0x4954, 0x49D4, 0x7954, 0x1154, 0x11D4, 0x5154, 0x5D54, 0x5168, 0x53C8, 0x5848, 0xE054, 0x54, 0x62}},
    {36378, 2, {0x20, 0x7C20, 0x4450, 0x4488, 0x4504, 0x7EFA, 0x1000, 0x11FC, 0x1154, 0x5D54, 0x51FC, 0x5154, 0x5154, 0x5D54, 0xE104, 0x10C}},
    {36379, 2, {0x20, 0x7C20, 0x45FC, 0x4420, 0x4420, 0x7DFE, 0x1088, 0x1104, 0x5E22, 0x5020, 0x51FC, 0x5020, 0x5C20, 0xE020, 0x3FE, 0x0}},
    {36380, 2, {0x20, 0x7820, 0x49FC, 0x4820, 0x4820, 0x7BFE, 0x1088, 0x1144, 0x1242, 0x5CF8, 0x5188, 0x5250, 0x5020, 0x5C50, 0xE188, 0x606}},
    {36381, 2, {0x0, 0x7DFC, 0x4524, 0x4524, 0x45FC, 0x7D24, 0x1124, 0x11FC, 0x1020, 0x5DFE, 0x5070, 0x50A8, 0x5124, 0x5E22, 0xE020, 0x20}},
    {36382, 2, {0x0, 0x7DFE, 0x4502, 0x4502, 0x45FE, 0x7D10, 0x1110, 0x11FE, 0x5D10, 0x5110, 0x517E, 0x5142, 0x5D42, 0xE242, 0x27E, 0x442}},
    {36383, 2, {0x100, 0x7900, 0x4900, 0x4BDE, 0x4A92, 0x7C92, 0x1092, 0x1092, 0x53F2, 0x5C92, 0x5092, 0x5152, 0x595E, 0xE220, 0x420, 0x800}},
    {36384, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4D04, 0x7900, 0x11DE, 0x1252, 0x1E52, 0x5352, 0x549A, 0x5094, 0x5110, 0x5D12, 0xE212, 0x40E}},
    {36385, 2, {0x20, 0x7D24, 0x44A8, 0x4420, 0x45FC, 0x7C40, 0x13FE, 0x1088, 0x5D04, 0x52FA, 0x5488, 0x5088, 0x5CA8, 0xE092, 0x82, 0x7E}},
    {36386, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x44FC, 0x7C84, 0x1084, 0x10FC, 0x5C40, 0x50FE, 0x512A, 0x524A, 0x5C92, 0xE122, 0x4A, 0x84}},
    {36387, 2, {0x40, 0x7C20, 0x45FC, 0x4400, 0x4508, 0x7C90, 0x1000, 0x13FE, 0x1000, 0x5C00, 0x51FC, 0x5104, 0x5104, 0x5D04, 0xE1FC, 0x104}},
    {36388, 2, {0x40, 0x7C20, 0x45FE, 0x4400, 0x4488, 0x7C88, 0x1154, 0x1222, 0x1000, 0x5C20, 0x53FE, 0x5020, 0x5020, 0x5C20, 0xE020, 0x20}},
    {36389, 2, {0x40, 0x7C20, 0x45FC, 0x4400, 0x4488, 0x7C50, 0x13FE, 0x1040, 0x5C40, 0x53FE, 0x5088, 0x5188, 0x5C50, 0xE030, 0xC8, 0x304}},
    {36390, 2, {0x20, 0x7C20, 0x45FC, 0x4450, 0x4488, 0x7D04, 0x13FE, 0x1008, 0x5DE8, 0x5128, 0x5128, 0x51E8, 0x5D28, 0xE008, 0x28, 0x10}},
    {36391, 2, {0x80, 0x7C80, 0x449C, 0x44F4, 0x4494, 0x7C94, 0x13F4, 0x1014, 0x1094, 0x5C94, 0x51C8, 0x52A8, 0x5088, 0x5C94, 0xE094, 0x1A2}},
    {36392, 2, {0x88, 0x7888, 0x4888, 0x4888, 0x4954, 0x7A22, 0x1442, 0x1020, 0x5020, 0x5D20, 0x513C, 0x5120, 0x5AA0, 0xE260, 0x43E, 0x800}},
    {36393, 2, {0x8, 0x7C3C, 0x45C0, 0x4444, 0x4524, 0x7CA8, 0x1080, 0x1020, 0x5DFE, 0x5070, 0x50A8, 0x50A8, 0x5D24, 0xE222, 0x420, 0x20}},
    {36394, 2, {0x40, 0x7820, 0x4BFE, 0x4A02, 0x4C04, 0x79F8, 0x1000, 0x1000, 0x53FE, 0x5C20, 0x5128, 0x5124, 0x5A22, 0xE422, 0xA0, 0x40}},
    {36395, 2, {0x108, 0x7C90, 0x4400, 0x45FC, 0x4490, 0x7C90, 0x1092, 0x1292, 0x1194, 0x5D98, 0x5090, 0x5090, 0x5090, 0x5C90, 0xE3FE, 0x0}},
    {36396, 2, {0x1C, 0x7DE0, 0x4510, 0x4510, 0x45FE, 0x7D10, 0x1110, 0x117C, 0x5D44, 0x5154, 0x5154, 0x5154, 0x5D54, 0xE228, 0x244, 0x482}},
    {36397, 2, {0x80, 0x7C80, 0x44F8, 0x4508, 0x4610, 0x7DFC, 0x1024, 0x1024, 0x5DFE, 0x5024, 0x5024, 0x51FC, 0x5C24, 0xE020, 0xA0, 0x40}},
    {36398, 2, {0x20, 0x7C10, 0x45FE, 0x4500, 0x4520, 0x7D20, 0x113C, 0x1120, 0x5D20, 0x517C, 0x5144, 0x5144, 0x5D44, 0xE244, 0x27C, 0x444}},
    {36399, 2, {0x440, 0xF25E, 0x9292, 0x9012, 0x97D4, 0xF114, 0x2118, 0x2114, 0xA7F2, 0xB912, 0xA112, 0xA11A, 0xBA94, 0xE250, 0x450, 0x810}},
    {36400, 2, {0x20, 0x7C50, 0x4488, 0x4504, 0x46FA, 0x7C00, 0x11E2, 0x112A, 0x5D2A, 0x51EA, 0x512A, 0x512A, 0x5DEA, 0xE122, 0x12A, 0x164}},
    {36401, 2, {0x40, 0x7C20, 0x45FE, 0x4548, 0x4548, 0x7DFE, 0x1148, 0x1178, 0x5D00, 0x51FC, 0x5144, 0x5148, 0x5D28, 0xE210, 0x228, 0x4C6}},
    {36402, 2, {0x8, 0x7C1C, 0x45F0, 0x4510, 0x4510, 0x7DFE, 0x1110, 0x117C, 0x5D44, 0x517C, 0x5144, 0x517C, 0x5D44, 0xE244, 0x27C, 0x444}},
    {36403, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x45FC, 0x7C40, 0x13FE, 0x1088, 0x5D04, 0x52FA, 0x5088, 0x5088, 0x5CF8, 0xE088, 0x88, 0xF8}},
    {36404, 2, {0x0, 0x7DFC, 0x4448, 0x4430, 0x45FC, 0x7D24, 0x11FC, 0x1124, 0x5DFC, 0x5124, 0x5040, 0x51FC, 0x5C44, 0xE084, 0x114, 0x208}},
    {36405, 2, {0x8, 0x7C3C, 0x45E0, 0x4420, 0x47FE, 0x7C20, 0x11FC, 0x1124, 0x5DFC, 0x5124, 0x51FC, 0x5020, 0x5DFC, 0xE020, 0x3FE, 0x0}},
    {36406, 2, {0x0, 0x7DFC, 0x4504, 0x4504, 0x45FC, 0x7D04, 0x1104, 0x11FC, 0x5C00, 0x53FE, 0x5020, 0x5120, 0x5D3C, 0xE120, 0x2A0, 0x47E}},
    {36407, 2, {0x20, 0x7C22, 0x45FA, 0x4424, 0x4428, 0x7FFE, 0x1020, 0x1040, 0x10FC, 0x5D84, 0x5284, 0x50FC, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {36408, 2, {0x88, 0x7C88, 0x47FE, 0x4488, 0x44F8, 0x7C88, 0x10F8, 0x1088, 0x5C88, 0x53FE, 0x5100, 0x5148, 0x5D84, 0xE100, 0x1FE, 0x0}},
    {36409, 2, {0x20, 0x7D24, 0x4524, 0x4524, 0x45FC, 0x7C00, 0x13FE, 0x1020, 0x5C40, 0x51FC, 0x5154, 0x5154, 0x5D54, 0xE154, 0x154, 0x10C}},
    {36410, 2, {0x10, 0x7810, 0x4F7C, 0x4914, 0x49FE, 0x7A14, 0x127C, 0x1710, 0x517C, 0x5D10, 0x55FE, 0x5210, 0x5A10, 0xE500, 0x8FE, 0x0}},
    {36411, 2, {0x0, 0x7CFC, 0x4484, 0x4484, 0x44F4, 0x7C94, 0x1094, 0x11FE, 0x5D02, 0x517A, 0x514A, 0x514A, 0x5D7A, 0xE102, 0x10A, 0x104}},
    {36412, 2, {0x0, 0x7CFC, 0x4484, 0x44FC, 0x4484, 0x7CFC, 0x1000, 0x11FE, 0x5C80, 0x50FE, 0x512A, 0x524A, 0x5C92, 0xE122, 0x54, 0x88}},
    {36413, 2, {0x4, 0x7C1E, 0x47E0, 0x4420, 0x45FC, 0x7D24, 0x1124, 0x11FC, 0x5C20, 0x53FE, 0x5222, 0x522A, 0x5EFA, 0xE202, 0x20A, 0x204}},
    {36414, 2, {0x0, 0x7DFE, 0x4400, 0x44FC, 0x4484, 0x7C84, 0x10FC, 0x1000, 0x11FE, 0x5D22, 0x5122, 0x51FE, 0x5122, 0x5D22, 0xE1FE, 0x102}},
    {36415, 2, {0x8, 0x78C8, 0x4B08, 0x4908, 0x492A, 0x792A, 0x13AC, 0x1148, 0x1D08, 0x5388, 0x5554, 0x5114, 0x5114, 0x5D24, 0xE124, 0x142}},
    {36416, 2, {0x48, 0x7D48, 0x4548, 0x47FE, 0x4548, 0x7D48, 0x1178, 0x1100, 0x5DFE, 0x5020, 0x53FE, 0x5070, 0x5CA8, 0xE124, 0x622, 0x20}},
    {36417, 2, {0x40, 0x7C20, 0x45FE, 0x4502, 0x4502, 0x7DFE, 0x1100, 0x1100, 0x5DFE, 0x51AA, 0x51AA, 0x52FE, 0x5EAA, 0xE2AA, 0x4A2, 0x86}},
    {36418, 2, {0x0, 0x7DFC, 0x4448, 0x4430, 0x47FE, 0x7C52, 0x1094, 0x1110, 0x5E30, 0x5020, 0x53FE, 0x5070, 0x5CA8, 0xE124, 0x622, 0x20}},
    {36419, 2, {0x88, 0x7C88, 0x47FE, 0x4488, 0x44A8, 0x7C20, 0x13FE, 0x1040, 0x1080, 0x5CFC, 0x5184, 0x5284, 0x5084, 0x5C84, 0xE0FC, 0x84}},
    {36420, 2, {0x40, 0x7C20, 0x45FC, 0x4400, 0x4488, 0x7C50, 0x13FE, 0x1222, 0x5C24, 0x51FC, 0x5124, 0x5124, 0x5D34, 0xE128, 0x20, 0x20}},
    {36421, 2, {0x20, 0x7C20, 0x47FE, 0x4470, 0x44A8, 0x7D24, 0x1222, 0x11FC, 0x5D04, 0x51FC, 0x5104, 0x51FC, 0x5D04, 0xE000, 0x3FE, 0x0}},
    {36422, 2, {0x0, 0x7A7C, 0x4944, 0x4944, 0x487C, 0x7844, 0x1744, 0x117C, 0x1150, 0x5D4A, 0x5144, 0x5154, 0x5162, 0x5D42, 0xE280, 0x47E}},
    {36423, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xCFE6, 0x100, 0x9F0, 0x1500, 0x23FC}},
    {36424, 2, {0x1E, 0x7DE0, 0x4422, 0x4512, 0x4494, 0x7C80, 0x1020, 0x11CE, 0x1102, 0x5D02, 0x51CE, 0x5102, 0x5102, 0x5D02, 0xE1FE, 0x102}},
    {36425, 2, {0x84, 0x7C44, 0x4448, 0x45FE, 0x4420, 0x7CFC, 0x1020, 0x11FE, 0x5C40, 0x5080, 0x50FE, 0x5110, 0x5E10, 0xE410, 0x1FE, 0x0}},
    {36426, 2, {0x1E, 0x7BE0, 0x4922, 0x4894, 0x4840, 0x7888, 0x11F0, 0x1020, 0x50C4, 0x5DFE, 0x5022, 0x5020, 0x5BFE, 0xE050, 0x88, 0x306}},
    {36427, 2, {0x0, 0x7DFC, 0x4504, 0x45FC, 0x4504, 0x7DFC, 0x1000, 0x1000, 0x13DE, 0x5C42, 0x5252, 0x514A, 0x5252, 0x5C42, 0xE14A, 0x84}},
    {36428, 2, {0x20, 0x7C20, 0x4450, 0x4448, 0x44A4, 0x7DFE, 0x1284, 0x10FC, 0x5C84, 0x50FC, 0x5080, 0x50FC, 0x5D44, 0xE144, 0x27C, 0x44}},
    {36429, 2, {0x0, 0x7BFC, 0x4A04, 0x4A04, 0x4BFC, 0x7A48, 0x1248, 0x12FC, 0x1248, 0x5E48, 0x53FE, 0x52A4, 0x52A8, 0x5C90, 0xE4C8, 0x886}},
    {36430, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x45FC, 0x7D04, 0x11FC, 0x1104, 0x11FC, 0x5D04, 0x51FC, 0x5104, 0x53FE, 0x5C88, 0xE104, 0x202}},
    {36431, 2, {0x4, 0x781E, 0x4BF0, 0x4A1E, 0x4A10, 0x7AFE, 0x1292, 0x1298, 0x52F2, 0x5E8E, 0x5280, 0x52B8, 0x5AA8, 0xE4AA, 0x54A, 0xA86}},
    {36432, 2, {0x20, 0x7D24, 0x44A8, 0x4554, 0x4488, 0x7D04, 0x12FA, 0x1488, 0x1088, 0x5CF8, 0x5088, 0x5088, 0x50F8, 0x5C88, 0xE088, 0x98}},
    {36433, 2, {0x0, 0x7DFE, 0x4484, 0x44FC, 0x4484, 0x7CFC, 0x1086, 0x13FC, 0x5C04, 0x53DE, 0x5052, 0x5152, 0x5C94, 0xE148, 0x254, 0x422}},
    {36434, 2, {0x48, 0x7C48, 0x45FE, 0x4448, 0x4400, 0x7DFE, 0x1048, 0x1048, 0x5DFE, 0x514A, 0x514A, 0x51B6, 0x5D22, 0xE102, 0x10A, 0x104}},
    {36435, 2, {0x40, 0x7DBC, 0x4514, 0x4594, 0x4554, 0x7DA4, 0x114C, 0x1000, 0x11FC, 0x5D24, 0x5124, 0x51FC, 0x5124, 0x5D24, 0xE1FC, 0x104}},
    {36436, 2, {0x804, 0x7F78, 0x840, 0x3E40, 0x2A7E, 0x3E48, 0x2A48, 0xFF48, 0x888, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36437, 2, {0x0, 0x7DFC, 0x4524, 0x45FC, 0x4524, 0x7DFC, 0x10A8, 0x10A8, 0x5DFE, 0x50A8, 0x50A8, 0x51FC, 0x5C20, 0xE3FE, 0x20, 0x20}},
    {36438, 2, {0x20, 0x7C20, 0x47FE, 0x4420, 0x45FC, 0x7C40, 0x13FE, 0x1088, 0x5D04, 0x5242, 0x559C, 0x5104, 0x5DDC, 0xE104, 0x1FC, 0x104}},
    {36439, 2, {0x40, 0x7820, 0x4BFE, 0x4A50, 0x4A50, 0x7BFE, 0x1252, 0x1252, 0x53FE, 0x5E00, 0x5292, 0x52D4, 0x5A98, 0xE492, 0x4D2, 0x88E}},
    {36440, 2, {0x0, 0x7BDE, 0x4842, 0x4A52, 0x494A, 0x7A52, 0x1028, 0x10C4, 0x5312, 0x5C60, 0x5188, 0x5032, 0x59C4, 0xE018, 0x60, 0x380}},
    {36441, 2, {0x24, 0x3FFE, 0x2220, 0x23A4, 0x2224, 0x3FA8, 0x2B10, 0x529A, 0x4626, 0x9FF2, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36442, 2, {0x20, 0x7924, 0x48A8, 0x4BFE, 0x4A02, 0x78F8, 0x1088, 0x1088, 0x1CF8, 0x5020, 0x5020, 0x51FC, 0x5020, 0x5C20, 0xE3FE, 0x0}},
    {36443, 2, {0x2A8, 0xF2A8, 0x97FC, 0x92A8, 0x92AA, 0xF4E6, 0x2800, 0x27FC, 0xA444, 0xB840, 0xA3F8, 0xA248, 0xA248, 0xB248, 0xC258, 0x40}},
    {36444, 2, {0x20, 0x7810, 0x4BFE, 0x4A02, 0x4880, 0x78FE, 0x1110, 0x1320, 0x1D7C, 0x5144, 0x5144, 0x517C, 0x5144, 0x5D44, 0xE17C, 0x144}},
    {36445, 2, {0x108, 0x7908, 0x4A28, 0x4CAE, 0x48A8, 0x7928, 0x137E, 0x1500, 0x1D08, 0x5128, 0x512E, 0x5128, 0x5158, 0x5D48, 0xE186, 0x100}},
    {36446, 2, {0x0, 0x78FE, 0x4810, 0x4A20, 0x4A7C, 0x7A44, 0x13C4, 0x127C, 0x5244, 0x5E7C, 0x52C4, 0x5344, 0x5A7C, 0xE028, 0x44, 0x82}},
    {36447, 2, {0x20, 0x7BFE, 0x4820, 0x49FC, 0x4820, 0x7BFE, 0x1000, 0x11FC, 0x1D04, 0x51FC, 0x5104, 0x51FC, 0x5104, 0x5DFC, 0xE088, 0x104}},
    {36448, 2, {0x40, 0x7820, 0x4BFE, 0x4A00, 0x4A48, 0x7A48, 0x13FE, 0x1248, 0x5248, 0x5E48, 0x5278, 0x5200, 0x5D54, 0xE52A, 0xA2A, 0x1000}},
    {36449, 2, {0x48, 0x7D48, 0x455E, 0x456A, 0x4554, 0x7DC8, 0x1050, 0x1064, 0x13C4, 0x5D7E, 0x5144, 0x5164, 0x5154, 0x5D44, 0xE254, 0x448}},
    {36450, 2, {0x40, 0x7820, 0x4BFE, 0x4888, 0x4850, 0x7BFE, 0x1222, 0x12FA, 0x5222, 0x5EFA, 0x528A, 0x528A, 0x5AFA, 0xE202, 0x20A, 0x204}},
    {36451, 2, {0x88, 0x7C88, 0x47FE, 0x4488, 0x4488, 0x7CF8, 0x1020, 0x13FE, 0x5E22, 0x5332, 0x52AA, 0x5376, 0x5E22, 0xE222, 0x22A, 0x224}},
    {36452, 2, {0x124, 0x7924, 0x4A24, 0x4CA4, 0x4954, 0x794A, 0x1392, 0x1508, 0x1108, 0x5D28, 0x512E, 0x5128, 0x5128, 0x5D58, 0xE14E, 0x180}},
    {36453, 2, {0x10, 0x7C10, 0x46FE, 0x4510, 0x457C, 0x7C54, 0x107C, 0x1354, 0x5D7C, 0x5110, 0x51FE, 0x5110, 0x5D10, 0xE290, 0x47E, 0x0}},
    {36454, 2, {0x20, 0x7A22, 0x4A22, 0x4BFE, 0x4800, 0x7BDE, 0x1252, 0x1252, 0x53DE, 0x5E52, 0x5252, 0x53DE, 0x5A52, 0xE252, 0x55A, 0x8A4}},
    {36455, 2, {0x90, 0x7890, 0x4FFE, 0x4890, 0x4BFC, 0x7A94, 0x13FC, 0x1294, 0x53FC, 0x5C00, 0x51F8, 0x5108, 0x59F8, 0xE108, 0x1F8, 0x108}},
    {36456, 2, {0x110, 0x7914, 0x49D2, 0x4A50, 0x4A7E, 0x7D50, 0x1090, 0x12A8, 0x5128, 0x5D44, 0x5244, 0x5482, 0x5808, 0xE2A4, 0x252, 0x452}},
    {36457, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x1100, 0x2900, 0x47FE}},
    {36458, 2, {0x20, 0x79FC, 0x4924, 0x49FC, 0x4820, 0x7BFE, 0x1000, 0x11FC, 0x1104, 0x5DFC, 0x5104, 0x51FC, 0x5104, 0x5DFC, 0xE088, 0x104}},
    {36459, 2, {0x0, 0x79FC, 0x4848, 0x4830, 0x4BFE, 0x7852, 0x1094, 0x1350, 0x1C20, 0x53FE, 0x5252, 0x528A, 0x5376, 0x5E52, 0xE272, 0x206}},
    {36460, 2, {0x10, 0x7BD2, 0x4854, 0x4948, 0x488A, 0x7904, 0x12FA, 0x1400, 0x51FC, 0x5D04, 0x5104, 0x51FC, 0x5904, 0xE088, 0x7FE, 0x0}},
    {36461, 2, {0x104, 0x7888, 0x4800, 0x4BFE, 0x4A22, 0x7AAA, 0x1272, 0x1222, 0x53FE, 0x5C00, 0x51FC, 0x5104, 0x59FC, 0xE104, 0x1FC, 0x104}},
    {36462, 2, {0x0, 0x7DFC, 0x4450, 0x45FC, 0x4554, 0x7DFC, 0x1020, 0x13FE, 0x5C48, 0x5084, 0x51FA, 0x5288, 0x5CF8, 0xE084, 0x84, 0x7C}},
    {36463, 2, {0x3C, 0x7DE0, 0x4524, 0x44A8, 0x45FE, 0x7CA8, 0x1124, 0x1202, 0x5DFC, 0x5124, 0x5124, 0x51FC, 0x5D24, 0xE124, 0x1FC, 0x104}},
    {36464, 2, {0x0, 0x7BFE, 0x4A00, 0x4A04, 0x4BF4, 0x7A04, 0x12EE, 0x12A4, 0x52A4, 0x5EEC, 0x5204, 0x52A4, 0x5A44, 0xE274, 0x594, 0x8}},
    {36465, 2, {0x20, 0x7DFC, 0x4488, 0x4450, 0x47FE, 0x7C00, 0x11FC, 0x1124, 0x5DFC, 0x5124, 0x51FC, 0x5020, 0x5DFC, 0xE020, 0x3FE, 0x0}},
    {36466, 2, {0x88, 0x7850, 0x4BFE, 0x4850, 0x49FC, 0x7954, 0x118C, 0x1174, 0x5104, 0x5DFC, 0x5008, 0x53FE, 0x5908, 0xE088, 0xA8, 0x10}},
    {36467, 2, {0x24, 0x7BA8, 0x4892, 0x4D14, 0x4A08, 0x7C04, 0x13BA, 0x10A8, 0x1CC6, 0x5380, 0x523C, 0x5384, 0x50A8, 0x5C90, 0xE2A8, 0x144}},
    {36468, 2, {0x214, 0xF112, 0x9FD2, 0x9010, 0x907E, 0xF790, 0x2490, 0x2490, 0xA790, 0xB928, 0xA5A8, 0xA568, 0xA928, 0xB14A, 0xC54A, 0x286}},
    {36469, 2, {0x828, 0x7F24, 0xFE, 0x3E20, 0x2250, 0x3E50, 0x4952, 0x988E, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36470, 2, {0x0, 0xF7FE, 0x9400, 0x9628, 0x9548, 0xF7EE, 0x2492, 0x2484, 0xA6A0, 0xBEA8, 0xA7E8, 0xA488, 0xBC94, 0xE514, 0x924, 0x1242}},
    {36471, 2, {0x3FFE, 0x2890, 0x2510, 0x3FDE, 0x2222, 0x2A88, 0x2F88, 0x2294, 0x2422, 0x2BF8, 0x2208, 0x23F8, 0x2040, 0x427C, 0x4540, 0x88FE}},
    {36472, 2, {0x20, 0x7924, 0x48A8, 0x4BFE, 0x48A8, 0x7924, 0x1222, 0x1104, 0x1104, 0x5DDE, 0x5244, 0x5554, 0x509E, 0x5D04, 0xE204, 0x404}},
    {36473, 2, {0x20, 0x7C50, 0x4488, 0x4574, 0x4602, 0x7CF8, 0x1088, 0x10F8, 0x1000, 0x5DDC, 0x5044, 0x5154, 0x50CC, 0x5D54, 0xE044, 0xCC}},
    {36474, 2, {0x20, 0x7820, 0x49FC, 0x4820, 0x4BFE, 0x7908, 0x139C, 0x1108, 0x1D88, 0x563E, 0x5000, 0x53FE, 0x5090, 0x5C90, 0xE112, 0x20E}},
    {36475, 2, {0x18, 0x7DE0, 0x4440, 0x47FE, 0x4488, 0x7D74, 0x1252, 0x1070, 0x1000, 0x5DFC, 0x5104, 0x5174, 0x5154, 0x5D74, 0xE104, 0x10C}},
    {36476, 2, {0x50, 0x7A52, 0x4954, 0x4850, 0x4BFE, 0x7888, 0x1050, 0x13FE, 0x5020, 0x5DFC, 0x5020, 0x53FE, 0x5850, 0xE088, 0x104, 0x602}},
    {36477, 2, {0x20, 0x7C20, 0x45FC, 0x4420, 0x4554, 0x7C88, 0x1104, 0x12FA, 0x1088, 0x5CF8, 0x5088, 0x50F8, 0x5020, 0x5CA8, 0xE124, 0x60}},
    {36478, 2, {0x108, 0x7888, 0x4BC8, 0x4810, 0x4BDE, 0x7A64, 0x13D4, 0x1014, 0x13D4, 0x5C54, 0x5094, 0x50E8, 0x5388, 0x5C94, 0xE294, 0x122}},
    {36479, 2, {0x40, 0x7820, 0x4BFE, 0x4A8A, 0x4924, 0x7A22, 0x11FC, 0x1124, 0x51FC, 0x5C20, 0x53FE, 0x5222, 0x5BFE, 0xE222, 0x20, 0x20}},
    {36480, 2, {0x0, 0x7DFC, 0x4524, 0x47FE, 0x4524, 0x7DFC, 0x1000, 0x11FC, 0x5D04, 0x51FC, 0x5104, 0x51FC, 0x5D04, 0xE1FC, 0x88, 0x104}},
    {36481, 2, {0xF8, 0x7C88, 0x4488, 0x44F8, 0x4400, 0x7DDC, 0x1154, 0x1154, 0x5DDC, 0x5020, 0x53FE, 0x5070, 0x5CA8, 0xE124, 0x222, 0x20}},
    {36482, 2, {0x10, 0x7A10, 0x497C, 0x4810, 0x48FE, 0x7844, 0x1328, 0x11FE, 0x1110, 0x5D7C, 0x5110, 0x51FE, 0x5110, 0x5D10, 0xE2FE, 0x400}},
    {36483, 2, {0x10, 0x7808, 0x4BC8, 0x4A7E, 0x4A40, 0x7A54, 0x13C8, 0x127E, 0x5208, 0x5FC8, 0x537E, 0x5548, 0x5D48, 0xE5C8, 0x948, 0x8}},
    {36484, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36485, 2, {0x0, 0x7DFC, 0x4554, 0x4554, 0x45FC, 0x7C80, 0x11FC, 0x1244, 0x11F4, 0x5D54, 0x5154, 0x51F4, 0x5044, 0x5DF4, 0xE014, 0x8}},
    {36486, 2, {0x40, 0x787C, 0x4840, 0x4BFE, 0x4A42, 0x7A78, 0x13C4, 0x123C, 0x1E00, 0x53FE, 0x5240, 0x53A4, 0x5258, 0x5DB4, 0xE452, 0x9B0}},
    {36487, 2, {0x88, 0x7888, 0x4BFE, 0x48A8, 0x4822, 0x79FA, 0x1024, 0x1028, 0x53FE, 0x5C40, 0x50FC, 0x5184, 0x5AFC, 0xE484, 0xFC, 0x84}},
    {36488, 2, {0x88, 0x7908, 0x4BC8, 0x4A50, 0x4BDE, 0x7A64, 0x13D4, 0x1114, 0x5094, 0x5FF4, 0x5114, 0x51C8, 0x5948, 0xE254, 0x354, 0x4A2}},
    {36489, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x7FFC, 0x4124, 0x5FFC, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36490, 2, {0x20, 0x7BFE, 0x4820, 0x49FC, 0x4800, 0x7BFE, 0x1002, 0x11FC, 0x1C20, 0x53FE, 0x5000, 0x53FE, 0x5004, 0x5FBE, 0xE2A4, 0x38C}},
    {36491, 2, {0x80, 0xF040, 0x97FC, 0x9110, 0x90A4, 0xFF58, 0x2554, 0x2552, 0xAB58, 0xB800, 0xA208, 0xA3F8, 0xBA08, 0xE3F8, 0x208, 0x408}},
    {36492, 2, {0x80, 0x7900, 0x4BFE, 0x4D54, 0x4954, 0x7BFE, 0x1154, 0x1154, 0x13FE, 0x5D04, 0x51DE, 0x5244, 0x5154, 0x5C9E, 0xE104, 0x204}},
    {36493, 2, {0x0, 0x7BDE, 0x4842, 0x4A52, 0x494A, 0x7A52, 0x1090, 0x1108, 0x13FE, 0x5D10, 0x51FE, 0x5110, 0x51FE, 0x5D10, 0xE1FE, 0x100}},
    {36494, 2, {0x0, 0xF7FC, 0x9248, 0x9444, 0x9FFE, 0xF444, 0x26EC, 0x2554, 0xA6EC, 0xBC44, 0xA6EC, 0xA554, 0xA6EC, 0xB444, 0xC454, 0x408}},
    {36495, 2, {0x48, 0x7DFE, 0x4448, 0x4500, 0x44BE, 0x7C82, 0x1152, 0x117E, 0x11D2, 0x5D7E, 0x5152, 0x517E, 0x5152, 0x5D7E, 0xE142, 0x106}},
    {36496, 2, {0x124, 0x7A48, 0x4924, 0x4800, 0x4BFC, 0x7A94, 0x1264, 0x1294, 0x13FC, 0x5E48, 0x5368, 0x5248, 0x536A, 0x5E4A, 0xE246, 0x362}},
    {36497, 2, {0x440, 0xF28E, 0x9FEA, 0x928A, 0x9FEA, 0xFAAA, 0x2C6C, 0x2BAA, 0xA82A, 0xBFEA, 0xA10A, 0xAFEA, 0xA28C, 0xB448, 0xC828, 0x8}},
    {36498, 2, {0x124, 0x7974, 0x4A58, 0x4ADA, 0x4BFE, 0x7954, 0x12DA, 0x13FE, 0x10A2, 0x5C20, 0x53FE, 0x5070, 0x50A8, 0x5D24, 0xE222, 0x20}},
    {36499, 2, {0x42, 0x7B9C, 0x4A10, 0x4BDE, 0x4A94, 0x7A94, 0x1420, 0x13FC, 0x5204, 0x5FFC, 0x5204, 0x53FC, 0x5A04, 0xE3FC, 0x108, 0x204}},
    {36500, 2, {0x20, 0xF7FE, 0x9400, 0x95FC, 0x9524, 0xF5FC, 0x2524, 0x25FC, 0xA420, 0xBDFC, 0xA420, 0xA7FE, 0xBD24, 0xEAFA, 0x820, 0x13FE}},
    {36501, 2, {0x40, 0x7820, 0x4BFE, 0x4A44, 0x4BF4, 0x7A44, 0x12EE, 0x1204, 0x52F4, 0x5EAC, 0x52E4, 0x5214, 0x5AA4, 0xE474, 0x594, 0x808}},
    {36502, 2, {0x110, 0x7A12, 0x4BAA, 0x493A, 0x4A92, 0x7BBA, 0x108A, 0x13FE, 0x1D12, 0x5212, 0x53AA, 0x513A, 0x5292, 0x5FBA, 0xE08A, 0x3FE}},
    {36503, 2, {0x2200, 0x4F9C, 0x9480, 0x2FDE, 0x6108, 0xAFC8, 0x2508, 0x27D8, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36504, 2, {0x110, 0x7890, 0x4BDE, 0x4810, 0x4A5E, 0x7982, 0x13DE, 0x1010, 0x13DE, 0x5E50, 0x53DE, 0x5250, 0x53DE, 0x5E50, 0xE252, 0x2CE}},
    {36505, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A22, 0x1292, 0x12FA, 0x1FA2, 0x52FA, 0x52A2, 0x52FA, 0x52A2, 0x5EFA, 0xE282, 0x206}},
    {36506, 2, {0x0, 0xF0FE, 0x9428, 0x92FE, 0x92AA, 0xF0FE, 0x2010, 0x21FE, 0xA644, 0xBAFE, 0xA344, 0xA27C, 0xBA42, 0xE23E, 0x500, 0x8FE}},
    {36507, 2, {0x2200, 0x4F9C, 0x9480, 0x2FDE, 0x6208, 0xAFC8, 0x2A48, 0x2AD8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36508, 2, {0x108, 0xF528, 0x97BE, 0x9948, 0x97BE, 0xF318, 0x25AA, 0x2946, 0xA000, 0xBBF8, 0xA208, 0xA248, 0xBA48, 0xE0B0, 0x108, 0x604}},
    {36509, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A22, 0x12FA, 0x1222, 0x52FA, 0x5EAA, 0x52FA, 0x52AA, 0x5AFA, 0xE272, 0x2AA, 0x226}},
    {36510, 2, {0x248, 0xF2E8, 0x9308, 0x96EA, 0x9A1C, 0xF2E8, 0x2208, 0x22E8, 0xA5B4, 0xB8E2, 0xA000, 0xA3F8, 0xB910, 0xE0E0, 0x318, 0xC06}},
    {36511, 2, {0x20, 0x7BFE, 0x4800, 0x49DC, 0x4954, 0x79DC, 0x1088, 0x13FE, 0x1C88, 0x53FE, 0x5088, 0x53FE, 0x5094, 0x5D88, 0xE2A4, 0xC2}},
    {36512, 2, {0x820, 0xFFFE, 0x2820, 0x7DFC, 0x4488, 0x7DFE, 0x4020, 0x7DFC, 0x4420, 0x7FF0, 0x1010, 0x1FF0, 0x100, 0x11F8, 0x3100, 0xCFFE}},
    {36513, 2, {0x3FC, 0xF108, 0x91F8, 0x9108, 0x91F8, 0xF10E, 0x27F8, 0x2008, 0xA7FE, 0xBA94, 0xA39C, 0xA294, 0xA39C, 0xB2D6, 0xC7BC, 0x84}},
    {36514, 2, {0x79E, 0xF492, 0x979E, 0x9492, 0x97FE, 0xF512, 0x25F2, 0x2512, 0xA5F2, 0xBC02, 0xA7BA, 0xA6AA, 0xA59A, 0xB6AA, 0xC59A, 0x406}},
    {36515, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x7A52, 0x13DE, 0x1090, 0x11FE, 0x5D10, 0x53FE, 0x5510, 0x51FE, 0x5D10, 0xE1FE, 0x100}},
    {36516, 2, {0x108, 0xF7FE, 0x9108, 0x9000, 0x9114, 0xF7BE, 0x2114, 0x2794, 0xB17E, 0xAF80, 0xA13E, 0xA3A2, 0xB57E, 0xC922, 0x13E, 0x122}},
    {36517, 2, {0x40, 0xF7FE, 0x9492, 0x9108, 0x9080, 0xF31C, 0x2204, 0x239C, 0xA204, 0xBBFC, 0xA248, 0xA368, 0xA248, 0xB36A, 0xC246, 0x362}},
    {36518, 2, {0x528, 0xF7BE, 0x9948, 0x97BE, 0x9318, 0xF5AA, 0x2946, 0x23FC, 0xA204, 0xBBFC, 0xA204, 0xA3FC, 0xBA04, 0xE3FC, 0x108, 0x204}},
    {36519, 2, {0x3DE, 0x7800, 0x4BDE, 0x4A52, 0x4B5A, 0x7A52, 0x1020, 0x13FE, 0x1E50, 0x53FE, 0x5252, 0x53FE, 0x5292, 0x5EDC, 0xE292, 0x4CE}},
    {36520, 2, {0x110, 0xF7FE, 0x9040, 0x92EE, 0x92AA, 0xFBEA, 0x2AAE, 0x2AE8, 0xABAA, 0xBCE6, 0xA288, 0xA5F4, 0xA290, 0xB060, 0xC198, 0x606}},
    {36521, 2, {0x3DE, 0x7A52, 0x4BDE, 0x4A52, 0x4BDE, 0x78A0, 0x11FE, 0x1320, 0x15FC, 0x5D20, 0x51FE, 0x5000, 0x51FC, 0x5C88, 0xE070, 0x78E}},
    {36522, 2, {0x88, 0x7BFE, 0x4888, 0x4BDE, 0x4A52, 0x7BDE, 0x1252, 0x13DE, 0x124A, 0x5EFE, 0x5392, 0x52FE, 0x5292, 0x5EFE, 0xE282, 0x206}},
    {36523, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF2, 0x1014, 0x1018, 0x7FF0, 0x50, 0x90, 0x310, 0x1C10, 0xE050, 0x20}},
    {36524, 2, {0x800, 0x10FC, 0x3E04, 0x2204, 0x3E04, 0x227C, 0x3E40, 0x2240, 0x2280, 0xFEFC, 0x604, 0xA04, 0x1204, 0x2204, 0x4A28, 0x410}},
    {36525, 2, {0x1020, 0x2020, 0x7C20, 0x45FE, 0x7D22, 0x4624, 0x7C20, 0x4420, 0x4450, 0xFC50, 0xC50, 0x1450, 0x2490, 0x4492, 0x9512, 0xA0E}},
    {36526, 2, {0x1010, 0x2050, 0x7C50, 0x4450, 0x7C88, 0x4488, 0x7D04, 0x46FA, 0x4448, 0xFC48, 0xC48, 0x1448, 0x2488, 0x4488, 0x9528, 0xA10}},
    {36527, 2, {0x1000, 0x21FE, 0x7D00, 0x4504, 0x7D44, 0x4528, 0x7D28, 0x4510, 0x4510, 0xFD28, 0xD28, 0x1544, 0x2584, 0x4500, 0x95FE, 0x800}},
    {36528, 2, {0x1020, 0x2020, 0x7C20, 0x4420, 0x7DFE, 0x4470, 0x7CA8, 0x44A8, 0x4524, 0xFD24, 0xE22, 0x14F8, 0x2420, 0x4420, 0x9420, 0x820}},
    {36529, 2, {0x1000, 0x21FC, 0x7C44, 0x4448, 0x7C5E, 0x4482, 0x7C8A, 0x4524, 0x4420, 0xFDFE, 0xC20, 0x1470, 0x24A8, 0x4524, 0x9622, 0x820}},
    {36530, 2, {0x1000, 0x20F8, 0x7C88, 0x4488, 0x7C88, 0x4488, 0x7D26, 0x4620, 0x45FE, 0xFC20, 0xC70, 0x14A8, 0x2524, 0x4622, 0x9420, 0x820}},
    {36531, 2, {0x800, 0x107C, 0x3E44, 0x2244, 0x3E44, 0x2244, 0x3E7C, 0x2200, 0x2200, 0xFEFE, 0x682, 0xA82, 0x1282, 0x2282, 0x4AFE, 0x482}},
    {36532, 2, {0x820, 0x1010, 0x3EFC, 0x2284, 0x3E84, 0x22FC, 0x3E84, 0x2284, 0x22FC, 0xFEA2, 0x6A4, 0xA98, 0x1290, 0x2288, 0x4AC4, 0x482}},
    {36533, 2, {0x1000, 0x21FC, 0x7C24, 0x44A4, 0x7CA4, 0x4524, 0x7C54, 0x4448, 0x44A0, 0xFC10, 0xC54, 0x1542, 0x2542, 0x454A, 0x9638, 0x800}},
    {36534, 2, {0x1000, 0x21FC, 0x7D24, 0x4524, 0x7DFC, 0x4524, 0x7D24, 0x45FC, 0x4420, 0xFDFE, 0xC70, 0x14A8, 0x2524, 0x4622, 0x9420, 0x820}},
    {36535, 2, {0x101C, 0x21E0, 0x7C20, 0x4420, 0x7FFE, 0x44A8, 0x7D24, 0x4642, 0x4440, 0xFFFE, 0xC88, 0x1508, 0x24D0, 0x4430, 0x9448, 0x984}},
    {36536, 2, {0x1020, 0x2020, 0x7DFC, 0x4450, 0x7C88, 0x4504, 0x7FFE, 0x4408, 0x45E8, 0xFD28, 0xD28, 0x15E8, 0x2528, 0x4408, 0x9428, 0x810}},
    {36537, 2, {0x1080, 0x2080, 0x7CFE, 0x4502, 0x7E22, 0x44AA, 0x7C72, 0x4422, 0x45FE, 0xFC22, 0xC72, 0x14AA, 0x2522, 0x4422, 0x940A, 0x804}},
    {36538, 2, {0x1020, 0x2124, 0x7CA4, 0x44A8, 0x7C20, 0x45FC, 0x7D04, 0x4504, 0x4574, 0xFD54, 0xD54, 0x1554, 0x2574, 0x4504, 0x9514, 0x908}},
    {36539, 2, {0x1040, 0x2020, 0x7DFE, 0x4502, 0x7E04, 0x4450, 0x7C88, 0x4504, 0x4400, 0xFDFC, 0xC20, 0x1420, 0x2420, 0x4420, 0x97FE, 0x800}},
    {36540, 2, {0x1000, 0x20FC, 0x7C80, 0x44F8, 0x7C80, 0x44F8, 0x7C80, 0x47FE, 0x4540, 0xFD24, 0xD28, 0x1510, 0x2508, 0x4544, 0x9582, 0x900}},
    {36541, 2, {0x1000, 0x23FE, 0x7A00, 0x4AFC, 0x7A84, 0x4AFC, 0x7A84, 0x4AFC, 0x4A20, 0xFBFE, 0x1A48, 0x2AC8, 0x4A30, 0x8A48, 0x2A84, 0x13FE}},
    {36542, 2, {0x1084, 0x2048, 0x7DFE, 0x4420, 0x7C20, 0x44FC, 0x7C20, 0x4420, 0x45FE, 0xFC00, 0xC20, 0x15FE, 0x2420, 0x4450, 0x9488, 0xB06}},
    {36543, 2, {0x1020, 0x2010, 0x79FE, 0x4910, 0x797C, 0x4914, 0x79FE, 0x4914, 0x497C, 0xF910, 0x1992, 0x2954, 0x4938, 0x8A54, 0x2A92, 0x1430}},
    {36544, 2, {0x1000, 0x23FE, 0x7A00, 0x4A7C, 0x7A44, 0x4A44, 0x7A7C, 0x4A00, 0x4AEE, 0xFAAA, 0x1AAA, 0x2AAA, 0x4AEE, 0x8A00, 0x2BFE, 0x1000}},
    {36545, 2, {0x1020, 0x21FC, 0x7D24, 0x47FE, 0x7D24, 0x45FC, 0x7C20, 0x45FC, 0x4524, 0xFDFC, 0xC40, 0x17FE, 0x2488, 0x45D0, 0x9470, 0xB8C}},
    {36546, 2, {0x1048, 0x216A, 0x7CDC, 0x4448, 0x7CB4, 0x4522, 0x7DFE, 0x4502, 0x4420, 0xFC20, 0xCFC, 0x1424, 0x2444, 0x4444, 0x9494, 0x908}},
    {36547, 2, {0x1000, 0x21DC, 0x7D54, 0x4554, 0x7DDC, 0x4400, 0x7DFC, 0x4524, 0x45FC, 0xFD24, 0xDFC, 0x1420, 0x27FE, 0x4420, 0x9420, 0x820}},
    {36548, 2, {0x1110, 0x2094, 0x7BD2, 0x4812, 0x7A50, 0x4990, 0x7FFE, 0x4810, 0x4BD2, 0xFA52, 0x1A54, 0x2BD4, 0x4A4A, 0x8A4A, 0x2BD6, 0x1022}},
    {36549, 2, {0x1000, 0x23FE, 0x7A00, 0x4A58, 0x7A54, 0x4ABE, 0x7BA8, 0x4AE8, 0x4ABE, 0xFAA8, 0x1AA8, 0x2ABE, 0x4AA8, 0x8AA8, 0x2ABE, 0x14A0}},
    {36550, 2, {0x1050, 0x21FC, 0x7D54, 0x45FC, 0x7D54, 0x45FC, 0x7C00, 0x45FC, 0x4400, 0xFDFC, 0xD04, 0x15FC, 0x2488, 0x4450, 0x97FE, 0x800}},
    {36551, 2, {0x1020, 0x23FE, 0x7820, 0x49FC, 0x7800, 0x4BFE, 0x7802, 0x49FC, 0x4820, 0xFBFE, 0x1800, 0x2BFE, 0x4804, 0x8BBE, 0x2AA4, 0x138C}},
    {36552, 2, {0x1040, 0x2020, 0x7BFE, 0x4A54, 0x7A7E, 0x4AA8, 0x7BBC, 0x4AA8, 0x4ABC, 0xFAA8, 0x1ABE, 0x2AA0, 0x4A54, 0x8B42, 0x2D4A, 0x1238}},
    {36553, 2, {0x1020, 0x23FE, 0x7A12, 0x49DC, 0x78A8, 0x49DC, 0x78AA, 0x4BFE, 0x4800, 0xF9FC, 0x1904, 0x29FC, 0x4904, 0x89FC, 0x2888, 0x1104}},
    {36554, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {36555, 2, {0x820, 0x820, 0xFFA0, 0x820, 0x7F20, 0x4920, 0x4920, 0x7F20, 0x4920, 0x4920, 0x7F20, 0x822, 0xFFA2, 0x822, 0x81E, 0x800}},
    {36556, 2, {0x1080, 0x1080, 0x1080, 0xFC80, 0x13F0, 0x7C90, 0x5490, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE92, 0x1112, 0x1112, 0x120E, 0x1400}},
    {36557, 2, {0x0, 0x7FFE, 0x4102, 0x8104, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {36558, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {36559, 2, {0x1000, 0x1000, 0x13FC, 0xFC90, 0x1090, 0x7C90, 0x5490, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE92, 0x1112, 0x1112, 0x120E, 0x1400}},
    {36560, 2, {0x1000, 0x11F0, 0x1090, 0xFE90, 0x1090, 0x7C90, 0x5490, 0x7DF0, 0x5490, 0x7C90, 0x1090, 0xFE92, 0x108A, 0x108A, 0x1086, 0x1082}},
    {36561, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x1020, 0x7DFE, 0x5420, 0x7C20, 0x5420, 0x7C50, 0x1050, 0xFE50, 0x1088, 0x1088, 0x1104, 0x1202}},
    {36562, 2, {0x800, 0x87C, 0x810, 0xFF10, 0x810, 0x7F10, 0x4910, 0x7F7E, 0x4910, 0x7F10, 0x810, 0xFF10, 0x810, 0x810, 0x810, 0x810}},
    {36563, 2, {0x1000, 0x1000, 0x10F0, 0xFE90, 0x1090, 0x7C90, 0x5490, 0x7CD0, 0x54B0, 0x7C90, 0x1090, 0xFE92, 0x1092, 0x1112, 0x110E, 0x1200}},
    {36564, 2, {0x1000, 0x1000, 0x11FC, 0xFE24, 0x1024, 0x7CA4, 0x54A4, 0x7CA4, 0x5524, 0x7C24, 0x1024, 0xFE44, 0x1044, 0x1084, 0x1114, 0x1208}},
    {36565, 2, {0x1020, 0x1020, 0x1020, 0xFC20, 0x1124, 0x7D24, 0x5524, 0x7D24, 0x5524, 0x7D24, 0x1124, 0xFF24, 0x1124, 0x11FC, 0x1004, 0x1000}},
    {36566, 2, {0x1000, 0x1000, 0x11FC, 0xFE20, 0x1020, 0x7C20, 0x5420, 0x7C20, 0x55FC, 0x7C20, 0x1020, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {36567, 2, {0x1000, 0x10F8, 0x1088, 0xFE88, 0x1088, 0x7D06, 0x5600, 0x7DFC, 0x5484, 0x7C84, 0x1048, 0xFE50, 0x1020, 0x1050, 0x1088, 0x1306}},
    {36568, 2, {0x1020, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x7D24, 0x5524, 0x7D24, 0x5524, 0x7DFC, 0x1024, 0xFE20, 0x1022, 0x1022, 0x101E, 0x1000}},
    {36569, 2, {0x1020, 0x1020, 0x1020, 0xFE3E, 0x1020, 0x7C20, 0x5420, 0x7DFC, 0x5484, 0x7C88, 0x1048, 0xFE50, 0x1020, 0x1050, 0x1188, 0x1606}},
    {36570, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x1020, 0x7DFE, 0x5420, 0x7C20, 0x5420, 0x7C50, 0x1050, 0xFE50, 0x1088, 0x10C8, 0x1124, 0x1202}},
    {36571, 2, {0x1000, 0x11FE, 0x1100, 0xFD00, 0x1178, 0x7D48, 0x5548, 0x7D48, 0x5548, 0x7D68, 0x1150, 0xFD42, 0x1142, 0x1142, 0x123E, 0x1400}},
    {36572, 2, {0x1020, 0x1020, 0x1020, 0xFDFE, 0x1122, 0x7D22, 0x5522, 0x7D22, 0x5552, 0x7D4A, 0x118A, 0xFF02, 0x1102, 0x1102, 0x110A, 0x1104}},
    {36573, 2, {0x1004, 0x101E, 0x10F0, 0xFE90, 0x1090, 0x7C90, 0x5490, 0x7CFE, 0x5490, 0x7C90, 0x1090, 0xFE88, 0x108A, 0x10AA, 0x10C6, 0x1082}},
    {36574, 2, {0x1008, 0x101C, 0x10E0, 0xFE20, 0x1020, 0x7C3C, 0x54E0, 0x7C20, 0x5420, 0x7C3E, 0x11E0, 0xFE20, 0x1024, 0x1024, 0x1024, 0x101C}},
    {36575, 2, {0x1040, 0x1040, 0x1040, 0xFE7C, 0x1084, 0x7C88, 0x5520, 0x7C20, 0x5420, 0x7C50, 0x1050, 0xFE50, 0x1088, 0x1088, 0x1104, 0x1202}},
    {36576, 2, {0x1008, 0x103C, 0x11E0, 0xFE20, 0x1020, 0x7C20, 0x5420, 0x7DFE, 0x5420, 0x7C20, 0x1020, 0xFE20, 0x1020, 0x1020, 0x11FC, 0x1000}},
    {36577, 2, {0x1020, 0x1020, 0x1050, 0xFE50, 0x1088, 0x7D44, 0x5622, 0x7C20, 0x5400, 0x7DFC, 0x1004, 0xFE08, 0x1008, 0x1010, 0x1010, 0x1020}},
    {36578, 2, {0x1000, 0x1000, 0x10FC, 0xFE00, 0x1000, 0x7C00, 0x55FE, 0x7C20, 0x5420, 0x7C40, 0x1040, 0xFE88, 0x1084, 0x11FE, 0x1082, 0x1000}},
    {36579, 2, {0x100, 0x7FFC, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x600C, 0x1830, 0x6C0, 0x1830, 0x600C}},
    {36580, 2, {0x1008, 0x103C, 0x13E0, 0xFC20, 0x1124, 0x7CA4, 0x54A8, 0x7C20, 0x57FE, 0x7C20, 0x1020, 0xFE20, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {36581, 2, {0x1040, 0x1040, 0x1040, 0xFEFC, 0x1084, 0x7D04, 0x54F4, 0x7C94, 0x5494, 0x7C94, 0x1094, 0xFEF4, 0x1094, 0x1004, 0x1028, 0x1010}},
    {36582, 2, {0x1000, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x7D04, 0x55FC, 0x7C50, 0x5450, 0x7C50, 0x1050, 0xFE92, 0x1092, 0x1112, 0x120E, 0x1400}},
    {36583, 2, {0x1004, 0x101E, 0x10F0, 0xFE90, 0x1090, 0x7C90, 0x5490, 0x7CFE, 0x5490, 0x7C90, 0x1090, 0xFE88, 0x108A, 0x10AA, 0x10D6, 0x108A}},
    {36584, 2, {0x1020, 0x1020, 0x1050, 0xFE50, 0x1088, 0x7D24, 0x5612, 0x7C10, 0x55FC, 0x7C04, 0x1008, 0xFE88, 0x1050, 0x1020, 0x1010, 0x1010}},
    {36585, 2, {0x1020, 0x1020, 0x1020, 0xFE40, 0x1048, 0x7C84, 0x55FE, 0x7C82, 0x5400, 0x7CFC, 0x1084, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {36586, 2, {0x2010, 0x2090, 0x2090, 0xF910, 0x217E, 0xFA52, 0xAB92, 0xF892, 0xA912, 0xF912, 0x2252, 0xFBD2, 0x2062, 0x2022, 0x204A, 0x2084}},
    {36587, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1104, 0x7E12, 0x5420, 0x7C40, 0x5588, 0x7C10, 0x1020, 0xFE44, 0x1188, 0x1010, 0x1060, 0x1380}},
    {36588, 2, {0x400, 0x820, 0x1FF0, 0x210, 0xFFFE, 0x920, 0x3FF8, 0xC106, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {36589, 2, {0x1000, 0x10FE, 0x1080, 0xFE80, 0x10FE, 0x7C90, 0x5490, 0x7CBC, 0x5490, 0x7C90, 0x1090, 0xFEFE, 0x1080, 0x1080, 0x10FE, 0x1000}},
    {36590, 2, {0x1020, 0x1020, 0x1020, 0xFDFC, 0x1124, 0x7D24, 0x5524, 0x7D24, 0x5524, 0x7FFE, 0x1020, 0xFE50, 0x1050, 0x1088, 0x1104, 0x1202}},
    {36591, 2, {0x1000, 0x11FC, 0x1020, 0xFC20, 0x1124, 0x7CA4, 0x54A8, 0x7C20, 0x57FE, 0x7C20, 0x1020, 0xFE20, 0x1020, 0x1020, 0x1020, 0x1020}},
    {36592, 2, {0x444, 0x7C58, 0x462, 0x1C42, 0xE43E, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {36593, 2, {0x1008, 0x101C, 0x11F0, 0xFD50, 0x1150, 0x7D50, 0x5550, 0x7D50, 0x5550, 0x7D48, 0x1148, 0xFD68, 0x1154, 0x1274, 0x1252, 0x1400}},
    {36594, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x11FE, 0x7C20, 0x5420, 0x7C20, 0x54FC, 0x7C84, 0x1084, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {36595, 2, {0x1040, 0x1040, 0x10FC, 0xFE84, 0x1104, 0x7EF4, 0x5494, 0x7C94, 0x54F4, 0x7C84, 0x1094, 0xFE88, 0x1082, 0x1082, 0x107E, 0x1000}},
    {36596, 2, {0x1040, 0x1020, 0x1000, 0xFDFE, 0x1020, 0x7C20, 0x5420, 0x7C20, 0x55FC, 0x7C20, 0x1020, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {36597, 2, {0x1044, 0x1044, 0x1044, 0xFE84, 0x10BE, 0x7D84, 0x5684, 0x7CA4, 0x5494, 0x7C94, 0x1084, 0xFE84, 0x1084, 0x1084, 0x1094, 0x1088}},
    {36598, 2, {0x1040, 0x1020, 0x11FC, 0xFD04, 0x1104, 0x7DFC, 0x5500, 0x7D00, 0x557C, 0x7D08, 0x1110, 0xFD20, 0x1142, 0x1282, 0x127E, 0x1400}},
    {36599, 2, {0x1050, 0x1048, 0x1048, 0xFC40, 0x13FE, 0x7C80, 0x5480, 0x7CFC, 0x5544, 0x7D44, 0x1128, 0xFD28, 0x1210, 0x1228, 0x1444, 0x1182}},
    {36600, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x11FC, 0x7D24, 0x5524, 0x7D24, 0x5524, 0x7DFC, 0x1124, 0xFF24, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {36601, 2, {0x1000, 0x1000, 0x11FC, 0xFF04, 0x1104, 0x7D04, 0x5504, 0x7D04, 0x55FC, 0x7D04, 0x1000, 0xFE90, 0x1088, 0x1104, 0x1202, 0x1402}},
    {36602, 2, {0x1000, 0x11FC, 0x1044, 0xFE44, 0x1044, 0x7C44, 0x5494, 0x7C88, 0x5500, 0x7CFC, 0x1084, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {36603, 2, {0x1000, 0x1000, 0x13FE, 0xFC08, 0x1008, 0x7DE8, 0x5528, 0x7D28, 0x5528, 0x7D28, 0x11E8, 0xFD28, 0x1008, 0x1008, 0x1028, 0x1010}},
    {36604, 2, {0x1020, 0x1020, 0x1120, 0xFD20, 0x11FC, 0x7D20, 0x5620, 0x7C20, 0x55FE, 0x7C20, 0x1050, 0xFE50, 0x1088, 0x1088, 0x1104, 0x1202}},
    {36605, 2, {0x1000, 0x11FC, 0x1084, 0xFE88, 0x1050, 0x7C20, 0x5450, 0x7C88, 0x5726, 0x7C20, 0x11FC, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {36606, 2, {0x1010, 0x1014, 0x1012, 0xFC10, 0x13FE, 0x7C10, 0x5410, 0x7DD0, 0x5490, 0x7C90, 0x1090, 0xFC88, 0x10EA, 0x138A, 0x1006, 0x1002}},
    {36607, 2, {0x1084, 0x1048, 0x1000, 0xFEFC, 0x1048, 0x7C48, 0x5448, 0x7C48, 0x55FE, 0x7C48, 0x1048, 0xFE48, 0x1048, 0x1088, 0x1088, 0x1108}},
    {36608, 2, {0x1000, 0x13FE, 0x1020, 0xFE20, 0x1040, 0x7DFC, 0x5554, 0x7D54, 0x5554, 0x7D54, 0x1154, 0xFF54, 0x1154, 0x1144, 0x1114, 0x1108}},
    {36609, 2, {0x1048, 0x1048, 0x1048, 0xFE48, 0x11FE, 0x7C48, 0x5448, 0x7C48, 0x5448, 0x7DFE, 0x1000, 0xFE48, 0x1044, 0x1084, 0x1082, 0x1102}},
    {36610, 2, {0x820, 0x7FFC, 0x820, 0xFFFE, 0x1010, 0x2108, 0x5FF4, 0x8102, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {36611, 2, {0x1040, 0x1020, 0x1020, 0xFDFE, 0x1000, 0x7C88, 0x5504, 0x7E02, 0x5488, 0x7C88, 0x1050, 0xFE50, 0x1020, 0x1050, 0x1088, 0x1306}},
    {36612, 2, {0x1020, 0x1020, 0x1124, 0xFEA4, 0x10A8, 0x7C20, 0x57FE, 0x7C90, 0x5490, 0x7C90, 0x1090, 0xFE92, 0x1112, 0x1112, 0x120E, 0x1400}},
    {36613, 2, {0x1040, 0x1040, 0x1078, 0xFE88, 0x1150, 0x7C20, 0x5450, 0x7C88, 0x5506, 0x7CF8, 0x1088, 0xFE88, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {36614, 2, {0x1020, 0x1010, 0x1010, 0xFDFE, 0x1020, 0x7C24, 0x5444, 0x7CF8, 0x5412, 0x7C22, 0x1044, 0xFD88, 0x1010, 0x1028, 0x1044, 0x1182}},
    {36615, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1104, 0x7E02, 0x55FC, 0x7C20, 0x5420, 0x7C20, 0x11FC, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {36616, 2, {0x1010, 0x1020, 0x10FC, 0xFE84, 0x10A4, 0x7C94, 0x5494, 0x7DFE, 0x5484, 0x7CA4, 0x1094, 0xFE94, 0x1084, 0x1104, 0x1114, 0x1208}},
    {36617, 2, {0x820, 0x7F28, 0x824, 0xFFFE, 0x820, 0xFFA0, 0x824, 0x7F24, 0x4924, 0x7F28, 0x4928, 0x7F10, 0x812, 0xFFAA, 0x846, 0x882}},
    {36618, 2, {0x1000, 0x11FE, 0x1020, 0xFE40, 0x1088, 0x7D04, 0x55FE, 0x7C22, 0x5420, 0x7C20, 0x11FE, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {36619, 2, {0x100, 0x4104, 0x7FFC, 0x200, 0xFFFE, 0x920, 0x3FF8, 0xC106, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {36620, 2, {0x2000, 0x23FE, 0x2020, 0xF820, 0x2020, 0xFBFE, 0xAA22, 0xFA22, 0xAAAA, 0xFAAA, 0x22AA, 0xFAFA, 0x2202, 0x2202, 0x220A, 0x2204}},
    {36621, 2, {0x1000, 0x1050, 0x1048, 0xFE84, 0x1124, 0x7C20, 0x5450, 0x7C88, 0x5506, 0x7EF8, 0x1088, 0xFE88, 0x1088, 0x1088, 0x10F8, 0x1088}},
    {36622, 2, {0x1010, 0x1092, 0x1052, 0xFE54, 0x1010, 0x7CFE, 0x5482, 0x7C82, 0x54FE, 0x7C82, 0x1082, 0xFEFE, 0x1082, 0x1082, 0x108A, 0x1084}},
    {36623, 2, {0x1000, 0x13FE, 0x1050, 0xFC50, 0x11FC, 0x7D54, 0x5554, 0x7D54, 0x558C, 0x7D04, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {36624, 2, {0x1020, 0x1010, 0x11FE, 0xFD02, 0x1204, 0x7CF8, 0x5400, 0x7C00, 0x55FE, 0x7C50, 0x1050, 0xFE50, 0x1092, 0x1092, 0x110E, 0x1200}},
    {36625, 2, {0x1000, 0x10FC, 0x1024, 0xFE24, 0x11FE, 0x7C24, 0x5424, 0x7CFC, 0x5440, 0x7C40, 0x10FC, 0xFEC4, 0x1144, 0x1044, 0x107C, 0x1044}},
    {36626, 2, {0x1000, 0x13FE, 0x1120, 0xFD28, 0x1128, 0x7DE8, 0x5528, 0x7D28, 0x5528, 0x7DE8, 0x1128, 0xFD2A, 0x113A, 0x13EA, 0x1026, 0x1020}},
    {36627, 2, {0x1080, 0x1080, 0x10F8, 0xFD08, 0x1210, 0x7DFC, 0x5524, 0x7D24, 0x5524, 0x7DFC, 0x1050, 0xFE50, 0x1090, 0x1092, 0x1112, 0x120E}},
    {36628, 2, {0x1028, 0x1024, 0x13FE, 0xFC20, 0x1020, 0x7DFC, 0x5524, 0x7D24, 0x55FC, 0x7D24, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x1124, 0x110C}},
    {36629, 2, {0x1000, 0x11FE, 0x1000, 0xFE92, 0x1124, 0x7E48, 0x5524, 0x7C92, 0x5400, 0x7DFE, 0x1020, 0xFE20, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {36630, 2, {0x1000, 0x10FE, 0x1092, 0xFE92, 0x10BA, 0x7C92, 0x54FE, 0x7C82, 0x54BA, 0x7CAA, 0x10AA, 0xFEBA, 0x1082, 0x1102, 0x110A, 0x1204}},
    {36631, 2, {0x1000, 0x1040, 0x119C, 0xFF04, 0x1104, 0x7DDC, 0x5504, 0x7D04, 0x55FC, 0x7C50, 0x1050, 0xFE50, 0x1090, 0x1092, 0x1112, 0x120E}},
    {36632, 2, {0x1020, 0x1020, 0x10FC, 0xFC20, 0x1020, 0x7DFE, 0x5448, 0x7C84, 0x5522, 0x7C78, 0x1088, 0xFD48, 0x1050, 0x1020, 0x1050, 0x118C}},
    {36633, 2, {0x1000, 0x13E0, 0x115C, 0xFD54, 0x1154, 0x7DD4, 0x5554, 0x7D54, 0x55D4, 0x7D54, 0x1148, 0xFD68, 0x13C8, 0x1054, 0x1054, 0x1062}},
    {36634, 2, {0x1048, 0x1044, 0x105E, 0xFDE0, 0x1028, 0x7C12, 0x546A, 0x7D96, 0x5448, 0x7C5E, 0x11E0, 0xFE24, 0x1028, 0x1012, 0x106A, 0x1186}},
    {36635, 2, {0x2000, 0x23FE, 0x2020, 0xF820, 0x23FE, 0xFA22, 0xAA22, 0xFB32, 0xAAAA, 0xFAAA, 0x2376, 0xFA66, 0x2222, 0x2222, 0x222A, 0x2204}},
    {36636, 2, {0x1092, 0x1092, 0x1124, 0xFE48, 0x1124, 0x7C92, 0x5492, 0x7C00, 0x55FE, 0x7D22, 0x1122, 0xFDFE, 0x1122, 0x1122, 0x11FE, 0x1102}},
    {36637, 2, {0x1000, 0x13FE, 0x9222, 0x5420, 0x5BFE, 0x1020, 0xFDFC, 0x2924, 0x29FC, 0x2924, 0x29FC, 0x2820, 0x2BFE, 0x4C20, 0x4820, 0x8020}},
    {36638, 2, {0x2000, 0x23FE, 0x2202, 0xFA8A, 0x2252, 0xFBFE, 0xAA42, 0xFA22, 0xABFE, 0xFA82, 0x2282, 0xFA82, 0x22FA, 0x2202, 0x220A, 0x2204}},
    {36639, 2, {0x2000, 0x23DE, 0x2042, 0xF94A, 0x2084, 0xF94A, 0xAA52, 0xFC20, 0xA800, 0xFBDE, 0x2052, 0xF952, 0x2094, 0x2148, 0x2254, 0x2422}},
    {36640, 2, {0x1000, 0x11FC, 0x1124, 0xFD24, 0x11FC, 0x7D24, 0x5524, 0x7DFC, 0x5420, 0x7DFE, 0x1070, 0xFEA8, 0x1124, 0x1222, 0x1020, 0x1020}},
    {36641, 2, {0x1040, 0x1040, 0x107C, 0xFC84, 0x1108, 0x7E00, 0x5420, 0x7DCE, 0x5502, 0x7D02, 0x11CE, 0xFD02, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {36642, 2, {0x1020, 0x1020, 0x11FC, 0xFC50, 0x1088, 0x7D04, 0x57FE, 0x7C08, 0x55E8, 0x7D28, 0x1128, 0xFDE8, 0x1128, 0x1008, 0x1028, 0x1010}},
    {36643, 2, {0x1000, 0x11DC, 0x1154, 0xFD54, 0x1154, 0x7DDC, 0x5554, 0x7D54, 0x5554, 0x7DDC, 0x1154, 0xFD54, 0x1154, 0x12D4, 0x1224, 0x144C}},
    {36644, 2, {0x1020, 0x1020, 0x11FE, 0xFE20, 0x10FC, 0x7C20, 0x55FE, 0x7C00, 0x54FC, 0x7C84, 0x10FC, 0xFE84, 0x10FC, 0x1084, 0x1094, 0x1088}},
    {36645, 2, {0x1000, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x7D04, 0x5504, 0x7DFC, 0x5400, 0x7D12, 0x11D4, 0xFD18, 0x1110, 0x1152, 0x1192, 0x110E}},
    {36646, 2, {0x820, 0x3EF8, 0x820, 0x7EFC, 0x1450, 0x2388, 0x5FF4, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0xFFFE, 0x100}},
    {36647, 2, {0x1044, 0x1044, 0x1088, 0xFDDC, 0x1088, 0x7C88, 0x5488, 0x7C88, 0x55DE, 0x7C88, 0x1088, 0xFE88, 0x1088, 0x1088, 0x1088, 0x1108}},
    {36648, 2, {0x1020, 0x1010, 0x11FE, 0xFD02, 0x1000, 0x7CF8, 0x5488, 0x7C88, 0x54F8, 0x7C80, 0x1080, 0xFEFC, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {36649, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x540, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100, 0x100}},
    {36650, 2, {0x1020, 0x1020, 0x1050, 0xFE88, 0x1104, 0x7EFA, 0x5400, 0x7DFC, 0x5554, 0x7D54, 0x11FC, 0xFF54, 0x1154, 0x1154, 0x1104, 0x110C}},
    {36651, 2, {0x1048, 0x1048, 0x1048, 0xFDCE, 0x1048, 0x7C48, 0x5448, 0x7DCE, 0x5448, 0x7C48, 0x1048, 0xFDCE, 0x1048, 0x1048, 0x1048, 0x1048}},
    {36652, 2, {0x1040, 0x1020, 0x13FE, 0xFC00, 0x1000, 0x7DFC, 0x5504, 0x7D04, 0x5504, 0x7DFC, 0x1020, 0xFEA8, 0x1124, 0x1222, 0x10A0, 0x1040}},
    {36653, 2, {0x1000, 0x13FE, 0x1020, 0xFC40, 0x11FC, 0x7D54, 0x5554, 0x7D54, 0x5554, 0x7D2C, 0x1020, 0xFDFE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {36654, 2, {0x1000, 0x10FC, 0x1048, 0xFE30, 0x11FE, 0x7C52, 0x5494, 0x7D10, 0x5630, 0x7C20, 0x11FE, 0xFE70, 0x10A8, 0x1124, 0x1222, 0x1020}},
    {36655, 2, {0x1000, 0x10F8, 0x1088, 0xFE88, 0x10F8, 0x7C00, 0x55FE, 0x7C88, 0x54F8, 0x7C88, 0x10F8, 0xFE88, 0x109E, 0x13E8, 0x1008, 0x1008}},
    {36656, 2, {0x1000, 0x10FC, 0x1084, 0xFEFC, 0x1084, 0x7CFC, 0x5400, 0x7DFE, 0x5480, 0x7CFE, 0x112A, 0xFE4A, 0x1092, 0x1122, 0x1054, 0x1088}},
    {36657, 2, {0x2008, 0x200C, 0x200A, 0xFBFE, 0x2208, 0xFA08, 0xAAE8, 0xFA0A, 0xAA0A, 0xFAEC, 0x22AC, 0xFAA8, 0x22EA, 0x221A, 0x2226, 0x2442}},
    {36658, 2, {0x1020, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x7C00, 0x57FE, 0x7C20, 0x5440, 0x7DFC, 0x1154, 0xFF54, 0x1154, 0x1154, 0x1154, 0x110C}},
    {36659, 2, {0x1020, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x7C40, 0x57FE, 0x7C88, 0x5504, 0x7EFA, 0x1020, 0xFDFC, 0x1020, 0x1050, 0x1088, 0x1304}},
    {36660, 2, {0x1008, 0x101C, 0x11F0, 0xFD10, 0x1110, 0x7DFE, 0x5510, 0x7D7C, 0x5544, 0x7D7C, 0x1144, 0xFD7C, 0x1144, 0x1244, 0x127C, 0x1444}},
    {36661, 2, {0x1000, 0x10FC, 0x1084, 0xFEFC, 0x1084, 0x7CFC, 0x5440, 0x7CFE, 0x5512, 0x7C92, 0x10AA, 0xFE82, 0x10FA, 0x1002, 0x1014, 0x1008}},
    {36662, 2, {0x1088, 0x1048, 0x1050, 0xFDFE, 0x1050, 0x7C50, 0x55FC, 0x7D54, 0x5554, 0x7D8C, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {36663, 2, {0x1080, 0x1080, 0x10FE, 0xFD02, 0x1242, 0x7DFA, 0x5402, 0x7DF2, 0x5402, 0x7DF2, 0x1002, 0xFDF2, 0x1112, 0x11F2, 0x1014, 0x1008}},
    {36664, 2, {0x2040, 0x20A0, 0x2110, 0xFA08, 0x25F6, 0xF800, 0xABC4, 0xFA54, 0xAA54, 0xFBD4, 0x2254, 0xFA54, 0x23D4, 0x2244, 0x2254, 0x22C8}},
    {36665, 2, {0x1040, 0x1040, 0x10FE, 0xFE80, 0x11FC, 0x7E84, 0x54FC, 0x7C84, 0x54FC, 0x7C40, 0x107C, 0xFEC4, 0x1128, 0x1010, 0x1068, 0x1186}},
    {36666, 2, {0x1052, 0x1052, 0x10A4, 0xFDFE, 0x10A4, 0x7C52, 0x5452, 0x7C00, 0x54FE, 0x7C92, 0x1092, 0xFEFE, 0x1092, 0x1092, 0x10FE, 0x1082}},
    {36667, 2, {0x1000, 0x10FE, 0x1000, 0xFE7C, 0x1044, 0x7C44, 0x547C, 0x7C00, 0x54FE, 0x7C92, 0x1092, 0xFEFE, 0x1092, 0x1092, 0x10FE, 0x1082}},
    {36668, 2, {0x1000, 0x10F8, 0x1088, 0xFE88, 0x10F8, 0x7C88, 0x5488, 0x7CF8, 0x5400, 0x7DFC, 0x1154, 0xFF54, 0x1154, 0x1154, 0x13FE, 0x1000}},
    {36669, 2, {0x1040, 0x1048, 0x1084, 0xFDFE, 0x1020, 0x7FFE, 0x5488, 0x7D04, 0x57FE, 0x7D24, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {36670, 2, {0x2000, 0x23FC, 0x2204, 0xFA04, 0x23FC, 0xFA48, 0xAA48, 0xFAFC, 0xAA48, 0xFA48, 0x23FE, 0xFAA4, 0x22A8, 0x2490, 0x24C8, 0x2886}},
    {36671, 2, {0x100, 0x17C0, 0x611C, 0x47C4, 0x4544, 0x77DC, 0x4544, 0x47C4, 0x711C, 0x47C4, 0x4104, 0xFFFE, 0x0, 0x1010, 0x2008, 0x4004}},
    {36672, 2, {0x1000, 0x10F8, 0x10A8, 0xFEA8, 0x10D8, 0x7C88, 0x54F8, 0x7C00, 0x5400, 0x7DFC, 0x1154, 0xFF54, 0x1154, 0x1154, 0x13FE, 0x1000}},
    {36673, 2, {0x101E, 0x11E0, 0x1022, 0xFD12, 0x1094, 0x7C80, 0x5420, 0x7DCE, 0x5502, 0x7D02, 0x11CE, 0xFF02, 0x1102, 0x1102, 0x11FE, 0x1102}},
    {36674, 2, {0x800, 0xFFB8, 0x828, 0x7F28, 0x28, 0xFFC6, 0x8880, 0x7F7C, 0x824, 0x7F24, 0x4928, 0x7F28, 0x4910, 0xFFA8, 0x844, 0x882}},
    {36675, 2, {0x1020, 0x1020, 0x13FE, 0xFC20, 0x11FC, 0x7C40, 0x57FE, 0x7C88, 0x5534, 0x7EE2, 0x1020, 0xFDFC, 0x1070, 0x10AC, 0x1322, 0x1020}},
    {36676, 2, {0x2040, 0x2020, 0x23FE, 0xFA02, 0x2424, 0xF9FC, 0xA820, 0xF9FC, 0xA820, 0xFBFE, 0x2020, 0xF9FC, 0x2104, 0x2104, 0x21FC, 0x2104}},
    {36677, 2, {0x1020, 0x1020, 0x11FC, 0xFC20, 0x13FE, 0x7C00, 0x55FC, 0x7D04, 0x5504, 0x7DFC, 0x1052, 0xFC94, 0x1188, 0x16A4, 0x10C2, 0x1080}},
    {36678, 2, {0x2040, 0x2020, 0x23FE, 0xFA50, 0x2250, 0xFBFE, 0xAA52, 0xFA52, 0xABFE, 0xFA00, 0x2292, 0xFAD4, 0x2298, 0x2492, 0x24D2, 0x288E}},
    {36679, 2, {0x2000, 0x23DE, 0x2042, 0xFA52, 0x214A, 0xFA52, 0xA828, 0xF8C4, 0xAB12, 0xF860, 0x2188, 0xF832, 0x21C4, 0x2018, 0x2060, 0x2380}},
    {36680, 2, {0x1124, 0x1248, 0x1124, 0xFC00, 0x11FC, 0x7D24, 0x55FC, 0x7D24, 0x55FC, 0x7C20, 0x13FE, 0xFC70, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {36681, 2, {0x1020, 0x13FE, 0x1020, 0xFDFC, 0x1124, 0x7DFC, 0x5524, 0x7DFC, 0x5422, 0x7DFE, 0x1008, 0xFDFE, 0x1108, 0x1088, 0x1028, 0x1010}},
    {36682, 2, {0x1088, 0x13DE, 0x1088, 0xFDDC, 0x1088, 0x7FDE, 0x5488, 0x7C00, 0x57FC, 0x7C04, 0x1004, 0xFDFC, 0x1004, 0x1004, 0x13FC, 0x1004}},
    {36683, 2, {0x1010, 0x1010, 0x12FE, 0xFD10, 0x117C, 0x7C54, 0x547C, 0x7F54, 0x557C, 0x7D10, 0x11FE, 0xFD10, 0x1110, 0x1290, 0x147E, 0x1000}},
    {36684, 2, {0x1000, 0x11FC, 0x1020, 0xFFFE, 0x1222, 0x7DAC, 0x5420, 0x7DAC, 0x5400, 0x7DFC, 0x1004, 0xFDFC, 0x1004, 0x1004, 0x11FC, 0x1004}},
    {36685, 2, {0x2108, 0x2088, 0x23E8, 0xF910, 0x229E, 0xFFD4, 0xA864, 0xFBD4, 0xAA54, 0xFBD4, 0x2254, 0xFBC8, 0x2248, 0x2254, 0x2264, 0x22C2}},
    {36686, 2, {0x201C, 0x21E0, 0x2040, 0xFBFE, 0x2088, 0xF904, 0xAAFA, 0xFC88, 0xA8F8, 0xF800, 0x23FE, 0xFA02, 0x22FA, 0x228A, 0x22FA, 0x2206}},
    {36687, 2, {0x23FE, 0x2202, 0x23FE, 0xFA00, 0x22FC, 0xFA08, 0xABFE, 0xFA10, 0xAA30, 0xFBCE, 0x2242, 0xFA84, 0x24EE, 0x2784, 0x2884, 0x218C}},
    {36688, 2, {0x1050, 0x1152, 0x10D4, 0xFC50, 0x11FE, 0x7C88, 0x5450, 0x7DFE, 0x5420, 0x7CFC, 0x1020, 0xFDFE, 0x1050, 0x1088, 0x1104, 0x1202}},
    {36689, 2, {0x1020, 0x1020, 0x11FC, 0xFC20, 0x1154, 0x7C88, 0x5504, 0x7EFA, 0x5488, 0x7CF8, 0x1088, 0xFEF8, 0x1020, 0x10A8, 0x1124, 0x1060}},
    {36690, 2, {0x1020, 0x11FC, 0x1020, 0xFC88, 0x13FE, 0x7C88, 0x55FC, 0x7D04, 0x55FC, 0x7D04, 0x11FC, 0xFD04, 0x11FC, 0x1088, 0x1104, 0x1202}},
    {36691, 2, {0x103C, 0x13E0, 0x1124, 0xFCA8, 0x13FE, 0x7CA8, 0x5524, 0x7E02, 0x55FC, 0x7D24, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x11FC, 0x1104}},
    {36692, 2, {0x2020, 0x2124, 0x20A8, 0xFBFE, 0x20A8, 0xF924, 0xAA22, 0xF904, 0xA904, 0xF9DE, 0x2244, 0xFD54, 0x209E, 0x2104, 0x2204, 0x2404}},
    {36693, 2, {0x1088, 0x13FE, 0x1088, 0xFDFC, 0x1104, 0x7DFC, 0x5504, 0x7DFC, 0x5480, 0x7DFE, 0x1222, 0xFD52, 0x1102, 0x11FA, 0x100A, 0x1004}},
    {36694, 2, {0x2020, 0x2020, 0x23FE, 0xF924, 0x2124, 0xFAAA, 0xAFFE, 0xF800, 0xA9FC, 0xF904, 0x2174, 0xF954, 0x2174, 0x2104, 0x21FC, 0x2104}},
    {36695, 2, {0x2014, 0x2012, 0x23FE, 0xFA10, 0x22F0, 0xFA14, 0xAAF4, 0xFAB8, 0xAAEA, 0xFA16, 0x2442, 0xF824, 0x22A2, 0x228A, 0x2478, 0x2000}},
    {36696, 2, {0x1000, 0x11FC, 0x1154, 0xFD54, 0x11FC, 0x7C00, 0x57FE, 0x7C00, 0x55FC, 0x7D04, 0x11FC, 0xFC62, 0x1094, 0x1188, 0x12A4, 0x10C2}},
    {36697, 2, {0x1104, 0x1088, 0x13FE, 0xFC20, 0x11FC, 0x7C20, 0x57FE, 0x7C54, 0x5592, 0x7C90, 0x13FE, 0xFC90, 0x10D4, 0x138A, 0x1096, 0x11A2}},
    {36698, 2, {0x800, 0x7F3C, 0x2A24, 0x3E42, 0x2A3C, 0xFFA4, 0x4918, 0x7F66, 0x100, 0x7FFC, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100, 0x100}},
    {36699, 2, {0x2144, 0x2144, 0x2554, 0xFB64, 0x215E, 0xFBE4, 0xA804, 0xFA34, 0xA94C, 0xFBE4, 0x2084, 0xFBE4, 0x2084, 0x20E4, 0x2714, 0x2008}},
    {36700, 2, {0x2000, 0x21FC, 0x2020, 0xFBFE, 0x2222, 0xF9AC, 0xA820, 0xF9AC, 0xA800, 0xFBFE, 0x2020, 0xF9FC, 0x2154, 0x2154, 0x2154, 0x210C}},
    {36701, 2, {0x1A38, 0x2388, 0x3A38, 0x2388, 0x38B8, 0x2288, 0xFFFE, 0x1110, 0x3FF8, 0xC106, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {36702, 2, {0x2008, 0x23E8, 0x2288, 0xFBEE, 0x2228, 0xFBF4, 0xAA82, 0xFBE2, 0xA800, 0xF9FC, 0x2154, 0xF954, 0x2154, 0x2154, 0x27FE, 0x2000}},
    {36703, 2, {0x100, 0x3FF8, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100, 0x1110, 0xFEFE, 0x5454, 0x7C7C, 0x5454, 0xFEFE, 0x1010}},
    {36704, 2, {0x2000, 0x21FC, 0x2124, 0xF9FC, 0x2124, 0xF9FC, 0xA904, 0xF800, 0xABFE, 0xFAAA, 0x22AA, 0xFBFE, 0x22AA, 0x22AA, 0x23FE, 0x2222}},
    {36705, 2, {0x2104, 0x27C4, 0x4108, 0x97D2, 0xE55C, 0x47C8, 0x9552, 0xF7DE, 0x100, 0xAFEA, 0xA92A, 0x100, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {36706, 2, {0x2124, 0x2174, 0x2258, 0xFADA, 0x23FE, 0xF954, 0xAADA, 0xFBFE, 0xA8A2, 0xF820, 0x23FE, 0xF870, 0x20A8, 0x2124, 0x2222, 0x2020}},
    {36707, 2, {0x2000, 0x23FE, 0x2222, 0xFACC, 0x2244, 0xFBEE, 0xAA44, 0xFAEE, 0xAB54, 0xFA44, 0x2210, 0xFA90, 0x229E, 0x2490, 0x2490, 0x2BFE}},
    {36708, 2, {0x2040, 0x207C, 0x2040, 0xFBFC, 0x2244, 0xFBF0, 0xAA44, 0xFAFC, 0xAAA8, 0xFAF8, 0x22A8, 0xFAF8, 0x2200, 0x25FC, 0x2554, 0x2BFE}},
    {36709, 2, {0x2088, 0x23FE, 0x2088, 0xFBDE, 0x2252, 0xFBDE, 0xAA52, 0xFBDE, 0xAA4A, 0xFAFE, 0x2392, 0xFAFE, 0x2292, 0x22FE, 0x2282, 0x2206}},
    {36710, 2, {0x200, 0x200, 0x200, 0x7FFC, 0x400, 0x900, 0x1100, 0x2100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {36711, 2, {0x1040, 0x1040, 0x1040, 0xFE40, 0x2040, 0x2840, 0x4840, 0x7E40, 0x840, 0x840, 0xE40, 0xF842, 0x4842, 0x842, 0x83E, 0x800}},
    {36712, 2, {0x2080, 0x2080, 0x2080, 0xFC80, 0x43F0, 0x5090, 0x9090, 0xFC90, 0x1090, 0x1090, 0x1C90, 0xF092, 0x5112, 0x1112, 0x120E, 0x1400}},
    {36713, 2, {0x1000, 0x1000, 0x10FE, 0xFE10, 0x2010, 0x2810, 0x4810, 0x7E10, 0x9FE, 0x810, 0xE10, 0xF810, 0x4810, 0x810, 0x810, 0x810}},
    {36714, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x2020, 0x29FE, 0x4820, 0x7E20, 0x820, 0x850, 0xE50, 0xF850, 0x4888, 0x888, 0x904, 0xA02}},
    {36715, 2, {0x1000, 0x1000, 0x10FC, 0xFE24, 0x2024, 0x28A4, 0x48A4, 0x7EA4, 0x8A4, 0x924, 0xE24, 0xF844, 0x4844, 0x884, 0x928, 0xA10}},
    {36716, 2, {0x2020, 0x2020, 0x2020, 0xFDFC, 0x4020, 0x5040, 0x93FE, 0xFC40, 0x1080, 0x11FC, 0x1C04, 0xF088, 0x5050, 0x1020, 0x1010, 0x1010}},
    {36717, 2, {0x2000, 0x21FE, 0x2100, 0xFD00, 0x4178, 0x5148, 0x9148, 0xFD48, 0x1148, 0x1168, 0x1D50, 0xF142, 0x5142, 0x1142, 0x123E, 0x1400}},
    {36718, 2, {0x2040, 0x2040, 0x20A0, 0xFCA0, 0x4110, 0x5208, 0x9406, 0xFD10, 0x1120, 0x1140, 0x1D80, 0xF100, 0x5104, 0x1104, 0x10FC, 0x1000}},
    {36719, 2, {0x1040, 0x1040, 0x1040, 0xFE7C, 0x2084, 0x2888, 0x4920, 0x7E20, 0x820, 0x850, 0xE50, 0xF850, 0x4888, 0x888, 0x904, 0xA02}},
    {36720, 2, {0x200, 0x200, 0x7FFC, 0x800, 0x1100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x3EF8, 0x288, 0x1450, 0x820, 0x3458, 0xC286}},
    {36721, 2, {0x1020, 0x1020, 0x1020, 0xFE20, 0x21FE, 0x2820, 0x4820, 0x7E20, 0x8FC, 0x884, 0xE84, 0xF884, 0x4884, 0x884, 0x8FC, 0x884}},
    {36722, 2, {0x2000, 0x2000, 0x21FE, 0xFC08, 0x4008, 0x51E8, 0x9128, 0xFD28, 0x1128, 0x1128, 0x1DE8, 0xF128, 0x5008, 0x1008, 0x1028, 0x1010}},
    {36723, 2, {0x2020, 0x2020, 0x203E, 0xFC20, 0x4020, 0x51FC, 0x9104, 0xFD04, 0x1104, 0x11FC, 0x1D04, 0xF100, 0x5100, 0x1200, 0x1200, 0x1400}},
    {36724, 2, {0x2020, 0x2020, 0x2020, 0xFC20, 0x41FC, 0x5124, 0x9124, 0xFD24, 0x1124, 0x11FC, 0x1D24, 0xF124, 0x5124, 0x1124, 0x11FC, 0x1104}},
    {36725, 2, {0x2000, 0x2000, 0x21FC, 0xFD04, 0x4104, 0x5104, 0x9104, 0xFD04, 0x11FC, 0x1104, 0x1C00, 0xF090, 0x5088, 0x1104, 0x1202, 0x1402}},
    {36726, 2, {0x2020, 0x2020, 0x2120, 0xFD20, 0x41FC, 0x5120, 0x9220, 0xFC20, 0x13FE, 0x1020, 0x1C50, 0xF050, 0x5088, 0x1088, 0x1104, 0x1202}},
    {36727, 2, {0x2008, 0x203C, 0x23E0, 0xFC20, 0x4124, 0x50A4, 0x90A8, 0xFC20, 0x13FE, 0x1020, 0x1C20, 0xF020, 0x5020, 0x1020, 0x10A0, 0x1040}},
    {36728, 2, {0x2020, 0x2020, 0x2050, 0xFC88, 0x4104, 0x5212, 0x9020, 0xFC40, 0x1188, 0x1010, 0x1C20, 0xF044, 0x5188, 0x1010, 0x1060, 0x1380}},
    {36729, 2, {0x2008, 0x201C, 0x21E0, 0xFD00, 0x4120, 0x5120, 0x9120, 0xFDFE, 0x1020, 0x1020, 0x1CA8, 0xF0A4, 0x5122, 0x1222, 0x10A0, 0x1040}},
    {36730, 2, {0x1000, 0x11FC, 0x1044, 0xFE44, 0x2044, 0x2844, 0x4894, 0x7E88, 0x900, 0x8FC, 0xE84, 0xF884, 0x4884, 0x884, 0x8FC, 0x884}},
    {36731, 2, {0x2000, 0x21FC, 0x2008, 0xFC10, 0x4030, 0x5048, 0x9084, 0xFF02, 0x1000, 0x11FC, 0x1C20, 0xF020, 0x5020, 0x1020, 0x13FE, 0x1000}},
    {36732, 2, {0x2010, 0x2014, 0x2012, 0xFC10, 0x41FE, 0x5010, 0x9010, 0xFDD0, 0x1090, 0x1090, 0x1C90, 0xF088, 0x50EA, 0x138A, 0x1006, 0x1002}},
    {36733, 2, {0x820, 0x828, 0x7F24, 0x820, 0xFFFE, 0x1020, 0x1024, 0xFF24, 0x2024, 0x4828, 0x7F28, 0x810, 0xF12, 0xF82A, 0x846, 0x882}},
    {36734, 2, {0x1000, 0x10FE, 0x1010, 0xFE20, 0x2044, 0x2882, 0x48FE, 0x7E12, 0x810, 0x810, 0xEFE, 0xF810, 0x4810, 0x810, 0x9FE, 0x800}},
    {36735, 2, {0x2008, 0x203C, 0x21E0, 0xFC20, 0x4020, 0x53FE, 0x9050, 0xFC88, 0x1104, 0x128A, 0x1C88, 0xF088, 0x5088, 0x1108, 0x1108, 0x1208}},
    {36736, 2, {0x2010, 0x2020, 0x20FC, 0xFC84, 0x40A4, 0x5094, 0x9094, 0xFDFE, 0x1084, 0x10A4, 0x1C94, 0xF094, 0x5084, 0x1104, 0x1114, 0x1208}},
    {36737, 2, {0x2020, 0x2020, 0x2050, 0xFC88, 0x4104, 0x5202, 0x91FC, 0xFC20, 0x1020, 0x1020, 0x1DFC, 0xF020, 0x5020, 0x1020, 0x13FE, 0x1000}},
    {36738, 2, {0x1040, 0x1040, 0x1078, 0xFE88, 0x2150, 0x2820, 0x4850, 0x7E88, 0x906, 0x8F8, 0xE88, 0xF888, 0x4888, 0x888, 0x8F8, 0x888}},
    {36739, 2, {0x2040, 0x2020, 0x2020, 0xFDFE, 0x4000, 0x5088, 0x9104, 0xFE02, 0x1088, 0x1088, 0x1C50, 0xF050, 0x5020, 0x1050, 0x1088, 0x1306}},
    {36740, 2, {0x2000, 0x23FE, 0x2120, 0xFD28, 0x4128, 0x51E8, 0x9128, 0xFD28, 0x1128, 0x11E8, 0x1D28, 0xF12A, 0x513A, 0x13EA, 0x1026, 0x1020}},
    {36741, 2, {0x2028, 0x2024, 0x23FE, 0xFC20, 0x4020, 0x51FC, 0x9124, 0xFD24, 0x11FC, 0x1124, 0x1D24, 0xF1FC, 0x5124, 0x1124, 0x1124, 0x110C}},
    {36742, 2, {0x2000, 0x23FE, 0x2050, 0xFC50, 0x4050, 0x53FE, 0x9252, 0xFE52, 0x1252, 0x12AA, 0x1EA6, 0xF302, 0x5202, 0x1202, 0x120A, 0x1204}},
    {36743, 2, {0x820, 0x820, 0x7EFC, 0x820, 0xFEFE, 0x1450, 0x2288, 0x4206, 0xBFF8, 0x500, 0x900, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {36744, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x640, 0x200, 0x3FF8, 0x500, 0x900, 0x1FF0, 0x100, 0xFFFE, 0x100, 0x100}},
    {36745, 2, {0x1000, 0x13FE, 0x9202, 0x5444, 0x5840, 0x13FC, 0xFC80, 0x28A0, 0x2920, 0x29FC, 0x2820, 0x2820, 0x2BFE, 0x4C20, 0x4820, 0x8020}},
    {36746, 2, {0x2000, 0x21FC, 0x2104, 0xFD04, 0x41FC, 0x5104, 0x9104, 0xFDFC, 0x1000, 0x1112, 0x1DD4, 0xF118, 0x5110, 0x1152, 0x1192, 0x110E}},
    {36747, 2, {0x2000, 0x23FE, 0x2202, 0xFA8A, 0x4252, 0x53FE, 0x9242, 0xFE22, 0x13FE, 0x1282, 0x1E82, 0xF282, 0x52FA, 0x1202, 0x120A, 0x1204}},
    {36748, 2, {0x2040, 0x2020, 0x23FE, 0xFC00, 0x4000, 0x51FC, 0x9104, 0xFD04, 0x11FC, 0x1020, 0x1D28, 0xF124, 0x5222, 0x1422, 0x10A0, 0x1040}},
    {36749, 2, {0x2000, 0x23DE, 0x2042, 0xFD4A, 0x4084, 0x514A, 0x9252, 0xFC20, 0x1000, 0x13DE, 0x1C52, 0xF152, 0x5094, 0x1148, 0x1254, 0x1422}},
    {36750, 2, {0x2092, 0x2092, 0x2124, 0xFA48, 0x4124, 0x5092, 0x9092, 0xFC00, 0x11FE, 0x1122, 0x1D22, 0xF1FE, 0x5122, 0x1122, 0x11FE, 0x1102}},
    {36751, 2, {0x2020, 0x2020, 0x23FE, 0xFC20, 0x41FC, 0x5040, 0x93FE, 0xFC88, 0x1104, 0x12FA, 0x1C20, 0xF1FC, 0x5020, 0x1050, 0x1088, 0x1304}},
    {36752, 2, {0x2000, 0x21FE, 0x2000, 0xFCFC, 0x4084, 0x5084, 0x90FC, 0xFC00, 0x11FE, 0x1122, 0x1D22, 0xF1FE, 0x5122, 0x1122, 0x11FE, 0x1102}},
    {36753, 2, {0x2000, 0x21F8, 0x2108, 0xFD08, 0x41F8, 0x5000, 0x97FE, 0xFD08, 0x11F8, 0x1108, 0x1DF8, 0xF108, 0x513E, 0x17C8, 0x1008, 0x1008}},
    {36754, 2, {0x1000, 0x10F8, 0x1088, 0xFE88, 0x20F8, 0x2888, 0x4888, 0x7EF8, 0x800, 0x9FC, 0xF54, 0xF954, 0x4954, 0x954, 0xBFE, 0x800}},
    {36755, 2, {0x2040, 0x20A0, 0x2110, 0xFA08, 0x25F6, 0x4000, 0x53C4, 0x9254, 0xFA54, 0x13D4, 0x1A54, 0xF254, 0x53D4, 0x1244, 0x1254, 0x12C8}},
    {36756, 2, {0x2208, 0x2208, 0x47D0, 0x8A22, 0xF53C, 0x27C8, 0x4110, 0xF93E, 0x7C0, 0x193E, 0xE100, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {36757, 2, {0x2020, 0x2020, 0x21FC, 0xFC20, 0x43FE, 0x5000, 0x91FC, 0xFD04, 0x1104, 0x11FC, 0x1C52, 0xF094, 0x5188, 0x16A4, 0x10C2, 0x1080}},
    {36758, 2, {0x2040, 0x2020, 0x23FE, 0xFE02, 0x4424, 0x51FC, 0x9020, 0xFDFC, 0x1020, 0x13FE, 0x1C20, 0xF1FC, 0x5104, 0x1104, 0x11FC, 0x1104}},
    {36759, 2, {0x2000, 0x23FC, 0x2204, 0xFA04, 0x43FC, 0x5248, 0x9248, 0xFEFC, 0x1248, 0x1248, 0x1FFE, 0xF2A4, 0x52A8, 0x1490, 0x14C8, 0x1886}},
    {36760, 2, {0x2040, 0x2020, 0x23FE, 0xFA50, 0x4250, 0x53FE, 0x9252, 0xFE52, 0x13FE, 0x1200, 0x1E92, 0xF2D4, 0x5298, 0x1492, 0x14D2, 0x188E}},
    {36761, 2, {0x2208, 0x2108, 0x27C8, 0xFA10, 0x249E, 0x4FD4, 0x5064, 0x9794, 0xFC94, 0x1794, 0x1C94, 0xF788, 0x5488, 0x1494, 0x14A4, 0x15C2}},
    {36762, 2, {0x2020, 0x2124, 0x20A8, 0xFBFE, 0x20A8, 0x4124, 0x5222, 0x9104, 0xF904, 0x11DE, 0x1A44, 0xF554, 0x509E, 0x1104, 0x1204, 0x1404}},
    {36763, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x100, 0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x100, 0x100, 0x100}},
    {36764, 2, {0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x820, 0x440, 0xFFFE, 0x100, 0x3FF8, 0x100, 0x100, 0x100}},
    {36765, 2, {0x1040, 0x1020, 0x2020, 0x29FC, 0x4400, 0xFE88, 0x4250, 0x1FE, 0x7C20, 0x4420, 0x45FC, 0x4420, 0x4420, 0x7C20, 0x4420, 0x20}},
    {36766, 2, {0x20, 0x1E10, 0xF010, 0x11FF, 0x1000, 0xFE44, 0x1024, 0x1028, 0x11FF, 0x7C10, 0x4410, 0x45FF, 0x4410, 0x4410, 0x7C10, 0x4410}},
    {36767, 2, {0x40, 0x7C20, 0x4420, 0x45FC, 0x4400, 0x7C88, 0x4050, 0x41FE, 0x7C20, 0x6420, 0xA5FC, 0xA420, 0x2420, 0x3C20, 0x2420, 0x20}},
    {36768, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x820, 0xFFFE, 0x100, 0x3FF8, 0x100, 0x100}},
    {36769, 2, {0x2040, 0x1020, 0x1020, 0xFDFC, 0x0, 0x4488, 0x2850, 0xFDFE, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x2020, 0x2020, 0x4020, 0x8020}},
    {36770, 2, {0x1040, 0x1020, 0xFE20, 0x11FC, 0x1000, 0xFE88, 0x9250, 0x93FE, 0xFE20, 0x1020, 0x39FC, 0x5420, 0x9220, 0x1020, 0x1020, 0x1020}},
    {36771, 2, {0x2020, 0x1020, 0x11FC, 0xFC20, 0x20, 0x45FC, 0x2924, 0xFD24, 0x11FC, 0x1020, 0xFC70, 0x10A8, 0x2124, 0x2222, 0x4020, 0x8020}},
    {36772, 2, {0x420, 0x1E10, 0xE010, 0x22FE, 0x9200, 0x5444, 0xFE28, 0x82FE, 0x10, 0x7C10, 0x47C, 0x6810, 0x1010, 0x2810, 0x4410, 0x8210}},
    {36773, 2, {0x820, 0x4A10, 0x5210, 0x7EFE, 0x2400, 0x4844, 0xBE28, 0x22FE, 0x2210, 0x3E10, 0x207C, 0x3E10, 0x2210, 0x2210, 0x3E10, 0x2210}},
    {36774, 2, {0x4210, 0x2208, 0x2208, 0xFA3E, 0x780, 0x5294, 0x2288, 0xFABE, 0x2288, 0x2288, 0x729C, 0x2288, 0x2488, 0x4488, 0x4A88, 0x9108}},
    {36775, 2, {0x4010, 0x2008, 0x2788, 0xFABE, 0x280, 0x5294, 0x2288, 0xFABE, 0x2288, 0x2288, 0x729C, 0x2288, 0x2488, 0x4488, 0x4A88, 0x9108}},
    {36776, 2, {0x2010, 0x1088, 0x1088, 0x7CBE, 0x80, 0x2A94, 0x1288, 0xFEBE, 0x1288, 0x1088, 0x7C9C, 0x1108, 0x1108, 0x2208, 0x2408, 0x4008}},
    {36777, 2, {0x2010, 0x1208, 0x1108, 0x7D3E, 0x0, 0x2B14, 0x1108, 0xFD3E, 0x1108, 0x1108, 0x7D1C, 0x1148, 0x1188, 0x2108, 0x2008, 0x4008}},
    {36778, 2, {0x820, 0x4F10, 0x4810, 0x48FE, 0xFF00, 0x844, 0x1028, 0x7EFE, 0x4210, 0x7E10, 0x407C, 0x7F10, 0x4110, 0x7F10, 0x4110, 0x10}},
    {36779, 2, {0x2010, 0x1088, 0x1088, 0x7D3E, 0x140, 0x2BD4, 0x1088, 0xFCBE, 0x1108, 0x13C8, 0x7C1C, 0x1008, 0x10C8, 0x2308, 0x2008, 0x4008}},
    {36780, 2, {0x4010, 0x2208, 0x2108, 0xF93E, 0x0, 0x57D4, 0x2088, 0xFCBE, 0x2288, 0x2108, 0x711C, 0x2288, 0x2288, 0x4488, 0x4808, 0x8008}},
    {36781, 2, {0xF20, 0xF010, 0x4910, 0x22FE, 0xFE00, 0x1444, 0x828, 0xFFFE, 0x9510, 0xBD10, 0x837C, 0xBD10, 0x9510, 0x8910, 0x9510, 0xA310}},
    {36782, 2, {0x4110, 0x2108, 0x2288, 0xFABE, 0x780, 0x5114, 0x2108, 0xFABE, 0x27C8, 0x2148, 0x711C, 0x2588, 0x2548, 0x4948, 0x4508, 0x8208}},
    {36783, 2, {0x4210, 0x2108, 0x2008, 0xFBBE, 0x0, 0x5394, 0x2008, 0xFBBE, 0x2008, 0x2388, 0x729C, 0x2288, 0x2288, 0x4388, 0x4288, 0x8008}},
    {36784, 2, {0x0, 0x3FFC, 0x2000, 0x2000, 0x2FF8, 0x2000, 0x2000, 0x3FFC, 0x2480, 0x2488, 0x2450, 0x2420, 0x4410, 0x4508, 0x8606, 0x400}},
    {36785, 2, {0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2888, 0x4850, 0x4A30, 0x8C0E, 0x20, 0xFFFE, 0x820, 0x420, 0x4A0, 0x40}},
    {36786, 2, {0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2888, 0x4850, 0x4A30, 0x8C0E}},
    {36787, 2, {0x820, 0x820, 0x7EFC, 0x820, 0x1C70, 0x2AA8, 0xC824, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFE, 0x2888, 0x4850, 0x4A30, 0x8C0E}},
    {36788, 2, {0x0, 0xEEFE, 0xAA80, 0xAA80, 0xEEBC, 0x80, 0xFE80, 0x92FE, 0xFEA0, 0x92B2, 0xFEB4, 0x10A8, 0xFEA8, 0x1124, 0x1132, 0x1220}},
    {36789, 2, {0x38, 0x3C0, 0x7C00, 0x38, 0x3C0, 0x3C0C, 0x70, 0x780, 0x7800, 0x100, 0x11F8, 0x1100, 0x1100, 0x2900, 0x47FE, 0x8000}},
    {36790, 2, {0x0, 0x2000, 0x1000, 0x1000, 0x0, 0x0, 0xF000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36791, 2, {0x0, 0x2000, 0x1000, 0x1000, 0x0, 0x7FC, 0xF000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36792, 2, {0x0, 0x2000, 0x17F8, 0x1108, 0x110, 0x120, 0xF13C, 0x1104, 0x1104, 0x1204, 0x1204, 0x1428, 0x1810, 0x2800, 0x47FE, 0x0}},
    {36793, 2, {0x40, 0x2040, 0x1040, 0x13FC, 0x44, 0x44, 0xF044, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410, 0x2800, 0x47FE, 0x0}},
    {36794, 2, {0x0, 0x2000, 0x17F8, 0x1088, 0x88, 0x88, 0xF088, 0x1088, 0x1088, 0x1108, 0x1108, 0x1228, 0x1410, 0x2800, 0x47FE, 0x0}},
    {36795, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x40, 0x7FE, 0xF040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36796, 2, {0x200, 0x2100, 0x1080, 0x1040, 0x40, 0xA0, 0xF0A0, 0x1110, 0x1110, 0x1208, 0x1404, 0x1802, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36797, 2, {0x0, 0x23F8, 0x1010, 0x1020, 0x40, 0x40, 0xF040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x1080, 0x2800, 0x47FE, 0x0}},
    {36798, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x7FE, 0x40, 0xF040, 0x10A0, 0x1090, 0x1108, 0x1204, 0x1404, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36799, 2, {0x40, 0x2040, 0x1040, 0x1444, 0x444, 0x444, 0xF444, 0x1444, 0x1444, 0x1444, 0x17FC, 0x1004, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36800, 2, {0x0, 0x21FC, 0x1020, 0x1020, 0x20, 0x20, 0xF3FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x1020, 0x2800, 0x47FE, 0x0}},
    {36801, 2, {0x10, 0x2078, 0x13C0, 0x1040, 0x40, 0x40, 0xF7FE, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36802, 2, {0x0, 0x23F8, 0x1040, 0x1040, 0x40, 0x40, 0xF7FC, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1140, 0x2880, 0x47FE, 0x0}},
    {36803, 2, {0x0, 0x23F8, 0x1000, 0x1000, 0x7FE, 0x80, 0xF100, 0x11F8, 0x1008, 0x1008, 0x1008, 0x1050, 0x1020, 0x2800, 0x47FE, 0x0}},
    {36804, 2, {0x100, 0x2100, 0x11FE, 0x1200, 0x400, 0x1F8, 0xF010, 0x1060, 0x1080, 0x1104, 0x1104, 0x10FC, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36805, 2, {0x0, 0x4FE0, 0x2220, 0x2220, 0x220, 0x220, 0xEFA0, 0x2220, 0x2224, 0x2214, 0x2214, 0x220C, 0x2204, 0x5000, 0x8FFE, 0x0}},
    {36806, 2, {0x40, 0x2040, 0x1040, 0x1258, 0x2E8, 0x748, 0xF248, 0x1268, 0x1254, 0x1244, 0x1204, 0x11FC, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36807, 2, {0x10, 0x2010, 0x1010, 0x17FE, 0x10, 0x10, 0xF210, 0x1110, 0x1110, 0x1010, 0x1010, 0x1050, 0x1020, 0x2800, 0x47FE, 0x0}},
    {36808, 2, {0x0, 0x2000, 0x17FE, 0x1080, 0x80, 0xFC, 0xF084, 0x1084, 0x1084, 0x1104, 0x1104, 0x1228, 0x1410, 0x2800, 0x47FE, 0x0}},
    {36809, 2, {0x0, 0x21FC, 0x1104, 0x1104, 0x104, 0x1FC, 0xF104, 0x1100, 0x1100, 0x1100, 0x1100, 0x1200, 0x1400, 0x2800, 0x47FE, 0x0}},
    {36810, 2, {0x0, 0x4FFE, 0x2040, 0x2040, 0x7FC, 0x444, 0xE444, 0x2444, 0x2444, 0x2454, 0x2448, 0x2040, 0x2040, 0x5000, 0x8FFE, 0x0}},
    {36811, 2, {0x0, 0x27FC, 0x1040, 0x1040, 0x40, 0x40, 0xF3F8, 0x1040, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36812, 2, {0x0, 0x21FC, 0x1104, 0x1104, 0x1FC, 0x104, 0xF104, 0x11FC, 0x1104, 0x1104, 0x1104, 0x1214, 0x1408, 0x2800, 0x47FE, 0x0}},
    {36813, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x248, 0x248, 0xF248, 0x13F8, 0x1044, 0x1044, 0x1044, 0x103C, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36814, 2, {0x0, 0x2080, 0x133C, 0x1224, 0x224, 0x224, 0xF224, 0x1224, 0x1224, 0x12B4, 0x1328, 0x1220, 0x1020, 0x2820, 0x47FE, 0x0}},
    {36815, 2, {0x40, 0x2040, 0x1040, 0x1040, 0x7FE, 0x40, 0xF040, 0x10A0, 0x1090, 0x1108, 0x1244, 0x1424, 0x1020, 0x2800, 0x47FE, 0x0}},
    {36816, 2, {0x0, 0x23F8, 0x1000, 0x1000, 0x0, 0x7FC, 0xF040, 0x1080, 0x1110, 0x1208, 0x17FC, 0x1204, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36817, 2, {0x8, 0x201C, 0x11E0, 0x1100, 0x100, 0x1FE, 0xF110, 0x1110, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410, 0x2800, 0x47FE, 0x0}},
    {36818, 2, {0x80, 0x2040, 0x1040, 0x17FE, 0x0, 0x0, 0xF1F0, 0x1110, 0x1110, 0x1110, 0x1212, 0x1212, 0x140E, 0x2800, 0x47FE, 0x0}},
    {36819, 2, {0x0, 0x23FC, 0x1010, 0x1110, 0x110, 0x210, 0xF3FE, 0x1030, 0x1050, 0x1090, 0x1110, 0x1650, 0x1020, 0x2800, 0x47FE, 0x0}},
    {36820, 2, {0x8, 0x203C, 0x13C0, 0x1200, 0x200, 0x3FC, 0xF204, 0x1288, 0x1250, 0x1220, 0x1250, 0x1488, 0x1504, 0x2800, 0x47FE, 0x0}},
    {36821, 2, {0x80, 0x2080, 0x1080, 0x11FC, 0x120, 0x220, 0xF020, 0x1020, 0x17FE, 0x1020, 0x1020, 0x1020, 0x1020, 0x2820, 0x47FE, 0x0}},
    {36822, 2, {0x50, 0x2048, 0x1048, 0x1040, 0x7FE, 0x40, 0xF040, 0x10A0, 0x1090, 0x1108, 0x1204, 0x1404, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36823, 2, {0x0, 0x23FC, 0x1040, 0x1040, 0x40, 0x40, 0xF7FE, 0x1040, 0x10A0, 0x1090, 0x1108, 0x1204, 0x1404, 0x2800, 0x47FE, 0x0}},
    {36824, 2, {0x0, 0x23FC, 0x1020, 0x1020, 0x40, 0x40, 0xF0D0, 0x1148, 0x1244, 0x1444, 0x1040, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36825, 2, {0x80, 0x2040, 0x1040, 0x17FC, 0x10, 0x210, 0xF120, 0x10A0, 0x1040, 0x10A0, 0x1110, 0x1208, 0x1408, 0x2800, 0x47FE, 0x0}},
    {36826, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x444, 0x444, 0xF444, 0x1444, 0x17FC, 0x1444, 0x1040, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36827, 2, {0x90, 0x2090, 0x1090, 0x13FC, 0x90, 0x90, 0xF090, 0x17FE, 0x1090, 0x1090, 0x1110, 0x1110, 0x1210, 0x2800, 0x47FE, 0x0}},
    {36828, 2, {0x0, 0x23F8, 0x1000, 0x1000, 0x0, 0x7FC, 0xF120, 0x1120, 0x1120, 0x1120, 0x1124, 0x1224, 0x1224, 0x141C, 0x2800, 0x47FE}},
    {36829, 2, {0x80, 0x4080, 0x2FFC, 0x2080, 0x80, 0x7F8, 0xE080, 0x2080, 0x2FFC, 0x2084, 0x2084, 0x2094, 0x2088, 0x5080, 0x8FFE, 0x0}},
    {36830, 2, {0x40, 0x2040, 0x17FE, 0x1080, 0xA0, 0x120, 0xF3FC, 0x1020, 0x1020, 0x1020, 0x17FE, 0x1020, 0x1020, 0x2820, 0x47FE, 0x0}},
    {36831, 2, {0x0, 0x21F8, 0x1108, 0x1108, 0x108, 0x1F8, 0xF108, 0x1120, 0x1110, 0x1208, 0x1204, 0x1404, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36832, 2, {0x40, 0x2040, 0x1040, 0x107E, 0x40, 0x40, 0xF3FC, 0x1204, 0x1204, 0x1204, 0x1204, 0x13FC, 0x1204, 0x2800, 0x47FE, 0x0}},
    {36833, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x3FC, 0x200, 0xF280, 0x1284, 0x1298, 0x12E0, 0x1282, 0x1282, 0x147E, 0x2800, 0x47FE, 0x0}},
    {36834, 2, {0x0, 0x23FC, 0x1084, 0x1084, 0x84, 0x114, 0xF208, 0x11FC, 0x1104, 0x1104, 0x1104, 0x11FC, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36835, 2, {0x48, 0x2248, 0x1248, 0x1248, 0x7FE, 0x248, 0xF248, 0x1248, 0x1278, 0x1200, 0x1200, 0x13FE, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36836, 2, {0x200, 0x2200, 0x13FC, 0x1440, 0x440, 0xA58, 0xF2E8, 0x1748, 0x1268, 0x1250, 0x1244, 0x1204, 0x11FC, 0x2800, 0x47FE, 0x0}},
    {36837, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x2F4, 0x294, 0xF294, 0x1294, 0x12F4, 0x1294, 0x1204, 0x1214, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36838, 2, {0x400, 0x4400, 0x243C, 0x2FA4, 0x4A4, 0x4A4, 0xE4A4, 0x24A4, 0x24A4, 0x28A4, 0x28BC, 0x32A4, 0x2100, 0x5000, 0x8FFE, 0x0}},
    {36839, 2, {0x40, 0x2040, 0x17FC, 0x1444, 0x444, 0x7FC, 0xF444, 0x1444, 0x17FC, 0x1444, 0x1040, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36840, 2, {0x80, 0x2080, 0x1110, 0x1208, 0x7FC, 0x204, 0xF000, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36841, 2, {0x80, 0x2080, 0x1100, 0x11FE, 0x202, 0x424, 0xF020, 0x1128, 0x1124, 0x1222, 0x1422, 0x10A0, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36842, 2, {0x40, 0x4040, 0x2040, 0x27FC, 0x444, 0x444, 0xE444, 0x27FC, 0x2444, 0x2444, 0x2444, 0x27FC, 0x2000, 0x5000, 0x8FFE, 0x0}},
    {36843, 2, {0x40, 0x2040, 0x1080, 0x13F8, 0x208, 0x208, 0xF208, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36844, 2, {0x80, 0x2040, 0x1000, 0x17FC, 0x40, 0x40, 0xF040, 0x13F8, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36845, 2, {0x40, 0x2240, 0x1240, 0x13FC, 0x240, 0x440, 0xF040, 0x17FE, 0x1040, 0x10A0, 0x1090, 0x1108, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36846, 2, {0x100, 0x2100, 0x1100, 0x13FE, 0x280, 0x480, 0xF0F8, 0x1080, 0x1080, 0x10FC, 0x1080, 0x1080, 0x1080, 0x2800, 0x47FE, 0x0}},
    {36847, 2, {0x110, 0x2110, 0x1110, 0x11D0, 0x258, 0x254, 0xF252, 0x1552, 0x1890, 0x1090, 0x1110, 0x1210, 0x1410, 0x2800, 0x47FE, 0x0}},
    {36848, 2, {0x40, 0x2050, 0x1048, 0x1040, 0x7FC, 0xC0, 0xF0E0, 0x1150, 0x1148, 0x1244, 0x1444, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36849, 2, {0x80, 0x4040, 0x27FC, 0x2404, 0x808, 0x200, 0xE210, 0x2260, 0x2380, 0x2204, 0x2204, 0x21FC, 0x2000, 0x5000, 0x8FFE, 0x0}},
    {36850, 2, {0x40, 0x2040, 0x1040, 0x13FC, 0x40, 0x40, 0xF7FE, 0x1040, 0x1080, 0x1110, 0x1208, 0x17FC, 0x1204, 0x2800, 0x47FE, 0x0}},
    {36851, 2, {0x0, 0x23F8, 0x1010, 0x1020, 0x50, 0x188, 0xF604, 0x13F8, 0x1040, 0x1040, 0x1040, 0x17FC, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36852, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x2F4, 0x294, 0xF294, 0x12F4, 0x1204, 0x1204, 0x13FC, 0x1204, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36853, 2, {0x0, 0x23FC, 0x1204, 0x12F4, 0x204, 0x204, 0xF2F4, 0x1294, 0x1294, 0x12F4, 0x1204, 0x1214, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36854, 2, {0x40, 0x2040, 0x17FE, 0x1080, 0x100, 0x3FC, 0xF504, 0x19FC, 0x1104, 0x11FC, 0x1104, 0x1114, 0x1108, 0x2800, 0x47FE, 0x0}},
    {36855, 2, {0x40, 0x2248, 0x1148, 0x1150, 0x40, 0x7FC, 0xF040, 0x10E0, 0x1150, 0x1248, 0x1444, 0x1040, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36856, 2, {0x208, 0x2110, 0x1000, 0x17FC, 0x110, 0x110, 0xF110, 0x17FE, 0x1110, 0x1110, 0x1210, 0x1210, 0x1410, 0x2800, 0x47FE, 0x0}},
    {36857, 2, {0x40, 0x2020, 0x1020, 0x17FE, 0x90, 0x90, 0xF294, 0x1292, 0x1492, 0x1110, 0x1110, 0x1250, 0x1420, 0x2800, 0x47FE, 0x0}},
    {36858, 2, {0x0, 0x4FFE, 0x20A0, 0x20A0, 0x7FC, 0x4A4, 0xE4A4, 0x24A4, 0x251C, 0x2604, 0x2404, 0x27FC, 0x2404, 0x5000, 0x8FFE, 0x0}},
    {36859, 2, {0x40, 0x2080, 0x11F8, 0x1610, 0x120, 0xC0, 0xF120, 0x167C, 0x1084, 0x1348, 0x1030, 0x10C0, 0x1700, 0x2800, 0x47FE, 0x0}},
    {36860, 2, {0x40, 0x2040, 0x1040, 0x17FC, 0x40, 0x40, 0xF3F8, 0x1000, 0x1000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE}},
    {36861, 2, {0x20, 0x2040, 0x13FC, 0x1204, 0x204, 0x3FC, 0xF200, 0x1200, 0x13FE, 0x1202, 0x1202, 0x13FE, 0x1202, 0x2800, 0x47FE, 0x0}},
    {36862, 2, {0x4, 0x27E4, 0x1104, 0x1114, 0x114, 0x3D4, 0xF254, 0x1254, 0x1554, 0x1094, 0x1094, 0x1104, 0x1214, 0x1408, 0x2800, 0x47FE}},
    {36863, 2, {0x100, 0x2100, 0x11FC, 0x1204, 0x404, 0x3E4, 0xF224, 0x1224, 0x13E4, 0x1224, 0x1224, 0x13E4, 0x1004, 0x1028, 0x2810, 0x47FE}},
    {36864, 2, {0x3F8, 0x2208, 0x1208, 0x13F8, 0x208, 0x208, 0xF3F8, 0x1244, 0x1228, 0x1210, 0x1288, 0x1304, 0x1204, 0x2800, 0x47FE, 0x0}},
    {36865, 2, {0x202, 0x4104, 0x3088, 0x1010, 0x3FE, 0x20, 0xF020, 0x1020, 0x17FF, 0x1020, 0x1050, 0x1088, 0x1104, 0x2A02, 0x4401, 0x83FF}},
    {36866, 2, {0x78, 0x23C0, 0x1040, 0x1040, 0x7FE, 0x40, 0xF040, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36867, 2, {0xA0, 0x20A0, 0x14A4, 0x12A8, 0x1B0, 0xA0, 0xF1B0, 0x12A8, 0x14A4, 0x1120, 0x1124, 0x1224, 0x141C, 0x2800, 0x47FE, 0x0}},
    {36868, 2, {0x80, 0x2080, 0x10FC, 0x1108, 0x290, 0x60, 0xF198, 0x1626, 0x1020, 0x11FC, 0x1020, 0x1220, 0x13FE, 0x1020, 0x2820, 0x47FE}},
    {36869, 2, {0x3C, 0x23C0, 0x1200, 0x1200, 0x3FE, 0x200, 0xF200, 0x12FC, 0x1284, 0x1484, 0x1484, 0x18FC, 0x1084, 0x2800, 0x47FE, 0x0}},
    {36870, 2, {0x208, 0x4108, 0x2110, 0x2FFE, 0x40, 0x40, 0xE444, 0x2444, 0x2444, 0x27FC, 0x2084, 0x2100, 0x2200, 0x5400, 0x8FFE, 0x0}},
    {36871, 2, {0x210, 0x4250, 0x2250, 0x225C, 0xF74, 0x2D4, 0xE254, 0x2254, 0x225C, 0x2350, 0x2E42, 0x2442, 0x203E, 0x5000, 0x8FFE, 0x0}},
    {36872, 2, {0x40, 0x2080, 0x13FC, 0x1204, 0x204, 0x2F4, 0xF294, 0x1294, 0x1294, 0x12F4, 0x1204, 0x1214, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36873, 2, {0x40, 0x2240, 0x1240, 0x13F8, 0x440, 0x40, 0xF7FC, 0x1120, 0x1120, 0x1120, 0x1224, 0x1224, 0x141C, 0x2800, 0x47FE, 0x0}},
    {36874, 2, {0x10, 0x4F90, 0x2090, 0x2110, 0x210, 0x2D8, 0xE354, 0x2254, 0x2692, 0x2A92, 0x2312, 0x2210, 0x2A50, 0x2420, 0x5000, 0x8FFE}},
    {36875, 2, {0x48, 0x4044, 0x2FFE, 0x2040, 0x7FC, 0x444, 0xE7FC, 0x2444, 0x27FC, 0x2444, 0x2444, 0x2454, 0x2408, 0x5000, 0x8FFE, 0x0}},
    {36876, 2, {0x40, 0x2040, 0x107C, 0x1040, 0x40, 0x3FC, 0xF204, 0x13F4, 0x1214, 0x13F4, 0x1204, 0x13FC, 0x1204, 0x2800, 0x47FE, 0x0}},
    {36877, 2, {0x40, 0x2248, 0x1148, 0x1150, 0x40, 0x3F8, 0xF208, 0x1208, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x1228, 0x2A10, 0x47FE}},
    {36878, 2, {0x0, 0x27FE, 0x1090, 0x1090, 0x3FC, 0x294, 0xF294, 0x130C, 0x1204, 0x13FC, 0x1204, 0x1204, 0x13FC, 0x1204, 0x2800, 0x47FE}},
    {36879, 2, {0x38, 0x23C0, 0x1040, 0x17FC, 0x150, 0x248, 0xF446, 0x13F0, 0x1110, 0x113C, 0x1104, 0x1214, 0x1408, 0x2800, 0x47FE, 0x0}},
    {36880, 2, {0x0, 0x23FC, 0x1040, 0x1080, 0x144, 0x668, 0xF0B0, 0x1128, 0x1668, 0x10A4, 0x1124, 0x1620, 0x10A0, 0x2840, 0x47FE, 0x0}},
    {36881, 2, {0x40, 0x2050, 0x1048, 0x17FC, 0x40, 0x248, 0xF150, 0x10E0, 0x1150, 0x1248, 0x1444, 0x1140, 0x1080, 0x2800, 0x47FE, 0x0}},
    {36882, 2, {0x110, 0x20A0, 0x13F8, 0x1048, 0x48, 0x3F8, 0xF240, 0x1240, 0x13FC, 0x10C4, 0x1144, 0x1254, 0x1448, 0x2840, 0x47FE, 0x0}},
    {36883, 2, {0x3C, 0x47E0, 0x2420, 0x25FC, 0x420, 0x7FE, 0xE420, 0x2420, 0x25FC, 0x2524, 0x2524, 0x2934, 0x2928, 0x3020, 0x5000, 0x8FFE}},
    {36884, 2, {0x40, 0x20A0, 0x1110, 0x1208, 0x5F6, 0x40, 0xF040, 0x17FC, 0x1040, 0x1248, 0x1244, 0x1444, 0x1140, 0x2880, 0x47FE, 0x0}},
    {36885, 2, {0x0, 0x27FC, 0x1000, 0x1248, 0x490, 0x248, 0xF000, 0x17FC, 0x1040, 0x1040, 0x1040, 0x1FFE, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36886, 2, {0x910, 0x4510, 0x2210, 0x2652, 0xA52, 0x254, 0xE290, 0x2628, 0x2A24, 0x3244, 0x2242, 0x2A82, 0x2500, 0x5000, 0x8FFE, 0x0}},
    {36887, 2, {0x0, 0x27FC, 0x1000, 0x1000, 0x3F8, 0x208, 0xF208, 0x13F8, 0x1000, 0x1208, 0x1110, 0x17FC, 0x1000, 0x2800, 0x47FE, 0x0}},
    {36888, 2, {0x80, 0x2100, 0x1208, 0x17FC, 0x104, 0x100, 0xF3F8, 0x1440, 0x1040, 0x17FE, 0x1040, 0x10A0, 0x1110, 0x2A08, 0x47FE, 0x0}},
    {36889, 2, {0x80, 0x2040, 0x17FC, 0x1000, 0x3F8, 0x0, 0xF3F8, 0x1000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36890, 2, {0x0, 0x47F8, 0x2010, 0x21A0, 0x40, 0x7FC, 0xE444, 0x2444, 0x27FC, 0x2444, 0x2444, 0x27FC, 0x2444, 0x2454, 0x5408, 0x8FFE}},
    {36891, 2, {0x0, 0x4880, 0x257C, 0x2210, 0xD10, 0x110, 0xE17C, 0x2310, 0x2D10, 0x2110, 0x2110, 0x2A7E, 0x2400, 0x5000, 0x8FFE, 0x0}},
    {36892, 2, {0x0, 0x4FF8, 0x2080, 0x2080, 0x7F0, 0x110, 0xE110, 0x2FFC, 0x2000, 0x27F8, 0x2408, 0x2408, 0x27F8, 0x5000, 0x8FFE, 0x0}},
    {36893, 2, {0x208, 0x421C, 0x2260, 0x2240, 0xFC0, 0x27E, 0xE2C8, 0x2348, 0x2E48, 0x2248, 0x2248, 0x2A48, 0x2488, 0x5100, 0x8FFE, 0x0}},
    {36894, 2, {0x3F8, 0x2208, 0x1208, 0x1208, 0x3F8, 0x0, 0xF7FC, 0x1040, 0x1040, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x2800, 0x47FE, 0x0}},
    {36895, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x3F8, 0x248, 0xF248, 0x13F8, 0x10E0, 0x1150, 0x1248, 0x1444, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36896, 2, {0x40, 0x2240, 0x13F8, 0x1440, 0x40, 0x7FC, 0xF000, 0x1000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36897, 2, {0x80, 0x2110, 0x1208, 0x17FC, 0x4, 0x110, 0xF208, 0x1504, 0x11F8, 0x1208, 0x1590, 0x1060, 0x1190, 0x2E08, 0x47FE, 0x0}},
    {36898, 2, {0x100, 0x21F8, 0x1308, 0x1490, 0x60, 0x198, 0xF606, 0x1040, 0x13F8, 0x1040, 0x13F8, 0x1040, 0x17FC, 0x1040, 0x2840, 0x47FE}},
    {36899, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x7FC, 0x444, 0xE7FC, 0x2444, 0x27FC, 0x2040, 0x2FFE, 0x2040, 0x2040, 0x5040, 0x8FFE, 0x0}},
    {36900, 2, {0x10, 0x2210, 0x1110, 0x1454, 0x252, 0x92, 0xF010, 0x1114, 0x1604, 0x1208, 0x1210, 0x1220, 0x10C0, 0x2800, 0x47FE, 0x0}},
    {36901, 2, {0x0, 0x23FC, 0x1294, 0x1294, 0x2F4, 0x294, 0xF294, 0x12F4, 0x1294, 0x1294, 0x1294, 0x13FC, 0x1204, 0x2800, 0x47FE, 0x0}},
    {36902, 2, {0x0, 0x4FFE, 0x2000, 0x2000, 0x7BC, 0x4A4, 0xE4A4, 0x26B4, 0x25AC, 0x24A4, 0x24A4, 0x24A4, 0x25AC, 0x5000, 0x8FFE, 0x0}},
    {36903, 2, {0x110, 0x2248, 0x1444, 0x10A0, 0x110, 0x208, 0xF406, 0x13F8, 0x1208, 0x1208, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36904, 2, {0x40, 0x2040, 0x17FC, 0x1040, 0x150, 0x150, 0xF2E8, 0x1444, 0x10E0, 0x1150, 0x1248, 0x1444, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36905, 2, {0x0, 0x2080, 0x17FC, 0x1110, 0x248, 0xC46, 0xF3F8, 0x1040, 0x1150, 0x1110, 0x17FE, 0x1110, 0x1210, 0x1410, 0x2810, 0x47FE}},
    {36906, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x110, 0xFFE, 0xF000, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36907, 2, {0x0, 0x47BC, 0x2084, 0x2528, 0x210, 0x528, 0xE840, 0x27BC, 0x2084, 0x2528, 0x2210, 0x2528, 0x2844, 0x5000, 0x8FFE, 0x0}},
    {36908, 2, {0x84, 0x2084, 0x1108, 0x17BC, 0x108, 0x108, 0xF108, 0x1108, 0x17BE, 0x1108, 0x1108, 0x1108, 0x1108, 0x2A08, 0x47FE, 0x0}},
    {36909, 2, {0x80, 0x2040, 0x17FC, 0x1404, 0x3F0, 0x210, 0xF210, 0x13F0, 0x1200, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x2800, 0x47FE, 0x0}},
    {36910, 2, {0x40, 0x2040, 0x13F8, 0x1048, 0x7FE, 0x48, 0xF3F8, 0x1444, 0x1248, 0x1150, 0x1248, 0x1444, 0x1140, 0x2880, 0x47FE, 0x0}},
    {36911, 2, {0x0, 0x47F8, 0x2008, 0x23F8, 0x8, 0xFFE, 0xE040, 0x2444, 0x22E8, 0x2150, 0x2248, 0x2444, 0x2140, 0x5080, 0x8FFE, 0x0}},
    {36912, 2, {0x220, 0x4220, 0x223E, 0x2F40, 0x280, 0x27C, 0xE304, 0x2208, 0x2E7E, 0x2208, 0x2208, 0x2208, 0x2A28, 0x2410, 0x5000, 0x8FFE}},
    {36913, 2, {0x0, 0x47FC, 0x2444, 0x2444, 0x5F4, 0x444, 0xE7FC, 0x2404, 0x25F4, 0x2514, 0x25F4, 0x2404, 0x280C, 0x5000, 0x8FFE, 0x0}},
    {36914, 2, {0x140, 0x2120, 0x13FE, 0x1220, 0x620, 0xBFC, 0xF220, 0x1220, 0x13FC, 0x1220, 0x1220, 0x13FE, 0x1200, 0x2800, 0x47FE, 0x0}},
    {36915, 2, {0x80, 0x2040, 0x17FE, 0x1100, 0x208, 0x7FC, 0xF004, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1228, 0x2A10, 0x47FE}},
    {36916, 2, {0x40, 0x2040, 0x107E, 0x1040, 0x3F8, 0x208, 0xF3F8, 0x1208, 0x13F8, 0x1040, 0x1FFE, 0x1040, 0x1040, 0x2840, 0x47FE, 0x0}},
    {36917, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x40, 0x7FC, 0xF110, 0x1248, 0x1444, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x1000, 0x2800, 0x47FE}},
    {36918, 2, {0x38, 0x23C0, 0x1040, 0x17FC, 0x150, 0x248, 0xF444, 0x1080, 0x1FFE, 0x1108, 0x1390, 0x1060, 0x1198, 0x1604, 0x2800, 0x47FE}},
    {36919, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x208, 0x3F8, 0xF100, 0x13FC, 0x1494, 0x1924, 0x1244, 0x14A8, 0x1110, 0x2A00, 0x47FE, 0x0}},
    {36920, 2, {0x100, 0x4100, 0x21F0, 0x2210, 0x420, 0xFFC, 0xE444, 0x2444, 0x27FC, 0x20A0, 0x2130, 0x212A, 0x2222, 0x241E, 0x5000, 0x8FFE}},
    {36921, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x40, 0x7FC, 0xF110, 0x10A0, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x1040, 0x1040, 0x2840, 0x47FE}},
    {36922, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x7FE, 0x80, 0xF110, 0x13E0, 0x1044, 0x11A8, 0x1690, 0x1088, 0x10C4, 0x2800, 0x47FE, 0x0}},
    {36923, 2, {0x0, 0x47FC, 0x24A4, 0x24A4, 0x7FC, 0x80, 0xE1F8, 0x2208, 0x2510, 0x20A0, 0x2040, 0x2180, 0x2E00, 0x5000, 0x8FFE, 0x0}},
    {36924, 2, {0x0, 0x27FC, 0x1000, 0x13F8, 0x208, 0x3F8, 0xF000, 0x17FC, 0x1444, 0x17FC, 0x1444, 0x17FC, 0x1404, 0x2800, 0x47FE, 0x0}},
    {36925, 2, {0x108, 0x2108, 0x17FE, 0x1108, 0x40, 0x7FE, 0xF080, 0x1100, 0x13FC, 0x1504, 0x1904, 0x11FC, 0x1104, 0x2800, 0x47FE, 0x0}},
    {36926, 2, {0x40, 0x40A0, 0x2110, 0x2208, 0xDF6, 0x0, 0xE784, 0x24A4, 0x27A4, 0x24A4, 0x27A4, 0x2484, 0x2494, 0x2588, 0x5000, 0x8FFE}},
    {36927, 2, {0x1F8, 0x2108, 0x11F8, 0x1108, 0x1F8, 0x0, 0xF7FE, 0x1100, 0x13FC, 0x14A4, 0x1124, 0x1254, 0x1088, 0x2800, 0x47FE, 0x0}},
    {36928, 2, {0x80, 0x27FC, 0x1100, 0x13F8, 0x440, 0xBFC, 0xF000, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1228, 0x2A10, 0x47FE}},
    {36929, 2, {0x1E, 0x23E0, 0x1220, 0x1220, 0x3FE, 0x220, 0xF2FC, 0x1284, 0x12FC, 0x1284, 0x12FC, 0x1484, 0x14FC, 0x2800, 0x47FE, 0x0}},
    {36930, 2, {0x110, 0x2090, 0x10A0, 0x17FC, 0x80, 0x144, 0xF668, 0x10B0, 0x1128, 0x1668, 0x10A4, 0x1124, 0x16A0, 0x2840, 0x47FE, 0x0}},
    {36931, 2, {0x40, 0x2020, 0x13FC, 0x1000, 0x108, 0x90, 0xF3FE, 0x1220, 0x12C8, 0x1210, 0x1224, 0x12C8, 0x1210, 0x1460, 0x2980, 0x47FE}},
    {36932, 2, {0x40, 0x4444, 0x2444, 0x2444, 0x7FC, 0x0, 0xEFFE, 0x2080, 0x27FC, 0x24A4, 0x24A4, 0x24A4, 0x24A4, 0x240C, 0x5000, 0x8FFE}},
    {36933, 2, {0x0, 0x4FFC, 0x2804, 0x2804, 0xFFC, 0x910, 0xE8A0, 0x2FFC, 0x2840, 0x2BF8, 0x2840, 0x37FC, 0x2040, 0x5040, 0x8FFE, 0x0}},
    {36934, 2, {0x80, 0x2040, 0x17FC, 0x1110, 0xA0, 0x7FC, 0xF444, 0x1040, 0x13F8, 0x1248, 0x1248, 0x1268, 0x1250, 0x1040, 0x2840, 0x47FE}},
    {36935, 2, {0x0, 0x23F8, 0x1248, 0x13F8, 0x248, 0x3F8, 0xF040, 0x17FC, 0x1444, 0x1454, 0x1474, 0x1594, 0x1404, 0x140C, 0x2800, 0x47FE}},
    {36936, 2, {0x0, 0x43F8, 0x2208, 0x23F8, 0x208, 0x3F8, 0xE000, 0x2FFC, 0x2040, 0x2278, 0x2240, 0x2540, 0x28FE, 0x5000, 0x8FFE, 0x0}},
    {36937, 2, {0x40, 0x2040, 0x107C, 0x1040, 0x3F8, 0x208, 0xF3F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x10A0, 0x1110, 0x2A08, 0x47FE, 0x0}},
    {36938, 2, {0x420, 0x4220, 0x2FBE, 0x2440, 0x43C, 0x784, 0xE488, 0x2488, 0x24FE, 0x2488, 0x2488, 0x2888, 0x29A8, 0x5010, 0x8FFE, 0x0}},
    {36939, 2, {0x0, 0x27FC, 0x1444, 0x13F8, 0x40, 0x3F8, 0xF248, 0x13F8, 0x1248, 0x13F8, 0x1040, 0x17FC, 0x1040, 0x2840, 0x47FE, 0x0}},
    {36940, 2, {0x0, 0x27BC, 0x14A4, 0x17BC, 0x0, 0x3F8, 0xF000, 0x17FE, 0x1100, 0x11F8, 0x1008, 0x1008, 0x1050, 0x2820, 0x47FE, 0x0}},
    {36941, 2, {0x80, 0x2040, 0x13FC, 0x1204, 0x204, 0x3FC, 0xF200, 0x13FC, 0x1354, 0x1554, 0x15FC, 0x1954, 0x1144, 0x110C, 0x2800, 0x47FE}},
    {36942, 2, {0x1F8, 0x2108, 0x11E8, 0x1128, 0x128, 0x3FC, 0xF204, 0x12F4, 0x1294, 0x12F4, 0x1204, 0x1214, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36943, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x208, 0x3F8, 0xF100, 0x13FC, 0x1444, 0x1AA4, 0x1204, 0x13F4, 0x1014, 0x2808, 0x47FE, 0x0}},
    {36944, 2, {0x0, 0x47BC, 0x2484, 0x2484, 0x7BC, 0x400, 0xE400, 0x27BC, 0x2404, 0x2404, 0x27A8, 0x2410, 0x2428, 0x5444, 0x8FFE, 0x0}},
    {36945, 2, {0x40, 0x2080, 0x13F8, 0x1208, 0x3F8, 0x208, 0xF3F8, 0x1000, 0x17FC, 0x1040, 0x13F8, 0x1040, 0x17FC, 0x2800, 0x47FE, 0x0}},
    {36946, 2, {0x208, 0x4110, 0x2FFE, 0x20A0, 0x7FC, 0x4A4, 0xE4A4, 0x2524, 0x261C, 0x2404, 0x27FC, 0x2404, 0x27FC, 0x5000, 0x8FFE, 0x0}},
    {36947, 2, {0x208, 0x2110, 0x1000, 0x17FC, 0x80, 0x3F8, 0xF208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x2800, 0x47FE, 0x0}},
    {36948, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x7FC, 0x110, 0xF0A0, 0x17FC, 0x1040, 0x13F8, 0x1040, 0x17FC, 0x1040, 0x2840, 0x47FE, 0x0}},
    {36949, 2, {0x80, 0x23F8, 0x1108, 0x17FE, 0x0, 0x3F8, 0xF208, 0x13F8, 0x1040, 0x17FC, 0x1040, 0x1440, 0x17FE, 0x1040, 0x2840, 0x47FE}},
    {36950, 2, {0x40, 0x4040, 0x2FFE, 0x2040, 0x7FC, 0x514, 0xE4A4, 0x25F4, 0x2444, 0x25F4, 0x2444, 0x2454, 0x2448, 0x5000, 0x8FFE, 0x0}},
    {36951, 2, {0x40, 0x23F8, 0x1248, 0x13F8, 0x40, 0xFFE, 0xF000, 0x13F8, 0x1208, 0x1248, 0x1248, 0x10A0, 0x1110, 0x2A08, 0x47FE, 0x0}},
    {36952, 2, {0x110, 0x27FC, 0x1110, 0x13F8, 0x110, 0x7FC, 0xF040, 0x13F8, 0x1248, 0x13F8, 0x1248, 0x17FE, 0x1208, 0x1228, 0x2A10, 0x47FE}},
    {36953, 2, {0x40, 0x20FC, 0x1324, 0x1098, 0xE0, 0x700, 0xF1F8, 0x1240, 0x1040, 0x17FC, 0x1040, 0x1244, 0x13FC, 0x2800, 0x47FE, 0x0}},
    {36954, 2, {0x40, 0x2140, 0x165C, 0x1444, 0x75C, 0x444, 0xF7FC, 0x1040, 0x17FC, 0x1008, 0x1190, 0x1060, 0x1198, 0x1604, 0x2800, 0x47FE}},
    {36955, 2, {0x80, 0x477C, 0x2424, 0x2524, 0x4A4, 0x554, 0xE648, 0x2080, 0x27FC, 0x2444, 0x27FC, 0x2444, 0x27FC, 0x5000, 0x8FFE, 0x0}},
    {36956, 2, {0x6, 0x4FB8, 0x2090, 0x2120, 0x244, 0x278, 0xE310, 0x2E24, 0x227E, 0x2212, 0x2254, 0x2A92, 0x2450, 0x5020, 0x8FFE, 0x0}},
    {36957, 2, {0x0, 0x27FC, 0x14A4, 0x14A4, 0x7FC, 0x40, 0xF150, 0x1248, 0x1444, 0x1150, 0x1248, 0x1444, 0x1040, 0x2800, 0x47FE, 0x0}},
    {36958, 2, {0x1E, 0x47E0, 0x243C, 0x2420, 0x5FE, 0x522, 0xE538, 0x25E2, 0x251E, 0x2500, 0x2978, 0x2A48, 0x328A, 0x2506, 0x5000, 0x8FFE}},
    {36959, 2, {0x0, 0x23FC, 0x1204, 0x1204, 0x3FC, 0x240, 0xF220, 0x13FC, 0x1288, 0x1250, 0x13FE, 0x1220, 0x15FC, 0x1420, 0x2820, 0x47FE}},
    {36960, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x7FC, 0x0, 0xF3F8, 0x1208, 0x13F8, 0x10C4, 0x11A8, 0x1290, 0x1488, 0x1084, 0x2800, 0x47FE}},
    {36961, 2, {0x440, 0x225E, 0x1292, 0x1012, 0x7D2, 0x11E, 0xF552, 0x1552, 0x1552, 0x17DE, 0x1152, 0x1212, 0x1222, 0x142A, 0x2844, 0x47FE}},
    {36962, 2, {0x0, 0x23F8, 0x1208, 0x13F8, 0x208, 0x3F8, 0xF000, 0x17BC, 0x1084, 0x14A4, 0x1294, 0x14A4, 0x1294, 0x1108, 0x2800, 0x47FE}},
    {36963, 2, {0x40, 0x23F8, 0x1248, 0x13F8, 0x40, 0x7FE, 0xF000, 0x13F8, 0x1208, 0x13F8, 0x1200, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE}},
    {36964, 2, {0x0, 0x23FC, 0x1240, 0x13F8, 0x240, 0x3F8, 0xF240, 0x13FC, 0x1004, 0x1554, 0x1554, 0x1554, 0x1428, 0x2810, 0x47FE, 0x0}},
    {36965, 2, {0x3C, 0x23C0, 0x1004, 0x1244, 0x128, 0x1FC, 0xF220, 0x1020, 0x13FE, 0x1020, 0x1124, 0x1124, 0x1124, 0x11FC, 0x2800, 0x47FE}},
    {36966, 2, {0x3FC, 0x2244, 0x17FE, 0x1244, 0x3FC, 0x0, 0xF3FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1204, 0x13FC, 0x1108, 0x2A04, 0x47FE}},
    {36967, 2, {0x40, 0x47FC, 0x2110, 0x2FFE, 0x0, 0x3F8, 0xE208, 0x23F8, 0x2208, 0x23F8, 0x2040, 0x2FFE, 0x2040, 0x5040, 0x8FFE, 0x0}},
    {36968, 2, {0x210, 0x4210, 0x2F90, 0x221E, 0x7A4, 0x244, 0xEF94, 0x2414, 0x2788, 0x2488, 0x2494, 0x2AA2, 0x2942, 0x5000, 0x8FFE, 0x0}},
    {36969, 2, {0x80, 0x4040, 0x27FC, 0x2110, 0xA0, 0x7FC, 0xE444, 0x27FC, 0x2444, 0x25F4, 0x2514, 0x25F4, 0x2404, 0x240C, 0x5000, 0x8FFE}},
    {36970, 2, {0x80, 0x2110, 0x13F8, 0x1210, 0x4A4, 0x7FC, 0xF0A0, 0x1318, 0x1C46, 0x1390, 0x1064, 0x1398, 0x1060, 0x1380, 0x2800, 0x47FE}},
    {36971, 2, {0x210, 0x4210, 0x2F90, 0x223E, 0xFA4, 0xAC4, 0xEAA4, 0x2F94, 0x2208, 0x2708, 0x2A94, 0x3224, 0x2244, 0x5000, 0x8FFE, 0x0}},
    {36972, 2, {0x220, 0x4220, 0x2FA0, 0x223C, 0xFC4, 0xA90, 0xEA90, 0x2F90, 0x2210, 0x2728, 0x2AA8, 0x3244, 0x2284, 0x5000, 0x8FFE, 0x0}},
    {36973, 2, {0xA0, 0x2FFE, 0x10A0, 0x17FC, 0x4A4, 0x7FC, 0xF4A4, 0x17FC, 0x1000, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2800, 0x47FE}},
    {36974, 2, {0x40, 0x4020, 0x27FE, 0x2400, 0x488, 0x488, 0xE7FE, 0x2488, 0x2488, 0x24F8, 0x2800, 0x2AA4, 0x3252, 0x2452, 0x5000, 0x8FFE}},
    {36975, 2, {0x0, 0x477E, 0x2510, 0x2530, 0x74A, 0x59A, 0xE52C, 0x274C, 0x259A, 0x252A, 0x254A, 0x25A8, 0x2B10, 0x5000, 0x8FFE, 0x0}},
    {36976, 2, {0x2A8, 0x22A8, 0x17FC, 0x12AA, 0x4E6, 0x800, 0xF7FC, 0x1444, 0x1040, 0x13F8, 0x1248, 0x1248, 0x1258, 0x2840, 0x47FE, 0x0}},
    {36977, 2, {0x40, 0x43F8, 0x2248, 0x2FFE, 0x248, 0x3F8, 0xE040, 0x23F8, 0x2248, 0x23F8, 0x2080, 0x2FFE, 0x2110, 0x20E0, 0x5318, 0x8FFE}},
    {36978, 2, {0x7FC, 0x4404, 0x27FC, 0x2420, 0x4A8, 0x470, 0xE4A8, 0x2504, 0x24A0, 0x24FC, 0x2520, 0x2420, 0x2BFE, 0x2820, 0x5020, 0x8FFE}},
    {36979, 2, {0x110, 0x2110, 0x17FC, 0x1110, 0x40, 0x248, 0xF248, 0x1554, 0x1040, 0x13F8, 0x1040, 0x1040, 0x17FC, 0x2800, 0x47FE, 0x0}},
    {36980, 2, {0x40, 0x2248, 0x1150, 0x17FC, 0x150, 0x248, 0xF444, 0x1210, 0x1390, 0x14BC, 0x1A90, 0x1150, 0x117E, 0x1210, 0x2A10, 0x47FE}},
    {36981, 2, {0x110, 0x4FFE, 0x20A0, 0x27FC, 0x4A4, 0x51C, 0xE604, 0x27FC, 0x2404, 0x27FC, 0x2010, 0x2FFE, 0x2110, 0x20B0, 0x5000, 0x8FFE}},
    {36982, 2, {0x40, 0x2040, 0x13F8, 0x1040, 0x7FC, 0x210, 0xF738, 0x1210, 0x1F7C, 0x1000, 0x17FC, 0x1120, 0x1224, 0x141C, 0x2800, 0x47FE}},
    {36983, 2, {0x7FC, 0x20A0, 0x17FC, 0x14A4, 0x7FC, 0x80, 0xF7FE, 0x1208, 0x15F4, 0x1912, 0x11F0, 0x1104, 0x10FC, 0x2800, 0x47FE, 0x0}},
    {36984, 2, {0xF3C, 0x4924, 0x2F3C, 0x28A2, 0x79E, 0x110, 0xE7FC, 0x2110, 0x2110, 0x2FFE, 0x2110, 0x2208, 0x2404, 0x5000, 0x8FFE, 0x0}},
    {36985, 2, {0x0, 0x23F8, 0x1090, 0x1060, 0x7FC, 0x124, 0xF6A8, 0x1040, 0x17FC, 0x14A4, 0x1514, 0x16EC, 0x14A4, 0x14EC, 0x2800, 0x47FE}},
    {36986, 2, {0x40, 0x23F8, 0x1248, 0x13F8, 0x40, 0x7FE, 0xF000, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x13F8, 0x2910, 0x47FE}},
    {36987, 2, {0x0, 0x27BC, 0x14A4, 0x14A4, 0x7BC, 0x110, 0xF0A0, 0x17FC, 0x1040, 0x1248, 0x1248, 0x13F8, 0x1040, 0x1080, 0x2900, 0x47FE}},
    {36988, 2, {0x40, 0x4040, 0x2FFC, 0x20A0, 0x514, 0x3F8, 0xE60C, 0x2BFA, 0x2208, 0x23F8, 0x2040, 0x2248, 0x2444, 0x2944, 0x5080, 0x8FFE}},
    {36989, 2, {0x40, 0x407C, 0x2040, 0x27FE, 0x442, 0x5F8, 0xE444, 0x27FC, 0x24C0, 0x2722, 0x28D4, 0x2B18, 0x3074, 0x2392, 0x5020, 0x8FFE}},
    {36990, 2, {0x210, 0x4210, 0x23DE, 0x2528, 0x884, 0x7FC, 0xE040, 0x2248, 0x2248, 0x2554, 0x28E2, 0x2040, 0x2FFE, 0x5000, 0x8FFE, 0x0}},
    {36991, 2, {0x10, 0x4788, 0x24BE, 0x2480, 0x4A2, 0x794, 0xE43E, 0x2808, 0x2F88, 0x34BE, 0x2488, 0x2788, 0x2488, 0x5000, 0x8FFE, 0x0}},
    {36992, 2, {0x200, 0x4410, 0x2F90, 0x2890, 0xF9E, 0x8A4, 0xEFC4, 0x2214, 0x3FD4, 0x2408, 0x2788, 0x2894, 0x32A2, 0x2142, 0x5000, 0x8FFE}},
    {36993, 2, {0x110, 0x4FFE, 0x2110, 0x23F8, 0x248, 0x3F8, 0xE248, 0x23F8, 0x2040, 0x27FC, 0x2454, 0x25F4, 0x2414, 0x240C, 0x5000, 0x8FFE}},
    {36994, 2, {0x400, 0x4F3C, 0x2914, 0x3214, 0xF94, 0xAA8, 0xEAA8, 0x2FBC, 0x2AC8, 0x2A88, 0x2FBE, 0x2A88, 0x2A88, 0x3188, 0x5008, 0x8FFE}},
    {36995, 2, {0x80, 0x4040, 0x2FFE, 0x2A12, 0x140C, 0x120, 0xE7FC, 0x2080, 0x2144, 0x2668, 0x21B0, 0x2668, 0x21A4, 0x26A4, 0x5040, 0x8FFE}},
    {36996, 2, {0x0, 0x47FC, 0x24A4, 0x27FC, 0x0, 0xFFE, 0xE000, 0x23F8, 0x2208, 0x23F8, 0x2144, 0x2328, 0x2D90, 0x5108, 0x8FFE, 0x0}},
    {36997, 2, {0x20, 0x27FE, 0x1000, 0x13FC, 0x204, 0x2F4, 0xF294, 0x13FC, 0x1108, 0x11F8, 0x1108, 0x11F8, 0x1000, 0x17FE, 0x2800, 0x47FE}},
    {36998, 2, {0x24, 0x23A8, 0x1092, 0x1514, 0x208, 0x5F4, 0xF002, 0x13F8, 0x1208, 0x13F8, 0x1208, 0x1110, 0x17FC, 0x2800, 0x47FE, 0x0}},
    {36999, 2, {0x7FC, 0x4248, 0x2444, 0x2FFE, 0x444, 0x6EC, 0xE554, 0x26EC, 0x2444, 0x26EC, 0x2554, 0x26EC, 0x2444, 0x240C, 0x5000, 0x8FFE}},
    {37000, 2, {0x110, 0x4220, 0x2D7C, 0x25C4, 0x244, 0xD7C, 0xE344, 0x2544, 0x297C, 0x2328, 0x2528, 0x292A, 0x2546, 0x5280, 0x8FFE, 0x0}},
    {37001, 2, {0x80, 0x47FC, 0x2404, 0x27FC, 0x404, 0x7FC, 0xE404, 0x27FC, 0x2040, 0x2FFE, 0x2912, 0x23F8, 0x2514, 0x21F0, 0x5000, 0x8FFE}},
    {37002, 2, {0x80, 0x47F8, 0x2548, 0x24A8, 0x7F8, 0x40, 0xEFFC, 0x2A14, 0x2488, 0x2BF4, 0x2100, 0x21F0, 0x2210, 0x2430, 0x5000, 0x8FFE}},
    {37003, 2, {0x248, 0x4490, 0x2248, 0x27F8, 0x528, 0x4C8, 0xE528, 0x27F8, 0x2490, 0x26D0, 0x2490, 0x26D4, 0x248C, 0x26C4, 0x5000, 0x8FFE}},
    {37004, 2, {0x190, 0x461E, 0x222A, 0x2F8A, 0x312, 0x6AA, 0xEA44, 0x20A0, 0x2318, 0x2C46, 0x2250, 0x2160, 0x2250, 0x2548, 0x5080, 0x8FFE}},
    {37005, 2, {0x100, 0x43F0, 0x2510, 0x20E0, 0x71C, 0x3F8, 0xE248, 0x23F8, 0x2248, 0x23F8, 0x2008, 0x2FFE, 0x2248, 0x2150, 0x5248, 0x8FFE}},
    {37006, 2, {0x400, 0x4F8C, 0x3570, 0x2A20, 0x444, 0x1A78, 0xEF90, 0x2024, 0x2F7E, 0x2010, 0x2F54, 0x2992, 0x2F30, 0x5000, 0x8FFE, 0x0}},
    {37007, 2, {0x0, 0x4FFE, 0x2912, 0x2FFE, 0x228, 0x424, 0xE97E, 0x2FC8, 0x227E, 0x2548, 0x2F7E, 0x2148, 0x2AFE, 0x2A40, 0x5000, 0x8FFE}},
    {37008, 2, {0xFBE, 0x4000, 0x2FBE, 0x2CB2, 0xAAA, 0x40, 0xEFFE, 0x2920, 0x2FFC, 0x2924, 0x2FFC, 0x2A28, 0x2BB2, 0x2A1E, 0x5380, 0x8FFE}},
    {37009, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2008, 0x2002, 0x2002, 0x1FFE, 0x0}},
    {37010, 2, {0x0, 0xFF7C, 0x844, 0x848, 0x848, 0x850, 0x848, 0x848, 0x844, 0x844, 0x844, 0x868, 0x850, 0x2840, 0x1040, 0x40}},
    {37011, 2, {0x0, 0x7E7C, 0x244, 0x248, 0x4248, 0x2450, 0x1448, 0x848, 0x844, 0x1444, 0x1444, 0x2268, 0x4250, 0x8040, 0x40, 0x40}},
    {37012, 2, {0x0, 0x7F7C, 0x144, 0x148, 0x148, 0x150, 0x3F48, 0x2148, 0x2044, 0x2044, 0x2044, 0x2168, 0x2650, 0x3840, 0x2040, 0x40}},
    {37013, 2, {0x888, 0x1110, 0x2220, 0x1110, 0x888, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2002, 0x2002, 0x1FFE}},
    {37014, 2, {0x800, 0x83E, 0x822, 0x824, 0x4924, 0x4928, 0x4924, 0x4924, 0x4922, 0x4922, 0x4922, 0x4F34, 0x7928, 0x20, 0x20, 0x20}},
    {37015, 2, {0x0, 0x7F7C, 0x844, 0x848, 0x848, 0x850, 0x848, 0xFFC8, 0x844, 0x844, 0x844, 0x868, 0x850, 0x840, 0x840, 0x840}},
    {37016, 2, {0x0, 0x7F7C, 0x844, 0x848, 0x848, 0x850, 0x848, 0xFFC8, 0x844, 0x844, 0x844, 0x868, 0x850, 0x840, 0x2840, 0x1040}},
    {37017, 2, {0x1000, 0x83E, 0x822, 0x24, 0xFFA4, 0x2028, 0x2024, 0x2024, 0x2022, 0x2022, 0x2022, 0x2034, 0x2028, 0x3F20, 0x20, 0x20}},
    {37018, 2, {0x1000, 0x107C, 0x1044, 0x1048, 0xFE48, 0x2250, 0x2248, 0x2248, 0x4244, 0x2444, 0x1444, 0x868, 0x1450, 0x2240, 0x4240, 0x8040}},
    {37019, 2, {0x0, 0x7C, 0x7F44, 0x848, 0x848, 0x850, 0x848, 0x848, 0x844, 0x844, 0x844, 0xF68, 0xF050, 0x4040, 0x40, 0x40}},
    {37020, 2, {0x1000, 0x107C, 0x1044, 0x1F48, 0x2148, 0x2150, 0x5148, 0x8A48, 0xA44, 0x444, 0x444, 0x868, 0x1050, 0x2040, 0x4040, 0x8040}},
    {37021, 2, {0x800, 0x43E, 0x422, 0x3FA4, 0x2024, 0x2028, 0x2024, 0x2024, 0x2022, 0x2022, 0x2022, 0x2034, 0x2028, 0x4020, 0x4020, 0x8020}},
    {37022, 2, {0x800, 0x87C, 0x844, 0x7F48, 0x848, 0x850, 0x848, 0xFFC8, 0x844, 0x1444, 0x1244, 0x2268, 0x2150, 0x4140, 0x8040, 0x40}},
    {37023, 2, {0x1000, 0x87C, 0x844, 0xFF48, 0x48, 0x50, 0x3C48, 0x2448, 0x2444, 0x2444, 0x2444, 0x2568, 0x4650, 0x4440, 0x8040, 0x40}},
    {37024, 2, {0x0, 0x147C, 0x1244, 0x1248, 0x2148, 0x2150, 0x4048, 0xBE48, 0x1244, 0x1244, 0x1244, 0x1268, 0x2250, 0x2240, 0x4A40, 0x8440}},
    {37025, 2, {0x1000, 0x83E, 0x822, 0xFFA4, 0x1024, 0x1028, 0x1F24, 0x1124, 0x1122, 0x1122, 0x1122, 0x1134, 0x2128, 0x2520, 0x4220, 0x8020}},
    {37026, 2, {0x0, 0x7FBE, 0x1222, 0x1224, 0x1224, 0x1228, 0x1224, 0xFFE4, 0x1222, 0x1222, 0x1222, 0x1234, 0x2228, 0x2220, 0x4220, 0x8220}},
    {37027, 2, {0x0, 0x7E7C, 0x1244, 0x1248, 0x1248, 0x7E50, 0x1248, 0x1248, 0x7E44, 0x1244, 0x1244, 0x2268, 0x2250, 0x4A40, 0x8440, 0x40}},
    {37028, 2, {0x200, 0x77C, 0x3844, 0x2048, 0x2048, 0x2050, 0x3FC8, 0x2448, 0x2444, 0x2444, 0x2444, 0x2468, 0x2450, 0x4440, 0x4440, 0x8440}},
    {37029, 2, {0x800, 0x87C, 0x844, 0x7F48, 0x4948, 0x850, 0x848, 0x1448, 0x1444, 0x1444, 0x1444, 0x2568, 0x2650, 0x4440, 0x8040, 0x40}},
    {37030, 2, {0x800, 0x87C, 0x844, 0xFF48, 0x848, 0x850, 0x7E48, 0x848, 0x844, 0xFF44, 0x844, 0x1068, 0x1050, 0x2040, 0x4040, 0x8040}},
    {37031, 2, {0x0, 0x3F3E, 0x22, 0x24, 0x24, 0xFFA8, 0x1224, 0x1224, 0x1222, 0x1222, 0x1222, 0x22B4, 0x2328, 0x4220, 0x8020, 0x20}},
    {37032, 2, {0x800, 0x83E, 0x822, 0xFFA4, 0x824, 0x4928, 0x4924, 0x4924, 0x4922, 0x7F22, 0x822, 0x834, 0x928, 0xA20, 0xC20, 0x820}},
    {37033, 2, {0x800, 0x87C, 0x844, 0x948, 0x4948, 0x4A50, 0x4C48, 0x8848, 0x844, 0x844, 0x1444, 0x1268, 0x2150, 0x2140, 0x4040, 0x8040}},
    {37034, 2, {0x0, 0x7FBE, 0x222, 0x2224, 0x2224, 0x2228, 0x4224, 0x7FE4, 0x622, 0xA22, 0x1222, 0x2234, 0x4228, 0x8220, 0xA20, 0x420}},
    {37035, 2, {0x1000, 0x10F8, 0x1088, 0xFC88, 0x10F8, 0x1000, 0x7DFC, 0x1124, 0x1124, 0xFF24, 0x11FC, 0x1100, 0x2100, 0x2102, 0x4102, 0x80FE}},
    {37036, 2, {0x1000, 0x203E, 0x7E22, 0x4224, 0x4224, 0x4A28, 0x4424, 0x4024, 0x7F22, 0x122, 0x122, 0x1D34, 0xE128, 0x4120, 0xA20, 0x420}},
    {37037, 2, {0x1000, 0x103E, 0x2022, 0x3F24, 0x4124, 0x8128, 0x3D24, 0x2524, 0x2522, 0x2522, 0x3D22, 0x2534, 0x128, 0x120, 0xA20, 0x420}},
    {37038, 2, {0x800, 0x87C, 0x844, 0x7F48, 0x4948, 0x4950, 0x4948, 0x4948, 0x7F44, 0x4944, 0x4944, 0x4968, 0x4950, 0x7F40, 0x4140, 0x40}},
    {37039, 2, {0x2200, 0x223E, 0x2222, 0x2224, 0xFFA4, 0x2228, 0x2224, 0x2224, 0x3E22, 0x2222, 0x2222, 0x2234, 0x2228, 0x3E20, 0x2220, 0x20}},
    {37040, 2, {0x800, 0x83E, 0x1022, 0x2224, 0x4124, 0xFFA8, 0xA4, 0x24, 0x7F22, 0x4122, 0x4122, 0x4134, 0x4128, 0x4120, 0x7F20, 0x4120}},
    {37041, 2, {0x0, 0x33E, 0x3C22, 0x2024, 0x2024, 0x2028, 0x3FA4, 0x2224, 0x2222, 0x2222, 0x2222, 0x2234, 0x27A8, 0xF820, 0x4020, 0x20}},
    {37042, 2, {0x1000, 0x83E, 0xA22, 0x224, 0x1224, 0x1328, 0x54A4, 0x54A4, 0x5422, 0x9822, 0x1922, 0x1134, 0x3128, 0x4F20, 0x8020, 0x20}},
    {37043, 2, {0x0, 0xFF7C, 0x844, 0x848, 0x1048, 0x1450, 0x3248, 0x5148, 0x9044, 0x1044, 0x1044, 0x68, 0xE50, 0xF040, 0x4040, 0x40}},
    {37044, 2, {0x0, 0x7FBE, 0x822, 0x824, 0x824, 0x7FA8, 0x48A4, 0x48A4, 0x54A2, 0x52A2, 0x62A2, 0x40B4, 0x40A8, 0x40A0, 0x42A0, 0x4120}},
    {37045, 2, {0x0, 0x7F3E, 0x1122, 0x1124, 0x2124, 0x2528, 0x4224, 0x8024, 0x3F22, 0x2122, 0x2122, 0x2134, 0x2128, 0x3F20, 0x2120, 0x20}},
    {37046, 2, {0x1200, 0x123E, 0x1222, 0x1224, 0x1224, 0xF3A8, 0x1224, 0x1224, 0x1222, 0x1222, 0x1222, 0x3234, 0x52A8, 0x9320, 0x1220, 0x20}},
    {37047, 2, {0x0, 0xFF7C, 0x2044, 0x2048, 0x3C48, 0x2450, 0x2448, 0x3448, 0x4C44, 0x4C44, 0x4444, 0x4568, 0x5550, 0x6540, 0x4340, 0x40}},
    {37048, 2, {0x100, 0x7BE, 0x7C22, 0x4424, 0x4424, 0x4428, 0x4424, 0x7FA4, 0x4422, 0x4422, 0x4422, 0x4234, 0x42A8, 0x52A0, 0x69A0, 0x44A0}},
    {37049, 2, {0x2000, 0x207C, 0x3E44, 0x4248, 0x8448, 0x850, 0xFE48, 0x248, 0x244, 0x7E44, 0x244, 0x268, 0x250, 0xFE40, 0x240, 0x40}},
    {37050, 2, {0x1400, 0x143E, 0x1422, 0x14A4, 0x94A4, 0x5528, 0x5524, 0x5624, 0x1422, 0x1422, 0x1422, 0x17B4, 0xF828, 0x4020, 0x20, 0x20}},
    {37051, 2, {0x800, 0x87C, 0x1444, 0x1248, 0x2148, 0x48D0, 0x8448, 0x448, 0x7F44, 0x144, 0x2244, 0x1468, 0x850, 0x440, 0x440, 0x40}},
    {37052, 2, {0x0, 0x7FBE, 0x4022, 0x4024, 0x7FA4, 0x4428, 0x4424, 0x5F24, 0x4422, 0x4422, 0x7FA2, 0x4034, 0x4028, 0x7FA0, 0x20, 0x20}},
    {37053, 2, {0x800, 0x87C, 0x7F44, 0x848, 0x848, 0x850, 0xFFC8, 0x48, 0x844, 0x844, 0x7F44, 0x868, 0x850, 0xF40, 0xF040, 0x4040}},
    {37054, 2, {0x800, 0x487C, 0x4844, 0x7F48, 0x4848, 0x8850, 0x848, 0xFFC8, 0x1844, 0x1C44, 0x2A44, 0x2968, 0x4950, 0x8840, 0x840, 0x840}},
    {37055, 2, {0x800, 0x83E, 0x822, 0x7F24, 0x824, 0x828, 0xFFA4, 0x224, 0x222, 0x7FA2, 0x222, 0x2234, 0x1228, 0x220, 0xA20, 0x420}},
    {37056, 2, {0x800, 0x83E, 0xFFA2, 0x1424, 0x2224, 0x4128, 0xBEA4, 0x24, 0xFFA2, 0x1022, 0x2022, 0x7F34, 0x128, 0x120, 0xA20, 0x420}},
    {37057, 2, {0x800, 0x83E, 0xFF22, 0x1024, 0x1024, 0x3F28, 0x2124, 0x6124, 0xBF22, 0x2122, 0x2122, 0x3F34, 0x2128, 0x2120, 0x2520, 0x2220}},
    {37058, 2, {0x1000, 0x83E, 0x22, 0xFFA4, 0x824, 0x1128, 0x2124, 0x7E24, 0x422, 0x922, 0x1222, 0x2434, 0xCC28, 0x1220, 0x2120, 0xC0A0}},
    {37059, 2, {0x800, 0x87C, 0x1444, 0x2248, 0x4148, 0xBED0, 0x48, 0x48, 0x3E44, 0x2244, 0x2244, 0x2268, 0x2250, 0x3E40, 0x2240, 0x40}},
    {37060, 2, {0x100, 0x323E, 0xC22, 0x1224, 0x6924, 0x828, 0xFFA4, 0x1024, 0x2422, 0x2422, 0x4822, 0x8A34, 0x1128, 0x3F20, 0x1120, 0x20}},
    {37061, 2, {0x0, 0xFFBE, 0x822, 0x1024, 0x2224, 0x4128, 0xFFA4, 0x8A4, 0x822, 0x822, 0x7F22, 0x834, 0x828, 0xFA0, 0xF820, 0x4020}},
    {37062, 2, {0x800, 0x83E, 0x822, 0xFFA4, 0x824, 0x828, 0x7F24, 0x24, 0x22, 0x7F22, 0x4122, 0x4134, 0x4128, 0x7F20, 0x4120, 0x20}},
    {37063, 2, {0x2000, 0x203E, 0x7F22, 0x4124, 0x8124, 0x7928, 0x4924, 0x4924, 0x7922, 0x4922, 0x4922, 0x7934, 0x4128, 0x120, 0xA20, 0x420}},
    {37064, 2, {0x200, 0x73E, 0x7822, 0x4024, 0x4024, 0x7FA8, 0x4024, 0x4024, 0x5F22, 0x5122, 0x5122, 0x5134, 0x5128, 0x9F20, 0x9120, 0x20}},
    {37065, 2, {0x0, 0x773E, 0x2222, 0x2224, 0x2224, 0x2228, 0x2224, 0xF7A4, 0x2222, 0x2222, 0x2222, 0x2234, 0x2228, 0x4220, 0x4220, 0x8220}},
    {37066, 2, {0x1000, 0x87C, 0x44, 0xFF48, 0x48, 0x2450, 0x4248, 0x8148, 0x2444, 0x1444, 0x844, 0x1468, 0x2250, 0x4240, 0x8040, 0x40}},
    {37067, 2, {0x800, 0x83E, 0x1022, 0x7F24, 0x4124, 0x4128, 0x7F24, 0x4124, 0x4122, 0x7F22, 0x4122, 0x4134, 0x4128, 0x7F20, 0x4120, 0x20}},
    {37068, 2, {0x1000, 0x10F8, 0x1088, 0x7C88, 0x10F8, 0x1000, 0xFDFC, 0x124, 0x1124, 0x1124, 0x7DFC, 0x1100, 0x1100, 0x1D02, 0xE102, 0x40FE}},
    {37069, 2, {0x800, 0x103E, 0x3F22, 0x2124, 0x2924, 0x2528, 0x2524, 0xFF24, 0x2122, 0x2922, 0x2522, 0x2534, 0x2128, 0x4120, 0x4520, 0x8220}},
    {37070, 2, {0x800, 0x43E, 0x3F22, 0x2124, 0x2124, 0x3F28, 0x2124, 0x2124, 0x3F22, 0x2022, 0x2422, 0x2234, 0x2528, 0x2920, 0x3020, 0x2020}},
    {37071, 2, {0x800, 0x83E, 0x822, 0xFFA4, 0x824, 0x4928, 0x2924, 0x2A24, 0xFFA2, 0x822, 0x822, 0x1434, 0x1228, 0x2120, 0x4120, 0x8020}},
    {37072, 2, {0x800, 0x83E, 0x1422, 0x2224, 0x4124, 0xBEA8, 0x24, 0x24, 0xFFA2, 0x1022, 0x1022, 0x2234, 0x4128, 0xFF20, 0x4120, 0x20}},
    {37073, 2, {0x2200, 0x123E, 0x1422, 0x7F24, 0x824, 0x828, 0x824, 0xFFA4, 0x822, 0x822, 0x1422, 0x1234, 0x2128, 0x4120, 0x8020, 0x20}},
    {37074, 2, {0x0, 0x20F8, 0x1088, 0x7C88, 0x44F8, 0x4400, 0x7DFC, 0x4524, 0x4524, 0x7D24, 0x41FC, 0x5100, 0x4900, 0x5502, 0x6502, 0x40FE}},
    {37075, 2, {0x0, 0x7FBE, 0x48A2, 0x824, 0x7FA4, 0x1028, 0x1424, 0x2424, 0x7FA2, 0x422, 0x422, 0x7B4, 0xFC28, 0x4420, 0x420, 0x420}},
    {37076, 2, {0x80, 0x1DE, 0x7712, 0x1112, 0x1114, 0x2514, 0x25D8, 0x7514, 0x1512, 0x5512, 0x5512, 0x27DA, 0x2014, 0x5010, 0x8FF0, 0x10}},
    {37077, 2, {0x400, 0x53E, 0x4A2, 0x424, 0x7FA4, 0x4428, 0x4424, 0x7524, 0x5522, 0x5522, 0x5522, 0x5234, 0x52A8, 0xA6A0, 0x89A0, 0x10A0}},
    {37078, 2, {0x0, 0x7FBE, 0x22, 0x24, 0x3F24, 0x2128, 0x2124, 0x2124, 0x3F22, 0x22, 0x2122, 0x1134, 0x1228, 0x7A0, 0x7820, 0x2020}},
    {37079, 2, {0x3080, 0xD3E, 0x622, 0x1924, 0x60A4, 0x828, 0x7FE4, 0x1424, 0x2422, 0x7FA2, 0xA4A2, 0x24B4, 0x26A8, 0x2520, 0x420, 0x420}},
    {37080, 2, {0x0, 0x3E3E, 0x2222, 0x2224, 0x2224, 0x2228, 0x3E24, 0x24, 0x22, 0x7F22, 0x4122, 0x4134, 0x4128, 0x4120, 0x7F20, 0x4120}},
    {37081, 2, {0xA00, 0x93E, 0xFFA2, 0x824, 0x824, 0x7F28, 0x4924, 0x4924, 0x7F22, 0x4922, 0x4922, 0x7F34, 0x4928, 0x4920, 0x4920, 0x4320}},
    {37082, 2, {0x0, 0xFFBE, 0x822, 0x824, 0x7F24, 0x1128, 0x1124, 0xFFE4, 0x22, 0x22, 0x7F22, 0x4134, 0x4128, 0x4120, 0x7F20, 0x4120}},
    {37083, 2, {0x200, 0xF3E, 0xF022, 0x124, 0x9124, 0x4A28, 0x24, 0xFE24, 0x422, 0x822, 0xF22, 0xF834, 0x4828, 0x820, 0x2820, 0x1020}},
    {37084, 2, {0x800, 0x483E, 0x4822, 0x7F24, 0x4824, 0x8828, 0x824, 0xFFA4, 0x22, 0x22, 0x7F22, 0x4134, 0x4128, 0x4120, 0x7F20, 0x4120}},
    {37085, 2, {0x800, 0x83E, 0x7F22, 0x824, 0x824, 0x828, 0xFFA4, 0x1224, 0x1222, 0x5322, 0x52A2, 0x92B4, 0x2228, 0x2220, 0x4A20, 0x8420}},
    {37086, 2, {0x1000, 0x87C, 0x7E44, 0x4248, 0x4248, 0x7E50, 0x4248, 0x4248, 0x7E44, 0x5144, 0x4A44, 0x4468, 0x4450, 0x5240, 0x6240, 0x4040}},
    {37087, 2, {0x800, 0x83E, 0x822, 0xFFA4, 0x824, 0x2A28, 0x2A24, 0x2A24, 0x5D22, 0x88A2, 0x822, 0x1434, 0x1228, 0x2120, 0x4120, 0x8020}},
    {37088, 2, {0x0, 0x7FBE, 0x822, 0x824, 0x7F24, 0x4928, 0x4924, 0x7F24, 0x4922, 0x4922, 0x7F22, 0x834, 0x6828, 0x1820, 0x2620, 0xC120}},
    {37089, 2, {0x0, 0x7F3E, 0x1122, 0x1124, 0xFFE4, 0x1128, 0x1124, 0x7F24, 0x2022, 0x2022, 0x7F22, 0xA134, 0x2128, 0x2120, 0x3F20, 0x2120}},
    {37090, 2, {0x0, 0x7F7C, 0x4144, 0x4148, 0x4148, 0x7F50, 0x48, 0x7F48, 0x844, 0x844, 0x7F44, 0x868, 0x850, 0xF40, 0xF040, 0x4040}},
    {37091, 2, {0x800, 0x87C, 0x7F44, 0x848, 0x848, 0x7F50, 0x4148, 0x8248, 0x3C44, 0x444, 0x844, 0xF68, 0x7850, 0x840, 0x2840, 0x1040}},
    {37092, 2, {0x1400, 0x127C, 0x2144, 0x4948, 0x848, 0x1450, 0x2248, 0x4148, 0xBE44, 0x2244, 0x2244, 0x2268, 0x2250, 0x3E40, 0x2240, 0x40}},
    {37093, 2, {0x0, 0x7E7C, 0x4244, 0x4248, 0x7E48, 0x4250, 0x4248, 0x7E48, 0x4244, 0x4244, 0x7E44, 0x68, 0x2450, 0x2240, 0x4140, 0x8140}},
    {37094, 2, {0x0, 0x7FDE, 0x12, 0x12, 0x7BD4, 0x4A54, 0x4A58, 0x4A54, 0x6B52, 0x5AD2, 0x4A52, 0x4A5A, 0x4A54, 0x4A50, 0x4A50, 0x5AD0}},
    {37095, 2, {0x0, 0x3E3E, 0x2222, 0x2224, 0x3E24, 0x28, 0x7F24, 0x4124, 0x4922, 0x4922, 0x4922, 0x4934, 0x1428, 0x1220, 0x2120, 0xC120}},
    {37096, 2, {0x1000, 0x83E, 0x7FA2, 0x24, 0x2124, 0x1228, 0xFFE4, 0x24, 0x22, 0x3F22, 0x2122, 0x2134, 0x2128, 0x3F20, 0x2120, 0x20}},
    {37097, 2, {0x6100, 0x1A3E, 0xC22, 0x3224, 0xC924, 0x828, 0xFFA4, 0x1024, 0x3F22, 0x6122, 0xBF22, 0x2134, 0x3F28, 0x2120, 0x2520, 0x2220}},
    {37098, 2, {0x800, 0x83E, 0xFFA2, 0x824, 0x7F24, 0x928, 0xFFA4, 0x924, 0x7F22, 0x1022, 0xFF22, 0x2134, 0x7228, 0xC20, 0x3320, 0xC0A0}},
    {37099, 2, {0x1000, 0x203E, 0x7F22, 0x4924, 0x4924, 0x7F28, 0x4924, 0x5124, 0x7F22, 0x2422, 0x4422, 0xFFB4, 0x428, 0x420, 0x420, 0x420}},
    {37100, 2, {0x800, 0x83E, 0xFFA2, 0x824, 0x7F24, 0x828, 0xFFA4, 0x24, 0x7F22, 0x4122, 0x7F22, 0x4134, 0x7F28, 0x4120, 0x4520, 0x4220}},
    {37101, 2, {0x1000, 0x87C, 0xFF44, 0x48, 0x7E48, 0x4250, 0x7E48, 0x48, 0x7E44, 0x444, 0x844, 0xF68, 0xF850, 0x840, 0x2840, 0x1040}},
    {37102, 2, {0x0, 0x3FBE, 0x24A2, 0x24A4, 0x2EA4, 0x24A8, 0x3FA4, 0x20A4, 0x2EA2, 0x2AA2, 0x2AA2, 0x2EB4, 0x20A8, 0x40A0, 0x42A0, 0x8120}},
    {37103, 2, {0x800, 0x93E, 0x4922, 0x4A24, 0x8824, 0x1428, 0x2224, 0x4924, 0x8822, 0x4922, 0x4922, 0x8A34, 0x1428, 0x2220, 0x4120, 0x80A0}},
    {37104, 2, {0x0, 0xFC1E, 0x4812, 0x4BD2, 0x7854, 0x4A54, 0x4958, 0x7954, 0x4892, 0x4892, 0x5C92, 0xE95A, 0x4954, 0xA50, 0x810, 0x810}},
    {37105, 2, {0x2100, 0x113E, 0x1222, 0x24, 0x7FA4, 0x1228, 0x1224, 0x1224, 0x1222, 0xFFA2, 0x1222, 0x1234, 0x1228, 0x1220, 0x2220, 0x4220}},
    {37106, 2, {0x800, 0x87C, 0x844, 0xFF48, 0x848, 0x2A50, 0x2A48, 0x2A48, 0x5D44, 0x8844, 0x1C44, 0x2A68, 0x4950, 0x8840, 0x840, 0x840}},
    {37107, 2, {0x1000, 0x677C, 0x4144, 0x4148, 0x7748, 0x4150, 0x4148, 0x7F48, 0x1444, 0x1444, 0x1444, 0x1468, 0x2550, 0x2640, 0x4440, 0x8040}},
    {37108, 2, {0x1100, 0x111E, 0x1112, 0x1112, 0x7BD4, 0x1114, 0x1118, 0x3314, 0x3B92, 0x5552, 0x5512, 0x991A, 0x1114, 0x1110, 0x1110, 0x1110}},
    {37109, 2, {0x200, 0xF3E, 0x7822, 0x824, 0xFFA4, 0x2A28, 0x2A24, 0xFFA4, 0x2A22, 0x2A22, 0xFFA2, 0x834, 0x828, 0xF20, 0xF020, 0x4020}},
    {37110, 2, {0x2000, 0x10F8, 0x7C88, 0x88, 0x44F8, 0x2800, 0x1FC, 0xFF24, 0x124, 0x124, 0x7DFC, 0x4500, 0x4500, 0x4502, 0x7D02, 0x44FE}},
    {37111, 2, {0x2000, 0x23DE, 0x2252, 0x4A52, 0x4BD4, 0xF254, 0x1258, 0x23D4, 0x4A12, 0xFA92, 0xA52, 0x12DA, 0x1354, 0x2210, 0x4010, 0x8010}},
    {37112, 2, {0x4100, 0x223E, 0x1422, 0x7F24, 0x4924, 0x4928, 0x7F24, 0x4924, 0x4922, 0x7F22, 0x822, 0x834, 0xFFA8, 0x820, 0x820, 0x820}},
    {37113, 2, {0x0, 0x7E7C, 0x4244, 0x7E48, 0x4248, 0x7E50, 0x4248, 0x7E48, 0x1444, 0x1244, 0xFF44, 0x1068, 0x2850, 0x2440, 0x4240, 0x8040}},
    {37114, 2, {0x2124, 0x2248, 0x3D24, 0x4400, 0xA8F8, 0x1088, 0x20F8, 0xC800, 0x11FC, 0x3D24, 0x4524, 0x85FC, 0x2900, 0x1104, 0x2104, 0xC0FC}},
    {37115, 2, {0x3F00, 0xA13E, 0xBF22, 0xA124, 0xBF24, 0xA128, 0xBF24, 0x8024, 0xFFA2, 0x22, 0x4922, 0x4934, 0x9228, 0x9220, 0x4920, 0x4920}},
    {37116, 2, {0x1000, 0x103E, 0x7F22, 0x2124, 0xFFE4, 0x28, 0x7F24, 0x4124, 0x7F22, 0x822, 0xFFA2, 0x834, 0x8828, 0xFFE0, 0x820, 0x820}},
    {37117, 2, {0x800, 0x8BE, 0x7EA2, 0x924, 0xA24, 0xFFE8, 0x824, 0x1024, 0x3F22, 0x6122, 0xA122, 0x3F34, 0x2128, 0x2120, 0x3F20, 0x2120}},
    {37118, 2, {0x0, 0x7FDE, 0x4012, 0x5F92, 0x5094, 0x5F94, 0x5098, 0x5F94, 0x4412, 0x7FD2, 0x4912, 0x591A, 0x4614, 0x4910, 0x5090, 0x7FD0}},
    {37119, 2, {0x0, 0x7F7C, 0x4944, 0x4948, 0x7F48, 0x4050, 0x5F48, 0x5148, 0x5144, 0x5F44, 0x5144, 0x5F68, 0x5150, 0x5140, 0x9F40, 0x1140}},
    {37120, 2, {0x2200, 0x223E, 0xFFA2, 0x2224, 0x2A24, 0x828, 0xFFA4, 0x1024, 0x2022, 0x3F22, 0x6122, 0xA134, 0x2128, 0x2120, 0x3F20, 0x2120}},
    {37121, 2, {0x1400, 0x14BE, 0xF722, 0x1424, 0x34A4, 0xD3A8, 0x24, 0x7F24, 0x4122, 0x7F22, 0x4122, 0x7F34, 0x4128, 0x4120, 0x4520, 0x4220}},
    {37122, 2, {0x0, 0x773E, 0x5522, 0x5524, 0x7724, 0x28, 0x7F24, 0x24, 0xFFA2, 0x2022, 0x3F22, 0x134, 0x128, 0x120, 0xA20, 0x420}},
    {37123, 2, {0x800, 0x143E, 0x2222, 0x4124, 0xBEA4, 0x28, 0x78A4, 0x4AA4, 0x4AA2, 0x7AA2, 0x4AA2, 0x4AB4, 0x7AA8, 0x48A0, 0x4AA0, 0x5920}},
    {37124, 2, {0x0, 0xFF7C, 0x1444, 0x1448, 0x7F48, 0x5550, 0x5548, 0x7F48, 0x844, 0x844, 0x7F44, 0x868, 0x850, 0xF40, 0xF040, 0x4040}},
    {37125, 2, {0x100, 0x7BE, 0xF822, 0x824, 0x7F24, 0x4928, 0x4924, 0x7F24, 0x822, 0xFFA2, 0x88A2, 0x8AB4, 0xBEA8, 0x80A0, 0x82A0, 0x8120}},
    {37126, 2, {0x0, 0xFFBE, 0x88A2, 0x824, 0xFFA4, 0x828, 0x7F24, 0x4924, 0x7F22, 0x4922, 0x7F22, 0x834, 0xFFA8, 0x820, 0x820, 0x820}},
    {37127, 2, {0x1000, 0x179E, 0x1092, 0x2092, 0x2FF4, 0x6414, 0xA418, 0x27D4, 0x2912, 0x2112, 0x2FF2, 0x211A, 0x2294, 0x2250, 0x2450, 0x2810}},
    {37128, 2, {0xA00, 0x747C, 0x1544, 0x5248, 0x2248, 0x2150, 0x4148, 0xBE48, 0x844, 0x844, 0x7F44, 0x868, 0x1450, 0x1240, 0x2240, 0x4040}},
    {37129, 2, {0x2100, 0x209E, 0x23D2, 0x4A52, 0x4A54, 0xF3D4, 0x1258, 0x2254, 0x4BD2, 0xFA12, 0xA92, 0x125A, 0x12D4, 0x2350, 0x4210, 0x8010}},
    {37130, 2, {0x2000, 0x27FE, 0x2252, 0x4A52, 0x4BD4, 0xF254, 0x1258, 0x23D4, 0x4A52, 0xFA52, 0xA72, 0x17DA, 0x1054, 0x2050, 0x4050, 0x8050}},
    {37131, 2, {0x2800, 0x4B7C, 0x4944, 0x4948, 0x6B48, 0x4950, 0x4948, 0x7F48, 0x844, 0x7F44, 0x144, 0x3268, 0xC50, 0x1240, 0x6140, 0x40}},
    {37132, 2, {0x800, 0x43E, 0x7FA2, 0x4424, 0x5F24, 0x4528, 0x7FA4, 0x4524, 0x5F22, 0x4422, 0x5F22, 0x5134, 0x5128, 0x5120, 0x5F20, 0x9120}},
    {37133, 2, {0x0, 0xFFBE, 0x80A2, 0x24, 0x7F24, 0x4128, 0x7F24, 0x4124, 0x7F22, 0x1022, 0x822, 0xFFB4, 0x28, 0x2220, 0x4120, 0x80A0}},
    {37134, 2, {0x800, 0x103E, 0x7F22, 0x4124, 0x7F24, 0x4128, 0x7F24, 0x4124, 0x7F22, 0x1022, 0x822, 0x2934, 0xA2A8, 0xA2A0, 0x1E20, 0x20}},
    {37135, 2, {0x0, 0x7FBE, 0x4022, 0x5F24, 0x4024, 0x7FA8, 0x5524, 0x5224, 0x5922, 0x50A2, 0x4222, 0x7FB4, 0x5228, 0x4A20, 0x8220, 0x620}},
    {37136, 2, {0x1000, 0x83E, 0xFFA2, 0x1024, 0x2224, 0x7C28, 0x824, 0x1124, 0x7FA2, 0xA2, 0x7F22, 0x4934, 0x7F28, 0x4920, 0x7F20, 0x4120}},
    {37137, 2, {0x0, 0x7FBE, 0x1222, 0x52A4, 0x3324, 0x1228, 0xFFE4, 0x24, 0x3F22, 0x2122, 0x2122, 0x3F34, 0x2128, 0x2120, 0x3F20, 0x2120}},
    {37138, 2, {0x1000, 0x103E, 0x3FA2, 0x48A4, 0x2AA4, 0x3EA8, 0x10A4, 0x2524, 0x5222, 0x3FA2, 0x48A2, 0xAAB4, 0x3EA8, 0x10A0, 0x2520, 0x4220}},
    {37139, 2, {0x800, 0x103E, 0x7F22, 0x4124, 0x7F24, 0x4128, 0x7F24, 0x4124, 0x7F22, 0xA22, 0x922, 0xFFB4, 0x828, 0x1420, 0x2220, 0x4120}},
    {37140, 2, {0x1000, 0x203E, 0x7F22, 0x4124, 0x4124, 0x7F28, 0x4024, 0x7FE4, 0x4022, 0x7FA2, 0xA2, 0xAAB4, 0xAAA8, 0xA0, 0x520, 0x220}},
    {37141, 2, {0x2080, 0x211E, 0x23D2, 0x4A52, 0x4A54, 0xF3D4, 0x1258, 0x2254, 0x4BD2, 0xF812, 0xA52, 0x139A, 0x1214, 0x2250, 0x41D0, 0x8010}},
    {37142, 2, {0x0, 0x3F3E, 0x2122, 0x2124, 0x3F24, 0x28, 0x7FA4, 0x40A4, 0x7FA2, 0x40A2, 0x7FA2, 0x40B4, 0x7FA8, 0x1220, 0x2120, 0x40A0}},
    {37143, 2, {0x1000, 0x87C, 0xFF44, 0x48, 0x7E48, 0x4250, 0x7E48, 0x48, 0xFF44, 0x8144, 0xBD44, 0xA568, 0xBD50, 0x8140, 0x8540, 0x8240}},
    {37144, 2, {0x200, 0x3FDE, 0x2212, 0x2F92, 0x2294, 0x3FD4, 0x2298, 0x2F94, 0x2212, 0x2F92, 0x2A92, 0x2F9A, 0x4A94, 0x4F90, 0x8A90, 0x990}},
    {37145, 2, {0x1F00, 0x113E, 0x1122, 0x1F24, 0x424, 0x7FE8, 0x424, 0x3FA4, 0x20A2, 0x2EA2, 0x2AA2, 0x2EB4, 0x20A8, 0x3FA0, 0x20A0, 0x20}},
    {37146, 2, {0x2200, 0x223E, 0xFFA2, 0x2224, 0x7F24, 0x4128, 0x7F24, 0x4124, 0x7F22, 0x822, 0xFFA2, 0x834, 0x1428, 0x2220, 0x4120, 0x8120}},
    {37147, 2, {0x4900, 0x927C, 0x4944, 0x48, 0x7F48, 0x4950, 0x7F48, 0x4948, 0x7F44, 0x844, 0xFFC4, 0x1C68, 0x2A50, 0x4940, 0x8840, 0x840}},
    {37148, 2, {0x800, 0x43E, 0x7FA2, 0x5224, 0x5224, 0x7FA8, 0x52A4, 0x52A4, 0x7FA2, 0x5222, 0x5222, 0x5BB4, 0x5228, 0x52A0, 0x9B20, 0x1220}},
    {37149, 2, {0x0, 0x773E, 0x1122, 0x5524, 0x3324, 0x5528, 0xC24, 0x3324, 0xC4A2, 0x1822, 0x6222, 0xCB4, 0x7128, 0x620, 0x1820, 0xE020}},
    {37150, 2, {0x2200, 0x227C, 0xFFC4, 0x2248, 0x3E48, 0x850, 0x7F48, 0x4948, 0x7F44, 0x844, 0x7F44, 0x868, 0x7F50, 0x840, 0xF40, 0xF040}},
    {37151, 2, {0x800, 0xFFBE, 0x822, 0x7F24, 0x4924, 0x7F28, 0x4924, 0x7F24, 0x8A2, 0xFFA2, 0x222, 0xFFB4, 0x4228, 0x2220, 0xA20, 0x420}},
    {37152, 2, {0x7F00, 0x83E, 0xFFA2, 0x88A4, 0x6B24, 0x828, 0x6B24, 0x24, 0x7F22, 0x22, 0xFFA2, 0x2034, 0x3F28, 0x120, 0xA20, 0x420}},
    {37153, 2, {0x1000, 0x7F3E, 0x4122, 0x7F24, 0x4124, 0x7F28, 0x4024, 0x7FA4, 0x4022, 0x7FA2, 0x8A2, 0x7EB4, 0x1DA8, 0x2A20, 0xC920, 0x820}},
    {37154, 2, {0x0, 0x7FBE, 0x422, 0x2724, 0x2424, 0xFFA8, 0x2024, 0x3FA4, 0x4022, 0x7FA2, 0xA2, 0xAAB4, 0xAAA8, 0xA0, 0x520, 0x220}},
    {37155, 2, {0x800, 0x7F3E, 0x2222, 0x1424, 0xFFA4, 0x28, 0x7F24, 0x4124, 0x7F22, 0x4122, 0x7F22, 0x834, 0xFFA8, 0x820, 0x820, 0x820}},
    {37156, 2, {0x7E00, 0x427C, 0x7E44, 0x4248, 0x7E48, 0x50, 0xFF48, 0xA548, 0xFF44, 0x44, 0x7E44, 0x268, 0x3450, 0x840, 0x1440, 0x6240}},
    {37157, 2, {0x1000, 0x203E, 0x7F22, 0x4124, 0x7F24, 0x4128, 0x7F24, 0x4024, 0x7FA2, 0x4022, 0x7FA2, 0xB4, 0xAAA8, 0xAAA0, 0x80A0, 0x320}},
    {37158, 2, {0x2000, 0x203E, 0x7FA2, 0xD524, 0x5524, 0x5528, 0xFFA4, 0x5524, 0x5522, 0x5522, 0xFFA2, 0x34, 0x6A28, 0x5520, 0x9520, 0x20}},
    {37159, 2, {0x500, 0x723E, 0x12A2, 0x5124, 0x2124, 0x5EA8, 0x80A4, 0x3F24, 0x2122, 0x2122, 0x3F22, 0x2134, 0x1228, 0x7A0, 0x7820, 0x2020}},
    {37160, 2, {0x4920, 0x2A3E, 0x7F44, 0x49A8, 0x5D10, 0x6B28, 0x4946, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2002, 0x1FFE}},
    {37161, 2, {0x0, 0x7F3E, 0x122, 0x3F24, 0x124, 0x7F28, 0x24, 0x7724, 0x2522, 0xF722, 0x222, 0xFFB4, 0x4228, 0x2220, 0x2A20, 0x420}},
    {37162, 2, {0x1200, 0xFF9E, 0x1292, 0x7F92, 0x5214, 0x7FD4, 0x2258, 0x7F54, 0xA1D2, 0x3F12, 0x2112, 0x3F1A, 0x2114, 0x3F10, 0x1210, 0x2110}},
    {37163, 2, {0x4100, 0x223E, 0x22, 0xFFA4, 0x88A4, 0xAAA8, 0x9CA4, 0x88A4, 0xFFA2, 0x22, 0x7F22, 0x4134, 0x7F28, 0x4120, 0x7F20, 0x4120}},
    {37164, 2, {0x3C0, 0x7C1E, 0x2892, 0x1512, 0x3F14, 0x2114, 0x2118, 0x3F94, 0x2092, 0x2092, 0x3FD2, 0x205A, 0x5554, 0x5550, 0xA150, 0x90}},
    {37165, 2, {0x2200, 0x147C, 0xFF44, 0x1448, 0x7F48, 0x5550, 0x6348, 0x5D48, 0x4144, 0x7F44, 0x844, 0xFF68, 0x1450, 0x2240, 0x4140, 0x8040}},
    {37166, 2, {0x1000, 0xEFBE, 0xA4A2, 0x94A4, 0xAAA4, 0xC928, 0x1024, 0x7F24, 0x4122, 0x7F22, 0x4122, 0x7F34, 0x4128, 0x7F20, 0x2220, 0x4120}},
    {37167, 2, {0x2200, 0x143E, 0xFFA2, 0x824, 0x7F24, 0x828, 0xFFA4, 0x4924, 0x2A22, 0xFFA2, 0x22, 0x7F34, 0x4128, 0x4120, 0x7F20, 0x4120}},
    {37168, 2, {0x800, 0x493E, 0x2A22, 0xFFA4, 0x2A24, 0x4928, 0x88A4, 0x4124, 0x4122, 0x77A2, 0x9122, 0x5534, 0x27A8, 0x4120, 0x8120, 0x120}},
    {37169, 2, {0xF00, 0xF83E, 0x4922, 0x2A24, 0xFFA4, 0x2A28, 0x4924, 0x80A4, 0x7F22, 0x4922, 0x4922, 0x7F34, 0x4928, 0x4920, 0x7F20, 0x4120}},
    {37170, 2, {0x0, 0xF7BE, 0x94A2, 0x94A4, 0xF7A4, 0x28, 0x7F24, 0x4924, 0x7F22, 0x4922, 0x7F22, 0x834, 0xFFA8, 0x820, 0x820, 0x820}},
    {37171, 2, {0x0, 0x7F3E, 0x5522, 0x5524, 0x7724, 0x1428, 0x7724, 0x5524, 0x5522, 0x7722, 0x5522, 0x5534, 0x7728, 0x5420, 0x14A0, 0xFA0}},
    {37172, 2, {0x1400, 0x557C, 0x3644, 0x1448, 0xFF48, 0x2250, 0x1448, 0x7F48, 0x844, 0x7F44, 0x844, 0xFF68, 0x1C50, 0x2A40, 0x4940, 0x840}},
    {37173, 2, {0x0, 0x7F3E, 0x4122, 0x7F24, 0x24, 0xF7A8, 0x94A4, 0xF7A4, 0x822, 0x822, 0xFFA2, 0x1C34, 0x2A28, 0x4920, 0x88A0, 0x820}},
    {37174, 2, {0x400, 0xB1E, 0x30D2, 0xDF32, 0x14, 0x7FD4, 0x5558, 0x4E54, 0x7FD2, 0x12, 0x3F92, 0x209A, 0x3F94, 0x2090, 0x3F90, 0x2090}},
    {37175, 2, {0x1200, 0x7FBE, 0x52A2, 0x7FA4, 0x52A4, 0x7FA8, 0x24, 0x7FA4, 0x22, 0x7FA2, 0x40A2, 0x7FB4, 0x2128, 0x1220, 0xFA0, 0xF020}},
    {37176, 2, {0x2200, 0xFFBE, 0x2222, 0x7F24, 0x5524, 0x5528, 0x7F24, 0x24, 0xFFA2, 0x90A2, 0x3E22, 0x2234, 0x5228, 0xC20, 0x1820, 0xE020}},
    {37177, 2, {0xFC00, 0x4BBE, 0x78A2, 0x4AA4, 0x7924, 0x4D28, 0xFAA4, 0xCA4, 0x7A2, 0xF822, 0x28A2, 0x4D34, 0x1A28, 0x2920, 0xC8A0, 0x820}},
    {37178, 2, {0x800, 0x43E, 0x7FA2, 0x4A24, 0x5F24, 0x4A28, 0x7FA4, 0x4424, 0x5F22, 0x5522, 0x5F22, 0x5534, 0x5F28, 0x8020, 0x9120, 0x20A0}},
    {37179, 2, {0x2200, 0xF7BE, 0x2222, 0xF7A4, 0x5524, 0x88A8, 0x7F24, 0x824, 0x7F22, 0x4922, 0x7F22, 0x4934, 0x7F28, 0x820, 0xFFA0, 0x820}},
    {37180, 2, {0x2200, 0x773E, 0x2222, 0xF7A4, 0x2224, 0x5528, 0x88A4, 0x7F24, 0x4122, 0x7F22, 0x4122, 0x7F34, 0x4128, 0x7F20, 0x2220, 0x4120}},
    {37181, 2, {0x400, 0x7FDE, 0x4012, 0x5F92, 0x5494, 0x5F94, 0x5498, 0x5F94, 0x4412, 0x5F92, 0x4412, 0x7FDA, 0x5494, 0xAF50, 0x8410, 0x3FD0}},
    {37182, 2, {0x7F80, 0x83E, 0x3F22, 0x2124, 0x3F24, 0x2128, 0x7FA4, 0x48A4, 0x2522, 0x52A2, 0xE22, 0x1034, 0x3F28, 0x5120, 0xE20, 0x7120}},
    {37183, 2, {0x2200, 0xFFBE, 0x2222, 0x7724, 0x5524, 0x7728, 0x24, 0x7F24, 0x4922, 0x7F22, 0x4922, 0x7F34, 0x828, 0xFFA0, 0x820, 0x820}},
    {37184, 2, {0x2200, 0x221E, 0xFF92, 0x2212, 0x3E14, 0x4114, 0x5D58, 0xD594, 0x5512, 0x5552, 0xDD52, 0x40DA, 0x14, 0x5490, 0x4A50, 0x8A50}},
    {37185, 2, {0x3E00, 0x421E, 0xFF92, 0x4092, 0x7F94, 0x4494, 0x7718, 0x4454, 0x3FD2, 0x4212, 0xFF92, 0x489A, 0x7F94, 0x1A10, 0x2950, 0xC7D0}},
    {37186, 2, {0x1080, 0x529E, 0x7BD2, 0x9492, 0xFBF4, 0x3194, 0x5AD8, 0x9494, 0x12, 0x3F92, 0x2092, 0x249A, 0x2494, 0xB10, 0x1090, 0x6050}},
    {37187, 2, {0x0, 0x7FCE, 0x40A, 0xFFEA, 0x842A, 0x358A, 0x40C, 0x358A, 0xA, 0xA, 0xEEEA, 0xAAAA, 0xAAAC, 0xAAA8, 0xEEE8, 0xAAA8}},
    {37188, 2, {0x1100, 0x7FDE, 0x1112, 0x7BD2, 0x4A54, 0x7BD4, 0x1418, 0x1214, 0x3FD2, 0x2412, 0x7F92, 0xA41A, 0x3F94, 0x2410, 0x3FD0, 0x2010}},
    {37189, 2, {0x4440, 0x7FDE, 0x1212, 0x3FD2, 0x6214, 0xBFD4, 0x2218, 0x3FD4, 0x2212, 0x3FD2, 0x2012, 0x7FDA, 0x5154, 0x6ED0, 0x4A50, 0x4ED0}},
    {37190, 2, {0x2A00, 0x7F3E, 0xAAA2, 0xFFA4, 0xAAA4, 0xFFA8, 0xAAA4, 0xFFA4, 0x22, 0xFFA2, 0x22, 0x7F34, 0x4128, 0x7F20, 0x2220, 0xFFA0}},
    {37191, 2, {0x5280, 0x7BDE, 0x9492, 0xFBF2, 0x3194, 0x5AD4, 0x9498, 0x3F94, 0x2092, 0x3F92, 0x2092, 0x3F9A, 0x2094, 0x3F90, 0x1110, 0x2090}},
    {37192, 2, {0x7BC0, 0x1E, 0x7BD2, 0x4A52, 0x6B54, 0x4A54, 0x418, 0x7FD4, 0x4A12, 0x7FD2, 0x4A52, 0x7FDA, 0x5254, 0x5B90, 0x5250, 0x99D0}},
    {37193, 2, {0x0, 0xFFFE, 0x440, 0x440, 0x7FFC, 0x4444, 0x4444, 0x4844, 0x503C, 0x6004, 0x4004, 0x7FFC, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {37194, 2, {0x0, 0xFF80, 0x147E, 0x1410, 0x7F10, 0x5510, 0x5510, 0x5510, 0x5710, 0x6110, 0x4110, 0x7F10, 0x4110, 0x4110, 0x7F50, 0x4120}},
    {37195, 2, {0x1010, 0x810, 0x820, 0xFFFE, 0x440, 0x440, 0x7FFC, 0x4444, 0x4844, 0x503C, 0x6004, 0x7FFC, 0x4004, 0x4004, 0x7FFC, 0x4004}},
    {37196, 2, {0x20, 0xFE20, 0x2820, 0x287C, 0xFE44, 0xAA84, 0xAA04, 0xAA44, 0xAE24, 0xC214, 0x8214, 0xFE04, 0x8204, 0x8204, 0xFE28, 0x8210}},
    {37197, 2, {0x0, 0xFE00, 0x28F8, 0x2808, 0xFE08, 0xAA08, 0xAA08, 0xAAF8, 0xAE88, 0xC280, 0x8280, 0xFE80, 0x8282, 0x8282, 0xFE7E, 0x8200}},
    {37198, 2, {0x8, 0xFE08, 0x2808, 0x2808, 0xFEFE, 0xAA08, 0xAA08, 0xAA08, 0xAE48, 0xC228, 0x8228, 0xFE08, 0x8208, 0x8208, 0xFE28, 0x8210}},
    {37199, 2, {0x20, 0xFE20, 0x2820, 0x28A0, 0xFEAC, 0xAAB4, 0xAAE4, 0xABA4, 0xAEA4, 0xC2B4, 0x82A8, 0xFEA2, 0x82A2, 0x8282, 0xFE7E, 0x8200}},
    {37200, 2, {0x0, 0xFE7C, 0x2810, 0x2810, 0xFE10, 0xAA10, 0xAA10, 0xAAFE, 0xAE10, 0xC210, 0x8210, 0xFE10, 0x8210, 0x8210, 0xFE10, 0x8210}},
    {37201, 2, {0x0, 0xFEFC, 0x2820, 0x2820, 0xFE20, 0xAA20, 0xAA20, 0xABFE, 0xAE20, 0xC220, 0x8220, 0xFE20, 0x8220, 0x8220, 0xFEA0, 0x8240}},
    {37202, 2, {0x0, 0x4FFE, 0x20A0, 0x20A0, 0x7FC, 0x84A4, 0x44A4, 0x54A4, 0x151C, 0x2604, 0xE404, 0x27FC, 0x2404, 0x2404, 0x27FC, 0x404}},
    {37203, 2, {0x100, 0x280, 0xC60, 0x3118, 0xCFE6, 0x40, 0x80, 0xFFFE, 0x480, 0x3FF8, 0x2488, 0x2878, 0x3008, 0x3FF8, 0x2008, 0x3FF8}},
    {37204, 2, {0x40, 0xFE40, 0x29F8, 0x2848, 0xFE88, 0xAA8A, 0xAB0A, 0xAA26, 0xAE20, 0xC220, 0x83FE, 0xFE20, 0x8220, 0x8220, 0xFE20, 0x8220}},
    {37205, 2, {0x8, 0xFE1C, 0x28F0, 0x2810, 0xFE10, 0xAA1C, 0xAAF0, 0xAA10, 0xAE10, 0xC21E, 0x82F0, 0xFE10, 0x8212, 0x8212, 0xFE12, 0x820E}},
    {37206, 2, {0x10, 0xFE10, 0x2810, 0x2810, 0xFEFE, 0xAA92, 0xAA92, 0xAA10, 0xAE28, 0xC228, 0x8228, 0xFE28, 0x8248, 0x824A, 0xFE8A, 0x8306}},
    {37207, 2, {0x0, 0xFE44, 0x2844, 0x2828, 0xFEAA, 0xAAAA, 0xAA92, 0xAA92, 0xAEAA, 0xC2AA, 0x82A6, 0xFEC6, 0x8282, 0x8282, 0xFEFE, 0x8202}},
    {37208, 2, {0x0, 0xFE78, 0x2848, 0x2848, 0xFE48, 0xAA86, 0xAB00, 0xAAFC, 0xAE44, 0xC244, 0x8228, 0xFE28, 0x8210, 0x8228, 0xFE44, 0x8382}},
    {37209, 2, {0x8, 0xFE88, 0x2848, 0x2848, 0xFE08, 0xAA88, 0xAA48, 0xAA48, 0xAE08, 0xC20E, 0x82F8, 0xFE08, 0x8208, 0x8208, 0xFE08, 0x8208}},
    {37210, 2, {0x8, 0xFE28, 0x2828, 0x2828, 0xFE44, 0xAA44, 0xAA82, 0xAB7C, 0xAE24, 0xC224, 0x8224, 0xFE24, 0x8244, 0x8244, 0xFE94, 0x8308}},
    {37211, 2, {0x0, 0xFEFC, 0x2800, 0x2800, 0xFE00, 0xAAFE, 0xAA50, 0xAA50, 0xAE50, 0xC250, 0x8250, 0xFE52, 0x8252, 0x8292, 0xFE8E, 0x8300}},
    {37212, 2, {0x20, 0xFE20, 0x2820, 0x28FC, 0xFE20, 0xAA20, 0xAA20, 0xABFE, 0xAE20, 0xC250, 0x8250, 0xFE48, 0x8288, 0x8284, 0xFF04, 0x8202}},
    {37213, 2, {0x0, 0xFE00, 0x287C, 0x2800, 0xFE00, 0xAA00, 0xAAFE, 0xAA10, 0xAE10, 0xC220, 0x8220, 0xFE44, 0x8242, 0x82FE, 0xFE42, 0x8200}},
    {37214, 2, {0x20, 0xFE20, 0x2820, 0x2820, 0xFE20, 0xABFE, 0xAA20, 0xAA20, 0xAE20, 0xC250, 0x8250, 0xFE50, 0x8288, 0x82C8, 0xFF24, 0x8202}},
    {37215, 2, {0x10, 0xFF90, 0x1410, 0x1410, 0x7F1E, 0x5510, 0x5510, 0x5510, 0x577C, 0x6144, 0x4144, 0x7F44, 0x4144, 0x4144, 0x7F7C, 0x4144}},
    {37216, 2, {0x0, 0xFE00, 0x28FE, 0x2804, 0xFE04, 0xAAF4, 0xAA94, 0xAA94, 0xAE94, 0xC294, 0x82F4, 0xFE94, 0x8204, 0x8204, 0xFE14, 0x8208}},
    {37217, 2, {0x20, 0xFE10, 0x2810, 0x28FE, 0xFE82, 0xAA84, 0xAA40, 0xAA44, 0xAE48, 0xC250, 0x8260, 0xFE42, 0x8242, 0x8242, 0xFE3E, 0x8200}},
    {37218, 2, {0x40, 0xFE40, 0x2840, 0x287E, 0xFEA0, 0xAAA0, 0xAB20, 0xAA3C, 0xAE20, 0xC220, 0x8220, 0xFE3E, 0x8220, 0x8220, 0xFE20, 0x8220}},
    {37219, 2, {0x44, 0xFE44, 0x2844, 0x2844, 0xFEFE, 0xAA44, 0xAA44, 0xAA44, 0xAE44, 0xC27C, 0x8244, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {37220, 2, {0x10, 0xFE10, 0x2810, 0x2810, 0xFEFE, 0xAA10, 0xAA10, 0xAA10, 0xAE7C, 0xC244, 0x8244, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {37221, 2, {0x4, 0xFE1E, 0x28F0, 0x2810, 0xFE10, 0xAA10, 0xAAFE, 0xAA10, 0xAE38, 0xC234, 0x8254, 0xFE52, 0x8290, 0x8210, 0xFE10, 0x8210}},
    {37222, 2, {0x20, 0xFE28, 0x28A4, 0x28A0, 0xFEFE, 0xAA40, 0xAA40, 0xAA7C, 0xAEA4, 0xC2A4, 0x82A8, 0xFEA8, 0x8290, 0x8328, 0xFE44, 0x8282}},
    {37223, 2, {0x20, 0xFE10, 0x28FE, 0x2882, 0xFF0C, 0xAA08, 0xAA08, 0xAAFE, 0xAE08, 0xC248, 0x8228, 0xFE28, 0x8208, 0x8208, 0xFE28, 0x8210}},
    {37224, 2, {0x820, 0x828, 0x7F24, 0x820, 0xFFFE, 0x20, 0xFFA4, 0x1424, 0x7F24, 0x5528, 0x6328, 0x4110, 0x7F12, 0x412A, 0x7F46, 0x4182}},
    {37225, 2, {0x20, 0xFE20, 0x287E, 0x2882, 0xFF44, 0xAA28, 0xAA10, 0xAA20, 0xAE40, 0xC2FE, 0x8342, 0xFE42, 0x8242, 0x8242, 0xFE7E, 0x8242}},
    {37226, 2, {0x20, 0xFE20, 0x283C, 0x2844, 0xFEA8, 0xAA10, 0xAA28, 0xAA44, 0xAE82, 0xC27C, 0x8244, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {37227, 2, {0x10, 0xFE10, 0x2828, 0x2828, 0xFE44, 0xAA82, 0xAA7C, 0xAA10, 0xAE10, 0xC210, 0x827C, 0xFE10, 0x8210, 0x8210, 0xFEFE, 0x8200}},
    {37228, 2, {0x4, 0xFE44, 0x2854, 0x2854, 0xFE54, 0xAA54, 0xAADC, 0xAB74, 0xAE54, 0xC254, 0x8254, 0xFE54, 0x8254, 0x8254, 0xFE84, 0x8304}},
    {37229, 2, {0x20, 0xFE20, 0x28FE, 0x2820, 0xFE40, 0xAA7C, 0xAA44, 0xAAC4, 0xAF7C, 0xC244, 0x8244, 0xFE7C, 0x8244, 0x8244, 0xFE44, 0x824C}},
    {37230, 2, {0x0, 0xFEFE, 0x2882, 0x2882, 0xFEBA, 0xAA82, 0xAA82, 0xAABA, 0xAEAA, 0xC2AA, 0x82AA, 0xFEBA, 0x8282, 0x8282, 0xFE8A, 0x8284}},
    {37231, 2, {0x80, 0xFE82, 0x288C, 0x28F0, 0xFE82, 0xAA82, 0xAA7E, 0xAA00, 0xAEFC, 0xC284, 0x8284, 0xFEFC, 0x8284, 0x8284, 0xFEFC, 0x8284}},
    {37232, 2, {0x10, 0xFE50, 0x2850, 0x287C, 0xFE90, 0xAA10, 0xAA10, 0xAAFE, 0xAE28, 0xC228, 0x8228, 0xFE28, 0x824A, 0x824A, 0xFE86, 0x8300}},
    {37233, 2, {0x840, 0x4880, 0x29FC, 0x1A48, 0x2830, 0xC9C0, 0x800, 0xFFFE, 0x480, 0x3FF8, 0x2488, 0x2878, 0x3008, 0x3FF8, 0x2008, 0x3FF8}},
    {37234, 2, {0x0, 0xFE7C, 0x2844, 0x2844, 0xFE44, 0xAA7C, 0xAA00, 0xAA00, 0xAEFE, 0xC210, 0x8210, 0xFE7C, 0x8210, 0x8210, 0xFEFE, 0x8200}},
    {37235, 2, {0x20, 0xFE44, 0x28F8, 0x2810, 0xFE24, 0xAAFE, 0xAA02, 0xAAFC, 0xAE84, 0xC2FC, 0x8284, 0xFEFC, 0x8284, 0x8284, 0xFE94, 0x8288}},
    {37236, 2, {0x10, 0xFE10, 0x2828, 0x2828, 0xFE44, 0xAABA, 0xAA10, 0xAA10, 0xAEFE, 0xC210, 0x8258, 0xFE54, 0x8292, 0x8312, 0xFE50, 0x8220}},
    {37237, 2, {0x20, 0xFE22, 0x29FA, 0x2824, 0xFE24, 0xABFE, 0xAA10, 0xAA20, 0xAFFC, 0xC288, 0x8310, 0xFEFE, 0x8210, 0x8210, 0xFE50, 0x8220}},
    {37238, 2, {0x40, 0xFE40, 0x287E, 0x2880, 0xFF7C, 0xAA44, 0xAA64, 0xAA54, 0xAEFE, 0xC244, 0x82A4, 0xFE94, 0x82FE, 0x8204, 0xFE28, 0x8210}},
    {37239, 2, {0x10, 0xFE50, 0x2850, 0x287C, 0xFE50, 0xAA90, 0xAA10, 0xAAFE, 0xAE00, 0xC200, 0x827C, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {37240, 2, {0x10, 0xFE10, 0x2820, 0x2844, 0xFEFE, 0xAA28, 0xAA44, 0xAAA2, 0xAE3C, 0xC244, 0x8244, 0xFEA8, 0x8210, 0x8228, 0xFE44, 0x8282}},
    {37241, 2, {0x4, 0xFE1E, 0x28E0, 0x2822, 0xFE92, 0xAA54, 0xAA40, 0xAA08, 0xAFFE, 0xC208, 0x8288, 0xFE48, 0x8248, 0x8208, 0xFE28, 0x8210}},
    {37242, 2, {0x14, 0xFE12, 0x28FE, 0x2810, 0xFE10, 0xAAFE, 0xAA92, 0xAA92, 0xAEFE, 0xC292, 0x8292, 0xFEFE, 0x8292, 0x8292, 0xFE92, 0x8286}},
    {37243, 2, {0xE, 0xFEF0, 0x2822, 0x2892, 0xFE54, 0xAA00, 0xAA7C, 0xAA08, 0xAE10, 0xC210, 0x82FE, 0xFE10, 0x8210, 0x8210, 0xFE50, 0x8220}},
    {37244, 2, {0x20, 0xFE10, 0x29FE, 0x2820, 0xFE48, 0xAA84, 0xABFE, 0xAA02, 0xAEA8, 0xC2A8, 0x82A8, 0xFEA8, 0x82A8, 0x82AA, 0xFF2A, 0x8206}},
    {37245, 2, {0x0, 0xFEFE, 0x2828, 0x28AA, 0xFE6C, 0xAA28, 0xAAFE, 0xAA80, 0xAE80, 0xC280, 0x8280, 0xFE80, 0x8280, 0x8300, 0xFF00, 0x8200}},
    {37246, 2, {0x0, 0xFBFE, 0x2000, 0x2000, 0xFBDE, 0xAA52, 0xAA52, 0xAA52, 0xDB5A, 0x8AD6, 0x8A52, 0xFA52, 0x8A52, 0x8A52, 0xFA52, 0x8AD6}},
    {37247, 2, {0x20, 0xFE10, 0x28FC, 0x2884, 0xFE84, 0xAAFC, 0xAA84, 0xAA84, 0xAEFC, 0xC2A2, 0x82A4, 0xFE98, 0x8290, 0x8288, 0xFEC4, 0x8282}},
    {37248, 2, {0x50, 0xFE48, 0x2880, 0x28FE, 0xFF90, 0xAA90, 0xAAFC, 0xAA90, 0xAE90, 0xC2FC, 0x8290, 0xFE90, 0x8290, 0x82FE, 0xFE80, 0x8280}},
    {37249, 2, {0x0, 0xFEFC, 0x2804, 0x2804, 0xFE7C, 0xAA04, 0xAA04, 0xAAFE, 0xAE10, 0xC292, 0x8254, 0xFE38, 0x8254, 0x8292, 0xFE50, 0x8220}},
    {37250, 2, {0x48, 0xFE48, 0x2848, 0x2848, 0xFEFE, 0xAA48, 0xAA48, 0xAAEC, 0xAEDC, 0xC35A, 0x8248, 0xFE48, 0x8248, 0x8248, 0xFE48, 0x8248}},
    {37251, 2, {0x20, 0xFE20, 0x29FE, 0x2850, 0xFEA8, 0xAB26, 0xAAF8, 0xAAA8, 0xAEA8, 0xC2F8, 0x82A8, 0xFEA8, 0x82F8, 0x82AA, 0xFE22, 0x821E}},
    {37252, 2, {0x40, 0xFE40, 0x28FE, 0x2882, 0xFF42, 0xAA7A, 0xAAA2, 0xAA22, 0xAFFE, 0xC222, 0x82AA, 0xFEAA, 0x82FA, 0x8202, 0xFE14, 0x8208}},
    {37253, 2, {0x20, 0xFE10, 0x28FC, 0x2800, 0xFE84, 0xAA48, 0xAA00, 0xAAFE, 0xAE00, 0xC200, 0x82FC, 0xFE84, 0x8284, 0x8284, 0xFEFC, 0x8284}},
    {37254, 2, {0x28, 0xFE24, 0x282E, 0x28F0, 0xFE28, 0xAA12, 0xAA2A, 0xAAD6, 0xAE28, 0xC22E, 0x82F0, 0xFE24, 0x8228, 0x8212, 0xFE2A, 0x82C6}},
    {37255, 2, {0x20, 0xFE10, 0x28FE, 0x2800, 0xFE7C, 0xAA44, 0xAA7C, 0xAA00, 0xAE7C, 0xC208, 0x8210, 0xFEFE, 0x8210, 0x8210, 0xFE50, 0x8220}},
    {37256, 2, {0x20, 0xFEA2, 0x28A2, 0x2924, 0xFE50, 0xAA88, 0xAB04, 0xAA22, 0xAE20, 0xC2A4, 0x82A4, 0xFF28, 0x8250, 0x8288, 0xFF04, 0x8202}},
    {37257, 2, {0x40, 0xFE20, 0x29FE, 0x2800, 0xFE88, 0xAA88, 0xAB54, 0xAA22, 0xAE00, 0xC220, 0x83FE, 0xFE20, 0x8220, 0x8220, 0xFE20, 0x8220}},
    {37258, 2, {0x0, 0xFEEE, 0x2822, 0x28AA, 0xFE44, 0xAAAA, 0xAB12, 0xAA00, 0xAEEE, 0xC22A, 0x82AA, 0xFE4A, 0x8244, 0x82A4, 0xFF2A, 0x8210}},
    {37259, 2, {0x48, 0xFE48, 0x2848, 0x29FE, 0xFE48, 0xAA48, 0xABFE, 0xAA00, 0xAEFC, 0xC284, 0x8284, 0xFEFC, 0x8284, 0x8284, 0xFEFC, 0x8284}},
    {37260, 2, {0x0, 0xFEFC, 0x2884, 0x2884, 0xFEFC, 0xAA84, 0xAA84, 0xAAFC, 0xAE00, 0xC290, 0x8292, 0xFEF4, 0x8298, 0x8292, 0xFED2, 0x828E}},
    {37261, 2, {0x0, 0xFE7C, 0x2844, 0x287C, 0xFE44, 0xAA7C, 0xAA00, 0xABFE, 0xAE10, 0xC210, 0x825E, 0xFE50, 0x8250, 0x82B0, 0xFE9E, 0x8300}},
    {37262, 2, {0x8, 0xF80C, 0x200A, 0x23FE, 0xFA08, 0xAA08, 0xAAE8, 0xAA0A, 0xDA0A, 0x8AEC, 0x8AAC, 0xFAA8, 0x8AEA, 0x8A1A, 0xFA26, 0x8C42}},
    {37263, 2, {0x20, 0xFE20, 0x28FA, 0x2824, 0xFE28, 0xABFE, 0xAA20, 0xAA40, 0xAEFC, 0xC344, 0x8244, 0xFE7C, 0x8244, 0x8244, 0xFE7C, 0x8244}},
    {37264, 2, {0x100, 0xF93C, 0x2124, 0x27E4, 0xF924, 0xA93C, 0xA924, 0xABA4, 0xDAA4, 0x8ABC, 0x8AA4, 0xFAA4, 0x8BA4, 0x8844, 0xF854, 0x8888}},
    {37265, 2, {0x0, 0xFEFC, 0x2814, 0x2850, 0xFE5C, 0xAA50, 0xAA7E, 0xAA80, 0xAE7C, 0xC244, 0x827C, 0xFE44, 0x827C, 0x8244, 0xFE44, 0x824C}},
    {37266, 2, {0x0, 0xFE7C, 0x2844, 0x287C, 0xFE44, 0xAA7C, 0xAA10, 0xAA50, 0xAE7C, 0xC290, 0x8210, 0xFE7C, 0x8210, 0x8210, 0xFEFE, 0x8200}},
    {37267, 2, {0x10, 0xFE10, 0x28FE, 0x2892, 0xFE92, 0xAA28, 0xAA28, 0xAA4A, 0xAE86, 0xC300, 0x82FE, 0xFEAA, 0x82AA, 0x82AA, 0xFEAA, 0x83FE}},
    {37268, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x904, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8}},
    {37269, 2, {0x20, 0xFE10, 0x29FE, 0x2800, 0xFE7C, 0xAA44, 0xAA7C, 0xAA00, 0xAEFE, 0xC282, 0x8282, 0xFEFE, 0x8282, 0x8282, 0xFEFE, 0x8282}},
    {37270, 2, {0x0, 0xFE7C, 0x2844, 0x2844, 0xFE7C, 0xAA44, 0xAA44, 0xAA7C, 0xAE00, 0xC2FE, 0x82AA, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {37271, 2, {0x8, 0xFEEC, 0x2828, 0x28AA, 0xFE44, 0xAA44, 0xAA84, 0xAB7A, 0xAE28, 0xC228, 0x82FE, 0xFE28, 0x8228, 0x824A, 0xFE4A, 0x8286}},
    {37272, 2, {0x10, 0xFE10, 0x287E, 0x2810, 0xFE10, 0xAAFE, 0xAA20, 0xAA42, 0xAEFE, 0xC200, 0x82FE, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {37273, 2, {0x50, 0xFE96, 0x2892, 0x2892, 0xFED6, 0xAA92, 0xAA92, 0xAAFE, 0xAE10, 0xC2FC, 0x8244, 0xFE44, 0x8228, 0x8210, 0xFE28, 0x82C6}},
    {37274, 2, {0x8, 0xFE4A, 0x292A, 0x28AC, 0xFE88, 0xAA3E, 0xAA08, 0xAB98, 0xAE9C, 0xC2AA, 0x82CA, 0xFE88, 0x8288, 0x8340, 0xFE3E, 0x8200}},
    {37275, 2, {0x44, 0xFE44, 0x28FE, 0x2844, 0xFE10, 0xAA28, 0xAA44, 0xAA82, 0xAE7C, 0xC210, 0x8210, 0xFE7C, 0x8210, 0x8210, 0xFEFE, 0x8200}},
    {37276, 2, {0x10, 0xFE20, 0x28FE, 0x2892, 0xFE92, 0xAAFE, 0xAA92, 0xAA92, 0xAEFE, 0xC220, 0x8224, 0xFE5A, 0x825E, 0x8290, 0xFE92, 0x830E}},
    {37277, 2, {0x44, 0xFE24, 0x2828, 0x28FE, 0xFE10, 0xAA7C, 0xAA10, 0xAAFE, 0xAE20, 0xC220, 0x827C, 0xFE50, 0x8290, 0x8310, 0xFEFE, 0x8200}},
    {37278, 2, {0x0, 0xFE7C, 0x2854, 0x2854, 0xFE6C, 0xAA44, 0xAA7C, 0xAA00, 0xAE00, 0xC2FE, 0x82AA, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {37279, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4002, 0xBFF4, 0x480, 0x3FF8, 0x2488, 0x2878, 0x3008, 0x3FF8, 0x2008, 0x3FF8}},
    {37280, 2, {0x10, 0xFE10, 0x287C, 0x2854, 0xFE54, 0xAAFE, 0xAA28, 0xAA44, 0xAE82, 0xC300, 0x82FE, 0xFEAA, 0x82AA, 0x82AA, 0xFFFE, 0x8200}},
    {37281, 2, {0x40, 0xFE20, 0x29FE, 0x2902, 0xFE50, 0xAA88, 0xAB44, 0xAA40, 0xAEFE, 0xC320, 0x823C, 0xFE20, 0x823E, 0x8220, 0xFE20, 0x8220}},
    {37282, 2, {0x20, 0xF820, 0x23FC, 0x2040, 0xF880, 0xA9F8, 0xAA88, 0xAC88, 0xD8F8, 0x8800, 0x89FC, 0xF954, 0x8954, 0x8954, 0xFBFE, 0x8800}},
    {37283, 2, {0x20, 0xFE10, 0x28FE, 0x2890, 0xFEBC, 0xAA94, 0xAAFE, 0xAA94, 0xAEBC, 0xC290, 0x82BC, 0xFEA4, 0x82A4, 0x82A4, 0xFEBC, 0x8324}},
    {37284, 2, {0x83E, 0x4BC4, 0x2A48, 0x2928, 0x1BFE, 0x2908, 0xC8A8, 0x810, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8}},
    {37285, 2, {0x0, 0xFEFE, 0x2828, 0x28FE, 0xFEAA, 0xAAAA, 0xAAFE, 0xAA00, 0xAE7C, 0xC200, 0x82FE, 0xFE10, 0x8254, 0x8292, 0xFE50, 0x8220}},
    {37286, 2, {0x10, 0xFE20, 0x2844, 0x28FE, 0xFE44, 0xAAAA, 0xAAEE, 0xAA10, 0xAE28, 0xC254, 0x82AA, 0xFE50, 0x8224, 0x8248, 0xFE10, 0x8260}},
    {37287, 2, {0x0, 0xFBFE, 0x2200, 0x227C, 0xFA44, 0xAA44, 0xAA7C, 0xAA00, 0xDAEE, 0x8AAA, 0x8AAA, 0xFAAA, 0x8AEE, 0x8A00, 0xFBFE, 0x8800}},
    {37288, 2, {0x20, 0xFE10, 0x28FE, 0x2828, 0xFE92, 0xAAAA, 0xAA82, 0xAAFE, 0xAE10, 0xC2FE, 0x82A2, 0xFEAA, 0x82BA, 0x8282, 0xFE8A, 0x8284}},
    {37289, 2, {0x28, 0xFE28, 0x28FE, 0x2828, 0xFEFE, 0xAAAA, 0xAAFE, 0xAAAA, 0xAEFE, 0xC200, 0x827C, 0xFE44, 0x827C, 0x8244, 0xFE7C, 0x8244}},
    {37290, 2, {0x0, 0xFEEE, 0x2822, 0x28AA, 0xFE66, 0xAAAA, 0xAA10, 0xAA28, 0xAE44, 0xC392, 0x8220, 0xFEC8, 0x8232, 0x82C4, 0xFE18, 0x82E0}},
    {37291, 2, {0x7F78, 0x4848, 0x5E86, 0x6478, 0x5F48, 0x4A30, 0x7F48, 0x84, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8}},
    {37292, 2, {0x440, 0x247C, 0x3CA8, 0x550, 0x7C28, 0x25FE, 0x4488, 0x8058, 0xFFFE, 0x440, 0x3FF8, 0x2848, 0x3038, 0x27C8, 0x2008, 0x3FF8}},
    {37293, 2, {0x28, 0xFEAA, 0x286C, 0x2828, 0xFEFE, 0xAA44, 0xAA28, 0xAAFE, 0xAE10, 0xC27C, 0x8210, 0xFEFE, 0x8210, 0x8228, 0xFE44, 0x8282}},
    {37294, 2, {0x28, 0xFE24, 0x2840, 0x287E, 0xFEC8, 0xAB48, 0xAA7E, 0xAA48, 0xAE48, 0xC27E, 0x8248, 0xFE48, 0x827E, 0x8240, 0xFEAA, 0x832A}},
    {37295, 2, {0x40, 0xF820, 0x23FE, 0x2040, 0xF884, 0xA9FE, 0xA8A8, 0xA8AA, 0xD926, 0x8A00, 0x89FC, 0xF954, 0x8954, 0x8954, 0xFBFE, 0x8800}},
    {37296, 2, {0x0, 0xFEFE, 0x2828, 0x28FE, 0xFEAA, 0xAAFE, 0xAA00, 0xAA7C, 0xAE44, 0xC27C, 0x8244, 0xFE7C, 0x8210, 0x82FE, 0xFE10, 0x8210}},
    {37297, 2, {0x24, 0xFBA8, 0x2092, 0x2514, 0xFA08, 0xAC04, 0xABBA, 0xA8A8, 0xD8C6, 0x8B80, 0x8A3C, 0xFB84, 0x88A8, 0x8890, 0xFAA8, 0x8944}},
    {37298, 2, {0x28, 0xFEFE, 0x28AA, 0x28FE, 0xFEAA, 0xAAFE, 0xAA00, 0xAAFE, 0xAE80, 0xC2BE, 0x8280, 0xFEFE, 0x82A8, 0x82AA, 0xFEA4, 0x8332}},
    {37299, 2, {0x0, 0xFEFE, 0x28AA, 0x28FE, 0xFE10, 0xAAFE, 0xAA10, 0xABFE, 0xAE44, 0xC228, 0x82FE, 0xFE10, 0x83FE, 0x8210, 0xFE10, 0x8210}},
    {37300, 2, {0x28, 0xFEFE, 0x28AA, 0x28FE, 0xFEAA, 0xAAFE, 0xAA00, 0xAAFE, 0xAE00, 0xC2FE, 0x8282, 0xFEFE, 0x8244, 0x8228, 0xFEFE, 0x8200}},
    {37301, 2, {0x40, 0xF87C, 0x2040, 0x23FE, 0xFA42, 0xAA78, 0xABC4, 0xAA3C, 0xDA00, 0x8BFE, 0x8A40, 0xFBA4, 0x8A58, 0x8DB4, 0xFC52, 0x89B0}},
    {37302, 2, {0x10, 0xFE10, 0x2828, 0x2844, 0xFE82, 0xAA7C, 0xAA00, 0xAAEE, 0xAEAA, 0xC2AA, 0x82EE, 0xFE44, 0x8244, 0x82AA, 0xFEAA, 0x8312}},
    {37303, 2, {0x10, 0xFEFE, 0x2844, 0x2828, 0xFEFE, 0xAA00, 0xAA7C, 0xAA44, 0xAE7C, 0xC244, 0x827C, 0xFE20, 0x8294, 0x8342, 0xFE44, 0x823C}},
    {37304, 2, {0x20, 0xFE10, 0x29FE, 0x2848, 0xFE84, 0xAB4A, 0xAAFC, 0xAA48, 0xAEFC, 0xC248, 0x83FE, 0xFE54, 0x82C8, 0x8354, 0xFE62, 0x8240}},
    {37305, 2, {0x0, 0xFE7C, 0x2810, 0x28FE, 0xFE92, 0xAA54, 0xAA10, 0xAA54, 0xAE00, 0xC2FE, 0x8210, 0xFEFE, 0x82AA, 0x82AA, 0xFEAA, 0x8286}},
    {37306, 2, {0x1C, 0xF9E0, 0x2020, 0x23FE, 0xF820, 0xA9FC, 0xA9AC, 0xA974, 0xD9FC, 0x8820, 0x89FC, 0xF820, 0x8BFE, 0x8800, 0xF954, 0x8A2A}},
    {37307, 2, {0x20, 0xFBFE, 0x2020, 0x21FC, 0xF800, 0xABFE, 0xA802, 0xA9FC, 0xD820, 0x8BFE, 0x8800, 0xFBFE, 0x8804, 0x8BBE, 0xFAA4, 0x8B8C}},
    {37308, 2, {0x110, 0xF910, 0x27FC, 0x2110, 0xF9F0, 0xAA08, 0xAAEA, 0xAEAC, 0xDAA8, 0x8AAA, 0x8EEA, 0xFA06, 0x8800, 0x8AA4, 0xFA52, 0x8C52}},
    {37309, 2, {0x0, 0xF9FC, 0x2020, 0x23FE, 0xFA22, 0xA9AC, 0xA820, 0xA9AC, 0xD800, 0x8BEE, 0x8AAA, 0xFAAA, 0x8AAA, 0x8AAA, 0xFBEE, 0x8AAA}},
    {37310, 2, {0x40, 0xF820, 0x23FE, 0x2248, 0xFBFE, 0xAA48, 0xAAEC, 0xAB5A, 0xDA48, 0x8B24, 0x8AA8, 0xFBFE, 0x8A70, 0x8AA8, 0xFD26, 0x8820}},
    {37311, 2, {0x20, 0xFBFE, 0x2248, 0x23FE, 0xFA48, 0xAAEC, 0xAB5A, 0xAA48, 0xDA20, 0x8A44, 0x8A78, 0xFA14, 0x8AFE, 0x8A10, 0xFA54, 0x8CB2}},
    {37312, 2, {0x10, 0xFEFE, 0x2800, 0x28EE, 0xFEAA, 0xAAEE, 0xAA44, 0xAAFE, 0xAE44, 0xC2FE, 0x8244, 0xFEFE, 0x8250, 0x82CA, 0xFF64, 0x8242}},
    {37313, 2, {0x2FEC, 0x4824, 0x6BAC, 0x4824, 0x6BAC, 0x4AA4, 0xFFFE, 0x9292, 0x1D70, 0x1010, 0x1FF0, 0x440, 0x1830, 0xEFEE, 0x420, 0x1860}},
    {37314, 2, {0x1E, 0xFBE0, 0x2122, 0x2094, 0xFBFE, 0xAA52, 0xABFE, 0xA804, 0xDBC4, 0x8A7E, 0x8BC4, 0xFA54, 0x8BCC, 0x8A84, 0xFAD4, 0x8B48}},
    {37315, 2, {0x3DE, 0xF800, 0x23DE, 0x2252, 0xFB5A, 0xAA52, 0xA820, 0xABFE, 0xDA50, 0x8BFE, 0x8A52, 0xFBFE, 0x8A92, 0x8ADC, 0xFA92, 0x8CCE}},
    {37316, 2, {0x20, 0xFBFE, 0x2248, 0x23FE, 0xFA48, 0xAAEC, 0xAB5A, 0xAA48, 0xDA50, 0x8BDE, 0x8A50, 0xFBDC, 0x8A50, 0x8BDE, 0xFA50, 0x8C50}},
    {37317, 2, {0x3DE, 0xFA52, 0x23DE, 0x2000, 0xFBFE, 0xAA00, 0xAAC8, 0xAA4E, 0xDBEA, 0x8AB2, 0x8AEA, 0xFAAA, 0x8AEA, 0x8AB4, 0xFDE4, 0x882A}},
    {37318, 2, {0x10, 0xF8, 0x3F00, 0x100, 0x1110, 0x910, 0x920, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {37319, 2, {0x10, 0xF8, 0x3F00, 0x200, 0x1110, 0x910, 0x820, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x1110, 0x2108, 0xC106, 0x100}},
    {37320, 2, {0xC00, 0xF1FC, 0x1104, 0x5504, 0x3904, 0x1104, 0xFDFC, 0x1120, 0x3120, 0x3920, 0x5520, 0x5110, 0x9110, 0x1108, 0x1204, 0x1402}},
    {37321, 2, {0xC20, 0xF020, 0x1020, 0x5420, 0x39FC, 0x1124, 0xFD24, 0x1124, 0x3124, 0x39FC, 0x5524, 0x5124, 0x9124, 0x1124, 0x11FC, 0x1104}},
    {37322, 2, {0xC00, 0xF3FC, 0x1104, 0x5488, 0x3850, 0x1020, 0xFCD8, 0x1326, 0x3020, 0x39FC, 0x5420, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020}},
    {37323, 2, {0xC00, 0xF1FC, 0x1154, 0x55FC, 0x3820, 0x11FC, 0xFC20, 0x13FE, 0x3088, 0x3850, 0x55FC, 0x5020, 0x93FE, 0x1020, 0x1020, 0x1020}},
    {37324, 2, {0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0xFFFE, 0x0}},
    {37325, 2, {0x10, 0xF8, 0x3F00, 0x100, 0xFFFE, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0}},
    {37326, 2, {0x0, 0x7F7C, 0x4904, 0x4928, 0x7F10, 0x4908, 0x497E, 0x7F12, 0x814, 0x810, 0x7F10, 0x810, 0x810, 0xF10, 0xF050, 0x4020}},
    {37327, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x0, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x7FFC}},
    {37328, 2, {0x820, 0x3E7E, 0x8A4, 0x7F28, 0x1C10, 0x2A28, 0x4944, 0x3FFE, 0x2000, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x4FF8, 0x4080, 0x9FFC}},
    {37329, 2, {0x100, 0x100, 0x280, 0x440, 0x820, 0x1010, 0x2FE8, 0xC106, 0x100, 0x3FF8, 0x100, 0x1110, 0x910, 0x920, 0xFFFE, 0x0}},
    {37330, 2, {0x1000, 0x1000, 0x2800, 0x2400, 0x4200, 0xBC00, 0x1000, 0x1000, 0xFE00, 0x1000, 0x9400, 0x5800, 0x5000, 0x1E00, 0xF000, 0x4000}},
    {37331, 2, {0x1080, 0x1080, 0x2880, 0x2480, 0x4280, 0xBC80, 0x1080, 0x1080, 0xFE80, 0x1080, 0x9480, 0x5884, 0x5084, 0x1E84, 0xF07C, 0x4000}},
    {37332, 2, {0x1000, 0x1000, 0x28FE, 0x2404, 0x4208, 0xBC10, 0x1020, 0x1020, 0xFE40, 0x1040, 0x9480, 0x5882, 0x5082, 0x1E82, 0xF07E, 0x4000}},
    {37333, 2, {0x1000, 0x11FC, 0x2804, 0x2408, 0x4210, 0xBC20, 0x1020, 0x1020, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF0A0, 0x4040}},
    {37334, 2, {0x1000, 0x1000, 0x29FC, 0x2444, 0x4244, 0xBC44, 0x1044, 0x1044, 0xFE44, 0x1044, 0x9444, 0x5884, 0x5084, 0x1F04, 0xF228, 0x4410}},
    {37335, 2, {0x1004, 0x1004, 0x2804, 0x2444, 0x4244, 0xBC44, 0x1044, 0x1044, 0xFE44, 0x1044, 0x9444, 0x5804, 0x5004, 0x1E04, 0xF014, 0x4008}},
    {37336, 2, {0x1000, 0x1000, 0x2BFE, 0x2420, 0x4220, 0xBC20, 0x1020, 0x1020, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF0A0, 0x4040}},
    {37337, 2, {0x1040, 0x1040, 0x2840, 0x2440, 0x4240, 0xBC50, 0x1048, 0x1044, 0xFE44, 0x1040, 0x9440, 0x5840, 0x5040, 0x1E40, 0xF040, 0x4040}},
    {37338, 2, {0x1080, 0x1080, 0x2880, 0x2480, 0x43F0, 0xB890, 0x1090, 0x1090, 0xFC90, 0x1090, 0x9490, 0x5892, 0x5112, 0x1D12, 0xE20E, 0x4400}},
    {37339, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x42FC, 0xBC24, 0x1024, 0x1024, 0xFE24, 0x1044, 0x9444, 0x5844, 0x5084, 0x1E84, 0xF128, 0x4210}},
    {37340, 2, {0x440, 0x820, 0x1850, 0x2488, 0x300, 0xCC0, 0x3030, 0xC00E, 0x1FF0, 0x100, 0x100, 0x3FF8, 0x100, 0x1110, 0x920, 0x7FFC}},
    {37341, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x4220, 0xBC20, 0x13FE, 0x1020, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF020, 0x4020}},
    {37342, 2, {0x1080, 0x1040, 0x2840, 0x2420, 0x4220, 0xBC20, 0x1020, 0x1050, 0xFE50, 0x1050, 0x9448, 0x5888, 0x5088, 0x1E84, 0xF104, 0x4202}},
    {37343, 2, {0x1000, 0x1010, 0x2890, 0x2490, 0x4090, 0xB890, 0x1090, 0x1090, 0xFC90, 0x1090, 0x9508, 0x5908, 0x5108, 0x1E04, 0xE204, 0x4402}},
    {37344, 2, {0x1000, 0x10F0, 0x2890, 0x2490, 0x4290, 0xBC90, 0x1090, 0x1090, 0xFE90, 0x1090, 0x9490, 0x5892, 0x5092, 0x1F12, 0xF10E, 0x4200}},
    {37345, 2, {0x820, 0x1010, 0x2108, 0x4284, 0x440, 0x820, 0x3018, 0xCFE6, 0x100, 0x100, 0x3FF8, 0x100, 0x1110, 0x920, 0xFFFE, 0x0}},
    {37346, 2, {0x1000, 0x1000, 0x29FC, 0x2444, 0x4248, 0xBC48, 0x1050, 0x105E, 0xFE42, 0x1042, 0x9482, 0x5882, 0x5082, 0x1F02, 0xF114, 0x4208}},
    {37347, 2, {0x1040, 0x1040, 0x2840, 0x247C, 0x4284, 0xBC84, 0x1104, 0x1044, 0xFE24, 0x1024, 0x9404, 0x5804, 0x5004, 0x1E04, 0xF028, 0x4010}},
    {37348, 2, {0x804, 0x804, 0x1408, 0x2210, 0x4120, 0xBE44, 0x804, 0x808, 0x7F10, 0x822, 0x4942, 0x2A04, 0x808, 0xF10, 0xF020, 0x40C0}},
    {37349, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x4020, 0xB8A8, 0x10A4, 0x10A4, 0xFD22, 0x1122, 0x9622, 0x5820, 0x5020, 0x1C20, 0xE0A0, 0x4040}},
    {37350, 2, {0x1000, 0x1000, 0x28FC, 0x2484, 0x4284, 0xBC84, 0x1084, 0x1084, 0xFE84, 0x1084, 0x9484, 0x5884, 0x5084, 0x1EFC, 0xF084, 0x4000}},
    {37351, 2, {0x1104, 0x1124, 0x2924, 0x2524, 0x4124, 0xB924, 0x1124, 0x1124, 0xFD24, 0x1124, 0x9524, 0x5924, 0x5124, 0x1E24, 0xE204, 0x4404}},
    {37352, 2, {0x1000, 0x10FC, 0x2804, 0x2408, 0x4210, 0xBC20, 0x1020, 0x11FE, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF0A0, 0x4040}},
    {37353, 2, {0x1000, 0x11F0, 0x2910, 0x2510, 0x4110, 0xB910, 0x1190, 0x1150, 0xFD50, 0x1110, 0x9510, 0x5912, 0x5112, 0x1E12, 0xE20E, 0x4400}},
    {37354, 2, {0x1000, 0x10FC, 0x2820, 0x2420, 0x4220, 0xBC20, 0x1020, 0x11FE, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF0A0, 0x4040}},
    {37355, 2, {0x1000, 0x10FC, 0x2800, 0x2400, 0x4200, 0xBDFE, 0x1040, 0x1040, 0xFE80, 0x10FC, 0x9404, 0x5804, 0x5004, 0x1E04, 0xF028, 0x4010}},
    {37356, 2, {0x1000, 0x10FC, 0x2820, 0x2420, 0x4220, 0xBC20, 0x1020, 0x11FE, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF020, 0x4020}},
    {37357, 2, {0x1000, 0x1000, 0x28FC, 0x2420, 0x4220, 0xBC20, 0x1020, 0x1020, 0xFE20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1E20, 0xF1FE, 0x4000}},
    {37358, 2, {0x1010, 0x1010, 0x2810, 0x2410, 0x43FE, 0xBC10, 0x1030, 0x1030, 0xFE50, 0x1050, 0x9490, 0x5910, 0x5010, 0x1E10, 0xF050, 0x4020}},
    {37359, 2, {0x1040, 0x1020, 0x2820, 0x2400, 0x43FE, 0xBC80, 0x1080, 0x1080, 0xFE80, 0x1080, 0x9480, 0x5880, 0x5080, 0x1E80, 0xF0FC, 0x4000}},
    {37360, 2, {0x1000, 0x1000, 0x29FC, 0x2424, 0x4024, 0xB8A4, 0x10A4, 0x10A4, 0xFD24, 0x1024, 0x9424, 0x5844, 0x5044, 0x1C84, 0xE114, 0x4208}},
    {37361, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x4220, 0xBDFE, 0x1020, 0x1020, 0xFE20, 0x1050, 0x9450, 0x5850, 0x5088, 0x1E88, 0xF104, 0x4202}},
    {37362, 2, {0x1000, 0x1000, 0x28FC, 0x2484, 0x4284, 0xBC84, 0x1084, 0x10FC, 0xFE84, 0x1080, 0x9480, 0x5882, 0x5082, 0x1E82, 0xF07E, 0x4000}},
    {37363, 2, {0x1040, 0x1040, 0x2880, 0x24FE, 0x4100, 0xBA00, 0x10FC, 0x1008, 0xFC10, 0x1020, 0x9440, 0x5880, 0x5102, 0x1D02, 0xE0FE, 0x4000}},
    {37364, 2, {0x1050, 0x1048, 0x2848, 0x2440, 0x425E, 0xBDE0, 0x1040, 0x1040, 0xFE40, 0x1040, 0x9420, 0x5822, 0x5012, 0x1E0A, 0xF006, 0x4002}},
    {37365, 2, {0x1000, 0x11F8, 0x2888, 0x24C8, 0x42A8, 0xBCA8, 0x1088, 0x1050, 0xFE50, 0x1050, 0x9420, 0x5820, 0x5050, 0x1E88, 0xF104, 0x4202}},
    {37366, 2, {0x1020, 0x1020, 0x2820, 0x2520, 0x412C, 0xB934, 0x1164, 0x13A4, 0xFD24, 0x1134, 0x9528, 0x5922, 0x5122, 0x1D02, 0xE0FE, 0x4000}},
    {37367, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x4020, 0xB820, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37368, 2, {0x1040, 0x1040, 0x2840, 0x247C, 0x4244, 0xBC84, 0x10C4, 0x1124, 0xFE18, 0x1008, 0x9410, 0x5810, 0x5020, 0x1E40, 0xF080, 0x4100}},
    {37369, 2, {0x1040, 0x1040, 0x2840, 0x2440, 0x43FE, 0xB888, 0x1088, 0x1088, 0xFC88, 0x1108, 0x9490, 0x5850, 0x5020, 0x1C50, 0xE088, 0x4304}},
    {37370, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x4020, 0xB820, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37371, 2, {0x1040, 0x1040, 0x2840, 0x2440, 0x41F8, 0xB848, 0x1048, 0x1048, 0xFD48, 0x10C8, 0x9448, 0x58A8, 0x50AA, 0x1D0A, 0xE206, 0x4402}},
    {37372, 2, {0x1000, 0x1000, 0x29FC, 0x2444, 0x4044, 0xB944, 0x10C4, 0x1044, 0xFC64, 0x1054, 0x944C, 0x5886, 0x5084, 0x1D04, 0xE228, 0x4410}},
    {37373, 2, {0x1008, 0x101C, 0x29F0, 0x2550, 0x4150, 0xB950, 0x1150, 0x1150, 0xFD50, 0x1148, 0x9548, 0x5948, 0x5144, 0x1E44, 0xE242, 0x4400}},
    {37374, 2, {0x1000, 0x11FE, 0x2808, 0x2488, 0x4088, 0xB888, 0x1108, 0x11FE, 0xFC18, 0x1028, 0x9448, 0x5888, 0x5108, 0x1E08, 0xE028, 0x4010}},
    {37375, 2, {0x1004, 0x100E, 0x28F0, 0x2480, 0x4280, 0xBC80, 0x10FE, 0x1088, 0xFE88, 0x1088, 0x9488, 0x5888, 0x5088, 0x1F08, 0xF108, 0x4208}},
    {37376, 2, {0x1000, 0x1000, 0x29FC, 0x2524, 0x4124, 0xB924, 0x1124, 0x1124, 0xFDFC, 0x1100, 0x9500, 0x5900, 0x5102, 0x1D02, 0xE0FE, 0x4000}},
    {37377, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x4240, 0xBC40, 0x1040, 0x107C, 0xFE44, 0x1044, 0x9444, 0x5844, 0x5084, 0x1E84, 0xF128, 0x4210}},
    {37378, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4122, 0xBA24, 0x1020, 0x1020, 0xFC50, 0x1050, 0x9450, 0x5850, 0x5090, 0x1C92, 0xE112, 0x420E}},
    {37379, 2, {0x1000, 0x10FC, 0x2848, 0x2448, 0x4248, 0xBC48, 0x1048, 0x11FE, 0xFE48, 0x1048, 0x9448, 0x5848, 0x5048, 0x1E48, 0xF088, 0x4108}},
    {37380, 2, {0x1008, 0x1088, 0x2848, 0x2448, 0x4208, 0xBC88, 0x1048, 0x1048, 0xFE08, 0x100E, 0x95F8, 0x5808, 0x5008, 0x1E08, 0xF008, 0x4008}},
    {37381, 2, {0x1000, 0x10FC, 0x2884, 0x2484, 0x4284, 0xBCFC, 0x1084, 0x1084, 0xFE84, 0x10FC, 0x9484, 0x5884, 0x5084, 0x1E84, 0xF114, 0x4208}},
    {37382, 2, {0x1008, 0x1048, 0x2848, 0x2448, 0x4044, 0xB884, 0x10A4, 0x1122, 0xFC20, 0x1040, 0x9440, 0x5848, 0x5084, 0x1DFE, 0xE082, 0x4000}},
    {37383, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x41FC, 0xB820, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x9450, 0x5850, 0x5088, 0x1C88, 0xE104, 0x4202}},
    {37384, 2, {0x1000, 0x1000, 0x29FE, 0x2410, 0x4210, 0xBC20, 0x1020, 0x1068, 0xFE64, 0x10A2, 0x9522, 0x5820, 0x5020, 0x1E20, 0xF020, 0x4020}},
    {37385, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4122, 0xB922, 0x1122, 0x1122, 0xFD52, 0x114A, 0x958A, 0x5902, 0x5102, 0x1D02, 0xE10A, 0x4104}},
    {37386, 2, {0x1000, 0x1040, 0x2820, 0x2410, 0x4050, 0xB840, 0x1040, 0x1144, 0xFD42, 0x1142, 0x9542, 0x5A44, 0x5044, 0x1C44, 0xE03C, 0x4000}},
    {37387, 2, {0x10A0, 0x10A0, 0x28A0, 0x2524, 0x4124, 0xBB28, 0x1530, 0x1120, 0xFD60, 0x11A0, 0x9520, 0x5922, 0x5122, 0x1D22, 0xE11E, 0x4100}},
    {37388, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4024, 0xB824, 0x1024, 0x1024, 0xFDFE, 0x1020, 0x9450, 0x5850, 0x5088, 0x1C88, 0xE104, 0x4202}},
    {37389, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4020, 0xB924, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x9424, 0x5820, 0x5022, 0x1C22, 0xE01E, 0x4000}},
    {37390, 2, {0x1080, 0x1080, 0x2880, 0x24FC, 0x4104, 0xB904, 0x1244, 0x1044, 0xFC84, 0x10A4, 0x9514, 0x59F4, 0x5014, 0x1C04, 0xE028, 0x4010}},
    {37391, 2, {0x1004, 0x11E4, 0x2824, 0x2424, 0x4024, 0xB9E4, 0x1104, 0x1104, 0xFD04, 0x11E4, 0x9424, 0x5824, 0x5024, 0x1C24, 0xE144, 0x4084}},
    {37392, 2, {0x1020, 0x1020, 0x2850, 0x2450, 0x4088, 0xB944, 0x1222, 0x1020, 0xFC00, 0x11FC, 0x9404, 0x5808, 0x5008, 0x1C10, 0xE010, 0x4020}},
    {37393, 2, {0x1008, 0x101C, 0x29E0, 0x2500, 0x4100, 0xB9FC, 0x1144, 0x1144, 0xFD44, 0x1128, 0x9528, 0x5910, 0x5110, 0x1E28, 0xE244, 0x4482}},
    {37394, 2, {0x1000, 0x13FC, 0x2884, 0x2488, 0x4088, 0xB890, 0x109C, 0x1084, 0xFD44, 0x1144, 0x9528, 0x5928, 0x5210, 0x1E28, 0xE444, 0x4182}},
    {37395, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x4020, 0xB820, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE1FC, 0x4000}},
    {37396, 2, {0x1020, 0x1020, 0x2820, 0x24A8, 0x42A4, 0xBCA2, 0x1122, 0x1120, 0xFE24, 0x1024, 0x9428, 0x5808, 0x5010, 0x1E20, 0xF0C0, 0x4300}},
    {37397, 2, {0x1000, 0x11F8, 0x2848, 0x2448, 0x4048, 0xB848, 0x1048, 0x11F8, 0xFC88, 0x1088, 0x9488, 0x5888, 0x5088, 0x1C88, 0xE3FE, 0x4000}},
    {37398, 2, {0x1010, 0x1090, 0x2890, 0x2488, 0x4108, 0xB904, 0x1204, 0x15FA, 0xFC88, 0x1088, 0x9488, 0x5888, 0x5108, 0x1D08, 0xE228, 0x4410}},
    {37399, 2, {0x1040, 0x1040, 0x2880, 0x2488, 0x4104, 0xBBFE, 0x1002, 0x1090, 0xFC90, 0x1090, 0x9490, 0x5890, 0x5112, 0x1D12, 0xE20E, 0x4400}},
    {37400, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4020, 0xB820, 0x1020, 0x11FC, 0xFC84, 0x1088, 0x9448, 0x5850, 0x5020, 0x1C50, 0xE188, 0x4606}},
    {37401, 2, {0x1020, 0x1020, 0x2820, 0x243E, 0x4220, 0xBC20, 0x1020, 0x11FC, 0xFE84, 0x1088, 0x9448, 0x5850, 0x5020, 0x1E50, 0xF188, 0x4606}},
    {37402, 2, {0x1010, 0x1110, 0x2910, 0x2512, 0x4112, 0xB914, 0x11D8, 0x1110, 0xFD10, 0x1110, 0x9510, 0x5912, 0x5152, 0x1D92, 0xE10E, 0x4000}},
    {37403, 2, {0x1048, 0x1044, 0x2844, 0x2440, 0x425E, 0xBDE0, 0x1040, 0x1044, 0xFE44, 0x1048, 0x9430, 0x5822, 0x5052, 0x1E8A, 0xF106, 0x4002}},
    {37404, 2, {0x1040, 0x1040, 0x2840, 0x2440, 0x43FE, 0xB840, 0x1080, 0x1090, 0xFC90, 0x1120, 0x9520, 0x5A48, 0x5244, 0x1C84, 0xE1FE, 0x4082}},
    {37405, 2, {0x1010, 0x1010, 0x2890, 0x2490, 0x42FE, 0xBC90, 0x1110, 0x1010, 0xFE10, 0x11FE, 0x9410, 0x5810, 0x5010, 0x1E10, 0xF010, 0x4010}},
    {37406, 2, {0x1040, 0x1040, 0x2880, 0x24FC, 0x4104, 0xBA04, 0x1084, 0x1044, 0xFC44, 0x1014, 0x9424, 0x5844, 0x5184, 0x1C04, 0xE028, 0x4010}},
    {37407, 2, {0x1000, 0x11FC, 0x2824, 0x2424, 0x4224, 0xBDFC, 0x1120, 0x1120, 0xFF20, 0x11FE, 0x9422, 0x5822, 0x502A, 0x1E24, 0xF020, 0x4020}},
    {37408, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x4288, 0xBD06, 0x1200, 0x11FC, 0xFE84, 0x1084, 0x9448, 0x5850, 0x5020, 0x1E50, 0xF088, 0x4306}},
    {37409, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x41FC, 0xB924, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x9524, 0x5820, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37410, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x41FE, 0xB820, 0x1070, 0x1070, 0xFCA8, 0x10A8, 0x9524, 0x5924, 0x5222, 0x1C20, 0xE020, 0x4020}},
    {37411, 2, {0x1000, 0x10FE, 0x2810, 0x2410, 0x4290, 0xBC9E, 0x1090, 0x1090, 0xFE90, 0x10FE, 0x9402, 0x5802, 0x5002, 0x1E02, 0xF014, 0x4008}},
    {37412, 2, {0x1000, 0x1000, 0x28FC, 0x2484, 0x4284, 0xBC84, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x9484, 0x5884, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37413, 2, {0x1020, 0x1020, 0x2820, 0x2424, 0x42A4, 0xBCA4, 0x10A8, 0x1120, 0xFE20, 0x1050, 0x9450, 0x5850, 0x5088, 0x1E88, 0xF104, 0x4202}},
    {37414, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x4220, 0xBDFE, 0x1020, 0x1020, 0xFE20, 0x1050, 0x9450, 0x5850, 0x5088, 0x1EC8, 0xF124, 0x4202}},
    {37415, 2, {0x1040, 0x1020, 0x2800, 0x25FC, 0x4000, 0xB800, 0x10F0, 0x1090, 0xFC90, 0x1090, 0x9490, 0x5892, 0x5092, 0x1D12, 0xE10E, 0x4200}},
    {37416, 2, {0x1000, 0x1000, 0x29FC, 0x2400, 0x4000, 0xB800, 0x13FE, 0x1090, 0xFC90, 0x1090, 0x9490, 0x5890, 0x5112, 0x1D12, 0xE20E, 0x4400}},
    {37417, 2, {0x1020, 0x1010, 0x2810, 0x24FE, 0x4282, 0xBC82, 0x1082, 0x10FE, 0xFE82, 0x1080, 0x9480, 0x5880, 0x5080, 0x1F00, 0xF100, 0x4200}},
    {37418, 2, {0x1000, 0x11FE, 0x2900, 0x2500, 0x4178, 0xB948, 0x1148, 0x1148, 0xFD48, 0x1168, 0x9550, 0x5942, 0x5142, 0x1D42, 0xE23E, 0x4400}},
    {37419, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x4288, 0xBC88, 0x1088, 0x1088, 0xFE50, 0x1050, 0x9420, 0x5820, 0x5050, 0x1E88, 0xF104, 0x4202}},
    {37420, 2, {0x1000, 0x10FC, 0x2884, 0x2484, 0x4284, 0xBC84, 0x10FC, 0x1090, 0xFE90, 0x1090, 0x9490, 0x5888, 0x5088, 0x1E84, 0xF104, 0x4202}},
    {37421, 2, {0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF17C, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0x540, 0xFFFE, 0x0}},
    {37422, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4104, 0xB9FC, 0x1100, 0x1140, 0xFD44, 0x1148, 0x9570, 0x5940, 0x5142, 0x1E42, 0xE23E, 0x4400}},
    {37423, 2, {0x1020, 0x1020, 0x2924, 0x2524, 0x4124, 0xB924, 0x11FC, 0x1020, 0xFC20, 0x1124, 0x9524, 0x5924, 0x5124, 0x1D24, 0xE1FC, 0x4004}},
    {37424, 2, {0x1040, 0x1020, 0x2800, 0x25FE, 0x4020, 0xB820, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x9524, 0x5924, 0x5134, 0x1D28, 0xE020, 0x4020}},
    {37425, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4104, 0xB9FC, 0x1120, 0x1120, 0xFDFE, 0x1120, 0x9520, 0x5910, 0x5112, 0x1D4A, 0xE186, 0x4102}},
    {37426, 2, {0x1008, 0x101C, 0x29F0, 0x2550, 0x4150, 0xB950, 0x1150, 0x1150, 0xFD50, 0x1148, 0x9548, 0x5968, 0x5154, 0x1E74, 0xE252, 0x4400}},
    {37427, 2, {0x1000, 0x1000, 0x2BFE, 0x2408, 0x4008, 0xB9E8, 0x1128, 0x1128, 0xFD28, 0x1128, 0x95E8, 0x5928, 0x5008, 0x1C08, 0xE028, 0x4010}},
    {37428, 2, {0x1020, 0x1020, 0x2850, 0x2450, 0x4088, 0xB924, 0x1212, 0x1010, 0xFDFC, 0x1004, 0x9408, 0x5888, 0x5050, 0x1C20, 0xE010, 0x4010}},
    {37429, 2, {0x1000, 0x11FE, 0x2820, 0x2420, 0x4020, 0xB9FE, 0x1122, 0x1122, 0xFD52, 0x114A, 0x958A, 0x5902, 0x5102, 0x1D02, 0xE10A, 0x4104}},
    {37430, 2, {0x1020, 0x1020, 0x2820, 0x2440, 0x4248, 0xBC84, 0x11FE, 0x1082, 0xFE00, 0x10FC, 0x9484, 0x5884, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37431, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x43FE, 0xB820, 0x1020, 0x1020, 0xFDFC, 0x1104, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37432, 2, {0x1050, 0x1048, 0x2848, 0x2440, 0x43FE, 0xBC40, 0x1040, 0x107C, 0xFEA4, 0x10A4, 0x94A8, 0x5928, 0x5110, 0x1E28, 0xF044, 0x4082}},
    {37433, 2, {0x1010, 0x1010, 0x2810, 0x25FE, 0x4112, 0xB914, 0x1110, 0x11FC, 0xFD44, 0x1144, 0x9528, 0x5928, 0x5110, 0x1E28, 0xE244, 0x4482}},
    {37434, 2, {0x1000, 0x1000, 0x29FC, 0x2420, 0x4020, 0xB820, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x9428, 0x5824, 0x5024, 0x1C20, 0xE3FE, 0x4000}},
    {37435, 2, {0x1000, 0x10FC, 0x2884, 0x2484, 0x4284, 0xBC84, 0x10FC, 0x1080, 0xFE80, 0x10FE, 0x9482, 0x5882, 0x5082, 0x1E82, 0xF0FE, 0x4082}},
    {37436, 2, {0x1080, 0x1080, 0x2880, 0x24FE, 0x4340, 0xBD40, 0x1240, 0x107C, 0xFE40, 0x1040, 0x9440, 0x587E, 0x5040, 0x1E40, 0xF040, 0x4040}},
    {37437, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4040, 0xB850, 0x1090, 0x10FE, 0xFD92, 0x1292, 0x9492, 0x5892, 0x509A, 0x1C94, 0xE010, 0x4010}},
    {37438, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x41FC, 0xB924, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x9524, 0x5924, 0x5124, 0x1D24, 0xE1FC, 0x4104}},
    {37439, 2, {0x1000, 0x1000, 0x29FC, 0x2524, 0x4124, 0xB924, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x9524, 0x5924, 0x5124, 0x1D24, 0xE1FC, 0x4104}},
    {37440, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x4124, 0xB9FC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x9524, 0x5820, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37441, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4104, 0xBA12, 0x1020, 0x1040, 0xFD88, 0x1010, 0x9420, 0x5844, 0x5188, 0x1C10, 0xE060, 0x4380}},
    {37442, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4124, 0xB924, 0x1124, 0x11FC, 0xFD24, 0x1020, 0x9520, 0x58C0, 0x5040, 0x1CA0, 0xE118, 0x4206}},
    {37443, 2, {0x1040, 0x1040, 0x2880, 0x24FC, 0x4120, 0xBA20, 0x1020, 0x1020, 0xFDFE, 0x1020, 0x9450, 0x5850, 0x5088, 0x1C88, 0xE104, 0x4202}},
    {37444, 2, {0x1020, 0x1020, 0x2920, 0x2520, 0x41FC, 0xB920, 0x1220, 0x1020, 0xFDFE, 0x1020, 0x9450, 0x5850, 0x5088, 0x1C88, 0xE104, 0x4202}},
    {37445, 2, {0x1000, 0x10FE, 0x2880, 0x2480, 0x4280, 0xBCFC, 0x1084, 0x1084, 0xFE84, 0x1084, 0x94FC, 0x5880, 0x5080, 0x1E80, 0xF0FE, 0x4000}},
    {37446, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x403E, 0xB820, 0x1020, 0x1020, 0xFDFC, 0x1104, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37447, 2, {0x1080, 0x1080, 0x28FE, 0x2500, 0x4220, 0x7D20, 0x912C, 0x1174, 0x7BA4, 0x1124, 0x9534, 0x5928, 0x5122, 0x1D02, 0xE0FE, 0x4000}},
    {37448, 2, {0x1020, 0x1010, 0x2810, 0x25FE, 0x4102, 0xBA04, 0x1080, 0x1088, 0xFC90, 0x10A0, 0x94C0, 0x5882, 0x5082, 0x1C82, 0xE07E, 0x4000}},
    {37449, 2, {0x1020, 0x1010, 0x2810, 0x25FE, 0x4020, 0xB820, 0x1044, 0x1084, 0xFDF8, 0x1010, 0x9420, 0x5844, 0x5082, 0x1DFE, 0xE082, 0x4000}},
    {37450, 2, {0x1000, 0x11FC, 0x2844, 0x2444, 0x4044, 0xB844, 0x1094, 0x1088, 0xFD00, 0x10FC, 0x9484, 0x5884, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37451, 2, {0x1080, 0x1080, 0x29FC, 0x2504, 0x4204, 0xBDF4, 0x1114, 0x1114, 0xFD14, 0x11F4, 0x9504, 0x5928, 0x5112, 0x1D02, 0xE0FE, 0x4000}},
    {37452, 2, {0x101C, 0x11E0, 0x2820, 0x2420, 0x4020, 0xBBFE, 0x1020, 0x1070, 0xFC70, 0x10A8, 0x94A8, 0x5924, 0x5222, 0x1C20, 0xE020, 0x4020}},
    {37453, 2, {0x1040, 0x1020, 0x2828, 0x2408, 0x4008, 0xB848, 0x1054, 0x1152, 0xFD52, 0x1160, 0x9660, 0x5844, 0x50C4, 0x1D44, 0xE23C, 0x4000}},
    {37454, 2, {0x1020, 0x1020, 0x2920, 0x2520, 0x41FC, 0xB920, 0x1220, 0x1020, 0xFC20, 0x11FC, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37455, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x4088, 0xB8F8, 0x1088, 0x1088, 0xFC88, 0x10F8, 0x9488, 0x5888, 0x5088, 0x1C88, 0xE3FE, 0x4000}},
    {37456, 2, {0x1000, 0x1000, 0x2BFE, 0x2420, 0x4020, 0xB840, 0x1040, 0x10FC, 0xFD84, 0x1284, 0x9484, 0x5884, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37457, 2, {0x1010, 0x1010, 0x2820, 0x24FC, 0x4284, 0xBC84, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x9484, 0x5884, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37458, 2, {0x1040, 0x1020, 0x2800, 0x25FE, 0x4020, 0xB820, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37459, 2, {0x1080, 0x1080, 0x2880, 0x24FE, 0x4120, 0xBA20, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x9524, 0x5924, 0x5134, 0x1D28, 0xE020, 0x4020}},
    {37460, 2, {0x1000, 0x11FE, 0x2910, 0x2510, 0x4110, 0xB97C, 0x1154, 0x1154, 0xFD54, 0x1154, 0x9554, 0x595C, 0x5110, 0x1D10, 0xE1FE, 0x4000}},
    {37461, 2, {0x1000, 0x10FE, 0x2880, 0x2480, 0x42BC, 0xBCA4, 0x10A4, 0x10A4, 0xFEA4, 0x10A4, 0x94BC, 0x58A4, 0x5080, 0x1E80, 0xF0FE, 0x4000}},
    {37462, 2, {0x1040, 0x1040, 0x287C, 0x2484, 0x4188, 0xBA50, 0x1020, 0x1050, 0xFC88, 0x1306, 0x9460, 0x5810, 0x5008, 0x1CC0, 0xE020, 0x4010}},
    {37463, 2, {0x1088, 0x1088, 0x2888, 0x2488, 0x43FE, 0xBC88, 0x1088, 0x1088, 0xFE88, 0x10F8, 0x9488, 0x5888, 0x5088, 0x1E88, 0xF0F8, 0x4088}},
    {37464, 2, {0x1050, 0x1050, 0x2850, 0x25FC, 0x4054, 0xB854, 0x11FC, 0x1150, 0xFD50, 0x11FE, 0x9452, 0x5852, 0x505A, 0x1C94, 0xE090, 0x4110}},
    {37465, 2, {0x1000, 0x1000, 0x29FC, 0x2504, 0x4104, 0xB904, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x9400, 0x5890, 0x5088, 0x1D04, 0xE202, 0x4402}},
    {37466, 2, {0x2000, 0x2080, 0x531E, 0x4A52, 0x8252, 0x7A52, 0x2252, 0x2252, 0xFA52, 0x22D2, 0x2B5A, 0xB254, 0x6090, 0x3890, 0xE110, 0x210}},
    {37467, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x4088, 0xB888, 0x1106, 0x1200, 0xFC00, 0x10FC, 0x9484, 0x5884, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37468, 2, {0x1088, 0x1088, 0x2888, 0x2508, 0x417E, 0xBB08, 0x1508, 0x1148, 0xFD28, 0x1128, 0x9508, 0x5908, 0x5108, 0x1D08, 0xE128, 0x4110}},
    {37469, 2, {0x1020, 0x1010, 0x2810, 0x2400, 0x41FE, 0xB800, 0x1004, 0x1084, 0xFC84, 0x1048, 0x9448, 0x5848, 0x5050, 0x1C10, 0xE1FE, 0x4000}},
    {37470, 2, {0x2028, 0x2024, 0x5024, 0x4820, 0x83FE, 0x7A20, 0x2224, 0x2224, 0xFA24, 0x2228, 0x2A28, 0xB290, 0x6312, 0x3A2A, 0xE046, 0x82}},
    {37471, 2, {0x1000, 0x10FE, 0x2810, 0x2410, 0x4220, 0xBC20, 0x1068, 0x1064, 0xFEA2, 0x1122, 0x9420, 0x5820, 0x5020, 0x1E00, 0xF1FE, 0x4000}},
    {37472, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4124, 0xB924, 0x1124, 0x1124, 0xFD24, 0x13FE, 0x9420, 0x5850, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37473, 2, {0x1020, 0x1020, 0x2924, 0x24A4, 0x40A8, 0xB820, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x97FE, 0x5820, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37474, 2, {0x1020, 0x1020, 0x2820, 0x2420, 0x43FE, 0xB820, 0x1070, 0x1070, 0xFCA8, 0x10A8, 0x9524, 0x5AFA, 0x5420, 0x1C20, 0xE020, 0x4020}},
    {37475, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4020, 0xB820, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x9440, 0x5848, 0x5084, 0x1DFE, 0xE082, 0x4000}},
    {37476, 2, {0x1080, 0x1080, 0x2900, 0x25FC, 0x4204, 0xBC04, 0x11E4, 0x1124, 0xFD24, 0x1124, 0x95E4, 0x5924, 0x5004, 0x1C04, 0xE028, 0x4010}},
    {37477, 2, {0x1020, 0x1028, 0x2824, 0x2420, 0x41FC, 0xB820, 0x1070, 0x1070, 0xFCA8, 0x10A8, 0x9524, 0x5924, 0x5222, 0x1C20, 0xE020, 0x4020}},
    {37478, 2, {0x1000, 0x1000, 0x29FE, 0x2410, 0x4010, 0xB810, 0x1090, 0x1090, 0xFC9E, 0x1090, 0x9490, 0x5890, 0x5090, 0x1C90, 0xE3FE, 0x4000}},
    {37479, 2, {0x1000, 0x10FC, 0x2884, 0x24A4, 0x4094, 0xB884, 0x1084, 0x13FE, 0xFD04, 0x1144, 0x9524, 0x5904, 0x51FE, 0x1C04, 0xE028, 0x4010}},
    {37480, 2, {0x1080, 0x1080, 0x2880, 0x24FE, 0x4102, 0xB904, 0x1220, 0x1020, 0xFCA8, 0x10A4, 0x9524, 0x5922, 0x5222, 0x1C20, 0xE0A0, 0x4040}},
    {37481, 2, {0x1020, 0x1020, 0x2850, 0x2450, 0x4088, 0xB904, 0x1222, 0x1020, 0xFCA8, 0x10A4, 0x9524, 0x5922, 0x5222, 0x1C20, 0xE0A0, 0x4040}},
    {37482, 2, {0x1024, 0x10A4, 0x28A4, 0x24A4, 0x42A4, 0xBDFE, 0x10A4, 0x10A4, 0xFEA4, 0x10A4, 0x94BC, 0x5880, 0x5080, 0x1E80, 0xF0FE, 0x4000}},
    {37483, 2, {0x1100, 0x1100, 0x2900, 0x251C, 0x43D4, 0xB954, 0x1154, 0x1154, 0xFD54, 0x1154, 0x9554, 0x5954, 0x5154, 0x1E5C, 0xE354, 0x4480}},
    {37484, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4104, 0xB9FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37485, 2, {0x1000, 0x1000, 0x28FC, 0x2484, 0x4284, 0xBC84, 0x10FC, 0x1084, 0xFE84, 0x1084, 0x94FC, 0x5884, 0x5000, 0x1E00, 0xF1FE, 0x4000}},
    {37486, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4124, 0xB924, 0x1124, 0x11FC, 0xFD24, 0x1124, 0x9524, 0x59FC, 0x5124, 0x1C20, 0xE020, 0x4020}},
    {37487, 2, {0x1008, 0x1048, 0x2928, 0x2528, 0x4128, 0xB908, 0x1108, 0x1108, 0xFD08, 0x1108, 0x9548, 0x5994, 0x5114, 0x1C22, 0xE042, 0x4082}},
    {37488, 2, {0x1000, 0x11FC, 0x2804, 0x2404, 0x43F4, 0xB804, 0x1004, 0x11E4, 0xFD24, 0x1124, 0x9524, 0x59E4, 0x5004, 0x1C04, 0xE014, 0x4008}},
    {37489, 2, {0x1020, 0x1010, 0x2810, 0x25FE, 0x4100, 0xB910, 0x1110, 0x1110, 0xFD10, 0x1120, 0x9528, 0x5924, 0x5244, 0x1EFE, 0xE442, 0x4800}},
    {37490, 2, {0x1020, 0x1020, 0x2820, 0x243C, 0x4220, 0xBC20, 0x1020, 0x11FE, 0xFE20, 0x1020, 0x9428, 0x5824, 0x5022, 0x1E20, 0xF020, 0x4020}},
    {37491, 2, {0x1090, 0x1090, 0x2890, 0x2490, 0x4090, 0xBB92, 0x1094, 0x1098, 0xFC90, 0x1090, 0x9490, 0x5992, 0x5292, 0x1C92, 0xE08E, 0x4080}},
    {37492, 2, {0x440, 0x2440, 0x247C, 0x2490, 0x2488, 0x2508, 0x280, 0xC60, 0x3018, 0xCFE6, 0x100, 0x3FF8, 0x920, 0x540, 0x7FFC, 0x0}},
    {37493, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4124, 0xB924, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x9420, 0x5828, 0x5024, 0x1DFE, 0xE082, 0x4000}},
    {37494, 2, {0x1002, 0x13E2, 0x2942, 0x254A, 0x414A, 0xB94A, 0x114A, 0x13EA, 0xFD4A, 0x114A, 0x954A, 0x594A, 0x5142, 0x1E42, 0xE24A, 0x4444}},
    {37495, 2, {0x1048, 0x1048, 0x2848, 0x2448, 0x41FE, 0xB848, 0x1048, 0x1048, 0xFC48, 0x11FE, 0x9400, 0x5848, 0x5044, 0x1C84, 0xE082, 0x4102}},
    {37496, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x4000, 0xB888, 0x1104, 0x1202, 0xFC88, 0x1088, 0x9450, 0x5850, 0x5020, 0x1C50, 0xE088, 0x4306}},
    {37497, 2, {0x1020, 0x1020, 0x287C, 0x2484, 0x4148, 0xB830, 0x1020, 0x1048, 0xFD90, 0x103E, 0x9442, 0x59A4, 0x5018, 0x1C10, 0xE060, 0x4180}},
    {37498, 2, {0x1000, 0x13FE, 0x2888, 0x2488, 0x40F8, 0xB888, 0x1088, 0x10F8, 0xFC88, 0x1088, 0x948E, 0x5BF8, 0x5008, 0x1C08, 0xE008, 0x4008}},
    {37499, 2, {0x1040, 0x1040, 0x2878, 0x2488, 0x4150, 0xB820, 0x1050, 0x1088, 0xFD06, 0x10F8, 0x9488, 0x5888, 0x5088, 0x1C88, 0xE0F8, 0x4088}},
    {37500, 2, {0x1104, 0x1084, 0x2888, 0x2400, 0x41FE, 0xB888, 0x1088, 0x1088, 0xFC88, 0x13FE, 0x9488, 0x5888, 0x5108, 0x1D08, 0xE208, 0x4408}},
    {37501, 2, {0x2010, 0x2014, 0x5012, 0x4810, 0x83FE, 0x7810, 0x2010, 0x21D0, 0xF890, 0x2090, 0x2890, 0xB088, 0x60EA, 0x3B8A, 0xE006, 0x2}},
    {37502, 2, {0x1020, 0x1020, 0x2848, 0x2484, 0x41FE, 0xB812, 0x1090, 0x1090, 0xFCFE, 0x1110, 0x9410, 0x59FE, 0x5010, 0x1C10, 0xE010, 0x4010}},
    {37503, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4104, 0xBA02, 0x11FC, 0x1000, 0xFC00, 0x11FC, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37504, 2, {0x1000, 0x11F8, 0x2908, 0x2508, 0x41F8, 0xB908, 0x1108, 0x11F8, 0xFD44, 0x1148, 0x9530, 0x5920, 0x5110, 0x1D48, 0xE186, 0x4100}},
    {37505, 2, {0x1040, 0x1040, 0x2880, 0x24FE, 0x4102, 0xBA02, 0x11F2, 0x1112, 0xFD12, 0x11F2, 0x9512, 0x5912, 0x51F2, 0x1C02, 0xE014, 0x4008}},
    {37506, 2, {0x1082, 0x1092, 0x2892, 0x2492, 0x4092, 0xB892, 0x12DA, 0x12B6, 0xFC92, 0x1092, 0x9492, 0x5892, 0x5092, 0x1D12, 0xE102, 0x4202}},
    {37507, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x4040, 0xB888, 0x1104, 0x11FE, 0xFC92, 0x1090, 0x9490, 0x5890, 0x5112, 0x1D12, 0xE20E, 0x4400}},
    {37508, 2, {0x1020, 0x1020, 0x2840, 0x25FC, 0x4104, 0xB904, 0x1174, 0x1154, 0xFD54, 0x1154, 0x9554, 0x5974, 0x5104, 0x1D04, 0xE114, 0x4108}},
    {37509, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4174, 0xB904, 0x1104, 0x1174, 0xFD54, 0x1154, 0x9554, 0x5974, 0x5104, 0x1D04, 0xE114, 0x4108}},
    {37510, 2, {0x1000, 0x13FE, 0x2820, 0x2420, 0x4040, 0xB9FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37511, 2, {0x1020, 0x1020, 0x29FE, 0x2420, 0x4020, 0xB8FC, 0x1020, 0x1020, 0xFDFE, 0x1070, 0x94A8, 0x58A8, 0x5124, 0x1E22, 0xE420, 0x4020}},
    {37512, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x4020, 0xB820, 0x13FE, 0x1000, 0xFC20, 0x1020, 0x95FC, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37513, 2, {0x1020, 0x1020, 0x29FC, 0x2424, 0x43FE, 0xB824, 0x11FC, 0x1020, 0xFC20, 0x11FC, 0x9420, 0x5820, 0x53FE, 0x1C20, 0xE020, 0x4020}},
    {37514, 2, {0x2028, 0x2024, 0x5024, 0x4820, 0x83FE, 0x7A20, 0x2224, 0x2224, 0xFA24, 0x23A8, 0x2A28, 0xB210, 0x6212, 0x3C2A, 0xE446, 0x882}},
    {37515, 2, {0x1080, 0x1088, 0x289C, 0x2570, 0x4110, 0xBB10, 0x1510, 0x1110, 0xFDFE, 0x1110, 0x9510, 0x5910, 0x5110, 0x1D10, 0xE17C, 0x4100}},
    {37516, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4040, 0xB840, 0x10BC, 0x1084, 0xFD88, 0x1288, 0x94FE, 0x5888, 0x5088, 0x1C88, 0xE0A8, 0x4090}},
    {37517, 2, {0x1000, 0x11FE, 0x2820, 0x2440, 0x4088, 0xB904, 0x11FE, 0x1022, 0xFC20, 0x1020, 0x95FE, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37518, 2, {0x0, 0x7DF0, 0x1110, 0x1190, 0x1152, 0x1D12, 0xE20E, 0x4500, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37519, 2, {0x1000, 0x11DC, 0x2954, 0x2554, 0x4154, 0xB954, 0x1154, 0x13FE, 0xFD54, 0x1154, 0x9554, 0x5954, 0x5154, 0x1ED4, 0xE224, 0x444C}},
    {37520, 2, {0x4, 0x7F84, 0x824, 0x1F24, 0x2124, 0x5224, 0xC24, 0x3104, 0xC6CC, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37521, 2, {0x1020, 0x1120, 0x2920, 0x25FC, 0x4120, 0xBA20, 0x1020, 0x13FE, 0xFC90, 0x1090, 0x9490, 0x5890, 0x5112, 0x1D12, 0xE212, 0x440E}},
    {37522, 2, {0x1000, 0x11DC, 0x2888, 0x2488, 0x4088, 0xB888, 0x1088, 0x13DE, 0xFC88, 0x1088, 0x9488, 0x5888, 0x5088, 0x1D08, 0xE108, 0x4208}},
    {37523, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4104, 0xBA02, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x95FC, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37524, 2, {0x1000, 0x13FE, 0x2810, 0x2420, 0x4068, 0xB8A4, 0x1322, 0x1020, 0xFC00, 0x1020, 0x9420, 0x5BFE, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37525, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x41FC, 0xB824, 0x11FC, 0x1120, 0xFDFE, 0x1022, 0x942A, 0x5854, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37526, 2, {0x1020, 0x1120, 0x2920, 0x25FC, 0x4120, 0xBA20, 0x1020, 0x13FE, 0xFC70, 0x10A8, 0x94A8, 0x5924, 0x5124, 0x1E22, 0xE020, 0x4020}},
    {37527, 2, {0x1008, 0x101C, 0x29E0, 0x2500, 0x4100, 0xB9FE, 0x1100, 0x1100, 0xFD7C, 0x1144, 0x9544, 0x5944, 0x5144, 0x1E7C, 0xE244, 0x4400}},
    {37528, 2, {0x1040, 0x1040, 0x28FC, 0x2504, 0x4288, 0xB850, 0x1020, 0x1040, 0xFC80, 0x11FC, 0x9684, 0x5884, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37529, 2, {0x2020, 0x2020, 0x53FE, 0x4850, 0x8088, 0x7904, 0x22FA, 0x2000, 0xFBFE, 0x2040, 0x2880, 0xB1FC, 0x6004, 0x3804, 0xE028, 0x10}},
    {37530, 2, {0x1050, 0x1050, 0x2850, 0x2552, 0x42D4, 0xBC58, 0x1050, 0x1058, 0xFED4, 0x1152, 0x9450, 0x5850, 0x5092, 0x1E92, 0xF112, 0x420E}},
    {37531, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x4020, 0xBBFE, 0x1020, 0x1020, 0xFC20, 0x11FC, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37532, 2, {0x1100, 0x1100, 0x229C, 0x4240, 0x8420, 0x1BC0, 0x213E, 0x6108, 0xAFC8, 0x2108, 0x2948, 0x2588, 0x2508, 0x21C8, 0x2E28, 0x2410}},
    {37533, 2, {0x1090, 0x1090, 0x2890, 0x2510, 0x41FE, 0xBB10, 0x1510, 0x1138, 0xFD38, 0x1154, 0x9554, 0x5992, 0x5110, 0x1D10, 0xE110, 0x4110}},
    {37534, 2, {0x1000, 0x1FF8, 0x2008, 0x5FC8, 0x9048, 0x1FC8, 0x1048, 0x1FE8, 0x110, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37535, 2, {0x1010, 0x1092, 0x2892, 0x2492, 0x42FE, 0xBC00, 0x1010, 0x1010, 0xFEFE, 0x1092, 0x9492, 0x589A, 0x5094, 0x1E10, 0xF010, 0x4010}},
    {37536, 2, {0x1040, 0x1040, 0x2844, 0x25F4, 0x4048, 0xB850, 0x13FE, 0x1040, 0xFC80, 0x1184, 0x9698, 0x58E0, 0x5082, 0x1C82, 0xE07E, 0x4000}},
    {37537, 2, {0x1020, 0x1020, 0x2820, 0x27FE, 0x4020, 0xB820, 0x11FC, 0x1000, 0xFC00, 0x11FC, 0x9504, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37538, 2, {0x2000, 0x21FC, 0x5000, 0x4800, 0x83FE, 0x7800, 0x2020, 0x2022, 0xFBB2, 0x20B4, 0x28A8, 0xB128, 0x6124, 0x3A24, 0xE4A2, 0x40}},
    {37539, 2, {0x2100, 0x2100, 0x5100, 0x491E, 0x87D2, 0x7A52, 0x2252, 0x2252, 0xFA52, 0x2492, 0x2A92, 0xB112, 0x6292, 0x3A5E, 0xE452, 0x800}},
    {37540, 2, {0x1020, 0x1020, 0x2924, 0x24A4, 0x40A8, 0xB820, 0x1020, 0x13FE, 0xFC70, 0x10A8, 0x94A8, 0x5924, 0x5124, 0x1E22, 0xE020, 0x4020}},
    {37541, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x4040, 0xB840, 0x10A0, 0x10A2, 0xFDA4, 0x1298, 0x9490, 0x5888, 0x5088, 0x1CA4, 0xE0C2, 0x4080}},
    {37542, 2, {0x1000, 0x11FE, 0x2902, 0x2522, 0x4122, 0xB922, 0x11FE, 0x1122, 0xFD22, 0x1152, 0x954A, 0x598A, 0x5102, 0x1D02, 0xE1FE, 0x4102}},
    {37543, 2, {0x1040, 0x1044, 0x2A44, 0x2548, 0x4150, 0xB840, 0x17FE, 0x1090, 0xFC90, 0x1090, 0x9490, 0x5892, 0x5112, 0x1D12, 0xE20E, 0x4400}},
    {37544, 2, {0x1040, 0x1020, 0x2820, 0x27FE, 0x4202, 0xBC44, 0x1040, 0x13FE, 0xFC88, 0x1088, 0x9508, 0x58D0, 0x5020, 0x1C50, 0xE088, 0x4304}},
    {37545, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x4020, 0xB9FC, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x9420, 0x5840, 0x5088, 0x1D04, 0xE3FE, 0x4102}},
    {37546, 2, {0x1020, 0x1020, 0x29FE, 0x2440, 0x4240, 0xBCFC, 0x1084, 0x1184, 0xFEFC, 0x1084, 0x9484, 0x58FC, 0x5084, 0x1E84, 0xF094, 0x4088}},
    {37547, 2, {0x1080, 0x1080, 0x28F8, 0x2508, 0x4210, 0xBDFC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x9500, 0x5900, 0x5102, 0x1D02, 0xE0FE, 0x4000}},
    {37548, 2, {0x1040, 0x1044, 0x29F4, 0x2448, 0x4050, 0xBBFE, 0x1040, 0x1080, 0xFDFE, 0x1240, 0x9480, 0x58FC, 0x5004, 0x1C04, 0xE028, 0x4010}},
    {37549, 2, {0x1028, 0x1024, 0x283E, 0x25E0, 0x4220, 0xBC3C, 0x11E0, 0x1020, 0xFE3E, 0x11E0, 0x9422, 0x5824, 0x5018, 0x1E6A, 0xF186, 0x4002}},
    {37550, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4444, 0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37551, 2, {0x1020, 0x1020, 0x2848, 0x2488, 0x41F0, 0xB820, 0x1040, 0x1088, 0xFDFC, 0x1024, 0x9420, 0x58A8, 0x5124, 0x1E22, 0xE0A0, 0x4040}},
    {37552, 2, {0x1088, 0x1088, 0x29FE, 0x2488, 0x4088, 0xB800, 0x1104, 0x1104, 0xFC88, 0x1088, 0x9450, 0x5820, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37553, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x40F8, 0xB820, 0x1020, 0x11FC, 0xFD24, 0x1124, 0x9524, 0x5924, 0x5134, 0x1D28, 0xE020, 0x4020}},
    {37554, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1000, 0x1000, 0xFDFC, 0x1020, 0x9420, 0x5BFE, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37555, 2, {0x1010, 0x1090, 0x2888, 0x2508, 0x4204, 0xBDFA, 0x1108, 0x1108, 0xFD08, 0x11F8, 0x9490, 0x5890, 0x5090, 0x1D12, 0xE212, 0x440E}},
    {37556, 2, {0x80C, 0x8F0, 0x7E80, 0x880, 0xEFE, 0x7888, 0x888, 0x2908, 0x16C8, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37557, 2, {0x1020, 0x1040, 0x28F8, 0x2488, 0x40F8, 0xB888, 0x10F8, 0x108A, 0xFC8C, 0x13F8, 0x9418, 0x5828, 0x5048, 0x1C88, 0xE328, 0x4010}},
    {37558, 2, {0x1028, 0x1024, 0x2824, 0x2420, 0x43FE, 0xB820, 0x1120, 0x10B2, 0xFCB4, 0x1068, 0x94A8, 0x5924, 0x5222, 0x1C20, 0xE0A0, 0x4040}},
    {37559, 2, {0x1010, 0x1092, 0x2852, 0x2454, 0x4210, 0xBCFE, 0x1082, 0x1082, 0xFEFE, 0x1082, 0x9482, 0x58FE, 0x5082, 0x1E82, 0xF08A, 0x4084}},
    {37560, 2, {0x1000, 0x13FE, 0x2920, 0x2528, 0x4128, 0xB9E8, 0x1128, 0x1128, 0xFD28, 0x11E8, 0x9528, 0x592A, 0x513A, 0x1FEA, 0xE026, 0x4020}},
    {37561, 2, {0x101C, 0x11E0, 0x2820, 0x2420, 0x43FE, 0xB8A8, 0x1124, 0x1222, 0xFDF8, 0x1088, 0x9490, 0x58BE, 0x5102, 0x1D02, 0xE214, 0x4408}},
    {37562, 2, {0x2420, 0x2420, 0x3DFC, 0x420, 0xFC20, 0x24F8, 0x4500, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0x540, 0xFFFE, 0x0}},
    {37563, 2, {0x1108, 0x1088, 0x2890, 0x25FC, 0x4024, 0xB824, 0x11FC, 0x1120, 0xFD20, 0x11FE, 0x9462, 0x58A2, 0x512A, 0x1E24, 0xE420, 0x4020}},
    {37564, 2, {0x1020, 0x1020, 0x28A8, 0x24A8, 0x40A8, 0xB974, 0x1222, 0x1020, 0xFC20, 0x11FC, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37565, 2, {0x101C, 0x13E0, 0x2A20, 0x2620, 0x43FE, 0xBA20, 0x1290, 0x130A, 0xFE06, 0x11FA, 0x9508, 0x5908, 0x5108, 0x1D08, 0xE1F8, 0x4108}},
    {37566, 2, {0x2000, 0x21FC, 0x5020, 0x4820, 0x83FE, 0x7800, 0x2020, 0x2022, 0xFBB2, 0x20B4, 0x28A8, 0xB128, 0x6124, 0x3A24, 0xE4A2, 0x40}},
    {37567, 2, {0x1000, 0x11F8, 0x2808, 0x24D0, 0x4020, 0xB9FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x9524, 0x59FC, 0x5124, 0x1D24, 0xE124, 0x410C}},
    {37568, 2, {0x1000, 0x11FE, 0x2800, 0x2400, 0x42FC, 0xBC84, 0x1084, 0x1084, 0xFEFC, 0x1000, 0x9484, 0x5844, 0x5048, 0x1E00, 0xF1FE, 0x4000}},
    {37569, 2, {0x1000, 0x10FC, 0x2884, 0x2484, 0x4284, 0xBC84, 0x10FC, 0x1000, 0xFE00, 0x11FE, 0x9502, 0x5902, 0x5102, 0x1F02, 0xF1FE, 0x4102}},
    {37570, 2, {0x1080, 0x1080, 0x28FE, 0x2500, 0x42FC, 0xB884, 0x10A4, 0x1094, 0xFDFE, 0x1084, 0x9524, 0x5914, 0x51FE, 0x1C04, 0xE028, 0x4010}},
    {37571, 2, {0x1040, 0x1020, 0x29FC, 0x2504, 0x4104, 0xB9FC, 0x1104, 0x1104, 0xFDFC, 0x1120, 0x9522, 0x5914, 0x5108, 0x1D44, 0xE182, 0x4100}},
    {37572, 2, {0x1000, 0x13FE, 0x2820, 0x24A4, 0x4128, 0xB850, 0x1088, 0x1104, 0xFC00, 0x10FC, 0x9484, 0x5888, 0x5050, 0x1C20, 0xE0D8, 0x4706}},
    {37573, 2, {0x1040, 0x1020, 0x29FC, 0x2400, 0x4088, 0xB850, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x95FC, 0x5820, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37574, 2, {0x1080, 0x10FC, 0x7D04, 0x1044, 0x1034, 0x1CC4, 0xE114, 0x6C8, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0x540, 0xFFFE, 0x0}},
    {37575, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x41FC, 0xB904, 0x1104, 0x11FC, 0xFD04, 0x1104, 0x95FC, 0x5800, 0x5088, 0x1C84, 0xE102, 0x4202}},
    {37576, 2, {0x101C, 0x9E0, 0x4020, 0x2420, 0xBFE, 0x7050, 0x1088, 0x1104, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37577, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x4020, 0xB9FC, 0x1124, 0x1124, 0xFDFC, 0x1020, 0x9470, 0x58A8, 0x5124, 0x1E22, 0xE020, 0x4020}},
    {37578, 2, {0x1088, 0x1084, 0x2902, 0x2622, 0x4020, 0xB850, 0x1088, 0x1104, 0xFE02, 0x10FC, 0x9484, 0x5884, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37579, 2, {0x2004, 0x200E, 0x53B8, 0x4888, 0x8088, 0x7928, 0x212E, 0x23A8, 0xF8A8, 0x22A8, 0x2AA8, 0xB13E, 0x6100, 0x3A80, 0xE47E, 0x800}},
    {37580, 2, {0x2000, 0x2006, 0x53B8, 0x4888, 0x8088, 0x7908, 0x213E, 0x2388, 0xF888, 0x2288, 0x2A88, 0xB13E, 0x6100, 0x3A80, 0xE47E, 0x800}},
    {37581, 2, {0x2020, 0x2020, 0x53FE, 0x4820, 0x8020, 0x7BFE, 0x2202, 0x2404, 0xF9F8, 0x2010, 0x2820, 0xB3FE, 0x6020, 0x3820, 0xE0A0, 0x40}},
    {37582, 2, {0x1020, 0x1010, 0x29FE, 0x2502, 0x4204, 0xB8F8, 0x1000, 0x1000, 0xFDFE, 0x1050, 0x9450, 0x5850, 0x5092, 0x1C92, 0xE10E, 0x4200}},
    {37583, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x4020, 0xB8A8, 0x10A8, 0x10A8, 0xFD74, 0x1222, 0x9450, 0x5850, 0x5088, 0x1C88, 0xE104, 0x4202}},
    {37584, 2, {0x1040, 0x1020, 0x29FE, 0x2502, 0x4244, 0xB840, 0x11FE, 0x1040, 0xFC50, 0x1090, 0x9490, 0x58A0, 0x5124, 0x1D42, 0xE27E, 0x4422}},
    {37585, 2, {0x1040, 0x1040, 0x2888, 0x2504, 0x43FE, 0xB802, 0x1088, 0x1144, 0xFE42, 0x10F8, 0x9588, 0x5A50, 0x5020, 0x1C50, 0xE188, 0x4606}},
    {37586, 2, {0x1040, 0x1040, 0x28FC, 0x2488, 0x4150, 0xB820, 0x10D8, 0x1326, 0xFCF8, 0x1020, 0x94F8, 0x5820, 0x53FE, 0x1C20, 0xE020, 0x4020}},
    {37587, 2, {0x2042, 0x20E2, 0x5382, 0x488A, 0x808A, 0x788A, 0x23EA, 0x208A, 0xF98A, 0x21CA, 0x2AAA, 0xB282, 0x6482, 0x3882, 0xE08A, 0x84}},
    {37588, 2, {0x1080, 0x1080, 0x28F8, 0x2508, 0x4210, 0xB9FC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x9450, 0x5850, 0x5090, 0x1C92, 0xE112, 0x420E}},
    {37589, 2, {0x2020, 0x2020, 0x5020, 0x4BFE, 0x8020, 0x7820, 0x2020, 0x21FC, 0xF840, 0x2020, 0x28A4, 0xB282, 0x628A, 0x3A8A, 0xE478, 0x0}},
    {37590, 2, {0x1004, 0x101E, 0x29E0, 0x2422, 0x4112, 0xB894, 0x1080, 0x1020, 0xFDFE, 0x1044, 0x9484, 0x59C8, 0x5030, 0x1C28, 0xE0C4, 0x4302}},
    {37591, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x40F8, 0xB800, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x9504, 0x5904, 0x51FC, 0x1D04, 0xE114, 0x4108}},
    {37592, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x41FC, 0xB800, 0x1000, 0x11FC, 0xFC20, 0x1020, 0x97FE, 0x5820, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37593, 2, {0x1000, 0x11FE, 0x2820, 0x2420, 0x42FC, 0xBC44, 0x1044, 0x11FE, 0xFE00, 0x1000, 0x94FC, 0x5884, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37594, 2, {0x1040, 0x1040, 0x24FE, 0x6488, 0xA550, 0x2420, 0x2450, 0x2188, 0x6C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37595, 2, {0x2020, 0x2020, 0x51FC, 0x4924, 0x8124, 0x79FC, 0x2020, 0x2020, 0xFBFE, 0x2222, 0x2A22, 0xB3FE, 0x6222, 0x3820, 0xE020, 0x20}},
    {37596, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4104, 0xB904, 0x11FC, 0x1020, 0xFC20, 0x1120, 0x953E, 0x5920, 0x5120, 0x1EA0, 0xE47E, 0x4800}},
    {37597, 2, {0x1004, 0x101E, 0x29E0, 0x2422, 0x4112, 0xB894, 0x1080, 0x1008, 0xFDFE, 0x1008, 0x9488, 0x5848, 0x5048, 0x1C08, 0xE028, 0x4010}},
    {37598, 2, {0x1000, 0x11FE, 0x2800, 0x2492, 0x4124, 0xBA48, 0x1124, 0x1092, 0xFC00, 0x11FE, 0x9420, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37599, 2, {0x1000, 0x11FC, 0x2804, 0x24FC, 0x4004, 0xB9FC, 0x1000, 0x13FE, 0xFE02, 0x11F8, 0x9488, 0x5888, 0x5050, 0x1C20, 0xE0D8, 0x4306}},
    {37600, 2, {0x1000, 0x11FE, 0x2900, 0x2500, 0x417C, 0xB900, 0x1100, 0x11FE, 0xFD50, 0x1152, 0x9554, 0x5948, 0x5148, 0x1E44, 0xE252, 0x4460}},
    {37601, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4144, 0xBA22, 0x10F8, 0x1008, 0xFC10, 0x1020, 0x95FC, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37602, 2, {0x1008, 0x103C, 0x29C0, 0x2404, 0x4144, 0xB8A8, 0x1000, 0x11F8, 0xFC10, 0x1020, 0x97FE, 0x5820, 0x5020, 0x1C20, 0xE0A0, 0x4040}},
    {37603, 2, {0x2000, 0x27DE, 0x5092, 0x4C92, 0x8494, 0x7C94, 0x27D8, 0x2094, 0xF992, 0x2192, 0x2A92, 0xB29A, 0x6494, 0x3890, 0xE290, 0x110}},
    {37604, 2, {0x2008, 0x2008, 0x53C8, 0x4A48, 0x825E, 0x7A4A, 0x23CA, 0x224A, 0xFA4A, 0x23CA, 0x2A4A, 0xB24A, 0x626A, 0x3BD2, 0xE612, 0x26}},
    {37605, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x4104, 0xB9FC, 0x1000, 0x1000, 0xFDFE, 0x1020, 0x9420, 0x59FC, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37606, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x41FC, 0xB900, 0x1100, 0x11FE, 0xFD02, 0x117A, 0x954A, 0x594A, 0x527A, 0x1E02, 0xE414, 0x4008}},
    {37607, 2, {0x1000, 0x11F8, 0x2908, 0x2508, 0x41F8, 0xB908, 0x1108, 0x11F8, 0xFD08, 0x1108, 0x95F8, 0x5890, 0x5090, 0x1D12, 0xE212, 0x440E}},
    {37608, 2, {0x2010, 0x20D8, 0x5394, 0x4894, 0x8090, 0x7BFE, 0x2090, 0x2094, 0xF894, 0x20D8, 0x2B98, 0xB090, 0x60AA, 0x38CA, 0xE286, 0x102}},
    {37609, 2, {0x1048, 0x1048, 0x29FE, 0x2448, 0x4248, 0xBC20, 0x1010, 0x11FE, 0xFE80, 0x1080, 0x9480, 0x5880, 0x5080, 0x1E80, 0xF0FC, 0x4000}},
    {37610, 2, {0x1028, 0x1024, 0x2BFE, 0x2420, 0x4020, 0xB9FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x9524, 0x59FC, 0x5124, 0x1D24, 0xE124, 0x410C}},
    {37611, 2, {0x304, 0x3C04, 0x424, 0x7FA4, 0xE24, 0x1524, 0x6484, 0x50C, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37612, 2, {0x101C, 0x11E0, 0xFD00, 0x11FC, 0x1D04, 0xF168, 0x1210, 0x3128, 0x6C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37613, 2, {0x1104, 0x1084, 0x2888, 0x2410, 0x41FC, 0xB904, 0x1104, 0x1104, 0xFDFC, 0x1050, 0x9450, 0x5890, 0x5092, 0x1D12, 0xE20E, 0x4400}},
    {37614, 2, {0x2028, 0x2024, 0x5024, 0x4820, 0x83FE, 0x7A20, 0x2224, 0x2224, 0xFBA4, 0x22A8, 0x2AA8, 0xB290, 0x6292, 0x3DAA, 0xE446, 0x882}},
    {37615, 2, {0x1020, 0x1120, 0x2920, 0x25FC, 0x4120, 0xBA20, 0x1020, 0x13FE, 0xFC00, 0x1000, 0x95FC, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37616, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x41FC, 0xB924, 0x1124, 0x11FC, 0xFC20, 0x1020, 0x95FC, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37617, 2, {0x2010, 0x2014, 0x5012, 0x4810, 0x83FE, 0x7810, 0x2090, 0x2050, 0xF848, 0x2008, 0x2888, 0xB2A8, 0x629A, 0x3AAA, 0xE466, 0x2}},
    {37618, 2, {0x1010, 0x1038, 0x29C0, 0x2500, 0x4100, 0xB9FC, 0x1110, 0x1110, 0xFD10, 0x1110, 0x97FE, 0x5800, 0x5090, 0x1D08, 0xE204, 0x4402}},
    {37619, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x41FC, 0xB840, 0x13FE, 0x1048, 0xFC88, 0x10FE, 0x9508, 0x5948, 0x5228, 0x1C08, 0xE028, 0x4010}},
    {37620, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x4020, 0xB9FC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x9450, 0x5850, 0x5092, 0x1C92, 0xE112, 0x420E}},
    {37621, 2, {0x101C, 0x11E0, 0x2820, 0x2420, 0x43FE, 0xB8A8, 0x1124, 0x1222, 0xFC00, 0x10F8, 0x9488, 0x5888, 0x5088, 0x1C8A, 0xE10A, 0x4206}},
    {37622, 2, {0x1020, 0x1010, 0x29FE, 0x2420, 0x4248, 0xBC84, 0x11FE, 0x1002, 0xFEA8, 0x10A8, 0x94A8, 0x58A8, 0x50A8, 0x1EAA, 0xF12A, 0x4206}},
    {37623, 2, {0x2000, 0x27E0, 0x525E, 0x4A52, 0x8252, 0x7BD2, 0x2252, 0x2254, 0xFBD4, 0x2254, 0x2A48, 0xB2E8, 0x6754, 0x3854, 0xE062, 0x40}},
    {37624, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x41FC, 0xB910, 0x1110, 0x11FE, 0xFD10, 0x1110, 0x957C, 0x5944, 0x5144, 0x1D44, 0xE17C, 0x4244}},
    {37625, 2, {0x1000, 0x10FC, 0x2880, 0x24F8, 0x4080, 0xB8F8, 0x1080, 0x13FE, 0xFD40, 0x1124, 0x9528, 0x5910, 0x5108, 0x1D44, 0xE182, 0x4100}},
    {37626, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x8504, 0x7900, 0x21DE, 0x2252, 0xFA52, 0x2352, 0x2C9A, 0xB094, 0x6110, 0x3912, 0xE212, 0x40E}},
    {37627, 2, {0x7E7C, 0x4804, 0x7E44, 0x4228, 0x7E10, 0x4828, 0x7EC4, 0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37628, 2, {0x2000, 0x23FE, 0x5202, 0x4A8A, 0x8252, 0x7BFE, 0x2222, 0x2222, 0xFAAA, 0x22AA, 0x2AAA, 0xB2FA, 0x6202, 0x3A02, 0xE20A, 0x204}},
    {37629, 2, {0x1020, 0x1020, 0x283E, 0x2420, 0x41FC, 0xB904, 0x11FC, 0x1104, 0xFDFC, 0x1124, 0x9420, 0x5BFE, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37630, 2, {0x1040, 0x1040, 0x28FE, 0x2482, 0x4142, 0xB87A, 0x10A2, 0x1022, 0xFDFE, 0x1022, 0x94AA, 0x58AA, 0x50FA, 0x1C02, 0xE014, 0x4008}},
    {37631, 2, {0x1020, 0x1124, 0x28A4, 0x24A8, 0x4020, 0xB9FC, 0x1104, 0x1104, 0xFD74, 0x1154, 0x9554, 0x5954, 0x5174, 0x1D04, 0xE114, 0x4108}},
    {37632, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4104, 0xBAFA, 0x1000, 0x11FC, 0xFD54, 0x1154, 0x95FC, 0x5954, 0x5154, 0x1D54, 0xE104, 0x410C}},
    {37633, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x41FC, 0xB924, 0x1124, 0x11FC, 0xFC20, 0x11FE, 0x9470, 0x58A8, 0x5124, 0x1E22, 0xE020, 0x4020}},
    {37634, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x4020, 0xBBFE, 0x1088, 0x1144, 0xFE42, 0x10F8, 0x9588, 0x5A50, 0x5020, 0x1C50, 0xE188, 0x4606}},
    {37635, 2, {0x1020, 0x9FE, 0x4122, 0x2520, 0x9FC, 0x1104, 0x7268, 0x1410, 0x1128, 0x6C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37636, 2, {0x1080, 0x10F8, 0x2888, 0x2508, 0x41F0, 0xB810, 0x1010, 0x13FE, 0xFC20, 0x1222, 0x9574, 0x58A8, 0x5124, 0x1E22, 0xE0A0, 0x4040}},
    {37637, 2, {0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4908, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0x540, 0xFFFE, 0x0}},
    {37638, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x41FC, 0xB820, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x95FC, 0x5904, 0x51FC, 0x1D04, 0xE114, 0x4108}},
    {37639, 2, {0x1080, 0x1040, 0x29FC, 0x2400, 0x4108, 0xB888, 0x1090, 0x13FE, 0xFC00, 0x1000, 0x95FC, 0x5904, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37640, 2, {0x1020, 0x1124, 0x28A8, 0x2420, 0x41FC, 0xB840, 0x13FE, 0x1088, 0xFD04, 0x12FA, 0x9488, 0x5888, 0x50A8, 0x1C92, 0xE082, 0x407E}},
    {37641, 2, {0x101C, 0x13E0, 0x2A20, 0x27FE, 0x4220, 0xBA92, 0x130A, 0x1206, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37642, 2, {0x1040, 0x1020, 0x29FE, 0x2400, 0x4088, 0xB888, 0x1154, 0x1222, 0xFC00, 0x1020, 0x95FE, 0x5820, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37643, 2, {0x2000, 0x23DE, 0x5252, 0x4A52, 0x8252, 0x7BDE, 0x2252, 0x2252, 0xFA52, 0x23DE, 0x2A52, 0xB252, 0x6252, 0x3A52, 0xE552, 0x8A6}},
    {37644, 2, {0x1020, 0x1124, 0x2924, 0x2524, 0x41FC, 0xB800, 0x11FE, 0x1100, 0xFD7C, 0x1110, 0x9510, 0x59FE, 0x5110, 0x1E10, 0xE210, 0x4410}},
    {37645, 2, {0x1040, 0x1080, 0x29FC, 0x2524, 0x4124, 0xB9FC, 0x1124, 0x1144, 0xFDFC, 0x1090, 0x9510, 0x5BFE, 0x5010, 0x1C10, 0xE010, 0x4010}},
    {37646, 2, {0x1040, 0x1040, 0x287C, 0x2484, 0x4108, 0xBA00, 0x1020, 0x11CE, 0xFD02, 0x1102, 0x95CE, 0x5902, 0x5102, 0x1D02, 0xE1FE, 0x4102}},
    {37647, 2, {0x1000, 0x13FE, 0x2850, 0x2450, 0x4050, 0xB9DC, 0x1104, 0x1104, 0xFD04, 0x11DC, 0x9450, 0x5850, 0x5050, 0x1C50, 0xE3FE, 0x4000}},
    {37648, 2, {0x10A0, 0x1090, 0x2880, 0x25FE, 0x4110, 0xBB10, 0x15FC, 0x1110, 0xFD10, 0x11FC, 0x9510, 0x5910, 0x5110, 0x1DFE, 0xE100, 0x4100}},
    {37649, 2, {0x2040, 0x2020, 0x53FC, 0x4A04, 0x8204, 0x7BFC, 0x2200, 0x2228, 0xFA24, 0x23FE, 0x2A20, 0xB250, 0x6250, 0x3C88, 0xE504, 0xA02}},
    {37650, 2, {0x2000, 0x277E, 0x5504, 0x4D04, 0x8574, 0x7554, 0x2654, 0x2554, 0xF554, 0x2554, 0xA574, 0x7554, 0x2604, 0x3404, 0xC414, 0x408}},
    {37651, 2, {0x1088, 0x1048, 0x2850, 0x25FC, 0x4020, 0xB820, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x9450, 0x5850, 0x5090, 0x1C92, 0xE112, 0x420E}},
    {37652, 2, {0x1020, 0x1022, 0x2BB4, 0x24A8, 0x40A8, 0xB924, 0x12A2, 0x1040, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37653, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x41FC, 0xB904, 0x1104, 0x11FC, 0xFC00, 0x1112, 0x95D4, 0x5918, 0x5110, 0x1D52, 0xE192, 0x410E}},
    {37654, 2, {0x800, 0xE7C, 0x804, 0x7F44, 0x828, 0x2A10, 0x4928, 0x1944, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37655, 2, {0x101C, 0x11E0, 0x2820, 0x2420, 0x43FE, 0xB8A8, 0x1124, 0x1242, 0xFC40, 0x13FE, 0x9488, 0x5908, 0x50D0, 0x1C30, 0xE048, 0x4184}},
    {37656, 2, {0x2008, 0x203C, 0x51E0, 0x4820, 0x83FE, 0x78A8, 0x20A8, 0x20A8, 0xFBFE, 0x20A8, 0x28A8, 0xB3FE, 0x6020, 0x3820, 0xE1FC, 0x0}},
    {37657, 2, {0x1092, 0x1092, 0x2924, 0x2648, 0x4124, 0xB892, 0x1092, 0x1000, 0xFDFE, 0x1122, 0x9522, 0x59FE, 0x5122, 0x1D22, 0xE1FE, 0x4102}},
    {37658, 2, {0x1080, 0x1080, 0x28F8, 0x2508, 0x4210, 0xB9FC, 0x1024, 0x1024, 0xFDFE, 0x1024, 0x9424, 0x59FC, 0x5024, 0x1C20, 0xE0A0, 0x4040}},
    {37659, 2, {0x1020, 0x1020, 0x2BFE, 0x2450, 0x40A8, 0xB924, 0x12FA, 0x1020, 0xFCA8, 0x1088, 0x97FE, 0x5888, 0x5088, 0x1C88, 0xE108, 0x4208}},
    {37660, 2, {0x2020, 0x2020, 0x5050, 0x4888, 0x8144, 0x7A22, 0x21F8, 0x2008, 0xF850, 0x2020, 0x28A4, 0xB282, 0x628A, 0x3A8A, 0xE478, 0x0}},
    {37661, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x8000, 0x79FC, 0x2000, 0x2000, 0xFBFE, 0x2020, 0x2928, 0xB124, 0x6222, 0x3C22, 0xE0A0, 0x40}},
    {37662, 2, {0x1040, 0x1020, 0x2BFE, 0x2400, 0x41FC, 0xB904, 0x11FC, 0x1000, 0xFDFC, 0x1008, 0x9410, 0x5BFE, 0x5020, 0x1C20, 0xE0A0, 0x4040}},
    {37663, 2, {0x1020, 0x10A2, 0x28A2, 0x2524, 0x4050, 0xB888, 0x1304, 0x1022, 0xFC20, 0x10A4, 0x94A4, 0x5928, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37664, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x8404, 0x7800, 0x23FE, 0x2020, 0xF820, 0x2120, 0x293C, 0xB120, 0x62A0, 0x3A60, 0xE43E, 0x800}},
    {37665, 2, {0x1020, 0x1020, 0x29FC, 0x2450, 0x4088, 0xB904, 0x13FE, 0x1008, 0xFDE8, 0x1128, 0x9528, 0x59E8, 0x5128, 0x1C08, 0xE028, 0x4010}},
    {37666, 2, {0x1048, 0x1044, 0x285E, 0x25E0, 0x4228, 0xBC12, 0x106A, 0x1196, 0xFE48, 0x105E, 0x95E0, 0x5824, 0x5028, 0x1E12, 0xF06A, 0x4186}},
    {37667, 2, {0x2000, 0x23DE, 0x5042, 0x494A, 0x8084, 0x794A, 0x2252, 0x2420, 0xF800, 0x23DE, 0x2852, 0xB152, 0x6094, 0x3948, 0xE254, 0x422}},
    {37668, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x4288, 0xBCF8, 0x1088, 0x1088, 0xFEF8, 0x1088, 0x9488, 0x5BFE, 0x5000, 0x1E88, 0xF104, 0x4202}},
    {37669, 2, {0x1040, 0x1020, 0x2BFE, 0x2480, 0x4104, 0xBBFE, 0x1002, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE114, 0x4108}},
    {37670, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x40FC, 0xB884, 0x10FC, 0x1020, 0xFC20, 0x11FE, 0x9522, 0x5922, 0x512A, 0x1D24, 0xE020, 0x4020}},
    {37671, 2, {0x1020, 0x1010, 0x29FE, 0x2502, 0x4204, 0xBCF8, 0x1088, 0x1088, 0xFEF8, 0x1080, 0x9480, 0x58FC, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37672, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x4088, 0xB800, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x95FC, 0x5924, 0x5124, 0x1D24, 0xE1FC, 0x4104}},
    {37673, 2, {0x10F8, 0x1088, 0x2888, 0x24F8, 0x4088, 0xB888, 0x10F8, 0x1000, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37674, 2, {0x1050, 0x1050, 0x2850, 0x25FC, 0x4154, 0xB954, 0x1154, 0x11FC, 0xFD54, 0x1154, 0x9554, 0x5BFE, 0x5000, 0x1C50, 0xE088, 0x4104}},
    {37675, 2, {0x1000, 0x10FC, 0x2884, 0x2484, 0x40FC, 0xB884, 0x1084, 0x10FC, 0xFC40, 0x10FE, 0x952A, 0x5A4A, 0x5092, 0x1D22, 0xE04A, 0x4084}},
    {37676, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x41FC, 0xB924, 0x11FC, 0x1124, 0xFDFC, 0x1020, 0x9470, 0x58A8, 0x5124, 0x1E22, 0xE020, 0x4020}},
    {37677, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x4174, 0xB924, 0x11FC, 0x1104, 0xFD74, 0x1154, 0x9554, 0x5974, 0x5104, 0x1E04, 0xE214, 0x4408}},
    {37678, 2, {0x1000, 0x11FC, 0x2904, 0x2524, 0x4124, 0xB9FC, 0x1124, 0x1124, 0xFD74, 0x1154, 0x9554, 0x5974, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37679, 2, {0x1048, 0x1048, 0x2848, 0x25FE, 0x4248, 0xBC48, 0x11FE, 0x1000, 0xFEFC, 0x1084, 0x9484, 0x58FC, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37680, 2, {0x1020, 0x1020, 0x2820, 0x25FE, 0x4020, 0xB870, 0x10A8, 0x10A8, 0xFD24, 0x1222, 0x9420, 0x5800, 0x5154, 0x1D2A, 0xE22A, 0x4000}},
    {37681, 2, {0x100, 0x280, 0xC60, 0x3018, 0xDFF6, 0x100, 0x3FF8, 0x1110, 0x920, 0x7FFC, 0x0, 0x7EFE, 0x810, 0x3E7C, 0x810, 0x7EFE}},
    {37682, 2, {0x1000, 0x11F8, 0x2808, 0x2408, 0x41F8, 0xB808, 0x1008, 0x13FE, 0xFC20, 0x1222, 0x9574, 0x58A8, 0x5124, 0x1E22, 0xE0A0, 0x4040}},
    {37683, 2, {0x1000, 0x11FC, 0x2808, 0x2410, 0x4020, 0xBBFE, 0x1020, 0x10A0, 0xFC40, 0x11FC, 0x9554, 0x5954, 0x5154, 0x1D54, 0xE3FE, 0x4000}},
    {37684, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x4020, 0xBBFE, 0x1088, 0x1104, 0xFE22, 0x1020, 0x95FC, 0x5820, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37685, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x4088, 0xB850, 0x1050, 0x1094, 0xFD94, 0x1298, 0x9490, 0x58B2, 0x50D2, 0x1C92, 0xE08E, 0x4080}},
    {37686, 2, {0x2020, 0x2020, 0x53FE, 0x4820, 0x8020, 0x79FC, 0x2020, 0x2020, 0xFBFE, 0x2040, 0x28A4, 0xB1A8, 0x6290, 0x3C88, 0xE0C6, 0x80}},
    {37687, 2, {0x2088, 0x2088, 0x2088, 0x57E8, 0x8888, 0x73E8, 0x22A8, 0x23E8, 0xFAA8, 0x23E8, 0x2888, 0xB7E8, 0x608A, 0x388A, 0xE08A, 0x86}},
    {37688, 2, {0x2020, 0x2020, 0x5020, 0x4BFE, 0x8020, 0x7924, 0x2124, 0x2124, 0xFAAA, 0x2070, 0x28A8, 0xB0A8, 0x6124, 0x3A22, 0xE420, 0x20}},
    {37689, 2, {0x1020, 0x1120, 0x293C, 0x2520, 0x4120, 0xBBFE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE114, 0x4108}},
    {37690, 2, {0x1108, 0x1108, 0x2BFE, 0x2508, 0x4148, 0xB820, 0x13FE, 0x1080, 0xFC80, 0x10F8, 0x9488, 0x5888, 0x5088, 0x1D08, 0xE128, 0x4210}},
    {37691, 2, {0x2010, 0x2014, 0x51D2, 0x4812, 0x8010, 0x7BFE, 0x2010, 0x2090, 0xF890, 0x22D0, 0x2A90, 0xB290, 0x628A, 0x3AEA, 0xE706, 0x202}},
    {37692, 2, {0x2020, 0x2020, 0x53FE, 0x4850, 0x8088, 0x7904, 0x22FA, 0x2000, 0xF800, 0x23FE, 0x2820, 0xB124, 0x6122, 0x3A22, 0xE0A0, 0x40}},
    {37693, 2, {0x1000, 0x13FE, 0x2820, 0x24A4, 0x4128, 0xB850, 0x1088, 0x1144, 0xFC40, 0x10FC, 0x9584, 0x5A88, 0x5050, 0x1C20, 0xE0D8, 0x4706}},
    {37694, 2, {0x1004, 0xFF78, 0x2840, 0x7F7E, 0x848, 0xF48, 0xF888, 0x908, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37695, 2, {0x1020, 0x1020, 0x283E, 0x2420, 0x41FE, 0xB922, 0x1138, 0x11E0, 0xFD22, 0x111E, 0x9500, 0x5978, 0x5248, 0x1E4A, 0xE48A, 0x4906}},
    {37696, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1000, 0x11FE, 0xFC08, 0x13FE, 0x9408, 0x5908, 0x5088, 0x1C08, 0xE028, 0x4010}},
    {37697, 2, {0x2010, 0x2090, 0x5710, 0x4C3E, 0x8422, 0x7444, 0x2790, 0x2510, 0xF510, 0x2510, 0xA528, 0x7528, 0x2528, 0x3944, 0xC944, 0x1082}},
    {37698, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4104, 0xBA02, 0x10F8, 0x1020, 0xFC20, 0x11FC, 0x9420, 0x5924, 0x50A4, 0x1CA8, 0xE3FE, 0x4000}},
    {37699, 2, {0x2080, 0x2080, 0x50FC, 0x4954, 0x8254, 0x78A4, 0x2124, 0x2244, 0xF894, 0x2108, 0x2840, 0xB0A4, 0x62AA, 0x3A8A, 0xE478, 0x0}},
    {37700, 2, {0x1040, 0x1020, 0x2BFE, 0x2400, 0x4000, 0xB9FC, 0x1104, 0x1104, 0xFDFC, 0x1020, 0x9528, 0x5924, 0x5222, 0x1C22, 0xE0A0, 0x4040}},
    {37701, 2, {0x2210, 0x2110, 0x2110, 0x547C, 0x8A10, 0x7210, 0x2010, 0x217E, 0xF910, 0x2210, 0x2E20, 0xB224, 0x6242, 0x3AFE, 0xE242, 0x0}},
    {37702, 2, {0x2000, 0x23DE, 0x5252, 0x4A52, 0x83DE, 0x7A52, 0x2252, 0x23DE, 0xFA02, 0x2202, 0x2A02, 0xB202, 0x6202, 0x3A02, 0xE20A, 0x204}},
    {37703, 2, {0x1110, 0x1112, 0x29D4, 0x2518, 0x4152, 0xB992, 0x112E, 0x1040, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37704, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x40A8, 0xB820, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x97FE, 0x5820, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37705, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x41FC, 0xB904, 0x1104, 0x11FC, 0xFC00, 0x13FE, 0x9420, 0x5920, 0x513C, 0x1D20, 0xE2A0, 0x447E}},
    {37706, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x41FC, 0xB924, 0x11AC, 0x1174, 0xFD24, 0x11FC, 0x9420, 0x5870, 0x50A8, 0x1D24, 0xE222, 0x4020}},
    {37707, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x40E8, 0xB8A8, 0x10A8, 0x11FC, 0xFD04, 0x1174, 0x9554, 0x5954, 0x5174, 0x1D04, 0xE114, 0x4108}},
    {37708, 2, {0x20, 0x1120, 0x49FC, 0x2220, 0xBFE, 0x1090, 0x7092, 0x110E, 0x16C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37709, 2, {0x2040, 0x2020, 0x53FE, 0x4A48, 0x8248, 0x7BFE, 0x2248, 0x2278, 0xFA00, 0x22FC, 0x2A44, 0xB248, 0x6228, 0x3C10, 0xE468, 0x986}},
    {37710, 2, {0x1008, 0x101C, 0x29F0, 0x2510, 0x4110, 0xB9FE, 0x1110, 0x117C, 0xFD44, 0x117C, 0x9544, 0x597C, 0x5144, 0x1E44, 0xE27C, 0x4444}},
    {37711, 2, {0x1020, 0x1020, 0x28FC, 0x2444, 0x41FE, 0xB800, 0x10FC, 0x1084, 0xFCFC, 0x1020, 0x95FE, 0x5820, 0x5120, 0x1DFE, 0xE020, 0x4020}},
    {37712, 2, {0x2000, 0x2088, 0x5252, 0x4A22, 0x8252, 0x7A8A, 0x2202, 0x23FE, 0xF888, 0x2144, 0x2A7A, 0xB088, 0x6150, 0x3820, 0xE0D8, 0x706}},
    {37713, 2, {0x1040, 0x1040, 0x28FE, 0x2480, 0x41FC, 0xBA84, 0x10FC, 0x1084, 0xFCFC, 0x1040, 0x947C, 0x58C4, 0x5128, 0x1C10, 0xE068, 0x4186}},
    {37714, 2, {0x1000, 0x10FC, 0x2848, 0x2430, 0x41FE, 0xB852, 0x1094, 0x1110, 0xFE30, 0x1020, 0x95FE, 0x5870, 0x50A8, 0x1D24, 0xE222, 0x4020}},
    {37715, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x40F8, 0xB800, 0x13FE, 0x1088, 0xFCF8, 0x1088, 0x94F8, 0x5888, 0x509E, 0x1DE8, 0xE008, 0x4008}},
    {37716, 2, {0x2000, 0x23DE, 0x5252, 0x4A52, 0x83DE, 0x7800, 0x21FC, 0x2000, 0xFBFE, 0x2080, 0x2900, 0xB1FC, 0x6004, 0x3804, 0xE028, 0x10}},
    {37717, 2, {0x1000, 0x13FE, 0x2A22, 0x2420, 0x43FE, 0xB820, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x95FC, 0x5820, 0x53FE, 0x1C20, 0xE020, 0x4020}},
    {37718, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x40F8, 0xB888, 0x10F8, 0x1088, 0xFC88, 0x13FE, 0x9500, 0x5948, 0x5184, 0x1D00, 0xE1FE, 0x4000}},
    {37719, 2, {0x1040, 0x1020, 0x29FC, 0x2400, 0x4088, 0xB850, 0x13FE, 0x1222, 0xFC24, 0x11FC, 0x9524, 0x5924, 0x5134, 0x1D28, 0xE020, 0x4020}},
    {37720, 2, {0x2002, 0x23C2, 0x5242, 0x4A52, 0x83D2, 0x7A52, 0x2252, 0x23D2, 0xFA52, 0x2252, 0x2BD2, 0xB002, 0x6182, 0x3942, 0xE24A, 0x404}},
    {37721, 2, {0x1090, 0x890, 0x43FC, 0x2090, 0x890, 0x17FE, 0x7090, 0x1108, 0x16C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37722, 2, {0x1000, 0x10FC, 0x2884, 0x24FC, 0x4084, 0xB8FC, 0x1000, 0x11FE, 0xFC80, 0x10FE, 0x952A, 0x5A4A, 0x5092, 0x1D22, 0xE054, 0x4088}},
    {37723, 2, {0x2080, 0x2338, 0x5228, 0x4A28, 0x83A8, 0x7A46, 0x2200, 0x227C, 0xFBA4, 0x2224, 0x2A28, 0xB3A8, 0x6610, 0x3A28, 0xE244, 0x282}},
    {37724, 2, {0x2000, 0x23DE, 0x5242, 0x4A42, 0x8242, 0x7BDE, 0x2200, 0x223E, 0xFBD2, 0x2212, 0x2A14, 0xB3D4, 0x6208, 0x3A14, 0xE224, 0x242}},
    {37725, 2, {0x11FC, 0x1124, 0x2924, 0x25FC, 0x4124, 0xB924, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x9622, 0x5A2A, 0x52FA, 0x1E0A, 0xE202, 0x4206}},
    {37726, 2, {0x1020, 0x1020, 0x283E, 0x2420, 0x4020, 0xB9FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5000, 0x1C88, 0xE104, 0x4202}},
    {37727, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1020, 0x1120, 0xFDFE, 0x1220, 0x9420, 0x59FC, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37728, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x40FC, 0xB884, 0x10FC, 0x1000, 0xFDFE, 0x1020, 0x9420, 0x58FC, 0x5020, 0x1C20, 0xE1FE, 0x4000}},
    {37729, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x41FC, 0xB924, 0x1124, 0x11FC, 0xFC00, 0x13FE, 0x9520, 0x5922, 0x5114, 0x1D48, 0xE184, 0x4102}},
    {37730, 2, {0x1000, 0x10FE, 0x2800, 0x247C, 0x4244, 0xBC44, 0x107C, 0x1000, 0xFEFE, 0x1092, 0x9492, 0x58FE, 0x5092, 0x1E92, 0xF0FE, 0x4082}},
    {37731, 2, {0x1048, 0x1048, 0x29FE, 0x2448, 0x4200, 0xBDFC, 0x1044, 0x1044, 0xFE94, 0x1108, 0x96FC, 0x5884, 0x5084, 0x1E84, 0xF0FC, 0x4084}},
    {37732, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x4020, 0xBBFE, 0x1020, 0x10A0, 0xFD2C, 0x1124, 0x9524, 0x59AC, 0x5124, 0x1D24, 0xE1FC, 0x4104}},
    {37733, 2, {0x1080, 0x109E, 0x2BEA, 0x248A, 0x41CA, 0xB88A, 0x13EA, 0x1092, 0xFCA6, 0x1020, 0x97FE, 0x5820, 0x5050, 0x1C88, 0xE104, 0x4602}},
    {37734, 2, {0x2210, 0x2110, 0x2110, 0x57BE, 0x8A28, 0x7248, 0x23AA, 0x22AE, 0xFABA, 0x22EA, 0x2AAE, 0xB2A8, 0x64AA, 0x3CA2, 0xE9A2, 0x101E}},
    {37735, 2, {0x1080, 0x1080, 0x28FE, 0x2502, 0x4242, 0xB9FA, 0x1002, 0x11F2, 0xFC02, 0x11F2, 0x9402, 0x59F2, 0x5112, 0x1DF2, 0xE014, 0x4008}},
    {37736, 2, {0x1010, 0x11D4, 0x2858, 0x2652, 0x418C, 0xB888, 0x1104, 0x12FA, 0xFC20, 0x1020, 0x95FE, 0x5820, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37737, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x40A8, 0xB820, 0x13FE, 0x1040, 0xFC80, 0x10FC, 0x9584, 0x5A84, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37738, 2, {0x20, 0x7E20, 0x247E, 0x1884, 0xFF28, 0x2910, 0x4A28, 0x9944, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37739, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x904, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37740, 2, {0x2088, 0x21C8, 0x5708, 0x4908, 0x812A, 0x712A, 0x27AC, 0x2148, 0xF308, 0x2388, 0xA554, 0x7514, 0x2914, 0x3124, 0xC124, 0x142}},
    {37741, 2, {0x1080, 0x10BC, 0x2884, 0x2508, 0x41FE, 0xBB20, 0x1520, 0x117C, 0xFD90, 0x1110, 0x95FE, 0x5910, 0x5128, 0x1D28, 0xE144, 0x4182}},
    {37742, 2, {0x2040, 0x20A0, 0x5110, 0x4A08, 0x85F6, 0x7800, 0x23C4, 0x2254, 0xFA54, 0x23D4, 0x2A54, 0xB254, 0x63D4, 0x3A44, 0xE254, 0x2C8}},
    {37743, 2, {0x2080, 0x2080, 0x50FC, 0x4954, 0x82D4, 0x78B4, 0x212C, 0x2244, 0xF894, 0x2108, 0x2840, 0xB0A4, 0x62AA, 0x3A8A, 0xE478, 0x0}},
    {37744, 2, {0x101E, 0x13E0, 0x2844, 0x2524, 0x4088, 0xB9FC, 0x1040, 0x1040, 0xFDFE, 0x1080, 0x94FC, 0x5944, 0x5128, 0x1E10, 0xE068, 0x4186}},
    {37745, 2, {0x2048, 0x2148, 0x5148, 0x4BFE, 0x8148, 0x7948, 0x2178, 0x2100, 0xF9FE, 0x2020, 0x2BFE, 0xB070, 0x60A8, 0x3924, 0xE622, 0x20}},
    {37746, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4110, 0xB9FE, 0x1110, 0x114A, 0xFD86, 0x1000, 0x95FC, 0x5904, 0x51FC, 0x1D04, 0xE1FC, 0x4104}},
    {37747, 2, {0x3F20, 0x2420, 0x3F3E, 0x2140, 0x3F50, 0x2488, 0x3F08, 0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37748, 2, {0x1020, 0x1124, 0x2924, 0x2524, 0x41FC, 0xB800, 0x13FE, 0x1020, 0xFC40, 0x11FC, 0x9554, 0x5954, 0x5154, 0x1D54, 0xE154, 0x410C}},
    {37749, 2, {0x2010, 0x2010, 0x577C, 0x4914, 0x81FE, 0x7214, 0x227C, 0x2710, 0xF17C, 0x2510, 0xA5FE, 0x7210, 0x2210, 0x3500, 0xC8FE, 0x0}},
    {37750, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x41FC, 0xB924, 0x1124, 0x11FC, 0xFC20, 0x1010, 0x9454, 0x5942, 0x5142, 0x1D4A, 0xE238, 0x4000}},
    {37751, 2, {0x1020, 0x1020, 0x2BFE, 0x2450, 0x4088, 0xBB26, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x9400, 0x5820, 0x51FC, 0x1C20, 0xE020, 0x43FE}},
    {37752, 2, {0x1080, 0x109E, 0x2892, 0x25D2, 0x4092, 0xB89E, 0x1092, 0x11D2, 0xFD52, 0x115E, 0x9552, 0x5952, 0x51D2, 0x1C22, 0xE02A, 0x4044}},
    {37753, 2, {0x1040, 0x1020, 0x2BFE, 0x2602, 0x4404, 0xB9FC, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1C00, 0xE3FE, 0x4000}},
    {37754, 2, {0x1020, 0x1022, 0x29FA, 0x2424, 0x4028, 0xBBFE, 0x1020, 0x1040, 0xFCFC, 0x1184, 0x9684, 0x58FC, 0x5084, 0x1C84, 0xE0FC, 0x4084}},
    {37755, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1080, 0x11FE, 0xFE22, 0x1122, 0x9552, 0x5902, 0x51FA, 0x1C02, 0xE014, 0x4008}},
    {37756, 2, {0x2008, 0x200C, 0x500A, 0x4BFE, 0x8208, 0x7A08, 0x22E8, 0x220A, 0xFA0A, 0x22EC, 0x2AAC, 0xB2A8, 0x62EA, 0x3A1A, 0xE226, 0x442}},
    {37757, 2, {0x1040, 0x1020, 0x29FE, 0x2502, 0x4102, 0xB9FE, 0x1100, 0x1100, 0xFDFE, 0x11AA, 0x96AA, 0x5AFE, 0x54AA, 0x1CAA, 0xE0A2, 0x4086}},
    {37758, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x43FE, 0xB820, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x95FC, 0x5820, 0x51FC, 0x1C20, 0xE3FE, 0x4000}},
    {37759, 2, {0x1052, 0x1052, 0x28A4, 0x25FE, 0x42A4, 0xBC52, 0x1052, 0x1000, 0xFEFE, 0x1092, 0x9492, 0x58FE, 0x5092, 0x1E92, 0xF0FE, 0x4082}},
    {37760, 2, {0x1090, 0x109E, 0x28A2, 0x2554, 0x4108, 0xBB54, 0x1562, 0x1140, 0xFD44, 0x1158, 0x9542, 0x5944, 0x5118, 0x1D02, 0xE10C, 0x4130}},
    {37761, 2, {0x2000, 0x2FDE, 0x5492, 0x4C92, 0x8494, 0x7794, 0x2498, 0x2494, 0xF792, 0x2492, 0xA492, 0x75DA, 0x2E94, 0x3090, 0xC090, 0x90}},
    {37762, 2, {0x1104, 0x1088, 0x2BFE, 0x2420, 0x4020, 0xB9FC, 0x1020, 0x1020, 0xFDFE, 0x1000, 0x9420, 0x5BFE, 0x5020, 0x1C50, 0xE088, 0x4306}},
    {37763, 2, {0x2008, 0x241C, 0x52E0, 0x4A80, 0x888C, 0x74B0, 0x24A8, 0x22AA, 0xF2AC, 0x24A8, 0xACA8, 0x74A8, 0x24A4, 0x3524, 0xC532, 0x220}},
    {37764, 2, {0x1040, 0x1020, 0x2BFE, 0x2400, 0x41FC, 0xB904, 0x1104, 0x1104, 0xFDFC, 0x1052, 0x9494, 0x5988, 0x5288, 0x1CA4, 0xE0C2, 0x4080}},
    {37765, 2, {0x11FC, 0x1124, 0x2924, 0x25FC, 0x4124, 0xB924, 0x11FC, 0x1050, 0xFC88, 0x1104, 0x968A, 0x5888, 0x5088, 0x1C88, 0xE108, 0x4208}},
    {37766, 2, {0x2108, 0x2088, 0x5090, 0x4BFE, 0x8000, 0x7BC4, 0x2254, 0x2254, 0xFBD4, 0x2254, 0x2A54, 0xB3D4, 0x6254, 0x3A44, 0xE254, 0x2C8}},
    {37767, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x41FC, 0xB900, 0x117C, 0x1144, 0xFD44, 0x117C, 0x9544, 0x597C, 0x5144, 0x1E44, 0xE27C, 0x4444}},
    {37768, 2, {0x1084, 0x1044, 0x2848, 0x25FE, 0x4020, 0xB8FC, 0x1020, 0x11FE, 0xFC40, 0x1080, 0x94FE, 0x5910, 0x5210, 0x1C10, 0xE1FE, 0x4000}},
    {37769, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1000, 0x13DE, 0xFC42, 0x1252, 0x954A, 0x58C6, 0x535A, 0x1C42, 0xE14A, 0x4084}},
    {37770, 2, {0x1020, 0x1010, 0x29FE, 0x2484, 0x4048, 0xB9FE, 0x1102, 0x1224, 0xFC10, 0x11FE, 0x9440, 0x587C, 0x5044, 0x1C84, 0xE094, 0x4108}},
    {37771, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x8424, 0x79FC, 0x2020, 0x21FC, 0xF820, 0x23FE, 0x2820, 0xB1FC, 0x6104, 0x3904, 0xE1FC, 0x104}},
    {37772, 2, {0x1088, 0x1050, 0x2800, 0x25FE, 0x4050, 0xB9FC, 0x1054, 0x13FE, 0xFC54, 0x11FC, 0x9450, 0x58D8, 0x5154, 0x1E52, 0xE050, 0x4050}},
    {37773, 2, {0x2020, 0x2020, 0x53FE, 0x4820, 0x83FE, 0x7A42, 0x248C, 0x21F0, 0xF820, 0x2044, 0x2BFE, 0xB022, 0x6128, 0x3A24, 0xE4A4, 0x40}},
    {37774, 2, {0x2100, 0x2100, 0x51FC, 0x4A00, 0x85F8, 0x7800, 0x23F8, 0x2088, 0xFAA8, 0x21C8, 0x2BF8, 0xB08A, 0x61CA, 0x3AAA, 0xE086, 0x82}},
    {37775, 2, {0x2040, 0x17FE, 0x8090, 0x4908, 0x13FC, 0xE150, 0x2252, 0x244E, 0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37776, 2, {0x1008, 0x103C, 0x2BC0, 0x2404, 0x4244, 0xB928, 0x11FC, 0x1220, 0xFC20, 0x13FE, 0x9420, 0x5924, 0x5124, 0x1D24, 0xE1FC, 0x4004}},
    {37777, 2, {0x1020, 0x1020, 0x28FC, 0x2420, 0x4020, 0xB9FE, 0x1040, 0x1084, 0xFDFE, 0x1002, 0x95FC, 0x5954, 0x5154, 0x1D54, 0xE3FE, 0x4000}},
    {37778, 2, {0x1000, 0x11FE, 0x2900, 0x257E, 0x4100, 0xB9FE, 0x1154, 0x1148, 0xFD64, 0x1142, 0x9504, 0x59FE, 0x5144, 0x1E24, 0xE204, 0x440C}},
    {37779, 2, {0x2090, 0x2088, 0x5104, 0x4A42, 0x8488, 0x79FC, 0x2004, 0x2000, 0xFBDE, 0x2042, 0x2A52, 0xB14A, 0x6252, 0x3842, 0xE14A, 0x84}},
    {37780, 2, {0x1040, 0x1020, 0x2BFE, 0x2602, 0x4050, 0xB888, 0x1124, 0x1050, 0xFC88, 0x1104, 0x96FA, 0x5888, 0x5088, 0x1C88, 0xE0F8, 0x4088}},
    {37781, 2, {0x1020, 0x1010, 0x29FE, 0x2510, 0x417C, 0xB914, 0x11FE, 0x1114, 0xFD7C, 0x1110, 0x957C, 0x5944, 0x5144, 0x1E44, 0xE27C, 0x4444}},
    {37782, 2, {0x1020, 0x1124, 0x28A8, 0x2420, 0x41FC, 0xB904, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x5904, 0x51FC, 0x1C00, 0xE088, 0x4104}},
    {37783, 2, {0x1020, 0x1020, 0x2850, 0x2448, 0x40A4, 0xB9FE, 0x1284, 0x10FC, 0xFC84, 0x10FC, 0x9480, 0x58FC, 0x5144, 0x1D44, 0xE27C, 0x4044}},
    {37784, 2, {0x2000, 0x23FE, 0x5000, 0x49FC, 0x8104, 0x7904, 0x21FC, 0x2000, 0xFBFE, 0x228A, 0x2A52, 0xB3FE, 0x6222, 0x3A22, 0xE22A, 0x204}},
    {37785, 2, {0x2440, 0x225E, 0x5292, 0x4812, 0x87D2, 0x791E, 0x2552, 0x2552, 0xFD52, 0x27DE, 0x2952, 0xB112, 0x6212, 0x3A22, 0xE42A, 0x844}},
    {37786, 2, {0x1010, 0x1220, 0x297C, 0x2544, 0x4044, 0xB87C, 0x1340, 0x1140, 0xFD7C, 0x1144, 0x9544, 0x597C, 0x5144, 0x1E80, 0xE47E, 0x4000}},
    {37787, 2, {0x1028, 0x1024, 0x2BFE, 0x2420, 0x41FC, 0xB924, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x9408, 0x5BFE, 0x5088, 0x1C48, 0xE048, 0x4018}},
    {37788, 2, {0x1000, 0x3E78, 0x2A48, 0x2248, 0xFE86, 0x2B78, 0x4248, 0x4A30, 0x8548, 0x6C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37789, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x4020, 0xB850, 0x1088, 0x1104, 0xFEFA, 0x1000, 0x9400, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37790, 2, {0x1020, 0x1040, 0x29FC, 0x2504, 0x4154, 0xB924, 0x1154, 0x1104, 0xFDFC, 0x1000, 0x9512, 0x59D4, 0x5118, 0x1D52, 0xE192, 0x410E}},
    {37791, 2, {0x21F8, 0x2088, 0x5070, 0x498C, 0x8000, 0x7BDE, 0x2252, 0x218C, 0xFA52, 0x2020, 0x2BFE, 0xB070, 0x60A8, 0x3924, 0xE622, 0x20}},
    {37792, 2, {0x2000, 0x23FE, 0x5252, 0x4A52, 0x83FE, 0x7800, 0x2000, 0x23FE, 0xF820, 0x2020, 0x293C, 0xB120, 0x6120, 0x3920, 0xE7FE, 0x0}},
    {37793, 2, {0x2084, 0x2044, 0x5048, 0x4BFE, 0x8084, 0x7884, 0x2108, 0x214A, 0xFA52, 0x239C, 0x2884, 0xB108, 0x6108, 0x3A52, 0xE3DE, 0x42}},
    {37794, 2, {0x1040, 0x1080, 0x29FC, 0x2504, 0x4104, 0xB9FC, 0x1100, 0x11FE, 0xFD00, 0x11FE, 0x9402, 0x5AAA, 0x52AA, 0x1C02, 0xE014, 0x4008}},
    {37795, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4102, 0x86C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0x540, 0xFFFE, 0x0}},
    {37796, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1020, 0x1124, 0xFCA8, 0x1020, 0x97FE, 0x5890, 0x5090, 0x1D12, 0xE212, 0x440E}},
    {37797, 2, {0x1080, 0x11F8, 0x2288, 0x6870, 0xAB8E, 0x2820, 0x29FC, 0x28A8, 0x2164, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37798, 2, {0x2040, 0x23BE, 0x5212, 0x4A92, 0x8252, 0x7AAA, 0x2324, 0x2040, 0xFBFE, 0x2222, 0x2A22, 0xB3FE, 0x6222, 0x3A22, 0xE3FE, 0x202}},
    {37799, 2, {0x1020, 0x1124, 0x2924, 0x2524, 0x41FC, 0xB800, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5088, 0x1C50, 0xE3FE, 0x4000}},
    {37800, 2, {0x10A0, 0x1090, 0x29FE, 0x2720, 0x41FC, 0xB920, 0x11FC, 0x1120, 0xFDFE, 0x1100, 0x9420, 0x5BFE, 0x5020, 0x1C20, 0xE020, 0x4020}},
    {37801, 2, {0x2000, 0x245C, 0x5294, 0x4914, 0x8294, 0x7454, 0x2126, 0x2100, 0xF7DC, 0x2114, 0xA594, 0x7554, 0x2948, 0x3108, 0xC514, 0x222}},
    {37802, 2, {0x10A0, 0x112C, 0x2924, 0x2524, 0x41AC, 0xB924, 0x1124, 0x11FC, 0xFC20, 0x11FC, 0x9488, 0x5850, 0x5020, 0x1C50, 0xE088, 0x4306}},
    {37803, 2, {0x1020, 0x1040, 0x29FC, 0x2504, 0x4154, 0xB924, 0x1154, 0x1104, 0xFC88, 0x1144, 0x967A, 0x5888, 0x5150, 0x1C20, 0xE0D8, 0x4706}},
    {37804, 2, {0x1020, 0x1010, 0x29FE, 0x2400, 0x40FC, 0xB884, 0x10FC, 0x1000, 0xFDFE, 0x1102, 0x957A, 0x594A, 0x517A, 0x1D02, 0xE10A, 0x4104}},
    {37805, 2, {0x1118, 0x11E0, 0x2902, 0x24FE, 0x4000, 0xB97C, 0x1144, 0x117C, 0xFD44, 0x117C, 0x9544, 0x597C, 0x5100, 0x1DFE, 0xE088, 0x4104}},
    {37806, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x41FC, 0xB904, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x95FC, 0x5904, 0x53FE, 0x1C88, 0xE104, 0x4202}},
    {37807, 2, {0x2100, 0x209E, 0x53D2, 0x4A52, 0x8254, 0x7BD4, 0x2258, 0x2254, 0xFBD2, 0x2212, 0x2A92, 0xB25A, 0x62D4, 0x3B50, 0xE210, 0x10}},
    {37808, 2, {0x1088, 0x1050, 0x2800, 0x25FC, 0x4000, 0xB850, 0x1088, 0x1104, 0xFC00, 0x11FC, 0x9554, 0x5954, 0x5154, 0x1D54, 0xE3FE, 0x4000}},
    {37809, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x43FE, 0xB800, 0x11FC, 0x1104, 0xFD04, 0x11FC, 0x9452, 0x5894, 0x5188, 0x1EA4, 0xE0C2, 0x4080}},
    {37810, 2, {0x1020, 0x1124, 0x28A8, 0x2420, 0x43FE, 0xBA02, 0x1000, 0x11FC, 0xFD04, 0x1104, 0x95FC, 0x5850, 0x5050, 0x1C92, 0xE112, 0x420E}},
    {37811, 2, {0x1020, 0x1040, 0x29FC, 0x2504, 0x41FC, 0xB904, 0x11FC, 0x1104, 0xFDFC, 0x1020, 0x95FE, 0x5870, 0x50A8, 0x1D24, 0xE222, 0x4020}},
    {37812, 2, {0x1020, 0x1040, 0x29FC, 0x2504, 0x41FC, 0xB904, 0x11FC, 0x1104, 0xFDFC, 0x1040, 0x9420, 0x58A4, 0x528A, 0x1E8A, 0xE478, 0x4000}},
    {37813, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x8000, 0x79FC, 0x2040, 0x20A2, 0xFB34, 0x2058, 0x2894, 0xB334, 0x6052, 0x3890, 0xE350, 0x20}},
    {37814, 2, {0x1000, 0x13FE, 0x2808, 0x25E8, 0x4128, 0xB9E8, 0x1000, 0x13FE, 0xFC08, 0x11E8, 0x9528, 0x5928, 0x51E8, 0x1C08, 0xE028, 0x4010}},
    {37815, 2, {0x2000, 0x21FE, 0x5120, 0x49FC, 0x8120, 0x79FC, 0x2120, 0x2120, 0xF9FE, 0x2002, 0x2AAA, 0xB2AA, 0x62AA, 0x3A02, 0xE014, 0x8}},
    {37816, 2, {0x20A0, 0x2090, 0x51FE, 0x4B20, 0x85FC, 0x7920, 0x21FC, 0x2120, 0xF9FE, 0x2100, 0x2BFC, 0xB088, 0x609E, 0x3902, 0xE20A, 0x404}},
    {37817, 2, {0x1022, 0x1012, 0x2914, 0x2480, 0x40BE, 0xB808, 0x1008, 0x11BE, 0xFC88, 0x1088, 0x9494, 0x5892, 0x50A2, 0x1D40, 0xE23E, 0x4000}},
    {37818, 2, {0x2200, 0x2100, 0x513C, 0x4824, 0x87A4, 0x78A4, 0x213C, 0x2124, 0xFBA4, 0x2564, 0x293C, 0xB124, 0x6124, 0x3924, 0xE17E, 0x100}},
    {37819, 2, {0x1092, 0x1124, 0x2A48, 0x2524, 0x4092, 0xB800, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x95FC, 0x5904, 0x51FC, 0x1C88, 0xE104, 0x4202}},
    {37820, 2, {0x1020, 0x103C, 0x2820, 0x25FE, 0x4122, 0xB938, 0x11E2, 0x111E, 0xFD04, 0x1178, 0x9510, 0x597C, 0x5110, 0x1EFE, 0xE210, 0x4430}},
    {37821, 2, {0x2020, 0x2222, 0x53FE, 0x4880, 0x81F8, 0x7A88, 0x2070, 0x238E, 0xF820, 0x21FC, 0x2820, 0xB1FC, 0x6020, 0x3BFE, 0xE020, 0x20}},
    {37822, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x40F8, 0xB888, 0x1088, 0x10F8, 0xFC00, 0x11FC, 0x9554, 0x5954, 0x5154, 0x1D54, 0xE3FE, 0x4000}},
    {37823, 2, {0x1020, 0x1050, 0x2888, 0x2574, 0x4202, 0xB8F8, 0x1088, 0x10F8, 0xFC0C, 0x11F0, 0x9420, 0x59FC, 0x5020, 0x1DFE, 0xE020, 0x4060}},
    {37824, 2, {0x2020, 0x2020, 0x503E, 0x4820, 0x8020, 0x7BFE, 0x2202, 0x2326, 0xFA8A, 0x2252, 0x2B26, 0xB252, 0x628A, 0x3B26, 0xE3FE, 0x202}},
    {37825, 2, {0x1124, 0x1248, 0x2924, 0x2400, 0x41FC, 0xB924, 0x11FC, 0x1124, 0xFDFC, 0x1020, 0x97FE, 0x5870, 0x50A8, 0x1D24, 0xE622, 0x4020}},
    {37826, 2, {0x2000, 0x23FE, 0x5200, 0x4A7C, 0x8244, 0x7A44, 0x227C, 0x2200, 0xFAEE, 0x22AA, 0x2AAA, 0xB2AA, 0x62EE, 0x3A00, 0xE3FE, 0x0}},
    {37827, 2, {0x2220, 0x2120, 0x5120, 0x4FBE, 0x8240, 0x7A20, 0x23BC, 0x22D0, 0xFA90, 0x2290, 0x2AFE, 0xB290, 0x64A8, 0x3DA8, 0xE844, 0x1082}},
    {37828, 2, {0x1020, 0x13FE, 0x2820, 0x25FC, 0x4124, 0xB9FC, 0x1124, 0x11FC, 0xFC22, 0x13FE, 0x9408, 0x5BFE, 0x5108, 0x1C88, 0xE028, 0x4010}},
    {37829, 2, {0x2120, 0x213C, 0x2144, 0x52A8, 0x8A10, 0x76A8, 0x2AC6, 0x22BC, 0xFAA4, 0x22BC, 0x2AA4, 0xB2BC, 0x6224, 0x3A24, 0xE224, 0x22C}},
    {37830, 2, {0x1000, 0x11FC, 0x2924, 0x27FE, 0x4124, 0xB9FC, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1DFC, 0xE088, 0x4104}},
    {37831, 2, {0x2210, 0x2110, 0x2110, 0x57BE, 0x8A40, 0x7200, 0x23BE, 0x228A, 0xFA88, 0x22A8, 0x2AAE, 0xB2A8, 0x64A8, 0x3DA8, 0xE85E, 0x1080}},
    {37832, 2, {0x2010, 0x2010, 0x52FE, 0x4910, 0x817C, 0x7854, 0x207C, 0x2354, 0xF97C, 0x2110, 0x29FE, 0xB110, 0x6110, 0x3A90, 0xE47E, 0x0}},
    {37833, 2, {0x2108, 0x2108, 0x57C8, 0x4908, 0x811E, 0x77D2, 0x2564, 0x2540, 0xF7C8, 0x2108, 0xA388, 0x7548, 0x2954, 0x3114, 0xC124, 0x142}},
    {37834, 2, {0x820, 0x7F20, 0x83E, 0x7E44, 0x884, 0xFF28, 0x2010, 0x3E28, 0x4344, 0x86C2, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37835, 2, {0x2088, 0x2088, 0x53FE, 0x4888, 0x8088, 0x78F8, 0x2020, 0x23FE, 0xFA22, 0x2332, 0x2AAA, 0xB376, 0x6222, 0x3A22, 0xE22A, 0x224}},
    {37836, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x4000, 0xB9FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x9420, 0x5BFE, 0x5050, 0x1C88, 0xE104, 0x4202}},
    {37837, 2, {0x1000, 0x11FC, 0x2924, 0x25FC, 0x4124, 0xB9FC, 0x1040, 0x1088, 0xFDF0, 0x1020, 0x9444, 0x5BFE, 0x5022, 0x1D24, 0xE222, 0x4060}},
    {37838, 2, {0x2000, 0x21FC, 0x5124, 0x49FC, 0x8124, 0x79FC, 0x20A8, 0x20A8, 0xFBFE, 0x20A8, 0x28A8, 0xB1FC, 0x6020, 0x3BFE, 0xE020, 0x20}},
    {37839, 2, {0x2088, 0x23DE, 0x5088, 0x49DC, 0x8088, 0x7BDE, 0x2088, 0x2000, 0xFBFC, 0x2004, 0x2804, 0xB1FC, 0x6004, 0x3804, 0xE3FC, 0x4}},
    {37840, 2, {0x2000, 0x23DE, 0x5042, 0x4A52, 0x814A, 0x7A52, 0x2028, 0x20C4, 0xFB12, 0x2060, 0x2988, 0xB032, 0x61C4, 0x3818, 0xE060, 0x380}},
    {37841, 2, {0x1040, 0x1020, 0x2BFE, 0x2488, 0x4050, 0xBBFE, 0x1222, 0x12FA, 0xFE22, 0x12FA, 0x968A, 0x5A8A, 0x52FA, 0x1E02, 0xE20A, 0x4204}},
    {37842, 2, {0x2040, 0x2088, 0x51FC, 0x4908, 0x8252, 0x7BFE, 0x2050, 0x2188, 0xFE26, 0x20C0, 0x2B10, 0xB064, 0x6388, 0x3830, 0xE0C0, 0x700}},
    {37843, 2, {0x1020, 0x1040, 0x29FC, 0x2524, 0x417C, 0xB98C, 0x1154, 0x1124, 0xFD54, 0x11FC, 0x9420, 0x5810, 0x5154, 0x1D42, 0xE14A, 0x4238}},
    {37844, 2, {0x1040, 0x1020, 0x2BFE, 0x2602, 0x4000, 0xB9FC, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x9524, 0x59FC, 0x5000, 0x1C88, 0xE104, 0x4202}},
    {37845, 2, {0x2040, 0x2020, 0x53FE, 0x4A50, 0x8250, 0x7BFE, 0x2252, 0x2252, 0xFBFE, 0x2200, 0x2A92, 0xB2D4, 0x6298, 0x3C92, 0xE4D2, 0x88E}},
    {37846, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2462, 0x27BE, 0x2318, 0x2DF6, 0x2040, 0x47FC, 0x4248, 0x8FFE}},
    {37847, 2, {0x2000, 0x23DC, 0x5294, 0x4BD4, 0x8254, 0x7BC8, 0x2288, 0x2294, 0xFBE2, 0x2000, 0x2820, 0xB1FC, 0x6020, 0x3820, 0xE3FE, 0x0}},
    {37848, 2, {0x2088, 0x2288, 0x529E, 0x4AAA, 0x8294, 0x7B88, 0x2090, 0x20A4, 0xF884, 0x27BE, 0x2A84, 0xB2A4, 0x6294, 0x3A84, 0xE494, 0x88}},
    {37849, 2, {0x1020, 0x1222, 0x2BFE, 0x2490, 0x4088, 0xB9FE, 0x1310, 0x1510, 0xFDFE, 0x1110, 0x9510, 0x59FE, 0x5110, 0x1D10, 0xE1FE, 0x4100}},
    {37850, 2, {0x200C, 0x200A, 0x5008, 0x4BFE, 0x8208, 0x7A48, 0x226A, 0x224A, 0xFBFA, 0x224C, 0x2A4C, 0xB2EC, 0x635A, 0x3C4A, 0xE4D6, 0x822}},
    {37851, 2, {0x1020, 0x1124, 0x28A8, 0x27FE, 0x4202, 0xB8F8, 0x1088, 0x1088, 0xFCF8, 0x1020, 0x95FC, 0x5924, 0x5124, 0x1D34, 0xE128, 0x4020}},
    {37852, 2, {0x1020, 0x1124, 0x28A8, 0x27FE, 0x4202, 0xB8F8, 0x1088, 0x1088, 0xFCF8, 0x1020, 0x9420, 0x59FC, 0x5020, 0x1C20, 0xE3FE, 0x4000}},
    {37853, 2, {0x1000, 0x11F8, 0x2908, 0x25F8, 0x4108, 0xB9F8, 0x1000, 0x13FC, 0xFE94, 0x13FC, 0x9400, 0x59F8, 0x5090, 0x1C60, 0xE198, 0x4606}},
    {37854, 2, {0x1010, 0x11FE, 0x2910, 0x257C, 0x4114, 0xB9FE, 0x1114, 0x117C, 0xFD10, 0x117C, 0x9554, 0x597C, 0x5154, 0x1E7C, 0xE254, 0x444C}},
    {37855, 2, {0x1020, 0x1010, 0x29FE, 0x2400, 0x4084, 0xB848, 0x11FE, 0x1110, 0xFD50, 0x117E, 0x9590, 0x5910, 0x517E, 0x1E10, 0xE210, 0x44FE}},
    {37856, 2, {0x2020, 0x247C, 0x5284, 0x4948, 0x8030, 0x78CE, 0x2610, 0x227C, 0xFA10, 0x227C, 0x2A10, 0xB2FE, 0x6210, 0x3A10, 0xE5FE, 0x800}},
    {37857, 2, {0x1040, 0x1020, 0x29FC, 0x2488, 0x4050, 0xBBFE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5090, 0x1C92, 0xE112, 0x460E}},
    {37858, 2, {0x1000, 0x13FE, 0x2850, 0x25FC, 0x4154, 0xB954, 0x11FC, 0x1000, 0xFDFC, 0x1000, 0x97FE, 0x5820, 0x50A8, 0x1D24, 0xE2A2, 0x4040}},
    {37859, 2, {0x1040, 0x1020, 0x29FE, 0x2500, 0x4148, 0xB948, 0x11FE, 0x1148, 0xFD48, 0x1148, 0x9578, 0x5900, 0x52D4, 0x1EAA, 0xE52A, 0x4800}},
    {37860, 2, {0x1020, 0x11FC, 0x2924, 0x27FE, 0x4124, 0xB9FC, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x9440, 0x5BFE, 0x5088, 0x1DD0, 0xE070, 0x438C}},
    {37861, 2, {0x1020, 0x1010, 0x2BFE, 0x2602, 0x4080, 0xB8FE, 0x1110, 0x1320, 0xFD7C, 0x1144, 0x9544, 0x597C, 0x5144, 0x1D44, 0xE17C, 0x4144}},
    {37862, 2, {0x1124, 0x1124, 0x2A24, 0x24A4, 0x4154, 0x7D4A, 0x9392, 0x1508, 0x7908, 0x1128, 0x952E, 0x5928, 0x5128, 0x1D58, 0xE14E, 0x4180}},
    {37863, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x4148, 0x86C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37864, 2, {0x804, 0x7F78, 0x840, 0x3E40, 0x2A7E, 0x3E48, 0x2A48, 0xFF48, 0x988, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37865, 2, {0x2100, 0x2102, 0x511C, 0x4FD0, 0x8110, 0x77D0, 0x255E, 0x27D4, 0xF554, 0x27D4, 0xA114, 0x77D4, 0x2114, 0x3124, 0xC124, 0x144}},
    {37866, 2, {0x2090, 0x2090, 0x57FE, 0x4890, 0x83FC, 0x7A94, 0x23FC, 0x2294, 0xFBFC, 0x2000, 0x29F8, 0xB108, 0x61F8, 0x3908, 0xE1F8, 0x108}},
    {37867, 2, {0x2062, 0x238A, 0x208A, 0x57EA, 0x888A, 0x71CA, 0x22A2, 0x248A, 0xF8A4, 0x2020, 0x2BFE, 0xB070, 0x60A8, 0x3924, 0xE222, 0x20}},
    {37868, 2, {0x2040, 0x207C, 0x5040, 0x4BFE, 0x8242, 0x7A78, 0x23C4, 0x223C, 0xFA08, 0x22F0, 0x2B24, 0xB2A8, 0x65FE, 0x3C20, 0xE8A0, 0x40}},
    {37869, 2, {0x201C, 0x21E0, 0x5020, 0x4924, 0x80A8, 0x7BFE, 0x2070, 0x20A8, 0xF924, 0x2222, 0x2840, 0xB024, 0x62A2, 0x3A8A, 0xE488, 0x78}},
    {37870, 2, {0x1020, 0x1010, 0x29FE, 0x2510, 0x417C, 0xB914, 0x11FE, 0x1114, 0xFD7C, 0x1110, 0x9592, 0x5954, 0x5138, 0x1E54, 0xE292, 0x4430}},
    {37871, 2, {0x2020, 0x2020, 0x53FE, 0x4820, 0x82AA, 0x7924, 0x22AA, 0x2020, 0xFAAA, 0x2124, 0x2AAA, 0xB050, 0x6050, 0x3888, 0xE104, 0x602}},
    {37872, 2, {0x2020, 0x2222, 0x5222, 0x4BFE, 0x8000, 0x7BDE, 0x2252, 0x2252, 0xFBDE, 0x2252, 0x2A52, 0xB3DE, 0x6252, 0x3A52, 0xE55A, 0x8A4}},
    {37873, 2, {0x1020, 0x11FC, 0x2888, 0x2450, 0x43FE, 0xB800, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x95FC, 0x5820, 0x53FE, 0x1C20, 0xE020, 0x4020}},
    {37874, 2, {0x2020, 0x23FE, 0x5048, 0x48F0, 0x8020, 0x7BFE, 0x2242, 0x2090, 0xF9F8, 0x2028, 0x2920, 0xB1FC, 0x6220, 0x3BFE, 0xE020, 0x20}},
    {37875, 2, {0x2104, 0x2088, 0x5000, 0x4BFE, 0x8222, 0x7AAA, 0x2272, 0x2222, 0xFBFE, 0x2000, 0x29FC, 0xB104, 0x61FC, 0x3904, 0xE1FC, 0x104}},
    {37876, 2, {0x2010, 0x2790, 0x54BC, 0x4CA4, 0x84C4, 0x77A8, 0x2110, 0x2128, 0xF146, 0x25C0, 0xA53C, 0x7524, 0x2524, 0x35A4, 0xCE3C, 0x24}},
    {37877, 2, {0x1088, 0x13FE, 0x2888, 0x2400, 0x41FC, 0xB8A8, 0x10A8, 0x13FE, 0xFCA8, 0x10A8, 0x95FC, 0x5820, 0x53FE, 0x1C20, 0xE020, 0x4020}},
    {37878, 2, {0x10A0, 0x1090, 0x29FE, 0x2720, 0x41FC, 0xB920, 0x11FC, 0x1120, 0xFDFE, 0x1100, 0x9420, 0x5BFE, 0x5070, 0x1CA8, 0xE326, 0x4020}},
    {37879, 2, {0x1050, 0x1252, 0x2954, 0x2450, 0x43FE, 0xB888, 0x1050, 0x13FE, 0xFC20, 0x11FC, 0x9420, 0x5BFE, 0x5050, 0x1C88, 0xE104, 0x4602}},
    {37880, 2, {0x2020, 0x2020, 0x53FE, 0x4820, 0x81FC, 0x7924, 0x21FC, 0x2124, 0xF9FC, 0x2022, 0x2BFE, 0xB042, 0x6024, 0x3AA2, 0xE28A, 0x478}},
    {37881, 2, {0x2000, 0x23BE, 0x50A2, 0x48A2, 0x80BE, 0x7B88, 0x2208, 0x223E, 0xFA2A, 0x23AA, 0x28AA, 0xB0BE, 0x6088, 0x388A, 0xE57E, 0x202}},
    {37882, 2, {0x2024, 0x23A8, 0x2092, 0x5514, 0x8A08, 0x7404, 0x23BA, 0x20A8, 0xF8C6, 0x2380, 0x2A3C, 0xB384, 0x60A8, 0x3890, 0xE2A8, 0x144}},
    {37883, 2, {0x2020, 0x2124, 0x50A8, 0x4BFE, 0x80A8, 0x7924, 0x2222, 0x2104, 0xF904, 0x21DE, 0x2A44, 0xB554, 0x609E, 0x3904, 0xE204, 0x404}},
    {37884, 2, {0x2210, 0x2210, 0x53DE, 0x4D28, 0x80A4, 0x7820, 0x23FE, 0x2020, 0xFBFE, 0x2222, 0x2A2A, 0xB274, 0x60A8, 0x3924, 0xE622, 0x20}},
    {37885, 2, {0x2040, 0x23F8, 0x5048, 0x4FFE, 0x8048, 0x73F8, 0x2040, 0x2554, 0xF75C, 0x2444, 0xA7FC, 0x7444, 0x275C, 0x3554, 0xC554, 0x844}},
    {37886, 2, {0x2288, 0x2288, 0x57C8, 0x4A90, 0x829E, 0x77D4, 0x2024, 0x27D4, 0xF454, 0x27D4, 0xA454, 0x77C8, 0x2448, 0x3454, 0xC464, 0x4C2}},
    {37887, 2, {0x2020, 0x2124, 0x50A8, 0x4BFE, 0x8202, 0x78F8, 0x2088, 0x2088, 0xF8F8, 0x2020, 0x2BFE, 0xB070, 0x60A8, 0x3924, 0xE222, 0x20}},
    {37888, 2, {0x1020, 0x11FC, 0x2924, 0x25FC, 0x4020, 0xBBFE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1DFC, 0xE088, 0x4104}},
    {37889, 2, {0x2240, 0x2242, 0x525C, 0x4FF0, 0x8250, 0x7BD0, 0x225E, 0x23D4, 0xFA54, 0x2254, 0x2FF4, 0xB014, 0x6294, 0x3C64, 0xE824, 0x44}},
    {37890, 2, {0x2000, 0x23FE, 0x5050, 0x4BDE, 0x8252, 0x7A52, 0x23DE, 0x2050, 0xFBFE, 0x2222, 0x2A22, 0xB3FE, 0x6222, 0x3A22, 0xE3FE, 0x202}},
    {37891, 2, {0x2020, 0x2020, 0x51FC, 0x4820, 0x83FE, 0x7908, 0x239C, 0x2108, 0xF988, 0x263E, 0x2800, 0xB3FE, 0x6090, 0x3890, 0xE112, 0x20E}},
    {37892, 2, {0x1088, 0x1088, 0x29FC, 0x2488, 0x4088, 0xBBFE, 0x1020, 0x11FC, 0xFD24, 0x11FC, 0x9524, 0x59FC, 0x5000, 0x1C88, 0xE104, 0x4202}},
    {37893, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4144, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37894, 2, {0x108, 0x7890, 0x4BFC, 0x5040, 0x49A4, 0x4458, 0x55B4, 0x4852, 0x41B0, 0x46C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37895, 2, {0x103C, 0x13E0, 0x2924, 0x24A8, 0x43FE, 0xB8A8, 0x1124, 0x1202, 0xFDFC, 0x1124, 0x9524, 0x59FC, 0x5124, 0x1D24, 0xE1FC, 0x4104}},
    {37896, 2, {0x1018, 0x11E0, 0x2840, 0x27FE, 0x4088, 0xB974, 0x1252, 0x1070, 0xFC00, 0x11FC, 0x9504, 0x5974, 0x5154, 0x1D74, 0xE104, 0x410C}},
    {37897, 2, {0x1000, 0x13DE, 0x2A52, 0x27DE, 0x4210, 0xBA52, 0x11CE, 0x1000, 0xFC88, 0x13FE, 0x9488, 0x5888, 0x53FE, 0x1C88, 0xE104, 0x4202}},
    {37898, 2, {0x2000, 0x277C, 0x5544, 0x4D7C, 0x8544, 0x757C, 0x2600, 0x25FE, 0xF540, 0x257E, 0xA5AA, 0x752A, 0x264A, 0x3492, 0xC42A, 0x444}},
    {37899, 2, {0x2000, 0x227C, 0x5144, 0x497C, 0x8444, 0x7A7C, 0x2200, 0x21FE, 0xF940, 0x227E, 0x2EAA, 0xB32A, 0x624A, 0x3A92, 0xE22A, 0x44}},
    {37900, 2, {0x2080, 0x20F8, 0x5108, 0x4BFE, 0x8512, 0x7922, 0x21FE, 0x2040, 0xF8A2, 0x2354, 0x2898, 0xB334, 0x6054, 0x3892, 0xE350, 0x20}},
    {37901, 2, {0x11F8, 0x1050, 0x2820, 0x27FC, 0x40A4, 0xB928, 0x12A0, 0x1040, 0xFDFC, 0x1154, 0x958C, 0x5974, 0x5154, 0x1D74, 0xE104, 0x410C}},
    {37902, 2, {0x1090, 0x1088, 0x29FE, 0x2510, 0x4310, 0xBDFE, 0x1110, 0x1110, 0xFDFE, 0x1110, 0x9510, 0x59FE, 0x5100, 0x1EA4, 0xE252, 0x4452}},
    {37903, 2, {0x1088, 0x1050, 0x2BFE, 0x2450, 0x41FC, 0xB954, 0x118C, 0x1174, 0xFD04, 0x11FC, 0x9408, 0x5BFE, 0x5108, 0x1C88, 0xE0A8, 0x4010}},
    {37904, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x4154, 0xB888, 0x1104, 0x12FA, 0xFC88, 0x10F8, 0x9488, 0x58F8, 0x5020, 0x1CA8, 0xE124, 0x4060}},
    {37905, 2, {0x2088, 0x2088, 0x5108, 0x493E, 0x8248, 0x7B88, 0x209C, 0x2100, 0xFA40, 0x23DC, 0x2814, 0xB054, 0x62B4, 0x3A94, 0xE21C, 0x14}},
    {37906, 2, {0x2088, 0x22AA, 0x22DC, 0x5488, 0x8954, 0x7222, 0x2000, 0x23FE, 0xFA42, 0x2444, 0x2BFC, 0xB044, 0x6044, 0x3884, 0xE114, 0x208}},
    {37907, 2, {0x2108, 0x2088, 0x53C8, 0x4810, 0x83DE, 0x7A64, 0x23D4, 0x2014, 0xFBD4, 0x2054, 0x2894, 0xB0E8, 0x6388, 0x3894, 0xE294, 0x122}},
    {37908, 2, {0x2000, 0x23FE, 0x5050, 0x4BFE, 0x8252, 0x7BFE, 0x2000, 0x21FC, 0xF904, 0x21FC, 0x2904, 0xB1FC, 0x6020, 0x3BFE, 0xE020, 0x20}},
    {37909, 2, {0x2000, 0x23DE, 0x5088, 0x4AA8, 0x83FE, 0x7988, 0x22DA, 0x24A6, 0xF9F8, 0x2108, 0x2908, 0xB1F8, 0x6108, 0x3908, 0xE1F8, 0x108}},
    {37910, 2, {0x2124, 0x2124, 0x52AA, 0x4BAE, 0x8124, 0x7AAA, 0x23AE, 0x2124, 0xFBFE, 0x2110, 0x2914, 0xB114, 0x628A, 0x3A4A, 0xE216, 0x422}},
    {37911, 2, {0x23DE, 0x2252, 0x53DE, 0x4A52, 0x83DE, 0x7A02, 0x22F2, 0x2292, 0xFAF2, 0x2292, 0x2AF2, 0xB292, 0x6292, 0x3B32, 0xE202, 0x206}},
    {37912, 2, {0x1020, 0x11FC, 0x2888, 0x2450, 0x43FE, 0xB800, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x95FC, 0x5820, 0x51FC, 0x1C20, 0xE3FE, 0x4000}},
    {37913, 2, {0x1010, 0x10D4, 0x2848, 0x254A, 0x4284, 0xBC84, 0x117A, 0x1000, 0xFEFC, 0x1084, 0x94FC, 0x5800, 0x5084, 0x1E48, 0xF1FE, 0x4000}},
    {37914, 2, {0x1000, 0x13FE, 0x2850, 0x25DC, 0x4104, 0xB904, 0x11DC, 0x1050, 0xFC50, 0x13FE, 0x9440, 0x5820, 0x50A4, 0x1E8A, 0xE28A, 0x4478}},
    {37915, 2, {0x1000, 0x11FC, 0x2904, 0x25FC, 0x4104, 0xB9FC, 0x1020, 0x13FE, 0xFC00, 0x11FC, 0x9504, 0x59FC, 0x5020, 0x1D24, 0xE222, 0x4060}},
    {37916, 2, {0x820, 0x7F20, 0x223E, 0x3E44, 0x84, 0x3E28, 0x410, 0x7E28, 0x944, 0x2EC2, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37917, 2, {0x2000, 0x27FE, 0x5400, 0x4E28, 0x8548, 0x77EE, 0x2492, 0x2484, 0xF6A0, 0x26A8, 0xA7E8, 0x7488, 0x2494, 0x3514, 0xC924, 0x1242}},
    {37918, 2, {0x1000, 0x11FC, 0x2804, 0x24FC, 0x4004, 0xB9FC, 0x1000, 0x11DC, 0xFC94, 0x13DC, 0x9408, 0x5BFE, 0x5108, 0x1C88, 0xE0A8, 0x4010}},
    {37919, 2, {0x1088, 0x1088, 0x2BDE, 0x2488, 0x47FE, 0xB888, 0x1154, 0x1222, 0xFDFC, 0x1104, 0x9504, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37920, 2, {0x2108, 0x2090, 0x53FC, 0x4890, 0x8294, 0x7998, 0x2090, 0x27FE, 0xF800, 0x21F8, 0x2908, 0xB108, 0x61F8, 0x3908, 0xE108, 0x1F8}},
    {37921, 2, {0x2094, 0x2092, 0x53F2, 0x4890, 0x87FE, 0x7010, 0x23D2, 0x2252, 0xF252, 0x23D4, 0xA454, 0x728C, 0x20EA, 0x370A, 0xC016, 0x22}},
    {37922, 2, {0x2288, 0x2108, 0xFABE, 0x2008, 0x729C, 0xA92A, 0x2288, 0x2108, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37923, 2, {0x1124, 0x10A8, 0x2BFE, 0x2602, 0x40F8, 0xB888, 0x10F8, 0x100C, 0xFDF0, 0x1020, 0x95FC, 0x5820, 0x53FE, 0x1C20, 0xE0A0, 0x4040}},
    {37924, 2, {0x21F8, 0x2108, 0x51F8, 0x4D0A, 0x85FA, 0x750A, 0x25FA, 0x2402, 0xF79E, 0x2090, 0xA090, 0x7F9E, 0x2492, 0x3492, 0xC492, 0x892}},
    {37925, 2, {0x1088, 0x1050, 0x2BFE, 0x2420, 0x41FC, 0xB820, 0x13FE, 0x1124, 0xFCA8, 0x13FE, 0x9400, 0x59FC, 0x5104, 0x1D04, 0xE1FC, 0x4104}},
    {37926, 2, {0x2000, 0x23DE, 0x5252, 0x4BDE, 0x8252, 0x7BDE, 0x2202, 0x22FA, 0xFA52, 0x2252, 0x2BFE, 0xB252, 0x6252, 0x3A92, 0xE20A, 0x204}},
    {37927, 2, {0x2000, 0x23DE, 0x5252, 0x4BDE, 0x8252, 0x7BDE, 0x2202, 0x22FA, 0xFA8A, 0x22FA, 0x2A8A, 0xB2FA, 0x628A, 0x3A02, 0xE20A, 0x204}},
    {37928, 2, {0x1090, 0x17FC, 0x2894, 0x27FC, 0x4290, 0xBBFE, 0x1112, 0x13FA, 0xFD0E, 0x11F8, 0x9508, 0x59F8, 0x5108, 0x1DF8, 0xE090, 0x4108}},
    {37929, 2, {0x1044, 0x1224, 0x2928, 0x247E, 0x4010, 0xB828, 0x134A, 0x111C, 0xFD28, 0x114C, 0x951A, 0x592A, 0x5148, 0x1D10, 0xE2FE, 0x4400}},
    {37930, 2, {0x1020, 0x103E, 0x2820, 0x25FE, 0x4122, 0xB9FC, 0x1122, 0x117E, 0xFD54, 0x117C, 0x9554, 0x597C, 0x5110, 0x1EFE, 0xE222, 0x44C6}},
    {37931, 2, {0x20A0, 0x2090, 0x51FE, 0x4B20, 0x81FC, 0x7920, 0x21FC, 0x2120, 0xF9FE, 0x2100, 0x2BDE, 0xB252, 0x6252, 0x3A72, 0xE202, 0x206}},
    {37932, 2, {0x2020, 0x213C, 0x5120, 0x4BFE, 0x8010, 0x7814, 0x23FE, 0x2210, 0xFBF0, 0x2254, 0x2A54, 0xB2D4, 0x6368, 0x3C4A, 0xE896, 0x322}},
    {37933, 2, {0x2040, 0x2080, 0x53FE, 0x4A22, 0x82AA, 0x7A22, 0x23FE, 0x2272, 0xFAAA, 0x2202, 0x2820, 0xB3FE, 0x6050, 0x3888, 0xE104, 0x602}},
    {37934, 2, {0x2020, 0x23FE, 0x5288, 0x4A50, 0x83FE, 0x7A50, 0x22FC, 0x2254, 0xFBFE, 0x2254, 0x2AFC, 0xB250, 0x62D8, 0x3D54, 0xE652, 0x850}},
    {37935, 2, {0x1088, 0x1088, 0x2BFE, 0x24A8, 0x4022, 0x7DFA, 0x9024, 0x1028, 0x7BFE, 0x1040, 0x94FC, 0x5984, 0x52FC, 0x1C84, 0xE0FC, 0x4084}},
    {37936, 2, {0x10F8, 0x1088, 0x2888, 0x24F8, 0x4000, 0xB9DC, 0x1154, 0x1154, 0xFDDC, 0x1020, 0x95FE, 0x5870, 0x50A8, 0x1D24, 0xE222, 0x4020}},
    {37937, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x4104, 0xBAFA, 0x1000, 0x1000, 0xFDDC, 0x1154, 0x9554, 0x59DC, 0x5088, 0x1C88, 0xE154, 0x4222}},
    {37938, 2, {0x1000, 0x11FC, 0x2954, 0x2554, 0x41FC, 0xB880, 0x11FC, 0x1244, 0xFDF4, 0x1154, 0x9554, 0x59F4, 0x5044, 0x1DF4, 0xE014, 0x4008}},
    {37939, 2, {0x1000, 0x11FC, 0x2820, 0x27FE, 0x4222, 0xB9AC, 0x1020, 0x11AC, 0xFC00, 0x11FC, 0x9524, 0x5924, 0x51FC, 0x1D24, 0xE124, 0x41FC}},
    {37940, 2, {0x2010, 0x23C8, 0x527E, 0x4A40, 0x8262, 0x7BD4, 0x2200, 0x223E, 0xFBC8, 0x2248, 0x2E7E, 0xB248, 0x6248, 0x3BC8, 0xE248, 0x8}},
    {37941, 2, {0x208C, 0x23EA, 0x5088, 0x4BFE, 0x8008, 0x7BE8, 0x222A, 0x23EA, 0xF80A, 0x23EC, 0x288C, 0xB3EC, 0x608A, 0x38EA, 0xE396, 0x22}},
    {37942, 2, {0x1000, 0x11FC, 0x2954, 0x2554, 0x41FC, 0xB800, 0x13FE, 0x1000, 0xFDFC, 0x1104, 0x95FC, 0x5862, 0x5094, 0x1D88, 0xE2A4, 0x40C2}},
    {37943, 2, {0x1088, 0x13FE, 0x2888, 0x2400, 0x4148, 0xBBFE, 0x1148, 0x1178, 0xFD00, 0x11FC, 0x9420, 0x5BFE, 0x5070, 0x1CA8, 0xE326, 0x4020}},
    {37944, 2, {0x1000, 0x11FC, 0x2954, 0x25FC, 0x4020, 0xB9FC, 0x1020, 0x13FE, 0xFC88, 0x1050, 0x95FC, 0x5820, 0x53FE, 0x1C20, 0xE020, 0x4020}},
    {37945, 2, {0x2000, 0x227C, 0x5144, 0x4974, 0x8054, 0x7854, 0x26FE, 0x2282, 0xFABA, 0x22AA, 0x2ABA, 0xB282, 0x628A, 0x3A84, 0xE500, 0x8FE}},
    {37946, 2, {0x1020, 0x1124, 0x28A8, 0x27FE, 0x4202, 0xB8F8, 0x1088, 0x1088, 0xFCF8, 0x1000, 0x95FC, 0x5924, 0x51FC, 0x1D24, 0xE1FC, 0x4104}},
    {37947, 2, {0x2040, 0x207C, 0x5040, 0x4BFE, 0x8242, 0x7A78, 0x23C4, 0x223C, 0xFA00, 0x23FE, 0x2A40, 0xB3A4, 0x6258, 0x3DB4, 0xE452, 0x9B0}},
    {37948, 2, {0x1020, 0x11FC, 0x2820, 0x2488, 0x43FE, 0xB888, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x95FC, 0x5904, 0x51FC, 0x1C88, 0xE104, 0x4202}},
    {37949, 2, {0x1010, 0x1210, 0x297C, 0x2410, 0x40FE, 0xB844, 0x1328, 0x11FE, 0xFD10, 0x117C, 0x9510, 0x59FE, 0x5110, 0x1D10, 0xE2FE, 0x4400}},
    {37950, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2310, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37951, 2, {0x1040, 0x1020, 0x29FC, 0x2488, 0x4050, 0xBBFE, 0x1000, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5020, 0x1E94, 0xE28A, 0x447A}},
    {37952, 2, {0x100E, 0x11F0, 0x28A2, 0x2454, 0x41FE, 0xB922, 0x1014, 0x1152, 0xFD4A, 0x1238, 0x9440, 0x587C, 0x5084, 0x1D48, 0xE030, 0x41CE}},
    {37953, 2, {0x2020, 0x2088, 0x5252, 0x4B26, 0x8252, 0x7A8A, 0x2222, 0x23FE, 0xF888, 0x2144, 0x2A7A, 0xB088, 0x6150, 0x3820, 0xE0D8, 0x706}},
    {37954, 2, {0x101C, 0x11E0, 0x2820, 0x27FE, 0x4020, 0xB9FC, 0x11AC, 0x1174, 0xFDFC, 0x1020, 0x95FC, 0x5820, 0x53FE, 0x1C00, 0xE154, 0x422A}},
    {37955, 2, {0x1000, 0x11DC, 0x2844, 0x2554, 0x40CC, 0xB954, 0x1000, 0x1048, 0xFCFE, 0x1190, 0x96FC, 0x5890, 0x50FC, 0x1C90, 0xE0FE, 0x4080}},
    {37956, 2, {0x2020, 0x23FE, 0x5020, 0x49FC, 0x8000, 0x7BFE, 0x2002, 0x21FC, 0xF820, 0x23FE, 0x2800, 0xB3FE, 0x6004, 0x3BBE, 0xE2A4, 0x38C}},
    {37957, 2, {0x2088, 0x22AA, 0x22DC, 0x5488, 0x8954, 0x7222, 0x2000, 0x23FE, 0xFA22, 0x2020, 0x2BFE, 0xB070, 0x60A8, 0x3924, 0xE622, 0x20}},
    {37958, 2, {0x2144, 0x2144, 0x5554, 0x4B64, 0x814E, 0x7BE4, 0x2004, 0x222C, 0xF944, 0x23E4, 0x2884, 0xB3E4, 0x6084, 0x38E4, 0xE714, 0x8}},
    {37959, 2, {0x2040, 0x2020, 0x53FE, 0x4888, 0x8052, 0x77AC, 0x22AA, 0x22A8, 0xF5AC, 0x2000, 0xA104, 0x71FC, 0x2104, 0x31FC, 0xC104, 0x204}},
    {37960, 2, {0x2000, 0x27FC, 0x2248, 0x5444, 0x8FFE, 0x7444, 0x26EC, 0x2554, 0xFEEC, 0x2444, 0x2EEC, 0xB554, 0x66EC, 0x3C44, 0xE454, 0x408}},
    {37961, 2, {0x1088, 0x13FE, 0x2888, 0x2420, 0x41FC, 0xB820, 0x13FE, 0x1040, 0xFC88, 0x11FC, 0x9400, 0x59FC, 0x5154, 0x1D54, 0xE7FE, 0x4000}},
    {37962, 2, {0x1088, 0x13FE, 0x28A8, 0x2490, 0x41FE, 0xBB20, 0x11FC, 0x1120, 0xFDFC, 0x1120, 0x95FE, 0x5900, 0x53FC, 0x1C88, 0xE070, 0x438E}},
    {37963, 2, {0x1000, 0xFEFE, 0x1054, 0x7CA8, 0x5454, 0x7C7C, 0x5410, 0xFEFE, 0x1100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37964, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x81FC, 0x7948, 0x2250, 0x21FC, 0xFB04, 0x21FC, 0x2904, 0xB1FC, 0x6104, 0x39FC, 0xE088, 0x104}},
    {37965, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2388, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {37966, 2, {0x23FE, 0x2220, 0x52FC, 0x4AA4, 0x82FC, 0x7A20, 0x23FE, 0x2284, 0xFAFC, 0x2284, 0x2AFC, 0xB284, 0x62FC, 0x3A48, 0xE284, 0x3FE}},
    {37967, 2, {0x1040, 0x1020, 0x2BFE, 0x2622, 0x4154, 0xB94A, 0x123A, 0x1000, 0xFDFC, 0x1154, 0x97FE, 0x5800, 0x51FC, 0x1C20, 0xE0A0, 0x4040}},
    {37968, 2, {0x1000, 0x11FC, 0x2820, 0x27FE, 0x4222, 0xB9AC, 0x1020, 0x11AC, 0xFC00, 0x13FE, 0x9420, 0x59FC, 0x5154, 0x1D54, 0xE154, 0x410C}},
    {37969, 2, {0x2008, 0x23E8, 0x5288, 0x4BEE, 0x8228, 0x7BF4, 0x2282, 0x23E2, 0xF800, 0x21FC, 0x2954, 0xB154, 0x6154, 0x3954, 0xE7FE, 0x0}},
    {37970, 2, {0x7E40, 0x487E, 0x7E90, 0x4308, 0x7EFE, 0x48AA, 0x7EFE, 0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x1110, 0x920, 0xFFFE}},
    {37971, 2, {0x1010, 0x107C, 0x2A54, 0x257C, 0x4110, 0xB8FE, 0x1000, 0x137C, 0xFD44, 0x117C, 0x9540, 0x597C, 0x5144, 0x1D7C, 0xE280, 0x447E}},
    {37972, 2, {0x2040, 0x2020, 0x53FE, 0x4A02, 0x8514, 0x79DE, 0x2252, 0x2554, 0xFA88, 0x2174, 0x2A02, 0xB1FC, 0x6020, 0x3924, 0xE222, 0x60}},
    {37973, 2, {0x2042, 0x239C, 0x5210, 0x4BDE, 0x8294, 0x7A94, 0x2420, 0x23FC, 0xFA04, 0x23FC, 0x2A04, 0xB3FC, 0x6204, 0x3BFC, 0xE108, 0x204}},
    {37974, 2, {0x2108, 0x27FE, 0x5108, 0x4800, 0x83FC, 0x7A94, 0x2294, 0x23FC, 0xF828, 0x23FE, 0x2A20, 0xB324, 0x62A8, 0x3A92, 0xE42A, 0x8C6}},
    {37975, 2, {0x1050, 0x1190, 0x289E, 0x27EA, 0x408A, 0xB9D2, 0x128A, 0x10A4, 0xFC50, 0x1088, 0x9726, 0x58A8, 0x5070, 0x1CA8, 0xE124, 0x4060}},
    {37976, 2, {0x23F8, 0x2248, 0x5248, 0x4BF8, 0x8248, 0x7248, 0x23F8, 0x2000, 0xFFBE, 0x2AAA, 0xAAAA, 0x7FBE, 0x2AAA, 0x3AAA, 0xCFBE, 0x8A2}},
    {37977, 2, {0x2394, 0x2110, 0x27BE, 0x5118, 0x8BAA, 0x7546, 0x2218, 0x23E0, 0xFA04, 0x21FC, 0x2800, 0xB3F8, 0x6208, 0x3BF8, 0xE208, 0x3F8}},
    {37978, 2, {0x1088, 0x11DC, 0x2888, 0x25DE, 0x4088, 0xB954, 0x1222, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1DFC, 0xE088, 0x4104}},
    {37979, 2, {0x1020, 0x1010, 0x29FE, 0x2528, 0x417C, 0xB928, 0x11FE, 0x1110, 0xFD7C, 0x1154, 0x957C, 0x5954, 0x517C, 0x1E00, 0xE244, 0x4482}},
    {37980, 2, {0x1124, 0x10A8, 0x2BFE, 0x2602, 0x40F8, 0xB888, 0x11FC, 0x1104, 0xFDFC, 0x1104, 0x95FC, 0x5904, 0x51FC, 0x1C00, 0xE088, 0x4104}},
    {37981, 2, {0x2088, 0x23FE, 0x5088, 0x4820, 0x823C, 0x7944, 0x2038, 0x20D6, 0xFB38, 0x2110, 0x297C, 0xB110, 0x617E, 0x3910, 0xE310, 0x4FE}},
    {37982, 2, {0x2124, 0x2248, 0x5124, 0x4800, 0x83FC, 0x7A94, 0x2264, 0x2294, 0xFBFC, 0x2248, 0x2B68, 0xB248, 0x636A, 0x3A4A, 0xE246, 0x362}},
    {37983, 2, {0x1020, 0x13FE, 0x2820, 0x25FC, 0x4000, 0xBBFE, 0x1252, 0x13FE, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5104, 0x1DFC, 0xE088, 0x4104}},
    {37984, 2, {0x2444, 0x24E4, 0x58A8, 0x4AAA, 0x8EEE, 0x74A4, 0x2AAA, 0x2EEE, 0xF242, 0x2040, 0xAFFE, 0x70E0, 0x2150, 0x3248, 0xCC46, 0x40}},
    {37985, 2, {0x2020, 0x2120, 0x513C, 0x4920, 0x87FE, 0x7890, 0x2294, 0x236C, 0xFA04, 0x23FC, 0x2A94, 0xB294, 0x636C, 0x3A04, 0xE3FC, 0x4}},
    {37986, 2, {0x2020, 0x203C, 0x5020, 0x4BFE, 0x8222, 0x7AF8, 0x2224, 0x22FC, 0xFAA4, 0x22FC, 0x2AA4, 0xB2FC, 0x6420, 0x3D54, 0xE94A, 0x23A}},
    {37987, 2, {0x2040, 0x2020, 0x53FE, 0x4A50, 0x83FE, 0x7A52, 0x23FE, 0x2200, 0xFA94, 0x22D8, 0x2A90, 0xB2D2, 0x628E, 0x3C00, 0xE554, 0xA2A}},
    {37988, 2, {0x11FC, 0x1104, 0x29FC, 0x2504, 0x41FC, 0x7C88, 0x91FC, 0x1088, 0x7BFE, 0x1088, 0x9524, 0x5AAA, 0x5070, 0x1CA8, 0xE124, 0x4060}},
    {37989, 2, {0x1080, 0x11F8, 0x2A08, 0x25FC, 0x4124, 0xB9FC, 0x1124, 0x11FC, 0xFD54, 0x122A, 0x95FC, 0x5904, 0x51FC, 0x1D04, 0xE1FC, 0x4104}},
    {37990, 2, {0x27DE, 0x2512, 0x27D2, 0x5454, 0x8FC8, 0x7514, 0x27E2, 0x2000, 0xFBFC, 0x2204, 0x2BFC, 0xB204, 0x63FC, 0x3A04, 0xE3FC, 0x108}},
    {37991, 2, {0x1020, 0x13FE, 0x2A8A, 0x25FC, 0x4088, 0xB9FC, 0x1104, 0x11FC, 0xFD04, 0x11FC, 0x9504, 0x59FC, 0x5090, 0x1C94, 0xE112, 0x420E}},
    {37992, 2, {0x2110, 0x2090, 0x53DE, 0x4810, 0x825E, 0x7982, 0x23DE, 0x2010, 0xFBDE, 0x2250, 0x2BDE, 0xB250, 0x63DE, 0x3A50, 0xE252, 0x2CE}},
    {37993, 2, {0x2000, 0x23FE, 0x5020, 0x4BAE, 0x82AA, 0x7AAA, 0x23AE, 0x2020, 0xFBFE, 0x2020, 0x2BAE, 0xB2AA, 0x62AA, 0x3BAE, 0xE020, 0x3FE}},
    {37994, 2, {0x2040, 0x207C, 0x5040, 0x4BFC, 0x8244, 0x7BF0, 0x2244, 0x22FC, 0xFAA8, 0x22F8, 0x2AA8, 0xB2F8, 0x6200, 0x3DFC, 0xE554, 0xBFE}},
    {37995, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x1FF0, 0x920, 0x3FF8, 0x1010, 0x2828, 0x7C7C, 0x9292, 0x7C7C, 0x1010, 0x5454, 0xFEFE}},
    {37996, 2, {0x13D0, 0x1290, 0x2BDE, 0x2668, 0x43C4, 0xBA84, 0x13F8, 0x1108, 0xFDF8, 0x1108, 0x95F8, 0x5908, 0x51F8, 0x1C90, 0xE112, 0x460E}},
    {37997, 2, {0x23DE, 0x2252, 0x53DE, 0x4A52, 0x83DE, 0x7A22, 0x22FA, 0x2222, 0xFAFA, 0x22AA, 0x2AFA, 0xB2AA, 0x62FA, 0x3A72, 0xE2AA, 0x226}},
    {37998, 2, {0x2088, 0x23FE, 0x2088, 0x5412, 0x8AFE, 0x7010, 0x24FE, 0x2292, 0xFAFE, 0x2092, 0x29FE, 0xB292, 0x6604, 0x3AFE, 0xE244, 0x22C}},
    {37999, 2, {0x2148, 0x214C, 0x52AA, 0x4808, 0x83FE, 0x7948, 0x2368, 0x214A, 0xFB6A, 0x214A, 0x2B6C, 0xB14C, 0x616A, 0x3B8A, 0xE016, 0x22}},
    {38000, 2, {0x1020, 0x1050, 0x2888, 0x2574, 0x4202, 0xB9FC, 0x1154, 0x11FC, 0xFC00, 0x11FC, 0x9554, 0x5954, 0x51FC, 0x1D54, 0xE154, 0x410C}},
    {38001, 2, {0x21F0, 0x2210, 0x57FC, 0x4A04, 0x83FC, 0x7A24, 0x23B8, 0x2222, 0xF9FE, 0x2210, 0x2FFC, 0xB244, 0x63FC, 0x38D0, 0xE14A, 0x63E}},
    {38002, 2, {0x1020, 0x13FE, 0x2800, 0x25DC, 0x4154, 0xB9DC, 0x1088, 0x13FE, 0xFC88, 0x13FE, 0x9488, 0x5BFE, 0x5094, 0x1D88, 0xE2A4, 0x40C2}},
    {38003, 2, {0x2040, 0x27FE, 0x5492, 0x4BFC, 0x8090, 0x7BFC, 0x2090, 0x27FE, 0xF908, 0x23FC, 0x2D0A, 0xB1F8, 0x6240, 0x3A7C, 0xE540, 0x8FE}},
    {38004, 2, {0x2222, 0x23FE, 0x2090, 0x51FE, 0x8B10, 0x75FE, 0x2110, 0x21FE, 0xF910, 0x21FE, 0x2900, 0xB3FE, 0x628A, 0x3B76, 0xE252, 0x276}},
    {38005, 2, {0x2088, 0x23FE, 0x5088, 0x4BDE, 0x8252, 0x7BDE, 0x20A0, 0x2090, 0xF9FE, 0x2120, 0x2BFC, 0xB520, 0x61FC, 0x3920, 0xE1FE, 0x100}},
    {38006, 2, {0x2090, 0x2090, 0x57FE, 0x4894, 0x800A, 0x71FE, 0x2508, 0x27E8, 0xF14A, 0x21EA, 0xAF2A, 0x75EA, 0x254C, 0x39EA, 0xC216, 0x462}},
    {38007, 2, {0x23FC, 0x2108, 0x51F8, 0x4908, 0x81F8, 0x790E, 0x27F8, 0x2008, 0xFFFE, 0x2294, 0x2B9C, 0xB294, 0x639C, 0x3AD6, 0xE7BC, 0x84}},
    {38008, 2, {0x27FC, 0x2444, 0x57FC, 0x4C44, 0x87FC, 0x7000, 0x2FBE, 0x2AAA, 0xFFBE, 0x2AAA, 0xAFBE, 0x7040, 0x2040, 0x37FC, 0xC040, 0xFFE}},
    {38009, 2, {0x2040, 0x27FE, 0x5492, 0x4908, 0x8080, 0x7B1C, 0x2204, 0x239C, 0xFA04, 0x23FC, 0x2A48, 0xB368, 0x6248, 0x3B6A, 0xE246, 0x362}},
    {38010, 2, {0x23DE, 0x2252, 0x53DE, 0x4A52, 0x83DE, 0x7A52, 0x23DE, 0x2090, 0xF9FE, 0x2110, 0x2BFE, 0xB510, 0x61FE, 0x3910, 0xE1FE, 0x100}},
    {38011, 2, {0x22A8, 0x2248, 0x57BE, 0x4AA8, 0x875C, 0x7AAA, 0x2288, 0x2FFE, 0xF218, 0x25E4, 0xA842, 0x73F8, 0x2040, 0x37FC, 0xC040, 0xC0}},
    {38012, 2, {0x2000, 0x23FE, 0x5252, 0x4A52, 0x83FE, 0x7108, 0x2154, 0x225E, 0xF7B4, 0x211E, 0xA254, 0x77DE, 0x2014, 0x3554, 0xC55E, 0x10}},
    {38013, 2, {0x2528, 0x27BE, 0x2948, 0x57BE, 0x8B18, 0x75AA, 0x2946, 0x23FC, 0xFA04, 0x23FC, 0x2A04, 0xB3FC, 0x6204, 0x3BFC, 0xE108, 0x204}},
    {38014, 2, {0x2108, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0x100, 0x6C0, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {38015, 2, {0x5578, 0x1448, 0xFF48, 0x2286, 0x7F78, 0x848, 0x6B28, 0x4910, 0x7F28, 0x6C4, 0x1830, 0xEFEE, 0x100, 0x3FF8, 0x920, 0xFFFE}},
    {38016, 2, {0x2108, 0x21EC, 0x510A, 0x4FEA, 0x8928, 0x7BC8, 0x293E, 0x2FE8, 0xF948, 0x2BE8, 0xAA28, 0x7B68, 0x2AA8, 0x3BF4, 0xCAB4, 0x12A2}},
    {38017, 2, {0x23DE, 0x2252, 0x53DE, 0x4A52, 0x83DE, 0x78A0, 0x21FE, 0x2320, 0xFDFC, 0x2120, 0x29FE, 0xB000, 0x61FC, 0x3888, 0xE070, 0x78E}},
    {38018, 2, {0x2248, 0x2150, 0x57FC, 0x4C04, 0x81F0, 0x7110, 0x27FC, 0x2554, 0xF4E4, 0x27FC, 0xA040, 0x77FC, 0x2040, 0x3FFE, 0xC2A4, 0x452}},
    {38019, 2, {0x23FE, 0x2202, 0x53FE, 0x4A92, 0x8254, 0x7A92, 0x22FE, 0x22AA, 0xFAFE, 0x2322, 0x2AFA, 0xB2AA, 0x62FA, 0x3A22, 0xE5FA, 0x4}},
    {38020, 2, {0x2088, 0x23FE, 0x5088, 0x4BDE, 0x8252, 0x7BDE, 0x2252, 0x23DE, 0xFA22, 0x23FE, 0x2AAA, 0xB2FA, 0x6222, 0x3A72, 0xE2AA, 0x224}},
    {38021, 2, {0x1000, 0x1000, 0x1E00, 0x2000, 0x2000, 0x7C00, 0x9000, 0x1000, 0xFE00, 0x1000, 0x1000, 0x1200, 0x1400, 0x1800, 0x1000, 0x0}},
    {38022, 2, {0x1080, 0x1080, 0x1E80, 0x2080, 0x2080, 0x7C80, 0x9080, 0x1080, 0xFE80, 0x1080, 0x1080, 0x1284, 0x1484, 0x1884, 0x107C, 0x0}},
    {38023, 2, {0x1000, 0x1000, 0x3DFC, 0x2008, 0x4010, 0xBC20, 0x1040, 0x1040, 0xFC80, 0x1080, 0x1100, 0x1104, 0x1504, 0x1904, 0x10FC, 0x0}},
    {38024, 2, {0x1020, 0x1020, 0x1E20, 0x2020, 0x2020, 0x7C20, 0x93FE, 0x1020, 0xFE20, 0x1020, 0x1020, 0x1220, 0x1420, 0x1820, 0x1020, 0x20}},
    {38025, 2, {0x1000, 0x1000, 0x3DFE, 0x2020, 0x4020, 0xBC20, 0x1020, 0x1020, 0xFC20, 0x1020, 0x1020, 0x1020, 0x1420, 0x1820, 0x10A0, 0x40}},
    {38026, 2, {0x1004, 0x1004, 0x1E04, 0x2044, 0x2044, 0x7C44, 0x9044, 0x1044, 0xFE44, 0x1044, 0x1044, 0x1204, 0x1404, 0x1804, 0x1014, 0x8}},
    {38027, 2, {0x1040, 0x1040, 0x1E40, 0x2040, 0x2040, 0x7C60, 0x9050, 0x1048, 0xFE44, 0x1044, 0x1040, 0x1240, 0x1440, 0x1840, 0x1040, 0x40}},
    {38028, 2, {0x1000, 0x11FC, 0x1E04, 0x2008, 0x2010, 0x7C20, 0x9020, 0x1020, 0xFE20, 0x1020, 0x1020, 0x1220, 0x1420, 0x1820, 0x10A0, 0x40}},
    {38029, 2, {0x1020, 0x1020, 0x3C20, 0x2020, 0x4020, 0xBC20, 0x11FC, 0x1020, 0xFC20, 0x1020, 0x1020, 0x1020, 0x1420, 0x1820, 0x13FE, 0x0}},
    {38030, 2, {0x1008, 0x103C, 0x3DE0, 0x2020, 0x4020, 0xBC20, 0x1020, 0x13FE, 0xFC20, 0x1020, 0x1020, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {38031, 2, {0x1104, 0x1124, 0x3D24, 0x2124, 0x4124, 0xBD24, 0x1124, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1524, 0x1A24, 0x1204, 0x404}},
    {38032, 2, {0x1008, 0x1008, 0x1E10, 0x2020, 0x2040, 0x7C88, 0x9008, 0x1010, 0xFE20, 0x1044, 0x1084, 0x1208, 0x1410, 0x1820, 0x10C0, 0x300}},
    {38033, 2, {0x1000, 0x13FC, 0x3C84, 0x2088, 0x4088, 0xBC90, 0x109C, 0x1084, 0xFD44, 0x1144, 0x1128, 0x1128, 0x1610, 0x1A28, 0x1444, 0x182}},
    {38034, 2, {0x1000, 0x11F0, 0x3D10, 0x2110, 0x4110, 0xBD10, 0x1190, 0x1150, 0xFD50, 0x1110, 0x1110, 0x1112, 0x1512, 0x1A12, 0x120E, 0x400}},
    {38035, 2, {0x1040, 0x1040, 0x3C80, 0x20FC, 0x4104, 0xBE04, 0x1004, 0x1104, 0xFC84, 0x1044, 0x1044, 0x1004, 0x1404, 0x1804, 0x1028, 0x10}},
    {38036, 2, {0x1080, 0x1040, 0x3C5E, 0x2102, 0x4102, 0xBD02, 0x1102, 0x1102, 0xFD02, 0x1102, 0x1102, 0x1102, 0x1502, 0x1902, 0x110A, 0x104}},
    {38037, 2, {0x1040, 0x1040, 0x3C40, 0x2040, 0x43FE, 0xBC88, 0x1088, 0x1088, 0xFC88, 0x1108, 0x1090, 0x1050, 0x1420, 0x1850, 0x1088, 0x304}},
    {38038, 2, {0x1000, 0x11F8, 0x3C10, 0x2020, 0x4040, 0xBC80, 0x11FC, 0x1054, 0xFC54, 0x1094, 0x1124, 0x1024, 0x1444, 0x1884, 0x1128, 0x10}},
    {38039, 2, {0x1000, 0x13FC, 0x3D04, 0x2144, 0x4124, 0xBCA8, 0x1088, 0x1088, 0xFC50, 0x1050, 0x1020, 0x1020, 0x1450, 0x1888, 0x1104, 0x602}},
    {38040, 2, {0x1000, 0x11FC, 0x3C88, 0x2088, 0x4088, 0xBC88, 0x1088, 0x13FE, 0xFC88, 0x1088, 0x1088, 0x1088, 0x1488, 0x1908, 0x1108, 0x208}},
    {38041, 2, {0x1000, 0x13FE, 0x3C20, 0x2020, 0x4120, 0xBD3E, 0x1120, 0x1120, 0xFD20, 0x11FE, 0x1002, 0x1002, 0x1402, 0x1802, 0x1014, 0x8}},
    {38042, 2, {0x1000, 0x1000, 0x3DFE, 0x2010, 0x4010, 0xBC20, 0x1020, 0x1068, 0xFCA4, 0x1122, 0x1222, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {38043, 2, {0x1020, 0x1020, 0x1E20, 0x2020, 0x2020, 0x7DFE, 0x9020, 0x1020, 0xFE20, 0x1050, 0x1050, 0x1250, 0x1488, 0x18C8, 0x1124, 0x202}},
    {38044, 2, {0x1000, 0x11FE, 0x3D00, 0x2100, 0x4100, 0xBDFC, 0x1104, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1100, 0x1500, 0x1900, 0x11FE, 0x0}},
    {38045, 2, {0x2040, 0x2040, 0x3840, 0x27FC, 0x4040, 0x7A48, 0xA248, 0x2248, 0xFA48, 0x23F8, 0x2048, 0x2040, 0x2842, 0x3042, 0x203E, 0x0}},
    {38046, 2, {0x1020, 0x1020, 0x3C20, 0x20A8, 0x40A4, 0xBCA2, 0x1122, 0x1120, 0xFE24, 0x1024, 0x1028, 0x1008, 0x1410, 0x1820, 0x10C0, 0x300}},
    {38047, 2, {0x1020, 0x1020, 0x3C20, 0x2020, 0x41FC, 0xBD24, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {38048, 2, {0x1020, 0x1020, 0x3C20, 0x21FE, 0x4122, 0xBD22, 0x1122, 0x1122, 0xFD52, 0x114A, 0x118A, 0x1102, 0x1502, 0x1902, 0x110A, 0x104}},
    {38049, 2, {0x1000, 0x11FC, 0x3D04, 0x2104, 0x4124, 0xBD24, 0x1124, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1050, 0x1448, 0x1884, 0x1102, 0x602}},
    {38050, 2, {0x1000, 0x11FC, 0x3D04, 0x2104, 0x4194, 0xBD54, 0x1154, 0x1124, 0xFD24, 0x1154, 0x1154, 0x1194, 0x1504, 0x1904, 0x1114, 0x108}},
    {38051, 2, {0x1008, 0x101C, 0x3DE0, 0x2100, 0x4100, 0xBDFC, 0x1144, 0x1144, 0xFD44, 0x1128, 0x1128, 0x1110, 0x1510, 0x1A28, 0x1244, 0x482}},
    {38052, 2, {0x1020, 0x1020, 0x3C50, 0x2050, 0x4088, 0xBD44, 0x1222, 0x1020, 0xFC00, 0x11FC, 0x1004, 0x1008, 0x1408, 0x1810, 0x1010, 0x20}},
    {38053, 2, {0x1000, 0x10FC, 0x3C84, 0x2084, 0x4084, 0xBCFC, 0x1084, 0x1084, 0xFC84, 0x10FC, 0x1084, 0x1084, 0x1504, 0x1904, 0x1214, 0x408}},
    {38054, 2, {0x1040, 0x1040, 0x3C40, 0x207C, 0x4084, 0xBC88, 0x1120, 0x1020, 0xFC20, 0x1050, 0x1050, 0x1050, 0x1488, 0x1888, 0x1104, 0x202}},
    {38055, 2, {0x1040, 0x1040, 0x3C80, 0x20FC, 0x4104, 0xBE04, 0x1084, 0x1044, 0xFC44, 0x1014, 0x1024, 0x1044, 0x1584, 0x1804, 0x1028, 0x10}},
    {38056, 2, {0x1020, 0x1040, 0x3CFC, 0x2084, 0x4084, 0xBC84, 0x1094, 0x1088, 0xFC80, 0x10FE, 0x1002, 0x1002, 0x15FA, 0x1802, 0x1014, 0x8}},
    {38057, 2, {0x1080, 0x1080, 0x3C80, 0x20FC, 0x4104, 0xBD04, 0x1244, 0x1044, 0xFC84, 0x10A4, 0x1114, 0x11F4, 0x1414, 0x1804, 0x1028, 0x10}},
    {38058, 2, {0x1040, 0x1020, 0x3C00, 0x21FC, 0x4000, 0xBC00, 0x10F0, 0x1090, 0xFC90, 0x1090, 0x1090, 0x1092, 0x1492, 0x1912, 0x110E, 0x200}},
    {38059, 2, {0x1040, 0x1020, 0x3C20, 0x23FE, 0x4080, 0xBC80, 0x1080, 0x10FC, 0xFC84, 0x1084, 0x1084, 0x1084, 0x1504, 0x1904, 0x1228, 0x410}},
    {38060, 2, {0x1020, 0x1020, 0x1E20, 0x2024, 0x20A4, 0x7CA4, 0x90A8, 0x1120, 0xFE20, 0x1050, 0x1050, 0x1250, 0x1488, 0x1888, 0x1104, 0x202}},
    {38061, 2, {0x1010, 0x1110, 0x3C90, 0x2090, 0x4010, 0xBD10, 0x1090, 0x1090, 0xFC10, 0x101E, 0x13F0, 0x1010, 0x1410, 0x1810, 0x1010, 0x10}},
    {38062, 2, {0x1000, 0x11F8, 0x3C48, 0x2048, 0x4048, 0xBC48, 0x1048, 0x11F8, 0xFC88, 0x1088, 0x1088, 0x1088, 0x1488, 0x1888, 0x13FE, 0x0}},
    {38063, 2, {0x1000, 0x11FC, 0x3D24, 0x2124, 0x4124, 0xBD24, 0x1124, 0x11FC, 0xFD04, 0x1100, 0x1100, 0x1100, 0x1502, 0x1902, 0x10FE, 0x0}},
    {38064, 2, {0x1000, 0x1000, 0x3DFC, 0x2020, 0x4020, 0xBC20, 0x1020, 0x1020, 0xFDFC, 0x1020, 0x1028, 0x1024, 0x1424, 0x1820, 0x13FE, 0x0}},
    {38065, 2, {0x1050, 0x1048, 0x3C40, 0x205C, 0x41E0, 0xBC40, 0x105E, 0x11E0, 0xFC44, 0x1048, 0x1030, 0x1022, 0x1452, 0x188A, 0x1306, 0x2}},
    {38066, 2, {0x1000, 0x1000, 0x3DFE, 0x2010, 0x4010, 0xBC10, 0x1090, 0x1090, 0xFC9E, 0x1090, 0x1090, 0x1090, 0x1490, 0x1890, 0x13FE, 0x0}},
    {38067, 2, {0x1088, 0x1088, 0x3C88, 0x2088, 0x43FE, 0xBC88, 0x1088, 0x1088, 0xFC88, 0x10F8, 0x1088, 0x1088, 0x1488, 0x1888, 0x10F8, 0x88}},
    {38068, 2, {0x1020, 0x1020, 0x3C20, 0x2020, 0x43FE, 0xBC20, 0x1020, 0x1020, 0xFDFC, 0x1104, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38069, 2, {0x1020, 0x1020, 0x3C20, 0x2020, 0x41FE, 0xBC70, 0x10A8, 0x10A8, 0xFD24, 0x1124, 0x1222, 0x10F8, 0x1420, 0x1820, 0x1020, 0x20}},
    {38070, 2, {0x1000, 0x1000, 0x3DFE, 0x2008, 0x4008, 0xBDE8, 0x1128, 0x1128, 0xFD28, 0x1128, 0x11E8, 0x1128, 0x1408, 0x1808, 0x1028, 0x10}},
    {38071, 2, {0x1000, 0x11FE, 0x3D00, 0x2100, 0x417C, 0xBD44, 0x1144, 0x1144, 0xFD44, 0x1144, 0x117C, 0x1144, 0x1500, 0x1900, 0x11FE, 0x0}},
    {38072, 2, {0x2040, 0x2040, 0x3840, 0x23FE, 0x4080, 0x78A0, 0xA120, 0x21FC, 0xFB24, 0x2524, 0x2124, 0x2124, 0x2934, 0x3128, 0x2020, 0x20}},
    {38073, 2, {0x1050, 0x1048, 0x3C48, 0x2040, 0x43FE, 0xBC80, 0x1080, 0x10FC, 0xFD44, 0x1144, 0x1128, 0x1128, 0x1610, 0x1A28, 0x1444, 0x182}},
    {38074, 2, {0x2028, 0x2024, 0x3824, 0x2020, 0x43FE, 0x7A20, 0xA224, 0x2224, 0xFA24, 0x2228, 0x2228, 0x2290, 0x2B12, 0x322A, 0x2046, 0x82}},
    {38075, 2, {0x1020, 0x1020, 0x3C20, 0x2020, 0x403E, 0xBC20, 0x1020, 0x1020, 0xFDFC, 0x1104, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38076, 2, {0x1000, 0x11FC, 0x3D04, 0x2104, 0x4104, 0xBDFC, 0x1104, 0x1104, 0xFD04, 0x11FC, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38077, 2, {0x1000, 0x1000, 0x3DFC, 0x2104, 0x4104, 0xBD04, 0x11FC, 0x1104, 0xFD04, 0x1104, 0x11FC, 0x1104, 0x1400, 0x1800, 0x13FE, 0x0}},
    {38078, 2, {0x1000, 0x11FC, 0x3D24, 0x2124, 0x4124, 0xBDFC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {38079, 2, {0x1000, 0x1000, 0x3DFC, 0x2124, 0x4124, 0xBD24, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x1124, 0x1124, 0x1524, 0x1924, 0x11FC, 0x104}},
    {38080, 2, {0x1020, 0x1020, 0x3C20, 0x2020, 0x41FC, 0xBD24, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x1124, 0x1124, 0x1524, 0x1924, 0x11FC, 0x104}},
    {38081, 2, {0x1020, 0x1020, 0x3D20, 0x2120, 0x41FC, 0xBD20, 0x1220, 0x1020, 0xFDFE, 0x1020, 0x1050, 0x1050, 0x1488, 0x1888, 0x1104, 0x202}},
    {38082, 2, {0x1020, 0x1020, 0x3C40, 0x21FC, 0x4104, 0xBD04, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38083, 2, {0x2040, 0x2040, 0x38A0, 0x20A0, 0x4110, 0x7A48, 0xA426, 0x2020, 0xFBF8, 0x2008, 0x2010, 0x2110, 0x28A0, 0x3040, 0x2020, 0x20}},
    {38084, 2, {0x2008, 0x203C, 0x3BC0, 0x2200, 0x4220, 0x7A20, 0xA220, 0x23FE, 0xF820, 0x2020, 0x2128, 0x2124, 0x2A22, 0x3422, 0x20A0, 0x40}},
    {38085, 2, {0x1000, 0x10F8, 0x3C88, 0x2088, 0x4088, 0xBC88, 0x1106, 0x1200, 0xFC00, 0x11FC, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38086, 2, {0x2000, 0x2080, 0x3B1E, 0x2252, 0x4252, 0x7A52, 0xA252, 0x2252, 0xFA52, 0x22D2, 0x235A, 0x2254, 0x2890, 0x3090, 0x2110, 0x210}},
    {38087, 2, {0x1080, 0x1080, 0x3DFC, 0x2104, 0x4204, 0xBDF4, 0x1114, 0x1114, 0xFD14, 0x11F4, 0x1104, 0x1128, 0x1512, 0x1902, 0x10FE, 0x0}},
    {38088, 2, {0x1040, 0x1020, 0x3C00, 0x23FE, 0x4020, 0xBC20, 0x11FC, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1534, 0x1928, 0x1020, 0x20}},
    {38089, 2, {0x1020, 0x1010, 0x3C10, 0x21FE, 0x4020, 0xBC20, 0x1044, 0x1084, 0xFDF8, 0x1010, 0x1020, 0x1044, 0x1482, 0x19FE, 0x1082, 0x0}},
    {38090, 2, {0x1020, 0x1010, 0x3C10, 0x21FE, 0x4102, 0xBE04, 0x1080, 0x1088, 0xFC90, 0x10A0, 0x10C0, 0x1082, 0x1482, 0x1882, 0x107E, 0x0}},
    {38091, 2, {0x1040, 0x1020, 0x3C28, 0x2008, 0x4008, 0xBC48, 0x1054, 0x1152, 0xFD52, 0x1160, 0x1260, 0x1044, 0x14C4, 0x1944, 0x123C, 0x0}},
    {38092, 2, {0x1000, 0x11FC, 0x3D04, 0x2104, 0x4104, 0xBDFC, 0x1100, 0x1140, 0xFD44, 0x1148, 0x1170, 0x1140, 0x1542, 0x1A42, 0x123E, 0x400}},
    {38093, 2, {0x1010, 0x1010, 0x3C10, 0x21FE, 0x4112, 0xBD14, 0x1110, 0x11FC, 0xFD44, 0x1144, 0x1128, 0x1128, 0x1510, 0x1A28, 0x1244, 0x482}},
    {38094, 2, {0x1000, 0x13FC, 0x3D04, 0x2088, 0x4050, 0xBC20, 0x10D8, 0x1326, 0xFC20, 0x11FC, 0x1020, 0x1020, 0x17FE, 0x1820, 0x1020, 0x20}},
    {38095, 2, {0x2004, 0x27C4, 0x3A84, 0x2294, 0x4294, 0x7A94, 0xA294, 0x27D4, 0xFA94, 0x2294, 0x2294, 0x2294, 0x2A84, 0x3484, 0x2494, 0x888}},
    {38096, 2, {0x1040, 0x1044, 0x3DF4, 0x2048, 0x4050, 0xBDFE, 0x1040, 0x1080, 0xFDFE, 0x1240, 0x1480, 0x10FC, 0x1404, 0x1804, 0x1028, 0x10}},
    {38097, 2, {0x1040, 0x1040, 0x3C44, 0x21F4, 0x4048, 0xBC50, 0x13FE, 0x1040, 0xFC80, 0x1184, 0x1298, 0x10E0, 0x1482, 0x1882, 0x107E, 0x0}},
    {38098, 2, {0x1000, 0x11FE, 0x3C88, 0x2088, 0x4088, 0xBCF8, 0x1088, 0x1088, 0xFCF8, 0x1088, 0x1088, 0x109E, 0x15E8, 0x1808, 0x1008, 0x8}},
    {38099, 2, {0x1048, 0x1048, 0x3DFE, 0x2048, 0x4048, 0xBC20, 0x1010, 0x11FE, 0xFC80, 0x1080, 0x1080, 0x1080, 0x1480, 0x1880, 0x10FC, 0x0}},
    {38100, 2, {0x2000, 0x23FC, 0x3890, 0x2090, 0x4090, 0x7892, 0xA492, 0x2294, 0xFA94, 0x2298, 0x2090, 0x2090, 0x2890, 0x3090, 0x27FE, 0x0}},
    {38101, 2, {0x1020, 0x1020, 0x3DFE, 0x2040, 0x4040, 0xBCFC, 0x1084, 0x1184, 0xFEFC, 0x1084, 0x1084, 0x10FC, 0x1484, 0x1884, 0x1094, 0x88}},
    {38102, 2, {0x2028, 0x2024, 0x3824, 0x2020, 0x43FE, 0x7A20, 0xA224, 0x2224, 0xFBA4, 0x22A8, 0x22A8, 0x2290, 0x2A92, 0x35AA, 0x2446, 0x882}},
    {38103, 2, {0x1020, 0x1020, 0x3C20, 0x21FC, 0x4020, 0xBD24, 0x10A4, 0x10A8, 0xFC20, 0x13FE, 0x1050, 0x1050, 0x1488, 0x1888, 0x1104, 0x202}},
    {38104, 2, {0x2000, 0x27DE, 0x3892, 0x4492, 0x4494, 0x7C94, 0xA7D8, 0x2094, 0xF992, 0x2192, 0x2292, 0x2A9A, 0x3494, 0x2890, 0x290, 0x110}},
    {38105, 2, {0x2080, 0x2080, 0x38BC, 0x23C0, 0x4050, 0x7824, 0xA0D4, 0x230C, 0xF800, 0x23FE, 0x2090, 0x2090, 0x2912, 0x3112, 0x220E, 0x400}},
    {38106, 2, {0x1000, 0x11FE, 0x3C20, 0x2040, 0x4088, 0xBD04, 0x11FE, 0x1022, 0xFC20, 0x1020, 0x11FE, 0x1020, 0x1420, 0x1820, 0x13FE, 0x0}},
    {38107, 2, {0x1020, 0x1124, 0x3CA4, 0x20A4, 0x40A8, 0xBC20, 0x13FC, 0x1004, 0xFC04, 0x1004, 0x11FC, 0x1004, 0x1404, 0x1804, 0x13FC, 0x4}},
    {38108, 2, {0x2000, 0x23FC, 0x3A04, 0x2204, 0x42F4, 0x7A04, 0xA204, 0x22F4, 0xFA94, 0x2294, 0x2294, 0x22F4, 0x2A04, 0x3204, 0x2214, 0x208}},
    {38109, 2, {0x1000, 0x10FC, 0x3C84, 0x2084, 0x4084, 0xBC84, 0x10FC, 0x1000, 0xFC00, 0x11FE, 0x1102, 0x1102, 0x1502, 0x1902, 0x11FE, 0x102}},
    {38110, 2, {0x2000, 0x21FC, 0x3904, 0x2104, 0x41FC, 0x7820, 0xA020, 0x23FE, 0xFA22, 0x2222, 0x2222, 0x2222, 0x2A2A, 0x3224, 0x2020, 0x20}},
    {38111, 2, {0x2000, 0x23FE, 0x3A02, 0x2222, 0x4222, 0x7A22, 0xA3FE, 0x2222, 0xFA22, 0x2252, 0x224A, 0x228A, 0x2B02, 0x3202, 0x23FE, 0x202}},
    {38112, 2, {0x1020, 0x1124, 0x3D24, 0x2124, 0x41FC, 0xBC00, 0x11FC, 0x1004, 0xFC04, 0x11FC, 0x1100, 0x1100, 0x1502, 0x1902, 0x10FE, 0x0}},
    {38113, 2, {0x2002, 0x23E2, 0x3A22, 0x222A, 0x42AA, 0x7AAA, 0xA2AA, 0x22AA, 0xFAAA, 0x22AA, 0x22AA, 0x2082, 0x2942, 0x3122, 0x220A, 0x404}},
    {38114, 2, {0x2020, 0x2120, 0x3920, 0x21FC, 0x4120, 0x7A20, 0xA020, 0x23FE, 0xF870, 0x20A8, 0x20A8, 0x2124, 0x2924, 0x3222, 0x2020, 0x20}},
    {38115, 2, {0x1020, 0x1120, 0x3D20, 0x21FC, 0x4120, 0xBE20, 0x1020, 0x13FE, 0xFC90, 0x1090, 0x1090, 0x1090, 0x1512, 0x1912, 0x1212, 0x40E}},
    {38116, 2, {0x2000, 0x2006, 0x3BB8, 0x2088, 0x4088, 0x7908, 0xA13E, 0x2388, 0xF888, 0x2288, 0x2288, 0x213E, 0x2900, 0x3280, 0x247E, 0x800}},
    {38117, 2, {0x1008, 0x103C, 0x3DE0, 0x2020, 0x4020, 0xBDFC, 0x1020, 0x1020, 0xFC20, 0x13FE, 0x1020, 0x1040, 0x1488, 0x1904, 0x13FE, 0x102}},
    {38118, 2, {0x1008, 0x103C, 0x3DE0, 0x2020, 0x4020, 0xBDFE, 0x1020, 0x1020, 0xFC20, 0x11FC, 0x1104, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38119, 2, {0x1090, 0x1094, 0x3D14, 0x2118, 0x4310, 0xBD32, 0x1152, 0x110E, 0xFD20, 0x1020, 0x13FE, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {38120, 2, {0x2040, 0x2040, 0x38A0, 0x2110, 0x4208, 0x7C06, 0xA3F8, 0x2040, 0xF840, 0x2040, 0x23F8, 0x2040, 0x2840, 0x3040, 0x27FE, 0x0}},
    {38121, 2, {0x1004, 0x1188, 0x3C50, 0x2020, 0x4050, 0xBC88, 0x1124, 0x1020, 0xFFFE, 0x1020, 0x10A8, 0x10A4, 0x1522, 0x1A22, 0x10A0, 0x40}},
    {38122, 2, {0x2040, 0x2040, 0x38A0, 0x2110, 0x4208, 0x7C06, 0xA3F8, 0x2000, 0xF800, 0x23F8, 0x2208, 0x2208, 0x2A08, 0x3208, 0x23F8, 0x208}},
    {38123, 2, {0x1050, 0x1050, 0x3C50, 0x2152, 0x40D4, 0xBC58, 0x1050, 0x1058, 0xFCD4, 0x1152, 0x1250, 0x1050, 0x1492, 0x1892, 0x1112, 0x20E}},
    {38124, 2, {0x1040, 0x1040, 0x3C78, 0x2088, 0x4150, 0xBC20, 0x1050, 0x1088, 0xFD06, 0x10F8, 0x1088, 0x1088, 0x1488, 0x1888, 0x10F8, 0x88}},
    {38125, 2, {0x1040, 0x1040, 0x3CFC, 0x2104, 0x4288, 0xBC50, 0x1020, 0x1040, 0xFC80, 0x11FC, 0x1284, 0x1084, 0x1484, 0x1884, 0x10FC, 0x84}},
    {38126, 2, {0x2080, 0x2080, 0x39F8, 0x2208, 0x4010, 0x7BFC, 0xA024, 0x2024, 0xFBFE, 0x2024, 0x2024, 0x23FC, 0x2824, 0x3020, 0x20A0, 0x40}},
    {38127, 2, {0x1080, 0x1080, 0x3CF8, 0x2108, 0x4210, 0xBDFC, 0x1124, 0x1124, 0xFD24, 0x11FC, 0x1100, 0x1100, 0x1502, 0x1902, 0x10FE, 0x0}},
    {38128, 2, {0x1040, 0x1020, 0x3C20, 0x21FE, 0x4000, 0xBC88, 0x1104, 0x1202, 0xFC88, 0x1088, 0x1050, 0x1050, 0x1420, 0x1850, 0x1088, 0x306}},
    {38129, 2, {0x1040, 0x1020, 0x3C20, 0x23FE, 0x4040, 0xBC40, 0x10A0, 0x10A2, 0xFDA4, 0x1298, 0x1490, 0x1088, 0x1484, 0x18A2, 0x10C0, 0x80}},
    {38130, 2, {0x2040, 0x2020, 0x3BFC, 0x2000, 0x4108, 0x7890, 0xA3FE, 0x2200, 0xFA00, 0x2200, 0x2200, 0x2200, 0x2A00, 0x3400, 0x2400, 0x800}},
    {38131, 2, {0x1040, 0x1020, 0x3C20, 0x23FE, 0x4040, 0xBC88, 0x1104, 0x13FE, 0xFC92, 0x1090, 0x1090, 0x1090, 0x1512, 0x1912, 0x120E, 0x400}},
    {38132, 2, {0x2200, 0x217C, 0x3904, 0x2408, 0x4210, 0x7A20, 0xA07E, 0x212A, 0xF92A, 0x222A, 0x264A, 0x2252, 0x2A92, 0x3322, 0x224A, 0x84}},
    {38133, 2, {0x2040, 0x2020, 0x3820, 0x23FE, 0x4202, 0x7C44, 0xA040, 0x23FE, 0xF888, 0x2088, 0x2108, 0x20D0, 0x2820, 0x3050, 0x2088, 0x304}},
    {38134, 2, {0x1000, 0x11F8, 0x3D08, 0x2108, 0x41F8, 0xBD08, 0x1108, 0x11F8, 0xFD44, 0x1148, 0x1130, 0x1120, 0x1510, 0x1948, 0x1186, 0x100}},
    {38135, 2, {0x2100, 0x2100, 0x3900, 0x211E, 0x47D2, 0x7A52, 0xA252, 0x2252, 0xFA52, 0x2492, 0x2292, 0x2112, 0x2A92, 0x325E, 0x2452, 0x800}},
    {38136, 2, {0x1020, 0x1020, 0x3DFC, 0x2020, 0x40FC, 0xBC20, 0x11FE, 0x1048, 0xFC48, 0x10FE, 0x1088, 0x1148, 0x1528, 0x1A08, 0x1028, 0x10}},
    {38137, 2, {0x2088, 0x2088, 0x3BFE, 0x2088, 0x4000, 0x7BFE, 0xA202, 0x2444, 0xF840, 0x23FC, 0x2044, 0x2084, 0x2884, 0x3104, 0x2228, 0x410}},
    {38138, 2, {0x2048, 0x2044, 0x3BFE, 0x2040, 0x4040, 0x7BFC, 0xA244, 0x2244, 0xFBFC, 0x2244, 0x2244, 0x23FC, 0x2A44, 0x3244, 0x2254, 0x208}},
    {38139, 2, {0x1000, 0x13FE, 0x3C40, 0x2040, 0x41FC, 0xBC84, 0x1084, 0x13FE, 0xFC00, 0x1000, 0x11FC, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38140, 2, {0x2020, 0x2020, 0x3820, 0x23FE, 0x4020, 0x7924, 0xA0A4, 0x20A8, 0xFBFE, 0x2070, 0x20A8, 0x20A8, 0x2924, 0x3222, 0x2420, 0x20}},
    {38141, 2, {0x2010, 0x2014, 0x3812, 0x2010, 0x43FE, 0x7810, 0xA090, 0x2050, 0xF848, 0x2008, 0x2088, 0x22A8, 0x2A9A, 0x32AA, 0x2466, 0x2}},
    {38142, 2, {0x2010, 0x2210, 0x3910, 0x217E, 0x4020, 0x7828, 0xA748, 0x217E, 0xF908, 0x2108, 0x21FE, 0x2108, 0x2908, 0x3288, 0x247E, 0x0}},
    {38143, 2, {0x2080, 0x22FC, 0x3AA4, 0x22A4, 0x42A8, 0x7A90, 0xA2A8, 0x20C6, 0xF820, 0x2020, 0x21FC, 0x2020, 0x2820, 0x3020, 0x23FE, 0x0}},
    {38144, 2, {0x1020, 0x1124, 0x3CA4, 0x20A8, 0x4020, 0xBDFC, 0x1104, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1504, 0x1904, 0x1114, 0x108}},
    {38145, 2, {0x1020, 0x1124, 0x3CA4, 0x20A8, 0x4020, 0xBDFC, 0x1104, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1450, 0x1888, 0x1104, 0x202}},
    {38146, 2, {0x1000, 0x11FC, 0x3D24, 0x2124, 0x41FC, 0xBD24, 0x1124, 0x11FC, 0xFC20, 0x1020, 0x11FC, 0x1020, 0x1420, 0x1820, 0x13FE, 0x0}},
    {38147, 2, {0x1000, 0x11FC, 0x3D04, 0x2104, 0x4104, 0xBDFC, 0x1000, 0x1000, 0xFDFE, 0x1020, 0x1020, 0x11FC, 0x1420, 0x1820, 0x13FE, 0x0}},
    {38148, 2, {0x2008, 0x2008, 0x3BC8, 0x2248, 0x425E, 0x7A4A, 0xA3CA, 0x224A, 0xFA4A, 0x23CA, 0x224A, 0x224A, 0x2A6A, 0x33D2, 0x2612, 0x26}},
    {38149, 2, {0x2000, 0x21FC, 0x3904, 0x2104, 0x4104, 0x79FC, 0xA020, 0x2020, 0xFBFE, 0x2222, 0x2252, 0x228A, 0x2B0A, 0x3202, 0x220A, 0x204}},
    {38150, 2, {0x2020, 0x2120, 0x3920, 0x21FC, 0x4120, 0x7A20, 0xA020, 0x23FE, 0xF800, 0x2000, 0x21FC, 0x2104, 0x2904, 0x3104, 0x21FC, 0x104}},
    {38151, 2, {0x2010, 0x20D8, 0x3B94, 0x2094, 0x4090, 0x7BFE, 0xA090, 0x2094, 0xF894, 0x20D8, 0x2398, 0x2090, 0x28AA, 0x30CA, 0x2286, 0x102}},
    {38152, 2, {0x101C, 0x11E0, 0x3C20, 0x2020, 0x43FE, 0xBCA8, 0x1124, 0x1222, 0xFDF8, 0x1088, 0x1090, 0x10BE, 0x1502, 0x1902, 0x1214, 0x408}},
    {38153, 2, {0x1020, 0x1020, 0x3D24, 0x2124, 0x4124, 0xBD24, 0x12AA, 0x1472, 0xFC20, 0x1020, 0x11FC, 0x1020, 0x1420, 0x1820, 0x13FE, 0x0}},
    {38154, 2, {0x1004, 0x101E, 0x3DE0, 0x2022, 0x4112, 0xBC94, 0x1080, 0x1008, 0xFDFE, 0x1008, 0x1088, 0x1048, 0x1448, 0x1808, 0x1028, 0x10}},
    {38155, 2, {0x1040, 0x1040, 0x3CFC, 0x2088, 0x4150, 0xBC20, 0x10D8, 0x1326, 0xFCF8, 0x1020, 0x10F8, 0x1020, 0x17FE, 0x1820, 0x1020, 0x20}},
    {38156, 2, {0x1040, 0x1020, 0x3DFC, 0x2000, 0x4088, 0xBC50, 0x13FE, 0x1020, 0xFC20, 0x1020, 0x11FC, 0x1020, 0x1420, 0x1820, 0x1020, 0x20}},
    {38157, 2, {0x2040, 0x2020, 0x3BFC, 0x2040, 0x4090, 0x7908, 0xA3FC, 0x2004, 0xF950, 0x2150, 0x2150, 0x2150, 0x2A50, 0x3252, 0x2452, 0x80E}},
    {38158, 2, {0x2100, 0x20BE, 0x3A82, 0x2202, 0x42FA, 0x7A52, 0xA252, 0x2252, 0xFBFE, 0x2252, 0x2252, 0x2292, 0x2B12, 0x3202, 0x220A, 0x204}},
    {38159, 2, {0x2100, 0x20BE, 0x3A82, 0x2202, 0x4202, 0x7AFA, 0xA28A, 0x228A, 0xFAFA, 0x228A, 0x228A, 0x22FA, 0x2A02, 0x3202, 0x220A, 0x204}},
    {38160, 2, {0x2208, 0x2108, 0x3910, 0x4020, 0x43F8, 0x7A08, 0xA208, 0x2208, 0xFBF8, 0x20A0, 0x20A0, 0x2920, 0x3124, 0x2224, 0x41C, 0x800}},
    {38161, 2, {0x2108, 0x2088, 0x3890, 0x23FC, 0x4024, 0x7824, 0xA3FC, 0x2220, 0xFA20, 0x23FE, 0x2062, 0x20A2, 0x292A, 0x3224, 0x2420, 0x20}},
    {38162, 2, {0x1040, 0x1020, 0x3DFC, 0x2104, 0x4104, 0xBDFC, 0x1104, 0x1104, 0xFDFC, 0x1120, 0x1122, 0x1114, 0x1508, 0x1944, 0x1182, 0x100}},
    {38163, 2, {0x1000, 0x11FC, 0x3C04, 0x20FC, 0x4004, 0xBDFC, 0x1000, 0x13FE, 0xFE02, 0x11F8, 0x1088, 0x1088, 0x1450, 0x1820, 0x10D8, 0x306}},
    {38164, 2, {0x2000, 0x23F8, 0x3A08, 0x4208, 0x43F8, 0x7A00, 0xA200, 0x23FC, 0xFA04, 0x22F4, 0x2294, 0x2A94, 0x34F4, 0x2404, 0x828, 0x10}},
    {38165, 2, {0x2000, 0x277E, 0x3D04, 0x2504, 0x4574, 0x7D54, 0xA654, 0x2554, 0xFD54, 0x2554, 0x2574, 0x2554, 0x2E04, 0x3404, 0x2414, 0x408}},
    {38166, 2, {0x2020, 0x2020, 0x3BFE, 0x2020, 0x41FC, 0x7820, 0xA3FE, 0x2000, 0xF9FC, 0x2104, 0x21FC, 0x2104, 0x29FC, 0x3104, 0x2114, 0x108}},
    {38167, 2, {0x2020, 0x2022, 0x39FA, 0x2024, 0x4028, 0x7BFE, 0xA020, 0x2040, 0xF8FC, 0x2184, 0x2284, 0x20FC, 0x2884, 0x3084, 0x20FC, 0x84}},
    {38168, 2, {0x2088, 0x2088, 0x3BFE, 0x2088, 0x40A8, 0x7820, 0xA3FE, 0x2040, 0xF880, 0x20FC, 0x2184, 0x2284, 0x2884, 0x3084, 0x20FC, 0x84}},
    {38169, 2, {0x2110, 0x2110, 0x3910, 0x27FC, 0x4110, 0x7910, 0xAFFE, 0x2000, 0xFBF8, 0x2208, 0x2208, 0x23F8, 0x2A08, 0x3208, 0x23F8, 0x208}},
    {38170, 2, {0x2088, 0x2088, 0x3BFE, 0x2088, 0x4088, 0x7800, 0xA1FC, 0x2124, 0xF924, 0x2124, 0x21FC, 0x2124, 0x2924, 0x3124, 0x21FC, 0x104}},
    {38171, 2, {0x2020, 0x2020, 0x3BFE, 0x2050, 0x40A8, 0x7924, 0xA2FA, 0x2020, 0xF8A8, 0x2088, 0x23FE, 0x2088, 0x2888, 0x3088, 0x2108, 0x208}},
    {38172, 2, {0x1020, 0x1020, 0x3DFC, 0x2050, 0x4088, 0xBD04, 0x13FE, 0x1008, 0xFDE8, 0x1128, 0x1128, 0x11E8, 0x1528, 0x1808, 0x1028, 0x10}},
    {38173, 2, {0x1000, 0x11FC, 0x3D04, 0x21FC, 0x4104, 0xBDFC, 0x1000, 0x11FE, 0xFC08, 0x13FE, 0x1008, 0x1108, 0x1488, 0x1808, 0x1028, 0x10}},
    {38174, 2, {0x1000, 0x11FC, 0x3D24, 0x2124, 0x41FC, 0xBD24, 0x1124, 0x11FC, 0xFC20, 0x13FE, 0x1070, 0x10A8, 0x1524, 0x1A22, 0x1020, 0x20}},
    {38175, 2, {0x1000, 0x11FC, 0x3D04, 0x2104, 0x41FC, 0xBD04, 0x1104, 0x11FC, 0xFC00, 0x1112, 0x11D4, 0x1118, 0x1510, 0x1952, 0x1192, 0x10E}},
    {38176, 2, {0x10FC, 0x1084, 0x3C84, 0x20FC, 0x4084, 0xBC84, 0x10FC, 0x1000, 0xFDFE, 0x1102, 0x1102, 0x11FE, 0x1502, 0x1902, 0x11FE, 0x102}},
    {38177, 2, {0x1000, 0x11F8, 0x3D08, 0x2108, 0x41F8, 0xBD08, 0x1108, 0x11F8, 0xFC80, 0x11FC, 0x1254, 0x1094, 0x1524, 0x1A44, 0x1094, 0x108}},
    {38178, 2, {0x1000, 0x11FC, 0x3D04, 0x2124, 0x4124, 0xBDFC, 0x1124, 0x1124, 0xFD74, 0x1154, 0x1154, 0x1174, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38179, 2, {0x2000, 0x23FE, 0x3A52, 0x2252, 0x4252, 0x7BFE, 0xA020, 0x2040, 0xF8FC, 0x2104, 0x2288, 0x2050, 0x2820, 0x3040, 0x2180, 0x600}},
    {38180, 2, {0x2008, 0x203C, 0x39E0, 0x2020, 0x4020, 0x7BFE, 0xA124, 0x2124, 0xFBFE, 0x2124, 0x2124, 0x23FE, 0x2820, 0x3020, 0x23FE, 0x0}},
    {38181, 2, {0x1050, 0x1048, 0x3C80, 0x20FE, 0x4190, 0xBE90, 0x10FC, 0x1090, 0xFC90, 0x10FC, 0x1090, 0x1090, 0x1490, 0x18FE, 0x1080, 0x80}},
    {38182, 2, {0x1020, 0x1040, 0x3DFC, 0x2104, 0x41FC, 0xBD04, 0x11FC, 0x1020, 0xFC20, 0x13FE, 0x1222, 0x1222, 0x162A, 0x1A24, 0x1020, 0x20}},
    {38183, 2, {0x101C, 0x11E0, 0x3D10, 0x2110, 0x41FE, 0xBD10, 0x1110, 0x117C, 0xFD44, 0x1154, 0x1154, 0x1154, 0x1554, 0x1A28, 0x1244, 0x482}},
    {38184, 2, {0x2010, 0x2050, 0x3B90, 0x221E, 0x4212, 0x7A24, 0xA3C8, 0x2288, 0xFA88, 0x2288, 0x2288, 0x2294, 0x2A94, 0x3494, 0x2494, 0x822}},
    {38185, 2, {0x1020, 0x1124, 0x3CA8, 0x2020, 0x41FC, 0xBC40, 0x13FE, 0x1088, 0xFD04, 0x12FA, 0x1488, 0x1088, 0x14A8, 0x1892, 0x1082, 0x7E}},
    {38186, 2, {0x2080, 0x2080, 0x38FC, 0x2154, 0x4254, 0x78A4, 0xA124, 0x2244, 0xF894, 0x2108, 0x2040, 0x20A4, 0x2AAA, 0x328A, 0x2478, 0x0}},
    {38187, 2, {0x1040, 0x1020, 0x3DFC, 0x2000, 0x4108, 0xBC90, 0x1000, 0x13FE, 0xFC00, 0x1000, 0x11FC, 0x1104, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38188, 2, {0x2040, 0x2244, 0x3A44, 0x2448, 0x40A0, 0x7910, 0xA608, 0x2044, 0xF840, 0x2248, 0x2248, 0x2450, 0x28A0, 0x3110, 0x2208, 0xC06}},
    {38189, 2, {0x2040, 0x2020, 0x3BFE, 0x2202, 0x4404, 0x7800, 0xA3FE, 0x2020, 0xF820, 0x2120, 0x213C, 0x2120, 0x2AA0, 0x3260, 0x243E, 0x800}},
    {38190, 2, {0x2010, 0x2010, 0x3B7C, 0x2114, 0x41FE, 0x7A14, 0xA27C, 0x2710, 0xF97C, 0x2510, 0x25FE, 0x2210, 0x2A10, 0x3500, 0x28FE, 0x0}},
    {38191, 2, {0x2000, 0x23FC, 0x3A04, 0x2204, 0x43FC, 0x7A20, 0xA220, 0x23FE, 0xFA20, 0x2220, 0x22FC, 0x2284, 0x2A84, 0x3484, 0x24FC, 0x884}},
    {38192, 2, {0x2000, 0x21FC, 0x3808, 0x2010, 0x4020, 0x7BFE, 0xA020, 0x20A0, 0xF840, 0x21FC, 0x2154, 0x2154, 0x2954, 0x3154, 0x27FE, 0x0}},
    {38193, 2, {0x1092, 0x1092, 0x3D24, 0x2248, 0x4124, 0xBC92, 0x1092, 0x1000, 0xFDFE, 0x1122, 0x1122, 0x11FE, 0x1522, 0x1922, 0x11FE, 0x102}},
    {38194, 2, {0x2100, 0x213C, 0x3BD4, 0x2114, 0x4394, 0x7914, 0xA7D4, 0x2124, 0xF94C, 0x2040, 0x27FE, 0x2040, 0x28A0, 0x3110, 0x2208, 0xC06}},
    {38195, 2, {0x2088, 0x2088, 0x3BFE, 0x2088, 0x40A8, 0x7820, 0xA1FC, 0x2124, 0xF924, 0x2124, 0x27FE, 0x2020, 0x2850, 0x3088, 0x2104, 0x202}},
    {38196, 2, {0x1110, 0x1112, 0x3DD4, 0x2118, 0x4152, 0xBD92, 0x112E, 0x1040, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1504, 0x1904, 0x11FC, 0x104}},
    {38197, 2, {0x1090, 0x1090, 0x3C9E, 0x22A2, 0x41D4, 0xBC88, 0x1094, 0x10A4, 0xFDFE, 0x1284, 0x14A4, 0x1094, 0x1494, 0x1884, 0x1094, 0x88}},
    {38198, 2, {0x2000, 0x21FC, 0x3924, 0x2124, 0x41FC, 0x7924, 0xA124, 0x21FC, 0xF800, 0x2040, 0x2024, 0x22A2, 0x2A8A, 0x3488, 0x2078, 0x0}},
    {38199, 2, {0x2000, 0x23DE, 0x3A52, 0x2252, 0x43DE, 0x7800, 0xA1FC, 0x2000, 0xFBFE, 0x2080, 0x2100, 0x21FC, 0x2804, 0x3004, 0x2028, 0x10}},
    {38200, 2, {0x1008, 0x103C, 0x3DE0, 0x2020, 0x4020, 0xBDFE, 0x1020, 0x10A0, 0xFD2C, 0x1124, 0x1124, 0x11AC, 0x1524, 0x1924, 0x11FC, 0x104}},
    {38201, 2, {0x2008, 0x20C8, 0x3B08, 0x2108, 0x412A, 0x792A, 0xA3AC, 0x2148, 0xF908, 0x2388, 0x2554, 0x2114, 0x2914, 0x3124, 0x2124, 0x142}},
    {38202, 2, {0x1008, 0x103C, 0x3DE0, 0x2020, 0x43FE, 0xBC20, 0x11FC, 0x1124, 0xFDFC, 0x1124, 0x11FC, 0x1020, 0x15FC, 0x1820, 0x13FE, 0x0}},
    {38203, 2, {0x2080, 0x2338, 0x3A28, 0x2228, 0x43A8, 0x7A46, 0xA200, 0x227C, 0xFBA4, 0x2224, 0x2228, 0x23A8, 0x2E10, 0x3228, 0x2244, 0x282}},
    {38204, 2, {0x10A0, 0x112C, 0x3D24, 0x2124, 0x41AC, 0xBD24, 0x1124, 0x11FC, 0xFC20, 0x11FC, 0x1088, 0x1050, 0x1420, 0x1850, 0x1088, 0x306}},
    {38205, 2, {0x1020, 0x1040, 0x3CFC, 0x2084, 0x40FC, 0xBC84, 0x10FC, 0x1000, 0xFDFE, 0x1020, 0x1020, 0x10FC, 0x1420, 0x1820, 0x11FE, 0x0}},
    {38206, 2, {0x101E, 0x13E0, 0x3C44, 0x2124, 0x4088, 0xBDFC, 0x1040, 0x1040, 0xFDFE, 0x1080, 0x10FC, 0x1144, 0x1528, 0x1A10, 0x1068, 0x186}},
    {38207, 2, {0x2040, 0x2020, 0x3BFE, 0x2000, 0x41FC, 0x7904, 0xA104, 0x2104, 0xF9FC, 0x2052, 0x2094, 0x2188, 0x2A88, 0x34A4, 0x20C2, 0x80}},
    {38208, 2, {0x2040, 0x2020, 0x3BFE, 0x2248, 0x4248, 0x7BFE, 0xA248, 0x2278, 0xFA00, 0x22FC, 0x2244, 0x2248, 0x2A28, 0x3410, 0x2468, 0x986}},
    {38209, 2, {0x2104, 0x2088, 0x3BFE, 0x2020, 0x4020, 0x79FC, 0xA020, 0x2020, 0xFBFE, 0x2000, 0x2020, 0x23FE, 0x2820, 0x3050, 0x2088, 0x306}},
    {38210, 2, {0x1020, 0x1124, 0x3CA8, 0x2020, 0x43FE, 0xBCA8, 0x1124, 0x1202, 0xFC40, 0x13FE, 0x1088, 0x1108, 0x1590, 0x1860, 0x1198, 0x604}},
    {38211, 2, {0x2084, 0x2044, 0x3848, 0x23FE, 0x4084, 0x7884, 0xA108, 0x214A, 0xFA52, 0x239C, 0x2084, 0x2108, 0x2908, 0x3252, 0x23DE, 0x42}},
    {38212, 2, {0x2090, 0x27FC, 0x3894, 0x23FC, 0x4290, 0x7BFE, 0xA112, 0x2216, 0xFDFC, 0x2104, 0x2124, 0x2124, 0x2924, 0x3050, 0x2088, 0x304}},
    {38213, 2, {0x2000, 0x23FC, 0x3A24, 0x4224, 0x43FC, 0x7A00, 0xA2FC, 0x2284, 0xFA84, 0x22FC, 0x2284, 0x2AFC, 0x3484, 0x2484, 0x8FC, 0x1084}},
    {38214, 2, {0x2088, 0x2088, 0x3BFE, 0x2088, 0x4000, 0x79FC, 0xA104, 0x21FC, 0xF904, 0x21FC, 0x2020, 0x23FE, 0x2850, 0x3088, 0x2104, 0x202}},
    {38215, 2, {0x2020, 0x2020, 0x3BFE, 0x4020, 0x41FC, 0x7904, 0xA1FC, 0x2104, 0xF9FC, 0x2104, 0x21FC, 0x2904, 0x37FE, 0x2088, 0x104, 0x202}},
    {38216, 2, {0x2028, 0x2024, 0x3BFE, 0x2020, 0x41FC, 0x7924, 0xA1FC, 0x2124, 0xF9FC, 0x2124, 0x2008, 0x23FE, 0x2888, 0x3048, 0x2048, 0x18}},
    {38217, 2, {0x2000, 0x23FE, 0x3800, 0x21FC, 0x4104, 0x7904, 0xA1FC, 0x2000, 0xFBFE, 0x228A, 0x2252, 0x23FE, 0x2A22, 0x3222, 0x222A, 0x204}},
    {38218, 2, {0x2000, 0x23FC, 0x3908, 0x21F8, 0x4108, 0x79F8, 0xA10E, 0x27F8, 0xF808, 0x27BC, 0x20A4, 0x22A4, 0x2928, 0x3290, 0x24A8, 0x846}},
    {38219, 2, {0x2020, 0x2124, 0x38A8, 0x2020, 0x43FE, 0x7A02, 0xA000, 0x21FC, 0xF904, 0x2104, 0x21FC, 0x2050, 0x2850, 0x3092, 0x2112, 0x20E}},
    {38220, 2, {0x20A0, 0x2090, 0x39FE, 0x2320, 0x45FC, 0x7920, 0xA1FC, 0x2120, 0xF9FE, 0x2100, 0x23FC, 0x2088, 0x289E, 0x3102, 0x220A, 0x404}},
    {38221, 2, {0x2040, 0x2080, 0x3BF8, 0x2208, 0x43F8, 0x7A08, 0xA3F8, 0x2208, 0xFBF8, 0x2040, 0x27FE, 0x20E0, 0x2950, 0x3248, 0x2446, 0x40}},
    {38222, 2, {0x1020, 0x1050, 0x3C88, 0x2174, 0x4202, 0xBCF8, 0x1088, 0x10F8, 0xFC0C, 0x11F0, 0x1020, 0x11FC, 0x1420, 0x1BFE, 0x1020, 0x60}},
    {38223, 2, {0x2040, 0x23BE, 0x3A12, 0x2292, 0x4252, 0x7AAA, 0xA324, 0x2040, 0xFBFE, 0x2222, 0x2222, 0x23FE, 0x2A22, 0x3222, 0x23FE, 0x202}},
    {38224, 2, {0x2040, 0x2020, 0x3BFE, 0x2000, 0x41FC, 0x7904, 0xA1FC, 0x2000, 0xFBFE, 0x2202, 0x22FA, 0x228A, 0x2AFA, 0x3202, 0x220A, 0x204}},
    {38225, 2, {0x2040, 0x2020, 0x3BFC, 0x2108, 0x4090, 0x7BFE, 0xA202, 0x2444, 0xF820, 0x23FC, 0x2080, 0x20F8, 0x2888, 0x3108, 0x2128, 0x210}},
    {38226, 2, {0x2104, 0x2084, 0x3888, 0x2000, 0x43FE, 0x7800, 0xA088, 0x2104, 0xFA02, 0x21FC, 0x2154, 0x2154, 0x2954, 0x3154, 0x27FE, 0x0}},
    {38227, 2, {0x2040, 0x2020, 0x3BFE, 0x2202, 0x4000, 0x79FC, 0xA040, 0x20A2, 0xFB34, 0x2058, 0x2094, 0x2334, 0x2852, 0x3090, 0x2350, 0x20}},
    {38228, 2, {0x2040, 0x2020, 0x3BFE, 0x2202, 0x4414, 0x79E0, 0xA100, 0x2100, 0xF9FC, 0x2110, 0x2110, 0x27FE, 0x2800, 0x3090, 0x2108, 0x204}},
    {38229, 2, {0x2040, 0x2020, 0x3BFE, 0x2202, 0x4050, 0x7888, 0xA124, 0x2050, 0xF888, 0x2104, 0x22FA, 0x2088, 0x2888, 0x3088, 0x20F8, 0x88}},
    {38230, 2, {0x1000, 0x13FE, 0x3C50, 0x21FC, 0x4154, 0xBD54, 0x11FC, 0x1000, 0xFDFC, 0x1000, 0x13FE, 0x1020, 0x14A8, 0x1924, 0x12A2, 0x40}},
    {38231, 2, {0x2020, 0x2124, 0x38A8, 0x23FE, 0x4202, 0x78F8, 0xA088, 0x2088, 0xF8F8, 0x2020, 0x2020, 0x21FC, 0x2820, 0x3020, 0x23FE, 0x0}},
    {38232, 2, {0x2000, 0x21F8, 0x3908, 0x21F8, 0x4108, 0x79F8, 0xA000, 0x23FC, 0xFA94, 0x23FC, 0x2000, 0x21F8, 0x2890, 0x3060, 0x2198, 0x606}},
    {38233, 2, {0x1000, 0x11FC, 0x3D24, 0x21FC, 0x4124, 0xBDFC, 0x1040, 0x1088, 0xFDF0, 0x1020, 0x1044, 0x13FE, 0x1422, 0x1924, 0x1222, 0x60}},
    {38234, 2, {0x2020, 0x2222, 0x3A22, 0x23FE, 0x4000, 0x7BDE, 0xA252, 0x2252, 0xFBDE, 0x2252, 0x2252, 0x23DE, 0x2A52, 0x3252, 0x255A, 0x8A4}},
    {38235, 2, {0x2020, 0x27FE, 0x3C20, 0x25FC, 0x4424, 0x77FE, 0xA424, 0x25FC, 0xFC20, 0x25FC, 0x2524, 0x25FC, 0x2D24, 0x39FC, 0x2924, 0x112C}},
    {38236, 2, {0x2080, 0x2040, 0x3BF8, 0x2110, 0x40A0, 0x77FE, 0xA000, 0x23F8, 0xFA08, 0x23F8, 0x2208, 0x23F8, 0x2920, 0x3122, 0x2222, 0xC1E}},
    {38237, 2, {0x2040, 0x2020, 0x3BFE, 0x2088, 0x4050, 0x7BFE, 0xA222, 0x22FA, 0xFA22, 0x22FA, 0x228A, 0x228A, 0x2AFA, 0x3202, 0x220A, 0x204}},
    {38238, 2, {0x2220, 0x2120, 0x3920, 0x47BE, 0x4240, 0x7A20, 0xA3BC, 0x22D0, 0xFA90, 0x2290, 0x22FE, 0x2A90, 0x34A8, 0x25A8, 0x844, 0x1082}},
    {38239, 2, {0x2210, 0x2110, 0x3910, 0x47BE, 0x4240, 0x7A00, 0xA3BE, 0x228A, 0xFA88, 0x22A8, 0x22AE, 0x2AA8, 0x34A8, 0x25A8, 0x85E, 0x1080}},
    {38240, 2, {0x2000, 0x23DE, 0x3842, 0x2252, 0x414A, 0x7A52, 0xA028, 0x20C4, 0xFB12, 0x2060, 0x2188, 0x2032, 0x29C4, 0x3018, 0x2060, 0x380}},
    {38241, 2, {0x2000, 0x23FE, 0x3850, 0x23FE, 0x4252, 0x7BFE, 0xA000, 0x21FC, 0xF904, 0x21FC, 0x2104, 0x21FC, 0x2820, 0x33FE, 0x2020, 0x20}},
    {38242, 2, {0x2000, 0x2FFE, 0x3800, 0x4A28, 0x4948, 0x7BEE, 0xA892, 0x2884, 0xFAA0, 0x2AA8, 0x2BE8, 0x2888, 0x3894, 0x2914, 0x924, 0x1242}},
    {38243, 2, {0x2020, 0x2020, 0x3BFE, 0x2050, 0x428A, 0x7904, 0xA3FE, 0x2504, 0xF9FC, 0x2104, 0x21FC, 0x2020, 0x2924, 0x3222, 0x24A2, 0x40}},
    {38244, 2, {0x20A0, 0x24A4, 0x3AA8, 0x20A0, 0x47FC, 0x7910, 0xA0A0, 0x27FC, 0xF840, 0x23F8, 0x2040, 0x27FC, 0x28A0, 0x3110, 0x2208, 0xC06}},
    {38245, 2, {0x2080, 0x20F8, 0x3910, 0x23FC, 0x4524, 0x79FC, 0xA124, 0x21FC, 0xF800, 0x23FE, 0x2000, 0x21FC, 0x2904, 0x31FC, 0x2104, 0x1FC}},
    {38246, 2, {0x2108, 0x2088, 0x3BC8, 0x2010, 0x43DE, 0x7A64, 0xA3D4, 0x2014, 0xFBD4, 0x2054, 0x2094, 0x20E8, 0x2B88, 0x3094, 0x2294, 0x122}},
    {38247, 2, {0x2200, 0x217C, 0x3D04, 0x2444, 0x45F4, 0x7C44, 0xA5F4, 0x2554, 0xFD74, 0x25D4, 0x2554, 0x25F4, 0x2CE4, 0x3554, 0x2444, 0x44C}},
    {38248, 2, {0x2108, 0x2090, 0x3BFC, 0x2090, 0x4294, 0x7998, 0xA090, 0x27FE, 0xF800, 0x21F8, 0x2108, 0x2108, 0x29F8, 0x3108, 0x2108, 0x1F8}},
    {38249, 2, {0x2040, 0x2020, 0x3BFE, 0x228A, 0x4124, 0x7A22, 0xA1FC, 0x2124, 0xF9FC, 0x2020, 0x23FE, 0x2222, 0x2BFE, 0x3222, 0x2020, 0x20}},
    {38250, 2, {0x2000, 0x23BE, 0x38A2, 0x20A2, 0x40BE, 0x7B88, 0xA208, 0x223E, 0xFA2A, 0x23AA, 0x20AA, 0x20BE, 0x2888, 0x308A, 0x257E, 0x202}},
    {38251, 2, {0x2020, 0x27A4, 0x38A8, 0x2292, 0x4114, 0x7A08, 0xA5F4, 0x2802, 0xFBF8, 0x2208, 0x2208, 0x23F8, 0x2A08, 0x3110, 0x2FFE, 0x0}},
    {38252, 2, {0x2088, 0x23FE, 0x38A8, 0x2090, 0x41FE, 0x7B20, 0xA1FC, 0x2120, 0xF9FC, 0x2120, 0x21FE, 0x2100, 0x2BFC, 0x3088, 0x2070, 0x38E}},
    {38253, 2, {0x1000, 0x11FC, 0x3C20, 0x23FE, 0x4222, 0xBDAC, 0x1020, 0x11AC, 0xFC00, 0x11FC, 0x1124, 0x1124, 0x15FC, 0x1924, 0x1124, 0x1FC}},
    {38254, 2, {0x2000, 0x23FC, 0x3A94, 0x2294, 0x43FC, 0x7800, 0xA7FE, 0x2000, 0xFBFC, 0x2204, 0x23FC, 0x20A2, 0x2914, 0x3308, 0x2D44, 0x182}},
    {38255, 2, {0x2000, 0x23F8, 0x3AA8, 0x22A8, 0x43F8, 0x7900, 0xA3FC, 0x2484, 0xFBE4, 0x22A4, 0x22A4, 0x23E4, 0x2894, 0x37F4, 0x2014, 0x8}},
    {38256, 2, {0x2020, 0x23FE, 0x3A88, 0x2250, 0x43FE, 0x7A50, 0xA2FC, 0x2254, 0xFBFE, 0x2254, 0x22FC, 0x2250, 0x2AD8, 0x3554, 0x2652, 0x850}},
    {38257, 2, {0x2040, 0x2020, 0x39FC, 0x2088, 0x4050, 0x7BFE, 0xA000, 0x21FC, 0xF904, 0x21FC, 0x2104, 0x21FC, 0x2820, 0x3294, 0x228A, 0x47A}},
    {38258, 2, {0x2080, 0x2040, 0x37FE, 0x4402, 0x4A24, 0x73BC, 0xA4A4, 0x2AA8, 0xF510, 0x22E8, 0x2404, 0x23F8, 0x2840, 0x3248, 0x2444, 0xC0}},
    {38259, 2, {0x2040, 0x2020, 0x3BFE, 0x2250, 0x43FE, 0x7A52, 0xA3FE, 0x2200, 0xFA94, 0x22D8, 0x2290, 0x22D2, 0x2A8E, 0x3400, 0x2554, 0xA2A}},
    {38260, 2, {0x2124, 0x2248, 0x3924, 0x2000, 0x43FC, 0x7A94, 0xA264, 0x2294, 0xFBFC, 0x2248, 0x2368, 0x2248, 0x2B6A, 0x324A, 0x2246, 0x362}},
    {38261, 2, {0x21F0, 0x2210, 0x3BFC, 0x2604, 0x43FC, 0x7A24, 0xA3B8, 0x2222, 0xF9FE, 0x2210, 0x27FC, 0x2244, 0x2BFC, 0x30D0, 0x214A, 0x63E}},
    {38262, 2, {0x2040, 0x27FC, 0x3800, 0x23B8, 0x42A8, 0x7BB8, 0xA110, 0x27FC, 0xF910, 0x27FC, 0x2110, 0x27FE, 0x2928, 0x3310, 0x2548, 0x186}},
    {38263, 2, {0x0, 0xFF8, 0x800, 0xFF0, 0x800, 0xFF0, 0x800, 0xFFFE, 0x1200, 0x1110, 0x10A0, 0x1040, 0x1220, 0x1418, 0x1806, 0x1000}},
    {38264, 2, {0x0, 0xFF8, 0x800, 0x800, 0xFF0, 0x800, 0x800, 0xFF0, 0x800, 0x800, 0xFFFE, 0x400, 0x820, 0x1010, 0x3FF8, 0x1008}},
    {38265, 2, {0x40, 0x3E40, 0x2040, 0x2078, 0x3C88, 0x2088, 0x3C88, 0x2110, 0x2010, 0xFE10, 0x1028, 0x2028, 0x4448, 0xFE44, 0x284, 0x102}},
    {38266, 2, {0x8, 0x3E3C, 0x21E0, 0x2020, 0x3C20, 0x2020, 0x3DFE, 0x2020, 0x2020, 0xFE50, 0x1050, 0x2050, 0x4488, 0xFE88, 0x304, 0x202}},
    {38267, 2, {0x20, 0x3E20, 0x20A0, 0x20A0, 0x3CFC, 0x20A0, 0x3D20, 0x2020, 0x21FE, 0xFE20, 0x1030, 0x2050, 0x4448, 0xFE88, 0x304, 0x202}},
    {38268, 2, {0x0, 0x3EFE, 0x2082, 0x20FE, 0x3C88, 0x20AA, 0x3CAA, 0x20AA, 0x20BE, 0xFE88, 0x10AA, 0x20AA, 0x44AA, 0xFEAA, 0x2BE, 0x102}},
    {38269, 2, {0x20, 0x3E20, 0x21FC, 0x2020, 0x3D54, 0x2088, 0x3D04, 0x22FA, 0x2088, 0xFEF8, 0x1088, 0x20F8, 0x4420, 0xFEA8, 0x324, 0x60}},
    {38270, 2, {0x0, 0x3BFE, 0x2124, 0x3A22, 0x27FE, 0x3A22, 0x2376, 0x22AA, 0xFB76, 0x2222, 0x2376, 0x4AAA, 0xFB76, 0x4A22, 0x22A, 0x204}},
    {38271, 2, {0x800, 0x810, 0x820, 0x840, 0x880, 0x900, 0x800, 0xFFFE, 0xA00, 0x900, 0x880, 0x840, 0x920, 0xA18, 0xC06, 0x800}},
    {38272, 2, {0x0, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x2244, 0x3E7C, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2004, 0x2014, 0x2008}},
    {38273, 2, {0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x4008, 0x4088, 0x4088, 0x4108, 0x4108, 0x4208, 0x4408, 0x4808, 0x4028, 0x4010}},
    {38274, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4004, 0x4004, 0x5FF4, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {38275, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4104, 0x4104, 0x4104, 0x4284, 0x4244, 0x4424, 0x4804, 0x4014, 0x4008}},
    {38276, 2, {0x10, 0x7710, 0x5510, 0x5510, 0x7710, 0x5510, 0x5510, 0x7710, 0x4110, 0x4128, 0x4128, 0x4128, 0x4128, 0x4144, 0x4544, 0x4282}},
    {38277, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x4504, 0x4214, 0x4008}},
    {38278, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4004, 0x4FE4, 0x4004, 0x5FF4, 0x4004, 0x4004, 0x4014, 0x4008}},
    {38279, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4204, 0x4284, 0x4244, 0x4224, 0x4204, 0x4204, 0x4214, 0x4008}},
    {38280, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4104, 0x4114, 0x4008}},
    {38281, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4084, 0x4084, 0x4FF4, 0x4184, 0x4284, 0x4484, 0x4884, 0x4284, 0x4114, 0x4008}},
    {38282, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x4924, 0x4924, 0x4924, 0x4924, 0x4FE4, 0x4004, 0x4014, 0x4008}},
    {38283, 2, {0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x4008, 0x4FC8, 0x4488, 0x4488, 0x5FE8, 0x4488, 0x4488, 0x4888, 0x50A8, 0x4010}},
    {38284, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4104, 0x5FF4, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4854, 0x5034, 0x4004, 0x400C}},
    {38285, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4104, 0x7FFC, 0x4404, 0x47E4, 0x4424, 0x4824, 0x5144, 0x6094, 0x4008}},
    {38286, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x5FF4, 0x4204, 0x4484, 0x4484, 0x4904, 0x5224, 0x67F4, 0x4214, 0x400C}},
    {38287, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38288, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x41C4, 0x5E04, 0x43C4, 0x5E04, 0x43E4, 0x5E04, 0x4224, 0x41E4, 0x4014, 0x4008}},
    {38289, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4384, 0x4544, 0x4924, 0x5114, 0x4104, 0x4114, 0x4008}},
    {38290, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x47E4, 0x4424, 0x47E4, 0x4424, 0x47E4, 0x4424, 0x48A4, 0x5044, 0x4014, 0x4008}},
    {38291, 2, {0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x4008, 0x4FC8, 0x4848, 0x4FC8, 0x4848, 0x4FC8, 0x4848, 0x4008, 0x4028, 0x4010}},
    {38292, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4104, 0x5FF4, 0x4044, 0x4444, 0x4284, 0x4104, 0x4284, 0x4444, 0x4854, 0x4008}},
    {38293, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4084, 0x4884, 0x4FF4, 0x4284, 0x4484, 0x5884, 0x4194, 0x4008}},
    {38294, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x5D14, 0x45A4, 0x4544, 0x4924, 0x4914, 0x5114, 0x6504, 0x4214, 0x4008}},
    {38295, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4084, 0x4484, 0x4284, 0x4884, 0x4484, 0x40F4, 0x5F84, 0x4084, 0x4094, 0x4088}},
    {38296, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4104, 0x4104, 0x4014, 0x4008}},
    {38297, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4924, 0x49A4, 0x4944, 0x4114, 0x4108}},
    {38298, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4104, 0x4FE4, 0x4444, 0x4244, 0x4284, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38299, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4104, 0x4924, 0x4544, 0x5FF4, 0x4104, 0x4104, 0x4104, 0x4114, 0x4008}},
    {38300, 2, {0x0, 0x7CF8, 0x4488, 0x7CF8, 0x4488, 0x7CF8, 0x4008, 0x5FE8, 0x4048, 0x4F48, 0x4948, 0x4F48, 0x4048, 0x4148, 0x40A8, 0x4010}},
    {38301, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4104, 0x4284, 0x4444, 0x4A24, 0x511C, 0x6FE4, 0x4024, 0x4044, 0x4284, 0x4114, 0x4088}},
    {38302, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4404, 0x4824, 0x5FF4, 0x4454, 0x4444, 0x5FF4, 0x4444, 0x4444, 0x4444, 0x4854, 0x4008}},
    {38303, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4144, 0x4484, 0x5524, 0x5614, 0x6454, 0x4C44, 0x73C4, 0x4014, 0x4008}},
    {38304, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4104, 0x4FE4, 0x4104, 0x4144, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38305, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7E7C, 0x4104, 0x5FF4, 0x4204, 0x4444, 0x4F84, 0x4124, 0x4644, 0x5884, 0x4344, 0x4C34, 0x4008}},
    {38306, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4444, 0x4284, 0x4FE4, 0x4104, 0x5FF4, 0x4104, 0x4284, 0x4444, 0x4824, 0x400C}},
    {38307, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x47E4, 0x4824, 0x5444, 0x4384, 0x4C64, 0x701C, 0x47C4, 0x4444, 0x4454, 0x47C8}},
    {38308, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4284, 0x4444, 0x5BB4, 0x4004, 0x47C4, 0x4444, 0x4444, 0x47C4, 0x4454, 0x4008}},
    {38309, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4544, 0x4924, 0x59F4, 0x6F24, 0x4944, 0x4884, 0x4954, 0x4A34, 0x4814, 0x400C}},
    {38310, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4284, 0x4444, 0x5834, 0x4444, 0x4444, 0x4444, 0x4AA4, 0x5114, 0x4014, 0x4008}},
    {38311, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4444, 0x5FF4, 0x4444, 0x4444, 0x7FFC, 0x4004, 0x4444, 0x4424, 0x4824, 0x400C}},
    {38312, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38313, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4104, 0x4FE4, 0x4924, 0x4924, 0x4FE4, 0x4114, 0x5FF4, 0x4014, 0x4004, 0x400C}},
    {38314, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4284, 0x4FE4, 0x4AA4, 0x4AA4, 0x4CE4, 0x4824, 0x4FE4, 0x4824, 0x400C}},
    {38315, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x5114, 0x5114, 0x5FF4, 0x5394, 0x5554, 0x5934, 0x5114, 0x5FF4, 0x400C}},
    {38316, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7E7C, 0x4104, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4914, 0x48A4, 0x4A44, 0x4C24, 0x481C}},
    {38317, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5FF4, 0x5014, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38318, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4034, 0x5DC4, 0x4444, 0x4844, 0x5DF4, 0x4444, 0x5444, 0x49F4, 0x5604, 0x61F4, 0x400C}},
    {38319, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x5084, 0x4884, 0x62A4, 0x5294, 0x4C84, 0x7024, 0x5044, 0x5184, 0x5614, 0x4008}},
    {38320, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4204, 0x4CE4, 0x4824, 0x4824, 0x4EE4, 0x4824, 0x4824, 0x4FE4, 0x4004, 0x400C}},
    {38321, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x4824, 0x5FF4, 0x682C, 0x4824, 0x4FE4, 0x4284, 0x4284, 0x4494, 0x4874, 0x500C}},
    {38322, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x4284, 0x4FE4, 0x4824, 0x4FE4, 0x4284, 0x4284, 0x4494, 0x5874, 0x4004, 0x400C}},
    {38323, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4104, 0x5FF4, 0x4104, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x5FF4, 0x410C}},
    {38324, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4444, 0x4824, 0x400C}},
    {38325, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4504, 0x4484, 0x4FF4, 0x5904, 0x6FE4, 0x4904, 0x4FE4, 0x4904, 0x4FF4, 0x4804, 0x400C}},
    {38326, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5FF4, 0x5014, 0x5FF4, 0x5014, 0x5FF4, 0x4008}},
    {38327, 2, {0x800, 0x8BDE, 0x5252, 0x23DE, 0x5252, 0x8BDE, 0x2202, 0x2222, 0xFA22, 0x2222, 0xB252, 0xAA4A, 0x2A8A, 0x2202, 0xA20A, 0x4204}},
    {38328, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4284, 0x4EE4, 0x4824, 0x4824, 0x4EE4, 0x4284, 0x4284, 0x5FF4, 0x400C}},
    {38329, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4204, 0x5FF4, 0x4924, 0x5FF4, 0x692C, 0x4FE4, 0x4924, 0x4FE4, 0x4114, 0x40F4, 0x400C}},
    {38330, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4104, 0x5FF4, 0x4444, 0x4384, 0x5C6C}},
    {38331, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7E7C, 0x4204, 0x47E4, 0x4824, 0x5244, 0x4CE4, 0x4824, 0x4824, 0x4EE4, 0x4824, 0x4824, 0x4FEC}},
    {38332, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4804, 0x4444, 0x5E44, 0x48A4, 0x4F14, 0x4A4C, 0x4A24, 0x4A84, 0x5244, 0x6624, 0x400C}},
    {38333, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4064, 0x5F84, 0x5104, 0x5FE4, 0x5094, 0x5874, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FEC}},
    {38334, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4084, 0x40A4, 0x5FF4, 0x4084, 0x4EA4, 0x4AA4, 0x4EA4, 0x4054, 0x46B4, 0x5914, 0x420C}},
    {38335, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x40E4, 0x5F24, 0x4944, 0x5FF4, 0x5014, 0x47C4, 0x4044, 0x4684, 0x4104, 0x4284, 0x4C4C}},
    {38336, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x5FF4, 0x4444, 0x7FFC, 0x4824, 0x57D4, 0x644C, 0x47C4, 0x4424, 0x43E4, 0x400C}},
    {38337, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4EE4, 0x4AA4, 0x4EE4, 0x4AA4, 0x4EE4, 0x4824, 0x4824, 0x4824, 0x4864, 0x400C}},
    {38338, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4924, 0x4544, 0x5FF4, 0x4204, 0x7FFC, 0x4824, 0x57D4, 0x644C, 0x44C4, 0x4424, 0x43EC}},
    {38339, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4144, 0x5FF4, 0x4284, 0x444C}},
    {38340, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4284, 0x4FE4, 0x4AA4, 0x4FE4, 0x4204, 0x5FF4, 0x4444, 0x4384, 0x5C6C}},
    {38341, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4204, 0x43F4, 0x4204, 0x5FE4, 0x4444, 0x4384, 0x5C6C}},
    {38342, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x47C4, 0x4444, 0x47C4, 0x4004, 0x4EE4, 0x4AA4, 0x4AA4, 0x4EE4, 0x4014, 0x4008}},
    {38343, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7E7C, 0x4104, 0x4FE4, 0x4444, 0x4284, 0x5FF4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FEC}},
    {38344, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4204, 0x4FC4, 0x4444, 0x5FF4, 0x4824, 0x4FE4, 0x4104, 0x4FE4, 0x5104, 0x5FF4, 0x4108}},
    {38345, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5FF4, 0x4284, 0x5FF4, 0x5294, 0x5FF4, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x400C}},
    {38346, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4834, 0x45C4, 0x4044, 0x53FC, 0x4844, 0x41F4, 0x4514, 0x49F4, 0x5014, 0x4008}},
    {38347, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4094, 0x4EA4, 0x424C, 0x5454, 0x4824, 0x57D4, 0x410C, 0x4FE4, 0x4284, 0x4444, 0x582C}},
    {38348, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x5FF4, 0x4104, 0x5FF4, 0x5554, 0x5394, 0x5FF4, 0x4544, 0x4924, 0x5114, 0x410C}},
    {38349, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4114, 0x4FE4, 0x4144, 0x5FF4, 0x4204, 0x4FE4, 0x7424, 0x47E4, 0x4424, 0x47E4, 0x400C}},
    {38350, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4204, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4114, 0x5DA4, 0x4544, 0x4924, 0x531C}},
    {38351, 2, {0xFC7E, 0x8442, 0xFC7E, 0x8442, 0xFC7E, 0x8002, 0xBFE2, 0xA022, 0xAFA2, 0xA222, 0xAFA2, 0xAAAA, 0xAFAA, 0xA25A, 0xDFCA, 0x8046}},
    {38352, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4FE4, 0x4104, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x7FFC, 0x4444, 0x482C}},
    {38353, 2, {0x7C7C, 0x4444, 0x7D7C, 0x4644, 0x7FFC, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4104, 0x5FF4, 0x4544, 0x4924, 0x410C}},
    {38354, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x47C4, 0x4444, 0x47C4, 0x4444, 0x47C4, 0x4004, 0x4EE4, 0x4AA4, 0x4664, 0x4AA4, 0x466C}},
    {38355, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4924, 0x4FE4, 0x4004, 0x5FF4, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4444, 0x5FF4, 0x400C}},
    {38356, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4104, 0x4FE4, 0x4104, 0x5FF4, 0x4444, 0x4FE4, 0x4004, 0x4FE4, 0x4AA4, 0x5FF4, 0x400C}},
    {38357, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x5144, 0x4A44, 0x5F74, 0x4494, 0x5524, 0x5544, 0x5F44, 0x4444, 0x48A4, 0x4914, 0x5208}},
    {38358, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FF4, 0x4904, 0x4FE4, 0x4904, 0x4FE4, 0x4904, 0x4FF4, 0x4014, 0x5554, 0x402C}},
    {38359, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4844, 0x4844, 0x5294, 0x5CE4, 0x4844, 0x5294, 0x5EF4, 0x4004, 0x4AA4, 0x5154, 0x400C}},
    {38360, 2, {0x0, 0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4024, 0x5E24, 0x40F4, 0x5E24, 0x52A4, 0x5E64, 0x5224, 0x4F24, 0x78B4, 0x5048}},
    {38361, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4804, 0x4444, 0x5E7C, 0x5294, 0x5E94, 0x5154, 0x5E24, 0x6A54, 0x4E8C, 0x4004, 0x400C}},
    {38362, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4804, 0x49E4, 0x5D24, 0x49E4, 0x4924, 0x7DE4, 0x4924, 0x49E4, 0x54C4, 0x5154, 0x623C}},
    {38363, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4924, 0x4544, 0x5FF4, 0x5014, 0x47C4, 0x4444, 0x47C4, 0x4104, 0x4FE4, 0x4104, 0x5FEC}},
    {38364, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4844, 0x5294, 0x5CE4, 0x4A54, 0x5EF4, 0x4284, 0x5294, 0x5EF4, 0x4284, 0x4494, 0x4888}},
    {38365, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4024, 0x5F24, 0x517C, 0x5F94, 0x5114, 0x5F54, 0x5124, 0x5F24, 0x4A54, 0x5194, 0x400C}},
    {38366, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x5C44, 0x447C, 0x7F54, 0x5294, 0x5E14, 0x5254, 0x5E24, 0x53A4, 0x7E54, 0x4298}},
    {38367, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x46C4, 0x5834, 0x47C4, 0x4444, 0x47C4, 0x4004, 0x4EE4, 0x4AA4, 0x4664, 0x4AA4, 0x466C}},
    {38368, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7D7C, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x7FFC, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4444, 0x482C}},
    {38369, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7EFC, 0x4AA4, 0x4EE4, 0x4004, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x5FF4, 0x4108}},
    {38370, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4084, 0x5E44, 0x53F4, 0x5EA4, 0x5044, 0x5EF4, 0x6A44, 0x4AE4, 0x4E44, 0x4A54, 0x4008}},
    {38371, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4924, 0x4544, 0x5FF4, 0x5014, 0x47C4, 0x4444, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FEC}},
    {38372, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7FFC, 0x4AA4, 0x4FE4, 0x4004, 0x5FF4, 0x4004, 0x4FC4, 0x4844, 0x4FC4, 0x4524, 0x4CC4, 0x562C}},
    {38373, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x5084, 0x4BE4, 0x4084, 0x47F4, 0x5944, 0x4BE4, 0x4884, 0x4FF4, 0x4884, 0x4884, 0x57EC}},
    {38374, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4284, 0x4FE4, 0x4AA4, 0x4FE4, 0x4AA4, 0x5FF4, 0x4444, 0x47C4, 0x4444, 0x4284, 0x4FEC}},
    {38375, 2, {0x7C7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4084, 0x63E4, 0x5084, 0x47F4, 0x7224, 0x53E4, 0x5154, 0x5324, 0x5514, 0x6FF4, 0x400C}},
    {38376, 2, {0x2000, 0x13FC, 0x1004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {38377, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x4004, 0x4004, 0x4004, 0x5FF4, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4004, 0x4014, 0x4008}},
    {38378, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x4104, 0x4104, 0x4104, 0x4104, 0x4284, 0x4244, 0x4424, 0x4824, 0x4004, 0x4004, 0x4014, 0x4008}},
    {38379, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x5FF4, 0x4004, 0x4004, 0x4004, 0x4FE4, 0x4004, 0x4004, 0x4004, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38380, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x4FE4, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4104, 0x4104, 0x4104, 0x4014, 0x4008}},
    {38381, 2, {0x2000, 0x17FC, 0x4, 0x4084, 0x4084, 0x5FF4, 0x4084, 0x4184, 0x4284, 0x4484, 0x4884, 0x5084, 0x4284, 0x4104, 0x4014, 0x4008}},
    {38382, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4004, 0x4014, 0x4008}},
    {38383, 2, {0x2000, 0x13FC, 0x1004, 0x4004, 0x5F84, 0x4084, 0x4884, 0x4884, 0x4FE4, 0x4024, 0x4024, 0x5FA4, 0x4024, 0x4144, 0x4094, 0x4008}},
    {38384, 2, {0x2000, 0x13FC, 0x1004, 0x4004, 0x4004, 0x5FF4, 0x4104, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4014, 0x4008}},
    {38385, 2, {0x2000, 0x17FC, 0x104, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4114, 0x4114, 0x4154, 0x4124, 0x400C}},
    {38386, 2, {0x2000, 0x17FC, 0x4, 0x4104, 0x4104, 0x4104, 0x5FF4, 0x4104, 0x4384, 0x4544, 0x4924, 0x5114, 0x4104, 0x4104, 0x4014, 0x4008}},
    {38387, 2, {0x2000, 0x17FC, 0x204, 0x4204, 0x4204, 0x5FF4, 0x4404, 0x4484, 0x4884, 0x4904, 0x5144, 0x5224, 0x67F4, 0x4214, 0x4014, 0x4008}},
    {38388, 2, {0x2000, 0x13FC, 0x1004, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4444, 0x4444, 0x47C4, 0x4004, 0x4014, 0x4008}},
    {38389, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x4104, 0x5FF4, 0x4044, 0x4444, 0x4284, 0x4104, 0x4284, 0x4444, 0x4844, 0x4004, 0x4014, 0x4008}},
    {38390, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x4104, 0x5FF4, 0x4004, 0x4004, 0x47C4, 0x4444, 0x4444, 0x4454, 0x4834, 0x5004, 0x4014, 0x4008}},
    {38391, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x4104, 0x4104, 0x4404, 0x5424, 0x5414, 0x6454, 0x4444, 0x43C4, 0x4004, 0x4004, 0x4014, 0x4008}},
    {38392, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x4FE4, 0x4924, 0x4924, 0x4FE4, 0x4924, 0x4924, 0x4FE4, 0x4924, 0x4104, 0x4104, 0x4114, 0x4008}},
    {38393, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4FE4, 0x4924, 0x4924, 0x49A4, 0x4944, 0x4104, 0x4114, 0x4008}},
    {38394, 2, {0x2000, 0x17FC, 0x4, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x4104, 0x4FE4, 0x4104, 0x4104, 0x5FF4, 0x4004, 0x400C}},
    {38395, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x5FF4, 0x4444, 0x47C4, 0x4444, 0x47C4, 0x4444, 0x4474, 0x5FC4, 0x4044, 0x4044, 0x4054, 0x4008}},
    {38396, 2, {0x2000, 0x17FC, 0x4, 0x4844, 0x4444, 0x45F4, 0x4044, 0x5C44, 0x44A4, 0x4494, 0x4514, 0x4A04, 0x51F4, 0x4004, 0x4014, 0x4008}},
    {38397, 2, {0x2000, 0x17FC, 0x4, 0x4104, 0x4104, 0x4FE4, 0x4924, 0x4924, 0x4924, 0x4FE4, 0x4904, 0x4124, 0x41F4, 0x5E14, 0x4804, 0x400C}},
    {38398, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x47C4, 0x4444, 0x4444, 0x47C4, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4004, 0x4014, 0x4008}},
    {38399, 2, {0x2000, 0x17FC, 0x104, 0x4924, 0x4924, 0x4FE4, 0x4004, 0x4FE4, 0x4024, 0x4024, 0x4FE4, 0x4804, 0x4814, 0x47F4, 0x4004, 0x400C}},
    {38400, 2, {0x2000, 0x17FC, 0x4, 0x4544, 0x4524, 0x4904, 0x59F4, 0x6F04, 0x4924, 0x4924, 0x48C4, 0x4894, 0x4954, 0x4A34, 0x4814, 0x4008}},
    {38401, 2, {0x2000, 0x17FC, 0x204, 0x4204, 0x47E4, 0x4C44, 0x5284, 0x4104, 0x4284, 0x4C64, 0x77DC, 0x4444, 0x4444, 0x47C4, 0x4014, 0x4008}},
    {38402, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x4104, 0x5FF4, 0x4204, 0x4444, 0x4F84, 0x4124, 0x4644, 0x5884, 0x4344, 0x4C24, 0x4004, 0x400C}},
    {38403, 2, {0x2000, 0x17FC, 0x4, 0x4004, 0x5FF4, 0x5114, 0x5114, 0x5FF4, 0x5114, 0x5394, 0x5554, 0x5934, 0x5114, 0x5FF4, 0x4004, 0x400C}},
    {38404, 2, {0x2000, 0x17FC, 0x4, 0x4404, 0x47C4, 0x4884, 0x5104, 0x6FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4114, 0x4114, 0x40F4, 0x400C}},
    {38405, 2, {0x2000, 0x17FC, 0x4, 0x4444, 0x4284, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4284, 0x4284, 0x4484, 0x4894, 0x5074, 0x4008}},
    {38406, 2, {0x2000, 0x17FC, 0x204, 0x4104, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4914, 0x48A4, 0x4A44, 0x4C24, 0x4814, 0x4004, 0x400C}},
    {38407, 2, {0x2000, 0x17FC, 0x104, 0x4114, 0x4FE4, 0x4144, 0x5FF4, 0x4204, 0x4FE4, 0x7424, 0x47E4, 0x4424, 0x47E4, 0x4424, 0x4014, 0x4008}},
    {38408, 2, {0x2000, 0x17FC, 0x84, 0x40A4, 0x4094, 0x5FF4, 0x4084, 0x4EA4, 0x4AA4, 0x4EA4, 0x40C4, 0x4654, 0x58B4, 0x4114, 0x4204, 0x400C}},
    {38409, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x5FF4, 0x4444, 0x4924, 0x711C, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4114, 0x40F4, 0x4008}},
    {38410, 2, {0x2000, 0x17FC, 0x4, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5FF4, 0x5014, 0x5FF4, 0x5014, 0x5FF4, 0x4004, 0x400C}},
    {38411, 2, {0x2000, 0x17FC, 0x4, 0x4204, 0x4CE4, 0x4824, 0x4EE4, 0x4824, 0x4FE4, 0x4284, 0x4284, 0x4494, 0x4494, 0x4874, 0x5004, 0x400C}},
    {38412, 2, {0x2000, 0x17FC, 0x4, 0x40E4, 0x5F04, 0x4A24, 0x4544, 0x5FF4, 0x5014, 0x47C4, 0x4044, 0x4684, 0x4104, 0x4284, 0x4444, 0x482C}},
    {38413, 2, {0x2000, 0x17FC, 0x4, 0x4064, 0x5F84, 0x5104, 0x5FF4, 0x5084, 0x5C54, 0x5034, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x482C}},
    {38414, 2, {0x2000, 0x17FC, 0x204, 0x4204, 0x47E4, 0x4824, 0x5244, 0x4CE4, 0x4824, 0x4824, 0x4EE4, 0x4824, 0x4824, 0x4FE4, 0x4014, 0x4008}},
    {38415, 2, {0x2000, 0x17FC, 0x4, 0x4844, 0x4444, 0x5EA4, 0x48A4, 0x4914, 0x4E0C, 0x4A44, 0x4A24, 0x4A04, 0x4A84, 0x5244, 0x5624, 0x600C}},
    {38416, 2, {0x2000, 0x17FC, 0x4, 0x4444, 0x4284, 0x4FE4, 0x4924, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x5FF4, 0x4104, 0x4104, 0x4014, 0x4008}},
    {38417, 2, {0x2000, 0x17FC, 0x104, 0x4104, 0x5FF4, 0x4104, 0x5FF4, 0x5114, 0x5554, 0x5394, 0x5FF4, 0x4544, 0x4924, 0x5124, 0x4104, 0x410C}},
    {38418, 2, {0x2000, 0x17FC, 0x4, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4144, 0x5FF4, 0x4104, 0x4284, 0x4444, 0x482C}},
    {38419, 2, {0x2000, 0x17FC, 0x104, 0x4FE4, 0x4924, 0x4FE4, 0x4104, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4924, 0x4924, 0x4284, 0x4444, 0x482C}},
    {38420, 2, {0x2000, 0x17FC, 0x4, 0x4834, 0x45C4, 0x6444, 0x5044, 0x51F4, 0x4444, 0x4844, 0x79F4, 0x4914, 0x4914, 0x49F4, 0x4004, 0x400C}},
    {38421, 2, {0x2000, 0x17FC, 0x4, 0x4084, 0x5E94, 0x42A4, 0x4A4C, 0x4454, 0x4824, 0x57D4, 0x610C, 0x4FE4, 0x4104, 0x4284, 0x4444, 0x582C}},
    {38422, 2, {0x2000, 0x13FC, 0x1104, 0x4FE4, 0x4104, 0x5FF4, 0x4204, 0x4444, 0x4FE4, 0x4004, 0x4FE4, 0x4AA4, 0x4AA4, 0x5FF4, 0x4014, 0x4008}},
    {38423, 2, {0x2000, 0x17FC, 0x104, 0x5FF4, 0x4104, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x7FFC, 0x4444, 0x482C}},
    {38424, 2, {0x2000, 0x17FC, 0x4, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x5EF4, 0x4214, 0x5294, 0x4A54, 0x5294, 0x4A54, 0x442C}},
    {38425, 2, {0x2000, 0x17FC, 0x4, 0x5144, 0x4A44, 0x5F74, 0x4494, 0x5524, 0x5544, 0x5F44, 0x4444, 0x48A4, 0x4914, 0x5204, 0x4014, 0x4008}},
    {38426, 2, {0x2000, 0x17FC, 0x4, 0x5C44, 0x447C, 0x7F54, 0x5294, 0x5E14, 0x5254, 0x5E24, 0x53A4, 0x7E54, 0x4294, 0x4304, 0x4014, 0x4008}},
    {38427, 2, {0x2000, 0x17FC, 0x4, 0x5FF4, 0x5294, 0x5FF4, 0x4004, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4514, 0x4CA4, 0x7444, 0x462C}},
    {38428, 2, {0x200, 0x400, 0x3FF0, 0x2010, 0x3FF0, 0x2000, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {38429, 2, {0x0, 0x7C00, 0x4400, 0x4800, 0x4800, 0x5000, 0x4800, 0x4800, 0x4400, 0x4400, 0x4400, 0x6800, 0x5000, 0x4000, 0x4000, 0x4000}},
    {38430, 2, {0x40, 0x7C40, 0x4440, 0x4840, 0x4BFC, 0x5044, 0x4844, 0x4844, 0x4444, 0x4484, 0x4484, 0x6884, 0x5104, 0x4104, 0x4228, 0x4410}},
    {38431, 2, {0x40, 0x7C40, 0x4440, 0x4840, 0x4840, 0x5040, 0x4840, 0x4840, 0x44A0, 0x44A0, 0x44A0, 0x6910, 0x5110, 0x4208, 0x4404, 0x4802}},
    {38432, 2, {0x0, 0x7DF0, 0x4490, 0x4890, 0x4890, 0x5090, 0x4890, 0x49F0, 0x4490, 0x4490, 0x4490, 0x6892, 0x508A, 0x408A, 0x4086, 0x4082}},
    {38433, 2, {0x8, 0x7C3C, 0x45E0, 0x4820, 0x4820, 0x5020, 0x4820, 0x4BFE, 0x4420, 0x4420, 0x4420, 0x6820, 0x5020, 0x4020, 0x4020, 0x4020}},
    {38434, 2, {0x0, 0x7C00, 0x47FC, 0x4890, 0x4890, 0x5090, 0x4890, 0x4890, 0x4490, 0x4490, 0x4490, 0x6892, 0x5112, 0x4112, 0x420E, 0x4400}},
    {38435, 2, {0x40, 0x7C40, 0x4480, 0x48FE, 0x4900, 0x5200, 0x48FC, 0x4808, 0x4410, 0x4420, 0x4440, 0x6880, 0x5102, 0x4102, 0x40FE, 0x4000}},
    {38436, 2, {0x20, 0x7C20, 0x4420, 0x4920, 0x492C, 0x5134, 0x4964, 0x4BA4, 0x4524, 0x4534, 0x4528, 0x6922, 0x5122, 0x4102, 0x40FE, 0x4000}},
    {38437, 2, {0x40, 0x7840, 0x4840, 0x5044, 0x5068, 0x6770, 0x5160, 0x4950, 0x4950, 0x4A48, 0x6A48, 0x5444, 0x4842, 0x4040, 0x4140, 0x4080}},
    {38438, 2, {0x40, 0x7C40, 0x4440, 0x4840, 0x4944, 0x5144, 0x4948, 0x4A50, 0x4440, 0x44A0, 0x44A0, 0x6890, 0x5110, 0x4108, 0x4204, 0x4402}},
    {38439, 2, {0x10, 0x7D10, 0x4490, 0x4890, 0x4810, 0x5110, 0x4890, 0x4890, 0x4410, 0x441E, 0x45F0, 0x6810, 0x5010, 0x4010, 0x4010, 0x4010}},
    {38440, 2, {0x0, 0x7DFE, 0x4500, 0x4900, 0x4978, 0x5148, 0x4948, 0x4948, 0x4548, 0x4568, 0x4550, 0x6942, 0x5142, 0x4142, 0x423E, 0x4400}},
    {38441, 2, {0x8, 0x7C68, 0x4788, 0x4888, 0x4888, 0x5088, 0x4888, 0x4BFE, 0x4488, 0x4488, 0x4488, 0x6888, 0x5108, 0x4108, 0x4208, 0x4408}},
    {38442, 2, {0x8, 0x7C1C, 0x45E0, 0x4900, 0x4900, 0x51FC, 0x4944, 0x4944, 0x4544, 0x4528, 0x4528, 0x6910, 0x5110, 0x4228, 0x4244, 0x4482}},
    {38443, 2, {0x0, 0x7C00, 0x45FE, 0x4810, 0x4810, 0x5020, 0x4820, 0x4868, 0x44A4, 0x4522, 0x4622, 0x6820, 0x5020, 0x4020, 0x4020, 0x4020}},
    {38444, 2, {0x80, 0x7C40, 0x4440, 0x4BFC, 0x4800, 0x5000, 0x49F0, 0x4910, 0x4510, 0x4510, 0x4510, 0x6912, 0x5112, 0x4212, 0x420E, 0x4400}},
    {38445, 2, {0x40, 0x7C40, 0x4480, 0x4888, 0x4904, 0x53FE, 0x4802, 0x4890, 0x4490, 0x4490, 0x4490, 0x6890, 0x5112, 0x4112, 0x420E, 0x4400}},
    {38446, 2, {0x0, 0x7DFC, 0x4400, 0x4800, 0x4800, 0x53FE, 0x4890, 0x4890, 0x4490, 0x4490, 0x4490, 0x6892, 0x5112, 0x4112, 0x420E, 0x4400}},
    {38447, 2, {0x20, 0x7C20, 0x4420, 0x4920, 0x4920, 0x5120, 0x493C, 0x4920, 0x4520, 0x4520, 0x4520, 0x6920, 0x5120, 0x4120, 0x47FE, 0x4000}},
    {38448, 2, {0x10, 0x7D10, 0x4510, 0x4912, 0x4912, 0x5114, 0x49D8, 0x4910, 0x4510, 0x4510, 0x4510, 0x6912, 0x5152, 0x4192, 0x410E, 0x4000}},
    {38449, 2, {0x88, 0x7C88, 0x4488, 0x4888, 0x4BFE, 0x5088, 0x4888, 0x4888, 0x4488, 0x47FE, 0x4488, 0x6888, 0x5088, 0x4108, 0x4108, 0x4208}},
    {38450, 2, {0x40, 0x7C20, 0x4420, 0x4BFE, 0x4880, 0x5080, 0x4880, 0x48FC, 0x4484, 0x4484, 0x4484, 0x6884, 0x5104, 0x4104, 0x4228, 0x4410}},
    {38451, 2, {0x0, 0x7C00, 0x45FC, 0x4904, 0x4904, 0x5104, 0x4904, 0x4904, 0x45FC, 0x4504, 0x4504, 0x6904, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {38452, 2, {0x0, 0x7DFC, 0x4504, 0x4904, 0x4904, 0x51FC, 0x4904, 0x4904, 0x4504, 0x45FC, 0x4504, 0x6904, 0x5204, 0x4204, 0x4414, 0x4808}},
    {38453, 2, {0x40, 0x7C40, 0x4440, 0x4BFE, 0x4880, 0x50A0, 0x4920, 0x49FC, 0x4420, 0x4420, 0x4420, 0x6BFE, 0x5020, 0x4020, 0x4020, 0x4020}},
    {38454, 2, {0x20, 0x7C20, 0x4450, 0x4850, 0x4888, 0x5104, 0x4A02, 0x4888, 0x4488, 0x4488, 0x4488, 0x6888, 0x5088, 0x4108, 0x4108, 0x4208}},
    {38455, 2, {0x0, 0x7C00, 0x45FE, 0x4820, 0x4820, 0x5020, 0x4920, 0x4920, 0x453C, 0x4520, 0x4520, 0x6920, 0x5120, 0x4120, 0x47FE, 0x4000}},
    {38456, 2, {0x40, 0x7C20, 0x45FC, 0x4904, 0x4904, 0x51FC, 0x4900, 0x4900, 0x457C, 0x4508, 0x4510, 0x6920, 0x5142, 0x4282, 0x427E, 0x4400}},
    {38457, 2, {0x20, 0x7C20, 0x4420, 0x49FC, 0x4820, 0x5020, 0x4820, 0x4BFE, 0x4420, 0x4440, 0x4440, 0x6888, 0x5104, 0x43FE, 0x4102, 0x4000}},
    {38458, 2, {0x4, 0x7C1E, 0x45F0, 0x4910, 0x4910, 0x5110, 0x4910, 0x49FE, 0x4510, 0x4510, 0x4510, 0x6908, 0x510A, 0x414A, 0x41A6, 0x4112}},
    {38459, 2, {0x0, 0x7DF8, 0x4508, 0x4908, 0x4908, 0x51F8, 0x4908, 0x4908, 0x4508, 0x45F8, 0x4508, 0x6908, 0x5108, 0x4108, 0x47FE, 0x4000}},
    {38460, 2, {0x80, 0x7C80, 0x4480, 0x48FE, 0x4940, 0x5140, 0x4A40, 0x487C, 0x4440, 0x4440, 0x4440, 0x687E, 0x5040, 0x4040, 0x4040, 0x4040}},
    {38461, 2, {0x20, 0x7C20, 0x4420, 0x4820, 0x483E, 0x5020, 0x4820, 0x4820, 0x45FC, 0x4504, 0x4504, 0x6904, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {38462, 2, {0x20, 0x7C20, 0x4450, 0x4850, 0x4888, 0x5124, 0x4A12, 0x4810, 0x45FC, 0x4404, 0x4408, 0x6888, 0x5050, 0x4020, 0x4010, 0x4010}},
    {38463, 2, {0x0, 0x7DFE, 0x4408, 0x4808, 0x49E8, 0x5128, 0x4928, 0x4928, 0x4528, 0x4528, 0x45E8, 0x6928, 0x5008, 0x4008, 0x4028, 0x4010}},
    {38464, 2, {0x40, 0x7C20, 0x4420, 0x49FE, 0x4902, 0x5204, 0x4900, 0x4908, 0x4510, 0x4520, 0x45C0, 0x6904, 0x5104, 0x4104, 0x40FC, 0x4000}},
    {38465, 2, {0x80, 0x7C80, 0x44FE, 0x4900, 0x4A20, 0x5120, 0x492C, 0x4974, 0x47A4, 0x4524, 0x4534, 0x6928, 0x5122, 0x4102, 0x40FE, 0x4000}},
    {38466, 2, {0x20, 0x7820, 0x4820, 0x53FE, 0x5222, 0x6224, 0x5220, 0x4BFC, 0x4A84, 0x4A88, 0x6A48, 0x5250, 0x4220, 0x4450, 0x4488, 0x4906}},
    {38467, 2, {0x0, 0x7DFE, 0x4420, 0x4820, 0x4820, 0x51FE, 0x4922, 0x4922, 0x4552, 0x454A, 0x458A, 0x6902, 0x5102, 0x4102, 0x410A, 0x4104}},
    {38468, 2, {0x88, 0x7C88, 0x4488, 0x4908, 0x497E, 0x5308, 0x4D08, 0x4948, 0x4528, 0x4528, 0x4508, 0x6908, 0x5108, 0x4108, 0x4128, 0x4110}},
    {38469, 2, {0x0, 0x79FC, 0x4800, 0x5000, 0x5000, 0x63FE, 0x5020, 0x4820, 0x4928, 0x4924, 0x6A24, 0x5222, 0x4422, 0x4020, 0x40A0, 0x4040}},
    {38470, 2, {0x20, 0x7C20, 0x4420, 0x49FC, 0x4820, 0x5020, 0x4820, 0x4BFE, 0x4420, 0x4420, 0x4524, 0x6924, 0x5124, 0x4124, 0x41FC, 0x4004}},
    {38471, 2, {0x90, 0x7C88, 0x4488, 0x4880, 0x4BFE, 0x50A0, 0x48A0, 0x48A4, 0x44A4, 0x4528, 0x4528, 0x6932, 0x5222, 0x4262, 0x449E, 0x4800}},
    {38472, 2, {0x40, 0x7840, 0x4840, 0x57FE, 0x5080, 0x6120, 0x5120, 0x4A20, 0x4BFC, 0x4820, 0x6928, 0x5124, 0x4222, 0x4422, 0x40A0, 0x4040}},
    {38473, 2, {0x0, 0x7DFC, 0x4408, 0x4810, 0x4830, 0x5048, 0x4884, 0x4B02, 0x4400, 0x45FC, 0x4420, 0x6820, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38474, 2, {0x20, 0x7C20, 0x447C, 0x4884, 0x4948, 0x5030, 0x4820, 0x4848, 0x4590, 0x443E, 0x4442, 0x69A4, 0x5018, 0x4010, 0x4060, 0x4180}},
    {38475, 2, {0x0, 0x7BFE, 0x4820, 0x5020, 0x52FC, 0x62A4, 0x52A4, 0x4AA4, 0x4AD4, 0x4A84, 0x6A84, 0x5294, 0x4288, 0x4200, 0x43FE, 0x4000}},
    {38476, 2, {0x0, 0x7DFE, 0x4420, 0x4820, 0x4840, 0x51FC, 0x4904, 0x4904, 0x4504, 0x45FC, 0x4504, 0x6904, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {38477, 2, {0x80, 0x7C80, 0x44FC, 0x4908, 0x4A90, 0x5060, 0x4998, 0x4E26, 0x4420, 0x45FC, 0x4420, 0x6A20, 0x53FE, 0x4020, 0x4020, 0x4020}},
    {38478, 2, {0x20, 0x7D20, 0x4520, 0x49FC, 0x4920, 0x5220, 0x4820, 0x4BFE, 0x4470, 0x44A8, 0x44A8, 0x6924, 0x5124, 0x4222, 0x4020, 0x4020}},
    {38479, 2, {0x20, 0x7C20, 0x45FE, 0x4840, 0x4840, 0x50FC, 0x4884, 0x4984, 0x46FC, 0x4484, 0x4484, 0x68FC, 0x5084, 0x4084, 0x4094, 0x4088}},
    {38480, 2, {0x0, 0x7BF8, 0x4A08, 0x5208, 0x53F8, 0x6208, 0x5208, 0x4BF8, 0x4A44, 0x4A48, 0x6A30, 0x5220, 0x4210, 0x4288, 0x4306, 0x4200}},
    {38481, 2, {0x0, 0x7DFE, 0x4420, 0x4820, 0x4840, 0x51FE, 0x4952, 0x4952, 0x4552, 0x4552, 0x4552, 0x6952, 0x5152, 0x4142, 0x410A, 0x4104}},
    {38482, 2, {0x40, 0x7C40, 0x44FC, 0x4904, 0x4A08, 0x51FE, 0x4900, 0x497C, 0x4544, 0x4544, 0x4554, 0x6948, 0x5142, 0x4242, 0x423E, 0x4400}},
    {38483, 2, {0x20, 0x7820, 0x4BFE, 0x4850, 0x5088, 0x5104, 0x62FA, 0x5000, 0x4BFE, 0x4840, 0x4880, 0x69FC, 0x5004, 0x4004, 0x4028, 0x4010}},
    {38484, 2, {0x40, 0x7820, 0x4820, 0x57FE, 0x5040, 0x6084, 0x5108, 0x4BF0, 0x4822, 0x4844, 0x6988, 0x5610, 0x4020, 0x4050, 0x4188, 0x4604}},
    {38485, 2, {0x40, 0x7840, 0x4840, 0x57FC, 0x5040, 0x6248, 0x5150, 0x4840, 0x4FFE, 0x48A0, 0x68A0, 0x5110, 0x4110, 0x4208, 0x4404, 0x4802}},
    {38486, 2, {0x40, 0x7C40, 0x4488, 0x4904, 0x4BFE, 0x5002, 0x4888, 0x4944, 0x4642, 0x44F8, 0x4588, 0x6A50, 0x5020, 0x4050, 0x4188, 0x4606}},
    {38487, 2, {0x20, 0x7D24, 0x44A4, 0x48A8, 0x4820, 0x51FC, 0x4904, 0x4904, 0x45FC, 0x4504, 0x4504, 0x69FC, 0x5104, 0x4104, 0x4114, 0x4108}},
    {38488, 2, {0x0, 0x7BFC, 0x4800, 0x5124, 0x5248, 0x6490, 0x5248, 0x4924, 0x4800, 0x4BFC, 0x6840, 0x5040, 0x4040, 0x4040, 0x47FE, 0x4000}},
    {38489, 2, {0x0, 0x7DFE, 0x4500, 0x4900, 0x497C, 0x5100, 0x4900, 0x49FE, 0x4550, 0x4552, 0x4554, 0x6948, 0x5148, 0x4244, 0x4252, 0x4460}},
    {38490, 2, {0x10, 0x7818, 0x4BD4, 0x5014, 0x5010, 0x67FE, 0x5010, 0x4890, 0x4890, 0x4AD0, 0x6A90, 0x5290, 0x428A, 0x42EA, 0x4706, 0x4202}},
    {38491, 2, {0x220, 0x7A24, 0x4A28, 0x53B0, 0x5220, 0x6222, 0x52A2, 0x4B1E, 0x4840, 0x4840, 0x6BFC, 0x5040, 0x4040, 0x4040, 0x47FE, 0x4000}},
    {38492, 2, {0x20, 0x7C20, 0x4420, 0x4BFE, 0x4820, 0x5020, 0x4924, 0x4924, 0x4524, 0x46AA, 0x4420, 0x6850, 0x5050, 0x4088, 0x4104, 0x4202}},
    {38493, 2, {0x40, 0x7840, 0x4840, 0x57FE, 0x5040, 0x6450, 0x5248, 0x4A48, 0x4D54, 0x48A2, 0x68A0, 0x5110, 0x4110, 0x4208, 0x4404, 0x4802}},
    {38494, 2, {0x28, 0x7DC8, 0x4488, 0x4888, 0x4BFE, 0x5088, 0x4888, 0x4908, 0x4628, 0x4420, 0x45FC, 0x6820, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38495, 2, {0x20, 0x7820, 0x4920, 0x513C, 0x5120, 0x6120, 0x57FE, 0x4800, 0x4820, 0x4924, 0x6924, 0x5228, 0x4410, 0x4020, 0x40C0, 0x4700}},
    {38496, 2, {0x28, 0x7C24, 0x45FE, 0x4820, 0x4820, 0x51FC, 0x4924, 0x4924, 0x45FC, 0x4524, 0x4524, 0x69FC, 0x5124, 0x4124, 0x4124, 0x410C}},
    {38497, 2, {0x20, 0x7C20, 0x45FC, 0x4820, 0x4820, 0x5020, 0x4BFE, 0x4820, 0x4420, 0x4520, 0x453C, 0x6920, 0x52A0, 0x4260, 0x443E, 0x4800}},
    {38498, 2, {0x40, 0x7820, 0x4BFE, 0x5202, 0x5404, 0x61F8, 0x5000, 0x4800, 0x4BFE, 0x4890, 0x6890, 0x5090, 0x4112, 0x4112, 0x420E, 0x4400}},
    {38499, 2, {0x20, 0x7C20, 0x45FE, 0x4820, 0x49FC, 0x5124, 0x4924, 0x49FC, 0x4524, 0x4524, 0x45FC, 0x6820, 0x53FE, 0x4020, 0x4020, 0x4020}},
    {38500, 2, {0x0, 0x7840, 0x4840, 0x50A0, 0x5110, 0x6208, 0x55F6, 0x4840, 0x4840, 0x4BFC, 0x6840, 0x5150, 0x4248, 0x4444, 0x4144, 0x4080}},
    {38501, 2, {0x40, 0x7C40, 0x447C, 0x4884, 0x4908, 0x5210, 0x493E, 0x4922, 0x4522, 0x4522, 0x453E, 0x6922, 0x5122, 0x4122, 0x413E, 0x4122}},
    {38502, 2, {0x20, 0x7C20, 0x47FE, 0x4820, 0x49FC, 0x5040, 0x4BFE, 0x4848, 0x4488, 0x44FE, 0x4508, 0x6948, 0x5228, 0x4408, 0x4028, 0x4010}},
    {38503, 2, {0x0, 0x7DFC, 0x4504, 0x4904, 0x49FC, 0x5104, 0x4904, 0x49FC, 0x4420, 0x4420, 0x45FC, 0x6820, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38504, 2, {0x0, 0x7CF8, 0x4488, 0x4888, 0x48F8, 0x5000, 0x49FC, 0x4904, 0x4524, 0x4524, 0x4524, 0x6924, 0x5050, 0x4048, 0x4084, 0x4304}},
    {38505, 2, {0x40, 0x7840, 0x48A0, 0x50A0, 0x5110, 0x6208, 0x55F6, 0x4800, 0x4888, 0x4848, 0x6A48, 0x5150, 0x4110, 0x4020, 0x47FE, 0x4000}},
    {38506, 2, {0x40, 0x7C20, 0x45FC, 0x4800, 0x4908, 0x5090, 0x4800, 0x4BFE, 0x4400, 0x4400, 0x45FC, 0x6904, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {38507, 2, {0x90, 0x7C90, 0x4490, 0x4B9E, 0x4890, 0x5090, 0x4890, 0x4B9C, 0x4490, 0x4490, 0x4490, 0x6B9E, 0x5090, 0x4090, 0x4090, 0x4090}},
    {38508, 2, {0x0, 0x7FE0, 0x4A5E, 0x5252, 0x5252, 0x63D2, 0x5252, 0x4A54, 0x4BD4, 0x4A54, 0x6A48, 0x52E8, 0x4754, 0x4054, 0x4062, 0x4040}},
    {38509, 2, {0x20, 0x7C20, 0x45FC, 0x4850, 0x4888, 0x5104, 0x4BFE, 0x4808, 0x45E8, 0x4528, 0x4528, 0x69E8, 0x5128, 0x4008, 0x4028, 0x4010}},
    {38510, 2, {0xA0, 0x7C90, 0x4480, 0x49FE, 0x4910, 0x5310, 0x4DFC, 0x4910, 0x4510, 0x45FC, 0x4510, 0x6910, 0x5110, 0x41FE, 0x4100, 0x4100}},
    {38511, 2, {0x20, 0x7C20, 0x4450, 0x4888, 0x4904, 0x52FA, 0x4800, 0x49FC, 0x4554, 0x4554, 0x45FC, 0x6954, 0x5154, 0x4154, 0x4104, 0x410C}},
    {38512, 2, {0x20, 0x7C20, 0x4450, 0x4888, 0x4944, 0x5222, 0x49F8, 0x4808, 0x4410, 0x45FC, 0x4400, 0x6BFE, 0x5040, 0x4088, 0x41FC, 0x4084}},
    {38513, 2, {0x80, 0x7C80, 0x44FE, 0x4902, 0x4A42, 0x524A, 0x4952, 0x4842, 0x47FE, 0x4442, 0x44E2, 0x6952, 0x524A, 0x4042, 0x4054, 0x4008}},
    {38514, 2, {0x8, 0x783C, 0x49E0, 0x5020, 0x5020, 0x63FE, 0x5124, 0x4924, 0x4FFE, 0x4924, 0x6924, 0x53FE, 0x4020, 0x4020, 0x43FE, 0x4000}},
    {38515, 2, {0x40, 0x7840, 0x4FFC, 0x5040, 0x53F8, 0x6248, 0x53F8, 0x4A48, 0x4BF8, 0x4840, 0x68E0, 0x5150, 0x4248, 0x4C46, 0x4040, 0x4040}},
    {38516, 2, {0x40, 0x7880, 0x4BFC, 0x5224, 0x5224, 0x63FC, 0x5224, 0x4A44, 0x4BFC, 0x4890, 0x6910, 0x57FE, 0x4010, 0x4010, 0x4010, 0x4010}},
    {38517, 2, {0x20, 0x7820, 0x49FC, 0x5020, 0x5020, 0x63FE, 0x5088, 0x4944, 0x4A42, 0x48F8, 0x6988, 0x5250, 0x4020, 0x4050, 0x4188, 0x4606}},
    {38518, 2, {0x100, 0x7900, 0x49FC, 0x5204, 0x5504, 0x61E4, 0x5284, 0x4884, 0x4FF4, 0x4884, 0x6AA4, 0x52A4, 0x43E4, 0x4004, 0x4028, 0x4010}},
    {38519, 2, {0x80, 0x7880, 0x48F8, 0x5108, 0x5210, 0x6400, 0x5040, 0x4B9C, 0x4A04, 0x4A04, 0x6B9C, 0x5204, 0x4204, 0x4204, 0x43FC, 0x4204}},
    {38520, 2, {0x20, 0x7C20, 0x45FC, 0x4820, 0x4820, 0x53FE, 0x4888, 0x4904, 0x4622, 0x4420, 0x45FC, 0x6820, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38521, 2, {0x28, 0x7DC8, 0x4488, 0x4888, 0x4BFE, 0x5088, 0x4888, 0x4908, 0x4600, 0x45FC, 0x4504, 0x6904, 0x51FC, 0x4104, 0x4104, 0x41FC}},
    {38522, 2, {0x20, 0x7C20, 0x4450, 0x4888, 0x4904, 0x52FA, 0x4820, 0x4820, 0x45FC, 0x4524, 0x4524, 0x69FC, 0x5050, 0x4088, 0x4104, 0x4202}},
    {38523, 2, {0x0, 0x7BFE, 0x4850, 0x5050, 0x53FE, 0x6252, 0x5252, 0x4A52, 0x4BFE, 0x4820, 0x6820, 0x53FE, 0x4020, 0x4020, 0x47FE, 0x4000}},
    {38524, 2, {0x20, 0x7C22, 0x45FA, 0x4824, 0x4828, 0x53FE, 0x4820, 0x4840, 0x44FC, 0x4584, 0x4684, 0x68FC, 0x5084, 0x4084, 0x40FC, 0x4084}},
    {38525, 2, {0x0, 0x7CFC, 0x4484, 0x48FC, 0x4884, 0x50FC, 0x4800, 0x49FE, 0x4480, 0x44FE, 0x452A, 0x6A4A, 0x5092, 0x4122, 0x4054, 0x4088}},
    {38526, 2, {0x0, 0x7BFE, 0x4820, 0x5040, 0x51FC, 0x6154, 0x5154, 0x4954, 0x4954, 0x492C, 0x6820, 0x53FE, 0x4050, 0x4088, 0x4104, 0x4202}},
    {38527, 2, {0x0, 0x7DFE, 0x4510, 0x4910, 0x49FC, 0x5110, 0x4954, 0x4954, 0x45BA, 0x4510, 0x4528, 0x6924, 0x5142, 0x4180, 0x41FE, 0x4000}},
    {38528, 2, {0x8, 0x7C3C, 0x45E0, 0x4820, 0x4BFE, 0x5020, 0x49FC, 0x4924, 0x45FC, 0x4524, 0x45FC, 0x6820, 0x51FC, 0x4020, 0x43FE, 0x4000}},
    {38529, 2, {0x0, 0x7BFE, 0x4A00, 0x52FC, 0x5284, 0x62FC, 0x5284, 0x4AFC, 0x4A20, 0x4BFE, 0x6A48, 0x52C8, 0x4230, 0x4248, 0x4284, 0x43FE}},
    {38530, 2, {0x20, 0x7C50, 0x4488, 0x4904, 0x4AFE, 0x5080, 0x48F8, 0x4880, 0x44F8, 0x4480, 0x47FE, 0x6840, 0x5088, 0x4104, 0x43FE, 0x4002}},
    {38531, 2, {0x40, 0x78A0, 0x4910, 0x5208, 0x55F6, 0x6000, 0x53C4, 0x4A54, 0x4A54, 0x4BD4, 0x6A54, 0x5254, 0x43D4, 0x4244, 0x4254, 0x42C8}},
    {38532, 2, {0x0, 0x7DFC, 0x4504, 0x4904, 0x49FC, 0x5104, 0x4904, 0x49FC, 0x4400, 0x47FE, 0x4420, 0x6920, 0x513C, 0x4120, 0x42A0, 0x447E}},
    {38533, 2, {0x1FC, 0x7924, 0x4924, 0x49FC, 0x5124, 0x5124, 0x61FC, 0x5020, 0x4820, 0x4BFE, 0x4A22, 0x6A2A, 0x52FA, 0x420A, 0x4202, 0x4206}},
    {38534, 2, {0x80, 0x79F8, 0x4A08, 0x5510, 0x50E0, 0x6318, 0x5C06, 0x4BF8, 0x4840, 0x4A40, 0x6BFC, 0x5440, 0x43F8, 0x4040, 0x4FFE, 0x4000}},
    {38535, 2, {0xC, 0x780A, 0x4808, 0x4BFE, 0x5208, 0x53F8, 0x624A, 0x524A, 0x4BEA, 0x4AAC, 0x4AAC, 0x6A4C, 0x52AA, 0x430A, 0x4216, 0x4422}},
    {38536, 2, {0x0, 0x7BF8, 0x4A48, 0x4A48, 0x53F8, 0x5248, 0x6248, 0x53F8, 0x4800, 0x4FFE, 0x4A40, 0x6A44, 0x5228, 0x4290, 0x4308, 0x4206}},
    {38537, 2, {0x40, 0x7D9C, 0x4504, 0x4904, 0x49DC, 0x5104, 0x4904, 0x49FC, 0x4400, 0x4400, 0x45FC, 0x6820, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38538, 2, {0x104, 0x7888, 0x4800, 0x53FE, 0x5040, 0x60C2, 0x5122, 0x4A54, 0x4898, 0x4938, 0x6A54, 0x5094, 0x4112, 0x4610, 0x4050, 0x4020}},
    {38539, 2, {0x40, 0x7840, 0x4BFE, 0x5080, 0x51FC, 0x6220, 0x55FE, 0x4800, 0x49FC, 0x4904, 0x69FC, 0x5104, 0x41FC, 0x4104, 0x4114, 0x4108}},
    {38540, 2, {0x40, 0x7C20, 0x45FE, 0x4800, 0x4884, 0x5048, 0x4BFE, 0x4800, 0x44FC, 0x4484, 0x4484, 0x68FC, 0x5084, 0x4084, 0x40FC, 0x4084}},
    {38541, 2, {0x20, 0x7C40, 0x45FC, 0x4904, 0x49FC, 0x5104, 0x49FC, 0x4800, 0x47FE, 0x4420, 0x4420, 0x69FC, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38542, 2, {0x110, 0x7D12, 0x45D4, 0x4918, 0x4952, 0x5192, 0x492E, 0x4840, 0x45FC, 0x4504, 0x4504, 0x69FC, 0x5104, 0x4104, 0x41FC, 0x4104}},
    {38543, 2, {0x10, 0x7810, 0x49FE, 0x5420, 0x5240, 0x60FC, 0x5144, 0x4844, 0x4E7C, 0x4A44, 0x6A7C, 0x5244, 0x4254, 0x4248, 0x4500, 0x48FE}},
    {38544, 2, {0x80, 0x7880, 0x49F8, 0x4A08, 0x5410, 0x53FC, 0x6004, 0x51FC, 0x4804, 0x4BFC, 0x4804, 0x6840, 0x5124, 0x452A, 0x450A, 0x48F8}},
    {38545, 2, {0x20, 0x7D24, 0x4524, 0x4924, 0x49FC, 0x5000, 0x4BFE, 0x4800, 0x45FC, 0x4504, 0x4504, 0x69FC, 0x5088, 0x4050, 0x43FE, 0x4000}},
    {38546, 2, {0x88, 0x7C50, 0x4400, 0x49FE, 0x4850, 0x51FC, 0x4854, 0x4BFE, 0x4454, 0x45FC, 0x4450, 0x68D8, 0x5154, 0x4252, 0x4050, 0x4050}},
    {38547, 2, {0x20, 0x7C20, 0x45FE, 0x4840, 0x48FC, 0x5110, 0x4AFE, 0x4800, 0x4420, 0x45FE, 0x4440, 0x68FC, 0x5110, 0x4210, 0x41FE, 0x4000}},
    {38548, 2, {0x0, 0x7BFE, 0x4800, 0x51FC, 0x5104, 0x6104, 0x51FC, 0x4800, 0x4BFE, 0x4A8A, 0x6A52, 0x53FE, 0x4222, 0x4222, 0x422A, 0x4204}},
    {38549, 2, {0x0, 0x7CFC, 0x4484, 0x4884, 0x48FC, 0x5000, 0x49FE, 0x4902, 0x45FE, 0x4502, 0x45FE, 0x6902, 0x51FE, 0x4048, 0x4084, 0x4102}},
    {38550, 2, {0x40, 0x7C80, 0x45FC, 0x4904, 0x4904, 0x51FC, 0x4900, 0x49FE, 0x4500, 0x45FE, 0x4402, 0x6AAA, 0x52AA, 0x4402, 0x4014, 0x4008}},
    {38551, 2, {0x20, 0x7840, 0x4BFC, 0x4A24, 0x5224, 0x53FC, 0x6224, 0x5244, 0x4BFC, 0x4840, 0x48A8, 0x68B4, 0x513C, 0x4122, 0x4222, 0x441E}},
    {38552, 2, {0x208, 0x7908, 0x4910, 0x4800, 0x57FC, 0x5000, 0x6110, 0x5208, 0x4C04, 0x4BF8, 0x4AA8, 0x6AA8, 0x52A8, 0x42A8, 0x4FFE, 0x4000}},
    {38553, 2, {0x20, 0x7CA8, 0x4524, 0x4A22, 0x4800, 0x51FC, 0x4904, 0x49FC, 0x4504, 0x45FC, 0x4420, 0x68A8, 0x5124, 0x4222, 0x40A0, 0x4040}},
    {38554, 2, {0x20, 0x7D24, 0x44A8, 0x4BFE, 0x4A02, 0x50F8, 0x4888, 0x4888, 0x44F8, 0x4420, 0x4420, 0x69FC, 0x5020, 0x4020, 0x43FE, 0x4000}},
    {38555, 2, {0x100, 0x7910, 0x4BDC, 0x5254, 0x5554, 0x6288, 0x5108, 0x4AF4, 0x4C02, 0x4800, 0x6BFC, 0x5040, 0x4150, 0x4248, 0x4544, 0x4080}},
    {38556, 2, {0x20, 0x7DFC, 0x4488, 0x4850, 0x4BFE, 0x5000, 0x49FC, 0x4904, 0x45FC, 0x4504, 0x45FC, 0x6820, 0x53FE, 0x4020, 0x4020, 0x4020}},
    {38557, 2, {0x40, 0x7C80, 0x45FC, 0x4904, 0x49FC, 0x5104, 0x49FC, 0x4900, 0x45FE, 0x4500, 0x45FE, 0x6802, 0x52AA, 0x42AA, 0x4202, 0x400C}},
    {38558, 2, {0x108, 0x7908, 0x4908, 0x57D0, 0x511E, 0x6794, 0x5124, 0x4FD4, 0x4A14, 0x4BD4, 0x6A54, 0x5248, 0x4248, 0x44D4, 0x4424, 0x4842}},
    {38559, 2, {0x124, 0x7A48, 0x4924, 0x5000, 0x51FC, 0x6124, 0x51FC, 0x4924, 0x49FC, 0x4820, 0x6BFE, 0x5070, 0x40A8, 0x4124, 0x4622, 0x4020}},
    {38560, 2, {0x8, 0x7C3C, 0x45C0, 0x4844, 0x4924, 0x50A8, 0x49FC, 0x4804, 0x44FC, 0x4404, 0x45FC, 0x6840, 0x50A4, 0x42AA, 0x428A, 0x4478}},
    {38561, 2, {0x40, 0x7820, 0x4BFC, 0x5000, 0x5108, 0x6090, 0x53FE, 0x4A20, 0x4AA0, 0x4AFC, 0x6B20, 0x5220, 0x42FC, 0x4420, 0x4420, 0x49FE}},
    {38562, 2, {0x20, 0x7C20, 0x45FC, 0x4820, 0x4BFE, 0x5108, 0x4B9C, 0x4908, 0x4588, 0x463E, 0x4400, 0x6BFE, 0x5090, 0x4090, 0x4112, 0x420E}},
    {38563, 2, {0x20, 0x7924, 0x48A8, 0x53FE, 0x50A8, 0x6124, 0x5222, 0x4904, 0x4904, 0x49DE, 0x6A44, 0x5554, 0x409E, 0x4104, 0x4204, 0x4404}},
    {38564, 2, {0x20, 0x7DFC, 0x4524, 0x49FC, 0x4820, 0x53FE, 0x4800, 0x49FC, 0x4504, 0x45FC, 0x4504, 0x69FC, 0x5104, 0x41FC, 0x4088, 0x4104}},
    {38565, 2, {0x10, 0x7BD2, 0x4854, 0x5148, 0x508A, 0x6104, 0x52FA, 0x4C00, 0x49FC, 0x4904, 0x6904, 0x51FC, 0x4104, 0x4088, 0x47FE, 0x4000}},
    {38566, 2, {0x10, 0x7BC8, 0x4A7E, 0x5240, 0x5262, 0x63D4, 0x5200, 0x4A3E, 0x4BC8, 0x4A48, 0x6E7E, 0x5248, 0x4248, 0x43C8, 0x4248, 0x4008}},
    {38567, 2, {0x88, 0x7C48, 0x4A50, 0x50FC, 0x5020, 0x6050, 0x5694, 0x4A38, 0x4A50, 0x4A98, 0x6A34, 0x5254, 0x4290, 0x4220, 0x45FE, 0x4800}},
    {38568, 2, {0x20, 0x79FE, 0x4C40, 0x4AFC, 0x5110, 0x52FE, 0x6000, 0x56FC, 0x4A84, 0x4AFC, 0x4A84, 0x6AFC, 0x5284, 0x428C, 0x4500, 0x48FE}},
    {38569, 2, {0x40, 0x7880, 0x4BFE, 0x5222, 0x52AA, 0x6222, 0x53FE, 0x4A72, 0x4AAA, 0x4A02, 0x6820, 0x53FE, 0x4050, 0x4088, 0x4104, 0x4602}},
    {38570, 2, {0x20, 0x7C20, 0x4450, 0x4888, 0x4904, 0x52FA, 0x4800, 0x4800, 0x45DC, 0x4554, 0x4554, 0x69DC, 0x5088, 0x4088, 0x4154, 0x4222}},
    {38571, 2, {0x20, 0x7DFC, 0x4420, 0x4888, 0x4BFE, 0x5088, 0x49FC, 0x4904, 0x45FC, 0x4504, 0x45FC, 0x6904, 0x51FC, 0x4088, 0x4104, 0x4202}},
    {38572, 2, {0x0, 0x7BFE, 0x4924, 0x4A22, 0x57FE, 0x5222, 0x6376, 0x52AA, 0x4B76, 0x4A22, 0x4B76, 0x6AAA, 0x5376, 0x4222, 0x422A, 0x4204}},
    {38573, 2, {0x0, 0x79FC, 0x4820, 0x4BFE, 0x5222, 0x51AC, 0x6020, 0x51AC, 0x4800, 0x4BFE, 0x4820, 0x69FC, 0x5154, 0x4154, 0x4154, 0x410C}},
    {38574, 2, {0x40, 0x7820, 0x4BFE, 0x4888, 0x5052, 0x57AC, 0x62AA, 0x52A8, 0x4DAC, 0x4800, 0x4904, 0x69FC, 0x5104, 0x41FC, 0x4104, 0x4204}},
    {38575, 2, {0x20, 0x7BFE, 0x4820, 0x49FC, 0x5000, 0x53FE, 0x6002, 0x51FC, 0x4820, 0x4BFE, 0x4800, 0x6BFE, 0x5004, 0x43BE, 0x42A4, 0x438C}},
    {38576, 2, {0x0, 0x7BFE, 0x4A02, 0x4BFE, 0x5202, 0x53FE, 0x6108, 0x5252, 0x4B9C, 0x4908, 0x4A52, 0x6BDE, 0x5000, 0x42A4, 0x4252, 0x4452}},
    {38577, 2, {0x3E, 0x7FC0, 0x4A44, 0x4928, 0x53F8, 0x5040, 0x67FC, 0x5000, 0x4BF8, 0x4808, 0x4BF8, 0x6808, 0x53F8, 0x4544, 0x452A, 0x48FA}},
    {38578, 2, {0x20, 0x7CA8, 0x4524, 0x4A2A, 0x4830, 0x50C0, 0x4BFE, 0x4920, 0x45FC, 0x4520, 0x45FC, 0x6920, 0x51FE, 0x4002, 0x42AA, 0x4004}},
    {38579, 2, {0x40, 0x7BFC, 0x4880, 0x51F8, 0x4A20, 0x4DFC, 0x6888, 0x50F8, 0x4188, 0x46C0, 0x1930, 0xE10E, 0x1190, 0x2148, 0x4524, 0x200}},
    {38580, 2, {0x110, 0x7890, 0x4BDE, 0x4810, 0x525E, 0x5182, 0x63DE, 0x5010, 0x4BDE, 0x4A50, 0x4BDE, 0x6A50, 0x53DE, 0x4250, 0x4252, 0x42CE}},
    {38581, 2, {0x108, 0xF1EC, 0x910A, 0xA7E8, 0xA528, 0xC5DE, 0xA728, 0x94EA, 0x940A, 0x97EA, 0xD40C, 0xA7EC, 0x8A2A, 0x8BEA, 0x9156, 0x87E2}},
    {38582, 2, {0x100, 0x100, 0x3FF8, 0x108, 0x108, 0xFFFE, 0x108, 0x108, 0x3FF8, 0x2108, 0x1190, 0x560, 0x920, 0x3118, 0xC506, 0x200}},
    {38583, 2, {0x1020, 0x1020, 0xFCFC, 0x1024, 0x11FE, 0x7C24, 0xFC, 0x7C20, 0x124, 0xFCA8, 0x1070, 0x54A8, 0x5324, 0x9222, 0x10A0, 0x3040}},
    {38584, 2, {0x1020, 0x1020, 0xFEFC, 0x3824, 0x55FE, 0x9224, 0x10FC, 0x7C20, 0x124, 0xFEA8, 0x1070, 0x54A8, 0x5324, 0x9222, 0x50A0, 0x2040}},
    {38585, 2, {0x900, 0x880, 0x880, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1080, 0x1FFC, 0x1000, 0x1000}},
    {38586, 2, {0x200, 0x200, 0x7FFC, 0x4504, 0x880, 0x1FFC, 0x3080, 0x5080, 0x9FF8, 0x1080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000}},
    {38587, 2, {0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {38588, 2, {0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1100, 0x100, 0xFFFE, 0x100, 0x100, 0x100, 0x100}},
    {38589, 2, {0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3FE0, 0x420, 0x47C, 0x804, 0x3014, 0xC008}},
    {38590, 2, {0xA0, 0x90, 0xFC80, 0x5FE, 0x510, 0x4B10, 0x2DFC, 0x2910, 0x1110, 0x11FC, 0x2910, 0x2510, 0x4510, 0x81FE, 0x100, 0x100}},
    {38591, 2, {0x2850, 0x2448, 0x2480, 0x20FE, 0x3D90, 0xE290, 0x20FC, 0x2090, 0x2090, 0x20FC, 0x2090, 0x2090, 0x1490, 0x14FE, 0xC80, 0x480}},
    {38592, 2, {0x100, 0x1110, 0x1108, 0x2124, 0x40C0, 0x300, 0xC80, 0x3040, 0xDFFC, 0x1080, 0x1FF8, 0x1080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {38593, 2, {0x0, 0x3FFE, 0x2000, 0x2250, 0x2248, 0x24FE, 0x2490, 0x2D90, 0x36FE, 0x2490, 0x2490, 0x24FE, 0x2490, 0x4490, 0x44FE, 0x8480}},
    {38594, 2, {0x850, 0x848, 0x1480, 0x14FE, 0x2390, 0x5290, 0x88FC, 0x890, 0x90, 0x7EFC, 0x290, 0x490, 0x490, 0x8FE, 0x880, 0x80}},
    {38595, 2, {0x50, 0x7E48, 0x2480, 0x24FE, 0x2590, 0x2690, 0x24FC, 0xFF90, 0x2490, 0x24FC, 0x2490, 0x2490, 0x2490, 0x24FE, 0x4480, 0x8480}},
    {38596, 2, {0x850, 0x848, 0x880, 0x7EFE, 0x1190, 0x1290, 0x20FC, 0x2890, 0x4890, 0x48FC, 0x9090, 0x1490, 0x2290, 0x7EFE, 0x2280, 0x80}},
    {38597, 2, {0x50, 0x7E48, 0x480, 0x24FE, 0x2590, 0x2690, 0x44FC, 0x7F90, 0xC90, 0x14FC, 0x1490, 0x2490, 0x4490, 0x84FE, 0x1480, 0x880}},
    {38598, 2, {0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1100, 0xFFFE, 0x540, 0x920, 0x3118, 0xC106, 0x100}},
    {38599, 2, {0x100, 0x80, 0x3FFC, 0x2004, 0x3FFC, 0x2280, 0x2240, 0x27FC, 0x2C40, 0x37F8, 0x2440, 0x27F8, 0x4440, 0x4440, 0x87FC, 0x400}},
    {38600, 2, {0x440, 0xFC7E, 0x440, 0x100, 0x880, 0x1FFC, 0x1080, 0x3080, 0x5FF8, 0x9080, 0x1080, 0x1FF8, 0x1080, 0x1080, 0x1FFC, 0x1000}},
    {38601, 2, {0x20A0, 0x2090, 0x2080, 0x7DFE, 0x5110, 0x9310, 0x15FC, 0xFD10, 0x1110, 0x11FC, 0x1110, 0x2910, 0x2510, 0x45FE, 0x4100, 0x8100}},
    {38602, 2, {0x2050, 0x2048, 0x2080, 0x7EFE, 0x4390, 0x8290, 0x7AFC, 0x4A90, 0x4A90, 0x4AFC, 0x4A90, 0x7A90, 0x4A90, 0x2FE, 0x1480, 0x880}},
    {38603, 2, {0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000, 0x3E7C, 0x2244, 0x2244, 0x23C4, 0x2014, 0x2008}},
    {38604, 2, {0x1450, 0x1448, 0x1480, 0x54FE, 0x5590, 0x5E90, 0x54FC, 0x5490, 0x5490, 0x54FC, 0x5490, 0x5490, 0x5D90, 0x76FE, 0xC480, 0x80}},
    {38605, 2, {0x200, 0x100, 0xFFFE, 0x1050, 0x1048, 0x24FE, 0x4590, 0x7A90, 0x10FC, 0x2490, 0x4490, 0x78FC, 0x890, 0x1090, 0x20FE, 0xC080}},
    {38606, 2, {0x50, 0x7C48, 0x4480, 0x44FE, 0x4590, 0x7E90, 0x44FC, 0x4490, 0x7C90, 0x44FC, 0x4490, 0x4490, 0x4E90, 0xF0FE, 0x80, 0x80}},
    {38607, 2, {0x2050, 0x2048, 0x3E80, 0x42FE, 0x8590, 0xA90, 0xFEFC, 0x290, 0x290, 0x7EFC, 0x290, 0x290, 0x290, 0xFEFE, 0x280, 0x80}},
    {38608, 2, {0x100, 0x1F8, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x22FC, 0x2240, 0x27FE, 0x2C40, 0x37FC, 0x2440, 0x47FC, 0x4440, 0x87FE}},
    {38609, 2, {0x2050, 0x2048, 0xF880, 0x28FE, 0x2990, 0x2A90, 0x46FC, 0x9090, 0x1090, 0xFEFC, 0x1090, 0x3490, 0x5290, 0x92FE, 0x5080, 0x2080}},
    {38610, 2, {0x2050, 0x2048, 0x3C80, 0x44FE, 0xC590, 0x2A90, 0x10FC, 0x2890, 0x4490, 0x82FC, 0x7C90, 0x4490, 0x4490, 0x44FE, 0x7C80, 0x4480}},
    {38611, 2, {0x1050, 0x1048, 0x2880, 0x44FE, 0x8390, 0x7E90, 0x10FC, 0x1090, 0xFE90, 0x10FC, 0x5490, 0x5290, 0x9290, 0x10FE, 0x5080, 0x2080}},
    {38612, 2, {0x1428, 0x1224, 0x2040, 0x3F7E, 0x64C8, 0xA548, 0x3F7E, 0x2448, 0x2448, 0x3F7E, 0x2448, 0x2448, 0x2448, 0x3F7E, 0x2040, 0x2040}},
    {38613, 2, {0x28, 0x7F24, 0x4940, 0x497E, 0x5DC8, 0x4948, 0x7F7E, 0x4148, 0x5D48, 0x557E, 0x5548, 0x5D48, 0x4148, 0x417E, 0x4540, 0x8240}},
    {38614, 2, {0x50, 0x7C48, 0x4480, 0x44FE, 0x7D90, 0x1290, 0x10FC, 0xFE90, 0x9290, 0x92FC, 0xFE90, 0x1090, 0x1490, 0x12FE, 0xFE80, 0x280}},
    {38615, 2, {0x1040, 0x1040, 0x107E, 0xFE80, 0x1150, 0x7C48, 0x44FE, 0x7D90, 0x4690, 0x7CFC, 0x1090, 0xFEFC, 0x1090, 0x1090, 0x10FE, 0x1080}},
    {38616, 2, {0x44, 0x3DFE, 0x2454, 0x2448, 0x34FE, 0x2D90, 0x24FE, 0x2490, 0xFEFE, 0x2490, 0x24FE, 0x2480, 0x25FE, 0x2444, 0x4438, 0x8DC6}},
    {38617, 2, {0x810, 0x2448, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x2448, 0x3F7E, 0x2040, 0x1FF0, 0x820, 0x440, 0x380, 0x1C70, 0xE00E}},
    {38618, 2, {0x820, 0xFFFE, 0x820, 0x3EF8, 0x2288, 0x3EF8, 0x900, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {38619, 2, {0x2028, 0x2024, 0x7F40, 0x917E, 0x55C8, 0x7D48, 0x217E, 0x4A48, 0xA448, 0x7F7E, 0x9148, 0x5548, 0x7D48, 0x217E, 0x4A40, 0x8440}},
    {38620, 2, {0x1050, 0x848, 0xFE80, 0xFE, 0x2590, 0x2690, 0x5AFC, 0x9190, 0x1090, 0xFEFC, 0x1090, 0x5490, 0x9290, 0x12FE, 0x5080, 0x2080}},
    {38621, 2, {0x4928, 0x9224, 0x4940, 0x7E, 0x7CC8, 0x4548, 0x7C7E, 0x48, 0xFE48, 0x927E, 0x9248, 0xFE48, 0x8148, 0x817E, 0x7F40, 0x40}},
    {38622, 2, {0xE50, 0xF048, 0x2280, 0x92FE, 0x4590, 0x2290, 0x44FC, 0xF890, 0x1090, 0x24FC, 0xFE90, 0x1090, 0xFE90, 0x28FE, 0x4480, 0x8280}},
    {38623, 2, {0x100, 0x4204, 0x7FFC, 0x880, 0x3440, 0xCFFC, 0x1880, 0x2FF8, 0x4880, 0xFFE, 0x800, 0x3FFC, 0x2424, 0x2BD4, 0x2244, 0x23CC}},
    {38624, 2, {0x3018, 0x2A14, 0x4120, 0x7D3E, 0xD068, 0x5328, 0x7D3E, 0x5128, 0x5128, 0x7D3E, 0x5128, 0x5168, 0x51A8, 0x7D3E, 0x4020, 0x4020}},
    {38625, 2, {0x28, 0xEE24, 0x2240, 0xAA7E, 0x66C8, 0xAB48, 0x187E, 0x6648, 0x8948, 0x307E, 0xC448, 0x1948, 0xE248, 0xC7E, 0x3040, 0xC040}},
    {38626, 2, {0x1028, 0x824, 0x7F40, 0x147E, 0x49C8, 0x5548, 0x417E, 0x7F48, 0x848, 0x7F7E, 0x5148, 0x5548, 0x5D48, 0x417E, 0x4540, 0x4240}},
    {38627, 2, {0x2228, 0x2224, 0xFF40, 0x227E, 0x3EC8, 0x948, 0x7F7E, 0x4948, 0x7F48, 0x87E, 0x7F48, 0x848, 0xFF48, 0x147E, 0x2240, 0x4140}},
    {38628, 2, {0x3AB8, 0x2108, 0x3AB8, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4882, 0x9FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {38629, 2, {0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x2448, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x2448, 0x3F7E}},
    {38630, 2, {0x318C, 0x294A, 0x2210, 0x7BDE, 0x56B4, 0xD294, 0x7BDE, 0x5294, 0x5294, 0x7BDE, 0x5294, 0x5294, 0x5294, 0x7BDE, 0x4210, 0x4210}},
    {38631, 2, {0x880, 0x1FFC, 0x3080, 0xDFF8, 0x1080, 0x1FFC, 0x1428, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x100, 0xFFFE, 0x920, 0x711C}},
    {38632, 2, {0x0, 0xFFFE, 0x100, 0x100, 0x100, 0x7FFC, 0x4104, 0x4104, 0x4944, 0x4524, 0x4104, 0x4944, 0x4524, 0x4104, 0x4114, 0x4008}},
    {38633, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x400, 0xFF0, 0x10, 0xA0, 0x40}},
    {38634, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x8, 0x8, 0x1FF8, 0x8, 0x8, 0x3FF8, 0x8}},
    {38635, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0xFFFE, 0x200, 0x240, 0x220, 0x210, 0x200, 0x200}},
    {38636, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x100, 0x7FFC, 0x380, 0x560, 0x1918, 0xE106, 0x100}},
    {38637, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FE0, 0x840, 0xCF8, 0x1210, 0x1120, 0x20C0, 0x4330, 0x9C0E}},
    {38638, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE}},
    {38639, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x440, 0x1830, 0xE00E}},
    {38640, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x440, 0x820, 0x3018, 0xDFE6, 0x420, 0x420, 0x820, 0x10A0, 0x2040}},
    {38641, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x200, 0x100, 0x7FFC, 0x400, 0x7F0, 0x410, 0x810, 0x10A0, 0x6040}},
    {38642, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x820, 0x1010, 0x3FF8, 0x8}},
    {38643, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FFC, 0x2000, 0x2100, 0x2FF8, 0x2108, 0x4208, 0x4450, 0x9820}},
    {38644, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0}},
    {38645, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x100, 0x1FF0, 0x1110, 0x1110, 0xFFFE, 0x280, 0x1C70, 0xE00E}},
    {38646, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x280, 0xC60, 0x3218, 0xC106, 0x1FE0, 0x20, 0x640, 0x180, 0x40}},
    {38647, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8}},
    {38648, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x820, 0x7FFC, 0x820, 0xFE0, 0x820, 0xFE0, 0x820}},
    {38649, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1FD0, 0x1020, 0x1004, 0xFFC}},
    {38650, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x1FF0, 0x220, 0x140, 0x7FFC, 0x484, 0x1888, 0x6280, 0x100}},
    {38651, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x102, 0x102, 0xFE}},
    {38652, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x7FFC, 0x400, 0x800, 0x1FF8, 0x2808, 0xC808, 0xFF8, 0x808}},
    {38653, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0xF8, 0x7F00, 0x1110, 0x920, 0xFFFE, 0x100, 0x500, 0x200}},
    {38654, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x400, 0xFE0, 0x1440, 0x380, 0x1C70, 0xE20E, 0xFE0, 0x420, 0x1860}},
    {38655, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x2448, 0x1450, 0xC60, 0x1450, 0x6448, 0x842, 0x1042, 0x603E}},
    {38656, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0xFFFE, 0x200, 0x3FF8, 0x2488, 0x2488, 0x24A8, 0x2010}},
    {38657, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x80, 0x7FFC, 0x820, 0x7C0, 0x1830, 0xE44E, 0x440, 0x840, 0x1040}},
    {38658, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2040, 0x1040, 0x87FE, 0x50E0, 0x2150, 0xE248, 0x2444, 0x2842}},
    {38659, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2040, 0x13FC, 0x8244, 0x54A8, 0x10A0, 0xE122, 0x2222, 0x241E, 0x2800}},
    {38660, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x1110, 0x920, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x2018}},
    {38661, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x100, 0x7FFC, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8}},
    {38662, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x8, 0x7C3C, 0x9E0, 0x1020, 0x3DFC, 0x4420, 0x2BFE, 0x1000, 0x6FFE}},
    {38663, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3FFC, 0x2000, 0x2FF8, 0x2000, 0x3FFC, 0x2448, 0x4430, 0x860E}},
    {38664, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2040, 0x17FC, 0x4040, 0x2BF8, 0xA48, 0x7248, 0x1258, 0x1040}},
    {38665, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x1000, 0x1FFC, 0x2000, 0x5FF0, 0x1210, 0xFFFE, 0x2090, 0x3FFC, 0x30}},
    {38666, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x2448, 0x1450, 0xFFFE, 0x0}},
    {38667, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x200, 0x1FF0, 0x210, 0x7FFC, 0x210, 0x1FF0, 0x400, 0x7FFC, 0x820, 0x7C0, 0x7838}},
    {38668, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3FF8, 0x2108, 0x27C8, 0x2108, 0x2FE8, 0x2008, 0x27C8, 0x4448, 0x47C8, 0x8018}},
    {38669, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x880, 0x1FFC, 0x3080, 0x5FF8, 0x9080, 0x1FF8, 0x1080, 0x1FFC, 0x1000}},
    {38670, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x200, 0x3FF8, 0x820, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x7C0, 0x7838}},
    {38671, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440}},
    {38672, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x2788, 0x1088, 0x8788, 0x5410, 0x2790, 0xE0A4, 0x22BE, 0x2102}},
    {38673, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2000, 0x1040, 0x807C, 0x4840, 0xBF8, 0x1208, 0xE208, 0x23F8, 0x2208}},
    {38674, 2, {0x10, 0x7F10, 0x810, 0xFFA8, 0x88A8, 0x6B44, 0x8A2, 0x6B10, 0x10, 0x7E7C, 0x4, 0xFF04, 0x1008, 0x2208, 0x7F10, 0x110}},
    {38675, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x600, 0x38F8, 0x2008, 0x3CF8, 0x2008, 0x3FF8, 0x842, 0x1042, 0xE03E}},
    {38676, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2080, 0x1040, 0x87FC, 0x4040, 0x1040, 0x23F8, 0xE040, 0x2040, 0x27FC}},
    {38677, 2, {0x10, 0x7F10, 0x87E, 0xFF90, 0x8890, 0x6B54, 0x854, 0x6B54, 0x54, 0x7F7C, 0x14, 0xFF90, 0x1012, 0x2212, 0x7F12, 0x210E}},
    {38678, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x820, 0x820, 0x7EFC, 0x830, 0x1C68, 0x2AA4, 0xC922, 0x820, 0x820}},
    {38679, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2040, 0x10A0, 0x158, 0x4626, 0x29F0, 0x810, 0x70A0, 0x1040, 0x1020}},
    {38680, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2000, 0x13FC, 0x204, 0x42F4, 0x2A04, 0xAF4, 0x7294, 0x12F4, 0x120C}},
    {38681, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x820, 0xFFFE, 0x920, 0x1FF0, 0x1110, 0xFFFE, 0x280, 0x1C70, 0xE00E}},
    {38682, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x98C6}},
    {38683, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x7BDE, 0x842, 0x7BDE, 0x4210, 0x7BDE, 0x842, 0x294A, 0x1084}},
    {38684, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x8FC, 0x7E84, 0x18FC, 0x2C84, 0x4AFC, 0x8884, 0x8FC, 0x884}},
    {38685, 2, {0x0, 0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x7BBC, 0x4AA4, 0x4AA4, 0x4AA4, 0x4AA4, 0x7BBC, 0x4AA4}},
    {38686, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3EF8, 0x2208, 0x3EF8, 0x2000, 0x3EF8, 0x2048, 0x3E30, 0x20CC}},
    {38687, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x1090, 0x890, 0x43FC, 0x2090, 0x890, 0x17FE, 0x7090, 0x1108, 0x1204}},
    {38688, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2020, 0x1050, 0xFC88, 0x4944, 0x4A22, 0x10F8, 0x1E08, 0xF010, 0x4020}},
    {38689, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x20, 0x3C10, 0x24E0, 0x3C22, 0x25B4, 0x3CA8, 0x2524, 0x26A2, 0x4C40}},
    {38690, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0xC, 0x7DF0, 0x4508, 0x7D70, 0x4544, 0x7D68, 0x4550, 0x5548, 0x8A66}},
    {38691, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0xFE0, 0x820, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x820, 0x1010}},
    {38692, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x400, 0x79FC, 0x4044, 0x4844, 0x5454, 0x6288, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8}},
    {38693, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x7FFE, 0x4002, 0x9FF4, 0x600, 0x3B08, 0x4B0, 0x19C0, 0x62A0, 0xC98, 0x7306}},
    {38694, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2204, 0x2208, 0xF7B0, 0x2204, 0x7708, 0xAAB2, 0x2244, 0x2218, 0x2260}},
    {38695, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x40, 0x7EFC, 0x1544, 0x838, 0xFE44, 0x1AA2, 0x2CFC, 0x4844, 0xA894, 0x1108}},
    {38696, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3F08, 0x2108, 0x3F7E, 0x2008, 0x2F48, 0x2028, 0x5F88, 0x4408, 0x9528, 0x2C90}},
    {38697, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x87C, 0x7F44, 0x2248, 0x3E50, 0x48, 0x3E44, 0x464, 0x7F58, 0x840, 0x1840}},
    {38698, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x203C, 0x17C0, 0x8244, 0x4128, 0x17FC, 0x2040, 0xEFFE, 0x2040, 0x27FC}},
    {38699, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x7EFC, 0x2244, 0x1A34, 0x62C4, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {38700, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010}},
    {38701, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x23F8, 0x1208, 0x3F8, 0x7208, 0x13F8, 0x1100, 0x13FC, 0x1444, 0x1AA4, 0x13F4, 0xC}},
    {38702, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x2110, 0x17FC, 0x9110, 0x41F0, 0x4110, 0x11F0, 0x2110, 0xE7FC, 0x2250, 0x2288, 0x23FC}},
    {38703, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x90, 0x7EA0, 0x444, 0x2BA8, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x440, 0x7FFC}},
    {38704, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x1220, 0x3F20, 0x123E, 0x7F44, 0xA4, 0x3F24, 0x2128, 0x3F10, 0x2128, 0x2346}},
    {38705, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x9FF4, 0x220, 0x7FFC, 0x488, 0x7A80, 0x100, 0x3FF8, 0x2828, 0x37D8, 0x2448, 0x27C8, 0x2018}},
    {38706, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x40, 0x7CF8, 0x4548, 0x7C30, 0x11CE, 0x5C00, 0x50F8, 0x5C88, 0xE0F8}},
    {38707, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x80, 0x79F8, 0x4A88, 0x4870, 0x538E, 0x4920, 0x49FC, 0x6A20, 0x50F8, 0x4020, 0x43FE}},
    {38708, 2, {0x20, 0x7C20, 0x10FC, 0xFE24, 0x93FE, 0x5424, 0x10FC, 0x5420, 0x124, 0x7CA8, 0x70, 0xFEA8, 0x2124, 0x4822, 0xFCA0, 0x440}},
    {38709, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x3E28, 0x2224, 0x3E24, 0x20, 0xFFFE, 0x2224, 0x3E28, 0x2392, 0xFE2A, 0x246}},
    {38710, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x80, 0x47FC, 0x2110, 0xFFE, 0x8842, 0x57FC, 0x1100, 0xE1F8, 0x2208, 0x2228, 0x2410}},
    {38711, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x1F8, 0x7908, 0x49F8, 0x4908, 0x4BFE, 0x7900, 0x4BFC, 0x4CA4, 0x4924, 0x7A54, 0x4888}},
    {38712, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x2200, 0x7F7C, 0x2244, 0x3E44, 0x87C, 0x7F44, 0x4944, 0x7F7C, 0x844, 0xFF54, 0x888}},
    {38713, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x10, 0x7EFE, 0x4244, 0x7E28, 0x40FE, 0x7E10, 0x627C, 0xBE10, 0x2210}},
    {38714, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x2010, 0x4550, 0x97DE, 0x2024, 0x6FD4, 0xA014, 0x2788, 0x24D4, 0x28A2}},
    {38715, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x7FFC, 0x1010, 0x1FF0, 0x820, 0xFFFE, 0x0}},
    {38716, 2, {0x80, 0x7C80, 0x10FC, 0xFF00, 0x92F8, 0x5400, 0x11F8, 0x5448, 0x168, 0x7CC8, 0x1F8, 0xFE4A, 0x20EA, 0x495A, 0xFC46, 0x442}},
    {38717, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8514, 0xFFFE, 0x288, 0x7D70, 0x2548, 0x4D66, 0x9010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x2010}},
    {38718, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x3000, 0xCBFC, 0x2D24, 0x31FC, 0xC924, 0x15FC, 0x6420, 0xDFC, 0x3420, 0xC5FE, 0x1800}},
    {38719, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x80, 0x7DF8, 0x688, 0x2870, 0x1088, 0xFD06, 0x34F8, 0x5088, 0x90F8, 0x5088, 0x20F8}},
    {38720, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x820, 0x7FFC, 0x820, 0x7FFE, 0x4002, 0x9FF4, 0x600, 0x1920, 0x6C0, 0x38A0, 0x318}},
    {38721, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFEFE, 0x9292, 0xFEFE, 0x9292, 0xFEFE}},
    {38722, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x7FFE, 0x4000, 0x5F3C, 0x4410, 0x7F7E, 0x4E34, 0x5552, 0x6090, 0x44F8, 0x4480, 0xBFFE}},
    {38723, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x810, 0x2448, 0x3F7E, 0x64C8, 0xBF7E, 0x2448, 0x3F7E, 0x2448, 0x3F7E}},
    {38724, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0xA414, 0x11FC, 0xFD04, 0x1FC, 0x7904, 0x1FC, 0x7880, 0x1FE, 0x7A22, 0x4D52, 0x79FA, 0x4806}},
    {38725, 2, {0x28, 0x7CFE, 0x102A, 0xFEFE, 0x92A8, 0x54FE, 0x104A, 0x54FE, 0x82, 0x7CFE, 0x82, 0xFEFE, 0x2082, 0x48FE, 0xFC44, 0x482}},
    {38726, 2, {0x10, 0x7E10, 0x117C, 0xFE14, 0x92FE, 0x5414, 0x137C, 0x5592, 0x154, 0x7D38, 0x154, 0xFF92, 0x2130, 0x4900, 0xFEFE, 0x400}},
    {38727, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E}},
    {38728, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x100, 0x1D70, 0x0, 0x3BB8, 0x2AA8, 0x3BB8, 0x0, 0x7FFC, 0x1110, 0x2928, 0xFFFE}},
    {38729, 2, {0xE, 0x7DF0, 0x10A2, 0xFE54, 0x93FE, 0x5522, 0x1014, 0x5552, 0x14A, 0x7E38, 0x40, 0xFE7C, 0x2084, 0x4948, 0xFC30, 0x4CE}},
    {38730, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x0, 0x4924, 0x5D74, 0x4924, 0x7FFC, 0x0, 0x3FF8, 0x820, 0xFE0, 0x440, 0x7FFC}},
    {38731, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x2000, 0x17FE, 0x8400, 0x47DC, 0x1488, 0x27FE, 0xE5CC, 0x26AA, 0x293C, 0x2920, 0x13FE}},
    {38732, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8514, 0x7FFC, 0x4084, 0x3CF8, 0x928, 0x3CB0, 0x928, 0x7FF0, 0x1110, 0x1FF0, 0x820, 0x1010}},
    {38733, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x2840, 0x24FC, 0x7E84, 0x48FC, 0xFE84, 0x48FE, 0x7E80, 0x48FE, 0x4802, 0x7EAA, 0x4004}},
    {38734, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x840, 0x14FC, 0x2284, 0x5CFC, 0x8884, 0x3EFE, 0x880, 0x4AFE, 0x2802, 0xEAA, 0x7004}},
    {38735, 2, {0x3FF8, 0x100, 0x7FFE, 0x4102, 0x9D74, 0x1040, 0x7EFC, 0x5284, 0x28FC, 0x3E84, 0x68FE, 0xBE80, 0x28FE, 0x3E02, 0x28AA, 0x3E04}},
    {38736, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x9FF4, 0x1110, 0x1FF0, 0x0, 0x7C7C, 0x1010, 0xFEFE, 0x9292, 0x5454, 0x7C7C, 0x5454, 0x7C7C}},
    {38737, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {38738, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1050, 0x1020}},
    {38739, 2, {0x1000, 0x10FC, 0xFE84, 0x1084, 0x7CA4, 0x10A4, 0xFEA4, 0xA4, 0x7CA4, 0x44A4, 0x7C50, 0x4450, 0x7C50, 0x4492, 0x5492, 0x490E}},
    {38740, 2, {0x1040, 0x1040, 0xFE7E, 0x1080, 0x7D00, 0x107C, 0xFE00, 0x0, 0x7CF8, 0x4408, 0x7C08, 0x4408, 0x7C0A, 0x440A, 0x5406, 0x4802}},
    {38741, 2, {0x20, 0x20, 0xFDFE, 0x1020, 0x10FC, 0x1020, 0x51FE, 0x5C00, 0x50FC, 0x5084, 0x50FC, 0x5084, 0x5CFC, 0xF084, 0x94, 0x88}},
    {38742, 2, {0x2020, 0x1020, 0x13FE, 0xFC20, 0x1FC, 0x820, 0x8BFE, 0x8800, 0x49FC, 0x4904, 0x51FC, 0x5104, 0x1DFC, 0xE104, 0x4114, 0x108}},
    {38743, 2, {0x1020, 0x1020, 0xFD24, 0x10A4, 0x7CA8, 0x1020, 0xFDFE, 0x50, 0x7C50, 0x4450, 0x7C50, 0x4450, 0x7C92, 0x4492, 0x550E, 0x4A00}},
    {38744, 2, {0x1040, 0x1040, 0xFCFC, 0x1088, 0x7D10, 0x10FC, 0xFEA4, 0xA4, 0x7CA4, 0x44FC, 0x7C80, 0x4480, 0x7C82, 0x4482, 0x447E, 0x4C00}},
    {38745, 2, {0x1040, 0x1040, 0xFE78, 0x1088, 0x7C10, 0x11FC, 0xFE24, 0x24, 0x7DFE, 0x4424, 0x7C24, 0x45FC, 0x7C24, 0x4420, 0x54A0, 0x4840}},
    {38746, 2, {0x1000, 0x11FC, 0xFF04, 0x1104, 0x7DFC, 0x1104, 0xFF04, 0x1FC, 0x7D04, 0x4504, 0x7DFC, 0x4450, 0x7C90, 0x4492, 0x5512, 0x4A0E}},
    {38747, 2, {0x1020, 0x1010, 0xFEFE, 0x1082, 0x7D04, 0x1000, 0xFEFE, 0x10, 0x7C10, 0x4490, 0x7C9E, 0x4490, 0x7C90, 0x44D0, 0x553E, 0x4A00}},
    {38748, 2, {0x101E, 0x11E0, 0xFE92, 0x104C, 0x7C00, 0x10FC, 0xFE24, 0x24, 0x7DFE, 0x4424, 0x7C24, 0x44FC, 0x7C24, 0x4420, 0x54A0, 0x4840}},
    {38749, 2, {0x1080, 0x1080, 0xFCFC, 0x1100, 0x7EF8, 0x1000, 0xFDF8, 0x48, 0x7D68, 0x44C8, 0x7DF8, 0x444A, 0x7CEA, 0x475A, 0x5446, 0x4842}},
    {38750, 2, {0x440, 0x440, 0x440, 0xFC7E, 0x440, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x440, 0x440, 0xFC7E, 0x440, 0x440, 0x440}},
    {38751, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0xF0, 0x3F00, 0x1F0, 0x3F00, 0x1F8, 0x7F02, 0x102, 0xFE}},
    {38752, 2, {0x1100, 0x1FF0, 0x2100, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7C7C, 0x440, 0x440}},
    {38753, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2120, 0x2120, 0x3F3E, 0x2120, 0x2F3C, 0x2120, 0x5F3E, 0x4120, 0x8120}},
    {38754, 2, {0x0, 0xFFFE, 0x200, 0x200, 0x400, 0x3FF8, 0x2448, 0x2448, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2448, 0x2448, 0x3FF8, 0x2008}},
    {38755, 2, {0x0, 0xFFFE, 0x200, 0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x27C8, 0x2448, 0x2448, 0x27C8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {38756, 2, {0x40, 0xFE40, 0x10FC, 0x2084, 0xFF04, 0xAAF4, 0xAA94, 0xBA94, 0xAAF4, 0xAA84, 0xBA94, 0xAA88, 0xAA82, 0xFE82, 0x827E, 0x0}},
    {38757, 2, {0x3FFC, 0x2090, 0x3FFC, 0x2140, 0x2630, 0x380C, 0x27F0, 0x2080, 0x2FF8, 0x2A28, 0x2BE8, 0x2A28, 0x4BE8, 0x4A28, 0x8FF8, 0x808}},
    {38758, 2, {0x0, 0xFEFC, 0x1084, 0x2084, 0xFEFC, 0xAA84, 0xAA84, 0xBAFC, 0xAA84, 0xAA84, 0xBAFC, 0xAA50, 0xAA50, 0xFE52, 0x8292, 0x10E}},
    {38759, 2, {0x10, 0xFEFE, 0x1092, 0x20FE, 0xFE10, 0xABFE, 0xAA00, 0xBAFE, 0xAA82, 0xAAFE, 0xBA82, 0xAAFE, 0xAA82, 0xFEFE, 0x8244, 0x82}},
    {38760, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x3FFE, 0x2080, 0x2FFC, 0x2924, 0x49E4, 0x4924, 0x8FFC}},
    {38761, 2, {0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x100, 0x100}},
    {38762, 2, {0x2200, 0x2200, 0xFF7E, 0x2208, 0x3E08, 0x808, 0x7F08, 0x4908, 0x4908, 0x7F08, 0x808, 0xFF88, 0x808, 0x808, 0x828, 0x810}},
    {38763, 2, {0x2800, 0x29F8, 0xFE88, 0x28C8, 0x38A8, 0x10A8, 0x7C88, 0x5450, 0x5450, 0x7C50, 0x1020, 0xFE20, 0x1050, 0x1088, 0x1104, 0x1202}},
    {38764, 2, {0x2200, 0x227C, 0xFF10, 0x2210, 0x3E10, 0x810, 0x7F10, 0x49FE, 0x4910, 0x7F10, 0x810, 0xFF90, 0x810, 0x810, 0x810, 0x810}},
    {38765, 2, {0x2200, 0x2200, 0xFF7E, 0x2212, 0x3E12, 0x852, 0x7F52, 0x4952, 0x4952, 0x7F92, 0x812, 0xFFA2, 0x822, 0x842, 0x894, 0x908}},
    {38766, 2, {0x2220, 0x2220, 0xFF20, 0x227E, 0x3E42, 0x882, 0x7F02, 0x4982, 0x4942, 0x7F22, 0x822, 0xFF82, 0x802, 0x802, 0x814, 0x808}},
    {38767, 2, {0x2210, 0x2210, 0xFF10, 0x2210, 0x3E10, 0x87C, 0x7F10, 0x4910, 0x4910, 0x7F10, 0x810, 0xFF90, 0x810, 0x8FE, 0x800, 0x800}},
    {38768, 2, {0x2800, 0x2800, 0xFDFC, 0x2890, 0x3890, 0x1090, 0x7C90, 0x5490, 0x5490, 0x7C90, 0x1090, 0xFC92, 0x1092, 0x1092, 0x110E, 0x1200}},
    {38769, 2, {0x2800, 0x2800, 0xFDFC, 0x2844, 0x3844, 0x1144, 0x7CC4, 0x5444, 0x5464, 0x7C54, 0x104C, 0xFC86, 0x1084, 0x1104, 0x1228, 0x1410}},
    {38770, 2, {0x2820, 0x2820, 0xFE20, 0x2850, 0x3850, 0x1088, 0x7D44, 0x5622, 0x5420, 0x7CF8, 0x1008, 0xFE08, 0x1010, 0x1010, 0x1020, 0x1020}},
    {38771, 2, {0x2204, 0x220E, 0xFF70, 0x2240, 0x3E40, 0x840, 0x7F7E, 0x4948, 0x4948, 0x7F48, 0x848, 0xFFC8, 0x848, 0x888, 0x888, 0x908}},
    {38772, 2, {0x2850, 0x2850, 0xFE50, 0x2892, 0x3892, 0x1194, 0x7E98, 0x5490, 0x54B0, 0x7CD0, 0x1090, 0xFE90, 0x1092, 0x1092, 0x108E, 0x1080}},
    {38773, 2, {0x2800, 0x29F8, 0xFE48, 0x2848, 0x3848, 0x1048, 0x7C48, 0x55F8, 0x5488, 0x7C88, 0x1088, 0xFE88, 0x1088, 0x1088, 0x13FE, 0x1000}},
    {38774, 2, {0x2800, 0x29FC, 0xFD24, 0x2924, 0x3924, 0x1124, 0x7D24, 0x55FC, 0x5504, 0x7D00, 0x1100, 0xFD00, 0x1102, 0x1102, 0x10FE, 0x1000}},
    {38775, 2, {0x2804, 0x29E4, 0xFE24, 0x2824, 0x3824, 0x11E4, 0x7D04, 0x5504, 0x5504, 0x7DE4, 0x1024, 0xFE24, 0x1024, 0x1024, 0x1144, 0x1084}},
    {38776, 2, {0x2800, 0x2BFC, 0xFC84, 0x2888, 0x3888, 0x1090, 0x7C9C, 0x5484, 0x5544, 0x7D44, 0x1128, 0xFD28, 0x1210, 0x1228, 0x1444, 0x1182}},
    {38777, 2, {0x2820, 0x2820, 0xFE20, 0x29FE, 0x3922, 0x1122, 0x7D22, 0x5522, 0x5552, 0x7D4A, 0x118A, 0xFF02, 0x1102, 0x1102, 0x110A, 0x1104}},
    {38778, 2, {0x2820, 0x2820, 0xFC20, 0x2BFE, 0x3820, 0x1020, 0x7C20, 0x55FC, 0x5470, 0x7CA8, 0x10A8, 0xFD24, 0x1124, 0x1222, 0x1020, 0x1020}},
    {38779, 2, {0x2800, 0x28F8, 0xFE88, 0x2888, 0x3888, 0x10F8, 0x7C88, 0x5488, 0x5488, 0x7CF8, 0x1088, 0xFE88, 0x1088, 0x1088, 0x13FE, 0x1000}},
    {38780, 2, {0x2800, 0x2800, 0xFEFC, 0x2884, 0x3884, 0x1084, 0x7CFC, 0x5484, 0x5484, 0x7C84, 0x10FC, 0xFE84, 0x1000, 0x1000, 0x11FE, 0x1000}},
    {38781, 2, {0x2820, 0x2820, 0xFD24, 0x28A4, 0x38A8, 0x1020, 0x7DFC, 0x5420, 0x5420, 0x7C20, 0x13FE, 0xFC20, 0x1020, 0x1020, 0x1020, 0x1020}},
    {38782, 2, {0x2828, 0x28A8, 0xFEA8, 0x28A8, 0x38A8, 0x13FE, 0x7CA8, 0x54A8, 0x54A8, 0x7CA8, 0x10B8, 0xFE80, 0x1080, 0x1080, 0x10FE, 0x1000}},
    {38783, 2, {0x5010, 0x5090, 0xF890, 0x5110, 0x717E, 0x2252, 0xFB92, 0xA892, 0xA912, 0xF912, 0x2252, 0xFBD2, 0x2062, 0x2022, 0x204A, 0x2084}},
    {38784, 2, {0x2800, 0x29FC, 0xFE44, 0x2844, 0x3844, 0x1044, 0x7C94, 0x5488, 0x5500, 0x7CFC, 0x1084, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {38785, 2, {0x2810, 0x2810, 0xFE10, 0x29FE, 0x3912, 0x1114, 0x7D10, 0x55FC, 0x5544, 0x7D44, 0x1128, 0xFF28, 0x1110, 0x1228, 0x1244, 0x1482}},
    {38786, 2, {0x2808, 0x283C, 0xFDE0, 0x2820, 0x3820, 0x1020, 0x7DFE, 0x5420, 0x5470, 0x7C68, 0x10A8, 0xFCA4, 0x1122, 0x1020, 0x1020, 0x1020}},
    {38787, 2, {0x2808, 0x29C8, 0xFE48, 0x2848, 0x3848, 0x11C8, 0x7D08, 0x5508, 0x5510, 0x7DD0, 0x1054, 0xFE54, 0x1052, 0x105E, 0x1282, 0x1100}},
    {38788, 2, {0x2840, 0x2840, 0xFEFC, 0x2884, 0x3904, 0x12F4, 0x7C94, 0x5494, 0x54F4, 0x7C84, 0x1094, 0xFE88, 0x1082, 0x1082, 0x107E, 0x1000}},
    {38789, 2, {0x2820, 0x2820, 0xFE20, 0x29FC, 0x3924, 0x1124, 0x7D24, 0x5524, 0x5524, 0x7FFE, 0x1020, 0xFE50, 0x1050, 0x1088, 0x1104, 0x1202}},
    {38790, 2, {0x2800, 0x29FE, 0xFC20, 0x2820, 0x3820, 0x11FE, 0x7D22, 0x5522, 0x5552, 0x7D4A, 0x118A, 0xFD02, 0x1102, 0x1102, 0x110A, 0x1104}},
    {38791, 2, {0x2800, 0x29FC, 0xFD04, 0x2924, 0x3924, 0x1124, 0x7DFC, 0x5524, 0x5524, 0x7D54, 0x114C, 0xFD8C, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {38792, 2, {0x2820, 0x2820, 0xFE50, 0x2848, 0x3884, 0x1102, 0x7CFC, 0x5400, 0x5400, 0x7CFC, 0x1084, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {38793, 2, {0x2850, 0x2850, 0xFE50, 0x2952, 0x38D4, 0x1058, 0x7C50, 0x5458, 0x54D4, 0x7D52, 0x1050, 0xFE50, 0x1092, 0x1092, 0x1112, 0x120E}},
    {38794, 2, {0x2820, 0x2820, 0xFC20, 0x2BFE, 0x3820, 0x1020, 0x7DFC, 0x5400, 0x5400, 0x7DFC, 0x1104, 0xFD04, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {38795, 2, {0x2820, 0x2820, 0xFEFC, 0x2820, 0x3820, 0x1020, 0x7DFE, 0x5400, 0x5420, 0x7C20, 0x10FC, 0xFE20, 0x1020, 0x1020, 0x11FE, 0x1000}},
    {38796, 2, {0x100, 0x7FFC, 0x4404, 0x3FF8, 0x820, 0x7C0, 0x7838, 0x440, 0x7FFC, 0x440, 0x7C0, 0x100, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {38797, 2, {0x2840, 0x2820, 0xFC20, 0x2BFE, 0x3A02, 0x1444, 0x7C40, 0x57FE, 0x5488, 0x7C88, 0x1108, 0xFCD0, 0x1020, 0x1050, 0x1088, 0x1304}},
    {38798, 2, {0x2800, 0x28FC, 0xFE84, 0x2884, 0x38FC, 0x1084, 0x7C84, 0x54FC, 0x54A2, 0x7CA4, 0x1098, 0xFE90, 0x1088, 0x10A4, 0x10C2, 0x1080}},
    {38799, 2, {0x7CF8, 0x1088, 0x10CA, 0x1D2A, 0xF206, 0x820, 0x7FFC, 0x820, 0xFE0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0xFFFE, 0x100}},
    {38800, 2, {0x2820, 0x2820, 0xFE3E, 0x2820, 0x3820, 0x11FE, 0x7C00, 0x5400, 0x55FE, 0x7C20, 0x1028, 0xFE24, 0x1022, 0x1020, 0x1020, 0x1020}},
    {38801, 2, {0x2810, 0x2910, 0xFC90, 0x2890, 0x387E, 0x1010, 0x7D90, 0x5490, 0x5498, 0x7CA8, 0x10A4, 0xFCC4, 0x1080, 0x1140, 0x123E, 0x1000}},
    {38802, 2, {0x2808, 0x281C, 0xFEE0, 0x2820, 0x3820, 0x11FE, 0x7C50, 0x5488, 0x5504, 0x7E8A, 0x1088, 0xFE88, 0x1088, 0x1088, 0x1088, 0x1108}},
    {38803, 2, {0x2800, 0x28FC, 0xFE84, 0x2884, 0x3884, 0x10FC, 0x7C00, 0x5400, 0x55FE, 0x7C20, 0x1020, 0xFEFC, 0x1020, 0x1020, 0x11FE, 0x1000}},
    {38804, 2, {0x2880, 0x2880, 0xFCF8, 0x2908, 0x3A10, 0x11FC, 0x7D24, 0x5524, 0x5524, 0x7DFC, 0x1050, 0xFC50, 0x1090, 0x1092, 0x1112, 0x120E}},
    {38805, 2, {0x2800, 0x29FE, 0xFC20, 0x2820, 0x39FC, 0x1124, 0x7D24, 0x55FC, 0x5524, 0x7D24, 0x11FC, 0xFD20, 0x10A0, 0x1040, 0x10B0, 0x130E}},
    {38806, 2, {0x2804, 0x281E, 0xFDE0, 0x2822, 0x3912, 0x1094, 0x7C80, 0x5420, 0x55FE, 0x7C44, 0x1084, 0xFDC8, 0x1030, 0x1028, 0x10C4, 0x1302}},
    {38807, 2, {0x1080, 0x11F8, 0x1288, 0x2070, 0x2B8E, 0x6888, 0x6BFE, 0xA888, 0x28F8, 0x2820, 0x29FC, 0x2924, 0x29FC, 0x2020, 0x23FE, 0x2020}},
    {38808, 2, {0x2810, 0x2892, 0xFE52, 0x2854, 0x3810, 0x10FE, 0x7C82, 0x5482, 0x54FE, 0x7C82, 0x1082, 0xFEFE, 0x1082, 0x1082, 0x108A, 0x1084}},
    {38809, 2, {0x2800, 0x28F8, 0xFE88, 0x2888, 0x38F8, 0x1000, 0x7DFC, 0x5504, 0x5504, 0x7DFC, 0x1104, 0xFF04, 0x11FC, 0x1104, 0x1114, 0x1108}},
    {38810, 2, {0x2820, 0x2810, 0xFE10, 0x29FE, 0x3902, 0x1254, 0x7C88, 0x5504, 0x5400, 0x7CFC, 0x1020, 0xFE20, 0x1020, 0x1020, 0x11FE, 0x1000}},
    {38811, 2, {0x2820, 0x2810, 0xFEFC, 0x2800, 0x3888, 0x1050, 0x7C00, 0x55FE, 0x5400, 0x7C00, 0x10FC, 0xFE84, 0x1084, 0x1084, 0x10FC, 0x1084}},
    {38812, 2, {0x2820, 0x2822, 0xFDB4, 0x28A8, 0x38A8, 0x1124, 0x7EA2, 0x5440, 0x55FC, 0x7D04, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {38813, 2, {0x2820, 0x2924, 0xFCA4, 0x28A8, 0x3820, 0x11FC, 0x7D04, 0x5504, 0x5574, 0x7D54, 0x1154, 0xFD54, 0x1174, 0x1104, 0x1114, 0x1108}},
    {38814, 2, {0x2820, 0x2840, 0xFCFE, 0x2892, 0x3892, 0x10FE, 0x7C92, 0x54A2, 0x54FE, 0x7C48, 0x1088, 0xFDFE, 0x1008, 0x1008, 0x1008, 0x1008}},
    {38815, 2, {0x2820, 0x2810, 0xFDFE, 0x2800, 0x38FC, 0x1084, 0x7CFC, 0x5400, 0x54FC, 0x7C08, 0x1010, 0xFDFE, 0x1010, 0x1010, 0x1050, 0x1020}},
    {38816, 2, {0x2880, 0x2880, 0xFEFE, 0x2902, 0x3A22, 0x10AA, 0x7C72, 0x5422, 0x55FE, 0x7C22, 0x1072, 0xFEAA, 0x1122, 0x1022, 0x100A, 0x1004}},
    {38817, 2, {0x2890, 0x2888, 0xFC88, 0x28BE, 0x3BC0, 0x1082, 0x7CA2, 0x54E2, 0x5492, 0x7D92, 0x1294, 0xFC94, 0x1084, 0x1088, 0x12BE, 0x1100}},
    {38818, 2, {0x2848, 0x2948, 0xFD48, 0x2BFE, 0x3948, 0x1148, 0x7D78, 0x5500, 0x55FE, 0x7C20, 0x13FE, 0xFC70, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {38819, 2, {0x2800, 0x29FC, 0xFC48, 0x2830, 0x3BFE, 0x1052, 0x7C94, 0x5510, 0x5630, 0x7C20, 0x13FE, 0xFC70, 0x10A8, 0x1124, 0x1622, 0x1020}},
    {38820, 2, {0x808, 0x7F08, 0x8FE, 0xFF08, 0x848, 0x7F28, 0x808, 0xF28, 0xF010, 0x440, 0x7FFC, 0x440, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {38821, 2, {0x2820, 0x2830, 0xFE48, 0x2884, 0x397A, 0x1000, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x1048, 0xFE48, 0x11FE, 0x1048, 0x1088, 0x1108}},
    {38822, 2, {0x5008, 0x50C8, 0xFB08, 0x5108, 0x712A, 0x212A, 0xFBAC, 0xA948, 0xA908, 0xFB88, 0x2554, 0xF914, 0x2114, 0x2124, 0x2124, 0x2142}},
    {38823, 2, {0x2888, 0x2848, 0xFC50, 0x2BFE, 0x3850, 0x1050, 0x7DFC, 0x5554, 0x5554, 0x7D8C, 0x1104, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {38824, 2, {0x2800, 0x29FC, 0xFD04, 0x29FC, 0x3904, 0x11FC, 0x7C80, 0x55FE, 0x5622, 0x7D22, 0x1152, 0xFD02, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {38825, 2, {0x5082, 0x5082, 0xFAA2, 0x51CA, 0x708A, 0x23EA, 0xFA2A, 0xAA2A, 0xABEA, 0xFA2A, 0x222A, 0xFBEA, 0x2222, 0x2222, 0x22AA, 0x2244}},
    {38826, 2, {0x7E20, 0x243E, 0x1848, 0xFFA8, 0x2910, 0x4A28, 0x9846, 0x440, 0x7FFC, 0x440, 0x7C0, 0x100, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {38827, 2, {0x2880, 0x2880, 0xFCFE, 0x2902, 0x3A42, 0x11FA, 0x7C02, 0x55F2, 0x5402, 0x7DF2, 0x1002, 0xFDF2, 0x1112, 0x11F2, 0x1014, 0x1008}},
    {38828, 2, {0x5010, 0x5010, 0xF87C, 0x5714, 0x71FE, 0x2114, 0xFA7C, 0xAA10, 0xAF7C, 0xF910, 0x21FE, 0xFD10, 0x2210, 0x2300, 0x24FE, 0x2800}},
    {38829, 2, {0x2880, 0x28FE, 0xFE90, 0x297C, 0x3954, 0x1354, 0x7D7C, 0x5554, 0x5554, 0x7D7C, 0x1190, 0xFF50, 0x1120, 0x1130, 0x1148, 0x1186}},
    {38830, 2, {0x2800, 0x28FC, 0xFE84, 0x2884, 0x38FC, 0x1084, 0x7C84, 0x54FC, 0x5400, 0x7DFE, 0x1020, 0xFEA0, 0x10BC, 0x10A0, 0x1160, 0x123E}},
    {38831, 2, {0x2848, 0x2848, 0xFDFE, 0x2848, 0x3820, 0x11FE, 0x7C40, 0x5440, 0x54BC, 0x7D84, 0x1288, 0xFCBE, 0x1088, 0x1088, 0x10A8, 0x1090}},
    {38832, 2, {0x2800, 0x28F8, 0xFE88, 0x2888, 0x38F8, 0x1088, 0x7C88, 0x54F8, 0x5400, 0x7DFC, 0x1154, 0xFF54, 0x1154, 0x1154, 0x13FE, 0x1000}},
    {38833, 2, {0x281E, 0x28E0, 0xFE22, 0x2892, 0x3854, 0x1040, 0x7C20, 0x54CE, 0x5482, 0x7C82, 0x10EE, 0xFE82, 0x1082, 0x1082, 0x10FE, 0x1082}},
    {38834, 2, {0x5088, 0x5088, 0xFBFE, 0x5088, 0x71FC, 0x2088, 0xFBFE, 0xA820, 0xA9FC, 0xF924, 0x21FC, 0xF924, 0x23FE, 0x2104, 0x2114, 0x2108}},
    {38835, 2, {0x2888, 0x2888, 0xFDFE, 0x2888, 0x3820, 0x1050, 0x7C88, 0x5504, 0x56FA, 0x7C00, 0x1000, 0xFDFC, 0x1104, 0x1104, 0x11FC, 0x1104}},
    {38836, 2, {0x5088, 0x5088, 0xFBFE, 0x5088, 0x7088, 0x27FE, 0xF880, 0xA9FC, 0xAB24, 0xFDFC, 0x2124, 0xF9FC, 0x2124, 0x2124, 0x2124, 0x210C}},
    {38837, 2, {0x281E, 0x2BE0, 0xFD22, 0x2894, 0x3840, 0x1088, 0x7DF0, 0x5420, 0x54C4, 0x7DFE, 0x1022, 0xFC20, 0x13FE, 0x1050, 0x1088, 0x1306}},
    {38838, 2, {0x878, 0x3E48, 0x228E, 0x2A00, 0x7EF8, 0x2248, 0x2A30, 0x46CC, 0x7FFC, 0x440, 0x7C0, 0x100, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {38839, 2, {0x7FFC, 0x820, 0xFE0, 0x0, 0x3FF8, 0x2448, 0x2FE8, 0x2108, 0x2118, 0x440, 0x7FFC, 0x440, 0x1FF0, 0x1110, 0xFFFE, 0x100}},
    {38840, 2, {0x2800, 0x29FC, 0xFD24, 0x29FC, 0x3924, 0x11FC, 0x7CA8, 0x54A8, 0x57FE, 0x7CA8, 0x10A8, 0xFDFC, 0x1020, 0x13FE, 0x1020, 0x1020}},
    {38841, 2, {0x5100, 0x509E, 0xFBD2, 0x5012, 0x73D4, 0x2254, 0xFBD8, 0xA814, 0xABD2, 0xF852, 0x2092, 0xF8DA, 0x2394, 0x2090, 0x2290, 0x2110}},
    {38842, 2, {0x2820, 0x2924, 0xFCA8, 0x2BFE, 0x3A02, 0x10F8, 0x7C88, 0x5488, 0x54F8, 0x7C20, 0x1020, 0xFDFC, 0x1020, 0x1020, 0x13FE, 0x1000}},
    {38843, 2, {0x2820, 0x29FC, 0xFD24, 0x2BFE, 0x3924, 0x11FC, 0x7C20, 0x55FC, 0x5524, 0x7DFC, 0x1040, 0xFDFE, 0x1088, 0x11D0, 0x1070, 0x138C}},
    {38844, 2, {0x2820, 0x29FC, 0xFD24, 0x29FC, 0x3820, 0x13FE, 0x7C00, 0x55FC, 0x5504, 0x7DFC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {38845, 2, {0x2818, 0x29E0, 0xFC40, 0x2BFE, 0x3888, 0x1174, 0x7E52, 0x5470, 0x5400, 0x7DFC, 0x1104, 0xFD74, 0x1154, 0x1174, 0x1104, 0x110C}},
    {38846, 2, {0x2888, 0x2BFE, 0xFC88, 0x2800, 0x39FC, 0x10A8, 0x7CA8, 0x57FE, 0x54A8, 0x7CA8, 0x11FC, 0xFC20, 0x13FE, 0x1020, 0x1020, 0x1020}},
    {38847, 2, {0x5124, 0x5124, 0xFAAA, 0x53AE, 0x7124, 0x22AA, 0xFBAE, 0xA924, 0xAFFE, 0xF910, 0x2114, 0xF914, 0x228A, 0x224A, 0x2216, 0x2422}},
    {38848, 2, {0x5040, 0x5020, 0xFBFE, 0x5250, 0x73FE, 0x2252, 0xFBFE, 0xAA80, 0xAAFE, 0xFA80, 0x22FE, 0xFA02, 0x2556, 0x2556, 0x2A02, 0x2006}},
    {38849, 2, {0x2800, 0x2BFE, 0xFC00, 0x29FC, 0x3924, 0x11FC, 0x7D24, 0x57FE, 0x5400, 0x7DFC, 0x1124, 0xFDFC, 0x1124, 0x11FC, 0x1000, 0x13FE}},
    {38850, 2, {0x2880, 0x28F8, 0xFD08, 0x2BFE, 0x3944, 0x1192, 0x7DFE, 0x5500, 0x557C, 0x7D00, 0x117C, 0xFD00, 0x117C, 0x1244, 0x127C, 0x1444}},
    {38851, 2, {0x2810, 0x2A10, 0xFD7C, 0x2810, 0x38FE, 0x1044, 0x7F28, 0x55FE, 0x5510, 0x7D7C, 0x1110, 0xFDFE, 0x1110, 0x1110, 0x12FE, 0x1400}},
    {38852, 2, {0x2888, 0x2BFE, 0xFCA8, 0x2890, 0x39FE, 0x1320, 0x7DFC, 0x5520, 0x55FC, 0x7D20, 0x11FE, 0xFD00, 0x13FC, 0x1088, 0x1070, 0x138E}},
    {38853, 2, {0x2800, 0x29FC, 0xFD04, 0x29FC, 0x3904, 0x11FC, 0x7C88, 0x5554, 0x55DC, 0x7C88, 0x1154, 0xFDDC, 0x1000, 0x11A8, 0x1154, 0x1254}},
    {38854, 2, {0x5000, 0x50FE, 0xFC28, 0x52FE, 0x72AA, 0x20FE, 0xF810, 0xA9FE, 0xAE44, 0xFAFE, 0x2344, 0xFA7C, 0x2242, 0x223E, 0x2500, 0x28FE}},
    {38855, 2, {0x2820, 0x2BFE, 0xFC20, 0x29FC, 0x3800, 0x13FE, 0x7E52, 0x57FE, 0x5504, 0x7DFC, 0x1104, 0xFDFC, 0x1104, 0x11FC, 0x1088, 0x1104}},
    {38856, 2, {0x2848, 0x29FE, 0xFC48, 0x2800, 0x39FE, 0x114A, 0x7D4A, 0x55FE, 0x5414, 0x7DFE, 0x1110, 0xFD92, 0x114C, 0x114A, 0x1116, 0x1222}},
    {38857, 2, {0x5088, 0x53FE, 0xF888, 0x5020, 0x73FE, 0x2250, 0xFBFE, 0xAA52, 0xABFE, 0xFA80, 0x22FE, 0xFA80, 0x22FE, 0x2402, 0x2AAA, 0x2004}},
    {38858, 2, {0x5088, 0x53FE, 0xF888, 0x53DE, 0x7252, 0x23DE, 0xFA52, 0xABDE, 0xAA22, 0xFBFE, 0x22AA, 0xFAFA, 0x2222, 0x2272, 0x22AA, 0x2224}},
    {38859, 2, {0x200, 0x200, 0x3FF0, 0x410, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x100, 0x2100, 0x3FFC, 0x100, 0x100}},
    {38860, 2, {0x1000, 0x1000, 0x7EFE, 0x2212, 0xFF12, 0x52, 0x7E52, 0x4252, 0x7E92, 0x812, 0x7E12, 0x822, 0x4822, 0x7E42, 0x88A, 0x904}},
    {38861, 2, {0x1028, 0x1024, 0x7E24, 0x2220, 0xFF7E, 0x20, 0x7E20, 0x423C, 0x7E34, 0x854, 0x7E54, 0x854, 0x4888, 0x7E88, 0x914, 0x822}},
    {38862, 2, {0x1010, 0x1010, 0x7E10, 0x22FE, 0xFF10, 0x10, 0x7E10, 0x427C, 0x7E10, 0x838, 0x7E38, 0x854, 0x4854, 0x7E92, 0x810, 0x810}},
    {38863, 2, {0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1210, 0x2FE8, 0xC426, 0x3FF8, 0x1010, 0x1FF0, 0x100, 0x1FF0, 0x2100, 0x3FF8, 0x100}},
    {38864, 2, {0x1020, 0x1020, 0x7C50, 0x2450, 0xFE88, 0x104, 0x7EFA, 0x4400, 0x7C00, 0x10F8, 0x7C88, 0x1088, 0x5088, 0x7C88, 0x10F8, 0x1088}},
    {38865, 2, {0x1020, 0x11FC, 0x9444, 0x55FE, 0x5800, 0x11FC, 0xFD04, 0x29FC, 0x2820, 0x29FC, 0x2820, 0x2920, 0x29FC, 0x4820, 0x4822, 0x87FE}},
    {38866, 2, {0x1010, 0x1092, 0x7E52, 0x2254, 0xFF10, 0xFE, 0x7E82, 0x4282, 0x7EFE, 0x882, 0x7E82, 0x8FE, 0x4882, 0x7E82, 0x88A, 0x884}},
    {38867, 2, {0x1020, 0x1020, 0x10FC, 0xFE44, 0x11FE, 0x7C00, 0x44FC, 0x7C84, 0x44FC, 0x7C10, 0x10FC, 0xFE10, 0x1090, 0x10FE, 0x1010, 0x1010}},
    {38868, 2, {0x1000, 0x107C, 0x7E40, 0x2278, 0xFF40, 0x78, 0x7E40, 0x43FE, 0x7EA0, 0x8A4, 0x7E98, 0x890, 0x4888, 0x7EA4, 0x8C2, 0x880}},
    {38869, 2, {0x1020, 0x1010, 0x7EFE, 0x2200, 0xFF7C, 0x44, 0x7E7C, 0x4200, 0x7E7C, 0x808, 0x7E10, 0x8FE, 0x4810, 0x7E10, 0x850, 0x820}},
    {38870, 2, {0x1000, 0x10FC, 0x7C48, 0x2430, 0xFEFE, 0x52, 0x7C94, 0x4550, 0x7C20, 0x1010, 0x7DFE, 0x1038, 0x5054, 0x7C92, 0x1110, 0x1010}},
    {38871, 2, {0x1000, 0x11FE, 0x7D02, 0x2420, 0xFEFC, 0x20, 0x7DFC, 0x4524, 0x7DFC, 0x1124, 0x7DFC, 0x1020, 0x53FE, 0x7C20, 0x1020, 0x1020}},
    {38872, 2, {0x1014, 0x1054, 0x7E54, 0x22FE, 0xFF54, 0x54, 0x7E5C, 0x4240, 0x7E7E, 0x810, 0x7EFE, 0x810, 0x4838, 0x7E54, 0x892, 0x810}},
    {38873, 2, {0x10, 0x3E7C, 0x2224, 0x3EFE, 0x2200, 0x3E7C, 0x44, 0xFF7C, 0x810, 0x83C, 0x2F50, 0x287E, 0x2810, 0x5810, 0x4FFE, 0x8000}},
    {38874, 2, {0x1088, 0x1088, 0x7DFE, 0x2488, 0xFEF8, 0x20, 0x7DFC, 0x4524, 0x7D24, 0x11FC, 0x7C20, 0x13FE, 0x5020, 0x7C20, 0x1020, 0x1020}},
    {38875, 2, {0x1044, 0x1044, 0x7DFE, 0x2444, 0xFE44, 0x1FE, 0x7C40, 0x44FE, 0x7D92, 0x12FE, 0x7C92, 0x10FE, 0x5092, 0x7C92, 0x1092, 0x1086}},
    {38876, 2, {0x101E, 0x11E0, 0x7C22, 0x2512, 0xFE94, 0x80, 0x7C20, 0x45CE, 0x7D02, 0x1102, 0x7DCE, 0x1102, 0x5102, 0x7D02, 0x11FE, 0x1102}},
    {38877, 2, {0x1048, 0x1048, 0x7DFE, 0x2448, 0xFEFC, 0x48, 0x7DFE, 0x4410, 0x7CFC, 0x1094, 0x7CFC, 0x1094, 0x51FE, 0x7C84, 0x1094, 0x1088}},
    {38878, 2, {0x1000, 0x10F8, 0x7C88, 0x2488, 0xFEF8, 0x88, 0x7C88, 0x44F8, 0x7C00, 0x11FC, 0x7D54, 0x1154, 0x5154, 0x7D54, 0x13FE, 0x1000}},
    {38879, 2, {0x1010, 0x1020, 0x7E7C, 0x2244, 0xFF7C, 0x44, 0x7E7C, 0x4210, 0x7EFE, 0x828, 0x7E54, 0x892, 0x487C, 0x7E10, 0x810, 0x810}},
    {38880, 2, {0x1000, 0x10FE, 0x7E92, 0x22FE, 0xFF92, 0xFE, 0x7E54, 0x4254, 0x7EFE, 0x854, 0x7E54, 0x8FE, 0x4810, 0x7EFE, 0x810, 0x810}},
    {38881, 2, {0x1044, 0x10FE, 0x7E44, 0x2200, 0xFFFE, 0x54, 0x7E54, 0x42FE, 0x7E54, 0x854, 0x7EFE, 0x810, 0x48FE, 0x7E10, 0x810, 0x810}},
    {38882, 2, {0x1010, 0x1010, 0x7DFE, 0x2410, 0xFEFC, 0x94, 0x7CFC, 0x4494, 0x7CFC, 0x1012, 0x7DFE, 0x1022, 0x5014, 0x7D52, 0x114A, 0x1238}},
    {38883, 2, {0x1000, 0x11FC, 0x7D54, 0x2554, 0xFFFC, 0x80, 0x7DFC, 0x4644, 0x7DF4, 0x1154, 0x7D54, 0x11F4, 0x5044, 0x7DF4, 0x1014, 0x1008}},
    {38884, 2, {0x1048, 0x11FE, 0x7C48, 0x2400, 0xFFFE, 0x14A, 0x7D4A, 0x45FE, 0x7C14, 0x11FE, 0x7D10, 0x1192, 0x514C, 0x7D4A, 0x1116, 0x1222}},
    {38885, 2, {0x1010, 0x10FE, 0x7E10, 0x227C, 0xFF00, 0xFE, 0x7EAA, 0x42FE, 0x7E44, 0x87C, 0x7E44, 0x87C, 0x4844, 0x7E7C, 0x828, 0x844}},
    {38886, 2, {0x100, 0x100, 0x100, 0x7FFE, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x104, 0x104, 0x104, 0x128, 0x110, 0x100}},
    {38887, 2, {0x2000, 0x2000, 0x21FC, 0xFC24, 0x2024, 0x20A4, 0xF8A4, 0x20A4, 0x2124, 0xFC24, 0x2424, 0x2444, 0x3444, 0x2884, 0x2114, 0x2208}},
    {38888, 2, {0x2050, 0x2048, 0x2048, 0xFC40, 0x21FE, 0x2040, 0xF840, 0x207C, 0x20A4, 0xFCA4, 0x24A8, 0x2528, 0x3510, 0x2A28, 0x2044, 0x2082}},
    {38889, 2, {0x1020, 0x1020, 0x1020, 0xFEFE, 0x1020, 0x7C20, 0x44FC, 0x7C20, 0x4420, 0x7DFE, 0x1022, 0xFE22, 0x102A, 0x1024, 0x1020, 0x1020}},
    {38890, 2, {0x20, 0x3E20, 0x22FC, 0x3E20, 0x2220, 0x3EF8, 0x20, 0xFF20, 0x8FC, 0x824, 0x2F24, 0x2834, 0x2828, 0x5820, 0x4FFE, 0x8000}},
    {38891, 2, {0x2000, 0x20F8, 0x2088, 0xFC88, 0x20F8, 0x2088, 0xF888, 0x20F8, 0x2000, 0xFDFC, 0x2554, 0x2554, 0x3554, 0x2954, 0x23FE, 0x2000}},
    {38892, 2, {0x201E, 0x21E0, 0x2022, 0xFD12, 0x2094, 0x2080, 0xF820, 0x21CE, 0x2102, 0xFD02, 0x25CE, 0x2502, 0x3502, 0x2902, 0x21FE, 0x2102}},
    {38893, 2, {0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x440, 0xFFFE, 0x0}},
    {38894, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x440, 0x7C7C, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0x7C7C, 0x440, 0x440, 0xFFFE}},
    {38895, 2, {0x820, 0x828, 0x7F24, 0x820, 0xFFFE, 0x20, 0x1224, 0x73A4, 0x1224, 0x73A8, 0x1228, 0x7390, 0x1212, 0x17AA, 0xF846, 0x4082}},
    {38896, 2, {0x800, 0xF7C, 0x2844, 0x7F28, 0xA110, 0x1628, 0x38C6, 0xC280, 0x7EFC, 0x280, 0x3EF8, 0x280, 0x7EFC, 0x280, 0xFFFE, 0x0}},
    {38897, 2, {0x1120, 0x1128, 0x2AA4, 0x4464, 0x20, 0xFFFE, 0x1220, 0x73A4, 0x1224, 0x73A8, 0x1228, 0x7390, 0x1212, 0x17AA, 0xF846, 0x4082}},
    {38898, 2, {0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x280, 0x7EFC, 0x280, 0x3EF8, 0x280, 0x7EFC, 0x280, 0xFFFE, 0x0}},
    {38899, 2, {0x200, 0x100, 0x3FF8, 0x0, 0x820, 0x440, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {38900, 2, {0x2000, 0x13FE, 0x7C20, 0x20, 0x4420, 0x29FC, 0xFF24, 0x124, 0x7D24, 0x4524, 0x4524, 0x7D34, 0x4528, 0x4420, 0x7C20, 0x4420}},
    {38901, 2, {0x2040, 0x1040, 0x7C80, 0xFC, 0x4504, 0x2A04, 0xFE84, 0x44, 0x7C44, 0x4414, 0x4424, 0x7C44, 0x4584, 0x4404, 0x7C28, 0x4410}},
    {38902, 2, {0x2000, 0x11FC, 0x7C44, 0x44, 0x4444, 0x2844, 0xFE94, 0x88, 0x7D00, 0x44FC, 0x4484, 0x7C84, 0x4484, 0x4484, 0x7CFC, 0x4484}},
    {38903, 2, {0x2010, 0x1010, 0x7C92, 0x92, 0x4492, 0x2892, 0xFEFE, 0x10, 0x7C10, 0x4492, 0x4492, 0x7C92, 0x4492, 0x4492, 0x7CFE, 0x4402}},
    {38904, 2, {0x2040, 0x1040, 0x7CFC, 0x88, 0x4550, 0x2820, 0xFED8, 0x326, 0x7CF8, 0x4420, 0x44F8, 0x7C20, 0x47FE, 0x4420, 0x7C20, 0x4420}},
    {38905, 2, {0x2020, 0x1040, 0x7CFC, 0x84, 0x44FC, 0x2884, 0xFEFC, 0x0, 0x7DFE, 0x4420, 0x4420, 0x7CFC, 0x4420, 0x4420, 0x7DFE, 0x4400}},
    {38906, 2, {0x2088, 0x1088, 0x7DFC, 0x88, 0x44A8, 0x2820, 0xFEF8, 0xA8, 0x7CA8, 0x44A8, 0x44A8, 0x7DFC, 0x4420, 0x4450, 0x7C88, 0x4504}},
    {38907, 2, {0x2000, 0x107C, 0x7C44, 0x44, 0x447C, 0x2800, 0xFEFE, 0x82, 0x7CFE, 0x4482, 0x44FE, 0x7C82, 0x44FE, 0x4428, 0x7C44, 0x4482}},
    {38908, 2, {0x2020, 0x1220, 0x797C, 0xA4, 0x4818, 0x30E6, 0xFB10, 0x17E, 0x7910, 0x497C, 0x4910, 0x79FE, 0x4910, 0x4910, 0x7AFE, 0x4C00}},
    {38909, 2, {0x1020, 0x2810, 0x44FE, 0x9200, 0x7C44, 0x828, 0xFEFE, 0x2800, 0xFE7C, 0xAA44, 0xCE44, 0x827C, 0xFE44, 0x8244, 0xFE7C, 0x8244}},
    {38910, 2, {0x800, 0x7F78, 0x848, 0x3E48, 0x86, 0x3E78, 0x2A48, 0x3E30, 0x414E, 0xBFFC, 0x440, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {38911, 2, {0x1100, 0x23DE, 0x7A52, 0x13D4, 0x2A52, 0x7BDA, 0x1294, 0x2350, 0x100, 0x3FF8, 0x440, 0xFFFE, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {38912, 2, {0x2044, 0x11FE, 0x7C54, 0x48, 0x44FE, 0x2990, 0xFEFE, 0x90, 0x7CFE, 0x4490, 0x44FE, 0x7C80, 0x45FE, 0x4444, 0x7C38, 0x45C6}},
    {38913, 2, {0x0, 0x7FFE, 0x200, 0x400, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x1FF8, 0x1008, 0x1008, 0x1FF8, 0x810, 0x1008, 0x2004}},
    {38914, 2, {0x0, 0x1FE, 0xFE20, 0x1040, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1088, 0x5104, 0x2202}},
    {38915, 2, {0x0, 0x21FE, 0x2020, 0x2040, 0x21FC, 0x3D04, 0x2104, 0x21FC, 0x2104, 0x25FC, 0x2904, 0x3104, 0x21FC, 0x88, 0x104, 0x202}},
    {38916, 2, {0x2000, 0x20FE, 0x2010, 0x2020, 0xFCFC, 0x2484, 0x2484, 0x24FC, 0x2484, 0x24FC, 0x2584, 0x2684, 0x44FC, 0x4048, 0x8084, 0x102}},
    {38917, 2, {0x0, 0x1FE, 0x20, 0xFC40, 0x11FC, 0x1104, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1D04, 0xE1FC, 0x4088, 0x104, 0x202}},
    {38918, 2, {0x400, 0x45FE, 0x5420, 0x5440, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x5488, 0x8504, 0x602}},
    {38919, 2, {0x0, 0x7CFE, 0x1010, 0x1020, 0x10FC, 0x1084, 0x1084, 0xFEFC, 0x1084, 0x10FC, 0x1084, 0x1084, 0x10FC, 0x1048, 0x1084, 0x1102}},
    {38920, 2, {0x800, 0x9FE, 0x1020, 0x2040, 0x41FC, 0x8904, 0x904, 0x11FC, 0x2104, 0x45FC, 0x8504, 0x904, 0x11FC, 0x2088, 0x4104, 0x8202}},
    {38921, 2, {0x800, 0x9FE, 0x810, 0x820, 0x48FC, 0x4E84, 0x4884, 0x48FC, 0x4884, 0x48FC, 0x4884, 0x4E84, 0x58FC, 0xE048, 0x84, 0x102}},
    {38922, 2, {0x0, 0x1FE, 0x7C20, 0x1040, 0x11FC, 0x1104, 0x7D04, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x1D04, 0xE1FC, 0x4088, 0x104, 0x202}},
    {38923, 2, {0x0, 0x7EFE, 0x4010, 0x4020, 0x5EFC, 0x5284, 0x5284, 0x52FC, 0x5A84, 0x54FC, 0x5084, 0x5284, 0x54FC, 0x5848, 0x5084, 0x8102}},
    {38924, 2, {0x0, 0x28FE, 0x2410, 0x2420, 0x42FC, 0x5284, 0x9084, 0x10FC, 0x2084, 0x20FC, 0x4884, 0x4484, 0xFCFC, 0x4448, 0x84, 0x102}},
    {38925, 2, {0x1000, 0x11FE, 0x1020, 0xFE40, 0x11FC, 0x1104, 0xFD04, 0x5FC, 0x504, 0x49FC, 0x2904, 0x1104, 0x29FC, 0x4488, 0x8104, 0x202}},
    {38926, 2, {0x800, 0x1DFE, 0x6020, 0x4040, 0x41FC, 0x4104, 0x7F04, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x4904, 0x49FC, 0x4888, 0x8904, 0xA02}},
    {38927, 2, {0x1000, 0x8FE, 0x810, 0xFF20, 0xFC, 0x84, 0x3C84, 0x24FC, 0x2484, 0x24FC, 0x2484, 0x2584, 0x46FC, 0x4448, 0x8084, 0x102}},
    {38928, 2, {0x0, 0x7CFE, 0x410, 0x2820, 0x10FC, 0x884, 0xFE84, 0x12FC, 0x1484, 0x10FC, 0x1084, 0x1084, 0x10FC, 0x1048, 0x5084, 0x2102}},
    {38929, 2, {0x0, 0x7CFE, 0x10, 0x20, 0xFC, 0xFE84, 0x2884, 0x28FC, 0x2884, 0x28FC, 0x2884, 0x2A84, 0x4CFC, 0x4848, 0x8084, 0x102}},
    {38930, 2, {0x0, 0x28FE, 0x2410, 0x2420, 0x42FC, 0x4284, 0x8084, 0x7CFC, 0x2484, 0x24FC, 0x2484, 0x2484, 0x44FC, 0x4448, 0x9484, 0x902}},
    {38931, 2, {0x1000, 0x11FE, 0x1020, 0xFE40, 0x11FC, 0x5504, 0x5504, 0x55FC, 0x5504, 0x7DFC, 0x1104, 0x1104, 0x15FC, 0x1888, 0x1104, 0x202}},
    {38932, 2, {0x1000, 0x11FE, 0x1020, 0x7C40, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x7D04, 0x55FC, 0x5504, 0x5504, 0x55FC, 0x7C88, 0x4504, 0x202}},
    {38933, 2, {0x1000, 0x10FE, 0x1010, 0x1E20, 0x10FC, 0x1084, 0x1084, 0x10FC, 0x7E84, 0x42FC, 0x4284, 0x4284, 0x42FC, 0x7E48, 0x4284, 0x102}},
    {38934, 2, {0x1000, 0x11FE, 0x9420, 0x5840, 0x11FC, 0x7D04, 0x1104, 0x11FC, 0xFF04, 0x11FC, 0x1104, 0x1104, 0x21FC, 0x2088, 0x4104, 0x8202}},
    {38935, 2, {0x800, 0x8FE, 0x810, 0x7F20, 0x497C, 0x4A44, 0x4844, 0x7F7C, 0x4144, 0x527C, 0x4A44, 0x4444, 0x4A7C, 0x9100, 0x2128, 0x44}},
    {38936, 2, {0x1000, 0x11FE, 0x2820, 0x2440, 0x43FC, 0x9104, 0x904, 0x1FC, 0xFD04, 0x5FC, 0x904, 0x5104, 0x21FC, 0x1088, 0x1104, 0x202}},
    {38937, 2, {0x0, 0x1FE, 0xFC20, 0x1040, 0x11FC, 0x1104, 0x5104, 0x5DFC, 0x5104, 0x51FC, 0x5104, 0x5104, 0x5DFC, 0xF088, 0x104, 0x202}},
    {38938, 2, {0x0, 0x7CFE, 0x410, 0x6820, 0x10FC, 0x2884, 0x4484, 0x92FC, 0x1084, 0x7CFC, 0x1084, 0x1084, 0x1EFC, 0xF048, 0x4084, 0x102}},
    {38939, 2, {0x1000, 0x10FE, 0x1010, 0xFE20, 0x10FC, 0x7C84, 0x1084, 0xFEFC, 0x3084, 0x38FC, 0x5484, 0x5484, 0x90FC, 0x1048, 0x1084, 0x1102}},
    {38940, 2, {0x1000, 0x11FE, 0x2820, 0x2440, 0x43FC, 0x8104, 0x7D04, 0x1FC, 0x104, 0x7DFC, 0x4504, 0x4504, 0x45FC, 0x7C88, 0x4504, 0x202}},
    {38941, 2, {0x2000, 0x10FE, 0x10, 0xFE20, 0xFC, 0x2884, 0x4484, 0x82FC, 0x484, 0x44FC, 0x2884, 0x1084, 0x28FC, 0x4448, 0x8484, 0x102}},
    {38942, 2, {0x1000, 0x8FE, 0x7E10, 0x4220, 0x94FC, 0x1084, 0x1084, 0xFEFC, 0x2484, 0x24FC, 0x4884, 0x2884, 0x10FC, 0x2848, 0x4484, 0x8102}},
    {38943, 2, {0x2000, 0x20FE, 0x3C10, 0x4420, 0xC4FC, 0x2884, 0x1084, 0x28FC, 0x4484, 0x82FC, 0x7C84, 0x4484, 0x44FC, 0x4448, 0x7C84, 0x4502}},
    {38944, 2, {0x1000, 0x107E, 0x3F10, 0x2120, 0x427C, 0xBFC4, 0x2044, 0x2F7C, 0x2944, 0x297C, 0x2D44, 0x2A44, 0x287C, 0x4A00, 0x4C28, 0x8844}},
    {38945, 2, {0x1000, 0x10FE, 0x1010, 0xFE20, 0x10FC, 0x1084, 0x7C84, 0xFC, 0x84, 0x7CFC, 0x4484, 0x4484, 0x44FC, 0x7C48, 0x4484, 0x102}},
    {38946, 2, {0x800, 0x1DFE, 0x7020, 0x1040, 0x11FC, 0xFD04, 0x1104, 0x11FC, 0x7D04, 0x45FC, 0x4504, 0x4504, 0x45FC, 0x7C88, 0x4504, 0x202}},
    {38947, 2, {0x0, 0x7EFE, 0x4810, 0x4820, 0x48FC, 0x7E84, 0x4284, 0x42FC, 0x7E84, 0x48FC, 0x4884, 0x4884, 0x48FC, 0x7E48, 0x84, 0x102}},
    {38948, 2, {0x0, 0x7DFE, 0x4820, 0x4840, 0x49FC, 0x5D04, 0x5504, 0x55FC, 0x5504, 0x5DFC, 0x4904, 0x4904, 0x49FC, 0x7E88, 0x104, 0x202}},
    {38949, 2, {0x4000, 0x5F7E, 0x5410, 0x5420, 0x547C, 0x5F44, 0x5144, 0x517C, 0x5144, 0x5F7C, 0x5444, 0x5444, 0x547C, 0x5400, 0x5F28, 0x4044}},
    {38950, 2, {0x1000, 0x8FE, 0x810, 0xFF20, 0x8FC, 0x1084, 0x2284, 0x7CFC, 0x984, 0x12FC, 0x2484, 0xC884, 0x14FC, 0x2248, 0xC084, 0x102}},
    {38951, 2, {0x800, 0x11FE, 0x7C20, 0x4440, 0x45FC, 0x4504, 0x7D04, 0x41FC, 0x4104, 0x7DFC, 0x4504, 0x4504, 0x45FC, 0x7C88, 0x4504, 0x202}},
    {38952, 2, {0x0, 0xEEFE, 0x2210, 0x2220, 0xAAFC, 0x6684, 0x2284, 0x22FC, 0x2684, 0x6AFC, 0xB284, 0x2284, 0x22FC, 0x2248, 0xAA84, 0x4502}},
    {38953, 2, {0x4200, 0x24FE, 0x10, 0x7E20, 0x24FC, 0x2484, 0x2484, 0x24FC, 0xFF84, 0x24FC, 0x2484, 0x2484, 0x24FC, 0x4448, 0x4484, 0x8502}},
    {38954, 2, {0x800, 0x8FE, 0x4A10, 0x2A20, 0x2CFC, 0x884, 0xFE84, 0x18FC, 0x1C84, 0x2AFC, 0x2A84, 0x4884, 0x88FC, 0x848, 0x884, 0x902}},
    {38955, 2, {0x2800, 0x28FE, 0x2810, 0x2820, 0xAAFC, 0x6C84, 0x2884, 0x28FC, 0x6C84, 0xAAFC, 0x2884, 0x2884, 0x2AFC, 0x4C48, 0x4884, 0x8102}},
    {38956, 2, {0x800, 0x8FE, 0x810, 0x7E20, 0x8FC, 0x4A84, 0x2A84, 0x2CFC, 0x884, 0xFEFC, 0x884, 0x1484, 0x12FC, 0x2248, 0x4084, 0x8102}},
    {38957, 2, {0x0, 0xFEFE, 0x20, 0x40, 0x7CFC, 0x4484, 0x4484, 0x44FC, 0x7C84, 0xFC, 0x4484, 0x2884, 0x2EFC, 0xF000, 0x4048, 0x84}},
    {38958, 2, {0x800, 0x97E, 0xEA10, 0x2C20, 0x4A7C, 0x4944, 0xA844, 0x107C, 0x2444, 0x247C, 0xFF44, 0x2444, 0x247C, 0x4400, 0x4428, 0x8444}},
    {38959, 2, {0x1000, 0x10FE, 0x3E10, 0x6220, 0x94FC, 0x4884, 0x3084, 0xC0FC, 0x2484, 0x24FC, 0xFF84, 0x2484, 0x24FC, 0x4448, 0x4484, 0x8502}},
    {38960, 2, {0x800, 0x8FE, 0x810, 0xFFA0, 0x87C, 0x2A44, 0x2A44, 0x2A7C, 0x5D44, 0x88FC, 0x844, 0x1444, 0x127C, 0x2100, 0x4128, 0x8044}},
    {38961, 2, {0x400, 0x1EFE, 0xE010, 0x2220, 0x92FC, 0x5484, 0x4084, 0x4FC, 0xFF84, 0x4FC, 0x4484, 0x2484, 0x24FC, 0x448, 0x1484, 0x902}},
    {38962, 2, {0x0, 0x1BE, 0x7608, 0x1210, 0x123E, 0x2222, 0x27A2, 0x723E, 0x1222, 0x523E, 0x5222, 0x27A2, 0x203E, 0x5000, 0x8F94, 0x22}},
    {38963, 2, {0x1000, 0x10FE, 0x1010, 0x7C20, 0x10FC, 0x1084, 0xFE84, 0x28FC, 0x2884, 0x6CFC, 0xAA84, 0x2884, 0x28FC, 0x4848, 0x4884, 0x9902}},
    {38964, 2, {0x4000, 0x4CFE, 0x7010, 0x4220, 0x42FC, 0x3E84, 0x84, 0x7CFC, 0x84, 0xFEFC, 0x1084, 0x5484, 0x52FC, 0x9248, 0x5084, 0x2102}},
    {38965, 2, {0x0, 0x7EFE, 0x1210, 0x1220, 0xFF7C, 0x1244, 0x1244, 0x7E7C, 0x2044, 0x207C, 0x7E44, 0x6244, 0xA27C, 0x2200, 0x3E28, 0x2244}},
    {38966, 2, {0x800, 0x28FE, 0x2810, 0x3E20, 0x48FC, 0x884, 0xFF84, 0xFC, 0x84, 0x3EFC, 0x2284, 0x2284, 0x22FC, 0x3E48, 0x2284, 0x102}},
    {38967, 2, {0x800, 0x8FE, 0x1410, 0x2220, 0x51FC, 0x8884, 0x7E84, 0x2FC, 0x484, 0x8FC, 0x7E84, 0x4284, 0x42FC, 0x7E48, 0x4284, 0x102}},
    {38968, 2, {0x0, 0xFEFE, 0x5410, 0x5420, 0xA8FC, 0x5484, 0x5484, 0xFC, 0xFE84, 0x10FC, 0x1084, 0x1084, 0x1EFC, 0xF048, 0x4084, 0x102}},
    {38969, 2, {0xC00, 0x70FE, 0x1010, 0xFE20, 0x10FC, 0x3884, 0x5484, 0x92FC, 0x84, 0x28FC, 0x2884, 0x2884, 0x2AFC, 0x4C48, 0x4884, 0x8102}},
    {38970, 2, {0xC00, 0x70FE, 0x1010, 0xFE20, 0x10FC, 0x3884, 0x5484, 0x92FC, 0x7C84, 0x24FC, 0x2884, 0x2E84, 0x22FC, 0x4248, 0x4A84, 0x8502}},
    {38971, 2, {0x800, 0x8FE, 0x2810, 0x2E20, 0x28FC, 0x2884, 0xFF84, 0xFC, 0x884, 0x2AFC, 0x2A84, 0x4C84, 0x84FC, 0x848, 0x3084, 0xC102}},
    {38972, 2, {0x1000, 0x10FE, 0xFE10, 0x1020, 0x10FC, 0xFE84, 0x9284, 0x92FC, 0xFE84, 0x10FC, 0x3884, 0x5484, 0x92FC, 0x1048, 0x1084, 0x1102}},
    {38973, 2, {0xC00, 0x70FE, 0x1010, 0xFE20, 0x10FC, 0x3884, 0x5484, 0x92FC, 0x84, 0x38FC, 0x2884, 0x2884, 0x2AFC, 0x4C48, 0x4884, 0x8102}},
    {38974, 2, {0xBFE, 0x1040, 0x61FC, 0x904, 0x11FC, 0x6104, 0x5FC, 0x904, 0x31FC, 0xC088, 0x904, 0x2880, 0x2E98, 0x28E2, 0x2E82, 0xF07E}},
    {38975, 2, {0x600, 0x18FE, 0x6210, 0xC20, 0x737C, 0xC44, 0xF444, 0x147C, 0x5544, 0x5E7C, 0x5444, 0x5544, 0x557C, 0x5D00, 0xF328, 0x44}},
    {38976, 2, {0x1400, 0x12FE, 0x2010, 0x3F20, 0x647C, 0xA444, 0x3F44, 0x247C, 0x2444, 0x3F7C, 0x2444, 0x2444, 0x247C, 0x3F00, 0x2028, 0x2044}},
    {38977, 2, {0x1000, 0x87E, 0x7F10, 0x4120, 0x827C, 0x44, 0x7F44, 0x87C, 0x844, 0x287C, 0x2E44, 0x2844, 0x287C, 0x5800, 0x4F28, 0x8044}},
    {38978, 2, {0x1000, 0x10FE, 0x1010, 0xFE20, 0x10FC, 0x5484, 0x5484, 0x54FC, 0xBA84, 0x10FC, 0x3884, 0x5484, 0x92FC, 0x1048, 0x1084, 0x1102}},
    {38979, 2, {0x1000, 0x52FE, 0x5410, 0x9020, 0x28FC, 0x4484, 0x8284, 0x10FC, 0x1084, 0x52FC, 0x5484, 0x9084, 0x28FC, 0x2448, 0x4284, 0x8102}},
    {38980, 2, {0x0, 0x7F7E, 0x210, 0x420, 0x497C, 0x6B44, 0x5D44, 0x497C, 0x5D44, 0x6B7C, 0x4944, 0x6944, 0x517C, 0x4700, 0x7928, 0x44}},
    {38981, 2, {0x1000, 0x8FE, 0x7E10, 0x4220, 0x42FC, 0x7E84, 0x4084, 0x5EFC, 0x5284, 0x5EFC, 0x5284, 0x5E84, 0x52FC, 0x5248, 0x9284, 0x1702}},
    {38982, 2, {0x0, 0x7F7E, 0x4910, 0x4920, 0x7F7C, 0x4944, 0x4944, 0x7F7C, 0x844, 0xFF7C, 0x1C44, 0x2A44, 0x497C, 0x8800, 0x828, 0x844}},
    {38983, 2, {0x1000, 0x8FE, 0x10, 0x7F20, 0x7C, 0x2244, 0x2244, 0x557C, 0x88C4, 0x87C, 0xFF44, 0x844, 0x87C, 0x800, 0x828, 0x844}},
    {38984, 2, {0x4000, 0x4CFE, 0x7010, 0x4220, 0x42FC, 0x3E84, 0x2084, 0x44FC, 0x7884, 0x10FC, 0x2484, 0xFE84, 0x10FC, 0x5448, 0x9284, 0x3102}},
    {38985, 2, {0x1000, 0x10FE, 0x2810, 0x2420, 0x42FC, 0xBC84, 0x1084, 0x10FC, 0xFE84, 0x10FC, 0x9484, 0x5884, 0x50FC, 0x1E48, 0xF084, 0x4102}},
    {38986, 2, {0x1000, 0x20FE, 0x7E10, 0x4220, 0x7EFC, 0x4084, 0x7E84, 0x42FC, 0x7E84, 0x10FC, 0x1084, 0xFE84, 0x10FC, 0x1048, 0x1084, 0x1102}},
    {38987, 2, {0x0, 0x7F7E, 0x4910, 0x4920, 0x7F7C, 0x4944, 0x4944, 0x7F7C, 0x844, 0x47C, 0x1544, 0x50C4, 0x50FC, 0x5280, 0x8E28, 0x44}},
    {38988, 2, {0x0, 0x3EFE, 0x2210, 0x3E20, 0x227C, 0x3E44, 0x7C, 0xFF44, 0x87C, 0x844, 0x2F7C, 0x2828, 0x2844, 0x5882, 0x4FFE, 0x8000}},
    {38989, 2, {0x2000, 0x10FE, 0xFE10, 0x8220, 0x20FC, 0x3C84, 0x4484, 0xA8FC, 0x1084, 0x28FC, 0x4484, 0xFE84, 0x44FC, 0x4448, 0x7C84, 0x4502}},
    {38990, 2, {0x0, 0xEEFE, 0xAA10, 0xAA20, 0xEEFC, 0x84, 0x7C84, 0xFC, 0xFE84, 0x20FC, 0x4084, 0x7C84, 0x4FC, 0x448, 0x2884, 0x1102}},
    {38991, 2, {0x1000, 0x8FE, 0x7F10, 0x2220, 0x1CFC, 0x2284, 0x7F84, 0x44FC, 0x4884, 0x72FC, 0x4484, 0x4884, 0x71FC, 0x4248, 0x8C84, 0x7102}},
    {38992, 2, {0x0, 0xFEFE, 0x9210, 0x1020, 0xFEFC, 0x1084, 0x7C84, 0x54FC, 0x7C84, 0x54FC, 0x7C84, 0x1084, 0xFEFC, 0x1048, 0x1084, 0x1102}},
    {38993, 2, {0x200, 0x2BE, 0x208, 0x7F90, 0x423E, 0x4222, 0x5E22, 0x42BE, 0x42A2, 0x5EBE, 0x56A2, 0x5522, 0x5D7E, 0x4140, 0x42D4, 0x8462}},
    {38994, 2, {0x0, 0x7CFE, 0x5410, 0x5420, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x1084, 0xFEFC, 0x9284, 0x9684, 0xBEFC, 0x8248, 0x8A84, 0x8502}},
    {38995, 2, {0x1000, 0x92FE, 0x9210, 0x9220, 0xFEFC, 0x84, 0xFE84, 0x10FC, 0x2084, 0xFEFC, 0xAA84, 0xAA84, 0xAAFC, 0xAA48, 0xAA84, 0x8702}},
    {38996, 2, {0x1000, 0x8FE, 0x7F10, 0x2220, 0x147C, 0x7F44, 0x4444, 0x487C, 0x5244, 0x447C, 0x4844, 0x5144, 0x427C, 0x4400, 0x8828, 0x3044}},
    {38997, 2, {0x0, 0x7CFE, 0x4410, 0x4420, 0x7CFC, 0x4484, 0x4484, 0x7CFC, 0x2884, 0x28FC, 0xAA84, 0x6C84, 0x28FC, 0x2E48, 0xF084, 0x4102}},
    {38998, 2, {0x800, 0x107E, 0x7F10, 0x4120, 0x557C, 0x4944, 0x5544, 0x417C, 0x7F44, 0x87C, 0x544, 0x54C4, 0x50FC, 0x9200, 0xE28, 0x44}},
    {38999, 2, {0x800, 0x497E, 0x4910, 0x7F20, 0x7C, 0xFF44, 0x44, 0x7F7C, 0x4144, 0x417C, 0x7F44, 0x4244, 0x247C, 0xF00, 0xF028, 0x4044}},
    {39000, 2, {0x0, 0x7F7E, 0x4410, 0x4820, 0x5F7C, 0x5144, 0x5F44, 0x517C, 0x5F44, 0x447C, 0x5644, 0x5544, 0x657C, 0x4500, 0x9428, 0x844}},
    {39001, 2, {0x0, 0x7EFE, 0x2210, 0x1C20, 0x227C, 0x7744, 0x5544, 0x227C, 0x5544, 0x87C, 0x7F44, 0x1C44, 0x2A7C, 0x4900, 0x8828, 0x844}},
    {39002, 2, {0x4600, 0x787E, 0x4090, 0x3FA0, 0x7C, 0x5F44, 0x5144, 0x5F7C, 0x5144, 0x5F7C, 0x5144, 0x5F44, 0x407C, 0x7F80, 0x2228, 0x4144}},
    {39003, 2, {0x1000, 0x10FE, 0xFE10, 0x1020, 0x7CFC, 0x4484, 0x7C84, 0x44FC, 0x7C84, 0x44FC, 0x7C84, 0x4484, 0xFEFC, 0x2848, 0x4484, 0x8302}},
    {39004, 2, {0x2200, 0x223E, 0xFF88, 0x2210, 0x7F3E, 0x2222, 0xFFA2, 0x83E, 0x7F22, 0x493E, 0x7F22, 0x4922, 0xFFBE, 0x4100, 0x4514, 0x4222}},
    {39005, 2, {0x0, 0x7CFE, 0x4410, 0x7420, 0x54FC, 0xFE84, 0x8284, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7C84, 0x44FC, 0x4448, 0x5484, 0x4902}},
    {39006, 2, {0x1000, 0x92FE, 0x5410, 0xFE20, 0x30FC, 0x5484, 0x9284, 0xFC, 0x1484, 0x12FC, 0xFE84, 0x1084, 0x28FC, 0x2448, 0x4484, 0x8102}},
    {39007, 2, {0x0, 0x777E, 0x1110, 0x5520, 0x337C, 0x5544, 0xC44, 0x337C, 0xC4C4, 0x187C, 0x6244, 0xCC4, 0x717C, 0x600, 0x1828, 0xE044}},
    {39008, 2, {0x0, 0xFEFE, 0x2810, 0xFE20, 0xAAFC, 0xAA84, 0xFE84, 0xFC, 0x7C84, 0xFC, 0xFE84, 0x1084, 0x54FC, 0x9248, 0x5084, 0x2102}},
    {39009, 2, {0x1000, 0x8FE, 0x7F10, 0x2220, 0x147C, 0xFF44, 0x1044, 0x297C, 0xCE44, 0x147C, 0x2644, 0xCD44, 0x157C, 0x2400, 0xD428, 0x844}},
    {39010, 2, {0x2200, 0x223E, 0xFF88, 0x2210, 0x223E, 0x3E22, 0x822, 0xFFBE, 0x88A2, 0xCCBE, 0xAAA2, 0xDDA2, 0x88BE, 0x8880, 0x8A94, 0x8922}},
    {39011, 2, {0x200, 0x29E, 0x204, 0x7FC8, 0x421E, 0x4A12, 0x4E12, 0x4A9E, 0x7E92, 0x4A9E, 0x4A92, 0x5D12, 0x6B5E, 0x4940, 0x5ACC, 0x8452}},
    {39012, 2, {0x1000, 0x10FE, 0x7C10, 0x1020, 0xFEFC, 0x4484, 0xEE84, 0x44FC, 0x6484, 0xDEFC, 0x84, 0xFE84, 0x28FC, 0x2A48, 0x4C84, 0x8902}},
    {39013, 2, {0x0, 0x7CFE, 0x4410, 0x7C20, 0x44FC, 0x7C84, 0x1084, 0xFEFC, 0x84, 0x7CFC, 0x4484, 0x7C84, 0x10FC, 0x5448, 0x9284, 0x3102}},
    {39014, 2, {0x2800, 0x24FE, 0x7E10, 0xC820, 0x48FC, 0x7E84, 0x4884, 0x48FC, 0x7E84, 0x48FC, 0x4884, 0x7E84, 0x40FC, 0xAA48, 0xAA84, 0x102}},
    {39015, 2, {0x1000, 0x8FE, 0x7F10, 0x4120, 0x417C, 0x7F44, 0x5444, 0x527C, 0x7F44, 0x547C, 0x5F44, 0x5444, 0x5F7C, 0x5400, 0x9F28, 0x1044}},
    {39016, 2, {0xFEFE, 0x1020, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x2448, 0x4284, 0x0, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {39017, 2, {0x800, 0x8FE, 0x1410, 0x2220, 0x417C, 0xBE44, 0x44, 0x7C, 0x7744, 0x557C, 0x5544, 0x7744, 0x227C, 0x2200, 0x5528, 0x8844}},
    {39018, 2, {0x400, 0x27BE, 0x2408, 0xFFD0, 0x23E, 0x2A2, 0x7FE2, 0x423E, 0x7E22, 0x4ABE, 0x4AA2, 0x5AA2, 0x6D3E, 0x8940, 0x12D4, 0x6462}},
    {39019, 2, {0x1000, 0xFEFE, 0x10, 0xFE20, 0x82FC, 0xBA84, 0xAA84, 0xFEFC, 0x84, 0x7CFC, 0x4484, 0x7C84, 0x44FC, 0x7C48, 0x84, 0xFF02}},
    {39020, 2, {0x0, 0x7F3E, 0x808, 0xFF90, 0x88BE, 0x6B22, 0x822, 0x6B3E, 0x22, 0xFFBE, 0x822, 0x7F22, 0x553E, 0x5500, 0x5514, 0x4322}},
    {39021, 2, {0x2200, 0xFFBE, 0x2208, 0x7F10, 0x553E, 0x5522, 0x7F22, 0x3E, 0xFFA2, 0x90BE, 0x3E22, 0x2222, 0x523E, 0xC00, 0x1814, 0xE022}},
    {39022, 2, {0x1000, 0x87E, 0xFF90, 0x80A0, 0x7F7C, 0x5244, 0x9444, 0x7F7C, 0xC144, 0x7F7C, 0x4144, 0x7F44, 0x417C, 0x7F00, 0x2228, 0x4144}},
    {39023, 2, {0x0, 0xFEFE, 0x8210, 0xFE20, 0x82FC, 0xFE84, 0x4484, 0xAAFC, 0xEE84, 0x44FC, 0xAA84, 0xEE84, 0xFC, 0xAA48, 0xAA84, 0x102}},
    {39024, 2, {0x87E, 0x2E10, 0x287E, 0xFF42, 0x2A7E, 0x4C42, 0x327E, 0xC424, 0x3FFA, 0x2108, 0x3FF8, 0x2208, 0x3FF8, 0x880, 0xFFFE, 0x80}},
    {39025, 2, {0x800, 0xFBE, 0x808, 0x7F90, 0x48BE, 0x7E22, 0x48A2, 0x5FBE, 0x5522, 0x5F3E, 0x5522, 0x5F22, 0x403E, 0xBF80, 0x2A94, 0x7FE2}},
    {39026, 2, {0x2900, 0x4A7E, 0x1410, 0x6220, 0x97C, 0x2A44, 0x5444, 0x227C, 0x4844, 0xFF7C, 0x44, 0x7F44, 0x417C, 0x5D00, 0x5528, 0x7F44}},
    {39027, 2, {0x7F80, 0x213E, 0x3F08, 0x2110, 0x3F3E, 0x21E2, 0xFF22, 0x13E, 0xFFE2, 0x52BE, 0x73A2, 0x52A2, 0x73BE, 0x5AC0, 0xF794, 0x10A2}},
    {39028, 2, {0x2200, 0x227E, 0xFF90, 0x2220, 0x777C, 0x5544, 0x7744, 0x127C, 0x3F44, 0x647C, 0xBF44, 0x2444, 0x3F7C, 0x2400, 0x3F28, 0x2044}},
    {39029, 2, {0x0, 0x7FFC, 0x200, 0x400, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1290, 0x240, 0x420, 0x1810, 0x6008}},
    {39030, 2, {0x0, 0x1FE, 0xFE20, 0x1040, 0x11FC, 0x1104, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1144, 0x1050, 0x5088, 0x2104, 0x202}},
    {39031, 2, {0x0, 0x21FE, 0x2020, 0x2040, 0x21FC, 0x3D04, 0x2124, 0x2124, 0x2124, 0x2524, 0x2924, 0x3144, 0x2050, 0x88, 0x104, 0x202}},
    {39032, 2, {0x0, 0x79FE, 0x1020, 0x1040, 0x11FC, 0x1104, 0x1124, 0xFD24, 0x1124, 0x1124, 0x1124, 0x1144, 0x1050, 0x1088, 0x1104, 0x1202}},
    {39033, 2, {0x0, 0x1FE, 0x20, 0xFC40, 0x11FC, 0x1104, 0x1124, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D44, 0xE050, 0x4088, 0x104, 0x202}},
    {39034, 2, {0x400, 0x45FE, 0x5420, 0x5440, 0x55FC, 0x5504, 0x5524, 0x5524, 0x5524, 0x5524, 0x5524, 0x5544, 0x5450, 0x5488, 0x8504, 0x602}},
    {39035, 2, {0x800, 0x9FE, 0x1020, 0x2040, 0x41FC, 0x8904, 0x924, 0x1124, 0x2124, 0x4524, 0x8524, 0x944, 0x1050, 0x2088, 0x4104, 0x8202}},
    {39036, 2, {0x0, 0x1FE, 0x7C20, 0x1040, 0x11FC, 0x1104, 0x7D24, 0x1124, 0x1124, 0x1124, 0x1124, 0x1D44, 0xE050, 0x4088, 0x104, 0x202}},
    {39037, 2, {0x0, 0x7CFE, 0x20, 0x40, 0xFC, 0xFE84, 0x2894, 0x2894, 0x2894, 0x2894, 0x2894, 0x2AA4, 0x4C30, 0x4848, 0x8084, 0x302}},
    {39038, 2, {0x0, 0x7EFE, 0x4020, 0x4040, 0x5EFC, 0x5284, 0x5294, 0x5294, 0x5A94, 0x5494, 0x5094, 0x52A4, 0x5430, 0x5848, 0x5084, 0x8302}},
    {39039, 2, {0x1000, 0x11FE, 0x1020, 0xFE40, 0x11FC, 0x5504, 0x5524, 0x5524, 0x5524, 0x7D24, 0x1124, 0x1144, 0x1450, 0x1888, 0x1104, 0x202}},
    {39040, 2, {0x800, 0x1CFE, 0x6020, 0x4040, 0x40FC, 0x4084, 0x7E94, 0x4894, 0x4894, 0x4894, 0x4894, 0x48A4, 0x4830, 0x4848, 0x8884, 0xB02}},
    {39041, 2, {0x0, 0x29FE, 0x2820, 0x2440, 0x45FC, 0x4504, 0x8124, 0x7D24, 0x2524, 0x2524, 0x2524, 0x2544, 0x2450, 0x4488, 0x5504, 0x8A02}},
    {39042, 2, {0x0, 0x28FE, 0x2420, 0x2440, 0x42FC, 0x5284, 0x9094, 0x1094, 0x2094, 0x2094, 0x4894, 0x44A4, 0xFC30, 0x4448, 0x84, 0x302}},
    {39043, 2, {0x1000, 0x8FE, 0x820, 0xFF40, 0xFC, 0x84, 0x3C94, 0x2494, 0x2494, 0x2494, 0x2494, 0x25A4, 0x4630, 0x4448, 0x8084, 0x302}},
    {39044, 2, {0x0, 0xF9FE, 0x820, 0x5040, 0x21FC, 0x1104, 0xFD24, 0x2524, 0x2924, 0x2124, 0x2124, 0x2144, 0x2050, 0x2088, 0xA104, 0x4202}},
    {39045, 2, {0x1000, 0x10FE, 0x1E20, 0x1040, 0x10FC, 0x7C84, 0x4494, 0x4494, 0x4494, 0x7C94, 0x4494, 0x40A4, 0x4030, 0x4048, 0x8084, 0x302}},
    {39046, 2, {0x1000, 0x11FE, 0x2820, 0x2440, 0x43FC, 0x9104, 0x924, 0x124, 0xFD24, 0x524, 0x924, 0x5144, 0x2050, 0x1088, 0x1104, 0x202}},
    {39047, 2, {0x800, 0x8FE, 0x810, 0x7F20, 0x497C, 0x4A44, 0x4854, 0x7F54, 0x4154, 0x5254, 0x4A54, 0x4454, 0x4A28, 0x9124, 0x2142, 0x82}},
    {39048, 2, {0x0, 0x7EFE, 0x410, 0x820, 0x187C, 0x2444, 0x4254, 0x8154, 0x54, 0x7E54, 0x854, 0x854, 0x828, 0xE24, 0xF042, 0x4082}},
    {39049, 2, {0x1000, 0x10FE, 0x1020, 0xFE40, 0x10FC, 0x1084, 0x7C94, 0x94, 0x94, 0x7C94, 0x4494, 0x44A4, 0x4430, 0x7C48, 0x4484, 0x302}},
    {39050, 2, {0x800, 0x8FE, 0x820, 0x7E40, 0x8FC, 0x4A84, 0x2A94, 0x2C94, 0x894, 0xFE94, 0x894, 0x14A4, 0x1230, 0x2248, 0x4084, 0x8302}},
    {39051, 2, {0x0, 0x37E, 0xEC10, 0x2420, 0x247C, 0x4444, 0x4F54, 0xE454, 0x2454, 0xA454, 0xA454, 0x4F54, 0x4020, 0xA028, 0x1F44, 0x82}},
    {39052, 2, {0x1000, 0x10FE, 0x2820, 0x2440, 0x42FC, 0x8084, 0x7C94, 0x94, 0x94, 0x7C94, 0x4494, 0x44A4, 0x4430, 0x7C48, 0x4484, 0x302}},
    {39053, 2, {0x2000, 0x24FE, 0x3820, 0x2240, 0x22FC, 0x1E84, 0x94, 0x894, 0xA94, 0xEC94, 0x2894, 0x2CA4, 0x4A30, 0x4A48, 0xA884, 0x1302}},
    {39054, 2, {0x4000, 0x44FE, 0x7820, 0x4240, 0x42FC, 0x3E84, 0x94, 0x1094, 0x5294, 0x5294, 0x5494, 0x90A4, 0x2830, 0x2448, 0x4484, 0x8302}},
    {39055, 2, {0x1000, 0x8FE, 0x820, 0xFF40, 0x8FC, 0x1084, 0x2294, 0x7C94, 0x994, 0x1294, 0x2494, 0xC8A4, 0x1430, 0x2248, 0xC084, 0x302}},
    {39056, 2, {0x0, 0x7DFE, 0x4820, 0x4840, 0x49FC, 0x5D04, 0x5524, 0x5524, 0x5524, 0x5D24, 0x4924, 0x4944, 0x4850, 0x7E88, 0x104, 0x202}},
    {39057, 2, {0x1000, 0x11FE, 0x5020, 0x5C40, 0x51FC, 0x5104, 0xFF24, 0x124, 0x1124, 0x5524, 0x5524, 0x5544, 0x8450, 0x888, 0x3104, 0xC202}},
    {39058, 2, {0x800, 0x9FE, 0xEA10, 0x2C20, 0x4A7C, 0x4944, 0xA854, 0x1054, 0x2454, 0x2454, 0xFF54, 0x2454, 0x2428, 0x4424, 0x4442, 0x8482}},
    {39059, 2, {0xC00, 0x71FE, 0x1020, 0xFE40, 0x11FC, 0x3904, 0x5524, 0x9324, 0x124, 0x3924, 0x2924, 0x2944, 0x2A50, 0x4C88, 0x4904, 0x8202}},
    {39060, 2, {0x800, 0x8FE, 0x1410, 0x2220, 0x517C, 0x8844, 0x7E54, 0x254, 0x454, 0x854, 0x7E54, 0x4254, 0x4228, 0x7E24, 0x4242, 0x82}},
    {39061, 2, {0x4000, 0x4CFE, 0x7020, 0x4240, 0x42FC, 0x3E84, 0x94, 0x7C94, 0x94, 0xFE94, 0x1094, 0x54A4, 0x5230, 0x9248, 0x5084, 0x2302}},
    {39062, 2, {0x4000, 0x4DFE, 0x7020, 0x4440, 0x45FC, 0x3D04, 0x124, 0xD24, 0x7124, 0x1124, 0xFF24, 0x1144, 0x3850, 0x5488, 0x9104, 0x1202}},
    {39063, 2, {0x0, 0x7F7E, 0x4910, 0x4920, 0x7F7C, 0x4944, 0x4954, 0x7F54, 0x854, 0xFF54, 0x1C54, 0x2A54, 0x4920, 0x8828, 0x844, 0x882}},
    {39064, 2, {0x0, 0x3EFE, 0x2210, 0x3E20, 0x227C, 0x3E44, 0x54, 0xFF54, 0x854, 0x854, 0x2F28, 0x2844, 0x2882, 0x5800, 0x4FFE, 0x8000}},
    {39065, 2, {0x0, 0x3EFE, 0x2A10, 0x2A20, 0x3E7C, 0x2A44, 0x2A54, 0x3E54, 0x854, 0x7F54, 0x4954, 0x4B54, 0x5F28, 0x4124, 0x4542, 0x4282}},
    {39066, 2, {0x0, 0xEEFE, 0xAA10, 0xAA20, 0xEE7C, 0x44, 0xFE54, 0x54, 0xFF54, 0x4054, 0x7E54, 0x254, 0x228, 0x224, 0x1442, 0x882}},
    {39067, 2, {0x800, 0x497E, 0x4910, 0x4920, 0x7F7C, 0x44, 0xFFD4, 0x854, 0x1054, 0x7F54, 0x5554, 0x5554, 0x5520, 0x5528, 0x5544, 0x4382}},
    {39068, 2, {0x1000, 0x8FE, 0x7F10, 0x2220, 0x147C, 0x7F44, 0x4454, 0x4854, 0x5254, 0x4454, 0x4854, 0x5154, 0x4228, 0x4424, 0x8842, 0x3082}},
    {39069, 2, {0x1000, 0x8FE, 0x7F10, 0x4120, 0x107C, 0x1E44, 0x2254, 0x5454, 0x854, 0x1454, 0x2254, 0x7F54, 0xA228, 0x2224, 0x3E42, 0x2282}},
    {39070, 2, {0x0, 0x7F7E, 0x2210, 0x3E20, 0x227C, 0x3E44, 0x23D4, 0xFE54, 0x254, 0xF754, 0x1154, 0x5554, 0x2220, 0x5528, 0x8944, 0x1082}},
    {39071, 2, {0x2400, 0x24FE, 0xFF10, 0x2420, 0x7C, 0xFF44, 0x2454, 0x2454, 0xFF54, 0xA554, 0xA554, 0xDB54, 0x9128, 0x8124, 0x8542, 0x8282}},
    {39072, 2, {0x800, 0x8FE, 0x7F10, 0x820, 0x3E7C, 0x2244, 0x3E54, 0x2254, 0x3E54, 0x2254, 0x3E54, 0x2254, 0x7F28, 0x1424, 0x2242, 0x4182}},
    {39073, 2, {0x0, 0x7EFE, 0x2210, 0x1C20, 0x227C, 0x7744, 0x5554, 0x2254, 0x5554, 0x854, 0x7F54, 0x1C54, 0x2A28, 0x4924, 0x8842, 0x882}},
    {39074, 2, {0x0, 0x7F7E, 0x4110, 0x7F20, 0x417C, 0x7F44, 0x854, 0xFFD4, 0x54, 0x7F54, 0x4154, 0x7F54, 0x820, 0x4928, 0x88C4, 0x1882}},
    {39075, 2, {0x9200, 0x54FE, 0xFE10, 0x3020, 0x547C, 0x9244, 0x2054, 0x4454, 0x7854, 0x1054, 0x2454, 0xFE54, 0x1028, 0x5424, 0x9242, 0x3082}},
    {39076, 2, {0x800, 0xFF7E, 0x10, 0xFF20, 0x817C, 0xBD44, 0xA554, 0xFF54, 0x54, 0x7E54, 0x4254, 0x7E54, 0x4220, 0x7E28, 0x44, 0xFE82}},
    {39077, 2, {0x0, 0x7F7E, 0x810, 0xFFA0, 0x88FC, 0x6B44, 0x854, 0x6B54, 0x54, 0xFFD4, 0x854, 0x7F54, 0x5520, 0x5528, 0x5544, 0x4382}},
    {39078, 2, {0x800, 0x2EFE, 0x2820, 0xFF7C, 0x2A54, 0x4C54, 0x3228, 0xC444, 0x3FF8, 0x2108, 0x3FF8, 0x2208, 0x3FF8, 0x880, 0xFFFE, 0x80}},
    {39079, 2, {0x2200, 0x227E, 0xFF90, 0x2220, 0x777C, 0x5544, 0x7754, 0x1254, 0x3F54, 0x6454, 0xBF54, 0x2454, 0x3F20, 0x2428, 0x3F44, 0x2082}},
    {39080, 2, {0x0, 0x3FF8, 0x2008, 0x20E8, 0x2F08, 0x2108, 0x2FE8, 0x2928, 0x2928, 0x2FE8, 0x2908, 0x214A, 0x21EA, 0x5E2A, 0x4806, 0x8002}},
    {39081, 2, {0x2, 0x7FC2, 0x4044, 0x4348, 0x5C52, 0x4442, 0x5F44, 0x5548, 0x5550, 0x5F42, 0x4444, 0x4548, 0x5F50, 0x4922, 0x801E, 0x0}},
    {39082, 2, {0x0, 0x7FFC, 0x100, 0x160, 0x110, 0x100, 0x1FF0, 0x10D0, 0x1710, 0x1110, 0x17D0, 0x1550, 0x27CA, 0x212A, 0x4FE6, 0x8022}},
    {39083, 2, {0x8, 0x7FC8, 0x4048, 0x435C, 0x5C48, 0x4448, 0x5F7E, 0x5548, 0x5548, 0x5F54, 0x4452, 0x4562, 0x5F40, 0x4922, 0x801E, 0x0}},
    {39084, 2, {0x0, 0x7FDE, 0x4044, 0x4354, 0x5C54, 0x4464, 0x5F7E, 0x5544, 0x554C, 0x5F54, 0x4464, 0x4554, 0x5F48, 0x4922, 0x801E, 0x0}},
    {39085, 2, {0x8, 0x7FC8, 0x4048, 0x434E, 0x5C48, 0x4448, 0x5F48, 0x555C, 0x5554, 0x5F54, 0x4454, 0x455C, 0x5F54, 0x4922, 0x801E, 0x0}},
    {39086, 2, {0x40, 0x7F40, 0x417C, 0x4D84, 0x7174, 0x5154, 0x7D54, 0x5574, 0x7D48, 0x5142, 0x5542, 0x7D3E, 0x4480, 0x4042, 0x803E, 0x0}},
    {39087, 2, {0x8000, 0x4FF8, 0x4808, 0x868, 0xEB88, 0x888, 0xA888, 0xABE8, 0xAAA8, 0xAAA8, 0x4BE8, 0x388A, 0xC8AA, 0x10EA, 0x1726, 0x2002}},
    {39088, 2, {0x8, 0x7FCA, 0x4048, 0x437E, 0x5C48, 0x4450, 0x5F5E, 0x5552, 0x556A, 0x5F44, 0x444A, 0x4552, 0x5F40, 0x4922, 0x801E, 0x0}},
    {39089, 2, {0x8, 0x7FC8, 0x4048, 0x4352, 0x5C5E, 0x4442, 0x5F40, 0x555E, 0x5552, 0x5F52, 0x445E, 0x4552, 0x5F40, 0x4922, 0x801E, 0x0}},
    {39090, 2, {0x2, 0x7DF2, 0x4482, 0x4C8A, 0x748A, 0x55EA, 0x552A, 0x7D2A, 0x56AA, 0x7C4A, 0x544A, 0x5482, 0x530A, 0x7D04, 0x4482, 0x807E}},
    {39091, 2, {0x4, 0x7FD8, 0x4048, 0x4348, 0x5C7E, 0x4448, 0x5F48, 0x555C, 0x5554, 0x5F54, 0x4454, 0x455C, 0x5F40, 0x4922, 0x801E, 0x0}},
    {39092, 2, {0x20, 0x7F20, 0x413E, 0x4D40, 0x7180, 0x517E, 0x7D12, 0x5550, 0x7D5C, 0x5150, 0x5550, 0x7DBE, 0x4480, 0x4042, 0x803E, 0x0}},
    {39093, 2, {0x10, 0x7F54, 0x4138, 0x4D10, 0x717C, 0x5144, 0x7D7C, 0x5544, 0x7D7C, 0x5144, 0x5544, 0x7D4C, 0x4480, 0x4042, 0x803E, 0x0}},
    {39094, 2, {0x0, 0x7F7C, 0x4144, 0x4D7C, 0x7144, 0x517C, 0x7D44, 0x557C, 0x7D44, 0x51FE, 0x5500, 0x7D28, 0x44C4, 0x4042, 0x803E, 0x0}},
    {39095, 2, {0x10, 0x7F52, 0x4194, 0x4D28, 0x7144, 0x5184, 0x7D10, 0x5512, 0x7D54, 0x51A8, 0x5524, 0x7D44, 0x4480, 0x4042, 0x803E, 0x0}},
    {39096, 2, {0x0, 0x7F7C, 0x4154, 0x4D7C, 0x7154, 0x517C, 0x7D00, 0x5524, 0x7D12, 0x5160, 0x55A4, 0x7D1C, 0x4480, 0x4042, 0x803E, 0x0}},
    {39097, 2, {0x10, 0x7F7C, 0x4124, 0x4DFE, 0x7100, 0x517C, 0x7D44, 0x557C, 0x7D10, 0x513C, 0x5550, 0x7D7E, 0x4490, 0x4052, 0x803E, 0x0}},
    {39098, 2, {0x7C, 0x7F44, 0x417C, 0x4D44, 0x717C, 0x5100, 0x7DFE, 0x5540, 0x7D7E, 0x51AA, 0x5552, 0x7D2A, 0x4484, 0x4042, 0x803E, 0x0}},
    {39099, 2, {0xE00, 0xF1FC, 0x2304, 0x9334, 0x45C4, 0x2144, 0x3DF4, 0x5154, 0x1154, 0xFFF4, 0x1144, 0x5554, 0x55F4, 0x5D16, 0x6206, 0x402}},
    {39100, 2, {0xA0, 0x7D2C, 0x4524, 0x4DAC, 0x7524, 0x55FC, 0x5420, 0x7DFC, 0x5404, 0x7CC8, 0x5430, 0x5448, 0x5384, 0x7D00, 0x4482, 0x807E}},
    {39101, 2, {0x1000, 0x55FC, 0x5504, 0x5534, 0x7DC4, 0x144, 0x7DF4, 0x154, 0x7D54, 0x45F4, 0x7D44, 0x4554, 0x29F4, 0x1D16, 0xE206, 0x4402}},
    {39102, 2, {0x0, 0x7DF8, 0x44A8, 0x4C88, 0x7450, 0x5420, 0x54D8, 0x7F26, 0x54F8, 0x7CA8, 0x54F8, 0x5424, 0x53FC, 0x7D04, 0x4482, 0x807E}},
    {39103, 2, {0x0, 0x7DFC, 0x5104, 0x7D34, 0x51C4, 0x7D44, 0x51F4, 0x5154, 0x7D54, 0x5F4, 0x5544, 0x5554, 0x55F4, 0x8516, 0x1606, 0xC02}},
    {39104, 2, {0x40, 0x7DBC, 0x4514, 0x4D54, 0x75A4, 0x552C, 0x5440, 0x7DFC, 0x5524, 0x7DFC, 0x5524, 0x55FC, 0x5200, 0x7D00, 0x4482, 0x807E}},
    {39105, 2, {0x0, 0xFDFC, 0x2504, 0xB534, 0x6DC4, 0xB544, 0x25F4, 0x1154, 0x7D54, 0x45F4, 0x4544, 0x7D54, 0x45F4, 0x4516, 0x7E06, 0x4402}},
    {39106, 2, {0x0, 0x7DDC, 0x4444, 0x4D54, 0x74CC, 0x5554, 0x5420, 0x7C50, 0x54A8, 0x7D54, 0x54A0, 0x5448, 0x5290, 0x7D20, 0x44C2, 0x807E}},
    {39107, 2, {0x0, 0x7DFC, 0x4450, 0x4DFC, 0x7554, 0x55FC, 0x5400, 0x7CF8, 0x5400, 0x7DFC, 0x54A8, 0x5524, 0x5260, 0x7D00, 0x4482, 0x807E}},
    {39108, 2, {0x0, 0xFEF8, 0x2888, 0xFE98, 0xAAE8, 0xAAA8, 0xFEF8, 0xA8, 0x7CF8, 0xA8, 0xFEA8, 0x10BA, 0x54CA, 0x928A, 0x5106, 0x2202}},
    {39109, 2, {0x0, 0x7DFE, 0x4450, 0x4DDC, 0x7554, 0x55DC, 0x5450, 0x7DFC, 0x5524, 0x7DFC, 0x5524, 0x55FC, 0x5200, 0x7D00, 0x4482, 0x807E}},
    {39110, 2, {0x1400, 0x11FC, 0xFF04, 0x2934, 0x45C4, 0x8344, 0x25F4, 0x3754, 0x2554, 0xFFF4, 0x2544, 0x5554, 0x4BF4, 0x8916, 0x1206, 0x402}},
    {39111, 2, {0x28, 0x7C20, 0x45FC, 0x4C50, 0x7488, 0x5504, 0x5448, 0x7C6C, 0x5448, 0x7DFE, 0x5448, 0x54A8, 0x5294, 0x7D12, 0x44A2, 0x807E}},
    {39112, 2, {0x20, 0x7CA4, 0x44A4, 0x4D28, 0x7450, 0x5488, 0x5504, 0x7C88, 0x56AA, 0x7DDC, 0x5488, 0x5554, 0x5222, 0x7D00, 0x4482, 0x807E}},
    {39113, 2, {0x20, 0x7C20, 0x45FC, 0x4D54, 0x7488, 0x55FC, 0x568A, 0x7CF8, 0x5488, 0x7CF8, 0x54A8, 0x5524, 0x5260, 0x7D00, 0x4482, 0x807E}},
    {39114, 2, {0x800, 0x2AF8, 0x2C88, 0x4898, 0x14E8, 0x22A8, 0x41F8, 0x24A8, 0x2DF8, 0xB6A8, 0x64A8, 0x24BA, 0x54CA, 0x4A8A, 0x9106, 0x202}},
    {39115, 2, {0x0, 0x7DDE, 0x4484, 0x4DCE, 0x7484, 0x55DE, 0x5420, 0x7C18, 0x5448, 0x7CD4, 0x5562, 0x5448, 0x53B8, 0x7D00, 0x4482, 0x807E}},
    {39116, 2, {0x4400, 0xFEF8, 0x4488, 0x98, 0xEEE8, 0xAAA8, 0xEEF8, 0x24A8, 0x7EF8, 0xC8A8, 0x7EA8, 0x48BA, 0x7ECA, 0x488A, 0x7F06, 0x4202}},
    {39117, 2, {0x1FF0, 0x10D0, 0x1710, 0x1110, 0x17D0, 0x1550, 0x27CA, 0x212A, 0x4FE6, 0x8022, 0x7CF8, 0x54A8, 0x7CF8, 0x54AA, 0x7DFE, 0x8202}},
    {39118, 2, {0x0, 0x3FF0, 0x2010, 0x2010, 0x2850, 0x2450, 0x2290, 0x2290, 0x2110, 0x2110, 0x2290, 0x2292, 0x244A, 0x484A, 0x4006, 0x8002}},
    {39119, 2, {0x0, 0x7DF8, 0x4410, 0x4420, 0x44C0, 0x4DFC, 0x6C54, 0x5494, 0x5524, 0x5444, 0x6C84, 0x4528, 0x4410, 0x4202, 0x41FE, 0x8000}},
    {39120, 2, {0x10, 0x3F10, 0x211E, 0x2110, 0x2510, 0x357C, 0x2944, 0x2944, 0x2944, 0x297C, 0x3544, 0x2500, 0x2082, 0x4062, 0x401E, 0x8000}},
    {39121, 2, {0x40, 0x7C40, 0x44FC, 0x4504, 0x46F4, 0x4C94, 0x6CF4, 0x5484, 0x5494, 0x548A, 0x6C82, 0x447E, 0x4400, 0x4202, 0x41FE, 0x8000}},
    {39122, 2, {0x2000, 0x13F8, 0x1208, 0xFE08, 0x328, 0xAA8, 0x8AA8, 0x8A48, 0x4A48, 0x4AA8, 0x52A8, 0x532A, 0x1C0A, 0xE40A, 0x4806, 0x1002}},
    {39123, 2, {0x0, 0x7CF8, 0x4488, 0x44F8, 0x4488, 0x4CF8, 0x6C88, 0x54F8, 0x5488, 0x55FC, 0x6C00, 0x4450, 0x4488, 0x4202, 0x41FE, 0x8000}},
    {39124, 2, {0x0, 0x7DFC, 0x4524, 0x45FC, 0x4524, 0x4DFC, 0x6C20, 0x5414, 0x5552, 0x5542, 0x6E48, 0x4438, 0x4400, 0x4202, 0x41FE, 0x8000}},
    {39125, 2, {0xA0, 0x7D2C, 0x4524, 0x45AC, 0x4524, 0x4DFC, 0x6C20, 0x55FC, 0x5404, 0x54C8, 0x6C30, 0x4448, 0x4584, 0x4202, 0x41FE, 0x8000}},
    {39126, 2, {0xE00, 0xF0F8, 0x2288, 0x9288, 0x4488, 0x2098, 0x3CD8, 0x50A8, 0x10A8, 0xFEA8, 0x10D8, 0x548A, 0x548A, 0x5D0A, 0x6506, 0x202}},
    {39127, 2, {0x40, 0x7DBC, 0x4514, 0x4554, 0x45A4, 0x4D2C, 0x6C40, 0x55FC, 0x5524, 0x55FC, 0x6D24, 0x45FC, 0x4400, 0x4202, 0x41FE, 0x8000}},
    {39128, 2, {0x0, 0xFEF8, 0x2888, 0xFE88, 0xAA88, 0xAA98, 0xFED8, 0xA8, 0x7CA8, 0xA8, 0xFED8, 0x108A, 0x548A, 0x930A, 0x5106, 0x2202}},
    {39129, 2, {0x0, 0x14F8, 0x1088, 0xFE88, 0x2888, 0x4498, 0x82D8, 0x24A8, 0x36A8, 0x24A8, 0xFFD8, 0x248A, 0x548A, 0x4B0A, 0x8906, 0x1202}},
    {39130, 2, {0x20, 0x7CA4, 0x44A4, 0x4528, 0x4450, 0x4C88, 0x6D04, 0x5488, 0x56AA, 0x55DC, 0x6C88, 0x4554, 0x4622, 0x4202, 0x41FE, 0x8000}},
    {39131, 2, {0x0, 0x7FE0, 0x24, 0xA28, 0x7224, 0x1220, 0x1212, 0x120A, 0xFFC6, 0x1248, 0x1250, 0x1248, 0x2240, 0x2224, 0x4214, 0x820C}},
    {39132, 2, {0x800, 0x1DFA, 0xF00C, 0x544A, 0x3988, 0xFEAA, 0x34A6, 0x52A2, 0x81F8, 0x7CAA, 0x54AC, 0x54AA, 0x7CA8, 0x54AA, 0x54A6, 0x7D22}},
    {39133, 2, {0x4, 0x7FE8, 0x824, 0x7212, 0x120E, 0xFFD0, 0x122A, 0x221E, 0xFDF8, 0x2448, 0xD5AA, 0x56AC, 0xFDFA, 0x54A8, 0x55AA, 0x9326}},
    {39134, 2, {0x0, 0xFFC0, 0x40, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x20, 0x20, 0x12, 0xA, 0x6, 0x2}},
    {39135, 2, {0x100, 0x280, 0x440, 0xA20, 0x3118, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1008, 0x1190, 0x1060, 0x1210, 0x1408, 0x1804}},
    {39136, 2, {0x800, 0x800, 0x1400, 0x1200, 0x2900, 0x7E00, 0xA200, 0x3E00, 0x2200, 0x3E00, 0x2000, 0x2800, 0x2400, 0x2A00, 0x3200, 0x2000}},
    {39137, 2, {0x40, 0x4040, 0x20A0, 0x2110, 0x288, 0xC46, 0xBF8, 0x1208, 0x13F8, 0xE208, 0x23F8, 0x2244, 0x2228, 0x2290, 0x2308, 0x204}},
    {39138, 2, {0x800, 0x878, 0x1448, 0x1248, 0x2948, 0x7E48, 0xA248, 0x3E48, 0x2248, 0x3E48, 0x2048, 0x284A, 0x244A, 0x2A8A, 0x3286, 0x2100}},
    {39139, 2, {0x800, 0x800, 0x14FE, 0x1210, 0x2910, 0x7E10, 0xA210, 0x3E10, 0x2210, 0x3E10, 0x2010, 0x2810, 0x2410, 0x2A10, 0x3250, 0x2020}},
    {39140, 2, {0x820, 0x820, 0x1420, 0x1220, 0x2920, 0x7E20, 0xA220, 0x3E20, 0x2250, 0x3E50, 0x2050, 0x2850, 0x2448, 0x2A88, 0x3284, 0x2102}},
    {39141, 2, {0x808, 0x81C, 0x14E0, 0x1220, 0x2920, 0x7E20, 0xA220, 0x3E3E, 0x22E0, 0x3E20, 0x2020, 0x2822, 0x2422, 0x2A22, 0x321E, 0x2000}},
    {39142, 2, {0x800, 0x87C, 0x1410, 0x1210, 0x2910, 0x7E10, 0xA210, 0x3EFE, 0x2210, 0x3E10, 0x2010, 0x2810, 0x2410, 0x2A10, 0x3210, 0x2010}},
    {39143, 2, {0x2020, 0x2020, 0x2050, 0x3C48, 0x24A4, 0x45FA, 0x4688, 0xA4F8, 0x1488, 0x8F8, 0x8A2, 0x1094, 0x1088, 0x20A4, 0x40C4, 0x8080}},
    {39144, 2, {0x1020, 0x1020, 0x1050, 0x2448, 0x24A4, 0x79FA, 0xA88, 0x10F8, 0x2488, 0x7CF8, 0x4A2, 0x894, 0x888, 0x10A4, 0x20C4, 0x4080}},
    {39145, 2, {0x1020, 0x1020, 0x2820, 0x25FC, 0x5220, 0xFCA8, 0x44A8, 0x7CA8, 0x44A8, 0x7CA8, 0x40F8, 0x5022, 0x4822, 0x5422, 0x641E, 0x4000}},
    {39146, 2, {0x800, 0x808, 0x141C, 0x1270, 0x2910, 0x7E10, 0xA210, 0x3E10, 0x22FE, 0x3E10, 0x2010, 0x2810, 0x2410, 0x2A10, 0x327C, 0x2000}},
    {39147, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x5220, 0xFC20, 0x45FE, 0x7C20, 0x4420, 0x7C50, 0x4050, 0x5050, 0x4888, 0x5488, 0x6504, 0x4202}},
    {39148, 2, {0x100, 0x1110, 0x920, 0x3FF8, 0x200, 0x7FFC, 0x820, 0x1110, 0x2FE8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39149, 2, {0x1040, 0x1040, 0x2840, 0x24FE, 0x5280, 0xFD20, 0x4420, 0x7DFC, 0x4424, 0x7C24, 0x4044, 0x5044, 0x4844, 0x5484, 0x64A8, 0x4110}},
    {39150, 2, {0x820, 0x820, 0x1420, 0x127E, 0x2842, 0x4484, 0xBE10, 0x2210, 0x3E10, 0x2210, 0x3E28, 0x2028, 0x3E48, 0x2044, 0x3E84, 0x2102}},
    {39151, 2, {0x1004, 0x100E, 0x28F0, 0x2480, 0x5280, 0xFCFC, 0x44A4, 0x7CA4, 0x44A4, 0x7CA8, 0x40A8, 0x5090, 0x4910, 0x5528, 0x6644, 0x4082}},
    {39152, 2, {0x820, 0x810, 0x1410, 0x1200, 0x29FE, 0x7E20, 0xA220, 0x3E30, 0x2228, 0x3E28, 0x2024, 0x2824, 0x2420, 0x2A20, 0x3220, 0x2020}},
    {39153, 2, {0x20, 0x20, 0xFC50, 0x2088, 0x2144, 0x3E22, 0x45FC, 0x4504, 0x65FC, 0x9504, 0x9FC, 0x922, 0x1114, 0x2148, 0x4184, 0x8102}},
    {39154, 2, {0x1040, 0x1040, 0x2840, 0x247C, 0x5284, 0xFC88, 0x4520, 0x7C20, 0x4420, 0x7C50, 0x4050, 0x5050, 0x4888, 0x5488, 0x6504, 0x4202}},
    {39155, 2, {0x1020, 0x1010, 0x2800, 0x24FE, 0x5210, 0xFC10, 0x4410, 0x7C10, 0x44FE, 0x7C10, 0x4010, 0x5010, 0x4810, 0x5410, 0x65FE, 0x4000}},
    {39156, 2, {0x1020, 0x1020, 0x2820, 0x2440, 0x5248, 0xFC84, 0x45FE, 0x7C82, 0x4400, 0x7CFC, 0x4084, 0x5084, 0x4884, 0x5484, 0x64FC, 0x4084}},
    {39157, 2, {0x840, 0x840, 0x1440, 0x127E, 0x29A0, 0x7EA0, 0xA320, 0x3E3C, 0x2220, 0x3E20, 0x2020, 0x283E, 0x2420, 0x2A20, 0x3220, 0x2020}},
    {39158, 2, {0x1040, 0x1020, 0x2828, 0x2408, 0x5208, 0xFC48, 0x4454, 0x7D52, 0x4552, 0x7D60, 0x4260, 0x5044, 0x48C4, 0x5544, 0x663C, 0x4000}},
    {39159, 2, {0x800, 0x87C, 0x1444, 0x1244, 0x2944, 0x7E7C, 0xA244, 0x3E44, 0x2244, 0x3E7C, 0x2044, 0x2844, 0x2444, 0x2A44, 0x32FE, 0x2000}},
    {39160, 2, {0x20, 0x7C20, 0x4450, 0x4448, 0x44A4, 0x7DFA, 0x288, 0xFEF8, 0x2088, 0x40F8, 0x7CA2, 0x494, 0x488, 0x4A4, 0x28C4, 0x1080}},
    {39161, 2, {0x1000, 0x1080, 0x2B1E, 0x2652, 0x5252, 0xFA52, 0x4A52, 0x7A52, 0x4A52, 0x7AD2, 0x435A, 0x5254, 0x4890, 0x5890, 0x6910, 0x4210}},
    {39162, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF17C, 0x280, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39163, 2, {0x1020, 0x1020, 0x2850, 0x2488, 0x5304, 0xFE12, 0x4420, 0x7C40, 0x4588, 0x7C10, 0x4020, 0x5044, 0x4988, 0x5410, 0x6460, 0x4380}},
    {39164, 2, {0x800, 0x8FE, 0x1402, 0x1202, 0x29FA, 0x7E02, 0xA202, 0x3EFA, 0x228A, 0x3E8A, 0x20FA, 0x288A, 0x2402, 0x2A02, 0x3214, 0x2008}},
    {39165, 2, {0x840, 0x840, 0x14FC, 0x1284, 0x2904, 0x7EF4, 0xA294, 0x3E94, 0x22F4, 0x3E84, 0x2094, 0x2888, 0x2482, 0x2A82, 0x327E, 0x2000}},
    {39166, 2, {0x1080, 0x1080, 0x2880, 0x24FE, 0x5320, 0xFE20, 0x4420, 0x7DFC, 0x4524, 0x7D24, 0x4124, 0x5124, 0x4934, 0x5528, 0x6420, 0x4020}},
    {39167, 2, {0x810, 0x810, 0x1492, 0x1292, 0x2992, 0x7E92, 0xA2FE, 0x3E10, 0x2210, 0x3E92, 0x2092, 0x2892, 0x2492, 0x2A92, 0x32FE, 0x2002}},
    {39168, 2, {0x1088, 0x1088, 0x29FE, 0x2488, 0x5288, 0xFC00, 0x4504, 0x7D04, 0x4488, 0x7C88, 0x4050, 0x5020, 0x4850, 0x5488, 0x6504, 0x4202}},
    {39169, 2, {0x1040, 0x1044, 0x284E, 0x24B8, 0x5288, 0xFD88, 0x4688, 0x7C88, 0x44FE, 0x7C88, 0x4088, 0x5088, 0x4888, 0x5488, 0x64BE, 0x4080}},
    {39170, 2, {0x1008, 0x103C, 0x29E0, 0x2420, 0x5220, 0xFDFE, 0x4420, 0x7C20, 0x4420, 0x7DFC, 0x4104, 0x5104, 0x4904, 0x5504, 0x65FC, 0x4104}},
    {39171, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x5200, 0xFC88, 0x4504, 0x7E02, 0x4488, 0x7C88, 0x4050, 0x5050, 0x4820, 0x5450, 0x6488, 0x4306}},
    {39172, 2, {0x810, 0x810, 0x1428, 0x1228, 0x2944, 0x7E82, 0xA27C, 0x3E00, 0x2200, 0x3E7C, 0x2044, 0x2844, 0x2444, 0x2A44, 0x327C, 0x2044}},
    {39173, 2, {0x884, 0x848, 0x1400, 0x12FC, 0x2948, 0x7E48, 0xA248, 0x3E48, 0x23FE, 0x3E48, 0x2048, 0x2848, 0x2448, 0x2A88, 0x3288, 0x2108}},
    {39174, 2, {0x1050, 0x1050, 0x2850, 0x2552, 0x52D4, 0xFC58, 0x4450, 0x7C58, 0x44D4, 0x7D52, 0x4050, 0x5050, 0x4892, 0x5492, 0x6512, 0x420E}},
    {39175, 2, {0x800, 0x8FE, 0x1482, 0x1282, 0x29BA, 0x7E82, 0xA282, 0x3EBA, 0x22AA, 0x3EAA, 0x20AA, 0x28BA, 0x2482, 0x2A82, 0x328A, 0x2084}},
    {39176, 2, {0x4100, 0x21FC, 0xA44, 0x1440, 0xE0B0, 0x2108, 0x2284, 0x2D60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39177, 2, {0x810, 0x810, 0x1420, 0x12FE, 0x2982, 0x7E82, 0xA2BA, 0x3EAA, 0x22AA, 0x3EAA, 0x20AA, 0x28BA, 0x2482, 0x2A82, 0x328A, 0x2084}},
    {39178, 2, {0x820, 0x440, 0x7FFC, 0x100, 0x3FF8, 0x400, 0xFFFE, 0x1110, 0x2FE8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39179, 2, {0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1110, 0x22C8, 0xCD36, 0x3FE8, 0xC824, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39180, 2, {0x800, 0x8FE, 0x1444, 0x1244, 0x2944, 0x7E7C, 0xA244, 0x3E44, 0x227C, 0x3E44, 0x2044, 0x284E, 0x24F4, 0x2A04, 0x3204, 0x2004}},
    {39181, 2, {0x0, 0x3FFE, 0x2108, 0x3FFC, 0x2410, 0x2888, 0x3364, 0x2C98, 0x37F6, 0x2410, 0x27F0, 0x2410, 0x27F4, 0x4468, 0x4518, 0x8604}},
    {39182, 2, {0x840, 0x840, 0x1478, 0x1288, 0x2950, 0x7E20, 0xA250, 0x3E88, 0x2306, 0x3EF8, 0x2088, 0x2888, 0x2488, 0x2A88, 0x32F8, 0x2088}},
    {39183, 2, {0x1040, 0x1020, 0x2820, 0x25FE, 0x5240, 0xFC40, 0x44A0, 0x7CA2, 0x45A4, 0x7E98, 0x4490, 0x5088, 0x4888, 0x54A4, 0x64C2, 0x4080}},
    {39184, 2, {0x800, 0xF78, 0x2808, 0x7F28, 0xA110, 0x1728, 0x3AC4, 0xCD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39185, 2, {0x1020, 0x1020, 0x29FE, 0x2420, 0x5220, 0xFDFE, 0x4502, 0x7E04, 0x44F8, 0x7C10, 0x4020, 0x51FE, 0x4820, 0x5420, 0x64A0, 0x4040}},
    {39186, 2, {0x1004, 0x101E, 0x29E0, 0x2422, 0x5312, 0xFC94, 0x4480, 0x7C20, 0x45FE, 0x7C44, 0x4084, 0x51C8, 0x4830, 0x5428, 0x64C4, 0x4302}},
    {39187, 2, {0x1010, 0x10D8, 0x2B94, 0x2494, 0x5090, 0xFBFE, 0x4890, 0x7894, 0x4894, 0x78D8, 0x4398, 0x5090, 0x48AA, 0x58CA, 0x6A86, 0x4102}},
    {39188, 2, {0x1028, 0x1024, 0x2BFE, 0x2420, 0x5220, 0xFDFC, 0x4524, 0x7D24, 0x45FC, 0x7D24, 0x4124, 0x51FC, 0x4924, 0x5524, 0x6524, 0x410C}},
    {39189, 2, {0x810, 0x810, 0x1420, 0x1244, 0x29FE, 0x7E28, 0xA244, 0x3EA2, 0x223C, 0x3E44, 0x2044, 0x28A8, 0x2410, 0x2A28, 0x3244, 0x2082}},
    {39190, 2, {0x1000, 0x11FE, 0x2800, 0x2400, 0x52FC, 0xFC84, 0x4484, 0x7C84, 0x44FC, 0x7C00, 0x4084, 0x5044, 0x4848, 0x5400, 0x65FE, 0x4000}},
    {39191, 2, {0x1020, 0x1020, 0x2BFE, 0x2420, 0x5220, 0xFDFC, 0x4524, 0x7D24, 0x45FC, 0x7C20, 0x4070, 0x50A8, 0x4924, 0x5622, 0x6420, 0x4020}},
    {39192, 2, {0x810, 0x810, 0x1428, 0x1228, 0x2944, 0x7EBA, 0xA210, 0x3E10, 0x22FE, 0x3E10, 0x2058, 0x2854, 0x2492, 0x2B12, 0x3250, 0x2020}},
    {39193, 2, {0x10C2, 0x1034, 0x2818, 0x2464, 0x5382, 0xFC20, 0x45FE, 0x7C50, 0x4490, 0x7DFE, 0x4292, 0x5092, 0x489A, 0x5494, 0x6410, 0x4010}},
    {39194, 2, {0x1184, 0x1068, 0x2830, 0x24C8, 0x5324, 0xFC20, 0x47FE, 0x7C40, 0x44FC, 0x7D84, 0x42FC, 0x5084, 0x48FC, 0x5484, 0x6494, 0x4088}},
    {39195, 2, {0x1000, 0x11FC, 0x2904, 0x2504, 0x53FC, 0xFD04, 0x4504, 0x7DFC, 0x4400, 0x7D12, 0x41D4, 0x5118, 0x4910, 0x5552, 0x6592, 0x410E}},
    {39196, 2, {0x1000, 0x11FC, 0x2924, 0x2524, 0x53FC, 0xFD24, 0x4524, 0x7DFC, 0x4420, 0x7DFE, 0x4070, 0x50A8, 0x4924, 0x5622, 0x6420, 0x4020}},
    {39197, 2, {0x1084, 0x1084, 0x29FE, 0x2484, 0x52A4, 0xFC10, 0x45FE, 0x7C40, 0x4440, 0x7C7C, 0x4044, 0x5044, 0x4844, 0x5484, 0x6494, 0x4108}},
    {39198, 2, {0x1048, 0x1044, 0x285E, 0x25E0, 0x5228, 0xFC12, 0x446A, 0x7D96, 0x4448, 0x7C5E, 0x41E0, 0x5024, 0x4828, 0x5412, 0x646A, 0x4186}},
    {39199, 2, {0x1000, 0x13DE, 0x2842, 0x254A, 0x5084, 0xF94A, 0x4A52, 0x7C20, 0x4800, 0x7BDE, 0x4052, 0x5152, 0x4894, 0x5948, 0x6A54, 0x4422}},
    {39200, 2, {0x1044, 0x1044, 0x2888, 0x25DC, 0x5288, 0xFC88, 0x4488, 0x7C88, 0x45DE, 0x7C88, 0x4088, 0x5088, 0x4888, 0x5488, 0x6488, 0x4108}},
    {39201, 2, {0x1040, 0x1040, 0x287C, 0x2484, 0x5308, 0xFE00, 0x4420, 0x7DCE, 0x4502, 0x7D02, 0x41CE, 0x5102, 0x4902, 0x5502, 0x65FE, 0x4102}},
    {39202, 2, {0x820, 0x810, 0x147C, 0x1200, 0x2944, 0x7E24, 0xA228, 0x3EFE, 0x2200, 0x3E00, 0x207C, 0x2844, 0x2444, 0x2A44, 0x327C, 0x2044}},
    {39203, 2, {0x1020, 0x1020, 0x2BFE, 0x2450, 0x5288, 0xFD24, 0x4622, 0x7DFC, 0x4524, 0x7DFC, 0x4124, 0x51FC, 0x4822, 0x5422, 0x641E, 0x4000}},
    {39204, 2, {0x1020, 0x10A2, 0x28A2, 0x2524, 0x5250, 0xFC88, 0x4704, 0x7C22, 0x4420, 0x7CA4, 0x40A4, 0x5128, 0x4850, 0x5488, 0x6504, 0x4202}},
    {39205, 2, {0x440, 0x7C7C, 0x440, 0x3C78, 0x440, 0x7D7C, 0x6C0, 0xD60, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE8, 0x8D0, 0xA30, 0xC08}},
    {39206, 2, {0x1000, 0x10FC, 0x2880, 0x24F8, 0x5280, 0xFCF8, 0x4480, 0x7FFE, 0x4540, 0x7D24, 0x4128, 0x5110, 0x4908, 0x5544, 0x6582, 0x4100}},
    {39207, 2, {0x1038, 0x10E0, 0x2820, 0x25FC, 0x5270, 0xFCA8, 0x4524, 0x7C20, 0x4440, 0x7DFE, 0x4048, 0x5088, 0x4850, 0x5420, 0x6450, 0x4188}},
    {39208, 2, {0x820, 0x810, 0x14FE, 0x1282, 0x2904, 0x7E7C, 0xA244, 0x3E44, 0x227C, 0x3E40, 0x2040, 0x287E, 0x2442, 0x2A42, 0x327E, 0x2042}},
    {39209, 2, {0x820, 0x820, 0x1450, 0x1288, 0x2926, 0x7E10, 0xA3FE, 0x3E20, 0x2244, 0x3EF8, 0x2012, 0x2864, 0x2588, 0x2A18, 0x3264, 0x2182}},
    {39210, 2, {0x1000, 0x13FE, 0x2820, 0x2440, 0x53FC, 0xFD54, 0x4554, 0x7D54, 0x4554, 0x7D2C, 0x4020, 0x53FE, 0x4850, 0x5488, 0x6504, 0x4202}},
    {39211, 2, {0x1000, 0x13FE, 0x2A22, 0x2420, 0x53FE, 0xFC20, 0x45FC, 0x7D24, 0x45FC, 0x7D24, 0x41FC, 0x5020, 0x4BFE, 0x5420, 0x6420, 0x4020}},
    {39212, 2, {0x1080, 0x109E, 0x2892, 0x25D2, 0x5292, 0xFC9E, 0x4492, 0x7DD2, 0x4552, 0x7D5E, 0x4152, 0x5152, 0x49D2, 0x5422, 0x642A, 0x4044}},
    {39213, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x52FC, 0xFC84, 0x44FC, 0x7C00, 0x45FE, 0x7C20, 0x4020, 0x50FC, 0x4820, 0x5420, 0x65FE, 0x4000}},
    {39214, 2, {0x7F20, 0x1050, 0x1E88, 0x2346, 0x5490, 0x924, 0x1048, 0x6190, 0x660, 0x1FF8, 0xE816, 0xFF0, 0x810, 0xFF4, 0x868, 0xC1C}},
    {39215, 2, {0x1040, 0x107C, 0x2884, 0x24F8, 0x5208, 0xFDFE, 0x4440, 0x7CA4, 0x4538, 0x7C58, 0x4094, 0x5134, 0x4854, 0x5492, 0x6550, 0x4020}},
    {39216, 2, {0x1080, 0x108E, 0x2920, 0x2610, 0x54C0, 0xF8AE, 0x4904, 0x7B14, 0x4D24, 0x7964, 0x4124, 0x5124, 0x4924, 0x5904, 0x6914, 0x4108}},
    {39217, 2, {0x1080, 0x10BC, 0x2884, 0x2508, 0x51FE, 0xFB20, 0x4D20, 0x797C, 0x4990, 0x7910, 0x41FE, 0x5110, 0x4928, 0x5928, 0x6944, 0x4182}},
    {39218, 2, {0x800, 0x8FC, 0x1484, 0x12FC, 0x2984, 0x7EFC, 0xA240, 0x3EFE, 0x2312, 0x3E92, 0x20AA, 0x2882, 0x24FA, 0x2A02, 0x3214, 0x2008}},
    {39219, 2, {0x800, 0x87C, 0x1444, 0x127C, 0x2944, 0x7E7C, 0xA200, 0x3EFE, 0x2240, 0x3E7E, 0x20AA, 0x292A, 0x244A, 0x2A92, 0x322A, 0x2044}},
    {39220, 2, {0x1020, 0x1020, 0x2BFE, 0x2450, 0x50A8, 0xF924, 0x4AFA, 0x7820, 0x48A8, 0x7888, 0x43FE, 0x5088, 0x4888, 0x5888, 0x6908, 0x4208}},
    {39221, 2, {0x1000, 0x10FC, 0x28A4, 0x24A4, 0x52FC, 0xFCA4, 0x44A4, 0x7CFC, 0x4400, 0x7DFE, 0x4090, 0x5094, 0x4888, 0x54A8, 0x64C4, 0x4082}},
    {39222, 2, {0x1000, 0x10FC, 0x2884, 0x24F4, 0x5294, 0xFDFE, 0x4502, 0x7CFC, 0x4484, 0x7CFC, 0x4084, 0x50FC, 0x4884, 0x5484, 0x6494, 0x4088}},
    {39223, 2, {0x810, 0x810, 0x14FE, 0x1238, 0x2954, 0x7E92, 0xA200, 0x3E7C, 0x2244, 0x3E7C, 0x2044, 0x287C, 0x2444, 0x2A00, 0x32FE, 0x2000}},
    {39224, 2, {0x1022, 0x1012, 0x2914, 0x2480, 0x52BE, 0xFC08, 0x4408, 0x7DBE, 0x4488, 0x7C88, 0x4094, 0x5092, 0x48A2, 0x5540, 0x663E, 0x4000}},
    {39225, 2, {0x820, 0x810, 0x14FE, 0x1290, 0x29BC, 0x7E94, 0xA2FE, 0x3E94, 0x22BC, 0x3E90, 0x20BC, 0x28A4, 0x24A4, 0x2AA4, 0x32BC, 0x2124}},
    {39226, 2, {0x1028, 0x1024, 0x2BFE, 0x2420, 0x53FC, 0xFD24, 0x45FC, 0x7D24, 0x45FC, 0x7D24, 0x4008, 0x53FE, 0x4888, 0x5448, 0x6448, 0x4018}},
    {39227, 2, {0x844, 0x824, 0x1428, 0x1200, 0x29FE, 0x7E10, 0xA210, 0x3E7C, 0x2210, 0x3E10, 0x20FE, 0x2800, 0x24AA, 0x2AAA, 0x32AA, 0x2000}},
    {39228, 2, {0x1080, 0x1080, 0x28FC, 0x2500, 0x52F8, 0xFC00, 0x45F8, 0x7C48, 0x4568, 0x7CC8, 0x41F8, 0x504A, 0x48EA, 0x555A, 0x6446, 0x4042}},
    {39229, 2, {0x1020, 0x1040, 0x29FC, 0x2524, 0x5324, 0xFDFC, 0x4524, 0x7D44, 0x45FC, 0x7C40, 0x40B0, 0x50A4, 0x493C, 0x5522, 0x6622, 0x441E}},
    {39230, 2, {0x1040, 0x13BE, 0x2A12, 0x2692, 0x5252, 0xFAAA, 0x4B24, 0x7840, 0x4BFE, 0x7A22, 0x4222, 0x53FE, 0x4A22, 0x5A22, 0x6BFE, 0x4202}},
    {39231, 2, {0x10A0, 0x112C, 0x2924, 0x2524, 0x53AC, 0xFD24, 0x4524, 0x7DFC, 0x4420, 0x7DFC, 0x4088, 0x5050, 0x4820, 0x5450, 0x6488, 0x4306}},
    {39232, 2, {0x101E, 0x11E0, 0x2822, 0x2512, 0x5294, 0xFC80, 0x4420, 0x7DCE, 0x4502, 0x7D02, 0x41CE, 0x5102, 0x4902, 0x5502, 0x65FE, 0x4102}},
    {39233, 2, {0x1020, 0x1020, 0x28FC, 0x2420, 0x5220, 0xFDFE, 0x4440, 0x7C84, 0x45FE, 0x7C02, 0x41FC, 0x5154, 0x4954, 0x5554, 0x67FE, 0x4000}},
    {39234, 2, {0x1000, 0x10F8, 0x2888, 0x2488, 0x52F8, 0xFC88, 0x4488, 0x7CF8, 0x4400, 0x7DFC, 0x4154, 0x5154, 0x4954, 0x5554, 0x67FE, 0x4000}},
    {39235, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x5200, 0xFDFC, 0x4504, 0x7DFC, 0x4504, 0x7DFC, 0x4020, 0x53FE, 0x4850, 0x5488, 0x6504, 0x4202}},
    {39236, 2, {0x1020, 0x1124, 0x28A8, 0x27FE, 0x5202, 0xFCF8, 0x4488, 0x7C88, 0x44F8, 0x7C20, 0x4020, 0x51FC, 0x4820, 0x5420, 0x67FE, 0x4000}},
    {39237, 2, {0x87C, 0x844, 0x147C, 0x1244, 0x297C, 0x7E00, 0xA2FE, 0x3EAA, 0x22FE, 0x3E00, 0x20FC, 0x2844, 0x2428, 0x2A10, 0x3268, 0x2186}},
    {39238, 2, {0x1000, 0x11FC, 0x2924, 0x25FC, 0x5324, 0xFDFC, 0x44A8, 0x7CA8, 0x47FE, 0x7CA8, 0x40A8, 0x51FC, 0x4820, 0x57FE, 0x6420, 0x4020}},
    {39239, 2, {0x1000, 0x13FE, 0x2A00, 0x267C, 0x5244, 0xFA44, 0x4A7C, 0x7A00, 0x4AEE, 0x7AAA, 0x42AA, 0x52AA, 0x4AEE, 0x5A00, 0x6BFE, 0x4000}},
    {39240, 2, {0x1084, 0x1048, 0x29FE, 0x2410, 0x52FC, 0xFC20, 0x45FE, 0x7C40, 0x447C, 0x7CA4, 0x4124, 0x50FC, 0x4844, 0x5444, 0x65FE, 0x4000}},
    {39241, 2, {0x1088, 0x1088, 0x2BFE, 0x2488, 0x52F8, 0xFC20, 0x45FC, 0x7D24, 0x45FC, 0x7C20, 0x43FE, 0x5020, 0x49FC, 0x5420, 0x67FE, 0x4000}},
    {39242, 2, {0x1148, 0x1148, 0x2BE8, 0x2550, 0x515E, 0xFBE4, 0x4814, 0x7BD4, 0x4A54, 0x7BD4, 0x4254, 0x53C8, 0x4A48, 0x5A54, 0x6A54, 0x42E2}},
    {39243, 2, {0x1020, 0x11FC, 0x2924, 0x25FC, 0x5220, 0xFFFE, 0x4400, 0x7DFC, 0x4504, 0x7DFC, 0x4104, 0x51FC, 0x4904, 0x55FC, 0x6488, 0x4104}},
    {39244, 2, {0x1000, 0x13DE, 0x2A52, 0x27DE, 0x5210, 0xFA52, 0x49CE, 0x7800, 0x4888, 0x7BFE, 0x4088, 0x5088, 0x4BFE, 0x5888, 0x6904, 0x4202}},
    {39245, 2, {0x1088, 0x1050, 0x2BFE, 0x2420, 0x51FC, 0xF820, 0x4BFE, 0x7924, 0x48A8, 0x7BFE, 0x4000, 0x51FC, 0x4904, 0x5904, 0x69FC, 0x4104}},
    {39246, 2, {0x1020, 0x13FE, 0x2820, 0x25FC, 0x5200, 0xFDFC, 0x4504, 0x7DFC, 0x4488, 0x7FFE, 0x4000, 0x51FC, 0x4904, 0x5504, 0x65FC, 0x4104}},
    {39247, 2, {0x3C20, 0x43E, 0xFF44, 0x22A8, 0x3E10, 0x2328, 0x7E44, 0x380, 0x660, 0x1FF8, 0xE816, 0xFF0, 0x810, 0xFF4, 0x868, 0xC1C}},
    {39248, 2, {0x1020, 0x13FE, 0x2820, 0x25FC, 0x5000, 0xFBFE, 0x4A02, 0x79FC, 0x4800, 0x79FC, 0x4104, 0x51FC, 0x4904, 0x5888, 0x6BFE, 0x4000}},
    {39249, 2, {0x1124, 0x1124, 0x2AAA, 0x27AE, 0x5124, 0xFAAA, 0x4BAE, 0x7924, 0x4FFE, 0x7910, 0x4114, 0x5114, 0x4A8A, 0x5A4A, 0x6A16, 0x4422}},
    {39250, 2, {0x1020, 0x1020, 0x29FC, 0x2420, 0x53FE, 0xF908, 0x4B9C, 0x7908, 0x4988, 0x7E3E, 0x4000, 0x53FE, 0x4890, 0x5890, 0x6912, 0x420E}},
    {39251, 2, {0x1020, 0x1124, 0x28A8, 0x27FE, 0x5202, 0xFCF8, 0x4488, 0x7C88, 0x44F8, 0x7C20, 0x43FE, 0x5070, 0x48A8, 0x5524, 0x6622, 0x4020}},
    {39252, 2, {0x100, 0xFFFE, 0x10A0, 0x25FE, 0x7B20, 0x15FC, 0x7920, 0x11FE, 0xE660, 0x1FF8, 0xE816, 0xFF0, 0x810, 0xFF4, 0x868, 0xC1C}},
    {39253, 2, {0x7C1C, 0x4410, 0xFEFE, 0x2092, 0x7CF8, 0x494, 0x28BC, 0x11AA, 0x666, 0x1FF8, 0xE816, 0xFF0, 0x810, 0xFF4, 0x868, 0xC1C}},
    {39254, 2, {0x1020, 0x113C, 0x2920, 0x27FE, 0x5010, 0xF814, 0x4BFE, 0x7A10, 0x4BF0, 0x7A54, 0x4254, 0x52D4, 0x4B68, 0x544A, 0x6896, 0x4322}},
    {39255, 2, {0x1100, 0x23DE, 0x7A52, 0x13D4, 0x2A52, 0x7BDA, 0x1294, 0x23D0, 0x660, 0x1FF8, 0xE816, 0xFF0, 0x810, 0xFF4, 0x868, 0xC1C}},
    {39256, 2, {0x1010, 0x11FE, 0x2800, 0x25FE, 0x5302, 0xFD7A, 0x454A, 0x7DFE, 0x4400, 0x7CFC, 0x4084, 0x50FC, 0x4884, 0x54FC, 0x6400, 0x41FE}},
    {39257, 2, {0x1020, 0x11FC, 0x2820, 0x2488, 0x53FE, 0xFC88, 0x45FC, 0x7D04, 0x45FC, 0x7D04, 0x41FC, 0x5104, 0x49FC, 0x5488, 0x6504, 0x4202}},
    {39258, 2, {0x1088, 0x13FE, 0x2888, 0x2420, 0x53FC, 0xFC20, 0x47FE, 0x7C40, 0x4488, 0x7DFC, 0x4000, 0x51FC, 0x4954, 0x5554, 0x67FE, 0x4000}},
    {39259, 2, {0x1088, 0x13FE, 0x2888, 0x27FE, 0x5202, 0xF9FC, 0x4800, 0x7BFE, 0x4840, 0x78A2, 0x4354, 0x50B8, 0x4B54, 0x5892, 0x6B50, 0x4020}},
    {39260, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x28D4, 0x29A2, 0x23F8, 0x2616, 0x2BF0, 0x2210, 0x43F4, 0x4268, 0x831C}},
    {39261, 2, {0x1040, 0x1020, 0x2BFE, 0x2648, 0x53FE, 0xFA48, 0x4AEC, 0x7B5A, 0x4A48, 0x7BFE, 0x4220, 0x527C, 0x4AC4, 0x5344, 0x647C, 0x4844}},
    {39262, 2, {0x11F0, 0x1210, 0x2BFC, 0x2604, 0x53FC, 0xFA24, 0x4BB8, 0x7A22, 0x49FE, 0x7A10, 0x47FC, 0x5244, 0x4BFC, 0x58D0, 0x694A, 0x423E}},
    {39263, 2, {0x1020, 0x13FE, 0x2800, 0x25DC, 0x5154, 0xF9DC, 0x4888, 0x7BFE, 0x4888, 0x7BFE, 0x4088, 0x53FE, 0x4894, 0x5988, 0x6AA4, 0x40C2}},
    {39264, 2, {0x1000, 0x13FE, 0x2A52, 0x2652, 0x53FE, 0xF908, 0x4954, 0x7A5E, 0x4FB4, 0x791E, 0x4254, 0x57DE, 0x4814, 0x5554, 0x655E, 0x4010}},
    {39265, 2, {0x1528, 0x17BE, 0x2948, 0x27BE, 0x5318, 0xFDAA, 0x4946, 0x7BFC, 0x4A04, 0x7BFC, 0x4204, 0x53FC, 0x4A04, 0x5BFC, 0x6908, 0x4204}},
    {39266, 2, {0x1040, 0x17FC, 0x2A48, 0x27F8, 0x5040, 0xFFFE, 0x4AAA, 0x7BB8, 0x4910, 0x7BFC, 0x4110, 0x57FE, 0x4928, 0x5310, 0x6D48, 0x4186}},
    {39267, 2, {0x2000, 0x2000, 0x2000, 0x3E00, 0x4400, 0x4800, 0x8000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1200, 0x1400, 0x1800, 0x1000, 0x0}},
    {39268, 2, {0x2000, 0x2000, 0x20FE, 0x3E10, 0x4410, 0x4810, 0x8010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1210, 0x1410, 0x1810, 0x1050, 0x20}},
    {39269, 2, {0x2000, 0x21F0, 0x2110, 0x3D10, 0x4510, 0x4910, 0xA110, 0x2110, 0x2110, 0x2110, 0x2110, 0x2112, 0x2912, 0x3212, 0x220E, 0x400}},
    {39270, 2, {0x2010, 0x2078, 0x23C0, 0x3C40, 0x4440, 0x4840, 0x8040, 0x207E, 0x27C0, 0x2040, 0x2040, 0x2442, 0x2842, 0x3042, 0x203E, 0x0}},
    {39271, 2, {0x2000, 0x21F8, 0x2010, 0x3C20, 0x4440, 0x4880, 0xA1FE, 0x2092, 0x2092, 0x2092, 0x2112, 0x2122, 0x2A22, 0x3442, 0x2094, 0x108}},
    {39272, 2, {0x2040, 0x2040, 0x2040, 0x3BFC, 0x4840, 0x5248, 0x8248, 0x2248, 0x2248, 0x23F8, 0x2048, 0x2040, 0x2842, 0x3042, 0x203E, 0x0}},
    {39273, 2, {0x2080, 0x2080, 0x20FC, 0x3D00, 0x4600, 0x48F8, 0x8000, 0x2000, 0x21F8, 0x2008, 0x2008, 0x2408, 0x280A, 0x300A, 0x2006, 0x2}},
    {39274, 2, {0x2008, 0x203C, 0x21E0, 0x3C20, 0x4420, 0x4820, 0x8020, 0x23FE, 0x2020, 0x2020, 0x2020, 0x2420, 0x2820, 0x3020, 0x21FC, 0x0}},
    {39275, 2, {0x2008, 0x203C, 0x21E0, 0x3C20, 0x4420, 0x4820, 0x83FE, 0x2020, 0x2020, 0x2050, 0x2050, 0x2450, 0x2888, 0x3088, 0x2104, 0x202}},
    {39276, 2, {0x2080, 0x2080, 0x2100, 0x39FE, 0x4A00, 0x5440, 0x8040, 0x23FC, 0x2044, 0x2044, 0x2084, 0x2084, 0x2904, 0x3104, 0x2228, 0x410}},
    {39277, 2, {0x2008, 0x201C, 0x21E0, 0x3D00, 0x4500, 0x49FC, 0xA144, 0x2144, 0x2144, 0x2128, 0x2128, 0x2110, 0x2910, 0x3228, 0x2244, 0x482}},
    {39278, 2, {0x2040, 0x2040, 0x2040, 0x3E7C, 0x4484, 0x4888, 0x8120, 0x1020, 0x1020, 0x1050, 0x1050, 0x1250, 0x1488, 0x1888, 0x1104, 0x202}},
    {39279, 2, {0x2050, 0x2048, 0x2040, 0x3C5C, 0x45E0, 0x4840, 0xA05E, 0x23E0, 0x2044, 0x2048, 0x2030, 0x2022, 0x2852, 0x308A, 0x2306, 0x2}},
    {39280, 2, {0x2080, 0x2080, 0x2080, 0x3CFE, 0x4520, 0x4A20, 0x8020, 0x21FC, 0x2124, 0x2124, 0x2124, 0x2524, 0x2934, 0x3128, 0x2020, 0x20}},
    {39281, 2, {0x2080, 0x2080, 0x21FC, 0x3904, 0x4A04, 0x55F4, 0x8114, 0x2114, 0x2114, 0x21F4, 0x2104, 0x2128, 0x2912, 0x3102, 0x20FE, 0x0}},
    {39282, 2, {0x2000, 0x23FC, 0x2004, 0x3C04, 0x47F4, 0x4804, 0xA004, 0x23E4, 0x2224, 0x2224, 0x2224, 0x23E4, 0x2804, 0x3004, 0x2014, 0x8}},
    {39283, 2, {0x2020, 0x2020, 0x2124, 0x3D24, 0x4524, 0x4924, 0x81FC, 0x2020, 0x2020, 0x2124, 0x2124, 0x2524, 0x2924, 0x3124, 0x21FC, 0x4}},
    {39284, 2, {0x2020, 0x2020, 0x2040, 0x3C88, 0x4504, 0x4BFE, 0x8002, 0x2000, 0x21FC, 0x2104, 0x2104, 0x2504, 0x2904, 0x3104, 0x21FC, 0x104}},
    {39285, 2, {0x2000, 0x23FE, 0x2108, 0x3D08, 0x4508, 0x49F8, 0xA108, 0x2108, 0x21F8, 0x2108, 0x2108, 0x211E, 0x2BE8, 0x3008, 0x2008, 0x8}},
    {39286, 2, {0x2080, 0x2080, 0x20BC, 0x3BC0, 0x4850, 0x5024, 0x80D4, 0x230C, 0x2000, 0x27FE, 0x2090, 0x2090, 0x2912, 0x3112, 0x220E, 0x400}},
    {39287, 2, {0x2040, 0x2040, 0x2080, 0x3BFC, 0x4A04, 0x5204, 0x82F4, 0x2294, 0x2294, 0x2294, 0x2294, 0x22F4, 0x2A04, 0x3204, 0x2214, 0x208}},
    {39288, 2, {0x2020, 0x2020, 0x2050, 0x3C88, 0x4504, 0x4A02, 0x81FC, 0x2000, 0x2000, 0x21FC, 0x2104, 0x2504, 0x2904, 0x3104, 0x21FC, 0x104}},
    {39289, 2, {0x2080, 0x2080, 0x20F8, 0x3908, 0x4B10, 0x54A0, 0x8040, 0x20A0, 0x2118, 0x2206, 0x25F8, 0x2108, 0x2908, 0x3108, 0x21F8, 0x108}},
    {39290, 2, {0x2040, 0x2020, 0x2020, 0x3DFE, 0x4400, 0x4888, 0xA104, 0x2202, 0x2088, 0x2088, 0x2050, 0x2050, 0x2820, 0x3050, 0x2088, 0x306}},
    {39291, 2, {0x2040, 0x2020, 0x2020, 0x3BFE, 0x4840, 0x5040, 0x80A0, 0x20A2, 0x21A4, 0x2298, 0x2490, 0x2088, 0x2884, 0x30A2, 0x20C0, 0x80}},
    {39292, 2, {0x2104, 0x2084, 0x2088, 0x3800, 0x4BFE, 0x5088, 0x8088, 0x2088, 0x2088, 0x27FE, 0x2088, 0x2088, 0x2908, 0x3108, 0x2208, 0x408}},
    {39293, 2, {0x2020, 0x2020, 0x23FE, 0x3C20, 0x4420, 0x4BFE, 0xA202, 0x2404, 0x21F8, 0x2010, 0x2020, 0x23FE, 0x2820, 0x3020, 0x20A0, 0x40}},
    {39294, 2, {0x2000, 0x23FC, 0x2000, 0x3C00, 0x45F8, 0x4908, 0x8108, 0x2108, 0x21F8, 0x2000, 0x2108, 0x2488, 0x2890, 0x3000, 0x27FE, 0x0}},
    {39295, 2, {0x2010, 0x20D8, 0x2394, 0x3C94, 0x4490, 0x4BFE, 0xA090, 0x2094, 0x2094, 0x20D8, 0x2398, 0x2090, 0x28AA, 0x30CA, 0x2286, 0x102}},
    {39296, 2, {0x2020, 0x2020, 0x2050, 0x3C88, 0x4504, 0x4AFA, 0xA020, 0x2020, 0x23FE, 0x2020, 0x2128, 0x2124, 0x2A22, 0x3422, 0x20A0, 0x40}},
    {39297, 2, {0x2008, 0x203C, 0x23C0, 0x3844, 0x4A24, 0x5128, 0x8100, 0x2040, 0x23FE, 0x2088, 0x2108, 0x2390, 0x2860, 0x3050, 0x2188, 0x604}},
    {39298, 2, {0x2040, 0x2040, 0x2088, 0x3D04, 0x47FE, 0x4802, 0xA088, 0x2144, 0x2242, 0x20F8, 0x2188, 0x2250, 0x2820, 0x3050, 0x2188, 0x606}},
    {39299, 2, {0x2000, 0x21FC, 0x2124, 0x3D24, 0x45FC, 0x4924, 0x8124, 0x21FC, 0x2020, 0x23FE, 0x2070, 0x24A8, 0x2924, 0x3222, 0x2020, 0x20}},
    {39300, 2, {0x2000, 0x21FC, 0x2104, 0x3D04, 0x45FC, 0x4904, 0xA104, 0x21FC, 0x2000, 0x2112, 0x21D4, 0x2118, 0x2910, 0x3152, 0x2192, 0x10E}},
    {39301, 2, {0x2080, 0x2080, 0x20F8, 0x3908, 0x4A10, 0x5400, 0x8040, 0x239C, 0x2204, 0x2204, 0x239C, 0x2204, 0x2A04, 0x3204, 0x23FC, 0x204}},
    {39302, 2, {0x2040, 0x2020, 0x23FE, 0x3A02, 0x4C04, 0x51F8, 0x8108, 0x2108, 0x21F8, 0x2100, 0x2100, 0x21FC, 0x2904, 0x3104, 0x21FC, 0x104}},
    {39303, 2, {0x2040, 0x2040, 0x27FC, 0x38E0, 0x4950, 0x5248, 0x8446, 0x23F8, 0x2208, 0x23F8, 0x2208, 0x23F8, 0x2A08, 0x3000, 0x27FE, 0x0}},
    {39304, 2, {0x2020, 0x2020, 0x21FC, 0x3D24, 0x45FC, 0x4820, 0x83FE, 0x2000, 0x21FC, 0x2104, 0x2124, 0x2524, 0x2924, 0x3050, 0x2088, 0x304}},
    {39305, 2, {0x2000, 0x21FC, 0x2104, 0x3DE4, 0x4524, 0x4BFE, 0x8202, 0x21FC, 0x2104, 0x21FC, 0x2104, 0x25FC, 0x2904, 0x3104, 0x2114, 0x108}},
    {39306, 2, {0x20A0, 0x232E, 0x2222, 0x3E22, 0x47AE, 0x4A22, 0xA222, 0x23FE, 0x2020, 0x23FE, 0x2104, 0x2088, 0x2850, 0x3020, 0x20D8, 0x706}},
    {39307, 2, {0x2100, 0x21F0, 0x2210, 0x3C20, 0x4BFC, 0x5244, 0x8244, 0x23FC, 0x20A0, 0x2122, 0x261E, 0x20C0, 0x2830, 0x3180, 0x2060, 0x10}},
    {39308, 2, {0x2020, 0x2020, 0x21FC, 0x3C20, 0x4420, 0x4BFE, 0xA080, 0x2104, 0x23FE, 0x2002, 0x23FC, 0x2294, 0x2A94, 0x3294, 0x27FE, 0x0}},
    {39309, 2, {0x2110, 0x2110, 0x27FC, 0x3910, 0x4800, 0x53F8, 0x8208, 0x23F8, 0x2208, 0x23F8, 0x2040, 0x23FC, 0x28A0, 0x3110, 0x2208, 0x406}},
    {39310, 2, {0x2028, 0x2024, 0x23FE, 0x3C20, 0x45FC, 0x4924, 0xA1FC, 0x2124, 0x21FC, 0x2124, 0x2008, 0x23FE, 0x2888, 0x3048, 0x2048, 0x18}},
    {39311, 2, {0x2040, 0x23BE, 0x2212, 0x3A92, 0x4A52, 0x52AA, 0x8324, 0x2040, 0x23FE, 0x2222, 0x2222, 0x23FE, 0x2A22, 0x3222, 0x23FE, 0x202}},
    {39312, 2, {0x2208, 0x2110, 0x27FC, 0x3840, 0x4BF8, 0x5080, 0x87FC, 0x2100, 0x23F8, 0x2488, 0x2888, 0x23F8, 0x2908, 0x3108, 0x27FE, 0x0}},
    {39313, 2, {0x2088, 0x2088, 0x23FE, 0x3C88, 0x44F8, 0x4820, 0xA1FC, 0x2124, 0x21FC, 0x2020, 0x23FE, 0x2020, 0x29FC, 0x3020, 0x23FE, 0x0}},
    {39314, 2, {0x23F8, 0x2208, 0x23F8, 0x3A08, 0x4BF8, 0x5000, 0x87FC, 0x24A4, 0x24A4, 0x27FC, 0x2000, 0x23F8, 0x2910, 0x30E0, 0x2318, 0xC06}},
    {39315, 2, {0x2288, 0x2288, 0x27C8, 0x3A90, 0x4A9E, 0x57D4, 0x8024, 0x27D4, 0x2454, 0x27D4, 0x2454, 0x27C8, 0x2C48, 0x3454, 0x2464, 0x4C2}},
    {39316, 2, {0x2000, 0x23DE, 0x2252, 0x3BDE, 0x4A10, 0x5252, 0x81CE, 0x2000, 0x2088, 0x23FE, 0x2088, 0x2088, 0x2BFE, 0x3088, 0x2104, 0x202}},
    {39317, 2, {0x2040, 0x27FC, 0x2248, 0x3BF8, 0x4840, 0x5FFE, 0x8AAA, 0x23B8, 0x2110, 0x27FC, 0x2110, 0x2FFE, 0x2928, 0x3310, 0x2D48, 0x186}},
    {39318, 2, {0x1010, 0x820, 0xFFFE, 0x200, 0x400, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008, 0x2008, 0x3FF8, 0x2008}},
    {39319, 2, {0x2110, 0x20A0, 0x2000, 0x23FC, 0xFC80, 0x25F8, 0x2508, 0x25F8, 0x2508, 0x25F8, 0x2508, 0x25FA, 0x250A, 0x4402, 0x43FE, 0x8000}},
    {39320, 2, {0x4414, 0x2412, 0x2810, 0xFDFE, 0x1010, 0x2010, 0x7DD2, 0x4552, 0x4552, 0x7D54, 0x45D4, 0x7C08, 0x44CA, 0x471A, 0x7C26, 0x4442}},
    {39321, 2, {0x70, 0x1F80, 0x100, 0x100, 0xFFFE, 0x540, 0x920, 0x3118, 0xDFF6, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010}},
    {39322, 2, {0x410, 0x7850, 0x1050, 0xFE50, 0x1088, 0x3888, 0x5504, 0x92FA, 0x7C48, 0x4448, 0x4448, 0x7C48, 0x4488, 0x4488, 0x7D28, 0x4610}},
    {39323, 2, {0x450, 0x7848, 0x1048, 0xFE40, 0x11FE, 0x3840, 0x5440, 0x907C, 0x7CA4, 0x44A4, 0x44A8, 0x7D28, 0x4510, 0x4628, 0x7C44, 0x4482}},
    {39324, 2, {0x400, 0x78FC, 0x1084, 0xFE84, 0x1084, 0x38FC, 0x5480, 0x90A0, 0x7CA4, 0x44A8, 0x44B0, 0x7CA0, 0x44A2, 0x4522, 0x7D1E, 0x4600}},
    {39325, 2, {0x440, 0x7820, 0x1028, 0xFE08, 0x1008, 0x3848, 0x5454, 0x9152, 0x7D52, 0x4560, 0x4660, 0x7C44, 0x44C4, 0x4544, 0x7E3C, 0x4400}},
    {39326, 2, {0x420, 0x7820, 0x11FE, 0xFE20, 0x1020, 0x39FE, 0x5502, 0x9204, 0x7CF8, 0x4410, 0x4420, 0x7DFE, 0x4420, 0x4420, 0x7CA0, 0x4440}},
    {39327, 2, {0x408, 0x783C, 0x11C0, 0xFE04, 0x1144, 0x38A8, 0x5400, 0x91F8, 0x7C10, 0x4420, 0x47FE, 0x7C20, 0x4420, 0x4420, 0x7CA0, 0x4440}},
    {39328, 2, {0x420, 0x7820, 0x1050, 0xFE88, 0x1144, 0x3A22, 0x54F8, 0x9008, 0x7C10, 0x4420, 0x45FC, 0x7D04, 0x4504, 0x4504, 0x7DFC, 0x4504}},
    {39329, 2, {0x448, 0x7848, 0x1048, 0xFDCE, 0x1048, 0x3848, 0x5448, 0x91CE, 0x7C48, 0x4448, 0x4448, 0x7DCE, 0x4448, 0x4448, 0x7C48, 0x4448}},
    {39330, 2, {0x448, 0x7844, 0x105E, 0xFDE0, 0x1028, 0x3812, 0x546A, 0x9196, 0x7C48, 0x445E, 0x45E0, 0x7C24, 0x4428, 0x4412, 0x7C6A, 0x4586}},
    {39331, 2, {0x420, 0x7820, 0x11FE, 0xFE50, 0x10A8, 0x3926, 0x54F8, 0x90A8, 0x7CA8, 0x44F8, 0x44A8, 0x7CA8, 0x44F8, 0x44AA, 0x7C22, 0x441E}},
    {39332, 2, {0x400, 0x79FC, 0x1104, 0xFDFC, 0x1104, 0x39FC, 0x5480, 0x91FE, 0x7E22, 0x4522, 0x4552, 0x7D02, 0x45FA, 0x4402, 0x7C14, 0x4408}},
    {39333, 2, {0x440, 0x7840, 0x10FE, 0xFE80, 0x11FC, 0x3A84, 0x54FC, 0x9084, 0x7CFC, 0x4440, 0x447C, 0x7CC4, 0x4528, 0x4410, 0x7C68, 0x4586}},
    {39334, 2, {0x488, 0x7850, 0x1000, 0xFDFE, 0x1050, 0x39FC, 0x5454, 0x93FE, 0x7C54, 0x45FC, 0x4450, 0x7CD8, 0x4554, 0x4652, 0x7C50, 0x4450}},
    {39335, 2, {0x400, 0x78F8, 0x1088, 0xFE88, 0x10F8, 0x3888, 0x5488, 0x90F8, 0x7C00, 0x45FC, 0x4554, 0x7D54, 0x4554, 0x4554, 0x7FFE, 0x4400}},
    {39336, 2, {0x878, 0xFF48, 0x886, 0x3E78, 0x2A48, 0x3E30, 0x4048, 0x9FE4, 0x100, 0x7FFC, 0x920, 0x3FF8, 0xC826, 0xFE0, 0x820, 0xFE0}},
    {39337, 2, {0x420, 0x79FC, 0x1020, 0xFC88, 0x13FE, 0x3888, 0x55FC, 0x9104, 0x7DFC, 0x4504, 0x45FC, 0x7D04, 0x45FC, 0x4488, 0x7D04, 0x4602}},
    {39338, 2, {0x420, 0x7810, 0x11FE, 0xFF02, 0x10FC, 0x38A8, 0x5530, 0x90FC, 0x7D84, 0x44FC, 0x4484, 0x7CFC, 0x4484, 0x44FC, 0x7C48, 0x4484}},
    {39339, 2, {0x1FE0, 0x100, 0x7FFC, 0x920, 0x3FF8, 0xC826, 0xFE0, 0x824, 0x7FF8, 0x1010, 0xFEFE, 0x5454, 0x9292, 0x7C7C, 0x4444, 0x7C7C}},
    {39340, 2, {0x0, 0x3FF8, 0x2100, 0x3FF0, 0x2100, 0x3FF0, 0x2100, 0x2100, 0x3FF8, 0x8, 0x1248, 0x4928, 0x4928, 0x8008, 0x50, 0x20}},
    {39341, 2, {0x0, 0x7F00, 0x487C, 0x7F24, 0x4824, 0x7F24, 0x4824, 0x4828, 0x7FA8, 0xA8, 0xAA90, 0xAA90, 0xAAA8, 0x80A8, 0x544, 0x282}},
    {39342, 2, {0x0, 0x43FC, 0x2240, 0x23F8, 0x240, 0xBF8, 0xA40, 0x1240, 0x13FC, 0xE004, 0x2554, 0x2554, 0x2554, 0x2404, 0x2028, 0x10}},
    {39343, 2, {0x0, 0x7F7C, 0x4810, 0x7F10, 0x4810, 0x7F10, 0x4810, 0x487E, 0x7F90, 0x90, 0xAA90, 0xAA90, 0xAA90, 0x8090, 0x510, 0x210}},
    {39344, 2, {0x10, 0x7F10, 0x4810, 0x7F3C, 0x4824, 0x7F44, 0x4804, 0x4824, 0x7F94, 0x94, 0xAA84, 0xAA84, 0xAA84, 0x8084, 0x528, 0x210}},
    {39345, 2, {0x10, 0x7F10, 0x4810, 0x7F10, 0x487E, 0x7F10, 0x4810, 0x4810, 0x7F90, 0xA8, 0xAAA8, 0xAAA8, 0xAAA8, 0x80C4, 0x544, 0x282}},
    {39346, 2, {0x8, 0x7F1C, 0x4870, 0x7F10, 0x4810, 0x7F10, 0x4810, 0x481E, 0x7FF0, 0x90, 0xAA90, 0xAA92, 0xAA92, 0x8092, 0x50E, 0x200}},
    {39347, 2, {0x20, 0x7E20, 0x4820, 0x7EA0, 0x48AC, 0x7EB4, 0x48E4, 0x49A4, 0x7EA4, 0x2B4, 0xAAA8, 0xAAA2, 0xAAA2, 0x8282, 0x147E, 0x800}},
    {39348, 2, {0x2, 0x7F22, 0x482A, 0x7F2A, 0x482A, 0x7F2A, 0x482A, 0x482A, 0x7FAA, 0xAA, 0xAAAA, 0xAAAA, 0xAAAA, 0x80AA, 0x542, 0x282}},
    {39349, 2, {0x3FF8, 0x2100, 0x3FF0, 0x2100, 0x3FF0, 0x2100, 0x3FFC, 0x4924, 0x8494, 0x828, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020, 0x4020}},
    {39350, 2, {0x10, 0x7E10, 0x4810, 0x7EFE, 0x4810, 0x7E10, 0x48FC, 0x4844, 0x7E44, 0x244, 0xAA28, 0xAA28, 0xAA10, 0x8228, 0x1444, 0x882}},
    {39351, 2, {0x8, 0x7F08, 0x4808, 0x7F3E, 0x4808, 0x7F08, 0x483E, 0x482A, 0x7FAA, 0xAA, 0xAAAA, 0xAAAA, 0xAAAE, 0x8088, 0x508, 0x208}},
    {39352, 2, {0x0, 0x7F0C, 0x4830, 0x7F20, 0x4820, 0x7F20, 0x483E, 0x4824, 0x7FA4, 0xA4, 0xAAA4, 0xAAA4, 0xAAA4, 0x80A4, 0x544, 0x284}},
    {39353, 2, {0x0, 0x7F3E, 0x4822, 0x7F22, 0x4822, 0x7F22, 0x4822, 0x483E, 0x7FA2, 0xA2, 0xAAA2, 0xAAA2, 0xAAA2, 0x80BE, 0x522, 0x200}},
    {39354, 2, {0x0, 0x7EFC, 0x4844, 0x7E48, 0x4848, 0x7E50, 0x485C, 0x4844, 0x7E44, 0x244, 0xAA68, 0xAAA8, 0xAA90, 0x8298, 0x1524, 0x842}},
    {39355, 2, {0x10, 0x7F10, 0x4820, 0x7F28, 0x4844, 0x7F7C, 0x4804, 0x4828, 0x7FA8, 0xA8, 0xAAA8, 0xAAA8, 0xAAAA, 0x80CA, 0x546, 0x280}},
    {39356, 2, {0x20, 0x7E10, 0x4810, 0x7EFE, 0x4844, 0x7E44, 0x4844, 0x4844, 0x7E44, 0x228, 0xAA28, 0xAA10, 0xAA10, 0x8228, 0x1444, 0x882}},
    {39357, 2, {0x3FF8, 0x2100, 0x3FF0, 0x2100, 0x3FF0, 0x2100, 0x3FFC, 0x4924, 0x8494, 0x108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x100, 0x100}},
    {39358, 2, {0x10, 0x7E10, 0x4810, 0x7E10, 0x48FE, 0x7E92, 0x4892, 0x4810, 0x7E28, 0x228, 0xAA28, 0xAA28, 0xAA48, 0x824A, 0x148A, 0x906}},
    {39359, 2, {0x10, 0x7F08, 0x4800, 0x7F3E, 0x4822, 0x7F22, 0x4822, 0x483E, 0x7FA2, 0xA0, 0xAAA0, 0xAAA0, 0xAAA0, 0x80A0, 0x540, 0x280}},
    {39360, 2, {0x48, 0x7E44, 0x4844, 0x7E40, 0x49FE, 0x7E50, 0x4850, 0x4850, 0x7E50, 0x250, 0xAA50, 0xAA52, 0xAA52, 0x8292, 0x148E, 0x900}},
    {39361, 2, {0x2, 0x7F22, 0x4814, 0x7F08, 0x4814, 0x7F22, 0x4842, 0x4800, 0x7FA2, 0xA2, 0xAA94, 0xAA94, 0xAA88, 0x8094, 0x524, 0x242}},
    {39362, 2, {0x4000, 0x49FC, 0x5120, 0x6120, 0x45FC, 0x4520, 0x3DFC, 0x1120, 0x1120, 0x11FE, 0xFC02, 0x12AA, 0x12AA, 0x1202, 0x1014, 0x1008}},
    {39363, 2, {0x10, 0x7F10, 0x4810, 0x7F10, 0x487C, 0x7F14, 0x4814, 0x4814, 0x7F94, 0xFE, 0xAA90, 0xAA90, 0xAAA8, 0x80A8, 0x544, 0x282}},
    {39364, 2, {0x20, 0x7E20, 0x4820, 0x7E20, 0x48FE, 0x7E20, 0x4820, 0x4830, 0x7E50, 0x250, 0xAA48, 0xAA48, 0xAAC8, 0x82A4, 0x1524, 0xA02}},
    {39365, 2, {0x0, 0x7E7C, 0x4844, 0x7E44, 0x4844, 0x7E44, 0x487C, 0x4854, 0x7E50, 0x250, 0xAA48, 0xAA48, 0xAA48, 0x8284, 0x1484, 0x902}},
    {39366, 2, {0x0, 0x7DFE, 0x5100, 0x7D04, 0x5144, 0x7D28, 0x5128, 0x5110, 0x7D10, 0x528, 0x5528, 0x5544, 0x5584, 0x8500, 0x15FE, 0x800}},
    {39367, 2, {0x10, 0x7F10, 0x4810, 0x7F20, 0x483E, 0x7F44, 0x48A4, 0x4824, 0x7FA8, 0xA8, 0xAA90, 0xAA90, 0xAAA8, 0x80A8, 0x544, 0x282}},
    {39368, 2, {0x8, 0x7E1C, 0x4860, 0x7E40, 0x4840, 0x7E40, 0x4840, 0x487E, 0x7E48, 0x248, 0xAA48, 0xAA48, 0xAA48, 0x8248, 0x15FE, 0x800}},
    {39369, 2, {0x0, 0x7EFE, 0x4882, 0x7E82, 0x4882, 0x7EBA, 0x48AA, 0x48AA, 0x7EAA, 0x2AA, 0xAABA, 0xAAAA, 0xAA82, 0x8282, 0x148A, 0x884}},
    {39370, 2, {0x10, 0x7E10, 0x4810, 0x7EFE, 0x4892, 0x7E94, 0x4890, 0x48FC, 0x7EA4, 0x2A4, 0xAAA8, 0xAAA8, 0xAA90, 0x8328, 0x1544, 0xA82}},
    {39371, 2, {0x0, 0x7F7E, 0x4812, 0x7F12, 0x4812, 0x7F22, 0x482A, 0x4844, 0x7F80, 0xBE, 0xAAA2, 0xAAA2, 0xAAA2, 0x80A2, 0x53E, 0x222}},
    {39372, 2, {0x1000, 0x3EF8, 0x4288, 0xA2AA, 0x1492, 0x187E, 0xE000, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x492C}},
    {39373, 2, {0x0, 0x7F7C, 0x4810, 0x7F10, 0x4854, 0x7F34, 0x4838, 0x4810, 0x7FFE, 0x90, 0xAA90, 0xAA90, 0xAA90, 0x8090, 0x510, 0x210}},
    {39374, 2, {0x8, 0x7F08, 0x4808, 0x7F08, 0x483E, 0x7F2A, 0x482A, 0x482A, 0x7FAA, 0xBE, 0xAAAA, 0xAAAA, 0xAAAA, 0x80AA, 0x53E, 0x222}},
    {39375, 2, {0x0, 0x7F3E, 0x4820, 0x7F20, 0x4820, 0x7F3E, 0x4822, 0x4822, 0x7FA2, 0xA2, 0xAABE, 0xAAA0, 0xAAA0, 0x80A0, 0x53E, 0x200}},
    {39376, 2, {0x10, 0x7F08, 0x4800, 0x7F3E, 0x4808, 0x7F08, 0x4808, 0x4808, 0x7FBE, 0x88, 0xAA88, 0xAA88, 0xAA88, 0x8088, 0x57E, 0x200}},
    {39377, 2, {0x800, 0x7F7C, 0x1144, 0x3228, 0xC10, 0x1228, 0x6146, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x492C}},
    {39378, 2, {0x20, 0x7F20, 0x4820, 0x7F7E, 0x4842, 0x7F82, 0x483A, 0x482A, 0x7FAA, 0xAA, 0xAABA, 0xAAAA, 0xAA82, 0x8082, 0x514, 0x208}},
    {39379, 2, {0x0, 0x7F3E, 0x4808, 0x7F08, 0x4808, 0x7F18, 0x481C, 0x482A, 0x7FAA, 0xCA, 0xAA88, 0xAA88, 0xAA88, 0x8080, 0x53E, 0x200}},
    {39380, 2, {0x0, 0x7F3C, 0x4824, 0x7F24, 0x4824, 0x7F3C, 0x4824, 0x4824, 0x7FA4, 0xBC, 0xAAA4, 0xAAA4, 0xAAA4, 0x80A4, 0x57E, 0x200}},
    {39381, 2, {0x1000, 0x1000, 0x7E7C, 0x1244, 0x2244, 0x2A7C, 0x4400, 0x9FF0, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x492C}},
    {39382, 2, {0x10, 0x7F10, 0x4828, 0x7F28, 0x4844, 0x7F92, 0x4808, 0x4800, 0x7FBC, 0x84, 0xAA88, 0xAAA8, 0xAA90, 0x8088, 0x504, 0x200}},
    {39383, 2, {0x10, 0x7F10, 0x4828, 0x7F28, 0x4844, 0x7F92, 0x4820, 0x4840, 0x7F88, 0x90, 0xAAA0, 0xAA84, 0xAA88, 0x8090, 0x520, 0x240}},
    {39384, 2, {0x8, 0x7F08, 0x4810, 0x7F10, 0x4824, 0x7F7E, 0x4822, 0x4800, 0x7FBC, 0xA4, 0xAAA4, 0xAAA4, 0xAAA4, 0x80A4, 0x53C, 0x224}},
    {39385, 2, {0x48, 0x7E48, 0x4848, 0x7E88, 0x48FE, 0x7E88, 0x4988, 0x4888, 0x7EC8, 0x2A8, 0xAAA8, 0xAA88, 0xAA88, 0x8288, 0x14A8, 0x890}},
    {39386, 2, {0x20, 0x7C20, 0x5020, 0x7DFC, 0x5124, 0x7D24, 0x5124, 0x5124, 0x7D24, 0x7FE, 0x5420, 0x5450, 0x5450, 0x8488, 0x1504, 0xA02}},
    {39387, 2, {0x8, 0x7F08, 0x4808, 0x7F3E, 0x482A, 0x7F2A, 0x482A, 0x482A, 0x7FBE, 0x88, 0xAAC8, 0xAAA8, 0xAA90, 0x8098, 0x524, 0x242}},
    {39388, 2, {0x10, 0x7E08, 0x4808, 0x7E02, 0x4822, 0x7E24, 0x4824, 0x48A8, 0x7EAC, 0x2B2, 0xAB32, 0xAA20, 0xAA62, 0x82A2, 0x151E, 0x800}},
    {39389, 2, {0x10, 0x7F08, 0x4808, 0x7F7E, 0x4842, 0x7F42, 0x4820, 0x4824, 0x7FA8, 0xB0, 0xAAA0, 0xAAA2, 0xAAA2, 0x80A2, 0x51E, 0x200}},
    {39390, 2, {0x40, 0x7C40, 0x507E, 0x7C80, 0x5120, 0x7CA0, 0x50AC, 0x50B4, 0x7DE4, 0x4A4, 0x54B4, 0x54A8, 0x54A2, 0x8482, 0x147E, 0x800}},
    {39391, 2, {0x0, 0x7E00, 0x48FE, 0x7EAA, 0x48AA, 0x7EAA, 0x48AA, 0x48AA, 0x7EAA, 0x2AA, 0xAAAE, 0xAAC2, 0xAA82, 0x8282, 0x14FE, 0x882}},
    {39392, 2, {0x0, 0x7E20, 0x48CE, 0x7EAA, 0x48AA, 0x7EAA, 0x48AA, 0x48AA, 0x7EAA, 0x2AA, 0xAAEA, 0xAAAC, 0xAA28, 0x8248, 0x1448, 0x888}},
    {39393, 2, {0x0, 0x3EF8, 0x2288, 0x3EF8, 0x0, 0x1FF8, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x4, 0x2924, 0x4494, 0x8}},
    {39394, 2, {0x84, 0x7E48, 0x4800, 0x7EFC, 0x4848, 0x7E48, 0x4848, 0x4848, 0x7FFE, 0x248, 0xAA48, 0xAA48, 0xAA48, 0x8288, 0x1488, 0x908}},
    {39395, 2, {0x50, 0x7E50, 0x4850, 0x7F52, 0x48D4, 0x7E58, 0x4850, 0x4858, 0x7ED4, 0x352, 0xAA50, 0xAA50, 0xAA92, 0x8292, 0x1512, 0xA0E}},
    {39396, 2, {0x0, 0x7F7E, 0x4810, 0x7F10, 0x4824, 0x7F22, 0x487E, 0x480A, 0x7F88, 0x88, 0xAABE, 0xAA88, 0xAA88, 0x8088, 0x57E, 0x200}},
    {39397, 2, {0x14, 0x7E12, 0x4812, 0x7E10, 0x48FE, 0x7E10, 0x4852, 0x4852, 0x7E52, 0x2F4, 0xAA54, 0xAA54, 0xAA8A, 0x828A, 0x1516, 0x822}},
    {39398, 2, {0x1110, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x3018, 0xDFF6, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x492C}},
    {39399, 2, {0x0, 0x7EFE, 0x4882, 0x7E82, 0x48BA, 0x7E82, 0x4882, 0x48BA, 0x7EAA, 0x2AA, 0xAAAA, 0xAABA, 0xAA82, 0x8282, 0x148A, 0x884}},
    {39400, 2, {0x10, 0x7F10, 0x4810, 0x7F3E, 0x4822, 0x7F42, 0x483A, 0x482A, 0x7FAA, 0xBA, 0xAAAA, 0xAAAA, 0xAABA, 0x8082, 0x514, 0x208}},
    {39401, 2, {0x8, 0x7F08, 0x4814, 0x7F14, 0x4822, 0x7F40, 0x483E, 0x4808, 0x7F88, 0x88, 0xAABE, 0xAA88, 0xAA88, 0x8088, 0x57E, 0x200}},
    {39402, 2, {0x10, 0x7E50, 0x4850, 0x7E7C, 0x4890, 0x7E10, 0x4810, 0x48FE, 0x7E28, 0x228, 0xAA28, 0xAA28, 0xAA4A, 0x824A, 0x1486, 0x900}},
    {39403, 2, {0x20, 0x7E20, 0x4924, 0x7EA4, 0x48A8, 0x7E20, 0x49FE, 0x4850, 0x7E50, 0x250, 0xAA50, 0xAA52, 0xAA52, 0x8292, 0x148E, 0x900}},
    {39404, 2, {0x0, 0x7F7E, 0x4824, 0x7F24, 0x4824, 0x7F3C, 0x4824, 0x4824, 0x7FBC, 0xA4, 0xAAA4, 0xAAAE, 0xAAF4, 0x8084, 0x504, 0x204}},
    {39405, 2, {0x20, 0x7C10, 0x5010, 0x7DFE, 0x5020, 0x7C24, 0x5044, 0x50F8, 0x7C12, 0x422, 0x5444, 0x5588, 0x5410, 0x8428, 0x1444, 0x982}},
    {39406, 2, {0x20, 0x7E10, 0x4810, 0x7EFE, 0x4800, 0x7E44, 0x4882, 0x4900, 0x7E44, 0x244, 0xAA28, 0xAA28, 0xAA10, 0x8228, 0x1444, 0x982}},
    {39407, 2, {0x10, 0x7E50, 0x4850, 0x7E7C, 0x4890, 0x7E10, 0x4810, 0x48FE, 0x7E10, 0x238, 0xAA38, 0xAA54, 0xAA54, 0x8292, 0x1410, 0x810}},
    {39408, 2, {0x0, 0x7EFE, 0x4882, 0x7E92, 0x4892, 0x7E92, 0x48FE, 0x4892, 0x7E92, 0x2AA, 0xAAA6, 0xAAC6, 0xAA82, 0x8282, 0x14FE, 0x882}},
    {39409, 2, {0x40, 0x7E40, 0x4878, 0x7E88, 0x4950, 0x7E20, 0x4850, 0x4888, 0x7F06, 0x2F8, 0xAA88, 0xAA88, 0xAA88, 0x8288, 0x14F8, 0x888}},
    {39410, 2, {0x82, 0x7C92, 0x5092, 0x7C92, 0x5092, 0x7C92, 0x52DA, 0x52B6, 0x7C92, 0x492, 0x5492, 0x5492, 0x5492, 0x8512, 0x1502, 0xA02}},
    {39411, 2, {0x4, 0xF80E, 0xA3B8, 0xF888, 0xA088, 0xF928, 0xA12E, 0xA3A8, 0xF8A8, 0xAA8, 0xAAA8, 0xA93E, 0xA900, 0x8A80, 0xC7E, 0x3800}},
    {39412, 2, {0x10, 0xF814, 0xA012, 0xF810, 0xA3FE, 0xF810, 0xA150, 0xA154, 0xFFF4, 0x954, 0xA954, 0xA948, 0xAA4A, 0x8A5A, 0xC26, 0x3042}},
    {39413, 2, {0x0, 0x7EFE, 0x4828, 0x7E28, 0x4828, 0x7EEE, 0x48AA, 0x48AA, 0x7EAA, 0x2AA, 0xAAEE, 0xAA28, 0xAA28, 0x8228, 0x1448, 0x888}},
    {39414, 2, {0x0, 0x7EFC, 0x4884, 0x7E84, 0x48FC, 0x7E80, 0x48FE, 0x4882, 0x7EBA, 0x2AA, 0xAAAA, 0xAABA, 0xAA82, 0x8282, 0x1494, 0x908}},
    {39415, 2, {0x10, 0x7E10, 0x48FE, 0x7E10, 0x4810, 0x7EFE, 0x4892, 0x4892, 0x7EFE, 0x210, 0xAA38, 0xAA54, 0xAA92, 0x8310, 0x1410, 0x810}},
    {39416, 2, {0x0, 0x7EFC, 0x4804, 0x7E7C, 0x4804, 0x7EFC, 0x4800, 0x49FE, 0x7F02, 0x2FC, 0xAA44, 0xAA44, 0xAA28, 0x8210, 0x1468, 0x986}},
    {39417, 2, {0x48, 0x7E44, 0x4840, 0x7EFE, 0x4850, 0x7E54, 0x4858, 0x4852, 0x7E54, 0x258, 0xAA52, 0xAA54, 0xAA9A, 0x8292, 0x152E, 0xA40}},
    {39418, 2, {0x20, 0x7E10, 0x48FC, 0x7E84, 0x4884, 0x7EFC, 0x4884, 0x4884, 0x7EFC, 0x2A2, 0xAAA4, 0xAA98, 0xAA90, 0x8288, 0x14C4, 0x882}},
    {39419, 2, {0x0, 0x7F3E, 0x4822, 0x7F3E, 0x4822, 0x7F3E, 0x4800, 0x4800, 0x7FBE, 0x88, 0xAA88, 0xAAFE, 0xAA88, 0x8088, 0x508, 0x208}},
    {39420, 2, {0x10, 0x7E10, 0x4828, 0x7E28, 0x4844, 0x7EBA, 0x4810, 0x4810, 0x7EFE, 0x210, 0xAA58, 0xAA54, 0xAA92, 0x8312, 0x1450, 0x820}},
    {39421, 2, {0x0, 0x7F3E, 0x4822, 0x7F22, 0x483E, 0x7F00, 0x483E, 0x4822, 0x7FA2, 0xBE, 0xAAA2, 0xAABE, 0xAAA2, 0x80A2, 0x52A, 0x224}},
    {39422, 2, {0x84, 0x7E44, 0x4848, 0x7E00, 0x48FC, 0x7E84, 0x4884, 0x4884, 0x7EFC, 0x248, 0xAA48, 0xAA48, 0xAA48, 0x828A, 0x148A, 0x906}},
    {39423, 2, {0x10, 0x7E10, 0x4820, 0x7E44, 0x48FE, 0x7E28, 0x4844, 0x48A2, 0x7E3C, 0x244, 0xAA44, 0xAAA8, 0xAA10, 0x8228, 0x1444, 0x882}},
    {39424, 2, {0x10, 0xF8D8, 0xA394, 0xF894, 0xA090, 0xFBFE, 0xA090, 0xA094, 0xF894, 0x8D8, 0xAB98, 0xA890, 0xA8AA, 0x88CA, 0xA86, 0x3102}},
    {39425, 2, {0x10, 0x7E10, 0x48FC, 0x7E94, 0x48FC, 0x7E94, 0x48FC, 0x4800, 0x7FFE, 0x240, 0xAA80, 0xAAFC, 0xAA04, 0x8204, 0x1428, 0x810}},
    {39426, 2, {0x20, 0x7F10, 0x4810, 0x7F7C, 0x4800, 0x7F28, 0x4810, 0x48FE, 0x7F90, 0x90, 0xAAFC, 0xAA90, 0xAA90, 0x8090, 0x510, 0x210}},
    {39427, 2, {0x10, 0x7E20, 0x4844, 0x7E82, 0x48FE, 0x7E00, 0x4840, 0x487E, 0x7E90, 0x210, 0xAAFE, 0xAA10, 0xAA28, 0x8244, 0x1482, 0x900}},
    {39428, 2, {0x0, 0x7EFC, 0x4804, 0x7E04, 0x487C, 0x7E04, 0x4804, 0x48FE, 0x7E10, 0x292, 0xAA54, 0xAA38, 0xAA54, 0x8292, 0x1450, 0x820}},
    {39429, 2, {0x50, 0x7E48, 0x4880, 0x7EFE, 0x4990, 0x7E90, 0x48FC, 0x4890, 0x7E90, 0x2FC, 0xAA90, 0xAA90, 0xAA90, 0x82FE, 0x1480, 0x880}},
    {39430, 2, {0x0, 0x7EFE, 0x4892, 0x7E92, 0x48BA, 0x7E92, 0x48FE, 0x4882, 0x7EBA, 0x2AA, 0xAAAA, 0xAABA, 0xAA82, 0x8282, 0x148A, 0x904}},
    {39431, 2, {0x20, 0xF820, 0xA050, 0xF888, 0xA104, 0xFAFA, 0xA020, 0xA020, 0xFBFE, 0x820, 0xA820, 0xA9FC, 0xA904, 0x8904, 0x9FC, 0x3104}},
    {39432, 2, {0x22, 0x7E22, 0x4844, 0x7EEE, 0x4844, 0x7E44, 0x4844, 0x4844, 0x7EEE, 0x244, 0xAA44, 0xAA44, 0xAA44, 0x8244, 0x1444, 0x884}},
    {39433, 2, {0x0, 0x7EFC, 0x4884, 0x7E84, 0x48FC, 0x7E84, 0x4884, 0x48FC, 0x7E00, 0x290, 0xAA92, 0xAAF4, 0xAA98, 0x8292, 0x14D2, 0x88E}},
    {39434, 2, {0x40, 0x7E40, 0x48FE, 0x7E82, 0x4942, 0x7E7A, 0x48A2, 0x4822, 0x7EFA, 0x222, 0xAAAA, 0xAAAA, 0xAAFA, 0x8202, 0x1414, 0x808}},
    {39435, 2, {0x10, 0x7E10, 0x4810, 0x7EFE, 0x4810, 0x7E54, 0x4854, 0x4854, 0x7EBA, 0x210, 0xAA38, 0xAA54, 0xAA92, 0x8210, 0x1410, 0x810}},
    {39436, 2, {0x20, 0x7E10, 0x48FE, 0x7E82, 0x4800, 0x7E7C, 0x4800, 0x4800, 0x7EFE, 0x210, 0xAA10, 0xAA54, 0xAA52, 0x8292, 0x1450, 0x820}},
    {39437, 2, {0x0, 0x7E7C, 0x4854, 0x7E54, 0x487C, 0x7E54, 0x4854, 0x487C, 0x7E10, 0x2FE, 0xAA10, 0xAA38, 0xAA54, 0x8292, 0x1410, 0x810}},
    {39438, 2, {0x20, 0x7E20, 0x49FC, 0x7E20, 0x4850, 0x7E88, 0x49FE, 0x4808, 0x7EE8, 0x2A8, 0xAAA8, 0xAAE8, 0xAAA8, 0x8208, 0x1428, 0x810}},
    {39439, 2, {0x24, 0x7F24, 0x487E, 0x7F24, 0x4824, 0x7F3C, 0x4824, 0x4824, 0x7FBC, 0xA4, 0xAAA4, 0xAAFE, 0xAA80, 0x80A4, 0x522, 0x242}},
    {39440, 2, {0x20, 0xF820, 0xA050, 0xF888, 0xA144, 0xFA22, 0xA1F8, 0xA008, 0xF850, 0x820, 0xA8A4, 0xAA82, 0xAA8A, 0x8A8A, 0xC78, 0x3000}},
    {39441, 2, {0x28, 0x7E28, 0x4828, 0x7EEE, 0x4828, 0x7E28, 0x4828, 0x48EE, 0x7E28, 0x228, 0xAA28, 0xAAEE, 0xAA28, 0x8228, 0x1428, 0x828}},
    {39442, 2, {0x0, 0x7EFC, 0x4844, 0x7E28, 0x4810, 0x7E68, 0x4996, 0x4810, 0x7EFE, 0x292, 0xAA92, 0xAAFE, 0xAA10, 0x8214, 0x15FE, 0x882}},
    {39443, 2, {0x10, 0x7E28, 0x4844, 0x7E82, 0x487C, 0x7E10, 0x4810, 0x48FE, 0x7E92, 0x292, 0xAAFE, 0xAA10, 0xAA28, 0x8228, 0x1444, 0x882}},
    {39444, 2, {0x0, 0x7EFC, 0x4884, 0x7EFC, 0x4884, 0x7EFC, 0x4840, 0x48FE, 0x7F12, 0x292, 0xAAAA, 0xAA82, 0xAAFA, 0x8202, 0x1414, 0x808}},
    {39445, 2, {0x0, 0x7EFE, 0x4828, 0x7E28, 0x48FE, 0x7EAA, 0x48AA, 0x48FE, 0x7E10, 0x3FE, 0xAA24, 0xAA44, 0xAA68, 0x8218, 0x1464, 0x982}},
    {39446, 2, {0x7E20, 0x243E, 0x1848, 0xFFA8, 0x2910, 0x4A28, 0x9846, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x492C}},
    {39447, 2, {0x1000, 0x87C, 0x7E50, 0x427C, 0x4250, 0x7E7C, 0x4250, 0x4050, 0x7F7C, 0x5504, 0x5554, 0x7F54, 0xD554, 0x5584, 0x5514, 0x4308}},
    {39448, 2, {0x40, 0x7E40, 0x487E, 0x7EAA, 0x496A, 0x7E5A, 0x4896, 0x4922, 0x7E4A, 0x284, 0xAA20, 0xAA94, 0xAB42, 0x8244, 0x143C, 0x800}},
    {39449, 2, {0x40, 0x7C20, 0x51FE, 0x7D02, 0x5102, 0x7DFE, 0x5100, 0x5100, 0x7DFE, 0x5AA, 0x56AA, 0x56FE, 0x54AA, 0x84AA, 0x14A2, 0x886}},
    {39450, 2, {0x84, 0x7C44, 0x5048, 0x7DFE, 0x5000, 0x7DE2, 0x512A, 0x512A, 0x7DEA, 0x52A, 0x552A, 0x55EA, 0x552A, 0x8522, 0x152A, 0x964}},
    {39451, 2, {0x0, 0x7DFA, 0x500C, 0x7C4A, 0x5188, 0x7CAA, 0x50A6, 0x50A2, 0x7DF8, 0x4AA, 0x54AC, 0x54AA, 0x54A8, 0x852A, 0x1526, 0xA22}},
    {39452, 2, {0x10, 0x7E20, 0x487C, 0x7E44, 0x487C, 0x7E44, 0x487C, 0x4800, 0x7EFE, 0x210, 0xAA10, 0xAA7C, 0xAA10, 0x8210, 0x14FE, 0x800}},
    {39453, 2, {0x10, 0xF810, 0xA03C, 0xFB94, 0xA0FE, 0xF894, 0xA13C, 0xA110, 0xFBBC, 0x890, 0xA8FE, 0xAA90, 0xA910, 0x8980, 0xA7E, 0x3400}},
    {39454, 2, {0x10, 0x7E10, 0x48FE, 0x7E10, 0x48FC, 0x7E10, 0x48FE, 0x4810, 0x7EFE, 0x220, 0xAA40, 0xAAFC, 0xAB44, 0x8244, 0x147C, 0x844}},
    {39455, 2, {0x20, 0x7C50, 0x5088, 0x7D04, 0x52FA, 0x7C00, 0x51E2, 0x512A, 0x7D2A, 0x5EA, 0x552A, 0x552A, 0x55EA, 0x8522, 0x152A, 0x964}},
    {39456, 2, {0x0, 0x7EFC, 0x4884, 0x7E84, 0x48FC, 0x7E84, 0x4884, 0x48FC, 0x7E00, 0x3FE, 0xAA20, 0xAAA0, 0xAABC, 0x82A0, 0x1560, 0xA3E}},
    {39457, 2, {0x10, 0x7E20, 0x48FE, 0x7E82, 0x48FE, 0x7E82, 0x48FE, 0x4810, 0x7E12, 0x2D4, 0xAA58, 0xAA54, 0xAA94, 0x8292, 0x1550, 0x820}},
    {39458, 2, {0x0, 0x7DDC, 0x5144, 0x7D44, 0x5144, 0x7DDC, 0x5100, 0x5100, 0x7DDC, 0x514, 0x5514, 0x55D4, 0x5508, 0x8508, 0x1514, 0x922}},
    {39459, 2, {0x28, 0x7E92, 0x48AA, 0x7E82, 0x48FE, 0x7E28, 0x4844, 0x48A2, 0x7E3C, 0x244, 0xAA44, 0xAAA8, 0xAA10, 0x8228, 0x1444, 0x882}},
    {39460, 2, {0x10, 0x7ED4, 0x4848, 0x7F4A, 0x4884, 0x7E84, 0x4902, 0x48F8, 0x7E20, 0x220, 0xABFC, 0xAA20, 0xAA50, 0x8248, 0x1484, 0x904}},
    {39461, 2, {0x0, 0x7CFC, 0x5048, 0x7C30, 0x51FE, 0x7C52, 0x5094, 0x5110, 0x7E30, 0x420, 0x55FE, 0x5470, 0x54A8, 0x8524, 0x1622, 0x820}},
    {39462, 2, {0x0, 0x7DFC, 0x5124, 0x7D24, 0x51FC, 0x7D24, 0x5124, 0x51FC, 0x7C20, 0x410, 0x5454, 0x5542, 0x5542, 0x854A, 0x1638, 0x800}},
    {39463, 2, {0x0, 0x7E7C, 0x4844, 0x7E44, 0x4874, 0x7E54, 0x4854, 0x48FE, 0x7E82, 0x2BA, 0xAAAA, 0xAAAA, 0xAABA, 0x8282, 0x148A, 0x884}},
    {39464, 2, {0x242, 0x7D24, 0x5000, 0x7DFC, 0x5124, 0x7D24, 0x51FC, 0x5124, 0x7D24, 0x5FC, 0x5420, 0x5420, 0x57FE, 0x8420, 0x1420, 0x820}},
    {39465, 2, {0x10, 0x7E20, 0x48FE, 0x7E92, 0x4892, 0x7EFE, 0x4892, 0x4892, 0x7EFE, 0x220, 0xAA24, 0xAA5A, 0xAA5E, 0x8290, 0x1492, 0x90E}},
    {39466, 2, {0x50, 0x7E96, 0x4892, 0x7E92, 0x48D6, 0x7E92, 0x4892, 0x48FE, 0x7E10, 0x2FC, 0xAA44, 0xAA44, 0xAA28, 0x8210, 0x1428, 0x8C6}},
    {39467, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x920, 0x1FF0, 0x2908, 0xCFE6, 0x900, 0xFF8, 0x8, 0x1558}},
    {39468, 2, {0x1C, 0xF9E0, 0xA020, 0xFBFE, 0xA0A8, 0xF8AA, 0xA3AC, 0xA0AA, 0xF9AA, 0xAA6, 0xA870, 0xA8A8, 0xA924, 0x8A22, 0xC20, 0x3020}},
    {39469, 2, {0x20, 0x793C, 0x4920, 0x53FE, 0x48A4, 0x6B28, 0x5070, 0x4380, 0x0, 0x3FF8, 0x2100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496}},
    {39470, 2, {0x20, 0x7EDE, 0x488A, 0x7ECA, 0x48AA, 0x7ED2, 0x48A6, 0x4800, 0x7EFE, 0x292, 0xAA92, 0xAAFE, 0xAA92, 0x8292, 0x14FE, 0x882}},
    {39471, 2, {0x20, 0x7E10, 0x48FE, 0x7E00, 0x4844, 0x7E28, 0x48FE, 0x48A2, 0x7E10, 0x2FE, 0xAA20, 0xAA3C, 0xAA24, 0x8244, 0x1454, 0x888}},
    {39472, 2, {0x124, 0x78A8, 0x49FC, 0x4840, 0x4BFE, 0x7908, 0x4BFC, 0x4D22, 0x49FC, 0x7920, 0x49FC, 0x4920, 0x49FE, 0x4802, 0x4AAA, 0x9804}},
    {39473, 2, {0xE, 0x7EF0, 0x4822, 0x7E92, 0x4844, 0x7E20, 0x4844, 0x48F8, 0x7E10, 0x224, 0xAAFE, 0xAA10, 0xAAFE, 0x8228, 0x1444, 0x882}},
    {39474, 2, {0x88, 0x7C88, 0x53FE, 0x7C88, 0x5000, 0x7DFC, 0x5104, 0x51FC, 0x7D04, 0x5FC, 0x5524, 0x5420, 0x57FE, 0x8420, 0x1420, 0x820}},
    {39475, 2, {0x0, 0x7EFE, 0x4890, 0x7EFC, 0x4890, 0x7EFC, 0x4890, 0x4890, 0x7EFE, 0x202, 0xAAAA, 0xAAAA, 0xAAAA, 0x8282, 0x1414, 0x808}},
    {39476, 2, {0x0, 0x7E7C, 0x4844, 0x7E7C, 0x4844, 0x7E7C, 0x4810, 0x48FE, 0x7E82, 0x220, 0xAAFE, 0xAA44, 0xAA68, 0x8210, 0x1428, 0x844}},
    {39477, 2, {0x0, 0x7EFE, 0x4888, 0x7E90, 0x48BE, 0x7EA2, 0x48BE, 0x48A2, 0x7EBE, 0x288, 0xAAAC, 0xAAAA, 0xAACA, 0x828A, 0x1528, 0xA10}},
    {39478, 2, {0x40, 0x7E40, 0x48FE, 0x7F22, 0x48AA, 0x7EFA, 0x4842, 0x4894, 0x7F48, 0x2FE, 0xAB22, 0xAAAA, 0xAAFA, 0x8242, 0x1494, 0x908}},
    {39479, 2, {0x0, 0x7DFC, 0x50A4, 0x7C88, 0x5050, 0x7C20, 0x50D8, 0x5326, 0x7C20, 0x5FC, 0x5524, 0x5524, 0x55FC, 0x8422, 0x17FE, 0x802}},
    {39480, 2, {0x20, 0x7C10, 0x51FE, 0x7D02, 0x5102, 0x7DFE, 0x5100, 0x51EE, 0x7D22, 0x5AA, 0x5566, 0x5522, 0x5566, 0x86AA, 0x1622, 0xC66}},
    {39481, 2, {0x88, 0x7C88, 0x53FE, 0x7C88, 0x50F8, 0x7C20, 0x51FC, 0x5124, 0x7DFC, 0x420, 0x55FC, 0x5420, 0x55FC, 0x8420, 0x17FE, 0x800}},
    {39482, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0x882, 0x3FF8, 0x2100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496}},
    {39483, 2, {0x10, 0x7E10, 0x48FE, 0x7E10, 0x48BA, 0x7E54, 0x48BA, 0x4810, 0x7EBA, 0x254, 0xAABA, 0xAA10, 0xAA28, 0x8228, 0x1444, 0x882}},
    {39484, 2, {0x20, 0x7E10, 0x48FE, 0x7EA8, 0x48A8, 0x7EFE, 0x48AA, 0x48AA, 0x7EFE, 0x280, 0xAACA, 0xAAEC, 0xAB48, 0x824A, 0x146A, 0x846}},
    {39485, 2, {0x0, 0x7DDE, 0x5042, 0x7D52, 0x50CA, 0x7D52, 0x5042, 0x5020, 0x7CFC, 0x484, 0x5484, 0x54FC, 0x5484, 0x8484, 0x14FC, 0x884}},
    {39486, 2, {0x0, 0x7DFC, 0x5124, 0x7DFC, 0x5124, 0x7DFC, 0x5040, 0x5088, 0x7DF0, 0x420, 0x5444, 0x57FE, 0x5422, 0x8524, 0x1622, 0x860}},
    {39487, 2, {0x10, 0x7EFE, 0x4844, 0x7E28, 0x48FE, 0x7E00, 0x487C, 0x4844, 0x7E7C, 0x244, 0xAA7C, 0xAA10, 0xAAFE, 0x8210, 0x1410, 0x810}},
    {39488, 2, {0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x3FF8, 0xD106, 0x1FF0, 0x1100, 0x1FFC, 0x2A44, 0x452C}},
    {39489, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x7FFA, 0xA100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496}},
    {39490, 2, {0x10, 0x7E20, 0x4844, 0x7EFE, 0x4844, 0x7EAA, 0x48EE, 0x4810, 0x7E28, 0x254, 0xAAAA, 0xAA50, 0xAA24, 0x8248, 0x1410, 0x860}},
    {39491, 2, {0x0, 0x7EFE, 0x4828, 0x7EFE, 0x48AA, 0x7EAA, 0x48FE, 0x4800, 0x7E7C, 0x200, 0xAAFE, 0xAA10, 0xAA54, 0x8292, 0x1450, 0x820}},
    {39492, 2, {0x20, 0x7C40, 0x51FC, 0x7D24, 0x517C, 0x7D8C, 0x5154, 0x5124, 0x7D54, 0x5FC, 0x5420, 0x5410, 0x5554, 0x8542, 0x154A, 0xA38}},
    {39493, 2, {0x0, 0xFBFE, 0xA200, 0xFA7C, 0xA244, 0xFA44, 0xA27C, 0xA200, 0xFAEE, 0xAAA, 0xAAAA, 0xAAAA, 0xAAEE, 0x8A00, 0xBFE, 0x3000}},
    {39494, 2, {0x0, 0x7EFE, 0x4892, 0x7EFE, 0x4892, 0x7EFE, 0x4854, 0x4854, 0x7EFE, 0x254, 0xAA54, 0xAAFE, 0xAA10, 0x82FE, 0x1410, 0x810}},
    {39495, 2, {0x820, 0x3E20, 0x8F8, 0x7F28, 0x2A68, 0x5D2A, 0x856, 0x3E82, 0x0, 0x3FF8, 0x2100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496}},
    {39496, 2, {0xFC, 0x7E28, 0x4810, 0x7FFE, 0x4852, 0x7E94, 0x4950, 0x4820, 0x7EFE, 0x2AA, 0xAAC6, 0xAABA, 0xAAAA, 0x82BA, 0x1482, 0x886}},
    {39497, 2, {0x20, 0x7C3C, 0x5020, 0x7DFE, 0x5122, 0x7D38, 0x51E4, 0x511C, 0x7D00, 0x528, 0x55AA, 0x556C, 0x5528, 0x8628, 0x14FE, 0x800}},
    {39498, 2, {0x44, 0x7EFE, 0x4844, 0x7E00, 0x48FE, 0x7E54, 0x4854, 0x48FE, 0x7E54, 0x254, 0xAAFE, 0xAA10, 0xAAFE, 0x8210, 0x1410, 0x810}},
    {39499, 2, {0x24, 0xFBA8, 0xA092, 0xFD14, 0xA208, 0xFC04, 0xA3BA, 0xA0A8, 0xF8C6, 0xB80, 0xAA3C, 0xAB84, 0xA8A8, 0x8890, 0xAA8, 0x3144}},
    {39500, 2, {0x20, 0xF9FC, 0xA024, 0xFBFE, 0xA024, 0xF9FC, 0xA020, 0xA2AA, 0xFBAE, 0xA22, 0xABFE, 0xAA22, 0xABAE, 0x8AAA, 0xAAA, 0x3422}},
    {39501, 2, {0x20, 0xF820, 0xA1FC, 0xF820, 0xA3FE, 0xF908, 0xA39C, 0xA108, 0xF988, 0xE3E, 0xA800, 0xABFE, 0xA890, 0x8890, 0x912, 0x320E}},
    {39502, 2, {0x20, 0xF924, 0xA0A8, 0xFBFE, 0xA0A8, 0xF924, 0xA222, 0xA104, 0xF904, 0x9DE, 0xAA44, 0xAD54, 0xA89E, 0x8904, 0xA04, 0x3404}},
    {39503, 2, {0x3FE, 0xFA02, 0xA3FE, 0xFA00, 0xA2FC, 0xFA08, 0xA3FE, 0xA210, 0xFA30, 0xBCE, 0xAA42, 0xAA84, 0xACEE, 0x8F84, 0x884, 0x318C}},
    {39504, 2, {0x108, 0xF888, 0xA3C8, 0xF810, 0xA3DE, 0xFA64, 0xA3D4, 0xA014, 0xFBD4, 0x854, 0xA894, 0xA8E8, 0xAB88, 0x8894, 0xA94, 0x3122}},
    {39505, 2, {0x0, 0x7EFE, 0x4828, 0x7EEE, 0x48AA, 0x7EAA, 0x48EE, 0x4828, 0x7EFE, 0x292, 0xAA92, 0xAAFE, 0xAA92, 0x8292, 0x14FE, 0x882}},
    {39506, 2, {0x0, 0x7EEE, 0x48AA, 0x7EAA, 0x48EE, 0x7E00, 0x48FE, 0x4892, 0x7EFE, 0x292, 0xAAFE, 0xAA10, 0xABFE, 0x8210, 0x1410, 0x810}},
    {39507, 2, {0x44, 0x7E24, 0x4828, 0x7EFE, 0x4892, 0x7ED6, 0x48BA, 0x4892, 0x7EFE, 0x200, 0xAA7C, 0xAA44, 0xAA7C, 0x8244, 0x147C, 0x844}},
    {39508, 2, {0x0, 0x7EFE, 0x4828, 0x7EFE, 0x48AA, 0x7EFE, 0x4800, 0x487C, 0x7E44, 0x27C, 0xAA44, 0xAA7C, 0xAA10, 0x82FE, 0x1410, 0x810}},
    {39509, 2, {0x18, 0xF9E0, 0xA040, 0xFBFE, 0xA088, 0xF974, 0xA252, 0xA070, 0xF800, 0x9FC, 0xA904, 0xA974, 0xA954, 0x8974, 0x904, 0x310C}},
    {39510, 2, {0x8C, 0xFBEA, 0xA088, 0xFBFE, 0xA008, 0xFBE8, 0xA22A, 0xA3EA, 0xF80A, 0xBEC, 0xA88C, 0xABEC, 0xA88A, 0x88EA, 0xB96, 0x3022}},
    {39511, 2, {0x20, 0x7C20, 0x5050, 0x7C88, 0x5104, 0x7EFA, 0x5000, 0x5000, 0x7DDC, 0x554, 0x5554, 0x55DC, 0x5488, 0x8488, 0x1554, 0xA22}},
    {39512, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x0, 0x77DC, 0x5514, 0x77D4, 0x551C, 0x77D6, 0x5056, 0xB562}},
    {39513, 2, {0x10, 0x7EFE, 0x4800, 0x7EFE, 0x4882, 0x7EBA, 0x48AA, 0x48FE, 0x7E00, 0x27C, 0xAA44, 0xAA7C, 0xAA44, 0x827C, 0x1400, 0x8FE}},
    {39514, 2, {0x2420, 0xFF20, 0x247E, 0x7EC4, 0x8228, 0x7A10, 0x4A28, 0x7AC6, 0x400, 0x3FF8, 0x2100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496}},
    {39515, 2, {0x0, 0x7DFC, 0x5154, 0x7DFC, 0x5020, 0x7DFC, 0x5020, 0x53FE, 0x7C88, 0x450, 0x55FC, 0x5420, 0x57FE, 0x8420, 0x1420, 0x820}},
    {39516, 2, {0x50, 0x7D52, 0x50D4, 0x7C50, 0x51FE, 0x7C88, 0x5050, 0x51FE, 0x7C20, 0x4FC, 0x5420, 0x55FE, 0x54A8, 0x8524, 0x1622, 0x820}},
    {39517, 2, {0x10, 0x7EFE, 0x4810, 0x7E7C, 0x4800, 0x7EFE, 0x4882, 0x487C, 0x7E20, 0x27C, 0xAAC4, 0xAB7C, 0xAA10, 0x82FE, 0x1454, 0x992}},
    {39518, 2, {0x40, 0xF820, 0xA3FE, 0xFA02, 0xA1FC, 0xF948, 0xA250, 0xA1FC, 0xFB04, 0x9FC, 0xA904, 0xA9FC, 0xA904, 0x89FC, 0x888, 0x3104}},
    {39519, 2, {0x3F0, 0x7D2E, 0x51E2, 0x7D2A, 0x51E4, 0x7D34, 0x53EA, 0x5032, 0x7C1E, 0x7E0, 0x54A2, 0x5534, 0x5468, 0x84A4, 0x1722, 0x820}},
    {39520, 2, {0x110, 0xF910, 0xA7FC, 0xF910, 0xA1F0, 0xFA08, 0xA2EA, 0xA6AC, 0xFAA8, 0xAAA, 0xAEEA, 0xAA06, 0xA800, 0x8AA4, 0xA52, 0x3452}},
    {39521, 2, {0x1040, 0x7E7C, 0x2440, 0xFF7C, 0x4204, 0x7E7C, 0x4240, 0x7E78, 0x4242, 0x3FFE, 0x2100, 0x3FF8, 0x2100, 0x3FFE, 0x2922, 0x4496}},
    {39522, 2, {0x40, 0xF87C, 0xA040, 0xFBFC, 0xA244, 0xFBF0, 0xA244, 0xA2FC, 0xFAA8, 0xAF8, 0xAAA8, 0xAAF8, 0xAA00, 0x8DFC, 0x954, 0x33FE}},
    {39523, 2, {0x124, 0x7CA8, 0x51FC, 0x7C40, 0x53FE, 0x7D08, 0x53FC, 0x5522, 0x7DFC, 0x520, 0x55FC, 0x5520, 0x55FE, 0x8402, 0x16AA, 0x804}},
    {39524, 2, {0x10, 0x7EFE, 0x4800, 0x7EEE, 0x48AA, 0x7EEE, 0x4844, 0x48FE, 0x7E44, 0x2FE, 0xAA44, 0xAAFE, 0xAA50, 0x82CA, 0x1564, 0x842}},
    {39525, 2, {0x94, 0xFB98, 0xA092, 0xF98E, 0xA280, 0xF9FC, 0xA124, 0xA1FC, 0xF924, 0x9FC, 0xA888, 0xA9FC, 0xA888, 0x8BFE, 0x888, 0x3104}},
    {39526, 2, {0x1FC, 0xF820, 0xA3FE, 0xFA22, 0xA1AC, 0xF820, 0xA1AC, 0xA000, 0xF89E, 0xBF2, 0xA89E, 0xA9D2, 0xAABE, 0x8C92, 0x89E, 0x3092}},
    {39527, 2, {0x290, 0xFA90, 0xA7D0, 0xFABE, 0xA3CA, 0xF92E, 0xA7DA, 0xA54A, 0xFD7E, 0xFCA, 0xA91E, 0xAFEA, 0xA90A, 0x8902, 0x90A, 0x3104}},
    {39528, 2, {0x222, 0xFBFE, 0xA090, 0xF9FE, 0xA310, 0xFDFE, 0xA110, 0xA1FE, 0xF910, 0x9FE, 0xA900, 0xABFE, 0xAA8A, 0x8B76, 0xA52, 0x3276}},
    {39529, 2, {0x88, 0xFBFE, 0xA088, 0xFBDE, 0xA252, 0xFBDE, 0xA0A0, 0xA090, 0xF9FE, 0x920, 0xABFC, 0xAD20, 0xA9FC, 0x8920, 0x9FE, 0x3100}},
    {39530, 2, {0x3DE, 0xF800, 0xA3DE, 0xFA52, 0xA35A, 0xFA52, 0xA020, 0xA3FE, 0xFA50, 0xBFE, 0xAA52, 0xABFE, 0xAA92, 0x8ADC, 0xA92, 0x34CE}},
    {39531, 2, {0x7FF8, 0x4200, 0x7FF0, 0x4200, 0x7FFC, 0x5244, 0x892C, 0x7E7E, 0x4848, 0x7E7E, 0x4848, 0x7E7E, 0x4848, 0x7E7E, 0x202, 0x9696}},
    {39532, 2, {0x0, 0x7FE0, 0x20, 0x20, 0x1020, 0x1020, 0x1020, 0x1FFC, 0x4, 0x4, 0x4, 0xFFE4, 0x4, 0x4, 0x28, 0x10}},
    {39533, 2, {0x0, 0x7C00, 0x5FC, 0x2484, 0x2484, 0x2484, 0x2488, 0x3E88, 0x250, 0x250, 0x1A20, 0xE220, 0x4250, 0x288, 0x1504, 0xA02}},
    {39534, 2, {0x20, 0x7C20, 0x420, 0x2420, 0x2420, 0x25FE, 0x2420, 0x3E20, 0x220, 0x250, 0x1A50, 0xE250, 0x4288, 0x288, 0x1504, 0xA02}},
    {39535, 2, {0x104, 0xF924, 0x924, 0x4924, 0x4924, 0x4924, 0x4924, 0x7D24, 0x524, 0x524, 0x1D24, 0xE524, 0x4524, 0x624, 0x2A04, 0x1404}},
    {39536, 2, {0x20, 0xF820, 0x820, 0x4920, 0x492C, 0x4934, 0x4964, 0x7FA4, 0x524, 0x534, 0x1D28, 0xE522, 0x4522, 0x502, 0x28FE, 0x1000}},
    {39537, 2, {0x0, 0xF9FE, 0x900, 0x4904, 0x4944, 0x4928, 0x4928, 0x7D10, 0x510, 0x528, 0x1D28, 0xE544, 0x4584, 0x500, 0x29FE, 0x1000}},
    {39538, 2, {0x0, 0x7CFC, 0x484, 0x2484, 0x2484, 0x2484, 0x2484, 0x3EFC, 0x284, 0x284, 0x1A84, 0xE284, 0x4284, 0x284, 0x14FC, 0x884}},
    {39539, 2, {0x8, 0xF888, 0x850, 0x4820, 0x4850, 0x4888, 0x4900, 0x7C08, 0x488, 0x488, 0x1C50, 0xE450, 0x4420, 0x450, 0x2888, 0x1306}},
    {39540, 2, {0x40, 0xF820, 0x820, 0x48FC, 0x4884, 0x4884, 0x4884, 0x7CFC, 0x484, 0x480, 0x1C80, 0xE480, 0x4480, 0x500, 0x2900, 0x1200}},
    {39541, 2, {0x0, 0xF8F8, 0x888, 0x4888, 0x4888, 0x48F8, 0x4888, 0x7C88, 0x488, 0x4F8, 0x1C88, 0xE488, 0x4488, 0x488, 0x2BFE, 0x1000}},
    {39542, 2, {0x20, 0xF820, 0x820, 0x49FC, 0x4924, 0x4924, 0x4924, 0x7D24, 0x5FC, 0x420, 0x1CA0, 0xE460, 0x4460, 0x490, 0x2908, 0x1206}},
    {39543, 2, {0x0, 0xF9FC, 0x954, 0x4954, 0x4954, 0x4954, 0x4954, 0x7D54, 0x594, 0x58C, 0x1D04, 0xE504, 0x4504, 0x5FC, 0x2904, 0x1000}},
    {39544, 2, {0x88, 0xF888, 0x888, 0x4908, 0x497E, 0x4B08, 0x4D08, 0x7D48, 0x528, 0x528, 0x1D08, 0xE508, 0x4508, 0x508, 0x2928, 0x1110}},
    {39545, 2, {0x80, 0xF880, 0x900, 0x49FC, 0x4A04, 0x4C04, 0x49E4, 0x7D24, 0x524, 0x524, 0x1DE4, 0xE524, 0x4404, 0x404, 0x2828, 0x1010}},
    {39546, 2, {0x80, 0xF880, 0x8F8, 0x4908, 0x4A10, 0x4C20, 0x4BFC, 0x7C04, 0x404, 0x404, 0x1DFC, 0xE404, 0x4404, 0x404, 0x2BFC, 0x1004}},
    {39547, 2, {0x40, 0xF820, 0x800, 0x4BFE, 0x4820, 0x4820, 0x4820, 0x7C20, 0x5FC, 0x420, 0x1C20, 0xE420, 0x4420, 0x420, 0x2BFE, 0x1000}},
    {39548, 2, {0x20, 0xF810, 0x810, 0x49FE, 0x4902, 0x4A04, 0x4880, 0x7C88, 0x490, 0x4A0, 0x1CC0, 0xE482, 0x4482, 0x482, 0x287E, 0x1000}},
    {39549, 2, {0x800, 0x7F7C, 0x1144, 0x3228, 0xC10, 0x1228, 0x6146, 0x1FE0, 0x20, 0x820, 0x820, 0xFFC, 0x4, 0x7FC4, 0x14, 0x8}},
    {39550, 2, {0x800, 0x800, 0x7F7C, 0x1144, 0x1144, 0x257C, 0x4200, 0x1FE0, 0x20, 0x820, 0x820, 0xFFC, 0x4, 0x7FC4, 0x14, 0x8}},
    {39551, 2, {0x0, 0xFBFC, 0x904, 0x4888, 0x4850, 0x4820, 0x48D8, 0x7F26, 0x420, 0x5FC, 0x1C20, 0xE420, 0x47FE, 0x420, 0x2820, 0x1020}},
    {39552, 2, {0x20, 0xF820, 0x840, 0x4888, 0x4904, 0x4BFE, 0x4802, 0x7C00, 0x5FC, 0x504, 0x1D04, 0xE504, 0x4504, 0x504, 0x29FC, 0x1104}},
    {39553, 2, {0x80, 0xF880, 0x8BC, 0x4BC0, 0x4850, 0x4824, 0x48D4, 0x7F0C, 0x400, 0x7FE, 0x1C90, 0xE490, 0x4512, 0x512, 0x2A0E, 0x1400}},
    {39554, 2, {0x0, 0x7CF8, 0x4488, 0x4488, 0x7CF8, 0x0, 0x3FE0, 0x20, 0x820, 0x820, 0xFFC, 0x4, 0x7FE4, 0x4, 0x28, 0x10}},
    {39555, 2, {0x0, 0xF9FC, 0x904, 0x4924, 0x4924, 0x4924, 0x49FC, 0x7D24, 0x524, 0x554, 0x1D4C, 0xE58C, 0x4504, 0x504, 0x29FC, 0x1104}},
    {39556, 2, {0x8, 0xF83C, 0x9E0, 0x4820, 0x4820, 0x4BFE, 0x4850, 0x7C88, 0x504, 0x68A, 0x1C88, 0xE488, 0x4488, 0x508, 0x2908, 0x1208}},
    {39557, 2, {0x90, 0xF894, 0x914, 0x4918, 0x4B10, 0x4D32, 0x4952, 0x7D0E, 0x520, 0x420, 0x1FFE, 0xE420, 0x4420, 0x420, 0x2820, 0x1020}},
    {39558, 2, {0x80, 0xF880, 0x8F8, 0x4908, 0x4B10, 0x4CA0, 0x4840, 0x7CA0, 0x518, 0x606, 0x1DF8, 0xE508, 0x4508, 0x508, 0x29F8, 0x1108}},
    {39559, 2, {0x20, 0xF810, 0x810, 0x49FE, 0x4820, 0x4824, 0x4844, 0x7CF8, 0x412, 0x422, 0x1C44, 0xE588, 0x4410, 0x428, 0x2844, 0x1182}},
    {39560, 2, {0x84, 0xF844, 0x848, 0x4800, 0x49FE, 0x4848, 0x4848, 0x7C48, 0x448, 0x7FE, 0x1C48, 0xE448, 0x4448, 0x448, 0x2888, 0x1108}},
    {39561, 2, {0x3FF0, 0x10, 0x1010, 0x1FFC, 0x4, 0x7FE4, 0x28, 0x10, 0xFCFC, 0x404, 0x4444, 0x7E7E, 0x202, 0xF2F2, 0xA0A, 0x404}},
    {39562, 2, {0x0, 0xF3FE, 0x1000, 0x5000, 0x53DE, 0x5252, 0x5252, 0x7A52, 0xB5A, 0xAD6, 0x3A52, 0xCA52, 0xA52, 0xA52, 0x2A52, 0x12D6}},
    {39563, 2, {0x20, 0xF820, 0x9FC, 0x4924, 0x49FC, 0x4924, 0x49FC, 0x7C00, 0x7FE, 0x480, 0x1D00, 0xE5FC, 0x4404, 0x404, 0x2828, 0x1010}},
    {39564, 2, {0x20, 0xF820, 0x850, 0x4850, 0x4888, 0x4904, 0x4AFA, 0x7C00, 0x444, 0x424, 0x1D24, 0xE4A8, 0x4488, 0x410, 0x2BFE, 0x1000}},
    {39565, 2, {0x40, 0xF820, 0x9FC, 0x4800, 0x4888, 0x4850, 0x4BFE, 0x7C20, 0x420, 0x420, 0x1DFC, 0xE420, 0x4420, 0x420, 0x2820, 0x1020}},
    {39566, 2, {0x0, 0xF8FC, 0x804, 0x487C, 0x4804, 0x48FC, 0x4800, 0x7DFE, 0x502, 0x4FC, 0x1C44, 0xE444, 0x4428, 0x410, 0x2868, 0x1186}},
    {39567, 2, {0x40, 0xF840, 0x888, 0x4904, 0x4BFE, 0x4802, 0x4888, 0x7D44, 0x642, 0x4F8, 0x1D88, 0xE650, 0x4420, 0x450, 0x2988, 0x1606}},
    {39568, 2, {0x88, 0xF888, 0xBFE, 0x4888, 0x4888, 0x48F8, 0x4888, 0x7C88, 0x4F8, 0x488, 0x1C88, 0xE7FE, 0x4400, 0x488, 0x2884, 0x1104}},
    {39569, 2, {0x20, 0xF820, 0x9FC, 0x4850, 0x4888, 0x4904, 0x4BFE, 0x7C08, 0x5E8, 0x528, 0x1D28, 0xE5E8, 0x4528, 0x408, 0x2828, 0x1010}},
    {39570, 2, {0x0, 0xF9FC, 0x924, 0x4924, 0x49FC, 0x4924, 0x4924, 0x7DFC, 0x420, 0x7FE, 0x1C70, 0xE4A8, 0x4524, 0x622, 0x2820, 0x1020}},
    {39571, 2, {0x50, 0xF848, 0x880, 0x48FE, 0x4990, 0x4A90, 0x48FC, 0x7C90, 0x490, 0x4FC, 0x1C90, 0xE490, 0x4490, 0x4FE, 0x2880, 0x1080}},
    {39572, 2, {0x40, 0xF820, 0xBFE, 0x4A02, 0x4C04, 0x48F8, 0x4800, 0x7C00, 0x5FE, 0x420, 0x1CA8, 0xE4A4, 0x4522, 0x622, 0x28A0, 0x1040}},
    {39573, 2, {0x20, 0xF820, 0x9FC, 0x4824, 0x4BFE, 0x4824, 0x49FC, 0x7C20, 0x524, 0x524, 0x1D74, 0xE5AC, 0x4524, 0x524, 0x2A24, 0x1424}},
    {39574, 2, {0x40, 0xF848, 0x884, 0x49FE, 0x4820, 0x4BFE, 0x4888, 0x7D24, 0x642, 0x588, 0x1C10, 0xE462, 0x4584, 0x418, 0x2860, 0x1380}},
    {39575, 2, {0x40, 0xF820, 0x9FE, 0x4902, 0x4902, 0x49FE, 0x4900, 0x7D00, 0x5FE, 0x5AA, 0x1EAA, 0xE6FE, 0x44AA, 0x4AA, 0x28A2, 0x1086}},
    {39576, 2, {0x20, 0x793C, 0x4920, 0x53FE, 0x48A4, 0x6B28, 0x5070, 0x4380, 0x1FE0, 0x20, 0x820, 0xFFC, 0x4, 0x7FC4, 0x14, 0x8}},
    {39577, 2, {0x10, 0xFBD4, 0x858, 0x4A52, 0x498C, 0x4888, 0x4904, 0x7EFA, 0x420, 0x420, 0x1DFE, 0xE420, 0x4450, 0x488, 0x2904, 0x1202}},
    {39578, 2, {0x0, 0xF9FC, 0x8A4, 0x4888, 0x4850, 0x4820, 0x48D8, 0x7F26, 0x420, 0x5FC, 0x1D24, 0xE524, 0x45FC, 0x422, 0x2BFE, 0x1002}},
    {39579, 2, {0x7E20, 0x243E, 0x1848, 0xFFA8, 0x2910, 0x4A28, 0x9846, 0x0, 0x1FE0, 0x20, 0x820, 0xFFC, 0x4, 0x7FC4, 0x14, 0x8}},
    {39580, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x5FE2, 0x8020, 0x820, 0xFFC, 0x4, 0x7FD4, 0x8}},
    {39581, 2, {0x40, 0xF9BC, 0x914, 0x4994, 0x4954, 0x49A4, 0x494C, 0x7C00, 0x5FC, 0x524, 0x1D24, 0xE5FC, 0x4524, 0x524, 0x29FC, 0x1104}},
    {39582, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1FD0, 0x2048, 0xC846, 0xFF8, 0x8, 0x7FA8, 0x10}},
    {39583, 2, {0x20, 0xF810, 0x9FE, 0x4902, 0x4902, 0x49FE, 0x4900, 0x7DEE, 0x522, 0x5AA, 0x1D66, 0xE522, 0x4566, 0x6AA, 0x2A22, 0x1466}},
    {39584, 2, {0x0, 0xFBFE, 0x850, 0x49FC, 0x4954, 0x4954, 0x49FC, 0x7C00, 0x5FC, 0x400, 0x1FFE, 0xE420, 0x44A8, 0x524, 0x2AA2, 0x1040}},
    {39585, 2, {0x0, 0xF9FC, 0x924, 0x49FC, 0x4924, 0x49FC, 0x4840, 0x7C88, 0x5F0, 0x420, 0x1C44, 0xE7FE, 0x4422, 0x524, 0x2A22, 0x1060}},
    {39586, 2, {0x20, 0xF840, 0x9FC, 0x4924, 0x497C, 0x498C, 0x4954, 0x7D24, 0x554, 0x5FC, 0x1C20, 0xE410, 0x4554, 0x542, 0x294A, 0x1238}},
    {39587, 2, {0x3FE, 0xF202, 0x13FE, 0x5200, 0x52FC, 0x5208, 0x53FE, 0x7A10, 0xA30, 0xBCE, 0x3A42, 0xCA84, 0xCEE, 0xF84, 0x2884, 0x118C}},
    {39588, 2, {0x7E0, 0xF25C, 0x13C4, 0x5254, 0x53C8, 0x5268, 0x57D4, 0x7864, 0x83C, 0xFC0, 0x3944, 0xCA68, 0x8D0, 0x948, 0x2E46, 0x1040}},
    {39589, 2, {0x94, 0xFB98, 0x892, 0x498E, 0x4A80, 0x49FC, 0x4924, 0x7DFC, 0x524, 0x5FC, 0x1C88, 0xE5FC, 0x4488, 0x7FE, 0x2888, 0x1104}},
    {39590, 2, {0x1FC, 0xF820, 0xBFE, 0x4A22, 0x49AC, 0x4820, 0x49AC, 0x7C00, 0x49E, 0x7F2, 0x1C9E, 0xE5D2, 0x46BE, 0x492, 0x289E, 0x1092}},
    {39591, 2, {0x20, 0xF3FE, 0x1000, 0x51DC, 0x5154, 0x51DC, 0x5088, 0x7BFE, 0x888, 0xBFE, 0x3888, 0xCBFE, 0x894, 0x988, 0x2AA4, 0x10C2}},
    {39592, 2, {0x1FF0, 0x1010, 0x1F90, 0x1090, 0x1090, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1050, 0x1020}},
    {39593, 2, {0x40, 0x7C40, 0x4440, 0x7440, 0x55F8, 0xFE48, 0x8248, 0x7C48, 0x4448, 0x7C48, 0x4448, 0x7C4A, 0x448A, 0x448A, 0x5506, 0x4A00}},
    {39594, 2, {0x0, 0x7C00, 0x44F0, 0x7490, 0x5490, 0xFE90, 0x8290, 0x7CD0, 0x44B0, 0x7C90, 0x4490, 0x7C92, 0x4492, 0x4512, 0x550E, 0x4A00}},
    {39595, 2, {0x40, 0x7C40, 0x4440, 0x75F8, 0x5448, 0xFE48, 0x8248, 0x7D48, 0x44C8, 0x7C48, 0x44A8, 0x7CA8, 0x448A, 0x450A, 0x5506, 0x4A02}},
    {39596, 2, {0x0, 0x7CFC, 0x4420, 0x7420, 0x5420, 0xFE20, 0x8220, 0x7DFE, 0x4420, 0x7C20, 0x4420, 0x7C20, 0x4420, 0x4420, 0x54A0, 0x4840}},
    {39597, 2, {0x0, 0x7CFC, 0x4420, 0x7420, 0x5420, 0xFE20, 0x8220, 0x7DFE, 0x4420, 0x7C20, 0x4420, 0x7C20, 0x4420, 0x4420, 0x5420, 0x4820}},
    {39598, 2, {0x3F10, 0x2114, 0x3D12, 0x2512, 0x2510, 0x7F96, 0x40B8, 0xBF10, 0x2110, 0x3F10, 0x2110, 0x3F10, 0x210A, 0x210A, 0x2506, 0x2202}},
    {39599, 2, {0x40, 0x7C20, 0x4400, 0x75FC, 0x5400, 0xFE00, 0x82F0, 0x7C90, 0x4490, 0x7C90, 0x4490, 0x7C92, 0x4492, 0x4512, 0x550E, 0x4A00}},
    {39600, 2, {0x0, 0x7CF8, 0x4488, 0x7488, 0x5488, 0xFF06, 0x8200, 0x7DFC, 0x4484, 0x7C84, 0x4448, 0x7C50, 0x4420, 0x4450, 0x5488, 0x4B06}},
    {39601, 2, {0x20, 0x7C20, 0x4420, 0x7450, 0x5450, 0xFE88, 0x8306, 0x7C48, 0x4448, 0x7C48, 0x4448, 0x7C48, 0x4448, 0x4488, 0x5488, 0x4908}},
    {39602, 2, {0x40, 0x7C40, 0x44FC, 0x7484, 0x5504, 0xFEF4, 0x8294, 0x7C94, 0x44F4, 0x7C84, 0x4494, 0x7C88, 0x4482, 0x4482, 0x547E, 0x4800}},
    {39603, 2, {0x10, 0x7C10, 0x4410, 0x74FE, 0x5492, 0xFE94, 0x8290, 0x7CFC, 0x44A4, 0x7CA4, 0x44A8, 0x7CA8, 0x4490, 0x4528, 0x5544, 0x4A82}},
    {39604, 2, {0x8, 0x7C28, 0x4428, 0x7428, 0x54A8, 0xFEAA, 0x82BC, 0x7CA8, 0x44A8, 0x7CA8, 0x44A8, 0x7CA8, 0x44AA, 0x44BA, 0x54EA, 0x4986}},
    {39605, 2, {0x20, 0x7C20, 0x4420, 0x7420, 0x55FE, 0xFE70, 0x82A8, 0x7CA8, 0x4524, 0x7D24, 0x4622, 0x7CF8, 0x4420, 0x4420, 0x5420, 0x4820}},
    {39606, 2, {0x4, 0x7C1E, 0x44F0, 0x7490, 0x5490, 0xFE90, 0x8290, 0x7CFE, 0x4490, 0x7C90, 0x4490, 0x7C88, 0x448A, 0x44AA, 0x54D6, 0x488A}},
    {39607, 2, {0x20, 0x7C20, 0x4420, 0x7420, 0x55FE, 0xFE20, 0x8220, 0x7C20, 0x44FC, 0x7C84, 0x4484, 0x7C84, 0x4484, 0x4484, 0x54FC, 0x4884}},
    {39608, 2, {0x20, 0x7C10, 0x4410, 0x75FE, 0x5420, 0xFE24, 0x8244, 0x7CF8, 0x4412, 0x7C22, 0x4444, 0x7D88, 0x4410, 0x4428, 0x5444, 0x4982}},
    {39609, 2, {0x20, 0x7C10, 0x4410, 0x74FE, 0x5400, 0xFE44, 0x8282, 0x7D00, 0x4444, 0x7C44, 0x4428, 0x7C28, 0x4410, 0x4428, 0x5444, 0x4882}},
    {39610, 2, {0x8, 0x7C1C, 0x44E0, 0x7480, 0x5480, 0xFEFE, 0x8280, 0x7C80, 0x44BC, 0x7CA4, 0x44A4, 0x7CA4, 0x44A4, 0x453C, 0x5524, 0x4A00}},
    {39611, 2, {0x20, 0x7C20, 0x45FC, 0x7450, 0x5488, 0xFF04, 0x82FA, 0x7C00, 0x47FE, 0x7C40, 0x4480, 0x7DFC, 0x4404, 0x4404, 0x5428, 0x4810}},
    {39612, 2, {0x40, 0x7C40, 0x4478, 0x7488, 0x5550, 0xFE20, 0x8250, 0x7C88, 0x4506, 0x7CF8, 0x4488, 0x7C88, 0x4488, 0x4488, 0x54F8, 0x4888}},
    {39613, 2, {0x4, 0x7C1E, 0x45E0, 0x7422, 0x5512, 0xFE94, 0x8280, 0x7C20, 0x45FE, 0x7C44, 0x4484, 0x7DC8, 0x4430, 0x4428, 0x54C4, 0x4B02}},
    {39614, 2, {0x0, 0x7DFE, 0x4420, 0x7420, 0x55FC, 0xFF24, 0x8324, 0x7DFC, 0x4524, 0x7D24, 0x45FC, 0x7D20, 0x44A0, 0x4440, 0x54B0, 0x4B0E}},
    {39615, 2, {0x84, 0x7C48, 0x4400, 0x74FC, 0x5448, 0xFE48, 0x8248, 0x7C48, 0x45FE, 0x7C48, 0x4448, 0x7C48, 0x4448, 0x4488, 0x5488, 0x4908}},
    {39616, 2, {0x20, 0x7C40, 0x44FE, 0x7492, 0x5492, 0xFEFE, 0x8292, 0x7CA2, 0x44FE, 0x7C48, 0x4488, 0x7DFE, 0x4408, 0x4408, 0x5408, 0x4808}},
    {39617, 2, {0x0, 0x7CF8, 0x44A8, 0x74A8, 0x54F8, 0xFEA8, 0x82A8, 0x7CF8, 0x4420, 0x7DFC, 0x4420, 0x7C70, 0x44A8, 0x4526, 0x5420, 0x4820}},
    {39618, 2, {0x40, 0x7C20, 0x45FC, 0x7504, 0x5440, 0xFE78, 0x8288, 0x7D50, 0x4420, 0x7CD8, 0x4706, 0x7CF8, 0x4488, 0x4488, 0x54F8, 0x4888}},
    {39619, 2, {0xFC, 0x7CA4, 0x44A4, 0x74FC, 0x54A4, 0xFEA4, 0x82FC, 0x7C20, 0x4420, 0x7DFE, 0x4522, 0x7D2A, 0x45FA, 0x450A, 0x5502, 0x4906}},
    {39620, 2, {0x8, 0x7C08, 0x44FE, 0x7510, 0x54A0, 0xFE7E, 0x82A2, 0x7C22, 0x45BE, 0x7CA2, 0x44BE, 0x7CA2, 0x44AA, 0x44A4, 0x5540, 0x4A3E}},
    {39621, 2, {0x20, 0x7D24, 0x44A8, 0x7420, 0x55FE, 0xFE68, 0x82A4, 0x7D02, 0x4420, 0x7DFE, 0x4448, 0x7C88, 0x44D0, 0x4430, 0x544C, 0x4982}},
    {39622, 2, {0x14, 0x7C12, 0x45FE, 0x7410, 0x54FE, 0xFE92, 0x82FE, 0x7C92, 0x44FE, 0x7C92, 0x4404, 0x7DFE, 0x4444, 0x4424, 0x5424, 0x480C}},
    {39623, 2, {0x20, 0x7C10, 0x45FE, 0x7400, 0x54FC, 0xFE84, 0x82FC, 0x7C00, 0x45FE, 0x7D02, 0x457A, 0x7D4A, 0x457A, 0x4502, 0x550A, 0x4904}},
    {39624, 2, {0x20, 0x7C10, 0x44FE, 0x7400, 0x5444, 0xFE28, 0x82FE, 0x7CA2, 0x4410, 0x7CFE, 0x4420, 0x7C3C, 0x4424, 0x4444, 0x5454, 0x4888}},
    {39625, 2, {0x20, 0x7C20, 0x45FE, 0x7552, 0x5448, 0xFEFE, 0x8290, 0x7D90, 0x44FE, 0x7C90, 0x4490, 0x7CFE, 0x4490, 0x4490, 0x54FE, 0x4880}},
    {39626, 2, {0x84, 0x7C44, 0x4448, 0x75FE, 0x5420, 0xFEFC, 0x8220, 0x7DFE, 0x4440, 0x7C80, 0x44FE, 0x7D10, 0x4610, 0x4410, 0x55FE, 0x4800}},
    {39627, 2, {0x20, 0x7C10, 0x45FE, 0x7502, 0x5448, 0xFEFC, 0x8248, 0x7C00, 0x44FC, 0x7C84, 0x44A4, 0x7CA4, 0x44A4, 0x4450, 0x5492, 0x4B0E}},
    {39628, 2, {0x40, 0x7C20, 0x45FE, 0x7502, 0x5418, 0xFEE0, 0x8280, 0x7CFC, 0x4490, 0x7C90, 0x4490, 0x7DFE, 0x4400, 0x4488, 0x5504, 0x4A02}},
    {39629, 2, {0x20, 0x7C10, 0x45FE, 0x7548, 0x5548, 0xFF48, 0x83FE, 0x7D48, 0x4548, 0x7DEC, 0x455A, 0x7D68, 0x4548, 0x4548, 0x5648, 0x4848}},
    {39630, 2, {0x0, 0x7CEE, 0x4422, 0x74AA, 0x5466, 0xFEAA, 0x8210, 0x7C28, 0x4444, 0x7D92, 0x4420, 0x7CC8, 0x4432, 0x44C4, 0x5418, 0x48E0}},
    {39631, 2, {0x20, 0x7DFC, 0x4524, 0x77FE, 0x5524, 0xFFFC, 0x8220, 0x7DFC, 0x4524, 0x7DFC, 0x4440, 0x7DFE, 0x4488, 0x45D0, 0x5470, 0x4B8C}},
    {39632, 2, {0x20, 0x7C20, 0x44FC, 0x7420, 0x55FE, 0xFE88, 0x83DC, 0x7C88, 0x44C8, 0x7DBC, 0x4400, 0x7DFE, 0x4450, 0x4450, 0x5494, 0x490C}},
    {39633, 2, {0x0, 0x7CFE, 0x44AA, 0x74AA, 0x54FE, 0xFE40, 0x82FE, 0x7D22, 0x44FA, 0x7CAA, 0x44AA, 0x7CFA, 0x4422, 0x44FA, 0x540A, 0x4804}},
    {39634, 2, {0x48, 0x7DFE, 0x4448, 0x7400, 0x55FE, 0xFE88, 0x82EA, 0x7D2C, 0x44CA, 0x7C86, 0x4548, 0x7DFE, 0x4448, 0x4448, 0x5488, 0x4908}},
    {39635, 2, {0x10, 0x7C7E, 0x4520, 0x74BC, 0x54C8, 0xFEBE, 0x8200, 0x7DBC, 0x44A4, 0x7CBC, 0x44A4, 0x7CBC, 0x44A4, 0x44AC, 0x5540, 0x4A3E}},
    {39636, 2, {0x50, 0x7DFC, 0x4554, 0x75FC, 0x5554, 0xFFFC, 0x8200, 0x7DFC, 0x4400, 0x7DFC, 0x4504, 0x7DFC, 0x4488, 0x4450, 0x57FE, 0x4800}},
    {39637, 2, {0x20, 0x7C10, 0x45FE, 0x7502, 0x54FC, 0xFEA8, 0x8330, 0x7CFC, 0x4584, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x44FC, 0x5448, 0x4884}},
    {39638, 2, {0x20, 0x7DFE, 0x454A, 0x74FC, 0x5448, 0xFEFC, 0x8284, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x4450, 0x4454, 0x5492, 0x490E}},
    {39639, 2, {0x20, 0x7C3E, 0x4420, 0x75FE, 0x5522, 0xFFF8, 0x8322, 0x7D7E, 0x4554, 0x7D7C, 0x4554, 0x7D7C, 0x4500, 0x46FE, 0x54AA, 0x49FE}},
    {39640, 2, {0x200, 0x100, 0xFFFE, 0x0, 0xFE0, 0x820, 0x820, 0xFE0, 0x0, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x400C}},
    {39641, 2, {0x200, 0x100, 0x7FFC, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4004, 0x4FE4, 0x4824, 0x4824, 0x4FE4, 0x4004, 0x400C}},
    {39642, 2, {0x2020, 0x1010, 0xFE10, 0x1FE, 0x7C00, 0x4400, 0x7C78, 0x48, 0xFE48, 0x8248, 0xBA48, 0xAA4A, 0xAA4A, 0xBA8A, 0x8286, 0x8700}},
    {39643, 2, {0x2000, 0x10FC, 0xFE84, 0x84, 0x7CFC, 0x44A0, 0x7CA0, 0xA0, 0xFEF8, 0x82A8, 0xBAA8, 0xAAA8, 0xAAA8, 0xBAAA, 0x834A, 0x8686}},
    {39644, 2, {0x2000, 0x107E, 0xFE42, 0x7E, 0x7C42, 0x447E, 0x7C00, 0x14, 0xFE64, 0x8224, 0xBA24, 0xAAFE, 0xAA24, 0xBA24, 0x8244, 0x8684}},
    {39645, 2, {0x2048, 0x116A, 0xFEDC, 0x48, 0x7CB4, 0x4522, 0x7DFE, 0x102, 0xFE20, 0x8220, 0xBAFC, 0xAA24, 0xAA44, 0xBA44, 0x8294, 0x8708}},
    {39646, 2, {0x207C, 0x1044, 0xFE44, 0x7C, 0x7C00, 0x44EE, 0x7CAA, 0xAA, 0xFEEE, 0x8210, 0xBAFE, 0xAA38, 0xAA54, 0xBA92, 0x8310, 0x8610}},
    {39647, 2, {0x3F04, 0x2004, 0x2008, 0x3E10, 0x2020, 0x2044, 0x3E04, 0x2008, 0x2010, 0xFF22, 0x1042, 0x2404, 0x4208, 0xFF10, 0x4120, 0xC0}},
    {39648, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0xFE0, 0x820, 0x820, 0x1022, 0x2022, 0xC01E}},
    {39649, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x7FFC, 0x440, 0x440, 0x842, 0x3042, 0xC03E}},
    {39650, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x130, 0x11D0, 0x1F10, 0x7152, 0x1122, 0x1002, 0xFFE}},
    {39651, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x200, 0x7FFC, 0x400, 0x7F0, 0x810, 0x10A0, 0x6040}},
    {39652, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x100, 0x7FFC, 0x380, 0x540, 0x920, 0x3118, 0xC106}},
    {39653, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0xFF0, 0x810, 0x7FFE, 0x810, 0x7FFE, 0x810, 0x850, 0x820}},
    {39654, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x1F0, 0x3E00, 0x3F0, 0x3E00, 0x3FA, 0x7E02, 0x1FE}},
    {39655, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x100, 0x7FFC, 0x4104, 0x280, 0x482, 0x1882, 0xE07E}},
    {39656, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x440, 0x1842, 0xE03E}},
    {39657, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0xFFFE, 0x100, 0x21FC, 0x2104, 0x3FFC, 0x4, 0x28, 0x10}},
    {39658, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x100, 0xFFFE, 0x400, 0xFF0, 0x1220, 0x61C0, 0xE3C}},
    {39659, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x7FF8, 0x808, 0x1030, 0x3FF8, 0xD008, 0x1FF8}},
    {39660, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x7FFC, 0x300, 0xD70, 0x710C, 0x100, 0xFFFE}},
    {39661, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x880, 0x888, 0x2E90, 0x28E0, 0x2884, 0x2E84, 0xF07C}},
    {39662, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x110, 0xFFFE, 0x400, 0xFF0, 0x1220, 0x61C0, 0xE3C}},
    {39663, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0xFFFE, 0x2008, 0x2028, 0x2010}},
    {39664, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x1220, 0xFFFE, 0x1220, 0x13E0, 0x1000, 0x1FFC}},
    {39665, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x800, 0x1FF0, 0x2010, 0x5F90, 0x1090, 0x1FD0, 0x1020, 0x1004, 0xFFC}},
    {39666, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x80, 0x3FFC, 0x2084, 0x27F0, 0x2220, 0x41C0, 0x9E3C}},
    {39667, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x3FE0, 0x440, 0xFFFC, 0x484, 0x1888, 0xE180}},
    {39668, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x440, 0x7FF8, 0x448, 0x3FF8, 0x2440, 0x3FFC, 0x844, 0x1054, 0x2048}},
    {39669, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0xFFFE, 0x200, 0x3FF8, 0x2488, 0x24A8, 0x2010}},
    {39670, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3FF8, 0x820, 0xFE0, 0x820, 0xFE0, 0x83E, 0xFFE0, 0x20}},
    {39671, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x3FF8, 0x2008}},
    {39672, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {39673, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x840, 0x1040, 0x37FC, 0xD0E0, 0x1358, 0x1C46, 0x1040}},
    {39674, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0xF8, 0x3F00, 0x100, 0xFFFE, 0x100, 0x1FF0, 0x1010, 0x1FF0}},
    {39675, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x2008}},
    {39676, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE10E, 0x1FF0, 0x100, 0x7FFC, 0x100, 0x100}},
    {39677, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x1110, 0x1110, 0x2928, 0x4544, 0x3FF8, 0x100, 0xFFFE}},
    {39678, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1110, 0x920, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x2018}},
    {39679, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x1020, 0x48A8, 0x2524, 0xA22, 0x7028, 0x1030, 0x17C0}},
    {39680, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0x440, 0x7FF8, 0x108, 0x3FF8, 0x2100, 0x3FFC, 0x304, 0x514, 0x1908, 0x6100}},
    {39681, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x8, 0x3E48, 0x848, 0x7F48, 0x1C48, 0x2A08, 0x4918}},
    {39682, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x20, 0x1FC0, 0x1000, 0x1FF8, 0x1080, 0x1080, 0xFFFE, 0x1020, 0x2010}},
    {39683, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x7FFC, 0x4004, 0x1FF0, 0x0, 0x7FFC, 0x1110, 0x2508, 0x4204}},
    {39684, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x6444, 0x894, 0x3308}},
    {39685, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x4A94, 0x8508}},
    {39686, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x1010, 0x1088, 0xFD44, 0x3242, 0x5890, 0x95F8, 0x1008}},
    {39687, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x800, 0x1FF0, 0x2020, 0x5FF8, 0x108, 0xFFFE, 0x108, 0x1FF8, 0x108, 0x500, 0x200}},
    {39688, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1010, 0x2FE8, 0xC826, 0x8A0, 0x848, 0x7F8}},
    {39689, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x440, 0x2388, 0x2448, 0x3FF8, 0x1410, 0x2FE8, 0x1440, 0x380, 0x7C78}},
    {39690, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x200, 0xFFFE, 0x820, 0x1FF0, 0x2828, 0xCFE6, 0x820}},
    {39691, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1010, 0x820, 0xFFFE, 0x0, 0x3E08, 0x2248, 0x3E48, 0x2248, 0x3E48, 0x2208, 0x2618}},
    {39692, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x400, 0x7FFC, 0x880, 0x1FFC, 0x2810, 0x4FF0, 0x8810, 0xFF0, 0x810, 0x850, 0x820}},
    {39693, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x107C, 0xFF44, 0x107C, 0x7E44, 0x427C, 0x7E44, 0x428C}},
    {39694, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x404, 0x7FC4, 0x424, 0x3FA4, 0x24A4, 0x3FA4, 0x1504, 0x2494, 0x4408}},
    {39695, 2, {0x3E04, 0x2018, 0x3E60, 0x2004, 0x3E18, 0x2062, 0xFF04, 0x2218, 0x7F60, 0x0, 0x7C10, 0x1052, 0xFE94, 0x3828, 0x5444, 0x9282}},
    {39696, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x108, 0x1FD0, 0x120, 0x7FFC, 0x620, 0x1C44, 0x63FC, 0x0, 0x1FF8, 0x1108, 0x1FF8}},
    {39697, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x840, 0x7FFC, 0x480, 0x3FF0, 0x490, 0xFFFE, 0x490, 0x3FF0, 0xCA0, 0x3498, 0xC486}},
    {39698, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x7FFC, 0x100, 0x1FF0, 0x1010, 0x1110, 0x1110, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {39699, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x200, 0x100, 0x7FFE, 0x4022, 0x9FC4, 0x1000, 0x1FF8, 0x1080, 0xFFFE, 0x1020, 0x2010}},
    {39700, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x2100, 0x13F0, 0x1510, 0xE0, 0x775C, 0x11F0, 0x1040, 0x13FC, 0x2840, 0x47FE, 0x0}},
    {39701, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0xFFFE, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x820, 0x3018, 0xC006}},
    {39702, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0x1FF0, 0x1020, 0x2548, 0x7EFC, 0x440, 0x18B0, 0xE30E, 0xC60, 0x380, 0x1C00}},
    {39703, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0x7FFC, 0x820, 0xFE0, 0x100, 0x7FFC, 0x5144, 0x4924, 0x5554, 0x6184, 0x410C}},
    {39704, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x7FFC, 0x4444, 0x7FFC, 0x820, 0x7C0, 0xF83E}},
    {39705, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x820, 0x7FFC, 0x5114, 0x4924, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {39706, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0xBFE, 0x1040, 0x61FC, 0x904, 0x11FC, 0x6104, 0x5FC, 0x904, 0x31FC, 0xC088}},
    {39707, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x200, 0x3FF8, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x2948, 0x2528, 0x294A, 0x2526, 0x3182}},
    {39708, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x0, 0x7C7C, 0x4444, 0x7C7C, 0x47C4, 0x7C7C, 0x47C4, 0x4444, 0x47C4, 0x4444, 0x48CC}},
    {39709, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F7C, 0x4444, 0x7C7C, 0x4444, 0x7C7C, 0x4004, 0x4FE4, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x400C}},
    {39710, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x240, 0x1FF8, 0x1248, 0x1FF8, 0x1248, 0x3FFC, 0x2000, 0x3FFC, 0x2888, 0x4A70, 0x8C0E}},
    {39711, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x3FF8, 0x2448, 0xFFFE, 0x1010, 0x1FF0, 0xC88, 0x3470, 0xC60E}},
    {39712, 2, {0x3F08, 0x2030, 0x3EC4, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x280, 0xC60, 0x3FF8, 0xD556, 0x1390, 0x1FF0, 0xA20, 0x920, 0xFE0}},
    {39713, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x7FFC, 0x4204, 0xA928, 0x47E4, 0x0, 0x3FF8, 0x2448, 0xFFFE, 0x100, 0x300}},
    {39714, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0x7FFE, 0x4002, 0x9FF4, 0x920, 0x11C0, 0xFF8, 0x1888, 0x6FF8, 0x410, 0x808}},
    {39715, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x1084, 0x2108, 0x1084, 0x3FF8, 0x2448, 0x2388, 0x2448, 0x3FF8, 0x294A, 0x2526, 0x3182}},
    {39716, 2, {0x3F04, 0x2018, 0xFF62, 0x220C, 0x7F70, 0x100, 0xFFFE, 0x2288, 0x3EF8, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x490, 0x1C60, 0x671C}},
    {39717, 2, {0x0, 0x5E7A, 0x4422, 0x5E7A, 0x4422, 0x5E7A, 0x4002, 0x4002, 0x4002, 0x4002, 0x4002, 0x4002, 0x4002, 0x4002, 0x400A, 0x4004}},
    {39718, 2, {0x0, 0x5E7A, 0x4422, 0x5E7A, 0x4422, 0x5E7A, 0x4082, 0x4482, 0x4282, 0x4882, 0x4482, 0x40F2, 0x5F82, 0x4082, 0x408A, 0x4084}},
    {39719, 2, {0x0, 0x5E7A, 0x4422, 0x5E7A, 0x4422, 0x5E7A, 0x4202, 0x4102, 0x5FF2, 0x4102, 0x4FE2, 0x4922, 0x49A2, 0x4942, 0x410A, 0x4104}},
    {39720, 2, {0x0, 0x5E7A, 0x4422, 0x5E7A, 0x4422, 0x5E7A, 0x4002, 0x4442, 0x4442, 0x4FE2, 0x4442, 0x4442, 0x5FF2, 0x4442, 0x482A, 0x4004}},
    {39721, 2, {0x0, 0x5E7A, 0x4422, 0x5E7A, 0x4422, 0x5E7A, 0x4202, 0x4CE2, 0x4822, 0x4EE2, 0x4822, 0x4FE2, 0x4282, 0x4492, 0x587A, 0x4004}},
    {39722, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8002, 0xBF12, 0x8012, 0x9EFE, 0x9212, 0x9292, 0x9E52, 0x9212, 0x8F12, 0xB852, 0x9026}},
    {39723, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8002, 0x9C42, 0x847A, 0xBF52, 0x9292, 0x9E12, 0x9252, 0x9E22, 0x93A2, 0xFE52, 0x8296}},
    {39724, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8806, 0xB3BA, 0xAAA2, 0xBBA2, 0x8A3E, 0xBBAA, 0xA0AA, 0xBBAA, 0x8A2A, 0x8BAA, 0xBC46}},
    {39725, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8006, 0xBBBA, 0xAAA2, 0xBBA2, 0x8A3E, 0xBBAA, 0xA0AA, 0xBBAA, 0x8A2A, 0x8BAA, 0xBC46}},
    {39726, 2, {0xBEFA, 0x8822, 0xBEFA, 0x8822, 0xBEFA, 0x8102, 0xBFF2, 0xA012, 0xBCF2, 0x8482, 0xB4F2, 0x9492, 0xFCF2, 0x9482, 0xB492, 0x83F6}},
    {39727, 2, {0x920, 0x2448, 0x2288, 0x2928, 0x2288, 0x2448, 0x2928, 0x3FF8, 0x1000, 0x1038, 0x11C0, 0x1E04, 0x1004, 0x1004, 0xFFC, 0x0}},
    {39728, 2, {0x2288, 0xF93E, 0x2288, 0x729C, 0xA92A, 0x2288, 0x0, 0x7FFE, 0x4A12, 0xA4A4, 0x2AC8, 0x3F90, 0x1024, 0x1F48, 0x1010, 0x1F20}},
    {39729, 2, {0x2208, 0xF3BC, 0x2508, 0x739C, 0xAD6A, 0x27C8, 0x0, 0x7FFE, 0x4A12, 0xA4A4, 0x2AC8, 0x3F90, 0x1024, 0x1F48, 0x1010, 0x1F20}},
    {39730, 2, {0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4824, 0x4444, 0x5FF4, 0x4104, 0x4104, 0x4114, 0x4008}},
    {39731, 2, {0x1FC, 0x100, 0x3FFE, 0x2102, 0x21F0, 0x2F04, 0x20FC, 0x2000, 0x3FFE, 0x2208, 0x23F8, 0x2000, 0x2FFE, 0x2912, 0x4BFA, 0x8842}},
    {39732, 2, {0x14, 0xFE12, 0xFE, 0x7C10, 0x4410, 0x7CFE, 0x92, 0xFE92, 0x82FE, 0xAA92, 0x9292, 0xFEFE, 0x9292, 0x9292, 0x9292, 0x8686}},
    {39733, 2, {0x3E7C, 0x4890, 0x7EFC, 0x142A, 0x66C6, 0x400, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4444, 0x4FE4, 0x410C}},
    {39734, 2, {0x10F8, 0x7CA8, 0x10A8, 0xFE52, 0x288E, 0x4500, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4444, 0x4FE4, 0x410C}},
    {39735, 2, {0x28, 0xFE92, 0xAA, 0x7C82, 0x44FE, 0x7C28, 0x44, 0xFEA2, 0x823C, 0xAA44, 0x9244, 0xFEA8, 0x9210, 0x9228, 0x9244, 0x8682}},
    {39736, 2, {0x20, 0xFEDE, 0x8A, 0x7CCA, 0x44AA, 0x7CD2, 0xA6, 0xFE00, 0x82FE, 0xAA92, 0x9292, 0xFEFE, 0x9292, 0x9292, 0x92FE, 0x8682}},
    {39737, 2, {0x8F8, 0x3E88, 0x8F8, 0x7E88, 0x8F8, 0x1452, 0x228E, 0x4100, 0xFFFE, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4444, 0x4FE4, 0x410C}},
    {39738, 2, {0x40, 0xFE7E, 0x80, 0x7D7C, 0x4444, 0x7C7C, 0x44, 0xFE7C, 0x8200, 0xAAFE, 0x9240, 0xFEFE, 0x932A, 0x924A, 0x9292, 0x8626}},
    {39739, 2, {0x711C, 0x1544, 0x711C, 0x47D0, 0x739C, 0x1544, 0x310C, 0x0, 0xFFFE, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x4444, 0x4FE4, 0x410C}},
    {39740, 2, {0x200, 0x400, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2208, 0x2208, 0x3FF8, 0x510, 0x920, 0x948, 0x117C, 0x2102, 0x4102, 0x80FE}},
    {39741, 2, {0x804, 0x1004, 0x7F08, 0x4910, 0x4924, 0x7F04, 0x4908, 0x4910, 0x7F22, 0x1002, 0x1A04, 0x2C88, 0x2F90, 0x4820, 0x4802, 0x87FE}},
    {39742, 2, {0x810, 0x1010, 0x7F7E, 0x4910, 0x4910, 0x7F10, 0x497C, 0x4904, 0x7F44, 0x1028, 0x1A10, 0x2CA8, 0x2FC4, 0x4884, 0x4802, 0x87FE}},
    {39743, 2, {0x808, 0x101C, 0x7F60, 0x4940, 0x4940, 0x7F7E, 0x4948, 0x4948, 0x7F48, 0x1048, 0x1A48, 0x2C88, 0x2F88, 0x4800, 0x4802, 0x87FE}},
    {39744, 2, {0x810, 0x1010, 0x7F28, 0x4928, 0x4944, 0x7F82, 0x4928, 0x4928, 0x7F28, 0x1028, 0x1A28, 0x2CA8, 0x2FA8, 0x4848, 0x4802, 0x87FE}},
    {39745, 2, {0x808, 0x1048, 0x7F28, 0x4928, 0x4908, 0x7F48, 0x4928, 0x4928, 0x7F0E, 0x10F8, 0x1A08, 0x2C88, 0x2F88, 0x4808, 0x4802, 0x87FE}},
    {39746, 2, {0x20, 0x40, 0x7DFC, 0x124, 0x124, 0x1FC, 0xFF24, 0x2124, 0x21FC, 0x2040, 0x4868, 0x44B2, 0xFCBE, 0x4520, 0x122, 0x21E}},
    {39747, 2, {0x1028, 0x2024, 0x7C20, 0x55FE, 0x5420, 0x7C20, 0x547C, 0x5444, 0x7C54, 0x1088, 0x2A94, 0x2D24, 0x4F40, 0x4802, 0x87FE, 0x0}},
    {39748, 2, {0x1020, 0x1040, 0x21FC, 0x7D24, 0x4524, 0x45FC, 0x4524, 0x7D24, 0x45FC, 0x4440, 0x4468, 0x44B2, 0x7CBE, 0x4520, 0x122, 0x21E}},
    {39749, 2, {0x810, 0x1010, 0x7F10, 0x497C, 0x4910, 0x7F10, 0x49FE, 0x4910, 0x7F38, 0x1054, 0x1A92, 0x2C90, 0x2F90, 0x4810, 0x4802, 0x87FE}},
    {39750, 2, {0x1014, 0x2012, 0x7C10, 0x54FE, 0x5490, 0x7C94, 0x5494, 0x5488, 0x7C8A, 0x10DA, 0x2AA6, 0x2D42, 0x4F00, 0x4802, 0x87FE, 0x0}},
    {39751, 2, {0x3FFC, 0x2090, 0x3FFC, 0x2140, 0x2630, 0x388C, 0x2100, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x4150, 0x42FA, 0x8C8A, 0x307E}},
    {39752, 2, {0x810, 0x1054, 0x7F38, 0x4910, 0x497C, 0x7F44, 0x4944, 0x497C, 0x7F44, 0x1044, 0x1A7C, 0x2CC4, 0x2FC4, 0x484C, 0x4802, 0x87FE}},
    {39753, 2, {0x1000, 0x21FC, 0x7C50, 0x5450, 0x55FC, 0x7D54, 0x5554, 0x5554, 0x7DAC, 0x1104, 0x2B04, 0x2D14, 0x4F08, 0x4802, 0x87FE, 0x0}},
    {39754, 2, {0x1010, 0x2014, 0x7C12, 0x57FE, 0x5410, 0x7DD4, 0x5554, 0x55D4, 0x7C18, 0x10CA, 0x2B16, 0x2D22, 0x4F40, 0x4802, 0x87FE, 0x0}},
    {39755, 2, {0x828, 0x1024, 0x7F40, 0x497E, 0x49C8, 0x7F48, 0x497E, 0x4948, 0x7F48, 0x107E, 0x1A48, 0x2CC8, 0x2FFE, 0x4840, 0x4802, 0x87FE}},
    {39756, 2, {0x824, 0x1024, 0x7F7E, 0x4924, 0x4924, 0x7F3C, 0x4924, 0x493C, 0x7F24, 0x1024, 0x1A7E, 0x2C80, 0x2FA8, 0x4844, 0x4802, 0x87FE}},
    {39757, 2, {0x1000, 0x21FC, 0x7D04, 0x558C, 0x5554, 0x7DFC, 0x5544, 0x5524, 0x7DFC, 0x1144, 0x2B74, 0x2D04, 0x4F0C, 0x4802, 0x87FE, 0x0}},
    {39758, 2, {0x23FE, 0x4020, 0xF820, 0xABFE, 0xAA22, 0xFB32, 0xAAAA, 0xAB76, 0xFA66, 0x2222, 0x5622, 0x5A2A, 0x5E04, 0x5002, 0x4FFE, 0x8000}},
    {39759, 2, {0xC20, 0x7040, 0x11FC, 0xFF24, 0x3924, 0x55FC, 0x9324, 0x124, 0x11FC, 0xFC40, 0x2468, 0x44B2, 0x28BE, 0x1120, 0x2922, 0xC61E}},
    {39760, 2, {0x1088, 0x2050, 0x7DFE, 0x5450, 0x55FC, 0x7C54, 0x57FE, 0x5454, 0x7DFC, 0x1050, 0x2AD8, 0x2D54, 0x4F52, 0x4852, 0x87FE, 0x0}},
    {39761, 2, {0x1040, 0x2020, 0x7DFC, 0x5450, 0x5524, 0x7D54, 0x55FC, 0x5420, 0x7DFC, 0x1144, 0x2B54, 0x2D74, 0x4F0C, 0x4802, 0x87FE, 0x0}},
    {39762, 2, {0x1000, 0x21FC, 0x7C50, 0x55FC, 0x5554, 0x7DFC, 0x5400, 0x54F8, 0x7C00, 0x11FC, 0x2AA8, 0x2D24, 0x4F60, 0x4802, 0x87FE, 0x0}},
    {39763, 2, {0x11FC, 0x2124, 0x7DFC, 0x5524, 0x55FC, 0x7CA8, 0x55FC, 0x54A8, 0x7CA8, 0x11FC, 0x2A20, 0x2DFC, 0x4F20, 0x4822, 0x87FE, 0x0}},
    {39764, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3512, 0x2FF8, 0x2888, 0x2FF8, 0x2888, 0x2FF8, 0x2150, 0x42F8, 0x4C8A, 0xB07E}},
    {39765, 2, {0x1094, 0x2118, 0x7DDE, 0x5494, 0x555A, 0x7DDE, 0x5494, 0x55FE, 0x7C90, 0x10D4, 0x2AAA, 0x2D16, 0x4F20, 0x4802, 0x87FE, 0x0}},
    {39766, 2, {0x1020, 0x203E, 0x7C20, 0x55FE, 0x5522, 0x7D3C, 0x55E2, 0x551E, 0x7D28, 0x11AA, 0x2A6C, 0x2D28, 0x4F7C, 0x4802, 0x87FE, 0x0}},
    {39767, 2, {0x1020, 0xFE40, 0x10FC, 0x7C94, 0x94, 0xFEFC, 0x2A4, 0x7CA4, 0x10FC, 0xFE20, 0x44, 0xFF6A, 0x26E, 0xFFA0, 0xAAA2, 0xE71E}},
    {39768, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x27F8, 0x2488, 0x27F8, 0x2488, 0x47F8, 0x4252, 0x8C3E}},
    {39769, 2, {0x4204, 0x3FD8, 0x8A90, 0x4F9E, 0x2A94, 0x5FD4, 0x8224, 0x1FF0, 0x1110, 0x1FF0, 0x1210, 0x1FF0, 0x4A8, 0x8BA, 0x3082, 0xC07E}},
    {39770, 2, {0x400, 0x400, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x4888, 0x4444, 0x8444}},
    {39771, 2, {0x2000, 0x2000, 0x7CFE, 0x4422, 0x8822, 0x7E22, 0x5222, 0x5222, 0x7E22, 0x5222, 0x5222, 0x7E42, 0x42, 0xAA82, 0xAA94, 0x108}},
    {39772, 2, {0x2020, 0x2020, 0x7C20, 0x4420, 0x8820, 0x7E20, 0x5220, 0x5220, 0x7E50, 0x5250, 0x5250, 0x7E50, 0x48, 0xAA88, 0xAA84, 0x102}},
    {39773, 2, {0x1004, 0x1004, 0x3E04, 0x4224, 0x8424, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x4, 0x5504, 0x4A94, 0x8A88}},
    {39774, 2, {0x2080, 0x2040, 0x7C20, 0x4420, 0x8810, 0x7E10, 0x5210, 0x5228, 0x7E28, 0x5228, 0x5228, 0x7E48, 0x44, 0xAA44, 0xAA84, 0x102}},
    {39775, 2, {0x2000, 0x2000, 0x7C7C, 0x4410, 0x8810, 0x7E10, 0x5210, 0x5210, 0x7E10, 0x5210, 0x5210, 0x7E10, 0x10, 0xAA10, 0xAAFE, 0x0}},
    {39776, 2, {0x2008, 0x203C, 0x79E0, 0x4820, 0x9020, 0x7C20, 0x5420, 0x543E, 0x7DE0, 0x5420, 0x5420, 0x7C22, 0x22, 0xAA22, 0xAA1E, 0x0}},
    {39777, 2, {0x2040, 0x2040, 0x7C40, 0x447C, 0x8884, 0x7E84, 0x5304, 0x5244, 0x7E24, 0x5224, 0x5204, 0x7E04, 0x4, 0xAA04, 0xAA28, 0x10}},
    {39778, 2, {0x2000, 0x2000, 0x7CFC, 0x4404, 0x8804, 0x7E04, 0x5204, 0x52FC, 0x7E84, 0x5280, 0x5280, 0x7E80, 0x82, 0xAA82, 0xAA7E, 0x0}},
    {39779, 2, {0x2000, 0x21F8, 0x7C08, 0x4450, 0x8820, 0x7E10, 0x53FE, 0x5222, 0x7E24, 0x5220, 0x5220, 0x7E20, 0x20, 0xAA20, 0xAAA0, 0x40}},
    {39780, 2, {0x2050, 0x2050, 0x7850, 0x4892, 0x9092, 0x7D94, 0x5698, 0x5490, 0x7CB0, 0x54D0, 0x5490, 0x7C90, 0x92, 0xAA92, 0xAA8E, 0x80}},
    {39781, 2, {0x2000, 0x20FC, 0x7C44, 0x4448, 0x8848, 0x7E50, 0x525C, 0x5244, 0x7E44, 0x5244, 0x5268, 0x7EA8, 0x90, 0xAA98, 0xAB24, 0x42}},
    {39782, 2, {0x2010, 0x2010, 0x7C10, 0x4410, 0x8854, 0x7E52, 0x5252, 0x5290, 0x7E10, 0x5214, 0x5204, 0x7E08, 0x8, 0xAA10, 0xAA20, 0xC0}},
    {39783, 2, {0x2020, 0x2010, 0x7C10, 0x44FE, 0x8800, 0x7E00, 0x5278, 0x5248, 0x7E48, 0x5248, 0x5248, 0x7E4A, 0x4A, 0xAA8A, 0xAA86, 0x100}},
    {39784, 2, {0x2020, 0x2020, 0x7C20, 0x45FC, 0x8820, 0x7EA8, 0x52A8, 0x52A8, 0x7EA8, 0x52A8, 0x52F8, 0x7E22, 0x22, 0xAA22, 0xAA1E, 0x0}},
    {39785, 2, {0x2040, 0x2040, 0x7C40, 0x447E, 0x88AA, 0x7F2A, 0x522A, 0x524A, 0x7E4A, 0x5292, 0x5312, 0x7E22, 0x22, 0xAA42, 0xAA94, 0x8}},
    {39786, 2, {0x2010, 0x2010, 0x7C10, 0x4428, 0x8828, 0x7E44, 0x5282, 0x5248, 0x7E48, 0x5248, 0x5248, 0x7E48, 0x48, 0xAA88, 0xAA88, 0x108}},
    {39787, 2, {0x2010, 0x2010, 0x7C10, 0x4410, 0x88FE, 0x7E92, 0x5292, 0x5210, 0x7E28, 0x5228, 0x5228, 0x7E28, 0x48, 0xAA4A, 0xAA8A, 0x106}},
    {39788, 2, {0x2008, 0x201C, 0x79E0, 0x4900, 0x9100, 0x7DFC, 0x5544, 0x5544, 0x7D44, 0x5528, 0x5528, 0x7D10, 0x110, 0xAA28, 0xAA44, 0x482}},
    {39789, 2, {0x2000, 0x20FC, 0x7C00, 0x4400, 0x8800, 0x7EFE, 0x5250, 0x5250, 0x7E50, 0x5250, 0x5250, 0x7E52, 0x52, 0xAA92, 0xAA8E, 0x100}},
    {39790, 2, {0x2010, 0x2090, 0x7C90, 0x4492, 0x8892, 0x7E94, 0x52F4, 0x5298, 0x7E90, 0x5290, 0x5290, 0x7E92, 0xB2, 0xAAD2, 0xAA8E, 0x0}},
    {39791, 2, {0x800, 0x1FE0, 0x2020, 0x7FF8, 0xA108, 0x3FF8, 0x2108, 0x3FF8, 0x4444, 0x8222, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010}},
    {39792, 2, {0x2040, 0x2020, 0x7C20, 0x45FE, 0x8888, 0x7E88, 0x5288, 0x5288, 0x7E88, 0x5250, 0x5250, 0x7E20, 0x20, 0xAA50, 0xAA88, 0x106}},
    {39793, 2, {0x2000, 0x21FE, 0x7C20, 0x4420, 0x8820, 0x7E3C, 0x5244, 0x5244, 0x7E44, 0x5284, 0x52FC, 0x7E08, 0x8, 0xAA08, 0xABFE, 0x0}},
    {39794, 2, {0x2020, 0x2010, 0x7C10, 0x447E, 0x8842, 0x7E42, 0x5242, 0x527E, 0x7E42, 0x5240, 0x5240, 0x7E40, 0x40, 0xAA80, 0xAA80, 0x100}},
    {39795, 2, {0x2000, 0x23FE, 0x7820, 0x4820, 0x9020, 0x7DFC, 0x5524, 0x5524, 0x7D24, 0x5524, 0x5524, 0x7D34, 0x128, 0xAA20, 0xAA20, 0x20}},
    {39796, 2, {0x2020, 0x2010, 0x7C10, 0x45FE, 0x8840, 0x7E40, 0x5240, 0x527C, 0x7E44, 0x5244, 0x5244, 0x7E44, 0x44, 0xAA44, 0xAAA8, 0x110}},
    {39797, 2, {0x2008, 0x2028, 0x7C28, 0x4428, 0x8844, 0x7E44, 0x5282, 0x537C, 0x7E24, 0x5224, 0x5224, 0x7E24, 0x44, 0xAA44, 0xAA94, 0x108}},
    {39798, 2, {0x2010, 0x2010, 0x7C10, 0x44FE, 0x8892, 0x7E92, 0x5292, 0x52AA, 0x7EAA, 0x52CA, 0x5282, 0x7E82, 0x82, 0xAA82, 0xAA8A, 0x84}},
    {39799, 2, {0x2048, 0x2044, 0x7844, 0x4840, 0x91FE, 0x7C50, 0x5450, 0x5450, 0x7C50, 0x5450, 0x5450, 0x7C52, 0x52, 0xAA92, 0xAA8E, 0x100}},
    {39800, 2, {0x2010, 0x2090, 0x7890, 0x4890, 0x9090, 0x7CFE, 0x5480, 0x5480, 0x7C80, 0x54F8, 0x5488, 0x7C88, 0x88, 0xAA88, 0xAA88, 0x108}},
    {39801, 2, {0x2008, 0x201C, 0x7CE0, 0x4420, 0x8820, 0x7E3C, 0x52E0, 0x5220, 0x7E20, 0x523E, 0x53E0, 0x7E20, 0x22, 0xAA22, 0xAA22, 0x1E}},
    {39802, 2, {0x2000, 0x2000, 0x7DFE, 0x4404, 0x8804, 0x7EF4, 0x5294, 0x5294, 0x7E94, 0x5294, 0x52F4, 0x7E94, 0x4, 0xAA04, 0xAA14, 0x8}},
    {39803, 2, {0x2000, 0x21FC, 0x7924, 0x4924, 0x9124, 0x7DFC, 0x5524, 0x5524, 0x7D24, 0x55FC, 0x5524, 0x7C20, 0x20, 0xAA20, 0xAA20, 0x20}},
    {39804, 2, {0x2020, 0x2020, 0x7820, 0x49FC, 0x9020, 0x7C20, 0x5420, 0x55FE, 0x7C20, 0x5420, 0x5440, 0x7C48, 0x84, 0xABFE, 0xAA82, 0x0}},
    {39805, 2, {0x2044, 0x2044, 0x7C44, 0x4444, 0x88FE, 0x7E44, 0x5244, 0x5244, 0x7E44, 0x527C, 0x5244, 0x7E44, 0x44, 0xAA44, 0xAA7C, 0x44}},
    {39806, 2, {0x2000, 0x21FE, 0x7C10, 0x4410, 0x8820, 0x7E20, 0x5268, 0x52A4, 0x7F22, 0x5222, 0x5220, 0x7E20, 0x20, 0xAA00, 0xABFE, 0x0}},
    {39807, 2, {0x2020, 0x2020, 0x7850, 0x4850, 0x9088, 0x7D24, 0x5612, 0x5410, 0x7DFC, 0x5404, 0x5408, 0x7C88, 0x50, 0xAA20, 0xAA10, 0x10}},
    {39808, 2, {0x2020, 0x2010, 0x7C10, 0x44FE, 0x8882, 0x7E84, 0x5240, 0x5244, 0x7E48, 0x5250, 0x5260, 0x7E42, 0x42, 0xAA42, 0xAA3E, 0x0}},
    {39809, 2, {0x2050, 0x2048, 0x7848, 0x4840, 0x91FE, 0x7C40, 0x5440, 0x547C, 0x7CA4, 0x54A4, 0x54A8, 0x7D28, 0x110, 0xAA28, 0xAA44, 0x82}},
    {39810, 2, {0x2000, 0x20FE, 0x7C82, 0x4492, 0x8892, 0x7E92, 0x5292, 0x5292, 0x7E92, 0x52AA, 0x52A6, 0x7EC6, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39811, 2, {0x2000, 0x21FC, 0x7C20, 0x4420, 0x8924, 0x7EA4, 0x52A8, 0x5220, 0x7FFE, 0x5220, 0x5220, 0x7E20, 0x20, 0xAA20, 0xAA20, 0x20}},
    {39812, 2, {0x2050, 0x2050, 0x7850, 0x49FC, 0x9054, 0x7C54, 0x55FC, 0x5550, 0x7D50, 0x55FE, 0x5452, 0x7C52, 0x5A, 0xAA94, 0xAA90, 0x110}},
    {39813, 2, {0x2010, 0x2008, 0x7C08, 0x4402, 0x8822, 0x7E24, 0x5224, 0x52A8, 0x7EAC, 0x52B2, 0x5332, 0x7E20, 0x62, 0xAAA2, 0xAB1E, 0x0}},
    {39814, 2, {0x880, 0x2888, 0x2EF0, 0x2884, 0x2E84, 0xF07C, 0x800, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39815, 2, {0x2020, 0x2020, 0x7820, 0x49FC, 0x9020, 0x7C20, 0x5420, 0x57FE, 0x7C70, 0x54A8, 0x54A8, 0x7D24, 0x124, 0xAA22, 0xAA20, 0x20}},
    {39816, 2, {0x2040, 0x2040, 0x7C40, 0x44FC, 0x8884, 0x7F04, 0x52F4, 0x5294, 0x7E94, 0x5294, 0x5294, 0x7EF4, 0x94, 0xAA04, 0xAA28, 0x10}},
    {39817, 2, {0x2000, 0x21FC, 0x7C44, 0x4444, 0x8844, 0x7E44, 0x5294, 0x5288, 0x7F00, 0x52FC, 0x5284, 0x7E84, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39818, 2, {0x2010, 0x2010, 0x7C20, 0x44FE, 0x8882, 0x7E82, 0x5282, 0x5282, 0x7EFE, 0x5282, 0x5282, 0x7E82, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39819, 2, {0x2010, 0x2010, 0x7C10, 0x4410, 0x88FE, 0x7E92, 0x5292, 0x5292, 0x7E92, 0x52FE, 0x5292, 0x7E92, 0x92, 0xAA92, 0xAAFE, 0x82}},
    {39820, 2, {0x2010, 0x2008, 0x7C08, 0x44FE, 0x8810, 0x7E10, 0x5220, 0x5242, 0x7EFC, 0x5248, 0x5210, 0x7E20, 0x44, 0xAAFE, 0xAA42, 0x0}},
    {39821, 2, {0x2010, 0x2010, 0x7810, 0x49FE, 0x9112, 0x7D14, 0x5510, 0x55FC, 0x7D44, 0x5544, 0x5528, 0x7D28, 0x110, 0xAA28, 0xAA44, 0x482}},
    {39822, 2, {0x2010, 0x2010, 0x7C10, 0x4410, 0x881E, 0x7E10, 0x5210, 0x5210, 0x7EFE, 0x5282, 0x5282, 0x7E82, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39823, 2, {0x2020, 0x2020, 0x7920, 0x4920, 0x91FC, 0x7D20, 0x5620, 0x5420, 0x7C20, 0x55FC, 0x5420, 0x7C20, 0x20, 0xAA20, 0xABFE, 0x0}},
    {39824, 2, {0x2020, 0x2020, 0x7C20, 0x4440, 0x8848, 0x7E84, 0x53FE, 0x5282, 0x7E00, 0x52FC, 0x5284, 0x7E84, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39825, 2, {0x2040, 0x2040, 0x7CFC, 0x4484, 0x8904, 0x7EF4, 0x5294, 0x5294, 0x7EF4, 0x5284, 0x5294, 0x7E88, 0x82, 0xAA82, 0xAA7E, 0x0}},
    {39826, 2, {0x2044, 0x2044, 0x7844, 0x4884, 0x90BE, 0x7D84, 0x5684, 0x54A4, 0x7C94, 0x5494, 0x5484, 0x7C84, 0x84, 0xAA84, 0xAA94, 0x88}},
    {39827, 2, {0x2040, 0x2040, 0x7C40, 0x447E, 0x88A0, 0x7EA0, 0x5320, 0x523C, 0x7E20, 0x5220, 0x5220, 0x7E3E, 0x20, 0xAA20, 0xAA20, 0x20}},
    {39828, 2, {0x2000, 0x20FE, 0x7C80, 0x4480, 0x8880, 0x7EFC, 0x5284, 0x5284, 0x7E84, 0x5284, 0x52FC, 0x7E80, 0x80, 0xAA80, 0xAAFE, 0x0}},
    {39829, 2, {0x2010, 0x2010, 0x7C10, 0x4410, 0x88FE, 0x7E10, 0x5210, 0x5210, 0x7E7C, 0x5244, 0x5244, 0x7E44, 0x44, 0xAA44, 0xAA7C, 0x44}},
    {39830, 2, {0x2000, 0x2000, 0x7DFE, 0x4410, 0x8810, 0x7E20, 0x5220, 0x527C, 0x7E44, 0x52C4, 0x5344, 0x7E44, 0x44, 0xAA44, 0xAA7C, 0x44}},
    {39831, 2, {0x2040, 0x2040, 0x787C, 0x4884, 0x9188, 0x7E50, 0x5420, 0x5450, 0x7C88, 0x5706, 0x5460, 0x7C10, 0x8, 0xAAC0, 0xAA20, 0x10}},
    {39832, 2, {0x2050, 0x2054, 0x7852, 0x4892, 0x9090, 0x7D96, 0x55F8, 0x5690, 0x7C90, 0x5490, 0x5490, 0x7C90, 0x8A, 0xAA8A, 0xAA86, 0x82}},
    {39833, 2, {0x2000, 0x20EE, 0x7C22, 0x4422, 0x88AA, 0x7E66, 0x5222, 0x5222, 0x7E26, 0x526A, 0x52B2, 0x7E22, 0x22, 0xAA22, 0xAAAA, 0x44}},
    {39834, 2, {0x2010, 0x2010, 0x7C10, 0x44FE, 0x8810, 0x7E10, 0x527C, 0x5200, 0x7E00, 0x527C, 0x5244, 0x7E44, 0x44, 0xAA44, 0xAA7C, 0x44}},
    {39835, 2, {0x2020, 0x2020, 0x7820, 0x483C, 0x9020, 0x7C20, 0x55FE, 0x5400, 0x7C20, 0x5420, 0x54A8, 0x7CA4, 0x122, 0xAA22, 0xAAA0, 0x40}},
    {39836, 2, {0x2008, 0x201C, 0x79E0, 0x4900, 0x9100, 0x7DFE, 0x5500, 0x5500, 0x7D7C, 0x5544, 0x5544, 0x7D44, 0x144, 0xAA7C, 0xAA44, 0x400}},
    {39837, 2, {0x1110, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x3418, 0xCFE6, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39838, 2, {0x2000, 0x20FE, 0x7C10, 0x4410, 0x8820, 0x7EFE, 0x52AA, 0x52AA, 0x7EAA, 0x52AA, 0x52AA, 0x7EAA, 0xAA, 0xAAA2, 0xAA8A, 0x84}},
    {39839, 2, {0x2040, 0x2020, 0x7820, 0x4BFE, 0x9202, 0x7C44, 0x5440, 0x57FE, 0x7C88, 0x5488, 0x5508, 0x7CD0, 0x20, 0xAA50, 0xAA88, 0x304}},
    {39840, 2, {0x2040, 0x2040, 0x78FC, 0x4904, 0x9208, 0x7DFE, 0x5500, 0x557C, 0x7D44, 0x5544, 0x5554, 0x7D48, 0x142, 0xAA42, 0xAA3E, 0x400}},
    {39841, 2, {0x2050, 0x2050, 0x7850, 0x4952, 0x90D4, 0x7C58, 0x5450, 0x5458, 0x7CD4, 0x5552, 0x5450, 0x7C50, 0x92, 0xAA92, 0xAB12, 0x20E}},
    {39842, 2, {0x2020, 0x2120, 0x7920, 0x49FC, 0x9120, 0x7E20, 0x5420, 0x57FE, 0x7C70, 0x54A8, 0x54A8, 0x7D24, 0x124, 0xAA22, 0xAA20, 0x20}},
    {39843, 2, {0x2000, 0x2040, 0x799E, 0x4912, 0x9112, 0x7D12, 0x55D2, 0x5512, 0x7D12, 0x5512, 0x555A, 0x7D94, 0x110, 0xAA10, 0xAA10, 0x10}},
    {39844, 2, {0x7F84, 0x824, 0x1F24, 0x2124, 0x5624, 0x1804, 0xE414, 0xFE8, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39845, 2, {0x2040, 0x2040, 0x7C78, 0x4488, 0x8950, 0x7E20, 0x5250, 0x5288, 0x7F06, 0x52F8, 0x5288, 0x7E88, 0x88, 0xAA88, 0xAAF8, 0x88}},
    {39846, 2, {0x2000, 0x20FE, 0x7C82, 0x4482, 0x88BA, 0x7E82, 0x5282, 0x52BA, 0x7EAA, 0x52AA, 0x52AA, 0x7EBA, 0x82, 0xAA82, 0xAA8A, 0x84}},
    {39847, 2, {0x2020, 0x2020, 0x7BFE, 0x4820, 0x91FC, 0x7C24, 0x55FC, 0x5520, 0x7DFE, 0x5422, 0x542A, 0x7C54, 0x50, 0xAA88, 0xAB04, 0x202}},
    {39848, 2, {0x2080, 0x2084, 0x7898, 0x48E0, 0x9082, 0x7C82, 0x547E, 0x5400, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39849, 2, {0x2084, 0x2048, 0x7C00, 0x44FC, 0x8848, 0x7E48, 0x5248, 0x5248, 0x7FFE, 0x5248, 0x5248, 0x7E48, 0x48, 0xAA88, 0xAA88, 0x108}},
    {39850, 2, {0x2020, 0x2020, 0x79FE, 0x4840, 0x9040, 0x7CFC, 0x5484, 0x5584, 0x7EFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0xAA84, 0xAA94, 0x88}},
    {39851, 2, {0x2040, 0x2020, 0x7820, 0x49FC, 0x9000, 0x7C88, 0x5504, 0x5602, 0x7C88, 0x5488, 0x5450, 0x7C50, 0x20, 0xAA50, 0xAA88, 0x104}},
    {39852, 2, {0x2010, 0x2010, 0x7CFE, 0x4428, 0x8844, 0x7E82, 0x527C, 0x5200, 0x7EFE, 0x5220, 0x5240, 0x7E7C, 0x4, 0xAA04, 0xAA28, 0x10}},
    {39853, 2, {0x2010, 0x2010, 0x7C7C, 0x4410, 0x8810, 0x7E10, 0x52FE, 0x5200, 0x7E10, 0x5210, 0x527C, 0x7E10, 0x10, 0xAA10, 0xAAFE, 0x0}},
    {39854, 2, {0x2084, 0x2044, 0x7848, 0x4800, 0x91FE, 0x7C20, 0x5420, 0x54FC, 0x7C20, 0x5420, 0x55FE, 0x7C20, 0x20, 0xAA20, 0xAA20, 0x20}},
    {39855, 2, {0x2020, 0x2020, 0x7850, 0x4850, 0x9088, 0x7D04, 0x56FA, 0x5400, 0x7C00, 0x54F8, 0x5488, 0x7C88, 0x88, 0xAA88, 0xAAF8, 0x88}},
    {39856, 2, {0x2000, 0x2000, 0x7CFE, 0x4482, 0x8882, 0x7EBA, 0x52AA, 0x52AA, 0x7EAA, 0x52AA, 0x52BA, 0x7E82, 0x82, 0xAAFE, 0xAA82, 0x0}},
    {39857, 2, {0x2040, 0x2040, 0x7844, 0x49F4, 0x9048, 0x7C50, 0x57FE, 0x5440, 0x7C80, 0x5584, 0x5698, 0x7CE0, 0x82, 0xAA82, 0xAA7E, 0x0}},
    {39858, 2, {0x2050, 0x2054, 0x7852, 0x4892, 0x9090, 0x7DBE, 0x5690, 0x5490, 0x7C90, 0x54A8, 0x54A8, 0x7CA8, 0xA8, 0xAAA4, 0xAAA4, 0xC2}},
    {39859, 2, {0x2040, 0x2044, 0x79F4, 0x4848, 0x9050, 0x7DFE, 0x5440, 0x5480, 0x7DFE, 0x5640, 0x5480, 0x7CFC, 0x4, 0xAA04, 0xAA28, 0x10}},
    {39860, 2, {0x2048, 0x2048, 0x7848, 0x4888, 0x90FE, 0x7D88, 0x5688, 0x549C, 0x7C9C, 0x54AA, 0x54AA, 0x7CC8, 0x88, 0xAA88, 0xAA88, 0x88}},
    {39861, 2, {0x2084, 0x2044, 0x7C48, 0x4400, 0x88FC, 0x7E84, 0x5284, 0x5284, 0x7EFC, 0x5248, 0x5248, 0x7E48, 0x48, 0xAA8A, 0xAA8A, 0x106}},
    {39862, 2, {0x2000, 0x20FC, 0x7C24, 0x4424, 0x89FE, 0x7E24, 0x5224, 0x52FC, 0x7E40, 0x5240, 0x52FC, 0x7EC4, 0x144, 0xAA44, 0xAA7C, 0x44}},
    {39863, 2, {0x2088, 0x2048, 0x7850, 0x49FC, 0x9024, 0x7C24, 0x55FC, 0x5520, 0x7D20, 0x55FE, 0x5422, 0x7C22, 0x6A, 0xAAA4, 0xAB20, 0x20}},
    {39864, 2, {0x2080, 0x2080, 0x78F8, 0x4908, 0x9210, 0x7DFC, 0x5524, 0x5524, 0x7D24, 0x55FC, 0x5450, 0x7C50, 0x90, 0xAA92, 0xAB12, 0x20E}},
    {39865, 2, {0x2010, 0x2092, 0x7C52, 0x4454, 0x8810, 0x7EFE, 0x5282, 0x5282, 0x7EFE, 0x5282, 0x5282, 0x7EFE, 0x82, 0xAA82, 0xAA8A, 0x84}},
    {39866, 2, {0x440, 0x7FFC, 0x100, 0x3FF8, 0x400, 0xFFFE, 0x1400, 0x2FE0, 0xD040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39867, 2, {0x2010, 0x2010, 0x7C20, 0x4444, 0x88FE, 0x7E28, 0x5244, 0x52A2, 0x7E3C, 0x5244, 0x5244, 0x7EA8, 0x10, 0xAA28, 0xAA44, 0x82}},
    {39868, 2, {0x2000, 0x20FC, 0x7804, 0x487C, 0x9004, 0x7CFC, 0x5400, 0x55FE, 0x7D02, 0x54FC, 0x5444, 0x7C44, 0x28, 0xAA10, 0xAA68, 0x186}},
    {39869, 2, {0x2010, 0x2010, 0x7C28, 0x4428, 0x8844, 0x7EBA, 0x5210, 0x5210, 0x7EFE, 0x5210, 0x5258, 0x7E54, 0x92, 0xAB12, 0xAA50, 0x20}},
    {39870, 2, {0x2004, 0x201E, 0x79E0, 0x4822, 0x9112, 0x7C94, 0x5480, 0x5420, 0x7DFE, 0x5444, 0x5484, 0x7DC8, 0x30, 0xAA28, 0xAAC4, 0x302}},
    {39871, 2, {0x2000, 0x23FE, 0x7920, 0x4928, 0x9128, 0x7DE8, 0x5528, 0x5528, 0x7D28, 0x55E8, 0x5528, 0x7D2A, 0x13A, 0xABEA, 0xAA26, 0x20}},
    {39872, 2, {0x200C, 0x21F0, 0x7820, 0x4844, 0x9088, 0x7DF0, 0x5420, 0x5444, 0x7DFE, 0x5422, 0x5420, 0x7CA8, 0x124, 0xAA22, 0xAAA0, 0x40}},
    {39873, 2, {0x2000, 0x21FE, 0x7820, 0x4820, 0x91FC, 0x7D24, 0x5524, 0x55FC, 0x7D24, 0x5524, 0x55FC, 0x7D20, 0xA0, 0xAA40, 0xAAB0, 0x30E}},
    {39874, 2, {0x2000, 0x21FE, 0x7C28, 0x4428, 0x88FE, 0x7EAA, 0x52AA, 0x52AA, 0x7EC6, 0x5282, 0x5282, 0x7EFE, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39875, 2, {0x2000, 0x21FE, 0x7820, 0x4820, 0x90FC, 0x7C44, 0x5444, 0x55FE, 0x7C00, 0x5400, 0x54FC, 0x7C84, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39876, 2, {0x2028, 0x2024, 0x7824, 0x4820, 0x91FE, 0x7C20, 0x5520, 0x54B2, 0x7CB4, 0x5468, 0x54A8, 0x7D24, 0x222, 0xAA20, 0xAAA0, 0x40}},
    {39877, 2, {0x2004, 0x200E, 0x7BB8, 0x4888, 0x9088, 0x7D28, 0x552E, 0x57A8, 0x7CA8, 0x56A8, 0x56A8, 0x7D3E, 0x100, 0xAA80, 0xAA7E, 0x0}},
    {39878, 2, {0x2014, 0x2012, 0x7CFE, 0x4410, 0x8810, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x92, 0xAA92, 0xAA92, 0x86}},
    {39879, 2, {0x2020, 0x2010, 0x79FE, 0x4902, 0x9204, 0x7CF8, 0x5400, 0x5400, 0x7DFE, 0x5450, 0x5450, 0x7C50, 0x92, 0xAA92, 0xAB0E, 0x200}},
    {39880, 2, {0x1080, 0x11F8, 0x1288, 0x2070, 0x2B8E, 0x6880, 0x68F8, 0xA910, 0x2BFC, 0x2D24, 0x29FC, 0x2924, 0x29FC, 0x2000, 0x2154, 0x222A}},
    {39881, 2, {0x2000, 0x21FC, 0x7924, 0x4924, 0x91FC, 0x7D24, 0x5524, 0x55FC, 0x7C20, 0x5420, 0x55FC, 0x7C20, 0x20, 0xAA20, 0xABFE, 0x0}},
    {39882, 2, {0x2040, 0x1248, 0x4444, 0x2014, 0x860, 0x7380, 0x2800, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39883, 2, {0x2008, 0x2108, 0x7888, 0x4808, 0x912C, 0x7CAA, 0x54AA, 0x5448, 0x7C48, 0x548A, 0x558A, 0x7C84, 0x84, 0xAA88, 0xAA90, 0x60}},
    {39884, 2, {0x2010, 0x2050, 0x7C50, 0x447C, 0x8850, 0x7E90, 0x5210, 0x52FE, 0x7E00, 0x5200, 0x527C, 0x7E44, 0x44, 0xAA44, 0xAA7C, 0x44}},
    {39885, 2, {0x2020, 0x2010, 0x79FE, 0x4820, 0x9048, 0x7C84, 0x55FE, 0x5402, 0x7CA8, 0x54A8, 0x54A8, 0x7CA8, 0xA8, 0xAAAA, 0xAB2A, 0x206}},
    {39886, 2, {0x2010, 0x2014, 0x7812, 0x4810, 0x91FC, 0x7D10, 0x5510, 0x55D4, 0x7D54, 0x5554, 0x5554, 0x7D48, 0x14A, 0xAA9A, 0xAA26, 0x442}},
    {39887, 2, {0x2002, 0x2032, 0x78C2, 0x484A, 0x904A, 0x7C4A, 0x55FA, 0x544A, 0x7CCA, 0x54EA, 0x555A, 0x7D42, 0x242, 0xAA42, 0xAA4A, 0x44}},
    {39888, 2, {0x2010, 0x2010, 0x7C7C, 0x4410, 0x8810, 0x7E10, 0x52FE, 0x5210, 0x7E10, 0x5250, 0x525C, 0x7E50, 0x70, 0xAA90, 0xAA8E, 0x100}},
    {39889, 2, {0x20C2, 0x2034, 0x7818, 0x4864, 0x9182, 0x7C20, 0x55FE, 0x5450, 0x7C90, 0x55FE, 0x5692, 0x7C92, 0x9A, 0xAA94, 0xAA10, 0x10}},
    {39890, 2, {0x2000, 0x20FC, 0x7C04, 0x4468, 0x8810, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x92, 0xAA92, 0xAA92, 0x86}},
    {39891, 2, {0x2020, 0x2040, 0x78F8, 0x4888, 0x90F8, 0x7C88, 0x54F8, 0x548A, 0x7C8C, 0x55F8, 0x5418, 0x7C28, 0x48, 0xAA88, 0xAB28, 0x10}},
    {39892, 2, {0x2000, 0x2092, 0x7C92, 0x4524, 0x8924, 0x7E92, 0x5292, 0x5200, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x92, 0xAA92, 0xAAFE, 0x82}},
    {39893, 2, {0x2088, 0x2088, 0x79FC, 0x4888, 0x9088, 0x7CF8, 0x5488, 0x5488, 0x7CF8, 0x5488, 0x5488, 0x7DFE, 0x0, 0xAA50, 0xAA88, 0x104}},
    {39894, 2, {0x2020, 0x2020, 0x79FE, 0x4820, 0x90FC, 0x7C20, 0x55FE, 0x5400, 0x7CFC, 0x5484, 0x54FC, 0x7C84, 0xFC, 0xAA84, 0xAA94, 0x88}},
    {39895, 2, {0x440, 0x7FFC, 0x100, 0x3FF8, 0x400, 0xFFFE, 0x1410, 0x2FE8, 0xD046, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39896, 2, {0x201C, 0x21E0, 0x7820, 0x4820, 0x93FE, 0x7CA8, 0x5524, 0x5642, 0x7C40, 0x57FE, 0x5488, 0x7D08, 0xD0, 0xAA30, 0xAA48, 0x184}},
    {39897, 2, {0x2020, 0x2010, 0x79FE, 0x4800, 0x90FC, 0x7C84, 0x54FC, 0x5400, 0x7CFC, 0x5408, 0x5410, 0x7DFE, 0x10, 0xAA10, 0xAA50, 0x20}},
    {39898, 2, {0x201C, 0x20E0, 0x7820, 0x49FE, 0x9070, 0x7CA8, 0x5524, 0x5622, 0x7CF8, 0x5410, 0x5420, 0x7DFE, 0x20, 0xAA20, 0xAAA0, 0x40}},
    {39899, 2, {0x2000, 0x20FE, 0x7C92, 0x4492, 0x88BA, 0x7E92, 0x52FE, 0x5282, 0x7EBA, 0x52AA, 0x52AA, 0x7EBA, 0x82, 0xAA82, 0xAA8A, 0x104}},
    {39900, 2, {0x2020, 0x2010, 0x7CFE, 0x4400, 0x8844, 0x7E28, 0x53FE, 0x5220, 0x7E20, 0x53FE, 0x5244, 0x7EC4, 0x28, 0xAA18, 0xAA64, 0x182}},
    {39901, 2, {0x2000, 0x20FE, 0x7C82, 0x4492, 0x8892, 0x7EFE, 0x5292, 0x5292, 0x7EBA, 0x52AA, 0x52AA, 0x7EBA, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39902, 2, {0x2000, 0x20FC, 0x7804, 0x487C, 0x9004, 0x7CFC, 0x5400, 0x55FE, 0x7D22, 0x5420, 0x54FC, 0x7CA4, 0xA4, 0xAAB4, 0xAAA8, 0x20}},
    {39903, 2, {0x2010, 0x2010, 0x7CFE, 0x4410, 0x8810, 0x7EFE, 0x5292, 0x52FE, 0x7E92, 0x52FE, 0x5210, 0x7E38, 0x54, 0xAA92, 0xAA10, 0x10}},
    {39904, 2, {0x2010, 0x2010, 0x7C10, 0x44FE, 0x8810, 0x7E54, 0x5254, 0x5254, 0x7EBA, 0x5210, 0x5238, 0x7E54, 0x92, 0xAA10, 0xAA10, 0x10}},
    {39905, 2, {0x2050, 0x2050, 0x7850, 0x4BDE, 0x9050, 0x7C50, 0x5450, 0x55DC, 0x7C50, 0x5450, 0x5450, 0x7C50, 0x3DE, 0xAA50, 0xAA50, 0x50}},
    {39906, 2, {0x2000, 0x2020, 0x7CCE, 0x4482, 0x8882, 0x7EEE, 0x5282, 0x5282, 0x7EFE, 0x5228, 0x5228, 0x7E28, 0x48, 0xAA4A, 0xAA8A, 0x106}},
    {39907, 2, {0x2000, 0x20FC, 0x7C84, 0x4484, 0x88FC, 0x7E84, 0x5284, 0x52FC, 0x7E40, 0x52FE, 0x532A, 0x7E4A, 0x92, 0xAB22, 0xAA4A, 0x84}},
    {39908, 2, {0x2000, 0x20FC, 0x7884, 0x4884, 0x90FC, 0x7C84, 0x5484, 0x54FC, 0x7C00, 0x5490, 0x5492, 0x7CF4, 0x98, 0xAA92, 0xAAD2, 0x8E}},
    {39909, 2, {0x2020, 0x2020, 0x78FC, 0x4820, 0x9020, 0x7DFE, 0x5448, 0x5484, 0x7D22, 0x5420, 0x54FC, 0x7C20, 0x20, 0xAA20, 0xABFE, 0x0}},
    {39910, 2, {0x2088, 0x2088, 0x78E8, 0x492C, 0x912A, 0x7EA8, 0x5448, 0x54A0, 0x7D1E, 0x5600, 0x54FC, 0x7C84, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39911, 2, {0x207C, 0x2044, 0x7C44, 0x447C, 0x8844, 0x7E44, 0x527C, 0x5200, 0x7EFE, 0x5282, 0x5282, 0x7EFE, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39912, 2, {0x2040, 0x2020, 0x7DFE, 0x4400, 0x8800, 0x7EF8, 0x5288, 0x5288, 0x7E88, 0x52F8, 0x5220, 0x7EA8, 0x124, 0xAA22, 0xAAA0, 0x40}},
    {39913, 2, {0x2010, 0x2010, 0x7C28, 0x4444, 0x8882, 0x7F7C, 0x5200, 0x52FE, 0x7EAA, 0x52AA, 0x52FE, 0x7EAA, 0xAA, 0xAAAA, 0xAA82, 0x86}},
    {39914, 2, {0x2020, 0x2020, 0x78FC, 0x4820, 0x9020, 0x7DFE, 0x5448, 0x5484, 0x7D22, 0x5478, 0x5488, 0x7D48, 0x50, 0xAA20, 0xAA50, 0x18C}},
    {39915, 2, {0x2000, 0x23E0, 0x795C, 0x4954, 0x9154, 0x7DD4, 0x5554, 0x5554, 0x7DD4, 0x5554, 0x5548, 0x7D68, 0x3C8, 0xAA54, 0xAA54, 0x62}},
    {39916, 2, {0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4808, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39917, 2, {0x2000, 0x20FC, 0x7C08, 0x4410, 0x88FE, 0x7E10, 0x5210, 0x5250, 0x7E20, 0x52FE, 0x52AA, 0x7EAA, 0xAA, 0xAAAA, 0xABFE, 0x0}},
    {39918, 2, {0x2020, 0x2010, 0x7CFE, 0x4482, 0x8800, 0x7E7C, 0x5200, 0x5200, 0x7EFE, 0x5210, 0x5210, 0x7E54, 0x52, 0xAA92, 0xAA50, 0x20}},
    {39919, 2, {0x2082, 0x2082, 0x7282, 0x93EA, 0x248A, 0xF88A, 0xAFFA, 0xA88A, 0xFBEA, 0xAAAA, 0xAAAA, 0xFAAA, 0x2A2, 0xAAE2, 0xA88A, 0x84}},
    {39920, 2, {0x2010, 0x2010, 0x7C28, 0x4444, 0x88A2, 0x7F10, 0x52FC, 0x5204, 0x7E08, 0x5220, 0x5210, 0x7E04, 0xA2, 0xAAA2, 0xAAAA, 0x118}},
    {39921, 2, {0x2010, 0x2010, 0x7C1E, 0x4410, 0x88FE, 0x7E92, 0x5298, 0x52F0, 0x7E92, 0x528E, 0x5280, 0x7EB8, 0xA8, 0xAAAA, 0xAB4A, 0x286}},
    {39922, 2, {0x2088, 0x2048, 0x7808, 0x4BF4, 0x9114, 0x7D22, 0x5500, 0x55E8, 0x7D24, 0x5524, 0x5520, 0x7D30, 0x128, 0xAA24, 0xAAA4, 0x440}},
    {39923, 2, {0x2040, 0x2020, 0x73FE, 0x9200, 0x221C, 0xFAF0, 0xAA90, 0xAA90, 0xFAFE, 0xAA90, 0xAA90, 0xFA90, 0x2AA, 0xAACA, 0xACA6, 0x812}},
    {39924, 2, {0x2000, 0x21FC, 0x7804, 0x4834, 0x91C4, 0x7C44, 0x5444, 0x55F4, 0x7D54, 0x5554, 0x55F4, 0x7C44, 0x54, 0xAA76, 0xAB96, 0x2}},
    {39925, 2, {0x2040, 0x2048, 0x7884, 0x49FE, 0x9020, 0x7DFE, 0x5488, 0x5524, 0x7E42, 0x5588, 0x5410, 0x7C22, 0xC4, 0xAA18, 0xAA60, 0x180}},
    {39926, 2, {0x2000, 0x20FE, 0x7C92, 0x4410, 0x88FE, 0x7E10, 0x527C, 0x5254, 0x7E7C, 0x5254, 0x527C, 0x7E10, 0xFE, 0xAA10, 0xAA10, 0x10}},
    {39927, 2, {0x2000, 0x207C, 0x7C44, 0x447C, 0x8844, 0x7E7C, 0x5200, 0x53FE, 0x7E10, 0x5210, 0x525E, 0x7E50, 0x50, 0xAAB0, 0xAA9E, 0x100}},
    {39928, 2, {0x2040, 0x205C, 0x7C44, 0x4484, 0x88BE, 0x7E90, 0x5390, 0x529E, 0x7EA8, 0x5288, 0x52BE, 0x7E88, 0x88, 0xAA94, 0xAA94, 0xA2}},
    {39929, 2, {0x2000, 0x21FC, 0x7904, 0x49FC, 0x9104, 0x7DFC, 0x5420, 0x5520, 0x7DFE, 0x5620, 0x5420, 0x7DFC, 0x20, 0xAA20, 0xABFE, 0x0}},
    {39930, 2, {0x2020, 0x2024, 0x79F4, 0x4828, 0x9028, 0x7DFE, 0x5420, 0x5440, 0x7CFC, 0x5584, 0x5684, 0x7CFC, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39931, 2, {0x2082, 0x2082, 0x73E2, 0x9082, 0x208A, 0xFBEA, 0xAAAA, 0xAAAA, 0xFBEA, 0xA88A, 0xA9CA, 0xFAAA, 0x4A2, 0xA882, 0xA88A, 0x84}},
    {39932, 2, {0x2028, 0x2092, 0x7CAA, 0x4482, 0x88FE, 0x7E28, 0x5244, 0x52A2, 0x7E3C, 0x5244, 0x5244, 0x7EA8, 0x10, 0xAA28, 0xAA44, 0x82}},
    {39933, 2, {0x2000, 0x20EE, 0x7CAA, 0x44AA, 0x88AA, 0x7EEA, 0x52AA, 0x52AA, 0x7EEA, 0x528A, 0x52CA, 0x7EAE, 0xD8, 0xAA88, 0xAA08, 0x8}},
    {39934, 2, {0x2040, 0x207E, 0x7848, 0x48BE, 0x90AA, 0x7DAA, 0x56BE, 0x54AA, 0x7CAA, 0x54BE, 0x54C8, 0x7CA8, 0x90, 0xAA98, 0xAAA4, 0xC2}},
    {39935, 2, {0x2020, 0x2010, 0x79FE, 0x4902, 0x9102, 0x7DFE, 0x5500, 0x5500, 0x7DFE, 0x55AA, 0x56AA, 0x7CFE, 0xAA, 0xAAAA, 0xAAAA, 0x86}},
    {39936, 2, {0x201E, 0x23E0, 0x7844, 0x4924, 0x9088, 0x7DFC, 0x5440, 0x5440, 0x7DFE, 0x5480, 0x54FC, 0x7D44, 0x128, 0xAA10, 0xAA68, 0x186}},
    {39937, 2, {0x2010, 0x2020, 0x7CFE, 0x4482, 0x88FE, 0x7E82, 0x52FE, 0x5210, 0x7E12, 0x52D4, 0x5258, 0x7E54, 0x94, 0xAA92, 0xAB50, 0x20}},
    {39938, 2, {0x2002, 0x21E2, 0x7922, 0x492A, 0x91EA, 0x7D2A, 0x552A, 0x55EA, 0x7D2A, 0x552A, 0x55EA, 0x7C02, 0xC2, 0xAAA2, 0xAB2A, 0x204}},
    {39939, 2, {0x2000, 0x20FC, 0x78A4, 0x48A4, 0x90FC, 0x7CA4, 0x54A4, 0x54FC, 0x7C00, 0x55FE, 0x5490, 0x7C94, 0x88, 0xAAA8, 0xAAC4, 0x82}},
    {39940, 2, {0x200C, 0x200A, 0x7008, 0x93FE, 0x2208, 0xFBF8, 0xAA4A, 0xAA4A, 0xFBEA, 0xAAAC, 0xAAAC, 0xFA4C, 0x2AA, 0xAB0A, 0xAA16, 0x422}},
    {39941, 2, {0x2000, 0x207C, 0x7C54, 0x4454, 0x887C, 0x7E54, 0x5254, 0x527C, 0x7E10, 0x52FE, 0x5292, 0x7E96, 0xBE, 0xAA82, 0xAA8A, 0x84}},
    {39942, 2, {0x2020, 0x2020, 0x79FE, 0x4820, 0x90FC, 0x7C40, 0x55FE, 0x5488, 0x7D04, 0x56FA, 0x5488, 0x7C88, 0xF8, 0xAA88, 0xAA88, 0xF8}},
    {39943, 2, {0x2000, 0x20FC, 0x7848, 0x4830, 0x91FE, 0x7C52, 0x5494, 0x5510, 0x7E30, 0x5420, 0x55FE, 0x7C70, 0xA8, 0xAB24, 0xAA22, 0x20}},
    {39944, 2, {0x2014, 0x2054, 0x7C54, 0x44FE, 0x8854, 0x7E54, 0x525C, 0x5240, 0x7E7E, 0x5210, 0x52FE, 0x7E10, 0x38, 0xAA54, 0xAA92, 0x10}},
    {39945, 2, {0x2010, 0x2020, 0x7C7C, 0x4444, 0x887C, 0x7E44, 0x527C, 0x5200, 0x7EFE, 0x5210, 0x5210, 0x7E7C, 0x10, 0xAA10, 0xAAFE, 0x0}},
    {39946, 2, {0x2020, 0x2020, 0x7BFE, 0x4820, 0x91FC, 0x7D24, 0x55AC, 0x5574, 0x7D24, 0x55FC, 0x5420, 0x7C70, 0xA8, 0xAB24, 0xAA22, 0x20}},
    {39947, 2, {0x2000, 0x21FE, 0x7900, 0x497E, 0x9142, 0x7D7E, 0x5542, 0x557E, 0x7D10, 0x55FE, 0x5524, 0x7D64, 0x118, 0xAB24, 0xAB42, 0x1FE}},
    {39948, 2, {0x2044, 0x2024, 0x7C28, 0x44FE, 0x8828, 0x7E28, 0x52FE, 0x52AA, 0x7EAA, 0x52C6, 0x5282, 0x7EFE, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {39949, 2, {0x2008, 0x2068, 0x7988, 0x4888, 0x90AA, 0x7CAC, 0x55E8, 0x54A8, 0x7C88, 0x55C8, 0x56A8, 0x7C88, 0x94, 0xAA94, 0xAA94, 0xA2}},
    {39950, 2, {0x2010, 0x2010, 0x783C, 0x4B94, 0x90FE, 0x7C94, 0x553C, 0x5510, 0x7FBC, 0x5490, 0x54FE, 0x7E90, 0x190, 0xAAC0, 0xAABE, 0x100}},
    {39951, 2, {0x2000, 0x20FE, 0x7C00, 0x447C, 0x8844, 0x7E44, 0x527C, 0x5200, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x92, 0xAA92, 0xAAFE, 0x82}},
    {39952, 2, {0x2000, 0x21DC, 0x7954, 0x4954, 0x91DC, 0x7C00, 0x55FC, 0x5400, 0x7DFE, 0x5480, 0x54FC, 0x7C04, 0x4, 0xAA04, 0xAA28, 0x10}},
    {39953, 2, {0x2000, 0x20FC, 0x7884, 0x48FC, 0x9084, 0x7CFC, 0x5400, 0x55FE, 0x7C80, 0x54FE, 0x552A, 0x7E4A, 0x92, 0xAB22, 0xAA54, 0x88}},
    {39954, 2, {0x2040, 0x2040, 0x7CFE, 0x4480, 0x89FC, 0x7E84, 0x52FC, 0x5284, 0x7EFC, 0x5240, 0x527C, 0x7EC4, 0x128, 0xAA10, 0xAA68, 0x186}},
    {39955, 2, {0x2000, 0x21FC, 0x7924, 0x4924, 0x91FC, 0x7D24, 0x5524, 0x55FC, 0x7C20, 0x5410, 0x5454, 0x7D42, 0x142, 0xAB4A, 0xAA38, 0x0}},
    {39956, 2, {0x2008, 0x200C, 0x700A, 0x93FE, 0x2208, 0xFA08, 0xAAE8, 0xAA0A, 0xFA0A, 0xAAEC, 0xAAAC, 0xFAA8, 0x2EA, 0xAA1A, 0xAA26, 0x442}},
    {39957, 2, {0x2000, 0x21DC, 0x7944, 0x4944, 0x9144, 0x7DDC, 0x5500, 0x5500, 0x7DDC, 0x5514, 0x5514, 0x7DD4, 0x108, 0xAB08, 0xAB14, 0x122}},
    {39958, 2, {0x2020, 0x2020, 0x79FE, 0x4840, 0x90FE, 0x7D10, 0x56FE, 0x5400, 0x7CFE, 0x5482, 0x54FE, 0x7C82, 0xFE, 0xAA82, 0xAA8A, 0x84}},
    {39959, 2, {0x2080, 0x209E, 0x7892, 0x4BF2, 0x9092, 0x7C9E, 0x5492, 0x55D2, 0x7D52, 0x555E, 0x5552, 0x7D52, 0x1D2, 0xAA22, 0xAA2A, 0x44}},
    {39960, 2, {0x2040, 0x2020, 0x79FE, 0x4902, 0x9204, 0x7DFC, 0x5440, 0x5488, 0x7DFC, 0x5424, 0x5420, 0x7DFC, 0x20, 0xAA20, 0xABFE, 0x0}},
    {39961, 2, {0x2084, 0x2084, 0x79FE, 0x4884, 0x90A4, 0x7C20, 0x55FE, 0x5440, 0x7C80, 0x54FC, 0x5584, 0x7E84, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39962, 2, {0x2020, 0x2010, 0x79FE, 0x4902, 0x9204, 0x7CFC, 0x5400, 0x54FC, 0x7C84, 0x54FC, 0x5484, 0x7CFC, 0x84, 0xAA00, 0xAAFE, 0x0}},
    {39963, 2, {0x2000, 0x20F8, 0x7888, 0x4888, 0x90F8, 0x7C88, 0x5488, 0x54F8, 0x7C00, 0x55FC, 0x5554, 0x7D54, 0x154, 0xAB54, 0xABFE, 0x0}},
    {39964, 2, {0x2088, 0x2050, 0x7800, 0x49FE, 0x9050, 0x7DFC, 0x5454, 0x57FE, 0x7C54, 0x55FC, 0x5450, 0x7CD8, 0x154, 0xAA52, 0xAA50, 0x50}},
    {39965, 2, {0x2020, 0x2010, 0x7CFE, 0x4400, 0x887C, 0x7E44, 0x527C, 0x5200, 0x7EFE, 0x5282, 0x52BA, 0x7EAA, 0xBA, 0xAA82, 0xAA8A, 0x84}},
    {39966, 2, {0x2020, 0x2040, 0x7CFC, 0x4484, 0x8884, 0x7EFC, 0x5280, 0x52FE, 0x7E80, 0x52FE, 0x5202, 0x7EAA, 0xAA, 0xAA02, 0xAA14, 0x8}},
    {39967, 2, {0x2020, 0x2010, 0x7CFE, 0x4400, 0x8844, 0x7E28, 0x52FE, 0x52A2, 0x7E10, 0x52FE, 0x5220, 0x7E3C, 0x24, 0xAA44, 0xAA54, 0x88}},
    {39968, 2, {0x2000, 0x21FC, 0x78A4, 0x4888, 0x9050, 0x7C20, 0x54D8, 0x5726, 0x7C20, 0x55FC, 0x5524, 0x7D24, 0x1FC, 0xAA22, 0xABFE, 0x2}},
    {39969, 2, {0x2020, 0x20DE, 0x7C8A, 0x44CA, 0x88AA, 0x7ED2, 0x52A6, 0x5200, 0x7EFE, 0x5292, 0x5292, 0x7EFE, 0x92, 0xAA92, 0xAAFE, 0x82}},
    {39970, 2, {0x2000, 0x20FE, 0x7890, 0x48FC, 0x9090, 0x7CFC, 0x5490, 0x5490, 0x7CFE, 0x5402, 0x54AA, 0x7CAA, 0xAA, 0xAA82, 0xAA14, 0x8}},
    {39971, 2, {0x2008, 0x2008, 0x79C8, 0x497E, 0x9148, 0x7D48, 0x557E, 0x55C4, 0x7D44, 0x557E, 0x5544, 0x7D64, 0x1D4, 0xAA04, 0xAA14, 0x8}},
    {39972, 2, {0x2100, 0x227E, 0x7388, 0x9288, 0x22BE, 0xFAAA, 0xABAA, 0xAA2A, 0xFA2A, 0xABAA, 0xAAAA, 0xFAAE, 0x288, 0xAB88, 0xAA88, 0x8}},
    {39973, 2, {0x2000, 0x20FE, 0x7CAA, 0x44AA, 0x88AA, 0x7EFE, 0x5210, 0x5254, 0x7E52, 0x5292, 0x5210, 0x7E54, 0x52, 0xAA92, 0xAA10, 0x10}},
    {39974, 2, {0x2084, 0x2044, 0x7848, 0x4BFE, 0x9084, 0x7C84, 0x5508, 0x554A, 0x7E52, 0x579C, 0x5484, 0x7D08, 0x108, 0xAA52, 0xABDE, 0x42}},
    {39975, 2, {0x124, 0x78A8, 0x49FC, 0x4840, 0x4BFE, 0x7908, 0x4A84, 0x4CFA, 0x4910, 0x7BFC, 0x4D24, 0x49FC, 0x4924, 0x49FC, 0x4800, 0x9954}},
    {39976, 2, {0x2000, 0x21F8, 0x7908, 0x49F8, 0x9108, 0x7DF8, 0x5400, 0x5400, 0x7DDC, 0x5444, 0x5554, 0x7CCC, 0x154, 0xAA44, 0xAA54, 0xC8}},
    {39977, 2, {0x2004, 0x201E, 0x79E0, 0x4802, 0x9122, 0x7C94, 0x54FE, 0x5510, 0x7C10, 0x55FE, 0x5410, 0x7C92, 0x92, 0xAA92, 0xAAFE, 0x2}},
    {39978, 2, {0x2010, 0x2010, 0x7C7E, 0x4410, 0x8810, 0x7EFE, 0x5220, 0x5242, 0x7EFE, 0x5200, 0x52FE, 0x7EAA, 0xAA, 0xAAAA, 0xABFE, 0x0}},
    {39979, 2, {0x2040, 0x2020, 0x79FE, 0x4902, 0x9050, 0x7C88, 0x5524, 0x5450, 0x7C88, 0x5504, 0x56FA, 0x7C88, 0x88, 0xAA88, 0xAAF8, 0x88}},
    {39980, 2, {0x2020, 0x203C, 0x7820, 0x49FE, 0x9122, 0x7D38, 0x55E4, 0x551C, 0x7D20, 0x5510, 0x55FE, 0x7D44, 0x128, 0xAA10, 0xAA28, 0x4C6}},
    {39981, 2, {0x2020, 0x2024, 0x78F4, 0x4828, 0x91FE, 0x7C20, 0x5478, 0x54C2, 0x7D3E, 0x5600, 0x54FC, 0x7C84, 0xFC, 0xAA84, 0xAAFC, 0x84}},
    {39982, 2, {0x2000, 0x20F8, 0x78A8, 0x48A8, 0x90D8, 0x7C88, 0x54F8, 0x5400, 0x7C00, 0x55FC, 0x5554, 0x7D54, 0x154, 0xAB54, 0xABFE, 0x0}},
    {39983, 2, {0x2000, 0x21DC, 0x7844, 0x4844, 0x91DC, 0x7D10, 0x5510, 0x55DC, 0x7C44, 0x5554, 0x54CC, 0x7CCC, 0x154, 0xAA44, 0xAA54, 0xC8}},
    {39984, 2, {0x2208, 0x2108, 0x7908, 0x483E, 0x93AA, 0x7CAA, 0x54BE, 0x552A, 0x7DAA, 0x576A, 0x553E, 0x7D2A, 0x108, 0xAB08, 0xAB08, 0x108}},
    {39985, 2, {0x2010, 0x2010, 0x7AFE, 0x4910, 0x917C, 0x7C54, 0x547C, 0x5754, 0x7D7C, 0x5510, 0x55FE, 0x7D10, 0x110, 0xAA90, 0xAA7E, 0x0}},
    {39986, 2, {0x820, 0x7E20, 0x87E, 0xFEC4, 0x1028, 0x1E10, 0x2A28, 0x4FE6, 0x9040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39987, 2, {0x2148, 0x2148, 0x73E8, 0x9148, 0x21DE, 0xF88A, 0xABEA, 0xAAAA, 0xFAAA, 0xABEA, 0xA88A, 0xFBEA, 0x92, 0xA892, 0xA8AA, 0xC4}},
    {39988, 2, {0x1110, 0x1550, 0x2550, 0x47D0, 0x903E, 0x17A4, 0x2254, 0x6394, 0xA494, 0x27D4, 0x2D54, 0x27C8, 0x2548, 0x27D4, 0x2014, 0x2AA2}},
    {39989, 2, {0x2020, 0x3F20, 0x403E, 0xBE48, 0x2AA8, 0xFF28, 0x4A10, 0x7FE8, 0x1046, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {39990, 2, {0x2080, 0x2088, 0x79EE, 0x492A, 0x92AA, 0x7D44, 0x5484, 0x557A, 0x7E00, 0x5400, 0x55FE, 0x7C20, 0xA8, 0xAB24, 0xAAA2, 0x40}},
    {39991, 2, {0x2110, 0x211E, 0x7122, 0x9254, 0x2208, 0xFE94, 0xAAE2, 0xAA88, 0xFA88, 0xAABE, 0xAA88, 0xFAAC, 0x22A, 0xAA4A, 0xAA28, 0x210}},
    {39992, 2, {0x2000, 0x23FE, 0x7200, 0x927C, 0x2244, 0xFA44, 0xAA7C, 0xAA00, 0xFAEE, 0xAAAA, 0xAAAA, 0xFAAA, 0x2EE, 0xAA00, 0xABFE, 0x0}},
    {39993, 2, {0x2000, 0x21EE, 0x794A, 0x49EA, 0x912A, 0x7DE4, 0x5544, 0x554A, 0x7DF0, 0x5400, 0x5410, 0x7CFE, 0x10, 0xAA10, 0xABFE, 0x0}},
    {39994, 2, {0x2010, 0x2020, 0x7C44, 0x44FE, 0x8844, 0x7EAA, 0x52EE, 0x5210, 0x7E28, 0x5254, 0x52AA, 0x7E50, 0x24, 0xAA48, 0xAA10, 0x60}},
    {39995, 2, {0x207C, 0x2044, 0x7C7C, 0x4444, 0x887C, 0x7E00, 0x52FE, 0x52AA, 0x7EFE, 0x5200, 0x52FC, 0x7E44, 0x28, 0xAA10, 0xAA68, 0x186}},
    {39996, 2, {0x2000, 0x21DE, 0x7842, 0x4952, 0x90CA, 0x7D52, 0x5442, 0x5420, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0xAA84, 0xAAFC, 0x84}},
    {39997, 2, {0x2050, 0x2050, 0x79FE, 0x4850, 0x91FC, 0x7D54, 0x55FC, 0x5554, 0x7DFC, 0x5400, 0x54F8, 0x7C88, 0xF8, 0xAA88, 0xAAF8, 0x88}},
    {39998, 2, {0x2000, 0x23FE, 0x7850, 0x49FC, 0x9154, 0x7D54, 0x55FC, 0x5400, 0x7DFC, 0x5400, 0x57FE, 0x7C20, 0xA8, 0xAB24, 0xAAA2, 0x40}},
    {39999, 2, {0x2010, 0x21FE, 0x7810, 0x48FC, 0x9010, 0x7DFE, 0x5400, 0x54FC, 0x7C84, 0x54FC, 0x5484, 0x7CFC, 0x84, 0xAAFC, 0xAA48, 0x84}},
    {40000, 2, {0x7DFC, 0x4420, 0x7D20, 0x45FE, 0x7C50, 0x4852, 0x5492, 0x6FEE, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {40001, 2, {0x2000, 0x213E, 0x7888, 0x4890, 0x902A, 0x7C4A, 0x571C, 0x552C, 0x7D4A, 0x551A, 0x552A, 0x7D48, 0x128, 0xAA90, 0xAA7E, 0x0}},
    {40002, 2, {0x2048, 0x2148, 0x795E, 0x496A, 0x9154, 0x7DC8, 0x5450, 0x5464, 0x7FC4, 0x557E, 0x5544, 0x7D64, 0x154, 0xAB44, 0xAA54, 0x448}},
    {40003, 2, {0x2084, 0x2048, 0x79FE, 0x4810, 0x90FC, 0x7C20, 0x55FE, 0x5440, 0x7C7C, 0x54A4, 0x5524, 0x7CFC, 0x44, 0xAA44, 0xABFE, 0x0}},
    {40004, 2, {0x2020, 0x23FE, 0x7820, 0x49FC, 0x9124, 0x7DFC, 0x5524, 0x55FC, 0x7C22, 0x55FE, 0x5408, 0x7DFE, 0x108, 0xAA88, 0xAA28, 0x10}},
    {40005, 2, {0x2010, 0x21FE, 0x7910, 0x497C, 0x9114, 0x7DFE, 0x5514, 0x557C, 0x7D10, 0x557C, 0x5554, 0x7D7C, 0x154, 0xAA7C, 0xAA54, 0x44C}},
    {40006, 2, {0x2010, 0x20FE, 0x7C44, 0x4428, 0x88FE, 0x7E00, 0x527C, 0x5244, 0x7E7C, 0x5244, 0x527C, 0x7E10, 0xFE, 0xAA10, 0xAA10, 0x10}},
    {40007, 2, {0x2020, 0x2010, 0x79FE, 0x4910, 0x917C, 0x7D14, 0x55FE, 0x5514, 0x7D7C, 0x5510, 0x5592, 0x7D54, 0x138, 0xAA54, 0xAA92, 0x430}},
    {40008, 2, {0x2000, 0x21FC, 0x7820, 0x4BFE, 0x9222, 0x7DAC, 0x5420, 0x55AC, 0x7C00, 0x55FC, 0x5404, 0x7DFC, 0x4, 0xAA04, 0xABFC, 0x4}},
    {40009, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4FE6, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {40010, 2, {0x20FC, 0x2028, 0x7810, 0x49FE, 0x9052, 0x7C94, 0x5550, 0x5420, 0x7CFE, 0x54AA, 0x54C6, 0x7CBA, 0xAA, 0xAABA, 0xAA82, 0x86}},
    {40011, 2, {0x2020, 0x203C, 0x7820, 0x49FE, 0x9122, 0x7D38, 0x55E4, 0x551C, 0x7D00, 0x5528, 0x55AA, 0x7D6C, 0x128, 0xAA28, 0xAAFE, 0x0}},
    {40012, 2, {0x2080, 0x20F0, 0x7910, 0x49FC, 0x9324, 0x7D24, 0x55FC, 0x5440, 0x7CA4, 0x5558, 0x5498, 0x7D34, 0x54, 0xAA92, 0xAB50, 0x20}},
    {40013, 2, {0x2012, 0x21D4, 0x7848, 0x4A8A, 0x9104, 0x7E02, 0x55DC, 0x5454, 0x7C62, 0x55C0, 0x551E, 0x7DC2, 0x54, 0xAA48, 0xAB54, 0xA2}},
    {40014, 2, {0x2018, 0x21E0, 0x7840, 0x4BFE, 0x9088, 0x7D74, 0x5652, 0x5470, 0x7C00, 0x55FC, 0x5504, 0x7D74, 0x154, 0xAB74, 0xAB04, 0x10C}},
    {40015, 2, {0x2000, 0x20FE, 0x7C28, 0x44FE, 0x88AA, 0x7EFE, 0x5200, 0x527C, 0x7E44, 0x527C, 0x5244, 0x7E7C, 0x10, 0xAAFE, 0xAA10, 0x10}},
    {40016, 2, {0x2020, 0x21FC, 0x7024, 0x93FE, 0x2024, 0xF9FC, 0xA820, 0xAAAA, 0xFBAE, 0xAA22, 0xABFE, 0xFA22, 0x3AE, 0xAAAA, 0xAAAA, 0x422}},
    {40017, 2, {0x2048, 0x2048, 0x78FC, 0x4848, 0x9048, 0x7DFE, 0x5420, 0x54FC, 0x7CA4, 0x54FC, 0x54A4, 0x7CFC, 0x0, 0xAA48, 0xAA84, 0x102}},
    {40018, 2, {0x2088, 0x2050, 0x79FE, 0x4850, 0x91FC, 0x7D54, 0x559C, 0x5504, 0x7DFC, 0x5504, 0x55FC, 0x7C08, 0x1FE, 0xAA88, 0xAA48, 0x18}},
    {40019, 2, {0x2000, 0x21DC, 0x7954, 0x4954, 0x91DC, 0x7C00, 0x55FC, 0x5524, 0x7DFC, 0x5524, 0x55FC, 0x7C20, 0x3FE, 0xAA20, 0xAA20, 0x20}},
    {40020, 2, {0x2044, 0x2028, 0x79FE, 0x4810, 0x90FE, 0x7C10, 0x55FE, 0x5492, 0x7C54, 0x55FE, 0x5400, 0x7CFE, 0x82, 0xAA82, 0xAAFE, 0x82}},
    {40021, 2, {0x201C, 0x21E0, 0x78A4, 0x4868, 0x91FE, 0x7CA8, 0x5524, 0x5602, 0x7CFC, 0x54A4, 0x54A4, 0x7CFC, 0xA4, 0xAAA4, 0xAAFC, 0x84}},
    {40022, 2, {0x2000, 0x23FE, 0x7200, 0x9314, 0x22A4, 0xFBF6, 0xAA4A, 0xAA42, 0xFB50, 0xAB54, 0xABF4, 0xFA44, 0x24A, 0xAA8A, 0xAC92, 0x920}},
    {40023, 2, {0x2010, 0x2092, 0x7854, 0x4810, 0x91FE, 0x7C54, 0x5492, 0x5510, 0x7C84, 0x54EE, 0x5524, 0x7CB4, 0x5E, 0xAA44, 0xAA84, 0x104}},
    {40024, 2, {0x2000, 0x21FE, 0x7802, 0x49FE, 0x9002, 0x7DFE, 0x5400, 0x55DE, 0x7C92, 0x57DE, 0x5404, 0x7DFE, 0x84, 0xAA44, 0xAA54, 0x8}},
    {40025, 2, {0x2020, 0x2020, 0x78FC, 0x4820, 0x91FE, 0x7C88, 0x55DC, 0x5488, 0x7CC8, 0x55BC, 0x5400, 0x7DFE, 0x50, 0xAA50, 0xAA94, 0x10C}},
    {40026, 2, {0x2010, 0x2010, 0x7CFE, 0x4410, 0x887C, 0x7E00, 0x527C, 0x5244, 0x7E7C, 0x5228, 0x52FE, 0x7E00, 0x7C, 0xAA44, 0xAA7C, 0x44}},
    {40027, 2, {0x2044, 0x2024, 0x7C28, 0x44FE, 0x8892, 0x7ED6, 0x52BA, 0x5292, 0x7EFE, 0x5200, 0x527C, 0x7E44, 0x7C, 0xAA44, 0xAA7C, 0x44}},
    {40028, 2, {0x2280, 0x224E, 0x74EA, 0x95AA, 0x29AA, 0xFEEA, 0xAAAC, 0xACAA, 0xFDEA, 0xAF8A, 0xAA8A, 0xFAAA, 0x4AC, 0xACE8, 0xA8A8, 0x8}},
    {40029, 2, {0x2020, 0x21FC, 0x7820, 0x4888, 0x93FE, 0x7C88, 0x55FC, 0x5504, 0x7DFC, 0x5504, 0x55FC, 0x7D04, 0x1FC, 0xAA88, 0xAB04, 0x202}},
    {40030, 2, {0x2000, 0x21FC, 0x7954, 0x4954, 0x91FC, 0x7C00, 0x57FE, 0x5400, 0x7DFC, 0x5504, 0x55FC, 0x7C62, 0x94, 0xAB88, 0xAAA4, 0xC2}},
    {40031, 2, {0xA80, 0x3138, 0x2288, 0x3938, 0x2288, 0xFFFE, 0x8802, 0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x2488, 0x4244}},
    {40032, 2, {0x2020, 0x2050, 0x7888, 0x4974, 0x9202, 0x7DFC, 0x5524, 0x55AC, 0x7D24, 0x55FC, 0x5400, 0x7CF8, 0x88, 0xAAF8, 0xAA88, 0xF8}},
    {40033, 2, {0x2008, 0x200C, 0x738A, 0x92FE, 0x2288, 0xFBA8, 0xAAA8, 0xAAAA, 0xFBFA, 0xAAAA, 0xAAAC, 0xFBAC, 0x2A, 0xAAAA, 0xAC56, 0x22}},
    {40034, 2, {0x20F8, 0x2088, 0x7888, 0x48F8, 0x9000, 0x7DDC, 0x5554, 0x5554, 0x7DDC, 0x5420, 0x57FE, 0x7C70, 0xA8, 0xAB24, 0xAA22, 0x20}},
    {40035, 2, {0x2010, 0x21FE, 0x7800, 0x49FE, 0x9102, 0x7D7A, 0x554A, 0x55FE, 0x7C00, 0x54FC, 0x5484, 0x7CFC, 0x84, 0xAAFC, 0xAA00, 0x1FE}},
    {40036, 2, {0x2014, 0x2012, 0x73FE, 0x9210, 0x22F0, 0xFA14, 0xAAF4, 0xAAB8, 0xFAEA, 0xAA16, 0xAC42, 0xF824, 0x2A2, 0xAA8A, 0xAC78, 0x0}},
    {40037, 2, {0x2020, 0x2138, 0x7920, 0x4BFE, 0x9010, 0x7C14, 0x55FE, 0x5510, 0x7DF0, 0x5554, 0x5554, 0x7DD4, 0x168, 0xAA4A, 0xAA96, 0x122}},
    {40038, 2, {0x2000, 0x21FC, 0x7954, 0x4954, 0x91DC, 0x7C50, 0x55DC, 0x5554, 0x7D54, 0x55DC, 0x5554, 0x7D54, 0x1DC, 0xAB50, 0xAA52, 0x3E}},
    {40039, 2, {0x2050, 0x21FC, 0x7954, 0x49FC, 0x9154, 0x7DFC, 0x5400, 0x55FC, 0x7C00, 0x55FC, 0x5504, 0x7DFC, 0x88, 0xAA50, 0xABFE, 0x0}},
    {40040, 2, {0x2124, 0x20A8, 0x7BFE, 0x4A8A, 0x90F8, 0x7D0C, 0x55F0, 0x5502, 0x7CFE, 0x5400, 0x55FC, 0x7D04, 0x1FC, 0xAB04, 0xABFC, 0x104}},
    {40041, 2, {0x2000, 0x21FC, 0x7820, 0x4BFE, 0x9222, 0x7DAC, 0x5420, 0x55AC, 0x7C00, 0x55FC, 0x5524, 0x7D24, 0x1FC, 0xAB24, 0xAB24, 0x1FC}},
    {40042, 2, {0x21FC, 0x2104, 0x79FC, 0x4904, 0x91FC, 0x7C22, 0x55FC, 0x5428, 0x7FFE, 0x5440, 0x55FC, 0x7E84, 0xFC, 0xAA84, 0xAAFC, 0x84}},
    {40043, 2, {0x200E, 0x21F0, 0x78A2, 0x4854, 0x91FE, 0x7D22, 0x5414, 0x5552, 0x7D4A, 0x5638, 0x5440, 0x7C7C, 0x84, 0xAB48, 0xAA30, 0x1CE}},
    {40044, 2, {0x2000, 0x21FC, 0x7820, 0x4BFE, 0x9222, 0x7DAC, 0x5420, 0x55AC, 0x7C00, 0x57FE, 0x5420, 0x7DFC, 0x154, 0xAB54, 0xAB54, 0x10C}},
    {40045, 2, {0x2040, 0x2020, 0x73FE, 0x9088, 0x2052, 0xFFAC, 0xAAAA, 0xAAA8, 0xFDAC, 0xA800, 0xA904, 0xF9FC, 0x104, 0xA9FC, 0xA904, 0x204}},
    {40046, 2, {0x20A0, 0x212C, 0x7934, 0x4924, 0x91AC, 0x7D34, 0x5554, 0x55DC, 0x7D54, 0x5554, 0x57FE, 0x7C00, 0x48, 0xAA44, 0xAA82, 0x102}},
    {40047, 2, {0x2044, 0x21FE, 0x7854, 0x4848, 0x90FE, 0x7D90, 0x54FE, 0x5490, 0x7CFE, 0x5490, 0x54FE, 0x7C80, 0x1FE, 0xAA44, 0xAA38, 0x1C6}},
    {40048, 2, {0x2000, 0x23FE, 0x7A52, 0x4A52, 0x93FE, 0x7C24, 0x55FC, 0x5428, 0x7FFE, 0x5420, 0x54FC, 0x7D84, 0x2FC, 0xAA84, 0xAAFC, 0x84}},
    {40049, 2, {0x2000, 0x21FE, 0x7928, 0x49FE, 0x9128, 0x7D7C, 0x5554, 0x557C, 0x7D54, 0x557C, 0x5510, 0x7DFE, 0x192, 0xAABA, 0xAA8A, 0x484}},
    {40050, 2, {0x2124, 0x2248, 0x7124, 0x9000, 0x23FC, 0xFA94, 0xAA64, 0xAA94, 0xFBFC, 0xAA48, 0xAB68, 0xFA48, 0x36A, 0xAA4A, 0xAA46, 0x362}},
    {40051, 2, {0x2124, 0x2174, 0x7258, 0x92DA, 0x23FE, 0xF954, 0xAADA, 0xABFE, 0xF8A2, 0xA820, 0xABFE, 0xF870, 0xA8, 0xA924, 0xAA22, 0x20}},
    {40052, 2, {0x2044, 0x21FE, 0x7C44, 0x4400, 0x88FE, 0x7EAA, 0x52AA, 0x52FE, 0x7E14, 0x52FE, 0x5290, 0x7ED2, 0xAC, 0xAAAA, 0xAA96, 0x122}},
    {40053, 2, {0x2110, 0x2110, 0x71DE, 0x92A8, 0x2454, 0xF812, 0xABFE, 0xAA10, 0xFAF4, 0xAA14, 0xAAF4, 0xFAA8, 0x2AA, 0xAAFA, 0xAA16, 0x422}},
    {40054, 2, {0x2088, 0x23FE, 0x7820, 0x49FC, 0x9040, 0x7FFE, 0x5488, 0x5524, 0x7EFA, 0x5488, 0x54F8, 0x7C88, 0xFA, 0xAA94, 0xAAA8, 0xC4}},
    {40055, 2, {0x2000, 0x23FE, 0x7020, 0x93AE, 0x22AA, 0xFAAA, 0xABAE, 0xA820, 0xFBFE, 0xA820, 0xABAE, 0xFAAA, 0x2AA, 0xABAE, 0xA820, 0x3FE}},
    {40056, 2, {0x2020, 0x203E, 0x7820, 0x49FE, 0x9122, 0x7DF8, 0x5522, 0x557E, 0x7D54, 0x557C, 0x5554, 0x7D7C, 0x100, 0xAAFE, 0xAAAA, 0x5FE}},
    {40057, 2, {0x2088, 0x23FE, 0x7888, 0x4800, 0x91DC, 0x7D54, 0x55DC, 0x5488, 0x7DFE, 0x5690, 0x54FC, 0x7C90, 0xFC, 0xAA90, 0xAAFE, 0x80}},
    {40058, 2, {0x20EE, 0x2000, 0x7CEE, 0x44AA, 0x88EE, 0x7EAA, 0x5210, 0x52FE, 0x7EA8, 0x52FE, 0x52AA, 0x7EFE, 0xAA, 0xAABC, 0xAAAA, 0x136}},
    {40059, 2, {0xFE0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x24A8, 0x5E34, 0x2450, 0x7EFC, 0x2A54, 0x3E7C, 0x2A54, 0xBE7C, 0x95AA}},
    {40060, 2, {0x400, 0x400, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0xFFFE, 0x0}},
    {40061, 2, {0x2000, 0x2000, 0x7DFC, 0x4444, 0x8844, 0x7C44, 0x5444, 0x5444, 0x7C44, 0x5444, 0x5444, 0x7C84, 0x84, 0x1D04, 0xE128, 0x4210}},
    {40062, 2, {0x1000, 0x1000, 0x1EF8, 0x2208, 0x2408, 0x7E08, 0xAA08, 0x2AF8, 0x3E88, 0x2A80, 0x2A80, 0x3E80, 0x82, 0xE82, 0x707E, 0x2000}},
    {40063, 2, {0x2040, 0x2050, 0x7C48, 0x4448, 0x8840, 0x7DFE, 0x5450, 0x5450, 0x7C50, 0x5450, 0x5490, 0x7C90, 0x92, 0x1D12, 0xE10E, 0x4200}},
    {40064, 2, {0x2020, 0x2020, 0x7C20, 0x45FE, 0x8820, 0x7D24, 0x5524, 0x5524, 0x7D24, 0x55FC, 0x5424, 0x7C20, 0x22, 0x1C22, 0xE01E, 0x4000}},
    {40065, 2, {0x800, 0x1FE0, 0x2020, 0x7FF8, 0xA108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0}},
    {40066, 2, {0x1020, 0x1010, 0x1E10, 0x23FE, 0x2440, 0x7E40, 0xAA40, 0x2A7C, 0x3E44, 0x2A44, 0x2A44, 0x3E44, 0x84, 0xE84, 0x7128, 0x2210}},
    {40067, 2, {0x2000, 0x2000, 0x7DFC, 0x4524, 0x8924, 0x7D24, 0x5524, 0x5524, 0x7DFC, 0x5500, 0x5500, 0x7D00, 0x102, 0x1D02, 0xE0FE, 0x4000}},
    {40068, 2, {0x2000, 0x2000, 0x7DFE, 0x4408, 0x8808, 0x7DE8, 0x5528, 0x5528, 0x7D28, 0x5528, 0x55E8, 0x7D28, 0x8, 0x1C08, 0xE028, 0x4010}},
    {40069, 2, {0x2050, 0x2048, 0x7C48, 0x4440, 0x89FE, 0x7C40, 0x5440, 0x547C, 0x7CA4, 0x54A4, 0x54A8, 0x7D28, 0x110, 0x1E28, 0xE044, 0x4082}},
    {40070, 2, {0x2000, 0x2000, 0x7DFC, 0x4420, 0x8820, 0x7D24, 0x54A4, 0x54A8, 0x7C20, 0x57FE, 0x5420, 0x7C20, 0x20, 0x1C20, 0xE020, 0x4020}},
    {40071, 2, {0x1010, 0x1010, 0x1E10, 0x2210, 0x241E, 0x7E10, 0xAA10, 0x2A10, 0x3EFC, 0x2A84, 0x2A84, 0x3E84, 0x84, 0xE84, 0x70FC, 0x2084}},
    {40072, 2, {0x1010, 0x1010, 0x1E1E, 0x2210, 0x2410, 0x7EFE, 0xAA82, 0x2A82, 0x3E82, 0x2AFE, 0x2A82, 0x3E80, 0x80, 0xF00, 0x7100, 0x2200}},
    {40073, 2, {0x2020, 0x2020, 0x7C20, 0x4420, 0x89FC, 0x7D24, 0x5524, 0x5524, 0x7D24, 0x55FC, 0x5524, 0x7D24, 0x124, 0x1D24, 0xE1FC, 0x4104}},
    {40074, 2, {0x1040, 0x1040, 0x1E40, 0x227E, 0x24A0, 0x7EA0, 0xAB20, 0x2A3E, 0x3E20, 0x2A20, 0x2A20, 0x3E3E, 0x20, 0xE20, 0x7020, 0x2020}},
    {40075, 2, {0x2044, 0x2044, 0x7C44, 0x4484, 0x88BE, 0x7D84, 0x5684, 0x54A4, 0x7C94, 0x5494, 0x5484, 0x7C84, 0x84, 0x1C84, 0xE094, 0x4088}},
    {40076, 2, {0x1010, 0x1010, 0x1E20, 0x22FC, 0x2484, 0x7E84, 0xAA84, 0x2A84, 0x3EFC, 0x2A84, 0x2A84, 0x3E84, 0x84, 0xE84, 0x70FC, 0x2084}},
    {40077, 2, {0x2040, 0x2040, 0x7CFC, 0x4484, 0x8904, 0x7EF4, 0x5494, 0x5494, 0x7CF4, 0x5484, 0x5494, 0x7C88, 0x82, 0x1C82, 0xE07E, 0x4000}},
    {40078, 2, {0x2208, 0x1110, 0x20, 0x7FFC, 0x4404, 0x8408, 0xFC0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE, 0x0}},
    {40079, 2, {0x2010, 0x2010, 0x7C10, 0x45FE, 0x8912, 0x7D14, 0x5510, 0x55FC, 0x7D44, 0x5544, 0x5528, 0x7D28, 0x110, 0x1E28, 0xE244, 0x4482}},
    {40080, 2, {0x1020, 0x1020, 0x1E20, 0x2240, 0x2448, 0x7E84, 0xABFE, 0x2A82, 0x3E00, 0x2AFC, 0x2A84, 0x3E84, 0x84, 0xE84, 0x70FC, 0x2084}},
    {40081, 2, {0x2020, 0x2020, 0x7DFC, 0x4420, 0x8820, 0x7C20, 0x57FE, 0x5400, 0x7C20, 0x5420, 0x55FC, 0x7C20, 0x20, 0x1C20, 0xE3FE, 0x4000}},
    {40082, 2, {0x2020, 0x2020, 0x7C20, 0x47FE, 0x8820, 0x7C20, 0x55FC, 0x5400, 0x7C00, 0x55FC, 0x5504, 0x7D04, 0x104, 0x1D04, 0xE1FC, 0x4104}},
    {40083, 2, {0x2040, 0x2044, 0x7DF4, 0x4448, 0x8850, 0x7DFE, 0x5440, 0x5480, 0x7DFE, 0x5640, 0x5480, 0x7CFC, 0x4, 0x1C04, 0xE028, 0x4010}},
    {40084, 2, {0x2020, 0x2020, 0x7DFE, 0x4440, 0x8840, 0x7CFC, 0x5484, 0x5584, 0x7EFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0x1C84, 0xE094, 0x4088}},
    {40085, 2, {0x2000, 0x21FE, 0x7C20, 0x4420, 0x8840, 0x7DFC, 0x5554, 0x5554, 0x7D54, 0x5554, 0x5554, 0x7D54, 0x154, 0x1D54, 0xE104, 0x410C}},
    {40086, 2, {0x2000, 0x21FC, 0x7D04, 0x4504, 0x8974, 0x7D04, 0x5504, 0x5574, 0x7D54, 0x5554, 0x5554, 0x7D74, 0x104, 0x1D04, 0xE114, 0x4108}},
    {40087, 2, {0x4002, 0x43E2, 0x7A22, 0x8A2A, 0x12AA, 0xFAAA, 0xAAAA, 0xAAAA, 0xFAAA, 0xAAAA, 0xAAAA, 0xF882, 0x142, 0x1922, 0xE20A, 0x4404}},
    {40088, 2, {0x2008, 0x201C, 0x7DE0, 0x4500, 0x8900, 0x7DFE, 0x5500, 0x5500, 0x7D7C, 0x5544, 0x5544, 0x7D44, 0x144, 0x1E7C, 0xE244, 0x4400}},
    {40089, 2, {0x2020, 0x2020, 0x7C50, 0x4488, 0x8904, 0x7E02, 0x54F8, 0x5400, 0x7C00, 0x57FE, 0x5420, 0x7C40, 0x88, 0x1D04, 0xE3FE, 0x4102}},
    {40090, 2, {0x2040, 0x2020, 0x7DFE, 0x4488, 0x8850, 0x7C20, 0x5450, 0x5488, 0x7D06, 0x5488, 0x5488, 0x7C88, 0x88, 0x1D08, 0xE108, 0x4208}},
    {40091, 2, {0x2040, 0x2020, 0x7C20, 0x45FE, 0x8800, 0x7C88, 0x5504, 0x5602, 0x7C88, 0x5488, 0x5450, 0x7C50, 0x20, 0x1C50, 0xE088, 0x4306}},
    {40092, 2, {0x1082, 0x1044, 0x1E28, 0x22FE, 0x2410, 0x7E10, 0xAA10, 0x2A7C, 0x3E10, 0x2A10, 0x2A10, 0x3EFE, 0x10, 0xE10, 0x7010, 0x2010}},
    {40093, 2, {0x440, 0x7FFC, 0x100, 0x3FF8, 0x400, 0xFFFE, 0x1400, 0x2FC0, 0xD040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE}},
    {40094, 2, {0x1110, 0x920, 0x7FFC, 0x200, 0xFFFE, 0x820, 0x3418, 0xCFC6, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE}},
    {40095, 2, {0x2000, 0x21FC, 0x7C04, 0x4404, 0x88FC, 0x7C04, 0x5404, 0x55FC, 0x7C08, 0x5408, 0x55FE, 0x7C88, 0x48, 0x1C48, 0xE008, 0x4018}},
    {40096, 2, {0x2000, 0x21FE, 0x7C20, 0x4420, 0x89FC, 0x7D24, 0x5524, 0x55FC, 0x7D24, 0x5524, 0x55FC, 0x7D20, 0xA0, 0x1C40, 0xE0B0, 0x430E}},
    {40097, 2, {0x4000, 0x43FE, 0x7800, 0x8800, 0x13DE, 0xFA52, 0xAA52, 0xAA52, 0xFB5A, 0xAAD6, 0xAA52, 0xFA52, 0x252, 0x1A52, 0xE252, 0x42D6}},
    {40098, 2, {0x4010, 0x4210, 0x7910, 0x897E, 0x1020, 0xF828, 0xAF48, 0xA97E, 0xF908, 0xA908, 0xA9FE, 0xF908, 0x108, 0x1A88, 0xE47E, 0x4000}},
    {40099, 2, {0x2040, 0x217E, 0x7D52, 0x4552, 0x8954, 0x7D48, 0x5554, 0x5462, 0x7C10, 0x5410, 0x54FE, 0x7C10, 0x10, 0x1C10, 0xE1FE, 0x4000}},
    {40100, 2, {0x2000, 0x21FC, 0x7D24, 0x4524, 0x89FC, 0x7D24, 0x5524, 0x55FC, 0x7C20, 0x5420, 0x55FC, 0x7C20, 0x20, 0x1C20, 0xE3FE, 0x4000}},
    {40101, 2, {0x4004, 0x4004, 0x7BC4, 0x8A44, 0x127E, 0xFA44, 0xAA44, 0xABE4, 0xFA54, 0xAA54, 0xAA44, 0xFA44, 0x3C4, 0x1A44, 0xE014, 0x4008}},
    {40102, 2, {0x2040, 0x2040, 0x7CF8, 0x4488, 0x8950, 0x7C20, 0x54D8, 0x5726, 0x7C20, 0x55FC, 0x5420, 0x7D28, 0x124, 0x1E24, 0xE0A0, 0x4040}},
    {40103, 2, {0x200C, 0x21F0, 0x7C20, 0x4444, 0x8888, 0x7DF0, 0x5420, 0x5444, 0x7DFE, 0x5422, 0x5420, 0x7CA8, 0x124, 0x1E22, 0xE0A0, 0x4040}},
    {40104, 2, {0x2040, 0x1248, 0x4444, 0x2014, 0x860, 0x7380, 0x2800, 0xFC0, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE}},
    {40105, 2, {0x2020, 0x2010, 0x7DFE, 0x4502, 0x8A04, 0x7CF8, 0x5400, 0x5400, 0x7DFE, 0x5450, 0x5450, 0x7C50, 0x92, 0x1C92, 0xE10E, 0x4200}},
    {40106, 2, {0x1000, 0x10FC, 0x1E24, 0x2224, 0x25FE, 0x7E24, 0xAA24, 0x2AFC, 0x3E40, 0x2A40, 0x2AFC, 0x3EC4, 0x144, 0xE44, 0x707C, 0x2044}},
    {40107, 2, {0x4000, 0x43DE, 0x7A52, 0x8A52, 0x1252, 0xFBD2, 0xAA52, 0xAA52, 0xFA52, 0xABD2, 0xAA1A, 0xFA94, 0x250, 0x1AB0, 0xE310, 0x4010}},
    {40108, 2, {0x2000, 0x21F8, 0x7C08, 0x44D0, 0x8820, 0x7DFC, 0x5524, 0x5524, 0x7DFC, 0x5524, 0x5524, 0x7DFC, 0x124, 0x1D24, 0xE124, 0x410C}},
    {40109, 2, {0x2020, 0x2020, 0x7DFE, 0x4420, 0x88FC, 0x7C20, 0x55FE, 0x5400, 0x7CFC, 0x5484, 0x54FC, 0x7C84, 0xFC, 0x1C84, 0xE094, 0x4088}},
    {40110, 2, {0x2020, 0x2020, 0x7CFC, 0x4420, 0x8820, 0x7DFE, 0x5448, 0x5484, 0x7D22, 0x5478, 0x5488, 0x7D48, 0x50, 0x1C20, 0xE050, 0x418C}},
    {40111, 2, {0x2088, 0x2088, 0x7DFC, 0x4488, 0x8888, 0x7CF8, 0x5488, 0x5488, 0x7CF8, 0x5488, 0x5488, 0x7DFE, 0x0, 0x1C88, 0xE104, 0x4202}},
    {40112, 2, {0x2000, 0x23E0, 0x7D5C, 0x4554, 0x8954, 0x7DD4, 0x5554, 0x5554, 0x7DD4, 0x5554, 0x5548, 0x7D68, 0x3C8, 0x1C54, 0xE054, 0x4062}},
    {40113, 2, {0x2050, 0x2050, 0x7C50, 0x47DE, 0x8850, 0x7C50, 0x5450, 0x55DC, 0x7C50, 0x5450, 0x5450, 0x7C50, 0x3DE, 0x1C50, 0xE050, 0x4050}},
    {40114, 2, {0x2000, 0x21FC, 0x7D04, 0x4504, 0x89FC, 0x7D04, 0x5504, 0x55FC, 0x7C00, 0x5512, 0x55D4, 0x7D18, 0x110, 0x1D52, 0xE192, 0x410E}},
    {40115, 2, {0x20FC, 0x2084, 0x7C84, 0x44FC, 0x8884, 0x7C84, 0x54FC, 0x5400, 0x7DFE, 0x5502, 0x5502, 0x7DFE, 0x102, 0x1D02, 0xE1FE, 0x4102}},
    {40116, 2, {0x2000, 0x21FC, 0x7D04, 0x4524, 0x8924, 0x7DFC, 0x5524, 0x5524, 0x7D74, 0x5554, 0x5554, 0x7D74, 0x104, 0x1D04, 0xE1FC, 0x4104}},
    {40117, 2, {0x2000, 0x2040, 0x7D9C, 0x4504, 0x8904, 0x7DDC, 0x5504, 0x5504, 0x7DFC, 0x5450, 0x5450, 0x7C50, 0x90, 0x1C92, 0xE112, 0x420E}},
    {40118, 2, {0x4020, 0x4020, 0x7850, 0x8888, 0x1144, 0xFA22, 0xA9F8, 0xA808, 0xF850, 0xA820, 0xA8A4, 0xFA82, 0x28A, 0x1A8A, 0xE478, 0x4000}},
    {40119, 2, {0x2000, 0x21FC, 0x7D24, 0x4524, 0x8974, 0x7D24, 0x55FC, 0x5504, 0x7D74, 0x5554, 0x5554, 0x7D74, 0x104, 0x1E04, 0xE214, 0x4408}},
    {40120, 2, {0x2040, 0x2020, 0x7DFE, 0x4400, 0x8800, 0x7CF8, 0x5488, 0x5488, 0x7C88, 0x54F8, 0x5420, 0x7CA8, 0x124, 0x1E22, 0xE0A0, 0x4040}},
    {40121, 2, {0x2040, 0x2048, 0x7C84, 0x45FE, 0x8820, 0x7FFE, 0x5488, 0x5524, 0x7E42, 0x5588, 0x5410, 0x7C62, 0x184, 0x1C18, 0xE060, 0x4380}},
    {40122, 2, {0x2000, 0x21FC, 0x7C04, 0x4434, 0x89C4, 0x7C44, 0x5444, 0x55F4, 0x7D54, 0x5554, 0x55F4, 0x7C44, 0x54, 0x1C76, 0xE196, 0x4002}},
    {40123, 2, {0x2092, 0x2092, 0x7D24, 0x4648, 0x8924, 0x7C92, 0x5492, 0x5400, 0x7DFE, 0x5522, 0x5522, 0x7DFE, 0x122, 0x1D22, 0xE1FE, 0x4102}},
    {40124, 2, {0x2020, 0x21FC, 0x7C20, 0x4488, 0x8BFE, 0x7C88, 0x5400, 0x55FC, 0x7D04, 0x5524, 0x5524, 0x7D24, 0x124, 0x1C50, 0xE088, 0x4304}},
    {40125, 2, {0x4048, 0x4148, 0x7948, 0x8BFE, 0x1148, 0xF948, 0xA978, 0xA900, 0xF9FE, 0xA820, 0xABFE, 0xF870, 0xA8, 0x1924, 0xE622, 0x4020}},
    {40126, 2, {0x1000, 0x10FE, 0x1E00, 0x227C, 0x2444, 0x7E44, 0xAA7C, 0x2A00, 0x3EFE, 0x2A92, 0x2A92, 0x3EFE, 0x92, 0xE92, 0x70FE, 0x2082}},
    {40127, 2, {0x2020, 0x2124, 0x7CA8, 0x4420, 0x89FE, 0x7D02, 0x5400, 0x54F8, 0x7C00, 0x5400, 0x55FE, 0x7C20, 0x40, 0x1C88, 0xE1FC, 0x4084}},
    {40128, 2, {0x1000, 0x10FC, 0x1E84, 0x2284, 0x24FC, 0x7E84, 0xAA84, 0x2AFC, 0x3E00, 0x2BFE, 0x2A20, 0x3EA0, 0xBC, 0xEA0, 0x7160, 0x223E}},
    {40129, 2, {0x2000, 0x20F8, 0x7C88, 0x4488, 0x88F8, 0x7C88, 0x5488, 0x54F8, 0x7C00, 0x55FC, 0x5554, 0x7D54, 0x154, 0x1D54, 0xE3FE, 0x4000}},
    {40130, 2, {0x2000, 0x21FC, 0x7D24, 0x4524, 0x89FC, 0x7D24, 0x5524, 0x55FC, 0x7C00, 0x57FE, 0x5520, 0x7D22, 0x114, 0x1D48, 0xE184, 0x4102}},
    {40131, 2, {0x2000, 0x21FC, 0x7D24, 0x4524, 0x89FC, 0x7D24, 0x5524, 0x55FC, 0x7C20, 0x5410, 0x5454, 0x7D42, 0x142, 0x1D4A, 0xE238, 0x4000}},
    {40132, 2, {0x2000, 0x21DC, 0x7D54, 0x4554, 0x89DC, 0x7C00, 0x55FC, 0x5400, 0x7DFE, 0x5480, 0x54FC, 0x7C04, 0x4, 0x1C04, 0xE028, 0x4010}},
    {40133, 2, {0x4008, 0x40C8, 0x7B08, 0x8908, 0x112A, 0xF92A, 0xABAC, 0xA948, 0xF908, 0xAB88, 0xAD54, 0xF914, 0x114, 0x1924, 0xE124, 0x4142}},
    {40134, 2, {0x2040, 0x2040, 0x7CFE, 0x4480, 0x89FC, 0x7E84, 0x54FC, 0x5484, 0x7CFC, 0x5440, 0x547C, 0x7CC4, 0x128, 0x1C10, 0xE068, 0x4186}},
    {40135, 2, {0x2020, 0x2040, 0x7DFC, 0x4504, 0x89FC, 0x7D04, 0x55FC, 0x5400, 0x7FFE, 0x5420, 0x5420, 0x7DFC, 0x20, 0x1C20, 0xE3FE, 0x4000}},
    {40136, 2, {0x2040, 0x2080, 0x7DFC, 0x4504, 0x89FC, 0x7D04, 0x55FC, 0x5420, 0x7C32, 0x57B4, 0x54A8, 0x7D28, 0x124, 0x1E24, 0xE4A2, 0x4040}},
    {40137, 2, {0x2090, 0x2090, 0x7C9E, 0x46A2, 0x89D4, 0x7C88, 0x5494, 0x54A4, 0x7DFE, 0x5684, 0x54A4, 0x7C94, 0x94, 0x1C84, 0xE094, 0x4088}},
    {40138, 2, {0x2020, 0x2010, 0x7DFE, 0x4502, 0x8902, 0x7DFE, 0x5500, 0x5500, 0x7DFE, 0x55AA, 0x56AA, 0x7CFE, 0xAA, 0x1CAA, 0xE0AA, 0x4086}},
    {40139, 2, {0x2000, 0x21FC, 0x7CA4, 0x4488, 0x8850, 0x7C20, 0x54D8, 0x5726, 0x7C20, 0x55FC, 0x5524, 0x7D24, 0x1FC, 0x1C22, 0xE3FE, 0x4002}},
    {40140, 2, {0x820, 0x7E20, 0x87E, 0xFEC4, 0x1028, 0x1E10, 0x2228, 0x4FC6, 0x9040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE}},
    {40141, 2, {0x2040, 0x2044, 0x7DF8, 0x4450, 0x8BFE, 0x7C40, 0x54F8, 0x5582, 0x7E7E, 0x5400, 0x55FC, 0x7D04, 0x1FC, 0x1D04, 0xE1FC, 0x4104}},
    {40142, 2, {0x2000, 0x20FC, 0x7C84, 0x44FC, 0x8884, 0x7CFC, 0x5400, 0x5400, 0x7DDC, 0x5444, 0x5554, 0x7CCC, 0x154, 0x1C44, 0xE154, 0x4088}},
    {40143, 2, {0x2000, 0x21FC, 0x7D54, 0x4554, 0x8954, 0x7DFC, 0x5420, 0x54A8, 0x7CA4, 0x5524, 0x5420, 0x7CA8, 0xA4, 0x1D24, 0xE020, 0x4020}},
    {40144, 2, {0x2004, 0x201E, 0x7DE0, 0x4402, 0x8922, 0x7C94, 0x54FE, 0x5510, 0x7C10, 0x55FE, 0x5410, 0x7C92, 0x92, 0x1C92, 0xE0FE, 0x4002}},
    {40145, 2, {0x2020, 0x2010, 0x7DFE, 0x4484, 0x8848, 0x7DFE, 0x5502, 0x5624, 0x7C10, 0x55FE, 0x5440, 0x7C7C, 0x44, 0x1C84, 0xE094, 0x4108}},
    {40146, 2, {0x2088, 0x2050, 0x7C00, 0x45FE, 0x8850, 0x7DFC, 0x5454, 0x57FE, 0x7C54, 0x55FC, 0x5450, 0x7CD8, 0x154, 0x1E52, 0xE050, 0x4050}},
    {40147, 2, {0x4148, 0x4148, 0x7BE8, 0x8948, 0x11DE, 0xF88A, 0xABEA, 0xAAAA, 0xFAAA, 0xABEA, 0xA88A, 0xFBEA, 0x92, 0x1892, 0xE0AA, 0x40C4}},
    {40148, 2, {0x2000, 0x23FE, 0x7C50, 0x45FC, 0x8954, 0x7D54, 0x55FC, 0x5400, 0x7DFC, 0x5400, 0x57FE, 0x7C20, 0xA8, 0x1D24, 0xE2A2, 0x4040}},
    {40149, 2, {0x2000, 0x21FC, 0x7C20, 0x47FE, 0x8A22, 0x7DAC, 0x5420, 0x55AC, 0x7C00, 0x55FC, 0x5404, 0x7DFC, 0x4, 0x1C04, 0xE1FC, 0x4004}},
    {40150, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4FC6, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE}},
    {40151, 2, {0x4000, 0x41F8, 0x7908, 0x89F8, 0x1108, 0xF9F8, 0xA800, 0xABFC, 0xFA94, 0xABFC, 0xA800, 0xF9F8, 0x90, 0x1860, 0xE198, 0x4606}},
    {40152, 2, {0x2020, 0x3F20, 0x407E, 0xBEA8, 0x2A28, 0xFF10, 0x4A28, 0x7FC6, 0x1040, 0x3FF0, 0x5110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0xFFFE}},
    {40153, 2, {0x2010, 0x21FE, 0x7D10, 0x457C, 0x8914, 0x7DFE, 0x5514, 0x557C, 0x7D10, 0x557C, 0x5554, 0x7D7C, 0x154, 0x1E7C, 0xE254, 0x444C}},
    {40154, 2, {0x4004, 0x43E4, 0x7A24, 0x8A24, 0x13EE, 0xFA04, 0xAAE4, 0xAA04, 0xFA0C, 0xABF4, 0xAA44, 0xFA64, 0x554, 0x1D44, 0xEA54, 0x40C8}},
    {40155, 2, {0x2000, 0x21DE, 0x7C42, 0x4552, 0x88CA, 0x7D52, 0x5442, 0x5420, 0x7CFC, 0x5484, 0x5484, 0x7CFC, 0x84, 0x1C84, 0xE0FC, 0x4084}},
    {40156, 2, {0x4000, 0x43FE, 0x7A00, 0x8B14, 0x12A4, 0xFBF6, 0xAA4A, 0xAA42, 0xFB50, 0xAB54, 0xABF4, 0xFA44, 0x24A, 0x1A8A, 0xE492, 0x4920}},
    {40157, 2, {0x4088, 0x4050, 0x7BFE, 0x8820, 0x11FC, 0xF820, 0xABFE, 0xA924, 0xF8A8, 0xABFE, 0xA800, 0xF9FC, 0x104, 0x1904, 0xE1FC, 0x4104}},
    {40158, 2, {0x4020, 0x4124, 0x78A8, 0x8BFE, 0x10A8, 0xF924, 0xAA22, 0xA904, 0xF904, 0xA9DE, 0xAA44, 0xFD54, 0x9E, 0x1904, 0xE204, 0x4404}},
    {40159, 2, {0x2088, 0x2050, 0x7DFE, 0x4450, 0x89FC, 0x7D54, 0x559C, 0x5504, 0x7DFC, 0x5504, 0x55FC, 0x7C08, 0x1FE, 0x1C88, 0xE048, 0x4018}},
    {40160, 2, {0x2088, 0x23FE, 0x7CA8, 0x4490, 0x89FE, 0x7F20, 0x55FC, 0x5520, 0x7DFC, 0x5520, 0x55FE, 0x7D00, 0x3FC, 0x1C88, 0xE070, 0x438E}},
    {40161, 2, {0x4014, 0x4012, 0x7BFE, 0x8A10, 0x12F0, 0xFA14, 0xAAF4, 0xAAB8, 0xFAEA, 0xAA16, 0xAC42, 0xF824, 0x2A2, 0x1A8A, 0xE478, 0x4000}},
    {40162, 2, {0x2050, 0x21FC, 0x7D54, 0x45FC, 0x8954, 0x7DFC, 0x5400, 0x55FC, 0x7C00, 0x55FC, 0x5504, 0x7DFC, 0x88, 0x1C50, 0xE3FE, 0x4000}},
    {40163, 2, {0x2010, 0x21FE, 0x7C00, 0x45FE, 0x8902, 0x7D7A, 0x554A, 0x55FE, 0x7C00, 0x54FC, 0x5484, 0x7CFC, 0x84, 0x1CFC, 0xE000, 0x41FE}},
    {40164, 2, {0x2108, 0x2108, 0x7DEE, 0x4694, 0x8842, 0x7C10, 0x55FE, 0x5502, 0x7CF8, 0x5488, 0x54F8, 0x7C80, 0xFC, 0x1C84, 0xE0FC, 0x4084}},
    {40165, 2, {0x200, 0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x924, 0x2494, 0x2494, 0x4028, 0x10}},
    {40166, 2, {0x1000, 0x2000, 0x7F3E, 0x4102, 0x7F04, 0x4104, 0x7F08, 0x4008, 0x7F90, 0x4010, 0x7FA0, 0xA2, 0xAAA2, 0xAAA2, 0x811E, 0x600}},
    {40167, 2, {0x800, 0x3FE0, 0x2020, 0x3FE0, 0x2020, 0x3FE0, 0x2000, 0x3FF8, 0x2000, 0x3FF8, 0x5248, 0x8928, 0x1FD8, 0x1042, 0x2042, 0xC03E}},
    {40168, 2, {0x1010, 0x2010, 0x7F10, 0x4110, 0x7F7E, 0x4112, 0x7F12, 0x4012, 0x7F92, 0x4012, 0x7F92, 0x92, 0xAAA2, 0xAAA2, 0x814A, 0x684}},
    {40169, 2, {0x2040, 0x2080, 0x21FC, 0x2104, 0xF9FC, 0x2904, 0x29FC, 0x2900, 0x29FE, 0x2900, 0x2BFE, 0x2C02, 0x4AAA, 0x42AA, 0x8202, 0xC}},
    {40170, 2, {0x2040, 0x2080, 0x21FC, 0x2104, 0x21FC, 0x2904, 0x25FC, 0x2500, 0x25FE, 0x2100, 0x21FE, 0x2002, 0x22AA, 0x22AA, 0x2202, 0x200C}},
    {40171, 2, {0x0, 0x3FFE, 0x2100, 0x2200, 0x27F0, 0x2410, 0x27F0, 0x2410, 0x27F0, 0x2400, 0x27FC, 0x2400, 0x27FC, 0x4004, 0x4AA4, 0x8AAC}},
    {40172, 2, {0x400, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0xF94, 0x888, 0x882, 0x1082, 0x607E}},
    {40173, 2, {0x40, 0x80, 0xF9FC, 0x2904, 0x29FC, 0x2904, 0x29FC, 0x2900, 0x29FE, 0x2900, 0x29FE, 0x2802, 0x4AAA, 0x4AAA, 0x9A02, 0xC}},
    {40174, 2, {0x40, 0x80, 0x1FC, 0xFD04, 0x5FC, 0x504, 0x49FC, 0x2900, 0x11FE, 0x1100, 0x29FE, 0x2402, 0x46AA, 0x82AA, 0x202, 0xC}},
    {40175, 2, {0x0, 0x7FF8, 0x4408, 0x5FC8, 0x5048, 0x5FC8, 0x5048, 0x5FC8, 0x5008, 0x5FE8, 0x5008, 0x5FEA, 0x402A, 0xAAAA, 0xAAA6, 0x42}},
    {40176, 2, {0x40, 0x80, 0x81FC, 0x4104, 0x21FC, 0x2104, 0x11FC, 0x1100, 0x29FE, 0x2900, 0x25FE, 0x2402, 0x42AA, 0x42AA, 0x8202, 0xC}},
    {40177, 2, {0x40, 0x7C80, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0xFF00, 0x11FE, 0x1100, 0x11FE, 0x1002, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40178, 2, {0x40, 0x80, 0x3DFC, 0x2504, 0x25FC, 0x2504, 0x3DFC, 0x2500, 0x21FE, 0x2100, 0x21FE, 0x2002, 0x22AA, 0x42AA, 0x4202, 0x800C}},
    {40179, 2, {0x3FF8, 0x2008, 0x2FE8, 0x2208, 0x2FE8, 0x2828, 0x2FE8, 0x2828, 0x2FE8, 0x2808, 0x2FF8, 0x2808, 0x2FFA, 0x401A, 0x5556, 0x8032}},
    {40180, 2, {0x40, 0x80, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0x4900, 0x49FE, 0x4900, 0x49FE, 0x7802, 0x4AAA, 0x2AA, 0x202, 0xC}},
    {40181, 2, {0x40, 0xF880, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x49FC, 0xF900, 0x49FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x4AAA, 0x4602, 0x420C}},
    {40182, 2, {0x420, 0x210, 0x13E, 0xFFC0, 0x222, 0x1FFA, 0x1016, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40183, 2, {0x1040, 0x1080, 0x11FC, 0xFF04, 0x11FC, 0x1104, 0xFDFC, 0x500, 0x5FE, 0x4900, 0x29FE, 0x1002, 0x2AAA, 0x46AA, 0x8202, 0xC}},
    {40184, 2, {0x100, 0x7FFC, 0x4004, 0x7FFC, 0x4200, 0x5FF0, 0x5010, 0x5FF0, 0x5010, 0x5FF0, 0x5000, 0x5FFC, 0x5000, 0x5FFC, 0xAA44, 0x452C}},
    {40185, 2, {0x1040, 0x1080, 0x29FC, 0x2904, 0x45FC, 0xA504, 0x11FC, 0x1100, 0x1FE, 0xFD00, 0x5FE, 0x802, 0xAAA, 0x12AA, 0x1202, 0xC}},
    {40186, 2, {0x1040, 0x1080, 0x11FC, 0x7D04, 0x11FC, 0x1104, 0x11FC, 0xFD00, 0x11FE, 0x1100, 0x11FE, 0x2802, 0x26AA, 0x42AA, 0x8202, 0xC}},
    {40187, 2, {0x20, 0x2840, 0x24FC, 0x2484, 0x42FC, 0x4284, 0x80FC, 0x7C80, 0x24FE, 0x2480, 0x24FE, 0x2402, 0x44AA, 0x44AA, 0x9502, 0x80C}},
    {40188, 2, {0x40, 0x2080, 0x11FC, 0x104, 0xFDFC, 0x904, 0x9FC, 0x4900, 0x29FE, 0x1100, 0x11FE, 0x2802, 0x2AAA, 0x4AAA, 0x8202, 0xC}},
    {40189, 2, {0x40, 0x7C80, 0x29FC, 0x2904, 0x29FC, 0x2904, 0x29FC, 0xFF00, 0x29FE, 0x2900, 0x29FE, 0x2802, 0x2AAA, 0x2AAA, 0x4A02, 0x880C}},
    {40190, 2, {0x40, 0xFE80, 0x11FC, 0x1104, 0x7DFC, 0x5504, 0x55FC, 0x5500, 0x55FE, 0x5500, 0x55FE, 0x5C02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40191, 2, {0x40, 0x80, 0x7DFC, 0x1104, 0x11FC, 0x1104, 0x7DFC, 0x1100, 0x11FE, 0x1100, 0x11FE, 0x1C02, 0xE2AA, 0x42AA, 0x202, 0xC}},
    {40192, 2, {0x40, 0xFC80, 0x11FC, 0x1104, 0x21FC, 0x3104, 0x69FC, 0x6500, 0xA5FE, 0x2100, 0x21FE, 0x2002, 0x22AA, 0x22AA, 0x2202, 0x200C}},
    {40193, 2, {0x40, 0xC80, 0x71FC, 0x1104, 0x11FC, 0x1104, 0xFDFC, 0x1100, 0x11FE, 0x1100, 0x29FE, 0x2802, 0x42AA, 0x42AA, 0x8202, 0xC}},
    {40194, 2, {0x2040, 0x2080, 0x21FC, 0xF904, 0x29FC, 0x2904, 0x29FC, 0x2900, 0xFDFE, 0x2100, 0x21FE, 0x2002, 0x32AA, 0x4AAA, 0x4A02, 0x800C}},
    {40195, 2, {0x1010, 0x2010, 0x7F10, 0x4110, 0x7F7C, 0x4114, 0x7F14, 0x4014, 0x7F94, 0x407E, 0x7F90, 0x90, 0xAAA8, 0xAAA8, 0x8144, 0x682}},
    {40196, 2, {0x20, 0x40, 0xFEFC, 0xA884, 0xA8FC, 0xA884, 0xA8FC, 0xA880, 0xAAFE, 0xAA80, 0xA6FE, 0xC002, 0x82AA, 0xFAAA, 0x202, 0xC}},
    {40197, 2, {0x40, 0x7880, 0x49FC, 0x4904, 0x69FC, 0x5904, 0x49FC, 0xFD00, 0x49FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x4AAA, 0x4A02, 0x980C}},
    {40198, 2, {0x1040, 0x1080, 0x11FC, 0x7D04, 0x55FC, 0x9104, 0x11FC, 0x2900, 0x29FE, 0x2900, 0x29FE, 0x2C02, 0x4AAA, 0x42AA, 0x8202, 0xC}},
    {40199, 2, {0x4040, 0x4880, 0x51FC, 0x6104, 0x45FC, 0x4504, 0x3DFC, 0x1100, 0x11FE, 0x1100, 0xFDFE, 0x1002, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40200, 2, {0x0, 0x7FFE, 0x4440, 0x45FC, 0x4904, 0x49FC, 0x5904, 0x69FC, 0x4900, 0x49FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x8AAA, 0x804}},
    {40201, 2, {0x40, 0xFC80, 0x9FC, 0x4904, 0x49FC, 0x4904, 0x89FC, 0xFF00, 0x19FE, 0x1900, 0x29FE, 0x2802, 0x4AAA, 0x8AAA, 0x2A02, 0x100C}},
    {40202, 2, {0x40, 0x80, 0xFDFC, 0x1104, 0x11FC, 0x1104, 0x51FC, 0x5D00, 0x51FE, 0x5100, 0x51FE, 0x5002, 0x5EAA, 0xF2AA, 0x202, 0xC}},
    {40203, 2, {0x2020, 0x1040, 0xFC, 0xFE84, 0x20FC, 0x2084, 0x3CFC, 0x2480, 0x24FE, 0x2480, 0x24FE, 0x2402, 0x44AA, 0x54AA, 0x8902, 0xC}},
    {40204, 2, {0x0, 0x3FF8, 0x100, 0x7FFC, 0x440, 0x1830, 0xE20E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40205, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x7C0, 0x1830, 0xE20E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40206, 2, {0x40, 0xFC80, 0x81FC, 0x8904, 0xC9FC, 0xA904, 0x91FC, 0x9100, 0xA9FE, 0xA900, 0xC9FE, 0x8002, 0x82AA, 0xFEAA, 0x202, 0xC}},
    {40207, 2, {0x1050, 0x2054, 0x7C52, 0x4492, 0x7C90, 0x459E, 0x7DF0, 0x4290, 0x7E90, 0x4090, 0x7E90, 0x290, 0xAA8A, 0xAA8A, 0x8286, 0xC82}},
    {40208, 2, {0x1000, 0x7E78, 0x1248, 0x2678, 0xC200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40209, 2, {0x800, 0x7E78, 0x1248, 0xC30, 0x32CE, 0xDFF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40210, 2, {0x1040, 0x1080, 0x29FC, 0x2504, 0x43FC, 0x9104, 0x9FC, 0x100, 0xFDFE, 0x500, 0x9FE, 0x5002, 0x22AA, 0x12AA, 0x1202, 0xC}},
    {40211, 2, {0x2020, 0x1040, 0x14FC, 0x484, 0x24FC, 0x2484, 0xAAFC, 0xAA80, 0xA8FE, 0x3080, 0x32FE, 0x2202, 0x22AA, 0x5EAA, 0x8102, 0xC}},
    {40212, 2, {0x840, 0x1C80, 0xE1FC, 0x104, 0x21FC, 0x1104, 0xFDFC, 0x500, 0x9FE, 0x900, 0x11FE, 0x2002, 0x42AA, 0xA2AA, 0x1E02, 0xC}},
    {40213, 2, {0x1010, 0x2008, 0x7F08, 0x417E, 0x7F42, 0x4142, 0x7F20, 0x4024, 0x7FA8, 0x4030, 0x7FA0, 0xA2, 0xAAA2, 0xAAA2, 0x811E, 0x600}},
    {40214, 2, {0x20, 0x7E40, 0x42FC, 0x4284, 0x7EFC, 0x4884, 0x48FC, 0x4880, 0x7EFE, 0x4880, 0x48FE, 0x4802, 0x4AAA, 0x4AAA, 0x6702, 0x420C}},
    {40215, 2, {0x2040, 0x1080, 0x11FC, 0xFD04, 0x1FC, 0x904, 0x89FC, 0x8900, 0x49FE, 0x4900, 0x51FE, 0x5002, 0x1EAA, 0xE2AA, 0x4202, 0xC}},
    {40216, 2, {0x1040, 0x1080, 0x21FC, 0x4904, 0xFDFC, 0x504, 0x29FC, 0x2900, 0x29FE, 0xFD00, 0x29FE, 0x2802, 0x2AAA, 0x4AAA, 0x4A02, 0x880C}},
    {40217, 2, {0x2040, 0x2080, 0x21FC, 0x7D04, 0x51FC, 0x9104, 0x11FC, 0x1100, 0xFDFE, 0x1100, 0x11FE, 0x2802, 0x26AA, 0x42AA, 0x8202, 0xC}},
    {40218, 2, {0x1000, 0x2000, 0x7DFE, 0x4404, 0x7C04, 0x44F4, 0x7C94, 0x4094, 0x7E94, 0x4094, 0x7EF4, 0x294, 0xAA04, 0xAA04, 0x8214, 0xC08}},
    {40219, 2, {0x20F8, 0x3E88, 0x449A, 0xA882, 0x127E, 0x3FF0, 0xD010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40220, 2, {0x844, 0x2F78, 0x2842, 0x2F3E, 0xF200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40221, 2, {0x2020, 0x2040, 0x20FC, 0x7E84, 0x42FC, 0x8284, 0x7AFC, 0x4A80, 0x4AFE, 0x4A80, 0x7AFE, 0x4A02, 0x2AA, 0x2AA, 0x1502, 0x80C}},
    {40222, 2, {0x40, 0x7880, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x1FC, 0xFD00, 0x21FE, 0x4100, 0x79FE, 0x802, 0xAAA, 0xAAA, 0x5202, 0x200C}},
    {40223, 2, {0x440, 0xE80, 0xF1FC, 0x9104, 0x91FC, 0x9104, 0xFDFC, 0x9100, 0x91FE, 0x9100, 0x91FE, 0x8802, 0x8AAA, 0xC6AA, 0x9202, 0x80C}},
    {40224, 2, {0x40, 0x80, 0x7DFC, 0x4504, 0x45FC, 0x7D04, 0x45FC, 0x4500, 0x7DFE, 0x100, 0x1FE, 0x1C02, 0xE2AA, 0x42AA, 0x202, 0xC}},
    {40225, 2, {0x40, 0x7880, 0x49FC, 0x4904, 0x49FC, 0x7904, 0x49FC, 0x4900, 0x79FE, 0x4900, 0x49FE, 0x4802, 0x4EAA, 0xF2AA, 0x202, 0xC}},
    {40226, 2, {0x408, 0x2410, 0x243C, 0x4424, 0x5FBC, 0x94A4, 0xE4BC, 0x24A0, 0x44BE, 0x44A0, 0x94BE, 0xF482, 0x18AA, 0x8AA, 0x1282, 0x210C}},
    {40227, 2, {0x1020, 0x1040, 0x10FC, 0x1084, 0xFEFC, 0x1084, 0x10FC, 0x1080, 0x7CFE, 0x4480, 0x44FE, 0x4402, 0x44AA, 0x7CAA, 0x4502, 0xC}},
    {40228, 2, {0x1020, 0x2020, 0x7F3C, 0x4144, 0x7F68, 0x41A8, 0x7F10, 0x4028, 0x7FC4, 0x4082, 0x7FA0, 0x90, 0xAA88, 0xAAA0, 0x8110, 0x608}},
    {40229, 2, {0x1020, 0x2010, 0x7C10, 0x44FE, 0x7C82, 0x4504, 0x7C00, 0x4028, 0x7E28, 0x4028, 0x7E28, 0x228, 0xAA44, 0xAA44, 0x8244, 0xC82}},
    {40230, 2, {0x100, 0x1FF0, 0x1110, 0xFFFE, 0x280, 0x1C70, 0xE20E, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40231, 2, {0x2020, 0x1040, 0x10FC, 0xFE84, 0x82FC, 0x8484, 0xFC, 0x7880, 0x48FE, 0x4880, 0x48FE, 0x4A02, 0x4EAA, 0x4AAA, 0x8202, 0xC}},
    {40232, 2, {0x40, 0x7C80, 0x55FC, 0x5504, 0x55FC, 0x7D04, 0x55FC, 0x5500, 0x55FE, 0x7D00, 0x55FE, 0x1002, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40233, 2, {0x1040, 0x1080, 0x51FC, 0x5104, 0x7DFC, 0x9104, 0x11FC, 0x1100, 0xFDFE, 0x1100, 0x11FE, 0x2802, 0x26AA, 0x42AA, 0x8202, 0xC}},
    {40234, 2, {0x2020, 0x1040, 0x10FC, 0xFE84, 0x82FC, 0x8284, 0x28FC, 0x2880, 0x24FE, 0x2480, 0x44FE, 0x4002, 0x42AA, 0x82AA, 0x202, 0xC}},
    {40235, 2, {0x40, 0x80, 0xF9FC, 0xA904, 0xA9FC, 0xA904, 0xA9FC, 0xF900, 0xA9FE, 0xA900, 0xA9FE, 0xA802, 0xFAAA, 0x8AAA, 0x202, 0xC}},
    {40236, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4202, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40237, 2, {0x1040, 0x2080, 0x79FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4100, 0x41FE, 0x7900, 0x49FE, 0x4802, 0x4AAA, 0x7AAA, 0x4A02, 0xC}},
    {40238, 2, {0x1020, 0x1040, 0xFEFC, 0x2884, 0x44FC, 0x8284, 0x7CFC, 0x80, 0xFEFE, 0x2080, 0x40FE, 0x7C02, 0x4AA, 0x4AA, 0x2902, 0x100C}},
    {40239, 2, {0x20, 0xFE40, 0x10FC, 0x1084, 0x20FC, 0xFE84, 0xAAFC, 0xAA80, 0xAAFE, 0xAA80, 0xAAFE, 0xAA02, 0xAAAA, 0xA2AA, 0x8B02, 0x840C}},
    {40240, 2, {0x820, 0x1C40, 0x70FC, 0x1084, 0x10FC, 0xFC84, 0x10FC, 0x1080, 0x7CFE, 0x4480, 0x44FE, 0x4402, 0x44AA, 0x7CAA, 0x4502, 0xC}},
    {40241, 2, {0x2840, 0x2880, 0xFDFC, 0x2904, 0x29FC, 0x104, 0x5FC, 0x4500, 0x29FE, 0x2900, 0x11FE, 0x1002, 0x2AAA, 0x2AAA, 0x4A02, 0x800C}},
    {40242, 2, {0x8020, 0x8840, 0xB0FC, 0xC084, 0x84FC, 0x8484, 0x7CFC, 0x80, 0xF8FE, 0x8880, 0x88FE, 0xF802, 0x8AAA, 0x8AAA, 0xFA02, 0x880C}},
    {40243, 2, {0x1020, 0x840, 0x7EFC, 0x4284, 0x94FC, 0x1084, 0x10FC, 0xFE80, 0x24FE, 0x2480, 0x48FE, 0x2802, 0x12AA, 0x2AAA, 0x4602, 0x800C}},
    {40244, 2, {0x1010, 0x1020, 0x277C, 0x4044, 0x907C, 0x1044, 0x2FFC, 0x6240, 0xA27E, 0x2240, 0x227E, 0x2202, 0x22AA, 0x22AA, 0x2B02, 0x240C}},
    {40245, 2, {0x2820, 0x2840, 0x28FC, 0x2884, 0xAAFC, 0x6C84, 0x28FC, 0x2880, 0x6CFE, 0xAA80, 0x28FE, 0x2802, 0x2AAA, 0x4CAA, 0x4902, 0x800C}},
    {40246, 2, {0x1020, 0x1040, 0x10FC, 0xFE84, 0x10FC, 0x1084, 0x7CFC, 0x80, 0xFE, 0x7C80, 0x44FE, 0x4402, 0x44AA, 0x7CAA, 0x4502, 0xC}},
    {40247, 2, {0x7F04, 0x844, 0x1E44, 0x2244, 0x5444, 0x844, 0x320C, 0xDFF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40248, 2, {0x1040, 0x5080, 0x51FC, 0x7D04, 0x91FC, 0x1104, 0x11FC, 0xFF00, 0x31FE, 0x3900, 0x55FE, 0x5402, 0x92AA, 0x12AA, 0x1202, 0x100C}},
    {40249, 2, {0x4440, 0x2480, 0x29FC, 0xFF04, 0x11FC, 0x1104, 0x7DFC, 0x1100, 0x11FE, 0xFF00, 0x11FE, 0x1002, 0x22AA, 0x22AA, 0x4202, 0x800C}},
    {40250, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x1484, 0x7CFC, 0x5080, 0x7EFE, 0x1280, 0x12FE, 0x1602, 0x28AA, 0x24AA, 0x4502, 0x800C}},
    {40251, 2, {0x20, 0x8040, 0x40FC, 0x1F84, 0x84FC, 0x4484, 0x44FC, 0x480, 0x24FE, 0x4480, 0xC4FE, 0x4702, 0x5CAA, 0x48AA, 0x4102, 0xC}},
    {40252, 2, {0x1010, 0x2010, 0x7F3C, 0x4124, 0x7F44, 0x4128, 0x7F10, 0x4028, 0x7FC6, 0x4000, 0x7FBC, 0xA4, 0xAAA4, 0xAAA4, 0x813C, 0x624}},
    {40253, 2, {0x1000, 0xFE7C, 0x2244, 0x1C7C, 0xE200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40254, 2, {0x1040, 0x1080, 0x21FC, 0x4904, 0xFDFC, 0x4504, 0x11FC, 0x5100, 0x7DFE, 0x9100, 0x11FE, 0xFC02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40255, 2, {0x2040, 0x2080, 0x51FC, 0x4904, 0x85FC, 0x104, 0xF9FC, 0x100, 0x1FE, 0xF900, 0x89FE, 0x8802, 0x8AAA, 0xFAAA, 0x8A02, 0xC}},
    {40256, 2, {0x818, 0x13E0, 0x3040, 0x57FC, 0x9040, 0x13F8, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40257, 2, {0x2040, 0x1080, 0x1FC, 0xFD04, 0x1FC, 0x2904, 0x45FC, 0x8300, 0x9FE, 0x4900, 0x29FE, 0x1002, 0x2AAA, 0x4AAA, 0x8202, 0xC}},
    {40258, 2, {0x1208, 0x1210, 0x123C, 0x2224, 0x2FBC, 0x6224, 0xA63C, 0x2720, 0x2ABE, 0x2AA0, 0x323E, 0x2202, 0x222A, 0x222A, 0x2202, 0x220C}},
    {40259, 2, {0x1040, 0x2080, 0x79FC, 0x4904, 0x69FC, 0x5904, 0x49FC, 0xF900, 0x49FE, 0x4900, 0x69FE, 0x5802, 0x4AAA, 0x4AAA, 0x4A02, 0x980C}},
    {40260, 2, {0x40, 0xFE80, 0x21FC, 0x2904, 0x45FC, 0xFD04, 0x15FC, 0x1100, 0x11FE, 0x7D00, 0x11FE, 0x1002, 0x1EAA, 0xE2AA, 0x4202, 0xC}},
    {40261, 2, {0x2040, 0x2080, 0x39FC, 0x4904, 0x49FC, 0xA904, 0x11FC, 0x2900, 0x45FE, 0x8500, 0x79FE, 0x4802, 0x4AAA, 0x4AAA, 0x7A02, 0x480C}},
    {40262, 2, {0x1100, 0x1200, 0x279C, 0x4480, 0x8780, 0x1480, 0x27BE, 0x6408, 0xA7C8, 0x2408, 0x27C8, 0x2048, 0x2548, 0x2548, 0x2868, 0x2190}},
    {40263, 2, {0x2020, 0x2040, 0x3EFC, 0x2884, 0x48FC, 0x884, 0x3CFC, 0x2880, 0x28FE, 0x2880, 0xFEFE, 0x802, 0xAAA, 0xAAA, 0xA02, 0x80C}},
    {40264, 2, {0x40, 0xFC80, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x7900, 0x49FE, 0x4900, 0x4DFE, 0x7802, 0xCAAA, 0xAAA, 0xA02, 0x80C}},
    {40265, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4204, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40266, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x10FC, 0x5484, 0x54FC, 0x5480, 0xBAFE, 0x1180, 0x10FE, 0x2802, 0x28AA, 0x44AA, 0x8502, 0xC}},
    {40267, 2, {0x10, 0x7E20, 0x27C, 0x244, 0x7E7C, 0x4044, 0x417C, 0x4140, 0x3F7E, 0x1040, 0x87E, 0x2902, 0xA2AA, 0xA2AA, 0x1F02, 0xC}},
    {40268, 2, {0x1010, 0x2010, 0x7C28, 0x4428, 0x7C44, 0x44BA, 0x7C10, 0x4010, 0x7EFE, 0x4010, 0x7E58, 0x254, 0xAA92, 0xAB12, 0x8250, 0xC20}},
    {40269, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0xFC, 0x7C84, 0xFC, 0x80, 0xFEFE, 0x2880, 0x28FE, 0x2802, 0x28AA, 0x4AAA, 0x4D02, 0x800C}},
    {40270, 2, {0x420, 0x1E40, 0xE0FC, 0x2284, 0x92FC, 0x5484, 0x48FC, 0x1080, 0xFCFE, 0x2480, 0x44FE, 0x2802, 0x10AA, 0x28AA, 0x4502, 0x800C}},
    {40271, 2, {0x1020, 0x1440, 0x10FC, 0xFE84, 0x10FC, 0x7C84, 0x54FC, 0x5480, 0x7CFE, 0x5480, 0x54FE, 0x7C02, 0x54AA, 0x54AA, 0x5502, 0x4C0C}},
    {40272, 2, {0x20, 0x40, 0x7CFC, 0x1084, 0x10FC, 0x5484, 0x54FC, 0x5480, 0xBAFE, 0x1080, 0x10FE, 0x1C02, 0xE2AA, 0x42AA, 0x202, 0xC}},
    {40273, 2, {0x40, 0xF880, 0x89FC, 0x8904, 0xF9FC, 0x104, 0xF9FC, 0x8900, 0x89FE, 0xF900, 0x89FE, 0x8802, 0xFAAA, 0x8AAA, 0x8A02, 0x980C}},
    {40274, 2, {0x40, 0x4880, 0x45FC, 0x9504, 0x11FC, 0x2904, 0x45FC, 0x8100, 0x79FE, 0x4900, 0x49FE, 0x4802, 0x4AAA, 0x7AAA, 0x4A02, 0xC}},
    {40275, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x10FC, 0xFE84, 0x82FC, 0x480, 0x78FE, 0x880, 0x10FE, 0x1E02, 0xF0AA, 0x10AA, 0x5102, 0x200C}},
    {40276, 2, {0x1010, 0x2010, 0x7C20, 0x4444, 0x7CFE, 0x4428, 0x7C44, 0x40A2, 0x7E3C, 0x4044, 0x7E44, 0x2A8, 0xAA10, 0xAA28, 0x8244, 0xC82}},
    {40277, 2, {0x1020, 0x1040, 0x20FC, 0x4484, 0xFEFC, 0x2884, 0x44FC, 0xA280, 0x3CFE, 0x4480, 0x44FE, 0xA802, 0x12AA, 0x2AAA, 0x4602, 0x800C}},
    {40278, 2, {0x40, 0x80, 0x7DFC, 0x4504, 0x45FC, 0x7D04, 0x45FC, 0x4500, 0x7DFE, 0x4100, 0x51FE, 0x4802, 0x56AA, 0x62AA, 0x4202, 0xC}},
    {40279, 2, {0x4210, 0x2420, 0x187C, 0x2444, 0x527C, 0x1044, 0xFF7C, 0x2040, 0x287E, 0x7E40, 0xAA7E, 0x2A02, 0x2AAA, 0x2EAA, 0x902, 0x80C}},
    {40280, 2, {0x10, 0x7E20, 0x127C, 0x1244, 0xFF7C, 0x1244, 0x127C, 0x7E40, 0x207E, 0x2040, 0x7E7E, 0x6202, 0xA2AA, 0x22AA, 0x3F02, 0x220C}},
    {40281, 2, {0x40, 0x7880, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x49FC, 0x7900, 0x49FE, 0x4900, 0x79FE, 0x2, 0x52AA, 0x4AAA, 0x8A02, 0xC}},
    {40282, 2, {0xC20, 0x7040, 0x10FC, 0xFE84, 0x10FC, 0x3884, 0x54FC, 0x9280, 0xFE, 0x3880, 0x28FE, 0x2802, 0x2AAA, 0x4CAA, 0x4902, 0x800C}},
    {40283, 2, {0x20, 0xFE40, 0x54FC, 0x5484, 0xA8FC, 0x5484, 0x54FC, 0x80, 0xFEFE, 0x1080, 0x10FE, 0x1002, 0x1EAA, 0xF2AA, 0x4202, 0xC}},
    {40284, 2, {0x4420, 0x2440, 0x28FC, 0xFC84, 0x14FC, 0x1484, 0xFCFC, 0x9080, 0x90FE, 0xFE80, 0x32FE, 0x3202, 0x52AA, 0x96AA, 0x1102, 0x100C}},
    {40285, 2, {0x410, 0x3620, 0xE57C, 0x2544, 0x247C, 0xFFC4, 0x247C, 0x2540, 0x257E, 0x3640, 0xE67E, 0x2402, 0x2AAA, 0x32AA, 0xA1C2, 0x408C}},
    {40286, 2, {0x240, 0x3C50, 0x848, 0xFFFE, 0x848, 0xE50, 0x7824, 0x854, 0x1A8C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40287, 2, {0x2800, 0x11FC, 0x6820, 0x14F8, 0x2420, 0xD5FC, 0xA00, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40288, 2, {0x1020, 0x5040, 0x50FC, 0x7C84, 0x90FC, 0x1084, 0xFEFC, 0x80, 0xFE, 0x7C80, 0x44FE, 0x4402, 0x44AA, 0x7CAA, 0x4502, 0xC}},
    {40289, 2, {0x410, 0x620, 0x757C, 0x544, 0x47C, 0xFF44, 0x47C, 0x1440, 0x147E, 0x5C40, 0x547E, 0x5402, 0x552A, 0x5D2A, 0xE342, 0x410C}},
    {40290, 2, {0x1040, 0x2080, 0x79FC, 0x4904, 0x79FC, 0x4904, 0x79FC, 0x4900, 0x49FE, 0xF900, 0x19FE, 0x2802, 0x4AAA, 0x8AAA, 0x2A02, 0x100C}},
    {40291, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x10FC, 0x7C84, 0x54FC, 0x5480, 0x54FE, 0x7C80, 0x10FE, 0x3802, 0x56AA, 0x92AA, 0x1202, 0x100C}},
    {40292, 2, {0x2020, 0x2040, 0x3CFC, 0x4484, 0x48FC, 0xBE84, 0x2AFC, 0x2A80, 0x3EFE, 0x2A80, 0x2AFE, 0x3E02, 0x2AAA, 0x2AAA, 0x2B02, 0x440C}},
    {40293, 2, {0x4904, 0x2A24, 0x7F24, 0x824, 0xFFA4, 0x80C, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40294, 2, {0x1000, 0x20FC, 0x7C04, 0x4404, 0x7C7C, 0x4404, 0x7C04, 0x40FE, 0x7E10, 0x4092, 0x7E54, 0x238, 0xAA54, 0xAA92, 0x8250, 0xC20}},
    {40295, 2, {0x4440, 0x2880, 0x1FC, 0x7D04, 0x29FC, 0x2904, 0x29FC, 0x2900, 0xFDFE, 0x2900, 0x29FE, 0x2802, 0x2AAA, 0x4AAA, 0x4A02, 0x880C}},
    {40296, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x42FC, 0xB884, 0x10FC, 0x1080, 0xFEFE, 0x1080, 0x10FE, 0x7C02, 0x44AA, 0x44AA, 0x7D02, 0x440C}},
    {40297, 2, {0x8, 0x7790, 0x54BC, 0x54A4, 0x54BC, 0x75A4, 0x543C, 0x57A0, 0x54BE, 0x74A0, 0x56BE, 0x5502, 0x552A, 0x56AA, 0x5482, 0xB40C}},
    {40298, 2, {0x1020, 0x1040, 0xFEFC, 0x2884, 0x54FC, 0x9284, 0x7CFC, 0x5480, 0x7CFE, 0x5480, 0x7CFE, 0x1002, 0x16AA, 0x1AAA, 0x1202, 0xC}},
    {40299, 2, {0x1020, 0x1040, 0x1EFC, 0x1084, 0x7CFC, 0x4484, 0x7CFC, 0x4480, 0x7CFE, 0x5480, 0x10FE, 0xFE02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40300, 2, {0x10, 0x7720, 0x557C, 0x5544, 0x557C, 0x7744, 0x557C, 0x5540, 0x557E, 0x7740, 0x557E, 0x5502, 0x552A, 0xB52A, 0x8942, 0x130C}},
    {40301, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x42FC, 0xBC84, 0x10FC, 0x1080, 0xFEFE, 0x1080, 0x94FE, 0x5802, 0x52AA, 0x1EAA, 0xF202, 0x400C}},
    {40302, 2, {0x1020, 0x1040, 0x3EFC, 0x2284, 0x44FC, 0x8084, 0x10FC, 0x6680, 0x42FE, 0x4280, 0x66FE, 0x4202, 0x42AA, 0x7EAA, 0x4302, 0xC}},
    {40303, 2, {0x1040, 0x2080, 0x7DFC, 0x5504, 0x55FC, 0x7D04, 0x55FC, 0x5500, 0x7DFE, 0x2100, 0x51FE, 0xFC02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40304, 2, {0x10, 0x7F20, 0x497C, 0x4944, 0x5D7C, 0x4944, 0x7F7C, 0x4140, 0x5D7E, 0x5540, 0x557E, 0x5D02, 0x412A, 0x412A, 0x4542, 0x820C}},
    {40305, 2, {0x1020, 0x1040, 0x7CFC, 0x1084, 0x10FC, 0xFE84, 0xFC, 0x4480, 0x92FE, 0x1080, 0x7CFE, 0x1002, 0x12AA, 0x1EAA, 0xE202, 0x400C}},
    {40306, 2, {0x2820, 0x2840, 0x28FC, 0xFC84, 0x28FC, 0x2884, 0xFEFC, 0x80, 0x7CFE, 0x4480, 0x44FE, 0x7C02, 0x44AA, 0x44AA, 0x7D02, 0x440C}},
    {40307, 2, {0x1020, 0x840, 0x7EFC, 0x4284, 0x42FC, 0x7E84, 0x40FC, 0x5E80, 0x52FE, 0x5E80, 0x52FE, 0x5E02, 0x52AA, 0x52AA, 0x9302, 0x160C}},
    {40308, 2, {0x4020, 0x4040, 0x7EFC, 0x8284, 0x22FC, 0xAA84, 0x72FC, 0x2280, 0xFEFE, 0x2280, 0x72FE, 0xAA02, 0x22AA, 0x22AA, 0xB02, 0x40C}},
    {40309, 2, {0x2020, 0x2040, 0x3CFC, 0x4484, 0x88FC, 0x7E84, 0x52FC, 0x5280, 0x7EFE, 0x1080, 0x2CFE, 0x2A02, 0x28AA, 0x4AAA, 0x4B02, 0x860C}},
    {40310, 2, {0x20, 0xFE40, 0x28FC, 0x2884, 0x28FC, 0xEE84, 0x82FC, 0x8280, 0x82FE, 0xEE80, 0x28FE, 0x2802, 0x2EAA, 0xF0AA, 0x4102, 0xC}},
    {40311, 2, {0x1010, 0x820, 0xFF7C, 0x8144, 0x407C, 0x7744, 0x557C, 0x5540, 0x557E, 0xB540, 0x977E, 0x1402, 0x242A, 0x252A, 0x4642, 0x840C}},
    {40312, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x28FC, 0x4484, 0xFEFC, 0x480, 0x74FE, 0x5480, 0x54FE, 0x7402, 0x54AA, 0x4AA, 0x1502, 0x80C}},
    {40313, 2, {0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4A08, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40314, 2, {0x2020, 0x1040, 0xFEFC, 0x2884, 0x28FC, 0x4E84, 0x52FC, 0xDA80, 0x56FE, 0x6280, 0x54FE, 0x4C02, 0x48AA, 0x54AA, 0x6502, 0x400C}},
    {40315, 2, {0x1050, 0x2048, 0x7C80, 0x44FE, 0x7D90, 0x4690, 0x7CFC, 0x4090, 0x7E90, 0x40FC, 0x7E90, 0x290, 0xAA90, 0xAAFE, 0x8280, 0xC80}},
    {40316, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0x28FC, 0x4484, 0x82FC, 0x80, 0x7CFE, 0x1080, 0x10FE, 0x1002, 0x1EAA, 0xF2AA, 0x4202, 0xC}},
    {40317, 2, {0x10, 0xF720, 0x117C, 0x5544, 0x227C, 0x5544, 0x897C, 0x1040, 0x7E, 0xF740, 0x117E, 0x5502, 0x222A, 0x552A, 0x8942, 0x100C}},
    {40318, 2, {0x20, 0xFC40, 0x84FC, 0x8484, 0xFCFC, 0x8484, 0x84FC, 0xFC80, 0xFE, 0x8880, 0xEEFE, 0x8802, 0x88AA, 0xAAAA, 0xCD02, 0x880C}},
    {40319, 2, {0x1000, 0x20FE, 0x7C82, 0x44FE, 0x7C82, 0x44FE, 0x7C00, 0x4014, 0x7EE4, 0x4044, 0x7E44, 0x3FE, 0xAA44, 0xAA44, 0x8284, 0xD04}},
    {40320, 2, {0x1024, 0x2024, 0x7F7E, 0x4124, 0x7F24, 0x413C, 0x7F24, 0x4024, 0x7FBC, 0x4024, 0x7FA4, 0xFE, 0xAA80, 0xAAA4, 0x8122, 0x642}},
    {40321, 2, {0x2020, 0x1040, 0xFEFC, 0x84, 0xFC, 0x7C84, 0x44FC, 0x4480, 0x7CFE, 0x1080, 0x54FE, 0x5202, 0x92AA, 0x10AA, 0x5102, 0x200C}},
    {40322, 2, {0x2020, 0xCE40, 0x82FC, 0x8284, 0xEEFC, 0x8284, 0x82FC, 0xFE80, 0x28FE, 0x2880, 0x28FE, 0x2802, 0x4AAA, 0x4CAA, 0x8902, 0xC}},
    {40323, 2, {0x1000, 0x2020, 0x7CCE, 0x4482, 0x7C82, 0x44EE, 0x7C82, 0x4082, 0x7EFE, 0x4028, 0x7E28, 0x228, 0xAA48, 0xAA4A, 0x828A, 0xD06}},
    {40324, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x1084, 0xFEFC, 0x80, 0x7CFE, 0x4480, 0x7CFE, 0x4402, 0x7CAA, 0x44AA, 0x5502, 0x480C}},
    {40325, 2, {0x40, 0x5480, 0x55FC, 0xA904, 0xA9FC, 0x5504, 0x55FC, 0x100, 0xF9FE, 0xA900, 0xA9FE, 0xF802, 0xAAAA, 0xAAAA, 0xFA02, 0x880C}},
    {40326, 2, {0x1040, 0x1080, 0x11FC, 0x7D04, 0x11FC, 0x5504, 0x55FC, 0x5500, 0xBBFE, 0x1100, 0x39FE, 0x5402, 0x92AA, 0x12AA, 0x1202, 0x100C}},
    {40327, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x5484, 0x7CFC, 0x5480, 0x7CFE, 0x1080, 0x38FE, 0x5402, 0x92AA, 0x12AA, 0x1202, 0x100C}},
    {40328, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x1484, 0xFEFC, 0x1480, 0x7CFE, 0x1080, 0xFCFE, 0x2402, 0x44AA, 0x28AA, 0x3102, 0xC80C}},
    {40329, 2, {0x2020, 0x1040, 0xFEFC, 0x84, 0x7CFC, 0x4484, 0x7CFC, 0x80, 0x7CFE, 0x480, 0x8FE, 0x1E02, 0xF2AA, 0x12AA, 0x5202, 0x200C}},
    {40330, 2, {0x1010, 0x820, 0x7F7C, 0x4844, 0x5E7C, 0x4A44, 0x4A7C, 0x7F40, 0x4A7E, 0x4A40, 0x5E7E, 0x4802, 0x542A, 0x922A, 0x2242, 0x400C}},
    {40331, 2, {0x20, 0x7E40, 0x42FC, 0x4284, 0x7EFC, 0x4884, 0x48FC, 0x7E80, 0x48FE, 0x4880, 0x5CFE, 0x5402, 0x54AA, 0x54AA, 0x5D02, 0x940C}},
    {40332, 2, {0x10, 0x7F20, 0x417C, 0x4144, 0x7F7C, 0x4444, 0x557C, 0x5540, 0x557E, 0x5F40, 0x447E, 0x5502, 0x552A, 0x572A, 0x9D42, 0xC}},
    {40333, 2, {0x20, 0x7C40, 0x44FC, 0x7C84, 0x44FC, 0x4484, 0x7CFC, 0x2080, 0x7EFE, 0xAA80, 0x2AFE, 0x4A02, 0x92AA, 0x22AA, 0x4B02, 0x40C}},
    {40334, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0xFC, 0x7C84, 0xFC, 0x80, 0xFEFE, 0x1080, 0x10FE, 0x5402, 0x52AA, 0x92AA, 0x5102, 0x200C}},
    {40335, 2, {0xE20, 0xF040, 0x22FC, 0x9284, 0x44FC, 0x1084, 0x10FC, 0x7C80, 0x10FE, 0x1080, 0xFCFE, 0x1002, 0x2AAA, 0x26AA, 0x4202, 0x800C}},
    {40336, 2, {0x820, 0x8C40, 0x4AFC, 0x4884, 0x3EFC, 0x884, 0xC8FC, 0x5C80, 0x6AFE, 0x4880, 0x48FE, 0x4802, 0x4AAA, 0x62AA, 0x9A02, 0xC}},
    {40337, 2, {0x2AA4, 0x2CA8, 0x4920, 0x1450, 0x2288, 0x5FF4, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1000, 0x1FFC, 0x1000, 0x1FFC, 0x2A44, 0x452C}},
    {40338, 2, {0x2208, 0x2210, 0xFA3C, 0x2224, 0x27BC, 0xFAA4, 0xAABC, 0xAAA0, 0xFABE, 0x22A0, 0x72BE, 0xAA82, 0x2AAA, 0x24AA, 0x2482, 0x298C}},
    {40339, 2, {0x5040, 0x5080, 0xF9FC, 0x5104, 0x51FC, 0x104, 0xF9FC, 0xA900, 0xA9FE, 0xA900, 0xF9FE, 0xA802, 0xAAAA, 0xAAAA, 0xFA02, 0x880C}},
    {40340, 2, {0x10, 0x7E20, 0x247C, 0x1844, 0x7F7C, 0x2944, 0x4A7C, 0xA840, 0x107E, 0x840, 0xFF7E, 0x1C02, 0x2A2A, 0x492A, 0x8842, 0x80C}},
    {40341, 2, {0x2040, 0x1080, 0x79FC, 0x104, 0x49FC, 0x3104, 0xFDFC, 0x100, 0x79FE, 0x4900, 0x49FE, 0x7802, 0x4AAA, 0x4AAA, 0x7A02, 0x480C}},
    {40342, 2, {0x610, 0x3810, 0x852, 0x7E54, 0x890, 0x1C28, 0x2A28, 0x4844, 0xA82, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40343, 2, {0x40, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x80, 0xFEFE, 0x1080, 0x10FE, 0x5E02, 0x52AA, 0x52AA, 0xB004, 0x9FFE, 0x0}},
    {40344, 2, {0x2010, 0x2720, 0x257C, 0x7544, 0x257C, 0x2744, 0x257C, 0x7540, 0x557E, 0x5740, 0x557E, 0x5502, 0x752A, 0x492A, 0x942, 0x130C}},
    {40345, 2, {0x2020, 0x1040, 0xFEFC, 0x84, 0x44FC, 0x2884, 0xFEFC, 0x8280, 0x10FE, 0x7C80, 0x54FE, 0x5402, 0x54AA, 0x5CAA, 0x1102, 0x100C}},
    {40346, 2, {0x20, 0xEE40, 0xAAFC, 0xAA84, 0xEEFC, 0x84, 0x78FC, 0x80, 0xFCFE, 0x2080, 0x40FE, 0x7802, 0xAAA, 0xAAA, 0x2A02, 0x100C}},
    {40347, 2, {0x4820, 0x4A40, 0x6CFC, 0x4884, 0x4AFC, 0x6A84, 0x56FC, 0x2080, 0x7EFE, 0x4280, 0x42FE, 0x7E02, 0x42AA, 0x42AA, 0x7F02, 0x420C}},
    {40348, 2, {0x4420, 0xFE40, 0x44FC, 0x84, 0x7CFC, 0x484, 0x28FC, 0x1080, 0xFEFE, 0x3280, 0x54FE, 0x9002, 0x12AA, 0x12AA, 0x5202, 0x200C}},
    {40349, 2, {0x20, 0xFE40, 0xFC, 0x7C84, 0x44FC, 0x4484, 0x7CFC, 0x80, 0xFEFE, 0x9280, 0x92FE, 0xFE02, 0x92AA, 0x92AA, 0xFF02, 0x820C}},
    {40350, 2, {0x210, 0x720, 0x3C7C, 0x2444, 0x247C, 0x3F44, 0x247C, 0x2F40, 0x297E, 0x2F40, 0x297E, 0x2F02, 0x292A, 0x492A, 0x4F42, 0x890C}},
    {40351, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0x28FC, 0x4484, 0x92FC, 0x1480, 0x12FE, 0xFE80, 0x10FE, 0x1002, 0x2AAA, 0x26AA, 0x4202, 0x800C}},
    {40352, 2, {0x10, 0xFFA0, 0x807C, 0xBF44, 0xA17C, 0xBF44, 0xA17C, 0xBF40, 0x887E, 0xFFC0, 0x927E, 0xB202, 0x8CAA, 0x92AA, 0xA102, 0xFF8C}},
    {40353, 2, {0x20, 0x7C40, 0x44FC, 0x7C84, 0x44FC, 0x7C84, 0x20FC, 0x7E80, 0x92FE, 0x5280, 0x6AFE, 0x4202, 0x7AAA, 0x2AA, 0x1502, 0x80C}},
    {40354, 2, {0xE20, 0xF040, 0x22FC, 0x9284, 0x54FC, 0x4884, 0x7CFC, 0x1080, 0xFEFE, 0x2080, 0x3EFE, 0x4202, 0x54AA, 0x88AA, 0x1502, 0x620C}},
    {40355, 2, {0x1010, 0x820, 0x7E7C, 0x4244, 0x427C, 0x7E44, 0x427C, 0x4040, 0x7F7E, 0x5540, 0x557E, 0x7F02, 0xD52A, 0x552A, 0x5542, 0x430C}},
    {40356, 2, {0x20, 0xFE40, 0x92FC, 0x1084, 0xFEFC, 0x1084, 0x7CFC, 0x5480, 0x7CFE, 0x5480, 0x7CFE, 0x1002, 0xFEAA, 0x12AA, 0x1202, 0x100C}},
    {40357, 2, {0x10, 0x7F20, 0x497C, 0x4944, 0x7F7C, 0x4044, 0x5F7C, 0x5140, 0x517E, 0x5F40, 0x517E, 0x5F02, 0x512A, 0x512A, 0x9F42, 0x110C}},
    {40358, 2, {0x107C, 0xFF44, 0x107C, 0x7E44, 0x427C, 0x7E44, 0x28C, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40359, 2, {0x2210, 0x2220, 0x7F7C, 0x2244, 0x2A7C, 0x844, 0x3E7C, 0x2A40, 0x2A7E, 0x2A40, 0xFF7E, 0x802, 0x142A, 0x122A, 0x2242, 0x400C}},
    {40360, 2, {0x2020, 0x3C40, 0x44FC, 0x7884, 0x8FC, 0xFE84, 0x20FC, 0x5280, 0x9CFE, 0x2C80, 0x4AFE, 0x9802, 0x2AAA, 0x4AAA, 0xAA02, 0x100C}},
    {40361, 2, {0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x98C6, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40362, 2, {0x20, 0x7E40, 0x42FC, 0x7E84, 0x42FC, 0x7E84, 0x42FC, 0x7E80, 0x14FE, 0x1280, 0xFEFE, 0x1002, 0x2AAA, 0x26AA, 0x4202, 0x800C}},
    {40363, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0xFEFC, 0x9284, 0xD6FC, 0xBA80, 0x92FE, 0xFE80, 0x10FE, 0x3802, 0x54AA, 0x92AA, 0x1102, 0x100C}},
    {40364, 2, {0x810, 0xC20, 0x127C, 0x2944, 0x44FC, 0xBF44, 0x217C, 0x3F40, 0x217E, 0x3F40, 0x207E, 0x3F02, 0x512A, 0x512A, 0x9F42, 0x110C}},
    {40365, 2, {0x2020, 0x4410, 0xF9FE, 0x1040, 0x247C, 0xFE84, 0x84, 0x5514, 0x8208, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40366, 2, {0x2020, 0x1040, 0xFEFC, 0x84, 0x7CFC, 0x4484, 0x7CFC, 0x80, 0xFEFE, 0x8280, 0xBAFE, 0xAA02, 0xBAAA, 0x82AA, 0x8B02, 0x840C}},
    {40367, 2, {0x820, 0x2AA4, 0x4D28, 0x1450, 0x2288, 0x7FFE, 0x4202, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFC, 0x4, 0x492C}},
    {40368, 2, {0x20, 0x7C40, 0x44FC, 0x4484, 0x7CFC, 0x84, 0xFEFC, 0x8280, 0xFEFE, 0x8280, 0xFEFE, 0x8202, 0xFEAA, 0x28AA, 0x4502, 0x840C}},
    {40369, 2, {0x100, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x3FF8, 0x440, 0xFFFE, 0xA20, 0x1FF0, 0x2828, 0xCFE6, 0x820, 0xFF8, 0x8, 0x1558}},
    {40370, 2, {0x2820, 0x4440, 0x92FC, 0x2084, 0x48FC, 0xFC84, 0x4FC, 0x80, 0xEEFE, 0x2280, 0xAAFE, 0x6602, 0xAAAA, 0x22AA, 0xAB02, 0x440C}},
    {40371, 2, {0x2100, 0x427E, 0xF388, 0x9288, 0xF2BE, 0x92AA, 0xF3AA, 0x822A, 0xFA2A, 0x83AA, 0xFAAA, 0xAAE, 0xAA88, 0xAB88, 0x8A88, 0x3008}},
    {40372, 2, {0x810, 0x820, 0x7F7C, 0x5544, 0x127C, 0x3F44, 0x647C, 0xA440, 0x3F7E, 0x2440, 0x247E, 0x3F02, 0x24AA, 0x24AA, 0x3F02, 0x200C}},
    {40373, 2, {0x2010, 0x2020, 0x7F7C, 0x9144, 0x557C, 0x7D44, 0x217C, 0x4A40, 0xA47E, 0x7F40, 0x917E, 0x5502, 0x7D2A, 0x212A, 0x4A42, 0x840C}},
    {40374, 2, {0x1010, 0x820, 0x7F7C, 0x4844, 0x5E7C, 0x4A44, 0x7F7C, 0x4A40, 0x5E7E, 0x4840, 0x5E7E, 0x5202, 0x52AA, 0x52AA, 0x5F02, 0x920C}},
    {40375, 2, {0x2020, 0x1040, 0xFEFC, 0x8284, 0x10FC, 0x7C84, 0x10FC, 0x7C80, 0x10FE, 0xFE80, 0x10FE, 0x7C02, 0x44AA, 0x44AA, 0x7D02, 0x440C}},
    {40376, 2, {0x20, 0xEE40, 0x22FC, 0x2284, 0xEEFC, 0x8884, 0x88FC, 0xEE80, 0x22FE, 0xAA80, 0x66FE, 0x6602, 0xAAAA, 0x22AA, 0xAB02, 0x440C}},
    {40377, 2, {0x2020, 0xDE40, 0x8AFC, 0xCA84, 0xAAFC, 0xD284, 0xA6FC, 0x80, 0xFEFE, 0x9280, 0x92FE, 0xFE02, 0x92AA, 0x92AA, 0xFF02, 0x820C}},
    {40378, 2, {0x1020, 0x9240, 0x54FC, 0xAA84, 0x44FC, 0x8284, 0x7CFC, 0x4480, 0x44FE, 0x7C80, 0x44FE, 0x4402, 0x7CAA, 0x44AA, 0x4502, 0x4C0C}},
    {40379, 2, {0x20, 0x7C40, 0x44FC, 0x7484, 0x54FC, 0xFE84, 0x82FC, 0x7C80, 0x44FE, 0x7C80, 0x44FE, 0x7C02, 0x44AA, 0x44AA, 0x5502, 0x480C}},
    {40380, 2, {0x2210, 0x1420, 0x7C, 0x7FC4, 0x147C, 0x7F44, 0x157C, 0xFFC0, 0x157E, 0x7F40, 0x147E, 0x3602, 0x552A, 0x94AA, 0x1442, 0x140C}},
    {40381, 2, {0x2820, 0x2440, 0x7EFC, 0xC884, 0x7EFC, 0x4884, 0x7EFC, 0x4880, 0x7EFE, 0x5080, 0x10FE, 0xFE02, 0x12AA, 0x12AA, 0x1202, 0x100C}},
    {40382, 2, {0x1080, 0x10FE, 0x1140, 0xFEFC, 0x1084, 0x7CFC, 0x4484, 0x7CFC, 0x4480, 0x7CFE, 0x1080, 0xFEFE, 0x1002, 0x12AA, 0x12AA, 0x1004}},
    {40383, 2, {0x820, 0x440, 0x7FFC, 0x1248, 0x3CF0, 0x820, 0x1450, 0x3EF8, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40384, 2, {0x2108, 0x1110, 0x123C, 0xFFE4, 0x213C, 0x2124, 0x423C, 0x52A0, 0x94BE, 0xE720, 0x213E, 0x4202, 0x422A, 0x94AA, 0xF782, 0x108C}},
    {40385, 2, {0x2210, 0x1420, 0x7C, 0x7F44, 0x7C, 0x1444, 0x227C, 0x4140, 0x7E, 0x7F40, 0x557E, 0x5502, 0x552A, 0x57AA, 0xFC42, 0xC}},
    {40386, 2, {0xE20, 0xF040, 0x22FC, 0x9284, 0x44FC, 0x2084, 0x3CFC, 0x5080, 0x10FE, 0xFE80, 0x10FE, 0x5402, 0x54AA, 0x5CAA, 0x6502, 0xC}},
    {40387, 2, {0x10, 0x7E20, 0x427C, 0x7E44, 0x427C, 0x7E44, 0x87C, 0xFF40, 0x817E, 0x1040, 0xFF7E, 0x2202, 0x64AA, 0x18AA, 0x2502, 0xC20C}},
    {40388, 2, {0xE20, 0xF040, 0x22FC, 0x9284, 0x44FC, 0x2084, 0x44FC, 0xF880, 0x10FE, 0x2480, 0xFEFE, 0x1002, 0xFEAA, 0x28AA, 0x4502, 0x820C}},
    {40389, 2, {0x20, 0xFE40, 0x28FC, 0x2884, 0xFEFC, 0xAA84, 0xAAFC, 0xFE80, 0x10FE, 0x1080, 0xFEFE, 0x3802, 0x54AA, 0x92AA, 0x1102, 0x100C}},
    {40390, 2, {0x4610, 0x7820, 0x40FC, 0x3FC4, 0x7C, 0x5F44, 0x517C, 0x5F40, 0x517E, 0x5F40, 0x517E, 0x5F02, 0x402A, 0x7FAA, 0x2242, 0x410C}},
    {40391, 2, {0x1000, 0xFE78, 0x1048, 0x7C48, 0x48, 0xFE86, 0x9200, 0x7CFC, 0x4444, 0x7C44, 0x4428, 0x7E28, 0x4010, 0x7F28, 0x144, 0xAB82}},
    {40392, 2, {0x2004, 0x401E, 0xF3F0, 0x921E, 0xF210, 0x92FE, 0xF292, 0x8298, 0xFAF2, 0x828E, 0xFA80, 0xAB8, 0xAAA8, 0xACAA, 0x8D4A, 0x3A86}},
    {40393, 2, {0x310, 0x7C20, 0x477C, 0x4444, 0x5F7C, 0x5544, 0x567C, 0x5C40, 0x557E, 0x5340, 0x507E, 0x5702, 0x552A, 0x552A, 0xA9C2, 0xC}},
    {40394, 2, {0x20, 0xFE40, 0xFC, 0x7C84, 0x44FC, 0x7C84, 0xFC, 0xFE80, 0x82FE, 0xAA80, 0x92FE, 0xFE02, 0x92AA, 0x92AA, 0x9302, 0x860C}},
    {40395, 2, {0x2208, 0x2210, 0xFFBC, 0x2224, 0x2A3C, 0x1424, 0x223C, 0x5D20, 0x88BE, 0x820, 0x7F3E, 0x802, 0x2A2A, 0x492A, 0xA882, 0x100C}},
    {40396, 2, {0x1000, 0x20FE, 0x7C90, 0x44FC, 0x7C90, 0x44FC, 0x7C90, 0x4090, 0x7EFE, 0x4002, 0x7EAA, 0x2AA, 0xAAAA, 0xAA82, 0x8214, 0xC08}},
    {40397, 2, {0x1040, 0x2080, 0x7DFC, 0x4504, 0x7DFC, 0x4504, 0x7DFC, 0x4500, 0x7DFE, 0x1100, 0xFDFE, 0x1002, 0x3AAA, 0x56AA, 0x9202, 0x100C}},
    {40398, 2, {0x1020, 0x2040, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x1080, 0xFEFE, 0x2880, 0x54FE, 0x9202, 0x7CAA, 0x10AA, 0x1102, 0x100C}},
    {40399, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x4484, 0x7CFC, 0x4480, 0x7CFE, 0x4480, 0x7CFE, 0x4402, 0xFEAA, 0x28AA, 0x4502, 0x820C}},
    {40400, 2, {0x10, 0x7E20, 0x427C, 0x7E44, 0x427C, 0x7E44, 0x7C, 0x7F40, 0x407E, 0x5E40, 0x407E, 0x7F02, 0x552A, 0x522A, 0x5942, 0x910C}},
    {40401, 2, {0x4410, 0x4420, 0x777C, 0xAA44, 0x117C, 0x2044, 0x107C, 0xFF40, 0x7E, 0x440, 0x447E, 0x2402, 0x28AA, 0xEAA, 0xF102, 0x400C}},
    {40402, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x5484, 0x7CFC, 0x5480, 0x7CFE, 0x1280, 0xFEFE, 0x802, 0xFEAA, 0x48AA, 0x2902, 0x180C}},
    {40403, 2, {0x1010, 0x820, 0x7F7C, 0x4044, 0x527C, 0x5244, 0x7F7C, 0x5240, 0x527E, 0x5E40, 0x407E, 0x5502, 0x552A, 0x552A, 0x8042, 0xC}},
    {40404, 2, {0x820, 0x7F20, 0x87E, 0x7E44, 0x8A4, 0xFF28, 0x1010, 0x1E28, 0x2244, 0x5FF2, 0x9010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40405, 2, {0x90, 0x79FE, 0x4B20, 0x4DFC, 0x4920, 0x49FC, 0x7920, 0x1FE, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40406, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40407, 2, {0x20, 0xFE40, 0x80FC, 0x8084, 0xBCFC, 0xA484, 0xA4FC, 0xBC80, 0x80FE, 0xEE80, 0xAAFE, 0xAA02, 0xEEAA, 0x80AA, 0xFF02, 0xC}},
    {40408, 2, {0x820, 0xFFA0, 0x83E, 0x7F44, 0x49A8, 0x7F10, 0x2A28, 0x4946, 0x8200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40409, 2, {0x820, 0x7E20, 0x8F8, 0xFF28, 0x1428, 0x7F6A, 0x82A, 0xFF56, 0xA82, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40410, 2, {0x20, 0xEE40, 0x22FC, 0xAA84, 0x66FC, 0xAA84, 0x10FC, 0x2880, 0x44FE, 0x9280, 0x20FE, 0xC802, 0x32AA, 0xC4AA, 0x1902, 0xE00C}},
    {40411, 2, {0x208, 0x3FD0, 0x223C, 0x2FA4, 0x22BC, 0x3FE4, 0x22BC, 0x2FA0, 0x223E, 0x2FA0, 0x2ABE, 0x2F82, 0x4AAA, 0x4FAA, 0x8A82, 0x98C}},
    {40412, 2, {0x810, 0x7F20, 0x497C, 0xFFC4, 0x497C, 0x7F44, 0x87C, 0x7F40, 0x497E, 0x7F40, 0x107E, 0xFF02, 0x212A, 0x722A, 0xC42, 0xF30C}},
    {40413, 2, {0x20, 0xFE40, 0x92FC, 0xFE84, 0x92FC, 0xFE84, 0x54FC, 0x5480, 0xFEFE, 0x5480, 0x54FE, 0xFE02, 0x10AA, 0xFEAA, 0x1102, 0x100C}},
    {40414, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0xBAFC, 0x5484, 0xBAFC, 0x1080, 0xBAFE, 0x5480, 0xBAFE, 0x1002, 0x2AAA, 0x26AA, 0x4202, 0x800C}},
    {40415, 2, {0x2080, 0x10FC, 0xFD40, 0x20F8, 0x3D20, 0x24FC, 0x4450, 0x5588, 0x8A04, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40416, 2, {0x2040, 0x2080, 0x79FC, 0x4904, 0x91FC, 0x7D04, 0x55FC, 0x5500, 0x7DFE, 0x5500, 0x55FE, 0x7C02, 0x2AA, 0xAAAA, 0xAA02, 0xC}},
    {40417, 2, {0x2008, 0x2010, 0x7FBC, 0xD524, 0x553C, 0x5524, 0xFFBC, 0x5520, 0x553E, 0x5520, 0xFFBE, 0x2, 0x6A2A, 0x552A, 0x9502, 0xC}},
    {40418, 2, {0x3FFE, 0x2890, 0x2510, 0x3FDE, 0x2222, 0x2A88, 0x2F88, 0x2294, 0x2522, 0x2FF0, 0x2410, 0x27F0, 0x2410, 0x47FC, 0x4004, 0x8AAC}},
    {40419, 2, {0x20, 0xFE40, 0x28FC, 0xFE84, 0xAAFC, 0xFE84, 0xFC, 0x7C80, 0x44FE, 0x7C80, 0x44FE, 0x7C02, 0x10AA, 0xFEAA, 0x1102, 0x100C}},
    {40420, 2, {0x20, 0xEE40, 0xAAFC, 0xAA84, 0xEEFC, 0x84, 0x7CFC, 0x5480, 0x7CFE, 0x5480, 0x7CFE, 0x1002, 0xFEAA, 0x12AA, 0x1202, 0x100C}},
    {40421, 2, {0x1020, 0x2448, 0x78F0, 0x1428, 0x7EFC, 0x20, 0x54A8, 0x5564, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40422, 2, {0x2820, 0x2440, 0x7EFC, 0xC884, 0x48FC, 0x7E84, 0x48FC, 0x4880, 0x7EFE, 0x4880, 0x48FE, 0x7E02, 0x40AA, 0xAAAA, 0xAB02, 0xC}},
    {40423, 2, {0x1020, 0xFE40, 0x10FC, 0x7C84, 0xFC, 0xFE84, 0x82FC, 0x7C80, 0xFE, 0x7C80, 0x44FE, 0x7C02, 0x44AA, 0x28AA, 0xF02, 0xF00C}},
    {40424, 2, {0x4420, 0x4440, 0xFEFC, 0x4484, 0xFC, 0xFE84, 0x54FC, 0x5480, 0xFEFE, 0x5480, 0x54FE, 0xFE02, 0x10AA, 0xFEAA, 0x1102, 0x100C}},
    {40425, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x4146, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40426, 2, {0x0, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x7FFE, 0x4202, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40427, 2, {0x808, 0x7F10, 0x93C, 0xFFA4, 0x93C, 0x7F24, 0x83C, 0xAAA0, 0xEBBE, 0x88A0, 0xFFBE, 0x8882, 0xEBAA, 0xAAAA, 0xAA82, 0x88C}},
    {40428, 2, {0x2820, 0x2840, 0x7CFC, 0x2884, 0x28FC, 0xFE84, 0x10FC, 0x7C80, 0x54FE, 0x7C80, 0x54FE, 0x7C02, 0xAA, 0x28AA, 0x4502, 0x820C}},
    {40429, 2, {0xE20, 0xF040, 0x92FC, 0x5484, 0xFEFC, 0x3884, 0x54FC, 0x8280, 0x7CFE, 0x5480, 0x54FE, 0x7C02, 0x54AA, 0x54AA, 0x7D02, 0x440C}},
    {40430, 2, {0x610, 0x7820, 0x107C, 0xFFC4, 0x227C, 0x5D44, 0x94FC, 0x1C40, 0x7E, 0x7F40, 0x417E, 0x5D02, 0x552A, 0x5D2A, 0x4142, 0x430C}},
    {40431, 2, {0x810, 0x820, 0x7F7C, 0x844, 0x557C, 0x2244, 0x417C, 0xBEC0, 0x227E, 0x3E40, 0x227E, 0x3E02, 0x8AA, 0x2AAA, 0x4902, 0x180C}},
    {40432, 2, {0x440, 0xFFFE, 0x440, 0x17D2, 0xF01C, 0x17D0, 0x3452, 0xD7CE, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40433, 2, {0x2020, 0xFE40, 0x82FC, 0xFE84, 0x82FC, 0xFE84, 0x82FC, 0xFE80, 0x10FE, 0x9280, 0x54FE, 0x9202, 0x10AA, 0xFEAA, 0x1102, 0x100C}},
    {40434, 2, {0x828, 0x7F24, 0xFE, 0x3E20, 0x2250, 0x3E50, 0x4952, 0x988E, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40435, 2, {0xEE20, 0xAA40, 0xEEFC, 0xAA84, 0xEEFC, 0x8284, 0xBAFC, 0xAA80, 0xBAFE, 0xAA80, 0xBAFE, 0xAA02, 0xAAAA, 0xDAAA, 0x8302, 0x860C}},
    {40436, 2, {0x20, 0xEE40, 0xAAFC, 0xEE84, 0xAAFC, 0xEE84, 0x92FC, 0x9280, 0xFEFE, 0x9280, 0xBAFE, 0xD602, 0x92AA, 0x92AA, 0x8B02, 0x840C}},
    {40437, 2, {0x10, 0x7F20, 0x417C, 0x4144, 0x7F7C, 0x4844, 0x5D7C, 0x4A40, 0x7F7E, 0x4840, 0x5E7E, 0x7202, 0x5EAA, 0x52AA, 0x5F02, 0x920C}},
    {40438, 2, {0x20, 0xFE40, 0xAAFC, 0xAA84, 0xFEFC, 0x84, 0x7CFC, 0x4480, 0x7CFE, 0x4480, 0x7CFE, 0x4402, 0x7CAA, 0xAA, 0x2902, 0x440C}},
    {40439, 2, {0x2210, 0x1420, 0xFF7C, 0x1444, 0x7F7C, 0x5544, 0x637C, 0x5D40, 0x417E, 0x7F40, 0x47E, 0xFF02, 0x44AA, 0x24AA, 0x2502, 0xC0C}},
    {40440, 2, {0x7E10, 0x1420, 0x87C, 0xFF44, 0x297C, 0x4A44, 0xA87C, 0x1040, 0x7F7E, 0x5540, 0x637E, 0x5D02, 0x552A, 0x5D2A, 0x4142, 0x430C}},
    {40441, 2, {0x80, 0x3FFE, 0x2250, 0x24FC, 0x2D90, 0x36FC, 0x2490, 0x24FC, 0x2100, 0x27F0, 0x2410, 0x27F0, 0x4410, 0x47FC, 0x8004, 0xAAC}},
    {40442, 2, {0x40, 0x3EFC, 0x2348, 0x3E30, 0x9CE, 0x2E78, 0x2848, 0x7E78, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40443, 2, {0x2108, 0x4088, 0xF3C8, 0x9010, 0xF3DE, 0x9264, 0xF3D4, 0x8014, 0xFBD4, 0x8054, 0xF894, 0x8E8, 0xAB88, 0xA894, 0x8A94, 0x3122}},
    {40444, 2, {0x20, 0xEE40, 0xAAFC, 0xEE84, 0xAAFC, 0xEE84, 0x82FC, 0xBA80, 0xAAFE, 0xBA80, 0xAAFE, 0xBA02, 0xAAAA, 0x82AA, 0x8B02, 0x840C}},
    {40445, 2, {0xA80, 0x3138, 0x2288, 0x3838, 0x2288, 0x3938, 0x2288, 0x7FFE, 0x4202, 0x9FF4, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40446, 2, {0x1020, 0xFE40, 0x44FC, 0x2884, 0xFEFC, 0x84, 0x7CFC, 0x4480, 0x7CFE, 0x4480, 0x7CFE, 0x1002, 0xAAA, 0xA9AA, 0xA502, 0x1C0C}},
    {40447, 2, {0x3E10, 0x22FE, 0x3E44, 0x2028, 0x7EFE, 0xA210, 0x3EFC, 0x2210, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40448, 2, {0x10, 0x7F20, 0x557C, 0x5544, 0x7F7C, 0x2044, 0x7F7C, 0x9140, 0x7D7E, 0x5540, 0x557E, 0x7D02, 0x112A, 0x7D2A, 0x542, 0x20C}},
    {40449, 2, {0x100, 0xFFFE, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x2008, 0x3FF8, 0x200, 0x779C, 0x5494, 0x7794, 0x549C, 0x77D6, 0x5056, 0xB562}},
    {40450, 2, {0x310, 0xBC20, 0x44FC, 0x1144, 0x847C, 0x4944, 0x5E7C, 0x440, 0x297E, 0x5F40, 0xC47E, 0x5F02, 0x44AA, 0x4AAA, 0x5102, 0x208C}},
    {40451, 2, {0x2104, 0x4088, 0xF3FE, 0x9020, 0xF1FC, 0x9020, 0xF3FE, 0x8054, 0xF992, 0x8090, 0xFBFE, 0x890, 0xA8D4, 0xAB8A, 0x8896, 0x31A2}},
    {40452, 2, {0x1020, 0x7E20, 0x4A7E, 0x7EC4, 0x1028, 0xFF10, 0x2028, 0x3EC6, 0x4200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40453, 2, {0x10, 0x7F20, 0x557C, 0x7F44, 0x87C, 0x7F44, 0x87C, 0xFFC0, 0x227E, 0x1440, 0x7F7E, 0x802, 0xFFAA, 0x8AA, 0x902, 0x80C}},
    {40454, 2, {0x810, 0xF20, 0x87C, 0x7F44, 0x497C, 0x7C44, 0x497C, 0x5F40, 0x527E, 0x5E40, 0x407E, 0x5E02, 0x482A, 0x7F2A, 0x5242, 0xA10C}},
    {40455, 2, {0x1020, 0xFE40, 0xFC, 0xFE84, 0x82FC, 0xBA84, 0xAAFC, 0xFE80, 0xFE, 0x7C80, 0x44FE, 0x7C02, 0x44AA, 0x7CAA, 0x102, 0xFE0C}},
    {40456, 2, {0x1410, 0x5520, 0x367C, 0x1444, 0xFF7C, 0x2244, 0x147C, 0x7F40, 0x87E, 0x7F40, 0x87E, 0xFF02, 0x1CAA, 0x2AAA, 0x4902, 0x80C}},
    {40457, 2, {0x11FC, 0x1104, 0xFDFC, 0x3104, 0x59FC, 0x9480, 0x11FE, 0x1252, 0x2AA, 0x1FF4, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40458, 2, {0x208, 0x7910, 0x493C, 0x4BA4, 0x483C, 0x7AA4, 0x413C, 0x47E0, 0x793E, 0x6920, 0xABBE, 0xA902, 0x292A, 0x392A, 0x2902, 0x10C}},
    {40459, 2, {0x810, 0xFF20, 0x917C, 0x4A44, 0xA57C, 0x1C44, 0x7C, 0x7F40, 0x557E, 0xFFC0, 0x7E, 0x7F02, 0x8AA, 0x8AA, 0x2902, 0x100C}},
    {40460, 2, {0x1044, 0x21FE, 0x7C54, 0x4448, 0x7CFE, 0x4590, 0x7CFE, 0x4090, 0x7EFE, 0x4090, 0x7EFE, 0x280, 0xABFE, 0xAA44, 0x8238, 0xDC6}},
    {40461, 2, {0x8, 0xFF90, 0x493C, 0x88A4, 0xFFFC, 0x88A4, 0xDDBC, 0xAAA0, 0xDDBE, 0x88A0, 0xDDBE, 0xAA82, 0xDDAA, 0x88AA, 0x8A82, 0x810C}},
    {40462, 2, {0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x1450, 0x2288, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40463, 2, {0x2410, 0xFF20, 0x247C, 0xFF44, 0x817C, 0x7E44, 0x7C, 0xFF40, 0x107E, 0x28C0, 0xD57E, 0x2E02, 0xD5AA, 0x24AA, 0xD502, 0x80C}},
    {40464, 2, {0x10, 0xEE20, 0x227C, 0xAA44, 0x667C, 0xAA44, 0x7C, 0x2440, 0x7F7E, 0xC840, 0x7E7E, 0x4802, 0x7EAA, 0x48AA, 0x7F02, 0x400C}},
    {40465, 2, {0x510C, 0x278A, 0x5008, 0x97BE, 0x3008, 0x5788, 0x9494, 0x57A4, 0x2202, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40466, 2, {0x1A38, 0x2388, 0x3A38, 0x2388, 0x38B8, 0x2288, 0xFFFE, 0x1010, 0x2208, 0xDFF6, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40467, 2, {0x8, 0x7F10, 0x493C, 0x7F24, 0x493C, 0x7F24, 0x413C, 0x20, 0xFFBE, 0xAAA0, 0xAABE, 0xFF82, 0xAAAA, 0xAAAA, 0xFF82, 0x888C}},
    {40468, 2, {0x21FC, 0x4104, 0xF9FC, 0x8904, 0xF9FC, 0x8888, 0xF9FC, 0x8088, 0xFDFE, 0x8088, 0xFD24, 0x6AA, 0x5470, 0x54A8, 0x8524, 0x1860}},
    {40469, 2, {0x808, 0xF90, 0x83C, 0x7FA4, 0x48BC, 0x7E24, 0x48BC, 0x5FA0, 0x553E, 0x5F20, 0x553E, 0x5F02, 0x402A, 0xBFAA, 0x2A82, 0x7FCC}},
    {40470, 2, {0x7F08, 0x810, 0xFFBC, 0x88A4, 0x6B3C, 0x824, 0x6B3C, 0x2420, 0x7FBE, 0xC820, 0x7F3E, 0x4802, 0x7F2A, 0x482A, 0x7F82, 0x400C}},
    {40471, 2, {0x840, 0x3E7C, 0x1440, 0xFF78, 0x2208, 0x3E78, 0x2242, 0x263E, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40472, 2, {0xFE20, 0x1040, 0xFEFC, 0x9284, 0x54FC, 0x1084, 0x54FC, 0x80, 0x2EFE, 0xFA80, 0x2EFE, 0x7A02, 0xAEAA, 0x2AAA, 0x2F02, 0x2A0C}},
    {40473, 2, {0x1020, 0x2840, 0x44FC, 0xBA84, 0xFC, 0xFE84, 0xAAFC, 0xFE80, 0xFE, 0xFE80, 0xAAFE, 0xAA02, 0xFEAA, 0xAAAA, 0xAB02, 0x860C}},
    {40474, 2, {0x10, 0x7720, 0x557C, 0x7744, 0x557C, 0x7744, 0x557C, 0x7740, 0x227E, 0x5540, 0x107E, 0xFF02, 0x22AA, 0x64AA, 0x1D02, 0xE30C}},
    {40475, 2, {0x4420, 0xFE40, 0x44FC, 0x84, 0xEEFC, 0xAA84, 0xEEFC, 0x2480, 0x7EFE, 0xC880, 0x7EFE, 0x4802, 0x7EAA, 0x48AA, 0x7F02, 0x400C}},
    {40476, 2, {0x7710, 0x5520, 0x777C, 0x5544, 0x777C, 0x5544, 0x777C, 0x2240, 0x7FFE, 0xA440, 0x3F7E, 0x2402, 0x3F2A, 0x242A, 0x3FC2, 0x200C}},
    {40477, 2, {0x7710, 0x20, 0x777C, 0x5544, 0x777C, 0x5544, 0x87C, 0x7F40, 0x547E, 0x7F40, 0x557E, 0x7F02, 0x552A, 0x5E2A, 0x5542, 0x9B0C}},
    {40478, 2, {0x2108, 0x4FD2, 0xF03C, 0x2388, 0x4812, 0xFBBE, 0x280, 0xABAA, 0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x5244, 0x892C}},
    {40479, 2, {0x100, 0x200, 0x1FF0, 0x1010, 0x1210, 0x1110, 0x1150, 0x1020, 0x1000, 0x1FFC, 0x4, 0x4, 0x7FE4, 0x4, 0x28, 0x10}},
    {40480, 2, {0x2020, 0x2040, 0x20FC, 0x2084, 0xFCA4, 0x2484, 0x2494, 0x2488, 0x2480, 0x24FE, 0x2502, 0x2602, 0x45FA, 0x4002, 0x8014, 0x8}},
    {40481, 2, {0x20, 0x40, 0xFC, 0xFC84, 0x4A4, 0x484, 0x4894, 0x2888, 0x1080, 0x10FE, 0x2802, 0x2402, 0x45FA, 0x8002, 0x14, 0x8}},
    {40482, 2, {0x420, 0x410, 0x23E, 0xFFC0, 0x80, 0x42, 0x232, 0x1FEE, 0x1420, 0x1260, 0x1000, 0x1FF8, 0x8, 0x7F88, 0x28, 0x10}},
    {40483, 2, {0x20, 0x40, 0x79FC, 0x4904, 0x4944, 0x4924, 0x4924, 0x490C, 0x4900, 0x49FE, 0x7802, 0x4802, 0x3FA, 0x2, 0x14, 0x8}},
    {40484, 2, {0x20, 0x40, 0x3EFC, 0x2284, 0x22A4, 0x2284, 0x3E94, 0x2288, 0x2080, 0x20FE, 0x2002, 0x2002, 0x21FA, 0x4002, 0x4014, 0x8008}},
    {40485, 2, {0x20, 0x7E40, 0x40FC, 0x4484, 0x64A4, 0x5484, 0x4894, 0x4888, 0x5480, 0x54FE, 0x6402, 0x4002, 0x41FA, 0x7E02, 0x14, 0x8}},
    {40486, 2, {0x20, 0x7E40, 0x4FC, 0x2484, 0x24A4, 0x2484, 0x4494, 0x7F88, 0xC80, 0x14FE, 0x1402, 0x2402, 0x45FA, 0x8402, 0x1414, 0x808}},
    {40487, 2, {0x810, 0x820, 0x147C, 0x1244, 0x2164, 0x4054, 0xBE44, 0x224C, 0x2240, 0x227E, 0x2A02, 0x2402, 0x217A, 0x2102, 0x1F0A, 0x4}},
    {40488, 2, {0x4020, 0x4840, 0x50FC, 0x6084, 0x44A4, 0x4484, 0x3C94, 0x1088, 0x1080, 0x10FE, 0xFE02, 0x1002, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {40489, 2, {0x810, 0x820, 0x87C, 0x7F44, 0x4964, 0x854, 0x844, 0x144C, 0x1440, 0x147E, 0x1402, 0x2502, 0x267A, 0x4402, 0x800A, 0x4}},
    {40490, 2, {0x1020, 0x1040, 0x10FC, 0x1084, 0xFEA4, 0x1084, 0x1094, 0x1088, 0x7C80, 0x44FE, 0x4402, 0x4402, 0x45FA, 0x7C02, 0x4414, 0x8}},
    {40491, 2, {0x1010, 0x1020, 0x107C, 0xFF44, 0x2064, 0x2854, 0x4844, 0x484C, 0x7E40, 0x87E, 0x2C02, 0x2A02, 0x497A, 0x8902, 0x280A, 0x1004}},
    {40492, 2, {0x1020, 0x1040, 0x1EFC, 0x1084, 0x10A4, 0x7C84, 0x4494, 0x4488, 0x4480, 0x7CFE, 0x4402, 0x4002, 0x41FA, 0x4002, 0x8014, 0x8}},
    {40493, 2, {0x20, 0xFE40, 0x92FC, 0x9284, 0x92A4, 0xFE84, 0x9294, 0x9288, 0x9280, 0xFEFE, 0x9202, 0x1002, 0x11FA, 0x1002, 0x1014, 0x1008}},
    {40494, 2, {0x20, 0x7E40, 0x42FC, 0x4284, 0x42A4, 0x7E84, 0x94, 0xFF88, 0x2080, 0x40FE, 0x7E02, 0x202, 0x3FA, 0x202, 0x1414, 0x808}},
    {40495, 2, {0x100, 0x1FF0, 0x1110, 0x1110, 0xFFFE, 0x280, 0x1C70, 0xE20E, 0x400, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40496, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x42A4, 0x9184, 0x894, 0x888, 0xFE80, 0x2FE, 0x4402, 0x2802, 0x11FA, 0x802, 0x814, 0x8}},
    {40497, 2, {0x110, 0x7A0, 0x7C7C, 0x4444, 0x4464, 0x4454, 0x4444, 0x7F4C, 0x4440, 0x447E, 0x4402, 0x4202, 0x42FA, 0x5282, 0x698A, 0x4484}},
    {40498, 2, {0x2020, 0x2040, 0x40FC, 0x7E84, 0x82A4, 0x284, 0x7A94, 0x4A88, 0x4A80, 0x4AFE, 0x7A02, 0x4A02, 0x3FA, 0x202, 0x1414, 0x808}},
    {40499, 2, {0x1000, 0x3EF8, 0x4288, 0xA2AA, 0x1492, 0x187E, 0xE200, 0x1FE0, 0x1420, 0x1260, 0x1000, 0x1FF8, 0x8, 0x7F88, 0x28, 0x10}},
    {40500, 2, {0x2208, 0x1110, 0x20, 0x7FFE, 0x4102, 0x8204, 0x1FF0, 0x1010, 0x1210, 0x1150, 0x1020, 0x1FFC, 0x4, 0x4, 0x7FD4, 0x8}},
    {40501, 2, {0x1020, 0x2010, 0x7C10, 0x44FE, 0x5482, 0x4504, 0x5440, 0x4848, 0x4050, 0x7E60, 0x240, 0x242, 0xFA42, 0x242, 0x143E, 0x800}},
    {40502, 2, {0x1020, 0x2448, 0x78F0, 0x1020, 0x7EFC, 0x0, 0xFFFE, 0x200, 0x400, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40503, 2, {0x1080, 0x1080, 0xFBF0, 0x1090, 0x1990, 0xF094, 0x1154, 0x520C, 0x2404, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40504, 2, {0x10, 0x7F20, 0x87C, 0x844, 0x1064, 0x7F54, 0x5544, 0x554C, 0x5540, 0x557E, 0x5502, 0x5502, 0x557A, 0x5102, 0x450A, 0x4204}},
    {40505, 2, {0x420, 0xE40, 0x38FC, 0x884, 0x8A4, 0xFF84, 0x894, 0x888, 0x3E80, 0x22FE, 0x2202, 0x2202, 0x23FA, 0x3E02, 0x2214, 0x8}},
    {40506, 2, {0x1208, 0x1210, 0x123E, 0x2222, 0x2FB2, 0x622A, 0xA622, 0x2726, 0x2AA0, 0x2ABE, 0x3202, 0x2202, 0x227A, 0x2202, 0x220A, 0x2204}},
    {40507, 2, {0x1010, 0x1020, 0x277C, 0x4044, 0x9064, 0x1054, 0x2FC4, 0x624C, 0xA240, 0x227E, 0x2202, 0x2202, 0x227A, 0x2202, 0x2A0A, 0x2404}},
    {40508, 2, {0x820, 0x1040, 0x3EFC, 0x2284, 0x32A4, 0x2A84, 0x2A94, 0xFE88, 0x2280, 0x32FE, 0x2A02, 0x2A02, 0x23FA, 0x4202, 0x4A14, 0x8408}},
    {40509, 2, {0x1020, 0x1040, 0x28FC, 0x2484, 0x42A4, 0x8084, 0x7C94, 0x88, 0x80, 0x7CFE, 0x4402, 0x4402, 0x45FA, 0x7C02, 0x4414, 0x8}},
    {40510, 2, {0x200, 0x100, 0xFFFE, 0x440, 0x1450, 0x2448, 0x4204, 0x1FE0, 0x1420, 0x1260, 0x1000, 0x1FF8, 0x8, 0x7F88, 0x28, 0x10}},
    {40511, 2, {0x10, 0x8020, 0x407C, 0x1F44, 0x8464, 0x4454, 0x4444, 0x44C, 0x2440, 0x447E, 0xC402, 0x4702, 0x5C7A, 0x4802, 0x400A, 0x4}},
    {40512, 2, {0x10, 0x20, 0x7F7C, 0x844, 0x864, 0x2A54, 0x2A44, 0x2A4C, 0x5D40, 0x497E, 0x8802, 0x802, 0xF7A, 0xF002, 0x400A, 0x4}},
    {40513, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x10A4, 0xFE84, 0x8294, 0x488, 0x7880, 0x8FE, 0x1002, 0x1E02, 0xF1FA, 0x1002, 0x5014, 0x2008}},
    {40514, 2, {0x8, 0xFF90, 0x3E, 0x22, 0xF7B2, 0x94AA, 0x94A2, 0x94A6, 0xD6A0, 0xB5BE, 0x9482, 0x9482, 0x94FA, 0x9482, 0x948A, 0xB584}},
    {40515, 2, {0x10, 0x3E20, 0x227C, 0x2244, 0x3E64, 0x54, 0x7F44, 0x414C, 0x4140, 0x7F7E, 0x4102, 0x4102, 0x7F7A, 0x4102, 0x450A, 0x4204}},
    {40516, 2, {0x810, 0x4820, 0x487C, 0x7F44, 0x8864, 0x854, 0x844, 0xFF4C, 0x40, 0x7E, 0x7E02, 0x4202, 0x427A, 0x4202, 0x7E0A, 0x4204}},
    {40517, 2, {0x410, 0x3620, 0xE57C, 0x2544, 0x2464, 0xFFD4, 0x2444, 0x254C, 0x2540, 0x367E, 0xE602, 0x2402, 0x2AFA, 0x3282, 0xA18A, 0x4084}},
    {40518, 2, {0x2820, 0x2440, 0x42FC, 0x9284, 0x10A4, 0x2884, 0x4494, 0x8288, 0x7C80, 0x44FE, 0x4402, 0x4402, 0x45FA, 0x7C02, 0x4414, 0x8}},
    {40519, 2, {0x4010, 0x2F20, 0x17C, 0x4944, 0x4964, 0x4954, 0x7F44, 0x494C, 0x5D40, 0x6B7E, 0x4902, 0x4902, 0x497A, 0x4102, 0x450A, 0x4204}},
    {40520, 2, {0x4420, 0x2440, 0x28FC, 0xFC84, 0x14A4, 0x1484, 0xFC94, 0x9088, 0x9080, 0xFEFE, 0x3202, 0x3202, 0x53FA, 0x9602, 0x1014, 0x1008}},
    {40521, 2, {0x410, 0x620, 0x757C, 0x544, 0x464, 0xFF54, 0x444, 0x144C, 0x1440, 0x5C7E, 0x5402, 0x5402, 0x557A, 0x5D02, 0xE30A, 0x4104}},
    {40522, 2, {0x2410, 0x2420, 0x247C, 0xFF44, 0x2464, 0x2454, 0xFFC4, 0x4C, 0x7E40, 0x427E, 0x4202, 0x7E02, 0x427A, 0x4202, 0x7E0A, 0x4204}},
    {40523, 2, {0x2208, 0x2210, 0xFFBE, 0x2222, 0x2232, 0x2A, 0x7F22, 0x4926, 0x4920, 0x493E, 0x7F02, 0x4902, 0x497A, 0x4902, 0x7F0A, 0x4104}},
    {40524, 2, {0x808, 0x810, 0xFFBE, 0x1422, 0x2232, 0x492A, 0x88A2, 0x7F26, 0x4920, 0x7F3E, 0x4902, 0x7F02, 0x87A, 0xA02, 0xC0A, 0x804}},
    {40525, 2, {0x10, 0x7E20, 0x427C, 0x4244, 0x7E64, 0x4254, 0x4244, 0x7E4C, 0x40, 0x447E, 0x7702, 0x4402, 0x447A, 0x5502, 0x660A, 0x4404}},
    {40526, 2, {0x1010, 0x2020, 0x7F7C, 0x4944, 0x4964, 0x7F54, 0x4944, 0x514C, 0x7F40, 0x247E, 0x4402, 0xFF82, 0x47A, 0x402, 0x40A, 0x404}},
    {40527, 2, {0x10, 0x7720, 0x557C, 0x5544, 0x5564, 0x7754, 0x5544, 0x554C, 0x5540, 0x777E, 0x5502, 0x5502, 0x557A, 0xB502, 0x890A, 0x1304}},
    {40528, 2, {0x1020, 0x1040, 0x3EFC, 0x2284, 0x44A4, 0x8084, 0x1094, 0x6688, 0x4280, 0x42FE, 0x6602, 0x4202, 0x43FA, 0x7E02, 0x4214, 0x8}},
    {40529, 2, {0x1010, 0x820, 0xFF7C, 0x44, 0x7E64, 0x4254, 0x7E44, 0x4C, 0x7E40, 0x47E, 0x802, 0xF02, 0xF87A, 0x802, 0x280A, 0x1004}},
    {40530, 2, {0x1010, 0x820, 0x7F7C, 0x4844, 0x5E64, 0x4A54, 0x4A44, 0x7F4C, 0x4A40, 0x4A7E, 0x5E02, 0x4802, 0x547A, 0x9202, 0x220A, 0x4004}},
    {40531, 2, {0x1010, 0x820, 0xFF7C, 0x8144, 0x4064, 0x7754, 0x5544, 0x554C, 0x5540, 0xB57E, 0x9702, 0x1402, 0x247A, 0x2502, 0x460A, 0x8404}},
    {40532, 2, {0x810, 0x820, 0x7F7C, 0x944, 0xFFE4, 0x954, 0x7F44, 0x84C, 0x4940, 0x497E, 0x5D02, 0x6B02, 0x497A, 0x4902, 0x890A, 0x904}},
    {40533, 2, {0x2008, 0x2790, 0x24BE, 0xFCA2, 0x24B2, 0x27AA, 0x24A2, 0x74A6, 0x54A0, 0x57BE, 0x5482, 0x5482, 0x74FA, 0x882, 0xA8A, 0x1104}},
    {40534, 2, {0x10, 0x7E20, 0x427C, 0x7E44, 0x4264, 0x7E54, 0x2044, 0x7F4C, 0x8940, 0x497E, 0x5502, 0x4102, 0x7D7A, 0x102, 0xA0A, 0x404}},
    {40535, 2, {0x10, 0x7720, 0x557C, 0x5544, 0x7764, 0x54, 0x7F44, 0x4C, 0xFFC0, 0x207E, 0x3F02, 0x102, 0x17A, 0x102, 0xA0A, 0x404}},
    {40536, 2, {0x20, 0x7C40, 0x44FC, 0x7484, 0x54A4, 0xFE94, 0x8294, 0x7C8C, 0x4480, 0x7CFE, 0x4402, 0x7C02, 0x45FA, 0x4402, 0x5414, 0x4808}},
    {40537, 2, {0x620, 0x3820, 0x8A4, 0x7EA8, 0x1920, 0x2C50, 0x4A88, 0x1FE4, 0x1420, 0x1260, 0x1000, 0x1FF8, 0x8, 0x7F88, 0x28, 0x10}},
    {40538, 2, {0x2108, 0x1110, 0x123E, 0xFFE2, 0x2132, 0x212A, 0x4222, 0x52A6, 0x94A0, 0xE73E, 0x2102, 0x4202, 0x427A, 0x9482, 0xF78A, 0x1084}},
    {40539, 2, {0x10, 0x7F20, 0x497C, 0x4944, 0x7F64, 0x4054, 0x5F44, 0x514C, 0x5140, 0x5F7E, 0x5102, 0x5F02, 0x517A, 0x5102, 0x9F0A, 0x1104}},
    {40540, 2, {0x20, 0x7E20, 0x247E, 0x18A4, 0xFF28, 0x2910, 0x4A28, 0x9AC6, 0x400, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40541, 2, {0x20, 0xFE40, 0xFC, 0x7C84, 0x44A4, 0x7C84, 0x94, 0xFE88, 0x8280, 0xAAFE, 0x9202, 0xFE02, 0x93FA, 0x9202, 0x9214, 0x8608}},
    {40542, 2, {0xE20, 0xF040, 0x22FC, 0x9284, 0x44A4, 0x2084, 0x3C94, 0x5088, 0x1080, 0xFEFE, 0x1002, 0x5402, 0x55FA, 0x5C02, 0x6414, 0x8}},
    {40543, 2, {0x2820, 0x4440, 0x92FC, 0x2084, 0x48A4, 0xFC84, 0x494, 0x88, 0xEE80, 0x22FE, 0xAA02, 0x6602, 0xABFA, 0x2202, 0xAA14, 0x4408}},
    {40544, 2, {0x2020, 0xDE40, 0x8AFC, 0xCA84, 0xAAA4, 0xD284, 0xA694, 0x88, 0xFE80, 0x92FE, 0x9202, 0xFE02, 0x93FA, 0x9202, 0xFE14, 0x8208}},
    {40545, 2, {0x1020, 0x9240, 0x54FC, 0xAA84, 0x44A4, 0x8284, 0x7C94, 0x4488, 0x4480, 0x7CFE, 0x4402, 0x4402, 0x7DFA, 0x4402, 0x4414, 0x4C08}},
    {40546, 2, {0x2210, 0x1420, 0x7C, 0x7F44, 0x64, 0x1454, 0x2244, 0x414C, 0x40, 0x7F7E, 0x5502, 0x5502, 0x557A, 0x5782, 0xFC0A, 0x4}},
    {40547, 2, {0x2210, 0x1420, 0x7C, 0x7FC4, 0x1464, 0x7F54, 0x1544, 0xFFCC, 0x1540, 0x7F7E, 0x1402, 0x3602, 0x557A, 0x9482, 0x140A, 0x1404}},
    {40548, 2, {0x810, 0x820, 0x7F7C, 0x5544, 0x1264, 0x3F54, 0x6444, 0xA44C, 0x3F40, 0x247E, 0x2402, 0x3F02, 0x247A, 0x2402, 0x3F0A, 0x2004}},
    {40549, 2, {0x7F78, 0x4848, 0x5F4C, 0x6480, 0x5F78, 0x4A48, 0x5130, 0x7F4C, 0x400, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40550, 2, {0x10, 0x7720, 0x557C, 0x5544, 0x5564, 0x5554, 0x2244, 0x554C, 0x840, 0xFF7E, 0x1202, 0x2202, 0x347A, 0xC02, 0x320A, 0xC104}},
    {40551, 2, {0x1010, 0x820, 0x7F7C, 0x4044, 0x5264, 0x5254, 0x7F44, 0x524C, 0x5240, 0x5E7E, 0x4002, 0x5502, 0x557A, 0x5502, 0x800A, 0x4}},
    {40552, 2, {0x10, 0x7720, 0x117C, 0x5544, 0x3364, 0x5554, 0xC44, 0x334C, 0xC4C0, 0x187E, 0x6202, 0xC82, 0x717A, 0x602, 0x180A, 0xE004}},
    {40553, 2, {0x810, 0x820, 0x7F7C, 0x844, 0x5564, 0x2254, 0x4144, 0xBECC, 0x2240, 0x3E7E, 0x2202, 0x3E02, 0x87A, 0x2A02, 0x490A, 0x1804}},
    {40554, 2, {0x2820, 0x2440, 0x7EFC, 0xC884, 0x48A4, 0x7E84, 0x4894, 0x4888, 0x7E80, 0x48FE, 0x4802, 0x7E02, 0x41FA, 0xAA02, 0xAA14, 0x8}},
    {40555, 2, {0x828, 0x7F24, 0xFE, 0x3E20, 0x2250, 0x3E50, 0x4952, 0x9A8E, 0x400, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40556, 2, {0x7E10, 0x1420, 0x87C, 0xFF44, 0x2964, 0x4A54, 0xA844, 0x104C, 0x7F40, 0x557E, 0x6302, 0x5D02, 0x557A, 0x5D02, 0x410A, 0x4304}},
    {40557, 2, {0x40, 0x3EFC, 0x2348, 0x3E30, 0x9CE, 0x2E78, 0x2848, 0x7E78, 0x400, 0x1FE0, 0x1220, 0x1160, 0x1FFC, 0x4, 0x7FD4, 0x8}},
    {40558, 2, {0x20, 0xFE40, 0xAAFC, 0xAA84, 0xFEA4, 0x84, 0xFE94, 0x88, 0x7C80, 0x44FE, 0x7C02, 0x2A02, 0x65FA, 0xA202, 0x3114, 0x2008}},
    {40559, 2, {0x1020, 0xFE40, 0xFC, 0xFE84, 0x82A4, 0xBA84, 0xAA94, 0xFE88, 0x80, 0x7CFE, 0x4402, 0x7C02, 0x45FA, 0x7C02, 0x14, 0xFE08}},
    {40560, 2, {0x80, 0x3FFE, 0x2250, 0x24FC, 0x2D90, 0x36FC, 0x2490, 0x24FC, 0x2100, 0x27F0, 0x2490, 0x2420, 0x47FC, 0x4004, 0x9FD4, 0x8}},
    {40561, 2, {0x1044, 0x21FE, 0x7C54, 0x4448, 0x64FE, 0x5590, 0x44FE, 0x4C90, 0x40FE, 0x7C90, 0x4FE, 0x480, 0xF5FE, 0x444, 0x1438, 0x9C6}},
    {40562, 2, {0x2410, 0xFF20, 0x247C, 0xFF44, 0x8164, 0x7E54, 0x44, 0xFF4C, 0x1040, 0x28FE, 0xD502, 0x2E02, 0xD57A, 0x2482, 0xD40A, 0x804}},
    {40563, 2, {0x2208, 0xFF90, 0x223E, 0xF7A2, 0x94B2, 0xF7AA, 0x2822, 0x2426, 0x7FA0, 0x483E, 0xFF02, 0x4802, 0x7F7A, 0x4802, 0x7F8A, 0x4004}},
    {40564, 2, {0x7F10, 0x820, 0x7F7C, 0x4944, 0x2A64, 0x854, 0x2A44, 0x4C, 0x2740, 0xFD7E, 0x2702, 0x7502, 0xAF7A, 0x2502, 0x270A, 0x2504}},
    {40565, 2, {0x100, 0x100, 0x1FE, 0x100, 0x100, 0x3FF8, 0x2008, 0x2928, 0x2448, 0x2288, 0x2928, 0x2288, 0x2448, 0x2928, 0x3FF8, 0x2008}},
    {40566, 2, {0x1020, 0x1020, 0x1E20, 0x1050, 0x1050, 0xFE88, 0x9344, 0xC622, 0xAA20, 0xD6F8, 0xAA08, 0xC608, 0x9210, 0xFE10, 0x8220, 0x20}},
    {40567, 2, {0x1020, 0x1020, 0x1E50, 0x1050, 0x1088, 0xFF24, 0x9212, 0xC610, 0xABFC, 0xD604, 0xAA08, 0xC688, 0x9250, 0xFE20, 0x8210, 0x10}},
    {40568, 2, {0x1010, 0x1028, 0x1E44, 0x1082, 0x107C, 0xFE10, 0x9210, 0xC6FE, 0xAA92, 0xD692, 0xAAFE, 0xC610, 0x9228, 0xFE28, 0x8244, 0x82}},
    {40569, 2, {0x100C, 0x100A, 0x1E08, 0x10FE, 0x1088, 0xFE88, 0x92FA, 0xC68A, 0xAA8A, 0xD6EC, 0xAAAC, 0xC6AC, 0x92EA, 0xFE8A, 0x8296, 0x122}},
    {40570, 2, {0x1044, 0x1024, 0x1E28, 0x10FE, 0x1010, 0xFE7C, 0x9210, 0xC6FE, 0xAA20, 0xD620, 0xAA7C, 0xC650, 0x9290, 0xFF10, 0x82FE, 0x0}},
    {40571, 2, {0x1088, 0x1050, 0x1E00, 0x11FE, 0x1050, 0xFEFC, 0x9254, 0xC7FE, 0xAA54, 0xD6FC, 0xAA50, 0xC6D8, 0x9354, 0xFE52, 0x8250, 0x50}},
    {40572, 2, {0x1010, 0x1010, 0x1E28, 0x1044, 0x1082, 0xFE7C, 0x9200, 0xC6EE, 0xAAAA, 0xD6AA, 0xAAEE, 0xC644, 0x9244, 0xFEAA, 0x82AA, 0x112}},
    {40573, 2, {0x100, 0xF9FE, 0xA220, 0xA43E, 0xF820, 0x8BFE, 0x8AAA, 0xFA52, 0xA326, 0xA252, 0xA2AA, 0xFBFE, 0x0, 0x3FF8, 0x2448, 0xFFFE}},
    {40574, 2, {0x2084, 0x2044, 0x3C48, 0x21FE, 0x2020, 0xFCFC, 0x8420, 0x95FE, 0xD440, 0xA480, 0xD4FE, 0x9510, 0x8610, 0xFC10, 0x85FE, 0x0}},
    {40575, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x2220, 0x3FFC, 0x2224, 0x2224, 0x3FFC, 0x2840, 0x2844, 0x2F58, 0x4860, 0x4844, 0x8B44, 0xC3C}},
    {40576, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2F78, 0x2842, 0x2F3E, 0x2000, 0x2878, 0x4F82, 0x4802, 0x87FE}},
    {40577, 2, {0x100, 0x200, 0xFF0, 0x7060, 0x380, 0x1D00, 0xE080, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {40578, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2840, 0x2F78, 0x2842, 0x2F3E, 0x2000, 0x23F0, 0x4210, 0x4412, 0x980E}},
    {40579, 2, {0x100, 0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2844, 0x2F48, 0x2874, 0x2B44, 0x2C3C, 0x4000, 0x4A48, 0x8924, 0x1124}},
    {40580, 2, {0x440, 0x820, 0x3018, 0xDFF6, 0x410, 0x950, 0x10A0, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {40581, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2400, 0x2FF8, 0x3008, 0x2FC8, 0x2868, 0x4FD0, 0x4802, 0x87FE}},
    {40582, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x47F8, 0x4408, 0xBFFE}},
    {40583, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2008, 0x23F0, 0x2040, 0x4FFE, 0x4150, 0x8E4E}},
    {40584, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2080, 0x2FFE, 0x2040, 0x47FC, 0x4040, 0x9FFE}},
    {40585, 2, {0x0, 0x7C7C, 0x1010, 0x1010, 0xFEFE, 0x1110, 0x1090, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {40586, 2, {0x1110, 0x920, 0x7FFC, 0x540, 0x1930, 0x610C, 0x80, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {40587, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x279E, 0x2248, 0x2150, 0x2FFE, 0x4150, 0x4248, 0x8C46}},
    {40588, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x27F0, 0x2410, 0x27F0, 0x2000, 0x2FF8, 0x2080, 0x5FFC, 0x4220, 0x9C1C}},
    {40589, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2040, 0x3FFE, 0x2200, 0x2408, 0x2FFC, 0x4494, 0x4892, 0x908E}},
    {40590, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2000, 0x2FFC, 0x2800, 0x2BFC, 0x2800, 0x4FFE, 0x4A28, 0x939E}},
    {40591, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x2FFC, 0x2104, 0x3FFE, 0x2104, 0x2FFC, 0x2200, 0x47FC, 0x4A04, 0x93FC}},
    {40592, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2040, 0x2FFE, 0x2110, 0x20E0, 0x2F1E, 0x43F8, 0x4208, 0x83F8}},
    {40593, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2422, 0x27BE, 0x2100, 0x2E78, 0x2808, 0x2F78, 0x2808, 0x4FF8, 0x4222, 0x9C1E}},
    {40594, 2, {0x844, 0x7F44, 0x55FE, 0x5444, 0x7F44, 0x557C, 0x5544, 0x7F44, 0x547C, 0x5444, 0x5E44, 0x55FE, 0x5400, 0x5C28, 0x5644, 0x8082}},
    {40595, 2, {0x820, 0x820, 0x7EFC, 0x1C70, 0x2AA8, 0xC926, 0x8A0, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {40596, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2220, 0x27B0, 0x2D28, 0x3224, 0x2DFE, 0x3000, 0x27F8, 0x4408, 0x4408, 0x87F8}},
    {40597, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x2FFC, 0x2824, 0x2BC4, 0x2884, 0x2FF4, 0x2AA4, 0x4C94, 0x4FFC, 0x8804}},
    {40598, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2080, 0x3FFE, 0x2000, 0x2FF8, 0x2808, 0x2FF8, 0x2080, 0x4888, 0x5284, 0x8100}},
    {40599, 2, {0x3F7E, 0x0, 0x3E7C, 0x2244, 0x3264, 0x2B54, 0x80, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2844, 0x2F78, 0x4842, 0x8F3E}},
    {40600, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2010, 0x2FE0, 0x2080, 0x3FFC, 0x2490, 0x2FF8, 0x3416, 0x47F0, 0x4410, 0x87F0}},
    {40601, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x2028, 0x2FFE, 0x2820, 0x2BA4, 0x2828, 0x2BAC, 0x4A94, 0x4BAC, 0x9044}},
    {40602, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x2F7C, 0x2904, 0x2F7C, 0x2800, 0x2F7C, 0x2844, 0x4F28, 0x4830, 0x88CE}},
    {40603, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x2F7E, 0x2124, 0x2F3C, 0x2824, 0x2F3C, 0x2126, 0x417C, 0x4504, 0x8204}},
    {40604, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x2FFE, 0x2110, 0x27FC, 0x2514, 0x27FC, 0x2040, 0x4FFE, 0x4150, 0x864C}},
    {40605, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2420, 0x27BC, 0x2422, 0x2F9E, 0x2908, 0x2F7E, 0x2948, 0x5F28, 0x4508, 0x9B18}},
    {40606, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2040, 0x27FC, 0x2110, 0x2FFE, 0x2208, 0x23F8, 0x2208, 0x4FFE, 0x4040, 0x8040}},
    {40607, 2, {0x810, 0x7F54, 0x5438, 0x54FE, 0x7F10, 0x5538, 0x5556, 0x7F90, 0x5484, 0x54EE, 0x5EA4, 0x54B4, 0x555E, 0x5C44, 0x5684, 0x8104}},
    {40608, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2888, 0x2FF8, 0x2888, 0x3FFE, 0x2000, 0x2FF8, 0x2888, 0x4FF8, 0x4888, 0xBFFE}},
    {40609, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2040, 0x2FFE, 0x20A0, 0x2F5E, 0x2554, 0x2554, 0x2B5A, 0x4408, 0x47F8, 0x8808}},
    {40610, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x2000, 0x27FC, 0x2040, 0x2FFE, 0x2842, 0x2358, 0x2000, 0x4EEE, 0x4AAA, 0x8EEE}},
    {40611, 2, {0x13DE, 0xFE52, 0xABDE, 0xA800, 0xFFFE, 0xAA00, 0xAAC8, 0xFE4E, 0xABEA, 0xAAB2, 0xBEEA, 0xAAAA, 0xAAEA, 0xBAB4, 0xADE4, 0x2A}},
    {40612, 2, {0x80, 0x3FFE, 0x2220, 0x3FFC, 0x2224, 0x3FFC, 0x4810, 0xFF7E, 0x5448, 0x7F7E, 0x554A, 0x7F7E, 0x5264, 0x5B76, 0x52A4, 0x9B36}},
    {40613, 2, {0x100, 0x7FFC, 0x920, 0x920, 0x1550, 0x2388, 0x540, 0x1930, 0xE20E, 0x7E0, 0x820, 0x3440, 0x280, 0x300, 0xC80, 0x7040}},
    {40614, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x800, 0x1FF0, 0x2820, 0x4440, 0x380, 0x1C70, 0xE00E}},
    {40615, 2, {0x820, 0x820, 0xFF40, 0x2A7E, 0x2A80, 0x5D00, 0x887C, 0x1C04, 0x2A08, 0x5010, 0x1E20, 0x2240, 0x5442, 0x842, 0x143E, 0x6000}},
    {40616, 2, {0x810, 0x810, 0xFF10, 0x2A10, 0x2A54, 0x5D52, 0x8852, 0x1C90, 0x2A10, 0x5014, 0x1E04, 0x2208, 0x5408, 0x810, 0x1420, 0x60C0}},
    {40617, 2, {0x1020, 0x1020, 0xFE20, 0x5420, 0x55FC, 0xBA20, 0x1020, 0x3820, 0x55FE, 0xA220, 0x3C50, 0x4450, 0xA888, 0x1088, 0x2904, 0xC202}},
    {40618, 2, {0x1000, 0x11FE, 0xFE20, 0x5420, 0x54A0, 0xBABC, 0x10A4, 0x38A4, 0x54A4, 0xA2FC, 0x3C04, 0x4404, 0xA804, 0x1004, 0x2828, 0xC010}},
    {40619, 2, {0x800, 0x8FE, 0xFF10, 0x2A10, 0x2A50, 0x5D5E, 0x8850, 0x1C50, 0x2A50, 0x507E, 0x1E02, 0x2202, 0x5402, 0x802, 0x1414, 0x6008}},
    {40620, 2, {0x1010, 0x1010, 0xFE10, 0x54FE, 0x5492, 0xBA94, 0x1090, 0x38FC, 0x54A4, 0xA2A4, 0x3CA8, 0x44A8, 0xA890, 0x1128, 0x2944, 0xC282}},
    {40621, 2, {0x1040, 0x1040, 0xFEFC, 0x5484, 0x5504, 0xBAF4, 0x1094, 0x3894, 0x54F4, 0xA284, 0x3C94, 0x4488, 0xA882, 0x1082, 0x287E, 0xC000}},
    {40622, 2, {0x810, 0x810, 0xFF10, 0x2A7E, 0x2A10, 0x5D10, 0x8810, 0x1CFE, 0x2A10, 0x5010, 0x1E20, 0x2228, 0x5444, 0x8FE, 0x1442, 0x6000}},
    {40623, 2, {0x1050, 0x1050, 0xFE50, 0x5450, 0x55FC, 0xBB54, 0x1154, 0x3954, 0x5554, 0xA3FC, 0x3D54, 0x4554, 0xA954, 0x1154, 0x29FC, 0xC104}},
    {40624, 2, {0x1020, 0x1020, 0xFE48, 0x5484, 0x55FE, 0xBA12, 0x1090, 0x3890, 0x54FE, 0xA310, 0x3C10, 0x45FE, 0xA810, 0x1010, 0x2810, 0xC010}},
    {40625, 2, {0x1014, 0x1012, 0xFEFE, 0x5410, 0x5410, 0xBAFE, 0x1092, 0x3892, 0x54FE, 0xA292, 0x3C92, 0x44FE, 0xA892, 0x1092, 0x2892, 0xC086}},
    {40626, 2, {0x800, 0x8FC, 0xFF84, 0x2A84, 0x2AFC, 0x5D84, 0x8884, 0x1CFC, 0x2A84, 0x5084, 0x1EFC, 0x2250, 0x5450, 0x852, 0x1492, 0x610E}},
    {40627, 2, {0x810, 0x810, 0xFF10, 0x2AFE, 0x2A10, 0x5D54, 0x8854, 0x1C54, 0x2ABA, 0x5010, 0x1E38, 0x2254, 0x5492, 0x810, 0x1410, 0x6010}},
    {40628, 2, {0x1080, 0x1080, 0xFEFE, 0x5502, 0x5622, 0xBAAA, 0x1072, 0x3822, 0x55FE, 0xA222, 0x3C72, 0x44AA, 0xA922, 0x1022, 0x280A, 0xC004}},
    {40629, 2, {0x800, 0x8FE, 0xFF10, 0x2A20, 0x2AFE, 0x5DAA, 0x88AA, 0x1CBA, 0x2AAA, 0x50AA, 0x1EBA, 0x22AA, 0x54AA, 0x8AA, 0x14FE, 0x6082}},
    {40630, 2, {0x1020, 0x1010, 0xFEFE, 0x5428, 0x5492, 0xBAAA, 0x1082, 0x38FE, 0x5410, 0xA2FE, 0x3CA2, 0x44AA, 0xA8BA, 0x1082, 0x288A, 0xC084}},
    {40631, 2, {0x10A8, 0x11FC, 0xFEAA, 0x57FE, 0x56AA, 0xBBFE, 0x12AA, 0x3BFE, 0x5400, 0xA3FE, 0x3C00, 0x45FC, 0xA904, 0x11FC, 0x2888, 0xC3FE}},
    {40632, 2, {0x1020, 0x1020, 0xFE20, 0x1020, 0x7DFC, 0x1020, 0xFE20, 0x2020, 0x3DFE, 0x4420, 0x4850, 0xA850, 0x1088, 0x2888, 0x4104, 0x8202}},
    {40633, 2, {0x1080, 0x1080, 0xFEFE, 0x1102, 0x7E22, 0x10AA, 0xFE72, 0x2022, 0x3DFE, 0x4422, 0x4872, 0xA8AA, 0x1122, 0x2822, 0x400A, 0x8004}},
    {40634, 2, {0x1000, 0x11FE, 0xFE20, 0x1040, 0x7DFC, 0x1154, 0xFF54, 0x2174, 0x3D54, 0x4554, 0x4974, 0xA954, 0x1154, 0x2954, 0x41FC, 0x8104}},
    {40635, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0x2210, 0x3FBE, 0x2210, 0x2610, 0x2738, 0x2AB4, 0x2A54, 0x5292, 0x4210, 0x8210, 0x210}},
    {40636, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0x2FBC, 0x2318, 0x26B4, 0x2A52, 0x2100, 0x2210, 0x27E0, 0x40C0, 0x4308, 0x8FFC, 0x4}},
    {40637, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2210, 0x2FBC, 0x2318, 0x26B4, 0x2A52, 0x2100, 0x2240, 0x2480, 0x4910, 0x4208, 0x87FC, 0x4}},
    {40638, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2638, 0x2B54, 0x3292, 0x2078, 0x2F80, 0x20F8, 0x2F80, 0x20FC, 0x5F82, 0x4082, 0x807E}},
    {40639, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2000, 0x27F8, 0x2408, 0x27F8, 0x2000, 0x2FFC, 0x4804, 0x4804, 0x8FFC}},
    {40640, 2, {0x100, 0x80, 0x3FFE, 0x2210, 0x2F7C, 0x2638, 0x2B54, 0x3292, 0x3F80, 0x297C, 0x2F24, 0x2924, 0x2F28, 0x4990, 0x5F28, 0x8146}},
    {40641, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3412, 0x2FE0, 0x2080, 0x3FFC, 0x2490, 0x2FF8, 0x3416, 0x47F0, 0x4410, 0x87F0}},
    {40642, 2, {0x80, 0x3FFE, 0x2410, 0x3F7C, 0x2638, 0x2D54, 0x3492, 0x2FF8, 0x22A0, 0x2FFC, 0x2220, 0x2FF8, 0x2888, 0x4FF8, 0x4410, 0x8808}},
    {40643, 2, {0x820, 0x7FFC, 0x820, 0xFE0, 0x0, 0xFFFE, 0x100, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1FF0, 0x820, 0x1010}},
    {40644, 2, {0x440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x0, 0x820, 0x1010, 0x600C}},
    {40645, 2, {0x2210, 0x2210, 0x7F10, 0x2228, 0x2228, 0xFFC4, 0x844, 0x7FA2, 0x4910, 0x7F10, 0x497C, 0x7F04, 0x8, 0x2208, 0x2110, 0x4110}},
    {40646, 2, {0x2210, 0x2210, 0x7F10, 0x2210, 0x22FE, 0xFF92, 0x892, 0x7F10, 0x4928, 0x7F28, 0x4928, 0x7F28, 0x48, 0x224A, 0x218A, 0x4106}},
    {40647, 2, {0x2210, 0x2210, 0x7F10, 0x2210, 0x221E, 0xFF90, 0x810, 0x7F10, 0x497C, 0x7F44, 0x4944, 0x7F44, 0x44, 0x2244, 0x217C, 0x4144}},
    {40648, 2, {0x2220, 0x2210, 0x7F00, 0x22FE, 0x2210, 0xFF90, 0x810, 0x7F10, 0x497C, 0x7F10, 0x4910, 0x7F10, 0x10, 0x2210, 0x21FE, 0x4100}},
    {40649, 2, {0x1108, 0x890, 0x7FFE, 0x4442, 0x9FF4, 0x440, 0x7FFC, 0x100, 0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x440, 0x820, 0x1010}},
    {40650, 2, {0x2820, 0x2820, 0x7CF8, 0x2820, 0x2820, 0xFE20, 0x11FC, 0x7C00, 0x5420, 0x5420, 0x7CF8, 0x5420, 0x5420, 0x7C20, 0x29FC, 0x4400}},
    {40651, 2, {0x1088, 0x1088, 0x95FC, 0x5488, 0x5888, 0x13FE, 0xFC20, 0x29FC, 0x2924, 0x29FC, 0x2924, 0x29FC, 0x2888, 0x4904, 0x4A02, 0x87FE}},
    {40652, 2, {0xA80, 0x3138, 0x2288, 0x3938, 0x2288, 0xFFFE, 0x8442, 0x3FF8, 0x440, 0xFFFE, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x820, 0x1010}},
    {40653, 2, {0xF8, 0x3F00, 0x100, 0x7FFC, 0x920, 0x1110, 0x2288, 0xC444, 0x920, 0x3118, 0xCB26, 0x5C0, 0x1930, 0x6108, 0x500, 0x200}},
    {40654, 2, {0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4908, 0x280, 0xC60, 0x3118, 0xC926, 0x540, 0x920, 0x1110, 0x2508, 0x200}},
    {40655, 2, {0xC20, 0x7020, 0x1020, 0xFE20, 0x383E, 0x5420, 0x9020, 0x2820, 0x55FC, 0x9104, 0x5504, 0x3904, 0x5504, 0x9304, 0x51FC, 0x2104}},
    {40656, 2, {0xC40, 0x7020, 0x11FC, 0xFE50, 0x3924, 0x5554, 0x9104, 0x29FC, 0x5420, 0x91FC, 0x5544, 0x3954, 0x5574, 0x9304, 0x5114, 0x2108}},
    {40657, 2, {0x0, 0x3FF8, 0x2108, 0x2928, 0x2548, 0x3FF8, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x4888, 0x4444, 0x8444}},
    {40658, 2, {0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x3FF8, 0x100, 0x100, 0xFFFE, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {40659, 2, {0x28, 0x7F24, 0x4924, 0x6B20, 0x5D2E, 0x49F0, 0x7F20, 0x820, 0x7F20, 0x820, 0xFA0, 0xF010, 0x12, 0x550A, 0x4A86, 0x8002}},
    {40660, 2, {0x20, 0xFE20, 0x9220, 0xD650, 0xBA50, 0x9288, 0xFF44, 0x1222, 0xFE20, 0x10F8, 0x1E08, 0xE008, 0x210, 0xAA10, 0xA820, 0x8020}},
    {40661, 2, {0x10, 0xFE10, 0x9210, 0xD610, 0xBAFE, 0x9292, 0xFE92, 0x1010, 0xFE28, 0x1028, 0x1E28, 0xE028, 0x248, 0xAA4A, 0xA88A, 0x8106}},
    {40662, 2, {0x0, 0xFEFE, 0x9210, 0xD610, 0xBA90, 0x9290, 0xFE90, 0x10FE, 0xFE10, 0x1010, 0x1E28, 0xE028, 0x248, 0xAA4A, 0xA88A, 0x8106}},
    {40663, 2, {0x20, 0xFE20, 0x9220, 0xD7FC, 0xBA20, 0x92A8, 0xFEA8, 0x10A8, 0xFEA8, 0x10A8, 0x1EF8, 0xE022, 0x222, 0xAA22, 0xA81E, 0x8000}},
    {40664, 2, {0x20, 0xFE28, 0x9224, 0xD624, 0xBA20, 0x93FE, 0xFE20, 0x1020, 0xFE20, 0x1050, 0x1E50, 0xE050, 0x288, 0xAA88, 0xA904, 0x8202}},
    {40665, 2, {0x10, 0x7F14, 0x4912, 0x7F10, 0x497E, 0x7F10, 0x828, 0x7F28, 0x844, 0xF44, 0xF082, 0x0, 0x2488, 0x2244, 0x4244, 0x8004}},
    {40666, 2, {0x44, 0xFE44, 0x9244, 0xD644, 0xBAFE, 0x9244, 0xFE44, 0x1044, 0xFE44, 0x107C, 0x1E44, 0xE044, 0x244, 0xAA44, 0xA87C, 0x8044}},
    {40667, 2, {0x890, 0x1088, 0x307E, 0xD7C0, 0x1022, 0x101A, 0x6, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE, 0x4444}},
    {40668, 2, {0x10, 0xFE10, 0x9292, 0xD692, 0xBA92, 0x9292, 0xFEFE, 0x1010, 0xFE10, 0x1092, 0x1E92, 0xE092, 0x292, 0xAA92, 0xA8FE, 0x8002}},
    {40669, 2, {0x10, 0xFE50, 0x9250, 0xD690, 0xBAFC, 0x9354, 0xFFD4, 0x1094, 0xFE94, 0x1114, 0x1F54, 0xE1D4, 0x254, 0xAA24, 0xA834, 0x8048}},
    {40670, 2, {0x10, 0xFE10, 0x9210, 0xD610, 0xBA1E, 0x9210, 0xFE10, 0x1010, 0xFEFE, 0x1082, 0x1E82, 0xE082, 0x282, 0xAA82, 0xA8FE, 0x8082}},
    {40671, 2, {0x20, 0xFE20, 0x927C, 0xD684, 0xBB48, 0x9230, 0xFE20, 0x1048, 0xFE90, 0x103E, 0x1E42, 0xE0A4, 0x218, 0xAA10, 0xA820, 0x80C0}},
    {40672, 2, {0x20, 0xFE20, 0x9220, 0xD7FE, 0xBA20, 0x9220, 0xFEFC, 0x1000, 0xFE00, 0x10FC, 0x1E84, 0xE084, 0x284, 0xAA84, 0xA8FC, 0x8084}},
    {40673, 2, {0x3FFC, 0x2090, 0x3FFC, 0x2140, 0x2630, 0x380C, 0x2FF8, 0x2AA8, 0x29C8, 0x2FF8, 0x2080, 0x2FF8, 0x4080, 0x5FFC, 0x8A48, 0x1124}},
    {40674, 2, {0x10, 0xFE10, 0x9220, 0xD644, 0xBAFE, 0x9228, 0xFE44, 0x10A2, 0xFE3C, 0x1044, 0x1E44, 0xE0A8, 0x210, 0xAA28, 0xA844, 0x8082}},
    {40675, 2, {0x40, 0xFE40, 0x927E, 0xD680, 0xBB7C, 0x9244, 0xFE64, 0x1054, 0xFEFE, 0x1044, 0x1EA4, 0xE094, 0x2FE, 0xAA04, 0xA828, 0x8010}},
    {40676, 2, {0x20, 0xFE20, 0x93FE, 0xD650, 0xBAA8, 0x9326, 0xFEF8, 0x10A8, 0xFEA8, 0x10F8, 0x1EA8, 0xE0A8, 0x2F8, 0xAAAA, 0xA822, 0x801E}},
    {40677, 2, {0x20, 0xFE10, 0x92FE, 0xD600, 0xBA00, 0x927C, 0xFE44, 0x1044, 0xFE7C, 0x1010, 0x1E54, 0xE052, 0x292, 0xAA10, 0xA850, 0x8020}},
    {40678, 2, {0x20, 0xFE10, 0x92FE, 0xD682, 0xBA40, 0x924E, 0xFE6A, 0x10AA, 0xFEAA, 0x10AA, 0x1F6C, 0xE028, 0x24A, 0xAA4A, 0xA88A, 0x8106}},
    {40679, 2, {0x640, 0x387C, 0x894, 0x7E24, 0x1C44, 0x2A94, 0x4808, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x7FFC, 0x100, 0xFFFE, 0x4444}},
    {40680, 2, {0x1110, 0x920, 0x7FFE, 0x4002, 0x9FF4, 0x1010, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x2448, 0x4224}},
    {40681, 2, {0x10, 0xFE10, 0x92FC, 0xD610, 0xBA10, 0x93FE, 0xFE02, 0x1054, 0xFE30, 0x1090, 0x1E50, 0xE1FE, 0x228, 0xAA44, 0xA882, 0x8102}},
    {40682, 2, {0x20, 0xFE28, 0x9244, 0xD6FE, 0xBA20, 0x93FE, 0xFE88, 0x1114, 0xFE22, 0x1048, 0x1E10, 0xE022, 0x2C4, 0xAA18, 0xA860, 0x8180}},
    {40683, 2, {0x0, 0xFEFE, 0x9228, 0xD628, 0xBAFE, 0x92AA, 0xFEAA, 0x10AA, 0xFEFE, 0x1010, 0x1E10, 0xE0FE, 0x210, 0xAA10, 0xA9FE, 0x8000}},
    {40684, 2, {0xC, 0xFE0A, 0x9208, 0xD6FE, 0xBA88, 0x9288, 0xFEFA, 0x108A, 0xFE8A, 0x10EC, 0x1EAC, 0xE0AC, 0x2EA, 0xAA8A, 0xA896, 0x8122}},
    {40685, 2, {0x20, 0xFE30, 0x9248, 0xD684, 0xBB7A, 0x9200, 0xFEFC, 0x1084, 0xFE84, 0x10FC, 0x1E48, 0xE048, 0x3FE, 0xAA48, 0xA888, 0x8108}},
    {40686, 2, {0x48, 0xFE48, 0x93FC, 0xD648, 0xBA78, 0x9248, 0xFE78, 0x1048, 0xFE48, 0x11FE, 0x1E80, 0xE0A8, 0x2C4, 0xAA80, 0xA8FE, 0x8000}},
    {40687, 2, {0x20, 0xFE10, 0x92FE, 0xD600, 0xBA44, 0x9228, 0xFEFE, 0x1000, 0xFE7C, 0x1044, 0x1E44, 0xE07C, 0x244, 0xAA44, 0xA87C, 0x8044}},
    {40688, 2, {0x10, 0xFE10, 0x92FE, 0xD610, 0xBA7C, 0x9244, 0xFE7C, 0x1044, 0xFE7C, 0x1044, 0x1E7C, 0xE044, 0x2FE, 0xAA28, 0xA844, 0x8082}},
    {40689, 2, {0x124, 0x78A8, 0x49FC, 0x4840, 0x4BFE, 0x7908, 0x4BFC, 0x4DAE, 0x4974, 0x79FC, 0x4820, 0x49FC, 0x4820, 0x4BFE, 0x4954, 0x9A2A}},
    {40690, 2, {0x10, 0xFE20, 0x9244, 0xD6FE, 0xBA44, 0x92AA, 0xFEEE, 0x1010, 0xFE28, 0x1054, 0x1EAA, 0xE050, 0x224, 0xAA48, 0xA810, 0x8060}},
    {40691, 2, {0x7F78, 0x4848, 0x5E86, 0x6478, 0x5F48, 0x4A30, 0x7F4C, 0x0, 0x3FF8, 0x2928, 0x2548, 0x3FF8, 0x100, 0xFFFE, 0x2448, 0x4224}},
    {40692, 2, {0x1110, 0x1550, 0x2550, 0x47D0, 0x903E, 0x17A4, 0x2054, 0x6FD4, 0xA954, 0x2DD4, 0x2B54, 0x2FC8, 0x2108, 0x2FD4, 0x2014, 0x2AA2}},
    {40693, 2, {0x40, 0xFE7C, 0x9284, 0xD7FE, 0xBAA2, 0x92C8, 0xFEFE, 0x1080, 0xFEBE, 0x1080, 0x1EBE, 0xE080, 0x2BE, 0xAAA2, 0xA8BE, 0x8122}},
    {40694, 2, {0x3FFE, 0x2888, 0x2F8A, 0x2008, 0x2FBE, 0x2888, 0x2F88, 0x2894, 0x29A2, 0x2000, 0x2FFE, 0x2952, 0x2FFE, 0x4040, 0x5FFE, 0x8924}},
    {40695, 2, {0x10, 0xFEFE, 0x9210, 0xD67C, 0xBA00, 0x92FE, 0xFEAA, 0x10FE, 0xFE44, 0x107C, 0x1E44, 0xE07C, 0x244, 0xAA7C, 0xA828, 0x8044}},
    {40696, 2, {0x40, 0xF87C, 0xA840, 0xABFC, 0xFA44, 0xABF0, 0xFA44, 0x22FC, 0xFAA8, 0x22F8, 0x3AA8, 0xE2F8, 0xA00, 0xA5FC, 0xA554, 0xBFE}},
    {40697, 2, {0x440, 0x2448, 0x1450, 0x440, 0xFFFE, 0x100, 0x1110, 0x920, 0x3FF8, 0x2108, 0x2548, 0x2528, 0x2928, 0x2108, 0x2128, 0x2110}},
    {40698, 2, {0x1408, 0x5528, 0x3628, 0x1428, 0xFF44, 0x4944, 0x2A82, 0x7F7C, 0x4924, 0x4924, 0x5D24, 0x6B24, 0x4944, 0x4944, 0x4994, 0x4B08}},
    {40699, 2, {0x1428, 0x5524, 0x3624, 0x1420, 0xFF7E, 0x4920, 0x2A20, 0x7F3C, 0x4934, 0x4954, 0x5D54, 0x6B54, 0x4988, 0x4988, 0x4914, 0x4B22}},
    {40700, 2, {0x1414, 0x5512, 0x367E, 0x1410, 0xFF10, 0x497E, 0x2A52, 0x7F52, 0x497E, 0x4952, 0x5D52, 0x6B7E, 0x4952, 0x4952, 0x4952, 0x4B46}},
    {40701, 2, {0x0, 0x3FF8, 0x2288, 0x2288, 0x3EF8, 0x280, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x2288, 0x2288, 0x3EF8, 0x228A, 0x282, 0x1FE}},
    {40702, 2, {0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x210A, 0x102, 0x102, 0xFE}},
    {40703, 2, {0x1FF0, 0x0, 0xFFFE, 0x844, 0x303C, 0xDFF0, 0x1290, 0x1EF0, 0x280, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EFA, 0x282, 0x1FE}},
    {40704, 2, {0x100, 0x3FF8, 0x100, 0xFFFE, 0x1010, 0x3FF8, 0x5294, 0x1EF0, 0x280, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EFA, 0x282, 0x1FE}},
    {40705, 2, {0x100, 0x3FF8, 0x100, 0xFFFE, 0x420, 0x1FF0, 0x0, 0x1FF0, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40706, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x0, 0x1FF0, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40707, 2, {0x100, 0x3FF8, 0x100, 0xFFFE, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40708, 2, {0x900, 0xFF0, 0x1100, 0x7FFC, 0x540, 0x1930, 0xE10E, 0x1FF0, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40709, 2, {0x2000, 0x3F7C, 0x4844, 0xFFC4, 0x147C, 0x2244, 0x4100, 0x1FF0, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40710, 2, {0x0, 0xFDFC, 0x8554, 0x154, 0x79DC, 0x4850, 0x79DC, 0x4954, 0x7954, 0x21DC, 0x1154, 0xFD54, 0x1DC, 0x5150, 0x4852, 0x883E}},
    {40711, 2, {0x820, 0x7E20, 0x87E, 0xFEC4, 0x1028, 0x1E10, 0x2228, 0x5FF6, 0x9290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40712, 2, {0x4920, 0x2A3E, 0x7F48, 0x4948, 0x5DA8, 0x6B10, 0x4928, 0x5FF6, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40713, 2, {0x0, 0x3EF8, 0x2288, 0x3FF8, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40714, 2, {0x10, 0x7E7C, 0x4228, 0x7EFE, 0x4010, 0x7E7C, 0xA210, 0x3FF0, 0x1290, 0x3EF8, 0x2288, 0x3EF8, 0x2288, 0x3EF8, 0x282, 0x1FE}},
    {40715, 2, {0x1FF0, 0x0, 0xFFFE, 0x844, 0x303C, 0xDFF0, 0x1010, 0x1FF0, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {40716, 2, {0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0xFFFE, 0x0, 0x1FF0, 0x1010, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {40717, 2, {0x0, 0x3EF8, 0x2288, 0x3FF8, 0x1110, 0x1FF0, 0x1110, 0xFFFE, 0x100, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FFA, 0x102, 0xFE}},
    {40718, 2, {0xFE0, 0x820, 0xFE0, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x7C7C, 0x440, 0x440, 0xFC7C, 0x2444, 0x2444, 0x4444, 0x8444}},
    {40719, 2, {0x7FFE, 0x4002, 0x8FE4, 0x820, 0xFE0, 0x820, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x7C7C, 0x440, 0xFC7C, 0x2444, 0x4444, 0x8444}},
    {40720, 2, {0x7FF8, 0x1008, 0x101E, 0x2FE2, 0x482A, 0x8FE4, 0x820, 0x4FE4, 0x4824, 0x4FE4, 0x4004, 0x7C7C, 0x440, 0xFC7C, 0x2444, 0x4444}},
    {40721, 2, {0x100, 0x1FC, 0x100, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x5FF4, 0x5014, 0x5FF4, 0x4004, 0x7C7C, 0x440, 0xFC7C, 0x2444, 0x4444}},
    {40722, 2, {0x80, 0x7FFC, 0x480, 0x1980, 0x6FE0, 0x820, 0xFE0, 0x4824, 0x4FE4, 0x4824, 0x4FE4, 0x7C7C, 0x440, 0xFC7C, 0x2444, 0xC444}},
    {40723, 2, {0x810, 0x810, 0xFF10, 0x8FE, 0x810, 0x7E10, 0xFC, 0x7E44, 0x4244, 0x4244, 0x7E28, 0x4228, 0x2410, 0xF28, 0xF044, 0x4082}},
    {40724, 2, {0x810, 0x810, 0xFF10, 0x81E, 0x810, 0x7E10, 0xFC, 0x7E44, 0x4244, 0x4244, 0x7E28, 0x4228, 0x2410, 0xF28, 0xF044, 0x4082}},
    {40725, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4204, 0x7E48, 0x2438, 0x1E44, 0xE800, 0x1FF0, 0x2820, 0x7C0, 0x1830, 0xE18E, 0xC40, 0x300}},
    {40726, 2, {0x100, 0x7FFC, 0x920, 0xFFFE, 0x820, 0x810, 0xFF10, 0x8FE, 0x7E10, 0xFC, 0x7E44, 0x4248, 0x7E28, 0x2410, 0x1E28, 0xE0C6}},
    {40727, 2, {0x2490, 0x14A0, 0x2494, 0x884, 0x707C, 0x810, 0xFF10, 0x8FE, 0x7E10, 0xFC, 0x7E44, 0x4248, 0x7E28, 0x2410, 0x1E28, 0xE0C6}},
    {40728, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E28, 0x2418, 0xFEE6, 0x2448, 0x3C78, 0x2008, 0x3FF8, 0x2008, 0x3C78, 0x2448, 0x4448}},
    {40729, 2, {0x810, 0xFF7E, 0x810, 0x7F7C, 0x4144, 0x7F28, 0x2210, 0x1F28, 0xFFF6, 0x1110, 0x1FF0, 0x1210, 0x1FF0, 0x440, 0x7FFC, 0x40}},
    {40730, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4204, 0x7E48, 0x2438, 0x1E44, 0xEFF0, 0x800, 0xFF0, 0x800, 0xFFFE, 0x1110, 0x14E0, 0x181C}},
    {40731, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E38, 0x2444, 0x7EA2, 0x2238, 0x5424, 0x820, 0x37FE, 0xC000, 0x1FF0, 0x1010, 0x1FF0}},
    {40732, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E38, 0x2444, 0x7FF2, 0x920, 0x7C0, 0xF93E, 0x1FF0, 0x1110, 0x1FF0, 0x108, 0x7FFC}},
    {40733, 2, {0x24A4, 0x14A4, 0x87BC, 0x4404, 0x17FC, 0x2404, 0x4FBC, 0x90A4, 0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E28, 0x2418, 0xFEE6}},
    {40734, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E28, 0x2418, 0xFEE6, 0x1110, 0x7FFE, 0x4822, 0x8FE4, 0x100, 0x1FF0, 0x100, 0x7FFC}},
    {40735, 2, {0x810, 0xFF7E, 0x810, 0x7E7C, 0x4244, 0x7E38, 0x24D4, 0x7EA2, 0x444, 0x2BA8, 0x1010, 0x2FE8, 0xC826, 0xFE0, 0x440, 0x7FFC}},
    {40736, 2, {0x400, 0x3878, 0x2008, 0x3C78, 0x2008, 0x2008, 0x3FF8, 0x0, 0x2948, 0x2528, 0x2108, 0x2948, 0x252A, 0x210A, 0x2946, 0x3182}},
    {40737, 2, {0x2208, 0x1108, 0x1110, 0x0, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x210A, 0x410A, 0x4106, 0x8002}},
    {40738, 2, {0x2008, 0xCE28, 0x8228, 0xEE28, 0x8244, 0x8244, 0xFE82, 0x17C, 0x9224, 0x9224, 0xDA24, 0x9224, 0xDA44, 0x9244, 0x9394, 0xD908}},
    {40739, 2, {0x2010, 0xCE14, 0x8212, 0xEE12, 0x8210, 0x82FE, 0xFE10, 0x10, 0x9210, 0x9228, 0xDA28, 0x9228, 0xDA28, 0x9244, 0x9344, 0xD982}},
    {40740, 2, {0x2020, 0xCE10, 0x8210, 0xEEFE, 0x8244, 0x8244, 0xFE44, 0x44, 0x9244, 0x9228, 0xDA28, 0x9210, 0xDA10, 0x9228, 0x9344, 0xD982}},
    {40741, 2, {0x2028, 0xCE24, 0x8224, 0xEE20, 0x82FE, 0x8220, 0xFE20, 0x3C, 0x9234, 0x9254, 0xDA54, 0x9254, 0xDA88, 0x9288, 0x9314, 0xD922}},
    {40742, 2, {0x2000, 0xCEFC, 0x8224, 0xEE24, 0x8224, 0x8224, 0xFE54, 0x48, 0x9280, 0x927C, 0xDA44, 0x9244, 0xDA44, 0x9244, 0x937C, 0xD944}},
    {40743, 2, {0x2020, 0xCE10, 0x8210, 0xEEFE, 0x8282, 0x8284, 0xFE40, 0x44, 0x9248, 0x9250, 0xDA60, 0x9242, 0xDA42, 0x9242, 0x933E, 0xD900}},
    {40744, 2, {0x2020, 0xCE20, 0x823E, 0xEE42, 0x82C4, 0x8328, 0xFE10, 0x28, 0x9244, 0x9382, 0xDA30, 0x9208, 0xDA04, 0x9260, 0x9310, 0xD908}},
    {40745, 2, {0x2040, 0xCE40, 0x8240, 0xEEFC, 0x8284, 0x8304, 0xFEF4, 0x94, 0x9294, 0x9294, 0xDA94, 0x92F4, 0xDA94, 0x9204, 0x9328, 0xD910}},
    {40746, 2, {0x2010, 0xCE10, 0x8250, 0xEE50, 0x827C, 0x8250, 0xFE90, 0x10, 0x9210, 0x927C, 0xDA10, 0x9210, 0xDA10, 0x9210, 0x93FE, 0xD900}},
    {40747, 2, {0x2000, 0xCE00, 0x82FE, 0xEE10, 0x8210, 0x8220, 0xFE20, 0x7C, 0x9244, 0x92C4, 0xDB44, 0x9244, 0xDA44, 0x9244, 0x937C, 0xD944}},
    {40748, 2, {0x2010, 0xCE10, 0x8210, 0xEE10, 0x82FE, 0x8292, 0xFE92, 0x92, 0x9292, 0x92FE, 0xDA92, 0x9292, 0xDA92, 0x9292, 0x93FE, 0xD982}},
    {40749, 2, {0x2010, 0xCE10, 0x8210, 0xEE7C, 0x8210, 0x8210, 0xFEFE, 0x8, 0x9208, 0x92FE, 0xDA08, 0x9248, 0xDA28, 0x9208, 0x9328, 0xD910}},
    {40750, 2, {0x2000, 0xCC06, 0x85D8, 0xEC48, 0x8448, 0x8488, 0xFC9E, 0x1C8, 0xA848, 0xF948, 0xA948, 0xF89E, 0xA880, 0xA940, 0xFA3E, 0xA400}},
    {40751, 2, {0x2000, 0xCEFE, 0x8210, 0xEE10, 0x827C, 0x8224, 0xFE24, 0xFE, 0x9200, 0x9200, 0xDA7C, 0x9244, 0xDA44, 0x9244, 0x937C, 0xD944}},
    {40752, 2, {0x2000, 0xCE7C, 0x8244, 0xEE44, 0x827C, 0x8244, 0xFE44, 0x7C, 0x9244, 0x9244, 0xDA7C, 0x9200, 0xDA28, 0x9224, 0x9342, 0xD982}},
    {40753, 2, {0x2010, 0xCE10, 0x82FE, 0xEE10, 0x827C, 0x8210, 0xFEFE, 0x0, 0x927C, 0x9244, 0xDA7C, 0x9244, 0xDA7C, 0x9244, 0x9354, 0xD948}},
    {40754, 2, {0x2000, 0xCEFE, 0x8292, 0xEE10, 0x82FE, 0x8210, 0xFE7C, 0x54, 0x927C, 0x9254, 0xDA7C, 0x9210, 0xDAFE, 0x9210, 0x9310, 0xD910}},
    {40755, 2, {0x2000, 0xCEFE, 0x8282, 0xEEFE, 0x8282, 0x82FE, 0xFE82, 0xFE, 0x9214, 0x9212, 0xDAFE, 0x9210, 0xDA28, 0x9228, 0x9344, 0xD982}},
    {40756, 2, {0x2000, 0xCDFE, 0x8500, 0xED7E, 0x8542, 0x857E, 0xFD42, 0x17E, 0xA910, 0xF9FE, 0xA924, 0xF964, 0xA918, 0xA924, 0xFB42, 0xA5FE}},
    {40757, 2, {0x2020, 0xCE10, 0x82FE, 0xEE82, 0x8244, 0x8292, 0xFE14, 0x12, 0x92FE, 0x9210, 0xDA28, 0x9228, 0xDA28, 0x9244, 0x9344, 0xD982}},
    {40758, 2, {0x2008, 0xCC3C, 0x85D0, 0xED1C, 0x8510, 0x857C, 0xFD54, 0x158, 0xA970, 0xF954, 0xA94C, 0xF940, 0xAA78, 0xA8A8, 0xFA2A, 0xA446}},
    {40759, 2, {0x200E, 0xCEF0, 0x8222, 0xEE92, 0x8244, 0x8220, 0xFE44, 0xF8, 0x9210, 0x9224, 0xDAFE, 0x9210, 0xDAFE, 0x9228, 0x9344, 0xD982}},
    {40760, 2, {0x2088, 0xCC50, 0x8400, 0xEDFE, 0x8450, 0x85FC, 0xFC54, 0x3FE, 0xA854, 0xF9FC, 0xA850, 0xF8D8, 0xA954, 0xAA52, 0xFA50, 0xA450}},
    {40761, 2, {0x2000, 0xCE7C, 0x8244, 0xEE7C, 0x8244, 0x827C, 0xFE10, 0xFE, 0x9282, 0x9210, 0xDAFE, 0x9224, 0xDA68, 0x9210, 0x9328, 0xD944}},
    {40762, 2, {0x2000, 0xCDFC, 0x8524, 0xEDFC, 0x8524, 0x85FC, 0xFD04, 0x0, 0xABFE, 0xFAAA, 0xAAAA, 0xFBFE, 0xAAAA, 0xAAAA, 0xFBFE, 0xA622}},
    {40763, 2, {0x200, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0xFFFE, 0x820, 0x1020}},
    {40764, 2, {0x1004, 0x3E04, 0x2244, 0x3E44, 0x2244, 0x3E44, 0x44, 0x7F44, 0x4944, 0x7F4C, 0x4974, 0x7F44, 0x4, 0xFF84, 0x2204, 0x4204}},
    {40765, 2, {0x1020, 0x3E20, 0x2220, 0x3E20, 0x22F8, 0x3E28, 0x28, 0x7F28, 0x4928, 0x7F28, 0x4928, 0x7F2A, 0x2A, 0xFFCA, 0x2246, 0x4280}},
    {40766, 2, {0x1000, 0x3E00, 0x227C, 0x3E10, 0x2210, 0x3E10, 0x10, 0x7F10, 0x49FE, 0x7F10, 0x4910, 0x7F10, 0x10, 0xFF90, 0x2210, 0x4210}},
    {40767, 2, {0x1000, 0x3E00, 0x227C, 0x3E28, 0x2228, 0x3E28, 0x28, 0x7F28, 0x4928, 0x7F28, 0x4928, 0x7F2A, 0x2A, 0xFFAA, 0x2246, 0x4280}},
    {40768, 2, {0x1000, 0x3E7C, 0x2220, 0x3E20, 0x2220, 0x3E38, 0x28, 0x7F28, 0x4958, 0x7F48, 0x4948, 0x7F48, 0x5A, 0xFFEA, 0x224A, 0x4206}},
    {40769, 2, {0x1040, 0x3E40, 0x2240, 0x3EFC, 0x2284, 0x3F04, 0x74, 0x7F54, 0x4954, 0x7F54, 0x4974, 0x7F54, 0x4, 0xFF84, 0x2228, 0x4210}},
    {40770, 2, {0x2020, 0x7C20, 0x44FC, 0x7C24, 0x45FE, 0x7C24, 0xFC, 0xFE20, 0x9324, 0xFEA8, 0x9270, 0xFEA8, 0x124, 0xFE22, 0x44A0, 0x8440}},
    {40771, 2, {0x2000, 0x7CFC, 0x4484, 0x7CFC, 0x4484, 0x7CFC, 0x40, 0xFEFE, 0x9312, 0xFE92, 0x92AA, 0xFE82, 0xFA, 0xFE02, 0x4414, 0x8408}},
    {40772, 2, {0x1010, 0x3E10, 0x22FE, 0x3E38, 0x2254, 0x3E92, 0x0, 0x7F7C, 0x4944, 0x7F7C, 0x4944, 0x7F7C, 0x44, 0xFF80, 0x22FE, 0x4200}},
    {40773, 2, {0x2020, 0x7C40, 0x44FC, 0x7C84, 0x44FC, 0x7C84, 0xFC, 0xFE84, 0x92FC, 0xFE28, 0x9224, 0xFEFE, 0x20, 0xFE50, 0x4488, 0x8506}},
    {40774, 2, {0x2092, 0x7D24, 0x4492, 0x7C00, 0x447C, 0x7C44, 0x7C, 0xFE00, 0x92FE, 0xFE92, 0x9292, 0xFEFE, 0x80, 0xFE82, 0x4482, 0x847E}},
    {40775, 2, {0x2020, 0x7C3C, 0x4420, 0x7CFE, 0x44A2, 0x7CB8, 0xE4, 0xFE9C, 0x9280, 0xFEBC, 0x92A4, 0xFEBC, 0xA4, 0xFEBC, 0x44A4, 0x857E}},
    {40776, 2, {0x2028, 0x7CFE, 0x44AA, 0x7CFE, 0x44AA, 0x7CFE, 0x0, 0xFEFE, 0x9280, 0xFEBE, 0x9280, 0xFEFE, 0xA8, 0xFEAA, 0x44A4, 0x8532}},
    {40777, 2, {0x2020, 0xFBFE, 0x8924, 0xF9FC, 0x8820, 0xFBFE, 0x356, 0xF9DC, 0xA888, 0xFBFE, 0xA888, 0xFBFE, 0x94, 0xF988, 0x52A4, 0x90C2}},
    {40778, 2, {0x200, 0x100, 0x7FFC, 0x440, 0x28C, 0x7D70, 0x2550, 0x2550, 0x4548, 0x8D66, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x2010}},
    {40779, 2, {0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2108, 0x2548, 0x2528, 0x4928, 0x8108}},
    {40780, 2, {0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x2008, 0x2528, 0x2948, 0x4288, 0x8C48}},
    {40781, 2, {0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x2008, 0x3FF8, 0x2008, 0x3FF8, 0x4008, 0x3FF8, 0x2448, 0x2448, 0xFFFE}},
    {40782, 2, {0x100, 0xFFFE, 0x284, 0x7D78, 0x2550, 0x2548, 0x4D64, 0x2008, 0x3FF8, 0x2008, 0x2FE8, 0x2828, 0x2FE8, 0x2828, 0x4FE8, 0x8448}},
    {40783, 2, {0x100, 0xFFFE, 0x288, 0x7D70, 0x2548, 0x4D66, 0xA008, 0x3FF8, 0x2288, 0x2EE8, 0x2288, 0x2EE8, 0x2288, 0x4EE8, 0x4288, 0x8FE8}},
    {40784, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0xC60, 0x3018, 0xC826, 0x820, 0x820, 0x820, 0x820, 0x1020, 0x1020, 0x2020}},
    {40785, 2, {0x200, 0x100, 0x7FFC, 0x820, 0x440, 0x380, 0x1C70, 0xE28E, 0x2EE8, 0x2288, 0x2EE8, 0x2288, 0x2EE8, 0x2288, 0x5FF8, 0x8008}},
    {40786, 2, {0x100, 0x1100, 0x11F8, 0x1100, 0x1100, 0xFFFE, 0x440, 0x2448, 0x2AA8, 0x3118, 0x3FF8, 0x2448, 0x2AA8, 0x3118, 0x3FF8, 0x8}},
    {40787, 2, {0x820, 0x2820, 0x2F20, 0x2820, 0xFF20, 0x1420, 0x5520, 0x6B20, 0x4120, 0x7F20, 0x5520, 0x5522, 0x6B22, 0x4F22, 0x711E, 0x100}},
    {40788, 2, {0x820, 0x2820, 0x2F20, 0x2820, 0xFF22, 0x1424, 0x5528, 0x6B30, 0x4120, 0x7F60, 0x55A0, 0x5522, 0x6B22, 0x4F22, 0x711E, 0x100}},
    {40789, 2, {0x820, 0x2820, 0x2F20, 0x287E, 0xFF40, 0x1480, 0x553C, 0x6B04, 0x4108, 0x7F10, 0x5520, 0x5540, 0x6B42, 0x4F42, 0x713E, 0x100}},
    {40790, 2, {0x1000, 0x50FC, 0x5C08, 0x5048, 0xFE48, 0x2848, 0xAA88, 0xD6FE, 0x8218, 0xFE28, 0xAA28, 0xAA48, 0xD688, 0x9F08, 0xE228, 0x210}},
    {40791, 2, {0x800, 0x280C, 0x2F70, 0x2840, 0xFF40, 0x1440, 0x557E, 0x6B48, 0x4148, 0x7F48, 0x5548, 0x5548, 0x6B48, 0x4F48, 0x7148, 0x188}},
    {40792, 2, {0x810, 0x2810, 0x2F10, 0x2828, 0xFF28, 0x1444, 0x55AA, 0x6B28, 0x4128, 0x7F28, 0x5528, 0x5528, 0x6B28, 0x4F48, 0x7148, 0x188}},
    {40793, 2, {0x1040, 0x5040, 0x5CFC, 0x5084, 0xFF04, 0x2AF4, 0xAA94, 0xD694, 0x82F4, 0xFE84, 0xAA94, 0xAA88, 0xD682, 0x9E82, 0xE27E, 0x200}},
    {40794, 2, {0x820, 0x2820, 0x2F20, 0x283E, 0xFF50, 0x1450, 0x5590, 0x6B1E, 0x4110, 0x7F10, 0x5510, 0x551E, 0x6B10, 0x4F10, 0x7110, 0x110}},
    {40795, 2, {0x1028, 0x50A8, 0x5CA8, 0x50A8, 0xFEA8, 0x29FE, 0xAAA8, 0xD6A8, 0x82A8, 0xFEA8, 0xAAB8, 0xAA80, 0xD680, 0x9E80, 0xE2FE, 0x200}},
    {40796, 2, {0x1008, 0x5028, 0x5C28, 0x5028, 0xFEA8, 0x28AA, 0xAABC, 0xD6A8, 0x82A8, 0xFEA8, 0xAAA8, 0xAAA8, 0xD6AA, 0x9EBA, 0xE2EA, 0x386}},
    {40797, 2, {0x1020, 0x5020, 0x5C20, 0x5040, 0xFE48, 0x2884, 0xABFE, 0xD682, 0x8200, 0xFEFC, 0xAA84, 0xAA84, 0xD684, 0x9E84, 0xE2FC, 0x284}},
    {40798, 2, {0x800, 0x2800, 0x2F7C, 0x2844, 0xFF44, 0x1444, 0x5544, 0x6B44, 0x417C, 0x7F44, 0x5500, 0x5528, 0x6B24, 0x4F44, 0x7142, 0x182}},
    {40799, 2, {0x800, 0x287C, 0x2F44, 0x2844, 0xFF44, 0x147C, 0x5544, 0x6B44, 0x4144, 0x7F7C, 0x5544, 0x5544, 0x6B44, 0x4F44, 0x71FE, 0x100}},
    {40800, 2, {0x800, 0x28FC, 0x2F24, 0x2824, 0xFF24, 0x1424, 0x5554, 0x6B48, 0x4180, 0x7F7C, 0x5544, 0x5544, 0x6B44, 0x4F44, 0x717C, 0x144}},
    {40801, 2, {0x810, 0x2810, 0x2F10, 0x2828, 0xFF28, 0x1454, 0x558A, 0x6B08, 0x417C, 0x7F04, 0x5508, 0x5528, 0x6B10, 0x4F10, 0x7108, 0x108}},
    {40802, 2, {0x810, 0x2810, 0x2E10, 0x2828, 0x2828, 0xFF54, 0x8A, 0x4908, 0x6B7C, 0x4904, 0x7F08, 0x5D28, 0x6B10, 0x4F10, 0x7908, 0x8}},
    {40803, 2, {0x820, 0x2820, 0x2F20, 0x287C, 0xFF44, 0x1484, 0x5574, 0x6B54, 0x4154, 0x7F54, 0x5574, 0x5554, 0x6B04, 0x4F04, 0x7128, 0x110}},
    {40804, 2, {0x1110, 0x920, 0x7FFC, 0x400, 0xFFFE, 0x1110, 0x29E8, 0xC906, 0x7FFC, 0x440, 0x2AA8, 0x3FF8, 0x2448, 0x2AA8, 0x3FF8, 0x8}},
    {40805, 2, {0x1020, 0x5020, 0x5CFC, 0x50A4, 0xFEA4, 0x28FC, 0xAAA4, 0xD6A4, 0x82FC, 0xFE20, 0xAA24, 0xAA18, 0xD612, 0x9E6A, 0xE386, 0x202}},
    {40806, 2, {0x800, 0x87C, 0x2F44, 0x2844, 0xFF7C, 0x1444, 0x5544, 0x6B7C, 0x4150, 0x7F52, 0x5554, 0x5548, 0x6B54, 0x4166, 0x7F42, 0x0}},
    {40807, 2, {0x800, 0x7F7C, 0x824, 0x3E24, 0x844, 0x7F54, 0x9E8, 0x900, 0x7FFC, 0x440, 0x2AA8, 0x3FF8, 0x2448, 0x2AA8, 0x3FF8, 0x8}},
    {40808, 2, {0x800, 0x2820, 0x2F4C, 0x2844, 0xFF44, 0x1444, 0x5544, 0x6B6C, 0x4144, 0x7F44, 0x5544, 0x5544, 0x6B44, 0x4F7C, 0x7144, 0x100}},
    {40809, 2, {0x1020, 0x5010, 0x5C10, 0x50FE, 0xFE00, 0x2844, 0xAA82, 0xD700, 0x8244, 0xFE44, 0xAA28, 0xAA28, 0xD610, 0x9E28, 0xE244, 0x282}},
    {40810, 2, {0x800, 0x287C, 0x2F44, 0x2844, 0xFF44, 0x1444, 0x557C, 0x6B10, 0x4110, 0x7F50, 0x555E, 0x5550, 0x6B50, 0x4F70, 0x719E, 0x100}},
    {40811, 2, {0x1000, 0x50FE, 0x5C92, 0x5092, 0xFE92, 0x28FE, 0xAA92, 0xD6B2, 0x82BA, 0xFED6, 0xAA92, 0xAA92, 0xD692, 0x9E82, 0xE2FE, 0x282}},
    {40812, 2, {0x800, 0x28FC, 0x2F10, 0x2810, 0xFF7C, 0x1424, 0x5524, 0x6BFE, 0x4100, 0x7F00, 0x557C, 0x5544, 0x6B44, 0x4F44, 0x717C, 0x144}},
    {40813, 2, {0x1000, 0x5032, 0x5CCC, 0x5088, 0xFE88, 0x28E8, 0xAAAE, 0xD6AA, 0x82AA, 0xFEEA, 0xAA8A, 0xAA8A, 0xD68A, 0x9E8A, 0xE312, 0x222}},
    {40814, 2, {0x1020, 0x5020, 0x5DFC, 0x5020, 0xFE50, 0x2888, 0xABFE, 0xD608, 0x82E8, 0xFEA8, 0xAAA8, 0xAAE8, 0xD6A8, 0x9E08, 0xE228, 0x210}},
    {40815, 2, {0x1000, 0x5020, 0x5CCE, 0x5082, 0xFE82, 0x28EE, 0xAA82, 0xD682, 0x82FE, 0xFE28, 0xAA28, 0xAA28, 0xD648, 0x9E4A, 0xE28A, 0x306}},
    {40816, 2, {0x1048, 0x5048, 0x5C48, 0x51FE, 0xFE48, 0x2848, 0xABFE, 0xD600, 0x82FC, 0xFE84, 0xAA84, 0xAAFC, 0xD684, 0x9E84, 0xE2FC, 0x284}},
    {40817, 2, {0x1000, 0x51F0, 0x5CAE, 0x50AA, 0xFEEA, 0x28AA, 0xAAAA, 0xD6EA, 0x82AA, 0xFEAA, 0xAAB4, 0xABE4, 0xD624, 0x9E2A, 0xE22A, 0x230}},
    {40818, 2, {0x100E, 0x51F0, 0x5C10, 0x50FE, 0xFE92, 0x2892, 0xAAFE, 0xD610, 0x82FE, 0xFE92, 0xAA9A, 0xAA96, 0xD6FE, 0x9E82, 0xE28A, 0x284}},
    {40819, 2, {0x1000, 0x50FE, 0x5C92, 0x5010, 0xFEFE, 0x2810, 0xAA7C, 0xD654, 0x827C, 0xFE54, 0xAA7C, 0xAA10, 0xD6FE, 0x9E10, 0xE210, 0x210}},
    {40820, 2, {0x1020, 0x5010, 0x5CFE, 0x5044, 0xFE28, 0x28FE, 0xAA88, 0xD690, 0x82A4, 0xFE88, 0xAA90, 0xAAA2, 0xD684, 0x9E88, 0xE310, 0x260}},
    {40821, 2, {0x1000, 0x507C, 0x5C54, 0x5054, 0xFE7C, 0x2854, 0xAA54, 0xD67C, 0x8210, 0xFEFE, 0xAA92, 0xAA96, 0xD6BE, 0x9E82, 0xE28A, 0x284}},
    {40822, 2, {0x1000, 0x50EE, 0x5CAA, 0x50AA, 0xFEEE, 0x2800, 0xAA7C, 0xD600, 0x82FE, 0xFE20, 0xAA40, 0xAA7C, 0xD604, 0x9E04, 0xE228, 0x210}},
    {40823, 2, {0x1000, 0x50FE, 0x5C82, 0x5082, 0xFEFE, 0x2880, 0xAABE, 0xD690, 0x8294, 0xFEBE, 0xAA8A, 0xAA88, 0xD6BE, 0x9E88, 0xE308, 0x27E}},
    {40824, 2, {0x1044, 0x5028, 0x5C00, 0x50FE, 0xFE00, 0x2828, 0xAA44, 0xD682, 0x8200, 0xFEFE, 0xAAAA, 0xAAAA, 0xD6AA, 0x9EAA, 0xE3FE, 0x200}},
    {40825, 2, {0x820, 0x440, 0x7FFC, 0x100, 0xFFFE, 0x400, 0x1FF8, 0xE080, 0x1FFC, 0x100, 0x9F0, 0x900, 0x7FFC, 0x2448, 0x2AA8, 0x3FF8}},
    {40826, 2, {0x1040, 0x5040, 0x5CFE, 0x5122, 0xFEAA, 0x28FA, 0xAA42, 0xD694, 0x8348, 0xFEFE, 0xAB22, 0xAAAA, 0xD6FA, 0x9E42, 0xE294, 0x308}},
    {40827, 2, {0x1010, 0x5010, 0x5CFE, 0x5010, 0xFE7C, 0x2844, 0xAA7C, 0xD644, 0x827C, 0xFE44, 0xAA7C, 0xAA44, 0xD6FE, 0x9E28, 0xE244, 0x282}},
    {40828, 2, {0x1088, 0x5088, 0x5DFE, 0x5088, 0xFEDC, 0x29AA, 0xAA88, 0xD600, 0x83FE, 0xFE22, 0xAAA0, 0xAABC, 0xD6A0, 0x9F60, 0xE23E, 0x200}},
    {40829, 2, {0x1044, 0x5044, 0x5DEE, 0x5044, 0xFEEE, 0x2954, 0xAA44, 0xD600, 0x827C, 0xFE00, 0xAAFE, 0xAA10, 0xD654, 0x9E92, 0xE250, 0x220}},
    {40830, 2, {0xF14, 0x812, 0x7F10, 0x497E, 0x7E10, 0x4928, 0x7F28, 0x5244, 0x9EFA, 0x480, 0x7FFE, 0x440, 0x2AA4, 0x3114, 0x3FFC, 0x4}},
    {40831, 2, {0x100, 0x100, 0x11F8, 0x1100, 0x1100, 0x1100, 0xFFFE, 0x0, 0x2108, 0x2108, 0x2288, 0x2448, 0x2828, 0x2008, 0x3FF8, 0x8}},
    {40832, 2, {0x820, 0x820, 0x2820, 0x2F22, 0x2822, 0x2824, 0xFF28, 0x30, 0x4920, 0x4960, 0x55A0, 0x6322, 0x4122, 0x4722, 0x791E, 0x100}},
    {40833, 2, {0x820, 0x820, 0x2820, 0x2F7E, 0x2840, 0x2880, 0xFF3C, 0x4, 0x4908, 0x4910, 0x5520, 0x6340, 0x4142, 0x4742, 0x793E, 0x100}},
    {40834, 2, {0x800, 0x80C, 0x2870, 0x2F40, 0x2840, 0x2840, 0xFF7E, 0x48, 0x4948, 0x4948, 0x5548, 0x6348, 0x4148, 0x4748, 0x7948, 0x188}},
    {40835, 2, {0x800, 0x87C, 0x2844, 0x2F44, 0x2844, 0x287C, 0xFF44, 0x44, 0x4944, 0x497C, 0x5544, 0x6344, 0x4144, 0x4744, 0x79FE, 0x100}},
    {40836, 2, {0x1020, 0x1020, 0x5050, 0x5E50, 0x5088, 0x5124, 0xFE12, 0x10, 0x93FC, 0x9204, 0xAA08, 0xC688, 0x8250, 0x8E20, 0xF210, 0x210}},
    {40837, 2, {0x1040, 0x1040, 0x507C, 0x5E84, 0x5084, 0x51F4, 0xFE94, 0x94, 0x9294, 0x92F4, 0xAA84, 0xC6A8, 0x8292, 0x8E82, 0xF27E, 0x200}},
    {40838, 2, {0x800, 0x8FE, 0x2822, 0x2F22, 0x2822, 0x2822, 0xFF4A, 0x44, 0x4980, 0x497E, 0x5542, 0x6342, 0x4142, 0x4742, 0x797E, 0x142}},
    {40839, 2, {0x1008, 0x1028, 0x5028, 0x5E28, 0x50A8, 0x50AA, 0xFEBC, 0xA8, 0x92A8, 0x92A8, 0xAAA8, 0xC6A8, 0x82AA, 0x8EBA, 0xF2EA, 0x386}},
    {40840, 2, {0x1000, 0x10FC, 0x5084, 0x5E84, 0x50FC, 0x5084, 0xFE84, 0xFC, 0x92A2, 0x92A4, 0xAA98, 0xC690, 0x8288, 0x8EA4, 0xF2C2, 0x280}},
    {40841, 2, {0x1000, 0x11FC, 0x5020, 0x5E20, 0x50FC, 0x5044, 0xFE44, 0x1FE, 0x9200, 0x9200, 0xAAFC, 0xC684, 0x8284, 0x8E84, 0xF2FC, 0x284}},
    {40842, 2, {0x800, 0x87C, 0x2844, 0x2F44, 0x2844, 0x2844, 0xFF7C, 0x10, 0x4910, 0x4950, 0x555E, 0x6350, 0x4150, 0x4770, 0x799E, 0x100}},
    {40843, 2, {0x100E, 0x11F0, 0x5010, 0x5EFE, 0x5092, 0x5092, 0xFEFE, 0x10, 0x92FE, 0x9292, 0xAA9A, 0xC696, 0x82FE, 0x8E82, 0xF28A, 0x284}},
    {40844, 2, {0x1000, 0x10FE, 0x5082, 0x5E82, 0x50FE, 0x5080, 0xFEFE, 0x90, 0x92A4, 0x92FE, 0xAA92, 0xC690, 0x82FC, 0x8E90, 0xF310, 0x2FE}},
    {40845, 2, {0x1040, 0x840, 0x7E7E, 0x2440, 0x187E, 0xFF02, 0x7E, 0x7E40, 0x427E, 0x7E40, 0x427E, 0x7E40, 0x427E, 0x4240, 0x4A42, 0x443E}},
    {40846, 2, {0x0, 0x3FFE, 0x2220, 0x2FBC, 0x2520, 0x223C, 0x3F84, 0x203C, 0x2FA0, 0x28BC, 0x2FA0, 0x28BC, 0x2FA0, 0x48BC, 0x48A2, 0x899E}},
    {40847, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E78, 0x2242, 0x263E, 0x820, 0x820, 0xFFFE, 0x820, 0x1020, 0x2020}},
    {40848, 2, {0x80, 0x7FFE, 0x4220, 0x5FBC, 0x4920, 0x463C, 0x7F84, 0x403C, 0x5FA0, 0x50BC, 0x5FA0, 0x50BC, 0x5FA0, 0x50BC, 0x52A2, 0x911E}},
    {40849, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E78, 0x2242, 0x263E, 0x0, 0x3FF8, 0x200, 0xFFFE, 0x860, 0x701C}},
    {40850, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E7C, 0x2242, 0x263E, 0x0, 0x3FF8, 0x0, 0xFFFE, 0x1110, 0x2308}},
    {40851, 2, {0x2110, 0x2090, 0xFBDE, 0x2010, 0x425E, 0x7982, 0x4BDE, 0xC810, 0x7BDE, 0x4A50, 0x4BDE, 0x7A50, 0x4BDE, 0x4A50, 0x4A52, 0x5ACE}},
    {40852, 2, {0x840, 0x7F7C, 0x1440, 0xFF7C, 0x2204, 0x3E7C, 0x2240, 0x3E78, 0x2242, 0x263E, 0x440, 0x3FF8, 0x440, 0xFFFE, 0x820, 0x1010}},
    {40853, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x840, 0x3E7C, 0x1440, 0xFF78, 0x2208, 0x3E78, 0x2242, 0x263E}},
    {40854, 2, {0x8888, 0x4E4E, 0xF8F8, 0xA8A8, 0x4E4E, 0xF2F2, 0xE0E, 0xE8E8, 0xAEAE, 0xA8A8, 0xEEEE, 0xA8A8, 0xEEEE, 0xA8A8, 0xAAAA, 0xA6A6}},
    {40855, 2, {0x3FF8, 0x100, 0x7FFE, 0x4922, 0x8414, 0x3BB8, 0x2AA8, 0x3BB8, 0x840, 0x3E7C, 0x1440, 0xFF78, 0x2208, 0x3E78, 0x2242, 0x263E}},
    {40856, 2, {0x840, 0x3E7C, 0x1440, 0xFF78, 0x2208, 0x3E78, 0x2242, 0x487E, 0xE8E8, 0xAEAE, 0xF8F8, 0xE0E, 0xE2E2, 0xAEAE, 0xE8E8, 0xAEAE}},
    {40857, 2, {0x420, 0x410, 0x410, 0x400, 0xFFFE, 0x480, 0x488, 0x488, 0x490, 0x8A0, 0x8C0, 0x1082, 0x1182, 0x2282, 0x447E, 0x8000}},
    {40858, 2, {0x420, 0x410, 0x7FFC, 0x920, 0x9C4, 0x1704, 0x20FC, 0xC440, 0x440, 0x3FF8, 0x440, 0x440, 0xFFFE, 0x820, 0x1010, 0x2008}},
    {40859, 2, {0x100, 0x6C0, 0x1830, 0xEFEE, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x420, 0xFFFE, 0x910, 0x1160, 0x2184, 0x4F04, 0x80FC}},
    {40860, 2, {0x400, 0x3FF8, 0x2008, 0x3EF8, 0x280, 0x7A80, 0xAFC, 0xFE94, 0xAD4, 0x7AA4, 0x2A4, 0x7AD4, 0xA94, 0xFEFC, 0xA82, 0x79FE}},
    {40861, 2, {0x1080, 0xE7FC, 0x2404, 0x27BC, 0x20A0, 0xFEBE, 0x22A2, 0x37B6, 0x6AAA, 0x66AA, 0xA0B6, 0x26A2, 0x22BE, 0x27A0, 0x22A2, 0x267E}},
    {40862, 2, {0x4920, 0x2A3E, 0x7F48, 0x5DA8, 0x6B10, 0x456E, 0x3FF8, 0x2008, 0x3EF8, 0x280, 0x7AF8, 0xAD8, 0xFEA8, 0xAD8, 0x7A8A, 0x1FE}},
    {40863, 2, {0x400, 0x400, 0xFE0, 0x1020, 0x2040, 0x7FF8, 0xA108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x102, 0x102, 0xFE}},
    {40864, 2, {0x100, 0x280, 0xC60, 0x3018, 0xDFF6, 0x0, 0x3BB8, 0x2AA8, 0x3BB8, 0x0, 0x3FF8, 0x2488, 0x3FF8, 0x2488, 0x24A8, 0x2010}},
    {40865, 2, {0x810, 0x810, 0x1410, 0x2210, 0x5D3E, 0x80A2, 0x7F44, 0x5590, 0x7F10, 0x10, 0x7F10, 0x5528, 0x7F28, 0x5524, 0x5544, 0x4382}},
    {40866, 2, {0x808, 0x81C, 0x14F0, 0x2210, 0x5D10, 0x8090, 0x7F7C, 0x5510, 0x7F38, 0x34, 0x7F54, 0x5554, 0x7F92, 0x5510, 0x5510, 0x4310}},
    {40867, 2, {0x800, 0x8FC, 0x1404, 0x2204, 0x5D7C, 0x8084, 0x7F04, 0x55FE, 0x7F10, 0x92, 0x7F54, 0x5538, 0x7F54, 0x5592, 0x5550, 0x4320}},
    {40868, 2, {0x848, 0x84A, 0x146C, 0x2248, 0x5D4A, 0x80EA, 0x7F56, 0x5520, 0x7F7E, 0x42, 0x7F42, 0x557E, 0x7F42, 0x5542, 0x557E, 0x4342}},
    {40869, 2, {0x800, 0x87E, 0x1408, 0x2210, 0x5D7C, 0x80C4, 0x7F44, 0x557C, 0x7F44, 0x7C, 0x7F44, 0x5544, 0x7F7C, 0x5528, 0x5544, 0x4382}},
    {40870, 2, {0x1124, 0x10A4, 0x10A8, 0x13FE, 0x1602, 0x5AFA, 0x5088, 0x90F8, 0x1020, 0x11FC, 0x1124, 0x2924, 0x2524, 0x452C, 0x4020, 0x8020}},
    {40871, 2, {0x1FC, 0x104, 0x79FC, 0x4904, 0x49FC, 0x4844, 0x79F8, 0x4850, 0x4FFF, 0x4880, 0x4BFC, 0x7D04, 0x49FC, 0x104, 0x1FC, 0x104}},
    {40872, 2, {0x0, 0x3FFC, 0x2280, 0x2440, 0x2FFC, 0x3440, 0x2440, 0x27F8, 0x2440, 0x27F8, 0x2440, 0x2440, 0x27FC, 0x2400, 0x3FFC, 0x0}},
    {40873, 2, {0x420, 0x420, 0xFFFF, 0x420, 0x240C, 0xFF70, 0x2440, 0x3C40, 0x87F, 0x7E48, 0x4A48, 0x7E48, 0x848, 0xFF88, 0x888, 0x908}},
    {40874, 2, {0x820, 0x820, 0x820, 0x7E20, 0x820, 0x820, 0xFF20, 0x850, 0x850, 0xE50, 0x2848, 0x2888, 0x2884, 0x5B02, 0x4800, 0x87FF}},
    {40875, 2, {0x1010, 0x1010, 0xFF10, 0x1028, 0x7E44, 0x5282, 0x537C, 0x7E10, 0x5210, 0x5210, 0x7EFE, 0x1010, 0xFF10, 0x1010, 0x1010, 0x1010}},
    {40876, 2, {0x1000, 0x13FE, 0x1A02, 0x268A, 0x2252, 0x7FFE, 0x9242, 0x1222, 0x7FFE, 0x1282, 0x1682, 0x5A82, 0x32FA, 0x1E02, 0xE20A, 0x204}},
    {40877, 2, {0x7DFE, 0x5122, 0x7DFE, 0x5122, 0x5122, 0x7DFE, 0x5048, 0x5048, 0x7EFC, 0x248, 0x5648, 0x53FE, 0x8200, 0x248, 0x1484, 0x902}},
    {40878, 2, {0x7FDE, 0x514A, 0x7CC6, 0x514A, 0x5252, 0x7CFC, 0x50A4, 0x50FC, 0x7EA4, 0x2FC, 0x5648, 0x52FE, 0x8248, 0x3FE, 0x1484, 0x902}},
    {40879, 2, {0x1028, 0x1024, 0x2820, 0x25FE, 0x4120, 0xB920, 0x1124, 0x1124, 0x7D28, 0x1118, 0x1510, 0x5910, 0x312A, 0x1E2A, 0xF246, 0x4482}},
    {40880, 2, {0x0, 0x0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x11F0, 0x1100, 0x1100, 0x1100, 0x1100, 0x2D00, 0x4300, 0x80FC}},
    {40881, 2, {0x0, 0x7FFC, 0x4004, 0x4024, 0x4824, 0x4424, 0x4244, 0x4144, 0x4084, 0x4144, 0x4224, 0x4414, 0x5814, 0x4004, 0x7FFC, 0x4004}},
    {40882, 2, {0x1020, 0x1010, 0x29FE, 0x2510, 0x41FE, 0xBD10, 0x117C, 0x1154, 0x7D7C, 0x1154, 0x157C, 0x5910, 0x31FE, 0x1E10, 0xE210, 0x4410}},
    {40883, 2, {0x1010, 0x1010, 0x7EFC, 0x1010, 0x1010, 0xFEFE, 0x2448, 0x5294, 0x1010, 0x7E7C, 0x1010, 0x1010, 0x1010, 0x1E10, 0xE1FE, 0x0}},
    {40884, 2, {0x0, 0x0, 0x0, 0x0, 0x7FFC, 0x8, 0x10, 0x320, 0xC0, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40885, 2, {0xF8, 0x7F00, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x200, 0x400, 0x800, 0x3000, 0xC000, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40886, 2, {0x100, 0x100, 0x7FFC, 0x100, 0x3FF8, 0x100, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40887, 2, {0x820, 0x820, 0x7FFC, 0x820, 0x820, 0xFFFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40888, 2, {0x100, 0x1110, 0x920, 0x100, 0x7FFE, 0x4004, 0x8008, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40889, 2, {0x1110, 0x920, 0x100, 0x3FFC, 0x100, 0x200, 0x7FFE, 0x820, 0x1010, 0x2008, 0xC006, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40890, 2, {0x1000, 0x1000, 0xFE00, 0x1000, 0x7C00, 0x4400, 0x4400, 0x7C00, 0x4400, 0x7C00, 0x5400, 0x1000, 0xFE00, 0x1000, 0x1000, 0x1000}},
    {40891, 2, {0x2088, 0x27E8, 0x4812, 0xF3DC, 0x2008, 0x4BD2, 0xF83E, 0x3C0, 0xAA52, 0xABD2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {40892, 2, {0x40, 0x207C, 0x2040, 0x27FE, 0x2402, 0x23FC, 0xFA64, 0x2492, 0x2108, 0x23FC, 0x250A, 0x21F8, 0x3908, 0xC1F8, 0x108, 0x1F8}},
    {40893, 2, {0x1080, 0x1080, 0x10FC, 0x1108, 0x5610, 0x55FC, 0x5524, 0x9924, 0x11FC, 0x1124, 0x1124, 0x29FC, 0x2400, 0x4554, 0x412A, 0x822A}},
    {40894, 2, {0x840, 0x840, 0x87E, 0x7F42, 0x4AA2, 0x48AA, 0x4872, 0x7EFE, 0x4222, 0x6272, 0x54AA, 0x4D22, 0x8822, 0x9402, 0x220A, 0x4004}},
    {40895, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x440, 0x840, 0xBF8, 0x1040, 0x3040, 0x57FC, 0x9040, 0x13F8, 0x1040, 0x1040, 0x17FE, 0x1000}},
    {40896, 2, {0x420, 0xFFFE, 0x420, 0x7F78, 0x5048, 0x5E86, 0x6578, 0x4A48, 0x7F30, 0xC8, 0xFFFE, 0x440, 0x3FF8, 0x2448, 0x2BA8, 0x3FF8}},
    {40897, 2, {0x2090, 0x1090, 0x90, 0xFC90, 0x3FC, 0x7890, 0x90, 0x7890, 0x90, 0x7BFE, 0x4800, 0x4800, 0x4890, 0x7908, 0x4A04, 0x404}},
    {40898, 2, {0x1020, 0x1040, 0xFEFC, 0x1084, 0x7CFC, 0x5484, 0x7CFC, 0x5480, 0x54FE, 0x7C80, 0x54FE, 0x1002, 0xFEAA, 0x10AE, 0x1152, 0x100C}},
    {40899, 2, {0x40, 0x40, 0xF7FC, 0x9040, 0x9250, 0xF148, 0x9148, 0x92D4, 0xF460, 0x9040, 0x90A0, 0x90A0, 0xF110, 0x9208, 0x404, 0x802}},
    {40900, 2, {0x1000, 0xBF8, 0x4088, 0x2288, 0x988, 0x50C8, 0x2108, 0x2228, 0x2510, 0x100, 0xFFFE, 0x540, 0x920, 0x111C, 0x6108, 0x100}},
    {40901, 2, {0x21FC, 0x1020, 0x3FE, 0xFA22, 0x9AC, 0x1020, 0x118C, 0x3820, 0x5450, 0x9188, 0x1626, 0x11F8, 0x1008, 0x10D0, 0x1020, 0x1010}},
    {40902, 2, {0x4040, 0x2040, 0x40, 0xFBFE, 0x880, 0x1080, 0x20F8, 0x3088, 0x6948, 0xA150, 0x2150, 0x2220, 0x2220, 0x2450, 0x2888, 0x2104}},
    {40903, 2, {0x880, 0x8F8, 0x1110, 0x1220, 0x35FC, 0x5044, 0x9044, 0x17FE, 0x1044, 0x1044, 0x13FC, 0x1044, 0x1040, 0x1040, 0x1140, 0x1080}},
    {40904, 2, {0x800, 0x9FC, 0x1104, 0x2104, 0x4904, 0x9FC, 0x1124, 0x3120, 0x5120, 0x9110, 0x1110, 0x1110, 0x1208, 0x1208, 0x1404, 0x1802}},
    {40905, 2, {0x800, 0xBFE, 0x1040, 0x2040, 0x4840, 0x840, 0x13F8, 0x3048, 0x5048, 0x9088, 0x1088, 0x1088, 0x1088, 0x1088, 0x17FE, 0x1000}},
    {40906, 2, {0x820, 0x820, 0xFFFE, 0x820, 0x820, 0x4000, 0x2020, 0x9020, 0x4120, 0x293C, 0x1120, 0x1120, 0xE120, 0x2120, 0x2120, 0x27FE}},
    {40907, 2, {0x2040, 0x17FE, 0x40, 0xFBFC, 0x800, 0x1BF8, 0x1608, 0x3BF8, 0x5510, 0x9FFE, 0x1000, 0x13F8, 0x1208, 0x1208, 0x13F8, 0x1208}},
    {40908, 2, {0x80, 0x4040, 0x2FFE, 0x2000, 0x8000, 0x43F8, 0x4A08, 0xBF8, 0x1208, 0x13F8, 0xE040, 0x2248, 0x2244, 0x2442, 0x2942, 0x80}},
    {40960, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x7C10, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {40961, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1FF0, 0x1010, 0x7C10, 0x1210, 0x1210, 0x7C10, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {40962, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1FF0, 0x1010, 0x7C10, 0x1210, 0x1210, 0x7C10, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {40963, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0x50, 0x48, 0x40, 0x3BC0, 0x440, 0xA40, 0xA40, 0x440, 0x3BC0, 0x0, 0x0}},
    {40964, 2, {0x0, 0x0, 0x0, 0x0, 0x1390, 0x1450, 0xC60, 0x7C0, 0x440, 0x440, 0xE40, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {40965, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0x380, 0x440, 0x440, 0x380, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {40966, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x380, 0x440, 0x440, 0x380, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {40967, 2, {0x0, 0x0, 0x0, 0x0, 0x710, 0x8F0, 0x990, 0xA80, 0xA80, 0xA80, 0xA80, 0x990, 0x8F0, 0x710, 0x0, 0x0}},
    {40968, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x188, 0x3168, 0xD18, 0x300, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {40969, 2, {0x0, 0x380, 0x440, 0x0, 0x1A0, 0x160, 0x120, 0xD20, 0x320, 0x120, 0x120, 0x120, 0xD20, 0x320, 0x0, 0x0}},
    {40970, 2, {0x0, 0x0, 0x0, 0x0, 0x1A0, 0x160, 0x120, 0xD20, 0x320, 0x120, 0x120, 0x120, 0xD20, 0x320, 0x0, 0x0}},
    {40971, 2, {0x0, 0x0, 0x30, 0x40, 0x340, 0x2C0, 0x240, 0x1A30, 0x600, 0x200, 0x200, 0x200, 0x1A00, 0x600, 0x0, 0x0}},
    {40972, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1010, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x0, 0x0}},
    {40973, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1110, 0x1110, 0x1110, 0x1010, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x1110, 0x0, 0x0}},
    {40974, 2, {0x0, 0x0, 0x0, 0x420, 0xA50, 0xA50, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x420, 0x3C0, 0x0, 0x0}},
    {40975, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1110, 0x1110, 0xFE0, 0x100, 0x700, 0x800, 0x700, 0x800, 0x700, 0x0, 0x0}},
    {40976, 2, {0x0, 0x380, 0x440, 0x0, 0xC70, 0xA88, 0x988, 0x888, 0x8A8, 0x8A8, 0x8A8, 0x4888, 0x2888, 0x1870, 0x0, 0x0}},
    {40977, 2, {0x0, 0x0, 0x0, 0x0, 0xC70, 0xA88, 0x988, 0x888, 0x8A8, 0x8A8, 0x8A8, 0x4888, 0x2888, 0x1870, 0x0, 0x0}},
    {40978, 2, {0x0, 0x0, 0x0, 0x0, 0xC70, 0xA88, 0x988, 0x888, 0x888, 0x888, 0x888, 0x4888, 0x2888, 0x1870, 0x0, 0x0}},
    {40979, 2, {0x0, 0x380, 0x440, 0x0, 0x60, 0x3858, 0x440, 0x840, 0x1FC0, 0x840, 0x440, 0x3840, 0x340, 0xC0, 0x0, 0x0}},
    {40980, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0x3858, 0x440, 0x840, 0x1FC0, 0x840, 0x440, 0x3840, 0x340, 0xC0, 0x0, 0x0}},
    {40981, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x0, 0x440, 0x440, 0x440, 0x440, 0x0, 0x0}},
    {40982, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x550, 0x4E0, 0x4E0, 0x550, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {40983, 2, {0x0, 0x380, 0x440, 0x0, 0x10F0, 0x1100, 0x1200, 0x1200, 0x1200, 0x1FF0, 0x1200, 0x1200, 0x1100, 0x10F0, 0x0, 0x0}},
    {40984, 2, {0x0, 0x0, 0x0, 0x0, 0x10F0, 0x1100, 0x1200, 0x1200, 0x1200, 0x1FF0, 0x1200, 0x1200, 0x1100, 0x10F0, 0x0, 0x0}},
    {40985, 2, {0x0, 0x0, 0x0, 0x0, 0x620, 0x5A0, 0x460, 0x400, 0x4F0, 0x500, 0x7F0, 0x500, 0x4F0, 0x400, 0x0, 0x0}},
    {40986, 2, {0x0, 0x0, 0x0, 0x0, 0x1C18, 0x2220, 0x2240, 0x2240, 0x2240, 0x23F8, 0x2240, 0x2240, 0x2220, 0x1C18, 0x0, 0x0}},
    {40987, 2, {0x0, 0x380, 0x440, 0x0, 0x7D0, 0x820, 0x1050, 0x1090, 0x1110, 0x1110, 0x1210, 0x1410, 0x820, 0x7C0, 0x0, 0x0}},
    {40988, 2, {0x0, 0x0, 0x0, 0x0, 0x7D0, 0x820, 0x1050, 0x1090, 0x1110, 0x1110, 0x1210, 0x1410, 0x820, 0x7C0, 0x0, 0x0}},
    {40989, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x890, 0x888, 0x888, 0x890, 0x8E0, 0x890, 0x888, 0x888, 0x710, 0x0, 0x0}},
    {40990, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0x380, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {40991, 2, {0x0, 0x380, 0x440, 0x0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {40992, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {40993, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0x0, 0x0, 0x440, 0x820, 0x1010, 0x0, 0x0}},
    {40994, 2, {0x0, 0x380, 0x440, 0x0, 0x3E0, 0x410, 0x9D0, 0x90, 0x90, 0x1D0, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {40995, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x410, 0x9D0, 0x90, 0x90, 0x1D0, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {40996, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xFE0, 0x0, 0x0}},
    {40997, 2, {0x0, 0x0, 0x0, 0x0, 0xF0, 0x1700, 0x900, 0x1480, 0x1070, 0x1480, 0x900, 0x1700, 0x100, 0xF0, 0x0, 0x0}},
    {40998, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0xBA0, 0x920, 0x920, 0xBA0, 0x820, 0x820, 0x7C0, 0x0, 0x0}},
    {40999, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0xBA0, 0x920, 0x920, 0xBA0, 0x820, 0x820, 0x7C0, 0x0, 0x0}},
    {41000, 2, {0x0, 0x0, 0x0, 0x0, 0x1FF0, 0x100, 0x100, 0x1FF0, 0x0, 0x0, 0x1FF0, 0x100, 0x100, 0x1FF0, 0x0, 0x0}},
    {41001, 2, {0x0, 0x380, 0x440, 0x0, 0xEE0, 0x1110, 0x1110, 0x100, 0x100, 0x0, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41002, 2, {0x0, 0x0, 0x0, 0x0, 0xEE0, 0x1110, 0x1110, 0x100, 0x100, 0x0, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41003, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x100, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x100, 0xFE0, 0x0, 0x0}},
    {41004, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x120, 0x120, 0x100, 0x100, 0x0, 0x0}},
    {41005, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x100, 0x100, 0x100, 0x120, 0x120, 0x120, 0x120, 0xD20, 0x300, 0x0, 0x0}},
    {41006, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x120, 0x120, 0x120, 0x120, 0xD20, 0x300, 0x0, 0x0}},
    {41007, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x200, 0x4E0, 0x500, 0x500, 0x500, 0x500, 0x4E0, 0x200, 0x1E0, 0x0, 0x0}},
    {41008, 2, {0x0, 0x380, 0x440, 0x0, 0x860, 0x880, 0x880, 0xBE0, 0x880, 0xBE0, 0x880, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41009, 2, {0x0, 0x0, 0x0, 0x0, 0x860, 0x880, 0x880, 0xBE0, 0x880, 0xBE0, 0x880, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41010, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x200, 0x390, 0x250, 0x220, 0x220, 0x250, 0x1B90, 0x600, 0x0, 0x0}},
    {41011, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0x160, 0x180, 0x160, 0x110, 0x160, 0xD80, 0x360, 0x0, 0x0}},
    {41012, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x160, 0x180, 0x160, 0x110, 0x160, 0xD80, 0x360, 0x0, 0x0}},
    {41013, 2, {0x0, 0x0, 0x0, 0x0, 0xCC0, 0xAA0, 0x990, 0x880, 0x880, 0x1C80, 0x880, 0x880, 0x2280, 0x1C80, 0x0, 0x0}},
    {41014, 2, {0x0, 0x380, 0x440, 0x0, 0x30, 0x28, 0x24, 0x1FE0, 0x2220, 0x4AA0, 0x4220, 0x4AA0, 0x2220, 0x1FE0, 0x0, 0x0}},
    {41015, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0x28, 0x24, 0x1FE0, 0x2220, 0x4AA0, 0x4220, 0x4AA0, 0x2220, 0x1FE0, 0x0, 0x0}},
    {41016, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0x50, 0x48, 0xFC0, 0x1040, 0x3FC0, 0x2040, 0x3FC0, 0x1040, 0xFC0, 0x0, 0x0}},
    {41017, 2, {0x0, 0x380, 0x440, 0x0, 0x1C0, 0x220, 0x1220, 0x1220, 0x12E0, 0x1720, 0x1A20, 0x220, 0x220, 0x1C0, 0x0, 0x0}},
    {41018, 2, {0x0, 0x0, 0x0, 0x0, 0x1C0, 0x220, 0x1220, 0x1220, 0x12E0, 0x1720, 0x1A20, 0x220, 0x220, 0x1C0, 0x0, 0x0}},
    {41019, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1100, 0x1100, 0xA50, 0xA50, 0x450, 0xA50, 0xA50, 0x1100, 0x1100, 0x0, 0x0}},
    {41020, 2, {0x0, 0x380, 0x440, 0x0, 0xE00, 0x1100, 0x1100, 0x1150, 0x1150, 0x1F50, 0x1150, 0x1150, 0x1100, 0xE00, 0x0, 0x0}},
    {41021, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1100, 0x1100, 0x1150, 0x1150, 0x1F50, 0x1150, 0x1150, 0x1100, 0xE00, 0x0, 0x0}},
    {41022, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1110, 0x1110, 0x820, 0x7C0, 0x100, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {41023, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x0, 0x0}},
    {41024, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41025, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41026, 2, {0x0, 0x0, 0x0, 0x0, 0x880, 0x890, 0x890, 0x890, 0x8F0, 0x890, 0x890, 0x890, 0x880, 0x700, 0x0, 0x0}},
    {41027, 2, {0x0, 0x380, 0x440, 0x0, 0x7F0, 0x800, 0xFF0, 0x800, 0x7F0, 0x80, 0xAA0, 0xAA0, 0xAA0, 0x80, 0x0, 0x0}},
    {41028, 2, {0x0, 0x0, 0x0, 0x0, 0x7F0, 0x800, 0xFF0, 0x800, 0x7F0, 0x80, 0xAA0, 0xAA0, 0xAA0, 0x80, 0x0, 0x0}},
    {41029, 2, {0x0, 0x0, 0x0, 0x0, 0x7F0, 0x910, 0xFF0, 0x910, 0x7F0, 0x10, 0x810, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {41030, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x8A0, 0x890, 0x890, 0x1CA0, 0x8C0, 0x1CA0, 0x890, 0x890, 0x720, 0x0, 0x0}},
    {41031, 2, {0x0, 0x380, 0x440, 0x0, 0x920, 0x920, 0x920, 0x540, 0x380, 0x0, 0xFE0, 0x0, 0x0, 0xFE0, 0x0, 0x0}},
    {41032, 2, {0x0, 0x0, 0x0, 0x0, 0x920, 0x920, 0x920, 0x540, 0x380, 0x0, 0xFE0, 0x0, 0x0, 0xFE0, 0x0, 0x0}},
    {41033, 2, {0x0, 0x0, 0x0, 0x0, 0x7F0, 0x800, 0xFF0, 0x800, 0x7F0, 0x80, 0x7F0, 0x80, 0x7F0, 0x80, 0x0, 0x0}},
    {41034, 2, {0x0, 0x0, 0x0, 0x0, 0x620, 0x5A0, 0x460, 0x400, 0x400, 0x400, 0x400, 0x460, 0x5A0, 0x620, 0x0, 0x0}},
    {41035, 2, {0x0, 0x380, 0x440, 0x0, 0xFE0, 0x1020, 0x1020, 0xFE0, 0x1020, 0x1020, 0xFE0, 0x1020, 0x1020, 0xFE0, 0x0, 0x0}},
    {41036, 2, {0x0, 0x0, 0x0, 0x0, 0x7F0, 0x810, 0x810, 0x7F0, 0x810, 0x810, 0x7F0, 0x810, 0x810, 0x7F0, 0x0, 0x0}},
    {41037, 2, {0x0, 0x0, 0x0, 0xC0, 0xB0, 0x780, 0x880, 0xA80, 0xA80, 0x880, 0x780, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41038, 2, {0x0, 0x380, 0x440, 0x0, 0xE00, 0x1100, 0x1150, 0x1550, 0x15F0, 0x1550, 0x1150, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41039, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1100, 0x1150, 0x1550, 0x15F0, 0x1550, 0x1150, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41040, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x90, 0x780, 0x880, 0x1080, 0x1080, 0x1080, 0x880, 0x780, 0x0, 0x0}},
    {41041, 2, {0x0, 0x380, 0x440, 0x0, 0x0, 0x3018, 0x2828, 0x2448, 0x22A8, 0x2128, 0x22A8, 0x2448, 0x2828, 0x3018, 0x0, 0x0}},
    {41042, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3018, 0x2828, 0x2448, 0x22A8, 0x2128, 0x22A8, 0x2448, 0x2828, 0x3018, 0x0, 0x0}},
    {41043, 2, {0x0, 0x0, 0x0, 0x0, 0xCC0, 0xAA0, 0x990, 0x880, 0x1CA0, 0x8A0, 0x8A0, 0x8A0, 0x2280, 0x1C80, 0x0, 0x0}},
    {41044, 2, {0x0, 0x380, 0x440, 0x0, 0x8C0, 0x8A0, 0x890, 0x80, 0x780, 0x890, 0x1090, 0x1090, 0x880, 0x780, 0x0, 0x0}},
    {41045, 2, {0x0, 0x0, 0x0, 0x0, 0x8C0, 0x8A0, 0x890, 0x80, 0x780, 0x890, 0x1090, 0x1090, 0x880, 0x780, 0x0, 0x0}},
    {41046, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x7C0, 0x440, 0x7C0, 0x440, 0x380, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {41047, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x1110, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x1110, 0x100, 0x0, 0x0}},
    {41048, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x1110, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x1110, 0x100, 0x0, 0x0}},
    {41049, 2, {0x0, 0x0, 0x0, 0x0, 0x37D8, 0x820, 0x1450, 0x1290, 0x1290, 0x1290, 0x1290, 0x1450, 0x820, 0x37D8, 0x0, 0x0}},
    {41050, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x100, 0x0}},
    {41051, 2, {0x0, 0x380, 0x440, 0x0, 0x0, 0x78, 0x80, 0xD18, 0xB20, 0x978, 0x920, 0x918, 0x6880, 0x1878, 0x0, 0x0}},
    {41052, 2, {0x0, 0x0, 0x0, 0x0, 0x3C, 0x40, 0x8C, 0x690, 0x5BC, 0x490, 0x48C, 0x440, 0x343C, 0xC00, 0x0, 0x0}},
    {41053, 2, {0x0, 0x0, 0x0, 0x0, 0x1818, 0x420, 0x240, 0x240, 0x240, 0xFF0, 0x240, 0x240, 0x420, 0x1818, 0x0, 0x0}},
    {41054, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x300C, 0xFF0, 0x420, 0x240, 0x3C0, 0x240, 0x420, 0xFF0, 0x300C, 0x0, 0x0}},
    {41055, 2, {0x0, 0x380, 0x440, 0x0, 0x0, 0x406, 0x20A, 0x4712, 0x2AA2, 0x1242, 0x2AA2, 0x4712, 0x20A, 0x406, 0x0, 0x0}},
    {41056, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x406, 0x20A, 0x4712, 0x2AA2, 0x1242, 0x2AA2, 0x4712, 0x20A, 0x406, 0x0, 0x0}},
    {41057, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x0, 0xFE0, 0x0, 0x0, 0x1FF0, 0x0, 0x0}},
    {41058, 2, {0x0, 0x380, 0x440, 0x0, 0xCA0, 0xAA0, 0x9A0, 0x20, 0xFA0, 0x20, 0xFA0, 0x20, 0x1A0, 0x60, 0x0, 0x0}},
    {41059, 2, {0x0, 0x0, 0x0, 0x0, 0xCA0, 0xAA0, 0x9A0, 0x20, 0xFA0, 0x20, 0xFA0, 0x20, 0x1A0, 0x60, 0x0, 0x0}},
    {41060, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x1450, 0x1450, 0x1450, 0x440, 0x440, 0x1450, 0x1450, 0x1450, 0x380, 0x0, 0x0}},
    {41061, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x790, 0x890, 0x1090, 0x1090, 0x890, 0x790, 0x80, 0x80, 0x0, 0x0}},
    {41062, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x920, 0x540, 0x380, 0x540, 0x920, 0x100, 0xFE0, 0x100, 0xFE0, 0x0, 0x0}},
    {41063, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x920, 0x540, 0x380, 0x540, 0x920, 0x100, 0xFE0, 0x100, 0xFE0, 0x0, 0x0}},
    {41064, 2, {0x0, 0x0, 0x0, 0x0, 0xEE0, 0x1110, 0x100, 0x7C0, 0x820, 0x1110, 0x1110, 0x100, 0x820, 0x7C0, 0x0, 0x0}},
    {41065, 2, {0x0, 0x380, 0x440, 0x0, 0x640, 0x540, 0x34D0, 0x2C10, 0x2310, 0x20D0, 0x2030, 0x780, 0x0, 0x780, 0x0, 0x0}},
    {41066, 2, {0x0, 0x0, 0x0, 0x0, 0x320, 0x2A0, 0x1A68, 0x1608, 0x1188, 0x1068, 0x1018, 0x3C0, 0x0, 0x3C0, 0x0, 0x0}},
    {41067, 2, {0x0, 0x0, 0x0, 0x0, 0x3140, 0x2944, 0x2548, 0x2350, 0x60, 0x3F40, 0x40, 0x240, 0x140, 0xC0, 0x0, 0x0}},
    {41068, 2, {0x0, 0x0, 0x0, 0x0, 0x304C, 0x850, 0x850, 0x304C, 0x340, 0x2C0, 0x320C, 0xA10, 0xA10, 0x320C, 0x0, 0x0}},
    {41069, 2, {0x0, 0x380, 0x440, 0x0, 0x10, 0x190, 0x150, 0xD30, 0xB00, 0x920, 0x920, 0x2920, 0x1900, 0x100, 0x0, 0x0}},
    {41070, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0xC8, 0xA8, 0x698, 0x580, 0x490, 0x490, 0x1490, 0xC80, 0x80, 0x0, 0x0}},
    {41071, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x18B0, 0x1688, 0x1180, 0x90, 0x90, 0x1890, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41072, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x20, 0x1820, 0x1620, 0x11A0, 0x1060, 0x1C0, 0x220, 0x2A0, 0x2A0, 0x0, 0x0}},
    {41073, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x1820, 0x1620, 0x11A0, 0x1060, 0x1C0, 0x220, 0x2A0, 0x2A0, 0x0, 0x0}},
    {41074, 2, {0x0, 0x0, 0x0, 0x0, 0x660, 0x5D0, 0x440, 0x540, 0x540, 0x440, 0x7C0, 0x440, 0x1540, 0xCC0, 0x0, 0x0}},
    {41075, 2, {0x0, 0x380, 0x440, 0x0, 0x190, 0x170, 0x110, 0x100, 0x100, 0x100, 0x100, 0x1100, 0x1D00, 0x1300, 0x0, 0x0}},
    {41076, 2, {0x0, 0x0, 0x0, 0x0, 0x190, 0x170, 0x110, 0x100, 0x100, 0x100, 0x100, 0x1100, 0x1D00, 0x1300, 0x0, 0x0}},
    {41077, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0xD00, 0x300, 0x100, 0x100, 0x180, 0x160, 0xD00, 0x300, 0x0, 0x0}},
    {41078, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0xAC0, 0xA30, 0xA00, 0x2A0, 0x2A0, 0xBE0, 0xAA0, 0xAA0, 0x200, 0x0, 0x0}},
    {41079, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x920, 0xFE0, 0x920, 0x7C0, 0x100, 0x140, 0x150, 0x110, 0xE0, 0x0, 0x0}},
    {41080, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0xFE0, 0x920, 0x7C0, 0x100, 0x140, 0x150, 0x110, 0xE0, 0x0, 0x0}},
    {41081, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1650, 0x11C8, 0x1040, 0x1040, 0x11F0, 0x1040, 0x11F0, 0x1040, 0x1040, 0x0, 0x0}},
    {41082, 2, {0x0, 0x700, 0x880, 0x0, 0x1FC0, 0x220, 0x210, 0x200, 0x1FC0, 0x200, 0x200, 0x1FC0, 0x200, 0x200, 0x0, 0x0}},
    {41083, 2, {0x0, 0x0, 0x0, 0x0, 0x1FC0, 0x220, 0x210, 0x200, 0x1FC0, 0x200, 0x200, 0x1FC0, 0x200, 0x200, 0x0, 0x0}},
    {41084, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0xFE0, 0x100, 0x100, 0x100, 0x100, 0xFE0, 0x100, 0x100, 0x0, 0x0}},
    {41085, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1114, 0x1114, 0x2084, 0x20FC, 0x2084, 0x2094, 0x2494, 0x2480, 0x1B00, 0x0, 0x0}},
    {41086, 2, {0x0, 0x380, 0x440, 0x0, 0x0, 0x440, 0x440, 0x1FF0, 0x280, 0x100, 0x280, 0x1FF0, 0x440, 0x440, 0x0, 0x0}},
    {41087, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x1FF0, 0x280, 0x100, 0x280, 0x1FF0, 0x440, 0x440, 0x0, 0x0}},
    {41088, 2, {0x0, 0x0, 0x0, 0x1010, 0x1450, 0x1450, 0x1290, 0x1290, 0x1110, 0x1290, 0x1290, 0x1450, 0x1450, 0x1010, 0x0, 0x0}},
    {41089, 2, {0x0, 0x0, 0x0, 0x0, 0x388, 0x268, 0xE18, 0x1240, 0x2240, 0x2240, 0x2240, 0x218, 0x268, 0x388, 0x0, 0x0}},
    {41090, 2, {0x380, 0x440, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x7C0, 0x100, 0x100, 0x1010, 0xFE0, 0x0, 0x0}},
    {41091, 2, {0x0, 0x0, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x7C0, 0x100, 0x100, 0x1010, 0xFE0, 0x0, 0x0}},
    {41092, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x1110, 0x1010, 0xFE0, 0x0, 0x0}},
    {41093, 2, {0x0, 0x0, 0x0, 0x0, 0x3D0, 0x530, 0x910, 0x900, 0x9F0, 0x900, 0x900, 0x910, 0x520, 0x3C0, 0x0, 0x0}},
    {41094, 2, {0x0, 0x380, 0x440, 0x0, 0x9C0, 0xA20, 0x220, 0xE20, 0x1220, 0x1220, 0xE20, 0x220, 0xA20, 0x9C0, 0x0, 0x0}},
    {41095, 2, {0x0, 0x0, 0x0, 0x0, 0x9C0, 0xA20, 0x220, 0xE20, 0x1220, 0x1220, 0xE20, 0x220, 0xA20, 0x9C0, 0x0, 0x0}},
    {41096, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x7C0, 0x920, 0x1110, 0x1150, 0x1150, 0x1110, 0x920, 0x7C0, 0x100, 0x0, 0x0}},
    {41097, 2, {0x0, 0x380, 0x440, 0x0, 0x3838, 0x4544, 0x4544, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4444, 0x4544, 0x3938, 0x0, 0x0}},
    {41098, 2, {0x0, 0x0, 0x0, 0x0, 0x3838, 0x4544, 0x4544, 0x4444, 0x4444, 0x47C4, 0x4444, 0x4444, 0x4544, 0x3938, 0x0, 0x0}},
    {41099, 2, {0x0, 0x0, 0x0, 0x770, 0x880, 0x9A0, 0xAA0, 0xA80, 0xBE0, 0xA80, 0xAA0, 0x9A0, 0x880, 0x770, 0x0, 0x0}},
    {41100, 2, {0x0, 0x380, 0x440, 0x0, 0x1150, 0x1150, 0x1110, 0x1150, 0x1F50, 0x1110, 0x1150, 0x1150, 0x920, 0x7C0, 0x0, 0x0}},
    {41101, 2, {0x0, 0x0, 0x0, 0x0, 0x1150, 0x1150, 0x1110, 0x1150, 0x1F50, 0x1110, 0x1150, 0x1150, 0x920, 0x7C0, 0x0, 0x0}},
    {41102, 2, {0x0, 0x0, 0x0, 0x0, 0x6360, 0x1458, 0x1346, 0x1440, 0x6340, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41103, 2, {0x0, 0x380, 0x440, 0x0, 0x3118, 0x2928, 0x2548, 0x2388, 0x3FF8, 0x2388, 0x2548, 0x2928, 0x3118, 0x100, 0x0, 0x0}},
    {41104, 2, {0x0, 0x0, 0x0, 0x100, 0x3118, 0x2928, 0x2548, 0x2388, 0x3FF8, 0x2388, 0x2548, 0x2928, 0x3118, 0x100, 0x0, 0x0}},
    {41105, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x920, 0x920, 0x920, 0x920, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41106, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x920, 0x7C0, 0x100, 0x1FF0, 0x100, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41107, 2, {0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x1FF0, 0x100, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41108, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0x1110, 0x1290, 0x1290, 0x1290, 0x1390, 0x1450, 0x820, 0x7C0, 0x0, 0x0}},
    {41109, 2, {0x0, 0x380, 0x440, 0x0, 0x1550, 0x1550, 0x1110, 0x920, 0x7C0, 0x100, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41110, 2, {0x0, 0x0, 0x0, 0x0, 0x1550, 0x1550, 0x1110, 0x920, 0x7C0, 0x100, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41111, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x190, 0x150, 0x130, 0x100, 0x920, 0x920, 0x920, 0x100, 0x100, 0x0, 0x0}},
    {41112, 2, {0x0, 0x0, 0x0, 0x0, 0x600, 0x580, 0x400, 0x400, 0x580, 0x420, 0x710, 0x480, 0x3448, 0xC48, 0x0, 0x0}},
    {41113, 2, {0x0, 0x380, 0x440, 0x0, 0x10, 0x3010, 0x2C10, 0x2310, 0x20D0, 0x2038, 0x2044, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {41114, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x3010, 0x2C10, 0x2310, 0x20D0, 0x2038, 0x2044, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {41115, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x30B0, 0x88C, 0x480, 0x480, 0x3F80, 0x480, 0x480, 0x880, 0x3080, 0x0, 0x0}},
    {41116, 2, {0x0, 0x380, 0x440, 0x0, 0x240, 0x240, 0x1248, 0xA50, 0x460, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41117, 2, {0x0, 0x0, 0x0, 0x0, 0x240, 0x240, 0x1248, 0xA50, 0x460, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41118, 2, {0x0, 0x0, 0x0, 0x0, 0x980, 0x960, 0x918, 0x900, 0x180, 0x160, 0x3118, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41119, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0x50, 0x48, 0x7C0, 0x840, 0x1240, 0x1540, 0x1240, 0x840, 0x7C0, 0x0, 0x0}},
    {41120, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x1160, 0x1118, 0x100, 0xF88, 0x1148, 0x2520, 0x2500, 0x1100, 0xF00, 0x0, 0x0}},
    {41121, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x1160, 0x1118, 0x100, 0xF88, 0x1148, 0x2520, 0x2500, 0x1100, 0xF00, 0x0, 0x0}},
    {41122, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x400, 0x200, 0x180, 0x40, 0x120, 0x2A0, 0x920, 0x7C0, 0x0, 0x0}},
    {41123, 2, {0x0, 0x0, 0x0, 0xFE0, 0x820, 0x440, 0x280, 0x2A0, 0x120, 0x2A0, 0x280, 0x440, 0x820, 0xFE0, 0x0, 0x0}},
    {41124, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x920, 0x920, 0x20, 0x920, 0x920, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41125, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x920, 0x920, 0x20, 0x920, 0x920, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41126, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0x1550, 0x1550, 0x1110, 0x1110, 0x1550, 0x1550, 0x920, 0x7C0, 0x0, 0x0}},
    {41127, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1010, 0x820, 0x7C0, 0x0, 0x820, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {41128, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1010, 0x820, 0x7C0, 0x0, 0x820, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {41129, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x100, 0x120, 0x920, 0x920, 0x900, 0x1C0, 0x120, 0xD10, 0x310, 0x0, 0x0}},
    {41130, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x120, 0x920, 0x920, 0x900, 0x1C0, 0x120, 0xD10, 0x310, 0x0, 0x0}},
    {41131, 2, {0x0, 0x0, 0x0, 0x850, 0x850, 0x10, 0x850, 0x850, 0x10, 0x850, 0x850, 0x310, 0x2D0, 0x230, 0x0, 0x0}},
    {41132, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x920, 0xFE0, 0x920, 0x7C0, 0x100, 0x540, 0x1550, 0x1110, 0xEE0, 0x0, 0x0}},
    {41133, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0xFE0, 0x920, 0x7C0, 0x100, 0x540, 0x1550, 0x1110, 0xEE0, 0x0, 0x0}},
    {41134, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x16D0, 0x1290, 0x1110, 0x17D0, 0x1010, 0x1390, 0x820, 0x7C0, 0x0, 0x0}},
    {41135, 2, {0x0, 0x380, 0x440, 0x0, 0x310, 0x2D0, 0x230, 0x200, 0x230, 0x2C0, 0x300, 0x12C0, 0xA30, 0x600, 0x0, 0x0}},
    {41136, 2, {0x0, 0x0, 0x0, 0x0, 0x310, 0x2D0, 0x230, 0x200, 0x230, 0x2C0, 0x300, 0x12C0, 0xA30, 0x600, 0x0, 0x0}},
    {41137, 2, {0x0, 0x0, 0x0, 0xC8, 0xA8, 0x98, 0x80, 0xFF8, 0x80, 0xFF8, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41138, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1450, 0x2288, 0x2128, 0x2128, 0x2288, 0x1450, 0x820, 0x7C0, 0x0, 0x0}},
    {41139, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1450, 0x2288, 0x2128, 0x2128, 0x2288, 0x1450, 0x820, 0x7C0, 0x0, 0x0}},
    {41140, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0xC8, 0xA8, 0x1098, 0x1080, 0x1090, 0x90, 0x1890, 0x680, 0x180, 0x0, 0x0}},
    {41141, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1750, 0x12C8, 0x1240, 0x1240, 0x1240, 0x1FC0, 0x1240, 0x1240, 0x1240, 0x0, 0x0}},
    {41142, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41143, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41144, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x7C0, 0x100, 0x7C0, 0x100, 0x0, 0x0}},
    {41145, 2, {0x0, 0x0, 0x0, 0x1210, 0x1110, 0x1090, 0x1210, 0x1110, 0x1090, 0x1210, 0x1110, 0x1090, 0x820, 0x7C0, 0x0, 0x0}},
    {41146, 2, {0x0, 0x380, 0x440, 0x0, 0xC0, 0x6B0, 0x580, 0x4C0, 0x4B0, 0x1480, 0xCC0, 0xB0, 0x680, 0x180, 0x0, 0x0}},
    {41147, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x6B0, 0x580, 0x4C0, 0x4B0, 0x1480, 0xCC0, 0xB0, 0x680, 0x180, 0x0, 0x0}},
    {41148, 2, {0x0, 0x0, 0x0, 0x0, 0x638, 0x544, 0x4C4, 0x454, 0x1F54, 0x454, 0x1F54, 0x444, 0x3444, 0xC38, 0x0, 0x0}},
    {41149, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x90, 0x80, 0x780, 0x880, 0x1280, 0x1280, 0x880, 0x780, 0x0, 0x0}},
    {41150, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x1C0, 0x120, 0x1C0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41151, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x1C0, 0x120, 0x1C0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41152, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x810, 0x10, 0x10, 0x10, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41153, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x118, 0x124, 0x124, 0xFE4, 0x124, 0x124, 0x3118, 0xD00, 0x300, 0x0, 0x0}},
    {41154, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x118, 0x124, 0x124, 0xFE4, 0x124, 0x124, 0x3118, 0xD00, 0x300, 0x0, 0x0}},
    {41155, 2, {0x0, 0x0, 0x0, 0x0, 0x90, 0x88, 0xA88, 0xA90, 0xAE0, 0x90, 0x88, 0x1888, 0x690, 0x180, 0x0, 0x0}},
    {41156, 2, {0x0, 0x380, 0x440, 0x0, 0x420, 0x420, 0x420, 0x420, 0x7E0, 0x420, 0x420, 0x2420, 0x1420, 0xC20, 0x0, 0x0}},
    {41157, 2, {0x0, 0x0, 0x0, 0x0, 0x210, 0x210, 0x210, 0x210, 0x3F0, 0x210, 0x210, 0x1210, 0xA10, 0x610, 0x0, 0x0}},
    {41158, 2, {0x0, 0x0, 0x0, 0x0, 0x388, 0x450, 0x3E0, 0x40, 0x1F0, 0x248, 0x1F0, 0x1840, 0x640, 0x180, 0x0, 0x0}},
    {41159, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0x80, 0x8F0, 0x880, 0x880, 0x80, 0xF0, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41160, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x8F0, 0x880, 0x880, 0x80, 0xF0, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41161, 2, {0x0, 0x0, 0x0, 0x0, 0xC20, 0xB20, 0x8E0, 0x800, 0xF70, 0x880, 0x940, 0x880, 0xF70, 0x800, 0x0, 0x0}},
    {41162, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x820, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {41163, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x820, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {41164, 2, {0x0, 0x0, 0x0, 0x0, 0x200, 0x220, 0x220, 0x220, 0x200, 0xFF0, 0x200, 0x220, 0x220, 0x220, 0x0, 0x0}},
    {41165, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x1EF0, 0x820, 0x1EF0, 0x820, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {41166, 2, {0x0, 0x380, 0x440, 0x0, 0xC0, 0xB0, 0x480, 0x780, 0x480, 0x80, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41167, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0xB0, 0x480, 0x780, 0x480, 0x80, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41168, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xB0, 0x880, 0xF80, 0x880, 0x80, 0x80, 0xE80, 0x980, 0x800, 0x0, 0x0}},
    {41169, 2, {0x0, 0x0, 0x0, 0x920, 0xAB0, 0xAAC, 0x920, 0x820, 0xFE0, 0x820, 0x920, 0xAA0, 0x6AA0, 0x1920, 0x0, 0x0}},
    {41170, 2, {0x0, 0x380, 0x440, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x140, 0x948, 0x948, 0x410, 0x3E0, 0x0, 0x0}},
    {41171, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x140, 0x948, 0x948, 0x410, 0x3E0, 0x0, 0x0}},
    {41172, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x220, 0xAA8, 0x888, 0x410, 0x3E0, 0x0, 0x0}},
    {41173, 2, {0x0, 0x380, 0x440, 0x0, 0x120, 0xA0, 0x18E0, 0x520, 0x220, 0x220, 0x520, 0x18E0, 0xA0, 0x120, 0x0, 0x0}},
    {41174, 2, {0x0, 0x0, 0x0, 0x0, 0x90, 0x50, 0xC70, 0x290, 0x110, 0x110, 0x290, 0xC70, 0x50, 0x90, 0x0, 0x0}},
    {41175, 2, {0x0, 0x0, 0x0, 0x0, 0x920, 0x520, 0x720, 0x900, 0x11F0, 0x1100, 0x920, 0x720, 0x520, 0x900, 0x0, 0x0}},
    {41176, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x12C0, 0x1230, 0x1200, 0x220, 0x220, 0x1220, 0x1200, 0x1200, 0x200, 0x0, 0x0}},
    {41177, 2, {0x0, 0x380, 0x440, 0x0, 0x820, 0x2828, 0x2828, 0x2828, 0x1830, 0x820, 0xFE0, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {41178, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x2828, 0x2828, 0x2828, 0x1830, 0x820, 0xFE0, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {41179, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x240, 0x240, 0x1810, 0x1660, 0x1180, 0x1660, 0x1810, 0x240, 0x240, 0x0, 0x0}},
    {41180, 2, {0x0, 0x380, 0x440, 0x0, 0x8, 0x188, 0x3968, 0x518, 0x2300, 0x2300, 0x500, 0x3900, 0x100, 0x100, 0x0, 0x0}},
    {41181, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x3968, 0x518, 0x2300, 0x2300, 0x500, 0x3900, 0x100, 0x100, 0x0, 0x0}},
    {41182, 2, {0x0, 0x0, 0x0, 0x0, 0x90, 0xA0, 0xC0, 0x90, 0xA0, 0xC0, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41183, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1510, 0x920, 0x17C0, 0x100, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41184, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1510, 0x920, 0x17C0, 0x100, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41185, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x820, 0x7C0, 0x100, 0x120, 0x3140, 0x2D80, 0x2300, 0x2000, 0x0, 0x0}},
    {41186, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x820, 0x7C0, 0x100, 0x1900, 0x1500, 0x1300, 0x1000, 0x0, 0x0, 0x0}},
    {41187, 2, {0x0, 0x380, 0x440, 0x0, 0x18, 0x20, 0x40, 0x640, 0x5C0, 0x440, 0x440, 0x440, 0x3420, 0xC18, 0x0, 0x0}},
    {41188, 2, {0x0, 0x0, 0x0, 0x0, 0x18, 0x20, 0x40, 0x640, 0x5C0, 0x440, 0x440, 0x440, 0x3420, 0xC18, 0x0, 0x0}},
    {41189, 2, {0x0, 0x0, 0x0, 0x0, 0x1C10, 0x1390, 0x1170, 0x100, 0x1C0, 0x920, 0x920, 0x920, 0x1C0, 0x100, 0x0, 0x0}},
    {41190, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0x1110, 0x110, 0x120, 0x1C0, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41191, 2, {0x0, 0x380, 0x440, 0x0, 0x920, 0x540, 0x1390, 0x1390, 0x540, 0x920, 0x100, 0xFE0, 0x100, 0xFE0, 0x0, 0x0}},
    {41192, 2, {0x0, 0x0, 0x0, 0x0, 0x920, 0x540, 0x1390, 0x1390, 0x540, 0x920, 0x100, 0xFE0, 0x100, 0xFE0, 0x0, 0x0}},
    {41193, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0x9C0, 0x920, 0x920, 0x9C0, 0x100, 0x100, 0x3FF8, 0x0, 0x0}},
    {41194, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41195, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1390, 0x1450, 0x820, 0x1450, 0x1390, 0x1010, 0x1010, 0x1010, 0x0, 0x0}},
    {41196, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1390, 0x1450, 0x820, 0x1450, 0x1390, 0x1010, 0x1010, 0x1010, 0x0, 0x0}},
    {41197, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x200, 0x3238, 0xEC8, 0x308, 0xEC8, 0x3238, 0x1A00, 0x600, 0x0, 0x0}},
    {41198, 2, {0x0, 0x0, 0x0, 0x0, 0x240, 0x240, 0x1A48, 0x1648, 0x13C8, 0x1268, 0x1258, 0x240, 0x240, 0x180, 0x0, 0x0}},
    {41199, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0xC88, 0x1290, 0x12A0, 0x12C0, 0xC80, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41200, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0xC88, 0x1290, 0x12A0, 0x12C0, 0xC80, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41201, 2, {0x0, 0x0, 0x0, 0x24, 0x1A4, 0x1F74, 0x212C, 0x2120, 0x1100, 0xF00, 0x1100, 0x2100, 0x2100, 0x1F00, 0x0, 0x0}},
    {41202, 2, {0x0, 0x380, 0x440, 0x0, 0x1E0, 0x10, 0xC10, 0x1020, 0x11C0, 0x1020, 0xC10, 0x10, 0x10, 0x1E0, 0x0, 0x0}},
    {41203, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x10, 0xC10, 0x1020, 0x11C0, 0x1020, 0xC10, 0x10, 0x10, 0x1E0, 0x0, 0x0}},
    {41204, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0x28, 0x224, 0x1420, 0x8A0, 0x1520, 0x2220, 0x520, 0x8A0, 0x40, 0x0, 0x0}},
    {41205, 2, {0x380, 0x440, 0x0, 0x1C70, 0x380, 0x440, 0x820, 0x3E20, 0x820, 0x3E20, 0x820, 0x440, 0x380, 0x1C70, 0x0, 0x0}},
    {41206, 2, {0x0, 0x0, 0x0, 0x1C70, 0x380, 0x440, 0x820, 0x3E20, 0x820, 0x3E20, 0x820, 0x440, 0x380, 0x1C70, 0x0, 0x0}},
    {41207, 2, {0x0, 0x0, 0x0, 0x80, 0x40, 0xFE0, 0x820, 0x440, 0x280, 0x100, 0x280, 0x440, 0x920, 0x1110, 0x0, 0x0}},
    {41208, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x20, 0x20, 0x1E0, 0x220, 0x1F20, 0x420, 0x1F20, 0x220, 0x1E0, 0x0, 0x0}},
    {41209, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x1E0, 0x220, 0x1F20, 0x420, 0x1F20, 0x220, 0x1E0, 0x0, 0x0}},
    {41210, 2, {0x0, 0x0, 0x0, 0x180, 0x140, 0x2128, 0x2108, 0x1110, 0x4FE4, 0x4104, 0x4FE4, 0x1110, 0x2108, 0x2108, 0x0, 0x0}},
    {41211, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x820, 0x820, 0x920, 0x100, 0x100, 0x0, 0x0}},
    {41212, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x820, 0x820, 0x920, 0x100, 0x100, 0x0, 0x0}},
    {41213, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x2148, 0x2108, 0x1110, 0xFE4, 0x104, 0xFE4, 0x1110, 0x2508, 0x2308, 0x0, 0x0}},
    {41214, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x1110, 0x1110, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x0, 0x0}},
    {41215, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x1110, 0x1110, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x0, 0x0}},
    {41216, 2, {0x0, 0x0, 0x0, 0x0, 0x88, 0x3284, 0x2A84, 0x2688, 0xF0, 0x88, 0x3284, 0x2A84, 0x2688, 0x80, 0x0, 0x0}},
    {41217, 2, {0x0, 0x380, 0x440, 0x0, 0xC0, 0x4A0, 0x290, 0x1280, 0xC80, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41218, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x4A0, 0x290, 0x1280, 0xC80, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41219, 2, {0x0, 0x0, 0x100, 0x200, 0x100, 0x80, 0x1110, 0x1110, 0xEE0, 0x440, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {41220, 2, {0x0, 0x380, 0x440, 0x0, 0x1C, 0x20, 0x4C, 0x650, 0x5D0, 0x450, 0x450, 0x44C, 0x3420, 0xC1C, 0x0, 0x0}},
    {41221, 2, {0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x4C, 0x650, 0x5D0, 0x450, 0x450, 0x44C, 0x3420, 0xC1C, 0x0, 0x0}},
    {41222, 2, {0x0, 0x0, 0x0, 0x0, 0x420, 0x420, 0x7242, 0x4E42, 0x4182, 0x4272, 0x424E, 0x420, 0x420, 0x420, 0x0, 0x0}},
    {41223, 2, {0x0, 0x0, 0x0, 0x0, 0x1100, 0x1120, 0x1120, 0x1120, 0x1100, 0x1FF8, 0x1100, 0x1120, 0x1120, 0xE20, 0x0, 0x0}},
    {41224, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x1018, 0x828, 0x448, 0x1288, 0x1108, 0x1288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {41225, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1018, 0x828, 0x448, 0x1288, 0x1108, 0x1288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {41226, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x80C, 0x414, 0x1A24, 0x1544, 0x1084, 0x5144, 0x3224, 0x414, 0x80C, 0x0, 0x0}},
    {41227, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0xA0, 0xA0, 0x8A0, 0xEA0, 0x9A0, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41228, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0xA0, 0xA0, 0x8A0, 0xEA0, 0x9A0, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41229, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x7C0, 0x920, 0x920, 0x820, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0}},
    {41230, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41231, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41232, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x100, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41233, 2, {0x380, 0x440, 0x0, 0x1C0, 0x130, 0x100, 0x7C0, 0x540, 0x380, 0x380, 0x540, 0x7C0, 0x100, 0x1900, 0x700, 0x0}},
    {41234, 2, {0x0, 0x0, 0x1C0, 0x130, 0x100, 0x7C0, 0x540, 0x380, 0x380, 0x540, 0x7C0, 0x100, 0x1900, 0x700, 0x0, 0x0}},
    {41235, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x3E0, 0x490, 0x888, 0x888, 0x0, 0x0}},
    {41236, 2, {0x0, 0x0, 0x0, 0x0, 0x410, 0x890, 0x910, 0x710, 0x910, 0x890, 0x410, 0x10, 0x190, 0x70, 0x0, 0x0}},
    {41237, 2, {0x0, 0x380, 0x440, 0x0, 0x1E20, 0x2120, 0x21C0, 0x1E80, 0xF0, 0x880, 0x8F0, 0x880, 0x88, 0x70, 0x0, 0x0}},
    {41238, 2, {0x0, 0x0, 0x0, 0x0, 0x1E20, 0x2120, 0x21C0, 0x1E80, 0xF0, 0x880, 0x8F0, 0x880, 0x88, 0x70, 0x0, 0x0}},
    {41239, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x1F0, 0x100, 0x100, 0xF0, 0x0, 0x0}},
    {41240, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x500, 0x500, 0x500, 0x100, 0x0, 0x0}},
    {41241, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x500, 0x500, 0x500, 0x100, 0x0, 0x0}},
    {41242, 2, {0x0, 0x0, 0xFE0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x100, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41243, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0x80, 0x27F0, 0x2080, 0x2080, 0x27F0, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41244, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x13F8, 0x1040, 0x1040, 0x13F8, 0x40, 0xC40, 0x340, 0xC0, 0x0, 0x0}},
    {41245, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41246, 2, {0x0, 0x380, 0x440, 0x0, 0x8C0, 0x4B0, 0x2480, 0x1880, 0x80, 0x7F0, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41247, 2, {0x0, 0x0, 0x0, 0x0, 0x8C0, 0x4B0, 0x2480, 0x1880, 0x80, 0x7F0, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41248, 2, {0x0, 0x0, 0x0, 0x0, 0x850, 0x850, 0x790, 0x10, 0x790, 0x850, 0x850, 0x310, 0xD0, 0x30, 0x0, 0x0}},
    {41249, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x10, 0x390, 0x410, 0x410, 0x410, 0x410, 0x390, 0x10, 0xFE0, 0x0, 0x0}},
    {41250, 2, {0x0, 0x380, 0x440, 0x0, 0x820, 0x820, 0x0, 0x1FE0, 0x10, 0x10, 0x1FE0, 0x0, 0x840, 0x840, 0x0, 0x0}},
    {41251, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x0, 0x1FE0, 0x10, 0x10, 0x1FE0, 0x0, 0x840, 0x840, 0x0, 0x0}},
    {41252, 2, {0x0, 0x0, 0x0, 0xFE0, 0x10, 0x10, 0xF90, 0x10, 0xF90, 0x10, 0xF90, 0x10, 0x10, 0xFE0, 0x0, 0x0}},
    {41253, 2, {0x0, 0x0, 0x0, 0x0, 0x711C, 0x920, 0x920, 0x1010, 0x600C, 0x1010, 0x920, 0x920, 0x920, 0x701C, 0x0, 0x0}},
    {41254, 2, {0x380, 0x440, 0x0, 0x10, 0xFF0, 0x10, 0xFF0, 0x10, 0xFF0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0x0, 0x0}},
    {41255, 2, {0x0, 0x0, 0x0, 0x10, 0xFF0, 0x10, 0xFF0, 0x10, 0xFF0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0x0, 0x0}},
    {41256, 2, {0x0, 0x0, 0x0, 0x0, 0x804, 0xAC4, 0xAB4, 0x88C, 0xA80, 0xA80, 0x880, 0x4A80, 0x2A80, 0x1880, 0x0, 0x0}},
    {41257, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x180, 0x160, 0x100, 0x180, 0x160, 0xD00, 0x300, 0x0, 0x0}},
    {41258, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x1160, 0x1118, 0x1180, 0x160, 0x118, 0x3180, 0x2D60, 0x2318, 0x2000, 0x0, 0x0}},
    {41259, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x1160, 0x1118, 0x1180, 0x160, 0x118, 0x3180, 0x2D60, 0x2318, 0x2000, 0x0, 0x0}},
    {41260, 2, {0x0, 0x0, 0x0, 0x0, 0x3C78, 0x4104, 0x4104, 0x3C78, 0x4104, 0x4104, 0x3C78, 0x4104, 0x4104, 0x3C78, 0x0, 0x0}},
    {41261, 2, {0x0, 0x380, 0x440, 0x0, 0x1248, 0xA88, 0x708, 0x708, 0xA88, 0x1248, 0xF88, 0x208, 0x788, 0x78, 0x0, 0x0}},
    {41262, 2, {0x0, 0x0, 0x0, 0x0, 0x1248, 0xA88, 0x708, 0x708, 0xA88, 0x1248, 0xF88, 0x208, 0x788, 0x78, 0x0, 0x0}},
    {41263, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x230, 0x300, 0x2C0, 0x230, 0x200, 0xF80, 0x1240, 0x1240, 0x0, 0x0}},
    {41264, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1170, 0x1180, 0x1100, 0x1120, 0x1120, 0x1100, 0x1180, 0x1170, 0xE00, 0x0, 0x0}},
    {41265, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41266, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41267, 2, {0x0, 0x0, 0x0, 0x0, 0x7F0, 0x808, 0x808, 0x808, 0x7F0, 0x40, 0x1650, 0x1550, 0x14D0, 0x400, 0x0, 0x0}},
    {41268, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0xD60, 0xB00, 0x928, 0x928, 0x29F8, 0x1928, 0x128, 0xD00, 0x300, 0x0, 0x0}},
    {41269, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0xD60, 0xB00, 0x928, 0x928, 0x29F8, 0x1928, 0x128, 0xD00, 0x300, 0x0, 0x0}},
    {41270, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x490, 0x490, 0x490, 0x80, 0x0, 0x0}},
    {41271, 2, {0x0, 0x380, 0x440, 0x0, 0x10C0, 0x8B0, 0x480, 0x1080, 0x880, 0x480, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41272, 2, {0x0, 0x0, 0x0, 0x0, 0x10C0, 0x8B0, 0x480, 0x1080, 0x880, 0x480, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41273, 2, {0x0, 0x0, 0x0, 0x0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0xFF8, 0x80, 0x80, 0xFF8, 0x0, 0x0}},
    {41274, 2, {0x0, 0x380, 0x440, 0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x100, 0xFE0, 0x100, 0x100, 0xFE0, 0x0, 0x0}},
    {41275, 2, {0x0, 0x0, 0x0, 0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x100, 0xFE0, 0x100, 0x100, 0xFE0, 0x0, 0x0}},
    {41276, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x2108, 0x3FF8, 0x2108, 0x100, 0x7C0, 0x920, 0x1110, 0x1110, 0x1110, 0x0, 0x0}},
    {41277, 2, {0x0, 0x0, 0x0, 0x0, 0x1FF0, 0x100, 0x7C0, 0x920, 0x920, 0x920, 0x920, 0x7C0, 0x100, 0x1FF0, 0x0, 0x0}},
    {41278, 2, {0x380, 0x440, 0x0, 0x1FF0, 0x100, 0x311C, 0xD64, 0x384, 0xD64, 0x311C, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x0, 0x0}},
    {41279, 2, {0x0, 0x0, 0x0, 0x1FF0, 0x100, 0x311C, 0xD64, 0x384, 0xD64, 0x311C, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x0, 0x0}},
    {41280, 2, {0x0, 0x0, 0xFE0, 0x100, 0xFE0, 0x100, 0x3908, 0x2708, 0x21C8, 0x2138, 0x100, 0xFE0, 0x100, 0xFE0, 0x0, 0x0}},
    {41281, 2, {0x0, 0x380, 0x440, 0x0, 0xFE0, 0x100, 0x380, 0x540, 0x920, 0x920, 0x920, 0x920, 0x540, 0x380, 0x0, 0x0}},
    {41282, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x100, 0x380, 0x540, 0x920, 0x920, 0x920, 0x920, 0x540, 0x380, 0x0, 0x0}},
    {41283, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x100, 0xFE0, 0x820, 0x440, 0x280, 0x100, 0x280, 0x440, 0x820, 0x0, 0x0}},
    {41284, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x140, 0x120, 0x100, 0x100, 0x920, 0x7C0, 0x0, 0x920, 0x920, 0x0, 0x0}},
    {41285, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x920, 0x920, 0x0, 0x0}},
    {41286, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x920, 0x920, 0x0, 0x0}},
    {41287, 2, {0x0, 0x0, 0x0, 0x310, 0x2D0, 0x230, 0xF80, 0x200, 0x200, 0x1040, 0xF80, 0x0, 0x1240, 0x1240, 0x0, 0x0}},
    {41288, 2, {0x0, 0x380, 0x440, 0x0, 0xEE0, 0x100, 0x280, 0x440, 0x540, 0x540, 0x440, 0x280, 0x100, 0xEE0, 0x0, 0x0}},
    {41289, 2, {0x0, 0x0, 0x0, 0x0, 0xEE0, 0x100, 0x280, 0x440, 0x540, 0x540, 0x440, 0x280, 0x100, 0xEE0, 0x0, 0x0}},
    {41290, 2, {0x0, 0x0, 0x0, 0x0, 0x860, 0xFD0, 0x2848, 0x1840, 0x3C0, 0x440, 0x840, 0x840, 0x440, 0x3C0, 0x0, 0x0}},
    {41291, 2, {0x0, 0x0, 0x0, 0x1FC0, 0x1040, 0x880, 0x510, 0x510, 0x3F0, 0x510, 0x510, 0x880, 0x1040, 0x1FC0, 0x0, 0x0}},
    {41292, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x1100, 0x1100, 0x1100, 0x100, 0x0, 0x0}},
    {41293, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x1100, 0x1100, 0x1100, 0x100, 0x0, 0x0}},
    {41294, 2, {0x0, 0x0, 0x0, 0x0, 0xF80, 0x1080, 0x1088, 0x888, 0x7F8, 0x888, 0x1088, 0x1080, 0x1080, 0xF80, 0x0, 0x0}},
    {41295, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41296, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41297, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x0, 0x490, 0x490, 0x0, 0x0}},
    {41298, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x820, 0xFE0, 0x820, 0x20, 0x20, 0x820, 0xE20, 0x9A0, 0x60, 0x0, 0x0}},
    {41299, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x820, 0xFE0, 0x820, 0x20, 0x20, 0x820, 0xE20, 0x9A0, 0x60, 0x0, 0x0}},
    {41300, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x920, 0x920, 0x820, 0x820, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {41301, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x2108, 0x17D0, 0x1110, 0x17D0, 0x2108, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41302, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x2108, 0x17D0, 0x1110, 0x17D0, 0x2108, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41303, 2, {0x0, 0x0, 0x0, 0x0, 0xC60, 0x1250, 0x13C8, 0x1240, 0xC40, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41304, 2, {0x0, 0x380, 0x440, 0x0, 0xC60, 0x1250, 0x1248, 0xC40, 0x40, 0x1FF8, 0x40, 0x740, 0x4C0, 0x400, 0x0, 0x0}},
    {41305, 2, {0x0, 0x0, 0x0, 0x0, 0xC60, 0x1250, 0x1248, 0xC40, 0x40, 0x1FF8, 0x40, 0x740, 0x4C0, 0x400, 0x0, 0x0}},
    {41306, 2, {0x0, 0x0, 0x0, 0x0, 0x1888, 0x24A8, 0x24A8, 0x24A8, 0x18A8, 0xF0, 0xA0, 0x18A0, 0x6A0, 0x180, 0x0, 0x0}},
    {41307, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x7F0, 0x840, 0xFF0, 0x840, 0x7F0, 0x40, 0xE40, 0x9C0, 0x800, 0x0, 0x0}},
    {41308, 2, {0x380, 0x440, 0x0, 0x400, 0x990, 0x1250, 0x1250, 0x1250, 0x1190, 0x810, 0x310, 0x2D0, 0x230, 0x200, 0x0, 0x0}},
    {41309, 2, {0x0, 0x0, 0x0, 0x400, 0x990, 0x1250, 0x1250, 0x1250, 0x1190, 0x810, 0x310, 0x2D0, 0x230, 0x200, 0x0, 0x0}},
    {41310, 2, {0x0, 0x0, 0x400, 0x800, 0x1010, 0x1450, 0x1450, 0x1110, 0x1110, 0x810, 0x310, 0x2D0, 0x230, 0x200, 0x0, 0x0}},
    {41311, 2, {0x0, 0x380, 0x440, 0x0, 0x1830, 0x2428, 0x2424, 0x2420, 0x19E0, 0x220, 0x420, 0x420, 0x220, 0x1E0, 0x0, 0x0}},
    {41312, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x2428, 0x2424, 0x2420, 0x19E0, 0x220, 0x420, 0x420, 0x220, 0x1E0, 0x0, 0x0}},
    {41313, 2, {0x0, 0x0, 0x0, 0x0, 0x660, 0x958, 0x2940, 0x2940, 0x2640, 0x1040, 0xFC0, 0x40, 0x340, 0xC0, 0x0, 0x0}},
    {41314, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x880, 0x880, 0x520, 0x520, 0x3E0, 0x520, 0x520, 0x880, 0x880, 0x0, 0x0}},
    {41315, 2, {0x0, 0x380, 0x440, 0x0, 0x120, 0x120, 0x120, 0x120, 0x120, 0x1C0, 0x100, 0x1110, 0x1110, 0xFE0, 0x0, 0x0}},
    {41316, 2, {0x0, 0x0, 0x0, 0x0, 0x120, 0x120, 0x120, 0x120, 0x120, 0x1C0, 0x100, 0x1110, 0x1110, 0xFE0, 0x0, 0x0}},
    {41317, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x0, 0x0}},
    {41318, 2, {0x0, 0x380, 0x440, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x1F20, 0x120, 0x1F20, 0x100, 0x0, 0x0}},
    {41319, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x1F20, 0x120, 0x1F20, 0x100, 0x0, 0x0}},
    {41320, 2, {0x0, 0x0, 0x0, 0x0, 0x980, 0x1160, 0x1118, 0x900, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x100, 0x0, 0x0}},
    {41321, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x100, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x100, 0x0, 0x0}},
    {41322, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x100, 0x0, 0x0}},
    {41323, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x140, 0x2128, 0x2108, 0x1110, 0xFE0, 0x100, 0xFE0, 0x1110, 0x2108, 0x2108, 0x0}},
    {41324, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x140, 0x120, 0x100, 0x1FF0, 0x100, 0xFE0, 0x1010, 0x2108, 0x2108, 0x0, 0x0}},
    {41325, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x1900, 0x1500, 0x1340, 0x1040, 0x180, 0x0, 0x0}},
    {41326, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x1900, 0x1500, 0x1340, 0x1040, 0x180, 0x0, 0x0}},
    {41327, 2, {0x0, 0x0, 0x0, 0x460, 0x2558, 0x2546, 0x440, 0x6440, 0x1C40, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41328, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0x1FC8, 0x208, 0x1FC8, 0x200, 0x200, 0x0, 0x0}},
    {41329, 2, {0x0, 0x380, 0x440, 0x0, 0x8, 0x188, 0x168, 0x118, 0xD00, 0xB80, 0x2960, 0x1918, 0x100, 0x100, 0x0, 0x0}},
    {41330, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0xD00, 0xB80, 0x2960, 0x1918, 0x100, 0x100, 0x0, 0x0}},
    {41331, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x820, 0x440, 0x380, 0x440, 0x820, 0x180, 0x140, 0x500, 0x300, 0x0, 0x0}},
    {41332, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x540, 0x540, 0x3548, 0x2F48, 0x25E8, 0x2558, 0x540, 0x540, 0x380, 0x0, 0x0}},
    {41333, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x540, 0x540, 0x3548, 0x2F48, 0x25E8, 0x2558, 0x540, 0x540, 0x380, 0x0, 0x0}},
    {41334, 2, {0x0, 0x0, 0x0, 0x0, 0x2230, 0x5528, 0x5524, 0x2220, 0x820, 0x1420, 0x1420, 0x9A0, 0x160, 0x100, 0x0, 0x0}},
    {41335, 2, {0x0, 0x380, 0x440, 0x0, 0xC0, 0x22B0, 0x228C, 0x880, 0x880, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41336, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x22B0, 0x228C, 0x880, 0x880, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41337, 2, {0x380, 0x440, 0x0, 0x180, 0x540, 0x300, 0x0, 0xEE0, 0x100, 0x280, 0x100, 0xEE0, 0x0, 0x180, 0x540, 0x300}},
    {41338, 2, {0x0, 0x180, 0x540, 0x300, 0x0, 0xEE0, 0x100, 0x280, 0x100, 0xEE0, 0x0, 0x180, 0x540, 0x300, 0x0, 0x0}},
    {41339, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20C4, 0x12A8, 0x1188, 0x810, 0xFF0, 0x810, 0x10C8, 0x12A8, 0x2184, 0x0, 0x0}},
    {41340, 2, {0x0, 0x0, 0x0, 0x360, 0x458, 0x3346, 0x4040, 0x3340, 0x440, 0x340, 0x1840, 0x1640, 0x11C0, 0x1000, 0x0, 0x0}},
    {41341, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0x50, 0x48, 0x40, 0x23C8, 0x2448, 0x840, 0x840, 0x2448, 0x23C8, 0x0, 0x0}},
    {41342, 2, {0x0, 0x380, 0x440, 0x0, 0x90, 0xA90, 0xA90, 0xA90, 0xF0, 0xE90, 0x90, 0x490, 0x290, 0x190, 0x0, 0x0}},
    {41343, 2, {0x0, 0x0, 0x0, 0x0, 0x90, 0xA90, 0xA90, 0xA90, 0xF0, 0xE90, 0x90, 0x490, 0x290, 0x190, 0x0, 0x0}},
    {41344, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x90, 0x80, 0x7A0, 0x8C0, 0x1180, 0x1280, 0xC80, 0x780, 0x0, 0x0}},
    {41345, 2, {0x0, 0x380, 0x440, 0x0, 0xE0, 0x110, 0x110, 0x710, 0x590, 0x1550, 0xD30, 0x110, 0x110, 0xE0, 0x0, 0x0}},
    {41346, 2, {0x0, 0x0, 0x0, 0x0, 0xE0, 0x110, 0x110, 0x710, 0x590, 0x1550, 0xD30, 0x110, 0x110, 0xE0, 0x0, 0x0}},
    {41347, 2, {0x0, 0x0, 0x0, 0x0, 0x600, 0x580, 0x400, 0x798, 0x460, 0x5C0, 0x640, 0x780, 0x3400, 0xC00, 0x0, 0x0}},
    {41348, 2, {0x380, 0x440, 0x0, 0x660, 0x550, 0x14C8, 0xC60, 0x50, 0x40, 0x1F0, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41349, 2, {0x0, 0x0, 0x0, 0x660, 0x550, 0x14C8, 0xC60, 0x50, 0x40, 0x1F0, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41350, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x17C0, 0x1100, 0x17C0, 0x100, 0xD00, 0xB00, 0x800, 0x0, 0x0}},
    {41351, 2, {0x0, 0x380, 0x440, 0x0, 0x1D80, 0x360, 0x118, 0x100, 0x300, 0x1D00, 0x100, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41352, 2, {0x0, 0x0, 0x0, 0x0, 0x1D80, 0x360, 0x118, 0x100, 0x300, 0x1D00, 0x100, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41353, 2, {0x0, 0x0, 0x0, 0x0, 0x8E0, 0x910, 0x910, 0x950, 0x950, 0x950, 0x950, 0x910, 0x910, 0x8E0, 0x0, 0x0}},
    {41354, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x890, 0x890, 0xFF0, 0x890, 0x890, 0xFF0, 0x890, 0x890, 0x700, 0x0, 0x0}},
    {41355, 2, {0x380, 0x440, 0x0, 0x3E60, 0x4158, 0x846, 0x3E40, 0x840, 0x4140, 0x3E40, 0x40, 0xE40, 0x9C0, 0x800, 0x0, 0x0}},
    {41356, 2, {0x0, 0x0, 0x0, 0x3E60, 0x4158, 0x846, 0x3E40, 0x840, 0x4140, 0x3E40, 0x40, 0xE40, 0x9C0, 0x800, 0x0, 0x0}},
    {41357, 2, {0x0, 0x0, 0x0, 0x0, 0x24C0, 0x18B0, 0x8C, 0x2480, 0x1880, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41358, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0xD00, 0x300, 0x100, 0xD00, 0x300, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41359, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0xD00, 0x300, 0x100, 0xD00, 0x300, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41360, 2, {0x0, 0x0, 0x0, 0x0, 0x3180, 0xD60, 0x318, 0x3100, 0xD00, 0x300, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41361, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xB0, 0x80, 0xA80, 0xA80, 0xA80, 0x80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41362, 2, {0x380, 0x440, 0x0, 0x12C0, 0xCB0, 0x80, 0x1280, 0xC80, 0x80, 0x1280, 0xC80, 0x80, 0x1C80, 0x380, 0x0, 0x0}},
    {41363, 2, {0x0, 0x0, 0x0, 0x12C0, 0xCB0, 0x80, 0x1280, 0xC80, 0x80, 0x1280, 0xC80, 0x80, 0x1C80, 0x380, 0x0, 0x0}},
    {41364, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0xA58, 0xA40, 0xA40, 0x40, 0x1140, 0xE40, 0x40, 0x340, 0xC0, 0x0, 0x0}},
    {41365, 2, {0x0, 0x380, 0x440, 0x0, 0xC0, 0x14B0, 0x1480, 0x1480, 0xE0, 0x1C90, 0x90, 0x90, 0x680, 0x180, 0x0, 0x0}},
    {41366, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x14B0, 0x1480, 0x1480, 0xE0, 0x1C90, 0x90, 0x90, 0x680, 0x180, 0x0, 0x0}},
    {41367, 2, {0x0, 0x0, 0x410, 0x410, 0x3E0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0x3E0, 0x490, 0x490, 0x0, 0x0}},
    {41368, 2, {0x380, 0x440, 0x0, 0x820, 0x820, 0x7C0, 0x110, 0xFF0, 0x1100, 0x1100, 0xFF0, 0x100, 0x7C0, 0x820, 0x820, 0x0}},
    {41369, 2, {0x0, 0x0, 0x820, 0x820, 0x7C0, 0x110, 0xFF0, 0x1100, 0x1100, 0xFF0, 0x100, 0x7C0, 0x820, 0x820, 0x0, 0x0}},
    {41370, 2, {0x0, 0x0, 0x0, 0x0, 0x17D0, 0x820, 0x17D0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x1010, 0xFE0, 0x0, 0x0}},
    {41371, 2, {0x380, 0x440, 0x0, 0x380, 0x440, 0x440, 0x380, 0x3110, 0x2D10, 0x2310, 0x21D0, 0x2130, 0x380, 0x540, 0x540, 0x0}},
    {41372, 2, {0x0, 0x0, 0x1C0, 0x220, 0x220, 0x1C0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x1C0, 0x2A0, 0x2A0, 0x0, 0x0}},
    {41373, 2, {0x0, 0x0, 0x0, 0x0, 0x890, 0x710, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0x700, 0xA80, 0xA80, 0x0, 0x0}},
    {41374, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x880, 0x880, 0x880, 0xB90, 0xAD0, 0xAB0, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41375, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0x790, 0x890, 0x1080, 0x1090, 0x1090, 0x1080, 0x890, 0x790, 0x80, 0x0, 0x0}},
    {41376, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x790, 0x890, 0x1080, 0x1090, 0x1090, 0x1080, 0x890, 0x790, 0x80, 0x0, 0x0}},
    {41377, 2, {0x0, 0x0, 0x0, 0x0, 0x780, 0x840, 0x1020, 0xA8, 0xA8, 0xA8, 0x70, 0x1C20, 0x1320, 0x10C0, 0x0, 0x0}},
    {41378, 2, {0x380, 0x440, 0x0, 0xEE0, 0x100, 0x280, 0x440, 0x440, 0xFE0, 0x440, 0x540, 0x540, 0x280, 0x100, 0xEE0, 0x0}},
    {41379, 2, {0x0, 0x0, 0xEE0, 0x100, 0x280, 0x440, 0x440, 0xFE0, 0x440, 0x540, 0x540, 0x280, 0x100, 0xEE0, 0x0, 0x0}},
    {41380, 2, {0x0, 0x0, 0x0, 0x0, 0x27C8, 0x2828, 0x1010, 0x1830, 0x17D0, 0x1010, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41381, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x124, 0x124, 0x1F8, 0x120, 0x124, 0x124, 0x31F8, 0xD20, 0x320, 0x0, 0x0}},
    {41382, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x124, 0x124, 0x1F8, 0x120, 0x124, 0x124, 0x31F8, 0xD20, 0x320, 0x0, 0x0}},
    {41383, 2, {0x0, 0x0, 0x0, 0x0, 0x27C8, 0x2828, 0x1010, 0x1830, 0x17D0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41384, 2, {0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1E10, 0x11D0, 0x1030, 0x11D0, 0x1E10, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41385, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1E10, 0x11D0, 0x1030, 0x11D0, 0x1E10, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41386, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41387, 2, {0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x1D00, 0x1300, 0x1000, 0x0, 0x0}},
    {41388, 2, {0x0, 0x380, 0x440, 0x0, 0x4F0, 0x900, 0x900, 0x480, 0x70, 0x480, 0x900, 0x900, 0x900, 0x4F0, 0x0, 0x0}},
    {41389, 2, {0x0, 0x0, 0x0, 0x0, 0x4F0, 0x900, 0x900, 0x480, 0x70, 0x480, 0x900, 0x900, 0x900, 0x4F0, 0x0, 0x0}},
    {41390, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x100, 0x900, 0x8E0, 0x100, 0x900, 0x8E0, 0x0, 0x0}},
    {41391, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1450, 0x820, 0x17D0, 0x2008, 0x0}},
    {41392, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1450, 0x820, 0x17D0, 0x2008, 0x0}},
    {41393, 2, {0x0, 0x0, 0x0, 0x0, 0x110, 0x110, 0xE0, 0x40, 0x340, 0x2C0, 0x200, 0x700, 0x880, 0x880, 0x0, 0x0}},
    {41394, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x3900, 0x2700, 0x2100, 0x2100, 0x0, 0x0}},
    {41395, 2, {0x0, 0x0, 0x0, 0x0, 0x888, 0x888, 0x888, 0x490, 0x3E0, 0x80, 0x1C80, 0x1380, 0x1080, 0x1080, 0x0, 0x0}},
    {41396, 2, {0x0, 0x0, 0x0, 0x0, 0x1C0, 0x220, 0x630, 0x528, 0x420, 0x420, 0x420, 0x420, 0x14A0, 0xC60, 0x0, 0x0}},
    {41397, 2, {0x0, 0x380, 0x440, 0x0, 0x2C0, 0x1AB0, 0x168C, 0x5380, 0x3280, 0x280, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41398, 2, {0x0, 0x0, 0x0, 0x0, 0x2C0, 0x1AB0, 0x168C, 0x5380, 0x3280, 0x280, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41399, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x190, 0x170, 0x100, 0x3148, 0x2D48, 0x23C8, 0x2178, 0x140, 0x100, 0x0, 0x0}},
    {41400, 2, {0x0, 0x380, 0x440, 0x0, 0x40, 0x440, 0x440, 0x440, 0x440, 0x440, 0x40, 0xC40, 0x340, 0xC0, 0x0, 0x0}},
    {41401, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x440, 0x440, 0x440, 0x440, 0x440, 0x40, 0xC40, 0x340, 0xC0, 0x0, 0x0}},
    {41402, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0x1110, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41403, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x220, 0x220, 0xAA0, 0x8A0, 0x8A0, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41404, 2, {0x0, 0x380, 0x440, 0x0, 0x40, 0x440, 0x440, 0x440, 0x40, 0x440, 0xA40, 0xA40, 0x640, 0x980, 0x0, 0x0}},
    {41405, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x440, 0x440, 0x440, 0x40, 0x440, 0xA40, 0xA40, 0x640, 0x980, 0x0, 0x0}},
    {41406, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x620, 0x920, 0x920, 0x920, 0x620, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41407, 2, {0x0, 0x0, 0x0, 0x0, 0x1020, 0x1028, 0x1028, 0x1020, 0x1028, 0x1FE8, 0x1020, 0x1028, 0x1028, 0x1020, 0x0, 0x0}},
    {41408, 2, {0x0, 0x380, 0x440, 0x0, 0xC20, 0xA20, 0x920, 0x820, 0x820, 0x820, 0x820, 0x920, 0x8A0, 0x860, 0x0, 0x0}},
    {41409, 2, {0x0, 0x0, 0x0, 0x0, 0xC20, 0xA20, 0x920, 0x820, 0x820, 0x820, 0x820, 0x920, 0x8A0, 0x860, 0x0, 0x0}},
    {41410, 2, {0x0, 0x0, 0x0, 0xC0, 0xB0, 0x188C, 0x23E0, 0x1880, 0x23E0, 0x1880, 0x80, 0xE80, 0x980, 0x800, 0x0, 0x0}},
    {41411, 2, {0x0, 0x550, 0x550, 0x10, 0x3D0, 0x430, 0x810, 0x800, 0x800, 0x800, 0x800, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {41412, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x380, 0x540, 0x540, 0x1390, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41413, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x380, 0x540, 0x540, 0x1390, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41414, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0xFE0, 0x1450, 0x2448, 0x2448, 0xFE0, 0x1450, 0x2448, 0x2388, 0x0, 0x0}},
    {41415, 2, {0x0, 0x0, 0x0, 0x0, 0x1450, 0x1550, 0x1110, 0x820, 0x7C0, 0x100, 0xFE0, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {41416, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0x7C0, 0x920, 0x920, 0x7C0, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41417, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x7C0, 0x920, 0x920, 0x7C0, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41418, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x540, 0x540, 0x920, 0xFE0, 0x920, 0xFE0, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {41419, 2, {0x0, 0x380, 0x440, 0x0, 0x420, 0x920, 0x920, 0x820, 0x420, 0x20, 0x620, 0x5A0, 0x460, 0x400, 0x0, 0x0}},
    {41420, 2, {0x0, 0x0, 0x0, 0x0, 0x420, 0x920, 0x920, 0x820, 0x420, 0x20, 0x620, 0x5A0, 0x460, 0x400, 0x0, 0x0}},
    {41421, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1708, 0x10E8, 0x1018, 0x100, 0x100, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {41422, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1290, 0x1290, 0x1010, 0x1010, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41423, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x440, 0x440, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {41424, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x440, 0x440, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {41425, 2, {0x0, 0x0, 0x0, 0x1004, 0x808, 0x1C10, 0x2220, 0x4940, 0x4880, 0x4940, 0x2220, 0x1C10, 0x808, 0x1004, 0x0, 0x0}},
    {41426, 2, {0x0, 0x380, 0x440, 0x2008, 0x1110, 0x920, 0x440, 0x280, 0x100, 0x280, 0x440, 0x920, 0x1110, 0x2008, 0x0, 0x0}},
    {41427, 2, {0x0, 0x0, 0x0, 0x2008, 0x1110, 0x920, 0x440, 0x280, 0x100, 0x280, 0x440, 0x920, 0x1110, 0x2008, 0x0, 0x0}},
    {41428, 2, {0x0, 0x0, 0x0, 0x0, 0x1084, 0x1084, 0x888, 0x7F0, 0x80, 0xC80, 0xB80, 0x880, 0x6A80, 0x1A80, 0x0, 0x0}},
    {41429, 2, {0x0, 0x380, 0x440, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x0, 0x410, 0x3E0, 0x0, 0x0}},
    {41430, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x0, 0x410, 0x3E0, 0x0, 0x0}},
    {41431, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x680, 0x580, 0x400, 0x0, 0x0}},
    {41432, 2, {0x0, 0x3C0, 0x420, 0x0, 0x3C0, 0x420, 0x400, 0x400, 0x200, 0x180, 0x40, 0x20, 0x420, 0x3C0, 0x0, 0x0}},
    {41433, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x400, 0x400, 0x200, 0x180, 0x40, 0x20, 0x420, 0x3C0, 0x0, 0x0}},
    {41434, 2, {0x0, 0x0, 0x0, 0x0, 0xF00, 0x1080, 0x2048, 0x48, 0x48, 0x78, 0x48, 0x3848, 0x2648, 0x2180, 0x0, 0x0}},
    {41435, 2, {0x0, 0x7C0, 0x820, 0x0, 0x3C0, 0x420, 0x810, 0x110, 0x110, 0x110, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41436, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x810, 0x110, 0x110, 0x110, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41437, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0x100, 0x7C0, 0x2108, 0x2108, 0x1010, 0xFE0, 0x0, 0x0}},
    {41438, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0x1110, 0x1110, 0x100, 0x7C0, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41439, 2, {0x0, 0x380, 0x440, 0x0, 0x1450, 0x1450, 0x1390, 0x820, 0x7C0, 0x100, 0x100, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {41440, 2, {0x0, 0x0, 0x0, 0x0, 0x1450, 0x1450, 0x1390, 0x820, 0x7C0, 0x100, 0x100, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {41441, 2, {0x0, 0x0, 0x0, 0x0, 0x188, 0x168, 0x118, 0x100, 0x1FF0, 0x100, 0x2548, 0x2548, 0x1010, 0xFE0, 0x0, 0x0}},
    {41442, 2, {0x0, 0x0, 0x0, 0x0, 0x1450, 0x1390, 0x820, 0x7C0, 0x0, 0x0, 0x7C0, 0x820, 0x1390, 0x1450, 0x0, 0x0}},
    {41443, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x0, 0x0}},
    {41444, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x0, 0x0}},
    {41445, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1450, 0x1390, 0x1010, 0x1390, 0x1450, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41446, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4060, 0x20A0, 0x112C, 0xA30, 0x430, 0xA30, 0x1130, 0x20AC, 0x4060, 0x0, 0x0}},
    {41447, 2, {0x0, 0x380, 0x440, 0x0, 0x0, 0x406C, 0x20B0, 0x1130, 0xA2C, 0x420, 0xA2C, 0x1130, 0x20B0, 0x406C, 0x0, 0x0}},
    {41448, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x406C, 0x20B0, 0x1130, 0xA2C, 0x420, 0xA2C, 0x1130, 0x20B0, 0x406C, 0x0, 0x0}},
    {41449, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1290, 0x1290, 0x1010, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41450, 2, {0x0, 0x0, 0x0, 0x0, 0x4444, 0x4444, 0x4444, 0x4444, 0x3BB8, 0x1110, 0x1550, 0x1550, 0x1110, 0xEE0, 0x0, 0x0}},
    {41451, 2, {0x0, 0x380, 0x440, 0x0, 0x3100, 0xD08, 0x110, 0x3120, 0xD40, 0x180, 0x100, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41452, 2, {0x0, 0x0, 0x0, 0x0, 0x3100, 0xD08, 0x110, 0x3120, 0xD40, 0x180, 0x100, 0x3100, 0xD00, 0x300, 0x0, 0x0}},
    {41453, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x820, 0x0, 0x1FF0, 0x0, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41454, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x110, 0x1F0, 0x3110, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41455, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0x900, 0x900, 0x900, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41456, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x900, 0x900, 0x900, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41457, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x960, 0x918, 0x900, 0x100, 0x100, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41458, 2, {0x0, 0x0, 0x0, 0xE0, 0x1910, 0x550, 0x350, 0x310, 0x3F0, 0x310, 0x350, 0x550, 0x1910, 0xE0, 0x0, 0x0}},
    {41459, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41460, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41461, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1450, 0x17D0, 0x1450, 0x1450, 0x1390, 0x820, 0x7C0, 0x0, 0x0}},
    {41462, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1450, 0x820, 0x17D0, 0x2008, 0x0}},
    {41463, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1450, 0x820, 0x17D0, 0x2008, 0x0}},
    {41464, 2, {0x0, 0x0, 0x0, 0x0, 0x400, 0x5E0, 0x400, 0x5E0, 0x400, 0x5E0, 0x400, 0x460, 0x580, 0x600, 0x0, 0x0}},
    {41465, 2, {0x0, 0x380, 0x440, 0x0, 0x1C0, 0x220, 0x220, 0x410, 0x1F10, 0x410, 0x410, 0x490, 0x490, 0x360, 0x0, 0x0}},
    {41466, 2, {0x0, 0x0, 0x0, 0x0, 0x1C0, 0x220, 0x220, 0x410, 0xF10, 0x410, 0x410, 0x490, 0x490, 0x360, 0x0, 0x0}},
    {41467, 2, {0x0, 0x0, 0x0, 0x1FF0, 0x1010, 0x920, 0x540, 0x280, 0x100, 0x280, 0x540, 0x920, 0x1010, 0x1FF0, 0x0, 0x0}},
    {41468, 2, {0x0, 0x380, 0x440, 0x0, 0x8, 0x188, 0x168, 0x118, 0x3100, 0x2D08, 0x2388, 0x2168, 0x2118, 0x100, 0x0, 0x0}},
    {41469, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0x3100, 0x2D08, 0x2388, 0x2168, 0x2118, 0x100, 0x0, 0x0}},
    {41470, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x3118, 0x2928, 0x2548, 0x2388, 0x2388, 0x2548, 0x2928, 0x3118, 0x100, 0x0, 0x0}},
    {41471, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x260, 0x1218, 0x1240, 0x1240, 0x1E40, 0x1240, 0x1218, 0x1260, 0x380, 0x0, 0x0}},
    {41472, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x1018, 0x828, 0x448, 0x288, 0x1FF8, 0x288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {41473, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1018, 0x828, 0x448, 0x288, 0x1FF8, 0x288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {41474, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x960, 0x918, 0x900, 0x110, 0x1F0, 0x3110, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41475, 2, {0x0, 0x380, 0x440, 0x0, 0x60, 0x1158, 0x1140, 0x1140, 0x440, 0x470, 0x448, 0x48, 0x348, 0xC0, 0x0, 0x0}},
    {41476, 2, {0x0, 0x0, 0x0, 0x0, 0x60, 0x1158, 0x1140, 0x1140, 0x440, 0x470, 0x448, 0x48, 0x348, 0xC0, 0x0, 0x0}},
    {41477, 2, {0x0, 0x0, 0x180, 0x160, 0x100, 0x380, 0x540, 0x380, 0x100, 0xD00, 0x300, 0x7C0, 0x820, 0x820, 0x0, 0x0}},
    {41478, 2, {0x0, 0x0, 0x0, 0x0, 0x310, 0x2D0, 0x1E30, 0x200, 0x200, 0x3C0, 0x200, 0x1E00, 0x200, 0x200, 0x0, 0x0}},
    {41479, 2, {0x380, 0x440, 0x0, 0x1800, 0x6C0, 0xB0, 0x1880, 0x680, 0x80, 0x18E0, 0x690, 0x90, 0x690, 0x180, 0x0, 0x0}},
    {41480, 2, {0x0, 0x0, 0x0, 0x1800, 0x6C0, 0xB0, 0x1880, 0x680, 0x80, 0x18E0, 0x690, 0x90, 0x690, 0x180, 0x0, 0x0}},
    {41481, 2, {0x0, 0x0, 0x0, 0x300, 0x2C0, 0x230, 0x200, 0xFF0, 0x1200, 0x1200, 0x1200, 0xFF0, 0x200, 0x200, 0x0, 0x0}},
    {41482, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41483, 2, {0x0, 0x380, 0x440, 0x0, 0x820, 0x820, 0x820, 0x820, 0x0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41484, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x820, 0x820, 0x0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41485, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x1110, 0x1110, 0x1110, 0x1110, 0x0, 0x0}},
    {41486, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x0, 0x0}},
    {41487, 2, {0x0, 0x380, 0x440, 0x0, 0x10, 0x1A50, 0x1650, 0x1250, 0x1350, 0x12D0, 0x10, 0x310, 0xD0, 0x30, 0x0, 0x0}},
    {41488, 2, {0x0, 0x0, 0x0, 0x210, 0x1A50, 0x1650, 0x1250, 0x1350, 0x12D0, 0x210, 0x10, 0x310, 0xD0, 0x30, 0x0, 0x0}},
    {41489, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x490, 0x490, 0x80, 0x80, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41490, 2, {0x380, 0x440, 0x0, 0x40, 0x640, 0x540, 0x440, 0x1F40, 0x440, 0x1440, 0xC40, 0x40, 0x340, 0xC0, 0x0, 0x0}},
    {41491, 2, {0x0, 0x0, 0x0, 0x20, 0x320, 0x2A0, 0x220, 0xFA0, 0x220, 0xA20, 0x620, 0x20, 0x1A0, 0x60, 0x0, 0x0}},
    {41492, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0xC60, 0x1010, 0x1110, 0x1110, 0xD60, 0x100, 0x0, 0x0}},
    {41493, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x8A0, 0x520, 0x520, 0x520, 0x8A0, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41494, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x8A0, 0x520, 0x520, 0x520, 0x8A0, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41495, 2, {0x0, 0x380, 0x440, 0x0, 0xF60, 0x10D0, 0x1648, 0x940, 0x1640, 0x1040, 0x1640, 0x940, 0x1640, 0x40, 0x0, 0x0}},
    {41496, 2, {0x0, 0x0, 0x0, 0x0, 0xF60, 0x10D0, 0x1648, 0x940, 0x1640, 0x1040, 0x1640, 0x940, 0x1640, 0x40, 0x0, 0x0}},
    {41497, 2, {0x0, 0x0, 0x0, 0x0, 0x400, 0x400, 0x400, 0x1C08, 0x1608, 0x1188, 0x1068, 0x1038, 0x20, 0x20, 0x0, 0x0}},
    {41498, 2, {0x0, 0x0, 0x0, 0x0, 0x208, 0x208, 0x208, 0x208, 0x3F8, 0x208, 0x208, 0x1A08, 0x1608, 0x1008, 0x0, 0x0}},
    {41499, 2, {0x380, 0x440, 0x0, 0x7E0, 0x810, 0x800, 0x7E0, 0x10, 0x810, 0x7E0, 0x80, 0x7F0, 0x80, 0x7F0, 0x80, 0x0}},
    {41500, 2, {0x0, 0x0, 0x0, 0x7E0, 0x810, 0x800, 0x7E0, 0x10, 0x810, 0x7E0, 0x80, 0x7F0, 0x80, 0x7F0, 0x80, 0x0}},
    {41501, 2, {0x0, 0x0, 0x0, 0x0, 0x2138, 0x2140, 0x2180, 0x2180, 0x180, 0x140, 0x3138, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41502, 2, {0x0, 0x0, 0x0, 0x0, 0x840, 0x840, 0x840, 0x40, 0xC40, 0x340, 0xC0, 0x1E0, 0x210, 0x210, 0x0, 0x0}},
    {41503, 2, {0x0, 0x380, 0x440, 0x0, 0x1830, 0x2448, 0x440, 0x440, 0x440, 0x440, 0xFE0, 0x1450, 0x2448, 0x2388, 0x0, 0x0}},
    {41504, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x2448, 0x440, 0x440, 0x440, 0x440, 0xFE0, 0x1450, 0x2448, 0x2388, 0x0, 0x0}},
    {41505, 2, {0x0, 0x0, 0xFE0, 0x1010, 0x1110, 0x1110, 0x1010, 0xFE0, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x500, 0x300, 0x0, 0x0}},
    {41506, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x3108, 0x2D08, 0x2388, 0x2168, 0x2118, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41507, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x820, 0x820, 0xE20, 0x9A0, 0x860, 0x0, 0x0}},
    {41508, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x820, 0x820, 0xE20, 0x9A0, 0x860, 0x0, 0x0}},
    {41509, 2, {0x0, 0x0, 0x0, 0x0, 0x840, 0x840, 0x840, 0x840, 0x40, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41510, 2, {0x0, 0x0, 0x0, 0x0, 0xAC0, 0xAB0, 0x80, 0x1F80, 0x80, 0xA80, 0xA80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {41511, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1010, 0x1010, 0x1110, 0x1110, 0x1010, 0x1010, 0x1110, 0x1110, 0x0, 0x0}},
    {41512, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1010, 0x1010, 0x1110, 0x1110, 0x1010, 0x1010, 0x1110, 0x1110, 0x0, 0x0}},
    {41513, 2, {0x0, 0x0, 0x0, 0x0, 0x1960, 0x2558, 0x2440, 0x27C0, 0x2440, 0x2540, 0x2540, 0x1840, 0x340, 0xC0, 0x0, 0x0}},
    {41514, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x988, 0x1568, 0x1518, 0x1500, 0xF00, 0x500, 0x900, 0x1100, 0x100, 0x0, 0x0}},
    {41515, 2, {0x380, 0x440, 0x0, 0x1830, 0x6D0, 0x110, 0x6D0, 0x1830, 0x0, 0x1830, 0x6D0, 0x110, 0x6D0, 0x1830, 0x0, 0x0}},
    {41516, 2, {0x0, 0x0, 0x0, 0x1830, 0x6D0, 0x110, 0x6D0, 0x1830, 0x0, 0x1830, 0x6D0, 0x110, 0x6D0, 0x1830, 0x0, 0x0}},
    {41517, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x6D0, 0x110, 0x6D0, 0x1830, 0x100, 0x100, 0x1FF0, 0x100, 0x100, 0x0, 0x0}},
    {41518, 2, {0x380, 0x440, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x0, 0x0}},
    {41519, 2, {0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x0, 0x0}},
    {41520, 2, {0x0, 0x380, 0x440, 0x0, 0x618, 0x3D94, 0x4472, 0x4410, 0x2410, 0x1C10, 0x2410, 0x4410, 0x4410, 0x3C10, 0x0, 0x0}},
    {41521, 2, {0x0, 0x0, 0x0, 0x0, 0x618, 0x3D94, 0x4472, 0x4410, 0x2410, 0x1C10, 0x2410, 0x4410, 0x4410, 0x3C10, 0x0, 0x0}},
    {41522, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x410, 0x410, 0x7F0, 0x410, 0x510, 0x610, 0xC10, 0x1410, 0x3E0, 0x0, 0x0}},
    {41523, 2, {0x0, 0x380, 0x440, 0x0, 0x1860, 0x1658, 0x1040, 0x1040, 0x1040, 0x1FC0, 0x1240, 0x1240, 0x1240, 0x1240, 0x0, 0x0}},
    {41524, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1658, 0x1040, 0x1040, 0x1040, 0x1FC0, 0x1240, 0x1240, 0x1240, 0x1240, 0x0, 0x0}},
    {41525, 2, {0x0, 0x0, 0x0, 0x0, 0x10C, 0x110, 0x120, 0x120, 0x1E0, 0x120, 0x3120, 0x2D10, 0x230C, 0x2000, 0x0, 0x0}},
    {41526, 2, {0x380, 0x440, 0x100, 0xFE0, 0x1110, 0x1110, 0xFE0, 0x2108, 0x2108, 0xFE0, 0x1110, 0x1110, 0xFE0, 0x100, 0x0, 0x0}},
    {41527, 2, {0x0, 0x0, 0x100, 0xFE0, 0x1110, 0x1110, 0xFE0, 0x2108, 0x2108, 0xFE0, 0x1110, 0x1110, 0xFE0, 0x100, 0x0, 0x0}},
    {41528, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41529, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0xFE0, 0x920, 0x7C0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41530, 2, {0x380, 0x440, 0x0, 0x3E0, 0x410, 0x410, 0x410, 0x3E0, 0x80, 0x380, 0x200, 0xF80, 0x1240, 0x1240, 0x0, 0x0}},
    {41531, 2, {0x0, 0x0, 0x0, 0x3E0, 0x410, 0x410, 0x410, 0x3E0, 0x80, 0x380, 0x200, 0xF80, 0x1240, 0x1240, 0x0, 0x0}},
    {41532, 2, {0x0, 0x0, 0x0, 0x0, 0x78, 0x80, 0x380, 0x440, 0x1E38, 0x440, 0x380, 0x80, 0x80, 0x78, 0x0, 0x0}},
    {41533, 2, {0x380, 0x440, 0x0, 0x70, 0x80, 0x90, 0x790, 0x880, 0x1270, 0x1200, 0x1270, 0x880, 0x790, 0x90, 0x80, 0x70}},
    {41534, 2, {0x0, 0x70, 0x80, 0x90, 0x790, 0x880, 0x1270, 0x1200, 0x1270, 0x880, 0x790, 0x90, 0x80, 0x70, 0x0, 0x0}},
    {41535, 2, {0x0, 0x0, 0x0, 0x0, 0xFB8, 0x1040, 0x20A0, 0x20A0, 0x20B8, 0x20A0, 0x20A0, 0x20A0, 0x1040, 0xFB8, 0x0, 0x0}},
    {41536, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x2AA8, 0x1AB0, 0x1930, 0x1830, 0x1AB0, 0x1AB0, 0x2928, 0x820, 0x7C0, 0x0, 0x0}},
    {41537, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x6E0, 0x100, 0x2C0, 0x200, 0xFE0, 0x200, 0x2C0, 0x100, 0x6E0, 0x0, 0x0}},
    {41538, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6E0, 0x100, 0x2C0, 0x200, 0xFE0, 0x200, 0x2C0, 0x100, 0x6E0, 0x0, 0x0}},
    {41539, 2, {0x0, 0x0, 0x0, 0x0, 0x410, 0x410, 0x3E0, 0x80, 0x7F0, 0x80, 0x7F0, 0x1080, 0x1080, 0xF00, 0x0, 0x0}},
    {41540, 2, {0x0, 0x380, 0x440, 0x0, 0x1410, 0x1410, 0x1410, 0x1590, 0x1690, 0x1490, 0x1090, 0x1090, 0x8A0, 0x7C0, 0x0, 0x0}},
    {41541, 2, {0x0, 0x0, 0x0, 0x0, 0x1410, 0x1410, 0x1410, 0x1590, 0x1690, 0x1490, 0x1090, 0x1090, 0x8A0, 0x7C0, 0x0, 0x0}},
    {41542, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0xE10, 0x990, 0x870, 0x0, 0x0, 0xE10, 0x990, 0x870, 0x0, 0x0, 0x0}},
    {41543, 2, {0x380, 0x440, 0x0, 0x100, 0x3118, 0x2928, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2928, 0x3118, 0x100, 0x0, 0x0}},
    {41544, 2, {0x0, 0x0, 0x0, 0x100, 0x3118, 0x2928, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2928, 0x3118, 0x100, 0x0, 0x0}},
    {41545, 2, {0x0, 0x0, 0x0, 0x1FF0, 0x40, 0x7C0, 0x840, 0x840, 0x840, 0x7C0, 0x840, 0x840, 0x840, 0x7C0, 0x0, 0x0}},
    {41546, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x940, 0x940, 0x940, 0x40, 0xF40, 0x40, 0xC40, 0x340, 0xC0, 0x0, 0x0}},
    {41547, 2, {0x0, 0x380, 0x440, 0x0, 0x30, 0x40, 0x90, 0x390, 0x2F0, 0x280, 0x280, 0x240, 0x1A30, 0x600, 0x0, 0x0}},
    {41548, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x90, 0x390, 0x2F0, 0x280, 0x280, 0x240, 0x1A30, 0x600, 0x0, 0x0}},
    {41549, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x8A0, 0x8A0, 0x880, 0x8F0, 0x880, 0x8A0, 0x8A0, 0x880, 0x700, 0x0, 0x0}},
    {41550, 2, {0x0, 0x0, 0x0, 0x0, 0x10C0, 0x20B0, 0x248C, 0x2280, 0x1180, 0x80, 0x7E0, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41551, 2, {0x0, 0x380, 0x440, 0x0, 0x10C0, 0x20B0, 0x248C, 0x2280, 0x1180, 0x80, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41552, 2, {0x0, 0x0, 0x0, 0x0, 0x10C0, 0x20B0, 0x248C, 0x2280, 0x1180, 0x80, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41553, 2, {0x0, 0x0, 0x0, 0x0, 0x12C0, 0x22B0, 0x248C, 0x2280, 0x1180, 0x280, 0x280, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41554, 2, {0x0, 0x0, 0x0, 0x0, 0x2288, 0x2288, 0x2288, 0x1C70, 0x0, 0x820, 0x820, 0x820, 0x440, 0x380, 0x0, 0x0}},
    {41555, 2, {0x0, 0x380, 0x440, 0x0, 0x1E0, 0x200, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x200, 0x1E0, 0x0, 0x0}},
    {41556, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x200, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x200, 0x1E0, 0x0, 0x0}},
    {41557, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x380, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41558, 2, {0x0, 0x0, 0x0, 0x0, 0xFC0, 0x20, 0x7A0, 0x820, 0x720, 0xA0, 0xF20, 0x20, 0x20, 0xFC0, 0x0, 0x0}},
    {41559, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x900, 0x500, 0x300, 0x100, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41560, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x900, 0x500, 0x300, 0x100, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41561, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3018, 0x2828, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2828, 0x3018, 0x0, 0x0}},
    {41562, 2, {0x380, 0x440, 0x0, 0x920, 0x920, 0x7C0, 0xFE0, 0x1110, 0x1110, 0x1110, 0xFE0, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41563, 2, {0x0, 0x0, 0x0, 0x920, 0x920, 0x7C0, 0xFE0, 0x1110, 0x1110, 0x1110, 0xFE0, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41564, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x3160, 0x2D18, 0x2300, 0x2100, 0x100, 0x3100, 0x2D00, 0x2300, 0x2100, 0x0, 0x0}},
    {41565, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x560, 0x518, 0x500, 0x140, 0x940, 0x540, 0x1310, 0x1110, 0xFE0, 0x0, 0x0}},
    {41566, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x560, 0x518, 0x500, 0x140, 0x940, 0x540, 0x1310, 0x1110, 0xFE0, 0x0, 0x0}},
    {41567, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x1108, 0x1108, 0x7D3E, 0x1108, 0x1108, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41568, 2, {0x0, 0x0, 0x0, 0x0, 0x2020, 0x2020, 0x3828, 0x2628, 0x21A8, 0x2070, 0x2020, 0x2020, 0x1040, 0xF80, 0x0, 0x0}},
    {41569, 2, {0x0, 0x380, 0x440, 0x0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x680, 0x580, 0x400, 0x0, 0x0}},
    {41570, 2, {0x0, 0x0, 0x0, 0x0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x680, 0x580, 0x400, 0x0, 0x0}},
    {41571, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x8E0, 0x890, 0x3E88, 0x888, 0x888, 0x90, 0x18E0, 0x680, 0x180, 0x0, 0x0}},
    {41572, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x2448, 0x2288, 0x100, 0x100, 0x2288, 0x2448, 0x1010, 0xFE0, 0x0, 0x0}},
    {41573, 2, {0x380, 0x440, 0x0, 0x380, 0x440, 0x540, 0x540, 0x440, 0x7C0, 0x440, 0x280, 0x100, 0x280, 0x440, 0x0, 0x0}},
    {41574, 2, {0x0, 0x0, 0x0, 0x380, 0x440, 0x540, 0x540, 0x440, 0x7C0, 0x440, 0x280, 0x100, 0x280, 0x440, 0x0, 0x0}},
    {41575, 2, {0x0, 0x0, 0x0, 0x0, 0x1C70, 0x280, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x280, 0x1C70, 0x0, 0x0}},
    {41576, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x440, 0x440, 0x7C0, 0x820, 0x920, 0x920, 0x820, 0x7C0, 0x0, 0x0}},
    {41577, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x440, 0x7C0, 0x820, 0x920, 0x920, 0x820, 0x7C0, 0x0, 0x0}},
    {41578, 2, {0x0, 0x0, 0x0, 0x0, 0x88, 0x290, 0x2A0, 0x2C0, 0x12E0, 0xC90, 0x90, 0x1890, 0x6E0, 0x180, 0x0, 0x0}},
    {41579, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x220, 0x220, 0xFA0, 0x220, 0x220, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41580, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x220, 0x220, 0xFA0, 0x220, 0x220, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {41581, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1100, 0x1100, 0x1F20, 0x1150, 0x1FD0, 0x1150, 0x1F20, 0x1100, 0xE00, 0x0, 0x0}},
    {41582, 2, {0x0, 0x0, 0x0, 0x0, 0x980, 0x1160, 0x1118, 0x900, 0x700, 0x900, 0x1100, 0x1100, 0x900, 0x100, 0x0, 0x0}},
    {41583, 2, {0x0, 0x380, 0x440, 0x0, 0x1180, 0x1160, 0x918, 0x700, 0x900, 0x1100, 0x1100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41584, 2, {0x0, 0x0, 0x0, 0x0, 0x1180, 0x1160, 0x918, 0x700, 0x900, 0x1100, 0x1100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41585, 2, {0x0, 0x0, 0x0, 0x0, 0x50, 0x350, 0x2D0, 0x270, 0x240, 0x240, 0x200, 0xF80, 0x1240, 0x1240, 0x0, 0x0}},
    {41586, 2, {0x0, 0x380, 0x440, 0x0, 0x1C0, 0x2A0, 0x3E0, 0x2A0, 0x31C0, 0x2C80, 0x23C0, 0x220, 0x490, 0x490, 0x0, 0x0}},
    {41587, 2, {0x0, 0x0, 0x0, 0x0, 0xE0, 0x150, 0x1F0, 0x150, 0x18E0, 0x1640, 0x11E0, 0x110, 0x248, 0x248, 0x0, 0x0}},
    {41588, 2, {0x0, 0x0, 0x0, 0x0, 0xE0, 0x110, 0x110, 0x110, 0x18E0, 0x1640, 0x11E0, 0x110, 0x248, 0x248, 0x0, 0x0}},
    {41589, 2, {0x0, 0x0, 0x0, 0x0, 0x980, 0x1560, 0x1518, 0x900, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x100, 0x0, 0x0}},
    {41590, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x1960, 0x518, 0x1300, 0x2B00, 0x2B00, 0x1300, 0x500, 0x1900, 0x100, 0x0, 0x0}},
    {41591, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x1960, 0x518, 0x1300, 0x2B00, 0x2B00, 0x1300, 0x500, 0x1900, 0x100, 0x0, 0x0}},
    {41592, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x1FE0, 0x910, 0x1510, 0x910, 0x1FE0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41593, 2, {0x0, 0x0, 0x0, 0x0, 0xF00, 0x1038, 0x1040, 0x880, 0x780, 0x880, 0x1040, 0x1038, 0x1000, 0xF00, 0x0, 0x0}},
    {41594, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1390, 0x1450, 0x1450, 0x1450, 0x1450, 0x1390, 0x820, 0x7C0, 0x0, 0x0}},
    {41595, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1390, 0x1450, 0x1450, 0x1450, 0x1450, 0x1390, 0x820, 0x7C0, 0x0, 0x0}},
    {41596, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2020, 0x1040, 0x890, 0x528, 0x228, 0x528, 0x890, 0x1040, 0x2020, 0x0, 0x0}},
    {41597, 2, {0x0, 0x380, 0x440, 0x0, 0xC30, 0xB28, 0x8E4, 0x820, 0x820, 0x1C20, 0x2A20, 0x2A20, 0x2A20, 0x820, 0x0, 0x0}},
    {41598, 2, {0x0, 0x0, 0x0, 0x0, 0xC30, 0xB28, 0x8E4, 0x820, 0x820, 0x1C20, 0x2A20, 0x2A20, 0x2A20, 0x820, 0x0, 0x0}},
    {41599, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0xE40, 0x440, 0xE40, 0x440, 0x380, 0x440, 0x920, 0x920, 0x0, 0x0}},
    {41600, 2, {0x0, 0x0, 0x0, 0x0, 0x220, 0x220, 0xE24, 0xBA4, 0x2A74, 0x1A2C, 0x2A0, 0x2A0, 0x220, 0x1C0, 0x0, 0x0}},
    {41601, 2, {0x0, 0x380, 0x440, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x0, 0x7E0, 0x0, 0x0, 0x7E0, 0x0, 0x0}},
    {41602, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x0, 0x7E0, 0x0, 0x0, 0x7E0, 0x0, 0x0}},
    {41603, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x0, 0x420, 0x420, 0x420, 0x420, 0x0, 0x0}},
    {41604, 2, {0x0, 0x380, 0x440, 0x0, 0xE70, 0x180, 0x180, 0x250, 0x250, 0x250, 0x180, 0x580, 0x240, 0xD30, 0x0, 0x0}},
    {41605, 2, {0x0, 0x0, 0x0, 0x0, 0xE70, 0x180, 0x180, 0x250, 0x250, 0x250, 0x180, 0x580, 0x240, 0xD30, 0x0, 0x0}},
    {41606, 2, {0x0, 0x0, 0x0, 0x1C0, 0x130, 0xF80, 0x140, 0xF40, 0x140, 0xF80, 0x100, 0x380, 0x540, 0x540, 0x0, 0x0}},
    {41607, 2, {0x0, 0x380, 0x440, 0x0, 0x710, 0x890, 0x8A0, 0x7C0, 0x80, 0x3E0, 0x80, 0xBE0, 0x880, 0x700, 0x0, 0x0}},
    {41608, 2, {0x0, 0x0, 0x0, 0x0, 0x710, 0x890, 0x8A0, 0x7C0, 0x80, 0x3E0, 0x80, 0xBE0, 0x880, 0x700, 0x0, 0x0}},
    {41609, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x280, 0x2A0, 0x120, 0xAA0, 0x680, 0x640, 0x540, 0x0, 0x0}},
    {41610, 2, {0x380, 0x440, 0x0, 0x1800, 0x6C0, 0xB0, 0x1888, 0x680, 0x80, 0x1880, 0x680, 0x3E0, 0x490, 0x490, 0x0, 0x0}},
    {41611, 2, {0x0, 0x0, 0x1800, 0x6C0, 0xB0, 0x1888, 0x680, 0x80, 0x1880, 0x680, 0x80, 0x3E0, 0x490, 0x490, 0x0, 0x0}},
    {41612, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x140, 0x140, 0x1D10, 0x1390, 0x1170, 0x100, 0x100, 0x0, 0x0}},
    {41613, 2, {0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x1500, 0x1500, 0x1F00, 0x1500, 0x1500, 0x100, 0x0, 0x0}},
    {41614, 2, {0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x440, 0xAA0, 0xAA0, 0x0, 0x0}},
    {41615, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x440, 0xAA0, 0xAA0, 0x0, 0x0}},
    {41616, 2, {0x0, 0x0, 0x0, 0x0, 0x3D0, 0x430, 0x810, 0x900, 0x900, 0x900, 0x900, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {41617, 2, {0x0, 0x380, 0x440, 0x0, 0x1440, 0x1440, 0x1440, 0x1FF8, 0x1440, 0x1440, 0x1440, 0x1F0, 0x40, 0x1F0, 0x0, 0x0}},
    {41618, 2, {0x0, 0x0, 0x0, 0x0, 0x1440, 0x1440, 0x1440, 0x1FF8, 0x1440, 0x1440, 0x1440, 0x1F0, 0x40, 0x1F0, 0x0, 0x0}},
    {41619, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x1560, 0x1518, 0x1500, 0x1FE0, 0x1500, 0x1500, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41620, 2, {0x0, 0x0, 0x0, 0x8, 0x188, 0x1D68, 0x318, 0xB00, 0xB70, 0x380, 0x1DA0, 0x1A0, 0x180, 0x170, 0x0, 0x0}},
    {41621, 2, {0x0, 0x380, 0x440, 0x0, 0x78, 0x780, 0x880, 0x1078, 0x1480, 0x1480, 0x1078, 0x880, 0x780, 0x78, 0x0, 0x0}},
    {41622, 2, {0x0, 0x0, 0x0, 0x0, 0x78, 0x780, 0x880, 0x1078, 0x1480, 0x1480, 0x1078, 0x880, 0x780, 0x78, 0x0, 0x0}},
    {41623, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x0, 0x0, 0xEE0, 0x0, 0x0}},
    {41624, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x440, 0xE40, 0x440, 0x440, 0xE40, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41625, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0xE40, 0x440, 0x440, 0xE40, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41626, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x118, 0x124, 0x13C, 0x124, 0x1FC, 0x124, 0x3118, 0xD00, 0x300, 0x0, 0x0}},
    {41627, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x280, 0x440, 0x820, 0xFE0, 0x100, 0x280, 0x440, 0x820, 0xFE0, 0x0, 0x0}},
    {41628, 2, {0x0, 0x380, 0x440, 0x0, 0x804, 0x874, 0x8AC, 0x920, 0xA20, 0xC20, 0xA20, 0x6920, 0x58A0, 0x4060, 0x0, 0x0}},
    {41629, 2, {0x0, 0x0, 0x0, 0x0, 0x804, 0x874, 0x8AC, 0x920, 0xA20, 0xC20, 0xA20, 0x6920, 0x58A0, 0x4060, 0x0, 0x0}},
    {41630, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x0, 0x0}},
    {41631, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x100, 0x17D0, 0x1110, 0x1110, 0x17D0, 0x100, 0x100, 0x0, 0x0}},
    {41632, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x17D0, 0x1110, 0x1110, 0x17D0, 0x100, 0x100, 0x0, 0x0}},
    {41633, 2, {0x0, 0x0, 0x0, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0xBE0, 0xA20, 0x940, 0x880, 0x940, 0x220, 0x0, 0x0}},
    {41634, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x8A0, 0x8A0, 0x8C0, 0x880, 0x8A0, 0x8A0, 0x8C0, 0x880, 0x700, 0x0, 0x0}},
    {41635, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x640, 0x1C0, 0x640, 0x1C0, 0x0, 0x640, 0x1C0, 0x640, 0x1C0, 0x0, 0x0}},
    {41636, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x640, 0x1C0, 0x640, 0x1C0, 0x0, 0x640, 0x1C0, 0x640, 0x1C0, 0x0, 0x0}},
    {41637, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1110, 0x1110, 0x0, 0x0, 0x7C0, 0x820, 0x1110, 0x1110, 0x0, 0x0}},
    {41638, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x6030, 0x5054, 0x4894, 0x518, 0x210, 0x514, 0x894, 0x1058, 0x2030, 0x0, 0x0}},
    {41639, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6030, 0x5054, 0x4894, 0x518, 0x210, 0x514, 0x894, 0x1058, 0x2030, 0x0, 0x0}},
    {41640, 2, {0x0, 0x0, 0x0, 0x248, 0x150, 0xE0, 0x150, 0x248, 0x340, 0x12C0, 0xA80, 0x700, 0xA80, 0x1240, 0x0, 0x0}},
    {41641, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x280, 0x280, 0x100, 0x280, 0xA80, 0x840, 0x740, 0x0, 0x0}},
    {41642, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x280, 0x280, 0x100, 0x280, 0xA80, 0x840, 0x740, 0x0, 0x0}},
    {41643, 2, {0x0, 0x0, 0x0, 0x0, 0xF0, 0x900, 0x900, 0xF0, 0x900, 0x900, 0xF0, 0x900, 0x900, 0xF0, 0x0, 0x0}},
    {41644, 2, {0x0, 0x380, 0x440, 0x0, 0x3C0, 0x240C, 0x2410, 0x220, 0x21E0, 0x2220, 0x410, 0x240C, 0x2400, 0x3C0, 0x0, 0x0}},
    {41645, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x240C, 0x2410, 0x220, 0x21E0, 0x2220, 0x410, 0x240C, 0x2400, 0x3C0, 0x0, 0x0}},
    {41646, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1650, 0x11C8, 0x1040, 0x1240, 0x1240, 0x1240, 0x1040, 0x1040, 0x1040, 0x0, 0x0}},
    {41647, 2, {0x0, 0x380, 0x440, 0x0, 0x300, 0x2C0, 0x230, 0x200, 0x200, 0x200, 0x1FC0, 0x200, 0x200, 0x200, 0x0, 0x0}},
    {41648, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x230, 0x200, 0x200, 0x200, 0x1FC0, 0x200, 0x200, 0x200, 0x0, 0x0}},
    {41649, 2, {0x0, 0x0, 0x0, 0xFE0, 0x820, 0x440, 0x280, 0x280, 0x100, 0x280, 0x280, 0x440, 0x820, 0xFE0, 0x0, 0x0}},
    {41650, 2, {0x0, 0x0, 0x0, 0x1E08, 0x11CC, 0x103A, 0x1188, 0x1148, 0x1108, 0x1108, 0x1108, 0x1508, 0x5308, 0x3000, 0x0, 0x0}},
    {41651, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x450, 0x450, 0x450, 0x460, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41652, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x450, 0x450, 0x450, 0x460, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {41653, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x230, 0x200, 0x200, 0xF80, 0x1040, 0x1FC0, 0x1040, 0xF80, 0x0, 0x0}},
    {41654, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1090, 0x1690, 0x1190, 0x1090, 0x1690, 0x1190, 0x820, 0x7C0, 0x0, 0x0}},
    {41655, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1190, 0x1150, 0x1110, 0x1510, 0x1310, 0x820, 0x7C0, 0x0, 0x0}},
    {41656, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1190, 0x1150, 0x1110, 0x1510, 0x1310, 0x820, 0x7C0, 0x0, 0x0}},
    {41657, 2, {0x0, 0x0, 0x0, 0x0, 0x780, 0x840, 0x1328, 0x2A8, 0x228, 0xA30, 0x620, 0x1C20, 0x1320, 0x10C0, 0x0, 0x0}},
    {41658, 2, {0x380, 0x440, 0x0, 0x1C0, 0x130, 0x100, 0x1C0, 0x120, 0x7C0, 0x900, 0x700, 0x100, 0x1900, 0x700, 0x0, 0x0}},
    {41659, 2, {0x0, 0x0, 0x0, 0x1C0, 0x130, 0x100, 0x1C0, 0x120, 0x7C0, 0x900, 0x700, 0x100, 0x1900, 0x700, 0x0, 0x0}},
    {41660, 2, {0x0, 0x0, 0x0, 0x0, 0xC60, 0x280, 0x100, 0x280, 0x280, 0x440, 0x440, 0x920, 0x920, 0x920, 0x0, 0x0}},
    {41661, 2, {0x0, 0x0, 0x0, 0x0, 0x220, 0x220, 0x228, 0x228, 0xF28, 0x2AE8, 0x1A38, 0x228, 0x228, 0x1C0, 0x0, 0x0}},
    {41662, 2, {0x0, 0x380, 0x440, 0x0, 0x1C0, 0x220, 0x228, 0x228, 0xF28, 0x2AE8, 0x1A38, 0x228, 0x228, 0x1C0, 0x0, 0x0}},
    {41663, 2, {0x0, 0x0, 0x0, 0x0, 0x1C0, 0x220, 0x228, 0x228, 0xF28, 0x2AE8, 0x1A38, 0x228, 0x228, 0x1C0, 0x0, 0x0}},
    {41664, 2, {0x0, 0x0, 0x0, 0x0, 0x710, 0x890, 0x890, 0x890, 0x890, 0x1FF0, 0x890, 0x890, 0x890, 0x710, 0x0, 0x0}},
    {41665, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1390, 0x1050, 0x1390, 0x1050, 0x1390, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41666, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1390, 0x1050, 0x1390, 0x1050, 0x1390, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41667, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x440, 0x1550, 0x1550, 0x1110, 0xEE0, 0x0, 0x0}},
    {41668, 2, {0x0, 0x0, 0x0, 0x0, 0x11C4, 0x2224, 0x2224, 0x1224, 0xE3C, 0x1224, 0x2224, 0x2224, 0x1224, 0x1C4, 0x0, 0x0}},
    {41669, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x4060, 0x20A0, 0x1124, 0x4A24, 0x443C, 0x4A24, 0x1124, 0x20A0, 0x4060, 0x0, 0x0}},
    {41670, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4060, 0x20A0, 0x1124, 0x4A24, 0x443C, 0x4A24, 0x1124, 0x20A0, 0x4060, 0x0, 0x0}},
    {41671, 2, {0x0, 0x0, 0x0, 0x0, 0x10C0, 0x13B0, 0x1080, 0x880, 0x780, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41672, 2, {0x0, 0x380, 0x440, 0x0, 0x700, 0x890, 0x890, 0x890, 0x8F0, 0x890, 0x890, 0x890, 0x880, 0x700, 0x0, 0x0}},
    {41673, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x890, 0x890, 0x890, 0x8F0, 0x890, 0x890, 0x890, 0x880, 0x700, 0x0, 0x0}},
    {41674, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x420, 0x220, 0x120, 0xCA0, 0xA60, 0x900, 0x880, 0x840, 0x820, 0x0, 0x0}},
    {41675, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x280, 0x2A0, 0x120, 0x2A0, 0x280, 0x440, 0x440, 0x0, 0x0}},
    {41676, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x280, 0x2A0, 0x120, 0x2A0, 0x280, 0x440, 0x440, 0x0, 0x0}},
    {41677, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41678, 2, {0x0, 0x380, 0x440, 0x0, 0x1830, 0x6C0, 0x100, 0x6C0, 0x1830, 0x100, 0x100, 0x7C0, 0x100, 0x100, 0x0, 0x0}},
    {41679, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x6C0, 0x100, 0x6C0, 0x1830, 0x100, 0x100, 0x7C0, 0x100, 0x100, 0x0, 0x0}},
    {41680, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1588, 0x1468, 0x1458, 0x440, 0x7C0, 0x440, 0x440, 0x440, 0x0, 0x0}},
    {41681, 2, {0x0, 0x380, 0x440, 0x1000, 0x860, 0x2450, 0x1048, 0x840, 0x3C0, 0x440, 0x840, 0x840, 0x440, 0x3C0, 0x0, 0x0}},
    {41682, 2, {0x0, 0x0, 0x0, 0x1000, 0x860, 0x2450, 0x1048, 0x840, 0x3C0, 0x440, 0x840, 0x840, 0x440, 0x3C0, 0x0, 0x0}},
    {41683, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x620, 0x5A0, 0x460, 0x400, 0x520, 0x520, 0x400, 0x520, 0x520, 0x0, 0x0}},
    {41684, 2, {0x0, 0x0, 0x0, 0xFE0, 0x100, 0x1140, 0x920, 0x510, 0x1140, 0x920, 0x510, 0x1140, 0x920, 0x510, 0x0, 0x0}},
    {41685, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x17D0, 0x100, 0x100, 0x540, 0x540, 0x540, 0x100, 0x0, 0x0}},
    {41686, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x17D0, 0x100, 0x100, 0x540, 0x540, 0x540, 0x100, 0x0, 0x0}},
    {41687, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1118, 0x100, 0x100, 0x440, 0x440, 0x440, 0x0, 0x0}},
    {41688, 2, {0x0, 0x0, 0x0, 0x0, 0x1FC0, 0x200, 0x200, 0x1268, 0xAA8, 0x738, 0xAA8, 0x1268, 0x200, 0x1FC0, 0x0, 0x0}},
    {41689, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x1100, 0x1100, 0xA50, 0xA50, 0x400, 0xA50, 0xA50, 0x1100, 0x1100, 0x0, 0x0}},
    {41690, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1100, 0x1100, 0xA50, 0xA50, 0x400, 0xA50, 0xA50, 0x1100, 0x1100, 0x0, 0x0}},
    {41691, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1100, 0x1100, 0xA50, 0xA50, 0x400, 0xA70, 0xA00, 0x1100, 0x1100, 0x0, 0x0}},
    {41692, 2, {0x380, 0x440, 0x10, 0x310, 0x2D0, 0x230, 0xF80, 0x200, 0xF80, 0x200, 0x700, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41693, 2, {0x0, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0xF80, 0x200, 0xF80, 0x200, 0x700, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41694, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x7C0, 0x1110, 0x1110, 0x920, 0x7C0, 0x440, 0x820, 0x820, 0x0, 0x0}},
    {41695, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0x7C0, 0x100, 0x110, 0xE0, 0x0, 0x0}},
    {41696, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0xFE0, 0x820, 0xFE0, 0x820, 0x7C0, 0x100, 0x110, 0xE0, 0x0, 0x0}},
    {41697, 2, {0x0, 0x0, 0x0, 0x0, 0xF00, 0x1140, 0x1140, 0xF40, 0x1140, 0x1100, 0xF00, 0x110, 0x110, 0xE0, 0x0, 0x0}},
    {41698, 2, {0x0, 0x380, 0x440, 0x0, 0x300, 0x2C0, 0x230, 0x200, 0x700, 0x880, 0x880, 0x700, 0x1040, 0xF80, 0x0, 0x0}},
    {41699, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x230, 0x200, 0x700, 0x880, 0x880, 0x700, 0x1040, 0xF80, 0x0, 0x0}},
    {41700, 2, {0x0, 0x0, 0x180, 0x160, 0x100, 0xF28, 0x1128, 0x11F8, 0x1128, 0xF28, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41701, 2, {0x0, 0x380, 0x440, 0x0, 0x600, 0x580, 0x460, 0x400, 0x4A0, 0x4A0, 0x7E0, 0x4A0, 0x4A0, 0x400, 0x0, 0x0}},
    {41702, 2, {0x0, 0x0, 0x0, 0x0, 0x600, 0x580, 0x460, 0x400, 0x4A0, 0x4A0, 0x7E0, 0x4A0, 0x4A0, 0x400, 0x0, 0x0}},
    {41703, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x820, 0x1010, 0x440, 0x820, 0x1010, 0x440, 0x820, 0x1010, 0x0, 0x0}},
    {41704, 2, {0x380, 0x440, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0xA60, 0x6A0, 0x320, 0x6A0, 0xA60, 0x200, 0x0, 0x0}},
    {41705, 2, {0x0, 0x0, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0xA60, 0x6A0, 0x320, 0x6A0, 0xA60, 0x200, 0x0, 0x0}},
    {41706, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x220, 0x420, 0x420, 0x220, 0x1E0, 0x60, 0x9A0, 0xE20, 0x820, 0x0, 0x0}},
    {41707, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x440, 0x440, 0xFC0, 0x1440, 0x1440, 0x1440, 0x1440, 0x440, 0x0, 0x0}},
    {41708, 2, {0x0, 0x0, 0x0, 0x0, 0x1C0, 0x220, 0x220, 0x220, 0x7E0, 0xA20, 0xA20, 0xA20, 0xA20, 0x220, 0x0, 0x0}},
    {41709, 2, {0x0, 0x0, 0x0, 0x0, 0x7F0, 0x80, 0x1C0, 0x130, 0x100, 0x120, 0x120, 0x120, 0xD00, 0x300, 0x0, 0x0}},
    {41710, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x440, 0x0, 0xFE0, 0x0, 0x440, 0x440, 0x440, 0x0, 0x0}},
    {41711, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x440, 0x0, 0xFE0, 0x0, 0x440, 0x440, 0x440, 0x0, 0x0}},
    {41712, 2, {0x0, 0x0, 0x0, 0x0, 0x120, 0x910, 0x910, 0x920, 0x1C0, 0x120, 0x910, 0x910, 0x920, 0x100, 0x0, 0x0}},
    {41713, 2, {0x0, 0x380, 0x440, 0x0, 0x1120, 0x1510, 0x1510, 0x1120, 0x1FC0, 0x1120, 0x1510, 0x1510, 0x1120, 0xE00, 0x0, 0x0}},
    {41714, 2, {0x0, 0x0, 0x0, 0x0, 0x1120, 0x1510, 0x1510, 0x1120, 0x1FC0, 0x1120, 0x1510, 0x1510, 0x1120, 0xE00, 0x0, 0x0}},
    {41715, 2, {0x0, 0x0, 0x0, 0x0, 0x1120, 0x1110, 0x1110, 0x1120, 0x11C0, 0x1120, 0x1110, 0x1110, 0x1120, 0xE00, 0x0, 0x0}},
    {41716, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x7C0, 0x100, 0x0, 0x180, 0x160, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41717, 2, {0x0, 0x380, 0x440, 0x100, 0x7C0, 0x100, 0x0, 0x180, 0x160, 0x100, 0x7C0, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41718, 2, {0x0, 0x0, 0x0, 0x100, 0x7C0, 0x100, 0x0, 0x180, 0x160, 0x100, 0x7C0, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {41719, 2, {0x0, 0x0, 0x0, 0x0, 0x980, 0x1160, 0x1118, 0x900, 0x748, 0x948, 0x1100, 0x1148, 0x948, 0x100, 0x0, 0x0}},
    {41720, 2, {0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x380, 0x1110, 0x1390, 0xD60, 0x100, 0xD60, 0x380, 0xD60, 0x0, 0x0}},
    {41721, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x380, 0x1110, 0x1390, 0xD60, 0x100, 0xD60, 0x380, 0xD60, 0x0, 0x0}},
    {41722, 2, {0x0, 0x380, 0x440, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0xF80, 0x1240, 0x1240, 0x1240, 0xF80, 0x0, 0x0}},
    {41723, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0xF80, 0x1240, 0x1240, 0x1240, 0xF80, 0x0, 0x0}},
    {41724, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0x3E0, 0x80, 0x600, 0x580, 0x460, 0x400, 0x440, 0x440, 0x400, 0x0, 0x0}},
    {41725, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x3E0, 0x80, 0x600, 0x580, 0x460, 0x400, 0x440, 0x440, 0x400, 0x0, 0x0}},
    {41726, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x7C0, 0x100, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x7C0, 0x100, 0x0, 0x0}},
    {41727, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x7C0, 0x100, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x7C0, 0x100, 0x0, 0x0}},
    {41728, 2, {0x0, 0x0, 0x100, 0x7C0, 0x4104, 0x4C64, 0x3458, 0x440, 0x380, 0x100, 0x7C0, 0x100, 0x7C0, 0x100, 0x0, 0x0}},
    {41729, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41730, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41731, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x9F0, 0x1210, 0x1210, 0x910, 0xF0, 0x910, 0x1210, 0x1210, 0x9F0, 0x0, 0x0}},
    {41732, 2, {0x0, 0x380, 0x440, 0x0, 0x240, 0x280, 0x300, 0x3C0, 0x220, 0x220, 0x3C0, 0x300, 0x280, 0x240, 0x0, 0x0}},
    {41733, 2, {0x0, 0x0, 0x0, 0x0, 0x240, 0x280, 0x300, 0x3C0, 0x220, 0x220, 0x3C0, 0x300, 0x280, 0x240, 0x0, 0x0}},
    {41734, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1450, 0x1FF0, 0x1450, 0x1450, 0x1FF0, 0x1450, 0x820, 0x7C0, 0x0, 0x0}},
    {41735, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1450, 0x1450, 0x1450, 0x1010, 0x1110, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41736, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1450, 0x1450, 0x1450, 0x1010, 0x1110, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41737, 2, {0x0, 0x0, 0x0, 0x0, 0x88, 0x88, 0x88, 0x88, 0xC88, 0x10F8, 0x2088, 0x2488, 0x2288, 0x1188, 0x0, 0x0}},
    {41738, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x7C0, 0x820, 0x1110, 0x1110, 0x0, 0x0}},
    {41739, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1FF0, 0x1010, 0x1FF0, 0x820, 0x7C0, 0x820, 0x1110, 0x1110, 0x0, 0x0}},
    {41740, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x12C0, 0x1230, 0x1200, 0x200, 0x200, 0x1200, 0x1200, 0x1200, 0x200, 0x0, 0x0}},
    {41741, 2, {0x0, 0x380, 0x440, 0x0, 0x444, 0x444, 0x248, 0x1F0, 0x1840, 0x2040, 0x4040, 0x4840, 0x4640, 0x2180, 0x0, 0x0}},
    {41742, 2, {0x0, 0x0, 0x0, 0x0, 0x444, 0x444, 0x248, 0x1F0, 0x1840, 0x2040, 0x4040, 0x4840, 0x4640, 0x2180, 0x0, 0x0}},
    {41743, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x100, 0x780, 0x940, 0x640, 0x40, 0x0, 0x0}},
    {41744, 2, {0x0, 0x0, 0x0, 0x0, 0xB8, 0x8C0, 0x8C0, 0x8B8, 0x880, 0xB8, 0xC0, 0x18C0, 0x6B8, 0x180, 0x0, 0x0}},
    {41745, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0x890, 0x8F0, 0x890, 0x880, 0x80, 0x90, 0x18F0, 0x690, 0x180, 0x0, 0x0}},
    {41746, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x890, 0x8F0, 0x890, 0x880, 0x80, 0x90, 0x18F0, 0x690, 0x180, 0x0, 0x0}},
    {41747, 2, {0x0, 0xF90, 0x1050, 0x10, 0x3D0, 0x430, 0x810, 0x800, 0x800, 0x800, 0x800, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {41748, 2, {0x0, 0x380, 0x440, 0x0, 0x600, 0x580, 0x460, 0x400, 0x520, 0x520, 0x520, 0x520, 0x400, 0x400, 0x0, 0x0}},
    {41749, 2, {0x0, 0x0, 0x0, 0x0, 0x600, 0x580, 0x460, 0x400, 0x520, 0x520, 0x520, 0x520, 0x400, 0x400, 0x0, 0x0}},
    {41750, 2, {0x0, 0x0, 0x0, 0x0, 0xC00, 0xB00, 0x8C0, 0x800, 0xA20, 0x940, 0x880, 0x940, 0xA20, 0x800, 0x0, 0x0}},
    {41751, 2, {0x0, 0x0, 0x0, 0x3C0, 0x420, 0x500, 0x480, 0x200, 0x180, 0x40, 0x220, 0x120, 0x420, 0x3C0, 0x0, 0x0}},
    {41752, 2, {0x0, 0x380, 0x440, 0x0, 0x1180, 0x960, 0x500, 0x1100, 0x940, 0x520, 0x110, 0x140, 0xD20, 0x310, 0x0, 0x0}},
    {41753, 2, {0x0, 0x0, 0x0, 0x0, 0x1180, 0x960, 0x500, 0x1100, 0x940, 0x520, 0x110, 0x140, 0xD20, 0x310, 0x0, 0x0}},
    {41754, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0xA80, 0xA80, 0x800, 0x7C0, 0x20, 0x2A0, 0x2A0, 0x820, 0x7C0, 0x0, 0x0}},
    {41755, 2, {0x0, 0x380, 0x440, 0x0, 0x8C0, 0x8A0, 0x90, 0x780, 0x880, 0x880, 0x780, 0x80, 0x880, 0x880, 0x0, 0x0}},
    {41756, 2, {0x0, 0x0, 0x0, 0x0, 0x8C0, 0x8A0, 0x90, 0x780, 0x880, 0x880, 0x780, 0x80, 0x880, 0x880, 0x0, 0x0}},
    {41757, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x1050, 0x1050, 0x1350, 0x2C0, 0x200, 0x1210, 0x1210, 0x1210, 0x200, 0x0, 0x0}},
    {41758, 2, {0x0, 0x0, 0x0, 0x820, 0x820, 0x7C0, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x7C0, 0x820, 0x820, 0x0, 0x0}},
    {41759, 2, {0x0, 0x380, 0x440, 0x0, 0x240, 0x240, 0x240, 0x200, 0x3E0, 0x200, 0x240, 0x240, 0x1A40, 0x600, 0x0, 0x0}},
    {41760, 2, {0x0, 0x0, 0x0, 0x0, 0x120, 0x120, 0x120, 0x100, 0x1F0, 0x100, 0x120, 0x120, 0xD20, 0x300, 0x0, 0x0}},
    {41761, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x810, 0x810, 0x3E10, 0x810, 0x810, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {41762, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x920, 0x920, 0x1D20, 0x900, 0x9F0, 0x100, 0x120, 0xD20, 0x320, 0x0, 0x0}},
    {41763, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x920, 0x920, 0x1D20, 0x900, 0x9F0, 0x100, 0x120, 0xD20, 0x320, 0x0, 0x0}},
    {41764, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x100, 0x7C0, 0x100, 0x0, 0x0}},
    {41765, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41766, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41767, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x7F0, 0x80, 0x80, 0x7F0, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {41768, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41769, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41770, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x1188, 0x1F68, 0x1118, 0x100, 0x110, 0x1110, 0x1F10, 0x1110, 0x100, 0x0, 0x0}},
    {41771, 2, {0x0, 0x380, 0x440, 0x0, 0x20, 0x620, 0x5A0, 0x460, 0x400, 0x440, 0x440, 0x5F0, 0x440, 0x440, 0x0, 0x0}},
    {41772, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x620, 0x5A0, 0x460, 0x400, 0x440, 0x440, 0x5F0, 0x440, 0x440, 0x0, 0x0}},
    {41773, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x310, 0x1ED0, 0x230, 0x200, 0x220, 0x220, 0x1EF8, 0x220, 0x220, 0x0, 0x0}},
    {41774, 2, {0x0, 0x380, 0x440, 0x0, 0x1808, 0x1608, 0x1188, 0x1168, 0x1518, 0x540, 0x140, 0x540, 0x510, 0xE0, 0x0, 0x0}},
    {41775, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1168, 0x1518, 0x540, 0x140, 0x540, 0x510, 0xE0, 0x0, 0x0}},
    {41776, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x2A0, 0x2A0, 0x80, 0x7F0, 0x0, 0x0}},
    {41777, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1390, 0xD60, 0x100, 0x540, 0x1550, 0x1550, 0x1110, 0xEE0, 0x0, 0x0}},
    {41778, 2, {0x0, 0x380, 0x440, 0x0, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x2108, 0x2108, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x0, 0x0}},
    {41779, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x2108, 0x2108, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x0, 0x0}},
    {41780, 2, {0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x100, 0x1FF0, 0x100, 0xFE0, 0x1010, 0x1010, 0xFE0, 0x0, 0x0}},
    {41781, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x1188, 0x1F68, 0x1118, 0x1100, 0x1100, 0x1F00, 0x1100, 0x100, 0x100, 0x0, 0x0}},
    {41782, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41783, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1FF0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41784, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1650, 0x11C0, 0x1040, 0x1240, 0x1540, 0x1540, 0x1540, 0x1240, 0x1040, 0x0, 0x0}},
    {41785, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0x100, 0x100, 0x1C0, 0x120, 0x120, 0xD20, 0x300, 0x0, 0x0}},
    {41786, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x100, 0x100, 0x1C0, 0x120, 0x120, 0xD20, 0x300, 0x0, 0x0}},
    {41787, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1190, 0x1150, 0x1110, 0x1190, 0x1550, 0x1350, 0x820, 0x7C0, 0x0, 0x0}},
    {41788, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x1FF8, 0x40, 0x440, 0x478, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41789, 2, {0x0, 0x380, 0x440, 0x0, 0x1C60, 0x1350, 0x10C0, 0x240, 0xFC0, 0x240, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41790, 2, {0x0, 0x0, 0x0, 0x0, 0x1C60, 0x1350, 0x10C0, 0x240, 0xFC0, 0x240, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {41791, 2, {0x0, 0x0, 0x0, 0x490, 0x3E0, 0x80, 0x490, 0x3E0, 0x1880, 0x1680, 0x13E0, 0x1410, 0x490, 0x490, 0x0, 0x0}},
    {41792, 2, {0x0, 0x0, 0x0, 0x0, 0x1120, 0x1510, 0x1510, 0x1120, 0x15C0, 0x1520, 0x1110, 0x1510, 0x1510, 0x1120, 0x0, 0x0}},
    {41793, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x7D8, 0x900, 0x9C0, 0x900, 0x37C0, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41794, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x7D8, 0x900, 0x9C0, 0x900, 0x37C0, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {41795, 2, {0x0, 0x0, 0x0, 0x0, 0xF00, 0x1088, 0x2248, 0x240, 0xFC8, 0x248, 0x240, 0x3848, 0x2648, 0x2180, 0x0, 0x0}},
    {41796, 2, {0x380, 0x440, 0x0, 0xEE0, 0x1110, 0x1110, 0x1110, 0x100, 0x100, 0xFE0, 0x0, 0x920, 0x920, 0x920, 0x0, 0x0}},
    {41797, 2, {0x0, 0x0, 0x0, 0xEE0, 0x1110, 0x1110, 0x1110, 0x100, 0x100, 0xFE0, 0x0, 0x920, 0x920, 0x920, 0x0, 0x0}},
    {41798, 2, {0x0, 0x0, 0x70, 0xE80, 0x11A0, 0x11A0, 0x1180, 0x1570, 0x1500, 0x1570, 0x1180, 0x11A0, 0x11A0, 0xE80, 0x70, 0x0}},
    {41799, 2, {0x380, 0x440, 0x100, 0x100, 0x440, 0x440, 0x1810, 0x1610, 0x1110, 0x10D0, 0x1030, 0x440, 0x540, 0x100, 0x0, 0x0}},
    {41800, 2, {0x0, 0x0, 0x100, 0x100, 0x440, 0x440, 0x1810, 0x1610, 0x1110, 0x10D0, 0x1030, 0x440, 0x540, 0x100, 0x0, 0x0}},
    {41801, 2, {0x0, 0x0, 0x0, 0x0, 0x388, 0x3A68, 0x618, 0x1200, 0x1200, 0x1200, 0x1200, 0x618, 0x3A68, 0x388, 0x0, 0x0}},
    {41802, 2, {0x0, 0x380, 0x440, 0x0, 0x580, 0x560, 0x2518, 0x1900, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41803, 2, {0x0, 0x0, 0x0, 0x0, 0x580, 0x560, 0x2518, 0x1900, 0x100, 0x100, 0x3FF8, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41804, 2, {0x0, 0x0, 0x0, 0x0, 0x7E0, 0x820, 0x820, 0x820, 0x7E0, 0x20, 0x120, 0x920, 0x820, 0x7C0, 0x0, 0x0}},
    {41805, 2, {0x0, 0x0, 0x0, 0x0, 0x580, 0x560, 0x2518, 0x1900, 0x100, 0x100, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41806, 2, {0x0, 0x380, 0x440, 0x0, 0x1860, 0x1650, 0x11C0, 0x1040, 0x1240, 0x1240, 0x1040, 0x1FC0, 0x1040, 0x1040, 0x0, 0x0}},
    {41807, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1650, 0x11C0, 0x1040, 0x1240, 0x1240, 0x1040, 0x1FC0, 0x1040, 0x1040, 0x0, 0x0}},
    {41808, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x540, 0x440, 0x280, 0x1E90, 0x1110, 0x12F0, 0x280, 0x440, 0x540, 0x100, 0x0}},
    {41809, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0x100, 0x540, 0x2548, 0x2108, 0x1110, 0xFE0, 0x0, 0x0}},
    {41810, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x100, 0x100, 0x100, 0x540, 0x540, 0x2548, 0x2108, 0x1110, 0xFE0, 0x0, 0x0}},
    {41811, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x540, 0x540, 0x2548, 0x2108, 0x1110, 0xFE0, 0x0, 0x0}},
    {41812, 2, {0x0, 0x0, 0x0, 0x0, 0x1550, 0x1550, 0x1550, 0x1110, 0x1550, 0x1550, 0x1550, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41813, 2, {0x0, 0x0, 0x0, 0x0, 0x638, 0x544, 0x4C4, 0x444, 0x444, 0x444, 0x444, 0x2454, 0x1454, 0xC38, 0x10, 0x0}},
    {41814, 2, {0x0, 0x380, 0x440, 0x0, 0x190, 0xF68, 0x1128, 0x1128, 0x910, 0x700, 0x900, 0x1100, 0x1100, 0xF00, 0x0, 0x0}},
    {41815, 2, {0x0, 0x0, 0x0, 0x0, 0x190, 0xF68, 0x1128, 0x1128, 0x910, 0x700, 0x900, 0x1100, 0x1100, 0xF00, 0x0, 0x0}},
    {41816, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x280, 0x280, 0x100, 0x0, 0x440, 0xAA0, 0xAA0, 0x440, 0x0, 0x0}},
    {41817, 2, {0x0, 0x380, 0x440, 0x0, 0x310, 0x2E8, 0x1A28, 0x628, 0x210, 0x1A00, 0x600, 0x200, 0x1A00, 0x600, 0x0, 0x0}},
    {41818, 2, {0x0, 0x0, 0x0, 0x0, 0x310, 0x2E8, 0x1A28, 0x628, 0x210, 0x1A00, 0x600, 0x200, 0x1A00, 0x600, 0x0, 0x0}},
    {41819, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x800, 0x7C0, 0x20, 0x820, 0x7C0, 0x0, 0xFE0, 0x0, 0xFE0, 0x0, 0x0}},
    {41820, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x110, 0x110, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41821, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x110, 0x110, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41822, 2, {0x0, 0x0, 0x0, 0xE10, 0x990, 0x870, 0x0, 0xE10, 0x990, 0x870, 0x0, 0xE10, 0x990, 0x870, 0x0, 0x0}},
    {41823, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x260, 0x1218, 0x1200, 0x13E0, 0x1200, 0x1200, 0x1218, 0x260, 0x380, 0x0, 0x0}},
    {41824, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x260, 0x1218, 0x1200, 0x13E0, 0x1200, 0x1200, 0x1218, 0x260, 0x380, 0x0, 0x0}},
    {41825, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0xFE0, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x0, 0x0}},
    {41826, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0x7C0, 0x920, 0x7C0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41827, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x7C0, 0x920, 0x7C0, 0x100, 0x7C0, 0x920, 0x920, 0x0, 0x0}},
    {41828, 2, {0x0, 0x0, 0x0, 0x100, 0x380, 0x540, 0x440, 0x540, 0x380, 0x100, 0x380, 0x540, 0x920, 0x920, 0x0, 0x0}},
    {41829, 2, {0x0, 0x0, 0x0, 0x0, 0x30C0, 0x2CA8, 0x2398, 0x2088, 0x2480, 0x2480, 0x2080, 0x3F80, 0x2080, 0x2080, 0x0, 0x0}},
    {41830, 2, {0x0, 0x380, 0x440, 0x0, 0xC30, 0xA28, 0x820, 0x2820, 0x18A0, 0xC60, 0xA20, 0x820, 0x28A0, 0x1860, 0x0, 0x0}},
    {41831, 2, {0x0, 0x0, 0x0, 0x0, 0xC30, 0xA28, 0x820, 0x2820, 0x18A0, 0xC60, 0xA20, 0x820, 0x28A0, 0x1860, 0x0, 0x0}},
    {41832, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2230, 0x1250, 0x890, 0x510, 0x21C, 0x510, 0x890, 0x1250, 0x2230, 0x0, 0x0}},
    {41833, 2, {0x0, 0x0, 0x0, 0x0, 0x1818, 0x1514, 0x1112, 0x1010, 0x1FF0, 0x1010, 0x1110, 0x9110, 0x5050, 0x3030, 0x0, 0x0}},
    {41834, 2, {0x380, 0x440, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x0, 0x0}},
    {41835, 2, {0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x0, 0x0}},
    {41836, 2, {0x0, 0x0, 0x0, 0x0, 0xC30, 0xB2C, 0x820, 0x820, 0xFE0, 0x820, 0x820, 0x820, 0x69A0, 0x1860, 0x0, 0x0}},
    {41837, 2, {0x0, 0x0, 0x0, 0x0, 0x1FF0, 0x1110, 0x920, 0x540, 0x1390, 0x1110, 0x1390, 0x540, 0x920, 0x1110, 0x0, 0x0}},
    {41838, 2, {0x0, 0x380, 0x440, 0x0, 0x1FF0, 0x1110, 0x920, 0x540, 0x380, 0x1FF0, 0x280, 0x440, 0x920, 0x1110, 0x0, 0x0}},
    {41839, 2, {0x0, 0x0, 0x0, 0x0, 0x1FF0, 0x1110, 0x920, 0x540, 0x380, 0x1FF0, 0x280, 0x440, 0x920, 0x1110, 0x0, 0x0}},
    {41840, 2, {0x0, 0x0, 0x0, 0x0, 0x1FF0, 0x1110, 0x920, 0x540, 0x3B88, 0x27C8, 0x22B8, 0x440, 0x920, 0x1110, 0x0, 0x0}},
    {41841, 2, {0x0, 0x380, 0x440, 0x0, 0xE00, 0x1128, 0x1128, 0x15F8, 0x1528, 0x15F8, 0x1528, 0x1128, 0x1100, 0xE00, 0x0, 0x0}},
    {41842, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1128, 0x1128, 0x15F8, 0x1528, 0x15F8, 0x1528, 0x1128, 0x1100, 0xE00, 0x0, 0x0}},
    {41843, 2, {0x0, 0x0, 0x0, 0x0, 0x720, 0x8A0, 0x880, 0xAF0, 0xA80, 0xA80, 0xAF0, 0x880, 0x8A0, 0x720, 0x0, 0x0}},
    {41844, 2, {0x0, 0x380, 0x440, 0x0, 0xC04, 0xB34, 0x82C, 0x820, 0x820, 0x3E20, 0x820, 0x82C, 0x6834, 0x1804, 0x0, 0x0}},
    {41845, 2, {0x0, 0x0, 0x0, 0x0, 0xC04, 0xB34, 0x82C, 0x820, 0x820, 0x3E20, 0x820, 0x82C, 0x6834, 0x1804, 0x0, 0x0}},
    {41846, 2, {0x0, 0x0, 0x1C10, 0x1C8, 0x5A8, 0x598, 0x180, 0xE80, 0x80, 0xFF8, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {41847, 2, {0x0, 0x380, 0x440, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x1C0, 0x220, 0x220, 0x1C0, 0x0, 0x0}},
    {41848, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x1C0, 0x220, 0x220, 0x1C0, 0x0, 0x0}},
    {41849, 2, {0x0, 0x0, 0x0, 0x0, 0x550, 0x550, 0x110, 0x7D0, 0x110, 0x550, 0x450, 0xC10, 0x390, 0x70, 0x0, 0x0}},
    {41850, 2, {0x380, 0x440, 0x0, 0x1C08, 0x388, 0xB68, 0xB18, 0x300, 0x1D00, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x0, 0x0}},
    {41851, 2, {0x0, 0x0, 0x0, 0x1C08, 0x388, 0xB68, 0xB18, 0x300, 0x1D00, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x0, 0x0}},
    {41852, 2, {0x0, 0x380, 0x440, 0x0, 0x330, 0x12E8, 0x1224, 0x3A20, 0x1220, 0x1220, 0x220, 0x220, 0x1A20, 0x620, 0x0, 0x0}},
    {41853, 2, {0x0, 0x0, 0x0, 0x0, 0x330, 0x12E8, 0x1224, 0x3A20, 0x1220, 0x1220, 0x220, 0x220, 0x1A20, 0x620, 0x0, 0x0}},
    {41854, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x120, 0x920, 0x1500, 0x15F0, 0x1500, 0x920, 0x120, 0xD00, 0x300, 0x0, 0x0}},
    {41855, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x120, 0x920, 0x1500, 0x15F0, 0x1500, 0x920, 0x120, 0xD00, 0x300, 0x0, 0x0}},
    {41856, 2, {0x0, 0x0, 0x0, 0x0, 0x148, 0x1148, 0x1130, 0x1100, 0x11F8, 0x100, 0x130, 0x3148, 0xD48, 0x300, 0x0, 0x0}},
    {41857, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x7C0, 0x820, 0xFE0, 0x820, 0x820, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {41858, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x7C0, 0x820, 0xFE0, 0x820, 0x820, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {41859, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x450, 0x450, 0x110, 0x110, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41860, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0x100, 0x3FF8, 0x100, 0x1100, 0x7D00, 0x1100, 0x0, 0x0}},
    {41861, 2, {0x0, 0x380, 0x440, 0x0, 0xFC0, 0x20, 0x20, 0x20, 0xFE0, 0x20, 0x20, 0x20, 0x20, 0xFC0, 0x0, 0x0}},
    {41862, 2, {0x0, 0x0, 0x0, 0x0, 0xFC0, 0x20, 0x20, 0x20, 0xFE0, 0x20, 0x20, 0x20, 0x20, 0xFC0, 0x0, 0x0}},
    {41863, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x3E0, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41864, 2, {0x0, 0x0, 0x0, 0x0, 0x1810, 0x1410, 0x1210, 0x1010, 0x1110, 0x1110, 0x1010, 0x1090, 0x1050, 0x1030, 0x0, 0x0}},
    {41865, 2, {0x0, 0x380, 0x440, 0x0, 0x820, 0x1450, 0x1450, 0x1450, 0x820, 0x540, 0x280, 0x280, 0x280, 0x100, 0x0, 0x0}},
    {41866, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x1450, 0x1450, 0x1450, 0x820, 0x540, 0x280, 0x280, 0x280, 0x100, 0x0, 0x0}},
    {41867, 2, {0x0, 0x0, 0x0, 0x0, 0x36D8, 0x820, 0x1450, 0x1450, 0x1450, 0x3FF8, 0x1450, 0x1450, 0x820, 0x36D8, 0x0, 0x0}},
    {41868, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x40, 0x620, 0x820, 0x920, 0x920, 0x820, 0x620, 0x40, 0x180, 0x0, 0x0}},
    {41869, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x40, 0x620, 0x820, 0x920, 0x920, 0x820, 0x620, 0x40, 0x180, 0x0, 0x0}},
    {41870, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0xA8, 0xA8, 0xA8, 0xF8, 0xA8, 0x18A8, 0x16A8, 0x1180, 0x1000, 0x0, 0x0}},
    {41871, 2, {0x0, 0x380, 0x440, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x50, 0xC50, 0xB50, 0x8D0, 0x800, 0x0, 0x0}},
    {41872, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x50, 0xC50, 0xB50, 0x8D0, 0x800, 0x0, 0x0}},
    {41873, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x7A0, 0x890, 0x880, 0x4A0, 0x3A0, 0x4A0, 0x880, 0x880, 0x780, 0x0, 0x0}},
    {41874, 2, {0x0, 0x380, 0x440, 0x0, 0x808, 0x988, 0x1D68, 0x918, 0x900, 0x100, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41875, 2, {0x0, 0x0, 0x0, 0x0, 0x808, 0x988, 0x1D68, 0x918, 0x900, 0x100, 0x7C0, 0x920, 0x1110, 0x1110, 0x0, 0x0}},
    {41876, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1110, 0x1110, 0x1010, 0x1FF0, 0x1010, 0x1110, 0x1110, 0x820, 0x7C0, 0x0, 0x0}},
    {41877, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x1450, 0x1450, 0x440, 0xFE0, 0x440, 0x1450, 0x1450, 0x440, 0x380, 0x0, 0x0}},
    {41878, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x1450, 0x1450, 0x440, 0xFE0, 0x440, 0x1450, 0x1450, 0x440, 0x380, 0x0, 0x0}},
    {41879, 2, {0x0, 0x0, 0x0, 0x0, 0x240, 0x250, 0x260, 0x240, 0x240, 0x240, 0x240, 0x640, 0xA40, 0x240, 0x0, 0x0}},
    {41880, 2, {0x0, 0x380, 0x440, 0x0, 0x3C0, 0x420, 0x910, 0x110, 0x7D0, 0x110, 0x110, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41881, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x910, 0x110, 0x7D0, 0x110, 0x110, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41882, 2, {0x0, 0x0, 0x220, 0x220, 0x3E0, 0x630, 0xA30, 0x1D0, 0x10, 0x10, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {41883, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x168, 0x118, 0x100, 0x1FF0, 0x100, 0x920, 0x920, 0x100, 0x0, 0x0}},
    {41884, 2, {0x0, 0x380, 0x440, 0x0, 0x404, 0x434, 0x42C, 0x420, 0x420, 0x7E0, 0x420, 0x3420, 0x2C20, 0x2020, 0x0, 0x0}},
    {41885, 2, {0x0, 0x0, 0x0, 0x0, 0x404, 0x434, 0x42C, 0x420, 0x420, 0x7E0, 0x420, 0x3420, 0x2C20, 0x2020, 0x0, 0x0}},
    {41886, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x90, 0x80, 0x7B0, 0x8C0, 0x10F0, 0x10C0, 0x8B0, 0x780, 0x0, 0x0}},
    {41887, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x880, 0xBA0, 0xAE0, 0x880, 0x880, 0xBA0, 0xAE0, 0x880, 0x700, 0x0, 0x0}},
    {41888, 2, {0x0, 0x380, 0x440, 0x1000, 0x28C0, 0x2FB0, 0x288C, 0x1080, 0x80, 0x1080, 0x2880, 0x2F80, 0x2880, 0x1080, 0x0, 0x0}},
    {41889, 2, {0x0, 0x0, 0x0, 0x1000, 0x28C0, 0x2FB0, 0x288C, 0x1080, 0x80, 0x1080, 0x2880, 0x2F80, 0x2880, 0x1080, 0x0, 0x0}},
    {41890, 2, {0x0, 0x0, 0x0, 0x820, 0x820, 0x440, 0x1390, 0x1110, 0x7D7C, 0x1110, 0x1110, 0xD00, 0xB00, 0x800, 0x0, 0x0}},
    {41891, 2, {0x0, 0x380, 0x440, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41892, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41893, 2, {0x0, 0x0, 0x0, 0x0, 0x930, 0x550, 0x390, 0x390, 0x550, 0x930, 0x100, 0xD00, 0xB00, 0x800, 0x0, 0x0}},
    {41894, 2, {0x0, 0x380, 0x440, 0x0, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0xA10, 0xC10, 0x1810, 0x2810, 0x810, 0x0, 0x0}},
    {41895, 2, {0x0, 0x0, 0x0, 0x0, 0x810, 0x810, 0x810, 0xFF0, 0x810, 0xA10, 0xC10, 0x1810, 0x2810, 0x810, 0x0, 0x0}},
    {41896, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x880, 0x880, 0xF80, 0x880, 0x880, 0x80, 0x80, 0x80, 0xFF8, 0x0, 0x0}},
    {41897, 2, {0x0, 0x380, 0x440, 0x0, 0x300, 0x2C0, 0x200, 0x200, 0x3E0, 0x210, 0x248, 0x2A8, 0x1AA0, 0x640, 0x0, 0x0}},
    {41898, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x2C0, 0x200, 0x200, 0x3E0, 0x210, 0x248, 0x2A8, 0x1AA0, 0x640, 0x0, 0x0}},
    {41899, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1018, 0x828, 0x448, 0x288, 0x108, 0x288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {41900, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x2918, 0x2900, 0x2900, 0x3F00, 0x2900, 0x2900, 0x2900, 0x100, 0x0, 0x0}},
    {41901, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x2918, 0x2900, 0x2900, 0x3F00, 0x2900, 0x2900, 0x2900, 0x100, 0x0, 0x0}},
    {41902, 2, {0x0, 0x0, 0x0, 0x0, 0x930, 0x928, 0xFE4, 0x920, 0x920, 0x920, 0xFE0, 0x4920, 0x2920, 0x1920, 0x0, 0x0}},
    {41903, 2, {0x0, 0x380, 0x440, 0x0, 0xFE0, 0x100, 0xFE0, 0x100, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x0, 0x0}},
    {41904, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x100, 0xFE0, 0x100, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x0, 0x0}},
    {41905, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x10E8, 0x1098, 0x80, 0x2A0, 0x1C0, 0x1C0, 0x2A0, 0x0, 0x0}},
    {41906, 2, {0x0, 0x0, 0x0, 0x0, 0x9C0, 0x1130, 0x1100, 0x920, 0x720, 0x920, 0x1120, 0x1100, 0x930, 0x1C0, 0x0, 0x0}},
    {41907, 2, {0x380, 0x440, 0x0, 0x540, 0x540, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {41908, 2, {0x0, 0x0, 0x0, 0x540, 0x540, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {41909, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x190C, 0x2514, 0x2224, 0x1D44, 0x84, 0x1D44, 0x2224, 0x2514, 0x190C, 0x0, 0x0}},
    {41910, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x2A0, 0x2A0, 0x2A0, 0x1880, 0x1680, 0x11C0, 0x1220, 0x490, 0x490, 0x0, 0x0}},
    {41911, 2, {0x0, 0x380, 0x440, 0x1000, 0x1180, 0x1F60, 0x1118, 0x1140, 0x140, 0x1950, 0x1710, 0x11D0, 0x1130, 0x100, 0x0, 0x0}},
    {41912, 2, {0x0, 0x0, 0x0, 0x1000, 0x1180, 0x1F60, 0x1118, 0x1140, 0x140, 0x1950, 0x1710, 0x11D0, 0x1130, 0x100, 0x0, 0x0}},
    {41913, 2, {0x0, 0x0, 0x0, 0x1C00, 0x380, 0xB60, 0xB18, 0x340, 0x1D40, 0x140, 0x1910, 0x1710, 0x11D0, 0x1130, 0x0, 0x0}},
    {41914, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x920, 0x540, 0x380, 0x920, 0x540, 0x380, 0x920, 0x540, 0x380, 0x0, 0x0}},
    {41915, 2, {0x0, 0x380, 0x440, 0x0, 0x128, 0x128, 0x128, 0x128, 0x128, 0x128, 0x130, 0x31E0, 0xD20, 0x320, 0x0, 0x0}},
    {41916, 2, {0x0, 0x0, 0x0, 0x0, 0x128, 0x128, 0x128, 0x128, 0x128, 0x128, 0x130, 0x31E0, 0xD20, 0x320, 0x0, 0x0}},
    {41917, 2, {0x0, 0x0, 0x0, 0x0, 0x7E0, 0x810, 0x810, 0x810, 0x7E0, 0x80, 0x680, 0x900, 0x680, 0x80, 0x0, 0x0}},
    {41918, 2, {0x0, 0x380, 0x440, 0x0, 0x790, 0x860, 0x860, 0x7A0, 0x20, 0x20, 0x7A0, 0x860, 0x860, 0x790, 0x0, 0x0}},
    {41919, 2, {0x0, 0x0, 0x0, 0x0, 0x790, 0x860, 0x860, 0x7A0, 0x20, 0x20, 0x7A0, 0x860, 0x860, 0x790, 0x0, 0x0}},
    {41920, 2, {0x0, 0x0, 0x0, 0x0, 0x220, 0x220, 0x1A28, 0x1628, 0x13A8, 0x1268, 0x1238, 0x220, 0x220, 0x220, 0x0, 0x0}},
    {41921, 2, {0x0, 0x380, 0x440, 0x0, 0x3E0, 0x400, 0x400, 0x200, 0x1E0, 0x200, 0x400, 0x400, 0x400, 0x3E0, 0x0, 0x0}},
    {41922, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x400, 0x400, 0x200, 0x1E0, 0x200, 0x400, 0x400, 0x400, 0x3E0, 0x0, 0x0}},
    {41923, 2, {0x0, 0x0, 0x0, 0x490, 0x490, 0x2A0, 0x1C0, 0x80, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x1000, 0x0, 0x0}},
    {41924, 2, {0x0, 0x380, 0x440, 0x1000, 0x11E0, 0x1200, 0x400, 0x1400, 0x17E0, 0x1400, 0x400, 0x1400, 0x1200, 0x11E0, 0x0, 0x0}},
    {41925, 2, {0x0, 0x0, 0x0, 0x800, 0x8F0, 0x900, 0x200, 0xA00, 0xBF0, 0xA00, 0x200, 0xA00, 0x900, 0x8F0, 0x0, 0x0}},
    {41926, 2, {0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x7C0, 0x920, 0x1110, 0x17D0, 0x1110, 0x1110, 0x100, 0x0, 0x0}},
    {41927, 2, {0x0, 0x380, 0x440, 0x0, 0xE00, 0x1148, 0x1148, 0x1130, 0x11F8, 0x1130, 0x1148, 0x1148, 0x1100, 0xE00, 0x0, 0x0}},
    {41928, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1148, 0x1148, 0x1130, 0x11F8, 0x1130, 0x1148, 0x1148, 0x1100, 0xE00, 0x0, 0x0}},
    {41929, 2, {0x0, 0x0, 0x0, 0x0, 0x760, 0x880, 0x980, 0xA80, 0xA80, 0xA80, 0xA80, 0x980, 0x880, 0x760, 0x0, 0x0}},
    {41930, 2, {0x0, 0x380, 0x440, 0x0, 0xC20, 0xA20, 0x820, 0x920, 0xAA0, 0xAA0, 0x920, 0x820, 0x8A0, 0x860, 0x0, 0x0}},
    {41931, 2, {0x0, 0x0, 0x0, 0x0, 0xC20, 0xA20, 0x820, 0x920, 0xAA0, 0xAA0, 0x920, 0x820, 0x8A0, 0x860, 0x0, 0x0}},
    {41932, 2, {0x0, 0x0, 0x0, 0x0, 0x1000, 0x1210, 0x1120, 0x10C0, 0x1FF0, 0x10C0, 0x1120, 0x1210, 0x1000, 0x1000, 0x0, 0x0}},
    {41933, 2, {0x0, 0x0, 0x0, 0x440, 0x280, 0x100, 0x2288, 0x1450, 0x820, 0x1450, 0x2288, 0x100, 0x280, 0x440, 0x0, 0x0}},
    {41934, 2, {0x0, 0x380, 0x440, 0x0, 0x700, 0x880, 0x9C0, 0xAA0, 0xAA0, 0xAA0, 0xAA0, 0x9C0, 0x880, 0x700, 0x0, 0x0}},
    {41935, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x880, 0x9C0, 0xAA0, 0xAA0, 0xAA0, 0xAA0, 0x9C0, 0x880, 0x700, 0x0, 0x0}},
    {41936, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x1290, 0x2AA8, 0x2928, 0x2AA8, 0x1290, 0x440, 0x440, 0x0, 0x0}},
    {41937, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x440, 0x440, 0x380, 0x930, 0x550, 0x390, 0x390, 0x550, 0x930, 0x0, 0x0}},
    {41938, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x380, 0x930, 0x550, 0x390, 0x390, 0x550, 0x930, 0x0, 0x0}},
    {41939, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x210, 0x210, 0x12D0, 0xB50, 0x650, 0xB50, 0x12D0, 0x210, 0x1E0, 0x0, 0x0}},
    {41940, 2, {0x0, 0x0, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x540, 0x540, 0x100, 0xFE0, 0x100, 0xFE0, 0x0, 0x0}},
    {41941, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x3110, 0x2D10, 0x2310, 0x21D0, 0x2130, 0x100, 0x540, 0x540, 0x100, 0x0, 0x0}},
    {41942, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0x2A0, 0x2A0, 0x80, 0x0, 0x0}},
    {41943, 2, {0x0, 0x0, 0x0, 0x0, 0x1E00, 0x100, 0x930, 0x550, 0x390, 0x390, 0x550, 0x930, 0x100, 0x1E00, 0x0, 0x0}},
    {41944, 2, {0x0, 0x380, 0x440, 0x0, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x100, 0x540, 0x540, 0x100, 0x0, 0x0}},
    {41945, 2, {0x0, 0x0, 0x0, 0x0, 0x920, 0x540, 0x380, 0x380, 0x540, 0x920, 0x100, 0x540, 0x540, 0x100, 0x0, 0x0}},
    {41946, 2, {0x0, 0x0, 0x0, 0x0, 0x248, 0x248, 0x150, 0xE0, 0x340, 0x2C0, 0x700, 0xA80, 0x1240, 0x1240, 0x0, 0x0}},
    {41947, 2, {0x0, 0x380, 0x440, 0x0, 0x620, 0x5A0, 0x460, 0x400, 0x520, 0x7F0, 0x520, 0x520, 0x7F0, 0x520, 0x0, 0x0}},
    {41948, 2, {0x0, 0x0, 0x0, 0x0, 0x620, 0x5A0, 0x460, 0x400, 0x520, 0x7F0, 0x520, 0x520, 0x7F0, 0x520, 0x0, 0x0}},
    {41949, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0x28, 0x2624, 0x1520, 0xCA0, 0x3FA0, 0x620, 0x520, 0x4A0, 0x60, 0x0, 0x0}},
    {41950, 2, {0x0, 0x380, 0x440, 0x0, 0x78, 0x1880, 0x480, 0x278, 0x280, 0x280, 0x278, 0x480, 0x1880, 0x78, 0x0, 0x0}},
    {41951, 2, {0x0, 0x0, 0x0, 0x0, 0x78, 0x1880, 0x480, 0x278, 0x280, 0x280, 0x278, 0x480, 0x1880, 0x78, 0x0, 0x0}},
    {41952, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x100, 0xD20, 0xB20, 0x9A0, 0x960, 0x100, 0x820, 0x7C0, 0x0, 0x0}},
    {41953, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x100, 0x3110, 0x2C10, 0x2310, 0x20D0, 0x2030, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41954, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x1888, 0x1608, 0x1188, 0x1068, 0x1018, 0x80, 0x80, 0x80, 0x0, 0x0}},
    {41955, 2, {0x0, 0x0, 0x1FF8, 0x20, 0xA0, 0xA0, 0x620, 0x5A0, 0x460, 0x400, 0x480, 0x480, 0x400, 0x1FF8, 0x0, 0x0}},
    {41956, 2, {0x0, 0x0, 0x0, 0x730, 0x4E8, 0x424, 0x35A0, 0xC60, 0x420, 0x35A0, 0xC60, 0x420, 0x35A0, 0xC60, 0x0, 0x0}},
    {41957, 2, {0x0, 0x380, 0x440, 0x0, 0x804, 0x934, 0x92C, 0x820, 0xFE0, 0x820, 0x920, 0x6920, 0x5820, 0x4020, 0x0, 0x0}},
    {41958, 2, {0x0, 0x0, 0x0, 0x0, 0x804, 0x934, 0x92C, 0x820, 0xFE0, 0x820, 0x920, 0x6920, 0x5820, 0x4020, 0x0, 0x0}},
    {41959, 2, {0x0, 0x0, 0x0, 0x0, 0xEF0, 0x100, 0x280, 0xFE0, 0x280, 0x280, 0xFE0, 0x280, 0x100, 0xEF0, 0x0, 0x0}},
    {41960, 2, {0x0, 0x0, 0x1010, 0x17D0, 0x820, 0x17D0, 0x1010, 0x1110, 0x1110, 0x1010, 0x17D0, 0x820, 0x17D0, 0x1010, 0x0, 0x0}},
    {41961, 2, {0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41962, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x1390, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {41963, 2, {0x0, 0x0, 0x0, 0x0, 0xF80, 0x1058, 0x1360, 0x12C0, 0x1240, 0x1240, 0x1A40, 0x1660, 0x1058, 0xF80, 0x0, 0x0}},
    {41964, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x7C0, 0x100, 0x100, 0x0, 0x0}},
    {41965, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x900, 0x900, 0x600, 0x0, 0x0}},
    {41966, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x100, 0x900, 0x900, 0x600, 0x0, 0x0}},
    {41967, 2, {0x0, 0x0, 0x0, 0x440, 0x440, 0x380, 0x1110, 0x1110, 0xEE0, 0x0, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {41968, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1190, 0x1150, 0x1110, 0x17D0, 0x1110, 0x1510, 0x1310, 0x820, 0x7C0, 0x0, 0x0}},
    {41969, 2, {0x0, 0x380, 0x440, 0x0, 0x1E0, 0x200, 0x440, 0x4A0, 0x4A0, 0x4A0, 0x4A0, 0x440, 0x200, 0x1E0, 0x0, 0x0}},
    {41970, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x200, 0x440, 0x4A0, 0x4A0, 0x4A0, 0x4A0, 0x440, 0x200, 0x1E0, 0x0, 0x0}},
    {41971, 2, {0x0, 0x0, 0x0, 0x0, 0x3D0, 0x430, 0x910, 0xA80, 0xA80, 0xA80, 0xA80, 0x910, 0x420, 0x3C0, 0x0, 0x0}},
    {41972, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1190, 0x1150, 0x1110, 0x1110, 0x1110, 0x1510, 0x1310, 0x820, 0x7C0, 0x0, 0x0}},
    {41973, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1190, 0x1150, 0x1110, 0x1110, 0x1110, 0x1510, 0x1310, 0x820, 0x7C0, 0x0, 0x0}},
    {41974, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x918, 0x1100, 0x1100, 0x17C0, 0x1100, 0x1100, 0x900, 0x100, 0x0, 0x0}},
    {41975, 2, {0x0, 0x380, 0x440, 0x0, 0x380, 0x260, 0x200, 0xA00, 0xA00, 0xA00, 0xA00, 0xA00, 0x260, 0x380, 0x0, 0x0}},
    {41976, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x260, 0x200, 0xA00, 0xA00, 0xA00, 0xA00, 0xA00, 0x260, 0x380, 0x0, 0x0}},
    {41977, 2, {0x0, 0x0, 0x0, 0xE00, 0x9C0, 0x800, 0x880, 0x880, 0xBE0, 0x880, 0x880, 0x800, 0x9C0, 0xE00, 0x0, 0x0}},
    {41978, 2, {0x0, 0x380, 0x440, 0x0, 0x8, 0x188, 0x968, 0x1118, 0x1100, 0x1700, 0x1100, 0x1100, 0x900, 0x100, 0x0, 0x0}},
    {41979, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0x188, 0x968, 0x1118, 0x1100, 0x1700, 0x1100, 0x1100, 0x900, 0x100, 0x0, 0x0}},
    {41980, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0x28, 0x24, 0x1020, 0x23E0, 0x2420, 0x1CA0, 0x24A0, 0x2420, 0x13E0, 0x0, 0x0}},
    {41981, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x280, 0xAA0, 0x920, 0xAA0, 0x280, 0x440, 0x440, 0x0, 0x0}},
    {41982, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x280, 0xAA0, 0x920, 0xAA0, 0x280, 0x440, 0x440, 0x0, 0x0}},
    {41983, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1110, 0x1110, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41984, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3F80, 0x20A4, 0x1124, 0xA18, 0x7F8, 0xA18, 0x1124, 0x20A4, 0x3F80, 0x0, 0x0}},
    {41985, 2, {0x0, 0x380, 0x440, 0x0, 0x400, 0x488, 0x488, 0x450, 0x420, 0x7F8, 0x420, 0x2450, 0x1488, 0xC88, 0x0, 0x0}},
    {41986, 2, {0x0, 0x0, 0x0, 0x0, 0x400, 0x488, 0x488, 0x450, 0x420, 0x7F8, 0x420, 0x2450, 0x1488, 0xC88, 0x0, 0x0}},
    {41987, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1100, 0x1148, 0x1148, 0xA30, 0x7F8, 0xA30, 0x1148, 0x1148, 0x1100, 0x0, 0x0}},
    {41988, 2, {0x0, 0x0, 0x0, 0x0, 0x3F0, 0x410, 0x410, 0x3F0, 0x10, 0x3F0, 0x10, 0xBF0, 0x810, 0x7E0, 0x0, 0x0}},
    {41989, 2, {0x0, 0x380, 0x440, 0x0, 0x1010, 0x1010, 0x1010, 0x11F0, 0x1010, 0x1F10, 0x1010, 0x1010, 0x11F0, 0x1010, 0x0, 0x0}},
    {41990, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x11F0, 0x1010, 0x1F10, 0x1010, 0x1010, 0x11F0, 0x1010, 0x0, 0x0}},
    {41991, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x11F0, 0x1010, 0x1F10, 0x1010, 0x11F0, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {41992, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x1100, 0x1100, 0x1110, 0x110, 0x110, 0x1110, 0x1100, 0x1100, 0x100, 0x0, 0x0}},
    {41993, 2, {0x0, 0x380, 0x440, 0x0, 0xE00, 0x1900, 0x1500, 0x1120, 0x1150, 0x17D0, 0x1150, 0x1520, 0x1900, 0xE00, 0x0, 0x0}},
    {41994, 2, {0x0, 0x0, 0x0, 0x0, 0xE00, 0x1900, 0x1500, 0x1120, 0x1150, 0x17D0, 0x1150, 0x1520, 0x1900, 0xE00, 0x0, 0x0}},
    {41995, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x920, 0x1110, 0x1010, 0x1010, 0x1010, 0x1290, 0x1450, 0x820, 0x7C0, 0x0, 0x0}},
    {41996, 2, {0x0, 0x0, 0x0, 0x0, 0x700, 0x880, 0x8F0, 0x880, 0x8F0, 0x880, 0x8F0, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {41997, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41998, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1110, 0x920, 0x7C0, 0x0, 0x0}},
    {41999, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x810, 0x1010, 0x1090, 0xE90, 0x1090, 0x1010, 0xB10, 0xD0, 0x30, 0x0, 0x0}},
    {42000, 2, {0x0, 0x380, 0x440, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1290, 0xAA0, 0x7C0, 0x0, 0x0}},
    {42001, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1290, 0xAA0, 0x7C0, 0x0, 0x0}},
    {42002, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x1118, 0x928, 0x548, 0x388, 0x388, 0x548, 0x928, 0x1118, 0x100, 0x0, 0x0}},
    {42003, 2, {0x0, 0x380, 0x440, 0x0, 0x820, 0x820, 0x440, 0x380, 0x100, 0xD00, 0xB00, 0x800, 0x800, 0x800, 0x0, 0x0}},
    {42004, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x440, 0x380, 0x100, 0xD00, 0xB00, 0x800, 0x800, 0x800, 0x0, 0x0}},
    {42005, 2, {0x0, 0x0, 0x0, 0x920, 0xAA0, 0x920, 0x440, 0x380, 0x100, 0xD00, 0xB00, 0x800, 0x800, 0x800, 0x0, 0x0}},
    {42006, 2, {0x380, 0x440, 0x0, 0x1830, 0x1450, 0x1390, 0x1390, 0x1550, 0x1930, 0x100, 0x920, 0x540, 0x380, 0x540, 0x920, 0x0}},
    {42007, 2, {0x0, 0x0, 0x1830, 0x1450, 0x1390, 0x1390, 0x1550, 0x1930, 0x100, 0x920, 0x540, 0x380, 0x540, 0x920, 0x0, 0x0}},
    {42008, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x440, 0x380, 0x920, 0x920, 0x380, 0x440, 0x820, 0x820, 0x0, 0x0}},
    {42009, 2, {0x0, 0x380, 0x440, 0x0, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x0, 0x0}},
    {42010, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x0, 0x0}},
    {42011, 2, {0x0, 0x0, 0x0, 0x1800, 0x4C0, 0x1FA0, 0x490, 0x1880, 0x80, 0xF80, 0x1080, 0x1080, 0x1080, 0xF80, 0x0, 0x0}},
    {42012, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x440, 0x380, 0x100, 0x540, 0x540, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {42013, 2, {0x0, 0x380, 0x440, 0x0, 0x1830, 0x2448, 0x440, 0x440, 0x380, 0xFE0, 0x100, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {42014, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x2448, 0x440, 0x440, 0x380, 0xFE0, 0x100, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {42015, 2, {0x0, 0x0, 0x0, 0x0, 0x920, 0x920, 0x440, 0x380, 0x100, 0x920, 0x920, 0x920, 0x100, 0x100, 0x0, 0x0}},
    {42016, 2, {0x380, 0x440, 0x0, 0x1830, 0x440, 0x1290, 0x1290, 0x1290, 0x440, 0x1830, 0x380, 0x440, 0x920, 0x920, 0x0, 0x0}},
    {42017, 2, {0x0, 0x0, 0x0, 0x1830, 0x440, 0x1290, 0x1290, 0x1290, 0x440, 0x1830, 0x380, 0x440, 0x920, 0x920, 0x0, 0x0}},
    {42018, 2, {0x0, 0x0, 0x0, 0x0, 0x208, 0x208, 0x110, 0x8E0, 0x1400, 0x1400, 0x8E0, 0x110, 0x208, 0x208, 0x0, 0x0}},
    {42019, 2, {0x0, 0x0, 0x0, 0x0, 0xC04, 0xAC4, 0x9B4, 0x1C8C, 0x2A80, 0x2A80, 0x880, 0x1C80, 0x2A80, 0x2A80, 0x0, 0x0}},
    {42020, 2, {0x0, 0x380, 0x440, 0x0, 0x18C0, 0x16B0, 0x118C, 0x1080, 0x7C80, 0x1080, 0x7C80, 0x1080, 0x1080, 0x1080, 0x0, 0x0}},
    {42021, 2, {0x0, 0x0, 0x0, 0x0, 0x18C0, 0x16B0, 0x118C, 0x1080, 0x7C80, 0x1080, 0x7C80, 0x1080, 0x1080, 0x1080, 0x0, 0x0}},
    {42022, 2, {0x0, 0x0, 0x0, 0x1830, 0x440, 0x280, 0x280, 0x280, 0x440, 0x1830, 0x380, 0x440, 0x820, 0x820, 0x0, 0x0}},
    {42023, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x100, 0x700, 0x800, 0x820, 0x7C0, 0x0, 0x0}},
    {42024, 2, {0x380, 0x440, 0x0, 0x10, 0x7F0, 0x880, 0xFF0, 0x880, 0x7F0, 0x100, 0x1C0, 0x2A0, 0x490, 0x490, 0x0, 0x0}},
    {42025, 2, {0x0, 0x0, 0x0, 0x10, 0x7F0, 0x880, 0xFF0, 0x880, 0x7F0, 0x80, 0x1C0, 0x2A0, 0x490, 0x490, 0x0, 0x0}},
    {42026, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x920, 0x920, 0x440, 0x920, 0x920, 0x820, 0x7C0, 0x0, 0x0}},
    {42027, 2, {0x0, 0x380, 0x440, 0x0, 0x80, 0x7F0, 0x880, 0x880, 0x7F0, 0x80, 0x1C0, 0x2A0, 0x490, 0x490, 0x0, 0x0}},
    {42028, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x7F0, 0x880, 0x880, 0x7F0, 0x80, 0x1C0, 0x2A0, 0x490, 0x490, 0x0, 0x0}},
    {42029, 2, {0x0, 0x0, 0x0, 0x0, 0x3808, 0x2608, 0x21C8, 0x2538, 0x500, 0x140, 0x3948, 0x2708, 0x21C8, 0x2038, 0x0, 0x0}},
    {42030, 2, {0x0, 0x380, 0x440, 0x0, 0xEE0, 0x440, 0x440, 0x440, 0x380, 0x100, 0x7C0, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {42031, 2, {0x0, 0x0, 0x0, 0x0, 0xEE0, 0x440, 0x440, 0x440, 0x380, 0x100, 0x7C0, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {42032, 2, {0x0, 0x0, 0xAC0, 0xAA0, 0xF90, 0xA80, 0xA80, 0x80, 0xA80, 0xA80, 0xF80, 0xA80, 0xA80, 0x80, 0x0, 0x0}},
    {42033, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x1080, 0x1080, 0x1080, 0xFF0, 0x80, 0x80, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {42034, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x100, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {42035, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x920, 0x7C0, 0x100, 0x100, 0x100, 0x1100, 0x1100, 0xE00, 0x0, 0x0}},
    {42036, 2, {0x0, 0x0, 0x0, 0x0, 0x2020, 0x2020, 0x3FF8, 0x2024, 0x2024, 0x2038, 0x2020, 0x2020, 0x1040, 0xF80, 0x0, 0x0}},
    {42037, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x920, 0x920, 0x920, 0x0, 0xFE0, 0x0, 0x0}},
    {42038, 2, {0x380, 0x440, 0x0, 0x380, 0x440, 0x540, 0x540, 0x1E50, 0x1550, 0x14F0, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {42039, 2, {0x0, 0x0, 0x0, 0x380, 0x440, 0x540, 0x540, 0x1E50, 0x1550, 0x14F0, 0x540, 0x540, 0x440, 0x380, 0x0, 0x0}},
    {42040, 2, {0x0, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0xFE0, 0x100, 0x540, 0x1550, 0x1010, 0xFE0, 0x0, 0x0}},
    {42041, 2, {0x0, 0x380, 0x440, 0x0, 0x660, 0x5D0, 0x440, 0x440, 0x7C0, 0x440, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {42042, 2, {0x0, 0x0, 0x0, 0x0, 0x660, 0x5D0, 0x440, 0x440, 0x7C0, 0x440, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {42043, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6030, 0x5054, 0x4894, 0x4510, 0x4214, 0x4514, 0x4890, 0x5054, 0x6034, 0x0, 0x0}},
    {42044, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x6030, 0x5054, 0x4892, 0x4512, 0x421C, 0x4512, 0x4892, 0x5054, 0x6030, 0x0, 0x0}},
    {42045, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6030, 0x5054, 0x4892, 0x4512, 0x421C, 0x4512, 0x4892, 0x5054, 0x6030, 0x0, 0x0}},
    {42046, 2, {0x0, 0x0, 0x0, 0x0, 0xEE0, 0x100, 0x280, 0x280, 0xEA0, 0xBA0, 0xAE0, 0x280, 0x100, 0xEE0, 0x0, 0x0}},
    {42047, 2, {0x380, 0x440, 0x0, 0x1E0, 0x200, 0x440, 0x440, 0x400, 0x7E0, 0x400, 0x440, 0x440, 0x200, 0x1E0, 0x0, 0x0}},
    {42048, 2, {0x0, 0x0, 0x0, 0x1E0, 0x200, 0x440, 0x440, 0x400, 0x7E0, 0x400, 0x440, 0x440, 0x200, 0x1E0, 0x0, 0x0}},
    {42049, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x400, 0x820, 0x810, 0xF90, 0x810, 0x820, 0x800, 0x400, 0x3C0, 0x0, 0x0}},
    {42050, 2, {0x0, 0x380, 0x440, 0x0, 0x60, 0x58, 0x40, 0x3F8, 0x1840, 0x2040, 0x4C40, 0x4B40, 0x8C0, 0x800, 0x0, 0x0}},
    {42051, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0x2C, 0x20, 0x1FC, 0xC20, 0x1020, 0x2620, 0x25A0, 0x460, 0x400, 0x0, 0x0}},
    {42052, 2, {0x0, 0x0, 0x0, 0x0, 0xF20, 0x10D0, 0x1050, 0x1020, 0x800, 0x600, 0x100, 0x80, 0x1080, 0xF00, 0x0, 0x0}},
    {42053, 2, {0x380, 0x440, 0x0, 0x490, 0x2A0, 0x1C0, 0x1C0, 0x2A0, 0x490, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {42054, 2, {0x0, 0x0, 0x0, 0x490, 0x2A0, 0x1C0, 0x1C0, 0x2A0, 0x490, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {42055, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x180, 0x140, 0x120, 0x100, 0x100, 0x900, 0x500, 0x300, 0x0, 0x0}},
    {42056, 2, {0x380, 0x440, 0x0, 0x2C0, 0x1AA0, 0x690, 0x280, 0x1A80, 0x680, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {42057, 2, {0x0, 0x0, 0x0, 0x2C0, 0x1AA0, 0x690, 0x280, 0x1A80, 0x680, 0x80, 0x1880, 0x1680, 0x1180, 0x1000, 0x0, 0x0}},
    {42058, 2, {0x0, 0x0, 0x0, 0x1110, 0x1110, 0xFE0, 0x100, 0xFE0, 0x1110, 0x1110, 0x100, 0xD00, 0xB00, 0x800, 0x0, 0x0}},
    {42059, 2, {0x0, 0x380, 0x440, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x0, 0x0}},
    {42060, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x1FF0, 0x100, 0x0, 0x0}},
    {42061, 2, {0x0, 0x0, 0x0, 0x0, 0xC00, 0xB00, 0x8C0, 0x800, 0x840, 0x840, 0x840, 0x840, 0x800, 0x800, 0x0, 0x0}},
    {42062, 2, {0x0, 0x0, 0x0, 0x0, 0x6006, 0x1248, 0xA50, 0x990, 0x810, 0xFF0, 0x810, 0x990, 0x1248, 0x6246, 0x0, 0x0}},
    {42063, 2, {0x0, 0x380, 0x440, 0x0, 0x886, 0x888, 0x890, 0x810, 0x810, 0xFF0, 0x810, 0x4990, 0x2A48, 0x1A46, 0x0, 0x0}},
    {42064, 2, {0x0, 0x0, 0x0, 0x0, 0x886, 0x888, 0x890, 0x810, 0x810, 0xFF0, 0x810, 0x4990, 0x2A48, 0x1A46, 0x0, 0x0}},
    {42065, 2, {0x0, 0x0, 0x0, 0x0, 0x610C, 0x1110, 0x920, 0x820, 0x820, 0xFE0, 0x820, 0x920, 0x1110, 0x610C, 0x0, 0x0}},
    {42066, 2, {0x380, 0x440, 0x0, 0x380, 0x440, 0x920, 0x920, 0x440, 0x380, 0x0, 0xFF0, 0x60, 0x180, 0x600, 0xFF0, 0x0}},
    {42067, 2, {0x0, 0x0, 0x380, 0x440, 0x920, 0x920, 0x440, 0x380, 0x0, 0xFF0, 0x60, 0x180, 0x600, 0xFF0, 0x0, 0x0}},
    {42068, 2, {0x0, 0x0, 0x820, 0x2828, 0x2448, 0x2388, 0x1010, 0xFE0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x100, 0x0, 0x0}},
    {42069, 2, {0x0, 0x0, 0x0, 0x0, 0x3118, 0x920, 0x540, 0x540, 0x540, 0x540, 0x540, 0x540, 0x920, 0x3118, 0x0, 0x0}},
    {42070, 2, {0x0, 0x380, 0x440, 0x0, 0xD20, 0xBA0, 0x960, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0x1900, 0x700, 0x0, 0x0}},
    {42071, 2, {0x0, 0x180, 0x160, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0x1900, 0x700, 0x0, 0x0}},
    {42072, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x820, 0x820, 0x820, 0x820, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {42073, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1110, 0x1110, 0x1FF0, 0x1110, 0x1110, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {42074, 2, {0x0, 0x380, 0x440, 0x0, 0x880, 0x880, 0xE80, 0x9C0, 0x8B0, 0x880, 0x890, 0x890, 0x890, 0x880, 0x0, 0x0}},
    {42075, 2, {0x0, 0x0, 0x0, 0x0, 0x880, 0x880, 0xE80, 0x9C0, 0x8B0, 0x880, 0x890, 0x890, 0x890, 0x880, 0x0, 0x0}},
    {42076, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x7F0, 0x880, 0x880, 0x880, 0x7F0, 0x80, 0x90, 0x90, 0x90, 0x0, 0x0}},
    {42077, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x440, 0x280, 0x100, 0x280, 0x440, 0x0, 0x0}},
    {42078, 2, {0x0, 0x0, 0x380, 0x440, 0x0, 0x1108, 0x1108, 0xA08, 0xA08, 0x468, 0xA58, 0xA40, 0x1140, 0x1140, 0x0, 0x0}},
    {42079, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1108, 0x1108, 0xA08, 0xA08, 0x468, 0xA58, 0xA40, 0x1140, 0x1140, 0x0, 0x0}},
    {42080, 2, {0x0, 0x0, 0x0, 0x620, 0x920, 0x620, 0x20, 0x620, 0x920, 0x620, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {42081, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x440, 0x540, 0x380, 0x100, 0x900, 0x600, 0x0, 0x0}},
    {42082, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x440, 0x440, 0x440, 0x1540, 0x1540, 0x880, 0x0, 0x0}},
    {42083, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x440, 0x440, 0x440, 0x1540, 0x1540, 0x880, 0x0, 0x0}},
    {42084, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2208, 0x2208, 0x1468, 0x1458, 0x840, 0x14E0, 0x1440, 0x22E0, 0x2240, 0x0, 0x0}},
    {42085, 2, {0x0, 0x380, 0x440, 0x0, 0x920, 0x920, 0x440, 0x380, 0x440, 0x820, 0xFE0, 0x820, 0x440, 0x380, 0x0, 0x0}},
    {42086, 2, {0x0, 0x0, 0x0, 0x0, 0x920, 0x920, 0x440, 0x380, 0x440, 0x820, 0xFE0, 0x820, 0x440, 0x380, 0x0, 0x0}},
    {42087, 2, {0x0, 0x0, 0x0, 0x1110, 0x1110, 0xEE0, 0x0, 0x440, 0x380, 0x0, 0x1450, 0x1390, 0x820, 0x7C0, 0x0, 0x0}},
    {42088, 2, {0x0, 0x380, 0x440, 0x0, 0x660, 0x550, 0x4C8, 0x440, 0xE40, 0x1540, 0x1540, 0xE40, 0x440, 0x440, 0x0, 0x0}},
    {42089, 2, {0x0, 0x0, 0x0, 0x0, 0x660, 0x550, 0x4C8, 0x440, 0xE40, 0x1540, 0x1540, 0xE40, 0x440, 0x440, 0x0, 0x0}},
    {42090, 2, {0x0, 0x0, 0x0, 0x0, 0x1860, 0x1658, 0x11C6, 0x1040, 0x9640, 0x5A40, 0x3240, 0x5A40, 0x9640, 0x1040, 0x0, 0x0}},
    {42091, 2, {0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x380, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0x100, 0xFE0, 0x0, 0x0}},
    {42092, 2, {0x0, 0x380, 0x440, 0x0, 0x100, 0x100, 0x3110, 0x2D10, 0x2310, 0x21D0, 0x2130, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {42093, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x80, 0x0, 0x0}},
    {42094, 2, {0x0, 0x0, 0x0, 0x0, 0x220, 0x220, 0x1C0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x80, 0x80, 0x0, 0x0}},
    {42095, 2, {0x0, 0x380, 0x440, 0x0, 0x1110, 0x1110, 0x1110, 0x920, 0x7C0, 0x3110, 0x2D10, 0x2310, 0x21D0, 0x2130, 0x0, 0x0}},
    {42096, 2, {0x0, 0x0, 0x0, 0x0, 0x888, 0x888, 0x888, 0x490, 0x3E0, 0x1888, 0x1688, 0x1188, 0x10E8, 0x1098, 0x0, 0x0}},
    {42097, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0xC8, 0xA8, 0x98, 0x80, 0x1F80, 0x2480, 0x3F80, 0x2480, 0x1F80, 0x0, 0x0}},
    {42098, 2, {0x0, 0x380, 0x440, 0x0, 0x28F0, 0x2908, 0x3A04, 0x2804, 0x2804, 0x4, 0x4, 0x384, 0x264, 0x218, 0x0, 0x0}},
    {42099, 2, {0x0, 0x0, 0x0, 0x0, 0x28F0, 0x2908, 0x3A04, 0x2804, 0x2804, 0x4, 0x4, 0x384, 0x264, 0x218, 0x0, 0x0}},
    {42100, 2, {0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x920, 0x540, 0x380, 0x540, 0x920, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {42101, 2, {0x0, 0x0, 0x0, 0x0, 0x4, 0x64, 0x54, 0x4C, 0xFC0, 0x38E0, 0x1540, 0x1540, 0x38E0, 0x7C0, 0x0, 0x0}},
    {42102, 2, {0x0, 0x380, 0x440, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0x1FC0, 0x200, 0x1FC0, 0x200, 0x200, 0x0, 0x0}},
    {42103, 2, {0x0, 0x0, 0x0, 0x0, 0x10, 0x310, 0x2D0, 0x230, 0x200, 0x1FC0, 0x200, 0x1FC0, 0x200, 0x200, 0x0, 0x0}},
    {42104, 2, {0x0, 0x1C0, 0x130, 0x100, 0x7C0, 0x920, 0x1550, 0x1390, 0x1550, 0x920, 0x7C0, 0x100, 0x1900, 0x700, 0x0, 0x0}},
    {42105, 2, {0x0, 0x0, 0x0, 0x0, 0x12C0, 0x12B0, 0xC80, 0xC0, 0xCB0, 0x1280, 0x12C0, 0xB0, 0x680, 0x180, 0x0, 0x0}},
    {42106, 2, {0x0, 0x380, 0x440, 0x0, 0xCC0, 0x12B0, 0x1280, 0x1E80, 0x1280, 0x12E0, 0xC90, 0x90, 0x690, 0x180, 0x0, 0x0}},
    {42107, 2, {0x0, 0x0, 0x0, 0x0, 0xCC0, 0x12B0, 0x1280, 0x1E80, 0x1280, 0x12E0, 0xC90, 0x90, 0x690, 0x180, 0x0, 0x0}},
    {42108, 2, {0x0, 0x0, 0x0, 0x200, 0xFA0, 0x220, 0x20, 0xFA0, 0x1060, 0x1000, 0x1F00, 0x1020, 0x1040, 0xF80, 0x0, 0x0}},
    {42109, 2, {0x0, 0x0, 0x0, 0x0, 0xCC0, 0x12B0, 0x1280, 0x80, 0x1280, 0x1280, 0xC80, 0x80, 0x680, 0x180, 0x0, 0x0}},
    {42110, 2, {0x0, 0x380, 0x440, 0x0, 0x3D0, 0x430, 0x810, 0x800, 0xF80, 0x800, 0x800, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {42111, 2, {0x0, 0x0, 0x0, 0x0, 0x3D0, 0x430, 0x810, 0x800, 0xF80, 0x800, 0x800, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {42112, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x440, 0x380, 0x120, 0x140, 0x180, 0x110, 0x110, 0xE0, 0x0, 0x0}},
    {42113, 2, {0x0, 0x0, 0x180, 0x160, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0x7C0, 0x100, 0xD00, 0xB00, 0x800, 0x0, 0x0}},
    {42114, 2, {0x380, 0x440, 0x0, 0x1110, 0x1110, 0x17D0, 0x1110, 0x1110, 0x820, 0x7C0, 0x100, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {42115, 2, {0x0, 0x0, 0x0, 0x1110, 0x1110, 0x17D0, 0x1110, 0x1110, 0x820, 0x7C0, 0x100, 0x1110, 0x1110, 0xEE0, 0x0, 0x0}},
    {42116, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x920, 0x920, 0xFE0, 0x920, 0x920, 0x20, 0x620, 0x1A0, 0x60, 0x0, 0x0}},
    {42117, 2, {0x380, 0x440, 0x0, 0x180, 0x160, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0x7C0, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {42118, 2, {0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0xD20, 0xBA0, 0x960, 0x100, 0x7C0, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {42119, 2, {0x0, 0x0, 0x0, 0x920, 0x920, 0x440, 0x380, 0x0, 0xFE0, 0x0, 0x380, 0x440, 0x920, 0x920, 0x0, 0x0}},
    {42120, 2, {0x0, 0x380, 0x440, 0x0, 0x40, 0x40, 0x3C0, 0x440, 0x840, 0x840, 0x440, 0x3C0, 0x40, 0x40, 0x0, 0x0}},
    {42121, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x3C0, 0x440, 0x840, 0x840, 0x440, 0x3C0, 0x40, 0x40, 0x0, 0x0}},
    {42122, 2, {0x0, 0x0, 0x0, 0x0, 0xF0, 0x700, 0x900, 0x1080, 0x1070, 0x1080, 0x900, 0x700, 0x100, 0xF0, 0x0, 0x0}},
    {42123, 2, {0x380, 0x440, 0x0, 0x3938, 0x540, 0x540, 0x3938, 0x100, 0x1FF0, 0x100, 0x3938, 0x540, 0x540, 0x3938, 0x0, 0x0}},
    {42124, 2, {0x0, 0x0, 0x0, 0x3938, 0x540, 0x540, 0x3938, 0x100, 0x1FF0, 0x100, 0x3938, 0x540, 0x540, 0x3938, 0x0, 0x0}},
    {42128, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {42129, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0xC40, 0x340, 0xC0, 0x0, 0x0}},
    {42130, 2, {0x0, 0x0, 0x0, 0x0, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x460, 0x580, 0x600, 0x0, 0x0}},
    {42131, 2, {0x0, 0x0, 0x0, 0x0, 0x600, 0x580, 0x460, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x0, 0x0}},
    {42132, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xD00, 0x300, 0x0, 0x0}},
    {42133, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x620, 0x5A0, 0x460, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x0, 0x0}},
    {42134, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0xC40, 0xB40, 0x8C0, 0x800, 0x0, 0x0}},
    {42135, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x160, 0x118, 0x100, 0x100, 0x100, 0x3100, 0x2D00, 0x2300, 0x2000, 0x0, 0x0}},
    {42136, 2, {0x0, 0x0, 0x0, 0x0, 0xC8, 0xA8, 0x98, 0x80, 0x80, 0x80, 0x80, 0x1880, 0x680, 0x180, 0x0, 0x0}},
    {42137, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x1608, 0x1188, 0x1068, 0x1018, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42138, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x400, 0x400, 0x200, 0x180, 0x40, 0x20, 0x420, 0x3C0, 0x0, 0x0}},
    {42139, 2, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x340, 0x2C0, 0x200, 0x200, 0x200, 0x200, 0x200, 0x0, 0x0}},
    {42140, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x420, 0x810, 0x10, 0x10, 0x10, 0x10, 0xE10, 0x990, 0x860, 0x0, 0x0}},
    {42141, 2, {0x0, 0x0, 0x0, 0x0, 0x3D0, 0x430, 0x810, 0x800, 0x800, 0x800, 0x800, 0x810, 0x420, 0x3C0, 0x0, 0x0}},
    {42142, 2, {0x0, 0x0, 0x0, 0x0, 0xE20, 0x9A0, 0x860, 0x800, 0x800, 0x800, 0x800, 0x860, 0x9A0, 0xE20, 0x0, 0x0}},
    {42143, 2, {0x0, 0x0, 0x0, 0x0, 0x1E0, 0x200, 0x400, 0x400, 0x400, 0x400, 0x400, 0x400, 0x200, 0x1E0, 0x0, 0x0}},
    {42144, 2, {0x0, 0x0, 0x0, 0x0, 0xFC0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0xFC0, 0x0, 0x0}},
    {42145, 2, {0x0, 0x0, 0x0, 0x0, 0x820, 0x820, 0x440, 0x380, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42146, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x820, 0x820, 0x0, 0x0}},
    {42147, 2, {0x0, 0x0, 0x0, 0x0, 0x1808, 0x608, 0x188, 0x68, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42148, 2, {0x0, 0x0, 0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {42149, 2, {0x0, 0x0, 0x0, 0x0, 0x860, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x700, 0x0, 0x0}},
    {42150, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x2448, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {42151, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x820, 0x820, 0x820, 0x820, 0x920, 0x920, 0x6C0, 0x0, 0x0}},
    {42152, 2, {0x0, 0x0, 0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0}},
    {42153, 2, {0x0, 0x0, 0x0, 0x0, 0x380, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x440, 0x380, 0x0, 0x0}},
    {42154, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x1010, 0x1010, 0x1010, 0xFE0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42155, 2, {0x0, 0x0, 0x0, 0x0, 0x100, 0x280, 0x440, 0x820, 0xFE0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42156, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x90, 0x80, 0x780, 0x880, 0x1080, 0x1080, 0x880, 0x780, 0x0, 0x0}},
    {42157, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xA0, 0x90, 0x80, 0x7C0, 0x8A0, 0x1090, 0x1080, 0x880, 0x780, 0x0, 0x0}},
    {42158, 2, {0x0, 0x0, 0x0, 0x0, 0x8, 0xC8, 0xA8, 0x98, 0x780, 0x880, 0x1080, 0x1080, 0x880, 0x780, 0x0, 0x0}},
    {42159, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x1E0, 0x220, 0x420, 0x420, 0x220, 0x1E0, 0x0, 0x0}},
    {42160, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0x400, 0x400, 0x200, 0x1C0, 0x200, 0x400, 0x400, 0x400, 0x3C0, 0x0, 0x0}},
    {42161, 2, {0x0, 0x0, 0x0, 0x0, 0x780, 0x40, 0x40, 0x80, 0x700, 0x80, 0x40, 0x40, 0x40, 0x780, 0x0, 0x0}},
    {42162, 2, {0x0, 0x0, 0x0, 0x0, 0x3E0, 0x400, 0x400, 0x3E0, 0x400, 0x400, 0x3E0, 0x400, 0x400, 0x3E0, 0x0, 0x0}},
    {42163, 2, {0x0, 0x0, 0x0, 0x0, 0x1110, 0x1110, 0x1110, 0xEE0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42164, 2, {0x0, 0x0, 0x0, 0x0, 0x4444, 0x4444, 0x4C64, 0x3458, 0x440, 0x440, 0x380, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42165, 2, {0x0, 0x0, 0x0, 0x0, 0x1C70, 0x820, 0x820, 0x440, 0x380, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42166, 2, {0x0, 0x0, 0x0, 0x0, 0x4444, 0x4444, 0x4444, 0x3BB8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42167, 2, {0x0, 0x0, 0x0, 0x0, 0x1830, 0x2448, 0x440, 0x440, 0x380, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {42168, 2, {0x0, 0x0, 0x0, 0x0, 0x710, 0x890, 0x8A0, 0x7C0, 0x80, 0x80, 0x80, 0x80, 0xC80, 0x300, 0x0, 0x0}},
    {42169, 2, {0x0, 0x0, 0x0, 0x0, 0xE20, 0x1120, 0x1140, 0xF80, 0x100, 0x100, 0x100, 0x120, 0x120, 0xC0, 0x0, 0x0}},
    {42170, 2, {0x0, 0x0, 0x0, 0x0, 0x790, 0x860, 0x860, 0x7A0, 0x20, 0x20, 0x7A0, 0x860, 0x860, 0x790, 0x0, 0x0}},
    {42171, 2, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x220, 0x520, 0x520, 0x320, 0x4C0, 0x0, 0x0}},
    {42172, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x280, 0x280, 0x100, 0x280, 0x280, 0x440, 0x440, 0x0, 0x0}},
    {42173, 2, {0x0, 0x0, 0x0, 0x0, 0xC60, 0x280, 0x100, 0x280, 0x280, 0x440, 0x440, 0x820, 0x820, 0x820, 0x0, 0x0}},
    {42174, 2, {0x0, 0x0, 0x0, 0xFE0, 0x820, 0x440, 0x280, 0x280, 0x100, 0x280, 0x280, 0x440, 0x820, 0xFE0, 0x0, 0x0}},
    {42175, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3018, 0x2828, 0x2448, 0x2288, 0x2108, 0x2288, 0x2448, 0x2828, 0x3018, 0x0, 0x0}},
    {42176, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1018, 0x828, 0x448, 0x288, 0x108, 0x288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {42177, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3018, 0x2828, 0x2448, 0x288, 0x108, 0x288, 0x448, 0x828, 0x1018, 0x0, 0x0}},
    {42178, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x190C, 0x2514, 0x2224, 0x1D44, 0x84, 0x1D44, 0x2224, 0x2514, 0x190C, 0x0, 0x0}},
    {42179, 2, {0x0, 0x0, 0x0, 0xFE0, 0x820, 0x440, 0x440, 0x280, 0x280, 0x100, 0x280, 0x280, 0x440, 0x440, 0x0, 0x0}},
    {42180, 2, {0x0, 0x0, 0x0, 0x0, 0xFE0, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x0, 0x0}},
    {42181, 2, {0x0, 0x0, 0x0, 0x0, 0x1FC0, 0x220, 0x210, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x200, 0x0, 0x0}},
    {42182, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x440, 0x440, 0x440, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {44032, 2, {0x0, 0x0, 0x10, 0x10, 0x1F90, 0x90, 0x90, 0x9E, 0x110, 0x110, 0x210, 0x410, 0x1810, 0x10, 0x10, 0x0}},
    {44033, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44034, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44035, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44036, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44037, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44038, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44039, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44040, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44041, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44042, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44043, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44044, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44045, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44046, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44047, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44048, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44049, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44050, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44051, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44052, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44053, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44054, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44055, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44056, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44057, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44058, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44059, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x40E, 0x808, 0x3008, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44060, 2, {0x0, 0x0, 0x12, 0x12, 0x1F92, 0x92, 0x92, 0x9E, 0x112, 0x112, 0x212, 0x412, 0x1812, 0x12, 0x12, 0x0}},
    {44061, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44062, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44063, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44064, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44065, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44066, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44067, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44068, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44069, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44070, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44071, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44072, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44073, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44074, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44075, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44076, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44077, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44078, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44079, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44080, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44081, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44082, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44083, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44084, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44085, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44086, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44087, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x438, 0x828, 0x3028, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44088, 2, {0x0, 0x0, 0x10, 0x10, 0x1F90, 0x90, 0x9E, 0x90, 0x110, 0x11E, 0x210, 0x410, 0x1810, 0x10, 0x10, 0x0}},
    {44089, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44090, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44091, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44092, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44093, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44094, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44095, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44096, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44097, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44098, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44099, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44100, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44101, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44102, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44103, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44104, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44105, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44106, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44107, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44108, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44109, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44110, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44111, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44112, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44113, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44114, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44115, 2, {0x0, 0x8, 0x3E08, 0x208, 0x20E, 0x408, 0x80E, 0x3008, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44116, 2, {0x0, 0x0, 0x12, 0x12, 0x1F92, 0x92, 0x9E, 0x92, 0x112, 0x11E, 0x212, 0x412, 0x1812, 0x12, 0x12, 0x0}},
    {44117, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44118, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44119, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44120, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44121, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44122, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44123, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44124, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44125, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44126, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44127, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44128, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44129, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44130, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44131, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44132, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44133, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44134, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44135, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44136, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44137, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44138, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44139, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44140, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44141, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44142, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44143, 2, {0x0, 0x28, 0x3E28, 0x228, 0x238, 0x428, 0x838, 0x3028, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44144, 2, {0x0, 0x0, 0x2, 0x2, 0x1F82, 0x82, 0x82, 0x9E, 0x102, 0x102, 0x202, 0x402, 0x1802, 0x2, 0x2, 0x0}},
    {44145, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44146, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44147, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44148, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44149, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44150, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44151, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44152, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44153, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44154, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44155, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44156, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44157, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44158, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44159, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44160, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44161, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44162, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44163, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44164, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44165, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44166, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44167, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44168, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44169, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44170, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44171, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x438, 0x808, 0x3008, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44172, 2, {0x0, 0x0, 0xA, 0xA, 0x1F8A, 0x8A, 0x8A, 0xBA, 0x10A, 0x10A, 0x20A, 0x40A, 0x180A, 0xA, 0xA, 0x0}},
    {44173, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44174, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44175, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44176, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44177, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44178, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44179, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44180, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44181, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44182, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44183, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44184, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44185, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44186, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44187, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44188, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44189, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44190, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44191, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44192, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44193, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44194, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44195, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44196, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44197, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44198, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44199, 2, {0x0, 0x28, 0x3E28, 0x228, 0x228, 0x4E8, 0x828, 0x3028, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44200, 2, {0x0, 0x0, 0x2, 0x2, 0x1F82, 0x82, 0x9E, 0x82, 0x102, 0x11E, 0x202, 0x402, 0x1802, 0x2, 0x2, 0x0}},
    {44201, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44202, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44203, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44204, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44205, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44206, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44207, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44208, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44209, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44210, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44211, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44212, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44213, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44214, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44215, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44216, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44217, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44218, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44219, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44220, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44221, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44222, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44223, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44224, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44225, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44226, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44227, 2, {0x0, 0x8, 0x3E08, 0x208, 0x238, 0x408, 0x838, 0x3008, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44228, 2, {0x0, 0x0, 0xA, 0xA, 0x1F8A, 0x8A, 0xBA, 0x8A, 0x10A, 0x13A, 0x20A, 0x40A, 0x180A, 0xA, 0xA, 0x0}},
    {44229, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44230, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44231, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44232, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44233, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44234, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44235, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44236, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44237, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44238, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44239, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44240, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44241, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44242, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44243, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44244, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44245, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44246, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44247, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44248, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44249, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44250, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44251, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44252, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44253, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44254, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44255, 2, {0x0, 0x28, 0x3E28, 0x228, 0x2E8, 0x428, 0x8E8, 0x3028, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44256, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x10, 0x210, 0x200, 0x200, 0x200, 0x7FFC, 0x0, 0x0}},
    {44257, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44258, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44259, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44260, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44261, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44262, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44263, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44264, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44265, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44266, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44267, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44268, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44269, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44270, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44271, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44272, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44273, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44274, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44275, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44276, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44277, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44278, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44279, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44280, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44281, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44282, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44283, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44284, 2, {0x0, 0x0, 0x10, 0x10, 0x3FD0, 0x50, 0x50, 0x5E, 0x450, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x10, 0x0}},
    {44285, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44286, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44287, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44288, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44289, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44290, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44291, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44292, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44293, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44294, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44295, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44296, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44297, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44298, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44299, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44300, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44301, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44302, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44303, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44304, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44305, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44306, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44307, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44308, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44309, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44310, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44311, 2, {0x0, 0x8, 0x1F88, 0x88, 0x8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44312, 2, {0x0, 0x0, 0x12, 0x12, 0x3FD2, 0x52, 0x52, 0x5E, 0x452, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x12, 0x0}},
    {44313, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44314, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44315, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44316, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44317, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44318, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44319, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44320, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44321, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44322, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44323, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44324, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44325, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44326, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44327, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44328, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44329, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44330, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44331, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44332, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44333, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44334, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44335, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44336, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44337, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44338, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44339, 2, {0x0, 0x28, 0x1FA8, 0xA8, 0xB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44340, 2, {0x0, 0x0, 0x4, 0x4, 0x3FC4, 0x44, 0x44, 0x44, 0x444, 0x404, 0x404, 0x7FF4, 0x4, 0x4, 0x4, 0x0}},
    {44341, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44342, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44343, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44344, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44345, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44346, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44347, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44348, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44349, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44350, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44351, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44352, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44353, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44354, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44355, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44356, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44357, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44358, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44359, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44360, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44361, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44362, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44363, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44364, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44365, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44366, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44367, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44368, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x10, 0x890, 0x880, 0x880, 0x880, 0x7FFC, 0x0, 0x0}},
    {44369, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44370, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44371, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44372, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44373, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44374, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44375, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44376, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44377, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44378, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44379, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44380, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44381, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44382, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44383, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44384, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44385, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44386, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44387, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44388, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44389, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44390, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44391, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44392, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44393, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44394, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44395, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44396, 2, {0x0, 0x0, 0x0, 0x3FE0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {44397, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44398, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44399, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44400, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44401, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44402, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44403, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44404, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44405, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44406, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44407, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44408, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44409, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44410, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44411, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44412, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44413, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44414, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44415, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44416, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44417, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44418, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44419, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44420, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44421, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44422, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44423, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44424, 2, {0x0, 0x8, 0x8, 0x8, 0x3FE8, 0x28, 0x28, 0x28, 0x28, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {44425, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44426, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44427, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44428, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44429, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44430, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44431, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44432, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44433, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44434, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44435, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44436, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44437, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44438, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44439, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44440, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44441, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44442, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44443, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44444, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44445, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44446, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44447, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44448, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44449, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44450, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44451, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44452, 2, {0x0, 0xA, 0xA, 0xA, 0x3FEA, 0x2A, 0x2A, 0x2A, 0x2A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {44453, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44454, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44455, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44456, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44457, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44458, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44459, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44460, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44461, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44462, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44463, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44464, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44465, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44466, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44467, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44468, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44469, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44470, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44471, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44472, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44473, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44474, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44475, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44476, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44477, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44478, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44479, 2, {0x28, 0x28, 0x3F28, 0x128, 0x128, 0x128, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44480, 2, {0x0, 0x8, 0x8, 0x8, 0x3FE8, 0x28, 0x28, 0x28, 0x28, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {44481, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44482, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44483, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44484, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44485, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44486, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44487, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44488, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44489, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44490, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44491, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44492, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44493, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44494, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44495, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44496, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44497, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44498, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44499, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44500, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44501, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44502, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44503, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44504, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44505, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44506, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44507, 2, {0x8, 0x8, 0x3F08, 0x108, 0x108, 0x108, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44508, 2, {0x0, 0x0, 0x0, 0x3FE0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {44509, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44510, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44511, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44512, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44513, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44514, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44515, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44516, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44517, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44518, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44519, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44520, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44521, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44522, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44523, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44524, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44525, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44526, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44527, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44528, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44529, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44530, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44531, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44532, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44533, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44534, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44535, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44536, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {44537, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44538, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44539, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44540, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44541, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44542, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44543, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44544, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44545, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44546, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44547, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44548, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44549, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44550, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44551, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44552, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44553, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44554, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44555, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44556, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44557, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44558, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44559, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44560, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44561, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44562, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44563, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44564, 2, {0x0, 0x0, 0x8, 0x8, 0x3FC8, 0x48, 0x48, 0x48, 0x48, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {44565, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44566, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44567, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44568, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44569, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44570, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44571, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44572, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44573, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44574, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44575, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44576, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44577, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44578, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44579, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44580, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44581, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44582, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44583, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44584, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44585, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44586, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44587, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44588, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44589, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44590, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44591, 2, {0x0, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44592, 2, {0x0, 0x0, 0x8, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x108, 0x108, 0x208, 0x408, 0x1808, 0x8, 0x8, 0x0}},
    {44593, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44594, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44595, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44596, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44597, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44598, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44599, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44600, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44601, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44602, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44603, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44604, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44605, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44606, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44607, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44608, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44609, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44610, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44611, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44612, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44613, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44614, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44615, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44616, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44617, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44618, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44619, 2, {0x0, 0x8, 0x3E08, 0x208, 0x208, 0x408, 0x808, 0x3008, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44620, 2, {0x0, 0x0, 0x10, 0x10, 0x7B90, 0x890, 0x890, 0x89E, 0x1110, 0x1110, 0x2210, 0x4410, 0x810, 0x10, 0x10, 0x0}},
    {44621, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44622, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44623, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44624, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44625, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44626, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44627, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44628, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44629, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44630, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44631, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44632, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44633, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44634, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44635, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44636, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44637, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44638, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44639, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44640, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44641, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44642, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44643, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44644, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44645, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44646, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44647, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x108E, 0x2108, 0xC608, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44648, 2, {0x0, 0x0, 0x12, 0x12, 0x7B92, 0x892, 0x892, 0x89E, 0x1112, 0x1112, 0x2212, 0x4412, 0x812, 0x12, 0x12, 0x0}},
    {44649, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44650, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44651, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44652, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44653, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44654, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44655, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44656, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44657, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44658, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44659, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44660, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44661, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44662, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44663, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44664, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44665, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44666, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44667, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44668, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44669, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44670, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44671, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44672, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44673, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44674, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44675, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10B8, 0x2128, 0xC628, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44676, 2, {0x0, 0x0, 0x10, 0x10, 0x7B90, 0x890, 0x89E, 0x890, 0x1110, 0x111E, 0x2210, 0x4410, 0x810, 0x10, 0x10, 0x0}},
    {44677, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44678, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44679, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44680, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44681, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44682, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44683, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44684, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44685, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44686, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44687, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44688, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44689, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44690, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44691, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44692, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44693, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44694, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44695, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44696, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44697, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44698, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44699, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44700, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44701, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44702, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44703, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x210E, 0xC608, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44704, 2, {0x0, 0x0, 0x12, 0x12, 0x7B92, 0x892, 0x89E, 0x892, 0x1112, 0x111E, 0x2212, 0x4412, 0x812, 0x12, 0x12, 0x0}},
    {44705, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44706, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44707, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44708, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44709, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44710, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44711, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44712, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44713, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44714, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44715, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44716, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44717, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44718, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44719, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44720, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44721, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44722, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44723, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44724, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44725, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44726, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44727, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44728, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44729, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44730, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44731, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x2138, 0xC628, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44732, 2, {0x0, 0x0, 0x2, 0x2, 0x7B82, 0x882, 0x882, 0x89E, 0x1102, 0x1102, 0x2202, 0x4402, 0x802, 0x2, 0x2, 0x0}},
    {44733, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44734, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44735, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44736, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44737, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44738, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44739, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44740, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44741, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44742, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44743, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44744, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44745, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44746, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44747, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44748, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44749, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44750, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44751, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44752, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44753, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44754, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44755, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44756, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44757, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44758, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44759, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x10B8, 0x2108, 0xC608, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44760, 2, {0x0, 0x0, 0xA, 0xA, 0x7B8A, 0x88A, 0x88A, 0x8BA, 0x110A, 0x110A, 0x220A, 0x440A, 0x80A, 0xA, 0xA, 0x0}},
    {44761, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44762, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44763, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44764, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44765, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44766, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44767, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44768, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44769, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44770, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44771, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44772, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44773, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44774, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44775, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44776, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44777, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44778, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44779, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44780, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44781, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44782, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44783, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44784, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44785, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44786, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44787, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10E8, 0x2128, 0xC628, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44788, 2, {0x0, 0x0, 0x2, 0x2, 0x7B82, 0x882, 0x89E, 0x882, 0x1102, 0x111E, 0x2202, 0x4402, 0x802, 0x2, 0x2, 0x0}},
    {44789, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {44790, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44791, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44792, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {44793, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44794, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44795, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44796, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44797, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44798, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44799, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44800, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {44801, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44802, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44803, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44804, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44805, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44806, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {44807, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44808, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44809, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44810, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44811, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44812, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44813, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44814, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44815, 2, {0x0, 0x8, 0xF788, 0x1088, 0x10B8, 0x1088, 0x2138, 0xC608, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44816, 2, {0x0, 0x0, 0xA, 0xA, 0x7B8A, 0x88A, 0x8BA, 0x88A, 0x110A, 0x113A, 0x220A, 0x440A, 0x80A, 0xA, 0xA, 0x0}},
    {44817, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44818, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44819, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44820, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44821, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44822, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44823, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44824, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44825, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44826, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44827, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44828, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44829, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44830, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44831, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {44832, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44833, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44834, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44835, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {44836, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44837, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {44838, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44839, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44840, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44841, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44842, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44843, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10E8, 0x10A8, 0x21E8, 0xC628, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44844, 2, {0x0, 0x0, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x208, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {44845, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44846, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44847, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44848, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44849, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44850, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44851, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44852, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44853, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44854, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44855, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44856, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44857, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44858, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44859, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44860, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44861, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44862, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44863, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44864, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44865, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44866, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44867, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44868, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44869, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44870, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44871, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44872, 2, {0x0, 0x10, 0x10, 0x10, 0x7BD0, 0x850, 0x850, 0x85E, 0x850, 0x10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {44873, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44874, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44875, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44876, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44877, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44878, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44879, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44880, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44881, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44882, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44883, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44884, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44885, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44886, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44887, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44888, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44889, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44890, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44891, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44892, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44893, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44894, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44895, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44896, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44897, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44898, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44899, 2, {0x0, 0x8, 0xF788, 0x1088, 0x108E, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44900, 2, {0x0, 0x12, 0x12, 0x12, 0x7BD2, 0x852, 0x852, 0x85E, 0x852, 0x12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {44901, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44902, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44903, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44904, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44905, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44906, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44907, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44908, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44909, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44910, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44911, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44912, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44913, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44914, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44915, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44916, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44917, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44918, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44919, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44920, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44921, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44922, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44923, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44924, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44925, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44926, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44927, 2, {0x0, 0x28, 0xF7A8, 0x10A8, 0x10B8, 0x10A8, 0x14A8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44928, 2, {0x0, 0x8, 0x8, 0x8, 0x7BC8, 0x848, 0x848, 0x848, 0x848, 0x8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {44929, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {44930, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44931, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {44932, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {44933, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {44934, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {44935, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {44936, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {44937, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44938, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44939, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44940, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {44941, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44942, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44943, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {44944, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {44945, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {44946, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {44947, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {44948, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {44949, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {44950, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {44951, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {44952, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {44953, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {44954, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {44955, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1488, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {44956, 2, {0x0, 0x0, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x208, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {44957, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44958, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44959, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44960, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44961, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44962, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44963, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44964, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44965, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44966, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44967, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44968, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44969, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44970, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44971, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {44972, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {44973, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {44974, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {44975, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {44976, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {44977, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {44978, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44979, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {44980, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {44981, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44982, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {44983, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {44984, 2, {0x0, 0x0, 0x0, 0x0, 0x7DF0, 0x410, 0x410, 0x410, 0x410, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {44985, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {44986, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {44987, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {44988, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44989, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {44990, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {44991, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {44992, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {44993, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {44994, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {44995, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {44996, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {44997, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {44998, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {44999, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45000, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45001, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45002, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45003, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45004, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45005, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45006, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45007, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45008, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45009, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45010, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45011, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45012, 2, {0x0, 0x8, 0x8, 0x8, 0x7BC8, 0x848, 0x848, 0x848, 0x8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {45013, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45014, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45015, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45016, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45017, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45018, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45019, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45020, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45021, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45022, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45023, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45024, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45025, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45026, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45027, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45028, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45029, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45030, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45031, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45032, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45033, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45034, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45035, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45036, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45037, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45038, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45039, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45040, 2, {0x0, 0xA, 0xA, 0xA, 0x7BCA, 0x84A, 0x84A, 0x84A, 0xA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {45041, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45042, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45043, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45044, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45045, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45046, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45047, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45048, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45049, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45050, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45051, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45052, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45053, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45054, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45055, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45056, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45057, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45058, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45059, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45060, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45061, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45062, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45063, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45064, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45065, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45066, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45067, 2, {0x28, 0x28, 0xF7A8, 0x10A8, 0x10A8, 0x10A8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45068, 2, {0x0, 0x8, 0x8, 0x8, 0x7BC8, 0x848, 0x848, 0x848, 0x8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {45069, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45070, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45071, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45072, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45073, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45074, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45075, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45076, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45077, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45078, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45079, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45080, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45081, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45082, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45083, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45084, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45085, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45086, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45087, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45088, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45089, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45090, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45091, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45092, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45093, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45094, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45095, 2, {0x8, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45096, 2, {0x0, 0x0, 0x0, 0x0, 0x7DF0, 0x410, 0x410, 0x410, 0x410, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {45097, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45098, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45099, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45100, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45101, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45102, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45103, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45104, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45105, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45106, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45107, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45108, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45109, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45110, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45111, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45112, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45113, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45114, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45115, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45116, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45117, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45118, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45119, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45120, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45121, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45122, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45123, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45124, 2, {0x0, 0x0, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {45125, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45126, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45127, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45128, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45129, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45130, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45131, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45132, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45133, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45134, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45135, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45136, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45137, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45138, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45139, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45140, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45141, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45142, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45143, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45144, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45145, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45146, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45147, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45148, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45149, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45150, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45151, 2, {0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45152, 2, {0x0, 0x8, 0x8, 0x8, 0x7BC8, 0x848, 0x848, 0x848, 0x848, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {45153, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45154, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45155, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45156, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45157, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45158, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45159, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45160, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45161, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45162, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45163, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45164, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45165, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45166, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45167, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45168, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45169, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45170, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45171, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45172, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45173, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45174, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45175, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45176, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45177, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45178, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45179, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x1088, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45180, 2, {0x0, 0x0, 0x8, 0x8, 0x7B88, 0x888, 0x888, 0x888, 0x1108, 0x1108, 0x2208, 0x4408, 0x808, 0x8, 0x8, 0x0}},
    {45181, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45182, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45183, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45184, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45185, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45186, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45187, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45188, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45189, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45190, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45191, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45192, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45193, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45194, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45195, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45196, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45197, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45198, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45199, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45200, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45201, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45202, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45203, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45204, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45205, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45206, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45207, 2, {0x0, 0x8, 0xF788, 0x1088, 0x1088, 0x1088, 0x2108, 0xC608, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45208, 2, {0x0, 0x0, 0x10, 0x10, 0x10, 0x4010, 0x4010, 0x401E, 0x4010, 0x4010, 0x4010, 0x7F90, 0x10, 0x10, 0x10, 0x0}},
    {45209, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45210, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45211, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45212, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45213, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45214, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45215, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45216, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45217, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45218, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45219, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45220, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45221, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45222, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45223, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45224, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45225, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45226, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45227, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45228, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45229, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45230, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45231, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45232, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45233, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45234, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45235, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x400E, 0x4008, 0x4008, 0x7E08, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45236, 2, {0x0, 0x0, 0x12, 0x12, 0x12, 0x4012, 0x4012, 0x401E, 0x4012, 0x4012, 0x4012, 0x7F92, 0x12, 0x12, 0x12, 0x0}},
    {45237, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45238, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45239, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45240, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45241, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45242, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45243, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45244, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45245, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45246, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45247, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45248, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45249, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45250, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45251, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45252, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45253, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45254, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45255, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45256, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45257, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45258, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45259, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45260, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45261, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45262, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45263, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x4038, 0x4028, 0x4028, 0x7E28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45264, 2, {0x0, 0x0, 0x10, 0x10, 0x10, 0x4010, 0x401E, 0x4010, 0x4010, 0x401E, 0x4010, 0x7F90, 0x10, 0x10, 0x10, 0x0}},
    {45265, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45266, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45267, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45268, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45269, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45270, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45271, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45272, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45273, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45274, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45275, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45276, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45277, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45278, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45279, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45280, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45281, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45282, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45283, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45284, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45285, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45286, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45287, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45288, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45289, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45290, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45291, 2, {0x0, 0x8, 0x8, 0x8, 0x400E, 0x4008, 0x400E, 0x4008, 0x7E08, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45292, 2, {0x0, 0x0, 0x12, 0x12, 0x12, 0x4012, 0x401E, 0x4012, 0x4012, 0x401E, 0x4012, 0x7F92, 0x12, 0x12, 0x12, 0x0}},
    {45293, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45294, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45295, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45296, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45297, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45298, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45299, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45300, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45301, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45302, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45303, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45304, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45305, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45306, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45307, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45308, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45309, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45310, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45311, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45312, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45313, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45314, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45315, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45316, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45317, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45318, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45319, 2, {0x0, 0x28, 0x28, 0x28, 0x4038, 0x4028, 0x4038, 0x4028, 0x7E28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45320, 2, {0x0, 0x0, 0x2, 0x2, 0x2, 0x4002, 0x4002, 0x401E, 0x4002, 0x4002, 0x4002, 0x7F82, 0x2, 0x2, 0x2, 0x0}},
    {45321, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45322, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45323, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45324, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45325, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45326, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45327, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45328, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45329, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45330, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45331, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45332, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45333, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45334, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45335, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45336, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45337, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45338, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45339, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45340, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45341, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45342, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45343, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45344, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45345, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45346, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45347, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4038, 0x4008, 0x4008, 0x7E08, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45348, 2, {0x0, 0x0, 0xA, 0xA, 0xA, 0x400A, 0x400A, 0x403A, 0x400A, 0x400A, 0x400A, 0x7F8A, 0xA, 0xA, 0xA, 0x0}},
    {45349, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45350, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45351, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45352, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45353, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45354, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45355, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45356, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45357, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45358, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45359, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45360, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45361, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45362, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45363, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45364, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45365, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45366, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45367, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45368, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45369, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45370, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45371, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45372, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45373, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45374, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45375, 2, {0x0, 0x28, 0x28, 0x28, 0x4028, 0x40E8, 0x4028, 0x4028, 0x7E28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45376, 2, {0x0, 0x0, 0x2, 0x2, 0x2, 0x4002, 0x401E, 0x4002, 0x4002, 0x401E, 0x4002, 0x7F82, 0x2, 0x2, 0x2, 0x0}},
    {45377, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45378, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45379, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45380, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45381, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45382, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45383, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45384, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45385, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45386, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45387, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45388, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45389, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45390, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45391, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45392, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45393, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45394, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45395, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45396, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45397, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45398, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45399, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45400, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45401, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45402, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45403, 2, {0x0, 0x8, 0x8, 0x8, 0x4038, 0x4008, 0x4038, 0x4008, 0x7E08, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45404, 2, {0x0, 0x0, 0xA, 0xA, 0xA, 0x400A, 0x403A, 0x400A, 0x400A, 0x403A, 0x400A, 0x7F8A, 0xA, 0xA, 0xA, 0x0}},
    {45405, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45406, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45407, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45408, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45409, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45410, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45411, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45412, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45413, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45414, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45415, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45416, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45417, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45418, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45419, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45420, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45421, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45422, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45423, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45424, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45425, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45426, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45427, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45428, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45429, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45430, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45431, 2, {0x0, 0x28, 0x28, 0x28, 0x40E8, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45432, 2, {0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {45433, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45434, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45435, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45436, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45437, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45438, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45439, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45440, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45441, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45442, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45443, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45444, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45445, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45446, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45447, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45448, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45449, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45450, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45451, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45452, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45453, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45454, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45455, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45456, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45457, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45458, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45459, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45460, 2, {0x0, 0x10, 0x10, 0x4010, 0x4010, 0x4010, 0x4010, 0x401E, 0x7F90, 0x10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {45461, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45462, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45463, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45464, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45465, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45466, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45467, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45468, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45469, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45470, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45471, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45472, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45473, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45474, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45475, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45476, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45477, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45478, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45479, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45480, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45481, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45482, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45483, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45484, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45485, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45486, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45487, 2, {0x0, 0x8, 0x2008, 0x2008, 0x200E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45488, 2, {0x0, 0x12, 0x12, 0x4012, 0x4012, 0x4012, 0x4012, 0x401E, 0x7F92, 0x12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {45489, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45490, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45491, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45492, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45493, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45494, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45495, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45496, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45497, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45498, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45499, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45500, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45501, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45502, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45503, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45504, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45505, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45506, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45507, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45508, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45509, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45510, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45511, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45512, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45513, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45514, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45515, 2, {0x0, 0x28, 0x2028, 0x2028, 0x2038, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45516, 2, {0x0, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {45517, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45518, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45519, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45520, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45521, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45522, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45523, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45524, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45525, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45526, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45527, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45528, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45529, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45530, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45531, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45532, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45533, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45534, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45535, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45536, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45537, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45538, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45539, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45540, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45541, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45542, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45543, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45544, 2, {0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {45545, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45546, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45547, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45548, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45549, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45550, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45551, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45552, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45553, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45554, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45555, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45556, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45557, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45558, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45559, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45560, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45561, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45562, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45563, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45564, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45565, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45566, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45567, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45568, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45569, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45570, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45571, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45572, 2, {0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {45573, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45574, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45575, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45576, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45577, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45578, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45579, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45580, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45581, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45582, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45583, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45584, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45585, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45586, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45587, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45588, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45589, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45590, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45591, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45592, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45593, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45594, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45595, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45596, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45597, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45598, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45599, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45600, 2, {0x0, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {45601, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45602, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45603, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45604, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45605, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45606, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45607, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45608, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45609, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45610, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45611, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45612, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45613, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45614, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45615, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45616, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45617, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45618, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45619, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45620, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45621, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45622, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45623, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45624, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45625, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45626, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45627, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45628, 2, {0x0, 0xA, 0xA, 0x400A, 0x400A, 0x400A, 0x400A, 0x7F8A, 0xA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {45629, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45630, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45631, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45632, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45633, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45634, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45635, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45636, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45637, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45638, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45639, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45640, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45641, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45642, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45643, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45644, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45645, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45646, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45647, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45648, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45649, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45650, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45651, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45652, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45653, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45654, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45655, 2, {0x28, 0x2028, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45656, 2, {0x0, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {45657, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45658, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45659, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45660, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45661, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45662, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45663, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45664, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45665, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45666, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45667, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45668, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45669, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45670, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45671, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45672, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45673, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45674, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45675, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45676, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45677, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45678, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45679, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45680, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45681, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45682, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45683, 2, {0x8, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45684, 2, {0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {45685, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45686, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45687, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45688, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45689, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45690, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45691, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45692, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45693, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45694, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45695, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45696, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45697, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45698, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45699, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45700, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45701, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45702, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45703, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45704, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45705, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45706, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45707, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45708, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45709, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45710, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45711, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45712, 2, {0x0, 0x0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {45713, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45714, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45715, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45716, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45717, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45718, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45719, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {45720, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45721, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45722, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45723, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45724, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {45725, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45726, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45727, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {45728, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {45729, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {45730, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {45731, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {45732, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {45733, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {45734, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45735, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {45736, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {45737, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {45738, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {45739, 2, {0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {45740, 2, {0x0, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {45741, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45742, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45743, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45744, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45745, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45746, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45747, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45748, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45749, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45750, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45751, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45752, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45753, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45754, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45755, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45756, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45757, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45758, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45759, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45760, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45761, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45762, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45763, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45764, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45765, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45766, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45767, 2, {0x0, 0x8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45768, 2, {0x0, 0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x8, 0x0}},
    {45769, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45770, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45771, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45772, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45773, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45774, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45775, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45776, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45777, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45778, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45779, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45780, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45781, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45782, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45783, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45784, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45785, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45786, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45787, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45788, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45789, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45790, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45791, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45792, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45793, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45794, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45795, 2, {0x0, 0x8, 0x8, 0x8, 0x4008, 0x4008, 0x4008, 0x4008, 0x7E08, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45796, 2, {0x0, 0x0, 0x10, 0x10, 0x7F90, 0x4010, 0x4010, 0x401E, 0x4010, 0x4010, 0x7F90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45797, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45798, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45799, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45800, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45801, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45802, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45803, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45804, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45805, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45806, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45807, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45808, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45809, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45810, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45811, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45812, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45813, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45814, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45815, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45816, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45817, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45818, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45819, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45820, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45821, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45822, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45823, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x400E, 0x4008, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45824, 2, {0x0, 0x0, 0x12, 0x12, 0x7F92, 0x4012, 0x4012, 0x401E, 0x4012, 0x4012, 0x7F92, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {45825, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45826, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45827, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45828, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45829, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45830, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45831, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45832, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45833, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45834, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45835, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45836, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45837, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45838, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45839, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45840, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45841, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45842, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45843, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45844, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45845, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45846, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45847, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45848, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45849, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45850, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45851, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x4038, 0x4028, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45852, 2, {0x0, 0x0, 0x10, 0x10, 0x7F90, 0x4010, 0x401E, 0x4010, 0x4010, 0x401E, 0x7F90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {45853, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45854, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45855, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45856, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45857, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {45858, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {45859, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45860, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45861, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45862, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45863, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45864, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45865, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45866, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45867, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {45868, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45869, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45870, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45871, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45872, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45873, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45874, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45875, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45876, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45877, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45878, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45879, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x4008, 0x400E, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45880, 2, {0x0, 0x0, 0x12, 0x12, 0x7F92, 0x4012, 0x401E, 0x4012, 0x4012, 0x401E, 0x7F92, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {45881, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45882, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45883, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45884, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45885, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45886, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45887, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45888, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45889, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45890, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45891, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45892, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45893, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45894, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45895, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45896, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45897, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45898, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45899, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45900, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45901, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45902, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45903, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45904, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45905, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45906, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45907, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x4028, 0x4038, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45908, 2, {0x0, 0x0, 0x2, 0x2, 0x7F82, 0x4002, 0x4002, 0x401E, 0x4002, 0x4002, 0x7F82, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {45909, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45910, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45911, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45912, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45913, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45914, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45915, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45916, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45917, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45918, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45919, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45920, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45921, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45922, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45923, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45924, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45925, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45926, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45927, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45928, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45929, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45930, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45931, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45932, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45933, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45934, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45935, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4038, 0x4008, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45936, 2, {0x0, 0x0, 0xA, 0xA, 0x7F8A, 0x400A, 0x400A, 0x403A, 0x400A, 0x400A, 0x7F8A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {45937, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45938, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45939, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45940, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45941, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45942, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45943, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45944, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45945, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45946, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45947, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45948, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {45949, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45950, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45951, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45952, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45953, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45954, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {45955, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {45956, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45957, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {45958, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45959, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45960, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45961, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45962, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45963, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x40E8, 0x4028, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45964, 2, {0x0, 0x0, 0x2, 0x2, 0x7F82, 0x4002, 0x401E, 0x4002, 0x4002, 0x401E, 0x7F82, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {45965, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {45966, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45967, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {45968, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {45969, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45970, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45971, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {45972, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {45973, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {45974, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {45975, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {45976, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {45977, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {45978, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {45979, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {45980, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {45981, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {45982, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {45983, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {45984, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {45985, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {45986, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {45987, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {45988, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {45989, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {45990, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {45991, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x4008, 0x4038, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {45992, 2, {0x0, 0x0, 0xA, 0xA, 0x7F8A, 0x400A, 0x403A, 0x400A, 0x400A, 0x403A, 0x7F8A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {45993, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {45994, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {45995, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {45996, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {45997, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {45998, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {45999, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46000, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46001, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46002, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46003, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46004, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46005, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46006, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46007, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46008, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46009, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46010, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46011, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {46012, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46013, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {46014, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46015, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46016, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46017, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46018, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46019, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x4028, 0x40E8, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46020, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {46021, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46022, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46023, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46024, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46025, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46026, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46027, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46028, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46029, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46030, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46031, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46032, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46033, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46034, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46035, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46036, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46037, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46038, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46039, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46040, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46041, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46042, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46043, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46044, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46045, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46046, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46047, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46048, 2, {0x0, 0x10, 0x10, 0x7F90, 0x4010, 0x4010, 0x4010, 0x7F9E, 0x10, 0x10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {46049, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46050, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46051, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46052, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46053, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46054, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46055, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46056, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46057, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46058, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46059, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46060, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46061, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46062, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46063, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46064, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46065, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46066, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46067, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46068, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46069, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46070, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46071, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46072, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46073, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46074, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46075, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x200E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46076, 2, {0x0, 0x12, 0x12, 0x7F92, 0x4012, 0x4012, 0x4012, 0x7F9E, 0x12, 0x12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {46077, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46078, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46079, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46080, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46081, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46082, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46083, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46084, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46085, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46086, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46087, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46088, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46089, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46090, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46091, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46092, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46093, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46094, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46095, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46096, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46097, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46098, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46099, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46100, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46101, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46102, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46103, 2, {0x0, 0x28, 0x3FA8, 0x2028, 0x2038, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46104, 2, {0x0, 0x8, 0x8, 0x7F88, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {46105, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46106, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46107, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46108, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46109, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46110, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46111, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46112, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46113, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46114, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46115, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46116, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46117, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46118, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46119, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46120, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46121, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46122, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46123, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46124, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46125, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46126, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46127, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46128, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46129, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46130, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46131, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46132, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {46133, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46134, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46135, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46136, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46137, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46138, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46139, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46140, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46141, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46142, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46143, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46144, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46145, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46146, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46147, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46148, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46149, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46150, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46151, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46152, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46153, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46154, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46155, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46156, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46157, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46158, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46159, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46160, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {46161, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46162, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46163, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46164, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46165, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46166, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46167, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46168, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46169, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46170, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46171, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46172, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46173, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46174, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46175, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46176, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46177, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46178, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46179, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46180, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46181, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46182, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46183, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46184, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46185, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46186, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46187, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46188, 2, {0x0, 0x8, 0x8, 0x7FC8, 0x4008, 0x4008, 0x4008, 0x7FC8, 0x8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {46189, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46190, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46191, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46192, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46193, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46194, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46195, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46196, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46197, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46198, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46199, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46200, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46201, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46202, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46203, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46204, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46205, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46206, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46207, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46208, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46209, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46210, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46211, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46212, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46213, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46214, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46215, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46216, 2, {0x0, 0xA, 0xA, 0x7FCA, 0x400A, 0x400A, 0x400A, 0x7FCA, 0xA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {46217, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46218, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46219, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46220, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46221, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46222, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46223, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46224, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46225, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46226, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46227, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46228, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46229, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46230, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46231, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46232, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46233, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46234, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46235, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46236, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46237, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46238, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46239, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46240, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46241, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46242, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46243, 2, {0x28, 0x3FA8, 0x2028, 0x2028, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46244, 2, {0x0, 0x8, 0x8, 0x7FC8, 0x4008, 0x4008, 0x4008, 0x7FC8, 0x8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {46245, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46246, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46247, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46248, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46249, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46250, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46251, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46252, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46253, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46254, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46255, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46256, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46257, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46258, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46259, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46260, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46261, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46262, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46263, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46264, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46265, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46266, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46267, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46268, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46269, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46270, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46271, 2, {0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46272, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {46273, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46274, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46275, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46276, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46277, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46278, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46279, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46280, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46281, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46282, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46283, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46284, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46285, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46286, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46287, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46288, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46289, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46290, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46291, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46292, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46293, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46294, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46295, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46296, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46297, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46298, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46299, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46300, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {46301, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46302, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46303, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46304, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46305, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46306, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46307, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46308, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46309, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46310, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46311, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46312, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46313, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46314, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46315, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46316, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46317, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46318, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46319, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46320, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46321, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46322, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46323, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46324, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46325, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46326, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46327, 2, {0x0, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46328, 2, {0x0, 0x8, 0x8, 0x7F88, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {46329, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46330, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46331, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46332, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46333, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46334, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46335, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46336, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46337, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46338, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46339, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46340, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46341, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46342, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46343, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46344, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46345, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46346, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46347, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46348, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46349, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46350, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46351, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46352, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46353, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46354, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46355, 2, {0x0, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46356, 2, {0x0, 0x0, 0x8, 0x8, 0x7F88, 0x4008, 0x4008, 0x4008, 0x4008, 0x4008, 0x7F88, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46357, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46358, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46359, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46360, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46361, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46362, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46363, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46364, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46365, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46366, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46367, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46368, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46369, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46370, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46371, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46372, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46373, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46374, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46375, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46376, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46377, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46378, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46379, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46380, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46381, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46382, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46383, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x4008, 0x4008, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46384, 2, {0x0, 0x0, 0x10, 0x10, 0x7BD0, 0x4210, 0x4210, 0x421E, 0x4210, 0x4210, 0x7BD0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46385, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46386, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46387, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46388, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46389, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46390, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46391, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46392, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46393, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46394, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46395, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46396, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46397, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46398, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46399, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46400, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46401, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46402, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46403, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46404, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46405, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46406, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46407, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46408, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46409, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46410, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46411, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x440E, 0x4408, 0x7708, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46412, 2, {0x0, 0x0, 0x12, 0x12, 0x7BD2, 0x4212, 0x4212, 0x421E, 0x4212, 0x4212, 0x7BD2, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {46413, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46414, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46415, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46416, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46417, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46418, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46419, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46420, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46421, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46422, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46423, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46424, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46425, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46426, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46427, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46428, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46429, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46430, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46431, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {46432, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46433, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {46434, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46435, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46436, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46437, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46438, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46439, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x4438, 0x4428, 0x7728, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46440, 2, {0x0, 0x0, 0x10, 0x10, 0x7BD0, 0x4210, 0x421E, 0x4210, 0x4210, 0x421E, 0x7BD0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46441, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46442, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46443, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46444, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46445, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46446, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46447, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46448, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46449, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46450, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46451, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46452, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46453, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46454, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46455, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46456, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46457, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46458, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46459, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46460, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46461, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46462, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46463, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46464, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46465, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46466, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46467, 2, {0x0, 0x8, 0x8, 0x7708, 0x440E, 0x4408, 0x440E, 0x7708, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46468, 2, {0x0, 0x0, 0x12, 0x12, 0x7BD2, 0x4212, 0x421E, 0x4212, 0x4212, 0x421E, 0x7BD2, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {46469, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46470, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46471, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46472, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46473, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46474, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46475, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46476, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46477, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46478, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46479, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46480, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46481, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46482, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46483, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46484, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46485, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46486, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46487, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {46488, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46489, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {46490, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46491, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46492, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46493, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46494, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46495, 2, {0x0, 0x28, 0x28, 0x7728, 0x4438, 0x4428, 0x4438, 0x7728, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46496, 2, {0x0, 0x0, 0x2, 0x2, 0x7BC2, 0x4202, 0x4202, 0x421E, 0x4202, 0x4202, 0x7BC2, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {46497, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46498, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46499, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46500, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46501, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46502, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46503, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46504, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46505, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46506, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46507, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46508, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46509, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46510, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46511, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46512, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46513, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46514, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46515, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46516, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46517, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46518, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46519, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46520, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46521, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46522, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46523, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4438, 0x4408, 0x7708, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46524, 2, {0x0, 0x0, 0xA, 0xA, 0x7BCA, 0x420A, 0x420A, 0x423A, 0x420A, 0x420A, 0x7BCA, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {46525, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46526, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46527, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46528, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46529, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46530, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46531, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46532, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46533, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46534, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46535, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46536, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46537, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46538, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46539, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46540, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46541, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46542, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46543, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {46544, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46545, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {46546, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46547, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46548, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46549, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46550, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46551, 2, {0x0, 0x28, 0x28, 0x7728, 0x4428, 0x44E8, 0x4428, 0x7728, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46552, 2, {0x0, 0x0, 0x2, 0x2, 0x7BC2, 0x4202, 0x421E, 0x4202, 0x4202, 0x421E, 0x7BC2, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {46553, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46554, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46555, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46556, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46557, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46558, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46559, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46560, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46561, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46562, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46563, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46564, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46565, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46566, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46567, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46568, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46569, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46570, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46571, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46572, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46573, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46574, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46575, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46576, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46577, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46578, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46579, 2, {0x0, 0x8, 0x8, 0x7708, 0x4438, 0x4408, 0x4438, 0x7708, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46580, 2, {0x0, 0x0, 0xA, 0xA, 0x7BCA, 0x420A, 0x423A, 0x420A, 0x420A, 0x423A, 0x7BCA, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {46581, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46582, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46583, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46584, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46585, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46586, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46587, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46588, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46589, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46590, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46591, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46592, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46593, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46594, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46595, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46596, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46597, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46598, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46599, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {46600, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46601, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {46602, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46603, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46604, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46605, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46606, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46607, 2, {0x0, 0x28, 0x28, 0x7728, 0x44E8, 0x4428, 0x44E8, 0x7728, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46608, 2, {0x0, 0x0, 0x3EF8, 0x2080, 0x2080, 0x2080, 0x2080, 0x3EF8, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {46609, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46610, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46611, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46612, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46613, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46614, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46615, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46616, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46617, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46618, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46619, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46620, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46621, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46622, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46623, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46624, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46625, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46626, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46627, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46628, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46629, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46630, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46631, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46632, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46633, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46634, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46635, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46636, 2, {0x0, 0x10, 0x10, 0x7BD0, 0x4210, 0x4210, 0x4210, 0x421E, 0x7BD0, 0x10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {46637, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46638, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46639, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46640, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46641, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46642, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46643, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46644, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46645, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46646, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46647, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46648, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46649, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46650, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46651, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46652, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46653, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46654, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46655, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46656, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46657, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46658, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46659, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46660, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46661, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46662, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46663, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x220E, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46664, 2, {0x0, 0x12, 0x12, 0x7BD2, 0x4212, 0x4212, 0x4212, 0x421E, 0x7BD2, 0x12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {46665, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46666, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46667, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46668, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46669, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46670, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46671, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46672, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46673, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46674, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46675, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46676, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46677, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46678, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46679, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46680, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46681, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46682, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46683, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46684, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46685, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46686, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46687, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46688, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46689, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46690, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46691, 2, {0x0, 0x28, 0x3BA8, 0x2228, 0x2238, 0x3BA8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46692, 2, {0x0, 0x8, 0x8, 0x7BC8, 0x4208, 0x4208, 0x4208, 0x4208, 0x7BC8, 0x8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {46693, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46694, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46695, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46696, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46697, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46698, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46699, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46700, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46701, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46702, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46703, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46704, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46705, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46706, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46707, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46708, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46709, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46710, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46711, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46712, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46713, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46714, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46715, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46716, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46717, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46718, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46719, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46720, 2, {0x0, 0x0, 0x3EF8, 0x2080, 0x2080, 0x2080, 0x2080, 0x3EF8, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {46721, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46722, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46723, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46724, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46725, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46726, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46727, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46728, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46729, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46730, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46731, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46732, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46733, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46734, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46735, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46736, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46737, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46738, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46739, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46740, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46741, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46742, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46743, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46744, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46745, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46746, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46747, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46748, 2, {0x0, 0x0, 0x3EF8, 0x2080, 0x2080, 0x2080, 0x2080, 0x3EF8, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {46749, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46750, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46751, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46752, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46753, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46754, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46755, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46756, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46757, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46758, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46759, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46760, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46761, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46762, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46763, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46764, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46765, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46766, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46767, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46768, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46769, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46770, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46771, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46772, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46773, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46774, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46775, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46776, 2, {0x0, 0x8, 0x7BC8, 0x4208, 0x4208, 0x4208, 0x4208, 0x7BC8, 0x8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {46777, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46778, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46779, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46780, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46781, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46782, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46783, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46784, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46785, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46786, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46787, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46788, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46789, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46790, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46791, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46792, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46793, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46794, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46795, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46796, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46797, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46798, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46799, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46800, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46801, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46802, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46803, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46804, 2, {0x0, 0xA, 0x7BCA, 0x420A, 0x420A, 0x420A, 0x420A, 0x7BCA, 0xA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {46805, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46806, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46807, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46808, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46809, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46810, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46811, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46812, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46813, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46814, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46815, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46816, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46817, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46818, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46819, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46820, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46821, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46822, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46823, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46824, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46825, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46826, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46827, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46828, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46829, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46830, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46831, 2, {0x28, 0x7728, 0x4428, 0x4428, 0x7728, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46832, 2, {0x0, 0x8, 0x7BC8, 0x4208, 0x4208, 0x4208, 0x4208, 0x7BC8, 0x8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {46833, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46834, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46835, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46836, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46837, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46838, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46839, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46840, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46841, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46842, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46843, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46844, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46845, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46846, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46847, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46848, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46849, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46850, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46851, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46852, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46853, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46854, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46855, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46856, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46857, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46858, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46859, 2, {0x8, 0x7708, 0x4408, 0x4408, 0x7708, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46860, 2, {0x0, 0x0, 0x3EF8, 0x2080, 0x2080, 0x2080, 0x2080, 0x3EF8, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {46861, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46862, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46863, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46864, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46865, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46866, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46867, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46868, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46869, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46870, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46871, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46872, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46873, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46874, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46875, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46876, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46877, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46878, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46879, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46880, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46881, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46882, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46883, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46884, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46885, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46886, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46887, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46888, 2, {0x0, 0x0, 0x3EF8, 0x2080, 0x2080, 0x2080, 0x2080, 0x3EF8, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {46889, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {46890, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46891, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {46892, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46893, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46894, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46895, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {46896, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46897, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46898, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46899, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46900, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {46901, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46902, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46903, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {46904, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {46905, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {46906, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {46907, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {46908, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {46909, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {46910, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46911, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {46912, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {46913, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {46914, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {46915, 2, {0x0, 0x0, 0x1EF0, 0x1080, 0x1080, 0x1EF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {46916, 2, {0x0, 0x8, 0x8, 0x7BC8, 0x4208, 0x4208, 0x4208, 0x4208, 0x7BC8, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {46917, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46918, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46919, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46920, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46921, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46922, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46923, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46924, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46925, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46926, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46927, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46928, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46929, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46930, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46931, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46932, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46933, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46934, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46935, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46936, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46937, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46938, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46939, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46940, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46941, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46942, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46943, 2, {0x0, 0x8, 0x3B88, 0x2208, 0x2208, 0x3B88, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46944, 2, {0x0, 0x0, 0x8, 0x8, 0x7BC8, 0x4208, 0x4208, 0x4208, 0x4208, 0x4208, 0x7BC8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46945, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {46946, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46947, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46948, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {46949, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {46950, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {46951, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46952, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46953, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46954, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46955, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46956, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {46957, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46958, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46959, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {46960, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46961, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46962, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {46963, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46964, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46965, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46966, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46967, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46968, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46969, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46970, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46971, 2, {0x0, 0x8, 0x8, 0x7708, 0x4408, 0x4408, 0x4408, 0x7708, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {46972, 2, {0x0, 0x0, 0x10, 0x10, 0x3F90, 0x90, 0x90, 0x3F9E, 0x2010, 0x2010, 0x2010, 0x3F90, 0x10, 0x10, 0x10, 0x0}},
    {46973, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {46974, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {46975, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {46976, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {46977, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {46978, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {46979, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {46980, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {46981, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {46982, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {46983, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {46984, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {46985, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {46986, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {46987, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {46988, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {46989, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {46990, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {46991, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {46992, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {46993, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {46994, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {46995, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {46996, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {46997, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {46998, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {46999, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F0E, 0x2008, 0x3F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47000, 2, {0x0, 0x0, 0x12, 0x12, 0x3F92, 0x92, 0x92, 0x3F9E, 0x2012, 0x2012, 0x2012, 0x3F92, 0x12, 0x12, 0x12, 0x0}},
    {47001, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47002, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47003, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47004, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47005, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47006, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47007, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47008, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47009, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47010, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47011, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47012, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47013, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47014, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47015, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47016, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47017, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47018, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47019, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47020, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47021, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47022, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47023, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47024, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47025, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47026, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47027, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3F38, 0x2028, 0x3F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47028, 2, {0x0, 0x0, 0x10, 0x10, 0x3F90, 0x90, 0x9E, 0x3F90, 0x2010, 0x201E, 0x2010, 0x3F90, 0x10, 0x10, 0x10, 0x0}},
    {47029, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47030, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47031, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47032, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47033, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47034, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47035, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47036, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47037, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47038, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47039, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47040, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47041, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47042, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47043, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47044, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47045, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47046, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47047, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47048, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47049, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47050, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47051, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47052, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47053, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47054, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47055, 2, {0x0, 0x8, 0x8, 0x3F08, 0x10E, 0x3F08, 0x200E, 0x3F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47056, 2, {0x0, 0x0, 0x12, 0x12, 0x3F92, 0x92, 0x9E, 0x3F92, 0x2012, 0x201E, 0x2012, 0x3F92, 0x12, 0x12, 0x12, 0x0}},
    {47057, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47058, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47059, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47060, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47061, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47062, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47063, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47064, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47065, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47066, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47067, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47068, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47069, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47070, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47071, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47072, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47073, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47074, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47075, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47076, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47077, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47078, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47079, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47080, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47081, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47082, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47083, 2, {0x0, 0x28, 0x28, 0x3F28, 0x138, 0x3F28, 0x2038, 0x3F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47084, 2, {0x0, 0x0, 0x2, 0x2, 0x3F82, 0x82, 0x82, 0x3F9E, 0x2002, 0x2002, 0x2002, 0x3F82, 0x2, 0x2, 0x2, 0x0}},
    {47085, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47086, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47087, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47088, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47089, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47090, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47091, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47092, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47093, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47094, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47095, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47096, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47097, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47098, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47099, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47100, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47101, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47102, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47103, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47104, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47105, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47106, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47107, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47108, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47109, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47110, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47111, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F38, 0x2008, 0x3F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47112, 2, {0x0, 0x0, 0xA, 0xA, 0x3F8A, 0x8A, 0x8A, 0x3FBA, 0x200A, 0x200A, 0x200A, 0x3F8A, 0xA, 0xA, 0xA, 0x0}},
    {47113, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47114, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47115, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47116, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47117, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47118, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47119, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47120, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47121, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47122, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47123, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47124, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47125, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47126, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47127, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47128, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47129, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47130, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47131, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47132, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47133, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47134, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47135, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47136, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47137, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47138, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47139, 2, {0x0, 0x28, 0x28, 0x3F28, 0x128, 0x3FE8, 0x2028, 0x3F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47140, 2, {0x0, 0x0, 0x2, 0x2, 0x3F82, 0x82, 0x9E, 0x3F82, 0x2002, 0x201E, 0x2002, 0x3F82, 0x2, 0x2, 0x2, 0x0}},
    {47141, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47142, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47143, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47144, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47145, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47146, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47147, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47148, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47149, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47150, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47151, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47152, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47153, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47154, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47155, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47156, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47157, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47158, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47159, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47160, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47161, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47162, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47163, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47164, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47165, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47166, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47167, 2, {0x0, 0x8, 0x8, 0x3F08, 0x138, 0x3F08, 0x2038, 0x3F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47168, 2, {0x0, 0x0, 0xA, 0xA, 0x3F8A, 0x8A, 0xBA, 0x3F8A, 0x200A, 0x203A, 0x200A, 0x3F8A, 0xA, 0xA, 0xA, 0x0}},
    {47169, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47170, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47171, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47172, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47173, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47174, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47175, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47176, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47177, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47178, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47179, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47180, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47181, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47182, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47183, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47184, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47185, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47186, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47187, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47188, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47189, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47190, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47191, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47192, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47193, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47194, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47195, 2, {0x0, 0x28, 0x28, 0x3F28, 0x1E8, 0x3F28, 0x20E8, 0x3F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47196, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {47197, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47198, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47199, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47200, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47201, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47202, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47203, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47204, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47205, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47206, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47207, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47208, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47209, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47210, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47211, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47212, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47213, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47214, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47215, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47216, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47217, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47218, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47219, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47220, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47221, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47222, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47223, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47224, 2, {0x0, 0x10, 0x10, 0x3F90, 0x90, 0x90, 0x3F90, 0x201E, 0x2010, 0x3F90, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {47225, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47226, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47227, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47228, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47229, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47230, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47231, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47232, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47233, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47234, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47235, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47236, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47237, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47238, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47239, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47240, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47241, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47242, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47243, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47244, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47245, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47246, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47247, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47248, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47249, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47250, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47251, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47252, 2, {0x0, 0x12, 0x12, 0x3F92, 0x92, 0x92, 0x3F92, 0x201E, 0x2012, 0x3F92, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {47253, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47254, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47255, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47256, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47257, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47258, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47259, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47260, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47261, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47262, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47263, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47264, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47265, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47266, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47267, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47268, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47269, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47270, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47271, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47272, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47273, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47274, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47275, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47276, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47277, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47278, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47279, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0x2028, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47280, 2, {0x0, 0x8, 0x8, 0x3F88, 0x88, 0x88, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {47281, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47282, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47283, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47284, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47285, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47286, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47287, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47288, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47289, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47290, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47291, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47292, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47293, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47294, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47295, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47296, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47297, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47298, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47299, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47300, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47301, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47302, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47303, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47304, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47305, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47306, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47307, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47308, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {47309, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47310, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47311, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47312, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47313, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47314, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47315, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47316, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47317, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47318, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47319, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47320, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47321, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47322, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47323, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47324, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47325, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47326, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47327, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47328, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47329, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47330, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47331, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47332, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47333, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47334, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47335, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47336, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {47337, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47338, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47339, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47340, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47341, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47342, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47343, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47344, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47345, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47346, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47347, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47348, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47349, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47350, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47351, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47352, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47353, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47354, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47355, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47356, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47357, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47358, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47359, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47360, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47361, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47362, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47363, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47364, 2, {0x0, 0x8, 0x3FC8, 0x48, 0x48, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {47365, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47366, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47367, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47368, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47369, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47370, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47371, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47372, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47373, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47374, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47375, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47376, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47377, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47378, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47379, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47380, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47381, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47382, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47383, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47384, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47385, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47386, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47387, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47388, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47389, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47390, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47391, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47392, 2, {0x0, 0xA, 0x3FCA, 0x4A, 0x4A, 0x3FCA, 0x200A, 0x200A, 0x3FCA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {47393, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47394, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47395, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47396, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47397, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47398, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47399, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47400, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47401, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47402, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47403, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47404, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47405, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47406, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47407, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47408, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47409, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47410, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47411, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47412, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47413, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47414, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47415, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47416, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47417, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47418, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47419, 2, {0x7FA8, 0xA8, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47420, 2, {0x0, 0x8, 0x3FC8, 0x48, 0x48, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {47421, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47422, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47423, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47424, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47425, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47426, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47427, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47428, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47429, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47430, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47431, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47432, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47433, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47434, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47435, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47436, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47437, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47438, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47439, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47440, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47441, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47442, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47443, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47444, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47445, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47446, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47447, 2, {0x7F88, 0x88, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47448, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {47449, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47450, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47451, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47452, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47453, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47454, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47455, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47456, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47457, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47458, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47459, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47460, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47461, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47462, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47463, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47464, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47465, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47466, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47467, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47468, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47469, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47470, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47471, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47472, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47473, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47474, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47475, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47476, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x10, 0x1FF0, 0x1000, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {47477, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47478, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47479, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47480, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47481, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47482, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47483, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47484, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47485, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47486, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47487, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47488, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47489, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47490, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47491, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47492, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47493, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47494, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47495, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47496, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47497, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47498, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47499, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47500, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47501, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47502, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47503, 2, {0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47504, 2, {0x0, 0x8, 0x8, 0x3F88, 0x88, 0x88, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {47505, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47506, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47507, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47508, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47509, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47510, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47511, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47512, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47513, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47514, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47515, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47516, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47517, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47518, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47519, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47520, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47521, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47522, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47523, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47524, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47525, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47526, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47527, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47528, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47529, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47530, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47531, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x2008, 0x3F88, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47532, 2, {0x0, 0x0, 0x8, 0x8, 0x3F88, 0x88, 0x88, 0x3F88, 0x2008, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x8, 0x0}},
    {47533, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47534, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47535, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47536, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47537, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47538, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47539, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47540, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47541, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47542, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47543, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47544, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47545, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47546, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47547, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47548, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47549, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47550, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47551, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47552, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47553, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47554, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47555, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47556, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47557, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47558, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47559, 2, {0x0, 0x8, 0x8, 0x3F08, 0x108, 0x3F08, 0x2008, 0x3F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47560, 2, {0x0, 0x0, 0x10, 0x10, 0x7F90, 0x4090, 0x4090, 0x409E, 0x4090, 0x4090, 0x4090, 0x7F90, 0x10, 0x10, 0x10, 0x0}},
    {47561, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47562, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47563, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47564, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47565, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47566, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47567, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47568, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47569, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47570, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47571, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47572, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47573, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47574, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47575, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47576, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47577, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47578, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47579, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47580, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47581, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47582, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47583, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47584, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47585, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47586, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47587, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x420E, 0x4208, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47588, 2, {0x0, 0x0, 0x12, 0x12, 0x7F92, 0x4092, 0x4092, 0x409E, 0x4092, 0x4092, 0x4092, 0x7F92, 0x12, 0x12, 0x12, 0x0}},
    {47589, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47590, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47591, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47592, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47593, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47594, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47595, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47596, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47597, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47598, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47599, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47600, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47601, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47602, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47603, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47604, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47605, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47606, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47607, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47608, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47609, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47610, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47611, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47612, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47613, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47614, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47615, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x4238, 0x4228, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47616, 2, {0x0, 0x0, 0x10, 0x10, 0x7F90, 0x4090, 0x409E, 0x4090, 0x4090, 0x409E, 0x4090, 0x7F90, 0x10, 0x10, 0x10, 0x0}},
    {47617, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47618, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47619, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47620, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47621, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47622, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47623, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47624, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47625, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47626, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47627, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47628, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47629, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47630, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47631, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47632, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47633, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47634, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47635, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47636, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47637, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47638, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47639, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47640, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47641, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47642, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47643, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x420E, 0x4208, 0x420E, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47644, 2, {0x0, 0x0, 0x12, 0x12, 0x7F92, 0x4092, 0x409E, 0x4092, 0x4092, 0x409E, 0x4092, 0x7F92, 0x12, 0x12, 0x12, 0x0}},
    {47645, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47646, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47647, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47648, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47649, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47650, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47651, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47652, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47653, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47654, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47655, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47656, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47657, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47658, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47659, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47660, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47661, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47662, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47663, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47664, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47665, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47666, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47667, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47668, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47669, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47670, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47671, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4238, 0x4228, 0x4238, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47672, 2, {0x0, 0x0, 0x2, 0x2, 0x7F82, 0x4082, 0x4082, 0x409E, 0x4082, 0x4082, 0x4082, 0x7F82, 0x2, 0x2, 0x2, 0x0}},
    {47673, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47674, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47675, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47676, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47677, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47678, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47679, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47680, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47681, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47682, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47683, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47684, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47685, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47686, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47687, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47688, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47689, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47690, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47691, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47692, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47693, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47694, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47695, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47696, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47697, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47698, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47699, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4238, 0x4208, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47700, 2, {0x0, 0x0, 0xA, 0xA, 0x7F8A, 0x408A, 0x408A, 0x40BA, 0x408A, 0x408A, 0x408A, 0x7F8A, 0xA, 0xA, 0xA, 0x0}},
    {47701, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47702, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47703, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47704, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47705, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47706, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47707, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47708, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47709, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47710, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47711, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47712, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47713, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47714, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47715, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47716, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47717, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47718, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47719, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47720, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47721, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47722, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47723, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47724, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47725, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47726, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47727, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x4228, 0x42E8, 0x4228, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47728, 2, {0x0, 0x0, 0x2, 0x2, 0x7F82, 0x4082, 0x409E, 0x4082, 0x4082, 0x409E, 0x4082, 0x7F82, 0x2, 0x2, 0x2, 0x0}},
    {47729, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47730, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47731, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47732, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47733, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47734, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47735, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47736, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47737, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47738, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47739, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47740, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47741, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47742, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47743, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47744, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47745, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47746, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47747, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47748, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47749, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47750, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47751, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47752, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47753, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47754, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47755, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4238, 0x4208, 0x4238, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47756, 2, {0x0, 0x0, 0xA, 0xA, 0x7F8A, 0x408A, 0x40BA, 0x408A, 0x408A, 0x40BA, 0x408A, 0x7F8A, 0xA, 0xA, 0xA, 0x0}},
    {47757, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47758, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47759, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47760, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47761, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47762, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47763, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47764, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47765, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47766, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47767, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47768, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47769, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47770, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47771, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47772, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47773, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47774, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47775, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {47776, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47777, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {47778, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47779, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47780, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47781, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47782, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47783, 2, {0x0, 0x28, 0x7E28, 0x4228, 0x42E8, 0x4228, 0x42E8, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47784, 2, {0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {47785, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47786, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47787, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47788, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47789, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47790, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47791, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47792, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47793, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47794, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47795, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47796, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47797, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47798, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47799, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47800, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47801, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47802, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47803, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47804, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47805, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47806, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47807, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47808, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47809, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47810, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47811, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47812, 2, {0x0, 0x10, 0x10, 0x3F90, 0x2090, 0x2090, 0x2090, 0x209E, 0x2090, 0x3F90, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {47813, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47814, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47815, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47816, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47817, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47818, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47819, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47820, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47821, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47822, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47823, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47824, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47825, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47826, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47827, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47828, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47829, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47830, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47831, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47832, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47833, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47834, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47835, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47836, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47837, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47838, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47839, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x208E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47840, 2, {0x0, 0x12, 0x12, 0x3F92, 0x2092, 0x2092, 0x2092, 0x209E, 0x2092, 0x3F92, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {47841, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47842, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47843, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47844, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47845, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47846, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47847, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47848, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47849, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47850, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47851, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47852, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47853, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47854, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47855, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47856, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47857, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47858, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47859, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47860, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47861, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47862, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47863, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47864, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47865, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47866, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47867, 2, {0x0, 0x28, 0x3FA8, 0x20A8, 0x20B8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47868, 2, {0x0, 0x8, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {47869, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {47870, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47871, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47872, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {47873, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {47874, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {47875, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47876, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47877, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47878, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47879, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47880, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {47881, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47882, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47883, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {47884, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47885, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47886, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {47887, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47888, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47889, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47890, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47891, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47892, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47893, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47894, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47895, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47896, 2, {0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {47897, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47898, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47899, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47900, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47901, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47902, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47903, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47904, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47905, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47906, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47907, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47908, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47909, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47910, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47911, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47912, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47913, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47914, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47915, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47916, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47917, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47918, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47919, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47920, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47921, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47922, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47923, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47924, 2, {0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {47925, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {47926, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47927, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {47928, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47929, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47930, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47931, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {47932, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47933, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47934, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47935, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47936, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {47937, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47938, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47939, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {47940, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {47941, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {47942, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {47943, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {47944, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {47945, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {47946, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47947, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {47948, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {47949, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {47950, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {47951, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {47952, 2, {0x0, 0x8, 0x3FC8, 0x2048, 0x2048, 0x2048, 0x2048, 0x2048, 0x3FC8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {47953, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47954, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47955, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47956, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47957, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47958, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47959, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47960, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47961, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47962, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47963, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47964, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47965, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47966, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47967, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47968, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47969, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47970, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47971, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {47972, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {47973, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {47974, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {47975, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {47976, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {47977, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {47978, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {47979, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {47980, 2, {0x0, 0xA, 0x3FCA, 0x204A, 0x204A, 0x204A, 0x204A, 0x204A, 0x3FCA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {47981, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {47982, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {47983, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {47984, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {47985, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {47986, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {47987, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {47988, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {47989, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {47990, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {47991, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {47992, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {47993, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {47994, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {47995, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {47996, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {47997, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {47998, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {47999, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48000, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48001, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48002, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48003, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48004, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48005, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48006, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48007, 2, {0x28, 0x3FA8, 0x20A8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48008, 2, {0x0, 0x8, 0x3FC8, 0x2048, 0x2048, 0x2048, 0x2048, 0x2048, 0x3FC8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {48009, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48010, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48011, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48012, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48013, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48014, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48015, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48016, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48017, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48018, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48019, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48020, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48021, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48022, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48023, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48024, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48025, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48026, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48027, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48028, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48029, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48030, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48031, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48032, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48033, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48034, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48035, 2, {0x8, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48036, 2, {0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {48037, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48038, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48039, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48040, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48041, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48042, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48043, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48044, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48045, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48046, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48047, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48048, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48049, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48050, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48051, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48052, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48053, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48054, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48055, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48056, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48057, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48058, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48059, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48060, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48061, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48062, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48063, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48064, 2, {0x0, 0x0, 0x3FF8, 0x2008, 0x2008, 0x2008, 0x2008, 0x3FF8, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {48065, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48066, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48067, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48068, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48069, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48070, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48071, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48072, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48073, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48074, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48075, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48076, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48077, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48078, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48079, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48080, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48081, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48082, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48083, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48084, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48085, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48086, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48087, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48088, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48089, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48090, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48091, 2, {0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48092, 2, {0x0, 0x8, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {48093, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48094, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48095, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48096, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48097, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48098, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48099, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48100, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48101, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48102, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48103, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48104, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48105, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48106, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48107, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48108, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48109, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48110, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48111, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48112, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48113, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48114, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48115, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48116, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48117, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48118, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48119, 2, {0x0, 0x8, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48120, 2, {0x0, 0x0, 0x8, 0x8, 0x7F88, 0x4088, 0x4088, 0x4088, 0x4088, 0x4088, 0x4088, 0x7F88, 0x8, 0x8, 0x8, 0x0}},
    {48121, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48122, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48123, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48124, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48125, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48126, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48127, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48128, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48129, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48130, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48131, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48132, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48133, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48134, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48135, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48136, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48137, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48138, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48139, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48140, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48141, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48142, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48143, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48144, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48145, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48146, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48147, 2, {0x0, 0x8, 0x7E08, 0x4208, 0x4208, 0x4208, 0x4208, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48148, 2, {0x0, 0x0, 0x10, 0x10, 0x2090, 0x2090, 0x2090, 0x3F9E, 0x2090, 0x2090, 0x2090, 0x3F90, 0x10, 0x10, 0x10, 0x0}},
    {48149, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48150, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48151, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48152, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48153, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48154, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48155, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48156, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48157, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48158, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48159, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48160, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48161, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48162, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48163, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48164, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48165, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48166, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48167, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48168, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48169, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48170, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48171, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48172, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48173, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48174, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48175, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E0E, 0x4208, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48176, 2, {0x0, 0x0, 0x12, 0x12, 0x2092, 0x2092, 0x2092, 0x3F9E, 0x2092, 0x2092, 0x2092, 0x3F92, 0x12, 0x12, 0x12, 0x0}},
    {48177, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48178, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48179, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48180, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48181, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48182, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48183, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48184, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48185, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48186, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48187, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48188, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48189, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48190, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48191, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48192, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48193, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48194, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48195, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48196, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48197, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48198, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48199, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48200, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48201, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48202, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48203, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7E38, 0x4228, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48204, 2, {0x0, 0x0, 0x10, 0x10, 0x2090, 0x2090, 0x209E, 0x3F90, 0x2090, 0x209E, 0x2090, 0x3F90, 0x10, 0x10, 0x10, 0x0}},
    {48205, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48206, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48207, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48208, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48209, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48210, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48211, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48212, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48213, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48214, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48215, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48216, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48217, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48218, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48219, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48220, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48221, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48222, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48223, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48224, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48225, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48226, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48227, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48228, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48229, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48230, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48231, 2, {0x0, 0x8, 0x8, 0x4208, 0x420E, 0x7E08, 0x420E, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48232, 2, {0x0, 0x0, 0x12, 0x12, 0x2092, 0x2092, 0x209E, 0x3F92, 0x2092, 0x209E, 0x2092, 0x3F92, 0x12, 0x12, 0x12, 0x0}},
    {48233, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48234, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48235, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48236, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48237, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48238, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48239, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48240, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48241, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48242, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48243, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48244, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48245, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48246, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48247, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48248, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48249, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48250, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48251, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48252, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48253, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48254, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48255, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48256, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48257, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48258, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48259, 2, {0x0, 0x28, 0x28, 0x4228, 0x4238, 0x7E28, 0x4238, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48260, 2, {0x0, 0x0, 0x2, 0x2, 0x2082, 0x2082, 0x2082, 0x3F9E, 0x2082, 0x2082, 0x2082, 0x3F82, 0x2, 0x2, 0x2, 0x0}},
    {48261, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48262, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48263, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48264, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48265, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48266, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48267, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48268, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48269, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48270, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48271, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48272, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48273, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48274, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48275, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48276, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48277, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48278, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48279, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48280, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48281, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48282, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48283, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48284, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48285, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48286, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48287, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E38, 0x4208, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48288, 2, {0x0, 0x0, 0xA, 0xA, 0x208A, 0x208A, 0x208A, 0x3FBA, 0x208A, 0x208A, 0x208A, 0x3F8A, 0xA, 0xA, 0xA, 0x0}},
    {48289, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48290, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48291, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48292, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48293, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48294, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48295, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48296, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48297, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48298, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48299, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48300, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48301, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48302, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48303, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48304, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48305, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48306, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48307, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48308, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48309, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48310, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48311, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48312, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48313, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48314, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48315, 2, {0x0, 0x28, 0x28, 0x4228, 0x4228, 0x7EE8, 0x4228, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48316, 2, {0x0, 0x0, 0x2, 0x2, 0x2082, 0x2082, 0x209E, 0x3F82, 0x2082, 0x209E, 0x2082, 0x3F82, 0x2, 0x2, 0x2, 0x0}},
    {48317, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48318, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48319, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48320, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48321, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48322, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48323, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48324, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48325, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48326, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48327, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48328, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48329, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48330, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48331, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48332, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48333, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48334, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48335, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48336, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48337, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48338, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48339, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48340, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48341, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48342, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48343, 2, {0x0, 0x8, 0x8, 0x4208, 0x4238, 0x7E08, 0x4238, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48344, 2, {0x0, 0x0, 0xA, 0xA, 0x208A, 0x208A, 0x20BA, 0x3F8A, 0x208A, 0x20BA, 0x208A, 0x3F8A, 0xA, 0xA, 0xA, 0x0}},
    {48345, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48346, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48347, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48348, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48349, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48350, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48351, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48352, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48353, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48354, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48355, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48356, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48357, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48358, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48359, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48360, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48361, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48362, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48363, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48364, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48365, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48366, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48367, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48368, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48369, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48370, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48371, 2, {0x0, 0x28, 0x28, 0x4228, 0x42E8, 0x7E28, 0x42E8, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48372, 2, {0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {48373, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48374, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48375, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48376, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48377, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48378, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48379, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48380, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48381, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48382, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48383, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48384, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48385, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48386, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48387, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48388, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48389, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48390, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48391, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48392, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48393, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48394, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48395, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48396, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48397, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48398, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48399, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48400, 2, {0x0, 0x10, 0x10, 0x2090, 0x2090, 0x2090, 0x3F90, 0x209E, 0x2090, 0x3F90, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {48401, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48402, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48403, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48404, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48405, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48406, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48407, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48408, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48409, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48410, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48411, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48412, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48413, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48414, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48415, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48416, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48417, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48418, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48419, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48420, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48421, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48422, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48423, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48424, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48425, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48426, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48427, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F8E, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48428, 2, {0x0, 0x12, 0x12, 0x2092, 0x2092, 0x2092, 0x3F92, 0x209E, 0x2092, 0x3F92, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {48429, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48430, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48431, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48432, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48433, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48434, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48435, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48436, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48437, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48438, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48439, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48440, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48441, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48442, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48443, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48444, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48445, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48446, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48447, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48448, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48449, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48450, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48451, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48452, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48453, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48454, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48455, 2, {0x0, 0x28, 0x20A8, 0x20A8, 0x3FB8, 0x20A8, 0x3FA8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48456, 2, {0x0, 0x8, 0x8, 0x2088, 0x2088, 0x2088, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {48457, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48458, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48459, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48460, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48461, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48462, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48463, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48464, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48465, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48466, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48467, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48468, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48469, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48470, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48471, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48472, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48473, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48474, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48475, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48476, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48477, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48478, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48479, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48480, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48481, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48482, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48483, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48484, 2, {0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {48485, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48486, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48487, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48488, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48489, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48490, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48491, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48492, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48493, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48494, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48495, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48496, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48497, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48498, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48499, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48500, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48501, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48502, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48503, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48504, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48505, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48506, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48507, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48508, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48509, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48510, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48511, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48512, 2, {0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {48513, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48514, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48515, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48516, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48517, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48518, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48519, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48520, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48521, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48522, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48523, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48524, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48525, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48526, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48527, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48528, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48529, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48530, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48531, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48532, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48533, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48534, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48535, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48536, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48537, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48538, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48539, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48540, 2, {0x0, 0x8, 0x2088, 0x2088, 0x2088, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {48541, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48542, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48543, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48544, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48545, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48546, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48547, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48548, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48549, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48550, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48551, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48552, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48553, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48554, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48555, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48556, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48557, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48558, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48559, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48560, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48561, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48562, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48563, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48564, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48565, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48566, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48567, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48568, 2, {0x0, 0xA, 0x208A, 0x208A, 0x208A, 0x3F8A, 0x208A, 0x208A, 0x3F8A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {48569, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48570, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48571, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48572, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48573, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48574, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48575, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48576, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48577, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48578, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48579, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48580, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48581, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48582, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48583, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48584, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48585, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48586, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48587, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48588, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48589, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48590, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48591, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48592, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48593, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48594, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48595, 2, {0x20A8, 0x20A8, 0x3FA8, 0x20A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48596, 2, {0x0, 0x8, 0x2088, 0x2088, 0x2088, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {48597, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48598, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48599, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48600, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48601, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48602, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48603, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48604, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48605, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48606, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48607, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48608, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48609, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48610, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48611, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48612, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48613, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48614, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48615, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48616, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48617, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48618, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48619, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48620, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48621, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48622, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48623, 2, {0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48624, 2, {0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {48625, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48626, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48627, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48628, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48629, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48630, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48631, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48632, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48633, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48634, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48635, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48636, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48637, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48638, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48639, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48640, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48641, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48642, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48643, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48644, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48645, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48646, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48647, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48648, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48649, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48650, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48651, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48652, 2, {0x0, 0x0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {48653, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48654, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48655, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48656, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48657, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48658, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48659, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48660, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48661, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48662, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48663, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48664, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48665, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48666, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48667, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48668, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48669, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48670, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48671, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48672, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48673, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48674, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48675, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48676, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48677, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48678, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48679, 2, {0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48680, 2, {0x0, 0x8, 0x8, 0x2088, 0x2088, 0x2088, 0x3F88, 0x2088, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {48681, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48682, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48683, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48684, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48685, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48686, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48687, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48688, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48689, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48690, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48691, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48692, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48693, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48694, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48695, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48696, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48697, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48698, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48699, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48700, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48701, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48702, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48703, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48704, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48705, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48706, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48707, 2, {0x0, 0x8, 0x2088, 0x2088, 0x3F88, 0x2088, 0x3F88, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48708, 2, {0x0, 0x0, 0x8, 0x8, 0x2088, 0x2088, 0x2088, 0x3F88, 0x2088, 0x2088, 0x2088, 0x3F88, 0x8, 0x8, 0x8, 0x0}},
    {48709, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48710, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48711, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48712, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48713, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48714, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48715, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48716, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48717, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48718, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48719, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48720, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48721, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48722, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48723, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48724, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48725, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48726, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48727, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48728, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48729, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48730, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48731, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48732, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48733, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48734, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48735, 2, {0x0, 0x8, 0x8, 0x4208, 0x4208, 0x7E08, 0x4208, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48736, 2, {0x0, 0x0, 0x10, 0x10, 0x2490, 0x2490, 0x2490, 0x3F9E, 0x2490, 0x2490, 0x3F90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48737, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48738, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48739, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48740, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48741, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48742, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48743, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48744, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48745, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48746, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48747, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48748, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48749, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48750, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48751, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48752, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48753, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48754, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48755, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48756, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48757, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48758, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48759, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48760, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48761, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48762, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48763, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F0E, 0x4908, 0x7F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48764, 2, {0x0, 0x0, 0x12, 0x12, 0x2492, 0x2492, 0x2492, 0x3F9E, 0x2492, 0x2492, 0x3F92, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {48765, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48766, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48767, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48768, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48769, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48770, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48771, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48772, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48773, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48774, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48775, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48776, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48777, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48778, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48779, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48780, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48781, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48782, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48783, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48784, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48785, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48786, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48787, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48788, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48789, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48790, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48791, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7F38, 0x4928, 0x7F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48792, 2, {0x0, 0x0, 0x10, 0x10, 0x2490, 0x2490, 0x249E, 0x3F90, 0x2490, 0x249E, 0x3F90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48793, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48794, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48795, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48796, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48797, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48798, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48799, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48800, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48801, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48802, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48803, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48804, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48805, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48806, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48807, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {48808, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48809, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48810, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48811, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48812, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48813, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48814, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48815, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48816, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48817, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48818, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48819, 2, {0x0, 0x8, 0x4908, 0x4908, 0x490E, 0x7F08, 0x490E, 0x7F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48820, 2, {0x0, 0x0, 0x12, 0x12, 0x2492, 0x2492, 0x249E, 0x3F92, 0x2492, 0x249E, 0x3F92, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {48821, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48822, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48823, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48824, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48825, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48826, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48827, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48828, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48829, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48830, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48831, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48832, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48833, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48834, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48835, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48836, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48837, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48838, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48839, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48840, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48841, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48842, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48843, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48844, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48845, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48846, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48847, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4938, 0x7F28, 0x4938, 0x7F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48848, 2, {0x0, 0x0, 0x2, 0x2, 0x2482, 0x2482, 0x2482, 0x3F9E, 0x2482, 0x2482, 0x3F82, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {48849, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48850, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48851, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48852, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48853, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48854, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48855, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48856, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48857, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48858, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48859, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48860, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48861, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48862, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48863, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48864, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48865, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48866, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48867, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48868, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48869, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48870, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48871, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48872, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48873, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48874, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48875, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F38, 0x4908, 0x7F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48876, 2, {0x0, 0x0, 0xA, 0xA, 0x248A, 0x248A, 0x248A, 0x3FBA, 0x248A, 0x248A, 0x3F8A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {48877, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48878, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48879, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48880, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48881, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48882, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48883, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48884, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48885, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48886, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48887, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48888, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48889, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48890, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48891, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48892, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48893, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48894, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48895, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48896, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48897, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48898, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48899, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48900, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48901, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48902, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48903, 2, {0x0, 0x28, 0x4928, 0x4928, 0x4928, 0x7FE8, 0x4928, 0x7F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48904, 2, {0x0, 0x0, 0x2, 0x2, 0x2482, 0x2482, 0x249E, 0x3F82, 0x2482, 0x249E, 0x3F82, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {48905, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {48906, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48907, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48908, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {48909, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48910, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48911, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48912, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48913, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48914, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48915, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48916, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {48917, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48918, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48919, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48920, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48921, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48922, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {48923, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {48924, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48925, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {48926, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48927, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48928, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48929, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48930, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48931, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4938, 0x7F08, 0x4938, 0x7F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48932, 2, {0x0, 0x0, 0xA, 0xA, 0x248A, 0x248A, 0x24BA, 0x3F8A, 0x248A, 0x24BA, 0x3F8A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {48933, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48934, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48935, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48936, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48937, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48938, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48939, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48940, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48941, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48942, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48943, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48944, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {48945, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48946, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48947, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {48948, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {48949, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {48950, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {48951, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {48952, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {48953, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {48954, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {48955, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {48956, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {48957, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {48958, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {48959, 2, {0x0, 0x28, 0x4928, 0x4928, 0x49E8, 0x7F28, 0x49E8, 0x7F28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {48960, 2, {0x0, 0x0, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {48961, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {48962, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48963, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {48964, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48965, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {48966, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {48967, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {48968, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48969, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48970, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48971, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {48972, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {48973, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {48974, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {48975, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {48976, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {48977, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {48978, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {48979, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {48980, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {48981, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {48982, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48983, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {48984, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {48985, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {48986, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {48987, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {48988, 2, {0x0, 0x10, 0x10, 0x4450, 0x4450, 0x4450, 0x7FD0, 0x445E, 0x4450, 0x7FD0, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {48989, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {48990, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {48991, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {48992, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {48993, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {48994, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {48995, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {48996, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {48997, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {48998, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {48999, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49000, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49001, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49002, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49003, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49004, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49005, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49006, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49007, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49008, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49009, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49010, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49011, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49012, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49013, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49014, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49015, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x248E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49016, 2, {0x0, 0x12, 0x12, 0x4452, 0x4452, 0x4452, 0x7FD2, 0x445E, 0x4452, 0x7FD2, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {49017, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49018, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49019, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49020, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49021, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49022, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49023, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49024, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49025, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49026, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49027, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49028, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49029, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49030, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49031, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49032, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49033, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49034, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49035, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49036, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49037, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49038, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49039, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49040, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49041, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49042, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49043, 2, {0x0, 0x24A8, 0x24A8, 0x3FA8, 0x24B8, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49044, 2, {0x0, 0x8, 0x8, 0x4448, 0x4448, 0x4448, 0x7FC8, 0x4448, 0x4448, 0x7FC8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {49045, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49046, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49047, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49048, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49049, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49050, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49051, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49052, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49053, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49054, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49055, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49056, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49057, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49058, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49059, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49060, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49061, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49062, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49063, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49064, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49065, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49066, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49067, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49068, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49069, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49070, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49071, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49072, 2, {0x0, 0x0, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {49073, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49074, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49075, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49076, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49077, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49078, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49079, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49080, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49081, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49082, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49083, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49084, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49085, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49086, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49087, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49088, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49089, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49090, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49091, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49092, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49093, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49094, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49095, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49096, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49097, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49098, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49099, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49100, 2, {0x0, 0x0, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {49101, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49102, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49103, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49104, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49105, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49106, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49107, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49108, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49109, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49110, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49111, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49112, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49113, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49114, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49115, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49116, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49117, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49118, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49119, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49120, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49121, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49122, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49123, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49124, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49125, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49126, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49127, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49128, 2, {0x0, 0x8, 0x4448, 0x4448, 0x4448, 0x7FC8, 0x4448, 0x4448, 0x7FC8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {49129, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49130, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49131, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49132, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49133, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49134, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49135, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49136, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49137, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49138, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49139, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49140, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49141, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49142, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49143, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49144, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49145, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49146, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49147, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49148, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49149, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49150, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49151, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49152, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49153, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49154, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49155, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49156, 2, {0x0, 0xA, 0x444A, 0x444A, 0x444A, 0x7FCA, 0x444A, 0x444A, 0x7FCA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {49157, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49158, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49159, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49160, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49161, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49162, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49163, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49164, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49165, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49166, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49167, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49168, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49169, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49170, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49171, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49172, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49173, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49174, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49175, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49176, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49177, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49178, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49179, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49180, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49181, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49182, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49183, 2, {0x24A8, 0x24A8, 0x3FA8, 0x24A8, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49184, 2, {0x0, 0x8, 0x4448, 0x4448, 0x4448, 0x7FC8, 0x4448, 0x4448, 0x7FC8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {49185, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49186, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49187, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49188, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49189, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49190, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49191, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49192, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49193, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49194, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49195, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49196, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49197, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49198, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49199, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49200, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49201, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49202, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49203, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49204, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49205, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49206, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49207, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49208, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49209, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49210, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49211, 2, {0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49212, 2, {0x0, 0x0, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {49213, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49214, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49215, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49216, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49217, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49218, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49219, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49220, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49221, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49222, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49223, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49224, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49225, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49226, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49227, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49228, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49229, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49230, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49231, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49232, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49233, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49234, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49235, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49236, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49237, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49238, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49239, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49240, 2, {0x0, 0x0, 0x2108, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {49241, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49242, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49243, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49244, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49245, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49246, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49247, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49248, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49249, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49250, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49251, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49252, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49253, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49254, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49255, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49256, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49257, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49258, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49259, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49260, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49261, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49262, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49263, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49264, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49265, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49266, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49267, 2, {0x0, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49268, 2, {0x0, 0x8, 0x8, 0x4448, 0x4448, 0x4448, 0x7FC8, 0x4448, 0x4448, 0x7FC8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {49269, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49270, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49271, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49272, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49273, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49274, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49275, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49276, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49277, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49278, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49279, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49280, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49281, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49282, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49283, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49284, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49285, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49286, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49287, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49288, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49289, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49290, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49291, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49292, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49293, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49294, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49295, 2, {0x0, 0x2488, 0x2488, 0x3F88, 0x2488, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49296, 2, {0x0, 0x0, 0x8, 0x8, 0x2488, 0x2488, 0x2488, 0x3F88, 0x2488, 0x2488, 0x3F88, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49297, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49298, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49299, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49300, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49301, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49302, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49303, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49304, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49305, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49306, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49307, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49308, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49309, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49310, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49311, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49312, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49313, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49314, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49315, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49316, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49317, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49318, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49319, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49320, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49321, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49322, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49323, 2, {0x0, 0x8, 0x4908, 0x4908, 0x4908, 0x7F08, 0x4908, 0x7F08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49324, 2, {0x0, 0x0, 0x10, 0x10, 0x210, 0x210, 0x410, 0x41E, 0xA10, 0x1110, 0x6090, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49325, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49326, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49327, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49328, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49329, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49330, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49331, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49332, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49333, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49334, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49335, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49336, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49337, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49338, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49339, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49340, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49341, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49342, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49343, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49344, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49345, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49346, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49347, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49348, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49349, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49350, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49351, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC0E, 0x1208, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49352, 2, {0x0, 0x0, 0x12, 0x12, 0x212, 0x212, 0x412, 0x41E, 0xA12, 0x1112, 0x6092, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {49353, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49354, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49355, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49356, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49357, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49358, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49359, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49360, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49361, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49362, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49363, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49364, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49365, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49366, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49367, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49368, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49369, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49370, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49371, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {49372, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49373, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {49374, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49375, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49376, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49377, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49378, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49379, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xC38, 0x1228, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49380, 2, {0x0, 0x0, 0x10, 0x10, 0x210, 0x210, 0x41E, 0x410, 0xA10, 0x111E, 0x6090, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49381, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49382, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49383, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49384, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49385, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49386, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49387, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49388, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49389, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49390, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49391, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49392, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49393, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49394, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49395, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49396, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49397, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49398, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49399, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49400, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49401, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49402, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49403, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49404, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49405, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49406, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49407, 2, {0x0, 0x8, 0x8, 0x408, 0x40E, 0xC08, 0x120E, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49408, 2, {0x0, 0x0, 0x12, 0x12, 0x212, 0x212, 0x41E, 0x412, 0xA12, 0x111E, 0x6092, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {49409, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49410, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49411, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49412, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49413, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49414, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49415, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49416, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49417, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49418, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49419, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49420, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49421, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49422, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49423, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49424, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49425, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49426, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49427, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {49428, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49429, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {49430, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49431, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49432, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49433, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49434, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49435, 2, {0x0, 0x28, 0x28, 0x428, 0x438, 0xC28, 0x1238, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49436, 2, {0x0, 0x0, 0x2, 0x2, 0x202, 0x202, 0x402, 0x41E, 0xA02, 0x1102, 0x6082, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {49437, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49438, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49439, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49440, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49441, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49442, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49443, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49444, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49445, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49446, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49447, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49448, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49449, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49450, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49451, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49452, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49453, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49454, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49455, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49456, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49457, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49458, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49459, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49460, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49461, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49462, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49463, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC38, 0x1208, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49464, 2, {0x0, 0x0, 0xA, 0xA, 0x20A, 0x20A, 0x40A, 0x43A, 0xA0A, 0x110A, 0x608A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {49465, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49466, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49467, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49468, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49469, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49470, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49471, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49472, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49473, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49474, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49475, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49476, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49477, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49478, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49479, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49480, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49481, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49482, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49483, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {49484, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49485, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {49486, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49487, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49488, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49489, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49490, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49491, 2, {0x0, 0x28, 0x28, 0x428, 0x428, 0xCE8, 0x1228, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49492, 2, {0x0, 0x0, 0x2, 0x2, 0x202, 0x202, 0x41E, 0x402, 0xA02, 0x111E, 0x6082, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {49493, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49494, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49495, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49496, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49497, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49498, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49499, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49500, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49501, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49502, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49503, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49504, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49505, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49506, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49507, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49508, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49509, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49510, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49511, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49512, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49513, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49514, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49515, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49516, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49517, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49518, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49519, 2, {0x0, 0x8, 0x8, 0x408, 0x438, 0xC08, 0x1238, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49520, 2, {0x0, 0x0, 0xA, 0xA, 0x20A, 0x20A, 0x43A, 0x40A, 0xA0A, 0x113A, 0x608A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {49521, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49522, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49523, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49524, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49525, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49526, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49527, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49528, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49529, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49530, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49531, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49532, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49533, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49534, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49535, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49536, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49537, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49538, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49539, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {49540, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49541, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {49542, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49543, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49544, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49545, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49546, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49547, 2, {0x0, 0x28, 0x28, 0x428, 0x4E8, 0xC28, 0x12E8, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49548, 2, {0x0, 0x0, 0x80, 0x80, 0x100, 0x100, 0x280, 0x440, 0x1820, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {49549, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49550, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49551, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49552, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49553, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49554, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49555, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49556, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49557, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49558, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49559, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49560, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49561, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49562, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49563, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49564, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49565, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49566, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49567, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49568, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49569, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49570, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49571, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49572, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49573, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49574, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49575, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49576, 2, {0x0, 0x10, 0x10, 0x210, 0x210, 0x410, 0x410, 0xA1E, 0x1110, 0x6090, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {49577, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49578, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49579, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49580, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49581, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49582, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49583, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49584, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49585, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49586, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49587, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49588, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49589, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49590, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49591, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49592, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49593, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49594, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49595, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49596, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49597, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49598, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49599, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49600, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49601, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49602, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49603, 2, {0x0, 0x208, 0x208, 0x608, 0x90E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49604, 2, {0x0, 0x12, 0x12, 0x212, 0x212, 0x412, 0x412, 0xA1E, 0x1112, 0x6092, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {49605, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49606, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49607, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49608, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49609, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49610, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49611, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49612, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49613, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49614, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49615, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49616, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49617, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49618, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49619, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49620, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49621, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49622, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49623, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49624, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49625, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49626, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49627, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49628, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49629, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49630, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49631, 2, {0x0, 0x228, 0x228, 0x628, 0x938, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49632, 2, {0x0, 0x8, 0x8, 0x208, 0x208, 0x408, 0x408, 0xA08, 0x1108, 0x6088, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {49633, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49634, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49635, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49636, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49637, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49638, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49639, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49640, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49641, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49642, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49643, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49644, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49645, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49646, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49647, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49648, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49649, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49650, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49651, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49652, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49653, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49654, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49655, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49656, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49657, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49658, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49659, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49660, 2, {0x0, 0x0, 0x80, 0x80, 0x100, 0x100, 0x280, 0x440, 0x1820, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {49661, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49662, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49663, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49664, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49665, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49666, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49667, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49668, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49669, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49670, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49671, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49672, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49673, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49674, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49675, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49676, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49677, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49678, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49679, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49680, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49681, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49682, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49683, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49684, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49685, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49686, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49687, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49688, 2, {0x0, 0x0, 0x80, 0x80, 0x100, 0x100, 0x280, 0x440, 0x1820, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {49689, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49690, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49691, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49692, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49693, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49694, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49695, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49696, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49697, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49698, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49699, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49700, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49701, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49702, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49703, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49704, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49705, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49706, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49707, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49708, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49709, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49710, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49711, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49712, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49713, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49714, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49715, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49716, 2, {0x0, 0x8, 0x208, 0x208, 0x408, 0x408, 0xA08, 0x1108, 0x6088, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {49717, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49718, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49719, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49720, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49721, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49722, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49723, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49724, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49725, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49726, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49727, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49728, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49729, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49730, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49731, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49732, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49733, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49734, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49735, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49736, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49737, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49738, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49739, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49740, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49741, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49742, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49743, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49744, 2, {0x0, 0xA, 0x20A, 0x20A, 0x40A, 0x40A, 0xA0A, 0x110A, 0x608A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {49745, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49746, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49747, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49748, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49749, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49750, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49751, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49752, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49753, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49754, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49755, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49756, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49757, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49758, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49759, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49760, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49761, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49762, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49763, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49764, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49765, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49766, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49767, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49768, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49769, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49770, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49771, 2, {0x228, 0x228, 0x628, 0x928, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49772, 2, {0x0, 0x8, 0x208, 0x208, 0x408, 0x408, 0xA08, 0x1108, 0x6088, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {49773, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49774, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49775, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49776, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49777, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49778, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49779, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49780, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49781, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49782, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49783, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49784, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49785, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49786, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49787, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49788, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49789, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49790, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49791, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49792, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49793, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49794, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49795, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49796, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49797, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49798, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49799, 2, {0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49800, 2, {0x0, 0x0, 0x80, 0x80, 0x100, 0x100, 0x280, 0x440, 0x1820, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {49801, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49802, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49803, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49804, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49805, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49806, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49807, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49808, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49809, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49810, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49811, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49812, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49813, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49814, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49815, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49816, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49817, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49818, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49819, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49820, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49821, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49822, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49823, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49824, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49825, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49826, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49827, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49828, 2, {0x0, 0x0, 0x80, 0x80, 0x100, 0x100, 0x280, 0x440, 0x1820, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {49829, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49830, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49831, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49832, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49833, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49834, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49835, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {49836, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49837, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49838, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49839, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49840, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {49841, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49842, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49843, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {49844, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {49845, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {49846, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {49847, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {49848, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {49849, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {49850, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49851, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {49852, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {49853, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {49854, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {49855, 2, {0x0, 0x80, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {49856, 2, {0x0, 0x8, 0x8, 0x208, 0x208, 0x408, 0x408, 0xA08, 0x1108, 0x6088, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {49857, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49858, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49859, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49860, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49861, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49862, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49863, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49864, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49865, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49866, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49867, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49868, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49869, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49870, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49871, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49872, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49873, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49874, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49875, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49876, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49877, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49878, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49879, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49880, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49881, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49882, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49883, 2, {0x0, 0x208, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49884, 2, {0x0, 0x0, 0x8, 0x8, 0x208, 0x208, 0x408, 0x408, 0xA08, 0x1108, 0x6088, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49885, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {49886, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49887, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49888, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {49889, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49890, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49891, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49892, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49893, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49894, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49895, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49896, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {49897, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49898, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49899, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49900, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49901, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49902, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {49903, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49904, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49905, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49906, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49907, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49908, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49909, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49910, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49911, 2, {0x0, 0x8, 0x8, 0x408, 0x408, 0xC08, 0x1208, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49912, 2, {0x0, 0x0, 0x10, 0x10, 0x890, 0x890, 0x890, 0x191E, 0x1510, 0x2290, 0xCC50, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49913, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49914, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49915, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49916, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49917, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49918, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49919, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49920, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49921, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49922, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49923, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49924, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49925, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49926, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49927, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49928, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49929, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49930, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49931, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49932, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49933, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49934, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49935, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49936, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49937, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49938, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49939, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x340E, 0x2A08, 0x4908, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49940, 2, {0x0, 0x0, 0x12, 0x12, 0x892, 0x892, 0x892, 0x191E, 0x1512, 0x2292, 0xCC52, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {49941, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49942, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49943, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {49944, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49945, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {49946, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {49947, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49948, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49949, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49950, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49951, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49952, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49953, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49954, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49955, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {49956, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49957, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49958, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49959, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {49960, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49961, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {49962, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49963, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49964, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49965, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49966, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49967, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x3438, 0x2A28, 0x4928, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49968, 2, {0x0, 0x0, 0x10, 0x10, 0x890, 0x890, 0x89E, 0x1910, 0x1510, 0x229E, 0xCC50, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {49969, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49970, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49971, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {49972, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {49973, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {49974, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {49975, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {49976, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {49977, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {49978, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {49979, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {49980, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {49981, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {49982, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {49983, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {49984, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {49985, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {49986, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {49987, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {49988, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {49989, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {49990, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {49991, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {49992, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {49993, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {49994, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {49995, 2, {0x0, 0x8, 0x8, 0x1208, 0x120E, 0x3408, 0x2A0E, 0x4908, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {49996, 2, {0x0, 0x0, 0x12, 0x12, 0x892, 0x892, 0x89E, 0x1912, 0x1512, 0x229E, 0xCC52, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {49997, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {49998, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {49999, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50000, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50001, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50002, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50003, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50004, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50005, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50006, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50007, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50008, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50009, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50010, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50011, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50012, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50013, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50014, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50015, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50016, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50017, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50018, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50019, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50020, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50021, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50022, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50023, 2, {0x0, 0x28, 0x28, 0x1228, 0x1238, 0x3428, 0x2A38, 0x4928, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50024, 2, {0x0, 0x0, 0x2, 0x2, 0x882, 0x882, 0x882, 0x191E, 0x1502, 0x2282, 0xCC42, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {50025, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50026, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50027, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50028, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50029, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50030, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50031, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50032, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50033, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50034, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50035, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50036, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50037, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50038, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50039, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50040, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50041, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50042, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50043, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50044, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50045, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50046, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50047, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50048, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50049, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50050, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50051, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3438, 0x2A08, 0x4908, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50052, 2, {0x0, 0x0, 0xA, 0xA, 0x88A, 0x88A, 0x88A, 0x193A, 0x150A, 0x228A, 0xCC4A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {50053, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50054, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50055, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50056, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50057, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50058, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50059, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50060, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50061, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50062, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50063, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50064, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50065, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50066, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50067, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50068, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50069, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50070, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50071, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50072, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50073, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50074, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50075, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50076, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50077, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50078, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50079, 2, {0x0, 0x28, 0x28, 0x1228, 0x1228, 0x34E8, 0x2A28, 0x4928, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50080, 2, {0x0, 0x0, 0x2, 0x2, 0x882, 0x882, 0x89E, 0x1902, 0x1502, 0x229E, 0xCC42, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {50081, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50082, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50083, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50084, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50085, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50086, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50087, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50088, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50089, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50090, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50091, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50092, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50093, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50094, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50095, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50096, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50097, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50098, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50099, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50100, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50101, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50102, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50103, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50104, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50105, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50106, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50107, 2, {0x0, 0x8, 0x8, 0x1208, 0x1238, 0x3408, 0x2A38, 0x4908, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50108, 2, {0x0, 0x0, 0xA, 0xA, 0x88A, 0x88A, 0x8BA, 0x190A, 0x150A, 0x22BA, 0xCC4A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {50109, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50110, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50111, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50112, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50113, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50114, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50115, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50116, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50117, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50118, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50119, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50120, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50121, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50122, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50123, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50124, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50125, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50126, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50127, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50128, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50129, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50130, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50131, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50132, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50133, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50134, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50135, 2, {0x0, 0x28, 0x28, 0x1228, 0x12E8, 0x3428, 0x2AE8, 0x4928, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50136, 2, {0x0, 0x0, 0x0, 0x220, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {50137, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50138, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50139, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50140, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50141, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50142, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50143, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50144, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50145, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50146, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50147, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50148, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50149, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50150, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50151, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50152, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50153, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50154, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50155, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50156, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50157, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50158, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50159, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50160, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50161, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50162, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50163, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50164, 2, {0x0, 0x10, 0x10, 0x890, 0x890, 0x890, 0x1910, 0x151E, 0x2290, 0xCC50, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {50165, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50166, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50167, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50168, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50169, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50170, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50171, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50172, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50173, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50174, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50175, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50176, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50177, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50178, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50179, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50180, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50181, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50182, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50183, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50184, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50185, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50186, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50187, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50188, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50189, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50190, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50191, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50192, 2, {0x0, 0x12, 0x12, 0x892, 0x892, 0x892, 0x1912, 0x151E, 0x2292, 0xCC52, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {50193, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50194, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50195, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50196, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50197, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50198, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50199, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50200, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50201, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50202, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50203, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50204, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50205, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50206, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50207, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50208, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50209, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50210, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50211, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50212, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50213, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50214, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50215, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50216, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50217, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50218, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50219, 2, {0x0, 0x8A8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50220, 2, {0x0, 0x8, 0x8, 0x888, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {50221, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50222, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50223, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50224, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50225, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50226, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50227, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50228, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50229, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50230, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50231, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50232, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50233, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50234, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50235, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50236, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50237, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50238, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50239, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50240, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50241, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50242, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50243, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50244, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50245, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50246, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50247, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50248, 2, {0x0, 0x0, 0x0, 0x220, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3750, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {50249, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50250, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50251, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50252, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50253, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50254, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50255, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50256, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50257, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50258, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50259, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50260, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50261, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50262, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50263, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50264, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50265, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50266, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50267, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50268, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50269, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50270, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50271, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50272, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50273, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50274, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50275, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1550, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50276, 2, {0x0, 0x0, 0x220, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {50277, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50278, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50279, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50280, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50281, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50282, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50283, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50284, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50285, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50286, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50287, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50288, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50289, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50290, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50291, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50292, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50293, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50294, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50295, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50296, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50297, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50298, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50299, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50300, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50301, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50302, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50303, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50304, 2, {0x0, 0x8, 0x888, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {50305, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50306, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50307, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50308, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50309, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50310, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50311, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50312, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50313, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50314, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50315, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50316, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50317, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50318, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50319, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50320, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50321, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50322, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50323, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50324, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50325, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50326, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50327, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50328, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50329, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50330, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50331, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50332, 2, {0x0, 0xA, 0x88A, 0x88A, 0x88A, 0x190A, 0x150A, 0x228A, 0xCC4A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {50333, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50334, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50335, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50336, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50337, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50338, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50339, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50340, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50341, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50342, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50343, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50344, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50345, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50346, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50347, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50348, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50349, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50350, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50351, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50352, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50353, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50354, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50355, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50356, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50357, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50358, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50359, 2, {0x28, 0x8A8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50360, 2, {0x0, 0x8, 0x888, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {50361, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50362, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50363, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50364, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50365, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50366, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50367, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50368, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50369, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50370, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50371, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50372, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50373, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50374, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50375, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50376, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50377, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50378, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50379, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50380, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50381, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50382, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50383, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50384, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50385, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50386, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50387, 2, {0x8, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50388, 2, {0x0, 0x0, 0x220, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {50389, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50390, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50391, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50392, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50393, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50394, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50395, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50396, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50397, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50398, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50399, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50400, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50401, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50402, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50403, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50404, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50405, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50406, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50407, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50408, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50409, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50410, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50411, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50412, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50413, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50414, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50415, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50416, 2, {0x0, 0x0, 0x0, 0x220, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {50417, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50418, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50419, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50420, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50421, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50422, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50423, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50424, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50425, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50426, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50427, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50428, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50429, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50430, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50431, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50432, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50433, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50434, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50435, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50436, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50437, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50438, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50439, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50440, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50441, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50442, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50443, 2, {0x0, 0x0, 0x220, 0x220, 0x440, 0xAA0, 0x1110, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50444, 2, {0x0, 0x8, 0x8, 0x888, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {50445, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50446, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50447, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50448, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50449, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50450, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50451, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50452, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50453, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50454, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50455, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50456, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50457, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50458, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50459, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50460, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50461, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50462, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50463, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50464, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50465, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50466, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50467, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50468, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50469, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50470, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50471, 2, {0x0, 0x888, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50472, 2, {0x0, 0x0, 0x8, 0x8, 0x888, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50473, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50474, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50475, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50476, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50477, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50478, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50479, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50480, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50481, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50482, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50483, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50484, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50485, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50486, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50487, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50488, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50489, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50490, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50491, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50492, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50493, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50494, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50495, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50496, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50497, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50498, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50499, 2, {0x0, 0x8, 0x8, 0x1208, 0x1208, 0x3408, 0x2A08, 0x4908, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50500, 2, {0x0, 0x0, 0x10, 0x10, 0x1E10, 0x2110, 0x4090, 0x409E, 0x4090, 0x2110, 0x1E10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50501, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50502, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50503, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50504, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50505, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50506, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50507, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50508, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50509, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50510, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50511, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50512, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50513, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50514, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50515, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50516, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50517, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50518, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50519, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50520, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50521, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50522, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50523, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50524, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50525, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50526, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50527, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x210E, 0x2108, 0x1E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50528, 2, {0x0, 0x0, 0x12, 0x12, 0x1E12, 0x2112, 0x4092, 0x409E, 0x4092, 0x2112, 0x1E12, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {50529, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50530, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50531, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50532, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50533, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50534, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50535, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50536, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50537, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50538, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50539, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50540, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50541, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50542, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50543, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50544, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50545, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50546, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50547, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50548, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50549, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50550, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50551, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50552, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50553, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50554, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50555, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x2138, 0x2128, 0x1E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50556, 2, {0x0, 0x0, 0x10, 0x10, 0x1E10, 0x2110, 0x409E, 0x4090, 0x4090, 0x211E, 0x1E10, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50557, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50558, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50559, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50560, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50561, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50562, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50563, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50564, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50565, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50566, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50567, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50568, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50569, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50570, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50571, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50572, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50573, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50574, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50575, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50576, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50577, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50578, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50579, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50580, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50581, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50582, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50583, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x210E, 0x2108, 0x210E, 0x1E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50584, 2, {0x0, 0x0, 0x12, 0x12, 0x1E12, 0x2112, 0x409E, 0x4092, 0x4092, 0x211E, 0x1E12, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {50585, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50586, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50587, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50588, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50589, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50590, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50591, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50592, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50593, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50594, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50595, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50596, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50597, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50598, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50599, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50600, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50601, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50602, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50603, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50604, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50605, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50606, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50607, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50608, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50609, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50610, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50611, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2138, 0x2128, 0x2138, 0x1E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50612, 2, {0x0, 0x0, 0x2, 0x2, 0x1E02, 0x2102, 0x4082, 0x409E, 0x4082, 0x2102, 0x1E02, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {50613, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50614, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50615, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50616, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50617, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50618, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50619, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50620, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50621, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50622, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50623, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50624, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50625, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50626, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50627, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50628, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50629, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50630, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50631, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50632, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50633, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50634, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50635, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50636, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50637, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50638, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50639, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2138, 0x2108, 0x1E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50640, 2, {0x0, 0x0, 0xA, 0xA, 0x1E0A, 0x210A, 0x408A, 0x40BA, 0x408A, 0x210A, 0x1E0A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {50641, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50642, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50643, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50644, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50645, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50646, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50647, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50648, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50649, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50650, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50651, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50652, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50653, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50654, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50655, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50656, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50657, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50658, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50659, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50660, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50661, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50662, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50663, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50664, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50665, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50666, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50667, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x2128, 0x21E8, 0x2128, 0x1E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50668, 2, {0x0, 0x0, 0x2, 0x2, 0x1E02, 0x2102, 0x409E, 0x4082, 0x4082, 0x211E, 0x1E02, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {50669, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50670, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50671, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50672, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50673, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50674, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50675, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50676, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50677, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50678, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50679, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50680, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50681, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50682, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50683, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50684, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50685, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50686, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50687, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50688, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50689, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50690, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50691, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50692, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50693, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50694, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50695, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2138, 0x2108, 0x2138, 0x1E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50696, 2, {0x0, 0x0, 0xA, 0xA, 0x1E0A, 0x210A, 0x40BA, 0x408A, 0x408A, 0x213A, 0x1E0A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {50697, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50698, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50699, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50700, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50701, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50702, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50703, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50704, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50705, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50706, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50707, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50708, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50709, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50710, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50711, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50712, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50713, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50714, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50715, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {50716, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50717, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {50718, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50719, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50720, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50721, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50722, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50723, 2, {0x0, 0x28, 0x1E28, 0x2128, 0x21E8, 0x2128, 0x21E8, 0x1E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50724, 2, {0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {50725, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50726, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50727, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50728, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50729, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50730, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50731, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50732, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50733, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50734, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50735, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50736, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50737, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50738, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50739, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50740, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50741, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50742, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50743, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50744, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50745, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50746, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50747, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50748, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50749, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50750, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50751, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50752, 2, {0x0, 0x10, 0x10, 0x1F10, 0x2090, 0x4050, 0x4050, 0x405E, 0x2090, 0x1F10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {50753, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50754, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50755, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50756, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50757, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50758, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50759, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50760, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50761, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50762, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50763, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50764, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50765, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50766, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50767, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50768, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50769, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50770, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50771, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50772, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50773, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50774, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50775, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50776, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50777, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50778, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50779, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x208E, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50780, 2, {0x0, 0x12, 0x12, 0x1F12, 0x2092, 0x4052, 0x4052, 0x405E, 0x2092, 0x1F12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {50781, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50782, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50783, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50784, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50785, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50786, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50787, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50788, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50789, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50790, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50791, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50792, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50793, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50794, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50795, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50796, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50797, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50798, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50799, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50800, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50801, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50802, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50803, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50804, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50805, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50806, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50807, 2, {0x0, 0x28, 0x1F28, 0x20A8, 0x20B8, 0x20A8, 0x1F28, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50808, 2, {0x0, 0x8, 0x8, 0x1F08, 0x2088, 0x4048, 0x4048, 0x4048, 0x2088, 0x1F08, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {50809, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {50810, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50811, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50812, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {50813, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {50814, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {50815, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50816, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50817, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50818, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50819, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50820, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {50821, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50822, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50823, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {50824, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50825, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50826, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {50827, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50828, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50829, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50830, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50831, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50832, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50833, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50834, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50835, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50836, 2, {0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {50837, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50838, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50839, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50840, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50841, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50842, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50843, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50844, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50845, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50846, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50847, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50848, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50849, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50850, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50851, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50852, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50853, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50854, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50855, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50856, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50857, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50858, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50859, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50860, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50861, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50862, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50863, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50864, 2, {0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {50865, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50866, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50867, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50868, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50869, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50870, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50871, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50872, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50873, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50874, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50875, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50876, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50877, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50878, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50879, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50880, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50881, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50882, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50883, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50884, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50885, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50886, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50887, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50888, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {50889, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50890, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {50891, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {50892, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x4048, 0x4048, 0x4048, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {50893, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50894, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50895, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50896, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50897, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50898, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50899, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50900, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50901, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50902, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50903, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50904, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50905, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50906, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50907, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50908, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50909, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50910, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50911, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50912, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50913, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50914, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50915, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50916, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50917, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50918, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50919, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50920, 2, {0x0, 0xA, 0x1F0A, 0x208A, 0x404A, 0x404A, 0x404A, 0x208A, 0x1F0A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {50921, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50922, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50923, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50924, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50925, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50926, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50927, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50928, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50929, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50930, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50931, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50932, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50933, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50934, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50935, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50936, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50937, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50938, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50939, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50940, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50941, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50942, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50943, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50944, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50945, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50946, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50947, 2, {0x1F28, 0x20A8, 0x20A8, 0x20A8, 0x1F28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50948, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x4048, 0x4048, 0x4048, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {50949, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {50950, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50951, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {50952, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {50953, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50954, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50955, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {50956, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {50957, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50958, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50959, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50960, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {50961, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50962, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50963, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {50964, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {50965, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {50966, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {50967, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {50968, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {50969, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {50970, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {50971, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {50972, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {50973, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {50974, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {50975, 2, {0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {50976, 2, {0x0, 0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {50977, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {50978, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {50979, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {50980, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50981, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {50982, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {50983, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {50984, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {50985, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {50986, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {50987, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {50988, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {50989, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {50990, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {50991, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {50992, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {50993, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {50994, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {50995, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {50996, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {50997, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {50998, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {50999, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51000, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51001, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51002, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51003, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51004, 2, {0x0, 0x7C0, 0x820, 0x1010, 0x1010, 0x1010, 0x820, 0x7C0, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {51005, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51006, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51007, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51008, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51009, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51010, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51011, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51012, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51013, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51014, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51015, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51016, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51017, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51018, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51019, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51020, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51021, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51022, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51023, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51024, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51025, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51026, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51027, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51028, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51029, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51030, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51031, 2, {0x0, 0x7C0, 0x820, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51032, 2, {0x0, 0x8, 0x8, 0x1F08, 0x2088, 0x4048, 0x4048, 0x4048, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {51033, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51034, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51035, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51036, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51037, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51038, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51039, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51040, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51041, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51042, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51043, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51044, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51045, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51046, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51047, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51048, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51049, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51050, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51051, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51052, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51053, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51054, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51055, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51056, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51057, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51058, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51059, 2, {0x0, 0x8, 0x1F08, 0x2088, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51060, 2, {0x0, 0x0, 0x8, 0x8, 0x1E08, 0x2108, 0x4088, 0x4088, 0x4088, 0x2108, 0x1E08, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51061, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51062, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51063, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51064, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51065, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51066, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51067, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51068, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51069, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51070, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51071, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51072, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51073, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51074, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51075, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51076, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51077, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51078, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51079, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51080, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51081, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51082, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51083, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51084, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51085, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51086, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51087, 2, {0x0, 0x8, 0x1E08, 0x2108, 0x2108, 0x2108, 0x2108, 0x1E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51088, 2, {0x0, 0x0, 0x10, 0x10, 0x3F90, 0x110, 0x110, 0x21E, 0x610, 0x910, 0x3090, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51089, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51090, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51091, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51092, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51093, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51094, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51095, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51096, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51097, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51098, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51099, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51100, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51101, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51102, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51103, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51104, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51105, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51106, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51107, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51108, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51109, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51110, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51111, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51112, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51113, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51114, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51115, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x40E, 0x1A08, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51116, 2, {0x0, 0x0, 0x12, 0x12, 0x3F92, 0x112, 0x112, 0x21E, 0x612, 0x912, 0x3092, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {51117, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51118, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51119, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51120, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51121, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51122, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51123, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51124, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51125, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51126, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51127, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51128, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51129, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51130, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51131, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51132, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51133, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51134, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51135, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51136, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51137, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51138, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51139, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51140, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51141, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51142, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51143, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x438, 0x1A28, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51144, 2, {0x0, 0x0, 0x10, 0x10, 0x3F90, 0x110, 0x11E, 0x210, 0x610, 0x91E, 0x3090, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51145, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51146, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51147, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51148, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51149, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51150, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51151, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51152, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51153, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51154, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51155, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51156, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51157, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51158, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51159, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51160, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51161, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51162, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51163, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51164, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51165, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51166, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51167, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51168, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51169, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51170, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51171, 2, {0x0, 0x8, 0x8, 0x7E08, 0x40E, 0x408, 0x1A0E, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51172, 2, {0x0, 0x0, 0x12, 0x12, 0x3F92, 0x112, 0x11E, 0x212, 0x612, 0x91E, 0x3092, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {51173, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51174, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51175, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51176, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51177, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51178, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51179, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51180, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51181, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51182, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51183, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51184, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51185, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51186, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51187, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51188, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51189, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51190, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51191, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51192, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51193, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51194, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51195, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51196, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51197, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51198, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51199, 2, {0x0, 0x28, 0x28, 0x7E28, 0x438, 0x428, 0x1A38, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51200, 2, {0x0, 0x0, 0x2, 0x2, 0x3F82, 0x102, 0x102, 0x21E, 0x602, 0x902, 0x3082, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {51201, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51202, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51203, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51204, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51205, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51206, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51207, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51208, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51209, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51210, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51211, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51212, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51213, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51214, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51215, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51216, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51217, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51218, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51219, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51220, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51221, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51222, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51223, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51224, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51225, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51226, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51227, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x438, 0x1A08, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51228, 2, {0x0, 0x0, 0xA, 0xA, 0x3F8A, 0x10A, 0x10A, 0x23A, 0x60A, 0x90A, 0x308A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {51229, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51230, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51231, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51232, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51233, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51234, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51235, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51236, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51237, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51238, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51239, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51240, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51241, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51242, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51243, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51244, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51245, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51246, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51247, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51248, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51249, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51250, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51251, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51252, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51253, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51254, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51255, 2, {0x0, 0x28, 0x28, 0x7E28, 0x428, 0x4E8, 0x1A28, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51256, 2, {0x0, 0x0, 0x2, 0x2, 0x3F82, 0x102, 0x11E, 0x202, 0x602, 0x91E, 0x3082, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {51257, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51258, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51259, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51260, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51261, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51262, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51263, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51264, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51265, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51266, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51267, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51268, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51269, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51270, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51271, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51272, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51273, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51274, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51275, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51276, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51277, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51278, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51279, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51280, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51281, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51282, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51283, 2, {0x0, 0x8, 0x8, 0x7E08, 0x438, 0x408, 0x1A38, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51284, 2, {0x0, 0x0, 0xA, 0xA, 0x3F8A, 0x10A, 0x13A, 0x20A, 0x60A, 0x93A, 0x308A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {51285, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51286, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51287, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51288, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51289, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51290, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51291, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51292, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51293, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51294, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51295, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51296, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51297, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51298, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51299, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51300, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51301, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51302, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51303, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51304, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51305, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51306, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51307, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51308, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51309, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51310, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51311, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4E8, 0x428, 0x1AE8, 0x6128, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51312, 2, {0x0, 0x0, 0xFE0, 0x40, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {51313, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51314, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51315, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51316, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51317, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51318, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51319, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51320, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51321, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51322, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51323, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51324, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51325, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51326, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51327, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51328, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51329, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51330, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51331, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51332, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51333, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51334, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51335, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51336, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51337, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51338, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51339, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51340, 2, {0x0, 0x10, 0x10, 0x3F90, 0x110, 0x110, 0x210, 0x61E, 0x910, 0x3090, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {51341, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51342, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51343, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51344, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51345, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51346, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51347, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51348, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51349, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51350, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51351, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51352, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51353, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51354, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51355, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51356, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51357, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51358, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51359, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51360, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51361, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51362, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51363, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51364, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51365, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51366, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51367, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD0E, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51368, 2, {0x0, 0x12, 0x12, 0x3F92, 0x112, 0x112, 0x212, 0x61E, 0x912, 0x3092, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {51369, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51370, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51371, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51372, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51373, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51374, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51375, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51376, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51377, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51378, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51379, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51380, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51381, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51382, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51383, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51384, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51385, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51386, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51387, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51388, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51389, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51390, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51391, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51392, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51393, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51394, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51395, 2, {0x0, 0x28, 0x3F28, 0x228, 0xD38, 0x30A8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51396, 2, {0x0, 0x8, 0x8, 0x3F88, 0x108, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {51397, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51398, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51399, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51400, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51401, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51402, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51403, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51404, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51405, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51406, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51407, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51408, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51409, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51410, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51411, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51412, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51413, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51414, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51415, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51416, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51417, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51418, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51419, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51420, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51421, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51422, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51423, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51424, 2, {0x0, 0x0, 0xFE0, 0x40, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {51425, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51426, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51427, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51428, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51429, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51430, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51431, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51432, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51433, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51434, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51435, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51436, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51437, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51438, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51439, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51440, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51441, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51442, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51443, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51444, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51445, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51446, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51447, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51448, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51449, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51450, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51451, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51452, 2, {0x0, 0x0, 0xFE0, 0x40, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {51453, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51454, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51455, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51456, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51457, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51458, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51459, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51460, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51461, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51462, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51463, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51464, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51465, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51466, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51467, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51468, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51469, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51470, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51471, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51472, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51473, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51474, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51475, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51476, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51477, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51478, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51479, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51480, 2, {0x0, 0x8, 0x3F88, 0x108, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {51481, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51482, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51483, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51484, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51485, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51486, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51487, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51488, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51489, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51490, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51491, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51492, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51493, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51494, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51495, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51496, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51497, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51498, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51499, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51500, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51501, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51502, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51503, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51504, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51505, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51506, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51507, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51508, 2, {0x0, 0xA, 0x3F8A, 0x10A, 0x10A, 0x20A, 0x60A, 0x90A, 0x308A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {51509, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51510, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51511, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51512, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51513, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51514, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51515, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51516, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51517, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51518, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51519, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51520, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51521, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51522, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51523, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51524, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51525, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51526, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51527, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51528, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51529, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51530, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51531, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51532, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51533, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51534, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51535, 2, {0x3F28, 0x228, 0x228, 0xD28, 0x30A8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51536, 2, {0x0, 0x8, 0x3F88, 0x108, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {51537, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51538, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51539, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51540, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51541, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51542, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51543, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51544, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51545, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51546, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51547, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51548, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51549, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51550, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51551, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51552, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51553, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51554, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51555, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51556, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51557, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51558, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51559, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51560, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51561, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51562, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51563, 2, {0x3F08, 0x208, 0x208, 0xD08, 0x3088, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51564, 2, {0x0, 0x0, 0xFE0, 0x40, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {51565, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51566, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51567, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51568, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51569, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51570, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51571, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51572, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51573, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51574, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51575, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51576, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51577, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51578, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51579, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51580, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51581, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51582, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51583, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51584, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51585, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51586, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51587, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51588, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51589, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51590, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51591, 2, {0x0, 0xFC0, 0x80, 0x80, 0x340, 0xC20, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51592, 2, {0x0, 0x0, 0xFE0, 0x40, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {51593, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51594, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51595, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51596, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51597, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51598, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51599, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51600, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51601, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51602, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51603, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51604, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51605, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51606, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51607, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51608, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51609, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51610, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51611, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51612, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51613, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51614, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51615, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51616, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51617, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51618, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51619, 2, {0x0, 0xFC0, 0x80, 0x340, 0xC20, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51620, 2, {0x0, 0x8, 0x8, 0x3F88, 0x108, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {51621, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51622, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51623, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51624, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51625, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51626, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51627, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51628, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51629, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51630, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51631, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51632, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51633, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51634, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51635, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51636, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51637, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51638, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51639, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51640, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51641, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51642, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51643, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51644, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51645, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51646, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51647, 2, {0x0, 0x8, 0x3F08, 0x208, 0xD08, 0x3088, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51648, 2, {0x0, 0x0, 0x8, 0x8, 0x3F88, 0x108, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51649, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51650, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51651, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51652, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51653, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51654, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51655, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51656, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51657, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51658, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51659, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51660, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51661, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51662, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51663, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51664, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51665, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51666, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51667, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51668, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51669, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51670, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51671, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51672, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51673, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51674, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51675, 2, {0x0, 0x8, 0x8, 0x7E08, 0x408, 0x408, 0x1A08, 0x6108, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51676, 2, {0x0, 0x0, 0x10, 0x10, 0x7B90, 0x890, 0x890, 0x191E, 0x1510, 0x2290, 0xCC50, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51677, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51678, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51679, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51680, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51681, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51682, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51683, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51684, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51685, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51686, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51687, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51688, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51689, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51690, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51691, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51692, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51693, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51694, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51695, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51696, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51697, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51698, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51699, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51700, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51701, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51702, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51703, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x220E, 0x5508, 0x8888, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51704, 2, {0x0, 0x0, 0x12, 0x12, 0x7B92, 0x892, 0x892, 0x191E, 0x1512, 0x2292, 0xCC52, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {51705, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51706, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51707, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51708, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51709, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51710, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51711, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51712, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51713, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51714, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51715, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51716, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51717, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51718, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51719, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51720, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51721, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51722, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51723, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51724, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51725, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51726, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51727, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51728, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51729, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51730, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51731, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x2238, 0x5528, 0x88A8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51732, 2, {0x0, 0x0, 0x10, 0x10, 0x7B90, 0x890, 0x89E, 0x1910, 0x1510, 0x229E, 0xCC50, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51733, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51734, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51735, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51736, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51737, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51738, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51739, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51740, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51741, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51742, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51743, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51744, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51745, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51746, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51747, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51748, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51749, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51750, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51751, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51752, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51753, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51754, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51755, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51756, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51757, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51758, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51759, 2, {0x0, 0x8, 0x8, 0x7708, 0x220E, 0x2208, 0x550E, 0x8888, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51760, 2, {0x0, 0x0, 0x12, 0x12, 0x7B92, 0x892, 0x89E, 0x1912, 0x1512, 0x229E, 0xCC52, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {51761, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51762, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51763, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51764, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51765, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51766, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51767, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51768, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51769, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51770, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51771, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51772, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51773, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51774, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51775, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51776, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51777, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51778, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51779, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51780, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51781, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51782, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51783, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51784, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51785, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51786, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51787, 2, {0x0, 0x28, 0x28, 0x7728, 0x2238, 0x2228, 0x5538, 0x88A8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51788, 2, {0x0, 0x0, 0x2, 0x2, 0x7B82, 0x882, 0x882, 0x191E, 0x1502, 0x2282, 0xCC42, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {51789, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51790, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51791, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51792, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51793, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51794, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51795, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51796, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51797, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51798, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51799, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51800, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51801, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51802, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51803, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51804, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51805, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51806, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51807, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51808, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51809, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51810, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51811, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51812, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51813, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51814, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51815, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2238, 0x5508, 0x8888, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51816, 2, {0x0, 0x0, 0xA, 0xA, 0x7B8A, 0x88A, 0x88A, 0x193A, 0x150A, 0x228A, 0xCC4A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {51817, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51818, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51819, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51820, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51821, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51822, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51823, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51824, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51825, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51826, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51827, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51828, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51829, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51830, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51831, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51832, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51833, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51834, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51835, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51836, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51837, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51838, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51839, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51840, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51841, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51842, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51843, 2, {0x0, 0x28, 0x28, 0x7728, 0x2228, 0x22E8, 0x5528, 0x88A8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51844, 2, {0x0, 0x0, 0x2, 0x2, 0x7B82, 0x882, 0x89E, 0x1902, 0x1502, 0x229E, 0xCC42, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {51845, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {51846, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51847, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51848, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {51849, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51850, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51851, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51852, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51853, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51854, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51855, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51856, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {51857, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51858, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51859, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51860, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51861, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51862, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {51863, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51864, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51865, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51866, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51867, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51868, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51869, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51870, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51871, 2, {0x0, 0x8, 0x8, 0x7708, 0x2238, 0x2208, 0x5538, 0x8888, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51872, 2, {0x0, 0x0, 0xA, 0xA, 0x7B8A, 0x88A, 0x8BA, 0x190A, 0x150A, 0x22BA, 0xCC4A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {51873, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51874, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51875, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51876, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51877, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51878, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51879, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51880, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51881, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51882, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51883, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51884, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51885, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51886, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51887, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {51888, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51889, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51890, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51891, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {51892, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51893, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {51894, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51895, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51896, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51897, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51898, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51899, 2, {0x0, 0x28, 0x28, 0x7728, 0x22E8, 0x2228, 0x55E8, 0x88A8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51900, 2, {0x0, 0x0, 0x0, 0x1EE0, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {51901, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {51902, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51903, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {51904, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51905, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {51906, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {51907, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {51908, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51909, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51910, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51911, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51912, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {51913, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51914, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51915, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {51916, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {51917, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {51918, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {51919, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {51920, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {51921, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {51922, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51923, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {51924, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {51925, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {51926, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {51927, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {51928, 2, {0x0, 0x10, 0x10, 0x7B90, 0x890, 0x890, 0x1910, 0x151E, 0x2290, 0xCC50, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {51929, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51930, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51931, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51932, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51933, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51934, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51935, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51936, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51937, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51938, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51939, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51940, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51941, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51942, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51943, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51944, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51945, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51946, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51947, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51948, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51949, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51950, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51951, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51952, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51953, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51954, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51955, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A8E, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51956, 2, {0x0, 0x12, 0x12, 0x7B92, 0x892, 0x892, 0x1912, 0x151E, 0x2292, 0xCC52, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {51957, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51958, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51959, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51960, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51961, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51962, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51963, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51964, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51965, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51966, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51967, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51968, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51969, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51970, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51971, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {51972, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {51973, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {51974, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {51975, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {51976, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {51977, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {51978, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {51979, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {51980, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {51981, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {51982, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {51983, 2, {0x0, 0x7BE8, 0x8A8, 0x1128, 0x2AB8, 0x4468, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {51984, 2, {0x0, 0x8, 0x8, 0x7B88, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {51985, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {51986, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {51987, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {51988, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {51989, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {51990, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {51991, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {51992, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {51993, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {51994, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {51995, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {51996, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {51997, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {51998, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {51999, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52000, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52001, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52002, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52003, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52004, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52005, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52006, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52007, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52008, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52009, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52010, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52011, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52012, 2, {0x0, 0x0, 0x0, 0x1EE0, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3750, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {52013, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52014, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52015, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52016, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52017, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52018, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52019, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52020, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52021, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52022, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52023, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52024, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52025, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52026, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52027, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52028, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52029, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52030, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52031, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52032, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52033, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52034, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52035, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52036, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52037, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52038, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52039, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52040, 2, {0x0, 0x0, 0x1EE0, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {52041, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52042, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52043, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52044, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52045, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52046, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52047, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52048, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52049, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52050, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52051, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52052, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52053, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52054, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52055, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52056, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52057, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52058, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52059, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52060, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52061, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52062, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52063, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52064, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52065, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52066, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52067, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52068, 2, {0x0, 0x8, 0x7B88, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {52069, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52070, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52071, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52072, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52073, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52074, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52075, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52076, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52077, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52078, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52079, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52080, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52081, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52082, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52083, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52084, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52085, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52086, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52087, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52088, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52089, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52090, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52091, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52092, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52093, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52094, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52095, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52096, 2, {0x0, 0xA, 0x7B8A, 0x88A, 0x88A, 0x190A, 0x150A, 0x228A, 0xCC4A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {52097, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52098, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52099, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52100, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52101, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52102, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52103, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52104, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52105, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52106, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52107, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52108, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52109, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52110, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52111, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52112, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52113, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52114, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52115, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52116, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52117, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52118, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52119, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52120, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52121, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52122, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52123, 2, {0x7BE8, 0x8A8, 0x1128, 0x2AA8, 0x4468, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52124, 2, {0x0, 0x8, 0x7B88, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {52125, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52126, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52127, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52128, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52129, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52130, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52131, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52132, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52133, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52134, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52135, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52136, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52137, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52138, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52139, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52140, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52141, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52142, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52143, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52144, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52145, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52146, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52147, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52148, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52149, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52150, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52151, 2, {0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52152, 2, {0x0, 0x0, 0x1EE0, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {52153, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52154, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52155, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52156, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52157, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52158, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52159, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52160, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52161, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52162, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52163, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52164, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52165, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52166, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52167, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52168, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52169, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52170, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52171, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52172, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52173, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52174, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52175, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52176, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52177, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52178, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52179, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52180, 2, {0x0, 0x0, 0x0, 0x1EE0, 0x220, 0x220, 0x640, 0x540, 0x8A0, 0x3310, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {52181, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52182, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52183, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52184, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52185, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52186, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52187, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52188, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52189, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52190, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52191, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52192, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52193, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52194, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52195, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52196, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52197, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52198, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52199, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52200, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52201, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52202, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52203, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52204, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52205, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52206, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52207, 2, {0x0, 0x3DE0, 0x440, 0x880, 0x1540, 0x2220, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52208, 2, {0x0, 0x8, 0x8, 0x7B88, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {52209, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52210, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52211, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52212, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52213, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52214, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52215, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52216, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52217, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52218, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52219, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52220, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52221, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52222, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52223, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52224, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52225, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52226, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52227, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52228, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52229, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52230, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52231, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52232, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52233, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52234, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52235, 2, {0x0, 0x7BC8, 0x888, 0x1108, 0x2A88, 0x4448, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52236, 2, {0x0, 0x0, 0x8, 0x8, 0x7B88, 0x888, 0x888, 0x1908, 0x1508, 0x2288, 0xCC48, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52237, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52238, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52239, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52240, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52241, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52242, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52243, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52244, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52245, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52246, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52247, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52248, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52249, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52250, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52251, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52252, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52253, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52254, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52255, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52256, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52257, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52258, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52259, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52260, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52261, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52262, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52263, 2, {0x0, 0x8, 0x8, 0x7708, 0x2208, 0x2208, 0x5508, 0x8888, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52264, 2, {0x0, 0x0, 0x10, 0x410, 0x410, 0x3F90, 0x110, 0x21E, 0x610, 0x910, 0x3090, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52265, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52266, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52267, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52268, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52269, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52270, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52271, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52272, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52273, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52274, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52275, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52276, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52277, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52278, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52279, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52280, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52281, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52282, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52283, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52284, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52285, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52286, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52287, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52288, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52289, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52290, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52291, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x80E, 0x1408, 0x6208, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52292, 2, {0x0, 0x0, 0x12, 0x412, 0x412, 0x3F92, 0x112, 0x21E, 0x612, 0x912, 0x3092, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {52293, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52294, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52295, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52296, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52297, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52298, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52299, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52300, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52301, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52302, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52303, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52304, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52305, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52306, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52307, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52308, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52309, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52310, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52311, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {52312, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52313, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {52314, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52315, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52316, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52317, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52318, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52319, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x838, 0x1428, 0x6228, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52320, 2, {0x0, 0x0, 0x10, 0x410, 0x410, 0x3F90, 0x11E, 0x210, 0x610, 0x91E, 0x3090, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52321, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52322, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52323, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52324, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52325, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52326, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52327, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52328, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52329, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52330, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52331, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52332, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52333, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52334, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52335, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52336, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52337, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52338, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52339, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52340, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52341, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52342, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52343, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52344, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52345, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52346, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52347, 2, {0x0, 0x8, 0x8, 0x808, 0x7E0E, 0x808, 0x140E, 0x6208, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52348, 2, {0x0, 0x0, 0x12, 0x412, 0x412, 0x3F92, 0x11E, 0x212, 0x612, 0x91E, 0x3092, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {52349, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52350, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52351, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52352, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52353, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52354, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52355, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52356, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52357, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52358, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52359, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52360, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52361, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52362, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52363, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52364, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52365, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52366, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52367, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {52368, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52369, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {52370, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52371, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52372, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52373, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52374, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52375, 2, {0x0, 0x28, 0x28, 0x828, 0x7E38, 0x828, 0x1438, 0x6228, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52376, 2, {0x0, 0x0, 0x2, 0x402, 0x402, 0x3F82, 0x102, 0x21E, 0x602, 0x902, 0x3082, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {52377, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52378, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52379, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52380, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52381, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52382, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52383, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52384, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52385, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52386, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52387, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52388, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52389, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52390, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52391, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52392, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52393, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52394, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52395, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52396, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52397, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52398, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52399, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52400, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52401, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52402, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52403, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x838, 0x1408, 0x6208, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52404, 2, {0x0, 0x0, 0xA, 0x40A, 0x40A, 0x3F8A, 0x10A, 0x23A, 0x60A, 0x90A, 0x308A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {52405, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52406, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52407, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52408, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52409, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52410, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52411, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52412, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52413, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52414, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52415, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52416, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52417, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52418, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52419, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52420, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52421, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52422, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52423, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {52424, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52425, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {52426, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52427, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52428, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52429, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52430, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52431, 2, {0x0, 0x28, 0x28, 0x828, 0x7E28, 0x8E8, 0x1428, 0x6228, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52432, 2, {0x0, 0x0, 0x2, 0x402, 0x402, 0x3F82, 0x11E, 0x202, 0x602, 0x91E, 0x3082, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {52433, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52434, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52435, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52436, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52437, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52438, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52439, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52440, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52441, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52442, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52443, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52444, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52445, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52446, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52447, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52448, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52449, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52450, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52451, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52452, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52453, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52454, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52455, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52456, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52457, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52458, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52459, 2, {0x0, 0x8, 0x8, 0x808, 0x7E38, 0x808, 0x1438, 0x6208, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52460, 2, {0x0, 0x0, 0xA, 0x40A, 0x40A, 0x3F8A, 0x13A, 0x20A, 0x60A, 0x93A, 0x308A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {52461, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52462, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52463, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52464, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52465, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52466, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52467, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52468, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52469, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52470, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52471, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52472, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52473, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52474, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52475, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52476, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52477, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52478, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52479, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {52480, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52481, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {52482, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52483, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52484, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52485, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52486, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52487, 2, {0x0, 0x28, 0x28, 0x828, 0x7EE8, 0x828, 0x14E8, 0x6228, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52488, 2, {0x0, 0x100, 0x100, 0xFE0, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {52489, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52490, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52491, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52492, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52493, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52494, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52495, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52496, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52497, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52498, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52499, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52500, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52501, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52502, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52503, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52504, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52505, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52506, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52507, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52508, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52509, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52510, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52511, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52512, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52513, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52514, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52515, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52516, 2, {0x0, 0x10, 0x410, 0x410, 0x3F90, 0x110, 0x210, 0x61E, 0x910, 0x3090, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {52517, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52518, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52519, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52520, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52521, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52522, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52523, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52524, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52525, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52526, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52527, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52528, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52529, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52530, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52531, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52532, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52533, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52534, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52535, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52536, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52537, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52538, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52539, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52540, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52541, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52542, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52543, 2, {0x0, 0x208, 0x1F88, 0x208, 0x50E, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52544, 2, {0x0, 0x12, 0x412, 0x412, 0x3F92, 0x112, 0x212, 0x61E, 0x912, 0x3092, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {52545, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52546, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52547, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52548, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52549, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52550, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52551, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52552, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52553, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52554, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52555, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52556, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52557, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52558, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52559, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52560, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52561, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52562, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52563, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52564, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52565, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52566, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52567, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52568, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52569, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52570, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52571, 2, {0x0, 0x228, 0x1FA8, 0x228, 0x538, 0x18A8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52572, 2, {0x0, 0x8, 0x408, 0x408, 0x3F88, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {52573, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52574, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52575, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52576, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52577, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52578, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52579, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52580, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52581, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52582, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52583, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52584, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52585, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52586, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52587, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52588, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52589, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52590, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52591, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52592, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52593, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52594, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52595, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52596, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52597, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52598, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52599, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52600, 2, {0x0, 0x100, 0x100, 0xFE0, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {52601, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52602, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52603, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52604, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52605, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52606, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52607, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52608, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52609, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52610, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52611, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52612, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52613, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52614, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52615, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52616, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52617, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52618, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52619, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52620, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52621, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52622, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52623, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52624, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52625, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52626, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52627, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52628, 2, {0x0, 0x80, 0x80, 0x7F0, 0x20, 0x40, 0xC0, 0x120, 0x610, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {52629, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52630, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52631, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52632, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52633, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52634, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52635, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52636, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52637, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52638, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52639, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52640, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52641, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52642, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52643, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52644, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52645, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52646, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52647, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52648, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52649, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52650, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52651, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52652, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52653, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52654, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52655, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52656, 2, {0x0, 0x408, 0x408, 0x3F88, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {52657, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52658, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52659, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52660, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52661, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52662, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52663, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52664, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52665, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52666, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52667, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52668, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52669, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52670, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52671, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52672, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52673, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52674, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52675, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52676, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52677, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52678, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52679, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52680, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52681, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52682, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52683, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52684, 2, {0x0, 0x40A, 0x40A, 0x3F8A, 0x10A, 0x20A, 0x60A, 0x90A, 0x308A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {52685, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52686, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52687, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52688, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52689, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52690, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52691, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52692, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52693, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52694, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52695, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52696, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52697, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52698, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52699, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52700, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52701, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52702, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52703, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52704, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52705, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52706, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52707, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52708, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52709, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52710, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52711, 2, {0x428, 0x3F28, 0x428, 0xA28, 0x3128, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52712, 2, {0x0, 0x408, 0x408, 0x3F88, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {52713, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52714, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52715, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52716, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52717, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52718, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52719, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52720, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52721, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52722, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52723, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52724, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52725, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52726, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52727, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52728, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52729, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52730, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52731, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52732, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52733, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52734, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52735, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52736, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52737, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52738, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52739, 2, {0x408, 0x3F08, 0x408, 0xA08, 0x3108, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52740, 2, {0x0, 0x80, 0x80, 0x7F0, 0x20, 0x40, 0xC0, 0x120, 0x610, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {52741, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52742, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52743, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52744, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52745, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52746, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52747, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52748, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52749, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52750, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52751, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52752, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52753, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52754, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52755, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52756, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52757, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52758, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52759, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52760, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52761, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52762, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52763, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52764, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52765, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52766, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52767, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52768, 2, {0x0, 0x100, 0x100, 0xFE0, 0x40, 0x80, 0x180, 0x240, 0xC20, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {52769, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {52770, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52771, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52772, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52773, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52774, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52775, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {52776, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52777, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52778, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52779, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52780, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {52781, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52782, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52783, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {52784, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {52785, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {52786, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {52787, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {52788, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {52789, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {52790, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52791, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {52792, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {52793, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {52794, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {52795, 2, {0x0, 0x100, 0xFC0, 0x100, 0x280, 0xC40, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {52796, 2, {0x0, 0x8, 0x408, 0x408, 0x3F88, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {52797, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52798, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52799, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52800, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52801, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52802, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52803, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52804, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52805, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52806, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52807, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52808, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52809, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52810, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52811, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52812, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52813, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52814, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52815, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52816, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52817, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52818, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52819, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52820, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52821, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52822, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52823, 2, {0x0, 0x208, 0x1F88, 0x208, 0x508, 0x1888, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52824, 2, {0x0, 0x0, 0x8, 0x408, 0x408, 0x3F88, 0x108, 0x208, 0x608, 0x908, 0x3088, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52825, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52826, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52827, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52828, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52829, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52830, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52831, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52832, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52833, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52834, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52835, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52836, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52837, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52838, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52839, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52840, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52841, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52842, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52843, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52844, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52845, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52846, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52847, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52848, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52849, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52850, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52851, 2, {0x0, 0x8, 0x8, 0x808, 0x7E08, 0x808, 0x1408, 0x6208, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52852, 2, {0x0, 0x0, 0x10, 0x10, 0x1F90, 0x90, 0x90, 0x9E, 0x1F10, 0x110, 0x210, 0x410, 0x1810, 0x10, 0x10, 0x0}},
    {52853, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52854, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52855, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52856, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52857, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52858, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52859, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52860, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52861, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52862, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52863, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52864, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52865, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52866, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52867, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52868, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52869, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52870, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52871, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52872, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52873, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52874, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52875, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52876, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52877, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52878, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52879, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E0E, 0x408, 0x1808, 0x6008, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52880, 2, {0x0, 0x0, 0x12, 0x12, 0x1F92, 0x92, 0x92, 0x9E, 0x1F12, 0x112, 0x212, 0x412, 0x1812, 0x12, 0x12, 0x0}},
    {52881, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52882, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52883, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52884, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52885, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52886, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52887, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52888, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52889, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52890, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52891, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52892, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52893, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52894, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52895, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52896, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52897, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52898, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52899, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {52900, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52901, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {52902, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52903, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52904, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52905, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52906, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52907, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7E38, 0x428, 0x1828, 0x6028, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52908, 2, {0x0, 0x0, 0x10, 0x10, 0x1F90, 0x90, 0x9E, 0x90, 0x1F10, 0x11E, 0x210, 0x410, 0x1810, 0x10, 0x10, 0x0}},
    {52909, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52910, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52911, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52912, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52913, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {52914, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {52915, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52916, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52917, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52918, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52919, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52920, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52921, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52922, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52923, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {52924, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52925, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52926, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52927, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52928, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52929, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52930, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52931, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52932, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52933, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52934, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52935, 2, {0x0, 0x8, 0x7E08, 0x208, 0x20E, 0x7E08, 0x40E, 0x1808, 0x6008, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52936, 2, {0x0, 0x0, 0x12, 0x12, 0x1F92, 0x92, 0x9E, 0x92, 0x1F12, 0x11E, 0x212, 0x412, 0x1812, 0x12, 0x12, 0x0}},
    {52937, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52938, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52939, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52940, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52941, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52942, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52943, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52944, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52945, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52946, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52947, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52948, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {52949, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52950, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52951, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52952, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52953, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52954, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {52955, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {52956, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52957, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {52958, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52959, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52960, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52961, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52962, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52963, 2, {0x0, 0x28, 0x7E28, 0x228, 0x238, 0x7E28, 0x438, 0x1828, 0x6028, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52964, 2, {0x0, 0x0, 0x2, 0x2, 0x1F82, 0x82, 0x82, 0x9E, 0x1F02, 0x102, 0x202, 0x402, 0x1802, 0x2, 0x2, 0x0}},
    {52965, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {52966, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52967, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {52968, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {52969, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52970, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52971, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {52972, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {52973, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {52974, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {52975, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {52976, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {52977, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {52978, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {52979, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {52980, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {52981, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {52982, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {52983, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {52984, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {52985, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {52986, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {52987, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {52988, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {52989, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {52990, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {52991, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E38, 0x408, 0x1808, 0x6008, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {52992, 2, {0x0, 0x0, 0xA, 0xA, 0x1F8A, 0x8A, 0x8A, 0xBA, 0x1F0A, 0x10A, 0x20A, 0x40A, 0x180A, 0xA, 0xA, 0x0}},
    {52993, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {52994, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {52995, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {52996, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {52997, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {52998, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {52999, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53000, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53001, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53002, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53003, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53004, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53005, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53006, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53007, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53008, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53009, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53010, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53011, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {53012, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53013, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {53014, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53015, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53016, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53017, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53018, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53019, 2, {0x0, 0x28, 0x7E28, 0x228, 0x228, 0x7EE8, 0x428, 0x1828, 0x6028, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53020, 2, {0x0, 0x0, 0x2, 0x2, 0x1F82, 0x82, 0x9E, 0x82, 0x1F02, 0x11E, 0x202, 0x402, 0x1802, 0x2, 0x2, 0x0}},
    {53021, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53022, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53023, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53024, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53025, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53026, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53027, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53028, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53029, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53030, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53031, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53032, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53033, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53034, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53035, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53036, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53037, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53038, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53039, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53040, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53041, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53042, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53043, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53044, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53045, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53046, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53047, 2, {0x0, 0x8, 0x7E08, 0x208, 0x238, 0x7E08, 0x438, 0x1808, 0x6008, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53048, 2, {0x0, 0x0, 0xA, 0xA, 0x1F8A, 0x8A, 0xBA, 0x8A, 0x1F0A, 0x13A, 0x20A, 0x40A, 0x180A, 0xA, 0xA, 0x0}},
    {53049, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53050, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53051, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53052, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53053, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53054, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53055, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53056, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53057, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53058, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53059, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53060, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53061, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53062, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53063, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53064, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53065, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53066, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53067, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {53068, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53069, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {53070, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53071, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53072, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53073, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53074, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53075, 2, {0x0, 0x28, 0x7E28, 0x228, 0x2E8, 0x7E28, 0x4E8, 0x1828, 0x6028, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53076, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x208, 0x208, 0x200, 0x200, 0x7FFC, 0x0, 0x0}},
    {53077, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53078, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53079, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53080, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53081, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53082, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53083, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53084, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53085, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53086, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53087, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53088, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53089, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53090, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53091, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53092, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53093, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53094, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53095, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53096, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53097, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53098, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53099, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53100, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53101, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53102, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53103, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x110, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53104, 2, {0x0, 0x0, 0x10, 0x10, 0x3FD0, 0x50, 0x50, 0x3FDE, 0x450, 0x450, 0x410, 0x7FD0, 0x10, 0x10, 0x10, 0x0}},
    {53105, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53106, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53107, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53108, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53109, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53110, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53111, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53112, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53113, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53114, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53115, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53116, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53117, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53118, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53119, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53120, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53121, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53122, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53123, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53124, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53125, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53126, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53127, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53128, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53129, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53130, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53131, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F8E, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53132, 2, {0x0, 0x0, 0x12, 0x12, 0x3FD2, 0x52, 0x52, 0x3FDE, 0x452, 0x452, 0x412, 0x7FD2, 0x12, 0x12, 0x12, 0x0}},
    {53133, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53134, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53135, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53136, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53137, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53138, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53139, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53140, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53141, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53142, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53143, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53144, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53145, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53146, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53147, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53148, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53149, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53150, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53151, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53152, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53153, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53154, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53155, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53156, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53157, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53158, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53159, 2, {0x0, 0x28, 0x3FA8, 0xA8, 0x3FB8, 0xA8, 0x4A8, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53160, 2, {0x0, 0x0, 0x4, 0x4, 0x3FC4, 0x44, 0x44, 0x3FC4, 0x444, 0x444, 0x404, 0x7FF4, 0x4, 0x4, 0x4, 0x0}},
    {53161, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53162, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53163, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53164, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53165, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53166, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53167, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53168, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53169, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53170, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53171, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53172, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53173, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53174, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53175, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53176, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53177, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53178, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53179, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53180, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53181, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53182, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53183, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53184, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53185, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53186, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53187, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x488, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53188, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x888, 0x888, 0x880, 0x880, 0x7FFC, 0x0, 0x0}},
    {53189, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53190, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53191, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53192, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53193, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53194, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53195, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53196, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53197, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53198, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53199, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53200, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53201, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53202, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53203, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53204, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53205, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53206, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53207, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53208, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53209, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53210, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53211, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53212, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53213, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53214, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53215, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x490, 0x480, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53216, 2, {0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {53217, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53218, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53219, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53220, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53221, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53222, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53223, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53224, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53225, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53226, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53227, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53228, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53229, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53230, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53231, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53232, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53233, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53234, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53235, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53236, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53237, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53238, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53239, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53240, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53241, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53242, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53243, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53244, 2, {0x0, 0x8, 0x8, 0x3FC8, 0x48, 0x48, 0x3FC8, 0x48, 0x48, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {53245, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53246, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53247, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53248, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53249, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53250, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53251, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53252, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53253, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53254, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53255, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53256, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53257, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53258, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53259, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53260, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53261, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53262, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53263, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53264, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53265, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53266, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53267, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53268, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53269, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53270, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53271, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53272, 2, {0x0, 0xA, 0xA, 0x3FCA, 0x4A, 0x4A, 0x3FCA, 0x4A, 0x4A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {53273, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53274, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53275, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53276, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53277, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53278, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53279, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53280, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53281, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53282, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53283, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53284, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53285, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53286, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53287, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53288, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53289, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53290, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53291, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53292, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53293, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53294, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53295, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53296, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53297, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53298, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53299, 2, {0x28, 0x3FA8, 0xA8, 0x3FA8, 0xA8, 0xA8, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53300, 2, {0x0, 0x8, 0x8, 0x3FC8, 0x48, 0x48, 0x3FC8, 0x48, 0x48, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {53301, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53302, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53303, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53304, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53305, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53306, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53307, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53308, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53309, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53310, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53311, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53312, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53313, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53314, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53315, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53316, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53317, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53318, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53319, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53320, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53321, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53322, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53323, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53324, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53325, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53326, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53327, 2, {0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53328, 2, {0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {53329, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53330, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53331, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53332, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53333, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53334, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53335, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53336, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53337, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53338, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53339, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53340, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53341, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53342, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53343, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53344, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53345, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53346, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53347, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53348, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53349, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53350, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53351, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53352, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53353, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53354, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53355, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53356, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF8, 0x8, 0x8, 0x3FF8, 0x8, 0x8, 0x8, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {53357, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53358, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53359, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53360, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53361, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53362, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53363, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53364, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53365, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53366, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53367, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53368, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53369, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53370, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53371, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53372, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53373, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53374, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53375, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53376, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53377, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53378, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53379, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53380, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53381, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53382, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53383, 2, {0x0, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53384, 2, {0x0, 0x0, 0x8, 0x8, 0x3FC8, 0x48, 0x48, 0x3FC8, 0x48, 0x48, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {53385, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53386, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53387, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53388, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53389, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53390, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53391, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53392, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53393, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53394, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53395, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53396, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53397, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53398, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53399, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53400, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53401, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53402, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53403, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53404, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53405, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53406, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53407, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53408, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53409, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53410, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53411, 2, {0x0, 0x8, 0x3F88, 0x88, 0x3F88, 0x88, 0x88, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53412, 2, {0x0, 0x0, 0x8, 0x8, 0x1F88, 0x88, 0x88, 0x88, 0x1F08, 0x108, 0x208, 0x408, 0x1808, 0x8, 0x8, 0x0}},
    {53413, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53414, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53415, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53416, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53417, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53418, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53419, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53420, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53421, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53422, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53423, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53424, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53425, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53426, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53427, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53428, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53429, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53430, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53431, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53432, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53433, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53434, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53435, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53436, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53437, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53438, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53439, 2, {0x0, 0x8, 0x7E08, 0x208, 0x208, 0x7E08, 0x408, 0x1808, 0x6008, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53440, 2, {0x0, 0x0, 0x10, 0x10, 0x3F90, 0x2010, 0x2010, 0x3F9E, 0x2010, 0x2010, 0x3F90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53441, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53442, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53443, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53444, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53445, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53446, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53447, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53448, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53449, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53450, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53451, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53452, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53453, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53454, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53455, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53456, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53457, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53458, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53459, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53460, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53461, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53462, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53463, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53464, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53465, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53466, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53467, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E0E, 0x4008, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53468, 2, {0x0, 0x0, 0x12, 0x12, 0x3F92, 0x2012, 0x2012, 0x3F9E, 0x2012, 0x2012, 0x3F92, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {53469, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53470, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53471, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53472, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53473, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53474, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53475, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53476, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53477, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53478, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53479, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53480, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53481, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53482, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53483, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53484, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53485, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53486, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53487, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {53488, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53489, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {53490, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53491, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53492, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53493, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53494, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53495, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7E38, 0x4028, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53496, 2, {0x0, 0x0, 0x10, 0x10, 0x3F90, 0x2010, 0x201E, 0x3F90, 0x2010, 0x201E, 0x3F90, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53497, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53498, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53499, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53500, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53501, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53502, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53503, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53504, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53505, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53506, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53507, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53508, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53509, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53510, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53511, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53512, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53513, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53514, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53515, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53516, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53517, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53518, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53519, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53520, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53521, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53522, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53523, 2, {0x0, 0x8, 0x8, 0x7E08, 0x400E, 0x7E08, 0x400E, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53524, 2, {0x0, 0x0, 0x12, 0x12, 0x3F92, 0x2012, 0x201E, 0x3F92, 0x2012, 0x201E, 0x3F92, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {53525, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53526, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53527, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53528, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53529, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53530, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53531, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53532, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53533, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53534, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53535, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53536, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53537, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53538, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53539, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53540, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53541, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53542, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53543, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {53544, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53545, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {53546, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53547, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53548, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53549, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53550, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53551, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4038, 0x7E28, 0x4038, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53552, 2, {0x0, 0x0, 0x2, 0x2, 0x3F82, 0x2002, 0x2002, 0x3F9E, 0x2002, 0x2002, 0x3F82, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {53553, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53554, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53555, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53556, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53557, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53558, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53559, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53560, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53561, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53562, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53563, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53564, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53565, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53566, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53567, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53568, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53569, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53570, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53571, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53572, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53573, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53574, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53575, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53576, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53577, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53578, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53579, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E38, 0x4008, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53580, 2, {0x0, 0x0, 0xA, 0xA, 0x3F8A, 0x200A, 0x200A, 0x3FBA, 0x200A, 0x200A, 0x3F8A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {53581, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53582, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53583, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53584, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53585, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53586, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53587, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53588, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53589, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53590, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53591, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53592, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53593, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53594, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53595, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53596, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53597, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53598, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53599, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {53600, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53601, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {53602, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53603, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53604, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53605, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53606, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53607, 2, {0x0, 0x28, 0x28, 0x7E28, 0x4028, 0x7EE8, 0x4028, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53608, 2, {0x0, 0x0, 0x2, 0x2, 0x3F82, 0x2002, 0x201E, 0x3F82, 0x2002, 0x201E, 0x3F82, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {53609, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53610, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53611, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53612, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53613, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53614, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53615, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53616, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53617, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53618, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53619, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53620, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53621, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53622, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53623, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53624, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53625, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53626, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53627, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53628, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53629, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53630, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53631, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53632, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53633, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53634, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53635, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4038, 0x7E08, 0x4038, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53636, 2, {0x0, 0x0, 0xA, 0xA, 0x3F8A, 0x200A, 0x203A, 0x3F8A, 0x200A, 0x203A, 0x3F8A, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {53637, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53638, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53639, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53640, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53641, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53642, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53643, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53644, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53645, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53646, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53647, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53648, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53649, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53650, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53651, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53652, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53653, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53654, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53655, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {53656, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53657, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {53658, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53659, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53660, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53661, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53662, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53663, 2, {0x0, 0x28, 0x28, 0x7E28, 0x40E8, 0x7E28, 0x40E8, 0x7E28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53664, 2, {0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {53665, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53666, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53667, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53668, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53669, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53670, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53671, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53672, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53673, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53674, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53675, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53676, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53677, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53678, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53679, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53680, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53681, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53682, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53683, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53684, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53685, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53686, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53687, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53688, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53689, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53690, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53691, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53692, 2, {0x0, 0x10, 0x3FD0, 0x2010, 0x2010, 0x3FD0, 0x2010, 0x201E, 0x3FD0, 0x10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {53693, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53694, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53695, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53696, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53697, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53698, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53699, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53700, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53701, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53702, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53703, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53704, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53705, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53706, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53707, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53708, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53709, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53710, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53711, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53712, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53713, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53714, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53715, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53716, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53717, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53718, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53719, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F0E, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53720, 2, {0x0, 0x12, 0x3FD2, 0x2012, 0x2012, 0x3FD2, 0x2012, 0x201E, 0x3FD2, 0x12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {53721, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53722, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53723, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53724, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53725, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53726, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53727, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53728, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53729, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53730, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53731, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53732, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53733, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53734, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53735, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53736, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53737, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53738, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53739, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53740, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53741, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53742, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53743, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53744, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53745, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53746, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53747, 2, {0x0, 0x28, 0x7F28, 0x4028, 0x7F38, 0x4028, 0x7F28, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53748, 2, {0x0, 0x8, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {53749, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53750, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53751, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53752, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53753, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53754, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53755, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53756, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53757, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53758, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53759, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53760, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53761, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53762, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53763, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53764, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53765, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53766, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53767, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53768, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53769, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53770, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53771, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53772, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53773, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53774, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53775, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53776, 2, {0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {53777, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53778, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53779, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53780, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53781, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53782, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53783, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53784, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53785, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53786, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53787, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53788, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53789, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53790, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53791, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53792, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53793, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53794, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53795, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53796, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53797, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53798, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53799, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53800, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53801, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53802, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53803, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53804, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {53805, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53806, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53807, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53808, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53809, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53810, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53811, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53812, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53813, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53814, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53815, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53816, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53817, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53818, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53819, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53820, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53821, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53822, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53823, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53824, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53825, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53826, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53827, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53828, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53829, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53830, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53831, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53832, 2, {0x0, 0x8, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {53833, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53834, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53835, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53836, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53837, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53838, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53839, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53840, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53841, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53842, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53843, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53844, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53845, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53846, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53847, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53848, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53849, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53850, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53851, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53852, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53853, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53854, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53855, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53856, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53857, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53858, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53859, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53860, 2, {0x0, 0xA, 0x3FCA, 0x200A, 0x200A, 0x3FCA, 0x200A, 0x200A, 0x3FCA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {53861, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53862, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53863, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53864, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53865, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53866, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53867, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53868, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53869, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53870, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53871, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53872, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53873, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53874, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53875, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53876, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53877, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53878, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53879, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53880, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53881, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53882, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53883, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53884, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53885, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53886, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53887, 2, {0x7FA8, 0x4028, 0x7FA8, 0x4028, 0x7FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53888, 2, {0x0, 0x8, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {53889, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {53890, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53891, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53892, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {53893, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53894, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53895, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53896, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53897, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53898, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53899, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53900, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {53901, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53902, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53903, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {53904, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53905, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53906, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {53907, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53908, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53909, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53910, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53911, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53912, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53913, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53914, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53915, 2, {0x7F88, 0x4008, 0x7F88, 0x4008, 0x7F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {53916, 2, {0x0, 0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {53917, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53918, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53919, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53920, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53921, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53922, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53923, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53924, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53925, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53926, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53927, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53928, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53929, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53930, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53931, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53932, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53933, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53934, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53935, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53936, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53937, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53938, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53939, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53940, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53941, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53942, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53943, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53944, 2, {0x0, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x2000, 0x2000, 0x3FF8, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {53945, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {53946, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53947, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {53948, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53949, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {53950, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {53951, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {53952, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53953, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53954, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53955, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53956, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {53957, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53958, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53959, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {53960, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {53961, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {53962, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {53963, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {53964, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {53965, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {53966, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53967, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {53968, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {53969, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {53970, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {53971, 2, {0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {53972, 2, {0x0, 0x8, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x2008, 0x2008, 0x3FC8, 0x8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {53973, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {53974, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {53975, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {53976, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {53977, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {53978, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {53979, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {53980, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {53981, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {53982, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {53983, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {53984, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {53985, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {53986, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {53987, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {53988, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {53989, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {53990, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {53991, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {53992, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {53993, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {53994, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {53995, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {53996, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {53997, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {53998, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {53999, 2, {0x0, 0x8, 0x7F08, 0x4008, 0x7F08, 0x4008, 0x7F08, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54000, 2, {0x0, 0x0, 0x8, 0x8, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x2008, 0x2008, 0x3F88, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54001, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54002, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54003, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54004, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54005, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54006, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54007, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54008, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54009, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54010, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54011, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54012, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54013, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54014, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54015, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54016, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54017, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54018, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54019, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54020, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54021, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54022, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54023, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54024, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54025, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54026, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54027, 2, {0x0, 0x8, 0x8, 0x7E08, 0x4008, 0x7E08, 0x4008, 0x7E08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54028, 2, {0x0, 0x0, 0x10, 0x10, 0x7FD0, 0x1110, 0x1110, 0x111E, 0x1110, 0x1110, 0x7FD0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54029, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54030, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54031, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54032, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54033, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54034, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54035, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54036, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54037, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54038, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54039, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54040, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54041, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54042, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54043, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54044, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54045, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54046, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54047, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54048, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54049, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54050, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54051, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54052, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54053, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54054, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54055, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x120E, 0x1208, 0x7F88, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54056, 2, {0x0, 0x0, 0x12, 0x12, 0x7FD2, 0x1112, 0x1112, 0x111E, 0x1112, 0x1112, 0x7FD2, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {54057, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54058, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54059, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54060, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54061, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54062, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54063, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54064, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54065, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54066, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54067, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54068, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54069, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54070, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54071, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54072, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54073, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54074, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54075, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54076, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54077, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54078, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54079, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54080, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54081, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54082, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54083, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x1238, 0x1228, 0x7FA8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54084, 2, {0x0, 0x0, 0x10, 0x10, 0x7FD0, 0x1110, 0x111E, 0x1110, 0x1110, 0x111E, 0x7FD0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54085, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54086, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54087, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54088, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54089, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54090, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54091, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54092, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54093, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54094, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54095, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54096, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54097, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54098, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54099, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54100, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54101, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54102, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54103, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54104, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54105, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54106, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54107, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54108, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54109, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54110, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54111, 2, {0x0, 0x8, 0x8, 0x7F88, 0x120E, 0x1208, 0x120E, 0x7F88, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54112, 2, {0x0, 0x0, 0x12, 0x12, 0x7FD2, 0x1112, 0x111E, 0x1112, 0x1112, 0x111E, 0x7FD2, 0x12, 0x12, 0x12, 0x12, 0x0}},
    {54113, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54114, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54115, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54116, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54117, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54118, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54119, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54120, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54121, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54122, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54123, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54124, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54125, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54126, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54127, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54128, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54129, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54130, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54131, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54132, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54133, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54134, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54135, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54136, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54137, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54138, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54139, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1238, 0x1228, 0x1238, 0x7FA8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54140, 2, {0x0, 0x0, 0x2, 0x2, 0x7FC2, 0x1102, 0x1102, 0x111E, 0x1102, 0x1102, 0x7FC2, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {54141, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54142, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54143, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54144, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54145, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54146, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54147, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54148, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54149, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54150, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54151, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54152, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54153, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54154, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54155, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54156, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54157, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54158, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54159, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54160, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54161, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54162, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54163, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54164, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54165, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54166, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54167, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1238, 0x1208, 0x7F88, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54168, 2, {0x0, 0x0, 0xA, 0xA, 0x7FCA, 0x110A, 0x110A, 0x113A, 0x110A, 0x110A, 0x7FCA, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {54169, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54170, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54171, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54172, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54173, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54174, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54175, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54176, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54177, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54178, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54179, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54180, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54181, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54182, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54183, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54184, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54185, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54186, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54187, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54188, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54189, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54190, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54191, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54192, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54193, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54194, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54195, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x1228, 0x12E8, 0x1228, 0x7FA8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54196, 2, {0x0, 0x0, 0x2, 0x2, 0x7FC2, 0x1102, 0x111E, 0x1102, 0x1102, 0x111E, 0x7FC2, 0x2, 0x2, 0x2, 0x2, 0x0}},
    {54197, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54198, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54199, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54200, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54201, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54202, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54203, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54204, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54205, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54206, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54207, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54208, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54209, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54210, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54211, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54212, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54213, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54214, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54215, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54216, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54217, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54218, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54219, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54220, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54221, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54222, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54223, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1238, 0x1208, 0x1238, 0x7F88, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54224, 2, {0x0, 0x0, 0xA, 0xA, 0x7FCA, 0x110A, 0x113A, 0x110A, 0x110A, 0x113A, 0x7FCA, 0xA, 0xA, 0xA, 0xA, 0x0}},
    {54225, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54226, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54227, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54228, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54229, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54230, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54231, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54232, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54233, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54234, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54235, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54236, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54237, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54238, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54239, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54240, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54241, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54242, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54243, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54244, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54245, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54246, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54247, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54248, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54249, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54250, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54251, 2, {0x0, 0x28, 0x28, 0x7FA8, 0x12E8, 0x1228, 0x12E8, 0x7FA8, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54252, 2, {0x0, 0x0, 0x3FF8, 0x820, 0x820, 0x820, 0x820, 0x3FF8, 0x0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {54253, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54254, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54255, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54256, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54257, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54258, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54259, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54260, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54261, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54262, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54263, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54264, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54265, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54266, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54267, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54268, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54269, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54270, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54271, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54272, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54273, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54274, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54275, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54276, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54277, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54278, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54279, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54280, 2, {0x0, 0x10, 0x10, 0x10, 0x7FD0, 0x1110, 0x1110, 0x111E, 0x1110, 0x7FD0, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {54281, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54282, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54283, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54284, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54285, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54286, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54287, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54288, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54289, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54290, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54291, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54292, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54293, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54294, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54295, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54296, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54297, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54298, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54299, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54300, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54301, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54302, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54303, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54304, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54305, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54306, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54307, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x110E, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54308, 2, {0x0, 0x12, 0x12, 0x12, 0x7FD2, 0x1112, 0x1112, 0x111E, 0x1112, 0x7FD2, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {54309, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54310, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54311, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54312, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54313, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54314, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54315, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54316, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54317, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54318, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54319, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54320, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54321, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54322, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54323, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54324, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54325, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54326, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54327, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54328, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54329, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54330, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54331, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54332, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54333, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54334, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54335, 2, {0x0, 0x28, 0x3FA8, 0x1128, 0x1138, 0x3FA8, 0x428, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54336, 2, {0x0, 0x8, 0x8, 0x8, 0x7FC8, 0x1108, 0x1108, 0x1108, 0x1108, 0x7FC8, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {54337, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54338, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54339, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54340, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54341, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54342, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54343, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54344, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54345, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54346, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54347, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54348, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54349, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54350, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54351, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54352, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54353, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54354, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54355, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54356, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54357, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54358, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54359, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54360, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54361, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54362, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54363, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x408, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54364, 2, {0x0, 0x0, 0x3FF8, 0x820, 0x820, 0x820, 0x820, 0x3FF8, 0x0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {54365, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54366, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54367, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54368, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54369, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54370, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54371, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54372, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54373, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54374, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54375, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54376, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54377, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54378, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54379, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54380, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54381, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54382, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54383, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54384, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54385, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54386, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54387, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54388, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54389, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54390, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54391, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54392, 2, {0x0, 0x0, 0x3FF8, 0x820, 0x820, 0x820, 0x820, 0x3FF8, 0x0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {54393, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54394, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54395, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54396, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54397, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54398, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54399, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54400, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54401, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54402, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54403, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54404, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54405, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54406, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54407, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54408, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54409, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54410, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54411, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54412, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54413, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54414, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54415, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54416, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54417, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54418, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54419, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54420, 2, {0x0, 0x8, 0x8, 0x7FC8, 0x1108, 0x1108, 0x1108, 0x1108, 0x7FC8, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {54421, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54422, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54423, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54424, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54425, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54426, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54427, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54428, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54429, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54430, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54431, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54432, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54433, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54434, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54435, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54436, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54437, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54438, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54439, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54440, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54441, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54442, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54443, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54444, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54445, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54446, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54447, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54448, 2, {0x0, 0xA, 0xA, 0x7FCA, 0x110A, 0x110A, 0x110A, 0x110A, 0x7FCA, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {54449, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54450, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54451, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54452, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54453, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54454, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54455, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54456, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54457, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54458, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54459, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54460, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54461, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54462, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54463, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54464, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54465, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54466, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54467, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54468, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54469, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54470, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54471, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54472, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54473, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54474, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54475, 2, {0x28, 0x3FA8, 0x1128, 0x1128, 0x3FA8, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54476, 2, {0x0, 0x8, 0x8, 0x7FC8, 0x1108, 0x1108, 0x1108, 0x1108, 0x7FC8, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {54477, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54478, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54479, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54480, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54481, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54482, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54483, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54484, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54485, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54486, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54487, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54488, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54489, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54490, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54491, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54492, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54493, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54494, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54495, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54496, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54497, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54498, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54499, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54500, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54501, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54502, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54503, 2, {0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54504, 2, {0x0, 0x0, 0x3FF8, 0x820, 0x820, 0x820, 0x820, 0x3FF8, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {54505, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54506, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54507, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54508, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54509, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54510, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54511, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54512, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54513, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54514, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54515, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54516, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54517, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54518, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54519, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54520, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54521, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54522, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54523, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54524, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54525, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54526, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54527, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54528, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54529, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54530, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54531, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54532, 2, {0x0, 0x0, 0x3FF8, 0x820, 0x820, 0x820, 0x820, 0x3FF8, 0x0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {54533, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54534, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54535, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54536, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54537, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54538, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54539, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54540, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54541, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54542, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54543, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54544, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54545, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54546, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54547, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54548, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54549, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54550, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54551, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54552, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54553, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54554, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54555, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54556, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54557, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54558, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54559, 2, {0x0, 0x1FF0, 0x440, 0x440, 0x440, 0x1FF0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54560, 2, {0x0, 0x8, 0x8, 0x8, 0x7FC8, 0x1108, 0x1108, 0x1108, 0x1108, 0x7FC8, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {54561, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54562, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54563, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54564, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54565, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54566, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54567, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54568, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54569, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54570, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54571, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54572, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54573, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54574, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54575, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54576, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54577, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54578, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54579, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54580, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54581, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54582, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54583, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54584, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54585, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54586, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54587, 2, {0x0, 0x8, 0x3F88, 0x1108, 0x1108, 0x3F88, 0x8, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54588, 2, {0x0, 0x0, 0x8, 0x8, 0x7FC8, 0x1108, 0x1108, 0x1108, 0x1108, 0x1108, 0x7FC8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54589, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54590, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54591, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54592, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54593, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54594, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54595, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54596, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54597, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54598, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54599, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54600, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54601, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54602, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54603, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54604, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54605, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54606, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54607, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54608, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54609, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54610, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54611, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54612, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54613, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54614, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54615, 2, {0x0, 0x8, 0x8, 0x7F88, 0x1208, 0x1208, 0x1208, 0x7F88, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54616, 2, {0x0, 0x0, 0x410, 0x410, 0x3F90, 0xE10, 0x1110, 0x209E, 0x2090, 0x2090, 0x1110, 0xE10, 0x10, 0x10, 0x10, 0x0}},
    {54617, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54618, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54619, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54620, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x8, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54621, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54622, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54623, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54624, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54625, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54626, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54627, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54628, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54629, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54630, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54631, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54632, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54633, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54634, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54635, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54636, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54637, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54638, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54639, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54640, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54641, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54642, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54643, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x220E, 0x2208, 0x1C08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54644, 2, {0x0, 0x0, 0x412, 0x412, 0x3F92, 0xE12, 0x1112, 0x209E, 0x2092, 0x2092, 0x1112, 0xE12, 0x12, 0x12, 0x12, 0x0}},
    {54645, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54646, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54647, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54648, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54649, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54650, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54651, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54652, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54653, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54654, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54655, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54656, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54657, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54658, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54659, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54660, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54661, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54662, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54663, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54664, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54665, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54666, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54667, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54668, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54669, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54670, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54671, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x2238, 0x2228, 0x1C28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54672, 2, {0x0, 0x0, 0x410, 0x410, 0x3F90, 0xE10, 0x111E, 0x2090, 0x2090, 0x209E, 0x1110, 0xE10, 0x10, 0x10, 0x10, 0x0}},
    {54673, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54674, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54675, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54676, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54677, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54678, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54679, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54680, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54681, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54682, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54683, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54684, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54685, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54686, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54687, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54688, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54689, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54690, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54691, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54692, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54693, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54694, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54695, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54696, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54697, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54698, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54699, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x220E, 0x2208, 0x220E, 0x1C08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54700, 2, {0x0, 0x0, 0x412, 0x412, 0x3F92, 0xE12, 0x111E, 0x2092, 0x2092, 0x209E, 0x1112, 0xE12, 0x12, 0x12, 0x12, 0x0}},
    {54701, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54702, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54703, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54704, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54705, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54706, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54707, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54708, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54709, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54710, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54711, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54712, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54713, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54714, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54715, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54716, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54717, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54718, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54719, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54720, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54721, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54722, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54723, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54724, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54725, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54726, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54727, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2238, 0x2228, 0x2238, 0x1C28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54728, 2, {0x0, 0x0, 0x402, 0x402, 0x3F82, 0xE02, 0x1102, 0x209E, 0x2082, 0x2082, 0x1102, 0xE02, 0x2, 0x2, 0x2, 0x0}},
    {54729, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54730, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54731, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54732, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54733, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54734, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54735, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54736, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54737, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54738, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54739, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54740, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54741, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54742, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54743, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54744, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54745, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54746, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54747, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54748, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54749, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54750, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54751, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54752, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54753, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54754, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54755, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2238, 0x2208, 0x1C08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54756, 2, {0x0, 0x0, 0x40A, 0x40A, 0x3F8A, 0xE0A, 0x110A, 0x20BA, 0x208A, 0x208A, 0x110A, 0xE0A, 0xA, 0xA, 0xA, 0x0}},
    {54757, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54758, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54759, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54760, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54761, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54762, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54763, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54764, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54765, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54766, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54767, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54768, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54769, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54770, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54771, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54772, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54773, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54774, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54775, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54776, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54777, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54778, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54779, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54780, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54781, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54782, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54783, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x2228, 0x22E8, 0x2228, 0x1C28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54784, 2, {0x0, 0x0, 0x402, 0x402, 0x3F82, 0xE02, 0x111E, 0x2082, 0x2082, 0x209E, 0x1102, 0xE02, 0x2, 0x2, 0x2, 0x0}},
    {54785, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {54786, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54787, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54788, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {54789, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54790, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54791, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54792, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54793, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54794, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54795, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54796, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {54797, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54798, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54799, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54800, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54801, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54802, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {54803, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54804, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54805, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54806, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54807, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54808, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54809, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54810, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54811, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2238, 0x2208, 0x2238, 0x1C08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54812, 2, {0x0, 0x0, 0x40A, 0x40A, 0x3F8A, 0xE0A, 0x113A, 0x208A, 0x208A, 0x20BA, 0x110A, 0xE0A, 0xA, 0xA, 0xA, 0x0}},
    {54813, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54814, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54815, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x0, 0x3E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54816, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x28, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54817, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54818, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54819, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54820, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54821, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54822, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54823, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54824, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54825, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54826, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54827, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {54828, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54829, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54830, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54831, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x8, 0x8, 0x8, 0x34, 0xC2, 0x0}},
    {54832, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54833, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x0, 0x1F0, 0x208, 0x208, 0x1F0, 0x0}},
    {54834, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54835, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54836, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54837, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54838, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54839, 2, {0x0, 0x828, 0x7F28, 0x1C28, 0x22E8, 0x2228, 0x22E8, 0x1C28, 0x28, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54840, 2, {0x0, 0x100, 0x100, 0xFE0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x100, 0x100, 0x100, 0x7FFC, 0x0, 0x0}},
    {54841, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54842, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54843, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54844, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54845, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54846, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54847, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54848, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54849, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54850, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54851, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54852, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54853, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54854, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54855, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54856, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54857, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54858, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54859, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54860, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54861, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54862, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54863, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54864, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54865, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54866, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54867, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x100, 0x100, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54868, 2, {0x0, 0x10, 0x410, 0x410, 0x3F90, 0x10, 0x1F10, 0x209E, 0x2090, 0x1F10, 0x410, 0x410, 0x7FD0, 0x10, 0x10, 0x0}},
    {54869, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54870, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54871, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54872, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54873, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54874, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54875, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54876, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54877, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54878, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54879, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54880, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54881, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54882, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54883, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54884, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54885, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54886, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54887, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54888, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54889, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54890, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54891, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54892, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54893, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54894, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54895, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x110E, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54896, 2, {0x0, 0x12, 0x412, 0x412, 0x3F92, 0x12, 0x1F12, 0x209E, 0x2092, 0x1F12, 0x412, 0x412, 0x7FD2, 0x12, 0x12, 0x0}},
    {54897, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54898, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54899, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54900, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54901, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54902, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54903, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54904, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54905, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54906, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54907, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54908, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54909, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54910, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54911, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54912, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54913, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54914, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54915, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54916, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54917, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54918, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54919, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54920, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54921, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54922, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54923, 2, {0x0, 0x428, 0x3FA8, 0xE28, 0x1138, 0x1128, 0xE28, 0x428, 0x7FA8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54924, 2, {0x0, 0x8, 0x408, 0x408, 0x3F88, 0x8, 0x1F08, 0x2088, 0x2088, 0x1F08, 0x408, 0x408, 0x7FE8, 0x8, 0x8, 0x0}},
    {54925, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {54926, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54927, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {54928, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {54929, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {54930, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {54931, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {54932, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {54933, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54934, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54935, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54936, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {54937, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54938, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54939, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {54940, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {54941, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {54942, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {54943, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {54944, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {54945, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {54946, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {54947, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {54948, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {54949, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {54950, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {54951, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x408, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {54952, 2, {0x0, 0x100, 0x100, 0xFE0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x440, 0x440, 0x7FFC, 0x0, 0x0}},
    {54953, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54954, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54955, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54956, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54957, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54958, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54959, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54960, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54961, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54962, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54963, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54964, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54965, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54966, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54967, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54968, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54969, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54970, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54971, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {54972, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {54973, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {54974, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54975, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {54976, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {54977, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54978, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {54979, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x440, 0x440, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {54980, 2, {0x0, 0x100, 0x100, 0xFE0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x3FF8, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {54981, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {54982, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {54983, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {54984, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x100, 0x100, 0x1100, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54985, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {54986, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {54987, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {54988, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {54989, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {54990, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {54991, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {54992, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {54993, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {54994, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {54995, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {54996, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {54997, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {54998, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {54999, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {55000, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {55001, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {55002, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {55003, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {55004, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {55005, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {55006, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {55007, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x100, 0x100, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {55008, 2, {0x0, 0x408, 0x408, 0x3F88, 0x8, 0x1F08, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x408, 0x4F8, 0x408, 0x408, 0x0}},
    {55009, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {55010, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55011, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {55012, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x8, 0x7FE8, 0x278, 0x208, 0x208, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {55013, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {55014, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {55015, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {55016, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {55017, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55018, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55019, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55020, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {55021, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55022, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55023, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {55024, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {55025, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {55026, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {55027, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {55028, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {55029, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {55030, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {55031, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {55032, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {55033, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {55034, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {55035, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x278, 0x208, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {55036, 2, {0x0, 0x40A, 0x40A, 0x3F8A, 0xA, 0x1F0A, 0x208A, 0x208A, 0x1F0A, 0xA, 0x7FEA, 0x40A, 0x47A, 0x40A, 0x40A, 0x0}},
    {55037, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {55038, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55039, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {55040, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x28, 0x7FA8, 0x428, 0x5E8, 0x428, 0x1428, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {55041, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {55042, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {55043, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {55044, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {55045, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55046, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55047, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55048, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {55049, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55050, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55051, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {55052, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {55053, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {55054, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {55055, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {55056, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {55057, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {55058, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {55059, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {55060, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {55061, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {55062, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {55063, 2, {0x428, 0x3FA8, 0xE28, 0x1128, 0xE28, 0x28, 0x7FA8, 0x5E8, 0x428, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {55064, 2, {0x0, 0x408, 0x408, 0x3F88, 0x8, 0x1F08, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x408, 0x408, 0x408, 0x408, 0x0}},
    {55065, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {55066, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55067, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {55068, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x8, 0x7FE8, 0x208, 0x208, 0x208, 0x1008, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {55069, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {55070, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {55071, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {55072, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {55073, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55074, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55075, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55076, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {55077, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55078, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55079, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {55080, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {55081, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {55082, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {55083, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {55084, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {55085, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {55086, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {55087, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {55088, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {55089, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {55090, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {55091, 2, {0x408, 0x3F88, 0xE08, 0x1108, 0xE08, 0x8, 0x7FE8, 0x208, 0x200, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {55092, 2, {0x0, 0x100, 0x100, 0xFE0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x440, 0x440, 0x0}},
    {55093, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {55094, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55095, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {55096, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x440, 0x440, 0x1440, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {55097, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {55098, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {55099, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {55100, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {55101, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55102, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55103, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55104, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {55105, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55106, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55107, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {55108, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {55109, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {55110, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {55111, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {55112, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {55113, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {55114, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {55115, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {55116, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {55117, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {55118, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {55119, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x7FFC, 0x440, 0x440, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {55120, 2, {0x0, 0x100, 0x100, 0xFE0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x0, 0x0}},
    {55121, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {55122, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55123, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1E10, 0x210, 0x228, 0x2C4, 0x0}},
    {55124, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {55125, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {55126, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {55127, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1000, 0x1000, 0x1FF0, 0x0}},
    {55128, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {55129, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55130, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55131, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55132, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E10, 0x210, 0x3E10, 0x2028, 0x3EC4, 0x0}},
    {55133, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55134, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55135, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x3E20, 0x3FC, 0x3E70, 0x2088, 0x3E70, 0x0}},
    {55136, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1010, 0x1010, 0x1010, 0x1FF0, 0x0}},
    {55137, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1010, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x0}},
    {55138, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x2220, 0x2220, 0x3E20, 0x2250, 0x3E88, 0x0}},
    {55139, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x100, 0x100, 0x280, 0xC40, 0x0}},
    {55140, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x240, 0x240, 0xDA0, 0x3310, 0x0}},
    {55141, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0}},
    {55142, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {55143, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x80, 0x7E0, 0x80, 0x140, 0x620, 0x0}},
    {55144, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x10, 0x1FF0, 0x10, 0x10, 0x0}},
    {55145, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x1FF0, 0x1000, 0x1FF0, 0x1000, 0x1FF0, 0x0}},
    {55146, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x0, 0x1FF0, 0x440, 0x440, 0x1FF0, 0x0}},
    {55147, 2, {0x100, 0xFE0, 0x7C0, 0x820, 0x820, 0x7C0, 0x0, 0x0, 0x7FFC, 0x0, 0x100, 0x1FF0, 0x7C0, 0x820, 0x7C0, 0x0}},
    {55148, 2, {0x0, 0x8, 0x408, 0x408, 0x3F88, 0x8, 0x1F08, 0x2088, 0x2088, 0x1F08, 0x8, 0x7FE8, 0x8, 0x8, 0x8, 0x0}},
    {55149, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {55150, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55151, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {55152, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x800, 0x800, 0x800, 0x800, 0xFF8, 0x0}},
    {55153, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x0, 0x10F8, 0x1010, 0x1030, 0x1ECC, 0x0}},
    {55154, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x808, 0x87F, 0x81C, 0x822, 0xF9C, 0x0}},
    {55155, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {55156, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {55157, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55158, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55159, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55160, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x1F08, 0x108, 0x1F08, 0x1014, 0x1F62, 0x0}},
    {55161, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55162, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55163, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x1F08, 0x17F, 0x1F1C, 0x1022, 0x1F1C, 0x0}},
    {55164, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {55165, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {55166, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x1108, 0x1108, 0x1F08, 0x1114, 0x1F62, 0x0}},
    {55167, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {55168, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {55169, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {55170, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {55171, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {55172, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {55173, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {55174, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {55175, 2, {0x0, 0x408, 0x3F88, 0xE08, 0x1108, 0x1108, 0xE08, 0x8, 0x7FE8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {55176, 2, {0x0, 0x0, 0x408, 0x408, 0x3F88, 0xE08, 0x1108, 0x2088, 0x2088, 0x2088, 0x1108, 0xE08, 0x8, 0x8, 0x8, 0x0}},
    {55177, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x7F8, 0x8, 0x8, 0x8, 0x0}},
    {55178, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x3EF8, 0x208, 0x208, 0x208, 0x0}},
    {55179, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x1F08, 0x108, 0x114, 0x162, 0x0}},
    {55180, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x8, 0x0, 0x1000, 0x1000, 0x1000, 0x1FF8, 0x0}},
    {55181, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x20F8, 0x2010, 0x2030, 0x3ECC, 0x0}},
    {55182, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x2020, 0x21FC, 0x2070, 0x2088, 0x3E70, 0x0}},
    {55183, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x200, 0x200, 0x3F8, 0x0}},
    {55184, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x7F8, 0x8, 0x7F8, 0x400, 0x7F8, 0x0}},
    {55185, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x208, 0x3E08, 0x2008, 0x3E08, 0x0}},
    {55186, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x288, 0x3E88, 0x2088, 0x3EF8, 0x0}},
    {55187, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3E88, 0x288, 0x3EF8, 0x2088, 0x3EF8, 0x0}},
    {55188, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0xF88, 0x88, 0xF88, 0x814, 0xFA2, 0x0}},
    {55189, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3EF8, 0x280, 0x3EF8, 0x2080, 0x3EF8, 0x0}},
    {55190, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3EFC, 0x248, 0x3E48, 0x2048, 0x3EFC, 0x0}},
    {55191, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3E10, 0x2FE, 0x3E38, 0x2044, 0x3E38, 0x0}},
    {55192, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x208, 0x208, 0x208, 0x3F8, 0x0}},
    {55193, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x208, 0x208, 0x3F8, 0x208, 0x3F8, 0x0}},
    {55194, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x2108, 0x2108, 0x3F08, 0x2114, 0x3F62, 0x0}},
    {55195, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x10, 0x10, 0x10, 0x68, 0x184, 0x0}},
    {55196, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x48, 0x48, 0xA8, 0x114, 0x662, 0x0}},
    {55197, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0x3F0, 0x408, 0x408, 0x3F0, 0x0}},
    {55198, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x20, 0x20, 0xD0, 0x308, 0x0}},
    {55199, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x40, 0x3F8, 0x40, 0xA0, 0x318, 0x0}},
    {55200, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x8, 0x3F8, 0x8, 0x8, 0x0}},
    {55201, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x3F8, 0x200, 0x3F8, 0x200, 0x3F8, 0x0}},
    {55202, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x0, 0xFF8, 0x220, 0x220, 0xFF8, 0x0}},
    {55203, 2, {0x0, 0x808, 0x7F08, 0x1C08, 0x2208, 0x2208, 0x2208, 0x1C08, 0x8, 0x0, 0x80, 0xFF8, 0x3E0, 0x410, 0x3E0, 0x0}},
    {63744, 2, {0x80, 0x1088, 0x1088, 0x1FF8, 0x1008, 0x0, 0x7FFE, 0x0, 0xFF0, 0x810, 0xFF0, 0x810, 0x420, 0x244, 0x7FFE, 0x0}},
    {63745, 2, {0x4, 0x7FFE, 0x80, 0x1084, 0x1FFE, 0x1084, 0x1FFC, 0x1084, 0x1FFC, 0x1084, 0x880, 0x500, 0x300, 0xCC0, 0x703E, 0x0}},
    {63746, 2, {0x100, 0x104, 0x7FFE, 0x100, 0x1108, 0x1FFC, 0x1108, 0x1FF8, 0x1108, 0x1FF8, 0x1108, 0x104, 0x7FFE, 0x100, 0x100, 0x0}},
    {63747, 2, {0xFFFE, 0x440, 0x3FF8, 0x2448, 0x3FF8, 0x0, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0xC60, 0xF01E, 0x0}},
    {63748, 2, {0x4108, 0x21FC, 0x2108, 0x978, 0x8948, 0x57FE, 0x5404, 0x21F8, 0x2108, 0xC1F8, 0x4108, 0x41F8, 0xC108, 0xC138, 0x4110, 0x0}},
    {63749, 2, {0x100, 0x108, 0x3FFC, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x104, 0x7FFE, 0x4104, 0x4104, 0x7FFC, 0x4104, 0x100, 0x100, 0x0}},
    {63750, 2, {0x800, 0x800, 0x804, 0x1FFE, 0x2004, 0x4004, 0x1FC4, 0x1044, 0x1044, 0x1044, 0x1FC4, 0x1044, 0x44, 0x38, 0x10, 0x0}},
    {63751, 2, {0x840, 0xFE0, 0x840, 0x1080, 0x7FFC, 0x1144, 0x7F7C, 0x544, 0x7F6C, 0x554, 0x7D6C, 0x544, 0x7F7E, 0x502, 0x7CFC, 0x0}},
    {63752, 2, {0x840, 0xFE0, 0x840, 0x1080, 0x7FFC, 0x1144, 0x7F7C, 0x544, 0x7F6C, 0x554, 0x7D6C, 0x544, 0x7F7E, 0x502, 0x7CFC, 0x0}},
    {63753, 2, {0x800, 0x802, 0x7F7E, 0x812, 0x7F12, 0x812, 0x7F22, 0x84E, 0x984, 0x100, 0x7FFE, 0x200, 0x2C0, 0xC30, 0x700E, 0x0}},
    {63754, 2, {0x80, 0x180, 0x240, 0x420, 0x810, 0x1008, 0x6FF6, 0x100, 0x108, 0x3FFC, 0x100, 0x1118, 0x910, 0x924, 0x7FFE, 0x0}},
    {63755, 2, {0x102, 0x9142, 0xFFE2, 0x9112, 0x9112, 0x97D2, 0x9552, 0x9552, 0x97D2, 0x9552, 0xF392, 0x9542, 0x922, 0x111E, 0x104, 0x0}},
    {63756, 2, {0x200, 0x204, 0x7FFE, 0x240, 0x420, 0x1818, 0x67E6, 0x0, 0x8, 0x3FFC, 0x80, 0x1890, 0x108C, 0x2384, 0x4104, 0x0}},
    {63757, 2, {0x2214, 0x221E, 0x2FD4, 0x2224, 0xB27E, 0xAFA2, 0xAABE, 0xAAA2, 0x2FBE, 0x2222, 0x233E, 0x26A2, 0x2A94, 0x2222, 0x2242, 0x2000}},
    {63758, 2, {0x100, 0x1084, 0x1FFE, 0x9220, 0x5FBC, 0x5248, 0x1FBE, 0x12A2, 0x32BE, 0xDFA2, 0x923E, 0x2322, 0x26BE, 0x4A94, 0x9222, 0x0}},
    {63759, 2, {0x3FFC, 0x2244, 0x2244, 0x3FFC, 0x844, 0x1090, 0x24FE, 0x7990, 0xA90, 0x14FC, 0x7E90, 0xFC, 0x3490, 0x2A90, 0x4AFE, 0x0}},
    {63760, 2, {0x420, 0x7FFE, 0x420, 0x1FF8, 0x1248, 0x1FF8, 0x8A0, 0x1290, 0x3DFE, 0xA90, 0x14FC, 0x7E90, 0x28FC, 0x5490, 0x54FE, 0x0}},
    {63761, 2, {0x1104, 0x11FE, 0x1124, 0x55FC, 0x7F24, 0x55FC, 0x5444, 0x55F8, 0x7C24, 0x55FE, 0x1092, 0x1410, 0x1A94, 0xE372, 0x4222, 0x0}},
    {63762, 2, {0x41FC, 0x2124, 0x2924, 0xFDFC, 0x924, 0x11FC, 0x2524, 0x2820, 0x73FE, 0xAC20, 0x2470, 0x20A8, 0x2126, 0x2222, 0x2020, 0x0}},
    {63763, 2, {0x7FE, 0x4492, 0x3492, 0x17FE, 0x248, 0x4FE, 0x1748, 0xFAFE, 0x17C8, 0x107E, 0x1548, 0x1AFE, 0x3040, 0xC800, 0x87FE, 0x0}},
    {63764, 2, {0x2108, 0x2108, 0x47D0, 0x5454, 0xE7F8, 0x2448, 0x5454, 0xFFFE, 0x4550, 0x104, 0xFFFE, 0x520, 0x910, 0x310E, 0xC104, 0x0}},
    {63765, 2, {0x2100, 0x1108, 0x11FC, 0x108, 0x4290, 0x2C90, 0x2860, 0x1198, 0x1606, 0x61F8, 0x2108, 0x2108, 0x6108, 0x61F8, 0x2108, 0x0}},
    {63766, 2, {0x1080, 0x1088, 0x10FC, 0x1088, 0x5550, 0x5920, 0x5230, 0x50C8, 0x9306, 0x10FC, 0x1884, 0x2484, 0x2484, 0x40FC, 0x8084, 0x0}},
    {63767, 2, {0x80, 0x888, 0xFCFC, 0x1108, 0x1290, 0x1060, 0x1060, 0xFD98, 0x1206, 0x11F8, 0x1508, 0x1908, 0xE108, 0x41F8, 0x108, 0x0}},
    {63768, 2, {0x440, 0x7FFE, 0x440, 0x2440, 0x11F8, 0x1508, 0x4590, 0x2A60, 0x890, 0x1108, 0x17FE, 0x6108, 0x2108, 0x21F8, 0x6108, 0x0}},
    {63769, 2, {0x420, 0xFE24, 0x283E, 0x2824, 0xFE44, 0xAAA8, 0xAA10, 0xAA28, 0xCE44, 0x82FE, 0xFE44, 0x8244, 0x8244, 0xFE7C, 0x8244, 0x0}},
    {63770, 2, {0x4440, 0x7E48, 0x487C, 0x4888, 0x7F48, 0x4830, 0x7E30, 0x4848, 0x7E86, 0x43FC, 0x284, 0x5684, 0xAA84, 0xAAFC, 0xC84, 0x0}},
    {63771, 2, {0x720, 0xF920, 0x9120, 0x4A20, 0xFF20, 0x1420, 0xFFA0, 0x92A0, 0xBDA0, 0x80A0, 0xBEA0, 0x94A2, 0x8CA2, 0x92A2, 0xA39E, 0x0}},
    {63772, 2, {0x200, 0x2744, 0x387E, 0x2244, 0x3244, 0x2A54, 0x2A54, 0x224C, 0x2244, 0x2644, 0x3A44, 0x225C, 0x448, 0x840, 0x7040, 0x0}},
    {63773, 2, {0x13DE, 0x1252, 0x13DE, 0xFE52, 0x13DE, 0x1222, 0x3BFE, 0x5622, 0x53FE, 0x92AA, 0x93FE, 0x1222, 0x1272, 0x12AA, 0x1326, 0x0}},
    {63774, 2, {0x13DE, 0x1252, 0x13DE, 0x1252, 0x57DE, 0x5A22, 0x53FE, 0x5222, 0x93FE, 0x12AA, 0x1BFE, 0x2622, 0x22B2, 0x432E, 0x8224, 0x0}},
    {63775, 2, {0x420, 0x7FFE, 0x420, 0x7EFC, 0x4284, 0x7EFC, 0x4284, 0x7EFC, 0x4104, 0x5FF4, 0x4AA4, 0x4FE4, 0x4104, 0x4544, 0x593C, 0x0}},
    {63776, 2, {0x1208, 0x1108, 0x27D0, 0x783C, 0x2792, 0x783E, 0x2BA8, 0x5294, 0xFE0, 0x820, 0xFF8, 0x800, 0xFFE, 0x2922, 0x4494, 0x0}},
    {63777, 2, {0x80, 0x2084, 0x2084, 0x3FFC, 0x0, 0x3FF8, 0x2028, 0x2FC8, 0x2108, 0x2FE8, 0x2928, 0x2FEA, 0x410A, 0x4126, 0x9FD2, 0x0}},
    {63778, 2, {0x47D0, 0x2510, 0x2514, 0x7DE, 0x8450, 0x57D8, 0x5524, 0x2524, 0x27C0, 0xC000, 0x47FC, 0x44A4, 0xC4A4, 0xC4A4, 0x4FFE, 0x0}},
    {63779, 2, {0x420, 0x7FFE, 0x420, 0x3F00, 0x2424, 0x3F3E, 0x2140, 0x3F90, 0x2408, 0x3F88, 0x0, 0x1FF8, 0x1248, 0x1248, 0x7FFE, 0x0}},
    {63780, 2, {0x43E8, 0x2288, 0x2A88, 0xFFCE, 0xA50, 0x13E8, 0x2684, 0x2A84, 0x73E0, 0xA800, 0x27F8, 0x22A8, 0x22A8, 0x22A8, 0x27FE, 0x0}},
    {63781, 2, {0x1000, 0x1040, 0x1020, 0xFC24, 0x13FE, 0x1000, 0x140C, 0x1908, 0xF108, 0x5090, 0x1090, 0x1090, 0x9020, 0x7024, 0x27FE, 0x0}},
    {63782, 2, {0x4924, 0x7E48, 0x4A48, 0x4924, 0x4BFC, 0x7AA4, 0x4A54, 0x4BFC, 0x4840, 0x7B68, 0x4AD8, 0x4B68, 0x4ADA, 0xBB66, 0x9242, 0x0}},
    {63783, 2, {0x2124, 0x2248, 0x2248, 0xA924, 0xFFFC, 0xAAA4, 0xAA54, 0xABFC, 0xF800, 0xAB68, 0x22D8, 0x2B68, 0x36DA, 0xE266, 0x4342, 0x0}},
    {63784, 2, {0x100, 0x2084, 0x3FFE, 0x2400, 0x22BE, 0x2FE4, 0x28A4, 0x2FA8, 0x28A8, 0x2FA4, 0x28A2, 0x2932, 0x2BAC, 0x4CA0, 0x8820, 0x0}},
    {63785, 2, {0x1044, 0x4A7E, 0x7F44, 0x4244, 0x4244, 0x7E7C, 0x4244, 0x4244, 0x7E44, 0x4A7C, 0x4444, 0x4A44, 0x52A4, 0xE11C, 0x4208, 0x0}},
    {63786, 2, {0x2080, 0x1248, 0x13FC, 0x208, 0x4208, 0x2BF8, 0x2A08, 0x1208, 0x13F8, 0x6244, 0x2248, 0x2230, 0x6290, 0x6708, 0x2206, 0x0}},
    {63787, 2, {0x40, 0x4C20, 0x2924, 0x11FE, 0x3104, 0x49FC, 0x8904, 0x19FC, 0x2942, 0x4944, 0x8928, 0x910, 0x4948, 0x3B86, 0x1102, 0x0}},
    {63788, 2, {0x1024, 0x93E, 0x3FA4, 0x2124, 0x2128, 0x3F28, 0x2128, 0x2124, 0x3F24, 0x2122, 0x2422, 0x2232, 0x2D2C, 0x7120, 0x2020, 0x0}},
    {63789, 2, {0x100, 0x104, 0x7FFE, 0x100, 0x1110, 0x1110, 0x1110, 0x2928, 0x4384, 0x540, 0x920, 0x1110, 0x210C, 0x4104, 0x100, 0x0}},
    {63790, 2, {0x40, 0x8040, 0x60A0, 0x2890, 0x908, 0x1246, 0x1422, 0x2820, 0x23FC, 0xC008, 0x4010, 0x4020, 0x41C0, 0x4020, 0x10, 0x0}},
    {63791, 2, {0x410, 0x1554, 0x1658, 0x2490, 0xA28, 0x1144, 0x3FFE, 0x2002, 0x4104, 0x3FF8, 0x108, 0x208, 0x488, 0x870, 0x3020, 0x0}},
    {63792, 2, {0x107C, 0x1240, 0x13FE, 0xFE44, 0x13FA, 0x1242, 0x16FE, 0x1AA4, 0xF3FE, 0x52A4, 0x12FC, 0x1220, 0x95FE, 0x7842, 0x239C, 0x0}},
    {63793, 2, {0x1090, 0x10F8, 0x1110, 0xFFFC, 0x1124, 0x11FC, 0x3924, 0x55FC, 0x5554, 0x924A, 0x91FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x0}},
    {63794, 2, {0x107C, 0x1240, 0x13FE, 0x1244, 0x57FA, 0x5A42, 0x52FE, 0x52A8, 0x92F8, 0x12A8, 0x1AF8, 0x2400, 0x25FC, 0x4954, 0x97FE, 0x0}},
    {63795, 2, {0x110, 0x1F8, 0x2100, 0x3FFE, 0x2104, 0x2FF8, 0x2102, 0x23FE, 0x2248, 0x23F8, 0x2248, 0x2FFC, 0x2924, 0x4924, 0x9FFE, 0x0}},
    {63796, 2, {0x100, 0x100, 0x10C, 0x3FE8, 0x110, 0x124, 0x7FFE, 0x80, 0x100, 0x630, 0xCC0, 0x1704, 0x6404, 0x406, 0x3FC, 0x0}},
    {63797, 2, {0x420, 0x7FFE, 0x4A0, 0x10F8, 0x1FFE, 0x1104, 0x1FF8, 0x1102, 0x13FE, 0x1248, 0x13F8, 0x1248, 0x17F8, 0x2528, 0x5FFE, 0x0}},
    {63798, 2, {0x1F8, 0x100, 0x3FFE, 0x2104, 0x2FF0, 0x2102, 0x27FE, 0x2448, 0x3FFE, 0x2448, 0x27F8, 0x2080, 0x2FFC, 0x4104, 0x8E38, 0x0}},
    {63799, 2, {0x40, 0x7C44, 0x447E, 0x4484, 0x4548, 0x7C30, 0x5448, 0x1086, 0x5102, 0x5CFC, 0x5084, 0x5084, 0x5C84, 0xF0FC, 0x4084, 0x0}},
    {63800, 2, {0x1FF8, 0x4102, 0x7FFE, 0x8104, 0x3D78, 0x140, 0x3E78, 0x2248, 0x3ED0, 0x2B30, 0x84E, 0x2EF8, 0x2888, 0x7EF8, 0x2088, 0x0}},
    {63801, 2, {0x440, 0xFE0, 0x1088, 0x3FFC, 0x5108, 0x1FF8, 0x1108, 0x1FF8, 0x2444, 0x2222, 0x4FF2, 0x810, 0xFF0, 0x810, 0xFF0, 0x0}},
    {63802, 2, {0x3E7C, 0x2244, 0x3EB8, 0x844, 0x2EFA, 0x2888, 0x7FF8, 0x1010, 0x1FF0, 0x1010, 0x1FFC, 0x1000, 0x1FFE, 0x2922, 0x4494, 0x0}},
    {63803, 2, {0x44, 0x47C, 0xFE88, 0x1088, 0x11F0, 0x2010, 0x3FFE, 0x6420, 0xA522, 0x24B4, 0x2468, 0x27A4, 0x3D22, 0x24E2, 0x40, 0x0}},
    {63804, 2, {0x2084, 0x10FC, 0x1088, 0xFD08, 0x5F0, 0x810, 0x13FE, 0x3020, 0x592C, 0x94A8, 0x1470, 0x10A8, 0x1726, 0x12E2, 0x1040, 0x0}},
    {63805, 2, {0x1088, 0x10FC, 0x2088, 0x2508, 0x45F0, 0xF810, 0x13FE, 0x2820, 0xFD2C, 0x44A8, 0x70, 0x28A8, 0x5526, 0x56E2, 0x8040, 0x0}},
    {63806, 2, {0x420, 0x420, 0x7FFE, 0x420, 0x3E0, 0x220, 0x7E0, 0x44, 0x7FFE, 0x1088, 0x8D0, 0x6A0, 0x890, 0x738E, 0x2104, 0x0}},
    {63807, 2, {0x1044, 0x187C, 0x1444, 0x2288, 0x40F8, 0xFC10, 0x11FE, 0x1020, 0xFF26, 0x94B4, 0x5868, 0x51A8, 0x1E26, 0xF0E2, 0x4040, 0x0}},
    {63808, 2, {0x100, 0x2084, 0x3FFE, 0x2220, 0x2224, 0x3FFE, 0x2224, 0x3FFC, 0x2804, 0x2820, 0x2FA4, 0x2838, 0x29A2, 0x5E22, 0x883E, 0x0}},
    {63809, 2, {0x2020, 0x1020, 0x1050, 0xFC88, 0x106, 0x2FA, 0x7800, 0x3FE, 0x7A52, 0x252, 0x7BFE, 0x4A52, 0x4A52, 0x7A52, 0x4A46, 0x0}},
    {63810, 2, {0x87E, 0x7F40, 0x147C, 0x7F04, 0x227C, 0x3E40, 0x227C, 0x3E40, 0x227E, 0x2E42, 0x243E, 0x80, 0x1FF8, 0x84, 0x7FFE, 0x0}},
    {63811, 2, {0x7FFC, 0x100, 0x110, 0x3FF8, 0x100, 0x104, 0x7FFE, 0x820, 0x824, 0x7FFE, 0x820, 0x820, 0x1020, 0x2020, 0x4020, 0x0}},
    {63812, 2, {0x1040, 0x1F7E, 0x2890, 0x4548, 0x3F7E, 0x2240, 0x147C, 0x7F04, 0x227C, 0x3E40, 0x227C, 0x3E40, 0x227C, 0x2242, 0x263E, 0x0}},
    {63813, 2, {0x880, 0x7EFC, 0x1480, 0x7FFC, 0x2204, 0x3EFC, 0x2280, 0x3EFE, 0x2282, 0x7FFE, 0x810, 0xFF0, 0x810, 0x7FFE, 0x10, 0x0}},
    {63814, 2, {0x100, 0x2084, 0x3FFE, 0x2004, 0x4908, 0x910, 0xFF8, 0x1100, 0x2100, 0x104, 0x7FFE, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {63815, 2, {0x4, 0x7FFE, 0x200, 0x7F0, 0xC10, 0x1410, 0x67F0, 0x204, 0x7F7E, 0x1020, 0x1E3C, 0x3264, 0x52A4, 0x1E3C, 0x1224, 0x0}},
    {63816, 2, {0x4480, 0x7E84, 0x44FE, 0x4484, 0x7D48, 0x4630, 0x4448, 0x7C86, 0x4502, 0x44FC, 0x7C84, 0x4484, 0x2884, 0x44FC, 0x8484, 0x0}},
    {63817, 2, {0x1FF8, 0x4102, 0x7FFE, 0x8104, 0x3D78, 0x100, 0x3D78, 0x100, 0x2008, 0x3FFC, 0x2108, 0x3FF8, 0x2108, 0x3FF8, 0x2008, 0x0}},
    {63818, 2, {0x1FF0, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x1010, 0xFEFE, 0x9292, 0xFEFE, 0x9292, 0xFEFE, 0x110, 0x3FF8, 0x104, 0xFFFE, 0x0}},
    {63819, 2, {0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2488, 0x3FFE, 0x2488, 0x2FFC, 0x2884, 0x2FFC, 0x2100, 0x3FFE, 0x2220, 0x4160, 0x9F9E, 0x0}},
    {63820, 2, {0x1020, 0x11FC, 0x1124, 0x13FE, 0xFD24, 0x11FC, 0x1820, 0x35FC, 0x5524, 0x51FC, 0x9040, 0x97FE, 0x1090, 0x1070, 0x17CE, 0x0}},
    {63821, 2, {0x2080, 0x1244, 0x13FE, 0x204, 0x4204, 0x2BFC, 0x2A04, 0x1228, 0x1224, 0x63FE, 0x2220, 0x2450, 0x6488, 0x6906, 0x3202, 0x0}},
    {63822, 2, {0x4404, 0x27FE, 0x2404, 0x7FC, 0x8400, 0x57FE, 0x5420, 0x27FE, 0x2622, 0xC732, 0x46AA, 0x4B32, 0xCAAA, 0xD2AA, 0x6226, 0x0}},
    {63823, 2, {0x1FFC, 0x1084, 0x1FFC, 0x1084, 0x1FFC, 0x1104, 0x220, 0x460, 0x1F80, 0x208, 0x3FFC, 0x1084, 0x4A0, 0x1B98, 0x6106, 0x0}},
    {63824, 2, {0x1020, 0x1020, 0x21FC, 0x2524, 0x47FE, 0xF924, 0x13FE, 0x2A22, 0xFFFE, 0x4420, 0x3FE, 0x2888, 0x5590, 0x5460, 0x839E, 0x0}},
    {63825, 2, {0x8804, 0xF9FE, 0x8810, 0x9090, 0xA2FE, 0xA292, 0x9292, 0x9292, 0x8AAA, 0x8AC6, 0xCA82, 0xB29E, 0x8284, 0x8200, 0x83FE, 0x0}},
    {63826, 2, {0x2220, 0x2220, 0xFFA0, 0x2224, 0x3EFE, 0x824, 0x7F24, 0x4924, 0x4924, 0x7F24, 0x844, 0xFF44, 0x884, 0x8BC, 0x908, 0x0}},
    {63827, 2, {0x2420, 0x3E20, 0x2420, 0x2424, 0x25FE, 0x3C24, 0x2424, 0x2424, 0x2424, 0x3C24, 0x2444, 0x2444, 0x24A4, 0x5D1C, 0x8808, 0x0}},
    {63828, 2, {0x40, 0x8FFE, 0x6000, 0x2BFC, 0xA94, 0x12F4, 0x1204, 0x23FC, 0x2000, 0xC3F8, 0x4040, 0x4FFE, 0x4248, 0x4444, 0x842, 0x0}},
    {63829, 2, {0x40, 0x8048, 0x63FC, 0x2840, 0x844, 0x17FE, 0x1108, 0x2204, 0x248A, 0xC1FC, 0x4308, 0x4490, 0x4860, 0x419E, 0xE04, 0x0}},
    {63830, 2, {0x420, 0xE28, 0xF1FC, 0x1020, 0x15FE, 0xFE48, 0x1086, 0x1942, 0x3678, 0x5088, 0x9148, 0x1230, 0x1050, 0x118E, 0x1604, 0x0}},
    {63831, 2, {0x1020, 0x11FC, 0x2020, 0x2424, 0x47FE, 0xF888, 0x1104, 0x2A82, 0xFCF8, 0x4508, 0x288, 0x2850, 0x5460, 0x5598, 0x8606, 0x0}},
    {63832, 2, {0x420, 0x420, 0x7FFE, 0x520, 0x1FF8, 0x100, 0x7FFE, 0x420, 0x1A18, 0x63F6, 0x420, 0x1A40, 0x180, 0x670, 0x780E, 0x0}},
    {63833, 2, {0x8848, 0xFBFC, 0x8840, 0x9044, 0xA7FE, 0xA090, 0x910C, 0x9284, 0x88F8, 0x8988, 0xCA90, 0xB4A0, 0x8060, 0x8198, 0x8E06, 0x0}},
    {63834, 2, {0x2020, 0x13FE, 0xFC20, 0x1FC, 0x0, 0x7BFE, 0x252, 0x3FE, 0x7904, 0x1FC, 0x7904, 0x49FC, 0x4904, 0x79FC, 0x4B06, 0x0}},
    {63835, 2, {0x804, 0x8FE, 0x7F44, 0x1228, 0x2410, 0x1A28, 0x11C6, 0x6038, 0x1FC0, 0x88, 0x1FFC, 0x80, 0x7FFE, 0x80, 0x780, 0x0}},
    {63836, 2, {0x2108, 0x2108, 0x47D0, 0x5454, 0xE7F8, 0x2448, 0x5454, 0xFFFE, 0x4550, 0x104, 0xFFFE, 0x520, 0x910, 0x310E, 0xC104, 0x0}},
    {63837, 2, {0x2088, 0x1088, 0x13FE, 0xFC88, 0xA8, 0x20, 0x7824, 0x3FE, 0x7840, 0xFC, 0x7984, 0x4A84, 0x4C84, 0x78FC, 0x4884, 0x0}},
    {63838, 2, {0x10, 0xFF8, 0x810, 0x810, 0x910, 0x890, 0x890, 0x810, 0x7FFE, 0x810, 0x810, 0x1010, 0x1090, 0x2070, 0x4020, 0x0}},
    {63839, 2, {0x100, 0x2084, 0x3FFE, 0x2504, 0x54A8, 0x27E4, 0x0, 0x1FF8, 0x1248, 0x7FFE, 0x0, 0x3FFC, 0x80, 0x780, 0x100, 0x0}},
    {63840, 2, {0x800, 0x804, 0xBFE, 0x7E44, 0x1244, 0x2448, 0x1C30, 0xA48, 0x1186, 0x6002, 0x500, 0x2488, 0x24A6, 0x2422, 0x43F0, 0x0}},
    {63841, 2, {0x100, 0x84, 0x7FFE, 0x104, 0x2248, 0x1790, 0x100, 0x1A50, 0x67EC, 0x4524, 0x100, 0x7FFE, 0x100, 0x100, 0x100, 0x0}},
    {63842, 2, {0x1008, 0x1FFC, 0x1108, 0x1FF8, 0x1108, 0x1FF8, 0x1448, 0x440, 0x3FFC, 0x440, 0x444, 0x7FFE, 0x420, 0x1818, 0x6006, 0x0}},
    {63843, 2, {0x880, 0x880, 0x880, 0x88C, 0x888, 0xF890, 0x8E0, 0x880, 0x880, 0x880, 0x880, 0x1884, 0xE884, 0x4886, 0x87C, 0x0}},
    {63844, 2, {0xE, 0x9F0, 0xFC26, 0x2124, 0x20A8, 0x23FE, 0x3CA8, 0x6524, 0xA622, 0x25FE, 0x2522, 0x25FE, 0x3D22, 0x25FE, 0x2102, 0x0}},
    {63845, 2, {0x1004, 0x17FE, 0x1040, 0x2444, 0x27FE, 0x2444, 0x67FC, 0xA444, 0x27FC, 0x2444, 0x2240, 0x2140, 0x2080, 0x2140, 0x2E3E, 0x0}},
    {63846, 2, {0x1904, 0x11FE, 0x2200, 0x45FC, 0x9904, 0x11FC, 0x2104, 0x61FC, 0xA080, 0x21FC, 0x2308, 0x2490, 0x2060, 0x2198, 0x2606, 0x0}},
    {63847, 2, {0x4, 0x7FFE, 0x80, 0x80, 0x100, 0x300, 0x520, 0x910, 0x110C, 0x2106, 0x4102, 0x100, 0x100, 0x100, 0x100, 0x0}},
    {63848, 2, {0x2000, 0x1040, 0x1024, 0x424, 0x4488, 0x2888, 0x2888, 0x1290, 0x1294, 0x64A2, 0x24A2, 0x20C0, 0x6184, 0x6284, 0x247E, 0x0}},
    {63849, 2, {0x840, 0x7F40, 0x4940, 0xFFC4, 0x497E, 0x7FC8, 0x848, 0x7F48, 0x4948, 0x7F48, 0x1028, 0xFF90, 0x222C, 0x1C46, 0xF382, 0x0}},
    {63850, 2, {0x108, 0x3FFC, 0x100, 0x2104, 0x3FFE, 0x2104, 0x4228, 0x1FC0, 0x110, 0x208, 0x3FF4, 0x84, 0x4A0, 0x1B98, 0x6106, 0x0}},
    {63851, 2, {0x100, 0x220, 0x7D0, 0x1010, 0x2424, 0x7B7E, 0x182, 0x240, 0x7CBC, 0x300, 0x1C40, 0x198, 0x1E60, 0x180, 0x1E00, 0x0}},
    {63852, 2, {0x100, 0x2082, 0x3FFE, 0x4224, 0x1FFC, 0x220, 0x1FFC, 0x220, 0x7FFE, 0x490, 0x888, 0x17F6, 0x2082, 0x84, 0x7FFE, 0x0}},
    {63853, 2, {0x100, 0xD10, 0x908, 0x1166, 0x2142, 0x4080, 0x108, 0xFFC, 0x1808, 0x6FF8, 0x808, 0xFF8, 0x808, 0xFF8, 0x808, 0x0}},
    {63854, 2, {0x420, 0x7FFE, 0x420, 0x910, 0x7FFE, 0x910, 0x9F0, 0x800, 0xFFC, 0x80, 0x7FFE, 0x2C0, 0x4A0, 0x1898, 0x6086, 0x0}},
    {63855, 2, {0x2108, 0x1088, 0x1090, 0xFE20, 0x1FC, 0x104, 0xFD04, 0x104, 0xFDFC, 0x154, 0xFC50, 0x8450, 0x8492, 0xFD12, 0x860E, 0x0}},
    {63856, 2, {0x288, 0x42FC, 0x2488, 0x1888, 0x2688, 0x410E, 0x1400, 0x12FC, 0xFF84, 0x1048, 0x3848, 0x5430, 0x9228, 0x10C6, 0x1302, 0x0}},
    {63857, 2, {0x1004, 0x1FFE, 0x1000, 0x1010, 0x17F8, 0x1000, 0x1004, 0x1FFE, 0x1240, 0x1248, 0x1250, 0x2220, 0x2290, 0x4F08, 0x8406, 0x0}},
    {63858, 2, {0x2040, 0x1040, 0x1040, 0x244, 0x43FE, 0x2A44, 0x2C48, 0x1040, 0x1040, 0x6040, 0x2040, 0x20A0, 0x6122, 0x6222, 0x241E, 0x0}},
    {63859, 2, {0x1040, 0x1040, 0x10A0, 0xFD10, 0x1208, 0x1006, 0x15FA, 0x1800, 0xF108, 0x51FC, 0x1108, 0x1108, 0x9108, 0x71F8, 0x2108, 0x0}},
    {63860, 2, {0x420, 0x420, 0x7FFE, 0x420, 0x420, 0x100, 0x104, 0x7FFE, 0x200, 0x408, 0xFFC, 0x1408, 0x6408, 0x7F8, 0x408, 0x0}},
    {63861, 2, {0x1040, 0x1020, 0x1024, 0xFFFE, 0x1000, 0x11FC, 0x1504, 0x1904, 0xF1FC, 0x5124, 0x1020, 0x11A8, 0x9124, 0x72E2, 0x2442, 0x0}},
    {63862, 2, {0x80, 0x4484, 0x7EFE, 0x5484, 0x5548, 0x5430, 0x5448, 0x7C86, 0x5502, 0x54FC, 0x5484, 0x7C84, 0x4484, 0xFC, 0x84, 0x0}},
    {63863, 2, {0x200, 0x104, 0x7FFE, 0x0, 0xFF0, 0x810, 0xFF0, 0x0, 0x3FFE, 0x2002, 0x47C4, 0x440, 0x442, 0x1842, 0x603E, 0x0}},
    {63864, 2, {0x4, 0x7FFE, 0x100, 0x104, 0x7FFE, 0x4104, 0x5144, 0x4924, 0x4924, 0x5554, 0x6594, 0x4104, 0x4124, 0x411C, 0x4008, 0x0}},
    {63865, 2, {0x80, 0x8040, 0x6044, 0x27FE, 0x800, 0x13FC, 0x1204, 0x2204, 0x23FC, 0xC224, 0x4020, 0x4230, 0x4528, 0x48E4, 0x1044, 0x0}},
    {63866, 2, {0x1008, 0xBFC, 0x4048, 0x254C, 0x94A, 0xA4A, 0x30B8, 0x1310, 0x1084, 0x7FFE, 0x2C0, 0x4A0, 0x890, 0x108E, 0x6084, 0x0}},
    {63867, 2, {0x11FC, 0x1104, 0x95FC, 0x5504, 0x59FC, 0x1000, 0xFDFE, 0x1124, 0x19FC, 0x3524, 0x55FC, 0x9020, 0x11FC, 0x1020, 0x13FE, 0x0}},
    {63868, 2, {0x100, 0x890, 0xFF8, 0x810, 0x810, 0xFF0, 0x810, 0x810, 0xFF0, 0x900, 0x88C, 0x870, 0xB20, 0x3C18, 0x1006, 0x0}},
    {63869, 2, {0x2040, 0x1020, 0x1024, 0xFFFE, 0x0, 0x1FC, 0x7D04, 0x104, 0x7DFC, 0x124, 0x7CA8, 0x44A4, 0x4522, 0x7EE2, 0x4440, 0x0}},
    {63870, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x4, 0x7FFE, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x3FF8, 0x100, 0x7FFE, 0x0}},
    {63871, 2, {0x7FD0, 0x4910, 0x7FD0, 0x4912, 0x5FFE, 0x5252, 0x5FD2, 0x5252, 0x5FD2, 0x4212, 0x5FD2, 0x5252, 0x53D2, 0x9C6E, 0x90C4, 0x0}},
    {63872, 2, {0x810, 0xFF8, 0x810, 0x810, 0x810, 0xFF0, 0x910, 0x100, 0x1208, 0x1FFC, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x0}},
    {63873, 2, {0x200, 0x300, 0x200, 0x200, 0x204, 0x7FFE, 0x420, 0x420, 0x820, 0x820, 0x1040, 0xE80, 0x180, 0x670, 0x780E, 0x0}},
    {63874, 2, {0x2104, 0x3FFE, 0x2080, 0x20F8, 0x2880, 0x2FFE, 0x2884, 0x2FF8, 0x2A4A, 0x2BFE, 0x2A48, 0x2FF8, 0x2D28, 0x5528, 0xAFFE, 0x0}},
    {63875, 2, {0x1040, 0x840, 0x844, 0xFEFE, 0x2080, 0x251C, 0x3EE0, 0x24A6, 0x24A4, 0x2498, 0x2490, 0x2490, 0x24A8, 0x5DC6, 0x8882, 0x0}},
    {63876, 2, {0x4048, 0x207C, 0x2440, 0x7FE, 0x8444, 0x57FA, 0x5442, 0x25FE, 0x2524, 0xC5FC, 0x4524, 0x45FC, 0xCA94, 0xCA8A, 0x44FA, 0x0}},
    {63877, 2, {0x3FE, 0xA48, 0xFFFE, 0x2248, 0x22FC, 0x2294, 0x7AFC, 0xCA94, 0x4AFC, 0x4A10, 0x4AFE, 0x4A92, 0x7AFA, 0x4C86, 0x882, 0x0}},
    {63878, 2, {0x7E7E, 0x4242, 0x7E7E, 0x4242, 0x7E7E, 0x4002, 0x47E2, 0x4422, 0x47E2, 0x4102, 0x4FF2, 0x4812, 0x4FF2, 0x481E, 0x4004, 0x0}},
    {63879, 2, {0x89DC, 0xFC00, 0x93FE, 0x92AA, 0xFEAA, 0x9020, 0xFDFE, 0x9150, 0xFDFC, 0x8554, 0x5FC, 0x5594, 0xAED8, 0xA492, 0x18EE, 0x0}},
    {63880, 2, {0x3E7C, 0x0, 0x3E7C, 0x2A54, 0x2B54, 0x80, 0x3FFC, 0x2240, 0x3FF8, 0x2248, 0x3FF8, 0x2852, 0x2F62, 0x4842, 0x9F3E, 0x0}},
    {63881, 2, {0xE40, 0x7842, 0x87E, 0x7E92, 0xC22, 0x1A42, 0x299E, 0x4A44, 0x4A0, 0x1898, 0x6496, 0x3E0, 0x490, 0x3B8C, 0x1104, 0x0}},
    {63882, 2, {0x100, 0x100, 0x100, 0x108, 0x3FFC, 0x108, 0x108, 0x108, 0x108, 0x208, 0x208, 0x408, 0x888, 0x1070, 0x6020, 0x0}},
    {63883, 2, {0x3FFE, 0x2108, 0x2E70, 0x2214, 0x3FFE, 0x2210, 0x2738, 0x2AD4, 0x3292, 0x27F8, 0x2408, 0x27F8, 0x2408, 0x47F8, 0x8408, 0x0}},
    {63884, 2, {0x2004, 0x3FFE, 0x2084, 0x2F78, 0x2210, 0x3FFE, 0x2210, 0x2738, 0x2AD6, 0x3212, 0x2040, 0x2440, 0x247C, 0x4440, 0x9FFE, 0x0}},
    {63885, 2, {0x2124, 0x2124, 0xFD74, 0x22DA, 0xFB74, 0xA954, 0xFAFA, 0xAF56, 0xF820, 0xA824, 0x23FE, 0xFEB0, 0x2128, 0x2226, 0x2422, 0x0}},
    {63886, 2, {0xC00, 0x808, 0xFFC, 0x1080, 0x2080, 0x4888, 0xFFC, 0x880, 0x880, 0x884, 0x7FFE, 0x80, 0x80, 0x80, 0x80, 0x0}},
    {63887, 2, {0x124C, 0x1148, 0x1150, 0x17FE, 0x58D0, 0x5548, 0x5246, 0x9208, 0x13BE, 0x1488, 0x1AA8, 0x117E, 0x1108, 0x1208, 0x1408, 0x0}},
    {63888, 2, {0x1204, 0x1104, 0x27C8, 0x701C, 0x23C8, 0x4812, 0xFBFE, 0x5254, 0xAA6A, 0xABAA, 0x400, 0x2504, 0x2492, 0x2412, 0x43F8, 0x0}},
    {63889, 2, {0x1210, 0x1214, 0x13D2, 0xFA90, 0x12FE, 0x1590, 0x1290, 0x1990, 0xF128, 0x5246, 0x1C82, 0x1000, 0x9324, 0x7492, 0x2892, 0x0}},
    {63890, 2, {0x4020, 0x2824, 0x27FE, 0x420, 0x81FC, 0x5124, 0x55FC, 0x2F24, 0x25FC, 0xC420, 0x47FE, 0x4420, 0xC420, 0xDA20, 0x51FE, 0x0}},
    {63891, 2, {0x1020, 0x1024, 0x13FE, 0x1020, 0x57FE, 0x5A22, 0x5326, 0x52AA, 0x93FE, 0x1222, 0x1870, 0x24A8, 0x2526, 0x4222, 0x8020, 0x0}},
    {63892, 2, {0x10, 0x1410, 0xFAFE, 0x2210, 0x207E, 0x2052, 0x207E, 0xFF52, 0x217E, 0x2152, 0x2110, 0x39FE, 0xE110, 0x4690, 0x47E, 0x0}},
    {63893, 2, {0x20, 0xF0, 0x3F00, 0x104, 0xFFFE, 0x540, 0x920, 0x3118, 0xC076, 0x1F80, 0x108, 0xFFFC, 0x100, 0x100, 0x100, 0x0}},
    {63894, 2, {0x1024, 0x13FE, 0x2020, 0x27FE, 0x4622, 0xFB26, 0x12AA, 0x2BFE, 0xFA22, 0x4420, 0x70, 0x28A8, 0x5526, 0x5622, 0x8020, 0x0}},
    {63895, 2, {0x488, 0xFE88, 0x2552, 0x27BC, 0x3C88, 0x2554, 0x27DC, 0x3C22, 0x2554, 0x2754, 0x2D54, 0xF5DC, 0x4454, 0x490, 0x510, 0x0}},
    {63896, 2, {0x810, 0x3F7C, 0x810, 0x7F7E, 0x1428, 0x22C4, 0x4082, 0x3FFE, 0x888, 0xFF8, 0x888, 0xFF8, 0x80, 0x7FFE, 0x80, 0x0}},
    {63897, 2, {0x420, 0x7FFE, 0x420, 0x2040, 0x17FC, 0x1040, 0x3F8, 0x248, 0x73F8, 0x1248, 0x13F8, 0x1040, 0x17FC, 0x6840, 0x47FE, 0x0}},
    {63898, 2, {0x40, 0x4044, 0x37FE, 0x1040, 0x3FC, 0x244, 0x13FC, 0xFA44, 0x13FC, 0x1244, 0x1040, 0x17FE, 0x3040, 0xC840, 0x87FE, 0x0}},
    {63899, 2, {0x1020, 0x1824, 0x15FE, 0x2220, 0x41FE, 0xFD22, 0x11A6, 0x116A, 0xFFFE, 0x9522, 0x5870, 0x50A8, 0x1F26, 0xF222, 0x4020, 0x0}},
    {63900, 2, {0x102, 0x7F82, 0x822, 0x822, 0x922, 0xFA2, 0x1122, 0x3122, 0x4922, 0xA22, 0x422, 0x402, 0x812, 0x100E, 0x6004, 0x0}},
    {63901, 2, {0x100, 0x100, 0x1918, 0x1106, 0x211A, 0x4020, 0x1C0, 0x680, 0x1884, 0x67FE, 0x84, 0x104, 0x224, 0xC1C, 0x3008, 0x0}},
    {63902, 2, {0x202, 0x93FE, 0xFA22, 0x9222, 0x9222, 0x93FA, 0x9222, 0x9222, 0x9262, 0x9252, 0xF28A, 0x930A, 0x202, 0x3FE, 0x202, 0x0}},
    {63903, 2, {0x104, 0x7F84, 0x824, 0x824, 0xFA4, 0x1124, 0x3124, 0x4924, 0xA04, 0x41C, 0x7808, 0x0, 0x2444, 0x2222, 0x4222, 0x0}},
    {63904, 2, {0x104, 0x7FA4, 0x824, 0x1F24, 0x3124, 0xA24, 0x3D1C, 0x88, 0x7FFE, 0x280, 0x448, 0xC50, 0x7420, 0x718, 0xC06, 0x0}},
    {63905, 2, {0x2108, 0x1088, 0x1090, 0xFE20, 0x1FC, 0x104, 0xFD04, 0x104, 0xFDFC, 0x154, 0xFC50, 0x8450, 0x8492, 0xFD12, 0x860E, 0x0}},
    {63906, 2, {0x2104, 0x3FFE, 0x2220, 0x2144, 0x2FFE, 0x2140, 0x2FF8, 0x2148, 0x3FFE, 0x2148, 0x2FF8, 0x2140, 0x2360, 0x4550, 0x994C, 0x0}},
    {63907, 2, {0x100, 0x280, 0x440, 0x820, 0x1110, 0x608C, 0x0, 0xFE0, 0x20, 0x140, 0x480, 0x2484, 0x2412, 0x4412, 0x3F8, 0x0}},
    {63908, 2, {0x1020, 0x1020, 0x1050, 0xFC88, 0x1126, 0x1212, 0x15FC, 0x1808, 0xF010, 0x5010, 0x12C4, 0x12A2, 0x928A, 0x7488, 0x207C, 0x0}},
    {63909, 2, {0x420, 0xFE20, 0x2050, 0x2088, 0x2576, 0x3E02, 0x25DC, 0x6554, 0x9554, 0x15DC, 0x800, 0x888, 0x1088, 0x2156, 0xC222, 0x0}},
    {63910, 2, {0x1040, 0x1F7E, 0x2890, 0x7FFE, 0x2420, 0x2FFC, 0x2240, 0x2FF8, 0x2248, 0x3FFE, 0x2248, 0x2FF8, 0x2650, 0x4A48, 0xB246, 0x0}},
    {63911, 2, {0x124, 0x4E48, 0x2A48, 0x1124, 0x33FC, 0x4A54, 0x8AA4, 0x1BFC, 0x2840, 0x4B68, 0x8AD8, 0xB6A, 0x4A5A, 0x3B66, 0x1242, 0x0}},
    {63912, 2, {0x80, 0x180, 0x240, 0x420, 0x910, 0x1088, 0x2086, 0x4012, 0x1FF8, 0x10, 0x20, 0x640, 0x180, 0xC0, 0x40, 0x0}},
    {63913, 2, {0x2004, 0x3FFE, 0x2104, 0x2184, 0x2244, 0x2524, 0x389C, 0x2004, 0x2FF4, 0x2044, 0x2384, 0x2084, 0x2084, 0x3FFC, 0x2004, 0x0}},
    {63914, 2, {0x100, 0x2084, 0x3FFE, 0x2504, 0x54A8, 0x27E4, 0x0, 0x1FF8, 0x1248, 0x7FFE, 0x0, 0x3FFC, 0x80, 0x780, 0x100, 0x0}},
    {63915, 2, {0x1084, 0x1084, 0x1FFC, 0x0, 0xBFE, 0x1420, 0x22FC, 0x4984, 0x4FC, 0x7F84, 0x2FC, 0x484, 0x38FC, 0x448, 0x586, 0x0}},
    {63916, 2, {0x1040, 0x1040, 0x10A0, 0x1110, 0x5A08, 0x5446, 0x5422, 0x9020, 0x13FC, 0x1008, 0x1010, 0x10A0, 0x1060, 0x1030, 0x1010, 0x0}},
    {63917, 2, {0x20, 0x860, 0xFC50, 0x1088, 0x1104, 0x1242, 0x1022, 0xFC20, 0x1008, 0x11FC, 0x1408, 0x1810, 0xE0E0, 0x4010, 0x10, 0x0}},
    {63918, 2, {0x810, 0x2A54, 0x2C58, 0x4890, 0x1428, 0x2244, 0x7FFE, 0x4004, 0x3FF8, 0x100, 0x100, 0x1FF8, 0x120, 0x114, 0x7FFE, 0x0}},
    {63919, 2, {0x4420, 0x2420, 0x2850, 0xFE50, 0x1088, 0x10A6, 0x7D12, 0x1010, 0x10FC, 0xFE04, 0x1008, 0x10D0, 0x2020, 0x4010, 0x8010, 0x0}},
    {63920, 2, {0x220, 0xFF20, 0x2450, 0x2450, 0x3C88, 0x2546, 0x2622, 0x3C20, 0x2400, 0x25FC, 0x2E08, 0xF448, 0x4430, 0x410, 0x410, 0x0}},
    {63921, 2, {0x1020, 0x1820, 0x1450, 0x2288, 0x4106, 0xFE42, 0x1020, 0x1020, 0xFE04, 0x95FE, 0x5808, 0x5050, 0x1E20, 0xF010, 0x4010, 0x0}},
    {63922, 2, {0x1FF8, 0x4102, 0x7FFE, 0x8104, 0x3D78, 0x100, 0x3D78, 0x6C0, 0x1930, 0xE08E, 0xFF0, 0x20, 0x40, 0x380, 0x60, 0x0}},
    {63923, 2, {0x1FF8, 0x4102, 0x7FFE, 0x8104, 0x3D78, 0x100, 0x7BDE, 0x4A52, 0x7BDE, 0x0, 0x7FFC, 0x910, 0x1930, 0x2548, 0xFFFE, 0x0}},
    {63924, 2, {0x804, 0x9FE, 0x820, 0x14FC, 0x2284, 0x4984, 0x84FC, 0x484, 0x7FFC, 0x284, 0x2484, 0x18FC, 0x884, 0x858, 0x386, 0x0}},
    {63925, 2, {0x800, 0x842, 0xFE2, 0x1212, 0x1252, 0x33F2, 0x5452, 0x9652, 0x1952, 0x1092, 0x1092, 0x1112, 0x1102, 0x121E, 0x1404, 0x0}},
    {63926, 2, {0x2050, 0x11FC, 0x1154, 0xFDFC, 0x554, 0x9FC, 0x1000, 0x33FE, 0x5800, 0x95FC, 0x1504, 0x11FC, 0x1088, 0x1050, 0x13FE, 0x0}},
    {63927, 2, {0x428, 0xFEFE, 0x28AA, 0x28FE, 0xFEAA, 0xAAFE, 0xAA00, 0xABFE, 0xCE00, 0x82FC, 0xFE84, 0x82FC, 0x8284, 0xFE48, 0x83FE, 0x0}},
    {63928, 2, {0x1020, 0x1024, 0xFFFE, 0x1824, 0x3424, 0xD3FE, 0x1024, 0x7DFC, 0x124, 0xFEA2, 0x10B4, 0x5868, 0x97A4, 0xB122, 0x1062, 0x0}},
    {63929, 2, {0x4, 0x7FFE, 0x240, 0x3E7C, 0x2004, 0x2004, 0x3E7C, 0x240, 0x7FFE, 0x0, 0x480, 0x2444, 0x2452, 0x2412, 0x43F8, 0x0}},
    {63930, 2, {0x8, 0x3FFC, 0x8, 0x10, 0x20, 0xC0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x880, 0x780, 0x100, 0x0}},
    {63931, 2, {0x840, 0x844, 0xFFE, 0x10A2, 0x1514, 0x33F8, 0x560C, 0x9BFA, 0x1208, 0x13F8, 0x1040, 0x1248, 0x1444, 0x19C2, 0x1042, 0x0}},
    {63932, 2, {0x100, 0x2084, 0x3FFE, 0x2104, 0x5FFC, 0x2428, 0x1FF0, 0x2818, 0x4FF6, 0x810, 0xFF0, 0x80, 0xCB0, 0x138C, 0x2104, 0x0}},
    {63933, 2, {0x1004, 0x1FFE, 0x1004, 0x1004, 0x1FFC, 0x1044, 0x1040, 0x1F4C, 0x1268, 0x1250, 0x1250, 0x2448, 0x2446, 0x49C2, 0x4080, 0x0}},
    {63934, 2, {0x808, 0x4908, 0x4948, 0x2A28, 0x2828, 0xA08, 0xFF48, 0x828, 0xC28, 0x1A0E, 0x2A38, 0x49C8, 0x8888, 0x808, 0x808, 0x0}},
    {63935, 2, {0x2108, 0x2108, 0x47D0, 0x5454, 0xE7F8, 0x2448, 0x5454, 0xFFFE, 0x4550, 0x104, 0xFFFE, 0x520, 0x910, 0x310E, 0xC104, 0x0}},
    {63936, 2, {0x1040, 0x1040, 0x17FE, 0x14A2, 0x5314, 0x55F8, 0x5B0C, 0x55FA, 0x9108, 0x11F8, 0x1820, 0x2528, 0x2624, 0x44E2, 0x8842, 0x0}},
    {63937, 2, {0x80, 0x1044, 0x1FFE, 0x9040, 0x5FFE, 0x5514, 0x13F8, 0x120C, 0x37FA, 0xDA08, 0x93F8, 0x2040, 0x2250, 0x45C8, 0x8886, 0x0}},
    {63938, 2, {0x420, 0x7FFE, 0x420, 0x0, 0x7EFC, 0x1224, 0xE1C, 0x3264, 0x380, 0xC60, 0x771E, 0x60, 0xF98, 0x60, 0x3F80, 0x0}},
    {63939, 2, {0x40, 0x4044, 0x37FE, 0x1294, 0x108, 0x3FC, 0x150A, 0xF9F8, 0x1108, 0x11F8, 0x1128, 0x1226, 0x34E2, 0xC840, 0x87FE, 0x0}},
    {63940, 2, {0x1040, 0x844, 0x7F7E, 0x2240, 0x147C, 0x7F04, 0x7C, 0x3E40, 0x227C, 0x3E40, 0x227C, 0x3E40, 0x227E, 0x2E42, 0x243E, 0x0}},
    {63941, 2, {0x1010, 0x1FF8, 0x1010, 0x1FF0, 0x5012, 0x7FFE, 0x8104, 0x7FFC, 0x1110, 0x1FF0, 0x1110, 0x1FF0, 0x100, 0x7FFE, 0x100, 0x0}},
    {63942, 2, {0x4408, 0x7DFC, 0x4400, 0x4800, 0x5004, 0x53FE, 0x4890, 0x4890, 0x4490, 0x4490, 0x6490, 0x5912, 0x4212, 0x4412, 0x480E, 0x0}},
    {63943, 2, {0x1802, 0x67C2, 0x4252, 0x4A52, 0x5A52, 0x6592, 0x9E12, 0x2112, 0xDED2, 0x412, 0x7F92, 0x2492, 0x1502, 0x17DE, 0x7C04, 0x0}},
    {63944, 2, {0x1008, 0x13FC, 0x1088, 0x1088, 0xFC88, 0x1088, 0x1888, 0x37F8, 0x5488, 0x5088, 0x9088, 0x9088, 0x1088, 0x1088, 0x17FE, 0x0}},
    {63945, 2, {0x1040, 0x10D2, 0x131E, 0x1252, 0xFE52, 0x1252, 0x1A52, 0x3652, 0x5252, 0x53D2, 0x9252, 0x9056, 0x1092, 0x1110, 0x1210, 0x0}},
    {63946, 2, {0x2080, 0x1044, 0x17FE, 0x40, 0x4090, 0x2908, 0x2BFC, 0x1004, 0x1248, 0x6248, 0x2248, 0x224A, 0x624A, 0x644A, 0x2846, 0x0}},
    {63947, 2, {0x4080, 0x25FE, 0x2612, 0x412, 0x8492, 0x5562, 0x564E, 0x2484, 0x2000, 0xC7FC, 0x4444, 0x47FC, 0xC444, 0xC7FC, 0x4404, 0x0}},
    {63948, 2, {0x40, 0x824, 0xFDFE, 0x1020, 0x1040, 0x1088, 0x11FC, 0xFC04, 0x1128, 0x1128, 0x1528, 0x1928, 0xE22A, 0x442A, 0x826, 0x0}},
    {63949, 2, {0x200, 0x2704, 0x38FE, 0x2044, 0x2244, 0x7D5C, 0x2088, 0x1FF8, 0x1088, 0x1088, 0x1FF8, 0x1088, 0x1088, 0x1FF8, 0x1008, 0x0}},
    {63950, 2, {0x40, 0x424, 0xFFFE, 0x1040, 0x1088, 0x23FC, 0x3C84, 0x6400, 0xA4A8, 0x24A8, 0x24A8, 0x24A8, 0x3D2A, 0x222A, 0x426, 0x0}},
    {63951, 2, {0x1000, 0x1004, 0x23FE, 0x2444, 0x4444, 0xF844, 0x1044, 0x2844, 0xFDFC, 0x4484, 0x84, 0x2884, 0x5484, 0x5484, 0x87FE, 0x0}},
    {63952, 2, {0x4904, 0x2AFE, 0x2A10, 0x87C, 0x7F44, 0x1C44, 0x2A7C, 0x4944, 0xA7C, 0x944, 0x7F44, 0x87C, 0x1444, 0x2228, 0x42C6, 0x0}},
    {63953, 2, {0x200, 0x100, 0x180, 0x80, 0x4, 0x7FFE, 0x0, 0x0, 0x420, 0x420, 0x810, 0x810, 0x1008, 0x2006, 0x4002, 0x0}},
    {63954, 2, {0xF720, 0x1128, 0x5524, 0x3320, 0x5526, 0x99F8, 0x820, 0x1422, 0xE3A2, 0x824, 0x7228, 0xD12, 0x722A, 0xC46, 0xF182, 0x0}},
    {63955, 2, {0x8840, 0xF848, 0x8BFC, 0x9040, 0xA044, 0xA7FE, 0x9090, 0x910C, 0x8A44, 0x8840, 0xC848, 0xB3FC, 0x8040, 0x8044, 0x8FFE, 0x0}},
    {63956, 2, {0x840, 0x860, 0x890, 0x1108, 0x1216, 0x35FA, 0x5000, 0x97FE, 0x1492, 0x1492, 0x17FE, 0x1492, 0x1492, 0x1492, 0x1486, 0x0}},
    {63957, 2, {0x100, 0x2108, 0x2108, 0x3FF8, 0x2108, 0x280, 0xC60, 0xF7DE, 0x0, 0x3FF8, 0x2448, 0x3FF8, 0x2448, 0x2448, 0x2418, 0x0}},
    {63958, 2, {0x2040, 0x10C0, 0x10A0, 0x110, 0x4208, 0x2DF6, 0x2800, 0x17FC, 0x14A4, 0x64A4, 0x27FC, 0x24A4, 0x64A4, 0x64BC, 0x2408, 0x0}},
    {63959, 2, {0x1020, 0x1020, 0xFE50, 0x1088, 0x7D06, 0x56FA, 0x5400, 0x7DFC, 0x5554, 0x7D54, 0x55FC, 0x1154, 0xFF54, 0x1154, 0x110C, 0x0}},
    {63960, 2, {0x1848, 0x13FC, 0x2048, 0x4048, 0x9FFE, 0x1048, 0x23F8, 0x6048, 0xA040, 0x23FC, 0x2040, 0x2044, 0x27FE, 0x2040, 0x2040, 0x0}},
    {63961, 2, {0x1004, 0x17FE, 0x1090, 0x13FE, 0x5A92, 0x5692, 0x53FE, 0x9202, 0x1024, 0x17FE, 0x1020, 0x11B0, 0x1128, 0x1226, 0x1422, 0x0}},
    {63962, 2, {0x4, 0x7FFE, 0x240, 0x1FF8, 0x1248, 0x1248, 0x1FF8, 0x1088, 0x84, 0x7FFE, 0x2C0, 0x4A0, 0x890, 0x108E, 0x6084, 0x0}},
    {63963, 2, {0x100, 0x84, 0x7FFE, 0x104, 0x2248, 0x1790, 0x100, 0x1A50, 0x67EC, 0x4524, 0x100, 0x7FFE, 0x100, 0x100, 0x100, 0x0}},
    {63964, 2, {0x8908, 0xFDFC, 0x8A08, 0x9090, 0xA060, 0xA198, 0x9606, 0x91F8, 0x8800, 0x8928, 0xCBFC, 0xB420, 0x81FC, 0x8020, 0x87FE, 0x0}},
    {63965, 2, {0x102, 0x782, 0x7C12, 0x412, 0x492, 0xFFD2, 0x412, 0xE12, 0x1512, 0x2492, 0x4492, 0x8402, 0x412, 0x40E, 0x404, 0x0}},
    {63966, 2, {0x80, 0x84, 0x7FFE, 0x80, 0x1084, 0x1FFE, 0x1084, 0x1084, 0x1FFC, 0x1084, 0x880, 0x480, 0x300, 0x4C0, 0x783E, 0x0}},
    {63967, 2, {0x3FFC, 0x2004, 0x3FFC, 0x2080, 0x24FE, 0x2984, 0x32FC, 0x2284, 0x24FC, 0x2C40, 0x347C, 0x24C8, 0x2530, 0x4448, 0x8586, 0x0}},
    {63968, 2, {0x810, 0xFF8, 0x810, 0xFF0, 0x810, 0xFF0, 0x810, 0x404, 0xFFE, 0x1124, 0x2224, 0x4444, 0x884, 0x1138, 0x210, 0x0}},
    {63969, 2, {0x100, 0x104, 0x7FFE, 0x540, 0x920, 0x1118, 0x6106, 0xFF0, 0x40, 0x84, 0x7FFE, 0x80, 0x480, 0x380, 0x100, 0x0}},
    {63970, 2, {0x304, 0x3C24, 0x424, 0x7FA4, 0xC24, 0x1624, 0x251C, 0x4488, 0x84, 0x7FFE, 0x2C0, 0x4A0, 0x890, 0x108E, 0x2084, 0x0}},
    {63971, 2, {0x2204, 0x13FE, 0x1204, 0x204, 0x43FC, 0x2A04, 0x2A40, 0x124C, 0x1248, 0x6250, 0x2260, 0x2440, 0x6442, 0x6842, 0x303E, 0x0}},
    {63972, 2, {0x104, 0x9FE, 0xFD24, 0x1124, 0x11FC, 0x1124, 0x1124, 0xFDFC, 0x1124, 0x1020, 0x1420, 0x19FC, 0xE020, 0x4024, 0x7FE, 0x0}},
    {63973, 2, {0x100, 0x1084, 0x1FFE, 0x9080, 0x51C2, 0x5712, 0x1112, 0x17D2, 0x3112, 0xD192, 0x9352, 0x2552, 0x2902, 0x411E, 0x8104, 0x0}},
    {63974, 2, {0x3FFC, 0x2244, 0x2244, 0x3FFC, 0x2044, 0x1120, 0x11FE, 0x5B20, 0x5520, 0x91FC, 0x1120, 0x11FC, 0x1120, 0x1124, 0x11FE, 0x0}},
    {63975, 2, {0x104, 0x7FFE, 0x0, 0xFF8, 0x888, 0xFF8, 0x888, 0x3FFC, 0x80, 0x7FFE, 0x288, 0x450, 0x1CA0, 0x6518, 0x606, 0x0}},
    {63976, 2, {0x4104, 0x21FE, 0x2924, 0xFD24, 0x9FC, 0x1124, 0x2524, 0x29FC, 0x7124, 0xAC20, 0x2420, 0x21FC, 0x2020, 0x2024, 0x27FE, 0x0}},
    {63977, 2, {0x1008, 0x1FFC, 0x1108, 0x1108, 0x1FF8, 0x1108, 0x1108, 0x1FF8, 0x1108, 0x100, 0x108, 0x3FFC, 0x100, 0x104, 0x7FFE, 0x0}},
    {63978, 2, {0x1050, 0x848, 0xFFC8, 0x557E, 0x4948, 0x49C8, 0x557E, 0x7F48, 0x848, 0x7F48, 0x497E, 0x5548, 0x7B48, 0x417E, 0x4740, 0x0}},
    {63979, 2, {0x4004, 0x7FFE, 0x4220, 0x4220, 0x5FFC, 0x42A0, 0x4104, 0x5FFE, 0x4200, 0x47F8, 0x4A08, 0x5208, 0x43F8, 0x4208, 0x7FFE, 0x0}},
    {63980, 2, {0x4082, 0x27BE, 0x2082, 0x82, 0x879E, 0x5492, 0x5410, 0x27BE, 0x28A2, 0xC492, 0x4186, 0x469A, 0xD8E2, 0xC71C, 0x4208, 0x0}},
    {63981, 2, {0x100, 0x84, 0x7FFE, 0x810, 0x420, 0x240, 0x180, 0xE70, 0x700E, 0x1FF8, 0x1008, 0x1008, 0x1008, 0x1FF8, 0x1008, 0x0}},
    {63982, 2, {0x1222, 0x1124, 0x10A8, 0x13FE, 0x54A8, 0x5924, 0x5222, 0x5108, 0x91DE, 0x1248, 0x19A8, 0x24BE, 0x2508, 0x4208, 0x8408, 0x0}},
    {63983, 2, {0x222, 0x1124, 0xF8A8, 0x23FE, 0x20A8, 0x2124, 0x2222, 0xFA08, 0x23BE, 0x2288, 0x26A8, 0x39FE, 0xE108, 0x4208, 0x408, 0x0}},
    {63984, 2, {0x420, 0x7FFE, 0x420, 0x3E7C, 0x2244, 0x3E7C, 0x2244, 0x3E7C, 0x27E4, 0x2C84, 0x37E4, 0x2484, 0x27E4, 0x249C, 0x27E8, 0x0}},
    {63985, 2, {0x8A4C, 0xF948, 0x8950, 0x97FE, 0xA150, 0xA248, 0x9C46, 0x9208, 0x8BBE, 0x8C88, 0xCAA8, 0xB13E, 0x8208, 0x8408, 0x8808, 0x0}},
    {63986, 2, {0x2124, 0x3D24, 0x24A8, 0x4BFC, 0xFCA8, 0x5524, 0x5622, 0x7D08, 0x55FE, 0x5548, 0x7F48, 0x44E8, 0x54BE, 0xAB08, 0xAA08, 0x0}},
    {63987, 2, {0x8A4, 0x4464, 0x7F28, 0x54FE, 0x7F28, 0x5564, 0x55A2, 0x7F04, 0x54EE, 0x54A4, 0x5D24, 0x56B4, 0x545E, 0x5D84, 0x9704, 0x0}},
    {63988, 2, {0x820, 0x820, 0x820, 0x824, 0x7FFE, 0x820, 0x820, 0x1CB0, 0x2AA8, 0x2928, 0x4924, 0x4A24, 0xC22, 0x822, 0x820, 0x0}},
    {63989, 2, {0x4210, 0x2210, 0x2210, 0x294, 0x8FFE, 0x5210, 0x5310, 0x2290, 0x26B8, 0xCA54, 0x5252, 0x5212, 0xC210, 0xC210, 0x4210, 0x0}},
    {63990, 2, {0x8480, 0xFE84, 0x90FE, 0x9100, 0x92F8, 0xFC88, 0x8488, 0x84F8, 0x8400, 0xFFDE, 0x9252, 0x9252, 0x9252, 0xFFDE, 0x8252, 0x0}},
    {63991, 2, {0x200, 0x100, 0x80, 0x88, 0x3FFC, 0x0, 0x810, 0x810, 0x420, 0x420, 0x240, 0x240, 0x84, 0x7FFE, 0x0, 0x0}},
    {63992, 2, {0x1020, 0x1020, 0x1F3E, 0x2850, 0x4488, 0x100, 0x80, 0x88, 0x3FFC, 0x420, 0x420, 0x240, 0x240, 0x84, 0x7FFE, 0x0}},
    {63993, 2, {0x1040, 0x9220, 0x9220, 0x5404, 0x55FE, 0x1000, 0xFE84, 0x1084, 0x1888, 0x3448, 0x5248, 0x5250, 0x9010, 0x9010, 0x13FE, 0x0}},
    {63994, 2, {0x440, 0x2450, 0x2448, 0x2448, 0x2440, 0x3FFE, 0x450, 0x450, 0x7C50, 0x2450, 0x2448, 0x2448, 0x2484, 0x4506, 0x8602, 0x0}},
    {63995, 2, {0x200, 0x210, 0x7F8, 0xC20, 0x3240, 0x880, 0x700, 0xC80, 0x7098, 0x890, 0x8A0, 0x1140, 0x220, 0xC18, 0x3006, 0x0}},
    {63996, 2, {0x2110, 0x1090, 0x13F4, 0xFD52, 0x152, 0x90, 0x7BFE, 0x10, 0x7BD2, 0x252, 0x7BD4, 0x4A48, 0x4BCA, 0x7A56, 0x4822, 0x0}},
    {63997, 2, {0x840, 0x840, 0x840, 0x1040, 0x1040, 0x3040, 0x57FE, 0x9040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x1040, 0x0}},
    {63998, 2, {0x420, 0x420, 0x7FFE, 0x420, 0x580, 0x240, 0x420, 0x1898, 0x6086, 0x1FF8, 0x80, 0x4A0, 0x890, 0x108E, 0x6084, 0x0}},
    {63999, 2, {0x802, 0x882, 0xFFD2, 0x812, 0x8892, 0xFFD2, 0x8892, 0x8892, 0x8B92, 0x8892, 0x1C12, 0x2A02, 0x4992, 0x888E, 0x804, 0x0}},
    {64000, 2, {0x2000, 0x2004, 0x23FE, 0x2044, 0x2444, 0x2E44, 0xF044, 0x2044, 0x2244, 0x2444, 0x2844, 0x3084, 0x2144, 0x238, 0x410, 0x0}},
    {64001, 2, {0x100, 0x2084, 0x3FFE, 0x2220, 0x2224, 0x3FFE, 0x2220, 0x23E0, 0x2000, 0x27F8, 0x2210, 0x2120, 0x20C0, 0x4330, 0x9C0E, 0x0}},
    {64002, 2, {0x1000, 0x1004, 0x13FE, 0xFC40, 0x1040, 0x1040, 0x1484, 0x19FE, 0xF284, 0x5484, 0x1084, 0x1084, 0x9084, 0x70FC, 0x2084, 0x0}},
    {64003, 2, {0x1040, 0x1124, 0x95FE, 0x5520, 0x59FC, 0x1124, 0xFDFE, 0x1124, 0x19FC, 0x3520, 0x55FC, 0x9284, 0x1284, 0x14FC, 0x1884, 0x0}},
    {64004, 2, {0x200, 0x4108, 0x7FFC, 0x4008, 0x8090, 0x1C0, 0x7E00, 0x410, 0x438, 0x7C0, 0xFC00, 0x404, 0x404, 0x406, 0x3FC, 0x0}},
    {64005, 2, {0x2204, 0x13FE, 0x1204, 0x204, 0x46F4, 0x2A04, 0x2A04, 0x12F4, 0x1294, 0x6294, 0x22F4, 0x2294, 0x6204, 0x621C, 0x2208, 0x0}},
    {64006, 2, {0x1FF0, 0x1010, 0x1FF0, 0x1010, 0x1FF0, 0x440, 0x3FFC, 0x440, 0x7FFE, 0x1110, 0x2928, 0x4744, 0x5C0, 0x1930, 0x630C, 0x0}},
    {64007, 2, {0x11FE, 0x1000, 0xFEFC, 0x1084, 0x7C84, 0x54FC, 0x7C00, 0x55FE, 0x7D22, 0x5522, 0x11FE, 0xFF22, 0x1122, 0x11FE, 0x1102, 0x0}},
    {64008, 2, {0x1808, 0x11FC, 0x2000, 0x4000, 0xC04, 0xBFE, 0x1010, 0x3010, 0x5010, 0x9010, 0x1010, 0x1010, 0x1090, 0x1070, 0x1020, 0x0}},
    {64009, 2, {0x4488, 0x7CFC, 0x4488, 0x4950, 0x5020, 0x5050, 0x4B8E, 0x4820, 0x45FC, 0x44A0, 0x6524, 0x5BFE, 0x4020, 0x4020, 0x4020, 0x0}},
    {64010, 2, {0x20, 0x1FF0, 0x1020, 0x1020, 0x1FE0, 0x1020, 0x1020, 0x1FE0, 0x1020, 0x1020, 0x1FE0, 0x14A4, 0x884, 0x3086, 0xC07C, 0x0}},
    {64011, 2, {0x2084, 0x3FFE, 0x2200, 0x3FDE, 0x2012, 0x2F94, 0x2894, 0x2F98, 0x2014, 0x3FD2, 0x2112, 0x221A, 0x2F94, 0x4210, 0x8E10, 0x0}},
    {64012, 2, {0x8, 0xFFFC, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x880, 0x1080, 0x1084, 0x2084, 0x4086, 0x807C, 0x0, 0x0}},
    {64013, 2, {0x848, 0x87C, 0x7F48, 0x848, 0x3F48, 0x8E, 0x7F00, 0x82FE, 0x3E44, 0x44, 0x3E28, 0x2210, 0x2228, 0x3E46, 0x2282, 0x0}},
    {64014, 2, {0x1050, 0x848, 0xFEFE, 0x4590, 0x26FC, 0x1890, 0x8FC, 0x1490, 0x24FE, 0xC080, 0x3FF0, 0x820, 0x440, 0x380, 0xC60, 0xF01C}},
    {64015, 2, {0x2090, 0x2088, 0x2104, 0x2240, 0xF840, 0x20A0, 0x2090, 0x2108, 0x2204, 0x25FA, 0x2108, 0x3908, 0xC108, 0x108, 0x1F8, 0x108}},
    {64016, 2, {0x1202, 0x13FE, 0x1404, 0x1000, 0x13FE, 0xFC40, 0x1086, 0x1744, 0x12A8, 0x1130, 0x1EE8, 0xE0A8, 0x4124, 0x622, 0xC0, 0x0}},
    {64017, 2, {0x2040, 0x2020, 0x2BFE, 0xA800, 0xA908, 0xA890, 0xAFFE, 0xA808, 0xABC8, 0xAA48, 0xAA48, 0xBBC8, 0xCA48, 0x8, 0x28, 0x10}},
    {64018, 2, {0x20, 0x4BFE, 0x7C20, 0x49FC, 0x4820, 0x4FFE, 0x7800, 0x49FC, 0x4904, 0x49FC, 0x4904, 0x79FC, 0x4904, 0x11C, 0x108, 0x0}},
    {64019, 2, {0x1088, 0x1088, 0x13FE, 0x1088, 0xFC88, 0x1044, 0x3884, 0x357E, 0x5104, 0x5324, 0x9514, 0x1114, 0x1104, 0x1104, 0x1114, 0x1108}},
    {64020, 2, {0x1244, 0x1124, 0x1088, 0x13FE, 0xFC90, 0x1108, 0x3A14, 0x34E2, 0x5020, 0x51F8, 0x9020, 0x13FC, 0x1020, 0x1020, 0x10A0, 0x1040}},
    {64021, 2, {0x8BD4, 0x6A9E, 0x2A94, 0xA94, 0x1BD4, 0x2A5C, 0x2A54, 0xCBD0, 0x4A92, 0x5292, 0x63EE, 0x0, 0x2444, 0x2222, 0x4222, 0x0}},
    {64022, 2, {0x40, 0x4C40, 0x2846, 0x11F4, 0x3048, 0x4850, 0x8BFE, 0x1840, 0x29FC, 0x4B04, 0x8904, 0x9FC, 0x4904, 0x39FC, 0x1104, 0x0}},
    {64023, 2, {0x830, 0x420, 0x240, 0x244, 0x7FFE, 0x420, 0x810, 0x100C, 0x6004, 0x1FF8, 0x1248, 0x1248, 0x1248, 0x1248, 0x7FFE, 0x0}},
    {64024, 2, {0x40, 0x3E40, 0x40, 0x40, 0x7F40, 0x840, 0x2A40, 0x2A40, 0x2A40, 0x2A40, 0x4A40, 0x4A40, 0x842, 0x842, 0x83E, 0x0}},
    {64025, 2, {0x2020, 0x1020, 0x1124, 0xFDFE, 0x524, 0x924, 0x11FC, 0x3124, 0x5924, 0x9524, 0x15FC, 0x1124, 0x1020, 0x1020, 0x1020, 0x0}},
    {64026, 2, {0x210C, 0x1088, 0x1050, 0xFFFE, 0x420, 0x820, 0x1020, 0x31FC, 0x5820, 0x9420, 0x1424, 0x13FE, 0x1020, 0x1020, 0x1020, 0x0}},
    {64027, 2, {0x2004, 0x13FE, 0x1000, 0xFCF8, 0x488, 0x888, 0x10F8, 0x3000, 0x59FC, 0x9524, 0x15FC, 0x1124, 0x1124, 0x11FC, 0x1104, 0x0}},
    {64028, 2, {0x2020, 0x23FE, 0x1020, 0x11FC, 0xFC20, 0x3FE, 0x4400, 0x45FC, 0x2904, 0x29FC, 0x1104, 0x1DFC, 0xE104, 0x411C, 0x108, 0x0}},
    {64029, 2, {0x1024, 0x13FE, 0x9220, 0x54FC, 0x5420, 0x11FE, 0xFE00, 0x10FC, 0x1884, 0x34FC, 0x5284, 0x92FC, 0x1084, 0x109C, 0x1088, 0x0}},
    {64030, 2, {0x204, 0x7FFE, 0xE1C, 0x1214, 0x2224, 0x4244, 0x60C, 0xA14, 0x1224, 0x2244, 0x4284, 0x204, 0x1224, 0xE1C, 0x408, 0x0}},
    {64031, 2, {0x440, 0xFFFE, 0x440, 0x440, 0x3EFC, 0x2284, 0x22FC, 0x3E84, 0x22FC, 0x2240, 0x3EFE, 0x2322, 0x22D2, 0x428A, 0x4AF2, 0x840C}},
    {64032, 2, {0x820, 0xFFFE, 0x8A0, 0x1CF0, 0xF110, 0x12F8, 0x10A8, 0xFCF8, 0x10A8, 0x31FC, 0x3924, 0x55FC, 0x5524, 0x91FC, 0x1022, 0x101E}},
    {64033, 2, {0x2020, 0x2120, 0x2120, 0xF9FC, 0xAA20, 0xAC20, 0xA820, 0xA820, 0xA9FC, 0xF820, 0xA820, 0x2020, 0x2820, 0x3820, 0xCBFE, 0x0}},
    {64034, 2, {0x2040, 0x1040, 0x1044, 0xFDF4, 0x48, 0x50, 0x7BFE, 0x40, 0x79FC, 0x284, 0x7884, 0x48FC, 0x4884, 0x78FC, 0x4884, 0x0}},
    {64035, 2, {0x1008, 0x1048, 0x1028, 0xFE08, 0x1088, 0x1048, 0x1008, 0xFE0E, 0x11F8, 0x1008, 0x5C08, 0x5008, 0x5008, 0x7008, 0x8C00, 0x83FE}},
    {64036, 2, {0x0, 0x23FC, 0x1088, 0x888, 0x4090, 0x20BC, 0x1084, 0x84, 0xF0C4, 0x1148, 0x1128, 0x1210, 0x1428, 0x2844, 0x4482, 0x83FE}},
    {64037, 2, {0x110, 0x41F8, 0x3110, 0x1224, 0x7FE, 0x244, 0x1244, 0xFBFC, 0x1060, 0x10A8, 0x1126, 0x1622, 0x301E, 0xC800, 0x87FE, 0x0}},
    {64038, 2, {0x824, 0x8BE, 0x3EA4, 0x924, 0xA28, 0x7FE8, 0x428, 0x924, 0x1FA4, 0x3122, 0x5F22, 0x1132, 0x112C, 0x1F20, 0x1120, 0x0}},
    {64039, 2, {0x1020, 0x1020, 0x29FE, 0x4424, 0xB828, 0x13FE, 0x1020, 0x7C40, 0x10FC, 0x1308, 0x5410, 0x39FE, 0x1010, 0x1C10, 0xE050, 0x20}},
    {64040, 2, {0x1020, 0x1020, 0x29FC, 0x4420, 0xB820, 0x17FE, 0x1104, 0x7C88, 0x11FC, 0x1020, 0x5420, 0x3BFE, 0x1020, 0x1C20, 0xE020, 0x20}},
    {64041, 2, {0x40, 0x7C80, 0x45F8, 0x4908, 0x49F8, 0x5108, 0x51F8, 0x4900, 0x45FE, 0x4500, 0x45FC, 0x5884, 0x4494, 0x47F4, 0x4004, 0x4018}},
    {64042, 2, {0x1008, 0x189C, 0x14E0, 0x2A80, 0x4884, 0xBEFE, 0x22A4, 0x3EA4, 0x22A8, 0x3EA8, 0x20A8, 0x2490, 0x26A8, 0x7B46, 0x2282, 0x0}},
    {64043, 2, {0x1000, 0x19FE, 0x1402, 0x2A02, 0x48FA, 0xBE02, 0x22F2, 0x3E92, 0x2292, 0x3E92, 0x20F2, 0x2492, 0x2602, 0x7A1E, 0x2004, 0x0}},
    {64044, 2, {0x1020, 0x1810, 0x1492, 0x2AFE, 0x4904, 0xBEF8, 0x2288, 0x3E88, 0x22F8, 0x3E80, 0x20FC, 0x2484, 0x2684, 0x7AFC, 0x2084, 0x0}},
    {64045, 2, {0x1030, 0x1020, 0x92FC, 0xFE84, 0xA2FC, 0x2884, 0x7EFC, 0x4880, 0xFEFE, 0x4880, 0x7EFE, 0x4802, 0x48AA, 0x7F56, 0x4154, 0x0}},
    {64048, 2, {0x900, 0x904, 0xBFE, 0x1400, 0x13F8, 0x3288, 0x5248, 0x9208, 0x1FFE, 0x1288, 0x1248, 0x1208, 0x17FE, 0x1008, 0x1070, 0x0}},
    {64049, 2, {0x908, 0x890, 0x810, 0x13FE, 0x1222, 0x3326, 0x52AA, 0x93FE, 0x1000, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x0}},
    {64050, 2, {0x800, 0x880, 0x1FC0, 0x1080, 0x2110, 0x7FF8, 0xA210, 0x2210, 0x2210, 0x3FF0, 0x2290, 0x484, 0x884, 0x3086, 0xC07C, 0x0}},
    {64051, 2, {0x2020, 0x2220, 0x3F20, 0x2224, 0x44FE, 0xFF24, 0x4924, 0x4924, 0x7F24, 0x4944, 0x844, 0x1884, 0x291A, 0x4802, 0x87FE, 0x0}},
    {64052, 2, {0x2220, 0xFFA0, 0x2220, 0x3E24, 0x8FE, 0x7F24, 0x4924, 0x7F24, 0x824, 0xFFA4, 0x844, 0x7F44, 0x8A4, 0xF9C, 0xF108, 0x0}},
    {64053, 2, {0x100, 0x2208, 0x3FFC, 0x2088, 0x2088, 0x3FF8, 0x2108, 0x3FF8, 0x2288, 0x480, 0x884, 0x7FFE, 0x80, 0x80, 0x80, 0x0}},
    {64054, 2, {0x208, 0x93FC, 0xFA08, 0x93F8, 0x9208, 0x93F8, 0x9288, 0x9102, 0x93FE, 0x9622, 0xFA22, 0x9252, 0x28A, 0x3F2, 0x1C, 0x0}},
    {64055, 2, {0x110, 0x97FE, 0xF910, 0x91F0, 0x9040, 0x97FC, 0x9444, 0x97FC, 0x9040, 0x97FC, 0xF040, 0x9FFE, 0xA0, 0x318, 0xC06, 0x0}},
    {64056, 2, {0x2244, 0x3F7E, 0x2244, 0x2244, 0x3E7C, 0x120, 0x214, 0x7FFE, 0x810, 0x1208, 0x7F7E, 0x2244, 0x2244, 0x3E7C, 0x2244, 0x0}},
    {64057, 2, {0x11FE, 0x1102, 0x1102, 0x11FE, 0xFD44, 0x1128, 0x11FE, 0x1148, 0x1548, 0x19FE, 0xE248, 0x4248, 0x488, 0x908, 0x0, 0x0}},
    {64058, 2, {0x1008, 0x1FFC, 0x14A8, 0x12A8, 0x1FF8, 0x80, 0x1FFC, 0x80, 0x7FFE, 0x1448, 0x12A6, 0x2082, 0x1FF8, 0x84, 0x7FFE, 0x0}},
    {64059, 2, {0x3FFC, 0x2004, 0x2004, 0x3FFC, 0x2224, 0x2140, 0x2FFC, 0x2A94, 0x2924, 0x2FFC, 0x2408, 0x27F8, 0x4408, 0x47F8, 0x8408, 0x0}},
    {64060, 2, {0x100, 0x100, 0x100, 0x2104, 0x2104, 0x2104, 0x2104, 0x2104, 0x3FFC, 0x2104, 0x100, 0x100, 0x200, 0xC00, 0x7000, 0x0}},
    {64061, 2, {0x1200, 0x1204, 0x13FE, 0x1400, 0x5BF8, 0x5288, 0x5248, 0x9248, 0x1FFE, 0x1288, 0x1248, 0x1248, 0x17FE, 0x1008, 0x1070, 0x0}},
    {64062, 2, {0x2004, 0x24BE, 0x2788, 0x24A8, 0xB4A8, 0xAFA8, 0xA4A8, 0x27FE, 0x2410, 0x2518, 0x2498, 0x25A8, 0x2E4A, 0x248A, 0x210E, 0x0}},
    {64063, 2, {0x110C, 0x1088, 0x1090, 0x13FE, 0x5A22, 0x5726, 0x52AA, 0x93FE, 0x1202, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x11FC, 0x1104, 0x0}},
    {64064, 2, {0x110, 0x3550, 0x27D4, 0x405E, 0x97B4, 0x2014, 0x67D4, 0xA108, 0x27D4, 0x2126, 0x2FC2, 0x484, 0x2452, 0x2412, 0x43F8, 0x0}},
    {64065, 2, {0x2020, 0x2120, 0x3FA0, 0x2024, 0x407E, 0x3EC8, 0x2A48, 0x2A48, 0xFFC8, 0x5228, 0x4A30, 0x4A10, 0x7FA8, 0x246, 0xC82, 0x0}},
    {64066, 2, {0x4404, 0x7EFE, 0x4410, 0x4450, 0x7C50, 0x4490, 0x4494, 0x7DFE, 0x4410, 0x4010, 0x4830, 0x4C50, 0x5492, 0xE112, 0x420E, 0x0}},
    {64067, 2, {0xFF0, 0x810, 0xFF0, 0x810, 0xFF4, 0x908, 0x1FF0, 0x120, 0x7FFE, 0x80, 0x7F8, 0x7C08, 0x7F8, 0x408, 0x7F8, 0x0}},
    {64068, 2, {0x1100, 0x1104, 0x13FE, 0x1000, 0xFDFC, 0x1144, 0x1924, 0x3524, 0x57FE, 0x5288, 0x9248, 0x9248, 0x17FE, 0x1008, 0x1070, 0x0}},
    {64069, 2, {0x4200, 0x2204, 0x23FE, 0xC00, 0x8BF8, 0x5288, 0x5248, 0x2248, 0x2FFE, 0xC488, 0x4448, 0x4448, 0xC7FE, 0xC010, 0x40E0, 0x0}},
    {64070, 2, {0x40, 0x2040, 0x1044, 0x3F8, 0x4050, 0x2864, 0x2FFE, 0x1080, 0x11FC, 0x2304, 0x6DFC, 0x2104, 0x2104, 0x21FC, 0x2104, 0x0}},
    {64071, 2, {0x4110, 0x2FFE, 0x2110, 0x1F0, 0x8040, 0x57FC, 0x5444, 0x27FC, 0x2040, 0xC7FC, 0x4040, 0x4FFE, 0xC080, 0xC330, 0x4C0E, 0x0}},
    {64072, 2, {0x100, 0xFEC, 0x108, 0x110, 0x7FFE, 0x80, 0x7F0, 0xC10, 0x17F0, 0x6410, 0x7F0, 0x410, 0x2444, 0x2222, 0x4222, 0x0}},
    {64073, 2, {0x10, 0xF8, 0x3F00, 0x10C, 0x1088, 0x890, 0x820, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64074, 2, {0x4, 0xBFE, 0xFC40, 0x1080, 0x114C, 0x1248, 0x12B0, 0x7D20, 0x12B0, 0x1068, 0x14A8, 0x1924, 0xE222, 0x4520, 0x8C0, 0x0}},
    {64075, 2, {0x124, 0x5FE, 0xFF24, 0x1124, 0x11FC, 0x2124, 0x3D24, 0x65FC, 0xA544, 0x2490, 0x2514, 0x27FE, 0x3C10, 0x2410, 0x10, 0x0}},
    {64076, 2, {0x2020, 0x1020, 0x1020, 0xFC20, 0x420, 0x828, 0x11FC, 0x3020, 0x5820, 0x9420, 0x1420, 0x1020, 0x1020, 0x1024, 0x13FE, 0x0}},
    {64077, 2, {0x2020, 0x1020, 0x1020, 0xFC20, 0x520, 0x924, 0x113E, 0x3120, 0x5920, 0x9520, 0x1520, 0x1120, 0x1120, 0x1124, 0x17FE, 0x0}},
    {64078, 2, {0x2008, 0x109C, 0x10E0, 0xFC80, 0x480, 0x884, 0x10FE, 0x3088, 0x5888, 0x9488, 0x1488, 0x1108, 0x1108, 0x1208, 0x1408, 0x0}},
    {64079, 2, {0x2040, 0x1040, 0x1044, 0xFDFE, 0x440, 0x840, 0x1080, 0x3084, 0x59FE, 0x9684, 0x1484, 0x1084, 0x1084, 0x10FC, 0x1084, 0x0}},
    {64080, 2, {0x2108, 0x11FC, 0x1108, 0xFD08, 0x508, 0x908, 0x11F8, 0x3108, 0x5908, 0x9508, 0x15F8, 0x1108, 0x1108, 0x1108, 0x17FE, 0x0}},
    {64081, 2, {0x2104, 0x11FE, 0x1104, 0xFD04, 0x504, 0x904, 0x11FC, 0x3154, 0x5850, 0x9450, 0x1450, 0x1052, 0x1092, 0x1112, 0x120E, 0x0}},
    {64082, 2, {0x2088, 0x10FC, 0x1088, 0xFC88, 0x4B8, 0x8A8, 0x10A8, 0x33FE, 0x5A02, 0x96F2, 0x1292, 0x12F2, 0x1292, 0x120E, 0x1204, 0x0}},
    {64083, 2, {0x2020, 0x1024, 0x103E, 0xFC20, 0x5FC, 0x904, 0x11FC, 0x3104, 0x59FC, 0x9504, 0x15FC, 0x1104, 0x1050, 0x1088, 0x1306, 0x0}},
    {64084, 2, {0x848, 0xFFFC, 0x848, 0x7F48, 0x48, 0x7F8E, 0x8100, 0x3EFC, 0x48, 0x7E48, 0x848, 0x7F30, 0x2A30, 0x4948, 0x8986, 0x0}},
    {64085, 2, {0x100, 0x2084, 0x3FFE, 0x2004, 0x4448, 0x1830, 0x614C, 0x120, 0x104, 0x7FFE, 0x100, 0x280, 0x440, 0x1830, 0x600E, 0x0}},
    {64086, 2, {0x1040, 0x1040, 0x1F7E, 0x2890, 0x4508, 0x3E7C, 0x2244, 0x3E44, 0x2244, 0x3E44, 0x2244, 0x2444, 0x2E5C, 0x7348, 0x2140, 0x0}},
    {64087, 2, {0x1024, 0x13FE, 0x2020, 0x27FE, 0x4622, 0xFB26, 0x12AA, 0x2BFE, 0xFA22, 0x4420, 0x70, 0x28A8, 0x5526, 0x5622, 0x8020, 0x0}},
    {64088, 2, {0x1004, 0x13FE, 0x2088, 0x2488, 0x4552, 0xFBDE, 0x1000, 0x2BFE, 0xFC00, 0x45FC, 0x104, 0x29FC, 0x5504, 0x55FC, 0x8104, 0x0}},
    {64089, 2, {0x2020, 0x3FBE, 0x5544, 0x3FA4, 0x1518, 0x1FA6, 0x142, 0x210, 0x1FE0, 0x210, 0x7FFC, 0x2084, 0xA0, 0x1B98, 0x6106, 0x0}},
    {64090, 2, {0x3FFC, 0x2244, 0x2244, 0x3FFC, 0x2104, 0x1FE8, 0x110, 0x7FFE, 0x80, 0x7F8, 0xC08, 0x37F8, 0x408, 0x7F8, 0x408, 0x0}},
    {64091, 2, {0x200, 0x208, 0x3FC8, 0x210, 0x224, 0xFFFE, 0x100, 0x210, 0xFF8, 0x1810, 0x2810, 0xCFF0, 0x810, 0xFF0, 0x810, 0x0}},
    {64092, 2, {0x910, 0xFF8, 0x810, 0xFF0, 0x810, 0xFF0, 0x810, 0xFF0, 0x820, 0x94, 0x7FFE, 0x140, 0x220, 0xC1E, 0x7004, 0x0}},
    {64093, 2, {0x0, 0x0, 0x0, 0x420, 0x420, 0x7FFE, 0x420, 0x420, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64094, 2, {0x0, 0x0, 0x0, 0x420, 0x420, 0x7FFE, 0x420, 0x420, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64095, 2, {0x420, 0x420, 0x7FFE, 0x520, 0x108, 0x1FD0, 0x120, 0x7FFE, 0x100, 0x7F8, 0x1C08, 0x67F8, 0x408, 0x7F8, 0x408, 0x0}},
    {64096, 2, {0x41FC, 0x2104, 0x29FC, 0xFD04, 0x9FC, 0x1124, 0x2440, 0x28FE, 0x7122, 0xAF22, 0x2552, 0x218A, 0x2102, 0x21F2, 0x200C, 0x0}},
    {64097, 2, {0x21FC, 0x1104, 0x1104, 0x5FC, 0x7F04, 0x904, 0x11FC, 0x3904, 0x5504, 0x95FC, 0x1154, 0x1050, 0x1052, 0x1092, 0x170E, 0x0}},
    {64098, 2, {0x2104, 0x11FE, 0x1104, 0xFDFC, 0x104, 0x1FC, 0x7940, 0xFE, 0x7922, 0x322, 0x7952, 0x498A, 0x4902, 0x79F2, 0x480C, 0x0}},
    {64099, 2, {0x2088, 0x1088, 0x13FE, 0xFC88, 0xF8, 0x20, 0x79FC, 0x124, 0x79FC, 0x20, 0x7BFE, 0x4820, 0x49FC, 0x7820, 0x4BFE, 0x0}},
    {64100, 2, {0x2104, 0x3FFE, 0x2004, 0x5FF4, 0x510, 0x920, 0x3140, 0xFF0, 0x7810, 0xFF0, 0x810, 0xFF0, 0x810, 0xFF0, 0x381C, 0x0}},
    {64101, 2, {0x890C, 0xFC88, 0x8850, 0x8BFE, 0xFA22, 0x8B26, 0x8AAA, 0xFBFE, 0x8800, 0x89FC, 0xF904, 0x89FC, 0x5104, 0x89FC, 0x8904, 0x0}},
    {64102, 2, {0x4000, 0x3000, 0x1000, 0x0, 0x0, 0x1000, 0xF800, 0x1000, 0x1000, 0x1000, 0x1000, 0x3000, 0xC800, 0x87FE, 0x0, 0x0}},
    {64103, 2, {0x110, 0x41F8, 0x3110, 0x1224, 0x7FE, 0x244, 0x1244, 0xFBFC, 0x1060, 0x10A8, 0x1126, 0x1622, 0x301E, 0xC800, 0x87FE, 0x0}},
    {64104, 2, {0x22A0, 0xFF90, 0x2290, 0x3EFE, 0x990, 0x7E90, 0x4A90, 0x7EFC, 0x890, 0x7E90, 0x8FC, 0xFF90, 0x1494, 0x22FE, 0x4280, 0x0}},
    {64105, 2, {0x111E, 0x27D2, 0x7A54, 0x13D8, 0x2A56, 0x7A92, 0x135C, 0x6090, 0x1FF8, 0x444, 0x7FFE, 0x810, 0xFF0, 0x810, 0xFF0, 0x0}},
    {64106, 2, {0x804, 0x9FE, 0x2820, 0x2EFC, 0x2884, 0x2884, 0xFFFC, 0x884, 0x28FC, 0x2884, 0x4984, 0x8AFC, 0x484, 0x1858, 0x6186, 0x0}},
    {64256, 1, {0x0, 0x0, 0x0, 0x1B, 0x24, 0x24, 0x24, 0x7E, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {64257, 1, {0x0, 0x0, 0x0, 0x1A, 0x22, 0x20, 0x20, 0x7E, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {64258, 1, {0x0, 0x0, 0x0, 0x1E, 0x22, 0x22, 0x22, 0x7A, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {64259, 1, {0x0, 0x0, 0x0, 0x16, 0x2A, 0x28, 0x28, 0x7E, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x0, 0x0}},
    {64260, 1, {0x0, 0x0, 0x0, 0x1E, 0x2A, 0x2A, 0x2A, 0x7E, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x0, 0x0}},
    {64261, 1, {0x0, 0x0, 0x0, 0x38, 0x48, 0x48, 0x48, 0x5E, 0x48, 0x48, 0x48, 0x48, 0x48, 0x46, 0x0, 0x0}},
    {64262, 1, {0x0, 0x0, 0x0, 0x1C, 0x24, 0x24, 0x24, 0x36, 0x44, 0x44, 0x24, 0x14, 0x14, 0x62, 0x0, 0x0}},
    {64275, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x30, 0x5D, 0x55, 0x55, 0x55, 0x55, 0x55, 0x33, 0x0, 0x0}},
    {64276, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x14, 0x57, 0x54, 0x54, 0x55, 0x55, 0x55, 0x36, 0x0, 0x0}},
    {64277, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x14, 0x56, 0x55, 0x55, 0x55, 0x55, 0x55, 0x35, 0x4, 0x4}},
    {64278, 1, {0x0, 0x0, 0x0, 0x0, 0x50, 0x3C, 0x14, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x33, 0x10, 0x1C}},
    {64279, 1, {0x0, 0x0, 0x0, 0x0, 0x18, 0x28, 0x28, 0xAD, 0xAD, 0xAD, 0xAD, 0xAD, 0xAD, 0x6A, 0x8, 0x8}},
    {64285, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x4, 0x4, 0x4, 0x0, 0x4, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64286, 1, {0x0, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x24, 0x0, 0x42, 0x0, 0x24, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64287, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x77, 0x11, 0x11, 0x11, 0x0, 0x3E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64288, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x44, 0x22, 0x12, 0x12, 0x12, 0x12, 0x12, 0x14, 0xF8, 0x0, 0x0}},
    {64297, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x7F, 0x0, 0x0, 0x0, 0x0}},
    {64298, 1, {0x0, 0x0, 0x0, 0x1, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {64299, 1, {0x0, 0x0, 0x0, 0x40, 0x0, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {64300, 1, {0x0, 0x0, 0x0, 0x1, 0x0, 0x49, 0x49, 0x49, 0x49, 0x4D, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {64301, 1, {0x0, 0x0, 0x0, 0x40, 0x0, 0x49, 0x49, 0x49, 0x49, 0x4D, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {64302, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x42, 0x22, 0x12, 0x2C, 0x48, 0x44, 0x42, 0x41, 0x1C, 0x0}},
    {64303, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x42, 0x22, 0x12, 0x2C, 0x48, 0x44, 0x42, 0x41, 0x1C, 0x8}},
    {64304, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x42, 0x22, 0x12, 0x2C, 0x48, 0x44, 0x42, 0x41, 0x8, 0x0}},
    {64305, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x24, 0x4, 0x4, 0x4, 0x7F, 0x0, 0x0}},
    {64306, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x8, 0x8, 0x8, 0x44, 0xC, 0x14, 0x22, 0x42, 0x0, 0x0}},
    {64307, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x4, 0x4, 0x4, 0x24, 0x4, 0x4, 0x4, 0x4, 0x0, 0x0}},
    {64308, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x22, 0x2A, 0x22, 0x22, 0x22, 0x22, 0x0, 0x0}},
    {64309, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x8, 0x8, 0x8, 0x28, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {64310, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0xC, 0xA, 0x8, 0x28, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {64312, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x46, 0x4A, 0x42, 0x42, 0x52, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0}},
    {64313, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x8, 0x8, 0x8, 0x20, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64314, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x12, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2}},
    {64315, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x12, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x0}},
    {64316, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x3E, 0x2, 0x2, 0x2, 0x12, 0x4, 0x4, 0x8, 0x30, 0x0, 0x0}},
    {64318, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x48, 0x48, 0x34, 0x14, 0x22, 0x22, 0x49, 0x41, 0x4F, 0x0, 0x0}},
    {64320, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x4, 0x4, 0x4, 0x24, 0x4, 0x4, 0x4, 0x3C, 0x0, 0x0}},
    {64321, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x52, 0x42, 0x42, 0x44, 0x78, 0x0, 0x0}},
    {64323, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x72, 0x2, 0x2, 0x12, 0x2, 0x2, 0x2}},
    {64324, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x72, 0x2, 0x12, 0x2, 0x7E, 0x0, 0x0}},
    {64326, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x24, 0x18, 0x8, 0x44, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {64327, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x22, 0x2A, 0x22, 0x24, 0x28, 0x20, 0x20, 0x20, 0x20, 0x20}},
    {64328, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x12, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {64329, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x49, 0x49, 0x49, 0x49, 0x4D, 0x49, 0x49, 0x49, 0x7F, 0x0, 0x0}},
    {64330, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3E, 0x22, 0x22, 0x22, 0x2A, 0x22, 0x22, 0x22, 0x62, 0x0, 0x0}},
    {64331, 1, {0x0, 0x0, 0x0, 0x20, 0x0, 0x38, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {64332, 1, {0x0, 0x0, 0x0, 0x70, 0x0, 0x7C, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x7F, 0x0, 0x0}},
    {64333, 1, {0x0, 0x0, 0x0, 0x70, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7C, 0x0, 0x0}},
    {64334, 1, {0x0, 0x0, 0x0, 0x70, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x72, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {64335, 1, {0x0, 0x0, 0x60, 0x20, 0x20, 0x46, 0x42, 0x22, 0x12, 0x2C, 0x48, 0x44, 0x42, 0x42, 0x0, 0x0}},
    {64336, 1, {0x4, 0xA, 0x1E, 0x20, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64337, 1, {0x10, 0x28, 0x78, 0x80, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64338, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {64339, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {64340, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {64341, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {64342, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x14, 0x0, 0x8, 0x0}},
    {64343, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x14, 0x0, 0x8, 0x0}},
    {64344, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x14, 0x0, 0x8, 0x0}},
    {64345, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x14, 0x0, 0x8, 0x0}},
    {64346, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x14, 0x0, 0x14, 0x0}},
    {64347, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x14, 0x0, 0x14, 0x0}},
    {64348, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x14, 0x0, 0x14, 0x0}},
    {64349, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x14, 0x0, 0x14, 0x0}},
    {64350, 1, {0x0, 0x0, 0x8, 0x0, 0x8, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64351, 1, {0x0, 0x0, 0x8, 0x0, 0x8, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64352, 1, {0x0, 0x0, 0x8, 0x0, 0x8, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64353, 1, {0x0, 0x0, 0x8, 0x0, 0x8, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64354, 1, {0x0, 0x0, 0x14, 0x0, 0x14, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64355, 1, {0x0, 0x0, 0x14, 0x0, 0x14, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64356, 1, {0x0, 0x0, 0x14, 0x0, 0x14, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64357, 1, {0x0, 0x0, 0x14, 0x0, 0x14, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64358, 1, {0x0, 0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64359, 1, {0x0, 0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64360, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64361, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64362, 1, {0x0, 0x8, 0x0, 0x12, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64363, 1, {0x0, 0x8, 0x0, 0x12, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64364, 1, {0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64365, 1, {0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64366, 1, {0x0, 0xA, 0x0, 0xA, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64367, 1, {0x0, 0xA, 0x0, 0xA, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64368, 1, {0x0, 0x28, 0x0, 0x28, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64369, 1, {0x0, 0x28, 0x0, 0x28, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64370, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x48, 0x40, 0x48, 0x22, 0x1C}},
    {64371, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x14, 0x23, 0x48, 0x40, 0x48, 0x21, 0x1E}},
    {64372, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFC, 0x0, 0x20, 0x0, 0x20, 0x0}},
    {64373, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFF, 0x0, 0x20, 0x0, 0x20, 0x0}},
    {64374, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x54, 0x40, 0x22, 0x1C}},
    {64375, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x14, 0x23, 0x40, 0x54, 0x40, 0x21, 0x1E}},
    {64376, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFC, 0x0, 0x50, 0x0, 0x0, 0x0}},
    {64377, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFF, 0x0, 0x50, 0x0, 0x0, 0x0}},
    {64378, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x40, 0x54, 0x40, 0x48, 0x22, 0x1C}},
    {64379, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x22, 0x41, 0x54, 0x40, 0x48, 0x21, 0x1E}},
    {64380, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFC, 0x0, 0x28, 0x0, 0x10, 0x0}},
    {64381, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFF, 0x0, 0x28, 0x0, 0x10, 0x0}},
    {64382, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x20, 0x40, 0x54, 0x40, 0x54, 0x22, 0x1C}},
    {64383, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x43, 0x54, 0x40, 0x54, 0x21, 0x1E}},
    {64384, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFC, 0x0, 0x50, 0x0, 0x50, 0x0}},
    {64385, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFF, 0x0, 0x50, 0x0, 0x50, 0x0}},
    {64386, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {64387, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7F, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {64388, 1, {0x0, 0x0, 0x0, 0x28, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64389, 1, {0x0, 0x0, 0x0, 0x28, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64390, 1, {0x0, 0x0, 0x10, 0x0, 0x28, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64391, 1, {0x0, 0x0, 0x10, 0x0, 0x28, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64392, 1, {0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64393, 1, {0x0, 0x20, 0x20, 0x38, 0x38, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64394, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x14, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64395, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x14, 0x0, 0x0, 0x4, 0x2, 0x3, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64396, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0xE, 0xE, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64397, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0xE, 0xE, 0x0, 0x4, 0x2, 0x3, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64398, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64399, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64400, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64401, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64402, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64403, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64404, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64405, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64406, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x10, 0x0, 0x10, 0x0}},
    {64407, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x82, 0x7F, 0x0, 0x10, 0x0, 0x10, 0x0}},
    {64408, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFE, 0x0, 0x10, 0x0, 0x10, 0x0}},
    {64409, 1, {0x0, 0x8, 0x10, 0x24, 0x48, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFF, 0x0, 0x10, 0x0, 0x10, 0x0}},
    {64410, 1, {0x0, 0x50, 0x6, 0x18, 0x62, 0xC, 0x30, 0x40, 0x3C, 0x82, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64411, 1, {0x0, 0x50, 0x6, 0x18, 0x62, 0xC, 0x30, 0x40, 0x3C, 0x82, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64412, 1, {0x0, 0x50, 0x6, 0x18, 0x62, 0xC, 0x30, 0x40, 0x3C, 0x2, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64413, 1, {0x0, 0x50, 0x6, 0x18, 0x62, 0xC, 0x30, 0x40, 0x3C, 0x2, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64414, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {64415, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x43, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {64416, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {64417, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x0, 0x4, 0x22, 0x43, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {64418, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64419, 1, {0x0, 0x0, 0x10, 0x10, 0x1C, 0x1C, 0x0, 0x8, 0x4, 0x4, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64420, 1, {0x0, 0x18, 0x20, 0x18, 0x20, 0x0, 0x10, 0x18, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64421, 1, {0x18, 0x20, 0x18, 0x20, 0x4, 0x4, 0xC, 0x14, 0x24, 0x3C, 0x7, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64422, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x18, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64423, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2F, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {64424, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0xF0, 0x10, 0x10, 0x14, 0x8, 0x0}},
    {64425, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0xF3, 0x14, 0x14, 0x14, 0x8, 0x8}},
    {64426, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2C, 0x2A, 0x1A, 0x24, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {64427, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2C, 0x2A, 0x1A, 0x27, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {64428, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2C, 0x2A, 0x1A, 0xE4, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64429, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2C, 0x2A, 0x1A, 0xE7, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64430, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x10, 0x20, 0x40, 0x7F, 0x0}},
    {64431, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0x10, 0x20, 0x40, 0x7F, 0x0}},
    {64432, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x40, 0x8, 0x14, 0x10, 0x20, 0x40, 0x7F, 0x0}},
    {64433, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x40, 0x0, 0xF, 0x10, 0x20, 0x40, 0x7F, 0x0}},
    {64467, 1, {0x10, 0x0, 0x28, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64468, 1, {0x10, 0x0, 0x28, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64469, 1, {0x20, 0x0, 0x50, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64470, 1, {0x20, 0x0, 0x50, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64471, 1, {0x0, 0x10, 0x28, 0x28, 0x1C, 0x10, 0x20, 0x4C, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64472, 1, {0x0, 0x10, 0x28, 0x28, 0x1C, 0x10, 0x20, 0x4C, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64473, 1, {0x0, 0x0, 0x22, 0x14, 0x8, 0x8, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64474, 1, {0x0, 0x0, 0x22, 0x14, 0x8, 0x8, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64475, 1, {0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64476, 1, {0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64477, 1, {0x0, 0x23, 0x54, 0x53, 0x34, 0x20, 0x40, 0x8C, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64478, 1, {0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64479, 1, {0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64480, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x1C, 0x48, 0x30, 0x0}},
    {64481, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x1C, 0x48, 0x30, 0x0}},
    {64482, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64483, 1, {0x0, 0x0, 0x0, 0x8, 0x14, 0x22, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {64484, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x52, 0x48, 0x44, 0x38, 0x0, 0x10, 0x0, 0x10}},
    {64485, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4C, 0x52, 0x49, 0x30, 0x0, 0x10, 0x0, 0x10}},
    {64486, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xF8, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {64487, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xFB, 0x0, 0x8, 0x0, 0x8, 0x0}},
    {64488, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64489, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64490, 1, {0x0, 0x0, 0x0, 0x4C, 0x50, 0x4C, 0x50, 0x40, 0x44, 0x44, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64491, 1, {0x0, 0x0, 0x0, 0x4C, 0x50, 0x4C, 0x50, 0x40, 0x44, 0x44, 0x3B, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64492, 1, {0x0, 0x0, 0x0, 0x6, 0x8, 0x6, 0x8, 0x0, 0x2, 0x12, 0x2C, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {64493, 1, {0x0, 0x0, 0x0, 0x6, 0x8, 0x6, 0x8, 0x0, 0x2, 0x12, 0x2D, 0x40, 0x0, 0x0, 0x0, 0x0}},
    {64494, 1, {0x0, 0x0, 0x0, 0x3, 0x4, 0x3, 0x4, 0x10, 0x2A, 0x2A, 0x3C, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64495, 1, {0x0, 0x0, 0x0, 0x3, 0x4, 0x3, 0x4, 0x10, 0x2A, 0x2A, 0x3D, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64496, 1, {0x0, 0x0, 0x30, 0x33, 0x14, 0x23, 0x44, 0x10, 0x2A, 0x2A, 0x3C, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64497, 1, {0x0, 0x0, 0x30, 0x33, 0x14, 0x23, 0x44, 0x10, 0x2A, 0x2A, 0x3D, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64498, 1, {0x0, 0x0, 0x88, 0x53, 0x24, 0x23, 0x4, 0x10, 0x2A, 0x2A, 0x3C, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64499, 1, {0x0, 0x0, 0x88, 0x53, 0x24, 0x23, 0x4, 0x10, 0x2A, 0x2A, 0x3D, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64500, 1, {0x0, 0x0, 0x0, 0x23, 0x14, 0x13, 0x4, 0x10, 0x2A, 0x2A, 0x3C, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64501, 1, {0x0, 0x0, 0x0, 0x23, 0x14, 0x13, 0x4, 0x10, 0x2A, 0x2A, 0x3D, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64502, 1, {0x0, 0xC, 0x10, 0xC, 0x10, 0x4, 0x4, 0xC, 0x50, 0x48, 0x44, 0x38, 0x0, 0x10, 0x0, 0x10}},
    {64503, 1, {0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0x4, 0x4C, 0x52, 0x49, 0x30, 0x0, 0x10, 0x0, 0x10}},
    {64504, 1, {0x0, 0x0, 0x6, 0x8, 0x6, 0x8, 0x0, 0x22, 0x22, 0x12, 0xEC, 0x0, 0x20, 0x0, 0x20, 0x0}},
    {64505, 1, {0x0, 0x0, 0x6, 0x8, 0x6, 0x8, 0x2, 0x2, 0x2, 0x6, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {64506, 1, {0x0, 0x0, 0x6, 0x8, 0x6, 0x8, 0x0, 0x2, 0x2, 0x5, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {64507, 1, {0x0, 0x0, 0x6, 0x8, 0x6, 0x8, 0x0, 0x22, 0x22, 0x12, 0xEC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64508, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {64509, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x43, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {64510, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xF8, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {64511, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xFB, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {64512, 1, {0x0, 0x0, 0x0, 0x60, 0x80, 0x60, 0x8C, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64513, 1, {0x0, 0x0, 0x0, 0x60, 0x80, 0x60, 0x8C, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64514, 1, {0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x23, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64515, 1, {0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x44, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x0}},
    {64516, 1, {0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x44, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64517, 1, {0x0, 0x0, 0x0, 0x60, 0x80, 0x60, 0x8C, 0x30, 0x40, 0x7E, 0x20, 0x42, 0x48, 0x40, 0x22, 0x1C}},
    {64518, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x42, 0x40, 0x40, 0x22, 0x1C}},
    {64519, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x42, 0x40, 0x40, 0x22, 0x1C}},
    {64520, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x41, 0x40, 0x40, 0x40, 0x40}},
    {64521, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x1, 0x0}},
    {64522, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x1, 0x14}},
    {64523, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x42, 0x40, 0x40, 0x22, 0x1C}},
    {64524, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64525, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0x8C, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64526, 1, {0x0, 0x0, 0x0, 0x5, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64527, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64528, 1, {0x0, 0x0, 0x0, 0x14, 0x0, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x42, 0x22, 0x1C, 0x0, 0x14}},
    {64529, 1, {0x0, 0x0, 0x0, 0x40, 0x0, 0xA0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64530, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x28, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64531, 1, {0x0, 0x0, 0x0, 0x20, 0x0, 0x50, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x0}},
    {64532, 1, {0x0, 0x0, 0x0, 0x20, 0x0, 0x50, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64533, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0xC, 0x30, 0x40, 0x78, 0x22, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64534, 1, {0x0, 0x0, 0x0, 0x30, 0xC, 0x2, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x41, 0x40, 0x40, 0x40, 0x40}},
    {64535, 1, {0x0, 0x0, 0x0, 0x70, 0xC, 0x2, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x42, 0x40, 0x40, 0x22, 0x1C}},
    {64536, 1, {0x0, 0x0, 0x0, 0x30, 0xC, 0x2, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64537, 1, {0x0, 0x0, 0x4, 0x70, 0xC, 0x2, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64538, 1, {0x0, 0x0, 0x4, 0x70, 0xC, 0x2, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64539, 1, {0x0, 0x0, 0x0, 0x31, 0xC, 0x2, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64540, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x25, 0x5A, 0x40, 0x7E, 0x20, 0x42, 0x40, 0x40, 0x22, 0x1C}},
    {64541, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x25, 0x5A, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64542, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x25, 0x5A, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64544, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0x9, 0x1E, 0x28, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64546, 1, {0x0, 0x0, 0x0, 0x0, 0x26, 0x9, 0x1E, 0x28, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64547, 1, {0x0, 0x0, 0x0, 0x0, 0x26, 0x9, 0x1E, 0x28, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64548, 1, {0x0, 0x0, 0x0, 0x0, 0x26, 0x89, 0x1E, 0x28, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64550, 1, {0x0, 0x0, 0x0, 0x20, 0x2C, 0x32, 0x3C, 0x20, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64551, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0xE, 0x9, 0x16, 0x10, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64552, 1, {0x0, 0x0, 0x0, 0x8, 0xA, 0x8, 0xE, 0x9, 0x16, 0x10, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64553, 1, {0x0, 0x0, 0x0, 0x30, 0x40, 0x38, 0x10, 0x20, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64554, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x10, 0xF, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64555, 1, {0x0, 0x0, 0x80, 0x30, 0x40, 0x38, 0x10, 0x20, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64556, 1, {0x0, 0x0, 0x8, 0x0, 0xC, 0x10, 0xF, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64557, 1, {0x0, 0x0, 0x80, 0x20, 0x50, 0x30, 0x10, 0x20, 0x7E, 0x10, 0x20, 0x40, 0x48, 0x40, 0x21, 0x1E}},
    {64558, 1, {0x0, 0x0, 0x80, 0x20, 0x50, 0x30, 0x10, 0x20, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x21, 0x1E}},
    {64559, 1, {0x0, 0x10, 0x80, 0x20, 0x50, 0x30, 0x10, 0x20, 0x7E, 0x10, 0x20, 0x40, 0x48, 0x40, 0x21, 0x1E}},
    {64560, 1, {0x0, 0x0, 0x0, 0x40, 0x10, 0x28, 0x18, 0x8, 0x10, 0x20, 0x10, 0x38, 0x40, 0x40, 0x40, 0x40}},
    {64561, 1, {0x0, 0x0, 0x4, 0x0, 0x0, 0xC, 0xA, 0x6, 0x2, 0x46, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64562, 1, {0x0, 0x0, 0x4, 0x0, 0x0, 0xC, 0xA, 0x6, 0x2, 0x46, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64563, 1, {0x0, 0x0, 0x0, 0x50, 0x0, 0x20, 0x50, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64564, 1, {0x0, 0x0, 0x28, 0x0, 0x10, 0x28, 0x18, 0x8, 0x10, 0x20, 0x10, 0x38, 0x40, 0x40, 0x40, 0x40}},
    {64565, 1, {0x0, 0x0, 0x14, 0x0, 0x0, 0xC, 0xA, 0x6, 0x2, 0x46, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64566, 1, {0x0, 0x0, 0x14, 0x0, 0x0, 0xC, 0xA, 0x6, 0x2, 0x46, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64567, 1, {0x0, 0x0, 0x0, 0x4, 0x88, 0x90, 0xA0, 0xC0, 0xBC, 0x82, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64568, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x20, 0x10, 0x20, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64569, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x20, 0x10, 0x20, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64570, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x10, 0x20, 0x90, 0x20, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64571, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x12, 0x14, 0x18, 0x14, 0x12, 0x1C, 0x90, 0x90, 0x90, 0x90, 0x60}},
    {64572, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x10, 0x8, 0x10, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64575, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x20, 0x7E, 0x10, 0x20, 0x40, 0x48, 0x40, 0x21, 0x1E}},
    {64576, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x20, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x21, 0x1E}},
    {64577, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x50, 0x10, 0x20, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x21, 0x1E}},
    {64578, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64579, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x6, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {64580, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x6, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64581, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64582, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64583, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64584, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64585, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x8, 0x44, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {64586, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x8, 0x44, 0x42, 0x42, 0x3C, 0x0, 0x28}},
    {64587, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64588, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64589, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x20, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64590, 1, {0x0, 0x0, 0x0, 0x2, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64591, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64592, 1, {0x0, 0x0, 0x0, 0x4, 0x0, 0x4, 0x2, 0x2, 0x22, 0x44, 0x46, 0x42, 0x22, 0x1C, 0x0, 0x14}},
    {64593, 1, {0x0, 0x0, 0x0, 0x10, 0x38, 0x54, 0x52, 0x3C, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64594, 1, {0x0, 0x0, 0x0, 0x20, 0x18, 0x34, 0x54, 0x38, 0x8, 0x10, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64595, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0xD, 0x15, 0xE, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x0}},
    {64596, 1, {0x0, 0x0, 0x0, 0x0, 0x6, 0xD, 0x15, 0xE, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64597, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x4A, 0x40, 0x48, 0x22, 0x1C}},
    {64598, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x4A, 0x40, 0x40, 0x22, 0x1C}},
    {64599, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0xC, 0x30, 0x40, 0x7E, 0x20, 0x4A, 0x40, 0x40, 0x22, 0x1C}},
    {64600, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x41, 0x44, 0x40, 0x40, 0x40}},
    {64601, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x22, 0x4, 0x46, 0x62, 0x1C, 0x1, 0x4, 0x0}},
    {64602, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x1, 0x4, 0x50}},
    {64603, 1, {0x10, 0x10, 0x10, 0x0, 0x20, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64604, 1, {0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64605, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64606, 1, {0x0, 0x4, 0x6A, 0x24, 0x18, 0x0, 0x54, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64607, 1, {0x0, 0x54, 0x54, 0x7C, 0x6, 0x18, 0x66, 0x18, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64608, 1, {0x0, 0x0, 0x6, 0x18, 0x60, 0x0, 0x54, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64609, 1, {0x0, 0x4, 0xA, 0x4, 0x7A, 0x0, 0x54, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64610, 1, {0x0, 0x54, 0x54, 0x7C, 0x0, 0x6, 0x18, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64611, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x0, 0x54, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64612, 1, {0x0, 0x0, 0x18, 0x20, 0x18, 0x22, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64613, 1, {0x0, 0x0, 0x18, 0x20, 0x98, 0x22, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64614, 1, {0x0, 0x0, 0x0, 0x18, 0x20, 0x18, 0x23, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64615, 1, {0x0, 0x0, 0x0, 0x18, 0x20, 0x98, 0x2, 0x6, 0xA, 0x2A, 0x45, 0x44, 0x48, 0x30, 0x0, 0x0}},
    {64616, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64617, 1, {0x0, 0x0, 0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64618, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x12, 0x60, 0x0}},
    {64619, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x12, 0x60, 0x0}},
    {64620, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3D, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64621, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2, 0x6, 0xA, 0x2A, 0x45, 0x44, 0x48, 0x32, 0x0, 0x0}},
    {64622, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x1, 0x0}},
    {64623, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x1, 0x28}},
    {64624, 1, {0x0, 0x0, 0x5, 0x0, 0x0, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64625, 1, {0x0, 0x0, 0x5, 0x0, 0x20, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64626, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64627, 1, {0x0, 0x0, 0x0, 0x5, 0x0, 0x10, 0x2, 0x6, 0xA, 0x2A, 0x45, 0x44, 0x48, 0x30, 0x0, 0x0}},
    {64628, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64629, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x5, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64630, 1, {0x0, 0x2, 0x0, 0x5, 0x0, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64631, 1, {0x0, 0x2, 0x0, 0x5, 0x0, 0x22, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64632, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x28, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64633, 1, {0x0, 0x0, 0x2, 0x0, 0x5, 0x0, 0x12, 0x6, 0xA, 0x2A, 0x45, 0x44, 0x48, 0x30, 0x0, 0x0}},
    {64634, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x0, 0x5, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64635, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x0, 0x5, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64636, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x4, 0xA, 0x7, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x0}},
    {64637, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x4, 0xA, 0x7, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64638, 1, {0x0, 0x0, 0x0, 0x0, 0xA, 0x0, 0x4, 0xA, 0x7, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x0}},
    {64639, 1, {0x0, 0x0, 0x0, 0x0, 0xA, 0x0, 0x4, 0xA, 0x7, 0x22, 0x44, 0x46, 0x22, 0x1C, 0x0, 0x14}},
    {64640, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x21, 0x22, 0x24, 0x28, 0x30, 0x28, 0x2C, 0x12, 0x1}},
    {64641, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x21, 0x22, 0x24, 0x28, 0x30, 0xA8, 0xAC, 0xB2, 0x41}},
    {64642, 1, {0x0, 0x1, 0x2, 0x4, 0x8, 0x10, 0x8, 0x4, 0xA, 0x11, 0x8, 0x1C, 0x20, 0x20, 0x20, 0x20}},
    {64643, 1, {0x0, 0x0, 0x1, 0x2, 0x4, 0x8, 0x10, 0x8, 0x4, 0x46, 0x89, 0x8C, 0x44, 0x38, 0x0, 0x0}},
    {64644, 1, {0x0, 0x0, 0x1, 0x2, 0x4, 0x8, 0x10, 0x8, 0x4, 0x46, 0x89, 0x8C, 0x44, 0x38, 0x0, 0x28}},
    {64645, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x3C, 0x20, 0x30, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64646, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x8, 0x44, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {64647, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0xE, 0x8, 0x44, 0x42, 0x42, 0x3C, 0x0, 0x28}},
    {64648, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x20, 0x26, 0x18, 0xE}},
    {64649, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x4, 0x42, 0xA1, 0xBF, 0xA0, 0x80, 0x80, 0x80, 0x0}},
    {64650, 1, {0x0, 0x0, 0x0, 0x2, 0x0, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64651, 1, {0x0, 0x0, 0x0, 0x2, 0x10, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x60, 0x0}},
    {64652, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64653, 1, {0x0, 0x0, 0x0, 0x2, 0x10, 0x0, 0x2, 0x6, 0xA, 0x2A, 0x45, 0x44, 0x48, 0x30, 0x0, 0x0}},
    {64654, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64655, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64656, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64657, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x65, 0x0}},
    {64658, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x2, 0x6, 0xA, 0x12, 0xA, 0x9, 0x8, 0x8, 0x10, 0x65, 0x0}},
    {64659, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3, 0x4, 0x8, 0x8, 0x3C, 0x40, 0x4A, 0x40, 0x40, 0x40}},
    {64660, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x2, 0x6, 0xA, 0x2A, 0x45, 0x44, 0x48, 0x30, 0x5, 0x0}},
    {64661, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x1, 0x4}},
    {64662, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x1, 0x54}},
    {64663, 1, {0x2, 0x4, 0x6, 0x8, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x10, 0x0, 0x0, 0x0}},
    {64664, 1, {0x2, 0x4, 0x6, 0x8, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64665, 1, {0x2, 0x4, 0x6, 0x8, 0x0, 0x4C, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64666, 1, {0x2, 0x4, 0x6, 0x8, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64667, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0xC, 0x10, 0x0, 0x26, 0x69, 0xE8, 0x28, 0x24, 0x14, 0xC, 0x0}},
    {64668, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x4, 0x20, 0x0, 0x0}},
    {64669, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x4, 0x0, 0x0, 0x0}},
    {64670, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x4, 0x0, 0x0, 0x0}},
    {64671, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x8, 0x10, 0x10, 0xF8, 0x0, 0x4, 0x0, 0x0, 0x0}},
    {64672, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x26, 0x69, 0xE8, 0x28, 0x25, 0x14, 0xC, 0x0}},
    {64673, 1, {0x0, 0x0, 0xA, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x20, 0x0, 0x0}},
    {64674, 1, {0x0, 0x0, 0xA, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64675, 1, {0x0, 0x0, 0xA, 0x0, 0x20, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64676, 1, {0x0, 0x0, 0x0, 0xA, 0x0, 0x0, 0x6, 0x8, 0x10, 0x10, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64677, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xA, 0x0, 0x0, 0x26, 0x69, 0xE8, 0x28, 0x24, 0x14, 0xC, 0x0}},
    {64678, 1, {0x0, 0x4, 0x0, 0xA, 0x0, 0x0, 0x6, 0x8, 0x10, 0x10, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64679, 1, {0x0, 0x60, 0x58, 0x4, 0x1E, 0x20, 0x40, 0x7E, 0x18, 0x62, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64680, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x2, 0x1F, 0x28, 0xD8, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {64681, 1, {0x0, 0x60, 0x58, 0x4, 0x1E, 0x20, 0x40, 0x7E, 0x18, 0x60, 0x80, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {64682, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x3, 0x1F, 0x28, 0xD8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64683, 1, {0x8, 0x60, 0x58, 0x4, 0x1E, 0x20, 0x40, 0x7E, 0x18, 0x60, 0x80, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {64684, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x1C, 0x3, 0x1F, 0x28, 0xD8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64688, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x15, 0x15, 0x3E, 0x40, 0xE0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64696, 1, {0x10, 0x10, 0x16, 0x19, 0x1E, 0x20, 0x78, 0x6, 0x18, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64697, 1, {0x0, 0x0, 0x0, 0x0, 0x12, 0x10, 0x16, 0x19, 0x3E, 0x40, 0xF0, 0x48, 0x30, 0x0, 0x0, 0x0}},
    {64698, 1, {0x0, 0x0, 0x30, 0x42, 0x4C, 0x30, 0x40, 0x7E, 0x18, 0x60, 0x80, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {64699, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x4C, 0x30, 0x40, 0xF0, 0x48, 0x30, 0x0, 0x0, 0x0}},
    {64700, 1, {0x20, 0x0, 0x30, 0x42, 0x4C, 0x30, 0x40, 0x7E, 0x18, 0x60, 0x80, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {64701, 1, {0x0, 0x0, 0x0, 0x20, 0x0, 0x30, 0x40, 0x4C, 0x30, 0x40, 0xF0, 0x48, 0x30, 0x0, 0x0, 0x0}},
    {64702, 1, {0x40, 0x10, 0x28, 0x38, 0x8, 0x10, 0x20, 0x3E, 0xC, 0x30, 0xC0, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {64703, 1, {0x40, 0x10, 0x28, 0x38, 0x8, 0x10, 0x20, 0x3E, 0xC, 0x30, 0xC0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64704, 1, {0x0, 0x40, 0x10, 0x28, 0x38, 0x88, 0x10, 0x20, 0x3E, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64705, 1, {0x0, 0x10, 0x0, 0x10, 0x28, 0x38, 0x8, 0x10, 0x20, 0x20, 0xF0, 0x28, 0x10, 0x0, 0x0, 0x0}},
    {64706, 1, {0x28, 0x0, 0x10, 0x28, 0x38, 0x8, 0x10, 0x20, 0x3E, 0xC, 0xF0, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {64707, 1, {0x0, 0x28, 0x0, 0x10, 0x28, 0x38, 0x8, 0x10, 0x20, 0x20, 0xF0, 0x28, 0x10, 0x0, 0x0, 0x0}},
    {64708, 1, {0x2, 0x4, 0x8, 0x10, 0x30, 0x8, 0x10, 0x7E, 0x18, 0x60, 0x80, 0x8, 0x0, 0x0, 0x0, 0x0}},
    {64709, 1, {0x2, 0x4, 0x8, 0x10, 0x30, 0x8, 0x10, 0x7E, 0x18, 0x60, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64710, 1, {0x2, 0x4, 0x8, 0x10, 0x30, 0x88, 0x10, 0x7E, 0x18, 0x60, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64711, 1, {0x0, 0x2, 0x44, 0x48, 0x50, 0x60, 0x78, 0x44, 0x44, 0x24, 0xD8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64712, 1, {0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x10, 0x20, 0xF0, 0x28, 0x10, 0x0, 0x0, 0x0}},
    {64713, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x20, 0x0, 0x0}},
    {64714, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64715, 1, {0x0, 0x0, 0x10, 0x10, 0x50, 0x10, 0x10, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64716, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0xFC, 0x2, 0x0, 0x0, 0x0, 0x0}},
    {64717, 1, {0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x1C, 0x2A, 0x3A, 0xF6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64718, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0x1C, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x20, 0x0, 0x0}},
    {64719, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0x1C, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64720, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0xA, 0x1C, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64721, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0x1E, 0x20, 0x20, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64722, 1, {0x0, 0x0, 0x4, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x20, 0x0, 0x0}},
    {64723, 1, {0x0, 0x0, 0x4, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64724, 1, {0x0, 0x0, 0x4, 0x0, 0x20, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64725, 1, {0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x6, 0x8, 0x10, 0x10, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64726, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x26, 0x69, 0xE8, 0x28, 0x24, 0x14, 0xC, 0x0}},
    {64727, 1, {0x0, 0x3C, 0x12, 0x29, 0x29, 0x3E, 0x40, 0x7E, 0x18, 0x60, 0x80, 0x10, 0x0, 0x0, 0x0, 0x0}},
    {64728, 1, {0x0, 0x3C, 0x12, 0x29, 0x29, 0x1E, 0x8, 0x10, 0x20, 0x40, 0xE0, 0x50, 0x20, 0x0, 0x0, 0x0}},
    {64729, 1, {0x20, 0x20, 0x0, 0x78, 0x24, 0x52, 0x52, 0x3C, 0x20, 0x40, 0x80, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64730, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0xA, 0x20, 0x0, 0x0}},
    {64731, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0xA, 0x0, 0x0, 0x0}},
    {64732, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0xC, 0x14, 0x20, 0x78, 0xC, 0xF0, 0x0, 0xA, 0x0, 0x0, 0x0}},
    {64733, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x8, 0x10, 0x10, 0xF8, 0x0, 0xA, 0x0, 0x0, 0x0}},
    {64734, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x50, 0xD5, 0x50, 0x48, 0x28, 0x18, 0x0}},
    {64735, 1, {0x0, 0x0, 0xC, 0x10, 0xC, 0x30, 0x0, 0x10, 0x28, 0x24, 0xD3, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64736, 1, {0x0, 0x0, 0x0, 0x6, 0x8, 0x6, 0x18, 0x0, 0x24, 0x6A, 0xE9, 0x28, 0x24, 0x14, 0xC, 0x0}},
    {64737, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x24, 0xD3, 0x0, 0x8, 0x0, 0x0, 0x0}},
    {64738, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x6A, 0xE9, 0x28, 0x25, 0x14, 0xC, 0x0}},
    {64739, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x10, 0x28, 0x24, 0xD3, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64740, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x0, 0x24, 0x6A, 0xE9, 0x28, 0x24, 0x14, 0xC, 0x0}},
    {64741, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x14, 0x0, 0x10, 0x28, 0x24, 0xD3, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64742, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x14, 0x0, 0x0, 0x24, 0x6A, 0xE9, 0x28, 0x24, 0x14, 0xC, 0x0}},
    {64747, 1, {0x0, 0x2, 0x44, 0x48, 0x50, 0x60, 0x70, 0x48, 0x4C, 0x4A, 0xB1, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64748, 1, {0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x20, 0x10, 0x18, 0x24, 0xF3, 0x28, 0x10, 0x0, 0x0, 0x0}},
    {64749, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x18, 0x24, 0xF3, 0x28, 0x10, 0x0, 0x0, 0x0}},
    {64750, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x10, 0x28, 0x24, 0xD3, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64751, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0xC, 0x12, 0x52, 0xD1, 0x50, 0x48, 0x28, 0x18, 0x0}},
    {64752, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x28, 0x24, 0xD3, 0x0, 0xA, 0x0, 0x0, 0x0}},
    {64753, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x56, 0xD1, 0x50, 0x48, 0x2B, 0x18, 0x0}},
    {64754, 1, {0x0, 0xC, 0x30, 0x2, 0x2A, 0x14, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64755, 1, {0x4, 0xA, 0x4, 0xA, 0x30, 0x2, 0x2A, 0x14, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64756, 1, {0x0, 0xA, 0x54, 0x20, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64757, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0xE, 0x9, 0xE, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64758, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0xE, 0x9, 0xE, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64759, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x8, 0x6, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64760, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x8, 0x6, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64761, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x6, 0x8, 0x6, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64762, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x6, 0x8, 0x6, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64767, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x4, 0xE, 0x10, 0x38, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64768, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x4, 0xE, 0x38, 0x8, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64769, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x4, 0xE, 0x38, 0x9, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64770, 1, {0x0, 0x0, 0x0, 0x0, 0x60, 0x58, 0x4, 0xE, 0x38, 0x9, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64771, 1, {0x0, 0x0, 0x0, 0x4, 0x60, 0x58, 0x4, 0xE, 0x38, 0x8, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64772, 1, {0x0, 0x0, 0x0, 0x2, 0x60, 0x58, 0x4, 0xE, 0x38, 0x8, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64773, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0x5, 0x1E, 0x14, 0x10, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64774, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0x5, 0x1E, 0x14, 0x10, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64775, 1, {0x0, 0x0, 0x0, 0x8, 0x2, 0x5, 0x5, 0x1E, 0x14, 0x10, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64776, 1, {0x0, 0x0, 0x0, 0x8, 0x2, 0x5, 0x5, 0x1E, 0x14, 0x10, 0x8, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64777, 1, {0x0, 0x0, 0x4, 0x0, 0xA, 0x0, 0x25, 0x5A, 0x40, 0x7E, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {64778, 1, {0x0, 0x0, 0x4, 0x0, 0xA, 0x0, 0x25, 0x5A, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64779, 1, {0x0, 0x0, 0x4, 0x0, 0xA, 0x80, 0x25, 0x5A, 0x40, 0x7E, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {64780, 1, {0x0, 0x0, 0x4, 0x0, 0xA, 0x0, 0x25, 0x5A, 0x40, 0x60, 0x3C, 0x40, 0x40, 0x40, 0x40, 0x40}},
    {64781, 1, {0x0, 0x0, 0x0, 0x4, 0x0, 0xA, 0x0, 0x15, 0x2A, 0x20, 0x20, 0x20, 0x20, 0x40, 0x80, 0x0}},
    {64782, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x15, 0x2A, 0x20, 0x20, 0x20, 0x20, 0x40, 0x80, 0x0}},
    {64783, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0xD, 0x16, 0x24, 0x24, 0x20, 0x20, 0x20, 0x40, 0x80, 0x0}},
    {64784, 1, {0x0, 0x0, 0x0, 0x8, 0x2, 0x5, 0xD, 0x16, 0x24, 0x24, 0x20, 0x20, 0x20, 0x40, 0x80, 0x0}},
    {64785, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0xE, 0x9, 0xE, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64786, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0xE, 0x9, 0xE, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64787, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64788, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64789, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x2, 0x5, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x0}},
    {64790, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x2, 0x5, 0x6, 0x49, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {64795, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0xC, 0x13, 0x88, 0x88, 0x70, 0x0, 0x0}},
    {64796, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0xC, 0x13, 0x88, 0x88, 0x70, 0x0, 0x50}},
    {64797, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0xC, 0x13, 0x88, 0x8A, 0x70, 0x0, 0x0}},
    {64798, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x18, 0x60, 0x18, 0xC, 0x13, 0x88, 0x8A, 0x70, 0x0, 0x50}},
    {64799, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x6, 0x18, 0x60, 0x18, 0xC, 0x13, 0x88, 0x88, 0x70, 0x0, 0x0}},
    {64800, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x6, 0x18, 0x60, 0x18, 0xC, 0x13, 0x88, 0x88, 0x70, 0x0, 0x50}},
    {64801, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0xA, 0x3E, 0x29, 0x20, 0x90, 0x88, 0x70, 0x0, 0x0}},
    {64802, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0xA, 0xA, 0x3D, 0x28, 0x20, 0x10, 0x88, 0x84, 0x78, 0x0, 0x50}},
    {64803, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x4, 0xA, 0xA, 0x3E, 0x29, 0x20, 0x90, 0x88, 0x70, 0x0, 0x0}},
    {64804, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x4, 0xA, 0xA, 0x3E, 0x29, 0x20, 0x90, 0x88, 0x70, 0x0, 0x50}},
    {64805, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x28, 0x0, 0x54, 0xAA, 0x81, 0xFC, 0x40, 0x90, 0x44, 0x38}},
    {64806, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x28, 0x0, 0x54, 0xAA, 0x81, 0xFC, 0x40, 0x80, 0x44, 0x38}},
    {64807, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0xA8, 0x0, 0x54, 0xAA, 0x81, 0xFC, 0x40, 0x80, 0x44, 0x38}},
    {64808, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x14, 0x0, 0x4A, 0xB5, 0x80, 0xC0, 0x70, 0x80, 0x80, 0x80}},
    {64809, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0xA, 0x0, 0x15, 0x2A, 0x20, 0x20, 0x40, 0x80}},
    {64810, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x15, 0x2A, 0x20, 0x20, 0x40, 0x80}},
    {64811, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x5, 0xD, 0x16, 0x24, 0x24, 0x40, 0x80, 0x0}},
    {64812, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x12, 0x5, 0xD, 0x16, 0x24, 0x24, 0x40, 0x80, 0x0}},
    {64819, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x16, 0x19, 0x1E, 0x20, 0xD0, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {64826, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x16, 0x19, 0x1E, 0x22, 0xD1, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {64827, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x16, 0x19, 0x1E, 0x22, 0xD1, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {64828, 1, {0x1, 0x2, 0x4, 0x11, 0x12, 0x14, 0x10, 0x10, 0x10, 0x10, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {64829, 1, {0x1, 0x2, 0x4, 0x11, 0x12, 0x14, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65072, 2, {0x0, 0x0, 0x0, 0x0, 0x180, 0x180, 0x0, 0x0, 0x0, 0x0, 0x180, 0x180, 0x0, 0x0, 0x0, 0x0}},
    {65073, 2, {0x0, 0x0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0, 0x0}},
    {65074, 2, {0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x0, 0x0, 0x0, 0x0}},
    {65077, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E0, 0x1818, 0x2004, 0x4002, 0x0, 0x0}},
    {65078, 2, {0x0, 0x0, 0x4002, 0x2004, 0x1818, 0x7E0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65079, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x3E7C, 0x4002, 0x0, 0x0}},
    {65080, 2, {0x0, 0x0, 0x4002, 0x3E7C, 0x180, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65081, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1FF8, 0x2004, 0x4002, 0x0, 0x0}},
    {65082, 2, {0x0, 0x0, 0x4002, 0x2004, 0x1FF8, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65083, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7FFE, 0x7FFE, 0x781E, 0x6006, 0x4002, 0x0, 0x0}},
    {65084, 2, {0x0, 0x0, 0x4002, 0x6006, 0x781E, 0x7FFE, 0x7FFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65085, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x660, 0x1818, 0x6186, 0x660, 0x1818, 0x6006, 0x0, 0x0}},
    {65086, 2, {0x0, 0x0, 0x6006, 0x1818, 0x660, 0x6186, 0x1818, 0x660, 0x180, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65087, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x660, 0x1818, 0x6006, 0x0, 0x0}},
    {65088, 2, {0x0, 0x0, 0x6006, 0x1818, 0x660, 0x180, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65089, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7FFE, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {65090, 2, {0x0, 0x0, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x7FFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65091, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7FFE, 0x4002, 0x7FFA, 0xA, 0xA, 0xA, 0xE, 0x0, 0x0}},
    {65092, 2, {0x0, 0x0, 0x7000, 0x5000, 0x5000, 0x5000, 0x5FFE, 0x4002, 0x7FFE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65093, 2, {0x0, 0x0, 0x70, 0x38, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65094, 2, {0x0, 0x0, 0x78, 0x24, 0x14, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65097, 2, {0x0, 0x39CE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65098, 2, {0x0, 0x7DBE, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65099, 2, {0xC30, 0x9249, 0x6186, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65100, 2, {0xC30, 0x9249, 0x6186, 0xC30, 0x9249, 0x6186, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65101, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x39CE, 0x0}},
    {65102, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7DBE, 0x0}},
    {65103, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC30, 0x9249, 0x6186}},
    {65104, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x100, 0x200, 0x0}},
    {65105, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x180, 0x0}},
    {65106, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x180, 0x0, 0x0}},
    {65108, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x180, 0x0, 0x180, 0x100, 0x200, 0x0}},
    {65109, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x180, 0x0, 0x180, 0x180, 0x0, 0x0}},
    {65110, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x280, 0x80, 0x100, 0x0, 0x100, 0x0, 0x0}},
    {65111, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x80, 0x80, 0x80, 0x0, 0x80, 0x0, 0x0}},
    {65112, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C0, 0x0, 0x0, 0x0, 0x0}},
    {65113, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x100, 0x100, 0x100, 0x100, 0x80, 0x0, 0x0}},
    {65114, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x80, 0x80, 0x80, 0x80, 0x100, 0x0, 0x0}},
    {65115, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x80, 0x180, 0x80, 0x80, 0x40, 0x0, 0x0}},
    {65116, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x180, 0x100, 0x100, 0x200, 0x0, 0x0}},
    {65117, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C0, 0x100, 0x100, 0x100, 0x100, 0x1C0, 0x0, 0x0}},
    {65118, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x380, 0x80, 0x80, 0x80, 0x80, 0x380, 0x0, 0x0}},
    {65119, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x280, 0x7C0, 0x280, 0x7C0, 0x280, 0x0, 0x0}},
    {65120, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x280, 0x100, 0x280, 0x140, 0x0, 0x0}},
    {65121, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x280, 0x100, 0x380, 0x100, 0x280, 0x0, 0x0}},
    {65122, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x100, 0x100, 0x7C0, 0x100, 0x100, 0x0, 0x0}},
    {65123, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C0, 0x0, 0x0, 0x0, 0x0}},
    {65124, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x80, 0x100, 0x80, 0x40, 0x0, 0x0}},
    {65125, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x80, 0x100, 0x200, 0x0, 0x0}},
    {65126, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C0, 0x0, 0x3C0, 0x0, 0x0, 0x0}},
    {65128, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x100, 0x80, 0x40, 0x0, 0x0, 0x0}},
    {65129, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x380, 0x500, 0x380, 0x140, 0x380, 0x0, 0x0}},
    {65130, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x200, 0x240, 0x80, 0x100, 0x240, 0x40, 0x0, 0x0}},
    {65131, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x180, 0x240, 0x2C0, 0x2C0, 0x200, 0x180, 0x0, 0x0}},
    {65136, 1, {0x0, 0xC, 0x30, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65137, 1, {0x0, 0xC, 0x30, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65138, 1, {0x0, 0x6, 0x6, 0x32, 0x14, 0x38, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65139, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65140, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0xC, 0x30, 0x0, 0x0}},
    {65142, 1, {0x0, 0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65143, 1, {0x0, 0x0, 0x0, 0xC, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65144, 1, {0x0, 0x0, 0x10, 0x28, 0x28, 0x1C, 0x10, 0x20, 0x40, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65145, 1, {0x0, 0x0, 0x10, 0x28, 0x28, 0x1C, 0x10, 0x20, 0x40, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65146, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30}},
    {65147, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0xC, 0x30}},
    {65148, 1, {0x0, 0x0, 0x0, 0x4, 0x14, 0x58, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65149, 1, {0x0, 0x0, 0x0, 0x4, 0x14, 0x58, 0x60, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65150, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65151, 1, {0x0, 0x0, 0x0, 0x18, 0x24, 0x24, 0x18, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65152, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x20, 0x3C, 0x10, 0x20, 0x0, 0x0, 0x0, 0x0}},
    {65153, 1, {0x0, 0x1, 0x3E, 0x40, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65154, 1, {0x0, 0x8, 0x70, 0x80, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65155, 1, {0x18, 0x20, 0x18, 0x20, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65156, 1, {0x30, 0x40, 0x30, 0x40, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65157, 1, {0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65158, 1, {0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65159, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0xC, 0x10, 0xC, 0x10}},
    {65160, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x1F, 0x0, 0x30, 0x40, 0x30, 0x40}},
    {65161, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x40, 0x0, 0x46, 0x48, 0x48, 0x46, 0x42, 0x3C, 0x0}},
    {65162, 1, {0x0, 0x0, 0x0, 0x0, 0x30, 0x40, 0x30, 0x40, 0x0, 0x40, 0x43, 0x42, 0x42, 0x3C, 0x0, 0x0}},
    {65163, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65164, 1, {0x0, 0x0, 0x0, 0xC, 0x10, 0xC, 0x10, 0x0, 0x4, 0x4, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65165, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65166, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x1F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65167, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x8, 0x0, 0x0}},
    {65168, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x0, 0x8, 0x0, 0x0}},
    {65169, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFC, 0x0, 0x0, 0x10, 0x0, 0x0}},
    {65170, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x0, 0x10, 0x0, 0x0}},
    {65171, 1, {0x0, 0x0, 0x0, 0x24, 0x0, 0x0, 0x18, 0x24, 0x22, 0x22, 0x1C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65172, 1, {0x0, 0x0, 0x0, 0x12, 0x0, 0x4, 0xC, 0x14, 0x24, 0x3C, 0x7, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65173, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65174, 1, {0x0, 0x0, 0x0, 0x0, 0x14, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65175, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65176, 1, {0x0, 0x0, 0x0, 0x0, 0x28, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65177, 1, {0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65178, 1, {0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x0, 0x4, 0x22, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65179, 1, {0x0, 0x0, 0x20, 0x0, 0x48, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65180, 1, {0x0, 0x0, 0x20, 0x0, 0x48, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65181, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x48, 0x40, 0x22, 0x1C}},
    {65182, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x14, 0x23, 0x40, 0x48, 0x40, 0x21, 0x1E}},
    {65183, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x6, 0xFE, 0x0, 0x10, 0x0, 0x0, 0x0}},
    {65184, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x6, 0xFF, 0x0, 0x10, 0x0, 0x0, 0x0}},
    {65185, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {65186, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x14, 0x23, 0x40, 0x40, 0x40, 0x21, 0x1E}},
    {65187, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65188, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x70, 0xC, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65189, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x7E, 0x10, 0x20, 0x40, 0x40, 0x40, 0x22, 0x1C}},
    {65190, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x7E, 0x14, 0x23, 0x40, 0x40, 0x40, 0x21, 0x1E}},
    {65191, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0, 0x70, 0xC, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65192, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0, 0x70, 0xC, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65193, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65194, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65195, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65196, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x8, 0x4, 0x2, 0x42, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65197, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65198, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x2, 0x3, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65199, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x4, 0x2, 0x2, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65200, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x4, 0x2, 0x3, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65201, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65202, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5F, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65203, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2A, 0x2A, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65204, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x2A, 0x2A, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65205, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x0, 0x1, 0x15, 0x15, 0x5E, 0x48, 0x48, 0x30, 0x0}},
    {65206, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x12, 0x0, 0x1, 0x15, 0x15, 0x5F, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65207, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x2, 0x2A, 0x2A, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65208, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x24, 0x0, 0x2, 0x2A, 0x2A, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65209, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65210, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x19, 0x4F, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65211, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x52, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65212, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x52, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65213, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x6, 0x19, 0x4E, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65214, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x0, 0x0, 0x6, 0x19, 0x4F, 0x48, 0x48, 0x30, 0x0, 0x0}},
    {65215, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0xC, 0x52, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65216, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0xC, 0x52, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65217, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65218, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65219, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x2C, 0x32, 0x22, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65220, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x2C, 0x32, 0x22, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65221, 1, {0x0, 0x0, 0x0, 0x20, 0x24, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x7C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65222, 1, {0x0, 0x0, 0x0, 0x20, 0x24, 0x20, 0x20, 0x2C, 0x32, 0x22, 0x7F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65223, 1, {0x0, 0x0, 0x0, 0x20, 0x24, 0x20, 0x20, 0x2C, 0x32, 0x22, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65224, 1, {0x0, 0x0, 0x0, 0x20, 0x24, 0x20, 0x20, 0x2C, 0x32, 0x22, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65225, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x20, 0x1C, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {65226, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x30, 0x1F, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {65227, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x20, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65228, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x8, 0xF7, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65229, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x18, 0x24, 0x20, 0x1C, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {65230, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x30, 0x30, 0x1F, 0x20, 0x40, 0x40, 0x42, 0x3C}},
    {65231, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x18, 0x24, 0x20, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65232, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x0, 0x0, 0x18, 0x18, 0x8, 0xF7, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65233, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65234, 1, {0x0, 0x0, 0x0, 0x8, 0x0, 0x6, 0xA, 0x6, 0x42, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65235, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65236, 1, {0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65237, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x0, 0x0, 0x1C, 0x12, 0xA, 0x26, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {65238, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x0, 0x0, 0x1C, 0x12, 0xA, 0x27, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {65239, 1, {0x0, 0x0, 0x0, 0x48, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65240, 1, {0x0, 0x0, 0x0, 0x48, 0x0, 0x0, 0x38, 0x24, 0x14, 0xC, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65241, 1, {0x0, 0x0, 0x0, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65242, 1, {0x0, 0x0, 0x0, 0x2, 0xA, 0x12, 0xA, 0x22, 0x42, 0x42, 0x3F, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65243, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65244, 1, {0x0, 0x0, 0x0, 0x4, 0x8, 0x10, 0x20, 0x40, 0x3C, 0x2, 0xFF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65245, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x22, 0x42, 0x42, 0x44, 0x3C, 0x0, 0x0}},
    {65246, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x23, 0x42, 0x42, 0x44, 0x3C, 0x0, 0x0}},
    {65247, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65248, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65249, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x6, 0x1E, 0x20, 0x20, 0x20, 0x20, 0x20}},
    {65250, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1C, 0x6, 0x1F, 0x20, 0x20, 0x20, 0x20, 0x20}},
    {65251, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x32, 0xEE, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65252, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x14, 0x32, 0xEF, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65253, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x4, 0x22, 0x42, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {65254, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x4, 0x22, 0x43, 0x42, 0x44, 0x38, 0x0, 0x0}},
    {65255, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x8, 0x4, 0x4, 0xF8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65256, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x0, 0x0, 0x8, 0x4, 0x4, 0xFB, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65257, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x18, 0x24, 0x24, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65258, 1, {0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0xC, 0x14, 0x24, 0x3C, 0x7, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65259, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x8, 0x3C, 0x2A, 0x3A, 0xF6, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65260, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x14, 0x14, 0xFF, 0x14, 0xC, 0x0, 0x0, 0x0}},
    {65261, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1E, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65262, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x12, 0x12, 0x1F, 0x2, 0x4, 0x48, 0x30, 0x0}},
    {65263, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {65264, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x43, 0x44, 0x42, 0x42, 0x3C, 0x0}},
    {65265, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4A, 0x48, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {65266, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x43, 0x44, 0x42, 0x3C, 0x0, 0x28}},
    {65267, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xF8, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {65268, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x4, 0xFB, 0x0, 0x28, 0x0, 0x0, 0x0}},
    {65269, 1, {0x0, 0x0, 0x8, 0x72, 0x82, 0x22, 0x12, 0xA, 0x4, 0xC, 0x14, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {65270, 1, {0x0, 0x0, 0x8, 0x72, 0x82, 0x22, 0x12, 0xA, 0xE, 0xA, 0x13, 0x60, 0x0, 0x0, 0x0, 0x0}},
    {65271, 1, {0x0, 0x0, 0x30, 0x42, 0x32, 0x42, 0x12, 0xA, 0x4, 0xC, 0x14, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {65272, 1, {0x0, 0x0, 0x30, 0x42, 0x32, 0x42, 0x12, 0xA, 0xE, 0xA, 0x13, 0x60, 0x0, 0x0, 0x0, 0x0}},
    {65273, 1, {0x0, 0x0, 0x2, 0x2, 0x22, 0x12, 0xA, 0x4, 0xC, 0x14, 0x1C, 0x0, 0xC, 0x10, 0xC, 0x10}},
    {65274, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x22, 0x12, 0xA, 0xE, 0xA, 0x13, 0x60, 0xC, 0x10, 0xC, 0x10}},
    {65275, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x22, 0x12, 0xA, 0x4, 0xC, 0x14, 0x1C, 0x0, 0x0, 0x0, 0x0}},
    {65276, 1, {0x0, 0x0, 0x0, 0x2, 0x2, 0x22, 0x12, 0xA, 0xE, 0xA, 0x13, 0x60, 0x0, 0x0, 0x0, 0x0}},
    {65281, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0xC0, 0xC0, 0x0, 0x0}},
    {65282, 2, {0x0, 0x0, 0xC0C, 0xC0C, 0xC0C, 0xC0C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65283, 2, {0x0, 0x0, 0x0, 0x0, 0x30C, 0x30C, 0x30C, 0x3FFC, 0xC30, 0xC30, 0x3FFC, 0x30C0, 0x30C0, 0x30C0, 0x0, 0x0}},
    {65284, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xFFC, 0x30C3, 0x30C0, 0xFC0, 0xFC, 0xC3, 0x30C3, 0xFFC, 0xC0, 0x0, 0x0}},
    {65285, 2, {0x0, 0x0, 0x0, 0x0, 0xF03, 0x30CC, 0x30CC, 0xF30, 0xC0, 0xC0, 0x33C, 0xCC3, 0xCC3, 0x303C, 0x0, 0x0}},
    {65286, 2, {0x0, 0x0, 0x0, 0x0, 0x3F0, 0xC0C, 0xC0C, 0xC0C, 0x3F0, 0xFC3, 0x3033, 0x300C, 0x303C, 0xFC3, 0x0, 0x0}},
    {65287, 2, {0x0, 0x0, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65288, 2, {0x0, 0x0, 0x0, 0x30, 0xC0, 0xC0, 0x300, 0x300, 0x300, 0x300, 0x300, 0x300, 0xC0, 0xC0, 0x30, 0x0}},
    {65289, 2, {0x0, 0x0, 0x0, 0xC00, 0x300, 0x300, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0x300, 0x300, 0xC00, 0x0}},
    {65290, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC0, 0x30C3, 0xCCC, 0x3F0, 0xCCC, 0x30C3, 0xC0, 0x0, 0x0, 0x0}},
    {65291, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xC0, 0xC0, 0xC0, 0x3FFF, 0xC0, 0xC0, 0xC0, 0x0, 0x0, 0x0}},
    {65292, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C0, 0xC0, 0xC0, 0x300}},
    {65293, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3FFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65294, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C0, 0x3C0, 0x0, 0x0}},
    {65295, 2, {0x0, 0x0, 0x0, 0x0, 0xC, 0xC, 0x30, 0xC0, 0xC0, 0x300, 0x300, 0xC00, 0x3000, 0x3000, 0x0, 0x0}},
    {65296, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0xC30, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0xC30, 0x3C0, 0x0, 0x0}},
    {65297, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0x3C0, 0xCC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xFFC, 0x0, 0x0}},
    {65298, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0xC, 0xF0, 0x300, 0xC00, 0x3000, 0x3000, 0x3FFC, 0x0, 0x0}},
    {65299, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0xC, 0x3F0, 0xC, 0xC, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65300, 2, {0x0, 0x0, 0x0, 0x0, 0x30, 0xF0, 0x330, 0xC30, 0x3030, 0x3030, 0x3FFC, 0x30, 0x30, 0x30, 0x0, 0x0}},
    {65301, 2, {0x0, 0x0, 0x0, 0x0, 0x3FFC, 0x3000, 0x3000, 0x3000, 0x3FF0, 0xC, 0xC, 0xC, 0x300C, 0xFF0, 0x0, 0x0}},
    {65302, 2, {0x0, 0x0, 0x0, 0x0, 0x3F0, 0xC00, 0x3000, 0x3000, 0x3FF0, 0x300C, 0x300C, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65303, 2, {0x0, 0x0, 0x0, 0x0, 0x3FFC, 0xC, 0xC, 0x30, 0x30, 0x30, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0x0}},
    {65304, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x300C, 0xFF0, 0x300C, 0x300C, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65305, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x300C, 0xFFC, 0xC, 0xC, 0xC, 0x30, 0xFC0, 0x0, 0x0}},
    {65306, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C0, 0x3C0, 0x0, 0x0, 0x0, 0x3C0, 0x3C0, 0x0, 0x0, 0x0}},
    {65307, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C0, 0x3C0, 0x0, 0x0, 0x0, 0x3C0, 0xC0, 0xC0, 0x300, 0x0}},
    {65308, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0x30, 0xC0, 0x300, 0xC00, 0x300, 0xC0, 0x30, 0xC, 0x0, 0x0}},
    {65309, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3FFC, 0x0, 0x0, 0x0, 0x3FFC, 0x0, 0x0, 0x0, 0x0}},
    {65310, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x3000, 0xC00, 0x300, 0xC0, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x0, 0x0}},
    {65311, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0xC, 0x30, 0xC0, 0xC0, 0x0, 0xC0, 0xC0, 0x0, 0x0}},
    {65312, 2, {0x0, 0x0, 0x0, 0x0, 0x3F0, 0xC0C, 0x30CC, 0x333C, 0x330C, 0x330C, 0x330C, 0x30FC, 0xC00, 0x3FC, 0x0, 0x0}},
    {65313, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0xC30, 0xC30, 0x300C, 0x300C, 0x3FFC, 0x300C, 0x300C, 0x300C, 0x300C, 0x0, 0x0}},
    {65314, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x300C, 0x300C, 0x300C, 0x3FF0, 0x300C, 0x300C, 0x300C, 0x300C, 0x3FF0, 0x0, 0x0}},
    {65315, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x3000, 0x3000, 0x3000, 0x3000, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65316, 2, {0x0, 0x0, 0x0, 0x0, 0x3FC0, 0x3030, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x3030, 0x3FC0, 0x0, 0x0}},
    {65317, 2, {0x0, 0x0, 0x0, 0x0, 0x3FFC, 0x3000, 0x3000, 0x3000, 0x3FF0, 0x3000, 0x3000, 0x3000, 0x3000, 0x3FFC, 0x0, 0x0}},
    {65318, 2, {0x0, 0x0, 0x0, 0x0, 0x3FFC, 0x3000, 0x3000, 0x3000, 0x3FF0, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x0, 0x0}},
    {65319, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x3000, 0x3000, 0x30FC, 0x300C, 0x300C, 0x303C, 0xFCC, 0x0, 0x0}},
    {65320, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x300C, 0x300C, 0x3FFC, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x0, 0x0}},
    {65321, 2, {0x0, 0x0, 0x0, 0x0, 0xFFC, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xFFC, 0x0, 0x0}},
    {65322, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3030, 0x3030, 0xFC0, 0x0, 0x0}},
    {65323, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x3030, 0x30C0, 0x3300, 0x3C00, 0x3C00, 0x3300, 0x30C0, 0x3030, 0x300C, 0x0, 0x0}},
    {65324, 2, {0x0, 0x0, 0x0, 0x0, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x3FFC, 0x0, 0x0}},
    {65325, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x3C3C, 0x3C3C, 0x33CC, 0x33CC, 0x300C, 0x300C, 0x300C, 0x300C, 0x0, 0x0}},
    {65326, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x3C0C, 0x3C0C, 0x330C, 0x330C, 0x30CC, 0x30CC, 0x303C, 0x303C, 0x300C, 0x0, 0x0}},
    {65327, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65328, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x300C, 0x300C, 0x300C, 0x3FF0, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x0, 0x0}},
    {65329, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x33CC, 0x3C3C, 0xFF0, 0xF, 0x0}},
    {65330, 2, {0x0, 0x0, 0x0, 0x0, 0x3FF0, 0x300C, 0x300C, 0x300C, 0x3FF0, 0x30C0, 0x3030, 0x3030, 0x300C, 0x300C, 0x0, 0x0}},
    {65331, 2, {0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x3000, 0xF00, 0xF0, 0xC, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65332, 2, {0x0, 0x0, 0x0, 0x0, 0x3FFF, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0x0}},
    {65333, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65334, 2, {0x0, 0x0, 0x0, 0x0, 0x3003, 0x3003, 0x3003, 0xC0C, 0xC0C, 0xC0C, 0x330, 0x330, 0xC0, 0xC0, 0x0, 0x0}},
    {65335, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x300C, 0x300C, 0x33CC, 0x33CC, 0x3C3C, 0x3C3C, 0x300C, 0x300C, 0x0, 0x0}},
    {65336, 2, {0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0xC30, 0xC30, 0x3C0, 0x3C0, 0xC30, 0xC30, 0x300C, 0x300C, 0x0, 0x0}},
    {65337, 2, {0x0, 0x0, 0x0, 0x0, 0x3003, 0x3003, 0xC0C, 0xC0C, 0x330, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0x0}},
    {65338, 2, {0x0, 0x0, 0x0, 0x0, 0x3FFC, 0xC, 0xC, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x3000, 0x3FFC, 0x0, 0x0}},
    {65339, 2, {0x0, 0x0, 0x0, 0xFC, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xFC, 0x0}},
    {65340, 2, {0x0, 0x0, 0x0, 0x0, 0x3000, 0x3000, 0xC00, 0x300, 0x300, 0xC0, 0xC0, 0x30, 0xC, 0xC, 0x0, 0x0}},
    {65342, 2, {0x0, 0x0, 0x3C0, 0xC30, 0x300C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65343, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3FFF, 0x0}},
    {65344, 2, {0x0, 0xC00, 0x300, 0xC0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65345, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0xC, 0xFFC, 0x300C, 0x300C, 0x303C, 0xFCC, 0x0, 0x0}},
    {65346, 2, {0x0, 0x0, 0x0, 0x3000, 0x3000, 0x3000, 0x33F0, 0x3C0C, 0x300C, 0x300C, 0x300C, 0x300C, 0x3C0C, 0x33F0, 0x0, 0x0}},
    {65347, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x3000, 0x3000, 0x3000, 0x3000, 0x300C, 0xFF0, 0x0, 0x0}},
    {65348, 2, {0x0, 0x0, 0x0, 0xC, 0xC, 0xC, 0xFCC, 0x303C, 0x300C, 0x300C, 0x300C, 0x300C, 0x303C, 0xFCC, 0x0, 0x0}},
    {65349, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x3FFC, 0x3000, 0x3000, 0x300C, 0xFF0, 0x0, 0x0}},
    {65350, 2, {0x0, 0x0, 0x0, 0xF0, 0x300, 0x300, 0x300, 0x3FF0, 0x300, 0x300, 0x300, 0x300, 0x300, 0x300, 0x0, 0x0}},
    {65351, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0xC, 0xFCC, 0x3030, 0x3030, 0x3030, 0xFC0, 0xC00, 0xFF0, 0x300C, 0x300C, 0xFF0}},
    {65352, 2, {0x0, 0x0, 0x0, 0x3000, 0x3000, 0x3000, 0x33F0, 0x3C0C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x0, 0x0}},
    {65353, 2, {0x0, 0x0, 0x0, 0xC0, 0xC0, 0x0, 0x3C0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xFFC, 0x0, 0x0}},
    {65354, 2, {0x0, 0x0, 0x0, 0x30, 0x30, 0x0, 0xF0, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30C0, 0xF00}},
    {65355, 2, {0x0, 0x0, 0x0, 0x0, 0x3000, 0x3000, 0x3030, 0x30C0, 0x3300, 0x3C00, 0x3300, 0x30C0, 0x3030, 0x300C, 0x0, 0x0}},
    {65356, 2, {0x0, 0x0, 0x0, 0x0, 0x3C0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xFFC, 0x0, 0x0}},
    {65357, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3F3C, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0x0, 0x0}},
    {65358, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x33F0, 0x3C0C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x0, 0x0}},
    {65359, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0xFF0, 0x0, 0x0}},
    {65360, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x33F0, 0x3C0C, 0x300C, 0x300C, 0x300C, 0x300C, 0x3C0C, 0x33F0, 0x3000, 0x3000}},
    {65361, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFCC, 0x303C, 0x300C, 0x300C, 0x300C, 0x300C, 0x303C, 0xFCC, 0xC, 0xC}},
    {65362, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x33F0, 0x3C0C, 0x300C, 0x3000, 0x3000, 0x3000, 0x3000, 0x3000, 0x0, 0x0}},
    {65363, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF0, 0x300C, 0x3000, 0xF00, 0xF0, 0xC, 0x300C, 0xFF0, 0x0, 0x0}},
    {65364, 2, {0x0, 0x0, 0x0, 0x0, 0x300, 0x300, 0x3FF0, 0x300, 0x300, 0x300, 0x300, 0x300, 0x300, 0xF0, 0x0, 0x0}},
    {65365, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0x303C, 0xFCC, 0x0, 0x0}},
    {65366, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x300C, 0xC30, 0xC30, 0xC30, 0x3C0, 0x3C0, 0x0, 0x0}},
    {65367, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3003, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0x30C3, 0xF3C, 0x0, 0x0}},
    {65368, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0xC30, 0x3C0, 0x3C0, 0xC30, 0x300C, 0x300C, 0x0, 0x0}},
    {65369, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x300C, 0x300C, 0x300C, 0x300C, 0x300C, 0xC3C, 0x3CC, 0xC, 0xC, 0xFF0}},
    {65370, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3FFC, 0xC, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0x3FFC, 0x0, 0x0}},
    {65371, 2, {0x0, 0x0, 0x0, 0xF0, 0x300, 0x300, 0xC0, 0xC0, 0x300, 0x300, 0xC0, 0xC0, 0x300, 0x300, 0xF0, 0x0}},
    {65372, 2, {0x0, 0x0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0}},
    {65373, 2, {0x0, 0x0, 0x0, 0xF00, 0xC0, 0xC0, 0x300, 0x300, 0xC0, 0xC0, 0x300, 0x300, 0xC0, 0xC0, 0xF00, 0x0}},
    {65374, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF8, 0x8304, 0x7C00, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65375, 2, {0x0, 0x0, 0x0, 0x186, 0x618, 0x618, 0x1860, 0x1860, 0x1860, 0x1860, 0x1860, 0x1860, 0x618, 0x618, 0x186, 0x0}},
    {65376, 2, {0x0, 0x0, 0x0, 0x6180, 0x1860, 0x1860, 0x618, 0x618, 0x618, 0x618, 0x618, 0x618, 0x1860, 0x1860, 0x6180, 0x0}},
    {65377, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x50, 0x50, 0x20, 0x0}},
    {65378, 1, {0x0, 0x1E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0, 0x0}},
    {65379, 1, {0x0, 0x0, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x3C, 0x0}},
    {65380, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x60, 0x30, 0x18, 0x0, 0x0}},
    {65381, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65382, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x7E, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {65383, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x12, 0x14, 0x10, 0x10, 0x10, 0x20, 0x0}},
    {65384, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x4, 0x8, 0x18, 0x68, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {65385, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7E, 0x42, 0x42, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0}},
    {65386, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7E, 0x0, 0x0}},
    {65387, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x7E, 0x18, 0x18, 0x28, 0x48, 0x8, 0x18, 0x0}},
    {65388, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x22, 0x16, 0x1A, 0x32, 0x48, 0x8, 0x8, 0x4, 0x4, 0x0}},
    {65389, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x8, 0x8, 0x8, 0x8, 0x8, 0x7E, 0x0, 0x0}},
    {65390, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x4, 0x4, 0x3C, 0x4, 0x4, 0x3C, 0x0, 0x0}},
    {65391, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x52, 0x4A, 0x22, 0x4, 0x4, 0x8, 0x10, 0x20, 0x0}},
    {65392, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65393, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x12, 0x14, 0x14, 0x10, 0x10, 0x10, 0x20, 0x20, 0x40, 0x0}},
    {65394, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x4, 0x4, 0x8, 0x18, 0x28, 0xC8, 0x8, 0x8, 0x8, 0x8, 0x0}},
    {65395, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x7E, 0x42, 0x42, 0x42, 0x2, 0x4, 0x4, 0x8, 0x10, 0x20, 0x0}},
    {65396, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xFF, 0x0, 0x0}},
    {65397, 1, {0x0, 0x0, 0x8, 0x8, 0x8, 0xFF, 0x8, 0x18, 0x18, 0x28, 0x48, 0x88, 0x28, 0x18, 0x0, 0x0}},
    {65398, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x7E, 0x12, 0x12, 0x12, 0x22, 0x22, 0x22, 0x42, 0x4E, 0x0, 0x0}},
    {65399, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x16, 0x18, 0x30, 0xD0, 0x13, 0xC, 0x38, 0x48, 0x4, 0x4, 0x0}},
    {65400, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x1E, 0x12, 0x22, 0x22, 0x4, 0x4, 0x8, 0x8, 0x10, 0x20, 0x0}},
    {65401, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x3F, 0x44, 0x44, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x0}},
    {65402, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {65403, 1, {0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0xFF, 0x24, 0x24, 0x24, 0x4, 0x8, 0x8, 0x10, 0x20, 0x0}},
    {65404, 1, {0x0, 0x0, 0x0, 0x20, 0x20, 0x10, 0x42, 0x42, 0x22, 0x4, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0}},
    {65405, 1, {0x0, 0x0, 0x0, 0x0, 0x7C, 0x4, 0x4, 0x8, 0x8, 0x18, 0x14, 0x22, 0x42, 0x81, 0x0, 0x0}},
    {65406, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x2E, 0x32, 0xE2, 0x24, 0x20, 0x20, 0x20, 0x3E, 0x0, 0x0}},
    {65407, 1, {0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x22, 0x22, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x20, 0x0}},
    {65408, 1, {0x0, 0x0, 0x10, 0x10, 0x10, 0x1E, 0x12, 0x22, 0x32, 0x4A, 0x4, 0x8, 0x8, 0x10, 0x20, 0x0}},
    {65409, 1, {0x0, 0x0, 0x0, 0x2, 0x4, 0x78, 0x8, 0x8, 0xFF, 0x8, 0x8, 0x10, 0x10, 0x20, 0x40, 0x0}},
    {65410, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x52, 0x4A, 0x22, 0x4, 0x4, 0x8, 0x10, 0x10, 0x20, 0x40, 0x0}},
    {65411, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0xFF, 0x8, 0x8, 0x10, 0x10, 0x10, 0x20, 0x40, 0x0}},
    {65412, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x18, 0x14, 0x12, 0x10, 0x10, 0x10, 0x10, 0x0}},
    {65413, 1, {0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0xFF, 0x8, 0x8, 0x10, 0x10, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {65414, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x0, 0x0, 0x0}},
    {65415, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x44, 0x28, 0x18, 0x8, 0x14, 0x12, 0x20, 0x20, 0x0}},
    {65416, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x7E, 0x2, 0x4, 0x8, 0x18, 0x34, 0x52, 0x11, 0x10, 0x10}},
    {65417, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {65418, 1, {0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x22, 0x22, 0x22, 0x42, 0x41, 0x41, 0x81, 0x0, 0x0}},
    {65419, 1, {0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x46, 0x48, 0x70, 0x40, 0x40, 0x40, 0x20, 0x1E, 0x0, 0x0}},
    {65420, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {65421, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x30, 0x30, 0x48, 0x44, 0x84, 0x4, 0x2, 0x0, 0x0, 0x0, 0x0}},
    {65422, 1, {0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0xFF, 0x10, 0x10, 0x54, 0x52, 0x91, 0x10, 0x30, 0x0, 0x0}},
    {65423, 1, {0x0, 0x0, 0x0, 0x0, 0xFF, 0x1, 0x2, 0x4, 0x28, 0x10, 0x18, 0x8, 0x4, 0x4, 0x0, 0x0}},
    {65424, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x18, 0x4, 0x20, 0x10, 0xC, 0x20, 0x10, 0x8, 0x4, 0x0, 0x0}},
    {65425, 1, {0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x20, 0x24, 0x24, 0x22, 0x4E, 0x72, 0x81, 0x0, 0x0}},
    {65426, 1, {0x0, 0x0, 0x0, 0x0, 0x2, 0x2, 0x24, 0x24, 0x18, 0x8, 0x1C, 0x14, 0x22, 0x20, 0x40, 0x0}},
    {65427, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x10, 0x10, 0x10, 0xFE, 0x10, 0x10, 0x10, 0x10, 0x1E, 0x0, 0x0}},
    {65428, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x23, 0x2D, 0x31, 0xD2, 0x10, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {65429, 1, {0x0, 0x0, 0x0, 0x0, 0x78, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x7E, 0x0, 0x0}},
    {65430, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x7E, 0x0, 0x0}},
    {65431, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x4, 0x4, 0x8, 0x10, 0x20, 0x0}},
    {65432, 1, {0x0, 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x22, 0x2, 0x4, 0x4, 0x4, 0x8, 0x10, 0x0}},
    {65433, 1, {0x0, 0x0, 0x0, 0x0, 0x8, 0x28, 0x28, 0x29, 0x29, 0x29, 0x2A, 0x4A, 0x4C, 0x88, 0x0, 0x0}},
    {65434, 1, {0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x20, 0x20, 0x22, 0x22, 0x24, 0x24, 0x28, 0x30, 0x0, 0x0}},
    {65435, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, 0x0, 0x0, 0x0}},
    {65436, 1, {0x0, 0x0, 0x0, 0x0, 0x7E, 0x42, 0x42, 0x42, 0x2, 0x4, 0x4, 0x8, 0x10, 0x20, 0x0, 0x0}},
    {65437, 1, {0x0, 0x0, 0x0, 0x40, 0x20, 0x12, 0x2, 0x2, 0x4, 0x4, 0x8, 0x10, 0x20, 0x40, 0x0, 0x0}},
    {65438, 1, {0x0, 0x28, 0x28, 0x28, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65439, 1, {0x0, 0x10, 0x28, 0x28, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65440, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65441, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x2, 0x2, 0x2, 0x0, 0x0}},
    {65442, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x24, 0x24, 0x24, 0x24, 0x0, 0x0}},
    {65443, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x74, 0x14, 0x14, 0x1A, 0x11, 0x0, 0x0, 0x0}},
    {65444, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {65445, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x2A, 0x39, 0x0, 0x0}},
    {65446, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x22, 0x27, 0x22, 0x25, 0x3A, 0x0, 0x0}},
    {65447, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x40, 0x40, 0x7E, 0x0, 0x0}},
    {65448, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6C, 0x48, 0x48, 0x48, 0x6C, 0x0, 0x0}},
    {65449, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3C, 0x4, 0x3C, 0x20, 0x3C, 0x0, 0x0}},
    {65450, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x76, 0x12, 0x72, 0x42, 0x72, 0x0, 0x0}},
    {65451, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xEE, 0x2A, 0xEA, 0x8A, 0xEE, 0x0, 0x0}},
    {65452, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xEA, 0x2A, 0xEE, 0x8A, 0xEE, 0x0, 0x0}},
    {65453, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xE4, 0x24, 0xE4, 0x8A, 0xEA, 0x0, 0x0}},
    {65454, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xEE, 0x28, 0xEE, 0x88, 0xEE, 0x0, 0x0}},
    {65455, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFF, 0x2A, 0xEA, 0x8A, 0xFF, 0x0, 0x0}},
    {65456, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xE4, 0x2E, 0xE4, 0x8A, 0xE4, 0x0, 0x0}},
    {65457, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x44, 0x44, 0x44, 0x7C, 0x0, 0x0}},
    {65458, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x42, 0x42, 0x7E, 0x42, 0x7E, 0x0, 0x0}},
    {65459, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xAA, 0xAA, 0xEE, 0xAA, 0xEE, 0x0, 0x0}},
    {65460, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x54, 0x74, 0x5A, 0x71, 0x0, 0x0}},
    {65461, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x28, 0x44, 0x0, 0x0}},
    {65462, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x24, 0x24, 0x24, 0x5A, 0x91, 0x0, 0x0}},
    {65463, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x38, 0x44, 0x44, 0x44, 0x38, 0x0, 0x0}},
    {65464, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x28, 0x44, 0x0, 0x0}},
    {65465, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x24, 0x5A, 0x91, 0x0, 0x0}},
    {65466, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0x0, 0x7C, 0x10, 0x28, 0x44, 0x0, 0x0}},
    {65467, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x2, 0x7E, 0x2, 0x2, 0x0, 0x0}},
    {65468, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x40, 0x7E, 0x40, 0x7E, 0x0, 0x0}},
    {65469, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x24, 0x24, 0x24, 0x7E, 0x0, 0x0}},
    {65470, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x38, 0xFE, 0x38, 0x44, 0x38, 0x0, 0x0}},
    {65474, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x20, 0x38, 0x20, 0x20, 0x0, 0x0}},
    {65475, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x28, 0x38, 0x28, 0x28, 0x0, 0x0}},
    {65476, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x38, 0x20, 0x38, 0x20, 0x0, 0x0}},
    {65477, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x38, 0x28, 0x38, 0x28, 0x0, 0x0}},
    {65478, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x38, 0x8, 0x8, 0x0, 0x0}},
    {65479, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x14, 0x74, 0x14, 0x14, 0x0, 0x0}},
    {65482, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x38, 0x8, 0x38, 0x8, 0x0, 0x0}},
    {65483, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x14, 0x74, 0x14, 0x74, 0x14, 0x0, 0x0}},
    {65484, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x0, 0x0}},
    {65485, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x14, 0x14, 0x7E, 0x0, 0x0}},
    {65486, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xA, 0xA, 0x2A, 0x2A, 0xFE, 0x0, 0x0}},
    {65487, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x14, 0x14, 0x7C, 0x0, 0x0}},
    {65490, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x28, 0x28, 0x28, 0x28, 0x7C, 0x0, 0x0}},
    {65491, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x0, 0x0}},
    {65492, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x3C, 0x14, 0x1C, 0x14, 0x14, 0x0, 0x0}},
    {65493, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xA, 0xA, 0x7A, 0x2A, 0x3A, 0x2A, 0x2A, 0x0, 0x0}},
    {65494, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x4, 0x4, 0x3C, 0x14, 0x14, 0x14, 0x14, 0x0, 0x0}},
    {65495, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7C, 0x28, 0x28, 0x28, 0x28, 0x0, 0x0}},
    {65498, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x0, 0x0, 0x0, 0x0}},
    {65499, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x18, 0x28, 0x8, 0x0, 0x0}},
    {65500, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x0}},
    {65504, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xC0, 0xFFC, 0x30C3, 0x30C0, 0x30C0, 0x30C3, 0xFFC, 0xC0, 0xC0, 0x0, 0x0}},
    {65505, 2, {0x0, 0x0, 0x0, 0x0, 0xFC, 0x300, 0x300, 0x300, 0x3FF0, 0x300, 0x300, 0x300, 0xFFC, 0x3C03, 0x0, 0x0}},
    {65506, 2, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x3FFC, 0xC, 0xC, 0xC, 0x0, 0x0}},
    {65507, 2, {0x0, 0x0, 0x3FFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65508, 2, {0x0, 0x0, 0x0, 0x0, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0x0, 0xC0, 0xC0, 0xC0, 0xC0, 0x0, 0x0}},
    {65509, 2, {0x0, 0x0, 0x0, 0x0, 0x3003, 0xC0C, 0x330, 0xC0, 0x3FFF, 0xC0, 0x3FFF, 0xC0, 0xC0, 0xC0, 0x0, 0x0}},
    {65510, 2, {0x0, 0x0, 0x0, 0x0, 0xC0C, 0xC0C, 0xC0C, 0x3FFF, 0xCCC, 0x3FFF, 0xF3C, 0xF3C, 0xC0C, 0xC0C, 0x0, 0x0}},
    {65512, 1, {0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8}},
    {65513, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x10, 0x20, 0x7E, 0x20, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65514, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x1C, 0x2A, 0x8, 0x8, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65515, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x4, 0x7E, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65516, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x8, 0x8, 0x8, 0x2A, 0x1C, 0x8, 0x0, 0x0, 0x0, 0x0, 0x0}},
    {65517, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x0, 0x0, 0x0}},
    {65518, 1, {0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x24, 0x42, 0x42, 0x24, 0x18, 0x0, 0x0, 0x0, 0x0}},
    {65533, 1, {0x0, 0x0, 0x0, 0x7E, 0x66, 0x5A, 0x5A, 0x7A, 0x76, 0x76, 0x7E, 0x76, 0x76, 0x7E, 0x0, 0x0}},
};
